/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`timescale 1ns / 1ps
`default_nettype none



//==============================================================================
// 									DFF
//==============================================================================

module DFF
#(
	parameter BIT_WIDTH = 1	// must be bigger than 0
)
(
	input  wire clk,
	input  wire rst,
	input  wire [BIT_WIDTH-1:0] dataIn,
	output reg  [BIT_WIDTH-1:0] dataOut
);

always@(posedge clk)
begin

	if(rst)
		dataOut <= 0;
	else
		dataOut <= dataIn;

end

endmodule


//==============================================================================
// 								SHIFT_DFF
//==============================================================================

module SHIFT_DFF
#(
	parameter BIT_WIDTH = 1, // must be bigger than 0
	parameter LATENCY = 1
)
(
	input	wire clk,
	input	wire rst,
	input	wire [BIT_WIDTH-1:0] dataIn,
	output	wire [BIT_WIDTH-1:0] dataOut
);

wire [BIT_WIDTH-1:0] connection [LATENCY:0];

// conntects input and output, for LATENCY=0 its a direct connection
assign dataOut = connection[LATENCY];
assign connection[0] = dataIn;


genvar i;
generate
for (i=1; i<=LATENCY;i=i+1)
	begin:DFF_INIT
	
		DFF #(.BIT_WIDTH(BIT_WIDTH)) def (clk, rst, connection[i-1], connection[i]);
	
	end
endgenerate

endmodule



//==============================================================================
// 								MEM
//==============================================================================
// Dual-Port Block RAM with Two Write Ports, Write First!
// File: HDL_Coding_Techniques/rams/rams_16.v
// https://www.xilinx.com/support/documentation/sw_manuals/xilinx2014_1/ug901-vivado-synthesis.pdf
// ! dual clock removed !

module MEM
(
	clk,	
	
	ena,
	wea,
	addra,
	dia,
	doa,
	
	enb,
	web,
	addrb,
	dib,
	dob
	
);

function integer clog2;
input integer value;
begin
value = value-1;
for (clog2=0; value>0; clog2=clog2+1)
value = value>>1;
end
endfunction

	parameter REG_BWIDTH = 16;
	parameter REG_NUMBER = 1024;
	parameter ADDR_BWIDTH	= clog2(REG_NUMBER);

	input  wire clk;
	
	input  wire 							ena;
	input  wire								wea;
	input  wire [ADDR_BWIDTH-1:0]			addra;
	input  wire [REG_BWIDTH-1:0]			dia;
	output reg  [REG_BWIDTH-1:0]			doa;
	
	input  wire 							enb;
	input  wire								web;
	input  wire [ADDR_BWIDTH-1:0]			addrb;
	input  wire [REG_BWIDTH-1:0]			dib;
	output reg  [REG_BWIDTH-1:0]			dob;

reg[REG_BWIDTH-1:0] mem [REG_NUMBER-1:0];

always @(posedge clk) begin if (ena)
begin
if (wea)
begin
	mem[addra] <= dia;
	doa <= dia;
end
else
begin
	doa <= mem[addra];
end	
end
end

always @(posedge clk) begin if (enb)
begin
if (web)
begin
		mem[addrb] <= dib;
		dob <= dib;
end
else
begin
		dob <= mem[addrb];
end	
end
end

endmodule

//==============================================================================
// 								MEM_RESETTER
//==============================================================================

module MEM_RESETTER
#(
	parameter REG_BWIDTH = 16,
	parameter REG_NUMBER = 1024,
	parameter RESET_VALUE = 0
)
(
	clk,	
	rst,
	busyOut,	
	wen,	
	addra,
	di
	
);

function integer clog2;
input integer value;
begin
value = value-1;
for (clog2=0; value>0; clog2=clog2+1)
value = value>>1;
end
endfunction

	parameter ADDR_BWIDTH	= clog2(REG_NUMBER);

	input  wire clk;	
	input  wire rst;
	output reg busyOut;
	output reg wen;
	output reg [ADDR_BWIDTH-1:0] addra;
	output reg [REG_BWIDTH-1:0] di;

reg resettingMode;

always@(posedge clk)
begin

	di	<= RESET_VALUE;

	if(rst)
	begin
	
		resettingMode	<= 1;
		addra			<= 0;
		wen				<= 1;
		busyOut			<= 1;
		
	end
	else
	begin
		
		if(resettingMode)
		begin
			
			if(addra == REG_NUMBER-1)
			begin

				resettingMode	<= 0;
				busyOut		 	<= 0;
				wen				<= 0;
			end
			else
			begin
			
				addra	<= addra+1;
			
			end
		end
	end

end


endmodule



//==============================================================================
// 								MEM_RSTABLE
//==============================================================================

// whenever you get an rst as input, the memory will stall(busyOut) and will
// reset all memeory entries to RESET_VALUE. when finished, busyOut will become
// zero again

module MEM_RSTABLE
#(
	parameter REG_BWIDTH	= 16,
	parameter REG_NUMBER	= 1024,
	parameter RESET_VALUE	= 0
)
(
	clk,	
	rst,
	busyOut,
	ena,
	wena,
	addra,
	dia,
	doa,
	enb,
	wenb,
	addrb,
	dib,
	dob
);

function integer clog2;
input integer value;
begin
value = value-1;
for (clog2=0; value>0; clog2=clog2+1)
value = value>>1;
end
endfunction

	parameter ADDR_BWIDTH	= clog2(REG_NUMBER);

	input  wire clk;	
	
	input  wire rst;
	output wire busyOut;
	
	input  wire 							ena;
	input  wire								wena;
	input  wire [ADDR_BWIDTH-1:0]			addra;
	input  wire [REG_BWIDTH-1:0]			dia;
	output wire [REG_BWIDTH-1:0]			doa;
	
	input  wire 							enb;
	input  wire								wenb;
	input  wire [ADDR_BWIDTH-1:0]			addrb;
	input  wire [REG_BWIDTH-1:0]			dib;
	output wire [REG_BWIDTH-1:0]			dob;

wire rstWen, memWena, memEna, memEnb;
wire [clog2(REG_NUMBER)-1:0] rstAddra;
reg [clog2(REG_NUMBER)-1:0] memAddra;

wire [REG_BWIDTH-1:0] memDib, rstDi;
reg  [REG_BWIDTH-1:0] memDia;

MEM_RESETTER #(.REG_BWIDTH(REG_BWIDTH), .REG_NUMBER(REG_NUMBER))
mem_resetter  (.clk(clk), .rst(rst), .busyOut(busyOut), .wen(rstWen), .addra(rstAddra), .di(rstDi));


MEM #(.REG_BWIDTH(REG_BWIDTH), .REG_NUMBER(REG_NUMBER))
mem  ( .clk(clk),	.ena(memEna),	.wea(memWena), .addra(memAddra),  .dia(memDia), .doa(doa),
					.enb(memEnb),		.web(wenb),	.addrb(addrb), 		.dib(dib), 	.dob(dob));


assign memEna	= ena	|| rstWen;
assign memWena	= wena	|| rstWen;
assign memEnb	= enb	&& !rstWen;


always@(*)
begin
	
	if(rstWen)
	begin	
		memAddra	= rstAddra;
		memDia		= rstDi;
	end
	else
	begin
		memAddra	= addra;
		memDia		= dia;
	end

end






endmodule

//==============================================================================
// 							MEM_RSTABLE_RWPORT
//==============================================================================

// memory with only one read and one write port, with write first behaviour
//(multiplexed outputs)

module MEM_RSTABLE_BYPASS
#(
	parameter REG_BWIDTH	= 16,
	parameter REG_NUMBER	= 1024,
	parameter RESET_VALUE	= 0
)
(
	clk,	
	rst,
	busyOut,
	addra,
	wenb,
	addrb,
	dib,
	dout	
);

function integer clog2;
input integer value;
begin
value = value-1;
for (clog2=0; value>0; clog2=clog2+1)
value = value>>1;
end
endfunction

	parameter ADDR_BWIDTH	= clog2(REG_NUMBER);

	input  wire 					clk;	

	input  wire 					rst;
	output wire 					busyOut;
	
	input  wire [ADDR_BWIDTH-1:0]	addra;
	input  wire						wenb;
	input  wire [ADDR_BWIDTH-1:0]	addrb;
	input  wire [REG_BWIDTH-1:0]	dib;
	output reg [REG_BWIDTH-1:0]		dout;	

wire [REG_BWIDTH-1:0] doa,dob;

MEM_RSTABLE #(.REG_BWIDTH(REG_BWIDTH), .REG_NUMBER(REG_NUMBER), .RESET_VALUE(0))
mem (	.clk(clk), .rst(rst), .busyOut(busyOut), 
		.ena(1'b1), .wena(1'b0),		.addra(addra), .dia({REG_BWIDTH{1'b0}}), .doa(doa),
		.enb(1'b1), .wenb(wenb),	.addrb(addrb), .dib(dib), .dob(dob));


reg mux;

always@(posedge clk)
begin
	if(addra == addrb)
		mux <= 1;
	else
		mux <= 0;
end

always@(*)
begin

	if(mux)
		dout = dob;
	else
		dout = doa;

end




endmodule


//==============================================================================
// 							memReader
//==============================================================================


module memReader
#(
	parameter REG_BWIDTH	= 16,
	parameter REG_NUMBER	= 1024,
	parameter MAX_ADDR		= 1		// NEEDS TO BE SET! max addr till which the mem is read out
)
(
	clk,
	rst,
	rAddrInDefault,	// the default readAddres, will be replace at readOut
	startRead,	
	dataInMem,	
	rAddrToMem,	
	dataOutDefault,	
	dataEnOut,
	dataOutReader	
);

function integer clog2;
input integer value;
begin
value = value-1;
for (clog2=0; value>0; clog2=clog2+1)
value = value>>1;
end
endfunction

	parameter ADDR_BWIDTH	= clog2(REG_NUMBER);

	input  wire clk;
	input  wire rst;
	input  wire [ADDR_BWIDTH-1:0] rAddrInDefault;	// the default readAddres, will be replace at readOut
	input  wire startRead;
	input  wire [REG_BWIDTH-1:0]  dataInMem;
	output reg  [ADDR_BWIDTH-1:0] rAddrToMem;	
	output reg  [REG_BWIDTH-1:0]  dataOutDefault;	
	output wire dataEnOut;
	output reg  [REG_BWIDTH-1:0]  dataOutReader;

localparam memLatency = 1;

reg dataEn;
wire dataEnDelayed;
reg readMode;
reg [ADDR_BWIDTH-1:0] addr;

SHIFT_DFF #( .LATENCY(memLatency), .BIT_WIDTH(1)) sr (.clk(clk), .rst(rst), .dataIn(dataEn), .dataOut(dataEnDelayed));

assign dataEnOut = dataEn || dataEnDelayed;

always@(posedge clk)
begin

	if(rst)
	begin
	
		readMode		<= 0;
		addr			<= 0;
		
	end
	else
	begin
	
		if((startRead == 1 || readMode) && addr != MAX_ADDR)
		begin
			
			readMode	<= 1;
			addr		<= addr +1;
			dataEn		<= 1;
			
		end
		else
		begin
			
			readMode	<= 0;
			addr		<= 0;
			dataEn		<= 0;
			
		end
	end

end



always@(*)
begin

	if((startRead == 1 || readMode || dataEnOut))
	begin
	
		rAddrToMem		= addr;
		dataOutDefault	= 0;
		dataOutReader	= dataInMem;
	end
	else
	begin
	
		rAddrToMem		= rAddrInDefault;
		dataOutDefault	= dataInMem;
		dataOutReader	= 0;
		
	end

end


endmodule
