/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


//////////////////////////////////////////////////////////////////////////////////
//Module: 		AXI-Master (dma_arbiter)
//Author: 		Tobias Schladt
//Version:		23.11.2015
//Description:	Arbiter for the DMA memory port. Reads the control registers
//					and passes contained tasks on to the AXI logic.
//////////////////////////////////////////////////////////////////////////////////

`default_nettype none
`timescale	1ns / 10ps

module axi_arbiter(
	input wire clk,
	input wire reset,
	
	//DMA memory interface
	output wire [8:0] bram_addr,
	output wire [31:0] bram_din,
	input wire [31:0] bram_dout,
	output wire bram_we_0,
	output wire bram_we_1,
	output wire bram_we_2,
	output wire bram_we_3,
	
	//valid reg interface
	input wire [15:0] valid_reg_0,
	input wire [15:0] valid_reg_1,
	input wire [15:0] valid_reg_2,
	input wire [15:0] valid_reg_3,
	output wire [3:0] transaction_finished,
	
	//AXI write interface
	input wire w_done,
	output wire w_valid,
	input wire [8:0] w_addr,
	input wire [31:0] w_dout,
	output wire [31:0] w_din,
	input wire w_we_0,
	input wire w_we_1,
	input wire w_we_2,
	input wire w_we_3,
	
	//AXI read interface
	input wire r_done,
	output wire r_valid,
	input wire [8:0] r_addr,
	input wire [31:0] r_dout,
	output wire [31:0] r_din,
	input wire r_we_0,
	input wire r_we_1,
	input wire r_we_2,
	input wire r_we_3,
	
	//global AXI interface
	output reg [7:0] blen,
	output reg [1:0] burst_type,
	output reg bsel
   );

	parameter DOUBLE_BUFFERING	= 0;

	//DMA memory addresses of control/status registers
	parameter W_CTRL_ADDR = 0;		//write control register address
	parameter R_CTRL_ADDR = 1;		//read control register address
	parameter VALID_BIT_POS = 10;	//position of valid bit
	
	//internal wires and registers
	reg write_state;					//reg to save read/write state
	reg [8:0] bram_addr_reg;		//output address register for DMA memory
	wire valid;
	wire dma_valid;
	
	//FSM
	reg [2:0] state, next_state;
	
	localparam [2:0]
	RESET			 	 = 1,
	IDLE			 	 = 2,
	SET_FIRST_VALID = 3,
	POLL 				 = 4,
	CHECK				 = 5,
	SET_VALID    	 = 6,
	WAIT_FOR_DONE	 = 7;
	
	//BRAM port arbitration muxes
	assign bram_addr 	= (state == POLL || state == CHECK) ? bram_addr_reg : (!write_state) ? w_addr : r_addr;
	assign bram_din  	= (state == POLL || state == CHECK) ? 0 : (!write_state) ? w_dout : r_dout;
	assign bram_we_0	= (state == POLL || state == CHECK) ? 1'b0 : (!write_state) ? w_we_0 : r_we_0;
	assign bram_we_1	= (state == POLL || state == CHECK) ? 1'b0 : (!write_state) ? w_we_1 : r_we_1;
	assign bram_we_2	= (state == POLL || state == CHECK) ? 1'b0 : (!write_state) ? w_we_2 : r_we_2;
	assign bram_we_3	= (state == POLL || state == CHECK) ? 1'b0 : (!write_state) ? w_we_3 : r_we_3;
	assign r_din		= bram_dout;
	assign w_din		= bram_dout;
	//vaid outputs for the AXI modules
	assign w_valid   	= ((state == SET_VALID || state == SET_FIRST_VALID) && write_state == 0);
	assign r_valid		= ((state == SET_VALID || state == SET_FIRST_VALID) && write_state == 1);
	//valid bits of the control registers
	assign valid		= bram_dout[VALID_BIT_POS + (16*bsel)];
	assign dma_valid  = (valid_reg_0[VALID_BIT_POS] || valid_reg_1[VALID_BIT_POS] ||
								valid_reg_2[VALID_BIT_POS] || valid_reg_3[VALID_BIT_POS]);
	
	//detect finished AXI transactions to reset mirror ctrl registers and set interrupt registers
	assign transaction_finished[0]	= (w_done && (bsel == 0));
	assign transaction_finished[1]	= (w_done && (bsel == 1));
	assign transaction_finished[2]	= (r_done && (bsel == 0));
	assign transaction_finished[3]	= (r_done && (bsel == 1));
	
	always@(*) begin
		case(state)
			RESET:			next_state = IDLE;
			IDLE:				if(dma_valid == 0) next_state = IDLE;	//leave IDLE state only if one ctrl reg valid
								else next_state = SET_FIRST_VALID;
			SET_FIRST_VALID: next_state = WAIT_FOR_DONE;				//read control data directly from registers
			POLL:				if(dma_valid == 0) next_state = IDLE;	//go to IDLE if no more tasks
								else next_state = CHECK;
			CHECK:			if(valid) next_state = SET_VALID;		//if valid bit is set, init AXI transaction
								else next_state = POLL;
			SET_VALID:		next_state = WAIT_FOR_DONE;
			WAIT_FOR_DONE: if(w_done || r_done) next_state = POLL;//continue polling when AXI transaction finished
								else next_state = WAIT_FOR_DONE;
			default:			next_state = 0;
		endcase
	end
	
	always@(posedge clk) begin
		if(reset) begin
			state <= RESET;
			bram_addr_reg <= 0;
			write_state <= 1;
			blen <= 0;
			burst_type <= 0;
			bsel <= 0;
		end else begin
			state <= next_state;
			case(next_state)
				//read control data directly from registers and set write_state and bsel
				SET_FIRST_VALID:
						begin
							if(valid_reg_0[VALID_BIT_POS] == 1) begin
								write_state <= 0;
								bsel <= 0;
								blen <= valid_reg_0[7:0];
								burst_type <= valid_reg_0[9:8];
							end else if(valid_reg_1[VALID_BIT_POS] == 1 && DOUBLE_BUFFERING == 1) begin
								write_state <= 0;
								bsel <= 1;
								blen <= valid_reg_1[7:0];
								burst_type <= valid_reg_1[9:8];
							end else if(valid_reg_2[VALID_BIT_POS] == 1) begin
								write_state <= 1;
								bsel <= 0;
								blen <= valid_reg_2[7:0];
								burst_type <= valid_reg_2[9:8];
							end else if(valid_reg_3[VALID_BIT_POS] == 1 && DOUBLE_BUFFERING == 1) begin
								write_state <= 1;
								bsel <= 1;
								blen <= valid_reg_3[7:0];
								burst_type <= valid_reg_3[9:8];
							end
						end
				//set control register address and prepare write_state and bsel for next poll
				POLL:	begin
							write_state <= ~write_state;
							if(write_state == 1) bram_addr_reg <= W_CTRL_ADDR;
							else begin
								bram_addr_reg <= R_CTRL_ADDR;
								if(DOUBLE_BUFFERING) bsel <= ~bsel;
							end
						end
				//set outputs for the AXI modules to control register settings 
				SET_VALID:
						begin
							blen <= bram_dout[(0+(16*bsel)) +: 8];
							burst_type <= bram_dout[(8+(16*bsel)) +: 2];
						end
			endcase
		end
	end
	
endmodule
