/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale	1ns / 10ps

//////////////////////////////////////////////////////////////////////////////////
//Module: 		AXI-Master (read_dma)
//Author: 		Tobias Schladt
//Version:		23.11.2015
//Description:	AXI read logic. Performs single AXI read transactions and
//					writes status information to control register afterwards.
//////////////////////////////////////////////////////////////////////////////////

module axi_read_dma
	#(
	//user definable parameters	
	parameter AXI_BUS_WIDTH 				= 16,
	parameter DOUBLE_BUFFERING				= 0,
	//DMA memory addresses
	parameter R_CTRL_ADDR					= 1,	 //read control register address
	parameter R_AXIADDR_ADDR				= 3,	 //read axi address register address
	parameter R_TRANSACTION_ID				= 0,	 //read transaction id ARID
	parameter R_DATA_BASE_ADDR_SB			= 259, //single buffer base address
	parameter R_DATA_BASE_ADDR_DB_0		= 259, //double buffer 0 base address
	parameter R_DATA_BASE_ADDR_DB_1		= 386 //double buffer 1 base address
	)
	(
	input		wire								clk,
	input		wire								reset,

	//interface to arbiter
	output	wire 								done,
	input 	wire 								mem_valid,
	output 	reg	[8:0]						addr_out,
	output	wire	[31:0]					dout,
	input		wire	[31:0] 					din,
	output	wire								we_0,
	output	wire								we_1,
	output	wire								we_2,
	output	wire								we_3,
	input		wire	[7:0]						blen,
	input		wire	[1:0]						burst_type,
	input		wire								bsel,

	//AXI signals
	//read address channel
	output	reg	[31:0]					ARADDR,
	output	reg	[7:0]						ARLEN,
	output	wire								ARVALID,
	input		wire								ARREADY,
	output	wire	[3:0]						ARID,
	output	reg	[1:0]						ARBURST,
	output	wire	[2:0]						ARSIZE,
		
	//read data channel
	input		wire	[1:0]						RRESP,
	input		wire	[AXI_BUS_WIDTH-1:0]	RDATA,
	input		wire								RLAST,
	input		wire								RVALID,
	output	wire								RREADY
   );

	//set shift amount for addressing words smaller than 32 bit
	parameter DMA_ADDR_COUNT_SHIFT		= (AXI_BUS_WIDTH == 8) ? 2 :
													  (AXI_BUS_WIDTH == 16 ) ? 1 : 0;
	//internal wires and registers
	wire	[31:0] 	data;
	reg	[31:0] 	base_addr_reg;	//register to save base address of a transaction
	reg 	[1:0]	 	resp_reg;		//register to save AXI response during transactions
	wire				clear_resp_reg;
	
	//assign default values to optional AXI-signals
	assign	ARID			= R_TRANSACTION_ID;
	generate
	case(AXI_BUS_WIDTH)	//set transfer size to full AXI bus width
		8 : 	assign	ARSIZE		= 3'b000;
		16 : 	assign	ARSIZE		= 3'b001;
		32 : 	assign	ARSIZE		= 3'b010;
		64 :	assign	ARSIZE		= 3'b011;
		128 : assign	ARSIZE		= 3'b100;
		256 : assign	ARSIZE		= 3'b101;
		512 : assign	ARSIZE		= 3'b110;
		1024:	assign	ARSIZE		= 3'b111;
	endcase
	endgenerate
	//FSM
	reg [3:0] state, next_state;
	reg [8:0] count;
	
	localparam [3:0]
	RESET									= 1,
	IDLE									= 2,	
	LOAD_ADDR							= 3,
	AWAIT_ADDR							= 4,
	STORE_ADDR							= 5,
	AWAIT_DATA_HANDSHAKE				= 6,
	BURST									= 7,
	STORE_DONE							= 8;
	
	//assign AXI control signals
	assign ARVALID 			= (state == STORE_ADDR);
	assign RREADY				= (state == AWAIT_DATA_HANDSHAKE || state == BURST);	
	//assign data out to AXI data or status information
	assign dout					= RVALID ? data : bsel ? {2'b0, 1'b1, resp_reg, 27'b0} : 
																		{18'b0, 1'b1, resp_reg, 11'b0};
	assign done					= (state == STORE_DONE);
	assign clear_resp_reg 	= (state == RESET || state == IDLE);
	
	//generate data and write enable signals according to AXI bus width
	generate
		if(AXI_BUS_WIDTH == 32) begin
			assign data 				= RDATA;
			assign we_0					= (state == AWAIT_DATA_HANDSHAKE && RVALID) ||
											(state == BURST && RVALID) || (state == STORE_DONE && bsel == 0);
			assign we_1					= we_0;
			assign we_2					= (state == AWAIT_DATA_HANDSHAKE && RVALID) ||
											(state == BURST && RVALID) || (state == STORE_DONE && bsel == 1);
			assign we_3					= we_2;
		end else if(AXI_BUS_WIDTH == 16) begin
			assign data 				= ((count-1)%2) ? {RDATA, 16'b0} : {16'b0, RDATA};
			assign we_0					= (((state == AWAIT_DATA_HANDSHAKE && RVALID) ||
											(state == BURST && RVALID)) && ((count-1)%2 == 0)) || (state == STORE_DONE && bsel == 0);
			assign we_1					= we_0;
			assign we_2					= (((state == AWAIT_DATA_HANDSHAKE && RVALID) ||
											(state == BURST && RVALID)) && ((count-1)%2 == 1)) || (state == STORE_DONE && bsel == 1);
			assign we_3					= we_2;
		end else begin
			assign data 				= ((count-1)%4 == 0) ? {24'b0, RDATA} :
												((count-1)%4 == 1) ? {16'b0, RDATA, 8'b0} :
												((count-1)%4 == 2) ? {8'b0, RDATA, 16'b0} : {RDATA, 24'b0};
			assign we_0 				= (((state == AWAIT_DATA_HANDSHAKE && RVALID) ||
											(state == BURST && RVALID)) && ((count-1)%4 == 0)) || (state == STORE_DONE && bsel == 0);
			assign we_1 				= (((state == AWAIT_DATA_HANDSHAKE && RVALID) ||
											(state == BURST && RVALID)) && ((count-1)%4 == 1)) || (state == STORE_DONE && bsel == 0);
			assign we_2 				= (((state == AWAIT_DATA_HANDSHAKE && RVALID) ||
											(state == BURST && RVALID)) && ((count-1)%4 == 2)) || (state == STORE_DONE && bsel == 1);
			assign we_3 				= (((state == AWAIT_DATA_HANDSHAKE && RVALID) ||
											(state == BURST && RVALID)) && ((count-1)%4 == 3)) || (state == STORE_DONE && bsel == 1);
		end
	endgenerate
	
	always@(*) begin
		case(state)
			RESET:		next_state = IDLE;
			//leave IDLE state if valid from arbiter
			IDLE:			if(mem_valid) next_state = LOAD_ADDR;
							else next_state = IDLE;
			//load AXI address from dma mem
			LOAD_ADDR: 	next_state = AWAIT_ADDR;
			
			AWAIT_ADDR: next_state = STORE_ADDR;
			//handshake on read address channel
			STORE_ADDR: if(ARREADY) next_state = BURST;
							else next_state = STORE_ADDR;
			//handshake on read data channel
			AWAIT_DATA_HANDSHAKE:
							if(RVALID && ARLEN == 0) next_state = STORE_DONE; //single burst
							else if(RVALID) next_state = BURST;
							else next_state = AWAIT_DATA_HANDSHAKE;
			//burst state				
			BURST:				if(!RVALID) next_state = AWAIT_DATA_HANDSHAKE;
							else if(RLAST) next_state = STORE_DONE;
							else next_state = BURST;
			//store done bit and status information				
			STORE_DONE: next_state = IDLE;

			default:		next_state = 0;
		endcase
	end
	
	always@(posedge clk) begin
		if(reset) begin
			state <= RESET;
			count <= 0;
			ARLEN <= 0;
			ARBURST <= 0;
			ARADDR <= 0;
			addr_out <= 0;
			base_addr_reg <= 0;
		end else begin
		state <= next_state;
		case(next_state)
			LOAD_ADDR:	begin
								//set address to AXI address register according to buffer select
								addr_out <= R_AXIADDR_ADDR + (2*bsel);
								ARLEN <= blen;
								ARBURST <= burst_type;
								count <= 0;
								//set buffer base address according to double buffering and buffer select
								if(bsel == 1) base_addr_reg <= R_DATA_BASE_ADDR_DB_1;
								else if(DOUBLE_BUFFERING == 1 && bsel == 0) base_addr_reg <= R_DATA_BASE_ADDR_DB_0;
								else base_addr_reg <= R_DATA_BASE_ADDR_SB;
							end
							
			STORE_ADDR: ARADDR <= din;
			
			AWAIT_DATA_HANDSHAKE:
							begin
							end
							
			BURST:		begin
								//set output address according to AXI bus width
								addr_out <= base_addr_reg + (count >> DMA_ADDR_COUNT_SHIFT);
								count <= count + 1;
							end
							
			STORE_DONE: begin
								addr_out <= R_CTRL_ADDR;	
								count <= count + 1;
							end
		endcase
		end
	end
	
	//logic to set and reset AXI response register
	always@(posedge clk) begin
		if(clear_resp_reg) resp_reg <= 0;
		else if(RVALID) begin
			if(resp_reg[0] == 0) resp_reg[0] <= RRESP[0];
			if(resp_reg[1] == 0) resp_reg[1] <= RRESP[1];
		end
	end

endmodule
