/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


//////////////////////////////////////////////////////////////////////////////////
//Module:		AXI-Master (top)
//Author:		Tobias Schladt
//Version:		23.11.2015
//Description: AXI-Bus-Master DMA-Module
//////////////////////////////////////////////////////////////////////////////////

`default_nettype none
`timescale	1ns / 10ps

module axi_top#(
	//user definable parameters
	parameter AXI_BUS_WIDTH 				= 16,
	parameter TRANSACTION_ID				= 0, // used for AWID and ARID
	parameter DOUBLE_BUFFERING				= 1,
	parameter USE_INFER_BRAM				= 1,
	parameter DMA_BASE_ADR					= 18'h0
	)
	(
	//system bus
	input		wire								clk_axi,
	input		wire								reset,
	//DMA-Signals
	input		wire	[7:0]						addr_high,
	input 	wire								mem_access,
	input		wire								mem_clk,
	input	wire									store_access_high,
	input	wire									store_access_low,
	input		wire	[9:0]						addr_peri,
	input		wire	[17:0]					do_peri,
	output	wire	[17:0]					di_peri,
	//interrupt signals
	output	wire	[3:0]						intr,
	
	//AXI signals
	//global signals
	output	wire								ACLK,
	output	wire								ARESETn,
	
	//write address channel
	output	wire	[31:0]					AWADDR,
	output	wire	[7:0]						AWLEN,
	output	wire								AWVALID,
	input		wire								AWREADY,
	output	wire	[3:0]						AWID,
	output	wire	[1:0]						AWBURST,
	output	wire	[2:0]						AWSIZE,
	
	//write data channel
	output	wire	[AXI_BUS_WIDTH-1:0]	WDATA,
	output	wire	[AXI_BUS_WIDTH/8-1:0]		WSTRB,
	output	wire								WLAST,
	output	wire								WVALID,
	input		wire								WREADY,
	
	//write response channel
	input		wire	[3:0]						BID,
	input		wire								BVALID,
	output	wire								BREADY,
	input		wire	[1:0]						BRESP,
	
	//read address channel
	output	wire	[31:0]					ARADDR,
	output	wire	[7:0]						ARLEN,
	output	wire								ARVALID,
	input		wire								ARREADY,
	output	wire	[3:0]						ARID,
	output	wire	[1:0]						ARBURST,
	output	wire	[2:0]						ARSIZE,
		
	//read data channel
	input		wire	[1:0]						RRESP,
	input		wire	[3:0]						RID,
	input		wire	[AXI_BUS_WIDTH-1:0]	RDATA,
	input		wire								RLAST,
	input		wire								RVALID,
	output	wire								RREADY
   );
	
	
	//DMA memory register addresses
	parameter W_CTRL_ADDR = 0;		//write control register address
	parameter R_CTRL_ADDR = 1;		//read control register address
	

	
	//internal wires
	wire [7:0] blen;
	wire [1:0] burst_type;
	wire store_access, select, bsel;
	
	//arbiter - bram
	wire web_0, web_1, web_2, web_3;
	wire [8:0] addrb;
	wire [31:0] dinb, doutb;
	
	//arbiter - mirror control regs
	wire [3:0] transaction_finished;
	wire [15:0] valid_reg_0;
	wire [15:0] valid_reg_1;
	wire [15:0] valid_reg_2;
	wire [15:0] valid_reg_3;
	
	//arbiter - write
	wire w_done, w_mem_valid, w_we_0, w_we_1, w_we_2, w_we_3;
	wire [8:0] w_addr_out;
	wire [31:0] w_dout, w_din;
	
	//arbiter - read
	wire r_done, r_mem_valid, r_we_0, r_we_1, r_we_2, r_we_3;
	wire [8:0] r_addr_out;
	wire [31:0] r_dout, r_din;
	
	assign	ARESETn			= !reset;	//AXI bus reset
	assign	ACLK				= clk_axi;	//AXI bus clock
	assign	di_peri[17:16] = 2'b0;		//upper 2 bits of DMA bus not used
	assign	store_access	= store_access_high | store_access_low;	// byte-wide write not supported

	//pselect module to generate select signal for DMA memory
	pselect #(
		.ADDR_WIDTH(8),
		.BASE_WIDTH(8),
		.BASE_ADDR(DMA_BASE_ADR[17:10])
		) pselect_0 (
		.addr(addr_high),
		.activ_peri(mem_access),
		.select(select)
	);
	
	//instantiate DMA memory as inference BRAM or macro depending on parameter
	generate
		if(USE_INFER_BRAM) begin
			bram_infer DMARAM00 (
				.clkA(mem_clk), 
				.clkB(ACLK), 
				.enA(1'b1), 
				.enB(1'b1), 
				.weA((store_access && select)), 
				.weB({web_2, web_0}), 
				.rstA(!select), 
				.addrA(addr_peri), 
				.addrB(addrb), 
				.diA(do_peri[15:0]), 
				.diB(dinb), 
				.doA(di_peri[15:0]), 
				.doB(doutb)
			);	
		end else begin
			BRAM_TDP_MACRO #(
				.BRAM_SIZE("18Kb"), // Target BRAM: "9Kb" or "18Kb"
				.DEVICE("SPARTAN6"), // Target device: "VIRTEX5", "VIRTEX6", "SPARTAN6"
				.DOA_REG(0),        // Optional port A output register (0 or 1)
				.DOB_REG(0),        // Optional port B output register (0 or 1)
				.INIT_A(36'h00000000),  // Initial values on port A output port
				.INIT_B(36'h00000000), // Initial values on port B output port
				.INIT_FILE ("NONE"),
				.READ_WIDTH_A (16),   // Valid values are 1-36
				.READ_WIDTH_B (32),   // Valid values are 1-36
				.SIM_COLLISION_CHECK ("ALL"), // Collision check enable "ALL", "WARNING_ONLY",
				//   "GENERATE_X_ONLY" or "NONE"
				.SRVAL_A(36'h00000000), // Set/Reset value for port A output
				.SRVAL_B(36'h00000000), // Set/Reset value for port B output
				.WRITE_MODE_A("WRITE_FIRST"), // "WRITE_FIRST", "READ_FIRST", or "NO_CHANGE"
				.WRITE_MODE_B("WRITE_FIRST"), // "WRITE_FIRST", "READ_FIRST", or "NO_CHANGE"
				.WRITE_WIDTH_A(16), // Valid values are 1-36
				.WRITE_WIDTH_B(32) // Valid values are 1-36
			) DMARAM00 (
				.DOA(di_peri[15:0]),       // Output port-A data, width defined by READ_WIDTH_A parameter
				.DOB(doutb),       // Output port-B data, width defined by READ_WIDTH_B parameter
				.ADDRA(addr_peri),   // Input port-A address, width defined by Port A depth
				.ADDRB(addrb),   // Input port-B address, width defined by Port B depth
				.CLKA(mem_clk),     // 1-bit input port-A clock
				.CLKB(ACLK),     // 1-bit input port-B clock
				.DIA(do_peri[15:0]),       // Input port-A data, width defined by WRITE_WIDTH_A parameter
				.DIB(dinb),       // Input port-B data, width defined by WRITE_WIDTH_B parameter
				.ENA(1'b1),       // 1-bit input port-A enable
				.ENB(1'b1),       // 1-bit input port-B enable
				.REGCEA(1'b0), // 1-bit input port-A output register enable
				.REGCEB(1'b0), // 1-bit input port-B output register enable
				.WEA({(store_access && select), (store_access && select)}),       // Input port-A write enable, width defined by Port A depth
				.WEB({web_3, web_2, web_1, web_0}),   // Input port-B write enable, width defined by Port B depth
				.RSTA(!select),
				.RSTB(1'b0)
			);
		end
	endgenerate
	
	axi_arbiter #(
		.DOUBLE_BUFFERING(DOUBLE_BUFFERING),
		.W_CTRL_ADDR(W_CTRL_ADDR),
		.R_CTRL_ADDR(R_CTRL_ADDR)) axi_arbiter_0 (
		.clk(ACLK), 
		.reset(reset),
		//interface to DMA memory
		.bram_addr(addrb), 
		.bram_din(dinb), 
		.bram_dout(doutb), 
		.bram_we_0(web_0),
		.bram_we_1(web_1),
		.bram_we_2(web_2),
		.bram_we_3(web_3),
		//interface to mirror control registers and interrupt registers
		.valid_reg_0(valid_reg_0),
		.valid_reg_1(valid_reg_1),
		.valid_reg_2(valid_reg_2),
		.valid_reg_3(valid_reg_3),
		.transaction_finished(transaction_finished), 
		//interface to AXI modules
		.blen(blen),
		.burst_type(burst_type),
		.bsel(bsel),
		//AXI write
		.w_done(w_done), 
		.w_valid(w_mem_valid), 
		.w_addr(w_addr_out), 
		.w_dout(w_dout), 
		.w_din(w_din), 
		.w_we_0(w_we_0),
		.w_we_1(w_we_1),
		.w_we_2(w_we_2),
		.w_we_3(w_we_3),
		//AXI read
		.r_done(r_done), 
		.r_valid(r_mem_valid), 
		.r_addr(r_addr_out), 
		.r_dout(r_dout), 
		.r_din(r_din), 
		.r_we_0(r_we_0),
		.r_we_1(r_we_1),
		.r_we_2(r_we_2),
		.r_we_3(r_we_3)
    );
	
	//mirror control registers
	mirror_ctrl_regs mirror_ctrl_regs_0 (
		.clk(ACLK), 
		.reset(reset), 
		.valid_reg_0(valid_reg_0),
		.valid_reg_1(valid_reg_1),
		.valid_reg_2(valid_reg_2),
		.valid_reg_3(valid_reg_3), 
		.transaction_finished(transaction_finished), 
		.do_peri(do_peri), 
		.addr_peri(addr_peri), 
		.we_dma((store_access && select))
   );
	 
	//interrupt register
	intr_reg intr_reg_0 (
		.clk(ACLK), 
		.reset(reset), 
		.transaction_finished(transaction_finished),
		.do_peri(do_peri), 
		.addr_peri(addr_peri), 
		.we_dma((store_access && select)), 
		.intr_reg(intr)
   );

	axi_write_dma #(
		.AXI_BUS_WIDTH(AXI_BUS_WIDTH),
		.DOUBLE_BUFFERING(DOUBLE_BUFFERING),
		.W_CTRL_ADDR(W_CTRL_ADDR),
		.W_TRANSACTION_ID(TRANSACTION_ID)) write_0 (
		.clk					(ACLK), 
		.reset				(reset),
		//Interface to arbiter
		.done					(w_done), 
		.mem_valid			(w_mem_valid), 
		.addr_out			(w_addr_out), 
		.dout					(w_dout), 
		.din					(w_din), 
		.we_0					(w_we_0),
		.we_1					(w_we_1),
		.we_2					(w_we_2),
		.we_3					(w_we_3),
		.blen					(blen),
		.burst_type			(burst_type),
		.bsel					(bsel),
		//AXI
		.AWADDR				(AWADDR),
		.AWLEN				(AWLEN),
		.AWVALID				(AWVALID),
		.AWREADY				(AWREADY),
		.AWID					(AWID),
		.AWBURST				(AWBURST),
		.AWSIZE				(AWSIZE),
		.WDATA				(WDATA),
		.WSTRB				(WSTRB),
		.WLAST				(WLAST),
		.WVALID				(WVALID),
		.WREADY				(WREADY),
		.BVALID				(BVALID),
		.BREADY				(BREADY),
		.BRESP				(BRESP)
		);
		
	axi_read_dma #(
		.AXI_BUS_WIDTH(AXI_BUS_WIDTH),
		.DOUBLE_BUFFERING(DOUBLE_BUFFERING),
		.R_CTRL_ADDR(R_CTRL_ADDR),
		.R_TRANSACTION_ID(TRANSACTION_ID)) read_0 (
		.clk					(ACLK),
		.reset				(reset),
		//Interface to arbiter
		.done					(r_done), 
		.mem_valid			(r_mem_valid), 
		.addr_out			(r_addr_out), 
		.dout					(r_dout), 
		.din					(r_din), 
		.we_0					(r_we_0), 
		.we_1					(r_we_1),
		.we_2					(r_we_2), 
		.we_3					(r_we_3),
		.blen					(blen),
		.burst_type			(burst_type),
		.bsel					(bsel),
		//AXI
		.ARADDR				(ARADDR),
		.ARLEN				(ARLEN),
		.ARVALID				(ARVALID),
		.ARREADY				(ARREADY),
		.ARID					(ARID),
		.ARBURST				(ARBURST),
		.ARSIZE				(ARSIZE),
		.RRESP				(RRESP),
		.RDATA				(RDATA),
		.RLAST				(RLAST),
		.RVALID				(RVALID),
		.RREADY				(RREADY)
		);

endmodule
