/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale	1ns / 10ps

//////////////////////////////////////////////////////////////////////////////////
//Module: 		AXI-Master (write_dma)
//Author: 		Tobias Schladt
//Version:		23.11.2015
//Description:	AXI write logic. Performs single AXI write transactions and
//					writes status information to control register afterwards.
//////////////////////////////////////////////////////////////////////////////////

module axi_write_dma#(
		//user definable parameters
	parameter AXI_BUS_WIDTH 				= 16,
	parameter DOUBLE_BUFFERING				= 0,
	//DMA memory addresses
	parameter W_CTRL_ADDR					= 0,	 //write control register address
	parameter W_AXIADDR_ADDR				= 2,	 //write axi address register address
	parameter W_TRANSACTION_ID				= 0,	 //write transaction id AWID
	parameter W_DATA_BASE_ADDR_SB			= 6,   //single buffer base address
	parameter W_DATA_BASE_ADDR_DB_0			= 6,   //double buffer 0 base address
	parameter W_DATA_BASE_ADDR_DB_1			= 133 //double buffer 1 base address
	)
	(
	input 	wire 								clk,
	input 	wire 								reset,
	
	//interface to arbiter
	output	wire 								done,
	input 	wire 								mem_valid,
	output 	reg	[8:0]						addr_out,
	output	reg	[31:0]					dout,
	input		wire	[31:0] 					din,
	output	wire								we_0,
	output	wire								we_1,
	output	wire								we_2,
	output	wire								we_3,
	input		wire	[7:0]						blen,
	input		wire	[1:0]						burst_type,
	input		wire								bsel,

	//AXI signals	
	//write address channel
	output	reg	[31:0]					AWADDR,
	output	reg	[7:0]						AWLEN,
	output	wire								AWVALID,
	input		wire								AWREADY,
	output	wire	[3:0]						AWID,
	output	reg	[1:0]						AWBURST,
	output	wire	[2:0]						AWSIZE,
	
	//write data channel
	output	wire	[AXI_BUS_WIDTH-1:0]	WDATA,
	output	wire	[AXI_BUS_WIDTH/8-1:0]		WSTRB,
	output	wire								WLAST,
	output	wire								WVALID,
	input		wire								WREADY,
	
	//write response channel
	input		wire								BVALID,
	output	wire								BREADY,
	input		wire	[1:0]						BRESP
   );
	
	
	//derived parameters
	//calculate number of byte lanes
	parameter BYTE_LANES						= AXI_BUS_WIDTH/8;
	//set shift amount for addressing words smaller than 32 bit
	parameter DMA_ADDR_COUNT_SHIFT		= (AXI_BUS_WIDTH == 8) ? 2 :
													  (AXI_BUS_WIDTH == 16 ) ? 1 : 0;
	//internal wires and registers
	reg	[31:0] first_transfer_reg;	//register to buffer first transfer of a transaction
	reg 	[31:0] base_addr_reg;		//register to save base address of a transaction
	wire 	[31:0] data;

	//assign default values to optional AXI signals
	generate	//generate write strobe signal (enable all byte lanes)
		genvar i;
		for(i=0; i<BYTE_LANES; i=i+1) begin : GEN_WSTRB
			assign 	WSTRB[i]		= 'b1;
		end
	endgenerate

	assign	AWID			= W_TRANSACTION_ID;
generate	
	case(AXI_BUS_WIDTH)	//set transfer size to full AXI bus width
		8 : 	assign	AWSIZE		= 3'b000;
		16 : 	assign	AWSIZE		= 3'b001;
		32 : 	assign	AWSIZE		= 3'b010;
		64 :	assign	AWSIZE		= 3'b011;
		128 : assign	AWSIZE		= 3'b100;
		256 : assign	AWSIZE		= 3'b101;
		512 : assign	AWSIZE		= 3'b110;
		1024:	assign	AWSIZE		= 3'b111;
	endcase
endgenerate
	//FSM
	reg [3:0] state, next_state;
	reg [8:0] count;
	
	localparam [3:0]
	RESET									= 1,
	IDLE									= 2,	
	LOAD_ADDR							= 3,
	AWAIT_ADDR							= 4,
	STORE_ADDR							= 5,
	LOAD_FIRST_TRANSFER				= 6,	
	AWAIT_FIRST_TRANSFER				= 7,	
	AWAIT_DATA_HANDSHAKE_SINGLE	= 8,
	AWAIT_DATA_HANDSHAKE 			= 9,
	BURST									= 10,
	LAST_BURST							= 11,
	AWAIT_RESP							= 12,
	STORE_DONE							= 13,
	BURST_WAIT							= 14,
	BURST_WAIT2							= 15;
	
	//assign AXI control signals
	assign AWVALID = (state == STORE_ADDR);
	assign WVALID	= (state == AWAIT_DATA_HANDSHAKE || state == AWAIT_DATA_HANDSHAKE_SINGLE ||
							state == BURST || state == BURST_WAIT || state == BURST_WAIT2 || state == LAST_BURST);
	assign WLAST	= (state == LAST_BURST || state == AWAIT_DATA_HANDSHAKE_SINGLE);
	assign BREADY	= (state == AWAIT_RESP);
	//set write enables to store status information
	assign we_0		= (state == STORE_DONE && bsel == 0);
	assign we_1		= we_0;
	assign we_2		= (state == STORE_DONE && bsel == 1);
	assign we_3 	= we_2;
	//done signal for arbiter
	assign done		= (state == STORE_DONE);
	
	//assign AXI data out
	assign data		= (state == AWAIT_DATA_HANDSHAKE || state == AWAIT_DATA_HANDSHAKE_SINGLE || state == BURST_WAIT ||
						state == BURST_WAIT2) ? first_transfer_reg : (WVALID) ? din : 0;
	//select bytes to put on AXI bus according to bus width
	generate
		if(AXI_BUS_WIDTH == 32) assign WDATA = data;
		else if(AXI_BUS_WIDTH == 16) assign WDATA = data[0 + ((count)%2)*16 +: 16];
		else assign WDATA = data[0 + ((count-2)%4)*8 +: 8];
	endgenerate
	
	
	always@(*) begin
		case(state)
			RESET:		next_state = IDLE;
			//leave IDLE state if valid from arbiter is set
			IDLE:			if(mem_valid) next_state = LOAD_ADDR;
							else next_state = IDLE;
			//load AXI address from DMA mem
			LOAD_ADDR: 	next_state = AWAIT_ADDR;			
			AWAIT_ADDR: next_state = STORE_ADDR;
			//handshake on write address channel
			STORE_ADDR: if(AWREADY) next_state = LOAD_FIRST_TRANSFER;
							else next_state = STORE_ADDR;
			//load first AXI transfer				
			LOAD_FIRST_TRANSFER: 
							next_state = AWAIT_FIRST_TRANSFER;
			AWAIT_FIRST_TRANSFER:
							if(AWLEN == 0) next_state = AWAIT_DATA_HANDSHAKE_SINGLE;
							else next_state = AWAIT_DATA_HANDSHAKE;
			//handshake on write data channel
			//single burst
			AWAIT_DATA_HANDSHAKE_SINGLE:
							if(WREADY) next_state = AWAIT_RESP;
							else next_state = AWAIT_DATA_HANDSHAKE_SINGLE;
			//non-single burst
			AWAIT_DATA_HANDSHAKE:
							if(WREADY && AWLEN == 1) next_state = LAST_BURST;
							else if(WREADY) next_state = BURST;
							else next_state = BURST_WAIT;
			//burst state
			BURST:		if(count <= AWLEN && WREADY) next_state = BURST;
							else if(!WREADY) next_state = BURST_WAIT;
							else next_state = LAST_BURST;
			BURST_WAIT:
							if(count <= AWLEN && WREADY) next_state = BURST;
							else if(!WREADY) next_state = BURST_WAIT2;
							else next_state = LAST_BURST;
			BURST_WAIT2:
							if(count <= AWLEN && WREADY) next_state = BURST;
							else if(!WREADY) next_state = BURST_WAIT2;
							else next_state = LAST_BURST;
			//last burst state to set WLAST signal
			LAST_BURST:
							if(!WREADY) next_state = LAST_BURST;
							else next_state = AWAIT_RESP;
			//handshake on write response channel
			AWAIT_RESP: if(BVALID) next_state = STORE_DONE;
							else next_state = AWAIT_RESP;		
			//store status information in control regs
			STORE_DONE: next_state = IDLE;		
			default: next_state = 0;
		endcase
	end
	
	always@(posedge clk) begin
		if(reset) begin
			state <= RESET;
			count <= 0;
			AWADDR <= 0;
			AWLEN <= 0;
			AWBURST <= 0;
			first_transfer_reg <= 0;
			dout <= 0;
			addr_out <= 0;
			base_addr_reg <= 0;
		end else begin
		state <= next_state;
		case(next_state)
			LOAD_ADDR:	begin
								//set address to AXI address register according to buffer select
								addr_out <= W_AXIADDR_ADDR + (2*bsel);
								AWLEN <= blen;
								AWBURST <= burst_type;
								//set buffer base address according to double buffering and buffer select
								if(bsel == 1) base_addr_reg <= W_DATA_BASE_ADDR_DB_1;
								else if(DOUBLE_BUFFERING == 1 && bsel == 0) base_addr_reg <= W_DATA_BASE_ADDR_DB_0;
								else base_addr_reg <= W_DATA_BASE_ADDR_SB;
							end
							
			STORE_ADDR: AWADDR <= din;
			
			LOAD_FIRST_TRANSFER:	
							addr_out <= base_addr_reg;
							
			AWAIT_DATA_HANDSHAKE_SINGLE:
							begin
								first_transfer_reg <= din;
								count <= 2;
							end
			
			AWAIT_DATA_HANDSHAKE:
							begin
								first_transfer_reg <= din;
								addr_out <= base_addr_reg + (1 >> DMA_ADDR_COUNT_SHIFT);
								count <= 2;
							end
							
			BURST:		begin
								//set output address according to AXI bus width
								first_transfer_reg <= din;
								addr_out <= base_addr_reg + (count >> DMA_ADDR_COUNT_SHIFT);
								count <= count + 1;
							end
			
			BURST_WAIT:
							begin
								first_transfer_reg <= din;
								addr_out <= base_addr_reg + (count-1 >> DMA_ADDR_COUNT_SHIFT);
							end
			BURST_WAIT2:
							begin
								addr_out <= base_addr_reg + (count-1 >> DMA_ADDR_COUNT_SHIFT);
							end

			LAST_BURST: count <= count + 1;
			
			STORE_DONE: begin
								//store done bit and AXI repsonse to ctrl reg
								addr_out <= W_CTRL_ADDR;
								if(bsel == 0) dout <= {18'b0, 1'b1, BRESP, 11'b0};
								else dout <= {2'b0, 1'b1, BRESP, 27'b0};
							end
		endcase
		end
	end

endmodule
