/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


//////////////////////////////////////////////////////////////////////////////////
//Modul:			AXI-Master (DMA memory)
//Autor:			Tobias Schladt
//Version:		06.10.2015
//Description:	Code for asymmectric inference BRAM taken from
//					Xilinx XST User Guide (UG687) and extdened by reset for port A
//					and 16 bit write enable for port B.
//////////////////////////////////////////////////////////////////////////////////

`default_nettype none
`timescale	1ns / 10ps

module bram_infer
	#(
	parameter WIDTHA      = 16,
	parameter SIZEA       = 1024,
	parameter ADDRWIDTHA  = 10,
	parameter WIDTHB      = 32,
	parameter SIZEB       = 512,
	parameter ADDRWIDTHB  = 9
	)
	(
	input wire                  		    clkA,
	input wire                        		clkB,
	input wire                        		enA,
	input wire                        		enB,
	input wire                       		weA,
	input wire  [ratio(WIDTHA, WIDTHB)-1:0]	weB,
	input wire                        		rstA,
	input wire  [ADDRWIDTHA-1:0]      		addrA,
	input wire  [ADDRWIDTHB-1:0]      		addrB,
	input wire  [WIDTHA-1:0]          		diA,
	input wire  [WIDTHB-1:0]          		diB,
	output reg  [WIDTHA-1:0]          		doA,
	output reg  [WIDTHB-1:0]          		doB
	);



	

	`define max(a,b) {(a) > (b) ? (a) : (b)}
	`define min(a,b) {(a) < (b) ? (a) : (b)}
	
	function integer ratio;
		input integer a;
		input integer b;
		begin
			ratio = `max(a,b)/`min(a,b);
		end
	endfunction

	function integer log2;
	 input integer value;
	 reg [31:0] shifted;
	 integer res;
	begin
	 if (value < 2)
		log2 = value;
	 else
	 begin
		shifted = value-1;
		for (res=0; shifted>0; res=res+1)
		  shifted = shifted>>1;
		log2 = res;
	 end
	end
	endfunction

	localparam maxSIZE   = `max(SIZEA, SIZEB);
	localparam maxWIDTH  = `max(WIDTHA, WIDTHB);
	localparam minWIDTH  = `min(WIDTHA, WIDTHB);
	localparam RATIO     = ratio(WIDTHA, WIDTHB);
	localparam log2RATIO = log2(RATIO);

	reg     [minWIDTH-1:0]  RAM [0:maxSIZE-1];

	genvar i;

	//Port A description
	always @(posedge clkA) begin
		if(enA) begin
			if(weA) begin
				RAM[addrA] <= diA;
		   if(!rstA)
				doA <= diA;
			else
				doA <= 0;
			end else if(!rstA)
				doA <= RAM[addrA];
			else
				doA <= 0;
		end
	end
	
	//Port B description
	generate
		for (i = 0; i < RATIO; i = i+1) begin: portb
			wire [log2RATIO-1:0] lsbaddr;
			assign lsbaddr = i;
			always @(posedge clkB) begin
				if(enB) begin
					if(weB[i]) begin
						RAM[{addrB, lsbaddr}] <= diB[(i+1)*minWIDTH-1:i*minWIDTH];
						doB[(i+1)*minWIDTH-1:i*minWIDTH] <= diB[(i+1)*minWIDTH-1:i*minWIDTH];
					end else begin
						doB[(i+1)*minWIDTH-1:i*minWIDTH] <= RAM[{addrB, lsbaddr}];
					end
				end
			end
		end
	endgenerate
	
endmodule
