/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale 1ns / 10ps

module axi_cache #(
	parameter LINE_SIZE = 4, // Number of words in a cache line
	parameter WAYS = 1, // TODO Number of ways to describe associativity
	parameter SETS = 4, // Number of sets (if WAYS=1 number of cache lines)
	parameter DMA_BASE_ADR = 18'h2000,
	parameter DMA_SIZE = 'd11, // #Bits of DMA address range
	parameter DATA_RAMBLOCKS = 50, // Equivalent number of blockrams for DMA-SIZE
	parameter SINGLE_MASTER = 0 // only memory module on the bus
) (
	// DMA interface
	input  wire        reset,
	input  wire        clk_peri,
	input  wire        mem_access,
	input  wire        store_access_low,
	input  wire        store_access_high,
	input  wire [9:0]  addr_peri,
	input  wire [7:0]  addr_high,
	input  wire [17:0] do_peri,
	output wire [17:0] di_peri,

	output wire        stall,
	output wire        stall_wr,

	// TODO AXI global signals

	// AXI interface
	// Read
	output wire [31:0] ARADDR,
	output wire [7:0]  ARLEN,
	output wire        ARVALID,
	input  wire        ARREADY,
	output wire [3:0]  ARID,
	output wire [1:0]  ARBURST,
	output wire [2:0]  ARSIZE,

	input  wire [1:0]  RRESP,
	input  wire [3:0]  RID,
	input  wire [31:0] RDATA,
	input  wire        RLAST,
	input  wire        RVALID,
	output wire        RREADY,

	// Write
	output wire [31:0] AWADDR,
	output wire [7:0]  AWLEN,
	output wire        AWVALID,
	input  wire        AWREADY,
	output wire [3:0]  AWID,
	output wire [1:0]  AWBURST,
	output wire [2:0]  AWSIZE,

	output wire [31:0] WDATA,
	output wire [3:0]  WSTRB,
	output wire        WLAST,
	output wire        WVALID,
	input  wire        WREADY,

	input  wire [3:0]  BID,
	input  wire        BVALID,
	output wire        BREADY,
	input  wire [1:0]  BRESP,

	// Coherence bus
	output wire        last_i,
	input  wire        last_o,
	output wire        request,
	input  wire        grant,

	output wire [17:0] addr_i,
	input  wire [17:0] addr_o,
	output wire        addr_valid_i,
	input  wire        addr_valid_o,
	output wire [17:0] data_i,
	input  wire [17:0] data_o,
	output wire        hit_i,
	input  wire        hit_o,
	output wire [1:0]  type_i,
	input  wire [1:0]  type_o,
	output wire        hold_i,
	input  wire        hold_o
);

function integer log2;
	input integer value;
	begin
		value = value - 1;
		for (log2=0; value>0; log2=log2+1) begin
			value = value>>1;
		end
	end
endfunction

localparam ADDRWIDTH_DATA = log2(SETS)+log2(LINE_SIZE);
localparam LINE_BITS = log2(LINE_SIZE);
localparam SET_BITS = log2(SETS);
localparam WAYS_BITS = log2(WAYS);
wire [17:0] addr;
assign addr = {addr_high, addr_peri};
wire select;
wire store_access;
assign store_access = store_access_low | store_access_high;
reg store_low_reg, store_high_reg;
reg [17:0] data_in;
reg [17:0] addr_reg;

// Generate select signal from dma bus signals
generate
if(SINGLE_MASTER == 0) begin
	assign select = mem_access && addr[17:DMA_SIZE] >= DMA_BASE_ADR[17:DMA_SIZE] && addr[17:DMA_SIZE] < (DMA_BASE_ADR[17:DMA_SIZE] + DATA_RAMBLOCKS);
end else begin
	assign select = mem_access;
end
endgenerate

wire [WAYS*(18-ADDRWIDTH_DATA)-1:0] tag, tag_masked;
wire hit_reg;
wire [WAYS*3-1:0] status_ways, status_ways_masked;
wire [2:0] status; // status bits: dirty, valid, shared
reg write_transaction;
wire valid, dirty, shared, valid_all;
wire [WAYS-1:0] valid_ways, dirty_ways;
reg set_stall, set_write, set_byte_wide_write, store_tag, set_addr_reg, set_data_in, set_coherence, set_notify_write_coherence, set_doA_reg, set_lru_access;
reg [2:0] set_new_state;
reg [3:0] state, next_state, last_state;
reg stall_reg;
reg mem_access_reg;
wire [WAYS-1:0] way_select_ui, way_select_coherence;

reg coherence_request, coherence_write_notification;
wire [1:0] coherence_state;

wire [WAYS_BITS-1:0] lru_position;
wire [WAYS-1:0] first_invalid_way;
wire [WAYS-2:0] lru_data, lru_data_accessed;
reg  [WAYS-2:0] lru_write_data;
reg  [WAYS-1:0] lru_new_way, lru_new_way_sync;
reg lru_we;

localparam
	RESET           = 0,
	IDLE            = 1,
	CHECK_HIT       = 2,
	LOAD_CACHELINE  = 3,
	LOAD_DATA       = 6,
	WRITE_CACHELINE = 4,
	CHECK_COHERENCE = 5,
	NOTIFY_COHERENCE = 7,
	NEW_WAY         = 8;

localparam
	MOESI_MODIFIED  = 3'b110,
	MOESI_OWNED     = 3'b111,
	MOESI_EXCLUSIVE = 3'b010,
	MOESI_SHARED    = 3'b011,
	MOESI_INVALID   = 3'b000,
	MOESI_INVALID_DIRTY = 3'b100,
	MOESI_NO_CHANGE = 3'b001;
	// status bits: dirty valid shared

reg [2:0] status_write_data;
reg status_we;
wire [ADDRWIDTH_DATA-LINE_BITS-1:0] addr_tag, status_addr, tag_addr;
wire status_weB, tag_weB;
wire [2:0] status_di;
wire [3*WAYS-1:0] status_do;
assign addr_tag = (state != IDLE && (state != CHECK_HIT || !hit_reg) || status_we) ? addr_reg[ADDRWIDTH_DATA-1:LINE_BITS] : addr[ADDRWIDTH_DATA-1:LINE_BITS];

reg  [18-ADDRWIDTH_DATA-1:0] tag_write_data, tag_writeback_sync;
wire [18-ADDRWIDTH_DATA-1:0] tag_di, tag_writeback;
wire [(18-ADDRWIDTH_DATA)*WAYS-1:0] tag_do;
reg tag_we;

// detect a hit: tag same as requested AND cache entry is valid AND (if write transaction) is EXCLUSIVE/MODIFIED (==!shared)
//assign hit_reg = tag == addr_reg[17:ADDRWIDTH_DATA] && valid && !(write_transaction && shared);
assign hit_reg = |way_select_ui && valid;

wire cache_write_enable;
assign cache_write_enable = ((state == CHECK_HIT && hit_reg || state == LOAD_DATA) && write_transaction); // (write hit || write miss)
wire [ADDRWIDTH_DATA-1:0] addr_data, data_addr;
wire write_we, data_we;
wire [17:0] write_addr;
wire [17:0] write_data, write_data_axi, data_di;
wire [17:0] data_doA;
wire [18*WAYS-1:0] data_do, data_doA_ways, data_doA_ways_masked;
reg [17:0] data_doA_reg;
wire [17:0] read_addr;
assign addr_data = (state == LOAD_CACHELINE) ? write_addr[ADDRWIDTH_DATA-1:0] : 
					(state == WRITE_CACHELINE) ? read_addr[ADDRWIDTH_DATA-1:0] : 
					(stall || state == LOAD_DATA || (state == CHECK_HIT && write_transaction)) ? addr_reg[ADDRWIDTH_DATA-1:0] : addr[ADDRWIDTH_DATA-1:0];
assign write_data = (state == LOAD_CACHELINE) ? write_data_axi : data_in;

generate
if(SINGLE_MASTER == 0) begin
	assign di_peri = (state == CHECK_HIT || write_transaction || last_state == LOAD_DATA) ? (write_transaction ? data_doA_reg : data_doA) : 18'b0; // read hit / write after read / read miss
end else begin // keep output signal for hardware accelerator
	assign di_peri = (state == CHECK_HIT || write_transaction || last_state == LOAD_DATA) ? (write_transaction ? data_doA_reg : data_doA) : data_doA_reg;
end

genvar i, j;
for(i=0; i<WAYS; i=i+1) begin : ways_ram
	// store status
	dp_ram #(
		.WIDTH(3),
		.ADDRWIDTH(SET_BITS),
		.SIZE(SETS)
	) statuses (
		.clk(clk_peri),
		.enA(1'b1),
		.enB(1'b1),
		.weA(status_we && way_select_ui[i]),
		.addrA(addr_tag),
		.diA(status_write_data),
		.doA(status_ways[3*(i+1)-1:3*i]),
		.addrB(status_addr),
		.weB(status_weB && way_select_coherence[i]),
		.diB(status_di),
		.doB(status_do[3*(i+1)-1:3*i])
	);
	assign valid_ways[i] = status_ways[3*i+1];
	assign dirty_ways[i] = status_ways[3*i+2];
	assign status_ways_masked[3*(i+1)-1:3*i] = (way_select_ui[i] || lru_new_way_sync[i]) ? status_ways[3*(i+1)-1:3*i] : 3'b0;

	// store tags
	dp_ram #(
		.WIDTH(18-ADDRWIDTH_DATA),
		.ADDRWIDTH(SET_BITS),
		.SIZE(SETS)
	) tags (
		.clk(clk_peri),
		.enA(1'b1),
		.enB(1'b1),
		//~ .weA(tag_we && (i == lru_position)),
		.weA(tag_we && lru_new_way_sync[i]),
		.addrA(addr_tag),
		.diA(tag_write_data),
		.doA(tag[(18-ADDRWIDTH_DATA)*(i+1)-1:(18-ADDRWIDTH_DATA)*i]),
		.addrB(tag_addr),
		.weB(tag_weB && way_select_coherence[i]),
		.diB(tag_di),
		.doB(tag_do[(18-ADDRWIDTH_DATA)*(i+1)-1:(18-ADDRWIDTH_DATA)*i])
	);
	assign way_select_ui[i] = tag[(i+1)*(18-ADDRWIDTH_DATA)-1:i*(18-ADDRWIDTH_DATA)] == addr_reg[17:ADDRWIDTH_DATA];
	assign tag_masked[(18-ADDRWIDTH_DATA)*(i+1)-1:(18-ADDRWIDTH_DATA)*i] = (lru_new_way[i]) ? tag[(18-ADDRWIDTH_DATA)*(i+1)-1:(18-ADDRWIDTH_DATA)*i] : {{18-ADDRWIDTH_DATA}{1'b0}};

	// store data
	dp_ram #(
		.WIDTH(18),
		.ADDRWIDTH(ADDRWIDTH_DATA),
		.SIZE(SETS*LINE_SIZE),
		.NUM_COL(2)
	) data (
		.clk(clk_peri),
		.enA(1'b1),
		.enB(1'b1),
		.weA({(store_high_reg && cache_write_enable || write_we),(store_low_reg && cache_write_enable || write_we)} & {way_select_ui[i],way_select_ui[i]}),
		.addrA(addr_data),
		.diA(write_data),
		.doA(data_doA_ways[18*(i+1)-1:18*i]),
		.addrB(data_addr),
		.weB(data_we && way_select_coherence[i]),
		.diB(data_di),
		.doB(data_do[18*(i+1)-1:18*i])
	);
	if(WAYS == 1) // no need for mux with only one way
		assign data_doA_ways_masked[17:0] = data_doA_ways[18*(i+1)-1:18*i];
	else
		assign data_doA_ways_masked[18*(i+1)-1:18*i] = (way_select_ui[i]) ? data_doA_ways[18*(i+1)-1:18*i] : 18'b0;
end

// use one-hot-coded hit signal to select status and data
orgate #(
	.OR_WIDTH(WAYS),
	.BUS_WIDTH(3)
) status_or (
	.a(status_ways_masked),
	.y(status)
);
orgate #(
	.OR_WIDTH(WAYS),
	.BUS_WIDTH(18)
) data_doA_or (
	.a(data_doA_ways_masked),
	.y(data_doA)
);
orgate #(
	.OR_WIDTH(WAYS),
	.BUS_WIDTH(18-ADDRWIDTH_DATA)
) tag_or (
	.a(tag_masked),
	.y(tag_writeback)
);

if(WAYS_BITS > 0) begin
	// calculcate way index from lru tree
	assign lru_position[WAYS_BITS-1] = lru_data[0];
	for(i=1; i<WAYS_BITS; i=i+1) begin : lru_tree
		assign lru_position[WAYS_BITS-1-i] = lru_data[(1<<i)-1 + lru_position[WAYS_BITS-1:WAYS_BITS-i]];
	end

	// set accessed indices to opposite value (from way_select_ui)
	for(i=0; i<WAYS_BITS; i=i+1) begin : lru_reverse
		for(j=0; j<(1<<i); j=j+1) begin : lru_reverse_row
			wire [WAYS-1:0] mask_upper, mask_lower;
			assign mask_lower = {{WAYS>>(i+1)}{1'b1}} << (2*j*(WAYS>>(i+1)));
			assign mask_upper = {{WAYS>>(i+1)}{1'b1}} << ((2*j+1)*(WAYS>>(i+1)));
			assign lru_data_accessed[(1<<i)-1 + j] = (|(way_select_ui & mask_upper)) ? 1'b0 : (|(way_select_ui & mask_lower)) ? 1'b1 : lru_data[(1<<i)-1 + j];
		end
	end

	// find first invalid cache
	assign first_invalid_way[0] = !valid_ways[0];
	for(i=WAYS-1; i>0; i=i-1) begin : invalid_way
		assign first_invalid_way[i] = !valid_ways[i] && !(|first_invalid_way[i-1:0]);
	end
end else begin
	// only one way
	assign lru_position = 0;
	assign lru_data_accessed = 0;
	assign first_invalid_way = 0;
end
if(WAYS_BITS > 0) begin
	// pseudo lru data
	dp_ram #(
		.WIDTH(WAYS-1),
		.ADDRWIDTH(SET_BITS),
		.SIZE(SETS)
	) lru (
		.clk(clk_peri),
		.enA(1'b1),
		.enB(1'b0),
		.weA(lru_we),
		.addrA(addr_tag),
		.diA(lru_write_data),
		.doA(lru_data),
		.addrB(),
		.diB(),
		.doB(),
		.weB()
	);
end
endgenerate

assign stall = (state == CHECK_HIT && !hit_reg) || (state == CHECK_HIT && write_transaction && shared) || stall_reg;
assign stall_wr = stall || (state == IDLE || !write_transaction)  && !mem_access_reg;
assign dirty = status[2];
assign valid = status[1];
assign shared = status[0];
assign valid_all = &valid_ways;

wire read_ready;
wire read_enable;
wire write_ready;
wire write_enable;
assign read_enable = state == LOAD_CACHELINE && !hit_reg;
assign write_enable = state == WRITE_CACHELINE && dirty; // no need to check for hit

axi_read #(
	.LINE_SIZE(LINE_SIZE),
	.WAYS(WAYS),
	.SETS(SETS)
) reader (
	.reset(reset),
	.clk(clk_peri),
	.enable(read_enable),
	.ready(read_ready),
	.addr({addr_reg[17:LINE_BITS], {LINE_BITS+2{1'b0}}}),
	.data_we(write_we),
	.data_addr(write_addr),
	.data(write_data_axi),

	.ARADDR(ARADDR),
	.ARLEN(ARLEN),
	.ARVALID(ARVALID),
	.ARREADY(ARREADY),
	.ARID(ARID),
	.ARBURST(ARBURST),
	.ARSIZE(ARSIZE),

	.RRESP(RRESP),
	.RID(RID),
	.RDATA(RDATA),
	.RLAST(RLAST),
	.RVALID(RVALID),
	.RREADY(RREADY)
);

axi_write #(
	.LINE_SIZE(LINE_SIZE),
	.WAYS(WAYS),
	.SETS(SETS)
) writer (
	.reset(reset),
	.clk(clk_peri),
	.enable(write_enable),
	.ready(write_ready),
	.addr({tag_writeback_sync, addr_reg[ADDRWIDTH_DATA-1:LINE_BITS], {LINE_BITS+2{1'b0}}}),
	.data_addr(read_addr),
	.data(data_doA),

	.AWADDR(AWADDR),
	.AWLEN(AWLEN),
	.AWVALID(AWVALID),
	.AWREADY(AWREADY),
	.AWID(AWID),
	.AWBURST(AWBURST),
	.AWSIZE(AWSIZE),

	.WDATA(WDATA),
	.WSTRB(WSTRB),
	.WLAST(WLAST),
	.WVALID(WVALID),
	.WREADY(WREADY),

	.BID(BID),
	.BVALID(BVALID),
	.BREADY(BREADY),
	.BRESP(BRESP)
);

coherence_slave #(
	.LINE_SIZE(LINE_SIZE),
	.LINE_BITS(LINE_BITS),
	.WAYS(WAYS),
	.SET_BITS(SET_BITS),
	.ADDRWIDTH_DATA(ADDRWIDTH_DATA)
) coherence (
	.reset(reset),
	.clk(clk_peri),
	.last_i(last_i),
	.last_o(last_o),
	.request(request),
	.grant(grant),
	.addr_i(addr_i),
	.addr_o(addr_o),
	.addr_valid_i(addr_valid_i),
	.addr_valid_o(addr_valid_o),
	.data_i(data_i),
	.data_o(data_o),
	.hit_i(hit_i),
	.hit_o(hit_o),
	.type_i(type_i),
	.type_o(type_o),
	.hold_i(hold_i),
	.hold_o(hold_o),
	
	.status_addr(status_addr),
	.status_we(status_weB),
	.status_di(status_di),
	.status_do(status_do),
	.tag_addr(tag_addr),
	.tag_we(tag_weB),
	.tag_di(tag_di),
	.tag_do(tag_do),
	.data_addr(data_addr),
	.data_we(data_we),
	.data_di(data_di),
	.data_do(data_do),
	.way_select(way_select_coherence),

	.ui_miss(coherence_request),
	.ui_write_notification(coherence_write_notification),
	.ui_write_req(write_transaction),
	.ui_address(addr_reg),
	.ui_state(state),
	.ui_coherence_state(coherence_state)
);

always @(*) begin
	set_new_state = MOESI_NO_CHANGE;
	set_addr_reg = 0;
	set_stall = 0;
	set_write = write_transaction; // is it a write transaction?
	set_byte_wide_write = 0;
	set_data_in = 0;
	store_tag = 0;
	lru_new_way = 0;
	set_coherence = 0;
	set_notify_write_coherence = 0;
	set_doA_reg = 0;
	set_lru_access = 0;
	case(state)
		RESET: next_state = IDLE; // TODO invalidate all entries
		IDLE:
			if(select) begin
				next_state = CHECK_HIT;
				set_addr_reg = 1;
				set_write = 0;
				set_doA_reg = 1;
				if(store_access) begin
					set_write = 1;
					set_byte_wide_write = 1;
					set_data_in = 1;
					set_new_state = MOESI_MODIFIED;// if it is going to be a hit then it is the correct state
				end
			end else if(last_state == LOAD_DATA) begin
				// set doA_reg to keep output value for more than one clock cycle (for accelerator)
				set_doA_reg = 1;
				next_state = IDLE;
			end else next_state = IDLE;
		CHECK_HIT: begin
			set_lru_access = 1;
			if(!hit_reg && |way_select_ui) begin
				// not a hit but tag already exists -> invalid entry
				next_state = CHECK_COHERENCE;
				set_stall = 1;
				set_new_state = MOESI_INVALID;
				set_coherence = 1;
			end else if(!hit_reg) begin// && (!write_transaction || shared)) begin // if write and shared -> notify other caches
				// not a hit and tag not present -> have to find a new place
				if(!valid_all) begin
					// all caches used, but not all valid -> replace first invalid one
					if(WAYS_BITS > 0) begin
						lru_new_way = first_invalid_way;
					end else begin
						lru_new_way = 1;
					end
				end else begin
					// use way pointed to by lru tree
					lru_new_way = 1<<lru_position;
				end
				store_tag = 1;
				next_state = NEW_WAY;
				set_stall = 1;
			end else if(write_transaction && shared) begin // both MOESI_SHARED and MOESI_OWNED have to notify other caches
				// hit but have to invalidate entries in other caches
				set_stall = 1;
				set_notify_write_coherence = 1;
				next_state = NOTIFY_COHERENCE;
			end else begin
				set_doA_reg = 1;
				if(write_transaction) begin // write
					set_write = 0;
					next_state = IDLE;
				end else
				if(select) begin  // next cache access
					// not second time checking for write hit
					next_state = CHECK_HIT;
					set_addr_reg = 1;
					set_write = 0;
					if(store_access) begin
						set_write = 1;
						set_byte_wide_write = 1;
						set_data_in = 1;
						set_new_state = MOESI_MODIFIED;
					end
				end else next_state = IDLE;
			end
			end
		NEW_WAY: begin
				next_state = CHECK_COHERENCE;
				set_stall = 1;
				set_new_state = MOESI_INVALID;
				set_coherence = 1;
				if(valid_all && |(dirty_ways && lru_new_way_sync)) begin
					set_new_state = MOESI_INVALID_DIRTY;
					next_state = WRITE_CACHELINE;
					set_coherence = 0;
				end
			end
		LOAD_CACHELINE: 
			begin
				if(!read_ready) begin
					set_new_state = (write_transaction) ? MOESI_MODIFIED : MOESI_EXCLUSIVE;
				end
				if(read_ready) begin
					next_state = LOAD_DATA;
					if(!write_transaction) begin
						set_stall = 1; // one more stall for reading out data
					end
				end else begin
					set_stall = 1;
					next_state = LOAD_CACHELINE;
				end
			end
		LOAD_DATA:
			begin
				if(!write_transaction && select && store_access) begin
					// read and now a write
					set_stall = 1;
				end
				set_write = 0;
				next_state = IDLE;
			end
		WRITE_CACHELINE: begin
				set_stall = 1;
				if(write_ready) begin
					// remove dirty bit, don't have to notify others as everyone is shared
					set_new_state = MOESI_INVALID;
					set_coherence = 1;
					next_state = CHECK_COHERENCE;
				end else begin
					next_state = WRITE_CACHELINE;
				end
			end
		CHECK_COHERENCE: begin
				if(coherence_state == 2'b11) begin // coherence miss
					next_state = LOAD_CACHELINE;
					set_stall = 1;
				end else if(coherence_state == 2'b01) begin // coherence hit
					next_state = LOAD_DATA;
					set_stall = !write_transaction;
				end else begin
					next_state = CHECK_COHERENCE;
					set_stall = 1;
				end
			end
		NOTIFY_COHERENCE: begin
				if(coherence_state[1])
					next_state = LOAD_DATA;
				else begin
					set_stall = 1;
					set_notify_write_coherence = 1;
					next_state = NOTIFY_COHERENCE;
				end
			end
		default: next_state = RESET;
	endcase
end

always @(posedge clk_peri) begin
	if(reset) begin
		state <= RESET;
		store_low_reg <= 1'b0;
		store_high_reg <= 1'b0;
		coherence_request <= 1'b0;
		coherence_write_notification <= 1'b0;
		addr_reg <= 18'b0; // 0 is invalid address as it is location of local firmware
		tag_writeback_sync <= 0;
	end else begin
		state <= next_state;
		last_state <= state;
		tag_write_data <= addr_reg[17:ADDRWIDTH_DATA];
		mem_access_reg <= store_access && !select && write_transaction;
		if(set_new_state != MOESI_NO_CHANGE) begin
			status_we <= 1'b1;
			status_write_data <= set_new_state;
		end else begin
			status_we <= 1'b0;
		end
		if(set_stall) begin
			stall_reg <= 1'b1;
		end else begin
			stall_reg <= 1'b0;
		end
		if(set_addr_reg) begin
			addr_reg <= addr;
		end
		if(set_write) begin
			write_transaction <= 1'b1;
		end else begin
			write_transaction <= 1'b0;
		end
		if(set_byte_wide_write) begin
			store_low_reg <= store_access_low;
			store_high_reg <= store_access_high;
		end
		if(set_data_in) begin
			data_in <= do_peri;
		end
		if(store_tag) begin
			tag_we <= 1'b1;
			lru_new_way_sync <= lru_new_way;
			tag_writeback_sync <= tag_writeback;
		end else begin
			tag_we <= 1'b0;
			lru_new_way_sync <= 0;
		end
		if(set_coherence) begin
			coherence_request <= 1'b1;
		end else begin
			coherence_request <= 1'b0;
		end
		if(set_notify_write_coherence) begin
			coherence_write_notification <= 1'b1;
		end else begin
			coherence_write_notification <= 1'b0;
		end
		if(set_doA_reg) begin
			data_doA_reg <= data_doA;
		end
		if(set_lru_access) begin
			lru_we <= 1'b1;
			lru_write_data <= lru_data_accessed;
		end else begin
			lru_we <= 1'b0;
		end
	end
end

// hit / miss counter
reg[31:0] hit, coherence_hit, miss, writeback;

always @(posedge clk_peri) begin
	if(reset) begin
		hit <= 0;
		coherence_hit <= 0;
		miss <= 0;
		writeback <= 0;
	end else begin
		if(state == CHECK_HIT && hit_reg) begin
			hit <= hit + 1;
		end else if(state == CHECK_COHERENCE && coherence_state == 2'b01) begin
			coherence_hit <= coherence_hit + 1;
		end else if(state == CHECK_COHERENCE && coherence_state == 2'b11) begin
			// coherence miss => total miss
			miss <= miss + 1;
		end else if(state === WRITE_CACHELINE && write_ready) begin
			writeback <= writeback + 1;
		end
	end
end

endmodule
