/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale 1ns / 10ps

module axi_read #(
	parameter LINE_SIZE = 4,
	parameter WAYS = 1,
	parameter SETS = 4
) (
	// user interface
	input  wire        reset,
	input  wire        clk,
	input  wire        enable,
	output wire        ready,
	input  wire [19:0] addr,
	output wire        data_we,
	output reg  [17:0] data_addr, // log2(LINE_SIZE)
	output wire [17:0] data,

	// AXI interface
	output wire [31:0] ARADDR,
	output wire [7:0]  ARLEN,
	output wire        ARVALID,
	input  wire        ARREADY,
	output wire [3:0]  ARID,
	output wire [1:0]  ARBURST,
	output wire [2:0]  ARSIZE,

	input  wire [1:0]  RRESP, // ignore for now
	input  wire [3:0]  RID,
	input  wire [31:0] RDATA,
	input  wire        RLAST,
	input  wire        RVALID,
	output wire        RREADY
);

function integer log2;
	input integer value;
	begin
		value = value - 1;
		for (log2=0; value>0; log2=log2+1) begin
			value = value>>1;
		end
	end
endfunction

localparam ADDRWIDTH_DATA = log2(SETS)+log2(LINE_SIZE);
localparam LINE_BITS = log2(LINE_SIZE);
localparam SET_BITS = log2(SETS);

// Read Address Channel

reg [1:0] address_state;
localparam
	AR_IDLE = 0,
	AR_ADDRESS = 1,
	AR_RESPONSE = 2;

assign ready = RLAST && RVALID;

assign ARADDR = {12'b0,addr};
assign ARLEN = LINE_SIZE-1;
assign ARVALID = (address_state == AR_ADDRESS);
assign ARID = 4'b0000;
assign ARBURST = 2'b01; // incremental burst
assign ARSIZE = 3'b010; // 32 bit word size on AXI connection

always @(posedge clk) begin
	if(reset) begin
		address_state <= AR_IDLE;
	end else begin
		case(address_state)
			AR_IDLE: if(enable) begin
					address_state <= AR_ADDRESS;
				end else begin
					address_state <= AR_IDLE;
				end
			AR_ADDRESS: if(ARREADY) begin
					address_state <= AR_RESPONSE;
				end else address_state <= AR_ADDRESS;
			AR_RESPONSE: if(RLAST) address_state <= AR_IDLE;
				else address_state <= AR_RESPONSE;
			default: address_state <= AR_IDLE;
		endcase
	end
end

// Read Data Channel

reg [0:0] read_state;
localparam
	R_IDLE = 0,
	R_READ = 1;

assign data = RDATA[17:0];
assign data_we = RVALID && RREADY;

assign RREADY = (read_state == R_READ);

always @(posedge clk) begin
	if(reset) begin
		read_state <= R_IDLE;
	end else begin
		case(read_state)
			R_IDLE: 
				begin
					if(enable) read_state <= R_READ;
					else read_state <= R_IDLE;
					data_addr <= addr[19:2];
				end
			R_READ: 
				begin
					if(RVALID && RLAST) begin
						read_state <= R_IDLE;
					end else if(RVALID) begin
						data_addr <= data_addr + 1;
						read_state <= R_READ;
					end else read_state <= R_READ;
				end
			default: read_state <= R_IDLE;
		endcase
	end
end

endmodule
