/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale 1ns / 10ps

module axi_write #(
	parameter LINE_SIZE = 4,
	parameter WAYS = 1,
	parameter SETS = 4
) (
	// user interface
	input  wire        reset,
	input  wire        clk,
	input  wire        enable,
	output wire        ready,
	input  wire [19:0] addr,
	output reg  [17:0] data_addr, // log2(LINE_SIZE)
	input  wire [17:0] data,

	// AXI interface
	output wire [31:0] AWADDR,
	output wire [7:0]  AWLEN,
	output wire        AWVALID,
	input  wire        AWREADY,
	output wire [3:0]  AWID,
	output wire [1:0]  AWBURST,
	output wire [2:0]  AWSIZE,

	output wire [31:0] WDATA,
	output wire [3:0]  WSTRB,
	output wire        WLAST,
	output wire        WVALID,
	input  wire        WREADY,

	input  wire [3:0]  BID,
	input  wire        BVALID,
	output wire        BREADY,
	input  wire [1:0]  BRESP
);

function integer log2;
	input integer value;
	begin
		value = value - 1;
		for (log2=0; value>0; log2=log2+1) begin
			value = value>>1;
		end
	end
endfunction

localparam ADDRWIDTH_DATA = log2(SETS)+log2(LINE_SIZE);
localparam LINE_BITS = log2(LINE_SIZE);
localparam SET_BITS = log2(SETS);

reg [LINE_BITS-1:0] word_counter;
reg [17:0] data_reg;

// Write Address Channel

reg [1:0] address_state;
localparam
	AW_IDLE = 0,
	AW_ADDRESS = 1,
	AW_RESPONSE = 2;

assign AWADDR = {12'b0,addr};
assign AWLEN = LINE_SIZE-1;
assign AWVALID = (address_state == AW_ADDRESS);
assign AWID = 4'b0;
assign AWBURST = 2'b01; // incremental burst
assign AWSIZE = 3'b010; // 32 bit word size on AXI connection

always @(posedge clk) begin
	if(reset) begin
		address_state <= AW_IDLE;
	end else begin
		case(address_state)
			AW_IDLE: if(enable) begin
					address_state <= AW_ADDRESS;
				end else begin
					address_state <= AW_IDLE;
				end
			AW_ADDRESS: if(AWREADY) begin
					address_state <= AW_RESPONSE;
				end else address_state <= AW_ADDRESS;
			AW_RESPONSE: if(BREADY && BVALID) address_state <= AW_IDLE;
				else address_state <= AW_RESPONSE;
			default: address_state <= AW_IDLE;
		endcase
	end
end

// Write Data Channel

reg [1:0] write_state, last_write_state;
localparam
	W_IDLE = 0,
	WRITE_WAIT = 1,
	WRITE = 2,
	W_RESPONSE = 3;

assign ready = (write_state == W_RESPONSE && BVALID);

assign WDATA = (write_state == WRITE && last_write_state == WRITE) ? {14'b0, data} : {14'b0, data_reg};
assign WSTRB = 4'b1111;
assign WLAST = (word_counter == LINE_SIZE-1);
assign WVALID = (write_state == WRITE || write_state == WRITE_WAIT);

assign BREADY = (write_state == W_RESPONSE);

always @(posedge clk) begin
	if(reset) begin
		write_state <= W_IDLE;
		word_counter <= 0;
		data_addr <= 0;
		data_reg <= 0;
	end else begin
		last_write_state <= write_state;
		if(WVALID && WREADY) begin
			data_addr <= data_addr + 1;
			data_reg <= data;
			word_counter <= word_counter + 1;
		end
		case(write_state)
			W_IDLE: begin
					if(enable && AWREADY && AWVALID) begin
						write_state <= WRITE;
						data_addr <= addr[19:2] + 1;
					end else begin
						write_state <= W_IDLE;
						data_addr <= addr[19:2];
					end
					word_counter <= 0;
					data_reg <= data;
				end
			WRITE_WAIT: if(WREADY && WLAST) write_state <= W_RESPONSE; // Burst length of 0
				else if(WREADY) write_state <= WRITE;
				else write_state <= WRITE_WAIT;
			WRITE: if(WREADY && WLAST) write_state <= W_RESPONSE;
				else if(WREADY) write_state <= WRITE;
				else begin
					write_state <= WRITE_WAIT;
					data_reg <= data;
				end
			W_RESPONSE: if(BVALID) write_state <= W_IDLE;
				else write_state <= W_RESPONSE;
			default: write_state <= W_IDLE;
		endcase
	end
end

endmodule
