/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale 1ns / 10ps

module coherence_slave #(
	parameter LINE_SIZE = 4,
	parameter WAYS = 1,
	parameter LINE_BITS = 2,
	parameter SET_BITS = 2,
	parameter ADDRWIDTH_DATA = 4
) (
	input  wire                clk,
	input  wire                reset,
	output wire                last_i,
	input  wire                last_o,
	output wire                request,
	input  wire                grant,

	// coherence bus
	output reg  [17:0]         addr_i,
	input  wire [17:0]         addr_o,
	output reg                 addr_valid_i,
	input  wire                addr_valid_o,
	output wire [17:0]         data_i,
	input  wire [17:0]         data_o,
	output wire                hit_i,
	input  wire                hit_o,
	output reg  [1:0]          type_i, // read or write request, write notification
	input  wire [1:0]          type_o,
	output wire                hold_i,
	input  wire                hold_o,

	// blockram access
	output wire [SET_BITS-1:0] status_addr,
	output reg                 status_we,
	output reg  [2:0]          status_di,
	input  wire [3*WAYS-1:0]   status_do,
	output wire [SET_BITS-1:0] tag_addr,
	output reg                 tag_we,
	output reg  [18-ADDRWIDTH_DATA-1:0] tag_di,
	input  wire [WAYS*(18-ADDRWIDTH_DATA)-1:0] tag_do,
	output wire [ADDRWIDTH_DATA-1:0] data_addr,
	output wire                data_we,
	output wire [17:0]         data_di,
	input  wire [18*WAYS-1:0]  data_do,
	output wire [WAYS-1:0]     way_select,

	// user interface access
	input  wire                ui_miss,
	input  wire                ui_write_notification,
	input  wire                ui_write_req,
	input  wire [17:0]         ui_address,
	input  wire [3:0]          ui_state,
	output wire [1:0]          ui_coherence_state
);

localparam LINE_BITS_NOT_NULL = LINE_BITS == 0 ? 1 : LINE_BITS;

// keep in sync with axi_cache
localparam
	UI_RESET           = 0,
	UI_IDLE            = 1,
	UI_CHECK_HIT       = 2,
	UI_LOAD_CACHELINE  = 3,
	UI_LOAD_DATA       = 6,
	UI_WRITE_CACHELINE = 4,
	UI_CHECK_COHERENCE = 5;

// keep in sync with axi_cache
localparam
	MOESI_MODIFIED  = 3'b110,
	MOESI_OWNED     = 3'b111,
	MOESI_EXCLUSIVE = 3'b010,
	MOESI_SHARED    = 3'b011,
	MOESI_INVALID   = 3'b000,
	MOESI_INVALID_DIRTY = 3'b100,
	MOESI_NO_CHANGE = 3'b001;
	// status bits: dirty valid shared

reg [17:0] bus_addr_reg;
reg [17:0] base_addr_reg;
reg last_i_reg;
wire [WAYS*3-1:0] status_value_masked;
wire [2:0] status_value;
wire [WAYS*18-1:0] cache_data_masked;
wire [17:0] cache_data;
wire hit;
wire [WAYS-1:0] valid_masked;
wire valid;
wire [17:0] address, base_addr;
generate
if(LINE_BITS == 0) begin
	assign address = (addr_valid_o) ? {addr_o[17:LINE_BITS]} : bus_addr_reg;
	assign base_addr = (addr_valid_o) ? {addr_o[17:LINE_BITS]} : {base_addr_reg[17:LINE_BITS]};
end else begin
	assign address = (addr_valid_o) ? {addr_o[17:LINE_BITS],{LINE_BITS{1'b0}}} : bus_addr_reg;
	assign base_addr = (addr_valid_o) ? {addr_o[17:LINE_BITS],{LINE_BITS{1'b0}}} : {base_addr_reg[17:LINE_BITS],{LINE_BITS{1'b0}}};
end

if(SET_BITS > 0) begin
	assign status_addr = base_addr[ADDRWIDTH_DATA-1:LINE_BITS];
	assign tag_addr = base_addr[ADDRWIDTH_DATA-1:LINE_BITS];
end else begin
	assign status_addr = 0;
	assign tag_addr = 0;
end
if(ADDRWIDTH_DATA > 0) begin
	assign data_addr = address[ADDRWIDTH_DATA-1:0];
end else begin
	assign data_addr = 0;
end
assign hit = |way_select && valid;
assign valid = |valid_masked;

genvar i;
for(i=0; i<WAYS; i=i+1) begin : hit_ways
	assign way_select[i] = tag_do[(i+1)*(18-ADDRWIDTH_DATA)-1:i*(18-ADDRWIDTH_DATA)] == base_addr[17:ADDRWIDTH_DATA]; // status_do[1] == valid
	assign status_value_masked[3*(i+1)-1:3*i] = (way_select[i]) ? status_do[3*(i+1)-1:3*i] : 3'b0;
	assign cache_data_masked[18*(i+1)-1:18*i] = (way_select[i]) ? data_do[18*(i+1)-1:18*i] : 18'b0;
	assign valid_masked[i] = (way_select[i]) ? status_do[3*i+1] : 1'b0;
end
endgenerate

orgate #(
	.OR_WIDTH(WAYS),
	.BUS_WIDTH(3)
) status_value_or (
	.a(status_value_masked),
	.y(status_value)
);

orgate #(
	.OR_WIDTH(WAYS),
	.BUS_WIDTH(18)
) cache_data_or (
	.a(cache_data_masked),
	.y(cache_data)
);

// Listening state machine
reg [2:0] state, next_state;
localparam
	WAIT_REQUEST = 0,
	CHECK_HIT    = 1,
	HOLD_REQUEST = 3,
	TRANSFER_DATA = 2,
	CHECK_HIT_INVALIDATE = 4;
// UI state machine
reg [2:0] coherence_state, coherence_next_state;
reg set_coherence_miss;
localparam
	COHERENCE_IDLE = 4,
	REQUEST     = 0,
	GRANTED     = 1,
	CHECK_RESPONSE = 5,
	CHECK_RESPONSE2 = 6,
	HOLD        = 7,
	RECEIVE     = 2,
	NOT_AVAILABLE = 3;

// put signals on coherence bus
assign hit_i = (state == CHECK_HIT && hit);
assign hold_i = (hit_i && ui_state == UI_LOAD_CACHELINE && ui_address[17:LINE_BITS] == address[17:LINE_BITS]) || (state == HOLD_REQUEST);
assign data_i = (hit_i || state == TRANSFER_DATA || last_i) ? cache_data : 18'b0;
assign last_i = (LINE_BITS == 0) ? (hit_i || last_i_reg) : last_i_reg;

reg set_last;
reg[2:0] set_new_state;

generate
always @(*) begin
	set_last = 0;
	set_new_state = MOESI_NO_CHANGE;
	case(state)
		WAIT_REQUEST:
			if(addr_valid_o && !addr_valid_i && type_o[1]) next_state = CHECK_HIT_INVALIDATE;
			else if(addr_valid_o && !addr_valid_i) next_state = CHECK_HIT;
			else next_state = WAIT_REQUEST;
		CHECK_HIT:
			if(hit && ui_state == UI_LOAD_CACHELINE && ui_address[17:LINE_BITS] == address[17:LINE_BITS]) begin
				next_state = HOLD_REQUEST;
			end else if(hit) begin
				if(LINE_BITS == 0) begin
					next_state = WAIT_REQUEST;
				end else begin
					if(bus_addr_reg[LINE_BITS_NOT_NULL-1:0] == {LINE_BITS_NOT_NULL{1'b1}}) begin
						set_last = 1;
						next_state = WAIT_REQUEST;
					end else begin
						next_state = TRANSFER_DATA;
					end
				end
				if(type_o[0]) begin // write request
					set_new_state = MOESI_INVALID;
				end else begin // read request
					case(status_value)
						MOESI_MODIFIED: set_new_state = MOESI_OWNED;
						MOESI_EXCLUSIVE: set_new_state = MOESI_SHARED;
						default: set_new_state = status_value; // no state transition for MOESI_SHARED and MOESI_OWNED
					endcase
				end
			end else next_state = WAIT_REQUEST;
		CHECK_HIT_INVALIDATE: begin
				next_state = WAIT_REQUEST;
				if(hit) begin
					set_new_state = MOESI_INVALID;
				end
			end
		HOLD_REQUEST:
			if(ui_state != UI_LOAD_CACHELINE) next_state = WAIT_REQUEST;
			else next_state = HOLD_REQUEST;
		TRANSFER_DATA:
			if(LINE_BITS > 0) begin
				if(bus_addr_reg[LINE_BITS_NOT_NULL-1:0] == {LINE_BITS_NOT_NULL{1'b1}}) begin
					set_last = 1;
					next_state = WAIT_REQUEST;
				end else next_state = TRANSFER_DATA;
			end
		default:
			next_state = state;
	endcase
end

always @(posedge clk) begin
	if(reset) begin
		state <= WAIT_REQUEST;
	end else begin
		state <= next_state;
		last_i_reg <= 1'b0;
		status_we <= 1'b0;
		tag_we <= 1'b0;
		if(addr_valid_o && (coherence_state == COHERENCE_IDLE || coherence_state == REQUEST)) begin
			// address for reading out requested cache line
			if(LINE_BITS == 0)
				bus_addr_reg <= addr_o[17:LINE_BITS];
			else
				bus_addr_reg <= {addr_o[17:LINE_BITS],{LINE_BITS{1'b0}}} + 1;
			base_addr_reg <= addr_o;
		end else if(addr_valid_o) begin
			// address for storing needs to be one clock after reading
			if(LINE_BITS == 0)
				bus_addr_reg <= addr_o[17:LINE_BITS];
			else
				bus_addr_reg <= {addr_o[17:LINE_BITS],{LINE_BITS{1'b0}}};
			base_addr_reg <= addr_o;
		end else begin
			bus_addr_reg <= bus_addr_reg + 1;
		end
		if(set_last || set_coherence_miss) begin
			last_i_reg <= 1'b1;
		end
		if(LINE_BITS == 0) begin
			if(coherence_state == RECEIVE || coherence_state == CHECK_RESPONSE && hit_o && !hold_o) begin
				// set valid
				status_we <= 1'b1;
				status_di <= ui_write_req ? MOESI_MODIFIED : MOESI_SHARED;
				// set tag
				tag_we <= 1'b1;
				tag_di <= address[17:ADDRWIDTH_DATA];
			end else if(set_new_state != MOESI_NO_CHANGE) begin
				status_we <= 1'b1;
				status_di <= set_new_state;
			end
		end else begin
			if((coherence_state == RECEIVE || coherence_state == CHECK_RESPONSE) && (bus_addr_reg[LINE_BITS_NOT_NULL-1:0]+1) == {LINE_BITS_NOT_NULL{1'b1}}) begin // same cycle as last_o is active
				// set valid
				status_we <= 1'b1;
				status_di <= ui_write_req ? MOESI_MODIFIED : MOESI_SHARED;
				// set tag
				tag_we <= 1'b1;
				tag_di <= address[17:ADDRWIDTH_DATA];
			end else if(set_new_state != MOESI_NO_CHANGE) begin
				status_we <= 1'b1;
				status_di <= set_new_state;
			end
		end
	end
end
endgenerate

// UI logic
assign request = (coherence_state == COHERENCE_IDLE && (ui_miss || ui_write_notification)) || (coherence_state == REQUEST);
assign data_we = (coherence_state == CHECK_RESPONSE && hit_o) || (coherence_state == RECEIVE);
assign data_di = data_o;
assign ui_coherence_state[1] = coherence_state == NOT_AVAILABLE || addr_valid_i;
assign ui_coherence_state[0] = coherence_state == CHECK_RESPONSE && last_o && !hold_o || coherence_state == RECEIVE && last_o && !hold_o || last_i && grant;

always @(*) begin
	set_coherence_miss = 0;
	case(coherence_state)
		COHERENCE_IDLE: if(ui_miss || ui_write_notification) coherence_next_state = REQUEST;
			else coherence_next_state = COHERENCE_IDLE;
		REQUEST: if(grant) coherence_next_state = GRANTED;
			else coherence_next_state = REQUEST;
		GRANTED: if(ui_write_notification) begin
				coherence_next_state = NOT_AVAILABLE; // one wait cycle
				set_coherence_miss = 1; // for sending last (releases grant)
			end else coherence_next_state = CHECK_RESPONSE;
		CHECK_RESPONSE: if(last_o && !hold_o) begin
				coherence_next_state = COHERENCE_IDLE;
			end else if(hold_o) coherence_next_state = HOLD;
			else if(hit_o) coherence_next_state = RECEIVE;
			else begin
				set_coherence_miss = 1;
				coherence_next_state = NOT_AVAILABLE;
			end
		HOLD: if(hold_o) coherence_next_state = HOLD;
			else coherence_next_state = REQUEST;
		RECEIVE: if(last_o) begin
				coherence_next_state = COHERENCE_IDLE;
			end else coherence_next_state = RECEIVE;
		NOT_AVAILABLE: coherence_next_state = COHERENCE_IDLE;
		default: coherence_next_state = coherence_state;
	endcase
end

always @(posedge clk) begin
	coherence_state <= coherence_next_state;
	addr_i <= 18'b0;
	addr_valid_i <= 1'b0;
	type_i <= 2'b00;
	if(reset) begin
		coherence_state <= COHERENCE_IDLE;
	end
	case(coherence_state)
		REQUEST: if(grant) begin
				addr_i <= {ui_address[17:LINE_BITS],{LINE_BITS{1'b0}}};
				addr_valid_i <= 1'b1;
				type_i <= {ui_write_notification, ui_write_req};
			end
		GRANTED:
			type_i <= {ui_write_notification, ui_write_req};
		CHECK_RESPONSE:
			type_i <= {ui_write_notification, ui_write_req}; // keep signal for more than one cycle
		default: type_i <= 2'b00;
	endcase
end

endmodule
