/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale 1ns / 10ps

module spartanmc_mem_cache #(
	parameter BASE_ADDR = 0,
	parameter LOCAL_RAMBLOCKS = 4,
	parameter LOCAL_SHOW_MEM = 1,

	parameter LINE_SIZE = 4, // Number of words in a cache line
	parameter WAYS = 1, // TODO Number of ways to describe associativity
	parameter SETS = 4, // Number of sets (if WAYS=1 number of cache lines)
	parameter DATA_RAMBLOCKS = 50,

	parameter SECOND_PORT = 0
) (
	// Code memory bus
	input  wire        clk,
	input  wire        reset,
	input  wire        code_access,
	input  wire        code_jmp,
	input  wire [17:0] code_addr,
	output wire [17:0] code_do,
	output wire        code_busy,

	// Memory bus
	input  wire        mem_access,
	input  wire        mem_wr_low,
	input  wire        mem_wr_high,
	input  wire [9:0]  mem_addr_block,
	input  wire [7:0]  mem_addr_high,
	input  wire [17:0] mem_di,
	output wire [17:0] mem_do,
	output wire        mem_busy,
	output wire        mem_busy_wr,

	// Cache Memory bus
	input  wire        cache_clk,
	input  wire        cache_reset,
	input  wire        cache_access,
	input  wire        cache_wr_low,
	input  wire        cache_wr_high,
	input  wire [9:0]  cache_addr_block,
	input  wire [7:0]  cache_addr_high,
	input  wire [17:0] cache_di,
	output wire [17:0] cache_do,
	//output wire        cache_busy,

	// AXI interface
	// Read
	output wire [31:0] ARADDR,
	output wire [7:0]  ARLEN,
	output wire        ARVALID,
	input  wire        ARREADY,
	output wire [3:0]  ARID,
	output wire [1:0]  ARBURST,
	output wire [2:0]  ARSIZE,

	input  wire [1:0]  RRESP,
	input  wire [3:0]  RID,
	input  wire [31:0] RDATA,
	input  wire        RLAST,
	input  wire        RVALID,
	output wire        RREADY,

	// Write
	output wire [31:0] AWADDR,
	output wire [7:0]  AWLEN,
	output wire        AWVALID,
	input  wire        AWREADY,
	output wire [3:0]  AWID,
	output wire [1:0]  AWBURST,
	output wire [2:0]  AWSIZE,

	output wire [31:0] WDATA,
	output wire [3:0]  WSTRB,
	output wire        WLAST,
	output wire        WVALID,
	input  wire        WREADY,

	input  wire [3:0]  BID,
	input  wire        BVALID,
	output wire        BREADY,
	input  wire [1:0]  BRESP,

	// Coherence bus
	output wire        last_i,
	input  wire        last_o,
	output wire        request,
	input  wire        grant,

	output wire [17:0] addr_i,
	input  wire [17:0] addr_o,
	output wire        addr_valid_i,
	input  wire        addr_valid_o,
	output wire [17:0] data_i,
	input  wire [17:0] data_o,
	output wire        hit_i,
	input  wire        hit_o,
	output wire [1:0]  type_i,
	input  wire [1:0]  type_o,
	output wire        hold_i,
	input  wire        hold_o
);

wire [17:0] mem_do_local, mem_do_cache, do_cache;
assign mem_do = mem_do_local | mem_do_cache;
assign cache_do = do_cache;

wire mem_busy_local, mem_busy_cache, mem_busy_wr_local, mem_busy_wr_cache;
assign mem_busy = mem_busy_local | mem_busy_cache;
assign mem_busy_wr = mem_busy_wr_local | mem_busy_wr_cache;

// local memory for code sections
spartanmc_mem_local #(
	.BASE_ADDR(BASE_ADDR),
	.RAMBLOCKS(LOCAL_RAMBLOCKS),
	.SHOW_MEM(LOCAL_SHOW_MEM)
) code_mem (
	.clk(clk),
	.reset(reset),
	.code_access(code_access),
	.code_jmp(code_jmp),
	.code_addr(code_addr),
	.code_do(code_do),
	.code_busy(code_busy),

	.mem_addr_block(mem_addr_block),
	.mem_addr_high(mem_addr_high),
	.mem_di(mem_di),
	.mem_wr_low(mem_wr_low),
	.mem_wr_high(mem_wr_high),
	.mem_access(mem_access && ~mem_busy_cache),
	.mem_do(mem_do_local),
	.mem_busy(mem_busy_local),
	.mem_busy_wr(mem_busy_wr_local)
);

wire access;
generate
if (SECOND_PORT) begin
	reg cache_access_reg;
	always @(posedge clk)
		cache_access_reg <= cache_access;
	assign access = cache_access;
	assign mem_do_cache = (cache_access_reg) ? 18'b0 : do_cache;
end else begin
	assign access = 1'b0;
	assign mem_do_cache = do_cache;
end
endgenerate

// axi cache for data sections
axi_cache #(
	.LINE_SIZE(LINE_SIZE),
	.WAYS(WAYS),
	.SETS(SETS),
	.DMA_BASE_ADR(BASE_ADDR + LOCAL_RAMBLOCKS*2048),
	.DATA_RAMBLOCKS(DATA_RAMBLOCKS)
) data_cache (
	.reset(reset),
	.clk_peri(clk),
	.mem_access(mem_access | access),
	// assumption: mem_access is not accessing cache when cache_access is asserted
	.store_access_low(access ? cache_wr_low : mem_wr_low),
	.store_access_high(access ? cache_wr_high : mem_wr_high),
	.addr_peri(access ? cache_addr_block : mem_addr_block),
	.addr_high(access ? cache_addr_high : mem_addr_high),
	.do_peri(access ? cache_di : mem_di), // input
	.di_peri(do_cache), // output
	.stall(mem_busy_cache),
	.stall_wr(mem_busy_wr_cache),

	// AXI interface
	// Read
	.ARADDR(ARADDR),
	.ARLEN(ARLEN),
	.ARVALID(ARVALID),
	.ARREADY(ARREADY),
	.ARID(ARID),
	.ARBURST(ARBURST),
	.ARSIZE(ARSIZE),

	.RRESP(RRESP),
	.RID(RID),
	.RDATA(RDATA),
	.RLAST(RLAST),
	.RVALID(RVALID),
	.RREADY(RREADY),

	// Write
	.AWADDR(AWADDR),
	.AWLEN(AWLEN),
	.AWVALID(AWVALID),
	.AWREADY(AWREADY),
	.AWID(AWID),
	.AWBURST(AWBURST),
	.AWSIZE(AWSIZE),
	
	.WDATA(WDATA),
	.WSTRB(WSTRB),
	.WLAST(WLAST),
	.WVALID(WVALID),
	.WREADY(WREADY),

	.BID(BID),
	.BVALID(BVALID),
	.BREADY(BREADY),
	.BRESP(BRESP),

	// Coherence bus
	.last_i(last_i),
	.last_o(last_o),
	.request(request),
	.grant(grant),

	.addr_i(addr_i),
	.addr_o(addr_o),
	.addr_valid_i(addr_valid_i),
	.addr_valid_o(addr_valid_o),
	.data_i(data_i),
	.data_o(data_o),
	.hit_i(hit_i),
	.hit_o(hit_o),
	.type_i(type_i),
	.type_o(type_o),
	.hold_i(hold_i),
	.hold_o(hold_o)
);

endmodule
