/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale 1ns / 10ps

module spartanmc_mem_cache_system #(
	parameter BASE_ADDR = 0, // constant
	parameter LOCAL_RAMBLOCKS = 1,
	parameter DATA_RAMBLOCKS = 1,
	parameter DISABLE_CACHE = 0,
	parameter SECOND_PORT = 0,
	parameter CACHES = 1,
	parameter ACC_CONS = CACHES+SECOND_PORT-1, // width of memory-bus connected to accelerator
	parameter LINE_SIZE = 4,
	parameter WAYS = 1,
	parameter SETS = 4,
	parameter LOCAL_SHOW_MEM = 0
) (
	// Code memory bus
	input  wire        clk,
	input  wire        reset,
	input  wire        code_access,
	input  wire        code_jmp,
	input  wire [17:0] code_addr,
	output wire [17:0] code_do,
	output wire        code_busy,

	// Memory bus
	input  wire        mem_access,
	input  wire        mem_wr_low,
	input  wire        mem_wr_high,
	input  wire [9:0]  mem_addr_block,
	input  wire [7:0]  mem_addr_high,
	input  wire [17:0] mem_di,
	output wire [17:0] mem_do,
	output wire        mem_busy,
	output wire        mem_busy_wr,

	// caches data-bus
	input  wire        cache_clk,
	input  wire        cache_reset,
	input  wire [ACC_CONS-1:0]    cache_access,
	input  wire [ACC_CONS-1:0]    cache_wr_low,
	input  wire [ACC_CONS-1:0]    cache_wr_high,
	input  wire [ACC_CONS*10-1:0] cache_addr_block,
	input  wire [ACC_CONS*8-1:0]  cache_addr_high,
	input  wire [ACC_CONS*18-1:0] cache_di,
	output wire [ACC_CONS*18-1:0] cache_do,
	output wire [ACC_CONS-1:0]    cache_mem_busy,
	output wire [ACC_CONS-1:0]    cache_mem_busy_wr,

	// AXI interface
	// Read
	output wire [31:0] ARADDR,
	output wire [7:0]  ARLEN,
	output wire        ARVALID,
	input  wire        ARREADY,
	output wire [3:0]  ARID,
	output wire [1:0]  ARBURST,
	output wire [2:0]  ARSIZE,

	input  wire [1:0]  RRESP,
	input  wire [3:0]  RID,
	input  wire [31:0] RDATA,
	input  wire        RLAST,
	input  wire        RVALID,
	output wire        RREADY,

	// Write
	output wire [31:0] AWADDR,
	output wire [7:0]  AWLEN,
	output wire        AWVALID,
	input  wire        AWREADY,
	output wire [3:0]  AWID,
	output wire [1:0]  AWBURST,
	output wire [2:0]  AWSIZE,

	output wire [31:0] WDATA,
	output wire [3:0]  WSTRB,
	output wire        WLAST,
	output wire        WVALID,
	input  wire        WREADY,

	input  wire [3:0]  BID,
	input  wire        BVALID,
	output wire        BREADY,
	input  wire [1:0]  BRESP
);

parameter INTERCONNECT_PORTS = 16; // use full size of interconnect, but only assign used wires

generate
if(DISABLE_CACHE == 1) begin : cache_disabled
	// Caches are disabled, use dma memory
	
	spartanmc_mem_acc_dma #(
		.BASE_ADDR(BASE_ADDR),
		.CODE_RAMBLOCKS(LOCAL_RAMBLOCKS),
		.DATA_RAMBLOCKS(DATA_RAMBLOCKS),
		.SHOW_MEM(LOCAL_SHOW_MEM)
	) dma (
		.clk(clk),
		.reset(reset),
		.code_access(code_access),
		.code_jmp(code_jmp),
		.code_addr(code_addr),
		.code_do(code_do),
		.code_busy(code_busy),

		.mem_access(mem_access),
		.mem_wr_low(mem_wr_low),
		.mem_wr_high(mem_wr_high),
		.mem_addr_block(mem_addr_block),
		.mem_addr_high(mem_addr_high),
		.mem_di(mem_di),
		.mem_do(mem_do),
		.mem_busy(mem_busy),
		.mem_busy_wr(mem_busy_wr),

		// accelerator memory-bus for first cache
		.cache_clk(cache_clk),
		.cache_reset(cache_reset),
		.cache_access(cache_access[0]),
		.cache_wr_low(cache_wr_low[0]),
		.cache_wr_high(cache_wr_high[0]),
		.cache_addr_block(cache_addr_block[9:0]),
		.cache_addr_high(cache_addr_high[7:0]),
		.cache_di(cache_di[17:0]),
		.cache_do(cache_do[17:0])
	);

	assign cache_mem_busy = mem_busy;
	assign cache_mem_busy_wr = mem_busy_wr;

	// disable axi outputs
	assign ARADDR = 32'b0;
	assign ARLEN = 8'b0;
	assign ARVALID = 1'b0;
	assign ARID = 4'b0;
	assign ARBURST = 2'b0;
	assign ARSIZE = 3'b0;
	assign RREADY = 1'b0;
	assign AWADDR = 32'b0;
	assign AWLEN = 8'b0;
	assign AWVALID = 1'b0;
	assign AWID = 4'b0;
	assign AWBURST = 2'b0;
	assign AWSIZE = 3'b0;
	assign WDATA = 32'b0;
	assign WSTRB = 4'b0;
	assign WLAST = 1'b0;
	assign WVALID = 1'b0;
	assign BREADY = 1'b0;

end else begin : cache_enabled
	// AXI wires
	// Read
	wire [INTERCONNECT_PORTS*32-1:0] axi_araddr;
	wire [INTERCONNECT_PORTS*8-1:0]  axi_arlen;
	wire [INTERCONNECT_PORTS-1:0]    axi_arvalid;
	wire [INTERCONNECT_PORTS-1:0]    axi_arready;
	wire [INTERCONNECT_PORTS*4-1:0]  axi_arid;
	wire [INTERCONNECT_PORTS*2-1:0]  axi_arburst;
	wire [INTERCONNECT_PORTS*3-1:0]  axi_arsize;

	wire [INTERCONNECT_PORTS*2-1:0]  axi_rresp;
	wire [INTERCONNECT_PORTS*4-1:0]  axi_rid;
	wire [INTERCONNECT_PORTS*32-1:0] axi_rdata;
	wire [INTERCONNECT_PORTS-1:0]    axi_rlast;
	wire [INTERCONNECT_PORTS-1:0]    axi_rvalid;
	wire [INTERCONNECT_PORTS-1:0]    axi_rready;

	// Write
	wire [INTERCONNECT_PORTS*32-1:0] axi_awaddr;
	wire [INTERCONNECT_PORTS*8-1:0]  axi_awlen;
	wire [INTERCONNECT_PORTS-1:0]    axi_awvalid;
	wire [INTERCONNECT_PORTS-1:0]    axi_awready;
	wire [INTERCONNECT_PORTS*4-1:0]  axi_awid;
	wire [INTERCONNECT_PORTS*2-1:0]  axi_awburst;
	wire [INTERCONNECT_PORTS*3-1:0]  axi_awsize;

	wire [INTERCONNECT_PORTS*32-1:0] axi_wdata;
	wire [INTERCONNECT_PORTS*4-1:0]  axi_wstrb;
	wire [INTERCONNECT_PORTS-1:0]    axi_wlast;
	wire [INTERCONNECT_PORTS-1:0]    axi_wvalid;
	wire [INTERCONNECT_PORTS-1:0]    axi_wready;

	wire [INTERCONNECT_PORTS*4-1:0]  axi_bid;
	wire [INTERCONNECT_PORTS-1:0]    axi_bvalid;
	wire [INTERCONNECT_PORTS-1:0]    axi_bready;
	wire [INTERCONNECT_PORTS*2-1:0]  axi_bresp;

	// Wires for second port of spartanmc_mem_cache
	wire        second_port_cache_access;
	wire        second_port_cache_wr_low;
	wire        second_port_cache_wr_high;
	wire [9:0]  second_port_cache_addr_block;
	wire [7:0]  second_port_cache_addr_high;
	wire [17:0] second_port_cache_di;
	wire [17:0] second_port_cache_do;

	// Coherence bus wires
	wire [CACHES-1:0] last_i;
	wire              last_o;
	wire [CACHES-1:0] request;
	wire [CACHES-1:0] grant;

	wire [CACHES*18-1:0] addr_i;
	wire [17:0]          addr_o;
	wire [CACHES-1:0]    addr_valid_i;
	wire                 addr_valid_o;
	wire [CACHES*18-1:0] data_i;
	wire [17:0]          data_o;
	wire [CACHES-1:0]    hit_i;
	wire                 hit_o;
	wire [CACHES*2-1:0]  type_i;
	wire [1:0]           type_o;
	wire [CACHES-1:0]    hold_i;
	wire                 hold_o;

	// Use caches
	spartanmc_mem_cache #(
		.BASE_ADDR(BASE_ADDR),
		.LOCAL_RAMBLOCKS(LOCAL_RAMBLOCKS),
		.LOCAL_SHOW_MEM(LOCAL_SHOW_MEM),
		.LINE_SIZE(LINE_SIZE),
		.WAYS(WAYS),
		.SETS(SETS),
		.DATA_RAMBLOCKS(DATA_RAMBLOCKS),
		.SECOND_PORT(SECOND_PORT)
	) spartanmc_cache (
		.clk(clk),
		.reset(reset),
		.code_access(code_access),
		.code_jmp(code_jmp),
		.code_addr(code_addr),
		.code_do(code_do),
		.code_busy(code_busy),

		.mem_access(mem_access),
		.mem_wr_low(mem_wr_low),
		.mem_wr_high(mem_wr_high),
		.mem_addr_block(mem_addr_block),
		.mem_addr_high(mem_addr_high),
		.mem_di(mem_di),
		.mem_do(mem_do),
		.mem_busy(mem_busy),
		.mem_busy_wr(mem_busy_wr),

		.cache_clk(cache_clk),
		.cache_reset(cache_reset),
		.cache_access(second_port_cache_access),
		.cache_wr_low(second_port_cache_wr_low),
		.cache_wr_high(second_port_cache_wr_high),
		.cache_addr_block(second_port_cache_addr_block),
		.cache_addr_high(second_port_cache_addr_high),
		.cache_di(second_port_cache_di),
		.cache_do(second_port_cache_do),

		// AXI
		.ARADDR(axi_araddr[31:0]),
		.ARLEN(axi_arlen[7:0]),
		.ARVALID(axi_arvalid[0]),
		.ARREADY(axi_arready[0]),
		.ARID(axi_arid[3:0]),
		.ARBURST(axi_arburst[1:0]),
		.ARSIZE(axi_arsize[2:0]),

		.RRESP(axi_rresp[1:0]),
		.RID(axi_rid[3:0]),
		.RDATA(axi_rdata[31:0]),
		.RLAST(axi_rlast[0]),
		.RVALID(axi_rvalid[0]),
		.RREADY(axi_rready[0]),

		.AWADDR(axi_awaddr[31:0]),
		.AWLEN(axi_awlen[7:0]),
		.AWVALID(axi_awvalid[0]),
		.AWREADY(axi_awready[0]),
		.AWID(axi_awid[3:0]),
		.AWBURST(axi_awburst[1:0]),
		.AWSIZE(axi_awsize[2:0]),

		.WDATA(axi_wdata[31:0]),
		.WSTRB(axi_wstrb[3:0]),
		.WLAST(axi_wlast[0]),
		.WVALID(axi_wvalid[0]),
		.WREADY(axi_wready[0]),

		.BID(axi_bid[3:0]),
		.BVALID(axi_bvalid[0]),
		.BREADY(axi_bready[0]),
		.BRESP(axi_bresp[1:0]),

		// Coherence bus
		.last_i(last_i[0]),
		.last_o(last_o),
		.request(request[0]),
		.grant(grant[0]),

		.addr_i(addr_i[17:0]),
		.addr_o(addr_o),
		.addr_valid_i(addr_valid_i[0]),
		.addr_valid_o(addr_valid_o),
		.data_i(data_i[17:0]),
		.data_o(data_o),
		.hit_i(hit_i[0]),
		.hit_o(hit_o),
		.type_i(type_i[1:0]),
		.type_o(type_o),
		.hold_i(hold_i[0]),
		.hold_o(hold_o)
	);

	if(SECOND_PORT == 1) begin
		assign second_port_cache_access = cache_access[0];
		assign second_port_cache_wr_low = cache_wr_low[0];
		assign second_port_cache_wr_high = cache_wr_high[0];
		assign second_port_cache_addr_block = cache_addr_block[9:0];
		assign second_port_cache_addr_high = cache_addr_high[7:0];
		assign second_port_cache_di = cache_di[17:0];
		assign cache_do[17:0] = second_port_cache_do;
		assign cache_mem_busy[0] = mem_busy;
		assign cache_mem_busy_wr[0] = mem_busy_wr;
	end

	genvar i;
	for(i=1; i<CACHES; i=i+1) begin : gen_caches
		axi_cache #(
			.LINE_SIZE(LINE_SIZE),
			.WAYS(WAYS),
			.SETS(SETS),
			.SINGLE_MASTER(1)
		) cache (
			// Memory bus
			.reset(cache_reset),
			.clk_peri(cache_clk),
			.mem_access(cache_access[i-1+SECOND_PORT]),
			.store_access_low(cache_wr_low[i-1+SECOND_PORT]),
			.store_access_high(cache_wr_high[i-1+SECOND_PORT]),
			.addr_peri(cache_addr_block[(i-1+SECOND_PORT+1)*10-1:(i-1+SECOND_PORT)*10]),
			.addr_high(cache_addr_high[(i-1+SECOND_PORT+1)*8-1:(i-1+SECOND_PORT)*8]),
			.do_peri(cache_di[(i-1+SECOND_PORT+1)*18-1:(i-1+SECOND_PORT)*18]),
			.di_peri(cache_do[(i-1+SECOND_PORT+1)*18-1:(i-1+SECOND_PORT)*18]),
			.stall(cache_mem_busy[i-1+SECOND_PORT]),
			.stall_wr(cache_mem_busy_wr[i-1+SECOND_PORT]),

			// AXI
			.ARADDR(axi_araddr[(i+1)*32-1:i*32]),
			.ARLEN(axi_arlen[(i+1)*8-1:i*8]),
			.ARVALID(axi_arvalid[i]),
			.ARREADY(axi_arready[i]),
			.ARID(axi_arid[(i+1)*4-1:i*4]),
			.ARBURST(axi_arburst[(i+1)*2-1:i*2]),
			.ARSIZE(axi_arsize[(i+1)*3-1:i*3]),

			.RRESP(axi_rresp[(i+1)*2-1:i*2]),
			.RID(axi_rid[(i+1)*4-1:i*4]),
			.RDATA(axi_rdata[(i+1)*32-1:i*32]),
			.RLAST(axi_rlast[i]),
			.RVALID(axi_rvalid[i]),
			.RREADY(axi_rready[i]),

			.AWADDR(axi_awaddr[(i+1)*32-1:i*32]),
			.AWLEN(axi_awlen[(i+1)*8-1:i*8]),
			.AWVALID(axi_awvalid[i]),
			.AWREADY(axi_awready[i]),
			.AWID(axi_awid[(i+1)*4-1:i*4]),
			.AWBURST(axi_awburst[(i+1)*2-1:i*2]),
			.AWSIZE(axi_awsize[(i+1)*3-1:i*3]),

			.WDATA(axi_wdata[(i+1)*32-1:i*32]),
			.WSTRB(axi_wstrb[(i+1)*4-1:i*4]),
			.WLAST(axi_wlast[i]),
			.WVALID(axi_wvalid[i]),
			.WREADY(axi_wready[i]),

			.BID(axi_bid[(i+1)*4-1:i*4]),
			.BVALID(axi_bvalid[i]),
			.BREADY(axi_bready[i]),
			.BRESP(axi_bresp[(i+1)*2-1:i*2]),

			// Coherence bus
			.last_i(last_i[i]),
			.last_o(last_o),
			.request(request[i]),
			.grant(grant[i]),

			.addr_i(addr_i[(i+1)*18-1:i*18]),
			.addr_o(addr_o),
			.addr_valid_i(addr_valid_i[i]),
			.addr_valid_o(addr_valid_o),
			.data_i(data_i[(i+1)*18-1:i*18]),
			.data_o(data_o),
			.hit_i(hit_i[i]),
			.hit_o(hit_o),
			.type_i(type_i[(i+1)*2-1:i*2]),
			.type_o(type_o),
			.hold_i(hold_i[i]),
			.hold_o(hold_o)
		);
	end

	wire [1:0] type_i_merged;
	orgate #(
		.OR_WIDTH(CACHES),
		.BUS_WIDTH(2)
	) type_or (
		.a(type_i),
		.y(type_i_merged)
	);
	wire [17:0] addr_i_merged;
	orgate #(
		.OR_WIDTH(CACHES),
		.BUS_WIDTH(18)
	) addr_or (
		.a(addr_i),
		.y(addr_i_merged)
	);
	wire [17:0] data_i_merged;
	orgate #(
		.OR_WIDTH(CACHES),
		.BUS_WIDTH(18)
	) data_or (
		.a(data_i),
		.y(data_i_merged)
	);

	// Coherence bus master
	coherence #(
		.RESET_LEVEL("HIGH_ACTIVE"),
		.N(CACHES)
	) coherence_master (
		.clk(clk),
		.reset(reset),
		.last_i(|last_i),
		.last_o(last_o),
		.request(request),
		.grant(grant),

		.addr_i(addr_i_merged),
		.addr_o(addr_o),
		.addr_valid_i(|addr_valid_i),
		.addr_valid_o(addr_valid_o),
		.data_i(data_i_merged),
		.data_o(data_o),
		.hit_i(|hit_i),
		.hit_o(hit_o),
		.type_i(type_i_merged),
		.type_o(type_o),
		.hold_i(|hold_i),
		.hold_o(hold_o)
	);

	// AXI Interconnect
	axi_interconnect_v1_06_a #(
		.C_FAMILY                ( "artix7"                                 ),
		.C_NUM_SLAVE_PORTS       ( CACHES                                   ),
		.C_THREAD_ID_WIDTH       ( 0                                        ),
		.C_THREAD_ID_PORT_WIDTH  ( 4                                        ),
		.C_AXI_ADDR_WIDTH        ( 32                                       )
	) interconnect (
		 .INTERCONNECT_ACLK    ( clk                                                ),
		 .INTERCONNECT_ARESETN ( ~reset                                             ),
		 .M00_AXI_ACLK         ( clk                                                ),
		 .M00_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S00_AXI_ACLK         ( clk                                                ),
		 .S00_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S01_AXI_ACLK         ( clk                                                ),
		 .S01_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S02_AXI_ACLK         ( clk                                                ),
		 .S02_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S03_AXI_ACLK         ( clk                                                ),
		 .S03_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S04_AXI_ACLK         ( clk                                                ),
		 .S04_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S05_AXI_ACLK         ( clk                                                ),
		 .S05_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S06_AXI_ACLK         ( clk                                                ),
		 .S06_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S07_AXI_ACLK         ( clk                                                ),
		 .S07_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S08_AXI_ACLK         ( clk                                                ),
		 .S08_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S09_AXI_ACLK         ( clk                                                ),
		 .S09_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S10_AXI_ACLK         ( clk                                                ),
		 .S10_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S11_AXI_ACLK         ( clk                                                ),
		 .S11_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S12_AXI_ACLK         ( clk                                                ),
		 .S12_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S13_AXI_ACLK         ( clk                                                ),
		 .S13_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S14_AXI_ACLK         ( clk                                                ),
		 .S14_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .S15_AXI_ACLK         ( clk                                                ),
		 .S15_AXI_ARESET_OUT_N ( /* N/C */                                          ),
		 .M00_AXI_ARADDR       ( ARADDR                                             ),
		 .M00_AXI_ARBURST      ( ARBURST                                            ),
		 .M00_AXI_ARID         ( ARID                                               ),
		 .M00_AXI_ARLEN        ( ARLEN                                              ),
		 .M00_AXI_ARREADY      ( ARREADY                                            ),
		 .M00_AXI_ARSIZE       ( ARSIZE                                             ),
		 .M00_AXI_ARVALID      ( ARVALID                                            ),
		 .M00_AXI_AWADDR       ( AWADDR                                             ),
		 .M00_AXI_AWBURST      ( AWBURST                                            ),
		 .M00_AXI_AWID         ( AWID                                               ),
		 .M00_AXI_AWLEN        ( AWLEN                                              ),
		 .M00_AXI_AWREADY      ( AWREADY                                            ),
		 .M00_AXI_AWSIZE       ( AWSIZE                                             ),
		 .M00_AXI_AWVALID      ( AWVALID                                            ),
		 .M00_AXI_BID          ( BID                                                ),
		 .M00_AXI_BREADY       ( BREADY                                             ),
		 .M00_AXI_BRESP        ( BRESP                                              ),
		 .M00_AXI_BVALID       ( BVALID                                             ),
		 .M00_AXI_RDATA        ( RDATA                                              ),
		 .M00_AXI_RID          ( RID                                                ),
		 .M00_AXI_RLAST        ( RLAST                                              ),
		 .M00_AXI_RREADY       ( RREADY                                             ),
		 .M00_AXI_RRESP        ( RRESP                                              ),
		 .M00_AXI_RVALID       ( RVALID                                             ),
		 .M00_AXI_WDATA        ( WDATA                                              ),
		 .M00_AXI_WLAST        ( WLAST                                              ),
		 .M00_AXI_WREADY       ( WREADY                                             ),
		 .M00_AXI_WSTRB        ( WSTRB                                              ),
		 .M00_AXI_WVALID       ( WVALID                                             ),
		 .S00_AXI_ARADDR       ( axi_araddr[31:0]                                   ),
		 .S01_AXI_ARADDR       ( axi_araddr[63:32]                                  ),
		 .S02_AXI_ARADDR       ( axi_araddr[95:64]                                  ),
		 .S03_AXI_ARADDR       ( axi_araddr[127:96]                                 ),
		 .S04_AXI_ARADDR       ( axi_araddr[159:128]                                ),
		 .S05_AXI_ARADDR       ( axi_araddr[191:160]                                ),
		 .S06_AXI_ARADDR       ( axi_araddr[223:192]                                ),
		 .S07_AXI_ARADDR       ( axi_araddr[255:224]                                ),
		 .S08_AXI_ARADDR       ( axi_araddr[287:256]                                ),
		 .S09_AXI_ARADDR       ( axi_araddr[319:288]                                ),
		 .S10_AXI_ARADDR       ( axi_araddr[351:320]                                ),
		 .S11_AXI_ARADDR       ( axi_araddr[383:352]                                ),
		 .S12_AXI_ARADDR       ( axi_araddr[415:384]                                ),
		 .S13_AXI_ARADDR       ( axi_araddr[447:416]                                ),
		 .S14_AXI_ARADDR       ( axi_araddr[479:448]                                ),
		 .S15_AXI_ARADDR       ( axi_araddr[511:480]                                ),
		 .S00_AXI_ARID         ( axi_arid[3:0]                                      ),
		 .S01_AXI_ARID         ( axi_arid[7:4]                                      ),
		 .S02_AXI_ARID         ( axi_arid[11:8]                                     ),
		 .S03_AXI_ARID         ( axi_arid[15:12]                                    ),
		 .S04_AXI_ARID         ( axi_arid[19:16]                                    ),
		 .S05_AXI_ARID         ( axi_arid[23:20]                                    ),
		 .S06_AXI_ARID         ( axi_arid[27:24]                                    ),
		 .S07_AXI_ARID         ( axi_arid[31:28]                                    ),
		 .S08_AXI_ARID         ( axi_arid[35:32]                                    ),
		 .S09_AXI_ARID         ( axi_arid[39:36]                                    ),
		 .S10_AXI_ARID         ( axi_arid[43:40]                                    ),
		 .S11_AXI_ARID         ( axi_arid[47:44]                                    ),
		 .S12_AXI_ARID         ( axi_arid[51:48]                                    ),
		 .S13_AXI_ARID         ( axi_arid[55:52]                                    ),
		 .S14_AXI_ARID         ( axi_arid[59:56]                                    ),
		 .S15_AXI_ARID         ( axi_arid[63:60]                                    ),
		 .S00_AXI_ARLEN        ( axi_arlen[7:0]                                     ),
		 .S01_AXI_ARLEN        ( axi_arlen[15:8]                                    ),
		 .S02_AXI_ARLEN        ( axi_arlen[23:16]                                   ),
		 .S03_AXI_ARLEN        ( axi_arlen[31:24]                                   ),
		 .S04_AXI_ARLEN        ( axi_arlen[39:32]                                   ),
		 .S05_AXI_ARLEN        ( axi_arlen[47:40]                                   ),
		 .S06_AXI_ARLEN        ( axi_arlen[55:48]                                   ),
		 .S07_AXI_ARLEN        ( axi_arlen[63:56]                                   ),
		 .S08_AXI_ARLEN        ( axi_arlen[71:64]                                   ),
		 .S09_AXI_ARLEN        ( axi_arlen[79:72]                                   ),
		 .S10_AXI_ARLEN        ( axi_arlen[87:80]                                   ),
		 .S11_AXI_ARLEN        ( axi_arlen[95:88]                                   ),
		 .S12_AXI_ARLEN        ( axi_arlen[103:96]                                  ),
		 .S13_AXI_ARLEN        ( axi_arlen[111:104]                                 ),
		 .S14_AXI_ARLEN        ( axi_arlen[119:112]                                 ),
		 .S15_AXI_ARLEN        ( axi_arlen[127:120]                                 ),
		 .S00_AXI_ARREADY      ( axi_arready[0]                                     ),
		 .S01_AXI_ARREADY      ( axi_arready[1]                                     ),
		 .S02_AXI_ARREADY      ( axi_arready[2]                                     ),
		 .S03_AXI_ARREADY      ( axi_arready[3]                                     ),
		 .S04_AXI_ARREADY      ( axi_arready[4]                                     ),
		 .S05_AXI_ARREADY      ( axi_arready[5]                                     ),
		 .S06_AXI_ARREADY      ( axi_arready[6]                                     ),
		 .S07_AXI_ARREADY      ( axi_arready[7]                                     ),
		 .S08_AXI_ARREADY      ( axi_arready[8]                                     ),
		 .S09_AXI_ARREADY      ( axi_arready[9]                                     ),
		 .S10_AXI_ARREADY      ( axi_arready[10]                                    ),
		 .S11_AXI_ARREADY      ( axi_arready[11]                                    ),
		 .S12_AXI_ARREADY      ( axi_arready[12]                                    ),
		 .S13_AXI_ARREADY      ( axi_arready[13]                                    ),
		 .S14_AXI_ARREADY      ( axi_arready[14]                                    ),
		 .S15_AXI_ARREADY      ( axi_arready[15]                                    ),
		 .S00_AXI_ARSIZE       ( axi_arsize[2:0]                                    ),
		 .S01_AXI_ARSIZE       ( axi_arsize[5:3]                                    ),
		 .S02_AXI_ARSIZE       ( axi_arsize[8:6]                                    ),
		 .S03_AXI_ARSIZE       ( axi_arsize[11:9]                                   ),
		 .S04_AXI_ARSIZE       ( axi_arsize[14:12]                                  ),
		 .S05_AXI_ARSIZE       ( axi_arsize[17:15]                                  ),
		 .S06_AXI_ARSIZE       ( axi_arsize[20:18]                                  ),
		 .S07_AXI_ARSIZE       ( axi_arsize[23:21]                                  ),
		 .S08_AXI_ARSIZE       ( axi_arsize[26:24]                                  ),
		 .S09_AXI_ARSIZE       ( axi_arsize[29:27]                                  ),
		 .S10_AXI_ARSIZE       ( axi_arsize[32:30]                                  ),
		 .S11_AXI_ARSIZE       ( axi_arsize[35:33]                                  ),
		 .S12_AXI_ARSIZE       ( axi_arsize[38:36]                                  ),
		 .S13_AXI_ARSIZE       ( axi_arsize[41:39]                                  ),
		 .S14_AXI_ARSIZE       ( axi_arsize[44:42]                                  ),
		 .S15_AXI_ARSIZE       ( axi_arsize[47:45]                                  ),
		 .S00_AXI_ARBURST      ( axi_arburst[1:0]                                   ),
		 .S01_AXI_ARBURST      ( axi_arburst[3:2]                                   ),
		 .S02_AXI_ARBURST      ( axi_arburst[5:4]                                   ),
		 .S03_AXI_ARBURST      ( axi_arburst[7:6]                                   ),
		 .S04_AXI_ARBURST      ( axi_arburst[9:8]                                   ),
		 .S05_AXI_ARBURST      ( axi_arburst[11:10]                                 ),
		 .S06_AXI_ARBURST      ( axi_arburst[13:12]                                 ),
		 .S07_AXI_ARBURST      ( axi_arburst[15:14]                                 ),
		 .S08_AXI_ARBURST      ( axi_arburst[17:16]                                 ),
		 .S09_AXI_ARBURST      ( axi_arburst[19:18]                                 ),
		 .S10_AXI_ARBURST      ( axi_arburst[21:20]                                 ),
		 .S11_AXI_ARBURST      ( axi_arburst[23:22]                                 ),
		 .S12_AXI_ARBURST      ( axi_arburst[25:24]                                 ),
		 .S13_AXI_ARBURST      ( axi_arburst[27:26]                                 ),
		 .S14_AXI_ARBURST      ( axi_arburst[29:28]                                 ),
		 .S15_AXI_ARBURST      ( axi_arburst[31:30]                                 ),
		 .S00_AXI_ARVALID      ( axi_arvalid[0]                                     ),
		 .S01_AXI_ARVALID      ( axi_arvalid[1]                                     ),
		 .S02_AXI_ARVALID      ( axi_arvalid[2]                                     ),
		 .S03_AXI_ARVALID      ( axi_arvalid[3]                                     ),
		 .S04_AXI_ARVALID      ( axi_arvalid[4]                                     ),
		 .S05_AXI_ARVALID      ( axi_arvalid[5]                                     ),
		 .S06_AXI_ARVALID      ( axi_arvalid[6]                                     ),
		 .S07_AXI_ARVALID      ( axi_arvalid[7]                                     ),
		 .S08_AXI_ARVALID      ( axi_arvalid[8]                                     ),
		 .S09_AXI_ARVALID      ( axi_arvalid[9]                                     ),
		 .S10_AXI_ARVALID      ( axi_arvalid[10]                                    ),
		 .S11_AXI_ARVALID      ( axi_arvalid[11]                                    ),
		 .S12_AXI_ARVALID      ( axi_arvalid[12]                                    ),
		 .S13_AXI_ARVALID      ( axi_arvalid[13]                                    ),
		 .S14_AXI_ARVALID      ( axi_arvalid[14]                                    ),
		 .S15_AXI_ARVALID      ( axi_arvalid[15]                                    ),
		 .S00_AXI_AWADDR       ( axi_awaddr[31:0]                                   ),
		 .S01_AXI_AWADDR       ( axi_awaddr[63:32]                                  ),
		 .S02_AXI_AWADDR       ( axi_awaddr[95:64]                                  ),
		 .S03_AXI_AWADDR       ( axi_awaddr[127:96]                                 ),
		 .S04_AXI_AWADDR       ( axi_awaddr[159:128]                                ),
		 .S05_AXI_AWADDR       ( axi_awaddr[191:160]                                ),
		 .S06_AXI_AWADDR       ( axi_awaddr[223:192]                                ),
		 .S07_AXI_AWADDR       ( axi_awaddr[255:224]                                ),
		 .S08_AXI_AWADDR       ( axi_awaddr[287:256]                                ),
		 .S09_AXI_AWADDR       ( axi_awaddr[319:288]                                ),
		 .S10_AXI_AWADDR       ( axi_awaddr[351:320]                                ),
		 .S11_AXI_AWADDR       ( axi_awaddr[383:352]                                ),
		 .S12_AXI_AWADDR       ( axi_awaddr[415:384]                                ),
		 .S13_AXI_AWADDR       ( axi_awaddr[447:416]                                ),
		 .S14_AXI_AWADDR       ( axi_awaddr[479:448]                                ),
		 .S15_AXI_AWADDR       ( axi_awaddr[511:480]                                ),
		 .S00_AXI_AWBURST      ( axi_awburst[1:0]                                   ),
		 .S01_AXI_AWBURST      ( axi_awburst[3:2]                                   ),
		 .S02_AXI_AWBURST      ( axi_awburst[5:4]                                   ),
		 .S03_AXI_AWBURST      ( axi_awburst[7:6]                                   ),
		 .S04_AXI_AWBURST      ( axi_awburst[9:8]                                   ),
		 .S05_AXI_AWBURST      ( axi_awburst[11:10]                                 ),
		 .S06_AXI_AWBURST      ( axi_awburst[13:12]                                 ),
		 .S07_AXI_AWBURST      ( axi_awburst[15:14]                                 ),
		 .S08_AXI_AWBURST      ( axi_awburst[17:16]                                 ),
		 .S09_AXI_AWBURST      ( axi_awburst[19:18]                                 ),
		 .S10_AXI_AWBURST      ( axi_awburst[21:20]                                 ),
		 .S11_AXI_AWBURST      ( axi_awburst[23:22]                                 ),
		 .S12_AXI_AWBURST      ( axi_awburst[25:24]                                 ),
		 .S13_AXI_AWBURST      ( axi_awburst[27:26]                                 ),
		 .S14_AXI_AWBURST      ( axi_awburst[29:28]                                 ),
		 .S15_AXI_AWBURST      ( axi_awburst[31:30]                                 ),
		 .S00_AXI_AWID         ( axi_awid[3:0]                                      ),
		 .S01_AXI_AWID         ( axi_awid[7:4]                                      ),
		 .S02_AXI_AWID         ( axi_awid[11:8]                                     ),
		 .S03_AXI_AWID         ( axi_awid[15:12]                                    ),
		 .S04_AXI_AWID         ( axi_awid[19:16]                                    ),
		 .S05_AXI_AWID         ( axi_awid[23:20]                                    ),
		 .S06_AXI_AWID         ( axi_awid[27:24]                                    ),
		 .S07_AXI_AWID         ( axi_awid[31:28]                                    ),
		 .S08_AXI_AWID         ( axi_awid[35:32]                                    ),
		 .S09_AXI_AWID         ( axi_awid[39:36]                                    ),
		 .S10_AXI_AWID         ( axi_awid[43:40]                                    ),
		 .S11_AXI_AWID         ( axi_awid[47:44]                                    ),
		 .S12_AXI_AWID         ( axi_awid[51:48]                                    ),
		 .S13_AXI_AWID         ( axi_awid[55:52]                                    ),
		 .S14_AXI_AWID         ( axi_awid[59:56]                                    ),
		 .S15_AXI_AWID         ( axi_awid[63:60]                                    ),
		 .S00_AXI_AWLEN        ( axi_awlen[7:0]                                     ),
		 .S01_AXI_AWLEN        ( axi_awlen[15:8]                                    ),
		 .S02_AXI_AWLEN        ( axi_awlen[23:16]                                   ),
		 .S03_AXI_AWLEN        ( axi_awlen[31:24]                                   ),
		 .S04_AXI_AWLEN        ( axi_awlen[39:32]                                   ),
		 .S05_AXI_AWLEN        ( axi_awlen[47:40]                                   ),
		 .S06_AXI_AWLEN        ( axi_awlen[55:48]                                   ),
		 .S07_AXI_AWLEN        ( axi_awlen[63:56]                                   ),
		 .S08_AXI_AWLEN        ( axi_awlen[71:64]                                   ),
		 .S09_AXI_AWLEN        ( axi_awlen[79:72]                                   ),
		 .S10_AXI_AWLEN        ( axi_awlen[87:80]                                   ),
		 .S11_AXI_AWLEN        ( axi_awlen[95:88]                                   ),
		 .S12_AXI_AWLEN        ( axi_awlen[103:96]                                  ),
		 .S13_AXI_AWLEN        ( axi_awlen[111:104]                                 ),
		 .S14_AXI_AWLEN        ( axi_awlen[119:112]                                 ),
		 .S15_AXI_AWLEN        ( axi_awlen[127:120]                                 ),
		 .S00_AXI_AWREADY      ( axi_awready[0]                                     ),
		 .S01_AXI_AWREADY      ( axi_awready[1]                                     ),
		 .S02_AXI_AWREADY      ( axi_awready[2]                                     ),
		 .S03_AXI_AWREADY      ( axi_awready[3]                                     ),
		 .S04_AXI_AWREADY      ( axi_awready[4]                                     ),
		 .S05_AXI_AWREADY      ( axi_awready[5]                                     ),
		 .S06_AXI_AWREADY      ( axi_awready[6]                                     ),
		 .S07_AXI_AWREADY      ( axi_awready[7]                                     ),
		 .S08_AXI_AWREADY      ( axi_awready[8]                                     ),
		 .S09_AXI_AWREADY      ( axi_awready[9]                                     ),
		 .S10_AXI_AWREADY      ( axi_awready[10]                                    ),
		 .S11_AXI_AWREADY      ( axi_awready[11]                                    ),
		 .S12_AXI_AWREADY      ( axi_awready[12]                                    ),
		 .S13_AXI_AWREADY      ( axi_awready[13]                                    ),
		 .S14_AXI_AWREADY      ( axi_awready[14]                                    ),
		 .S15_AXI_AWREADY      ( axi_awready[15]                                    ),
		 .S00_AXI_AWSIZE       ( axi_awsize[2:0]                                    ),
		 .S01_AXI_AWSIZE       ( axi_awsize[5:3]                                    ),
		 .S02_AXI_AWSIZE       ( axi_awsize[8:6]                                    ),
		 .S03_AXI_AWSIZE       ( axi_awsize[11:9]                                   ),
		 .S04_AXI_AWSIZE       ( axi_awsize[14:12]                                  ),
		 .S05_AXI_AWSIZE       ( axi_awsize[17:15]                                  ),
		 .S06_AXI_AWSIZE       ( axi_awsize[20:18]                                  ),
		 .S07_AXI_AWSIZE       ( axi_awsize[23:21]                                  ),
		 .S08_AXI_AWSIZE       ( axi_awsize[26:24]                                  ),
		 .S09_AXI_AWSIZE       ( axi_awsize[29:27]                                  ),
		 .S10_AXI_AWSIZE       ( axi_awsize[32:30]                                  ),
		 .S11_AXI_AWSIZE       ( axi_awsize[35:33]                                  ),
		 .S12_AXI_AWSIZE       ( axi_awsize[38:36]                                  ),
		 .S13_AXI_AWSIZE       ( axi_awsize[41:39]                                  ),
		 .S14_AXI_AWSIZE       ( axi_awsize[44:42]                                  ),
		 .S15_AXI_AWSIZE       ( axi_awsize[47:45]                                  ),
		 .S00_AXI_AWVALID      ( axi_awvalid[0]                                     ),
		 .S01_AXI_AWVALID      ( axi_awvalid[1]                                     ),
		 .S02_AXI_AWVALID      ( axi_awvalid[2]                                     ),
		 .S03_AXI_AWVALID      ( axi_awvalid[3]                                     ),
		 .S04_AXI_AWVALID      ( axi_awvalid[4]                                     ),
		 .S05_AXI_AWVALID      ( axi_awvalid[5]                                     ),
		 .S06_AXI_AWVALID      ( axi_awvalid[6]                                     ),
		 .S07_AXI_AWVALID      ( axi_awvalid[7]                                     ),
		 .S08_AXI_AWVALID      ( axi_awvalid[8]                                     ),
		 .S09_AXI_AWVALID      ( axi_awvalid[9]                                     ),
		 .S10_AXI_AWVALID      ( axi_awvalid[10]                                    ),
		 .S11_AXI_AWVALID      ( axi_awvalid[11]                                    ),
		 .S12_AXI_AWVALID      ( axi_awvalid[12]                                    ),
		 .S13_AXI_AWVALID      ( axi_awvalid[13]                                    ),
		 .S14_AXI_AWVALID      ( axi_awvalid[14]                                    ),
		 .S15_AXI_AWVALID      ( axi_awvalid[15]                                    ),
		 .S00_AXI_BID          ( axi_bid[3:0]                                       ),
		 .S01_AXI_BID          ( axi_bid[7:4]                                       ),
		 .S02_AXI_BID          ( axi_bid[11:8]                                      ),
		 .S03_AXI_BID          ( axi_bid[15:12]                                     ),
		 .S04_AXI_BID          ( axi_bid[19:16]                                     ),
		 .S05_AXI_BID          ( axi_bid[23:20]                                     ),
		 .S06_AXI_BID          ( axi_bid[27:24]                                     ),
		 .S07_AXI_BID          ( axi_bid[31:28]                                     ),
		 .S08_AXI_BID          ( axi_bid[35:32]                                     ),
		 .S09_AXI_BID          ( axi_bid[39:36]                                     ),
		 .S10_AXI_BID          ( axi_bid[43:40]                                     ),
		 .S11_AXI_BID          ( axi_bid[47:44]                                     ),
		 .S12_AXI_BID          ( axi_bid[51:48]                                     ),
		 .S13_AXI_BID          ( axi_bid[55:52]                                     ),
		 .S14_AXI_BID          ( axi_bid[59:56]                                     ),
		 .S15_AXI_BID          ( axi_bid[63:60]                                     ),
		 .S00_AXI_BREADY       ( axi_bready[0]                                      ),
		 .S01_AXI_BREADY       ( axi_bready[1]                                      ),
		 .S02_AXI_BREADY       ( axi_bready[2]                                      ),
		 .S03_AXI_BREADY       ( axi_bready[3]                                      ),
		 .S04_AXI_BREADY       ( axi_bready[4]                                      ),
		 .S05_AXI_BREADY       ( axi_bready[5]                                      ),
		 .S06_AXI_BREADY       ( axi_bready[6]                                      ),
		 .S07_AXI_BREADY       ( axi_bready[7]                                      ),
		 .S08_AXI_BREADY       ( axi_bready[8]                                      ),
		 .S09_AXI_BREADY       ( axi_bready[9]                                      ),
		 .S10_AXI_BREADY       ( axi_bready[10]                                     ),
		 .S11_AXI_BREADY       ( axi_bready[11]                                     ),
		 .S12_AXI_BREADY       ( axi_bready[12]                                     ),
		 .S13_AXI_BREADY       ( axi_bready[13]                                     ),
		 .S14_AXI_BREADY       ( axi_bready[14]                                     ),
		 .S15_AXI_BREADY       ( axi_bready[15]                                     ),
		 .S00_AXI_BRESP        ( axi_bresp[1:0]                                     ),
		 .S01_AXI_BRESP        ( axi_bresp[3:2]                                     ),
		 .S02_AXI_BRESP        ( axi_bresp[5:4]                                     ),
		 .S03_AXI_BRESP        ( axi_bresp[7:6]                                     ),
		 .S04_AXI_BRESP        ( axi_bresp[9:8]                                     ),
		 .S05_AXI_BRESP        ( axi_bresp[11:10]                                   ),
		 .S06_AXI_BRESP        ( axi_bresp[13:12]                                   ),
		 .S07_AXI_BRESP        ( axi_bresp[15:14]                                   ),
		 .S08_AXI_BRESP        ( axi_bresp[17:16]                                   ),
		 .S09_AXI_BRESP        ( axi_bresp[19:18]                                   ),
		 .S10_AXI_BRESP        ( axi_bresp[21:20]                                   ),
		 .S11_AXI_BRESP        ( axi_bresp[23:22]                                   ),
		 .S12_AXI_BRESP        ( axi_bresp[25:24]                                   ),
		 .S13_AXI_BRESP        ( axi_bresp[27:26]                                   ),
		 .S14_AXI_BRESP        ( axi_bresp[29:28]                                   ),
		 .S15_AXI_BRESP        ( axi_bresp[31:30]                                   ),
		 .S00_AXI_BVALID       ( axi_bvalid[0]                                      ),
		 .S01_AXI_BVALID       ( axi_bvalid[1]                                      ),
		 .S02_AXI_BVALID       ( axi_bvalid[2]                                      ),
		 .S03_AXI_BVALID       ( axi_bvalid[3]                                      ),
		 .S04_AXI_BVALID       ( axi_bvalid[4]                                      ),
		 .S05_AXI_BVALID       ( axi_bvalid[5]                                      ),
		 .S06_AXI_BVALID       ( axi_bvalid[6]                                      ),
		 .S07_AXI_BVALID       ( axi_bvalid[7]                                      ),
		 .S08_AXI_BVALID       ( axi_bvalid[8]                                      ),
		 .S09_AXI_BVALID       ( axi_bvalid[9]                                      ),
		 .S10_AXI_BVALID       ( axi_bvalid[10]                                     ),
		 .S11_AXI_BVALID       ( axi_bvalid[11]                                     ),
		 .S12_AXI_BVALID       ( axi_bvalid[12]                                     ),
		 .S13_AXI_BVALID       ( axi_bvalid[13]                                     ),
		 .S14_AXI_BVALID       ( axi_bvalid[14]                                     ),
		 .S15_AXI_BVALID       ( axi_bvalid[15]                                     ),
		 .S00_AXI_RDATA        ( axi_rdata[31:0]                                    ),
		 .S01_AXI_RDATA        ( axi_rdata[63:32]                                   ),
		 .S02_AXI_RDATA        ( axi_rdata[95:64]                                   ),
		 .S03_AXI_RDATA        ( axi_rdata[127:96]                                  ),
		 .S04_AXI_RDATA        ( axi_rdata[159:128]                                 ),
		 .S05_AXI_RDATA        ( axi_rdata[191:160]                                 ),
		 .S06_AXI_RDATA        ( axi_rdata[223:192]                                 ),
		 .S07_AXI_RDATA        ( axi_rdata[255:224]                                 ),
		 .S08_AXI_RDATA        ( axi_rdata[287:256]                                 ),
		 .S09_AXI_RDATA        ( axi_rdata[319:288]                                 ),
		 .S10_AXI_RDATA        ( axi_rdata[351:320]                                 ),
		 .S11_AXI_RDATA        ( axi_rdata[383:352]                                 ),
		 .S12_AXI_RDATA        ( axi_rdata[415:384]                                 ),
		 .S13_AXI_RDATA        ( axi_rdata[447:416]                                 ),
		 .S14_AXI_RDATA        ( axi_rdata[479:448]                                 ),
		 .S15_AXI_RDATA        ( axi_rdata[511:480]                                 ),
		 .S00_AXI_RID          ( axi_rid[3:0]                                       ),
		 .S01_AXI_RID          ( axi_rid[7:4]                                       ),
		 .S02_AXI_RID          ( axi_rid[11:8]                                      ),
		 .S03_AXI_RID          ( axi_rid[15:12]                                     ),
		 .S04_AXI_RID          ( axi_rid[19:16]                                     ),
		 .S05_AXI_RID          ( axi_rid[23:20]                                     ),
		 .S06_AXI_RID          ( axi_rid[27:24]                                     ),
		 .S07_AXI_RID          ( axi_rid[31:28]                                     ),
		 .S08_AXI_RID          ( axi_rid[35:32]                                     ),
		 .S09_AXI_RID          ( axi_rid[39:36]                                     ),
		 .S10_AXI_RID          ( axi_rid[43:40]                                     ),
		 .S11_AXI_RID          ( axi_rid[47:44]                                     ),
		 .S12_AXI_RID          ( axi_rid[51:48]                                     ),
		 .S13_AXI_RID          ( axi_rid[55:52]                                     ),
		 .S14_AXI_RID          ( axi_rid[59:56]                                     ),
		 .S15_AXI_RID          ( axi_rid[63:60]                                     ),
		 .S00_AXI_RLAST        ( axi_rlast[0]                                       ),
		 .S01_AXI_RLAST        ( axi_rlast[1]                                       ),
		 .S02_AXI_RLAST        ( axi_rlast[2]                                       ),
		 .S03_AXI_RLAST        ( axi_rlast[3]                                       ),
		 .S04_AXI_RLAST        ( axi_rlast[4]                                       ),
		 .S05_AXI_RLAST        ( axi_rlast[5]                                       ),
		 .S06_AXI_RLAST        ( axi_rlast[6]                                       ),
		 .S07_AXI_RLAST        ( axi_rlast[7]                                       ),
		 .S08_AXI_RLAST        ( axi_rlast[8]                                       ),
		 .S09_AXI_RLAST        ( axi_rlast[9]                                       ),
		 .S10_AXI_RLAST        ( axi_rlast[10]                                      ),
		 .S11_AXI_RLAST        ( axi_rlast[11]                                      ),
		 .S12_AXI_RLAST        ( axi_rlast[12]                                      ),
		 .S13_AXI_RLAST        ( axi_rlast[13]                                      ),
		 .S14_AXI_RLAST        ( axi_rlast[14]                                      ),
		 .S15_AXI_RLAST        ( axi_rlast[15]                                      ),
		 .S00_AXI_RREADY       ( axi_rready[0]                                      ),
		 .S01_AXI_RREADY       ( axi_rready[1]                                      ),
		 .S02_AXI_RREADY       ( axi_rready[2]                                      ),
		 .S03_AXI_RREADY       ( axi_rready[3]                                      ),
		 .S04_AXI_RREADY       ( axi_rready[4]                                      ),
		 .S05_AXI_RREADY       ( axi_rready[5]                                      ),
		 .S06_AXI_RREADY       ( axi_rready[6]                                      ),
		 .S07_AXI_RREADY       ( axi_rready[7]                                      ),
		 .S08_AXI_RREADY       ( axi_rready[8]                                      ),
		 .S09_AXI_RREADY       ( axi_rready[9]                                      ),
		 .S10_AXI_RREADY       ( axi_rready[10]                                     ),
		 .S11_AXI_RREADY       ( axi_rready[11]                                     ),
		 .S12_AXI_RREADY       ( axi_rready[12]                                     ),
		 .S13_AXI_RREADY       ( axi_rready[13]                                     ),
		 .S14_AXI_RREADY       ( axi_rready[14]                                     ),
		 .S15_AXI_RREADY       ( axi_rready[15]                                     ),
		 .S00_AXI_RRESP        ( axi_rresp[1:0]                                     ),
		 .S01_AXI_RRESP        ( axi_rresp[3:2]                                     ),
		 .S02_AXI_RRESP        ( axi_rresp[5:4]                                     ),
		 .S03_AXI_RRESP        ( axi_rresp[7:6]                                     ),
		 .S04_AXI_RRESP        ( axi_rresp[9:8]                                     ),
		 .S05_AXI_RRESP        ( axi_rresp[11:10]                                   ),
		 .S06_AXI_RRESP        ( axi_rresp[13:12]                                   ),
		 .S07_AXI_RRESP        ( axi_rresp[15:14]                                   ),
		 .S08_AXI_RRESP        ( axi_rresp[17:16]                                   ),
		 .S09_AXI_RRESP        ( axi_rresp[19:18]                                   ),
		 .S10_AXI_RRESP        ( axi_rresp[21:20]                                   ),
		 .S11_AXI_RRESP        ( axi_rresp[23:22]                                   ),
		 .S12_AXI_RRESP        ( axi_rresp[25:24]                                   ),
		 .S13_AXI_RRESP        ( axi_rresp[27:26]                                   ),
		 .S14_AXI_RRESP        ( axi_rresp[29:28]                                   ),
		 .S15_AXI_RRESP        ( axi_rresp[31:30]                                   ),
		 .S00_AXI_RVALID       ( axi_rvalid[0]                                      ),
		 .S01_AXI_RVALID       ( axi_rvalid[1]                                      ),
		 .S02_AXI_RVALID       ( axi_rvalid[2]                                      ),
		 .S03_AXI_RVALID       ( axi_rvalid[3]                                      ),
		 .S04_AXI_RVALID       ( axi_rvalid[4]                                      ),
		 .S05_AXI_RVALID       ( axi_rvalid[5]                                      ),
		 .S06_AXI_RVALID       ( axi_rvalid[6]                                      ),
		 .S07_AXI_RVALID       ( axi_rvalid[7]                                      ),
		 .S08_AXI_RVALID       ( axi_rvalid[8]                                      ),
		 .S09_AXI_RVALID       ( axi_rvalid[9]                                      ),
		 .S10_AXI_RVALID       ( axi_rvalid[10]                                     ),
		 .S11_AXI_RVALID       ( axi_rvalid[11]                                     ),
		 .S12_AXI_RVALID       ( axi_rvalid[12]                                     ),
		 .S13_AXI_RVALID       ( axi_rvalid[13]                                     ),
		 .S14_AXI_RVALID       ( axi_rvalid[14]                                     ),
		 .S15_AXI_RVALID       ( axi_rvalid[15]                                     ),
		 .S00_AXI_WDATA        ( axi_wdata[31:0]                                    ),
		 .S01_AXI_WDATA        ( axi_wdata[63:32]                                   ),
		 .S02_AXI_WDATA        ( axi_wdata[95:64]                                   ),
		 .S03_AXI_WDATA        ( axi_wdata[127:96]                                  ),
		 .S04_AXI_WDATA        ( axi_wdata[159:128]                                 ),
		 .S05_AXI_WDATA        ( axi_wdata[191:160]                                 ),
		 .S06_AXI_WDATA        ( axi_wdata[223:192]                                 ),
		 .S07_AXI_WDATA        ( axi_wdata[255:224]                                 ),
		 .S08_AXI_WDATA        ( axi_wdata[287:256]                                 ),
		 .S09_AXI_WDATA        ( axi_wdata[319:288]                                 ),
		 .S10_AXI_WDATA        ( axi_wdata[351:320]                                 ),
		 .S11_AXI_WDATA        ( axi_wdata[383:352]                                 ),
		 .S12_AXI_WDATA        ( axi_wdata[415:384]                                 ),
		 .S13_AXI_WDATA        ( axi_wdata[447:416]                                 ),
		 .S14_AXI_WDATA        ( axi_wdata[479:448]                                 ),
		 .S15_AXI_WDATA        ( axi_wdata[511:480]                                 ),
		 .S00_AXI_WLAST        ( axi_wlast[0]                                       ),
		 .S01_AXI_WLAST        ( axi_wlast[1]                                       ),
		 .S02_AXI_WLAST        ( axi_wlast[2]                                       ),
		 .S03_AXI_WLAST        ( axi_wlast[3]                                       ),
		 .S04_AXI_WLAST        ( axi_wlast[4]                                       ),
		 .S05_AXI_WLAST        ( axi_wlast[5]                                       ),
		 .S06_AXI_WLAST        ( axi_wlast[6]                                       ),
		 .S07_AXI_WLAST        ( axi_wlast[7]                                       ),
		 .S08_AXI_WLAST        ( axi_wlast[8]                                       ),
		 .S09_AXI_WLAST        ( axi_wlast[9]                                       ),
		 .S10_AXI_WLAST        ( axi_wlast[10]                                      ),
		 .S11_AXI_WLAST        ( axi_wlast[11]                                      ),
		 .S12_AXI_WLAST        ( axi_wlast[12]                                      ),
		 .S13_AXI_WLAST        ( axi_wlast[13]                                      ),
		 .S14_AXI_WLAST        ( axi_wlast[14]                                      ),
		 .S15_AXI_WLAST        ( axi_wlast[15]                                      ),
		 .S00_AXI_WREADY       ( axi_wready[0]                                      ),
		 .S01_AXI_WREADY       ( axi_wready[1]                                      ),
		 .S02_AXI_WREADY       ( axi_wready[2]                                      ),
		 .S03_AXI_WREADY       ( axi_wready[3]                                      ),
		 .S04_AXI_WREADY       ( axi_wready[4]                                      ),
		 .S05_AXI_WREADY       ( axi_wready[5]                                      ),
		 .S06_AXI_WREADY       ( axi_wready[6]                                      ),
		 .S07_AXI_WREADY       ( axi_wready[7]                                      ),
		 .S08_AXI_WREADY       ( axi_wready[8]                                      ),
		 .S09_AXI_WREADY       ( axi_wready[9]                                      ),
		 .S10_AXI_WREADY       ( axi_wready[10]                                     ),
		 .S11_AXI_WREADY       ( axi_wready[11]                                     ),
		 .S12_AXI_WREADY       ( axi_wready[12]                                     ),
		 .S13_AXI_WREADY       ( axi_wready[13]                                     ),
		 .S14_AXI_WREADY       ( axi_wready[14]                                     ),
		 .S15_AXI_WREADY       ( axi_wready[15]                                     ),
		 .S00_AXI_WSTRB        ( axi_wstrb[3:0]                                     ),
		 .S01_AXI_WSTRB        ( axi_wstrb[7:4]                                     ),
		 .S02_AXI_WSTRB        ( axi_wstrb[11:8]                                    ),
		 .S03_AXI_WSTRB        ( axi_wstrb[15:12]                                   ),
		 .S04_AXI_WSTRB        ( axi_wstrb[19:16]                                   ),
		 .S05_AXI_WSTRB        ( axi_wstrb[23:20]                                   ),
		 .S06_AXI_WSTRB        ( axi_wstrb[27:24]                                   ),
		 .S07_AXI_WSTRB        ( axi_wstrb[31:28]                                   ),
		 .S08_AXI_WSTRB        ( axi_wstrb[35:32]                                   ),
		 .S09_AXI_WSTRB        ( axi_wstrb[39:36]                                   ),
		 .S10_AXI_WSTRB        ( axi_wstrb[43:40]                                   ),
		 .S11_AXI_WSTRB        ( axi_wstrb[47:44]                                   ),
		 .S12_AXI_WSTRB        ( axi_wstrb[51:48]                                   ),
		 .S13_AXI_WSTRB        ( axi_wstrb[55:52]                                   ),
		 .S14_AXI_WSTRB        ( axi_wstrb[59:56]                                   ),
		 .S15_AXI_WSTRB        ( axi_wstrb[63:60]                                   ),
		 .S00_AXI_WVALID       ( axi_wvalid[0]                                      ),
		 .S01_AXI_WVALID       ( axi_wvalid[1]                                      ),
		 .S02_AXI_WVALID       ( axi_wvalid[2]                                      ),
		 .S03_AXI_WVALID       ( axi_wvalid[3]                                      ),
		 .S04_AXI_WVALID       ( axi_wvalid[4]                                      ),
		 .S05_AXI_WVALID       ( axi_wvalid[5]                                      ),
		 .S06_AXI_WVALID       ( axi_wvalid[6]                                      ),
		 .S07_AXI_WVALID       ( axi_wvalid[7]                                      ),
		 .S08_AXI_WVALID       ( axi_wvalid[8]                                      ),
		 .S09_AXI_WVALID       ( axi_wvalid[9]                                      ),
		 .S10_AXI_WVALID       ( axi_wvalid[10]                                     ),
		 .S11_AXI_WVALID       ( axi_wvalid[11]                                     ),
		 .S12_AXI_WVALID       ( axi_wvalid[12]                                     ),
		 .S13_AXI_WVALID       ( axi_wvalid[13]                                     ),
		 .S14_AXI_WVALID       ( axi_wvalid[14]                                     ),
		 .S15_AXI_WVALID       ( axi_wvalid[15]                                     )
	);
end // DISABLE_CACHE
endgenerate

endmodule
