/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale 1ns / 10ps

module axi_bram_slave #(
	parameter MEM_SIZE        = 4096,
	parameter AXI_DATA_WIDTH  = 32,
	parameter AXI_ADDR_WIDTH  = 32,
	parameter ID_WIDTH        = 4
) (
	input wire s_aclk,
	input wire s_aresetn,

	// Slave Interface Write Address Ports
	input  wire [ID_WIDTH-1:0]  s_axi_awid,
	input  wire [AXI_ADDR_WIDTH-1:0] s_axi_awaddr,
	input  wire [7:0]           s_axi_awlen,
	input  wire [2:0]           s_axi_awsize,
	input  wire [1:0]           s_axi_awburst,
	input  wire [0:0]           s_axi_awlock,
	input  wire [3:0]           s_axi_awcache,
	input  wire [2:0]           s_axi_awprot,
	input  wire [3:0]           s_axi_awqos,
	input  wire                 s_axi_awvalid,
	output reg                  s_axi_awready,
	// Slave Interface Write Data Ports
	input  wire [AXI_DATA_WIDTH-1:0]   s_axi_wdata,
	input  wire [AXI_DATA_WIDTH/8-1:0] s_axi_wstrb,
	input  wire                 s_axi_wlast,
	input  wire                 s_axi_wvalid,
	output reg                  s_axi_wready,
	// Slave Interface Write Response Ports
	input  wire                 s_axi_bready,
	output reg  [ID_WIDTH-1:0]  s_axi_bid,
	output wire [1:0]           s_axi_bresp,
	output reg                  s_axi_bvalid,
	// Slave Interface Read Address Ports
	input  wire [ID_WIDTH-1:0]  s_axi_arid,
	input  wire [AXI_ADDR_WIDTH-1:0] s_axi_araddr,
	input  wire [7:0]           s_axi_arlen,
	input  wire [2:0]           s_axi_arsize,
	input  wire [1:0]           s_axi_arburst,
	input  wire [0:0]           s_axi_arlock,
	input  wire [3:0]           s_axi_arcache,
	input  wire [2:0]           s_axi_arprot,
	input  wire [3:0]           s_axi_arqos,
	input  wire                 s_axi_arvalid,
	output reg                  s_axi_arready,
	// Slave Interface Read Data Ports
	input  wire                 s_axi_rready,
	output reg  [ID_WIDTH-1:0]  s_axi_rid,
	output reg  [AXI_DATA_WIDTH-1:0] s_axi_rdata,
	output wire [1:0]           s_axi_rresp,
	output reg                  s_axi_rlast,
	output reg                  s_axi_rvalid
);

/*
TODO:
Read: buffer rdata in case of deassertion of rready
*/

function integer log2;
	input integer value;
	begin
		value = value - 1;
		for (log2=0; value>0; log2=log2+1) begin
			value = value>>1;
		end
	end
endfunction

reg[AXI_DATA_WIDTH-1:0] mem [MEM_SIZE/(AXI_DATA_WIDTH/8)-1:0];

localparam MEM_ADDR_WIDTH = log2(MEM_SIZE/(AXI_DATA_WIDTH/8));

reg [MEM_ADDR_WIDTH-1:0] write_addr;
reg [1:0] state_write;
localparam
	WRITE_IDLE = 0,
	WRITE_DATA = 1,
	WRITE_RESPONSE = 2;

// Write logic
always @(posedge s_aclk) begin
	if(s_axi_wvalid) begin
		mem[write_addr] <= s_axi_wdata;
	end
end

always @(posedge s_aclk) begin
	if(~s_aresetn) begin
		state_write <= WRITE_IDLE;
		write_addr <= 0;
		s_axi_awready <= 1;
		s_axi_wready <= 0;
		s_axi_bvalid <= 0;
	end else begin
		s_axi_awready <= 0;
		s_axi_wready <= 0;
		s_axi_bvalid <= 0;
		case(state_write)
			WRITE_IDLE:
				if(s_axi_awvalid) begin
					s_axi_wready <= 1;
					s_axi_bid <= s_axi_awid;
					write_addr <= s_axi_awaddr/(AXI_DATA_WIDTH/8);
					state_write <= WRITE_DATA;
				end else begin
					s_axi_awready <= 1;
					s_axi_wready <= 0;
					state_write <= WRITE_IDLE;
				end
			WRITE_DATA: begin
					if(s_axi_wvalid) begin
						write_addr <= write_addr + 1;
					end
					if(s_axi_wlast && s_axi_wvalid) begin
						s_axi_wready <= 0;
						s_axi_bvalid <= 1;
						state_write <= WRITE_RESPONSE;
					end else begin
						s_axi_wready <= 1;
						state_write <= WRITE_DATA;
					end
				end
			WRITE_RESPONSE:
				if(s_axi_bready) begin
					s_axi_awready <= 1;
					state_write <= WRITE_IDLE;
				end else begin
					s_axi_bvalid <= 1;
					state_write <= WRITE_RESPONSE;
				end
			default: state_write <= WRITE_IDLE;
		endcase
	end
end

reg [MEM_ADDR_WIDTH-1:0] read_addr;
wire [MEM_ADDR_WIDTH-1:0] araddr;
assign araddr = s_axi_araddr/(AXI_DATA_WIDTH/8);
reg [7:0] read_burst_length;
reg [2:0] state_read;
localparam
	READ_IDLE = 0,
	READ_DATA = 1,
	READ_DATA_LAST = 2,
	READ_RESPONSE = 3;

// Read logic
always @(posedge s_aclk) begin
	if(state_read == READ_IDLE && s_axi_arvalid || state_read == READ_DATA && s_axi_rready) begin
		s_axi_rdata <= mem[(state_read == READ_IDLE) ? araddr : read_addr];
	end else begin
		s_axi_rdata <= s_axi_rdata;
	end
end
always @(posedge s_aclk) begin
	if(~s_aresetn) begin
		state_read <= READ_IDLE;
		s_axi_arready <= 1;
		read_burst_length <= 0;
		s_axi_rvalid <= 0;
		s_axi_rlast <= 0;
	end else begin
		s_axi_arready <= 0;
		s_axi_rvalid <= 0;
		s_axi_rlast <= 0;
		case(state_read)
			READ_IDLE:
				if(s_axi_arvalid && s_axi_arlen == 0) begin
					s_axi_rvalid <= 1;
					s_axi_rlast <= 1;
					s_axi_rid <= s_axi_arid;
					read_burst_length <= 0;
					state_read <= READ_DATA_LAST;
				end else if(s_axi_arvalid) begin
					read_addr <= araddr + 1;
					s_axi_rvalid <= 1;
					read_burst_length <= s_axi_arlen - 1;
					s_axi_rid <= s_axi_arid;
					state_read <= READ_DATA;
				end else begin
					s_axi_arready <= 1;
					state_read <= READ_IDLE;
				end
			READ_DATA: begin
					if(s_axi_rready && read_burst_length == 0) begin
						s_axi_rlast <= 1;
						state_read <= READ_DATA_LAST;
					end else if(s_axi_rready) begin
						read_burst_length <= read_burst_length - 1;
						read_addr <= read_addr + 1;
						state_read <= READ_DATA;
					end else begin
						state_read <= READ_DATA;
					end
					s_axi_rvalid <= 1;
				end
			READ_DATA_LAST:
				if(s_axi_rready) begin
					s_axi_arready <= 1;
					s_axi_rvalid <= 0;
					state_read <= READ_IDLE;
				end else begin
					s_axi_rlast <= 1;
					s_axi_rvalid <= 1;
					state_read <= READ_DATA_LAST;
				end
			default: state_read <= READ_IDLE;
		endcase
	end
end

endmodule
