/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`timescale      1ns / 10ps
`default_nettype none
module bus_address_verification_top #(
	parameter COUNT = 1,
	parameter LOCATION = "UNKNOWN",
	parameter ADDRESS_RANGE = "UNKNOWN",
	parameter ADDR_WIDTH = 18,
	parameter DATA_WIDTH = 18
)  (
	input wire clk,
	input wire reset,
	input wire [ADDR_WIDTH-1:0] addr,
	input wire [DATA_WIDTH-1:0] data,
	input wire [COUNT-1:0] select,
	input wire busy,
	input wire access
);

reg[ADDR_WIDTH-1:0] addr_last;

reg accessRunning;
reg stop = 0;
always @(posedge clk) begin

	if (reset) begin
		accessRunning <= 0;
		stop <= 0;
	end else begin
		if ((busy !== 1'd1)) begin
			addr_last <= addr;
			accessRunning <= |select;
		end

		if (|data === 1'bX && accessRunning && ~busy) begin
			$display("[ %9d ] Error: Data Output on %s contains undefined bits: 0x%x", $time, LOCATION, data);
			stop <= 1;
		end 

		if ((data !== 18'd0) && !accessRunning ) begin
			if (access) begin
				$display("[ %9d ] Error: Detected access on %s while not selected. Output is 0x%x but should be 0. Module is supposed to respond to addresses %s, but responded to address 0x%x", $time, LOCATION, data, ADDRESS_RANGE, addr_last);
			end else begin
				$display("[ %9d ] Error: Detected access on %s while not selected. Output is 0x%x but should be 0. Module is supposed to respond to addresses %s, but responded when there was no memory access (at address 0x%x)", $time, LOCATION, data, ADDRESS_RANGE, addr_last);
			end
			stop <= 1;
		end

		if ((busy !== 1'd0) && !accessRunning) begin
			if (access) begin
				$display("[ %9d ] Error: Detected access on busy signal of %s while not selected. Output is %x but should be 0. Module is supposed to respond to addresses %s, but responded to address 0x%x", $time, LOCATION, busy, ADDRESS_RANGE, addr_last);
			end else begin
				$display("[ %9d ] Error: Detected access on busy signal of %s while not selected. Output is %x but should be 0. Module is supposed to respond to addresses %s, but responded when there was no memory access (at address 0x%x)", $time, LOCATION, busy, ADDRESS_RANGE, addr_last);
			end
			stop <= 1;
		end

		if (busy && select) begin
			$display("[ %9d ] Error: Processor fail! Processor did not lower its access signal despite presence of busy signal of %s", $time, LOCATION);
			stop <= 1;
		end

		if (stop)
			$stop();
	end
end


endmodule

