/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  BSL                                                        ////
////  Fuegt Stuffing-Bits ein und entfernt Stuffing-Bits.        ////
////                                                             ////
////  Author: Gerald Hempel                                      ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`default_nettype none
// synopsys translate_off
`include "can/src/timescale.vh"
// synopsys translate_on

module can_bsl
	(
	can_clk, 
	reset,
	config_mode,
	
	//Input von BTL
	sample_bit, 
	sample_bit_pre,
	sample_strobe, 
	tx_strobe, 
	
	//Input von BSP
	tx, 
	tx_next,
	go_overload_frame, 
	go_error_frame, 
	go_crc_lim,
	go_id1,
	rx_crc_lim_stuffing_disable,
	transmitting,
	
	//Output
	sample_strobe_q,
	tx_strobe_q,	
	tx_stuffed,				
	stuff_error
	);
   
	parameter 	tp=1;				//Nur zur Simulation
	
	input wire  		can_clk;
	input wire 		reset;
	input wire		config_mode;
	input wire 		sample_bit;			//Abgetastetes Sample
	input wire		sample_bit_pre;			//Vorgoegerbit
	input wire 		sample_strobe;			//Lesepunkt des Sample-Bits
	input wire 		tx_strobe;			//Uebergabepkt. an Bus
	input wire 		tx;				//Signal von BSP wird an Bus weitergegeben
	input wire		tx_next;			//naechstes Bit von BSP auf TX
	input wire 		go_overload_frame;		//erlaubt kein Bitstuffing
	input wire 		go_error_frame;			//erlaubt kein Bitstuffing
	input wire 		go_id1;				//markiert Beginn des Bitstuffin
	input wire 		go_crc_lim;			//markiert Ende des Bitstuffing
	input wire		rx_crc_lim_stuffing_disable;	//aussetzen des Bitstuffings am Frame-Ende eine Takt frueher
	input wire  		transmitting;			//deaktivieren das tx-Bitstuffing wenn nciht gesendet wird
	
	output reg		sample_strobe_q;		//Uebergabepkt. des Samples an BSP
	output reg		tx_strobe_q;			//Uebernahmepkt. von Bits von BSP
	output reg		tx_stuffed;			//Ausgabe von Bits mit Stuffing an Bus zum Zeitpkt. tx_strobe
	output wire		stuff_error;			//Ausgabe eines Stuff-Errors an EML

	
	reg	[2:0]	bit_stuff_cnt_rx;		//Anzahl aufgetretener gleicher Bits in RX (max 5)
	reg	[2:0]	bit_stuff_cnt_tx;		//Anzahl aufgetretener gleicher Bits in TX (max 5)
	reg		bit_stuff_cnt_en;		//Schaltet Bitstuffing an
		
	reg		set_bit_to_stuff_tx;
		
	wire 		bit_stuff_reset;		//zuruecksetzen von bit_stuff_cnt
	wire		bit_to_stuff_rx;
	wire		bit_to_stuff_tx;

	assign bit_stuff_reset = go_id1 & (~bit_stuff_cnt_en);
	assign bit_to_stuff_rx = bit_stuff_cnt_rx == 3'h5;
	//assign bit_to_stuff_tx = bit_stuff_cnt_tx == 3'h5;
	assign bit_to_stuff_tx = set_bit_to_stuff_tx;

	// Bitstuffing Error (natuerlich nur fuer empfangene Bits)
	assign stuff_error = /*sample_strobe &*/ bit_stuff_cnt_en & bit_to_stuff_rx & (sample_bit == sample_bit_pre);
	
	// Einschalten Bitstuffing
	always @ (posedge can_clk or posedge reset)
	begin
	  if (reset)
		 bit_stuff_cnt_en <= 1'b0;
	  else if (go_id1)
		 bit_stuff_cnt_en <=#tp 1'b1;
	  else if (go_crc_lim | go_error_frame | go_overload_frame | config_mode)
		 bit_stuff_cnt_en <=#tp 1'b0;
	end

	//Erzeugen von sample_strobe_q und Filtern der Stuffing-Bits
	always @ (posedge can_clk or posedge reset)
	begin
	  if (reset | config_mode)
		 sample_strobe_q <= 1'b0;
	  else if (bit_to_stuff_rx & sample_strobe)
		 sample_strobe_q <=#tp 1'b0;
	  else if (sample_strobe)
		 sample_strobe_q <=#tp 1'b1;
	  else
	    sample_strobe_q <=#tp 1'b0;
	end
	
	//Erzeugen von tx_strobe_q und erzeugen des tx_stuffed-Streams
	always @ (posedge can_clk or posedge reset)
	begin
	  if (reset | config_mode)
	    begin
			tx_strobe_q <= 1'b0;
			tx_stuffed <= 1'b1;
	    end
	  else if (bit_to_stuff_tx & tx_strobe)
		 begin
			tx_strobe_q <=#tp 1'b0;
			tx_stuffed <=#tp (~sample_bit_pre);
		 end
	  else if (tx_strobe)
		begin
			tx_strobe_q <=#tp 1'b1;
			tx_stuffed <=#tp tx;
		end
	  else
	  begin
	      tx_strobe_q <=#tp 1'b0;
	      if (~bit_to_stuff_tx)
	         tx_stuffed <=#tp tx;
	      else
	         tx_stuffed <=#tp (~sample_bit_pre);
	   end	
	end

 
	// bit_stuff_cnt_rx
	always @ (posedge can_clk or posedge reset)
	begin
	  if (reset)
		 bit_stuff_cnt_rx <= 3'h1;
	  else if (bit_stuff_reset | config_mode)
		 bit_stuff_cnt_rx <=#tp 3'h1;
	  else if (sample_strobe & bit_stuff_cnt_en)
		 begin
			if (bit_stuff_cnt_rx == 3'h5)
			  bit_stuff_cnt_rx <=#tp 3'h1;
			else if (sample_bit == sample_bit_pre)
			  bit_stuff_cnt_rx <=#tp bit_stuff_cnt_rx + 1'b1;
			else
			  bit_stuff_cnt_rx <=#tp 3'h1;
		 end
 	   else if (sample_strobe) 
		  bit_stuff_cnt_rx <= 3'h1;
	end
	

	// bit_stuff_cnt_tx
	always @ (posedge can_clk or posedge reset)
	begin
	  if (reset)
	  begin
		 bit_stuff_cnt_tx <= 3'h1;
		 set_bit_to_stuff_tx <= 1'b0;
	  end
	  else if (bit_stuff_reset | ~bit_stuff_cnt_en | config_mode)//+reset_mode
	  begin
		 bit_stuff_cnt_tx <=#tp 3'h1;
		 set_bit_to_stuff_tx <= 1'b0;
	  end
	  else if (tx_strobe & bit_stuff_cnt_en & transmitting)
		 begin
			if (bit_stuff_cnt_tx == 3'h5)
			begin
			  //bit_stuff_cnt_tx <=#tp 3'h1;
			  set_bit_to_stuff_tx <= 1'b1;
			  if (tx == tx_next)
              bit_stuff_cnt_tx <= 3'h0;
           else
              bit_stuff_cnt_tx <= 3'h1;
			end
			else if (tx == tx_next)
			begin
			  bit_stuff_cnt_tx <=#tp bit_stuff_cnt_tx + 1'b1;
			  set_bit_to_stuff_tx <= 1'b0;
			end
			else
			begin
			  bit_stuff_cnt_tx <=#tp 3'h1;
			  set_bit_to_stuff_tx <= 1'b0;
			end
		 end
	end
	
endmodule
