/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  BTL                                                        ////
////  Tastet den Bus ab und fuehrt die CAN-Synchronisation durch.////
////                                                             ////
////  Author: Gerald Hempel                                      ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`default_nettype none
// synopsys translate_off
`include "can/src/timescale.vh"
// synopsys translate_on

module can_btl
	(
	can_clk,
	reset,
	rx, 			//rx vom Bus
	tx,			//tx Status
	
	//Bus-Timing Register
	baudrate_presc,
	sync_jump_width,
	
	time_segment1,		 //Prob-Seg + Puffer Seg 1
	time_segment2,
	samplemode,
	
	//Input von CAN_bsp (Busstatus--> Info fuer Hard- und Soft-Sync. )
	rx_idle,
	rx_inter,
	transmitter, 		//Knoten sendet eigene Nachricht
	transmitting,		//Knoten sendet eigene Nachricht oder Error Frame, Overload Frame
	
	go_rx_inter,
	go_overload_frame,
	go_error_frame,
	go_tx,
	send_ack,
	node_error_passive,
		
	//Output von CAN_btl
	sample_bit, 		//abgetastetes Bit (Sample)
	sample_strobe,		//Strobe zur Uebername des Samples
	clk_fsm,
	tx_strobe,		//Strobe zum Senden eines Bits auf den Bus
	hard_sync,		//Zeigt an ob eine hard_sync moeglich ist
	tx_next,		//naechstes Sendebit
	sample_bit_pre		//Vorgaengerbit
	);
	
	parameter 	tp=1;	//Nur zur Simulation
		
	input wire			can_clk;
	input wire 			reset;
	input wire			rx;
	input wire			tx;
	
	//Bus-Timing Register
	input wire	[6:0]		baudrate_presc;
	input wire	[1:0]		sync_jump_width;
	
	input wire	[4:0]		time_segment1;
	input wire	[2:0]		time_segment2;
	input wire			samplemode;
	
	//CAN_bsp (Busstatus--> Info fuer Hard- und Soft-Sync. )
	input wire			rx_idle;
	input wire			rx_inter;
	input wire			transmitting;
	input wire			transmitter;
	
	input wire			go_rx_inter;
	input wire			go_overload_frame;
	input wire			go_error_frame;
	input wire			go_tx;
	input wire			send_ack;
	input wire			node_error_passive;
	input wire			tx_next;
	
	//Output von CAN_btl
	output 			sample_bit;	
	output			sample_bit_pre;		//Vorgaengerbit
	output 			sample_strobe;
	output			clk_fsm;		
	output 			tx_strobe;	
	output wire		hard_sync;
	
	reg	[6:0]		clk_cnt;
	reg			clk_q;			//Takt eines Zeitquantums
	reg			clk_q_q;
	wire			clk_fsm;
	reg 			resync_blocked;
	reg			hard_sync_blocked;
	reg 			sample_bit;
	reg 			sample_bit_pre;
	reg 			sample_strobe;
	reg	[2:0]		sample;			//speichert Sample fuer Triple-Sampling
	reg	[5:0]		tq_cnt;			//Zeitquantenzaehler
	reg	[3:0]		delay;			//Resync.-Delay
	reg			sync;			//Sync. Segment
	reg 			seg1;			//Bitsegment 1
	reg 			seg2;			//Bitsegment 2
	reg			resync_latched;		//Resync. merken
	reg 			tx_strobe;		//Sende-Strobe an Bsp
	reg 			tx_next_sp;				
	
	wire	[7:0]		preset_cnt;			
	wire	 		sync_window;
	wire			resync;
	//Sync. Statemachine
	wire			go_sync;
	wire 			go_seg1;
	wire			go_seg2;
	
	assign preset_cnt = (baudrate_presc+2);	//Prescalerwert verarbeiten preset_cnt=(BRP+2)
	assign hard_sync  = (~hard_sync_blocked)&(~rx)&( rx_idle |  rx_inter) & sample_bit; //harte Sync. (evt. "&sample_bit")
	assign resync	  = (~resync_blocked) &(~rx)&(~rx_idle)&(~rx_inter) & sample_bit;   //Resync. (evt. "&sample_bit")
	
	/* Erzeugen der Bautrate clk_q=can_clk/preset_cnt*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			clk_cnt <= 7'h0;
		else if (clk_cnt >= (preset_cnt-1'b1))//preset_cnt-1 wegen Nulldurchgang (merkwuerdigerweise keine identische Bitlaenge erforderlich)
			clk_cnt <=#tp 7'h0;
		else
			clk_cnt <=#tp clk_cnt+1'b1;
	end
	
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			clk_q <= 1'b0;
		else if ({1'b0, clk_cnt} == (preset_cnt-1'b1))	//Umschalten bei Erreichen genau eines Zeitquantums
			clk_q <=#tp 1'b1;
		else
			clk_q <=#tp 1'b0;
	end
	
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			clk_q_q <= 1'b0;
		else
			clk_q_q <=#tp sample_strobe;
	end

	assign	clk_fsm	= sample_strobe | clk_q_q;
	
	/* Bit-Segmente */
	assign go_sync = clk_q & (seg2 & (tq_cnt[2:0]==(time_segment2-1'b1))&(~hard_sync)&(~resync))|(hard_sync);
	assign go_seg1 = clk_q & (sync | hard_sync | (resync & seg2 & sync_window)|(resync_latched & sync_window));
	assign go_seg2 = clk_q & seg1 & (tq_cnt[5:0]==((time_segment1-1'b1) + delay))&(~hard_sync);
	
	/* Sync.-Segment */
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			sync <= 1'b0;
		else if (clk_q)
			sync <=#tp go_sync;
	end
		
	/* Puffer-Segment 1 + Ausbreitungsverzoegerung */
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			seg1 <= 1'b1;
		else if (go_seg1)
			seg1 <=#tp 1'b1;
		else if (go_seg2)
			seg1 <=#tp 1'b0;			
	end
	
	/* Puffer-Segment 2 */
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			seg2 <= 1'b0;
		else if (go_seg2)
			seg2 <=#tp 1'b1;
		else if (go_seg1 | go_sync)
			seg2 <=#tp 1'b0;			
	end
	
	/* Setzen des Quantenzaehlers */
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			tq_cnt <= 6'h0;
		else if (go_sync | go_seg1 | go_seg2)
			tq_cnt <=#tp 6'h0;
		else if (clk_q)
			tq_cnt <=#tp tq_cnt + 1'b1;			
	end
	
	/*Beseitigen des positiven und negativen Phasenfehlers*/
	/*Verlaengern von Segment 1 um den Wert delay wenn Flanke zu spaet (pos. Fehler)*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			delay <= 4'h0;
		else if (resync & seg1 & ((transmitting & (tx_next_sp | (tx & (~rx))))|(~transmitting)))//bedingung zur late dedection (bei transmitter sende 1 und empfange 0 benoetigt sync. (anderesherum ist wegen rez. und dom. bits nicht moeglich))-->wenn naechste Bit eine 1 kann auch sync. werden
			delay <=#tp (tq_cnt > {3'h0,sync_jump_width})?({2'h0,sync_jump_width} + 1'b1):(tq_cnt + 1'b1);//Erhoehe nur um tq_cnt wenn noetiger Wert nicht ausserhalb SJW liegt sonst nimm SJW
		else if (go_sync | go_seg1)
			delay <=#tp 4'h0;			
	end

	/*Verkuerzen von Segment 2 wenn Flanke zu frueh (neg. Fehler)
	--> liegt Wert im sync_window kann der Flankenfehler innerhalb einer Bitzeit kompensiert werden*/
	assign sync_window = (((time_segment2-1'b1) - tq_cnt[2:0]) < (sync_jump_width + 1'b1));//+1 evt. fuer sync segment
	
	/*Ist neg. Phasenfehler zu gross (ausserhalb SJW) wird das Korrekturvorhaben gespeichert bis sync_window erreicht
	--> vollstaendige Korrektur erfolgt erst mit der naechsten Bitzeit*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			resync_latched <= 1'b0;
		else if (resync & seg2 &(~sync_window))
			resync_latched <=#tp 1'b1;
		else if (go_seg1)
			resync_latched <= 1'b0;
	end
	
	/*Ermitteln des Bits am naechsten Sample Point*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			tx_next_sp <= 1'b0;
		else if (go_overload_frame | (go_error_frame & (~node_error_passive)) | go_tx | send_ack)
			tx_next_sp <= 1'b0;
		else if (go_error_frame & node_error_passive)
			tx_next_sp <= 1'b1;//Sende rezessiv
		else if (sample_strobe)
			tx_next_sp <=#tp tx_next;		
	end
	
	/*Ermittelne des Sendepunkts eines Bits fuer Can_bsp*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			tx_strobe <= 1'b0;
		else
			tx_strobe <=#tp ~tx_strobe & seg2 & (clk_q & (tq_cnt[2:0]==(time_segment2-1'b1))/*&(resync | hard_sync)*/);
	end
	
	/*Resync ist nur einmal pro Bit moeglich --> wird geblockt wenn einmal aufgetreten*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			resync_blocked <=#tp 1'b1;
		else if (clk_q)
			begin
				if (resync)
					resync_blocked <=#tp 1'b1;
				else if (sync)
					resync_blocked <=#tp 1'b0;
			end
	end
	
	/*Hard_sync wird blockiert waerend der Uebermittlung von Nachrichten und nach einmaligen auftreten*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
			hard_sync_blocked <=#tp 1'b0;
		else if (hard_sync & clk_q | (transmitting & transmitter | go_tx) & tx_strobe & (~tx_next))
			hard_sync_blocked <=#tp 1'b1;
		else if (go_rx_inter | (rx_inter | rx_idle) & sample_strobe & sample_bit) //viele Abfragen um zu verhindern das eine Stoerung eine Hard Sync. ausloest
			hard_sync_blocked <=#tp 1'b0;
	end
	
	/*Vorbereitung des Triple_Samples*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
		begin
			sample <= 3'h0;
		end
	   else if (seg1 & (tq_cnt==((time_segment1-1'b1) + delay)))
	   begin
	      case (clk_cnt % 4)
		2'h3,
		2'h0:	sample[0] <= rx;
		2'h1:	sample[1] <= rx;
		2'h2:	sample[2] <= rx;
		default:sample[0] <= rx;
	      endcase
		//sample <= 3'h3;
	  end
	end
	
	/*Sampeln von Bits (RX)*/
	always @ (posedge can_clk or posedge reset)
	begin
		if (reset)
		begin
			sample_bit	<= 1'b1;
			sample_bit_pre	<= 1'b1;
			sample_strobe	<= 1'b0;
		end
		else if (go_error_frame)
		begin
			sample_bit_pre <=#tp sample_bit;
			sample_strobe  <=#tp 1'b0;
		end
		else if (clk_q & (~hard_sync))
			begin
				if (seg1 & (tq_cnt==((time_segment1-1'b1) + delay)))
				begin
					sample_bit_pre <=#tp sample_bit;
					if (samplemode)	//hoffentlich serielles samplen von 3 Werten
						sample_bit <=#tp (sample[0] & sample[1])|(sample[0] & sample[2])|(sample[1] & sample[2]); //Mehrheitsentscheidung
					else
						sample_bit <=#tp rx;//einfaches erfassen eines Samples
					sample_strobe <=#tp 1'b1;
				end
			end
		else
			sample_strobe <=#tp 1'b0;
	end
		
endmodule

