/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package chipscope

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IIntegerParameterHandler
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.ModuleInternalNames
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil
import de.tu_darmstadt.rs.util.StreamUtil
import org.slf4j.Logger
import org.slf4j.LoggerFactory

import java.nio.file.Files
import java.nio.file.Path
import java.util.stream.Collectors
import java.util.stream.IntStream

class ChipscopeProject {

	private static final Logger logger = LoggerFactory.getLogger(ChipscopeProject)


	static void output(int chainDeviceIndex, IModuleConfiguration<?> icon, Path outputFile) throws IOException {
		Files.createDirectories(outputFile.parent);
		outputFile.toFile().withWriter('utf-8') { writer ->
			writer.println("#ChipScope Pro Analyzer Project File, Version 3.0");

			int numIlas = (int) DocumentUtil.getParameterHandler(icon, "NUMBER_CONTROL_PORTS", IIntegerParameterHandler.class).getIntegerValue().get();
			IntStream.range(0, numIlas).forEach { i ->
				String portName = "CONTROL" + i;
				IPortHandler<? extends IPort> portHandler = DocumentUtil.getPortHandler(icon, portName);
				if (portHandler.isConnected()) {
					Optional<IPortHandler<?>> optIla = portHandler.streamOppositeConnectors().collect(StreamUtil.atMostOne());
					optIla.ifPresent { ila -> outputIla(chainDeviceIndex, i, (IModuleConfiguration<?>) ila.getSourceElement(), new PrintWriter(writer)) }
				}
			}
		}
	}

	static void buildProject(IConfigurationDocument document, Path outputFile) {
		def iconCores = document.streamModulesRecursive()
				.filter {it.getModuleDescriptionName().equals("chipscope_icon")}
			.collect(Collectors.toList())
		if (iconCores.isEmpty())
			return;

		def icon = iconCores.get(0)

		if (iconCores.size()>1)
			logger.warn("Multiple chipscope_icon, using "+icon.toString())


		output(Integer.valueOf(document.getTarget().getDevice().getJtagid())-1, icon, outputFile)
	}

	static void outputNumberedPort(int chainDeviceIndex, int unitIndex, int portIndex, IPortHandler<IPort> ph, PrintWriter writer) {
		writer.format('unit.%1$d.%2$d.port.%3$d.b.0.name=%4$s%nunit.%1$d.%2$d.port.%3$d.channelcount=%5$d%n',
				chainDeviceIndex,
				unitIndex,
				portIndex,
				ph.getPortName(),
				ph.getConnectorWidth()
		);
		ModuleInternalNames.getFriendlyConnectedNames(ph)
				.map(StreamUtil.numberElements())
				.forEach(StreamUtil.IntPair.accessC { pinIndex, name->
			writer.format('unit.%1$d.%2$d.port.%3$d.s.%4$d.alias=%5$s%n' +
					'unit.%1$d.%2$d.port.%3$d.s.%4$d.name=TriggerPort%3$d[%4$d]%n',
					chainDeviceIndex,
					unitIndex,
					portIndex,
					pinIndex,
					name
			)
		})
	}


	static void outputIla(int chainDeviceIndex, int unitIndex, IModuleConfiguration<?> ila, PrintWriter writer) {
		writer.format("unit.%d.%d.coretype=ILA%n",chainDeviceIndex,unitIndex);
		writer.format("unit.%d.%d.username=%s%n",chainDeviceIndex,unitIndex,ila.getGlobalName());


		//Data port
		ila.getPortGroupDescriptions().stream()
				.flatMap { portGroup->portGroup.getPorts().stream() }
				.map(ila.&getPortHandler)
				.filter{ph->ph.getPortName().equals("DATA")}
				.forEach{ ph ->
			outputNumberedPort(chainDeviceIndex, unitIndex, -1, ph, writer)
		}

		int numTrig = (int) DocumentUtil.getParameterHandler(ila, "NUMBER_OF_TRIGGER_PORTS", IIntegerParameterHandler.class).getIntegerValue().get();
		writer.format("unit.%d.%d.portcount=%d%n",chainDeviceIndex,unitIndex,numTrig);

		//Trigger ports
		ila.getPortGroupDescriptions().stream()
				.flatMap { portGroup->portGroup.getPorts().stream() }
				.map(ila.&getPortHandler)
				.filter{ph->ph.getPortName().startsWith("TRIG")}
				.sorted(Comparator.comparing { ph -> ph.getPortName() })
				.map(StreamUtil.numberElements())
				.forEach(StreamUtil.IntPair.accessC { portIndex, ph ->
					outputNumberedPort(chainDeviceIndex, unitIndex, portIndex, ph, writer)
				})

	}

}