/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/* Konfigurierbarer FIFO
 * Der Parameter DATA_WIDTH gibt die Breite des FIFO an (Standard 8 Bit).
 * Der Parameter DEPTH gibt die Tiefe des FIFO an (Standard 8, Maximum 31).
 * Die Tiefe ist durch die Breite der Adresszaehler begrenzt.
 * Es ist gleichzeitiges Lesen und Schreiben moeglich, jedoch kein
 * "Kurzschluss" zwischen Ein- und Ausgang.
 * Das Signal empty zeigt, ob der FIFO leer ist und somit keine Daten gelesen
 * werden koenen. Genauso signalisiert full einen komplett gefuellten FIFO.
 * Ein Schreiben ist solange nicht moeglich.
 * Ueber die Parameter READ- und WRITEMODE kann die Flankensteuerung
 * konfiguriert werden. Bei SYNCHRON werden die Daten mit jeder steigenden
 * Taktflanke uebernommen solange das Enablesignal (read oder write) gesetzt
 * ist. Bei der Betriebsart ASYNCHRON wird die Flanke am Lese- bzw.
 * Schreibeingang untersucht. Ist das Signal im vergleich zum vorherigen Takt
 * jetzt gesetzt, erfolgt die Uebernahme.
 */

`default_nettype none

module fifo(
    clk,
    data_out,
    data_in,
    read,
    write,
    empty,
    full,
    reset
    );

	parameter DATA_WIDTH = 'd8;		//Breite Datenbus
    	parameter DEPTH      = 'd8;		//FIFO-Tiefe
	parameter READMODE   = "SYNCHRON";	//Steuerung eingangsseitig
	parameter WRITEMODE  = "SYNCHRON";	//Steuerung ausgangsseitig

	input wire			clk;		//Takt
	output [DATA_WIDTH-1:0]	data_out;	//Datenausgang
	input  [DATA_WIDTH-1:0]	data_in;	//Dateneingang
	input wire		read;		//Leseaufforderung
	input wire		write;		//Schreibaufforderung
	output			empty;		//Puffer enthaellt keine Daten
	output			full;		//Puffer komplett gefuellt
	input wire		reset;		//Reset

	reg    [DATA_WIDTH-1:0]	data_out;	//Datenausgang
	wire   [DATA_WIDTH-1:0]	data_in;	//Dateneingang
	reg    [DATA_WIDTH-1:0]	buffer[DEPTH-1:0]; //Datenpuffer
	reg               [4:0]	addr_in;	//Eingangsadresse (zeigt auf naechste zu schreibende Zelle)
	reg			empty;		//FIFO leer Flag
	reg			full;		//FIFO voll Flag
	reg			readp;		//Eingang mit Clk synchron
	reg			writep;		//Eingang mit Clk synchron
	reg			read_old;	//Speicher fuer Lesesignal
	reg			write_old;	//Speicher fuer Schreibsignal
	wire			read_sync;
	wire			write_sync;
	wire			ff_read;
	wire			ff_write;
	reg			ff_reado;
	reg			ff_writeo;
	integer			i;		//Laufvariable

	assign read_sync  = (READMODE  == "SYNCHRON");
	assign write_sync = (WRITEMODE == "SYNCHRON");

	assign ff_read	= (((!read_sync  & read_old  & !read)  |
			     (read_sync  & read))   & !empty);
	assign ff_write	= (((!write_sync & write_old & !write) |
			     (write_sync & write))  & !full);

// FIFO lesen und schreiben
	always @(posedge clk)
		if (reset) begin
			addr_in		<= 0;
			buffer[0]	<= 0;
			read_old	<= 0;
			write_old	<= 0;
		end else begin
			// FIFO lesen und schreiben
			if (ff_read && !ff_reado && ff_write && !ff_writeo)
			begin
				buffer[addr_in] <= data_in;
				for (i = 0; i< (DEPTH - 1); i = i + 1)
					buffer[i] <= buffer [i+1];
			end
			// FIFO schreiben
			else if (ff_write && !ff_writeo)
			begin
				buffer[addr_in] <= data_in;
				addr_in         <= addr_in + 1;
			end
			// FIFO lesen
			else if (ff_read && !ff_reado)
			begin
				addr_in <= addr_in - 1;
				for (i = 0; i< (DEPTH - 1); i = i + 1)
					buffer[i] <= buffer [i+1];
			end 
			//Signale zur Flankendetektion speichern
			read_old	<= readp;
			write_old	<= writep;
		end

// FlipFlops fuer full und empty nach jedem Schreiben oder lesen
// aktualisieren.
	always @(posedge clk)
		if (reset) begin
			empty		<= 1;
			full		<= 0;
			ff_reado	<= 0;
			ff_writeo	<= 0;
		end else begin
			if	((addr_in == 0) && ff_reado)
				begin
				empty	<= 1'b1;
				full	<= 1'b0;
				end
			else	if	((addr_in == DEPTH) && ff_writeo)
				begin
				empty	<= 1'b0;
				full	<= 1'b1;
				end
			else	begin
				if	((addr_in > 0) && ff_reado)
				full	<= 1'b0;

				if	((addr_in < DEPTH) && ff_writeo)
				empty	<= 1'b0;
				end
			ff_reado	<= ff_read;
			ff_writeo	<= ff_write;
		end

// Signale setzen, die wahrend FIFO lesen oder schreiben sich nicht
// aendern duerfen. Sonst Fehler bei addr_in Bildung!
	always @(negedge clk)
		if (reset) begin
			readp		<= 0;
			writep		<= 0;
//			data_inp	<= 0;
		end else begin
			readp		<= read;
			writep		<= write;
//			if (write)
//				data_inp	<= data_in;
		end

// FIFO Output Latch setzen wenn kein read angelegt wird und Daten im FIFO sind.
	always	@(read, read_old, buffer[0], reset)
		if (reset)
			data_out	= 0;
		else if (!read && !read_old)
			data_out	= buffer[0];
		else
			data_out	= data_out;

endmodule

