/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/*! Generic peripheral register access decoder
 * This module provides a generic decoder to simplify the implementation of
 * peripheral registers. It uses the @link pselect @endlink module to
 * generate the chip select signal for the peripheral unit.  Additionally it
 * provides a read/write/general select signal for each peripheral register
 * depending on the current address. The number of registers can be
 * configured using the generic parameter @link REG_COUNT @endlink.
 *
 */

`timescale 1 ns / 1 ps
`default_nettype none

//% @copydoc reg_access_decoder.v
module reg_access_decoder #(

    /*! Number of registers
    * Integer value specifying the number of registers.  This value determines
    * the number of select lines the module instance will provide.
    */
    parameter REG_COUNT = 'd8,                                  
    /*! @copydoc spartanmc::addr_peri() */
    parameter BASE_ADR = 18'h0,                                 //% Hexadecimal value of peripheral base address
    parameter SIM_DEBUG = "NO"                                  //% Set to "YES" to get simulation debug output

)(

    input wire                          reset,                  //% @copydoc global::spartanmc::reset
    input wire                          clk_peri,               //% @copydoc global::spartanmc::clk_peri
    input wire      [9:0]               addr_peri,              //% @copydoc global::spartanmc::addr_peri
    input wire                          access_peri,            //% @copydoc global::spartanmc::access_peri
    input wire                          wr_peri,                //% @copydoc global::spartanmc::wr_peri

    output wire                         peri_select,            //% Peripheral unit chip select
    output wire     [REG_COUNT-1:0]     reg_select,             //% Select for each register
    output reg      [REG_COUNT-1:0]     reg_read,               //% Read enable for each register (1 clk delay)
    output wire     [REG_COUNT-1:0]     reg_write               //% Write enable for each register

);

`include "common/functions/bitwidth.vh"


localparam
    REG_ADDR_WIDTH = bitwidth(REG_COUNT - 1),                   //% Number of bits required to address all implemented registers
    IO_SPACE_WIDTH = REG_ADDR_WIDTH > 3 ? REG_ADDR_WIDTH : 3;   //% Number of bits occupied in IO address space (must be at least 8)

//% Decoder for peripheral chip select signal
pselect #(
	.ADDR_WIDTH	(10 - IO_SPACE_WIDTH		),
    .BASE_WIDTH	(10 - IO_SPACE_WIDTH		),
    .BASE_ADDR	(BASE_ADR >> IO_SPACE_WIDTH	)
) pselect0 (
    .addr(addr_peri[9:IO_SPACE_WIDTH]),
    .activ_peri(access_peri),
    .select(peri_select)
);

genvar i;
generate

    if (REG_COUNT == 1)
        assign reg_select[0] = peri_select & !addr_peri[0];
    else
        for (i=0; i<REG_COUNT; i=i+1) begin : REGISTER_SELECT_ACCESS
            assign reg_select[i] = peri_select & (addr_peri[REG_ADDR_WIDTH-1:0] == i);
        end

    for (i=0; i<REG_COUNT; i=i+1)
    begin : REGISTER_ACCESS
        assign reg_write[i]    = wr_peri & reg_select[i];

	always @ (posedge clk_peri) begin
		if (reset)
			reg_read[i]	<= 0;			//% pipe_2 Register
		else
        		reg_read[i]	<= ~wr_peri & reg_select[i];
	end

        //synopsys translate_off
        if (SIM_DEBUG == "YES")
        begin

            always @ (posedge access_peri)
            begin : DEBUG
                #1
                if (reg_select[i])
                begin
                    $write("[ %9d ]  CS  ", $time);
                    if (reg_write[i])
                        $write("write");
                    else
                        $write("read");
                    $display(" select reg[%02d] @ 0x%H (%m)", i, addr_peri);
                end

            end

        end
        //synopsys translate_on

    end

endgenerate

endmodule
