/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Tobias Engelmann
// Create Date:    17:35:36 04/13/2010 
// Design Name:    SpartanMC generic modules
// Module Name:    spmc_async_fifo 
//
// Description:    Asynchronous fifo implementation according to Xilinx app note XAPP 051
//		   Configurable data width and fifo size.
//
//		   The fifo size is set by the number of address bits (n).
//		   Therefore the size will result in 2^n. 
//		   The background is, that we can implement simple read and write address counter,
//                 without maximum check. The simple overflow on increment...
//
// Dependencies:   spmc_graycounter
//                 spmc_fifo_mem
//
//////////////////////////////////////////////////////////////////////////////////

`default_nettype none

module spmc_async_fifo
#(
	parameter	FIFO_ADDR_BITS			= 'd10,
	parameter	FIFO_WIDTH			= 'd18
)(
	input	wire					reset,
	input	wire					clear,

	// Write Clock Domain
	input 	wire					write_clk,
	input	wire					write_enable,
	input	wire	[FIFO_WIDTH-1:0]		write_data,
	output	reg					write_full,
	
	// Read Clock Domain
	input 	wire					read_clk,
	input	wire					read_enable,
	output	wire	[FIFO_WIDTH-1:0]		read_data,
	output	reg					read_empty,
	
	output	wire	[FIFO_ADDR_BITS-1:0]		read_addr,		// Read Address sync to read_clk
	output	wire	[FIFO_ADDR_BITS-1:0]		write_addr		// Write Address sync to write_clk
    );
    

///////////////////////////////////////////////////////////////////
//
// Xilinx Application Note XAPP 051
// Synchronous and Asynchronous FIFO Designs
// See description on top ^^^


///////////////////////////////////////////////////////////////////
//
// Local Wires/ Signals
//
wire				equal_addresses;
wire  	[FIFO_ADDR_BITS-1:0]	next_write_addr_gray, next_read_addr_gray;	// Using Graycode, since 'Quadrant selectros' need to be glitchfree
wire                            next_write_addr_enable, next_read_addr_enable;
wire                            set_quadrant_direction, reset_quadrant_direction;
reg                             quadrant_direction;
wire                            preset_full, preset_empty;

///////////////////////////////////////////////////////////////////
//
// FIFO Addresses/ Control support logic
//
//'Next Addresses' enable logic:
assign next_write_addr_enable = write_enable & ~write_full;
assign next_read_addr_enable  = read_enable  & ~read_empty;
    
//'EqualAddresses' logic:
assign equal_addresses 	= (next_write_addr_gray == next_read_addr_gray);

//'Quadrant selectors' logic:
assign set_quadrant_direction 	= 	(next_write_addr_gray[FIFO_ADDR_BITS-2] ~^ next_read_addr_gray[FIFO_ADDR_BITS-1]) &
					(next_write_addr_gray[FIFO_ADDR_BITS-1] ^  next_read_addr_gray[FIFO_ADDR_BITS-2]);
		    
assign reset_quadrant_direction = 	(next_write_addr_gray[FIFO_ADDR_BITS-2] ^  next_read_addr_gray[FIFO_ADDR_BITS-1]) &
					(next_write_addr_gray[FIFO_ADDR_BITS-1] ~^ next_read_addr_gray[FIFO_ADDR_BITS-2]);
                         
//'quadrant_direction' latch logic:
//D Latch w/ Asynchronous Clear & Preset.
always @ (set_quadrant_direction, reset_quadrant_direction, clear) 
	if (reset_quadrant_direction | clear)
		quadrant_direction = 1'b0;  //Going 'Empty'.
	else if (set_quadrant_direction)
		quadrant_direction = 1'b1;  //Going 'Full'.
	else
		quadrant_direction = 1'b0; //This should never happen
            
//'Full_out' logic for the writing port:
assign preset_full = quadrant_direction & equal_addresses;  //'Full' Fifo.
    
//D Flip-Flop w/ Asynchronous Preset.  
always @ (posedge write_clk, posedge preset_full, posedge reset)//
 	if(reset) begin
		write_full <= 0;
	end
	else begin 
		if (preset_full)
            		write_full <= 1;
        	else
            		write_full <= 0;
	end  
/*always @ (posedge write_clk, posedge preset_full) 
	if (preset_full)
            write_full <= 1;
        else
            write_full <= 0;
*/            
//'Empty_out' logic for the reading port:
assign preset_empty = ~quadrant_direction & equal_addresses;  //'Empty' Fifo.
    
//D Flip-Flop w/ Asynchronous Preset.
always @ (posedge read_clk, posedge preset_empty, posedge reset)
	if(reset) begin
		read_empty <= 1;
	end
	else begin
        	if (preset_empty)
            		read_empty <= 1;
        	else
            		read_empty <= 0;
   	end    
/*
always @ (posedge read_clk, posedge preset_empty)  
        if (preset_empty)
            read_empty <= 1;
        else
            read_empty <= 0;
*/   
///////////////////////////////////////////////////////////////////
//
// Module Instantiations
//
spmc_graycounter
#(
	.COUNTER_WIDTH(			FIFO_ADDR_BITS		)
)
cnt_write_addr0(
	.clk(				write_clk		),
	.reset(				reset			),
	.enable(			next_write_addr_enable	),
	.clear(				clear			),
	.bin_value(			write_addr		),
	.gray_value(			next_write_addr_gray	)
);

spmc_graycounter
#(
	.COUNTER_WIDTH(			FIFO_ADDR_BITS		)
)
cnt_read_addr0(
	.clk(				read_clk		),
	.reset(				reset			),
	.enable(			next_read_addr_enable	),
	.clear(				clear			),
	.bin_value(			read_addr		),
	.gray_value(			next_read_addr_gray		)
);

///////////////////////////////////////////////////////////////////
// Dual ported generic memory module
// With write full protection! Written data will be ignored, if memory is full
spmc_fifo_mem  
#(
	.RAM_ADDR_BITS(			FIFO_ADDR_BITS		),
	.RAM_WIDTH(			FIFO_WIDTH		)
)
fifo_mem0(
	.reset(				reset			),
	.clkA(				write_clk		),
	.addrA(				write_addr		),
	.input_dataA(			write_data		),
	.enableA(			!write_full		),
	.write_enableA(			write_enable 		),
	
	.clkB(				read_clk		),
	.addrB(				read_addr		),
	.output_dataB(			read_data		),
	.enableB(			1'b1			)	
);


endmodule
