/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Create Date:    10:49:53 12/16/2011 
// Module Name:    fifo 
//////////////////////////////////////////////////////////////////////////////////
`default_nettype none

module yaf 
#(
    parameter DSIZE = 32,
    parameter ASIZE = 10,
    parameter ENABLE_PREREAD = "NO",
    parameter ENABLE_PREWRITE = "NO"
)(
    output wire [DSIZE-1:0] rdata,
    output reg  wfull,
    output reg  rfull,
    output wire rwfull,
    output reg  wempty,
    output reg  rempty,
    output wire rwempty,
    output wire [ASIZE:0] fill,
    input wire [DSIZE-1:0] wdata,
    input wire winc, 
    input wire wclk, 
    input wire wrst_n,
    input wire rinc, 
    input wire rclk, 
    input wire rrst_n,
    input wire rptr_upd,
    input wire rptr_rest,
    input wire wptr_upd,
    input wire wptr_rest
);

wire [ASIZE:0] rw_diff;
wire [ASIZE:0] wptr;
wire [ASIZE:0] rptr;
wire [ASIZE:0] wbin;
wire [ASIZE:0] rbin;
wire [ASIZE:0] wq2_rptr;
wire [ASIZE:0] rq2_wptr;

wire rempty_val;
wire wfull_val;

assign fill = wbin - rbin;

//---------------------------------------------------------------
// Synchroisation FIFO rempty rfull
//---------------------------------------------------------------

generate 
  if (ENABLE_PREREAD == "YES")
  begin
	always @(posedge rclk)
	begin
	  if (!rrst_n)
	    begin
	      rempty <= 1'b1;
	      rfull  <= 1'b0;
	    end
	  else 
	    begin
	      rempty <= rempty_val;
	      rfull  <= wfull_val;
	    end
	end
  end else begin
	always @(posedge rclk)
	begin
	  if (!rrst_n)
	    begin
	      rempty <= 1'b1;
	      rfull  <= 1'b0;
	    end
	  else 
	    begin
	      rempty <= rempty_val & (fill == 0);
	      rfull  <= wfull_val;
	    end
	end
  end
endgenerate

//---------------------------------------------------------------
// Synchroisation FIFO wempty wfull
//---------------------------------------------------------------

generate 
  if (ENABLE_PREWRITE == "YES")
  begin
	always @(posedge wclk)
	begin
	  if (!wrst_n) 
	    begin
	      wempty <= 1'b1;
	      wfull  <= 1'b0;
	    end
	  else 
	    begin
	      wempty <= rempty_val;
	      wfull  <= wfull_val;
	    end
	end
  end else begin
	always @(posedge wclk)
	begin
	  if (!wrst_n) 
	    begin
	      wempty <= 1'b1;
	      wfull  <= 1'b0;
	    end
	  else 
	    begin
	      wempty <= rempty_val & (fill == 0);
	      wfull  <= wfull_val;
	    end
	end
  end
endgenerate

assign rwempty = rempty & wempty;
assign rwfull  = rfull  | wfull;


yaf_sync_r2w #(.ADDRSIZE(ASIZE))
yaf_sync_r2w (
    .wq2_rptr(wq2_rptr), 
    .rptr(rptr),
    .wclk(wclk), 
    .wrst_n(wrst_n),
    .rinc(rinc),
    .winc(winc)
);
yaf_sync_w2r #(.ADDRSIZE(ASIZE))
yaf_sync_w2r (
    .rq2_wptr(rq2_wptr), 
    .wptr(wptr),
    .rclk(rclk), 
    .rrst_n(rrst_n),
    .rinc(rinc),
    .winc(winc)
);

yaf_mem #(
    .DSIZE (DSIZE),
    .ASIZE (ASIZE)
) 
yaf_mem (
    .rdata(rdata), 
    .wdata(wdata),
    .waddr(wbin[ASIZE-1:0]), 
    .raddr(rbin[ASIZE-1:0]),
    .wclken(winc), 
    .wfull(wfull),
    .wclk(wclk)
);

yaf_rptr_empty #(
    .ASIZE(ASIZE),
    .ENABLE_PREREAD(ENABLE_PREREAD)
) 
yaf_rptr_empty (
    .rempty_val(rempty_val),
    .rptr(rptr),
    .rbin(rbin),
    .rq2_wptr(rq2_wptr),
    .rempty(rempty),
    .rinc(rinc),
    .rclk(rclk),
    .rrst_n(rrst_n),
    .rptr_upd(rptr_upd),
    .rptr_rest(rptr_rest)
);
yaf_wptr_full #(
    .ASIZE(ASIZE),
    .ENABLE_PREWRITE(ENABLE_PREWRITE)
)
yaf_wptr_full (
    .wfull_val(wfull_val), 
    .wptr(wptr), 
    .wbin(wbin),
    .wq2_rptr(wq2_rptr),
    .wfull(wfull),
    .winc(winc), 
    .wclk(wclk),
    .wrst_n(wrst_n),
    .wptr_upd(wptr_upd),
    .wptr_rest(wptr_rest)
);
endmodule

