/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// TU-Dresden Chair for Embedded Systems
//////////////////////////////////////////////////////////////////////////////////

//Geschrieben von: Markus Noll, mail@markus-noll.de, TU-Darmstadt 2015


module concentrator_master
#(
	// SPMC Interface
	parameter BASE_ADR           = 10'h0,
	parameter BUFFER_SIZE        = 'd8,
	parameter NUMBER_OF_SLAVES   = 'd3,
	parameter SOFTWARE_ARBITRATION = 1

)
(
	//*** Verbindungen mit dem MC ***
	input   wire            clk_peri,	//Systemtakt
	input   wire [17:0]     do_peri,	//Datenbus von MC
	output  wire [17:0]     di_peri,	//Datenbus zu MC
	input   wire [9:0]      addr_peri,	//Adressbus von MC
	input   wire            access_peri,	//Signale fuer Peripheriezugriffe
	input   wire            wr_peri,	//Schreibsignal
	input   wire            reset,		//Reset-Signal
	
	// Verbindungen zum slave
	output reg [((NUMBER_OF_SLAVES*2)-1):0]		status,
	input wire [((NUMBER_OF_SLAVES*18)-1):0] 	data,
  	input wire [((NUMBER_OF_SLAVES*12)-1):0]	msg_size,
  	input wire [(NUMBER_OF_SLAVES-1):0]	 		data_available,
 	input wire [(NUMBER_OF_SLAVES-1):0]    		send
	
);
/////////////////////////////////////////////////////////////////////////////////////
//
// Local Functions and Parameters
//
function integer log2;
input integer value;
	for (log2=0; value>0; log2=log2+1)
	value = value>>1;
endfunction

localparam
    
	P_STATUS			= 0,
	P_MSG_SIZE			= 1,
    P_DATA_OUT      	= 2,
	P_PEEK_DATA			= 3,
	P_SOFT_ARBIT		= 4,
    
    ALL_PORTS			= 5,
    DECODSIM_DEB		= "NO",
    FIFO_ADDR_BITS   	= log2(BUFFER_SIZE - 1);


// Addressdekoder fuer alle Geraeteregister
wire							select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

wire		status_read_sel 	= reg_read[P_STATUS];
wire		msg_size_read_sel	= reg_read[P_MSG_SIZE];
wire  		data_out_read_sel 	= reg_read[P_DATA_OUT];
wire		peek_data_read_sel	= reg_read[P_PEEK_DATA];
wire		soft_arbit_read_sel = reg_read[P_SOFT_ARBIT];

wire		status_write_sel 	= reg_write[P_STATUS];
wire		msg_size_write_sel 	= reg_write[P_MSG_SIZE];
wire    	data_out_write_sel 	= reg_write[P_DATA_OUT];
wire		peek_data_write_sel	= reg_write[P_PEEK_DATA];
wire		soft_arbit_write_sel= reg_write[P_SOFT_ARBIT];

//reg_access_decoder anlegen
reg_access_decoder	
#(
	.REG_COUNT(ALL_PORTS),
	.BASE_ADR(BASE_ADR),
	.SIM_DEBUG(DECODSIM_DEB)
)
decod1			
(
	.reset(reset),
	.clk_peri(clk_peri),
	.addr_peri(addr_peri),
	.access_peri(access_peri),
	.wr_peri(wr_peri),

	.peri_select(select),
	.reg_select(reg_select),
	.reg_read(reg_read),
	.reg_write(reg_write)
);


//msg_size_in von slave
wire [11:0] msg_size_in;

//Ausgangssignale von fifo
wire[17:0] data_out_fifo0;
wire [FIFO_ADDR_BITS:0] fifo_fill;

//Eingangssignale für fifo
wire [17:0] data_to_buffer_x;
reg [17:0] data_to_buffer;

wire last_value_to_buffer;

wire send_to_buffer_x;
reg send_to_buffer;


//wires für arbitrierer
wire [8:0] slave_index_from_encoder;

//Status-, MSG_SIZE und Software-Arbitration-Register-Logik
reg [11:0] message_size;
reg [8:0] software_arbitration;

always @(posedge clk_peri)
begin
    if (reset) begin
        message_size <= 18'h1;
		software_arbitration <= 0;
    end else begin
		if (msg_size_write_sel) begin
        message_size <= do_peri;
   		end
		else if (soft_arbit_write_sel) begin
		software_arbitration <= do_peri;
		end
	end
end

wire [17:0] used_entries = 18'h00000 | fifo_fill;

wire fifo_empty = used_entries < message_size;

//Signale für di_peri
assign	di_peri [17:0]	= 	(status_read_sel) ? {17'h0000, ~fifo_empty} :
							(data_out_read_sel) ? data_out_fifo0 :
							(peek_data_read_sel && !fifo_empty) ? data_out_fifo0 : 18'b0;


//Daten von ausgewähltem Slave
assign data_to_buffer_x 		= {data[(slave_index_from_encoder*18) +: 18]};
assign last_value_to_buffer 	= !({data_available[slave_index_from_encoder]});
assign send_to_buffer_x 		= {send[slave_index_from_encoder]};
assign msg_size_in 				= {msg_size[(slave_index_from_encoder*12) +: 12]};


//Data to buffer, last value to buffer und send to buffer buffern, um speed > 60 MHz zu erreichen
always @ (posedge clk_peri) begin
	if (reset) begin
		data_to_buffer <= 18'b0;
		send_to_buffer <= 1'b0;
	end else begin
		data_to_buffer <= data_to_buffer_x;
		send_to_buffer <= send_to_buffer_x;
	end
end


//Generierung des Status-Signals inklusive Bufferung über Register um Speed > 60 MHz zu erreichen
wire [((NUMBER_OF_SLAVES*2)-1):0] status_x;

//Status-Signal zurück zum slave
assign status_x = 	((BUFFER_SIZE - fifo_fill) >= msg_size_in) ? (1000'd2<<(slave_index_from_encoder*2)) :	//genügend Platz frei
					((BUFFER_SIZE - fifo_fill) < msg_size_in ) ? (1000'd1<<(slave_index_from_encoder*2)) :	//nicht genügend Platz frei
					1000'd0;

//Signal buffern		
always @ (posedge clk_peri) begin
	if (reset) begin
		status <= 1000'd0;
	end else begin
		status <= status_x;
	end
end


generate
	if (SOFTWARE_ARBITRATION) begin
		//direct software arbitration
		assign slave_index_from_encoder = software_arbitration;

	end else begin
		//Round-Robin-Arbiter

		wire [NUMBER_OF_SLAVES-1:0] grant;

		concentrator_arbiter #
		(
			.NUMBER_OF_SLAVES(NUMBER_OF_SLAVES)
		) arbiter0 (
			.clk (clk_peri),
			.reset (reset),
			.last_value_write_sel (last_value_to_buffer),
			.req (data_available),   
			.grant(grant)  
		);
		   
		//Encoder für arbiter 
		//Wandelt one-hot grant in eine index-Zahl um
		one_hot_to_bin # 
		(
			.ONE_HOT_WIDTH(NUMBER_OF_SLAVES),
			.BIN_WIDTH(9)
		) encoder0 (   
			.reset(reset),
			.clk(clk_peri),
			.one_hot(grant),
			.binary_encoded(slave_index_from_encoder)
		);
	end
endgenerate
	

///////////////////////////////////////////////////////////////////
// Fifo logic module
// Using async fifo module with the same clock on both ports

xilinx_fifo_with_fill_counter #
(
 .WIDTH(18),
 .DEPTH(BUFFER_SIZE),
 .ADDR_BITS(FIFO_ADDR_BITS)
) 
	fifo0(
.clk(clk_peri),
.fill_count(fifo_fill),
.reset(reset),
.write_enable(send_to_buffer),
.write_data(data_to_buffer),
.read_enable(data_out_read_sel), 
.read_data(data_out_fifo0)
);

endmodule

