/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// TU-Dresden Chair for Embedded Systems
//////////////////////////////////////////////////////////////////////////////////

//Geschrieben von: Markus Noll, mail@markus-noll.de, TU-Darmstadt 2015

module concentrator_slave
#(
        // SPMC Interface
        parameter BASE_ADR    	= 10'h0,        
        parameter SLAVE_ID		= 'd0		
)
(
        //*** Connections with SpartanMC master ***
        input   wire            clk_peri,	//Systemtakt
        input   wire [17:0]     do_peri,	//Datenbus von MC
        output  wire [17:0]     di_peri,	//Datenbus zu MC
        input   wire [9:0]      addr_peri,	//Adressbus von MC
        input   wire            access_peri,//Signale fuer Peripheriezugriffe
        input   wire            wr_peri,	//Schreibsignal
        input   wire            reset,		//Reset-Signal
	
        // Verbindungen zu Konzentrator-Master
        input  wire [1:0]	    status,
        output wire [17:0]	    data,
        output reg  [11:0]		msg_size,
        output wire				data_available,
        output wire             send
);
/////////////////////////////////////////////////////////////////////////////////////
//
// Local Functions and Parameters
//

//Register
parameter 

    P_STATUS 			= 0,
    P_MSG_SIZE 			= 1,
    P_DATA_IN 			= 2,
	P_DATA_AVAILABLE	= 3, 
 	
    ALL_PORTS	  		= 4,
    DECODSIM_DEB   		="NO";


//Event-Wires anlegen
wire							select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

wire    data_in_read_sel 			= reg_read[P_DATA_IN];
wire    data_available_read_sel 	= reg_read[P_DATA_AVAILABLE];
wire	msg_size_read_sel			= reg_read[P_MSG_SIZE];
wire    status_read_sel 			= reg_read[P_STATUS];

wire    data_in_write_sel			= reg_write[P_DATA_IN];
wire    data_available_write_sel	= reg_write[P_DATA_AVAILABLE];
wire	msg_size_write_sel			= reg_write[P_MSG_SIZE];
wire    status_write_sel			= reg_write[P_STATUS];

reg_access_decoder	
#(
	.REG_COUNT(ALL_PORTS),
	.BASE_ADR(BASE_ADR),
	.SIM_DEBUG(DECODSIM_DEB)
)
decod1	
(
	.reset(reset),
	.clk_peri(clk_peri),
	.addr_peri(addr_peri),
	.access_peri(access_peri),
	.wr_peri(wr_peri),
	.peri_select(select),
	.reg_select(reg_select),
	.reg_read(reg_read),
	.reg_write(reg_write)
);


reg data_available_reg;
reg write_header;
reg [5:0] slave_id;
reg [11:0] msg_size_for_header;


//Reiche status-wire an di_peri weiter, wenn abgefragt wird von code
assign	di_peri [17:0]	=		(status_read_sel) ? (status == 2) ?	{18'd1} :	//buffer space available
													(status == 1) ?	{18'd0} :	//not enough buffer space available
													{18'd2} : 18'd0;			//no send permission/slave not selected

//Daten eingang do_peri auf data-bus
assign data = 	(data_in_write_sel & !write_header) ? do_peri[17:0] : 
	    		(data_in_write_sel & write_header) ? {(msg_size_for_header[11:0]),slave_id[5:0]} : 18'b0;

//send signal bei Dateneingang
assign send = data_in_write_sel;

//Datenverfügbar ab data_available_write_sel
assign data_available = data_available_reg;


always @(posedge clk_peri)
begin
	if (reset) begin
      //Register zurücksetzen
      data_available_reg <= 1'b0;       
      slave_id <= SLAVE_ID;      
      msg_size <= 12'd0;
      write_header <= 1'b0;
                          
    end else if (data_available_write_sel) begin
    	//Data_available merken und beim ersten Schreibvorgang header schreiben
    	data_available_reg <= do_peri[0];
		msg_size_for_header <= do_peri[12:1];
    	write_header <= 1'b1;
      	
    end else if (msg_size_write_sel) begin
    	//+1 wegen header
    	msg_size <= do_peri[11:0];
    	
    end else if (data_in_write_sel) begin
    	write_header <= 1'b0;
    	
    end 
end

endmodule
