/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  ALU                                                        ////
////  Dies ist die ALU des SpartanMC. Sie wird durch eine        ////
////  4-Bit breiten Fuctioncode und ein Signal was angibt        ////
////  ob die Operanten fuer die Vergleich vorzeichenbehafetet    ////
////  oder frei sind gesteuert. Das Ergbnis der Berechnung und   ////
////  der neue Wert fuer das cc Register werden von diesem       ////
////  Modul geliefert.                                           ////
////                                                             ////
////  Author: Falk Hassler                                       ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`include	"core/src/spartanmc_def.vh"
`default_nettype none

module alu	(
		opa,
		opb,
		result,
		cc_result,
		mem_addr,
		is_signed,
		do_ov,
		func
		);

// Die Folgenden Parameter werden durch die system.v bei der Erzeugung des Modules ueberschrieben.
parameter	RESET_VECTOR	=	0;
parameter	INTR_VECTOR	=	32;
parameter	SINGLE_SHIFT	=	1;
parameter	HARDWARE_MUL	=	1;

input wire	[17:0]	opa;
input wire	[17:0]	opb;
output	[17:0]	result;
output		cc_result;
output	[17:0]	mem_addr;
input wire		is_signed;
input		do_ov;
input wire	[3:0]	func;


wire	[17:0]	op_a;
wire	[17:0]	op_b;
wire	[18:0]  shiftr;
wire	[17:0]	summe;
wire	[18:0]	differenz;
wire		insert;
wire		is_null;
wire		is_negativ;
reg	[17:0]	result;
reg		cc_bit;
reg		cc_shift;
wire		cc_result = cc_bit | cc_shift;
wire		do_ov;
wire		cc_ova;
wire		cc_ovs;
wire	[17:0]	mem_addr;

wire	[4:0]	sh_op;

function [17:0] sign_ex;
input [17:0]	a;
input [4:0]	b;
	
	case(b)
		default	: sign_ex = {{10 {a[7] }}, a[7:0] };
		5'd9	: sign_ex = {{9  {a[8] }}, a[8:0] };
		5'd16	: sign_ex = {{2  {a[15]}}, a[15:0]};
	endcase

endfunction

// shift Schritte festlegen
generate
	if	(SINGLE_SHIFT == 1)
	assign	sh_op	= 5'b1;
	else
	assign	sh_op	= op_b[4:0];
endgenerate

assign	cc_ova		= ((opa[17] ==  opb[17])&(result[17] != opa[17])& do_ov)? 1'b1 : 1'b0;
assign	cc_ovs		= ((opa[17] == !opb[17])&(result[17] != opa[17])& do_ov)? 1'b1 : 1'b0;

assign	op_a[16:0]	= opa[16:0];
assign	op_a[17]	= is_signed ? !opa[17] : opa[17];
assign	op_b[16:0]	= opb[16:0];
assign	op_b[17]	= is_signed ? !opb[17] : opb[17];
assign	insert		= (func == `ALU_SRA) ? op_a[17] : 1'b0;
assign	shiftr[18:0]	= { {18{insert}}, op_a, 1'b0 } >> sh_op[4:0];

assign	summe		= op_a + op_b;
assign	differenz	= op_a - op_b;
assign	mem_addr	= summe[17:0];

assign	is_null		= ! ( differenz[17] | differenz[16] | differenz[15] | differenz[14] | 
			      differenz[13] | differenz[12] | differenz[11] | differenz[10] | 
			      differenz[ 9] | differenz[ 8] | differenz[ 7] | differenz[ 6] | 
			      differenz[ 5] | differenz[ 4] | differenz[ 3] | differenz[ 2] |
			      differenz[ 1] | differenz[ 0] );

assign	is_negativ	= differenz[18];

always @(*)
	case(func)
		`ALU_ADD:	{cc_shift,result} = {cc_ova,summe};
		`ALU_AND:	{cc_shift,result} = {1'b0,(op_a & op_b)};
		`ALU_OR:	{cc_shift,result} = {1'b0,(op_a | op_b)};
		`ALU_XOR:	{cc_shift,result} = {1'b0,(op_a ^ op_b)};
		`ALU_SLL:	{cc_shift,result} = {1'b0, op_a} << sh_op[4:0];
		`ALU_SLR:	{result,cc_shift} = shiftr[18:0];
		`ALU_SRA:	{result,cc_shift} = shiftr[18:0];
		`ALU_SIGNEX:	{cc_shift,result} = {1'b0,(sign_ex(op_a, op_b[4:0]))};
		`ALU_PASSB:	{cc_shift,result} = {1'b0,op_b};
		default:	{cc_shift,result} = {cc_ovs,differenz[17:0]};
	endcase

always @(is_null or is_negativ or func)
	case(func)
		`ALU_SEQ:	cc_bit = is_null;
		`ALU_SNE:	cc_bit = !is_null;
		`ALU_SLT:	cc_bit = is_negativ;
		`ALU_SGT:	cc_bit = !is_negativ& !is_null;
		`ALU_SLE:	cc_bit = is_negativ | is_null;
		`ALU_SGE:	cc_bit = !is_negativ;
		default:	cc_bit = 1'b0;
	endcase

endmodule
