/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`include	"core/src/spartanmc_def.vh"
`include	"core/util/pack_unpack.vh"
`default_nettype none

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

module breakpoint_mgr #(
	parameter COUNT_BREAKPOINTS = 4,
	parameter COUNT_WATCHPOINTS = 4,
	parameter BRKPT_TRAP = 1,
	parameter WTCHPT_TRAP = 2,
	parameter SINGLE_STEP_TRAP = 3
) (
	input wire							clk,
	input wire							reset,
	input wire							clk_en,
	input wire		[`ADDR_WIDTH-1:0]	pc_next,//next ir
	input wire		[`ADDR_WIDTH-1:0]	mem_addr, // word addressed
	input wire					[1:0]	mem_we, // which byte is actually written: upper, lower
	input wire					[1:0]	mem_re,// which byte is actually read: upper, lower
	
	input wire		[`ADDR_WIDTH *COUNT_BREAKPOINTS-1:0]	brkpts_packed,
	input wire		[COUNT_BREAKPOINTS-1:0]					brkpts_config,
	input wire		[`ADDR_WIDTH *COUNT_WATCHPOINTS-1:0]	wtchpts_packed,
	input wire		[5 *COUNT_WATCHPOINTS-1:0]				wtchpts_config_packed,
	input wire							enable,
	
	input wire							step_next,
	
	input wire		[6:0]				reg_window,
	
	input wire							delayed_ir,
	
	output reg							spring_trap,
	output reg					[7:0]	trap_no,
	output reg		[`ADDR_WIDTH-1:0] 	trapped_at,
	output reg		[`ADDR_WIDTH-1:0]	trapped_mem_addr // in byte TODO: should have 19 bits width
	//~ ,output wire		[56:0]				chipscope_out
);

wire		[`ADDR_WIDTH-1:0]	brkpts [0:COUNT_BREAKPOINTS-1];
wire		[`ADDR_WIDTH-1:0]	wtchpts [0:COUNT_WATCHPOINTS-1];
wire		[4:0]				wtchpts_config [0:COUNT_WATCHPOINTS-1];

`UNPACK_GEN
generate
`UNPACK_ARRAY(`ADDR_WIDTH, COUNT_BREAKPOINTS, brkpts, brkpts_packed, UPK1) `PACK_ARRAY_END
`UNPACK_ARRAY(`ADDR_WIDTH, COUNT_WATCHPOINTS, wtchpts, wtchpts_packed, UPK3) `PACK_ARRAY_END
`UNPACK_ARRAY(5, COUNT_WATCHPOINTS, wtchpts_config, wtchpts_config_packed, UPK4) `PACK_ARRAY_END
endgenerate

wire [COUNT_BREAKPOINTS-1:0] hit_brkpts;
wire [COUNT_WATCHPOINTS-1:0] hit_wtchpts;

wire hit_brkpt, hit_wtchpt, hit_single;

assign hit_brkpt = |hit_brkpts;
assign hit_wtchpt = |hit_wtchpts;

wire mem_addr_0;
assign mem_addr_0 = (!(&mem_re) && mem_re[0]) || (!(&mem_we) && mem_we[0]);


function eval_breakpoint;
	input [`ADDR_WIDTH-1:0] pc;
	input [`ADDR_WIDTH-1:0] breakpoint;
	input					breakpoint_active;
	
	begin
		if (breakpoint_active) begin
			eval_breakpoint = pc == breakpoint;
		end else begin
			eval_breakpoint = 0;
		end
	end
endfunction

function eval_watchpoint;
	input [`ADDR_WIDTH-1:0] mem_addr;
	input [1:0] mem_we; //write access to bytes at addr
	input [1:0] mem_re; //read access to bytes at addr
	input [`ADDR_WIDTH-1:0] watchpoint;
	input [4:0] wtchpt_config;
	// active, read_en, write_en, bytes_active_1, bytes_active_0

	reg read_hit;
	reg write_hit;
	reg addr_hit;

	begin
		if (wtchpt_config[4]) begin
			addr_hit = mem_addr == watchpoint;
			read_hit = |(mem_re & wtchpt_config[1:0]) & addr_hit;
			write_hit = |(mem_we & wtchpt_config[1:0]) & addr_hit;
			eval_watchpoint = (read_hit && wtchpt_config[3]) || (write_hit && wtchpt_config[2]);
		end else begin
			eval_watchpoint = 0;
		end
	end
endfunction

genvar i;
generate
	for (i=0; i<COUNT_BREAKPOINTS; i=i+1) begin: eval_breakpoints
		assign hit_brkpts[i] = eval_breakpoint(pc_next, brkpts[i], brkpts_config[i]);
	end
	for (i=0; i<COUNT_WATCHPOINTS; i=i+1) begin: eval_watchpoints
		assign hit_wtchpts[i] = eval_watchpoint(mem_addr, mem_we, mem_re, wtchpts[i], wtchpts_config[i]);
	end

endgenerate

reg [`ADDR_WIDTH-1:0]	pc;
reg [`ADDR_WIDTH-1:0]	pc_last;
reg [`ADDR_WIDTH-1:0]	pc_last2;
reg [`ADDR_WIDTH-1:0]	pc_last3;
//~ reg						hit_wtchpt_last;

//~ reg						delayed_ir_last;

reg [6:0] 				hit_window;
reg [`ADDR_WIDTH-1:0]	hit_pc;
reg [1:0]				hit_ignore;
reg						enable_last;
wire allow_brkpt, allow_step;
wire enable_internal;

assign hit_single = step_next;
assign enable_internal = (hit_ignore == 1 && hit_window != reg_window) || (hit_ignore >= 2);
assign allow_brkpt = (hit_ignore == 2) || ((hit_ignore == 1) && (pc_next != hit_pc) && (hit_window != reg_window));
assign allow_step = (hit_ignore == 2);

always @(*) begin
	if (enable_internal) begin // only really enable after return from breakpoint handler
		if (hit_single && allow_step) begin
			spring_trap = 1;
			trap_no = SINGLE_STEP_TRAP;
		end else if (hit_brkpt && allow_brkpt) begin
			spring_trap = 1;
			trap_no = BRKPT_TRAP;
		end else if (hit_wtchpt) begin
			spring_trap = 1;
			trap_no = WTCHPT_TRAP;
		end else begin
			spring_trap = 0;
			trap_no = 0;
		end
	end else begin
		spring_trap = 0;
		trap_no = 0;
	end
end

always @(posedge clk) begin
	if (reset) begin
		pc <= 0;
		pc_last <= 0;
		pc_last2 <= 0;
		pc_last3 <= 0;
		//~ hit_wtchpt_last <= 0;
		hit_window <= 0;
		hit_pc <= 0;
		hit_ignore <= 0;
		//~ delayed_ir_last <= 0;
		trapped_at <= 0;
		trapped_mem_addr <= 0;
		enable_last <= enable;
	end else if (clk_en) begin
		pc <= pc_next;
		pc_last <= pc;
		pc_last2 <= pc_last;
		pc_last3 <= pc_last2;
		enable_last <= enable;
		//~ hit_wtchpt_last <= hit_wtchpt;
		//~ delayed_ir_last <= delayed_ir;
		if (spring_trap) begin // remember locations on trap
			trapped_at <= (hit_wtchpt)? ((delayed_ir)? pc_last2 : pc_last) : pc_next;
			if (hit_wtchpt)
				trapped_mem_addr <= {mem_addr,mem_addr_0};
			hit_pc <= pc_next;
		end
		
		if (hit_ignore == 0 && enable) begin// 0: off
			hit_ignore <= 1;
			hit_window <= reg_window;
		end else if (hit_ignore == 1 && hit_window != reg_window) begin// 1: wait for return from breakpoint handler // : accept traps on first insn after breakpoint handler
			hit_ignore <= 2; // 2: accept traps on second insn after breakpoint handler (if the first insn is a step or last trapped insn)
		end else if (!enable) begin // if disabled, reset to 0
			hit_ignore <= 0;
		end
		
	end
end

//~ assign chipscope_out[0] = enable;
//~ assign chipscope_out[1] = step_next;
//~ assign chipscope_out[7:2] = reg_window;
//~ assign chipscope_out[13:8] = hit_window;
//~ assign chipscope_out[15:14] = hit_ignore;
//~ assign chipscope_out[34:16] = {pc_next, 1'b0};
//~ assign chipscope_out[53:35] = {hit_pc, 1'b0};
//~ assign chipscope_out[54] = spring_trap;


endmodule


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Decodes the 19 bit address + the load / store 9 bit into a word level address (18 bit) + a read enable signal per byte
*/
/*module mem_access_decoder (
	input wire		[`ADDR_WIDTH:0]		addr, // full 19 bit
	input wire							sl9,
	input wire		[1:0]				we,
	output wire	[`ADDR_WIDTH-1:0]	word_addr,
	output wire	[1:0]				re
);

wire read;
wire which_byte;

assign word_addr = addr[`ADDR_WIDTH:1];
assign which_byte = addr[0];
assign read = |we;
assign re = read & ((sl9)? ((which_byte)? 2'b10 : 2'b01 ) : 2'b11);


endmodule
*/
