/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  Control                                                    ////
////  In dieser Stufe werden alle Steuersignale, die die         ////
////  einzelnen Stufen benoetigen, um fuer jeden einzelnen       ////
////  Befehl den richtigen Datenpfad einzustellen, erzeugt.      ////
////  Dazu wird der Opcode des Befehles durch einen grossen      ////
////  Decoder verarbeitet.                                       ////
////                                                             ////
////  Author: Falk Hassler                                       ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`include	"core/src/spartanmc_def.vh"
`default_nettype none

// Darf nicht control heißen, da es sonst Konflikte mit IP von Lattice gibt
module control_logic	(
		opcode,
		intr,
		run_intr,
		inc_window,
		dec_window,
		reg_to_pc,
		link_pc,
		is_jump,
		is_branch,
		is_delay,
		branch_typ,
		branch_neg,
		is_mem,
		is_store,
		is_imm,
		sign_ext,
		load_high,
		short_trans,
		set_sfr,
		get_sfr,
		change_bit,
		set_bit,
		do_wb_cc,
		do_ov,
		do_wb,
		conditional_wb,
		multiply,
		is_signed,
		alu_func,
		ir_return,
		trap,
		is_not,
		exc_trap_exec
		);


input wire	[9:0]	opcode;
input wire			intr;		// intr darf nur bis zum Start der Interruptbehandlung 1 sein
input wire			exc_trap_exec;
output reg			run_intr;	// Start der Interuptbehandlung
output wire			inc_window;
output wire			dec_window;
output wire			reg_to_pc;
output wire			link_pc;
output wire			is_jump;
output wire			is_branch;
output wire			is_delay;
output wire			branch_typ;
output wire			branch_neg;
output wire			is_mem;
output wire			is_store;
output wire			is_imm;
output wire			sign_ext;
output wire			load_high;
output wire			short_trans;
output wire			set_sfr;
output wire			get_sfr;
output wire			change_bit;
output reg			set_bit;
output wire			do_wb_cc;
output wire			do_ov;
output wire			do_wb;
output wire			conditional_wb;
output wire			multiply;
output wire			is_signed;
output wire	[3:0]	alu_func;
output wire			ir_return;
output wire			trap;
output wire			is_not;

/* Steuersignal
	inc_window	auf positive Flanke wird das WindowReg um 1 erhoeht
	dec_window	auf negative Flanke wird das WindowReg um 1 decrementiert
	reg_to_pc	ist 1 wenn das Register a in den PC geladen werden soll
	link_pc		ist 1 wenn PC ins Register R11 geschoben werden soll
	is_jump		ist 1 wenn es sich um einen Jump   Befehl handelt
	is_branch	ist 1 wenn es sich um einen Branch Befehl handelt
	branch_typ	ist 1 wenn Vergleich mit Register cc und 0 wenn Vergleich mit
			Register a
	branch_neg	ist 1 wenn ungleich und 0 wenn gleich
	is_mem		ist 1 wenn es sich um einen Speicherbefehl handelta
	is_store	ist 1 wenn es ein Store Befehl ist und ist 0 wenn es ein
			Load Befehl ist
	is_imm		ist 1 wenn es ein Immediate Befehl ist
	sign_ext	ist 1 wenn eine Vorzeichenerweiterung beim Immediate
			vorgenommen werden soll
	load_high	ist 1 wenn die oberen 9 Bit des Immediate verwendet werden sollen
	short_trans	ist 1 bei s9 oder l9
	do_wb_cc	ist 1 wenn das Ergebnis in das CC Bit geschrieben werden soll
	do_ov		ist 1 wenn ein OverFlow in das CC Bit geschrieben werden soll
	do_wb		ist 1 wenn das Ergebnis in ein Register geschrieben werden soll
	conditional_wb	Writeback ist Abhaenging vom cc Bit. cc == 1 dann writeback 
			sonst nicht.
	multiply	ist 1 wenn MUL Befehl ausgefuehrt wird sonst 0
	is_signed	ist 1 wenn es sich um einen Vorzeichenbehafteten Ver-
			gleichsbefehl handet, sonst 0
	alu_func	ALU Funktionscode
	bypass_activ	ist 1 wenn er Bypass der Ex/Mem Stufe freigegeben ist
	ir_return	ist 1 wenn RFE Befehl ausgefuehrt wird, sonst 0
	trap		ist 1 wenn TRAP Befehl ausgefuehrt wird, sonst 0
*/

reg	[28:0]	control;
// Aufteilung der Steuersignale im control Word
assign	ir_return	=	control[28];
assign	trap		=	control[27];
assign	is_not		=	control[26];

assign	inc_window	=	control[25];
assign	dec_window	=	control[24];

assign	reg_to_pc	=	control[23];
assign	link_pc		=	control[22];
assign	is_jump		=	control[21];
assign	is_branch	=	control[20];
assign	is_delay	=	control[19];

assign	is_mem		=	control[18];
assign	is_store	=	control[17];

assign	is_imm		=	control[16];
assign	sign_ext	=	control[15];
assign	load_high	=	control[14];
assign	short_trans	=	control[13];

assign	set_sfr		=	control[12];
assign	get_sfr		=	control[11];
assign	change_bit	=	control[10];

assign	do_wb_cc	=	control[ 9];
assign	do_ov		=	control[ 8];
assign	do_wb		=	control[ 7];
assign	conditional_wb	=	control[ 6];

assign	multiply	=	control[ 5];
assign	is_signed	=	control[ 4];

assign	alu_func	=	control[3:0];
// Ende der Bitaufteilung im control Word

assign	branch_typ	=	opcode[6];
assign	branch_neg	=	opcode[5];

always @(opcode or intr or exc_trap_exec)
begin
	run_intr	=	1'b0;
	set_bit		=	(opcode == `SBITS) ? 1'b1 : 1'b0;
	if (exc_trap_exec)
					control = {25'b000_10_01100_00_0000_000_0010_00, `ALU_PASSB};
	else if (intr)
		begin
					control = {25'b000_10_01100_00_0000_000_0010_00, `ALU_PASSB};
					run_intr = 1'b1;
		end 
	else
		casex (opcode)
			default: control = {25'b000_00_00000_00_0000_000_0000_00, `ALU_DONTCARE};

			`JALRS	: control = {25'b000_10_11101_00_0000_000_0010_00, `ALU_PASSB};
			`JALS	: control = {25'b000_10_01101_00_0000_000_0010_00, `ALU_PASSB};
			`TRAP	: control = {25'b010_10_01100_00_0000_000_0010_00, `ALU_PASSB};
			`JRS	: control = {25'b000_01_10101_00_0000_000_0000_00, `ALU_DONTCARE};
			`RFE	: control = {25'b100_01_10101_00_0000_000_0000_00, `ALU_DONTCARE};

			`JALR	: control = {25'b000_00_11101_00_0000_000_0010_00, `ALU_PASSB};
			`JR		: control = {25'b000_00_10101_00_0000_000_0000_00, `ALU_DONTCARE};
			`J		: control = {25'b000_00_00101_00_0000_000_0000_00, `ALU_DONTCARE};

			`BEQZ	: control = {25'b000_00_00011_00_0100_000_0000_00, `ALU_DONTCARE};
			`BNEZ	: control = {25'b000_00_00011_00_0100_000_0000_00, `ALU_DONTCARE};
			`BEQZC	: control = {25'b000_00_00011_00_0100_000_0000_00, `ALU_DONTCARE};
			`BNEZC	: control = {25'b000_00_00011_00_0100_000_0000_00, `ALU_DONTCARE};

		
			`S9		: control = {25'b000_00_00000_11_0001_000_0000_00, `ALU_ADD};
			`S18	: control = {25'b000_00_00000_11_0000_000_0000_00, `ALU_ADD};
			`L9		: control = {25'b000_00_00000_10_0001_000_0010_00, `ALU_ADD};
			`L18	: control = {25'b000_00_00000_10_0000_000_0010_00, `ALU_ADD};
			`SWAP	: control = {25'b000_00_00000_11_0000_000_0010_00, `ALU_ADD};
		
			`MULI	: control = {25'b000_00_00000_00_1100_100_0010_10, `ALU_DONTCARE};
			`SLTI	: control = {25'b000_00_00000_00_1100_000_1000_01, `ALU_SLT};
			`SGTI	: control = {25'b000_00_00000_00_1100_000_1000_01, `ALU_SGT};
			`SLEI	: control = {25'b000_00_00000_00_1100_000_1000_01, `ALU_SLE};
			`SGEI	: control = {25'b000_00_00000_00_1100_000_1000_01, `ALU_SGE};
			`SEQI	: control = {25'b000_00_00000_00_1100_000_1000_00, `ALU_SEQ};
			`SNEI	: control = {25'b000_00_00000_00_1100_000_1000_00, `ALU_SNE};
		
			`ADDI	: control = {25'b000_00_00000_00_1100_000_0010_00, `ALU_ADD};
			`LHI	: control = {25'b000_00_00000_00_1010_000_0010_00, `ALU_PASSB};
			`IFADDUI: control = {25'b000_00_00000_00_1000_000_0011_00, `ALU_ADD};
			`IFSUBUI: control = {25'b000_00_00000_00_1000_000_0011_00, `ALU_SUB};

			`SIGEX	: control = {25'b000_00_00000_00_1000_000_0010_00, `ALU_SIGNEX};
			`MOVI	: control = {25'b000_00_00000_00_1000_000_0010_00, `ALU_PASSB};

			`ANDI	: control = {25'b000_00_00000_00_1000_000_0010_00, `ALU_AND};
			`ORI	: control = {25'b000_00_00000_00_1000_000_0010_00, `ALU_OR};
			`XORI	: control = {25'b000_00_00000_00_1000_000_0010_00, `ALU_XOR};
			`SLLI	: control = {25'b000_00_00000_00_1000_000_1010_00, `ALU_SLL};
			`SRLI	: control = {25'b000_00_00000_00_1000_000_1010_00, `ALU_SLR};
			`SRAI	: control = {25'b000_00_00000_00_1000_000_1010_00, `ALU_SRA};

			`MUL	: control = {25'b000_00_00000_00_0000_100_0010_10, `ALU_DONTCARE};
			`MOVI2S	: control = {25'b000_00_00000_00_0000_100_0000_00, `ALU_DONTCARE};
			`MOVS2I	: control = {25'b000_00_00000_00_0000_010_0010_00, `ALU_DONTCARE};

			`SBITS	: control = {25'b000_00_00000_00_0000_001_0000_00, `ALU_DONTCARE};
			`CBITS	: control = {25'b000_00_00000_00_0000_001_0000_00, `ALU_DONTCARE};


			`ADD	: control = {25'b000_00_00000_00_0000_000_1110_00, `ALU_ADD};
			`ADDU	: control = {25'b000_00_00000_00_0000_000_0010_00, `ALU_ADD};
			`SUB	: control = {25'b000_00_00000_00_0000_000_1110_00, `ALU_SUB};
			`SUBU	: control = {25'b000_00_00000_00_0000_000_0010_00, `ALU_SUB};
			`AND	: control = {25'b000_00_00000_00_0000_000_0010_00, `ALU_AND};
			`OR		: control = {25'b000_00_00000_00_0000_000_0010_00, `ALU_OR};
			`XOR	: control = {25'b000_00_00000_00_0000_000_0010_00, `ALU_XOR};
			`NOT	: control = {25'b001_00_00000_00_0000_000_0010_00, `ALU_XOR};

			`SLL	: control = {25'b000_00_00000_00_0000_000_1010_00, `ALU_SLL};
			`SRL	: control = {25'b000_00_00000_00_0000_000_1010_00, `ALU_SLR};
			`SRA	: control = {25'b000_00_00000_00_0000_000_1010_00, `ALU_SRA};
			`MOV	: control = {25'b000_00_00000_00_0000_000_0010_00, `ALU_PASSB};

			`SLT	: control = {25'b000_00_00000_00_0000_000_1000_01, `ALU_SLT};
			//`SGT	: control = {25'b000_00_00000_00_0000_000_1000_01, `ALU_SGT};
			`SLE	: control = {25'b000_00_00000_00_0000_000_1000_01, `ALU_SLE};
			//`SGE	: control = {25'b000_00_00000_00_0000_000_1000_01, `ALU_SGE};
			`SEQ	: control = {25'b000_00_00000_00_0000_000_1000_00, `ALU_SEQ};
			//`SNE	: control = {25'b000_00_00000_00_0000_000_1000_00, `ALU_SNE};

			`SLTU	: control = {25'b000_00_00000_00_0000_000_1000_00, `ALU_SLT};
			//`SGTU	: control = {25'b000_00_00000_00_0000_000_1000_00, `ALU_SGT};
			`SLEU	: control = {25'b000_00_00000_00_0000_000_1000_00, `ALU_SLE};
			//`SGEU	: control = {25'b000_00_00000_00_0000_000_1000_00, `ALU_SGE}; // not used by gcc
			//`SEQU	: control = {25'b000_00_00000_00_0000_000_1000_00, `ALU_SEQ}; // senseless!!
			//`SNEU	: control = {25'b000_00_00000_00_0000_000_1000_00, `ALU_SNE};

		endcase
end

endmodule
