/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

/*
	Register numbers:
	+---------+------------------------------------------------+
	| Number  | Register                                       |
	+---------+------------------------------------------------+
	| 0x00    | Cycle/Event counter configuration              |
	+---------+------------------------------------------------+
	| 0x01    | Cycle/Event counter value (bits 0-17)          |
	+---------+------------------------------------------------+
	| 0x02    | Cycle/Event counter value (bits 18-35)         |
	+---------+------------------------------------------------+
	

	Register layouts:
	
	reg_idx:
	+---------+------------------------------------------------+
	| Bit     | Function                                       |
	+---------+------------------------------------------------+
	| 0-3     | Register number.                               |
	+---------+------------------------------------------------+
	| 4-7     | Selected cycle/event counter.                  |
	+---------+------------------------------------------------+
	| 8       | 0: Access performance counter register.        |
	|         | 1: Access event counter register.              |
	+---------+------------------------------------------------+
	| 10      | Writing 1 disables counter 3.                  |
	+---------+------------------------------------------------+
	| 11      | Writing 1 enables counter 3.                   |
	+---------+------------------------------------------------+
	| 12      | Writing 1 disables counter 2.                  |
	+---------+------------------------------------------------+
	| 13      | Writing 1 enables counter 2.                   |
	+---------+------------------------------------------------+
	| 14      | Writing 1 disables counter 1.                  |
	+---------+------------------------------------------------+
	| 15      | Writing 1 enables counter 1.                   |
	+---------+------------------------------------------------+
	| 16      | Writing 1 disables counter 0.                  |
	+---------+------------------------------------------------+
	| 17      | Writing 1 enables counter 0.                   |
	+---------+------------------------------------------------+
	
	ccnt_conf:
	+---------+------------------------------------------------+
	| Bit     | Function                                       |
	+---------+------------------------------------------------+
	| 0       | 1: Counter enabled. 0: Counter disabled.       |
	+---------+------------------------------------------------+
	| 1       | Counter reset.                                 |
	+---------+------------------------------------------------+
	| 2       | Interrupt filter enable.                       |
	+---------+------------------------------------------------+
	| 4-7     | Prescaler:                                     |
	|         |   0: No prescaler                              |
	|         |   1: Prescaler 2                               |
	|         |   2: Prescaler 4                               |
	|         |   3: Prescaler 8                               |
	|         |   ...                                          |
	|         |  15: Prescaler 32768                           |
	+---------+------------------------------------------------+
	| 8       | Cycle counter overflow flag.                   |
	|         | Writing 1 will clear the flag.                 |
	+---------+------------------------------------------------+
	| 9-11    | Number of cycle counters.                      |
	+---------+------------------------------------------------+
	| 12-16   | Number of event counters.                      |
	+---------+------------------------------------------------+
	
	evcnt_conf:
	+---------+------------------------------------------------+
	| Bit     | Function                                       |
	+---------+------------------------------------------------+
	| 0-3     | Selected event number.                         |
	+---------+------------------------------------------------+
	| 4-5     | Associated cycle counter (see bit 6).          |
	+---------+------------------------------------------------+
	| 6       | Tie event counter to cycle counter.            |
	+---------+------------------------------------------------+
	| 7       | 1: Counter enabled. 0: Counter disabled.       |
	+---------+------------------------------------------------+
	| 8       | Counter reset.                                 |
	+---------+------------------------------------------------+
	| 9       | Interrupt filter enable.                       |
	+---------+------------------------------------------------+
	| 10      | Enable edge detection.                         |
	+---------+------------------------------------------------+
	| 11      | Event counter overflow flag.                   |
	|         | Writing 1 will clear the flag.                 |
	+---------+------------------------------------------------+
	| 12-15   | Prescaler:                                     |
	|         |   0: No prescaler                              |
	|         |   1: Prescaler 2                               |
	|         |   2: Prescaler 4                               |
	|         |   3: Prescaler 8                               |
	|         |   ...                                          |
	|         |  15: Prescaler 32768                           |
	+---------+------------------------------------------------+
*/

`include	"core/src/spartanmc_def.vh"
`default_nettype none

`define PCNT_RESETVALUE		18'h0
`define EVCNT_RESETVALUE	18'h0

/*
* This is not a typical prescaler. Instead, clk_out will be high for one
* period of clk_in after 2^div rising edges on clk_in.
*/
module clk_prescaler (
	input wire clk_in,
	input wire [3:0] div,
	output wire clk_out
);
	reg [15:0] n = 16'h1;
	
	assign clk_out = div ? n[div] : 1'h1;

	always @(posedge clk_in)
		if(n[div])
			n <= 1'h1;
		else
			n <= n + 1'h1;
endmodule

module event_prescaler (
	input wire clk_in,
	input wire evt_in,
	input wire [3:0] div,
	output wire evt_out
);
	reg [15:0] n = 16'h0;
	
	assign evt_out = div ? n[div] : evt_in;
	
	always @(posedge clk_in)
		if(n[div])
			n <= evt_in;
		else if(evt_in)
			n <= n + 1'h1;
endmodule

module core_edge_detector (
	input wire clk,
	input wire reset,
	input wire in,
	input wire en,
	output wire out
);

wire o;
reg filter = 1'h1;

assign o = in & filter;
assign out = en ? o : in;

always @(posedge clk) begin
	if(reset)
		filter <= 1'h1;
	else
		filter <= ~in;
end

endmodule

module intr_filter (
	input wire clk,
	input wire reset,
	input wire intr_start,
	input wire intr_end,
	output wire out
);

reg state = 1'h1;

assign out = state;

always @(posedge clk) begin
	if(reset | intr_end)
		state <= 1'h1;
	else if(intr_start)
		state <= 1'h0;
end

endmodule

module perfcounter #(
	parameter COUNT_CYCLECOUNTERS = 1,
	parameter COUNT_EVENTCOUNTERS = 2,
	parameter WIDTH_CYCLECOUNTER = 36,
	parameter WIDTH_EVENTCOUNTER = 36
) (
	input wire clk,
	input wire reset,
	input wire reg_write,
	input wire [17:0] reg_idx,
	output reg [17:0] reg_data_out = 0,
	input wire [17:0] reg_data_in,
	
	input wire data_read,
	input wire data_write,
	input wire instr_read,
	input wire instr_hit,
	input wire instr_miss,
	input wire pipeline_stall,
	input wire is_nop,
	input wire intr_start,
	input wire intr_end,
	input wire ext0,
	input wire ext1
);

wire [2:0] num_cyclecnt = COUNT_CYCLECOUNTERS;
wire [4:0] num_eventcnt = COUNT_EVENTCOUNTERS;
wire [6:0] selected_reg = reg_idx[3:0];
wire [4:0] selected_cnt = reg_idx[7:4];
wire sel_evcnt = reg_idx[8];

// Cycle counters:
reg [17:0] ccnt_conf[COUNT_CYCLECOUNTERS-1:0];
reg [WIDTH_CYCLECOUNTER:0] ccnt_counter[COUNT_CYCLECOUNTERS-1:0]; // Extra bit for overflow flag.
reg ccnt_overflow[COUNT_CYCLECOUNTERS-1:0];
wire [35:0] read_ccnt = {{(36-WIDTH_CYCLECOUNTER){1'h0}}, ccnt_counter[selected_cnt][WIDTH_CYCLECOUNTER-1:0]};
wire [3:0] ccnt_ovf_event;

// Event counters:
wire events_edge_detected [COUNT_EVENTCOUNTERS-1:0];
wire events_prescaled [COUNT_EVENTCOUNTERS-1:0];
reg [17:0] evcnt_conf [COUNT_EVENTCOUNTERS-1:0];
reg [WIDTH_EVENTCOUNTER:0] evcnt_counter [COUNT_EVENTCOUNTERS-1:0]; // Extra bit for overflow flag.
reg evcnt_overflow [COUNT_EVENTCOUNTERS-1:0];
wire [35:0] read_evcnt = {{(36-WIDTH_EVENTCOUNTER){1'h0}}, evcnt_counter[selected_cnt][WIDTH_EVENTCOUNTER-1:0]};

// Event lines.
wire [13:0] events;
assign events = {
	ccnt_ovf_event[3],
	ccnt_ovf_event[2],
	ccnt_ovf_event[1],
	ccnt_ovf_event[0],
	ext0,
	ext1,
	is_nop,
	intr_start,
	data_read,
	data_write,
	instr_read,
	instr_hit,
	instr_miss,
	pipeline_stall
	};

// Read registers.
always @(*) begin
	if(sel_evcnt)
		if(selected_cnt >= COUNT_EVENTCOUNTERS)
			reg_data_out <= 18'h0;
		else case (selected_reg)
			7'h0: reg_data_out <= {2'h0,
			                       evcnt_conf[selected_cnt][15:12],
			                       evcnt_overflow[selected_cnt] | evcnt_counter[selected_cnt][WIDTH_EVENTCOUNTER],
			                       evcnt_conf[selected_cnt][10:0]};
			7'h1: reg_data_out <= read_evcnt[17:0];
			7'h2: reg_data_out <= read_evcnt[35:18];
			default: reg_data_out <= 18'h0;
		endcase
	else
		if(selected_cnt >= COUNT_CYCLECOUNTERS)
			reg_data_out <= 18'h0;
		else case (selected_reg)
			7'h0: reg_data_out <= {1'h0,
			                       num_eventcnt,
			                       num_cyclecnt,
			                       ccnt_overflow[selected_cnt] | ccnt_counter[selected_cnt][WIDTH_CYCLECOUNTER],
			                       ccnt_conf[selected_cnt][7:0]};
			7'h1: reg_data_out <= read_ccnt[17:0];
			7'h2: reg_data_out <= read_ccnt[35:18];
			default: reg_data_out <= 18'h0;
		endcase
end

// Interrupt filter.
wire is_nointr; // 1 if the processor is NOT handling an interrupt.
intr_filter ifilter(clk, reset, intr_start, intr_end, is_nointr);

genvar j;
generate

for(j = 0; j < COUNT_CYCLECOUNTERS; j = j + 1) begin: ccnt

// Initial values.
initial begin
	ccnt_conf[j] = `PCNT_RESETVALUE;
	ccnt_counter[j] = 0;
	ccnt_overflow[j] = 0;
end

wire counter_enable = reg_idx[17 - (2 * j)];
wire counter_disable = reg_idx[16 - (2 * j)];

// Write performance counter configuration registers.
always @(posedge clk) begin
	if(reset)
		ccnt_conf[j] <= `PCNT_RESETVALUE;
	else if(reg_write && !sel_evcnt && selected_cnt == j && selected_reg == 0)
		ccnt_conf[j] <= reg_data_in;
	else if(counter_disable)
		ccnt_conf[j][0] <= 1'h0;
	else if(counter_enable)
		ccnt_conf[j][0] <= 1'h1;
	else begin
		if(ccnt_conf[j][1])
			ccnt_conf[j][1] <= 1'h0; // Reset reset bit.
		if(ccnt_conf[j][8])
			ccnt_conf[j][8] <= 1'h0; // Reset overflow clear bit.
	end
end

// Cycle counter overflow flags.
always @(posedge clk) begin
	if(ccnt_conf[j][1] | ccnt_conf[j][8]) // Reset overflow flag.
		ccnt_overflow[j] <= 1'h0;
	else if(ccnt_counter[j][WIDTH_CYCLECOUNTER]) // Set overflow flag.
		ccnt_overflow[j] <= 1'h1;
end

// Cycle counter prescaler.
wire prescaled_clock;
clk_prescaler p(clk, ccnt_conf[j][7:4], prescaled_clock);

// Interrupt filter.
wire intr_filter = ccnt_conf[j][2] ? is_nointr : 1'h1;

// Cycle counting.
always @(posedge clk) begin
	if(ccnt_conf[j][1])
		ccnt_counter[j] <= 0; // Reset counter.
	else if(ccnt_conf[j][0] & prescaled_clock & intr_filter)
		ccnt_counter[j] <= ccnt_counter[j] + 1'h1;
end

// Generate event for cycle counter overflow.
reg ovf_old = 0;
always @(posedge clk) begin
	ovf_old <= ccnt_counter[j][WIDTH_CYCLECOUNTER];
end
assign ccnt_ovf_event[j] = ovf_old ^ ccnt_counter[j][WIDTH_CYCLECOUNTER];

end //for

for(j = COUNT_CYCLECOUNTERS; j < 4; j = j + 1) begin: assign_ccnt_ovf_event

assign ccnt_ovf_event[j] = 0;

end


for(j = 0; j < COUNT_EVENTCOUNTERS; j = j + 1) begin: evcnt

// Initial values.
initial begin
	evcnt_conf[j] = `EVCNT_RESETVALUE;
	evcnt_counter[j] = 0;
	evcnt_overflow[j] = 0;
end

// Write event counter configuration registers.
always @(posedge clk) begin
	if(reset)
		evcnt_conf[j] <= `EVCNT_RESETVALUE;
	else if(reg_write && sel_evcnt && selected_cnt == j)
		evcnt_conf[j] <= reg_data_in;
	else begin
		if(evcnt_conf[j][11])
			evcnt_conf[j][11] <= 1'h0; // Reset overflow clear bit.
		if(evcnt_conf[j][8])
			evcnt_conf[j][8] <= 1'h0; // Reset reset bit.
	end
end

wire evcnt_reset = evcnt_conf[j][6] ? ccnt_conf[evcnt_conf[j][5:4]][1] : evcnt_conf[j][8];
wire evcnt_enabled = evcnt_conf[j][6] ? ccnt_conf[evcnt_conf[j][5:4]][0] : evcnt_conf[j][7];
wire evcnt_intrfilter_enabled = evcnt_conf[j][6] ? ccnt_conf[evcnt_conf[j][5:4]][2] : evcnt_conf[j][9];
wire evcnt_intrfilter = evcnt_intrfilter_enabled ? is_nointr : 1'h1;

// Event counter overflow flags.
always @(posedge clk) begin
	if(evcnt_reset | evcnt_conf[j][11]) // Reset overflow flag.
		evcnt_overflow[j] <= 1'h0;
	else if(evcnt_counter[j][WIDTH_EVENTCOUNTER]) // Set overflow flag.
		evcnt_overflow[j] <= 1'h1;
end

// Event counter edge detection.
core_edge_detector e(clk, (reset | evcnt_conf[j][8]), events[evcnt_conf[j][3:0]], evcnt_conf[j][10], events_edge_detected[j]);

// Event counter prescalers.
event_prescaler p(clk, events_edge_detected[j], evcnt_conf[j][14:11], events_prescaled[j]);

// Event counting.
always @(posedge clk)
	if(evcnt_reset)
		evcnt_counter[j] <= 0; // Reset counter.
	else if(evcnt_enabled & events_prescaled[j] & evcnt_intrfilter)
		evcnt_counter[j] <= evcnt_counter[j] + 1'h1;

end //for
endgenerate

endmodule
