/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/*! SpartanMC processor core
 * 
 */

/////////////////////////////////////////////////////////////////////
////                                                             ////
////  Toplevel Modul des SpartanMC                               ////
////  Hier werden die einzelen Teilmodule des Prozessor          ////
////  instanziert und mit einander verbunden. Weiterhin be-      ////
////  finden sich in diesem Modul die Pipelineregister.          ////
////                                                             ////
////  Author: Falk Hassler                                       ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`include    "core/src/spartanmc_def.vh"
`default_nettype none

//% @copydoc spartanmc.v
module spartanmc #(

    /*!
    * @name Parameters
    * @{
    */
    parameter   RESET_VECTOR                = 18'h00000,
    parameter   INTR_VECTOR                 = 18'h00020,
    parameter   SINGLE_SHIFT                = 1,
    parameter   HARDWARE_MUL                = 1,
    parameter   DISABLE_STALLS              = 0,
    parameter   SHOW_INSTR                  = 1,
    parameter   CLOCK_GATING    			= "NO",
	parameter   HARDWARE_DEBUGGING          = 0,
	parameter	COUNT_BREAKPOINTS			= 4,
	parameter	COUNT_WATCHPOINTS			= 4,
	parameter	PERFORMANCE_COUNTER			= 0,
	parameter	COUNT_CYCLECOUNTERS			= 1,
	parameter	COUNT_EVENTCOUNTERS			= 2,
	parameter	WIDTH_CYCLECOUNTER			= 36,
	parameter	WIDTH_EVENTCOUNTER			= 36,
	parameter	BRKPT_TRAP					= 1,
	parameter	WTCHPT_TRAP					= 2,
	parameter	SINGLE_STEP_TRAP			= 3,
    parameter   IO_BASE_ADR                 = 18'h1FC00,
	parameter	STOP_SIM_ON_X = 1,
	parameter	STOP_SIM_ON_PERI_S9 = 1
    //%@}

)(
    /*!
    * @name External inputs
    * External signals required for proper operation of the processor core
    * (e.g. clock and reset).
    * @{
    */
    input wire              locked,
    input wire              clk,
    //%@}



    /*!
    * @name Universal for all 4 buses
    * @{
    */

    /*!
    * When high, indicates that the processor is in reset state. Peripheral
    * units should implement an asynchronous reset using that signal.
    *
    * @brief Reset signal for peripheral units
    */
    output wire              bus_reset,

    /*!
    * This clock signal should be used to drive peripheral units. It runs
    * synchronously with the SpartanMC system clock.
    *
    * @brief Clock signal for peripheral units
    */
    output wire             bus_clk,

    /*!
    * All peripheral units can read on the bus at any time to receive data from
    * the processor. Usually, at one certain point in time, at most one peripheral
    * unit is requested to read data from the bus.
    * 
    * @brief Unidirectional data bus from the processor to all peripheral units.
    */
    output wire     [17:0]  bus_do,

    /*!
    * Processor data in. Should be all di_peri and di_ram_data combined
    * by a wide or gate.
    */
    input wire      [17:0] bus_di,


	/*!
    * @name Peripheral interface
    * Interface to connect peripheral units.
    * @{
    */
    output wire				peri_access,
    output wire				peri_wr,

    /*!
    * This signal provides the lower 10 bits of the SpartanMC address bus. These
    * bits are used to select a register within the IO space. The IO space
    * consists of all the registers implemented by the several peripheral units
    * in the system. At any given point in time at most one certain peripheral
    * register can be selected to avoid colissions on the peripheral bus. This
    * behaviour must be ensured by a proper chip select logic implemented by
    * each peripheral unit.
    * @see pselect
    *
    * @brief Address lines for peripheral register access 
    */
    output wire     [9:0]   bus_addr_block,

     
    /*!
    * @name DMA / Memory interface
    * Interface for DMA and Data Memory
    * @{
    */
    output wire             mem_access,
    output wire             bus_wr_low,
    output wire             bus_wr_high,
    output wire     [7:0]   bus_addr_high,
    //%@}

    /*!
    * @name Memory interface
    * Interface for Data Memory
    * @{
    */
    input wire				mem_busy,

    /*!
    * @name Code Memory interface
    * Interface for Code Memory
    * @{
    */
    output wire				code_access,
    output wire             code_jmp,
    input wire	[17:0]		code_di,
	output wire	[`ADDR_WIDTH-1:0]	code_addr,
	input wire				code_busy,


    /*!
    * @name Interrupt interface
    * Interface related to interrupt handling.
    * @{
    */
    input wire              intr,
    output wire             intr_enable,
    output wire             ir_return,
    //%@}

    input wire              step,
    input wire              step_en_on,
    input wire              step_en_off,
    input wire              alarm,

    input wire				evcnt_ext0,
    input wire				evcnt_ext1
);

/*!
* @name Internal wires and registers
* @{
*/
wire			clk_g;
reg				reset;
wire	[17:0]	pc_plusone;
wire    [17:0]  pc_intr;
wire    [9:0]   pc_trap;
wire            reg_cc;
wire            inc_window;
wire            dec_window;
wire            run_intr;
wire    [6:0]   curr_window;
wire    [6:0]   curr_window_p1;
wire    [6:0]   curr_window_ov_p1;
wire	[6:0]	curr_window_ov_p2;
wire    [9:0]   reg_no_a;
wire    [9:0]   reg_no_b;
wire    [3:0]   r_no_b;
wire    [17:0]  reg_a_p1;	// ist Ausgang 1 (A) des Registerfiles und Teil von pipe_1
wire    [17:0]  reg_b_p1;	// ist Ausgang 2 (B) des Registerfiles und Teil von pipe_1
wire			bypass1a;	// Register A wird gerade geschrieben
wire			bypass1b;	// Register B wird gerade geschrieben
wire    [17:0]  alu_imm;	// Konstanten fuer die ALU
wire    [3:0]   alu_func;
wire    [`ADDR_WIDTH-1:0] data_addr;
wire            mem_addr_0;
wire    [17:0]  write_data_data;
wire            write_data_low;
wire            write_data_high;
wire    [9:0]   write_reg_number;
wire    [17:0]  write_reg_data;
wire            write_reg;

wire            is_not;
wire            is_mem;
wire            is_store;
wire            is_imm;
wire			is_branch;
wire 			is_nop;
wire			branch_typ;
wire			branch_neg;
wire			is_jump;
wire			reg_to_pc;
wire            link_pc;
wire            short_trans;
wire			delay_bj;
wire			do_delay_bj;
wire            set_sfr;
wire            get_sfr;
wire            change_bit;
wire            set_bit;
wire            conditional_wb;
wire            do_wb_cc;
wire            do_ov;
wire            do_wb;
wire            multiply;
wire            is_signed;
wire    [17:0]  alu_result;
wire            interrupt;

// Pipeline Register
reg     [96:1]	pipe_1;     //Register zwischen ifid  und exmem + reg_a + reg_b
reg     [39:0]  pipe_2;     //Register zwischen exmem und writeback + lmd_p2 + di_peri +++

wire    [9:0]   reg_no_awb_p1;
wire    [3:0]   r_no_b_p1;
wire    [9:0]   reg_no_a_p1;
wire    [9:0]   reg_no_b_p1;
wire    		bypass1a_p1;
wire    		bypass1b_p1;
wire    [17:0]  alu_imm_p1;
wire	[17:0]	pc_delay;
wire    [17:0]  write_reg_data_p1;
wire    [3:0]   alu_func_p1;
wire            is_not_p1;
wire            is_mem_p1;
wire            is_store_p1;
wire            is_imm_p1;
wire			is_branch_p1;
wire 			is_nop_p1;
wire			branch_typ_p1;
wire			branch_neg_p1;
wire			is_jump_p1;
wire			reg_to_pc_p1;
wire            link_pc_p1;
wire            short_trans_p1;
wire			delay_bj_p1;
wire            set_sfr_p1;
wire            get_sfr_p1;
wire            change_bit_p1;
wire            set_bit_p1;
wire            conditional_wb_p1;
wire            do_wb_cc_p1;
wire            do_ov_p1;
wire            do_wb_p1;
wire            multiply_p1;
wire			is_signed_p1;

wire			do_delay_bj_p2;
wire			mem_addr_0_p2;
wire			short_trans_p2;
wire			is_mem_p2;
wire			is_store_p2;
wire    [9:0]   reg_no_awb_p2;
wire            do_wb_p2;
wire    [17:0]  alu_result_p2;


reg             por = 1'b1;
wire            res_start;
reg     [4:0]   res_cnt;

// synopsys translate_off
reg     [`INSTRUCTION_LEN*8:0]                      protocol_pipe_1;
wire    [(`INSTRUCTION_LEN+`OPERATION_LEN)*8:0]     protocol_pipe_2;
wire    [`INSTRUCTION_LEN*8:0]                      instruction;
wire    [`OPERATION_LEN*8:0]                        operation;
// synopsys translate_on

//%@}

//Breakpoint Support
wire			exc_trap;
wire	[7:0]	exc_trap_no;
wire	[`ADDR_WIDTH-1:0]	dbg_pc_next;


assign bus_reset = reset;
wire no_stall;

generate
if (CLOCK_GATING == "YES")
begin

	reg	clk_step;
	reg	clk_stepv;
	reg	step_en;
	reg	clk_e;

	always @ (posedge clk)
	if(reset)
		begin
			step_en		<= 1'b0;
		end
	else if (step_en_off)
		begin
			step_en		<= 1'b0;
		end
	else if (step_en_on)
		begin
			step_en		<= 1'b1;
		end

	always @ (posedge clk)
	if(reset)
		begin
			clk_step	<= 1'b0;
			clk_stepv	<= 1'b0;
		end
	else
		begin
			clk_step	<= step;
			clk_stepv	<= clk_step;
		end

	// Enable fuer einen BUFGCE aus clk1 bilden
	always @ (negedge clk)
	if(reset)
		begin
			clk_e	<= 1'b1;
		end
	else
		if((clk_step & (~clk_stepv))|(~step_en))
			begin
				clk_e	<= 1'b1;
			end
		else
			begin
				clk_e	<= 1'b0;
			end
	// CLK Treiber fuer den SpartanMC
	bufgce_prim BUFGCE_inst (
		.O(	clk_g),
		.CE(	clk_e),
		.I(	clk)
	);

end else begin

	assign	clk_g	= clk;

end
endgenerate

pselect        #(
			.ADDR_WIDTH	(8		),
			.BASE_WIDTH	(8		),
			.BASE_ADDR	(IO_BASE_ADR >> 10	)
		)
	peri_sel	(
			.addr		(bus_addr_high[7:0]	),
			.activ_peri	(mem_access),
			.select		(peri_access	)
		);

assign code_access =  ~mem_busy;

assign res_start            = !locked | por | alarm;
assign interrupt            = intr;
assign bus_clk				= clk_g;

assign peri_wr				= bus_wr_low | bus_wr_high;

assign mem_access           = is_mem_p1 && ~code_busy && ~mem_busy;

assign bus_wr_low			= write_data_low;
assign bus_wr_high			= write_data_high;
assign bus_addr_high		= data_addr[`ADDR_WIDTH-1:10];
assign bus_addr_block		= data_addr[9:0];
assign bus_do				= write_data_data;


assign is_nop_p1			= pipe_1[96];
assign reg_to_pc_p1         = pipe_1[95];
assign is_branch_p1         = pipe_1[94];
assign branch_typ_p1        = pipe_1[93];
assign branch_neg_p1        = pipe_1[92];
assign is_jump_p1           = pipe_1[91];
assign curr_window_p1       = pipe_1[90:84];
assign delay_bj_p1          = pipe_1[83];
assign link_pc_p1           = pipe_1[82];
assign r_no_b_p1            = pipe_1[81:78];
assign reg_no_a_p1          = pipe_1[77:68];
assign reg_no_b_p1          = pipe_1[67:58];
assign bypass1a_p1          = pipe_1[57];
assign bypass1b_p1          = pipe_1[56];
assign alu_imm_p1           = pipe_1[55:38];
assign write_reg_data_p1   = pipe_1[37:20];
assign alu_func_p1          = pipe_1[19:16];
assign is_not_p1            = pipe_1[15];
assign is_mem_p1            = pipe_1[14];
assign is_store_p1          = pipe_1[13];
assign is_imm_p1            = pipe_1[12];
assign short_trans_p1       = pipe_1[11];
assign set_sfr_p1           = pipe_1[10];
assign get_sfr_p1           = pipe_1[9];
assign change_bit_p1        = pipe_1[8];
assign set_bit_p1           = pipe_1[7];
assign conditional_wb_p1    = pipe_1[6];
assign do_wb_cc_p1          = pipe_1[5];
assign do_ov_p1             = pipe_1[4];
assign do_wb_p1             = pipe_1[3];
assign multiply_p1          = pipe_1[2];
assign is_signed_p1         = pipe_1[1];

assign mem_addr_0_p2        = pipe_2[39];
assign short_trans_p2       = pipe_2[38];
assign is_mem_p2            = pipe_2[37];
assign curr_window_ov_p2	= pipe_2[36:30];
assign do_delay_bj_p2       = pipe_2[29];
assign reg_no_awb_p2        = pipe_2[28:19];
assign do_wb_p2             = pipe_2[18];
assign alu_result_p2        = pipe_2[17:0];

assign curr_window_ov_p1 = delay_bj_p1? curr_window_p1 : curr_window; // Trap and Intr change Window immediately!

// synopsys translate_off
assign protocol_pipe_2      = {protocol_pipe_1, operation};
// synopsys translate_on

reg was_reset;
assign no_stall = ~mem_busy && ~code_busy && ~was_reset;

always @ (posedge clk)
    por <= 1'b0;

always  @ (posedge clk)
    if(res_start)
    begin
        reset   <= 1'b1;
        res_cnt <= 5'b11111;
    end
    else
        if(res_cnt != 5'b00000)
            res_cnt <= res_cnt - 1'b1;
        else
            reset   <= 1'b0;

always @(posedge clk_g)
	was_reset <= reset;

always @(posedge clk_g)
    if (reset)
    begin
        pipe_1 <= 95'b0;
        pipe_2 <= 41'b0;
        // synopsys translate_off
        protocol_pipe_1 <= { 10*8 {1'b0} };
        // synopsys translate_on
    end
    else if(no_stall)
    begin
        pipe_1 <=   {is_nop, reg_to_pc, is_branch, branch_typ, branch_neg, is_jump, curr_window, delay_bj, link_pc,
                 r_no_b, reg_no_a, reg_no_b, bypass1a, bypass1b, alu_imm, write_reg_data, alu_func, is_not,
                 is_mem, is_store, is_imm, short_trans, set_sfr, get_sfr, change_bit, set_bit, conditional_wb,
                 do_wb_cc, do_ov, do_wb, multiply, is_signed};
        pipe_2 <=   { mem_addr_0, short_trans_p1, is_mem_p1, curr_window_ov_p1,
		         delay_bj_p1, reg_no_awb_p1, do_wb_p1, alu_result};

        // synopsys translate_off
        protocol_pipe_1 <= instruction;
        // synopsys translate_on
    end


ifetch_idecode      #(
            .RESET_VECTOR(RESET_VECTOR),
            .INTR_VECTOR(INTR_VECTOR),
            .SINGLE_SHIFT(SINGLE_SHIFT),
            .HARDWARE_MUL(HARDWARE_MUL),
            .HARDWARE_DEBUGGING(HARDWARE_DEBUGGING),
	    .STOP_SIM_ON_X(STOP_SIM_ON_X))

ifid            (
            .clk1(              clk_g           ),
            .locked(            locked           ),
            .clk_en(            no_stall         ),
            .reset(             reset            ),
            .was_reset(         was_reset        ),
	    .code_busy(			code_busy		 ),
            .interrupt(         interrupt        ),
            .ir_return(         ir_return        ),
            .pc_plusone(        pc_plusone       ),
            .pc_delay(          pc_delay         ),
            .pc_intr(           pc_intr          ),
            .pc_trap(           pc_trap          ),
            .reg_cc(            reg_cc           ),
            .intr_enable(       intr_enable      ),
            .run_intr(          run_intr         ),
            .inc_window(        inc_window       ),
            .dec_window(        dec_window       ),
            .reg_window(        curr_window_ov_p1),
            .reg_no_a(          reg_no_a         ),
            .reg_no_b(          reg_no_b         ),
            .r_no_b(            r_no_b           ),
            .reg_a_p1(          reg_a_p1         ),
            .reg_b_p1(          reg_b_p1         ),
            .bypass1a(          bypass1a         ),
            .bypass1b(          bypass1b         ),
            .alu_imm(           alu_imm          ),
            .alu_func(          alu_func         ),
            .write_reg_number(  write_reg_number ),
            .write_reg_data(    write_reg_data   ),
            .write_reg(         write_reg        ),
            .code_jmp(          code_jmp         ),
            .code_addr(         code_addr        ),
            .ir(                code_di          ),
            .is_not(            is_not           ),
            .is_mem(            is_mem           ),
            .is_store(          is_store         ),
            .is_imm(            is_imm           ),
            .is_branch(         is_branch        ),
            .is_nop(            is_nop           ),
            .branch_typ(        branch_typ       ),
            .branch_neg(        branch_neg       ),
            .is_jump(           is_jump          ),
            .reg_to_pc(         reg_to_pc        ),
            .link_pc(           link_pc          ),
            .short_trans(       short_trans      ),
            .is_delay(          delay_bj         ),
            .delay_bj_p1(       delay_bj_p1      ),
            .do_delay_bj(       do_delay_bj      ),
            .set_sfr(           set_sfr          ),
            .get_sfr(           get_sfr          ),
            .change_bit(        change_bit       ),
            .set_bit(           set_bit          ),
            .conditional_wb(    conditional_wb   ),
            .do_wb_cc(          do_wb_cc         ),
            .do_ov(             do_ov            ),
            .do_wb(             do_wb            ),
            .multiply(          multiply         ),
            .is_signed(         is_signed        ),
            .mem_clk(           bus_clk          ),
            .dbg_pc_next(		dbg_pc_next		),//Breakpoint support
			.exc_trap(			exc_trap		),
			.exc_trap_no(		exc_trap_no	)
            // synopsys translate_off
            ,
            .instruction(       instruction      )
            // synopsys translate_on

            );

ex_mem #        (
            .RESET_VECTOR(RESET_VECTOR),
            .INTR_VECTOR(INTR_VECTOR),
            .SINGLE_SHIFT(SINGLE_SHIFT),
            .HARDWARE_MUL(HARDWARE_MUL),
            .HARDWARE_DEBUGGING(HARDWARE_DEBUGGING),
			.COUNT_BREAKPOINTS(COUNT_BREAKPOINTS),
			.COUNT_WATCHPOINTS(COUNT_WATCHPOINTS),
			.PERFORMANCE_COUNTER(PERFORMANCE_COUNTER),
			.COUNT_EVENTCOUNTERS(COUNT_EVENTCOUNTERS),
			.WIDTH_CYCLECOUNTER(WIDTH_CYCLECOUNTER),
			.WIDTH_EVENTCOUNTER(WIDTH_EVENTCOUNTER),
			.BRKPT_TRAP(BRKPT_TRAP),
			.WTCHPT_TRAP(WTCHPT_TRAP),
			.SINGLE_STEP_TRAP(SINGLE_STEP_TRAP))
exmem           (
            .clk(              clk_g           ),
            .clk_en(            no_stall         ),
            .reset(             reset            ),
            .inc_window(        inc_window       ),
            .dec_window(        dec_window       ),
            .window(            curr_window      ),
            .cc(                reg_cc           ),
            .intr_enable(       intr_enable      ),
            .run_intr(          run_intr         ),
            .pc_plusone(        pc_plusone       ),
            .pc_intr(           pc_intr          ),
            .pc_trap(           pc_trap          ),
            .reg_no_a(          reg_no_a_p1      ),
            .reg_no_b(          reg_no_b_p1      ),
            .reg_no_awb(		reg_no_awb_p1    ),
            .r_no_b(            r_no_b           ),
            .r_no_b_p1(         r_no_b_p1        ),
            .reg_a(             reg_a_p1         ),
            .reg_b(             reg_b_p1         ),
            .bypass1a(          bypass1a_p1      ),
            .bypass1b(          bypass1b_p1      ),
            .alu_imm(           alu_imm_p1       ),
            .pc_delay(			pc_delay         ),
            .alu_func(          alu_func_p1      ),
            .alu_result(        alu_result       ),
            .data_addr(         data_addr        ),
            .mem_addr_0(        mem_addr_0       ),
            .write_data_data(   write_data_data  ),
            .write_data_low(    write_data_low   ),
            .write_data_high(   write_data_high  ),
            .ir_return(         ir_return        ),
            .is_not(            is_not_p1        ),
            .is_mem(            is_mem_p1        ),
            .is_mem_p1(			is_mem_p2		 ),
            .is_store(          is_store_p1      ),
            .is_imm(            is_imm_p1        ),
            .is_branch(         is_branch_p1     ),
            .is_nop(            is_nop_p1        ),
            .branch_typ(        branch_typ_p1    ),
            .branch_neg(        branch_neg_p1    ),
            .do_delay_bj(       do_delay_bj      ),
            .do_delay_bj_p2(    do_delay_bj_p2   ),
            .is_jump(           is_jump_p1       ),
            .reg_to_pc(         reg_to_pc_p1     ),
            .link_pc(           link_pc_p1       ),
            .short_trans(		short_trans_p1	),
            .delay_bj_p1(		delay_bj_p1		),
            .delay_bj(			delay_bj		),
            .set_sfr(           set_sfr_p1       ),
            .get_sfr(           get_sfr_p1       ),
            .change_bit(        change_bit       ),
            .change_bit_p1(     change_bit_p1    ),
            .set_bit(           set_bit          ),
            .set_bit_p1(        set_bit_p1       ),
            .conditional_wb(    conditional_wb_p1),
            .do_wb_cc(          do_wb_cc_p1      ),
            .do_ov(             do_ov_p1         ),
            .multiply(          multiply_p1      ),
            .is_signed(         is_signed_p1     ),
            .last_do_wb(        do_wb_p2         ),
            .write_reg_no(     write_reg_number ),
            .write_reg_data(   write_reg_data   ),
            .last_write_reg_data(write_reg_data_p1),
            .evcnt_ext0(		evcnt_ext0), // Performance counter event lines
			.evcnt_ext1(		evcnt_ext1),
            .dbg_pc_next(		dbg_pc_next		),//Breakpoint support
			.dbg_trap(			exc_trap		),
			.dbg_trap_no(		exc_trap_no		)
            // synopsys translate_off
            ,
            .operation(		operation	 )
            // synopsys translate_on
            );

writeback #     (
            .RESET_VECTOR(RESET_VECTOR),
            .INTR_VECTOR(INTR_VECTOR),
            .SINGLE_SHIFT(SINGLE_SHIFT),
            .HARDWARE_MUL(HARDWARE_MUL),
            .SHOW_INSTR(SHOW_INSTR),
	    .DISABLE_STALLS(DISABLE_STALLS))
wb          (
            .clk1(              clk_g           ),
            .reset(             reset           ),
	    .mem_busy(		mem_busy),
	    .code_busy(		code_busy),
            .reg_no_a(          reg_no_awb_p2     ),
            .write_reg_number(  write_reg_number),
            .write_reg_data(    write_reg_data  ),
            .write_reg(         write_reg       ),
            .alu_result(        alu_result_p2   ),
            .lmd(               bus_di          ),
            .short_trans(       short_trans_p2  ),
            .mem_addr_0(        mem_addr_0_p2   ),
            .is_mem(            is_mem_p2       ),
            .do_wb(             do_wb_p2        )
            // synopsys translate_off
            ,
            .protocol(          protocol_pipe_2 ),
            .reg_window_read(   curr_window_ov_p1),
            .reg_window_write(  curr_window_ov_p2)
            // synopsys translate_on
            );


reg peri_s9_fail = 0;
reg[17:0] peri_s9_failaddr = 0;
always @(posedge clk_g) begin
	if (STOP_SIM_ON_PERI_S9 && ~reset && peri_access && (bus_wr_low^bus_wr_high)) begin
		peri_s9_fail <= 1;
		peri_s9_failaddr <= data_addr;
	end
	if (peri_s9_fail) begin
		$display("[ %9d ] Error: Tried to access peripheral at address 0x%x with a S9 instruction", $time, peri_s9_failaddr);
		$stop();
	end
end


reg mem_access_align_fail = 0;
reg[17:0] mem_access_align_failaddr = 0;
always @(posedge clk_g) begin
	if (~reset && mem_access && ~short_trans_p1 && mem_addr_0) begin
		mem_access_align_fail <= 1;
		mem_access_align_failaddr <= data_addr;
	end
	if (mem_access_align_fail) begin
		$display("[ %9d ] Error: Tried to access memory at unaligned address 0x%x with a L18/S18 instruction", $time, mem_access_align_failaddr);
		$stop();
	end
end


endmodule
