/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  Writeback                                                  ////
////  Hier wird das Ergebnis der ExMem Stufe in das entspr-      ////
////  echende Register zurueck geschrieben, wenn es im aktuellen ////
////  Befehl noegtig ist (was durch do_wb angezeigt wird).       ////
////                                                             ////
////  Author: Falk Hassler                                       ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`include	"core/src/spartanmc_def.vh"
`default_nettype none

module writeback(
		clk1,
		reset,
		mem_busy,
		code_busy,
		reg_no_a,
		write_reg_number,
		write_reg_data,
		write_reg,
		alu_result,
		lmd,
		short_trans,
		mem_addr_0,
		is_mem,
		do_wb
		// synopsys translate_off
		,
		protocol,
		reg_window_read,
		reg_window_write
		// synopsys translate_on
		);

// Die Folgenden Parameter werden durch die system.v bei der Erzeugung des Modules ueberschrieben.
parameter	RESET_VECTOR	=	0;
parameter	INTR_VECTOR	=	32;
parameter	SINGLE_SHIFT	=	1;
parameter	HARDWARE_MUL	=	1;
parameter	SHOW_INSTR	=	1;
parameter	DISABLE_STALLS  = 	0;

input	wire	[9:0]			reg_no_a;
output	wire	[9:0]			write_reg_number;
output	reg		[17:0]			write_reg_data;
output	wire					write_reg;
input	wire	[17:0]			alu_result;
input	wire	[17:0]			lmd;		// pipe_2 Datenspeicher
input	wire					short_trans;
input	wire					mem_addr_0;
input	wire					is_mem;
input	wire					do_wb;

input	wire	clk1;
input	wire	mem_busy, code_busy;
input	wire	reset;

reg		[17:0]					mem_buffer;

wire	[17:0]					mem_in;
assign mem_in = short_trans ? { 9'h000, mem_addr_0 ? lmd[8:0] : lmd[17:9] } : lmd;


reg stalling, mem_busy_delayed, code_busy_delayed;

//Memory result needs to be buffered when:
// a) mem stops being busy while code is still busy
// b) mem was never busy at all while code is busy
always @(posedge clk1) begin
	if (reset)
		stalling <= 0;
	stalling <= mem_busy || code_busy;
	mem_busy_delayed <= mem_busy;
	code_busy_delayed <= code_busy;

	if (stalling) begin
		//case a from above, detect falling edge on mem_busy
		if (mem_busy_delayed && !mem_busy)
			mem_buffer <= mem_in;
	end else begin
		//case b from above, just always save
		mem_buffer <= mem_in;
	end
end


// MUX fuer LMD und alu_result
generate
always @(*)
	if (is_mem & do_wb) begin
		//If code was busy longer than data
		if (code_busy_delayed && !mem_busy_delayed && !DISABLE_STALLS) begin
			write_reg_data = mem_buffer;
		end else begin
			write_reg_data = mem_in;
		end
	end else begin
		write_reg_data = alu_result;
	end
endgenerate


// die 3 Inputs mit den 3 Outputs verbinden.
assign	write_reg_number	= reg_no_a;
assign	write_reg		= do_wb;



// Debugausgaben fuer den Simulator:
// * gibt den abgearbeiteten Befehl aus
// * gibt die von der ALU ausgefuehrte Operation aus
// * gibt die Daten aus die wieder in ein Register
//   geschrieben werden
// synopsys translate_off
input wire	[6:0]	reg_window_read;
input wire	[6:0]	reg_window_write;
input wire	[(`INSTRUCTION_LEN+`OPERATION_LEN)*8:0]	protocol;

wire	[`INSTRUCTION_LEN*8:0]	instruction	= 
	protocol[(`INSTRUCTION_LEN+`OPERATION_LEN)*8:`OPERATION_LEN*8+1];
wire	[`OPERATION_LEN*8:0]	operation	= 
	protocol[`OPERATION_LEN*8:0];

reg	[6:0]	delay_window_r;
reg	[6:0]	delay_window_w;

generate
	if	(SHOW_INSTR	== 1)
	always @(posedge clk1) begin
		if (reset) begin
			delay_window_r	<= 0;
			delay_window_w	<= 0;
		end	else if (!code_busy && !mem_busy) begin
			delay_window_r	<= reg_window_read;
			delay_window_w	<= reg_window_write;
			if ((instruction[7:0]!=8'b00000000) | (operation[7:0]!=8'b00000000))
			begin
				$write("[ %9d ]     %s %s", $time, instruction, operation);
			end

			if (do_wb)
			begin
				$write("\t\t r%2d(%3d / %3d)=REGS[%4d] <= 0x%05H",
						 (write_reg_number < 6'h04) ?	write_reg_number :
							write_reg_number - (delay_window_w*8),
						delay_window_r,
						delay_window_w,
						write_reg_number,
						write_reg_data);
			end
			else	// REG_BASE immer anzeigen.
				$write("\t\t    (%3d / %3d)", delay_window_r,delay_window_w);

		$write("\n");
		end else begin
			if (code_busy && mem_busy)
				$write("[ %9d ]     STALL on CODE + MEM\n", $time);
			else if (code_busy)
				$write("[ %9d ]     STALL on CODE\n", $time);
			else
				$write("[ %9d ]     STALL on MEM\n", $time);
		end
	end
endgenerate
// synopsys translate_on

endmodule
