/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// TU-Dresden Chair for Embedded Systems
//////////////////////////////////////////////////////////////////////////////////
module core_connector_master
#(
        // SPMC Interface
        parameter BASE_ADR               = 10'h0,
        parameter FIFO_DEPTH             = 'd128,
        parameter FIFO_WIDTH             = 'd18
)
(
        //*** Connections with SpartanMC master ***
        input   wire            clk_peri,	//Systemtakt
        input   wire [17:0]     do_peri,	//Datenbus von MC
        output  wire [17:0]     di_peri,	//Datenbus zu MC
        input   wire [9:0]      addr_peri,	//Adressbus von MC
        input   wire            access_peri,//Signale fuer Peripheriezugriffe
        input   wire            wr_peri,	//Schreibsignal
        input   wire            reset,		//Reset-Signal
	
        // Connections with slave-connector
        input wire 			        	read_enable,
        output wire [FIFO_WIDTH-1:0]	data_to_slave,
        output wire [17:0]				free_entries,
        output wire [17:0]				used_entries
);
/////////////////////////////////////////////////////////////////////////////////////
//
// Local Functions and Parameters
//
function integer log2;
input integer value;
	for (log2=0; value>0; log2=log2+1)
	value = value>>1;
endfunction

parameter
    DATA_REG_ADDR           = 2,
    MSG_SIZE_REG_ADDR       = 1,
    STATUS_REG_ADDR         = 0,
	
	ALL_PORTS				= 3,
	DECODSIM_DEB			= "NO",
    FIFO_ADDR_BITS          = log2(FIFO_DEPTH - 1);


// Dekodierung der Adressen ********************** START *************************************
// chipselect
wire		status_read_sel;	// CPU liest von peri
wire		msg_size_read_sel;	// CPU liest von peri
wire		data_read_sel;		// CPU liest von peri
wire		status_write_sel;	// CPU schreibt auf peri
wire		msg_size_write_sel;	// CPU schreibt auf peri
wire		data_write_sel;		// CPU schreibt auf peri

wire							select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS		),
		.BASE_ADR(	BASE_ADR		),
		.SIM_DEBUG(	DECODSIM_DEB	)
) decod1 (
		.reset(			reset		),
		.clk_peri(		clk_peri	),
		.addr_peri(		addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(		wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(		reg_read	),
		.reg_write(		reg_write	)
);

assign		status_read_sel 	= reg_read[STATUS_REG_ADDR];
assign		msg_size_read_sel 	= reg_read[MSG_SIZE_REG_ADDR];
assign		data_read_sel 		= reg_read[DATA_REG_ADDR];

assign		status_write_sel 	= reg_write[STATUS_REG_ADDR];
assign		msg_size_write_sel 	= reg_write[MSG_SIZE_REG_ADDR];
assign		data_write_sel 		= reg_write[DATA_REG_ADDR];

reg [17:0] message_size;

wire fifo_write_enable = data_write_sel;
wire [FIFO_ADDR_BITS:0] fifo_fill;

assign free_entries = 18'h00000 | (FIFO_DEPTH - fifo_fill);
assign used_entries = 18'h00000 | fifo_fill;

wire fifo_full = free_entries < message_size;
//wire fifo_empty = used_entries < message_size;

assign	di_peri [17:0]	= 	(status_read_sel) ? {17'h0000, ~fifo_full} :
							(msg_size_read_sel) ? 18'h00000 | message_size :
							18'b0;

always @(posedge clk_peri)
begin
    if (reset)
    begin
        message_size <= 18'h1;
    end else if (msg_size_write_sel)
    begin
        message_size <= do_peri;
    end
end


///////////////////////////////////////////////////////////////////
// Fifo logic module
// Using async fifo module with the same clock on both ports
xilinx_fifo_with_fill_counter #
(
 .WIDTH(FIFO_WIDTH),
 .DEPTH(FIFO_DEPTH),
 .ADDR_BITS(FIFO_ADDR_BITS)
) 
	fifo0(
.clk(clk_peri),
.fill_count(fifo_fill),
.reset(reset),
.write_enable(fifo_write_enable),
.write_data(do_peri),
.read_enable(read_enable), 
.read_data(data_to_slave)
);

endmodule
