/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// TU-Dresden Chair for Embedded Systems
//////////////////////////////////////////////////////////////////////////////////
module core_connector_slave_dma
#(
        // SPMC Interface
        parameter DMA_BASE_ADR			= 18'h0,
        parameter DMA_SIZE				= 'd64,
        parameter DMA_WIDTH				= 'd18
)
(
        //*** Connections with SpartanMC master ***
        input   wire            clk_peri,	//Systemtakt
        input   wire            clk_dma,	//DMAtakt
        input   wire            reset,		//Reset-Signal
        
        //DMA-Signals
		input wire						mem_access,
		input wire						store_access_low,			//TODO
		input wire						store_access_high,			//TODO
		input wire	[9:0]				addr_peri,
		input wire	[7:0]				addr_high,
		input wire	[17:0]				do_peri,
		output wire	[17:0]				di_peri,  
	
        // Connections with slave-connector
		output wire                     read_enable,
		input wire [DMA_WIDTH-1:0]     data_from_master,
		input wire [17:0]				free_entries,
		input wire [17:0]				used_entries
);
/////////////////////////////////////////////////////////////////////////////////////
//
// Local Functions and Parameters
//
function integer log2;
input integer value;
	for (log2=0; value>0; log2=log2+1)
	value = value>>1;
endfunction

parameter
	DECODSIM_DEB			= "NO",
	DMA_RAMBLOCKS          = DMA_SIZE / 2048 + (DMA_SIZE % 2048 > 0 ? 1 : 0),
    DMA_ADDRWIDTH          = log2(DMA_SIZE - 1);

//inter wires & regs
wire [17:0] data_out_a;
wire [DMA_RAMBLOCKS-1:0] select;					//pselect zu Bram
wire [DMA_RAMBLOCKS-1:0] select_dma;
wire [17:0] full_addr;
assign full_addr = {addr_high,addr_peri};
//Bram Wires
wire write;
wire [17:0] doutb;
wire [17:0] dinb;
reg [DMA_ADDRWIDTH-1:0] addrb_reg;
wire [DMA_ADDRWIDTH-1:0]addrb;
wire store_access;

wire	[18*DMA_RAMBLOCKS-1:0]	mem_do_peri_all;
wire	[18*DMA_RAMBLOCKS-1:0]	mem_do_dma_all;

//FSM
reg [3:0] state;
reg [DMA_ADDRWIDTH-1:0] counter;//oder 17:0?!
reg fin_counter;
localparam [3:0]
	IDLE				= 0,
	GET_COUNT			= 1,
	GET_ADDR		 	= 2,
	REC					= 3,
	FIN					= 4;
generate
	if	(DMA_RAMBLOCKS	== 1) begin
		assign	di_peri	=	mem_do_peri_all[17:0];
		assign	doutb =	mem_do_dma_all[17:0];
	end else begin
// ****************************************************************
// OR Speicherports fuer Befehle bei mehr als einem Speicherblock
		orgate	#(
			.OR_WIDTH	(DMA_RAMBLOCKS	),
			.BUS_WIDTH	(18		)
			)
		perimem_or (
			.a		({mem_do_peri_all}	),
			.y		(di_peri		)
			);

		orgate	#(
			.OR_WIDTH	(DMA_RAMBLOCKS	),
			.BUS_WIDTH	(18		)
			)
		dmamem_or (
			.a		({mem_do_dma_all}	),
			.y		(doutb		)
			);
	end
endgenerate
	
genvar		i;
generate
for	(i=0; i<DMA_RAMBLOCKS; i=i+1)
begin:	MEM_BL
// Dekodierung der Adressen ********************** START *************************************
//pselect module to generate select signal for DMA memory
pselect #(
	.ADDR_WIDTH(7),
	.BASE_WIDTH(7),
	.BASE_ADDR(DMA_BASE_ADR[17:11]+i)
	) memselect_peri (
	.addr(full_addr[17:11]),
	.activ_peri(mem_access),
	.select(select[i])
);

if (DMA_RAMBLOCKS == 1)
	assign select_dma[0] = 1'b1;
else
	pselect #(
		.ADDR_WIDTH(DMA_ADDRWIDTH-11),
		.BASE_WIDTH(DMA_ADDRWIDTH-11),
		.BASE_ADDR(i)
		) memselect_dma (
		.addr(addrb[DMA_ADDRWIDTH-1:11]),
		.activ_peri(1'b1),
		.select(select_dma[i])
	);
	
ram_9_prim  #(
		.WRITE_MODE_A("READ_FIRST"),
		.WRITE_MODE_B("READ_FIRST"),
        .INIT_A(9'h000),
		.SRVAL_A(9'h000),
		.INIT_B(9'h000),
		.SRVAL_B(9'h000)
		)
DPRAMA		(
		.DIA(		do_peri[8:0]					),
		.ADDRA(		full_addr[10:0]			),
		.ENA(		1'b1				),
		.WEA(		store_access_low && select[i]				),
		.SSRA(		!select[i]			),
		.CLKA(		clk_peri				),
		.DOA(		mem_do_peri_all[18*i+8:18*i+0]		),
		.DIB(		dinb[8:0]			),
		.ADDRB(		addrb[10:0]			),
		.ENB(		1'b1				),
		.WEB(		write && select_dma[i]	),
		.SSRB(		!select_dma[i]			),
		.CLKB(		clk_dma			),
		.DOB(		mem_do_dma_all[18*i+8:18*i+0]		)
		);

ram_9_prim  #(
		.WRITE_MODE_A("READ_FIRST"),
		.WRITE_MODE_B("READ_FIRST"),
        .INIT_A(9'h000),
		.SRVAL_A(9'h000),
		.INIT_B(9'h000),
		.SRVAL_B(9'h000)
		)
DPRAMB		(
		.DIA(		do_peri[17:9]			),
		.ADDRA(		full_addr[10:0]			),
		.ENA(		1'b1				),
		.WEA(		store_access_high && select[i]			),
		.SSRA(		!select[i]			),
		.CLKA(		clk_peri				),
		.DOA(		mem_do_peri_all[18*i+17:18*i+9]		),
		.DIB(		dinb[17:9]			),
		.ADDRB(		addrb[10:0]			),
		.ENB(		1'b1				),
		.WEB(		write && select_dma[i]	),
		.SSRB(		!select_dma[i]			),
		.CLKB(		clk_dma			),
		.DOB(		mem_do_dma_all[18*i+17:18*i+9]		)
		);

/*
//Bram--evtl. nochmal in axi nach der 2ten Methode schauen
assign store_access = store_access_low || store_access_high;
ram_18_prim #(
	.WRITE_MODE_A(	"READ_FIRST"			),
	.WRITE_MODE_B(	"READ_FIRST"			)
	) DMARAM (
	.CLKA(clk_peri), 
	.CLKB(clk_dma), 
	.DOA(mem_do_peri_all[18*i+17:18*i]), 
	.DOB(mem_do_dma_all[18*i+17:18*i]),
	.DIA(do_peri), 
	.DIB(dinb),
	.ADDRA(full_addr[9:0]), 
	.ADDRB(addrb[9:0]),
	.ENA(1'b1),
	.ENB(1'b1),
	.WEA((store_access&&select[i])), 
	.WEB(write&&select_dma[i]),
	.SSRA(!select[i]),
	.SSRB(!select_dma[i])
);
*/
end
endgenerate
//FSM
always @(posedge clk_dma)begin
	if(reset)begin
		counter <= 0;
		addrb_reg <= 10'd0;
		state<=FIN;
		fin_counter <= 0;
	end else begin
		case(state)
			IDLE:begin
				fin_counter <= 0;
				state<=(doutb[1]==1'b1) ? GET_COUNT : IDLE;
			end
			GET_COUNT:begin
				counter<=doutb;
				state<=GET_ADDR;
			end
			GET_ADDR:begin
				addrb_reg<=doutb[DMA_ADDRWIDTH:1];
				state<=REC;
			end
			REC:begin
				counter<=(used_entries!=0) ? counter-1 : counter;
				state<=(used_entries!=0)&&(counter==1) ? FIN : REC;
				addrb_reg<=(used_entries!=0) ? 
							(counter==1) ? 0 : addrb_reg+1 : addrb_reg;
			end
			FIN:begin
				fin_counter <= 1;
				state<=(fin_counter == 1) ? IDLE : FIN;
			end
			default:begin
				counter <= 0;
				addrb_reg <= 10'd0;
				state<=IDLE;
				fin_counter <= 0;
			end
		endcase
	end
end
assign addrb = (state == IDLE && doutb[1]==1'b1) ? 1 :
				(state == GET_COUNT) ? 2 :
				(state == GET_ADDR) ? doutb[DMA_ADDRWIDTH:1] : 
				(state == FIN) ? 0 : addrb_reg;

assign read_enable = (state == REC)&&(used_entries!=0) ? 1 : 0;
assign dinb = (reset || state==FIN) ? 1 : data_from_master;
//Wann schreibe ich im BRAM?
assign write = ((state == REC)&&(used_entries!=0) || state == FIN || reset) ? 1 : 0;

endmodule






