/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
 
///////////////////////////////////////////////////////////
//
// Project  : AMIDAR
// Function : Pseudo LRU 4-way-set 
// File     : PLRU4.v
// Created  : 26.05.2015
//
// Author   : Sven Himer (sven.himer@gmail.com)
// Notes    : 4 Spaces = 1 Tab
//
//  This Module implements the pseudo-least-recently-used
// 	replacement strategy for 4-way-set assoziative caches.
//
//   two-way set associative - one bit
//
//   indicates which line of the two has been reference more recently
//
//   four-way set associative - three bits
//
//   each bit represents one branch point in a binary decision tree; let 1
//   represent that the left side has been referenced more recently than the
//   right side, and 0 vice-versa
//
//              are all 4 lines valid?
//                   /       \
//                 yes        no, use an invalid line
//                  |
//                  |
//                  |
//             bit_0 == 0?            state | replace      ref to | next state
//              /       \             ------+--------      -------+-----------
//             y         n             00x  |  line_0      line_0 |    11_
//            /           \            01x  |  line_1      line_1 |    10_
//     bit_1 == 0?    bit_2 == 0?      1x0  |  line_2      line_2 |    0_1
//       /    \          /    \        1x1  |  line_3      line_3 |    0_0
//      y      n        y      n
//     /        \      /        \        ('x' means       ('_' means unchanged)
//   line_0  line_1  line_2  line_3      don't care)
//
//   (see Figure 3-7, p. 3-18, in Intel Embedded Pentium Processor Family Dev.
//    Manual, 1998, http://www.intel.com/design/intarch/manuals/273204.htm)
//
///////////////////////////////////////////////////////////
`timescale 1ps / 1ps
`default_nettype none
module CachePLRU4
(
	input wire [3:0] use_vect_i,
	output wire [3:0] arbit_o,
	// Desicion Flags In/Out
	input wire [2:0] plruf_i,
	output wire [2:0] plruf_o
);

	// Assign arbit
	assign arbit_o[0] = ((plruf_i[2] == 0) & (plruf_i[1] == 0));
	assign arbit_o[1] = ((plruf_i[2] == 0) & (plruf_i[1] == 1));
	assign arbit_o[2] = ((plruf_i[2] == 1) & (plruf_i[0] == 0));
	assign arbit_o[3] = ((plruf_i[2] == 1) & (plruf_i[0] == 1));

	// Assign decision flags
	assign plruf_o[2] = (use_vect_i[0] | use_vect_i[1]);
	assign plruf_o[1] = (use_vect_i[0] | use_vect_i[1]) ? use_vect_i[0] : plruf_i[1];
	assign plruf_o[0] = (use_vect_i[2] | use_vect_i[3]) ? use_vect_i[2] : plruf_i[0];

endmodule
