/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
 
/////////////////////////////////////////////////////////////////////
////                                                             ////
////  Instruction Cache Controller                               ////
////  des SpartanMC Instruction Cache                            ////
////                                                             ////
////  Autor: Gregor Koch                                         ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`default_nettype none

module ICacheCtrl
#(
	parameter DIRECT_MEM_OUTPUT = 0
)(
	input wire clk,
	input wire rst,

	input wire read,
	input wire new_block,
	input wire index_equals_saved,
	input wire hit,
	output reg valid,

	output reg mem_req,
	input wire mem_ack,
	input wire mem_data_ready,

	output reg sample_data,
	output reg sample_offset,
	output wire we_lru_mem,
	output wire save_index,
	output reg output_from_reg,
	output wire reload
);

	reg has_reloaded = 1'b0;	// stored block from memory in cache in the last cycle

	assign  reload      = mem_data_ready;
	assign  save_index  = reload;
	assign  we_lru_mem  = valid & hit | reload;
		// write to the lru memory when hit or a way is arbited to store a new block

	always @(posedge clk)
		has_reloaded <= reload;


	// major state machine:

	// States Encoding
	localparam RESET = 2'b00;
	localparam READ = 2'b01;
	localparam GRANTED = 2'b11;
	localparam RELOADED = 2'b10;

	// State Regs
	reg [1:0] state;
	reg [1:0] next_state;

	// State Transition
	always @(posedge clk)
		if (rst)
			state <= RESET;
		else
			state <= next_state;

	// Output and Transition Logic
	always @(*) begin
		next_state = state;
		sample_data = 1'b0;
		sample_offset = 1'b0;
		valid = 1'b0;
		output_from_reg = 1'b0;
		mem_req = 1'b0;

		case (state)
			RESET:
				if (~rst & read) begin
					next_state = READ;
					sample_data = 1'b1;
					sample_offset = 1'b1;
				end
			READ: begin
				// read data is only valid when there was no write to the same index
				sample_data = read & new_block;
				sample_offset = read;
				if (!(has_reloaded & index_equals_saved)) begin
					valid = hit;
					if (!hit) begin
						sample_data = 1'b0;
						sample_offset = 1'b0;
						mem_req = 1'b1;
						if (mem_ack)
							next_state = GRANTED;
					end
				end
			end
			GRANTED: begin
				if (mem_data_ready) begin
					next_state = RELOADED;
					if (DIRECT_MEM_OUTPUT) begin
						sample_data = read & new_block;
						sample_offset = read;
						valid = 1'b1;
						if (read & new_block)
							next_state = READ;
					end
				end
			end
			RELOADED: begin
				sample_data = read & new_block;
				sample_offset = read;
				valid = 1'b1;
				output_from_reg = 1'b1;
				if (read & new_block)
					next_state = READ;
			end
			default: begin
				next_state = 2'bxx;
				sample_data = 1'bx;
				sample_offset = 1'bx;
				valid = 1'bx;
				output_from_reg = 1'bx;
				mem_req = 1'bx;
			end
		endcase
	end

endmodule
