/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
 
/////////////////////////////////////////////////////////////////////
////                                                             ////
////  SpartanMC Instruction Cache                                ////
////  zur Verwendung, wenn ein externer Firmware-Speicher        ////
////  eingesetzt wird.                                           ////
////                                                             ////
////  Autor: Gregor Koch                                         ////
////  Erstellt auf Basis einer Implementierung von Sven Himer    ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`default_nettype none

module InstructionCache
#(
	parameter p_mem_address_width = 18,
	parameter p_word_width = 8,
	parameter p_ways = 4,	// for n-way set associative cache
	parameter p_index_len = 8,
	parameter p_blockoffset_len = 2,
	parameter p_tag_len = (p_mem_address_width-p_index_len-p_blockoffset_len),
	parameter p_block_width = (2**p_blockoffset_len)*p_word_width,
	parameter p_direct_mem_output = 0
 )(
	input wire clk_i,
	input wire rst_i,

	input wire access_i,
	input wire [p_mem_address_width-1:0] address_i,
	input wire jmp_i,
	output wire [p_word_width-1:0] cache_data_o,
	output wire valid_o,

	// Global firmware memory interface
	output wire request,
	output wire [p_mem_address_width-1:0] load_addr,
	input wire mem_acknowledge,
	input wire mem_data_valid,
	input wire [p_block_width-1:0] mem_data_in
);

wire [p_blockoffset_len-1:0] offset_s;
wire [p_index_len-1:0] index_s;
wire [p_tag_len-1:0] tag_s;

generate
if( p_blockoffset_len > 0 )
	assign  {tag_s,index_s,offset_s} = address_i;
else begin
	assign  {tag_s,index_s}	= address_i;
end
endgenerate

reg [p_tag_len-1:0]         tag_reg;
reg [p_index_len-1:0]       index_reg;
reg [p_blockoffset_len-1:0] offset_reg;
reg [p_index_len-1:0]       index_buf;

wire sample_data;
wire sample_offset;
wire save_index;

always @(posedge clk_i) begin
	if (sample_data) begin
		tag_reg <= tag_s;
		index_reg <= index_s;
	end
	if (sample_offset)
		offset_reg <= offset_s;
	if (save_index)
		index_buf <= index_reg;
end

assign  load_addr = {tag_reg,index_reg,{p_blockoffset_len{1'b0}}};

// check if the next instruction to load is from the same block/cacheline:
wire    new_block;

generate
if( p_blockoffset_len > 0 ) begin
	wire    last_of_bl      = offset_reg == {p_blockoffset_len{1'b1}};
	assign  new_block       = last_of_bl | jmp_i;
end
else
	assign  new_block       = 1'b1;
endgenerate

wire    index_equals_saved = (index_reg == index_buf);


// cache memory and hit detection

wire [p_ways-1:0] we_bram;

wire [p_tag_len-1:0] tag_out[p_ways-1:0];
wire [p_block_width-1:0] data_mux[p_ways-1:0];
wire [p_ways-1:0] valid_data;

wire [p_ways-1:0] hit_ref;
wire [p_ways-1:0] hit_valid_ref;
wire hit;

reg     [p_block_width-1:0]     hit_data;

genvar xx;
generate
for (xx=0; xx<p_ways; xx=xx+1) begin:instantiations
	// MEMORY DATA
	sdp_ram #(
		.ADDRESS_WIDTH(p_index_len),
		.DATA_WIDTH(p_block_width)
	 ) DataMemory (
		.clk(clk_i),
		// Write Port
		.we(we_bram[xx]),
		.wr_address(index_reg),
		.data_in( mem_data_in ),
		// Read Port
		.rd_enable(sample_data),
		.rd_address(index_s),
		.data_out(data_mux[xx])
	);

	// MEMORY VALID+TAG
	sdp_ram #(
		.ADDRESS_WIDTH(p_index_len),
		.DATA_WIDTH(1+p_tag_len),
		.ASYNCHRONOUS_READ(0)
	 ) TagMemory (
		.clk(clk_i),
		// Write Port
		.we(we_bram[xx]),
		.wr_address(index_reg),
		.data_in( {1'b1,tag_reg} ),
		// Read Port
		.rd_enable(sample_data),
		.rd_address(index_s),
		.data_out({valid_data[xx],tag_out[xx]})
	);

	// hit, but maybe invalid
	assign hit_ref[xx] = (tag_out[xx] == tag_reg);
end
endgenerate

// valid hit
assign hit_valid_ref = hit_ref & valid_data;
assign hit = |hit_valid_ref;

// Set-associative Output Mux
integer i;
always @(*) begin
	hit_data = {(p_block_width){1'bx}};
	for (i=0; i<p_ways; i=i+1)
		if (hit_valid_ref == {{(p_ways-1){1'b0}}, 1'b1} << i)
			hit_data = data_mux[i];
end


// load buffer and output

wire    output_from_reg;

reg     [p_block_width-1:0]     load_data_buf;
wire    [p_block_width-1:0]     block_data;
wire    [p_word_width-1:0]      offset_mux[2**p_blockoffset_len-1:0];

always @(posedge clk_i)
	if (mem_data_valid)
		load_data_buf   <= mem_data_in;
	// Full cache line is assumed to be available at once
	// (needs to be changed for another memory type)

assign  block_data      = (p_direct_mem_output && mem_data_valid) ? mem_data_in
				: output_from_reg ? load_data_buf : hit_data;

// Offset Mux
generate
for (xx=0; xx<2**p_blockoffset_len; xx=xx+1) begin:offsetmux
	assign  offset_mux[xx]	= block_data[(xx+1)*p_word_width-1:xx*p_word_width];
end
endgenerate

assign  cache_data_o	= offset_mux[(p_blockoffset_len > 0) ? offset_reg : 0];


// Arbitration and LRU FLAGS
wire [p_ways-1:0] arbit_way;
wire [2:0] lruf_data;
wire [2:0] lruf_calc;
wire we_lru_mem;
localparam p_lru_bits = (p_ways == 1) ? 0	// number of needed LRU bits
		: (p_ways == 2) ? 1 : (p_ways == 4) ? 3 : -1;

generate
if (p_ways == 1) begin
	assign arbit_way = 1'b1;
end
else if (p_ways == 2) begin
	assign arbit_way = {~lruf_data[0],lruf_data[0]};
	assign lruf_calc[0] = !hit ? ~lruf_data[0] : hit_valid_ref[1];
end
else if (p_ways == 4) begin
	wire [p_ways-1:0] arbit_lru;
	wire [p_ways-1:0] arbit_uninit;
	wire [p_ways-1:0] valid_calc;

	// Arbit Valid Flags
	assign valid_calc = {valid_data[p_ways-2:0],1'b1};
	assign arbit_uninit = ~valid_data & valid_calc;

	CachePLRU4 PLRU4(
		.use_vect_i(hit ? hit_valid_ref : arbit_way),
		// Select the arbited
		.arbit_o(arbit_lru),
		// Desicion Flags In/Out
		.plruf_i(lruf_data),
		.plruf_o(lruf_calc)
	);

	wire allinit;	// Assign if all valid-flags are set
	assign allinit = &valid_data;
	assign arbit_way = allinit ? arbit_lru : arbit_uninit;
end

if (p_lru_bits > 0)
	// An asynchronously readable single-ported RAM is used
	// to read and write at the same address within a single cycle.
	sp_ram
	#(
		.ADDRESS_WIDTH(p_index_len),
		.DATA_WIDTH(p_lru_bits),
		.ASYNCHRONOUS_READ(1)
	 ) PLRUMemory (
		.clk(clk_i),
		// Write Port
		.we( we_lru_mem ),
		.address( index_reg ),
		.data_in(lruf_calc[p_lru_bits-1:0]),
		// Read Port
		.data_out(lruf_data[p_lru_bits-1:0])
	);
endgenerate

wire reload;
assign we_bram = {p_ways{reload}} & arbit_way;


// Cache Control FSM
ICacheCtrl #(
	.DIRECT_MEM_OUTPUT(p_direct_mem_output)
) cache_controller (
	// Reset & Clock
	.clk(clk_i),
	.rst(rst_i),
	// Control
	.read(access_i),
	.new_block(new_block),
	.hit(hit),
	.index_equals_saved(index_equals_saved),
	.valid(valid_o),
	// Memory interface
	.mem_req(request),
	.mem_ack(mem_acknowledge),
	.mem_data_ready(mem_data_valid),
	// Mux and reg control
	.sample_data(sample_data),
	.sample_offset(sample_offset),
	.we_lru_mem(we_lru_mem),
	.save_index(save_index),
	.output_from_reg(output_from_reg),
	.reload(reload)
);

// synopsys translate_off
/*// access and miss counter for evaluation purposes
integer miss_counter = 0;
reg sampled_data = 1'b0;
always @(posedge clk_i) begin
	if (reload) begin
		$display("Address 0x%H (index: 0x%H) reloaded to %b", {tag_reg,index_reg,{p_blockoffset_len{1'b0}}}, index_reg, arbit_way);
		miss_counter = miss_counter + 1;
	end
	if (sampled_data && valid_o) begin
		$display("Hit address 0x%H (index: 0x%H) in way %b", {tag_reg,index_reg,{p_blockoffset_len{1'b0}}}, index_reg, hit_valid_ref);
	end
	sampled_data <= sample_data;
end
*/
// synopsys translate_on

endmodule
