/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

/////////////////////////////////////////////////////////////////////
////                                                             ////
////  Global firmware memory module                              ////
////  Block RAM based global firmware memory                     ////
////                                                             ////
////  Autor: Gregor Koch                                         ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`default_nettype none

module firmware_mem
#(
	parameter RAMBLOCKS = 8,
	parameter DATA_WIDTH = 18,
	parameter CORES = 4,
	parameter USE_TWO_PORTS = 1,
	parameter PA_CORES = USE_TWO_PORTS ? CORES/2 : CORES
)
(
	clk,
	addrin,
	iout,
	request,
	grant
);

localparam PB_CORES = CORES - PA_CORES;
localparam OFFSET_BITS = (DATA_WIDTH == 18) ? 0 : (DATA_WIDTH == 36) ? 1 : (DATA_WIDTH == 72) ? 2 : -1;

input wire clk;
input wire	[CORES*18-1:0]		addrin;
output	[CORES*DATA_WIDTH-1:0]	iout;
input wire	[CORES-1:0]		request;
output wire	[CORES-1:0]		grant;


// arbiters:
generate

if (PA_CORES > 1)
	arbiter #(PA_CORES) core_arbiter (
		.clk(clk),
		.request(request[PA_CORES-1:0]),
		.grant(grant[PA_CORES-1:0])
	);
else if (PA_CORES == 1)
	assign grant[0] = request[0];

if (PB_CORES > 1)
	arbiter #(PB_CORES) core_arbiter (
		.clk(clk),
		.request(request[CORES-1:PA_CORES]),
		.grant(grant[CORES-1:PA_CORES])
	);
else if (PB_CORES == 1)
	assign grant[CORES-1] = request[CORES-1];

endgenerate


wire	[17:0]	addr_array[CORES-1:0];
reg	[17:0]	selected_addr;
reg	[17:0]	selected_addr_b;

wire	[CORES*DATA_WIDTH-1:0]	iout;
reg	[DATA_WIDTH-1:0]	iout_array[CORES-1:0];
reg	[DATA_WIDTH-1:0]	port_a_out;
reg	[DATA_WIDTH-1:0]	port_b_out;

// assign port buses to/from array:
genvar i;
generate
	for (i=0; i<CORES; i=i+1)
	begin : CORE
		assign addr_array[i] = addrin[18*i+17:18*i];
		assign iout[(i+1)*DATA_WIDTH-1:i*DATA_WIDTH] = iout_array[i];
	end
endgenerate

// selection of the arbitrated address and
// assignment of the correct output port to the
// instruction output of every core
integer j;
always @(*) begin
	selected_addr = {18{1'bx}};
	for (j = 0; j < PA_CORES; j = j+1) begin
		if (grant[j])
			selected_addr = addr_array[j];
		iout_array[j] = port_a_out;
	end
	selected_addr_b = {18{1'bx}};
	for (j = PA_CORES; j < CORES; j = j+1) begin
		if (grant[j])
			selected_addr_b = addr_array[j];
		iout_array[j] = port_b_out;
	end
end


wire [DATA_WIDTH-1:0] bl_out_a[RAMBLOCKS-1:0];
wire [DATA_WIDTH-1:0] bl_out_b[RAMBLOCKS-1:0];

// generation of the RAM blocks:
generate
	for (i=0; i<RAMBLOCKS; i=i+1)
	begin : MEM_BL
		// select signals for this memory block:
		wire sel_block;
		wire pb_sel_block;

		pselect #(
			.ADDR_WIDTH	( 7 ),
			.BASE_WIDTH	( 7 ),
			.BASE_ADDR	( i )
		)
		memsela (
			.addr	( selected_addr[17:11] ),
			.activ_peri ( 1'b1 ),
			.select ( sel_block )
		);

		if (USE_TWO_PORTS)
			pselect #(
				.ADDR_WIDTH	( 7 ),
				.BASE_WIDTH	( 7 ),
				.BASE_ADDR	( i )
			)
			memselb (
				.addr	( selected_addr_b[17:11] ),
				.activ_peri ( 1'b1 ),
				.select ( pb_sel_block )
			);
		else begin
			assign pb_sel_block = 1'b0;
		end

		// Block RAM outputs
		wire	[DATA_WIDTH/2-1:0]	bramA_out_a;
		wire	[DATA_WIDTH/2-1:0]	bramA_out_b;
		wire	[DATA_WIDTH/2-1:0]	bramB_out_a;
		wire	[DATA_WIDTH/2-1:0]	bramB_out_b;

		// Block RAM instantiation
		bram_container #(
			.DATA_WIDTH	( DATA_WIDTH/2 )
		)
		bram_cont_A (
			.dia(	{(DATA_WIDTH/2) {1'b0}} ),
			.addra(	selected_addr[10:OFFSET_BITS] ),
			.ena(	1'b1 ),
			.wea(	1'b0 ),
			.rsta(	! sel_block ),
			.clka(	clk ),
			.doa(	bramA_out_a ),
			.dib(	{(DATA_WIDTH/2) {1'b0}} ),
			.addrb(	selected_addr_b[10:OFFSET_BITS] ),
			.enb(	USE_TWO_PORTS != 0 ),
			.web(	1'b0 ),
			.rstb(	! pb_sel_block ),
			.clkb(	clk ),
			.dob(	bramA_out_b )
		);

		bram_container #(
			.DATA_WIDTH	( DATA_WIDTH/2 )
		)
		bram_cont_B (
			.dia(	{(DATA_WIDTH/2) {1'b0}} ),
			.addra(	selected_addr[10:OFFSET_BITS] ),
			.ena(	1'b1 ),
			.wea(	1'b0 ),
			.rsta(	! sel_block ),
			.clka(	clk ),
			.doa(	bramB_out_a ),
			.dib(	{(DATA_WIDTH/2) {1'b0}} ),
			.addrb(	selected_addr_b[10:OFFSET_BITS] ),
			.enb(	USE_TWO_PORTS != 0 ),
			.web(	1'b0 ),
			.rstb(	! pb_sel_block ),
			.clkb(	clk ),
			.dob(	bramB_out_b )
		);

		// assignment of BRAM outputs to the output of the RAMBLOCK:
		// ensure that the BRAMs can be initialized like
		// a pair of 9-bit-wide BRAMs
		genvar k;
		for (k=0; k<DATA_WIDTH; k=k+18) begin:bramoutputs
			assign	bl_out_a[i][k+8:k+0]	= bramA_out_a[k/2+8:k/2+0];
			assign	bl_out_a[i][k+17:k+9]	= bramB_out_a[k/2+8:k/2+0];
			assign	bl_out_b[i][k+8:k+0]	= bramA_out_b[k/2+8:k/2+0];
			assign	bl_out_b[i][k+17:k+9]	= bramB_out_b[k/2+8:k/2+0];
		end
	end
endgenerate

// or gate for both output ports
always @(*) begin
	port_a_out = {DATA_WIDTH{1'b0}};
	port_b_out = {DATA_WIDTH{1'b0}};
	for (j = 0; j < RAMBLOCKS; j = j+1) begin
		port_a_out = port_a_out | bl_out_a[j];
		if (USE_TWO_PORTS)
			port_b_out = port_b_out | bl_out_b[j];
	end
end

endmodule
