/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`default_nettype none
`timescale 1ns / 1ps

module global_mem_master #(
	parameter CONNECTIONS 	= 1,
	parameter CACHE			= "FALSE",
	parameter RAMBLOCKS		= 4,
	parameter BASE_ADDR		= 18'h10000,
	parameter DEBUG 		= "TRUE"
) (

	input wire							clk,
	input wire							reset,

	input wire		[10*CONNECTIONS-1:0]				mem_addr_block,
    input wire		[8*CONNECTIONS-1:0]					mem_addr_high,
	input wire		[18*CONNECTIONS-1:0]				mem_di,
	input wire		[CONNECTIONS-1:0]					mem_wr_low_in,

	input wire		[CONNECTIONS-1:0]					mem_wr_high_in,
	input wire		[CONNECTIONS-1:0]					mem_access,
	output wire		[18*CONNECTIONS-1:0]				mem_do,
	output wire		[CONNECTIONS-1:0]					mem_busy_out

);

function integer log2;
	input integer value;
	for (log2=0; value>0; log2=log2+1)
		value = value>>1;
endfunction

parameter BLOCK_ADR_LEN = log2(RAMBLOCKS-1);
parameter ADR_LEN = 11 + BLOCK_ADR_LEN;

wire [CONNECTIONS-1:0] grant;
wire [CONNECTIONS-1:0] mem_do_enable;
wire [ADR_LEN*CONNECTIONS-1:0] grant_fanout;
wire [18*CONNECTIONS-1:0] data_fanout;
wire [ADR_LEN*CONNECTIONS-1:0] addr_cleaned;
wire [17:0] data_out;

wire [ADR_LEN-1:0] selected_addr;
wire [ADR_LEN-1:0] addr_array[CONNECTIONS-1:0];
reg [ADR_LEN-1:0] addr_array_reg[CONNECTIONS-1:0];
wire [CONNECTIONS*ADR_LEN-1:0] addr_array_reg_vector;

wire [17:0] selected_data;
wire [18*CONNECTIONS-1:0] data_array;
reg [17:0] data_array_reg[CONNECTIONS-1:0];
wire [17:0] data_out_array[CONNECTIONS-1:0];
wire [CONNECTIONS*18-1:0] data_out_vector;

wire selected_wr_low;
wire selected_wr_high;


//regs to keep signals when core is in stall and changes inputs
reg		[CONNECTIONS-1:0]					mem_access_reg;
reg 	[8*CONNECTIONS-1:0]					mem_addr_high_reg;

//internal wires equivalents to either use input directly or stored value from regs if in stall
wire	[CONNECTIONS-1:0]					mem_access_intern;
wire 	[8*CONNECTIONS-1:0]					mem_addr_high_intern;



//only needed for CACHE == TRUE

wire [CONNECTIONS-1:0] 		use_cache;
wire [CONNECTIONS-1:0]		use_cache_read;
wire [17:0] 				cache_data_array [CONNECTIONS-1:0];
wire [CONNECTIONS*18-1:0]	cache_data;
reg [CONNECTIONS-1:0]		prev_grant;
reg [CONNECTIONS-1:0]		global_access;
wire [CONNECTIONS-1:0]		global_access_combined;
wire [CONNECTIONS-1:0]		global_select;

initial begin
	if (|(RAMBLOCKS & (RAMBLOCKS-1))) begin
		$display("Global memory at %m is not set to a power of two number of ramblocks. This is not supported!");
		$stop();
	end
end

always@(posedge clk) begin
	if(reset)
		prev_grant <= {CONNECTIONS { 1'b0 }};
	else
		prev_grant <= grant;
end


wire 	[CONNECTIONS-1:0] 	mem_wr_low;
reg 	[CONNECTIONS-1:0] 	mem_wr_low_reg;
wire 	[CONNECTIONS-1:0] 	mem_wr_high;
reg 	[CONNECTIONS-1:0] 	mem_wr_high_reg;

genvar i;
generate

	assign global_access_combined = global_access | global_select;

	for(i=0;i<CONNECTIONS;i=i+1) begin:global_acces_gen
		always@(posedge clk) begin
			if(reset) begin
				global_access[i] <= {CONNECTIONS{1'b0}}; 
			end else begin
				if (global_select[i] & ~grant[i] ) begin
					global_access[i] <= 1'b1;
				end else if(grant[i] || use_cache_read[i]) begin
					global_access[i] <= 1'b0;
				end
			end
		end		
	end

	assign mem_busy_out = global_access & ~(use_cache_read);

	for(i=0;i<CONNECTIONS;i=i+1) begin:gen_mem_wr

		assign mem_wr_low[i] =  mem_busy_out[i] ? mem_wr_low_reg[i] : mem_wr_low_in[i];
       	assign mem_wr_high[i] =  mem_busy_out[i] ? mem_wr_high_reg[i] : mem_wr_high_in[i];		

    	always@(posedge clk) begin
    		if(reset) begin
            	mem_wr_low_reg[i] <= 0;
            	mem_wr_high_reg[i] <= 0;
        	end else if ( mem_busy_out[i] == 0) begin
            	mem_wr_high_reg[i] <= mem_wr_high_in[i];
            	mem_wr_low_reg[i] <= mem_wr_low_in[i];
        	end
    	end
	end


	if(CONNECTIONS > 1) begin

		arbiter #(CONNECTIONS) core_arbiter (
			.clk(clk),
			//only request grant if cache miss
			.request(~use_cache_read & global_access_combined),
			.grant(grant)
		);
	end else begin
		assign grant = ~use_cache_read & global_access_combined;
	end


   	for (i=0; i<CONNECTIONS; i=i+1)
   	begin : CORE

		//if in stall keep previous values
		assign mem_access_intern[i] = mem_access_reg[i];
		assign mem_addr_high_intern[8*i+7:8*i] = mem_addr_high_reg[8*i+7:8*i];

		always@(posedge clk) begin
			if (reset) begin
				mem_access_reg[i] <= 0;
				mem_addr_high_reg[8*i+7:8*i] <= 0;
			end else if (mem_busy_out[i] == 1'b0) begin
				mem_access_reg[i] <= mem_access[i];
				mem_addr_high_reg[8*i+7:8*i] <= mem_addr_high[8*i+7:8*i];
			end
		end



   		assign data_array[18*i+17:18*i] = mem_busy_out[i] ? data_array_reg[i] : mem_di[18*i+17:18*i];
   		assign addr_array[i] = mem_busy_out[i] ? addr_array_reg[i] : { mem_addr_high[8*i+ADR_LEN-10-1:8*i], mem_addr_block[10*i+10-1:10*i] };
		assign mem_do_enable[i] = (prev_grant[i] | (use_cache_read[i] & mem_access_reg[i]));
		assign mem_do[18*i+17:18*i] = mem_do_enable[i] ? data_out_array[i] : 18'b0;
		assign data_out_vector[i*18+17:i*18] = data_out_array[i];
		assign addr_cleaned[ADR_LEN*i+ADR_LEN-1:ADR_LEN*i] = addr_array[i];
		assign addr_array_reg_vector[ADR_LEN*i+ADR_LEN-1:ADR_LEN*i] = addr_array_reg[i];
		assign cache_data_array[i] = cache_data[i*18+17:18*i];
		assign grant_fanout[ADR_LEN*i+ADR_LEN-1:ADR_LEN*i] = { ADR_LEN { grant[i] }};
		assign data_fanout[18*i+17:18*i] = { 18 { grant[i] }};

		always@(posedge clk) begin
			if (reset) begin
				data_array_reg[i] <= 0;
				addr_array_reg[i] <= 0;
			end else if (mem_busy_out[i] == 0) begin
					data_array_reg[i] <= mem_di[18*i+17:18*i];
   					addr_array_reg[i] <= { mem_addr_high[8*i+ADR_LEN-10-1:8*i], mem_addr_block[10*i+10-1:10*i] };
			end
		end



		pselect  #(
        			.ADDR_WIDTH	(7-BLOCK_ADR_LEN		),
        			.BASE_WIDTH	(7-BLOCK_ADR_LEN		),
        			.BASE_ADDR	(BASE_ADDR[17:11+BLOCK_ADR_LEN])
        		)
        	memsel	(
        			.addr		(mem_addr_high[8*i+7:8*i+1+BLOCK_ADR_LEN] ),
        			.activ_peri	(mem_access[i]	    ),
        			.select		(global_select[i]	)
        		);

		assign data_out_array[i] = use_cache_read[i] ? cache_data_array[i] : data_out;
   	end
if(CACHE == "TRUE") begin:cache_instantiation

	wire [ADR_LEN*CONNECTIONS-1:0] 	addr;	
	wire [CONNECTIONS-1:0] 			read;
	wire [CONNECTIONS-1:0] 			write;
	wire [18*CONNECTIONS-1:0] 		data_in;
	wire [CONNECTIONS-1:0] 			enable_input;
	

	assign read = ~(mem_wr_low_reg | mem_wr_high_reg) & prev_grant;
	assign write = (mem_wr_low_in | mem_wr_high_in)&global_select;

	for (i=0; i<CONNECTIONS; i=i+1)
	begin : cache_signals
		assign addr[i*ADR_LEN+ADR_LEN-1:i*ADR_LEN] = read[i] ? addr_array_reg[i] : addr_array[i];
		assign data_in[i*18+18-1:i*18] = ~write[i] ? data_out : mem_di[18*i+17:18*i];
		assign enable_input[i] = write[i] ? 1'b1 : read[i]? prev_grant[i] : 1'b0;
	end


	one_word_read_cache #(
				.CONNECTIONS(CONNECTIONS),
				.ADDR_LEN(ADR_LEN)
			) cache_inst (
				.clk(clk), 
				.reset(reset), 
				.addr(addr),
				.data_in(data_in), 
				.wr_low(mem_wr_low), 
				.wr_high(mem_wr_high), 
				.enable_input(enable_input), 
				.data_out(cache_data),
				.valid(use_cache)
			);
end else begin:cache_no_instantiation
	assign use_cache = {CONNECTIONS{1'b0}};
	assign cache_data = {CONNECTIONS{18'h0}};
end

	assign use_cache_read = use_cache & ~(mem_wr_low_reg | mem_wr_high_reg);

for (i=0; i<CONNECTIONS; i=i+1)begin:debug
if(DEBUG == "TRUE")begin:debug_output
	always@(negedge clk ) begin
		if(!reset & & global_select[i])begin
			if(mem_wr_low[i] | mem_wr_high[i])
				$display("[%0t][GMEM] WR-DATA:   core %0d  addr %x data %x",$time,i,addr_array[i],mem_di[18*i+17:18*i]);
			else
				$display("[%0t][GMEM] REQ-DATA:  core %0d  addr %x",$time,i,addr_array[i]);
		end
	end
	always@(negedge clk) begin
		if(!reset & mem_do_enable[i])
			$display("[%0t][GMEM] RESP:      core %0d data %x",$time,i,data_out_array[i]);
	end

end
end

endgenerate



orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	1			)
) wr_low_or (
	.a 			( mem_wr_low & grant ),
	.y 			( selected_wr_low )
);

orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	1			)
) wr_high_or (
	.a 			( mem_wr_high & grant ),
	.y 			( selected_wr_high )
);

orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	18			)
) data_in_or (
	.a 			( data_array & data_fanout ),
	.y 			( selected_data )
);

orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	ADR_LEN	)
) address_or (
	.a 			( addr_cleaned & grant_fanout),
	.y 			( selected_addr )
);


memory #(
	.RAMBLOCKS  ( RAMBLOCKS ),
	.ADR_LEN	( ADR_LEN	)
) mem (
	.clk		 ( clk				),
	.data_inA	 ( selected_data 	),
	.data_outA	 ( data_out			),
	.wr_high	 ( selected_wr_high	),
	.wr_low		 ( selected_wr_low	),
	.mem_addressA( selected_addr	),
	.data_outB(),
	.mem_addressB()
);



endmodule
