/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`default_nettype none


module global_mem_masterv2 #(
	parameter CONNECTIONS 			= 1,
	parameter CONNECTIONS_BARRIER	= 1,
	parameter CACHE					= 1,
	parameter BUFFER				= 0, //buffer is  unnecessary due to new access pattern
	parameter RAMBLOCKS				= 4,
	parameter BASE_ADDR				= 18'h10000
) (

	input wire							clk,
	input wire							reset,

	input wire		[10*CONNECTIONS-1:0]				mem_addr_block,
    input wire		[8*CONNECTIONS-1:0]					mem_addr_high,
	input wire		[18*CONNECTIONS-1:0]				mem_di,
	input wire		[CONNECTIONS-1:0]					mem_wr_low_in,

	input wire		[CONNECTIONS-1:0]					mem_wr_high_in,
	input wire		[CONNECTIONS-1:0]					mem_access,
	output wire		[18*CONNECTIONS-1:0]				mem_do,
	output reg		[CONNECTIONS-1:0]					mem_busy_out,

	output wire 										wait_done,
	input wire		[CONNECTIONS_BARRIER-1:0]			wait_for
);

function integer log2;
	input integer value;
	for (log2=0; value>0; log2=log2+1)
		value = value>>1;
endfunction

parameter BLOCK_ADR_LEN = log2(RAMBLOCKS-1);
parameter ADR_LEN = 11 + BLOCK_ADR_LEN;

wire [CONNECTIONS-1:0] wr_grant;
wire [CONNECTIONS-1:0] rd_grant;
reg [CONNECTIONS-1:0] write_port;
wire [ADR_LEN*CONNECTIONS-1:0] wr_grant_fanout;
wire [ADR_LEN*CONNECTIONS-1:0] rd_grant_fanout;
wire [18*CONNECTIONS-1:0] data_fanout;
wire [ADR_LEN*CONNECTIONS-1:0] addr_cleaned;
wire [17:0] data_out_wr;
wire [17:0] data_out_rd;

wire [ADR_LEN-1:0] selected_wr_addr;
wire [ADR_LEN-1:0] selected_rd_addr;
wire [ADR_LEN-1:0] addr_array[CONNECTIONS-1:0];
reg [ADR_LEN-1:0] addr_array_reg[CONNECTIONS-1:0];

wire [17:0] selected_data;
wire [18*CONNECTIONS-1:0] data_array;
reg [17:0] data_array_reg[CONNECTIONS-1:0];
wire [17:0] data_out_array[CONNECTIONS-1:0];

wire selected_wr_low;
wire selected_wr_high;


//regs to keep signals when core is in stall and changes inputs
reg		[CONNECTIONS-1:0]					mem_access_reg;
reg		[CONNECTIONS_BARRIER-1:0]			wait_for_reg;
reg 	[8*CONNECTIONS-1:0]					mem_addr_high_reg;

//internal wires equivalents to either use input directly or stored value from regs if in stall
wire	[CONNECTIONS-1:0]					mem_access_intern;
wire	[CONNECTIONS_BARRIER-1:0]			wait_for_intern;
wire 	[8*CONNECTIONS-1:0]					mem_addr_high_intern;



//only needed for CACHE == TRUE
reg [CONNECTIONS-1:0] 	cache_valid;
reg [CONNECTIONS-1:0] 	use_cache;
reg [CONNECTIONS-1:0]	write_to_cache;
reg [ADR_LEN-1:0]		cache_addr [CONNECTIONS-1:0];
reg [17:0] 				cache_data [CONNECTIONS-1:0];
reg [CONNECTIONS-1:0]	prev_access;
wire [CONNECTIONS-1:0]  cache_hit;
wire [CONNECTIONS-1:0]	global_access;
wire [CONNECTIONS-1:0]	global_select;
wire [CONNECTIONS-1:0]	mem_busy;

initial begin
	if (|(RAMBLOCKS & (RAMBLOCKS-1))) begin
		$display("Global memory at %m is not set to a power of two number of ramblocks. This is not supported!");
		$stop();
	end
end

always@(posedge clk) begin
	if(reset)
		prev_access <= {CONNECTIONS { 1'b0 }};
	else begin
		prev_access <= global_access;
		write_port <= wr_grant;
	end
end

assign wait_done = &wait_for_intern;

wire 	[CONNECTIONS-1:0] 	mem_wr_low;
reg 	[CONNECTIONS-1:0] 	mem_wr_low_reg;
wire 	[CONNECTIONS-1:0] 	mem_wr_high;
reg 	[CONNECTIONS-1:0] 	mem_wr_high_reg;

genvar i;
generate

	assign global_access =  global_select;

	always@(posedge clk) begin
		if(reset) begin
			mem_busy_out <= 0; 
		end else begin
			mem_busy_out <= mem_busy;
		end
	end


	for(i=0;i<CONNECTIONS;i=i+1) begin:gen_mem_wr

		assign mem_wr_low[i] =  mem_wr_low_reg[i];
       	assign mem_wr_high[i] =  mem_wr_high_reg[i];		

    	always@(posedge clk) begin
    		if(reset) begin
            	mem_wr_low_reg[i] <= 0;
            	mem_wr_high_reg[i] <= 0;
        	end else if ( mem_busy[i] == 0 ) begin
            	mem_wr_high_reg[i] <= mem_wr_high_in[i];
            	mem_wr_low_reg[i] <= mem_wr_low_in[i];
        	end
    	end
	end


	if(CACHE == "TRUE") begin:CACHE_gen

		always@(posedge clk) begin
			if(reset)
				use_cache <= {CONNECTIONS { 1'b0 }};
			else
				use_cache <= cache_hit & cache_valid;
		end

		if(CONNECTIONS > 1) begin

			arbiter #(CONNECTIONS) core_arbiter_wr (
				.clk(clk),
				.request(global_access & (mem_wr_low | mem_wr_high)),
				.grant(wr_grant)
			);

			arbiter #(CONNECTIONS) core_arbiter_rd (
            	.clk(clk),
            	//only request grant iff cache miss
            	.request(~cache_hit & global_access & (~mem_wr_low) & (~mem_wr_high)),
            	.grant(rd_grant)
            );
		end else begin
			assign wr_grant = global_access & (mem_wr_low | mem_wr_high);
			assign rd_grant = ~cache_hit & global_access & (~mem_wr_low) & (~mem_wr_high);
		end

		//memory is busy iff trying to access and no cache hit and no grant was given
		assign mem_busy = global_access & (~wr_grant) & ~cache_hit & (~rd_grant);

		initial begin
    		cache_valid <= {CONNECTIONS { 1'b0 }};
    		write_to_cache <= {CONNECTIONS { 1'b0 }};
		end

	end else begin:NOCACHE_gen

		if(CONNECTIONS > 1) begin:begin_arbit_inst
			arbiter #(CONNECTIONS) core_arbiter_wr (
				.clk(clk),
				.request(global_access & (mem_wr_low | mem_wr_high)),
				.grant(wr_grant)
			);

			arbiter #(CONNECTIONS) core_arbiter_rd (
				.clk(clk),
				.request(global_access & (~mem_wr_low) & (~mem_wr_high)),
				.grant(rd_grant)
			);
		end else begin
        	assign wr_grant = global_access & (mem_wr_low | mem_wr_high);
        	assign rd_grant = global_access & (~mem_wr_low) & (~mem_wr_high);
        end

		assign mem_busy = global_access & (~wr_grant) & (~rd_grant);

	end

	for (i=0; i<CONNECTIONS_BARRIER; i=i+1)
   	begin : BARRIERS

		//if in stall keep previous values
		assign wait_for_intern[i] = wait_for_reg[i];

		always@(posedge clk) begin
			if (reset) begin
				wait_for_reg[i] <= 0;
			end else if (mem_busy[i] == 0) begin
				wait_for_reg[i] <= wait_for[i];
			end
		end
	end


   	for (i=0; i<CONNECTIONS; i=i+1)
   	begin : CORE

		//if in stall keep previous values
		assign mem_access_intern[i] = mem_access_reg[i];
		assign mem_addr_high_intern[8*i+7:8*i] = mem_addr_high_reg[8*i+7:8*i];

		always@(posedge clk) begin
			if (reset) begin
				mem_access_reg[i] <= 0;
				mem_addr_high_reg[8*i+7:8*i] <= 0;
			end else if (mem_busy[i] != 1) begin
				mem_access_reg[i] <= mem_access[i];
				mem_addr_high_reg[8*i+7:8*i] <= mem_addr_high[8*i+7:8*i];
			end
		end



   		assign data_array[18*i+17:18*i] = data_array_reg[i];
   		assign addr_array[i] = addr_array_reg[i];
		assign mem_do[18*i+17:18*i] = prev_access[i] & !mem_busy_out[i] ? data_out_array[i] : 18'b0;
		assign addr_cleaned[ADR_LEN*i+ADR_LEN-1:ADR_LEN*i] = addr_array[i];
		assign wr_grant_fanout[ADR_LEN*i+ADR_LEN-1:ADR_LEN*i] = { ADR_LEN { wr_grant[i] }};
		assign rd_grant_fanout[ADR_LEN*i+ADR_LEN-1:ADR_LEN*i] = { ADR_LEN { rd_grant[i] }};
		assign data_fanout[18*i+17:18*i] = { 18 { wr_grant[i] }};

		always@(posedge clk) begin
			if (reset) begin
				data_array_reg[i] <= 0;
				addr_array_reg[i] <= 0;
			end else if (mem_busy[i] == 0) begin
					data_array_reg[i] <= mem_di[18*i+17:18*i];
   					addr_array_reg[i] <= { mem_addr_high[8*i+ADR_LEN-10-1:8*i], mem_addr_block[10*i+10-1:10*i] };
			end
		end



		pselect  #(
        			.ADDR_WIDTH	(7-BLOCK_ADR_LEN		),
        			.BASE_WIDTH	(7-BLOCK_ADR_LEN		),
        			.BASE_ADDR	(BASE_ADDR[17:11+BLOCK_ADR_LEN])
        		)
        	memsel	(
        			.addr		(mem_addr_high_intern[8*i+7:8*i+1+BLOCK_ADR_LEN] ),
        			.activ_peri	(mem_access_intern[i]	    ),
        			.select		(global_select[i]	)
        		);


		if(CACHE == "TRUE") begin

			//it's s cache hit iff we try to access the right address and have a valid cache and it's a read access
			assign cache_hit[i] = (addr_array[i] == cache_addr[i]) && cache_valid[i] && !(mem_wr_low[i] || mem_wr_high[i]);
			assign data_out_array[i] = write_port[i] ? data_out_wr : (use_cache[i] ? cache_data[i] : data_out_rd);

   			always@(posedge clk) begin
   				if(reset) begin
            		cache_valid[i] <= 1'b0;
            		write_to_cache[i] <= 1'b0;

				//if anyone writes to the cached address...
   				end else if(selected_wr_addr == cache_addr[i] && (selected_wr_high || selected_wr_low)) begin

                   	if(wr_grant[i]) begin
                   		//when we are the writer invalidate cache to allow swap instruction
                   		write_to_cache[i] <= 1'b0;
						cache_valid[i] <= 1'b0;
					end else begin
						//when someone else is the writer
						//cache remains valid if it was valid before or if completely overwritten
                	 	cache_valid[i] <= cache_valid[i] || (selected_wr_high && selected_wr_low);

						//abort any writes to cache
						write_to_cache[i] <= 1'b0;

                	 	//overwrite correct part
                        if(selected_wr_low)
                        	cache_data[i][8:0] <= selected_data[8:0];
                        if(selected_wr_high)
                            cache_data[i][17:9] <= selected_data[17:9];
					end


                //grant signal already implies cache miss
                //only if we execute a new read do we want to overwrite the cache
                //a write does not overwrite the cache if it is to a different address
                end else if(wr_grant[i] && !(mem_wr_low[i] || mem_wr_high[i])) begin
                	//newly cached address
                	cache_addr[i] <= selected_wr_addr;
					cache_valid[i] <= 1'b0;
					write_to_cache[i] <= 1'b1;

                //cache should be refilled with data read
   				end else if(write_to_cache[i]) begin
                    write_to_cache[i] <= 1'b0;
                    cache_valid[i] <= 1'b1;
                    cache_data[i] <= data_out_rd;
   				end
   			end

   		end else begin
   			assign data_out_array[i] = write_port[i] ? data_out_wr : data_out_rd;
   		end

   	end
endgenerate


orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	1			)
) wr_low_or (
	.a 			( mem_wr_low & wr_grant ),
	.y 			( selected_wr_low )
);

orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	1			)
) wr_high_or (
	.a 			( mem_wr_high & wr_grant ),
	.y 			( selected_wr_high )
);

orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	18			)
) data_in_or (
	.a 			( data_array & data_fanout ),
	.y 			( selected_data )
);

orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	ADR_LEN	)
) address_wr_or (
	.a 			( addr_cleaned & wr_grant_fanout),
	.y 			( selected_wr_addr )
);

orgate #(
	.OR_WIDTH	(	CONNECTIONS	),
	.BUS_WIDTH	(	ADR_LEN	)
) address_rd_or (
	.a 			( addr_cleaned & rd_grant_fanout),
	.y 			( selected_rd_addr )
);

memory #(
	.RAMBLOCKS  ( RAMBLOCKS ),
	.ADR_LEN	( ADR_LEN	)
) mem (
	.clk		 ( clk				),
	.data_inA	 ( selected_data 	),
	.data_outA	 ( data_out_wr		),
	.wr_high	 ( selected_wr_high	),
	.wr_low		 ( selected_wr_low	),
	.mem_addressA( selected_wr_addr	),
	.mem_addressB( selected_rd_addr	),
	.data_outB	 ( data_out_rd		)
);



endmodule
