/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`default_nettype none

module memory #(
	parameter RAMBLOCKS		= 4,
	parameter ADR_LEN		= 13
) (
	input wire							clk,

	input wire		[17:0]				data_inA,
    output wire		[17:0]				data_outA,
	input wire		  					wr_low,

	input wire							wr_high,
	input wire		[ADR_LEN-1:0]		mem_addressA,
	input wire		[ADR_LEN-1:0]		mem_addressB,
	output wire		[17:0]				data_outB
);

parameter BLOCK_ADR_LEN = ADR_LEN - 11;

wire	[RAMBLOCKS-1:0]	access_bramA;
wire	[RAMBLOCKS-1:0]	access_bramB;

wire	[18*RAMBLOCKS-1:0]	data_out_allA;
wire	[18*RAMBLOCKS-1:0]	data_out_allB;

generate
	if	(RAMBLOCKS	== 1) begin
		assign	data_outA =	data_out_allA[17:0];
		assign	data_outB =	data_out_allB[17:0];
	end else begin
		orgate #(
			.OR_WIDTH	(RAMBLOCKS	),
			.BUS_WIDTH	(18		)
			)
		dmem_orA (
			.a		( data_out_allA	),
			.y		( data_outA		)
		);

		orgate #(
        	.OR_WIDTH	(RAMBLOCKS	),
        	.BUS_WIDTH	(18		)
        	)
        dmem_orB (
        	.a		( data_out_allB	),
        	.y		( data_outB		)
        );
	end
endgenerate


genvar i;
generate
for(i = 0; i < RAMBLOCKS; i = i+1)
begin : MEM_BL

	if (ADR_LEN < 12)
	begin

		assign access_bramA = 1;
		assign access_bramB = 1;

	end else begin

		pselect        #(
			.ADDR_WIDTH	(BLOCK_ADR_LEN		),
			.BASE_WIDTH	(BLOCK_ADR_LEN		),
			.BASE_ADDR	(i		)
			)
		memselA	(
			.addr		(mem_addressA[ADR_LEN-1:11]	),
			.activ_peri	(1'b1	),
			.select		(access_bramA[i]	)
			);

		pselect        #(
			.ADDR_WIDTH	(BLOCK_ADR_LEN		),
			.BASE_WIDTH	(BLOCK_ADR_LEN		),
			.BASE_ADDR	(i		)
			)
		memselB	(
			.addr		(mem_addressB[ADR_LEN-1:11]	),
			.activ_peri	(1'b1	),
			.select		(access_bramB[i]	)
			);

	end

ram_9_prim  #(
		.WRITE_MODE_A(	"READ_FIRST"	),
		.INIT_A(	9'h000				),
		.SRVAL_A(	9'h000				),
		.WRITE_MODE_B(	"READ_FIRST"	),
		.INIT_B(	9'h000				),
		.SRVAL_B(	9'h000				)
		)
DPRAMA		(
		.DIA(		data_inA[8:0]					),
		.ADDRA(		mem_addressA[10:0]				),
		.ENA(		1'b1							),
		.WEA(		wr_low	& access_bramA[i]		),
		.SSRA(		!access_bramA[i]				),
		.CLKA(		clk								),
		.DOA(		data_out_allA[18*i+8:18*i+0]	),


		.DIB(		9'b0							),
		.ADDRB(		mem_addressB[10:0]				),
		.ENB(		1'b1							),
		.WEB(		1'b0							),
		.SSRB(		!access_bramB[i]				),
		.CLKB(		clk								),
		.DOB(		data_out_allB[18*i+8:18*i+0]	)
		);

ram_9_prim  #(
		.WRITE_MODE_A(	"READ_FIRST"	),
		.INIT_A(	9'h000				),
		.SRVAL_A(	9'h000				),
		.WRITE_MODE_B(	"READ_FIRST"	),
		.INIT_B(	9'h000				),
		.SRVAL_B(	9'h000				)
		)
DPRAMB		(
		.DIA(		data_inA[17:9]					),
		.ADDRA(		mem_addressA[10:0]				),
		.ENA(		1'b1							),
		.WEA(		wr_high	& access_bramA[i]		),
		.SSRA(		!access_bramA[i]				),
		.CLKA(		clk								),
		.DOA(		data_out_allA[18*i+17:18*i+9]	),


		.DIB(		9'b0							),
		.ADDRB(		mem_addressB[10:0]				),
		.ENB(		1'b1							),
		.WEB(		1'b0							),
		.SSRB(		!access_bramB[i]				),
		.CLKB(		clk								),
		.DOB(		data_out_allB[18*i+17:18*i+9]	)
		);
// ****************************************************************
end
endgenerate

endmodule
