/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`default_nettype none
`timescale 1ns / 1ps

module one_word_read_cache #(
	parameter CONNECTIONS 	= 1,
	parameter ADDR_LEN	= 1
) (

	input wire							clk,
	input wire							reset,

    input wire		[ADDR_LEN*CONNECTIONS-1:0]			addr,
	input wire		[18*CONNECTIONS-1:0]				data_in,
	input wire		[CONNECTIONS-1:0]					wr_low,

	input wire		[CONNECTIONS-1:0]					wr_high,
	input wire		[CONNECTIONS-1:0]					enable_input,
	output wire		[18*CONNECTIONS-1:0]				data_out,
	output reg		[CONNECTIONS-1:0]					valid

);

wire 	[ADDR_LEN-1:0]					addr_array					[CONNECTIONS-1:0];
reg 	[CONNECTIONS-1:0] 				data_valid;
reg 	[ADDR_LEN-1:0]					cache_addr 					[CONNECTIONS-1:0];
reg 	[17:0] 							cache_data 					[CONNECTIONS-1:0];
wire	[CONNECTIONS*CONNECTIONS-1:0]	other_write_on_my_addr_wide;
wire	[CONNECTIONS-1:0]				other_write_on_my_addr;
wire 	[17:0]							data_in_array				[CONNECTIONS-1:0];
wire	[CONNECTIONS-1:0]				overwrite_low;
wire	[CONNECTIONS-1:0]				overwrite_high;
wire	[CONNECTIONS-1:0]				write;
wire	[CONNECTIONS-1:0]				valid_wire;
wire	[CONNECTIONS-1:0]				match;

genvar i;
genvar j;
generate

	for(i=0;i<CONNECTIONS;i=i+1) begin:any_other_write_on_addr
		for(j=0;j<CONNECTIONS;j=j+1) begin:double_loop
			assign other_write_on_my_addr_wide[i*CONNECTIONS+j] = (i==j) ? 1'b0 : enable_input[j] & data_valid[i] & (addr_array[j] == cache_addr[i]) & (wr_high[j] | wr_low[j]);
		end
	end


	for(i=0;i<CONNECTIONS;i=i+1) begin:any_other_write_on_addr_combine
		assign other_write_on_my_addr[i] = !(other_write_on_my_addr_wide[i*CONNECTIONS+CONNECTIONS-1:i*CONNECTIONS]=={CONNECTIONS{1'b0}});
	end


	for(i=0;i<CONNECTIONS;i=i+1) begin:data_set

		assign addr_array[i] = addr[i*ADDR_LEN+ADDR_LEN-1:i*ADDR_LEN];
		assign data_out[18*i+17:18*i] = cache_data[i];
		assign data_in_array[i] = data_in[18*i+17:18*i];
		assign overwrite_low[i] = enable_input[i] & wr_low[i] & data_valid[i] & (cache_addr[i] == addr_array[i]);
		assign overwrite_high[i] = enable_input[i] & wr_high[i] & data_valid[i] & (cache_addr[i] == addr_array[i]);
		assign write[i] = wr_low[i] | wr_high[i];

		assign match[i] = data_valid[i] & (addr_array[i] == cache_addr[i]);
		assign valid_wire[i] = ~other_write_on_my_addr[i] & (match[i] | (enable_input[i] & (~write[i] | (write[i] & match[i]))));

		always@(posedge clk) begin
			if(reset) begin				
				cache_data[i] <= 1'b0;
				cache_addr[i] <= {ADDR_LEN{1'b0}};
				data_valid[i] <= 1'b0;
			end else if(other_write_on_my_addr[i]) begin
				data_valid[i] <= 1'b0;
			end else if(overwrite_low[i]) begin
				cache_data[i][8:0] <= data_in[i*18+8:i*18];
			end else if (overwrite_high[i]) begin
                cache_data[i][17:9] <= data_in[i*18+17:i*18+9];
			end else if(enable_input[i] && !write[i]) begin
				cache_data[i] <= data_in_array[i];
				cache_addr[i] <= addr_array[i];
				data_valid[i] <= 1'b1;
			end
		end


		/*always@(posedge clk) begin
			if(reset) begin				
				data_out[18*i+17:18*i] <= {CONNECTIONS{18'h0}};
			end else begin
				data_out[18*i+17:18*i] <= cache_data[i];
			end
		end*/

		always@(posedge clk) begin
			if(reset) begin				
				valid[i] <= {CONNECTIONS{1'b0}};
			end else begin
				valid[i] <= valid_wire[i];
			end
		end
	end
endgenerate
endmodule
