/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`timescale 1ns / 10ps

module spartanmc_mem_acc_dma #(
	parameter BASE_ADDR = 0,
	parameter CODE_RAMBLOCKS = 4,
	parameter DATA_RAMBLOCKS = 4,
	parameter SHOW_MEM = 1
) (
	// Code memory bus
	input  wire        clk,
	input  wire        reset,
	input  wire        code_access,
	input  wire        code_jmp,
	input  wire [17:0] code_addr,
	output wire [17:0] code_do,
	output wire        code_busy,

	// Memory bus
	input  wire        mem_access,
	input  wire        mem_wr_low,
	input  wire        mem_wr_high,
	input  wire [9:0]  mem_addr_block,
	input  wire [7:0]  mem_addr_high,
	input  wire [17:0] mem_di,
	output wire [17:0] mem_do,
	output wire        mem_busy,
	output wire        mem_busy_wr,

	// Cache Memory bus
	input  wire        cache_clk,
	input  wire        cache_reset,
	input  wire        cache_access,
	input  wire        cache_wr_low,
	input  wire        cache_wr_high,
	input  wire [9:0]  cache_addr_block,
	input  wire [7:0]  cache_addr_high,
	input  wire [17:0] cache_di,
	output wire [17:0] cache_do
);

wire [18*DATA_RAMBLOCKS-1:0] mem_do_all, cache_do_all;
wire [DATA_RAMBLOCKS-1:0] access_bram, access_cache;

assign mem_busy = 1'b0; // BlockRAM can always deliver in next clock cycle
assign mem_busy_wr = 1'b0; // BlockRAM can always deliver in next clock cycle

// local memory for code sections
spartanmc_mem_local #(
	.BASE_ADDR(BASE_ADDR),
	.RAMBLOCKS(CODE_RAMBLOCKS),
	.SHOW_MEM(0)
) code_mem (
	.clk(clk),
	.reset(reset),
	.code_access(code_access),
	.code_jmp(code_jmp),
	.code_addr(code_addr),
	.code_do(code_do),
	.code_busy(code_busy),

// do not use data port.
	.mem_addr_block(10'b0),
	.mem_addr_high(8'b0),
	.mem_di(18'b0),
	.mem_wr_low(1'b0),
	.mem_wr_high(1'b0),
	.mem_access(1'b0),
	.mem_do(),
	.mem_busy(),
	.mem_busy_wr()
);

// OR data output from all blockrams
orgate #(
	.OR_WIDTH (DATA_RAMBLOCKS),
	.BUS_WIDTH (18)
) dmem_or (
	.a ({mem_do_all}),
	.y (mem_do)
);

orgate #(
	.OR_WIDTH (DATA_RAMBLOCKS),
	.BUS_WIDTH (18)
) dcache_or (
	.a ({cache_do_all}),
	.y (cache_do)
);

genvar i;
generate
// Instantiate blockrams for data memory
for (i=0; i<DATA_RAMBLOCKS; i=i+1) begin: MEM_BL
	// Decode address
	// SpartanMC data port
	pselect #(
		.ADDR_WIDTH(7),
		.BASE_WIDTH(7),
		.BASE_ADDR(BASE_ADDR[17:11] + CODE_RAMBLOCKS + i)
	) memsel (
		.addr(mem_addr_high[7:1]),
		.activ_peri(mem_access),
		.select(access_bram[i])
	);

	// Accelerator data port
	pselect #(
		.ADDR_WIDTH(7),
		.BASE_WIDTH(7),
		.BASE_ADDR(BASE_ADDR[17:11] + CODE_RAMBLOCKS + i)
	) cachesel (
		.addr(cache_addr_high[7:1]),
		.activ_peri(cache_access),
		.select(access_cache[i])
	);

	// use legacy ram 2*9 bit to be compatible with initramj (from spartanmc_mem_local)
	// Port A is used for accelerator (instead of instruction fetch)
	// Port B used for spartanmc
	ram_9_prim  #(
		.WRITE_MODE_A(  "READ_FIRST"                     ),
		.INIT_A(        9'h000                          ),
		.SRVAL_A(       9'h000                          ),
		.WRITE_MODE_B(  "READ_FIRST"                     ),
		.INIT_B(        9'h000                          ),
		.SRVAL_B(       9'h000                          )
	) DPRAMA (
		.DIA(           cache_di[8:0]                   ),
		.ADDRA(         {cache_addr_high[0],cache_addr_block[9:0]}),
		.ENA(           1'b1                            ),
		.WEA(           cache_wr_low & access_cache[i]  ),
		.SSRA(          ! access_cache[i]               ),
		.CLKA(          cache_clk                       ),
		.DOA(           cache_do_all[18*i+8:18*i+0]     ),
		.DIB(           mem_di[8:0]                     ),
		.ADDRB(         {mem_addr_high[0],mem_addr_block[9:0]}),
		.ENB(           1'b1                            ),
		.WEB(           mem_wr_low & access_bram[i]     ),
		.SSRB(          ! access_bram[i]                ),
		.CLKB(          clk                             ),
		.DOB(           mem_do_all[18*i+8:18*i+0]       )
	);

	ram_9_prim  #(
		.WRITE_MODE_A(  "READ_FIRST"                     ),
		.INIT_A(        9'h000                          ),
		.SRVAL_A(       9'h000                          ),
		.WRITE_MODE_B(  "READ_FIRST"                     ),
		.INIT_B(        9'h000                          ),
		.SRVAL_B(       9'h000                          )
	) DPRAMB (
		.DIA(           cache_di[17:9]                  ),
		.ADDRA(         {cache_addr_high[0],cache_addr_block[9:0]}),
		.ENA(           1'b1                            ),
		.WEA(           cache_wr_high & access_cache[i] ),
		.SSRA(          ! access_cache[i]               ),
		.CLKA(          cache_clk                       ),
		.DOA(           cache_do_all[18*i+17:18*i+9]    ),
		.DIB(           mem_di[17:9]                    ),
		.ADDRB(         {mem_addr_high[0],mem_addr_block[9:0]}),
		.ENB(           1'b1                            ),
		.WEB(           mem_wr_high & access_bram[i]    ),
		.SSRB(          ! access_bram[i]                ),
		.CLKB(          clk                             ),
		.DOB(           mem_do_all[18*i+17:18*i+9]      )
	);
end

// Simulator Outputs
if (SHOW_MEM  == 1)
	always @(posedge clk)
		if (mem_access)
		begin
			$write("[ %9d ]  ", $time);
			if (mem_wr_low | mem_wr_high)
				$display("WR  mem[0x%05H] <= 0x%05H", {mem_addr_high, mem_addr_block}, mem_di);
			else
				$display("RD  mem[0x%05H] => Reg", {mem_addr_high, mem_addr_block});
end
endgenerate

endmodule
