/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`include	"core/src/spartanmc_def.vh"
`default_nettype none

module spartanmc_mem_global #(
	parameter RAMBLOCKS_GLOBAL	= 4,
	parameter LOCAL_BASE_ADDR	= RAMBLOCKS_GLOBAL*2048,
	parameter RAMBLOCKS_LOCAL	= 4,
	parameter CORES_COUNT		= 4,
	parameter GLOBAL_COUNT		= CORES_COUNT,
	parameter LOCAL_COUNT		= CORES_COUNT,
	parameter CACHE_WAYS		= 1,
	parameter INDEX_LENGTH		= 9,
	parameter OFFSET_LENGTH		= 1,
	parameter USE_TWO_PORTS_GLOBAL	= 0,
	parameter SHOW_MEM_ALL		= 0,
	parameter SHOW_MEM_SINGLE	= 1
) (
	input wire								clk,
	input wire								reset,

	input wire		[GLOBAL_COUNT-1:0]		code_access,
	input wire		[GLOBAL_COUNT-1:0]		code_jmp,
	input wire		[GLOBAL_COUNT*18-1:0]	code_addr,
	output wire		[GLOBAL_COUNT*18-1:0]	code_do,
	output wire		[GLOBAL_COUNT-1:0]		code_busy,

	input wire		[LOCAL_COUNT*10-1:0]	mem_addr_block,
	input wire		[LOCAL_COUNT*8-1:0]		mem_addr_high,
	input wire		[LOCAL_COUNT*18-1:0]	mem_di,
	input wire		[LOCAL_COUNT-1:0]		mem_wr_low,
	input wire		[LOCAL_COUNT-1:0]		mem_wr_high,
	input wire		[LOCAL_COUNT-1:0]		mem_access,
	output wire		[LOCAL_COUNT*18-1:0]	mem_do,
	output wire		[LOCAL_COUNT-1:0]		mem_busy
);

// local memories for data

genvar i;

generate

for (i = 0; i < LOCAL_COUNT; i = i + 1) begin : LOCAL_MEM
	spartanmc_mem_local #(
		.BASE_ADDR( LOCAL_BASE_ADDR ),
		.RAMBLOCKS( RAMBLOCKS_LOCAL ),
		.SHOW_MEM( SHOW_MEM_ALL || SHOW_MEM_SINGLE && (i == 0) )
	) data_mem (
		.clk( clk ),
		.reset( reset ),
		.code_access( 1'b0 ),
		.code_jmp( 1'b0 ),
		.code_addr( 18'b0 ),
		.code_do( ),
		.code_busy( ),

		.mem_addr_block( mem_addr_block[(i+1)*10-1:i*10] ),
		.mem_addr_high( mem_addr_high[(i+1)*8-1:i*8] ),
		.mem_di( mem_di[(i+1)*18-1:i*18] ),
		.mem_wr_low( mem_wr_low[i] ),
		.mem_wr_high( mem_wr_high[i] ),
		.mem_access( mem_access[i] ),
		.mem_do( mem_do[(i+1)*18-1:i*18] ),
		.mem_busy( mem_busy[i] )
	);
end

endgenerate


// global firmware memory with caches

localparam	GLOBAL_MEM_WIDTH	= (1<<OFFSET_LENGTH) * 18;

wire	[GLOBAL_COUNT-1:0]		request_vect;
wire	[GLOBAL_COUNT-1:0]		grant_vect;
wire	[GLOBAL_COUNT*18-1:0]	fw_mem_addr_in;
wire	[GLOBAL_COUNT*GLOBAL_MEM_WIDTH-1:0]	fw_mem_do;

generate

for (i = 0; i < GLOBAL_COUNT; i = i + 1) begin : CACHE
	reg		mem_data_valid = 1'b0;

	always @(posedge clk)
		mem_data_valid <= grant_vect[i];

	wire valid_internal;

	InstructionCache #(
		.p_mem_address_width( 18 ),
		.p_word_width( 18 ),
		.p_ways( CACHE_WAYS ),
		.p_index_len( INDEX_LENGTH ),
		.p_blockoffset_len( OFFSET_LENGTH ),
		.p_direct_mem_output( 1 )
	) cache (
		.clk_i( clk ),
		.rst_i( reset ),
		.access_i( code_access[i] ),
		.address_i( code_addr[(i+1)*18-1:i*18] ),
		.jmp_i( code_jmp[i] ),
		.cache_data_o( code_do[(i+1)*18-1:i*18] ),
		.request( request_vect[i] ),
		.valid_o( valid_internal ),
		.load_addr( fw_mem_addr_in[(i+1)*18-1:i*18] ),
		.mem_acknowledge( grant_vect[i] ),
		.mem_data_valid( mem_data_valid ),
		.mem_data_in( fw_mem_do[ (i+1)*GLOBAL_MEM_WIDTH-1 : i*GLOBAL_MEM_WIDTH ] )
	);

	assign code_busy[i] = ~ valid_internal;
	
end

endgenerate

firmware_mem #(
	.RAMBLOCKS( RAMBLOCKS_GLOBAL ),
	.DATA_WIDTH( GLOBAL_MEM_WIDTH ),
	.CORES( GLOBAL_COUNT ),
	.USE_TWO_PORTS( USE_TWO_PORTS_GLOBAL )
) global_fw_mem (
	.clk( clk ),
	.addrin( fw_mem_addr_in ),
	.iout( fw_mem_do ),
	.request( request_vect ),
	.grant( grant_vect )
);

endmodule
