/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// TU-Dresden Chair for Embedded Systems
//////////////////////////////////////////////////////////////////////////////////

//Geschrieben von: Markus Noll, mail@markus-noll.de, TU-Darmstadt 2015
module dispatcher_master
#(
	parameter BASE_ADR         = 10'h0,
	parameter BUFFER_SIZE      = 'd8,
	parameter DATA_WIDTH       = 'd18,
 	parameter NUMBER_OF_SLAVES = 'd3

)
(
	//spmc interface
	input   wire         clk_peri,
	input   wire [17:0]  do_peri,
	output  wire [17:0]  di_peri,
	input   wire [9:0]   addr_peri,
	input   wire         access_peri,
	input   wire         wr_peri,
	input   wire         reset,
	
	//interface to dispatcher slaves
	output wire[17:0]	data_out,
	output reg [5:0]	receiver_id_out,
	output reg [11:0]	msg_size_out,
	output wire			status_out,

	input wire [((NUMBER_OF_SLAVES*12)-1):0]	msg_size_in,
	input wire [NUMBER_OF_SLAVES-1:0]			read,
	input wire [NUMBER_OF_SLAVES-1:0] 			received
);

/////////////////////////////////////////////////////////////////////////////////////
//
// Local Functions and Parameters
//

function integer log2;
input integer value;
	for (log2=0; value>0; log2=log2+1)
	value = value>>1;
endfunction

localparam
    P_STATUS				= 0,
    P_MSG_SIZE	   			= 1,
    P_DATA_IN       		= 2,
    
    ALL_PORTS		   		= 3,
    DECODSIM_DEB			= "NO",
    FIFO_ADDR_BITS 	  		= log2(BUFFER_SIZE - 1),
    FIFO_ADDR_BITS_COUNT 	= FIFO_ADDR_BITS+1;

//event wires
wire					select;
wire	[ALL_PORTS-1:0]	reg_select;
wire	[ALL_PORTS-1:0]	reg_read;
wire	[ALL_PORTS-1:0]	reg_write;

wire	status_read_sel 	= reg_read[P_STATUS];
wire    msg_size_read_sel 	= reg_read[P_MSG_SIZE];
wire    data_in_read_sel 	= reg_read[P_DATA_IN];

wire	status_write_sel	= reg_write[P_STATUS];
wire    msg_size_write_sel 	= reg_write[P_MSG_SIZE];
wire    data_in_write_sel 	= reg_write[P_DATA_IN];

reg_access_decoder	
#(
	.REG_COUNT(ALL_PORTS),
	.BASE_ADR(BASE_ADR),
	.SIM_DEBUG(DECODSIM_DEB)
)
decod1			
(
	.reset(reset),
	.clk_peri(clk_peri),
	.addr_peri(addr_peri),
	.access_peri(access_peri),
	.wr_peri(wr_peri),
	.peri_select(select),
	.reg_select(reg_select),
	.reg_read(reg_read),
	.reg_write(reg_write)
);

//slave IDs from aribters
wire[5:0] slave_id_lb;
wire[5:0] slave_id_rr;

//fifo signals
wire [FIFO_ADDR_BITS:0] fifo_fill;

wire [17:0] data_to_buffer;
wire [17:0] data_out_fifo;

assign data_out = data_out_fifo;

wire read_from_fifo;
wire buffer_write;

//count and status registers
reg [FIFO_ADDR_BITS_COUNT*NUMBER_OF_SLAVES-1:0] count;
reg slave_is_reading;
reg header_readable;
reg [17:0] package_counter;
reg [17:0] msg_in_counter;
reg [5:0] writing_receiver_id;

//status and msg_size peripheral register logic
reg [17:0] message_size;

always @(posedge clk_peri)
begin
    if (reset)
    begin
        message_size <= 18'h1;
    end else if (msg_size_write_sel)
    begin
        message_size <= do_peri;
    end
end

wire fifo_full = (18'h00000 | (BUFFER_SIZE - fifo_fill)) < message_size;

//status of master to spmc
assign di_peri[17:0] = status_read_sel ? (fifo_full ? 18'd0 : 18'd1) : 18'd0;

//status of master to slave (amount of data specified by msg_size_in available)
assign status_out = (fifo_fill + 1 >= msg_size_in[(receiver_id_out*12) +: 12]) && slave_is_reading;


//write and read access to buffer
wire read_header;
wire read_data;
wire last_value_write_sel;

//read events
assign read_header 		= ((fifo_fill>0) && !slave_is_reading && header_readable/* package_counter > 0*/);
assign read_data 		= ((fifo_fill>0) && slave_is_reading && read > 0);
assign read_from_fifo	= (read_header | read_data);

//write events
assign buffer_write = data_in_write_sel;
//special care must be taken in case of empty payload, then the header is the last_value to write
assign last_value_write_sel = (data_in_write_sel && (msg_in_counter == 1 || data_to_buffer[17:6] == 0 && msg_in_counter == 0));

//multiplexor to buffer
assign data_to_buffer = (data_in_write_sel && !msg_in_counter && do_peri[5:0] <  62) ? {do_peri} :						//direct id assignment
						(data_in_write_sel && !msg_in_counter && do_peri[5:0] == 62) ? {do_peri[17:6],slave_id_rr} :	//round robin arbitration
						(data_in_write_sel && !msg_in_counter && do_peri[5:0] == 63) ? {do_peri[17:6],slave_id_lb} :	//load balancing arbitration
						(data_in_write_sel) ? {do_peri} : 18'd0;			

//buffer data_to_buffer and buffer_write to obtain speed above 60MHz
reg [17:0] data_to_buffer_x;
reg buffer_write_x;

always @ (posedge clk_peri) begin
	if (reset) begin
		data_to_buffer_x 	<= 18'd0;
		buffer_write_x 		<= 1'b0;
	end else begin
		data_to_buffer_x 	<= data_to_buffer;
		buffer_write_x 		<= buffer_write;
	end
end

//main logic
always@(posedge clk_peri) begin

	if (reset) begin
		msg_in_counter 			<= 18'd0;
		writing_receiver_id		<= 6'd0;
		slave_is_reading 		<= 1'b0;
		receiver_id_out			<= 6'b0;
		msg_size_out 			<= 12'd0;
		header_readable 		<= 1'b0;
	end

	else begin	
		//----------------------------write access			
		//write header
		if (data_in_write_sel && !msg_in_counter) begin
			msg_in_counter <= data_to_buffer[17:6];
			writing_receiver_id <= data_to_buffer[5:0];
			//header not really readable yet, but read_header only high when fifo_fill > 0;
			header_readable <= 1'b1;
		//write data
		end else if (data_in_write_sel && msg_in_counter > 0) begin
			msg_in_counter <= msg_in_counter - 1;
		end

		//----------------------------read access
		//read header
		if (read_header) begin
			slave_is_reading <= 1'b1;
			header_readable <= 1'b0;
			msg_size_out <= data_out_fifo[17:6];
			receiver_id_out <= data_out_fifo[5:0];
		//read data
		end else if (read_data) begin
		//read finished
		end else if (received > 0) begin
			slave_is_reading <= 1'b0;
			if(package_counter > 1) begin
				header_readable <= 1'b1;
			end
		end
	end
end

//package counter (keeps track of number of packages in buffer)
always @ (posedge clk_peri) begin

	if (reset) begin
		package_counter <= 18'd0;

	end else begin
		if (last_value_write_sel && received == 0) begin
			package_counter <= package_counter + 1;

		end else if (!last_value_write_sel && received > 0) begin
			package_counter <= package_counter - 1;

		end else begin
			package_counter <= package_counter;
		end
	end
end

//count register for load balancing logic 
always @ (posedge clk_peri) begin

	if (reset) begin
		count <= 1000'd0;	

	end else begin
		//write access
		if (data_in_write_sel & !read_data) begin
			//increase written count register
			count[writing_receiver_id*FIFO_ADDR_BITS_COUNT +: FIFO_ADDR_BITS_COUNT] <= count[writing_receiver_id*FIFO_ADDR_BITS_COUNT +: FIFO_ADDR_BITS_COUNT] + 1;

		//read access
		end else if (read_data & !data_in_write_sel) begin			
			//decrease read count register
			count[receiver_id_out*FIFO_ADDR_BITS_COUNT +: FIFO_ADDR_BITS_COUNT] <= count[receiver_id_out*FIFO_ADDR_BITS_COUNT +: FIFO_ADDR_BITS_COUNT] -1;

		//simultaneous read and write access
		end else if (read_data & data_in_write_sel) begin
			//increase written count register
			count[writing_receiver_id*FIFO_ADDR_BITS_COUNT +: FIFO_ADDR_BITS_COUNT] <= count[writing_receiver_id*FIFO_ADDR_BITS_COUNT +: FIFO_ADDR_BITS_COUNT] + 1;
			//deacrease read count register
			count[receiver_id_out*FIFO_ADDR_BITS_COUNT +: FIFO_ADDR_BITS_COUNT] <= count[receiver_id_out*FIFO_ADDR_BITS_COUNT +: FIFO_ADDR_BITS_COUNT] -1;
		end
	end
end


//load balancing arbiter
load_balancing_arbiter #
(
   	.NUMBER_OF_SLAVES(NUMBER_OF_SLAVES),
   	.FIFO_ADDR_BITS(FIFO_ADDR_BITS_COUNT)
)
lba
(
   	.data_in(count),
   	.slave_id(slave_id_lb)   
);

//round robin arbiter
rr_arbiter #
(
	.NUMBER_OF_SLAVES(NUMBER_OF_SLAVES)
)
rra
(
	.reset(reset),
	.clk(clk_peri),
	.last_value_write_sel(last_value_write_sel),
	.slave_id(slave_id_rr)
);
	

///////////////////////////////////////////////////////////////////
// Fifo logic module
// Using async fifo module with the same clock on both ports

xilinx_fifo_with_fill_counter #
(
 .WIDTH(18),
 .DEPTH(BUFFER_SIZE),
 .ADDR_BITS(FIFO_ADDR_BITS)
) 
	fifo0(
.clk(clk_peri),
.fill_count(fifo_fill),
.reset(reset),
.write_enable(buffer_write_x),
.write_data(data_to_buffer_x),
.read_enable(read_from_fifo), 
.read_data(data_out_fifo)
);

endmodule
