/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Seidowski
//
// Create Date:    10:46:23 10/21/2008
// Module Name:    display
// Project Name:   Display-Controller fuer SpartanMC
//
// Description:    Top-Modul zur Ansteuerung eines pixelbasierten Displays
//////////////////////////////////////////////////////////////////////////////////

`define SYSTEM_CLOCK		25_000_000		// Takt des Systems
`define IO_OFFSET_WIDTH		3			// Breite des "privaten" Adressbereichs eines Peripheriemoduls (idR. die untersten drei Bit)

// Peripherieregister
`define REG_DISPLAYSTATUS	`IO_OFFSET_WIDTH'd0	// Statusregister (siehe unten)
`define REG_TEXT_COLOR		`IO_OFFSET_WIDTH'd1	// Textfarbe
`define REG_TEXT_BGCOLOR	`IO_OFFSET_WIDTH'd2	// Texthintergrundfarbe
`define REG_TEXT_CHARPOS	`IO_OFFSET_WIDTH'd3	// Position des aktuell gezeichneten Textzeichens (fuer CharLineInterrupt)
`define REG_TEXT_CURSORPOS	`IO_OFFSET_WIDTH'd4	// Position des Cursors (Bit 17...9 fuer X, 8...0 fuer Y)
`define REG_GRAPH_COORDSELECT	`IO_OFFSET_WIDTH'd5	// Auswahl der Koordinate fuer die Grafikfunktionen (0 = X0 / 1 = Y0; 2 = X1 / 3 = Y1; usw.)
`define REG_GRAPH_COORDVALUE	`IO_OFFSET_WIDTH'd6	// Wert der ausgewaehlten Koordinate
`define REG_GRAPH_COLOR		`IO_OFFSET_WIDTH'd7	// Farbe fuer die Grafikfunktion

// Bits im Statusregister
`define STATUSBIT_DISP_BACKLIGHT 0	// Hintergrundbeleuchtung
`define STATUSBIT_DISP_ON	1	// Display-Logik
`define STATUSBIT_TEXTMODE	2	// Textmodus
`define STATUSBIT_FUNCTIONSELECT 5:3	// Hardware-Grafikfunktionsauswahl
`define STATUSBIT_FUNCTIONFLAG1	6	// Flag fuer die ausgewaehlte Funktion (derzeit nur fuer CopyRect genutzt)
`define STATUSBIT_FUNCTIONDRAW	7	// Ruft die Zeichenfunktion mit den gesetzen Parametern auf
`define STATUSBIT_FUNCTIONREADY	8	// Gibt an, ob das Modul fuer die Grafikfuntionen bereit ist (bzw. der Vorgang fertiggestellt wurde)
`define STATUSBIT_OVERLAY	9	// (De-)Aktiviert den Overlay-Modus. Bei Aktivierung wird die im Farbregister angegebene Farbe als transparente Farbe uebernommen

// Anzahl an erforderlichen Koordinaten fuer die Grafikfunktionen (abhaengig von den maximal noetigen fuer eine Funktion)
`define GRAPH_FUNCTION_MAXCOORDINATES 3

// Standardwerte fuer LCD Winstar WC320240C-TCI-N
/*
`define SCREEN_WIDTH		320
`define SCREEN_HEIGHT		240
`define DATA_WORD_WIDTH		8
`define BIT_PER_PIXEL		3
`define CL2_CYCLE		8
`define ROW_BEGIN_DELAY		CL2_CYCLE
`define ROW_END_DELAY		CL2_CYCLE
`define DATA_WORDS_PER_ROW	((SCREEN_WIDTH*BIT_PER_PIXEL) / DATA_WORD_WIDTH)
`define ROW_CYCLE		(DATA_WORDS_PER_ROW*CL2_CYCLE + ROW_BREAK_DELAY)
`define SCREEN_CYCLE		(ROW_CYCLE*SCREEN_HEIGHT + SCREEN_FINISH_DELAY)
*/

// Speichereinstellungen
`define BIT_PER_MEMORY_PIXEL	2	// Anzahl an Bits, die ein Pixel im Speicher benötigt
`define GRAPHMEM_DATA_WIDTH	9	// Breite eines Datenworts im Grafikspeicher
`define TEXTMEM_DATA_WIDTH	9	// Breite eines Datenworts im Textspeicher
`define CODEPAGE_DATA_WIDTH	9	// Breite eines Datenworts in der Codepage

// Frequenzeinstelungen fuer den Textmodus
`define TEXT_BLINK_FREQUENCY	2	// Frequenz in Hz von blinkendem Text
`define CURSOR_BLINK_FREQUENCY	4	// Frequenz in Hz des Cursors

`include "display/src/display_graph_function_common.vh"	// Enthaelt Standardwerte fuer die Grafikfunktionen

module display #(
		parameter BASE_ADR			= 18'h00010,
		parameter GRAPHMEM_BASE_ADDR		= 18'h08000,	// Der Grafikspeicher sollte auf jeden Fall hinter den Bereich fuer IO-Geraete gelegt werden,
									// damit dieser nicht ungewollt ueber die IO-Adressen angesteuert wird
		parameter TEXTMEM_BASE_ADDR		= 18'h1b000,	// Die uebrigen Speicher konnen auch vor die IO-Geraete gelegt werden
		parameter CODEPAGE_BASE_ADDR		= 18'h1c000,	// sofern ihre interne Adressbreite klein genug ist.
		parameter COLOR_LUT_BASE_ADDR		= 18'h1d000,
		parameter CURSORMEM_BASE_ADDR		= 18'h1e000,

		parameter SCREEN_WIDTH			= 320,		// Bildschirmaufloesung: Breite
		parameter SCREEN_HEIGHT			= 240,		// Bildschirmaufloesung: Hoehe
		parameter BIT_PER_PIXEL			= 3,		// Farbtiefe des Displays
		parameter BIT_PER_MEMORY_PIXEL		= 2,		// Farbtiefe eines Pixels im Speicher
		parameter DATA_WORD_WIDTH		= 8,		// Groesse eines Datenworts fuer das Display

		parameter TIMING_INIT_CLOCKS_TO_VCON_ON	= 1,		// Initialisierung: Verzoegerung der Aktivierung der Kontrastspannung
		parameter TIMING_INIT_CLOCKS_TO_DISPOFF	= 1,		// Initialisierung: Verzoegerung der Aktivierung der Display-Logik

		/* Zeilentakt */
		parameter TIMING_CL1_CLOCKS_AFTER_RESET	= 0,		// Low-Zeit nach einem Reset
		parameter TIMING_CL1_CLOCKS_HIGH	= 0,		// Dauer der High-Phase
		parameter TIMING_CL1_CLOCKS_LOW		= 0,		// Dauer der Low-Phase
		parameter TIMING_CL1_BEGIN_HIGH_LOW	= 1,		// Zeilenbeginn bei High->Low (1) oder bei Low->High (0) von CL1
		parameter TIMING_SCREEN_FINISH_DELAY	= 0,		// Pause nach Abschluss der Uebertragung eines kompletten Bildes

		/* Datentakt */
		parameter TIMING_CL2_CLOCKS_AFTER_RESET	= 0,		// Low-Zeit nach einem Reset
		parameter TIMING_CL2_CLOCKS_HIGH	= 0,		// Dauer der High-Phase
		parameter TIMING_CL2_CLOCKS_LOW		= 0,		// Dauer der Low-Phase
		parameter TIMING_ROW_END_DELAY		= 0,		// Pause nach Ende der Uebertragung einer kompletten Zeile
		parameter TIMING_ROW_BEGIN_DELAY	= 0,		// Pause vor Beginn der Datenuebertragung einer Zeile

		/* Framestart-Signal */
		parameter TIMING_FRAMESTART_CLOCKS_AFTER_RESET = 0,	// Low-Zeit nach einem Reset
		parameter TIMING_FRAMESTART_CLOCKS_HIGH	= 0,		// Dauer der High-Phase
		parameter TIMING_FRAMESTART_CLOCKS_LOW	= 0,		// Dauer der Low-Phase

		/* Invertierungssignal (jedes zweite Bild) */
		parameter TIMING_INVERT_CLOCKS_AFTER_RESET = 0,		// Low-Zeit nach einem Reset
		parameter TIMING_INVERT_CLOCKS_TOGGLE	= 0,		// Zeit, nach der das Signal invertiert wird

		/* Codepage */
		parameter CODEPAGE_CHAR_WIDTH		= 8,		// Breite in Pixeln eines Textzeichens
		parameter CODEPAGE_CHAR_HEIGHT		= 16,		// Hoehe in Pixeln eines Textzeichens
		parameter CODEPAGE_SIZE			= 256,		// Anzahl aller Zeichen der Codepage

		parameter FPGA_BRAM_SIZE		= 2048,		// Groesse eines Block-RAMs auf dem FPGA

		parameter USE_TEXT_MODE			= 1,		// Textmodus synthetisieren?
		parameter USE_GRAPHIC_MODE		= 1,		// Grafikmodus synthetisieren?

		/* Optional verfuegbare Graifkfunktion */
		parameter USE_GRAPHFUNCTION_SETPIXEL	= 1,
		parameter USE_GRAPHFUNCTION_LINE	= 1,
		parameter USE_GRAPHFUNCTION_COPYRECT	= 1,
		parameter USE_GRAPHFUNCTION_GETPIXEL	= 1,
		parameter USE_GRAPHFUNCTION_FILLRECT	= 1,
		parameter ALL_PORTS			= 8,		// Anzahl der Ports fuer den Dekoder
		parameter DECODSIM_DEB			= "NO"		// DEBUG mode YES oder NO
	)
	(
		/* Spartan-MC-Signale */
		input wire		reset,

		input wire		clk_peri,
		input wire	[17:0]	do_peri,
		output wire	[17:0]	di_peri,
		input wire	[9:0]	addr_peri,
		input wire		access_peri,
		input wire		wr_peri,

		/* Block-RAM-Interface */
		input wire		mem_clk,
		input wire		mem_access,
		input wire		store_access,
		input wire	[7:0]	addr_high,

		/* Display-Signale */
		output	reg	display_framestart,
		output	reg	display_cl1,
		output	reg	display_cl2,
		output	reg	display_dispoff,
		output	reg	display_invert,
		output	reg [DATA_WORD_WIDTH-1:0] display_data,
		output	reg	display_vdd,
		output	reg	display_backlight,
		output	reg	display_vcon,

		/* Interrupts */
		output wire		charLine_ir,		// High, wenn eine Zeile eines Zeichens im Textmodus gezeichnet wurde
		output wire		graphFunctionReady_ir	// High, wenn eine Grafikfunktion abgeschlossen wurde
	);

	`include "display/src/ld.vh"
	function integer max(input integer v1, input integer v2);
		max = (v1 > v2) ? v1 : v2;
	endfunction


	localparam PIXEL_PER_MEMORY_WORD = `GRAPHMEM_DATA_WIDTH / BIT_PER_MEMORY_PIXEL;
	localparam CHARS_PER_ROW = SCREEN_WIDTH / CODEPAGE_CHAR_WIDTH;
	localparam CHARS_PER_COLUMN = SCREEN_HEIGHT / CODEPAGE_CHAR_HEIGHT;

	/* Berechnung der Adressbreiten der jeweiligen Speicher */
	localparam ADDR_WIDTH_GRAPHMEM = ld(SCREEN_WIDTH*SCREEN_HEIGHT*BIT_PER_MEMORY_PIXEL/(PIXEL_PER_MEMORY_WORD*BIT_PER_MEMORY_PIXEL)-1);
	localparam ADDR_WIDTH_TEXTMEM  = ld(CHARS_PER_ROW*CHARS_PER_COLUMN-1);
	localparam ADDR_WIDTH_CODEPAGE = ld(CODEPAGE_CHAR_HEIGHT*CODEPAGE_SIZE-1);
	localparam ADDR_WIDTH_COLOR_LUT = BIT_PER_MEMORY_PIXEL;

	// Adressierungsbreite innerhalb eines DMA-Speichers. Diese gilt fuer alle Speicher mit Ausnahme
	// des Grafikspeichers. Es sollten alle Speicher die gleichen Bits als Offset und Adresse haben,
	// damit keine Mehrfachbelegung bestimmter Adressen auftreten kann. Daher wird hier die
	// Maximalbreite aller Speicher ermittelt, welche dann fuer die Adressdekodierung benutzt wird.
	localparam ADDR_WIDTH_DMA = max(ADDR_WIDTH_TEXTMEM, max(ADDR_WIDTH_CODEPAGE, ADDR_WIDTH_COLOR_LUT));

	wire module_select;
//	wire [`IO_OFFSET_WIDTH-1:0] io_offset = addr_peri[`IO_OFFSET_WIDTH-1:0];
	wire [17:0] addr_bus = { addr_high, addr_peri };

	reg pixelWait;	// Veranlasst das Text- bzw. das Grafikmodul dazu, bei 1 keine weiteren Pixeldaten zu senden


/* Konfigurationsregister */
	// Hier waere eine bedingte Deklaration bei (USE_TEXT_MODE == 1) sinnvoll.
		// Getrennte Register fuer Vorder- und Hintergrundfarbe, da bei vergroesserter Farbtiefe (BIT_PER_PIXEL) beide
		// zusammen evtl. nicht mehr mit einem Datenwort des SpartanMC uebertragbar sind.
		reg [BIT_PER_PIXEL-1:0] text_color;		// Farbe fuer auszugebende Zeichen
		reg [BIT_PER_PIXEL-1:0] text_bgColor;		// Hintergrundfarbe fuer auszugebende Zeichen
		// Cursor
		reg [8:0] char_cursorX;					// Position des Cursors
		reg [8:0] char_cursorY;

	// Grafikmodul
	reg [17:0] displayStatus;						// Statusregister des Displays
	// Hier waere eine bedingte Deklaration bei (USE_GRAPHIC_MODE == 1) sinnvoll.
		reg [BIT_PER_PIXEL-1:0] graphics_colorLut [0:2**BIT_PER_MEMORY_PIXEL-1]; // Color-Lookup-Table
		reg [17:0] graphics_function_color;					// Farbe, die an das Modul fuer die Grafikfunktionen uebergeben wird
		wire dispBacklight 	= displayStatus[`STATUSBIT_DISP_BACKLIGHT];	// Hintergrundbeleuchtung an/aus
		wire dispOn		= displayStatus[`STATUSBIT_DISP_ON];		// Display an/aus

		wire textMode		= (USE_GRAPHIC_MODE == 0) ? 1 :			// Wenn kein Grafikmodus vorhanden, immer 1
					(USE_TEXT_MODE == 0) ? 0:			// Wenn kein Textmodus vorhanden, immer 0
					displayStatus[`STATUSBIT_TEXTMODE];		// Textmodus an/aus

	// chipselect fuer die Status-, Daten- und Kommandoregister
		wire	[ALL_PORTS-1:0]	reg_select;	// r/w ADR
		wire	[ALL_PORTS-1:0]	reg_read;	// read ADR
		wire	[ALL_PORTS-1:0]	reg_write;	// write ADR
/* Zugriff auf Konfigurationsregister */
		wire wr_displayStatus	= reg_write[`REG_DISPLAYSTATUS];
		wire rd_displayStatus	= reg_read[`REG_DISPLAYSTATUS];

	// Hier waere eine bedingte Deklaration bei (USE_TEXT_MODE == 1) sinnvoll.
		wire wr_text_color	= reg_write[`REG_TEXT_COLOR];
		wire rd_text_color	= reg_read[`REG_TEXT_COLOR];

		wire wr_text_bgColor	= reg_write[`REG_TEXT_BGCOLOR];
		wire rd_text_bgColor	= reg_read[`REG_TEXT_BGCOLOR];

		wire rd_text_charpos	= reg_read[`REG_TEXT_CHARPOS];

		wire rd_text_cursorpos	= reg_read[`REG_TEXT_CURSORPOS];
		wire wr_text_cursorpos	= reg_write[`REG_TEXT_CURSORPOS];

	// Hier waere eine bedingte Deklaration bei (USE_GRAPHIC_MODE == 1) sinnvoll.
		wire wr_graph_coordAddr	= reg_write[`REG_GRAPH_COORDSELECT];
		wire rd_graph_coordAddr	= reg_read[`REG_GRAPH_COORDSELECT];

		wire wr_graph_coordValue = reg_write[`REG_GRAPH_COORDVALUE];
		wire rd_graph_coordValue = reg_read[`REG_GRAPH_COORDVALUE];

		wire wr_graph_color	= reg_write[`REG_GRAPH_COLOR];
		wire rd_graph_color	= reg_read[`REG_GRAPH_COLOR];

/* Puffer fuer empfangene Pixeldaten */
	reg receiveNewData;
	reg [1:0] dataReceiveState;
	reg [ld(BIT_PER_PIXEL*DATA_WORD_WIDTH)-1:0] dataReceiveCount;
	reg [BIT_PER_PIXEL*DATA_WORD_WIDTH-1:0] dataReceiveShiftReg;
	reg [BIT_PER_PIXEL-1:0] overlayColor;

/* Puffer fuer ausgehende Pixeldaten */
	reg [0:BIT_PER_PIXEL*DATA_WORD_WIDTH-1]	dataShiftReg;
	reg [ld(BIT_PER_PIXEL)-1:0] dataWordModCount;

/* Signalerzeugung */
	wire sig_framestart;
	wire sig_cl1;
	wire sig_cl2;
	wire sig_dispoff;
	wire sig_invert;
	wire sig_vdd;
	wire sig_vcon;

/* Textdarstellung */
	wire [ld((SCREEN_WIDTH/CODEPAGE_CHAR_WIDTH) * (SCREEN_HEIGHT/CODEPAGE_CHAR_HEIGHT))-1:0] text_memAddr;
									// Adressleitungen in den Textspeicher
	wire [`TEXTMEM_DATA_WIDTH-1:0] text_memData;			// Datenleitungen aus dem Textspeicher
	wire [ld(CODEPAGE_CHAR_HEIGHT*CODEPAGE_SIZE)-1:0] text_cpAddr;	// Adressleitungen in den Codepage-ROM
	wire [`CODEPAGE_DATA_WIDTH-1:0] text_cpData;			// Datenleitungen aus dem Codepage-ROM
	wire [BIT_PER_PIXEL-1:0] char_pixelOut;				// Ausgabedaten vom Textmodul
	wire [ld(SCREEN_WIDTH/CODEPAGE_CHAR_WIDTH)-1:0] text_charPosX;	// Spalte des aktuell gezeichneten Zeichens
	wire [ld(SCREEN_HEIGHT/CODEPAGE_CHAR_HEIGHT)-1:0] text_charPosY;	// Zeile des aktuell gezeichneten Zeichens
	wire [ld(CODEPAGE_CHAR_HEIGHT)-1:0] text_cursorAddr;

	// Hier waere eine bedingte Deklaration bei (USE_TEXT_MODE == 1) sinnvoll.
		reg [CODEPAGE_CHAR_WIDTH-1:0] text_cursorMem [CODEPAGE_CHAR_HEIGHT-1:0];// Speicher fuer das Cursor-Zeichen


/* Grafikdarstellung */
	wire [ld(SCREEN_WIDTH*SCREEN_HEIGHT*BIT_PER_MEMORY_PIXEL/(PIXEL_PER_MEMORY_WORD*BIT_PER_MEMORY_PIXEL))-1:0] graphics_memAddr;
								// Addressleitungen zum Grafikspeicher
	wire [`GRAPHMEM_DATA_WIDTH-1:0] graphics_memData;	// Datenleitungen vom Grafikspeicher
	wire [BIT_PER_PIXEL-1:0] graphics_pixelOut;		// Ausgabedaten vom Grafikmodul

	// Hier waere eine bedingte Deklaration bei (USE_GRAPHIC_MODE == 1) sinnvoll.
		reg  [PIXEL_PER_MEMORY_WORD*BIT_PER_PIXEL-1:0] graphics_pixelData;	// Aus Color-LUT konvertierte Pixeldaten


/* Grafikfunktionen */
	// Hier waere eine bedingte Deklaration bei (USE_GRAPHIC_MODE == 1) sinnvoll.
		reg [ld(`GRAPH_FUNCTION_MAXCOORDINATES * 2)-1:0] graphics_function_coordAddr;
		reg [ld(max(SCREEN_WIDTH, SCREEN_HEIGHT))-1:0] graphics_function_coordinate[0:2*`GRAPH_FUNCTION_MAXCOORDINATES-1];

	wire [BIT_PER_MEMORY_PIXEL-1:0] graphics_function_colorOut;
	wire graphics_function_ready;
	wire graphics_function_readyOld; // Verzoegertes Signal zur Erzeugung des ready-interrupts
	wire graphics_function_memWe;
	wire [ld(SCREEN_WIDTH*SCREEN_HEIGHT*BIT_PER_MEMORY_PIXEL/(PIXEL_PER_MEMORY_WORD*BIT_PER_MEMORY_PIXEL))-1:0] graphics_function_memAddr;
	wire [`GRAPHMEM_DATA_WIDTH-1:0] graphics_function_memDataIn;

/* Leitungen fuer zweiten Port des Grafikspeichers */
	wire [ld(SCREEN_WIDTH*SCREEN_HEIGHT*BIT_PER_MEMORY_PIXEL/(PIXEL_PER_MEMORY_WORD*BIT_PER_MEMORY_PIXEL))-1:0] graphics_memAddrB;
	wire [`GRAPHMEM_DATA_WIDTH-1:0] graphics_memDataB;
	wire [`GRAPHMEM_DATA_WIDTH-1:0] graphics_memDataInB;
	wire graphics_memWeB;

/* DMA Speicher */
	// Adressen innerhalb des jeweiligen BRAMs
	wire [ADDR_WIDTH_GRAPHMEM-1:0]	addr_graphmem	= addr_bus[0 +: ADDR_WIDTH_GRAPHMEM];	// Adresse im Grafikspeicher
	wire [ADDR_WIDTH_TEXTMEM-1:0]	addr_textmem	= addr_bus[0 +: ADDR_WIDTH_TEXTMEM];	// Adresse im Textspeicher
	wire [ADDR_WIDTH_CODEPAGE-1:0]	addr_codepage	= addr_bus[0 +: ADDR_WIDTH_CODEPAGE];	// Adresse im Codepage-Speicher
	wire [ADDR_WIDTH_COLOR_LUT-1:0]	addr_colorLut	= addr_bus[0 +: ADDR_WIDTH_COLOR_LUT];	// Adresse in der Color-Lookup-Table
	wire [ld(CODEPAGE_CHAR_HEIGHT)-1:0] addr_cursor = addr_bus[0 +: ld(CODEPAGE_CHAR_HEIGHT)]; // Adresse in den Speicher fuer den Cursor

	// Chip-Select fuer die BRAMs
	wire active_dma_graphmem;
	wire active_dma_textmem;
	wire active_dma_codepage;
	wire active_dma_colorLut;
	wire active_dma_cursormem;

	// Write-Enable-Signale
	wire we_dma_graph	= store_access & active_dma_graphmem;
	wire we_dma_char	= store_access & active_dma_textmem;
	wire we_dma_codepage	= store_access & active_dma_codepage;
	wire we_dma_colorLut	= store_access & active_dma_colorLut;
	wire we_dma_cursor	= store_access & active_dma_cursormem;

	// Datenausgaenge der DMA-Speicher
	wire [`GRAPHMEM_DATA_WIDTH-1:0]	graphmem_do;
	wire [`TEXTMEM_DATA_WIDTH-1:0]	textmem_do;
	wire [`CODEPAGE_DATA_WIDTH-1:0]	codepage_do;
	wire [BIT_PER_PIXEL-1:0]	colorlut_do;
	wire [CODEPAGE_CHAR_WIDTH-1:0]	cursormem_do;

	generate
	begin: ASSIGN_MEM_DO
		assign cursormem_do = (USE_TEXT_MODE == 1) ? text_cursorMem[addr_cursor] : 0;
		assign colorlut_do = (USE_GRAPHIC_MODE == 1) ? graphics_colorLut[addr_colorLut] : 0;
	end
	endgenerate

/* Ausgabe auf Peripheriebus */
	wire [17:0] charPosData;
	assign charPosData[17:9] = text_charPosX;
	assign charPosData[8:0] = text_charPosY;

	assign di_peri= active_dma_graphmem ? graphmem_do :			// DMA-Zugriff Grafikspeicher
			active_dma_textmem ? textmem_do :			// DMA-Zugriff Textspeicher
			active_dma_codepage ? codepage_do :			// DMA-Zugriff Codepage
			active_dma_colorLut ? colorlut_do :			// DMA-Zugriff Color-Lookup-Table
			active_dma_cursormem ? cursormem_do :			// DMA-Zugriff auf den Speicher fuer den Cursor

			rd_displayStatus ? displayStatus :			// Statusregister des Displays
			rd_text_color ? text_color :				// Register fuer Textfarbe
			rd_text_bgColor ? text_bgColor :			// Register fuer Texthintergrundfarbe
			rd_text_charpos ? charPosData :				// Register der Position des aktuell gezeichneten Textzeichens
			rd_text_cursorpos ? { char_cursorX, char_cursorY } :	// aktuelle Position des Text-Cursors
			rd_graph_coordAddr ? graphics_function_coordAddr :	// bisher angewaehlte Koordinate
			rd_graph_coordValue ? graphics_function_coordinate[graphics_function_coordAddr] : // Wert der Koordinate
			rd_graph_color ? graphics_function_color :		// Farbe fuer den Funktionsaufruf bzw. des Funktionsaufrufs
			0;

/* Ausgabe der Interrupts (charLine_ir wird direkt im character-Modul erzeugt). Jeder Interrupt ist nur einen Takt lang high! */
	assign graphFunctionReady_ir = graphics_function_ready && ~displayStatus[`STATUSBIT_FUNCTIONREADY];

/**************************************************************/

	// Pixeldaten aus Color-LUT errechnen, sobald sich die Daten aus dem Grafikspeicher aendern
	always @(graphics_memData)
	begin : fillPixelData
		reg [ld(PIXEL_PER_MEMORY_WORD)-1:0] i;
		for (i = 0; i < PIXEL_PER_MEMORY_WORD; i = i + 1)
			graphics_pixelData[i*BIT_PER_PIXEL +: BIT_PER_PIXEL] <= graphics_colorLut[graphics_memData[i*BIT_PER_MEMORY_PIXEL +: BIT_PER_MEMORY_PIXEL]];
	end

	// "Einfangen" neuer Pixeldaten bei Bedarf vom Character- oder Grafik-Modul
	always @(posedge clk_peri or posedge reset)
	begin : dataReceive
		if (reset)
		begin
			dataReceiveState <= 0;
			dataReceiveCount <= 0;
			dataReceiveShiftReg <= 0;
			pixelWait <= 1;
		end
		else begin
			case (dataReceiveState)
				0:	if (receiveNewData)
					begin
						// es sollen weitere Daten geholt werden
						pixelWait <= 0;
						dataReceiveCount <= 0;
						dataReceiveState <= 1;
					end
					else
						dataReceiveState <= 0;
				1:	dataReceiveState <= 2;	// Wartetakt, da jetzt noch keine neuen Daten anliegen wuerden
				2:	begin
						// neue Daten ins Pufferregister schieben
						// Wenn Overlay aktiviert, pruefen, ob die Farbe des Pixels des aktuellen Modus gleich der Overlay-Farbe ist und diesen bei Bedarf ersetzen.
						if (displayStatus[`STATUSBIT_OVERLAY])
							dataReceiveShiftReg <= { dataReceiveShiftReg, (overlayColor[0 +: BIT_PER_PIXEL] == (textMode ? char_pixelOut : graphics_pixelOut) ? (textMode ? graphics_pixelOut : char_pixelOut) : (textMode ? char_pixelOut : graphics_pixelOut))};
						else
							//dataReceiveShiftReg <= textMode ? { dataReceiveShiftReg, char_pixelOut } : { dataReceiveShiftReg, graphics_pixelOut };
							dataReceiveShiftReg <= { dataReceiveShiftReg, (textMode ? char_pixelOut : graphics_pixelOut) };

						// Pruefen, ob genug Daten gelesen wurden
						if (DATA_WORD_WIDTH-1 == dataReceiveCount)
							// erst nach dem letzten Datum in den Ausgangszusand zurueckkehren
							dataReceiveState <= 3;
						else begin
							// Senden weiterer Daten unterdruecken, jedoch noch das letzte Datum abwarten
							// Achtung: Aenderung hat einen Takt Verzoegerung. Im Folgenden Takt kommt noch ein Datum!
							if (DATA_WORD_WIDTH-2 == dataReceiveCount)
								pixelWait <= 1;

							dataReceiveState <= 2;
							dataReceiveCount <= dataReceiveCount + 1;
						end
					end
				3:	dataReceiveState <= receiveNewData ? 3 : 0;
				default: dataReceiveState <= 0;
			endcase
		end
	end

	// Vorbereiten der als naechstes auszugebenden Daten, ggf. Anfordern neuer Daten, wenn Puffer leer
	always @(posedge sig_cl2 or posedge reset)
	begin
		if (reset)
		begin
			receiveNewData	<= 0;
			dataWordModCount<= BIT_PER_PIXEL-1;
			dataShiftReg	<= 0;
		end
		else begin
			receiveNewData <= 0;
			if (BIT_PER_PIXEL-1 == dataWordModCount)
			begin
				// neue Daten anfordern, wenn der aktuelle Puffer rausgeschrieben wurde
				dataWordModCount<= 0;
				dataShiftReg	<= dataReceiveShiftReg;
				receiveNewData	<= 1;
			end
			else begin
				// sonst Puffer nur weiterschieben
				dataWordModCount<= dataWordModCount + 1;
				dataShiftReg	<= dataShiftReg << DATA_WORD_WIDTH;
			end
		end
	end

	// Ausgabe der Daten ans Display
	always @(posedge clk_peri or posedge reset)
	begin : dataOutput
		if (reset)
		begin
			display_framestart <= 0;
			display_cl1	<= 0;
			display_cl2	<= 0;
			display_dispoff	<= 0;
			display_invert	<= 0;
			display_vdd	<= 0;
			display_backlight <= 0;
			display_vcon	<= 0;
			display_data	<= 0;
		end
		else begin
			// Um einen Systemtakt verzoegerte Ausgabe der Signale ans Display
			display_framestart <= sig_framestart;
			display_cl1	<= sig_cl1;
			display_cl2	<= sig_cl2;
			display_dispoff	<= sig_dispoff;
			display_invert	<= sig_invert;
			display_vdd	<= sig_vdd;
			display_backlight <= dispBacklight;
			display_vcon	<= sig_vcon;
			display_data	<= dataShiftReg[0:DATA_WORD_WIDTH-1];
		end
	end

	// Schreibzugriffe auf die Konfigurationsregister
	always @(posedge clk_peri or posedge reset)
	begin
		if (reset)
		begin : registerReset
			reg [ld(`GRAPH_FUNCTION_MAXCOORDINATES * 2)-1:0] i;
			// Ruecksetzen der Koordinatenregister
			for (i = 0; i < `GRAPH_FUNCTION_MAXCOORDINATES * 2; i = i + 1)
				graphics_function_coordinate[i] <= 0;
			graphics_function_coordAddr <= 0;
			graphics_function_color <= 1;

			text_color	<= { BIT_PER_PIXEL {1'b1}};	// Vordergrundfarbe weiß
			text_bgColor	<= { BIT_PER_PIXEL {1'b0}};	// Hintergrundfarbe schwarz
			char_cursorX	<= 0;
			char_cursorY	<= 0;
			overlayColor	<= 0;
			displayStatus	<= 0;
			displayStatus[`STATUSBIT_TEXTMODE] <= 1;	// Display ist ausgeschaltet, lediglich Textmodus ist aktiviert
			//displayStatus	<= 18'b111;
		end
		else begin
			displayStatus[`STATUSBIT_FUNCTIONREADY] <= graphics_function_ready;	// Ready-Signal ins Register schreiben
			displayStatus[`STATUSBIT_FUNCTIONDRAW]	<= 1'b0;			// Draw-Signal immer auf 0 setzen, somit ist es hoechsten einen Takt lang 1

			if (displayStatus[`STATUSBIT_FUNCTIONSELECT] == `FUNCTION_GETPIXEL)
				graphics_function_color <= graphics_function_colorOut;		// Wenn Pixel gelesen wurde, die Farbe ins Farbregister schreiben, damit es ueber die selbe Adresse auslesbar ist

			if (wr_text_color)
				text_color <= do_peri[0 +: BIT_PER_PIXEL];
			else if (wr_text_bgColor)
				text_bgColor <= do_peri[0 +: BIT_PER_PIXEL];
			else if (wr_displayStatus)
			begin
				displayStatus <= do_peri;
				// Read-only-Bits duerfen nicht ueberschrieben werden!
				displayStatus[`STATUSBIT_FUNCTIONREADY] <= graphics_function_ready;

				// Falls das Overlay-Bit gesetzt wird, jetzt auch die Farbe aus dem regulaeren Farbregister sichern
				if ((do_peri[`STATUSBIT_OVERLAY] == 1) && (displayStatus[`STATUSBIT_OVERLAY] == 0))
					overlayColor <= graphics_function_color;
			end
			else if (wr_graph_coordAddr)
				graphics_function_coordAddr <= do_peri[0 +: ld(`GRAPH_FUNCTION_MAXCOORDINATES*2)];
			else if (wr_graph_coordValue)
				graphics_function_coordinate[graphics_function_coordAddr] <= do_peri[0 +: ld(max(SCREEN_WIDTH, SCREEN_HEIGHT))];
			else if (wr_graph_color)
				//graphics_function_color <= do_peri[0 +: BIT_PER_MEMORY_PIXEL];
				graphics_function_color <= do_peri;
			else if (wr_text_cursorpos)
			begin
				char_cursorX <= do_peri[17:9];
				char_cursorY <= do_peri[8:0];
			end
		end
	end

/****** Modulauswahl ******/
/*	pselect #(
			.ADDR_WIDTH(10 - `IO_OFFSET_WIDTH),
			.BASE_WIDTH(10 - `IO_OFFSET_WIDTH),
			.BASE_ADDR(BASE_ADR >> `IO_OFFSET_WIDTH)
		)
		CHIPSELECT (
			.addr(addr_peri[9:`IO_OFFSET_WIDTH]),
			.activ_peri(access_peri),
			.select(module_select)
		); */
// Dekodierung der Adressen ********************** START *************************************


// Addressdekoder fuer alle Geraeteregister (verwendet pselect.v und
// verzoegert die read Signale um einen Takt fuer die neue Pipeline)
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	module_select	),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// Dekodierung der Adressen ********************** END **************************************

/****** Signalerzeugung fuer das Pixel-Display ******/
	display_siggen #(
			.SCREEN_WIDTH(SCREEN_WIDTH),
			.SCREEN_HEIGHT(SCREEN_HEIGHT),

			.BIT_PER_PIXEL(BIT_PER_PIXEL),
			.DATA_WORD_WIDTH(DATA_WORD_WIDTH),

			.SCREEN_FINISH_DELAY(TIMING_SCREEN_FINISH_DELAY),

			// Initialisierung
			.INIT_CLOCKS_TO_VCON_ON(TIMING_INIT_CLOCKS_TO_VCON_ON),
			.INIT_CLOCKS_TO_DISPOFF(TIMING_INIT_CLOCKS_TO_DISPOFF),

			// Zeilentakt
			.CL1_CLOCKS_AFTER_RESET(TIMING_CL1_CLOCKS_AFTER_RESET),
			.CL1_CLOCKS_HIGH(TIMING_CL1_CLOCKS_HIGH),
			.CL1_CLOCKS_LOW(TIMING_CL1_CLOCKS_LOW),
			.CL1_BEGIN_HIGH_LOW(TIMING_CL1_BEGIN_HIGH_LOW),

			// Datentakt
			.CL2_CLOCKS_AFTER_RESET(TIMING_CL2_CLOCKS_AFTER_RESET),
			.CL2_CLOCKS_HIGH(TIMING_CL2_CLOCKS_HIGH),
			.CL2_CLOCKS_LOW(TIMING_CL2_CLOCKS_LOW),
			.ROW_END_DELAY(TIMING_ROW_END_DELAY),
			.ROW_BEGIN_DELAY(TIMING_ROW_BEGIN_DELAY),

			// Framestart-Timing
			.FRAMESTART_CLOCKS_AFTER_RESET(TIMING_FRAMESTART_CLOCKS_AFTER_RESET),
			.FRAMESTART_CLOCKS_HIGH(TIMING_FRAMESTART_CLOCKS_HIGH),
			.FRAMESTART_CLOCKS_LOW(TIMING_FRAMESTART_CLOCKS_LOW),

			// Invertierung (jeden zweiten Frame)
			.INVERT_CLOCKS_AFTER_RESET(TIMING_INVERT_CLOCKS_AFTER_RESET),
			.INVERT_CLOCKS_TOGGLE(TIMING_INVERT_CLOCKS_TOGGLE)
		)
		DISPLAYSIGNALS (
			.clk(clk_peri),
			.reset(reset),

			.displayOn(dispOn),
			.framestart(sig_framestart),
			.cl1(sig_cl1),
			.cl2(sig_cl2),
			.invert(sig_invert),
			.vcon(sig_vcon),
			.dispoff(sig_dispoff),
			.vdd(sig_vdd)
		);

	generate
/****** Modul fuer Grafikmodus ******/
		if (USE_GRAPHIC_MODE == 1)
		begin: MODULES_GRAPHICMODE
			display_graph #(
					.SCREEN_WIDTH(SCREEN_WIDTH),
					.SCREEN_HEIGHT(SCREEN_HEIGHT),
					.BIT_PER_PIXEL(BIT_PER_PIXEL),
					.BIT_PER_MEMORY_PIXEL(BIT_PER_MEMORY_PIXEL),
					.PIXEL_START_AFTER_RESET(DATA_WORD_WIDTH),
					.PIXEL_PER_MEMORY_WORD(PIXEL_PER_MEMORY_WORD)
				)
				GRAPHICMODE (
					.clk(clk_peri),
					.reset(reset),

					.mem_addr(graphics_memAddr),

					.pixelWait(pixelWait),
					.pixelOut(graphics_pixelOut),

					.lutPixelData(graphics_pixelData)
				);

/****** Modul fuer Grafikfunktionen ******/
			display_graph_function #(
					.SCREEN_WIDTH(SCREEN_WIDTH),
					.SCREEN_HEIGHT(SCREEN_HEIGHT),
					.BIT_PER_PIXEL(BIT_PER_PIXEL),
					.BIT_PER_MEMORY_PIXEL(BIT_PER_MEMORY_PIXEL),
					.PIXEL_PER_MEMORY_WORD(PIXEL_PER_MEMORY_WORD),
					.MEM_DATA_WIDTH(`GRAPHMEM_DATA_WIDTH),

					.USE_SETPIXEL(USE_GRAPHFUNCTION_SETPIXEL),
					.USE_LINE(USE_GRAPHFUNCTION_LINE),
					.USE_COPYRECT(USE_GRAPHFUNCTION_COPYRECT),
					.USE_GETPIXEL(USE_GRAPHFUNCTION_GETPIXEL),
					.USE_FILLRECT(USE_GRAPHFUNCTION_FILLRECT)
				)
				GRAPHFUNCTION (
					.clk(clk_peri),
					.reset(reset),
					.mem_addr(graphics_function_memAddr),
					.mem_readData(graphics_memDataB),
					.mem_writeData(graphics_function_memDataIn),
					.mem_writeEnable(graphics_function_memWe),
					.coord_x1(graphics_function_coordinate[0] [ld(SCREEN_WIDTH)-1:0]),
					.coord_y1(graphics_function_coordinate[1] [ld(SCREEN_HEIGHT)-1:0]),
					.coord_x2(graphics_function_coordinate[2] [ld(SCREEN_WIDTH)-1:0]),
					.coord_y2(graphics_function_coordinate[3] [ld(SCREEN_HEIGHT)-1:0]),
					.coord_x3(graphics_function_coordinate[4] [ld(SCREEN_WIDTH)-1:0]),
					.coord_y3(graphics_function_coordinate[5] [ld(SCREEN_HEIGHT)-1:0]),
					.color(graphics_function_color[0 +: BIT_PER_MEMORY_PIXEL]),
					.colorOut(graphics_function_colorOut),
					.optionalFlag1(displayStatus[`STATUSBIT_FUNCTIONFLAG1]),
					.ready(graphics_function_ready),
					.draw(displayStatus[`STATUSBIT_FUNCTIONDRAW]),
					.functionSelect(displayStatus[`STATUSBIT_FUNCTIONSELECT])
				);
		end
	endgenerate

	generate
/****** Modul fuer Textausgabe ******/
		if (USE_TEXT_MODE == 1)
		begin: MODULES_TEXTMODE
			display_char #(
					.CHAR_WIDTH(CODEPAGE_CHAR_WIDTH),
					.CHAR_HEIGHT(CODEPAGE_CHAR_HEIGHT),
					.CHAR_CODEPAGE_SIZE(CODEPAGE_SIZE),

					.CHARS_PER_ROW(SCREEN_WIDTH / CODEPAGE_CHAR_WIDTH),
					.CHARS_PER_COLUMN(SCREEN_HEIGHT / CODEPAGE_CHAR_HEIGHT),
					.CHAR_CODEPAGE_DATA_WIDTH(`CODEPAGE_DATA_WIDTH),

					.BIT_PER_PIXEL(BIT_PER_PIXEL),
					.CHARACTER_START_AFTER_RESET(CODEPAGE_CHAR_WIDTH / DATA_WORD_WIDTH),

					.TEXT_BLINK_FREQUENCY(`TEXT_BLINK_FREQUENCY),
					.CURSOR_BLINK_FREQUENCY(`CURSOR_BLINK_FREQUENCY),
					.SYSTEM_CLOCK(`SYSTEM_CLOCK)
				)
				TEXTMODE (
					.clk(clk_peri),
					.reset(reset),

					.cursorX(char_cursorX[0 +: ld(CHARS_PER_ROW)]),
					.cursorY(char_cursorY[0 +: ld(CHARS_PER_COLUMN)]),
					.cursor_addr(text_cursorAddr),
					.cursor_data(text_cursorMem[text_cursorAddr]),

					.mem_addr(text_memAddr),
					.mem_data(text_memData),

					.cp_addr(text_cpAddr),
					.cp_data(text_cpData),

					.pixelWait(pixelWait),
					.pixelOut(char_pixelOut),

					.charPosX(text_charPosX),
					.charPosY(text_charPosY),

					.text_color(text_color),
					.text_bgColor(text_bgColor),
					.lineInterrupt(charLine_ir)
				);
		end
	endgenerate

/****** Auswahl der DMA-Puffer ******/
/*	Via pselect gibt es aus ungeklaerter Ursache einen wiederholbaren Absturz in ModelSim. Daher sollte
	die Simulation mittels der auskommentierten boolesche Gleichungen erfolgen. */

	generate
	begin: DMA_SELECT
		if (USE_GRAPHIC_MODE == 1)
		begin: GRAPHIC
/*			assign active_dma_graphmem	= mem_access & (addr_bus[17:ADDR_WIDTH_GRAPHMEM] == (GRAPHMEM_BASE_ADDR >> ADDR_WIDTH_GRAPHMEM));
			assign active_dma_colorLut	= mem_access & (addr_bus[17:ADDR_WIDTH_DMA] == (COLOR_LUT_BASE_ADDR >> ADDR_WIDTH_DMA));
*/

			// Grafikspeicher
			pselect #(
					.ADDR_WIDTH(18 - ADDR_WIDTH_GRAPHMEM),
					.BASE_WIDTH(18 - ADDR_WIDTH_GRAPHMEM),
					.BASE_ADDR(GRAPHMEM_BASE_ADDR >> ADDR_WIDTH_GRAPHMEM)
				)
				SEL_GRAPHMEM (
					.addr(addr_bus[17:ADDR_WIDTH_GRAPHMEM]),
					.activ_peri(mem_access),
					.select(active_dma_graphmem)
				);

			// Color-Lookup-Table
			pselect #(
					.ADDR_WIDTH(18 - ADDR_WIDTH_DMA),
					.BASE_WIDTH(18 - ADDR_WIDTH_DMA),
					.BASE_ADDR(COLOR_LUT_BASE_ADDR >> ADDR_WIDTH_DMA)
				)
				SEL_COLORLUT (
					.addr(addr_bus[17:ADDR_WIDTH_DMA]),
					.activ_peri(mem_access),
					.select(active_dma_colorLut)
				);


		end
		else begin
			assign active_dma_graphmem	= 0;
			assign active_dma_colorLut	= 0;
		end

		if (USE_TEXT_MODE == 1)
		begin: TEXT
/*			assign active_dma_textmem	= mem_access & (addr_bus[17:ADDR_WIDTH_DMA] == (TEXTMEM_BASE_ADDR >> ADDR_WIDTH_DMA));
			assign active_dma_codepage	= mem_access & (addr_bus[17:ADDR_WIDTH_DMA] == (CODEPAGE_BASE_ADDR >> ADDR_WIDTH_DMA));
			assign active_dma_cursormem	= mem_access & (addr_bus[17:ADDR_WIDTH_DMA] == (CURSORMEM_BASE_ADDR >> ADDR_WIDTH_DMA));
*/

			// Textspeicher
			pselect #(
					.ADDR_WIDTH(18 - ADDR_WIDTH_DMA),
					.BASE_WIDTH(18 - ADDR_WIDTH_DMA),
					.BASE_ADDR(TEXTMEM_BASE_ADDR >> ADDR_WIDTH_DMA)
				)
				SEL_TEXTMEM (
					.addr(addr_bus[17:ADDR_WIDTH_DMA]),
					.activ_peri(mem_access),
					.select(active_dma_textmem)
				);

			// Codepage
			pselect #(
					.ADDR_WIDTH(18 - ADDR_WIDTH_DMA),
					.BASE_WIDTH(18 - ADDR_WIDTH_DMA),
					.BASE_ADDR(CODEPAGE_BASE_ADDR >> ADDR_WIDTH_DMA)
				)
				SEL_CODEPAGE (
					.addr(addr_bus[17:ADDR_WIDTH_DMA]),
					.activ_peri(mem_access),
					.select(active_dma_codepage)
				);

			// Cursor-Speicher
			pselect #(
					.ADDR_WIDTH(18 - ADDR_WIDTH_DMA),
					.BASE_WIDTH(18 - ADDR_WIDTH_DMA),
					.BASE_ADDR(CURSORMEM_BASE_ADDR >> ADDR_WIDTH_DMA)
				)
				SEL_CURSORMEM (
					.addr(addr_bus[17:ADDR_WIDTH_DMA]),
					.activ_peri(mem_access),
					.select(active_dma_cursormem)
				);

		end
		else begin
			assign active_dma_textmem	= 0;
			assign active_dma_codepage	= 0;
			assign active_dma_cursormem	= 0;
		end
	end
	endgenerate

/****** DMA-Puffer ******/

	// Belegung der Leitungen fuer den zweiten Port des Grafikspeichers.
	// Standardmaessig wird das display_graph_function-Modul geschaltet; nur wenn ein Zugriff
	// vom SpartanMC erfolgt, wird diesem Vorrang gewaehrt.

	generate
	begin: MEMORIES
		if (USE_GRAPHIC_MODE == 1)
		begin: GRAPHIC
			assign graphics_memAddrB = active_dma_graphmem ? addr_graphmem : graphics_function_memAddr;
			assign graphmem_do = active_dma_graphmem ? graphics_memDataB : { `GRAPHMEM_DATA_WIDTH {1'b0}};
			assign graphics_memDataInB = active_dma_graphmem && store_access ?
							do_peri[PIXEL_PER_MEMORY_WORD*BIT_PER_MEMORY_PIXEL-1:0] :
							graphics_function_memDataIn;
			assign graphics_memWeB = active_dma_graphmem ? store_access : graphics_function_memWe;

			// Grafikspeicher insanziieren
			display_memory_select #(
					.DATA_WORDS_REQUIRED((SCREEN_WIDTH*SCREEN_HEIGHT) / PIXEL_PER_MEMORY_WORD),
					.BRAM_SIZE(FPGA_BRAM_SIZE)
				)
				GRAPHMEM (
					// Zugriff vom Grafikmodul aus
					.clk_a(mem_clk),
					.addr_a(graphics_memAddr),
					.do_a(graphics_memData),
					.di_a(9'b0),
					.we_a(1'b0),
					.en_a(1'b1),

					// Zugriff vom Zeichenmodul/SpartanMC aus
					.clk_b(mem_clk),
					.addr_b(graphics_memAddrB),
					.do_b(graphics_memDataB),
					.di_b(graphics_memDataInB),
					.we_b(graphics_memWeB),
					.en_b(1'b1)
				);

			// Color-LUT
			always @(posedge mem_clk or posedge reset)
			begin
				if (reset)
				begin : lutReset
					reg [ld(BIT_PER_MEMORY_PIXEL)-1:0] i;

					graphics_colorLut[0] <= { BIT_PER_PIXEL {1'b0}};			// Kleinste Farbnummer ist schwarz
					graphics_colorLut[2**BIT_PER_MEMORY_PIXEL-1] <= { BIT_PER_PIXEL {1'b1}};// Groesste Farbnummer ist weiss

					for (i = 1; i < 2**BIT_PER_MEMORY_PIXEL-1; i = i + 1)			// Farben dazwischen berechnen
						graphics_colorLut[i] <= ~(i*(2**BIT_PER_PIXEL)/(2**BIT_PER_MEMORY_PIXEL));
				end
				else begin
					if (we_dma_colorLut)
						graphics_colorLut[addr_colorLut] <= do_peri[0 +: BIT_PER_PIXEL];
				end
			end
		end
		else begin
			assign graphics_memAddrB	= 0;
			assign graphmem_do		= 0;
			assign graphics_memDataInB	= 0;
			assign graphics_memWeB		= 0;
		end

		if (USE_TEXT_MODE == 1)
		begin : TEXT
			// Textmem
		/*	Ist der Textbereich klein genug, kann statt eines kompletten Block-RAMS auch ein kleiner Speicher uber LUTs genutzt werden.
			display_lutmemory #(
					.DEPTH((SCREEN_WIDTH/CODEPAGE_CHAR_WIDTH) * (SCREEN_HEIGHT/CODEPAGE_CHAR_HEIGHT)),
					.WIDTH(9)
				)
				TEXTMEM (
					// Zugriff vom Textmodul aus
					.addr_a(text_memAddr),
					.do_a(text_memData),
					.di_a({`TEXTMEM_DATA_WIDTH {1'b0}}),
					.we_a(1'b0),
					.en_a(1'b1),

					// Zugriff vom SpartanMC aus
					.addr_b(addr_textmem),
					.do_b(textmem_do),
					.di_b(do_peri[8:0]),
					.we_b(we_dma_char),
					.en_b(active_dma_textmem)
				);
		*/
			display_memory_select #(
					.DATA_WORDS_REQUIRED((SCREEN_WIDTH/CODEPAGE_CHAR_WIDTH) * (SCREEN_HEIGHT/CODEPAGE_CHAR_HEIGHT)),
					.BRAM_SIZE(FPGA_BRAM_SIZE)
				)
				TEXTMEM (
					// Zugriff vom Textmodul aus
					.clk_a(mem_clk),
					.addr_a(text_memAddr),
					.do_a(text_memData),
					.di_a({`TEXTMEM_DATA_WIDTH {1'b0}}),
					.we_a(1'b0),
					.en_a(1'b1),

					// Zugriff vom SpartanMC aus
					.clk_b(mem_clk),
					.addr_b(addr_textmem),
					.do_b(textmem_do),
					.di_b(do_peri[8:0]),
					.we_b(we_dma_char),
					.en_b(active_dma_textmem)
				);

			// Codepage
			display_memory_select #(
					.DATA_WORDS_REQUIRED(CODEPAGE_CHAR_HEIGHT * CODEPAGE_SIZE),
					.BRAM_SIZE(FPGA_BRAM_SIZE)
				)
				CP_ROM (
					// Zugriff vom Textmodul aus
					.clk_a(mem_clk),
					.addr_a(text_cpAddr),
					.do_a(text_cpData),
					.di_a(9'b0),
					.we_a(1'b0),
					.en_a(1'b1),

					// Zugriff vom SpartanMC aus
					.clk_b(mem_clk),
					.addr_b(addr_codepage),
					.do_b(codepage_do),
					.di_b(do_peri[8:0]),
					.we_b(we_dma_codepage),
					.en_b(active_dma_codepage)
				);


			// Cursor-Speicher
			always @(posedge mem_clk or posedge reset)
			begin
				if (reset)
				begin : cursorReset
					reg [ld(CODEPAGE_CHAR_HEIGHT)-1:0] i;
					for (i = 0; i < CODEPAGE_CHAR_HEIGHT; i = i + 1)
						text_cursorMem[i] <= 0;
					text_cursorMem[CODEPAGE_CHAR_HEIGHT-2] <= { CODEPAGE_CHAR_WIDTH {1'b1} };
					text_cursorMem[CODEPAGE_CHAR_HEIGHT-3] <= { CODEPAGE_CHAR_WIDTH {1'b1} };
				end
				else begin
					if (we_dma_cursor)
						text_cursorMem[addr_cursor] <= do_peri[0 +: CODEPAGE_CHAR_WIDTH];
				end
			end
		end
	end
	endgenerate
/*
// synopsys translate_off
	generate
		if (USE_TEXT_MODE == 1)
		begin
			// Initialisiert die Codepage

			`ifdef toolchain_ise

			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_3F = 256'h0000000000fec6c6c66c38100000000000000000000000000000000000dc7600;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_3E = 256'h0000000070181818180e18181870000000000018181818180018181818180000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_3D = 256'h000000000e18181818701818180e000000000000fec6603018ccfe0000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_3C = 256'h00f80c067ec6c6c6c6c6c6000000000000000000c3663c183c66c30000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_3B = 256'h0000000066ffdbdbc3c3c3000000000000000000386cc6c6c6c6c60000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_3A = 256'h0000000076cccccccccccc0000000000000000001c3630303030fc3030100000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_39 = 256'h000000007cc6067cc0c67c000000000000000000f06060606076dc0000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_38 = 256'h001e0c0c7ccccccccccc76000000000000f060607c6666666666dc0000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_37 = 256'h000000007cc6c6c6c6c67c000000000000000000666666666666dc0000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_36 = 256'h00000000dbdbdbdbdbffe60000000000000000003c1818181818181818380000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_35 = 256'h00000000e6666c78786c666060e00000003c66060606060606060e0006060000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_34 = 256'h000000003c181818181838001818000000000000e666666666766c6060e00000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_33 = 256'h0078cc0c7ccccccccccc760000000000000000007830303030783032361c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_32 = 256'h000000007cc6c0c0fec67c00000000000000000076cccccccc6c3c0c0c1c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_31 = 256'h000000007cc6c0c0c0c67c000000000000000000dc666666666c786060e00000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_30 = 256'h0000000076cccccc7c0c78000000000000000000000000000000000018303000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_2F = 256'h0000ff00000000000000000000000000000000000000000000000000c66c3810;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_2E = 256'h000000003c0c0c0c0c0c0c0c0c3c00000000000002060e1c3870e0c080000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_2D = 256'h000000003c30303030303030303c000000000000fec6c26030180c86c6fe0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_2C = 256'h000000003c181818183c66666666000000000000c6c66c7c38387c6cc6c60000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_2B = 256'h000000006ceefed6d6d6c6c6c6c600000000000010386cc6c6c6c6c6c6c60000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_2A = 256'h000000007cc6c6c6c6c6c6c6c6c60000000000003c1818181818185a7e7e0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_29 = 256'h000000007cc6c6060c3860c6c67c000000000000e66666666c7c666666fc0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_28 = 256'h00000e0c7cd6c6c6c6c6c6c6c67c000000000000f0606060607c666666fc0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_27 = 256'h000000007cc6c6c6c6c6c6c6c67c000000000000c6c6c6c6cedefef6e6c60000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_26 = 256'h00000000c6c6c6c6c6d6fefeeec6000000000000fe6662606060606060f00000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_25 = 256'h00000000e666666c78786c6666e600000000000078cccccc0c0c0c0c0c1e0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_24 = 256'h000000003c18181818181818183c000000000000c6c6c6c6c6fec6c6c6c60000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_23 = 256'h000000003a66c6c6dec0c0c2663c000000000000f06060606878686266fe0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_22 = 256'h00000000fe6662606878686266fe000000000000f86c6666666666666cf80000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_21 = 256'h000000003c66c2c0c0c0c0c2663c000000000000fc666666667c666666fc0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_20 = 256'h00000000c6c6c6c6fec6c66c38100000000000007cc0dcdededec6c67c000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_1F = 256'h0000000018180018180c06c6c67c0000000000006030180c060c183060000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_1E = 256'h000000000000007e00007e000000000000000000060c18306030180c06000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_1D = 256'h0000000030181800000018180000000000000000001818000000181800000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_1C = 256'h00000000780c0606067ec6c6c67c0000000000007cc6c6c6c67cc6c6c67c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_1B = 256'h000000003030303030180c06c6fe0000000000007cc6c6c6c6fcc0c0603c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_1A = 256'h000000007cc6060606fcc0c0c0fe0000000000001e0c0c0cfecc6c3c1c0c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_19 = 256'h000000007cc60606061c0606c67c000000000000fec6c06030180c06c67c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_18 = 256'h000000007e181818181818783818000000000000386cc6c6d6d6c6c66c380000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_17 = 256'h0000000080c06030180c06020000000000000000181800000000000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_16 = 256'h0000000000000000fe0000000000000000000030181818000000000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_15 = 256'h00000000000018187e18180000000000000000000000663cff3c660000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_14 = 256'h0000000030180c0c0c0c0c0c18300000000000000c18303030303030180c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_13 = 256'h000000000000000000000060303030000000000076ccccccdc76386c6c380000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_12 = 256'h0000000083c36030180c06c3c1000000000018187cc68606067cc0c2c67c1818;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_11 = 256'h000000006666ff666666ff666600000000000000000000000000002466666600;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_10 = 256'h0000000018180018183c3c3c3c18000000000000000000000000000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_0F = 256'h0000000000001038387c7cfefe000000000000000000fefe7c7c383810000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_0E = 256'h00000000000000286cfe6c2800000000000000000000fec0c0c0000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_0D = 256'h0000000000002060fe60200000000000000000000000080cfe0c080000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_0C = 256'h00000000183c7e18181818181818000000000000181818181818187e3c180000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_0B = 256'h000000007e183c7e1818187e3c1800000000000000fefefe0000000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_0A = 256'h0000007cc60c386cc6c66c3860c67c00000000001b1b1b1b1b7bdbdbdb7f0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_09 = 256'h000000006666006666666666666600000000000000183c7e1818187e3c180000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_08 = 256'h0000000002060e1e3efe3e1e0e0602000000000080c0e0f0f8fef8f0e0c08000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_07 = 256'h00000000001818db3ce73cdb1818000000000000c0e6e7676363637f637f0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_06 = 256'h00000000e0f070303030303f333f00000000000018187e183c666666663c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_05 = 256'h000000003c666666663c190d070f0000ffffffffffc399bdbd99c3ffffffffff;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_04 = 256'h00000000003c664242663c0000000000ffffffffffffe7c3c3e7ffffffffffff;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_03 = 256'h00000000000000183c3c18000000000000000000003c18187effff7e3c180000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_02 = 256'h00000000003c1818e7e7e73c3c180000000000000000081c3e7f3e1c08000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_01 = 256'h0000000000081c3e7f7f7f7f36000000000000007effffe7c3ffffdbff7e0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_00 = 256'h000000007e818199bd8181a5817e000000000000000000000000000000000000;

			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_3F = 256'h0000000000000000000000000000000000000000007e7e7e7e7e7e7e00000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_3E = 256'h000000000000000000f8c86030d870000000000000000000006c6c6c6c6cd800;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_3D = 256'h000000001c3c6c6cec0c0c0c0c0c0f0000000000000000180000000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_3C = 256'h000000000000001818000000000000000000000000000000000000386c6c3800;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_3B = 256'h000000000000dc7600dc76000000000000000000001818007e00181800000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_3A = 256'h0000000070d8d8d81818181818181818181818181818181818181b1b1b0e0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_39 = 256'h000000007c000c18306030180c000000000000007c006030180c183060000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_38 = 256'h000000007e000018187e1818000000000000000000fe0000fe0000fe00000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_37 = 256'h00000000cccccccccccccccccc78000000000000183060606078606030180000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_36 = 256'h00000000c0607ce6d6de7c0c06000000000000000000007cd6d67c0000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_35 = 256'h000000003c666666663c1830603e000000000000ee6c6c6c6cc6c6c66c380000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_34 = 256'h00000000386cc6c6fec6c66c38000000000000007e183c6666663c187e000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_33 = 256'h00000000181818181818dc760000000000c060607c6666666666660000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_32 = 256'h0000000070d8d8d8d8d87e000000000000000000fee670381c3870e6fe000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_31 = 256'h000000006c6c6c6c6c6c6cfe0000000000000000c0c0c0c0c0c0c0c6c6fe0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_30 = 256'h00000000ccc6c6c6ccd8cccccc7800000000000076dcd8d8d8dc760000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_2F = 256'h0000000000000000ffffffffffffffff0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_2E = 256'hf0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0ffffffffffffffff0000000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_2D = 256'hffffffffffffffffffffffffffffffff18181818181818181f00000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_2C = 256'h0000000000000000f81818181818181818181818181818ff18ff181818181818;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_2B = 256'h3636363636363636ff3636363636363636363636363636363f00000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_2A = 256'h181818181818181f181f000000000000000000000000001f181f181818181818;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_29 = 256'h00000000000000003f363636363636363636363636363636ff00000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_28 = 256'h18181818181818ff00ff0000000000000000000000000000ff36363636363636;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_27 = 256'h00000000000000ff00ff18181818181836363636363636f700f7363636363636;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_26 = 256'h00000000000000ff00ff00000000000036363636363636373037363636363636;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_25 = 256'h36363636363636f700ff00000000000000000000000000ff00f7363636363636;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_24 = 256'h3636363636363637303f000000000000000000000000003f3037363636363636;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_23 = 256'h36363636363636363736363636363636181818181818181f181f181818181818;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_22 = 256'h1818181818181818ff181818181818180000000000000000ff00000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_21 = 256'h18181818181818181f181818181818181818181818181818ff00000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_20 = 256'h0000000000000000ff1818181818181800000000000000001f18181818181818;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_1F = 256'h1818181818181818f80000000000000000000000000000f818f8181818181818;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_1E = 256'h0000000000000000fe3636363636363600000000000000fe06f6363636363636;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_1D = 256'h36363636363636f606fe00000000000036363636363636363636363636363636;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_1C = 256'h36363636363636f606f636363636363618181818181818f818f8000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_1B = 256'h3636363636363636fe000000000000003636363636363636f636363636363636;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_1A = 256'h18181818181818f818f81818181818181818181818181818f818181818181818;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_19 = 256'h1818181818181818181818181818181877dd77dd77dd77dd77dd77dd77dd77dd;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_18 = 256'haa55aa55aa55aa55aa55aa55aa55aa5544114411441144114411441144114411;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_17 = 256'h000000000000d86c366cd80000000000000000000000366cd86c360000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_16 = 256'h00000000183c3c3c3c18180018180000000006063f9ace6630186c6662e06000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_15 = 256'h00003e180c86dc6030186c6662e06000000000000006060606fe000000000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_14 = 256'h0000000000c0c0c0c0fe000000000000000000007cc6c6c06030303000303000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_13 = 256'h00000000000000007c00386c6c38000000000000000000007e003e6c6c3c0000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_12 = 256'h00000000c6c6c6c6cedef6e6c600dc7600000000c6c6c6c6c6e6dc00dc760000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_11 = 256'h0000000076cccccccccccc0060301800000000007cc6c6c6c6c67c0060301800;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_10 = 256'h000000003c1818181818380030180c000000000076cccccc7c0c780060301800;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_0F = 256'h0000000070d81818187e1818181b0e0000000000f36666666f66627c6666fc00;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_0E = 256'h00000000181818ff18ff183c66c3000000000000fce660606060f060646c3800;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_0D = 256'h00000018187cc6c0c0c0c67c18180000000000007cc6c6c6c6c6c6c6c600c600;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_0C = 256'h000000007cc6c6c6c6c6c6c67c00c60000780c067ec6c6c6c6c6c60000c60000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_0B = 256'h0000000076cccccccccccc00183060000000000076cccccccccccc00cc783000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_0A = 256'h000000007cc6c6c6c6c67c0018306000000000007cc6c6c6c6c67c0000c60000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_09 = 256'h000000007cc6c6c6c6c67c006c38100000000000ceccccccfecccc6c3e000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_08 = 256'h0000000077dcd87e3b7bde000000000000000000fe6660607c6066fe00603018;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_07 = 256'h00000000c6c6c6fec6c66c3800386c3800000000c6c6c6fec6c66c381000c600;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_06 = 256'h000000003c1818181818380018306000000000003c18181818183800663c1800;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_05 = 256'h000000003c1818181818380000660000000000007cc6c0c0fec67c0018306000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_04 = 256'h000000007cc6c0c0fec67c0000c60000000000007cc6c0c0fec67c006c381000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_03 = 256'h0000003c060c3c666060663c000000000000000076cccccc7c0c7800386c3800;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_02 = 256'h0000000076cccccc7c0c7800183060000000000076cccccc7c0c7800cc000000;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_01 = 256'h0000000076cccccc7c0c78006c381000000000007cc6c0c0fec67c0030180c00;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.INIT_00 = 256'h0000000076cccccccccccc00cc000000007c060c3c66c2c0c0c0c0c2663c0000;

			defparam TEXT_MODE.TEXTMEM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_00 = 256'h000000000000000000000000000000000000000000646c726f57206f6c6c6548;

			`endif

			`ifdef toolchain_quartus

			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.init_file = "TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.ram_9_inst.mif";
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.init_file = "TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.ram_9_inst.mif";

			defparam TEXT_MODE.TEXTMEM.BRAM_ARRAY[0].BRAM.init_file = "TEXT_MODE.TEXTMEM.BRAM_ARRAY[0].BRAM.ram_9_inst.mif";

			`endif

			`ifdef toolchain_lattice

			defparam TEXT_MODE.TEXTMEM.BRAM_ARRAY[0].BRAM.INITVAL_00 = "0x00000000000000000000000000000000000000000000000000000640D8720DE570406F0D86C0CA48" ;

			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_3F = "0x0000000000000FE18CC618C6C070100000000000000000000000000000000000000000000DC0EC00" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_3E = "0x00000000000E018030180300E0301803070000000000000018030180301800018030180301800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_3D = "0x000000000001C180301803070030180300E0000000000000001FCC60C030030CC1FC000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_3C = "0x000F8018060FCC618CC618CC618C0000000000000000000000186660781807866186000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_3B = "0x00000000000CCFF1B6DB186C318600000000000000000000000706C18CC618CC618C000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_3A = "0x00000000000ECCC198CC198CC19800000000000000000000000383606030060301F8300601000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_39 = "0x00000000000F8C600C7C180C60F800000000000000000000001E0600C0600C0761B8000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_38 = "0x0001E0180C0F8CC198CC198CC0EC000000000000000F00C0600F8660CC660CC661B8000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_37 = "0x00000000000F8C618CC618CC60F800000000000000000000000CC660CC660CC661B8000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_36 = "0x00000000001B6DB1B6DB1B6FF1CC0000000000000000000000078180301803018030180303800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_35 = "0x00000000001CC660D8780F06C0CC600C0E0000000003C0CC0600C0600C0600C0601C0000C0600000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_34 = "0x000000000007818030180301807000030180000000000000001CC660CC660CC760D8600C0E000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_33 = "0x000781980C0F8CC198CC198CC0EC00000000000000000000000F03006030060780603206C1C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_32 = "0x00000000000F8C6180C01FCC60F800000000000000000000000ECCC198CC1986C0780C0181C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_31 = "0x00000000000F8C6180C0180C60F800000000000000000000001B8660CC660CC6C0F0600C0E000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_30 = "0x00000000000ECCC198CC0F80C0F00000000000000000000000000000000000000000000303006000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_2F = "0x000001FE0000000000000000000000000000000000000000000000000000000000000018C6C07010" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_2E = "0x00000000000780C0180C0180C0180C0183C0000000000000000040601C1C070701C0C01000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_2D = "0x0000000000078300603006030060300603C0000000000000001FCC618460060180188618CFE00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_2C = "0x000000000007818030180303C0CC660CC6600000000000000018CC60D87C070380F86C18CC600000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_2B = "0x00000000000D8EE1FCD61ACD618CC618CC6000000000000000020380D8C618CC618CC618CC600000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_2A = "0x00000000000F8C618CC618CC618CC618CC60000000000000000781803018030180305A0FC7E00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_29 = "0x00000000000F8C618C06018380C0C618C7C0000000000000001CC660CC660D87C0CC660CCFC00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_28 = "0x0000001C0C0F8D618CC618CC618CC618C7C0000000000000001E0600C0600C07C0CC660CCFC00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_27 = "0x00000000000F8C618CC618CC618CC618C7C00000000000000018CC618CC619CDE1FCF61CCC600000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_26 = "0x000000000018CC618CC618CD61FCFE1DCC60000000000000001FC660C4600C0600C0600C0F000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_25 = "0x00000000001CC660CC6C0F0780D8660CCE60000000000000000F0CC198CC0180C0180C0181E00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_24 = "0x0000000000078180301803018030180303C00000000000000018CC618CC618CFE18CC618CC600000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_23 = "0x00000000000746618CC61BCC0180C20CC3C0000000000000001E0600C0600D0780D0620CCFE00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_22 = "0x00000000001FC660C4600D0780D0620CCFE0000000000000001F06C0CC660CC660CC660D8F800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_21 = "0x000000000007866184C0180C0180C20CC3C0000000000000001F8660CC660CC7C0CC660CCFC00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_20 = "0x000000000018CC618CC61FCC618C6C070100000000000000000F8C01B8DE1BCDE18CC60F80000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_1F = "0x000000000003018000180300C00CC618C7C0000000000000000C0300300C00C0C030300C00000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_1E = "0x0000000000000000007E000000FC000000000000000000000000C0C030300C0300300C00C0000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_1D = "0x00000000000601803000000000301800000000000000000000000180300000000030180000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_1C = "0x00000000000F00C00C0600C7E18CC618C7C0000000000000000F8C618CC618C7C18CC618C7C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_1B = "0x00000000000603006030060180180618CFE0000000000000000F8C618CC618CFC180C00C03C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_1A = "0x00000000000F8C600C0600CFC180C0180FE00000000000000003C0C0180C1FCCC0D83C0380C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_19 = "0x00000000000F8C600C0600C1C00C0618C7C0000000000000001FCC618060060180180618C7C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_18 = "0x00000000000FC18030180301803078070180000000000000000706C18CC61ACD618CC60D83800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_17 = "0x0000000000100C00C0300300C00C0200000000000000000000030180000000000000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_16 = "0x000000000000000000001FC000000000000000000000000030030180300000000000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_15 = "0x000000000000000030180FC180300000000000000000000000000000CC3C1FE3C0CC000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_14 = "0x0000000000060180180C0180C0180C03030000000000000000018180603006030060300300C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_13 = "0x000000000000000000000000000060060300600000000000000ECCC198CC1B8760706C0D83800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_12 = "0x0000000000106C30C0300300C00CC3182000000000000030180F8C610C0600C7C180C218C7C03018" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_11 = "0x00000000000CC661FE660CC661FE660CC00000000000000000000000000000000000240CC660CC00" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_10 = "0x000000000003018000180303C0783C07818000000000000000000000000000000000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_0F = "0x000000000000000020380707C0F8FE1FC00000000000000000000001FCFE0F87C070380200000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_0E = "0x000000000000000000280D8FE0D82800000000000000000000000001FCC0180C0000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_0D = "0x000000000000000040601FC600400000000000000000000000000000100C1FC0C010000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_0C = "0x00000000000303C0FC180301803018030180000000000000000301803018030180307E0781800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_0B = "0x00000000000FC180787E030180307E07818000000000000000000FE1FCFE00000000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_0A = "0x000000007C18C0C0706C18CC60D8380C0C60F80000000000000361B0361B0367B1B6DB1B67F00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_09 = "0x00000000000CC66000660CC660CC660CC66000000000000000000180787E030180307E0781800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_08 = "0x00000000000040601C1E07CFE07C1E01C06004000000000000100C01C0F01F0FE1F0F01C0C010000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_07 = "0x000000000000018030DB078E7078DB03018000000000000000180E61CE670C6630C67F0C67F00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_06 = "0x00000000001C0F00E030060300603F0663F000000000000000030180FC18078660CC660CC3C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_05 = "0x0000000000078660CC660CC3C0320D00E0F000001FEFF1FEFF1FEC3132BD17A99186FF1FEFF1FEFF" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_04 = "0x00000000000003C0CC42084660780000000000001FEFF1FEFF1FEFF1CEC3186E71FEFF1FEFF1FEFF" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_03 = "0x000000000000000000180783C03000000000000000000000000003C030180FCFF1FE7E0781800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_02 = "0x00000000000003C030181CEE71CE3C07818000000000000000000000101C07C7F07C1C0100000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_01 = "0x0000000000000080383E0FE7F0FE7F06C000000000000000000FCFF1FEE7186FF1FEDB1FE7E00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[0].BRAM.INITVAL_00 = "0x00000000000FC811029917A81102A51027E000000000000000000000000000000000000000000000" ;

			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_3F = "0x000000000000000000000000000000000000000000000000000007E0FC7E0FC7E0FC7E0000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_3E = "0x00000000000000000000000F819060060D80E000000000000000000000000006C0D86C0D86C1B000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_3D = "0x00000000000383C0D86C1D80C0180C0180C01E000000000000000000001800000000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_3C = "0x00000000000000000018030000000000000000000000000000000000000000000000380D86C07000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_3B = "0x0000000000000001B876000DC0EC000000000000000000000000018030000FC00030180000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_3A = "0x00000000000E0D81B0D80301803018030180301803018030180301803018030180361B0360E00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_39 = "0x00000000000F800018180606006018018000000000000000000F8000C0300300C030300C00000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_38 = "0x00000000000FC00000180307E0301800000000000000000000000FE000001FC00000FE0000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_37 = "0x0000000000198CC198CC198CC198CC19878000000000000000030300C0600C0780C0600601800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_36 = "0x0000000000180600F8E61ACDE0F80C00C00000000000000000000000007C1ACD60F8000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_35 = "0x0000000000078660CC660CC3C030300C03E0000000000000001DC6C0D86C0D8C618CC60D83800000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_34 = "0x00000000000706C18CC61FCC618C6C070000000000000000000FC18078660CC66078180FC0000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_33 = "0x00000000000301803018030181B8760000000000000C00C0600F8660CC660CC660CC000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_32 = "0x00000000000E0D81B0D81B0D80FC00000000000000000000001FCE60E038038380E0E61FC0000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_31 = "0x00000000000D86C0D86C0D86C0D8FE00000000000000000000180C0180C0180C0180C618CFE00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_30 = "0x0000000000198C618CC6198D8198CC198780000000000000000ECDC1B0D81B0DC0EC000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_2F = "0x000000000000000000001FEFF1FEFF1FEFF1FEFF01E0F01E0F01E0F01E0F01E0F01E0F01E0F01E0F" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_2E = "0x1E0F01E0F01E0F01E0F01E0F01E0F01E0F01E0F01FEFF1FEFF1FEFF1FEFF00000000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_2D = "0x1FEFF1FEFF1FEFF1FEFF1FEFF1FEFF1FEFF1FEFF0301803018030180301803E00000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_2C = "0x000000000000000000001F018030180301803018030180301803018030FF030FF030180301803018" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_2B = "0x06C3606C3606C3606C361FE3606C3606C3606C3606C3606C3606C3606C3607E00000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_2A = "0x0301803018030180301F0301F0000000000000000000000000000000001F0301F030180301803018" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_29 = "0x0000000000000000000007E3606C3606C3606C3606C3606C3606C3606C361FE00000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_28 = "0x030180301803018030FF000FF000000000000000000000000000000000001FE3606C3606C3606C36" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_27 = "0x000000000000000000FF000FF03018030180301806C3606C3606C3606CF7000F706C3606C3606C36" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_26 = "0x000000000000000000FF000FF00000000000000006C3606C3606C3606C370603706C3606C3606C36" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_25 = "0x06C3606C3606C3606CF7000FF000000000000000000000000000000000FF000F706C3606C3606C36" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_24 = "0x06C3606C3606C3606C370603F0000000000000000000000000000000003F0603706C3606C3606C36" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_23 = "0x06C3606C3606C3606C3606E3606C3606C3606C360301803018030180301F0301F030180301803018" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_22 = "0x030180301803018030181FE18030180301803018000000000000000000001FE00000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_21 = "0x0301803018030180301803E18030180301803018030180301803018030181FE00000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_20 = "0x000000000000000000001FE180301803018030180000000000000000000003E18030180301803018" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_1F = "0x030180301803018030181F000000000000000000000000000000000000F8030F8030180301803018" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_1E = "0x000000000000000000001FC3606C3606C3606C36000000000000000000FE00CF606C3606C3606C36" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_1D = "0x06C3606C3606C3606CF600CFE00000000000000006C3606C3606C3606C3606C3606C3606C3606C36" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_1C = "0x06C3606C3606C3606CF600CF606C3606C3606C36030180301803018030F8030F8000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_1B = "0x06C3606C3606C3606C361FC0000000000000000006C3606C3606C3606C361EC3606C3606C3606C36" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_1A = "0x030180301803018030F8030F8030180301803018030180301803018030181F018030180301803018" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_19 = "0x03018030180301803018030180301803018030180EEDD0EEDD0EEDD0EEDD0EEDD0EEDD0EEDD0EEDD" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_18 = "0x15455154551545515455154551545515455154550881108811088110881108811088110881108811" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_17 = "0x0000000000000001B06C06C6C1B000000000000000000000000000006C6C1B06C06C000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_16 = "0x00000000000303C0783C078180300003018000000000000C0607E9A19C66060180D8660C4E00C000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_15 = "0x0000007C18018861B860060180D8660C4E00C00000000000000000600C0600CFE000000000000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_14 = "0x0000000000000C0180C0180FE00000000000000000000000000F8C618CC00C030060300003006000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_13 = "0x000000000000000000000F8000706C0D83800000000000000000000000000FC0007C6C0D83C00000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_12 = "0x000000000018CC618CC619CDE1ECE618C001B876000000000018CC618CC618CE61B8001B87600000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_11 = "0x00000000000ECCC198CC198CC198000C0300300000000000000F8C618CC618CC60F8000C03003000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_10 = "0x000000000007818030180301807000060180180000000000000ECCC198CC0F80C0F0000C03003000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_0F = "0x00000000000E0D8030180307E030180301B01C0000000000001E6660CC660DE660C47C0CC661F800" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_0E = "0x000000000003018030FF030FF0303C0CCC30000000000000001F8E60C0600C0601E0600C86C07000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_0D = "0x00000000180307C18CC0180C018C7C030180000000000000000F8C618CC618CC618CC618C0018C00" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_0C = "0x00000000000F8C618CC618CC618CC60F80018C0000078018060FCC618CC618CC618C00000C600000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_0B = "0x00000000000ECCC198CC198CC19800030300C00000000000000ECCC198CC198CC198001987806000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_0A = "0x00000000000F8C618CC618CC60F800030300C00000000000000F8C618CC618CC60F800000C600000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_09 = "0x00000000000F8C618CC618CC60F8000D83802000000000000019CCC198CC1FCCC1986C07C0000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_08 = "0x00000000000EEDC1B07E0767B1BC00000000000000000000001FC660C0600F8600CCFE0006006018" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_07 = "0x000000000018CC618CFE18CC60D838000380D838000000000018CC618CFE18CC60D8380200018C00" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_06 = "0x000000000007818030180301807000030300C0000000000000078180301803018070000CC3C03000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_05 = "0x000000000007818030180301807000000660000000000000000F8C6180C01FCC60F800030300C000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_04 = "0x00000000000F8C6180C01FCC60F800000C60000000000000000F8C6180C01FCC60F8000D83802000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_03 = "0x000000003C00C0C078660C0600CC3C000000000000000000000ECCC198CC0F80C0F0000706C07000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_02 = "0x00000000000ECCC198CC0F80C0F000030300C00000000000000ECCC198CC0F80C0F0001980000000" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_01 = "0x00000000000ECCC198CC0F80C0F0000D8380200000000000000F8C6180C01FCC60F8000601801800" ;
			defparam TEXT_MODE.CP_ROM.BRAM_ARRAY[1].BRAM.INITVAL_00 = "0x00000000000ECCC198CC198CC1980019800000000007C00C0C07866184C0180C0180C20CC3C00000" ;

			`endif

		end

		if (USE_GRAPHIC_MODE == 1)

		begin
			// Viertes Datenwort (4 Pixel pro Datenwort) alles weiss

			`ifdef toolchain_ise

			defparam GRAPHIC_MODE.GRAPHMEM.BRAM_ARRAY[0].BRAM.ram_9_inst.INIT_00 = 256'h00000000000000000000000000000000000000000000000000000000ff000000;

			`endif

			`ifdef toolchain_quartus

			defparam GRAPHIC_MODE.GRAPHMEM.BRAM_ARRAY[0].BRAM.init_file = "GRAPHIC_MODE.GRAPHMEM.BRAM_ARRAY[0].BRAM.ram_9_inst.mif";

			`endif

			`ifdef toolchain_lattice

			defparam GRAPHIC_MODE.GRAPHMEM.BRAM_ARRAY[0].BRAM.INITVAL_00 = "0x00000000000000000000000000000000000000000000000000000000000000000000001FE0000000" ;

			`endif
			
		end
	endgenerate
// synopsys translate_on
*/
endmodule
