/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Seidowski
// 
// Create Date:    09:29:13 09/29/2008 
// Module Name:    display_char 
// Project Name:   Display-Controller fuer SpartanMC
//
// Description:    Ausgabe von Schriftzeichen auf einem Pixel-Display
//
// Additional Comments: 
//	Funktionsweise: Das Modul arbeitet intern Zeichenweise, d.h. es puffert immer
//		die naechst auszugebene Pixelzile des naechsten Zeichens. Diese Pixel-
//		zeile wird jedoch pixelweise im Systemtakt ueber pixelOut an das
//		uebergeordnete Modul ausgegeben, wenn pixelWait==0. Dort muessen die
//		einzelnen Pixel dann weiter der Ausgabe ans Display entsprechend
//		gepuffert werden. Der korrekte Vorlauf des Char-Moduls nach einem
//		Reset wird ueber den Parameter CHARACTER_START_AFTER_RESET angegeben
//		und bezeichnet, mit welchem Zeichen die Ausgabe nach einem Reset
//		beginnen soll.
//
//	OPTIMIERUNGSPOTENTIAL:
//		Wenn die Zeichenhoehe auf eine Potenz von 2 festgelegt wird, laesst
//		sich die Adressberechnung in der Codepage wesentlich verkleinern,
//		da dann lediglich die Konkatenation verschiedener bereits vorhandener
//		Daten die neue Adresse ergeben wuerde.
//////////////////////////////////////////////////////////////////////////////////
module display_char #(
		parameter BIT_PER_PIXEL = 3,		// Farbtiefe eines Pixels
		parameter CHARACTER_START_AFTER_RESET = 2, // mit welchem Zeichen soll nach einem Reset
							// begonnen werden, damit ausreichend Vorlauf bestehen bleibt

		parameter CHAR_WIDTH = 8,		// Breite eines Zeichens auf dem Display
		parameter CHAR_HEIGHT = 16,		// Hoehe eines Zeichens auf dem Display
		parameter CHAR_CODEPAGE_SIZE = 256,	// Anzahl moeglicher Zeichen
		parameter CHAR_CODEPAGE_DATA_WIDTH = 9,	// Breite eines Datenworts in der Codepage

		parameter CHARS_PER_ROW = 40,		// Zeichen pro Zeile
		parameter CHARS_PER_COLUMN = 15,	// Zeichen pro Spalte
		parameter TEXT_BLINK_FREQUENCY = 2,	// Frequenz fuer blinkenden Text
		parameter CURSOR_BLINK_FREQUENCY = 4,	// Frequenz fuer den blinkenden Cursor
		parameter SYSTEM_CLOCK = 50_000_000	// Systemtakt
	)(
		input wire		clk,
		input wire		reset,		// high-aktiv
		
		input wire [ld(CHARS_PER_ROW)-1:0] cursorX,		// X-Position des Cursors
		input wire [ld(CHARS_PER_COLUMN)-1:0] cursorY,	// Y-Position des Cursors
		output wire [ld(CHAR_HEIGHT)-1:0] cursor_addr,	// Adresse der Cursor-Zeile
		input wire [CHAR_WIDTH-1:0] cursor_data,		// Jeweilige Daten der Cursor-Zeile

		output reg[ld(CHARS_PER_ROW*CHARS_PER_COLUMN)-1:0] mem_addr,	// Adressleitungen fuer Textspeicher
		input wire	[8:0]	mem_data,			// Blink-Bit, 8 Bit Zeichen
		
		// Aufbau der Codepage:
		//	Adresse n*Zeichenbreite bis n*Zeichenbreite+Zeichenhoehe ist fuer das n-te Zeichen
		//	Speicherwortbreite entspricht Zeichenbreite (bspw. 8).
		//	Adressbereich dementsprechend Codepagegroesse * Zeichenbreite * Zeichenhoehe
		
		output reg[ld(CHAR_HEIGHT*CHAR_CODEPAGE_SIZE)-1:0]cp_addr,	// Adresse in die Codepage
		input wire	[CHAR_CODEPAGE_DATA_WIDTH-1:0]	cp_data,		// Daten aus der Codepage
		
		input wire		pixelWait,			// Haltesignal, waehrend High werden keine neuen Daten ausgegeben
		output reg[ld(BIT_PER_PIXEL):0]	pixelOut,	// Pixelausgabe (hier fuer eine Farbtiefe von 3 Bit)
		
		input wire	[BIT_PER_PIXEL-1:0]	text_color,	// Gibt die Farbe fuer den aktuell bearbeiteten Char an
		input wire	[BIT_PER_PIXEL-1:0]	text_bgColor,	// Gibt die Hitnergrundfarbe fuer den aktuell bearbeiteten Char an
		output wire	[ld(CHARS_PER_ROW)-1:0]	charPosX,	// Ausgabe der Position des aktuellen Char (Spalte)
		output wire	[ld(CHARS_PER_COLUMN)-1:0] charPosY,	// Ausgabe der Position des aktuellen Char (Zeile)
		output	reg	lineInterrupt 		// Interrupt, der ausgeloest wird, jedes mal wenn eine neue Char-Zeile vorbereitet werden muss
	);

	`include "display/src/ld.vh"

	localparam BLINK_CLOCKS = SYSTEM_CLOCK / (2 * TEXT_BLINK_FREQUENCY);
	localparam CURSOR_BLINK_CLOCKS = SYSTEM_CLOCK / (2 * CURSOR_BLINK_FREQUENCY);

	reg [BIT_PER_PIXEL*CHAR_WIDTH-1:0] pixelOutShiftReg; // fertiges Pixelregister einer Zeichenzeile zur Ausgabe
	reg [BIT_PER_PIXEL*CHAR_WIDTH-1:0] pixelOutBuf; // fertiges Pixelregister einer Zeichenzeile zur Ausgabe							

	reg [ld(CHARS_PER_ROW*CHARS_PER_COLUMN)-1:0] mem_addrRowBegin; // Adresse des Beginns der aktuellen Zeichenzeile

	reg [ld(CHAR_HEIGHT)-1:0]	count_charRow;	// Zaehler fuer die Ausgabe der Zeile eines Zeichens

	reg [ld(CHARS_PER_ROW)-1:0]	count_charPosX;	// Zaehler fuer die ausgegebenen Zeichen der aktuellen Zeile
	reg [ld(CHARS_PER_COLUMN)-1:0]	count_charPosY;	// Spaltenzaehler

	reg [ld(CHAR_WIDTH)-1:0]	count_pixelOutput;	// Zaehler zur Festsellung, wann pixelOutShiftReg vollstaendig ausgegeben wurde
	reg [ld(BLINK_CLOCKS)-1:0] 	count_blink;		// Zaehler fuer die Erzeugung des Blinksignals
	reg [ld(CURSOR_BLINK_CLOCKS)-1:0] count_cursorBlink;	// Zaehler fuer die Erzeugung des Blinksignals

	reg blink;					// Blinksignal
	reg cursorBlink;				// Blinksignal fuer den Cursor

	assign charPosX = count_charPosX;
	assign charPosY = count_charPosY;
	assign cursor_addr = count_charRow;

	// Erzeugung des Blinksignals fuer blinkende Zeichen
	always @(posedge clk or posedge reset)
	begin
		if (reset)
		begin
			count_blink <= 1;
			blink <= 0;
		end
		else begin
			if (BLINK_CLOCKS == count_blink)
			begin
				blink <= ~blink;
				count_blink <= 1;
			end
			else
				count_blink <= count_blink + 1;
		end
	end
	
	// Erzeugung des Blinksignals fuer den Cursor
	always @(posedge clk or posedge reset)
	begin
		if (reset)
		begin
			count_cursorBlink <= 1;
			cursorBlink <= 0;
		end
		else begin
			if (CURSOR_BLINK_CLOCKS == count_cursorBlink)
			begin
				cursorBlink <= ~cursorBlink;
				count_cursorBlink <= 1;
			end
			else
				count_cursorBlink <= count_cursorBlink + 1;

		end
	end
	
	// Ausgabe der Daten
	always @(posedge clk or posedge reset)
	begin
		if (reset)
		begin
			count_pixelOutput <= 0;
			count_charRow	<= 0;
			count_charPosX	<= CHARACTER_START_AFTER_RESET + 1;
			count_charPosY	<= 0;
			lineInterrupt <= 0;
			pixelOutShiftReg<= 0;
			pixelOutBuf	<= 0;
			mem_addrRowBegin<= 0;
			mem_addr	<= CHARACTER_START_AFTER_RESET + 1;
			cp_addr	<= 0;
			pixelOut	<= 0;
		end
		else begin
			lineInterrupt <= 0;
			if (!pixelWait)
			begin
				/* Ablauf neue Daten lesen:
					nach 0 Takten: "Strecken" der zuvor gelesenen Daten der Zeichen-/Cursorzeile auf komplette Pixeldaten
					nach 1 Takt: Anlegen der naechsten Adresse an Textspeicher
					nach CHAR_WIDTH-2 Takten: Speichern der gelesenen Daten fuer den Cursor
						Anlegen der Adresse entsprechend der gelesenen Daten aus dem Textspeicher an den Char-ROM
				*/

				// Daten ausgeben (die obersten Pixeldaten entsprechen den Ausgabedaten)
				pixelOut <= pixelOutShiftReg[BIT_PER_PIXEL*CHAR_WIDTH-1:BIT_PER_PIXEL*CHAR_WIDTH-3];

				// Pixeldaten weiterschieben
				pixelOutShiftReg <= pixelOutShiftReg << BIT_PER_PIXEL;

				count_pixelOutput <= count_pixelOutput + 1;

				case (count_pixelOutput)
					0:	begin: buildPixelData
							// falls neue Daten verfuegbar sind, diese zuerst auf die
							// jeweilige Farbtiefe "strecken", da die Zeichen selbst nur
							// monochrom abgespeichert sind
							reg [ld(CHAR_WIDTH)-1:0] i;
							for (i = 0; i < CHAR_WIDTH; i = i + 1)
								/* Ein Pixel des aktuellen Zeichens wird unter folgenden Bedingungen durch die Vordergrundfarbe ersetzt:
									Auszugebende Zeichenposition ist gleich der Cursor-Position
										& das Cursor-Zeichen hat an der aktuellen Position in der Zeile einen gesetzten Pixel
										& das Cursor-Zeichen soll gerade dargestellt werden (cursorBlink == 1)
										
								oder
									Das aktuelle Textzeichen hat an der aktuellen Zeilenposition einen gesetzten Pixel
										& nicht (das Blink-Bit ist gesetzt & es soll gerade angezeigt werden (blink == 1))
										
								andernfalls
									wird die Hintergrundfarbe gesetzt.
									
								*/
								pixelOutBuf[(i*BIT_PER_PIXEL) +: BIT_PER_PIXEL] <=
									// Bedingungen fuer den Cursor
									(((count_charPosX == cursorX) && (count_charPosY == cursorY)) && cursor_data[i] && cursorBlink)
									||
									// Bedingungen fuer das normale Textzeichen
									(cp_data[i] && !(mem_data[8] && blink)) ?

									text_color :		// Bei einer von beiden Bedingungen, Vordergrundfarbe nehmen
									text_bgColor;		// sonst Hintergrundfarbe
						end
					1:	begin
							// Anlegen der Adresse des naechsten Zeichens an Textspeicher
							// und anlegen der Adresse an den Char-ROM zum lesen des Cursor-Zeichens

							// Am Zeilenende angekommen?
							if (CHARS_PER_ROW-1 == count_charPosX)
							begin
								count_charPosX <= 0;
								// schon alle Pixelzeilen einer Zeichenreihe dargestellt?
								if (CHAR_HEIGHT-1 == count_charRow)
								begin
									count_charRow <= 0;

									// am unteren Bildschirmrand angekommen?
									if (CHARS_PER_COLUMN-1 == count_charPosY)
									begin
										count_charPosY <= 0;
										mem_addr <= 0;
										mem_addrRowBegin <= 0;
									end
									else begin
										count_charPosY <= count_charPosY + 1;
										mem_addr <= mem_addr + 1;
										mem_addrRowBegin <= mem_addr + 1;
									end
								end
								else begin
									count_charRow <= count_charRow + 1;

									// Zeichenspeicher wieder am Zeilenanfang beginnen zu lesen
									mem_addr <= mem_addrRowBegin;
								end
							end
							else begin
								count_charPosX <= count_charPosX + 1;
								mem_addr <= mem_addr + 1;
							end

							lineInterrupt <= 1;
						end

					CHAR_WIDTH-2: begin
							// Anlegen der Adresse des als naechstes darzustellenden Textzeichens an den Char-ROM mit
							// der darzustellenden Zeile als Offset.
// Optimierungsmoeglichkeit: Wenn die Zeichenhoehe als Potenz von 2 festgelegt wird, kann der eventuelle Multiplizierer gespart werden
							cp_addr <= mem_data[0 +: CHAR_WIDTH] * CHAR_HEIGHT + count_charRow;
						end
					CHAR_WIDTH-1: begin
							count_pixelOutput <= 0;
							pixelOutShiftReg <= pixelOutBuf;
						end
				endcase
			end
		end
	end

endmodule
