/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Seidowski
// Create Date:    16:03:40 10/01/2008 
// Module Name:    display_graph 
// Project Name:   Display-Controller fuer Spartan MC
//
// Description:    Grafikausgabe (pixelgenau)
//
// Additional Comments: 
//	Funktionsweise: Das Modul gibt im Systemtakt pixelweise Daten an das ueber-
//		geordnete Modul, solange pixelWait == 0. Dieses muss die Daten dann
//		dem Display entsprechend puffern. Parallel dazu werden im display_graph-
//		Modul permament neue Daten aus dem Grafikspeicher geladen. Die Daten
//		im Speicher sind momentan aus platzgruenden nur als 2 Bit pro Pixel abgelegt.
//		Mit der entsprechend angelegten Color-Lookup-Tabelle wird daraus dann
//		die Belegung von R, G und B ermittelt.
//
//	WICHTIG!
//		Sollte eine Aenderung von BIT_PER_PIXEL oder BIT_PER_MEMORY_PIXEL erfolgen,
//		so muss unbedingt auch die Color-Lookup-Tabelle entsprechend angepasst werden!!
//
//////////////////////////////////////////////////////////////////////////////////

`default_nettype none

module display_graph #(
		parameter SCREEN_WIDTH = 320,
		parameter SCREEN_HEIGHT = 240,
		parameter BIT_PER_PIXEL = 3,
		parameter BIT_PER_MEMORY_PIXEL = 2,
		parameter PIXEL_START_AFTER_RESET = 8,
		parameter PIXEL_PER_MEMORY_WORD = 4
	) (
		input wire clk,
		input wire reset,
		output reg[ld(SCREEN_WIDTH*SCREEN_HEIGHT*BIT_PER_MEMORY_PIXEL/(PIXEL_PER_MEMORY_WORD*BIT_PER_MEMORY_PIXEL))-1:0] mem_addr,
								// Adressbereich entsprechend noetiger Speicherwoerter
		input wire pixelWait,				// Haltesignal der Ausgabe
		output reg[BIT_PER_PIXEL-1:0] pixelOut,		// Ausgehende Pixeldaten

		input wire [PIXEL_PER_MEMORY_WORD*BIT_PER_PIXEL-1:0] lutPixelData // von der Color-LUT zurueckgelieferte konvertierte Pixeldaten
	);

	`include "display/src/ld.vh"

	reg [ld(PIXEL_PER_MEMORY_WORD*BIT_PER_PIXEL)-1:0] count_pixelOutput;
	reg [PIXEL_PER_MEMORY_WORD*BIT_PER_PIXEL-1:0] pixelOutShiftReg;
	reg [PIXEL_PER_MEMORY_WORD*BIT_PER_PIXEL-1:0] pixelOutBuf;

/***********************************************************************/

	// Ausgabe der Pixeldaten
	always @(posedge clk or posedge reset)
	begin
		if (reset)
		begin
			// +1 bei der Startadresse, da eine doppelte Pufferung noetig ist (einmal Pufferung
			// ueber das Schieberegister und einmal zum Lesen und Umcodieren der Daten aus dem Grafikspeicher)
			mem_addr <= PIXEL_START_AFTER_RESET/PIXEL_PER_MEMORY_WORD + 1;

			count_pixelOutput <= 0;
			pixelOutShiftReg <= 0;
			pixelOutBuf <= 0;
			pixelOut <= 0;
		end
		else begin
			if (!pixelWait)
			begin : pixelOutput
				/* Ablauf lesen neuer Daten:
					Takt 0: Puffern der von der Color-LUT zurueckgegebenen konvertierten, an der
						im letzten Zyklus angelegten Speicheradresse gelesenen Pixeldaten

					Takt 1: Anlegen der naechsten Adresse an den Grafikspeicher

					Takt PIXEL_PER_MEMORY_WORD-1: Ausgabeschieberegister mit den zuvor
						gepufferten Daten beschreiben. Diese doppelte Pufferung ist
						noetig, da der Pfad in Takt 1 wesentlich laenger als einen
						Takt benoetigt und die Daten so nicht ausreichend schnell
						verfuegbar waeren.
				*/
				
				// Die rechtesten BIT_PER_PIXEL bit sind immer die aktuell auszugebenden
				pixelOut <= pixelOutShiftReg[0 +: BIT_PER_PIXEL];

				// Schieberegister um BIT_PER_PIXEL weiter nach rechts schieben
				pixelOutShiftReg <= pixelOutShiftReg >> BIT_PER_PIXEL;

				count_pixelOutput <= count_pixelOutput + 1;

				case (count_pixelOutput)
					0:	// neue Daten in den Puffer schreiben
						pixelOutBuf <= lutPixelData;

					1:	// naechste Adresse an Grafikspeicher anlegen
						mem_addr <= (SCREEN_WIDTH*SCREEN_HEIGHT / PIXEL_PER_MEMORY_WORD - 1 == mem_addr) ? 0 : mem_addr + 1;

					PIXEL_PER_MEMORY_WORD-1: begin
							count_pixelOutput <= 0;			// Zaehler zuruecksetzen
							pixelOutShiftReg <= pixelOutBuf;	// Lesepuffer in den Ausgabepuffer schreiben
						end
				endcase
			end
		end
	end
endmodule
