/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Seidowski
// 
// Create Date:    09:34:18 10/20/2008 
// Module Name:    display_graph_function_copyrect 
// Project Name:   Display-Controller fuer SpartanMC
//
// Description:    Modul zum Kopieren eines rechteckigen Bildbereichs aus dem
//                 uebergebenen Grafikspeicher an eine andere Position innerhalb
//                 des Speichers. Dieser Vorgang ist optional unter Angabe
//                 einer Farbe transparent moeglich.
//
//////////////////////////////////////////////////////////////////////////////////
module display_graph_function_copyrect #(
		parameter SCREEN_WIDTH = 320,
		parameter SCREEN_HEIGHT = 240,
		parameter BIT_PER_MEMORY_PIXEL = 2,
		parameter PIXEL_PER_MEMORY_WORD = 4,
		parameter MEM_DATA_WIDTH = 9
	)(
		input wire clk,
		input wire reset,
		output reg [ld(SCREEN_WIDTH*SCREEN_HEIGHT * BIT_PER_MEMORY_PIXEL / (PIXEL_PER_MEMORY_WORD*BIT_PER_MEMORY_PIXEL))-1:0] mem_addr,
		input wire [MEM_DATA_WIDTH-1:0] mem_readData,

		input wire [ld(SCREEN_WIDTH-1)-1:0]	coord_src_x1,
		input wire [ld(SCREEN_WIDTH-1)-1:0]	coord_src_x2,
		input wire [ld(SCREEN_HEIGHT-1)-1:0]	coord_src_y1,
		input wire [ld(SCREEN_HEIGHT-1)-1:0]	coord_src_y2,
		input wire [ld(SCREEN_WIDTH-1)-1:0]	coord_dest_x,
		input wire [ld(SCREEN_HEIGHT-1)-1:0]	coord_dest_y,

		output wire [ld(SCREEN_WIDTH-1)-1:0] pixelx,
		output wire [ld(SCREEN_HEIGHT-1)-1:0]pixely,
		output reg [BIT_PER_MEMORY_PIXEL-1:0] pixelColor,
		output reg pixelDraw,
		input wire pixelReady,

		input wire [BIT_PER_MEMORY_PIXEL-1:0] transparentColor,
		input wire copyTransparent,

		input wire draw,
		output reg ready
	);

	`include "display/src/ld.vh"

	reg [2:0] copyState;
	reg [ld(SCREEN_WIDTH-1)-1:0] currentSrcX, currentDestX;
	reg [ld(SCREEN_HEIGHT-1)-1:0] currentSrcY, currentDestY;

	assign pixelx = currentDestX;
	assign pixely = currentDestY;


	always @(posedge clk or posedge reset)
	begin
		if (reset)
		begin
			copyState	<= 0;
			mem_addr	<= 0;
			pixelDraw	<= 0;
			pixelColor	<= 0;
			ready		<= 1;
		end
		else begin
			case (copyState)
				// Idle
				0:	begin
						if (draw)
						begin
							currentSrcX <= coord_src_x1;
							currentSrcY <= coord_src_y1;
							currentDestX <= coord_dest_x;
							currentDestY <= coord_dest_y;
							copyState <= 1;
							ready <= 0;
						end
						else begin
							copyState <= 0;
							ready <= 1;
						end
					end

				// Pixel aus Quellrechteck lesen
				1:	begin
						mem_addr <= (currentSrcX/PIXEL_PER_MEMORY_WORD) + (currentSrcY*(SCREEN_WIDTH/PIXEL_PER_MEMORY_WORD));
						`ifdef WAIT_FOR_MEMORY
							copyState <= 2;
						`else
							copyState <= 3;
						`endif
					end

			`ifdef WAIT_FOR_MEMORY
				// Wartetakt fuer Quelldaten
				2:	copyState <= 3;
			`endif

				// Gelesenen Pixel an Zielkoordinaten schreiben
				3:	begin : writeTargetPixel
						if (copyTransparent && (mem_readData[(currentSrcX % PIXEL_PER_MEMORY_WORD)*BIT_PER_MEMORY_PIXEL +: BIT_PER_MEMORY_PIXEL] == transparentColor))
							// Ist die Farbe des aktuellen Pixels gleich der transparenten Farbe und soll transparent kopiert werden?
							copyState <= 5;
						else begin
							// Pixel nur zeichnen, wenn dessen Farbe ungleich der als Transparent angegebenen Farbe ist
							pixelColor <= mem_readData[(currentSrcX % PIXEL_PER_MEMORY_WORD)*BIT_PER_MEMORY_PIXEL +: BIT_PER_MEMORY_PIXEL];
							pixelDraw <= 1;
							copyState <= 4;
						end
					end

				// Verzoegerung, damit pixelReady erst einmal auf 0 gehen kann
				4:	copyState <= 5;

				// Naechste Koordinaten festlegen
				5:	begin
						pixelDraw <= 0;

						if (pixelReady)
						begin
							// Pruefen, ob das rechte Ende des Quellrechtecks oder des Bildschirms erreicht wurde
							if ((currentSrcX == coord_src_x2) ||
								(currentDestX == SCREEN_WIDTH-1))
							begin
								// wenn ja: wieder mit den jeweils linken Koordinaten beginnen
								currentSrcX <= coord_src_x1;
								currentDestX <= coord_dest_x;

								// Pruefen, ob das untere Ende des Quellrechtecks oder des Bildschirms erreicht wurde
								if ((currentSrcY == coord_src_y2) ||
									(currentDestY == SCREEN_HEIGHT-1))
								begin
									// wenn ja: fertig
									copyState <= 0;
									ready <= 1;
								end
								else begin
									// sonst Quell- und Zielzeile um eins weiterschalten
									currentSrcY <= currentSrcY + 1;
									currentDestY <= currentDestY + 1;
									copyState <= 1;
								end
							end
							else begin
								currentSrcX <= currentSrcX + 1;
								currentDestX <= currentDestX + 1;
								copyState <= 1;
							end
						end
						else
							copyState <= 5;
					end
				default: copyState <= 0;
			endcase
		end
	end
endmodule
