/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Seidowski
// 
// Create Date:    10:12:42 10/10/2008 
// Module Name:    display_graph_function_line 
// Project Name:   Display-Controller fuer SpartanMC
//
// Description:    Zeichnet eine Linie nach Bresenham in den Grafikspeicher
//
// TODO:
//	Pruefen, ob die Vorzeichenunterscheidung in den Berechnungen mit
//		Einfuerung von "reg signed" entfernt werden koennen.
//////////////////////////////////////////////////////////////////////////////////
module display_graph_function_line #(
		parameter SCREEN_WIDTH = 320,
		parameter SCREEN_HEIGHT = 240
	) (
		input wire clk,
		input wire reset,
		input wire [ld(SCREEN_WIDTH)-1:0] xstart,
		input wire [ld(SCREEN_WIDTH)-1:0] xend,
		input wire [ld(SCREEN_HEIGHT)-1:0] ystart,
		input wire [ld(SCREEN_HEIGHT)-1:0] yend,
		input wire draw,
		output reg ready,

		output reg [ld(SCREEN_WIDTH)-1:0] pixelx,
		output reg [ld(SCREEN_HEIGHT)-1:0] pixely,
		output reg pixelDraw,
		input wire pixelReady
	);

	`include "display/src/ld.vh"

	localparam MAX_COORD = (SCREEN_HEIGHT > SCREEN_WIDTH) ? SCREEN_HEIGHT  : SCREEN_WIDTH;

	// Je nach dem, welche Koordinate groesser werden kann, muessen folgende Register dimensioniert werden
	reg [ld(MAX_COORD)-1:0] es;
	reg [ld(MAX_COORD)-1:0] el;
	reg [ld(MAX_COORD)  :0] err;	// ein Bit mehr, da ein Vorzeichen gebraucht wird
	reg [ld(MAX_COORD)-1:0] count;

	reg [ld(SCREEN_WIDTH):0] dx;	// ein Bit mehr, da ein Vorzeichen gebraucht wird
	reg [ld(SCREEN_HEIGHT):0] dy;	// ein Bit mehr, da ein Vorzeichen gebraucht wird

	reg [1:0] incx, incy, pdx, pdy, ddx, ddy; // nehmen nur Werte -1,0,1 an
	reg [2:0] lineState;

	always @(posedge clk or posedge reset)
	begin
		if (reset)
		begin
			lineState <= 0;
			pixelx <= 0;
			pixely <= 0;
			dx <= 0;
			dy <= 0;
			es <= 0;
			el <= 0;
			pdx <= 0;
			pdy <= 0;
			ddx <= 0;
			ddy <= 0;
			incx <= 0;
			incy <= 0;

			err <= 0;
			pixelDraw <= 0;
			ready <= 1;
		end
		else begin
			pixelDraw <= 0;
			case (lineState)
				// Initialisierung, erster Teil
				0:	if (draw)
					begin
						ready <= 0;

						dx <= xend-xstart;
						dy <= yend-ystart;
						lineState <= 1;
					end
					else begin
						ready <= 1;
						lineState <= 0;
					end

				// Initialisierung, zweiter Teil (benoetigt dx und dy)
				1:	begin
						ready <= 0;
						
						if (dx[ld(SCREEN_WIDTH)])		// entspricht "<0"
						begin
							dx <= ~dx + 1;			// entspricht dx = -dx
							incx <= -1;			// Vorzeichen des Inkrements in x-Richtung
						end
						else
							incx <= 1;

						if (dy[ld(SCREEN_HEIGHT)])		// entspricht "<0"
						begin
							dy <= ~dy + 1;			// entspricht dy = -dy
							incy <= -1;			// Vorzeichen des Inkrements in y-Richtung
						end
						else
							incy <= 1;

						lineState <= 2;
					end


				// Zeichnen des ersten Pixels vor "Schleifenbeginn"
				2:	begin
						// feststellen, welche Entfernung groesser ist
						if (dx > dy)
						begin
							// pixelx ist "schnellere" Richtung
							pdx <= incx;
							pdy <= 0;
							ddx <= incx;
							ddy <= incy;
							es <= dy;
							el <= dx;
						end
						else begin
							pdx <= 0;
							pdy <= incy;
							ddx <= incx;
							ddy <= incy;
							es <= dx;
							el <= dy;
						end

						pixelx <= xstart;
						pixely <= ystart;

						if (pixelReady)
						begin
							pixelDraw <= 1;
							count <= 0;
							lineState <= 3;
						end
						else begin
							lineState <= 2;
						end
					end

				// restliche Initialisierung
				3:	begin
						err <= (el / 2) - es;
						lineState <= 4;
					end
				// Schleife
				4:	if (pixelReady)
					begin
						// "Schleife" des Linienzeichnens
						if (count < el)
						begin
							/* Anmerkung: Im Gegensatz zum urspruenglichen Algorithmus
								wird der Fehlerterm hier immer einen Zyklus im Voraus
								berechnet. Der erste Wert entsteht daher schon in Zustand
								2. In der Schleife selbst wird somit auch immer der Wert
								fuer den naechsten Durchlauf berechnet
							*/
							
							/* Anmerkung 2: Leider kann in Verilog ein Register nicht als
								vorzeichenbehaftet oder -los deklariert werden. Ein
								2-Bit-Register mit dem Wert -1 (=2'b11) wuerde nach
								der Addition auf ein anderes Register einer Addition
								+2 entsprechen. Somit muss nach dem Vorzeichenbit
								unterschieden werden, um dann wahlweise + oder - zu
								rechnen.
							*/


							if (err[ld(MAX_COORD)])	// entspricht "<0"
							begin
								// Fehlerterm wieder positiv machen
								err <= err + el - es;
								// Schritt in langsamere Richtung (Diagonalschritt)
								pixelx <= ddx[1] ? pixelx - ddx[0] : pixelx + ddx[0];
								pixely <= ddy[1] ? pixely - ddy[0] : pixely + ddy[0];
							end
							else begin
								// Schritt in schnellere Richtung (Parallelschritt)
								pixelx <= pdx[1] ? pixelx - pdx[0] : pixelx + pdx[0];
								pixely <= pdy[1] ? pixely - pdy[0] : pixely + pdy[0];
								err <= err - es;							
							end

							count <= count + 1;

							pixelDraw <= 1;
							lineState <= 5;
						end
						else begin
							ready <= 1;
							lineState <= 0;
						end
					end
					else
						lineState <= 4;

				// Wartezustand
				5:	lineState <= pixelReady ? 4 : 5;

				default: lineState <= 0;
			endcase
		end
	end
endmodule
