/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Seidowski
// Create Date:    12:02:54 10/02/2008 
// Module Name:    display_graph_function_setpixel 
// Project Name:   Display-Controller fuer SpartanMC
//
// Description:    Zeichnet einen Pixel an die gegebene Stelle in den Grafikspeicher
//
// Additional Comments: 
//	Derzeit wird nach dem Anlegen der Adresse an den Speicher ein Takt gewartet,
//	bevor die Daten gelesen werden. Dies ist nicht erforderlich, wenn der Speicher
//	wesentlich schneller getaktet ist wie das System selbst (zb. mit doppeltem Takt).
//	Dazu muss das Macro WAIT_FOR_MEMORY in "display_graph_function_common.v"
//      entfernt werden.
//
//////////////////////////////////////////////////////////////////////////////////

module display_graph_function_setpixel #(
		parameter SCREEN_WIDTH = 320,
		parameter SCREEN_HEIGHT = 240,
		parameter BIT_PER_PIXEL = 3,
		parameter BIT_PER_MEMORY_PIXEL = 2,
		parameter PIXEL_PER_MEMORY_WORD = 4,
		parameter MEM_DATA_WIDTH = 9
	) (
		input wire clk,
		input wire reset,

		output reg [ld(SCREEN_WIDTH*SCREEN_HEIGHT * BIT_PER_MEMORY_PIXEL / (PIXEL_PER_MEMORY_WORD*BIT_PER_MEMORY_PIXEL))-1:0] mem_addr,
		input wire [MEM_DATA_WIDTH-1:0] mem_readData,
		output reg [MEM_DATA_WIDTH-1:0] mem_writeData,
		output reg mem_writeEnable,

		input wire [ld(SCREEN_WIDTH)-1:0] coord_x,
		input wire [ld(SCREEN_HEIGHT)-1:0] coord_y,

		input wire [BIT_PER_MEMORY_PIXEL-1:0] color,

		input wire draw,
		output reg ready
	);

	`include "display/src/ld.vh"

	reg [1:0] memoryWriteState;

	always @(posedge clk or posedge reset)
	begin
		if (reset)
		begin
			ready <= 1;
			mem_writeEnable <= 0;
			mem_addr <= 0;
			mem_writeData <= 0;
			memoryWriteState <= 0;
		end
		else begin
			case (memoryWriteState)
				0:	begin : waitForWriteRequest
						if (draw)
						begin
							// Adresse anlegen, die dem Raster im Speicher entspricht
							mem_addr <= (coord_x/PIXEL_PER_MEMORY_WORD) + (coord_y*(SCREEN_WIDTH/PIXEL_PER_MEMORY_WORD));
							`ifdef WAIT_FOR_MEMORY
								memoryWriteState <= 1;
							`else
								memoryWriteState <= 2;
							`endif
							ready <= 0;
						end
						else
							memoryWriteState <= 0;
					end
			`ifdef WAIT_FOR_MEMORY
				1:	memoryWriteState <= 2;	// Wartetakt zum Lesen aus Speicher
			`endif
				2:	begin : fillWriteData
						reg [ld(BIT_PER_MEMORY_PIXEL * PIXEL_PER_MEMORY_WORD)-1:0] i;
						mem_writeData <= mem_readData;
						
						for (i = 0; i < BIT_PER_MEMORY_PIXEL; i = i + 1)
							mem_writeData[(coord_x % PIXEL_PER_MEMORY_WORD) * BIT_PER_MEMORY_PIXEL + i] <= color[i];
						
						mem_writeEnable <= 1;
						memoryWriteState <= 3;
					end
				3:	begin : waitWriteFinish
						ready <= 1;
						mem_writeEnable <= 0;
						memoryWriteState <= 0;
					end

				default: memoryWriteState <= 0;
			endcase
		end
	end
endmodule
