/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Create Date:    12:24:00 10/29/2008 
// Module Name:    display_memory_select 
// Project Name:   Display-Controller fuer SpartanMC
//
// Description:    Vereinigt viele FPGA-interne 9-Bit Block-RAMs zu einem grossen
//
//////////////////////////////////////////////////////////////////////////////////

`define SYNTHESIS 1

module display_memory_select #(
		parameter USE_SSR = 1,			// SSRVAL_ Ausgabewerte bei Deaktivierung nutzen, sonst bleibe Ausgang bei Deaktivierung auf vorherigen Daten stehen
		parameter DATA_WORDS_REQUIRED = 2048,	// benoetigtes Datenvolumen in Speicherwoertern
		parameter BRAM_SIZE = 2048		// Groesse eines BRAMS in Speicherwoertern
	) (
		input wire	we_a,				// Write-Enable Port A
		input wire	en_a,				// Enable Port A
		input wire	clk_a,				// Clock Port A
		input wire [ld(DATA_WORDS_REQUIRED)-1:0] addr_a, // Adresse Port A
		input wire	[8:0] di_a,			// DataIn Port A
		output wire	[8:0] do_a,			// DataOut Port A

		input wire	we_b,
		input wire	en_b,
		input wire	clk_b,
		input wire [ld(DATA_WORDS_REQUIRED)-1:0] addr_b,
		input wire	[8:0]	di_b,
		output wire	[8:0]	do_b
	);

	`include "display/src/ld.vh"

/*	Da bei der Division der Rest immer abgeschnitten wird, jedoch prinzipiell aufgerundet
	werden muesste, kommt auf der Ergebnis noch ein +1. Dies ist nur problematisch, wenn
	kein Rest bliebe. Daher wird DATA_WORDS_REQUIRED immer um 1 verringert, sofern es nicht
	schon 0 ist. */
	localparam REQUIRED_BRAMS = (((DATA_WORDS_REQUIRED != 0) ? DATA_WORDS_REQUIRED-1 : 0) / BRAM_SIZE) + 1;

	// Adressbreite eines Block-RAMs bei gegebener Groesse, sofern mehr als ein BRAM benoetigt werden
	localparam MSB_ADDR = (REQUIRED_BRAMS > 1) ? ld(BRAM_SIZE-1)-1 : ld(DATA_WORDS_REQUIRED)-1;

	// Gibt die Nummer des anzusprechenden BRAMs an
	wire [ld(REQUIRED_BRAMS)-1:0] bram_number_a;
	wire [ld(REQUIRED_BRAMS)-1:0] bram_number_b;
	wire [REQUIRED_BRAMS-1:0] bram_select_a;
	wire [REQUIRED_BRAMS-1:0] bram_select_b;

	generate
		if (REQUIRED_BRAMS > 1)
		begin
			assign bram_number_a = addr_a[MSB_ADDR+1 +: ld(REQUIRED_BRAMS-1)];
			assign bram_number_b = addr_b[MSB_ADDR+1 +: ld(REQUIRED_BRAMS-1)];
		end
		else begin
			assign bram_number_a = 0;
			assign bram_number_b = 0;
		end
	endgenerate	
	
	// Ausgangsleitungen der BRAMs
	wire [8:0] bram_do_a [0:REQUIRED_BRAMS-1];
	wire [8:0] bram_do_b [0:REQUIRED_BRAMS-1];

	// Zuweisung der Ausgabe
	assign do_a = bram_do_a[bram_number_a];
	assign do_b = bram_do_b[bram_number_b];

	wire [10:0] bramaddr_a = addr_a[MSB_ADDR:0];
	wire [10:0] bramaddr_b = addr_b[MSB_ADDR:0];

	generate
		genvar i;
		for (i = 0; i < REQUIRED_BRAMS; i = i + 1)
		begin : BRAM_ARRAY
			assign bram_select_a[i] = (bram_number_a == i);
			assign bram_select_b[i] = (bram_number_b == i);
		`ifdef SYNTHESIS
			ram_9_prim BRAM (
					.WEA(we_a && bram_select_a[i]),
 					.ENA(en_a && bram_select_a[i]),
					.SSRA(1'b0),
					.CLKA(clk_a),
					.ADDRA(bramaddr_a),
					.DIA(di_a[8:0]),
					.DOA(bram_do_a[i] [8:0]),
	
					.WEB(we_b && bram_select_b[i]),
 					.ENB(en_b && bram_select_b[i]),
					.SSRB(1'b0),
					.CLKB(clk_b),
					.ADDRB(bramaddr_b),
					.DIB(di_b[8:0]),
					.DOB(bram_do_b[i] [8:0])
				);

		`else
			// Speicher durch ein entsprechend grosses Registers emulieren
			reg [8:0] data[0:BRAM_SIZE-1];		// Speicher
			wire ena = en_a && bram_select_a[i];	// Enable-Leitung fuer diesen Speicher
			wire enb = en_b && bram_select_b[i];
			wire wea = we_a && bram_select_a[i];	// Write-Enable-Leitung
			wire web = we_b && bram_select_b[i];
			assign bram_do_a[i] = ena ? data[addr_a[MSB_ADDR-1:0]] : 0;
			assign bram_do_b[i] = enb ? data[addr_b[MSB_ADDR-1:0]] : 0;

			// Schreibzugriff auf den Speicher (von beiden Seiten aus)
			always @(posedge clk_a or posedge clk_b)
			begin
				if (clk_a && wea)
					data[addr_a[MSB_ADDR-1:0]] <= di_a;
				if (clk_b && web)
					data[addr_b[MSB_ADDR-1:0]] <= di_b;
					
			end
			
			// Initialisierung
			initial begin : memReset
				integer i;
				for (i = 0; i < BRAM_SIZE; i = i + 1)
					data[i] = 0;
			end
		`endif

		end
	endgenerate

endmodule
