/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */



`timescale      1ns / 1ps
`default_nettype none

module dvi(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	
	clk_pixel,
	dvi_data,
	xclk_n,
	xclk_p,
	clk_test,
	data_en,
	h_sync,
	v_sync,

      data_from_mcb,
      data_to_mcb,
	debug

	);


// Die Parameter fuer die Adressen und die Portbreite werden 
// in der system.v durch jConfig eingestellt.
parameter BASE_ADR	= 10'h0;

parameter CONTROL_ADR	= 0; 

parameter ALL_PORTS	= 1;
parameter DECODSIM_DEB	= "NO";

// inputs & outputs des Systembus
input	wire				clk_peri;
input	wire	[18:0]	do_peri;	// Datenbus 
output	wire	[17:0]			di_peri;	// Datenbus
input	wire	[9:0]			addr_peri;	// Adressbus
input	wire				access_peri;	// chipselect Peripherie
input	wire				wr_peri;	// Peripherie-Schreibzugriff
input	wire				reset;
// Signale des Ports
output	wire	[11:0]	dvi_data;		// Ausgangspins
	
input wire clk_pixel;
output wire 				xclk_n;
output wire 				xclk_p;
output wire					clk_test;
output reg 				h_sync;
output reg					v_sync;
output wire 					data_en;



input   wire [56:0] data_from_mcb;
output  wire [78:0] data_to_mcb;


reg enable;

output wire [127:0] debug;


parameter H_VISIBLE = 1280;
parameter H_FRONT_PORCH = 48;
parameter H_SYNC = 112;
parameter H_BACK_PORCH = 248;

parameter V_VISIBLE = 1024;
parameter V_FRONT_PORCH = 1;
parameter V_SYNC = 3;
parameter V_BACK_PORCH = 38;

parameter H_SYNC_INVERT = 0;
parameter V_SYNC_INVERT = 0;

parameter COLOR_MODE = "RGB_24";


localparam DATA_WIDTH = COLOR_MODE=="RGB_24"? 24 : 16;



// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					cs_control_r;	// Lesen
wire					cs_control_w;	// Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder #(
		.REG_COUNT   ( ALL_PORTS    ),
		.BASE_ADR    ( BASE_ADR     ),
		.SIM_DEBUG   ( DECODSIM_DEB )
	) decod1 (
		.reset       ( reset        ),
		.clk_peri    ( clk_peri     ),
		.addr_peri   ( addr_peri    ),
		.access_peri ( access_peri  ),
		.wr_peri     ( wr_peri      ),

		.peri_select ( select       ),
		.reg_select  ( reg_select   ),
		.reg_read    ( reg_read	    ),
		.reg_write   ( reg_write    )
);



// chipselect fuer die FIFO-,Status- und Steuerregister
assign	cs_control_r	= reg_read[CONTROL_ADR];	// Lesen
assign	cs_control_w	= reg_write[CONTROL_ADR];	// Schreiben

// Dekodierung der Adressen ********************** END **************************************


always @ (posedge clk_peri)
if(reset)
	begin
	enable <= 0;
	end
else
begin	
	if(cs_control_w)				
		enable <= do_peri[0];
end


wire [2:0] mem_error;
reg[2:0] mem_error2;
always@(posedge clk_peri) mem_error2 <= mem_error;
// lesen der Ports
assign	di_peri [17:0]	= (cs_control_r)?	{14'b0, mem_error2,enable}: 18'b0;

//We don't want any metastability on the enable signal when it crosses
//clock domains. Therefore we buffer it a few times to eliminate any.
reg[3:0] enable_buf;
wire enable_pix;
always @(posedge clk_pixel) begin
	enable_buf[2:0] <= enable_buf[3:1];
	enable_buf[3] <= enable;
end
assign enable_pix = enable_buf[0];


//Clock generation
//Use ODDR2 to output clock

oddr2_prim ODDR2_xclk_p (
	.D0(enable_pix),
	.D1(1'b0),
	.C0(clk_pixel),
	.C1(~clk_pixel),
	.Q(xclk_p),
	.R(1'b0),
	.S(1'b0),
	.CE()
);


oddr2_prim ODDR2_clk_test (
	.D0(enable_pix),
	.D1(1'b0),
	.C0(clk_pixel),
	.C1(~clk_pixel),
	.Q(clk_test),
	.R(1'b0),
	.S(1'b0),
	.CE()
);

oddr2_prim ODDR2_xclk_n (
	.D0(1'b0),
	.D1(enable_pix),
	.C0(clk_pixel),
	.C1(~clk_pixel),
	.Q(xclk_n),
	.R(1'b0),
	.S(1'b0),
	.CE()
);

//Displayed pixel data
wire [DATA_WIDTH-1:0] pixel_data;

//DDR data outputs
generate 

	genvar i;
	for (i=0;i<(DATA_WIDTH/2);i=i+1) begin:oddr2
	
		oddr2_prim #(
			.DDR_ALIGNMENT("C1"),
			.INIT(0),
			.SRTYPE("ASYNC")
		) ODDR2_data (
			.D0(pixel_data[i]),
			.D1(pixel_data[i+DATA_WIDTH/2]),
			.C0(clk_pixel),
			.C1(~clk_pixel),
			.Q(dvi_data[i]),
			.R(1'b0),
			.S(1'b0),
			.CE()
		);
	
	end
	
	if (COLOR_MODE=="YCRCB_16") begin
		assign dvi_data[8]=0;
		assign dvi_data[9]=0;
		assign dvi_data[10]=0;
		assign dvi_data[11]=0;
	end

endgenerate


localparam H_TOTAL = H_BACK_PORCH + H_SYNC + H_FRONT_PORCH + H_VISIBLE;
localparam V_TOTAL = V_BACK_PORCH + V_SYNC + V_FRONT_PORCH + V_VISIBLE;
localparam H_BITS  = clog2(H_TOTAL);
localparam V_BITS  = clog2(V_TOTAL);

reg [H_BITS-1:0] current_x;
reg [V_BITS-1:0] current_y;

reg [H_BITS-1:0] next_x;
reg [V_BITS-1:0] next_y;



function integer clog2;
	input integer value;
	begin
		value = value-1;
		for (clog2=0; value>0; clog2=clog2+1)
			value = value>>1;
	end
endfunction



always @(posedge clk_pixel) begin
	if (reset) begin
		h_sync <= H_SYNC_INVERT;
		v_sync <= V_SYNC_INVERT;
	end else begin
		h_sync <= H_SYNC_INVERT ^ ((current_x >= (H_VISIBLE+H_FRONT_PORCH)) && (current_x < (H_VISIBLE+H_FRONT_PORCH+H_SYNC)));
		v_sync <= V_SYNC_INVERT ^ ((current_y >=  (V_VISIBLE+V_FRONT_PORCH)) && (current_y  < (V_VISIBLE+V_FRONT_PORCH+V_SYNC)));
	end
end


//Delay image transmission because data has not yet been read from dram (important during startup)
wire stall;

always @(posedge clk_pixel) begin
	if(reset || ~enable_pix) begin
		current_x <= H_TOTAL-1;
		current_y <= V_TOTAL-1;
	end else begin
		if (~stall) begin
			current_x <= next_x;
			current_y <= next_y;
		end
	end
end

always @* begin
	if(current_x == H_TOTAL-1) begin 
		next_x <= 0;
		
		if(current_y == V_TOTAL-1) begin 
			next_y <= 0;
		end else begin 
			next_y <= current_y +1;
		end
		
	end else begin 
		next_x <= current_x +1;
		next_y <= current_y;
	end
end



mem_access #(
		.X_BITS        ( H_BITS        ),
		.Y_BITS        ( V_BITS        ),
		.X_VISIBLE     ( H_VISIBLE     ),
		.Y_VISIBLE     ( V_VISIBLE     ),
		.X_TOTAL       ( H_TOTAL       ),
		.Y_TOTAL       ( V_TOTAL       ),
		.COLOR_MODE    ( COLOR_MODE    )
    ) mem_access_inst (

		.clk_pixel     ( clk_pixel     ),
		.reset         ( reset         ),
		.stall         ( stall         ),

		.enable        ( enable_pix    ),

      		.data_from_mcb ( data_from_mcb ),
		.data_to_mcb   ( data_to_mcb   ),

		.pixel_data    ( pixel_data    ),
		.data_en       ( data_en       ),

		.current_x     ( current_x     ),
		.current_y     ( current_y     ),

		.next_x        ( next_x        ),
		.next_y        ( next_y        ),
		.error         ( mem_error     ),
		.debug         ( debug         )
);



endmodule
