/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */



`timescale      1ns / 1ps
`default_nettype none

module mem_access #(
	parameter X_BITS    = 10,
	parameter Y_BITS    = 10,
	parameter X_VISIBLE = 640,
	parameter Y_VISIBLE = 480,
	parameter X_TOTAL   = 800,
	parameter Y_TOTAL   = 600,
	parameter BURST_SIZE = 8, //X_VISIBLE * Y_VISIBLE / (COLOR_MODE="RGB_24"?1:2) needs to be a multiple of burst size
	parameter START_ADDRESS = 0,
	parameter LIMIT_SIZE = 2,
	parameter COLOR_MODE = "RGB_24"
)(
	
	clk_pixel,
	reset,
	enable,

    data_from_mcb,
    data_to_mcb,

	pixel_data,
	data_en,

	current_x,
	current_y,

	next_x,
	next_y,

	stall,

	error,
	debug

	);
	

localparam DATA_WIDTH = COLOR_MODE=="RGB_24"? 24 : 16;


input   wire clk_pixel;
input   wire reset;
input   wire enable;
input   wire [56:0]       data_from_mcb;
output  wire [78:0]       data_to_mcb;

output  reg  [DATA_WIDTH-1:0]       pixel_data;
output  reg               data_en;
input   wire [X_BITS-1:0] current_x;
input   wire [Y_BITS-1:0] current_y;
input   wire [X_BITS-1:0] next_x;
input   wire [Y_BITS-1:0] next_y;

output reg [2:0] error;
output wire[127:0] debug;

output wire stall;


wire            cmd_empty;
wire            cmd_full;
wire            calib_done;

wire            wr_full;
wire            wr_empty;
wire    [6:0]   wr_count;
wire            wr_underrun;
wire            wr_error;


wire    [31:0]  rd_data;
wire            rd_full;
wire            rd_empty;
wire    [6:0]   rd_count;
wire            rd_overflow;
wire            rd_error;

assign cmd_empty   = data_from_mcb[0];
assign cmd_full    = data_from_mcb[1];
assign wr_count    = data_from_mcb[8:2];
assign wr_full     = data_from_mcb[9];
assign wr_empty    = data_from_mcb[10];
assign wr_underrun = data_from_mcb[11];
assign wr_error    = data_from_mcb[12];
assign rd_full     = data_from_mcb[13];
assign rd_empty    = data_from_mcb[14];
assign rd_count    = data_from_mcb[21:15];
assign rd_overflow = data_from_mcb[22];
assign rd_error    = data_from_mcb[23];
assign calib_done  = data_from_mcb[24];
assign rd_data     = data_from_mcb[56:25];

wire         mem_clk;
reg         cmd_en;
wire  [2:0]   cmd_instr;
wire  [5:0]   cmd_bl;
reg [29:0]  cmd_byte_addr;
wire         wr_en;
reg         rd_en;
wire [3:0]   wr_mask;
wire [31:0]  wr_data;

assign data_to_mcb[0]                   = mem_clk;
assign data_to_mcb[1]                   = cmd_en;
assign data_to_mcb[4:2]                 = cmd_instr;
assign data_to_mcb[11:5]                = cmd_bl;
assign data_to_mcb[40:12]               = cmd_byte_addr;
assign data_to_mcb[41]                  = wr_en;
assign data_to_mcb[42]                  = rd_en;
assign data_to_mcb[46:43]               = wr_mask;
assign data_to_mcb[78:47]               = wr_data;

assign mem_clk = clk_pixel;


assign wr_en = 0;
assign wr_mask = 0;
assign wr_data = 0;

//Count how many words will be in the FIFO when all pending read requests are done
//  We may only issue read requests that will not overflow the fifo, but have to 
//  keep track of how many bytes are still pending ourselves.
reg[7:0] inFifo;

localparam FINISH_ADDRESS = START_ADDRESS + (X_VISIBLE*Y_VISIBLE)* (COLOR_MODE=="RGB_24"?4:2)-BURST_SIZE*4;


assign cmd_instr = 1'b1; // Read from memory
assign cmd_bl = (BURST_SIZE - 1);


reg noRead;
always @(posedge clk_pixel) 
	noRead <= reset || ~enable;



//Issue read commands when there is space left
always @(posedge clk_pixel) begin

	if (noRead) begin
		cmd_en <= 0;
		inFifo <= 0;
		cmd_byte_addr <= START_ADDRESS;
	end else begin: blah

		reg[7:0] inFifoTemp;
		inFifoTemp = inFifo;

		if (rd_en)
			inFifoTemp = inFifoTemp - 1;

		if (inFifoTemp<=(64-BURST_SIZE) && ~cmd_full && ~cmd_en) begin
			cmd_en <= 1;
			inFifoTemp = inFifoTemp + BURST_SIZE;
		end else begin
			cmd_en <= 0;
		end

		inFifo <= inFifoTemp;

		if (cmd_en) begin
			if (cmd_byte_addr == FINISH_ADDRESS)
				cmd_byte_addr <= START_ADDRESS; //TODO switch to other frame here
			else
				//address is in bytes, burst size in words --> *4
				cmd_byte_addr <= cmd_byte_addr + BURST_SIZE * 4;
		end
	end
end

//Only used for ycrcb
reg isSecondWord = 1;

wire h_in_visible;
wire v_in_visible;
assign h_in_visible = next_x < X_VISIBLE;
assign v_in_visible = next_y < Y_VISIBLE;

wire next_data_en;
assign next_data_en = h_in_visible && v_in_visible;


assign stall = next_data_en && rd_empty;


reg data_en_2;
always @(posedge clk_pixel) begin
	if (noRead) begin
		data_en <= 0;
		data_en_2 <= 0;
	end else
		if (~stall) begin
			data_en_2 <= next_data_en;
			data_en <= data_en_2;

		end
end

//Read data from fifo 
always @(posedge clk_pixel) begin

	if (reset) begin
		rd_en <= 0;
		isSecondWord <= 1;
	end else if (~enable && calib_done) begin
		//If enable was switched off during frame, drain fifo
		//hopefully enable doesn't get switched on while the fifo is not yet empty ;)
		rd_en <= ~rd_empty;
	end else begin
		if (COLOR_MODE=="YCRCB_16") begin
			if (~stall)
				isSecondWord <= ~ isSecondWord;
		end
		if (~stall && next_data_en && ~isSecondWord)
			rd_en <= 1;
		else
			rd_en <= 0;
	end

end

always @(posedge clk_pixel) begin
	if (COLOR_MODE=="RGB_24")
		pixel_data <= stall? 24'hFFFFFF : rd_data[23:0];
	else
		if (isSecondWord)
			pixel_data <= stall? 16'hFFFF : rd_data[15:0];
		else
			pixel_data <= stall? 16'hFFFF : rd_data[31:16];
end

//error detection
always @(posedge clk_pixel) begin
	if (noRead) begin
		error <= 0;
	end else begin
		error <= error | {rd_error, rd_overflow, stall && data_en };
	end
end

/*
assign debug[2:0] = error;
assign debug[3] = enable;
assign debug[4] = data_en;
assign debug[16:5] = current_x;
assign debug[26:17] = current_y;
assign debug[27] = stall;
assign debug[34:28] = rd_count;
assign debug[35] = rd_empty;
assign debug[36] = rd_en;
assign debug[37] = rd_overflow;
assign debug[38] = rd_error;
assign debug[39] = cmd_empty;
assign debug[40] = cmd_full;
assign debug[41] = cmd_en;
assign debug[65:42] = pixel_data;
assign debug[72:66] = inFifo;
assign debug[73] = first_read;
assign debug[74] = higher_limit;*/
assign debug[127:0] = 0;


endmodule
