/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`timescale 1ns / 1ns
//////////////////////////////////////////////////////////////////////////////////
// Company: 
// Engineer: 
// 
// Create Date:    14:23:31 12/29/2010 
// Design Name: 
// Module Name:    Mdio 
// Project Name: 
// Target Devices: 
// Tool versions: 
// Description: 
//
// Dependencies: 
//
// Revision: 
// Revision 0.01 - File Created
// Additional Comments: 
//
//////////////////////////////////////////////////////////////////////////////////
`default_nettype none
module MdioModule #(
	parameter WIDTH=1
) (
	input wire Clk,
	input wire Reset,
	output wire Mdc,
	inout wire [WIDTH-1:0] Mdio,
	//control signals
	input wire Start,
	input wire WriteEnable,
	input wire [4:0] PhyAddr,
	input wire [4:0] RegAddr,
	input wire [15:0] Data,
	output wire [15:0] Result,
	output wire Ready,
	output wire mdioOut,
	output reg [4:0] cnt,
	output wire Mdio_actual
);

//parameters
// clock divider to generate Mdc
parameter MDC_DIVIDER = 4;
// after each frame a gab is required, the length differs between PHY
// 32 bit or shorter, the DP83849 from national need only a 1 bit gab 
parameter FRAME_GAB = 1;

//states
parameter M_BLOCK 	= 3'b000,
			 M_IDLE		= 3'b010,
			 M_SENDING	= 3'b100;

// local signals
reg localStart;
reg [2:0] state;
reg [15:0] buffer;
reg [MDC_DIVIDER-1:0] mdcCnt;
reg write;
reg [4:0] phyA;
reg [4:0] regA;

assign Mdc = mdcCnt[MDC_DIVIDER-1];
assign Ready = state == M_IDLE;
assign Result = buffer;

assign mdioOut = // start
					(cnt == 0) ? 1'b0 :
					(cnt == 1) ? 1'b1 :
					// opcode
					(cnt == 2) ? ~write :
					(cnt == 3) ? write :
					//phy address
					(cnt == 4) ? phyA[4] :
					(cnt == 5) ? phyA[3] :
					(cnt == 6) ? phyA[2] :
					(cnt == 7) ? phyA[1] :
					(cnt == 8) ? phyA[0] :
					//reg address
					(cnt < 14) ? regA[4] :
					//turnaround
					(cnt == 14) ? 1'b0 :
					(cnt == 15) ? 1'b0 :
					//data
					 buffer[15] ;
					
wire reading;
assign reading = ~write && cnt>=14;

generate
	genvar i;
	for (i=0;i<WIDTH;i=i+1) begin:tristrate

		assign Mdio[i] = (state != M_SENDING || reading) ? 1'bZ :
					mdioOut;

	end
endgenerate

//Only one PHY will pull down their line, all other are kept high by pullup -> and together
assign Mdio_actual = &Mdio;

always@(posedge Clk or posedge Reset)
begin
	if(Reset)
	begin
		localStart <= 0;
		state <= M_BLOCK;
		buffer <= 0;
		mdcCnt <= 0;
		cnt <= 31;
	end
	else
	begin
		mdcCnt <= mdcCnt + 1;
		//latch input signals
		if(state == M_IDLE && Start && !localStart)
		begin
			localStart <= 1;	//buffer start signal
			write <= WriteEnable;
			buffer <= Data;
			phyA <= PhyAddr;
			regA <= RegAddr;
		end
		
		if(mdcCnt[MDC_DIVIDER-1] == 1 && mdcCnt[MDC_DIVIDER-2:0] == 0)
		begin
			case(state)
			M_BLOCK:
			begin
				if(cnt == 0) state <= M_IDLE;
				else cnt <= cnt - 1;
			end
			M_IDLE:
			begin
				if(localStart)
				begin
					localStart <= 0;
					state <= M_SENDING;
					//input signals already sampled
				end
			end
			M_SENDING:
			begin
				cnt <= cnt + 1;
				
				//shift register address
				if(cnt > 8 && cnt < 14) regA <= {regA[3:0],regA[4]};
				//shift data buffer
				if(cnt > 15)
				begin
					if(write == 1)buffer <= {buffer[14:0],1'b0};
					//sample mdio at this time is too late, spec only garantee value 10 ns after posedge of mdc
				end
				
				if(cnt == 31)
				begin
					cnt <= FRAME_GAB;
					state <= M_BLOCK;
				end
			end
			endcase

		end	//end if mdcCnt
		
		//maybe cnt > 15 can be omitted
		if(mdcCnt == 4'b0111 && state == M_SENDING && write == 0  && cnt > 15)
		begin
			buffer <= {buffer[14:0], Mdio_actual};
		end
	end
end

endmodule
