/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none

module heartbeat (
    input wire              clk_peri,
    input wire              reset,
    input wire  [9:0]       addr_peri,
    input wire              access_peri,
    input wire              wr_peri,
    input wire  [17:0]      do_peri,
    output wire [17:0]      di_peri,
    output wire             heartbeat_o,
    
    input wire              reset_ext_i,
    input wire              clk_ext_i,
    input wire              sysclock_locked_i,
    input wire              uart_tx_i
);

parameter
    CLOCK_EXT_FREQ_HZ   = 50000000,
    BASE_ADR            = 10'h0,
    HEARTBEAT_PATTERN   = 18'h0,
    RESET_EXT_LEVEL     = "LOW_ACTIVE",
    CHECK_UART_TX       = "NO",
    SIM_DEBUG           = "NO";

localparam
    CYCLES_MAX          = CLOCK_EXT_FREQ_HZ / 10 * 7,
    CYCLES_ON           = CLOCK_EXT_FREQ_HZ / 6;


reg     [31:0]  counter;
reg             heartbeat;

wire            counter_reset;
wire            counter_en;
wire            reg_write;

wire            startup0_led_error_o;
wire            startup0_all_passed_o;

reg_access_decoder	#(
		.REG_COUNT(	1	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	SIM_DEBUG	)
			)
reg_access_decoder0			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(),
		.reg_select(),
		.reg_read(),
		.reg_write(	reg_write	)
);

startup startup0 (
    .reset_ext_i(reset_ext_i),
    .clk_ext_i(clk_ext_i),
    .sysclock_locked_i(sysclock_locked_i),
    .spmc_reset_i(reset),
    .spmc_clk_peri_i(clk_peri),
    .uart_tx_i(uart_tx_i),
    .led_error_o(startup0_led_error_o),
    .all_passed_o(startup0_all_passed_o)
);
defparam 
    startup0.CLOCK_EXT_FREQ_HZ  = CLOCK_EXT_FREQ_HZ,
    startup0.CHECK_SPMC_RESET   = "YES",
    startup0.CHECK_UART_TX      = CHECK_UART_TX,
    startup0.RESET_EXT_LEVEL    = RESET_EXT_LEVEL;

assign counter_en       = (counter < CYCLES_MAX);
assign counter_reset    = reset || (heartbeat && ~counter_en);
assign heartbeat_o      = startup0_all_passed_o ? (counter < CYCLES_ON) : startup0_led_error_o;

assign di_peri          = 18'b0;

always @ (posedge clk_peri or posedge reset)
begin
    if (reset)
    begin
        heartbeat <= 1'b0;
    end else begin
        heartbeat <= (reg_write & (do_peri == HEARTBEAT_PATTERN));
    end
end

always @ (posedge clk_ext_i or posedge counter_reset)
begin
    if (counter_reset)
    begin
        counter <= 32'b0;
    end else begin
        if (counter_en) 
        begin
            counter <= counter + 1;
        end
    end
end

endmodule
