/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`ifndef __startup
`define __startup
module startup(
    input wire              reset_ext_i,
    input wire              clk_ext_i,
    input wire              sysclock_locked_i,
    input wire              spmc_reset_i,
    input wire              spmc_clk_peri_i,
    input wire              uart_tx_i,

    output wire             all_passed_o,
    output wire             led_error_o
);

parameter
    CHECK_UART_TX       = "NO",
    CHECK_SPMC_RESET    = "NO",
    RESET_EXT_LEVEL     = "LOW_ACTIVE",
    CLOCK_EXT_FREQ_HZ   = 50000000;


localparam
    NUM_TESTS       = 6;

// generate reset pulse on power up
reg                     reset_int = 1'b1;
reg                     reset_ext_last;
wire [NUM_TESTS-1:0]    passed;
reg  [2:0]              passed_count;
wire                    led_encoder0_led_o;

assign all_passed_o     = passed[NUM_TESTS-1:0] == {NUM_TESTS{1'b1}};
assign led_error_o            = all_passed_o ? 1'b0 : led_encoder0_led_o;

// generate internal reset pulse on any
// transition of external reset signal
always @ (posedge clk_ext_i) 
begin
    reset_ext_last  <= reset_ext_i;
    reset_int       <= (reset_ext_last != reset_ext_i);
end

// check external clock
check_clock check_clock_ext0 (
    .clk_i(clk_ext_i),
    .reset_i(reset_int),
    .passed_o(passed[0])
);
defparam check_clock_ext0.CYCLES = 256;

// check release of external reset
generate
  if (RESET_EXT_LEVEL == "LOW_ACTIVE")
    assign passed[1] = reset_ext_i;
  else
    assign passed[1] = ~reset_ext_i;
endgenerate

// check syslock locked signal
assign passed[2] = sysclock_locked_i;

//check_transition check_transition_locked0 (
//    .clk_i(clk_ext_i),
//    .reset_i(reset_int),
//    .signal_i(sysclock_locked_i),
//    .passed_o(passed[2])
//);
//defparam check_transition_locked0.TRANSITION = "RISING";

generate
if (CHECK_SPMC_RESET=="YES")
begin
// check release of spartanmc generated 
// reset
// check_transition check_transition_reset0 (
//    .clk_i(clk_ext_i),
//    .reset_i(reset_int),
//    .signal_i(spmc_reset_i),
//    .passed_o(passed[3])
//);
//defparam check_transition_reset0.TRANSITION = "FALLING";
    assign passed[3] = ~spmc_reset_i;

end else begin

    assign passed[3] = 1'b1;

end
endgenerate

// check spartanmc peripheral clock
check_clock check_clock_peri0 (
    .clk_i(spmc_clk_peri_i),
    .reset_i(reset_int),
    .passed_o(passed[4])
);
defparam check_clock_peri0.CYCLES = 256;

generate
if (CHECK_UART_TX=="YES")
begin

// check for activity on uart tx line 
check_clock check_uart0 (
    .clk_i(uart_tx_i),
    .reset_i(reset_int),
    .passed_o(passed[5])
);
defparam check_uart0.CYCLES = 32;

end else begin

    assign passed[5] = 1'b1;

end
endgenerate

// priority encoder for passed
// tests
integer i;
always @(*)
begin
    passed_count <= NUM_TESTS;
    for (i=(NUM_TESTS-1); i>=0; i=i-1)
    begin
        if (~passed[i])
        begin
            passed_count <= i % 4'b1000;
        end
    end
end

// encoder for blinking led
led_encoder led_encoder0 (
    .clk_i(clk_ext_i),
    .reset_i(reset_int),
    .value_i(passed_count),
    .led_o(led_encoder0_led_o)
);
defparam 
    led_encoder0.BIT_COUNT          = 3,
    led_encoder0.CLOCK_FREQ_HZ      = CLOCK_EXT_FREQ_HZ;

//synthesis translate_off
//defparam led_encoder0.CLOCK_FREQ_HZ = 2500;
//synthesis translate_on

endmodule
`endif
