/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none

module accelerator_stub #(
    parameter BASE_ADR          = 18'h1A000,
    parameter DMA_BASE_ADR      = 18'h08000,
    parameter DECODSIM_DEB      = "NO",     //DEBUG mode YES oder NO
    parameter RAMBLOCKS         = 2,
    parameter HW_ACCELERATOR    = 1'b1,
    parameter CACHES            = 1,
    parameter SHOW_MEM_ACCESSES = 0
) (
    input  wire        clk_peri,            //Systemtakt
    input  wire [17:0] do_peri,             //Datenbus  vom MC
    output wire [17:0] di_peri,             //Datenbus  zum MC
    input  wire [9:0]  addr_peri,           //Adressbus vom MC
    input  wire        access_peri,         //Signale fuer Peripheriezugriffe
    input  wire        wr_peri,             //Schreibsignal
    input  wire        mem_access_dma,
    input  wire [7:0]  addr_high,
    input  wire        store_access_high,   //DMA-WE for upper spmc-byte
    input  wire        store_access_low,    //DMA-WE for lower spmc-byte
    input  wire        reset,               //reset
    output wire        cache_clk,
    output wire        cache_reset,
    output wire [CACHES-1:0] mem_access,
    output wire [CACHES-1:0] mem_wr_high,
    output wire [CACHES-1:0] mem_wr_low,
    input  wire [CACHES-1:0] mem_busy,
    input  wire [CACHES-1:0] mem_busy_wr,
    output wire [CACHES*10-1:0] mem_addr_block,
    output wire [CACHES*8-1:0] mem_addr_high,
    output wire [CACHES*18-1:0] mem_do,
    input  wire [CACHES*18-1:0] mem_di
);

localparam [6:0] REG_HIGH   = 127;
localparam [3:0] REG_ABITS  = 7;

wire [CACHES*18-1:0]          acc2mem_data;
wire [CACHES*18-1:0]          mem2acc_data;
reg  [CACHES*18-1:0]          align_mem2acc_data;
reg  [CACHES*18-1:0]          align_acc2mem_data;
wire [CACHES*18-1:0]          acc2mem_addr;
reg  [CACHES*18-1:0]          acc2mem_addr_reg; // debugging only
reg  [CACHES-1:0]             acc_wr_req_start;
wire                 peri_select;
wire [REG_ABITS-1:0] reg_addr;
reg  [CACHES*2-1:0]           align_acc2mem_we;
wire [CACHES*2-1:0]           acc2mem_wr_req;
wire [CACHES*2-1:0]           acc2mem_rd_req;
reg  [CACHES*2-1:0]           acc2mem_rd_req_reg;
wire [CACHES*3-1:0]           acc2mem_align_ctrl_rd;
wire [CACHES*3-1:0]           acc2mem_align_ctrl_wr;
reg  [CACHES-1:0]             align_error;
reg  [CACHES-1:0]             acc2mem_lsb;

parameter [2:0] I_WR_EVE = {1'b0, 2'b10};
parameter [2:0] C_WR_EVE = {1'b0, 2'b01}; 
parameter [2:0] N_WR_EVE = {1'b0, 2'b00};  
parameter [2:0] I_WR_ODD = {1'b1, 2'b10}; 
parameter [2:0] C_WR_ODD = {1'b1, 2'b01}; 
parameter [2:0] N_WR_ODD = {1'b1, 2'b00};


parameter [2:0] I_RD_EVE = {1'b0, 2'b10};   
parameter [2:0] C_RD_EVE = {1'b0, 2'b01};
parameter [2:0] N_RD_EVE = {1'b1, 2'b00}; 
parameter [2:0] I_RD_ODD = {1'b1, 2'b10}; 
parameter [2:0] C_RD_ODD = {1'b1, 2'b01}; 
parameter [2:0] N_RD_ODD = {1'b1, 2'b00}; 

genvar i;
generate
for (i = 0; i < CACHES; i = i + 1) begin
assign acc2mem_align_ctrl_rd[3*(i+1)-1:3*i] = {acc2mem_lsb[i], acc2mem_rd_req_reg[2*(i+1)-1:2*i]};
assign acc2mem_align_ctrl_wr[3*(i+1)-1:3*i] = {acc2mem_addr[18*i], acc2mem_wr_req[2*(i+1)-1:2*i]};

// store lsb for one cycle
always @(posedge clk_peri) begin
    if(|acc2mem_rd_req[2*(i+1)-1:2*i]) begin
        // SpartanMC memory bus does not require valid address during mem operation, but alignment logic does
        acc2mem_lsb[i] <= acc2mem_addr[18*i];
    end
    if(mem_access[i]) begin
        // keep addr for debugging
        acc2mem_addr_reg[18*(i+1)-1:18*i] <= acc2mem_addr[18*(i+1)-1:18*i];
    end
    if(~|mem_busy) begin
        // don't refresh if at least one cache is busy, does not allow interleaved memory requests
        acc2mem_rd_req_reg[2*(i+1)-1:2*i] <= acc2mem_rd_req[2*(i+1)-1:2*i];
    end
end

/* 
 * Alignment logic for SpartanMC  
 * - Generates an error for unaligned (addr == odd) integer-accesses.
 *   (This should never happen with a correct compiler-toolchain.)
 * - Read/write the upper byte for odd byte accesses.
 */
always@(*) begin
    align_acc2mem_we[2*(i+1)-1:2*i] <= 2'b0;
    align_error[i] <= 1'b0;
    align_mem2acc_data[18*(i+1)-1:18*i] <= 18'h00000;
    align_acc2mem_data[18*(i+1)-1:18*i] <= 18'h00000;
    //WRITE
    case (acc2mem_align_ctrl_wr[3*(i+1)-1:3*i])
        I_WR_EVE: begin
                     align_acc2mem_we[2*(i+1)-1:2*i] <= 2'b11;
                     align_acc2mem_data[18*(i+1)-1:18*i] <= acc2mem_data[18*(i+1)-1:18*i];
                  end
        C_WR_EVE: begin
                     align_acc2mem_we[2*(i+1)-1:2*i] <= 2'b10;
                     align_acc2mem_data[18*(i+1)-1:18*i] <= acc2mem_data[18*(i+1)-1:18*i] << 9;
                  end
        N_WR_EVE: begin
                     align_acc2mem_we[2*(i+1)-1:2*i] <= 2'b0;
                     align_acc2mem_data[18*(i+1)-1:18*i] <= 18'h00000;
                  end

        I_WR_ODD: begin
                     align_error[i] <= 1'b1;
                     align_acc2mem_data[18*(i+1)-1:18*i] <= acc2mem_data[18*(i+1)-1:18*i];
                  end
        C_WR_ODD: begin
                     align_acc2mem_we[2*(i+1)-1:2*i] <= 2'b01;
                     align_acc2mem_data[18*(i+1)-1:18*i] <= acc2mem_data[18*(i+1)-1:18*i] & 18'h1FF;
                  end
        N_WR_ODD: begin
                     align_acc2mem_we[2*(i+1)-1:2*i] <= 2'b0;
                     align_acc2mem_data[18*(i+1)-1:18*i] <= 18'h00000;
                  end
    endcase
    
    //READ
    case (acc2mem_align_ctrl_rd[3*(i+1)-1:3*i])
        I_RD_EVE: begin align_mem2acc_data[18*(i+1)-1:18*i] <= mem2acc_data[18*(i+1)-1:18*i]; end
        C_RD_EVE: begin align_mem2acc_data[18*(i+1)-1:18*i] <= (mem2acc_data[18*(i+1)-1:18*i] & 18'h3FE00) >> 9; end
        N_RD_EVE: begin align_mem2acc_data[18*(i+1)-1:18*i] <= 18'h00000; end
        I_RD_ODD: begin align_error[i] <= 1'b1; end
        C_RD_ODD: begin align_mem2acc_data[18*(i+1)-1:18*i] <= mem2acc_data[18*(i+1)-1:18*i] & 18'h1FF; end
        N_RD_ODD: begin align_mem2acc_data[18*(i+1)-1:18*i] <= 18'h00000; end

    endcase
end
end // for-loop

assign  reg_addr      = addr_peri & {REG_ABITS{1'b1}};

reg_access_decoder #(
    .REG_COUNT ( REG_HIGH + 1 ),
    .BASE_ADR  ( BASE_ADR     ),
    .SIM_DEBUG ( DECODSIM_DEB )
) reg_access_decoder0 (
    .reset       ( reset              ),
    .clk_peri    ( clk_peri           ),
    .addr_peri   ( addr_peri          ),
    .access_peri ( access_peri        ),
    .wr_peri     ( wr_peri            ),
    .peri_select ( peri_select        ),
    .reg_select  ( /* not required */ ),
    .reg_read    ( /* not required */ ),
    .reg_write   ( /* not required */ )
);

// Connect data caches
assign cache_clk = clk_peri;
assign cache_reset = reset;

for (i = 0; i < CACHES; i = i + 1) begin
	assign mem_access[i] = (|acc2mem_rd_req[2*(i+1)-1:2*i]) | (|acc2mem_wr_req[2*(i+1)-1:2*i]);
	assign mem_wr_low[i] = align_acc2mem_we[2*i];
	assign mem_wr_high[i] = align_acc2mem_we[2*i+1];
	assign mem_addr_block[10*(i+1)-1:10*i] = acc2mem_addr[18*(i+1)-8:18*i+1];
	assign mem_addr_high[8*(i+1)-1:8*i] = {1'b0, acc2mem_addr[18*(i+1)-1:18*i+11]};
	assign mem2acc_data[18*(i+1)-1:18*i] = mem_di[18*(i+1)-1:18*i];
	assign mem_do[18*(i+1)-1:18*i] = align_acc2mem_data[18*(i+1)-1:18*i];
end

if (HW_ACCELERATOR == 1'b1) begin
    acceleration_top_module acc_top (
        .bus2ip_clk_peri_i     ( clk_peri               ),
        .bus2ip_clk_acc_i      ( clk_peri               ),
        .bus2ip_reg_data_i     ( do_peri                ),
        .ip2bus_reg_data_o     ( di_peri                ),
        .bus2ip_reg_addr_i     ( reg_addr               ),
        .bus2ip_reg_re_i       ( peri_select & ~wr_peri ),
        .bus2ip_reg_we_i       ( peri_select & wr_peri  ),
        .bus2ip_reset_i        ( reset                  ),
        .acc2mem_data_o        ( acc2mem_data           ),
        .mem2acc_data_i        ( align_mem2acc_data     ),
        .acc2mem_addr_o        ( acc2mem_addr           ),
        .acc2mem_wr_req_o      ( acc2mem_wr_req         ),
        .acc2mem_rd_req_o      ( acc2mem_rd_req         ),
        .mem2acc_wr_ack_i      ( ~mem_busy_wr           ),
        .mem2acc_rd_ack_i      ( ~mem_busy              ),
        .mem2acc_align_error_i ( align_error            ),
        .store_access_low_i    ( store_access_low       ),
        .store_access_high_i   ( store_access_high      )
    );

	if(SHOW_MEM_ACCESSES) begin
	// print information about memory access
	// accelerator samples data after all caches are finished
	for (i = 0; i < CACHES; i = i + 1) begin
		always @(posedge |acc2mem_wr_req[2*i+1:2*i] or acc2mem_addr[18*(i+1)-1:18*i] or mem_do[18*(i+1)-1:18*i]) begin
			if(|acc2mem_wr_req[2*i+1:2*i])
				acc_wr_req_start[i] <= 1;
		end
		always @(posedge clk_peri) begin
			if (acc_wr_req_start[i]) begin
				// print mem write at operation request
				if(align_acc2mem_we[2*(i+1)-1] && align_acc2mem_we[2*i]) begin
					$display("[ %9d ]  WR18 ACC  mem[0x%05H] <= 0x%05H Port %2d", $time, acc2mem_addr[18*(i+1)-1:18*i], mem_do[18*(i+1)-1:18*i], i);
				end else begin
					$display("[ %9d ]  WR9 ACC  mem[0x%05H] <= 0x%05H Port %2d", $time, acc2mem_addr[18*(i+1)-1:18*i], mem_do[18*(i+1)-1:18*i], i);
				end
				acc_wr_req_start[i] <= 0;
			end
			if ((|acc2mem_rd_req_reg[2*i+1:2*i]) && ~|mem_busy) begin
				// print mem read at end of operation
				$display("[ %9d ]  RD ACC  mem[0x%05H] => 0x%05H Port %2d", $time, {1'b0, acc2mem_addr_reg[18*(i+1)-1:18*i+1]}, mem_di[18*(i+1)-1:18*i], i);
			end
		end
	end
	end // SHOW_MEM_ACCESSES

end else begin
    assign di_peri = 18'h00000;
end

endgenerate

//wire [35 : 0] control;
//
//chipscope_icon icon(
//    .CONTROL0 (control)
//);
//
//chipscope_ila ila(
//    .CONTROL (control),
//    .CLK (clk_peri),
//    .TRIG0 ({1'b0,align_acc2mem_data, acc2mem_align_ctrl, align_acc2mem_we, acc2mem_rd_req, acc2mem_wr_req, align_error, reset}),
//    .TRIG1 ({align_mem2acc_data[13:0], spmc2mem_addr}),
//    .TRIG2 ({do_peri[13:0], acc2mem_addr}),
//    .TRIG3 ({acc2mem_data[13:0], mem2acc_data})
//);

endmodule
