/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : icap.v
// Autor : Rico Backasch
// Modul : ICAP
//
// Beschreibung : Schnittstelle zum internen Konfigurationsport von FPGAs
//
///////////////////////////////////////////////////////////////////////////////

`timescale      1ns / 10ps
`default_nettype none

module icap_if(
    input	wire		clk_peri,
    input	wire [17:0]		do_peri,	// Datenbus
    output	wire [17:0]		di_peri,	// Datenbus
    input	wire [9:0]		addr_peri,	// Adressbus
    input	wire		access_peri,	// chipselect peripherie
    input	wire		wr_peri,	// signalisiert Schreibzugriff Peripherie
    input	wire		reset
);

// Die Parameter fuer die Adressen werden in system.v eingestellt.
parameter ARCHITECTURE   = "VIRTEX6";
parameter BASE_ADR	= 10'h0;

parameter CTRL_ADR	= 3'h0;
parameter DATI_ADR	= 3'h1;
parameter DATO_ADR	= 3'h2;
parameter DECODSIM_DEB	= "NO";

wire		[2:0]		reg_select;
wire		[2:0]		reg_read;
wire		[2:0]		reg_write;
wire    peri_select;
// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	3	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.addr_peri	(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri		(	wr_peri		),

		.peri_select( peri_select			),
		.reg_select	(       reg_select		),
		.reg_read	(	reg_read	),
		.reg_write	(	reg_write	)
			);
reg [1:0] ctrl_reg = 0;
wire fifo_empty;
wire fifo_full;
wire enable_reconfiguration = ctrl_reg[1];
wire enable_reconfiguration_port = ctrl_reg[0];
wire write_to_icap = enable_reconfiguration && !fifo_empty;
wire [15:0] icap_data_i;
wire [15:0] icap_data_o;
wire busy;
reg[16:0] icap_output_reg = 0;

always @(posedge clk_peri) 
	if(reset) ctrl_reg <= 0;
	else if (reg_write[CTRL_ADR]) ctrl_reg <= do_peri[1:0];
	else ctrl_reg <= ctrl_reg;


/*always @(*) if(|reg_select) case (reg_read)
 CTRL_ADR: begin di_peri <= {13'h00,fifo_full, fifo_empty, icap_output_reg[16], ctrl_reg}; end
 DATI_ADR: begin di_peri <= 0; end
 DATO_ADR: begin di_peri <= {2'h0, icap_output_reg[15:0]}; end
 default: begin di_peri <= 0; end
endcase
*/

assign	di_peri	= (reg_read[CTRL_ADR])  ?  {13'h00,fifo_full, fifo_empty, icap_output_reg[16], ctrl_reg} : 
		  (reg_read[DATO_ADR])  ?  {2'h0, icap_output_reg[15:0]} : 18'd0;


fifo16i16o icapfifo (
  .rst	(reset), // input srst
  .wr_clk	(clk_peri), // input clk
  .rd_clk	(clk_peri), // input clk
  .din	(do_peri[15:0]), // input [15 : 0] din
  .wr_en	(reg_write[DATI_ADR]), // input wr_en
  .rd_en	(write_to_icap), // input rd_en
  .dout	(icap_data_i), // output [15 : 0] dout
  .full	(fifo_full), // output full
  .empty	(fifo_empty) // output empty
);




always @(posedge clk_peri) icap_output_reg <= {busy,icap_data_o};

//wire[16:0] icap_output_reg;
//assign icap_output_reg = {busy,icap_data_o};

generate
	case (ARCHITECTURE)
   "SPARTAN6": begin: ICAPSP6
		// spartan 6
		ICAP_SPARTAN6 #(
			.DEVICE_ID(30'h4000093),     // Specifies the pre-programmed Device ID value
			.SIM_CFG_FILE_NAME("NONE")  // Specifies the Raw Bitstream (RBT) file to be parsed by the simulation
                           // model
		)
		ICAP_SPARTAN6_inst (
			.BUSY		(busy),   			// 1-bit output: Busy/Ready output
			.O			(icap_data_o),    // 16-bit output: Configuartion data output bus
			.CE		(!write_to_icap), // 1-bit input: Active-Low ICAP Enable input
			.CLK		(clk_peri),     	// 1-bit input: Clock input
			.I			(icap_data_i),    // 16-bit input: Configuration data input bus
			.WRITE	(!enable_reconfiguration_port)  // 1-bit input: Read/Write control input
		);
	end
   "SERIE7": begin: ICAPSE7
		//artix, Kintex, virtex 7
		ICAPE2 #(
			.DEVICE_ID(30'h3651093),   // Specifies the pre-programmed Device ID value to be used for simulation
												// purposes.
			.ICAP_WIDTH("X16"),        // Specifies the input and output data width.
			.SIM_CFG_FILE_NAME("None") // Specifies the Raw Bitstream (RBT) file to be parsed by the simulation
												// model.
		)
		ICAPE2_inst (
			.O			(icap_data_o),    // 32-bit output: Configuration data output bus
			.CLK		(clk_peri),     	// 1-bit input: Clock Input
			.CSIB		(!write_to_icap), // 1-bit input: Active-Low ICAP Enable
			.I			(icap_data_i),    // 32-bit input: Configuration data input bus
			.RDWRB	(!enable_reconfiguration_port)  // 1-bit input: Read/Write Select input
		);
	end
	"VIRTEX6": begin: ICAPV6
		// virtex 6
		ICAP_VIRTEX6 #(
			.DEVICE_ID(30'h4244093),   // Specifies the pre-programmed Device ID value
			.ICAP_WIDTH("X16"),        // Specifies the input and output data width to be used with the
												// ICAP_VIRTEX6.
			.SIM_CFG_FILE_NAME("NONE") // Specifies the Raw Bitstream (RBT) file to be parsed by the simulation
												// model
		)
		ICAP_VIRTEX6_inst (
			.BUSY		(busy),   			// 1-bit output: Busy/Ready output
			.O			(icap_data_o),    // 32-bit output: Configuration data output bus
			.CLK		(clk_peri),     	// 1-bit input: Clock Input
			.CSB		(!write_to_icap), // 1-bit input: Active-Low ICAP input Enable
			.I			(icap_data_i),    // 32-bit input: Configuration data input bus
			.RDWRB	(!enable_reconfiguration_port)  // 1-bit input: Read/Write Select input
		);
	end
	"VIRTEX5": begin: ICAPV5
		// virtex 5
		ICAP_VIRTEX5 #(
			.ICAP_WIDTH("X16") // "X8", "X16" or "X32" 
		) ICAP_VIRTEX5_inst (
			.BUSY		(busy),      		// Busy output
			.O			(icap_data_o),    // 32-bit data output
			.CE		(!write_to_icap), // Clock enable input
			.CLK		(clk_peri),       // Clock input
			.I			(icap_data_i),    // 32-bit data input
			.WRITE	(!enable_reconfiguration_port)     // Write input
		);
	end
	default: begin: ICAPDEF
                     
   end	
	endcase
endgenerate
endmodule
