/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt	: SpartanMC
// Dateiname	: intr_ctrlp.v
// Autor	: Falk Hassler / Steffen Kunze
// Modul	: intr_ctrlp
//
// Beschreibung : Ein Interrupt-Controller mit Unterbrechung bei hoeerer
//                Prioritaet als der aktuelle Interrupt.
///////////////////////////////////////////////////////////////////////////////

`timescale 1 ns / 1 ps
`default_nettype none

module intr_ctrlp (
			clk_peri,
			do_peri,
			di_peri,
			addr_peri,
			access_peri,
			wr_peri,
			reset,
			intr_out,
			intr_in,
			intr_return,
			intr_enable
		);

	// ld-Funktion (dualer Logarithmus, aufgerundet)
	function integer ld(input integer value);
		integer i;
		begin
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction

// Die Folgenden Parameter werden durch die system.v bei der Erzeugung des Modules ueberschrieben.
parameter BASE_ADR	= 10'h8;
parameter IR_SOURCES	= 18;

parameter DECODSIM_DEB	= "NO";

// Anzahl der PORTS fuer die Anzeige der Bits berechnen
parameter PORTS		= ((IR_SOURCES) / 18) + 1 * ( |((IR_SOURCES) % 18));
parameter ALL_PORTS	= PORTS + 1;
parameter PORT_ADDR	= ld(PORTS+1);	// =ld(groesste benoetigte Adresse)

// Port 0 die Nummer des Interrupt mit der hoechsten Prioritaet
// Port 1 die 1. 18 Interruptbits
// Port 2 die 2. 18 Interruptbits
// Port 3 die 3. 18 Interruptbits
// ... weitere Ports sind moeglich

input	wire				clk_peri;
input	wire	[17:0]			do_peri;
output	wire	[17:0]			di_peri;
input	wire	[9:0]			addr_peri;
input	wire				access_peri;
input	wire				wr_peri;
input	wire				reset;
output	wire				intr_out;
input	wire	[IR_SOURCES-1:0]	intr_in;
//input	wire				interrupt_lowactiv;
input	wire				intr_return;
input	wire				intr_enable;

wire		[17:0]			ioreg_data;
reg					allow_interrupt;

reg		[7:0]			active_priority;	// Prioritaet des aktiven Interrupts
reg		[63:0]			priority_list;		// Liste der angesprungenen Interrupts
reg		[3:0]			intr_cnt;		// Anzahl der geschachtelten Interrupts
reg					allow_interrupt_buf;	// Puffer			
reg					intr_return_buf;	// Puffer

wire		[17:0]			priority;		// Geraeteregister fuer Interrupt Prioritaet

wire					intr_or_out;

// chipselect fuer die Statusregister
wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);


wire[PORTS*18-1:0] intr_in_padded;
assign intr_in_padded[IR_SOURCES-1:0] = intr_in;

// Priority Port und Ports fuer Interrupt Bits erstellen.
di_peri_selector peri_sel	(
		.data_in({
				intr_in_padded,
				priority[17:0]
								}),
		.data_out(	ioreg_data			),
		.select(	reg_read			)
				);

defparam peri_sel.SEL_BITS	= ALL_PORTS;

assign di_peri[18-1:0]    = ioreg_data;

// intr_out fuer die Signalisierung in der CPU bilden.
assign intr_or_out = |intr_in;


reg	[7:0]	new_priority;
integer 	i;

always @(intr_in)
begin
	new_priority = 0;
	for (i=0; i<IR_SOURCES; i=i+1)
	begin
	if(intr_in[i]) new_priority = i+1;
	end
end



always @(posedge clk_peri)

if(reset)
	begin
	intr_return_buf		<= 1'b0;
	active_priority		<= 0;
	priority_list		<= 0;
	allow_interrupt		<= 1'b0;
	allow_interrupt_buf	<= 1'b0;
	intr_cnt		<= 0;
	end
else
   begin
	// steigende Flanke intr return -- eine ISR wurde beendet
	if(intr_return == 1'b1 && intr_return_buf == 1'b0)
	/* Das Ende einer ISR Funktion (also auch der RFE Befehl) muss
	   immer im diable Interrupt abgearbeitet werden!	 */
		// Der naechste Interrupt aus der Liste rueckt nach (wird weiter bearbeitet).
		begin
		active_priority	<= priority_list[55:48];
		priority_list	<= {priority_list[55:0],8'b0};
		intr_cnt	<= intr_cnt - 1;
		end
	else
		// Veraenderung des Interruptvektors -- eine neue ISR soll beginnen
		// neuer Interrupt ist hoeher priorisiert als aktiver
		if(!allow_interrupt && allow_interrupt_buf)	// Interrupt wurde angenommen
			begin
			// neuer Interrupt wird aktiv + in die Liste genommen
			priority_list	<= {new_priority,priority_list[63:8]};
			intr_cnt	<= intr_cnt + 1;
			  // keine weiteren Interrups annehmen, wenn Liste voll ist
//			  if (intr_cnt > 7)
//				begin
//				active_priority	<= (IR_SOURCES + 1); // maximale Prioritaet
//				end
//			  else
//				begin
				active_priority	<= new_priority;
//				end
			end
			
   // buffer zur Flankenerkennung und Synchronisation von interrupt_lowactiv
   intr_return_buf	<= intr_return;
   allow_interrupt_buf	<= allow_interrupt;

   // keine weiteren Interrups annehmen, wenn Liste voll ist oder die Prioritaet nicht hoeher ist.
   allow_interrupt	<= ((new_priority > active_priority)&(intr_enable)&(intr_cnt < 8));
   end

// keine weiteren Interrups annehmen, wenn Liste voll ist oder die Prioritaet nicht hoeher ist.
//assign allow_interrupt	= ((new_priority > active_priority)&(intr_enable)&(intr_cnt < 8));

// Der Ausgang des Oder-Gatters wird durch allow_interrupt maskiert
assign intr_out = allow_interrupt & intr_or_out;
assign priority[17:0]	= {10'h000, priority_list[63:56]};

// Offene Eingaenge in der Simulation suchen und melden
// synopsys translate_off
always @(intr_in)
begin
	for (i = 0; i < IR_SOURCES; i = i + 1)
		if (intr_in[i] === 1'bz)
		begin
			$display("Interruptcontroller: intr_in[%0d] === 1'bz", i);
			$stop;
		end
end
// synopsys translate_on

endmodule
