/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none

//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Meyer
// Create Date:    17:08:07 03/30/2007
// Module Name:    JTAG
// Project Name:   SpartanMC 18
//////////////////////////////////////////////////////////////////////////////////
module jtag(

	//*** Verbindungen mit dem MC ***

	input	wire 			clk_peri,	//Systemtakt
	input wire			[17:0]	do_peri,	//Datenbus von MC
	output wire 		[17:0]	di_peri,	//Datenbus zu MC
	input	wire 			[9:0]	addr_peri,	//Adressbus von MC
	input	wire 			access_peri,	//Signale fuer Peripheriezugriffe
	input	wire			wr_peri,	//Schreibsignal
	input wire			reset,		//Reset-Signal
	output wire			intr,		//Interrupt

	//*** JTAG-Interface ***

	output wire			tms,
	output wire  		tck,
	output wire			tdi,
	input wire			tdo

	);

// Die Parameter fuer die Adressen werden in der system.v eingestellt.
parameter SHIFT_MSB_FIRST_OUT = 0;		//Wenn gesetzt wird MSB als erstes aus geschoben

parameter BASE_ADR	= 10'h0;		//Basisadresse fuer TPA-Geraet
parameter CONT_ADR	= 6;			//Anzahl der Adressen
parameter READ_ADR	= 1;			//Anzahl der Eingabe Register

parameter CONFIG_ADR	= 5'b00001;			//Taktteiler  setzen/lesen
parameter TMSTAP_ADR	= 5'b00010;			//TMS-Sequenz setzen/lesen
parameter TDIDATA_ADR	= 5'b00100;			//TDI-Daten   setzen/lesen
parameter TDILAE_ADR	= 5'b01000;			//TDI-Laenge  setzen
parameter TDODATA_ADR	= 5'b10000;			//TDO-Daten   lesen

parameter ALL_PORTS		= 5;
parameter DECODSIM_DEB		= "NO";


// parameter für verschiedene Zustände
`include "jtag_master/src/jtag_parameter.vh"

wire 		select;
//wire	[17:0]	status;

// Adressdekoder, der aus der angelegten Adresse ein chipselect erzeugt
//pselect iCSL (
//	.addr		(	addr_peri[9:3]  ),
//	.activ_peri	(	access_peri     ),
//	.select		(	select          )
//	);
//defparam iCSL.ADDR_WIDTH  = 7;
//defparam iCSL.BASE_WIDTH  = 7;
//defparam iCSL.BASE_ADDR   = BASE_ADR >> 3;	//BASE_ADR muss durch 8 teilbar sein.

wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);




// Register die benötigt werden

reg [9:0] clk_limit = 10'h1;	// Konfigregister für Taktteiler
reg [9:0] clk_counter = 10'h0; // Taktteiler
reg clk_div = 0; 		// geteilter Takt


reg saved_tdi;



// Modul um 4 Taktphasen zu erstellen:
wire clk_0;
wire clk_90;
wire clk_180;
wire clk_270;

four_phase_clk four_phase_clk_0(
	.clk 		(clk_div),
	.reset 	(reset),
	.clk_0 	(clk_0),
	.clk_90 	(clk_90),
	.clk_180 (clk_180),
	.clk_270 (clk_270)
);

//Taktteilung
always @ (posedge clk_peri)
	if (reset) begin
		clk_counter <= 0;
		clk_div <= 0;
	end else begin
		clk_counter <= clk_counter +1;
		if (clk_counter >= clk_limit) begin
			clk_counter <= 0;
			clk_div <= !clk_div;
		end
	end

reg man_tms = 0;
reg man_tdi = 0;
reg man_tck = 0;

					// Register die vom Spartan zugreifbar sind
reg run = 0; 				// Komponente aktiv schalten idle <-> waiting
					// ClockLimit setzen: Register weiter oben definiert

reg auto_tdi = 0;
reg auto_tms = 0;
reg gated_tck = 0;			// config value: if 1 than tck stands when no tck is needed

					// tck generieren
reg tck_en = 0;				// globales enable für tck
wire tck_run;				// tck enable für gated tck
reg tckneeded = 0;
assign tck_run = (gated_tck)?tckneeded : 1;
assign tck = (tck_en & tck_run)?(clk_180 | clk_270):man_tck;

// Umschaltung Auto-Manuell
assign tdi = run ? auto_tdi : man_tdi;
assign tms = run ? auto_tms : man_tms;


reg tclk = 1;			// TCLK aktivieren oder deaktivieren
reg old_tclk = 1;		// waitung -> setTclk -> waiting
						// letzter Zustand bleibt im Bit gespeichert

reg irEn	= 0;		// Interrupts enable
wire resIr;
reg setIr = 0;
reg ir = 0;
assign intr = ir & irEn;

reg [3:0]   target_state = 0;		// Target in welches die Daten geladen werden
reg [3:0]   reset_state = 0;	// State in den die TAP fährt nach shiften
reg [17:0]  shift_data = 0;			// Daten die geschiftert werden sollen
reg [17:0]  shift_data_jtag = 0;		// Daten die durch die JTAG-Kette
reg [17:0]  read_data = 0;			
reg [4:0]   data_len = 0; 			// Länge der Daten 
reg		reverse_shift_out = 0;	// Daten werden Rückwärts geschoben (MSB first) wird mit Länge übergeben
reg 		reverse_shift_in  = 0;	// Daten werden Rückwärsts vom TDO gelesen, wird mit Länge übergeben

reg reset_data_len = 0;					// signal welches reset der Datenlänge auslöst

reg busy = 0;								// register das anzeigt ob Controller beschäftigt
// Lesen der Daten vom Spartan
always @ (posedge clk_peri) begin
    if(reset) begin
	clk_limit[9:0]      <= 1;
	irEn                <= 0;
	run                 <= 0;
	tclk                <= 0;
        man_tms             <= 0;
	man_tdi             <= 0;
	man_tck             <= 0;
	gated_tck           <= 0;
	target_state        <= 0;
	reset_state         <= 0;
	reverse_shift_out   <= 0;
	reverse_shift_in    <= 0;
	shift_data          <= 0;
	data_len	    <= 0;
    end else begin    
    if (select & |reg_write)
	case (reg_write)
            CONFIG_ADR:	begin //Taktteiler oder Interruptenable setzen
		clk_limit[9:0] <= do_peri[9:0];
		irEn <= do_peri[10];
		run <= do_peri[11];
		tclk <= do_peri[12];
		man_tms <= do_peri[13];
		man_tdi <= do_peri[14];
		man_tck <= do_peri[15];
		gated_tck <= do_peri[16];
		end
	    TMSTAP_ADR: begin	//TMS-Sequenz setzen
		target_state <= do_peri[3:0];
		reset_state <= do_peri[7:4];
		reverse_shift_out <= do_peri[8];
		reverse_shift_in  <= do_peri[9];
		end
            TDIDATA_ADR: begin	//TDI-Daten setzen
		shift_data <= do_peri;
		//read_data <= 0;
		end
	    TDILAE_ADR:	//TDI-Laenge setzen
		begin
		data_len	   		<= do_peri[4:0];
		end
	endcase
	if (reset_data_len) data_len <= 0;
	if (resIr) ir <= 0;
	if (setIr) ir <= 1; 
    end
end
// Daten die vom Spartan gelesen werden
wire [17:0] status;
//Daten werden vom MC gelesen
assign status[17] = busy;			//Bit 17 busy-Signal
assign status[16] = 0;
assign status[15] = tck;
assign status[14]	= tdi;
assign status[13] = tms;
assign status[12] = tclk;
assign status[11] = run;
assign status[10] = irEn;
assign status[9:0]	= clk_limit;		//Bit 10:0! Taktteiler


assign	di_peri	= (select & reg_read==CONFIG_ADR)  ? status  :
		  (select & reg_read==TMSTAP_ADR) ? {8'h0, reverse_shift_in, reverse_shift_out ,reset_state, target_state} :
		  (select & reg_read==TDILAE_ADR) ? data_len	  :
		  (select & reg_read==TDIDATA_ADR) ? shift_data :
		  (select & reg_read==TDODATA_ADR) ? read_data : 18'd0;
assign	resIr	= (select & reg_read==CONFIG_ADR)  ? 1  : 0;

// Statemachine der Komponente
reg 	[3:0] peri_state 			= 3'h0;
reg 	[3:0] next_peri_state	= 3'h0;
reg	[2:0] shift_state 		= 3'h0;
reg 	[2:0] next_shift_state	= 2'h0;
reg	[2:0] tms_state 			= 3'h0;
reg 	[2:0] next_tms_state 	= 3'h0;

(* FSM_ENCODING="USER" *)reg 	[3:0] curr_internal_tap_state = 0;
reg	[3:0] next_internal_tap_state = 0;
reg	[3:0] target_tap_state = 0;

reg [4:0] bitcount = 0;		

reg 		trigger_bitcount_neg = 0;
reg 		trigger_bitcount_load = 0;
reg 		trigger_setTCLKcount = 0;
reg 		trigger_correctBitcount = 0;
reg 		trigger_resetcount = 0;
		
reg 		trigger_setTDI = 0;
reg 		trigger_resetTDI = 0;
reg 		trigger_saveTDI = 0;
reg 		trigger_restoreTDI = 0;
reg 		trigger_setTMS = 0;
reg 		trigger_resetTMS =0;
reg 		trigger_enableTCK = 0;
reg 		trigger_disableTCK = 0;
reg 		trigger_setBusy = 0;
reg 		trigger_resetBusy = 0;
reg 		trigger_storeTCKL = 0;
//reg 		trigger_storeTMS = 0;
//reg 		trigger_storeTDI = 0;
reg 		trigger_shiftData = 0;
reg 		trigger_loadShiftData = 0;

reg 		trigger_shiftReadData = 0;
reg		trigger_deleteReadData = 0;
reg 		trigger_setTargetTapState = 0;
reg 		trigger_setResetTapState = 0;
reg 		trigger_sertckneeded = 0;
reg 		trigger_resettckneeded = 0;

always @ (posedge clk_div) begin
	if (reset) begin 
		peri_state <= State_Reset;
		shift_state <= idle;
		tms_state <= idle;
		curr_internal_tap_state <= RUN_TEST_IDLE;  
		// Interne TAP-State-Machine befindet sich nach Reset im IDLE Zustand.
		// Befor die Komponente in den Automatik-Modus geschalten
		// werden kann muss der TAP-Controller im Manuellen Modus in den IDLE Modus durch
		// eine Software-Init-Routine gebracht werden.
	end else begin 
		// set calculated next states for state machines
		if (clk_270) peri_state <= next_peri_state;
		shift_state <= next_shift_state;
		tms_state <= next_tms_state;
		if (clk_0) curr_internal_tap_state <= next_internal_tap_state;
	end

	// trigger auswertung
	if (trigger_bitcount_load) bitcount <= data_len[4:0]-1;
	if (trigger_setTCLKcount) bitcount <= 3;
	if (trigger_correctBitcount) bitcount <= 18 - data_len[4:0];
	if (trigger_resetcount) bitcount <= 5;
	if (trigger_bitcount_neg) bitcount <= bitcount-1;
	if (trigger_setTDI) auto_tdi <= 1;
	if (trigger_resetTDI) auto_tdi <= 0;
	if (trigger_setTMS) auto_tms <= 1;
	if (trigger_resetTMS) auto_tms <= 0;
	if (trigger_saveTDI) saved_tdi <= auto_tdi;
	if (trigger_restoreTDI) auto_tdi <= saved_tdi;
	if (trigger_enableTCK) tck_en <= 1;
	if (trigger_disableTCK) tck_en <= 0;
	if (trigger_setBusy) busy <= 1;
	if (trigger_resetBusy) busy <= 0;
	if (trigger_storeTCKL) old_tclk <= tclk;
//	if (trigger_storeTMS) old_man_tms <= man_tms;
//	if (trigger_storeTDI) old_man_tdi <= man_tdi;
	if (trigger_shiftData) begin 
		if (reverse_shift_out) begin
			// Wenn everse dann schifte MSB
			auto_tdi <= shift_data_jtag[17];
			shift_data_jtag <= {shift_data_jtag[16:0], 1'b0};
			end 
		else begin
			// Sonst LSB
			auto_tdi <= shift_data_jtag[0];
			shift_data_jtag <= {1'b0, shift_data_jtag[17:1]};
			end
	end
	if (trigger_loadShiftData) shift_data_jtag <= shift_data;
	if (trigger_deleteReadData) read_data <= 0;
	if (trigger_shiftReadData) begin 
		if (reverse_shift_in) begin
			// Wenn reverse shift in dann setze TDO als MSB
			read_data <= {tdo,read_data[17:1]};
			end 
		else begin
			// Sonst als LSB
			read_data <= {read_data[16:0],tdo};
		end
	end

	if (trigger_setTargetTapState) target_tap_state <= target_state;
	if (trigger_setResetTapState) target_tap_state <= reset_state;

	if(trigger_sertckneeded) tckneeded <= 1;
	if(trigger_resettckneeded) tckneeded <= 0;

end
always @* begin		
		
		
		next_peri_state			<= peri_state;
		next_shift_state			<= shift_state;
		next_tms_state 			<= tms_state ;
		next_internal_tap_state <= curr_internal_tap_state;

		
		
		trigger_bitcount_load <= 0;
		trigger_setTCLKcount <= 0;
		trigger_bitcount_neg <= 0;
		trigger_correctBitcount <= 0;
 		trigger_resetcount <= 0;
		
		trigger_setTDI <= 0;
		trigger_resetTDI <= 0;
		trigger_saveTDI <= 0;
		trigger_restoreTDI <= 0;
		trigger_setTMS <= 0;
		trigger_resetTMS <=0;
		trigger_enableTCK <= 0;
		trigger_disableTCK <= 0;
		trigger_setBusy <= 0;
		trigger_resetBusy <= 0;
		trigger_storeTCKL <= 0;
//		trigger_storeTMS <= 0;
//		trigger_storeTDI <= 0;
		trigger_shiftData <= 0;
		trigger_loadShiftData <= 0;
		reset_data_len <= 0;
		trigger_setTargetTapState <= 0;
		trigger_setResetTapState <= 0;
		trigger_shiftReadData <= 0;
		trigger_deleteReadData <= 0;
		setIr <= 0;
		trigger_sertckneeded <= 0;
 		trigger_resettckneeded <= 0;
		
		case (peri_state) 
			State_Reset			: if (clk_270)begin 
				next_peri_state <= State_Idle;
				trigger_disableTCK <= 1; 
				trigger_resetTMS <= 1;
				trigger_resetTDI <= 1;
				trigger_setBusy <= 1;
				end
			State_Idle			: if (clk_270)begin 
				trigger_resetBusy <= 1;
				trigger_disableTCK <= 1;
				//if(man_tms) trigger_setTMS <= 1;
				//else trigger_resetTMS <= 1;
				//if (man_tdi) trigger_setTDI <= 1;
				//else trigger_resetTDI <= 1;
				if (run) next_peri_state <= State_Waiting;
				end
			State_Waiting		: if (clk_270)begin 
				trigger_enableTCK <= 1;	
				//trigger_resettckneeded <= 1;
				if (!run) next_peri_state <= State_Idle;								//verlasse aktiven Modus
				else if (tclk != old_tclk)  begin
					// nur für MSP: Wenn sich TCLK vom alten TCLK unterscheidet dann setze den neuen
					trigger_setBusy <= 1;		// busy setzen
					trigger_disableTCK <= 1;	// normalen tck ausschalten
					trigger_saveTDI <= 1;		// tdi sichern
					trigger_setTCLKcount <= 1;	// Counter für taktanzahl setzen
					if (tclk) trigger_setTDI <= 1;	// wenn 1 gesetzt werden muss
					else trigger_resetTDI <= 1;		// wenn 0 gesetzt werden muss
					next_peri_state <= State_setTclk;	//ändere tckl
					end
				
				else if (reverse_shift_out & (data_len[4:0] != 0)) begin
					trigger_setBusy <= 1;				// busy setzen
					trigger_saveTDI <= 1;
					trigger_correctBitcount <= 1;		// Zähler auf 18 - Länge setzen
					trigger_loadShiftData <= 1;		// Daten zum shiften laden
					trigger_deleteReadData <= 1;		// letze gelesene Daten löschen
					next_peri_state <= State_Correct_1; 		//shifte Daten das MSB an position 18 steht
						// Daten sollen mit MSB zuerst geshiftet werden und müssen bis ans Ende des Regs geschoben werden
					end
					
				else if (data_len[4:0] != 0) begin
					next_peri_state <= State_TapToStart; 	// Hauptstate-machine in den TAP-Change-Modus setzen
					trigger_setTargetTapState <= 1; 			// Target Tap-State setzen
					next_tms_state <= prepare_for_working;	// TAP-Controller initialisieren
					trigger_loadShiftData <= 1;				// Daten zum shiften laden
					trigger_setBusy <= 1;						// Busy setzen
					trigger_saveTDI <= 1;						// TDI sichern
					trigger_deleteReadData <= 1;				// alte Daten löschen
					end
				else begin // do nothing 
																				// Wenn aktiver Modus dann schalte TCK ein
				end
			end
		State_TapToStart	: begin
			// Bleibe so lange hier bis TMS state machine fertig ist
			if (tms_state == idle) if (clk_270) begin
				trigger_bitcount_load <= 1;				// Zähler mit Länge der Daten laden
				//if (clk_270) trigger_shiftData <= 1;
				next_peri_state <= State_Shift;			// Haupt state machine in shift state setzen
				next_shift_state <= prepare_for_working;	// shift state machine init
				end
			end
		State_Shift			: begin 
			trigger_setResetTapState <= 1;			// lade reset tap state
			// bleibe so lange bis shiften fast beendet (3 Tackte vorher)
			if (shift_state == shifting1 & bitcount == 3) if (clk_0) begin
				next_tms_state <= prepare_for_working;		// init tms machine
				end
			// wenn shift state machine fertig fahre tap-machine in endzustand
			if (shift_state == idle) next_peri_state <= State_TapToEnd;
			end
		State_TapToEnd		: if (clk_270)begin 
			// bleibe hier bis tms state machine fertig
			if (tms_state == idle) begin
				trigger_resettckneeded <= 1;
				if (!reverse_shift_in) begin 			// Wenn nicht reverse_shift_in dann fertig
					next_peri_state <= State_Waiting;// haupt state machine in waiting zustand
					reset_data_len <= 1;					// daten länge zurück setzen
					trigger_resetBusy <= 1;				// busy aufheben
					setIr <= 1;								// interrupt auslösen
					
					

					end
				else begin 									// wenn reverse shift in:
					trigger_correctBitcount <= 1;		// zähler laden
					next_peri_state <= State_Correct_2;// in Correct state gehen
					end
				end
			end
		State_Correct_1		: if (clk_270)begin 
				//Correctur zustand um shift data in richtige position zu bringen
				if (bitcount == 0) begin
					// wenn zähler bei null verlasse zustand
					next_peri_state <= State_TapToStart; // Starte Datenshift-Vorgang
					trigger_setTargetTapState <= 1; 
					next_tms_state <= prepare_for_working;
					trigger_setBusy <= 1;
				//	trigger_saveTDI <= 1;  viel zu spät weil tdi schon geändert durch shiften
				end else begin
					trigger_bitcount_neg <= 1;
					trigger_shiftData <= 1;
				end
			end
		State_Correct_2		: if (clk_270)begin 
				//trigger_resettckneeded <= 1;
				//Correctur zustand um rückwärts geschobene Daten zu korrigieren
				if (bitcount == 0)
					// wenn zähler bei null verlasse zustand
					begin		// wenn 
						next_peri_state <= State_Waiting;
						trigger_restoreTDI <= 1;
						reset_data_len <= 1;
						trigger_resetBusy <= 1;
						setIr <= 1;
						if(gated_tck) trigger_disableTCK <= 1;
						
					end
				else begin
					trigger_bitcount_neg <= 1;
					trigger_shiftReadData <= 1;
					end
			end
		State_setTclk		: if (clk_270)begin 
			if (bitcount == 0) begin 
				next_peri_state <= State_Waiting;
				trigger_storeTCKL <= 1;
				trigger_restoreTDI <= 1;
				trigger_enableTCK <=1;
				trigger_resetBusy <= 1;
				setIr <= 1;
			end else begin
				trigger_bitcount_neg <= 1;
				end
			end
//		State_setTMS		: if (clk_270)begin 
//			next_peri_state <= State_Idle;
//			trigger_storeTMS <= 1;
//			trigger_resetBusy <= 1;
//			setIr <= 1;
//			end
//		State_setTDI		: if (clk_270)begin 
//			next_peri_state <= State_Idle;
//			trigger_storeTDI <= 1;
//			trigger_resetBusy <= 1;
//			setIr <= 1;
//			end
	endcase
		
		
	
	case (shift_state) 
	 idle						: begin end					
	 prepare_for_working	: if (clk_0) begin
		//if (bitcount == 0) next_shift_state <= idle;
		//else 
		begin 
			trigger_shiftData <= 1;
			trigger_shiftReadData <= 1;
			next_shift_state <= shifting1;
			//bitcount <= 0;
			
			end
		end
	 shifting1				: if (clk_0) begin
		if (bitcount==1) begin
			//if (curr_internal_tap_state != reset_state) begin
			//	trigger_setTMS <= 1;
			//	next_internal_tap_state <= curr_internal_tap_state + 1;
		//	end
			
		end else begin
		
		end
		//TDI schieben
		trigger_shiftData <= 1;
		trigger_shiftReadData <= 1;
		next_shift_state <= shifting2;
		end
	 shifting2				: if (clk_90) begin
		 // tck geht auf 1 Daten werden im Target übernommen
		 next_shift_state <= shifting3;
	 end
	 shifting3				: if (clk_180) begin
		//TDO schieben
		
		 next_shift_state <= shifting4;
		 if (bitcount==1) begin 
			//next_tms_state <= prepare_for_working;
			end
		end
	 shifting4				: if (clk_270) begin
		if (bitcount==1) begin 
			next_shift_state <= idle; 
			//next_tms_state <= prepare_for_working;
			trigger_restoreTDI <= 1;
			end 
		else begin 
			next_shift_state <= shifting1; 
			trigger_bitcount_neg <= 1;
			end
		end
	 shift_unused1			: begin
		if (clk_div) next_shift_state <= idle;
		
	 end
	 shift_unused2			: begin
		if (clk_div) next_shift_state <= idle;
	 end
	endcase
	
	case (tms_state)
	 idle						: begin 
		//next_tms_state <= prepare_for_working;
	 end						
	 prepare_for_working	: if (clk_0) begin 
		if ( target_tap_state == TEST_LOGIC_RESET ) next_tms_state <= working_softreset1; 
		else if (curr_internal_tap_state != target_tap_state) next_tms_state <= working_normal1;
		
		else if ((curr_internal_tap_state == PAUSE_DR) & (target_tap_state == PAUSE_DR) & (peri_state == State_TapToStart)) 
			next_tms_state <= working_normal1;
		
		else next_tms_state <= idle;
	end
	 working_normal1		: if (clk_0) begin 
		trigger_sertckneeded <= 1;
		case (curr_internal_tap_state)
		 TEST_LOGIC_RESET:
			if (target_tap_state == TEST_LOGIC_RESET)	begin
				trigger_setTMS <= 1;
				next_internal_tap_state <= TEST_LOGIC_RESET;
			end
			else	begin
				trigger_resetTMS <= 1;
				next_internal_tap_state <= RUN_TEST_IDLE;	
			end
		 RUN_TEST_IDLE:
			if (target_tap_state == RUN_TEST_IDLE) begin 
				trigger_resetTMS <= 1;
				next_internal_tap_state <= RUN_TEST_IDLE;
			end
			else	begin
				trigger_setTMS <= 1;
				next_internal_tap_state <= SELECT_DR;
			end
		 SELECT_DR:
			if ((target_tap_state == TEST_LOGIC_RESET ) 	|
				( target_tap_state == RUN_TEST_IDLE ) 		|
				( target_tap_state == SELECT_IR ) 			| 
				( target_tap_state == CAPTURE_IR ) 			| 
				( target_tap_state == SHIFT_IR ) 			| 
				( target_tap_state == EXIT1_IR ) 			| 
				( target_tap_state == PAUSE_IR ) 			| 
				( target_tap_state == EXIT2_IR ) 			|
				( target_tap_state == UPDATE_IR ) )	
			begin
				trigger_setTMS <= 1;
				next_internal_tap_state <= SELECT_IR;
			end
			else begin
			trigger_resetTMS <= 1;
			next_internal_tap_state <= CAPTURE_DR;
			end
		CAPTURE_DR: 
			if (target_tap_state == EXIT1_DR ) begin
				trigger_setTMS <= 1;	
				next_internal_tap_state <= EXIT1_DR;	
			end
			else begin 
				next_internal_tap_state <= SHIFT_DR;
				trigger_resetTMS <= 1;
			end											
		SHIFT_DR:
			if (target_tap_state == SHIFT_DR) begin 
				trigger_resetTMS <= 1;
				next_internal_tap_state <= SHIFT_DR;
			end else begin
				trigger_setTMS <= 1;
				next_internal_tap_state <= EXIT1_DR;
			end
		EXIT1_DR :
			if (target_tap_state == UPDATE_DR | target_tap_state == RUN_TEST_IDLE) begin
				trigger_setTMS <= 1;
				next_internal_tap_state <= UPDATE_DR;
			end else begin 
				trigger_resetTMS <= 1;
				next_internal_tap_state <= PAUSE_DR;
			end
		PAUSE_DR :
			if (target_tap_state == PAUSE_DR & (peri_state == State_TapToStart)) begin
				trigger_setTMS <= 1;
				next_internal_tap_state <= EXIT2_DR;
			end else if (target_tap_state == PAUSE_DR) begin
				trigger_resetTMS <= 1;
				next_internal_tap_state <= PAUSE_DR;
			end else begin
				trigger_setTMS <= 1;
				next_internal_tap_state <= EXIT2_DR;
			end 
		EXIT2_DR :
			if (target_tap_state == SHIFT_DR) begin
			trigger_resetTMS <= 1;
			next_internal_tap_state <= SHIFT_DR;
			end else begin
			trigger_setTMS <= 1;
			next_internal_tap_state <= UPDATE_DR;
			end 
		UPDATE_DR :
			if (target_tap_state == RUN_TEST_IDLE) begin
			trigger_resetTMS <= 1;
			next_internal_tap_state <= RUN_TEST_IDLE;
			end else begin
			trigger_setTMS <= 1;
			next_internal_tap_state <= SELECT_DR;
			end
		SELECT_IR:
			if (target_tap_state == TEST_LOGIC_RESET) begin
			trigger_setTMS <= 1;
			next_internal_tap_state <= TEST_LOGIC_RESET;
			end else begin
			trigger_resetTMS <= 1;
			next_internal_tap_state <= CAPTURE_IR;		
			end				
		CAPTURE_IR:
			if (target_tap_state == EXIT1_IR) begin
			trigger_setTMS <= 1;	
			next_internal_tap_state <= EXIT1_IR;
			end else begin
			trigger_resetTMS <= 1;
			next_internal_tap_state <= SHIFT_IR;
			end  			
		SHIFT_IR :
			if (target_tap_state == SHIFT_IR)begin 
			trigger_resetTMS <= 1;
			next_internal_tap_state <= SHIFT_IR;
			end else begin
			trigger_setTMS <= 1;
			next_internal_tap_state <= EXIT1_IR;
			end	
		EXIT1_IR :
			if (target_tap_state == UPDATE_IR)begin
			trigger_setTMS <= 1;
			next_internal_tap_state <= UPDATE_IR;
			end else begin
			trigger_resetTMS <= 1;
			next_internal_tap_state <= PAUSE_IR;
			end 
		PAUSE_IR :
			if (target_tap_state == PAUSE_IR)begin
			trigger_resetTMS <= 1;
			next_internal_tap_state <= PAUSE_IR;
			end else begin
			trigger_setTMS <= 1;
			next_internal_tap_state <= EXIT2_IR;
			end
		EXIT2_IR :
			if (target_tap_state == SHIFT_IR)begin 
			trigger_resetTMS <= 1;
			next_internal_tap_state <= SHIFT_IR;
			end else begin
			trigger_setTMS <= 1;
			next_internal_tap_state <= UPDATE_IR;
			end 
		UPDATE_IR :
			if (target_tap_state == RUN_TEST_IDLE)begin
				trigger_resetTMS <= 1;
				next_internal_tap_state <= RUN_TEST_IDLE; 
			end else begin
				trigger_setTMS <= 1;
				next_internal_tap_state <= SELECT_DR;
			end 
		endcase
		next_tms_state <= working_normal2;
	 end
	 working_normal2		: begin 
		next_tms_state <= working_normal3;
	 end
	 working_normal3 		:  begin 
		if (curr_internal_tap_state == target_tap_state) next_tms_state <= idle; 
		else next_tms_state <= working_normal1;
	 end
	 working_softreset1	: if (clk_0) begin 
		trigger_sertckneeded <= 1;
		trigger_setTMS <= 1;
		trigger_resetcount <= 1;
		next_tms_state <= working_softreset2;
	 end
	 working_softreset2	: begin 
		next_tms_state <= working_softreset3;
	 end
	 working_softreset3	:   begin 
		if (bitcount > 0) begin
			next_tms_state <= working_softreset2;
			trigger_bitcount_neg <=1;
			end
		else begin
			next_internal_tap_state <= TEST_LOGIC_RESET;
			next_tms_state <= idle; 
		end
	 end
	endcase
	
end
endmodule
