/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale 1ns / 1ps
`default_nettype none


module mem_swap_dual_master
#(
		parameter BASE_ADR              = 10'h0,
        parameter DMA_BASE_ADR			= 18'h0,
        parameter DMA_SIZE				= 'd2048,
		parameter DMA_ADDRWIDTH			= 'd11
)
(
        //*** Connections with SpartanMC master ***
        input   wire    	        		clk_peri,	//Systemtakt
        input   wire    	        		reset,		//Reset-Signal
        
		//register wires
        input   wire    	        		access_peri,//Signale fuer Peripheriezugriffe
        input   wire    	        		wr_peri,	//Schreibsignal

        //DMA-Signals
		input wire							mem_access,
		input wire							store_access_low,
		input wire							store_access_high,
		input wire		[9:0]				addr_peri,
		input wire		[7:0]				addr_high,
		input wire		[17:0]				do_peri,
		output wire		[17:0]				di_peri,  
	
        // Connections with slave-connector
        input wire 			        		swap_request,
        output wire							swap_complete,
        
		input wire							clk,
		input wire		[17:0]				data_inA,
    	output wire		[17:0]				data_outA,
		input wire							wr_low,
		input wire							wr_high,
		input wire		[DMA_ADDRWIDTH-1:0]	mem_addressA
);


parameter
    STATUS_REG_ADDR         	= 0,
	ALL_PORTS					= 1,
	RAMBLOCK_SIZE				= 2048,
	RAMBLOCKS_PER_CONNECTION 	= (DMA_SIZE - 1 + RAMBLOCK_SIZE - ((DMA_SIZE - 1) % RAMBLOCK_SIZE))/RAMBLOCK_SIZE,
	DECODSIM_DEB				= "NO";

wire [17:0] module_addr;
wire 		dma_select;
reg			ram_select;
reg			swap_complete_master_reg;
reg			swap_complete_slave_reg;
reg			dma_select_reg;
reg			swap_request_reg;
reg			swap_request_master_reg;
wire [17:0]	data_out_mem;
wire		store_access_low_mem;
wire		store_access_high_mem;
wire 		swap_request_master;

//register access signals
wire		status_read_sel;	// CPU liest von peri
wire		status_write_sel;	// CPU schreibt auf peri
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

function integer log2;
input integer value;
	for (log2=0; value>0; log2=log2+1)
	value = value>>1;
endfunction



pselect #(
	.ADDR_WIDTH(	18-DMA_ADDRWIDTH				),
	.BASE_WIDTH(	18-DMA_ADDRWIDTH				),
	.BASE_ADDR(		DMA_BASE_ADR[17:DMA_ADDRWIDTH]	)
) memselect_dma (
	.addr(			module_addr[17:DMA_ADDRWIDTH]	),
	.activ_peri(	mem_access						),
	.select(		dma_select						)
);


// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS		),
		.BASE_ADR(	BASE_ADR		),
		.SIM_DEBUG(	DECODSIM_DEB	)
) register_access_decoder (
		.reset(			reset		),
		.clk_peri(		clk_peri	),
		.addr_peri(		addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(		wr_peri		),

		.peri_select(		),
		.reg_select(	reg_select	),
		.reg_read(		reg_read	),
		.reg_write(		reg_write	)
);

assign		status_read_sel 	= reg_read[STATUS_REG_ADDR];
assign		status_write_sel 	= reg_write[STATUS_REG_ADDR];


dual_mem_switcher #(
    .RAMBLOCKS_PER_CONNECTION	( RAMBLOCKS_PER_CONNECTION	),
	.ADR_LEN					( DMA_ADDRWIDTH				)
) memory_switcher(
    .clkA			( clk_peri				 			),
	.data_inA		( do_peri				 			),
    .data_outA		( data_out_mem			 			),
	.wr_lowA		( store_access_low_mem	 			),
	.wr_highA		( store_access_high_mem	 			),
	.mem_addressA	( module_addr[DMA_ADDRWIDTH-1:0]	),

    .clkB			( clk		 						),
	.data_inB		( data_inA 	 						),
    .data_outB		( data_outA 	 					),
	.wr_lowB		( wr_low 		 					),
	.wr_highB		( wr_high 		 					),
	.mem_addressB	( mem_addressA	 					),

	.ram_select		( ram_select 						)      
);


//address of this module on SpartanMC
assign module_addr = {addr_high,addr_peri};

assign store_access_low_mem 	= dma_select ? store_access_low 	: 1'b0;
assign store_access_high_mem 	= dma_select ? store_access_high 	: 1'b0;

always@(posedge clk_peri)
begin
	if (reset )
	begin
		ram_select 					<= 1'b0;
		swap_request_master_reg 	<= 1'b0;
		swap_complete_master_reg 	<= 1'b0;
		swap_complete_slave_reg 	<= 1'b0;
		dma_select_reg 				<= 1'b0;
		swap_request_reg 			<= 1'b0;
	end
	else
	begin
		dma_select_reg 	<= dma_select;
		if (swap_request_master_reg && swap_request_reg)
		begin
			ram_select 					<= !ram_select;
			swap_complete_master_reg 	<= 1'b1;
			swap_complete_slave_reg 	<= 1'b1;
			swap_request_master_reg 	<= 1'b0;
			swap_request_reg 			<= 1'b0;
		end
		if (swap_request_master)
		begin
			swap_request_master_reg 	<= 1'b1;
			swap_complete_master_reg 	<= 1'b0;
		end
		if (swap_request)
		begin
			swap_request_reg 			<= 1'b1;
			swap_complete_slave_reg 	<= 1'b0;
		end
	end
end

assign swap_complete 		= swap_complete_slave_reg;

assign swap_request_master 	= status_write_sel && do_peri[0];

assign	di_peri				= 	status_read_sel ? {swap_complete_master_reg, 1'b0} : 
									dma_select_reg ? data_out_mem : 18'b0;

endmodule






