/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale 1ns / 1ps
`default_nettype none


module mem_swap_multi_master
#(
		parameter BASE_ADR              = 10'h0,
        parameter DMA_BASE_ADR			= 18'h0,
        parameter DMA_SIZE				= 'd2048,
		parameter DMA_ADDRWIDTH			= 'd11,
		parameter SLAVE_CONNECTIONS		= 'd1,
		parameter SOFTWARE_ARBITRATION	= 1'b0
)
(
        //*** Connections with SpartanMC master ***
        input   wire    	        		clk_peri,	//Systemtakt
        input   wire    	        		reset,		//Reset-Signal
        
		//register wires
        input   wire    	        		access_peri,//Signale fuer Peripheriezugriffe
        input   wire    	        		wr_peri,	//Schreibsignal

        //DMA-Signals
		input wire							mem_access,
		input wire							store_access_low,
		input wire							store_access_high,
		input wire		[9:0]				addr_peri,
		input wire		[7:0]				addr_high,
		input wire		[17:0]				do_peri,
		output wire		[17:0]				di_peri,  
	
        // Connections with slave-connector
        input wire 		[SLAVE_CONNECTIONS-1:0]	        		swap_request,
        output wire		[SLAVE_CONNECTIONS-1:0]					swap_complete,
        
		input wire		[SLAVE_CONNECTIONS-1:0]					clk,
		input wire		[SLAVE_CONNECTIONS*18-1:0]				data_inA,
    	output wire		[SLAVE_CONNECTIONS*18-1:0]				data_outA,
		input wire		[SLAVE_CONNECTIONS-1:0]					wr_low,
		input wire		[SLAVE_CONNECTIONS-1:0]					wr_high,
		input wire		[SLAVE_CONNECTIONS*DMA_ADDRWIDTH-1:0]	mem_addressA
);

function integer log2;
input integer value;
	for (log2=0; value>0; log2=log2+1)
	value = value>>1;
endfunction

localparam
    STATUS_REG_ADDR         	= 0,
	ALL_PORTS					= 1,
	RAMBLOCK_SIZE				= 2048,
	RAM_SEL_BITS				= log2(SLAVE_CONNECTIONS+1),
	RAMBLOCKS_PER_CONNECTION 	= (DMA_SIZE - 1 + RAMBLOCK_SIZE - ((DMA_SIZE - 1) % RAMBLOCK_SIZE))/RAMBLOCK_SIZE,
	DECODSIM_DEB				= "NO";

wire 	[17:0] 									module_addr;
wire 											dma_select;
wire	[RAM_SEL_BITS-1:0]						ram_select;
reg		[RAM_SEL_BITS-1:0]						ram_select_reg;
wire	[SLAVE_CONNECTIONS*RAM_SEL_BITS-1:0]	ram_select_array;
reg		[SLAVE_CONNECTIONS*RAM_SEL_BITS-1:0]	ram_select_slave_reg;
wire	[RAM_SEL_BITS-1:0]						ram_select_slave_array[SLAVE_CONNECTIONS-1:0];
reg		[RAM_SEL_BITS-1:0]						slave_ID_arbiter;
reg												swap_complete_master_reg;
reg		[SLAVE_CONNECTIONS-1:0]					swap_complete_slave_reg;
reg												dma_select_reg;
reg		[SLAVE_CONNECTIONS-1:0]					swap_request_slave_reg;
reg												swap_request_master_reg;
wire 	[17:0]									data_out_mem;
wire											store_access_low_mem;
wire											store_access_high_mem;
wire 											swap_request_master;	

//mem_switcher wires

//register access signals
wire		status_read_sel;	// CPU liest von peri
wire		status_write_sel;	// CPU schreibt auf peri
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;



pselect #(
	.ADDR_WIDTH(	18-DMA_ADDRWIDTH				),
	.BASE_WIDTH(	18-DMA_ADDRWIDTH				),
	.BASE_ADDR(		DMA_BASE_ADR[17:DMA_ADDRWIDTH]	)
) memselect_dma (
	.addr(			module_addr[17:DMA_ADDRWIDTH]	),
	.activ_peri(	mem_access						),
	.select(		dma_select						)
);


// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS		),
		.BASE_ADR(	BASE_ADR		),
		.SIM_DEBUG(	DECODSIM_DEB	)
) register_access_decoder (
		.reset(			reset		),
		.clk_peri(		clk_peri	),
		.addr_peri(		addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(		wr_peri		),

		.peri_select(		),
		.reg_select(	reg_select	),
		.reg_read(		reg_read	),
		.reg_write(		reg_write	)
);

assign		status_read_sel 	= reg_read[STATUS_REG_ADDR];
assign		status_write_sel 	= reg_write[STATUS_REG_ADDR];




mem_switcher #(
	.CONNECTIONS				(SLAVE_CONNECTIONS+1),
    .RAMBLOCKS_PER_CONNECTION	(RAMBLOCKS_PER_CONNECTION),
	.ADR_LEN					(DMA_ADDRWIDTH),
	.RAM_SEL_BITS				(RAM_SEL_BITS)
) memory_switcher(
    .clk			( {clk			, clk_peri				} ),
	.data_inA		( {data_inA 	, do_peri				} ),
    .data_outA		( {data_outA 	, data_out_mem			} ),
	.wr_low			( {wr_low 		, store_access_low_mem	} ),
	
	.wr_high		( {wr_high 		, store_access_high_mem	} ),
	.mem_addressA	( {mem_addressA	, module_addr[DMA_ADDRWIDTH-1:0]} ),

	.ram_select		( {ram_select_slave_reg, ram_select_reg} )     
);


//address of this module on SpartanMC
assign module_addr = {addr_high,addr_peri};

assign store_access_low_mem 	= dma_select ? store_access_low 	: 1'b0;
assign store_access_high_mem 	= dma_select ? store_access_high 	: 1'b0;


always@(posedge clk_peri)
begin
	if (reset )
	begin
		swap_request_master_reg 	<= 1'b0;
		swap_complete_master_reg 	<= 1'b0;
		dma_select_reg 				<= 1'b0;
		slave_ID_arbiter			<= 1'b0;
		ram_select_reg				<= {RAM_SEL_BITS{1'b0}};
	end
	else
	begin
		dma_select_reg 	<= dma_select;
		if (swap_request_master_reg && swap_request_slave_reg[slave_ID_arbiter])
		begin
			ram_select_reg								<= ram_select;
			//ram_select_slave_array[slave_ID_arbiter]	<= 1'b0;//ram_select_reg;
			
			swap_complete_master_reg					<= 1'b1;
			//swap_complete_slave_reg[slave_ID_arbiter] 	<= 1'b1;
			swap_request_master_reg					 	<= 1'b0;
			//swap_request_slave_reg[slave_ID_arbiter]	<= 1'b0;
			if(SOFTWARE_ARBITRATION==1'b0)
			begin
				slave_ID_arbiter			<= (slave_ID_arbiter<SLAVE_CONNECTIONS-1)?slave_ID_arbiter+1:{RAM_SEL_BITS{1'b0}};
			end
		end
		else
		if (swap_request_master)
		begin
			swap_request_master_reg 	<= 1'b1;
			swap_complete_master_reg 	<= 1'b0;
		end
		if(SOFTWARE_ARBITRATION==1'b1)
		begin
			if (status_write_sel)
			begin
				slave_ID_arbiter	<= do_peri[RAM_SEL_BITS-1+9:9];
			end
		end
	end
end

orgate #(
			.OR_WIDTH	(	SLAVE_CONNECTIONS	),
			.BUS_WIDTH	(	RAM_SEL_BITS		)
			)
		dmem_orA (
			.a		( 	ram_select_array	),
			.y		( 	ram_select		)
		);


genvar i;
generate
for(i = 0; i < SLAVE_CONNECTIONS; i = i+1)
begin : MEM_SW_WIRES

assign ram_select_array[RAM_SEL_BITS*i+RAM_SEL_BITS-1:RAM_SEL_BITS*i] = (swap_request_master_reg && swap_request_slave_reg[i] && slave_ID_arbiter==i) ? ram_select_slave_reg[i*RAM_SEL_BITS+RAM_SEL_BITS-1:i*RAM_SEL_BITS]:{RAM_SEL_BITS{1'b0}};

always@(posedge clk_peri)
begin
	if (reset )
	begin
		//ram_select_reg				<= {RAM_SEL_BITS{1'b0}};
		swap_complete_slave_reg[i] 	<= 1'b0;
		swap_request_slave_reg[i]	<= 1'b0;
		ram_select_slave_reg[i*RAM_SEL_BITS+RAM_SEL_BITS-1:i*RAM_SEL_BITS]		<= i+1;
	end
	else
	begin

		
		if (swap_request_master_reg && swap_request_slave_reg[i] && slave_ID_arbiter==i)
		begin
			//ram_select_reg															<= ram_select_slave_reg[i*RAM_SEL_BITS+RAM_SEL_BITS-1:i*RAM_SEL_BITS];
			ram_select_slave_reg[i*RAM_SEL_BITS+RAM_SEL_BITS-1:i*RAM_SEL_BITS]		<= ram_select_reg;
			
			//swap_complete_master_reg[i]	<= 1'b1;
			swap_complete_slave_reg[i] 	<= 1'b1;
			//swap_request_master_reg[i] 	<= 1'b0;
			swap_request_slave_reg[i]	<= 1'b0;
			if(SOFTWARE_ARBITRATION==1'b0)
			begin
				//slave_ID_arbiter			<= (slave_ID_arbiter<SLAVE_CONNECTIONS-1)?slave_ID_arbiter+1:{RAM_SEL_BITS{1'b0}};
			end
		end
		
		if (swap_request[i])
		begin
			swap_request_slave_reg[i]			<= 1'b1;
			swap_complete_slave_reg[i]	<= 1'b0;
		end
	end
	end
	
end
endgenerate


assign 	swap_complete 		= swap_complete_slave_reg;

assign 	swap_request_master	= status_write_sel && do_peri[0];

assign	di_peri				= status_read_sel ? {swap_complete_master_reg, 1'b0} : 
									dma_select_reg ? data_out_mem : 18'b0;

endmodule






