/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale 1ns / 1ps
`default_nettype none


module mem_switcher
#(
    parameter CONNECTIONS							= 18'h2,
    parameter RAMBLOCKS_PER_CONNECTION				= 'd2,
	parameter ADR_LEN								= 'd12,
	parameter RAM_SEL_BITS							= 'd1

)
(
        
    input wire		[CONNECTIONS-1:0]					clk,
	input wire		[CONNECTIONS*18-1:0]				data_inA,
    output wire		[CONNECTIONS*18-1:0]				data_outA,
	input wire		[CONNECTIONS-1:0]					wr_low,

	input wire		[CONNECTIONS-1:0]					wr_high,
	input wire		[CONNECTIONS*ADR_LEN-1:0]			mem_addressA,
	input wire		[CONNECTIONS*RAM_SEL_BITS-1:0]		ram_select
        
);

	//incoming wires in a array
	wire							clk_array			[CONNECTIONS-1:0];
	wire		[17:0]				data_inA_array		[CONNECTIONS-1:0];
    wire		[17:0]				data_outA_array		[CONNECTIONS-1:0];
	wire							wr_low_array		[CONNECTIONS-1:0];
	wire							wr_high_array		[CONNECTIONS-1:0];
	wire		[ADR_LEN-1:0]		mem_addressA_array	[CONNECTIONS-1:0];
	wire		[RAM_SEL_BITS-1:0]	ram_select_array	[CONNECTIONS-1:0];

	//mem interfaces
	wire		[CONNECTIONS-1:0]					clk_mem;
	wire		[CONNECTIONS*18-1:0]				data_inA_mem;
    wire		[CONNECTIONS*18-1:0]				data_outA_mem;
	wire		[18-1:0]							data_outA_mem_array	[CONNECTIONS-1:0];
	wire		[CONNECTIONS-1:0]					wr_low_mem;
	wire		[CONNECTIONS-1:0]					wr_high_mem;
	wire		[CONNECTIONS*ADR_LEN-1:0]			mem_addressA_mem;

		
		//clk		
		switch #(
			.CONNECTIONS(	CONNECTIONS		),
			.SELECT_BITS(	RAM_SEL_BITS	),
			.WIDTH(			1				)
			)clk_switch(
			.in(		clk			),
			.select(	ram_select	),
			.out(		clk_mem		) 
		);
		//data_inA
		switch #(
			.CONNECTIONS(	CONNECTIONS		),
			.SELECT_BITS(	RAM_SEL_BITS	),
			.WIDTH(			18				)
			)data_inA_switch(
			.in(		data_inA		),
			.select(	ram_select		),
			.out(		data_inA_mem	) 
		);
		//wr_low
		switch #(
			.CONNECTIONS(	CONNECTIONS		),
			.SELECT_BITS(	RAM_SEL_BITS	),
			.WIDTH(			1				)
			)wr_low_switch(
			.in(		wr_low		),
			.select(	ram_select	),
			.out(		wr_low_mem	) 
		);
		//wr_high
		switch #(
			.CONNECTIONS(CONNECTIONS),
			.SELECT_BITS(	RAM_SEL_BITS),
			.WIDTH(			1			)
			)wr_high_switch(
			.in(		wr_high			),
			.select(	ram_select		),
			.out(		wr_high_mem		) 
		);
		//mem_addressA
		switch #(
			.CONNECTIONS(	CONNECTIONS		),
			.SELECT_BITS(	RAM_SEL_BITS	),
			.WIDTH(			ADR_LEN			)
			)mem_addressA_switch(
			.in(		mem_addressA		),
			.select(	ram_select			),
			.out(		mem_addressA_mem	)	 
		);	

	
genvar		i;
genvar		j;
generate

	//TODO solve comment in
	/*for (i=0; i<CONNECTIONS;i=i+1)
	begin : TESTS
		for (j=0; j<CONNECTIONS;j=j+1)
		begin
			if(!i==j && ram_select_array[i] == ram_select_array[j])
			begin
				$error("ram select %d and %d access the same memory instance %d simultanously",i,j,ram_select_array[j]);
			end
		end
	end*/
	

	for (i=0; i<CONNECTIONS; i=i+1)
   	begin : ARRAY_GEN
		assign clk_array[i] 			= clk				[i];
		assign data_inA_array[i]		= data_inA			[18*i+17:18*i];
		assign data_outA[18*i+17:18*i]	= data_outA_array	[i];
		assign wr_low_array[i]			= wr_low			[i];
		assign wr_high_array[i]			= wr_high			[i];
		assign mem_addressA_array[i]	= mem_addressA		[ADR_LEN*i+ADR_LEN-1:ADR_LEN*i];
		assign ram_select_array[i]		= ram_select		[RAM_SEL_BITS*i+RAM_SEL_BITS-1:RAM_SEL_BITS*i];
	end


	for (i=0; i<CONNECTIONS; i=i+1)
   	begin : SWITCHES
		assign data_outA_array[i]		= data_outA_mem_array[ram_select_array[i]];
		
		assign data_outA_mem_array[i]	= data_outA_mem[i*18+18-1:i*18];
	end


	for	(i=0; i<CONNECTIONS; i=i+1)
	begin:	SWITCHABLE_MEM
	
		memory #(
			.RAMBLOCKS  ( RAMBLOCKS_PER_CONNECTION 	),
			.ADR_LEN	( ADR_LEN					)
		) memA (
			.clk		 ( clk_mem[i]											),
			.data_inA	 ( data_inA_mem[i*18+18-1:i*18] 						),
			.data_outA	 ( data_outA_mem[i*18+18-1:i*18]						),
			.wr_high	 ( wr_high_mem[i]										),
			.wr_low		 ( wr_low_mem[i]										),
			.mem_addressA( mem_addressA_mem[i*ADR_LEN+ADR_LEN-1:i*ADR_LEN]		),
			.data_outB	 (														),
			.mem_addressB( {ADR_LEN{1'b0}}										)
		);
	end

endgenerate

endmodule






