/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns/1ps
`default_nettype none

module memguard #(
	parameter BASE_ADR	= 10'h0,
	parameter NUM_WATCHERS = 4,
	parameter STOP_SIMULATION = "FALSE"
)(
	input wire core_clk,
	input wire core_code_access,
	input wire [17:0] core_code_addr,
	output wire [17:0] core_code_do,
	input wire core_code_jmp,
	output wire core_code_busy,
	input wire core_mem_access,
	input wire [9:0] core_mem_addr_block,
	input wire [7:0] core_mem_addr_high,
	output wire core_mem_busy,
	input wire [17:0] core_mem_di,
	output wire [17:0] core_mem_do,
	input wire core_mem_wr_high,
	input wire core_mem_wr_low,
	input wire core_reset,

	output wire mem_clk,
	output wire mem_code_access,
	output wire [17:0] mem_code_addr,
	input wire [17:0] mem_code_do,
	output wire mem_code_jmp,
	input wire mem_code_busy,
	output wire mem_mem_access,
	output wire [9:0] mem_mem_addr_block,
	output wire [7:0] mem_mem_addr_high,
	input wire mem_mem_busy,
	input wire [17:0] mem_mem_di,
	output wire [17:0] mem_mem_do,
	output wire mem_mem_wr_high,
	output wire mem_mem_wr_low,
	output wire mem_reset,

	input wire access_peri,
	input wire [9:0] addr_peri,
	input wire clk_peri,
	output wire [17:0] di_peri,
	input wire [17:0] do_peri,
	input wire reset,
	input wire wr_peri,

	output wire intr
);

`include "common/functions/bitwidth.vh"



wire [NUM_WATCHERS-1:0] error_code;
wire [NUM_WATCHERS-1:0] error_wr;
wire [NUM_WATCHERS-1:0] error_rd;

reg error_rd_last;
reg error_code_last;

//Forward bus connections
assign 	core_code_do = error_code_last? 18'h2455 : mem_code_do; //OR r2,r2 if not executable
assign 	core_code_busy = mem_code_busy;
assign 	core_mem_busy = mem_mem_busy;
assign 	core_mem_do = error_rd_last ? 18'h3FFFF : mem_mem_di; //Return all bits set if not readable
assign 	mem_clk = core_clk;
assign 	mem_code_access = core_code_access;
assign 	mem_code_addr = core_code_addr;
assign 	mem_code_jmp = core_code_jmp;
assign 	mem_mem_access = core_mem_access;
assign 	mem_mem_addr_block = core_mem_addr_block;
assign 	mem_mem_addr_high = core_mem_addr_high;
assign 	mem_mem_do = core_mem_di;
assign 	mem_mem_wr_high = core_mem_wr_high && !(|error_wr); //Ignore write if not writable
assign 	mem_mem_wr_low = core_mem_wr_low && !(|error_wr); //Ignore write if not writable
assign 	mem_reset = core_reset;



reg[bitwidth(NUM_WATCHERS)-1:0] selectedWatcher = 0;

parameter ALL_PORTS	= 8;
parameter DECODSIM_DEB	= "NO";

wire					select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregisterg
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	),
		.reg_select()
			);

wire cs_stat_r, cs_stat_w;
wire cs_control_r, cs_control_w;
wire cs_error_addr_r, cs_error_addr_w;
wire cs_error_pc_r, cs_error_pc_w;
wire cs_watcher_idx_r, cs_watcher_idx_w;
wire cs_watcher_conf_r, cs_watcher_conf_w;
wire cs_watcher_offs_r, cs_watcher_offs_w;
wire cs_watcher_size_r, cs_watcher_size_w;

assign {cs_watcher_size_r, cs_watcher_offs_r, cs_watcher_conf_r, cs_watcher_idx_r, cs_error_pc_r, cs_error_addr_r, cs_control_r, cs_stat_r} = reg_read;
assign {cs_watcher_size_w, cs_watcher_offs_w, cs_watcher_conf_w, cs_watcher_idx_w, cs_error_pc_w, cs_error_addr_w, cs_control_w, cs_stat_w} = reg_write;


always @(posedge clk_peri) begin
	if (reset)
		selectedWatcher <= 0;
	else if (cs_watcher_idx_w)
		selectedWatcher <= do_peri;

	error_rd_last <= |error_rd;
	error_code_last <= |error_code;
end



reg [17:0] error_addr;
reg [17:0] error_pc;
reg [bitwidth(NUM_WATCHERS)-1:0] error_watcher;
reg [1:0] error_type;

wire error_now;
assign error_now = |error_code || |error_wr || |error_rd;

wire [bitwidth(NUM_WATCHERS)-1:0] error_code_source, error_wr_source, error_rd_source;

assign error_code_source = bitwidth(error_code)-1;
assign error_wr_source = bitwidth(error_wr)-1;
assign error_rd_source = bitwidth(error_rd)-1;

//Interrupt if we have a saved error, or an error exists right now (avoid 1 cycle delay until error is saved)
assign intr = |error_type || error_now;


//Simulation only stuff

always @(posedge clk_peri) begin
	if (error_now) begin: errordetectionsim
		reg[10*8-1:0] type;
		reg[17:0] source;
		if (|error_code) begin
			type = "execute";
			source = error_code_source;
		end else if (|error_wr) begin
			type = "write";
			source = error_wr_source;
		end else begin
			type = "read";
			source = error_rd_source;
		end

    	$display("memory violation at 0x%x detected by %d on %s at %t", 42, source, type, $time);

		if (STOP_SIMULATION == "TRUE") begin
			$stop();
		end
	end
end

reg enable;

always @(posedge clk_peri) begin
	if (reset) begin
		enable <= 0;
	end else begin
		if (cs_control_w)
			enable <= do_peri[0];
	end
end

reg [17:0] core_code_addr_last, core_code_addr_last2;

always @(posedge clk_peri) begin
	core_code_addr_last <= core_code_addr;
	core_code_addr_last2 <= core_code_addr_last;
end

always @(posedge clk_peri) begin

	if (reset) begin
		error_type <= 0;
		error_addr <= 0;
		error_pc <= 0;
		error_watcher <= 0;
	end else begin
		if (error_type==0) begin
			if (|error_code) begin
				error_addr <= core_code_addr;
				error_pc <= core_code_addr_last2;
				error_watcher <= error_code_source;
				error_type <= 1;
			end else if (|error_wr) begin
				error_addr <= {core_mem_addr_high, core_mem_addr_block,1'b0};
				error_pc <= core_code_addr_last2;
				error_watcher <= error_wr_source;
				error_type <= 2;
			end else if (|error_rd) begin
				error_addr <= {core_mem_addr_high, core_mem_addr_block,1'b0};
				error_pc <= core_code_addr_last2;
				error_watcher <= error_rd_source;
				error_type <= 3;
			end
		end 
		if (cs_stat_r) begin
			error_type <= 0;
			error_watcher <= 0;
		end
	end
end

wire [18*NUM_WATCHERS-1:0] watcher_read_individual;
genvar i;
generate

	for (i=0;i<NUM_WATCHERS;i=i+1) begin: access_watchers
		wire currentWatcher;
		assign currentWatcher = selectedWatcher == i;

		access_watcher watcher(
			.clk(clk_peri),
			.reset(reset),
			.code_addr(core_code_addr<<1),
			.mem_addr({core_mem_addr_high, core_mem_addr_block}),
			.mem_wr(core_mem_wr_high | core_mem_wr_low),
			.mem_rd(core_mem_access),

			.global_enable(enable),

			.read_conf(currentWatcher && cs_watcher_conf_r),
			.read_offs(currentWatcher && cs_watcher_offs_r),
			.read_size(currentWatcher && cs_watcher_size_r),

			.write_conf(currentWatcher && cs_watcher_conf_w),
			.write_offs(currentWatcher && cs_watcher_offs_w),
			.write_size(currentWatcher && cs_watcher_size_w),

			.write_data(do_peri),
			.read_data(watcher_read_individual[i*18+17:i*18]),	

			.error_code(error_code[i]),
			.error_wr(error_wr[i]),
			.error_rd(error_rd[i])

		);

	end

endgenerate

wire [17:0] watcher_read;
orgate #(
	.OR_WIDTH(NUM_WATCHERS),
	.BUS_WIDTH(18)
) or_read(
			watcher_read_individual,
			watcher_read
		);


assign 	di_peri = 	cs_stat_r ? {error_watcher, error_type} :
			cs_control_r ? {17'h0, enable} :
			cs_error_addr_r ? error_addr :
			cs_error_pc_r ? error_pc :
			cs_watcher_idx_r ? selectedWatcher :
			watcher_read;

 
endmodule


