/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`default_nettype none
//`include	"core/src/spartanmc_def.vh"

module microblaze_mem #(
	parameter RAMBLOCKS		= 4,
	parameter BASE_ADDR		= 0
) (
	input wire							clk,

	input wire		[31:0]				addra,
	input wire		[31:0]				dina,
    output wire		[31:0]				douta,
	input wire							ena,
	input wire		[3:0]				wea,
	input wire							rsta,

	input wire		[31:0]				addrb,
	input wire		[31:0]				dinb,
    output wire		[31:0]				doutb,
	input wire							enb,
	input wire		[3:0]				web,
	input wire							rstb
);

wire	[RAMBLOCKS-1:0]	access_vector_port_a;
wire	[RAMBLOCKS-1:0]	access_vector_port_b;

wire	[(RAMBLOCKS*32)-1:0] data_out_concat_a;
wire	[(RAMBLOCKS*32)-1:0] data_out_concat_b;

wire	[(RAMBLOCKS*32)-1:0] data_out_concat_pre_reset_a;
wire	[(RAMBLOCKS*32)-1:0] data_out_concat_pre_reset_b;

//buffer addresses to hold output values between read accesses
reg     [31:0] addra_reg, addrb_reg;
wire    [31:0] addra_buffered = ena ? addra : addra_reg;
wire    [31:0] addrb_buffered = enb ? addrb : addrb_reg;

always@(posedge clk) begin
    if(rsta) begin
        addra_reg <= 0;
    end else if(ena) begin
        addra_reg <= addra;
    end else begin
        addra_reg <= addra_reg;
    end
end

always@(posedge clk) begin
    if(rstb) begin
        addrb_reg <= 0;
    end else if(enb) begin
        addrb_reg <= addrb;
    end else begin
        addrb_reg <= addrb_reg;
    end
end

//concatenation or-gate
generate
	if	(RAMBLOCKS	== 1) begin
		assign	douta =	data_out_concat_a[31:0];	
		assign	doutb =	data_out_concat_b[31:0];
	end else begin
		orgate #(
			.OR_WIDTH(RAMBLOCKS),
			.BUS_WIDTH(32)
			)
		dmem_orA (
			.a(data_out_concat_a),
			.y(douta)
		);

		orgate #(
        	.OR_WIDTH(RAMBLOCKS),
        	.BUS_WIDTH(32)
        	)
        dmem_orB (
        	.a(data_out_concat_b),
        	.y(doutb)
        );
	end
endgenerate

//generate BRAM modules
genvar i;

generate
for(i = 0; i < RAMBLOCKS; i = i+1)
begin : MEM_BL

    //one-hot address decoder
	assign access_vector_port_a[i] = (addra_buffered[31:12] == i);
	assign access_vector_port_b[i] = (addrb_buffered[31:12] == i);
	
	//combinatorical reset
	assign data_out_concat_a[(i*32)+:32] = (rsta || ~access_vector_port_a[i]) ? 32'b0 : data_out_concat_pre_reset_a[(i*32)+:32];
	assign data_out_concat_b[(i*32)+:32] = (rstb || ~access_vector_port_b[i]) ? 32'b0 : data_out_concat_pre_reset_b[(i*32)+:32];
	
	
	RAMB36E1 #(
		// Address Collision Mode: "PERFORMANCE" or "DELAYED_WRITE" 
		.RDADDR_COLLISION_HWCONFIG("PERFORMANCE"),
		// Collision check: Values ("ALL", "WARNING_ONLY", "GENERATE_X_ONLY" or "NONE")
		.SIM_COLLISION_CHECK("ALL"),
		// DOA_REG, DOB_REG: Optional output register (0 or 1)
		.DOA_REG(0),
		.DOB_REG(0),
		.EN_ECC_READ("FALSE"),                                                            // Enable ECC decoder,
				                                                                        // FALSE, TRUE
		.EN_ECC_WRITE("FALSE"),                                                           // Enable ECC encoder,
				                                                                        // FALSE, TRUE
		// INIT_A, INIT_B: Initial values on output ports
		.INIT_A(36'h000000000),
		.INIT_B(36'h000000000),
		// Initialization File: RAM initialization file
		.INIT_FILE("NONE"),
		// RAM Mode: "SDP" or "TDP" 
		.RAM_MODE("TDP"),
		// RAM_EXTENSION_A, RAM_EXTENSION_B: Selects cascade mode ("UPPER", "LOWER", or "NONE")
		.RAM_EXTENSION_A("NONE"),
		.RAM_EXTENSION_B("NONE"),
		// READ_WIDTH_A/B, WRITE_WIDTH_A/B: Read/write width per port
		.READ_WIDTH_A(36),                                                                 // 0-72
		.READ_WIDTH_B(36),                                                                 // 0-36
		.WRITE_WIDTH_A(36),                                                                // 0-36
		.WRITE_WIDTH_B(36),                                                                // 0-72
		// RSTREG_PRIORITY_A, RSTREG_PRIORITY_B: Reset or enable priority ("RSTREG" or "REGCE")
		.RSTREG_PRIORITY_A("RSTREG"),
		.RSTREG_PRIORITY_B("RSTREG"),
		// SRVAL_A, SRVAL_B: Set/reset value for output
		.SRVAL_A(36'h000000000),
		.SRVAL_B(36'h000000000),
		// Simulation Device: Must be set to "7SERIES" for simulation behavior
		.SIM_DEVICE("7SERIES"),
		// WriteMode: Value on output upon a write ("WRITE_FIRST", "READ_FIRST", or "NO_CHANGE")
		.WRITE_MODE_A("WRITE_FIRST"),
		.WRITE_MODE_B("WRITE_FIRST") 
  	 ) DPRAM (
		// Cascade Signals: 1-bit (each) output: BRAM cascade ports (to create 64kx1)
		.CASCADEOUTA(),     // 1-bit output: A port cascade
		.CASCADEOUTB(),     // 1-bit output: B port cascade
		// ECC Signals: 1-bit (each) output: Error Correction Circuitry ports
		.DBITERR(),             // 1-bit output: Double bit error status
		.ECCPARITY(),         // 8-bit output: Generated error correction parity
		.RDADDRECC(),         // 9-bit output: ECC read address
		.SBITERR(),             // 1-bit output: Single bit error status
		// Port A Data: 32-bit (each) output: Port A data
		.DOADO(data_out_concat_pre_reset_a[(i*32)+:32]),                 // 32-bit output: A port data/LSB data
		.DOPADOP(),             // 4-bit output: A port parity/LSB parity
		// Port B Data: 32-bit (each) output: Port B data
		.DOBDO(data_out_concat_pre_reset_b[(i*32)+:32]),                 // 32-bit output: B port data/MSB data
		.DOPBDOP(),             // 4-bit output: B port parity/MSB parity
		// Cascade Signals: 1-bit (each) input: BRAM cascade ports (to create 64kx1)
		.CASCADEINA(),       // 1-bit input: A port cascade
		.CASCADEINB(),       // 1-bit input: B port cascade
		// ECC Signals: 1-bit (each) input: Error Correction Circuitry ports
		.INJECTDBITERR(), // 1-bit input: Inject a double bit error
		.INJECTSBITERR(), // 1-bit input: Inject a single bit error
		// Port A Address/Control Signals: 16-bit (each) input: Port A address and control signals (read port
		// when RAM_MODE="SDP")
		.ADDRARDADDR({1'b0, addra_buffered[11:0], 3'b0}),     // 16-bit input: A port address/Read address
		.CLKARDCLK(clk),         // 1-bit input: A port clock/Read clock
		.ENARDEN(access_vector_port_a[i] && ena),             // 1-bit input: A port enable/Read enable
		.REGCEAREGCE(0),     // 1-bit input: A port register enable/Register enable
		.RSTRAMARSTRAM(rsta || ~access_vector_port_a[i]), // 1-bit input: A port set/reset
		.RSTREGARSTREG(0), // 1-bit input: A port register set/reset
		.WEA(wea),                     // 4-bit input: A port write enable
		// Port A Data: 32-bit (each) input: Port A data
		.DIADI(dina),                 // 32-bit input: A port data/LSB data
		.DIPADIP(),             // 4-bit input: A port parity/LSB parity
		// Port B Address/Control Signals: 16-bit (each) input: Port B address and control signals (write port
		// when RAM_MODE="SDP")
		.ADDRBWRADDR({1'b0, addrb_buffered[11:0], 3'b0}),     // 16-bit input: B port address/Write address
		.CLKBWRCLK(clk),         // 1-bit input: B port clock/Write clock
		.ENBWREN(access_vector_port_b[i] && enb),             // 1-bit input: B port enable/Write enable
		.REGCEB(0),               // 1-bit input: B port register enable
		.RSTRAMB(rstb || ~access_vector_port_b[i]),             // 1-bit input: B port set/reset
		.RSTREGB(0),             // 1-bit input: B port register set/reset
		.WEBWE({4'b0, web}),                 // 8-bit input: B port write enable/Write enable
		// Port B Data: 32-bit (each) input: Port B data
		.DIBDI(dinb),                 // 32-bit input: B port data/MSB data
		.DIPBDIP()              // 4-bit input: B port parity/MSB parity
   );

    /*BRAM_TDP_MACRO #(
      .BRAM_SIZE("36Kb"), // Target BRAM: "18Kb" or "36Kb" 
      .DEVICE("7SERIES"), // Target device: "7SERIES" 
      .DOA_REG(0),        // Optional port A output register (0 or 1)
      .DOB_REG(0),        // Optional port B output register (0 or 1)
      .INIT_A(36'h0000000),  // Initial values on port A output port
      .INIT_B(36'h00000000), // Initial values on port B output port
      .INIT_FILE ("NONE"),
      .READ_WIDTH_A (32),   // Valid values are 1-36 (19-36 only valid when BRAM_SIZE="36Kb")
      .READ_WIDTH_B (32),   // Valid values are 1-36 (19-36 only valid when BRAM_SIZE="36Kb")
      .SIM_COLLISION_CHECK ("ALL"), // Collision check enable "ALL", "WARNING_ONLY", 
                                    //   "GENERATE_X_ONLY" or "NONE" 
      .SRVAL_A(36'h00000000), // Set/Reset value for port A output
      .SRVAL_B(36'h00000000), // Set/Reset value for port B output
      .WRITE_MODE_A("WRITE_FIRST"), // "WRITE_FIRST", "READ_FIRST", or "NO_CHANGE" 
      .WRITE_MODE_B("WRITE_FIRST"), // "WRITE_FIRST", "READ_FIRST", or "NO_CHANGE" 
      .WRITE_WIDTH_A(32), // Valid values are 1-36 (19-36 only valid when BRAM_SIZE="36Kb")
      .WRITE_WIDTH_B(32) // Valid values are 1-36 (19-36 only valid when BRAM_SIZE="36Kb")

    ) BRAM_TDP_MACRO_inst (
      .DOA(data_out_concat_a[(i*32)+:32]),       // Output port-A data, width defined by READ_WIDTH_A parameter
      .DOB(data_out_concat_b[(i*32)+:32]),       // Output port-B data, width defined by READ_WIDTH_B parameter
      .ADDRA(addra[9:0]),   // Input port-A address, width defined by Port A depth
      .ADDRB(addrb[9:0]),   // Input port-B address, width defined by Port B depth
      .CLKA(clk),     // 1-bit input port-A clock
      .CLKB(clk),     // 1-bit input port-B clock
      .DIA(dina),       // Input port-A data, width defined by WRITE_WIDTH_A parameter
      .DIB(dinb),       // Input port-B data, width defined by WRITE_WIDTH_B parameter
      .ENA(access_vector_port_a[i]),       // 1-bit input port-A enable
      .ENB(access_vector_port_b[i]),       // 1-bit input port-B enable
      .REGCEA(0), // 1-bit input port-A output register enable
      .REGCEB(0), // 1-bit input port-B output register enable
      .RSTA(rsta || !access_vector_port_a[i]),     // 1-bit input port-A reset
      .RSTB(rstb || !access_vector_port_b[i]),     // 1-bit input port-B reset
      .WEA(wea),       // Input port-A write enable, width defined by Port A depth
      .WEB(web)        // Input port-B write enable, width defined by Port B depth
   );*/

end
endgenerate


endmodule
