/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none

module microstepper(	
		clk_peri,	
		do_peri, di_peri,	
		addr_peri,	
		access_peri,	
		wr_peri,	
		reset,	
                topspeed_intr,
                deceleration_intr,
		stop_intr,
		stpr_drv, 
		stpr_en, 
		opt_enc, 
	        vref_pwm,
                isense_comp
	);

	// ld-Funktion (dualer Logarithmus, aufgerundet)
	function integer ld(input integer value);
		integer i;
		begin
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction
  
        localparam NO_IDLE_PWR      = 2'b00;
        localparam QUARTER_IDLE_PWR = 2'b01;
        localparam HALF_IDLE_PWR    = 2'b10;
        localparam FULL_IDLE_PWR    = 2'b11;

	// Die Parameter für die Adressen und die Portbreite werden 
	// in der system.v durch jConfig eingestellt.
        

	parameter BASE_ADR			=	10'h0;
	parameter RAMP_INT_CNT			=	5;

	parameter USE_OPT_ENC 			=	0;
	parameter SYSTEM_FREQUENCY 		=	25_000;		// Peripherietaktfrequenz in kHz

	parameter INITIAL_ACC_DELAY 		=	20_000;
        parameter F_CHOP                	=	25;		// Choppingfrequenz in kHz
        parameter PHASE_CURRENT         	=	670;		// Nennstrom je Motorwicklung in mA
        parameter INITIAL_PWR_DOWN_MODE 	=	"HALF_IDLE_PWR";
	parameter INITIAL_F_TOP 		=	10*1000*8;
	parameter INITIAL_F_START 		=	500*8;
        parameter MICROSTEPPING_F_MAX   	=	2*1000*8;
        parameter F_STEP_MAX            	=	11*1000*8;

        localparam F_STEP_START_MIN     	=	(SYSTEM_FREQUENCY*1000)>>18;
        localparam F_STEP_DELTA         	=	(F_STEP_MAX - F_STEP_START_MIN) >> 10;
        localparam INITIAL_STARTSPEED   	=	(INITIAL_F_START - F_STEP_START_MIN) / F_STEP_DELTA;
        localparam INITIAL_TOPSPEED     	=	(INITIAL_F_TOP - F_STEP_START_MIN) / F_STEP_DELTA;
        localparam MICROSTEPPING_TOPSPEED	=	(MICROSTEPPING_F_MAX - F_STEP_START_MIN) / F_STEP_DELTA;


        localparam INITIAL_PWR_DOWN_VALUE	= (INITIAL_PWR_DOWN_MODE == "NO_IDLE_PWR") ?	  NO_IDLE_PWR :
						  (INITIAL_PWR_DOWN_MODE == "QUARTER_IDLE_PWR") ? QUARTER_IDLE_PWR :
						  (INITIAL_PWR_DOWN_MODE == "HALF_IDLE_PWR") ?	  HALF_IDLE_PWR :
						  (INITIAL_PWR_DOWN_MODE == "FULL_IDLE_PWR") ?	  FULL_IDLE_PWR : 2'b00;


	// Festlegen der Adresssen der Peripherieregister

	localparam STARTSPEED_ADR		=	0;
	localparam RAMP_CONF_BASE_ADR		=	1; 
	localparam RAMP_CONF_END_ADR		= 	RAMP_CONF_BASE_ADR + RAMP_INT_CNT * 2 - 2; // Rampe wird mit RAMP_INT_CNT Beschleunigungswerten und (RAMP_INT_CNT-1) Drehzahlschwellen konfiguriert
	localparam TOPSPEED_ADR			=	RAMP_CONF_END_ADR + 1;
	localparam STEP_MSB_ADR			= 	RAMP_CONF_END_ADR + 2; 
	localparam STEP_LSB_ADR 		=	RAMP_CONF_END_ADR + 3;
	localparam MODE_ADR			=	RAMP_CONF_END_ADR + 4;
	localparam SPEED_ADR			=	RAMP_CONF_END_ADR + 5;
	localparam ACC_ADR			= 	RAMP_CONF_END_ADR + 6;

	localparam ADDITIONAL_REG_CNT		= 	7;
	localparam PERI_REG_CNT 		= 	RAMP_INT_CNT * 2 -1 + ADDITIONAL_REG_CNT;
	localparam PERI_REG_ADR_WIDTH		= 	ld(PERI_REG_CNT-1);
	localparam ACC_DELAY_REG_CNT		= 	RAMP_INT_CNT;
	localparam RAMP_SWITCH_SPEED_REG_CNT 	= 	RAMP_INT_CNT-1;

	parameter ALL_PORTS			=	PERI_REG_CNT;
	parameter DECODSIM_DEB			=	"NO";

	// inputs & outputs des Systembus
	input	wire		clk_peri;  		//Peripherietakt
	input	wire	[17:0]	do_peri;		// Datenbus 
	output	wire	[17:0]	di_peri;		// Datenbus
	input	wire	[9:0]	addr_peri;		// Adressbus
	input	wire		access_peri;		// chipselect Peripherie
	input	wire		wr_peri;		// Peripherie-Schreibzugriff
	input	wire		reset; 			//Reset
	output 	wire		topspeed_intr;
	output 	wire		deceleration_intr;
	output 	wire		stop_intr;
	input 	wire		opt_enc;
	output 	wire	[3:0] 	stpr_drv; 		//Schrittmotortreiber
	output 	wire	[1:0] 	stpr_en; 		//Enable für Schrittmotortreiber
        output	wire	[1:0]	vref_pwm;
        input	wire	[1:0]	isense_comp;
/*
	wire 			select;  		//wird vom Adressdecoder erzeugt

	// Adressdekoder, der aus der angelegten Adresse ein chipselect erzeugt	
	pselect p1	(
		.addr(addr_peri[9:PERI_REG_ADR_WIDTH]),
		.activ_peri(access_peri),
		.select(select)
	);
	defparam p1.ADDR_WIDTH	= 10-PERI_REG_ADR_WIDTH;
	defparam p1.BASE_WIDTH	= 10-PERI_REG_ADR_WIDTH;
	defparam p1.BASE_ADDR	= BASE_ADR >> PERI_REG_ADR_WIDTH;
*/

// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Status- und Steuerregister
wire					rd_startspeed;		// Lesen
wire					rd_ramp_conf_base;	// Lesen
wire					rd_topspeed;		// Lesen
wire					rd_step_msb;		// Lesen
wire					rd_step_lsb;		// Lesen
wire					rd_mode;		// Lesen
wire					rd_speed;		// Lesen
wire					rd_acc;			// Lesen
wire					wr_startspeed;		// Schreiben
wire					wr_ramp_conf_base;	// Schreiben
wire					wr_topspeed;		// Schreiben
wire					wr_step_msb;		// Schreiben
wire					wr_step_lsb;		// Schreiben
wire					wr_mode;		// Schreiben
wire					wr_speed;		// Schreiben
wire					wr_acc;			// Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die Status- und Steuerregister
assign	rd_startspeed		= reg_select[STARTSPEED_ADR];		// Lesen
assign	rd_ramp_conf_base	= reg_select[RAMP_CONF_BASE_ADR];	// Lesen
assign	rd_topspeed		= reg_select[TOPSPEED_ADR];		// Lesen
assign	rd_step_msb		= reg_select[STEP_MSB_ADR];		// Lesen
assign	rd_step_lsb		= reg_select[STEP_LSB_ADR];		// Lesen
assign	rd_mode			= reg_select[MODE_ADR];			// Lesen
assign	rd_speed		= reg_select[SPEED_ADR];		// Lesen
assign	rd_acc			= reg_select[ACC_ADR];			// Lesen
assign	wr_startspeed		= reg_write[STARTSPEED_ADR];		// Schreiben
assign	wr_ramp_conf_base	= reg_write[RAMP_CONF_BASE_ADR];	// Schreiben
assign	wr_topspeed		= reg_write[TOPSPEED_ADR];		// Schreiben
assign	wr_step_msb		= reg_write[STEP_MSB_ADR];		// Schreiben
assign	wr_step_lsb		= reg_write[STEP_LSB_ADR];		// Schreiben
assign	wr_mode			= reg_write[MODE_ADR];			// Schreiben
assign	wr_speed		= reg_write[SPEED_ADR];			// Schreiben
assign	wr_acc			= reg_write[ACC_ADR];			// Schreiben

// Dekodierung der Adressen ********************** END **************************************



	wire 	[35:0]	remaining_steps_outp;
	peri_reg PR0 (
		.clk_peri(clk_peri), 
		.reset(reset), 
		.inp(do_peri),
		.we(wr_step_msb), 
		.outp(remaining_steps_outp[35:18])
	);
	defparam PR0.REG_WIDTH 	= 	18;
	defparam PR0.PRESET	= 	0;

	peri_reg PR1 (
		.clk_peri(clk_peri), 
		.reset(reset), 
		.inp(do_peri),
		.we(wr_step_lsb), 
		.outp(remaining_steps_outp[17:0])
	);
	defparam PR1.REG_WIDTH 	= 	18;
	defparam PR1.PRESET	= 	0;

	wire 	[17:0]	startspeed_outp;
	peri_reg PR2 (
		.clk_peri(clk_peri), 
		.reset(reset), 
		.inp(do_peri),
		.we(wr_startspeed), 
		.outp(startspeed_outp)
	);
	defparam PR2.REG_WIDTH 	= 	18;
	defparam PR2.PRESET	= 	INITIAL_STARTSPEED;

	wire 	[17:0]	topspeed_outp;
	peri_reg PR3 (
		.clk_peri(clk_peri), 
		.reset(reset), 
		.inp(do_peri),
		.we(wr_topspeed), 
		.outp(topspeed_outp)
	);
	defparam PR3.REG_WIDTH 	= 	18;
	defparam PR3.PRESET	= 	INITIAL_TOPSPEED;

	wire 	[4:0]	mode_outp;
	reg 	[4:0]	mode_value;
	always @ (posedge clk_peri) begin
		if (reset) begin
			mode_value <= {INITIAL_PWR_DOWN_VALUE<<3,3'b000};
		end else begin
			if (wr_mode) begin
				mode_value <= do_peri;
			end else begin
				mode_value <= mode_value & 5'b11100;   // start-und notstoppbit loeschen
			end
		end
	end
	
	assign mode_outp = mode_value;

	genvar i;

	wire [17:0] acc_delay_outp[RAMP_INT_CNT-1:0];   
	generate
	for (i = 0; i < (RAMP_INT_CNT); i = i + 1) begin: WRITE_ACC_DELAY_REGS_GEN
		peri_reg ACC_DELAY_REG (
			.clk_peri(clk_peri), 
			.reset(reset), 
			.inp(do_peri),
			.we(reg_write[(RAMP_CONF_BASE_ADR + i * 2)]), 
			.outp(acc_delay_outp[i])
		);
		defparam ACC_DELAY_REG.REG_WIDTH 	= 	18;
		defparam ACC_DELAY_REG.PRESET		= 	INITIAL_ACC_DELAY;
	end
	endgenerate

	wire [17:0] ramp_switch_speed_outp[RAMP_INT_CNT-2:0];   
	generate
	for (i = 0; i < (RAMP_INT_CNT-1); i = i + 1) begin: WRITE_SWITCH_SPEED_REGS_GEN
		peri_reg SWITCH_SPEED_REG (
			.clk_peri(clk_peri), 
			.reset(reset), 
			.inp(do_peri),
			.we(reg_write[(RAMP_CONF_BASE_ADR + i * 2 + 1)]),
			.outp(ramp_switch_speed_outp[i])
		);
		defparam SWITCH_SPEED_REG.REG_WIDTH 	= 	18;
		defparam SWITCH_SPEED_REG.PRESET	= 	(INITIAL_TOPSPEED-INITIAL_STARTSPEED)/2+INITIAL_STARTSPEED;
	end
	endgenerate

	wire selected_acc_delay;
	wire idle_ind;

	wire [17:0] acc_mux_chain_nodes_out [RAMP_INT_CNT-2:0];

// Lesen der Ports

	wire [17:0] speed;
	wire [17:0] const_mux_out;
	wire [17:0] startspeed_mux_out;
	wire [17:0] mux_chain_nodes_out [RAMP_INT_CNT*2-2:0];

	reg	[17:0]	read_data;
	reg		select_v;
/*
 * Register lesen in den MEM2_WB schieben. read_data ist synchron zu pipe_2
 * ========================================================================
 */
always@(posedge clk_peri)
	if(reset) begin
		read_data <= 18'b0;
		select_v  <= 1'b0;
	end	else begin
		read_data <= mux_chain_nodes_out[RAMP_INT_CNT*2-2];
		select_v  <= select & !wr_peri;
	end

	assign	const_mux_out =	(rd_step_msb)?	remaining_steps_outp[35:18]:
				(rd_step_lsb)?	remaining_steps_outp[17:0]:
				(rd_topspeed)?	topspeed_outp:	
				(rd_speed)?	speed:	
				(rd_acc)?	selected_acc_delay:
				(rd_mode)?	{12'b0000_00000000,idle_ind,mode_outp}: 18'b0;

	generate
	for (i = 0; i < (RAMP_INT_CNT); i = i + 1) begin:READ_MUX_GEN
		if (i == 0) begin
                        di_bus_mux MUX_CHAIN_STARTSPEED_NODE (startspeed_outp, const_mux_out, (rd_startspeed), startspeed_mux_out);	
			di_bus_mux MUX_CHAIN_FIRST_NODE (acc_delay_outp[0], startspeed_mux_out, (rd_ramp_conf_base), mux_chain_nodes_out[0]);
		end else begin
			di_bus_mux MUX_CHAIN_A(ramp_switch_speed_outp[i-1], mux_chain_nodes_out[(i-1)*2], reg_select[(RAMP_CONF_BASE_ADR + i*2 -1)], mux_chain_nodes_out[i*2-1]);
			di_bus_mux MUX_CHAIN_B(acc_delay_outp[i], mux_chain_nodes_out[i*2-1], reg_select[(RAMP_CONF_BASE_ADR + i*2)], mux_chain_nodes_out[i*2]);
		end
	end
	endgenerate

	// read_data wird im EX_MEM1 geschrieben und mit select_v erst im MEM2_WB gelesen.
	assign di_peri[17:0] = select_v ? read_data : 18'b0;

// ============================ Ende des Register Lesen =================================


	wire		step_cnt_trg;
	wire		step_seq_en;
	wire		rot_dir_out;
	wire	[4:0]	step_seq_index;
	wire		step_mode;
	wire		stop_trg;

	//assign step_mode = (speed < MICROSTEPPING_TOPSPEED) ? 1'b1 : 1'b0;
       // assign step_mode = 1'b0;
        assign step_mode = 1'b1;
	//assign step_mode = (speed < 1023) ? 1'b1 : 1'b0;

        wire stop_intr_trg, topspeed_intr_trg, deceleration_intr_trg;

        generate
            if (~USE_OPT_ENC) begin

 	ramp_fsm RFSM(
		.clk(clk_peri), 
		.reset(reset), 
		.steps(remaining_steps_outp), 
		.rot_dir(mode_outp[2]), 
		.start(mode_outp[0]), 
                .estop(mode_outp[1]),
		.speed(speed), 
		.step_cnt_trg(step_cnt_trg), 
                .obs_edge(step_cnt_trg),
		.step_seq_en(step_seq_en), 
		.rot_dir_out(rot_dir_out), 
		.startspeed(startspeed_outp), 
		.acc_delay(acc_mux_chain_nodes_out[RAMP_INT_CNT-2]), 
                .topspeed_intr(topspeed_intr_trg),
                .deceleration_intr(deceleration_intr_trg),
		.topspeed(topspeed_outp),
		.stop_intr(stop_intr_trg),
                .idle(idle_ind),
		.stop_trg(stop_trg)
	);
	defparam RFSM.SYSTEM_FREQUENCY 	= 	SYSTEM_FREQUENCY;

            end else begin

	wire opt_edge;
	edge_detector ed2 (clk_peri, reset, opt_enc, opt_edge);

 	ramp_fsm RFSM(
		.clk(clk_peri), 
		.reset(reset), 
		.steps(remaining_steps_outp), 
		.rot_dir(mode_outp[2]), 
		.start(mode_outp[0]), 
                .estop(mode_outp[1]),
		.speed(speed), 
		.step_cnt_trg(step_cnt_trg), 
                .obs_edge(opt_edge),
		.step_seq_en(step_seq_en), 
		.rot_dir_out(rot_dir_out), 
		.startspeed(startspeed_outp), 
		.acc_delay(acc_mux_chain_nodes_out[RAMP_INT_CNT-2]), 
                .topspeed_intr(topspeed_intr_trg),
                .deceleration_intr(deceleration_intr_trg),
		.topspeed(topspeed_outp),
		.stop_intr(stop_intr_trg),
                .idle(idle_ind),
		.stop_trg(stop_trg)
	);
	defparam RFSM.SYSTEM_FREQUENCY 	= 	SYSTEM_FREQUENCY;


            end
        endgenerate

	reg topspeed_intr_out_reg;
	reg deceleration_intr_out_reg;
	reg stop_intr_out_reg;

	always@(posedge clk_peri)
	if(reset) begin
		topspeed_intr_out_reg <= 1'b0;
	end	else begin
		// Registerzugriff setzt Interrupt zurück
		if(rd_mode) begin
			topspeed_intr_out_reg <= 1'b0;
		end else begin
		  	if(topspeed_intr_trg == 1'b1) begin
		     	topspeed_intr_out_reg <= 1'b1;
	   		end
		end
	end

	always@(posedge clk_peri)
	if(reset) begin
		deceleration_intr_out_reg <= 1'b0;
	end	else begin
		// Registerzugriff setzt Interrupt zurück
		if(rd_mode) begin
			deceleration_intr_out_reg <= 1'b0;
		end else begin
		  	if(deceleration_intr_trg == 1'b1) begin
		     	deceleration_intr_out_reg <= 1'b1;
	   		end
		end
	end

	always@(posedge clk_peri)
	if(reset) begin
		stop_intr_out_reg <= 1'b0;
	end	else begin
		// Registerzugriff setzt Interrupt zurück
		if(rd_mode) begin
			stop_intr_out_reg <= 1'b0;
		end else begin
		  	if(stop_intr_trg == 1'b1) begin
		     	stop_intr_out_reg <= 1'b1;
	   		end
		end
	end


	assign topspeed_intr = topspeed_intr_out_reg;
	assign deceleration_intr = deceleration_intr_out_reg;
	assign stop_intr = stop_intr_out_reg;


	stepper_clk_gen SCG(
		.clk_in(clk_peri), 
		.reset(reset), 
		.step_seq_en(step_seq_en), 
		.rot_dir(rot_dir_out), 
		.speed(speed), 
		.step_seq_index(step_seq_index), 
		.step_cnt_trg(step_cnt_trg)
	);

	stepper_sequencer SS(
		.reset(reset), 
		.clk(clk_peri), 
		.mode(step_mode), 
		.step_seq_index(step_seq_index), 
		.stpr_drv(stpr_drv), 
		.stpr_en(stpr_en),
                .vref_pwm(vref_pwm),
                .isense_comp(isense_comp),
                .idle(idle_ind),
                .pwr_down_sel(mode_outp[4:3]),
		.stop_trg(stop_trg)
	);
	defparam SS.SYSTEM_FREQUENCY	= 	SYSTEM_FREQUENCY;
	defparam SS.PHASE_CURRENT	= 	PHASE_CURRENT;
	defparam SS.F_CHOP		= 	F_CHOP;


	generate
	for (i = 0; i < (RAMP_INT_CNT - 1); i = i + 1) begin:ACC_DELAY_MUX_GEN
		if (i == 0) begin
			di_bus_mux ACC_MUX_CHAIN_FIRST_NODE (acc_delay_outp[i], acc_delay_outp[i+1], speed < ramp_switch_speed_outp[i], acc_mux_chain_nodes_out[i]);
		end else begin
			di_bus_mux ACC_MUX_CHAIN(acc_mux_chain_nodes_out[i-1], acc_delay_outp[i+1], speed < ramp_switch_speed_outp[i], acc_mux_chain_nodes_out[i]);
		end
	end
	endgenerate

	assign selected_acc_delay = acc_mux_chain_nodes_out[RAMP_INT_CNT-2];


endmodule
