/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 10ns / 1ns
`default_nettype none

module ramp_fsm(
		clk, 
		reset, 
		steps, 
		rot_dir, 
		start, 
                estop,
		speed, 
		step_cnt_trg, 
                obs_edge,
		step_seq_en, 
		rot_dir_out, 
		startspeed, 
		acc_delay, 
		topspeed,
                topspeed_intr,
                deceleration_intr,
		stop_intr,
                idle,
		stop_trg
	);

	// ld-Funktion (dualer Logarithmus, aufgerundet)
	function integer ld(input integer value);
		integer i;
		begin
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction

  parameter 	SYSTEM_FREQUENCY 	= 25_000; //Peripherietaktfrequenz in kHz
  localparam 	CC_PER_US		= SYSTEM_FREQUENCY / 1000;
  localparam 	ACC_CLK_DIV_WIDTH	= ld(CC_PER_US);
  localparam 	ACC_US_CNT_WIDTH 	= 18;
  localparam 	STOP_TIME		= 6600 * SYSTEM_FREQUENCY / 25000;
  localparam	STOP_WIDTH		= ld(STOP_TIME);


  input wire			clk, reset, rot_dir, start, estop, stop_trg;
  output wire		stop_intr;
  output wire		topspeed_intr;
  output wire		deceleration_intr;
  input wire		[35:0]	steps;
  output wire		step_seq_en;
  output wire	[17:0]	speed;
  output wire		rot_dir_out;
  output wire		idle;
  input wire			step_cnt_trg;
  input wire			obs_edge;
  input wire		[17:0]	startspeed;
  input wire		[17:0]	acc_delay; // Beschleunigungsdelay in µs
  input wire		[17:0]	topspeed;
  reg			rot_dir_reg;
  reg			step_seq_en_reg;
  reg		[35:0]	steps_reg;
  reg		[35:0]	steps_mem;
  reg		[35:0]	accel_steps;
  wire		[35:0]	steps_half;
  reg		[17:0]	speed_reg;
  reg [STOP_WIDTH-1:0]	stop_clk;
  reg			stop_end;


  assign steps_half = {1'b0,steps_mem[35:1]};
  assign step_seq_en = step_seq_en_reg;
  assign speed = speed_reg;
  
  reg [35:0] steps_observer;
  reg [1:0] acceleration_mode;
   
  parameter	[2:0] IDLE		= 3'b000;  
  parameter	[2:0] APPLY		= 3'b001; 
  parameter	[2:0] ACCELERATE 	= 3'b010;  
  parameter	[2:0] TOP_SPEED		= 3'b011;  
  parameter	[2:0] DECELERATE	= 3'b100;  
  parameter	[2:0] CORRECTION	= 3'b101; 
  parameter	[2:0] FINISHED		= 3'b110; 
  parameter	[2:0] STOP		= 3'b111;	// wie IDLE aber noch kein IDLE_PWR anlegen

  reg 		[2:0] state, next_state;
  
  assign idle = (state == IDLE);
  
  wire step_error;
  assign step_error = !(steps_observer == steps_mem);


always @ (posedge stop_trg or posedge reset)
if (reset) begin 
	stop_clk <= STOP_TIME;
	stop_end <= 1'b0;
end else begin
        if ((stop_clk > 0) && (!step_seq_en_reg)) begin
	    stop_clk <= stop_clk - 1;
		if (stop_clk < 3) begin
			stop_end <= 1'b1;
		end else begin
			stop_end <= 1'b0;
		end
        end else begin
	    stop_clk <= STOP_TIME;
            stop_end <= 1'b0;
        end
end 


//  always @ (state or start or estop or steps_half or accel_steps or speed_reg or steps_reg or step_error or steps_observer or topspeed or steps_mem) begin  
always @* begin
    next_state = IDLE;
    case (state) 
      IDLE: 
        if (start) begin
          next_state = APPLY;
        end else begin
          next_state = IDLE;
        end
      APPLY:
        if (estop) begin
          next_state = FINISHED;
        end else begin
          if (steps < 2) begin
              next_state = DECELERATE;
          end else begin
      	      next_state = ACCELERATE;
          end
        end
      ACCELERATE:
        if (estop) begin
          next_state = FINISHED;
        end else begin
      	if ((speed_reg == topspeed) && (steps_reg >= steps_half)) begin
      		next_state = TOP_SPEED;
      	end else if (steps_reg < steps_half) begin
      		next_state = DECELERATE;
      	end else begin
      		next_state = ACCELERATE;
      	end 
      	end 
      TOP_SPEED:
        if (estop) begin
          next_state = FINISHED;
        end else begin
      	if (steps_reg == accel_steps) begin
      		next_state = DECELERATE;
        end else begin
        	next_state = TOP_SPEED;
        end
      	end
      DECELERATE:
        if (estop) begin
          next_state = FINISHED;
        end else begin
      	if ((steps_reg == 0) && step_error) begin
      		next_state = CORRECTION;
      	end else if ((steps_reg == 0) && !step_error) begin
      		next_state = FINISHED;
      	end else begin
      		next_state = DECELERATE;
      	end 
      	end
      CORRECTION:
        if (estop) begin
          next_state = FINISHED;
        end else begin
      	if (steps_observer < steps_mem) begin
      		next_state = CORRECTION;
      	end else begin
      		next_state = FINISHED;
      	end
      	end
      FINISHED:
		next_state = STOP;
      STOP:
	if (stop_end) begin
		next_state = IDLE;
	end else begin
		next_state = STOP;
	end

      default:	next_state = IDLE;
    endcase
  end 
  
  wire topspeed_edge;
  wire topspeed_state;
  assign topspeed_state = (state == TOP_SPEED);
  edge_detector EDTS(
        .clk_peri(clk),
        .reset(reset),
        .det_sig(topspeed_state),
        .edge_ind(topspeed_edge)
  );
  assign topspeed_intr = (topspeed_edge & topspeed_state);

  wire deceleration_edge;
  wire deceleration_state;
  assign deceleration_state = (state == DECELERATE);
  edge_detector EDDC(
        .clk_peri(clk),
        .reset(reset),
        .det_sig(deceleration_state),
        .edge_ind(deceleration_edge)
  );
  assign deceleration_intr = (deceleration_edge & deceleration_state);

   //Ausgänge der FSM:
  reg apply_inputs;
  reg stop_intr_reg;
  assign stop_intr = stop_intr_reg;

  always @ (posedge clk) begin:OUTPUT_LOGIC
    if (reset) begin
	  state <= IDLE;
	  apply_inputs		<= 1'b0;
	  acceleration_mode	<= 2'b00;
	  step_seq_en_reg	<= 1'b0;
	  stop_intr_reg		<= 1'b0;
    end else begin
	    state <= next_state;
	    case (state)
	IDLE: begin
		apply_inputs	  <= 1'b0;
		acceleration_mode <= 2'b00;
		step_seq_en_reg	  <= 1'b0;
		stop_intr_reg	  <= 1'b0;
		end
	APPLY: begin
		apply_inputs	  <= 1'b1;
		acceleration_mode <= 2'b00;
		step_seq_en_reg	  <= 1'b0;
		stop_intr_reg	  <= 1'b0;
		end
	ACCELERATE: begin
		apply_inputs	  <= 1'b0;
		acceleration_mode <= 2'b01;   	   
		step_seq_en_reg	  <= 1'b1;
		stop_intr_reg	  <= 1'b0;
		end
	TOP_SPEED: begin
		apply_inputs	  <= 1'b0;
		acceleration_mode <= 2'b11;	// ebenfalls nicht beschleunigen, aber accel_steps nicht löschen sondern für Bremsvorgang speichern
		step_seq_en_reg	  <= 1'b1;
		stop_intr_reg	  <= 1'b0;
		end
	DECELERATE: begin
		apply_inputs	  <= 1'b0;
		acceleration_mode <= 2'b10;
		step_seq_en_reg	  <= 1'b1;
		stop_intr_reg	  <= 1'b0;
		end
	CORRECTION: begin
		apply_inputs	  <= 1'b0;
		acceleration_mode <= 2'b00;
		step_seq_en_reg	  <= 1'b1;
		stop_intr_reg	  <= 1'b0;
		end
	FINISHED: begin
		apply_inputs	  <= 1'b0;
		acceleration_mode <= 2'b00;
		step_seq_en_reg	  <= 1'b1;
		stop_intr_reg	  <= 1'b1;
		end			
	STOP: begin				// wie IDLE aber noch kein IDLE_PWR anlegen
		apply_inputs	  <= 1'b0;
		acceleration_mode <= 2'b00;
		step_seq_en_reg	  <= 1'b0;
		stop_intr_reg	  <= 1'b0;
		end
	default: begin
		apply_inputs	  <= 1'b0;
		acceleration_mode <= 2'b00;
		step_seq_en_reg	  <= 1'b0;
		end
	    endcase
    end
  end // block: OUTPUT_LOGIC


  wire acc_trg;
  reg [17:0] startspeed_mem;

  always @ (posedge clk) begin
    if (reset) begin
      rot_dir_reg <= 0;
      steps_reg <= 0;
      steps_mem <= 0;
      accel_steps <= 0;
      steps_observer <= 0;
      startspeed_mem <= startspeed;
      speed_reg <= startspeed;
    end else begin
      if (apply_inputs) begin
	    rot_dir_reg <= rot_dir;
	    steps_reg <= steps;
	    steps_mem <= steps;
	    steps_observer <= 0;
	    speed_reg <= startspeed;
		startspeed_mem <= startspeed;
	  end else if (state == IDLE) begin
	  	steps_mem <= 0;
	  end else begin
	    case (acceleration_mode)
	      2'b00 : begin  //Geschwindigkeit halten
	      	speed_reg <= speed_reg;
	      	accel_steps <= 0;
	      end
	      2'b01 : begin //beschleunigen
	      	if (acc_trg) begin
	      		if (speed_reg < topspeed) speed_reg <= speed_reg + 1;
	      	end
		   	if (step_cnt_trg) accel_steps <= accel_steps + 1;
	      end
	      2'b10 : begin  //bremsen
	      	if (acc_trg) begin
	      		if (speed_reg > startspeed_mem) speed_reg <= speed_reg - 1;
	      	end
	      end
	      2'b11 : begin // Geschwindigkeit halten, aber accel_steps nicht löschen
	      	speed_reg <= speed_reg;
	      end
	    endcase
	  	if (step_cnt_trg) begin
	  		steps_reg <= steps_reg - 1;
	  	end
	  	if (obs_edge) begin
	  		steps_observer <= steps_observer + 1;
	  	end
	  end
    end
  end // always @ (posedge clk)
  
  assign rot_dir_out = rot_dir_reg;


  
////////////////////////////////////////////////////////////////////////////////////////

	reg [ACC_CLK_DIV_WIDTH-1:0] acc_clk_div;
	reg acc_trg_reg;
	reg [ACC_US_CNT_WIDTH-1:0] acc_us_cnt_reg;
	reg acc_en;

	always @ (posedge clk) begin
		if (reset) begin
			acc_en <= 1'b0;
		end else begin
			acc_en <= ((acceleration_mode == 2'b01) | (acceleration_mode == 2'b10));
		end
	end

	always @ (posedge clk) begin
		if (reset) begin
			acc_clk_div	<= 0;
			acc_us_cnt_reg	<= 0;
			acc_trg_reg	<= 1'b0;
		end else begin 
			if (!acc_en) begin
				acc_clk_div	<= 0;
				acc_us_cnt_reg	<= 0;
				acc_trg_reg	<= 1'b0;
			end else begin          		
				if (acc_clk_div < (CC_PER_US-1)) begin
					acc_clk_div <= acc_clk_div + 1;
					acc_trg_reg <= 1'b0;
				end else begin 
				acc_clk_div <= 0;
					if (acc_us_cnt_reg < acc_delay) begin
	            				acc_us_cnt_reg <= acc_us_cnt_reg + 1;
						acc_trg_reg <= 1'b0;
					end else begin
						acc_us_cnt_reg <= 0;
						acc_trg_reg <= 1'b1;
				end
			end
		end
	end
	end

	assign acc_trg = acc_trg_reg;
  
endmodule
