/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 10ns / 1ns
`default_nettype none

module stepper_sequencer(
		reset, 
		clk, 
		mode, 
		step_seq_index, 
		stpr_drv, 
		stpr_en,
                vref_pwm,
                isense_comp,
                idle,
                pwr_down_sel,
		stop_trg
	);

	// ld-Funktion (dualer Logarithmus, aufgerundet)
	function integer ld(input integer value);
		integer i;
		begin
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction

input wire		reset, clk, mode; // mode=0 --> Vollschrittbetrieb ;  mode=1 --> Mikroschrittbetrieb
input wire	[4:0]	step_seq_index; // Index für H-Brücken-LUT
output wire	[3:0]	stpr_drv;	
output wire	[1:0]	stpr_en;
output wire	[1:0]	vref_pwm;
input wire	[1:0]	isense_comp;
input wire		idle;
input wire	[1:0]	pwr_down_sel;
output wire		stop_trg;
  
parameter PHASE_CURRENT		= 670;
parameter SYSTEM_FREQUENCY	= 25_000;
parameter F_CHOP		= 25;

localparam CHOP_CLK_DIV_LIMIT	= SYSTEM_FREQUENCY / F_CHOP;
localparam CHOP_CLK_DIV_WIDTH	= ld(CHOP_CLK_DIV_LIMIT);

localparam PHASE_A_CURRENT	= PHASE_CURRENT;
localparam PHASE_B_CURRENT	= PHASE_CURRENT;

reg	[7:0]			 pwm_cnt;
reg	[CHOP_CLK_DIV_WIDTH-1:0] chop_clk;
reg				 chop_trg;
reg				 chopper_a;
reg				 chopper_b;

reg	[7:0]			 pwm1_value;
reg	[7:0]			 pwm2_value;

  always @*
  if (reset) begin
    pwm1_value <= 0;
  end else begin
    if (mode) begin
  	case (step_seq_index[3:0])
	  	4'b0000 	: pwm1_value <= 0 * PHASE_A_CURRENT * 56 / (2000 * 50);
		4'b0001	  	: pwm1_value <= 50 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b0010 	: pwm1_value <= 98 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b0011 	: pwm1_value <= 142 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b0100 	: pwm1_value <= 180 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b0101 	: pwm1_value <= 212 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b0110 	: pwm1_value <= 236 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b0111 	: pwm1_value <= 250 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b1000 	: pwm1_value <= 255 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b1001 	: pwm1_value <= 250 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b1010 	: pwm1_value <= 236 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b1011 	: pwm1_value <= 212 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b1100 	: pwm1_value <= 180 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b1101 	: pwm1_value <= 142 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b1110 	: pwm1_value <= 98 * PHASE_A_CURRENT * 56 / (2000 * 50);
	   	4'b1111 	: pwm1_value <= 50 * PHASE_A_CURRENT * 56 / (2000 * 50);                  
	   	default 	: pwm1_value <= 0 * PHASE_A_CURRENT * 56 / (2000 * 50);
  	endcase
  	end else begin
  	   pwm1_value <= 255 * PHASE_A_CURRENT * 56 / (2000 * 50);
  	end
  end
  
  reg [3:0] stpr_drv_reg;
  always @*
  if (reset) begin
    stpr_drv_reg <= 4'b0000;
  end else begin
	case (step_seq_index[4:3])
		2'b00 	: stpr_drv_reg <= 4'b1001;
		2'b01 	: stpr_drv_reg <= 4'b0101;
		2'b10 	: stpr_drv_reg <= 4'b0110;
		2'b11 	: stpr_drv_reg <= 4'b1010;
		default : stpr_drv_reg <= 4'b0000;
	endcase
  end	

  always @*
  if (reset) begin
    pwm2_value <= 0;
  end else begin
    if (mode) begin
  	case (step_seq_index[3:0])
	  	4'b0000 	: pwm2_value <= 255 * PHASE_B_CURRENT * 56 / (2000 * 50);
		4'b0001	  	: pwm2_value <= 250 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b0010 	: pwm2_value <= 236 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b0011 	: pwm2_value <= 212 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b0100 	: pwm2_value <= 180 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b0101 	: pwm2_value <= 142 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b0110 	: pwm2_value <= 98 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b0111 	: pwm2_value <= 50 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b1000 	: pwm2_value <= 0 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b1001 	: pwm2_value <= 50 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b1010 	: pwm2_value <= 98 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b1011 	: pwm2_value <= 142 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b1100 	: pwm2_value <= 180 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b1101 	: pwm2_value <= 212 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b1110 	: pwm2_value <= 236 * PHASE_B_CURRENT * 56 / (2000 * 50);
	   	4'b1111 	: pwm2_value <= 250 * PHASE_B_CURRENT * 56 / (2000 * 50);                  
	   	default 	: pwm2_value <= 0 * PHASE_B_CURRENT * 56 / (2000 * 50);
  	endcase
  	end else begin
  	   pwm2_value <= 255 * PHASE_B_CURRENT * 56 / (2000 * 50);
  	end
  end


always @ (posedge clk or posedge reset)
if (reset) begin
	pwm_cnt <= 0;
end else begin
	pwm_cnt <= pwm_cnt + 1;
end

reg [7:0] pwm1_cmp_value, pwm2_cmp_value;

always @* begin
    case ({idle,pwr_down_sel}) 
        3'b000  :   pwm1_cmp_value = pwm1_value;
        3'b001  :   pwm1_cmp_value = pwm1_value;
        3'b010  :   pwm1_cmp_value = pwm1_value;
        3'b011  :   pwm1_cmp_value = pwm1_value;
        3'b100  :   pwm1_cmp_value = 0;
        3'b101  :   pwm1_cmp_value = {2'b00,pwm1_value[7:2]};
        3'b110  :   pwm1_cmp_value = {1'b0,pwm1_value[7:1]};
        3'b111  :   pwm1_cmp_value = pwm1_value;
        default :   pwm1_cmp_value = 0;
    endcase
end

always @* begin
    case ({idle,pwr_down_sel}) 
        3'b000  :   pwm2_cmp_value = pwm2_value;
        3'b001  :   pwm2_cmp_value = pwm2_value;
        3'b010  :   pwm2_cmp_value = pwm2_value;
        3'b011  :   pwm2_cmp_value = pwm2_value;
        3'b100  :   pwm2_cmp_value = 0;
        3'b101  :   pwm2_cmp_value = {2'b00,pwm2_value[7:2]};
        3'b110  :   pwm2_cmp_value = {1'b0,pwm2_value[7:1]};
        3'b111  :   pwm2_cmp_value = pwm2_value;
        default :   pwm2_cmp_value = 0;
    endcase
end

assign vref_pwm[0] = pwm_cnt < pwm1_cmp_value;
assign vref_pwm[1] = pwm_cnt < pwm2_cmp_value;


always @ (posedge clk or posedge reset)
if (reset) begin 
	chop_clk <= 0;
	chop_trg <= 1'b0;
end else begin
        if (chop_clk < CHOP_CLK_DIV_LIMIT) begin
	    chop_clk <= chop_clk + 1;
            chop_trg <= 1'b0;
        end else begin
	    chop_clk <= 0;
            chop_trg <= 1'b1;
        end
end

assign	stop_trg = chop_trg;

always @ (posedge clk) begin
if (reset) 
	chopper_a <= 1'b0;
else if(~isense_comp[0]) 
	chopper_a <= 1'b0;
else if(chop_trg)
	chopper_a <= 1'b1;
end

always @ (posedge clk) begin
if (reset) 
	chopper_b <= 1'b0;
else if(~isense_comp[1]) 
	chopper_b <= 1'b0;
else if(chop_trg)
	chopper_b <= 1'b1;
end


assign stpr_en = {chopper_b, chopper_a};
assign stpr_drv[3:0] = stpr_drv_reg;


endmodule
