/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : port_bi.v
// Autor : Steffen Kunze
// Modul : port_bi
//
// Beschreibung : Port_bi ist ein I/O-Port mit konfigurierbarer Breite.
//                Er kann sowohl als Eingang als auch als Ausgang programmiert
//                werden. Fuer den Ausgang gibt es ein output enable. Beim
//                Eingang ist Interrupterzeugung moeglich, programmierbar fuer
//                steigende oder fallende Flanke.
//
///////////////////////////////////////////////////////////////////////////////

`default_nettype none

module port_bi(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	intr,
	pin
	);


parameter BASE_ADR	= 10'h0;
parameter PORT_WIDTH	= 18;
parameter OD_OUTPUT	= 0;

parameter DAT_ADR	= 0;
parameter IE_ADR	= 1;
parameter OE_ADR	= 2;
parameter DIR_ADR	= 3;
parameter EDGSEL_ADR	= 4;
parameter IR_STATUS_ADR	= 5;

parameter ALL_PORTS	= 6;
parameter DECODSIM_DEB	= "NO";

// inputs & outputs
input	wire				clk_peri;
input	wire	[17:0]			do_peri;	// Datenbus
output	wire	[17:0]			di_peri;	// Datenbus
input	wire	[9:0]			addr_peri;	// interner Adressbus
input	wire				access_peri;	// chipselect peripherie
input	wire				wr_peri;	// signalisiert Schreibzugriff Peripherie
input	wire				reset;
output	wire				intr;		// interrupt flag
inout	wire	[PORT_WIDTH-1:0]	pin;		// I/O-Pins


reg		[PORT_WIDTH-1:0]	ir_sig;		// interrupt Register
reg		[PORT_WIDTH-1:0]	oe;		// output enable, 1=enable
reg		[PORT_WIDTH-1:0]	int_e;		// interrupt enable, 1=enable
reg		[PORT_WIDTH-1:0]	dir;		// 0=Ausgang,1=Eingang
reg		[PORT_WIDTH-1:0]	edge_sel;	// interrupt edge select, 1=steigend,0=fallend

wire		[PORT_WIDTH-1:0]	in;
reg		[PORT_WIDTH-1:0]	out;		// datenreg fuer output
reg		[PORT_WIDTH-1:0]	dat;		// datenreg fuer input
reg		[PORT_WIDTH-1:0]	dat_a;		// buffer fuer input

// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					rd_dat;		// Lesen
wire					rd_ie;		// Lesen
wire					rd_oe;		// Lesen
wire					rd_dir;		// Lesen
wire					rd_edgsel;	// Lesen
wire					rd_irstatus;	// Lesen
wire					wr_dat;		// Schreiben
wire					wr_ie;		// Schreiben
wire					wr_oe;		// Schreiben
wire					wr_dir;		// Schreiben
wire					wr_edgsel;	// Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	rd_dat		= reg_read[DAT_ADR];		// Lesen
assign	rd_ie		= reg_read[IE_ADR];		// Lesen
assign	rd_oe		= reg_read[OE_ADR];		// Lesen
assign	rd_dir		= reg_read[DIR_ADR];		// Lesen
assign	rd_edgsel	= reg_read[EDGSEL_ADR];		// Lesen
assign	rd_irstatus	= reg_read[IR_STATUS_ADR];	// Lesen
assign	wr_dat		= reg_write[DAT_ADR];		// Schreiben
assign	wr_ie		= reg_write[IE_ADR];		// Schreiben
assign	wr_oe		= reg_write[OE_ADR];		// Schreiben
assign	wr_dir		= reg_write[DIR_ADR];		// Schreiben
assign	wr_edgsel	= reg_write[EDGSEL_ADR];	// Schreiben

// Dekodierung der Adressen ********************** END **************************************


// Wenn Adresse des Datenregister anliegt, wird selbiges an den Bus gelegt.
// immer die Leitungen selbst lesen und nicht das out oder dat Register.
assign di_peri[17:0]	= (rd_dat)?		dat:
       			  (rd_ie)?		int_e:
			  (rd_oe)?		oe:
			  (rd_dir)?		dir:
			  (rd_edgsel)?		edge_sel:
			  (rd_irstatus)?	ir_sig: 18'b0;

// Wenn Interrupts enabled sind, wird das Interruptregister ausgegeben.
assign	 intr		= |(int_e & ir_sig);

// ***********************************************************
// Bei Verwendung einer buffer instance sind die Einstellungen
// in der UCF Datei fuer das Pin nicht wirksam!
// 
// Spartan-3 inout buffer instance
//IOBUF inoutp[PORT_WIDTH-1:0] (
//	.O(in),		// Buffer output
//	.IO(pin),	// Buffer inout port
//	.I(out),	// Buffer input
//	.T(~(oe&~dir))	// 3-state enable input
//	);
//   defparam IOBUF_inst.DRIVE = 12;
//   defparam IOBUF_inst.IOSTANDARD = "LVCMOS25";
//   defparam IOBUF_inst.SLEW = "SLOW";
// ***********************************************************

// Generierung der PIN Treiber

genvar  j;
generate
for (j=0; j<PORT_WIDTH; j=j+1)
begin:GEN_IOTBUF
	if	(OD_OUTPUT	== 0)
        	assign  pin[j]  = (oe[j]&~dir[j])?		out[j]: 1'bz;	// tri-state output
	else
        	assign  pin[j]  = (oe[j]&~dir[j]&~out[j])?	1'b0: 1'bz;	// open drain output
        end
endgenerate


// In diesem Block ist das Schreiben der Steuer- und Datenregister
// durch den Datenbus implementiert
	always @ (posedge clk_peri)
	if (reset)
		begin
		int_e	 <= 0;
		edge_sel <= 0;
		out	 <= 0;
		dir	 <= 0;
		oe	 <= 0;
		end
	else
		begin
		if(wr_dat)			// Daten
			out	 <= do_peri;
		if(wr_oe)			// Output enable
			oe	 <= do_peri;
		if(wr_ie)			// interrupt enable
			int_e	 <= do_peri;
		if(wr_edgsel)			// edge select
			edge_sel <= do_peri;
		if(wr_dir)			// I/O-Richtung
			dir	 <= do_peri;
		end

reg	[4:0]	i;	//Hilfsvariable

// In diesem Block werden Daten vom Inputpin gelesen und
// es werden Interrupts generiert/zurueckgesetzt
	always @ (posedge clk_peri)
	if(reset)
		begin
		ir_sig	<= 0;
		dat	<= 0;
		dat_a	<= 0;
		end
	else
		begin
		dat_a	<= dat;
		dat	<= pin;	// Inputs oder Zuruecklesen der Outputs
		// Die Schleife ueberprueft nacheinander alle pins.
		for(i=0; i<PORT_WIDTH; i=i+1)

		// Falls es inputpins sind, nach der Flanke suchen.
		if(dir[i])
			begin

			// Interrupt setzen fuer:

			// steigende Flanke
			if(edge_sel[i] && dat_a[i] == 1'b0 && dat[i] == 1'b1)
				ir_sig[i] <= 1'b1;
			else
			// oder fallende Flanke
			if(!edge_sel[i] && dat_a[i] == 1'b1 && dat[i] == 1'b0)
				ir_sig[i] <= 1'b1;

			end
		// Interrupt zuruecksetzen
		if(rd_dat || rd_ie || rd_edgsel || rd_dir || rd_oe)
			ir_sig <= 0;
		end

endmodule
