/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : port_in_f.v
// Autor : Siegmar Schöne
// Modul : port_in_f
//
// Beschreibung : Port_in_f ist ein Eingangs-Port mit konfigurierbarer Breite.
//                Es ist eine Ereignisabfrage, programmierbar fuer steigende
//                oder fallende Flanke moeglich. Die Ereignisse werden mit dem
//                clk_ext, welcher groesser ist als clk_peri abgetastet und in
//                den FIFO geschrieben. Sobald ein Ereinis gefunden wurde
//                wird Interrupt ausgeloest. Die Anzahl der Ereignisse darf
//                die Groesse des FIFO bis zur Abholung nicht uebersteigen!
///////////////////////////////////////////////////////////////////////////////

`timescale      1ns / 10ps
`default_nettype none

module port_in_f(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	intr,
	clk_ext,
	pin
	);

// synthesis attribute iob of port_in is TRUE;

function integer log2;

        input integer value;
        
        for (log2=0; value>0; log2=log2+1)
                value = value>>1;
                
endfunction

// Die Parameter fuer die Adressen und die Portbreite werden
// in der system.v durch jConfig eingestellt.
parameter PORT_WIDTH	= 18;
parameter FIFO_DEPTH	= 8;
parameter BASE_ADR	= 10'h9;

parameter DAT_ADR	= 3'h0;
parameter IE_ADR	= 3'h1;
parameter MASKE_ADR	= 3'h2;
parameter DAT_OLD_ADR	= 3'h3;
parameter FF_STATE	= 3'h4;

parameter FIFO_ADDR_BITS   = log2(FIFO_DEPTH - 1);

parameter ALL_PORTS	= 5;
parameter DECODSIM_DEB	= "NO";

// inputs & outputs des Systembus
input wire				clk_peri;
input wire	[17:0]			do_peri;	// Datenbus
output wire	[17:0]			di_peri;	// Datenbus
input wire	[9:0]			addr_peri;	// Adressbus
input wire				access_peri;	// CS Peripherie
input wire				wr_peri;	// signalisiert Schreibzugriff Peripherie
input wire				reset;
output wire				intr;		// interrupt flag
// Signale des Ports
input wire				clk_ext;	// clk der pin Quelle
input wire	[PORT_WIDTH-1:0]	pin; 		// Eingangspins

wire	[PORT_WIDTH-1:0]	in;		// Eingangssignal

assign				in	=	pin;

reg	[PORT_WIDTH-1:0]	int_e;		// interrupt enable, 0=disable
reg	[PORT_WIDTH-1:0]	bit_maske;	// int edge select, 0-fallend,1-steigend
reg	[PORT_WIDTH-1:0] 	ir_sig;		// interrupt register
reg	[PORT_WIDTH-1:0] 	dat;		// Datenregister
reg	[PORT_WIDTH-1:0] 	dat_o;		// input buffer

reg				fifo_wr;
reg				fifo_ue;	// FIFO Ueberlauf
reg				ir_sigl;	// Interrups loeschen

wire	[2*PORT_WIDTH-1:0]	fifo_di;	// FIFO Daten Input

// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					rd_dat;		// Lesen
wire					rd_ie;		// Lesen
wire					rd_maske;	// Lesen
wire					rd_dat_old;	// Lesen
wire					rd_ff_state;	// Lesen
wire					wr_ie;		// Schreiben
wire					wr_maske;	// Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	rd_dat		= reg_read[DAT_ADR];		// Lesen
assign	rd_ie		= reg_read[IE_ADR];		// Lesen
assign	rd_maske	= reg_read[MASKE_ADR];		// Lesen
assign	rd_dat_old	= reg_read[DAT_OLD_ADR];	// Lesen
assign	rd_ff_state	= reg_read[FF_STATE];		// Lesen
assign	wr_ie		= reg_write[IE_ADR];		// Schreiben
assign	wr_maske	= reg_write[MASKE_ADR];		// Schreiben

// Dekodierung der Adressen ********************** END **************************************

wire	[2*PORT_WIDTH-1:0]	dat_ff;
wire	[PORT_WIDTH-1:0]	dat_ff_h;
wire	[PORT_WIDTH-1:0]	dat_ff_l;
wire	[PORT_WIDTH-1:0]	state_ff;
wire				fifo_full;
wire				fifo_empty;
wire	[PORT_WIDTH-1:0]	event_bit;

assign	state_ff	= {15'b0, fifo_ue, fifo_full, fifo_empty};
assign	dat_ff_h	= {dat_ff[2*PORT_WIDTH-1:PORT_WIDTH]};
assign	dat_ff_l	= {dat_ff[PORT_WIDTH-1:0]};

// Wenn Adresse des Datenregister anliegt, wird selbiges an den Bus gelegt.
assign	di_peri[17:0]	= (rd_dat)?	   dat_ff_h:
			  (rd_ie)?	   int_e:
			  (rd_maske)?	   bit_maske:
			  (rd_dat_old)?    dat_ff_l:
			  (rd_ff_state)?   state_ff:18'h0;

// Wenn Interrupts enabled sind, wird das Interruptregister ausgegeben solange der FIFO nicht voll ist.
//assign	intr		= (|(int_e & ir_sig)) & !fifo_ue;
assign	intr		= (|(int_e & ir_sig));


// In diesem Block werden Daten vom Bus in die Register uebernommen.
always @ (posedge clk_peri)
if (reset)
	begin
	int_e		<= 0;
	bit_maske	<= 0;
	ir_sigl		<= 0;
	end
else
	begin
	if(wr_ie)
		int_e		<= do_peri;
	if(wr_maske)
		bit_maske	<= do_peri;
	// Interrupt zuruecksetzen
	if (rd_dat || rd_ie || rd_maske)
		ir_sigl <= 1;
	else if (ir_sig == 0)
		ir_sigl <= 0;
	end

// Spartan-3 input buffer instance
//IBUF inp[PORT_WIDTH-1:0] (
//			.O (in),
//			.I (pin));


// In diesem Block werden Daten vom Inputpin gelesen und
// es werden Interrupts generiert/zurueckgesetzt
// Die Abtastfrequenz clk_ext kann groesser als clk_peri sein. Die
// erkannten Ereignisse werden daher in einen FIFO geladen.
assign	event_bit	= ((bit_maske & dat_o) ^ (bit_maske & dat));

always @ (posedge clk_ext)
if(reset)
	begin
	ir_sig	<= 0;
	fifo_wr	<= 1'b0;
	fifo_ue	<= 1'b0;
	dat	<= 0;
	dat_o	<= 0;
	end
else
	begin
	// Daten lesen
	fifo_wr	<= 1'b0;
	dat_o	<= dat;
	dat	<= in;
	if (fifo_empty) fifo_ue	<= 1'b0;

	// Dieser Block ueberprueft welche maskierten pins sich geaendert haben.
	// Wenn an einem der Pins eine Flanke erkannt, dann wird
	// das Interruptregister gesetzt und der FIFO geschrieben.
	if (|event_bit)
		begin
		ir_sig	<= ir_sig | event_bit;
		fifo_wr	<= 1'b1;
		fifo_ue	<= fifo_full;
		end

	// Interrupt zuruecksetzen
	if (ir_sigl)
		ir_sig 	<= 0;
//		fifo_ue	<= 1'b0;
	end

// Bei jedem gefundenen Ereignis wird das Datenport und das
// Interrupt Status Port in den FIFO gespeichert.
assign	fifo_di	= {dat, ir_sig};

        yaf #
        (
                .DSIZE (	2*PORT_WIDTH	),
                .ASIZE (	FIFO_ADDR_BITS	)
        )
        yaf_rx0
        (
                .rdata(		dat_ff		),      //Empfangsdaten zum MC
                .wfull(                         ),      //Empfangs-FIFO voll mit wclk synchronisiert
                .rfull(                         ),      //Empfangs-FIFO voll mit rclk synchronisiert
                .rwfull(	fifo_full	),      //Empfangs-FIFO voll = rfull | wfull
                .wempty(                        ),      //Empfangs-FIFO leer mit wclk synchronisiert
                .rempty(                        ),      //Empfangs-FIFO leer mit rclk synchronisiert
                .rwempty(	fifo_empty	),      //Empfangs-FIFO leer = rempty & wempty
		.fill(				),	//Output wird nicht verwendet.
                .wdata(		fifo_di		),	//mit der Abtastfrequenz gelesene Daten
                .winc(		fifo_wr		), 
                .wclk(		clk_ext		),	//Abtastfrequenz > clk_peri
                .wrst_n(	~reset		),
                .rinc(		rd_dat		),  
                .rclk(		clk_peri	),  
                .rrst_n(	~reset		),
		.rptr_upd(	1'b0		),
		.rptr_rest(	1'b0		),
		.wptr_upd(	1'b0		),
		.wptr_rest(	1'b0		)

        );

endmodule
