/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : rot_ta.v
// Autor : Siegmar Schoene
// Modul : rot_ta
//
// Beschreibung : Interface fuer einen Rotations Geber und Taster..
//                Es ist eine Interrupterzeugung moeglich, programmierbar fuer
//                steigende oder fallende Flanke.
///////////////////////////////////////////////////////////////////////////////

`default_nettype none

module rot_ta(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	intr,
	rot_a,
	rot_b,
	rot_p
	);

// Der Parameter fuer die Adressen des Moduls wird
// in der system.v durch jConfig eingestellt.

parameter BASE_ADR	= 10'h9;
parameter PORT_WIDTH	= 5;			// rechts, links, rot_a, rot_b, tast

parameter DAT_ADR	= 0;
parameter IE_ADR	= 1;
parameter EDGSEL_ADR	= 2;
parameter IR_STATUS_ADR	= 3;
parameter COUNT		= 4;

parameter ALL_PORTS	= 5;
parameter DECODSIM_DEB	= "NO";


// inputs & outputs des Systembus
input	wire				clk_peri;
input	wire	[17:0]			do_peri;	// Datenbus
output	wire	[17:0]			di_peri;	// Datenbus
input	wire	[9:0]			addr_peri;	// Adressbus
input	wire				access_peri;	// CS Peripherie
input	wire				wr_peri;	// signalisiert Schreibzugriff Peripherie
input	wire				reset;
output	wire				intr;		// interrupt flag
// zusaetzliche Signale fuer Zaehlen der Impulse



// Signale des Ports
input	wire				rot_a;		// Eingangspins vom Rotations Taster
input	wire				rot_b;
input	wire				rot_p;
wire		[PORT_WIDTH-3:0]	rot_ta = {rot_a, rot_b, rot_p};

wire		[PORT_WIDTH-3:0]	in;		// Eingangssignal vom Rotations Taster

assign					in	=	rot_ta;

reg		[PORT_WIDTH-1:0] 	dat;		// Datenregister
reg		[PORT_WIDTH-1:0] 	dat_o;		// input old buffer

reg					rot_q1;		// zum Entprellen
reg					rot_q2;		// zum Entprellen
reg					q1_old;		// zum Entprellen

reg		[PORT_WIDTH-1:0]	int_e;		// interrupt enable, 0=disable
reg		[PORT_WIDTH-1:0]	edge_sel;	// int edge select, 0-fallend,1-steigend
reg		[PORT_WIDTH-1:0] 	ir_sig;		// interrupt register

reg		[17:0]			count;		// Zaehlsignal
reg		[17:0]			count1;		// zaehlt Impulse +1 1*rechts, -1 1*links

reg					count_reset;

// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					rd_dat;		// Lesen
wire					rd_ie;		// Lesen
wire					rd_edgsel;	// Lesen
wire					rd_irstatus;	// Lesen
wire					rd_count;	// Lesen
wire					wr_ie;		// Schreiben
wire					wr_edgsel;	// Schreiben
wire					wr_count;	// Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	rd_dat		= reg_read[DAT_ADR];		// Lesen
assign	rd_ie		= reg_read[IE_ADR];		// Lesen
assign	rd_edgsel	= reg_read[EDGSEL_ADR];		// Lesen
assign	rd_irstatus	= reg_read[IR_STATUS_ADR];	// Lesen
assign	rd_count	= reg_read[COUNT];		// Lesen
assign	wr_ie		= reg_write[IE_ADR];		// Schreiben
assign	wr_edgsel	= reg_write[EDGSEL_ADR];	// Schreiben
assign	wr_count	= reg_write[COUNT];		// Schreiben

// Dekodierung der Adressen ********************** END **************************************

reg					rd_ff1;		// Daten lesen merken

// Wenn Adresse des Datenregister anliegt, wird selbiges an den Bus gelegt.
assign	di_peri[17:0]	= (rd_dat)?	   dat:
			  (rd_ie)?	   int_e:
			  (rd_edgsel)?	   edge_sel:
			  (rd_irstatus)?   ir_sig:	
			  (rd_count)?	   count1:	18'h0;

// Wenn Interrupts enabled sind, wird das Interruptregister ausgegeben.
assign	intr		= |(int_e & ir_sig);

// In diesem Block werden Daten vom Bus in die Register uebernommen.
always @ (posedge clk_peri)
if (reset)
	begin
	int_e		<= 0;
	edge_sel	<= 0;
	count		<= 0;
	count_reset	<= 1;
	end
else
	begin
	count_reset	<= 0;
	if(wr_ie)
		int_e		<= do_peri;
	if(wr_edgsel)
		edge_sel	<= do_peri;
	if(wr_count)
		begin
			count		<= do_peri;
			count_reset	<= 1;
		end
	end

// ***********************************************************
// Bei Verwendung einer buffer instance sind die Einstellungen
// in der UCF Datei fuer das PIN nicht wirksam!
//
// Spartan-3 input buffer instance
//IBUF inp[PORT_WIDTH-3:0] (
//			.O (in),
//			.I (rot_ta));
// ***********************************************************

reg	[2:0]	i;	// Hilfsvariable (Geber hat nur 5 Bit)

// In diesem Block werden Daten vom Inputpin gelesen und
// es werden Interrupts generiert/zurueckgesetzt
always @ (posedge clk_peri)
if(reset)
	begin
	ir_sig	<= 0;
	dat	<= 0;
	dat_o	<= 0;
	rd_ff1	<= 0;
	q1_old	<= 0;
	rot_q1	<= 0;
	rot_q2	<= 0;
	end

else
	
	begin
	// zaehler neu setzen
	if (count_reset)		
		count1 <= count;
	
	// Daten lesen und vorherige Daten merken
	dat_o		<= {2'b00, dat[2:0]};
	dat[2:0]	<= in;
	q1_old		<= rot_q1;



	// Rotationsgeber entprellen
	case	(dat[2:1])
		2'b00:	begin
			rot_q1	<= 0;
			end
		2'b01:	begin
			rot_q2	<= 0;
			end
		2'b10:	begin
			rot_q2	<= 1;
			end
		2'b11:	begin
			rot_q1	<= 1;
			end
	endcase

	if (rd_ff1)
			// loeschen der Bits nach dem Lesen
			begin
			dat[4]	<=	1'b0;
			dat[3]	<=	1'b0;
			rd_ff1	<=	1'b0;
			end

	// Zugriff auf Datenregister merken
	else	if(rd_irstatus && (dat[4] || dat[3]))
			begin
			rd_ff1	<=	1'b1;		// Bits wurden gelesen --> loeschen
			end



	if (rot_q2 && !q1_old && rot_q1)
	     	begin
			if (!dat[4] && !dat[3])
				dat[4]	<= 1'b1;		// Rotation im Uhrzeigersinn.
			count1  <=       count1+1;
		end	
	else	if (!rot_q2 && !q1_old && rot_q1)
	   	begin
			if (!dat[4] && !dat[3])
				dat[3]	<= 1'b1;		// Rotation gegen den Uhrzeigersinn.
			count1  <= 	count1-1;
		end	

	// Die Schleife ueberprueft nacheinander alle Pins.
	// Wenn an einem der Pins eine Flanke detektiert wird,
	// wird das Interruptregister gesetzt.
	for (i=0; i < PORT_WIDTH; i = i+1)
		begin
		// steigende Flanke
		if ( edge_sel[i] && dat_o[i] == 1'b0 && dat[i] == 1'b1)
			ir_sig[i] <= 1'b1;
		else
		// fallende Flanke
		if (!edge_sel[i] && dat_o[i] == 1'b1 && dat[i] == 1'b0)
			ir_sig[i] <= 1'b1;
		end

	// Interrupt zuruecksetzen beim lesen der Register
	if (rd_dat || rd_ie || rd_edgsel)
		ir_sig <= 0;
	end

endmodule
