-------------------------------------------------------------------------------------------
--  Projekt	 : SpartanMC
--  Dateiname	 : rot_ta_vhdl.vhd
--  Autor	 : Siegmar Schoene, umgeshrieben in VHDL 29.09.2008
--  Modul	 : rot_ta_vhdl
--  Beschreibung : Interface fuer einen Rotations Geber und Taster..
--                 Es ist eine Interrupterzeugung moeglich, programmierbar fuer
--                 steigende oder fallende Flanke.
--
--
--  Revision 1.0  2007/11/23 10:55:00  schoene
--  Dekodierung fuer Rotation rechts und Rotation links.
--
--  Revision 0.0  2007/11/23 09:54:00  aus port_in von kunze
--  no message
--
--------------------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL, IEEE.NUMERIC_STD.all;

entity rot_ta_vhdl is


generic	(	DAT_ADR		        : std_ulogic_vector(2 downto 0):= "000";
		    IE_ADR		: std_ulogic_vector(2 downto 0):= "001";
		    EDGSEL_ADR   	: std_ulogic_vector(2 downto 0):= "010";
		    IR_STATUS_ADR	: std_ulogic_vector(2 downto 0):= "011";
		    COUNT	       	: std_ulogic_vector(2 downto 0):= "100";	-- Register zaehlt rechts/links Impulse
		    PORT_WIDTH	        : integer := 5  );


port 	(	clk_peri	 : in  std_ulogic;
		do_peri		 : in  std_ulogic_vector(17 downto 0);	-- Datenbus
		di_peri		 : out std_ulogic_vector(17 downto 0);	-- Datenbus
		addr_peri	 : in  std_ulogic_vector( 2 downto 0);	-- Adressbus
		access_peri	 : in  std_ulogic;			-- CS Peripherie
		wr_peri		 : in  std_ulogic;			-- signalisiert Schreibzugriff Peripherie
		reset	     	 : in  std_ulogic;
		intr	      	 : out std_ulogic;		 	-- interrupt flag
		select_copy	 : in  std_ulogic;			-- Eine Adresse des Moduls wird angesprochen

		rot_a		 : in  std_ulogic;			-- Eingangspins vom Rotations Taster
		rot_b		 : in  std_ulogic;
		rot_p		 : in  std_ulogic  );

end rot_ta_vhdl;

architecture only of rot_ta_vhdl is

	signal	in_rot_all	  : std_ulogic_vector(2 downto 0);
	signal	dat		  : std_ulogic_vector(4 downto 0);		    -- Datenregister
	signal	dat_old		  : std_ulogic_vector(4 downto 0);		-- input old buffer

	signal	rot_q1		   : std_ulogic;				-- zum Entprellen
	signal	rot_q2		   : std_ulogic;				-- zum Entprellen
	signal	q1_old		   : std_ulogic;				-- zum Entprellen

	signal	int_e		  : std_ulogic_vector(4 downto 0);		     -- interrupt enable, 0=disable 	
	signal	edge_sel	  : std_ulogic_vector(4 downto 0);		     -- int edge select, 0-fallend,1-steigen
	signal	ir_sig		  : std_ulogic_vector(4 downto 0);		     -- interrupt register  

	signal	count_neu   	: unsigned(17 downto 0);	  -- Zaehlsignal
	signal	count1  	: unsigned(17 downto 0);	  -- zaehlt Impulse +1 1*rechts, -1 1*links
	signal	count_reset	: std_ulogic;

	signal	rd_ff1		: std_ulogic; 

	signal	temp1		: std_ulogic_vector(4 downto 0);



begin
  
--  Wenn Adresse eines Datenregister anliegt, wird selbiges an den Bus gelegt

	di_peri(17 downto 0) 		<=	("0000000000000"&dat)	      when ((addr_peri(2 downto 0) = DAT_ADR)     and (select_copy = '1')) else
 				               ("0000000000000"&int_e)	    when ((addr_peri(2 downto 0) = IE_ADR)        and (select_copy = '1')) else
				               ("0000000000000"&edge_sel)  when ((addr_peri(2 downto 0) = EDGSEL_ADR)    and (select_copy = '1')) else
				               ("0000000000000"&ir_sig)    when ((addr_peri(2 downto 0) = IR_STATUS_ADR) and (select_copy = '1')) else 
				               std_ulogic_vector(count1)   when ((addr_peri(2 downto 0) = COUNT)		       and (select_copy = '1')) else	
				               "000000000000000000";


--  Wenn Interrupts enabled sind, wird das Interruptregister ausgegeben.

	temp1	<= (int_e  and ir_sig);
	intr	<= (temp1(0) or temp1(1) or temp1(2) or temp1(3) or temp1(4));


-- In diesem Block werden Daten vom Bus in die Register uebernommen.

proc1: process (clk_peri,reset)

begin
	if (reset = '1')  then
		int_e       	<= "00000";
		edge_sel 	   <= "00000";
		count_neu 	  <= "000000000000000000";
		count_reset 	<= '1';

	elsif (rising_edge(clk_peri)) then
 		count_reset	<= '0';

		if (((addr_peri(2 downto 0))= IE_ADR)     and (select_copy = '1') and (wr_peri = '1')) then
	    		int_e     <= do_peri(4 downto 0); 
		end if;

		if (((addr_peri(2 downto 0))= EDGSEL_ADR) and (select_copy = '1') and (wr_peri = '1')) then
			edge_sel  <= do_peri(4 downto 0);
		end if;

		if (((addr_peri(2 downto 0)) = COUNT)     and (select_copy = '1') and(wr_peri = '1')) then
			count_neu    <= unsigned(do_peri);
			count_reset  <= '1';
		end if;
	end if;

end process;


-- Interface fuer der Rotations Taster auf dem Spartan 3e Starter Kit

--  In diesem Block werden Daten von den Inputpins gelesen und
--  es werden Interrupts generiert/zurueckgesetzt

proc2: process (clk_peri) 

	begin
	    
	if (rising_edge(clk_peri)) then
		if (reset = '1') then
			ir_sig	 <= "00000";
			dat     <= "00000";
			dat_old	<= "00000";
			rd_ff1	 <= '0';
			q1_old	 <= '0';
			rot_q1	 <= '0';
			rot_q2	 <= '0';	
			
		else					                    
			if (count_reset = '1') then		   --  Zaehler neu setzen
				  count1 <= count_neu;
			end if;

			in_rot_all(0)		<= rot_p;        --  Daten lesen und vorherige Daten merken
			in_rot_all(1)		<= rot_b;
			in_rot_all(2)		<= rot_a;

			dat_old(2 downto 0)	<= dat(2 downto 0);
			dat_old(4 downto 3)	<= "00";
			
			dat(2 downto 0)		<= in_rot_all;
			q1_old			<= rot_q1;

			-- Rotationsgeber entprellen
			case  dat(2 downto 1) is
				when "00"   =>	rot_q1	<= '0';
						         rot_q2	<= rot_q2;

				when "01"   =>	rot_q1	<= rot_q1;
						         rot_q2	<= '0';

				when "10"   =>	rot_q1	<= rot_q1;
						         rot_q2	<= '1';
				                
	
				when "11"   =>	rot_q1	<= '1';
						         rot_q2	<= rot_q2;

				when others =>	rot_q1	<= rot_q1;
						         rot_q2	<= rot_q2;
			end case;

			if	(rd_ff1 = '1') then
				-- loeschen der Bits nach dem Lesen
				dat(4)	<=  '0';
				dat(3)	<=  '0';
				rd_ff1	<=  '0';

			elsif	(( ( rot_q2  and  not(q1_old) and rot_q1 ) ) = '1')	then
				if ((not(dat(4)) and not(dat(3))) = '1') then
				   dat(4)	<= '1';	-- Rotation im Uhrzeigersinn.
				end if;
				count1 <= count1 + "000000000000000001";

			elsif	(( ( not(rot_q2) and  not(q1_old) and rot_q1 ) ) = '1')	then
				if ((not(dat(4)) and not(dat(3))) = '1') then
				   dat(3)	<= '1'; -- Rotation gegen den Uhrzeigersinn.
				end if;
				count1 <= count1 - "000000000000000001";
				
				
			 elsif	((addr_peri(2 downto 0)= IR_STATUS_ADR) and (select_copy and (dat(4) or dat(3))) = '1')	then
                                     
				  rd_ff1	<= '1'; -- Bits wurden gelesen --> loeschen
                                     
		     end if;	
                                     
		--  Die Schleife ueberprueft nacheinander alle Pins.
	   --  Wenn an einem der Pins eine Flanke detektiert wird,                                    					
		--  wird das Interruptregister gesetzt.
		
		
		

			for i in 0 to 4 loop

				-- steigende Flanke

				if	(((    edge_sel(i)  and dat_old(i)) = '0') and (dat(i) = '1')) then
					ir_sig(i) <= '1';

			   -- fallende Flanke

		   		elsif	(((not(edge_sel(i)) and dat_old(i)) = '1') and (dat(i) = '0')) then
					ir_sig(i) <= '1';
				
				end if;
			
			end loop;
      
			--  Interrupt zuruecksetzen beim Lesen der Register

			if	((addr_peri(2 downto 0) = DAT_ADR)    and (select_copy = '1'))	then
				ir_sig <= "00000";

			elsif	((addr_peri(2 downto 0) = IE_ADR)     and (select_copy = '1'))	then
				ir_sig <= "00000";

			elsif	((addr_peri(2 downto 0) = EDGSEL_ADR) and (select_copy = '1'))	then
				ir_sig <= "00000";
           
			end if;
	   
		end if; 
	end if; 
  end process;
end;
