/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/* 

---OOOO---------OOOO               .//
   O   O--------OOOOO------     .//
-x-OOOO    /----OOOO         .//
 | O  O    |              .//
 | O   O   |           .//
 |         |        .//
 |-OOOO    |     .//        
   OOOO----/  .//          //==\\
---OOOO    .//             ||
        .//           010  \\==\\
     .//           010001      ||
  .//           100010100  \\==//
//                         


Created spring 2016 by Felix Sterzelmaier
TU Darmstadt
---Fachgebiet Rechnersysteme---
License: CC0
https://creativecommons.org/publicdomain/zero/1.0

Thanks to Maximilian Flaig and Kris Heid


Receiver
This module receives messages and passes them to the spartanmc core.
It is only synthesized if the router has at least one input.
It is connected to the first port of the splitter via datain,requestin and returnout
and to the spartanmc core via reset,clk_peri,do_peri,addr_peri,access_peri,wr_peri and di_peri.
It uses the modules YAF(yet another fifo) and reg_access_decoder

returnout (the same as the current state) can be:
00 (default output)
01 (if buffer/fifo/yaf has not enough free entries)
10 (ACK/Receiving)
11 (FINISH) (will never reach sender)

*/




`timescale 1ns / 1ps
`default_nettype none

module receiver
#(
    parameter BUFFER = 'd64, //buffer size
    parameter REG_COUNT = 3, //number of registers for communication with spmc core (always 3)
    parameter BASE_ADR = 10'h0, //base address
    parameter SIM_DEBUG = "NO", //debug?
    parameter HAS_SENDER = 0 //1 if there is a sender in this routermodule
) 
(
    //connected with spmc core
    input   wire            reset,
 	input   wire            clk_peri,
	input   wire [9:0]      addr_peri,
	input   wire            access_peri,
	input   wire            wr_peri,
    input   wire [17:0]     do_peri,
    output  wire [17:0]     di_peri,
    
    //connected with splitter
    input   wire [17:0]     datain,
    input   wire            requestin,
    output  wire [1:0]      returnout
);
`include "router/src/router_defines.vh"

localparam
    //state definitions
    S0 = 2'b00,
    S1 = 2'b01,
    S2 = 2'b10,
    S3 = 2'b11,
    
    //reg addresses
    DATA_AVAILABLE_REG_ADDR     = 2,
    FREE_ENTRIES_REG_ADDR       = 1,
    DATA_REG_ADDR               = 0,
    
    //amount of fifo address bits
    FIFO_ADDR_BITS              = log2(BUFFER - 1);
    
reg [1:0] state, nextstate; //for final-state-machine
wire [5:0] messagesize; //needed checking if buffer has enough free entries
assign messagesize = datain[11:6]; //only if datain==messageheader!

// chipselect reg_access_decoder
wire	[REG_COUNT-1:0]	reg_read;

wire		data_available_write_sel;   // CPU writes peri
wire		free_entries_write_sel;		// CPU writes peri
wire		data_write_sel;		  	    // CPU writes peri
wire		data_available_read_sel;	// CPU reads peri
wire		free_entries_read_sel;	 	// CPU reads peri
wire		data_read_sel;		    	// CPU reads peri
assign		data_available_read_sel     = reg_read[DATA_AVAILABLE_REG_ADDR];
assign		free_entries_read_sel 	    = reg_read[FREE_ENTRIES_REG_ADDR];
assign		data_read_sel 		        = reg_read[DATA_REG_ADDR];

reg_access_decoder	#(
		.REG_COUNT(	REG_COUNT		),
		.BASE_ADR(	BASE_ADR		),
		.SIM_DEBUG(	SIM_DEBUG	)
)
decod1 (
		.reset(			reset		),
		.clk_peri(		clk_peri	),
		.addr_peri(		addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(		wr_peri		),
		.peri_select(	),
		.reg_select(	),
		.reg_read(		reg_read	),
		.reg_write(		)
);




wire [FIFO_ADDR_BITS : 0]   fifo_fill;      //how many items are stored in fifo
wire [17:0]                 data_to_spmc;   //output from fifo, input to SpartanMC core

//signal to spmc core
assign	di_peri [17:0]	=   (data_available_read_sel) ? { (18'h00000 | fifo_fill  ) } : //data_available_read_sel ? di_peri=number of filled buffer entries
                            (free_entries_read_sel) ? (HAS_SENDER ? 18'bz : 18'b0) :    //free_entries_read_sel ? di_peri=0 (because only used in sender). If a sender exists in this routermodule di_peri=z (because sender is setting signal)
                            (data_read_sel) ? data_to_spmc : 18'b0;                     //data_read_sel ? di_peri=data_to_spmc ELSE = 0



//GENERATE BUFFER
generate
  //INPUT FIFO
	if(BUFFER > 0) begin
    
		xilinx_fifo_with_fill_counter #
			(
				.WIDTH(18),
				.DEPTH(BUFFER),
				.ADDR_BITS(FIFO_ADDR_BITS)
			) 
			fifo0(
				.clk(clk_peri),
				.fill_count(fifo_fill),
				.reset(reset),
				.write_enable((state==S2) & (requestin==1)),
				.write_data(datain),
				.read_enable(data_read_sel), 
				.read_data(data_to_spmc)
			);
	end
endgenerate

//final-state-machine
//see see $SPARTANMC_ROOT/src/doc/users-manual/src/router/receiver.svg
always @ (posedge clk_peri) begin
    if(reset)   begin
        state <= S0;
    end
    else begin
        state <= nextstate;
    end
end
    
always @ (*) begin
    case(state)
        S0: if(requestin==1) begin
                if(messagesize+1>(BUFFER-fifo_fill)) nextstate=S1; else nextstate=S2;
            end else nextstate=S0;
        S1: nextstate=S0;
        S2: if(requestin==0) nextstate=S3; else nextstate=S2;
        S3: nextstate=S0;
        default: nextstate = S0;
    endcase
end
    
//returnout is the state (00,01,10,11)
assign returnout = state;


endmodule
