/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/* 

---OOOO---------OOOO               .//
   O   O--------OOOOO------     .//
-x-OOOO    /----OOOO         .//
 | O  O    |              .//
 | O   O   |           .//
 |         |        .//
 |-OOOO    |     .//        
   OOOO----/  .//          //==\\
---OOOO    .//             ||
        .//           010  \\==\\
     .//           010001      ||
  .//           100010100  \\==//
//                         


Created spring 2016 by Felix Sterzelmaier
TU Darmstadt
---Fachgebiet Rechnersysteme---
License: CC0
https://creativecommons.org/publicdomain/zero/1.0

Thanks to Maximilian Flaig and Kris Heid


Selector
This module chooses one of the inputs or the sender to be connected
to the three internal ports data,request,return connected to the splitter.
It uses the module arbiter which implements round-robin.
Also the header of the message is stored to tell the decoder module to
which router the message should be passed.

If the Router has at least one output, it also has a sender module. So
requestin[0] is connected to the sender
requestin[1] is connected to the first input
requestin[2] is connected to the second input

If the Router has at no output, it has no sender module. So
requestin[0] is connected to the first input
requestin[1] is connected to the second input
requestin[2] is connected to the third input


The first 18bits of datain, the first bit of requestin and the 
first 2bits of returnout have to be connected to the same router(or the sender module)

The second 18bits of datain, the second bit of requestin and the 
second 2bits of returnout have to be connected to the same router

etc...

*/

`timescale 1ns / 1ps
`default_nettype none

module selector
#(
    parameter SELECTOR_IN = 'd3 //number of inputs
) 
(
    input   wire            reset,      //reset
    input   wire            clk_peri,   //clock
    
    //wires connected in direction of receiver of data /splitter)
    output  wire [17:0]     dataout,
    output  wire            requestout,
    input   wire [1:0]      returnin,
    
    output  wire [5:0]      routernr, //to which router the message should be passed? connected to decoder.
    
    //wires connected in direction of sender of data
    input   wire [(18*SELECTOR_IN)-1:0] datain,
    input   wire [SELECTOR_IN-1:0] requestin,
    output  wire [(2*SELECTOR_IN)-1:0] returnout
);


//state definitions
localparam
    S0 = 1'b0,  //sleepstate
    S1 = 1'b1;  //busystate

reg [5:0] storedto;         // "to" field from header (6bits from first 18bits of data) are stored here
reg state, nextstate;       //state registers for final-state-machine
wire roundrobinhasonehot;   //1 if arbiter has choosen one input. 0 otherwise
wire [SELECTOR_IN-1:0] roundrobinoutonehot; //choosen input from arbiter
wire resetroundrobin;       //arbiter should be resetted next clockcycle
wire currentrequest;        //current "request" from input , or 0 if no input is choosen by arbiter
wire [17:0] datatmp;        //current "data" from input , or 0 if no input is choosen by arbiter. Also needed to store destination (storedto)

assign routernr = storedto; //output "routernr" is the stored destination
assign roundrobinhasonehot = roundrobinoutonehot==0 ? 1'b0 : 1'b1; //1 if arbiter has choosen an input, 0 otherwise
assign resetroundrobin = (state==S1 && nextstate==S0); //arbiter should be resetted next clockcycle
assign dataout = (state==S0) ? 18'b0 : datatmp;  //dataout should not be send, if state=S0
assign requestout = (state==S0) ? 1'b0 : currentrequest; //requestout should not be send, if state=S0


generate
    genvar i;
    for(i = 0; i < SELECTOR_IN; i = i + 1 ) begin: selector_outfield
        assign datatmp      =   (roundrobinoutonehot==0) ? 18'b0 : //not active -> do nothing
                                roundrobinoutonehot[i] ? datain[((1+i)*18)-1:i*18] : //active and this input should be passed -> pass it
                                18'bz; //active and this input should NOT be passed -> DO NOT pass it
        
        assign currentrequest   =   (roundrobinoutonehot==0) ? 1'b0 : //no request -> request is 0
                                    roundrobinoutonehot[i] ? requestin[i] : //this is the selected request -> pass it
                                    1'bz; //this is NOT the selected request -> do NOT pass it
        
        assign returnout[((1+i)*2)-1:i*2] = (state==S0) ? 2'b00 : //not active -> do nothing
                                            roundrobinoutonehot[i] ? returnin[1:0] : //active and this input should be passed -> pass it 
                                            (roundrobinoutonehot!=0) ? 2'b01 : //active and this input should NOT be passed -> Send 01=blocked 
                                            2'b00; // do nothing
    end
endgenerate


generate
    if(SELECTOR_IN==1) begin //if only one input no arbiter/roundrobin is needed
        assign roundrobinoutonehot = requestin;
    end
    if(SELECTOR_IN>1) begin //roundrobin arbiter
        concentrator_arbiter # (
            .NUMBER_OF_SLAVES (SELECTOR_IN)
        )
        arbiter1 (
            .clk(			        clk_peri),
            .reset(		            reset | resetroundrobin),
            .last_value_write_sel(  1'b0),
            .req(	                requestin),
            .grant(		            roundrobinoutonehot)
        );
    end
endgenerate



//final-state-machine
//see see $SPARTANMC_ROOT/src/doc/users-manual/src/router/selector.svg
always @ (posedge clk_peri) begin
    if(reset)   begin 
        state <= S0;
        storedto <= 6'b0;
    end
    else begin
        state <= nextstate;
        if(state==S0 && nextstate==S1) begin
            storedto <= datatmp[5:0];
        end
        if(state==S1 && nextstate==S0) begin
            storedto <= 0;
        end
    end
end
    
always @ (*) begin
    case(state)
        S0: if(roundrobinhasonehot==1) nextstate=S1; else nextstate=S0;
        S1: if(currentrequest==0) nextstate=S0; else nextstate=S1;
        default: nextstate = S0;
    endcase
end
    


endmodule
