/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/* 

---OOOO---------OOOO               .//
   O   O--------OOOOO------     .//
-x-OOOO    /----OOOO         .//
 | O  O    |              .//
 | O   O   |           .//
 |         |        .//
 |-OOOO    |     .//        
   OOOO----/  .//          //==\\
---OOOO    .//             ||
        .//           010  \\==\\
     .//           010001      ||
  .//           100010100  \\==//
//                         


Created spring 2016 by Felix Sterzelmaier
TU Darmstadt
---Fachgebiet Rechnersysteme---
License: CC0
https://creativecommons.org/publicdomain/zero/1.0

Thanks to Maximilian Flaig and Kris Heid


Sender
This module receives messages from spartanmc core and sends them
to other routers.
It is only synthesized if the router has at least one output.
It is connected to the first port of the selector via dataout,requestout and returnin
and to the spartanmc core via reset,clk_peri,do_peri,addr_peri,access_peri,wr_peri and di_peri.
It uses the modules YAF(yet another fifo) and reg_access_decoder

returnin can be:
00 (default output)
01 (if buffer/fifo/yaf has not enough free entries, or path is blocked)
10 (ACK/Receiving)
11 (FINISH) (will never happen, because selector is closing too fast)

*/

`timescale 1ns / 1ps
`default_nettype none

module sender
#(
    parameter BUFFER = 'd64,        //buffer size
    parameter REG_COUNT = 3,        //number of registers for communication with spmc core (always 3)
    parameter BASE_ADR = 10'h0,     //base address
    parameter SIM_DEBUG = "NO",     //debug?
    parameter HAS_RECEIVER = 0,     //1 if there is an receiver in this routermodule
    parameter ROUTER_ID = 13        //routerid
) 
(
    //connected with spmc core
    input   wire            reset,
 	input   wire            clk_peri,
	input   wire [9:0]      addr_peri,
	input   wire            access_peri,
	input   wire            wr_peri,
    input   wire [17:0]     do_peri,
    output  wire [17:0]     di_peri,
    
    //connected with selector
    output  wire [17:0]     dataout,
    output  wire            requestout,
    input   wire [1:0]      returnin
);
`include "router/src/router_defines.vh"

localparam
    //state definitions
    S0 = 3'b000,
    S1 = 3'b001,
    S2 = 3'b010,
    S3 = 3'b011,
    S4 = 3'b100,
    
    //reg addresses
    DATA_AVAILABLE_REG_ADDR     = 2,
    FREE_ENTRIES_REG_ADDR       = 1,
    DATA_REG_ADDR               = 0,
    
    //amount of fifo address bits
    FIFO_ADDR_BITS              = log2(BUFFER - 1);
    

reg [17:0] storedheader;        //storedheader
reg [5:0] counter;              //used while waiting or transmitting
reg [2:0] state, nextstate;     //for final-state-machine

// chipselect reg_access_decoder
wire	[REG_COUNT-1:0]	reg_write;
wire	[REG_COUNT-1:0]	reg_read;
wire		data_available_write_sel;   // CPU writes peri
wire		free_entries_write_sel;		// CPU writes peri
wire		data_write_sel;		  	    // CPU writes peri
wire		free_entries_read_sel;	 	// CPU reads peri
wire		data_read_sel;		    	// CPU reads peri
assign		free_entries_read_sel 	    = reg_read[FREE_ENTRIES_REG_ADDR];
assign		data_write_sel 		        = reg_write[DATA_REG_ADDR];

reg_access_decoder	#(
		.REG_COUNT(	REG_COUNT		),
		.BASE_ADR(	BASE_ADR		),
		.SIM_DEBUG(	SIM_DEBUG	)
)
decod1 (
		.reset(			reset		),
		.clk_peri(		clk_peri	),
		.addr_peri(		addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(		wr_peri		),
		.peri_select(	),
		.reg_select(	),
		.reg_read(		reg_read	),
		.reg_write(		reg_write	)
);

//wires connected to fifo
wire                            readEna;        //1 if we read from fifo. firfo reduces counter.
wire [17:0]                     data_to_router; //output from fifo, input to this module
wire [FIFO_ADDR_BITS : 0]       fifoitems;      //how many items are stored in fifo

wire [5:0] messagesize;                         //how large is current message
assign messagesize = storedheader[11:6];

//signal to spmc core
assign di_peri [17:0]	=   (free_entries_read_sel) ? { ( 18'b0 | (BUFFER[17:0] - fifoitems) ) } :  //free_entries_read_sel ? di_peri=number of free buffer entries
                            (HAS_RECEIVER ? 18'bz : //receiver should set di_peri
                            18'b0); //we set di_peri to 0

/*
//not needed wires from yaf
wire wfull;
wire rfull;
wire rwfull;
wire wempty;
wire rempty;
wire rwempty;
*/

//for spmc_async_fifo_wrapper
wire write_full;
wire read_empty;
wire [FIFO_ADDR_BITS-1:0] read_addr;
wire [FIFO_ADDR_BITS-1:0] write_addr;

// yaf
//GENERATE BUFFER
generate
  //OUTPUT FIFO
	if(BUFFER > 0) begin

		xilinx_fifo_with_fill_counter #
			(
				.WIDTH(18),
				.DEPTH(BUFFER),
				.ADDR_BITS(FIFO_ADDR_BITS)
			) 
			fifo0(
				.clk(clk_peri),
				.fill_count(fifoitems),
				.reset(reset),
				.write_enable(data_write_sel),
				.write_data(do_peri),
				.read_enable(readEna), 
				.read_data(data_to_router)
			);
	end
endgenerate


//final-state-machine
//see see $SPARTANMC_ROOT/src/doc/users-manual/src/router/sender.svg
always @ (posedge clk_peri) begin
    if(reset)   begin
        state <= S0;
        counter <= 6'b0;
        storedheader <= 18'b0;
    end
    else begin
        state <= nextstate;
        if(state==S0 && nextstate==S1) storedheader <= data_to_router; //store header
        if(state==S2 && nextstate==S4) counter <= messagesize-1'b1;  //set counter to messagesize
        if(state==S2 && nextstate==S3) counter <= ROUTER_ID; //waitcounter starts counting (to 0) from ROUTER_ID. could also be a random number
        if(state==S3 && nextstate==S3) counter <= counter - 1'b1; //decrement waitcounter
        if(state==S4 && nextstate==S4) counter <= counter - 1'b1; //decrement transfercounter
        if(state==S4 && nextstate==S0) storedheader <= 18'b0; //reset stored header
    end
end
    
always @ (*) begin
    case(state)
        S0: if(fifoitems>0) nextstate=S1; else nextstate=S0;
        S1: if(fifoitems>=messagesize) nextstate=S2; else nextstate=S1;
        S2: if(returnin==2'b01) nextstate=S3; else begin if (returnin==2'b10) nextstate=S4; else nextstate=S2; end
        S3: if(counter==0) nextstate=S2; else nextstate=S3;
        S4: if(counter==0) nextstate=S0; else nextstate=S4;
        default: nextstate = S0;
    endcase
end
    
assign dataout = (state==S2) ? storedheader : (state==S4) ? data_to_router : 18'b0; //dataout = header or data
assign requestout = (state==S2 || state==S4) ? 1'b1 : 1'b0; //requestout =1 if state is s2 or s4
assign readEna = (state==S0 && nextstate==S1) | (state==S4 && nextstate==S4) | (state==S2 && nextstate==S4); //remove one fifoitem if: reading header, transmitting, transmitting first 


endmodule
