/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Seidowski
// Create Date:    12:52:14 10/21/2008 
// Module Name:    segment_display
// Project Name:   spartanmc
//
// Description:    Controller module for a segment based display (7, 14 segment or custom)
//                 with a user defined number of digits, configurable for use of LED or LC
//                 display.
//////////////////////////////////////////////////////////////////////////////////
`default_nettype none

module segment_display #(
		parameter BASE_ADR		= 10'h0,
		parameter SYSTEM_FREQUENCY	= 50_000_000,	// Systemtakt
		parameter SEGMENT_FREQUENCY	= 32,		// Wiederholfrequenz eines Segments (siehe Datenblatt)
		parameter NUMBER_OF_COMMONS	= 4,		// Anzahl an Backplanes/Commons
		parameter NUMBER_OF_SEGMENTS	= 4,		// Anzahl an Segmentleitungen pro Stelle
		parameter NUMBER_OF_DIGITS	= 8,		// Anzahl an Stellen
		parameter BIT_PER_DIGIT		= 16,		// Groesse einer Stelle im Speicher
		parameter IS_LCD		= 1,		// Controller nutzen fuer LED oder LC-Anzeige?
		parameter ALL_PORTS		= 3,		// Anzahl der Ports fuer den Dekoder
		parameter DECODSIM_DEB		= "NO"		// DEBUG mode YES oder NO
	)
	(
		input wire		clk_peri,
		input wire	[17:0]	do_peri,
		output wire	[17:0]	di_peri,
		input wire	[9:0]	addr_peri,
		input wire		access_peri,
		input wire		wr_peri,

		input wire		reset,

		/* Signale fuer das Display */
		output wire	[NUMBER_OF_COMMONS-1:0]			commons,
		output wire	[NUMBER_OF_SEGMENTS*NUMBER_OF_DIGITS-1:0] segments,
		output wire						vdd
	);
	
	// ld-Funktion (dualer Logarithmus, aufgerundet)
	function integer ld(input integer value);
		integer i;
		begin
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction
	
	localparam ENABLE_ADDR		= 3'h0;
	localparam TEXTMEM_ADDR		= 3'h1;
	localparam TEXTMEM_DATA		= 3'h2;

// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Status-, Daten- und Kommandoregister
	wire	[ALL_PORTS-1:0]	reg_select;	// r/w ADR
	wire	[ALL_PORTS-1:0]	reg_read;	// read ADR
	wire	[ALL_PORTS-1:0]	reg_write;	// write ADR
	wire select;		// Chip select signal
// Addressdekoder fuer alle Geraeteregister (verwendet pselect.v und
// verzoegert die read Signale um einen Takt fuer die neue Pipeline)
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// Dekodierung der Adressen ********************** END **************************************

	// Zugriffsmoeglichkeiten auf das Modul
	wire wr_enable	= reg_write[ENABLE_ADDR];
	wire rd_enable	= reg_read[ENABLE_ADDR];
	wire wr_memaddr	= reg_write[TEXTMEM_ADDR];
	wire rd_memaddr = reg_read[TEXTMEM_ADDR];
	wire wr_memdata = reg_write[TEXTMEM_DATA];
	wire rd_memdata = reg_read[TEXTMEM_DATA];

	// Enable-Register
	reg enable;

	// Textspeicher
	reg  [BIT_PER_DIGIT-1:0] mem [0:NUMBER_OF_DIGITS-1];
	reg  [ld(NUMBER_OF_DIGITS-1)-1:0] mem_addr_buf;
	wire [ld(NUMBER_OF_DIGITS-1)-1:0] mem_addr;
	wire [BIT_PER_DIGIT-1:0] mem_data = mem[mem_addr];


	// Daten auf den Lesebus legen
	assign di_peri	=	rd_enable ? enable :
				rd_memaddr ? mem_addr_buf :
				rd_memdata ? mem[mem_addr_buf] : 0;

	// Chip select erzeugen lassen
/*	pselect # (
			.ADDR_WIDTH(7),
			.BASE_WIDTH(7),
			.BASE_ADDR(BASE_ADDR >> 3)
		)
		CHIPSELECT (
			.addr(addr_peri[9:3]),
			.activ_peri(access_peri),
			.select(select)
                ); */


		
	// Signalerzeugung fuer das LC-Display
	segment_display_siggen # (
			.SYSTEM_FREQUENCY(SYSTEM_FREQUENCY),
			.SEGMENT_FREQUENCY(SEGMENT_FREQUENCY),
			.NUMBER_OF_COMMONS(NUMBER_OF_COMMONS),
			.NUMBER_OF_SEGMENTS(NUMBER_OF_SEGMENTS),
			.NUMBER_OF_DIGITS(NUMBER_OF_DIGITS),
			.BIT_PER_DIGIT(BIT_PER_DIGIT),
			.IS_LCD(IS_LCD)
		)
		SIGNALS (
			.clk(clk_peri),
			.reset(reset),
			.mem_addr(mem_addr),
			.mem_data(mem_data),
			.enable(enable),
			.com_out(commons),
			.seg_out(segments),
			.vdd(vdd)
		);

	// Schreibzugriff auf die Register
	/* Der Zugriff auf den Textspeicher erfolg zweiphasig: Im Ersten Schritt wird die Textadresse auf den
		Datenbus zusammen mit dem Offset TEXTMEM_ADDR auf den Adressbus gelegt. Diese Adresse wird dabei
		gepuffert. Im naechsten Schritt kann von dieser Adresse mittels des Offsets TEXTMEM_READ gelesen
		bzw. mittels TEXTMEM_WRITE geschrieben werden. */
	always @(posedge clk_peri)
	begin
		if (reset)
		begin : resetMemory
			reg[NUMBER_OF_DIGITS-1:0] i;
			for (i = 0; i < NUMBER_OF_DIGITS; i = i + 1)
				mem[i] <= 0;

/*			mem[0] <= 16'h0aaaa; // *
			mem[1] <= 16'h0aaaa; // *
			mem[2] <= 16'h08006; // T
			mem[3] <= 16'h01394; // E
			mem[4] <= 16'h0160c; // S
			mem[5] <= 16'h08006; // T
			mem[6] <= 16'h0aaaa; // *
			mem[7] <= 16'h0aaaa; // *
*/
			mem_addr_buf <= 0;
			enable <= 0;
		end
		else begin
			if (wr_enable)
				// Display an-/ausschalten
				enable <= do_peri[0];
			else if (wr_memaddr)
				// Textadresse puffern
				mem_addr_buf <= do_peri; //do_peri[0 +: ld(NUMBER_OF_DIGITS-1)];
			else if (wr_memdata)
				// Zeichen in Textspeicher an gepufferte Position schreiben
				mem[mem_addr_buf] <= do_peri; //do_peri[0 +: BIT_PER_DIGIT];
		end
	end
endmodule
