/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Engineer:       Andre Seidowski
// 
// Create Date:    11:19:53 08/18/2008 
// Module Name:    segment_display_siggen 
// Project Name:   Display Controller fuer SpartanMC
//
// Description:    Signalerzeugung fuer das segmentbasiertes LC-Display
//                 DE 301
//
// Dependencies:   used by "segment_display.v" module
//
//////////////////////////////////////////////////////////////////////////////////

module segment_display_siggen #(
		// Systemfrequenz in Hz
		parameter SYSTEM_FREQUENCY = 50_000_000,
		// Wiederholfrequenz eines einzelnen Segments in Hz
		parameter SEGMENT_FREQUENCY = 32,
		// Anzahl an commons _eines_ Displays
		parameter NUMBER_OF_COMMONS = 4,
		// Anzahl an Segment-Pins pro Ziffer
		parameter NUMBER_OF_SEGMENTS = 4,
		// Anzahl an Ziffern pro Display
		parameter NUMBER_OF_DIGITS = 8,
		// Bit pro Ziffer im Speicher
		parameter BIT_PER_DIGIT = 16,
		// Wird der Controller fuer ein LED- (0) oder LC-Display (1) benutzt
		parameter IS_LCD = 1
	)
	(
		input wire		clk,
		input wire		reset,				// High-aktiv!!
		input wire [BIT_PER_DIGIT-1:0] mem_data,		// Datenleitungen des Textspeichers
		output reg [ld(NUMBER_OF_DIGITS)-1:0] mem_addr, // Adressleitungen des Textspeichers
		input wire		enable	,			// Display an/ausschalten
		output reg [NUMBER_OF_COMMONS-1:0] com_out,	// "Anoden-Pins", beim LCD also die commons
		output reg [NUMBER_OF_SEGMENTS*NUMBER_OF_DIGITS-1:0] seg_out, // Daten-Pins des LCDs
		output reg	vdd				// Vdd-Leitung fuer Spannungsteiler am Display (um es an-/abzuschalten)
	);
	
	// Dualer Logarithmus aufgerundet
	function integer ld(input integer value);
		integer i;
		begin
			value = value > 0 ? value - 1 : 0;
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction

	
	localparam FRAME_CLOCK_COUNT = SYSTEM_FREQUENCY / (2* NUMBER_OF_COMMONS * SEGMENT_FREQUENCY);

	reg [NUMBER_OF_SEGMENTS*NUMBER_OF_DIGITS-1:0] seg_out_buf;	// Ausgangspuffer
	reg [ld(NUMBER_OF_COMMONS)-1:0] com_count;	// Zaehler fuer die aktuelle Common-Leitung
	
	reg [ld(FRAME_CLOCK_COUNT)-1:0] clockCount;	// Zaehler zum Erzeugen von frameClock
	reg 	frameClock;				// Takt, mit dem die Displayansteuerung geschaltet wird
	reg	currentLogic;				// Register fuer bedingte Signalinvertierung der Ausgabe


	// Werte zum Lesen aus der Codepage
	reg [ld(NUMBER_OF_DIGITS)-1:0] memReadCount;
	reg [2:0]	memReadState;

	// Generierung des langsamen Taktes, der die Commons durchschalten
	always @(posedge clk)
	begin : frameClockGenerate
		if (reset)
		begin
			clockCount <= 0;
			frameClock <= 1;
		end
		else begin
			if (FRAME_CLOCK_COUNT-1 == clockCount)
			begin
				frameClock <= !frameClock;
				clockCount <= 0;
			end
			else
				clockCount <= clockCount + 1;
		end
	end
	
	// Vdd je nach Belegung von enable beschalten. Im LCD-Betrieb muss jedoch vor allem beim Abschalten
	// abgewartet werden, bis ein kompletter Zyklus der Anzeige durchlaufen wurde, um keine Ladung auf dem
	// Display zu hinterlassen (Stichwort Gleichspannungsanteil)
	always @(posedge clk)
	begin
		if (reset)
			vdd <= 0;
		else
			if (IS_LCD)
			begin
				if ((com_count == NUMBER_OF_COMMONS-1) && !currentLogic)
					vdd <= enable;
			end
			else
				vdd <= enable;
	end
	
	// Erzeugung der Signale fuer das Display aus den Ausgangspuffern
	always @(posedge frameClock)
	begin : outputSignals
		if (reset)
		begin
			currentLogic 	<= 0;
			seg_out		<= 0;
			com_out		<= 0;
			com_count	<= 0;
		end
		else begin : dataWrite
			if (enable)
			begin
				if (IS_LCD)
				begin
					// Ausgabe der gepufferten Daten
					seg_out 	<= seg_out_buf ^ { NUMBER_OF_SEGMENTS*NUMBER_OF_DIGITS {currentLogic}};	// invertieren oder nicht (currentLogic auf 32 bit verbreitert)
					
					// Commons erst einmal alles "z-ten"
					com_out		<= { NUMBER_OF_COMMONS {1'bz} };
					// danach das entsprechende Bit auf 1 bzw. 0 setzen
					com_out[com_count] <= currentLogic;
				end
				else begin
					seg_out		<= seg_out_buf;
					com_out		<= 0;
					com_out[com_count] <= 1;
				end
			end else
			begin
				seg_out <= 0;
				com_out <= 0;
			end
			
			if (com_count == NUMBER_OF_COMMONS-1)
			begin
				com_count <= 0;
				currentLogic <= IS_LCD ? !currentLogic : 0;
			end
			else
				com_count <= com_count + 1;
		end
	end
	
	
	// Auslesen der Daten aus dem Speicher
	always @(posedge clk)
	begin : memoryRead
		if (reset)
		begin
			memReadState	<= 0;
			memReadCount	<= 0;
			mem_addr	<= 0;
			seg_out_buf	<= 0;
		end
		else begin
			case (memReadState)
					// Wenn frameClock das naechste Mal wieder auf High geht,
					// koennen die Daten fuer die Pinbelegung des naechsten
					// Commons geladen werden. Sonst im selben Zustand bleiben
				0:	memReadState <= frameClock ? 1 : 0;

					// Anlegen der Adresse an Speicher
				1:	begin
						mem_addr <= memReadCount;
						memReadState <= 2;
					end

					// Wartetakt fuer Daten aus dem Speicher
				2:	memReadState <= 3;

					// Umcodieren der zurueckgegebenen Daten
				3:	begin : fillDataBuf
						//reg [ld(NUMBER_OF_SEGMENTS)-1:0] i;
						integer i;
						for (i = 0; i < NUMBER_OF_SEGMENTS; i = i + 1)
							seg_out_buf[i+NUMBER_OF_SEGMENTS*memReadCount] <= mem_data[i+NUMBER_OF_COMMONS*com_count];

						// Wenn die Daten fuer alle Ziffern des Displays geladen wurden,
						// zum naechsten Zustand gehen
						if (NUMBER_OF_DIGITS-1 == memReadCount)
						begin
							memReadCount <= 0;
							memReadState <= 4;
						end
						else begin
							// ...sonst die naechste Ziffer aus dem ROM holen
							memReadCount <= memReadCount + 1;
							memReadState <= 1;
						end
					end

					// Warten, bis frameClock wieder auf 0 geht
				4:	memReadState <= frameClock ? 4 : 0;
			endcase
		end
	end
endmodule
