/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`ifndef __shi_fifo
`define __shi_fifo
`default_nettype none
module shi_fifo #(

    parameter DATA_WIDTH        = 18,
    parameter ADDRESS_WIDTH     = 8,
    parameter BLOCK_SIZE        = 0,
    parameter TIMEOUT_CYCLES    = 1,
    parameter TIMEOUT_PRESCALE  = 1,
    parameter OFF_CLK_CNT       = 0,
    parameter OFF_WORD_CNT      = 1

)(

    /* system clock and reset */
    input wire                          clk_tx_i,
    input wire                          clk_rx_i,
    input wire                          reset_i,

    /* generic interface */
    output wire     [DATA_WIDTH-1:0]    tx_data_o,
    output wire                         tx_valid_o,
    input wire                          tx_en_i,
    input wire      [DATA_WIDTH-1:0]    rx_data_i,
    input wire                          rx_en_i,
    output wire                         rx_rdy_o

);

localparam
    DIVIDER_VALUES = TIMEOUT_PRESCALE - 1;

wire                    yaf0_rempty;
wire                    yaf0_wfull;
wire                    yaf0_reset_n;
wire [ADDRESS_WIDTH:0]  yaf0_fill;
wire                    yaf0_rinc;
wire                    yaf0_winc;

//assign rx_rdy_o = ~yaf0_wfull;
assign yaf0_reset_n = ~reset_i;
assign yaf0_rinc = tx_en_i & tx_valid_o;

yaf yaf0(
    .rdata(tx_data_o),
    .rempty(yaf0_rempty),
    .rrst_n(yaf0_reset_n),
    .rinc(yaf0_rinc),
    .rclk(clk_tx_i),
    .wdata(rx_data_i),
    .wfull(yaf0_wfull),
    .wrst_n(yaf0_reset_n),
//    .winc(rx_en_i),
    .winc(yaf0_winc),
    .wclk(clk_rx_i),
    .fill(yaf0_fill)
);

defparam
    yaf0.DSIZE = DATA_WIDTH,
    yaf0.ASIZE = ADDRESS_WIDTH;

fifo_throttle
#(
  .OFF_CLK_CNT(OFF_CLK_CNT),
  .OFF_WORD_CNT(OFF_WORD_CNT)
)
fifo_throttle0(
  .clk_i(clk_rx_i),
  .rst_i(reset_i),
  .rx_en_i(rx_en_i),
  .rx_rdy_o(rx_rdy_o),
  .fifo_full_i(yaf0_wfull),
  .fifo_we_o(yaf0_winc)
);

genvar i;
generate 
if (BLOCK_SIZE > 1)
begin : BLOCK_LOGIC

    reg  [4:0]              flush_counter;
    reg                     block_valid;

    wire [5:0]              divide_chain;

    assign divide_chain[0] = 1'b1;

    for (i=0; i<5; i=i+1) 
    begin : DIVIDER_CHAIN
        divide_counter divide_counter0(
            .reset_i(reset_i),
            .clk_i(clk_rx_i),
            .en_i(divide_chain[i]),
            .pulse_o(divide_chain[i+1])
        );

        defparam
            divide_counter0.DIVIDE_BY = (DIVIDER_VALUES[i*2+1:i*2] + 1);
    end

    always @ (posedge clk_rx_i)
    begin
        if (reset_i || rx_en_i || tx_valid_o)
        begin
            flush_counter <= 5'b0;
        end else if ((flush_counter != (TIMEOUT_CYCLES - 1)) && divide_chain[5])
        begin
            flush_counter <= flush_counter + 1;
        end
    end

    always @ (posedge clk_rx_i)
    begin
        if (reset_i || yaf0_rempty)
        begin
            block_valid <= 0;
        end else if (((flush_counter == (TIMEOUT_CYCLES - 1)) && divide_chain[5]) || (yaf0_fill == BLOCK_SIZE))
        begin
            block_valid <= 1;
        end
    end

    assign tx_valid_o = ~yaf0_rempty && block_valid;

end else begin

    assign tx_valid_o = ~yaf0_rempty;

end

endgenerate

endmodule
`endif
