/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
`default_nettype none

module GmiiControllerDp
#(
  parameter   p_mac_addr_d    = 48'h00_0a_35_01_fd_dc,
  parameter   p_mac_addr_s    = 48'h00_0a_35_01_fd_dc,
  parameter   p_tpid          = 16'h81_00,
  parameter   p_tci           = 16'h55_55,
  parameter   p_type          = 16'h08_00,   
  parameter   p_tx_word_width = 'd8,
  parameter   p_dp_size       = 'd1498 //Die Größe eines Datenpakets. Min:44; Max:1498, da die letzten zwei bytes genutzt werden, als Length-Field. 
)(
  input  wire Rst_i,
  input  wire Clk_i,

  //interface to FIFO
  input  wire [p_tx_word_width-1:0] Rx_Data_i,
  input  wire Rx_En_i,
  output wire Rx_Rdy_o,

  //interface to PHY 
  output wire [p_tx_word_width-1:0] Tx_Data_o,
  output wire Tx_Enable_o,
  output wire Tx_Error_o,
  output wire Tx_Rst_o,//ähnlich dem Locked signal 
  
  //interface between Ctrl and Dp
  input  wire Send_Head_i,
  input  wire Send_Data_i,
  input  wire Send_Length_i,
  input  wire Send_Crc_i,
  input  wire Wait_i,//96 Bits Gap zwischen zwei Paketen notwendig!
  input  wire Clr_i,  
  output wire Data_Ready_o, 
  output wire Head_Done_o,
  output wire Data_Done_o,
  output wire Length_Done_o,
  output wire Crc_Done_o,
  output wire Wait_Done_o
);

  `include "common/functions/bitwidth.vh"
  localparam lp_head_size      = 'd26;
  localparam lp_crc_size       = 'd4;
  localparam lp_wait_num       = 'd12; //96 bits
  localparam lp_dp_size_min    = 'd44;
  localparam lp_data_cnt_size  = bitwidth(p_dp_size);
  localparam lp_ndata_cnt_size = bitwidth(lp_head_size);
  
  reg [lp_data_cnt_size-1:0]  r_data_counter;
  reg [lp_ndata_cnt_size-1:0]  r_non_data_counter;
  reg r_data_ready, r_head_done, r_data_done, r_length_done, r_crc_done, r_wait_done;
  reg [p_tx_word_width-1:0] r_data;
  reg [10:0] r_length;
  reg [4:0]  r_number;  
  reg r_length_low8;
  reg [7:0] r_lock_counter;
  reg [3:0] r_crc_enable;
  
  wire w_sending, w_in_use;
  wire [p_tx_word_width-1:0] w_head_data;
  wire [p_tx_word_width-1:0] w_len_and_num;
  wire w_crc_enable;
  wire [7:0] w_crc_byte;
  wire w_rst_stop;
  
  assign w_sending        = (Send_Head_i || Send_Data_i || Send_Length_i || Send_Crc_i);
  assign w_in_use         = (w_sending   || Wait_i);
  assign Data_Ready_o     = r_data_ready;
  assign Head_Done_o      = r_head_done; 
  assign Data_Done_o      = r_data_done;
  assign Length_Done_o    = r_length_done;
  assign Crc_Done_o       = r_crc_done;
  assign Wait_Done_o      = r_wait_done; 
  assign Rx_Rdy_o         = (Send_Head_i && (r_non_data_counter == (lp_head_size - 1'b1))) || (Send_Data_i && (r_data_counter < p_dp_size) && Rx_En_i); 
  assign Tx_Data_o        = Send_Head_i   ? w_head_data    :
			   (Send_Data_i   ? r_data         :
			   (Send_Length_i ? w_len_and_num  : 
			   (Send_Crc_i    ? w_crc_byte     : {(p_tx_word_width-1){1'b0}})));
  assign Tx_Enable_o      = w_sending;
  assign Tx_Rst_o         = !r_lock_counter[6];
  assign w_rst_stop       = r_lock_counter[7];
  assign w_len_and_num    = r_length_low8 ? r_length[7:0] : {r_number, r_length[10:8]};
  assign Tx_Error_o       = 1'b0;
  assign w_crc_enable     = (!r_crc_enable[1] & !r_crc_enable[2] & r_crc_enable[3]);
  
  always@(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_lock_counter <= 8'b0;
    end
    else if(!w_rst_stop) begin
      r_lock_counter <= r_lock_counter + 1'b1;
    end	 
  end
  
  always@(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_data_ready <= 1'b0;
    end
    else if(!w_in_use   && Rx_En_i && Tx_Rst_o) begin
      r_data_ready <= 1'b1;
    end
    else begin
      r_data_ready <= 1'b0;
    end
  end
  
  always@(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_data <= {(p_tx_word_width){1'b0}};
    end
    else begin
      r_data <= Rx_Data_i;
    end
  end
  
  always @(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_crc_enable <= 4'b0;
    end
    else if((Send_Head_i && !w_crc_enable) || Send_Length_i) begin
      r_crc_enable <= r_crc_enable + 1'b1;
    end
    else if(Send_Crc_i) begin
      r_crc_enable <= 4'b0;
    end
  end
  
  always @(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_data_counter     <= {{(lp_data_cnt_size-1){1'b0}}, 1'b1};
      r_non_data_counter <= {(lp_ndata_cnt_size){1'b0}};
      r_length           <= {10'b0, 1'b1};
//      r_length           <=11'b0;
      r_number           <= 5'b0;
      r_length_low8      <= 1'b0;
      r_head_done        <= 1'b0; 
      r_data_done        <= 1'b0;
      r_length_done	 <= 1'b0;
      r_crc_done         <= 1'b0; 
      r_wait_done        <= 1'b0;
    end
    else if(Send_Head_i) begin
      if(r_non_data_counter == (lp_head_size - 1'b1)) begin
        r_head_done    <= 1'b1;
        r_non_data_counter <= {(lp_ndata_cnt_size){1'b0}};
      end
      else begin
        r_non_data_counter <= r_non_data_counter + 1'b1;
      end
    end
    else if(Send_Data_i) begin
      if(Rx_En_i) begin
        if(r_data_counter == (p_dp_size)) begin
	  r_data_done    <= 1'b1;
        end 
        else begin
          r_data_counter <= r_data_counter + 1'b1;
	  r_length       <= r_length + 1'b1;
        end
      end
      else begin
        if(r_data_counter >= (lp_dp_size_min - 1)) begin
          r_data_done    <= 1'b1;
        end
        else begin
          r_data_counter <= r_data_counter + 1'b1;
        end
      end

/*
      if(Rx_En_i) begin
        if(r_data_counter == (p_dp_size)) begin
          r_length       <= p_dp_size;
          r_data_done    <= 1'b1;
        end 
        else begin
          r_data_counter <= r_data_counter + 1'b1;
        end
      end
      else begin
        if(r_length == 11'b0) begin
          r_length <= {{(11-lp_data_cnt_size){1'b0}}, r_data_counter};
        end 
         
        if(r_data_counter >= (lp_dp_size_min - 1)) begin
          r_data_done    <= 1'b1;
        end
        else begin
	  r_data_counter <= r_data_counter + 1'b1;
        end
      end
*/
    end
    else if(Send_Length_i) begin
      if(r_length_low8) begin
        r_length_done <= 1'b1;
      end
      else begin
        r_length_low8 <= 1'b1;
      end
    end
    else if(Send_Crc_i) begin
      if(r_non_data_counter == (lp_crc_size - 1'b1)) begin
        r_crc_done         <= 1'b1;
        r_non_data_counter <= {(lp_ndata_cnt_size){1'b0}};
      end
      else begin
        r_non_data_counter <= r_non_data_counter + 1'b1;
      end
    end
    else if(Wait_i) begin
      if(r_non_data_counter == (lp_wait_num - 1'b1)) begin
        r_wait_done        <= 1'b1;
      end
      else begin
        r_non_data_counter <= r_non_data_counter + 1'b1;
      end
    end
    else if(Clr_i) begin
        r_data_counter     <= {{(lp_data_cnt_size-1){1'b0}}, 1'b1};
        r_non_data_counter <= {(lp_ndata_cnt_size){1'b0}};
	r_length           <= {10'b0, 1'b1};
//	r_length           <= 11'b0;
	r_number           <= r_number + 1'b1;
	r_length_low8      <= 1'b0;
	r_head_done        <= 1'b0; 
	r_data_done        <= 1'b0; 
	r_length_done	   <= 1'b0;
	r_crc_done         <= 1'b0; 
	r_wait_done        <= 1'b0;
    end
  end
  
  FrameHeadGenerator#(
    .p_mac_addr_d(p_mac_addr_d),
    .p_mac_addr_s(p_mac_addr_s),
    .p_tpid(p_tpid),
    .p_tci(p_tci),
    .p_type(p_type),   
    .p_tx_word_width(p_tx_word_width)
  )
  fhg0(
    .Rst_i(Rst_i),
    .Clk_i(Clk_i),
    .En_i(Send_Head_i),
    .Clr_i(Clr_i),
    .Head_Data_o(w_head_data)
  );
  
  Crc32Generator
  crc32
  (
    .Rst_i(Rst_i),
    .Clk_i(Clk_i),
    .En_i(w_crc_enable),
    .Clr_i(Clr_i),
    .Input_Byte_i(Tx_Data_o),
    .Crc_Byte_o(w_crc_byte)
  );
  
endmodule

/**
 * Der folgende Code kann die Timing-Constraints nicht erfüllen, weil Rx_Data_i nicht zwischengepuffert wird. 
 * Im obigen Code wird der Register r_data genutzt, wo Rx_Data_i zuerst gespeichert wird. Dies löst aber ein 
 * Taktverzögerung aus. Deswegen muss Rx_Rdy_o einen Takt vorgezogen werden.  
 */
  
/*
module GmiiControllerDp
#(
  parameter   p_mac_addr_d    = 48'h00_0a_35_01_fd_dc,
  parameter   p_mac_addr_s    = 48'h00_0a_35_01_fd_dc,
  parameter   p_tpid          = 16'h81_00,
  parameter   p_tci           = 16'h45_55,
  parameter   p_type          = 16'h08_00,   
  parameter   p_tx_word_width = 'd8,
  parameter   p_dp_size       = 'd1498 
)(
  input  wire Rst_i,
  input  wire Clk_i,

  //interface to FIFO
  input  wire [p_tx_word_width-1:0] Rx_Data_i,
  input  wire Rx_En_i,
  output wire Rx_Rdy_o,
  
  //interface to PHY 
  output wire [p_tx_word_width-1:0] Tx_Data_o,
  output wire Tx_Enable_o,
  output wire Tx_Error_o,
  output wire Tx_Rst_o,
 
  //interface between Ctrl and Dp
  input  wire Send_Head_i,
  input  wire Send_Data_i,
  input  wire Send_Length_i,
  input  wire Send_Crc_i,
  input  wire Wait_i,
  input  wire Clr_i,  
  output wire Data_Ready_o, 
  output wire Head_Done_o,
  output wire Data_Done_o,
  output wire Length_Done_o,
  output wire Crc_Done_o,
  output wire Wait_Done_o
);
  `include "common/functions/bitwidth.vh"
  localparam lp_head_size      = 'd26;
  localparam lp_crc_size       = 'd4;
  localparam lp_wait_num       = 'd12; //96 bits
  localparam lp_dp_size_min    = 'd44;//'d60;
  localparam lp_data_cnt_size  = bitwidth(p_dp_size);
  localparam lp_ndata_cnt_size = bitwidth(lp_head_size);
  
  reg [lp_data_cnt_size-1:0]  r_data_counter;
  reg [lp_ndata_cnt_size-1:0]  r_non_data_counter;
  reg r_data_ready, r_head_done, r_data_done, r_length_done, r_crc_done, r_wait_done;
//  reg [p_tx_word_width-1:0] r_data;
//  reg [15:0] r_length;//zwei Bytes
  reg [10:0] r_length;
  reg [4:0]  r_number;
  reg r_length_low8;
  reg [7:0] r_lock_counter;
  reg [3:0] r_crc_enable;
  
  wire w_sending, w_in_use;
  wire [p_tx_word_width-1:0] w_head_data;
//  wire [p_tx_word_width-1:0] w_data;
  wire [p_tx_word_width-1:0] w_len_and_num;
  wire w_crc_enable;
  wire [7:0] w_crc_byte;
  wire w_rst_stop;
  
  assign w_sending        = (Send_Head_i || Send_Data_i || Send_Length_i || Send_Crc_i);
  assign w_in_use         = (w_sending   || Wait_i);
  assign Data_Ready_o     = r_data_ready;
  assign Head_Done_o      = r_head_done; 
  assign Data_Done_o      = r_data_done;
  assign Length_Done_o    = r_length_done;
  assign Crc_Done_o       = r_crc_done;
  assign Wait_Done_o      = r_wait_done; 
  assign Rx_Rdy_o         =(Send_Data_i && (r_data_counter < p_dp_size) && Rx_En_i); 
  assign Tx_Data_o        = Send_Head_i   ? w_head_data    :
                           (Send_Data_i   ? Rx_Data_i      :
//                           (Send_Data_i   ? r_data         : 
//			   (Send_Length_i ? r_length[15:8] :
                           (Send_Length_i ? w_len_and_num  : 
			   (Send_Crc_i    ? w_crc_byte     : {(p_tx_word_width-1){1'b0}})));
  assign Tx_Enable_o      = w_sending;
  assign Tx_Rst_o         = !r_lock_counter[6];
  assign w_rst_stop       = r_lock_counter[7];
//  assign w_data           = Rx_En_i ? Rx_Data_i : 8'haa;
  assign w_len_and_num    = r_length_low8 ? r_length[7:0] : {r_number, r_length[10:8]};
  assign Tx_Error_o       = 1'b0;
  assign w_crc_enable     = (!r_crc_enable[1] & !r_crc_enable[2] & r_crc_enable[3]);
  
  always@(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_lock_counter <= 8'b0;
    end
    else if(!w_rst_stop) begin
      r_lock_counter <= r_lock_counter + 1'b1;
    end	 
  end
  
  always@(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_data_ready <= 1'b0;
    end
    else if(!w_in_use   && Rx_En_i && Tx_Rst_o) begin
      r_data_ready <= 1'b1;
    end
    else begin
      r_data_ready <= 1'b0;
    end
  end

//  always@(posedge Clk_i, posedge Rst_i) begin
//    if(Rst_i) begin
//      r_data <= {(p_tx_word_width){1'b0}};
//    end
//    else begin
//      r_data <= Rx_Data_i;
//    end
//  end
  
  always @(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_crc_enable <= 4'b0;
    end
    else if((Send_Head_i && !w_crc_enable) || Send_Length_i) begin
      r_crc_enable <= r_crc_enable + 1'b1;
    end
    else if(Send_Crc_i) begin
      r_crc_enable <= 4'b0;
    end
  end
  
  always @(posedge Clk_i, posedge Rst_i) begin
    if(Rst_i) begin
      r_data_counter     <= {(lp_data_cnt_size){1'b0}};
      r_non_data_counter <= {(lp_ndata_cnt_size){1'b0}};
      //r_length           <= 16'b0;
      r_length           <= 11'b0;
      r_number           <= 5'b0;
      r_length_low8      <= 1'b0;
      r_head_done        <= 1'b0; 
      r_data_done        <= 1'b0;
      r_length_done	 <= 1'b0;
      r_crc_done         <= 1'b0; 
      r_wait_done        <= 1'b0;
    end
    else if(Send_Head_i) begin
      if(r_non_data_counter == (lp_head_size - 1'b1)) begin
        r_head_done    <= 1'b1;
        r_non_data_counter <= {(lp_ndata_cnt_size){1'b0}};
      end
      else begin
        r_non_data_counter <= r_non_data_counter + 1'b1;
      end
    end
    else if(Send_Data_i) begin
      if(Rx_En_i) begin
        if(r_data_counter == (p_dp_size - 1)) begin
          r_length       <= p_dp_size;
          r_data_done    <= 1'b1;
        end 
        else begin
          r_data_counter <= r_data_counter + 1'b1;
        end
      end
      else begin
 //       if(r_length == 16'b0) begin
        if(r_length == 11'b0) begin
          r_length <= {{(11-lp_data_cnt_size){1'b0}}, r_data_counter};
        end
		          
        if(r_data_counter >= (lp_dp_size_min - 1)) begin
          r_data_done    <= 1'b1;
        end
        else begin
          r_data_counter <= r_data_counter + 1'b1;
        end
      end
    end
    else if(Send_Length_i) begin
      if(r_length_low8) begin
        r_length_done <= 1'b1;
      end
      else begin
//        r_length      <= {r_length[7:0], 8'b0};
        r_length_low8 <= 1'b1;
      end
    end
    else if(Send_Crc_i) begin
      if(r_non_data_counter == (lp_crc_size - 1'b1)) begin
        r_crc_done         <= 1'b1;
        r_non_data_counter <= {(lp_ndata_cnt_size){1'b0}};
      end
      else begin
        r_non_data_counter <= r_non_data_counter + 1'b1;
      end
    end
    else if(Wait_i) begin
      if(r_non_data_counter == (lp_wait_num - 1)) begin
        r_wait_done        <= 1'b1;
      end
      else begin
        r_non_data_counter <= r_non_data_counter + 1'b1;
      end
    end
    else if(Clr_i) begin
      r_data_counter     <= {(lp_data_cnt_size){1'b0}};
      r_non_data_counter <= {(lp_ndata_cnt_size){1'b0}};
//      r_length           <= 16'b0;
      r_length           <= 11'b0;
      r_number           <= r_number + 1'b1;
      r_length_low8      <= 1'b0;
      r_head_done        <= 1'b0; 
      r_data_done        <= 1'b0; 
      r_length_done      <= 1'b0;
      r_crc_done         <= 1'b0; 
      r_wait_done        <= 1'b0;
    end
  end
  
  FrameHeadGenerator#(
    .p_mac_addr_d(p_mac_addr_d),
    .p_mac_addr_s(p_mac_addr_s),
    .p_tpid(p_tpid),
    .p_tci(p_tci),
    .p_type(p_type),   
    .p_tx_word_width(p_tx_word_width)
  )
  fhg0(
    .Rst_i(Rst_i),
    .Clk_i(Clk_i),
    .En_i(Send_Head_i),
    .Clr_i(Clr_i),
    .Head_Data_o(w_head_data)
  );
  
  Crc32Generator
  crc32
  (
    .Rst_i(Rst_i),
    .Clk_i(Clk_i),
    .En_i(w_crc_enable),
    .Clr_i(Clr_i),
    .Input_Byte_i(Tx_Data_o),
    .Crc_Byte_o(w_crc_byte)
  );
  
endmodule*/
