/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none

module shi_arbiter #(
    parameter DATA_WIDTH            = 8,
    parameter INPUT_PORTS           = 2,
    parameter CHANNEL_SELECT_WORD   = 8'h80,
    parameter MAX_WORDS_PER_CYCLE   = 16
)(
    input wire                                  clk_i,
    input wire                                  reset_i,

    output reg [DATA_WIDTH-1:0]                 tx_data_o,
    output wire                                 tx_valid_o,
    input wire                                  tx_en_i,
    input wire [(INPUT_PORTS * DATA_WIDTH)-1:0] rx_data_i,
    input wire [INPUT_PORTS-1:0]                rx_en_i,
    output wire [INPUT_PORTS-1:0]               rx_rdy_o
);

`include "common/functions/bitwidth.vh"

localparam
    WC_WIDTH        = bitwidth(MAX_WORDS_PER_CYCLE-1),
    SEL_WIDTH       = bitwidth(INPUT_PORTS-1),

    STATE_QUERY     = 0,
    STATE_SELECT    = 1,
    STATE_TX        = 2;

reg [31:0]              state;
reg [31:0]              next_state;

reg [WC_WIDTH-1:0]      word_counter;
reg                     wc_reset;
reg                     wc_inc;

reg [SEL_WIDTH-1:0]     ch_sel;
wire                    ch_sel_reset;
reg                     ch_switch;
wire                    ch_force_switch;

wire [DATA_WIDTH-1:0]   shi_mux_0_tx_data_o;
reg                     shi_mux_0_tx_en_i;

shi_mux shi_mux_0(
    .tx_data_o(shi_mux_0_tx_data_o),
    .tx_valid_o(tx_valid_o),
    .tx_en_i(shi_mux_0_tx_en_i),
    .rx_data_i(rx_data_i),
    .rx_en_i(rx_en_i),
    .rx_rdy_o(rx_rdy_o),
    .sel_i(ch_sel)
);

defparam
    shi_mux_0.DATA_WIDTH    = DATA_WIDTH,
    shi_mux_0.INPUT_PORTS   = INPUT_PORTS;    

assign ch_sel_reset = ((ch_sel == (INPUT_PORTS-1)) && ch_switch) || reset_i;
assign ch_force_switch = (word_counter == (MAX_WORDS_PER_CYCLE-1)) && tx_en_i;

always @ (posedge clk_i) begin
    if (wc_reset)
        word_counter <= {WC_WIDTH{1'b0}};
    else if (wc_inc)
        word_counter <= (word_counter + 1) % (2**WC_WIDTH);
end

always @ (posedge clk_i) begin
    if (ch_sel_reset)
        ch_sel <= {SEL_WIDTH{1'b0}};
    else if (ch_switch)
        ch_sel <= (ch_sel + 1) % (2**SEL_WIDTH);
end

always @ (posedge clk_i or posedge reset_i) begin
    if (reset_i)
        state <= STATE_QUERY;
    else
        state <= next_state;
end

always @ (*) begin
    tx_data_o <= CHANNEL_SELECT_WORD | ch_sel;
    wc_reset <= 1'b1;
    wc_inc <= 1'b0;
    ch_switch <= 1'b0;
    shi_mux_0_tx_en_i <= 1'b0;
    next_state  <= state;

    case (state)
        STATE_QUERY: begin
            if (tx_valid_o)
                next_state <= STATE_SELECT;
            else
                ch_switch  <= 1'b1;
        end

        STATE_SELECT: begin
            if (tx_en_i)
                next_state <= STATE_TX;
        end

        STATE_TX: begin
            wc_reset <= 1'b0;
            wc_inc <= tx_en_i;
            tx_data_o <= shi_mux_0_tx_data_o;
            shi_mux_0_tx_en_i <= tx_en_i;
            if (~tx_valid_o || ch_force_switch) begin
                ch_switch <= 1'b1;
                next_state <= STATE_QUERY;
            end
        end
    endcase
end

endmodule

