/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Create Date:    13:45:33 06/11/2012 
// Module Name:    shi_uart_rx 
//////////////////////////////////////////////////////////////////////////////////
module shi_uart_rx
#(
  parameter WORD_SIZE     = 8,
  parameter CLKIN_FREQ_HZ = 25000000,
  parameter BAUDRATE_BPS  = 115200
)(
  input  wire reset_i,
  input  wire clk_i,
  
  //Interface to external pin
  input  wire rx_line_i,

  //Internal SHI interfaces
  output wire [WORD_SIZE-1:0] tx_data_o,
  output wire tx_valid_o
);
  `include "shi/uart/util/uart_functions.vh"
  
  //Datapath variables
  localparam 
  BC_SIZE             = log2(WORD_SIZE - 1),
  SAMPLING_COUNT      = find_best_ovs_rate(CLKIN_FREQ_HZ, BAUDRATE_BPS), 
  SAMPLING_COUNT_HALB = (SAMPLING_COUNT - 1) / 2,
  SC_SIZE             = log2(SAMPLING_COUNT),
  BR_DIVISOR_RX       = baud_divisor(ovs_freq_10x(CLKIN_FREQ_HZ, SAMPLING_COUNT), BAUDRATE_BPS);
  
  wire w_enable;
  wire w_pre_enable;
  wire w_bit_eq_0, w_sc_halb, w_sc_full, w_frame_done;
  reg [WORD_SIZE-1:0] r_shf_rx;
  reg [BC_SIZE:0]     r_bit_cnt;
  reg [SC_SIZE-1:0]   r_sampling_cnt;

  //Controller variables
  localparam
  STATE_COUNT = 3,
  IDLE        = 3'b001,
  STARTING    = 3'b010,
  RECEIVING   = 3'b100;
  
  reg [STATE_COUNT-1:0] r_state_cur, r_state_next;
  reg r_sc_inc, r_sc_clr, r_bc_inc, r_bc_clr, r_shift, r_tx_valid;
  
  //Datapath
  assign w_bit_eq_0   = (rx_line_i == 1'b0);
  assign w_sc_halb    = (r_sampling_cnt == SAMPLING_COUNT_HALB - 1'b1);
  assign w_sc_full    = (r_sampling_cnt == SAMPLING_COUNT - 1'b1);
  assign w_frame_done = (r_bit_cnt == WORD_SIZE);
  assign tx_data_o    = r_shf_rx; 
  assign tx_valid_o   = r_tx_valid && w_enable;

  always @(posedge clk_i, posedge reset_i) begin
    if(reset_i) begin
      r_bit_cnt      <= {(BC_SIZE+1){1'b0}};
      r_sampling_cnt <= {(SC_SIZE){1'b0}};
      r_shf_rx       <= {(WORD_SIZE){1'b0}};
    end
    else if(w_enable)begin
      if(r_sc_clr)
        r_sampling_cnt <= {(SC_SIZE){1'b0}};
      else if(r_sc_inc) 
        r_sampling_cnt <= r_sampling_cnt + 1'b1;

      if(r_bc_clr)
        r_bit_cnt <= {(BC_SIZE+1){1'b0}};
      else if(r_bc_inc)
        r_bit_cnt <= r_bit_cnt + 1'b1;
			
      if(r_shift)
        r_shf_rx <= {rx_line_i, r_shf_rx[WORD_SIZE-1:1]};		  
    end
  end
  
  //Controller 
  always @(r_state_cur, w_bit_eq_0, w_sc_halb, w_sc_full, w_frame_done) begin
    r_sc_inc = 1'b0;
    r_sc_clr = 1'b0;
    r_bc_inc = 1'b0;
    r_bc_clr = 1'b0;
    r_shift  = 1'b0;
    r_tx_valid   = 1'b0;
    r_state_next = IDLE;
		
    case(r_state_cur)
      IDLE:
        if(w_bit_eq_0) begin
          r_state_next = STARTING;
        end
      STARTING:
        if(!w_bit_eq_0) begin
          r_sc_clr = 1'b1;
          r_state_next = IDLE;
        end
        else if(w_sc_halb) begin
          r_sc_clr = 1'b1;
          r_state_next = RECEIVING;
        end
        else begin
          r_sc_inc = 1'b1;
          r_state_next = STARTING;
        end
      RECEIVING:
        if(w_sc_full) begin
          r_sc_clr = 1'b1;
          if(w_frame_done) begin
            r_bc_clr = 1'b1;
            r_tx_valid = 1'b1;
            r_state_next = IDLE;
          end
          else begin
            r_shift  = 1'b1;
            r_bc_inc = 1'b1;
            r_state_next = RECEIVING;
          end
        end
        else begin
          r_sc_inc = 1'b1;
          r_state_next = RECEIVING;
        end
      default: r_state_next = IDLE;
    endcase
  end
	 
  always @(posedge clk_i, posedge reset_i) begin
    if(reset_i) begin
      r_state_cur <= IDLE;
    end
    else if(w_enable) begin 
      r_state_cur <= r_state_next;
    end
  end

  //Baudrate Generator
  shi_uart_baudgen #(
    .BAUDRATE_DIVISOR(BR_DIVISOR_RX)
  )
  shi_uart_baudgen_inst0(
    .clk_i(clk_i),
    .reset_i(reset_i),
    .enable_o(w_enable)
  );

endmodule
