/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Create Date:    17:07:11 06/11/2012 
// Module Name:    shi_uart_tx 
//////////////////////////////////////////////////////////////////////////////////
module shi_uart_tx
#(
  parameter WORD_SIZE     = 8,
  parameter CLKIN_FREQ_HZ = 25000000,
  parameter BAUDRATE_BPS  = 115200
)(
  input  wire reset_i,
  input  wire clk_i,
  
  //Interface to external pin
  output wire tx_line_o,

  //Internal SHI interfaces
  input  wire [WORD_SIZE-1:0] rx_data_i,
  input  wire rx_en_i,
  output wire rx_rdy_o
);
  `include "shi/uart/util/uart_functions.vh"
  
  //Datapath variables
  localparam 
  BC_SIZE       = log2(WORD_SIZE - 1),
  BR_DIVISOR_TX = CLKIN_FREQ_HZ/BAUDRATE_BPS;
  
  wire w_enable;
  wire w_frame_done;
  reg [WORD_SIZE:0] r_shf_tx;
  reg [BC_SIZE:0]   r_bit_cnt;

  //Controller variables
  localparam
  STATE_COUNT  = 2,
  IDLE         = 2'b01,
  TRANSMITTING = 2'b10;

  reg [STATE_COUNT-1:0] r_state_cur, r_state_next;
  reg r_start, r_transmit, r_clear;
  
  //Datapath	
  assign tx_line_o    = r_shf_tx[0];
  assign w_frame_done = (r_bit_cnt == WORD_SIZE);
  assign rx_rdy_o     = r_start && w_enable;
	
  always @(posedge clk_i, posedge reset_i) begin
    if(reset_i) begin
      r_shf_tx  <= {(WORD_SIZE+1){1'b1}};
      r_bit_cnt <= {(BC_SIZE+1){1'b0}};
    end
    else if(w_enable) begin
      if(r_start) begin 
        r_shf_tx <= {rx_data_i,1'b0};
      end
      else if(r_transmit) begin
        r_shf_tx  <= {1'b1, r_shf_tx[WORD_SIZE:1]};
        r_bit_cnt <= r_bit_cnt + 1'b1;
      end
      else if(r_clear) begin
        r_shf_tx  <= {(WORD_SIZE+1){1'b1}};
        r_bit_cnt <= {(BC_SIZE+1){1'b0}};
      end
    end
  end

  //Controller
  always @(r_state_cur, rx_en_i, w_frame_done) begin
    r_transmit   = 1'b0;
    r_clear      = 1'b0;
    r_start      = 1'b0;
    r_state_next = IDLE;
		
    case(r_state_cur)
      IDLE:
        if(rx_en_i) begin
          r_start      = 1'b1;
          r_state_next = TRANSMITTING;
        end
      TRANSMITTING:
        if(w_frame_done) begin
          r_clear      = 1'b1;
          r_state_next = IDLE;
        end
        else begin
          r_transmit   = 1'b1;
          r_state_next = TRANSMITTING;
        end
      default: r_state_next = IDLE;
    endcase
  end 
	
  always @(posedge clk_i, posedge reset_i) begin
    if(reset_i) begin
      r_state_cur <= IDLE;
    end
    else if(w_enable) begin 
      r_state_cur <= r_state_next;
    end
  end
  
  //Baudrate Generator
  shi_uart_baudgen #(
    .BAUDRATE_DIVISOR(BR_DIVISOR_TX)
  )
  shi_uart_baudgen_inst0(
    .clk_i(clk_i),
    .reset_i(reset_i),
    .enable_o(w_enable)
  );

endmodule
