/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`include "shi/usb/src/usb1_defines.vh"
`include "misc/memory/src/generic_memory.v"

module blockramusb( 

    // global
    input wire              clk,
    input wire              rst,

    // packet assembler interface
    input wire              rx_data_valid,
    input wire              rx_data_done,
    input wire  [15:0]      rx_data_st,
    output reg              zero_length,

    // packet disassembler interface
    output reg              send_data,
    input wire              rd_next,
    inout wire              tx_valid,
    output wire [7:0]       tx_data_st,

    // protocol engine
    input wire              tx_dma_en,
    input wire              rx_dma_en,
    output reg              idma_done,
    input wire  [3:0]       ep_sel,
    output wire [6:0]       funct_adr,

    // config register interface
    input wire              get_cfg,
    output reg              cfg_valid,
    output reg              ep_armed,
    output reg  [1:0]       ep_type,

    // status interface
    output reg              disconnect,
    output wire             send_stall,
    output wire             halt,
    output wire             configured,
    output wire             addressed,

    // device tx interface
    input wire [7:0]        ep_tx_data,
    input wire [6:0]        ep_tx_data_cnt,

    // device rx interface
    output reg [7:0]        ep_rx_data,
    output reg              ep_rx_valid,
    input wire              ep_rx_rdy

);

parameter   
    NUM_EP_IN           = 0,
    NUM_EP_OUT          = 0,
    VENDOR_STRING       = "",
    PRODUCT_STRING      = "",
    SERIAL_STRING       = "";

parameter [15:0]
    EP_TYPE_OUT_MAP     = 16'h00,
    EP_TYPE_IN_MAP      = 16'h00,
    VENDOR_ID           = 16'h00,
    PRODUCT_ID          = 16'h00,
    LANGUAGE_ID         = 16'h0000;

localparam [31:0]
    IDLE                =   1,
    WRITE_DATA          =   2,
    WRITE_CTRL          =   5,
    READ_DATA           =   4,
    READ_CTRL           =   12,
    READ_ZERO           =   11,
    READ_DATA_PREPARE   =   8,
    GET_STATUS_IN_EP     =   6,
    GET_STATUS_OUT_EP     =   3,
    GET_STATUS_CTRL_EP     =   10,
    GET_CFG_SET_VALID   =   9;

reg [31:0]  
    state, 
    next_state;

reg [6:0]  
    tx_size,
    tx_size_next;

reg [1:0]
    ep_type_next;

reg 
    ep_armed_next;

wire    [9:0]   ram_start;
wire    [6:0]   ctrl_size;

reg             rx_data_valid_r;
reg     [6:0]   rx_cnt_r;
reg             rx_cnt_clear;

reg     [6:0]   tx_cnt_r;
reg             tx_cnt_clear;

wire            ctrl_ep;
wire            out_ep;

// CTRL Interface
reg [7:0]
    hdr0,
    hdr1,
    hdr2,
    hdr3,
    hdr4,
    hdr5,
    hdr6,
    hdr7;

reg             hdr_done;
reg             get_header;
reg             ctrl_in;

wire    [7:0]   device_rom_data_rd;
wire    [7:0]   config_rom_data_rd;
wire    [7:0]   langid_rom_data_rd;
wire    [7:0]   vdescr_rom_data_rd;
wire    [7:0]   pdescr_rom_data_rd;
wire    [7:0]   sdescr_rom_data_rd;

wire    [7:0]   rom_data_rd;

assign rom_data_rd = device_rom_data_rd
    | config_rom_data_rd
    | langid_rom_data_rd
    | vdescr_rom_data_rd
    | pdescr_rom_data_rd
    | sdescr_rom_data_rd;

assign ctrl_ep = (ep_sel == 4'b0000);
assign out_ep = EP_TYPE_OUT_MAP[ep_sel];

assign tx_data_st = ctrl_ep ? rom_data_rd : ep_tx_data;

always @(posedge clk)
    rx_data_valid_r <= rx_data_valid;
    
//////////////////////////////////////////////////////////////////////
//
// Global Config Register Latch
//

always @(posedge clk or negedge rst)
    if(!rst)
        disconnect <= 1'b0;
    else
        disconnect <= 1'b1;

//////////////////////////////////////////////////////////////////////
//
// RX Byte Counter
//

always @(posedge clk)
    if(rx_cnt_clear)
        rx_cnt_r <= 7'h00;
    else if (rx_data_valid_r)
        rx_cnt_r <= rx_cnt_r + 7'h01;

//////////////////////////////////////////////////////////////////////
//
// TX Byte Counter
//

always @(posedge clk)
    if(tx_cnt_clear)
        tx_cnt_r <= 7'h00;
    else if (rd_next)
        tx_cnt_r <= tx_cnt_r + 7'h01;


//////////////////////////////////////////////////////////////////////
//
// Request Memory Registers
//


always @(posedge clk or negedge rst)
    if(!rst)
        begin
            hdr0 <= 8'h00;
            hdr1 <= 8'h00;
            hdr2 <= 8'h00;
            hdr3 <= 8'h00;
            hdr4 <= 8'h00;
            hdr5 <= 8'h00;
            hdr6 <= 8'h00;
            hdr7 <= 8'h00;
        end
    else
        if(get_header & rx_data_valid)
            case(rx_cnt_r)
                7'h00:  hdr0 <= rx_data_st[7:0];
                7'h01:  hdr1 <= rx_data_st[7:0];
                7'h02:  hdr2 <= rx_data_st[7:0];
                7'h03:  hdr3 <= rx_data_st[7:0];
                7'h04:  hdr4 <= rx_data_st[7:0];
                7'h05:  hdr5 <= rx_data_st[7:0];
                7'h06:  hdr6 <= rx_data_st[7:0];
                7'h07:  hdr7 <= rx_data_st[7:0];
            endcase

//////////////////////////////////////////////////////////////////////
//
// Status and Config Register FSM
//

always @(posedge clk or negedge rst) 
    if(!rst) begin
        state       <= IDLE;
        tx_size     <= 7'b0;
        ep_type     <= `EP_TYPE_NONE;
        ep_armed    <= 1'b0;
    end else begin
        state       <= next_state;
        tx_size     <= tx_size_next;
        ep_type     <= ep_type_next;
        ep_armed    <= ep_armed_next;
    end

always @(*)
begin
    next_state <= state;            // Default don't change current state

    case(state)
    IDLE:
        begin
            if(rx_dma_en)
                if (ctrl_ep)
                    next_state <= WRITE_CTRL;
                else
                    next_state <= WRITE_DATA;
            if(tx_dma_en)
                next_state <= READ_DATA_PREPARE;
            if(get_cfg)
                if (ctrl_ep)
                    next_state <= GET_STATUS_CTRL_EP;
                else if (out_ep)
                    next_state <= GET_STATUS_OUT_EP;
                else
                    next_state <= GET_STATUS_IN_EP;
        end
    
    GET_STATUS_IN_EP:
        next_state <= GET_CFG_SET_VALID;

    GET_STATUS_OUT_EP:
        next_state <= GET_CFG_SET_VALID;

    GET_STATUS_CTRL_EP:
        next_state <= GET_CFG_SET_VALID;

    GET_CFG_SET_VALID:
        next_state <= IDLE;

    WRITE_DATA:
        if(rx_data_done)
            next_state <= IDLE;

    WRITE_CTRL:
        if(rx_data_done)
            next_state <= IDLE;

    READ_DATA_PREPARE:
        if (tx_size == 0)
            next_state <= READ_ZERO;
        else
            if (ctrl_ep) 
                next_state <= READ_CTRL;
            else
                next_state <= READ_DATA;    

    READ_ZERO:
        next_state <= IDLE;

    READ_DATA:
        if (tx_cnt_r >= tx_size)
            next_state <= IDLE;

    READ_CTRL:
        if (tx_cnt_r >= tx_size)
            next_state <= IDLE;

    default:
        next_state <= IDLE;
    endcase
end

always @(*) 
begin

    send_data        <= 1'b0;            
    idma_done        <= 1'b1;
    cfg_valid        <= 1'b0;
    zero_length      <= 1'b0;
    get_header       <= 1'b0;
    hdr_done         <= 1'b0;
    ctrl_in          <= 1'b0;
    tx_cnt_clear     <= 1'b0;
    rx_cnt_clear     <= 1'b0;
    ep_type_next     <= ep_type; 
    ep_armed_next    <= ep_armed;
    tx_size_next     <= tx_size;
    ep_rx_valid      <= 1'b0;
    ep_rx_data       <= 8'h00;
            
    case(state)         

    IDLE: begin
        tx_cnt_clear    <= 1'b1;
        rx_cnt_clear    <= 1'b1;
    end

    GET_STATUS_IN_EP:
        begin       
            ep_armed_next   <= (ep_tx_data_cnt != 0);
            ep_type_next    <= `EP_TYPE_IN; 
            tx_size_next    <= ep_tx_data_cnt;
        end

    GET_STATUS_OUT_EP:
        begin       
            ep_armed_next   <= ep_rx_rdy;
            ep_type_next    <= `EP_TYPE_OUT; 
        end

    GET_STATUS_CTRL_EP:
        begin
            ep_armed_next   <= 1'b1;
            ep_type_next    <= `EP_TYPE_CTRL;
            tx_size_next    <= ctrl_size[5:0];
        end

    GET_CFG_SET_VALID:
        begin
            cfg_valid   <= 1'b1;
        end

    WRITE_CTRL:
        begin
            get_header <= 1'b1;
            
            if(rx_cnt_r >= 7'h7)
                hdr_done    <= 1'b1;
        
        end

    WRITE_DATA:     
        begin
            ep_rx_valid <= rx_data_valid & ep_armed;
            ep_rx_data <= rx_data_st[7:0];
        end

    READ_DATA_PREPARE:
        begin
            tx_cnt_clear <= 1'b1;
            ctrl_in <= 1'b1;
        end
    READ_ZERO:
        begin
            idma_done <= 1'b0;
            send_data <= 1'b1;
            zero_length <= 1'b1;
        end
    READ_CTRL:
        begin
            idma_done <= 1'b0;
            
            if(tx_cnt_r < tx_size)
                send_data <= 1'b1;
                            
            ctrl_in <= 1'b1;

        end
    READ_DATA:
        begin
            idma_done <= 1'b0;
            
            if(tx_cnt_r < tx_size)
                send_data <= 1'b1;
                            
        end
    endcase
end

usb1_ctrl   ctrl(   
            .clk(       clk     ),
            .rst(       rst     ),
            .ctrl_setup(    1'b0        ),
            .ctrl_in(   ctrl_in     ),
            .ctrl_out(  1'b0        ),
            .send_stall( send_stall ),
            .addressed(addressed),
            .funct_adr( funct_adr   ),
            .configured( configured ),
            .halt(      halt        ),
            .hdr0(      hdr0        ),
            .hdr1(      hdr1        ),
            .hdr2(      hdr2        ),
            .hdr3(      hdr3        ),
            .hdr4(      hdr4        ),
            .hdr5(      hdr5        ),
            .hdr6(      hdr6        ),
            .hdr7(      hdr7        ),
            .hdr_done(  hdr_done    ),
            .rom_start_d(   ram_start   ),
            .ctrl_size( ctrl_size   )
        );

usb1_device_rom usb1_device_rom_0(
    .addr_i(tx_cnt_r[4:0]),
    .cs_i(ram_start[0]),
    .data_rd_o(device_rom_data_rd)
);
defparam
    usb1_device_rom_0.PRODUCT_ID    = PRODUCT_ID,
    usb1_device_rom_0.VENDOR_ID     = VENDOR_ID;

usb1_config_rom usb1_config_rom_0(
    .addr_i(tx_cnt_r),
    .cs_i(ram_start[1]),
    .data_rd_o(config_rom_data_rd)
);
defparam
    usb1_config_rom_0.EP_TYPE_IN_MAP    = EP_TYPE_IN_MAP,
    usb1_config_rom_0.EP_TYPE_OUT_MAP   = EP_TYPE_OUT_MAP,
    usb1_config_rom_0.NUM_EP_OUT        = NUM_EP_OUT,
    usb1_config_rom_0.NUM_EP_IN         = NUM_EP_IN;

usb1_strdescr_rom usb1_langid_rom_0(
    .addr_i(tx_cnt_r[2:0]),
    .cs_i(ram_start[2]),
    .data_rd_o(langid_rom_data_rd)
);
defparam
    usb1_langid_rom_0.DATA_SIZE         = 2,
    usb1_langid_rom_0.RAW_DATA          = {LANGUAGE_ID[7:0],LANGUAGE_ID[15:8]};

usb1_strdescr_rom usb1_vdescr_rom_0(
    .addr_i(tx_cnt_r[usb1_vdescr_rom_0.AWIDTH-1:0]),
    .cs_i(ram_start[3]),
    .data_rd_o(vdescr_rom_data_rd)
);
defparam
    usb1_vdescr_rom_0.STR_DATA = VENDOR_STRING;

usb1_strdescr_rom usb1_pdescr_rom_0(
    .addr_i(tx_cnt_r[usb1_pdescr_rom_0.AWIDTH-1:0]),
    .cs_i(ram_start[4]),
    .data_rd_o(pdescr_rom_data_rd)
);
defparam
    usb1_pdescr_rom_0.STR_DATA = PRODUCT_STRING;

usb1_strdescr_rom usb1_sdescr_rom_0(
    .addr_i(tx_cnt_r[usb1_sdescr_rom_0.AWIDTH-1:0]),
    .cs_i(ram_start[5]),
    .data_rd_o(sdescr_rom_data_rd)
);
defparam
    usb1_sdescr_rom_0.STR_DATA = SERIAL_STRING;

endmodule
