/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
`include "shi/usb/src/usb1_defines.vh"

module usb1_config_rom(

  input wire [6:0]        addr_i,
  input wire              cs_i,
  output wire [7:0]       data_rd_o

);

`include "shi/usb/include/usb1_functions.vh"

parameter
  NUM_EP_IN       = 0,
  NUM_EP_OUT      = 0;

parameter [15:0]
  EP_TYPE_OUT_MAP = 16'h00,
  EP_TYPE_IN_MAP  = 16'h00;

localparam
  NUM_ENDPOINTS   = NUM_EP_OUT + NUM_EP_IN;

localparam [7:0]
  NUM_INTERFACES      = 8'h01,
  CONFIG_INDEX        = 8'h01,
  CONFIG_DESCR_INDEX  = 8'h00,
  POWER_MODE          = `USB_SELF_POWERED,
  POWER_CONS          = 8'h00,
  IFACE_INDEX         = 8'h00,
  ALT_SETTING         = 8'h00,
  CLASS               = `USB_CLASS_VENDOR_SPECIFIC,
  SUBCLASS            = 8'h01,
  PROTOCOL            = 8'hFF,
  IFACE_DESCR_INDEX   = 8'h02,
  ATTR                = `USB_ATTRIBUTE_BULK,
  POLL_INTERVAL       = 8'h01;

localparam
  CONFIG_SIZE     = `USB_CONFIG_DESCR_SIZE
  + `USB_INTERFACE_DESCR_SIZE
  + (NUM_ENDPOINTS * `USB_ENDPOINT_DESCR_SIZE),
    PACKET_SIZE     = 16'h0040;



reg [55:0] rom_data_rd;

assign data_rd_o = cs_i ? 
  rom_data_rd[55:48] | 
  rom_data_rd[47:40] | 
  rom_data_rd[39:32] | 
  rom_data_rd[31:24] |
  rom_data_rd[23:16] |
  rom_data_rd[15:8] |
  rom_data_rd[7:0]
  : 8'h00;

always @(*)
begin
  case(addr_i)
    // configuration descriptor
    7'h00: rom_data_rd[7:0] <= `USB_CONFIG_DESCR_SIZE;
    7'h01: rom_data_rd[7:0] <= `USB_CONFIG_DESCRIPTOR;
    7'h02: rom_data_rd[7:0] <= CONFIG_SIZE[7:0];
    7'h03: rom_data_rd[7:0] <= CONFIG_SIZE[15:8];
    7'h04: rom_data_rd[7:0] <= NUM_INTERFACES;
    7'h05: rom_data_rd[7:0] <= CONFIG_INDEX;
    7'h06: rom_data_rd[7:0] <= CONFIG_DESCR_INDEX;
    7'h07: rom_data_rd[7:0] <= POWER_MODE;
    7'h08: rom_data_rd[7:0] <= POWER_CONS;

    // interface descriptor
    7'h09: rom_data_rd[7:0] <= `USB_INTERFACE_DESCR_SIZE;
    7'h0A: rom_data_rd[7:0] <= `USB_INTERFACE_DESCRIPTOR;
    7'h0B: rom_data_rd[7:0] <= IFACE_INDEX;
    7'h0C: rom_data_rd[7:0] <= ALT_SETTING;
    7'h0D: rom_data_rd[7:0] <= NUM_ENDPOINTS[7:0];
    7'h0E: rom_data_rd[7:0] <= CLASS;
    7'h0F: rom_data_rd[7:0] <= SUBCLASS;
    7'h10: rom_data_rd[7:0] <= PROTOCOL;
    7'h11: rom_data_rd[7:0] <= IFACE_DESCR_INDEX;
    default: 
           rom_data_rd[7:0] <= 8'h00;
  endcase
end

genvar i;

// endpoint descriptors
generate
  for (i=0; i<NUM_ENDPOINTS; i=i+1)
  begin:EP_DESCR_ROM
    localparam [6:0] 
      ROM_BASE = 7'h12 + (i*7'h7);
    localparam [7:0]
      EP_NUM = get_nth_one_bit_pos(EP_TYPE_IN_MAP | EP_TYPE_OUT_MAP,i),
      EP_ADDR = EP_TYPE_IN_MAP[EP_NUM] ? 8'h80 + EP_NUM : EP_NUM;
    always @(*)
    begin
      case(addr_i)
        ROM_BASE+7'h0: rom_data_rd[i*8+15:i*8+8] <= `USB_ENDPOINT_DESCR_SIZE;
        ROM_BASE+7'h1: rom_data_rd[i*8+15:i*8+8] <= `USB_ENDPOINT_DESCRIPTOR;
        ROM_BASE+7'h2: rom_data_rd[i*8+15:i*8+8] <= EP_ADDR;
        ROM_BASE+7'h3: rom_data_rd[i*8+15:i*8+8] <= ATTR;
        ROM_BASE+7'h4: rom_data_rd[i*8+15:i*8+8] <= PACKET_SIZE[7:0];
        ROM_BASE+7'h5: rom_data_rd[i*8+15:i*8+8] <= PACKET_SIZE[15:8];
        ROM_BASE+7'h6: rom_data_rd[i*8+15:i*8+8] <= POLL_INTERVAL;
        default: rom_data_rd[i*8+15:i*8+8] <= 8'h00;
      endcase
    end
  end
endgenerate

endmodule
