/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none

module usb1_fifos_rx #(

  parameter EP_TYPE_OUT_MAP = "",
  parameter MAX_NUM_EP = 1,
  parameter DEVICE_RX_VALID_MSB = 1,
  parameter DEVICE_RX_EN_MSB = 1,
  parameter DEVICE_RX_DATA_MSB = 1,
  parameter FIFO_SIZE = 64

)(

  input wire [MAX_NUM_EP-1:0] clk_rd_i,
  input wire          clk_wr_i,
  input wire          reset_i,

  input wire [3:0]    ep_sel_i,

  input wire [7:0]    ep_rx_data_i,
  input wire          ep_rx_valid_i,
  output reg          ep_rx_rdy_o,

  output wire [DEVICE_RX_DATA_MSB:0]  device_rx_data_o,
  output wire [DEVICE_RX_VALID_MSB:0] device_rx_valid_o,
  input wire [DEVICE_RX_EN_MSB:0]     device_rx_en_i,

  input wire          transfer_failed_i,
  input wire          transfer_finished_i

);

`include "common/functions/bitwidth.vh"

localparam
  FIFO_RX_ASIZE = bitwidth(FIFO_SIZE - 1);

wire [7:0] yaf_rx_rdata [MAX_NUM_EP-1:0];
wire [FIFO_RX_ASIZE:0] yaf_rx_fill [MAX_NUM_EP-1:0];
wire [MAX_NUM_EP-1:0] yaf_rx_rinc;
wire [MAX_NUM_EP-1:0] yaf_rx_winc;
wire [MAX_NUM_EP-1:0] yaf_rx_wfull;
wire [MAX_NUM_EP-1:0] yaf_rx_rempty;
wire [MAX_NUM_EP-1:0] ep_out_sel;

integer j;

always @(*)
begin:EP_OUT_SEL_ENCODER

  ep_rx_rdy_o <= 1'b0;

  for (j=0; j<MAX_NUM_EP; j=j+1)
    if (ep_out_sel[j])
    begin
      ep_rx_rdy_o <= (yaf_rx_fill[j][FIFO_RX_ASIZE:0] <= ((2**FIFO_RX_ASIZE) - 64));
    end

end

genvar i;

generate
  for (i=0; i<MAX_NUM_EP; i=i+1)
  begin : DEVICE_RX_FIFO

    if (EP_TYPE_OUT_MAP[i+1])
    begin
      assign ep_out_sel[i]        = (ep_sel_i == (i+1));
      assign device_rx_valid_o[i] = ~yaf_rx_rempty[i];
      assign yaf_rx_winc[i]       = ep_out_sel[i] & ep_rx_valid_i;
      assign yaf_rx_rinc[i]       = device_rx_en_i[i] & device_rx_valid_o[i];
      
      yaf #(
        .ASIZE(FIFO_RX_ASIZE),
        .DSIZE(8),
        .ENABLE_PREWRITE("YES")
      ) yaf_rx (

        .wclk(clk_wr_i),
        .wrst_n(~reset_i),
        .winc(yaf_rx_winc[i]),
        .wdata(ep_rx_data_i),
        .wfull(yaf_rx_wfull[i]),

        .rclk(clk_rd_i[i]),
        .rrst_n(~reset_i),
        .rinc(yaf_rx_rinc[i]),
        .rempty(yaf_rx_rempty[i]),
        .rdata(device_rx_data_o[(i*8)+7:i*8]),
        .rptr_rest(),
        .rptr_upd(),
        .wptr_upd(transfer_finished_i),
        .wptr_rest(transfer_failed_i),

        .fill(yaf_rx_fill[i][FIFO_RX_ASIZE:0])

      );
    end else begin
      assign ep_out_sel[i]                  = 1'b0;
      assign device_rx_valid_o[i]           = 1'b0;
      assign device_rx_data_o[(i*8)+7:i*8]  = 8'h00;
    end
  end

endgenerate


endmodule
