/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none

module usb1_fifos_tx #(

  parameter EP_TYPE_IN_MAP = "",
  parameter MAX_NUM_EP = 0,
  parameter DEVICE_TX_DATA_MSB = 1,
  parameter DEVICE_TX_RDY_MSB = 1,
  parameter DEVICE_TX_VALID_MSB = 1,
  parameter FIFO_SIZE = 64

)(

  input wire          clk_rd_i,
  input wire [MAX_NUM_EP-1:0] clk_wr_i,
  input wire          reset_i,

  input wire [3:0]    ep_sel_i,
  output reg [7:0]    ep_tx_data_o,
  output wire [6:0]   ep_tx_data_cnt_o,
  input wire          ep_tx_next_i,

  input wire [DEVICE_TX_DATA_MSB:0]   device_tx_data_i,
  input wire [DEVICE_TX_VALID_MSB:0]  device_tx_valid_i,
  output wire [DEVICE_TX_RDY_MSB:0]   device_tx_rdy_o,

  input wire          transfer_finished_i,
  input wire          transfer_failed_i

);

`include "common/functions/bitwidth.vh"

localparam
  FIFO_TX_ASIZE = bitwidth(FIFO_SIZE - 1);

reg [FIFO_TX_ASIZE:0] ep_tx_data_cnt_int;

wire [7:0] yaf_tx_rdata [MAX_NUM_EP-1:0];
wire [FIFO_TX_ASIZE:0] yaf_tx_fill [MAX_NUM_EP-1:0];
wire [MAX_NUM_EP-1:0] yaf_tx_rinc;
wire [MAX_NUM_EP-1:0] yaf_tx_winc;
wire [MAX_NUM_EP-1:0] yaf_tx_wfull;
wire [MAX_NUM_EP-1:0] yaf_tx_rempty;
wire [MAX_NUM_EP-1:0] ep_in_sel;

assign ep_tx_data_cnt_o = ep_tx_data_cnt_int > 64 ? 7'd64 : ep_tx_data_cnt_int;

always @(*)
begin:EP_IN_SEL_ENCODER

  integer i;

  ep_tx_data_o        <= 8'h00;
  ep_tx_data_cnt_int  <= {FIFO_TX_ASIZE+1{1'b0}};

  for (i=0; i<MAX_NUM_EP; i=i+1)
    if (ep_in_sel[i])
    begin
      ep_tx_data_o <= yaf_tx_rdata[i][7:0];
      ep_tx_data_cnt_int <= yaf_tx_fill[i][FIFO_TX_ASIZE:0];
    end

end

genvar i;

generate
  for (i=0; i<MAX_NUM_EP; i=i+1)
  begin : DEVICE_TX_FIFO

    if (EP_TYPE_IN_MAP[i+1])
    begin
      assign ep_in_sel[i]       = (ep_sel_i == (i+1));
      assign device_tx_rdy_o[i] = ~yaf_tx_wfull[i];
      assign yaf_tx_winc[i]     = device_tx_rdy_o[i] & device_tx_valid_i[i];
      assign yaf_tx_rinc[i]     = ep_in_sel[i] & ep_tx_next_i;
      
      yaf #(
        .ASIZE(FIFO_TX_ASIZE),
        .DSIZE(8),
        .ENABLE_PREREAD("YES")
      ) yaf_tx (

        .wclk(clk_wr_i[i]),
        .wrst_n(~reset_i),
        .winc(yaf_tx_winc[i]),
        .wdata(device_tx_data_i[(i*8)+7:i*8]),
        .wfull(yaf_tx_wfull[i]),

        .rclk(clk_rd_i),
        .rrst_n(~reset_i),
        .rinc(yaf_tx_rinc[i]),
        .rempty(yaf_tx_rempty[i]),
        .rdata(yaf_tx_rdata[i][7:0]),

        .fill(yaf_tx_fill[i][FIFO_TX_ASIZE:0]),

        .rptr_upd(transfer_finished_i),
        .rptr_rest(transfer_failed_i),
        .wptr_upd(),
        .wptr_rest()

      );
    end else begin
      assign ep_in_sel[i]       = 1'b0;
      assign device_tx_rdy_o[i] = 1'b0;
    end
  end

endgenerate

endmodule
