/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 10ps
`default_nettype none

module usb_tx_phy2(
  input wire          clk_i,
  input wire          reset_i,
  input wire          bit_shift_en_i,

  input wire [7:0]    data_i,
  input wire          data_valid_i,
  output wire         data_next_o,

  output reg          tx_dp_o,
  output reg          tx_dn_o,
  output reg          tx_en_o
);

localparam
  IDLE    = 0,
  DATA    = 1,
  EOP1    = 2,
  EOP2    = 3,
  EOP3    = 4;

reg   [2:0]   bit_cnt;

reg           tx_en_d;
wire          tx_nrz;

reg           tx_dp_d;
reg           tx_dn_d;

reg           bit_shift_res;
reg           bit_shift_ld;
reg   [7:0]   bit_shift_r;
wire          stuff;

reg   [31:0]
  state,
  next_state;

assign data_next_o = bit_shift_ld && bit_shift_en_i && ~stuff;

// bit shift register
always @(posedge clk_i)
begin
  if (bit_shift_res && bit_shift_en_i)
    bit_shift_r <= 8'b10000000;
  else if (bit_shift_en_i && ~stuff)
    if (bit_shift_ld)
      bit_shift_r <= data_i;
    else
      bit_shift_r <= {1'b0,bit_shift_r[7:1]};
end

// tx enable
always @(posedge clk_i or posedge reset_i)
begin
  if (reset_i) 
  begin
    tx_en_o <= 1'b1;
    tx_dp_o <= 1'b1;
    tx_dn_o <= 1'b0;
  end else begin
    tx_en_o <= tx_en_d;
    tx_dp_o <= tx_dp_d;
    tx_dn_o <= tx_dn_d;
  end
end

// bit counter
always @(posedge clk_i)
begin
  if (bit_shift_res && bit_shift_en_i)
    bit_cnt <= 3'b000;
  else
    if (bit_shift_en_i & ~stuff)
      if (bit_shift_ld)
        bit_cnt <= 3'b000;
      else
        bit_cnt <= (bit_cnt + 3'b001) & 3'b111;
end

// state machine
always @(posedge clk_i or posedge reset_i)
begin
  if (reset_i)
    state <= IDLE;
  else
    if (bit_shift_en_i)
      state <= next_state;
end

always @(*)
begin
  next_state    <= state;
  bit_shift_ld  <= 1'b0;
  bit_shift_res <= 1'b0;
  tx_en_d       <= 1'b1;
  tx_dp_d       <= 1'b1;
  tx_dn_d       <= 1'b0;

  case (state)
    IDLE: begin
      bit_shift_res <= 1'b1;
      if (data_valid_i)
      begin
        next_state <= DATA;
      end
    end
    DATA: begin
      tx_en_d <= 1'b0;
      tx_dp_d <= tx_nrz;
      tx_dn_d <= ~tx_nrz;
      if ((bit_cnt == 3'h7) && ~stuff)
        if (!data_valid_i)
          next_state <= EOP1;
        else
          bit_shift_ld <= 1'b1;
    end
    EOP1:
    begin
      tx_en_d <= 1'b0;
      tx_dp_d <= tx_nrz;
      tx_dn_d <= ~tx_nrz;
      if (~stuff)
      begin
        next_state <= EOP2;
        bit_shift_res <= 1'b1;
      end
    end
    EOP2: begin
      tx_en_d <= 1'b0;
      tx_dp_d <= 1'b0;
      tx_dn_d <= 1'b0;
      if (bit_cnt == 3'h1)
        next_state <= EOP3;
    end
    EOP3: begin
      tx_en_d <= 1'b0;
      next_state <= IDLE;
    end

  endcase

end

usb_tx_nrz nrz0(
  .clk_i(clk_i),
  .reset_i(bit_shift_res),
  .bit_shift_en_i(bit_shift_en_i),
  .bit_i(bit_shift_r[0]),
  .stuff_o(stuff),
  .tx_nrz(tx_nrz)
);

//wire [35:0] CONTROL;
//
//chipscope_icon icon0(
//    .CONTROL0(CONTROL)
//);
//
//chipscope_ila(
//    .CLK(clk_i),
//    .CONTROL(CONTROL),
//    .DATA({
//      reset_i,
//      bit_shift_en_i,
//
//      data_i[7:0],
//      data_valid_i,
//      data_next_o,
//
//      tx_dp_o,
//      tx_dn_o,
//      tx_en_o,
//
//      bit_cnt[2:0],
//
//      tx_en_d,
//      tx_nrz,
//
//      tx_dp_d,
//      tx_dn_d,
//
//      bit_shift_res,
//      bit_shift_ld,
//      bit_shift_r[0],
//      stuff,
//      state[2:0]
//    }),
//    .TRIG0(data_valid_i),
//    .TRIG1(1'b0),
//    .TRIG2(1'b0),
//    .TRIG3(1'b0),
//    .TRIG4(1'b0),
//    .TRIG5(1'b0),
//    .TRIG6(1'b0),
//    .TRIG7(1'b0)
//);

endmodule
