/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  USB 1.1 function IP core                                   ////
////                                                             ////
////                                                             ////
////  Author: Rudolf Usselmann                                   ////
////          rudi@asics.ws                                      ////
////                                                             ////
////                                                             ////
////  Based on: http://www.opencores.org/cores/usb1_funct/       ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`timescale 10ps/1ps
`default_nettype none

module usbcore_fifo  (

            reset_i,

            clk_fifo_i,
            clk_48_mhz_i,
            // USB PHY Interface
            usb_data_p_io,
            usb_data_n_io,
            // Global Config Register Interface
            disconnect_o,
            // Debug Interface
            // funct_adr,
            // dpid,

            //Status
            tx_active_o,
            rx_active_o,

            send_stall_o,
            configured_o,
            halt_o,
            addressed_o,

            transfer_failed_o,
            transfer_finished_o,

            // Fifo interface
            device_tx_data_i,
            device_tx_valid_i,
            device_tx_rdy_o,

            device_rx_data_o,
            device_rx_valid_o,
            device_rx_en_i
        );

parameter
    NUM_EP_IN           = 1,
    NUM_EP_OUT          = 1,
    EP_01_TYPE          = "IN",
    EP_02_TYPE          = "OUT",
    EP_03_TYPE          = "",
    EP_04_TYPE          = "",
    EP_05_TYPE          = "",
    EP_06_TYPE          = "",
    PRODUCT_STRING      = "",
    VENDOR_STRING       = "",
    SERIAL_STRING       = "",
    VENDOR_ID           = 0,
    PRODUCT_ID          = 0,
    LANGUAGE_ID         = 0,
    FIFO_SIZE           = 64,
    PRETEND_CRC16_ERROR_INTERVAL = 0,
    TRANSMIT_CORRUPT_CRC16_INTERVAL = 0;

localparam

// due to limited packet size of 64 byte for control transfers we currently
// only can support a maximum of 6 EPs

    EP_07_TYPE          = "(not supported)",
    EP_08_TYPE          = "(not supported)",
    EP_09_TYPE          = "(not supported)",
    EP_10_TYPE          = "(not supported)",
    EP_11_TYPE          = "(not supported)",
    EP_12_TYPE          = "(not supported)",
    EP_13_TYPE          = "(not supported)",
    EP_14_TYPE          = "(not supported)",
    EP_15_TYPE          = "(not supported)";

localparam [15:0]
    EP_TYPE_IN_MAP = {
        EP_15_TYPE == "IN",
        EP_14_TYPE == "IN",
        EP_13_TYPE == "IN",
        EP_12_TYPE == "IN",
        EP_11_TYPE == "IN",
        EP_10_TYPE == "IN",
        EP_09_TYPE == "IN",
        EP_08_TYPE == "IN",
        EP_07_TYPE == "IN",
        EP_06_TYPE == "IN",
        EP_05_TYPE == "IN",
        EP_04_TYPE == "IN",
        EP_03_TYPE == "IN",
        EP_02_TYPE == "IN",
        EP_01_TYPE == "IN",
        1'b0
    },
    EP_TYPE_OUT_MAP = {
        EP_15_TYPE == "OUT",
        EP_14_TYPE == "OUT",
        EP_13_TYPE == "OUT",
        EP_12_TYPE == "OUT",
        EP_11_TYPE == "OUT",
        EP_10_TYPE == "OUT",
        EP_09_TYPE == "OUT",
        EP_08_TYPE == "OUT",
        EP_07_TYPE == "OUT",
        EP_06_TYPE == "OUT",
        EP_05_TYPE == "OUT",
        EP_04_TYPE == "OUT",
        EP_03_TYPE == "OUT",
        EP_02_TYPE == "OUT",
        EP_01_TYPE == "OUT",
        1'b0
    },
    MAX_NUM_EP            = 6,
    DEVICE_TX_DATA_MSB    = (MAX_NUM_EP * 8) - 1,
    DEVICE_TX_VALID_MSB   = MAX_NUM_EP - 1,
    DEVICE_TX_RDY_MSB     = DEVICE_TX_VALID_MSB,
    DEVICE_RX_DATA_MSB    = (MAX_NUM_EP * 8) - 1,
    DEVICE_RX_VALID_MSB   = MAX_NUM_EP - 1,
    DEVICE_RX_EN_MSB      = DEVICE_RX_VALID_MSB;

input wire [DEVICE_TX_DATA_MSB:0] device_tx_data_i;
input wire [DEVICE_TX_VALID_MSB:0] device_tx_valid_i;
output wire [DEVICE_TX_RDY_MSB:0] device_tx_rdy_o;

output wire [DEVICE_RX_DATA_MSB:0] device_rx_data_o;
output wire [DEVICE_RX_VALID_MSB:0] device_rx_valid_o;
input wire [DEVICE_RX_EN_MSB:0] device_rx_en_i;

input wire [MAX_NUM_EP-1:0] clk_fifo_i;

output wire send_stall_o;
output wire configured_o;
output wire halt_o;
output wire addressed_o;

output wire transfer_finished_o;
output wire transfer_failed_o;

output  wire    tx_active_o;
output  wire  rx_active_o;

wire        txoe;
assign      tx_active_o = ~txoe;

input wire  reset_i;

input   wire clk_48_mhz_i;

inout   wire    usb_data_p_io;
inout   wire    usb_data_n_io;

// Global Config Register Interface
output wire disconnect_o;

// Debug Interface
// output   [6:0]   funct_adr;
// output       dpid;

wire    [6:0]   funct_adr;
wire        dpid;

///////////////////////////////////////////////////////////////////
//
// Local Wires and Registers
//

// UTMI Interface
wire    [7:0]   DataOut;
wire        TxValid;
wire        TxReady;
wire    [7:0]   DataIn;
wire        RxValid;
wire        RxActive;
wire        RxError;
wire    [1:0]   LineState;

wire    [7:0]   rx_data;
wire        rx_valid, rx_err;
wire    [7:0]   tx_data;
wire        tx_valid;
wire        tx_ready;
wire        tx_first;
wire        tx_valid_last;

wire    [3:0]   ep_sel;     // Endpoint Number Input

reg [7:0]   tx_data_st;
wire    [7:0]   rx_data_st;

reg     rst_local;  // internal reset
wire        usb_rst;
wire        rst_core;
reg [7:0]   rst_cnt;

wire            ep_tx_start;
wire    [7:0]   ep_tx_data;
wire    [6:0]   ep_tx_data_cnt;
wire            ep_tx_next;

wire    [7:0]   ep_rx_data;
wire            ep_rx_rdy;
wire            ep_rx_valid;

wire    [35:0]  CONTROL0;
wire    [35:0]  CONTROL1;
///////////////////////////////////////////////////////////////////
//
// Misc Logic
//

assign rst_core = rst_cnt[7];

always @(posedge clk_48_mhz_i or posedge reset_i)
    if(reset_i)
        rst_cnt <= 8'h00;
    else
        if (!rst_cnt[7])
            rst_cnt <= #1 rst_cnt + 8'h01;

always @(posedge clk_48_mhz_i)
    rst_local <= #1 rst_core & ~(usb_rst & disconnect_o);

///////////////////////////////////////////////////////////////////
//
// Module Instantiations
//

usb_phy phy (
            .clk(       clk_48_mhz_i        ),
            .rst(       rst_core    ),
            // ONLY external reset_i
            .phy_tx_mode(   1'b1        ),
            .usb_rst(   usb_rst     ),
            // Transceiver Interface
            .dp(        usb_data_p_io       ),
            .dn(        usb_data_n_io       ),
            // UTMI Interface
            .DataIn_o(  DataIn      ),
            .RxValid_o( RxValid     ),
            .RxActive_o(    RxActive    ),
            .RxError_o( RxError     ),
            .DataOut_i( DataOut     ),
            .TxValid_i( TxValid     ),
            .TxReady_o( TxReady     ),
            .LineState_o(   LineState   ),
            .txoe(      txoe        )
        );

// UTMI Interface
usb1_utmi_if utmi (
            .phy_clk(   clk_48_mhz_i        ),
            .rst(       rst_local   ),
            .DataOut(   DataOut     ),
            .TxValid(   TxValid     ),
            .TxReady(   TxReady     ),
            .RxValid(   RxValid     ),
            .RxActive(  RxActive    ),
            .RxError(   RxError     ),
            .DataIn(    DataIn      ),
            .rx_data(   rx_data     ),
            .rx_valid(  rx_valid    ),
            .rx_active( rx_active_o ),
            .rx_err(    rx_err      ),
            .tx_data(   tx_data     ),
            .tx_valid(  tx_valid    ),
            .tx_valid_last(tx_valid_last    ),
            .tx_ready(  tx_ready    ),
            .tx_first(  tx_first    ),
            .data_ready( /* N/C */  )
        );

// Protocol Layer
usb1_pl  #(
        .EP_TYPE_IN_MAP(EP_TYPE_IN_MAP),
        .EP_TYPE_OUT_MAP(EP_TYPE_OUT_MAP),
        .NUM_EP_OUT(NUM_EP_OUT),
        .NUM_EP_IN(NUM_EP_IN),
        .VENDOR_STRING(VENDOR_STRING),
        .PRODUCT_STRING(PRODUCT_STRING),
        .SERIAL_STRING(SERIAL_STRING),
        .VENDOR_ID(VENDOR_ID[15:0]),
        .PRODUCT_ID(PRODUCT_ID[15:0]),
        .LANGUAGE_ID(LANGUAGE_ID[15:0]),
        .PRETEND_CRC16_ERROR_INTERVAL(PRETEND_CRC16_ERROR_INTERVAL),
        .TRANSMIT_CORRUPT_CRC16_INTERVAL(TRANSMIT_CORRUPT_CRC16_INTERVAL)
    ) pl (
            .clk(       clk_48_mhz_i        ),
            .rst(       rst_local   ),
            .rx_data(   rx_data     ),
            .rx_valid(  rx_valid    ),
            .rx_active( rx_active_o ),
            .rx_err(    rx_err      ),
            .tx_data(   tx_data     ),
            .tx_valid(  tx_valid    ),
            .tx_valid_last( tx_valid_last   ),
            .tx_ready(  tx_ready    ),
            .tx_first(  tx_first    ),
            .tx_valid_out(  TxValid     ),
            .ep_sel(    ep_sel      ),
            .disconnect(    disconnect_o    ),
            .funct_adr( funct_adr   ),
            .dpid(      dpid        ),
            .configured(configured_o),
            .halt(halt_o),
            .addressed(addressed_o),
            .send_stall(send_stall_o),
            .ep_tx_start(ep_tx_start),
            .ep_tx_data_cnt(ep_tx_data_cnt),
            .ep_tx_data(ep_tx_data),
            .rd_next(ep_tx_next),
            .ep_rx_data(ep_rx_data),
            .ep_rx_valid(ep_rx_valid),
            .ep_rx_rdy(ep_rx_rdy),
            .transfer_failed(transfer_failed_o),
            .transfer_finished(transfer_finished_o)
    );

// Device TX FIFOs
usb1_fifos_tx #(
        .MAX_NUM_EP(MAX_NUM_EP),
        .EP_TYPE_IN_MAP(EP_TYPE_IN_MAP),
        .DEVICE_TX_RDY_MSB(DEVICE_TX_RDY_MSB),
        .DEVICE_TX_DATA_MSB(DEVICE_TX_DATA_MSB),
        .DEVICE_TX_VALID_MSB(DEVICE_TX_VALID_MSB),
        .FIFO_SIZE(FIFO_SIZE)
    ) usb1_fifos_tx_0 (
        .clk_rd_i(clk_48_mhz_i),
        .clk_wr_i(clk_fifo_i),
        .reset_i(reset_i),
        .ep_sel_i(ep_sel),
        .ep_tx_data_o(ep_tx_data),
        .ep_tx_data_cnt_o(ep_tx_data_cnt),
        .ep_tx_next_i(ep_tx_next),
        .device_tx_rdy_o(device_tx_rdy_o),
        .device_tx_valid_i(device_tx_valid_i),
        .device_tx_data_i(device_tx_data_i),
        .transfer_finished_i(transfer_finished_o),
        .transfer_failed_i(transfer_failed_o)
    );

// Device RX FIFOs
usb1_fifos_rx #(
        .MAX_NUM_EP(MAX_NUM_EP),
        .EP_TYPE_OUT_MAP(EP_TYPE_OUT_MAP),
        .DEVICE_RX_VALID_MSB(DEVICE_RX_VALID_MSB),
        .DEVICE_RX_EN_MSB(DEVICE_RX_EN_MSB),
        .DEVICE_RX_DATA_MSB(DEVICE_RX_DATA_MSB),
        .FIFO_SIZE(FIFO_SIZE)
    ) usb1_fifos_rx_0 (
        .clk_rd_i(clk_fifo_i),
        .clk_wr_i(clk_48_mhz_i),
        .reset_i(reset_i),
        .ep_sel_i(ep_sel),
        .ep_rx_data_i(ep_rx_data),
        .ep_rx_valid_i(ep_rx_valid),
        .ep_rx_rdy_o(ep_rx_rdy),
        .device_rx_en_i(device_rx_en_i),
        .device_rx_valid_o(device_rx_valid_o),
        .device_rx_data_o(device_rx_data_o),
        .transfer_finished_i(transfer_finished_o),
        .transfer_failed_i(transfer_failed_o)
    );

endmodule
