/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none

module usbcore_fifo_peri(
  /* SpartanMC peripheral interface */
  input wire                  clk_peri,
  input wire                  reset,
  input wire                  access_peri,
  input wire                  wr_peri,
  input wire      [9:0]       addr_peri,
  input wire      [17:0]      do_peri,
  output wire     [17:0]      di_peri,
  /* USB phy */
  input wire                  clk_48_mhz_i,
  inout wire                  usb_data_p_io,
  inout wire                  usb_data_n_io,
  inout wire                  disconnect_o,
  /* USB status */
  output wire                 addressed_o,
  output wire                 configured_o,
  output wire                 tx_active_o,
  output wire                 rx_active_o,
  output wire                 transfer_failed_o,
  output wire                 transfer_finished_o
);

parameter
  BASE_ADR            = 18'h00000,
  NUM_EP_IN           = 1,
  NUM_EP_OUT          = 1,
  EP_01_TYPE          = "IN",
  EP_02_TYPE          = "OUT",
  EP_03_TYPE          = "",
  EP_04_TYPE          = "",
  EP_05_TYPE          = "",
  EP_06_TYPE          = "",
  PRODUCT_STRING      = "",
  VENDOR_STRING       = "",
  SERIAL_STRING       = "",
  VENDOR_ID           = 0,
  PRODUCT_ID          = 0,
  LANGUAGE_ID         = 0,
  FIFO_SIZE           = 64,
  PRETEND_CRC16_ERROR_INTERVAL = 0,
  TRANSMIT_CORRUPT_CRC16_INTERVAL = 0;

localparam
  MAX_NUM_EP          = 6;

wire [(MAX_NUM_EP*8)-1:0] 
    device_tx_data_i,
    device_rx_data_o;

wire [MAX_NUM_EP-1:0] 
    device_tx_valid_i,
    device_tx_rdy_o,
    device_rx_valid_o,
    device_rx_en_i;

wire [12:0] 
    reg_read,
    reg_write;

usbcore_fifo #(
  .NUM_EP_IN(NUM_EP_IN),
  .NUM_EP_OUT(NUM_EP_OUT),
  .EP_01_TYPE(EP_01_TYPE),
  .EP_02_TYPE(EP_02_TYPE),
  .EP_03_TYPE(EP_03_TYPE),
  .EP_04_TYPE(EP_04_TYPE),
  .EP_05_TYPE(EP_05_TYPE),
  .EP_06_TYPE(EP_06_TYPE),
  .PRODUCT_STRING(PRODUCT_STRING),
  .VENDOR_STRING(VENDOR_STRING),
  .SERIAL_STRING(SERIAL_STRING),
  .VENDOR_ID(VENDOR_ID),
  .PRODUCT_ID(PRODUCT_ID),
  .LANGUAGE_ID(LANGUAGE_ID),
  .FIFO_SIZE(FIFO_SIZE),
  .PRETEND_CRC16_ERROR_INTERVAL(PRETEND_CRC16_ERROR_INTERVAL),
  .TRANSMIT_CORRUPT_CRC16_INTERVAL(TRANSMIT_CORRUPT_CRC16_INTERVAL)
) usbcore_fifo0 (
  .reset_i(reset),
  .clk_fifo_i({MAX_NUM_EP{clk_peri}}),
  .clk_48_mhz_i(clk_48_mhz_i),
  .usb_data_n_io(usb_data_n_io),
  .usb_data_p_io(usb_data_p_io),
  .disconnect_o(disconnect_o),
  .device_rx_en_i(device_rx_en_i),
  .device_rx_valid_o(device_rx_valid_o),
  .device_rx_data_o(device_rx_data_o),
  .device_tx_rdy_o(device_tx_rdy_o),
  .device_tx_valid_i(device_tx_valid_i),
  .device_tx_data_i(device_tx_data_i),
  .addressed_o(addressed_o),
  .configured_o(configured_o),
  .rx_active_o(rx_active_o),
  .tx_active_o(tx_active_o),
  .send_stall_o(),
  .halt_o(),
  .transfer_failed_o(transfer_failed_o),
  .transfer_finished_o(transfer_finished_o)
);

reg_access_decoder #( 
  .REG_COUNT(13),
  .BASE_ADR(BASE_ADR),
  .SIM_DEBUG("NO")
) reg_access_decoder0 (
  .addr_peri(addr_peri),
  .access_peri(access_peri),
  .wr_peri(wr_peri),
  .peri_select(),
  .reg_select(),
  .reg_read(reg_read),
  .reg_write(reg_write)
);

assign di_peri[17:0] = 
    reg_read[0] ? {4'b0,addressed_o,configured_o,device_rx_valid_o[5:0],device_tx_rdy_o[5:0]} :
    reg_read[7] ? {10'b0,device_rx_data_o[7:0]} :
    reg_read[8] ? {10'b0,device_rx_data_o[15:8]} :
    reg_read[9] ? {10'b0,device_rx_data_o[23:16]} :
    reg_read[10] ? {10'b0,device_rx_data_o[31:24]} :
    reg_read[11] ? {10'b0,device_rx_data_o[39:32]} :
    reg_read[12] ? {10'b0,device_rx_data_o[47:40]} :
    18'h00000;

assign device_tx_valid_i[5:0] = reg_write[6:1];
assign device_rx_en_i[5:0] = reg_read[12:7];

assign device_tx_data_i = {
        do_peri[7:0],
        do_peri[7:0],
        do_peri[7:0],
        do_peri[7:0],
        do_peri[7:0],
        do_peri[7:0]
    };

endmodule
