/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`timescale      1ns / 10ps
`default_nettype none

module sim_loader(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	locked_in,
	locked_out
	);


parameter BASE_ADR	= 10'h0;
parameter PORT_WIDTH	= 18;

parameter ALL_PORTS	= 3;
parameter DECODSIM_DEB	= "NO";

parameter OUTPUT_FILENAME = "output.txt";
parameter MEMORY_FILE = "../build/spartanmc_programram.txt";
parameter START_ADDRESS = 17'h2000;
localparam REAL_START_ADDRESS = 2*START_ADDRESS;
parameter FLUSH = 1;
parameter VERIFY = 1;

// inputs & outputs des Systembus
input	wire				clk_peri;
input	wire	[PORT_WIDTH-1:0]	do_peri;	// Datenbus 
output	wire	[17:0]			di_peri;	// Datenbus
input	wire	[9:0]			addr_peri;	// Adressbus
input	wire				access_peri;	// chipselect Peripherie
input	wire				wr_peri;	// Peripherie-Schreibzugriff
input	wire				reset;
input	wire				locked_in;
output	wire				locked_out;


// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					cs_out_status_r;// Lesen
wire					cs_out_data_r;	// Lesen
wire					cs_out_data_w;	// Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

wire [17:0] port_status;
wire fifo_tx_full, fifo_tx_empty, fifo_rx_full, fifo_rx_empty;
assign port_status = {14'b0, fifo_rx_full, fifo_rx_empty, fifo_tx_full, fifo_tx_empty}; // rx and tx switched compared to uart_light module!

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	cs_out_status_r	= reg_read[0];
assign	cs_out_data_r	= reg_read[1];	// Lesen
assign	cs_out_data_w	= reg_write[2];	// Schreiben
// Dekodierung der Adressen ********************** END **************************************

localparam FIFO_WIDTH = 10;
reg[17:0] programram [131071:0];
reg[7:0] fifo_rx [2**FIFO_WIDTH-1:0];
reg[7:0] fifo_tx [2**FIFO_WIDTH-1:0];
reg[FIFO_WIDTH:0] rx_rd, rx_wr, tx_rd, tx_wr;
localparam MASK = {FIFO_WIDTH{1'b1}};
assign fifo_rx_empty = (rx_rd == rx_wr);
assign fifo_tx_empty = (tx_rd == tx_wr);
assign fifo_rx_full = (rx_rd[FIFO_WIDTH] != rx_wr[FIFO_WIDTH]) && (rx_rd[FIFO_WIDTH-1:0] == rx_wr[FIFO_WIDTH-1:0]);
assign fifo_tx_full = (tx_rd[FIFO_WIDTH] != tx_wr[FIFO_WIDTH]) && (tx_rd[FIFO_WIDTH-1:0] == tx_wr[FIFO_WIDTH-1:0]);
wire[17:0] port_rx;
assign port_rx = {10'b0, fifo_tx[tx_rd&MASK]};


integer txFile, rxFile, rxIn;
reg restart; // low for restart
reg after_restart; // high after restart -> skip second initialization
wire locked_pulldown = (locked_in === 1'bx) ? 1'b0 : locked_in;
assign locked_out = locked_pulldown && restart;

initial begin
	$display("Writing STDOUT to %s",OUTPUT_FILENAME);
	txFile = $fopen(OUTPUT_FILENAME, "w");
	$display("Using %s as programram",MEMORY_FILE);
	rxFile = $fopen(MEMORY_FILE, "r");

	rx_rd = 0;
	rx_wr = 0;
	tx_rd = 0;
	tx_wr = 0;
	restart = 1;
	after_restart = 0;
end

wire [7:0] rx_read_data = fifo_rx[rx_rd&MASK];
wire [7:0] rx_write_data = fifo_rx[rx_wr&MASK];
wire [7:0] tx_read_data = fifo_tx[tx_rd&MASK];
wire [7:0] tx_write_data = fifo_tx[tx_wr&MASK];
reg [17:0] input_data;
reg [FIFO_WIDTH-1:0] i;
integer c;

initial begin
	@(negedge reset); // wait for reset
	while(fifo_rx_empty || fifo_rx[rx_rd] !== "S") begin // wait for "S"
		if(!fifo_rx_empty)
			rx_rd = rx_rd + 1;
		@(fifo_rx_empty or fifo_rx[rx_rd]);
	end
	repeat (22) begin // wait for string "SpMC loader v20140908\n"
		@(posedge clk_peri)
		if(fifo_rx_empty)
			@(!fifo_rx_empty);
		rx_rd = rx_rd + 1;
	end
	$display("found loader");

	// send sync sequence (64 times "*")
	repeat (64) begin
		fifo_tx[tx_wr] = "*";
		tx_wr = tx_wr + 1;
	end

	// rx fifo should now contain 64 "v"
	repeat (64) begin
		@(posedge clk_peri);
		if(fifo_rx_empty)
			@(negedge fifo_rx_empty);
		if(fifo_rx[rx_rd] !== "v") begin
			$display("Loader returned wrong value.");
			$finish;
		end
		rx_rd = rx_rd + 1;
	end
	// receive core id (three times 0x00)
	repeat (3) begin
		@(posedge clk_peri);
		if(fifo_rx_empty)
			@(negedge fifo_rx_empty);
		if(fifo_rx[rx_rd] !== 0) begin
			$display("Wrong core id. Continue anyway...");
		end
		rx_rd = rx_rd + 1;
	end
	$display("init sequence completed");

	// send address

	// lower 8 bit data
	// middle 8 bit data
	// 6 bit command + 2 bit data; commands: LOADER_CMD_NOP: 0xf; LOADER_CMD_SET_ADDR: 0x9; LOADER_CMD_WRITE_DATA: 0x5; LOADER_CMD_READ_DATA: 0x6;
	fifo_tx[tx_wr[FIFO_WIDTH-1:0]]   = REAL_START_ADDRESS[7:0];
	fifo_tx[tx_wr[FIFO_WIDTH-1:0]+1] = REAL_START_ADDRESS[15:8];
	fifo_tx[tx_wr[FIFO_WIDTH-1:0]+2] = {6'h9, REAL_START_ADDRESS[17:16]};
	tx_wr = tx_wr + 3;
	for(i=0; i < 3; i=i+1) begin
		@(posedge clk_peri);
		if(fifo_rx_empty) @(negedge fifo_rx_empty);
		if(fifo_rx[rx_rd[FIFO_WIDTH-1:0]] !== fifo_tx[tx_wr[FIFO_WIDTH-1:0]-3+i]) begin
			$display("Wrong start address response.");
			$finish;
		end
		rx_rd = rx_rd + 1;
	end
	$display("address set to %x", START_ADDRESS);

	// transmit data
	for(c=0; !$feof(rxFile); c=c+1) begin
		rxIn = $fscanf(rxFile, "%h\n", input_data);
		fifo_tx[(tx_wr)&MASK]   = input_data[7:0];
		fifo_tx[(tx_wr+1)&MASK] = input_data[15:8];
		fifo_tx[(tx_wr+2)&MASK] = {6'h5, input_data[17:16]};
		tx_wr = tx_wr + 3;
		for(i=0; i < 3; i=i+1) begin
			@(posedge clk_peri);
			if(fifo_rx_empty) @(negedge fifo_rx_empty);
			if(fifo_rx[rx_rd&MASK] !== fifo_tx[(tx_wr-3+i)&MASK]) begin
				$display("Wrong write data response. %x %x", fifo_rx[rx_rd&MASK], fifo_tx[(tx_wr-3+i)&MASK]);
				$finish;
			end
			rx_rd = rx_rd + 1;
		end
	end
	$display("data written.");

	// verify data
	if(VERIFY) begin
		i = $fseek(rxFile, 0, 0);
		fifo_tx[tx_wr&MASK]   = REAL_START_ADDRESS[7:0];
		fifo_tx[(tx_wr+1)&MASK] = REAL_START_ADDRESS[15:8];
		fifo_tx[(tx_wr+2)&MASK] = {6'h9, REAL_START_ADDRESS[17:16]};
		tx_wr = tx_wr + 3;
		for(i=0; i < 3; i=i+1) begin
			@(posedge clk_peri);
			if(fifo_rx_empty) @(negedge fifo_rx_empty);
			if(fifo_rx[rx_rd&MASK] !== fifo_tx[(tx_wr-3+i)&MASK]) begin
				$display("Wrong start address response.");
				$finish;
			end
			rx_rd = rx_rd + 1;
		end
		$display("address set to %x", START_ADDRESS);
		for(c=0; !$feof(rxFile); c=c+1) begin
			rxIn = $fscanf(rxFile, "%h\n", input_data);
			fifo_tx[tx_wr&MASK]   = 8'b0;
			fifo_tx[(tx_wr+1)&MASK] = 8'b0;
			fifo_tx[(tx_wr+2)&MASK] = {6'h6, 2'b0};
			tx_wr = tx_wr + 3;
			for(i=0; i < 3; i=i+1) begin
				@(posedge clk_peri);
				if(fifo_rx_empty) @(negedge fifo_rx_empty);
				if((i == 0 && fifo_rx[rx_rd&MASK] !== input_data[7:0]) || 
						(i == 1 && fifo_rx[rx_rd&MASK] !== input_data[15:8]) || 
						(i == 2 && fifo_rx[rx_rd&MASK] !== {6'h6, input_data[17:16]})) begin
					$display("Wrong read data response. %x %x %x %x", c, i, fifo_rx[(rx_rd)&MASK], input_data);
					$finish;
				end
				rx_rd = rx_rd + 1;
			end
		end
		$display("data verified.");
	end
	$display("Reset SpartanMC core");
	restart = 0;
	repeat(5) @(posedge clk_peri);
	restart = 1;
	after_restart = 1;
	
	// stop spartanmc loader (timeout takes too long)
	fifo_tx[tx_wr&MASK] = 8'h18;
	tx_wr = tx_wr + 1;
end

always @ (posedge clk_peri)
	if(cs_out_data_w & ~reset) begin
		// TODO check for fifo_rx_full
		if(!after_restart) begin // dont put data in fifo after first loading process
			fifo_rx[rx_wr&MASK] = do_peri[7:0];
			rx_wr <= rx_wr + 1;
		end

		if(after_restart) begin // write only application output to file
			$fwrite(txFile,"%c",do_peri);
			if (FLUSH==1) begin
				$fflush(txFile);
			end
		end
		if (do_peri>31)
			$display("TB tx %X %c",do_peri,do_peri);
		else
			$display("TB tx %X",do_peri);
	end else if(cs_out_data_r & ~reset) begin
		tx_rd <= tx_rd + 1;
		if (port_rx>31)
			$display("TB rx %X %c",port_rx,port_rx);
		else
			$display("TB rx %X",port_rx);
	end

assign di_peri = cs_out_status_r ? port_status : (cs_out_data_r ? port_rx : 18'b0);

endmodule
