/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : spi.v
// Autor : Steffen Kunze
// Modul : SPI
//
// Beschreibung : Der SPI Master ist fuer Datenuebertragung mit n*8 oder 9 bit
//		  ausgelegt. Cpha und cpol koennen programmeirt werden. Es
//		  stehen 16 slave select-Leitungen zur Verfuegung.
//
///////////////////////////////////////////////////////////////////////////////

`timescale      1ns / 10ps
`default_nettype none

module spi(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	intr,
	clk_spi,
	miso,
	mosi,
	sclk,
	ss
	//~ ,chipscope_out
);

// Die Parameter fuer die Adressen werden im Top-Modul z.B.: module_subsystem_0.v eingestellt.
parameter BASE_ADR	= 10'h0;

parameter CTRL_ADR	= 0;
parameter DATO_ADR	= 1;
parameter DATI_ADR	= 2;
parameter STAT_ADR	= 3;

parameter SPI_SS	= 4'h1;

// parameter ALL_PORTS	= 3;		//Anzahl der Ports fuer den Dekoder
parameter ALL_PORTS	= 4;		//Anzahl der Ports fuer den Dekoder
parameter DECODSIM_DEB	= "NO";		//DEBUG mode YES oder NO


input wire			clk_peri;
input wire	[17:0]		do_peri;	// Datenbus
output wire	[17:0]		di_peri;	// Datenbus
input wire	[9:0]		addr_peri;	// Adressbus
input wire			access_peri;	// chipselect peripherie
input wire			wr_peri;	// signalisiert Schreibzugriff Peripherie
input wire			reset;
output wire			intr;		// Interrupt flag

input wire			clk_spi;	// CLK aus dem sclk < 400KHz oder < 1MHz moeglich ist

input wire			miso;		// Daten: Master In Slave Out
output wire			mosi;		// Daten: Master Out Slave In
output wire			sclk;		// Takt fuer Uebertragung
output wire	[SPI_SS-1:0]	ss;		// Slave select

wire			ss_set;		// ist 1 wenn ss_mo aktualisiert wurde
reg	[14:0]		ss_mo;		// slave select master output
reg			sclk_int;	// Puffer fuer Uebertragungstakt
wire			clk_en;		// aus SPI-Takt mit Vorteiler resultierendes clk enable

// reg	[bpw*n-1:0]	data_transmit;	// Datenregister, in das vom Bus geschrieben wird
// reg	[bpw*n-1:0]	data_recieve;	// Datenregister, das vom Bus gelesen wird
reg	[17:0]		data_transmit;	// Datenregister, in das vom Bus geschrieben wird
reg	[17:0]		data_recieve;	// Datenregister, das vom Bus gelesen wird

//control- und statusregister
reg	[7:0]		control;
wire			en	= control[0];	// SPI enable
wire			int_en	= control[1];	// interrupt enable
wire			cpol	= control[2];	// idle-Pegel des Taktsignals
wire			cpha 	= control[3];	// Phasenlage des Taktes
						// 0 = Flanke(Daten) von Takt aktiv zu  Takt idle
						// 1 = Flanke(Daten) von Takt idle zu Takt aktiv
wire	[3:0]	slave_addr	= control[7:4];	// Slave Adresse fuer aktuellen Transfer
reg			ir_sig;			// Interrupt register
reg			transmit_end;		// zeigt an, dass Senden beendet wird
reg			transmit_empty;		// zeigt an, dass Sendedatenregister leer ist

//Register zur Steuerung von FSM, Schieberegister und Vorteiler
reg	[1:0]		state;			// Zustandsregister FSM
reg 		load, store, shift_i, shift_o;	// Schieberegistersteuersignale
//reg	[2:0]		s_cnt;			// Bitzaehler Schieberegister
reg	[4:0]		s_cnt;			// Bitzaehler Schieberegister
//reg	[2:0]		ir_cnt;			// Bit Anzahl fuer Interrupt
reg	[4:0]		ir_cnt;			// Bit Anzahl fuer Interrupt
//reg	[bpw*n-1:0]	sr_i,sr_o;		// Schieberegister
reg	[17:0]		sr_i,sr_o;		// Schieberegister
wire 			din, dout;		// Datenleitungen von/zu Schieberegister

reg	[2:0]		teiler;			// Vorteiler SPI-Takt
reg	[6:0]		cnt;			// Zaehler fuer Vorteiler SPI-Takt
reg			sclk_vor;		// SPI-Takt


// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Status-, Daten- und Kommandoregister
wire			rd_ctrl;	// Lesen
wire			rd_datao;	// Lesen
wire			rd_datai;	// Lesen
wire			rd_stat;	// Lesen
wire			wr_ctrl;	// Schreiben
wire			wr_datao;	// Schreiben
wire			cs_datao;	// r/w
wire			cs_datai;	// r/w

wire			select;		// Modul wird angesprochen
wire	[ALL_PORTS-1:0]	reg_select;	// r/w ADR
wire	[ALL_PORTS-1:0]	reg_read;	// read ADR
wire	[ALL_PORTS-1:0]	reg_write;	// write ADR

//~ output wire [44:0] chipscope_out;
//~ 
//~ assign chipscope_out[0] = wr_ctrl;
//~ assign chipscope_out[1] = rd_stat;
//~ assign chipscope_out[16:2] = ss_mo;
//~ assign chipscope_out[35:32] = slave_addr;
//~ assign chipscope_out[41:40] = state;
//~ assign chipscope_out[42] = ss_set;

// Addressdekoder fuer alle Geraeteregister (verwendet pselect.v und
// verzoegert die read Signale um einen Takt fuer die neue Pipeline) 
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	rd_ctrl		= reg_read[CTRL_ADR];		// Lesen
assign	rd_datao	= reg_read[DATO_ADR];		// Lesen
assign	rd_datai	= reg_read[DATI_ADR];		// Lesen
assign	rd_stat		= reg_read[STAT_ADR];		// Lesen
assign	wr_ctrl		= reg_write[CTRL_ADR];		// Schreiben
assign	wr_datao	= reg_write[DATO_ADR];		// Schreiben
assign	cs_datao	= reg_select[DATO_ADR];		// r/w
assign	cs_datai	= reg_select[DATI_ADR];		// r/w

// Dekodierung der Adressen ********************** END **************************************

// lesen der SPI Register
//assign di_peri[17:0]	= (rd_ctrl)?	{ss_set,(|ss_mo),teiler,ir_cnt,ir_sig,transmit_empty,control} :
//       			  (rd_datao)?	data_transmit :
//			  (rd_datai)?	data_recieve : 18'b0;
assign di_peri[17:0]	= (rd_stat) ?	{ss_set,(|ss_mo),ir_sig,transmit_empty} : 
			  (rd_datai)?	data_recieve : 
			  (rd_ctrl) ?	{teiler,ir_cnt,control} :
       			  (rd_datao)?	data_transmit :  18'b0;

// ***********************************************************
// Bei Verwendung einer buffer instance sind die Einstellungen
// in der UCF Datei fuer das Pin nicht wirksam!
//
// Spartan-3 input + output buffer
//IBUF	mi		(.O (din),
//			 .I (miso));
//OBUFT	mo		(.O (mosi),
//			 .I (dout),
//			 .T (~en));
//OBUFT sss[SPI_SS-1:0]	(.O (ss),		//an ss sollte pullup!
//			 .I (slave_sel[SPI_SS-1:0]),
//			 .T (~ss_mo[SPI_SS-1:0]));
//OBUF	sc		(.O (sclk),
//			 .I (sclk_int));
// ***********************************************************
assign	din		= miso;
assign	mosi		= en?	dout:	1'bz;

// Generierung der Slave Select Treiber
genvar	j;
generate
for (j=0; j<SPI_SS; j=j+1)
begin:GEN_OUTBUF
	assign  ss[j]	= en?	~ss_mo[j]: 1'bz;
        end
endgenerate
assign	sclk		= en?	sclk_int: 1'bz;


assign intr = int_en? ir_sig : 1'b0;

reg	store_buf, load_buf;

reg [14:0] wanted_ss;

always @(*) begin
	wanted_ss = 15'b0;

	case(slave_addr)
		4'b0001:wanted_ss[0]	= 1'b1;
		4'b0010:wanted_ss[1]	= 1'b1;
		4'b0011:wanted_ss[2]	= 1'b1;
		4'b0100:wanted_ss[3]	= 1'b1;
		4'b0101:wanted_ss[4]	= 1'b1;
		4'b0110:wanted_ss[5]	= 1'b1;
		4'b0111:wanted_ss[6]	= 1'b1;
		4'b1000:wanted_ss[7]	= 1'b1;
		4'b1001:wanted_ss[8]	= 1'b1;
		4'b1010:wanted_ss[9]	= 1'b1;
		4'b1011:wanted_ss[10]	= 1'b1;
		4'b1100:wanted_ss[11]	= 1'b1;
		4'b1101:wanted_ss[12]	= 1'b1;
		4'b1110:wanted_ss[13]	= 1'b1;
		4'b1111:wanted_ss[14]	= 1'b1;
	endcase
end

assign ss_set = (ss_mo == wanted_ss);

// In diesem Block werden Daten vom Bus in die Register uebernommen.
always @ (posedge clk_peri)
if(reset)
	begin
	control		<= 8'h00;
	data_transmit	<= 0;
	ir_cnt		<= 5'b00000;			// 8 Bit einstellen
	teiler		<= 3'b111;			// 2^7 einstellen
	end
else
	begin

	if(wr_datao)
		data_transmit <= do_peri[17:0];	// Sendedaten
	if(wr_ctrl)
		begin
		control	<= do_peri[7:0];		// Steuerdaten
		ir_cnt	<= do_peri[12:8];		// Bit Anzahl fuer Interrupt
		teiler	<= do_peri[15:13];		// Vorteiler
		end
	end

// Ueberpruefen, ob die aktuelle Uebertragung schon fertig ist
always@(posedge clk_peri)
if(reset)
   begin
 	ir_sig		<= 1'b0;
 	transmit_empty	<= 1'b1;
 	transmit_end	<= 1'b1;
	store_buf	<= 1'b0;
	load_buf	<= 1'b0;
   end
 else
    begin
	if( ~load_buf && load )
		transmit_end <= 1'b1;

 	//erst bei fallender Flanke store ist der gelesene Wert tatsaechlich im Register
	if(store_buf && ~store)
		begin
		ir_sig		<= 1'b1;
		transmit_empty 	<= 1'b1;
		end

	// Bei Zugriff auf die Datenregister Interrupt zuruecksetzen.
 	if((cs_datai) || (cs_datao) || ~en)
		ir_sig	<= 1'b0;

	// Nach Schreiben des Sendedatenregisters das Senden freigeben
	if(wr_datao)
		begin
		transmit_empty	<= 1'b0;
		transmit_end	<= 1'b0;
		end

	store_buf	<= store;
	load_buf	<= load;
    end


reg sclk_vor_posedge;

// MSB des Ausgangsschieberegisters wird uebertragen
//assign dout = sr_o[bpw-1];
assign dout = (ir_cnt < 19) ? sr_o[ir_cnt - 1] : 1'b0;

// Die Schieberegister fuer Empfang und Senden sind getrennt, sie werden
// zu unterschiedlichen Zeitpunkten weitergeschoben. Die Schieberegister
// koennen von den/in die entsprechenden Datenregister geladen werden.
always @ (posedge clk_spi)
if(reset)
	begin
	s_cnt	<= 5'b00000;
	sr_o	<= 0;
	sr_i	<= 0;
	data_recieve <= 0;
	end
else if (sclk_vor_posedge) begin
	// rausgehende Daten ins Schieberegister laden
	if(load)
		begin
		sr_o	<= data_transmit;
		s_cnt	<= 5'b00000;
		end

	// empfangene Daten aus dem Schieberegister speichern
	if(store)
		begin
		data_recieve <= sr_i;
		sr_i <= 0;
		end
	// Register schieben
	// Master -> Slave
	if(shift_o)
		begin
		s_cnt	<= s_cnt + 1;
//		sr_o	<= {sr_o[bpw-2:0],1'b0};
		sr_o	<= {sr_o[16:0],1'b0};
		end

	//Slave -> Master
	if(shift_i)
//		sr_i	<= {sr_i[bpw-2:0],din};
		sr_i	<= {sr_i[16:0],din};
	end


// Der SPI-Takt kann durch einen Vorteiler geteilt werden. Bei 60 MHz kann kein SPI_CLK
// kleiner 400 KHz eingestellt werden. Neuer Takt Eingang Notwendig!
always @ (posedge clk_spi)
if(reset)
	begin
	cnt	 <= 7'b000_0000;
	sclk_vor <= 0;
	sclk_vor_posedge <= 0;
	end
else if(en)
	if(~|cnt)
		begin
		case(teiler)
			3'b000: cnt <= 7'b0000000; //Teiler 2^0
			3'b001: cnt <= 7'b0000001; //Teiler 2^1
			3'b010: cnt <= 7'b0000011; //Teiler 2^2
			3'b011: cnt <= 7'b0000111; //Teiler 2^3
			3'b100: cnt <= 7'b0001111; //Teiler 2^4
			3'b101: cnt <= 7'b0011111; //Teiler 2^5
			3'b110: cnt <= 7'b0111111; //Teiler 2^6
			3'b111: cnt <= 7'b1111111; //Teiler 2^7
		endcase
		sclk_vor_posedge <= ~sclk_vor;
		sclk_vor <= ~sclk_vor;		// clk_spi / 2 * teiler_wert
		end
	else begin
		cnt <= cnt - 1;
		sclk_vor_posedge <= 0;
	end

// Zustandsdeklaration
parameter idle		= 2'b00;
parameter transmit_a	= 2'b01;
parameter transmit_b	= 2'b10;
parameter ss_pre	= 2'b11;

// kombinatorischer Block der FSM
// Die Taktperiode ist in die zwei Abschnitte transmit_a und transmit_b
// unterteilt, je nach Phasenlage und Polaritaet werden darin die Steuersignale
// zugewiesen. Der Modul halbiert die Frequenz von "sclk_vor"! --> sclk = clk_spi / (4 * teiler_wert)
always @ (posedge clk_spi)
if(reset)
	begin
	shift_i		<= 1'b0;
	shift_o		<= 1'b0;
	load		<= 1'b0;
	store		<= 1'b0;
	sclk_int	<= 1'b0;
	ss_mo		<= 15'h0000;
	state		<= idle;
	end
else if (sclk_vor_posedge)
	begin
	shift_i		<= 1'b0;
	shift_o		<= 1'b0;
	load		<= 1'b0;
	store		<= 1'b0;
	sclk_int	<= cpol;
	case(state)
		// Im Idlezustand wird nach abgeschlossenem Transfer gewartet und
		// bei uebertragungsbeginn slave select zugewiesen
		idle:
			if(transmit_end)
			    begin
				// ss_mo aus wenn kein en oder SS_ADDR = 0
				if(~en || (slave_addr == 0))
					ss_mo	<= 15'h0000;
				else
					ss_mo	<= wanted_ss;
				state	<= idle;
			    end
			else
			    begin
				// Daten in Schieberegister laden
				load	<= 1'b1;
				state	<= ss_pre;
			    end

		// Slave select ist schon angelegt und der
		// naechste Zustand anhand cpha vorbereitet.
		ss_pre:
				begin
				if(~cpha)
					begin
					state	<= transmit_b;
					shift_i	<= 1;
					end
				else
					begin
					state	<= transmit_a;
					sclk_int <= ~cpol;
					shift_i	<= 1;
					end
				end

		// Transmit_a ist der Abschnitt der Taktperiode, in der der Takt den
		// "aktiv"-Pegel hat (Halbe Periode von sclk)
		transmit_a:
			// Empfang beendet, Daten speichern

			// wenn Daten mit Flanke von aktiv zu idle uebernommen werden, werden
			// jetzt(idle->aktiv) die Ausgangsdaten angelegt
			if(cpha == 1'b1)
				begin
				shift_o <= 1'b1;
				state <= transmit_b;
				end

			// wenn Daten mit Flanke von idle zu aktiv uebernommen werden, werden
			// jetzt(idle->aktiv) die Daten am Eingang uebernommen
			else
//				if(s_cnt == ~ir_cnt)
				if(s_cnt == (ir_cnt - 1))
					begin
					store <= 1'b1;
					state <= idle;
					end
				else
					begin
					state <= transmit_b;
					shift_i <= 1'b1;
					end

		// Transmit_b ist der Abschnitt der Taktperiode, in der der Takt den
		// "idle"-Pegel hat (Halbe Periode von sclk)
		transmit_b:
			begin
			// Takt auf aktiven Pegel legen
			sclk_int <= ~cpol;

			// wenn Daten mit Flanke von idle zu aktiv uebernommen werden, werden
			// jetzt(aktiv->idle) die Daten am Ausgang angelegt
			if(cpha == 1'b0)
				begin
				shift_o	<= 1'b1;
				state	<= transmit_a;
				end

			// wenn Daten mit Flanke von aktiv zu idle uebernommen werden, werden
			// jetzt(aktiv->idle) die Daten am Eingang uebernommen
			else
			   begin
			   shift_i <= 1'b1;
//				if(s_cnt == ~ir_cnt)
				if(s_cnt == (ir_cnt - 1))
					begin
					state	<= idle;
					store	<= 1'b1;
					sclk_int <= cpol;
					shift_i	<= 1'b0;
					end
				else
					state	<= transmit_a;
			   end
			end
	endcase
	end

endmodule
