/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : spi_slave.v
// Autor : Steffen Kunze
// Modul : SPI_slave
//
// Beschreibung : Der SPI Slave empfaengt und sendet Daten mit dem vom Master
//		  vorgegebenen Takt, wenn Slave select angelegt wird(lowaktiv).
//
///////////////////////////////////////////////////////////////////////////////

`timescale      1ns / 10ps
`default_nettype none

module spi_slave(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	intr,
	miso,
	mosi,
	sclk,
	ss
);

// Die Parameter fuer die Adressen werden im Top-Modul z.B.: module_subsystem_0.v eingestellt.
parameter BASE_ADR	= 10'h0;

parameter CTRL_ADR	= 0;
parameter DATO_ADR	= 1;
parameter DATI_ADR	= 2;

parameter bpw		= 8;			//bits pro wort
parameter n		= 1;			//Anzahl Worte im Datenregister(1 bis 4)

parameter ALL_PORTS	= 3;			//Anzahl der Ports fuer den Dekoder
parameter DECODSIM_DEB	= "NO";			//DEBUG mode YES oder NO

input wire			clk_peri;
input wire	[17:0]		do_peri;		// Datenbus
output wire	[17:0]		di_peri;		// Datenbus
input wire	[9:0]		addr_peri;		// Adressbus
input wire			access_peri;		// chipselect peripherie
input wire			wr_peri;		// signalisiert Schreibzugriff Peripherie
input wire			reset;
output wire			intr;			// Interrupt flag

output wire			miso;			// Daten: Master In Slave Out
input wire			mosi;			// Daten: Master Out Slave In
input wire			sclk;			// Takt fuer Uebertragung
input wire			ss;			// Slave select

wire			miso_oen;		// Output enable Daten

reg	[bpw*n-1:0]	data_transmit;		// Datenregister, in das vom Bus geschrieben wird
reg	[bpw*n-1:0]	data_recieve;		// Datenregister, das vom Bus gelesen wird

//control- und statusregister
reg	[3:0]		control;
wire			en	= control[0];	// SPI Slave enable
wire			int_en	= control[1];	// interrupt enable
wire			cpol	= control[2];	// idle-Pegel des Taktsignals
wire			cpha 	= control[3];	// Phasenlage des Taktes
						// 0 = Flanke von aktiv zu idle 
						// 1 = Flanke von idle zu aktiv
reg			ir_sig;			// Interrupt register
reg	[2:0]		ir_cnt;			// Bit Anzahl fuer Interrupt
//, i_cnt;		// Wortzaehler fuer Interrupt
reg			done, done_t;		// zeigt Ende des Transfers an

//Register zur Steuerung von FSM, Schieberegister und Vorteiler							 
reg	[1:0]		state;			// Zustandsregister FSM
reg		load, store, shift_i, shift_o;	// Schieberegistersteuersignale
reg	[2:0]		s_cnt;			// Bitzaehler Schieberegister
reg	[bpw*n-1:0]	sr_i,sr_o;		// Schieberegister
wire			din, dout;		// Datenleitungen von/zu Schieberegister 


// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Status-, Daten- und Kommandoregister
wire			rd_ctrl;	// Lesen
wire			rd_datao;	// Lesen
wire			rd_datai;	// Lesen
wire			wr_ctrl;	// Schreiben
wire			wr_datao;	// Schreiben
wire			cs_datao;	// r/w
wire			cs_datai;	// r/w

wire			select;		// Modul wird angesprochen
wire	[ALL_PORTS-1:0]	reg_select;	// r/w ADR
wire	[ALL_PORTS-1:0]	reg_read;	// read ADR
wire	[ALL_PORTS-1:0]	reg_write;	// write ADR

// Addressdekoder fuer alle Geraeteregister (verwendet pselect.v und
// verzoegert die read Signale um einen Takt fuer die neue Pipeline) 
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	rd_ctrl		= reg_read[CTRL_ADR];		// Lesen
assign	rd_datao	= reg_read[DATO_ADR];		// Lesen
assign	rd_datai	= reg_read[DATI_ADR];		// Lesen
assign	wr_ctrl		= reg_write[CTRL_ADR];		// Schreiben
assign	wr_datao	= reg_write[DATO_ADR];		// Schreiben
assign	cs_datao	= reg_select[DATO_ADR];		// r/w
assign	cs_datai	= reg_select[DATI_ADR];		// r/w

// Dekodierung der Adressen ********************** END **************************************

assign di_peri[17:0]	= (rd_ctrl)?	{ir_cnt,ir_sig,done,4'b0,control} :
       			  (rd_datao)?	data_transmit :
			  (rd_datai)?	data_recieve : 18'b0;

assign intr = int_en? ir_sig : 1'b0;

// ***********************************************************
// Bei Verwendung einer buffer instance sind die Einstellungen
// in der UCF Datei fuer das Pin nicht wirksam!
// 
// Spartan-3 input + output buffer
wire			cs_slave;
wire			sclk_inb;
//
//IBUF	si	(.O (din),
//		 .I (mosi));
//IBUF	scs	(.O (sclk_inb),
//		 .I (sclk));
//IBUF	css	(.O (cs_slave),
//		 .I (ss));
//OBUFT	so	(.O (miso),
//		 .I (dout),
//		 .T (~miso_oen));
// ***********************************************************
assign	din		= mosi;
assign	sclk_inb	= sclk;
assign	cs_slave	= ss;
assign	miso		= (~cs_slave & en)?	dout: 1'bz;

//assign miso_oen	= ~cs_slave;


always@(posedge clk_peri)
if(reset) 
   begin
     ir_sig	<= 1'b0;
     done	<= 1'b0;
   end
 else
    begin  
	if(store || (wr_ctrl&&do_peri[0]))
		done	<= 1'b1;
	if(store)
		ir_sig	<= 1'b1;
			
	// Bei Zugriff auf die Datenregister Interrupt zuruecksetzen.
	if((cs_datai) || (cs_datao) || ~en)		   
		ir_sig	<= 1'b0;
	
	// Nach Schreiben des Sendedatenregisters mit Senden beginnen	   
	if((wr_datao) || ~en)		   
		done	<= 1'b0;   
   end
	
// MSB des Ausgangsschieberegisters wird uebertragen
assign dout = sr_o[bpw*n-1];	
// Die Schieberegister fuer Empfang und Senden sind getrennt, sie werden
// zu unterschiedlichen Zeitpunkten weitergeschoben. Die Schieberegister
// koennen von den/in die entsprechenden Datenregister geladen werden.
always @ (posedge clk_peri)
if(reset)
	begin
	s_cnt	<= 3'b000;
	sr_o	<= 0;
	sr_i	<= 0;
	data_recieve <= 0;
	end
else
	begin
	// rausgehende Daten ins Schieberegister laden
	if(load)
		begin
		sr_o	<= data_transmit;
		s_cnt	<= 3'b000;
		end

	// empfangene Daten aus dem Schieberegister speichern
	if(store)
		data_recieve <= sr_i;

	// Register schieben
	if(shift_o)
		begin
		s_cnt	<= s_cnt + 1;
		sr_o	<= {sr_o[bpw*n-2:0],1'b0};
		end
	if(shift_i)
		sr_i	<= {sr_i[bpw*n-2:0],din};
	end


// In diesem Block werden Daten vom Bus in die Register uebernommen.
always @ (posedge clk_peri)
if(reset)
	begin
	control		<= 8'b00000000;
	data_transmit	<= 0;
	end
else
	begin
	if(wr_datao)
		data_transmit <= do_peri[bpw-1:0];	// Sendedaten
	if(wr_ctrl)
		begin
		control	<= do_peri[3:0];		// Steuerdaten
		ir_cnt	<= do_peri[12:10];		// Wortzahl fuer Interrupt
		end
	end


// Zustandsdeklaration	 
parameter idle		= 2'b00;
parameter transmit_a	= 2'b01;
parameter transmit_b	= 2'b10;

reg	[4:0]	fang;
reg	[2:0]	ncnt, i;
reg		sclk_slave;

always @ (posedge clk_peri)
if(reset)
	begin
	sclk_slave	 = 0;
	fang		<= 0;
	ncnt		 = 0;
	end
else
	begin
	ncnt = 0;
	for (i=0;i < 5; i = i+1)			
			if(fang[i] == 0)
				ncnt = ncnt + 1;
	
	if(ncnt < 3)
		sclk_slave = 1;
	else 
		sclk_slave = 0;
		
	fang <= {fang[3:0],sclk_inb};
	
	end
	

// kombinatorischer Block der FSM
// Die Taktperiode ist in die zwei Abschnitte transmit_a und transmit_b 
// unterteilt, je nach Phasenlage und Polaritaet werden darin die Steuersignale
// zugewiesen
always @ (posedge clk_peri)
if(reset)
	begin
	shift_i	= 1'b0;
	shift_o	= 1'b0;
	load	= 1'b0;
	state	= idle;
	store	= 1'b0; 
	end
else
	begin
	load	= 1'b0;
	shift_i	= 1'b0;
	shift_o	= 1'b0;
	store	= 1'b0;
	//next_state = idle;	 	
	case(state)
		// Im Idlezustand wird nach abgeschlossenem Transfer gewartet und
		// bei low-Pegel am slave select die Uebertragung begonnen
		idle:	
			//if(done)
			//	next_state = idle;
			//else
			   if(~cs_slave && en)		
					begin		
					if(cpha && (sclk_slave == ~cpol))
					   state = transmit_b;
					else if (!cpha)
					   state = transmit_a;
					else 
						state = idle;
					load = 1'b1;
					end 
				else
					state = idle;
					
		// Transmit_a ist der Abschnitt der Taktperiode, in der auf den 
		// Uebergang von idle zu aktiv gewartet wird			 				
		transmit_a:
			if (!en)
				state = idle;
			// Empfang beendet, Daten speichern
			else if( (s_cnt == ~ir_cnt) && cpha)
				begin
				store = 1'b1;
				state = idle;
				end
			else
			// Beim idle->aktiv-Uebergang Daten uebernehmen(cpha = 0)
			// oder an Ausgang legen(cpha = 1)
			if(sclk_slave == ~cpol)
				begin 
				if(cpha == 1'b1)
					shift_o = 1'b1;
				else
				   shift_i = 1'b1;
				state = transmit_b;
				end			
			else
			   state = transmit_a;
		
		// Transmit_b ist der Abschnitt der Taktperiode, in der auf den 
		// Uebergang von aktiv zu idle gewartet wird				
		transmit_b:
			if(!en)
				state = idle;
			// Beim aktiv->idle-Uebergang Daten uebernehmen(cpha = 1)
			// oder an Ausgang legen(cpha = 0)
			else if( (s_cnt == ~ir_cnt) && !cpha)
				begin
				store = 1'b1;
				state = idle;
				end
			else if(sclk_slave == cpol)
				begin
				if(cpha == 1'b0)
					shift_o = 1'b1;
			   else 
			      shift_i = 1'b1;
				state = transmit_a;
				end
			else			    
				state = transmit_b;
			
	endcase
	end

endmodule

