/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : timer.v
// Autor : Steffen Kunze
// Modul : timer
//
// Beschreibung : Der Timer beinhaltet einen von 2 bis 256 in 8 Stufen
//		  konfigurierbaren Vorteiler, der auch ausgeschalten werden
//		  kann. Der Takteingang dieses Vorteiler kann beliebig mit
//		  dem Systemtakt, einem DCM Ausgang oder dem Ausgang enes
//		  zweiten Timers verbunden werden. Der Ausgang des Vorteiler
//                treibt den Takt eines 18bit-Zaehlers, der bis zu einem
//                programmiebaren Wert hochzaehlt und dann wieder neu startet.
//                Zusammen mit diesem Modul koennen RTI-, Watchdog-, Capture-,
//		  Compare-  und Pulse Accumulator-Module in beliebiger Anzahl
//                generiert werden.
///////////////////////////////////////////////////////////////////////////////

`timescale      1ns / 10ps
`default_nettype none

module timer(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	clk_i,
    hw_enable_i,
	reset,
	timer_out
);

parameter BASE_ADR		= 10'h40;

parameter TIMER_CTRL_ADR	= 0;
parameter TIMER_DATA_ADR	= 1;
parameter TIMER_VALUE_ADR	= 2;
parameter USE_EXT_HW_ENABLE = "FALSE";

parameter ALL_PORTS		= 3;
parameter DECODSIM_DEB		= "NO";


// inputs & outputs des Systembus
input	wire				clk_peri;
input	wire	[17:0]			do_peri;	// Datenbus von CPU
output	wire	[17:0]			di_peri;	// Datenbus zur CPU
input	wire	[9:0]			addr_peri;	// Adressbus
input	wire				access_peri;	// chipselect peripherie
input	wire				wr_peri;	// signalisiert Schreibzugriff Peripherie
input	wire				clk_i;		// Takt vom Timer-DCM
input   wire                hw_enable_i;    //External enable signal   
input	wire				reset;

// Signale des Modul
output	wire	[17:0]			timer_out;	// Timerwert als Input fuer die anderen Module


reg     [17:0]          count_d;  
reg		[17:0]			count_q;		// 18 bit-Zaehler
reg		[17:0]			count_syn;	// 18 bit-Zaehler mit clk_peri synchronisiert
reg		[17:0]			count_dat;	// Zaehlerendwert
reg		[7:0]			cnt_d; 		// Vorteilerzaehler
reg		[7:0]			cnt_q; 		// Vorteilerzaehler
reg		[2:0]			teiler_sel;	// Vorteiler
reg					teiler_en;	// Vorteiler enable
reg					en;	 	// Timer enable
reg		[17:0]			count_buf;
reg					set_cnt_sig, count_set_q, count_set_d;
wire                hw_enable;      //hardware enable   


// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					cs_ti_ctrl_r;	// Lesen
wire					cs_ti_data_r;	// Lesen
wire					cs_ti_value_r;	// Lesen
wire					cs_ti_ctrl_w;	// Schreiben
wire					cs_ti_data_w;	// Schreiben
wire					cs_ti_value_w;	// Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	cs_ti_ctrl_r	= reg_read[TIMER_CTRL_ADR];	// Lesen
assign	cs_ti_data_r	= reg_read[TIMER_DATA_ADR];	// Lesen
assign	cs_ti_value_r	= reg_read[TIMER_VALUE_ADR];	// Lesen
assign	cs_ti_ctrl_w	= reg_write[TIMER_CTRL_ADR];	// Schreiben
assign	cs_ti_data_w	= reg_write[TIMER_DATA_ADR];	// Schreiben
assign	cs_ti_value_w	= reg_write[TIMER_VALUE_ADR];	// Schreiben

// Dekodierung der Adressen ********************** END **************************************


assign	di_peri[17:0]	= (cs_ti_ctrl_r)?	{teiler_sel, teiler_en, en}:
			  (cs_ti_data_r)?	count_dat:
			  (cs_ti_value_r)?	count_syn:	18'h0;

assign	timer_out = count_q;

generate
    if (USE_EXT_HW_ENABLE == "TRUE")
        assign hw_enable = hw_enable_i;
    else
        assign hw_enable = 1'b1;
endgenerate

// In diesem Block werden Daten vom Bus in die Register uebernommen.
always @ (posedge clk_peri)
if(reset)
	begin
	en		<=  1'b0;
	teiler_en	<=  1'b0;
	teiler_sel	<=  3'b0;
	count_dat	<= 18'b0;
	count_syn	<= 18'b0;
	set_cnt_sig	<=  1'b0;
	end
else
    begin
        count_syn	<= count_q;
        if(cs_ti_ctrl_w)
            begin
                en		<= do_peri[0];				// Timer enable
                teiler_en	<= do_peri[1];				// Vorteiler enable
                teiler_sel	<= do_peri[4:2];			// Vorteilerwert
            end
        if(cs_ti_data_w)
            count_dat	<= do_peri;				// Zaehlerendwert
        if(cs_ti_value_w)
            begin
                count_buf	<= do_peri;				// aktueller Zaehlerwert
                set_cnt_sig	<= 1'b1;
            end
        if (count_set_q == 1'b1)
            set_cnt_sig	<= 1'b0;
    end

// 18bit-Timer, der im Takt der DCM zaehlt, geteilt durch Vorteiler.
// Wird bei Erreichen eines einprogrammierten Endwertes zurueckgesetzt.
always @(posedge clk_i or posedge reset) begin
    if(reset) begin
        count_q		<= 18'b0;
        cnt_q	    <=  8'b0;
        count_set_q	<= 1'b0;
    end
    else begin
        count_q     <= count_d;
        cnt_q       <= cnt_d;
        count_set_q <= count_set_d;
    end
end
    
    
always @(*) begin
    count_d     = count_q;
    cnt_d       = cnt_q;
    count_set_d = 1'b0;
    
    if (set_cnt_sig == 1'b1) begin
        count_d	    = count_buf;
        count_set_d	= 1'b1;
    end
        
	if(en & hw_enable) begin
		if(teiler_en)
			// mit Vorteiler:
			// Der Vorteilerzaehler vom Vorteilerwert runter zählen,
			// dann der Timer erhoeht. Und so weiter.
			if(~|cnt_q) begin
				case(teiler_sel)
					3'b000: cnt_d = 8'b00000001; //Teiler 2^1
					3'b001: cnt_d = 8'b00000011; //Teiler 2^2
					3'b010: cnt_d = 8'b00000111; //Teiler 2^3
					3'b011: cnt_d = 8'b00001111; //Teiler 2^4
					3'b100: cnt_d = 8'b00011111; //Teiler 2^5
					3'b101: cnt_d = 8'b00111111; //Teiler 2^6
					3'b110: cnt_d = 8'b01111111; //Teiler 2^7
					3'b111: cnt_d = 8'b11111111; //Teiler 2^8
				endcase
            end else
				cnt_d = cnt_q - 1;


        if(~|cnt_q)
            if(count_q == count_dat)
				count_d = 18'b00_0000_0000_0000_0000;
			else
                if(set_cnt_sig == 1'b1)
                    count_d = count_buf +1;
                else
                    count_d = count_q + 1;
    end
end


endmodule
