/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : timer_capture.v
// Autor : Steffen Kunze
// Modul : capture
//
// Beschreibung : Das Captureregister speichert den Timerwert bei Eintreten
//		  eines eingestellten Ereignisses: bestimmter Pegel,
// 		  bestimmte Flanke, irgendeine Flanke.
//
///////////////////////////////////////////////////////////////////////////////
`default_nettype none

module timer_capture(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	intr,
	timer_data,
	in
	);

// Der Parameter fuer die Adresse wird in der
// system/subsystem_0/hardware.h durch jConfig eingestellt.
parameter BASE_ADR	= 10'h2;

parameter CTRL_ADR	= 0;
parameter DATA_ADR	= 1;

parameter ALL_PORTS	= 2;
parameter DECODSIM_DEB	= "NO";


// inputs & outputs des Systembus
input	wire		clk_peri;
input	wire	[17:0]	do_peri;			// Datenbus
output	wire	[17:0]	di_peri;			// Datenbus
input	wire	[9:0]	addr_peri;			// Adressbus
input	wire		access_peri;			// chipselect peripherie
input	wire		wr_peri;			// signalisiert Schreibzugriff Peripherie
input	wire		reset;
output	wire		intr;				// interrupt flag
// Signale des Modul
input	wire	[17:0]	timer_data;			// Zaehlerstand des Timers
input	wire		in;				// IO-pin


reg		[17:0]	cap_reg;			// 18 bit Capture-Register
reg		[4:0]	control;
wire		[1:0]	mode_sel	= control[4:3];	// 00 = aus; 01 = Flanke; 10 = Pegel; 11 = beide Flanken
wire			pf_sel		= control[2];	// Pegel/Flanke einstellen:
							// 1 = rising edge bzw. high; 0 = falling edge bzw. low
wire			int_en		= control[1];	// interrupt enable
wire			en 		= control[0];	// capture enable
reg			ir_sig; 			// interrupt register

assign	intr	= int_en?ir_sig:1'b0;

reg			zw;	// Eingangspuffer


// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire				cs_tca_ctrl_r;	// Lesen
wire				cs_tca_data_r;	// Lesen
wire				cs_tca_ctrl_w;	// Schreiben
wire				cs_tca_data_w;	// Schreiben

wire				select;
wire		[ALL_PORTS-1:0]	reg_select;
wire		[ALL_PORTS-1:0]	reg_read;
wire		[ALL_PORTS-1:0]	reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	cs_tca_ctrl_r	= reg_read[CTRL_ADR];	// Lesen
assign	cs_tca_data_r	= reg_read[DATA_ADR];	// Lesen
assign	cs_tca_ctrl_w	= reg_write[CTRL_ADR];	// Schreiben
assign	cs_tca_data_w	= reg_write[DATA_ADR];	// Schreiben

// Dekodierung der Adressen ********************** END **************************************



// Wenn Adresse des Datenregister anliegt, wird selbiges an den Bus gelegt.
assign	di_peri[17:0]	= (cs_tca_data_r)?	cap_reg:
			  (cs_tca_ctrl_r)?	control:	18'b0;


// In diesem Block werden Daten vom Bus in die Register uebernommen und
// entsprechend der Programmierung Interrupts ausgeloest.
always @ (posedge clk_peri)
if(reset)
	begin
	control	<= 5'b00000;
	cap_reg	<= 18'h00000;
	ir_sig	<= 1'b0;
	zw	<= 1'b0;
	end
else
	begin
	// Bei Lesezugriff auf das Datenregister oder Schreibzugriff auf das
	// Steuerregister wird der Interrupt zurueckgesetzt
	if(cs_tca_data_r || cs_tca_ctrl_w )
		ir_sig	<= 1'b0;
	// Steuerregister schreiben
	if(cs_tca_ctrl_w)
		control	<= do_peri[4:0];

	// Auswertung des Eingangssignals in verschiedenen Betriebsmodi
	else if(en && ~ir_sig)
		case(mode_sel)
			// aus
			2'b00:begin end

			//entweder steigende oder fallende Flanke erkennen
			2'b01:if(pf_sel && ~zw && in || ~pf_sel && zw && ~in)
					begin
					cap_reg		<= timer_data;
					ir_sig		<= 1'b1;
					control[0]	<= 1'b0;
					end

			//Pegel erkennen
			2'b10:if(pf_sel == in)
					begin
					cap_reg		<= timer_data;
					ir_sig		<= 1'b1;
					control[0]	<= 1'b0;
					end

			//allgemein Flanke erkennen
			2'b11:if(zw != in)
					begin
					cap_reg		<= timer_data;
					ir_sig		<= 1'b1;
					control[0]	<= 1'b0;
					end
		endcase

	// Wert des Eingangssignals zwischenspeichern fuer Flankendetektion
	zw	<= in;
	end

endmodule
