/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : timer_pulseacc.v
// Autor : Steffen Kunze
// Modul : pulse_acc
//
// Beschreibung : Der Pulse Accumulator zaehlt Impulse am Eingang pin oder an
//		  rti_in, bis zu einem Impuls am Eingang.pin.
//
///////////////////////////////////////////////////////////////////////////////

`default_nettype none

module timer_pulseacc(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	rti_in,
	reset,
	pin
	);

// Der Parameter fuer die Adresse wird
// in der system.v durch jConfig eingestellt.
parameter BASE_ADR	= 10'h0;

parameter CTRL_ADR	= 0;
parameter DATA_ADR	= 1;

parameter ALL_PORTS	= 2;
parameter DECODSIM_DEB	= "NO";


// inputs und outputs des Systembus
input	wire		clk_peri;
input	wire	[17:0]	do_peri;	// Datenbus
output	wire	[17:0]	di_peri;	// Zaehlerstand
input	wire	[9:0]	addr_peri;	// Adressbus
input	wire		access_peri;	// chipselect peripherie
input	wire		wr_peri;	// signalisiert Schreibzugriff Peripherie
input	wire		rti_in;		// Eingang RTI-Impulse (clk_i)
input	wire		reset;

// Signale des Modul
input	wire		pin;		// Eingangspin


reg		[17:0]	cnt;		// 18 bit-Zaehlerstand 
reg			en;		// Zaehler enable
reg			mode_sel;	// 0 = Impulse am Eingang zaehlen
					// 1 = RTI bis Impuls am Eingang zaehlen

reg			in_o, rti_o;	// Zwischenspeicher fuer Eingaenge rti und pin
reg			warte_sig;


// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire				cs_tpa_ctrl_rw;	// Lesen oder Schreiben
wire				cs_tpa_data_rw;	// Lesen oder Schreiben
wire				cs_tpa_ctrl_r;	// Lesen
wire				cs_tpa_data_r;	// Lesen
wire				cs_tpa_ctrl_w;	// Schreiben
wire				cs_tpa_data_w;	// Schreiben

wire				select;
wire		[ALL_PORTS-1:0]	reg_select;
wire		[ALL_PORTS-1:0]	reg_read;
wire		[ALL_PORTS-1:0]	reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	cs_tpa_ctrl_rw	= reg_select[CTRL_ADR];	// Lesen oder Schreiben
assign	cs_tpa_data_rw	= reg_select[DATA_ADR];	// Lesen oder Schreiben
assign	cs_tpa_ctrl_r	= reg_read[CTRL_ADR];	// Lesen
assign	cs_tpa_data_r	= reg_read[DATA_ADR];	// Lesen
assign	cs_tpa_ctrl_w	= reg_write[CTRL_ADR];	// Schreiben
assign	cs_tpa_data_w	= reg_write[DATA_ADR];	// Schreiben

// Dekodierung der Adressen ********************** END **************************************


// Wenn Adresse des Datenregister anliegt, wird selbiges an den Bus gelegt.
assign di_peri[17:0]	= (cs_tpa_data_r)?	cnt:
			  (cs_tpa_ctrl_r)?	{mode_sel,en}:	18'b0;

// In diesem Block werden Daten vom Bus in die Register uebernommen.
always @ (posedge clk_peri)
if(reset)
	begin
	 mode_sel	<= 1'b0;
	 in_o		<= 1'b0;
	 rti_o		<= 1'b0;
	 en		<= 1'b0;
	end
else
	begin
	 if(cs_tpa_ctrl_w)
		begin
		 mode_sel	<= do_peri[1];	// Betriebsmodus
		 en		<= do_peri[0];	// enable
		end

	 // Eingangssignale zwischenspeichern fuer Flankendetektion
	 in_o	<= pin;
	 rti_o	<= rti_in;
	end


// Dieser Block beherbergt den Zaehler. Eingangsimpulse werden fortlaufend
// gezaehlt, beim Zaehlen von RTI-Impulsen wird nach Auftreten eines Impulses 
// am Eingang gewartet, bis entweder das Datenregister gelesen wurde oder das
// Steuerregister geschrieben wurde.
always @ (posedge clk_peri)

if(reset)
	begin
	 cnt		= 18'h00000;
	 warte_sig	= 1'b0;
	end
else
	begin
	if(en && ~warte_sig)
		begin
		 // Warten, wenn Impuls am Eingang waehrend RTI-Modus
		 if(mode_sel && ~in_o && pin)
			warte_sig = 1'b1;

		 // Impulse am Eingang oder RTI-Impulse zaehlen
		 if(mode_sel && ~rti_o && rti_in || ~mode_sel && ~in_o && pin)
			cnt = cnt + 1;

		 // Zaehlerueberlauf
		 //if(cnt == 18'b11_1111_1111_1111_1111)
		 //	cnt = 18'h00000; 
		end

		// Zaehler zuruecksetzen
	if(cs_tpa_data_rw && mode_sel && warte_sig || cs_tpa_ctrl_rw)
		begin
		 warte_sig	= 1'b0;
		 cnt		= 18'h00000;
		end
	end

endmodule
