/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : timer_rti.v
// Autor : Steffen Kunze
// Modul : rti
//
// Beschreibung : Das RTI-Modul erzeugt in einstellbaren Abstaenden
//                Interruptsignale. Diese Impulse werden auch an einem Ausgang
//		  zur Verfuegung gestellt (fuer die Pulse Accumulator-Module).
//
///////////////////////////////////////////////////////////////////////////////
`default_nettype none

module timer_rti(	
	clk_peri,
 	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	clk_i,
	reset,
	intr,
	rti_out
	);

// Der Parameter fuer die Adresse wird
// in der system.v durch jConfig eingestellt.
parameter BASE_ADR	= 10'hD;
parameter CTRL_ADR	= 0;

parameter ALL_PORTS	= 2;
parameter DECODSIM_DEB	= "NO";

parameter SAMECLK = "NO";

// inputs & outputs des Systembus
input	wire				clk_peri;	// SpartanMC-Takt   
input	wire	[17:0]			do_peri;	// Datenbus
output	wire	[17:0]			di_peri;	// Datenbus
input	wire	[9:0]			addr_peri;	// Adressbus
input	wire				access_peri;	// chipselect peripherie
input	wire				wr_peri;	// signalisiert Schreibzugriff Peripherie
input	wire				clk_i;		// Takt fuer Vorteiler
input	wire				reset;
output	wire	  			intr;		// Interrupt flag

// Signale des Modul
output	wire				rti_out;	// Ausgangssignal

reg		[14:0]			cnt; 		// Vorteilerzaehler
reg		[3:0]			teiler_sel;	// Vorteiler
reg					int_en;		// interrupt enable
reg					en;		// RTI enable
reg					c_reg;		// clear intr
reg					rti_reg;
reg					int_sig;

// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					cs_rti_ctrl_r;	// Lesen
wire					cs_rti_ctrl_w;	// Schreiben
wire					cs_rti_ctrl;	// Lesen + Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	cs_rti_ctrl_r	= reg_read[CTRL_ADR];	// Lesen
assign	cs_rti_ctrl_w	= reg_write[CTRL_ADR];	// Schreiben
assign	cs_rti_ctrl	= reg_select[CTRL_ADR];	// Lesen + Schreiben

// Dekodierung der Adressen ********************** END **************************************

assign	rti_out	= en		?	c_reg:		1'b0;
assign	intr	= int_en	?	int_sig:	1'b0;

assign	di_peri	= cs_rti_ctrl_r	?	{12'b0, teiler_sel, int_en, en}:	18'b0;
//				?	{1'b0, rti_reg, cnt[7:0], int_sig, c_reg, teiler_sel, int_en, en}: 18'b0;

// In diesem Block werden Daten vom Bus in die Register uebernommen.
always @ (posedge clk_peri)
if(reset)
	begin
	teiler_sel	<= 4'b0000;
	en		<= 1'b0;
	int_en		<= 1'b0;
	end
else
   begin
	if(cs_rti_ctrl_w)
		begin
		teiler_sel	<= do_peri[5:2];	// Vorteiler
		int_en		<= do_peri[1];		// interrupt enable
		en		<= do_peri[0];		// enable
		end
end


// Der Block erzeugt beim Erreichen des Vorteilerwertes einen Impuls, der
// fuer die Dauer einer Taktperiode high ist.
always @ (posedge clk_i or posedge reset)
if(reset)
	begin
	rti_reg	<= 1'b0;
	cnt	<= 15'h0000;
	end
else
	if(en)
		if(~|cnt)
		   begin
		   case(teiler_sel)
		   4'b0000: cnt <= 15'h0000; //Teiler 2^0
		   4'b0001: cnt <= 15'h0001; //Teiler 2^1
		   4'b0010: cnt <= 15'h0003; //Teiler 2^2
		   4'b0011: cnt <= 15'h0007; //Teiler 2^3
		   4'b0100: cnt <= 15'h000f; //Teiler 2^4
		   4'b0101: cnt <= 15'h001f; //Teiler 2^5
		   4'b0110: cnt <= 15'h003f; //Teiler 2^6
		   4'b0111: cnt <= 15'h007f; //Teiler 2^7
		   4'b1000: cnt <= 15'h00ff; //Teiler 2^8
		   4'b1001: cnt <= 15'h01ff; //Teiler 2^9
		   4'b1010: cnt <= 15'h03ff; //Teiler 2^10
		   4'b1011: cnt <= 15'h07ff; //Teiler 2^11
		   4'b1100: cnt <= 15'h0fff; //Teiler 2^12
		   4'b1101: cnt <= 15'h1fff; //Teiler 2^13
		   4'b1110: cnt <= 15'h3fff; //Teiler 2^14
		   4'b1111: cnt <= 15'h7fff; //Teiler 2^15
		   endcase
		   rti_reg	<= 1'b1;
		   end
		else
		   begin
		   cnt		<= cnt - 1;
		   rti_reg	<= 1'b0;
		   end

// Hier wird die Flanke des erzeugten Impulses erkannt und daraufhin
// ein Impuls erzeugt, der eine SpartanMC-Taktperiode lang high ist und
// es wird ein Interrupt erzeugt.
always@(posedge clk_peri)
if(reset)
   begin
   c_reg	<= 1'b0;
   int_sig	<= 1'b0;
   end
else
   begin
   	if (SAMECLK=="YES") begin
		if ((teiler_sel == 0) && en)
			c_reg	<= 1;
		else
			c_reg	<= rti_reg;
   		if(cs_rti_ctrl)
	      		int_sig	<= 1'b0;
	   	else // nur an der Flake von rti_reg oder clk_i
	      		if (en && !c_reg && ((rti_reg && (teiler_sel > 0)) || (teiler_sel == 0)))
				int_sig	<= 1'b1;
   	end else begin
		if ((teiler_sel == 0) && en)
			c_reg	<= clk_i;
		else
			c_reg	<= rti_reg;
		// Registerzugriff setzt Interrupt zurueck
	   if(cs_rti_ctrl)
	      int_sig	<= 1'b0;
	   else // nur an der Flake von rti_reg oder clk_i
	      if (en && !c_reg && ((rti_reg && (teiler_sel > 0)) || (clk_i && (teiler_sel == 0))))
		int_sig	<= 1'b1;

	end
   end
endmodule
