/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt : SpartanMC
// Dateiname : timer_watchdog.v
// Autor : Steffen Kunze
// Modul : watchdog
//
// Beschreibung : Der Watchdog Timer gibt ein Alarmsignal, wenn sein Zaehler
//		  einen bestimmten, voreinstellbaren Wert erreicht, bevor das
//                in der system.v konfigurierte Codewort in das Pruefregister
//                geschrieben wurde.
//
///////////////////////////////////////////////////////////////////////////////

`default_nettype none

module timer_watchdog(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	wr_peri,
	access_peri,
	clk_1,
	reset,
	reset_lowactiv,
	clk_x,
	alarm
	);

// Die Parameter fuer die Adresse und den WDT_CODE
// werden in der system.v durch jConfig eingestellt.
parameter BASE_ADR	= 10'h0;
parameter WDT_CODE	= 18'h0;

parameter CTRL_ADR	= 0;
parameter DAT_ADR	= 1;
parameter CHK_ADR	= 2;

parameter ALL_PORTS	= 3;
parameter DECODSIM_DEB	= "NO";

// inputs & outputs des Systembus
input	wire				clk_peri;	// SpartanMC-Takt
input	wire	[17:0]			do_peri;	// Datenbus
output	wire	[17:0]			di_peri;	// Datenbus
input	wire	[9:0]			addr_peri;	// Adressbus
input	wire				wr_peri;	// signalisiert Schreibzugriff Peripherie
input	wire				access_peri;	// chipselect peripherie
input	wire				clk_1;		// Timer-"Takt" fuer Watchdog Zaehler.
input	wire				reset;
input	wire				reset_lowactiv;	// Externer RESET Taster oder IC.

// Signale des Modul
input	wire				clk_x;		// Takt, der bei RESET(ALARM) nicht angehalten wird!
output	wire				alarm;		// Alarmsignal

	
reg		[17:0]			count;		// 18 bit-Zaehler
reg		[17:0]			count_syn;	// 18 bit-Zaehler mit clk_peri synchronisiert
reg		[7:0]			cnt;		// Vorteilerzaehler
reg		[17:0]			check;		// Pruefregister
reg					del_check;
reg		[17:0]			comp;		// Endwert des Zaehlers

reg		[2:0]			teiler;		// Vorteilerwert
reg			 		teiler_en;	// Teiler ein/aus, 0=aus
reg 					en;		// Watchdog enable, 1=enable
reg			 		alarm_r;	// zeigt Erreichen des Endwerts an 
	
reg		[15:0]			alarm_a; 	// Puffer fuer Alarmsignal
wire					alarm1;
assign	alarm1	= |{alarm_a,alarm_r};
reg					alarm_m;	// Merken fuer Reset Behandlung. Bit wird nur
							// beim Schreiben auf CTRL_ADR auf 0 gesetzt.
assign	alarm	= alarm1 & alarm_m;


// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					rd_ctrl;	// Lesen
wire					rd_dat;		// Lesen
wire					rd_chk;		// Lesen
wire					wr_ctrl;	// Schreiben
wire					wr_dat;		// Schreiben
wire					wr_chk;		// Schreiben

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	rd_ctrl		= reg_read[CTRL_ADR];		// Lesen
assign	rd_dat		= reg_read[DAT_ADR];		// Lesen
assign	rd_chk		= reg_read[CHK_ADR];		// Lesen
assign	wr_ctrl		= reg_write[CTRL_ADR];		// Schreiben
assign	wr_dat		= reg_write[DAT_ADR];		// Schreiben
assign	wr_chk		= reg_write[CHK_ADR];		// Schreiben

// Dekodierung der Adressen ********************** END **************************************


// Wenn Adresse eines Datenregister anliegt, wird selbiges an den Bus gelegt.
assign	di_peri[17:0]	= (rd_chk)?	count_syn:
			  (rd_dat)?	comp:
			  (rd_ctrl)?	{alarm_m,teiler,teiler_en,en}:	18'b0;


// In diesem Block werden Daten vom Bus in die Register uebernommen.
always @(posedge clk_peri)
begin
if(reset)
	begin
	en		<= 1'b0;
	count_syn	<= 18'h00000;
	check		<= 18'h00000;
	comp		<= 18'h00000;
	teiler_en	<= 1'b0;
	teiler		<= 3'b000;
	if(!reset_lowactiv)
		alarm_m	<= 1'b0;
	end
else
	begin
	count_syn	<= count;
	if(wr_chk)	// Pruefregister
		check		<= do_peri;
	if(del_check)		// Pruefregister zuruecksetzen nach Ruecksetzen des Timers
		check		<= 18'b00_0000_0000_0000_0000;
	if(wr_dat)	// Endwert
		comp <= do_peri;
	if(wr_ctrl)
		begin
		en		<= do_peri[0];		// watchdog enable
		teiler_en	<= do_peri[1];		// Teiler enable
		teiler		<= do_peri[4:2];	// Teilerwert
		if(reset_lowactiv && !alarm1)		// kein Reset && kein Alarm
			alarm_m		<= 1'b0;	// letzten Alarm loeschen.
		end
	if(alarm1 && reset_lowactiv)			// kein Reset && Alarm
		alarm_m	<= 1'b1;
	end
end


// Wenn ein Alarm auftritt, wird dieser 16 Systemtakte lang gehalten,
// u.a. damit die DCM nach dem Reset wieder korrekt synchronisiert sind.
always@(posedge clk_x)
	alarm_a	<= {alarm_a[14:0],alarm_r};

// In diesem Block zaehlt der Zaehler mit dem Timertakt, bis der Code geschrieben
// wird oder Endwert erreicht ist und Alarm gegeben wird.
always @(posedge clk_1 or posedge reset)
if(reset)
	begin
	count		= 18'h00000;
	cnt		= 8'b0000_0000;
	del_check	= 1'b0;
	alarm_r		= 1'b0;
	end
else
	if(en)
		begin
		// Default fuer Alarm ist 0
		if(alarm)
			alarm_r = 1'b0;

		// Korrekter Wert geschrieben, Timer zurueckgesetzt
		if(check == WDT_CODE)
			begin
			count		= 18'h00000;
			del_check	= 1'b1;
			end
		else
			del_check	= 1'b0;

		// Endwert erreicht, Alarm!
		if(count == comp)
			begin
			alarm_r	= 1'b1;
			count	= 18'h00000;
			cnt	= 8'b0000_0000;
			end
		else
			begin
			if(teiler_en)
			// mit Vorteiler:
			// Der Vorteilerzaehler wird bis zu 0 herunter gezaehlt, dann
			// der Timer erhoeht und der Vorteiler wieder auf den Startwert
			// aus der Tabelle gesetzt..
				if(~|cnt)
					begin
					count = count +1;
					case(teiler)
						3'b000: cnt = 8'b00000001; //Teiler 2^1
						3'b001: cnt = 8'b00000011; //Teiler 2^2
						3'b010: cnt = 8'b00000111; //Teiler 2^3
						3'b011: cnt = 8'b00001111; //Teiler 2^4
						3'b100: cnt = 8'b00011111; //Teiler 2^5
						3'b101: cnt = 8'b00111111; //Teiler 2^6
						3'b110: cnt = 8'b01111111; //Teiler 2^7
						3'b111: cnt = 8'b11111111; //Teiler 2^8
					endcase
					end
				else
					cnt = cnt - 1;

			else
			//ohne Vorteiler
				count = count +1;
			end
		end

	else
	begin		// Falls enable = 0
	count	= 18'h00000;
	alarm_r = 1'b0;
	cnt	= 8'b0000_0000;
	end

endmodule
