/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Create Date:    17:40:44 01/22/2007 
// Module Name:    ReadCtrl 
//////////////////////////////////////////////////////////////////////////////////

`include "trace_port_analyzer/src/defs.v"

module ReadCtrl(

	//*** allgemeine Signale ***

	input wire 					i_clk,			//Systemtakt
	input wire 		[12:3]	i_cols,			//Spaltenanzahl des Speichers
	input wire 		[13:0]	i_rows,			//Zeilenanzahl des Speichers
	input wire 					i_pb,				//zweite physikalische Bank vorhanden
	input wire 					i_res,			//Resetsignal

	//*** Steuerausgaenge ***
	
	output reg				o_read,			//neue Daten aus Speicher holen
	
	output wire 		[1:0]		o_ba,				//Adresse der logischen Bank
	output wire 		[12:0]	o_col,			//Spaltenadresse
	output wire 					o_pb,				//Adresse der physikalischen Bank
	output wire 		[13:0]	o_row,			//Zeilenadresse
	
	//*** Stauseingaenge ***

	input wire 		[1:0]		i_mode,			//Sequenzmodus
	input wire 		[4:0]		i_state,			//aktueller Takt in der Sequenz
	input wire						i_fetch,			//neue Daten lesen (Lesezeiger inkrementieren)
	input wire						i_resFetch,		//Leseteil reseten
	
	//*** Datensignale ***
	
	output reg 	[15:0]	o_data,			//Datenleitungen fuer Datenbus zum MC
	output reg				o_validData,	//gibt an, ob die Daten im Leseregister gueltig sind
	input wire 		[127:0]	i_data			//Datenleitungen von RAM

	);
	
	reg [127:0]	data;
	reg			reading;
	
	reg [2:0]	hword;
	reg 			inc;
	wire			incA;
	wire 			incB;
	wire 			incC;
	wire 			incR;
	wire 			res;
	
	assign res = i_res | i_resFetch;
	//unteren vier Stellen immer null, da mit einem Zugriff immer 2*8 Byte gelesen werden
	assign o_col[3:0] = 4'd0;
	
	//Adresszaehler
	CntL #(2)  iCnt_Burst(o_col[5:4] , incA, i_clk,  inc      , 2'b11       , res);
	CntL #(2)  iCnt_BA	(o_ba       , incR, i_clk, (inc&incA), 2'b11       , res);
	CntL #(14) iCnt_Rows (o_row      , incB, i_clk, (inc&incR), i_rows      , res);
	CntL #(1)  iCnt_PB   (o_pb       , incC, i_clk, (inc&incB), i_pb        , res);
	CntL #(7) iCnt_Cols  (o_col[12:6],     , i_clk, (inc&incC), i_cols[12:3], res);

	
	
	always @(res, o_read, reading)
		if (res)
			o_validData = 0;
		else
			o_validData = !o_read & !reading;

	//Multiplexer fuer Daten
	always @(posedge i_fetch, posedge res)
		if (res)
		begin
			hword = 0;
			o_read = 0;
			o_data = 16'bxxxx_xxxx_xxxx_xxxx;
		end
		else
		begin
			if (o_validData)
				hword = hword + 1;
			
			if (0 == hword)
				o_read = ~reading;
			
			case (hword)
				3'd0:	o_data[15:0] = data[127:112];
				3'd1:	o_data[15:0] = data[111: 96];
				3'd2:	o_data[15:0] = data[ 95: 80];
				3'd3:	o_data[15:0] = data[ 79: 64];
				3'd4:	o_data[15:0] = data[ 63: 48];
				3'd5:	o_data[15:0] = data[ 47: 32];
				3'd6:	o_data[15:0] = data[ 31: 16];
				3'd7:	o_data[15:0] = data[ 15:  0];
			endcase
		end
	
	always @(posedge i_clk)
		if (res)
		begin
			reading = 0;
		end
		else
		begin
			if (o_read)
			begin
				reading = 1;
			end
			
			case ({i_mode, o_ba, i_state})
				{`M_READ,2'd0,5'd6}:		begin	data = i_data;	inc = 1;	reading = 0;	end
				{`M_READ,2'd1,5'd1}:		begin	data = i_data;	inc = 1;	reading = 0;	end
				{`M_READ,2'd2,5'd1}:		begin	data = i_data;	inc = 1;	reading = 0;	end
				{`M_READ,2'd3,5'd1}:		begin	data = i_data;	inc = 1;	reading = 0;	end
					
				{`M_DUAL,2'd0,5'd17}:	begin	data = i_data;	inc = 1;	reading = 0;	end
				{`M_DUAL,2'd1,5'd1}:		begin	data = i_data;	inc = 1;	reading = 0;	end
				{`M_DUAL,2'd2,5'd9}:		begin	data = i_data;	inc = 1;	reading = 0;	end
				{`M_DUAL,2'd3,5'd13}:	begin	data = i_data;	inc = 1;	reading = 0;	end
					
				default: 					begin						inc = 0;	end
			endcase
		end
	

endmodule
