/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Create Date:    12:46:02 01/23/2007 
// Module Name:    TPACtrl 
//////////////////////////////////////////////////////////////////////////////////

`include "trace_port_analyzer/src/TPA_glb.v"

module TPACtrl(

	//*** allgemeine Signale ***
	
	output wire			o_res,
	output	reg		o_trace,
	input wire			i_res,
	input wire 			i_clk,
	
	//*** Signale zur Kommunikation mit dem MC ***

	output wire			o_int,		//Interrupssignal
	output wire		[17:0]	o_data,		//Datenbus (Periph->CPU)
	input wire		[17:0]	i_data,		//Datenbus (CPU->Periph)
	input wire		[9:0]	i_addr,		//Adressbus
	input wire			i_act_peri,	//Peripherie Aktiv
	input wire			i_wr_peri,	//Schreien in Peripherie
	
	//*** interne Parameter ***
	
	output	reg	[23:0]	o_traceDepth,	//Tracetiefe in Schreibzyklen (16 Takte)
	output wire 		[13:0]	o_rows,		//Zeilenanzahl des Speichers
	output wire 		[12:3]	o_cols,		//Spaltenanzahl des Speichers
	output	reg		o_pb,		//2. physikalische Bank
	output 	reg		o_ddr,		//gibt an, ob DDR an der ETM-Seite verwendet wird
	
	output	reg		o_fetch,	//Lesezeiger inkrementieren
	output	reg		o_resFetch,	//Leseeinheit resetten
	input wire		[15:0]	i_readData,	//Datenausschnitt aus Leseregister
	
	///*** Stausangaben ***
	
	//von IntBuffer
	input wire			i_overflow,	//Pufferueberlauf
	input wire			i_readAble,	//duales Lesen moeglich
	input wire			i_recTrace,	//Aufzeichnung laeuft
	input wire			i_initRuns,	//RAM-Initialisierung laeuft
	
	input wire		[9:0]	i_frq,		//Takte fuer die Frequenzmessung
	input wire			i_fin,		//gibt an, dass Aufzeichnung beendet ist
	
	//von ReadCtrl
	input wire			i_validData,	//Lesegesiter enthaelt gueltige Daten
	
	output	reg		o_enInitRam	//RAM-Initialisuerng freigegeben
	
	);
	
	parameter
		P_baseAddr = 0;			//Basisadresse fuer TPA-Geraet

	wire		[16:0]	stateReg;	//Statusregister
	wire		[3:0]	op;		//Operationscode von MC
	wire		[12:0]	dataIn;		//Dateneingang von MC
	wire 			cs;		//Chip-Select
	reg		[1:0]	state;
	reg 			intRes;		//Manuelle Reset-Erzeugung
	reg		[13:12]	rows;
	reg		[11:8]	cols;
	reg			overflow;
	reg		[6:0]	intEn;
	reg			clearInts;
	
	assign op	= i_data[16:13];	//Operationscode
	assign dataIn	= i_data[12:0];		//Operationsdaten
	// Adressdekoder, der aus der angelegten Adresse ein chipselect erzeugt
	pselect iCSL (
		.addr		(	i_addr[9:2]	),
		.activ_peri	(	i_act_peri	),
		.select		(	cs		)
		);
	defparam iCSL.ADDR_WIDTH  = 8;
	defparam iCSL.BASE_WIDTH  = 8;
	defparam iCSL.BASE_ADDR   = P_baseAddr >> 2;	//BASE_ADR muss durch 4 teilbar sein.
	
	
	assign o_res		= i_res | intRes;
	assign o_rows[11:0]	= 12'hFFF;
	assign o_cols[12]	= 0;
	assign o_cols[7:3]	= 4'hF;
	
	assign o_rows[13:12]	= rows;
	assign o_cols[11:8]	= cols;
	
	assign stateReg[16:7] 	= i_frq;
	assign stateReg[6] 	= i_fin;
	assign stateReg[5] 	= i_recTrace;
	assign stateReg[ 4: 3]	= state;
	assign stateReg[2] 	= overflow;
	assign stateReg[1] 	= i_readAble;
	assign stateReg[0] 	= i_validData;
	
	assign o_data = (cs & !i_addr[0]) ? {1'h0,stateReg} :	//Adresse 0: Statusregister
			(cs &  i_addr[0]) ? {2'h0,i_readData} :	//Adresse 1: Leseregister
						 18'h0;
//	assign o_data =  {1'h0,stateReg}; 			//Adresse 0: Statusregister

	IntGen iIntGen(
		.i_fin		(	i_fin		),
		.i_recTrace	(	i_recTrace	),
		.i_overflow	(	i_overflow	),
		.i_validData	(	i_validData	),
		.i_readAble	(	i_readAble	),
		.i_intEn	(	intEn		),
		.i_clr		(	clearInts	),
		.o_int		(	o_int		)
		);
	
	//State ermitteln
	always @(negedge i_clk)
		if (!o_enInitRam)
			state = `TPA_LOAD;
		else if (i_initRuns)
			state = `TPA_INIT;
		else if (o_trace)
			state = `TPA_TRACE;
		else
			state = `TPA_IDLE;
	
	//Overflow-Reset
	always @(i_overflow, o_res)
		if (o_res)
			overflow = 0;
		else
			if (i_overflow)
				overflow = 1;
	
	//Hauptroutine
	always @(posedge i_clk)
		if (i_res)
		begin
			cols = 0;
			intRes = 0;
			rows = 0;
			
			o_pb = 0;
			o_ddr = 0;
			o_enInitRam = 0;
			o_traceDepth[23:13] = 0;
			o_traceDepth[12:0] = 13'b1_1111_1111_1111;
			o_trace = 0;
			o_fetch = 0;
			o_resFetch = 0;
		end
		else if (cs & i_wr_peri)
		begin
			if ( (`TPA_LOAD == state) && (`TPA_SET == op) )
			begin
				//Speichergroesse
				o_pb = dataIn[0];
				rows[13:12] = dataIn[2:1];
				cols[11:8] = dataIn[6:3];
				
				//sonstige Parameter
				o_ddr = dataIn[8];
				
				o_enInitRam = 1;
			end
			
			if ( ((`TPA_IDLE  == state) && (`TPA_INTSSET == op)) ||
				  ((`TPA_TRACE == state) && (`TPA_INTSSET == op)) )
				intEn = dataIn[6:0];
				
			if ( ((`TPA_IDLE  == state) && (`TPA_INTSCLR == op)) ||
				  ((`TPA_TRACE == state) && (`TPA_INTSCLR == op)) )
				clearInts = 1;
			else
				clearInts = 0;
			
			if ( (`TPA_IDLE == state) && (`TPA_SETDEPH == op) )
				o_traceDepth[23:13] = dataIn[10:0];
			
			if ( (`TPA_IDLE == state) && (`TPA_SETDEPL == op) )
				o_traceDepth[12:0] = dataIn;
			
			if ( (`TPA_IDLE == state) && (`TPA_START == op) )
				o_trace = 1;
			
			if ( (`TPA_IDLE == state) && (`TPA_CLEAR == op) )
				intRes = 1;
			else
				intRes = 0;
			
			if ( ((`TPA_IDLE  == state) && (`TPA_FETCH == op)) ||
				  ((`TPA_TRACE == state) && (`TPA_FETCH == op)) )
				o_fetch = 1;
			else
				o_fetch = 0;
			
			if ( ((`TPA_IDLE  == state) && (`TPA_RESFETCH == op)) ||
				  ((`TPA_TRACE == state) && (`TPA_RESFETCH == op)) )
				begin
					o_resFetch = 1;
				end
			else
				o_resFetch = 0;
			
			if ( (`TPA_TRACE == state) && (`TPA_CANCEL == op) )
				o_trace = 0;
			
		end
		else
		begin
			intRes = 0;
			o_fetch = 0;
			o_resFetch = 0;
		end
	
endmodule
