/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 1ns / 1ps
`default_nettype none
//////////////////////////////////////////////////////////////////////////////////
// Create Date:    11:39:57 01/22/2007 
// Module Name:    WriteCtrl 
//////////////////////////////////////////////////////////////////////////////////

`include "trace_port_analyzer/src/defs.v"

module WriteCtrl(

	//*** allgemeine Signale ***

	input wire 					i_clk,		//Systemtakt
	input wire 					i_clk90,		//Systemtakt 90� Phasenverschoben
	input wire 					i_clk180,	//Systemtakt 180� Phasenverschoben
	input wire 					i_clk270,	//Systemtakt 270� Phasenverschoben
	input wire 					i_res,		//Resetsignal
	input wire 		[13:0]	i_rows,		//Zeilenanzahl des Speichers
	input wire			[12:3]	i_cols,		//Spaltenanzahl des Speichers
	input wire 					i_pb,			//zweite physikalische Bank vorhanden
	
	//*** Stauseingaenge ***

	input wire 		[1:0]		i_ba,			//Bank, auf der gelesen werden soll
	input wire 		[1:0]		i_mode,		//Mode, mit Delay fuer Daten (Zuweisung erst in 2. Takt)
	input wire 		[4:0]		i_state,		//Zustand der Zustandsmachine (aktueller Takt)
	
	//*** Steuerausgaenge ***
	
	output wire 	 				o_ceDQ,
	output wire 					o_ceDQS,
	output wire 					o_enBuff,
	output wire 	 				o_fetchBuff,	//Zeigerinkrement im Buffer an RAM-Seite
	output wire 		[13:0]	o_row,		//Zeilenadresse fuer Schreibvorgang
	output wire 		[12:0]	o_col,		//Spaltenadresse fuer Schreibvorgang
	output wire 					o_pb			//Pysikalische Adresse fuer Schreibvorghang

	);
	
	reg mode;
	reg incR;
	wire incB;
	wire incC;
	reg a;
	reg b;
	reg c;
	reg x;
	reg y;
	reg z;
	reg write;
	
	//unteren 6 Bit immer null, da Burstlaenge von 8 * 8 Byte pro burst
	assign o_col[5:0] = 6'd0;
	
	//Adresszaehler
	//1. Zeilen
	CntL #(14) iCnt_Rows(o_row      , incB, i_clk, incR       , i_rows      , i_res);
	//2. physikalische Baenke
	CntL #(1)  iCnt_PB  (o_pb       , incC, i_clk, (incR&incB), i_pb        , i_res);
	//3. Spalten
	CntL #(7) iCnt_Cols(o_col[12:6],     , i_clk, (incR&incC), i_cols[12:6], i_res);
	
	assign o_ceDQ = y | z;
	assign o_ceDQS = c | z;
	assign o_fetchBuff = b | c;
	assign o_enBuff = a | o_fetchBuff;

	always @(posedge i_clk, posedge i_res)
		if (i_res)
		begin
			incR = 1'b0;
			write = 1'b0;
		end
		else
		begin
			//Schreibadresse erhoehen
			casex ({i_mode,i_ba,i_state})
				{`M_WRITE, 2'bxx, 5'd15}:	incR = 1'b1;
				{`M_DUAL , 2'd0 , 5'd18}:	incR = 1'b1;
				{`M_DUAL , 2'd1 , 5'd19}:	incR = 1'b1;
				{`M_DUAL , 2'd2 , 5'd18}:	incR = 1'b1;
				{`M_DUAL , 2'd3 , 5'd18}:	incR = 1'b1;
				default:							incR = 1'b0;
			endcase
			
			//Schreibsignale
			casex ({i_mode,i_ba,i_state})
				{`M_WRITE, 2'bxx, 5'd1}:	write = 1'b1;
				{`M_WRITE, 2'bxx, 5'd0}:	write = 1'b0;
				
				//Dual Bank 0
				{`M_DUAL , 2'd0 , 5'd1}:	write = 1'b1;
				{`M_DUAL , 2'd0 , 5'd12}:	write = 1'b0;
				{`M_DUAL , 2'd0 , 5'd16}:	write = 1'b1;
				{`M_DUAL , 2'd0 , 5'd0}:	write = 1'b0;
				
				//Dual Bank 1
				{`M_DUAL , 2'd1 , 5'd1}:	write = 1'b1;
				{`M_DUAL , 2'd1 , 5'd16}:	write = 1'b0;
				
				//Dual Bank 2
				{`M_DUAL , 2'd2 , 5'd1}:	write = 1'b1;
				{`M_DUAL , 2'd2 , 5'd4}:	write = 1'b0;
				{`M_DUAL , 2'd2 , 5'd8}:	write = 1'b1;
				{`M_DUAL , 2'd2 , 5'd0}:	write = 1'b0;
				
				//Dual Bank 3
				{`M_DUAL , 2'd3 , 5'd1}:	write = 1'b1;
				{`M_DUAL , 2'd3 , 5'd8}:	write = 1'b0;
				{`M_DUAL , 2'd3 , 5'd12}:	write = 1'b1;
				{`M_DUAL , 2'd3 , 5'd0}:	write = 1'b0;
				
			endcase
		end
	
	//always @(posedge i_clk, posedge i_res)
	//	if (i_res)
	//		o_enBuff = 1'b0;
	//	else
	//		o_enBuff = write | o_fetchBuff;
	
	always @(posedge i_clk90, posedge i_res)
	//always @(posedge i_clk, posedge i_res)
		if (i_res)
			c = 1'b0;
		else
			c = b;
	
	always  @(negedge i_clk, posedge i_res)
	//always  @(posedge i_clk180, posedge i_res)
	//always  @(posedge i_clk, posedge i_res)
		if (i_res)
			b = 1'b0;
		else
			b = a;
	
	always @(negedge i_clk90, posedge i_res)
	//always @(posedge i_clk270, posedge i_res)
	//always @(negedge i_clk90, posedge i_res)
	begin
		if (i_res)
		begin
			a = 1'b0;
			x = 1'b0;
			y = 1'b0;
			z = 1'b0;
		end
		else
		begin
			z = y;
			y = x;
			x = a;
			a = write;
		end
	end

	/*** DUBUGAUSGABEN ***/
	/*integer file;

	initial
	begin
		file = $fopen("writeS.txt");
		$fdisplay(file, "Fileno.: %d", file);
		$fdisplay(file, "");
		$fmonitor(file, "%t MODE:%b%b State=%d write=%b   a=%b b=%b c=%b   x=%b y=%b z=%b   ceDQ=%b  ceDQS=%b", $time,
			i_mode[1],i_mode[0], i_state, write, a, b, c, x, y, z, o_ceDQ, o_ceDQS);
			
	end*/


/*	
	always @(posedge i_clk, posedge i_res)
		if (i_res)
			inc = 0;
		else
			casex ({i_mode,i_ba,i_state})
				{`M_WRITE, 2'bxx, 5'd15}:	incR = 1;
				{`M_DUAL , 2'd0 , 5'd18}:	incR = 1;
				{`M_DUAL , 2'd1 , 5'd19}:	incR = 1;
				{`M_DUAL , 2'd2 , 5'd18}:	incR = 1;
				{`M_DUAL , 2'd3 , 5'd18}:	incR = 1;
				default:							incR = 0;
			endcase
	
	always @(posedge i_clk90, posedge i_res)
	if (i_res)
	begin
		o_dirOut = 1'b0;
		o_fetchBuff = 0;
		mode = `M_REFR;
	end
	else
	begin
		o_dirOut = (i_mode == `M_WRITE);
		
		casex ({i_mode,i_ba})
			{`M_WRITE,2'bxx}:
				o_fetchBuff = 1;
				
			{`M_DUAL,2'd0}:
				if((5'd15 == i_state) |
					(5'd16 == i_state) |
					(5'd17 == i_state) )
					o_fetchBuff = 0;
				else
					o_fetchBuff = 1;
				
			{`M_DUAL,2'd1}:
				if((5'd19 == i_state) |
					(5'd0  == i_state) |
					(5'd1  == i_state) |
					(5'd2  == i_state) )
					o_fetchBuff = 0;
				else
					o_fetchBuff = 1;
				
			{`M_DUAL,2'd2}:
				if((5'd7 == i_state) |
					(5'd8 == i_state) |
					(5'd9 == i_state) )
					o_fetchBuff = 0;
				else
					o_fetchBuff = 1;
				
			{`M_DUAL,2'd3}:
				if((5'd11 == i_state) |
					(5'd12 == i_state) |
					(5'd13 == i_state) )
					o_fetchBuff = 0;
				else
					o_fetchBuff = 1;
				
			default:
				o_fetchBuff = 0;
			
		endcase
	end
*/

endmodule
