/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt	: SpartanMC
// Dateiname	: uart_top.v
// Autor	: Steffen Kunze
// Modul	: uart_top
//
// Beschreibung : Das uart_top-Modul enthaelt die FIFO- und Transceiver-Instanzen
// 		  uns stellt optional Modemleitungen zur Verfuegung.
///////////////////////////////////////////////////////////////////////////////

`timescale 1 ns / 1 ps
`default_nettype none

module uart_top(
		clk_peri, 
		do_peri,
		di_peri,
		addr_peri,
		access_peri,
		wr_peri,
		uart_clk,
		reset,
		intr,
		rx,
		tx,
		rts,
		dcd,
		dtr
		);

function integer log2;

	input integer value;

	for (log2=0; value>0; log2=log2+1)
		value = value>>1;
endfunction


parameter CLKIN_FREQ	= 25000000;

parameter BASE_ADR	= 10'h0;
parameter MODEM		= 0;
parameter FIFO_RX_DEPTH = 'd8;
parameter FIFO_TX_DEPTH	= 'd8;

parameter FIFO_RX_ADDR_BITS   = log2(FIFO_RX_DEPTH - 1);
parameter FIFO_TX_ADDR_BITS   = log2(FIFO_TX_DEPTH - 1);

// E/A - Adressen der UART (Bitnummer der Select Signale)
parameter USTATUS_ADR	= 0;
parameter UREAD_ADR	= 1;
parameter UWRITE_ADR	= 2;
parameter UCTRL_ADR	= 3;
parameter UMODEM_ADR	= 4;
parameter URX_ADR	= 5;

parameter ALL_PORTS	= 6;
parameter DECODSIM_DEB	= "NO";

// inputs & outputs
input wire		clk_peri;
input wire	[17:0]	do_peri;		// Datenbus
output wire	[17:0]	di_peri;		// Datenbus
input wire	[9:0]	addr_peri;		// Adressbus
input wire		access_peri;		// Peripherie aktiviert
input wire		wr_peri;
input wire		uart_clk;		// vorgeteilter Takt mit ca. 6*16*115200Hz=11MHz
input wire		reset;
output wire		intr;			// Interrupt Tx-FIFO frei/ Zeichen im Rx FIFO

input wire		rx;	 		// serieller Input
output wire		tx;			// serieller Output
inout wire		rts, dcd;		// Modemleitungen
inout wire		dtr;			// Modemleitungen

wire		tx_stop;		// tx Schieberegister arbeitet nicht
wire		rst_uart;		// Reset für die UART
wire	[7:0]	datout;		 	// Daten vom FIFO
wire	[8:0]	fullemptystat;	 	// FIFO-Status
wire	[2:0]	modemstatus;		// Statusleitungen des Modem zusammengefasst

// Verbindungen FIFO-Transceiver
wire		clk_div;		// Rx Interrups koennen nur synchron zum Schiebetakt entstehen.
wire	[7:0]	rx_data;
wire	[7:0]	tx_data;
wire	[3:0]	fullempty;
wire		tx_ff_rd;
wire		rx_ff_wr;
// Rx Fehler Signale
wire		err_fr;			// offene Leitung/BREAK
wire		err_pa;			// Paritaetsfehler
wire		err_da;			// Pufferueberlauf

//
// Die negative Flanke von rx_ff_rd beendet das Lesen und
// die negative Flanke von tx_ff_wr beendet das Schreiben der Daten.
//
// Interruptleitungen
wire		ir_tx;
wire		ir_rx;
wire		rx_int_e;
wire		tx_int_e;
reg		tx_ir_en;
reg		tx_ir_env;
reg		tx_ir_sig;
reg		rx_ir_sig;
reg		cs_statold;

wire		read_en;
wire		write_en;

// Signale zur Fehlersuche und zum Test
wire	[3:0]	i;		// Bitzaehler Rx
wire	[3:0]	cnt;		// Zaehler Rx
wire		rx_clk;		// CLK fuer Rx Schieberegister
wire		cnt_stop;	// Rx-Zaehler Null setzen
wire	[3:0]	j;		// Bitzaehler Tx
wire	[3:0]	cnt0;		// Zaehler Tx
wire		tx_clk;		// CLK fuer Tx Schieberegister
//
// Ports zum lesen der UART Daten und Statusinformationen
//
wire	[17:0]	port1v	= {1'b0, tx_stop, rst_uart, tx_clk, cnt_stop, rx_clk, modemstatus, fullemptystat};
//wire	[17:0]	port2v	= {tx_clk, rx_clk, cnt0[3:0], j[3:0], cnt[3:0], i[3:0]};
wire	[17:0]	port2v	= {1'b0, rx_clk, cnt0[3:0], j[3:0], cnt[3:0], i[3:0]};
//
// Informationen mit clk_peri synchronisiert
//
reg	[17:0]	port1;
reg	[17:0]	port2;
reg	[17:0]	port3;

// Dekodierung der Adressen ********************** START *************************************
// chipselect fuer die Statusregister (ohne Modemsignale)
wire					cs_status;	// Lesen
wire					rx_rd;		// Lesen / Reset Rx Interrupt
wire					tx_ff_wr;	// Schreiben
wire					cs_stat2;	// lesen / Reset Tx Interrupt
wire					cs_ctrli;	// Schreiben
wire					sh_stat;	// lesen
wire					rx_ff_rd;	// FIFO lesen

wire					select;
wire		[ALL_PORTS-1:0]		reg_select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_select(	reg_select	),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	)
			);

// chipselect fuer die FIFO-,Status- und Steuerregister
assign	cs_status	= reg_read[USTATUS_ADR];	// Lesen
assign	rx_rd		= reg_read[UREAD_ADR];		// Lesen / Reset Rx Interrupt
assign	tx_ff_wr	= reg_write[UWRITE_ADR];	// Schreiben
assign	cs_stat2	= reg_read[UCTRL_ADR];		// lesen / Reset Tx Interrupt
assign	cs_ctrli	= reg_write[UCTRL_ADR];		// Schreiben
assign	sh_stat		= reg_read[URX_ADR];		// lesen
assign	rx_ff_rd	= reg_select[UREAD_ADR]&!wr_peri; // FIFO lesen

// Dekodierung der Adressen ********************** END **************************************

// Daten an den Bus legen
assign di_peri [17:0]	= (cs_status)?	port1:
			  (cs_stat2)?	port1:
			  (sh_stat)?	port2:
       			  (rx_rd)?	port3: 18'h0;


assign		intr  = ir_tx | ir_rx;
assign		ir_tx = tx_int_e ? tx_ir_sig:1'b0;
assign		ir_rx = rx_int_e ? rx_ir_sig:1'b0;

assign		fullemptystat = {err_da,err_fr,err_pa,tx_ir_sig,tx_ir_en,fullempty};

// Interrupts generieren. Sie koennen nur synchron zum Schiebetakt entstehen.
always@(posedge clk_peri)
	if(reset) begin
		tx_ir_env	<= 1'b0;
		tx_ir_en	<= 1'b1;	// fuer Tx Int. beim Neustart
		tx_ir_sig	<= 1'b0;
		rx_ir_sig	<= 1'b0;
		cs_statold	<= 1'b0;
		port1		<= 18'b0;
		port2		<= 18'b0;
		port3		<= 18'b0;
	end
	else
	begin
		port1		<= port1v;
		port2		<= port2v;
		// FIFO in Puffer port3 lesen
		if(rx_ff_rd)
			port3	  <= {10'h0, datout};
		
		// Rx Interrupt zuruecksetzen beim lesen des
		// Datenregisters und setzen wenn RxFIFO nicht leer ist.
		if(rx_rd)
			rx_ir_sig <= 1'b0;
		else if(!fullempty[0])
			rx_ir_sig <= 1'b1;		// rxfifo noch nicht leer

		// Tx Interrupt setzen, solange der FIFO
		// nicht voll ist (txfull = 0 = fullempty[3])..
		if(!fullempty[3] && tx_ir_en && !cs_stat2 && !cs_statold)
			begin
			tx_ir_en  <= 1'b0;
			tx_ir_sig <= 1'b1;
			end
		// Interrupt fuer Tx-FIFO Zuruecksetzen nach dem lesen des Statusregisters.
		// Da in new_pipe cs_statold einen Takt spaeter, geaendert auf cs_stat2
		else if(cs_stat2 && tx_ir_sig)
			begin
			tx_ir_sig <= 1'b0;
			end
		// Merken einer Interrupt freigabe
		// beim Schreiben eines zu sendenden Bytes.
		if(tx_ff_wr && !tx_ir_env)
			begin
			tx_ir_env <= 1'b1;
			end
		// Setzen der Interrupt freigabe, wenn eine Freigabe vorgemerkt ist.
		else if(tx_ir_env && !tx_ir_sig)
			begin
			tx_ir_en  <= 1'b1;
			tx_ir_env <= 1'b0;
			end
	cs_statold	<= cs_stat2;
	end

// Transceiverinstanz
uart_transceiver t1 (	.clk(		clk_peri	),
			.uart_clk(	uart_clk	),
			.clk_divg(	clk_div		),
			.reset(		reset		),
			.rst_uart(	rst_uart	),
			.rx_data(	rx_data		),
			.tx_data(	tx_data		),
			.fifo_rx_flag(	fullempty[1]	),
			.fifo_tx_flag(	fullempty[2]	),
			.rx_ff_wr(	rx_ff_wr	),
			.tx_ff_rd(	tx_ff_rd	),
			.tx_stop(	tx_stop		),
			.data(		do_peri		),
			.cs_ctrli(	cs_ctrli	),
			.Rx(		rx		),
			.Tx(		tx		),
			.rx_int_e(	rx_int_e	),
			.tx_int_e(	tx_int_e	),
			.read_en(	read_en		),
			.write_en(	write_en	),
			.err_fr(	err_fr		),
			.err_pa(	err_pa		),
			.err_da(	err_da		),

// Signale zur Fehlersuche und zum Test
			.i(		i		),
			.cnt(		cnt		),
			.rx_clk(	rx_clk		),
			.cnt_stop(	cnt_stop	),
			.j(		j		),
			.cnt0(		cnt0		),
			.tx_clk(	tx_clk		)
		);

defparam t1.CLKIN_FREQ = CLKIN_FREQ;

// Empfangs-FIFO
	yaf #
	(
		.DSIZE (	'd8		),
		.ASIZE (	FIFO_RX_ADDR_BITS  )
	)
	yaf_rx0
	(
		.rdata(		datout		),	//Empfangsdaten zum MC
		.wfull(				),	//Empfangs-FIFO voll mit wclk synchronisiert
		.rfull(				),	//Empfangs-FIFO voll mit rclk synchronisiert
		.rwfull(	fullempty[1]	),	//Empfangs-FIFO voll = rfull | wfull
		.wempty(			),	//Empfangs-FIFO leer mit wclk synchronisiert
		.rempty(			),	//Empfangs-FIFO leer mit rclk synchronisiert
		.rwempty(	fullempty[0]	),	//Empfangs-FIFO leer = rempty & wempty
		.fill(				),	//Output wird nicht verwendet.
		.wdata(		rx_data		),	//Empfangsdaten vom Transceiver
		.winc(		rx_ff_wr	),	//Transceiver schreibt Daten
		.wclk(		clk_div		),
		.wrst_n(	~reset		),
		.rinc(		rx_ff_rd	),	//MC liest Daten solange read=1 ist.
		.rclk(		clk_peri	),
		.rrst_n(	~reset		),
		.rptr_upd(	1'b0		),
		.rptr_rest(	1'b0		),
		.wptr_upd(	1'b0		),
		.wptr_rest(	1'b0		)
	);


// Sende-FIFO
	yaf #
	(
		.DSIZE (	'd8		),
		.ASIZE (	FIFO_TX_ADDR_BITS  )
	)
	yaf_tx0
	(
		.rdata(		tx_data		),	//Sendedaten zum Tranceiver
		.wfull(				),	//Sende-FIFO voll mit wclk synchronisiert
		.rfull(				),	//Sende-FIFO voll mit rclk synchronisiert
		.rwfull(	fullempty[3]	),	//Sende-FIFO voll = rfull | wfull
		.wempty(			),	//Sende-FIFO leer mit wclk synchronisiert
		.rempty(	fullempty[2]	),	//Sende-FIFO leer mit rclk synchronisiert
		.rwempty(			),	//Sende-FIFO leer = rempty & wempty
		.fill(				),	//Output wird nicht verwendet.
		.wdata(		do_peri[7:0]	),	//Sendedatem vom MC
		.winc(		tx_ff_wr	),	//MC schreibt Daten solange write=1 ist.
		.wclk(		clk_peri	),
		.wrst_n(	~reset		),
		.rinc(		tx_ff_rd	),	//Transceiver liest Daten
		.rclk(		clk_div		),
		.rrst_n(	~reset		),
		.rptr_upd(	1'b0		),
		.rptr_rest(	1'b0		),
		.wptr_upd(	1'b0		),
		.wptr_rest(	1'b0		)
	);


// Modemleitungen hinzufuegen **************************************************
generate if (MODEM != 0) begin
	reg		dsr1;
	reg		cts1;
	reg		dcdr1;
	reg		dsr;
	reg		cts;
	reg		dcdr; 
	reg	[2:0]	modem_dir;
	wire		modem_cs;	// Schreiben

assign	modem_cs	= reg_write[UMODEM_ADR];	// Schreiben

	// Hier koennen die Register der Modemleitungen geschrieben und
	// gelesen werden.
	always @ (posedge clk_peri)
	begin
		if(reset)
		begin
			dsr	= 1'b0;
			cts	= 1'b0;
			dcdr	= 1'b0;
			dsr1	= 1'b0;
			cts1	= 1'b0;
			dcdr1	= 1'b0;
			modem_dir = 3'b000;
		end
		else
		begin
			// Daten in die Modemregister schreiben
			if(modem_cs)
			begin
				dsr = do_peri[0];
				cts = do_peri[1];
				dcdr = do_peri[2];
				modem_dir = do_peri[5:3];
			end
			else
			begin
			// Wenn dcd als Eingang geschalten ist, Daten Uebernehmen
				if(~modem_dir[0])
				begin
					dcdr  = dcdr1;
					dcdr1 = dcd;
				end
			// Wenn cts als Eingang geschalten ist, Daten Uebernehmen
				if(~modem_dir[1])
				begin
					cts  = cts1;
					cts1 = rts;
				end
			// Wenn dsr als Eingang geschalten ist, Daten Uebernehmen
				if(~modem_dir[2])
				begin
					dsr  = dsr1;
					dsr1 = dtr;
				end
			end
		end
	end

	assign dcd = (modem_dir[0]) ? dcdr : 1'bz;
	assign rts = (modem_dir[1]) ? cts  : 1'bz;
	assign dtr = (modem_dir[2]) ? dsr  : 1'bz;
	assign modemstatus = {dsr,cts,dcdr};
end

// Modemleitungen hochohmig ****************************************************
else begin
	assign dcd = 1'bz;
	assign rts = 1'bz;
	assign dtr = 1'bz;
	assign modemstatus = 0;
end endgenerate

endmodule
