/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


///////////////////////////////////////////////////////////////////////////////
// Projekt	: SpartanMC
// Dateiname	: uart_transceiver.v
// Autor	: Steffen Kunze
// Modul	: uart_transceiver
//
// Beschreibung : Der UART empfaengt und sendet serielle Daten nach dem ueblichen
// 		  Verfahren: Startbit, 5-8 Datenbit, optional Paritaetsbit,
//                1-2 Stopbits. Die Baudrate kann programmiert werden. Beim
//		  Empfang wird 16-fach ueberabgetastet. Die Datenworte werden
//		  zum FIFO Uebergeben/von diesem gelesen.
///////////////////////////////////////////////////////////////////////////////
// Baudraten:
// 115200
// 57600
// 38400
// 31250     MIDI
// 19200
// 9600
// 7812.5    hc11-BOOT
// 4800
// 2400
// 1200
// 600
// 300
// 150
// 75
// 50

`timescale 1 ns / 1 ps
`default_nettype none

module uart_transceiver (
				clk,
				uart_clk,
				clk_divg,
				reset,
				rst_uart,
				rx_data,
				tx_data,
				fifo_tx_flag,
				fifo_rx_flag,
				data,
				tx_stop,
				tx_ff_rd,
				rx_ff_wr,
				Tx,
				Rx,
				tx_int_e,
				rx_int_e,
				read_en,
				write_en,
				err_fr,
				err_pa,
				err_da,
				cs_ctrli,

// Signale zur Fehlersuche und zum Test
				i,
				cnt,
				rx_clk,
				cnt_stop,
				j,
				cnt0,
				tx_clk
			);

`include "uart/src/include/uart_functions.vh"

// Vorteiler mit definierter Baudrate*16 wegen 16 fachem Oversampling erzeugen
// Baud = CLKIN_FREQ / (OVS_RATE * (Teiler + 1))   mit Teiler = BAUD_115200
// Baud = CLKIN_FREQ / (OVS_RATE * ((CLKIN_FREQ_OVS / 115200 +5)/10 -1 + 1))

//      =  25000000  / (  12     * ((  20833333     / 115200 +5)/10))
//      =  25000000  / (  12     * ((  18  )) = 115740,741

//      =  62500000  / (  16     * ((  39062500     / 115200 +5)/10))
//      =  62500000  / (  16     * ((  34  )) = 114889,706

parameter CLKIN_FREQ	= 25000000;				//Eingangstakt, benoetigt fuer Baudratenberechnung

parameter BAUD_ERR12	= max_baud_error(CLKIN_FREQ,12);
parameter BAUD_ERR14	= max_baud_error(CLKIN_FREQ,14);
parameter BAUD_ERR16	= max_baud_error(CLKIN_FREQ,16);
parameter BAUD_ERR18	= max_baud_error(CLKIN_FREQ,18);
parameter BAUD_ERR20	= max_baud_error(CLKIN_FREQ,20);
parameter OVS_RATE	= find_best_ovs_rate(CLKIN_FREQ);	//Berechnete Oversampling-Rate mit kleinstem Baudratenfehler
parameter OVS_CNT_WIDTH	= log2(OVS_RATE-1);			//Breite des Zaehlers fuer Oversampling
parameter SAMPLING_POS	= OVS_RATE/2-3;				//Position des Zaehlers, bei der das Bit abgetastet wird

parameter CLKIN_FREQ_OVS= clkin_freq_ovs(CLKIN_FREQ,OVS_RATE);

parameter BAUD_115200	= baud_teiler(CLKIN_FREQ_OVS,115200)-1;
parameter BAUD_57600	= baud_teiler(CLKIN_FREQ_OVS,57600)-1;
parameter BAUD_38400	= baud_teiler(CLKIN_FREQ_OVS,38400)-1;
parameter BAUD_31250	= baud_teiler(CLKIN_FREQ_OVS,31250)-1;
parameter BAUD_19200	= baud_teiler(CLKIN_FREQ_OVS,19200)-1;
parameter BAUD_9600	= baud_teiler(CLKIN_FREQ_OVS,9600)-1;
parameter BAUD_7812	= baud_teiler(CLKIN_FREQ_OVS*10,78125)-1;
parameter BAUD_4800	= baud_teiler(CLKIN_FREQ_OVS,4800)-1;
parameter BAUD_2400	= baud_teiler(CLKIN_FREQ_OVS,2400)-1;
parameter BAUD_1200	= baud_teiler(CLKIN_FREQ_OVS,1200)-1;
parameter BAUD_600	= baud_teiler(CLKIN_FREQ_OVS,600)-1;
parameter BAUD_300	= baud_teiler(CLKIN_FREQ_OVS,300)-1;
parameter BAUD_150	= baud_teiler(CLKIN_FREQ_OVS,150)-1;
parameter BAUD_75	= baud_teiler(CLKIN_FREQ_OVS,75)-1;
parameter BAUD_50	= baud_teiler(CLKIN_FREQ_OVS,50)-1;

parameter PRESCALER_WIDTH	= log2(BAUD_50);

// inputs & outputs
input wire		clk;
input wire		uart_clk;	// vorgeteilter Takt mit ca. 50MHz/4,5=11MHz
output wire		clk_divg;	// Interrups koennen nur synchron zum Schiebetakt entstehen.
input wire		reset;
output reg		rst_uart;	// RESET bis zum naehsten Tackt des Schieberegisters
input wire	[7:0]	tx_data;	// Daten vom FIFO
output wire	[7:0]	rx_data;	// Daten zum FIFO
input wire		fifo_tx_flag;	// Tx-FIFO leer
input wire		fifo_rx_flag;	// Rx-FIFO voll
input wire		cs_ctrli;	// chipselect Steuerregister
input wire	[17:0]	data;		// Datenbus
output reg		tx_stop;	// tx Schieberegister arbeitet nicht
output reg		tx_ff_rd;	// Signal, das Wort vom FIFO gelesen wurde
output wire		rx_ff_wr;	// Signal, das Wort fuer FIFO bereitsteht
output wire		tx_int_e;	// Tx Interrupt enable
output wire		rx_int_e;	// Rx Interrupt enable
output reg		read_en;	// enable Rx Transceiver
output reg		write_en;	// enable Tx Transceiver

output wire		err_fr;		// Frame Error
output wire		err_pa;		// Paritaets Error
output wire		err_da;		// Datenverlust - FIFO ist voll.

input wire		Rx; 		// serieller Input
output reg		Tx; 		// serieller Output

// Signale zur Fehlersuche und zum Test
output wire	[3:0]	i;		// Bitzaehler Rx
output reg	[OVS_CNT_WIDTH-1:0]	cnt;		// Zaehler Rx
output reg		rx_clk;		// CLK fuer Rx Schieberegister
output wire		cnt_stop;	// Rx-Zaehler Null setzen
output reg	[3:0]	j;		// Bitzaehler Tx
output reg	[OVS_CNT_WIDTH-1:0]	cnt0;		// Zaehler Tx
output reg		tx_clk;		// CLK fuer Tx Schieberegister

reg	[7:0]	tx_datap;	// Tx Datenpuffer
reg			fifo_tx_empty;
reg 		tx_ff_rdv;


reg		parity_en;	// ueertragung mit Paritaetsbit
reg		parity_sel; 	// Paritae: 1=gerade, 0=ungerade
reg		stopbit;	// Stopbits: 0=1,1=2
reg	[3:0]	baudrate;	// Baudrateneinstellung
reg	[2:0]	word_length;	// Wortlaenge
reg		rx_ie, tx_ie;	// Interrupt enable
reg		tx_br;		// Senden von Break

assign rx_int_e = rx_ie;
assign tx_int_e = tx_ie;

// Signal zur Fehlersuche und zum Test

reg	[7:0]	tx_sreg;	// Schieberegister Tx
reg		tx_par;	 	// Register fuer Paritaetsberechnung

wire	[3:0]	wdl_1	=	word_length + 1;
wire	[3:0]	wdl_2	=	word_length + 2;
wire	[3:0]	wdl_3	=	word_length + 3;
wire	[3:0]	wdl_4	=	word_length + 4;


reg	[PRESCALER_WIDTH-1:0]	clk_cnt;

reg		clk_div;
reg		clk_dive;

// RESET fuer die UART
reg	        rst_uv;		// RESET bis zum naehsten Tackt des Schieberegisters

always @ (posedge uart_clk)
if(reset)
	begin
		clk_cnt	<= {PRESCALER_WIDTH{1'b0}};
		clk_div	<= 1'b0;
	end
else
	if (~|clk_cnt)
		begin
		clk_div	<= 1'b1;
		case(baudrate)		// Wert = Teiler - 1
			4'b0000: clk_cnt<=BAUD_115200;	//115200
			4'b0001: clk_cnt<=BAUD_57600;	//57600
			4'b0010: clk_cnt<=BAUD_38400;	//38400
			4'b0011: clk_cnt<=BAUD_31250;	//31250  MIDI clock
			4'b0100: clk_cnt<=BAUD_19200;	//19200
			4'b0101: clk_cnt<=BAUD_9600;	//9600
			4'b0110: clk_cnt<=BAUD_4800;	//4800
			4'b0111: clk_cnt<=BAUD_2400;	//2400
			4'b1000: clk_cnt<=BAUD_1200;	//1200
			4'b1001: clk_cnt<=BAUD_600;	//600
			4'b1010: clk_cnt<=BAUD_300;	//300
			4'b1011: clk_cnt<=BAUD_150;	//150
			4'b1100: clk_cnt<=BAUD_75;	//75
			4'b1101: clk_cnt<=BAUD_50;	//50
			default: clk_cnt<=BAUD_7812;	//7812   hc11 boot
		endcase
		end
	else
		begin
		clk_div	<= 1'b0;
		clk_cnt <= clk_cnt - 1;
		end

// Enable fuer einen BUFGCE aus clk_div bilden
always @ (negedge uart_clk)
if(reset)
	begin
		clk_dive	<= 1'b0;
	end
else
	if(clk_div)
		begin
			clk_dive	<= 1'b1;
		end
	else
		begin
			clk_dive	<= 1'b0;
		end

// CLK Treiber fuer die Schieberegister

bufgce_prim bufgce_prim_inst (
	.O(	clk_divg),
	.CE(	clk_dive),
	.I(	uart_clk)
);

// In diesem Block werden Daten vom Bus in die Register Uebernommen.
always @ (posedge clk)
	if(reset)
	begin
	rst_uv		<= 1'b1;
	read_en		<= 1'b0;
	write_en	<= 1'b0;
	parity_en	<= 1'b0;
	parity_sel	<= 1'b0;
	stopbit		<= 1'b0;
	word_length	<= 3'b000;
	baudrate	<= 4'b0000;
	rx_ie		<= 1'b0;
	tx_ie		<= 1'b0;
	tx_br		<= 1'b0;
	end
	else
	   begin
		if(cs_ctrli)
		begin
		read_en		<= data[0];
		write_en	<= data[1];
		parity_en	<= data[2];
		parity_sel	<= data[3];
		stopbit		<= data[4];
		word_length	<= data[7:5];
		baudrate	<= data[12:9];
		rx_ie		<= data[13];
		tx_ie		<= data[14];
		tx_br		<= data[15];
		end

		if(cnt0==(OVS_RATE-1))
		rst_uv		<= 1'b0;	// RESET fuer die Schieberegister ausschalten
	   end


// In diesem Block laeuft der Zaehler fuer das Oversampling beim Empfang und
// fuer das Senden laeuft ebenfalls ein Zaehler, der den Takt durch 16 teilt,
// um die letztendliche Baudrate zu erzielen.
always@(posedge clk_divg)
	if(reset)
	begin
		rst_uart	<= 1'b1;	// RESET fuer die Schieberegister einschalten
		cnt0		<= {OVS_CNT_WIDTH{1'b0}};
		cnt		<= {OVS_CNT_WIDTH{1'b0}};
		tx_clk		<= 1'b0;
		rx_clk		<= 1'b0;
	end
	else
	begin
		if (cnt0==(OVS_RATE-1))
		begin
			cnt0 <= {OVS_CNT_WIDTH{1'b0}};
			tx_clk	<= 1'b1;
		end else begin
			cnt0	<= cnt0 + 1;	// tx Zaehler
			tx_clk	<= 1'b0;
		end

		if(cnt_stop)
		begin
			cnt	<= {OVS_CNT_WIDTH{1'b0}};
			rx_clk	<= 1'b0;
		end
		else
		begin
			if (cnt==(OVS_RATE-1))
				cnt <= {OVS_CNT_WIDTH{1'b0}};
			else
			begin
				cnt	<= cnt + 1;	// rx Zaehler
				if (cnt==SAMPLING_POS)	// in der Mitte des Bits abtasten
					rx_clk	<= 1'b1;// rx lesen, wenn cnt von 6 auf 7 geht.
				else
					rx_clk	<= 1'b0;
			end
		end
		if(!rst_uv)
		rst_uart	<= 1'b0;	// RESET fuer die Schieberegister ausschalten
	end


// Daten aus TX-FIFO synchron zum IO-Clk laden
always@(posedge clk_divg)
	if(rst_uv)
	begin
		tx_ff_rd	<= 1'b0;
		tx_datap	<= 7'b0;
	end
	else
	begin
		if(tx_ff_rdv && !tx_ff_rd)
			begin
			tx_ff_rd	<= 1'b1;
			tx_datap	<= tx_data;
			end
		else
			tx_ff_rd	<= 1'b0;
	end

// Hier werden die Daten aus dem Sendeschieberegister geschoben und Start-,
// Stop- und Paritaetsbits eingefuegt.
always@(posedge clk_divg)
	if(rst_uart)
	begin
		tx_ff_rdv	<= 1'b0;
		fifo_tx_empty	<= 1'b1;
		tx_sreg 	<= 0;
		j       	<= 4'b0000;
		Tx      	<= 1;	// Ruhezustand auf der Leitung
		tx_par		<= 1'b0;
		tx_stop		<= 1'b1;
	end
	else
	if(write_en && ~tx_br)
		// Alle 16 Takte von clk_div das naechste bit erzeugen
		if(tx_clk)
		begin
			fifo_tx_empty <= fifo_tx_flag; // Status synchronisieren
			// das Starbit wird gesendet und die Daten vom FIFO geladen
			if((j == 4'b0000) && ~fifo_tx_empty)    // Start, wenn Daten im FIFO
			begin
				tx_ff_rdv <= 1'b1;	// Datenuebernahme
							// signalisieren
				Tx	  <= 1'b0;	// Startbit senden
				tx_stop	  <= 1'b0;
			end

			// Falls FIFO leer ist, warten
			else if((j == 4'b0000) && fifo_tx_empty)
				begin
				Tx	<= 1'b1;
				tx_stop	<= 1'b1;
				end

			// Datenbits senden und Schieberegister schieben
			else if (j <= wdl_1)
			begin
			//	tx_ff_rdv	<= 1'b0; // Datenuebernahme beenden
				{tx_sreg,Tx}	<= {1'b0,tx_sreg};
				j <= j + 1;
			end

			// Wenn Paritaetsbit gesetzt, dieses senden
			else if((j == wdl_2) && parity_en)
			begin   //parity-bit
				Tx	<= tx_par;
				j	<= j+1;
			end

			// 1. Stopbit, wenn zwei Stopbits eingestellt sind
			else if(stopbit && ((~parity_en && (j==wdl_2))
				|| (parity_en && (j==wdl_3))))
			begin	//Stopbit
				Tx	<= 1'b1;
				j	<= j+1;
			end

			// 2. Stopbit
			else if(stopbit && ((~parity_en && (j>=wdl_3))
				|| (parity_en && (j>=wdl_4))))
			begin
				j	<= 0;	// alles gesendet
				Tx	<= 1'b1;
			end

			// 1. Stopbit, wenn nur ein Stopbit eingestellt ist
			else if(~stopbit && ((~parity_en && (j>=wdl_2))
				|| (parity_en && (j>=wdl_3))))
			begin
				j	<= 0;	// alles gesendet
				Tx	<= 1'b1;
		   	end
			else

			// kein bit senden
			begin
				Tx	<= 1'b1;
			end
		end
		else
		// die restlichen 15 Takte von clk_div den Pegel halten
		// und Daten aus dem FIFO laden.
		begin
			Tx <= Tx;
			// Daten aus FIFO in das
			if((j == 4'b0000) && tx_ff_rd)
			begin
				// Schieberegister laden.
				tx_sreg	<= tx_datap;
				// Paritaet berechnen
				if(^tx_datap == parity_sel)
					tx_par <= 1'b1;
				else
					tx_par <= 1'b0;
				j	<= 4'b0001;
				tx_ff_rdv	<= 1'b0; // Datenuebernahme beenden
			end
		end
	else
	// Tx disabled oder Break senden
	begin
		if (tx_br)
			Tx <= 1'b0;	// BREAK senden
		else
			Tx <= 1'b1;
	end

// Instanz des Empfaenger
uart_receiver r1(
	.rx_pin		(Rx),
	.clk		(clk),
	.clk_div	(clk_divg),
	.rx_clk		(rx_clk),
	.reset		(reset),
	.rst		(rst_uart),
	.enable		(read_en),
	.write_enable	(!fifo_rx_flag),

	.parity_en	(parity_en),
	.parity_sel	(parity_sel),
	.word_length	(word_length),

	.cnt_reset	(cnt_stop),
	.write		(rx_ff_wr),
	.data_out	(rx_data),

	.err_fr		(err_fr),
	.err_pa		(err_pa),
	.err_da		(err_da),

	.i		(i)
);

endmodule

// Empfangseinheit.
// Hier werden die Daten ins Empfangsschieberegister geschoben und Start-,
// Stop- und Paritaetsbits Ueberprueft.
// Abtastung mit Mehrheitsbildung aus drei Werten
module uart_receiver(
	input	wire	rx_pin,		// serieller Input
	input	wire	clk,		// FIFO clk
	input	wire	clk_div,	// UART-Takt mit 16fachem Oversampling
	input	wire	rx_clk,		// Samplingzaehler Ausgang
	input	wire	reset,		// Reseteingang
	input	wire	rst,
	input	wire	enable,		// Empfaenger aktiv
	input	wire	write_enable,	// Empfangsdaten duerfen geschrieben werden

	input	wire	parity_en,	// Uebertragung mit Paritaetsbit
	input	wire	parity_sel, 	// Paritaet: 1=gerade, 0=ungerade
	input	wire [2:0] word_length,	// Wortlaenge

	output	wire	cnt_reset,	// Reset fuer Samplingzaehler
	output	reg	write,		// Schreibsignal an FIFO
	output reg [7:0] data_out,	// Empfangene Daten

	output	reg	err_fr,		// Frame Error
	output	reg	err_pa,		// Paritaets Error
	output	reg	err_da,		// Datenverlust - FIFO ist voll.

// Signale zur Fehlersuche und zum Test

	output reg [3:0] i		// Bitzaehler Rx
);

	reg		rx_in;		// Latch zum auffangen von rx_pin
	reg		rx0;		// Synchronisations Schieberegister Bit Rx0
	reg		rx1;		// Synchronisations Schieberegister Bit Rx1
	reg		sampling;	// zeigt an, ob gerade ein Wort gelesen wird
	reg		samplingp;	// Master FF dazu
//	reg	[3:0]	i;		// Bitzaehler Rx
	reg	[7:0]	sreg;		// Schieberegister Rx
	reg		rx_wr;		// FIFO schreiben starten
	reg		s0bit;		// Start Bit
	reg		dbits;		// Daten Bits empfangen
	reg		pbit;		// Paritaetsbit
	reg		sbit;		// Stopbit
	reg		rx_clko;	// rx_clk old

wire	[3:0]	wdl_1	=	word_length + 1;
wire	[3:0]	wdl_2	=	word_length + 2;
wire	[3:0]	wdl_3	=	word_length + 3;

wire		startbit	= ((i == 4'b0000) & ~rx1);
wire		databits	= ((i <= wdl_1) & (i > 0));
wire		paribit		= ((i == wdl_2) & parity_en);
wire		stopbit		= ((i >= wdl_3) && parity_en || (i >= wdl_2) && !parity_en);

assign cnt_reset = !sampling;	//Zaehler anhalten solange nichts empfangen wird.

/***

Latch-Konstrukt verursacht nicht nachvollziehbare Synthesefehler
während bitgen für den Spartan6, obwohl alle Signale und Pins
ordnungsgemäß zugeordnet und verbunden sind.

ERROR:PhysDesignRules:368 - The signal
   <subsystem_0_inst/uart_monitor/t1/r1/rx_pin> is incomplete. The signal is not
   driven by any source pin in the design.
ERROR:PhysDesignRules:10 - The network
   <subsystem_0_inst/uart_monitor/t1/r1/rx_pin> is completely unrouted.

Markus Vogt, 08.04.2011

***/
/*
 * Synchronisation des Rx Signals mit clk_div
 */
always @(posedge clk_div)
begin
	if (reset)
		begin
		rx_in	<= 1'b1;
		rx0	<= 1'b1;
		rx1	<= 1'b1;
		end
	else
		begin
		rx_in	<= rx_pin;
		rx0	<= rx_in;
		rx1	<= rx0;
		end
end


// rx1 synchron in sreg schieben
always@(posedge clk_div)
	if(rst)
	begin
		rx_clko		<= 1'b0;
		rx_wr		<= 1'b0;
		samplingp	<= 1'b0;
		i		<= 4'b0000;
		sreg		<= 0;
		err_fr		<= 1'b0;
		err_pa		<= 1'b0;
		err_da		<= 1'b0;
	end
	else if(enable)
	begin
		if(sampling && rx_clk)	// in der Mitte des Bits abtasten
		begin
			// Startbit wird Ueberprueft
			if(s0bit)
			begin   //startbit
				//rx_wr	<= 1'b0;
				i	<= 4'b0001;
			end

			// Datenbits werden ins Schieberegister geschoben
			else if(dbits)

			begin
				sreg	<= {rx1,sreg[7:1]};	// Rx in sreg schieben
				i	<= i+1;
			end

			// Wenn Paritaesbit eingestellt ist, wird dieses ueberprueft
			// und bei Erfolg dem FIFO der Wortempfang signalisiert

			//parity-bit
			//stimmt die Paritaet nicht Fehler anzeigen
			else	if (pbit)
				begin
				if (rx1 == (^sreg ~^ parity_sel))
					err_pa	<= 1'b0;
				else
					err_pa	<= 1'b1;	//Fehler melden
				i <= i+1;
				end

			// Stopbit (nur erstes von Interesse)
			// jetzt muss Stopbit anliegen
			else if(sbit)
			begin
				samplingp	<= 1'b0;	//Empfang beenden
				err_fr		<= ~rx1;	//Frame Error, wenn Rx nicht 1 ist.
				err_da		<= ~write_enable; //FIFO ist voll!
				rx_wr		<= 1'b1;	//FIFO Empfang signalisieren.
				i		<= i+1;
			end
		end
		// Warten auf das Startbit (Rx == 0).
		// Bei lowpegel wird auf Empfang geschalten und der Zaehler gestartet
		else
		begin
			rx_wr		<= 1'b0;	//FIFO schreiben beenden
			if(cnt_reset && rx1 && ~rx0)	// neg. Flanke von Rx
			begin
				samplingp	<= 1'b1;
				i		<= 4'b0000;	//Bitanzahl=0
				rx_clko		<= 1'b1;	//Start clk
			end
		end
		if (sampling)
			rx_clko	<= rx_clk;
	end

// Slave FF fuer sampling
always@(posedge clk_div)
	if(rst)
	begin
		sampling <= 1'b0;
		s0bit	 <= 1'b0;
		dbits	 <= 1'b0;
		pbit	 <= 1'b0;
		sbit	 <= 1'b0;
	end
	else 	if (samplingp & rx_clko)
		begin
			sampling <= 1'b1;
			s0bit	 <= startbit;
			dbits	 <= (databits | (!startbit & !paribit & !stopbit));
			pbit	 <= paribit;
			sbit	 <= stopbit;
		end
		else	if (!samplingp & !rx_clk)
			sampling <= 1'b0;

// Rx FIFO synchron zum FIFO clk_div schreiben
always	@(posedge clk_div)
if (rst)
begin
	write	 <= 1'b0;
	data_out <= 0;
end
else
begin
	if (/*!write && */rx_wr)
	begin
		case(word_length)
			3'b100: data_out <= {3'b000,sreg[7:3]};
			3'b101: data_out <= {2'b00, sreg[7:2]};
			3'b110: data_out <= {1'b0,  sreg[7:1]};
			3'b111: data_out <= {       sreg[7:0]};
			default:data_out <= 8'b00000000;
		endcase
		write		<= 1'b1;
	end
	else
	if (!rx_wr)
	begin
		write		<= 1'b0;
	end
end

endmodule
