/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none

module uart_mux
#(
    parameter                   CHANNELS = 2,
    parameter   [7:0]           BREAK_CHARACTER = 8'hA
)(
    input wire                  reset_n,
    input wire                  bit_clk,
    input wire  [CHANNELS-1:0]  tx_in,
    input wire  [CHANNELS-1:0]  tx_empty,
    output wire [CHANNELS-1:0]  tx_wait,
    output reg                  tx_muxed,
    output reg                  state_out
);

localparam
    STATE_TRANSMIT      = 1,
    STATE_BREAK         = 2,
    STATE_LOAD_HEADER   = 3,
    STATE_HEADER        = 4,
    STATE_WAIT          = 5;

// log2 function for bitwidth calculation
function integer log2;
input integer value;
    for (log2=0; value>0; log2=log2+1)
    value = value>>1;
endfunction

reg     [9:0]                   data_byte;
reg     [log2(CHANNELS-1)-1:0]  selected;
reg                             tx_wait_all;
reg                             bit_cnt_clear;
reg     [3:0]                   bit_cnt;
reg     [2:0]                   state;
reg     [2:0]                   next_state;
reg                             load_header;
reg                             switch;

wire    [7:0]                   header_char;

assign header_char      = 8'h41 + selected;

genvar i;

generate
for (i=0; i<CHANNELS; i=i+1)
begin:TX_WAIT_GEN
    assign tx_wait[i] = (selected != i) || tx_wait_all;
end
endgenerate

// byte buffer
always @(posedge bit_clk) 
begin
    if (!load_header)
        data_byte <= {tx_in[selected],data_byte[9:1]};
    else
        data_byte <= {1'b1,header_char[7:0],1'b0};
end

// channel selector
always @(posedge bit_clk)
begin
    if (!reset_n)
        selected <= 0;
    else if (switch)
        if (selected == (CHANNELS - 1))
            selected <= 0;
        else
            selected <= selected + 1;
end

// bit counter
always @(posedge bit_clk)
begin
    if (!reset_n)
        bit_cnt <= 4'b0000;
    else
        if (bit_cnt_clear)
            bit_cnt <= 4'b0000;
        else
            bit_cnt <= bit_cnt + 1;
end

// state machine
always @(*)
begin
    next_state      <= state;
    tx_wait_all     <= 1;
    bit_cnt_clear   <= 0;
    load_header     <= 0;
    switch          <= 0;
    tx_muxed        <= 1;
    state_out       <= 2'd0;
    case (state)
        STATE_WAIT:
        begin
            state_out           <= 2'd0;
            if (!tx_empty[selected]) 
            begin
                load_header     <= 1;
                bit_cnt_clear   <= 1;
                next_state      <= STATE_HEADER;
            end else begin
                switch          <= 1;
            end
        end
        STATE_HEADER:
        begin
            state_out           <= 2'd1;
            tx_muxed            <= data_byte[0];
            if (bit_cnt == 4'd9)
            begin
                next_state      <= STATE_TRANSMIT;
            end
        end
        STATE_TRANSMIT:
        begin
            state_out           <= 2'd2;
            if (data_byte == {1'b1,BREAK_CHARACTER,1'b0})
            begin
                next_state      <= STATE_BREAK;
                bit_cnt_clear   <= 1;
            end else begin
                tx_wait_all     <= 0;
                tx_muxed        <= tx_in[selected];
            end
        end
        STATE_BREAK:
        begin
            state_out           <= 2'd3;
            if (bit_cnt == 4'd9)
            begin
                next_state      <= STATE_WAIT;
            end
        end
        default
            next_state          <= STATE_WAIT;
    endcase
end

always @(posedge bit_clk)
begin
    if (!reset_n)
        state   <= STATE_LOAD_HEADER;
    else
        state   <= next_state;
end

endmodule
