/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 10ns / 1ns
`default_nettype none
module gain_sequencer(	clk_peri, reset, start, gain_ctrl);

	// ld-Funktion (dualer Logarithmus, aufgerundet)
	function integer ld(input integer value);
		integer i;
		begin
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction

	parameter SYSTEM_FREQUENCY 		= 25_000;		// in kHz
	parameter GAIN_INTERVAL_LENGTH 	= 256;
	localparam CC_PER_MI 			= SYSTEM_FREQUENCY / 1_000 * GAIN_INTERVAL_LENGTH;   // Verstärkung soll alle 256µs erhöht werden
	localparam CC_PER_MI_WIDTH 		= ld(CC_PER_MI);

  	input 	wire				clk_peri, reset, start;
  	output 	wire	[5:0] 			gain_ctrl;

	reg 		[CC_PER_MI_WIDTH-1:0] 	clk_div;
	reg 		[3:0] 			meas_int; // Verstärkung wird in 16 Schritten verstellt

  	reg 					en;

//synthesis translate_off
	always @ (posedge start) begin
		$display("CC_PER_MI=%d",CC_PER_MI);	
	end
//synthesis translate_on

// Flankenerkennung auf start
	reg [1:0] edgedet;
	always @ (posedge clk_peri) begin
		if (reset) begin
			edgedet <= 0;
		end else begin
			edgedet [1] <= start;
			edgedet [0] <= edgedet[1];
		end
	end
	
	wire start_edge;
	assign start_edge = (edgedet[1] & (~edgedet[0]));

	always @ (posedge clk_peri) begin
		if (reset) begin
			clk_div <= 0;
			meas_int <= 15;
    		end else begin 
      			case (en)
			        1'b0: begin
				          clk_div <= 0;
				          meas_int <= 0;
			        end
			        1'b1: begin
					if (clk_div < (CC_PER_MI-1)) begin
						clk_div <= clk_div + 1;
          				end else begin 
			            		clk_div <= 0;
						if (meas_int < 15) begin
	            					meas_int <= meas_int + 1;
						end
			          	end
        			end
      			endcase
		end
	end


  	reg [1:0] state, next_state;
	parameter IDLE = 2'b00;
	parameter RESET = 2'b01;
	parameter RUNNING = 2'b10;
  

  
	always @ (state or start_edge or meas_int) begin
		next_state = 0;
		case(state)
      		IDLE: if(start_edge) begin
				next_state = RESET;
			end else begin
				next_state = IDLE;
			end
			RESET: next_state = RUNNING;
      		RUNNING: if (start_edge) begin
				next_state = IDLE;
			end else begin
				if (meas_int < 15) begin
					next_state = RUNNING;
				end else begin
					next_state = IDLE;
				end
			end
   			default: next_state = IDLE;
   		endcase
  	end
  
	always @ (posedge clk_peri) begin:OUTPUT_LOGIC
    		if(reset) begin
      			en <= 1'b0;
      			state <= IDLE;
    		end else begin
      			state <= next_state;
      			case (state)
        			IDLE: begin
          				en <= 1'b1;
		        	end
				RESET: begin
					en <= 1'b0;
				end
        		RUNNING: begin
          			en <= 1'b1;
		        end
        		default: begin
          			en <= 1'b0;
        		end
      		endcase
    	end
  	end

											//   012765
	assign gain_ctrl = 	(meas_int == 0) 	? 6'b000000 : 				//amp = 200
				(meas_int == 1) 			? 6'b100000 : 				//amp = 250
				(meas_int == 2) 			? 6'b110000 :				//amp = 300
				(meas_int == 3)			 	? 6'b111000 :				//amp = 350
				(meas_int == 4) 			? 6'b000001 : 				//amp = 400
				(meas_int == 5) 			? 6'b100001 : 				//amp = 500
				(meas_int == 6) 			? 6'b110001 : 				//amp = 600
				(meas_int == 7) 			? 6'b111001 : 				//amp = 700
				(meas_int == 8) 			? 6'b000010 : 				//amp = 1000
				(meas_int == 9) 			? 6'b100010 : 				//amp = 1250
				(meas_int == 10)		 	? 6'b110010 : 				//amp = 1500
				(meas_int == 11)		 	? 6'b111010 : 				//amp = 1750
				(meas_int == 12)		 	? 6'b000100 : 				//amp = 2000
				(meas_int == 13)		 	? 6'b100100 : 				//amp = 2500
				(meas_int == 14)		 	? 6'b110100 : 				//amp = 3000
				(meas_int == 15)		 	? 6'b111100 : 6'b000000; 	//amp = 3500

endmodule
