/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 10ns / 1ns
`default_nettype none

module ultrasonic(	clk_peri, 
			do_peri, 
			di_peri, 
			addr_peri, 
			access_peri, 
			wr_peri, 
			reset,
			echo_intr,
			timeout_intr, 
			us_tx_direct, 
			us_rx_tl084, 
			gain_ctrl, 
			leds, 
			sevenseg_leds, 
			mvcm 
		);

	// ld-Funktion (dualer Logarithmus, aufgerundet)
	function integer ld(input integer value);
		integer i;
		begin
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction


	// Die Parameter für die Adressen und die Portbreite werden 
	// in der system.v durch jConfig eingestellt.
	parameter BASE_ADR  			= 10'h0;
	parameter SYSTEM_FREQUENCY = 25_000;		// Frequenz in kHz
	localparam MAX_ECHO_CNT 	= 11;  //nicht verändern!
	parameter ECHO_CNT      = 3;  //darf nicht größer 11 sein!

	parameter MEAS_INT_END = 29_155; // Ab 5 Metern Entfernung Timeout  (in us)
	parameter MEAS_CYC_END = 60_000; // Abstand zwischen den Messungen in us

	parameter TIMEOUT_VALUE = 1023;	// Wert der bei einem Timeout in das aktuelle Distanz-Register geschrieben wird

	parameter INITIAL_US_FREQUENCY 	= 309;  // entspricht 40,453 kHz (Beispielhaft für einen Schaltungsaufbau durch Strommessung bestimmt)
	parameter LOWEST_US_FREQUENCY = 35;   // in kHz
	localparam FREQ_REG_WIDTH = ld(SYSTEM_FREQUENCY/2/LOWEST_US_FREQUENCY);
	localparam GUARD_REG_WIDTH = ld(MEAS_INT_END);

	parameter LOWEST_US_SPEED 		= 30000;  // niedrigste einstellbare Schallgeschwindigkeit in cm pro s
	parameter INITIAL_US_SPEED 		= 34300;  // Schallgeschwindigkeit bei 20°C
	localparam INITIAL_CC_PER_CM 	= 2 * SYSTEM_FREQUENCY * 1000 / INITIAL_US_SPEED;
	localparam CC_PER_CM_WIDTH		= ld(2 * SYSTEM_FREQUENCY * 1000 / LOWEST_US_SPEED);

	parameter GAIN_INTERVAL_LENGTH 	= 256;  // Länge eines Messabschnitts in dem sich die Verstärkung nicht ändert (in us)

	parameter INITIAL_PULSE_FORM			= 16'h0505;
	parameter INITIAL_GUARD_WIDTH 			= 8;   			// Länge des Totzeitintervalls nach Aussenden des Pulses in us
	parameter INITIAL_MODE				= 3'b010;

	parameter FILTER_WIDTH			= 100;
	parameter FILTER_SAMPLE_DELAY 	= 1;
	parameter FILTER_TRESHOLD		= 75;   // sollte kleiner als FILTER_WIDTH sein!  

	localparam DIST_BASE_ADR 		= 0; 
	localparam CC_PER_CM_ADR			= ECHO_CNT;
	localparam MODE_ADR  			= ECHO_CNT + 1;
	localparam FREQ_ADR  			= ECHO_CNT + 2;
	localparam GUARD_ADR 			= ECHO_CNT + 3;
	localparam PULSE_FORM_ADR  		= ECHO_CNT + 4;




	// inputs & outputs des Systembus
	input	wire			clk_peri;  //Peripherietakt
	input	wire	[17:0]	do_peri;	// Datenbus 
	output 	wire	[17:0]	di_peri;	// Datenbus
	input	wire	 [9:0]	addr_peri;	// Adressbus
	input	wire			access_peri;	// chipselect Peripherie
	input	wire			wr_peri;	// Peripherie-Schreibzugriff
	input	wire			reset; //Reset
	output 	wire			echo_intr;
	output 	wire			timeout_intr;
	output 	wire	[5:0]	gain_ctrl;
	output 	wire	[6:0]	sevenseg_leds;
	output 	wire 			mvcm;

	wire select;  //wird vom Adressdecoder erzeugt

	output wire		[1:0] 	us_tx_direct; //US-Lautsprecher-Treiber
	input wire				us_rx_tl084; //Input vom Mikroverstärker

	output wire		[3:0] 	leds;

	reg 			[2:0] 					mode_reg;
	reg 			[FREQ_REG_WIDTH-1:0]	freq_reg;
	reg 			[GUARD_REG_WIDTH-1:0] 	guard_reg;
	reg 			[15:0] 					per_cnt_reg;
	wire 			[8:0] 					distance;

	reg				[CC_PER_CM_WIDTH-1:0] 	cc_per_cm_reg;  

	reg 			[9:0] 	distance_reg[ECHO_CNT-1:0];             //aktuell gemessene Entfernung (Messbereich 0 - 512 cm; ein Bit mehr für TIMEOUT_VALUE

	wire									fsm_enable;
	assign fsm_enable = mode_reg[0];

	wire [7:0] mvc_value;
	assign mvc_value = (distance_reg[0] < 256) ? distance_reg[0][7:0] : 8'b11111111;

	movingcoil_meter MM1 (
		.clk_peri (clk_peri),
		.reset (reset),
		.val (mvc_value),
		.pwm_out (mvcm)
	);

////////////////////////////////////////////////////////////////////////////////////////////////////
//                               Einstellung der Verstärkung                                      //
////////////////////////////////////////////////////////////////////////////////////////////////////
	reg [5:0] gain_ctrl_reg;
	genvar i;
	generate
		for (i=0; i<6; i=i+1) begin:GAINCTRL
    		assign gain_ctrl[i] = gain_ctrl_reg[i] ? 1'b0 : 1'bZ ; 
  		end
	endgenerate

	wire [5:0] gain_ctrl_w;
	reg start_g_s;
	gain_sequencer G0 (
		.clk_peri(clk_peri),
		.reset(reset),
		.start(start_g_s),
		.gain_ctrl(gain_ctrl_w)
	);
	defparam G0.SYSTEM_FREQUENCY 		= SYSTEM_FREQUENCY;
	defparam G0.GAIN_INTERVAL_LENGTH 	= GAIN_INTERVAL_LENGTH;

	always @ (posedge clk_peri) begin
		if (reset) begin
			gain_ctrl_reg <= 0;
		end else begin
			gain_ctrl_reg <= gain_ctrl_w;
		end
	end
////////////////////////////////////////////////////////////////////////////////////////////////////


	
  	parameter [3:0] IDLE            	= 4'b0000;
	parameter [3:0] TRANSMIT_PULSE  	= 4'b0001;
	parameter [3:0] GUARD_INT	 	  	= 4'b0010;
	parameter [3:0] WAIT_FOR_ECHO   	= 4'b0011;
	parameter [3:0] BUFFER_DISTANCE 	= 4'b0100;
	parameter [3:0] MEAS_ECHO_WIDTH		= 4'b0101;
	parameter [3:0] STORE_DISTANCE  	= 4'b0110;
	parameter [3:0] INC_ECHO_CNT  		= 4'b0111;
	parameter [3:0] STORE_TIMEOUT      	= 4'b1000;
	parameter [3:0] WAIT_FOR_MEAS_END  	= 4'b1001;
	  
	reg [3:0] state, next_state;

	reg [8:0] distance_buffer;
	reg [ld(ECHO_CNT-1)-1:0] echo_cnt;
	reg [ld(ECHO_CNT-1)-1:0] echo_cnt_buffer;

	  
	wire pulse_transmitted;
	wire guard_rdy;
	wire meas_int_timeout;
	wire meas_cyc_rdy;

	wire us_rx_in;
	wire us_rx;
	  
	assign us_rx_in = us_rx_tl084;

	wire us_rx_filt;

// Eingangssignal glätten //
	ma_filter MA_F (
		.clk_peri(clk_peri),
		.reset(reset),
		.in(~us_rx_in),
		.out(us_rx_filt)
	);
	defparam MA_F.SYSTEM_FREQUENCY = SYSTEM_FREQUENCY;
	defparam MA_F.SAMPLE_REG_SIZE = FILTER_WIDTH;  
	defparam MA_F.SAMPLE_DELAY = FILTER_SAMPLE_DELAY;
	defparam MA_F.FILTER_TRESHOLD = FILTER_TRESHOLD;

//	assign us_rx_filt = ~us_rx_in;

	wire us_rx_filt_e;
	edge_detector E0 (clk_peri, reset, us_rx_filt, us_rx_filt_e);  // auf Flanke triggern
	assign us_rx = us_rx_filt_e & us_rx_filt;	// nur steigende Flanke
	  
	wire wide_pulse;
	wire narrow_pulse;
	pulse_width_obs PWO0 (
		.clk_peri(clk_peri),
		.reset(reset),
		.signal_in(us_rx_filt),
		.wide_pulse(wide_pulse),
		.narrow_pulse(narrow_pulse)
	);
	defparam PWO0.SYSTEM_FREQUENCY = SYSTEM_FREQUENCY;
	defparam PWO0.REQ_PULSE_WIDTH = 240;

  	always @ (state or pulse_transmitted or guard_rdy or us_rx or meas_int_timeout or meas_cyc_rdy or wide_pulse or echo_cnt or narrow_pulse or fsm_enable) begin
	 	next_state = 0;
    	case(state)
      		IDLE: if (fsm_enable) begin
	    			next_state = TRANSMIT_PULSE;
				end else begin
					next_state = IDLE;
				end
      		TRANSMIT_PULSE:   
            	if(pulse_transmitted) begin
              		next_state = GUARD_INT;
            	end else begin
              		next_state = TRANSMIT_PULSE;
            	end
      		GUARD_INT:             
            	if(guard_rdy) begin
              		next_state = WAIT_FOR_ECHO;
            	end else begin
              		next_state = GUARD_INT;
            	end
      		WAIT_FOR_ECHO:             
            	if(us_rx) begin
              		next_state = BUFFER_DISTANCE;
            	end else begin
            		if (meas_int_timeout) begin
	              		next_state = STORE_TIMEOUT;
   	         		end else begin
   	           			next_state = WAIT_FOR_ECHO;
   	         		end
   	      		end
      		BUFFER_DISTANCE:             
            	next_state = MEAS_ECHO_WIDTH;
      		MEAS_ECHO_WIDTH: 
	    		if (wide_pulse) begin
					next_state = STORE_DISTANCE;
	    		end else begin 
					if (narrow_pulse) begin
						next_state = WAIT_FOR_ECHO;
					end else begin
						if (meas_int_timeout) begin
			      			next_state = STORE_TIMEOUT;
						end else begin
			      			next_state = MEAS_ECHO_WIDTH;
						end
					end
	    		end
      		STORE_DISTANCE: 
	    		next_state = INC_ECHO_CNT;
      		INC_ECHO_CNT:
	    		if (meas_int_timeout) begin
					if (echo_cnt == (ECHO_CNT - 1)) begin
						next_state = WAIT_FOR_MEAS_END;
					end else begin
						next_state = STORE_TIMEOUT;
					end
	    		end else begin
					if (echo_cnt == (ECHO_CNT - 1)) begin
	    				next_state = WAIT_FOR_MEAS_END;
					end else begin
						next_state = WAIT_FOR_ECHO;
					end
	    		end
      		STORE_TIMEOUT: 
            	next_state = WAIT_FOR_MEAS_END;
      		WAIT_FOR_MEAS_END:             
            	if(meas_cyc_rdy) begin
              		next_state = IDLE;
            	end else begin
              		next_state = WAIT_FOR_MEAS_END;
            	end
      		default: 
	    		next_state = IDLE;
    	endcase
  	end
  

  
// Outputs: meas_en, start_g_s, store_timeout_en, store_buffer_en_en, echo_cnt
  	reg meas_en, store_buffer_en, store_timeout_en, echo_intr_reg, timeout_intr_reg;
  
	always @ (posedge clk_peri) begin:OUTPUT_LOGIC
	    if(reset) begin
      		meas_en <= 1'b0;
			start_g_s <= 1'b0;
		  	store_buffer_en <= 1'b0;
		  	store_timeout_en <= 1'b0;
		  	distance_buffer <= 0;
		 	echo_cnt <= 0;
		  	echo_intr_reg <= 1'b0;
			timeout_intr_reg <= 1'b0;
		  	state <= IDLE;
    	end else begin
      		state <= next_state;
      		case (state)
        		IDLE: begin
          			meas_en <= 1'b0;
          			start_g_s <= 1'b0;
				  	store_buffer_en <= 1'b0;
				  	store_timeout_en <= 1'b0;
				  	echo_cnt <= 0;
			 	  	echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
        		end
        		TRANSMIT_PULSE: begin
          			meas_en <= 1'b1;
          			start_g_s <= 1'b0;
				  	store_buffer_en <= 1'b0;
				  	store_timeout_en <= 1'b0;
			 	  	echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
        		end
        		GUARD_INT: begin
          			meas_en <= 1'b1;
          			start_g_s <= 1'b0;
		  			store_buffer_en <= 1'b0;
		  			store_timeout_en <= 1'b0;
	 	  			echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
        		end
        		WAIT_FOR_ECHO: begin
          			meas_en <= 1'b1;
          			start_g_s <= 1'b1;
		  			store_buffer_en <= 1'b0;
		  			store_timeout_en <= 1'b0;
	 	  			echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
        		end
        		BUFFER_DISTANCE: begin
          			meas_en <= 1'b1;
          			start_g_s <= 1'b1;
          			distance_buffer <= distance;
		  			store_buffer_en <= 1'b0;
		  			store_timeout_en <= 1'b0;
	 	  			echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
        		end
				MEAS_ECHO_WIDTH: begin
				  	meas_en <= 1'b1;
				  	start_g_s <= 1'b1;
				  	store_buffer_en <= 1'b0;
				  	store_timeout_en <= 1'b0;
	 	  			echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
				end
				STORE_DISTANCE: begin
				  	meas_en <= 1'b1;
				  	start_g_s <= 1'b1;
				  	store_buffer_en <= 1'b1;
				  	store_timeout_en <= 1'b0;
	 	  			echo_intr_reg <= 1'b1;
					timeout_intr_reg <= 1'b0;
				end
				INC_ECHO_CNT: begin
				  	meas_en <= 1'b1;
				  	start_g_s <= 1'b1;
				  	store_buffer_en <= 1'b0;
				  	store_timeout_en <= 1'b0;
				  	echo_cnt <= echo_cnt + 1;
	 	  			echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
				end
				STORE_TIMEOUT: begin
				  	meas_en <= 1'b1;
				  	start_g_s <= 1'b0;
				  	store_buffer_en <= 1'b0;
				  	store_timeout_en <= 1'b1;
	 	  			echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b1;
				end
				WAIT_FOR_MEAS_END: begin
				  	meas_en <= 1'b0;
				  	start_g_s <= 1'b0;
				  	store_buffer_en <= 1'b0;
				  	store_timeout_en <= 1'b0;
	 	  			echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
				end
				default: begin
				  	meas_en <= 1'b0;
				  	start_g_s <= 1'b0;
				  	store_buffer_en <= 1'b0;
				  	store_timeout_en <= 1'b0;
	 	  			echo_intr_reg <= 1'b0;
					timeout_intr_reg <= 1'b0;
				end
			endcase
		end
	end


	always @ (posedge clk_peri) begin
   		if (reset) begin
			echo_cnt_buffer <= 0;
   		end else begin
       		if (store_timeout_en) begin
				echo_cnt_buffer <= echo_cnt;
       		end
   		end
	end

	// Debugausgabe auf die LEDs des Memec-Boards (Leds zeigen die 4 LSB
	// des ersten empfangenen Echos
	// 7-Segmentanzeige zeigt die Anzahl der empfangenen Echos
  	bin2seg S0 (
		.clk_peri(clk_peri),
		.reset(reset),
		.bin_in({1'b0, echo_cnt_buffer}),
		.seg_out(sevenseg_leds[6:0])
  	);
  	assign leds[2:0] = ~distance_reg[0][2:0];
	assign leds[3] = state[3];

	generate
	for (i = 0; i < ECHO_CNT; i = i + 1) begin:ECHO_STORAGE  
  		always @ (posedge clk_peri) begin
    		if (reset) begin
        		distance_reg[i] <= TIMEOUT_VALUE;
    		end else begin
        		if (echo_cnt == i) begin
            		if (store_buffer_en) begin
                		distance_reg[i] <= distance_buffer;
            		end else begin
                		if (store_timeout_en) begin
                    		distance_reg[i] <= TIMEOUT_VALUE;
                		end
            		end
        		end
    		end
  		end
	end
	endgenerate
                  

  	wire us_tx;
  	wire us_tx_pulse;
  	wire us_tx_con;
  
  	assign us_tx 			= 	mode_reg[1] ? us_tx_pulse : 
								mode_reg[2] ? us_tx_con : 1'b0;
  
  	assign us_tx_direct[0] 	= 	~us_tx;
  	assign us_tx_direct[1] 	= 	us_tx;
  
  	us_pulse_gen U0(	
    	.clk_peri (clk_peri), 
    	.reset (reset), 
    	.pulse_finished (pulse_transmitted), 
    	.clk_40khz (us_tx_pulse),
    	.tune_frequency (freq_reg),
    	.period_count (per_cnt_reg),
    	.guard_int_size (guard_reg), 
    	.guard_rdy (guard_rdy), 
    	.meas_int_timeout (meas_int_timeout), 
    	.meas_cyc_rdy(meas_cyc_rdy),
	.start()
  	);
  	defparam U0.SYSTEM_FREQUENCY = SYSTEM_FREQUENCY;
  	defparam U0.MEAS_INT_END = MEAS_INT_END;
  	defparam U0.MEAS_CYC_END = MEAS_CYC_END;
  	defparam U0.LOWEST_US_FREQUENCY = LOWEST_US_FREQUENCY;


//Start der Messung ebenfalls verzögern:

  	wire meas_en_filt;
	pulse_width_obs PWO_FILT_MEAS (
		.clk_peri(clk_peri),
		.reset(reset),
		.signal_in(meas_en),
		.wide_pulse(meas_en_filt),
		.narrow_pulse()
	);
	defparam PWO_FILT_MEAS.SYSTEM_FREQUENCY = SYSTEM_FREQUENCY;
	defparam PWO_FILT_MEAS.REQ_PULSE_WIDTH = FILTER_SAMPLE_DELAY * FILTER_TRESHOLD;
    
  	dist_meas U2(	
    	.clk_peri (clk_peri), 
    	.reset (reset), 
    	.meas_en (meas_en_filt),
    	.distance (distance),
		.cc_per_cm(cc_per_cm_reg)
  	);
  	defparam U2.SYSTEM_FREQUENCY = SYSTEM_FREQUENCY;
  	defparam U2.LOWEST_US_SPEED = LOWEST_US_SPEED;

  	us_con_gen U3(	
  	 	.clk_peri (clk_peri), 
  	 	.reset (reset), 
  	 	.clk_40khz (us_tx_con), 
  	 	.tune_frequency (freq_reg)
  	);
  	defparam U3.SYSTEM_FREQUENCY = SYSTEM_FREQUENCY;
  	defparam U3.LOWEST_US_FREQUENCY = LOWEST_US_FREQUENCY;



////////////////////////////////////////////////////////////////////////////////////////////////////
//                               Anbindung an Spartan-Bus                                         //
////////////////////////////////////////////////////////////////////////////////////////////////////

// Adressdekoder, der aus der angelegten Adresse ein chipselect erzeugt	
	pselect p1(
		.addr(addr_peri[9:4]),
		.activ_peri(access_peri),
		.select(select)
	);
	defparam p1.ADDR_WIDTH	= 6;
	defparam p1.BASE_WIDTH	= 6;
	defparam p1.BASE_ADDR	= BASE_ADR >> 4;

	wire [17:0] const_mux_out;
	wire [17:0] mux_chain_nodes_out [ECHO_CNT-1:0];

	assign	const_mux_out	= (addr_peri[3:0]==CC_PER_CM_ADR)?	cc_per_cm_reg:
		                      (addr_peri[3:0]==MODE_ADR)?		{echo_cnt,9'b000000000}|{15'h0000,mode_reg}:
		                      (addr_peri[3:0]==GUARD_ADR)?  	guard_reg:	
		                      (addr_peri[3:0]==PULSE_FORM_ADR)?  		{2'b00,per_cnt_reg}:	
		                      (addr_peri[3:0]==FREQ_ADR)?		freq_reg : 18'b0;

	generate
	for (i = 0; i < ECHO_CNT; i = i + 1) begin:READ_MUX_GEN
		if (i == 0) begin
			di_bus_mux MUX_CHAIN_FIRST_NODE ({8'b00000000, distance_reg[0]}, const_mux_out, (addr_peri[3:0]==DIST_BASE_ADR), mux_chain_nodes_out[i]);
		end else begin
			di_bus_mux MUX_CHAIN({8'b00000000, distance_reg[i]}, mux_chain_nodes_out[i-1], (addr_peri[3:0]==(DIST_BASE_ADR + i)), mux_chain_nodes_out[i]);
		end
	end
	endgenerate

	assign di_peri[17:0] = select ? mux_chain_nodes_out[ECHO_CNT-1] : 18'b0;


	// In diesem Block werden Daten vom Bus in die Register übernommen.
	always @ (posedge clk_peri)
	if(reset) begin
		mode_reg <= INITIAL_MODE;
		freq_reg <= INITIAL_US_FREQUENCY;
		guard_reg <= INITIAL_GUARD_WIDTH;
		per_cnt_reg <= INITIAL_PULSE_FORM;
		cc_per_cm_reg <= INITIAL_CC_PER_CM;
	end else begin
		if(addr_peri[3:0]==MODE_ADR && select && wr_peri)	begin			
			 mode_reg[2:0]	<= do_peri[2:0];
		end
		if(addr_peri[3:0]==GUARD_ADR && select && wr_peri)	begin			
			 guard_reg	<= do_peri[GUARD_REG_WIDTH-1:0];
		end
		if(addr_peri[3:0]==CC_PER_CM_ADR && select && wr_peri)	begin			
			 cc_per_cm_reg	<= do_peri[CC_PER_CM_WIDTH-1:0];
		end
		if(addr_peri[3:0]==PULSE_FORM_ADR && select && wr_peri)	begin			
			 per_cnt_reg	<= do_peri[15:0];
		end
		if(addr_peri[3:0]==FREQ_ADR && select && wr_peri)	begin			
			 freq_reg	<= do_peri[(FREQ_REG_WIDTH-1):0];
		end
	end



	reg echo_intr_out_reg;
	always@(posedge clk_peri)
	if(reset) begin
		echo_intr_out_reg = 1'b0;
	end	else begin
		// Registerzugriff setzt Interrupt zurück
		if(select && addr_peri[3:0]==MODE_ADR) begin
			echo_intr_out_reg = 1'b0;
		end else begin
		  	if(echo_intr_reg == 1'b1) begin
		     	echo_intr_out_reg = 1'b1;
	   		end
		end
	end
	assign echo_intr = echo_intr_out_reg;


	reg timeout_intr_out_reg;
	always@(posedge clk_peri)
	if(reset) begin
		timeout_intr_out_reg = 1'b0;
	end	else begin
		// Registerzugriff setzt Interrupt zurück
		if(select && addr_peri[3:0]==MODE_ADR) begin
			timeout_intr_out_reg = 1'b0;
		end else begin
		  	if(timeout_intr_reg == 1'b1) begin
		     	timeout_intr_out_reg = 1'b1;
	   		end
		end
	end
	assign timeout_intr = timeout_intr_out_reg;

		  
endmodule
