/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale 10ns / 1ns
`default_nettype none
module us_pulse_gen(	
		clk_peri, 
		reset, 
		start, 
		pulse_finished, 
		clk_40khz, 
		tune_frequency, 
		period_count, 
		guard_int_size, 
		guard_rdy, 
		meas_int_timeout, 
		meas_cyc_rdy
	);

	// ld-Funktion (dualer Logarithmus, aufgerundet)
	function integer ld(input integer value);
		integer i;
		begin
			for (i = 0; value > 0; i = i + 1)
				value = value >> 1;
			ld = i;
		end
	endfunction


	parameter SYSTEM_FREQUENCY = 25_000;   	// in kHz
	parameter LOWEST_US_FREQUENCY = 35;		// in kHz
	localparam FREQ_REG_WIDTH = ld(SYSTEM_FREQUENCY/2/LOWEST_US_FREQUENCY);
	localparam CC_PER_US = SYSTEM_FREQUENCY / 1_000;

	parameter MEAS_INT_END = 29_155; // Ab 5 Metern Entfernung Timeout
	parameter MEAS_CYC_END = 50_000; // Abstand zwischen den Messungen
	localparam GUARD_REG_WIDTH = ld(MEAS_INT_END);

	input	wire				clk_peri, reset;
	input 	wire	[GUARD_REG_WIDTH-1:0] 	guard_int_size;
	output 	wire				guard_rdy;
	output 	wire				meas_int_timeout;
	output 	wire				meas_cyc_rdy;
		
	input 	wire				start;
	output 	wire				clk_40khz;
	output 	wire				pulse_finished;
	input 	wire	[FREQ_REG_WIDTH-1:0] 	tune_frequency;
	input 	wire	[15:0] 			period_count;

	reg [FREQ_REG_WIDTH-1:0] clk_div;
	reg clk_40khz_reg;
	reg [8:0] per_cnt;
	reg us_en;
	reg del_en;

	//Taktteiler für Messung der Zeit in Mikrosekunden so breit dimensionieren breit genug dimensionieren
	reg [ld(CC_PER_US)-1:0] del_clk_div;
	//Mikrosekundenzähler muss maximal bis MEAS_CYC_END zählen
	reg [ld(MEAS_CYC_END)-1:0] us_cnt_reg;

	parameter [2:0] IDLE      			= 3'b000;
	parameter [2:0] TRANS_PULSE  		= 3'b001;
	parameter [2:0] PULSE_FINISHED 		= 3'b010;
	parameter [2:0] GUARD_FINISHED		= 3'b011;
	parameter [2:0] MEAS_TIMEOUT		= 3'b100;
	parameter [2:0] MEAS_CYCLE_FINISHED	= 3'b101;
	  
	reg [2:0] state, next_state;
	  
	wire guard_rdy_trg;
	wire meas_int_timeout_trg;
	wire meas_cyc_rdy_trg;
	wire pulse_finished_trg;
	  
	always @ (state or start or pulse_finished_trg or guard_rdy_trg or meas_int_timeout_trg or meas_cyc_rdy_trg) begin
		next_state = 0;
	    	case(state)
	      		IDLE: next_state = TRANS_PULSE;
	      		TRANS_PULSE: 
				if(pulse_finished_trg) begin
	              			next_state = PULSE_FINISHED;
	            		end else begin
		              		next_state = TRANS_PULSE;
		            	end
	      		PULSE_FINISHED: 
				if (guard_rdy_trg) begin
					next_state = GUARD_FINISHED;
	      			end else begin
	      		  		next_state = PULSE_FINISHED;
	      			end
	      		GUARD_FINISHED: 	
				if (meas_int_timeout_trg) begin
		      		  	next_state = MEAS_TIMEOUT;
	      			end else begin
	      		  		next_state = GUARD_FINISHED;
	      			end
	      		MEAS_TIMEOUT: 
				if (meas_cyc_rdy_trg) begin
	      		  		next_state = MEAS_CYCLE_FINISHED;
	      			end else begin
	      		  		next_state = MEAS_TIMEOUT;
	      			end
	      		MEAS_CYCLE_FINISHED:	
				next_state = IDLE;
	      		default: 
				next_state = IDLE;
	    	endcase
	end
	  
	reg pulse_finished_reg, guard_rdy_reg, meas_int_timeout_reg, meas_cyc_rdy_reg;
	assign pulse_finished = pulse_finished_reg;
	assign guard_rdy = guard_rdy_reg;
	assign meas_int_timeout = meas_int_timeout_reg;
	assign meas_cyc_rdy = meas_cyc_rdy_reg;

	always @ (posedge clk_peri) begin:OUTPUT_LOGIC
		if(reset) begin
	      		us_en <= 1'b0;
	      		del_en <= 1'b0;
	      		pulse_finished_reg = 1'b0;
	      		guard_rdy_reg = 1'b0;
	      		meas_int_timeout_reg = 1'b0;
	      		meas_cyc_rdy_reg = 1'b0;
	      		state <= IDLE;
	    	end else begin
	      		state <= next_state;
		      	case (state)
		        	IDLE: begin
		          		us_en <= 1'b0;
			       		del_en <= 1'b0;
				    	pulse_finished_reg = 1'b0;
				    	guard_rdy_reg = 1'b0;
				    	meas_int_timeout_reg = 1'b0;
				    	meas_cyc_rdy_reg = 1'b0;
		        	end
		        	TRANS_PULSE: begin
		          		us_en <= 1'b1;
		          		del_en <= 1'b0;
				    	pulse_finished_reg = 1'b0;
				    	guard_rdy_reg = 1'b0;
				    	meas_int_timeout_reg = 1'b0;
				    	meas_cyc_rdy_reg = 1'b0;
		        	end
		        	PULSE_FINISHED: begin
		          		us_en <= 1'b0;
		          		del_en <= 1'b1;
				    	pulse_finished_reg = 1'b1;
				    	guard_rdy_reg = 1'b0;
				    	meas_int_timeout_reg = 1'b0;
				    	meas_cyc_rdy_reg = 1'b0;
		        	end
		        	GUARD_FINISHED: begin
		          		us_en <= 1'b0;
		          		del_en <= 1'b1;
				    	pulse_finished_reg = 1'b1;
				    	guard_rdy_reg = 1'b1;
				    	meas_int_timeout_reg = 1'b0;
				    	meas_cyc_rdy_reg = 1'b0;
		        	end
		        	MEAS_TIMEOUT: begin
		          		us_en <= 1'b0;
		          		del_en <= 1'b1;
				    	pulse_finished_reg = 1'b1;
				    	guard_rdy_reg = 1'b1;
				    	meas_int_timeout_reg = 1'b1;
				    	meas_cyc_rdy_reg = 1'b0;
		        	end
		        	MEAS_CYCLE_FINISHED: begin
		          		us_en <= 1'b0;
		          		del_en <= 1'b1;
				    	pulse_finished_reg = 1'b1;
				    	guard_rdy_reg = 1'b1;
				    	meas_int_timeout_reg = 1'b1;
				    	meas_cyc_rdy_reg = 1'b1;
		        	end
		        	default: begin
		          		state <= IDLE;
		        	end
		      	endcase
		end
	end
	      
		 
	// Erzeugen des 40kHz-Signals für den US-Lautsprecher (50% DutyCycle)	--> aktiv, falls us_en = 1
	always @ (posedge clk_peri) begin
		if (reset) begin
	      		clk_div <= 0;
	      		clk_40khz_reg <= 0;
	      		per_cnt <= 0;
	    	end else begin 
	      		case (us_en)
	        		1'b0: begin
	          			per_cnt <= 0;
	        		end
	        		1'b1: begin
	          			if (clk_div < tune_frequency) begin
	            				clk_div <= clk_div + 1;
	          			end else begin 
	            				clk_div <= 0;
	            				clk_40khz_reg <= !clk_40khz_reg;
	            				per_cnt <= per_cnt + 1;
	          			end
	        		end
			endcase
		end
	end
	    
	always @ (posedge clk_peri) begin
		if (reset) begin
	      		del_clk_div <= 0;
	      		us_cnt_reg <= 0;
	    	end else begin 
	      		case (del_en)
	        		1'b0: begin
	          			del_clk_div <= 0;
	          			us_cnt_reg <= 0;
	        		end
	        		1'b1: begin
	          			if (del_clk_div < (CC_PER_US-1)) begin
	            				del_clk_div <= del_clk_div + 1;
	          			end else begin 
	            				del_clk_div <= 0;
	            				us_cnt_reg <= us_cnt_reg + 1;
	          			end
	        		end
	      		endcase
	    	end
	end

	assign guard_rdy_trg = us_cnt_reg == guard_int_size;
	assign meas_int_timeout_trg = us_cnt_reg == MEAS_INT_END;
	assign meas_cyc_rdy_trg = us_cnt_reg == MEAS_CYC_END;

	wire invert_pulse_trg;
	assign pulse_finished_trg = per_cnt == {period_count[7:0],1'b0};
	assign invert_pulse_trg = per_cnt >= {period_count[15:8],1'b0};

	assign clk_40khz = ~ pulse_finished_trg & (invert_pulse_trg ? ~clk_40khz_reg : clk_40khz_reg);

endmodule
