/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

`default_nettype none
module usb_tester(	
			clk,
			clk_i,
			rst,
			disconnect,
			dp,
			dn,
			funct_sel,
			execute,
			result,
			upper_nibble,
			dpid
		);
                     
input wire		clk;
output wire		clk_i;
input wire		rst;
output wire		disconnect;
inout wire		dp;
inout wire		dn;
input wire	[3:0]	funct_sel;
input wire		execute;
output reg	[6:0]	result;
output wire		upper_nibble;

// DMA Interface
wire	[17:0]	do_dma;
wire	[17:0]	di_dma;
wire	[9:0]	addr_dma;
wire		we_dma;
wire		clk_dma;
wire		en_dma;
wire		ssr_dma;

// Debug Interface
output wire		dpid;

reg	[3:0]	result_data;
reg		execute_r;
reg		execute_r1;
reg	[26:0]	cnt;
reg	[1:0]	wr_size;
reg	[9:0]	addr, addr_read, addr_write;
reg	[17:0]	di;
wire	[17:0]	do;
wire		we;

assign upper_nibble = cnt[26];
assign we = (execute_r & execute_r1);

always @(posedge clk)
	execute_r <= !execute;

always @(posedge clk)
	execute_r1 <= execute_r;

always @(execute_r)
	if (execute_r)
		addr = addr_write;
	else
		addr = addr_read;

always @(posedge clk)
	if (!rst)
		cnt <= 27'h0000000;
	else
		cnt <= cnt + 27'h0000001;

always @(funct_sel[3:1])
	case (funct_sel[3:1])
		4'h0:  addr_read  = 10'h002;   // CFG   EP 1
		4'h1:  addr_read  = 10'h004;   // CFG   EP 2
		4'h2:  addr_read  = 10'h040;   // Daten EP 2
		4'h3:  addr_read  = 10'h001;   // INTR  Reg
		default: addr_read= 10'h040;   // Daten EP 2
	endcase

always @(funct_sel[3:1])
	case (funct_sel[3:1])
		4'h0:				// EP 1 Scharf machen
			begin
				addr_write  = 10'h002;
				di          = 18'h007;
			end
		4'h1:
			begin			 // EP 2 Scharf machen 
				addr_write  = 10'h004;
				di          = 18'h001;
			end
		4'h2:
			begin			// Disconnect
				addr_write  = 10'h000;
				di          = 18'h000;
			end
		4'h3:				// Connect
			begin
				addr_write  = 10'h000;
				di          = 18'h001;
			end
	endcase

always @(do or upper_nibble or funct_sel[0])
	if (funct_sel[0])
	begin
		if (upper_nibble)
			result_data = do[7:4];
		else
			result_data = do[3:0];
	end
	else
	begin
		if (upper_nibble)
			result_data = do[15:12];
		else
			result_data = do[11:8];
	end

always  @(result_data)
              case (result_data)
                      0:   result   <= ~(7'b0111111); 
                      1:   result   <= ~(7'b0000110); 
                      2:   result   <= ~(7'b1011011); 
                      3:   result   <= ~(7'b1001111); 
                      4:   result   <= ~(7'b1100110); 
                      5:   result   <= ~(7'b1101101); 
                      6:   result   <= ~(7'b1111101); 
                      7:   result   <= ~(7'b0000111); 
                      8:   result   <= ~(7'b1111111); 
                      9:   result   <= ~(7'b1101111); 
                   4'hA:   result   <= ~(7'b1110111); 
                   4'hb:   result   <= ~(7'b1111100); 
                   4'hC:   result   <= ~(7'b0111001); 
                   4'hd:   result   <= ~(7'b1011110); 
                   4'hE:   result   <= ~(7'b1111001); 
                   4'hF:   result   <= ~(7'b1110001); 
                default:   result   <= 7'b1111111;
              endcase


usb1_core core (
			.clk(		clk		),
			.clk_i(		clk_i		),
			.rst(		rst		),
			// USB PHY Interface
			.dp(		dp		),
			.dn(		dn		),
			// Blockram Interface
			.di_dma(	di_dma		),
			.addr_dma(	addr_dma	),
			.en_dma(	en_dma		),
			.we_dma(	we_dma		),
			.ssr_dma(	ssr_dma		),
			.clk_dma(	clk_dma		),
			.do_dma(	do_dma		),
			// Global Config Register Interface
			.disconnect(	disconnect	),
			.interrupt(			),
			// Debug Interface
			.funct_adr(			),
			.dpid(	dpid			)
		); 

ram_18_prim	DPRAM(
			.DIA(		di_dma[17:0]	),
			.ADDRA(		addr_dma	),
 			.ENA(		en_dma		),
			.WEA(		we_dma		),
			.SSRA(		ssr_dma		),
			.CLKA(		clk_dma		),
			.DOA(		do_dma[17:0]	),
			.DIB(		di[17:0]	),
			.ADDRB(		addr		),
			.ENB(		1'b1		),
			.WEB(		we		),
			.SSRB(		1'b0		),
			.CLKB(		clk		),
			.DOB(		do[17:0]	)
		);

`ifdef toolchain_ise
defparam DMARAM.ram_18_inst.INIT_07 = 256'h0000000000000000000000000000010100020205070101000281050700ff01ff;
defparam DMARAM.ram_18_inst.INIT_06 = 256'h02000004090040000101002002090100000000105678123440ff00ff01000112;
defparam DMARAM.ram_18_inst.INIT_02 = 256'h0000000000000000000000000000000000000000000000000000000000030201;
defparam DMARAM.ram_18_inst.INIT_00 = 256'h0000000000000000000000000000000000000000000000000000000300000001;

// synopsys translate_off
defparam	DPRAM.ram_18_inst.INIT_00 = 256'h0000000000000000000000000000000000000000000000010000000300000001;
defparam	DPRAM.ram_18_inst.INIT_06 = 256'h02000004090040000101002002090100000000105678123440FF00FF01000112;
defparam	DPRAM.ram_18_inst.INIT_07 = 256'h0000000000000000000000000000010100010205070101000181050700FF01FF;
// synopsys translate_on
`endif



`ifdef toolchain_ise

// synopsys translate_off

defparam DPRAM.ram_18_inst.INIT_07 = 256'h0000000000000000000000000000010100010205070101000181050700ff01ff;
defparam DPRAM.ram_18_inst.INIT_06 = 256'h02000004090040000101002002090100000000105678123440ff00ff01000112;
defparam DPRAM.ram_18_inst.INIT_00 = 256'h0000000000000000000000000000000000000000000000010000000300000001;

// synopsys translate_on

`endif

`ifdef toolchain_quartus

defparam DMARAM.init_file = "DMARAM.ram_18_inst.mif";

// synopsys translate_off

defparam DPRAM.init_file = "DPRAM.ram_18_inst.mif";

// synopsys translate_on

`endif

`ifdef toolchain_lattice

defparam DMARAM.INITVAL_07 = "0x0000000000000000000000000000000000000101000020020500701001000028100507000FF001FF" ;
defparam DMARAM.INITVAL_06 = "0x002000000400900040000010100020002090010000000000100567801234040FF000FF0010000112" ;
defparam DMARAM.INITVAL_02 = "0x00000000000000000000000000000000000000000000000000000000000000000000000000300201" ;
defparam DMARAM.INITVAL_00 = "0x00000000000000000000000000000000000000000000000000000000000000000000030000000001" ;

// synopsys translate_off

defparam DPRAM.INITVAL_07 = "0x0000000000000000000000000000000000000101000010020500701001000018100507000FF001FF" ;
defparam DPRAM.INITVAL_06 = "0x002000000400900040000010100020002090010000000000100567801234040FF000FF0010000112" ;
defparam DPRAM.INITVAL_00 = "0x00000000000000000000000000000000000000000000000000000000000100000000030000000001" ;

// synopsys translate_on

`endif
	
            
endmodule
           
