/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  USB DMA Blockram des SpartanMC                             ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`include "usb11/src/usb1_defines.vh"
`default_nettype none

module blockramusb(	clk, rst,
			// Packet Assembler Interface
			rx_data_valid,
			rx_data_done,
			rx_data_st,
			zero_length,
			// Packet Disassembler Interface
			send_data,
			rd_next,
			tx_valid,
			tx_data_st,
			// Protocol Engine
			tx_dma_en,
			rx_dma_en,
			idma_done,
			transfer_finished,
			ep_sel,
			funct_adr,
			state,
			// Config Register Interface
			get_cfg,
			cfg_valid,
			cfg_r,
			// Global Config Register Interface
			disconnect,
			interrupt,
			// Blockram Interface
			do_peri,
			di_peri,
			addr_peri,
			addr_high,
			mem_access,
			store_access,
			clkb_data,
			send_stall,
			halt,
			configured,
			addressed
	);

output wire	send_stall;
output wire	halt;
output wire	configured;
output wire	addressed;

parameter	IR_SOURCES	 = 2;
parameter	DOUBLE_BUFFERING = 0;
parameter	DMABASE_ADR	 = 18'h19c00;
parameter 	NOGAP		 = 0;
input wire		clk;
input wire		rst;

// Packet Assembler Interface
input wire		rx_data_valid;
input wire		rx_data_done;
input wire	[15:0]	rx_data_st;
output reg		zero_length;

// Packet Disassembler Interface
output reg		send_data;
input wire		rd_next;
input wire		tx_valid;
output wire	[7:0]	tx_data_st;

// Protocol Engine
input wire		tx_dma_en;
input wire		rx_dma_en;
output reg		idma_done;
input wire	[3:0]	ep_sel;
output wire	[6:0]	funct_adr;
output reg	[6:0]	state;
input wire		transfer_finished;

//Config Register Interface
input wire		get_cfg;
output reg		cfg_valid;
output reg	[13:0]	cfg_r;
						
// Global Config Register Interface
output reg		disconnect;
output reg	[IR_SOURCES-1:0]	interrupt;

// Blockram Interface
input wire	[17:0]	do_peri;
output wire	[17:0]	di_peri;
input wire	[9:0]	addr_peri;
input wire	[7:0]	addr_high;
input wire		mem_access;
input wire		store_access;
input wire		clkb_data;


parameter [6:0]
		IDLE		=	7'b0000001,
		WRITE_DATA	=	7'b0000010,
		SET_STS		=	7'b0000100,
		READ_DATA	=	7'b0001000,
		GET_CFG1	=	7'b0010000,
		GET_CFG2	=	7'b0100000,
		SET_CFG		=	7'b1000000;


reg	[6:0]	next_state;


wire	[17:0]	do_dma;
wire	[7:0]	do_dma_h;	// oberen 8 Datenbits fuer NOGAP = 0 oder 1
reg	[17:0]	di_dma;
wire	[17:0]	di_dma_h8;	// Nur die oberen 8 Bit fuer NOGAP = 0 oder 1
wire	[17:0]	di_dma_16;	// alle 16 Datenbits fuer NOGAP = 0 oder 1
reg	[9:0]	addr_dma;
reg		we_dma;
wire		clk_dma;
wire		en_dma = 1'b1;
wire		ssr_dma = 1'b0;

wire	[9:0]  	ram_start;
wire	[6:0]	ctrl_size;
reg		rx_data_valid_r;
reg	[6:0]	rx_cnt_r;
reg	[6:0]	tx_cnt_r;
reg		tx_cnt_high;
reg		ire_ep0;
reg	[IR_SOURCES-1:0]	ir_return;
reg		ir_enable;

wire		ctrl_ep;
wire	[6:0]	tx_size;
wire	[9:0]	cfg_addr;
wire	[9:0]	sts_addr;
wire	[9:0]	read_addr;
wire	[9:0]	write_addr;
wire	[9:0]	mem_addr;
reg		global_cfg_valid;



// CTRL Interface
reg	[7:0]	hdr0, hdr1, hdr2, hdr3, hdr4, hdr5, hdr6, hdr7;
reg		hdr_done;
reg		get_header;
reg		ctrl_in;
wire		ARMED_ep = cfg_r[13];

assign		clk_dma	  = !clk;
assign		ctrl_ep   = (ep_sel == 4'b0000);
assign 		tx_size   = cfg_r[6:0];
assign 		cfg_addr  = {6'h00, ep_sel[3:0], 1'b0};
assign 		sts_addr  = {6'h00, ep_sel[3:0], 1'b1};
assign 		mem_addr  = tx_cnt_r[5:1] + ram_start;

// Adressleitungen in Abhaengigkeit des double bufferings **********************
// Puffer beginnend bei 0x80 mit Endpunkt 0
/*generate if (DOUBLE_BUFFERING == 0) begin
	wire	[4:0]	ep_sel_plusfour = ep_sel[3:0] + 3;
	assign read_addr  = {ep_sel_plusfour,        tx_cnt_r[5:1]};
	assign write_addr = {ep_sel_plusfour,        rx_cnt_r[5:1]};
end else begin
	wire		bufsel   = cfg_r[ 7];
	wire	[3:0]	ep_sel_plustwo  = ep_sel[3:0] + 2;
	assign read_addr  = {ep_sel_plustwo, bufsel, tx_cnt_r[5:1]};
	assign write_addr = {ep_sel_plustwo, bufsel, rx_cnt_r[5:1]};
end endgenerate*/
wire	[4:0]	ep_sel_plusfour = ep_sel[3:0] + 4;
wire	[3:0]	ep_sel_plustwo  = ep_sel[3:0] + 2;
wire		bufsel		= cfg_r[7];

assign read_addr  = (DOUBLE_BUFFERING == 0)?
		{ep_sel_plusfour, tx_cnt_r[5:1]}:
		{ep_sel_plustwo, bufsel, tx_cnt_r[5:1]};
assign write_addr = (DOUBLE_BUFFERING == 0)?
		{ep_sel_plusfour, rx_cnt_r[5:1]}:
		{ep_sel_plustwo, bufsel, rx_cnt_r[5:1]};

// hier muss NOGAP aber nur für Datenzugriffe (!ctrl_ep & (state == READ_DATA)) eingefuegt werden
// und Bereitstellung der Bits fuer den Zustand WRITE_DATA
generate
	if (NOGAP==0) begin: DATA_W_GAP_RD
		assign do_dma_h  = do_dma[16:9];
		assign di_dma_h8 = {1'b0, rx_data_st[7:0], 9'h000 };
		assign di_dma_16 = {1'b0, rx_data_st[15:8], 1'b0, rx_data_st[7:0]};
	end else begin: DATA_WO_GAP_RD
		wire		tx16bit;
		wire		rx16bit;
//		assign tx16bit   = (!ctrl_ep & (state == READ_DATA) & (rd_next | tx_dma_en));
		assign tx16bit   = (!ctrl_ep & (state == READ_DATA));
//		assign rx16bit   = (!ctrl_ep & (state == WRITE_DATA) & (rx_data_valid_r | rx_dma_en));
		assign rx16bit   = (!ctrl_ep & (state == WRITE_DATA) & rx_data_valid & ARMED_ep);
		assign do_dma_h  = tx16bit ? do_dma[15:8] : do_dma[16:9];
		assign di_dma_h8 = rx16bit ? {2'b00, rx_data_st[7:0], 8'h00 } :
					     {1'b0, rx_data_st[7:0], 9'h000 };
		assign di_dma_16 = rx16bit ? {2'b00, rx_data_st[15:8], rx_data_st[7:0]} :
					     {1'b0, rx_data_st[15:8], 1'b0, rx_data_st[7:0]};
	end
endgenerate

assign tx_data_st = tx_cnt_high  ? do_dma [7:0] : do_dma_h[7:0];


always @(posedge clk)
	rx_data_valid_r <= rx_data_valid;
	
always @(negedge clk)
	tx_cnt_high <= tx_cnt_r[0];

//////////////////////////////////////////////////////////////////////
//
// Global Config Register Latch
//

wire		is_idle = state[0];

always @(posedge clk)
	if(!rst)
		global_cfg_valid <= 1'b0;
	else
		global_cfg_valid <= is_idle;

always @(global_cfg_valid or is_idle or do_dma)
	if(global_cfg_valid & is_idle)
	begin
		disconnect = do_dma[17];
		ire_ep0    = do_dma[16];
		ir_return  = do_dma[IR_SOURCES-1:0];
	end

//////////////////////////////////////////////////////////////////////
//
// Interrupt FipFlops
//

genvar i;
generate
begin:IR_REG
	for (i=0; i<IR_SOURCES; i=i+1)
	begin:IR_FF
		always @(posedge clk)
			if(ir_return[i])
				interrupt[i] <= 0;
			else if ((ep_sel[3:0] == i) & ir_enable)
				interrupt[i] <= 1;
	end
end
endgenerate

//////////////////////////////////////////////////////////////////////
//
// RX Byte Counter
//

always @(posedge clk)
	if(!rst)
		rx_cnt_r <= 7'h00;
	else
		if (rx_dma_en)
			rx_cnt_r <= 7'h00;
		else
			if (rx_data_valid_r)
				rx_cnt_r <= rx_cnt_r + 7'h01;


//////////////////////////////////////////////////////////////////////
//
// TX Byte Counter
//

always @(posedge clk)
	if(!rst)
		tx_cnt_r <= 7'h00;
	else
		if (tx_dma_en)
			tx_cnt_r <= 7'h00;
		else
			if (rd_next)
				tx_cnt_r <= tx_cnt_r + 7'h01;


//////////////////////////////////////////////////////////////////////
//
// Request Memory Registers
//


always @(posedge clk)
	if(!rst)
		begin
			hdr0 <= 8'h00;
			hdr1 <= 8'h00;
			hdr2 <= 8'h00;
			hdr3 <= 8'h00;
			hdr4 <= 8'h00;
			hdr5 <= 8'h00;
			hdr6 <= 8'h00;
			hdr7 <= 8'h00;
		end
	else
		if(get_header & rx_data_valid)
			case(rx_cnt_r)
				7'h00:	hdr0 <= rx_data_st[7:0];
				7'h01:	hdr1 <= rx_data_st[7:0];
				7'h02:	hdr2 <= rx_data_st[7:0];
				7'h03:	hdr3 <= rx_data_st[7:0];
				7'h04:	hdr4 <= rx_data_st[7:0];
				7'h05:	hdr5 <= rx_data_st[7:0];
				7'h06:	hdr6 <= rx_data_st[7:0];
				7'h07:	hdr7 <= rx_data_st[7:0];
			endcase

//////////////////////////////////////////////////////////////////////
//
// Config Register
//

always @(posedge cfg_valid or negedge rst)
	if(!rst)
		cfg_r = 14'h0000;
	else
		if (ctrl_ep)
		begin
			cfg_r = {2'b10, `CTRL | `BULK, 2'b00, ctrl_size[5:0]};	// cfg_r hat nur 14 Bit und nicht 18!
			// synopsys translate_off
			$display("Read CTRL cfg.reg: %h", cfg_r);
			// synopsys translate_on
		end
		else
		begin
			cfg_r = {do_dma[17:11], do_dma[6:0]};
			// synopsys translate_off
			$display("Read EP %d cfg.reg: %h", ep_sel, cfg_r);
			// synopsys translate_on
		end
		
//////////////////////////////////////////////////////////////////////
//
// Status and Config Register FSM
//

always @(posedge clk) 
	if(!rst)
		state <= IDLE;
	else
		state <= next_state;

always @(state or rx_dma_en or tx_dma_en or ep_sel or rx_data_done or
	 tx_cnt_r or get_cfg or transfer_finished or tx_size)
begin
	next_state = state;			// Default don't change current state

	case(state)
	IDLE:
		begin
			if(rx_dma_en)
				next_state = WRITE_DATA;
			if(tx_dma_en)
				next_state = READ_DATA;		
			if(get_cfg)
				next_state = GET_CFG1;
			if(transfer_finished)
				next_state = SET_STS;
		end
	
	GET_CFG1:	next_state = GET_CFG2;
	GET_CFG2:	next_state = IDLE;
	SET_STS:	next_state = SET_CFG;
	SET_CFG:	next_state = IDLE;

	WRITE_DATA:
			if(rx_data_done)
					next_state = IDLE;
	
	READ_DATA:
			if (tx_cnt_r >= tx_size)
					next_state = IDLE;
	default:	next_state = IDLE;
	endcase
end

always @(state or ctrl_ep or write_addr or read_addr or 
	 mem_addr or cfg_addr or sts_addr or tx_cnt_r or 
	 rx_cnt_r or tx_size or rx_data_st or rx_data_valid or
	 cfg_r or do_dma or ep_sel or ARMED_ep)
begin

	we_dma		= 1'b0;			
	send_data	= 1'b0;			
	idma_done	= 1'b1;
	cfg_valid	= 1'b0;
	zero_length	= 1'b0;
	get_header	= 1'b0;
	hdr_done	= 1'b0;
	ctrl_in		= 1'b0;
	addr_dma	= 10'h20; //Globales Konfigurationsregister
	di_dma		= 18'h00000;
	ir_enable	= 1'b0;
			
	case(state)								

	GET_CFG1:		addr_dma = cfg_addr;
	GET_CFG2:
			begin		
				addr_dma = cfg_addr;
				cfg_valid	= 1'b1;
			end
	SET_STS:
			begin
				we_dma    = 1'b1;			
				addr_dma  = sts_addr;
				di_dma= {11'h0, rx_cnt_r};
				if (ctrl_ep)
					ir_enable = ire_ep0;
				else
					ir_enable = cfg_r[12];
				// synopsys translate_off
				$display("Set EP %d statusreg.: %h", ep_sel, di_dma);
				// synopsys translate_on
			end
	SET_CFG:	begin
				we_dma   = 1'b1;			
				addr_dma = cfg_addr;
				// ARMED-Flag hier zurueckgesetzt
				di_dma   = {1'b0, cfg_r[12:7], 4'b0, cfg_r[6:0]};
				// synopsys translate_off
				$display("Set EP %d cfg.reg.: %h", ep_sel, di_dma);
				// synopsys translate_on
			end
	WRITE_DATA:		
			begin
				if(ctrl_ep)
				begin
					get_header = 1'b1;
						
					if(rx_cnt_r >= 7'h7)
						hdr_done	= 1'b1;
				end
			
				addr_dma = write_addr;
				if(!rx_cnt_r[0])
					di_dma = di_dma_h8;	// Datebits mit NOGAP = 0 oder 1
				else
					di_dma = di_dma_16;	// Datebits mit NOGAP = 0 oder 1

				we_dma = rx_data_valid & ARMED_ep;
			end
	READ_DATA:
			begin
				idma_done = 1'b0;
				
				if(tx_size == 0)
				begin
					send_data = 1'b1;
					zero_length = 1'b1;
				end
				else
					if(tx_cnt_r < tx_size)
						send_data = 1'b1;
								
				if(ctrl_ep)
				begin
					ctrl_in = 1'b1;
					addr_dma = mem_addr;
				end
				else
					addr_dma = read_addr;
			end
	endcase
end

// synopsys translate_off
always @(posedge clk_dma)
	if(we_dma)
		$display("MEM[%04h] <= %05h", addr_dma, di_dma);
// synopsys translate_on
	
usb1_ctrl	ctrl(	
			.clk(		clk		),
			.rst(		rst		),
            		.ctrl_setup(	1'b0		),
            		.ctrl_in(	ctrl_in		),
            		.ctrl_out(	1'b0		),
			.send_stall(	send_stall	),
			.funct_adr(	funct_adr	),
			.configured(	configured	),
			.halt(		halt		),
			.addressed(	addressed	),
		  	.hdr0(		hdr0		),
			.hdr1(		hdr1		),
			.hdr2(		hdr2		),
			.hdr3(		hdr3		),
			.hdr4(		hdr4		),
			.hdr5(		hdr5		),
			.hdr6(		hdr6		),
			.hdr7(		hdr7		),
			.hdr_done(	hdr_done	),
			.rom_start_d(	ram_start	),
			.ctrl_size(	ctrl_size	)
		);


// ****************************************************************
// Initialisierung fuer alle Speicher (Speicherinhalt beim Start im VSIM)
/*
// synopsys translate_off
`include "usb11/src/usbramsim.v"
// synopsys translate_on
*/
// ****************************************************************


// ****************************************************************
// DMA Puffer fuer den USB Modul und dem Dekoder
// Dekodierung des DMA-Speicherblocks
wire            activ_dma0;

pselect dmasel0 (
			.addr(		addr_high[7:0]	),
			.activ_peri(	mem_access	),
			.select(	activ_dma0	)
                );

defparam dmasel0.ADDR_WIDTH = 8;
defparam dmasel0.BASE_WIDTH = 8;
defparam dmasel0.BASE_ADDR  = DMABASE_ADR >> 10;        // Oberen 8 Bit der Speicherblockadresse

/*  NOGAP muss nur fuer die Datenbloecke an di_dma und do_dma realisiert werden
wire [17:0] tmp_di_peri;
wire [17:0] tmp_do_peri;

generate 
	if (NOGAP==0) begin: DATA_W_GAP
		assign di_peri = tmp_di_peri;
		assign tmp_do_peri = do_peri;
	end else begin: DATA_WO_GAP
		// Beim Transfer der Daten zum Spartan muss die Lücke entfernt werden
		assign di_peri = {tmp_di_peri[17],tmp_di_peri[8],tmp_di_peri[16:9],tmp_di_peri[7:0]};
		// Beim Transfer der Daten zum USB-DMA muss die Lücke hinzugefügt werden
		assign tmp_do_peri = {do_peri[17],do_peri[15:8],do_peri[16],do_peri[7:0]};
	end
endgenerate
*/

ram_18_prim		#(
			.WRITE_MODE_A(	"READ_FIRST"			),
			.INIT_A(	18'h00000			),
			.SRVAL_A(	18'h00000			),
			.WRITE_MODE_B(	"READ_FIRST"			),
			.INIT_B(	18'h00000			),
			.SRVAL_B(	18'h00000			),
			.SIM_COLLISION_CHECK("NONE"			)
			)
DMARAM00		(
                        .DIA(           di_dma[17:0]                    ),
                        .ADDRA(         addr_dma[9:0]                   ),
                        .ENA(           1'b1                            ),      // en_dma
                        .WEA(           we_dma                          ),
                        .SSRA(          1'b0                            ),      // ssr_dma
                        .CLKA(          clk_dma                         ),
                        .DOA(           do_dma[17:0]                    ),
                        .DIB(           do_peri[17:0]			),
                        .ADDRB(         addr_peri[9:0]			),
                        .ENB(           1'b1				),
                        .WEB(           store_access & activ_dma0	),
                        .SSRB(          ! activ_dma0                    ),
                        .CLKB(          clkb_data			),
                        .DOB(           di_peri[17:0]			)
                        );

// ****************************************************************


endmodule
