/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  Protocol Layer                                             ////
////  This block is typically referred to as the SEI in USB      ////
////  Specification. It encapsulates the Packet Assembler,       ////
////  disassembler, protocol engine and internal DMA             ////
////                                                             ////
////  Author: Rudolf Usselmann                                   ////
////          rudi@asics.ws                                      ////
////                                                             ////
////                                                             ////
////  Based on: http://www.opencores.org/cores/usb1_fucnt/       ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`include "usb11/src/usb1_defines.vh"
`default_nettype none

module usb1_pl(	clk, rst,

		// UTMI Interface
		rx_data, rx_valid, rx_active, rx_err,
		tx_data, tx_valid, tx_valid_last, tx_ready,
		tx_first, tx_valid_out,
		// EP Interface
		ep_sel, 
		nse_err,	// no such endpoint error
		fsel,		// This Function is selected
		match_o,

		do_peri,
		di_peri,
		addr_peri,
		addr_high,
		mem_access,
		store_access,
		clkb_data,

		// Global Config Register Interface
		disconnect, interrupt,
		// Debug Interface
		state_pe,
		state,
		get_cfg,
		transfer_finished,
		tx_dma_en,
		rx_dma_en,
		funct_adr,
		dpid,
		// Status
		int_crc16_set,	// Set CRC16 error interrupt
		int_to_set,	// Set time out interrupt
		int_seqerr_set, // Set PID sequence error interrupt
		configured,
		halt,
		send_stall,
		addressed
	);

parameter	IR_SOURCES	 = 2;
parameter	DOUBLE_BUFFERING = 0;
parameter	DMABASE_ADR	 = 18'h19c00;
parameter 	NOGAP		 = 0;

// UTMI Interface
input wire			clk, rst;
input wire	[7:0]	rx_data;
input wire			rx_valid, rx_active, rx_err;
output wire	[7:0]	tx_data;
output wire			tx_valid;
output wire			tx_valid_last;
input wire			tx_ready;
output wire			tx_first;
input wire			tx_valid_out;

// Endpoint Interfaces
output wire	[3:0]	ep_sel;		// Endpoint Number
output wire			nse_err;	// no such endpoint error
output wire			fsel;		// This Function is selected
output wire			match_o;

// Global Config Register Interface
output wire						disconnect;
output wire	[IR_SOURCES-1:0]	interrupt;


// SpartanMC Bus
input wire	[17:0]	do_peri;
output wire	[17:0]	di_peri;
input wire	[9:0]	addr_peri;
input wire	[7:0]	addr_high;
input wire			mem_access;
input wire			store_access;
input wire			clkb_data;


//Debug Interfaces
output wire	[6:0]	state_pe;
output wire	[6:0]	state;
output wire	[6:0]	funct_adr;
output wire			dpid;
output wire			int_crc16_set;	// Set CRC16 error interrupt
output wire			int_to_set;	// Set time out interrupt
output wire			int_seqerr_set; // Set PID sequence error interrupt


output wire	send_stall;
output wire	configured;
output wire	halt;
output wire	addressed;
///////////////////////////////////////////////////////////////////
//
// Local Wires and Registers
//

// Packet Disassembler Interface
wire		pid_OUT, pid_IN, pid_SOF, pid_SETUP;
wire		pid_DATA0, pid_DATA1, pid_DATA2, pid_MDATA;
wire		pid_ACK, pid_NACK, pid_STALL, pid_NYET;
wire		pid_PRE, pid_ERR, pid_SPLIT, pid_PING;
wire	[6:0]	token_fadr;
wire		token_valid;
wire	[10:0]	frame_no;
reg	[15:0]	rx_data_st;
wire	[15:0]	rx_data_st_d;
wire		rx_data_valid;
wire		rx_data_done;
wire		crc16_err;
wire		rx_seq_err;
wire		zero_length;

// Packet Assembler Interface
wire		send_token;
wire	[1:0]	token_pid_sel;
wire		send_data;
wire	[1:0]	data_pid_sel;
wire	[7:0]	tx_data_st;
wire	[7:0]	tx_data_st_o;
wire		rd_next;

// BlockRam Interface
output wire		transfer_finished;
output wire		tx_dma_en;	// Allows for data to be retrieved
output wire		rx_dma_en;	// Allows the data to be stored

wire		abort;		// Abort Transfer (time_out, crc_err or rx_error)
wire		idma_done;	// DMA is done
wire		pid_cs_err;	// pid checksum error
wire		crc5_err;	// crc5 error
wire	[31:0]	frm_nat;


// Config Register Interface
output wire		get_cfg;
wire		 	cfg_valid;
wire	[13:0]	cfg_r;

// Local signals
wire		pid_bad;

reg		hms_clk;	// 0.5 Micro Second Clock
reg	[4:0]	hms_cnt;		
reg	[10:0]	frame_no_r;	// Current Frame Number register
wire		frame_no_we;
reg	[11:0]	sof_time;	// Time since last sof
reg		clr_sof_time;

reg		frame_no_we_r;

///////////////////////////////////////////////////////////////////
//
// Misc Logic
//

// PIDs we should never receive
assign pid_bad = pid_ACK | pid_NACK | pid_STALL | pid_NYET | pid_PRE |
		 pid_ERR | pid_SPLIT |  pid_PING;

assign match_o = !pid_bad & token_valid & !crc5_err;

// Frame Number (from SOF token)
assign frame_no_we = token_valid & !crc5_err & pid_SOF;

always @(posedge clk)
	frame_no_we_r <= #1 frame_no_we;

always @(posedge clk)
	if(!rst)
		frame_no_r <= #1 11'h0;
	else
		if(frame_no_we_r)
			frame_no_r <= #1 frame_no;

//SOF delay counter
always @(posedge clk)
	clr_sof_time <= #1 frame_no_we;

always @(posedge clk)
	if(clr_sof_time)
		sof_time <= #1 12'h0;
	else
		if(hms_clk)
			sof_time <= #1 sof_time + 12'h1;

assign frm_nat = {4'h0, 1'b0, frame_no_r, 4'h0, sof_time};

// 0.5 Micro Seconds Clock Generator
always @(posedge clk)
	if(!rst)
		hms_cnt <= #1 5'h0;
	else
		if(hms_clk | frame_no_we_r)
			hms_cnt <= #1 5'h0;
		else
			hms_cnt <= #1 hms_cnt + 5'h1;

always @(posedge clk)
	hms_clk <= #1 (hms_cnt == `USBF_HMS_DEL);

always @(posedge clk)
	rx_data_st <= rx_data_st_d;

///////////////////////////////////////////////////////////////////

// This function is addressed
assign fsel = (token_fadr == funct_adr);

///////////////////////////////////////////////////////////////////
//
// Module Instantiations
//

//Packet Decoder
usb1_pd	pd(
		.clk(			clk		),
		.rst(			rst		),
		.rx_data(		rx_data		),
		.rx_valid(		rx_valid	),
		.rx_active(		rx_active	),
		.rx_err(		rx_err		),
		.pid_OUT(		pid_OUT		),
		.pid_IN(		pid_IN		),
		.pid_SOF(		pid_SOF		),
		.pid_SETUP(		pid_SETUP	),
		.pid_DATA0(		pid_DATA0	),
		.pid_DATA1(		pid_DATA1	),
		.pid_DATA2(		pid_DATA2	),
		.pid_MDATA(		pid_MDATA	),
		.pid_ACK(		pid_ACK		),
		.pid_NACK(		pid_NACK	),
		.pid_STALL(		pid_STALL	),
		.pid_NYET(		pid_NYET	),
		.pid_PRE(		pid_PRE		),
		.pid_ERR(		pid_ERR		),
		.pid_SPLIT(		pid_SPLIT	),
		.pid_PING(		pid_PING	),
		.pid_cks_err(		pid_cs_err	),
		.token_fadr(		token_fadr	),
		.token_endp(		ep_sel		),
		.token_valid(		token_valid	),
		.crc5_err(		crc5_err	),
		.frame_no(		frame_no	),
		.rx_data_st(		rx_data_st_d	),
		.rx_data_valid(		rx_data_valid	),
		.rx_data_done(		rx_data_done	),
		.crc16_err(		crc16_err	),
		.seq_err(		rx_seq_err	)
	);

// Packet Assembler
usb1_pa	pa(	
		.clk(			clk		),
		.rst(			rst		),
		.tx_data(		tx_data		),
		.tx_valid(		tx_valid	),
		.tx_valid_last(		tx_valid_last	),
		.tx_ready(		tx_ready	),
		.tx_first(		tx_first	),
		.send_token(		send_token	),
		.token_pid_sel(		token_pid_sel	),
		.send_data(		send_data	),
		.data_pid_sel(		data_pid_sel	),
		.tx_data_st(		tx_data_st_o	),
		.rd_next(		rd_next		),
		.zero_length(		zero_length	)
	);

blockramusb #(
		.IR_SOURCES(		IR_SOURCES	),
		.DOUBLE_BUFFERING(	DOUBLE_BUFFERING),
		.DMABASE_ADR(		DMABASE_ADR	),
		.NOGAP(			NOGAP		))
br (
		.clk(			clk		),
		.rst(			rst		),
		.rx_data_valid(		rx_data_valid	),
		.rx_data_done(		rx_data_done	),
		.rx_data_st(		rx_data_st	),
		.zero_length(		zero_length	),
		.send_data(		send_data	),
		.rd_next(		rd_next		),
		.tx_valid(		tx_valid	),
		.tx_data_st(		tx_data_st_o	),
		.tx_dma_en(		tx_dma_en	),
		.rx_dma_en(		rx_dma_en	),
		.idma_done(		idma_done	),
		.transfer_finished(	transfer_finished),
		.ep_sel(		ep_sel		),
		.funct_adr(		funct_adr	),
		.state(			state		),
		.get_cfg(		get_cfg		),
		.cfg_valid(		cfg_valid	),
		.cfg_r(			cfg_r		),
		.disconnect(		disconnect	),
		.interrupt(		interrupt	),

		.send_stall(		send_stall	),
		.halt(			halt		),
		.configured(		configured	),
		.addressed(		addressed	),

		.do_peri(		do_peri		),
		.di_peri(		di_peri		),
		.addr_peri(		addr_peri	),
		.addr_high(		addr_high	),
		.mem_access(		mem_access	),
		.store_access(		store_access	),
		.clkb_data(		clkb_data	)

	); 
wire int_upid_set;
// Protocol Engine
usb1_pe
	pe(
		.clk(			clk		),
		.rst(			rst		),
		.tx_valid(		tx_valid_out	),
		.rx_active(		rx_active	),
		.pid_OUT(		pid_OUT		),
		.pid_IN(		pid_IN		),
		.pid_SOF(		pid_SOF		),
		.pid_SETUP(		pid_SETUP	),
		.pid_DATA0(		pid_DATA0	),
		.pid_DATA1(		pid_DATA1	),
		.pid_DATA2(		pid_DATA2	),
		.pid_MDATA(		pid_MDATA	),
		.pid_ACK(		pid_ACK		),
		.pid_PING(		pid_PING	),
		.token_valid(		token_valid	),
		.rx_data_done(		rx_data_done	),
		.crc16_err(		crc16_err	),
		.send_token(		send_token	),
		.token_pid_sel(		token_pid_sel	),
		.data_pid_sel(		data_pid_sel	),
		.rx_dma_en(		rx_dma_en	),
		.tx_dma_en(		tx_dma_en	),
		.abort(			abort		),
		.idma_done(		idma_done	),
		.transfer_finished(	transfer_finished),
		.fsel(			fsel		),
		.ep_sel(		ep_sel		),
		.match(			match_o		),
		.nse_err(		nse_err		),
		.int_upid_set(		int_upid_set	),
		.int_crc16_set(		int_crc16_set	),
		.int_to_set(		int_to_set	),
		.get_cfg(		get_cfg		),
		.cfg_valid(		cfg_valid	),
		.cfg_r(			cfg_r		),
		.state(			state_pe	),
		.dpid(			dpid		)
	);

endmodule
