/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


/////////////////////////////////////////////////////////////////////
////                                                             ////
////  USB 1.1 function IP core                                   ////
////                                                             ////
////                                                             ////
////  Author: Rudolf Usselmann                                   ////
////          rudi@asics.ws                                      ////
////                                                             ////
////                                                             ////
////  Based on: http://www.opencores.org/cores/usb1_funct/       ////
////                                                             ////
/////////////////////////////////////////////////////////////////////

`include "usb11/src/usb1_defines.vh"
`default_nettype none

module usbcore	(

			reset,
			clk_peri,
			do_peri,
			di_peri,
			addr_peri,
			mem_access,
			store_access_low,
			store_access_high,
			addr_high,

			clk_48_mhz,
			intr,
			// USB PHY Interface
			dp,
			dn,
			// Global Config Register Interface
			disconnect,
			// Debug Interface
			ep_sel,
			nse_err,	// no such endpoint error
			fsel,		// This Function is selected
			match_o,
			state_pe,
			state,
			get_cfg,
			transfer_finished,
			tx_dma_en,
			rx_dma_en,
			funct_adr,
			dpid,

			tx_active,
			rx_active,

			int_crc16_set,	// Set CRC16 error interrupt
			int_to_set,	// Set time out interrupt
			int_seqerr_set, // Set PID sequence error interrupt
			send_stall_o,
			configured_o,
			halt_o,
			addressed_o
		);

parameter	USE_DCM			= 0;
parameter	ENDPOINTS		= 2;
parameter	DOUBLE_BUFFERING	= 0;
parameter	DMA_BASE_ADR		= 18'h0;

parameter	IR_SOURCES		= ENDPOINTS+1;
parameter 	NOGAP			= 0;
// Parameter for DCM
parameter	CLKIN_DIVIDE_BY_2	=	"FALSE";
parameter	CLKDV_DIVIDE10X		=	20.0;
parameter	CLKFX_MULTIPLY		=	24;
parameter	CLKFX_DIVIDE		=	25;
parameter	CLKIN_PERIOD		=	20;
parameter	CLK_FEEDBACK		=	"1X";
parameter	CLKOUT_PHASE_SHIFT	=	"NONE";
parameter	DESKEW_ADJUST		=	"SYSTEM_SYNCHRONOUS";
parameter	DFS_FREQUENCY_MODE	=	"LOW";
parameter	DLL_FREQUENCY_MODE	=	"LOW";
parameter	DUTY_CYCLE_CORRECTION	=	"TRUE";
parameter	PHASE_SHIFT		=	0;
parameter	STARTUP_WAIT		=	"FALSE";

output wire	send_stall_o;
output wire	configured_o;
output wire	halt_o;
output wire	addressed_o;
output wire	get_cfg;
output wire	transfer_finished;
output wire	tx_dma_en;
output wire	rx_dma_en;
output wire	int_crc16_set;	// Set CRC16 error interrupt
output wire	int_to_set;	// Set time out interrupt
output wire	int_seqerr_set; // Set PID sequence error interrupt
output wire	nse_err;	// no such endpoint error
output wire	fsel;		// This Function is selected
output wire	match_o;

output wire		tx_active;
output wire		rx_active;

wire		txoe;
assign		tx_active = ~txoe;

input	wire	reset;

input	wire	clk_48_mhz;

inout	wire	dp;
inout	wire	dn;

input wire		clk_peri;
input wire	[17:0]	do_peri;
output wire	[17:0]	di_peri;
input wire	[9:0]	addr_peri;
input wire		mem_access;
input wire		store_access_low;
input wire		store_access_high;
input wire	[7:0]	addr_high;



// Global Config Register Interface
output wire		disconnect;
output wire	[IR_SOURCES-1:0]	intr;


// Debug Interface
output wire	[3:0]	ep_sel;		// Endpoint Number
output wire	[6:0]	state_pe;
output wire	[6:0]	state;
output wire	[6:0]	funct_adr;
output wire		dpid;


///////////////////////////////////////////////////////////////////
//
// Local Wires and Registers
//

// UTMI Interface
wire	[7:0]	DataOut;
wire			TxValid;
wire			TxReady;
wire	[7:0]	DataIn;
wire			RxValid;
wire			RxActive;
wire			RxError;
wire	[1:0]	LineState;

wire	[7:0]	rx_data;
wire			rx_valid, rx_err;
wire	[7:0]	tx_data;
wire			tx_valid;
wire			tx_ready;
wire			tx_first;
wire			tx_valid_last;


reg		[7:0]	tx_data_st;
wire	[7:0]	rx_data_st;

reg				rst_local;	// internal reset
wire			usb_rst;
wire			rst_core;
reg		[7:0]	rst_cnt;

///////////////////////////////////////////////////////////////////
//
// Misc Logic
//

assign rst_core = rst_cnt[7];

always @(posedge clk_48_mhz )
	if(reset)
		rst_cnt <= 8'h00;
	else
		if (!rst_cnt[7])
			rst_cnt <= #1 rst_cnt + 8'h01;

always @(posedge clk_48_mhz)
	rst_local <= #1 rst_core & ~(usb_rst & disconnect);

///////////////////////////////////////////////////////////////////
//
// Module Instantiations
//

usb_phy	phy	(
			.clk(		clk_48_mhz	),
			.rst(		rst_core	),
			// ONLY external reset
			.phy_tx_mode(	1'b1		),
			.usb_rst(	usb_rst		),
			// Transceiver Interface
			.dp(		dp		),
			.dn(		dn		),
			// UTMI Interface
			.DataIn_o(	DataIn		),
			.RxValid_o(	RxValid		),
			.RxActive_o(	RxActive	),
			.RxError_o(	RxError		),
			.DataOut_i(	DataOut		),
			.TxValid_i(	TxValid		),
			.TxReady_o(	TxReady		),
			.LineState_o(	LineState	),
			.txoe(		txoe		)
		);

// UTMI Interface
usb1_utmi_if utmi (
			.phy_clk(	clk_48_mhz	),
			.rst(		rst_local	),
			.DataOut(	DataOut		),
			.TxValid(	TxValid		),
			.TxReady(	TxReady		),
			.RxValid(	RxValid		),
			.RxActive(	RxActive	),
			.RxError(	RxError		),
			.DataIn(	DataIn		),
			.rx_data(	rx_data		),
			.rx_valid(	rx_valid	),
			.rx_active(	rx_active	),
			.rx_err(	rx_err		),
			.tx_data(	tx_data		),
			.tx_valid(	tx_valid	),
			.tx_valid_last(	tx_valid_last	),
			.tx_ready(	tx_ready	),
			.tx_first(	tx_first	),
			.data_ready()
		);

// Protocol Layer
usb1_pl  #(
		.IR_SOURCES(		IR_SOURCES	),
		.DOUBLE_BUFFERING(	DOUBLE_BUFFERING),
		.DMABASE_ADR(		DMA_BASE_ADR	),
		.NOGAP(			NOGAP		))
pl (
			.clk(		clk_48_mhz	),
			.rst(		rst_local	),
			.rx_data(	rx_data		),
			.rx_valid(	rx_valid	),
			.rx_active(	rx_active	),
			.rx_err(	rx_err		),
			.tx_data(	tx_data		),
			.tx_valid(	tx_valid	),
			.tx_valid_last(	tx_valid_last	),
			.tx_ready(	tx_ready	),
			.tx_first(	tx_first	),
			.tx_valid_out(	TxValid		),
			.ep_sel(	ep_sel		),
			.nse_err(	nse_err		),	// no such endpoint error
			.fsel(		fsel		),	// This Function is selected
			.match_o(	match_o		),
			.disconnect(	disconnect	),
			.interrupt(	intr		),

			.int_crc16_set(	int_crc16_set	),	// Set CRC16 error interrupt
			.int_to_set(	int_to_set	),	// Set time out interrupt
			.int_seqerr_set(int_seqerr_set	),	// Set PID sequence error interrupt
			.configured(	configured_o	),
			.halt(		halt_o		),
			.addressed(	addressed_o	),
			.send_stall(	send_stall_o	),

			.do_peri(	do_peri		),
			.di_peri(	di_peri		),
			.addr_peri(	addr_peri	),
			.addr_high(	addr_high	),
			.mem_access(	mem_access	),
			.store_access(	store_access_low && store_access_high	),
			.clkb_data(	clk_peri	),

			.state_pe(	state_pe	),
			.state(		state		),
			.get_cfg(	get_cfg		),
			.transfer_finished(transfer_finished),
			.tx_dma_en(	tx_dma_en	),
			.rx_dma_en(	rx_dma_en	),
			.funct_adr(	funct_adr	),
			.dpid(		dpid		)
	);

endmodule
