/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

 `default_nettype none

module clk_prim
#(
	// default parameters for 60 MHz output frequency on DE0-Nano 
	// with 50 MHz onboard oscillator
	parameter CLKIN_PERIOD		=	20.000,
	parameter SHIFT			= 	0.0, // can be imprecise, e. g. 50 MHz input, 90.4 ° phase results in 5022 ps shift instead of 5022.222...
	parameter CLKFX_DIVIDE		=	5,
	parameter CLKFX_MULTIPLY	=	6,
	parameter CLKIN_FREQ_STR	=	"156.250000", // only needed for lattice ECP3
	parameter CLKOUT_FREQ_FLT	=	60.000000, // only needed for lattice ECP3
	parameter CPHASE_STEPS		=	0, // needed for Lattice ECP5
	parameter FPHASE_STEPS          =	0 // needed for Lattice ECP5

)
(
	input wire CLKIN, 
	input wire RST, 

	output wire CLKFX,
	output wire LOCKED
);

// CLKIN_PERIOD conversion from nanoseconds to picoseconds
localparam integer CLKIN_PERIOD_PS = ((CLKIN_PERIOD+0.0) * 1000) ;

// SHIFT conversion from degrees to picoseconds
localparam integer SHIFT_PS = ((SHIFT+0.0) / 360.0 ) * (CLKIN_PERIOD_PS+0.0) ;

// conversion from decimal to string (clk1_phase_shift parameter accepts string only)
// needed for simulation

// highest value tested 1.000.000.000 picoseconds ( = 1 kHz)
// clock range 1 kHz - 1 THz

localparam value = SHIFT_PS;
localparam digit_10 = value / 1000000000;
localparam digit_9 = (value / 100000000) - (digit_10 * 10);
localparam digit_8 = (value / 10000000) - (digit_10 * 100) - (digit_9 * 10);
localparam digit_7 = (value / 1000000) - (digit_10 * 1000) - (digit_9 * 100) - (digit_8 * 10);
localparam digit_6 = (value / 100000) - (digit_10 * 10000) - (digit_9 * 1000) - (digit_8 * 100) - (digit_7 * 10);
localparam digit_5 = (value / 10000) - (digit_10 * 100000) - (digit_9 * 10000) - (digit_8 * 1000) - (digit_7 * 100) - (digit_6 * 10);
localparam digit_4 = (value / 1000) - (digit_10 * 1000000) - (digit_9 * 100000) - (digit_8 * 10000) - (digit_7 * 1000) - (digit_6 * 100) - (digit_5 * 10);
localparam digit_3 = (value / 100) - (digit_10 * 10000000) - (digit_9 * 1000000) - (digit_8 * 100000) - (digit_7 * 10000) - (digit_6 * 1000) - (digit_5 * 100) - (digit_4 * 10);
localparam digit_2 = (value / 10) - (digit_10 * 100000000) - (digit_9 * 10000000) - (digit_8 * 1000000) - (digit_7 * 100000) - (digit_6 * 10000) - (digit_5 * 1000) - (digit_4 * 100) - (digit_3 * 10);
localparam digit_1 = value  - (digit_10 * 1000000000) - (digit_9 * 100000000) - (digit_8 * 10000000) - (digit_7 * 1000000) - (digit_6 * 100000) - (digit_5 * 10000) - (digit_4 * 1000) - (digit_3 * 100)- (digit_2 * 10);

localparam digit_10_ascii = digit_10 + 48;
localparam digit_9_ascii = digit_9 + 48;
localparam digit_8_ascii = digit_8 + 48;
localparam digit_7_ascii = digit_7 + 48;
localparam digit_6_ascii = digit_6 + 48;
localparam digit_5_ascii = digit_5 + 48;
localparam digit_4_ascii = digit_4 + 48;
localparam digit_3_ascii = digit_3 + 48;
localparam digit_2_ascii = digit_2 + 48;
localparam digit_1_ascii = digit_1 + 48;

localparam  SHIFT_PS_ASCII = 96'd0 + (8'd34 << (8 * 11)) + (digit_10_ascii << (8 * 10)) + (digit_9_ascii << (8 * 9)) + (digit_8_ascii << (8 * 8)) + (digit_7_ascii << (8 * 7)) + (digit_6_ascii << (8 * 6)) + (digit_5_ascii << (8 * 5)) + (digit_4_ascii << (8 * 4)) +(digit_3_ascii << (8 * 3)) + (digit_2_ascii << (8 * 2)) + (digit_1_ascii << (8 * 1)) + (8'd34);

//localparam  value_control = {8'd34,8'd49,8'd48,8'd48,8'd48,8'd48,8'd48,8'd48,8'd48,8'd48,8'd48,8'd34}; // ASCII "1000000000"

wire [4:0] outclk;
wire [1:0] inclk = {1'b0, CLKIN};

assign CLKFX = outclk[0:0];

wire locked_internal;
assign LOCKED = locked_internal;

	defparam
		altpll_component.bandwidth_type = "AUTO",
		altpll_component.clk0_divide_by = CLKFX_DIVIDE,
		altpll_component.clk0_duty_cycle = 50,
		altpll_component.clk0_multiply_by = CLKFX_MULTIPLY,
		altpll_component.clk0_phase_shift = 0, // the synthesis works fine with picoseconds as integer value
//		altpll_component.clk0_phase_shift = SHIFT_PS, // the synthesis works fine with picoseconds as integer value
		altpll_component.compensate_clock = "CLK0",
		altpll_component.inclk0_input_frequency = CLKIN_PERIOD_PS,
		altpll_component.intended_device_family = "Cyclone IV E",
		altpll_component.lpm_hint = "CBX_MODULE_PREFIX=clk_prim",
		altpll_component.lpm_type = "altpll",
		altpll_component.operation_mode = "NORMAL",
		altpll_component.pll_type = "AUTO",
		altpll_component.port_activeclock = "PORT_UNUSED",
		altpll_component.port_areset = "PORT_USED",
		altpll_component.port_clkbad0 = "PORT_UNUSED",
		altpll_component.port_clkbad1 = "PORT_UNUSED",
		altpll_component.port_clkloss = "PORT_UNUSED",
		altpll_component.port_clkswitch = "PORT_UNUSED",
		altpll_component.port_configupdate = "PORT_UNUSED",
		altpll_component.port_fbin = "PORT_UNUSED",
		altpll_component.port_inclk0 = "PORT_USED",
		altpll_component.port_inclk1 = "PORT_UNUSED",
		altpll_component.port_locked = "PORT_USED",
		altpll_component.port_pfdena = "PORT_UNUSED",
		altpll_component.port_phasecounterselect = "PORT_UNUSED",
		altpll_component.port_phasedone = "PORT_UNUSED",
		altpll_component.port_phasestep = "PORT_UNUSED",
		altpll_component.port_phaseupdown = "PORT_UNUSED",
		altpll_component.port_pllena = "PORT_UNUSED",
		altpll_component.port_scanaclr = "PORT_UNUSED",
		altpll_component.port_scanclk = "PORT_UNUSED",
		altpll_component.port_scanclkena = "PORT_UNUSED",
		altpll_component.port_scandata = "PORT_UNUSED",
		altpll_component.port_scandataout = "PORT_UNUSED",
		altpll_component.port_scandone = "PORT_UNUSED",
		altpll_component.port_scanread = "PORT_UNUSED",
		altpll_component.port_scanwrite = "PORT_UNUSED",
		altpll_component.port_clk0 = "PORT_USED",
		altpll_component.port_clk1 = "PORT_UNUSED",
		altpll_component.port_clk2 = "PORT_UNUSED",
		altpll_component.port_clk3 = "PORT_UNUSED",
		altpll_component.port_clk4 = "PORT_UNUSED",
		altpll_component.port_clk5 = "PORT_UNUSED",
		altpll_component.port_clkena0 = "PORT_UNUSED",
		altpll_component.port_clkena1 = "PORT_UNUSED",
		altpll_component.port_clkena2 = "PORT_UNUSED",
		altpll_component.port_clkena3 = "PORT_UNUSED",
		altpll_component.port_clkena4 = "PORT_UNUSED",
		altpll_component.port_clkena5 = "PORT_UNUSED",
		altpll_component.port_extclk0 = "PORT_UNUSED",
		altpll_component.port_extclk1 = "PORT_UNUSED",
		altpll_component.port_extclk2 = "PORT_UNUSED",
		altpll_component.port_extclk3 = "PORT_UNUSED",
		altpll_component.self_reset_on_loss_lock = "OFF",
		altpll_component.width_clock = 5;

// simulation in modelsim works only if parameter value is given in ascii converted format

// synopsys translate_off
defparam	altpll_component.clk0_phase_shift = SHIFT_PS_ASCII; // accepts only integer values in picoseconds as a string
// synopsys translate_on

	altpll	altpll_component (
				.areset (RST),
				.inclk (inclk),
				.clk (outclk),
 				.locked (locked_internal),
				.activeclock (),
				.clkbad (),
				.clkena ({6{1'b1}}),
				.clkloss (),
				.clkswitch (1'b0),
				.configupdate (1'b0),
				.enable0 (),
				.enable1 (),
				.extclk (),
				.extclkena ({4{1'b1}}),
				.fbin (1'b1),
				.fbmimicbidir (),
				.fbout (),
				.fref (),
				.icdrclk (),
				.pfdena (1'b1),
				.phasecounterselect ({4{1'b1}}),
				.phasedone (),
				.phasestep (1'b1),
				.phaseupdown (1'b1),
				.pllena (1'b1),
				.scanaclr (1'b0),
				.scanclk (1'b0),
				.scanclkena (1'b1),
				.scandata (1'b0),
				.scandataout (),
				.scandone (),
				.scanread (1'b0),
				.scanwrite (1'b0),
				.sclkout0 (),
				.sclkout1 (),
				.vcooverrange (),
				.vcounderrange ());

endmodule
