#ifndef _ARP_H_
#define _ARP_H_

#ifdef __cplusplus
extern "C" {
#endif

/**
 *  ARP Table
 *  store every mac, ip pair. It is a array of arpEntry structures.
 *  The two values offset and maxOffset describe the state of 
 *  the whole table.
 *  Both member values use only 16 bit. The remaining 2 bit may
 *  not be used.
 *
 *  int offset: is equal to the highest index of all valid entries.
 *  	Between 0 and offset every entry is valid.
 *
 *  int maxOffset: is the maximal size of the ARP Table index.
 */

struct arpEntry{
	unsigned int mac[3];
	unsigned int ip[2];
};

/**
 *  encapsulate ARP Table, offset and maxOffset.
 *  This way the number of parameters is decreased.
 *  This method is necessary, as handleArpFrame would require more than
 *  4 parameters. But only 4 parameters can be passed with the registers
 *  of the spartanmc. Also exceeding this limit can leed to
 *  uncompileable code. At least as long the lcc18 is used.
 */
struct arpData{
	struct arpEntry* table;
	unsigned char nextOffset;
	unsigned char maxOffset;
};

/**
 *  Add an arpEntry to the table. If no free space is available, the oldest
 *  entry is removed and the new entry is inserted.
 *  If an entry contain a matching mac address it is removed and the new entry
 *  is added.
 *  
 *  parameters:
 *  	struct arpEntry* table: a pointer to the ARP Table
 *
 *  	struct arpEntry* newEntry: a pointer to the new arpEntry
 *
 *  	int offset: the current offset
 *
 *  	int maxOffset: the maximal offset
 *
 *  return value:
 *  	the new nextOffset after the insert
 */
int addArpEntry(struct arpData* data, struct arpEntry* newEntry);

/**
 *  Search the ARP Table after the given mac address
 *
 *  parameters:
 *  	struct arpEntry* table: a pointer to the ARP Table
 *
 *  	const char* mac: a pointer to the 6 byte long mac address
 *
 *  	unsigned char offset: the current offset. The ARP Table is searched
 *  	until this value.
 *
 *  return value:
 *  	-1 if no matching entry is found.
 *  	If an entry match contain a matching mac address, its offset is
 *  	returned.
 */
int searchArpEntry(struct arpData* data, const unsigned int* mac);

/**
 *  Overwrite the entry dest with the data from the entry src.
 *
 *  parameters:
 *  	struct arpEntry* src: a pointer to the source entry.
 *
 *  	struct arpEntry* dest: a pointer to the destination entry.
 *
 */
void replaceArpEntry(struct arpEntry* dest, struct arpEntry* src);

/**
 *  If a ARP frame is received this function can handle it.
 *  (The Ethertype is not checked again)
 *  This function update the ARP Table and if the received frame
 *  is a ARP Request with matching MAC address, a ARP Reply is written
 *  to the location of sendFrame. The data starts with the 18 bit long
 *  length of the frame. It is correctly terminated and its length is
 *  written after all other data.
 *  This function can directly write into the tx section of the ethernet
 *  module. The TxAddrCnt is not changed. It have to be set properly before
 *  handleArpFrame is called.
 *
 *  parameters:
 *  	const unsigned int* frame: the pointer to the frame data. It
 *  	reference the first byte of the destionation MAC
 *
 *  	struct arpData* data: a pointer to the arpData structure.
 *  	It contain the required offsets and the ARP Table
 *
 *  	unsigned int* sendFrame: a pointer to the tx memory section
 *  	of the DMA memory from the ethernet module. After this pointer
 *  	the whole ARP Reply (60 byte + 18 bit frame length) have to
 *  	fit in the memory, without crossing any section borders.
 *  	It is not supported in this version, but it is theoretically
 *  	possible. 
 *
 *  return values:
 *  	-1 = error (malformed frame,...)
 *  	0  = ok
 *  	1  = a ARP reply was written
 */
int handleArpFrame(const volatile unsigned int* frame, struct arpData* data,
	volatile unsigned int* sendFrame, struct arpEntry* own);

/**
 *  Write a ARP Reply frame to the sendFrame location.
 *
 *  parameters:
 *  	unsigned int* sendFrame: a pointer to the tx section
 *  	After this pointer
 *  	the whole ARP Reply (60 byte + 18 bit frame length) have to
 *  	fit in the memory without crossing any section borders.
 *  	It is not supported in this version, but it is theoretically
 *  	possible. 
 *
 *  	struct arpEntry* own: a pointer to an entry containing
 *  	the own ip and mac address
 *
 *  	struct arpEntry* target: a pointer to an entry containing
 *  	the target ip and mac address
 *
 */
void writeArpReply(volatile unsigned int* sendFrame, struct arpEntry* own,
	struct arpEntry* target);

/**
 *  Write a ARP Request frame to the sendFrame location.
 *
 *  parameters:
 *  	unsigned int* sendFrame: a pointer to the tx section
 *  	After this pointer
 *  	the whole ARP Request (60 byte + 18 bit frame length) have to
 *  	fit in the memory without crossing any section borders.
 *  	It is not supported in this version, but it is theoretically
 *  	possible. 
 *
 *  	struct arpEntry* own: a pointer to an entry containing
 *  	the own ip and mac address
 *
 *  	struct arpEntry* target: a pointer to an entry containing
 *  	the target ip address. the mac field can contain any value.
 *  	It is ignored.
 *
 */
void writeArpRequest(volatile unsigned int* sendFrame, struct arpEntry* own,
	struct arpEntry* target);

#ifdef __cplusplus
}
#endif

#endif
