#ifndef ETHERNETLIB_H_
#define ETHERNETLIB_H_

#ifdef __cplusplus
extern "C" {
#endif

#include "peripherals/ethernet_tx.h"
#include "peripherals/ethernet_rx.h"
#include "peripherals/mdio.h"

#define E_NO_MEM -1
#define E_INVALID_ARG -2
#define E_NO_IDLE -3



/**
 * Initialize the ethernet_tx for sending.
 * It always uses the start of the memory, so any other frame in that area of memory gets overwritten.
 *
 *
 * @param ethernet_tx_regs_t * reg: Peripheral registers of ethernet_tx
 * @param ethernet_tx_dma_t * dma: DMA memory of ethernet_tx
 *
 * @return: The memory area that should be used for the data
 */
volatile unsigned int * ethernet_tx_init_frame(ethernet_tx_regs_t * reg, ethernet_tx_dma_t * dma);

/**
 * Check if the device is currently sending.
 *
 * @param ethernet_tx_regs_t * reg: Peripheral registers of ethernet_tx
 * @param ethernet_tx_dma_t * dma: DMA memory of ethernet_tx
 *
 * @return: !=0 if the controller is currently sending, 0 otherwise.
 */
unsigned int ethernet_tx_is_sending(ethernet_tx_regs_t * reg, ethernet_tx_dma_t * dma);


/**
 * Send data that was previously written to the DMA memory. Method returns immediately. To
 * find out if the packet was sent, use ethernet_tx_is_sending.
 *
 * @param ethernet_tx_regs_t * reg: Peripheral registers of ethernet_tx
 * @param ethernet_tx_dma_t * dma: DMA memory of ethernet_tx
 *
 * @param unsigned int size: size of the frame in byte, without
 * 	frame check sum. 60 <= size <= 2000
 *
 * @return: 0 On success and negative value on error
 */
int ethernet_tx_commit_frame(ethernet_tx_regs_t * reg, ethernet_tx_dma_t * dma, unsigned int size);

/**
 * Copy the frame data in the ethernet buffer and start transmission,
 * It always uses the start of the memory, so any other frame in that area of memory gets overwritten.
 * Method returns immediately after giving the memory to the hardware. To find out if the packet was
 * sent, use ethernet_tx_is_sending.
 *
 * @param ethernet_tx_regs_t * reg: Peripheral registers of ethernet_tx
 * @param ethernet_tx_dma_t * dma: DMA memory of ethernet_tx
 *
 * @param unsigned int* data: frame data buffer, array containing the frame 
 * 	which should be sent. Each int represent 2 byte of data.
 * 	IMPORTANT: Frame check sum is excluded.
 * 	It is calculated in hardware and appended to the data
 *
 * @param unsigned int size: size of the frame in byte, without
 * 	frame check sum. 60 <= size <= 2000
 *
 * @return: 0 On success and negative value on error
 */
int ethernet_tx_send_frame(ethernet_tx_regs_t * reg, ethernet_tx_dma_t * dma, unsigned int *data, unsigned int size);


/**
 * Delete all stored packets. Useful after reset.
 *
 * @param ethernet_rx_regs_t * reg: Peripheral registers of ethernet_rx
 * @param ethernet_rx_dma_t * dma: DMA memory of ethernet_rx
 */
void ethernet_rx_init(ethernet_rx_regs_t * reg, ethernet_rx_dma_t * dma);

/**
 * If there is a received frame available, get its start address and size.
 * The received packet may wrap around inside the ring buffer. The word following dma.x[1023]
 * is dma.x[0]. Using ethernet_rx_read correctly handles the wraparound.
 *
 * When the frame is no longer needed, call ethernet_rx_frame_free. If the memory is full, the hardware
 * starts dropping frames.
 *
 * @param ethernet_rx_regs_t * reg: Peripheral registers of ethernet_rx
 * @param ethernet_rx_dma_t * dma: DMA memory of ethernet_rx
 *
 * @param unsigned int * size: Is set to the length of a received frame in bytes if there is one available. Otherwise 0
 *
 * @return: If there is a frame available, its start. Otherwise 0.
 */
volatile unsigned int * ethernet_rx_get_frame_start(ethernet_rx_regs_t * reg, ethernet_rx_dma_t * dma, unsigned int * size);

/**
 * Free the oldest previously received frame (i.e. the one ethernet_rx_get_frame_start returns)
 * so that the hardware may overwrite it.
 *
 * @param ethernet_rx_regs_t * reg: Peripheral registers of ethernet_rx
 * @param ethernet_rx_dma_t * dma: DMA memory of ethernet_rx
 */
void ethernet_rx_frame_free(ethernet_rx_regs_t * reg, ethernet_rx_dma_t * dma);

/**
 * Handle wraparound for reading data from an ethernet frame. Subtracts the size of the dma mem from the pointer if it
 * is outside the address range.
 *
 * WARNING: Only one subtraction is done! Should be enough for regular wraparound.
 *
 * Implemented as a define for SPEED.
 *
 * @param ethernet_rx_dma_t * dma: DMA memory of ethernet_rx
 *
 * @param volatile unsigned int* mem: A location in memory, possibly after the end of dma.
 *
 * @return: The value of the memory
 */
#define ethernet_rx_read(dma,mem) (\
	((mem)>=(dma)->x+1024)?\
	(*((mem)-1024)):\
	(*(mem))\
)

//**************************
//	MDIO
//**************************

/**
 * Wait until a transmission process over the mdio line is done;
 *
 * @param mdio: the mdio peripheral
 */
void mdio_wait(const mdio_regs_t* const mdio);

/**
 * Write the specified value in the register of the desired port of the PHY.
 * Does not wait for the end of the write process. If waiting is necessary
 * use the function mdio_wait(). If this function is called while the bus is not idle,
 * it waits for the bus to become idle.
 *
 * @param mdio: the mdio peripheral
 * 
 * @param unsigned char phy: mdio address of the port to be accessed
 *
 * @param unsigned char reg: The number of the register. 0 - 0x1F
 * 
 * @param unsigned int value: The 16 bit value which is written to the
 * 	register.
 *
 *
 * @return: On success 0, if an error occured an negative value.
 */
int mdio_write(mdio_regs_t * const  mdio, const unsigned char phy, const unsigned char reg, const unsigned int value);

/**
 * Read the value of the specified register and port from the PHY.
 * This function waits for the completion of the read process and
 * saves the value in *value.  If this function is called while the
 * bus is not idle, it waits for the bus to become idle.
 *
 * @param mdio: the mdio peripheral
 *
 * @param unsigned char phy: mdio address of the port to be accessed
 * 
 * @param unsigned char reg: The number of the register. 0 - 0x1F
 * 
 * @param unsigned int *value: pointer to which the result of the
 * 	read process is stored.
 *
 *
 *
 * @return: On success 0, if an error occured an negative value.
 */
int mdio_read(mdio_regs_t * const  mdio, const unsigned char phy, const unsigned char reg, unsigned int * const value);

/**
 * Write the specified value in the register of the desired port of the PHY.
 * Does not wait for the end of the last transaction. If waiting is necessary
 * use the function mdio_wait(). If this function is called while the bus is not idle,
 * it waits for the bus to become idle.
 *
 * It can access the extended Register set in TI's DP83867IR.
 *
 * @param mdio: the mdio peripheral
 *
 * @param unsigned char phy: mdio address of the port to be accessed
 *
 * @param unsigned char reg: The number of the register. 0 - 0x1F
 *
 * @param unsigned int value: The 16 bit value which is written to the
 * 	register.
 *
 *
 * @return: On success 0, if an error occured an negative value.
 */
int mdio_write_indirect(mdio_regs_t * const mdio,const unsigned char phy, const unsigned int reg, const unsigned int value);

/**
 * Read the value of the specified register and port from the PHY.
 * This function waits for the completion of the read process and
 * saves the value in *value. If this function is called while the
 * bus is not idle, it waits for the bus to become idle.
 *
 * It can access the extended Register set in TI's DP83867IR.
 *
 * @param mdio: the mdio peripheral
 *
 * @param unsigned char phy: mdio address of the port to be accessed
 *
 * @param unsigned char reg: The number of the register. 0 - 0x1F
 *
 * @param unsigned int *value: pointer to which the result of the
 * 	read process is stored.
 *
 *
 *
 * @return: On success 0, if an error occured an negative value.
 */
int mdio_read_indirect(mdio_regs_t * const mdio,const unsigned char phy, const unsigned int reg, unsigned int * const value);



/**
 * Get the MAC Address of a RX Port as string
 * @param port: the Port
 * @return: The MAC Address as string
 */
char * format_mac_address(const ethernet_rx_regs_t *const port, char * printf_buffer);

/**
 * Format a three word mac address beginning at the specified word in memory
 *
 * Handles wraparound for ethernet_rx dma memory.
 */
char * format_mac_address_mem(ethernet_rx_dma_t * dma, const unsigned int * macAddress, char * printf_buffer);



#ifdef __cplusplus
}
#endif

#endif
