/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#ifndef __PERF_H
#define __PERF_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

#define MAX_CYCLECOUNTERS	4
#define MAX_EVENTCOUNTERS	16

#define EVENT_CCNT3_OVF 13
#define EVENT_CCNT2_OVF 12
#define EVENT_CCNT1_OVF 11
#define EVENT_CCNT0_OVF 10
#define EVENT_EXT0		9
#define EVENT_EXT1		8
#define EVENT_NOP		7
#define EVENT_INTR		6
#define EVENT_DREAD		5
#define EVENT_DWRITE	4
#define EVENT_IREAD		3
#define EVENT_IHIT		2
#define EVENT_IMISS		1
#define EVENT_STALL		0

#define PCNT_IDX_WRITE(a)	asm volatile("movi2s sfr_pcnt_idx, %0" : : "r"(a))
#define PCNT_IDX_READ(a)	asm volatile("movs2i %0, sfr_pcnt_idx" : "=r"(a))
#define PCNT_DAT_WRITE(a)	asm volatile("movi2s sfr_pcnt_dat, %0" : : "r"(a))
#define PCNT_DAT_READ(a)	asm volatile("movs2i %0, sfr_pcnt_dat" : "=r"(a))

/**
 * Configuration data for an event counter.
 */
struct evcnt_conf {
	uint9_t prescaler; /**< Prescaler. 0: No prescaler, 1: Prescaler 2, 2: Prescaler 4, ..., 15: Prescaler 32768 */
	uint9_t eventnum; /**< Number of the counted event. */
	uint9_t tied_ccnt; /**< Cycle counter that this event counter is tied to. */
	uint9_t edge_detect; /**< Edge detection enabled? */
	uint9_t intr_filter; /**< Interrupt filtering enabled? */
	uint9_t tie_ccnt; /**< Tie this event counter to a cycle counter? */
};

/**
 * Configuration data for a cycle counter.
 */
struct ccnt_conf {
	uint9_t prescaler; /**< Prescaler. 0: No prescaler, 1: Prescaler 2, 2: Prescaler 4, ..., 15: Prescaler 32768 */
	uint9_t intr_filter; /**< Interrupt filtering enabled? */
};
	
/**
 * Configuration data for all counters of the performance counter.
 */
struct perf_conf {
	uint9_t num_ccnt; /**< Number of used cycle counters. */
	uint9_t num_evcnt; /**< Number of used event counters. */
	struct ccnt_conf cycle[MAX_CYCLECOUNTERS]; /**< Configuration for the cycle counters. */
	struct evcnt_conf event[MAX_EVENTCOUNTERS]; /**< Configuration for the event counters. */
};

/**
 * Results of all counters of the performance counter.
 */
struct perf_result {
	struct {
		uint36_t counter; /**< Counted cycles. */
		uint9_t overflow; /**< Overflow flag. */
	} cycles[MAX_CYCLECOUNTERS]; /**< Configuration data for all cycle counters. */

	struct {
		uint36_t counter; /**< Counted events. */
		uint9_t overflow; /**< Overflow flag. */
	} events[MAX_EVENTCOUNTERS]; /**< Configuration data for all event counters. */
};

/**
 * Start a cycle counter.
 *
 * This macro sets the corresponding bit in sfr_pcnt_idx to start a cycle counter and all
 * event counters that are tied to that cycle counter.
 *
 * @param a Cycle counter to start.
 */
#define perf_start(a)	__asm__ volatile("movi2s sfr_pcnt_idx, %0" : : "r"(1<<(17 - 2*(a)) ))

/**
 * Stop a cycle counter.
 *
 * This macro sets the corresponding bit in sfr_pcnt_idx to stop a cycle counter and all
 * event counters that are tied to that cycle counter.
 *
 * @param a Cycle counter to stop.
 */
#define perf_stop(a)	__asm__ volatile("movi2s sfr_pcnt_idx, %0" : : "r"(1<<(16 - 2*(a)) ))

/**
 * Initialize all cycle and event counters.
 *
 * All cycle and event counters that are requested in conf are initialized according to the
 * parameters passed by conf.
 *
 * @param conf Pointer to a struct perf_conf.
 * @return 0 on success, -1 on failure.
 */
int perf_init(struct perf_conf *conf);

/**
 * Initializes a cycle counter.
 *
 * Cycle counter n gets configured according to the parameters passed by conf.
 *
 * @param n Cycle counter that should be initialized.
 * @param conf Pointer to a struct ccnt_conf.
 * @return 0 on success, -1 on failure.
 */
int perf_init_ccnt(int n, struct ccnt_conf *conf);

/**
 * Initializes an event counter.
 *
 * Event counter n gets configured according to the parameters passed by conf.
 *
 * @param n Event counter that should be initialized.
 * @param conf Pointer to a struct evcnt_conf.
 * @return 0 on success, -1 on failure.
 */
int perf_init_evcnt(int n, struct evcnt_conf *conf);

/**
 * Resets all cycle and event counters.
 *
 * All counter values and overflow flags are reset to 0. The configuration of
 * the counters is not changed by this function.
 */
void perf_reset(void);

/**
 * Resets a cycle counter.
 *
 * Sets the counter value and the overflow flag to 0. The configuration of the
 * counter is not changed.
 *
 * @param n The cycle counter that should be reset.
 */
void perf_reset_cycle(int n);

/**
 * Resets an event counter.
 *
 * Sets the counter value and the overflow flag to 0. The configuration of the
 * counter is not changed.
 *
 * @param n The event counter that should be reset.
 */
void perf_reset_event(int n);

/**
 * Read results of all counters.
 *
 * Reads the counter values and overflow flags of all cycle and event counters.
 *
 * @param res Pointer to a struct perf_result where the results are stored.
 */
void perf_read(struct perf_result *res);

/**
 * Read results of a cycle counter.
 *
 * @param n The cycle counter that is read.
 * @param overflow Where the overflow flag is stored. If NULL the overflow flag
 *                 is ignored.
 * @return Number of counted cycles.
 */
uint36_t perf_read_ccnt(int n, uint9_t *overflow);

/**
 * Read results of an event counter.
 *
 * @param n The event counter that is read.
 * @param overflow Where the overflow flag is stored. If NULL the overflow flag
 *                 is ignored.
 * @return Number of counted events.
 */
uint36_t perf_read_evcnt(int n, uint9_t *overflow);

/**
 * Print overview of performance counter results.
 *
 * This function prints an overview of the passed results. It prints the counted
 * cycles and events and a warning for every counter that overflowed.
 *
 * @note This function uses printf() for printing. Make sure printf() is set up
 *       correctly before calling this function.
 *
 * @param res struct perf_result that is printed.
 */
void perf_results_printf(struct perf_result *res); // Uses printf() to print the results.

// ***********************************************************************************
// ******* Everything below is used by AutoPerf, do not use in other programs! *******
// ***********************************************************************************
struct perf_auto_result {
	char *name;
	struct perf_result result;
};

#define perf_auto_start() perf_start(0)
#define perf_auto_stop(/*int*/a,/*struct perf_auto_result* */b)\
	do {\
		perf_stop(0);\
		perf_auto_stop_((a),(b));\
	} while(0)

void perf_auto_init(void);
void perf_auto_stop_(int n, struct perf_auto_result *r);
void perf_auto_print(int size, struct perf_auto_result *r);

void perf_auto_print_iter(int size, struct perf_auto_result *r);
void perf_auto_end_loop(void);
void perf_auto_stop_loopOpt();

extern struct perf_auto_result perf_results[];

// ***********************************************************************************
// ******* Everything below is used by µStreams, do not use in other programs! *******
// ***********************************************************************************
void microstreams_eval_init(void);
void microstreams_eval_restart(void);
uint36_t microstreams_eval_stop(void);

#ifdef __cplusplus
}
#endif

#endif
