/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
*/
#ifndef __DDR_MEM_PORT_18_H
#define __DDR_MEM_PORT_18_H

#ifdef __cplusplus
extern "C" {
#endif

/* hardware constants */
#define DDR_MEM_PORT_18_FIFO_SIZE      64

/* commands */
#define DDR_MEM_PORT_18_CMD_WR         0x0
#define DDR_MEM_PORT_18_CMD_RD         0x1
#define DDR_MEM_PORT_18_CMD_WR_PRE     0x2
#define DDR_MEM_PORT_18_CMD_RD_PRE     0x3
#define DDR_MEM_PORT_18_CMD_REF        0x4

/* burst lengths */
#define DDR_MEM_PORT_18_BURST(n)       ((((n)-1) & 0x3F) << 3)

/* status */
#define DDR_MEM_PORT_18_ST_CALIB_DONE  0x001
#define DDR_MEM_PORT_18_ST_CMD_EMPTY   0x002
#define DDR_MEM_PORT_18_ST_CMD_FULL    0x004
#define DDR_MEM_PORT_18_ST_WR_FULL     0x008
#define DDR_MEM_PORT_18_ST_WR_EMPTY    0x010
#define DDR_MEM_PORT_18_ST_WR_UN       0x020
#define DDR_MEM_PORT_18_ST_WR_ERROR    0x040
#define DDR_MEM_PORT_18_ST_RD_FULL     0x080
#define DDR_MEM_PORT_18_ST_RD_EMPTY    0x100
#define DDR_MEM_PORT_18_ST_RD_OV       0x200
#define DDR_MEM_PORT_18_ST_RD_ERROR    0x400

#include <stdint.h>

typedef struct {

    /* Setzen von Kommando und Burstlength für Zugriffe auf den DDR 
     * [2:0] cmd:   write 000
     *              Read  001
     *              Write with Auto Precharge
     *                    010
     *              Read with Auto Precharge
     *                    011
     *              Refresh
     *                    1xx
     *
     * [8:3] burstlength: Controller schreibt bzw. ließt die Anzahl in diesem Register +1
     *  bl = 0 -> r/w 1 Wort
     *  bl = 1 -> r/w 2 Wörter 
     *  usw.
     * */
    volatile uint18_t cmd_bl;

    /* Aufbau der DDR-Adresse (insgesammt 30bit breit):
     * [1:0] -> fest auf 0 gesetzt da bei 32Bit Zugriffen diese 0 sein müssen
     * [19:2] -> dr_addr_low- wert
     * [29:20] -> inhalt des Registers "upper_addr" -> emulation von Speicherfenster
     *
     * Schreiben dieses Register überträgt Kommando + Burstlength + gesammte Addresse an den Kommandofifo vom MCB
     */        
    volatile uint18_t addr_low;	

    /* setzen des Registers "upper_addr"
     * auswahl des Speicherfensters
     */
    volatile uint18_t addr_high;	

    /* Daten die in den DDR geschrieben werden sollen */
    volatile uint18_t data_wr;	

    /* Daten die aus dem DDR gelesen werden
    */
    volatile uint18_t data_rd;		

    /* DDR - Status:
     * [0]   calibration done
     * [1]   command fifo empty (cmd_fifo 4 Enträge)
     * [2]   command fifo full
     * [3]   wr_full     -> Schreibfifo voll  (data_fifo 64 Einträge)
     * [4]   wr_empty    -> Schreibfifo leer
     * [5]   wr_underrun -> Es sollten mehr Daten geschrieben werden als im Fifo vorhanden
     *                      (burst_length > wr_count bei setzen des Commandos)
     *                      fehlende Daten wurden durch 0 ersetzt
     * [6]   wr_error    -> Fehler, reset nötig1
     * [7]   rd_full     -> Lesefifo voll (data_fifo 64 Einträge)
     * [8]   rd_empty    -> Lesefifo leer
     * [9]   rd_overflow -> es wurden mehr Daten gelesen als Platz im Fifo war
     *                      überschüssige Daten verloren
     * [10]  rd_error    -> Fehler, reset nötig
     *
     * [17:11] unused
     */ 
    volatile uint18_t status;

    /**
     * Füllstand Schreibfifo
     * wird mehr, wenn Daten auf data_wr geschrieben werden
     * wird weniger wenn Schreib-Kommando abgesetzt wurde
     */
    volatile uint18_t fill_wr;
    
    /**
     * Füllstand Lesefifo
     * wird mehr, wenn Lese-Kommando abgesetzt wurde
     * wird weniger, wenn von data_rd gelesen wird
     */
    volatile uint18_t fill_rd;

} ddr_mem_port_18_regs_t;

#ifdef __cplusplus
}
#endif

#endif
