#ifndef __DDR_MEM_PORT_32_H
#define __DDR_MEM_PORT_32_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>

typedef union {
    struct {
        uint18_t low;
        uint18_t high;
    };
    uint36_t full;
} uint36_2x18_t;


/* control */
#define DDR_MEM_PORT_CT_WR_MASK           0x00F
#define DDR_MEM_PORT_CT_DATA_LAYOUT       0x010
#define DDR_MEM_PORT_CT_LOW_ONLY          0x020
#define DDR_MEM_PORT_CT_AUTO_PRECHARGE    0x040


/* status */
#define DDR_MEM_PORT_ST_RD_COUNT_MASK 0x007F
#define DDR_MEM_PORT_ST_CALIB_DONE    0x0080
#define DDR_MEM_PORT_ST_WR_EMPTY      0x0100
#define DDR_MEM_PORT_ST_WR_UN         0x0200
#define DDR_MEM_PORT_ST_WR_ERROR      0x0400
#define DDR_MEM_PORT_ST_WR_FULL       0x0800
#define DDR_MEM_PORT_ST_RD_EMPTY      0x1000
#define DDR_MEM_PORT_ST_RD_OV         0x2000
#define DDR_MEM_PORT_ST_RD_ERROR      0x4000

typedef struct {
    /* Control: r/w
     *  [3:0] wr_mask         -> Any bit set to 1 masks out the corresponding byte during write. The old value for masked
     *                           bytes is retained in memory. wr_mask needs to be set before entering the word into the
     *                           wr data FIFO.
     *  [4]   data_layout     -> Controls how to split the 32 bits of data into 18 bit registers.
     *                          0: Continuous. 4 bits of zero padding are inserted at the most significant bits.
     *                          1: 16 Bit + padding. Each 18 Bit word contains 16 Bits of data and two bits of zero
     *                              padding.
     *  [5]   low_only        -> Controls when data is read / written to the rd/wr FIFO
     *                          0: Both the upper and lower register need to be accessed to issue a request.
     *                              If accessing the 36-bit register, this mode should be used.
     *                          1: Only the lower register needs to be accessed to issue a request.
     *                              If only 18 Bits of data are ever used (i.e. only the lower reg), this mode
     *                              should be used.
     *  [6]   auto_precharge -> Append a auto precharge command after each read/write. Improves performance for random
     *                          access patterns.
     *  [17:7] unused
     */
    volatile uint18_t control;

    /* Status: read only
     * [6:0]   rd_count_reserved -> count in read fifo once all pending requests have been served
     * [7]     calibration done  -> do not write anything to other registers before this is 1.
     * [8]     wr_empty          -> Write FIFO empty, All data safely written
     * [9]     wr_underrun       -> Error: More data was to be written than was contained in FIFO. Indicates error in
     *                               periphery controller. Reset needed to clear error flag
     * [10]    wr_error          -> Clocking error. Reset needed to clear flag
     * [11]    wr_full           -> Write FIFO full. Do not write into the FIFO until this flag is deasserted
     * [12]    rd_empty          -> Read FIFO empty
     * [13]    rd_overflow       -> More data was requested than there was space in the read FIFO. Indicates error in
     *                               periphery controller. Reset needed to clear error flag
     * [14]    rd_error          -> Clocking error. Reset needed to clear flag
     *
     * [17:15] unused
     */
    volatile uint18_t status;

    /*
     * Address of first data to be read/written, in 32-bit words.
     * This value is updated automatically as burst requests are issued.
     * Do not change the address while a read is still in progress.
     * Only the lowest 30 bits are used.
     */
    volatile uint36_2x18_t address;

    /*
     * Writing to this register starts a read of the specified number of words. Because a FIFO of 64 words is used
     * internally, multiple smaller read requests are issued to the DRAM controller. Reading this register will show the
     * number of words that have not yet been requested from the controller.
     */
    volatile uint18_t rdLength;

    /*
     * Data.
     * Writing to this register will issue data to be saved to the DRAM, reading from it will serve previously requested
     * data.
     *
     * Depending on data_layout, either there are 4 bits of zero padding at the most significant bits, or at bits
     * [35:34] and [17:16] (i,e, the two most significant bits of the two 18-bit subwords).
     *
     * If low_only is 0, both the upper and lower register need to be accessed to enter a word into the FIFO / read from
     * it. If low_only is 1, requests are issued when accessing the lower register.
     */
    volatile uint36_2x18_t data;

} ddr_mem_port_32_regs_t;

#ifdef __cplusplus
}
#endif

#endif
