#pragma once
#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
	volatile unsigned int status;
	volatile unsigned int control;
	void * volatile error_addr;
	void * volatile error_pc;
	volatile unsigned int watcher_idx;
	volatile unsigned int watcher_conf;
	void * volatile watcher_offs;
	volatile unsigned int watcher_size;
} memguard_regs_t;

/**
 * Convert a code address to a data address
 */
#define CODE_TO_DATA(x) ((int *)(2*((unsigned int)(x))))

#define MEMGUARD_WATCHER_ENABLE 1
#define MEMGUARD_ALLOW_CODE 2
#define MEMGUARD_ALLOW_RD 4
#define MEMGUARD_ALLOW_WR 8

#define MEMGUARD_STAT_ERR_MASK 3
#define MEMGUARD_STAT_ERR_CODE 1
#define MEMGUARD_STAT_ERR_WR 2
#define MEMGUARD_STAT_ERR_RD 3

/**
 * Configure a watcher. It is automatially enabled
 *
 * @param memguard The memguard instance
 * @param watcher_idx The index of the watcher to configure
 * @param start start of memory area to watch as data address. code pointers passed in will need to be multiplied
 * 	by two or they are wrong. (Use CODE_TO_DATA)
 * @param size size in words of the memory area
 * @param flags allowed access modes. Use an or-combination of MEMGUARD_ALLOW_*
 */
void memguard_configure_watcher(memguard_regs_t *memguard, unsigned int watcher_idx, void *start, unsigned int size,
			   unsigned int flags);

/**
 * Enable or disable the whole memguard
 * @param memguard The memguard instance
 * @param enable 1 to enable, 0 to disable
 */
void memguard_enable(memguard_regs_t *memguard, int enable);

/**
 * Check if the memguard is enabled
 * @param memguard The memguard instance
 * @return 1 if enabled, 0 otherwise
 */
#define memguard_is_enabled(memguard) (memguard)->control

/**
 * To be used in memguard generated interrupts. Print the reason for the interrupt.
 * if trapno is != 0, trap with trap 0
 * Will not return in case of unallowed execution (it would only return into more unallowed code).
 */
void memguard_print_internal(memguard_regs_t * memguard, int trapnoPlus1);

/**
 * To be used in memguard generated interrupts. Print the reason for the interrupt.
 * Will not return in case of unallowed execution (it would only return into more unallowed code).
 */
#define memguard_print(memguard) memguard_print_internal(memguard,0)

/**
 * To be used in memguard generated interrupts. Print the reason for the interrupt.
 * Afterwards, trap with trap 0
 * Will not return in case of unallowed execution (it would only return into more unallowed code).
 */
#define memguard_print_trapping(memguard) memguard_print_internal(memguard,1)

/**
 * Set up default configuration. All Sections generated by the elf file will be protected so that they are only accessible
 * with the access type specified by the elf file.
 *
 * @param memguard the memguard to configure
 * @param numMatchers the number of matchers the memguard provides
 */
void memguard_configure_default(memguard_regs_t *memguard, unsigned int numMatchers);


/**
 * Print information about the memguard's configuration
 * @param memguard the memory guard module
 * @param numWatchers the number of watchers the memguard has
 */
void memguard_info(memguard_regs_t *memguard, unsigned int numWatchers) ;

#ifdef __cplusplus
}
#endif
