/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
*/

#ifndef __USB11_H
#define __USB11_H

#ifdef __cplusplus
extern "C" {
#endif

////////////////////////////////////////////////////////////////////////////////
//                                                                            //
// Konstanten und Datenfelder fuer das USB-Modul des SpartanMC                //
//                                                                            //
////////////////////////////////////////////////////////////////////////////////

#define USB11_ATTRIBUTE_BULK		0x02
#define USB11_DIRECTION_IN		0x01
#define USB11_DIRECTION_OUT		0x00
#define USB11_DEVICE_DESCRIPTOR		0x01
#define USB11_CONFIG_DESCRIPTOR		0x02
#define USB11_STRING_DESCRIPTOR		0x03
#define USB11_INTERFACE_DESCRIPTOR	0x04
#define USB11_ENDPOINT_DESCRIPTOR	0x05
#define USB11_SELF_POWERED		0x40

#define USB11_CLASS_VENDOR_SPECIFIC	0xFF

/**
 * Konstanten fuer epXXc-Register
 */
#define USB11_CTRL_EN			0x20000	//Bereit zum Datenaustausch mit dem Host
#define USB11_CTRL_INTR			0x10000	//Interruptfreigabe fuer den EP
#define USB11_CTRL_MODE			0x08000	//synchron
#define USB11_CTRL_CTRL			0x04000
#define USB11_CTRL_OUT			0x02000
#define USB11_CTRL_IN			0x01000
#define USB11_CTRL_BUFSEL		0x00800	//Buffer HIGH bei Doppelpufferung waehlen
#define USB11_CTRL_SIZE			0x0007F	//Werte von 0 bis 63, wobei 0 als 64 gewertet wird.

/**
 * Konstanten fuer das globale Konfigurationsregister
 */
#define	USB11_EN			0x20000
#define	USB11_DI			0x00000
#define	USB11_IEN_EP0			0x10000	//Interruptfreigabe fuer EP0
#define	USB11_IEP15			0x08000
#define	USB11_IEP14			0x04000
#define	USB11_IEP13			0x02000
#define	USB11_IEP12			0x01000
#define	USB11_IEP11			0x00800
#define	USB11_IEP10			0x00400
#define	USB11_IEP09			0x00200
#define	USB11_IEP08			0x00100
#define	USB11_IEP07			0x00080
#define	USB11_IEP06			0x00040
#define	USB11_IEP05			0x00020
#define	USB11_IEP04			0x00010
#define	USB11_IEP03			0x00008
#define	USB11_IEP02			0x00004
#define	USB11_IEP01			0x00002
#define	USB11_IEP00			0x00001

// descriptor header 
struct __attribute__((__packed__)) usb_descr_header {
        unsigned char size;             // size of descriptor structure
        unsigned char type;             // descriptor type
                                        // 0x01 = device descriptor
                                        // 0x02 = control descriptor
                                        // 0x03 = string descriptor
                                        // 0x04 = interface descriptor
                                        // 0x05 = endpoint descriptor
};

// device descriptor
struct __attribute__((__packed__)) usb_device_descr {
        struct usb_descr_header hdr;    // descriptor header
                                        // .size = 0x12
                                        // .type = 0x01
        unsigned char _r0;              // reserverd
        unsigned char usb_spec;         // usb specification number = 0x01
        unsigned char class;            // class code
                                        // 0xFF = class code is vendor specified
        unsigned char subclass;         // subclass code = 0x00
        unsigned char protocol;         // protocol code = 0xFF
        unsigned char packet_size;      // max packet size for EP0 = 0x40
        unsigned char vendor_id_low;
        unsigned char vendor_id_high;
        unsigned char product_id_low;
        unsigned char product_id_high;
        unsigned char release_low;      // device release number
        unsigned char release_high;
        unsigned char vendor_descr_idx; // index of vendor string descriptor
        unsigned char product_descr_idx;// index of product string descriptor
        unsigned char serial_descr_idx; // index of serial string descriptor
        unsigned char num_configs;      // number of possible configurations
};

// configuration descriptor
struct __attribute__((__packed__)) usb_config_descr {
        struct usb_descr_header hdr;    // descriptor header
                                        // .size = 0x09
                                        // .type = 0x02
        unsigned char config_size_low;  // number of bytes used for endpoint 
        unsigned char config_size_high; // descriptors, including this structure
        unsigned char num_ifaces;       // number of interfaces
        unsigned char index;            // value to use as argument to select this 
                                        // configuration
        unsigned char descr_idx;        // index of config string descriptor
        unsigned char power_mode;       // power mode for device
                                        // 0x40 = self powered
        unsigned char power_cons;       // max power consumption in 2 mA units
};

// interface descriptor
struct __attribute__((__packed__)) usb_iface_descr {
        struct usb_descr_header hdr;    // descriptor header
                                        // .size = 0x09
                                        // .type = 0x04
        unsigned char index;            // index of interface
        unsigned char alt_setting;      // alternativ setting
                                        // 0x00 = no
                                        // 0x01 = yes (?)
        unsigned char num_endpoints;    // number of endpoints
        unsigned char class;            // class code
        unsigned char subclass;         // subclass code
        unsigned char protocol;         // protocol code
        unsigned char descr_idx;        // index of string descriptor
};

// endpoint address
struct __attribute__((__packed__)) usb_endpoint_address {
        unsigned _r0:           1;      // reserved
        unsigned direction:     1;      // direction
                                        // 0 = out
                                        // 1 = in
        unsigned _r1:           3;      // reserved
        unsigned number:        4;      // endpoint number
};

// endpoint descriptor
struct __attribute__((__packed__)) usb_endpoint_descr {
        struct usb_descr_header hdr;    // descriptor header
                                        // .size = 0x07
                                        // .type = 0x05
        struct usb_endpoint_address adr;// endpoint address
        unsigned char attr;             // endpoint attributes
                                        // 0x02 = bulk
        unsigned char packet_size_low;  // maximum packet size this endpoint
        unsigned char packet_size_high; // is able to handle
        unsigned char poll_interval;    // polling interval
                                        // (ignored for bulk and 
                                        // control endpoints)
};

// language descriptor
struct __attribute__((__packed__)) usb_language_descr {
        struct usb_descr_header hdr;    // descriptor header
                                        // .size = 0x04
                                        // .type = 0x03
        unsigned char lang_id_low;      // language id
        unsigned char lang_id_high;     // 0x0409 = US English
};

/*
// Konfigurationsregister USB-Endpunkt
struct __attribute__((__packed__)) usb_control {
        unsigned en             :1;     //    [17] Enable
        unsigned intr           :1;     //    [16] Interrupt enable
        unsigned mode           :1;     //    [15] Uebertragungsmodus
                                        //              1 = synchron
                                        //              0 = asynchron
        unsigned control        :1;     //    [14] Kontrollendpunkt
        unsigned out            :1;     //    [13] Endpunkt OUT
        unsigned in             :1;     //    [12] Endpunkt IN
        unsigned bufsel         :1;     //    [11] Pufferauswahl
                                        //              1 = oberer
                                        //              0 = unterer
        unsigned                :4;     // [10: 7] reserviert
        unsigned size           :7;     // [ 6: 0] Zeichenanzahl EP IN
};

// Konfigurationsregister USB global
struct __attribute__((__packed__)) usb_global {
        unsigned en             :1;     // [17] USB-Interface Enable
        unsigned ien_ep0        :1;     // [16] Interruptfreigabe fuer EP0
        unsigned iep15          :1;     // [15] Reset Interrupt EP 15
        unsigned iep14          :1;     // [14] Reset Interrupt EP 14
        unsigned iep13          :1;     // [13] Reset Interrupt EP 13
        unsigned iep12          :1;     // [12] Reset Interrupt EP 12
        unsigned iep11          :1;     // [11] Reset Interrupt EP 11
        unsigned iep10          :1;     // [10] Reset Interrupt EP 10
        unsigned iep09          :1;     // [ 9] Reset Interrupt EP  9
        unsigned iep08          :1;     // [ 8] Reset Interrupt EP  8
        unsigned iep07          :1;     // [ 7] Reset Interrupt EP  7
        unsigned iep06          :1;     // [ 6] Reset Interrupt EP  6
        unsigned iep05          :1;     // [ 5] Reset Interrupt EP  5
        unsigned iep04          :1;     // [ 4] Reset Interrupt EP  4
        unsigned iep03          :1;     // [ 3] Reset Interrupt EP  3
        unsigned iep02          :1;     // [ 2] Reset Interrupt EP  2
        unsigned iep01          :1;     // [ 1] Reset Interrupt EP  1
        unsigned iep00          :1;     // [ 0] Reset Interrupt EP  0
};
*/
// USB-Register und -puffer
typedef struct usb {

        // 0x00 - 0x3F
        volatile unsigned int	ep0c0;  // Globales Kommandoregister 1
        volatile unsigned int	ep0c1;  // Globales Kommandoregister 2
        volatile unsigned int	ep01c;  // Kommados fuer Endpunkt 1
        volatile unsigned int	ep01s;  // Status vom Endpunkt 1
        volatile unsigned int	ep02c;  // Kommados fuer Endpunkt 2
        volatile unsigned int	ep02s;  // Status vom Endpunkt 2
        volatile unsigned int	ep03c;  // Kommados fuer Endpunkt 3
        volatile unsigned int	ep03s;  // Status vom Endpunkt 3
        volatile unsigned int	ep04c;  // Kommados fuer Endpunkt 4
        volatile unsigned int	ep04s;  // Status vom Endpunkt 4
        volatile unsigned int	ep05c;  // Kommados fuer Endpunkt 5
        volatile unsigned int	ep05s;  // Status vom Endpunkt 5
        volatile unsigned int	ep06c;  // Kommados fuer Endpunkt 6
        volatile unsigned int	ep06s;  // Status vom Endpunkt 6
        volatile unsigned int	ep07c;  // Kommados fuer Endpunkt 7
        volatile unsigned int	ep07s;  // Status vom Endpunkt 7
        volatile unsigned int	ep08c;  // Kommados fuer Endpunkt 8
        volatile unsigned int	ep08s;  // Status vom Endpunkt 8
        volatile unsigned int	ep09c;  // Kommados fuer Endpunkt 9
        volatile unsigned int	ep09s;  // Status vom Endpunkt 9
        volatile unsigned int	ep10c;  // Kommados fuer Endpunkt 10
        volatile unsigned int	ep10s;  // Status vom Endpunkt 10
        volatile unsigned int	ep11c;  // Kommados fuer Endpunkt 11
        volatile unsigned int	ep11s;  // Status vom Endpunkt 11
        volatile unsigned int	ep12c;  // Kommados fuer Endpunkt 12
        volatile unsigned int	ep12s;  // Status vom Endpunkt 12
        volatile unsigned int	ep13c;  // Kommados fuer Endpunkt 13
        volatile unsigned int	ep13s;  // Status vom Endpunkt 13
        volatile unsigned int	ep14c;  // Kommados fuer Endpunkt 14
        volatile unsigned int	ep14s;  // Status vom Endpunkt 14
        volatile unsigned int	ep15c;  // Kommados fuer Endpunkt 15
        volatile unsigned int	ep15s;  // Status vom Endpunkt 15

        // 0x40
        volatile unsigned int	glob;   // Globales Register

        // 0x42
        struct usb_device_descr device; // device descriptor
        // 0x54
        struct usb_config_descr config; // configuration descriptor
        // 0x5D
        struct usb_iface_descr iface;   // interface descriptor

        // 0x66 - 0xCE
        struct usb_endpoint_descr ep01;  // endpoint descriptor
        struct usb_endpoint_descr ep02;  // endpoint descriptor
        struct usb_endpoint_descr ep03;  // endpoint descriptor
        struct usb_endpoint_descr ep04;  // endpoint descriptor
        struct usb_endpoint_descr ep05;  // endpoint descriptor
        struct usb_endpoint_descr ep06;  // endpoint descriptor
        struct usb_endpoint_descr ep07;  // endpoint descriptor
        struct usb_endpoint_descr ep08;  // endpoint descriptor
        struct usb_endpoint_descr ep09;  // endpoint descriptor
        struct usb_endpoint_descr ep10;  // endpoint descriptor
        struct usb_endpoint_descr ep11;  // endpoint descriptor
        struct usb_endpoint_descr ep12;  // endpoint descriptor
        struct usb_endpoint_descr ep13;  // endpoint descriptor
        struct usb_endpoint_descr ep14;  // endpoint descriptor
        struct usb_endpoint_descr ep15;  // endpoint descriptor

        unsigned char           _padding;

        // 0xD0
        struct usb_language_descr lang;  // language descriptor

        // 0xD4
        struct usb_descr_header vendor_str_hdr;    // vendor string descriptor header
                                         // .size = 0x12
                                         // .type = 0x03
        unsigned char           vendor_str[16];
        // 0xE6
        struct usb_descr_header product_str_hdr;   // product string descriptor header
                                         // .size = 0x12
                                         // .type = 0x03
        unsigned char           product_str[16];
        // 0xF8
        struct usb_descr_header serial_str_hdr;    // serial string descriptor header
                                         // .size = 0x8
                                         // .type = 0x03
        unsigned char           serial_str[6];


        // 0x100 - 0x420
        volatile unsigned int     data00[32];    // Puffer 0     EP00
        volatile unsigned int     data01[32];    // Puffer 1     EP00 2 fuer Doppelpufferung

        volatile unsigned int     data02[32];    // Puffer 2     EP01
        volatile unsigned int     data03[32];    // Puffer 3
        volatile unsigned int     data04[32];    // Puffer 4     EP02
        volatile unsigned int     data05[32];    // Puffer 5
        volatile unsigned int     data06[32];    // Puffer 6     EP03
        volatile unsigned int     data07[32];    // Puffer 7
        volatile unsigned int     data08[32];    // Puffer 8     EP04
        volatile unsigned int     data09[32];    // Puffer 9
        volatile unsigned int     data10[32];    // Puffer 10    EP05
        volatile unsigned int     data11[32];    // Puffer 11
        volatile unsigned int     data12[32];    // Puffer 12    EP06
        volatile unsigned int     data13[32];    // Puffer 13
        volatile unsigned int     data14[32];    // Puffer 14    EP07
        volatile unsigned int     data15[32];    // Puffer 15
        volatile unsigned int     data16[32];    // Puffer 16    EP08
        volatile unsigned int     data17[32];    // Puffer 17
        volatile unsigned int     data18[32];    // Puffer 18    EP09
        volatile unsigned int     data19[32];    // Puffer 19
        volatile unsigned int     data20[32];    // Puffer 20    EP10
        volatile unsigned int     data21[32];    // Puffer 21
        volatile unsigned int     data22[32];    // Puffer 22    EP11
        volatile unsigned int     data23[32];    // Puffer 23
//      volatile unsigned int     data24[32];    // Puffer 24    EP12
//      volatile unsigned int     data25[32];    // Puffer 25
//      volatile unsigned int     data26[32];    // Puffer 26    EP13
//      volatile unsigned int     data27[32];    // Puffer 27
} usb11_dma_t;
#ifdef __cplusplus
}
#endif

#endif
