#ifndef RTOS_H
#define RTOS_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdint.h>
#include <stddef.h>

// ---------------------------------------------------------------------------
// types
// ---------------------------------------------------------------------------

typedef struct rtos_list_item {
	struct rtos_list_item *previous;
	struct rtos_list_item *next;
} rtos_list_item_t;

// Idee das list_item als Liste zu verwenden
// dem System RT-Thread entnommen:
typedef rtos_list_item_t rtos_list_t;

typedef struct regfile_block {
	rtos_list_item_t list_item;
	int9_t lower_limit;	//
	int9_t upper_limit;	// must be a signed type
} regfile_block_t;

struct tcb {
	rtos_list_item_t list_item;
	regfile_block_t regs;
	void *stack;
	uint9_t priority;
	uint9_t ready;
	uint9_t forbidden_count;
	uint9_t status;
};

typedef struct tcb *task_t;

typedef struct semaphore {
	rtos_list_t waiting_tasks;
	uint18_t value;
} semaphore_t;

#define DEFINE_INITIALIZED_MUTEX_SEMAPHORE(sem)	semaphore_t sem = {  \
		{&sem.waiting_tasks, &sem.waiting_tasks},  \
		1  \
	}

#define ISR_MAX_CALL_LEVEL_NO_INTERRUPT	-2
#define ISR_MAX_STACK_USAGE_NO_INTERRUPT	0


// ---------------------------------------------------------------------------
// system call macros
// ---------------------------------------------------------------------------

#ifdef RTOS_DONT_USE_TRAP
	#define SYSCALL_ASM(name) "jals "#name"_\n	or	r0,	r0"
#else
	#define SYSCALL_ASM(name) "trap "#name"_trap_nr"
#endif

#define INLINE_ONLY	static inline __attribute__((always_inline))

#define RETURN_IF_NOT_VOID

#define DEFINE_SYSCALL_0_VOID(name)		\
	INLINE_ONLY void name()			\
	{						\
		register uint18_t r12 __asm__("r12");	\
		__asm__ volatile (			\
			SYSCALL_ASM(name)		\
		: "=r"(r12)				\
		:					\
		: "r13", "r14", "r15", "cc", "memory"	\
		);					\
	}

#define DEFINE_SYSCALL_0(return_type, name)		\
	INLINE_ONLY return_type name()			\
	{						\
		register uint18_t r12 __asm__("r12");	\
		__asm__ volatile (			\
			SYSCALL_ASM(name)		\
		: "=r"(r12)				\
		:					\
		: "r13", "r14", "r15", "cc", "memory"	\
		);					\
		return (return_type) r12;		\
	}

#define DEFINE_SYSCALL_1_VOID(name, param1_type, param1)	\
	INLINE_ONLY void name(param1_type param1)		\
	{								\
		register uint18_t r12 __asm__("r12") = (uint18_t) param1;\
		__asm__ volatile (					\
			SYSCALL_ASM(name)				\
		: "+r"(r12)						\
		:							\
		: "r13", "r14", "r15", "cc", "memory"			\
		);							\
	}

#define DEFINE_SYSCALL_1(return_type, name, param1_type, param1)	\
	INLINE_ONLY return_type name(param1_type param1)		\
	{								\
		register uint18_t r12 __asm__("r12") = (uint18_t) param1;\
		__asm__ volatile (					\
			SYSCALL_ASM(name)				\
		: "+r"(r12)						\
		:							\
		: "r13", "r14", "r15", "cc", "memory"			\
		);							\
		return (return_type) r12;				\
	}

#define DEFINE_SYSCALL_2_VOID(name, 				\
		param1_type, param1, param2_type, param2)		\
	INLINE_ONLY void name(param1_type param1, param2_type param2)\
	{								\
		register uint18_t r12 __asm__("r12") = (uint18_t) param1;\
		register uint18_t r13 __asm__("r13") = (uint18_t) param2;\
		__asm__ volatile (					\
			SYSCALL_ASM(name)				\
		: "+r"(r12), "+r"(r13)					\
		:							\
		: "r14", "r15", "cc", "memory"				\
		);							\
	}

#define DEFINE_SYSCALL_2(return_type, name, 				\
		param1_type, param1, param2_type, param2)		\
	INLINE_ONLY return_type name(param1_type param1, param2_type param2)\
	{								\
		register uint18_t r12 __asm__("r12") = (uint18_t) param1;\
		register uint18_t r13 __asm__("r13") = (uint18_t) param2;\
		__asm__ volatile (					\
			SYSCALL_ASM(name)				\
		: "+r"(r12), "+r"(r13)					\
		:							\
		: "r14", "r15", "cc", "memory"				\
		);							\
		return (return_type) r12;				\
	}

#define DEFINE_SYSCALL_3_VOID(name, param1_type, param1,	\
		param2_type, param2, param3_type, param3)		\
	INLINE_ONLY void name(param1_type param1, 		\
			param2_type param2, param3_type param3)		\
	{								\
		register uint18_t r12 __asm__("r12") = (uint18_t) param1;\
		register uint18_t r13 __asm__("r13") = (uint18_t) param2;\
		register uint18_t r14 __asm__("r14") = (uint18_t) param3;\
		__asm__ volatile (					\
			SYSCALL_ASM(name)				\
		: "+r"(r12), "+r"(r13), "+r"(r14)			\
		:							\
		: "r15", "cc", "memory"					\
		);							\
	}

#define DEFINE_SYSCALL_3(return_type, name, param1_type, param1,	\
		param2_type, param2, param3_type, param3)		\
	INLINE_ONLY return_type name(param1_type param1, 		\
			param2_type param2, param3_type param3)		\
	{								\
		register uint18_t r12 __asm__("r12") = (uint18_t) param1;\
		register uint18_t r13 __asm__("r13") = (uint18_t) param2;\
		register uint18_t r14 __asm__("r14") = (uint18_t) param3;\
		__asm__ volatile (					\
			SYSCALL_ASM(name)				\
		: "+r"(r12), "+r"(r13), "+r"(r14)			\
		:							\
		: "r15", "cc", "memory"					\
		);							\
		return (return_type) r12;				\
	}

#define DEFINE_SYSCALL_4_VOID(name, param1_type, param1, 		\
		param2_type, param2, param3_type, param3, param4_type, param4)	\
	INLINE_ONLY void name(param1_type param1, param2_type param2, 	\
			param3_type param3, param4_type param4)			\
	{									\
		register uint18_t r12 __asm__("r12") = (uint18_t) param1;	\
		register uint18_t r13 __asm__("r13") = (uint18_t) param2;	\
		register uint18_t r14 __asm__("r14") = (uint18_t) param3;	\
		register uint18_t r15 __asm__("r15") = (uint18_t) param4;	\
		__asm__ volatile (						\
			SYSCALL_ASM(name)				\
		: "+r"(r12), "+r"(r13), "+r"(r14), "+r"(r15)		\
		:							\
		: "cc", "memory"					\
		);							\
	}

#define DEFINE_SYSCALL_4(return_type, name, param1_type, param1, 		\
		param2_type, param2, param3_type, param3, param4_type, param4)	\
	INLINE_ONLY return_type name(param1_type param1, param2_type param2, 	\
			param3_type param3, param4_type param4)			\
	{									\
		register uint18_t r12 __asm__("r12") = (uint18_t) param1;	\
		register uint18_t r13 __asm__("r13") = (uint18_t) param2;	\
		register uint18_t r14 __asm__("r14") = (uint18_t) param3;	\
		register uint18_t r15 __asm__("r15") = (uint18_t) param4;	\
		__asm__ volatile (						\
			SYSCALL_ASM(name)				\
		: "+r"(r12), "+r"(r13), "+r"(r14), "+r"(r15)		\
		:							\
		: "cc", "memory"					\
		);							\
		return (return_type) r12;				\
	}


// ---------------------------------------------------------------------------
// system calls
// ---------------------------------------------------------------------------

// void delete_task(task_t task);
DEFINE_SYSCALL_1_VOID(delete_task, task_t, task)

// void suspend_task(task_t task);
DEFINE_SYSCALL_1_VOID(suspend_task, task_t, task)

// void resume_task(task_t task);
DEFINE_SYSCALL_1_VOID(resume_task, task_t, task)

// task_t get_current_task();
DEFINE_SYSCALL_0(task_t, get_current_task)

// void forbid_preemption();
DEFINE_SYSCALL_0_VOID(forbid_preemption)

// void permit_preemption();
DEFINE_SYSCALL_0_VOID(permit_preemption)

// void task_yield();
DEFINE_SYSCALL_0_VOID(task_yield)

// void free(void *ptr);
DEFINE_SYSCALL_1_VOID(free, void*, ptr)

// void *malloc(size_t size);
DEFINE_SYSCALL_1(void*, malloc, size_t, size)

// void initialize_semaphore(semaphore_t *sem, uint18_t value);
DEFINE_SYSCALL_2_VOID(initialize_semaphore, semaphore_t*, sem, uint18_t, value)

// void semaphore_down(semaphore_t *sem);
DEFINE_SYSCALL_1_VOID(semaphore_down, semaphore_t*, sem)

// void semaphore_up(semaphore_t *sem);
DEFINE_SYSCALL_1_VOID(semaphore_up, semaphore_t*, sem)


INLINE_ONLY task_t create_task( void (*entry)(void *param), void *param,
	uint18_t priority, uint18_t max_call_level, size_t stack_size)
{
	register uint18_t r12 __asm__("r12") = (uint18_t) entry;
	register uint18_t r13 __asm__("r13") = (uint18_t) param;
	register uint18_t r14 __asm__("r14") = (uint18_t) priority;
	register uint18_t r15 __asm__("r15") = (uint18_t) max_call_level;
	uint18_t stack_param = stack_size;
	__asm__ volatile (
		"addi	r0,	-2	\n"
		"s18	0(r0),	%4	\n"
		SYSCALL_ASM(create_task)"\n"
		"addi	r0,	2"
	: "+r"(r12), "+r"(r13), "+r"(r14), "+r"(r15)
	: "r"(stack_param)
	: "cc", "memory"
	);
	return (task_t) r12;
}

#ifdef __cplusplus
}
#endif

#endif
