/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#ifndef RTOS_INTERNAL_H
#define RTOS_INTERNAL_H

#define RTOS_DONT_USE_TRAP
#include <rtos.h>
#include <rtos_config.h>


//-----------------------------------------------------------------------------
// internal functions
//-----------------------------------------------------------------------------

void idle_task_code(void *param);
__attribute__((section(".traps"))) task_t get_current_task_();
DEFINE_SYSCALL_2_VOID(switch_tasks, task_t, from_task, task_t, to_task)

DEFINE_SYSCALL_1_VOID(initialize_scheduler, task_t, current)
DEFINE_SYSCALL_1_VOID(scheduler_add_task, task_t, task)
DEFINE_SYSCALL_1_VOID(scheduler_remove_task, task_t, task)
DEFINE_SYSCALL_0_VOID(schedule_tasks)

DEFINE_SYSCALL_1_VOID(init_regfile_allocator, regfile_block_t*, initial_allocated_block)
DEFINE_SYSCALL_2(uint18_t, regalloc, regfile_block_t*, regs, uint18_t, ncalls)
DEFINE_SYSCALL_1_VOID(regfree, regfile_block_t*, regs)

DEFINE_SYSCALL_2_VOID(initialize_heap, void*, heap, size_t, size)


//-----------------------------------------------------------------------------
// constants and configuration
//-----------------------------------------------------------------------------

#ifndef PRIORITY_LEVELS
	#error "The configuration option 'PRIORITY_LEVELS' is not defined!"
#endif

#ifndef EXTENDED_CHECKS_ENABLED
	#error "The configuration option 'EXTENDED_CHECKS_ENABLED' is not defined!"
#endif

#define REGFILE_SIZE	127
#define MAX_REGBASE	REGFILE_SIZE - 1

#define FALSE	0
#define TRUE	!FALSE


//-----------------------------------------------------------------------------
// other macros
//-----------------------------------------------------------------------------

#define STATUS_DISABLE_INT( status )		\
	__asm__ volatile (			\
		"movs2i	%0, SFR_STATUS	\n"	\
		"cbits	INT	\n"		\
	: "=r"(status)				\
	:					\
	: "memory"				\
	);

#define STATUS_RESTORE( status )		\
	__asm__ volatile (			\
		"movi2s	SFR_STATUS, %0	\n"	\
	:					\
	: "r"(status)				\
	: "memory"				\
	);

#define INLINE_ONLY		static inline __attribute__((always_inline))

#define ROUNDED_DOWN_EVEN(a)	(((a) >> 1) << 1)
#define ROUNDED_UP_EVEN(a)	ROUNDED_DOWN_EVEN((a)+1)

#ifndef MAX
	#define MAX(a,b) (((a)<(b)) ? (b) : (a))
#endif

#ifndef MIN
	#define MIN(a,b) (((a)>(b)) ? (b) : (a))
#endif


//-----------------------------------------------------------------------------
// lists
//-----------------------------------------------------------------------------

#define DEFINE_INITIALIZED_EMPTY_LIST(name) rtos_list_t name = { &name, &name }

#define rtos_list_first( list ) ((list)->next)

#define rtos_list_last( list ) ((list)->previous)

#define rtos_list_next( list_item ) ((list_item)->next)

#define rtos_list_previous( list_item ) ((list_item)->previous)

// Idee dieses Makros dem System RT-Thread entnommen:
#define rtos_list_item_owner(list_item, owner_type, member_name)\
	((owner_type *) ((char *)(list_item) - offsetof(owner_type, member_name)))

DEFINE_SYSCALL_3_VOID(rtos_list_insert_between, rtos_list_item_t*, new_item, rtos_list_item_t*, previous, rtos_list_item_t*, next)

DEFINE_SYSCALL_1_VOID(rtos_list_remove, rtos_list_item_t*, item)

DEFINE_SYSCALL_1(rtos_list_item_t*, rtos_list_remove_first, rtos_list_t*, list)

DEFINE_SYSCALL_1_VOID(rtos_list_initialize, rtos_list_t*, list)

/*
INLINE_ONLY void rtos_list_initialize( rtos_list_t *list )
{
	list->next = list;
	list->previous = list;
}
*/

INLINE_ONLY uint18_t rtos_list_is_empty( rtos_list_t *list )
{
	return list->next == list;
}

INLINE_ONLY void rtos_list_insert_after( rtos_list_item_t *new_item, rtos_list_item_t *previous )
{
	rtos_list_insert_between(new_item, previous, previous->next);
}

INLINE_ONLY void rtos_list_insert_before( rtos_list_item_t *new_item, rtos_list_item_t *next )
{
	rtos_list_insert_between(new_item, next->previous, next);
}

#define rtos_list_add_first( list, new_item ) rtos_list_insert_after(new_item, list)

#define rtos_list_add_last( list, new_item ) rtos_list_insert_before(new_item, list)


#endif
