/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#include <string.h>

#pragma GCC diagnostic ignored "-Wimplicit-fallthrough"

static inline void memcpy_wordwise(void *dst, const void *src, size_t count);
static inline void memcpy_bytewise(unsigned char *dst, const void *src, size_t count);


void *memcpy(void *dst, const void *src, size_t count) {
	if(count == 0 || dst == src) {
		return dst;
	}
	unsigned char* dst_ptr = (unsigned char*) dst;
	unsigned char* src_ptr = (unsigned char*) src;

	if((((unsigned int) src) + count) % 2 != 0) {
        	// dst + count is odd -> copy one extra byte at end
            *(dst_ptr + count - 1)  = *(src_ptr + count - 1);
           	if(--count == 0) {
            	return dst;
            }
    }

    if((unsigned int) src % 2 != 0) {
    	//src is odd -> copy one extra byte at start
    	*((unsigned char*) dst) = *((unsigned char*) src);
        if(--count == 0) {
        	return dst;
        }
        dst_ptr++;
        src_ptr++;
    }

	//from now on we work on word level
	count = count >> 1;

	if(((unsigned int) src % 2 == 0) ^ ((unsigned int) dst % 2 == 0)) {
		//copying from odd to even address or even to odd
		memcpy_bytewise(dst_ptr, src_ptr, count);
	} else {
		//copying from odd to odd or even to even
		memcpy_wordwise(dst_ptr, src_ptr, count);
	}
	return dst;
}

/**
 * copies data word by word
 * dst and src must be already word aligned
 * count is number of words (!!!) to copy
 */
__attribute__((optimize("unroll-loops")))
static inline void memcpy_wordwise(void *dst, const void *src, size_t count) {
    int n = (count + 15) >> 4;
    int switch_var = count & 0xf;
    unsigned int * src_ptr;
    unsigned int * dst_ptr;

    src_ptr = (unsigned int *) ((unsigned int)src - 2*(16-switch_var));
    dst_ptr = (unsigned int *) ((unsigned int)dst - 2*(16-switch_var));

    switch (switch_var) {
        case 0	:do{src_ptr=src_ptr+16;
        			dst_ptr=dst_ptr+16;
                *(dst_ptr) = *(src_ptr);
                case 15	:	*(dst_ptr+1) = *(src_ptr+1);
                case 14	:	*(dst_ptr+2) = *(src_ptr+2);
                case 13	:	*(dst_ptr+3) = *(src_ptr+3);
                case 12	:	*(dst_ptr+4) = *(src_ptr+4);
                case 11	:	*(dst_ptr+5) = *(src_ptr+5);
                case 10	:	*(dst_ptr+6) = *(src_ptr+6);
                case 9	:	*(dst_ptr+7) = *(src_ptr+7);
                case 8	:	*(dst_ptr+8) = *(src_ptr+8);
                case 7	:	*(dst_ptr+9) = *(src_ptr+9);
                case 6	:	*(dst_ptr+10) = *(src_ptr+10);
                case 5	:	*(dst_ptr+11) = *(src_ptr+11);
                case 4	:	*(dst_ptr+12) = *(src_ptr+12);
                case 3	:	*(dst_ptr+13) = *(src_ptr+13);
                case 2	:	*(dst_ptr+14) = *(src_ptr+14);
                case 1	:	*(dst_ptr+15) = *(src_ptr+15);
            }while(--n > 0);
    }
}


/**
 * copies data with one l18 followed by s9 and s9
 * src must be already word aligned (is read wordwise)
 * dst must not be word aligned (is written bytewise)
 * count is number of words (!!!) to copy
 */
__attribute__((optimize("unroll-loops")))
static inline void memcpy_bytewise(unsigned char *dst, const void *src, size_t count) {
    int n = (count + 7) >> 3;
    int switch_var = count & 0x7;
    unsigned int * src_ptr;
    unsigned char * dst_ptr;

    src_ptr = (unsigned int *) ((unsigned int)src - 2*(8-switch_var));
    dst_ptr = (unsigned char *) ((unsigned int)dst - 2*(8-switch_var));

	register unsigned int data;
    switch (switch_var) {
        case 0	:do{src_ptr=src_ptr+8;
        			dst_ptr=dst_ptr+16;
        		data = 	*(src_ptr);
                *(dst_ptr+1) = (unsigned char) data;
                *(dst_ptr) = (unsigned char) (data >> 9);
                case 7	:	data = 	*(src_ptr+1);
                			*(dst_ptr+3) = (unsigned char) data;
                			*(dst_ptr+2) = (unsigned char) (data >> 9);
                case 6	:	data = 	*(src_ptr+2);
                			*(dst_ptr+5) = (unsigned char) data;
                            *(dst_ptr+4) = (unsigned char) (data >> 9);
                case 5	:	data = 	*(src_ptr+3);
                			*(dst_ptr+7) = (unsigned char) data;
                			*(dst_ptr+6) = (unsigned char) (data >> 9);
           		case 4	:	data = 	*(src_ptr+4);
           					*(dst_ptr+9) = (unsigned char) data;
                			*(dst_ptr+8) = (unsigned char) (data >> 9);
                case 3	:	data = 	*(src_ptr+5);
                			*(dst_ptr+11) = (unsigned char) data;
                			*(dst_ptr+10) = (unsigned char) (data >> 9);
                case 2	:	data = 	*(src_ptr+6);
                			*(dst_ptr+13) = (unsigned char) data;
                			*(dst_ptr+12) = (unsigned char) (data >> 9);
                case 1	:	data = 	*(src_ptr+7);
                			*(dst_ptr+15) = (unsigned char) data;
                			*(dst_ptr+14) = (unsigned char) (data >> 9);
            }while(--n > 0);
    }
}
