/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*************
*
*    The following gdb commands are supported:
*
* command          function                               Return value
*
*    g             return the value of the CPU registers  hex data or ENN
*    G             set the value of the CPU registers     OK or ENN
		*
		*    mAA..AA,LLLL  Read LLLL bytes at address AA..AA      hex data or ENN
		*    MAA..AA,LLLL: Write LLLL bytes at address AA.AA      OK or ENN
*
*    p				read single register					hex or ENN
*    P				write single register					OK or ENN
*    	*	pAA				read reg AA
*    	*	PAA=XX..XX		write XX to reg AA
*
*    c             Resume at current address              SNN   ( signal NN)
*    cAA..AA       Continue at address AA..AA             SNNf
*
*    s             Step one instruction                   SNN
		*    sAA..AA       Step one instruction from AA..AA       SNN
*
* 	z0			clear memory breakpoint						OK or ENN
* 	Z0			set memory breakpoint						OK or ENN
* 	z1			clear hardware breakpoint					OK or ENN
* 	Z1			set hardware breakpoint						OK or ENN
* 	Z2,Z3,Z4	set watchpoint (read, write, both)			OK or ENN
* 	z2,z3,z4	clear watchpoint (read, write, both)		OK or ENN
* 		*	Zx,AA..AA,4		for watchpoints, 1/2/3 possible, 1: watch only lower halfword, 2: watch only upper, 3/4: watch both
* 		*	zx,AA..AA,4		for watchpoints last param is "Don't Care"
*
* 	w			read register window (only 12 local regs)	hex or ENN
* 		* wAA		read reg window AA (0..124)
*
*    k             kill (remove all breakpoints / watchpoints), proceed with r
*
*    r				reset reg window, jump 0
*
*    ?             What was the last sigval ?             SNN   (signal NN)
*
* All commands and responses are sent with a packet which includes a
* checksum.  A packet consists of
*
* $<packet info>#<checksum>.
*
* where
* <packet info> :: <characters representing the command or response>
* <checksum>    :: < two hex digits computed as modulo 256 sum of <packetinfo>>
*
* When a packet is received, it is first acknowledged with either '+' or '-'.
* '+' indicates a successful transfer.  '-' indicates a failed transfer.
*
* Example:
*
* Host:                  Reply:
* $m0,10#2a               +$00010203040506070809101112131415#42
*
****************************************************************************/

/*
 * The following Defines are supported to further influence function:
 * ALLOW_XMIT_SHORTCUT: accept '==' as valid checksum on all packets
 * WITHOUT_MEMORY_BREAKPOINTS / WITHOUT_HARDWARE_BREAKPOINTS / WITHOUT_WATCHPOINTS: leaves out code and variables for the specified functionality to save on code size
 * SOFT_SLOTS x : allocate memory for x Software Breakpoints, 0 is equivalent to WITHOUT_MEMORY_BREAKPOINTS
 */

#include <stdint.h>
#include <string.h>
#include <debugging.h>
#include <peripherals/uart_light.h>
#include <stdio.h>
#include <bitmagic.h>
#include "dbg_int.h"

#define UPDATE_GPRS				0x8
#define UPDATE_RETURN_ADDR		0x1
#define UPDATE_RESET_REGWINDOW	0x10
#define UPDATE_FIXUP_BRK		0x2
#define SINGLE_STEP_NEXT		0x4


// accept the checksum '==' as correct, always, very helpful if using a serial terminal to skip gdb and directly talk with us (otherwise you'd have to calculate checksums all the time...)
#define ALLOW_XMIT_SHORTCUT
// for breakpoints, use halfword addresses (the ones used in .elf and .lst), recommended ON
#define PC_BYTE_ADDRESSES
// use 'd' to report all hdw debug config data
//#define DEBUG_HDW_DATA

#ifdef HAS_HARDWARE_BREAKPOINTS
uint9_t count_breakpoints = 0;
#endif

#ifdef HAS_WATCHPOINTS
uint9_t count_watchpoints = 0;
#endif

uint9_t single_step_support = 0;

#define OFFSET_BREAKPOINTS 2
#define OFFSET_WATCHPOINTS (OFFSET_BREAKPOINTS+count_breakpoints)
#define OFFSET_BREAK_CFG (OFFSET_BREAKPOINTS+count_breakpoints+count_watchpoints)
#define OFFSET_WATCH_CFG (OFFSET_BREAKPOINTS+count_breakpoints+count_watchpoints+count_breakpoints)
#define DIST_CFG (count_breakpoints+count_watchpoints)

#define BRK_IS_ON(cfg) ((cfg) != 0)
#define WTCH_IS_ON(cfg) (((cfg) & 0x10) != 0)

#define CMD_JRS_R11 0x2164
#define CMD_TRAP_0 0x02001
#define CMD_J 0x04000
#define CMD_NOP 0x2015

#define IS_CMD_BRANCH_LMT(cmd) ((cmd & 0x3C000) == 0x08000)
#define IS_CMD_J_REL(cmd) ((cmd & 0x3C000) == 0x04000)
#define IS_CMD_BRANCH(cmd) ((cmd & 0x38000) == 0x08000)
#define IS_CMD_J_LINK(cmd) (((cmd & 0x3E01F) == 0x02003) || ((cmd & 0x3E01F) == 0x02005))

//#pragma GCC push_options
//#pragma GCC optimize ("Os")

extern void tsr_breakpoint();
extern void capture_window(uint18_t*, uint9_t);
void handle_exception(uint9_t);


/************************************************************************
*
* external low-level support routines
*/

static void __attribute__((section("traps"))) putDebugChar(char c) {
    	fputc(c, debugIO);
}

static char __attribute__((section("traps"))) getDebugChar() {
    	return getc(debugIO);
}

static int __attribute__((section("traps"))) getDebugChar_nb(char* c) {
	return getc_nb(debugIO, c);
}

/************************************************************************/
/* BUFMAX defines the maximum number of characters in inbound/outbound buffers*/
/* at least enough bytes to fit all regs in hex form + 1 \0 */
#define BUFMAX 146

static const char hexchars[]="0123456789abcdef";

#define NUMREGS 18

/* Number of bytes of registers.  */
#define NUMREGBYTES (NUMREGS * 4)
enum regnames {R0, R1, R2, R3, R4, R5, R6, R7,
	R8, R9, R10, R11, R12, R13, R14, R15, SFR_STATUS, PC};

#if NUMREGBYTES*2+1 > BUFMAX
#error "remcom Buffer too small for reg dump"
#endif

#define MAX_DUMP_BYTES (BUFMAX-1)/2
/***************************  ASSEMBLY CODE MACROS *************************/
/* 									   */

/* Convert ch from a hex digit to an int */

static int __attribute__((section("traps")))
hex (char ch)//TODO @see hexToInt(), integrate into libs under util?
{
	if (ch >= 'a' && ch <= 'f')
		return ch-'a'+10;
	if (ch >= '0' && ch <= '9')
		return ch-'0';
	if (ch >= 'A' && ch <= 'F')
		return ch-'A'+10;
	return -1;
}

static char remcomInBuffer[BUFMAX];
//static char remcomOutBuffer[BUFMAX];
//static char * remcomOutBuffer = &remcomInBuffer[0];
#define remcomOutBuffer remcomInBuffer


/* scan for the sequence $<data>#<checksum>     */

static char * __attribute__((section("traps"))) getpacket (void)
{
	char *buffer = &remcomInBuffer[0];
	char checksum;
	char xmitcsum;
	int count;
	char ch;

	while (1)
	{
		/* wait around for the start character, ignore all other characters */
		while ((ch = getDebugChar ()) != '$')
			;

		retry:
		checksum = 0;
		xmitcsum = -1;
		count = 0;

		/* now, read until a # or end of buffer is found */
		while (count < BUFMAX - 1)
		{
			ch = getDebugChar ();
			if (ch == '$')
				goto retry;
			if (ch == '#')
				break;
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}
		buffer[count] = 0;

		if (ch == '#')
		{
			ch = getDebugChar ();
#ifdef ALLOW_XMIT_SHORTCUT
			if (ch != '=') {// Convenience so i dont have to calculate checksums #== is always valid
#endif
				xmitcsum = hex (ch) << 4;
				ch = getDebugChar ();
				xmitcsum += hex (ch);
#ifdef ALLOW_XMIT_SHORTCUT
			} else {
				ch = getDebugChar();
			}

			if (ch != '=' && ((checksum & 0xFF) != xmitcsum))
#else
			if ((checksum & 0xFF) != xmitcsum)
#endif
			{
				putDebugChar ('-');	/* failed checksum */
			}
			else
			{
				putDebugChar ('+');	/* successful transfer */

				/* if a sequence char is present, reply the sequence ID */
				if (buffer[2] == ':')
				{
					putDebugChar (buffer[0]);
					putDebugChar (buffer[1]);

					return &buffer[3];
				}

				return &buffer[0];
			}
		}
	}
}

/* send the packet in buffer.  */

static void __attribute__((section("traps"))) putpacket (char *buffer)
{
	char checksum;
	int count;
	char ch;

	/*  $<packet info>#<checksum>. */
	do
	{
		putDebugChar('$');
		checksum = 0;
		count = 0;

		while ((ch = buffer[count]))
		{
			putDebugChar(ch);
			checksum += ch;
			count += 1;
		}

		putDebugChar('#');
		putDebugChar(hexchars[(checksum >> 4) & 0xf]);
		putDebugChar(hexchars[checksum & 0xf]);

	}
	while (getDebugChar() != '+');
}

__attribute__((constructor (1000)))
void __attribute__((section("traps"))) debugging_initialize() {
    uint18_t info;

    // load info reg to see if we have specialized hardware and how much points it holds
    __asm__(
			"movi %[info], 0\n\t"
            "movi2s SFR_DBG_IDX, %[info]\n\t"
            "movs2i %[info], SFR_DBG_DAT\n\t"
    :[info] "=r"(info)
    : /* none */
    );

#ifdef HAS_HARDWARE_BREAKPOINTS
    count_breakpoints = 0xFF & info;
#endif
#ifdef HAS_WATCHPOINTS
    count_watchpoints = (0x1FE00 & info) >> 9;
#endif
    single_step_support = (0x20000 & info) >> 17;
    

    __asm__ volatile(
			"movi r12, 3\n\t"
			"j capture_regs\n\t"
			"or r0,r0\n\t");
}

/* Convert the memory pointed to by mem into hex, placing result in buf.
 * Input is in form of 18 bit vars, output as 32bits filled with zeros
 * count: number of uint18 to write to hex
 * may_fault: legacy parameter (wether or not mem_fault results in a trap [spartanmc does not have that ability])
 */

static char* __attribute__((section(".traps"))) mem18_2_hex32 (uint18_t *mem, char *buf, unsigned int count)
{
	uint18_t word;

//	set_mem_fault_trap(may_fault);

	while (count-- > 0)
	{
		word = *mem++;
//		if (mem_err)
//			return 0;
		*buf++ = '0';
		*buf++ = '0';
		*buf++ = '0';
		*buf++ = hexchars[(word >> 16) & 0xf];
		*buf++ = hexchars[(word >> 12) & 0xf];
		*buf++ = hexchars[(word >> 8) & 0xf];
		*buf++ = hexchars[(word >> 4) & 0xf];
		*buf++ = hexchars[word & 0xf];
	}

	*buf = 0;

//	set_mem_fault_trap(0);

	return buf;
}

static char* __attribute__((section(".traps"))) mem9_2_hex16 (uint9_t *mem, char *buf, unsigned int count)
{
	uint9_t byte;

//	set_mem_fault_trap(may_fault);

	while (count-- > 0)
	{
		byte = *mem++;
//		if (mem_err)
//			return 0;
		*buf++ = '0';
		*buf++ = hexchars[(byte >> 8) & 0xf];
		*buf++ = hexchars[(byte >> 4) & 0xf];
		*buf++ = hexchars[byte & 0xf];
	}

	*buf = 0;

//	set_mem_fault_trap(uint18_t * hdw_points = (uint18_t*) &BREAKPOINT_MGR_0->point_start;0);

	return buf;
}

static char* __attribute__((section(".traps"))) mem8_2_hex8 (uint9_t *mem, char *buf, unsigned int count)
{
	uint9_t byte;

//	set_mem_fault_trap(may_fault);
	while (count-- > 0) {
		byte = *mem++;
		//		if (mem_err)
		//			return 0;
		*buf++ = hexchars[(byte >> 4) & 0xf];
		*buf++ = hexchars[byte & 0xf];
	}
	*buf = 0;

//	set_mem_fault_trap(uint18_t * hdw_points = (uint18_t*) &BREAKPOINT_MGR_0->point_start;0);

	return buf;
}

/* convert the hex array pointed to by buf into binary to be placed in mem
 * return a pointer to the character AFTER the last byte written */

static uint18_t * __attribute__((section(".traps"))) hex32_2_mem18 (char *buf, uint18_t *mem, unsigned int count)
{
	uint18_t word;
	uint9_t halfScrap;
	uint18_t scrap;

//	set_mem_fault_trap(may_fault);

	while (count-- > 0)
	{
		scrap = hex(*buf++);
		scrap |= hex(*buf++);
		scrap |= hex(*buf++);
		// the first three characters cannot be anything else than 0
		if (scrap != 0)
			return 0;
		halfScrap = hex(*buf++);
		if (halfScrap > 3) // the upper 2 bits cannot be antything else than 0
			return 0;

		word = halfScrap << 16;
		word |= hex(*buf++) << 12;
		word |= hex(*buf++) << 8;
		word |= hex(*buf++) << 4;
		word |= hex(*buf++);
		*mem++ = word;
//		if (mem_err)
//			return 0;
	}

//	set_mem_fault_trap(0);

	return mem;
}

static uint9_t * __attribute__((section(".traps"))) hex16_2_mem9 (char *buf, uint9_t *mem, unsigned int count)
{
	uint9_t byte;
	uint9_t halfScrap;
	uint9_t scrap;

//	set_mem_fault_trap(may_fault);

	while (count-- > 0)
	{
		scrap = hex(*buf++);
		// the first character cannot be anything else than 0
		if (scrap != 0)
			return 0;
		halfScrap = hex(*buf++);
		if (halfScrap > 1) // the upper 3 bits cannot be antything else than 0
			return 0;

		byte = halfScrap << 8;
		byte |= hex(*buf++) << 4;
		byte |= hex(*buf++);
		*mem++ = byte;
//		if (mem_err)
//			return 0;
	}

//	set_mem_fault_trap(0);

	return mem;
}

static uint9_t * __attribute__((section(".traps"))) hex8_2_mem8 (char *buf, uint9_t *mem, unsigned int count)
{
	uint9_t byte;

//	set_mem_fault_trap(may_fault);

	while (count-- > 0) {
		byte = hex(*buf++) << 4;
		byte |= hex(*buf++);
		*mem++ = byte;
	//		if (mem_err)
	//			return 0;
	}
//	set_mem_fault_trap(0);

	return mem;
}

#define SIGHUP 1
#define SIGTRAP 5
#define SIGSTOP 17

/* This table contains the mapping between trap types, and
   signals, which are primarily what GDB understands. */

static struct hard_trap_info
{
	char tt;		/* Trap type code for SPARTANMC */
	unsigned char signo;		/* Signal that we map this trap into */
} hard_trap_info[] = {
		{1+0, SIGTRAP}, //soft breakpoint
		{1+1, SIGTRAP}, // hard breakpoint
		{1+2, SIGTRAP}, // watchpoint
		{1+3, SIGTRAP}, // single step
		{1+4, SIGSTOP}, //external interrupt
		{0, 0}			/* Must be last */
};

/* Convert the SpartanMC trap type code to a unix signal number. */

static int __attribute__((section(".traps"))) computeSignal (int tt)
{
	struct hard_trap_info *ht;

	for (ht = hard_trap_info; ht->tt && ht->signo; ht++)
		if (ht->tt == tt)
			return ht->signo;

	return SIGHUP;		/* default for things we don't know about */
}

/*
 * While we find nice hex chars, build an int.
 * Return number of chars processed.
 */

static int __attribute__((section(".traps"))) hexToInt(char **ptr, unsigned int *intValue)//TODO: maybe integrate with util, somewhere
{
	int numChars = 0;
	int hexValue;

	*intValue = 0;

	while (**ptr)
	{
		hexValue = hex(**ptr);
		if (hexValue < 0)
			break;

		*intValue = (*intValue << 4) | hexValue;
		numChars ++;

		(*ptr)++;
	}

	return (numChars);
}

enum brkpt_types {SOFT_BRKPT, HARD_BRKPT, WATCHPOINT};
extern uint18_t regs[18];
extern uint9_t update_regs;
extern uint18_t pushwin[6];
extern uint18_t fixup_brk_addr;

#if defined(HAS_HARDWARE_BREAKPOINTS) || defined(HAS_WATCHPOINTS)
/**
 * iterate over hardware regs and return addresses for first empty / off one
 *
 * returns index, -1 if no empty, -2 if invalid
**/
static int9_t __attribute__((section(".traps"))) set_hdwPoint(uint18_t addr, uint18_t cfg, uint9_t type) {
#ifdef PC_BYTE_ADDRESSES
	addr = addr >> 1;
#endif

	uint18_t idx = (type == WATCHPOINT)? OFFSET_WATCHPOINTS : OFFSET_BREAKPOINTS;

	int9_t i;
	uint18_t current_cfg;

	for (i=0; i <((type == WATCHPOINT)? count_watchpoints : count_breakpoints); i++) {

		__asm__ ("movi2s SFR_DBG_IDX, %[i]\n\t"
                "movs2i %[cfg], SFR_DBG_DAT\n\t"
            : [cfg] "=r" (current_cfg)
            : [i] "r" (idx+ DIST_CFG)
        );

		if (((type == HARD_BRKPT) && !BRK_IS_ON(current_cfg)) || ((type == WATCHPOINT) && !WTCH_IS_ON(current_cfg))) {

            __asm__ ("movi2s SFR_DBG_IDX, %[i]\n\t"
                    "movi2s SFR_DBG_DAT, %[pt]\n\t"
                    "mov r12, %[i]\n\t"
                    "add r12, %[OFFSET]\n\t"
                    "movi2s SFR_DBG_IDX, r12\n\t"
                    "movi2s SFR_DBG_DAT, %[cfg]\n\t"
                    : /* none */
                    : [i] "r" (idx), [pt] "r" (addr), [OFFSET] "r" (DIST_CFG), [cfg] "r" (cfg)
                    : "r12"
            );
            return i;
        }

		idx++;
	}

	return -1;
}

static int9_t __attribute__((section(".traps"))) remove_hdwPoint(uint18_t addr, uint9_t type) {
#ifdef PC_BYTE_ADDRESSES
	addr = addr >> 1;
#endif

    uint18_t idx = (type == WATCHPOINT)? OFFSET_WATCHPOINTS : OFFSET_BREAKPOINTS;

	int9_t kills = 0;
	uint9_t i;
    uint18_t current_pt;
    uint18_t current_cfg;

	for (i=0; i<((type == WATCHPOINT)? count_watchpoints : count_breakpoints); i++) {

        __asm__ (
            "movi2s SFR_DBG_IDX, %[i]\n\t"
            "movs2i %[pt], SFR_DBG_DAT\n\t"
            "mov r12, %[i]\n\t"
            "add r12, %[OFFSET]\n\t"
            "movi2s SFR_DBG_IDX, r12\n\t"
            "movs2i %[cfg], SFR_DBG_DAT\n\t"
            : [cfg] "=r" (current_cfg), [pt] "=r" (current_pt)
            : [i] "r" (idx), [OFFSET] "r" (DIST_CFG)
            : "r12"
        );

		if (((type == WATCHPOINT) && (current_pt == addr)) || ((type == HARD_BRKPT) && (current_pt == addr))) {

			current_cfg = (type == WATCHPOINT)? current_cfg & 0xF : 0;

            __asm__ (
                "movi2s SFR_DBG_IDX, %[i]\n\t"
                "movi2s SFR_DBG_DAT, %[cfg]\n\t"
				: /* none */
				:[i] "r" (idx+ DIST_CFG), [cfg] "r" (current_cfg)
            );
			kills++;
		}
		idx++;
	}

	return kills;
}

static void __attribute__((section(".traps"))) remove_hdwPoints() {
	uint9_t i;

	for (i= OFFSET_BREAK_CFG; i< OFFSET_WATCH_CFG+count_watchpoints; i++) {
		__asm__ (
			"movi r12, 0\n\t"
			"movi2s SFR_DBG_IDX, %[idx]\n\t"
			"movi2s SFR_DBG_DAT, r12\n\t"
			: /* none */
			: [idx] "r" (i)
			: "r12"
		);
	}
}

#endif

#if defined(HAS_SOFT_BREAKPOINTS)
static uint18_t saves[SOFT_SLOTS*3];
//1n: original pc
//2n: original cmd
//3n: jump back

static char __attribute__((section(".traps"))) set_soft_breakpoint(uint18_t addr) { //TODO map modulo to ranges to increase search time if full
    uint9_t i;
    uint18_t addr_pc;
    uint18_t addr_mem;
#ifdef PC_BYTE_ADDRESSES
	addr_pc = addr >> 1;
	addr_mem = addr;
#else
	addr_pc = addr;
	addr_mem = addr << 1;
#endif

	uint18_t * cmd = (uint18_t*)(addr_mem);
	
	int18_t rel_j_back;
//TODO check prev. insn for delay slot, disallow
	if (IS_CMD_BRANCH(*cmd) || IS_CMD_J_LINK(*cmd) || IS_CMD_J_REL(*cmd)) { // impossible src cmd
		return -2; // relative jump / links depend on position in memory //TODO: rewrite displacement so that 'relative' is no more problem, LINK needs to be replaced by manual window shift+call
	}

    for(i=0; i< SOFT_SLOTS; i++) {

        if (saves[3*i+2] == 0x0) {// we look for empty jump entry, because, j some bits guaranteed
			rel_j_back = (addr_pc+1) - (((uint18_t)&saves[3*i+2]) >> 1);
			if (rel_j_back > 4095 || rel_j_back < -4096)
				return -3; // out of range for j xxx;

            saves[3*i+0] = addr_pc;//id by return address bc that is what we'll have when hitting the breakpoint
            saves[3*i+1] = *cmd;
            saves[3*i+2] = CMD_J | (0x1FFF & rel_j_back); // jump back (same window), use NO reg
            *cmd = CMD_TRAP_0;
            return i;
        }
    }

    return -1; // no space
}

static char __attribute__((section(".traps"))) makeup_soft_breakpoint() {
	uint18_t addr_pc = regs[PC]>>1;
	uint9_t i;


#ifdef PC_BYTE_ADDRESSES
	addr_pc = addr_pc >> 1;
#endif

	for(i=0; i< SOFT_SLOTS; i++) {

		if (saves[3*i+0] == addr_pc && saves[3*i+2] != 0x0) {
			fixup_brk_addr = ((uint18_t)&saves[3*i+1])>>1;
			update_regs |= UPDATE_FIXUP_BRK;
			return 0;
		}
	}
	return -1;
}

static signed char __attribute__((section(".traps"))) remove_soft_breakpoint(uint18_t addr) {
    uint9_t i;
    uint18_t addr_pc;
    uint18_t addr_mem;
#ifdef PC_BYTE_ADDRESSES
	addr_pc = addr >> 1;
	addr_mem = addr;
#else
	addr_pc = addr;
	addr_mem = addr << 1;
#endif
	uint18_t * cmd = (uint18_t*)(addr_mem);

	if (*cmd != CMD_TRAP_0)
		return -1;

    for(i=0; i< SOFT_SLOTS; i++) {

		if (saves[3*i+0] == addr_pc && saves[3*i+2] != 0x0) {
			*cmd = saves[3*i+1];
			saves[3*i+2] = 0x0;
			saves[3*i+0] = 0x0;

			if (fixup_brk_addr == ((uint18_t)&saves[3*i+1])>>1) { // we're removing the breakpoint we just hit -> force return to PC-1 to exec said cmd
				fixup_brk_addr = addr;
				update_regs |= UPDATE_FIXUP_BRK;
			}
			return 1;
		}
    }
	return 0;
}

static char __attribute__((section(".traps"))) remove_soft_breakpoints() {
	uint9_t i;
	uint18_t addr_pc;
	uint18_t addr_mem;
	uint18_t * cmd;
	char result = 0;

	for(i=0; i< SOFT_SLOTS; i++) {

		if (saves[3*i+2] != 0x0) {
			addr_pc = saves[3*i+0];
			addr_mem = addr_pc << 1;
			cmd = (uint18_t*)(addr_mem);
			*cmd = saves[3*i+1];
			saves[3*i+2] = 0x0;
			saves[3*i+0] = 0x0;

			if (fixup_brk_addr == ((uint18_t)&saves[3*i+1])>>1) { // we're removing the breakpoint we just hit -> force return to PC-1 to exec said cmd
				fixup_brk_addr = addr_pc;
				update_regs |= UPDATE_FIXUP_BRK;
			}
			result++;
		}
	}
	return result;
}
#endif

static void __attribute__((section(".traps"))) set_breakpoint(char * ptr) {
	uint18_t addr;
	uint9_t wtch_access;

	switch (*ptr++) {
#ifdef HAS_SOFT_BREAKPOINTS
		case '0': {
			char result;
			if (*ptr++ == ','
					&& hexToInt(&ptr, &addr)
					&& *ptr++ == ','
					&& *ptr++ == '4') { //only breakpoints that take up 4 real bytes (2 9bit blocks)
						
						addr >>= 1;

				result = set_soft_breakpoint(addr);
				if (result == 0x1ff)
					strcpy(remcomOutBuffer, "E02");//no space
				else if (result == 0x1fe)
					strcpy(remcomOutBuffer, "E03");//forbidden cmd
				else if (result == 0x1fd)
					strcpy(remcomOutBuffer, "E04");//unreachable for j xxx (return)
				else
				{
					strcpy(remcomOutBuffer, "OK");
				}
			} else {
				strcpy(remcomOutBuffer, "E01");//format
			}
			break;}
#else
		case '0': {
			//Send empty response to signal that this is unsupported
			remcomOutBuffer[0]=0;
			break;
		}
#endif
#ifdef HAS_HARDWARE_BREAKPOINTS
		case '1':
			if (*ptr++ == ','
					&& hexToInt(&ptr, &addr)
					&& *ptr++ == ','
					&& *ptr++ == '4') {
						
						addr >>= 1;

				if (set_hdwPoint(addr, 0x1, HARD_BRKPT) < 0) {
					strcpy(remcomOutBuffer, "E03"); //No more regs
					break;
				}

				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "E01");//format
			}
			break;
#else
		case '1': {
			//Send empty response to signal that this is unsupported
			remcomOutBuffer[0]=0;
			break;
		}
#endif
#ifdef HAS_WATCHPOINTS
		case '2':
			wtch_access = 0x4;
			goto create_watchpoint;
		case '3':
			wtch_access = 0x8;
			goto create_watchpoint;
		case '4':
			wtch_access = 0xC;
		create_watchpoint:
			if (*ptr++ == ','
					&& hexToInt(&ptr, &addr)
					&& *ptr++ == ',') {
						
						addr >>= 1;

				if (*ptr == '1')
					wtch_access |= 0x1;
				else if (*ptr == '2')
					wtch_access |= 0x2;
				else if (*ptr == '3' || *ptr == '4')
					wtch_access |= 0x3;
				else {
					strcpy(remcomOutBuffer, "E04"); // unsupported kind
					break;
				}

				if (set_hdwPoint(addr, 0x10 | wtch_access, WATCHPOINT) < 0) {
					strcpy(remcomOutBuffer, "E03"); //No more regs
					break;
				}

				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "E01");//format
			}
			break;
#endif
		default:
			strcpy(remcomOutBuffer, "E00");
			break;
	}
}

static void __attribute__((section(".traps"))) remove_breakpoint(char * ptr) {
	uint18_t addr;
	uint18_t temp;

	switch (*ptr++) {
#ifdef HAS_SOFT_BREAKPOINTS
		case '0': {
			signed char result;
			if (*ptr++ == ','
					&& hexToInt(&ptr, &addr)
					&& *ptr++ == ','
					&& *ptr++ == '4') {
						
						addr >>= 1;

				result = remove_soft_breakpoint(addr);

				if (result == -1)
					strcpy(remcomOutBuffer, "E02");//was no trap
				else if (result == 0)
					strcpy(remcomOutBuffer, "E03"); //could not restore
				else
					strcpy(remcomOutBuffer, "OK");// {result} removed
			} else {
				strcpy(remcomOutBuffer, "E01");//format
			}
			break; }
#endif
#ifdef HAS_HARDWARE_BREAKPOINTS
		case '1':
			if (*ptr++ == ','
					&& hexToInt(&ptr, &addr)
					&& *ptr++ == ','
					&& *ptr++ == '4') {
						
						addr >>= 1;

				if (remove_hdwPoint(addr, HARD_BRKPT) <= 0) {
					strcpy(remcomOutBuffer, "E04"); // none present
					break;
				}

				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "E01");//format
			}
			break;
#endif
#ifdef HAS_WATCHPOINTS
		case '2':
		case '3':
		case '4':
			if (*ptr++ == ','
					&& hexToInt(&ptr, &addr)
					&& *ptr++ == ','
					&& hexToInt(&ptr, &temp)) {
						
						addr >>= 1;

				if (remove_hdwPoint(addr, WATCHPOINT) <= 0) {
					strcpy(remcomOutBuffer, "E04"); // none present
					break;
				}

				strcpy(remcomOutBuffer, "OK");
			} else {
				strcpy(remcomOutBuffer, "E01");//format
			}
			break;
#endif
		default:
			strcpy(remcomOutBuffer, "E00");
			break;
	}
}


/*
 * This function does all command processing for interfacing to gdb.  It
 * Regs must have already been captured
 * reason is internal code:
 * 0 -> mem_trap
 * 1 -> hard_trap
 * 2 -> watchpoint
 * 3 -> single_step (used for init)
 * 4 -> interrupt
 */

void __attribute__((section(".traps"))) handle_exception(uint9_t reason)
{
	uint9_t sigval;
	uint18_t addr;
	unsigned int length;
	char *ptr;

	uint18_t * registers = regs;

	/* reply to host that an exception has occurred */
	sigval = computeSignal(1+reason);
	//sigval = tt;
	ptr = remcomOutBuffer;

	// when we have reached a breakpoint, we need to acknowledge it with a stop-reply packet (either S, or T, T carries user defined regs to simplify communication)
	*ptr++ = 'T';//SpartanMC: changed from T because we do not yet know the special registers
	*ptr++ = hexchars[(sigval >> 4) & 0xf];
	*ptr++ = hexchars[sigval & 0xf];

	//Report PC
	*ptr++ = hexchars[(PC >> 4) & 0xf];
	*ptr++ = hexchars[PC & 0xf];
	*ptr++ = ':';
	ptr = mem18_2_hex32(&registers[PC], ptr, 1);
	*ptr++ = ';';

#ifdef DEBUG_HDW_DATA
	int i=0;
	uint18_t brk[count_watchpoints+count_breakpoints+2];
	uint18_t * current;
	uint18_t current_data;
#endif

	if (reason == 4) {
		char c;
		getDebugChar_nb(&c);
	} else if (reason == 2){//watchpoint
		strcpy(ptr, "watch");
		ptr+=5;
		*ptr++ = ':';
		__asm__ (
			"movi %[trapped_addr],1\n\t"
			"movi2s SFR_DBG_IDX, %[trapped_addr]\n\t"
			"movs2i %[trapped_addr], SFR_DBG_DAT\n\t"
		:[trapped_addr] "=r" (addr)
		:/*none*/);
		ptr = mem18_2_hex32(&addr, ptr, 1);
		*ptr++ = ';';
	}
	*ptr++ = 0;

	putpacket(remcomOutBuffer);
	
	if (reason == 0) {// soft breakpoint
#ifdef HAS_SOFT_BREAKPOINTS
		if (makeup_soft_breakpoint()) {
			strcpy(remcomOutBuffer, "E33"); //cannot find original data // throws off GDB, violates protocol, but at this point target is unreliable anyway
			putpacket(remcomOutBuffer);
		}
#else
		fixup_brk_addr = (regs[PC]+2)>>1;
#endif
	}

	while (1)
	{
		remcomOutBuffer[0] = 0;

		ptr = getpacket();
		switch (*ptr++)
		{
			case '?':
				remcomOutBuffer[0] = 'S';
				remcomOutBuffer[1] = hexchars[(sigval >> 4) & 0xf];
				remcomOutBuffer[2] = hexchars[sigval & 0xf];
				remcomOutBuffer[3] = 0;
				break;
#ifdef DEBUG_HDW_DATA
			case 'd':		/* toggle debug flag */ // misused here
				current = &brk[0];
				for (i=0; i< OFFSET_WATCH_CFG+count_watchpoints; i++) {

					__asm__ (
							"movi2s SFR_DBG_IDX, %[i]\n\t"
							"movs2i %[pt], SFR_DBG_DAT\n\t"
							: [pt] "=r" (current_data)
							: [i] "r" (i)
					);

					*current++ = current_data;
				}

				mem18_2_hex32(brk, remcomOutBuffer, OFFSET_WATCH_CFG+ count_watchpoints);
				break;
#endif
			case 'g':		/* return the value of the CPU registers */
				ptr = remcomOutBuffer;
				ptr = mem18_2_hex32(registers, ptr, NUMREGS); /* all GP regs, SFR_STATUS, PC_NEXT */

				break;

			case 'G':	   /* set the value of the CPU registers - return OK */
				hex32_2_mem18(ptr, registers, NUMREGS); /* all GP regs */
				update_regs |= UPDATE_GPRS | UPDATE_RETURN_ADDR;

				strcpy(remcomOutBuffer,"OK");
				break;
			case 'p':// p[idx] //read single reg
				if (hexToInt(&ptr, &addr)) {
					if (addr < NUMREGS) {
						ptr = mem18_2_hex32((uint18_t*)&registers[addr], remcomOutBuffer, 1);
					} else {
						strcpy(remcomOutBuffer, "E02"); //reg out of bounds
					}
				} else {
					strcpy(remcomOutBuffer, "E01");// format
				}
				break;
			case 'P':// P[idx]=[data] //write single Reg
				if (hexToInt(&ptr, &addr)
						&& *ptr++ == '=') {
					if (addr < NUMREGS) {
						if (hex32_2_mem18(ptr, (uint18_t*)&registers[addr], 1)) {
							update_regs |= (addr == PC) ? UPDATE_RETURN_ADDR : UPDATE_GPRS;
							strcpy(remcomOutBuffer, "OK");
						} else
							strcpy(remcomOutBuffer, "E03"); //memory
					} else {
						strcpy(remcomOutBuffer, "E02"); //reg out of bounds
					}
				} else {
					strcpy(remcomOutBuffer, "E01");// format
				}
				break;
			case 'w': // w[idx] read all local regs of window [idx]
			{
				uint18_t window[12]; // for convenience return all local regs, even non unique
				if (hexToInt(&ptr, &addr)) {
					if (addr < 125) {

						capture_window(window, addr);

						ptr = mem18_2_hex32(window, remcomOutBuffer, 12);
					} else
						strcpy(remcomOutBuffer, "E02"); // window
				} else
					strcpy(remcomOutBuffer, "E01"); // format
			}
				break;
			//~ case 'W': // push a window on the stack, that sets up an inferior call, only local and output regs accepted
			//~ {// input regs must not be touched, as they belong to __THIS__ function
				//~ if (*ptr++ == 'p') {
						//~ hex32_2_mem18(ptr, pushwin, 6);
						//~ update_regs |= UPDATE_PUSH_REGWINDOW;
						//~ strcpy(remcomOutBuffer, "OK");
				//~ } else
					//~ strcpy(remcomOutBuffer, "E01"); // format
			//~ }
				//~ break;
			case 'm':	  /* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */ // AA..AA is just a single address
				/* Try to read %x,%x.  */

				if (hexToInt(&ptr, &addr)
						&& *ptr++ == ','
						&& hexToInt(&ptr, &length))
				{
					char* result;
					int count;
					
					if (length > MAX_DUMP_BYTES) {// 9bit block, every 2 blocks require 4 bytes
						strcpy(remcomOutBuffer, "E04"); //too large for buffer
						break;
					}
					
					addr >>= 1;
					
					result = 0;
					
					if (length >= 4 && !(length & 0x3) && !(addr & 0x1)) {
						count = length >> 2;
						result = mem18_2_hex32((uint18_t *)addr, remcomOutBuffer, count);
						//length -= count << 2;
						//addr += count << 1;
					} else if (length >= 2 && !(length & 0x1)) {
						count = length >> 1;
						result = mem9_2_hex16((uint9_t *) addr, remcomOutBuffer, count);
						//~ addr += count;
						//~ length -= count << 1;
					} else if (length > 0) {	
						result = mem8_2_hex8((uint9_t*) addr, remcomOutBuffer, length);
					}
					
					if (result)
						break;

					strcpy (remcomOutBuffer, "E03"); // read returned 0, 
				}
				else
					strcpy(remcomOutBuffer,"E01"); // format
				break;

			case 'M': /* MAA..AA,LLLL: Write LLLL bytes at address AA.AA return OK */
				/* Try to read '%x,%x:'.  */

				if (hexToInt(&ptr, &addr)
						&& *ptr++ == ','
						&& hexToInt(&ptr, &length)
						&& *ptr++ == ':')
				{
					void * mem = 0;
					uint9_t count;
					
					addr >>= 1;

					if (length >= 4 && !(length & 0x3) && !(addr & 0x1)) {
						count = length >> 2;
						mem = hex32_2_mem18(ptr, (uint18_t *)addr, count);
						//~ length -= count << 2;
					} else	if (length >= 2 && !(length & 0x1)) {
						count = length >> 1;
						mem = hex16_2_mem9(ptr, (uint9_t *)addr, count);
						//~ length -= count << 1;
					} else if (length > 0) {
						mem = hex8_2_mem8(ptr, (uint9_t*)addr, length);
					}
						
					if (mem)
						strcpy(remcomOutBuffer, "OK");
					else
						strcpy(remcomOutBuffer, "E03");
				}
				else
					strcpy(remcomOutBuffer, "E02"); // format
				break;

			case 'c':    /* cAA..AA    Continue at address AA..AA(optional) */
				/* try to read optional parameter, pc unchanged if no parm */

				if (hexToInt(&ptr, &addr))
				{
					regs[PC] = addr;
					update_regs |= UPDATE_RETURN_ADDR;
				}

				return;
			case 's': //single step
				if (single_step_support) {
			
					if (hexToInt(&ptr, &addr))
					{
						regs[PC] = addr;
						update_regs |= UPDATE_RETURN_ADDR;
					}
					
					update_regs |= SINGLE_STEP_NEXT;
					
					return;
				} else {
					strcpy(remcomOutBuffer, "E00"); // No hardware single step support
					break; //TODO: implement single stepping in software? should be done in GDB, requires rewrite, GDB needs to handle this error here
				}
			case 'Z':// set Breakpoints
				set_breakpoint(ptr);
				break;
			case 'z'://remove breakpoint
				remove_breakpoint(ptr);
				break;
				/* kill the program */
			case 'k' :		/* remove all memory breakpoints */
#ifdef HAS_SOFT_BREAKPOINTS
					remove_soft_breakpoints();
#endif
#if defined(HAS_HARDWARE_BREAKPOINTS) || defined(HAS_WATCHPOINTS)
					remove_hdwPoints();
#endif
				__attribute__ ((fallthrough));
			case 'r':		/* Reset */
				update_regs |= UPDATE_RESET_REGWINDOW;
				return;
			default:
				remcomOutBuffer[0] = 0;
				break;
		}			/* switch */

		/* reply to the request */
		putpacket(remcomOutBuffer);
	}
}


//#pragma GCC pop_options
