/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

#ifndef _MUTEX_
#define _MUTEX_

#include <mutex.h>
#include <semaphore.h>
#include "globalvariables.h"

/*
	attempts to lock the mutex
	if the mutex is already locked, it waits until it is opened
		returns once it successfully acquires the lock
*/
void wait(mutex* const m) {
	__asm(
		"movi		r8,		"LOCKED_S"	\n"			//load locked state to r8
		"wait_try%=:						\n"
		"swap		r8,		0(%0)		\n"			//swap with memory
		"seqi		r8,		"LOCKED_S"	\n"			//compare memory data to locked state
		"beqzc		wait_success%=		\n"			//if wasnt locked => go to end, else carry on
		"movi		r8,		"LOCKED_S"	\n"			//reload locked state to r8 while wait for previous command to execute (does no harm if comp was successful)

												//wait was unsuccessfull! do busy waiting
		"l18		r9,		0(%0)		\n"			//reload memory data (does no harm if comp was successfull, but possible stall)
		"wait_read%=:						\n"
		"seqi		r9,		"LOCKED_S"	\n"			//compare again to locked state
		"beqzc		wait_try%=			\n"			//if mutex isnt locked, try to close again

		"nop							\n"			//necessary nop
		"j			wait_read%=			\n"			//mutex still locked, go to wait loop
		"l18		r9,		0(%0)		\n"			//reload memory data

		"wait_success%=:					\n"



		: //output ops
		: "r" (m)
		: "r8", "r9", "cc", "memory"
	);
}

/*
	releases the mutex,
	call only on mutex whose lock you currently hold!
*/
void signal(mutex* const m) {
	__asm(
		"movi		r8,		"OPEN_S"	\n"
		"s18		0(%0),	r8			\n"
		: //output ops
		: "r" (m)
		: "r8", "memory"
	);
}

/*
	attempts to decrease the semaphore by 1,
	it the semaphore is less or equal to 0, it waits until another core increases the semaphore
		returns once the semaphore is decreased to a number greater or equal to zero
*/
void down(semaphore* s) {
	down_try:

	wait(&s->m);				//acquire lock
	uint18_t x = s->counter;	//read counter
	if(x > 0) {					//down success if greater 0
		s->counter = x - 1;		//dec
        signal(&s->m);			//release lock
        return;
	}
						//down unsuccessfull
	signal(&s->m);			//release lock
	while(s->counter == 0);	//wait for counter != 0
	goto down_try;			//try again

}

/*
	increases the semaphore by 1,
*/
void up(semaphore* s) {
	wait(&s->m);  	//acquire lock
	s->counter++;	//inc
	//should I worry about overflows of counter?
	signal(&s->m);	//release lock
}

/*
	returns the current value of the semaphore
*/
uint18_t value(const semaphore* s) {
	return s->counter;
}

#endif
