/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#include <perf.h>


static void conf_ccnt(int n, struct ccnt_conf *conf)
{
	uint18_t reg_idx, reg_dat;

	reg_idx = n << 4;
	reg_dat = (conf->prescaler & 0x0f) << 4 |
	          (conf->intr_filter & 0x01) << 2 |
	          2; // Reset counter.

	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_WRITE(reg_dat);
}

static void conf_evcnt(int n, struct evcnt_conf *conf)
{
	uint18_t reg_idx, reg_dat;

	reg_idx = (n << 4) | (1 << 8);
	reg_dat = ((uint18_t)conf->prescaler & 0x0f) << 12 |
	          (uint18_t)conf->edge_detect << 10 |
	          (uint18_t)conf->intr_filter << 9 |
	          1 << 8 | // Reset counter.
	          (conf->tie_ccnt ? (((uint18_t)(conf->tied_ccnt & 0x03) | 4)) << 4 : 0) |
	          conf->eventnum;

	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_WRITE(reg_dat);
}

int perf_init(struct perf_conf *conf)
{
	int i, num_cyclecnt, num_eventcnt;
	uint18_t reg_idx, reg_dat;

	reg_idx = 0;
	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_READ(reg_dat);
	num_cyclecnt = (reg_dat >> 9) & 0x07;
	num_eventcnt = (reg_dat >> 12) & 0x1f;

	// Check if the requested number of counters is available:
	if(conf->num_ccnt > num_cyclecnt || conf->num_evcnt > num_eventcnt) {
		return -1;
	}
	
	for(i = 0; i < conf->num_ccnt; i++) {
		conf_ccnt(i, &conf->cycle[i]);
	}
	
	for(i = 0; i < conf->num_evcnt; i++) {
		conf_evcnt(i, &conf->event[i]);
	}
	
	return 0;
}

int perf_init_ccnt(int n, struct ccnt_conf *conf)
{
	int num_cyclecnt;
	uint18_t reg_idx, reg_dat;

	reg_idx = 0;
	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_READ(reg_dat);
	num_cyclecnt = (reg_dat >> 9) & 0x07;

	if(n >= num_cyclecnt) {
		return -1;
	}

	conf_ccnt(n, conf);

	return 0;
}

int perf_init_evcnt(int n, struct evcnt_conf *conf)
{
	int num_eventcnt;
	uint18_t reg_idx, reg_dat;

	reg_idx = 0;
	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_READ(reg_dat);
	num_eventcnt = (reg_dat >> 12) & 0x1f;

	if(n >= num_eventcnt) {
		return -1;
	}

	conf_evcnt(n, conf);

	return 0;
}

void perf_reset(void)
{
	int i, num_cyclecnt, num_eventcnt;
	uint18_t reg_idx, reg_dat;
	
	reg_idx = 0;
	
	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_READ(reg_dat);
	num_cyclecnt = (reg_dat >> 9) & 0x07;
	num_eventcnt = (reg_dat >> 12) & 0x1f;

	for(i = 0; i < num_cyclecnt; i++) {
		reg_idx = i << 4;
		PCNT_IDX_WRITE(reg_idx);
		PCNT_DAT_READ(reg_dat);
		reg_dat |= 2;
		PCNT_DAT_WRITE(reg_dat);
	}

	for(i = 0; i < num_eventcnt; i++) {
		reg_idx = (i << 4) | (1 << 8);
		PCNT_IDX_WRITE(reg_idx);
		PCNT_DAT_READ(reg_dat);
		reg_dat |= 0x100;
		PCNT_DAT_WRITE(reg_dat);
	}
	
	return;
}

void perf_reset_cycle(int n)
{
	uint18_t reg_idx, reg_dat;
	
	reg_idx = n << 4;
	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_READ(reg_dat);
	reg_dat |= 2;
	PCNT_DAT_WRITE(reg_dat);
	
	return;
}

void perf_reset_event(int n)
{
	uint18_t reg_idx, reg_dat;
	
	reg_idx = (n << 4) | (1 << 8);
	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_READ(reg_dat);
	reg_dat |= 0x100;
	PCNT_DAT_WRITE(reg_dat);
	
	return;
}

#define ADDITIONAL_CYCLES 2
uint36_t perf_read_ccnt(int n, uint9_t *overflow)
{
	uint18_t reg_idx, reg_dat;
	uint36_t count;

	reg_idx = n << 4;

	if(overflow) {
		PCNT_IDX_WRITE(reg_idx);
		PCNT_DAT_READ(reg_dat);
		*overflow = (reg_dat >> 8) & 1;
	}

	PCNT_IDX_WRITE(reg_idx + 1);
	PCNT_DAT_READ(reg_dat);
	count = reg_dat;

	PCNT_IDX_WRITE(reg_idx + 2);
	PCNT_DAT_READ(reg_dat);
	count |= (uint36_t)reg_dat << 18;

	// Adjust cycle counter.
	if(count >= ADDITIONAL_CYCLES) {
		count -= ADDITIONAL_CYCLES;
	}


	return count;
}

uint36_t perf_read_evcnt(int n, uint9_t *overflow)
{
	uint18_t reg_idx, reg_dat;
	uint36_t count;

	reg_idx = (uint18_t)n << 4 | (1 << 8);

	if(overflow) {
		PCNT_IDX_WRITE(reg_idx);
		PCNT_DAT_READ(reg_dat);
		*overflow = (reg_dat >> 11) & 1;
	}

	PCNT_IDX_WRITE(reg_idx + 1);
	PCNT_DAT_READ(reg_dat);
	count = reg_dat;

	PCNT_IDX_WRITE(reg_idx + 2);
	PCNT_DAT_READ(reg_dat);
	count |= (uint36_t)reg_dat << 18;

	return count;
}

void perf_read(struct perf_result *res)
{
	int i, num_cyclecnt, num_eventcnt;
	uint18_t reg_idx, reg_dat;
	
	reg_idx = 0;
	PCNT_IDX_WRITE(reg_idx);
	PCNT_DAT_READ(reg_dat);
	num_cyclecnt = (reg_dat >> 9) & 0x07;
	num_eventcnt = (reg_dat >> 12) & 0x1f;
	
	for(i = 0; i < num_cyclecnt; i++) {
		res->cycles[i].counter = perf_read_ccnt(i, &res->cycles[i].overflow);
	}
	
	for(i = 0; i < num_eventcnt; i++) {
		 res->events[i].counter = perf_read_evcnt(i, &res->events[i].overflow);
	}

	return;
}

void perf_auto_init(void)
{
	struct ccnt_conf c;

	c.prescaler = 0;
	c.intr_filter = 0;
	perf_init_ccnt(0, &c);

	return;
}

void perf_auto_stop_(int n, struct perf_auto_result *r)
{
	perf_read(&r[n].result);
	perf_reset();

	return;
}

void microstreams_eval_init(void)
{
	struct ccnt_conf c;

	c.prescaler = 0;
	c.intr_filter = 0;

	perf_init_ccnt(0, &c);
}

void microstreams_eval_restart(void)
{
	perf_reset_cycle(0);	// Set counter value to 0.
	perf_start(0);			// Start counter.
}

uint36_t microstreams_eval_stop(void)
{
	perf_stop(0);

	return perf_read_ccnt(0, 0);
}
