/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#include <sleep.h>
#include <stdio.h>
#include "ethlib.h"
#include <memdump.h>

volatile unsigned int * ethernet_tx_init_frame(ethernet_tx_regs_t * reg, ethernet_tx_dma_t * dma) {
    //set first byte of tx port to 0 and set AddrCnt to 0
    //prevent accidently start of a transmission
    dma->x[0] = 0;
    reg->offset = 0;

    return dma->x+1;
}


unsigned int ethernet_tx_is_sending(ethernet_tx_regs_t * reg, ethernet_tx_dma_t * dma) {
    return dma->x[reg->offset/2];
}

int ethernet_tx_commit_frame(ethernet_tx_regs_t * reg, ethernet_tx_dma_t * dma, unsigned int size) {
    //check if args are valid
    //packet have to be between 64 and 2004
    //4 byte are appended (FCS)
    if((size<60) || (size>2000)) return E_INVALID_ARG;

    unsigned int maxindex = size/2+size%2;
    unsigned int offset = reg->offset/2;
    //set byte after frame to 0, prevent start of another transmission
    dma->x[offset+maxindex+1] = 0;

    //start transmission
    dma->x[offset]=size;

    return 0;
}

int ethernet_tx_send_frame(ethernet_tx_regs_t * reg, ethernet_tx_dma_t * dma, unsigned int *data, unsigned int size){
    ethernet_tx_init_frame(reg,dma);
    unsigned int maxindex = size/2+size%2;

    //copy data
    for(unsigned int i=0;i<maxindex;i++){
	dma->x[i+1] = data[i];
    }

    return ethernet_tx_commit_frame(reg,dma,size);
}

void ethernet_rx_init(ethernet_rx_regs_t * reg, ethernet_rx_dma_t * dma) {
    dma->x[0]=0;
    reg->offset = 0;
}

volatile unsigned int * ethernet_rx_get_frame_start(ethernet_rx_regs_t * reg, ethernet_rx_dma_t * dma, unsigned int * size) {
    //The hardware accesses the memory in 9 bit words. Therefore, it writes the size in two parts. It is possible
    //that the hardware has written one half word when the software accesses the size. It is required to read the
    //size twice to be sure it has been written completely.
    if (!dma->x[reg->offset]) {
	*size = 0;
	return 0;
    }
    *size = dma->x[reg->offset];
    return dma->x+reg->offset+1;
}

void ethernet_rx_frame_free(ethernet_rx_regs_t * reg, ethernet_rx_dma_t * dma){
	unsigned int offset = reg->offset;
	unsigned int size = dma->x[offset];
	
	//set the ByteCnt,if a frame is in the buffer
	if(size!=0){
		reg->offset = offset + size/2 + size%2 + 1;
	}
}


//*****************************
//	MDIO
//*****************************

void mdio_wait(const mdio_regs_t * const mdio){
	while(!(mdio->address & MDIO_ADDRESS_IDLE));
}

int mdio_write(mdio_regs_t * const  mdio, const unsigned char phy, const unsigned char reg, const unsigned int value){
	mdio_wait(mdio);
	if((reg>0x1F) || (phy>0x1F)){
		return E_INVALID_ARG;
	}

	//set port and register
	mdio->address = phy<<MDIO_ADDRESS_PHY_ADDR_OFFSET | reg<<MDIO_ADDRESS_REG_ADDR_OFFSET;

	mdio->data = MDIO_DATA_WEN | value;

	return 0;
}

int mdio_read(mdio_regs_t * const  mdio, const unsigned char phy, const unsigned char reg, unsigned int * const value){
	mdio_wait(mdio);
	if((reg>0x1F) || (phy>0x1F)){
		printf("fail at %x.%x\n",reg,phy);
		return E_INVALID_ARG;
	}

	//set port and register
	mdio->address = phy<<MDIO_ADDRESS_PHY_ADDR_OFFSET | reg<<MDIO_ADDRESS_REG_ADDR_OFFSET;

	//start read by writing a value to data
	mdio->data = 0;

	//wait for completion
	sleep(20);
	mdio_wait(mdio);

	*value = mdio->data & MDIO_DATA_RESULT_MASK;

	return 0;
}

char* format_mac_address_words(const unsigned int high, const unsigned int mid, const unsigned int low, char * printf_buffer) {
#define HI(x) ((x>>8)&0xFF)
#define LO(x) (x&0xFF)

	//We rely on the buffer having enough bits in front of the returned address
	char* b = printf_format_bits(LO(low), 4, 2, '0',1, printf_buffer);
	*(--b) =':';
	b = printf_format_bits_buf(HI(low), 4, 2, '0',b,1);
	*(--b) =':';

	b = printf_format_bits_buf(LO(mid), 4, 2, '0',b,1);
	*(--b) =':';
	b = printf_format_bits_buf(HI(mid), 4, 2, '0',b,1);
	*(--b) =':';

	b = printf_format_bits_buf(LO(high), 4, 2, '0',b,1);
	*(--b) =':';
	b = printf_format_bits_buf(HI(high), 4, 2, '0',b,1);

#undef HI
#undef LO

	return b;
}


char* format_mac_address(const ethernet_rx_regs_t *const port, char * printf_buffer) {

	return format_mac_address_words(port->mac3, port->mac2, port->mac1, printf_buffer);
}

char* format_mac_address_mem(ethernet_rx_dma_t * dma, const unsigned int * macAddress, char * printf_buffer) {
	return format_mac_address_words(
		ethernet_rx_read(dma,macAddress),
		ethernet_rx_read(dma,macAddress+1),
		ethernet_rx_read(dma,macAddress+2),
		printf_buffer
	);
}

#define RETURN_IF_NEQ_0(x) {int _r_ = (x); if (_r_) return _r_;}

const int REGCR = 0xD;
const int ADDAR = 0xE;
const int DEVAD = 0x1F;

int mdio_indirect_addr(mdio_regs_t * const mdio, const unsigned char phy, const unsigned int reg) {
    RETURN_IF_NEQ_0(mdio_write(mdio,phy,REGCR,DEVAD)); //We want to set reg address
    RETURN_IF_NEQ_0(mdio_write(mdio,phy,ADDAR,reg)); //Set reg address
    return mdio_write(mdio,phy,REGCR, 0x4000 | DEVAD); //We want to access reg content, no increment
}

int mdio_read_indirect(mdio_regs_t * const mdio,const unsigned char phy, const unsigned int reg, unsigned int * const value) {
    RETURN_IF_NEQ_0(mdio_indirect_addr(mdio,phy,reg));
    return mdio_read(mdio,phy,ADDAR,value);
}
int mdio_write_indirect(mdio_regs_t * const mdio,const unsigned char phy, const unsigned int reg, const unsigned int value) {
    RETURN_IF_NEQ_0(mdio_indirect_addr(mdio,phy,reg));
    return mdio_write(mdio,phy,ADDAR,value);
}
