/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#include <peripherals/memguard.h>
#include <stdio.h>
#include <perf.h>




void memguard_configure_watcher(memguard_regs_t *memguard, unsigned int watcher_idx, void *start, unsigned int size,
			   unsigned int flags) {

    memguard->watcher_idx = watcher_idx;

    //disable watcher before changing settings to avoid spurious errors
    memguard->watcher_conf = 0;

    memguard->watcher_offs = start;
    memguard->watcher_size = size;
    memguard->watcher_conf = flags | MEMGUARD_WATCHER_ENABLE;

}


void memguard_configure_watcher_if_avail(memguard_regs_t *memguard, unsigned int watcher_idx, unsigned int numWatchers, void *start, unsigned int size,
				unsigned int flags) {
    if (watcher_idx<numWatchers)
	memguard_configure_watcher(memguard, watcher_idx, start, size, flags);
}

void memguard_enable(memguard_regs_t *memguard, int enable) {
	memguard->control = enable;
}




void memguard_print_internal(memguard_regs_t * memguard, int trap) {
    int enabled = memguard->control;
    memguard->control = 0;
    //Status gets cleared on access, save it to a variable
    unsigned int status = memguard->status;
    unsigned int type = status & MEMGUARD_STAT_ERR_MASK;

    //Called without an error present?
    if (type==0)
	return;

    printf("memory violation at 0x%X detected by %d at pc 0x%X on ", (unsigned int) memguard->error_addr,status>>2, (unsigned int) memguard->error_pc);
    if (type == MEMGUARD_STAT_ERR_CODE)
	putsn("execute");
    else if (type == MEMGUARD_STAT_ERR_RD)
	putsn("read");
    else
	putsn("write");
    uint36_t cycles = perf_read_ccnt(0, 0);
    if (cycles != 0) {
    	printf(" in cycle %ld", cycles);
    }
    putchar('\n');

    if (trap)
	asm volatile ("trap 0");
    

    if (type == MEMGUARD_STAT_ERR_CODE)
	//We don't know how to continue execution, so just stay here.
	while(1);
    memguard->control = enabled;
}

void memguard_info(memguard_regs_t *memguard, unsigned int numWatchers) {
    for (unsigned int i=0;i<numWatchers;i++) {
	memguard->watcher_idx = i;
	unsigned int conf = memguard->watcher_conf;
	printf("Watcher %d: ",i);
	if (conf & MEMGUARD_WATCHER_ENABLE)
	    printf("enabled\n");
	else {
	    printf("disabled\n");
	    continue;
	}

	printf("  allowed:");
	if (conf & MEMGUARD_ALLOW_CODE)
	    printf(" code");
	if (conf & MEMGUARD_ALLOW_RD)
	    printf(" read");
	if (conf & MEMGUARD_ALLOW_WR)
	    printf(" write");
	printf("\n  start: %X\n  size: %X\n",(unsigned int) memguard->watcher_offs, memguard->watcher_size);
    }
}
// For spartanmc_0_spartanmc_mem_local_0:
extern int __startof__startup asm(".startof..startup");
extern int __sizeof__startup asm(".sizeof..startup");
extern int __startof__ctors asm(".startof..init_array");
extern int __sizeof__ctors asm(".sizeof..init_array");
extern int __sizeof__dtors asm(".sizeof..fini_array");
extern int __startof__init asm(".startof..init");
extern int __sizeof__init asm(".sizeof..init");
extern int __sizeof__fini asm(".sizeof..fini");
extern int __sizeof__text asm(".sizeof..text");
extern int __startof__data asm(".startof..data");
extern int __sizeof__data asm(".sizeof..data");
extern int __startof__rodata asm(".startof..rodata");
extern int __sizeof__rodata asm(".sizeof..rodata");
extern int __startof__traps asm(".startof..traps");
extern int __sizeof__traps asm(".sizeof..traps");
extern int __startof__intr_tab asm(".startof..intr_tab");
extern int __sizeof__intr_tab asm(".sizeof..intr_tab");
extern int __startof_fini asm(".startof.fini");
extern int __sizeof_fini asm(".sizeof.fini");

void memguard_configure_default(memguard_regs_t *memguard, unsigned int numWatchers) {
    //Generated by InitramJ from ELF file
    //Sections .startup
    memguard_configure_watcher_if_avail(memguard, 0, numWatchers, CODE_TO_DATA(&__startof__startup), (((unsigned int)&__sizeof__startup))/2, MEMGUARD_ALLOW_CODE);
    //Sections .ctors, .dtors
    memguard_configure_watcher_if_avail(memguard, 1, numWatchers, (&__startof__ctors), (((unsigned int)&__sizeof__ctors)+((unsigned int)&__sizeof__dtors))/2, MEMGUARD_ALLOW_RD | MEMGUARD_ALLOW_WR);
    //Sections .init, .fini, .text
    memguard_configure_watcher_if_avail(memguard, 2, numWatchers, CODE_TO_DATA(&__startof__init), (((unsigned int)&__sizeof__init)+((unsigned int)&__sizeof__fini)+((unsigned int)&__sizeof__text))/2, MEMGUARD_ALLOW_CODE);
    //Sections .data
    memguard_configure_watcher_if_avail(memguard, 3, numWatchers, (&__startof__data), (((unsigned int)&__sizeof__data))/2, MEMGUARD_ALLOW_RD | MEMGUARD_ALLOW_WR);
    //Sections .rodata
    memguard_configure_watcher_if_avail(memguard, 4, numWatchers, (&__startof__rodata), (((unsigned int)&__sizeof__rodata))/2, MEMGUARD_ALLOW_RD);
    //Sections .traps
    memguard_configure_watcher_if_avail(memguard, 5, numWatchers, CODE_TO_DATA(&__startof__traps), (((unsigned int)&__sizeof__traps))/2, MEMGUARD_ALLOW_CODE);
    //Sections .intr_tab
    memguard_configure_watcher_if_avail(memguard, 6, numWatchers, (&__startof__intr_tab), (((unsigned int)&__sizeof__intr_tab))/2, MEMGUARD_ALLOW_RD | MEMGUARD_ALLOW_WR);
    //Sections fini
    memguard_configure_watcher_if_avail(memguard, 7, numWatchers, CODE_TO_DATA(&__startof_fini), (((unsigned int)&__sizeof_fini))/2, MEMGUARD_ALLOW_CODE);
}
