/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#include <rtos_internal.h>


typedef struct free_block_info {
	rtos_list_item_t list_item;
	size_t size;
} free_block_info_t;

typedef struct allocated_block_info {
	size_t size;
} allocated_block_info_t;


static DEFINE_INITIALIZED_EMPTY_LIST(free_list);

#define listed_free_block( _list_item ) rtos_list_item_owner(_list_item, free_block_info_t, list_item)

static DEFINE_INITIALIZED_MUTEX_SEMAPHORE(heap_sem);


__attribute__((section(".traps"))) void initialize_heap_( void *heap, size_t size )
{
	free_block_info_t *initial_free_block;

	if (size >= sizeof(free_block_info_t))
	{
		initial_free_block = (free_block_info_t *)heap;
		initial_free_block->size = size;
		rtos_list_add_last(&free_list, &initial_free_block->list_item);
	}
}

__attribute__((section(".traps"))) void *malloc_( size_t size )
{
	rtos_list_item_t *iterator;
	free_block_info_t *free_block;
	allocated_block_info_t *new_allocated;
	free_block_info_t *new_free;
	size_t free_size, dif_size;

	size = ROUNDED_UP_EVEN(size + sizeof(allocated_block_info_t));
	// check if size overflowed or was zero:
	if( size <= sizeof(allocated_block_info_t) ) return NULL;
	
	// minimal block size, that it can be freed again later:
	if( size < sizeof(free_block_info_t) )
		size = sizeof(free_block_info_t);
	
	semaphore_down(&heap_sem);
	
	for( iterator = rtos_list_first(&free_list);
		iterator != &free_list;
		iterator = rtos_list_next(iterator))
	{
		free_block = listed_free_block(iterator);
		free_size = free_block->size;

		if( free_size >= size )
		{
			// this block is large enough
			dif_size = free_size - size;
			// check if there is enough space for a new free block
			if( dif_size >= sizeof(free_block_info_t) )
			{
				new_free = (free_block_info_t *) ((char *) free_block + size);
				new_free->size = dif_size;
				rtos_list_insert_between(&new_free->list_item, rtos_list_previous(iterator), rtos_list_next(iterator));
				// this also removes the old free block from the list
			}
			else
			{
				rtos_list_remove(iterator);
				size = free_size;
			}
			
			new_allocated = (allocated_block_info_t *)free_block;
			new_allocated->size = size;

			semaphore_up(&heap_sem);
			return new_allocated + 1;
		}
	}

	// no place found
	semaphore_up(&heap_sem);
	return NULL;
}

__attribute__((section(".traps"))) void free_( void *ptr )
{
	allocated_block_info_t *to_free;
	size_t size;
	rtos_list_item_t *next, *previous;
	free_block_info_t *new_free_block;

	if (ptr == NULL) return;
	
	to_free = (allocated_block_info_t *) ptr - 1;
	size = to_free->size;
	
	semaphore_down(&heap_sem);
	
	// find the right position to insert the new free block in the list
	for (next = rtos_list_first(&free_list); 1; next = rtos_list_next(next))
	{	
		if (next == &free_list)
		{
			// there are no free blocks after this
			previous = rtos_list_previous(next);
			break;
		}
		else if ((char *) listed_free_block(next) > (char *) to_free)
		{
			// there is a free block after this
			previous = rtos_list_previous(next); // do this before modifying "next"
			
			// check if the new free block can be combined with the next one
			if ((char *) to_free + size == (char *) listed_free_block(next))
			{
				size += listed_free_block(next)->size;
				next = rtos_list_next(next);
			}
			
			break;
		}
	}

	// if there is a previous free block, check if it can be combined with the new one
	if (previous != &free_list
		&& (char *) listed_free_block(previous) + listed_free_block(previous)->size == (char *) to_free)
	{
		new_free_block = listed_free_block(previous);
		size += listed_free_block(previous)->size;
		previous = rtos_list_previous(previous);
	}
	else new_free_block = (free_block_info_t *) to_free;

	// this removes the connected blocks from the list and inserts the new one:
	rtos_list_insert_between(&new_free_block->list_item, previous, next);
	new_free_block->size = size;
	
	semaphore_up(&heap_sem);
}
