/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#include <rtos_internal.h>


static regfile_block_t out_of_bounds = {
	.lower_limit = -1,
	.upper_limit = -1,
	.list_item = {&out_of_bounds.list_item, &out_of_bounds.list_item}
};

static regfile_block_t *next_block = &out_of_bounds;

#define listed_regs( _list_item ) rtos_list_item_owner(_list_item, regfile_block_t, list_item)

static DEFINE_INITIALIZED_MUTEX_SEMAPHORE(regfile_sem);


__attribute__((section(".traps"))) void init_regfile_allocator_(regfile_block_t *initial_allocated_block)
{
	out_of_bounds.lower_limit = MAX_REGBASE + 1;
	rtos_list_insert_before(&initial_allocated_block->list_item, &out_of_bounds.list_item);
}

__attribute__((section(".traps"))) uint18_t regalloc_(regfile_block_t *regs, uint18_t ncalls)
{
	rtos_list_item_t *iterator;
	int18_t lower_limit, upper_limit;
	uint18_t size, size_free;
	
	size = ncalls + 2;	// +1 for the entry, +1 for the limit
	
	semaphore_down(&regfile_sem);
	
	// start the search for the free space after next_block
	lower_limit = next_block->upper_limit;
	iterator = next_block->list_item.next;
	do
	{
		// The upper limit of the free space is the lower limit of the next listed block
		upper_limit = (int18_t) listed_regs(iterator)->lower_limit;
		
		// check the size of the free space
		size_free = upper_limit - lower_limit;
		if(size_free >= size)
		{
			// set the data of the new block and add it to the list
			regs->lower_limit = lower_limit;
			regs->upper_limit = lower_limit + size;
			rtos_list_insert_before(&regs->list_item, iterator);

			// start the next search for free space after the new block
			next_block = regs;
			
			semaphore_up(&regfile_sem);
			return TRUE;
		}
		
		// The lower limit of the space between this and the next allocated block
		// is the upper limit of this block.
		lower_limit = (int18_t) listed_regs(iterator)->upper_limit;

		iterator = rtos_list_next(iterator);
	}
	while (iterator != next_block->list_item.next);
	
	semaphore_up(&regfile_sem);
	
	return FALSE;
}

__attribute__((section(".traps"))) void regfree_(regfile_block_t *regs)
{
	semaphore_down(&regfile_sem);
	
	if (regs->lower_limit <= next_block->lower_limit)
	{
		// start the next search for free space where this block is deleted
		next_block = listed_regs(regs->list_item.previous);
	}
	rtos_list_remove(&regs->list_item);
	
	semaphore_up(&regfile_sem);
}
