/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#include <rtos_internal.h>


task_t current_task;

static rtos_list_t ready_queues[PRIORITY_LEVELS];

static task_t highest_priority_task;

static uint9_t stop_current = FALSE;


__attribute__((section(".traps"))) task_t get_current_task_()
{
	return current_task;
}

__attribute__((section(".traps"))) void initialize_scheduler_(task_t current)
{
	//uint18_t status;
	uint18_t i;
	
	for( i=0; i<PRIORITY_LEVELS; i++ )
		rtos_list_initialize(&ready_queues[i]);
	
	current_task = highest_priority_task = current;
	scheduler_add_task(current);
}

__attribute__((section(".traps"))) void scheduler_add_task_(task_t task)
{
	uint18_t status;
	
	STATUS_DISABLE_INT(status);
	
	rtos_list_add_last(&ready_queues[task->priority], &task->list_item);
	
	// change the highest priority task if necessary
	if( task->priority > highest_priority_task->priority )
		highest_priority_task = task;
	
	STATUS_RESTORE(status);
}

__attribute__((section(".traps"))) void scheduler_remove_task_(task_t task)
{
	uint18_t status;
	uint18_t highest_priority;
	rtos_list_item_t *queue_head;
	
	STATUS_DISABLE_INT(status);
	
	rtos_list_remove(&task->list_item);
	
	// In the case we are removing the current task,
	// a task switch is forced.
	if (task == current_task)
		stop_current = TRUE;

	// If the task was the highest priority task,
	// search for the new one.
	if (task == highest_priority_task)
	{
		highest_priority = highest_priority_task->priority;
		while (rtos_list_is_empty(&ready_queues[highest_priority]))
			highest_priority--;
		queue_head = rtos_list_first(&ready_queues[highest_priority]);
		highest_priority_task = rtos_list_item_owner(queue_head, struct tcb, list_item);
	}
	
	STATUS_RESTORE(status);
}

__attribute__((section(".traps"))) void schedule_tasks_()
{
	uint18_t status;
	task_t previous_task;
	
	STATUS_DISABLE_INT(status);

	// A task switch is only necessary when the current task is not the
	// highest priority task. In this case check if preemption is allowed or
	// the task switch is forced. Then do it.
	if ( (current_task != highest_priority_task) &&
		(stop_current || current_task->forbidden_count == 0) )
	{
		stop_current = FALSE;
		
		previous_task = current_task;
		current_task = highest_priority_task;
		switch_tasks(previous_task, highest_priority_task);
	}
	
	STATUS_RESTORE(status);
}

__attribute__((section(".traps"))) void forbid_preemption_()
{
	current_task->forbidden_count++;
}

__attribute__((section(".traps"))) void permit_preemption_()
{
	#if EXTENDED_CHECKS == 1
		if( current_task->forbidden_count == 0 )
		{
			// TODO: error
			return;
		}
	#endif
	
	current_task->forbidden_count--;
	if (current_task->forbidden_count == 0) schedule_tasks();
}

__attribute__((section(".traps"))) void task_yield_()
{
	rtos_list_item_t *queue_head;
	uint18_t status;
	
	STATUS_DISABLE_INT(status);
	
	// put the current task to the end of the ready queue
	rtos_list_remove(&current_task->list_item);
	rtos_list_add_last(&ready_queues[current_task->priority], &current_task->list_item);
	
	// select the next task
	queue_head = rtos_list_first(&ready_queues[highest_priority_task->priority]);
	highest_priority_task = rtos_list_item_owner(queue_head, struct tcb, list_item);
	
	schedule_tasks();
	
	STATUS_RESTORE(status);
}
