/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
#include <rtos_internal.h>
#include <interrupt.h>

extern uint18_t rtos_starting;

static DEFINE_INITIALIZED_EMPTY_LIST(to_delete_list);

__attribute__((section(".traps"), noinline)) void switch_tasks_(task_t from_task, task_t to_task)
{
	__asm__ volatile (
		// save the global registers
		"mov	r6,	r0	\n"
		"mov	r7,	r1	\n"
		"mov	r8,	r2	\n"
		"mov	r9,	r3	\n"
		// save the status register in the tcb, and restore the status of the next task
		"movs2i	r0,	sfr_status	\n"
		"l9	r1,	%1	\n"
		"movi2s	sfr_status,	r1	\n"
		"s9	%0,	r0	\n"		// still uses the old register window
		// restore the global registers of the next task and return
		"mov	r0,	r6	\n"
		"mov	r1,	r7	\n"
		"mov	r2,	r8	\n"
		"mov	r3,	r9	\n"
	: "=m"(from_task->status)
	: "m"(to_task->status)
	: "r6", "r7", "r8", "r9", "cc", "memory"
	// "cc" (condition code) and "memory" will have changed
	// when returning to from_task
	);
}




__attribute__((section(".traps"))) task_t create_task_( void (*entry)(void *param), void *param,
	uint18_t priority, uint18_t max_call_level, size_t stack_size )
{
	task_t new_task;
	void *stack, *stack_base;
	int18_t regbase;
	
	#if EXTENDED_CHECKS == 1
		if (priority >= PRIORITY_LEVELS)
		{
			// TODO: error
			return NULL;
		}
	#endif
	
	new_task = (struct tcb *) malloc(sizeof(struct tcb));
	if (new_task != NULL)
	{
		stack_size = ROUNDED_UP_EVEN(stack_size);
		stack = malloc(stack_size);
		if (stack != NULL)
		{
			new_task->stack = stack;
			new_task->priority = priority;
			new_task->forbidden_count = 0;
			new_task->ready = TRUE;
			
			if (regalloc(&new_task->regs, max_call_level))
			{
				regbase = new_task->regs.lower_limit + 1;
				stack_base = (char *)stack + stack_size;
				
				__asm__ volatile (
					// push entry, parameter, stack base and r3:
					"addi	r0,	-8	\n"
					"s18	6(r0),	r3	\n"
					"s18	4(r0),	%[stack_base]	\n"
					"s18	2(r0),	%[param]	\n"
					"s18	0(r0),	%[entry]	\n"
					// save the current status register in r3:
					"movs2i	r3,	sfr_status	\n"
					// set the status to initialize the registers:
					"movi	r8,	0x80	\n"	//
					"and	r8,	r3	\n"	// current MM-Bit
					"or	r8,	%[regbase]	\n"
					// disable interrrupts and go to the new tasks registers:
					"cbits	INT	\n"
					"movi2s	sfr_status,	r8	\n"
					"or	r0,	r0	\n"
					// give the task its parameter and set its return address:
					"l18	r4,	2(r0)	\n"
					"lhi	r11,	%%hi(task_return)	\n"
					// initialize the registers for the task switch:
					"jals	here	\n"
					"ori	r11,	%%lo(task_return)	\n"	// delay-slot
				"here:	\n"
					"l18	r11,	0(r0)	\n"
					"l18	r6,	4(r0)	\n"
					// restore the status register:
					"movi2s	sfr_status,	r3	\n"
					// restore r3:
					"l18	r3,	6(r0)	\n"
					// pop four times
					"addi	r0,	8	\n"
				:
				: [regbase] "r"(regbase),
				  [entry] "r"(entry),
				  [param] "r"(param),
				  [stack_base] "r"(stack_base)
				: "r8",
				  "memory"    // prevent reordering of the code
				);
				
				new_task->status = regbase + 1;
				
				scheduler_add_task(new_task);
				schedule_tasks();
			}
			else if (!rtos_starting)
			{
				free(stack);
				free(new_task);
				new_task = NULL;
			}
		}
		else
		{
			free(new_task);
			new_task = NULL;
		}
	}
	return new_task;
}

__attribute__((section(".traps"))) void delete_task_(task_t task)
{
	uint18_t status;
	
	STATUS_DISABLE_INT(status);
	
	if (task->ready)
	{
		task->ready = FALSE;
		scheduler_remove_task(task);
		
		if (task == get_current_task())
		{
			rtos_list_add_last(&to_delete_list, &task->list_item);
			STATUS_RESTORE(status);
			schedule_tasks();
		}
	}
	else rtos_list_remove(&task->list_item);
	
	STATUS_RESTORE(status);
	regfree(&task->regs);
	free(task->stack);
	free(task);
}

__attribute__((destructor)) __attribute__((section(".traps"))) void task_return()
{
	delete_task_(get_current_task_());
	while(1);
}


void idle_task_code(void *param)
{
	task_t to_delete;
	rtos_list_item_t *first;
	uint18_t status;
	
	__asm__("sbits INT\n");
	// enable interrupts without the use of interrupt.s
	
	while (TRUE)
	{
		__asm__ volatile ("": : : "memory");
		// forces the compiler to check the if-statement in every loop iteration

		if (!rtos_list_is_empty(&to_delete_list))
		{
			STATUS_DISABLE_INT(status);
			first = rtos_list_remove_first(&to_delete_list);
			to_delete = rtos_list_item_owner(first, struct tcb, list_item);
			STATUS_RESTORE(status);
			regfree(&to_delete->regs);
			free(to_delete->stack);
			free(to_delete);
		}
	}
}
