/**
 * Copyright (c) 2012, Michael Raitza
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the <ORGANIZATION> nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <libusb-1.0/libusb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>

#include "err.h"
#include "progname.h"
#define VERSION "0.2"

char attach = 0;
char detach = 0;
uint16_t vid = 0;
uint16_t pid = 0;
unsigned char devcnt = 0;

void usage() {
	fprintf(stdout, "Usage: %s [-hV] [-n NUM] -a|-d -v VENDOR_ID -p PRODUCT_ID\n\n"
		"  -a\t\tAttach kernel driver to device.\n"
		"  -d\t\tDetach kernel driver from device.\n"
		"  -h\t\tPrint this help and exit.\n"
		"  -n NUM\tDetach drivers from NUM device only. Detach from all matching devices if not given.\n"
		"  -p PRODUCT_ID\tMatch against PRODUCT_ID.\n"
		"  -v VENDOR_ID\tMathc against VENDOR_ID.\n"
		"  -V\t\tPrint version information and exit.\n\n", getprogname());
}

void parse_options(int argc, char** argv) {
	char opt;
	if (argc)
		setprogname(argv[0]);
	
	while ((opt = getopt(argc, argv, "hVadn:v:p:")) != -1) {
		switch (opt) {
		case 'a': attach = 1;
			break;
		case 'd': detach = 1;
			break;
		case 'v': vid = (uint16_t) strtol(optarg, NULL, 16);
			break;
		case 'p': pid = (uint16_t) strtol(optarg, NULL, 16);
			break;
		case 'V': fprintf(stdout, "%s: Version %s\n", getprogname(), VERSION);
			exit(EXIT_SUCCESS);
		case 'h': usage();
			exit(EXIT_SUCCESS);
		case 'n': devcnt = (unsigned char)strtol(optarg, NULL, 10);
			break;
		default:
			errx(EXIT_FAILURE, "%s %c", "Unknown option:", optopt);
		}
	}
		if (optind != argc)
			errx(EXIT_FAILURE, "%s", "Unknown non-option arguments found");
		
		if (attach && detach)
			errx(EXIT_FAILURE, "%s", "Only one action can be specified");
		if (!(attach || detach))
			errx(EXIT_FAILURE, "%s", "No action specified");
		if (!vid)
			errx(EXIT_FAILURE, "%s", "No vendor ID specified");
		if (!pid)
			errx(EXIT_FAILURE, "%s", "No product ID specified");
}

int main(int argc, char** argv) {
	libusb_context *ct;
	libusb_device **devlist;
	ssize_t len, i;
	libusb_device *dev;
	libusb_device_handle *devh;
	char handled = 0;
	int ret = EXIT_SUCCESS;

	parse_options(argc, argv);
	if (libusb_init(&ct))
		errx(EXIT_FAILURE, "%s", "Could not initialise Libusb");
	
	len = libusb_get_device_list(ct, &devlist);
	if (len < 0)
		errx(EXIT_FAILURE, "%s", "Could not acquire device list");
	for (i = 0; i < len; i++) {
		struct libusb_device_descriptor d;
		unsigned char cnt = 0;
		dev = devlist[i];
		libusb_get_device_descriptor(dev, &d);
		
		if (!((d.idVendor == vid) && (d.idProduct == pid)))
			continue;
		cnt++;
		if (!((devcnt == 0) || (cnt == devcnt)))
			continue;

		handled = 1;
		if (libusb_open(dev, &devh))
			errx(EXIT_FAILURE, "%s", "Could not open device");
		if (detach && (libusb_kernel_driver_active(devh, 0) == 1)) {
			if (libusb_detach_kernel_driver(devh, 0))
				errx(EXIT_FAILURE, "%s", "Could not detach kernel driver");
		} else if (attach && !libusb_kernel_driver_active(devh, 0))
			if (libusb_attach_kernel_driver(devh, 0))
				errx(EXIT_FAILURE, "%s", "Could not attach kernel driver");
		libusb_close(devh);
	}
	if (!handled) {
		warnx("%s %04x:%04x", "No device found with id ", vid, pid);
		ret = EXIT_FAILURE;
	}
	
	libusb_free_device_list(devlist, 1);

	libusb_exit(ct);
	return ret;
}
