package de.tu_darmstadt.rs;

import net.sf.xslthl.Block;
import net.sf.xslthl.Config;
import net.sf.xslthl.MainHighlighter;
import net.sf.xslthl.StyledBlock;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class IncludeCode {
    private static final String RS_NAMESPACE = "http://www.rs.tu-darmstadt.de/ns/spartanmc/manual/";
    private static final Document doc;

    private static File projectDir = null;

    private static MainHighlighter highlighter = null;

    public static void setProjectDir(File projectDir) {
        IncludeCode.projectDir = projectDir;
    }

    static {
        DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder;
        try {
            builder = dbf.newDocumentBuilder();
        } catch (ParserConfigurationException e) {
            throw new RuntimeException(e);
        }
        doc = builder.newDocument();

    }

    private static MainHighlighter getHighlighter() {

        if (IncludeCode.highlighter == null) {
            if (projectDir == null) {
                throw new NullPointerException("Need to set project dir first!");
            }
            Config c = Config.getInstance(new File(projectDir, "xslthl/xslthl-config.xml").toURI().toString());
            MainHighlighter hl = c.getMainHighlighter(new File(projectDir, "xslthl/c-hl.xml").toURI().toString());
            if (hl == null)
                throw new RuntimeException("Highlighter not found");
            IncludeCode.highlighter = hl;
        }
        return IncludeCode.highlighter;
    }

    private static Block copyStyle(Block reference, String text) {
        if (reference.isStyled()) {
            return new StyledBlock(text, ((StyledBlock) reference).getStyle());
        }
        return new Block(text);
    }

    /**
     * Split blocks by lines
     *
     * @param blocks
     * @return
     */
    private static List<List<Block>> blocksToLines(List<Block> blocks) {
        List<List<Block>> result = new ArrayList<>();
        List<Block> currentLine = new ArrayList<>();

        for (Block block : blocks) {
            final List<DomHelpers.SplitElement> split = DomHelpers.split("\n", block.getText());
            for (DomHelpers.SplitElement se : split) {

                if (!se.isSplitRegex) {
                    currentLine.add(copyStyle(block, se.text));

                } else {
                    result.add(currentLine);
                    currentLine = new ArrayList<>();
                }
            }
        }
        if (!currentLine.isEmpty())
            result.add(currentLine);
        return result;

    }

    private static List<Node> blockToNode(Block block) {
        List<Node> res = new ArrayList<>();
        String elmName;
        if (block.isStyled()) {
            elmName = ((StyledBlock) block).getStyle();
        } else {
            elmName = "regular";
        }
        DomHelpers.split("\t|    ", block.getText())
                .forEach(se -> {
                    if (se.isSplitRegex)
                        res.add(doc.createElementNS(RS_NAMESPACE, "tab"));
                    else {

                        Node elm = doc.createElementNS(RS_NAMESPACE, elmName);
                        elm.setTextContent(se.text);
                        res.add(elm);
                    }
                });
        return res;
    }

    private static Node blocksToNode(List<Block> blocks) throws ParserConfigurationException {

        Node node = doc.createElementNS(RS_NAMESPACE, "listingHighlighted");
        List<List<Block>> lines = blocksToLines(blocks);
        lines.stream()
                //Avoid xslht bug: first line is not correctly highlighted. we inserted dummy data before, so only that gets corrupted
                .skip(1)
                .map(line -> {
                    Node lineNode = doc.createElementNS(RS_NAMESPACE, "codeline");
                    line.forEach(b -> blockToNode(b).forEach(lineNode::appendChild));
                    return lineNode;
                }).forEach(node::appendChild);
        return node;
    }

    private static Node highlightText(String code, String language) throws IOException, ParserConfigurationException {
        MainHighlighter hl = getHighlighter();
        final List<Block> highlight = hl.highlight("//avoid xslthl bug: first line not correctly highlighted\n" + code);
        return blocksToNode(highlight);
    }

    private enum RemoveLicenseState {
        IN_LICENSE,
        REGULAR;
    }

    private static Stream<String> removeLicense(Stream<String> code, int commentCountP) {
        int[] commentCount = {commentCountP};
        RemoveLicenseState[] state = {RemoveLicenseState.REGULAR};
        return code.filter(line -> {
            System.out.print (line+": ");
            if (state[0]==RemoveLicenseState.REGULAR && commentCount[0]>0) {
                if (line.trim().isEmpty()) {
                    return false;
                } if (line.trim().startsWith("/*")) {
                    commentCount[0]--;
                    state[0] = RemoveLicenseState.IN_LICENSE;
                    return false;
                } else {
                    //regular text, not removing anything afterwards
                    commentCount[0] = 0;
                    state[0] = RemoveLicenseState.REGULAR;
                    return true;
                }
            } else if (state[0]==RemoveLicenseState.REGULAR) {
                return true;
            } else { //In License
                    if (line.trim().endsWith("*/")) {
                        state[0] = RemoveLicenseState.REGULAR;
                        return false;
                    } else if (line.contains("*/")) { //End of comment within the line
                        throw new RuntimeException("end of license within line not supported");
                    } else {
                        return false;
                    }
            }
        });
    }

    public static Node includeCode(String src, String language, String removeCommentCount) throws ParserConfigurationException, TransformerException, IOException {
        int removeCommentCountInt = tryParseInt(removeCommentCount,1);

        Stream<String> lines = Files.lines(projectDir.toPath().resolve("../../" + src).toAbsolutePath().normalize());
        lines = removeLicense(lines, removeCommentCountInt);
        return highlightText(lines.collect(Collectors.joining("\n")), language);
    }

    private static int tryParseInt(String s, int def) {
        int i;
        try {
            i = Integer.parseInt(s);
        } catch (NumberFormatException e) {
            i = def;
        }
        return i;
    }

    public static Node highlight(Node code, String language) throws IOException, ParserConfigurationException {

        final String lines = DomHelpers.streamNodeList(code.getChildNodes())
                .filter(n -> n.getNodeName().equals("codeline"))
                .map(IncludeCode::codeLineToStr)
                .collect(Collectors.joining("\n"));
        final Node result = highlightText(lines, language);
        return result;

    }

    private static String codeLineToStr(Node n) {
        return DomHelpers.streamNodeList(n.getChildNodes())
                .map(c -> {
                    if (c instanceof Text)
                        return c.getTextContent().replaceAll("\n", "").replaceAll("\\s+", " ");
                    else switch (c.getNodeName()) {
                        case "tab":
                            return "\t";
                        case "shl":
                            return "<<";
                        case "shr":
                            return ">>";
                        case "chl":
                            return "<";
                        case "chr":
                            return ">";
                        default:
                            return "unknown \'" + c.getNodeName() + "'";
                    }
                })
                .collect(Collectors.joining());
    }
}