/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.autoPerf.serialReader;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.nio.file.StandardOpenOption;
import java.util.ArrayList;
import java.util.List;

/**
 * Created by laurenzk on 24.04.17.
 */
public class AutoPerfResultsReader implements ResultReader {
	private static final String PROFILE_ARRAY = "PROFILE_ARRAY:";
	private static final String PERF_AUTO_START = "PERF_AUTO_START";
	private static final String PERF_AUTO_STOP = "PERF_AUTO_STOP";


	private SerialReader input;
	private Logger log = LoggerFactory.getLogger(AutoPerfResultsReader.class);
	private List<String> results;
	private List<String> arrayProfile;

	public AutoPerfResultsReader(SerialReader in) {
		input = in;
		results = new ArrayList<String>();
		arrayProfile = new ArrayList<String>();
	}

	/**
	 * Receives autoPerf results from an open serial port.
	 *
	 * This function blocks until PERF_AUTO_START is received. The following lines are
	 * interpreted as autoPerf results. Receiving is stopped after PERF_AUTO_STOP is
	 * stopped.
	 *
	 * @return true if the received data was parsed successfully. False if parsing failed.
	 */
	// TODO: Add timeout while receiving.
	public boolean readResults() {
		String line;

		while ((line = input.readLine()) != null) {
			if (line.startsWith(PROFILE_ARRAY)) {
				arrayProfile.add(line.replace(PROFILE_ARRAY, ""));
			} else if (PERF_AUTO_START.equals(line)) {
				break;
			} else {
				log.info(line);
			}
		}

		while ((line = input.readLine()) != null) {
			if (PERF_AUTO_STOP.equals(line)) {
				log.info("autoPerf results received successfully.");
				break;
			}

			String[] s = line.split(";");
			if (s.length < 2) {
				log.error("Unknown line received: " + line);
				return false;
			}

			results.add(line);
		}

		return true;
	}

	/**
	 * Saves the received results to a CSV file (with semicolons as separators).
	 *
	 * @return true if the file was written successfully, false otherwise.
	 */
	public boolean saveResults(String output) {
		File outputFile = Paths.get(output).toFile();

		try {
			FileUtils.writeLines(outputFile, "UTF-8", results, false);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}

		return true;
	}

	public boolean saveArrays(String output) {
		File arrayOutputFile = Paths.get(output).toFile();
		boolean createArrayProfileOutput = !arrayProfile.isEmpty();

		if(createArrayProfileOutput) {
			try {
				FileUtils.writeLines(arrayOutputFile, "UTF-8", arrayProfile, false);
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
		}

		return true;
	}
}
