/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.autoPerf.serialReader;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;

/**
 * Implements a {@link ResultReader}, that listens
 * for LoopOpt specific PerformanceCounters via UART
 * Created by dominikl on 12.05.17.
 */
public class LoopOptResultsReader implements ResultReader {
	private static final String PERF_AUTO_START_PRINT_ITER = "PERF_AUTO_START_PRINT_ITER";
	private static final String PERF_AUTO_STOP_PRINT_ITER = "PERF_AUTO_STOP_PRINT_ITER";
	private static final String PERF_AUTO_END_LOOP = "PERF_AUTO_END_LOOP";
	private static final String PERF_AUTO_STOP = "PERF_AUTO_STOP";

	private SerialReader input;
	private Logger log = LoggerFactory.getLogger(LoopOptResultsReader.class);
	private List<String> results;

	public LoopOptResultsReader(SerialReader in) {
		input = in;
		results = new ArrayList<String>();
	}

	/**
	 * Receives LoopOpt results from an open serial port
	 *
	 * @return true when all results have been recieved
	 */
	@Override
	public boolean readResults() {
		log.info("Listening for LoopOpt results via UART!");
		while (true) {
			while (true) {
				String line = input.readLine();

				if (PERF_AUTO_START_PRINT_ITER.equals(line)) {
					log.info("Loop Iteration started");
					break;
				}
				if (PERF_AUTO_STOP.equals(line)) {
					log.info("All LoopOpt Results received successfully!");
					return true;
				}
				if (PERF_AUTO_END_LOOP.equals(line)) {
					log.info("Full Loop received successfully!");
					results.add("Loop end;");
				}
			}

			input.putChar('\n');

			while (true) {
				String line = input.readLine();

				if (PERF_AUTO_STOP_PRINT_ITER.equals(line)) {
					log.info("One loop iteration received successfully!");
					results.add("Iteration end; ");
					break;
				} else if (line.length() > 0) {
					results.add(line);
				}
			}

		}
	}

	@Override
	public boolean saveResults(String fileName) {
		File outputFile = Paths.get(fileName).toFile();

		try {
			FileUtils.writeLines(outputFile, "UTF-8", results, false);
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}

		return true;
	}
}
