/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.autoPerf.serialReader;

import de.tu_darmstadt.rs.jcommander.JCommanderRunner;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This class contains the {@link Main#main(String[])} function that parses the passed arguments
 * and executes the autoPerf tool.
 */
public final class Main implements Runnable{


	public static final String PROGRAM_NAME = "autoPerf-SerialReader";

	private static Logger log = LoggerFactory.getLogger(Main.class);

	@Parameter(names = {"-t", "--target"}, description = "Set the target/ Or from which application the performance results are generated", converter = InputType.Converter.class)
	private InputType target = InputType.autoperf;

	@Parameter(names = {"-a", "--autoperf"}, description = "Same as '-t autoperf'. Overwrites -t")
	private boolean autoperf = false;

	@Parameter(names = {"-m", "--microstreams"}, description = "Same as '-t microstreams'.  Overwrites -t")
	private boolean microstreams = false;

	@Parameter(names = {"-p", "--port"}, description = "Set the serial port to listen to")
	private String port = "/dev/ttyUSB0";

	@Parameter(names = {"-o", "--output"}, description = "Specify the output File for the results.")
	private String output = "results.csv";

	@Parameter(names = {"-s", "--array-size-dump"}, description = "Specify the output File for the dumped array sizes.")
	private String arraySizeOutput = "arrays.csv";

	@Parameter(names = {"-l", "--loopOpt"}, description = "Same as '-t loopopt'.  Overwrites -t")
	private boolean loopOpt = false;

	public Main() {

	}

	/**
	 * The main function for autoPerf. Takes an array of arguments and executes autoPerf according
	 * to the arguments.
	 *
	 * @param args Array of strings containing the arguments.
	 */
	public static void main(String[] args) {
		JCommanderRunner.create(new Main())
				.setProgramName(PROGRAM_NAME)
				.run(args);
	}

	/**
	 * Starts the program
	 */
	public void run() {

		checkArguments();

		if (microstreams) {
			target = InputType.microstreams;
		} else if (autoperf) {
			target = InputType.autoperf;
		} else if (loopOpt) {
			target = InputType.loopopt;
		}

		SerialReader serialPort = new SerialReader();
		if (!serialPort.open(port)) {
			throw new RuntimeException("Can't open serial port '" + port + "'.");
		}

		switch (target) {
			case microstreams: {
				MicroStreamsEvalResultsReader reader = new MicroStreamsEvalResultsReader(serialPort);
				reader.readResults();
				reader.saveResults(output);
			}
			break;
			case loopopt: {
				LoopOptResultsReader reader = new LoopOptResultsReader(serialPort);
				reader.readResults();
				reader.saveResults(output);
			}
			break;
			case autoperf: {
				AutoPerfResultsReader reader = new AutoPerfResultsReader(serialPort);
				reader.readResults();
				reader.saveResults(output);
				reader.saveArrays(arraySizeOutput);
			}
			break;
			default:
				throw new ParameterException("No target selected");
		}

		serialPort.close();

		return;
	}

	private void checkArguments() {
		int autoperfInt = autoperf ? 1:0;
		int microstreamsInt = microstreams ? 1:0;
		int loopOptInt = loopOpt ? 1:0;
		int sum = autoperfInt + microstreamsInt + loopOptInt;
		if (sum > 1) {
			throw new ParameterException("-l and -a and -m can't be used together!");
		} else if (sum == 0) {
			throw new ParameterException("None of the following option is set: -l, -a, -m  ...please select one!");
		}
	}

}
