/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.autoPerf.transformationPasses;

import de.tu_darmstadt.rs.autoPerf.config.AutoPerfConfiguration;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.VariableUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;

import cetus.hir.CompoundStatement;
import cetus.hir.DFIterator;
import cetus.hir.DeclarationStatement;
import cetus.hir.ExpressionStatement;
import cetus.hir.FunctionCall;
import cetus.hir.IDExpression;
import cetus.hir.NameID;
import cetus.hir.Procedure;
import cetus.hir.Statement;
import cetus.hir.TranslationUnit;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.List;
import java.util.Optional;

/**
 * Adds code to analyze the size of arrays that might be send from core to core
 */
public class ArraySizeProfileInjector extends AbstractTransformationPass {

	private static final Logger log = LoggerFactory.getLogger(ArraySizeProfileInjector.class);

	public ArraySizeProfileInjector() {

	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				CetusProgram.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				CetusProgram.class
		);
	}

	@Override
	public List<Class<? extends TransformationPass>> requiredTransformationPasses() {
		return Arrays.asList(
				PerfCounterCallInjector.class
		);

	}

	@Override
	public boolean process(Model model) {
		if (model.getModelComponent(AutoPerfConfiguration.class).isArrayProfiling()) {
			CetusProgram cetusProgram = model.getModelComponent(CetusProgram.class);

			DFIterator<VariableDeclarator> vars = new DFIterator<>(cetusProgram.getProgram(), VariableDeclarator.class);

			while (vars.hasNext()) {
				VariableDeclarator variableDeclarator = vars.next();

				if (VariableUtils.isArray(variableDeclarator) && VariableUtils.getLength(variableDeclarator) < 0 && CetusUtils.findParentOfType(variableDeclarator, Procedure.class) != null) {
					CompoundStatement parentCompound = CetusUtils.findParentOfType(variableDeclarator, CompoundStatement.class);
					//wenn global definiert muss die Größe fix sein nur mit DEFINE möglich.
					final Statement printVarSize = printfVarSize(variableDeclarator);
					if (printVarSize == null) {
						return false;
					}
					if (parentCompound == null) {
						model.printIgnoreableError(log,"The definition " + variableDeclarator + "seems invalid in global scope.");
							/*final Procedure mainProcedure = CetusUtils.getMainProcedure(program);
							final CompoundStatement mainCompound = (CompoundStatement) mainProcedure.getChildren().stream().collect(StreamUtil.exactlyOne());
							final Statement firstChild = CetusUtils.findChildOfType(mainCompound, Statement.class);
							mainCompound.addStatementBefore(firstChild, printVarSize);*/
					} else { //wenn in funktion einfach nach deklaration
						parentCompound.addStatementAfter(CetusUtils.findParentOfType(variableDeclarator, DeclarationStatement.class), printVarSize);
					}
				}
			}

			return true;
		}
		return true;
	}

	private Statement printf(String msg, FunctionCall functionCall) {
		FunctionCall f = new FunctionCall(new NameID("printf"));
		f.addArgument(new NameID(msg));
		f.addArgument(functionCall);
		return new ExpressionStatement(f);
	}

	private FunctionCall sizeof(VariableDeclarator variableDeclarator) {
		FunctionCall f = new FunctionCall(new NameID("sizeof"));
		f.addArgument(variableDeclarator.getID().clone());
		return f;
	}

	private Statement printfVarSize(VariableDeclarator variableDeclarator) {
		final IDExpression varName = variableDeclarator.getID();
		String translationUnit = CetusUtils.findParentOfType(variableDeclarator, TranslationUnit.class).getOutputFilename();
		String function;
		int lineNumber;
		if (CetusUtils.findParentOfType(variableDeclarator, Procedure.class) == null) {
			log.error("global arrays shall not have a variable length array type");
			return null;
		} else {
			function = CetusUtils.findParentOfType(variableDeclarator, Procedure.class).getSymbolName();
		}
		if (CetusUtils.findParentOfType(variableDeclarator, Statement.class)!=null) {
			lineNumber = CetusUtils.findParentOfType(variableDeclarator, Statement.class).where();
		} else {
			lineNumber = -1; //should not happen, since arrays of variable length can't be global and therefore should always have parent statement (CompoundStatement(body) of Procedure at latest)
		}
		String msg = "\"\\nPROFILE_ARRAY:"+ translationUnit + ":" + function + ":" + variableDeclarator.toString() + ":" + lineNumber + ",%d \\n\"";
		return printf(msg, sizeof(variableDeclarator));
	}

}
