/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.autoPerf.transformationPasses;

import de.tu_darmstadt.rs.autoPerf.config.AutoPerfConfiguration;
import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.Uart;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;

import cetus.hir.AnnotationDeclaration;
import cetus.hir.AnnotationStatement;
import cetus.hir.CodeAnnotation;
import cetus.hir.CompoundStatement;
import cetus.hir.Declaration;
import cetus.hir.DoLoop;
import cetus.hir.ExpressionStatement;
import cetus.hir.ForLoop;
import cetus.hir.FunctionCall;
import cetus.hir.IfStatement;
import cetus.hir.IntegerLiteral;
import cetus.hir.Loop;
import cetus.hir.NameID;
import cetus.hir.Procedure;
import cetus.hir.Program;
import cetus.hir.ReturnStatement;
import cetus.hir.Statement;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import cetus.hir.WhileLoop;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;

/**
 * This class provides functions to transform an existing cetus program into a program
 * with autoPerf control statements.
 */
public class PerfCounterCallInjector extends AbstractTransformationPass {

	private static final String NO_PARENT = "no_parent";
	private static final String PERF_RESULTS = "perf_results";
	private static final String STRUCT_NAME = "struct perf_auto_result";
	private static final Logger log = LoggerFactory.getLogger(PerfCounterCallInjector.class);
	private static final String PERF_AUTO_INIT = "perf_auto_init";
	private static final String PERF_AUTO_START = "perf_auto_start";
	private static final String PERF_AUTO_STOP = "perf_auto_stop";
	private static final String PERF_AUTO_PRINT_ITER = "perf_auto_print_iter";
	private static final String PERF_AUTO_END_LOOP = "perf_auto_end_loop";
	private static final String PERF_AUTO_STOP_LOOP_OPT = "perf_auto_stop_loopOpt";
	private static final String PERF_AUTO_PRINT = "perf_auto_print";
	private Model model;

	/**
	 * Creates a new instance of CetusPerfPass.
	 */
	public PerfCounterCallInjector() {
	}

	@Override
	public boolean init() {
		return true;
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
			CetusProgram.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
			CetusProgram.class
		);
	}

	@Override
	public boolean process(Model m) {
		model = m;

		CetusProgram cetusProgram = model.getModelComponents(CetusProgram.class).stream().findAny().get();

		if (transform(cetusProgram.getProgram()) == null) {
			return false;
		}

		return true;
	}

	private ReturnStatement findReturn(Procedure p) {
		CompoundStatement compound = p.getBody();

		for (int i = compound.getChildren().size() - 1; i >= 0; i--) {
			Traversable ret = compound.getChildren().get(i);
			if (ret instanceof ReturnStatement) {
				return (ReturnStatement)ret;
			}
		}

		return null;
	}

	private List<Traversable> findPragmas(Traversable p) {
		List<Traversable> ret = new ArrayList<>();

		findPragmas(p, ret);

		return ret;
	}

	private void findPragmas(Traversable p, List<Traversable> pragmas) {
		if (p instanceof AnnotationStatement || p instanceof AnnotationDeclaration) {
			if (p.toString().contains(Constants.Pragma.AUTOPERF) && !model.getModelComponent(AutoPerfConfiguration.class).getLoopOptMode()) {
				pragmas.add(p);
			} else {
				//if loopOpt Mode is on, only pragma microStreams loopOpt should be found
				String[] lp = p.toString().split(" ");
				if (lp.length>2
						&& lp[0].equalsIgnoreCase(Constants.Pragma.SYMBOL)
						&& lp[1].equalsIgnoreCase(Constants.Pragma.IDENTIFIER)
						&& lp[2].equalsIgnoreCase(Constants.PragmaLoop.LOOPOPT)
						&& Arrays.stream(lp).anyMatch(s -> s.equalsIgnoreCase(Constants.PragmaLoop.AUTOPERF))) {
					pragmas.add(p);
				}
			}
		}

		if (p != null && p.getChildren() != null) {
			for (Traversable t : p.getChildren()) {
				findPragmas(t, pragmas);
			}
		}

		return;
	}

	private boolean containsClass(Traversable t, Class<?> a) {
		return findClass(t, a) != null;
	}

	private Traversable findClass(Traversable t, Class<?> a) {
		if (a.isInstance(t)) {
			return t;
		}

		for (Traversable tt : t.getChildren()) {
			Traversable c = findClass(tt, a);
			if (c != null) {
				return c;
			}
		}

		return null;
	}

	private boolean isEligiblePerfStatement(Statement e) {
		if (containsClass(e, ForLoop.class)
				|| containsClass(e, WhileLoop.class)
				|| containsClass(e, DoLoop.class)
				|| containsClass(e, FunctionCall.class)) {
			return true;
		} else if (e instanceof IfStatement) {
			return true;
		}

		return false;
	}

	private boolean isEligiblePerfBlockStatement(Statement e) {
		if (e instanceof ExpressionStatement) {
			return true;
		}

		return false;
	}

	private String getContainingFile(Statement s) {
		Traversable t = s;

		while (t != null) {
			if (t instanceof TranslationUnit) {
				return ((TranslationUnit) t).getOutputFilename();
			}
			t = t.getParent();
		}

		return NO_PARENT;
	}

	private String getContainingFunction(Statement s) {
		Traversable t = s;

		while (t != null) {
			if (t instanceof Procedure) {
				return ((Procedure) t).getSymbolName();
			}
			t = t.getParent();
		}

		return NO_PARENT;
	}

	private Declaration getResultsArray(List<List<Statement>> statements) {
		String initValue = "";
		initValue = initValue.concat("{\n");
		for (List<Statement> s : statements) {
			String identifier;
			if (s.size() == 1) {
				Statement ss = s.get(0);
				if (ss instanceof ExpressionStatement) {
					identifier = getContainingFile(ss) + ":"
							+ getContainingFunction(ss) + ":"
							+ ((ExpressionStatement) ss).getExpression().toString().replace("\\", "\\\\").replace("\"", "\\\"") + ":"
							+ String.valueOf(ss.where());
				} else if (ss instanceof WhileLoop) {
					identifier = getContainingFile(ss) + ":"
							+ getContainingFunction(ss) + ":"
							+ "while:"
							+ String.valueOf(ss.where());
				} else if (ss instanceof ForLoop || (ss instanceof CompoundStatement && containsClass(ss, ForLoop.class))) {
					identifier = getContainingFile(ss) + ":"
							+ getContainingFunction(ss) + ":"
							+ "for:"
							+ String.valueOf(ss.where());
				} else if (ss instanceof DoLoop) {
					identifier = getContainingFile(ss) + ":"
							+ getContainingFunction(ss) + ":"
							+ "do:"
							+ String.valueOf(ss.where());
				} else if (ss instanceof IfStatement) {
					identifier = getContainingFile(ss) + ":"
							+ getContainingFunction(ss) + ":"
							+ "if:"
							+ String.valueOf(ss.where());
				} else {
					identifier = getContainingFile(ss) + ":"
							+ getContainingFunction(ss) + ":"
							+ ss.getProcedure().getSymbolName().replace("\\", "\\\\").replace("\"", "\\\"") + ":"
							+ String.valueOf(ss.where());
				}

				initValue = initValue.concat("{.name = \"" + identifier + "\"},\n");
			} else {
				identifier = getContainingFile(s.get(0)) + ":";
				identifier += getContainingFunction(s.get(0)) + ":";
				identifier += "block:";
				identifier += String.valueOf(s.get(0).where());
				identifier += ":";
				identifier += String.valueOf(s.get(s.size() - 1).where());

				initValue = initValue.concat("{.name = \"" + identifier + "\"},\n");
			}
		}
		initValue = initValue.concat("}");
		VariableDeclarator dec = new VariableDeclarator(new NameID(STRUCT_NAME + " " + PERF_RESULTS + "[] = " + initValue));

		return new VariableDeclaration(dec);
	}

	private Statement getInitPrintf() {
		FunctionCall f = new FunctionCall(new NameID("stdio_uart_light_open"));
		f.addArgument(new NameID("&" + getUartHwAddr()));

		return new ExpressionStatement(f);
	}

	private Statement getInitCounters() {
		FunctionCall f = new FunctionCall(new NameID(PERF_AUTO_INIT));

		return new ExpressionStatement(f);
	}

	private Statement getStartCounters() {
		FunctionCall f = new FunctionCall(new NameID(PERF_AUTO_START));

		return new ExpressionStatement(f);
	}

	private Statement getStopCounters(int id) {
		FunctionCall f = new FunctionCall(new NameID(PERF_AUTO_STOP));
		f.addArgument(new IntegerLiteral(id));
		f.addArgument(new NameID(PERF_RESULTS));

		return new ExpressionStatement(f);
	}

	private Statement getLoopStopCounters(int statementID, int loopID) {
		FunctionCall f = new FunctionCall(new NameID(PERF_AUTO_STOP));
		f.addArgument(new IntegerLiteral(statementID));
		f.addArgument(new NameID("loop" + loopID));

		return new ExpressionStatement(f);
	}

	private Statement getIterationPrint(int itertionID, int loopID) {
		FunctionCall f = new FunctionCall(new NameID(PERF_AUTO_PRINT_ITER));
		f.addArgument(new IntegerLiteral(itertionID));
		f.addArgument((new NameID("loop" + loopID)));

		return new ExpressionStatement(f);
	}

	private Statement getLoopEnd() {
		FunctionCall f = new FunctionCall(new NameID(PERF_AUTO_END_LOOP));

		return new ExpressionStatement(f);
	}

	private Statement getStopLoopOpt() {
		FunctionCall f = new FunctionCall(new NameID(PERF_AUTO_STOP_LOOP_OPT));

		return new ExpressionStatement(f);
	}

	private Declaration getLoopResults(int size, int id, String file) {
		String initValue = "";
		initValue = initValue.concat("{\n");
		for (int i = 0; i < size; i++) {
			String identifier = "loop:" + file + ":" + id + ":" + (i+1);
			initValue = initValue.concat("{.name = \"" + identifier + "\"},\n");
		}
		initValue = initValue.concat("}");
		VariableDeclarator dec = new VariableDeclarator(new NameID(STRUCT_NAME + " loop" + id + "[] = " + initValue));

		return new VariableDeclaration(dec);

	}

	private Statement getPrintCounters(int size) {
		FunctionCall f = new FunctionCall(new NameID(PERF_AUTO_PRINT));
		f.addArgument(new IntegerLiteral(size));
		f.addArgument(new NameID(PERF_RESULTS));

		return new ExpressionStatement(f);
	}


	private List<List<Statement>> findStatements(CompoundStatement compound) {
		int blockcount = 0;
		List<List<Statement>> statements = new ArrayList<>();
		ArrayList<Statement> block = null;

		if (compound.getChildren() != null) {
			for (Traversable t : compound.getChildren()) {
				if (t instanceof Statement && isEligiblePerfStatement((Statement) t)) {
					if (blockcount != 0) {
						// End of block found.
						statements.add(block);
						blockcount = 0;
					}
					ArrayList<Statement> a = new ArrayList<>();
					a.add((Statement) t);
					statements.add(a);
				} else if (t instanceof Statement && isEligiblePerfBlockStatement((Statement) t)) {
					if (blockcount == 0) {
						block = new ArrayList<>();
					}
					block.add((Statement) t);
					blockcount++;
				} else {
					if (blockcount != 0) {
						// End of block found.
						statements.add(block);
						blockcount = 0;
					}
				}
			}
			if (blockcount != 0) {
				statements.add(block);
			}
		} else {
			log.warn("No statements found in selected compound statement.");
		}

		return statements;
	}

	/**
	 * Transforms the program and inserts autoPerf control statements.
	 *
	 * @return A new Cetus program with autoPerf control statements.
	 */
	private Program transform(Program program) {
		Procedure mainProcedure = CetusUtils.getMainProcedure(program);
		TranslationUnit mainFile = CetusUtils.findParentOfType(mainProcedure, TranslationUnit.class);

		if (program == null) {
			return null;
		}

		// Find all #pragma autoperfs or #pragma microStreams loopOpt ... autoperf in all source files (depending on mode)
		List<Traversable> pragmas = findPragmas(program);
		List<CompoundStatement> compounds = new ArrayList<>();
		if (pragmas.size() != 0) {
			// Find compound statement after each #pragma
			for (Traversable pragma : pragmas) {
				Traversable parent = pragma.getParent();
				boolean afterPragma = false;
				for (Traversable t : parent.getChildren()) {
					if (t == pragma) {
						afterPragma = true;
						continue;
					}
					if (afterPragma) {
						CompoundStatement compound = (CompoundStatement) findClass(t, CompoundStatement.class);
						if (compound == null) {
							model.printIgnoreableError(log,"Can't find compound statement after " + Constants.Pragma.AUTOPERF);
							return null;
						}
						compounds.add(compound);
						break;
					}
				}
				pragma.getParent().removeChild(pragma);

			}
		} else {
			if (model.getModelComponent(AutoPerfConfiguration.class).getLoopOptMode()) {
				model.printIgnoreableError(log,"No pragma microStreams loopOpt autoperf was found. You need at least one Pragma microStreams loopOpt autoperf when running in loopOpt-Mode");
				return null;
			}
			// No #pragma autoperf found, profile main() function.
			Procedure procedure = CetusUtils.getMainProcedure(program);
			if (procedure == null) {
				model.printIgnoreableError(log, "No main() procedure or \"" + Constants.Pragma.AUTOPERF +"\" found.");
				return null;
			}

			compounds.add((CompoundStatement) procedure.getChildren().get(0));
		}

		// Holds all profiled statements and statement blocks.
		List<List<Statement>> statements = new ArrayList<>();

		// Holds all TranslationUnits with profiled compounds.
		Set<TranslationUnit> translationUnits = new HashSet<>();

		//used to make the results array big enough to hold all loop results
		int loopIDCounter = 0;

		// Find statemants and statement blocks in profiled compounds and add autoperf statements around them.
		for (CompoundStatement compound : compounds) {
			TranslationUnit translationUnit = CetusUtils.findParentOfType(compound, TranslationUnit.class);
			if (translationUnit == null) {
				throw new RuntimeException("Can't find TranslationUnit (shouldn't happen)!");
			}

			translationUnits.add(translationUnit);

			if (!model.getModelComponent(AutoPerfConfiguration.class).getLoopOptMode()) {
				List<List<Statement>> newStatements = findStatements(compound);
				for (List<Statement> s : newStatements) {
					statements.add(s);
				}

				if (statements.size() == 0) {
					log.warn("No statements eligible for profiling found, not doing anything...");
					return program;
				}
				for (List<Statement> t : newStatements) {

					compound.addStatementBefore(t.get(0), getStartCounters());
					compound.addStatementAfter(t.get(t.size() - 1), getStopCounters(statements.indexOf(t)));
				}

			} else {
				Loop loop = null;

				//compound is pointing on the body (happens when there is no extra CompoundStatement between Pragma Statement and Loop)
				if (compound.getParent() instanceof Loop) {
					loop = (Loop) compound.getParent();
				} else { //compound is pointing on the CompoundStatement outside the Loop
					ListIterator iter = compound.getChildren().listIterator();
					while (iter.hasNext()) {
						Statement s = (Statement) iter.next();
						if (s instanceof Loop) {
							loop = (Loop) s;
						}
					}
				}
				if (loop==null) {
					model.printIgnoreableError(log, "No Loop found behind the pragma loopOpt autoperf.");
					return program;
				}
				int statementCounter = 0;

				CompoundStatement body = (CompoundStatement) loop.getBody();
				List<List<Statement>> states = findStatements(body);
				Statement lastStatement = null;
				String file = getContainingFile(body);
				for (List<Statement> lt : states) {
					//add Start and Stop Counters on every Statement in the Loop Body
					for (Statement st : lt) {
						body.addStatementBefore(st, getStartCounters());
						body.addStatementAfter(st, getLoopStopCounters(statementCounter, loopIDCounter));
						statementCounter++;
					}
					lastStatement = lt.get(lt.size()-1);
				}
				//Add function call to send auto_perf results via UART
				if (lastStatement != null) {
					body.addStatement(getIterationPrint(statementCounter, loopIDCounter));
				}
				mainFile.addDeclarationBefore(mainProcedure, getLoopResults(statementCounter, loopIDCounter, file));
				//add function calls to signal end of loop and add declaration for results of this loop
				CompoundStatement parent = (CompoundStatement) loop.getParent();
				parent.addStatementAfter((Statement) loop, getLoopEnd());
				loopIDCounter++;
			}
		}


		// Include perf.h in all modified TranslationUnits.
		for (TranslationUnit translationUnit : translationUnits) {
			translationUnit.addDeclarationFirst(new AnnotationDeclaration(new CodeAnnotation(Constants.SpartanMC.CLibraries.PERF.buildSystemInclude())));
		}

		// Add array for results before main function.
		if (!model.getModelComponent(AutoPerfConfiguration.class).getLoopOptMode()) {
			mainFile.addDeclarationBefore(mainProcedure, getResultsArray(statements));
		}

		// Add initialization code at the start of main.
		mainProcedure.getBody().getChildren().add(0, getInitCounters());

		// Add perf_auto_print() at the end of main() or, in case of loopOptMode, perf_auto_stop_loopOpt()
		ReturnStatement ret = findReturn(mainProcedure);
		if (ret != null) {
			if (!model.getModelComponent(AutoPerfConfiguration.class).getLoopOptMode()) {
				mainProcedure.getBody().addStatementBefore(ret, getPrintCounters(statements.size()));
			} else {
				mainProcedure.getBody().addStatementBefore(ret, getStopLoopOpt());
			}
		} else {
			if (!model.getModelComponent(AutoPerfConfiguration.class).getLoopOptMode()) {
				mainProcedure.getBody().addStatement(getPrintCounters(statements.size()));
			} else {
				mainProcedure.getBody().addStatement(getStopLoopOpt());
			}
		}

		return program;
	}

	private Statement getBody(Statement s) {
		if (s instanceof Loop) {
			log.info("Body: " + ((Loop) s).getBody().toString());
			return ((Loop) s).getBody();
		}
		return null;
	}

	private String getUartHwAddr() {
		final String hwAddress = model.getModelComponents(Core.class).stream().findAny().get().getPeripherals(Uart.class).stream().findAny().get().getHwAddress();
		return hwAddress;
	}
}
