/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.autoPerf.transformationPasses;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelSubComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.Uart;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.AnnotationDeclaration;
import cetus.hir.CodeAnnotation;
import cetus.hir.Declarator;
import cetus.hir.ExpressionStatement;
import cetus.hir.FunctionCall;
import cetus.hir.Initializer;
import cetus.hir.NameID;
import cetus.hir.Procedure;
import cetus.hir.Statement;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.UnaryExpression;
import cetus.hir.UnaryOperator;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;

/**
 * Intits UART at first line in main
 */
public class UartOpener extends AbstractTransformationPass {

	private static final Logger log = LoggerFactory.getLogger(UartOpener.class);

	private Model model;

	public UartOpener() {

	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				CetusProgram.class,
				Core.class,
				HardwareModelSubComponent.class);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				CetusProgram.class);
	}

	@Override
	public boolean process(Model model) {
		this.model = model;

		CetusProgram cetusProgram = model.getModelComponent(CetusProgram.class);

		final Optional<VariableDeclaration> ioDeclaration = CetusUtils.getStreamOfAllChildren(cetusProgram.getProgram(), VariableDeclaration.class)
				.filter(vd -> vd.getSpecifiers().stream().allMatch(s -> s.toString().equalsIgnoreCase("FILE")))
				.filter(vd -> vd.getDeclaredIDs().stream().anyMatch(id -> id.toString().equalsIgnoreCase("stdout")))
				.collect(StreamUtil.atMostOne());
		if (ioDeclaration.isPresent()) {
			final Declarator stdoutDeclarator = ioDeclaration.get().getDeclarators().stream().filter(x -> x.getID().getName().equals("stdout")).collect(StreamUtil.exactlyOne());
			final Optional<Traversable> initializer = Stream.of(stdoutDeclarator.getInitializer()).flatMap(CetusUtils::getStreamOfChildren).collect(StreamUtil.atMostOne());

			if (!initializer.toString().contains(getInitializerNameID().toString())) {
				if (initializer.isPresent()) {
					log.warn("IO File was set to " + initializer + " but " + getInitializerNameID() + " is expected. Set to the expected now");
				}
				stdoutDeclarator.setInitializer(getInitializer());
			}


		}
		else {
			final TranslationUnit parentTranslationUnit = CetusUtils.findParentOfType(CetusUtils.getMainProcedure(cetusProgram.getProgram()), TranslationUnit.class);
			final Procedure firstProcedure = CetusUtils.findChildOfType(parentTranslationUnit, Procedure.class);
			parentTranslationUnit.addDeclarationBefore(firstProcedure, getInitPrintfDec());
			final Procedure mainProcedure = CetusUtils.getMainProcedure(cetusProgram.getProgram());

			CetusUtils.findParentOfType(mainProcedure, TranslationUnit.class).addDeclarationFirst(new AnnotationDeclaration(new CodeAnnotation(Constants.C.buildSystemInclude(Constants.SpartanMC.CLibraries.STDIO.getHeader()))));
		}
		return true;
	}

	private VariableDeclaration getInitPrintfDec() {
		final Initializer initializer = getInitializer();
		final NameID stdout = new NameID("FILE * stdout");
		final VariableDeclarator variableDeclarator = new VariableDeclarator(/*PointerSpecifier.UNQUALIFIED,*/ stdout);
		variableDeclarator.setInitializer(initializer);
		final VariableDeclaration variableDeclaration = new VariableDeclaration(variableDeclarator);
		return variableDeclaration;
	}

	private Initializer getInitializer() {
		final NameID io = getInitializerNameID();

		final UnaryExpression rightHandside = new UnaryExpression(UnaryOperator.ADDRESS_OF, io);
		return new Initializer(rightHandside);
	}

	private NameID getInitializerNameID() {
		final NameID io;
		if (model.getConfig().isCreateSimulatedTarget()) {
			io = new NameID("SIM_PRINT_0_FILE");
		}  else {
			io = new NameID("UART_LIGHT_0_FILE");
		}
		return io;
	}

	@Deprecated
	private Statement getInitPrintf() {
		FunctionCall f = getUartOpenFunctionCall();

		return new ExpressionStatement(f);
	}

	@Deprecated
	private FunctionCall getUartOpenFunctionCall() {
		FunctionCall f = new FunctionCall(new NameID("stdio_uart_light_open"));
		f.addArgument(new NameID("&" + getUartHwAddr()));
		return f;
	}

	private String getUartHwAddr() {
		final String uartType = (new Uart()).getType();
		final String hwAddress = model.getModelComponents(Core.class).stream().findAny().get().getPeripherals().stream().filter(p -> p.getType().equalsIgnoreCase(uartType)).findAny().get().getHwAddress();
		return hwAddress;
	}
}
