/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package cetus.hir;

import java.math.BigInteger;
import java.util.Optional;

public class EvaluateExpression extends ExpressionVisitorAdapter<Optional<BigInteger>> {

	@Override
	protected Optional<BigInteger> visitDefault(Expression e) {
		return Optional.empty();
	}

	private BigInteger boolToBigInt(boolean b) {
		return  b? BigInteger.ONE : BigInteger.ZERO;
	}

	private BigInteger evaluateBinary(BigInteger lhs, BigInteger rhs, RegularBinaryOperator op) {
		switch (op) {
			case ADD: return lhs.add(rhs);
			case COMPARE_EQ: return boolToBigInt(lhs.equals(rhs));
			case COMPARE_GE: return boolToBigInt(lhs.compareTo(rhs) >= 0);
			case COMPARE_GT: return boolToBigInt(lhs.compareTo(rhs) > 0);
			case COMPARE_LE: return boolToBigInt(lhs.compareTo(rhs) <= 0);
			case COMPARE_LT: return boolToBigInt(lhs.compareTo(rhs) < 0);
			case COMPARE_NE: return boolToBigInt(!lhs.equals(rhs));
			case DIVIDE: return lhs.divide(rhs);
			case MODULUS: return lhs.mod(rhs);
			case MULTIPLY: return lhs.multiply(rhs);
			case SHIFT_LEFT: return lhs.shiftLeft(rhs.intValueExact());
			case SHIFT_RIGHT: return lhs.shiftRight(rhs.intValueExact());
			case SUBTRACT: return lhs.subtract(rhs);

			case BITWISE_AND:
			case BITWISE_EXCLUSIVE_OR:
			case BITWISE_INCLUSIVE_OR:
			case LOGICAL_AND:
			case LOGICAL_OR:
			case INSTANCEOF:
			default:
				return null;
		}
	}

	private BigInteger evaluateUnary(BigInteger i, UnaryOperator op) {
		switch (op) {
			case MINUS: return i.negate();
			case PLUS: return i;

			case ADDRESS_OF:
			case BITWISE_COMPLEMENT:
			case DEREFERENCE:
			case LOGICAL_NEGATION:
			case POST_DECREMENT:
			case POST_INCREMENT:
			case PRE_DECREMENT:
			case PRE_INCREMENT:
			default:
				return null;
		}
	}

	@Override
	public Optional<BigInteger> visitBinaryExpression(BinaryExpression e) {
		Optional<BigInteger> lhsOpt = e.getLHS().visit(this);
		Optional<BigInteger> rhsOpt = e.getRHS().visit(this);

		if (lhsOpt.isPresent() && rhsOpt.isPresent()) {

			BigInteger lhs = lhsOpt.get();
			BigInteger rhs = rhsOpt.get();

			if (!(e.op instanceof RegularBinaryOperator))
				return Optional.empty();

			return Optional.ofNullable(evaluateBinary(lhs, rhs, (RegularBinaryOperator) e.op));

		}
		return Optional.empty();
	}



	@Override
	public Optional<BigInteger> visitUnaryExpression(UnaryExpression e) {
		Optional<BigInteger> childOpt = e.getExpression().visit(this);

		if (childOpt.isPresent()) {

			BigInteger child = childOpt.get();

			return Optional.ofNullable(evaluateUnary(child, e.op));

		}
		return Optional.empty();
	}

	@Override
	public Optional<BigInteger> visitCharLiteral(CharLiteral e) {
		return Optional.of(BigInteger.valueOf(e.getValue()));
	}

	@Override
	public Optional<BigInteger> visitBooleanLiteral(BooleanLiteral e) {
		return Optional.of(boolToBigInt(e.getValue()));
	}


	@Override
	public Optional<BigInteger> visitIntegerLiteral(IntegerLiteral e) {
		return Optional.of(BigInteger.valueOf(e.getValue()));
	}

}
