package cetus.hir;

import java.io.PrintWriter;
import java.util.HashMap;

/**
* Infix operators that act on two expressions.
*/
public enum RegularBinaryOperator implements BinaryOperator {

	/**
	 * +
	 */
	ADD(0),

	/**
	 * &amp;
	 */
	BITWISE_AND(1),

	/**
	 * ^
	 */
	BITWISE_EXCLUSIVE_OR(2),

	/**
	 * |
	 */
	BITWISE_INCLUSIVE_OR(3),

	/**
	 * ==
	 */
	COMPARE_EQ(4),

	/**
	 * &gt;=
	 */
	COMPARE_GE(5),

	/**
	 * &gt;
	 */
	COMPARE_GT(6),

	/**
	 * &lt;=
	 */
	COMPARE_LE(7),

	/**
	 * &lt;
	 */
	COMPARE_LT(8),

	/**
	 * &#33;=
	 */
	COMPARE_NE(9),

	/**
	 * /
	 */
	DIVIDE(10),

	/**
	 * &amp;&amp;
	 */
	LOGICAL_AND(11),

	/**
	 * ||
	 */
	LOGICAL_OR(12),

	/**
	 * %
	 */
	MODULUS(13),

	/**
	 * *
	 */
	MULTIPLY(14),

	/**
	 * &lt;&lt;
	 */
	SHIFT_LEFT(15),

	/**
	 * &gt;&gt;
	 */
	SHIFT_RIGHT(16),

	/**
	 * -
	 */
	SUBTRACT(17),

	/**
	 * instanceof
	 */
	INSTANCEOF(18);




	private static String[] names = {
			"+", "&", "^", "|", "==", ">=", ">", "<=", "<",
			"!=", "/", "&&", "||", "%", "*", "<<", ">>", "-", " instanceof "};

	private static HashMap<String, RegularBinaryOperator> op_map =
			new HashMap<>(32);

	static {
		for (RegularBinaryOperator op : values()) {
			op_map.put(names[op.value], op);
		}
	}

    protected int value;

    /**
    * Used internally -- you may not create arbitrary binary operators
    * and may only use the ones provided as static members.
    *
    * @param value The numeric code of the operator.
    */
    private RegularBinaryOperator(int value) {
        this.value = value;
    }

    /**
    * Returns a binary operator that matches the specified string <tt>s</tt>.
    * @param s the string to be matched.
    * @return the matching operator or null if not found.
    */
    public static RegularBinaryOperator fromString(String s) {
        return op_map.get(s);
    }

    /* It is not necessary to override equals or provide cloning, because
       all possible operators are provided as static objects. */

    public void print(PrintWriter o) {
        o.print(names[value]);
    }

    @Override
    public String toString() {
        return names[value];
    }

    /**
    * Verifies this operator is valid.
    *
    * @throws IllegalStateException if the operator is invalid.
    */
    public void verify() throws IllegalStateException {
        if (value < 0 || value > 18) {
            throw new IllegalStateException();
        }
    }

    /**
    * Checks if this operator belongs to binary comparison operator.
    */
    public boolean isCompare() {
        return (value >= 4 && value <= 9);
    }

    /**
    * Checks if this operator belongs to boolean logic operator.
    */
    public boolean isLogical() {
        return (value >= 11 && value <= 12);
    }

}
