/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.synthesis;

import de.tu_darmstadt.rs.fpga.resources.HierarchicalUsage;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Pin;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Optional;

public class RapidSmithHardwareUsage implements IHardwareUsage {

	private static final Logger logger = LoggerFactory.getLogger(RapidSmithHardwareUsage.class);

	private final LibJConfigProject project;
	private final Design design;

	public RapidSmithHardwareUsage(LibJConfigProject project, Design design) {

		this.project = project;
		this.design = design;
	}

	@Override
	public String getInfo() {
		return "The report shows 5-input LUTs. 6-input LUTs are counted as two 5-input LUTs.";
	}

	@Override
	public HierarchicalUsage getUsage() throws IOException {
		RapidSmithHierarchicalUsage hierarchicalUsage = new RapidSmithHierarchicalUsage("Design");

		design.getInstances().forEach(hierarchicalUsage::addInstance);

		return hierarchicalUsage;

	}

	private Pin getTimingDomainSourcePin(LibJConfigProject project, Design design, String timingDomain) {
		//FPGA Pin?
		if (timingDomain.endsWith("_TNM")) {
			String pinName = timingDomain.substring(3, timingDomain.length() - 4); //String TS_ at beginning and _TNM at end
			logger.info("pin name {}", pinName);
			String pinLocation = DocumentUtil.getPinHandler(project.getDocumentHandler().getDocument(), pinName).getPinDescription().getFpgaPinName();
			logger.info("pin location {}", pinLocation);
			final Instance instance = design.getInstanceAtPrimitiveSite(design.getDevice().getPrimitiveSite(pinLocation));
			logger.info("instance {}", instance.getName());
			return instance.getPins().stream().collect(StreamUtil.exactlyOne());
		} else { //Derived clock
			String withoutPrefix = timingDomain.substring(3);
			logger.info(withoutPrefix);
			//Strip _dcm at end
			final String dcm1 = "_dcm";
			if (withoutPrefix.endsWith(dcm1))
				withoutPrefix = withoutPrefix.substring(0,withoutPrefix.length()-dcm1.length());
			int lastUnderscore = withoutPrefix.lastIndexOf('_');
			String instName = withoutPrefix.substring(0, lastUnderscore);
			String netName = withoutPrefix.substring(lastUnderscore + 1);
			logger.info("{} {}", instName, netName);
			final Instance dcm = design.getInstances().stream()
					.filter(this::isDCM)
					.filter(i -> i.getName().startsWith(instName))
					.collect(StreamUtil.exactlyOne());
			final Optional<Pin> pin = dcm.getPins().stream().filter(p -> p.getName().equalsIgnoreCase(netName)).collect(StreamUtil.atMostOne());
			return pin.orElseGet(() -> design.getNets().stream().filter(n -> n.getName().endsWith(netName)).collect(StreamUtil.exactlyOne()).getSource());

		}
	}

	private boolean isClockBuffer(Instance instance) {
		return instance.getType().toString().startsWith("BUFG");
	}

	private boolean isDCM(Instance instance) {
		final String name = instance.getType().toString();
		return name.startsWith("DCM") || name.startsWith("MMCM");
	}

	private int getUsageFromPin(Pin pin) {
		if (pin == null) {
			return 0;
		}
		return pin.getNet().getPins().stream()
				.filter(p -> p != pin)
				.mapToInt(p -> {
					if (isDCM(p.getInstance())) {
						return 1;
					}
					if (isClockBuffer(p.getInstance())) {
						return p.getInstance().getPins().stream()
								.filter(o -> o != p)
								.mapToInt(this::getUsageFromPin)
								.sum();
					}
					return 1;
				})
				.sum();
	}


	@Override
	public int getUsageByTimingDomain(String timingDomain) {
		logger.info("analyzing {}", timingDomain);
		final Pin pin = getTimingDomainSourcePin(project, design, timingDomain);
		logger.info("pin {}", pin);


		return getUsageFromPin(pin);
	}
}
