/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.synthesis;

import de.tu_darmstadt.rs.fpga.resources.HierarchicalUsage;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.BlockRamPrimitives;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.DspPrimitives;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.NonChangingPrimitives;

import edu.byu.ece.rapidSmith.design.Attribute;
import edu.byu.ece.rapidSmith.design.Instance;

import java.util.Optional;
import java.util.stream.Collectors;

public class RapidSmithHierarchicalUsage extends HierarchicalUsage {


	public RapidSmithHierarchicalUsage(String name) {
		super(name);
	}


	public void addInstance(Instance instance) {
		if (instance.isBRAM()) {
			handleBram(instance);
		} else if (instance.isDSP()) {
			addInstance(instance.getName(), DspPrimitives.DSP48A1);
		} else if (instance.isSLICE()) {
			handleSlice(instance);
		}
	}

	private void handleBram(Instance instance) {
		BlockRamPrimitives result;
		switch (instance.getType()) {
			case RAMB8BWER:
				result = BlockRamPrimitives.RAMB8BWER;
				break;
			case RAMB16BWER:
				result = BlockRamPrimitives.RAMB16BWER;
				break;
			case RAMB18E1:
				result = BlockRamPrimitives.RAMB18E1FIFO18E1;
				break;
			case RAMB36E1:
				result = BlockRamPrimitives.RAMB36E1FIFO36E1;
				break;
			default:
				throw new RuntimeException("Unsupported block ram primitive: " + instance.getType());
		}
		addInstance(instance.getName(), result);
	}

	private Optional<String> getLogicalNameIfPresent(Instance instance, String physicalName) {
		Attribute attr = instance.getAttribute(physicalName);
		if (attr == null) {
			if (!physicalName.endsWith("5FF")) {
				final String possible = instance.getAttributes().stream().map(Attribute::getPhysicalName).sorted().collect(Collectors.joining(", "));
				throw new NullPointerException(instance.getType() + " " + instance.getName() + " does not have attr " + physicalName + ".\nit has " + possible);
			} else {
				return Optional.empty();
			}
		} else {
			if (!attr.getValue().equals("#OFF")) {
				return Optional.of(attr.getLogicalName());
			}
			return Optional.empty();
		}
	}

	private void handleSlice(Instance instance) {
		//Routethrough?
		if (instance.getAttribute("_NO_USER_LOGIC") != null) {
			//Empty
			return;
		}

		addInstance(instance.getName(), NonChangingPrimitives.SLICE);

		for (char c = 'A'; c <= 'D'; c++) {
			Optional<String> lut6 = getLogicalNameIfPresent(instance, c + "6LUT");
			Optional<String> lut5 = getLogicalNameIfPresent(instance, c + "5LUT");

			//If lut6 is present but not lut5, the lut is not fractured. lut6 uses both lut5 halves --> counts double
			if (lut6.isPresent()) {
				addInstance(lut6.get(), NonChangingPrimitives.LUT);
				if (!lut5.isPresent()) {
					addInstance(lut6.get(), NonChangingPrimitives.LUT);
				}
			}
			lut5.ifPresent(n -> addInstance(n, NonChangingPrimitives.LUT));

			Optional<String> ff6 = getLogicalNameIfPresent(instance, c + "FF");
			Optional<String> ff5 = getLogicalNameIfPresent(instance, c + "5FF");

			ff6.ifPresent(n -> addInstance(n, NonChangingPrimitives.REGISTER));
			ff5.ifPresent(n -> addInstance(n, NonChangingPrimitives.REGISTER));
		}
	}
}
