/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.maxFreq;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.ClkPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.evaluation.Reports;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBooleanParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IFloatParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IIntegerParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IToolchainAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.SpartanMCMakeUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.stream.Collectors;

/**
 * Find the maximum frequency that a project can be synthesized with
 */
public class FindMaxFreq {

	/**
	 * Mega unit prefix
	 */
	public static final double MEGA = 1E6;

	/**
	 * Giga unit prefix
	 */
	public static final float GIGA = 1E9f;
	public static final double KILO = 1E3f;
	public static final String CLOCK_MODULE_NAME = "clk_xilinx";
	public static final double MINIMUM_FREQUENCY = 20E6;

	private static final Logger logger = LoggerFactory.getLogger(FindMaxFreq.class);
	private static final String clockDiv2 = "CLKIN_DIVIDE_BY_2";
	private static final String clockDiv7Series = "DIVCLK_DIVIDE";
	private static final float upperBoundMultiplier = 1.3f;
	private static final String OUTPUT_SEPERATOR = "-------------------------------------------------";

	private static Builder rapidSocBuilder = (pd, r) -> SpartanMCMakeUtil.runMakeTarget(pd, Collections.singletonMap("SPARTANMC_ROOT", r), "rapidSoc", "+args=\"\"");
	private static Builder regularBuilder = (pd, r) -> SpartanMCMakeUtil.runMakeTarget(pd, Collections.singletonMap("SPARTANMC_ROOT", r), "all");

	private final Builder builder;
	private final String outputFilePrefix;
	private final IModuleConfiguration<?> sysclk;
	private final boolean noLowerMaxFreq;
	private final LibJConfigProject project;

	private final Map<Float, StreamUtil.Pair<Float, Float>> frequencyMultipliers;
	private final List<Float> frequencies;
	private final float upperBound;
	private final float lowerBound;
	private final int tries;
	private boolean is7SeriesDCM;


	/**
	 * Create instance
	 *
	 * @param project The project to use
	 * @param builder How to build the project
	 * @param outputFilePrefix Name prefix of output files
	 * @param upperBound Initial upper bound to use when searching for upper frequency limit
	 * @param lowerBound the initial lower bound
	 * @param tries how many tries should be performed
	 * @param noLowerMaxFreq
	 */
	public FindMaxFreq(LibJConfigProject project, Builder builder, String outputFilePrefix, float upperBound, float lowerBound, int tries, boolean noLowerMaxFreq) {
		this.builder = builder;
		this.outputFilePrefix = outputFilePrefix;
		this.project = project;
		this.upperBound = upperBound;
		this.lowerBound = lowerBound;
		this.tries = tries;
		this.noLowerMaxFreq = noLowerMaxFreq;


		sysclk = project.getDocumentHandler().getDocument().streamModulesRecursive().filter(x -> x.getModuleDescriptionName().startsWith(CLOCK_MODULE_NAME)).collect(StreamUtil.exactlyOne());
		if (!sysclk.getModuleDescriptionName().equals(CLOCK_MODULE_NAME)) {
			is7SeriesDCM = true;
		}

		//CHECKSTYLE.OFF: MultipleStringLiterals
		if (is7SeriesDCM) {
			final Optional<? extends IParameter> clkin_divide = sysclk.findParameterDescription(clockDiv7Series);
			if (clkin_divide.isPresent()) {
				if (sysclk.getParameterHandler(IIntegerParameterHandler.class, clkin_divide.get()).getIntegerValue().get().intValue() != 1) {
					throw new RuntimeException("The maximum Frequency can not be found if " + clockDiv7Series + "is set!");
				}
			} else {
				logger.warn("Can not find DCM Parameter " + clockDiv7Series + " for a check");
			}
		} else {
			final Optional<? extends IParameter> clkin_divide_by_2 = sysclk.findParameterDescription(clockDiv2);
			if (clkin_divide_by_2.isPresent()) {
				if (sysclk.getParameterHandler(IBooleanParameterHandler.class, clkin_divide_by_2.get()).isSet()) {
					throw new RuntimeException("The maximum Frequency can not be found if " + clockDiv2 + "is set!");
				}
			} else {
				logger.warn("Can not find DCM Parameter " + clockDiv2 + " for a check");
			}
		}
		//CHECKSTYLE.ON: MultipleStringLiterals

		frequencyMultipliers = getPossibleFrequencies();
		frequencies = frequencyMultipliers.keySet().stream().sorted().collect(Collectors.toList());
	}


	/**
	 * Find the maximum frequency of a regular build
	 *
	 * @param project The project to use
	 * @param upperBound Initial upper bound to use when searching for upper frequency limit
	 * @param lowerBound the initial lower frequency
	 * @param tries how many tries should be performed
	 * @param noLowerMaxFreq disable searching for a lower max freq where all runs need to succeed
	 * @return Information about the maximum frequency
	 *
	 * @throws IOException on broken IO
	 * @throws InterruptedException when interrupted
	 */
	public static MaxFreqResultRange findRegularMaxFreq(LibJConfigProject project, float upperBound, float lowerBound, int tries, boolean noLowerMaxFreq) throws IOException, InterruptedException {
		FindMaxFreq f = new FindMaxFreq(project, regularBuilder, Reports.SPARTANMC_PAR_PREFIX, upperBound, lowerBound, tries, noLowerMaxFreq);

		return f.findMaxFreq();
	}

	/**
	 * Find the maximum frequency of a rapidSoc build
	 *
	 * @param project The project to use
	 * @param upperBound Initial upper bound to use when searching for upper frequency limit
	 * @param lowerBound the initial lower frequency
	 * @param tries how many tries should be performed
	 * @param noLowerMaxFreq disable searching for a lower max freq where all runs need to succeed
	 * @return Information about the maximum frequency
	 *
	 * @throws IOException on broken IO
	 * @throws InterruptedException when interrupted
	 */
	public static MaxFreqResultRange findRapidSocMaxFreq(LibJConfigProject project, float upperBound, float lowerBound, int tries, boolean noLowerMaxFreq) throws IOException, InterruptedException {
		FindMaxFreq f = new FindMaxFreq(project, rapidSocBuilder, Reports.RAPID_PROTO_ROUTED_PREFIX, upperBound, lowerBound, tries, noLowerMaxFreq);
		return f.findMaxFreq();
	}

	private Map<Float, StreamUtil.Pair<Float, Float>> getPossibleFrequencies() {

		Map<Float, StreamUtil.Pair<Float, Float>> result = new HashMap<>();

		IPortHandler<? extends IPort> p = DocumentUtil.getPortHandler(sysclk, "clk_gen");
		long frequency = sysclk.resolveClkPort((ClkPort) p.getConnectorDescription()).get().longValue();

		//CHECKSTYLE.OFF: MagicNumber
		float multiStart = 2;
		float multiMax = 32;
		float multiStepWidth = 1;
		float divStart = 2;
		float divMax = 32;
		float divStepWidth = 1;
		if (is7SeriesDCM) {
			multiStart = 6;
			multiMax = 12;
			multiStepWidth = 0.125f;
			divStart = 1;
			divMax = 128;
			divStepWidth = 0.125f;
		}
		//CHECKSTYLE.ON: magicNumber

		for (float multiply = multiStart; multiply <= multiMax; multiply += multiStepWidth) {
			for (float divide = divStart; divide <= divMax; divide += divStepWidth) {
				float outFreq = frequency * multiply / divide;

				//Set min freq because uart_light cannot handle too low frequencies
				if (outFreq < MINIMUM_FREQUENCY || outFreq < lowerBound * MEGA) {
					continue;
				}

				StreamUtil.Pair<Float, Float> oldValue = result.get(outFreq);
				//We only replace if the new values are smaller than the old ones
				if (oldValue == null || (oldValue.getA() + oldValue.getB() > multiply + divide)) {
					result.put(outFreq, new StreamUtil.Pair<>(multiply, divide));
				}
			}
		}
		return result;
	}

	private MaxFreqResult buildWithFrequency(float frequency, ResultCombiner combiner, BiConsumer<Float, MaxFreqResult> singleRunFailConsumer) throws IOException, InterruptedException {
		logger.info("building at {}MHz", frequency / MEGA);
		StreamUtil.Pair<Float, Float> p = frequencyMultipliers.get(frequency);
		return buildWithFrequency(frequency, p.getA(), p.getB(), combiner, singleRunFailConsumer);
	}

	enum ResultCombiner {
		FIND_ERROR,
		FIND_SUCCESS;
	}

	private MaxFreqResult buildWithFrequencyTry(int tryNumber, Float multiplier, Float divider) throws IOException, InterruptedException {
		if (!project.getDocumentHandler().getDocument().getTarget().getToolchain().getParams().isEmpty()) {
			//Set verilog define
			final IToolchainAttributeHandler mapParam = DocumentUtil.getToolchainParamHandler(project.getDocumentHandler().getDocument().getTarget(), "MAP");
			mapParam.setStoredValue(mapParam.getDefaultValue() + " -t " + (tryNumber + 1));
		} else {
			logger.warn("Setting seed is currently not supported on this toolchain.");
		}

		Path outFile = project.getProjectDir().resolve("findMaxFreq.jc3");
		project.getDocumentHandler().setSaveLocation(outFile);
		project.getDocumentHandler().save();


		project.getDocumentHandler().build(true);

		double buildTime = builder.build(project.getProjectDir(), project.getApplication().getSpartanRoot().toAbsolutePath().toString());

		final Collection<TimingInfo> timings = TimingResultParser.getParser(project).loadTimingInfo(project.getBuildDir(), outputFilePrefix);
		TimingInfo timing = getTimingDomain(timings);

		logger.info("Design has {} timing violations, runs at {}MHz", timing.errors,timing.actual/MEGA);
		return new MaxFreqResult(timing, buildTime, multiplier, divider);
	}

	private TimingInfo getTimingDomain(Collection<TimingInfo> timings) {
		//Vivado currently reports only one timing domain...
		if (timings.size() == 1) {
			return timings.iterator().next();
		}

		final TimingInfo info = timings.stream().filter(t -> t.clockName.startsWith("TS_" + sysclk.getName() + "_")).collect(StreamUtil.exactlyOne());
		final Set<TimingInfo> otherErrors = timings.stream()
				.filter(t -> t != info)
				.filter(t -> t.errors > 0)
				.collect(Collectors.toSet());
		if (!otherErrors.isEmpty()) {
			throw new RuntimeException("Errors exist in the following other timing domains: " + otherErrors);
		}
		return info;
	}

	private MaxFreqResult buildWithFrequency(float frequency, Float multiplier, Float divider, ResultCombiner combiner, BiConsumer<Float, MaxFreqResult> singleRunFailConsumer) throws IOException, InterruptedException {

		final String format = "Multiplier value: {}, Divider Value: {}";
		if (is7SeriesDCM) {
			IFloatParameterHandler multiplierParam = DocumentUtil.getParameterHandler(sysclk, "CLKFBOUT_MULT_F", IFloatParameterHandler.class);
			IFloatParameterHandler dividerParam = DocumentUtil.getParameterHandler(sysclk, "CLKOUT0_DIVIDE_F", IFloatParameterHandler.class);
			multiplierParam.setValue(multiplier);
			dividerParam.setValue(divider);
			logger.info(format, multiplierParam.getValue().orError(), dividerParam.getValue().orError());
		} else {
			IIntegerParameterHandler multiplierParam = DocumentUtil.getParameterHandler(sysclk, "CLKFX_MULTIPLY", IIntegerParameterHandler.class);
			IIntegerParameterHandler dividerParam = DocumentUtil.getParameterHandler(sysclk, "CLKFX_DIVIDE", IIntegerParameterHandler.class);
			multiplierParam.setValue(multiplier.intValue());
			dividerParam.setValue(divider.intValue());
			logger.info(format, multiplierParam.getValue().orError(), dividerParam.getValue().orError());
		}


		MaxFreqResult result = null;
		for (int i = 0; i < tries; i++) {
			result = buildWithFrequencyTry(i, multiplier, divider);

			if ((result.timing.hasErrors() && combiner == ResultCombiner.FIND_ERROR)
					|| (!result.timing.hasErrors() && combiner == ResultCombiner.FIND_SUCCESS)) {
				return result;
			}

			if (result.timing.hasErrors() && singleRunFailConsumer != null) {
				singleRunFailConsumer.accept(frequency, result);
			}
		}
		return result;
	}

	/**
	 * Find the maximum frequency
	 *
	 * @return Information about the maximum Frequency
	 *
	 * @throws IOException On broken IO
	 * @throws InterruptedException when Interrupted
	 */
	public MaxFreqResultRange findMaxFreq() throws IOException, InterruptedException {

		//First, build with lowest frequency to get a baseline
		logger.info(OUTPUT_SEPERATOR);
		logger.info("Establishing baseline");
		logger.info(OUTPUT_SEPERATOR);
		float lower = frequencies.get(0);

		MaxFreqResult lowerInfo = buildWithFrequency(frequencies.get(0), noLowerMaxFreq? ResultCombiner.FIND_SUCCESS : ResultCombiner.FIND_ERROR, null);

		if (lowerInfo.timing.hasErrors()) {
			logger.warn("Result: {}", lowerInfo);
			throw new RuntimeException("baseline already has timing violations");
		}


		//Then, try to find an upper bound
		logger.info(OUTPUT_SEPERATOR);
		logger.info("Finding upper bound");
		logger.info(OUTPUT_SEPERATOR);

		final float[] allOkUpper = {Float.POSITIVE_INFINITY};

		BiConsumer<Float, MaxFreqResult> singleRunFailConsumer = (t, r) -> {
			if (t < allOkUpper[0]) {
				allOkUpper[0] = t;
			}
		};

		float upper = getNearestFreqBelow((float) (upperBound * MEGA));
		MaxFreqResult info;
		do {
			info = buildWithFrequency(upper, ResultCombiner.FIND_SUCCESS, singleRunFailConsumer);
			if (!info.timing.hasErrors()) {
				lower = upper;
				lowerInfo = info;
				upper = getNearestFreqBelow(upper * upperBoundMultiplier);

				if (upper == lower) {
					throw new IOException("Failed to find an upper limit");
				}
			}
		}
		while (!info.timing.hasErrors());

		if (Float.isInfinite(allOkUpper[0])) {
			allOkUpper[0] = upper;
		}


		logger.info(OUTPUT_SEPERATOR);
		logger.info("Starting interval halving for upper end");
		MaxFreqResult oneOk = intervalHalving(lower, lowerInfo, upper, ResultCombiner.FIND_SUCCESS, singleRunFailConsumer);
		logger.info(OUTPUT_SEPERATOR);
		MaxFreqResult allOk;
		if (tries > 1 && !noLowerMaxFreq) {
			logger.info("Starting interval halving for lower end");
			allOk = intervalHalving(lower, lowerInfo, allOkUpper[0], ResultCombiner.FIND_ERROR, null);
		} else {
			allOk = oneOk;
		}
		return new MaxFreqResultRange(allOk, oneOk);
	}

	private MaxFreqResult intervalHalving(float lower, MaxFreqResult lowerInfo, float upper, ResultCombiner combiner, BiConsumer<Float, MaxFreqResult> singleRunFailConsumer) throws IOException, InterruptedException {
		MaxFreqResult info;
		while (true) {
			logger.info(OUTPUT_SEPERATOR);
			logger.info("searching between " + lower / MEGA + "MHz and " + upper / MEGA + "MHz");
			Float mid = findMidFreq(lower, upper);
			if (mid == null) {
				logger.info("No frequency exists between, finished.");
				break;
			}

			info = buildWithFrequency(mid, combiner, singleRunFailConsumer);
			if (info.timing.hasErrors()) {
				upper = mid;
			} else {
				lower = mid;
				lowerInfo = info;
			}
		}
		return lowerInfo;
	}

	/**
	 * Find the middle frequency between the given limits.
	 *
	 * First, get the true middle frequency and find the nearest allowed frequency below it. If that frequency
	 * is the lower limit, get the middle allowed frequency between the two. If there is none, return null.
	 *
	 * @param lower lower bound
	 * @param upper upper bound
	 * @return the middle allowed frequency or null if there is none
	 */
	private Float findMidFreq(float lower, float upper) {
		int lowerIdx = frequencies.indexOf(lower);
		int upperIdx = frequencies.indexOf(upper);
		//If there is no value inbetween, return null;
		if (upperIdx == lowerIdx + 1) {
			return null;
		}

		int between = upperIdx - lowerIdx - 1;
		int numTries = (int) Math.ceil(Math.log(between) / Math.log(2)) - 1;
		logger.info("About {} tries remaining", numTries);

		//Try to get middle frequency
		float trueMid = (upper + lower) / 2;
		float adjMid = getNearestFreqBelow(trueMid);
		//Only return it if it is not rounded to lower
		if (adjMid != lower) {
			return adjMid;
		}

		//return middle value by indices
		return frequencies.get((lowerIdx + upperIdx) / 2);
	}

	private float getNearestFreqBelow(float f) {
		Float last = 0f;
		for (Float frequency : frequencies) {
			if (frequency > f) {
				return last;
			}
			last = frequency;
		}
		return last;
	}

	/**
	 * Function that can build the project
	 */
	@FunctionalInterface
	public interface Builder {

		float build(Path t, String root) throws IOException, InterruptedException;
	}
}
