/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.maxFreq;

import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigBuilder;
import de.tu_darmstadt.rs.util.table.Column;
import de.tu_darmstadt.rs.util.table.TableFormat;
import de.tu_darmstadt.rs.util.table.TableFormatter;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import com.beust.jcommander.Parameters;
import com.beust.jcommander.ParametersDelegate;

import java.io.IOException;

/**
 * JCommander Command for finding the maximum frequency
 */
@SuppressWarnings("checkstyle:SystemOut")
@Parameters(commandNames = "maxFreq", commandDescription = "Find the maximum frequency a project will synthesize with by synthesizing it with increasing frequency constraints.")
public class FindMaxFreqCommand implements Runnable {

	public static final int DEFAULT_UPPER_BOUND = 85;
	public static final int DEFAULT_LOWER_BOUND = 20;
	public static final int DEFAULT_TRIES = 1;

	@Parameter(names = {"--type", "-t"}, description = "Type of build to use")
	private Type type = Type.Regular;
	@Parameter(names = {"--upper"}, description = "First Frequency in MHz to try when establishing an upper limit. Setting this value closely above the (expected) result reduces the number of tries.")
	private float upperBound = DEFAULT_UPPER_BOUND;
	@Parameter(names = {"--lower"}, description = "First Frequency in MHz to try when establishing an lower limit. Setting this value closely below the (expected) result reduces the number of tries. (Minimum = 20)")
	private float lowerBound = DEFAULT_LOWER_BOUND;
	@Parameter(names = {"--tries"}, description = "Number of synthesis tries")
	private int tries = DEFAULT_TRIES;
	@Parameter(names = "--no-lowest-result", description = "Do not try to find a lower max frequency where all tries need to succeed.")
	private boolean noLowerMaxFreq = false;
	@ParametersDelegate
	private LibJConfigBuilder builder = new LibJConfigBuilder("Evaluation", true);


	public FindMaxFreqCommand() {
	}

	private void outputRunInfo(TableFormatter table, MaxFreqResult info) {
		String indentation = ((tries > 1) && !noLowerMaxFreq)? "  " : "";
		table.addContentRow(indentation + "Resulting target Frequency (highest possible DCM frequency [x" + info.multiplier + ", ÷" + info.divider + "])", new Object[]{info.timing.requirement / FindMaxFreq.MEGA, "MHz"});
		if (info.timing.actual != null) {
			table.addContentRow(indentation + "Resulting actual Frequency (DCM freq + slack)", new Object[]{info.timing.actual / FindMaxFreq.MEGA, "MHz"});
		}
		table.addContentRow(indentation + "Time to synthesize single run in maximum frequency design", new Object[]{info.buildTime / FindMaxFreq.KILO, "s"});
	}

	@Override
	public void run() {
		try {
			if (tries < 1 || tries > 100) {
				throw new ParameterException("The number of tries must be between 1 and 100");
			}

			builder.getSpartanRoot(); // force check ROOT before start
			builder.requiresConfigFile();
			builder.useLastBuiltConfig();
			builder.autoOpenConfig();

			LibJConfigProject project = builder.build();

			long before = System.nanoTime();
			final MaxFreqResultRange result = type.freqFinder.findMaxFreq(project, upperBound, lowerBound, tries, noLowerMaxFreq);
			long after = System.nanoTime();
			double runtime = (after - before) / FindMaxFreq.GIGA;


			TableFormatter table = new TableFormatter(TableFormat.CHARACTERS, Column.STRING, new Column("%.2f %-3s", Column.Alignment.RIGHT));
			if (tries > 1 && !noLowerMaxFreq) {
				table.addContentRow("Lowest result (all runs succeeded)", null);
				outputRunInfo(table, result.allOk);
				table.addContentRow("", null);
				table.addContentRow("Highest result (at least one run succeeded)", null);
			}
			outputRunInfo(table, result.oneOk);
			table.addContentRow("", null);
			table.addContentRow("Total Runtime", new Object[]{runtime, "s"});


			table.output().forEach(System.out::println);
		} catch (InterruptedException | IOException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Function to call to get maximum frequency
	 */
	@FunctionalInterface
	interface MaxFreqFinder {

		MaxFreqResultRange findMaxFreq(LibJConfigProject project, float upperBound, float lowerBound, int tries, boolean noLowerMaxFreq) throws IOException, InterruptedException;
	}


	/**
	 * Types of maximum frequency finders
	 */
	public enum Type {
		Regular(FindMaxFreq::findRegularMaxFreq),
		RapidSoc(FindMaxFreq::findRapidSocMaxFreq);

		public final MaxFreqFinder freqFinder;

		Type(MaxFreqFinder freqFinder) {
			this.freqFinder = freqFinder;
		}
	}
}
