/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.maxFreq;

import de.tu_darmstadt.rs.spartanmc.evaluation.Reports;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

/**
 * Created by wenzel on 07.04.17.
 */
public class IseTimingResultParser implements TimingResultParser {

	private Path reportPath;

	/**
	 * Find the named subnode in a node's sublist.
	 * <ul>
	 * <li>Ignores comments and processing instructions.
	 * <li>Ignores TEXT nodes (likely to exist and contain
	 * ignorable whitespace, if not validating.
	 * <li>Ignores CDATA nodes and EntityRef nodes.
	 * <li>Examines element nodes to find one with
	 * the specified name.
	 * </ul>
	 *
	 * @param node the element node to start searching from
	 * @param name the tag name for the element to find
	 * @return the Node found
	 */
	static Node findSubNode(Node node, String name) {
		if (node.getNodeType() != Node.ELEMENT_NODE) {
			throw new RuntimeException("Error: Search node not of element type");
		}

		if (!node.hasChildNodes()) {
			return null;
		}

		NodeList list = node.getChildNodes();
		for (int i = 0; i < list.getLength(); i++) {
			Node subnode = list.item(i);
			if (subnode.getNodeType() == Node.ELEMENT_NODE) {
				if (subnode.getNodeName().equals(name)) {
					return subnode;
				}
			}
		}
		return null;
	}

	static Node findSubNode(Node node, String... names) {
		Node current = node;
		for (String name : names) {
			current = findSubNode(current, name);
			if (current == null) {
				return null;
			}
		}
		return current;
	}

	/**
	 * Load timing information from a twx file
	 *
	 * @param buildDir the build directory
	 * @param outputFilePrefix prefix of the filename
	 * @return Timing information stored in the file
	 *
	 * @throws IOException on broken IO
	 */
	@Override
	public Collection<TimingInfo> loadTimingInfo(Path buildDir, String outputFilePrefix) throws IOException {
		reportPath = buildDir.resolve(outputFilePrefix+ Reports.PAR_TIMING_FILE_ENDING);
		if (!Files.exists(reportPath)) {
			throw new IOException("ptwx does not exist at " + reportPath);
		}


		DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
		try {
			DocumentBuilder db = dbf.newDocumentBuilder();
			Document doc = db.parse(reportPath.toFile());

			Set<TimingInfo> results = new HashSet<>();

			Node summaryTable = findSubNode(doc.getDocumentElement(), "twBody", "twSumRpt", "twConstRollupTable");
			if (summaryTable != null) {
				for (int i = 0; i < summaryTable.getChildNodes().getLength(); i++) {
					Node child = summaryTable.getChildNodes().item(i);
					if (!child.getNodeName().equals("twConstRollup")) {
						continue;
					}


					Element e = (Element) child;

					String name = e.getAttribute("name");
					float requirement = Float.parseFloat(e.getAttribute("requirement"));
					float actual = Float.parseFloat(e.getAttribute("actual"));
					float requirementFreq = FindMaxFreq.GIGA / requirement;
					float actualFreq = FindMaxFreq.GIGA / actual;
					int errors = Integer.parseInt(e.getAttribute("errors"));


					results.add(new TimingInfo(name, requirementFreq, actualFreq, errors));

				}
			}
			return results;
		} catch (ParserConfigurationException | SAXException e) {
			throw new IOException(e);
		}
	}

	@Override
	public Path getTimingReportPath() {
		return reportPath;
	}
}
