/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.maxFreq;

import de.tu_darmstadt.rs.jcommander.validators.DirectoryExistsValidator;
import de.tu_darmstadt.rs.spartanmc.evaluation.Reports;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryMap;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.Parameters;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collection;

/**
 * The Command for showing timing violations
 */
@Parameters(commandNames = "timingViolation", commandDescription = "Tells if the current hardware design holds timing violations")
public class TimingViolationCheckerCommand implements Runnable {


	private static final Logger log = LoggerFactory.getLogger(TimingViolationCheckerCommand.class);
	private static final Path TIMING_CHECKFILE = Reports.BUILD.resolve(Paths.get("timingViolations.log"));
	private static final String OK_TOKEN = "OK";
	@Parameter(names = {"-q", "--quiet"}, description = "Supress info messages")
	private boolean quiet;
	@Parameter(names = {"-e", "--exception"}, description = "Throw exception on timing violation")
	private boolean throwException;
	@Parameter(names = {"--no-fat"}, description = "Don't show the fat timing violation error")
	private boolean noFatTimingError;
	@Parameter(names = {"--parseRapidSocTiming"}, description = "parse rapidSoc input")
	private boolean parseRapidSoc;
	@Parameter(names = {"--parsePlaceRouteTiming"}, description = "parse place and route report from ISE or Vivado input")
	private boolean parsePar;
	private String timingCheckfileContent;


	@Parameter(names = {"-p", "--project-dir"}, description = "Select a folder other than the current working directory as project directory", validateValueWith = DirectoryExistsValidator.class)
	private Path projectDir = Paths.get("."); //working dir is jc3 dir by default


	public TimingViolationCheckerCommand() {

	}

	@Override
	@SuppressWarnings("checkstyle:SystemOut")
	public void run() {

		try {
			if (parsePar || parseRapidSoc) {
				parseReport();
			} else {
				printInfo(!isTimingOK(), getUsedReportInTimingFile(), false);
			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private void parseReport() throws IOException {

		Path memFile = projectDir.resolve("system/memory.xml");
		final MemoryMap memoryMap = MemoryMap.fromFile(memFile);
		Path buildDir = projectDir.resolve("build");


		final TimingResultParser parser = TimingResultParser.getParser(memoryMap);
		Collection<TimingInfo> timingInfo;

		if (parseRapidSoc) {
			 timingInfo = parser.loadTimingInfo(buildDir, Reports.RAPID_PROTO_ROUTED_PREFIX);
		} else if (parsePar) {
			timingInfo = parser.loadTimingInfo(buildDir, Reports.SPARTANMC_PAR_PREFIX);
		} else {
			throw new RuntimeException("Nothing to parse...");
		}

		boolean hasErrors = timingInfo.stream().anyMatch(TimingInfo::hasErrors);

		Path relativePath = Paths.get(".").toAbsolutePath().getParent().relativize(parser.getTimingReportPath().toAbsolutePath());
		printInfo(hasErrors, relativePath.toString(), true);
	}

	private void printInfo(boolean hasErrors, String reportPath, boolean writeToFile) {
		if (hasErrors) {
			logError("The synthesis timing report holds not matching timing contraints! ...see " + reportPath + " for details");
			if (writeToFile) {
				writeTimingCheckfile(Paths.get(".").toAbsolutePath().resolve(TIMING_CHECKFILE), reportPath);
			}
			return;
		} else {
			if (!quiet) {
				log.info("All timing constraints met.");
			}
			if (writeToFile) {
				writeTimingCheckfile(Paths.get(".").toAbsolutePath().resolve(TIMING_CHECKFILE), OK_TOKEN);
			}
		}
	}

	private void logError(String msg) {
		if (!noFatTimingError) {
			showFatTimingError();
		}
		log.error(msg);
		if (throwException) {
			throw new RuntimeException(msg);
		}
	}


	@SuppressWarnings("checkstyle:SystemOut")
	private void showFatTimingError() {
		System.out.println("\033[0;31m");
		System.err.println("████████ ██ ███    ███ ██ ███    ██  ██████      ███████ ██████  ██████   ██████  ██████");
		System.err.println("   ██    ██ ████  ████ ██ ████   ██ ██           ██      ██   ██ ██   ██ ██    ██ ██   ██");
		System.err.println("   ██    ██ ██ ████ ██ ██ ██ ██  ██ ██   ███     █████   ██████  ██████  ██    ██ ██████");
		System.err.println("   ██    ██ ██  ██  ██ ██ ██  ██ ██ ██    ██     ██      ██  ██  ██  ██  ██    ██ ██  ██");
		System.err.println("   ██    ██ ██      ██ ██ ██   ████  ██████      ███████ ██   ██ ██   ██  ██████  ██   ██");
		System.out.println("\033[0m");
	}

	private void writeTimingCheckfile(Path reportpath, String content) {
		try {
			Files.write(reportpath.toAbsolutePath(), content.getBytes());
		} catch (IOException e) {
			log.error("Something went wrong with the output file");
			e.printStackTrace();
		}
	}

	private boolean isTimingOK() {
		if (readTimingCheckfile().equalsIgnoreCase(OK_TOKEN)) {
			return true;
		}
		return false;
	}

	private String getUsedReportInTimingFile() {
		if (isTimingOK()) {
			return null;
		} else {
			return readTimingCheckfile();
		}
	}

	private String readTimingCheckfile() {
		try {
			if (timingCheckfileContent == null) {
				timingCheckfileContent = new String(Files.readAllBytes(TIMING_CHECKFILE));
			}
			return timingCheckfileContent;
		} catch (IOException e) {
			log.error("Something went wrong with the input file");
			e.printStackTrace();
		}
		return timingCheckfileContent;
	}
}