/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.synthesis;

import de.tu_darmstadt.rs.fpga.resources.FpgaFamily;
import de.tu_darmstadt.rs.fpga.resources.HierarchicalUsage;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.FpgaPrimitive;
import de.tu_darmstadt.rs.spartanmc.evaluation.Reports;
import de.tu_darmstadt.rs.spartanmc.evaluation.maxFreq.TimingInfo;
import de.tu_darmstadt.rs.spartanmc.evaluation.maxFreq.TimingResultParser;
import de.tu_darmstadt.rs.spartanmc.evaluation.synthesis.argumentCollection.ArgumentParser;
import de.tu_darmstadt.rs.spartanmc.evaluation.synthesis.argumentCollection.Arguments;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigBuilder;
import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.table.Column;
import de.tu_darmstadt.rs.util.table.TableFormat;
import de.tu_darmstadt.rs.util.table.TableFormatter;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import com.beust.jcommander.Parameters;
import com.beust.jcommander.ParametersDelegate;
import com.beust.jcommander.converters.FileConverter;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.StandardOpenOption;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.function.IntFunction;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * The Command for creating the hardware report
 */
@Parameters(commandNames = "hardwareReport", commandDescription = "Shows the important numbers of the synthesis hardware report in a nice format")
public class ReportCreatorCommand implements Runnable {


	private static final Logger log = LoggerFactory.getLogger(ReportCreatorCommand.class);
	@Parameter(names = {"-o", "--outputFile"}, description = "File to write results to", converter = FileConverter.class)
	private File outputFile = null;
	@Parameter(names = {"--format"}, description = "output format")
	private TableFormat outputFormat = TableFormat.CHARACTERS;
	@Parameter(names = {"-a", "--append"}, description = "append output...simply don't print the header again")
	private boolean append;
	@Parameter(names = {"-s", "--source"}, description = "Data source. See below for details", converter = HardwareSourceConverter.class)
	private IHardwareUsageSource source = HardwareSourceConverter.getDefault();

	private List<Arguments> arguments;

	@ParametersDelegate
	private LibJConfigBuilder builder = new LibJConfigBuilder("Evaluation", true);

	private LibJConfigProject project;

	public ReportCreatorCommand() {

	}

	private void buildLibJconfig() {

		builder.getSpartanRoot(); // force check ROOT before start
		//builder.requiresConfigFile();
		if (!builder.hasExplicitConfigFile()) {
			builder.useLastBuiltConfig();
		}
		builder.autoOpenConfig();

		project = builder.build();
	}


	@Override
	@SuppressWarnings("checkstyle:SystemOut")
	public void run() {
		try {

			buildLibJconfig();

			if (!(builder.hasExplicitConfigFile() || project.getLastConfigFile().isPresent())) {
				log.error("No previous configuration file found and no jc3 file specified!");
				return;
			}


			String usedDevice = project.getDocumentHandler().getDocument().getTarget().getDevice().getName();
			final Optional<FpgaFamily> optionalFamily = FpgaFamily.getFamilyIfSupported(usedDevice);

			if (!optionalFamily.isPresent()) {
				log.error("The used device is not Supported.");
				return;
			}

			if (source == null) {
				throw new ParameterException("No source specified");
			}

			IHardwareUsage iusage = source.loadDesign(project, optionalFamily.get());

			Stream<String> frequencies = reportFrequencies(iusage);
			final Stream<String> usage = reportUsage(iusage);

			final Stream<String> argumentReport;
			if (!append) {
				argumentReport = Stream.concat(
						reportArguments(),
						Stream.of("")
				);
			} else {
				argumentReport = Stream.empty();
			}


			Stream<String> output = StreamUtil.concat(argumentReport, frequencies, usage);


			if (outputFile == null) {
				output.forEach(System.out::println);
			} else {

				try {
					StandardOpenOption[] openOption;
					if (append) {
						openOption = new StandardOpenOption[]{StandardOpenOption.APPEND, StandardOpenOption.WRITE};
					} else {
						openOption = new StandardOpenOption[]{StandardOpenOption.CREATE, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.WRITE};
					}

					Files.write(outputFile.toPath(), (Iterable<String>) output::iterator, openOption);
				} catch (IOException e) {
					log.error("Something went wrong with the output file");
					e.printStackTrace();
				}

			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private Stream<String> reportUsage(IHardwareUsage iusage) throws IOException {
		HierarchicalUsage usage = iusage.getUsage();


		final List<FpgaPrimitive> types = usage.getUsedPrimitives().sorted(FpgaPrimitive.comparator).collect(Collectors.toList());

		Stream<String> nameStream;
		Stream<Column> nameColumnStream;
		if (!usage.getSubs().isEmpty()) {
			nameStream = Stream.of("Name");
			nameColumnStream = Stream.of(new Column("%s", Column.Alignment.LEFT));
		} else {
			nameStream = Stream.empty();
			nameColumnStream = Stream.empty();
		}


		Column[] columns = Stream.concat(
				nameColumnStream,
				types.stream().map(x -> new Column("%d", Column.Alignment.RIGHT))
		).toArray(Column[]::new);

		TableFormatter table = new TableFormatter(outputFormat,
				columns
		);


		table.addHeaderRow(
				Stream.concat(
						nameStream,
						types.stream().map(FpgaPrimitive::getReadableString)
				).toArray(String[]::new)
		);


		if (!usage.getSubs().isEmpty()) {
			table.addTree(usage, HierarchicalUsage::getSubs, d -> d.toColumns(true, types));
		} else {
			table.addContentRow(usage.toColumns(false, types));
		}

		return Stream.concat(Stream.of(iusage.getInfo(),""),table.output());
	}

	@NotNull
	private Stream<String> reportArguments() {
		ArgumentParser argumentParser = new ArgumentParser(project);
		arguments = argumentParser.getArguments();
		return Stream.of("Most important options:",
				arguments.stream().map(Arguments::toStringForHWEval).collect(Collectors.joining("\n\t")));
	}

	private <T> T[] optionalColumn(T[] alwaysColumns, boolean include, Supplier<T> optionalColumnCreator, IntFunction<T[]> arrayGenerator) {
		if (include) {
			return Stream.concat(Arrays.stream(alwaysColumns), Stream.of(optionalColumnCreator.get())).toArray(arrayGenerator);
		}
		return alwaysColumns;
	}

	private Stream<String> reportFrequencies(IHardwareUsage usage) throws IOException {
		Collection<TimingInfo> timingInfos = TimingResultParser.getParser(project).loadTimingInfo(project.getBuildDir(), Reports.SPARTANMC_PAR_PREFIX);

		if (timingInfos.stream().anyMatch(TimingInfo::hasErrors)) {
			log.error("The synthesis timing report holds not matching timing contraints!");
		}
		Stream<String> frequencies;
		if (timingInfos.size() > 1) {
			boolean usageSupported = source.usageByTimingSupported();

			TableFormatter freqTable = new TableFormatter(outputFormat,
					optionalColumn(
							new Column[]{
									Column.STRING,
									new Column("%.3f", Column.Alignment.RIGHT),
									new Column("%.3f", Column.Alignment.RIGHT)
							},
							usageSupported,
							() -> new Column("%d", Column.Alignment.RIGHT),
							Column[]::new
					)
			);
			freqTable.addHeaderRow(optionalColumn(new String[]{"Name", "Requirement (MHz)", "Actual (MHz)"}, usageSupported, () -> "Usage", String[]::new));
			timingInfos.stream().sorted(Comparator.comparing(ti -> ti.clockName)).forEach(t -> {
				freqTable.addContentRow(optionalColumn(
						new Object[]{t.clockName, t.requirement / TimingInfo.MEGA, t.actual / TimingInfo.MEGA},
						usageSupported,
						() -> usage.getUsageByTimingDomain(t.clockName),
						Object[]::new
				));
			});
			frequencies = StreamUtil.concat(Stream.of("The design uses multiple frequencies:"), freqTable.output(), Stream.of(""));

		} else {
			TimingInfo t = timingInfos.iterator().next();
			frequencies = Stream.of(String.format("%.3f MHz required, %.3f MHz actual", t.requirement / TimingInfo.MEGA, t.actual / TimingInfo.MEGA));
		}
		return frequencies;
	}


}
