/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.synthesis;

import de.tu_darmstadt.rs.fpga.resources.FpgaFamily;
import de.tu_darmstadt.rs.fpga.resources.HierarchicalUsage;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.FpgaResourceType;
import de.tu_darmstadt.rs.spartanmc.evaluation.Reports;
import de.tu_darmstadt.rs.spartanmc.evaluation.synthesis.reportParsing.HardwareReportParser;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.stream.Collectors;

public class ReportHardwareUsageSource implements IHardwareUsageSource {

	private static final Logger logger = LoggerFactory.getLogger(ReportHardwareUsageSource.class);

	private HardwareReportParser parReport;
	private HardwareReportParser mapReport;

	private HierarchicalUsage getUsage(LibJConfigProject project, FpgaFamily fpgaFamily) throws IOException {

		Path parReportPath = project.getBuildDir().resolve(Reports.PAR_REPORT);
		Path mapReportPath = project.getBuildDir().resolve(Reports.MAP_REPORT);

		InputStreamReader parIsr = new InputStreamReader(new FileInputStream(parReportPath.toFile()));
		InputStreamReader mapIsr = new InputStreamReader(new FileInputStream(mapReportPath.toFile()));
		parReport = new HardwareReportParser(parIsr, fpgaFamily);
		mapReport = new HardwareReportParser(mapIsr, fpgaFamily);


		parReport.parse();
		mapReport.parse();

		boolean allTypesFoundInParReport = parReport.getResources().stream().map(res -> res.getPrimitive().getType()).collect(Collectors.toList()).containsAll(Arrays.asList(FpgaResourceType.values()));

		boolean allTypesFoundInMapReport = mapReport.getResources().stream().map(res -> res.getPrimitive().getType()).collect(Collectors.toList()).containsAll(Arrays.asList(FpgaResourceType.values()));


		if (!(allTypesFoundInMapReport || allTypesFoundInParReport)) {
			logger.error("Not all resource types have been found in the report. Propably you are reading an old report for a different device than the latest jConfig build. Run synthesis again!");
		}

		final HierarchicalUsage usage = new HierarchicalUsage("Design");

		parReport.getResources().forEach(res -> {
			if (res.getUsedResources() > 0) {
				usage.addResource(res);
			}
		});

		return usage;

	}

	@Override
	public IHardwareUsage loadDesign(LibJConfigProject project, FpgaFamily fpgaFamily) throws IOException {
		final HierarchicalUsage usage = getUsage(project, fpgaFamily);
		return new IHardwareUsage() {
			@Override
			public String getInfo() {
				return "The report shows 6-input LUTs. The two 5-input LUTs belonging together are only counted as one LUT.";
			}

			@Override
			public HierarchicalUsage getUsage() throws IOException {
				return usage;
			}

			@Override
			public int getUsageByTimingDomain(String timingDomain) {
				throw new UnsupportedOperationException();
			}
		};
	}

	@Override
	public boolean usageByTimingSupported() {
		return false;
	}

	@Override
	public String describe() {
		return "Less detailed, but always available. Parses MAP/PAR report";
	}

	@Override
	public String getName() {
		return "ISE";
	}

	@Override
	public int getPriority() {
		return 0;
	}


	@Override
	public String toString() {
		return getName();
	}
}
