/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.evaluation.synthesis.reportParsing;

import de.tu_darmstadt.rs.fpga.resources.FpgaResource;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.FpgaResourceType;
import de.tu_darmstadt.rs.fpga.resources.FpgaFamily;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.FpgaPrimitive;

import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Creates fpga resources for a given report line and device
 */
public final class FpgaResourceFactory {

	private FpgaResourceFactory() {

	}

	/**
	 * produces a FPGA resource if a resource can be found in the line for the given device
	 * @param reportLine a line from the report
	 * @param device a supported device
	 * @return a FPGA resource if a resource can be found in the line for the given device
	 */
	public static Optional<FpgaResource> produce(String reportLine, FpgaFamily device) {
		for (FpgaResourceType type : FpgaResourceType.values()) {
			for (Pattern pattern : type.getEvaluationReportPattern(device)) {
				final Matcher matcher = pattern.matcher(reportLine);
				if (matcher.matches()) {

					int used = new Integer(matcher.group(FpgaResourceReportFields.USED.getCaptureGroup()).replaceAll(",", "")).intValue();
					int total = new Integer(matcher.group(FpgaResourceReportFields.TOTAL.getCaptureGroup()).replaceAll(",", "")).intValue();
					int percent = new Integer(matcher.group(FpgaResourceReportFields.PERCENTAGE.getCaptureGroup()).replaceAll("%", "")).intValue();
					Optional<FpgaPrimitive> primitive = type.findPrimitiveInLine(reportLine);
					if (primitive.isPresent()) {
						return Optional.of(new FpgaResource(used, total, primitive.get()));
					}
				}
			}
		}
		return Optional.empty();
	}

	/**
	 * order of the capturing groups for the pattern
	 */
	private enum FpgaResourceReportFields {
		USED(1),
		TOTAL(2),
		PERCENTAGE(3);

		private int captureGroup;

		FpgaResourceReportFields(int captureGroup) {
			this.captureGroup = captureGroup;
		}

		public int getCaptureGroup() {
			return captureGroup;
		}
	}

}
