/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.evaluation.hardwareUsageTests;

import de.tu_darmstadt.rs.fpga.resources.FpgaFamily;
import de.tu_darmstadt.rs.fpga.resources.FpgaResource;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.BlockRamPrimitives;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.DspPrimitives;
import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.NonChangingPrimitives;
import de.tu_darmstadt.rs.spartanmc.evaluation.synthesis.reportParsing.HardwareReportParser;
import de.tu_darmstadt.rs.spartanmc.evaluation.synthesis.reportParsing.LibJConfigProjectUtils;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;

import com.googlecode.junittoolbox.ParallelParameterized;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

/**
 * tests the HW reports
 */
@RunWith(ParallelParameterized.class)
public class HardwareReportTester {

	private static final Logger log = LoggerFactory.getLogger(HardwareReportTester.class);
	private static final String xc6slxDocument = "/configuration/xc6slx.jc3";
	private static final String xc7aDocument = "/configuration/xc7a.jc3";

	@Rule
	public TemporaryFolder tempJC3File = new TemporaryFolder();

	@Parameterized.Parameter(0)
	public InputStreamReader inputReport;
	@Parameterized.Parameter(1)
	public List<FpgaResource> referenceResources;
	@Parameterized.Parameter(2)
	public String testName;
	@Parameterized.Parameter(3)
	public InputStream jconfigDocument;

	public HardwareReportTester() {

	}

	/**
	 * Collects the input data
	 * @return test data
	 * @throws IOException if test report reading goes wrong
	 */
	@Parameterized.Parameters(name = "{index}: {2}")
	public static Collection<Object[]> data() throws IOException {

		List tests = new ArrayList();
		tests.add(getxc6slxParTestData());
		tests.add(getxc6slxMapTestData());
		tests.add(getxc7aParTestData());
		tests.add(getxc7aMapTestData());
		return tests;
	}

	@SuppressWarnings("checkstyle:MagicNumber")
	private static Object[] getxc6slxParTestData() {
		//PAR tests

		InputStream parJconfigDocument = HardwareReportTester.class.getResourceAsStream(xc6slxDocument);
		InputStreamReader parIsr = new InputStreamReader(HardwareReportTester.class.getResourceAsStream("/synthesisReports/xc6slx.par"));
		FpgaResource bram16 = new FpgaResource(20 ,116, BlockRamPrimitives.RAMB16BWER);
		FpgaResource bram8 = new FpgaResource(0 ,232, BlockRamPrimitives.RAMB8BWER);
		FpgaResource registers = new FpgaResource(971, 54576, NonChangingPrimitives.REGISTER);
		FpgaResource luts = new FpgaResource(2696, 27288, NonChangingPrimitives.LUT);
		FpgaResource slices = new FpgaResource(928, 6822, NonChangingPrimitives.SLICE);
		FpgaResource dsps = new FpgaResource(2, 58, DspPrimitives.DSP48A1);
		List<FpgaResource> fpgaResources = new ArrayList<>();
		fpgaResources.add(bram16);
		fpgaResources.add(bram8);
		fpgaResources.add(registers);
		fpgaResources.add(luts);
		fpgaResources.add(slices);
		fpgaResources.add(dsps);

		return new Object[] { parIsr, fpgaResources,  "xc6slx par report parse test" , parJconfigDocument};
	}

	@SuppressWarnings("checkstyle:MagicNumber")
	private static Object[] getxc6slxMapTestData() {
		//MAP tests
		InputStream jConfigDocument = HardwareReportTester.class.getResourceAsStream(xc6slxDocument);
		InputStreamReader isr = new InputStreamReader(HardwareReportTester.class.getResourceAsStream("/synthesisReports/xc6slx.mrp"));
		FpgaResource bram16 = new FpgaResource(20 ,116, BlockRamPrimitives.RAMB16BWER);
		FpgaResource bram8 = new FpgaResource(0 ,232, BlockRamPrimitives.RAMB8BWER);
		FpgaResource registers = new FpgaResource(971, 54576, NonChangingPrimitives.REGISTER);
		FpgaResource luts = new FpgaResource(2696, 27288, NonChangingPrimitives.LUT);
		FpgaResource slices = new FpgaResource(928, 6822, NonChangingPrimitives.SLICE);
		FpgaResource dsps = new FpgaResource(2, 58, DspPrimitives.DSP48A1);
		List<FpgaResource> fpgaResources = new ArrayList<>();
		fpgaResources.add(bram16);
		fpgaResources.add(bram8);
		fpgaResources.add(registers);
		fpgaResources.add(luts);
		fpgaResources.add(slices);
		fpgaResources.add(dsps);
		return new Object[] { isr, fpgaResources, "xc6slx map report parse test", jConfigDocument };
	}

	@SuppressWarnings("checkstyle:MagicNumber")
	private static Object[] getxc7aParTestData() {
		//PAR tests
		InputStream parJconfigDocument = HardwareReportTester.class.getResourceAsStream(xc7aDocument);
		InputStreamReader parIsr = new InputStreamReader(HardwareReportTester.class.getResourceAsStream("/synthesisReports/xc7a.par"));
		FpgaResource bram16 = new FpgaResource(0 ,365, BlockRamPrimitives.RAMB36E1FIFO36E1);
		FpgaResource bram8 = new FpgaResource(20 ,730, BlockRamPrimitives.RAMB18E1FIFO18E1);
		FpgaResource registers = new FpgaResource(1033, 269200, NonChangingPrimitives.REGISTER);
		FpgaResource luts = new FpgaResource(2749, 134600, NonChangingPrimitives.LUT);
		FpgaResource slices = new FpgaResource(848, 33650, NonChangingPrimitives.SLICE);
		FpgaResource dsps = new FpgaResource(2, 740, DspPrimitives.DSP48E1);
		List<FpgaResource> fpgaResources = new ArrayList<>();
		fpgaResources.add(bram16);
		fpgaResources.add(bram8);
		fpgaResources.add(registers);
		fpgaResources.add(luts);
		fpgaResources.add(slices);
		fpgaResources.add(dsps);

		return new Object[] { parIsr, fpgaResources,  "xc7a par report parse test" , parJconfigDocument};
	}

	@SuppressWarnings("checkstyle:MagicNumber")
	private static Object[] getxc7aMapTestData() {
		//MAP tests
		InputStream mapJconfigDocument = HardwareReportTester.class.getResourceAsStream(xc7aDocument);
		InputStreamReader mapIsr = new InputStreamReader(HardwareReportTester.class.getResourceAsStream("/synthesisReports/xc7a.mrp"));
		FpgaResource bram16 = new FpgaResource(0 ,365, BlockRamPrimitives.RAMB36E1FIFO36E1);
		FpgaResource bram8 = new FpgaResource(20 ,730, BlockRamPrimitives.RAMB18E1FIFO18E1);
		FpgaResource registers = new FpgaResource(1033, 269200, NonChangingPrimitives.REGISTER);
		FpgaResource luts = new FpgaResource(2749, 134600, NonChangingPrimitives.LUT);
		FpgaResource slices = new FpgaResource(848, 33650, NonChangingPrimitives.SLICE);
		FpgaResource dsps = new FpgaResource(2, 740, DspPrimitives.DSP48E1);
		List<FpgaResource> fpgaResources = new ArrayList<>();
		fpgaResources.add(bram16);
		fpgaResources.add(bram8);
		fpgaResources.add(registers);
		fpgaResources.add(luts);
		fpgaResources.add(slices);
		fpgaResources.add(dsps);
		return new Object[] { mapIsr, fpgaResources, "xc7a map report parse test", mapJconfigDocument };
	}


	@Test
	public void test() throws IOException, IllegalArgumentException {

		Path jconfigDocumentTempPath = tempJC3File.getRoot().toPath().resolve("jconfig.jc3");

		final ILibJConfigProject libJconfigProject = LibJConfigProjectGenerator.getLibJconfigProjectFrom(jconfigDocument, jconfigDocumentTempPath);

		final Optional<FpgaFamily> familyIfSupported = FpgaFamily.getFamilyIfSupported(LibJConfigProjectUtils.getDeviceFrom(libJconfigProject));

		if (!familyIfSupported.isPresent()) {
			throw new IllegalArgumentException("The given jconfig file contains a not supported device");
		}

		HardwareReportParser parReport = new HardwareReportParser(inputReport, familyIfSupported.get());
		parReport.parse();
		List<FpgaResource> testResults = parReport.getResources();
		Collections.sort(testResults);
		Collections.sort(referenceResources);
		Assert.assertEquals(referenceResources, testResults);

		//Arguments testArguments = parReport.getArguments();
		//Assert.assertEquals(referenceArguments, testArguments);

	}

}
