/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.initram;

import de.tu_darmstadt.rs.jcommander.JCommanderRunner;
import de.tu_darmstadt.rs.jcommander.validators.FileExistsValidator;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBramBuilderFactory;
import de.tu_darmstadt.rs.spartanmc.initram.check.GlobalLayoutCheck;
import de.tu_darmstadt.rs.spartanmc.initram.output.HexOutput;
import de.tu_darmstadt.rs.spartanmc.initram.output.IOutputGen;
import de.tu_darmstadt.rs.spartanmc.initram.output.MemguardOutput;
import de.tu_darmstadt.rs.spartanmc.initram.output.SPHOutput;
import de.tu_darmstadt.rs.spartanmc.initram.output.lattice.LSEOutput;
import de.tu_darmstadt.rs.spartanmc.initram.output.lattice.VerilogOutputLattice;
import de.tu_darmstadt.rs.spartanmc.initram.output.quartus.MifFilesOutput;
import de.tu_darmstadt.rs.spartanmc.initram.output.quartus.VerilogOutputQuartus;
import de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.TCLOutput;
import de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.UCFOutput;
import de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.VerilogOutputXilinx;
import de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.bitfile.ISEBitfileReplace;
import de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.bitfile.VivadoBitfileReplace;
import de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.xdl.XDLReplace;
import de.tu_darmstadt.rs.spartanmc.initram.parameter.MemoryMapConverter;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryMap;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Path;
import java.util.HashSet;
import java.util.Set;

/**
 * Main entry point for Initram
 */
//Command line app, we are allowed to use System.out
@SuppressWarnings("checkstyle:SystemOut")
public class Initramj implements Runnable {
	private static final Logger logger = LoggerFactory.getLogger(Initramj.class);

	//Inputs
	@Parameter(names = {"--memory", "-m"}, converter = MemoryMapConverter.class, required = true, description = "Memory map")
	private MemoryMap memory;
	@Parameter(names = {"--xdlin", "-x"}, validateValueWith = FileExistsValidator.class, description = "ISE: XDL input")
	private Path xdlIn;

	@Parameter(names = {"--bitin"}, validateValueWith = FileExistsValidator.class, description = "ISE, Vivado: Bitfile input")
	private Path bitIn;
	@Parameter(names = {"--location"}, validateValueWith = FileExistsValidator.class, description = "ISE, Vivado: Logic location file")
	private Path location;

	@Parameter(names = {"--primLoc"}, validateValueWith = FileExistsValidator.class, description = "Vivado: Primitive Location File")
	private Path primLoc;

	//Filter
	@Parameter(names = {"--region", "-r"}, description = "Filter output for region")
	private String region;

	//Output
	@Parameter(names = {"--initData", "-s"}, description = "General: SPH output")
	private Path sphFile;
	@Parameter(names = {"--xdlout", "-y"}, description = "ISE: XDL output with replaced data. Needs xdlIn")
	private Path xdlOut;

	@Parameter(names = {"--bitout"}, description = "ISE, Vivado: Bitfile output with replaced data. For ISE, needs xdlIn, location and bitIn. For Vivado, needs primLoc, location and bitIn")
	private Path bitOut;

	@Parameter(names = {"--ucf", "-u"}, description = "ISE: UCF output")
	private Path ucfFile;

	/*
	@Parameter(names = {"--qsf","-q"}, description = "Altera Quartus: QSF output")
	private Path qsfFile;
	*/

	@Parameter(names = "--mif", description = "Altera Quartus: Mif files. Target directory")
	private Path mifTargetDir;

	//TODO what is this used for?
	/*
	@Parameter(names = {"--lpf","-l"}, description = "Lattice Diamond: LPF output")
	private Path lpfFile;
	 */

	@Parameter(names = {"--hex"}, description = "General: Hex memory file for volatile memories")
	private Path hexFile;

	@Parameter(names = {"--lse"}, description = "Lattice Diamond: output for LSE flow")
	private Path lseFile;

	@Parameter(names={"--xilinxverilog","-xv"}, description = "General: Verilog output for Xilinx")
	private Path xilinxVerilogFile;

	@Parameter(names={"--quartusverilog","-av"}, description = "General: Verilog output for Altera")
	private Path quartusVerilogFile;

	@Parameter(names={"--latticeverilog","-lv"}, description = "General: Verilog output for Lattice")
	private Path latticeVerilogFile;

	@Parameter(names = {"--tcl"}, description = "Vivado: TCL output")
	private Path tclFile;

	@Parameter(names = {"--memguard"}, description = "General: generate memguard initalization to filter access to sections")
	private Path memguardOut;

	public Initramj() {
	}

	/**
	 * Entry point
	 *
	 * @param args command line args
	 * @throws IOException On broken IO
	 */
	public static void main(String... args) throws IOException {
		JCommanderRunner.create(new Initramj())
				.run(args);
	}

	private void validateParameters() {
		if (sphFile != null && region == null) {
			throw new ParameterException("For SPH export, a memory region needs to be set");
		}

		if (bitOut != null) {
			if (location == null) {
				throw new ParameterException("Location is needed for bitOut");
			}
			if (bitIn == null) {
				throw new ParameterException("BitIn is needed for bitOut");
			}
			if (xdlIn != null && primLoc != null) {
				throw new ParameterException("Either set xdlIn or primLoc, not both");
			}
			if (xdlIn == null && primLoc == null) {
				throw new ParameterException("xdlIn or primLoc required for bitOut");
			}
		}

		if (xdlOut != null && xdlIn == null) {
			throw new ParameterException("If xdlOut is set, xdlIn also needs to be set.");
		}
	}

	/**
	 * Do the work
	 */
	public void run() {
		validateParameters();

		Set<IOutputGen> generators = new HashSet<>();

		if (sphFile != null) {
			generators.add(new SPHOutput(sphFile, region));
		}
		if (xdlOut != null) {
			generators.add(new XDLReplace(xdlIn, xdlOut));
		}

		if (ucfFile != null) {
			generators.add(new UCFOutput(ucfFile));
		}

		if (mifTargetDir!=null) {
			generators.add(new MifFilesOutput(mifTargetDir));
		}

		//TODO what is this used for?
		/*
		if (lpfFile!=null) {
			generators.add(new LPFOutput(lpfFile));
		}
		*/

		if (lseFile!=null) {
			generators.add(new LSEOutput(lseFile));
		}

		if (xilinxVerilogFile!=null) {
			generators.add(new VerilogOutputXilinx(xilinxVerilogFile));
		}

		if (quartusVerilogFile !=null) {
			generators.add(new VerilogOutputQuartus(quartusVerilogFile));
		}

		if (latticeVerilogFile!=null) {
			generators.add(new VerilogOutputLattice(latticeVerilogFile));
		}

		if (bitOut != null) {
			if (xdlIn != null) {
				generators.add(new ISEBitfileReplace(xdlIn, bitIn, bitOut, location));
			} else {
				generators.add(new VivadoBitfileReplace(primLoc, bitIn, bitOut, location));
			}
		}

		if (tclFile != null) {
			generators.add(new TCLOutput(tclFile));
		}

		if (memguardOut != null) {
			generators.add(new MemguardOutput(memguardOut));
		}

		if (hexFile != null) {
			generators.add(new HexOutput(hexFile));
		}

		Path projectDir = memory.getSourceFile().toAbsolutePath().getParent().getParent();

		IBramBuilderFactory bramBuilderFactory = IBramBuilderFactory.getImplementationFor(memory.getToolchain());
		logger.debug("Using {}, available bramConfigs: {}", bramBuilderFactory, bramBuilderFactory.getAvailableConfigs());
		MemorySource memorySource = new MemorySource(memory, projectDir, bramBuilderFactory);

		if (! new GlobalLayoutCheck(memorySource).check()) {
			System.err.println("The global memory layout contains errors.");
			System.exit(1);
		}

		for (IOutputGen generator : generators) {
			try {
				generator.output(memorySource);
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}

		if (generators.isEmpty()) {
			LoggerFactory.getLogger(Initramj.class).error("No output generator given, nothing to do.");
		}
	}
}
