/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.bramLayout;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryLocator;
import de.tu_darmstadt.rs.spartanmc.initram.exceptions.InitramJException;
import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection;
import de.tu_darmstadt.rs.util.BitUtil;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;
import java.util.function.IntFunction;


/**
 * This builder assumes that the number of initialization vectors required for a bram is equivalent to the number of lines of content.
 *
 * There is only 1 type of line and its width is equivalent to bramConfig.width
 *
 * @param <Config>
 * @param <Bram>
 * @param <BramLine>
 */
public abstract class BaseBramBuilder<Config extends IBramConfig, Bram extends GenericBram, BramLine extends GenericBram.GenericBramLine> implements IBramBuilder {

	private static final Logger logger = LoggerFactory.getLogger(BaseBramBuilder.class);

	@NotNull
	protected final MemoryLocator bramPattern;
	@NotNull
	protected final Integer bramColumnIndex;
	@NotNull
	protected final Config bramConfig;
	@Nullable
	protected final IntFunction<String> lineNameSupplier;

	protected List<BramLine> inits = new ArrayList<>();
	protected final List<Bram> brams = new ArrayList<>();
	private final int byteWidth;
	protected final int maxBlockCount;

	/**
	 *
	 *  @param bramPattern
	 * @param bramColumnIndex
	 * @param bramConfig
	 * @param lineNameSupplier
	 * @param maxBlockCount
	 */
	public BaseBramBuilder(@NotNull MemoryLocator bramPattern, Integer bramColumnIndex, Config bramConfig, @Nullable IntFunction<String> lineNameSupplier, int maxBlockCount) {
		this.bramPattern = bramPattern;
		this.bramColumnIndex = bramColumnIndex;
		this.bramConfig = bramConfig;
		this.lineNameSupplier = lineNameSupplier;
		byteWidth = BitUtil.ceilDiv(bramConfig.getDataWidth(), 8);
		this.maxBlockCount = maxBlockCount;
	}

	@Override
	public void starting(ISection section) {}

	protected int currentBramIdx = 0;

	protected int currentInitLine = 0;
	protected int initBits = 0;
	protected BigInteger currentInit = BigInteger.ZERO;

	@Override
	public void add(byte[] data) {


		if (data.length != byteWidth) {
			throw new IllegalArgumentException("Data must be "+ byteWidth +" bytes wide! Was: "+ data);
		}
		int width = bramConfig.getDataWidth();
		currentInit = new BigInteger(1, data);
		int initAdded = width;

		tryCommitFullLines(initAdded);

		tryCommitFullBram();
	}

	protected void tryCommitFullBram() {
		if (isBramFull()) {
			commitBram();
		}
	}

	protected boolean isBramFull() {
		return inits.size() == bramConfig.getDepth();
	}

	protected void tryCommitFullLines(int initAdded) {
		initBits += initAdded;

		if (isLineFull()) {
			commitInit();
		} else {
			currentInit = currentInit.shiftRight(initAdded);
		}
	}

	protected boolean isLineFull() {
		return initBits == bramConfig.getDataWidth();
	}

	protected void commitBram() {
		checkFullBram();
		String instanceName = bramPattern.resolve(currentBramIdx++, bramColumnIndex);
		currentInitLine = 0;

		brams.add(createNewBram(instanceName));
		inits = new ArrayList<>();
	}

	protected void checkFullBram() {
		if (inits.size() != bramConfig.getDepth()) {
			throw new InitramJException("Wrong amount of Bram Data lines: "+ inits.size() +", required: "+ bramConfig.getDepth());
		}
	}

	@NotNull
	protected abstract Bram createNewBram(String instanceName);

	protected abstract BramLine createNewLine(int lineIdx, BigInteger contents);

	protected void commitInit() {
		inits.add(createNewLine(currentInitLine++, currentInit));
		currentInit = BigInteger.ZERO;
		initBits = 0;
	}

	@Override
	public List<Bram> build() {
		if (!inits.isEmpty()) {
			commitInit();
			commitBram();
		}

		logger.debug("Bram pattern {}::{} was implemented with {} brams", bramPattern, bramColumnIndex, brams.size());

		return brams;
	}

	@Override
	public long getCurrentMaxAddressEx() {
		int incompleteCount = brams.size();
		if (hasUncommitedData()) {
			incompleteCount += 1;
		}
		return bramConfig.getDepth() * incompleteCount;
	}

	private boolean hasUncommitedData() {
		return !inits.isEmpty() || initBits != 0;
	}

	@Override
	public void padIfNeeded() {
		if (bramConfig.isPadToFullMemory()) {
			while (brams.size() < maxBlockCount) {
				addZeroBram();
			}
		}
	}

	protected abstract void addZeroBram();
}
