/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.bramLayout;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryLocator;
import de.tu_darmstadt.rs.spartanmc.initram.exceptions.InitramJException;
import de.tu_darmstadt.rs.spartanmc.initram.memory.extractors.IBlockLineExtractor;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.IMemoryBlock;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public abstract class BaseBramBuilderFactory<Config extends IBramConfig> implements IBramBuilderFactory {

	private static final Logger logger = LoggerFactory.getLogger(BaseBramBuilderFactory.class);

	protected Config findBramConfig(int bitsPerBlockLine) {
		return Arrays.stream(getAvailableConfigs()).filter(c -> c.getDataWidth() == bitsPerBlockLine).findFirst().orElseThrow(() -> new InitramJException(String.format("No Bram matches width %d: %s", bitsPerBlockLine, Arrays.toString(XilinxBramBuilderFactory.XBramConfig.values()))));
	}

	public abstract Config[] getAvailableConfigs();

	@Override
	public IBramBuilder getSplitBuilder(IMemoryBlock block, int bramColumnIndex) {
		return getBramBuilder(block, bramColumnIndex, getBramConfig(block));
	}

	@NotNull
	protected abstract IBramBuilder getBramBuilder(IMemoryBlock block, Integer bramColumnIndex, Config bramConfig);

	@Override
	public int getBramCountPerColumnFor(IMemoryBlock block) {
		Config bramConfig = getBramConfig(block);

		long length = block.getLength();
		int depth = bramConfig.getDepth();

		assert length % depth == 0;

		return Math.toIntExact(length / depth);
	}

	public Stream<String> streamAllBramInstanceNamesFor(IMemoryBlock block) {
		int cols = block.getColumnsPerMemoryLine();
		int bramCount = getBramCountPerColumnFor(block);
		MemoryLocator bramPattern = block.getBramPattern();

		return IntStream.range(0, bramCount)
				.boxed()
				.flatMap(bramIdx -> IntStream.range(0, cols)
						.mapToObj(colIdx -> bramPattern.resolve(bramIdx, colIdx))
				);
	}

	protected Config getBramConfig(IMemoryBlock block) {
		return findBramConfig(block.getBitsPerBlockLine());
	}

	@Override
	public IBramBuilder getBuilder(IMemoryBlock block) {
		int bitsPerBlockLine = block.getBitsPerBlockLine();
		Config bramConfig = findBramConfig(bitsPerBlockLine);
		logger.debug("For module '{}', chose {} bram configuration", block.getModuleName(), bramConfig);
		IBlockLineExtractor extractor = IBlockLineExtractor.getExtractor(block);
		if (block.isSplit()) {
			return new SplittingBramBuilder(block, this, extractor);
		} else {
			return new ExtractingBramBuilder(getBramBuilder(block, null, bramConfig), extractor);
		}
	}

	@Override
	public String toString() {
		return getClass().getSimpleName();
	}
}
