/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.bramLayout;

import de.tu_darmstadt.rs.spartanmc.initram.exceptions.InitramJException;
import de.tu_darmstadt.rs.spartanmc.initram.memory.extractors.IBlockLineExtractor;
import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.IMemoryBlock;

import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Will be utilized if memory lines need to be distributed across multiple brams. Data will be split according to big-endian: most-significant byte of memory line will be placed in first bram (A) and so on. If you'd like little-endianness, go connect the data-lines in reverse order, do not split the memory or add that here, preferably in the blockLineExtractor
 */
public class SplittingBramBuilder implements IBramBuilder {
	private static final Logger logger = LoggerFactory.getLogger(SplittingBramBuilder.class);

	private final int bitsPerMemoryLine;
	private final int bitsPerBlockLine;
	private final int totalBytesPerMemoryLine;
	private final int blockRamCount;
	private final IMemoryBlock block;
	private final IBlockLineExtractor blockLineExtractor;
	private final IBramBuilder[] blockRams;

	SplittingBramBuilder(IMemoryBlock block, IBramBuilderFactory builderFactory, IBlockLineExtractor blockLineExtractor) {
		this.bitsPerMemoryLine = block.getBitsPerMemoryLine();
		this.bitsPerBlockLine = block.getBitsPerBlockLine();
		this.block = block;
		this.blockLineExtractor = blockLineExtractor;

		assert bitsPerMemoryLine <= 32;

		this.totalBytesPerMemoryLine = block.getBytesPerMemoryLine();
		this.blockRamCount = block.getColumnsPerMemoryLine();
		this.blockRams = new IBramBuilder[blockRamCount];

		logger.debug("Splitting memory for '{}' into {} parts according to {} bitsPerMemoryLine, {} bitsPerBlockLine, expecting {} bytes per memory line, using {}", block.getModuleName(), blockRamCount, bitsPerMemoryLine, bitsPerBlockLine, totalBytesPerMemoryLine, blockLineExtractor);

		for (int i = 0; i < blockRams.length; i++) {
			blockRams[blockRamCount - 1 - i] = builderFactory.getSplitBuilder(block, i);
		}

		assert blockRamCount > 1;
	}

	@Nullable
	private ISection currentSection;

	@Override
	public void starting(ISection section) {
		currentSection = section;
	}

	@Override
	public void add(byte[] data) {
		if (data.length != totalBytesPerMemoryLine) {
			throw new InitramJException(String.format("Wrong number of bytes %d for Memory with %d bits per line", data.length, bitsPerMemoryLine));
		}

		byte[][] blockLine = blockLineExtractor.splitBlockLines(data, currentSection);
		for (int i = 0; i < blockRamCount; i++) {
			blockRams[i].add(blockLine[i]);
		}
	}

	@Override
	public List<? extends IBram> build() {
		return Arrays.stream(blockRams).flatMap(builder -> builder.build().stream()).collect(Collectors.toList());
	}

	@Override
	public long getCurrentMaxAddressEx() {
		return blockRams[0].getCurrentMaxAddressEx();
	}

	@Override
	public void padIfNeeded() {
		for (IBramBuilder blockRam : blockRams) {
			blockRam.padIfNeeded();
		}
	}
}