/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.bramLayout;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryLocator;
import de.tu_darmstadt.rs.spartanmc.initram.exceptions.InitramJException;
import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

import static de.tu_darmstadt.rs.spartanmc.initram.bramLayout.XilinxBramBuilderFactory.XBramConfig;

public class XilinxBramBuilder extends BaseBramBuilder<XBramConfig, XilinxBram, XilinxBram.XilinxBramLine> {

	private static final Logger logger = LoggerFactory.getLogger(XilinxBramBuilder.class);

	private List<XilinxBram.XilinxBramLine> initps = new ArrayList<>();

	public XilinxBramBuilder(@NotNull MemoryLocator bramPattern, Integer bramSuffix, @NotNull XBramConfig bramConfig) {
		super(bramPattern, bramSuffix, bramConfig, null, 0);
	}

	@Override
	public void starting(ISection section) {}


	private int currentInitpLine = 0;
	private int initpBits = 0;
	private BigInteger currentInitp = BigInteger.ZERO;

	private void commitInitp() {
		initps.add(new XilinxBram.XilinxBramLine(true, currentInitpLine++, currentInitp));
		currentInitp = BigInteger.ZERO;
		initpBits = 0;
	}

	@Override
	public void add(byte[] data) {
		int initAdded;
		int initpAdded;

		switch (bramConfig) {
			case XRam9x2048:
				currentInit = currentInit.or(new BigInteger(1, new byte[]{ data[1] }).shiftLeft(248));
				initAdded = 8;
				currentInitp = currentInitp.or(new BigInteger(1, new byte[]{ data[0] }).shiftLeft(255));
				initpAdded = 1;
				break;
			case XRam18x1024:
				currentInit = currentInit.or(new BigInteger(1, new byte[]{ data[1], data[2] }).shiftLeft(240));
				initAdded = 16;
				currentInitp = currentInitp.or(new BigInteger(1, new byte[]{ data[0] }).shiftLeft(254));
				initpAdded = 2;
				break;
			case XRam32x512:
				currentInit = currentInit.or(new BigInteger(1, data).shiftLeft(224));
				initAdded = 32;
				initpAdded = 0;
				break;
			default:
				throw new UnsupportedOperationException();
		}


		tryCommitFullLines(initAdded, initpAdded);

		tryCommitFullBram();
	}

	protected void tryCommitFullLines(int initAdded, int initpAdded) {
		super.tryCommitFullLines(initAdded);

		initpBits += initpAdded;

		if (initpBits == 256) {
			commitInitp();
		} else {
			currentInitp = currentInitp.shiftRight(initpAdded);
		}
	}

	@Override
	protected boolean isLineFull() {
		return initBits == 256;
	}

	@Override
	protected boolean isBramFull() {
		return inits.size() == bramConfig.getInitCount();
	}

	@Override
	protected void commitBram() {
		super.commitBram();
		currentInitpLine = 0;
		initps = new ArrayList<>();
	}

	@Override
	protected void checkFullBram() {
		if (inits.size() != bramConfig.getInitCount()) {
			throw new InitramJException("Wrong amount of Bram Data lines: "+ inits.size() +", required: "+ bramConfig.getInitCount());
		}
		if (initps.size() != bramConfig.getInitpCount()) {
			throw new InitramJException("Wrong amount of Bram Parity lines: "+ initps.size() +", required: "+ bramConfig.getInitpCount());
		}
	}

	@Override
	protected XilinxBram.XilinxBramLine createNewLine(int lineIdx, BigInteger contents) {
		return new XilinxBram.XilinxBramLine(false, lineIdx, contents);
	}

	@Override
	public List<XilinxBram> build() {
		if (!inits.isEmpty() || !initps.isEmpty()) {
			commitInit();
			commitInitp();
			commitBram();
		}

		logger.debug("Bram pattern {}::{} was implemented with {} brams", bramPattern, bramColumnIndex, brams.size());

		return brams;
	}

	@Override
	public void padIfNeeded() {
		if (bramConfig.isPadToFullMemory()) {
			throw new RuntimeException("not implemented, as the xilinxBramConfig will always return false!");
		}
	}

	@Override
	protected void addZeroBram() {
		throw new InitramJException("Not implemented");
	}

	@Override
	@NotNull
	protected XilinxBram createNewBram(String instanceName) {
		return new XilinxBram(instanceName, bramConfig, inits, initps);
	}

	@Override
	public String toString() {
		return getClass().getSimpleName()+"("+ bramPattern + ", "+ bramColumnIndex +")";
	}
}
