/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.check;

import de.tu_darmstadt.rs.spartanmc.initram.MemorySource;
import de.tu_darmstadt.rs.util.StreamUtil.Pair;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.nio.file.Path;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * This Class checks if all firmwares contain the same global variable layouts
 * the build process is interrupted if differences are detected
 *
 * Created by sherber on 04.10.16.
 */
public class GlobalLayoutCheck {

	private static final String SECTION = Pattern.quote(SectionLayout.LayoutGenerator.SECTION_NAME);

	private HashMap<String, Collection<SectionLayout>> layouts;
	private static final Logger logger = LoggerFactory.getLogger(GlobalLayoutCheck.class);


	public GlobalLayoutCheck(MemorySource memorySource) {
		generateGlobalLayoutData(memorySource);
	}


	/*
		reads an elf.map file linewise and returns the global layouts described therein
	 */
	private static Pair<String, SectionLayout> readInMap(File mapFile) throws IOException {
		BufferedReader reader = new BufferedReader(new FileReader(mapFile));

		String line = reader.readLine();
		String region;

		//while (!line.matches("\\.[^\\.\\s]+" + SECTION + ".*")) {
		while (!line.matches("\\s" + SECTION + ".*")) {
			line = reader.readLine();
			if (line == null) {
				return null;
			}

		}

		Pattern pattern = Pattern.compile("\\.([^\\.\\s]+)\\..*");
		Matcher matcher = pattern.matcher(line);

		if (!matcher.matches()) {
			region = "placeholder";
		} else {
			region = matcher.group(1);
		}


		SectionLayout l = new SectionLayout(mapFile);

		while (SectionLayout.LayoutGenerator.parseLine(line, l)) {

			line = reader.readLine();
		}
		reader.close();
		return new Pair<>(region, l);
	}

	/**
	 * Generate layouts data for a whole system
	 *
	 * @param memorySource entry point for memory
	 */
	private void generateGlobalLayoutData(MemorySource memorySource) {
		Map<Path, SectionLayout> firmwares = new HashMap<>();
		layouts = new HashMap<>();

		memorySource.streamMappedBlocks().forEach(block -> {

			Path mapFile = block.resolveFirmwareFile().resolveSibling("main.elf.map");

			SectionLayout layout = firmwares.get(mapFile);
			if (layout == null) {
				Pair<String, SectionLayout> map = null;
				try {
					map = readInMap(mapFile.toFile());
				} catch (IOException e) {
					logger.info("No ELF map found. Skipping global variables layout check.");
					return;
				}

				if (map != null) {
					firmwares.put(mapFile, map.getB());

					Collection<SectionLayout> ls = layouts.get(map.getA());
					if (ls != null) {
						ls.add(map.getB());
					} else {
						List<SectionLayout> list = new LinkedList<>();
						list.add(map.getB());
						layouts.put(map.getA(), list);
					}
				}
			}
		});

	}

	/**
	 * do a pairwise comparison of all Layouts
	 *
	 * @return true if all layouts are equal
	 */
	public boolean check() {
		logger.info("Found " + layouts.size() + " global layouts");
		if (layouts.size() == 0) {
			return true;
		}

		return layouts.keySet().stream().allMatch(globalmem -> {
			logger.info("Comparing " + layouts.get(globalmem).size() + " implementations...");
			Iterator<SectionLayout> it = layouts.get(globalmem).iterator();
			if (!it.hasNext()) {
				return true;
			}
			SectionLayout l1 = it.next();
			logger.info("Section size contains " + l1.getNumberOfVariables() + " variables in " + l1.getLength() + " Bytes.");
			while (it.hasNext()) {
				SectionLayout l2 = it.next();
				if (!l1.equals(l2)) {
					logger.error("The global memory layouts of " + l1.getSource() + " and " + l2.getSource() + " differ.");
					return false;
				}
				l1 = l2;
			}
			return true;
		});
	}
}
