/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.check;

import java.io.File;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by sherber on 04.10.16.
 */
public class SectionLayout {
	private long length;
	private File source;
	private HashMap<Long, String> map;

	public SectionLayout(File source) {
		this.source = source;
		map = new HashMap<>();
	}

	public void add(Long pos, String varname) {
		map.put(pos, varname);
	}

	public void setLength(long length) {
		this.length = length;
	}

	public long getLength() {
		return length;
	}

	public int getNumberOfVariables() {
		return map.size();
	}

	/**
		for equallity we check size and compare variable names and their position
	 */
	@Override
	public boolean equals(Object o) {
		if (o == this) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		SectionLayout other = (SectionLayout) o;

		if (other.length != this.length) {
			return false;
		}
		if (other.map.size() != map.size()) {
			return false;
		}
		for (Long l: map.keySet()) {
			if (!map.get(l).equals(other.map.get(l))) {
				return false;
			}
		}
		return true;

	}

	public File getSource() {
		return source;
	}

	public String toString() {
		return "[Size: " + length + ", " + map.toString() + "]";
	}

	/*
		helper class to extract lines from elf.map file and parse the data
		uses regular expressions :/
	 */
	public static class LayoutGenerator {
		public static final String SECTION_NAME = ".globaldata";

		private static final String HEX_NUM_GROUP = "0x([0-9a-f]+)";
		private static final String HEX_NUM = "0x[0-9a-f]+";
		private static final String HEX_NUM16_GROUP = "0x([0-9a-f]{16})";
		private static final String HEX_NUM16 = "0x[0-9a-f]{16}";
		private static final String WS = "[ \t]";
		private static final String WS_OPT = WS + "*";
		private static final String ANY = ".*";
		private static final String SECTION = Pattern.quote(SECTION_NAME);

		/**
		 * parses a line from the MemoryLayout file and places the content in SectionLayout
		 * @param line the line to parse
		 * @param l the sectionlayout this line belongs to
		 * @return true iff line could be parsed, false otherwise
		 */
		public static boolean parseLine(String line, SectionLayout l) {
			if (line == null) {
				return false;
			}
			//matches i.e.
			//                 0x000000000004001c                mutex
			if (line.matches(WS_OPT + HEX_NUM16 + ANY)) {
				return handleVariable(line, l);
			}
			//matches i.e.
			// .globaldata     0x0000000000040000       0x34
			//but also
			//  .globaldata    0x0000000000040000       0x34 /foo/bar/main.o
			if (line.matches(WS_OPT + SECTION + WS_OPT + HEX_NUM16 + WS_OPT + HEX_NUM + WS_OPT + ANY)) {
				return handleSource(line, l);
			}
			// matches i.e.
			// *(.globaldata)
			if (line.matches(WS_OPT + Pattern.quote("*(" + SECTION_NAME + ")") + WS_OPT)) {
				return true;
			}
			return false;
		}

		/**
			extracts from the string the variable name and its position in address space
			and adds it to the layout
			@return true on success
		 */
		private static boolean handleVariable(String line, SectionLayout l) {
			Pattern pattern = Pattern.compile(WS_OPT + HEX_NUM16_GROUP + WS_OPT + "(.*)");
			Matcher matcher = pattern.matcher(line);

			if (!matcher.matches()) {
				return false;
			}
			String number = matcher.group(1);
			String varname = matcher.group(2);

			long address = Long.parseLong(number, 16);
			l.add(address, varname);
			return true;
		}

		/**
		 extracts from the string the length of the global section, ignores file lines
		 and adds it to the layout
		 @return true on success
		 */
		private static boolean handleSource(String line, SectionLayout l) {

			Pattern pattern = Pattern.compile(WS_OPT + SECTION + WS_OPT + HEX_NUM16 + WS_OPT + HEX_NUM_GROUP + ANY);
			Matcher matcher = pattern.matcher(line);

			if (!matcher.matches()) {
				//ignore
				return true;
			}
			String length = matcher.group(1);

			long sectionLength = Long.parseLong(length, 16);
			l.setLength(sectionLength);
			return true;
		}
	}
}
