/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.memory;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryLocator;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBram;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBramBuilder;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBramBuilderFactory;
import de.tu_darmstadt.rs.spartanmc.initram.exceptions.MemoryLayoutException;
import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemFirmware;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryBlock;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryLineLayout;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.List;
import java.util.stream.Stream;

/**
 * MemoryBlock and Memory at the same time.
 * Addresses are in target words, but the sections give access to byte addresses
 * Reading of firmware is lazy.
 */
public class MappedBlock implements IMappedBlock {

	private static final Logger logger = LoggerFactory.getLogger(MappedBlock.class);

	private final MemoryBlock block;
	private final IBramBuilderFactory builderFactory;
	private final FirmwareCache firmwareCache;

	public MappedBlock(MemoryBlock block, FirmwareCache firmwareCache, IBramBuilderFactory builderFactory) {
		this.block = block;
		this.builderFactory = builderFactory;
		this.firmwareCache = firmwareCache;
	}

	@Override
	public Path resolveFirmwareFile() {
		return firmwareCache.resolveElfFile(block.getFirmware());
	}

	@Override
	public List<? extends IBram> generateBrams() {
		if (!block.isBitfileInitializable()) {
			throw new IllegalStateException("Cannot create blockrams for a non bitfile initializable memory block: "+block.getModuleName());
		}
		IBramBuilder builder = builderFactory.getBuilder(block);

		int bytesPerLine = block.getBytesPerMemoryLine();
		long start = block.getStart();
		long length = block.getLength();
		long endEx = start + length;
		byte[] buffer = new byte[bytesPerLine];

		logger.debug("Generating Brams for '{}' starting at target address {}, expecting {} lines of {} bytes", block.getModuleName(), start, length, bytesPerLine);

		ISectionedMemory sections = getContents();
		logger.debug("The following sections will be written to block {}: {}", block.getModuleName(), sections.getSections());
		long byteStart = sections.getStart();
		long firmwareEndEx = byteStart + sections.getLength();
		ISection currentSection = null;

		long lineAddr = start, byteAddr = byteStart, contentLines = 0, fillerLines = 0, alignmentLines = 0;
		while (byteAddr < firmwareEndEx) {
			if (lineAddr >= endEx) {
				throw new MemoryLayoutException("Contents of firmware "+ sections.getSections() +" for "+ block + " do not fit!");
			}
			ISection nextSection = sections.getSectionOf(byteAddr);
			if (nextSection != currentSection) {
				if (currentSection != null && !currentSection.exists() && nextSection != null && nextSection.exists()) {
					contentLines += fillerLines;
					alignmentLines += fillerLines;
					logger.trace("Output padded with {} empty lines ahead of {}", fillerLines, nextSection);
					fillerLines = 0;
				}

				currentSection = nextSection;
				builder.starting(currentSection);
			}

			sections.alignedRead(buffer, byteAddr);
			builder.add(buffer);

			lineAddr++;
			byteAddr += bytesPerLine;

			if (currentSection.exists()) {
				contentLines++;
			} else {
				fillerLines++;
			}
		}

		float percentFull = contentLines * 100 / (float) length;

		long bramEndEx = builder.getCurrentMaxAddressEx() + start;

		long unreachable = bramEndEx - endEx;

		logger.info("Memory of '{}': {} % full ({} / {} lines, incl. {} alignment, {} lines not reachable) with firmware, filling rest with 0", block.getModuleName(), percentFull, contentLines, length, alignmentLines, unreachable);

		for (int i = 0; i < buffer.length; i++) {
			buffer[i] = 0;
		}

		while (lineAddr < bramEndEx) {
			builder.add(buffer);
			lineAddr++;
		}

		builder.padIfNeeded();

		return builder.build();
	}

	@Override
	public Stream<String> streamAllBramNames() {
		return builderFactory.streamAllBramInstanceNamesFor(this);
	}

	@Override
	@NotNull
	public MemoryLocator getBramPattern() {
		return block.getBramPattern();
	}

	@Override
	public long getStart() {
		return block.getStart();
	}

	@Override
	public long getLength() {
		return block.getLength();
	}

//	@Override
//	public byte get(long addr) {
//		return firmwareCache.get(block).get(addr);
//	}
//
//	@Override
//	public void read(byte[] buffer, long addr, int bufferOffset, int size) {
//		firmwareCache.get(block).read(buffer, addr, bufferOffset, size);
//	}
//
//	@Override
//	public void read(byte[] buffer, long addr) {
//		firmwareCache.get(block).read(buffer, addr);
//	}
//
//	@Override
//	public void alignedRead(byte[] buffer, long addr) {
//		firmwareCache.get(block).alignedRead(buffer, addr);
//	}

	private ISection getSectionOf(long addr) {
		return firmwareCache.get(block).getSectionOf(addr);
	}

	@Override
	public boolean isBitfileInitializable() {
		return block.isBitfileInitializable();
	}

	@Override
	@NotNull
	public String getModuleName() {
		return block.getModuleName();
	}

	@Override
	public int getBitsPerBlockLine() {
		return block.getBitsPerBlockLine();
	}

	@Override
	public int getBitsPerMemoryLine() {
		return block.getBitsPerMemoryLine();
	}

	@Override
	@NotNull
	public MemFirmware getFirmware() {
		return block.getFirmware();
	}

	@Override
	public ISectionedMemory getFirmwareAsSectionedMemory() {
		return firmwareCache.get(block);
	}

	@Override
	public ISectionedMemory getContents() {
		SectionedMemory wholeFirmware = firmwareCache.get(block);
		int bytesPerMemoryLine = block.getBytesPerMemoryLine();
		return wholeFirmware.subRange(block.getStart() * bytesPerMemoryLine, block.getLength() * bytesPerMemoryLine);
	}

	@Override
	public @NotNull
	MemoryLineLayout getMemoryLineLayout() {
		return block.getMemoryLineLayout();
	}

	@Override
	public String toString() {
		return "MappedBlock{" +
				"block=" + block.getModuleName() +
				", builderFactory=" + builderFactory +
				'}';
	}
}
