/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.initram.memory;

import de.tu_darmstadt.rs.spartanmc.initram.exceptions.MemoryLayoutException;
import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.FillerSection;
import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection;

import com.sun.istack.Nullable;

import java.util.Collection;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeMap;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * Container for firmware images. Should be agnostic of what type of container (ELF) was used.
 * Only contains sections that should be placed in the targets embedded firmware.
 * Does not reorder data from the image, section contents should still be equivalent to ELF contents
 *
 * This is essentially an in memory copy. It should be possible to use this class for simulations, with enough context to sanity check memory accesses.
 * This data should not be preprocessed (e.g. dropping bits for 18-bit layout, because it would prevent us from keeping debug infos.
 *
 * All addresses are in bytes.
 *///TODO could be File-Backed, only loading sections when required!
public class SectionedMemory implements ISectionedMemory {

	private final TreeMap<Long, ISection> sections;
	private final long length;
	private final long start;

	public SectionedMemory(TreeMap<Long, ISection> sections) {
		this.sections = sections;
		start = sections.firstKey();
		Map.Entry<Long, ISection> entry = sections.lastEntry();
		length =  entry.getKey() + entry.getValue().getSize() - start;
	}

	public SectionedMemory(long start, long length) {
		this.start = start;
		this.length = length;

		FillerSection fillerSection = new FillerSection(start, Math.toIntExact(length));
		sections = new TreeMap<>();
		sections.put(start, fillerSection);
	}

	@Override
	public long getStart() {
		return start;
	}
	@Override
	public long getLength() {
		return length;
	}

	@Override
	public byte get(long addr) {
		Map.Entry<Long, ISection> entry = sections.floorEntry(addr);
		int offset = Math.toIntExact(addr - entry.getKey());
		return entry.getValue().get(offset);
	}

	@Override
	public void read(final byte[] buffer, final long addr, final int bufferOffset, final int size) {
		long currentAddr = addr;
		int currentBufferOffset = bufferOffset;
		final long exclEndAddr = addr + size;

		do {
			Map.Entry<Long, ISection> entry = sections.floorEntry(currentAddr);
			long sectionStart = entry.getKey();

			int offset = Math.toIntExact(addr - sectionStart);
			ISection section = entry.getValue();
			int readableSize = Math.toIntExact(section.getSize() + sectionStart - offset);

			section.read(buffer, offset, currentBufferOffset, Math.min(readableSize, size));

			currentAddr += readableSize;
			currentBufferOffset += readableSize;
		} while (currentAddr < exclEndAddr);
	}

	@Override
	public void read(byte[] buffer, long addr) {
		read(buffer, addr, 0, buffer.length);
	}

	@Override
	public void alignedRead(byte[] buffer, long addr) {
		Map.Entry<Long, ISection> entry = sections.floorEntry(addr);
		long sectionStart = entry.getKey();
		int offset = Math.toIntExact(addr - sectionStart);
		entry.getValue().read(buffer, offset, 0, buffer.length);
	}

	@Override
	public Collection<ISection> getSections() {
		return sections.values();
	}

	@Override
	@Nullable
	public ISection getSectionOf(long addr) {
		return sections.floorEntry(addr).getValue();
	}

	public ISectionedMemory subRange(long start, long length) {
		TreeMap<Long, ISection> range = new TreeMap<>();
		Map.Entry<Long, ISection> current = sections.floorEntry(start);
		final long rangeEndEx = start + length;

		while (current != null && current.getKey() >= start && current.getKey() + current.getValue().getSize() <= rangeEndEx) {
			range.put(current.getKey(), current.getValue());
			current = sections.higherEntry(current.getKey());
		}

		if (range.size() > 0) {
			return new SectionedMemory(range);
		} else {
			return new SectionedMemory(start, length);
		}
	}


	public static class Builder implements ISectionedMemoryBuilder {

		private final LinkedList<ISection> sections = new LinkedList<>();

		private long nextFreeAddress = 0L;

		@Override
		public void addSection(ISection section) {
			long nextStartAddress = section.getStartAddress();
			if (nextStartAddress < nextFreeAddress) {
				ISection last = sections.getLast();
				throw new MemoryLayoutException("Section "+ last +" and "+ section +" overlap!");
			} else if (nextStartAddress > nextFreeAddress) {
				ISection last = sections.getLast();
				throw new MemoryLayoutException("Non contigous memory in between "+ last + " and "+ section +"!");
			}

			sections.add(section);
			nextFreeAddress += section.getSize();
		}

		@Override
		public SectionedMemory build() {
			TreeMap<Long, ISection> sections = this.sections.stream().collect(Collectors.toMap(ISection::getStartAddress, Function.identity(), Builder::preventMerge, TreeMap<Long, ISection>::new));

			return new SectionedMemory(sections);
		}

		private static ISection preventMerge(ISection a, ISection b) {
			throw new IllegalStateException("Overlapping sections: "+ a +", "+ b);
		}

		/**
		 * Fill the memory with undefined data until (excl.) addr
		 * @param addr address after the filler
		 */
		@Override
		public void fillUntil(long addr) {
			if (addr == nextFreeAddress) {
				return;
			} else if (addr < nextFreeAddress) {
				throw new MemoryLayoutException("Cannot fill until "+ addr +" , already upto "+ (nextFreeAddress-1));
			}

			int size = Math.toIntExact(addr - nextFreeAddress - 1);
			sections.add(new FillerSection(nextFreeAddress, size));
			nextFreeAddress = addr;
		}
	}
}
