/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.initram.memory.elf;

import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection;
import de.tu_darmstadt.rs.spartanmc.initram.memory.SectionedMemory;
import de.tu_darmstadt.rs.spartanmc.initram.exceptions.InitramJException;

import com.oracle.max.elf.ELFHeader;
import com.oracle.max.elf.ELFLoader;
import com.oracle.max.elf.ELFSectionHeaderTable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.RandomAccessFile;
import java.nio.file.Path;
import java.util.Comparator;
import java.util.stream.Stream;

@SuppressWarnings("checkstyle:MagicNumber")
public class ElfReader {

	private static final Logger logger = LoggerFactory.getLogger(ElfReader.class);


	public ElfReader() {

	}

	/**
	 * Read an elf file
	 *
	 * @param elfFile the file
	 * @return file content with sectionInfo
	 *
	 * @throws InitramJException on invalid input ELF File
	 */
	public static SectionedMemory read(Path elfFile) {
		logger.info("Loading {}", elfFile);
		// read ELF File
		try {
			final RandomAccessFile randomAccessFile = new RandomAccessFile(elfFile.toFile(), "r");
			// read the ELF header
			final ELFHeader header = ELFLoader.readELFHeader(randomAccessFile);
			// read the section header table
			final ELFSectionHeaderTable sht = ELFLoader.readSHT(randomAccessFile, header);

			SectionedMemory.Builder memBuilder = new SectionedMemory.Builder();

			long currentAddress = 0;

			Stream.of(sht.entries)
					.sorted(Comparator.comparingLong(ELFSectionHeaderTable.Entry::getAddr))
					.filter(BaseElfSection::isAllocatableSection)
					.forEachOrdered(entry -> {
						addSectionToMemory(memBuilder, entry, randomAccessFile);
			});

			return memBuilder.build();

		} catch (Exception e) {
			throw new InitramJException("Failed to read "+ elfFile, e);
		}
	}

	private static void addSectionToMemory(SectionedMemory.Builder memBuilder, ELFSectionHeaderTable.Entry entry, RandomAccessFile randomAccessFile) {
		long addr = entry.getAddr();
		long size = entry.getSize();
		memBuilder.fillUntil(addr);

		logger.debug("adding section: '{}' begin: {} size: {} end: {}", entry.getName(), addr, size, addr + size);

		ISection section;
		if (BaseElfSection.isNoBitsSection(entry)) {
			section = new EmptyElfSection(entry);
		} else {
			section = ElfSection.from(entry, randomAccessFile);
		}
		memBuilder.addSection(section);
	}
}
