/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.memory.elf;

import com.oracle.max.elf.ELFSectionHeaderTable;
import com.oracle.max.elf.exceptions.ElfSectionException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.RandomAccessFile;

public class ElfSection extends BaseElfSection {

	private static final Logger logger = LoggerFactory.getLogger(ElfSection.class);

	private final byte[] contents;

	public ElfSection(ELFSectionHeaderTable.Entry sectionHeader, byte[] contents) {
		super(sectionHeader);
		this.contents = contents;
	}

	public static ElfSection from(ELFSectionHeaderTable.Entry entry, RandomAccessFile randomAccessFile) {
		long fileOffset = entry.getOffset();
		try {
			randomAccessFile.seek(fileOffset);
			byte[] contents = new byte[Math.toIntExact(entry.getSize())];
			randomAccessFile.read(contents);
			return new ElfSection(entry, contents);
		} catch (IOException e) {
			logger.error("Failed to seek to start of section contents for {} @ {}", entry, fileOffset, e);
			throw new ElfSectionException(entry, "Section claims data to be at "+ fileOffset +" which is not accessible!");
		}
	}

	@Override
	public byte get(int offset) {
		return contents[offset];
	}

	@Override
	public void set(int offset, byte data) {
		contents[offset] = data;
	}

	@Override
	public void read(byte[] buffer, int offset, int bufferOffset, int readSize) {
		boolean pad;
		int actualReadSize;
		int maxReadSize = getSize() - offset;
		if (readSize > maxReadSize) {
			pad = true;
			actualReadSize = maxReadSize;
		} else {
			pad = false;
			actualReadSize = readSize;
		}
		System.arraycopy(contents, offset, buffer, bufferOffset, actualReadSize);
		if (pad) {
			logger.debug("Reading {} at {} of {}, padding to {} with 0", actualReadSize, offset, this, readSize);
			for (int i = actualReadSize; i < readSize; i++) {
				buffer[bufferOffset + i] = 0;
			}
		}
	}

	@Override
	public String toString() {
		return "ElfSection('"+ getName() +"', start: "+ getStartAddress() +", size: "+ getSize() +")";
	}
}
