/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.memory.extractors;

import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;

public class SpartancMcExtractor {

	private static final Logger logger = LoggerFactory.getLogger(SpartanMc18BlockLineExtractor.class);
	private final InvalidDataHandling errorHandlingMode;

	public SpartancMcExtractor(InvalidDataHandling errorHandlingMode) {
		this.errorHandlingMode = errorHandlingMode;
	}

	public SpartancMcExtractor() {
		this.errorHandlingMode = InvalidDataHandling.Log;
	}

	public static boolean containsInvalid18Bits(byte[] buffer) {
		return buffer[0] != 0 || (buffer[1] & ~0x3) != 0;
	}

	public static boolean containsInvalid9x2Bits(byte[] buffer) {
		return (buffer[0] & ~0x1) != 0 || (buffer[2] & ~0x1) != 0;
	}

	public static int extractSpartan18Bits(byte[] data) {
		return (((data[2] & 0xFF) << 16) | ((data[1] & 0xFF) << 8) | (data[0] & 0xFF)) & 0x3FFFF;
	}

	public static int extractSpartan9x2Bits(byte[] data) {
		int lowerAddrByte = (((data[0] & 0x1) << 8) & 0x100) | (data[1] & 0xFF);
		int upperAddrByte = (((data[2] & 0x1) << 8) & 0x100) | (data[3] & 0xFF);

		return (lowerAddrByte << 9) | upperAddrByte;
	}

	public void checkOrFix9x2MemoryLine(byte[] data, ISection section) {
		boolean invalidSpartanBits = containsInvalid9x2Bits(data);
		if (invalidSpartanBits) {
			String msg = String.format("section %s contains invalid bits for a SpartanMC 9x2 section: 0x%X. Allowed Mask: 0x01FF01FF", section.toString(), new BigInteger(1, data));
			if (errorHandlingMode == InvalidDataHandling.Throw) {
				throw new IllegalArgumentException(msg);
			} else {
				logger.error(msg);
			}
			data[0] &= 0x1;
			data[2] &= 0x1;
		}
	}

	public void checkOrFix18MemoryLine(byte[] data, ISection section) {
		boolean invalidSpartanBits = containsInvalid18Bits(data);

		if (invalidSpartanBits) {
			String msg = String.format("section %s contains invalid bits for a SpartanMC 18 section: 0x%X. Allowed Mask: 0x3FFFF", section.toString(), new BigInteger(1, data));
			if (errorHandlingMode == InvalidDataHandling.Throw) {
				throw new IllegalArgumentException();
			} else {
				logger.error(msg);
			}
			data[0] = 0x0;
			data[1] &= 0x3;
		}
	}

	public byte[][] split18BlockLine(byte[] data, ISection section) {
		checkOrFix18MemoryLine(data, section);

		byte[] moreSignificant = new byte[2];
		byte[] lessSignificant = new byte[2];

		lessSignificant[1] = data[3];
		lessSignificant[0] = (byte) (data[2] & 0x1);

		moreSignificant[1] = (byte) (((data[2] >> 1) & 0x7F) | (data[1] << 7));
		moreSignificant[0] = (byte) ((data[1] >> 1) & 0x1);

		return new byte[][] { moreSignificant, lessSignificant };
	}

	public byte[][] split9x2BlockLines(byte[] data, ISection section) {
		checkOrFix9x2MemoryLine(data, section);

		return new byte[][] { { data[0], data[1] }, { data[2], data[3] } };
	}

	public byte[] extract18BlockLineFromCode(byte[] data, ISection section) {
		checkOrFix18MemoryLine(data, section);

		return data;
	}

	public byte[] extract9x2BlockLineFromData(byte[] data, ISection section) {
		checkOrFix9x2MemoryLine(data, section);

		return new byte[] { (byte)(((data[0] & 0xFF) << 1) | ((data[1] & 0xFF) >>> 7)), (byte)(((data[1] & 0xFF) << 1) | (data[2] & 0xFF)), data[3] };
	}

	public enum InvalidDataHandling {Throw, Log}
}
