/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.output;

import de.tu_darmstadt.rs.spartanmc.initram.MemorySource;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBram;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBramLine;
import de.tu_darmstadt.rs.spartanmc.initram.memory.IMappedBlock;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.IMemoryBlock;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.function.Function;
import java.util.stream.Stream;

/**
 * Output rams into individual files
 */
public abstract class BramIndividualTextOutput implements IOutputGen {

	private static final Logger logger = LoggerFactory.getLogger(BramSharedTextOutput.class);

	protected final Path outputDir;

	protected BramIndividualTextOutput(Path outputDir) {

		this.outputDir = outputDir;
	}

	protected abstract String getHeader(IBram bram);

	protected abstract String getFooter(IBram bram);

	protected abstract String formatBramLine(String bramInstanceName, IBramLine line);

	protected abstract String getBramFilename(String instanceName);

	public Path getBramPath(String instanceName) {
		return outputDir.resolve(instanceName);
	}

	protected Stream<? extends IBram> configureStream(Stream<? extends IMappedBlock> blockStream) {
		return blockStream.filter(IMemoryBlock::isBitfileInitializable).flatMap(block -> block.generateBrams().stream());
	}

	@Override
	public void output(MemorySource memorySource) throws IOException {
		final Stream<IBram> stream = configureStream(memorySource.streamMappedBlocks()).map(Function.identity());
		final Iterable<IBram> iterable = stream::iterator;
		for (IBram bram : iterable) {
			outputBram(bram);
		}
	}


	protected void outputBram(IBram bram) throws IOException {
		Path outputFile = outputDir.resolve(getBramFilename(bram.getInstanceName()));
		try (PrintStream out = new PrintStream(Files.newOutputStream(outputFile))) {
			out.println(getHeader(bram));
			bram.streamLines().forEach(val -> { // no sorting neccessary. Use of LinkedHashMap guarantees order
				out.println(formatBramLine(bram.getInstanceName(), val));
			});
			out.println(getFooter(bram));
		}
	}

	public Stream<Path> streamAllBramFilenames(MemorySource memorySource) {
		return memorySource.streamAllInitializableBramNames()
				.map(this::getBramFilename)
				.map(outputDir::resolve);
	}

}
