/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.output;

import de.tu_darmstadt.rs.spartanmc.initram.MemorySource;
import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection;
import de.tu_darmstadt.rs.util.StreamUtil;

import com.oracle.max.elf.ELFSectionHeaderTable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Output memguard initialization function that protects the sections as specified in the elf file
 */
public class MemguardOutput implements IOutputGen {

	/**
	 * Types of sections
	 */
	private enum SectionType {
		Code("MEMGUARD_ALLOW_CODE"),
		Data("MEMGUARD_ALLOW_RD | MEMGUARD_ALLOW_WR"),
		Readonly("MEMGUARD_ALLOW_RD");

		private final String code;

		SectionType(String code) {
			this.code = code;
		}

		public static SectionType fromElfFlags(int flags) {
			if ((flags & ELFSectionHeaderTable.SHF_EXECINSTR) != 0) {
				return Code;
			} else {
				if ((flags & ELFSectionHeaderTable.SHF_WRITE) != 0) {
					return Data;
				} else {
					return Readonly;
				}

			}
		}

		public static SectionType fromSection(ISection section) {
			if (section.isCode())
				return Code;
			else if (section.isWritable())
				return Data;
			else
				return Readonly;
		}

		public String toCode() {
			return code;
		}
	}


	private static String sanitizeVar(String var) {
		return "_"+var.replaceAll("\\.","_");
	}

	private static class MemguardWatcher {

		final String name;
		final String startVar;
		final List<String> lengthVars;
		final SectionType sectionType;

		private MemguardWatcher(String name, String startVar, List<String> lengthVars, SectionType sectionType) {
			this.name = name;
			this.startVar = startVar;
			this.lengthVars = Collections.unmodifiableList(lengthVars);
			this.sectionType = sectionType;
		}

		public static MemguardWatcher fromElfEntry(ISection section) {
			final String secName = section.getName().toUpperCase().replaceAll("[^A-Za-z0-9]", "_");
			return new MemguardWatcher(
					section.getName(),
					".startof."+secName,
					Collections.singletonList(".sizeof." + secName),
					SectionType.fromSection(section)
			);
		}

		public Optional<MemguardWatcher> tryAppend(MemguardWatcher watcher) {
			if (sectionType != watcher.sectionType) {
				return Optional.empty();
			}

			List<String> allLengthVars = new ArrayList<>(lengthVars);
			allLengthVars.addAll(watcher.lengthVars);

			return Optional.of(new MemguardWatcher(
					name + ", " + watcher.name,
					startVar,
					allLengthVars,
					sectionType
			));
		}

		@Override
		public String toString() {
			return "MemguardWatcher{" +
					"name='" + name + '\'' +
					", startVar='" + startVar + '\'' +
					", lengthVars='" + lengthVars + '\'' +
					", sectionType=" + sectionType +
					'}';
		}

		public String toCode(int i) {
			String codeMultiplier = sectionType == SectionType.Code? "CODE_TO_DATA" : "";

			String length = lengthVars.stream().map(MemguardOutput::sanitizeVar).map(v->"((unsigned int)&"+v+")").collect(Collectors.joining("+"));
			length = "("+length+")/2";

			return "\t//Sections " + name + "\n" +
					"\tmemguard_configure_watcher_if_avail(memguard, " + i + ", numWatchers, "
					+ codeMultiplier
					+ "(&" + sanitizeVar(startVar) + "), " + length + ", " + sectionType.toCode() + ");";
		}
	}


	private static final Logger logger = LoggerFactory.getLogger(MemguardOutput.class);

	private final Path memguardOut;

	public MemguardOutput(Path memguardOut) {
		this.memguardOut = memguardOut;
	}

	@Override
	public void output(MemorySource memorySource) throws IOException {
		try (PrintWriter writer = new PrintWriter(Files.newBufferedWriter(memguardOut))) {
			memorySource.streamMappedBlocks().forEach(mappedBlock -> {
				writer.println("// For " + mappedBlock.getFirmware().getName() + ":");


				final List<MemguardWatcher> watchers = mappedBlock.getContents().getSections().stream()
						.map(MemguardWatcher::fromElfEntry)
						.collect(ArrayList::new, this::entryAccumulator, (x, y) -> {
							throw new UnsupportedOperationException();
						});

				watchers.stream()
						.flatMap(w -> Stream.concat(Stream.of(w.startVar), w.lengthVars.stream()))
						.distinct()
						.forEach(v -> writer.format("extern int %s asm(\"%s\");%n", sanitizeVar(v), v));

				writer.println("void memguard_configure_default(memguard_regs_t *memguard, unsigned int numWatchers) {");
				writer.println("\t//Generated by InitramJ from ELF file");
				watchers.stream()
						.map(StreamUtil.numberElements())
						.forEach(StreamUtil.IntPair.accessC((i, w) -> writer.println(w.toCode(i))));
				writer.println("}");

			});
		}
	}

	private void entryAccumulator(List<MemguardWatcher> r, MemguardWatcher watcher) {
		if (r.isEmpty()) {
			r.add(watcher);
			return;
		}

		//Try to unify last two entries
		final MemguardWatcher lastItem = r.get(r.size() - 1);
		Optional<MemguardWatcher> appended = lastItem.tryAppend(watcher);
		if (appended.isPresent()) {
			r.remove(r.size() - 1);
			r.add(appended.get());
		} else {
			r.add(watcher);
		}

	}
}
