/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.bitfile;

import de.tu_darmstadt.rs.spartanmc.initram.MemorySource;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBramLine;
import de.tu_darmstadt.rs.spartanmc.initram.output.IOutputGen;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryMap;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Function;
import java.util.function.LongSupplier;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Replace content of a bitfile
 */
//So much bit shifting
@SuppressWarnings("checkstyle:MagicNumber")
public abstract class BitfileReplace implements IOutputGen {

	private static final String RAMB16 = "/RAMB16";
	private static final String PART_NOT_SUPPORTED_FOR_REPLACE_RAM = "Part not supported for Replace RAM: ";
	private static final Logger logger = LoggerFactory.getLogger(BitfileReplace.class);
	private static final int CHUNK_SIZE = 2048;
	private static final String OUR_CRC_VALUE = "\t Our crc Value: {}";
	private static Pattern locationPattern = Pattern.compile("Bit\\s+(\\d+) 0x([\\da-f]+)\\s+(\\d+) Block=(\\w+) Ram=(\\w):(\\w+)");
	private final Path bitIn;
	private final Path bitOut;
	private final Path location;

	/**
	 * Replace ram in a bitfile
	 *
	 * @param bitIn input Bitfile
	 * @param bitOut output bitfile
	 * @param location input location file
	 */
	public BitfileReplace(Path bitIn, Path bitOut, Path location) {
		super();

		this.bitIn = bitIn;
		this.bitOut = bitOut;
		this.location = location;
	}

	private static long readWord(int wordSize, InputStream is, OutputStream os, boolean writeToOs) throws IOException {
		long res = 0;
		for (int i = 0; i < wordSize; i++) {
			int current = is.read();
			if (current == -1) {
				return -1;
			}
			if (writeToOs) {
				os.write(current);
			}

			res = res << 8 | current;
		}
		return res;
	}

	/**
	 * Find the syncronization packet
	 * @param in input stream
	 * @param out output stream
	 * @throws IOException on broken IO
	 */
	public static void findSynchronization(InputStream in, OutputStream out) throws IOException {

		final int[] sequence = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xAA, 0x99, 0x55, 0x66};

		int inByte;
		int seqIdx = 0;
		while ((inByte = in.read()) != -1) {
			out.write(inByte);
			//Match at current position?
			if (inByte == sequence[seqIdx]) {
				seqIdx++;
				if (seqIdx == sequence.length) {
					break;
				}
			} else {
				//check for match at start
				//We only need to check for first byte because our sequence does not have any repetitions
				if (inByte == sequence[0]) {
					seqIdx = 1;
				} else {
					seqIdx = 0;
				}
			}
		}
	}

	private static int changeAddressToByteOrder(int offset, BitfilePacketDecoder decoder) {
		return offset ^ decoder.getAddressXorMask();
	}

	/**
	 * Parse the logic location file
	 * @param location the file
	 * @param usedSites list of used sites. Only used sites will be included in the result
	 * @param decoder bitfile decoder (needed to generate the right byte order)
	 * @return stream of PrimSizeBits
	 * @throws IOException on broken IO
	 */
	//Regex groups need numbers
	@SuppressWarnings("checkstyle:MagicNumber")
	public static Stream<PrimSiteBit> parseLocationFile(Path location, Collection<String> usedSites, BitfilePacketDecoder decoder) throws IOException {
		Set<Integer> seenBits = new HashSet<>();
		return Files.lines(location).flatMap(line -> {
			Matcher matcher = locationPattern.matcher(line);
			if (matcher.matches()) {
				int offset = changeAddressToByteOrder(Integer.parseInt(matcher.group(1)), decoder);
				String address = matcher.group(2);
				String frameOffset = matcher.group(3);
				String primitiveSite = matcher.group(4);
				String ramId = matcher.group(5);
				String bit = matcher.group(6);

				if (!usedSites.contains(primitiveSite)) {
					return Stream.empty();
				}

				if (!primitiveSite.startsWith("RAM")) {
					return Stream.empty();
				}

				if (!"B".equals(ramId)) {
					throw new RuntimeException("Ram ID is " + ramId + " in " + line);
				}

				if (!seenBits.add(offset)) {
					throw new RuntimeException("Duplicate offset: " + offset);
				}

				return Stream.of(new PrimSiteBit(offset, primitiveSite, bit));
			} else {
				return Stream.empty();
			}
		}).sorted(Comparator.comparingInt(psb -> psb.bitOffset));
	}

	private BitfilePacketDecoder createHeaderDecoder(MemoryMap memory) {
		String partName = Objects.requireNonNull(memory.getPartName()).toUpperCase();
		if (!partName.startsWith("XC")) {
			throw new RuntimeException(PART_NOT_SUPPORTED_FOR_REPLACE_RAM + partName);
		}

		String id = partName.substring(2, 4);
		switch (id) {
			case "6S":
				return new Sp6BitfilePacketDecoder();
			case "7A":
			case "7V":
			case "7K":
				return new X7BitfilePacketDecoder();
			default:
				throw new RuntimeException(PART_NOT_SUPPORTED_FOR_REPLACE_RAM + partName + " (id: " + id + ")");
		}
	}

	private void replaceBitfile(int byteLength, InputStream in, OutputStream out, Map<String, Map<String, IBramLine>> primitiveSiteInitData, Iterator<PrimSiteBit> primSiteBits, BitfilePacketDecoder decoder, BitfileChecksum crc, int bytesPerWord, int register) throws IOException {

		final int[] totalCount = {0};
		final int[] differenceCount0to1 = {0};
		final int[] differenceCount1to0 = {0};

		if (!primSiteBits.hasNext()) {
			copyUntil(0, byteLength, in, out, crc, bytesPerWord, register, (offset, actualLength, data) -> {

			});
			return;
		}

		final PrimSiteBit[] nextReplace = {primSiteBits.next()};


		copyUntil(0, byteLength, in, out, crc, bytesPerWord, register, (offset, actualLength, data) -> {
			while (nextReplace[0] != null) {
				int offsetWithinArray = nextReplace[0].bitOffset / 8 - offset;
				//Is the replacer in the next array?
				if (offsetWithinArray >= actualLength) {
					break;
				}


				Map<String, IBramLine> instData = primitiveSiteInitData.get(nextReplace[0].primitiveSite);
				if (instData != null) {

					byte input = data[offsetWithinArray];

					int offs = 7 - (nextReplace[0].bitOffset % 8);
					int refBit = (input >> offs) & 1;

					int bit = extractBit(nextReplace[0].primitiveSite, instData, nextReplace[0].bit, decoder);

					if (bit != refBit) {
						if (refBit == 1) {
							differenceCount1to0[0]++;
						} else {
							differenceCount0to1[0]++;
						}
					}


					totalCount[0]++;

					if (bit != 0) {
						input |= (1 << offs);
					} else {
						input &= ~(1 << offs);
					}

					data[offsetWithinArray] = input;

				} else {
					throw new RuntimeException("No data found for " + nextReplace[0].primitiveSite);
				}


				if (primSiteBits.hasNext()) {
					nextReplace[0] = primSiteBits.next();
				} else {
					nextReplace[0] = null;
				}

			}
		});

		if (nextReplace[0] != null) {
			throw new RuntimeException("Not all bits used! Next replacer: " + nextReplace[0] + ".");
		}
		float diffPercent0to1 = 100F * differenceCount0to1[0] / totalCount[0];
		float diffPercent1to0 = 100F * differenceCount1to0[0] / totalCount[0];
		logger.info(differenceCount0to1[0] + "/" + totalCount[0] + " program bits changed from 0 to 1. " + diffPercent0to1 + "%");
		logger.info(differenceCount1to0[0] + "/" + totalCount[0] + " program bits changed from 1 to 0. " + diffPercent1to0 + "%");
	}

	/**
	 * Copy rest of streams
	 *
	 * @param in input stream
	 * @param out output stream
	 * @throws IOException on broken IO
	 */
	private void copyRest(InputStream in, OutputStream out) throws IOException {
		copyUntil(0, null, in, out, null, 0, 0, null);
	}

	/**
	 * Directly copy parts of the Path in larger chunks for speedup
	 *
	 * @param byteOffset current offset
	 * @param byteLimit limit for copying or null if we should copy until the end of the file
	 * @param in input stream
	 * @param out output stream
	 * @param crc crc generator
	 * @param replacer modifier for the data
	 * @return offset after copying
	 */
	private int copyUntil(int byteOffset, Integer byteLimit, InputStream in, OutputStream out, BitfileChecksum crc, int bytesPerWord, int register, ReplaceData replacer) throws IOException {
		byte[] buffer = null;
		while (byteLimit == null || byteOffset < byteLimit) {
			int copySize;
			if (byteLimit != null) {
				int remaining = byteLimit - byteOffset;
				copySize = Math.min(remaining, CHUNK_SIZE);
			} else {
				copySize = CHUNK_SIZE;
			}

			if (buffer == null || copySize != buffer.length) {
				buffer = new byte[copySize];
			}

			int actual = in.read(buffer);

			if (actual == -1) {
				if (byteLimit != null) {
					throw new RuntimeException("Not enough data.");
				} else {
					return byteOffset;
				}
			}

			if (replacer != null) {
				replacer.replace(byteOffset, actual, buffer);
			}

			out.write(buffer, 0, actual);

			if (crc != null) {
				int wordOffset = 0;
				long word = 0;
				for (int i = 0; i < actual; i++) {

					long current = buffer[i];
					//Remove sign extension...
					current &= 0xFF;
					word = word << 8 | current;
					if (wordOffset == bytesPerWord - 1) {
						crc.updateCrc(word, register);
						word = 0;
						wordOffset = 0;
					} else {
						wordOffset++;
					}
				}
			}

			byteOffset += actual;
		}

		return byteOffset;
	}

	/**
	 * Read the packets
	 * @param crc the crc
	 * @param decoder the packet decoder
	 * @param primitiveSiteInitData new ram content
	 * @param primSiteBits bits representing bram data
	 * @throws IOException on broken IO
	 */
	public void readPackets(BitfileChecksum crc, BitfilePacketDecoder decoder, Map<String, Map<String, IBramLine>> primitiveSiteInitData, Iterator<PrimSiteBit> primSiteBits) throws IOException {
		try (InputStream in = Files.newInputStream(bitIn); OutputStream out = Files.newOutputStream(bitOut)) {
			findSynchronization(in, out);

			int wordSize = decoder.getWordSize();

			LongSupplier supplier = () -> {
				try {
					long l = readWord(wordSize, in, out, true);
					return l;
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
			};

			int nopCount = 0;
			while (in.available() > 0) {
				PacketData packet;
				packet = decoder.analyzePacket(supplier);

				if (packet.isConfigPacket) {
					logger.debug(packet.opCode + " " + packet.registerName + " configuration " + packet.wordCount + " {");
					replaceBitfile((int) packet.wordCount * wordSize, in, out, primitiveSiteInitData, primSiteBits, decoder, crc, wordSize, packet.registerAddress);
					//copyUntil(0, (int) ((packet.wordCount)*wordSize),in, out, crc, wordSize, packet.registerAddress);
					//					for (long i = 0; i < packet.wordCount; i++) {
					//						long x = supplier.getAsLong();
					//						crc.updateCrc(x,packet.registerAddress, false);
					//					}
					logger.debug("}");
				} else {
					if (packet.opCode == BitfilePacketDecoder.OpCode.NOP) {
						nopCount++;
					} else {
						if (nopCount > 0) {
							logger.debug("NOP * {}", nopCount);
							nopCount = 0;
						}

						logger.debug(packet.opCode + " " + packet.registerName + "(" + packet.wordCount + ") { ");
						if (packet.isCrcWrite) {
							int crcVal = crc.computeCRCRegValue();
							logger.debug(OUR_CRC_VALUE, Long.toHexString(crcVal));

							writePacketData(out, wordSize, packet.wordCount, crcVal);

							crc.reset();
						}

						for (int i = 0; i < packet.wordCount; i++) {
							long data = readWord(wordSize, in, out, !packet.isCrcWrite);
							logger.debug("\t" + Long.toHexString(data));

							if (!packet.isCrcWrite) {
								if (packet.opCode == BitfilePacketDecoder.OpCode.Write && !packet.isExcludedFromCrc) {
									crc.updateCrc(data, packet.registerAddress);
								} else {
									logger.info("no crc for this packet");
								}

								if (decoder.isCrcReset(packet, data)) {
									logger.debug("\tCRC RESET!");
									crc.reset();
								}
							}
						}

						logger.debug("}");
					}


				}
				if (packet.isAutoCrc) {
					long oldCrc = readWord(wordSize, in, out, false);
					oldCrc = oldCrc << 16 | readWord(wordSize, in, out, false);

					int crcVal = crc.getValue();
					writePacketData(out, wordSize, 2, crcVal);

					logger.debug(OUR_CRC_VALUE, Long.toHexString(crcVal));
					logger.debug("Auto CRC " + Long.toHexString(oldCrc));
				}
			}
		}
	}

	private void writePacketData(OutputStream out, int wordSize, long wordCount, int data) throws IOException {
		if (wordCount * wordSize > 4) {
			throw new RuntimeException("Too large");
		}

		int dataShift = (int) (8 * (wordSize * wordCount - 1));

		for (int i = 0; i < wordSize * wordCount; i++) {
			int b = (data >> dataShift) & 0xFF;
			out.write(b);

			dataShift -= 8;
		}
	}

	private int extractBitFromLine(IBramLine line, int bit) {
		return line.getContents().testBit(bit)? 1 : 0;
	}

	private int extractBit(String inst, Map<String, IBramLine> instData, String bit, BitfilePacketDecoder decoder) {
		boolean isParity = bit.startsWith("PARBIT");
		String offsetStr;
		if (isParity) {
			offsetStr = bit.substring(6);
		} else {
			offsetStr = bit.substring(3);
		}

		//For some reason, on 7 series FPGAs, bit offsets are *2... Correct that...
		int offset = Integer.parseInt(offsetStr) / decoder.getOffsetDivider();

		int line = offset / 256;


		String hexLine = Integer.toHexString(line).toUpperCase();
		if (hexLine.length() < 2) {
			hexLine = "0" + hexLine;
		}

		String lineName = (isParity? "INITP_" : "INIT_") + hexLine;
		IBramLine data = instData.get(lineName);

		int lineOffset = offset % 256;

		if (data == null) {
			throw new RuntimeException("Data " + lineName + " not found. from offset " + offset + " in " + inst);
		}

		return extractBitFromLine(data, lineOffset);
	}

	@SuppressWarnings("checkstyle:SystemOut")
	@Override
	public void output(MemorySource memorySource) throws IOException {
		Tracer.start();

		BitfilePacketDecoder decoder = createHeaderDecoder(memorySource.getMemoryMap());

		BitfileChecksum crc = decoder.createCrc();

		Tracer.enter("Load Primitive Locations");
		Map<String, String> locationMap = loadLocationMap();
		Tracer.exit();

		Tracer.enter("Reference ELF Data to placement");
		Map<String, Map<String, IBramLine>> primitiveSiteInitData = new HashMap<>();
		memorySource.streamAllBRams().forEach(d -> {
			String instance = d.getInstanceName();
			String location = locationMap.get(instance);
			if (location == null && !instance.endsWith(RAMB16)) {
				String instance2 = instance + RAMB16;
				location = locationMap.get(instance2);
				locationMap.put(instance, location);
			}

			if (location == null) {
				logger.warn("location map is {");
				locationMap.keySet().stream().sorted().forEach((a)->logger.warn("{} -> {}",a,locationMap.get(a)));
				logger.warn("}");

				throw new RuntimeException("Did not find location for " + instance);
			} else {
				primitiveSiteInitData.put(location, d.streamLines().collect(Collectors.toMap(IBramLine::getName, Function.identity())));
			}
		});
		Tracer.exit();

		Tracer.enter("Parse Location file");
		Stream<PrimSiteBit> primSiteBitStream = parseLocationFile(location, primitiveSiteInitData.keySet(), decoder);
		Tracer.exit();

		Tracer.enter("Replace Data in Bitfile");
		readPackets(crc, decoder, primitiveSiteInitData, primSiteBitStream.iterator());
		Tracer.exit();

		Tracer.printStats(System.out);

	}


	protected abstract Map<String,String> loadLocationMap() throws IOException;

	@FunctionalInterface
	interface ReplaceData {

		void replace(int offset, int actualLength, byte[] data);
	}
}
