/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.bitfile;

/**
 * Created by wenzel on 28.11.16.
 */
public class Crc implements BitfileChecksum {

	private final int crcSize;
	/** Value of the CRC */
	private int crcValue;

	/** Generator Polynomial */
	private final int polynomial;

	/**
	 * Constructor, simply sets crcValue to zero.
	 * @param crcSize size of the crc
	 * @param polynomial generator polynomial
	 */
	public Crc(int crcSize, int polynomial) {
		final int maxCrcSize = 32;
		if (crcSize> maxCrcSize || crcSize<1) {
			throw new IllegalArgumentException("Only Crc sizes between 1 and 32 supported!");
		}
		this.crcSize = crcSize;
		this.polynomial = polynomial;
		crcValue = 0;
	}

	@Override
	public void updateCrc(long content, int register) {
		for (int i = 0; i < crcSize; i++) {
			shiftInSingleBit((int) (content >> i));
		}
		final int contentSize = 5;
		for (int i = 0; i < contentSize; i++) {
			shiftInSingleBit(register >> i);
		}
	}

	@Override
	public void reset() {
		crcValue = 0;
	}


	/**
	 * Shifts in one bit from i and updates the crcValue
	 * @param i Input data
	 */
	private void shiftInSingleBit(int i) {
		int val = ((crcValue>>(crcSize-1)) ^ i) & 0x00000001;
		crcValue <<= 1;
		if (val != 0) {
			crcValue ^= polynomial;
		}
	}

	/**
	 * Gets the current CRC and returns it.
	 * @return The current crcValue
	 */
	@Override
	public int getValue() {
		return crcValue;
	}

	/**
	 * The method creates the CRC Register value need to be placed in the CRC register for
	 * a CRC check to pass.  Essentially, this reverses the 32 bits of the CRC.  Any time
	 * a CRC write is performed, this function should be called.
	 * @return The value to be placed in the CRC Register for the CRC check to pass
	 */
	@Override
	public int computeCRCRegValue() {
		int crcRegValue = 0x00000000;
		int mask = 0x00000001;
		for (int i = 0; i < crcSize; i++) {
			crcRegValue |= ((crcValue >> i) & mask) << (crcSize - 1 - i);
		}
		return crcRegValue;
	}

	public int getCrcSize() {
		return crcSize;
	}
}
