/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.bitfile;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.LongSupplier;

/**
 * Decode Spartan 6 Bistream Packets
 */
//So much bit shifting
@SuppressWarnings("checkstyle:MagicNumber")
public class Sp6BitfilePacketDecoder extends BitfilePacketDecoder {

	private static final Logger logger = LoggerFactory.getLogger(Sp6BitfilePacketDecoder.class);
	private static final String RDBK_SIGN = "RDBK_SIGN";
	private static final String LOUT = "LOUT";
	private static final String CSBO = "CSBO";
	private static final String FDRO = "FDRO";
	private static final String STAT = "STAT";
	private static final String CRC = "CRC";
	private static final String CMD = "CMD";
	private static final String BOOSTS = "BOOSTS";
	private static final String[] registers = {
			CRC,
			"FAR_MAJ",
			"FAR_MIN",
			"FDRI",
			FDRO,
			CMD,
			"CTL",
			"MASK",
			STAT,
			LOUT,
			"COR1",
			"COR2",
			"PWRDN_REG",
			"FLR",
			"IDCODE",
			"CWDT",
			"HC_OPT_REG",
			"!!INVALID",
			CSBO,
			"GENERAL1",
			"GENERAL2",
			"GENERAL3",
			"GENERAL4",
			"GENERAL5",
			"MODE_REG",
			"PU_GWE",
			"PU_GTS",
			"MFWR",
			"CCLK_FREQ",
			"SEU_OPT",
			"EXP_SIGN",
			RDBK_SIGN,
			"BOOTSTS",
			"EYE_MASK",
			"CBC_REG"
	};

	public Sp6BitfilePacketDecoder() {
	}

	@Override
	public int getAddressXorMask() {
		return 0;
	}

	@Override
	public int getOffsetDivider() {
		return 1;
	}

	@Override
	public boolean isCrcReset(PacketData packet, long data) {
		return packet.wordCount == 1 && packet.registerName.equals(CMD) && data == 7;
	}

	@Override
	public BitfileChecksum createCrc() {
		//		return new Crc(16,0x8005);
		return new Spartan6Crc();
	}

	@Override
	protected int getWordSize() {
		return 2;
	}

	@Override
	protected PacketData analyzePacket(LongSupplier wordSupplier) {

		long word = wordSupplier.getAsLong();
		//logger.debug("input word: {}",Long.toHexString(word));


		int frameType = (int) (word >> 13);
		OpCode operation = OpCode.decode((word >> 11) & 0x3);
		int rawRegister = (int) ((word >> 5) & 0x3F);
		String register = registers[rawRegister];
		long wordCount = (word) & 0x01F;

		boolean isConfigPacket = false;

		boolean isCrcWrite;
		if (frameType == 0x2) {
			wordCount = wordSupplier.getAsLong();
			wordCount = wordCount << 16 | wordSupplier.getAsLong();

			//Configuration data is expressed as writes to FDRI (coded as 3)
			if (operation == OpCode.Write && rawRegister == 3) {
				isConfigPacket = true;
			} else { //We don't know any other stuff
				throw new RuntimeException("Unsupported type 2 frame!");
			}
			isCrcWrite = false;
		} else {
			isCrcWrite = operation == OpCode.Write && register.equals(CRC);
		}

		boolean isExcludedFromCrc = (register.equals(BOOSTS) || register.equals(STAT)
				|| register.equals(LOUT) || register.equals(CSBO)
				|| register.equals(RDBK_SIGN) || register.equals(FDRO));

		return new PacketData(
				FrameType.fromInt(frameType),
				operation,
				rawRegister,
				register,
				wordCount,
				isConfigPacket, isCrcWrite, frameType == 0x2, isExcludedFromCrc);
	}
}
