/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.bitfile;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.LongSupplier;

/**
 * Decode 7 Series Bitfile packets
 */
//So much bit shifting
@SuppressWarnings("checkstyle:MagicNumber")
public class X7BitfilePacketDecoder extends BitfilePacketDecoder {

	private static final Logger logger = LoggerFactory.getLogger(X7BitfilePacketDecoder.class);

	private static final String INVALID = "!!!INVALID";
	private static final String CRC = "CRC";
	private static final String[] registers = {
			CRC,			//00000
			"FAR",			//00001
			"FDRI",			//00010
			"FDRO",			//00011
			"CMD",			//00100
			"CTL0",			//00101
			"MASK",			//00110
			"STAT",			//00111
			"LOUT",			//01000
			"COR0",			//01001
			"MFWR",			//01010
			"CBC",			//01011
			"IDCODE",		//01100
			"AXSS",			//01101
			"COR1",			//01110
			INVALID,		//01111
			"WBSTAR",		//10000
			"TIMER",		//10001
			"BSPI_READ",	//10010
			"FALL_EDGE",	//10011
			INVALID,		//10100
			INVALID,		//10101
			"BOOTSTS",		//10110
			INVALID,		//10111
			"CTL1",			//11000
			INVALID,		//11001
			INVALID,		//11010
			INVALID,		//11011
			INVALID,		//11100
			INVALID,		//11101
			INVALID,		//11110
			"BSPI"			//11111
	};

	private int regAddress;

	public X7BitfilePacketDecoder() {
	}

	@Override
	public int getAddressXorMask() {
		//Invert both the address bits within the byte, as well as the lower two byte address bits
		// -> 3 within byte, 2 byte addres --> 5 bits --> 2^5-1
		return 31;
	}

	@Override
	public int getOffsetDivider() {
		return 2;
	}

	@Override
	public boolean isCrcReset(PacketData packet, long data) {
		return packet.registerName.equals("CMD") && data==7;
	}

	@Override
	public BitfileChecksum createCrc() {
		return new Crc(32, 0x1EDC6F41);
	}

	@Override
	protected int getWordSize() {
		return 4;
	}


	@Override
	protected PacketData analyzePacket(LongSupplier wordSupplier) {
		long word = wordSupplier.getAsLong();

		FrameType frameType = FrameType.fromInt((int) (word >> 29));
		OpCode operation = OpCode.decode((word >> 27) & 0x3);
		int rawRegister = (int) ((word >> 13) & 0x3FFF);

		String register;
		long wordCount;
		boolean configDataFollows;
		boolean isCrcWrite;
		if (frameType==FrameType.TYPE1) {
			wordCount = (word ) & 0x3FF;
			configDataFollows = false;
		} else {
			rawRegister = this.regAddress;
			wordCount = word & 0x3FFFFFF;
			configDataFollows = true;
		}
		register = registers[rawRegister];
		isCrcWrite = operation==OpCode.Write && register.equals(CRC);

		this.regAddress = rawRegister;

		return new PacketData(
				frameType,
				operation,
				rawRegister,
				register,
				wordCount,
				configDataFollows,
				isCrcWrite, false, false);
	}
}
