/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.xdl;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Parse an XDL file
 */
public abstract class XDLParser<UserDataT, InstanceUserDataT, OutputT> {

	private static final Pattern instPattern = Pattern.compile("(inst \"([^\"]+)\" \"[^\"]+\",placed (\\w+) (\\w+)\\s+, cfg \")([^\"]*)(\" ;\\s*)?");
	private static final Pattern attrPattern = Pattern.compile("([^\"]*)(\" ;\\s*)?");
	private final Path inFile;

	public XDLParser(Path inFile) {

		this.inFile = inFile;
	}

	protected abstract String replaceAttribute(String attribute, InstanceUserDataT replacements, String bramName);

	protected abstract void output(OutputT outputData, String out);

	protected void outputLn(OutputT outputData, String out) {
		output(outputData, out + "\n");
	}

	private String handleAttributes(String input, InstanceUserDataT instanceUserData, String bramName) {
		int currentIndex = 0;
		StringBuilder output = new StringBuilder();

		while (currentIndex < input.length()) {
			if (input.charAt(currentIndex) == ' ') {
				output.append(' ');
				currentIndex++;
			} else {
				int nextSpace = input.indexOf(' ', currentIndex);
				if (nextSpace == -1) {
					nextSpace = input.length();
				}
				String attribute = input.substring(currentIndex, nextSpace);
				currentIndex = nextSpace;

				output.append(replaceAttribute(attribute, instanceUserData, bramName));
			}
		}
		return output.toString();
	}

	//Regex group numbers
	@SuppressWarnings("checkstyle:magicNumber")
	protected void parse(UserDataT userData, OutputT outputData) throws IOException {
		ParserState parserState = ParserState.IDLE;
		InstanceUserDataT currentInstanceUserData = null;
		String currentInstName = null;
		try (BufferedReader in = Files.newBufferedReader(inFile)) {
			String line;
			while ((line = in.readLine()) != null) {
				switch (parserState) {
					case IDLE: {
						Matcher m = instPattern.matcher(line);
						if (m.matches()) {
							String beginning = m.group(1);
							String instName = m.group(2);
							String tile = m.group(3);
							String primitiveSite = m.group(4);
							String attributes = m.group(5);
							String end = m.group(6);

							output(outputData, beginning);

							currentInstanceUserData = getUserData(userData, instName, tile, primitiveSite);

							currentInstName = instName;

							output(outputData, handleAttributes(attributes, currentInstanceUserData, instName));


							if (end != null) {
								outputLn(outputData, end);
							} else {
								outputLn(outputData, "");
								parserState = ParserState.IN_CONFIG;
							}
						} else {
							outputLn(outputData, line);
						}
						break;
					}
					case IN_CONFIG: {
						Matcher m = attrPattern.matcher(line);
						if (m.matches()) {

							String attributes = m.group(1);
							String end = m.group(2);

							output(outputData, handleAttributes(attributes, currentInstanceUserData, currentInstName));

							if (end != null) {
								outputLn(outputData, end);
								parserState = ParserState.IDLE;
								currentInstanceUserData = null;
								currentInstName = null;
							} else {
								outputLn(outputData, "");
							}

						} else {
							if (line.isEmpty()) {
								outputLn(outputData, "");
							} else {
								throw new RuntimeException("invalid input in parser state IN_CONFIG: " + line);
							}
						}


						break;
					}
					default:
						break;
				}
			}
		}

	}

	protected abstract InstanceUserDataT getUserData(UserDataT userData, String instName, String tile, String primitiveSite);

	enum ParserState {
		IDLE,
		IN_CONFIG
	}

}
