/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.initram.output.xilinx.xdl;

import de.tu_darmstadt.rs.spartanmc.initram.MemorySource;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBram;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBramLine;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.XilinxBram;
import de.tu_darmstadt.rs.spartanmc.initram.output.IOutputGen;

import java.io.IOException;
import java.io.PrintStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * Replace XDL Ram data
 */
public class XDLReplace extends XDLParser<Map<String, XilinxBram>, XilinxBram, PrintStream> implements IOutputGen {

	private static final String PRE_EMPHASIS = "PRE_EMPHASIS";
	private static final String RAMB16 = "/RAMB16";
	private final Path inFile;
	private final Path outFile;
	private Map<String, Set<IBramLine>> usedData;

	/**
	 * Replace Ram in an XDL file
	 * @param inFile XDL input
	 * @param outFile XDL output
	 */
	public XDLReplace(Path inFile, Path outFile) {
		super(inFile);
		this.inFile = inFile;
		this.outFile = outFile;
	}

	private Map<IBram, Map<String, IBramLine>> mappedLines = new HashMap<>();

	private IBramLine lookupLine(IBram bram, String name) {
		Map<String, IBramLine> map = mappedLines.computeIfAbsent(bram, this::mapBramLines);
		return map.get(name);
	}

	private Map<String, IBramLine> mapBramLines(IBram bram) {
		return bram.streamLines().collect(Collectors.toMap(IBramLine::getName, Function.identity()));
	}

	@Override
	protected String replaceAttribute(String attribute, XilinxBram replacements, String bramName) {
		int colonPos = attribute.indexOf(':');
		String attrName = attribute.substring(0, colonPos);

		//Work around XDL Bug in back conversion.
		//All Inputs are somehow considered as outputs if this attribute is not removed.
		if (attrName.equals(PRE_EMPHASIS)) {
			return "";
		}

		if (replacements == null) {
			return attribute;
		}

		IBramLine replacement = lookupLine(replacements, attrName);


		if (replacement != null) {
			Set<IBramLine> usedBram = usedData.computeIfAbsent(bramName, (x) -> new HashSet<>());
			if (!usedBram.add(replacement)) {
				throw new RuntimeException("used bram init data " + usedData + " twice");
			}

			return replacement.toString();
		} else {
			return attribute;
		}
	}

	@Override
	protected XilinxBram getUserData(Map<String, XilinxBram> data, String instName, String tile, String primitiveSite) {
		XilinxBram currentInitData = data.get(instName);

		//On Gen 7 FPGAS, "/RAMB16" seems to be added to inst name
		if (currentInitData == null && instName.endsWith(RAMB16)) {
			final int suffixLength = 7;
			instName = instName.substring(0, instName.length() - suffixLength);
			return data.get(instName);
		}
		return currentInitData;
	}


	@Override
	public void output(MemorySource memorySource) throws IOException {
		Map<String, XilinxBram> data = memorySource.streamAllBRams().map(bram -> (XilinxBram) bram).collect(Collectors.toMap(IBram::getInstanceName, Function.identity()));
		replaceData(data);
	}

	@Override
	protected void output(PrintStream ps,String out) {
		ps.print(out);
	}

	/**
	 * Replace the data
	 * @param data init data
	 * @throws IOException on broken IO
	 */
    public void replaceData(Map<String, XilinxBram> data) throws IOException {
		usedData = new HashMap<>();

		try (PrintStream out = new PrintStream(Files.newOutputStream(outFile))) {
			parse(data,out);
		}

		data.keySet().forEach(bram -> {
			Set<IBramLine> usedBrams = usedData.get(bram);
			if (usedBrams == null && !bram.endsWith(RAMB16)) {
				usedBrams = usedData.get(bram + RAMB16);
			}
			if (usedBrams == null) {
				throw new RuntimeException("bram " + bram + " was not found");
			}
			Set<IBramLine> usedBramsFinal = usedBrams;
			Set<IBramLine> fail = data.get(bram).streamLines().filter(x -> !usedBramsFinal.contains(x)).collect(Collectors.toSet());
			if (!fail.isEmpty()) {
				throw new RuntimeException("In bram " + bram + ", these attributes were not found: " + fail);
			}
		});
	}
}
