/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.initram

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryLocator
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.IBramBuilder
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.XilinxBram
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.XilinxBramBuilderFactory
import de.tu_darmstadt.rs.spartanmc.initram.memory.sections.ISection
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.IMemoryBlock
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryLineLayout
import spock.lang.Specification

class XilinxBramBuilderSpec extends Specification {

	private final MemoryLocator bramPat = Mock() {
		resolve(_, _) >> { bIdx, colIdx -> "BRAM_${bIdx}_${colIdx}" }
	}

	private final IMemoryBlock splitBlock = Spy() {
		isBitfileInitializable() >> true
		getStart() >> 0
		getLength() >> 2048
		getBitsPerMemoryLine() >> 18
		getBitsPerBlockLine() >> 9
		getModuleName() >> "MockModuleBlock"
		getMemoryLineLayout() >> MemoryLineLayout.SpartanMcSectionDependent
		getBramPattern() >> bramPat
	}

	private final IMemoryBlock nonSplitBlock = Spy() {
		isBitfileInitializable() >> true
		getStart() >> 0
		getLength() >> 1024
		getBitsPerMemoryLine() >> 18
		getBitsPerBlockLine() >> 18
		getModuleName() >> "MockModuleBlock"
		getMemoryLineLayout() >> MemoryLineLayout.SpartanMcSectionDependent
		getBramPattern() >> bramPat
	}

	private final ISection dataSection = Mock() {
		toString() >> "MockDataSection"
		getName() >> "DataSection"
		isCode() >> false
		isWritable() >> true
	}


	private List<XilinxBram> buildAndInitXilinxBrams(IBramBuilder blockBuilder, int fillUpTo) {
		blockBuilder.starting(dataSection)
		blockBuilder.add([0x00, 0x09, 0x00, 0x34] as byte[]) // 0x01234
		blockBuilder.add([0x00, 0x2b, 0x00, 0x78] as byte[]) // 0x05678
		blockBuilder.add([0x01, 0x5A, 0x00, 0x00] as byte[]) // 0x2B400
		blockBuilder.add([0x00, 0x00, 0x01, 0xA5] as byte[]) // 0x001A5
		for (int i = 16; i < fillUpTo; i+=4) {
			blockBuilder.add([0x0, 0x0, 0x0, 0x0] as byte[])
		}
		return blockBuilder.build() as List<XilinxBram>
	}

	def "SpartanMC Data 9x2x2048 layout in Xilinx Brams"() {
		def blockBuilder = new XilinxBramBuilderFactory().getBuilder(splitBlock)

		when:
		def brams = buildAndInitXilinxBrams(blockBuilder, 8196)

		then:
		brams.size() == 2
		brams[0].width == 9
		brams[0].depth == 2048
		brams[0].instanceName == "BRAM_0_1"
		brams[0].inits[0].contents == 0x005A2b09
		brams[0].initps[0].contents == 0x04// 0b0100
		(1..<64).each {
			brams[0].inits[it].contents == 0x0
		}
		(1..<8).each {
			brams[0].initps[it].contents == 0x0
		}
		brams[1].width == 9
		brams[1].depth == 2048
		brams[1].instanceName == "BRAM_0_0"

		brams[1].inits[0].contents == 0xA5007834
		brams[1].initps[0].contents == 0x08 // 0b1000
		(1..<64).each {
			brams[1].inits[it].contents == 0x0
		}
		(1..<8).each {
			brams[1].initps[it].contents == 0x0
		}
	}

	def "SpartanMC Data 18x1024 layout in Xilinx Brams"() {
		def blockBuilder = new XilinxBramBuilderFactory().getBuilder(nonSplitBlock)

		when:
		def brams = buildAndInitXilinxBrams(blockBuilder, 4096)
		// 0x01234
		// 0x05678
		// 0x2B400
		// 0x001A5

		then:
		brams.size() == 1
		brams[0].width == 18
		brams[0].depth == 1024
		brams[0].instanceName == "BRAM_0_null"
		brams[0].inits[0].contents == 0x01A5B40056781234
		brams[0].initps[0].contents == 0x20 // 0b00100000
		(1..<64).each {
			brams[0].inits[it].contents == 0x0
		}
		(1..<8).each {
			brams[0].initps[it].contents == 0x0
		}
	}
}