/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


package de.tu_darmstadt.rs.spartanmc.jconfig.application;

import de.tu_darmstadt.rs.jcommander.JCommanderRunner;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.microStreamsImport.MicroStreamsHardwareImporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.scripting.groovy.GroovyScriptPlugin;
import de.tu_darmstadt.rs.spartanmc.jconfig.scripting.shell.JConfigShell;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.JConfigFrame;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.LookAndFeel;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.FooterPanel;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.fileHandling.DocumentFileUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.JConfigRunnerFailureException;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigRunner;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParametersDelegate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.Optional;

/**
 * Application Entry Point
 * Part of a Controller of a Tiered MVC
 */
public class JConfigApplication implements Runnable {

	public static final String PROGRAM_NAME = "jconfig";
	public static final String JC3_CONFIG_FILE_DESC = "JConfig v3 Documents";
	public static final String ALL_CONFIG_FILE_DESC = "Compatible JConfig Documents";
	public static final String LEGACY_CONFIG_FILE_DESC = "Legacy JConfig Documents";
	public static final String JCONFIG_VERSION = "JConfig-3.0-alpha0";
	private static final Logger logger = LoggerFactory.getLogger(JConfigApplication.class);
	private final Thread mainThread;

	@Parameter(names = {"-q", "--quiet"}, description = "Do not show the GUI")
	private boolean quiet;

	@Parameter(names = {"--no-patch-gtk"}, description = "On GTK-System-LookAndFeel, do not replace the horrible FileChooser with a normal looking one")
	private boolean noRegularFileChooser;

	@ParametersDelegate
	private LibJConfigRunner projectRunner = new LibJConfigRunner(JCONFIG_VERSION, true);


	private ThreadGroup threadGroup;
	private JConfigFrame mainFrame;
	private RedirectingProgressIndicator progressIndicator;

	private JConfigApplication() {

		threadGroup = new ThreadGroup("jconfig") {
			@Override
			public void uncaughtException(Thread t, Throwable e) {
				logger.error("uncaught exception in {}", t.getName(), e);
				JConfigApplication.this.abort(e);
			}
		};

		mainThread = new Thread(threadGroup, "JConfig-Main-Thread") {
			@Override
			public void run() {
				JConfigApplication.this.run();
			}
		};
	}

	/**
	 * JConfig Entry Point.
	 * Creates instance to allow JCommander Parameter Processing
	 * @param args command line args. passed to jcommander unmodified
	 */
	@SuppressWarnings("SystemOut")
	public static void main(String[] args) {
		JCommanderRunner.create(new JConfigApplication()).setProgramName(PROGRAM_NAME).run(args);
	}

	public void run() {
		if (!quiet) { // early, so that dialogs can be themed as well
			LookAndFeel.setupLookAndFeel(!noRegularFileChooser);
		}

		projectRunner
			.useLastBuiltConfig()
			.configureLibraryBuilder(LibJConfigRunner::defaultLibraryConfig);

		if (!quiet) {
			progressIndicator = new RedirectingProgressIndicator();
			projectRunner
				.configureLibraryBuilder(builder -> builder.setProgressIndicator(progressIndicator))
				.executeWithProject(this::startGui)
				.autoOpenConfig((docHandler, configFile) -> DocumentFileUtil.openDocument(this.mainFrame, docHandler, configFile));
		} else {
			projectRunner
				.autoOpenConfig();
		}

		projectRunner
			.autoImport((path, targetSuggester) -> new MicroStreamsHardwareImporter(targetSuggester.getTargetName(), targetSuggester.getDeviceName().orElse(null)));//TODO switch to different importers

		projectRunner
			.allowAutoExecuteScript()
			.allowAutoBuild()
			.allowAutoCreateDocument();

		projectRunner
			.executeWithProject(prj -> prj.getDocumentHandler().setPluginEngine(new GroovyScriptPlugin()));
		//			.executeWithProject(prj -> prj.getDocumentHandler().setPluginEngine(new LuaScriptPlugin()));

		try {
			LibJConfigProject project = projectRunner.buildAndRun();

			if (quiet && !projectRunner.isUpdate()) {
				new JConfigShell(project);
			}
		} catch (JConfigRunnerFailureException e) {
			if (quiet) {
				logger.error("Failed to execute desired actions in console mode, terminating");
				System.exit(1);
			}
		}
	}

	private void startGui(LibJConfigProject project) {
		logger.info("Starting GUI");

		mainFrame = new JConfigFrame(this, project.getApplication(), project.getDocumentHandler());

	}

	protected void abort(final Throwable t) {
		try {
			File f = new File(
				String.format(
					"jconfig-crash-%d.xml",
					System.currentTimeMillis()
				)
			);
			logger.error("Ooops! We encountered a serious problem... :( might be you just discovered a bug. Trying emergency save!", t);
			//TODO try emergency save
		} catch (Exception e) {
			logger.error("Unkown exception during abort()", e);
		}
		logger.warn("Terminating application");
		System.exit(1);
	}

	public void onWindowClosed(JConfigFrame jConfigFrame) {
		logger.info("Shutting down");
		System.exit(0);
	}

	public RedirectingProgressIndicator getProgressIndicator() {
		return progressIndicator;
	}

	/**
	 * @author ramonw
	 */
	public class RedirectingProgressIndicator implements Library.IProgressIndicator {

		private int maxProgress;
		private int current;
		private boolean done = false;
		private String text;

		private FooterPanel footer;

		private RedirectingProgressIndicator() {
		}

		private synchronized Optional<FooterPanel> getFooter() {
			return Optional.ofNullable(footer);
		}

		public synchronized void attach(FooterPanel footer) {
			this.footer = footer;
			if (!done) {
				footer.setMaxProgress(maxProgress);
				footer.setProgress(current);
				footer.setProgressText(text);
			} else {
				footer.progressComplete();
			}

		}

		@Override
		public void setProgress(int progress) {
			current = progress;
			getFooter().ifPresent(f -> f.setProgress(progress));
		}

		@Override
		public void setMaxProgress(int maxProgress) {
			this.maxProgress = maxProgress;
			getFooter().ifPresent(f -> f.setMaxProgress(maxProgress));
		}

		@Override
		public void stepProgress(int steps) {
			current += steps;
			getFooter().ifPresent(f -> f.setProgress(current));
		}

		@Override
		public void setProgressText(String fmt, Object... args) {
			text = String.format(fmt, args);
			getFooter().ifPresent(f -> f.setProgressText(text));
		}

		@Override
		public void progressComplete() {
			done = true;
			getFooter().ifPresent(Library.IProgressIndicator::progressComplete);
		}
	}
}
