/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.settings;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.Properties;

/**
 * Settings Manager. Handles loading and storing of global settings, user settings, project specific settings vie the Java-BuiltIn Properties.
 *
 * Meant to be used to store last used document per project, so that the last document can be automatically opened.
 *
 * Created by Ramon on 31.03.2016.
 */
public class JConfigSettings {

	private static final Logger logger = LoggerFactory.getLogger(JConfigSettings.class);
	private final Properties jcProp;
	private final Path settingsFile;

	/**
	 * Start managing settings for current project
	 * @param workingDir current project dir root
	 */
	public JConfigSettings(Path workingDir) {
		settingsFile = workingDir.resolve(".jconfig");
		jcProp = new Properties();
	}

	/**
	 * Load all properties from files
	 * (could be hierarchical (system, user, project), but only project specific right now)
	 */
	public void load() {
		if (Files.exists(settingsFile)) {
			logger.info("Found project settings");
			Reader stream = null;
			try {
				stream = Files.newBufferedReader(settingsFile);
				jcProp.load(stream);
				stream.close();
			} catch (IOException e) {
				logger.error("Could not load settings {}", settingsFile, e);
			}
		} else {
			logger.info("No Settings File found");
		}
	}

	/**
	 * Announce configFile as last opened document
	 * @param configFile path of a valid document
	 */
	public void setLastBuild(Path configFile) {

		jcProp.setProperty("LAST_BUILT_FILE", configFile.toAbsolutePath().toString());
	}

	/**
	 * Write project specific properties to file
	 */
	public void storeProjectSpecific() {
		try {
			Writer stream = Files.newBufferedWriter(settingsFile, StandardOpenOption.CREATE, StandardOpenOption.WRITE, StandardOpenOption.TRUNCATE_EXISTING);
			jcProp.store(stream, "Local Project Settings");
			stream.close();
		} catch (IOException e) {
			logger.error("Failed to save settings {}", settingsFile, e);
		}
	}
}
