/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


package de.tu_darmstadt.rs.spartanmc.jconfig.ui;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.application.JConfigApplication;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.JConfigAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.Lua.AutoScriptAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.document.BuildAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.document.NewDocumentAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.document.OpenDocumentAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.document.SaveAsDocumentAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.document.SaveDocumentAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.document.ShowSchematicAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.document.ValidateAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.FooterPanel;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.fileHandling.DocumentFileUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.fileHandling.FilePicker;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.ConfigurationDocumentPane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.resources.IconUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.util.UserCanceledException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.nio.file.Path;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JToggleButton;
import javax.swing.filechooser.FileNameExtensionFilter;

/**
 * Main Window of JConfig
 */
public class JConfigFrame extends JFrame implements WindowListener, FilePicker, IApplicationConfiguration.ILibraryReadyListener, IDocumentHandler.IDocumentHandlerListener {

	public static final Marker uiMarker = MarkerFactory.getMarker("UI");
	public static final int STRUT_WIDTH = 3;
	public static final Dimension MINIMUM_SIZE = new Dimension(1000, 700);
	public static final Dimension PREFERRED_SIZE = new Dimension(1200, 800);
	public static final int FLOW_HGAP = 3;
	public static final int FLOW_VGAP = 1;
	private static final Logger logger = LoggerFactory.getLogger(JConfigFrame.class);
	private static final String APPLICATION_TITLE = "SpartanMC System-Builder";
	private final IApplicationConfiguration application;
	private final IDocumentHandler documentHandler;
	private final JConfigApplication rootClass;

	private JPanel toolBar;
	private FooterPanel footer;
	private ConfigurationDocumentPane documentPane;

	// Actions
	private JConfigAction saveAsAction;
	private JConfigAction saveAction;
	private JConfigAction newAction;
	private JConfigAction loadAction;
	private JConfigAction validateAction;
	private JConfigAction buildActionInc;
	private JConfigAction buildActionAll;
	private JConfigAction showSchematicAction;
	private JConfigAction autoScriptAction;


	/**
	 * Create an empty jconfig frame.
	 * Lazily initializes some of its components, when library / document are loaded
	 *
	 * @param rootClass used solely to signal program termination back to Entry Class
	 * @param application application wide configurations
	 * @param documentHandler a documentHandler to which this frame will attach itself
	 */
	public JConfigFrame(JConfigApplication rootClass, IApplicationConfiguration application, IDocumentHandler documentHandler) {
		super(APPLICATION_TITLE);
		this.rootClass = rootClass;

		logger.debug(uiMarker, "Main Window created");

		this.application = application;
		this.documentHandler = documentHandler;

		documentHandler.addDocumentHandlerListener(this);

		Container contentPane = getContentPane();
		contentPane.setLayout(new BorderLayout());

		saveAction = new SaveDocumentAction(documentHandler, this);
		saveAsAction = new SaveAsDocumentAction(documentHandler, this);
		validateAction = new ValidateAction(documentHandler);
		buildActionInc = new BuildAction(documentHandler, true);
		buildActionAll = new BuildAction(documentHandler, false);
		showSchematicAction = new ShowSchematicAction(application, documentHandler, this);
		newAction = new NewDocumentAction(this, documentHandler, application);
		newAction.setEnabled(false);
		loadAction = new OpenDocumentAction(this, documentHandler, this);
		autoScriptAction = new AutoScriptAction(documentHandler);

		JMenu fileMenu = new JMenu("File");
		fileMenu.add(new JMenuItem(newAction));
		fileMenu.add(new JMenuItem(loadAction));
		fileMenu.addSeparator();
		fileMenu.add(new JMenuItem(saveAction));
		fileMenu.add(new JMenuItem(saveAsAction));
		fileMenu.addSeparator();
		fileMenu.add(new JMenuItem(validateAction));
		fileMenu.add(new JMenuItem(buildActionInc));
		fileMenu.add(new JMenuItem(buildActionAll));
		fileMenu.addSeparator();
		fileMenu.add(new JMenuItem(showSchematicAction));
		fileMenu.addSeparator();
		JMenuItem menuItem = new JMenuItem(autoScriptAction);
		menuItem.setModel(new JToggleButton.ToggleButtonModel());
		fileMenu.add(menuItem);


		JMenuBar menuBar = new JMenuBar();
		menuBar.add(fileMenu);
		//		menuBar.setBorderPainted(true);
		//		menuBar.setBorder(new SingleLineBorder(SwingConstants.BOTTOM));
		setJMenuBar(menuBar);

		toolBar = new JPanel();
		toolBar.setLayout(new FlowLayout(FlowLayout.LEFT, FLOW_HGAP, FLOW_VGAP));
		toolBar.add(new JButton(newAction));
		toolBar.add(new JButton(loadAction));
		toolBar.add(new JButton(saveAction));
		toolBar.add(Box.createHorizontalStrut(STRUT_WIDTH));
		toolBar.add(new JButton(validateAction));
		toolBar.add(new JButton(buildActionInc));
		toolBar.add(Box.createHorizontalStrut(STRUT_WIDTH));
		toolBar.add(new JButton(showSchematicAction));
		toolBar.add(Box.createHorizontalStrut(STRUT_WIDTH));
		toolBar.add(new JToggleButton(autoScriptAction));

		contentPane.add(toolBar, BorderLayout.NORTH);

		documentPane = new ConfigurationDocumentPane(application, documentHandler);
		contentPane.add(documentPane, BorderLayout.CENTER);

		setIconImage(IconUtil.getIcon("48/application0.png").getImage());


		footer = new FooterPanel();
		documentHandler.setDocumentStatusReporter(footer);
		if (rootClass!=null)
			rootClass.getProgressIndicator().attach(footer);
		contentPane.add(footer, BorderLayout.SOUTH);

		addWindowListener(this);
		setMinimumSize(MINIMUM_SIZE);
		setPreferredSize(PREFERRED_SIZE);
		pack();
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
		setVisible(true);

		application.addOrCallReadyListener(this);

		onDocumentSwitched(documentHandler.hasLoadedDocument());
	}

	private void updateTitle() {
		Path projDir = documentHandler.getWorkingDir();

		Path saveLocation = documentHandler.getSaveLocation();
		Path relativeSaveLocation = saveLocation != null? projDir.relativize(saveLocation.toAbsolutePath()) : null;

		String saveString = relativeSaveLocation != null? " - " + relativeSaveLocation : "";

		setTitle(APPLICATION_TITLE + " - " + projDir + saveString);
	}

	public FooterPanel getFooter() {
		return footer;
	}

	@Override
	public File pickFile(String title, boolean creating) {
		Path initialLocation = (documentHandler.hasSaveLocation())?
			documentHandler.getSaveLocation() : documentHandler.getWorkingDir();
		JFileChooser fileChooser = new JFileChooser(initialLocation.toFile());
		fileChooser.setAcceptAllFileFilterUsed(false);
		fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
			JConfigApplication.ALL_CONFIG_FILE_DESC,
			ILibJConfigProject.JC3_CONFIG_FILE_EXT,
			ILibJConfigProject.LEGACY_CONFIG_FILE_EXT
		));
		fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
			JConfigApplication.JC3_CONFIG_FILE_DESC,
			ILibJConfigProject.JC3_CONFIG_FILE_EXT));
		fileChooser.addChoosableFileFilter(new FileNameExtensionFilter(
			JConfigApplication.LEGACY_CONFIG_FILE_DESC,
			ILibJConfigProject.LEGACY_CONFIG_FILE_EXT));
		fileChooser.setDialogTitle(title);

		if (documentHandler.hasSaveLocation()) {
			fileChooser.setSelectedFile(documentHandler.getSaveLocation().toFile());
		}

		File file;

		switch (fileChooser.showSaveDialog(this)) {
			case JFileChooser.APPROVE_OPTION:
				file = fileChooser.getSelectedFile();

				String name = file.getName();
				if (!name.endsWith(ILibJConfigProject.JC3_CONFIG_FILE_EXT)) {
					file = new File(name + "." + ILibJConfigProject.JC3_CONFIG_FILE_EXT);
				}

				if (creating && file.exists()) {
					int answer = JOptionPane.showConfirmDialog(
						this,
						"File '" + file + "' already exists!\nDo you want to OVERWRITE the existing file?",
						"File already exists",
						JOptionPane.YES_NO_OPTION,
						JOptionPane.WARNING_MESSAGE);
					return (answer == JOptionPane.YES_OPTION)? file : null;
				}

				return file;
			case JFileChooser.CANCEL_OPTION:
				logger.info("FileChooser was canceled");
				return null;
			default:
				logger.error("File chooser returned error");
				return null;
		}
	}

	@Override
	public void windowOpened(final WindowEvent e) {
		// unused
	}

	@Override
	public void windowClosing(final WindowEvent event) {
		try {
			DocumentFileUtil.ensureSavedDocument(this.getParent(), documentHandler, true);// autosave, do never delay closing
			JConfigFrame.this.dispose();
			logger.info("closing...");
		} catch (UserCanceledException e) {
			logger.info("closing canceled by user wish");
		}
	}

	@Override
	public void windowClosed(final WindowEvent e) {
		if (rootClass!=null)
			rootClass.onWindowClosed(this);
	}

	@Override
	public void windowIconified(final WindowEvent e) {
		// unused
	}

	@Override
	public void windowDeiconified(final WindowEvent e) {
		// unused
	}

	@Override
	public void windowActivated(final WindowEvent e) {
		// unused
	}

	@Override
	public void windowDeactivated(final WindowEvent e) {
		// unused
	}

	@Override
	public void onLibraryReady(Library library) {
		newAction.setEnabled(true);
	}

	@Override
	public void onDocumentSwitched(boolean valid) {
		updateTitle();
	}

	@Override
	public void onSaveLocationChanged() {
		updateTitle();
	}
}
