/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.ui.components;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.ConfigurationDocumentPane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.resources.IconUtil;

import java.awt.Component;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.swing.AbstractAction;
import javax.swing.AbstractListModel;
import javax.swing.DefaultListCellRenderer;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPopupMenu;
import javax.swing.ListCellRenderer;

/**
 * Created by Ramon on 20.04.2016.
 */
public class ConfigurationIssueList extends JList<ConfigurationIssue> {

	private static Comparator<ConfigurationIssue> ISSUE_COMP = Comparator.comparing(ConfigurationIssue::getLevel).thenComparing(ConfigurationIssue::getActuality);
	private final IssueModel model;
	private final ConfigurationDocumentPane parentPane;
	private IConfigurationDocument document;

	public ConfigurationIssueList(ConfigurationDocumentPane parentPane) {
		this.parentPane = parentPane;
		model = new IssueModel();
		setModel(model);
		setCellRenderer(new IssueCellRenderer());
		setComponentPopupMenu(new IssuePopupMenu());
	}

	public void setDocument(IConfigurationDocument document) {
		if (this.document != null) {
			this.document.removeConfigurationIssueListener(model);
		}

		this.document = document;

		if (document != null) {
			document.addConfigurationIssueListener(model);
		}

		model.update();
	}

	private void updateList() {
	}


	private class DismissIssueAction extends AbstractAction {

		private final IssuePopupMenu popupMenu;

		public DismissIssueAction(IssuePopupMenu popupMenu) {
			super("Dismiss Issue");
			this.popupMenu = popupMenu;
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			document.getIssueManager().dismissIssue(popupMenu.issue);
		}
	}


	private class GotoIssueAction extends AbstractAction {

		private final IssuePopupMenu popupMenu;

		public GotoIssueAction(IssuePopupMenu popupMenu) {
			super("Goto Issue...");
			this.popupMenu = popupMenu;
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			Optional<IConfigurationElement<?, ?, ?>> reference = popupMenu.issue.getReferencedModule();
			reference.ifPresent(r -> {
				parentPane.selectElement(r, true);
			});
		}
	}


	private class IssueModel extends AbstractListModel<ConfigurationIssue> implements IConfigurationDocument.IConfigurationIssueListener {

		private List<ConfigurationIssue> issues = Collections.EMPTY_LIST;

		@Override
		public int getSize() {
			return issues.size();
		}

		@Override
		public ConfigurationIssue getElementAt(int index) {
			return issues.get(index);
		}

		public void update() {
			List<ConfigurationIssue> oldList = issues;

			if (document != null) {
				issues = document.getConfigurationIssues().stream().sorted(ISSUE_COMP).collect(Collectors.toList());
			} else {
				issues = Collections.EMPTY_LIST;
			}

			if (oldList.size() > 0) {
				fireIntervalRemoved(this, 0, oldList.size() - 1);
			}
			if (issues.size() > 0) {
				fireIntervalAdded(this, 0, issues.size() - 1);
			}
		}

		@Override
		public void onNewConfigurationIssue(ConfigurationIssue issue, ConfigurationIssue.IssueType level) {
			if (level == ConfigurationIssue.IssueType.Error) {
				issues.add(0, issue);
				fireIntervalAdded(this, 0, 0);
			} else {
				issues.add(issue);
				int i = issues.size() - 1;
				fireIntervalAdded(this, i, i);
			}
		}

		@Override
		public void onConfigurationIssueDismissed(ConfigurationIssue issue) {
			onIssueRemoved(issue);
		}

		@Override
		public void onConfigurationIssueResolved(ConfigurationIssue issue) {
			onIssueRemoved(issue);
		}

		private void onIssueRemoved(ConfigurationIssue issue) {
			int i = issues.indexOf(issue);
			if (i >= 0) {
				issues.remove(i);
				fireIntervalRemoved(this, i, i);
			}
		}
	}


	private class IssueCellRenderer implements ListCellRenderer<ConfigurationIssue> {

		private final DefaultListCellRenderer cellRenderer;
		private final ImageIcon warnIcon;
		private final ImageIcon errorIcon;

		private IssueCellRenderer() {
			cellRenderer = new DefaultListCellRenderer();
			errorIcon = IconUtil.getIcon("48/overlay_error.png");
			warnIcon = IconUtil.getIcon("48/overlay1.png"); //TODO use actual warn icon
			//TODO add notice Icon
		}

		@Override
		public Component getListCellRendererComponent(JList<? extends ConfigurationIssue> list, ConfigurationIssue value, int index, boolean isSelected, boolean cellHasFocus) {
			JLabel comp = (JLabel) cellRenderer.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
			comp.setIcon(value.getLevel() == ConfigurationIssue.IssueType.Error? errorIcon : warnIcon);
			comp.setText(value.getMessage() + ": " + value.getHintText());
			comp.setToolTipText("Occurred while: " + value.getSituation() + ", Actuality: " + value.getActuality());
			return comp;
		}
	}


	private class IssuePopupMenu extends JPopupMenu {

		private final GotoIssueAction gotoItem;
		private final DismissIssueAction dismissItem;
		private ConfigurationIssue issue;

		private IssuePopupMenu() {
			dismissItem = new DismissIssueAction(this);
			gotoItem = new GotoIssueAction(this);
			add(dismissItem);
			add(gotoItem);
		}

		@Override
		public void show(Component invoker, int x, int y) {
			ConfigurationIssueList list = ConfigurationIssueList.this;
			int index = list.locationToIndex(new Point(x, y));
			if (index < 0) {
				return; // click on BG
			}

			issue = model.issues.get(index);

			super.show(invoker, x, y);
		}
	}
}
