/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


package de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.attributes;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.awt.Color;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import javax.swing.JTextField;
import javax.swing.UIManager;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;


public class AttributeTextField<T extends IAttributeHandler> extends JTextField implements DocumentListener, FocusListener, ActionListener, IAttributeWatcher<T> {

	private static final Marker attributeUiMarker = MarkerFactory.getMarker("AttributeUiHandling");
	private static final Logger logger = LoggerFactory.getLogger(AttributeTextField.class);

	protected T handler;
	private boolean valid = false;
	private String lastValue;

	public AttributeTextField() {
		getDocument().addDocumentListener(this);
		addFocusListener(this);
		addActionListener(this);
	}

	@Override
	public void setHandler(T handler) {
		this.handler = handler;
		update();
	}

	@Override
	public void update() {
		boolean active = handler != null;

		if (handler != null) {
			boolean editable = handler.isEditable().orElse(true);
			setEditable(editable);
			setEnabled(editable);
		}

		setText((active)? getValue().orElse("") : "");
	}

	protected CalculatedValueResult<String> getValue() {
		return handler.getValue();
	}

	@Override
	public void insertUpdate(DocumentEvent e) {
		checkValid();
	}

	@Override
	public void removeUpdate(DocumentEvent e) {
		checkValid();
	}

	@Override
	public void changedUpdate(DocumentEvent e) {
		if (isEditable()) {
			checkCommit();
		}
	}

	private void checkValid() {

		if (handler!=null) {
			if (isEditable()) {
				valid = handler.isValidValue(getText());
			} else {
				valid = handler.isValid(false);
			}
		} else {
			valid = false;
		}

		showValid();
	}

	private void showValid() {
		if (isEnabled() && !valid) {
			setForeground(Color.RED);
			setFont(getFont().deriveFont(Font.BOLD | Font.ITALIC));
		} else {
			setForeground((Color) UIManager.get("TextField.foreground"));
			setFont(getFont().deriveFont(Font.PLAIN));
		}
	}

	@Override
	public void focusGained(FocusEvent e) {
		// irrelevant
	}

	@Override
	public void focusLost(FocusEvent e) {
		if (isEditable()) {
			checkCommit();
		}
	}

	private void checkCommit() {
		String value = getText();

		if (valid) {
			handler.setStoredValue(value);
		} else {
			logger.warn(attributeUiMarker, "Could not commit to {}, value invalid: {}", handler, value);
		}
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		checkCommit();
	}
}
