/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.attributes;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBooleanParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IFloatParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IIntegerParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterConverter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Created by Ramon on 17.08.2016.
 */
public class UnitConverter implements IParameterConverter {

	private static final Logger logger = LoggerFactory.getLogger(UnitConverter.class);
	private static final Pattern pattern = Pattern.compile("^([0-9]+)\\.([0-9]+) ([a-zA-Z]+)$");

	private static final String[] factorNames1000 = {"", "k", "M", "G"};
	private static final String[] factorNames1024 = {"", "ki", "Mi", "Gi"};

	private String unitSuffix;
	private int factor;
	private int precision;


	public UnitConverter(String unitConfig) {
		Matcher match = pattern.matcher(unitConfig);
		if (match.matches()) {
			unitSuffix = match.group(3);
			factor = Integer.parseInt(match.group(1));
			precision = Integer.parseInt(match.group(2));
		} else {
			throw new IllegalStateException("Invalid UI Unit Config: " + unitConfig);
		}
	}

	private String getUnitizedValue(double value) {
		double result = value;
		double nextResult;
		String shortUnit;

		int exp = 0;

		if (factor>1) {

			while ((nextResult = result / factor) >= 1.0) {
				exp++;
				result = nextResult;
			}

		}

		if (factor == 1000) {
			shortUnit = factorNames1000[exp];
		} else if (factor == 1024) {
			shortUnit = factorNames1024[exp];
		} else {
			if (exp != 0) {
				shortUnit = String.format("* %d^%d ", factor, exp);
			} else {
				shortUnit = "";
			}
		}

		return String.format("%." + precision + "f %s%s", result, shortUnit, unitSuffix);
	}

	@Override
	public String convert(IIntegerParameterHandler param) {
		return param.getIntegerValue().map(this::getUnitizedValue).orError();
	}

	@Override
	public String convert(IFloatParameterHandler param) {
		return param.getFloatValue().map(this::getUnitizedValue).orError();
	}

	@Override
	public String convert(IParameterHandler param) {
		throw new UnsupportedOperationException("Only numbers");
	}

	@Override
	public String convert(IBooleanParameterHandler param) {
		throw new UnsupportedOperationException("Only numbers");
	}
}
