/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.model;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ResolvedConnector;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.handlers.AvailablePortsHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.handlers.IAvailablePortsHandler;

import java.util.ArrayList;
import java.util.List;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;

/**
 * @author Ramon Wirsch
 */
public abstract class AvailablePortsModel<T extends IConnectorHandler> implements AvailablePortsHandler.IAvailablePortsListener, IAvailablePortsModel {

	protected final IAvailablePortsHandler availablePorts;
	protected final IConfigurationElement<?, ?, ?> relativeTo;
	protected final T portHandler;
	private final List<ListDataListener> listeners = new ArrayList<>();
	private List<ResolvedConnector> resolvedPorts;
	private boolean registered = false;
	private Object selected;

	private AvailablePortsModel(IAvailablePortsHandler availablePorts, IConfigurationElement<?, ?, ?> relativeTo, T portHandler) {
		this.availablePorts = availablePorts;
		this.relativeTo = relativeTo;
		this.portHandler = portHandler;
		resolvedPorts = getResolvedConnectors();
	}

	public static AvailablePortsModel getPortModel(IAvailablePortsHandler availablePorts, IConfigurationElement<?, ?, ?> relativeTo, IPortHandler portHandler) {
		return new PortModel(availablePorts, relativeTo, portHandler);
	}

	public static AvailablePortsModel getBusModel(IAvailablePortsHandler availablePorts, IConfigurationElement<?, ?, ?> relativeTo, IBusHandler busHandler) {
		return new BusModel(availablePorts, relativeTo, busHandler);
	}

	public T getPortHandler() {
		return portHandler;
	}

	protected abstract List<ResolvedConnector> getResolvedConnectors();

	@Override
	public int getSize() {
		return resolvedPorts.size();
	}

	@Override
	public ResolvedConnector getElementAt(int index) {
		return resolvedPorts.get(index);
	}

	@Override
	public void addListDataListener(ListDataListener l) {
		listeners.add(l);

		if (!registered) {
			availablePorts.addListener(this);
			registered = true;
		}
	}
@Override
	public void removeListDataListener(ListDataListener l) {
		listeners.remove(l);

		if (listeners.size() == 0 && registered) {
			availablePorts.removeListener(this);
			registered = false;
		}
	}	@Override
	public void setSelectedItem(Object anItem) {
		if ((selected != null && !selected.equals(anItem)) ||
			selected == null && anItem != null) {
			selected = anItem;

			ListDataEvent e = new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, -1, -1);

			for (ListDataListener listener : listeners) {
				listener.contentsChanged(e);
			}
		}
	}

	@Override
	public void onAvailablePortsChanged() {
		resolvedPorts = getResolvedConnectors();

		ListDataEvent e = new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, 0, getSize());

		for (ListDataListener listener : listeners) {
			listener.contentsChanged(e);
		}
	}

	@Override
	public boolean isRecommendedConnection(ResolvedConnector connector) {
		if (connector.getConnectorHandler() == null) {
			return true;//Dummy Connector
		}
		return portHandler.isConnectibleTo(connector.getConnectorHandler());
	}

	@Override
	public Object getSelectedItem() {
		return selected;
	}


	private static class BusModel extends AvailablePortsModel<IBusHandler> {

		private BusModel(IAvailablePortsHandler availablePorts, IConfigurationElement<?, ?, ?> relativeTo, IBusHandler portHandler) {
			super(availablePorts, relativeTo, portHandler);
		}

		@Override
		protected List<ResolvedConnector> getResolvedConnectors() {
			return availablePorts.getResolvedBusesFor(relativeTo, portHandler);
		}
	}


	private static class PortModel extends AvailablePortsModel<IPortHandler> {

		private PortModel(IAvailablePortsHandler availablePorts, IConfigurationElement<?, ?, ?> relativeTo, IPortHandler portHandler) {
			super(availablePorts, relativeTo, portHandler);
		}

		@Override
		protected List<ResolvedConnector> getResolvedConnectors() {
			return availablePorts.getResolvedPortsFor(relativeTo, portHandler);
		}

	}


}
