/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.model;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.ResolvedConnector;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.ComponentUtil;
import de.tu_darmstadt.rs.util.LevenshteinDistance;

import com.jidesoft.swing.event.SearchableEvent;
import com.jidesoft.swing.event.SearchableListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import javax.swing.event.ListDataEvent;
import javax.swing.event.ListDataListener;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;

/**
 * Created by wenzel on 02.02.17.
 */
public class FilteringPortsModel implements IAvailablePortsModel, ListDataListener, SearchableListener, PopupMenuListener {

	private static final Logger logger = LoggerFactory.getLogger(FilteringPortsModel.class);

	private final AvailablePortsModel<?> model;
	private final List<ListDataListener> listeners = new ArrayList<>();

	private List<ResolvedConnector> resolvedConnectors;
	private ComponentUtil.CustomSearchable searchable;
	private final int maxLevenshteinDistance = 5;

	private String searchingText;

	public FilteringPortsModel(AvailablePortsModel<?> model) {

		this.model = model;
		filterList(null);

		model.addListDataListener(this);
	}

	private void updateConnectors() {
		if (resolvedConnectors==null) {
			if (searchable != null) {
				resolvedConnectors = model.getResolvedConnectors().stream()
						.filter(c -> searchable.compare(c, searchingText))
						.sorted(
								Comparator.comparing((ResolvedConnector c) -> Math.min(getLevenshteinDistance(c), maxLevenshteinDistance))
										.thenComparing(Comparator.comparing((ResolvedConnector c) -> c.getConnectorHandler().getAbsoluteConnectorName()))
						)
						.collect(Collectors.toList());
			} else {
				resolvedConnectors = model.getResolvedConnectors();
			}
		}
	}

	private void filterList(String searchingText) {
		resolvedConnectors = null;
		this.searchingText = searchingText;
	}

	@Override
	public void setSelectedItem(Object anItem) {
		model.setSelectedItem(anItem);
	}

	@Override
	public Object getSelectedItem() {
		return model.getSelectedItem();
	}

	@Override
	public int getSize() {
		updateConnectors();
		return resolvedConnectors.size();
	}

	@Override
	public ResolvedConnector getElementAt(int index) {
		updateConnectors();
		return resolvedConnectors.get(index);
	}

	@Override
	public void addListDataListener(ListDataListener l) {
		listeners.add(l);
	}

	@Override
	public void removeListDataListener(ListDataListener l) {
		listeners.remove(l);
	}

	@Override
	public boolean isRecommendedConnection(ResolvedConnector connector) {
		return model.isRecommendedConnection(connector);
	}

	@Override
	public void intervalAdded(ListDataEvent e) {
		listeners.forEach(l -> l.intervalAdded(e));
		filterList(null);
	}

	@Override
	public void intervalRemoved(ListDataEvent e) {
		listeners.forEach(l -> l.intervalRemoved(e));
		filterList(null);
	}

	@Override
	public void contentsChanged(ListDataEvent e) {
		listeners.forEach(l -> l.contentsChanged(e));
		filterList(null);
	}

	@Override
	public void searchableEventFired(SearchableEvent e) {
		//Do not change filters if the popup is not visible
		if (searchable.isPopupVisible()) {
			filterList(e.getSearchingText());
			ListDataEvent le = new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, -1, -1);
			listeners.forEach(l -> l.contentsChanged(le));
		}
	}

	public void setSearchable(ComponentUtil.CustomSearchable searchable) {
		searchable.addSearchableListener(this);
		this.searchable = searchable;
		filterList(null);
	}

	@Override
	public void popupMenuWillBecomeVisible(PopupMenuEvent e) {
		filterList(null);
		ListDataEvent le = new ListDataEvent(this, ListDataEvent.CONTENTS_CHANGED, -1, -1);
		listeners.forEach(l -> l.contentsChanged(le));

	}

	@Override
	public void popupMenuWillBecomeInvisible(PopupMenuEvent e) {
		searchingText = null;
	}

	@Override
	public void popupMenuCanceled(PopupMenuEvent e) {
		searchingText = null;
	}

	private int getLevenshteinDistance(ResolvedConnector conn) {
		Objects.requireNonNull(conn);
		if (conn.getConnectorHandler() == null) {
			return Integer.MAX_VALUE;
		}
		return LevenshteinDistance.get(conn.getConnectorHandler().getPortName(), model.getPortHandler().getPortName());

	}

	public boolean isLevenshteinClose(ResolvedConnector conn) {
		return getLevenshteinDistance(conn) < maxLevenshteinDistance;
	}

	public String getSearchingText() {
		return searchingText;
	}
}
