/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.ui.fileHandling;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.handlers.DocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.util.UserCanceledException;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibraryUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.Component;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

/**
 * Common utility functions to save, update, open a new file within a {@link DocumentHandler}.
 * Mostly used from Actions
 */
public class DocumentFileUtil {

	private static final Logger logger = LoggerFactory.getLogger(DocumentFileUtil.class);

	public static void ensureSavedDocument(Component parent, IDocumentHandler documentHandler, boolean autosave) throws UserCanceledException {
		if (documentHandler.needsSaving()) {
			if (!documentHandler.hasSaveLocation()) {
				if (autosave) {
					documentHandler.setSaveLocation(documentHandler.getWorkingDir().resolve(ILibJConfigProject.getAutosaveFileName(null)));
					documentHandler.save();

				} else {
					confirmSave(parent, null);
				}
			} else if (autosave) {
				logger.warn("Detected unsaved changes to document. Doing Auto-Save!");
				documentHandler.setSaveLocation(documentHandler.getSaveLocation().resolveSibling(ILibJConfigProject.getAutosaveFileName(documentHandler.getSaveLocation().getFileName().toString())));

				documentHandler.save();
			} else {
				confirmSave(parent, documentHandler);
			}
		}
	}

	private static void confirmSave(Component parent, IDocumentHandler handler) throws UserCanceledException { //TODO revert to autosave on timer
		int response;
		if (handler == null) {
			response = JOptionPane.showConfirmDialog(parent, "If you continue you will lose the jc3 document", "Unsaved changes", JOptionPane.OK_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
		} else {
			response = JOptionPane.showConfirmDialog(parent, "Do you want to save the jc3 document to " + handler.getSaveLocation(), "Unsaved changes", JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.WARNING_MESSAGE);
		}
		switch (response) {
			case JOptionPane.NO_OPTION:
				return; // user said not to save
			case JOptionPane.YES_OPTION://also OK
				if (handler != null) {
					handler.save(); // user said overwrite
				} else {
					return; // user said ok to loosing
				}
			default:
				throw new UserCanceledException(); // user hit cancel or closed window
		}
	}

	public static void openDocument(Component parent, IDocumentHandler documentHandler, final Path pick) {
		SwingUtilities.invokeLater(() -> {
			if (pick != null && Files.exists(pick)) {
				boolean usingAutosave = false;
				Path actualFile = pick;
				Optional<Path> autosave = documentHandler.findAutosave();
				try {
					if (autosave.isPresent() && Files.getLastModifiedTime(autosave.get()).toMillis() > Files.getLastModifiedTime(pick).toMillis() && ILibJConfigProject.doesAutosaveNameMatch(autosave.get(), pick)) {
						int response = JOptionPane.showConfirmDialog(parent, "The autosave: " + autosave.get().getFileName() + " is newer than: " + actualFile.getFileName() + "! Use the autosave?", "Use Autosave?", JOptionPane.OK_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE);
						if (response == JOptionPane.OK_OPTION) {
							usingAutosave = true;
							actualFile = autosave.get();
						}
					}
				} catch (IOException e) {
					logger.error("Failed to read lastModified of {} / {}", autosave, pick, e);
				}

				try {
					documentHandler.open(actualFile);
					if (usingAutosave) {
						documentHandler.setSaveLocation(pick);
					}
				} catch (Exception e) {
					logger.error("Failed to open document {}", actualFile, e);
					JOptionPane.showMessageDialog(parent,
						"<html><body><p style='width: 800px;'>"+e.getMessage()+"</p></body></html>",
						"Failed to open document", JOptionPane.ERROR_MESSAGE);
				}
			} else {
				logger.error("You picked an invalid File {}", pick);
			}
		});
	}

	public static void pickAndOpenDocument(Component parent, String title, IDocumentHandler documentHandler, FilePicker filePicker) {
		File pick = filePicker.pickFile(title, false);

		Path pPick = pick != null? pick.toPath() : null;

		openDocument(parent, documentHandler, pPick);
	}

	public static void pickSaveLocation(String title, IDocumentHandler documentHandler, FilePicker filePicker) {
		File pick = filePicker.pickFile(title, true);

		if (pick != null) {
			documentHandler.setSaveLocation(pick.toPath());
		}
	}

	public static TargetDescription askForTarget(Component parent, Library library) throws UserCanceledException {
		TargetDescription preselectedTarget = null;
		final Optional<TargetDescription> nexys = library.getAvailableTargets().stream()
				.filter(t -> t.getName().contains("Nexys-Video"))
				.findFirst();
		if (nexys.isPresent()) {
			preselectedTarget = nexys.get();
		}
		TargetDescription target = ((TargetDescription) JOptionPane.showInputDialog(parent, "Please select a Target", "New Configuration", JOptionPane.PLAIN_MESSAGE, null, LibraryUtil.getSortedTargetDescriptions(library.getAvailableTargets()).toArray(), preselectedTarget));

		if (target == null) {
			throw new UserCanceledException();
		}

		return target;
	}
}
