/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


package de.tu_darmstadt.rs.spartanmc.jconfig.ui.menus;

import de.tu_darmstadt.rs.process.ProcessUtil;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.configuration.AddBusBreakoutAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.configuration.AddModuleAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.configuration.RemoveModuleAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.fileHandling.DocumentationLocator;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.resources.IconUtil;

import com.jidesoft.dialog.BannerPanel;
import com.jidesoft.popup.JidePopup;
import com.jidesoft.swing.Resizable;
import com.jidesoft.utils.PortingUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.nio.file.Path;
import java.util.Collection;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JTree;
import javax.swing.MenuElement;
import javax.swing.MenuSelectionManager;
import javax.swing.SwingConstants;
import javax.swing.Timer;
import javax.swing.event.MenuKeyEvent;
import javax.swing.event.MenuKeyListener;

/**
 * Context Menu for use on the configuration tree.
 * Enables adding all applicable element types or removing applicable elements
 */
public class ModuleMenu extends JPopupMenu {
	private static final Logger logger = LoggerFactory.getLogger(ModuleMenu.class);

	private final JMenu glueMenu;
	private final JMenu procMenu;
	private final JMenu periMenu;
	private final IApplicationConfiguration application;
	private final JTree tree;
	private final IDocumentHandler documentHandler;
	private final JMenu busBreakoutMenu;
	private JMenuItem title;
	private JMenuItem deleteItem;

	private IConfigurationElement<?, ?, ?> currentElement;

	private ModulePopup popup;

	public ModuleMenu(IApplicationConfiguration application, JTree tree, Library library, IDocumentHandler documentHandler) {
		this.application = application;
		this.tree = tree;
		this.documentHandler = documentHandler;

		title = new JMenuItem();
		title.setEnabled(false);
		periMenu = new JMenu("Peripheral modules");
		procMenu = new JMenu("Subsystem modules");
		glueMenu = new JMenu("Common modules");
		busBreakoutMenu = new JMenu("Bus breakouts");
		deleteItem = new JMenuItem(new RemoveModuleAction(this));

		add(title);
		addSeparator();
		add(procMenu);
		add(periMenu);
		add(glueMenu);
		add(busBreakoutMenu);
		addSeparator();
		add(deleteItem);
		addSeparator();

		addModules(library.getPeripheralDescriptionsByCategory(), periMenu);
		addModules(library.getSubsystemDescriptionsByCategory(), procMenu);
		addModules(library.getCommonDescriptionsByCategory(), glueMenu);
		addModules(library.getAvailableBuses(), library, busBreakoutMenu);
	}

	public IDocumentHandler getDocumentHandler() {
		return documentHandler;
	}

	private void addModules(Collection<BusDescription> busDescriptions, Library library, JMenu menu) {
		busDescriptions.stream().sorted(Comparator.comparing(BusDescription::getName)).forEach(busDescription -> {
			JMenu busMenu = new JMenu(busDescription.getName());
			JMenuItem item1 = busMenu.add(new AddBusBreakoutAction(busDescription, Bus.Role.Master, this, library::getBusBreakoutDescription));
			JMenuItem item2 = busMenu.add(new AddBusBreakoutAction(busDescription, Bus.Role.Slave, this, library::getBusBreakoutDescription));
			busMenu.add(item1);
			busMenu.add(item2);
			busBreakoutMenu.add(busMenu);
		});
	}

	private void showPopupForItem(JMenuItem item) {
		ModuleDescription moduleDescription = ((AddModuleAction<?>) item.getAction()).getModuleDescription();

		//Already showing the right item?
		if (popup!=null && popup.isVisible() && popup.getModuleDescription()==moduleDescription) {
			return;
		}

		//Height is not correctly calculated if popup was closed once.
		//make a new one in that case
		if (popup==null || !popup.isVisible()) {
			popup = new ModulePopup(application);
		}

		popup.setModuleItem(moduleDescription,item);
		popup.showPopup();
	}

	private void menuSelectionChanged() {
		MenuElement[] path = MenuSelectionManager.defaultManager()
				.getSelectedPath();

		if (path.length>0) {
			Component c = path[path.length-1].getComponent();
			if (!(c instanceof JPopupMenu) && (c instanceof JMenuItem) ) {
				if (((JMenuItem) c).getAction()!=null) {
					showPopupForItem((JMenuItem) c);
					return;
				}
			}
		}

		if (popup!=null) {
			popup.startTimeout();
		}
	}

	private <T extends ModuleDescription> void addModules(Map<String, List<T>> modulesByCategory, JMenu menu) {
		modulesByCategory.entrySet().stream().sorted(Comparator.comparing(Map.Entry::getKey)).forEach(entry -> {
			JMenu catMenu = new JMenu(entry.getKey());
			entry.getValue().stream().sorted(Comparator.comparing(ModuleDescription::getName)).forEach(moduleDescription -> {
				JMenuItem item = catMenu.add(new AddModuleAction<>(moduleDescription, this));

				item.addMouseListener(new MouseAdapter() { //TODO: needed? (multiple instances)
					Timer timer = new Timer(250, evt -> showPopupForItem(item));

					@Override
					public void mouseEntered(MouseEvent e) {
						timer.setRepeats(false);
						timer.restart();
					}

					@Override
					public void mouseExited(MouseEvent e) {
						timer.stop();
						if (popup != null) {
							popup.startTimeout();
						}
					}
				});

				item.addMenuKeyListener(new MenuKeyListener() {
					@Override
					public void menuKeyTyped(MenuKeyEvent e) {
						menuSelectionChanged();
					}

					@Override
					public void menuKeyPressed(MenuKeyEvent e) {
						menuSelectionChanged();
					}

					@Override
					public void menuKeyReleased(MenuKeyEvent e) {
						menuSelectionChanged();
					}
				});
			});
			menu.add(catMenu);


		});
	}

	@Override
	public void show(Component invoker, int x, int y) {
		currentElement = (IConfigurationElement<?, ?, ?>) tree.getClosestPathForLocation(x, y).getLastPathComponent();

		update(currentElement);

		super.show(invoker, x, y);
	}

	/**
	 * Update which parts of our menu are shown
	 * @param element the entry that was clicked on
	 */
	private void update(IConfigurationElement element) {
		boolean canContainSubsystem = true;
		boolean canContainPeri = true;
		boolean canContainCommon = true;
		boolean removable = true;

		switch (element.getType()) {
			case Subsystem:
				canContainSubsystem = false;
				break;
			case Configuration:
				removable = false;
				canContainPeri = false;
				break;
			default:
				canContainPeri = false;
				canContainCommon = false;
				canContainSubsystem = false;
		}

		periMenu.setEnabled(canContainPeri);
		procMenu.setEnabled(canContainSubsystem);
		glueMenu.setEnabled(canContainCommon);
		busBreakoutMenu.setEnabled(canContainCommon);

		deleteItem.setEnabled(removable);

		title.setText(String.format("%s: %s", element.getType(), element.getName()));

		updateModules();
	}

	private void updateModulesForMenu(JMenu menu) {
		for (Component periGroup : menu.getMenuComponents()) {
			JMenu periGroupMenu = (JMenu) periGroup;

			Component[] groupPeris = periGroupMenu.getMenuComponents();

			for (Component groupPeri : groupPeris) {

				JMenuItem currentItem = (JMenuItem) groupPeri;
				AddModuleAction currentAction = (AddModuleAction) currentItem.getAction();

				if(currentAction.getModuleDescription().supports(documentHandler.getDocument().getTarget().getCurrentConfiguration())){
					groupPeri.setEnabled(true);
				}
				else{
					groupPeri.setEnabled(false);
				}
			}
		}
	}

	/**
	 * Enable only those modules in the menu which are supported on the target
	 */
	public void updateModules() {
		updateModulesForMenu(periMenu);
		updateModulesForMenu(procMenu);
		updateModulesForMenu(glueMenu);
	}

	public IConfigurationElement<?, ?, ?> getCurrentElement() {
		return currentElement;
	}


	private static class ModulePopup extends JidePopup {
		final BannerPanel panel;
		final JButton docuButton;
		private final IApplicationConfiguration application;
		private Path docsFile;
		private JMenuItem item;

		public ModuleDescription getModuleDescription() {
			return moduleDescription;
		}

		private ModuleDescription moduleDescription;

		public ModulePopup(IApplicationConfiguration application) {
			this.application = application;
			panel = new BannerPanel();
			panel.setTitleIconLocation(SwingConstants.WEST);

			docuButton = new JButton();
			docuButton.setText("Show Documentation");
			docuButton.setPreferredSize(new Dimension(200, 50));

			//We are using a mouselistener.mousePressed instead of an actionListener, because the popup closes when
			//the user clicks, because its parent is the menu item, which in turn closes when the user clicks outside.
			//ActionPerformed therefore never gets fired, but mousePressed does.

			docuButton.addMouseListener(new MouseAdapter() {
				@Override
				public void mousePressed(MouseEvent e) {
					logger.info("opening documentation {}",docsFile);
					ProcessUtil.openWithDefaultProgram(docsFile,ModulePopup.this);
				}
			});

			this.getContentPane().setLayout(new BoxLayout(this.getContentPane(), BoxLayout.LINE_AXIS));

			this.setMinimumSize(new Dimension(700,50));
			this.getContentPane().add(panel);
			this.getContentPane().add(docuButton);


			this.packPopup();
		}

		public void setModuleItem(ModuleDescription moduleDescription, JMenuItem item) {
			this.moduleDescription = moduleDescription;
			this.item = item;

			docsFile = DocumentationLocator.locateDocumentationPdf(application, moduleDescription);

			docuButton.setVisible(docsFile !=null);

			setTimeout(0);
			panel.setTitle(moduleDescription.getLabel()+" ("+moduleDescription.getName()+")");
			panel.setSubtitle(moduleDescription.getDescription().replaceAll("\\s+"," "));
			panel.setTitleIcon(IconUtil.getIconForElement(moduleDescription.getType(), 48));

			packPopup();

			setOwner(item);
		}

		public void startTimeout() {
			setTimeout(100);
		}


		//Make popup sit right of owner
		protected Point getPopupLocation(Point point, Dimension size, Component owner) {
			Component actualOwner = (owner != null) ? owner : getOwner();
			Dimension ownerSize = actualOwner != null ? actualOwner.getSize() : new Dimension(0, 0);
			if (size.width == 0) {
				size = this.getPreferredSize();
			}

			Point p = new Point(point.x + ownerSize.width + _insets.left, point.y +_insets.top+ownerSize.height/2-size.height/2);
			int left = p.x + size.width;
			int bottom = p.y + size.height;

			Rectangle screenBounds = PortingUtils.getContainingScreenBounds(new Rectangle(p, size), true);

			if (bottom > screenBounds.y + screenBounds.height) {
				p.y = point.y + _insets.top - size.height; // flip to upward
				if (p.y < screenBounds.y) {
					p.y = screenBounds.y;
				}
				if (isResizable()) {
					setupResizeCorner(Resizable.UPPER_RIGHT);
				}
			} else {
				if (isResizable()) {
					setupResizeCorner(Resizable.LOWER_RIGHT);
				}
			}

			Rectangle bounds = PortingUtils.containsInScreenBounds(actualOwner, new Rectangle(p, size));
			p.x = bounds.x;
			p.y = bounds.y;
			return p;
		}

		//Make minimum size actually do something
		protected void createWindow(Component owner, int x, int y) {
			super.createWindow(owner, x, y);
			_window.setMinimumSize(getMinimumSize());
		}

		@Override
		public boolean isVisible() {
			if (_window==null) {
				return false;
			}
			return _window.isVisible();
		}

	}
}
