/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.swing.BorderFactory;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;

/**
 * Base class for UI Elements watching / displaying information of a specific element of the Model hierarchy
 * You will receive all calls from {@link IConfigurationElement.IElementChangeListener}, if the contents change
 * or {@link #setHandlers(IConfigurationElement)} / {@link #clearHandlers()} if the element itself was swapped
 *
 * @author Ramon Wirsch
 */
public abstract class BaseElementEditorTab<T extends IConfigurationElement> extends JPanel {

	private static final Logger logger = LoggerFactory.getLogger(BaseElementEditorTab.class);
	public static final int UNIT_INCREMENT = 16;
	protected T element;
	private JScrollPane scrollPane;

	public BaseElementEditorTab() {
		setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
		scrollPane = new JScrollPane(this, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED, ScrollPaneConstants.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		scrollPane.getVerticalScrollBar().setUnitIncrement(UNIT_INCREMENT);
		scrollPane.getHorizontalScrollBar().setUnitIncrement(UNIT_INCREMENT);
	}

	/**
	 * Call this, if the underlying document or the watched element has changed
	 * @param newElement the new element that is represented here
	 */
	public final void setConfigurationElement(T newElement) {
		T oldElement = this.element;

		setConfigEl_Internal1(oldElement, newElement);

		this.element = newElement;

		setConfigEl_Internal2(oldElement, newElement);
	}

	/**
	 * Clean up after the old element
	 * @param oldElement still jc3 element
	 * @param newElement new contender
	 */
	protected void setConfigEl_Internal1(T oldElement, T newElement) {
		if (oldElement == newElement) {
			logger.warn("Ignored newElement. Same as before.");

			return;
		}

		logger.trace("Updating Element");


		if (oldElement != null) {
			detach(oldElement);
		}
	}

	protected abstract void detach(T oldElement);

	/**
	 * Register us on the new element
	 * @param oldElement previous element
	 * @param newElement jc3 element
	 */
	protected void setConfigEl_Internal2(T oldElement, T newElement) {
		if (newElement != null) {
			setHandlers(newElement);
			attach(newElement);
		} else {
			clearHandlers();
		}
	}

	protected abstract void attach(T newElement);

	/**
	 * Our element has changed, swap out / update handlers to be in sync with the new element
	 * @param element the already accepted element
	 */
	protected abstract void setHandlers(T element);

	/**
	 * The jc3 element has been removed, terminate all connections with it
	 */
	protected abstract void clearHandlers();

	public JScrollPane getScrollPane() {
		return scrollPane;
	}
}
