/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.attributes.IAttributeWatcher;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;

/**
 * Advanced version of {@link BaseElementEditorTab} that eases implementing UI for a variable amount of attributes
 * on a specific Model element. Instead of the generic callbacks you will need to provide a list of attribute handlers
 * and a list of elements that apply handlers.
 * Both will be matched up in order.
 *
 *
 * @author Ramon Wirsch
 */
public abstract class BaseVariableElementEditorTab<E extends IConfigurationElement, W extends IAttributeWatcher<? super H>, H> extends BaseElementEditorTab<E> {

	private static final Logger logger = LoggerFactory.getLogger(BaseVariableElementEditorTab.class);

	private List<W> attribComponents = new ArrayList<>();
	private List<H> handlers = new ArrayList<>();

	/**
	 * Hook in {@link #updateStructureInternal(IConfigurationElement, IConfigurationElement)}, before setConfigEl_Int2
	 * @param oldElement previous element
	 * @param newElement jc3 element
	 */
	@Override
	protected void setConfigEl_Internal2(E oldElement, E newElement) {
		updateStructureInternal(oldElement, newElement);

		super.setConfigEl_Internal2(oldElement, newElement);
	}

	private void updateStructureInternal(E oldElement, E newElement) {
		List<H> newHandlers = updateHandlers(handlers, newElement);

		if (newHandlers == null || (newHandlers != handlers || newHandlers.size() != handlers.size()) || requiresStructuralChange(oldElement, handlers, newElement, newHandlers)) {
			attribComponents = updateStructure(oldElement, handlers, newElement, newHandlers);
		}

		handlers = newHandlers;
	}

	/**
	 * Called while the element is being swapped
	 * update the list of handlers as necessary
	 *
	 *
	 * @param oldHandlers old list of handlers
	 * @param newElement new element
	 * @return old handlers or a new list or null.
	 */
	protected abstract List<H> updateHandlers(List<H> oldHandlers, E newElement);

	/**
	 * Called if the structue is required to change.
	 * Update your variable UI according to the new list of handlers.
	 * Return the List of Watchers that matches up 1:1 with the List of handlers
	 *
	 * @param oldElement for easy access, to allow reusing parts of old UI
	 * @param oldHandlers for easy access, to allow reusing parts of old UI
	 * @param newElement for easy access
	 * @param newHandlers for easy access
	 * @return a new List of an updated version
	 */
	protected abstract List<W> updateStructure(E oldElement, List<H> oldHandlers, E newElement, List<H> newHandlers);

	/**
	 * Called if this class is in doubt if the UI needs changing. (If the handler count changed, a UI change is always needed)
	 * If structural change is required, a call to {@link #updateStructure(IConfigurationElement, List, IConfigurationElement, List)} will follow.
	 * If uncertain return true
	 *
	 * @param oldElement for easy access
	 * @param oldHandlers for easy access
	 * @param newElement for easy access
	 * @param newHandlers for easy access
	 * @return whether or not the ui structure (fields that use handlers) need to change in response to the new model element
	 */
	protected abstract boolean requiresStructuralChange(E oldElement, List<H> oldHandlers, E newElement, List<H> newHandlers);

	public List<W> getWatchers() {
		return attribComponents;
	}

	public List<H> getHandlers() {
		return handlers;
	}

	@Override
	protected void setHandlers(E element) {
		if (handlers.size() != attribComponents.size()) {
			throw new RuntimeException("handlers do not match components! " + this);
		}

		for (int i = 0; i < attribComponents.size(); ++i) {
			attribComponents.get(i).setHandler(handlers.get(i));
		}
	}

	@Override
	protected void clearHandlers() {
		for (W attribComponent : attribComponents) {
			attribComponent.setHandler(null);
		}
	}

	protected void updateAllFields() {
		logger.trace("Updating Values");

		for (W field : attribComponents) {
			field.update();
		}
	}

	protected void forceUpdateHandlers() {
		updateStructureInternal(element, element);
		setHandlers(element);
	}
}
