/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


package de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes;

import de.tu_darmstadt.rs.process.ProcessUtil;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.fileHandling.DocumentationLocator;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.AddressSpacePane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.FirmwarePane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.GeneralPane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.GenericAttributePane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.IoPinConnectionPane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.IoPinPane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.ParameterPane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.PortConnectionPane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes.editorTabs.TargetOptionsPane;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.resources.IconUtil;

import com.jidesoft.dialog.BannerPanel;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.nio.file.Path;
import java.util.HashMap;
import java.util.Map;
import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;
import javax.swing.SwingConstants;

/**
 * Panel that shows configuration of  a single module
 */
public class ModuleEditorPane extends JPanel {

	private static final Logger logger = LoggerFactory.getLogger(ModuleEditorPane.class);

	// Tab Masks in order of appearance
	private static final int TAB_GENERAL = 0x1;
	private static final int TAB_TARGET = 0x2;
	private static final int TAB_PARAMETERS = 0x4;
	private static final int TAB_PIN_CONNECTIONS = 0x8;
	private static final int TAB_PORT_CONNECTIONS = 0x10;
	private static final int TAB_IO_CONF = 0x20;
	//	private static final int TAB_REGISTERS = 0x20;
	private static final int TAB_FIRMWARE = 0x40;
	private static final int TAB_ADDRESS_SPACE = 0x80;
	private static final int TAB_DEBUG = 0x100;

	// Tab titles same order
	private static final String[] TAB_TITLES = new String[]{
		"General",
		"Target",
		"Parameters",
		"Connections",
		"Connections",
		"I/O-Configuration",
		"Firmware",
		"Address Space",
		"Debug"
	};

	/**
	 * Mapping of ElementTypes to all tabs that are supported
	 */
	private static final Map<IConfigurationElement.ElementType, Integer> enabledTabs = new HashMap<>();

	static {
		enabledTabs.put(IConfigurationElement.ElementType.Configuration, TAB_GENERAL | TAB_TARGET | TAB_PIN_CONNECTIONS | TAB_IO_CONF | TAB_DEBUG);
		enabledTabs.put(IConfigurationElement.ElementType.Subsystem, TAB_GENERAL | TAB_PARAMETERS | TAB_PORT_CONNECTIONS | TAB_ADDRESS_SPACE | TAB_DEBUG);
		enabledTabs.put(IConfigurationElement.ElementType.CommonModule, TAB_GENERAL | TAB_PARAMETERS | TAB_PORT_CONNECTIONS | TAB_DEBUG);
		enabledTabs.put(IConfigurationElement.ElementType.Peripheral, TAB_GENERAL | TAB_PARAMETERS | TAB_PORT_CONNECTIONS | TAB_DEBUG);
		enabledTabs.put(IConfigurationElement.ElementType.MemoryModule, TAB_GENERAL | TAB_PARAMETERS | TAB_PORT_CONNECTIONS | TAB_FIRMWARE | TAB_DEBUG);
	}

	private final BannerPanel title;
	private final JButton docButton;
	private Path docFile;
	private final IDocumentHandler documentHandler;
	private final IApplicationConfiguration application;
	private final IConfigurationDocument document;

	/**
	 * All Tab Panes in order of titles and masks
	 */
	private final BaseElementEditorTab[] tabPanes;

	private JTabbedPane tabs;
	private IConfigurationElement selectedElement;
	private JPanel titlePane;


	public ModuleEditorPane(IDocumentHandler documentHandler, IApplicationConfiguration application) {
		this.documentHandler = documentHandler;
		this.document = documentHandler.getDocument();
		this.application = application;

		setBorder(BorderFactory.createEmptyBorder(3, 3, 3, 3));
		setLayout(new BorderLayout());

		title = new BannerPanel();

		titlePane = new JPanel();
		titlePane.setLayout(new BoxLayout(titlePane, BoxLayout.LINE_AXIS));
		docButton = new JButton("Show Documentation");
		docButton.setPreferredSize(new Dimension(200, 50));
		docButton.setVisible(false);
		docButton.addActionListener(e -> {
			ProcessUtil.openWithDefaultProgram(docFile, this);
		});

		titlePane.add(title);
		titlePane.add(docButton);
		add(titlePane, BorderLayout.NORTH);

		tabs = new JTabbedPane();
		tabs.setName("tabs");

		tabs.setBorder(BorderFactory.createEmptyBorder());
		add(tabs, BorderLayout.CENTER);

		tabPanes = new BaseElementEditorTab[]{
			new GeneralPane(),
			new TargetOptionsPane(this.application),
			new ParameterPane(this.application),
			new IoPinConnectionPane(documentHandler),
			new PortConnectionPane(documentHandler),
			new IoPinPane(),
			new FirmwarePane(),
			new AddressSpacePane(),
			new GenericAttributePane()
		};
	}

	private void updateActiveTabs() {
		tabs.removeAll();


		forEachActiveTab((index, tab) -> {
			tabs.add(TAB_TITLES[index], tab.getScrollPane());
			tab.setConfigurationElement(selectedElement);
		});
	}

	void onElementNameChange(IConfigurationElement<?, ?, ?> element) {
		if (selectedElement == element) {
			updateElementTitle();
		}
	}

	private void forEachActiveTab(TabConsumer forEach) {
		for (int tabMask = enabledTabs.get(selectedElement.getType()), i = 0; tabMask > 0; tabMask >>= 1, ++i) {
			if ((tabMask & 1) > 0) {
				forEach.consume(i, tabPanes[i]);
			}
		}
	}

	private void updateElementTitle() {
		String moduleName = "";
		String description = "";
		boolean dokuAvailable = false;
		if (selectedElement instanceof IModuleConfiguration) {
			IModuleDescription moduleDescription = ((IModuleConfiguration) selectedElement).getModuleDescription();
			moduleName = " (" + moduleDescription.getLabel() + ")";
			description = moduleDescription.getHeader().getDescription();

			docFile = DocumentationLocator.locateDocumentationPdf(application,moduleDescription);
			dokuAvailable = docFile!=null;
		}
		docButton.setVisible(dokuAvailable);

		title.setTitle(String.format("%s: %s%s", selectedElement.getType(), selectedElement.getName(), moduleName));
		title.setSubtitle(description);
		title.setTitleIcon(IconUtil.getIconForElement(selectedElement.getType(), 48));
		title.setTitleIconLocation(SwingConstants.WEST);
	}

	public void setSelectedElement(IConfigurationElement selectedElement) {
		int lastTab = tabs.getSelectedIndex();
		logger.info("Element changed to {}", selectedElement);
		this.selectedElement = selectedElement;
		updateElementTitle();
		updateActiveTabs();

		if (lastTab != -1 && tabs.getTabCount() > lastTab) {
			tabs.setSelectedIndex(lastTab);
		} else if (isActive(TAB_PARAMETERS)) {
			tabs.setSelectedComponent(tabPanes[2].getScrollPane());
		} else {
			tabs.setSelectedIndex(0);
		}
	}

	private boolean isActive(int tabMask) {
		return (enabledTabs.get(selectedElement.getType()) & tabMask) != 0;
	}

	private interface TabConsumer {

		void consume(int index, BaseElementEditorTab tab);
	}
}
