/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.ui.panes;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPluginEngine;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IScriptManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.DocumentAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.JConfigAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.components.AutoCompletionIntelliHints;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;

/**
 * Created by Ramon on 25.06.2016.
 */
public class ScriptReplPane extends JSplitPane implements IDocumentHandler.IPluginEngineListener {

	private static final Logger logger = LoggerFactory.getLogger(ScriptReplPane.class);

	private final JTextArea output;
	private final JTextArea input;
	private final JConfigAction evalAction;
	private final JConfigAction helpAction;
	private final IDocumentHandler documentHandler;
	private IScriptManager scriptManager;
	private LinkedList<String> history = new LinkedList<>();
	private int currentShownHistory = -1;
	private AutoCompletionIntelliHints intelliHints;

	public ScriptReplPane(IDocumentHandler documentHandler) {
		super(VERTICAL_SPLIT);
		this.documentHandler = documentHandler;

		output = new JTextArea();
		//		output.setPreferredSize(new Dimension());
		setTopComponent(new JScrollPane(output)); // ausgabe
		output.setEditable(false);
		output.setLineWrap(true);
		JPanel bottom = new JPanel(new BorderLayout());
		bottom.setPreferredSize(new Dimension(0, 20));
		input = new JTextArea();
		input.addKeyListener(new KeyAdapter() {
			@Override
			public void keyPressed(KeyEvent e) {

				if ((e.getKeyCode() == KeyEvent.VK_ENTER) && ((e.getModifiers() & KeyEvent.CTRL_MASK) != 0)) {
					eval();
				} else if ((e.getKeyCode() == KeyEvent.VK_TAB) || ((e.getKeyCode() == KeyEvent.VK_SPACE) && ((e.getModifiers() & KeyEvent.CTRL_MASK) != 0))) {
					e.consume();
					if (intelliHints.isHintsPopupVisible()) {
						intelliHints.setHintsEnabled(false);
					}
					autocomplete();
				} else if (e.getKeyCode() == KeyEvent.VK_UP) {
					if (!intelliHints.isHintsPopupVisible() && input.getLineCount() <= 1) {
						showHistory(1);
					}
				} else if (e.getKeyCode() == KeyEvent.VK_DOWN) {
					if (!intelliHints.isHintsPopupVisible() && input.getLineCount() <= 1) {
						showHistory(-1);
					}
				} else if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
					if (intelliHints.isHintsPopupVisible()) {
						intelliHints.setHintsEnabled(false);
					}
				}
			}
		});
		bottom.add(new JScrollPane(input), BorderLayout.CENTER);

		evalAction = new EvalScriptAction(documentHandler);
		helpAction = new HelpAction();

		JPanel buttons = new JPanel(new GridLayout(0, 2));
		buttons.add(new JButton(evalAction));
		buttons.add(new JButton(helpAction));

		bottom.add(buttons, BorderLayout.EAST);
		setBottomComponent(bottom);

		setResizeWeight(1.0);

		intelliHints = new AutoCompletionIntelliHints(input, new ArrayList<String>(), 0);

		documentHandler.addPluginEngineListener(this);
		if (documentHandler.getPluginEngine() != null) {
			onPluginEngineChanged(documentHandler.getPluginEngine()); // fake call so we notice if we're late to the party
		}

		//		//scriptRepl.setDividerSize(0);
		//		addComponentListener(new ComponentAdapter() {
		//			@Override
		//			public void componentResized(ComponentEvent e) {
		//				scriptRepl.setDividerLocation(0.5);
		//			}
		//		});
	}

	private void showHistory(int diff) {
		int max = history.size() - 1;
		int current = currentShownHistory + diff;
		current = Math.min(current, max);
		current = Math.max(current, -1);

		if (current < 0) {
			input.setText("");
		} else {
			input.setText(history.get(current));
		}

		currentShownHistory = current;
	}

	@Override
	public void onPluginEngineChanged(IPluginEngine pluginEngine) {
		Optional<IPluginEngine> engine = Optional.ofNullable(pluginEngine);

		Optional<IScriptManager> scriptManager = engine.flatMap(IPluginEngine::getScriptManager);

		if (!scriptManager.isPresent()) {
			logger.info("Scripting not available on current PluginEngine {}", documentHandler.getPluginEngine());
			output.setText("Scripting not available on current PluginEngine");
			input.setEnabled(false);
			evalAction.setEnabled(false);
			helpAction.setEnabled(false);
		} else {
			this.scriptManager = scriptManager.get();
			this.scriptManager.redirectStdIo(this::appendScriptOutput, this::appendScriptOutput, this::appendScriptOutput);
		}
	}

	private void appendScriptOutput(String s) {
		output.append(s);
	}

	private void eval() {
		currentShownHistory = -1; // reset history counter
		intelliHints.setHintsEnabled(false);

		String inputText = input.getText();
		history.addFirst(inputText);
		String text = inputText + "\n";
		output.append("> " + text);

		input.setText("");

		output.setCaretPosition(output.getDocument().getLength());

		try {
			Optional<Object> result = scriptManager.eval(text);
			if (result.isPresent()) {
				Object r = result.get();
				logger.info("Script returned: {}", r);
				output.append(String.format("Script returned: %s\n", r));
			}
		} catch (Exception ex) {
			logger.error("Terminal input was wrong", ex);
			output.append("Error: " + ex.getMessage() + " \n");
		}
	}

	private void autocomplete() {

		String inputText = input.getText();
		int cursorPos = input.getCaretPosition();

		if (cursorPos < inputText.length()) {
			return;
		}

		List<String> candidates = new ArrayList<String>();
		int identifierStart = scriptManager.autocomplete(inputText, cursorPos, candidates);

		//no candidates found
		if (candidates.size() == 0) {
			return;
		}

		//one candidate found
		if (candidates.size() == 1) {
			input.setText(input.getText().substring(0, identifierStart) + candidates.get(0));
			return;
		}

		//multiple candidates found
		String candidateIntersection = getCandidateIntersection(candidates);

		//if nothing to append, show intelliHints
		if (candidateIntersection.equals(input.getText().substring(identifierStart))) {

			//get extended list of possible candidates
			if (identifierStart > 0) {
				List<String> extendedCandidates = new ArrayList<String>();
				scriptManager.autocomplete(input.getText().substring(0, identifierStart), identifierStart, extendedCandidates);
				if (extendedCandidates.size() > candidates.size()) {
					candidates = extendedCandidates;
				}
			}

			intelliHints.setIdentifierStart(identifierStart); //TODO resize after reshow?
			intelliHints.setCompletionList(candidates);
			intelliHints.setHintsEnabled(true);
		}
		input.setText(input.getText().substring(0, identifierStart) + candidateIntersection);
	}

	private String getCandidateIntersection(List<String> candidates) {

		boolean candidatesIntersect = true;
		int intersectionIndex;
		for (intersectionIndex = 0; candidatesIntersect; intersectionIndex++) {
			Character oldChar = null;

			for (String candidate : candidates) {

				if (candidate.length() <= intersectionIndex) {
					candidatesIntersect = false;
					break;
				}

				Character currentChar = candidate.charAt(intersectionIndex);
				if (oldChar != null && currentChar != oldChar) {
					candidatesIntersect = false;
					break;
				}

				oldChar = currentChar;
			}
		}

		return intersectionIndex == 0? "" : candidates.get(0).substring(0, (intersectionIndex - 1));
	}

	/**
	 * Created by melanieb on 19.04.16.
	 */
	private class EvalScriptAction extends DocumentAction {

		public EvalScriptAction(IDocumentHandler documentHandler) {
			super("Eval", "Evaluates script",
					null, documentHandler);
		}

		/**
		 * Wenn noch cjconfig dazu kommt dafür einen passenden befehl dann rein machen der dies macht
		 */
		@Override
		public void actionPerformed(ActionEvent e) {
			eval();
		}
	}


	private class HelpAction extends JConfigAction {

		public HelpAction() {
			super("Help", "Shows quick help",
					null);
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			if (scriptManager != null) {
				String helpMessage = scriptManager.getHelpMessage();
				output.append((helpMessage == null)? "No help available\n" : helpMessage);
			}
		}
	}
}
