/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.ui.schematic;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.schematic.ExportSvgAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.schematic.LoadStyleXMLAction;
import de.tu_darmstadt.rs.spartanmc.jconfig.ui.actions.schematic.ResetStyleAction;

import org.apache.batik.svggen.SVGGraphics2D;
import org.apache.batik.swing.JSVGCanvas;
import org.w3c.dom.Element;
import org.w3c.dom.svg.SVGDocument;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Insets;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;

public class SchematicWindow extends JFrame implements WindowListener, ComponentListener, IDocumentHandler.IDocumentHandlerListener, IConfigurationDocument.IConfigurationHierarchyListener {

	private final IDocumentHandler documentHandler;
	private final IApplicationConfiguration application;
	private File pathToXMLStylesheet;
	private ResetStyleAction resetStyleAction;
	private SVGDocument svgDoc;

	//All window elements are listed here
	private JLabel label;
	private JSVGCanvas canvas;
	private JPanel toolBar;
	private JButton exportButton;
	private JButton loadStylesheetButton;
	private JButton resetStylesheetButton;


	//constructor
	//sourceElement
	//Action: The actions activating the schematic window; used to deactivate the actions
	//when the window is shown
	//
	public SchematicWindow(IApplicationConfiguration application, IDocumentHandler documentHandler, Component parent) {
		super("Schematic");
		this.documentHandler = documentHandler;
		this.application = application;

		addWindowListener(this);
		addComponentListener(this);
		setDefaultCloseOperation(DISPOSE_ON_CLOSE);
		createGUI();
		setVisible(true);

		documentHandler.addDocumentHandlerListener(this);
		documentHandler.getDocument().addHierarchyListener(this);

		repaintSchematic();
	}

	private void createGUI() {
		exportButton = new JButton(new ExportSvgAction(this));
		exportButton.setMargin(new Insets(0, 0, 0, 0));
		loadStylesheetButton = new JButton(new LoadStyleXMLAction(this));
		loadStylesheetButton.setMargin(new Insets(0, 0, 0, 0));
		resetStyleAction = new ResetStyleAction(this);
		resetStylesheetButton = new JButton(resetStyleAction);
		resetStylesheetButton.setMargin(new Insets(0, 0, 0, 0));
		toolBar = new JPanel();
		toolBar.setLayout(new FlowLayout(FlowLayout.LEFT, 3, 1));
		toolBar.add(exportButton);
		toolBar.add(loadStylesheetButton);
		toolBar.add(resetStylesheetButton);
		canvas = new JSVGCanvas(null, false, false);
		getContentPane().add(canvas);
		getContentPane().add(toolBar, BorderLayout.NORTH);
		setSize(1024, 768);
		setLocation(100, 100);//?

	}

	@Override
	public void componentResized(ComponentEvent e) {
		// react on changed window size and display previously hidden parts
		if (svgDoc != null) {
			canvas.setSVGDocument(svgDoc);
		}
	}

	public void componentMoved(ComponentEvent e) {
	}

	public void componentShown(ComponentEvent e) {
	}

	public void componentHidden(ComponentEvent e) {
	}

	public void repaintSchematic() {
		if (isVisible() && documentHandler.hasLoadedDocument()) {
			SchematicInformation schematicInformation = new SchematicInformation(documentHandler.getDocument(), application.getLibrary());
			reloadStylesheet(schematicInformation);
			Object[] svgInformation = schematicInformation.getSVGGraphicsInformation();
			if (svgInformation[0] != null) {
				svgDoc = (SVGDocument) svgInformation[0];
				SVGGraphics2D g = (SVGGraphics2D) svgInformation[1];
				Element root = svgDoc.getDocumentElement();
				g.getRoot(root);
				canvas.setSVGDocument(svgDoc);
			}
		}
	}

	private void reloadStylesheet(SchematicInformation schematicInformation) {
		if (pathToXMLStylesheet != null) {
			try {
				schematicInformation.loadVariables(pathToXMLStylesheet);
			} catch (Exception ex) {
				JOptionPane.showMessageDialog(null, "An error occured while trying to read the style file. Using default values instead. " + ex.getMessage(), "Reading error", JOptionPane.ERROR_MESSAGE);
				pathToXMLStylesheet = null;
				resetStyleAction.setEnabled(false);
			}
		}
	}

	public void saveSchematicAs(FileWriter fileWriter) throws IOException {
		if (documentHandler.hasLoadedDocument()) {
			SchematicInformation schematicInformation = new SchematicInformation(documentHandler.getDocument(), application.getLibrary());//TODO impl
			reloadStylesheet(schematicInformation);
			Object[] svgInformation = schematicInformation.getSVGGraphicsInformation();
			if (svgInformation[0] != null) {
				SVGDocument svgDoc = (SVGDocument) svgInformation[0];
				SVGGraphics2D g = (SVGGraphics2D) svgInformation[1];
				g.stream(fileWriter, false);
			}
		}
	}

	public void setXMLStylesheet(File filename) {
		pathToXMLStylesheet = filename;
		if (pathToXMLStylesheet != null) {
			resetStyleAction.setEnabled(true);
		}
		repaintSchematic();
	}

	public void windowOpened(WindowEvent e) {
	}

	public void windowClosing(WindowEvent e) {
	}

	public void windowClosed(WindowEvent e) {
		documentHandler.removeDocumentHandlerListener(this);
	}

	public void windowIconified(WindowEvent e) {
	}

	public void windowDeiconified(WindowEvent e) {
	}

	public void windowActivated(WindowEvent e) {
	}

	public void windowDeactivated(WindowEvent e) {
	}

	@Override
	public void onDocumentSwitched(boolean valid) {
		if (!valid) {
			this.dispose(); // without a valid document no schematic can be painted
		}
	}

	private void onHierarchyChanged() {
		repaintSchematic();
	}

	@Override
	public void onModuleAdded(IModuleConfiguration module) {
		onHierarchyChanged();
	}

	@Override
	public void onModuleRemoved(IConfigurationElement<?, ?, ?> parent, IModuleConfiguration module) {
		onHierarchyChanged();
	}

	@Override
	public void onElementNameChanged(IConfigurationElement<?, ?, ?> element) {
		onHierarchyChanged();
	}
}
