/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.fpga.resources;

import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.FpgaPrimitive;

/**
 * Represents an FPGA resource with the usage statistics
 */
public class FpgaResource implements Comparable<FpgaResource> {

	private int used;
	private int total;
	private FpgaPrimitive primitive;


	/**
	 * The constructor
	 * @param used the used resources
	 * @param total the total available resources
	 * @param primitive the exact FPGA primitive that this represents
	 */
	public FpgaResource(int used, int total, FpgaPrimitive primitive) {
		this.used = used;
		this.total = total;
		this.primitive = primitive;
	}

	/**
	 * The constructor neglecting total and percentage numbers
	 * @param used the used resources
	 * @param primitive the exact FPGA primitive that this represents
	 */
	public FpgaResource(int used, FpgaPrimitive primitive){
		this(used, 0, primitive);
	}

	public int getUsedResources() {
		return used;
	}

	public FpgaPrimitive getPrimitive() {
		return primitive;
	}

	public String getPrimitiveString() {
		return getPrimitive().getReadableString();
	}


	@Override
	public int hashCode() {
		int result = used;
		result = 31 * result + total;
		result = 31 * result + (primitive != null? primitive.hashCode() : 0);
		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		FpgaResource that = (FpgaResource) o;

		if (used != that.used) {
			return false;
		}
		if (total != that.total) {
			return false;
		}
		return primitive != null? primitive.equals(that.primitive) : that.primitive == null;
	}

	public String toString() {
		return primitive.getReadableString() + " used: " + used + " total: " + total + " percentage: " + (100f*used/total);
	}

	@Override
	public int compareTo(FpgaResource o) {
		return Integer.compare(primitive.getType().getOrder(),o.primitive.getType().getOrder());

	}


	public void incrementUsed(int count) {
		used += count;
	}
}
