/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.fpga.resources;

import de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives.FpgaPrimitive;

import com.google.common.collect.EnumMultiset;
import com.google.common.collect.Multiset;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class HierarchicalUsage implements Comparable<HierarchicalUsage> {

	private final Map<String, HierarchicalUsage> subDesigns = new HashMap<>();

	private final Map<FpgaPrimitive, FpgaResource> resources = new HashMap<>();

	public String getName() {
		return name;
	}

	private final String name;

	public HierarchicalUsage(String name) {
		this.name = name;
	}

	private FpgaResource getResource(FpgaPrimitive primitive) {
		return resources.computeIfAbsent(primitive, p -> new FpgaResource(0, p));
	}

	public void addInstance(String processedName, Multiset<? extends FpgaPrimitive> types) {
		final int slashPos = processedName.indexOf('/');
		if (slashPos == -1) {
			types.entrySet().forEach(t -> getResource(t.getElement()).incrementUsed(t.getCount()));

		} else {
			String subName = processedName.substring(0, slashPos);
			final String subsubName = processedName.substring(slashPos + 1, processedName.length());
			subDesigns.computeIfAbsent(subName, HierarchicalUsage::new).addInstance(subsubName, types);
		}
	}


	public <T extends Enum<T> & FpgaPrimitive> void addInstance(String processedName, T type) {
		addInstance(processedName, EnumMultiset.create(Collections.singletonList(type)));
	}

	public Collection<? extends HierarchicalUsage> getSubs() {
		final Stream<HierarchicalUsage> stream = subDesigns.values().stream().sorted();
		return stream.collect(Collectors.toList());
	}

	public int getRecursiveCount(FpgaPrimitive type) {
		int selfCount = getSelfCount(type);
		final int subCount = subDesigns.values().stream().mapToInt(s -> s.getRecursiveCount(type)).sum();
		return selfCount + subCount;
	}

	private int getSelfCount(FpgaPrimitive type) {
		return getResource(type).getUsedResources();
	}

	public Object[] toColumns(boolean includeName, List<FpgaPrimitive> types) {
		Stream<String> nameStream;
		if (includeName) {
			nameStream = Stream.of(name);
		} else {
			nameStream = Stream.empty();
		}
		return Stream.concat(
				nameStream,
				types.stream().map(this::getRecursiveCount)
		).toArray(Object[]::new);

	}

	@Override
	public int compareTo(HierarchicalUsage o) {
		return getName().compareTo(o.getName());
	}

	public Stream<FpgaPrimitive> getUsedPrimitives() {
		final Stream<FpgaPrimitive> subUsed = subDesigns.values().stream().flatMap(HierarchicalUsage::getUsedPrimitives);
		final Stream<FpgaPrimitive> ownUsed = resources.keySet().stream();
		return Stream.concat(subUsed, ownUsed).distinct();
	}

	public void addResource(FpgaResource res) {
		final FpgaResource previous = resources.put(res.getPrimitive(), res);
		if (previous != null) {
			throw new RuntimeException("Cannot replace resources! Tried to overwrite " + previous + " with " + res);
		}
	}
}
