/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.fpga.resources.fpgaPrimitives;

import de.tu_darmstadt.rs.fpga.resources.FpgaFamily;
import de.tu_darmstadt.rs.util.StreamUtil;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * represents the different types of resources on a FPGA
 */
public enum FpgaResourceType {
	Slice(1),
	LUT(2),
	Register(3),
	BlockRAM(4),
	DSP(5);

	int order;

	private List<StreamUtil.Pair<FpgaFamily, FpgaPrimitive>> blockRamNames = new ArrayList<>();
	private List<StreamUtil.Pair<FpgaFamily, FpgaPrimitive>> dspNames = new ArrayList<>();

	FpgaResourceType(int order) {

		this.order = order;

		FpgaFamily family = FpgaFamily.XC6SLX;
		blockRamNames.add(new StreamUtil.Pair<>(family, BlockRamPrimitives.RAMB8BWER));
		blockRamNames.add(new StreamUtil.Pair<>(family, BlockRamPrimitives.RAMB16BWER));
		dspNames.add(new StreamUtil.Pair<>(family, DspPrimitives.DSP48A1));

		family = FpgaFamily.XC7A;
		blockRamNames.add(new StreamUtil.Pair<>(family, BlockRamPrimitives.RAMB36E1FIFO36E1));
		blockRamNames.add(new StreamUtil.Pair<>(family, BlockRamPrimitives.RAMB18E1FIFO18E1));
		dspNames.add(new StreamUtil.Pair<>(family, DspPrimitives.DSP48E1));
	}

	/**
	 * gives a pattern to decode this resource for the given device
	 * @param device a fpga device
	 * @return a pattern to decode this resource for the given device
	 */
	public List<Pattern> getEvaluationReportPattern(FpgaFamily device) {
		return getResourcesForDevice(device).stream()
				.map(primitive -> Pattern.compile(" *Number of " + primitive.getReportSearchString() + ": *(\\d+,?\\d*) +out of +(\\d+,?\\d*) +(\\d+%)"))
				.collect(Collectors.toList());

	}

	/**
	 * Searches for a FPGA primitive in the given line for all available devices
	 * @param parsedLine a line from the report
	 * @return a FPGA primitive in the given line for all available devices
	 */
	public Optional<FpgaPrimitive> findPrimitiveInLine(String parsedLine) {
		return Arrays.stream(FpgaFamily.values())
				.flatMap(dev -> getResourcesForDevice(dev).stream())
				.filter(prim -> parsedLine.toLowerCase().contains(prim.getReportSearchString().toLowerCase()))
				.findAny();
	}


	public List<FpgaPrimitive> getResourcesForDevice(FpgaFamily device) {

		switch (this) {
			case BlockRAM:
				return filterForDevice(blockRamNames, device);
			case DSP:
				return filterForDevice(dspNames, device);
			case Slice:
				return Arrays.asList(NonChangingPrimitives.SLICE);
			case LUT:
				return Arrays.asList(NonChangingPrimitives.LUT);
			case Register:
				return Arrays.asList(NonChangingPrimitives.REGISTER);
			default:
				return null;
		}
	}

	private List<FpgaPrimitive> getResourcesForAllDevices() {
		return Arrays.stream(FpgaFamily.values()).map(dev -> getResourcesForDevice(dev)).flatMap(l -> l.stream()).collect(Collectors.toList());
	}

	public static List<FpgaPrimitive> getAllPrimitives() {
		return Arrays.stream(FpgaResourceType.values()).map(r -> r.getResourcesForAllDevices()).flatMap(r -> r.stream()).collect(Collectors.toList());
	}


	private List<FpgaPrimitive> filterForDevice(List<StreamUtil.Pair<FpgaFamily, FpgaPrimitive>> resourceNames, FpgaFamily device) {
		return resourceNames.stream()
				.filter(p -> p.getA() == device)
				.map(p -> p.getB())
				.collect(Collectors.toList());
	}

	public int getOrder() {
		return order;
	}
}

