/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml;

import de.tu_darmstadt.rs.jcommander.validators.DirectoryExistsValidator;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.CalculatedValue;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.DynamicParamRef;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.ParamRef;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.Constraint;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.KeepHierarchy;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.MixedParamRefs;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.TimingIgnore;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.VendorConstraint;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.BaseParameterGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.dynamic.DynamicParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.dynamic.DynamicParameterGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.dynamic.DynamicParameterImpl;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.dynamic.DynamicParamterGroupImpl;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.dynamic.DynamicClkPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.dynamic.DynamicClkPortImpl;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.dynamic.DynamicPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.dynamic.DynamicPortGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.dynamic.DynamicPortGroupImpl;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.dynamic.DynamicPortImpl;
import de.tu_darmstadt.rs.spartanmc.devxml.exceptions.LibraryException;
import de.tu_darmstadt.rs.spartanmc.devxml.exceptions.MissingModuleDescriptionException;
import de.tu_darmstadt.rs.spartanmc.devxml.files.base.XmlLoadedFile;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.BusBreakoutModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.CodeModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.CommonModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.DynamicStructureModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.SubsystemModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.jaxb.Unmarshaller;
import de.tu_darmstadt.rs.spartanmc.devxml.schemas.ResourceSchemaResolver;
import de.tu_darmstadt.rs.util.BuildInfo;
import de.tu_darmstadt.rs.util.FileFinder;
import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.StringSearchMatcher;

import com.beust.jcommander.Parameter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.xml.bind.JAXBException;
import javax.xml.bind.UnmarshalException;

import static java.util.stream.Collectors.toMap;

/**
 * Manages all targets, devices, modules.
 * Use the {@link Builder} to get an Instance
 */
public final class Library {

	public static final String INCLUDE_DIR = "include";
	private static final Logger logger = LoggerFactory.getLogger(Library.class);
	private static final String LIBDEVXML_VERSION = "libDevXml-3.0-alpha0";
	private static BuildInfo buildInfo;
	private final Map<String, DeviceDescription> devices;
	private final Map<String, TargetDescription> targets;
	private final Map<String, ModuleDescription> modules;
	private final Map<String, BusDescription> buses;
	private final Map<String, BusBreakoutModuleDescription> busBreakouts;
	private final Set<Path> invalidFiles;
	private final Set<Path> moduleSourceTrees;
	private final Set<Path> busSourceTrees;
	private final Set<Path> targetSourceTrees;
	private final Set<Path> deviceSourceTrees;

	private Library(
			Set<Path> deviceSourceTrees,
			Map<String, DeviceDescription> devices,
			Set<Path> targetSourceTrees,
			Map<String, TargetDescription> targets,
			Set<Path> moduleSourceTrees,
			Map<String, ModuleDescription> modules,
			Set<Path> busSourceTrees,
			Map<String, BusDescription> buses,
			Map<String, BusBreakoutModuleDescription> busBreakouts,
			Set<Path> invalidFiles) {
		this.deviceSourceTrees = deviceSourceTrees;

		this.devices = devices;
		this.targetSourceTrees = targetSourceTrees;
		this.targets = targets;
		this.moduleSourceTrees = moduleSourceTrees;
		this.modules = modules;
		this.busSourceTrees = busSourceTrees;
		this.buses = buses;
		this.busBreakouts = busBreakouts;
		this.invalidFiles = invalidFiles;
	}

	public static BuildInfo getBuildInfo() {
		if (buildInfo == null) {
			buildInfo = new BuildInfo(Library.class);
		}
		return buildInfo;
	}

	public static Path getCIncludeDir(ModuleDescription module) {
		return module.getSourceTree().resolveSibling(INCLUDE_DIR);
	}

	/**
	 * find the name prefix of a modules descriptions fileName
	 * Ex: (core).module.xml
	 *
	 * @param fileName a module descriptions file name
	 * @return empty optional if the filename has no prefix
	 */
	public static Optional<String> extractModuleXmlPrefix(String fileName) {
		Pattern pattern = Pattern.compile("^([^.]+)\\.module\\.xml$");
		Matcher matcher = pattern.matcher(fileName);
		if (matcher.matches()) {
			return Optional.of(matcher.group(1));
		} else {
			return Optional.empty();
		}
	}

	public Set<Path> getModuleSourceTrees() {
		return moduleSourceTrees;
	}

	public Set<Path> getBusSourceTrees() {
		return busSourceTrees;
	}

	public Set<Path> getTargetSourceTrees() {
		return targetSourceTrees;
	}

	public Set<Path> getDeviceSourceTrees() {
		return deviceSourceTrees;
	}

	public DeviceDescription getDeviceDescription(String name) {
		return devices.get(name);
	}

	public IModuleDescription getModuleDescription(String name) {
		return verifyIsType(modules.get(name), name, ModuleDescription.class);
	}

	public Map<String, List<PeripheralModuleDescription>> getPeripheralDescriptionsByCategory() {
		return modules.values().stream().filter(m -> m.getType() == IModuleDescription.ModuleType.Peripheral).map(m -> (PeripheralModuleDescription) m).collect(Collectors.groupingBy(p -> p.getHeader().getCategory()));
	}

	public Map<String, List<SubsystemModuleDescription>> getSubsystemDescriptionsByCategory() {
		return modules.values().stream().filter(m -> m.getType() == IModuleDescription.ModuleType.Subsystem).map(m -> (SubsystemModuleDescription) m).collect(Collectors.groupingBy(p -> p.getHeader().getCategory()));
	}

	public Map<String, List<ModuleDescription>> getCommonDescriptionsByCategory() {
		return modules.values().stream().filter(m -> m.getType() != IModuleDescription.ModuleType.Code && m.getType() != IModuleDescription.ModuleType.Subsystem && m.getType() != IModuleDescription.ModuleType.Peripheral).collect(Collectors.groupingBy(p -> p.getHeader().getCategory()));
	}

	public SubsystemModuleDescription getSubsystemDescription(String name) {
		return verifyIsType(modules.get(name), name, SubsystemModuleDescription.class);
	}

	private <T extends ModuleDescription> T verifyIsType(ModuleDescription module, String name, Class<T> clz) {
		if (module == null) {
			throw new MissingModuleDescriptionException("Module " + name + " does not exist");
		} else if (!clz.isAssignableFrom(module.getClass())) {
			throw new MissingModuleDescriptionException("Module " + module.getName() + " is of type " + module.getClass().getSimpleName() + ", required: " + clz.getSimpleName());
		} else {
			return (T) module;
		}
	}

	public CodeModuleDescription getCodeDescription(String name) {
		return verifyIsType(modules.get(name), name, CodeModuleDescription.class);
	}

	public PeripheralModuleDescription getPeripheralDescription(String name) {
		return verifyIsType(modules.get(name), name, PeripheralModuleDescription.class);
	}

	public CommonModuleDescription getCommonDescription(String name) {
		return verifyIsType(modules.get(name), name, CommonModuleDescription.class);
	}

	public MemoryModuleDescription getMemoryCommonDescription(String name) {
		return verifyIsType(modules.get(name), name, MemoryModuleDescription.class);
	}

	public DynamicStructureModuleDescription getDynamicCommonDescription(String name) {
		return verifyIsType(modules.get(name), name, DynamicStructureModuleDescription.class);
	}

	public BusBreakoutModuleDescription getBusBreakoutDescription(BusDescription busDescription, Bus.Role role) {
		return busBreakouts.computeIfAbsent(BusBreakoutModuleDescription.getModuleName(busDescription, role), name -> Builder.generateBusBreakout(busDescription, role));
	}

	public TargetDescription getTargetDescription(String name) {
		return targets.entrySet().stream()
				.filter(entry -> entry.getKey().equalsIgnoreCase(name))
				.map(Map.Entry::getValue)
				.collect(StreamUtil.exactlyOne());
	}

	/**
	 * Uses an advanced search to correct possible errors in the target name. For example My-Target will be found if searching for mytarget or mt.
	 *
	 * @param name of a target
	 * @return a target description of a matching target if only exactly one target is found.
	 */
	public TargetDescription getTargetDescriptionWithAdvancedSearch(String name) {
		return targets.entrySet().stream()
				.filter(entry -> StringSearchMatcher.match(entry.getKey(), name))
				.map(Map.Entry::getValue)
				.collect(StreamUtil.exactlyOne());
	}

	public BusDescription getBusDescription(String name) {
		return buses.get(name);
	}

	public Collection<BusDescription> getAvailableBuses() {
		return new ArrayList<>(buses.values());
	}

	public Collection<ModuleDescription> getAvailableModules() {
		return modules.values();
	}

	public Collection<DeviceDescription> getAvailableDevices() {
		return new ArrayList<>(devices.values());
	}

	public Collection<TargetDescription> getAvailableTargets() {
		return new ArrayList<>(targets.values());
	}

	public Collection<DeviceDescription> getSupportedDeviceDescriptions(TargetDescription targetDescription) {
		return targetDescription.getDevices().stream().map(TargetDescription.Device::getName).map(devices::get).collect(Collectors.toList());
	}

	public String getLibDevXmlVersion() {
		return LIBDEVXML_VERSION;
	}

	public Set<Path> getInvalidFiles() {
		return invalidFiles;
	}


	private Stream<String> getSourceTreeAsBuilderParameters(Set<Path> sourceTree, String parameterPrefix, DefaultSearchpathNames defaultSearchpath, Path spartanRoot) {
		Path defaultPath = spartanRoot != null ? defaultSearchpath.resolveDefaultSearchpath(spartanRoot) : null;
		return sourceTree.stream()
				.filter(p -> !p.equals(defaultPath))
				.map(p -> parameterPrefix+" "+p.toAbsolutePath().toString());
	}

	public String getSourceTreesAsBuilderParameters(Path spartanRoot) {
		return Stream.of(
				getSourceTreeAsBuilderParameters(busSourceTrees, "--include-buses", DefaultSearchpathNames.BUSES, spartanRoot),
				getSourceTreeAsBuilderParameters(deviceSourceTrees, "--include-devices", DefaultSearchpathNames.DEVICES, spartanRoot),
				getSourceTreeAsBuilderParameters(targetSourceTrees, "--include-targets", DefaultSearchpathNames.TARGETS, spartanRoot),
				getSourceTreeAsBuilderParameters(moduleSourceTrees, "--include-modules", DefaultSearchpathNames.MODULES, spartanRoot)
		)
				.flatMap(Function.identity())
				.collect(Collectors.joining(" "));
	}

	public interface IProgressIndicator {

		void setProgress(int progress);

		void setMaxProgress(int maxProgress);

		void stepProgress(int steps);

		void setProgressText(String fmt, Object... args);

		void progressComplete();
	}

	private enum DefaultSearchpathNames {
		DEVICES("devices"),
		TARGETS("targets"),
		MODULES("hardware"),
		BUSES("buses");

		private final String folderName;

		DefaultSearchpathNames(String folderName) {
			this.folderName = folderName;
		}

		public String getFolderName() {
			return folderName;
		}

		public Path resolveDefaultSearchpath(Path spartanRoot) {
			return spartanRoot.resolve("spartanmc").resolve(folderName);
		}
	}


	/**
	 * Searches and reads in all applicable descriptions across the configured searchpaths
	 */
	public static class Builder {

		public static final String DEVICE = "Device";
		public static final String TARGET = "Target";
		public static final String MODULE = "Module";
		public static final String BUS = "Bus";
		private static final Logger logger = LoggerFactory.getLogger(Builder.class);
		private static final String DEVICE_DESCR_PATTERN = "*.xml";
		private static final String TARGET_DESCR_PATTERN = DEVICE_DESCR_PATTERN;
		private static final String MODULE_FILE_PATTERN = "{*.module.xml,module.xml}";
		private static final String BUS_DESCR_PATTERN = DEVICE_DESCR_PATTERN;
		private static final Marker searchpathMarker = MarkerFactory.getMarker("Searchpath");
		private static final Marker fileMarker = MarkerFactory.getMarker("Files");
		private final Set<Path> deviceDescSearchpaths;
		private final Set<Path> targetDescSearchpaths;
		private final Set<Path> moduleDescSearchpaths;
		private final Set<Path> busDescSearchpaths;
		private IProgressIndicator progressIndicator;
		private Unmarshaller<ModuleDescription> moduleUnmarshaller;
		private Unmarshaller<DeviceDescription> deviceUnmarshaller;
		private Unmarshaller<TargetDescription> targetUnmarshaller;
		private Unmarshaller<BusDescription> busUnmarshaller;
		private Map<String, BusDescription> buses;
		private Map<String, DeviceDescription> devices;
		private Map<String, TargetDescription> targets;
		private boolean analyze = false;
		private boolean optimize = false;
		private boolean verify = true;
		private boolean generateBusBreakouts = false;

		private final Set<Path> invalidFiles;

		/**
		 * Searches and reads in all applicable descriptions across the configured searchpaths
		 */
		public Builder() {
			deviceDescSearchpaths = new HashSet<>();
			targetDescSearchpaths = new HashSet<>();
			moduleDescSearchpaths = new HashSet<>();
			busDescSearchpaths = new HashSet<>();
			invalidFiles = new HashSet<>();
		}

		private static <T extends XmlLoadedFile> T failMerge(T a, T b) {
			throw createDuplicateDescriptorError(a, b);
		}

		private static RuntimeException createDuplicateDescriptorError(XmlLoadedFile a, XmlLoadedFile b) {
			logger.error("Descriptor {} and {} both try to declare the same document {}", a.getSourceFile(), b.getSourceFile(), a.getName());
			return new IllegalStateException(String.format("Duplicate document %s", a.getName()));
		}

		private static BusBreakoutModuleDescription generateBusBreakout(BusDescription busDescription, Bus.Role role) {
			return new BusBreakoutModuleDescription(busDescription, role);
		}

		public Builder addDefaultSearchpaths(Path spartanRoot) {
			logger.info("Adding default Library search paths in {}", spartanRoot);
			addDeviceDescriptionSearchpath(DefaultSearchpathNames.DEVICES.resolveDefaultSearchpath(spartanRoot));
			addTargetDescriptionSearchpath(DefaultSearchpathNames.TARGETS.resolveDefaultSearchpath(spartanRoot));
			addModuleDescriptionSearchpath(DefaultSearchpathNames.MODULES.resolveDefaultSearchpath(spartanRoot));
			addBusDescriptionSearchpath(DefaultSearchpathNames.BUSES.resolveDefaultSearchpath(spartanRoot));

			return this;
		}

		private void addSearchpath(Path path, String type, Collection<Path> var) {
			if (Files.isDirectory(path)) {
				boolean success = var.add(path);
				if (success) {
					logger.debug(searchpathMarker, "Adding {} description location: {}", type, path);
				} else {
					logger.warn("{} description location already added: {}", type, path);
				}
			} else {
				throw new LibraryException(type + " description location " + path + " does not exist");
			}
		}

		@Parameter(names = {"-B", "--include-buses"}, description = "Add additional paths that will be searched for bus-descriptions", validateValueWith = DirectoryExistsValidator.class)
		public Builder addBusDescriptionSearchpath(Path path) {
			addSearchpath(path, BUS, busDescSearchpaths);
			return this;
		}

		@Parameter(names = {"-D", "--include-devices"}, description = "Add additional paths that will be searched for device-descriptions", validateValueWith = DirectoryExistsValidator.class)
		public Builder addDeviceDescriptionSearchpath(Path path) {
			addSearchpath(path, DEVICE, deviceDescSearchpaths);
			return this;
		}

		@Parameter(names = {"-T", "--include-targets"}, description = "Add additional paths that will be searched for target-descriptions", validateValueWith = DirectoryExistsValidator.class)
		public Builder addTargetDescriptionSearchpath(Path path) {
			addSearchpath(path, TARGET, targetDescSearchpaths);
			return this;
		}

		@Parameter(names = {"-I", "--include-modules"}, description = "Add additional paths that will be searched for module-descriptions", validateValueWith = DirectoryExistsValidator.class)
		public Builder addModuleDescriptionSearchpath(Path path) {
			addSearchpath(path, MODULE, moduleDescSearchpaths);
			return this;
		}

		/**
		 * If a progressIndicator is set, it will be notified of the loading progress, including short messages, the total number of files and progress reading those files in
		 *
		 * @param progressIndicator usually a wrapper around a JProgressBar and Label
		 * @return Builder
		 */
		public Builder setProgressIndicator(IProgressIndicator progressIndicator) {
			this.progressIndicator = progressIndicator;
			return this;
		}

		/**
		 * Synchronously build the library as configured
		 *
		 * @return Immutable Description Library
		 */
		public Library build() {
			if (!hasMinimumSearchpaths()) {
				throw new IllegalStateException("Missing Searchpaths");
			}

			try {
				logger.trace("searching for files...");
				Collection<SourceFile> deviceFiles = new HashSet<>();
				Collection<SourceFile> targetFiles = new HashSet<>();
				Collection<SourceFile> moduleFiles = new HashSet<>();
				Collection<SourceFile> busFiles = new HashSet<>();

				findLoadableFiles(deviceFiles, DEVICE_DESCR_PATTERN, deviceDescSearchpaths, DEVICE);
				findLoadableFiles(targetFiles, TARGET_DESCR_PATTERN, targetDescSearchpaths, TARGET);
				findLoadableFiles(moduleFiles, MODULE_FILE_PATTERN, moduleDescSearchpaths, MODULE);
				findLoadableFiles(busFiles, BUS_DESCR_PATTERN, busDescSearchpaths, BUS);

				logger.info("Found {} modules, {} devices, {} targets, {} buses",
						moduleFiles.size(), deviceFiles.size(), targetFiles.size(), busFiles.size());

				reportProgressBeginLoading(deviceFiles.size() + moduleFiles.size() + targetFiles.size() + busFiles.size());


				initializeUnmarshallers();


				buses = busFiles.stream().map(this::loadBus).filter(Objects::nonNull).collect(toMap(BusDescription::getName, Function.identity(), Builder::failMerge));
				devices = deviceFiles.stream().map(this::loadDevice).filter(Objects::nonNull).collect(toMap(DeviceDescription::getName, Function.identity(), Builder::failMerge));
				targets = targetFiles.stream().map(this::loadTarget).filter(Objects::nonNull).collect(toMap(TargetDescription::getName, Function.identity(), Builder::failMerge));
				Map<String, ModuleDescription> modules = moduleFiles.stream().map(this::loadModule).filter(Objects::nonNull).collect(toMap(ModuleDescription::getName, Function.identity(), Builder::failMerge));

				Map<String, BusBreakoutModuleDescription> busBreakouts;

				if (generateBusBreakouts) {
					busBreakouts = Collections.unmodifiableMap(buses.values().stream().flatMap(b -> Stream.of(generateBusBreakout(b, Bus.Role.Master), generateBusBreakout(b, Bus.Role.Slave))).collect(toMap(IModuleDescription::getName, Function.identity())));
				} else {
					busBreakouts = new HashMap<>();
				}

				logger.info("Loaded {} modules, {} devices, {} targets, {} buses, {} bus-breakouts",
						modules.size(),
						devices.size(),
						targets.size(),
						buses.size(),
						busBreakouts.size());
				reportProgressEndLoading();

				return new Library(Collections.unmodifiableSet(deviceDescSearchpaths), Collections.unmodifiableMap(devices),
						Collections.unmodifiableSet(targetDescSearchpaths), Collections.unmodifiableMap(targets),
						Collections.unmodifiableSet(moduleDescSearchpaths), Collections.unmodifiableMap(modules),
						Collections.unmodifiableSet(busDescSearchpaths), Collections.unmodifiableMap(buses), busBreakouts,
						invalidFiles);
			} catch (Exception e) {
				logger.error("Failed to build Library with dev:{}, targ:{}, bus:{}, mod:{}, verify:{}, analyze:{}, optimize:{}, pIndicator:{}, busBreakouts:{}", deviceDescSearchpaths, targetDescSearchpaths, busDescSearchpaths, moduleDescSearchpaths, verify, analyze, optimize, progressIndicator, generateBusBreakouts, e);
				throw new LibraryException("Failed to build Library", e);
			}
		}

		private void reportProgressEndLoading() {
			if (progressIndicator != null) {
				progressIndicator.progressComplete();
			}
		}

		private void reportProgressBeginLoading(int fileCount) {
			if (progressIndicator != null) {
				progressIndicator.setMaxProgress(fileCount);
				progressIndicator.setProgressText("loading modules, devices, targets ...");
			}
		}

		private void initializeUnmarshallers() throws JAXBException {
			moduleUnmarshaller = new Unmarshaller<>(ResourceSchemaResolver.getSchema("module-description.xsd"),
					new Class[]{
							CodeModuleDescription.class, CommonModuleDescription.class, PeripheralModuleDescription.class, SubsystemModuleDescription.class,
							CalculatedValue.Constant.class, CalculatedValue.Multiply.class, CalculatedValue.Add.class, ParamRef.class, CalculatedValue.AddressSize.class,
							DynamicParamRef.class, CalculatedValue.Invert.class, CalculatedValue.ClkPortRef.class, CalculatedValue.Divide.class, CalculatedValue.Switch.class,
							DynamicStructureModuleDescription.class, MemoryModuleDescription.class,
							DynamicParameterGroup.class, DynamicParamterGroupImpl.class,
							DynamicParameter.class, DynamicParameterImpl.class,
							DynamicPortGroup.class, DynamicPortGroupImpl.class,
							DynamicPort.class, DynamicPortImpl.class,
							DynamicClkPort.class, DynamicClkPortImpl.class,
							MixedParamRefs.class, Constraint.class,
							TimingIgnore.class, VendorConstraint.class,
							KeepHierarchy.class
					});
			deviceUnmarshaller = new Unmarshaller<>(ResourceSchemaResolver.getSchema("device-description.xsd"), DeviceDescription.class);
			targetUnmarshaller = new Unmarshaller<>(ResourceSchemaResolver.getSchema("target-description.xsd"), TargetDescription.class);
			busUnmarshaller = new Unmarshaller<>(ResourceSchemaResolver.getSchema("bus-description.xsd"),
					new Class[]{
							BusDescription.class, BusDescription.BusClkPort.class, BusPort.class, CalculatedValue.Switch.class, CalculatedValue.Divide.class,
							CalculatedValue.Constant.class, CalculatedValue.Multiply.class, CalculatedValue.Add.class, ParamRef.class, CalculatedValue.AddressSize.class, CalculatedValue.ClkPortRef.class
					});
		}

		private void findLoadableFiles(Collection<SourceFile> files, String pattern, Collection<Path> searchpaths, String type) throws IOException {

			for (Path sp : searchpaths) {
				reportProgressScanning(sp);

				Stream<Path> c = FileFinder.find(sp.toAbsolutePath(), pattern);

				c.map(fp -> new SourceFile(sp, fp)).forEach(files::add);
			}
		}

		private void reportProgressScanning(Path p) {
			if (progressIndicator != null) {
				Path displayPath;
				if (p.getNameCount() >= 2) {
					displayPath = p.subpath(p.getNameCount() - 2, p.getNameCount());
				} else {
					displayPath = p;
				}

				progressIndicator.setProgressText("Scanning:  ...\\%s", displayPath);
			}
		}

		private ModuleDescription loadModule(SourceFile sf) {
			return catchExceptionsWhenLoading(p->{

				ModuleDescription module = moduleUnmarshaller.unmarshal(p.file);
				module.setSourceFile(p.file, p.sourceTree);

				ModuleDescription.ModuleType type = module.getType();

				if (verify) {
					verifyModuleFileName(module, p.file);
					verifyBusesOnModule(module);
					verifyParameterValues(module);
				}

				if (analyze) {
					module.analyzeParamRefUsage();
				}

				reportProgressFileLoaded(p.file);
				logger.debug(fileMarker, "Loaded Module {} as {} from {}", module.getName(), type, p);
				return module;
			},"module",sf);
		}

		private void verifyParameterValues(ModuleDescription module) {
			if (module.getType() != IModuleDescription.ModuleType.DynamicStructure) {
				for (BaseParameterGroup parameterGroup : module.getParameterGroups()) {
					for (IParameter parameter : parameterGroup.getParameters()) {
						if (parameter.getDefaultValue() == null && !parameter.hasCalculatedValue()) {
							throw new IllegalStateException(String.format("Parameter %s in module %s has neither default value nor calculated value.", parameter.getName(), module.getName()));
						}
					}
				}
			}
		}

		private void verifyModuleFileName(ModuleDescription module, Path file) {
			String fileName = file.getFileName().toString();
			Optional<String> prefix = extractModuleXmlPrefix(fileName);
			if (!module.getName().equals(prefix.orElse(null))) {
				Path target = file.resolveSibling(module.getName() + ".module.xml");
				throw new IllegalStateException(String.format("Module %s has invalid file name at %s. Should be %s", module.getName(), file, target));
			}
		}

		private void verifyBusesOnModule(ModuleDescription module) throws UnmarshalException {
			for (Bus bus : module.getBuses()) {
				BusDescription busDescription = buses.get(bus.getBusDescriptionName());
				if (busDescription != null) {
					int maxDefCon = busDescription.getMaxConnections(bus.getRole());
					int maxImplCon = bus.getMaxConnections();
					if (maxImplCon != 0 && maxDefCon != 0 && maxImplCon > maxDefCon) {
						throw new IllegalStateException("Module '" + module.getName() + "' tries to implement bus '" + bus.getName() + "' with conflicting maxConnections of " + bus.getMaxConnections());
					}

					BusDescription.BusPortGroup busPortGroup = busDescription.getMatchingPorts(bus);
					bus.getParameterMappings().stream().forEach(pm -> {
						if (!busDescription.isDefinedParameter(pm.getBusName(), bus.getRole())) {
							throw new IllegalStateException("Module '" + module.getName() + "' tries to map non-existent parameter '" + pm.getBusName() + "' of bus '" + bus.getBusDescriptionName() + "'");
						}
					});
					bus.getPortMappings().stream().forEach(pm -> {
						if (!busPortGroup.getPort(pm.getBusName()).isPresent()) {
							throw new IllegalStateException("Module '" + module.getName() + "' tries to map non-existent port '" + pm.getBusName() + "' of bus '" + bus.getBusDescriptionName() + "'");
						}
					});
				} else {
					throw new UnmarshalException("Module references bus " + bus.getBusDescriptionName() + ", but that is unknown at this time");
				}
			}
		}

		private void invalidFile(SourceFile p) {
			invalidFiles.add(p.file);
		}

		@FunctionalInterface
		private interface UnmarshallFunction<T> {

			T apply(SourceFile p) throws JAXBException;
		}

		private <T> T catchExceptionsWhenLoading(UnmarshallFunction<T> loadFunction, String fileType, SourceFile p) {
			try {
				return loadFunction.apply(p);
			} catch (UnmarshalException e) {
				Throwable rootCause = e;
				if (e.getCause() != null) {
					rootCause = e.getCause();
				}
				logger.error("Could not parse file {} as {}:\n{}", p, fileType, rootCause.getMessage());
				invalidFile(p);
				return null;
			} catch (Exception e) {
				logger.error("Could not load file {}", p, e);
				invalidFile(p);
				return null;
			}
		}

		private DeviceDescription loadDevice(SourceFile sf) {
			return catchExceptionsWhenLoading(p -> {
				DeviceDescription device = deviceUnmarshaller.unmarshal(p.file);
				device.setSourceFile(p.file, p.sourceTree);

				reportProgressFileLoaded(p.file);
				logger.debug(fileMarker, "Loaded DeviceDescription {} from {}", device.getName(), p);
				return device;
			},"device", sf);
		}

		private TargetDescription loadTarget(SourceFile sf) {
			return catchExceptionsWhenLoading(p -> {
				TargetDescription target = targetUnmarshaller.unmarshal(p.file, TargetDescription.class);
				target.setSourceFile(p.file, p.sourceTree);

				if (verify) {
					target.verifyIoStds(devices);
					target.verifyBuses(buses);
				}

				reportProgressFileLoaded(p.file);
				logger.debug(fileMarker, "Loaded TargetDescription {} from {}", target.getName(), p);
				return target;
			}, "target", sf);
		}

		private BusDescription loadBus(SourceFile sf) {
			return catchExceptionsWhenLoading(p->{
				BusDescription bus = busUnmarshaller.unmarshal(p.file, BusDescription.class);
				bus.setSourceFile(p.file, p.sourceTree);

				reportProgressFileLoaded(p.file);
				logger.debug(fileMarker, "Loaded BusDescription {} from {}", bus.getName(), p);
				return bus;
			},"bus",sf);
		}

		private void reportProgressFileLoaded(Path p) {
			if (progressIndicator != null) {
				progressIndicator.setProgressText("Loading:  ...\\%s", p.subpath(p.getNameCount() - 2, p.getNameCount()));
				progressIndicator.stepProgress(1);
			}
		}

		/**
		 * At least one folder for each of the different searchpaths.
		 * This means it is at least possible to find sufficient descriptions
		 *
		 * @return true if proceeding is possible
		 */
		public boolean hasMinimumSearchpaths() {
			return !deviceDescSearchpaths.isEmpty() && !moduleDescSearchpaths.isEmpty() && !targetDescSearchpaths.isEmpty() && !busDescSearchpaths.isEmpty();
		}

		/**
		 * Do module analyses while loading. Defaults to false.
		 * If disabled, analysis has to be triggered manually on each module.
		 * This includes parameter reference analysis
		 *
		 * @param analyze true to enable
		 * @return Builder
		 */
		public Builder setAnalyze(boolean analyze) {
			this.analyze = analyze;
			return this;
		}

		/**
		 * Do module optimization while loading. Defaults to false;
		 * If disabled, will be automatically done per module on first use of said data.
		 * Optimizations are transparent to the API and can include HashMaps for revers lookups.
		 *
		 * @param optimize true to enable
		 * @return Builder
		 */
		public Builder setOptimize(boolean optimize) {
			this.optimize = optimize;
			return this;
		}

		/**
		 * Generate bus breakout modules during build().
		 * Otherwise lazily created when needed
		 *
		 * @param busBreakouts enable
		 * @return Builder
		 */
		public Builder setEagerBusBreakouts(boolean busBreakouts) {
			this.generateBusBreakouts = busBreakouts;
			return this;
		}

		/**
		 * Do module verification beyond validating against the Schemas, such as consistency check and verify referenced modules / descriptions.
		 * Defaults to true.
		 *
		 * @param verify true to enable
		 * @return Builder
		 */
		public Builder setVerify(boolean verify) {
			this.verify = verify;
			return this;
		}


		private static class SourceFile {

			private final Path sourceTree;
			private final Path file;

			public SourceFile(Path sourceTree, Path file) {
				this.sourceTree = sourceTree;
				this.file = file;
			}

			@Override
			public String toString() {
				return file.toString();
			}
		}
	}
}
