/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses;

import de.tu_darmstadt.rs.spartanmc.devxml.DevXmlNamespace;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.CalculatedValue;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IRedirectedResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ClkReference;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ParameterReference;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.Parameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlEnumValue;
import javax.xml.bind.annotation.XmlID;
import javax.xml.bind.annotation.XmlRootElement;

import static java.util.stream.Collectors.toMap;

/**
 * Created by Ramon on 22.03.2016.
 */
@XmlRootElement(name = "bus", namespace = DevXmlNamespace.NAMESPACE)
public class Bus implements IConnectorDescription {

	private static final Logger logger = LoggerFactory.getLogger(Bus.class);

	public static final String COUNT_PARAM_NAME = "COUNT";
	public static final String ID_PARAM_NAME = "ID";
	public static final Parameter COUNT_PARAM = new Parameter(COUNT_PARAM_NAME); //Dummy parameter, against which the connection count can be resolved
	public static final Parameter ID_PARAM = new Parameter(ID_PARAM_NAME); //Dummy parameter, against which the connection id can be resolved (index on opposite connector)
	@XmlID
	@XmlAttribute(required = true)
	private String name;
	@XmlAttribute(name = "type")
	private String busDescriptionName;
	@XmlAttribute(required = true)
	private Role role;
	@XmlAttribute
	private String maxConnections = "unlimited";
	private Integer maxConnectionsInternal;
	@XmlElement(name = "connectionCount", type = CalculatedValue.WrapperValue.class)
	private CalculatedValue connectionCount;
	@XmlElement(name = "setParam")
	private List<SetParameter> setParameters = new ArrayList<>();
	@XmlElement(name = "setClk")
	private List<SetClk> setClks = new ArrayList<>();
	@XmlElement(name = "paramMapping")
	private List<ParameterMapping> parameterMappings = new ArrayList<>();
	@XmlElement(name = "clkPortMapping", type = ClkPortMapping.class)
	private List<IClkPortMapping> clkPortMappings = new ArrayList<>();
	@XmlElement(name = "portMapping")
	private List<PortMapping> portMappings = new ArrayList<>();
	private List<IPortMapping> allPortMappings;

	private Bus() {

	}

	public Bus(
			String name,
			String busDescriptionName,
			Role role,
			String maxConnections,
			List<SetParameter> setParameters,
			List<SetClk> setClks,
			List<ParameterMapping> parameterMappings,
			List<IClkPortMapping> clkPortMappings,
			List<PortMapping> portMappings
	) { // TODO interfaces for all!!
		this.name = name;
		this.busDescriptionName = busDescriptionName;
		this.role = role;
		this.maxConnections = maxConnections;
		this.setParameters = setParameters;
		this.setClks = setClks;
		this.parameterMappings = parameterMappings;
		this.clkPortMappings = clkPortMappings;
		this.portMappings = portMappings;
	}

	public static boolean isParametersCompatible(Bus a, IValueResolver aResolver, Bus b, IValueResolver bResolver) {
		if (b.setParameters.size() != a.setParameters.size()) {
			return false;
		}
		if (a.setParameters.isEmpty()) {
			return true;
		} else {
			Map<String, CalculatedValueResult<? extends Number>> aVals = a.setParameters.stream()
					.collect(toMap(
							SetParameter::getBusName,
							param -> param.hasCalculatedValue()? param.resolveValue(aResolver) : CalculatedValueResult.of(param.getStaticValue())
					));

			return b.setParameters.stream().allMatch(param -> {
				CalculatedValueResult<? extends Number> theirValue = aVals.get(param.getBusName());
				CalculatedValueResult<? extends Number> outValue = param.hasCalculatedValue()? param.resolveValue(bResolver) : CalculatedValueResult.of(param.getStaticValue());
				return theirValue.equals(outValue);
			});
		}
	}

	public Optional<SetClk> findSetClk(BusPort port) {
		final Optional<SetClk> any = getSetClks().stream().filter(s -> s.getBusName().equals(port.getName())).findAny();
		if (!any.isPresent()) {
			logger.warn("No SetClk on " + this.getName() + "(" + this.getRole() + ")." + port.getName() + ". SetClks: " + getSetClks());
		}
		return any;
	}

	public int getBusConnectionCount(IValueResolver resolver) {
		if (connectionCount == null) {
			return 1;
		}
		return connectionCount.resolve(resolver).get().intValue();
	}

	public List<SetParameter> getSetParameters() {
		return setParameters;
	}

	public List<SetClk> getSetClks() {
		return setClks;
	}

	/**
	 * Get only clkPortMappings. Only relevant for clkPortRefs.
	 *
	 * @return clkPort mappings
	 */
	public List<IClkPortMapping> getClkPortMappings() {
		return clkPortMappings;
	}

	public List<ParameterMapping> getParameterMappings() {
		return parameterMappings;
	}

	/**
	 * Get all portMappings (clk+portMappings)
	 *
	 * @return concatenation of clkPort+port mappings
	 */
	public List<IPortMapping> getPortMappings() {
		if (allPortMappings == null) {
			allPortMappings = new ArrayList<>();
			allPortMappings.addAll(clkPortMappings);
			allPortMappings.addAll(portMappings);
		}
		return allPortMappings;
	}

	public String getBusDescriptionName() {
		return busDescriptionName;
	}

	public Role getRole() {
		return role;
	}

	public boolean isConnectionsLimited() {
		calcMaxConnections();
		return (maxConnectionsInternal != null && maxConnectionsInternal != 0);
	}

	/**
	 * maxConnection as in moduleDescription.
	 * irrelevant on single-master slaves (only 1 allowed)
	 *
	 * @return 0 if unlimited
	 */
	public int getMaxConnections() {
		calcMaxConnections();
		return maxConnectionsInternal;
	}

	private void calcMaxConnections() {
		if (maxConnectionsInternal == null) {
			try {
				maxConnectionsInternal = Integer.parseInt(maxConnections);
			} catch (NumberFormatException e) {
				maxConnectionsInternal = 0;
			}
		}
	}

	@Override
	public String getName() {
		return name;
	}

	@Override
	public ConnectorDirection getDirection() {
		return (role == Role.Master)? ConnectorDirection.Output : ConnectorDirection.Input;// Fake direction but allows easy checks
	}

	@Override
	public boolean isDirectionLocked() {
		return true;
	}

	@Override
	public Type getType() {
		return Type.Bus;
	}

	public Optional<SetParameter> findSetParameter(String name) {
		return setParameters.stream().filter(s -> s.getBusName().equals(name)).findAny();
	}

	@Override
	public String toString() {
		return "Bus{" +
				"name='" + name + '\'' +
				", type='" + busDescriptionName + '\'' +
				", role=" + role +
				", maxConn='" + maxConnections + '\'' +
				'}';
	}

	@XmlEnum
	public enum Role {
		@XmlEnumValue("master") Master,
		@XmlEnumValue("slave") Slave;

		public Role getOpposite() {
			if (this == Master) {
				return Slave;
			} else {
				return Master;
			}
		}
	}


	public IValueResolver busResolver(IValueResolver rootResolver) {
		return new BusResolver(rootResolver);
	}

	private class BusResolver implements IValueResolver {

		private final IValueResolver rootResolver;

		public BusResolver(IValueResolver rootResolver) {
			this.rootResolver = rootResolver;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolveParam(ParameterReference param) {
			return findSetParameter(param.getName())
					.map(sp -> sp.resolveValue(rootResolver))
					.orElseGet(() -> CalculatedValueResult.customError("No setParam found for "+param.getName()));
		}

		@Override
		public CalculatedValueResult<? extends Number> resolveClkPort(ClkReference clkPort) {
			throw new UnsupportedOperationException("Not implemented");
		}

		@Override
		public CalculatedValueResult<? extends String> resolveParamString(ParameterReference param) {
			throw new UnsupportedOperationException("Not implemented");
		}

		@Override
		public CalculatedValueResult<Double> resolveParamFloat(ParameterReference param) {
			throw new UnsupportedOperationException("Not implemented");
		}
	}


	/**
	 * Set a parameter of the bus to a static or dynamic value
	 * paramRef is just short-hand for value/paramRef
	 */
	public static class SetParameter extends BusSet {

		@XmlAttribute
		private long value;

		@XmlElement(name = "value", type = CalculatedValue.WrapperValue.class)
		private CalculatedValue valueRef;

		public SetParameter() {
		}

		public SetParameter(String busName, CalculatedValue valueRef) {
			super(busName);
			this.valueRef = valueRef;
		}

		public long getStaticValue() {
			return value;
		}

		public boolean hasCalculatedValue() {
			return valueRef != null;
		}

		public CalculatedValue getValueRef() {
			return valueRef;
		}

		public CalculatedValueResult<? extends Number> resolveValue(IValueResolver resolver) {
			if (valueRef != null) {
				return valueRef.resolve(resolver);
			} else {
				return CalculatedValueResult.of(value);
			}
		}

		@Override
		public String toString() {
			return "SetParameter{" +
					"busName='" + getBusName() + '\'' +
					", value=" + value +
					", valueRef=" + valueRef +
					'}';
		}
	}


	public static class SetClk extends BusSet {

		private CalculatedValue frequencyRef;

		private SetClk() {

		}

		public SetClk(String busName, CalculatedValue frequencyRef) {
			super(busName);
			this.frequencyRef = frequencyRef;
		}

		@XmlElement
		private void setFrequency(CalculatedValue.WrapperValue value) {
			this.frequencyRef = value;
		}

		@XmlElement
		private void setClkPortRef(CalculatedValue.ClkPortRef clkPortRef) {
			this.frequencyRef = clkPortRef;
		}

		public CalculatedValue getFrequencyRef() {
			return frequencyRef;
		}

		public CalculatedValueResult<? extends Number> resolveFrequency(IValueResolver resolver) {
			return frequencyRef.resolve(resolver);
		}

		@Override
		public String toString() {
			return "SetClk{" +
					"busName='" + getBusName() + '\'' +
					", frequencyRef=" + frequencyRef +
					'}';
		}
	}


	/**
	 * Exposes a parameter of the bus to the local module, so that it can be referenced with paramRef
	 * Only slaves can do this, as only master can set parameters at the moment
	 */
	public static class ParameterMapping extends ParameterReference {

		private Bus parent;

		@XmlAttribute
		private String busName;

		public String getBusName() {
			return busName;
		}

		private Parameter getBackingParameter() {
			switch (busName) {
				case ID_PARAM_NAME:
					return Bus.ID_PARAM;
				case COUNT_PARAM_NAME:
					return Bus.COUNT_PARAM;
				default:
					throw new UnsupportedOperationException("Buses are currently not capable of mapping params besides 'COUNT' & 'ID'. Requested: '"+ busName +"'");
			}
		}

		@Override
		public boolean isRedirected() {
			return true;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolveRedirected(IRedirectedResolver resolver) {
			return resolver.resolveAgainstBusHandler(parent, getBackingParameter());
		}

		private void afterUnmarshal(Unmarshaller u, Object parent) {
			this.parent = (Bus) parent;
		}
	}


	public static class BusSet {

		@XmlAttribute
		private String busName;

		private BusSet() {
		}

		public BusSet(String busName) {
			this.busName = busName;
		}

		/**
		 * Name local to BusDescription
		 *
		 * @return name local to bus
		 */
		public String getBusName() {
			return busName;
		}
	}


	public static class BusMapping extends BusSet {

		@XmlAttribute
		private String name;

		private BusMapping() {
		}

		public BusMapping(String busName, String name) {
			super(busName);
			this.name = name;
		}

		/**
		 * Name local to module
		 *
		 * @return module local name
		 */
		public String getName() {
			return name;
		}
	}


	public static class PortMapping extends BusMapping implements IPortMapping {

		@XmlAttribute
		private boolean allowMerging = false;
		@XmlAttribute
		private CombineType combineUsing = CombineType.Auto;

		private PortMapping() {
		}

		public PortMapping(String busName, String name, boolean allowMerging, CombineType combineUsing) {
			super(busName, name);
			this.allowMerging = allowMerging;
			this.combineUsing = combineUsing;
		}

		/**
		 * Behavior if this port conflicts with another port (same localName)
		 * merging means combining ports from different buses (matched by localName) into a single port by applying the {@link CombineType}.
		 *
		 * @return bool
		 */
		@Override
		public boolean isAllowMerging() {
			return allowMerging;
		}

		/**
		 * @return configured combineType
		 *
		 * @see CombineType
		 */
		@Override
		public CombineType getCombineUsing() {
			return combineUsing;
		}
	}


	/**
	 * Maps a clkPort of the bus into the local module where it can be referenced.
	 * This is why this has to extend ClkPort, so
	 */
	public static class ClkPortMapping extends ClkReference implements IClkPortMapping {

		private Bus parent;

		@XmlAttribute
		private String busName;

		@XmlAttribute
		private boolean allowMerging = false;
		@XmlAttribute
		private CombineType combineUsing = CombineType.Auto;


		public ClkPortMapping() {
		}

		public ClkPortMapping(String busName, String name, boolean allowMerging, CombineType combineUsing) {
			super(name);
			this.busName = busName;
			this.allowMerging = allowMerging;
			this.combineUsing = combineUsing;
		}

		@Override
		public boolean isRedirected() {
			return true;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolveRedirected(IRedirectedResolver redirector) {
			return redirector.resolveAgainstBus(parent, this);
		}

		@Override
		public CalculatedValueResult<? extends Number> resolveFrequency(IValueResolver resolver) {
			throw new UnsupportedOperationException();
		}

		@Override
		public CalculatedValue getFrequencyRef() {
			throw new UnsupportedOperationException();
		}

		private void afterUnmarshal(Unmarshaller u, Object parent) {
			this.parent = (Bus) parent;
		}

		public Bus getBus() {
			return parent;
		}

		@Override
		public String getBusName() {
			return busName;
		}

		/**
		 * Behavior if this port conflicts with another port (same localName)
		 * merging means combining ports from different buses (matched by localName) into a single port by applying the {@link CombineType}.
		 *
		 * @return bool
		 */
		@Override
		public boolean isAllowMerging() {
			return allowMerging;
		}

		/**
		 * @return configured combineType
		 *
		 * @see CombineType
		 */
		@Override
		public CombineType getCombineUsing() {
			return combineUsing;
		}

		@Override
		public String toString() {
			return "ClkPortMapping{" +
					"parent=" + parent +
					", busName='" + busName + '\'' +
					", name='" + name + '\'' +
					"}";
		}
	}
}
