/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ClkReference;
import de.tu_darmstadt.rs.spartanmc.devxml.jaxb.LongDecodeAdapter;

import com.fasterxml.jackson.annotation.JsonTypeInfo;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.function.Function;
import java.util.stream.Collectors;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlIDREF;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

/**
 * Abstract because JAXB cannot handle Interfaces!
 *
 * @author Ramon Wirsch
 */
@JsonTypeInfo(property = "class", use = JsonTypeInfo.Id.CLASS, include = JsonTypeInfo.As.PROPERTY)
public abstract class CalculatedValue {

	public abstract CalculatedValueResult<? extends Number> resolve(IValueResolver resolver);
	public CalculatedValueResult<? extends String> resolveString(IValueResolver resolver) {
		return resolve(resolver).map(Object::toString);
	}

	public abstract boolean isDynamic();

	/**
	 * Deep clone the hierarchy of CalculatedValues, but allow replacing TerminalCalculatedValues with values delivered
	 * by a mapper
	 *
	 * @param mapper Function that clones/maps TerminalCalculatedValue
	 * @return deep clone
	 */
	public abstract CalculatedValue cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper);


	/**
	 * A CalculatedValue that does not reference other CalculatedValues
	 */
	public abstract static class TerminalCalculatedValue extends CalculatedValue {

		@Override
		public CalculatedValue cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
			return mapper.apply(this);
		}
	}


	public static class WrapperValue extends CalculatedValue {

		@XmlElementRef(required = true)
		protected CalculatedValue value;

		private WrapperValue() {
		}

		public WrapperValue(CalculatedValue value) {
			this.value = value;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolve(IValueResolver resolver) {
			return value.resolve(resolver);
		}

		@Override
		public int hashCode() {
			return value != null? value.hashCode() : 0;
		}

		@Override
		public boolean isDynamic() {
			return value != null && value.isDynamic();
		}

		@Override
		public WrapperValue cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
			return new WrapperValue(value.cloneWithMapper(mapper));
		}

		@Override
		public String toString() {
			return value != null? value.toString() : "{Undefined}";
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			WrapperValue that = (WrapperValue) o;

			return value != null? value.equals(that.value) : that.value == null;

		}


	}


	public static class BooleanWrapper extends TerminalCalculatedValue {

		@XmlElement
		private ParamRef paramRef;

		@XmlAttribute
		private boolean inverted = false;

		public CalculatedValueResult<Boolean> resolveBoolean(IValueResolver resolver) {
			return paramRef.resolve(resolver).map(CalculatedValueResult.longOrDoubleMapper(
					v -> v != 0 ^ inverted,
					v -> Math.abs(v) < 1E-14 ^ inverted
			));
		}

		@Override
		public CalculatedValueResult<? extends Number> resolve(IValueResolver resolver) {
			return resolveBoolean(resolver).map(v -> v? 1 : 0);
		}

		@Override
		public boolean isDynamic() {
			return true;
		}

		@Override
		public String toString() {
			return inverted? "!" + paramRef : "" + paramRef;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			BooleanWrapper that = (BooleanWrapper) o;

			if (inverted != that.inverted) {
				return false;
			}
			return paramRef.equals(that.paramRef);

		}

		@Override
		public int hashCode() {
			int result = paramRef.hashCode();
			result = 31 * result + (inverted? 1 : 0);
			return result;
		}
	}


	@XmlRootElement(name = "constant")
	public static class Constant extends TerminalCalculatedValue {

		@XmlAttribute(required = true)
		@XmlJavaTypeAdapter(LongDecodeAdapter.class)
		@NotNull
		private Long value;

		private Constant() {
		}

		public Constant(long value) {
			this.value = value;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolve(IValueResolver resolver) {
			return CalculatedValueResult.of(value);
		}

		@Override
		public boolean isDynamic() {
			return false;
		}

		@Override
		public String toString() {
			return Long.toString(value);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			Constant constant = (Constant) o;

			return value == constant.value;

		}

		@Override
		public int hashCode() {
			return (int) (value ^ (value >>> 32));
		}
	}


	@XmlRootElement(name = "clkPortRef")
	public static class ClkPortRef extends TerminalCalculatedValue {

		@XmlAttribute(required = true)
		@XmlIDREF
		private ClkReference clkPort;

		private ClkPortRef() {

		}

		public ClkPortRef(ClkReference clkPort) {
			this.clkPort = clkPort;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolve(IValueResolver resolver) {
			if (clkPort == null) {
				throw new NullPointerException("no clkPort!");
			}
			return resolver.resolveClkPort(clkPort).map(x -> x);
		}

		@Override
		public boolean isDynamic() {
			return true;
		}

		@Override
		public String toString() {
			return clkPort != null? "{freq($." + clkPort.getName() + ")}" : "{UndefinedClkPortRef}";
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			ClkPortRef that = (ClkPortRef) o;

			return clkPort != null? clkPort.equals(that.clkPort) : that.clkPort == null;

		}

		@Override
		public int hashCode() {
			return clkPort != null? clkPort.hashCode() : 0;
		}
	}


	@XmlRootElement(name = "addressSize")
	public static class AddressSize extends CalculatedValue {

		@XmlElementRef(required = true)
		private CalculatedValue value;

		public AddressSize() {
		}

		public AddressSize(CalculatedValue value) {
			this.value = value;
		}

		public static int getAddressSize(long value) {
			return (int) Math.ceil(Math.log(value) / Math.log(2));
		}

		@Override
		public CalculatedValueResult<? extends Number> resolve(IValueResolver resolver) {
			return value.resolve(resolver).flatMap(CalculatedValueResult.castOrError(Long.class)).map(AddressSize::getAddressSize).map(i -> (long) i);
		}

		@Override
		public boolean isDynamic() {
			return value != null && value.isDynamic();
		}

		@Override
		public AddressSize cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
			return new AddressSize(value.cloneWithMapper(mapper));
		}

		@Override
		public String toString() {
			return value != null? "adrSize(" + value + ")" : "adrSize{Undefined}";
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			AddressSize that = (AddressSize) o;

			return value != null? value.equals(that.value) : that.value == null;

		}

		@Override
		public int hashCode() {
			return value != null? value.hashCode() : 0;
		}
	}


	/**
	 * assumes input to be boolean and inverts it.
	 * 0 is false, everything else is true
	 * will output 0 or 1
	 * Resolver must return boolean values as ints!
	 */
	@XmlRootElement(name = "invert")
	public static class Invert extends CalculatedValue {

		@XmlElementRef(required = true)
		private CalculatedValue value;

		public Invert() {
		}

		public Invert(CalculatedValue value) {
			this.value = value;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolve(IValueResolver resolver) {
			return value.resolve(resolver).map(CalculatedValueResult.longOrDoubleMapper(
					v -> (v == 0)? 1L : 0L,
					v -> Math.abs(v) < 1E-14? 1L : 0L
			));
		}

		@Override
		public boolean isDynamic() {
			return value != null && value.isDynamic();
		}

		@Override
		public Invert cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
			return new Invert(value.cloneWithMapper(mapper));
		}

		@Override
		public String toString() {
			return value != null? "!" + value : "!{Undefined}";
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			Invert invert = (Invert) o;

			return value != null? value.equals(invert.value) : invert.value == null;

		}

		@Override
		public int hashCode() {
			return value != null? value.hashCode() : 0;
		}
	}


	@XmlRootElement(name = "switch")
	public static class Switch extends CalculatedValue {

		private static class Case extends WrapperValue {

			@XmlAttribute(name = "value")
			private long caseValue;

			public Case() {
			}

			public Case(CalculatedValue value, long caseValue) {
				super(value);
				this.caseValue = caseValue;
			}

			public long getCaseValue() {
				return caseValue;
			}

			@Override
			public Case cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
				return new Case(value.cloneWithMapper(mapper), caseValue);
			}

			@Override
			public int hashCode() {
				return super.hashCode() * 31 + (int) caseValue;
			}

			@Override
			public boolean equals(Object o) {
				if (!super.equals(o)) {
					return false;
				}
				return caseValue == ((Case) o).caseValue;
			}
		}


		@XmlElement(name = "value", required = true, type = WrapperValue.class)
		private CalculatedValue switchValue;
		@XmlElement(name = "case")
		private List<Case> cases;
		@XmlElement(name = "default", type = WrapperValue.class)
		private CalculatedValue defaultValue;

		public Switch() {
		}

		public Switch(CalculatedValue switchValue, List<Case> cases, CalculatedValue defaultValue) {
			this.switchValue = switchValue;
			this.defaultValue = defaultValue;
			this.cases = cases;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolve(IValueResolver resolver) {
			return switchCase(resolver).flatMap(f -> f.resolve(resolver));
		}

		private CalculatedValueResult<CalculatedValue> switchCase(IValueResolver resolver) {
			return this.switchValue.resolve(resolver).flatMap(CalculatedValueResult.castOrError(Long.class)).map(resolvedSwitch -> {
				for (Case cs : cases) {
					if (resolvedSwitch == cs.getCaseValue()) {
						return cs;
					}
				}
				return defaultValue;
			});
		}

		@Override
		public boolean isDynamic() {
			return switchValue.isDynamic() || cases.stream().anyMatch(Case::isDynamic) || defaultValue.isDynamic();
		}

		@Override
		public CalculatedValue cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
			return new Switch(switchValue.cloneWithMapper(mapper),
					cases.stream().map(cs -> cs.cloneWithMapper(mapper)).collect(Collectors.toList()),
					defaultValue.cloneWithMapper(mapper));
		}

		@Override
		public int hashCode() {
			int result = 0;
			if (cases != null) {
				for (Case cs : cases) {
					result = result * 31 + (cs != null? cs.hashCode() : 0);
				}
			}
			result = result * 31 + (defaultValue != null? defaultValue.hashCode() : 0);
			result = result * 31 + (switchValue != null? switchValue.hashCode() : 0);

			return result;
		}

		@Override
		public boolean equals(Object obj) {
			if (obj == null || obj.getClass() != getClass()) {
				return false;
			}
			Switch other = (Switch) obj;

			return (switchValue != null? switchValue.equals(other.switchValue) : other.switchValue == null)
					&& (cases != null? cases.equals(other.cases) : other.cases == null)
					&& (defaultValue != null? defaultValue.equals(other.defaultValue) : other.defaultValue == null);
		}
	}


	private abstract static class BinaryCalculatedValue extends CalculatedValue {

		@XmlElement(name = "a", required = true, type = WrapperValue.class)
		protected CalculatedValue firstOperand;
		@XmlElement(name = "b", required = true, type = WrapperValue.class)
		protected CalculatedValue secondOperand;

		public BinaryCalculatedValue() {
		}

		public BinaryCalculatedValue(CalculatedValue firstOperand, CalculatedValue secondOperand) {
			this.firstOperand = firstOperand;
			this.secondOperand = secondOperand;
		}

		protected abstract long combine(long a, long b);

		protected abstract double combine(double a, double b);


		@Override
		public final CalculatedValueResult<? extends Number> resolve(IValueResolver resolver) {
			return CalculatedValueResult.dualMap(
				firstOperand.resolve(resolver),
				secondOperand.resolve(resolver),
				(a, b) -> {
					if (a instanceof Double || b instanceof Double) {
						return combine(a.doubleValue(), b.doubleValue());
					} else {
						return combine(a.longValue(), b.longValue());
					}
				}
			);
		}

		protected String firstOperandToString() {
			return firstOperand != null? firstOperand.toString() : "{Undefined}";
		}

		protected String secondOperandToString() {
			return firstOperand != null? secondOperand.toString() : "{Undefined}";
		}

		@Override
		public boolean isDynamic() {
			return firstOperand != null && secondOperand != null && (firstOperand.isDynamic() || secondOperand.isDynamic());
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			BinaryCalculatedValue that = (BinaryCalculatedValue) o;

			if (firstOperand != null? firstOperand.equals(that.firstOperand) : that.firstOperand == null) { // Commutative by default
				return secondOperand != null? secondOperand.equals(that.secondOperand) : that.secondOperand == null;
			} else {
				return (firstOperand != null? firstOperand.equals(that.secondOperand) : that.secondOperand == null) && (secondOperand != null? secondOperand.equals(that.firstOperand) : that.firstOperand == null);
			}
		}

		@Override
		public int hashCode() {
			int result = firstOperand != null? firstOperand.hashCode() : 0;
			result = result + (secondOperand != null? secondOperand.hashCode() : 0);
			return result;
		}


		@Override
		public abstract BinaryCalculatedValue cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper);
	}


	@XmlRootElement(name = "multiply")
	public static class Multiply extends BinaryCalculatedValue {

		public Multiply() {
		}

		public Multiply(CalculatedValue firstOperand, CalculatedValue secondOperand) {
			super(firstOperand, secondOperand);
		}


		protected long combine(long a, long b) {
			return a * b;
		}

		protected double combine(double a, double b) {
			return a * b;
		}

		@Override
		public Multiply cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
			return new Multiply(firstOperand.cloneWithMapper(mapper), secondOperand.cloneWithMapper(mapper));
		}

		@Override
		public String toString() {
			return "(" + firstOperandToString() + "*" + secondOperandToString() + ")";
		}
	}


	@XmlRootElement(name = "add")
	public static class Add extends BinaryCalculatedValue {

		public Add() {
		}

		public Add(CalculatedValue firstOperand, CalculatedValue secondOperand) {
			super(firstOperand, secondOperand);
		}

		protected long combine(long a, long b) {
			return a + b;
		}

		protected double combine(double a, double b) {
			return a + b;
		}

		@Override
		public String toString() {
			return "(" + firstOperandToString() + "+" + secondOperandToString() + ")";
		}


		@Override
		public Add cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
			return new Add(firstOperand.cloneWithMapper(mapper), secondOperand.cloneWithMapper(mapper));
		}
	}


	@XmlRootElement(name = "divide")
	public static class Divide extends BinaryCalculatedValue {

		public Divide() {
		}

		public Divide(CalculatedValue firstOperand, CalculatedValue secondOperand) {
			super(firstOperand, secondOperand);
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			BinaryCalculatedValue that = (BinaryCalculatedValue) o;

			return (firstOperand != null? firstOperand.equals(that.firstOperand) : that.firstOperand == null) && (secondOperand != null? secondOperand.equals(that.secondOperand) : that.secondOperand == null);
		}

		@Override
		public int hashCode() {
			int result = firstOperand != null? firstOperand.hashCode() : 0;
			result = 31 * result + (secondOperand != null? secondOperand.hashCode() : 0);
			return result;
		}


		protected long combine(long a, long b) {
			return a / b;
		}

		protected double combine(double a, double b) {
			return a / b;
		}

		@Override
		public String toString() {
			return "(" + firstOperandToString() + "/" + secondOperandToString() + ")";
		}


		@Override
		public Divide cloneWithMapper(Function<TerminalCalculatedValue, CalculatedValue> mapper) {
			return new Divide(firstOperand.cloneWithMapper(mapper), secondOperand.cloneWithMapper(mapper));
		}


	}
}
