/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.CalculatedValue;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.DynamicParamRef;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.ParamRef;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.INamedValueResolver;

import com.fasterxml.jackson.annotation.JsonProperty;

import java.io.Serializable;
import java.nio.file.Path;
import java.util.List;
import java.util.function.UnaryOperator;
import java.util.stream.Collectors;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlMixed;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * Mix between text and embedded param refs
 */
public class MixedParamRefs {


	/**
	 * Embeeded reference to module name
	 */
	@XmlRootElement(name = "moduleName")
	public static class ModuleName implements Serializable {

		public ModuleName() {
		}

		@Override
		public String toString() {
			return "$MODULE_NAME";
		}
	}

	/**
	 * Embedded reference to firmware build directory
	 */
	@XmlRootElement(name = "firmwareBuildDir")
	public static class FirmwareBuildDir implements Serializable {

		public FirmwareBuildDir() {
		}

		@Override
		public String toString() {
			return "$FIRMWARE_BUILD_DIR";
		}
	}


	/**
	 * Embedded reference to some unique id (it's the module name with special chars replaced)
	 */
	@XmlRootElement(name = "id")
	public static class Id implements Serializable {

		public Id() {
		}

		@Override
		public String toString() {
			return "$ID";
		}
	}


	/**
	 * Embedded reference to the current module's source directory
	 */
	@XmlRootElement(name = "moduleDir")
	public static class ModuleDir implements Serializable {

		public ModuleDir() {
		}

		@Override
		public String toString() {
			return "$ModuleDir";
		}
	}

	/**
	 * Embedded reference to the current module's base address
	 */
	@XmlRootElement(name = "moduleBaseAddress")
	public static class ModuleBaseAddress implements Serializable {

		public ModuleBaseAddress() {
		}

		@Override
		public String toString() {
			return "$ModuleBaseAddress";
		}
	}




	@XmlMixed
	@XmlElementRefs(value = {
			@XmlElementRef(type = ParamRef.class),
			@XmlElementRef(type = DynamicParamRef.class),
			@XmlElementRef(type = ModuleName.class),
			@XmlElementRef(type = Id.class),
			@XmlElementRef(type = ModuleDir.class),
			@XmlElementRef(type = ModuleBaseAddress.class)
	})
	@JsonProperty
	protected List<Object> content;

	public List<Object> getContent() {
		return content;
	}

	public MixedParamRefs() {
	}

	@Override
	public String toString() {
		return "MixedParamRefs" + content;
	}

	/**
	 * Resolve the value
	 *
	 * @param resolver value resolver
	 * @param nameFormatter formatter for the module name
	 * @param moduleDir directory the current module's source is located in
	 * @param baseAddress base address of current module
	 * @return the content with all values resolved
	 */
	public CalculatedValueResult<String> resolve(INamedValueResolver resolver, UnaryOperator<String> nameFormatter, Path moduleDir, long baseAddress) {
		final CalculatedValueResult[] error = {null};
		final String res = content.stream()
				.map(x -> {
					if (x instanceof CalculatedValue) {
						final CalculatedValueResult<? extends String> r = ((CalculatedValue) x).resolveString(resolver);
						if (!r.isPresent()) {
							error[0] = r;
							return "";
						}
						return r.get();
					} else if (x instanceof ModuleName) {
						return nameFormatter.apply(resolver.getGlobalName());
					} else if (x instanceof Id) {
						return nameFormatter.apply(resolver.getGlobalName()).replaceAll("[^A-Za-z0-9_]", "_");
					} else if (x instanceof ModuleDir) {
						return moduleDir.toAbsolutePath().toString();
					} else if (x instanceof ModuleBaseAddress) {
						return String.format("0x%x", baseAddress);
					} else {
						return x.toString();
					}
				}).collect(Collectors.joining());
		if (error[0] != null) {
			return (CalculatedValueResult<String>) error[0];
		}
		return CalculatedValueResult.of(res);
	}

	/**
	 * Resolve the value when not having a baseAddress element present
	 *
	 * @param resolver value resolver
	 * @param nameFormatter formatter for the module name
	 * @param moduleDir directory the current module's source is located in
	 * @return the content with all values resolved
	 */
	public CalculatedValueResult<String> resolve(INamedValueResolver resolver, UnaryOperator<String> nameFormatter, Path moduleDir) {
		return resolve(resolver, nameFormatter, moduleDir, 0);
	}
}
