/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies;

import java.util.Objects;
import java.util.function.BiFunction;
import java.util.function.Function;

/**
 * Like java.util.Optional, only with more info if a value is absent
 * @param <T> element type
 */
public class CalculatedValueResult<T> {

	private static ErrorValue<?> unknownValue = new ErrorValue<>("Unknown Value");
	private static ErrorValue<?> unconnectedValue = new ErrorValue<>("Unconnected");
	private static ErrorValue<?> absentValue = new ErrorValue<>("Value is absent");
	private static ErrorValue<?>[] errorValues = {
			unconnectedValue,
			absentValue,
			unknownValue
	};
	private final T value;

	//For jaxb
	private CalculatedValueResult() {
		this.value = null;
	}

	private CalculatedValueResult(T value) {
		this.value = value;
	}

	public static <T> CalculatedValueResult<T> of(T value) {
		return new CalculatedValueResult<T>(Objects.requireNonNull(value));
	}

	/**
	 * Create a CalculatedValueResult of a nullable element
	 * @param value the value
	 * @param <T> element type
	 * @return CVR that wraps the value, or absent if value is null
	 */
	public static <T> CalculatedValueResult<T> ofNullable(T value) {
		if (value == null) {
			return absent();
		}
		return new CalculatedValueResult<T>(value);
	}

	public static <U> ErrorValue<U> unconnected() {
		return (ErrorValue<U>) unconnectedValue;
	}

	public static <U> ErrorValue<U> unknown() {
		return (ErrorValue<U>) unknownValue;
	}

	public static <U> ErrorValue<U> absent() {
		return (ErrorValue<U>) absentValue;
	}

	public static <U> ErrorValue<U> customError(String error) {
		return new ErrorValue<>(error);
	}

	/**
	 * Map two CalculatedValueResults at the same time. If both are Errors, the more important error is returned
	 * @param a First Value
	 * @param b Second Value
	 * @param f Mapper
	 * @param <A> First element type
	 * @param <B> Second element type
	 * @param <R> Result type
	 * @return If one of the inputs is an error, the error. If both are errors, the more important error. Else the result of the mapper
	 */
	public static <A, B, R> CalculatedValueResult<R> dualMap(CalculatedValueResult<? extends A> a, CalculatedValueResult<? extends B> b, BiFunction<A, B, ? extends R> f) {
		if (a.isPresent() && b.isPresent()) {
			return new CalculatedValueResult<R>(f.apply(a.get(), b.get()));
		} else {
			if (a.isPresent()) {
				return (ErrorValue<R>) b;
			} else if (b.isPresent()) {
				return (ErrorValue<R>) a;
			} else { //Both absent.
				return getMoreImportantError((ErrorValue<R>) a, (ErrorValue<R>) b);
			}
		}
	}

	private static <T> ErrorValue<T> getMoreImportantError(ErrorValue<T> a, ErrorValue<T> b) {
		for (ErrorValue<?> listVal : errorValues) {
			if (listVal == a) {
				return a;
			}
			if (listVal == b) {
				return b;
			}
		}

		return a;

	}

	/**
	 * Map long or double values
	 * @param longMapper function to map longs
	 * @param doubleMapper function to map doubles
	 * @param <T> result type
	 * @return the mapped value
	 * @throws IllegalArgumentException if the element type is neither long or double
	 */
	public static <T> Function<Number, T> longOrDoubleMapper(Function<Long, ? extends T> longMapper, Function<Double, ? extends T> doubleMapper) {
		return x -> {
			if (x instanceof Long) {
				return longMapper.apply((Long) x);
			} else if (x instanceof Double) {
				return doubleMapper.apply((Double) x);
			} else {
				throw new IllegalArgumentException("Value is a " + x.getClass() + ", but should be a Long or Double");
			}

		};
	}

	/**
	 * Mapper function to try to cast the element type or return an error value
	 * @param clazz the class to map to
	 * @param <T> element type
	 * @param <U> result type
	 * @return the casted value or an error value
	 */
	public static <T, U extends T> Function<T, CalculatedValueResult<U>> castOrError(Class<U> clazz) {
		return (T x) -> {
			if (clazz.isInstance(x)) {
				return CalculatedValueResult.of((U) x);
			} else {
				return CalculatedValueResult.customError("Value is a " + x.getClass() + ", but should be a " + clazz);
			}
		};
	}

	public boolean isPresent() {
		return value != null;
	}

	public T get() {
		return value;
	}

	/**
	 * Apply the mapping function. Error values are left unchanged.
	 * @param f mapper
	 * @param <U> result type
	 * @return the mapper applied to the value
	 */
	public <U> CalculatedValueResult<U> map(Function<? super T, ? extends U> f) {
		if (isPresent()) {
			return new CalculatedValueResult<>(f.apply(get()));
		} else {
			return (CalculatedValueResult<U>) this;
		}
	}

	/**
	 * Apply the flat mapping function. Error values are left unchanged.
	 * @param f mapper
	 * @param <U> result type
	 * @return the mapper applied to the value
	 */
	public <U> CalculatedValueResult<U> flatMap(Function<? super T, CalculatedValueResult<U>> f) {
		if (isPresent()) {
			return f.apply(get());
		} else {
			return (CalculatedValueResult<U>) this;
		}
	}

	/**
	 * If the value is an error value, return the specified value
	 * @param other value to return in case of errors
	 * @return the element or other
	 */
	public T orElse(T other) {
		if (isPresent()) {
			return get();
		}
		return other;
	}

	/**
	 * If the value is an error value, return map the error to a value
	 * @param messageConsumer mapper to invoke in case of error values
	 * @return the element or mapped error
	 */
	public T orError(Function<String, T> messageConsumer) {
		if (isPresent()) {
			return get();
		}
		throw new UnsupportedOperationException("Should be overridden in child class!");
	}

	/**
	 * If the value is an error value, throw an exception
	 * @return the value
	 */
	public String orError() {
		if (isPresent()) {
			return get().toString();
		}
		throw new UnsupportedOperationException("Should be overridden in child class!");
	}
	/**
	 * If the value is an error value, throw an exception generated by the supplier
	 * @param exceptionSupplier created exception for errors
	 * @param <X> exception type
	 * @return the value
	 * @throws X specified exception
	 */
	public <X extends Throwable> T orElseThrow(Function<CalculatedValueResult, ? extends X> exceptionSupplier) throws X {
		if (isPresent()) {
			return get();
		}
		throw exceptionSupplier.apply(this);
	}

	public T orElseThrow() {
		return orElseThrow((x) -> new RuntimeException("Value is absent: " + x));
	}

	public String getError() {
		throw new RuntimeException("This is not an error");
	}

	@Override
	public int hashCode() {
		return value != null? value.hashCode() : 0;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (!(o instanceof CalculatedValueResult)) {
			return false;
		}

		CalculatedValueResult<?> that = (CalculatedValueResult<?>) o;

		return value != null? value.equals(that.value) : that.value == null;
	}

	@Override
	public String toString() {
		return "CalculatedValueResult{" +
				value +
				'}';
	}


	/**
	 * Error value. Holds a string describing the error
	 * @param <T> element type
	 */
	public static final class ErrorValue<T> extends CalculatedValueResult<T> {

		public final String errorValue;

		private ErrorValue(String errorValue) {
			super(null);
			this.errorValue = errorValue;
		}

		@Override
		public String toString() {
			return errorValue;
		}

		@Override
		public T orError(Function<String, T> messageConsumer) {
			return messageConsumer.apply(errorValue);
		}

		@Override
		public String orError() {
			return errorValue;
		}

		@Override
		public String getError() {
			return errorValue;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (!(o instanceof ErrorValue)) {
				return false;
			}
			if (!super.equals(o)) {
				return false;
			}

			ErrorValue<?> that = (ErrorValue<?>) o;

			return errorValue.equals(that.errorValue);
		}

		@Override
		public int hashCode() {
			int result = super.hashCode();
			result = 31 * result + errorValue.hashCode();
			return result;
		}
	}
}
