/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic;

import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.Constraint;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.MixedParamRefs;
import de.tu_darmstadt.rs.spartanmc.devxml.files.base.ILoadedFile;
import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;

import java.io.Serializable;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlIDREF;
import javax.xml.bind.annotation.XmlMixed;
import javax.xml.bind.annotation.XmlValue;

/**
 * Created by Ramon on 16.06.2016.
 */
public class HdlDescription {

	@XmlAttribute
	private String lang;

	@XmlAttribute
	private String toplevel;

	@XmlElement
	private Sources sources;

	@XmlAttribute
	private Verification verification = Verification.FULL;

	@XmlElementWrapper(name = "constraints")
	@XmlElementRef
	private List<Constraint> constraints = new ArrayList<>();


	@XmlElementWrapper(name = "tests")
	@XmlElement(name = "testbench")
	private List<ModuleTestbench> testbenches = new ArrayList<>();

	@XmlElement(name = "singletonSimulationRootModule")
	private List<SingletonSimRootModule> singletonSimRootModules = new ArrayList<>();

	private HdlDescription() {

	}

	public HdlDescription(String lang, String toplevel, Sources sources) {

		this.lang = lang;
		this.toplevel = toplevel;
		this.sources = sources;
	}

	public String getLang() {
		return lang;
	}

	public String getToplevel() {
		return toplevel;
	}

	public Sources getSources() {
		return sources;
	}

	public Verification getVerification() {
		return verification;
	}

	public List<Constraint> getConstraints() {
		return constraints;
	}

	public List<ModuleTestbench> getTestbenches() {
		return testbenches;
	}

	public List<SingletonSimRootModule> getSingletonSimRootModules() {
		return singletonSimRootModules;
	}

	public static class SourceFile {

		@XmlValue
		protected String filename;

		@XmlAttribute
		protected boolean simulationOnly;

		public SourceFile() {
		}

		public SourceFile(String filename, boolean simulationOnly) {
			this.filename = filename;
			this.simulationOnly = simulationOnly;
		}

		@Override
		public String toString() {
			return "SourceFile{" +
					"filename='" + filename + '\'' +
					", simulationOnly=" + simulationOnly +
					'}';
		}
	}


	public static class PlaceholderFile {

		@XmlMixed
		@XmlElementRefs(value = {
			@XmlElementRef(type = MixedParamRefs.ModuleName.class),
			@XmlElementRef(type = MixedParamRefs.FirmwareBuildDir.class)
		})
		protected List<Serializable> content;

		@XmlAttribute
		protected boolean simulationOnly;

		@Override
		public String toString() {
			return content.toString();
		}

		public Path resolve(Path parent, Path firmwareBuildPath, String moduleName) {
			if (firmwareBuildPath != null && content.get(0) instanceof MixedParamRefs.FirmwareBuildDir) {
				// if first entry is FirmwareBuildDir element, then use firmwareBuildPath as path
				parent = firmwareBuildPath;
			}
			String name = content.stream()
					.map(x -> {
						if (x instanceof MixedParamRefs.ModuleName) {
							return moduleName;
						} else if (x instanceof MixedParamRefs.FirmwareBuildDir) {
							return ""; // ignore
						} else {
							return x.toString();
						}
					})
					.collect(Collectors.joining());
			return parent.resolve(name);
		}

		public boolean isSimulationOnly() {
			return simulationOnly;
		}
	}


	public static class CoregenConstraint extends PlaceholderFile {

		@XmlAttribute
		public String modifier;

		public String getModifier() {
			return modifier;
		}

		public Path resolve(Path parent, String moduleName) {
			return resolve(parent, null, moduleName);
		}

		@Override
		public String toString() {
			return "CoregenConstraint{" +
					"modifier='" + modifier + '\'' +
					"} " + super.toString();
		}
	}


	public static abstract class SourceRoot {

		public static final String DEFAULT_SRC_DIR = "src";

		@XmlAttribute
		@XmlIDREF
		private ToolchainConfigurationDescription configuration;

		@XmlAttribute
		private String library;

		@XmlElement(name = "file")
		private List<SourceFile> files;

		@XmlElement(name = "coregenFile")
		private List<PlaceholderFile> coregenFiles;

		@XmlElement(name = "coregenConstraint")
		private List<CoregenConstraint> coregenConstraints;

		protected SourceRoot() {

			files = new ArrayList<>();
		}

		public SourceRoot(ToolchainConfigurationDescription configuration, List<SourceFile> files, List<PlaceholderFile> coregenFiles, List<CoregenConstraint> coregenConstraints) {
			this.configuration = configuration;
			this.files = files;
			this.coregenFiles = coregenFiles;
			this.coregenConstraints = coregenConstraints;
		}

		public ToolchainConfigurationDescription getConfiguration() {
			return configuration;
		}

		public List<SourceFile> getFiles() {
			return files;
		}

		public List<PlaceholderFile> getCoregenFiles() {
			return coregenFiles;
		}

		public List<CoregenConstraint> getCoregenConstraints() {
			if (coregenConstraints == null) {
				return Collections.emptyList();
			}
			return coregenConstraints;
		}

		public abstract String getPath();

		public String getLibrary() {
			return library;
		}

		public abstract Path resolveRootPath(ILoadedFile moduleDescription);

		public Stream<SourceRef> streamResolvedFilePaths(ILoadedFile moduleDescription, IToolchainConfiguration currentConfig) {
			if (configuration==null || configuration.supports(currentConfig)) {
				Path rootPath = resolveRootPath(moduleDescription);
				String lib = getLibrary();
				return files.stream().map(sf -> {
					if (sf.filename == null || sf.filename.isEmpty()) {
						throw new NullPointerException("empty filename in sources for " + moduleDescription.getName());
					}
					return new SourceRef(rootPath.resolve(sf.filename), lib, SourceRef.FileType.fromBoolean(sf.simulationOnly));
				});
			} else {
				return Stream.empty();
			}
		}

		protected Path relativeToDescription(ILoadedFile moduleDescription) {
			return moduleDescription.getSourceFile().resolveSibling(getPath());
		}
	}


	public static class ExternalSourcesRoot extends SourceRoot {

		@XmlAttribute(required = true)
		private String path;

		protected ExternalSourcesRoot() {

		}

		public ExternalSourcesRoot(String path, ToolchainConfigurationDescription configuration, List<SourceFile> files, List<PlaceholderFile> coregenFiles, List<CoregenConstraint> coregenConstraints) {
			super(configuration, files, coregenFiles, coregenConstraints);
			this.path = path;
		}

		@Override
		public String getPath() {
			return path;
		}

		@Override
		public Path resolveRootPath(ILoadedFile moduleDescription) {
			return Paths.get(path);
		}
	}


	public static class AdditionalSourcesRoot extends SourceRoot {

		@XmlAttribute
		private String path;

		protected AdditionalSourcesRoot() {
			path = DEFAULT_SRC_DIR;
		}

		public AdditionalSourcesRoot(ToolchainConfigurationDescription configuration, List<SourceFile> files, List<PlaceholderFile> coregenFiles, String path, List<CoregenConstraint> coregenConstraints) {
			super(configuration, files, coregenFiles, coregenConstraints);
			this.path = path;
		}

		@Override
		public String getPath() {
			return path;
		}

		@Override
		public Path resolveRootPath(ILoadedFile moduleDescription) {
			return relativeToDescription(moduleDescription);
		}
	}


	public static class TreeSourcesRoot extends SourceRoot {

		@Override
		public String getPath() {
			return null;
		}

		@Override
		public Path resolveRootPath(ILoadedFile moduleDescription) {
			return moduleDescription.getSourceTree();
		}
	}


	public static class Sources extends SourceRoot {

		@XmlElement(name = "root")
		private List<AdditionalSourcesRoot> additionalRoots;

		@XmlElement(name = "treeRoot")
		private List<TreeSourcesRoot> treeRoots;

		@XmlElement(name = "externalRoot")
		private List<ExternalSourcesRoot> externalRoots;

		@XmlElement(name = "module")
		private List<String> modules;

		protected Sources() {
			modules = new ArrayList<>();
			externalRoots = new ArrayList<>();
			treeRoots = new ArrayList<>();
			additionalRoots = new ArrayList<>();
		}

		public Sources(ToolchainConfigurationDescription configuration, List<SourceFile> files, List<PlaceholderFile> coregenFiles, List<CoregenConstraint> coregenConstraints, List<String> modules) {
			super(configuration, files, coregenFiles, coregenConstraints);
			this.modules = new ArrayList<>();
			this.modules = modules;
			externalRoots = new ArrayList<>();
			treeRoots = new ArrayList<>();
			additionalRoots = new ArrayList<>();
		}

		public List<String> getModules() {
			return modules;
		}

		public List<AdditionalSourcesRoot> getAdditionalRoots() {
			return additionalRoots;
		}

		public List<TreeSourcesRoot> getTreeRoots() {
			return treeRoots;
		}

		public List<ExternalSourcesRoot> getExternalRoots() {
			return externalRoots;
		}

		public Stream<? extends SourceRoot> streamChildRoots() {
			return Stream.concat(additionalRoots.stream(), Stream.concat(treeRoots.stream(), externalRoots.stream()));
		}

		public Stream<SourceRef> streamAllResolvedPaths(ILoadedFile moduleDescription, IToolchainConfiguration currentConfig) {
			return Stream.concat(
					streamResolvedFilePaths(moduleDescription, currentConfig),
					streamChildRoots().flatMap(root -> root.streamResolvedFilePaths(moduleDescription, currentConfig))
			);
		}

		@Override
		public String getPath() {
			return DEFAULT_SRC_DIR;
		}

		@Override
		public Path resolveRootPath(ILoadedFile moduleDescription) {
			return relativeToDescription(moduleDescription);
		}


	}

	public static class ModuleTestbench {
		@XmlAttribute
		private String name;

		@XmlAttribute
		private String runtime;

		@XmlAttribute
		@XmlIDREF
		private ToolchainConfigurationDescription configuration;

		public String getName() {
			return name;
		}

		public String getRuntime() {
			return runtime;
		}

		public ToolchainConfigurationDescription getConfiguration() {
			return configuration;
		}
	}

	public static class SingletonSimRootModule {
		@XmlAttribute
		private String name;

		@XmlAttribute
		private boolean includeInPostParSim = false;

		@XmlAttribute
		@XmlIDREF
		private ToolchainConfigurationDescription configuration;

		public SingletonSimRootModule() {
		}

		public SingletonSimRootModule(String name, boolean includeInPostParSim) {
			this.name = name;
			this.includeInPostParSim = includeInPostParSim;
		}

		public String getName() {
			return name;
		}

		public boolean isIncludeInPostParSim() {
			return includeInPostParSim;
		}

		public ToolchainConfigurationDescription getConfiguration() {
			return configuration;
		}
	}

	/*public static class FileRef {

		@XmlValue
		private String value;

		@XmlAttribute
		private String root;

		public boolean hasRoot() {
			return root != null;
		}

		public String getRoot() {
			return root;
		}

		public String getValue() {
			return value;
		}

		@Override
		public String toString() {
			return value;
		}
	}*/
}
