/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.jaxb.LongDecodeAdapter;

import com.fasterxml.jackson.annotation.JsonIgnore;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.Stream;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlEnumValue;
import javax.xml.bind.annotation.XmlIDREF;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

/**
 * Created by ramon on 23.03.2016.
 */
public class AddressLayoutSpec {

	@XmlElement(name = "addressSpace")
	private List<AddressSpace> addressSpaces = new ArrayList<>();
	@XmlElement(name = "autoMerge")
	private List<AutoMergeInfo> autoMerges = new ArrayList<>();

	public AddressLayoutSpec() {
	}

	public List<AddressSpace> getAddressSpaces() {
		return addressSpaces;
	}

	public List<AutoMergeInfo> getAutoMerges() {
		return autoMerges;
	}

	public Stream<AddressRange> streamImportingAddressRanges() {
		return addressSpaces.stream().flatMap(as -> as.addressRanges.stream());
	}

	@Override
	public int hashCode() {
		int result = addressSpaces.hashCode();
		result = 31 * result + autoMerges.hashCode();
		return result;
	}

	@Override
	public String toString() {
		return "AddressLayoutSpec{" +
				"addressSpaces=" + addressSpaces +
				", autoMerges=" + autoMerges +
				'}';
	}

	@XmlEnum
	public enum BlockType {
		@XmlEnumValue("SelfInitializing")SelfInitializing("mem"),
		@XmlEnumValue("Initializable")Initializable("dma"),
		@XmlEnumValue("Uninitializable")Unintializable("reg");

		private final String shortName;

		BlockType(String shortName) {
			this.shortName = shortName;
		}

		public String shortName() {
			return shortName;
		}
	}


	public static class AddressSpace {

		@XmlElement(name = "range")
		private List<AddressRange> addressRanges = new ArrayList<>();

		@XmlAttribute(required = true)
		private String name;

		@XmlJavaTypeAdapter(LongDecodeAdapter.class)
		@XmlAttribute(required = true)
		private Long start = 0L;

		@XmlJavaTypeAdapter(LongDecodeAdapter.class)
		@XmlAttribute(required = true)
		private Long end = 0L;

		public List<AddressRange> getAddressRanges() {
			return addressRanges;
		}

		public String getName() {
			return name;
		}

		public long getStart() {
			return start;
		}

		public long getEnd() {
			return end;
		}

		@Override
		public String toString() {
			return "AddressSpace{" +
					"addressRanges=" + addressRanges +
					", name='" + name + '\'' +
					", start=0x" + Long.toHexString(start) +
					", end=0x" + Long.toHexString(end) +
					'}';
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			AddressSpace that = (AddressSpace) o;

			if (!addressRanges.equals(that.addressRanges)) {
				return false;
			}
			if (!name.equals(that.name)) {
				return false;
			}
			if (!start.equals(that.start)) {
				return false;
			}
			return end.equals(that.end);

		}

		@Override
		public int hashCode() {
			int result = addressRanges.hashCode();
			result = 31 * result + name.hashCode();
			result = 31 * result + start.hashCode();
			result = 31 * result + end.hashCode();
			return result;
		}
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		AddressLayoutSpec that = (AddressLayoutSpec) o;

		if (!addressSpaces.equals(that.addressSpaces)) {
			return false;
		}
		return autoMerges.equals(that.autoMerges);

	}


	public static class AddressRange {

		@XmlAttribute
		private BlockType type = BlockType.Unintializable;

		@XmlAttribute(required = true)
		private String name;

		@XmlJavaTypeAdapter(LongDecodeAdapter.class)
		@XmlAttribute(required = true)
		private Long start = 0L;

		@XmlJavaTypeAdapter(LongDecodeAdapter.class)
		@XmlAttribute(required = true)
		private Long end = 0L;

		@XmlIDREF
		@XmlAttribute(required = true)
		@JsonIgnore
		private Bus implementedBy;

		@XmlAttribute
		private boolean writable = false;

		@XmlAttribute
		private boolean executable = false;

		@XmlAttribute
		private boolean readable = false;

		public String getName() {
			return name;
		}

		public long getStart() {
			return start;
		}

		public long getEnd() {
			return end;
		}

		public Bus getImplementedBy() {
			return implementedBy;
		}

		public boolean isWritable() {
			return writable;
		}

		public boolean isExecutable() {
			return executable;
		}

		public boolean isReadable() {
			return readable;
		}

		public BlockType getType() {
			return type;
		}

		@Override
		public String toString() {
			return "AddressRange{" +
					"name='" + name + '\'' +
					", start=0x" + Long.toHexString(start) +
					", end=0x" + Long.toHexString(end) +
					", implementedBy=" + implementedBy +
					", readable="+ readable +
					", writable=" + writable +
					", executable=" + executable +
					'}';
		}


		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			AddressRange that = (AddressRange) o;

			if (writable != that.writable) {
				return false;
			}
			if (executable != that.executable) {
				return false;
			}
			if (type != that.type) {
				return false;
			}
			if (!name.equals(that.name)) {
				return false;
			}
			if (!start.equals(that.start)) {
				return false;
			}
			return end.equals(that.end);

		}

		@Override
		public int hashCode() {
			int result = type.hashCode();
			result = 31 * result + name.hashCode();
			result = 31 * result + start.hashCode();
			result = 31 * result + end.hashCode();
			result = 31 * result + (writable? 1 : 0);
			result = 31 * result + (executable? 1 : 0);
			return result;
		}
	}


	public static class AutoMergeInfo {

		@XmlAttribute
		private String space;

		@XmlAttribute
		private String range;

		@XmlAttribute
		private String intoSpace;

		@XmlAttribute
		private String intoRange;

		public String getSpace() {
			return space;
		}

		public String getRange() {
			return range;
		}

		public String getIntoSpace() {
			return intoSpace;
		}

		public String getIntoRange() {
			return intoRange;
		}

		@Override
		public String toString() {
			return "AutoMergeInfo{" +
					"space='" + space + '\'' +
					", range='" + range + '\'' +
					", intoSpace='" + intoSpace + '\'' +
					", intoRange='" + intoRange + '\'' +
					'}';
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			AutoMergeInfo that = (AutoMergeInfo) o;

			if (!space.equals(that.space)) {
				return false;
			}
			if (!range.equals(that.range)) {
				return false;
			}
			if (!intoSpace.equals(that.intoSpace)) {
				return false;
			}
			return intoRange.equals(that.intoRange);

		}

		@Override
		public int hashCode() {
			int result = space.hashCode();
			result = 31 * result + range.hashCode();
			result = 31 * result + intoSpace.hashCode();
			result = 31 * result + intoRange.hashCode();
			return result;
		}
	}


}
