/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory;


import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementRefs;
import javax.xml.bind.annotation.XmlMixed;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * Memory instance containing placeholders
 */
public class MemoryLocator {

	private void entryAccumulator(List<Object> l, Object objSrc) {
		Object obj;
		if (objSrc instanceof String) {
			obj = ((String) objSrc).trim();
		} else {
			obj = objSrc;
		}

		if (l.isEmpty()) {
			l.add(obj);
			return;
		}
		int idx = l.size() - 1;
		Object last = l.get(idx);
		if (last instanceof String && obj instanceof String) {
			l.remove(idx);
			l.add(idx, ((String) last) + ((String) obj));
		} else {
			l.add(idx + 1, obj);
		}
	}

	private List<Object> joinNeighbouringStrings(List<Object> original) {
		final Stream<Object> stream = original.stream();
		return joinNeighbouringStrings(stream);
	}

	private List<Object> joinNeighbouringStrings(Stream<Object> stream) {
		return stream.collect(ArrayList::new, this::entryAccumulator, (x, y) -> {
			throw new UnsupportedOperationException();
		});
	}

	/**
	 * Embedded reference to memory index
	 */
	@XmlRootElement(name = "memoryId")
	public static class MemoryId implements Serializable {

		public MemoryId() {
		}

		@Override
		public String toString() {
			return "$memory";
		}
	}


	/**
	 * Embedded reference to block index
	 */
	@XmlRootElement(name = "block")
	public static class Block implements Serializable {

		public Block() {
		}

		@Override
		public String toString() {
			return "$index";
		}
	}


	/**
	 * Embedded reference to block index
	 */
	@XmlRootElement(name = "byte")
	public static class Byte implements Serializable {

		public Byte() {
		}

		@Override
		public String toString() {
			return "$byte";
		}
	}


	@XmlMixed
	@XmlElementRefs(value = {
			@XmlElementRef(type = MemoryId.class),
			@XmlElementRef(type = Block.class),
			@XmlElementRef(type = Byte.class)
	})
	protected List<Object> content;

	public List<Object> getContent() {
		return content;
	}

	public MemoryLocator() {
	}

	public MemoryLocator(Stream<Object> content) {
		this.content = joinNeighbouringStrings(content);
	}

	public MemoryLocator(String prepend, MemoryLocator memoryLocator) {
		content = joinNeighbouringStrings(Stream.concat(Stream.of(prepend), memoryLocator.content.stream()));
	}


	@Override
	public String toString() {
		return "MemoryLocator:" + content;
	}


	/**
	 * Partially resolve the value, replacing only memory index
	 *
	 * @return the content with all values resolved
	 */
	public MemoryLocator resolveMemoryIndex(int memIdx) {
		return new MemoryLocator(content.stream().map(x -> {
			if (x instanceof MemoryId) {
				return Integer.toString(memIdx);
			}
			return x;
		}));
	}

	/**
	 * Resolve the value
	 *
	 * @return the content with all values resolved
	 * @param blockIndex n-th bram
	 * @param bramColumnIndex n-th column of bram
	 */
	public String resolve(int blockIndex, Integer bramColumnIndex) {
		final boolean[] sawByteId = {false};
		final String result = content.stream()
				.map(x -> {
					if (x instanceof Memory) {
						throw new RuntimeException("Unresolved memory id in " + this);
					} else if (x instanceof Block) {
						return Integer.toString(blockIndex);
					} else if (x instanceof Byte) {
						if (bramColumnIndex == null) {
							throw new RuntimeException("Unresolved byteid in memory that does not expect one: " + this);
						}
						sawByteId[0] = true;
						return String.valueOf((char)('A' + bramColumnIndex));
					} else {
						return x.toString().trim();
					}
				}).collect(Collectors.joining());
		if (bramColumnIndex != null && !sawByteId[0]) {
			throw new RuntimeException("MemoryLocator does not contain a byteID to replace in " + this);
		}
		return result;

	}

}

