/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.CalculatedValue;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IRedirectedResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ParameterReference;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.DefaultableBoolean;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.references.IReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.jaxb.NumberDecodeAdapter;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.adapters.XmlJavaTypeAdapter;

/**
 * Created by Ramon on 22.03.2016.
 */
public class Parameter extends ParameterReference implements IParameter {

	private static final Logger logger = LoggerFactory.getLogger(Parameter.class);

	@XmlElement
	@NotNull
	private String description;
	@XmlElement
	private Range range;
	@XmlAttribute
	private boolean simulationOnly = false;
	@XmlAttribute
	private boolean virtual = false;
	@XmlAttribute
	private String value;
	@XmlElement(name = "value", type = CalculatedValue.WrapperValue.class)
	private CalculatedValue valueRef;
	@XmlElement(name = "relevantIf")
	private CalculatedValue.BooleanWrapper relevantIfRef;
	@XmlAttribute(required = true)
	private ParameterType type;
	@XmlAttribute(name = "editable")
	private DefaultableBoolean internalEditable = DefaultableBoolean.Default;
	private boolean choicesOnly;
	private boolean editable;
	@XmlElement(name = "choice")
	private List<Choice> choices;//null is checked to decide on choicesOnly
	@XmlElement(name = "instanceHint")
	private List<String> instanceHints;//null is checked

	@XmlAttribute
	private String uiUnit;

	@XmlAttribute
	private String tclName;

	protected Parameter() {

	}

	public Parameter(String name) {
		this.name = name;
	}

	public Parameter(String name, CalculatedValue valueRef, ParameterType type) {
		this.name = name;
		this.valueRef = valueRef;
		this.internalEditable = DefaultableBoolean.False; // reffed params always uneditable
		this.editable = false;
		this.type = type;
	}

	public Parameter(String name, String value, ParameterType type) {
		this.name = name;
		this.value = value;
		this.type = type;
		editable = true;
		internalEditable = DefaultableBoolean.True;
	}

	public static Parameter getBackingParameter(ParameterReference parameterReference) {
		return ((Parameter) parameterReference);
	}

	@Override
	public boolean isDynamicallyReferenced() {
		return getReferencers() != null && getReferencers().stream().anyMatch(IReferencer::isDynamic);
	}

	@Override
	public List<String> getInstanceHints() {
		return instanceHints;
	}

	@Override
	public boolean hasInstanceHints() {
		return instanceHints != null;
	}

	@Override
	public CalculatedValue.BooleanWrapper getRelevantIfRef() {
		return relevantIfRef;
	}

	@Override
	public boolean hasRelevantIf() {
		return relevantIfRef != null;
	}

	@Override
	public CalculatedValueResult<Boolean> resolveRelevantIf(IValueResolver resolver) {
		if (!hasRelevantIf()) {
			return CalculatedValueResult.of(true);
		}
		return relevantIfRef.resolveBoolean(resolver);
	}

	@Override
	public boolean isSimulationOnly() {
		return simulationOnly;
	}

	@Override
	public String getDescription() {
		return description;
	}

	@Override
	public Range getRange() {
		return range;
	}

	@Override
	public String getDefaultValue() {
		return value;
	}

	@Override
	public boolean hasCalculatedValue() {
		return valueRef != null;
	}

	@Override
	public CalculatedValue getValueRef() {
		return valueRef;
	}

	@Override
	public CalculatedValueResult<? extends Number> resolveValue(IValueResolver resolver) {
		if (valueRef != null) {
			return valueRef.resolve(resolver);
		} else {
			throw new IllegalStateException("No Value to resolve");
		}
	}

	@Override
	public ParameterType getType() {
		return type;
	}

	@Override
	public boolean isChoicesOnly() {
		return choicesOnly;
	}

	@Override
	public boolean hasChoices() {
		return choices != null;
	}

	@Override
	public boolean isEditable() {
		return editable;
	}

	@Override
	public boolean isVirtual() {
		return virtual;
	}

	@Override
	public List<Choice> getChoices() {
		return choices;
	}

	@Override
	public String getUiUnit() {
		return uiUnit;
	}

	@Override
	public boolean hasUiUnit() {
		return uiUnit != null;
	}

	@Override
	public boolean isRedirected() {
		return false;
	}

	@Override
	public String getTclName() {
		return tclName;
	}

	@Override
	public CalculatedValueResult<? extends Number> resolveRedirected(IRedirectedResolver resolver) {
		throw new UnsupportedOperationException("Baum");
	}

	private void afterUnmarshal(Unmarshaller u, Object parent) {
		boolean dynamic = hasCalculatedValue();

		if (dynamic) {
			editable = false;
			if (internalEditable == DefaultableBoolean.True) {
				throw new IllegalStateException("CalculatedValue cannot be editable!");
			} else if (internalEditable == DefaultableBoolean.Default) {
				logger.trace("Parameter {} has calculated value but is not marked uneditable, overriding", getName());
				internalEditable = DefaultableBoolean.False;
			}
		}

		if (internalEditable == DefaultableBoolean.Default) {
			choicesOnly = (getChoices() != null);
			editable = true;
		} else if (internalEditable == DefaultableBoolean.True) {
			choicesOnly = false;
			editable = true;
		} else {
			choicesOnly = (getChoices() != null);
			editable = (getChoices() != null);
		}
	}

	@Override
	public String toString() {
		return "Parameter{" +
				"name='" + name + '\'' +
				", range=" + range +
				", value='" + value + '\'' +
				", type=" + type +
				", choicesOnly=" + choicesOnly +
				", choices=" + choices +
				'}';
	}

	public static class Range {

		@XmlAttribute
		@XmlJavaTypeAdapter(type = long.class, value = NumberDecodeAdapter.class)
		private Number min;

		@XmlAttribute
		@XmlJavaTypeAdapter(type = long.class, value = NumberDecodeAdapter.class)
		private Number max;

		public Number getMin() {
			return min;
		}

		public Number getMax() {
			return max;
		}

		public boolean isInRange(long value) {
			return value >= min.longValue() && value <= max.longValue();
		}

		public boolean isInRange(double value) {
			return value >= min.doubleValue() && value <= max.doubleValue();
		}

		@Override
		public String toString() {
			return "Range{" +
					"min=" + min +
					", max=" + max +
					'}';
		}
	}


	public static class Choice {

		@XmlAttribute
		private String value;

		//TODO add configuration? was previously toolchain

		@XmlAttribute
		private String descr;

		public String getDescr() {
			return descr;
		}

		public String getValue() {
			return value;
		}

		@Override
		public String toString() {
			return "Choice{" +
					"value='" + value + '\'' +
					", descr='" + descr + '\'' +
					'}';
		}
	}
}
