/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.files.buses;

import de.tu_darmstadt.rs.spartanmc.devxml.DevXmlNamespace;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.IPortMapping;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.CalculatedValue;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.Parameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPortGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.files.base.XmlLoadedFile;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlEnumValue;
import javax.xml.bind.annotation.XmlRootElement;

import static de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus.Role.Master;
import static de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus.Role.Slave;

/**
 * Created by Ramon on 07.02.2016.
 */
@XmlRootElement(name = "bus", namespace = DevXmlNamespace.NAMESPACE)
public class BusDescription extends XmlLoadedFile {

	@XmlAttribute
	private String name;
	@XmlAttribute
	private BusType busType;
	@XmlElementWrapper(name = "definingParameters")
	@XmlElement(name = "parameter", namespace = DevXmlNamespace.NAMESPACE)
	private List<Parameter> definingParameters = new ArrayList<>();
	@XmlElement
	private BusPortGroup masterPorts;
	@XmlElement
	private BusPortGroup slavePorts;


	public List<Parameter> getDefiningParameters() {
		return definingParameters;
	}

	@Override
	public String toString() {
		return "BusDescription{" +
				"name='" + name +
				", type=" + busType + "\'}";
	}

	//	public Parameter getParameter(String busName) {
	//		return getMasterPorts().getParameters().stream().filter(p->p.getName().equals(busName)).findAny().get();
	//	}

	public BusPortGroup getMatchingPorts(Bus bus) {
		return getMatchingPorts(bus.getRole());
	}


	public BusPortGroup getMatchingPorts(Bus.Role role) {
		return role == Master? getMasterPorts() : getSlavePorts();
	}

	public BusPortGroup getOppositePorts(Bus bus) {
		return bus.getRole() == Master? getSlavePorts() : getMasterPorts();
	}

	public boolean isDefinedParameter(String name, Bus.Role role) {
		return isBuiltinParam(name, role) || definingParameters.stream().anyMatch(p -> p.getName().equals(name));
	}

	private boolean isBuiltinParam(String name, Bus.Role role) {
		switch (name) {
			case Bus.COUNT_PARAM_NAME:
				return (busType == BusType.MultiMaster
						|| (busType == BusType.SingleMaster && role == Master));
			case Bus.ID_PARAM_NAME:
				return (busType == BusType.MultiMaster
						|| (busType == BusType.SingleMaster && role == Slave));
			default:
				return false;
		}
	}

	public BusPort getOppositePort(Bus.Role portRole, BusPort port) {
		if (portRole == Master) {
			return slavePorts.getPort(port.getMapping()).orElseThrow(() -> new RuntimeException("could not find " + port.getMapping() + " on slave ports of " + getName()));
		} else {
			String target = port.getMapping();
			return masterPorts.ports.stream().filter(p -> p.getMapping().equals(target)).findAny().orElseThrow(() -> new RuntimeException("could not find " + target + " on master ports of " + getName()));
		}
	}

	public BusPort getOppositePort(Bus bus, BusPort port) {
		return getOppositePort(bus.getRole(),port);
	}

	public BusPort getOppositeBusPortDescription(Bus bus, IPortMapping portMapping) {
		return getOppositePort(bus, getPort(bus, portMapping.getBusName()));
	}

	public BusPort getPort(Bus bus, String busName) {
		if (bus.getRole() == Master) {
			return masterPorts.getPort(busName).orElseThrow(() -> new RuntimeException("could not find " + busName + " on master ports of " + bus));
		} else {
			return slavePorts.getPort(busName).orElseThrow(() -> new RuntimeException("could not find " + busName + " on slave ports of " + bus));
		}
	}

	public int getMaxConnections(Bus.Role role) {
		switch (busType) {
			case PeerToPeer:
				return 1;
			case SingleMaster:
				return role == Master? 0 : 1;
			case MultiMaster:
				return 0;
			default:
				throw new IllegalStateException("Invalid busType " + busType);
		}
	}

	@Override
	public String getName() {
		return name;
	}

	public BusType getBusType() {
		return busType;
	}

	public BusPortGroup getMasterPorts() {
		return masterPorts;
	}

	public BusPortGroup getSlavePorts() {
		return slavePorts;
	}

	@XmlEnum
	public enum BusType {
		@XmlEnumValue("single-master")SingleMaster,
		@XmlEnumValue("multi-master")MultiMaster,
		@XmlEnumValue("peer-to-peer")PeerToPeer
	}


	public static class BusPortGroup implements IPortGroup {

		@XmlElementRef(type = BusPort.class)
		private List<BusPort> ports;

		@XmlAttribute
		private ConnectorDirection direction;

		//		@XmlElement(name = "masterSetParameter")
		//		private List<Parameter> parameters = new ArrayList<>();

		@Override
		public List<BusPort> getPorts() {
			return ports;
		}

		@Override
		public String getName() {
			return "BusPortGroup";
		}

		@Override
		public ConnectorDirection getDirection() {
			return direction;
		}

		//		public List<Parameter> getParameters() {
		//			return parameters;
		//		}

		public Optional<BusPort> getPort(String busName) {
			return ports.stream().filter(p -> p.getName().equals(busName)).findAny();
		}
	}


	/**
	 * Not a full ClkPort, because it serves only to verify merging.
	 * Frequency cannot be set inside a bus (a bus is only theoretical after all), only via setClk on the bus implementation
	 */
	@XmlRootElement(name = "clkPort", namespace = DevXmlNamespace.NAMESPACE)
	public static class BusClkPort extends BusPort {

		@Override
		public Type getType() {
			return Type.Clk;
		}

		@Override
		public int getConstantWidth() {
			return 1;
		}

		@Override
		public boolean hasCalculatedWidth() {
			return false;
		}

		@Override
		public boolean hasDynamicWidth() {
			return false;
		}

		@Override
		public CalculatedValue getCalculatedWidth() {
			return null;
		}

		@Override
		public CalculatedValueResult<Long> resolveWidth(IValueResolver resolver) {
			return CalculatedValueResult.of(1L);
		}
	}

	//	private static class PerConnection extends CalculatedValue {
	//		private CalculatedValue value;
	//
	//		@XmlElementRef
	//		private void setWidth(CalculatedValue value) {
	//			this.value = value;
	//		}
	//
	//		@XmlAttribute(name = "width")
	//		private void setConstantWidth(long width) {
	//			value = new Constant(width);
	//		}
	//
	//		@Override
	//		public long resolve(IValueResolver resolver) {
	//			return value.resolve(resolver);
	//		}
	//
	//		@Override
	//		public double resolveFloat(IValueResolver resolver) {
	//			return value.resolveFloat(resolver);
	//		}
	//
	//		@Override
	//		public boolean isDynamic() {
	//			return value != null && value.isDynamic();
	//		}
	//
	//		@Override
	//		public String toString() {
	//			return value!=null? value.toString() : "{Undefined}";
	//		}
	//
	//		@Override
	//		public boolean equals(Object o) {
	//			if (this == o) return true;
	//			if (o == null || getClass() != o.getClass()) return false;
	//
	//			WrapperValue that = (WrapperValue) o;
	//
	//			return value != null ? value.equals(that.value) : that.value == null;
	//
	//		}
	//
	//		@Override
	//		public int hashCode() {
	//			return value != null ? value.hashCode() : 0;
	//		}
	//	}


}
