/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.files.devices;

import de.tu_darmstadt.rs.spartanmc.devxml.DevXmlNamespace;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.INamed;
import de.tu_darmstadt.rs.spartanmc.devxml.files.base.XmlLoadedFile;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlEnumValue;
import javax.xml.bind.annotation.XmlID;
import javax.xml.bind.annotation.XmlIDREF;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlValue;

/**
 * A FPGA device
 */
@XmlRootElement(name = "device", namespace = DevXmlNamespace.NAMESPACE)
public class DeviceDescription extends XmlLoadedFile implements INamed {

	@XmlElement(name = "toolchain")
	private List<Toolchain> toolchains;

	@XmlElementWrapper(name = "iostandards")
	@XmlElement(name = "iostd")
	private List<String> iostandards = new ArrayList<>();

	@XmlElementWrapper(name = "drives")
	@XmlElement(name = "drive")
	private List<String> drives = new ArrayList<>();

	@XmlElementWrapper(name = "slewRates")
	@XmlElement(name = "slewRate")
	private List<String> slewRates = new ArrayList<>();

	@XmlAttribute
	private String name;

	@XmlElement
	private Family family;

	@XmlElement
	private Terminations terminations;

	public DeviceDescription() {
	}

	public DeviceDescription(List<Toolchain> toolchain, String name) {
		this.toolchains = toolchains;
		this.name = name;
	}

	public List<String> getSlewRates() {
		return slewRates;
	}

	@Override
	public String getName() {
		return name;
	}

	public String getFamily() {
		return family.getName();
	}

	public List<Toolchain> getToolchains() {
		return toolchains;
	}

	public List<String> getIostandards() {
		return iostandards;
	}

	public List<String> getDrives() {
		return drives;
	}

	public boolean isPulldownSupported() {
		return terminations != null && terminations.pulldowns;
	}

	public boolean isPullupSupported() {
		return terminations != null && terminations.pullups;
	}

	/**
	 * Get the terminations supported on this FPGA
	 *
	 * @return the supported terminations
	 */
	public List<Termination> getTerminations() {
		List<Termination> res = new ArrayList<>();
		res.add(Termination.None);
		if (isPulldownSupported()) {
			res.add(Termination.PullDown);
		}
		if (isPullupSupported()) {
			res.add(Termination.PullUp);
		}
		return res;
	}

	public Toolchain getToolchain(String toolchain) {
		for (Toolchain t : toolchains) {
			if (t.getName().equals(toolchain)) {
				return t;
			}
		}
		throw new IllegalArgumentException("Unsupported toolchain: " + toolchain);
	}


	/**
	 * A termination
	 */
	@XmlEnum
	public enum Termination {
		@XmlEnumValue("none") None,
		@XmlEnumValue("pulldown") PullDown,
		@XmlEnumValue("pullup") PullUp
	}


	private static class Terminations {

		@XmlAttribute
		private boolean pulldowns = false;

		@XmlAttribute
		private boolean pullups = false;

		public Terminations() {
		}
	}


	public static class Family {

		@XmlAttribute
		private String name;

		public String getName() {
			return name;
		}

		public Family() {
		}

		public Family(String name) {
			this.name = name;
		}
	}


	public enum SimlibAssociation {
		@XmlEnumValue("both")
		Both,
		@XmlEnumValue("functional")
		Functional,
		@XmlEnumValue("timing")
		Timing;

		public boolean isIncludedIn(SimlibAssociation association) {
			if (association == Both) {
				return true;
			}
			if (this == Both) {
				return true;
			}
			return this == association;
		}
	}


	public static class Simlib implements INamed {

		@XmlValue
		private String name;

		@XmlAttribute
		private SimlibAssociation association = SimlibAssociation.Both;

		@Override
		public String getName() {
			return name;
		}

		public SimlibAssociation getAssociation() {
			if (association == null) {
				throw new NullPointerException("Simlib " + name + " has no association");
			}
			return association;
		}
	}


	public static class Toolchain implements INamed {

		@XmlAttribute
		@XmlID
		private String name;

		@XmlAttribute
		private String toolchain;

		@XmlElement(name = "param")
		private List<DevParam> params;


		@XmlElement(name = "simlib")
		private List<Simlib> simlibs = new ArrayList<>();

		private Map<SimlibAssociation, List<String>> simlibsWithParent = new HashMap<>();

		private Map<String, DevParam> paramsWithParent = null;

		@XmlAttribute
		@XmlIDREF
		private Toolchain parent;

		public Toolchain() {
		}

		public Toolchain(String name) {
			this.name = name;
		}

		public String getName() {
			return name;
		}

		public String getToolchain() {
			if (toolchain == null) {
				return name;
			}
			return toolchain;
		}

		private Stream<String> getOwnSimlibs(SimlibAssociation association) {
			return simlibs.stream().filter(s -> s.getAssociation().isIncludedIn(association)).map(Simlib::getName);
		}

		private Stream<String> getSimlibsWithParent(SimlibAssociation association) {
			if (parent != null) {
				return Stream.concat(parent.getSimlibsWithParent(association), getOwnSimlibs(association));
			} else {
				return getOwnSimlibs(association);
			}
		}

		public Stream<String> getSimlibsAll() {
			return getSimlibsWithParent(SimlibAssociation.Both);
		}

		public Stream<String> getSimlibsFunctional() {
			return getSimlibsWithParent(SimlibAssociation.Functional);
		}

		public Stream<String> getSimlibsTiming() {
			return getSimlibsWithParent(SimlibAssociation.Timing);
		}

		public Map<String, DevParam> getParams() {
			if (paramsWithParent == null) {
				if (parent == null) {
					paramsWithParent = new HashMap<>();
				} else {
					paramsWithParent = new HashMap<>(parent.getParams());
				}
				if (params != null) {
					params.forEach(param -> paramsWithParent.put(param.name, param));
				}
			}
			return Collections.unmodifiableMap(paramsWithParent);
		}

		@Override
		public String toString() {
			return name;
		}
	}

	@Override
	public String toString() {
		return name;
	}

	@XmlEnum
	public enum ParamOverrideOption {
		@XmlEnumValue("replace")
		Replace,
		@XmlEnumValue("prepend")
		Prepend,
		@XmlEnumValue("append")
		Append
	}


	/**
	 * A parameter that the toolchain has
	 */
	public static class DevParam {

		@XmlAttribute
		protected String name;

		@XmlAttribute
		private String value;

		@XmlValue
		private String multiLineValue;

		@XmlAttribute
		private boolean hidden;

		@XmlAttribute(name = "readonly")
		private boolean readOnly;

		@XmlAttribute
		private ParamOverrideOption override;

		private Toolchain toolchain;


		private void afterUnmarshal(Unmarshaller u, Object parent) {
			this.toolchain = (Toolchain) parent;
		}

		public DevParam() {
		}

		public String getName() {
			return name;
		}

		private String getOwnValue() {
			if (value != null) {
				return value;
			} else {
				return multiLineValue;
			}
		}

		private String getParentValue() {
			final Toolchain parentToolchain = toolchain.parent;
			if (parentToolchain == null) {
				throw new RuntimeException("To append or prepend, toolchain needs a parent, but " + toolchain + " has none");
			}

			return parentToolchain.getParams().get(name).getValue();
		}

		public String getValue() {
			if (override == null) {
				return getOwnValue();
			}

			switch (override) {
				case Prepend:
					return getOwnValue() + getParentValue();
				case Append:
					return getParentValue() + getOwnValue();
				default:
					return getOwnValue();
			}
		}

		public boolean isHidden() {
			return hidden;
		}

		public boolean isReadOnly() {
			return readOnly;
		}

		@Override
		public String toString() {
			return "Param{" +
					"name='" + name + '\'' +
					", value='" + value + '\'' +
					", multiLineValue='" + multiLineValue + '\'' +
					", hidden=" + hidden +
					", readOnly=" + readOnly +
					'}';
		}
	}
}
