/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.files.modules;

import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfigurationSelector;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.CombineType;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.IClkPortMapping;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.CalculatedValue;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.ParamRef;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IRedirectedResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.PortReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ClkReference;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ParamRefAnalyzer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.FirmwareDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HdlDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HeaderDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.TclDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.BaseParameterGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.Parameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.BasePortGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.ClkPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPortGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.Port;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by wenzel on 14.06.16.
 */
public class BusBreakoutModuleDescription implements IModuleDescription {

	public static final String BUS_SUFFIX = "_bus";
	private static final Logger logger = LoggerFactory.getLogger(BusBreakoutModuleDescription.class);
	private final HeaderDescription header;
	private final HdlDescription hdl;
	private final Bus bus;
	private final BusDescription busDescription;
	private final Bus.Role role;
	private final IPortGroup invertedPortGroup;
	private final InternalClkPortMapping clkPortMapping;
	private final Map<BusPort, IPort> portMap;
	private final Map<String, Parameter> busParameters;
	private final boolean outputFreqToBus;

	public BusBreakoutModuleDescription(BusDescription busDescription, Bus.Role role) {

		this.busDescription = busDescription;
		this.role = role;

		this.header = new HeaderDescription(getModuleName(busDescription, role),
				"Bus Breakout",
				"Bus Breakout",
				Collections.emptyList(),
				"Convert the bus " + busDescription.getName() + " (" + role + ") into individual wires");

		this.hdl = new HdlDescription("verilog", "bus_breakout_invalid", new HdlDescription.Sources(null, Collections.emptyList(), Collections.emptyList(), Collections.emptyList(), Collections.emptyList()));

		outputFreqToBus = busDescription.getMatchingPorts(role).getPorts().stream()
				.filter(p -> p instanceof BusDescription.BusClkPort)
				.map(p -> p.getDirection() == ConnectorDirection.Output)
				.collect(StreamUtil.AllOrAnyMatch.collectorSomeForbidden())
				.orElse(role == Bus.Role.Master);

		if (!outputFreqToBus) {
			this.clkPortMapping = busDescription.getSlavePorts().getPorts().stream().flatMap(StreamUtil.castOrEmpty(BusDescription.BusClkPort.class))
					.collect(StreamUtil.atMostOne())
					.map(clkPort -> new InternalClkPortMapping(clkPort.getName(), clkPort.getName() + BUS_SUFFIX))
					.orElse(null);
		} else {
			this.clkPortMapping = null;
		}

		this.busParameters = busDescription.getDefiningParameters().stream()
				.map(BusBreakoutModuleDescription::cloneDefiningParam)
				.collect(Collectors.toMap(Parameter::getName, Function.identity()));

		portMap = getPortGroup().getPorts().stream().collect(Collectors.toMap(Function.identity(), p -> {

			if (p instanceof BusDescription.BusClkPort) {
				CalculatedValue freqRef = null;
				if (!outputFreqToBus) {
					freqRef = new CalculatedValue.ClkPortRef(clkPortMapping);
				}
				return new ClkPort(p.getName(),
						p.getDirection().getReverse(),
						p.getDescription(),
						p.isRequired(),
						freqRef
				);
			} else {
				Port result = new Port(p.getName(),
						p.getDirection().getReverse(),
						p.getDescription(),
						p.isRequired(),
						p.getConstantWidth(),
						cloneWithReplacedParamRefs(p.getCalculatedWidth())
				);

				if (result.getCalculatedWidth() != null) {
					result.resolveWidth(new ParamRefAnalyzer<>(result, PortReferencer::new));
				}
				return result;
			}
		}));

		invertedPortGroup = new BasePortGroup() {
			@Override
			public List<? extends IPort> getPorts() {
				return portMap.values().stream().collect(Collectors.toList());
			}
		};


		List<Bus.SetParameter> setParameters = busParameters.values().stream().map(p -> new Bus.SetParameter(p.getName(), new ParamRef(p))).collect(Collectors.toList());


		this.bus = new Bus(
				busDescription.getName(),
				busDescription.getName(),
				role,
				"unlimited",
				setParameters,
				generateSetClks(),
				Collections.emptyList(),
				clkPortMapping != null? Collections.singletonList(clkPortMapping) : Collections.emptyList(),
				generatePortMappings()
		);

		if (clkPortMapping != null) {
			this.clkPortMapping.parent = bus;
		}


	}

	private static Parameter cloneDefiningParam(Parameter parameter) {
		if (parameter.hasCalculatedValue()) {//TODO maybe copy constraints too?
			return new Parameter(parameter.getName(), parameter.getValueRef(), parameter.getType());
		} else {
			return new Parameter(parameter.getName(), parameter.getDefaultValue(), parameter.getType());
		}
	}

	public static String getModuleName(BusDescription busDescription, Bus.Role role) {
		return "breakout-" + busDescription.getName() + "-" + role.toString().toLowerCase();
	}

	private CalculatedValue cloneWithReplacedParamRefs(CalculatedValue calculatedWidth) {
		if (calculatedWidth == null) {
			return null;
		}

		return calculatedWidth.cloneWithMapper(val -> {
			if (val instanceof CalculatedValue.Constant) {
				return val; //No need to clone constancts
			}
			if (val instanceof ParamRef) {
				ParamRef paramRef = (ParamRef) val;
				return new ParamRef(busParameters.get(paramRef.getParam().getName()));
			} else {
				throw new UnsupportedOperationException("Can only clone with paramRefs");
			}
		});
	}

	@Override
	public boolean hasConfigurations() {
		return false;
	}

	@Override
	public Stream<? extends IToolchainConfigurationSelector> streamConfigurations() {
		return Stream.empty();
	}

	public BusDescription getBusDescription() {
		return busDescription;
	}

	public Bus.Role getBusRole() {
		return role;
	}

	private List<Bus.PortMapping> generatePortMappings() {
		return getPortGroups().stream()
				.flatMap(pg -> pg.getPorts().stream())
				.filter(p -> p.getType() != IConnectorDescription.Type.Clk || outputFreqToBus)
				.map(p -> new Bus.PortMapping(p.getName(), p.getName() + BUS_SUFFIX, false, CombineType.Auto))
				.collect(Collectors.toList());
	}

	private List<Bus.SetClk> generateSetClks() {
		if (role == Bus.Role.Slave) {
			return Collections.emptyList();
		} else {
			final List<Bus.SetClk> collect = busDescription.getMasterPorts().getPorts().stream()
					.flatMap(StreamUtil.castOrEmpty(BusDescription.BusClkPort.class))
					.map(port -> new Bus.SetClk(port.getName(), new CalculatedValue.ClkPortRef((ClkPort) portMap.get(port))))
					.collect(Collectors.toList());
			logger.info("Generated setclks for " + getName() + ": " + collect);
			return collect;
		}
	}

	@Override
	public ModuleType getType() {
		return ModuleType.BusBreakout;
	}

	@Override
	public HeaderDescription getHeader() {
		return header;
	}

	@Override
	public List<BaseParameterGroup> getParameterGroups() {
		return Collections.singletonList(new BaseParameterGroup() {
			@Override
			public List<? extends IParameter> getParameters() {
				Optional<ClkReference> clkPort;
				if (role == Bus.Role.Slave) {
					clkPort = Optional.ofNullable(clkPortMapping);
				} else {
					try {
						clkPort = getInvertedPortGroup().getPorts().stream().flatMap(StreamUtil.castOrEmpty(ClkPort.class)).collect(StreamUtil.atMostOne());
					} catch (RuntimeException e) {
						throw new RuntimeException("fail in " + getName(), e);
					}
				}

				List<IParameter> res = new ArrayList<IParameter>();
				clkPort.ifPresent(clkP -> {
					logger.info("for " + BusBreakoutModuleDescription.this.getName() + ", adding parameter for " + clkP);
					res.add(new Parameter("Clock Freq", new CalculatedValue.ClkPortRef(clkP), IParameter.ParameterType.IntType));
				});


				res.addAll(busParameters.values());


				return res;
			}
		});
	}

	@Override
	public List<IPortGroup> getPortGroups() {
		//Return opposite ports
		return Collections.singletonList(getInvertedPortGroup());
	}

	@Override
	public Optional<Bus> findBus(String busName) {
		if (bus.getName().equals(busName)) {
			return Optional.of(bus);
		} else {
			return Optional.empty();
		}
	}

	@Override
	public List<Bus> getBuses() {
		return Collections.singletonList(bus);
	}

	@Override
	public HdlDescription getHdl() {
		return hdl;
	}

	public TclDescription getTcl() {
		return null;
	}

	@Override
	public FirmwareDescription getFirmwareDescription() {
		// return an empty list
		return new FirmwareDescription(new ArrayList<>());
	}

	@Override
	public Path getSourceFile() {
		return busDescription.getSourceFile();
	}

	@Override
	public Path getSourceTree() {
		return busDescription.getSourceTree();
	}

	@Override
	public Optional<String> getSourceFilePrefix() {
		return Optional.empty();
	}

	private BusDescription.BusPortGroup getPortGroup() {
		return busDescription.getMatchingPorts(role);
	}

	private IPortGroup getInvertedPortGroup() {
		return invertedPortGroup;
	}


	private static class InternalClkPortMapping extends ClkReference implements IClkPortMapping {

		private final String busName;
		private Bus parent;

		public InternalClkPortMapping(String busName, String name) {
			super(name);
			this.busName = busName;
		}

		@Override
		public boolean isRedirected() {
			return true;
		}

		@Override
		public CalculatedValueResult<? extends Number> resolveRedirected(IRedirectedResolver redirector) {
			return redirector.resolveAgainstBus(parent, this);
		}

		@Override
		public CalculatedValueResult<? extends Number> resolveFrequency(IValueResolver resolver) {
			throw new UnsupportedOperationException();
		}

		@Override
		public CalculatedValue getFrequencyRef() {
			return null;
		}

		@Override
		public boolean isAllowMerging() {
			return false;
		}

		@Override
		public CombineType getCombineUsing() {
			return CombineType.Auto;
		}

		@Override
		public String getBusName() {
			return busName;
		}

		@Override
		public String toString() {
			return "InternalClkPortMapping{" +
					"busName='" + busName + '\'' +
					", parent=" + parent +
					"} " + super.toString();
		}
	}
}
