/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.devxml.files.modules;

import de.tu_darmstadt.rs.spartanmc.devxml.DevXmlNamespace;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.MixedParamRefs;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.DynamicReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.ParamReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.PortReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.DynamicParamRefAnalyzer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dynamic.IDynamicParameterResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.IDynamicContainer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameterGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.dynamic.DynamicParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.dynamic.DynamicParameterGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPortGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.dynamic.DynamicPortGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.dynamic.IDynamicPort;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * Created by Ramon on 16.02.2016.
 */
@XmlRootElement(name = "dynamic-structure-module", namespace = DevXmlNamespace.NAMESPACE)
public class DynamicStructureModuleDescription extends ModuleDescription {

	private static final Logger logger = LoggerFactory.getLogger(DynamicStructureModuleDescription.class);
	@XmlElementRef(type = DynamicParameterGroup.class, namespace = DevXmlNamespace.NAMESPACE)
	private List<DynamicParameterGroup> parameterGroups;
	@XmlElementRef(type = DynamicPortGroup.class)
	private List<DynamicPortGroup> portGroups;
	@XmlElement
	private Coregen coregen;

	@Override
	public ModuleType getType() {
		return ModuleType.DynamicStructure;
	}

	@Override
	public List<DynamicParameterGroup> getParameterGroups() {
		throw new UnsupportedOperationException();
	}

	@Override
	public List<IPortGroup> getPortGroups() {
		throw new UnsupportedOperationException();
	}

	@Override
	public void analyzeParamRefUsage() {
		Stream.concat(parameterGroups.stream(), portGroups.stream()).filter(IDynamicContainer::hasDynamicImplementationCount).forEach(this::findParamUsages);
		parameterGroups.stream().filter(DynamicParameterGroup::isDynamicallyResolvable).flatMap(pg -> pg.getParameters().stream()).forEach(this::findParamUsages);
		portGroups.stream().filter(DynamicPortGroup::isDynamicallyResolvable).flatMap(pg -> pg.getPorts().stream()).forEach(this::findParamUsages);
	}

	private void findParamUsages(IDynamicContainer container) {
		try {
			container.resolveImplementationCount(new DynamicParamRefAnalyzer<>(container, DynamicReferencer::new));
		} catch (Exception e) {
			logger.error("Failed to resolve implementation count on container {}", container, e);
			throw new IllegalStateException("Could not resolve calculated parameter", e);
		}
	}

	private void findParamUsages(DynamicParameter parameter) {
		try {
			DynamicParamRefAnalyzer<DynamicParameter> analyzer = new DynamicParamRefAnalyzer<>(parameter, ParamReferencer::new);
			parameter.isEditable(analyzer);
			parameter.isEnabled(analyzer);
			if (parameter.hasCalculatedValue()) {
				parameter.resolveValue(analyzer);
			}
		} catch (Exception e) {
			logger.error("Failed to resolve calculated value on parameter '{}'", parameter.getName(), e);
			throw new IllegalStateException("Could not resolve calculated parameter", e);
		}
	}

	private void findParamUsages(IDynamicPort port) {
		try {
			DynamicParamRefAnalyzer<IDynamicPort> analyzer = new DynamicParamRefAnalyzer<>(port, PortReferencer::new);
			port.isEnabled(analyzer);
			if (port.hasCalculatedWidth()) {
				port.resolveWidth(analyzer);
			}
		} catch (Exception e) {
			logger.error("Failed to resolve calculated value on port '{}'", port.getName(), e);
			throw new IllegalStateException("Could not resolve calculated parameter", e);
		}
	}

	public List<IParameterGroup> getStaticParameterGroups(IDynamicParameterResolver resolver) {
		return parameterGroups.stream().flatMap(g -> g.getStaticParameterGroups(resolver).stream()).collect(Collectors.toList());
	}

	public List<IPortGroup> getStaticPortGroups(IDynamicParameterResolver resolver) {
		return portGroups.stream().flatMap(g -> g.getStaticPortGroups(resolver).stream()).collect(Collectors.toList());
	}


	public static class Coregen {

		@XmlElement
		public String description;

		@XmlElement
		public Select select;

		@XmlElement(name = "cset")
		public List<CSet> cSets;

		public String getDescription() {
			return description;
		}

		public Select getSelect() {
			return select;
		}

		public List<CSet> getcSets() {
			return cSets;
		}
	}


	public static class CSet {

		@XmlElement
		private String name;

		@XmlElement
		private MixedParamRefs value;

		@XmlElement(name = "default")
		private String defaultVal;

		public MixedParamRefs getValue() {
			return value;
		}

		public String getName() {
			return name;
		}

		public String getDefaultVal() {
			return defaultVal;
		}

		@Override
		public String toString() {
			return "CSet{" +
					"name='" + name + '\'' +
					", value=" + value +
					", defaultVal='" + defaultVal + '\'' +
					'}';
		}
	}


	public static class Select {

		@XmlAttribute
		private String value;

		public String getValue() {
			return value;
		}
	}

	public Coregen getLegacyCoregen() {
		return coregen;
	}

}
