/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 *
 */
package de.tu_darmstadt.rs.spartanmc.devxml.files.modules;

import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfigurationSelector;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.ClkPortReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.ParamReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.PortReferencer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ParamRefAnalyzer;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.FirmwareDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HdlDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HeaderDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.TclDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.ClkPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.devxml.files.base.XmlLoadedFile;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.stream.Stream;
import javax.xml.bind.annotation.XmlElement;

/**
 * Base JAXB implementation of all Module xmls.
 * Maps directly to schema/base-module-description.xsd
 */
public abstract class ModuleDescription extends XmlLoadedFile implements IModuleDescription {

	private static final Logger logger = LoggerFactory.getLogger(ModuleDescription.class);

	@XmlElement
	private HeaderDescription header;

	@XmlElement
	private HdlDescription hdl;

	@XmlElement
	private TclDescription tcl;

	@XmlElement
	private FirmwareDescription firmware;

	@XmlElement(name = "bus")
	private List<Bus> buses = new ArrayList<>();

	@Override
	public HeaderDescription getHeader() {
		return header;
	}

	@Override
	public FirmwareDescription getFirmwareDescription() {
		return firmware;
	}

	@Override
	public Optional<Bus> findBus(String busName) {
		return getBuses().stream().filter(bus -> bus.getName().equals(busName)).findAny();
	}

	@Override
	public List<Bus> getBuses() {
		return buses;
	}

	@Override
	public HdlDescription getHdl() {
		return hdl;
	}

	@Override
	public TclDescription getTcl() {
		return tcl;
	}

	@Override
	public String getName() {
		return IModuleDescription.super.getName();
	}

	@Override
	public boolean supports(IToolchainConfiguration currentConfig) {
		return !hasConfigurations() || streamConfigurations().anyMatch(sel -> sel.supports(currentConfig));
	}

	@Override
	public boolean hasConfigurations() {
		return !getHeader().getConfigurations().isEmpty();
	}

	@Override
	public Stream<? extends IToolchainConfigurationSelector> streamConfigurations() {
		return getHeader().getConfigurations().stream();
	}

	/**
	 * Hook that allows subclasses to index / optimize their internal data. Call is not guaranteed (depends on how library is built)
	 * If optimize is not called, stuff like indexing needs either be done on first use, or circumvented entirely
	 */
	public void optimize() {

	}

	private void findClkUsages(ClkPort port) {
		try {
			port.resolveFrequency(new ParamRefAnalyzer<>(port, ClkPortReferencer::new));
		} catch (Exception e) {
			logger.error("Failed to resolve frequency on port {}", port.getName(), e);
			throw new IllegalStateException("Could not resolve calculated frequency", e);
		}
	}

	private void findPortUsages(IPort port) {
		try {
			port.resolveWidth(new ParamRefAnalyzer<>(port, PortReferencer::new));
		} catch (Exception e) {
			logger.error("Failed to resolve calculated-value on port {}", port.getName(), e);
			throw new IllegalStateException("Could not resolve calculated port width", e);
		}
	}

	private void findParamUsages(IParameter parameter) {
		try {
			ParamRefAnalyzer<IParameter> refAnalyzer = new ParamRefAnalyzer<>(parameter, ParamReferencer::new);
			parameter.resolveValue(refAnalyzer);
			parameter.resolveRelevantIf(refAnalyzer);
		} catch (Exception e) {
			logger.error("Failed to resolve calculated-value on parameter {}", parameter.getName(), e);
			throw new IllegalStateException("Could not resolve calculated parameter", e);
		}
	}

	private void findBusUsages(Bus bus) {
		try {
			ParamRefAnalyzer<Bus> refAnalyzer = new ParamRefAnalyzer<>(bus, BusReferencer::new);
			bus.getSetClks().forEach(clk -> clk.resolveFrequency(refAnalyzer));
			bus.getSetParameters().stream().filter(Bus.SetParameter::hasCalculatedValue).forEach(clk -> clk.resolveValue(refAnalyzer));
		} catch (Exception e) {
			logger.error("Failed to determine bus dependencies {}", bus.getName(), e);
			throw new IllegalStateException("Could not resolve bus deps", e);
		}
	}

	/**
	 * Crawl all ports and parameters. Find dependencies and add them to the referenced parameters as "usages"
	 */
	public void analyzeParamRefUsage() {
		getPortGroups().stream().flatMap(pg -> pg.getPorts().stream()).filter(IPort::hasCalculatedWidth).forEach(this::findPortUsages);
		getPortGroups().stream().flatMap(pg -> pg.getPorts().stream()).filter(p -> p instanceof ClkPort).map(p -> ((ClkPort) p)).forEach(this::findClkUsages);
		getParameterGroups().stream().flatMap(pg -> pg.getParameters().stream()).filter(IParameter::hasCalculatedValue).forEach(this::findParamUsages);
		getBuses().forEach(this::findBusUsages);
	}

	@Override
	public String toString() {
		return getClass().getSimpleName() + "{"
			+ "name=" + getHeader().getName()
			+ "}";
	}

}
