/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.files.targets;

import de.tu_darmstadt.rs.spartanmc.devxml.DevXmlNamespace;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.INamed;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorGroupDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.base.XmlLoadedFile;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlEnumValue;
import javax.xml.bind.annotation.XmlID;
import javax.xml.bind.annotation.XmlIDREF;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * Created by ramonw on 19.11.15.
 */
@XmlRootElement(name = "target", namespace = DevXmlNamespace.NAMESPACE)
public class TargetDescription extends XmlLoadedFile implements INamed {

	private static final Logger logger = LoggerFactory.getLogger(TargetDescription.class);

	@XmlElement
	private String description;

	@XmlAttribute(required = true)
	private String name;

	@XmlElement(name = "device", required = true)
	private List<Device> devices;

	@XmlElement(name = "configurationHint")
	private List<ConfigurationHint> configurationHints;

	@XmlElementWrapper(name = "iopins")
	@XmlElement(name = "iogroup", required = true)
	private List<IoPinGroup> ioPinGroups;

	@XmlElementWrapper(name = "iostdGroups")
	@XmlElement(name = "iostdGroup")
	private List<IoStdGroup> ioStdGroups;

	private Map<String, IoPin> defaultPins;

	private static boolean verifyIoStdGroup(DeviceDescription device, IoStdGroup ioStdGroup, Set<IoStdGroup> verifiedIoStdGroups) {
		boolean verified = ioStdGroup.iostds.stream().allMatch(i -> device.getIostandards().contains(i));
		if (verified) {
			verifiedIoStdGroups.add(ioStdGroup);
		} else {
			logger.error("IoStdGroup '{}' could not be verified against Device '{}'", ioStdGroup.name, device.getName());
		}
		return verified;
	}

	@Override
	public String getName() {
		return name;
	}

	public String getDescription() {
		return description;
	}

	public List<Device> getDevices() {
		return devices;
	}

	public List<IoPinGroup> getIoPinGroups() {
		return ioPinGroups;
	}

	private void afterUnmarshal(Unmarshaller u, Object parent) {
		defaultPins = ioPinGroups.stream().flatMap(g -> g.getPorts().stream())
				.filter(p -> p.getDefaultFor() != null)
				.collect(Collectors.toMap(IoPin::getDefaultFor, Function.identity()));
	}


	public void verifyIoStds(Map<String, DeviceDescription> devices) {
		String defaultDeviceName = this.devices.get(0).name;
		Set<IoStdGroup> verifiedIoStdGroups = new HashSet<>();
		boolean verified = ioPinGroups.stream().allMatch(g -> {
			String deviceName = (g.device != null)? g.device.getName() : defaultDeviceName;
			DeviceDescription device = devices.get(deviceName);

			if (device == null) {
				logger.error("Could not find Device '{}' required by iogroup '{}'", deviceName, g.name);
				return false;
			}

			return g.pins.stream().allMatch(p -> {
				boolean success = (p.iostd == null || device.getIostandards().contains(p.iostd));
				if (!success) {
					logger.error("IoStd '{}' on pin '{}' is not supported on device '{}'", p.iostd, p.name, deviceName);
				}
				IoStdGroup ioStdGroup = p.getIoStdGroup();
				return success && (ioStdGroup == null || verifiedIoStdGroups.contains(ioStdGroup) || verifyIoStdGroup(device, ioStdGroup, verifiedIoStdGroups));
			});
		});

		if (!verified) {
			throw new IllegalStateException("Unable to verify used iostds on Target '" + name + "'. See log.");
		}
	}

	public void verifyBuses(Map<String, BusDescription> buses) {
		boolean success = ioPinGroups.stream().allMatch(g -> {
			if (g.implementsBus == null) {
				return true;
			} else {
				BusDescription bus = buses.get(g.implementsBus);
				if (bus == null) {
					logger.error("Could not find bus '{}', required by iogroup '{}'", g.implementsBus, g.name);
					return false;
				}
				if (bus.getBusType() != BusDescription.BusType.PeerToPeer) {
					logger.error("Bus '{}' is not PeerToPeer, requires by iogroup '{}'", g.implementsBus, g.name);
					return false;
				}
				Map<String, BusPort> busPorts = bus.getMasterPorts().getPorts().stream().collect(Collectors.toMap(BusPort::getName, Function.identity()));
				boolean verified = g.pins.stream().allMatch(p -> {
					try {
						boolean matches = busPorts.get(p.getName()).getConstantWidth() == 1;
						if (!matches) {
							logger.error("BusPort '{}' of Bus '{}' does not have the required constant width of 1", p.name, g.implementsBus);
						}
						return matches;
					} catch (Exception e) {
						logger.error("Pin '{}' does is not present on bus '{}'", p.name, g.implementsBus, e);
						return false;
					}
				});

				if (!verified) {
					logger.error("Failed to verify compliance with bus '{}' for iogroup '{}'", g.implementsBus, g.name);
				}

				return verified;
			}
		});
		if (!success) {
			throw new IllegalStateException("Unable to verify implementedBuses on Target '" + name + "'. See log");
		}
	}

	public Optional<IoPin> findPinDescription(String pinName) {
		return getIoPinGroups().stream().filter(pg -> pinName.startsWith(pg.getName())).flatMap((group) -> group.getPorts().stream()).filter((port) -> pinName.endsWith(port.getName())).findAny();//TODO "optimize"
	}

	public Optional<String> getConfigurationHint(String name) {
		return configurationHints.stream().filter(ch -> ch.name.equals(name)).findAny().map(ch -> ch.value);
	}

	public Optional<IoPin> getDefaultPin(String function) {
		return Optional.ofNullable(defaultPins.get(function));
	}

	public static class Device {

		@XmlAttribute
		private String name;

		@XmlAttribute
		private String jtagid;

		@XmlID
		@XmlAttribute
		private String id;

		public String getName() {
			return name;
		}

		public String getJtagid() {
			return jtagid;
		}

		public String getId() {
			return id;
		}
	}


	public static class IoStdGroup {

		@XmlID
		@XmlAttribute
		private String name;

		@XmlElement(name = "iostd")
		private List<String> iostds;

		public String getName() {
			return name;
		}

		public List<String> getIostds() {
			return iostds;
		}
	}


	public static class IoPinGroup implements IConnectorGroupDescription<IoPin> {

		@XmlElement(name = "pin")
		private List<IoPin> pins;

		@XmlAttribute
		private String name;

		@XmlAttribute
		private ConnectorDirection type = ConnectorDirection.Configurable;

		@XmlAttribute
		private ConnectorDirection defaultDirection = ConnectorDirection.Input;

		@XmlAttribute
		private String iostd;

		@XmlIDREF
		@XmlAttribute(name = "iostdGroup")
		private IoStdGroup ioStdGroup;

		@XmlAttribute
		private String implementsBus;

		@XmlIDREF
		@XmlAttribute
		private Device device;

		public IoStdGroup getIoStdGroup() {
			return ioStdGroup;
		}

		@Override
		public List<IoPin> getPorts() {
			return pins;
		}

		@Override
		public String getName() {
			return name;
		}

		@Override
		public ConnectorDirection getDirection() {
			return type.getReverse();
		}

		public ConnectorDirection getTargetDirection() {
			return type;
		}

		public ConnectorDirection getDefaultDirection() {
			return defaultDirection;
		}

		public String getIostd() {
			return iostd;
		}

		@Override
		public String toString() {
			return "IoPinGroup{" +
					"name='" + name + '\'' +
					", direction=" + type +
					'}';
		}
	}


	@XmlEnum
	public enum ActiveType {
		@XmlEnumValue("high")High,
		@XmlEnumValue("low")Low,
		@XmlEnumValue("unknown")Unknown;

		public static ActiveType fromString(String p) {
			String lower = p.toLowerCase();
			if (lower.equals("low") || lower.equals("active_low")|| lower.equals("low_active"))
				return Low;
			if (lower.equals("high") || lower.equals("active_high")|| lower.equals("high_active"))
				return High;
			throw new UnsupportedOperationException("Unknown active type: "+p);
		}

		public ActiveType getOpposite() {
			if (this == High) {
				return Low;
			}
			if (this == Low) {
				return High;
			}
			return Unknown;
		}
	}


	public static class IoPin implements IConnectorDescription {

		@XmlAttribute
		private String name;
		@XmlAttribute
		private ConnectorDirection type = ConnectorDirection.Inherit;
		@XmlAttribute
		private ConnectorDirection defaultDirection = ConnectorDirection.Inherit;
		@XmlAttribute
		private String fpgapin;
		@XmlAttribute
		private String descr;
		@XmlAttribute
		private String iostd;
		@XmlAttribute
		private long freq = 0;
		@XmlAttribute(required = false)
		private ActiveType activeType = ActiveType.Unknown;
		@XmlAttribute
		private String defaultFor;

		@XmlIDREF
		@XmlAttribute(name = "iostdGroup")
		private IoStdGroup ioStdGroup;
		private IoPinGroup ioStdGroupParent;

		@Override
		public String getName() {
			return name;
		}

		@Override
		public ConnectorDirection getDirection() {
			return type.getReverse();
		}

		@Override
		public boolean isDirectionLocked() {
			return type != ConnectorDirection.Configurable;
		}

		@Override
		public Type getType() {
			return Type.Pin;
		}

		public long getFreq() { //TODO model better
			return freq;
		}

		public ConnectorDirection getTargetDirection() {
			return type;
		}

		public ConnectorDirection getDefaultDirection() {
			return defaultDirection;
		}

		public String getFpgaPinName() {
			return fpgapin;
		}

		public String getDescr() {
			return descr;
		}

		public String getIostd() {
			return iostd;
		}

		public String getDefaultFor() {
			return defaultFor;
		}

		public IoStdGroup getIoStdGroup() {
			return (ioStdGroupParent != null)? ioStdGroupParent.ioStdGroup : ioStdGroup;
		}

		public ActiveType getActiveType() {
			return activeType;
		}

		private void afterUnmarshal(Unmarshaller u, Object parent) {
			IoPinGroup group = (IoPinGroup) parent;

			if (type == ConnectorDirection.Inherit) {
				type = group.getTargetDirection();
			}
			if (defaultDirection == ConnectorDirection.Inherit) {
				defaultDirection = group.getDefaultDirection();
			}
			if (iostd == null) {
				iostd = group.getIostd();
			}
			if (ioStdGroup == null) { //KeyRefs are resolved after afterUnmarshal!
				ioStdGroupParent = group;
			}
			if (group.name != null) {
				name = group.getName() + '_' + name;
			}
		}

		@Override
		public String toString() {
			return "Pin{" +
					"name='" + name + '\'' +
					", direction=" + type +
					", iostd='" + iostd + '\'' +
					", freq=" + freq +
					'}';
		}
	}

	@Override
	public String toString() {
		return name;
	}

	public static class ConfigurationHint {

		@XmlAttribute
		private String name;

		@XmlAttribute
		private String value;

	}
}
