/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.devxml.mappedBuses;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.CombineType;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * Basically a ReadOnly PortHandler IConnectorInstance for a {@link BusPort} that is implemented on a Module.
 * MappedBusPorts are the ports that the modules HDL-Code will really have to implement (after mapping to local names and merging)
 * <p>
 * Created by Ramon on 06.04.2016.
 */
public abstract class BaseMappedBusPort implements IMappedBusPort {

	public static final Comparator<BaseMappedBusPort> NAME_COMPARATOR = Comparator.comparing(BaseMappedBusPort::getPortName);

	protected final String localName;
	protected final CombineType combineType;
	protected final IModuleDescription moduleDescription;
	protected final Bus busDescription;


	protected BaseMappedBusPort(String localName, CombineType combineType, Bus bus, IModuleDescription moduleDescription) {
		this.localName = localName;
		this.combineType = combineType;
		this.moduleDescription = moduleDescription;
		this.busDescription = bus;
	}

	public static List<BaseMappedBusPort> mapAllBuses(IModuleDescription module, Library library) {
		Map<String, List<SingleMappedBusPort>> mappedPortsByName = module.getBuses().stream().flatMap(busDesc -> SingleMappedBusPort.mapBus(module, busDesc, library)).collect(Collectors.groupingBy(BaseMappedBusPort::getPortName));
		return mappedPortsByName.values().stream().map(BaseMappedBusPort::mergeIfMultiple).sorted(NAME_COMPARATOR).collect(Collectors.toList());
	}

	public static BaseMappedBusPort mergeIfMultiple(List<SingleMappedBusPort> mappedPorts) {
		return mappedPorts.size() > 1? MergedBusPort.merge(mappedPorts) : mappedPorts.get(0);
	}

	@Override
	public IModuleDescription getModuleDescription() {
		return moduleDescription;
	}

	/**
	 * The local name of this BusPort in this Module
	 *
	 * @return local name
	 */
	@Override
	public String getPortName() {
		return localName;
	}

	@Override
	public int getBusConnectionCount(IValueResolver resolver) {
		return busDescription.getBusConnectionCount(resolver);
	}

	@Override
	public CombineType getCombineType() {
		return combineType;
	}

	public abstract <T> T apply(IConverter<T> visitor);

	public interface IConverter<T> {

		T convert(SingleMappedBusPort mappedBusPort);

		T convert(MergedBusPort mergedBusPort, List<SingleMappedBusPort> ports);
	}
}


