/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.devxml.mappedBuses;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.CombineType;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.IPortMapping;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;

import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by Ramon on 06.04.2016.
 */
public class SingleMappedBusPort extends BaseMappedBusPort {

	private final BusPort port;
	private final Bus bus;
	private final boolean allowMerging;

	protected SingleMappedBusPort(String localName, BusPort port, boolean allowMerging, CombineType combineType, Bus bus, IModuleDescription module) {
		super(localName, combineType, bus, module);
		this.port = port;
		this.allowMerging = allowMerging;
		this.bus = bus;
	}

	public static Stream<SingleMappedBusPort> mapBus(IModuleDescription module, Bus bus, Library library) {
		Map<String, IPortMapping> mappings = bus.getPortMappings().stream().collect(Collectors.toMap(IPortMapping::getBusName, Function.identity()));
		BusDescription.BusPortGroup portGroup = library.getBusDescription(bus.getBusDescriptionName()).getMatchingPorts(bus);
		return portGroup.getPorts().stream().map(p -> create(bus, mappings.get(p.getName()), p, library, module));
	}

	public static SingleMappedBusPort create(Bus bus, IPortMapping mapping, BusPort port, Library library, IModuleDescription module) {
		boolean allowMerging;
		CombineType combineType;
		String name;

		if (mapping != null) {
			name = mapping.getName();
			allowMerging = mapping.isAllowMerging();
			combineType = mapping.getCombineUsing();
		} else {
			name = port.getName();
			allowMerging = false;
			combineType = port.getDefaultCombineType();
		}

		if (combineType == CombineType.Auto) {
			combineType = getDefaultCombineType(port, bus, library);
		}

		return new SingleMappedBusPort(name, port, allowMerging, combineType, bus, module);
	}

	private static CombineType getDefaultCombineType(BusPort port, Bus bus, Library library) {
		return CombineType.ForceSingle;
	}

	public IConnectorDescription.Type getType() {
		return port.getType();
	}

	@Override
	public boolean isAllowMerging() {
		return allowMerging;
	}

	public ConnectorDirection getDirection() {
		return port.getDirection();
	}

	@Override
	public boolean isPerConnectionWidth() {
		return combineType == CombineType.Parallel;
	}

	@Override
	public Stream<SingleMappedBusPort> streamSingleMappedBusPorts() {
		return Stream.of(this);
	}

	public boolean hasCalculatedBaseWidth() {
		return port.hasCalculatedWidth();
	}

	public int getStaticBaseWidth() {
		return port.getConstantWidth();
	}

	public CalculatedValueResult<Integer> resolveBaseWidth(IValueResolver resolver) {
		return port.resolveWidth(resolver).map(Math::toIntExact);
	}

	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder();
		toStringIntro(sb);
		toStringDescProp(sb);
		return sb.toString();
	}

	public void toStringDescProp(StringBuilder sb) {
		sb.append("(Bus: ").append(bus.getName()).append(", Type: ").append(bus.getBusDescriptionName()).append("::").append(bus.getRole());
		sb.append("), combineUsing: ").append(combineType);
		sb.append(", ").append(port);
	}

	public StringBuilder toStringIntro(StringBuilder sb) {
		return sb.append("Single{").append(getCanonicalName());
	}

	public BusPort getPort() {
		return port;
	}

	@Override
	public <T> T apply(IConverter<T> converter) {
		return converter.convert(this);
	}

	public Bus getBus() {
		return bus;
	}
}
