/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


package de.tu_darmstadt.rs.spartanmc.devxml.schemas;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.ls.LSInput;
import org.w3c.dom.ls.LSResourceResolver;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import javax.xml.XMLConstants;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

public final class ResourceSchemaResolver implements LSResourceResolver {

	private static final Logger logger = LoggerFactory.getLogger(ResourceSchemaResolver.class);
	public final Schema schema;
	private final String base;

	private ResourceSchemaResolver(String sch) {
		SchemaFactory sf = SchemaFactory.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);
		sf.setResourceResolver(this);
		this.base = ResourceSchemaResolver.class.getResource("").toString();


		try (InputStream schemaStream = ResourceSchemaResolver.class.getResourceAsStream(sch)) {
			if (schemaStream == null) {
				throw new RuntimeException(
						String.format(
								"no such schema document: %s",
								sch
						)
				);
			}
			schema = sf.newSchema(new StreamSource(schemaStream, sch));
		} catch (SAXParseException e) {
			logger.error("Schema invalid: {}", sch, e);
			throw new RuntimeException("schema document error");
		} catch (SAXException e) {
			throw new RuntimeException(e);
		} catch (IOException e) {
			logger.error("Failed to close schemaStream");
			throw new RuntimeException(e);
		}
	}

	public static Schema getSchema(String schema) {
		ResourceSchemaResolver resourceSchemaResolver = new ResourceSchemaResolver(schema);
		return resourceSchemaResolver.schema;
	}

	@Override
	public LSInput resolveResource(String t, String ns, String pubId, String sysId, String base) {
		//System.out.printf("Sysid: %s base %s%n", sysId, base);
		return new ResourceInput(sysId, base);
	}


	private class ResourceInput implements LSInput {

		private final InputStream fIs;
		private final String systemId;

		public ResourceInput(String res, String base) {
			this.systemId = ResourceSchemaResolver.class.getResource(res).toString();
			fIs = ResourceSchemaResolver.class.getResourceAsStream(res);
			if (fIs == null) {
				throw new RuntimeException(
						String.format(
								"no such resource: %s",
								res
						)
				);
			}
		}

		@Override
		public Reader getCharacterStream() {
			return null;
		}

		@Override
		public void setCharacterStream(Reader r) {
			throw new UnsupportedOperationException();
		}

		@Override
		public InputStream getByteStream() {
			return fIs;
		}

		@Override
		public void setByteStream(InputStream is) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getStringData() {
			return null;
		}

		@Override
		public void setStringData(String s) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getSystemId() {
			return systemId;
		}

		@Override
		public void setSystemId(String s) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getPublicId() {
			return null;
		}

		@Override
		public void setPublicId(String s) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getBaseURI() {
			return base;
		}

		@Override
		public void setBaseURI(String s) {
			throw new UnsupportedOperationException();
		}

		@Override
		public String getEncoding() {
			return null;
		}

		@Override
		public void setEncoding(String s) {
			throw new UnsupportedOperationException();
		}

		@Override
		public boolean getCertifiedText() {
			return false;
		}

		@Override
		public void setCertifiedText(boolean b) {
			throw new UnsupportedOperationException();
		}
	}

}
