/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.sources;

import java.nio.file.Path;
import java.util.EnumSet;
import java.util.Set;

/**
 * Created by Ramon on 17.08.2016.
 */
public final class SourceRef implements Comparable<SourceRef> {

	public enum FileType {
		SYNTHESIS,
		FUNCTIONAL_SIM,
		TIMESIM;

		public static Set<FileType> ALL_SIMULATION = EnumSet.of(FUNCTIONAL_SIM, TIMESIM);
		public static Set<FileType> REGULAR = EnumSet.of(FUNCTIONAL_SIM, SYNTHESIS, TIMESIM);

		public static Set<FileType> fromBoolean(boolean isSimulation) {
			if (isSimulation) {
				return ALL_SIMULATION;
			} else {
				return REGULAR;
			}
		}
	}


	private final Path path;
	private final String library;
	private final Set<FileType> fileAssociation;

	public SourceRef(Path path, FileType fileAssociation) {
		this(path, EnumSet.of(fileAssociation));
	}
	public SourceRef(Path path, String library, FileType fileAssociation) {
		this(path, library, EnumSet.of(fileAssociation));
	}

	public SourceRef(Path path, Set<FileType> fileAssociation) {
		this(path, null, fileAssociation);
	}
	public SourceRef(Path path, String library, Set<FileType> fileAssociation) {
		this.path = path.normalize();
		this.library = library;
		this.fileAssociation = fileAssociation;
	}

	public SourceRef(Path path) {
		this(path, null, FileType.REGULAR);
	}

	public Path getPath() {
		return path;
	}

	public String getLibrary() {
		return library;
	}

	public Set<FileType> getFileAssociation() {
		return fileAssociation;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (!(o instanceof SourceRef)) {
			return false;
		}

		SourceRef sourceRef = (SourceRef) o;

		if (!path.equals(sourceRef.path)) {
			return false;
		}
		if (library != null? !library.equals(sourceRef.library) : sourceRef.library != null) {
			return false;
		}
		return fileAssociation == sourceRef.fileAssociation;
	}

	@Override
	public int hashCode() {
		int result = path.hashCode();
		result = 31 * result + (library != null? library.hashCode() : 0);
		result = 31 * result + fileAssociation.hashCode();
		return result;
	}

	@Override
	public String toString() {
		return "SourceRef{" +
				"path=" + path +
				", library='" + library + '\'' +
				", fileAssociation=" + fileAssociation +
				'}';
	}

	@Override
	public int compareTo(SourceRef o) {
		return this.path.compareTo(o.path);
	}
}
