/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.devxml.tests

import de.tu_darmstadt.rs.spartanmc.devxml.Library
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription
import de.tu_darmstadt.rs.spartanmc.devxml.exceptions.MissingModuleDescriptionException
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription
import spock.lang.Shared
import spock.lang.Specification

import java.nio.file.Files
import java.nio.file.Paths
/**
 * Created by Ramon on 29.05.2016.
 */
class ModuleSpec extends Specification {

	@Shared Library lib = new Library.Builder().addDefaultSearchpaths(Paths.get("../../..")).setAnalyze(true).build()

	def "non-existent module does not exist"() {
		when:
		def mod = lib.getCommonDescription("foo")
		then:
		thrown(MissingModuleDescriptionException)
	}

	def "spartanmc core exists"() {
		when:
		def core = lib.getSubsystemDescription("spartanmc")

		then:
		core != null
		core.type == IModuleDescription.ModuleType.Subsystem
		Files.isSameFile core.sourceTree, Paths.get("../../../spartanmc/hardware")
		Files.isSameFile core.sourceFile, Paths.get("../../../spartanmc/hardware/core/spartanmc.module.xml")
	}

	def "spartanmc local memory exists"() {
		when:
		def mem = lib.getMemoryCommonDescription("spartanmc_mem_local")

		then:
		mem != null
		mem.type == IModuleDescription.ModuleType.Memory
		Files.isSameFile mem.sourceTree, Paths.get("../../../spartanmc/hardware")
		Files.isSameFile mem.sourceFile, Paths.get("../../../spartanmc/hardware/core_memory/spartanmc_mem_local.module.xml")
	}

	def "core resolves static clock"() {
		def core = lib.getSubsystemDescription("spartanmc")
		def bF = CalculatedValueResult.of(1000000L)
		def bP = 9L
		def resolver = Mock(IValueResolver)

		def setClk = core.buses[0].setClks[0]

		when:
		def freq = setClk.resolveFrequency(resolver)
		then:
		with(resolver){
			1 * resolveClkPort(!null) >> bF
			0 * _
		}
		freq ==  bF
	}

	def "uart resolves clock"() {
		def uart = lib.getPeripheralDescription("uart_light")
		def bF = CalculatedValueResult.of(25000000L)
		def resolver = Mock(IValueResolver)

		def param = uart.getParameterGroups().collectMany({it.getParameters()}).find {it.getName() == "CLOCK_FREQUENCY"}

		when:
		def freq = param.resolveValue(resolver)
		then:
		with(resolver){
			1 * resolveClkPort(!null) >> bF
			0 * _
		}
		freq ==  bF
	}

	def "port_in resolves width"() {
		def peri = lib.getPeripheralDescription("port_in")
		def resolver = Mock(IValueResolver)
		def bP = CalculatedValueResult.of(12L)

		when:
		def port = peri.portGroups[0].ports[0]
		then:
		port.name == "pin"
		port.directionLocked
		port.direction == ConnectorDirection.Input
		port.type == IConnectorDescription.Type.Port
		port.hasCalculatedWidth()

		when:
		def width = port.resolveWidth(resolver)
		then:
		with(resolver) {
			1 * resolveParam(_) >> bP
			0 * _
		}
		width == bP

	}
}
