/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.scripting

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.SubsystemModuleDescription
import de.tu_darmstadt.rs.spartanmc.jconfig.api.*
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument.ITargetConfiguration
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IModuleContainer
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.*
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil
import groovy.transform.stc.ClosureParams
import groovy.transform.stc.SimpleType

import java.util.stream.Collectors
import java.util.stream.Stream
/**
 * Groovy Extensions for Document and Modules to facilitate easiest access to params, buses, ports, pins and connection
 *
 * Allows stuff like
 * <pre>
 * 		module.ports["name"]
 * </pre>
 * and
 * <pre>
 * 		module.parameters["name"] as long
 * </pre>
 */
public class ModuleExtension {

	public static class ToolchainParams extends HashMap<String, IToolchainAttributeHandler> {
		private final ITargetConfiguration target

		public ToolchainParams(ITargetConfiguration target) {
			this.target = target;
			target.getToolchain().params.values().each {
				put(it.name, target.getToolchainParamHandler(it))
			}
		}

		public IToolchainAttributeHandler getAt(final String paramName) {
			def param = getToolchainParameter(target, paramName)
			if (param == null) {
				throw new NoSuchElementException("No Toolchain Parameter '$param'")
			}
			return param
		}

		public void putAt(final String paramName, final String value) {
			getAt(paramName).storedValue = value
		}

		public String tree(Map<String, Object> config = [:], StringBuilder sb = null, String indent = '') {
			def sbInt = sb != null ? sb : new StringBuilder()
			def ind = "${indent}|    "
			sbInt.append "Toolchain Parameters\n"
			def last = size() - 1
			this.eachWithIndex{ String k, IToolchainAttributeHandler p, int i ->
				sbInt.append indent
				if (i != last) {
					sbInt.append '+--- '
				} else {
					sbInt.append '\\--- '
				}
				sbInt.append "$p.prettyString\n"
			}
			if (sb == null)
				return sbInt.toString()
		}

		public String toString() {
			return tree()
		}
	}

	public static class Parameters extends HashMap<String, IParameterHandler> {

		private IModuleConfiguration module;

		public Parameters(IModuleConfiguration module) {
			this.module = module;
			putAll(DocumentUtil.streamAllParameters(module).collect(Collectors.toMap({
				it.name
			}, { it })));
		}

		public <P extends IParameterHandler> P getAt(final String paramName) {
			P parameter = getParameter(module, paramName);
			if (parameter == null) {
				throw new NoSuchElementException("No parameter " + paramName);
			}
			return parameter;
		}

		public void putAt(final String paramName, long value) {
			((IIntegerParameterHandler) getAt(paramName)).setValue(value);
		}

		public void putAt(final String paramName, String value) {
			getAt(paramName).setStoredValue(value);
		}

		public void putAt(final String paramName, double value) {
			((IFloatParameterHandler) getAt(paramName)).setValue(value);
		}

		public void putAt(final String paramName, Boolean value) {
			((IBooleanParameterHandler) getAt(paramName)).setValue(value);
		}

//		public void setProperty(final String paramName, Object value) {
//			//TODO wont work. Overriden by groovy's builtin stuff?
//			if (value == null) {
//				throw new UnsupportedOperationException("Null not allowed");
//			} else if (value.getClass() == Long.class || value.getClass() == Integer.class) {
//				setProperty(paramName, (long) value);
//			} else if (value.getClass() == Double.class || value.getClass() == Float.class) {
//				setProperty(paramName, (double) value);
//			} else if (value.getClass() == String.class) {
//				setProperty(paramName, (String) value);
//			} else if (value.getClass() == Boolean.class) {
//				setProperty(paramName, (boolean) value);
//			} else {
//				throw new UnsupportedOperationException("Incompatible type");
//			}
//		}

		public String tree(Map<String, Object> config = [:], StringBuilder sb = null, String indent = '') {
			def sbInt = sb != null ? sb : new StringBuilder()
			def ind = "${indent}|    "
			sbInt.append "Parameters of $module.name\n"
			def last = size() - 1
			this.eachWithIndex{ k, IParameterHandler p, i ->
				sbInt.append indent
				if (i != last) {
					sbInt.append '+--- '
				} else {
					sbInt.append '\\--- '
				}
				sbInt.append "$p.prettyString\n"
			}
			if (sb == null)
				return sbInt.toString()
		}

		public String toString() {
			return tree();
		}
	}


	public static class Ports extends ConnectorsBase<IPortHandler<? extends IPort>> {

		public Ports(IConfigurationElement module) {
			super(module)
			DocumentUtil.streamConnectorHandlers(module).forEach({ put(it.portName, it) })
		}
	}

	public static class Modules extends LinkedHashMap<String, IModuleConfiguration<?>> {

		private IConfigurationElement parent

		public Modules(IConfigurationElement parent) {
			this.parent = parent
			Stream<IModuleConfiguration<?>> stream;
			if (parent instanceof ISubsystemConfiguration) {
				stream = ((ISubsystemConfiguration<?>) parent).streamAllModules();
			} else {
				stream = ((IConfigurationDocument) parent).streamAllModules();
			}
			stream.forEach({ put(it.name, it) })
		}

		public String tree(Map<String, Object> config = [:], StringBuilder sb = null, String indent = '') {
			def sbInt = sb != null ? sb : new StringBuilder()
			def ind2 = "${indent}|    "
			sbInt.append "$parent.prettyString\n"
			def last = size() - 1
			eachWithIndex { k, m, i ->
				sbInt.append indent
				if (i != last) {
					sbInt.append '+--- '
					m.tree(config, sbInt, ind2)
				} else {
					sbInt.append '\\--- '
					m.tree(config, sbInt, "${indent}     ")
				}
			}
			if (sb == null)
				return sbInt.toString()
		}

		public String toString() {
			return tree();
		}
	}


	public static class Pins extends ConnectorsBase<IPinHandler> {

		public Pins(IConfigurationDocument document) {
			super(document)
			DocumentUtil.streamConnectorHandlers(document).forEach({ put(it.portName, it) });
		}
	}

	public static class Buses extends ConnectorsBase<IBusHandler> {

		public Buses(IModuleConfiguration module) {
			super(module)
			DocumentUtil.streamBusHandlers(module).forEach({ put(it.portName, it) })
		}
	}

	private static abstract class ConnectorsBase<T extends IConnectorHandler> extends LinkedHashMap<String, T> {
		protected final IConfigurationElement element;

		protected ConnectorsBase(IConfigurationElement element) {
			this.element = element
		}

		public String tree(Map<String, Object> config = [:], StringBuilder sb = null, String indent = '') {
			def sbInt = sb != null ? sb : new StringBuilder()
			def ind = "${indent}|    "
			def type = this.getClass().simpleName
			sbInt.append "$type of $element.name\n"
			def last = size() - 1
			this.eachWithIndex{ k, IConnectorHandler c, i ->
				sbInt.append indent
				if (i != last) {
					sbInt.append '+--- '
					c.tree(config, sbInt, ind)
				} else {
					sbInt.append '\\--- '
					c.tree(config, sbInt, "$indent     ")
				}

			}
			if (sb == null)
				return sbInt.toString()
		}

		public String toString() {
			return tree();
		}
	}

	public static Parameters getParameters(final IModuleConfiguration self) {
		return new Parameters(self);
	}

	public static Ports getPorts(final IConfigurationElement self) {
		return new Ports(self);
	}

	public static Pins getPorts(final IConfigurationDocument self) {
		return new Pins(self);
	}

	public static Buses getBuses(final IModuleConfiguration self) {
		return new Buses(self);
	}

	public static Pins getPins(final IConfigurationDocument self) {
		return new Pins(self);
	}

	public static IToolchainAttributeHandler getToolchainParameter(final ITargetConfiguration self, final String paramName) {
		def paramDesc = self.getToolchain().params[paramName]
		if (paramDesc != null) {
			return self.getToolchainParamHandler(paramDesc)
		} else {
			return null
		}
	}

	public static ToolchainParams getToolchainParameters(final ITargetConfiguration self) {
		return new ToolchainParams(self)
	}

	public static IToolchainAttributeHandler getToolchainParameter(final IConfigurationDocument self, final String paramName) {
		return getToolchainParameter(self.target, paramName)
	}

	public static ToolchainParams getToolchainParameters(final IConfigurationDocument self) {
		return getToolchainParameters(self.target)
	}

	public static <P extends IParameterHandler> P getParameter(
		final IModuleConfiguration self, final String paramName) {
		return (P) DocumentUtil.getParameterHandler(self, paramName, IParameterHandler.class);
	}

	public static IPortHandler<? extends IPort> getPort(final IModuleConfiguration self, final String portName) {
		return DocumentUtil.getPortHandler(self, portName);
	}

	public static IPinHandler getPin(final IConfigurationDocument self, final String pinName) {
		return DocumentUtil.getPinHandler(self, pinName);
	}

	public static IBusHandler getBus(final IModuleConfiguration self, final String busName) {
		return DocumentUtil.getBusHandler(self, busName);
	}

	public static ISubsystemConfiguration addSubsystem(
		final IConfigurationDocument self, final String descriptionName) {
		return self.addSubsystem(self.getLibrary().getSubsystemDescription(descriptionName));
	}

	public static IPeripheralModuleConfiguration<?> addPeripheral(
		final ISubsystemConfiguration<?> self, final String descriptionName) {
		return self.addPeripheral(self.getLibrary().getPeripheralDescription(descriptionName));
	}

	public static IModuleConfiguration<?> addCommonModule(
		final ISubsystemConfiguration<?> self, final String descriptionName) {
		return self.addCommonModule(self.getLibrary().getCommonDescription(descriptionName));
	}

	public static IModuleConfiguration addCommonModule(
		final IConfigurationDocument self, final String descriptionName) {
		return self.addModule(self.getLibrary().getCommonDescription(descriptionName));
	}

	public static IConnection<? extends IPortHandler<?>> connect(
		final IConfigurationDocument self, IPortHandler<?> a, IPortHandler<?> b) {
		IConnection<? extends IPortHandler<?>> con = self.createPortConnection(a, b);
		con.connect();
		return con;
	}

	public static final int AUTO = -1;

	public static IConnection<? extends IPortHandler<?>> connect(
		final IConfigurationDocument self, IPortHandler<?> a, int aStart, IPortHandler<?> b, int bStart=AUTO, int width=AUTO) {
		IConnection<? extends IPortHandler<?>> con = self.createPartialConnection(a, aStart, b, bStart, width);
		con.connect();
		return con;
	}

	public static IConnection<? extends IBusHandler> connect(
		final IConfigurationDocument self, IBusHandler a, IBusHandler b) {
		IConnection<? extends IBusHandler> con = self.createBusConnection(a, b);
		con.connect();
		return con;
	}

	public static boolean connectIfConnectable( final IConfigurationDocument self, IPortHandler<?> a, IPortHandler<?> b) {
		if (a.isConnectibleTo(b) && a.isAdditionalConnectionAllowed() && b.isAdditionalConnectionAllowed()) {
			self.connect(a, b);
			return true;
		}
		return false;
	}


	public static boolean connectIfConnectable( final IConfigurationDocument self, IBusHandler a, IBusHandler b) {
		if (a.isConnectibleTo(b) && a.isAdditionalConnectionAllowed() && b.isAdditionalConnectionAllowed()) {
			self.connect(a, b);
			return true;
		}
		return false;
	}

	public
	static List<IConnection<? extends IPortHandler<?>>> connectMultiple(IConfigurationDocument self, IPortHandler<?> a, int aStart = 0, int width = AUTO, IConfigurationElement other, List<String> portNames) {
		if (width < 0) {
			width = a.connectorWidth
		}
		def res = []
		for (int i = aStart; i < width; i++) {
			def otherPort = other.ports[portNames[i]]
			if (!otherPort)
				throw new NoSuchElementException("No port ${portNames[i]} on $other")
			res += connect(self, a, i, otherPort)
		}
		return res
	}

	public
	static List<IConnection<? extends IPortHandler<?>>> connectMultiple(IConfigurationDocument self, IPortHandler<?> a, int aStart = 0, int width = AUTO, IConfigurationElement other,
																		@ClosureParams(value = SimpleType, options = ['int']) Closure<String> portNames) {
		if (width < 0) {
			width = a.connectorWidth
		}
		def res = []
		for (int i = aStart; i < width; i++) {
			def otherPort = other.ports[portNames(i)]
			res += connect(self, a, i, otherPort)
		}
		return res
	}


	public static Parameters connectToDefault(final IPortHandler<?> self, String defaultPort) {
		IConfigurationElement parent = self.getSourceElement();

		//Get the document. Ugly stuff :(
		while (!(parent instanceof IConfigurationDocument)) {
			parent = parent.parent
		}
		IConfigurationDocument document = parent

		document.getDefaultPin(defaultPort).ifPresent { pin ->
			connectIfConnectable(document, self, pin)
		}
	}

	public static Object asType(final IParameterHandler handler, Class target) {
		if (target == Integer.class || target == Long.class) {
			return ((IIntegerParameterHandler) handler).getIntegerValue().get();
		} else if (target == Float.class || target == Double.class) {
			return ((IFloatParameterHandler) handler).getFloatValue();
		} else if (target == Boolean.class) {
			return ((IBooleanParameterHandler) handler).isSet();
		} else if (target == String.class) {
			return handler.getValue();
		} else {
			throw new ClassCastException("ParameterHandler cannot be coerced into " + target.getSimpleName());
		}
	}

	public
	static IModuleConfiguration<? extends IConfigurationDocument> add(IConfigurationDocument self, String moduleDescriptionName, String moduleName = null) {
		IModuleDescription<?> moduleDescription = self.getLibrary().getModuleDescription(moduleDescriptionName);
		if (moduleDescription==null) {
			throw new RuntimeException("No module description named \""+moduleDescriptionName+"\" found.");
		}
		if (moduleDescription instanceof SubsystemModuleDescription) {
			return self.addSubsystem((SubsystemModuleDescription) moduleDescription, moduleName);
		} else {
			return self.addModule(moduleDescription, moduleName);
		}
	}

	public
	static IModuleConfiguration<? extends IConfigurationDocument> plus(IConfigurationDocument self, String moduleDescriptionName) {
		return add(self, moduleDescriptionName);
	}

	public
	static IModuleConfiguration<? extends IConfigurationDocument> plus(IConfigurationDocument self, Map<String, String> modDesc) {
		String desc = modDesc.get("desc")
		String name = modDesc.get("name")
		return add(self, desc, name)
	}

	public
	static IModuleConfiguration<? extends ISubsystemConfiguration<?>> add(ISubsystemConfiguration<?> self, String moduleDescriptionName, String moduleName = null) {
		IModuleDescription moduleDescription = self.getLibrary().getModuleDescription(moduleDescriptionName);
		return self.addModule(moduleDescription, moduleName);
	}

	public
	static IModuleConfiguration<? extends ISubsystemConfiguration<?>> plus(ISubsystemConfiguration<?> self, String moduleDescriptionName) {
		return add(self, moduleDescriptionName);
	}

	public
	static IModuleConfiguration<? extends ISubsystemConfiguration<?>> plus(ISubsystemConfiguration<?> self, Map<String, String> modDesc) {
		String desc = modDesc.get("desc")
		String name = modDesc.get("name")
		return add(self, desc, name)
	}

	public static IModuleConfiguration<?> getModule(ISubsystemConfiguration<?> self, String moduleName) {
		return self.streamAllModules().filter({ c -> c.getName().equals(moduleName) }).findFirst().orElseThrow({ ->
			new NoSuchElementException("No child with name " + moduleName)
		});
	}

	public static IModuleConfiguration<?> getModule(IConfigurationDocument self, String moduleName) {
		return self.streamAllModules().filter({ c -> c.getName().equals(moduleName) }).findFirst().orElseThrow({ ->
			new NoSuchElementException("No child with name " + moduleName)
		});
	}

	public static boolean remove(ISubsystemConfiguration<?> self, String moduleName) {
		IModuleConfiguration<?> mod = self.modules[moduleName];
		return self.removeModule(mod);
	}

	public static boolean remove(IConfigurationDocument self, String moduleName) {
		IModuleConfiguration<?> mod = self.modules[moduleName];
		return self.removeModule(mod);
	}

	public static boolean minus(ISubsystemConfiguration<?> self, String moduleName) {
		return remove(self, moduleName);
	}

	public static boolean minus(IConfigurationDocument self, String moduleName) {
		return remove(self, moduleName);
	}

	public static boolean remove(IConfigurationDocument self, IModuleConfiguration module) {
		return self.removeModule(module)
	}

	public static boolean remove(ISubsystemConfiguration self, IModuleConfiguration module) {
		return self.removeModule(module)
	}

	public static boolean minus(IConfigurationDocument self, IModuleConfiguration module) {
		return self.removeModule(module)
	}

	public static boolean minus(ISubsystemConfiguration self, IModuleConfiguration module) {
		return self.removeModule(module)
	}

	public static Modules getModules(IConfigurationDocument self) {
		return new Modules(self)
	}

	public
	static Modules getModules(ISubsystemConfiguration<?> self) {
		return new Modules(self)
	}

	public static IModuleConfiguration<?> find(IModuleContainer<?> self, String moduleName) {
		def res = DocumentUtil.findModulesByName(self, moduleName);
		if (res.size() > 1) {
			throw new IllegalArgumentException("Mutliple matches: ${res.collect { it.absolutePath }}")
		}
		return res[0]
	}

	public static Set<IModuleConfiguration<?>> findByName(IModuleContainer<?> self, String moduleName) {
		return DocumentUtil.findModulesByName(self, moduleName);
	}

	public static Set<IModuleConfiguration<?>> findByType(IModuleContainer<?> self, String moduleType) {
		return DocumentUtil.findModulesByType(self, moduleType);
	}

	public static void setName(IConfigurationElement<?, ?, ?> self, String name) {
		self.getNameHandler().setStoredValue(name);
	}

	public static void each(IConfigurationDocument self,
							@ClosureParams(value = SimpleType, options = ['de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration']) Closure closure) {
		self.streamModulesRecursive().forEach(closure)
	}

	public static String getPrettyString(IConfigurationElement element) {
		"$element.type : $element.name"
	}

	public static String tree(IConfigurationDocument self, Map<String, Object> config = [:], StringBuilder sb = null) {
		def sbInt = sb != null ? sb : new StringBuilder()
		self.modules.tree(config, sbInt, '')
		if (sb == null)
			return sbInt.toString()
	}

	public
	static String tree(ISubsystemConfiguration self, Map<String, Object> config = [:], StringBuilder sb = null, String indent = '') {
		def sbInt = sb != null ? sb : new StringBuilder()
		self.modules.tree(config, sbInt, indent)
		if (sb == null)
			return sbInt.toString()
	}

	public
	static String tree(IModuleConfiguration self, Map<String, Object> config = [:], StringBuilder sb = null, String indent = '') {
		def sbInt = sb != null ? sb : new StringBuilder()

		def indMore = "${indent}|    "
		def indLast = "${indent}     "
		sbInt.append "$self.prettyString\n"
		sbInt.append indent
		if (config?.all) {
			sbInt.append '+--- '
			self.parameters.tree(config, sbInt, indMore)
			sbInt.append indent
			sbInt.append '+--- '
			self.buses.tree(config, sbInt, indMore)
			sbInt.append indent
			sbInt.append '\\--- '
			self.ports.tree(config, sbInt, indLast)
		} else {
			sbInt.append '\\--- '
			self.parameters.tree(config, sbInt, indLast)
		}

		if (sb == null)
			return sbInt.toString()
	}

	public
	static String tree(IConnectorHandler self, Map<String, Object> config = [:], StringBuilder sb = null, String indent = '') {
		def sbInt = sb != null ? sb : new StringBuilder()
		def ind2 = "${indent}|    "
		sbInt.append "$self.prettyString\n"
		def connections = self.connections
		def last = connections.size() - 1
		connections.eachWithIndex { c, i ->
			sbInt.append indent
			if (i != last) {
				sbInt.append '+--- '
			} else {
				sbInt.append '\\--- '
			}
			sbInt.append "[${c.getMyStart(self)}:${c.getMyEnd(self)}] ${c.getOppositeConnector(self).absoluteConnectorName} [${c.getOppositeStart(self)}:${c.getOppositeEnd(self)}]\n"
		}
		if (sb == null)
			return sbInt.toString()
	}

	public static List<IConnection<?>> getConnections(IConnectorHandler self) {
		self.streamConnections().collect(Collectors.toList())
	}

	public static String getPrettyString(IConnectorInstance self) {
		if (self instanceof IBusHandler)
			return "$self.portName ($self.connectorDescription.busDescriptionName:$self.portName ($self.connectorWidth): $self.direction)"
		else if (self instanceof IPinHandler)
			return "$self.portName ($self.direction, $self.directionLocked, $self.ioStd, $self.drive, $self.frequency, $self.termination, $self.slewRate, $self.inverted, $self.customUserConstraints)"
		else if (self instanceof IPortHandler)
			return "$self.portName ($self.direction [$self.connectorWidth])"
		else
			return "$self.portName"
	}

	public static String getPrettyString(IParameterHandler self) {
		return "$self.name ($self.parameterDescription.type): $self.value"
	}

	public static String getPrettyString(IToolchainAttributeHandler self) {
		return "$self.name: ${self.value.get()}"
	}
}
