/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.api;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.INamed;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorGroupDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectionContainer;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;

import java.util.Comparator;
import java.util.Map;
import java.util.Set;

/**
 * Part of Configuration that has a representation in the configuration hierarchy tree,
 * and therefor has a Name, ElementType, generic Attributes, Connectors and basic events
 * to access children it is required to know the actual implementation, as the types of children differ
 *
 * @author Ramon Wirsch
 */
public interface IConfigurationElement<PD extends IConnectorDescription, PG extends IConnectorGroupDescription, PH extends IConnectorHandler<?, ?>> extends INamed, IConnectionContainer<PG, PD, PH>, IValueResolver {

	Comparator<IConfigurationElement<?, ?, ?>> NAME_COMPARATOR = Comparator.comparing(IConfigurationElement::getName);
	Comparator<Map.Entry<? extends IConfigurationElement<?, ?, ?>, ?>> KEY_COMPARATOR = Comparator.comparing(Map.Entry::getKey, IConfigurationElement.NAME_COMPARATOR);

	/**
	 * Prefix name with subsystem name, if in subsystem
	 *
	 * @return [{SUB_NAME}_]{MODULE_NAME}
	 */
	String getGlobalName();

	boolean isValid(boolean report);

	/**
	 * Get parent
	 *
	 * @return null, if no parent
	 */
	IConfigurationElement<?, ?, ?> getParent();

	/**
	 * Track all parents up to root, return string representation of this chain
	 *
	 * @return / separated chain of elements
	 */
	String getAbsolutePath();

	/**
	 * Type of this element
	 *
	 * @return type
	 */
	ElementType getType();

	/**
	 * Get a Handler that allows validating / changing this elements name
	 *
	 * @return nameHandler, cached internally
	 */
	IAttributeHandler getNameHandler();

	/**
	 * List currently set generic attributes
	 *
	 * @return Attribute names
	 */
	Set<String> getKnownAttributes();

	/**
	 * Get a Handler that allows validating / changing / setting a generic attribute
	 * Generic attributes are all in string form and not otherwise limited, any name is valid. Since there is no description, validation will always return true
	 * <p>
	 * Here for legacy reasons, allows anybody to add information to any element and retrieve it later
	 *
	 * @param attribute name of an attribute.
	 * @return handler
	 */
	IAttributeHandler getAttributeHandler(String attribute);

	/**
	 * Register for events of this element
	 *
	 * @param listener listener
	 */
	void addElementListener(IElementChangeListener listener);

	/**
	 * Unregister from events of this element
	 *
	 * @param listener listener
	 */
	void removeElementListener(IElementChangeListener listener);

	boolean wasRemoved();

	/**
	 * Type of element, relates to implementation classes
	 */
	enum ElementType {
		CommonModule, Peripheral, Subsystem, MemoryModule, Configuration
	}


	/**
	 * Any change to the implementing element.
	 */
	interface IElementChangeListener {

		/**
		 * A connector has changed: either a connection was added or removed or the properties like width or direction of the connector changed
		 *
		 * @param element parent element of connector
		 * @param connector connector in question
		 * @param type type of change
		 * @param connection connection, null for property changes
		 */
		void onConnectorChanged(IConfigurationElement<?, ?, ?> element, IConnectorHandler<?, ?> connector, ConnectorChangeType type, IConnection<?> connection);

		/**
		 * Elements name has changed (Name is valid)
		 *
		 * @param element element in question
		 */
		void onElementNameChanged(IConfigurationElement<?, ?, ?> element);

		/**
		 * A generic attribute was set / changed
		 *
		 * @param element parent element
		 * @param attribute attribute name
		 */
		void onAttributeChanged(IConfigurationElement<?, ?, ?> element, String attribute);

		/**
		 * The Declaration / Description of an element has changed.
		 * This means the list of supported connectors (buses / ports / pins) and parameters may have changed, basically everything on this element should be reloaded
		 *
		 * @param element most likely a DynamicStructureModuleConfiguration (ie. chipsoce, generated ip-core)
		 */
		void onElementDeclarationChanged(IConfigurationElement<?, ?, ?> element);

		enum ConnectorChangeType {
			Removed, Added, Properties
		}
	}
}
