/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.api;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.INamedValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IModuleContainer;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentExporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentImporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager.BaseBuildManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.handlers.IAvailablePortsHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.IDocumentStatusReporter;

import org.jetbrains.annotations.Nullable;

import java.nio.file.Path;
import java.util.Optional;
import java.util.function.Function;

/**
 * The API for accessing the DocumentHandler. Reduces all return values to implementations of interfaces, equally present in the API package.
 * Hides all methods that for implementation purposes need to be public, but should not be exposed via the API
 *
 * @author Ramon Wirsch
 */
public interface IDocumentHandler {

	/**
	 * The document has been changed since last save operation
	 *
	 * @return unsaved changes
	 */
	boolean needsSaving();

	/**
	 * Create a new, empty Document. Target and device can be changed later, but must not be null
	 *
	 * @param target the Target Board, for which to build
	 * @param device device to target on the board
	 * @return newly created document
	 */
	IConfigurationDocument newDocument(TargetDescription target, DeviceDescription device);

	/**
	 * Extract target and device names from String: target[:device]
	 *
	 * @param targetConf {@link de.tu_darmstadt.rs.spartanmc.jconfig.util.TargetConf} String
	 * @return newly created document
	 */
	IConfigurationDocument newDocument(String targetConf);

	/**
	 * Get a DocumentBuilder that can build the current document.
	 *
	 * @return builder, toolchain specific helper
	 */
	DocumentBuilder getDocumentBuilder();

	void buildIncremental();

	/**
	 * Convenience Method, get the documentBuilder and build()
	 *
	 * @param incremental build only outputs detected as changed since last build
	 */
	void build(boolean incremental);

	void buildWith(Function<DocumentBuilder, BaseBuildManager> buildManagerSupplier, boolean requireValidate, boolean incremental);

	/**
	 * Check the validity of the current document.
	 * Will recursively check every property for its validity.
	 * Checking is robust, will continue after first invalid occurrence.
	 * Will generate issues for every detected invalidity
	 *
	 * @return all parts were valid
	 */
	boolean isValid();

	/**
	 * The documentHandler has a location to which the current document can be saved.
	 *
	 * @return {@link #save()} can be called
	 */
	boolean hasSaveLocation();

	/**
	 * Save the current document to the current saveLocation
	 * <p>
	 * Document will be saved to a temporary file, only on success will old data be overwritten.
	 */
	void save();

	/**
	 * Generic mechanism to generate a document from a file. Allows plugging in of different file formats
	 *
	 * @param document source file
	 * @param reader implementation, that can read file and build a document from it
	 */
	void importFile(Path document, IDocumentImporter reader);

	/**
	 * Generic mechanism to store a document to a file. Allows plugging in of different file formats
	 *
	 * @param document target file
	 * @param writer implementation, that can write any document to a file
	 */
	void exportFile(Path document, IDocumentExporter writer);

	/**
	 * Look if there is an autosave file for the current saveLocation
	 * Autosaves are detected by their names {org}.autosave.jc3
	 *
	 * @return file if there is one
	 */
	Optional<Path> findAutosave();

	/**
	 * Read in a JC3 configuration from file, replace current document, if there is one
	 *
	 * @param document existing file in JC3 format
	 */
	void open(Path document);

	/**
	 * Used to lookup possible ports for autocompletion. Only valid for the current document, cached internally
	 *
	 * @return AvailablePortsHandler
	 */
	IAvailablePortsHandler getAvailablePortsHandler();

	/**
	 * Read out the current save location
	 *
	 * @return null, if none is set
	 */
	@Nullable Path getSaveLocation();

	/**
	 * Change the current save location
	 *
	 * @param saveLocation null, to remove saveLocation
	 */
	void setSaveLocation(Path saveLocation);

	/**
	 * Get projects directory (set during creation of {@link ILibJConfigProject})
	 *
	 * @return Directory
	 */
	Path getWorkingDir();

	/**
	 * A Document is currently loaded and can be accessed through {@link #getDocument()}
	 *
	 * @return null, if none
	 */
	boolean hasLoadedDocument();

	/**
	 * @return Current Document or null
	 */
	IConfigurationDocument getDocument();

	/**
	 * Convenience Method for parent.addModule. Add a module of type to parent. Checks whether parent supports child of module.
	 * Use this if you do not know the class of parent, otherwise access parent itself
	 *
	 * @param parent the parent to which to add
	 * @param moduleDescription the type of module to create
	 * @param <P> Type of parent module
	 * @return the module created
	 */
	<P extends IModuleContainer & IConfigurationElement & INamedValueResolver> IModuleConfiguration<P> addModule(IConfigurationElement<?, ?, ?> parent, IModuleDescription moduleDescription);

	/**
	 * Convenience Method for parent.removeModule
	 *
	 * @param parent parent from which to remove
	 * @param child module to remove
	 * @return module was removed
	 */
	boolean removeModule(IConfigurationElement<?, ?, ?> parent, IConfigurationElement<?, ?, ?> child);

	/**
	 * Convenience Method for module.parent.removeModule(module)
	 *
	 * @param module module which to remove from its parent
	 * @return module was removed
	 */
	boolean removeModule(IConfigurationElement<?, ?, ?> module);

	/**
	 * Register for documentHandler events:
	 * - new document present
	 * - no document present
	 *
	 * @param listener listener
	 */
	void addDocumentHandlerListener(IDocumentHandlerListener listener);

	/**
	 * Unregister for documentHandler events
	 *
	 * @param listener listener
	 */
	void removeDocumentHandlerListener(IDocumentHandlerListener listener);

	/**
	 * Check if special conditions are met which lead to dependencies on module
	 * This will not be implemented in libjconfig, it can optionally be handled by the PluginEngine
	 *
	 * @param module any module
	 * @return true if dependency detected, defaults to false
	 */
	boolean isModuleInUse(IModuleConfiguration module);

	boolean hasPluginEngine();

	@Nullable IPluginEngine getPluginEngine();

	/**
	 * Attach a "PluginEngine" that gets notified of all events on the document and its elements, to allow injection of additional functionality.
	 * PluginEngine gets notified before the normal listeners
	 *
	 * @param pluginEngine plugin implementation
	 */
	void setPluginEngine(IPluginEngine pluginEngine);

	void addPluginEngineListener(IPluginEngineListener listener);

	void removePluginEngineListener(IPluginEngineListener listener);

	void setDocumentStatusReporter(@Nullable IDocumentStatusReporter statusReporter);

	void doNotCleanupFile(Path file);

	/**
	 * Listener for changes on the DocumentHandler itself
	 */
	interface IDocumentHandlerListener {

		/**
		 * Called whenever the jc3 document changes. E.g. a new document was begun or another document was opened
		 *
		 * @param valid whether we currently have a valid document
		 */
		void onDocumentSwitched(boolean valid);

		/**
		 * Called whenever the save location of an existing document changes.
		 */
		default void onSaveLocationChanged() {

		}
	}


	interface IPluginEngineListener {

		void onPluginEngineChanged(IPluginEngine pluginEngine);
	}
}
