/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.api;

import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentExporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentImporter;

import org.jetbrains.annotations.NotNull;

import java.util.Optional;

/**
 * Implementation of this interface can be announced on DocumentHandler, which will then ensure that this implementation becomes the single pluginEngine for all modules that exist now and in the future
 * PluginEngine gets notified of the same events as the normal listeners, but is guaranteed to be notified first and even during ongoing transaction. This allows to make changes here and notifying users / UI
 * only after all changes haven been made.
 * <p>
 * Created by ramon on 17.03.2016.
 */
public interface IPluginEngine extends IConfigurationElement.IElementChangeListener, IModuleConfiguration.IModuleChangeListener, IConfigurationDocument.IDocumentChangeListener, IMemoryModuleConfiguration.IMemoryChangeListener, IConfigurationDocument.IConfigurationHierarchyListener {

	void onAttached(ILibJConfigProject project);

	/**
	 * A JC3 document has been (re-)opened
	 *
	 * @param document a document of unknown contents
	 */
	void onOpenedDocument(IConfigurationDocument document);

	/**
	 * A new document has been imported. Allows adding to the import logic
	 *
	 * @param document new document
	 * @param importer importer to customize depending on format
	 */
	void onImportedDocument(IConfigurationDocument document, IDocumentImporter importer);

	/**
	 * A new, empty document has just been created.
	 * It will be announced to listeners next
	 *
	 * @param document empty document
	 */
	void onCreatedDocument(IConfigurationDocument document);

	/**
	 * Functionality not implemented in libjconfig. Determine whether the given module is somehow required at this time.
	 * Use of the response is not specified. DocumentHandler defaults to false
	 *
	 * @param module module in question
	 * @return true if required
	 */
	boolean isModuleInUse(IModuleConfiguration module);

	/**
	 * Document was unloaded
	 */
	void onNoDocument();

	/**
	 * Beginning the save process. Allows making final changes, tracking actually saved configs
	 *
	 * @param document document to save
	 */
	void onSaving(IConfigurationDocument document);

	/**
	 * Beginning the export process. Allows making final changes.
	 *
	 * @param document document to export
	 * @param exporter exporter, to customize depending on target
	 */
	void onExportingDocument(IConfigurationDocument document, IDocumentExporter exporter);

	/**
	 * Beginning of build process. Allows making final changes before handing off to DocumentBuilder
	 *
	 * @param document document building
	 */
	void onBuilding(IConfigurationDocument document);

	/**
	 * Additional logic to check, whether the module in question is valid.
	 * Allows checks that take context into account (like parameters that depend on each other)
	 * Note: if a check is used widely, it should be integrated into libjconfig
	 *
	 * @param module module under test
	 * @param report whether to create issues for detected problems
	 * @return true if valid according to our metrics
	 */
	boolean isValid(IModuleConfiguration<?> module, boolean report);

	/**
	 * @param configurationDocument document under test
	 * @param report whether to create issues for detected problems
	 * @return true if valid according to our metrics
	 * @see #isValid(IModuleConfiguration, boolean) for whole Document
	 */
	boolean isValid(IConfigurationDocument configurationDocument, boolean report);

	@NotNull
	Optional<IScriptManager> getScriptManager();
}
