/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.api;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;

import org.jetbrains.annotations.NotNull;

import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Iterator;
import java.util.Set;

public abstract class OrderedConnection<P extends IConnectorInstance> implements IConnection<P> {


	public static Comparator<OrderedConnection<? extends IConnectorInstance>> COMPARATOR = Comparator
			.<OrderedConnection<? extends IConnectorInstance>, String>comparing(con -> con.getPortA().getAbsoluteConnectorName())
			.thenComparing(con -> con.getPortB().getAbsoluteConnectorName());

	protected abstract P getPortA();

	protected abstract P getPortB();


	protected abstract int getStartA();

	protected abstract int getEndA();

	protected abstract int getStartB();

	protected abstract int getEndB();

	protected abstract boolean isPartialA();

	protected abstract boolean isPartialB();


	@Override
	public P getOppositeConnector(IConnectorInstance me) {
		return getPortA() == me? getPortB() : getPortA();
	}


	@Override
	public int getOppositeStart(IConnectorInstance me) {
		return getPortA() == me? getStartB() : getStartA();
	}

	//TODO document whether end is inclusive or exclusive
	@Override
	public int getOppositeEnd(IConnectorInstance me) {
		return getPortA() == me? getEndB() : getEndA();
	}

	@Override
	public int getMyStart(IConnectorInstance me) {
		return getPortA() != me? getStartB() : getStartA();
	}

	@Override
	public int getMyEnd(IConnectorInstance me) {
		return getPortA() != me? getEndB() : getEndA();
	}

	@Override
	public boolean isOppositePartial(IConnectorInstance me) {
		return getPortA() == me? isPartialB() : isPartialA();
	}

	@Override
	public boolean isMyPartial(IConnectorInstance me) {
		return getPortA() != me? isPartialB() : isPartialA();
	}


	@Override
	public boolean isPartial() {
		return isPartialA() || isPartialB();
	}


	private PortSet ports = null;

	@Override
	public Set<P> getPorts() {
		if (ports == null) {
			ports = new PortSet();
		}
		return ports;
	}

	private class PortSet implements Set<P> {

		@Override
		public int size() {
			return 2;
		}

		@Override
		public boolean isEmpty() {
			return false;
		}

		@Override
		public boolean contains(Object o) {
			return o.equals(getPortA()) || o.equals(getPortB());
		}

		@NotNull
		@Override
		public Iterator<P> iterator() {
			return new Iterator<P>() {
				int idx = 0;

				@Override
				public boolean hasNext() {
					return idx < 2;
				}

				@Override
				public P next() {
					if (idx == 0) {
						idx++;
						return getPortA();
					}
					if (idx == 1) {
						idx++;
						return getPortB();
					}
					throw new ArrayIndexOutOfBoundsException();
				}
			};
		}

		@NotNull
		@Override
		public Object[] toArray() {
			return new Object[]{getPortA(), getPortB()};
		}

		@NotNull
		@Override
		public <T> T[] toArray(@NotNull T[] a) {
			T[] res = a;
			if (a.length < 2) {
				res = Arrays.copyOf(a, 2);
			}
			res[0] = (T) getPortA();
			res[1] = (T) getPortB();

			for (int i = 2; i < res.length; i++)
				res[i] = null;

			return res;

		}

		@Override
		public boolean add(P p) {
			throw new UnsupportedOperationException();
		}

		@Override
		public boolean remove(Object o) {
			throw new UnsupportedOperationException();
		}

		@Override
		public boolean containsAll(@NotNull Collection<?> c) {
			P a = getPortA();
			P b = getPortB();
			for (Object o : c) {
				if (!o.equals(a) && !o.equals(b)) {
					return false;
				}
			}
			return true;
		}

		@Override
		public boolean addAll(@NotNull Collection<? extends P> c) {
			throw new UnsupportedOperationException();
		}

		@Override
		public boolean retainAll(@NotNull Collection<?> c) {
			throw new UnsupportedOperationException();
		}

		@Override
		public boolean removeAll(@NotNull Collection<?> c) {
			throw new UnsupportedOperationException();
		}

		@Override
		public void clear() {
			throw new UnsupportedOperationException();
		}
	}
}
