/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.api;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.LinkedList;
import java.util.List;

/**
 * A specific port described relative to another module.
 * <p>
 * Example:
 * {@link IConnectorDescription} + {@link IConfigurationElement} describe a specific port on a specific element (caler must guarantee, that the portDescription exists on element)
 * For convenience we want the path of a port in relation to what we are connecting:
 * when connecting port "subsystem_0.led_0", sourceElement is the current subsystem
 * "subsystem_0.led_1" should show up as ".led_1"
 * "subsystem_0.uart_light_0.tx" should show up as "uart_light_0.tx"
 * "subsystem_1.uart_light_1.tx" should show up as "/subsystem_1.uart_light_1.tx"
 * FPGA pin x should show up as #{targetPinName} ({devicePinName})
 *
 * @author Ramon Wirsch
 */
public class ResolvedConnector implements Comparable<ResolvedConnector> {

	private static final Logger logger = LoggerFactory.getLogger(ResolvedConnector.class);
	private final IConfigurationElement<?, ?, ?> relativeTo;
	private String string;
	private List<String> hints = new LinkedList<>();
	private String[] hintArr;
	private IConnectorHandler portHandler;
	private Object connectionHint;
	private boolean pin = false;

	private ResolvedConnector(String name) {
		this.relativeTo = null;
		this.string = name;

	}

	/**
	 * Create a resolved connector of a portHandler and the module to use to address it relatively
	 *
	 * @param portHandler port to resolve
	 * @param relativeTo current module
	 */
	public ResolvedConnector(IConnectorHandler portHandler, IConfigurationElement<?, ?, ?> relativeTo) {
		this.portHandler = portHandler;
		this.relativeTo = relativeTo;

		this.string = analyze();

		hints.add(this.string);
		hintArr = hints.toArray(new String[hints.size()]);
		hints = null;
	}

	/**
	 * Create a resolved connector wich is actually connected
	 * @param portHandler portHandler
	 * @param relativeTo current mocule
	 * @param connection connection of port
	 */
	public ResolvedConnector(IConnectorHandler portHandler, IConfigurationElement<?, ?, ?> relativeTo, IConnection<?> connection) {
		this(portHandler, relativeTo);

		this.string += "  " + VerilogLang.getVerilogBus(connection, connection.getOppositeConnector(portHandler));
	}

	/**
	 * Get a dummy port with the specified name.
	 * This should only be used for UI layout (JComboBox.PrototypeDisplayValue), not as actual values
	 *
	 * @param name name of dummy port
	 * @return dummy ResolvedConnector
	 */
	public static ResolvedConnector getDummyPort(String name) {
		return new ResolvedConnector(name);
	}

	private String analyze() {
		//TODO maybe reuse getAbsolutePath of IConnectorHandler

		IConnectorDescription port = portHandler.getConnectorDescription();
		hints.add(port.getName());//raw without prefix

		if (port instanceof TargetDescription.IoPin) { // External PIN

			hints.add(((TargetDescription.IoPin) port).getFpgaPinName());
			hints.add(portHandler.getPrefixedConnectorName());

			pin = true;
			return portHandler.getPrefixedConnectorName() + " (" + ((TargetDescription.IoPin) port).getFpgaPinName() + ")";
		}

		hints.add(portHandler.getPrefixedConnectorName());//prefixed to guarantee namespace uniqueness (@ . #)

		//		IConfigurationElement<?,?,?> sourceElement = portHandler.getSourceElement();

		//		if (sourceElement == relativeTo) { // on same module as we
		//			return portHandler.getPrefixedConnectorName();
		//		}

		//		String elementName = sourceElement.getName();
		//
		//		IConfigurationElement parent = sourceElement.getParent();

		//		if (parent != null && parent.getType() == IConfigurationElement.ElementType.Subsystem && parent == relativeTo) { //same processor
		//			return elementName + portHandler.getPrefixedConnectorName();
		//		}

		return portHandler.getAbsoluteConnectorName();
	}

	public String[] getHints() {
		return hintArr;
	}

	@Override
	public int hashCode() {
		return string.hashCode();
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		ResolvedConnector that = (ResolvedConnector) o;

		return string.equals(that.string);

	}

	@Override
	public String toString() {
		return string;
	}

	public IConnectorHandler getConnectorHandler() {
		return portHandler;
	}

	@Override
	public int compareTo(@NotNull ResolvedConnector o) {
		int v = (o.pin == this.pin)? 0 : (o.pin)? -1 : 1;
		return (v != 0)? v : String.CASE_INSENSITIVE_ORDER.compare(string, o.string);
	}

	public void setConnectionHint(Object hint) {
		this.connectionHint = hint;
	}

	public boolean isConnectionHintMatch(Object hint) {
		return hint != null && (connectionHint == hint || hint.equals(connectionHint));
	}


}
