/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;

import java.util.Optional;
import java.util.stream.Stream;

/**
 * Interface for all ConnectorHandlers (Bus, Pin, Port). Simple ReadOnly information has been extracted to IConnectorInstance
 *
 * @author Ramon Wirsch
 */
public interface IConnectorHandler<PH extends IConnectorHandler<?, ?>, CD extends IConnectorDescription> extends IConnectorInstance {

	static boolean isDirectionCompatible(IConnectorHandler a, IConnectorHandler b) {
		return !a.isDirectionLocked() || !b.isDirectionLocked()
			|| (a.getDirection() == ConnectorDirection.InOut && b.getDirection() == ConnectorDirection.InOut)
			|| (a.getDirection() == ConnectorDirection.Input && b.getDirection() == ConnectorDirection.Output)
			|| (a.getDirection() == ConnectorDirection.Output && b.getDirection() == ConnectorDirection.Input);
	}

	String getPrefixedConnectorName();

	String getAbsoluteConnectorName();

	CD getConnectorDescription();

	boolean isPresent(IConnection<PH> connection);

	/**
	 * Maximum number of connections that can be attached to this port.
	 * Only outputs can have multiple connections
	 *
	 * @return 0 if no limit
	 */
	int getMaxAllowedConnections();

	boolean isMultipleConnectionsAllowed();

	/**
	 * Finds connection that connects us to handler
	 *
	 * @param connectorHandler a handler that we are currently connected to
	 * @return optional connection if present
	 */
	Optional<? extends IConnection<PH>> findConnection(IConnectorInstance connectorHandler);

	boolean isConnectibleTo(IConnectorHandler toPort);

	Optional<? extends IPartialConnectorHandler<PH, CD>> getPartialConnectorHandler();

	boolean isAdditionalConnectionAllowed();

	/**
	 * Stream currently held connections.
	 *
	 * @return Could be partial connections
	 */
	Stream<? extends IConnection<PH>> streamConnections();

	default Stream<? extends PH> streamOppositeConnectors() {
		return streamConnections().map(c -> c.getOppositeConnector(this));
	}

	PH getSingleConnectedPort();

	boolean isValid(boolean report);

	/**
	 * Convenience Method to call {@link IConnection#disconnect()} on all active connections
	 */
	void disconnectAll();

	boolean isDirectionLocked();

	default boolean isConnectedTo(IConnectorHandler<IPortHandler<?>, ? extends IPort> portHandler) {
		return streamOppositeConnectors().anyMatch(oc -> oc == portHandler);
	}
}
