/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.util.FileUtil;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.nio.file.Path;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.List;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.BinaryOperator;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collector;
import java.util.stream.Collectors;

/**
 * Created by ramon on 01.04.2016.
 */
public abstract class BaseBuildManager implements IBuildManager {

	public static final Marker buildMarker = MarkerFactory.getMarker("Build");
	private static final Logger logger = LoggerFactory.getLogger(BaseBuildManager.class);

	protected final DocumentBuilder builder;
	protected List<Path> noCleanupFiles;

	public BaseBuildManager(DocumentBuilder builder) {
		this.builder = builder;
	}

	private static boolean buildGroup(BuildState state, DocumentBuilder.IOutputGroup group) {
		logger.trace(buildMarker, "Building {}", group);
		try {
			List<MessageDigest> digests = group.build();
			logger.trace(buildMarker, "Successfully built {}", group.getName());
			//			StreamUtil.streamZipped(digests.iterator(), group.getOutputFiles().iterator()).forEach(z -> logger.debug(buildMarker, "Generated {}, digest: {}", z.getB(), DigestUtil.prettifyDigest(z.getA())));
			state.addBuilt(group, digests);
			return true;
		} catch (Exception e) {
			logger.error(buildMarker, "Failed to build group {}", group.getName(), e);
			state.addException(new BuildException(group, e));
			return false;
		}
	}

	@NotNull
	protected List<Path> cleanupUnneededFiles(Path outputDir, List<? extends DocumentBuilder.IOutputGroup> outputGroups, Collection<Path> doNotCleanup) {
		List<Path> cleanedUpFiles = new ArrayList<>();

		Set<Path> outputFiles = outputGroups.stream().flatMap(gr -> gr.getOutputFiles().stream()).collect(Collectors.toSet());

		if (doNotCleanup != null && !doNotCleanup.isEmpty()) {
			outputFiles.addAll(doNotCleanup);
		}

		FileUtil.cleanupFiles(outputDir, outputFiles, cleanedUpFiles);
		return cleanedUpFiles;
	}

	protected Path getOutputDir(IConfigurationDocument document) {
		return document.getProject().getIntermediaryBuildDir();
	}

	public void setNoCleanupFiles(List<Path> noCleanupFiles) {
		this.noCleanupFiles = noCleanupFiles;
	}


	public static class BuildCollector implements Collector<DocumentBuilder.IOutputGroup, BuildState, BuildResult> {

		private final List<Path> cleanedUpFiles;
		private final Path configurationDocument;
		private final long start;
		private final List<? extends DocumentBuilder.IOutputGroup> upToDate;

		public BuildCollector(List<Path> cleanedUpFiles, Path configurationDocument, long start) {
			this(cleanedUpFiles, Collections.emptyList(), configurationDocument, start);
		}

		public BuildCollector(List<Path> cleanedUpFiles, List<? extends DocumentBuilder.IOutputGroup> upToDate, Path configurationDocument, long start) {
			this.upToDate = upToDate;
			this.cleanedUpFiles = cleanedUpFiles;
			this.configurationDocument = configurationDocument;
			this.start = start;
		}

		@Override
		public Supplier<BuildState> supplier() {
			return BuildState::new;
		}

		@Override
		public BiConsumer<BuildState, DocumentBuilder.IOutputGroup> accumulator() {
			return BaseBuildManager::buildGroup;
		}

		@Override
		public BinaryOperator<BuildState> combiner() {
			return BuildState::merge;
		}

		@Override
		public Function<BuildState, BuildResult> finisher() {
			return (state) -> {
				state.verifyUniqueOutputGroups(upToDate);
				return new BuildResult(state, cleanedUpFiles, upToDate, configurationDocument, System.currentTimeMillis() - start);
			};
		}

		@Override
		public Set<Characteristics> characteristics() {
			return EnumSet.of(Characteristics.UNORDERED);
		}
	}

}
