/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager;

import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.security.MessageDigest;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collector;
import java.util.stream.Collectors;

/**
 * Created by Ramon on 17.05.2016.
 */
public class BuildResult {

	private static final Logger logger = LoggerFactory.getLogger(BuildResult.class);

	private final List<BuildException> exceptions;
	private final List<BuiltGroup> built;
	private final List<Path> cleanedUpFiles;
	private final List<? extends DocumentBuilder.IOutputGroup> upToDate;
	private final Path configurationFile;
	private final long duration;

	public BuildResult(BuildState state, List<Path> cleanedUpFiles, List<? extends DocumentBuilder.IOutputGroup> upToDate, Path configurationFile, long duration) {
		this.exceptions = Collections.unmodifiableList(state.getExceptions());
		this.built = Collections.unmodifiableList(state.getBuilt());
		this.cleanedUpFiles = Collections.unmodifiableList(cleanedUpFiles);
		this.upToDate = upToDate;
		this.configurationFile = configurationFile;
		this.duration = duration;
	}

	public BuildResult() {
		exceptions = Collections.emptyList();
		built = Collections.emptyList();
		cleanedUpFiles = Collections.emptyList();
		upToDate = Collections.emptyList();
		configurationFile = null;
		duration = 0;
	}

	private static Collector<CharSequence, ?, String> getArrayJoiner() {
		return Collectors.joining(", ", "[", "]");
	}

	public void report() {
		Path file = getConfigurationFile();
		String time = String.format("%.2f secs", getDuration() / 1000f);
		logger.info("Built from {} in {}", file != null? file : "Unsaved File", time);
		List<Path> cleanedUpFiles = getCleanedUpFiles();
		logger.info("Cleaned Up: {} files {}", cleanedUpFiles.size(), cleanedUpFiles.stream().map(Path::toString).collect(getArrayJoiner()));
		List<? extends DocumentBuilder.IOutputGroup> upToDate = getUpToDate();
		logger.info("Up-To-Date: {} groups {}", upToDate.size(), upToDate.stream().map(DocumentBuilder.IOutputGroup::getName).collect(getArrayJoiner()));
		List<BuiltGroup> built = getBuilt();
		logger.info("Written: {} groups {}", built.size(), built.stream().map(BuiltGroup::getOutputGroup).map(DocumentBuilder.IOutputGroup::getName).collect(getArrayJoiner()));
	}

	public long getDuration() {
		return duration;
	}

	public boolean wasSuccessful() {
		return exceptions.isEmpty();
	}

	public List<BuildException> getExceptions() {
		return exceptions;
	}

	public List<BuiltGroup> getBuilt() {
		return built;
	}

	public List<Path> getCleanedUpFiles() {
		return cleanedUpFiles;
	}

	public List<? extends DocumentBuilder.IOutputGroup> getUpToDate() {
		return upToDate;
	}

	public Path getConfigurationFile() {
		return configurationFile;
	}

	public static class BuiltGroup {

		private final DocumentBuilder.IOutputGroup group;
		private final List<MessageDigest> digests;
		private final long timestamp;

		public BuiltGroup(DocumentBuilder.IOutputGroup group, List<MessageDigest> digests) {
			this.group = group;
			this.digests = digests;
			this.timestamp = System.currentTimeMillis();
		}

		public long getTimestamp() {
			return timestamp;
		}

		public DocumentBuilder.IOutputGroup getOutputGroup() {
			return group;
		}

		public List<MessageDigest> getDigests() {
			return digests;
		}
	}
}
