/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder;

import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.IToolchainProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.ToolchainRegistry;

import java.io.IOException;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

/**
 * This class is the workhorse of the build process.
 * Since you should get an instance from a {@link IToolchainProvider}, the provider is responsible for ensuring compatibility.
 * A DocumentBuilder instance is not bound to any specific document, but will return the specific grouped outputs for a specific document in {@link #getOutputGroups(IConfigurationDocument)}
 * <p>
 * Each OutputGroup can be built independently of other groups and specifies its dependencies.
 * <p>
 * Created by ramon on 01.04.2016.
 */
public abstract class DocumentBuilder {

	private static ToolchainRegistry registry;

	/**
	 * Use the global ToolchainRegistry to find a ToolchainImplementation that supports whatever toolchain that the current device requires.
	 *
	 * @param document only device-&gt;toolchain is read.
	 * @return a fresh DocumentBuilder that can build the given document. The Builder is not attached to any specific document. It could be reused.
	 */
	public static DocumentBuilder getBuilder(IConfigurationDocument document) {
		if (registry == null) {
			registry = new ToolchainRegistry();
		}

		Optional<DocumentBuilder> builder = registry.findToolchainBuilder(document);

		if (!builder.isPresent()) {
			throw new IllegalStateException("No Implementation available to build for toolchain: " + document);
		}

		return builder.get();
	}

	public static Path getLinkerScriptFile(Path systemDir, String name) {
		return systemDir.resolve("firmwares").resolve(String.format("%s.x", name));
	}

	public static Path getSubsystemFile(Path systemDir, String subName, String fileName) {
		return systemDir.resolve("subsystems").resolve(subName).resolve(fileName);
	}

	public abstract String getBuilderName();

	/**
	 * Return a description of this toolchain.
	 * Should give the same response as its {@link IToolchainProvider#getDescription()}
	 *
	 * @return arbitrary description of builder
	 */
	public abstract String getDescription();

	/**
	 * Return all outputs that can be build.
	 * Outputs are grouped by inputs/outputs, meaning that each group can be built or reused independently of the other groups.
	 * Members of a group though are built collectively.
	 * <p>
	 * Ex: A group per subsystem to build the [subsystem]/moduleParameters.h &amp; [subsystem]/moduleParameters.x.
	 *
	 * @param document ConfDoc that will be attached to all outputGroups.
	 * @return all output groups for document
	 */
	public abstract List<? extends IOutputGroup> getOutputGroups(IConfigurationDocument document);

	public abstract String getToolchainName();

	/**
	 * Collection of OutputFiles that can only be build together and their dependents.
	 * Note: this interface is meant to compatible to Gradle's incremental build facilities. It should map to a single Incremental Task
	 * <p>
	 * If desired, this may be used to check whether build of this specific group is required
	 * <p>
	 * IOutputGroups are short-lived and static. They become invalid if changes to the underlying document are made.
	 */
	public interface IOutputGroup {

		/**
		 * List all outputs that will be created / updated as a result of {@link #build()}
		 *
		 * @return all files of group
		 */
		List<Path> getOutputFiles();

		/**
		 * List all input files, on which this group's outputs depend. Any change in these files should merit a rebuild.
		 *
		 * @return all input files, this group depends on
		 */
		Set<Path> getInputFiles();

		/**
		 * List additional input properties, on which this group's outputs depend. These inputs only exist in memory and not as files.
		 * Properties means descriptive name and any instance.
		 * <p>
		 * Ex: settings in the configuration document.
		 * Contract: The Instances returned here, must implement a valid {@link #equals(Object)} method and must be Serializable.
		 * <p>
		 * (Gradle serializes all Properties and compares all of them using equals, adding/removing is determined by equality of their respective names)
		 *
		 * @return all properties this group depends on
		 */
		Map<String, ? extends Object> getInputProperties();

		/**
		 * Build all outputs of this group. Overwrite if they exist.
		 *
		 * @return MessageDigest for every output in {@link #getOutputFiles()}
		 * @throws IOException exception during output
		 */
		List<MessageDigest> build() throws IOException;

		/**
		 * Name to describe group
		 *
		 * @return unique name under which to file history
		 */
		String getName();
	}
}
