/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBooleanParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IFloatParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IIntegerParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterConverter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.IMappedBusPortInstance;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Locale;

/**
 * Created by Ramon on 11.05.2016.
 */
public class VerilogLang {

	private static final Logger logger = LoggerFactory.getLogger(VerilogLang.class);
	private static final IParameterConverter V_CONV = new VerilogParamConverter();

	public VerilogLang() {
		super();
	}

	/**
	 *
	 * @param direction direction of a signal on a module
	 * @return the direction as a string
	 */
	public static String getVerilogDirection(ConnectorDirection direction) {
		switch (direction) {
			case Input:
				return "input";
			case Output:
				return "output";
			case InOut:
				return "inout";
			default:
				throw new IllegalArgumentException("Invalid pin direction");
		}
	}

	public static String getVerilogName(String name) {
		return name.toLowerCase().replaceAll("[^_a-zA-Z0-9$]", "_");
	}

	public static String getVerilogBus(int width) {
		return width == 1L? "" : String.format("[%d:0]", width - 1);
	}

	public static String getVerilogBus(IConnection<?> connection, IConnectorInstance side) {
		boolean isBus = side.getConnectorWidth() > 1;
		return getVerilogBus(connection.getMyStart(side), connection.getMyEnd(side), connection.isMyPartial(side), isBus);
	}

	public static String getVerilogBus(int start, int end, boolean partial, boolean isBus) {
		if (start == end) {
			return (partial && isBus)? String.format("[%d]", start) : "";
		} else {
			return partial ? String.format("[%d:%d]", end, start) : "";
		}
	}

	public static String getVerilogBus(int start, int end, int totalWidth) {
		return getVerilogBus(start,end, (start!=0) || (end+1!=totalWidth), totalWidth>1);
	}

	public static String getModuleInstanceParam(String name, String value) {
		return String.format("\t.%-20s \t ( %-40s )", name, value);
	}

	public static String getWireName(IConnectorInstance port) {
		if (port.getType() != IConnectorDescription.Type.Pin) {
			return getVerilogName(port.getSourceElement().getGlobalName() + '_' + port.getPortName());
		} else {
			return getPinName(port);
		}
	}

	public static String getPinName(IConnectorInstance pin) {
		return getPinName(pin.getPortName());
	}

	public static String getPinName(String portName) {
		return String.format("pin_%s", portName);
	}

	public static String getVerilogAdr(long adr, int size) {
		return String.format("%d'h%x", size, adr);
	}

	public static String getMaskedAdr(long adr, int width) {
		adr = maskAddress(adr, width);
		return getVerilogAdr(adr, width);
	}

	public static long maskAddress(long adr, int width) {
		return adr & ((1 << width) - 1); // mask away anything beyond width bits
	}

	public static String getGlueName(IMappedBusPortInstance busPortInstance) {
		return getVerilogName(String.format("bus_%s_%s_%s", busPortInstance.getCombineType(), busPortInstance.getSourceElement().getName(), busPortInstance.getPortName()));
	}

	public static String getGlueOutputName(String glueName) {
		return glueName + "_out";
	}

	public static String getResolvedParameterValue(IParameterHandler param) {
		return param.getValueConverted(V_CONV);
	}

	public static class VerilogParamConverter implements IParameterConverter {

		public VerilogParamConverter() {
			super();
		}

		public static String convertInt(long value, int radix, String name) {
			String sign;
			if (value < 0) {
				sign = "-";
				value = Math.abs(value);
				logger.warn("Writing negative number {} literal to verilog as {}", value, name);
			} else {
				sign = "";
			}

			switch (radix) {
				case 8:
					return String.format("%s64'o%o", sign, value);
				case 2:
					return String.format("%s64'b%s", sign, Long.toBinaryString(value));
				case 16:
					return String.format("%s64'h%H", sign, value);
				default:
					return String.format("%s%d", sign, value);
			}
		}

		@Override
		public String convert(IIntegerParameterHandler param) {
			return convertInt(param.getIntegerValue().orElseThrow(), param.getRadix(), param.getCanonicalName());
		}

		@Override
		public String convert(IFloatParameterHandler param) {
			return String.format(Locale.US, "%.6f", param.getFloatValue().get());
		}

		@Override
		public String convert(IParameterHandler param) {
			return toWrappedString(param);
		}

		@Override
		public String convert(IBooleanParameterHandler param) {
			return shouldOutputAsInt(param)? Long.toString(param.getIntegerValue().orElseThrow()) : toWrappedString(param);
		}

		public static String toWrappedString(String value) {
			if (value==null)
				return "\"\"";
			return String.format("\"%s\"", value);

		}

		private String toWrappedString(IParameterHandler param) {
			return toWrappedString(param.getValue().orElseThrow());
		}

		private boolean shouldOutputAsInt(IBooleanParameterHandler param) {
			return param.getValue().get().equals("1") || param.getValue().get().equals("0");
		}
	}
}
