/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains;

import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.ServiceLoader;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

/**
 * Manages Known {@link IToolchainProvider}s detected via ServiceLoader
 * <p>
 * Created by ramon on 01.04.2016.
 */
public class ToolchainRegistry {

	private static final Logger logger = LoggerFactory.getLogger(ToolchainRegistry.class);
	private final Map<String, List<IToolchainProvider>> toolchains;

	public ToolchainRegistry() {
		ServiceLoader<IToolchainProvider> loader = ServiceLoader.load(IToolchainProvider.class);
		toolchains = StreamSupport.stream(loader.spliterator(), false).collect(Collectors.groupingBy(IToolchainProvider::getToolchainName));
		logger.info("Available Toolchains: "+toolchains.keySet());
	}

	/**
	 * Finds the best DocumentBuilder possible out of all known IToolchainProviders
	 *
	 * @param document the document for which a builder should be found
	 * @return the highest priority builder for the devices toolchain, that also returned true in {@link IToolchainProvider#isApplicable(IConfigurationDocument)}
	 */
	public Optional<DocumentBuilder> findToolchainBuilder(IConfigurationDocument document) {
		DeviceDescription.Toolchain toolchainConfig = document.getTarget().getToolchain();
		return Optional.ofNullable(toolchains.get(toolchainConfig.getToolchain()))
			.flatMap(list -> list.stream().sorted(Comparator.comparing(IToolchainProvider::getPriority).reversed())
				.filter(t -> t.isApplicable(document)).findFirst()
				.map(IToolchainProvider::getDocumentBuilder));
	}
}
