/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryLocator;

import org.jetbrains.annotations.NotNull;

import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlIDREF;

/**
 * Information about one block of initializable memory
 */
public class MemoryBlock implements IMemoryBlock {

	/**
	 * The hierarchical name of the block rams in the design
	 * <p>
	 * placeholder is an integer for the index of the bram
	 */
	@XmlElement
	private MemoryLocator bramPattern;

	/**
	 * Start address of the memory section, in target words
	 */
	@XmlAttribute
	private long start;

	/**
	 * Length of the memory section, in target words
	 */
	@XmlAttribute
	private long length;

	/**
	 * Name of the module that the memory is for
	 */
	@XmlAttribute
	@NotNull
	private String moduleName;

	/**
	 * How many bits are addressed as 1 line in memory. Ex: SpartanMC 18 bit
	 */
	@XmlAttribute
	private int bitsPerMemoryLine;

	/**
	 * How many of those bits are in each blockram. Ex: If there is only 1, then same as bitsPerMemoryLine. For SpartaMC split memories 9 bit.
	 */
	@XmlAttribute
	private int bitsPerBlockLine;

	@XmlIDREF
	@XmlAttribute
	@NotNull
	private MemFirmware firmware;

	@XmlAttribute
	@NotNull
	private MemoryLineLayout memoryLineLayout;

	/**
	 * Create a MemoryBlock
	 *
	 * @param bramPattern The hierarchical name of the block rams in the design
	 * @param start Start address of the memory section, in target words
	 * @param length Length of the memory section, in target words
	 * @param moduleName Name of the module that the memory is for
	 * @param bitsPerMemoryLine how many bits are accesses with a single address
	 * @param bitsPerBlockLine in which chunks is this memory line split, equal to bitsPerMemoryLine if no split
	 * @param firmware the firmware to program
	 * @param memoryLineLayout whether there are discardable bits within each line (for SpartanMC)
	 */
	public MemoryBlock(MemoryLocator bramPattern, long start, long length, @NotNull String moduleName, int bitsPerMemoryLine, int bitsPerBlockLine, @NotNull MemFirmware firmware, @NotNull MemoryLineLayout memoryLineLayout) {
		this.bramPattern = bramPattern;
		this.start = start;
		this.length = length;
		this.moduleName = moduleName;
		this.bitsPerMemoryLine = bitsPerMemoryLine;
		this.bitsPerBlockLine = bitsPerBlockLine;
		this.firmware = firmware;
		this.memoryLineLayout = memoryLineLayout;
	}

	/**
	 * Only for use by JAXB
	 */
	private MemoryBlock() {
	}

	@Override
	@NotNull
	public MemoryLocator getBramPattern() {
		return bramPattern;
	}

	@Override
	public long getStart() {
		return start;
	}

	@Override
	public long getLength() {
		return length;
	}

	@Override
	public boolean isBitfileInitializable() {
		return bramPattern != null;
	}

	@Override
	@NotNull
	public String getModuleName() {
		return moduleName;
	}

	@Override
	public int getBitsPerBlockLine() {
		return bitsPerBlockLine;
	}

	@Override
	public int getBitsPerMemoryLine() {
		return bitsPerMemoryLine;
	}

	@Override
	@NotNull
	public MemFirmware getFirmware() {
		return firmware;
	}

	@Override
	@NotNull
	public MemoryLineLayout getMemoryLineLayout() {
		return memoryLineLayout;
	}
}
