/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.BinaryOutputFormat;
import de.tu_darmstadt.rs.util.StreamUtil;

import java.nio.file.Path;
import java.util.HashSet;
import java.util.Set;
import javax.xml.bind.JAXBException;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementWrapper;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * Maps all memory in a design to block rams
 */

@XmlRootElement(name = "memory-map", namespace = "http://www.rs.tu-darmstadt.de/ns/spartanmc/memory/jc3")
public class MemoryMap {

	private static final int VERSION = 5;

	@XmlAttribute
	private int version;

	@XmlElement(name = "block")
	@XmlElementWrapper(name = "blocks")
	private final Set<MemoryBlock> blocks;
	@XmlAttribute
	private BinaryOutputFormat outputFormat;
	@XmlElement(name = "firmware")
	@XmlElementWrapper(name = "firmwares")
	private Set<MemFirmware> firmwares;
	private Path sourceFile;

	@XmlElement(name = "partName")
	private String partName;
	@XmlElement(name = "toolchain")
	private String toolchain;

	public MemoryMap(Set<MemoryBlock> blocks, Set<MemFirmware> firmwares, BinaryOutputFormat outputFormat, String partName, String toolchain) {
		this.blocks = blocks;
		this.firmwares = firmwares;
		this.outputFormat = outputFormat;
		this.partName = partName;
		this.toolchain = toolchain;
		this.version = VERSION;
	}

	private MemoryMap() {
		blocks = new HashSet<>();
	}

	public boolean isMatchingVersion() {
		return VERSION == version;
	}

	public BinaryOutputFormat getOutputFormat() {
		return outputFormat;
	}

	public Set<MemoryBlock> getBlocks() {
		return blocks;
	}

	public Path getSourceFile() {
		return sourceFile;
	}

	/**
	 * Place to store from which file the memory was unmarshalled, to resolve relative paths
	 *
	 * @param sourceFile xml file from which this data was loaded
	 */
	public void setSourceFile(Path sourceFile) {
		this.sourceFile = sourceFile;
	}

	public MemoryBlock findBlock(String peripheral) {
		return blocks.stream().filter(b -> b.getModuleName().equals(peripheral)).collect(StreamUtil.exactlyOne());//TODO separate validation and access
	}

	public String getPartName() {
		return partName;
	}

	public String getToolchain() {
		return toolchain;
	}

	public static MemoryMap fromFile(Path file) {
		try {
			MemoryMap map = new MemoryMapJaxB().unmarshall(file);
			map.setSourceFile(file);

			if (!map.isMatchingVersion()) {
				throw new RuntimeException("MemoryMap "+ file +" was not created by code compatible with this application! Please update it!");
			}

			return map;
		} catch (JAXBException e) {
			throw new RuntimeException("Failed to read "+ file +" as MemoryMap!", e);
		}
	}
}
