/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.issues;

import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.linker.api.ILinkerSection;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IMemoryAddressBlock;

import java.nio.file.Path;
import java.util.List;
import java.util.Optional;
import java.util.function.Predicate;

/**
 * Created by ramon on 17.03.2016.
 */
public abstract class ConfigurationIssue {

	protected IConfigurationDocument document;
	private final String message;
	private final String situation;
	private final IssueType level;

	public ConfigurationIssue(String message, String situation, IssueType level) {
		this.message = message;
		this.situation = situation;
		this.level = level;
	}

	public static ConfigurationIssue createSectionPlacementGuessed(ILinkerSection linkerSection, IMemoryAddressBlock picked, List<? extends IMemoryAddressBlock> alternatives, String situation) {
		return new SectionLayoutIssue(linkerSection, picked, alternatives, "Guessed placement of Linker Section", situation, IssueType.Warning);
	}

	public static ConfigurationIssue createMissingConnection(IConnectorHandler handler, String situation) {
		return new ConnectionIssue(handler, null, ConnectionIssue.Type.Unconnected, "Missing connection", situation, IssueType.Error);
	}

	public static ConfigurationIssue createMismatchedConnection(IConnectorHandler handler, IConnection connection, String situation) {
		return new ConnectionIssue(handler, null, ConnectionIssue.Type.InvalidConnection, "Mismatched Connection detected", situation, IssueType.Error);
	}

	public static ConfigurationIssue createMismatchedConnectionRemoved(IConnection connection, String situation) {
		return new ConnectionIssue(null, connection, ConnectionIssue.Type.RemovedConnection, "Invalid Connection detected and removed", situation, IssueType.Warning);
	}

	public static ConfigurationIssue createClockIssue(IConnectorHandler handler, String situation, String errorMessage) {
		return new ConnectionIssue(handler, null, ConnectionIssue.Type.Unconnected, "Has no clock Frequency: "+errorMessage, situation, IssueType.Error);
	}

	public static ConfigurationIssue createUnconnected(IConnectorHandler handler, String situation) {
		return new ConnectionIssue(handler, null, ConnectionIssue.Type.Unconnected, "Unconnected but required", situation, IssueType.Error);
	}

	public static ConfigurationIssue createRecursiveConnection(IConnectorHandler handler, IConnection connection, String situation) {
		return new ConnectionIssue(handler, connection, ConnectionIssue.Type.Notice, "Port is not allowed to be connected to sibling port!", situation, IssueType.Error);
	}

	public static ConfigurationIssue createTooManyConnections(IConnectorHandler handler, String situation) {
		return new ConnectionIssue(handler, null, ConnectionIssue.Type.TooMany, "Too many connections", situation, IssueType.Error);
	}

	public static ConfigurationIssue createConnectedButUnconfigured(IBusHandler busHandler, BusPort port, String situation) {
		return new BusConnectionIssue(busHandler, port, "Bus is connected, but unspecified frequency is used", situation, IssueType.Warning);
	}

	public static ConfigurationIssue createInvalidPinProperties(IPinHandler pinHandler, String message, String situation) {
		return new ConnectionIssue(pinHandler, null, ConnectionIssue.Type.Properties, "Pin has Invalid properties: "+message, situation, IssueType.Error);
	}

	public static ConfigurationIssue createInvalidPinProperties(IPinHandler pinHandler, String situation) {
		return new ConnectionIssue(pinHandler, null, ConnectionIssue.Type.Properties, "Pin has Invalid properties", situation, IssueType.Error);
	}

	public static ConfigurationIssue createModuleDoesNotExpose(IModuleConfiguration module, ISubsystemConfiguration subsystem, Bus implementedBy, Bus exposing, String situation) {
		return new AddressSpaceIssue(module, subsystem, implementedBy, exposing, "A peripheral module does not expose any usable addressBlocks although it is connected on a bus that requires it", situation);
	}

	public static ConfigurationIssue createPeripheralExposesUnmapped(IPeripheralModuleConfiguration<?> module, ISubsystemConfiguration subsystem, String type, String situation) {
		return new AddressSpaceIssue(module, subsystem, null, null, "A peripheral module exposes " + type + " but no mapping exists (most likely unconnected)", situation);
	}

	public static ConfigurationIssue createGhostConnection(IConnectorHandler problem, IConnection connection, String situation) {
		return new ConnectionIssue(problem, connection, ConnectionIssue.Type.RemovedConnection, "An incomplete Connection was detected and removed", situation, IssueType.Warning);
	}

	public static ConfigurationIssue createTargetDeclarationChange() {
		return new ConfigurationIssue("Check all connections manually! Validity across different Targets cannot be guaranteed", "TargetDescription change", IssueType.Warning) {
			@Override
			public String getHintText() {
				return "";
			}

			@Override
			public Actuality getActuality() {
				return Actuality.Unknown;
			}

			@Override
			public Optional<IConfigurationElement<?, ?, ?>> getReferencedModule() {
				return Optional.empty();
			}
		};
	}

	public static ConfigurationIssue createDescriptionChange(String message, String situation) {
		return new ConfigurationIssue(message, situation, IssueType.Warning) {
			@Override
			public String getHintText() {
				return "";
			}

			@Override
			public Actuality getActuality() {
				return Actuality.Unknown;
			}

			@Override
			public Optional<IConfigurationElement<?, ?, ?>> getReferencedModule() {
				return Optional.empty();
			}
		};
	}

	public static ConfigurationIssue createInvalidModuleName(IConfigurationElement module, String situation) {
		return new ModuleIssue(module, module.getNameHandler(), ModuleIssue.Type.InvalidHandler, "Invalid Name", situation, IssueType.Error);
	}

	public static ConfigurationIssue createParameterInvalidValue(IParameterHandler handler, String situation) {
		return new ModuleIssue(handler.getSourceElement(), handler, null, ModuleIssue.Type.InvalidParam, "Invalid parameter", situation, IssueType.Error);
	}

	public static ConfigurationIssue createParameterConflictResolved(IParameterHandler handler, String oldValue, String situation, boolean couldMigrate) {
		return new ModuleIssue(handler.getSourceElement(), handler, oldValue, ModuleIssue.Type.InvalidParam, "Parameter changed due to conflict with description", situation, couldMigrate? IssueType.Warning : IssueType.Error);
	}

	///// static creators

	public static ConfigurationIssue createModuleAddedButNotUsed(IModuleConfiguration module, String situation) {
		return new ModuleIssue(module, (IAttributeHandler) null, ModuleIssue.Type.NotUsed, "Module exists, but is not connected to anything.", situation, IssueType.Warning);
	}

	public static ConfigurationIssue createMissingModuleDescription(IConfigurationElement<?, ?, ?> parent, String name, String moduleDescriptionName, String situation) {
		return new DescriptionIssue(parent, name, moduleDescriptionName, "Module has been removed from parent because its Description is missing", situation, IssueType.Error);
	}

	public static <T extends IConfigurationElement> ConfigurationIssue createMissingProperty(T module, Predicate<T> fixedTester, String property, String situation, IssueType level) {
		return new PropertyIssue<T>(module, PropertyIssue.Type.Missing, property, fixedTester, "Property is required but was not set", situation, level);
	}

	public static ConfigurationIssue createMismatchedBusClks(IBusHandler busHandler, BusPort busPort, String situation) {
		return new BusConnectionIssue(busHandler, busPort, "Clock conflict: a bus connected to multiple clock sources is driven by different clocks", situation, IssueType.Error);
	}

	public static ConfigurationIssue createModuleNotSupportedIssue(IModuleConfiguration module, IToolchainConfiguration currentConfiguration, String situation) {
		return new ModuleIssue(module, currentConfiguration, ModuleIssue.Type.NotSupported, "Module not supported on this configuration", situation, IssueType.Error);
	}


	public static ConfigurationIssue createConstraintsFileDeprecatedIssue(IModuleConfiguration module, IToolchainConfiguration currentConfiguration, Path f, String situation) {
		return new ModuleIssue(module, currentConfiguration, ModuleIssue.Type.NotSupported, "Module constraints files (UCF,QSF,LPF) are no longer supported. Please enter constraints from "+f+" to module description instead.", situation, IssueType.Error);
	}

	public static ConfigurationIssue createMissingFileIssue(Path f, IAttributeHandler handler, String situation, IssueType type) {
		return new MissingFileIssue(f, handler,"File referenced but was not found", situation, type);
	}

	public IssueType getLevel() {
		return level;
	}

	public String getMessage() {
		return message;
	}

	public abstract String getHintText();

	public abstract Actuality getActuality();

	public abstract Optional<IConfigurationElement<?, ?, ?>> getReferencedModule();

	public String getSituation() {
		return situation;
	}

	public void setDocument(IConfigurationDocument document) {
		this.document = document;
	}

	@Override
	public int hashCode() {
		int result = message.hashCode();
		//result = 31 * result + situation.hashCode();
		result = 31 * result + level.hashCode();
		result = 31 * result + (document != null? document.hashCode() : 0);
		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		ConfigurationIssue that = (ConfigurationIssue) o;

		if (!message.equals(that.message)) {
			return false;
		}
		/*if (!situation.equals(that.situation)) {
			return false;
		}*/
		if (level != that.level) {
			return false;
		}
		return document != null? document.equals(that.document) : that.document == null;

	}

	@Override
	public String toString() {
		return '"' + getMessage() + "\"@\"" + getSituation() + "\": " + getHintText();
	}


	public enum IssueType {
		Error,
		Warning
	}


	public enum Actuality {
		/**
		 * The issue cannot be fixed in JConfig, Description changes are required
		 */
		Persistent,
		/**
		 * The Issue still exists
		 */
		Current,
		/**
		 * This is a warning, actuality cannot be checked automatically
		 */
		Unknown,
		/**
		 * The issue has been fixed and is no longer valid. An Issue should not be created with this state, it should only switch to this state after user interaction
		 * Issues that are resolved can be silently dismissed
		 */
		Resolved,

	}
}

