/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.issues;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;

import org.jetbrains.annotations.Nullable;

import java.util.Iterator;
import java.util.Optional;

/**
 * Created by ramon on 17.03.2016.
 */
class ConnectionIssue extends ConfigurationIssue {

	@Nullable
	private final IConnectorHandler handler;
	private final IConnection<? extends IConnectorHandler> connection;
	private final Type type;

	public ConnectionIssue(@Nullable IConnectorHandler handler, IConnection connection, Type type, String message, String situation, IssueType level) {
		super(message, situation, level);
		this.handler = handler;
		this.connection = connection;
		this.type = type;
		//TODO report runs before constructors of subclasses that may initialize fields. report adds the list to a set in the document, which requires the hashcode. Hashcode often uses these fields.
	}

	public IConnectorHandler getHandler() {
		return handler;
	}

	public IConnection getConnection() {
		return connection;
	}

	@Override
	public String getHintText() {
		StringBuilder sb = new StringBuilder();
		if (handler != null) {
			sb.append("Connector '").append(handler.getAbsoluteConnectorName()).append("'");
			sb.append(" (").append(handler.getType()).append(") on ").append(handler.getSourceElement().getName()).append(' ');
		}
		if (connection != null) {
			Iterator<? extends IConnectorHandler> it = connection.getPorts().iterator();
			String aName = it.next().getPortName();
			String bName = it.next().getPortName();
			sb.append("Connection: ").append(aName).append(" - ").append(bName);
			if (type == Type.RemovedConnection || connection.isValid(false)) {
				sb.append(", invalid");
			}
			sb.append('\n');
		}
		return sb.toString();
	}

	@Override
	public Actuality getActuality() {
		if (type == Type.Notice) {
			return Actuality.Unknown;
		} else if (type == Type.InvalidConnection) {
			return ((connection == null) || connection.isValid(false)) ? Actuality.Resolved : Actuality.Current;
		} else if (type == Type.Unconnected) {
			boolean resolved = handler.isConnected() || handler.getSourceElement().wasRemoved();
			return resolved? Actuality.Resolved : Actuality.Current;
		} else if (type == Type.Properties || type == Type.TooMany) {
			return handler.isValid(false)? Actuality.Resolved : Actuality.Current;
		} else {
			return Actuality.Unknown;
		}
	}

	@Override
	public Optional<IConfigurationElement<?, ?, ?>> getReferencedModule() {
		return handler != null? Optional.of(handler.getSourceElement()) : Optional.of(document);
	}

	@Override
	public int hashCode() {
		int result = super.hashCode();
		result = 31 * result + (handler != null? handler.hashCode() : 0);
		result = 31 * result + (connection != null? connection.hashCode() : 0);
		result = 31 * result + type.hashCode();
		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}
		if (!super.equals(o)) {
			return false;
		}

		ConnectionIssue that = (ConnectionIssue) o;

		if (handler != null? !handler.equals(that.handler) : that.handler != null) {
			return false;
		}
		if (connection != null? !connection.equals(that.connection) : that.connection != null) {
			return false;
		}
		return type == that.type;

	}

	public enum Type {
		InvalidConnection, Notice, Properties, TooMany, RemovedConnection, Unconnected
	}
}
