/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.issues;

import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;

import org.jetbrains.annotations.NotNull;

import java.util.Optional;

/**
 * Created by Ramon on 21.04.2016.
 */
public class ModuleIssue extends ConfigurationIssue {

	@NotNull
	private final IConfigurationElement element;
	private final IParameterHandler parameter;
	private final Type type;
	private final IAttributeHandler attribute;
	private final String oldValue;
	private final IToolchainConfiguration currentConfiguration;

	public ModuleIssue(@NotNull IConfigurationElement element, IParameterHandler parameter, String oldValue, Type type, String message, String situation, IssueType level) {
		super(message, situation, level);
		this.element = element;
		this.parameter = parameter;
		this.oldValue = oldValue;
		this.attribute = null;
		this.type = type;
		currentConfiguration = null;
	}

	public ModuleIssue(IConfigurationElement element, IAttributeHandler attribute, Type type, String message, String situation, IssueType level) {
		super(message, situation, level);
		this.element = element;
		this.attribute = attribute;
		this.parameter = null;
		oldValue = null;
		this.type = type;
		currentConfiguration = null;
	}

	public ModuleIssue(IModuleConfiguration module, IToolchainConfiguration currentConfiguration, Type type, String message, String situation, IssueType level) {
		super(message, situation, level);
		this.element = module;
		this.attribute = null;
		this.parameter = null;
		this.type = type;
		oldValue = null;
		this.currentConfiguration = currentConfiguration;
	}

	@Override
	public String getHintText() {
		StringBuilder sb = new StringBuilder();
		if (element != null) {
			sb.append("Module: ").append(element.getName());
			if (element.getType() != IConfigurationElement.ElementType.Configuration) {
				sb.append(", of type '").append(((IModuleConfiguration) element).getModuleDescriptionName()).append("' \n");
			}
		}
		if (attribute != null) {
			sb.append("Attribute: ").append(attribute).append('\n');
		}
		if (parameter != null) {
			sb.append("Parameter '").append(parameter.getCanonicalName()).append("', value = ").append(parameter.getValue());
			if (oldValue != null) {
				sb.append(", old value = ").append(oldValue);
			}
			addConstraintViolation(sb);
			sb.append('\n');
		}
		if (currentConfiguration != null) {
			sb.append("Target Config: ").append(currentConfiguration.getTargetName()).append(':').append(currentConfiguration.getDeviceName()).append(", toolchain: ").append(currentConfiguration.getToolchainImplementation());
			sb.append('\n');
		}
		return sb.toString();
	}

	@Override
	public Actuality getActuality() {
		if (type == Type.InvalidHandler || type == Type.InvalidParam) {
			return (parameter != null && parameter.isValid(false))? Actuality.Resolved : Actuality.Current;
		} else if (type == Type.NotUsed) {
			return element.isValid(false)? Actuality.Resolved : Actuality.Current;
		} else if (type == Type.ConflictedParam) {
			return Actuality.Unknown;
		} else {
			return Actuality.Persistent;
		}
	}

	@Override
	public Optional<IConfigurationElement<?, ?, ?>> getReferencedModule() {
		return Optional.of(element);
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}
		if (!super.equals(o)) {
			return false;
		}

		ModuleIssue that = (ModuleIssue) o;

		if (element != null? !element.equals(that.element) : that.element != null) {
			return false;
		}
		if (parameter != null? !parameter.equals(that.parameter) : that.parameter != null) {
			return false;
		}
		return type == that.type;

	}

	@Override
	public int hashCode() {
		int result = super.hashCode();
		result = 31 * result + (element != null? element.hashCode() : 0);
		result = 31 * result + (parameter != null? parameter.hashCode() : 0);
		result = 31 * result + type.hashCode();
		return result;
	}

	private void addConstraintViolation(StringBuilder sb) {
		if (parameter.hasChoices() || parameter.hasRange()) {
			sb.append(" does not match constraints ");
			if (parameter.hasChoices()) {
				sb.append("choices: ").append(parameter.getChoices()).append(' ');
			}
			if (parameter.hasRange()) {
				sb.append("range: ").append(parameter.getParameterDescription().getRange());
			}
		}
	}

	public enum Type {
		NotUsed, InvalidParam, InvalidHandler, NotSupported, ConflictedParam
	}
}
