/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.calc.CalculatedValue;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.AddressLayoutSpec;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryLocator;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IFirmwareConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;

import java.util.Map;
import java.util.Optional;

/**
 * Created by Ramon on 26.03.2016.
 */
public abstract class AddressBlock<M extends IModuleConfiguration<?>> implements IAddressChunk {

	protected final long length;
	protected final M source;
	protected IAddressRangeHandler rangeHandler;
	private final AddressLayoutSpec.BlockType blockType;
	private final String name;


	protected AddressBlock(String name, long length, M source, AddressRangeHandler rangeHandler, AddressLayoutSpec.BlockType blockType) {
		this.name = name;
		this.length = length;
		this.source = source;
		this.rangeHandler = rangeHandler;
		this.blockType = blockType;
	}

	public static RegAddressBlock of(PeripheralModuleDescription.RegisterInfo registers, IPeripheralModuleConfiguration<?> peri, AddressRangeHandler rangeHandler) {
		long size = registers.getRegisters().size();
		int alignExp = CalculatedValue.AddressSize.getAddressSize(size);
		return new RegAddressBlock(registers.getName(), size, alignExp, peri, registers, rangeHandler);
	}

	public static DmaAddressBlock of(PeripheralModuleDescription.DmaMemory memory, IPeripheralModuleConfiguration<?> module, AddressRangeHandler rangeHandler) {
		long size = memory.resolveSize(module).orElseThrow();
		int alignExp = CalculatedValue.AddressSize.getAddressSize(size);
		return new DmaAddressBlock(memory.getName(), size, alignExp, module, memory, rangeHandler);
	}

	public static MemAddressBlock of(MemoryModuleDescription.ExtMemory memory, Bus exposing, IMemoryModuleConfiguration<?> module, AddressRangeHandler rangeHandler) {
		long size = memory.resolveSize(module).orElseThrow();
		long start = memory.getExposedOn(exposing).resolveAt(module).orElseThrow();
		return new MemAddressBlock(memory.getName(), start, size, module, memory, rangeHandler);
	}

	public IModuleConfiguration getSourceModule() {
		return source;
	}

	@Override
	public long getLength() {
		return length;
	}

	public abstract boolean isValid();

	@Override
	public long getActualEnd() {
		return getActualStart() + getLength();
	}

	public abstract boolean isShared();

	public abstract boolean isInitializable();

	public abstract boolean isBitfileInitializable();

	@Override
	public void layoutSelf(AddressSpaceHandler analyzer) {
		//Nothing to to. Basic Blocks have a constant length
	}

	public abstract boolean hasAttachedFirmware();

	@Override
	public void setMinStart(long start) {
		setActualStart(start);
	}

	/**
	 * Return name of parameter with which to configurate the actual start (if it is configurable)
	 *
	 * @return String or null if {@link #hasFixedStart()} returns true
	 */
	public abstract String getConfigurationParameterName();

	public IAddressRangeHandler getRange() {
		return rangeHandler;
	}

	@Override
	public AddressLayoutSpec.BlockType getBlockType() {
		return blockType;
	}

	@Override
	public String getName() {
		return name;
	}


	public abstract static class AlignedAddressBlock<M extends IPeripheralModuleConfiguration<?>> extends AddressBlock<M> {

		protected final int alignExp;
		protected long start = Long.MIN_VALUE;

		protected AlignedAddressBlock(String name, long length, int alignExp, M source, AddressRangeHandler rangeHandler, AddressLayoutSpec.BlockType blockType) {
			super(name, length, source, rangeHandler, blockType);
			this.alignExp = Math.max(3, alignExp); // to save resources reg_access_decoder only supports ranges of 3 bits and more
		}

		@Override
		public boolean isValid() {
			return (start & ((1 << alignExp) - 1)) == 0;
		}

		@Override
		public int getAlignmentExp() {
			return alignExp;
		}

		@Override
		public int hashCode() {
			int result = (int) (start ^ (start >>> 32));
			result = 31 * result + (int) (length ^ (length >>> 32));
			result = 31 * result + alignExp;
			result = 31 * result + source.hashCode();
			return result;
		}

		@Override
		public void setActualStart(long start) {
			this.start = start;
		}

		@Override
		public long getActualStart() {
			return start;
		}

		@Override
		public boolean hasFixedStart() {
			return false;
		}


		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			AlignedAddressBlock that = (AlignedAddressBlock) o;

			if (start != that.start) {
				return false;
			}
			if (length != that.length) {
				return false;
			}
			if (alignExp != that.alignExp) {
				return false;
			}
			return source.equals(that.source);

		}

		@Override
		public String toString() {
			return "AddressBlock(" +
					source.getGlobalName() + "."+ getName() +
					", alignExp=" + alignExp +
					" @ 0x" + Long.toHexString(start) +
					", 0x" + Long.toHexString(length) + " lines)";
		}

		@Override
		public boolean isInitializable() {
			return false;
		}

		@Override
		public boolean hasAttachedFirmware() {
			return false;
		}
	}


	public static class RegAddressBlock extends AlignedAddressBlock<IPeripheralModuleConfiguration<?>> {

		private final PeripheralModuleDescription.RegisterInfo registers;

		protected RegAddressBlock(String name, long length, int alignExp, IPeripheralModuleConfiguration source, PeripheralModuleDescription.RegisterInfo registers, AddressRangeHandler rangeHandler) {
			super(name, length, alignExp, source, rangeHandler, AddressLayoutSpec.BlockType.Unintializable);
			this.registers = registers;
		}

		public PeripheralModuleDescription.RegisterInfo getRegisters() {
			return registers;
		}


		@Override
		public boolean isShared() {
			return false;
		}

		@Override
		public boolean isBitfileInitializable() {
			return false;
		}

		@Override
		public String getConfigurationParameterName() {
			return registers.getBaseAdrParam();
		}
	}


	public static class DmaAddressBlock extends AlignedAddressBlock<IPeripheralModuleConfiguration<?>> implements IMemoryAddressBlock {

		private final PeripheralModuleDescription.DmaMemory memory;

		protected DmaAddressBlock(String name, long length, int alignExp, IPeripheralModuleConfiguration source, PeripheralModuleDescription.DmaMemory memory, AddressRangeHandler rangeHandler) {
			super(name, length, alignExp, source, rangeHandler, AddressLayoutSpec.BlockType.Initializable);
			this.memory = memory;
		}

		@Override
		public PeripheralModuleDescription.DmaMemory getMemory() {
			return memory;
		}

		@Override
		public MemoryLocator resolveBRamInstance() {
			return getMemory().getInstance();
		}

		@Override
		public boolean isInitializable() {
			return true;
		}

		@Override
		public AddressLayoutSpec.BlockType getBlockType() {
			return AddressLayoutSpec.BlockType.Initializable;
		}

		@Override
		public String getConfigurationParameterName() {
			return memory.getBaseAdrParam();
		}

		@Override
		public boolean isShared() {
			return false;
		}

		@Override
		public boolean isBitfileInitializable() {
			return getMemory().isBitfileInitializable();
		}

	}


	public static class MemAddressBlock extends AddressBlock<IMemoryModuleConfiguration<?>> implements IMemoryAddressBlock {

		private final MemoryModuleDescription.ExtMemory memory;
		private final long start;
		/**
		 * This is mislabled, only the Block-instance is shared, the memory is explicitly not shared but exists multiple times
		 */
		private final Integer shareIndex;

		protected MemAddressBlock(String name, long start, long size, IMemoryModuleConfiguration<?> source, MemoryModuleDescription.ExtMemory memory, AddressRangeHandler rangeHandler) {
			super(name, size, source, rangeHandler, AddressLayoutSpec.BlockType.SelfInitializing);
			this.start = start;
			this.memory = memory;
			int memoryCount = memory.resolveCount(source).orElseThrow();
			if (memoryCount > 1) {
				shareIndex = assignShareIndex(memory, memoryCount);
			} else {
				shareIndex = null;
			}
		}

		private int assignShareIndex(MemoryModuleDescription.ExtMemory memory, int memoryCount) {
			Optional<Map<Integer, MemAddressBlock>> opShared = getRange().getGlobal().map(gl -> gl.getSharedMemoryIndexes(memory));
			if (!opShared.isPresent()) {
				return 0; // ignore,bc we are the only one in layout
			}

			Map<Integer, MemAddressBlock> sharedMemoryIndexes = opShared.get();

			for (int i = 0; i < memoryCount; i++) {
				if (!sharedMemoryIndexes.containsKey(i)) {
					sharedMemoryIndexes.put(i, this);
					return i;
				}
			}
			throw new IllegalStateException("No more memory instances to share!");
		}

		public IFirmwareConfiguration getFirmware() {
			return source.getFirmware();
		}

		@Override
		public MemoryLocator resolveBRamInstance() {
			MemoryLocator instance = memory.getInstance();
			if (shareIndex != null) {
				return instance.resolveMemoryIndex(shareIndex);
			}
			return instance;
		}


		public boolean hasFixedStart() {
			return true;
		}

		@Override
		public int getAlignmentExp() {
			return 0;
		}

		@Override
		public long getActualStart() {
			return start;
		}

		@Override
		public MemoryModuleDescription.ExtMemory getMemory() {
			return memory;
		}

		@Override
		public void setActualStart(long start) {
			throw new UnsupportedOperationException();
		}

		@Override
		public AddressLayoutSpec.BlockType getBlockType() {
			return AddressLayoutSpec.BlockType.SelfInitializing;
		}

		@Override
		public boolean isValid() {
			return true;
		}

		@Override
		public boolean isShared() {
			return memory.isShared();
		}

		@Override
		public boolean isBitfileInitializable() {
			return getMemory().isBitfileInitializable();
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			MemAddressBlock that = (MemAddressBlock) o;

			if (start != that.start) {
				return false;
			}
			if (!memory.equals(that.memory)) {
				return false;
			}
			if (length != that.length) {
				return false;
			}
			if (shareIndex != null) {
				return shareIndex.equals(that.shareIndex);
			}

			return source.equals(that.source);
		}

		@Override
		public int hashCode() {
			int result = memory.hashCode();
			result = 31 * result + (int) (start ^ (start >>> 32));
			result = 31 * result + (int) (length ^ (length >>> 32));
			result = 31 * result + (shareIndex != null? shareIndex + 1 : 0);// avoid hash collision between null and 0
			result = 31 * result + source.hashCode();
			return result;
		}

		@Override
		public String toString() {
			return "AddressBlock(" +
					source.getGlobalName() + "." + getName() +
					" @ 0x" + Long.toHexString(start) +
					", 0x" + Long.toHexString(length) + " lines)";
		}

		@Override
		public boolean isInitializable() {
			return true;
		}

		@Override
		public boolean hasAttachedFirmware() {
			return source.getModuleDescription().containsFirmware();
		}

		@Override
		public String getConfigurationParameterName() {
			return null;
		}
	}


}
