/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting;

import java.util.LinkedList;
import java.util.List;
import java.util.stream.Collectors;

import static de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.AddressSpaceHandler.logger;

/**
 * Created by Ramon on 26.03.2016.
 */
public abstract class AddressBlockContainer<B extends IAddressChunk> {

	protected final List<B> addressBlocks;
	protected long nextAddress;
	protected long actualStart;
	protected boolean manualChange;

	protected AddressBlockContainer() {
		actualStart = Long.MIN_VALUE;
		addressBlocks = new LinkedList<>();
		manualChange = false;
	}

	protected AddressBlockContainer(long nextAddress, long actualStart, List<B> addressBlocks, boolean manualChange) {
		this.nextAddress = nextAddress;
		this.actualStart = actualStart;
		this.addressBlocks = addressBlocks;
		this.manualChange = manualChange;
	}

	protected void layoutBlock(B block, AddressSpaceHandler analyzer) {
		if (block.hasFixedStart()) {
			if (block.getActualStart() < nextAddress && block.getLength() > 0) {
				throw createDoesNotFitException(block);
			}
		} else {
			long next = findNextPossibleAddress(block);
			block.setMinStart(next);
		}

		block.layoutSelf(analyzer);

		long blockActualStart = block.getActualStart();

		long length = block.getLength();

		if (length > 0) {
			nextAddress = blockActualStart + length;
			if (nextAddress > getMaxEnd()) {
				throw createDoesNotFitException(block);
			}

			if (addressBlocks.isEmpty()) {
				actualStart = blockActualStart;
			}

			addressBlocks.add(block);
		} else {
			logger.debug("Ignoring {} because its empty", block);
		}
	}

	protected boolean isValid(long containerStart, long containerEnd) {
		long address = actualStart;

		if (address < containerStart) {
			return false;
		}

		for (B block : addressBlocks) {
			if (block.getActualStart() < address || block.getLength() <= 0) {
				return false;
			}
			address = block.getActualEnd();
		}

		return address <= containerEnd;
	}

	protected abstract RuntimeException createDoesNotFitException(B block);

	protected abstract long findNextPossibleAddress(B block);

	protected abstract long getMaxEnd();

	/**
	 * Get end of address block
	 * Exclusive!
	 *
	 * @return native memory address
	 */
	public long getActualEnd() {
		return manualChange? addressBlocks.stream().reduce(getActualStart(), (s, block) -> Math.max(s, block.getActualEnd()), Math::max) : nextAddress; // Either find highest actualEnd or nextAddress if it is still valid
	}

	public long getActualStart() {
		return actualStart;
	}

	/**
	 * Inclusive
	 *
	 * @param start start address in native address
	 */
	public void setActualStart(long start) {
		manualChange = true;
		this.actualStart = start;
	}

	public long getLength() {
		return addressBlocks.isEmpty()? 0 : getActualEnd() - getActualStart();
	}

	public abstract long getMinStart();

	public void setMinStart(long start) {
		this.nextAddress = start;
	}

	public List<B> getAddressBlocks() {
		return addressBlocks;
	}

	@Override
	public String toString() {
		return "begin=0x" + Long.toHexString(getActualStart()) +
			", end=0x" + Long.toHexString(getActualEnd()) +
			", contents=" + addressBlocks.stream().map(Object::toString).collect(Collectors.joining("\n\t", "[", "]")) +
			'}';
	}
}
