/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.AddressLayoutSpec;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.SubsystemModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by Ramon on 18.04.2016.
 */
public final class AddressLayout {

	private static final Logger logger = LoggerFactory.getLogger(AddressLayout.class);

	private final ISubsystemConfiguration<?> subsystem;
	private final List<AddressSpaceHandler> spaceHandlers;
	private IGlobalAddressLayout global;

	/**
	 * AutoLayout a subsystem
	 *
	 * @param subsystem the subsystem
	 * @param global a global address layout
	 */
	private AddressLayout(@NotNull ISubsystemConfiguration subsystem, @Nullable IGlobalAddressLayout global) {
		this.subsystem = subsystem;
		this.global = global;
		SubsystemModuleDescription description = subsystem.getModuleDescription();
		AddressLayoutSpec addressLayoutSpec = description.getAddressLayoutSpec();
		spaceHandlers = addressLayoutSpec.getAddressSpaces().stream().map(space -> new AddressSpaceHandler(subsystem, space, global))
			.map(space -> {
				space.layoutSelf();
				return space;
			}).filter(space -> space.getLength() > 0).collect(Collectors.toList());

		addressLayoutSpec.getAutoMerges().forEach(info -> autoMerge(info, spaceHandlers));// merge / remove merged
	}

	public String getName() {
		return subsystem.getGlobalName();
	}

	/**
	 * Get the most current known layout.
	 * If the subsystem has static mappings set, return a layout based on those, otherwise return autoLayout
	 *
	 * @param subsystem subsystem which to layout
	 * @return new AddressLayout
	 */
	public static AddressLayout getCurrent(ISubsystemConfiguration subsystem) {
		return getCurrent(subsystem, null);
	}

	@NotNull
	static AddressLayout getCurrent(ISubsystemConfiguration subsystem, IGlobalAddressLayout global) {
		AddressLayout layout = new AddressLayout(subsystem, global);
		if (subsystem.hasStaticAddressSpaceMappings()) {
			layout.applyMappings(subsystem.getStaticAddressLayoutMapping());
		}
		return layout;
	}

	public void addDependencies(Set<Path> inputs, Map<String, Object> properties) {
		addAddressDependencies(properties);
		addSubsystemLayoutSpecDependencies(properties, subsystem);
		addAttachedMemoryModuleDependencies(inputs, properties, subsystem);
	}

	public void addAddressDependencies(Map<String, Object> properties) {
		properties.put("Layout_" + subsystem.getName(), extractStaticMappings());
		properties.put(subsystem.getName(), streamFirmwareAttachedInitializableBlocks()
				.map(AddressBlock::getSourceModule)
				.map(IModuleConfiguration::getGlobalName)
				.collect(Collectors.toList()));//TODO might be non-deterministic, but so was the old call that was broken by sebastian herber
	}

	public static void addSubsystemLayoutSpecDependencies(Map<String, Object> properties, ISubsystemConfiguration subsystem) {
		SubsystemModuleDescription sd = subsystem.getModuleDescription();
		properties.put("Spec_" + sd.getName(), sd.getAddressLayoutSpec());
	}

	public static void addAttachedMemoryModuleDependencies(Set<Path> inputs, Map<String, Object> properties, ISubsystemConfiguration<?> subsystem) {
		subsystem.streamMemoryModules().forEach(memory -> {
			inputs.add(memory.getModuleDescription().getSourceFile());
			properties.put("Firmware_" + memory.getGlobalName(), memory.getFirmware().getFirmwareName());
			int i = 0;
			for (MemoryModuleDescription.ExtMemory extMemory : memory.getModuleDescription().getMemories()) {
				properties.put(String.format("Memory_%s_%d", memory.getGlobalName(), i++), memory.getResolvedMemorySize(extMemory));
			}
		});
	}

	public Optional<AddressSpaceHandler> findSpaceHandler(String name) {
		return spaceHandlers.stream().filter(sh -> sh.getName().equals(name)).findAny();
	}

	public Optional<AddressBlock.RegAddressBlock> findRegAddressBlock(String range, IPeripheralModuleConfiguration peripheral) {
		return findAddressBlock(range, peripheral).map(block -> ((AddressBlock.RegAddressBlock) block));
	}

	/**
	 * Only works if every peripheral searched only exports a single RegisterSet
	 *
	 * @param peripheral peripheral to find
	 * @return addressBock if present
	 */
	public Optional<AddressBlock.RegAddressBlock> findRegAddressBlock(IPeripheralModuleConfiguration peripheral) {
		return streamUnInitializableBlocks().filter(block -> block.getSourceModule() == peripheral).findAny();
	}

	/**
	 * TODO Only works if every peripheral searched only exports a single DmaMemory
	 *
	 * @param peripheral peripheral to find
	 * @return addressBlock if present
	 */
	public Optional<AddressBlock.DmaAddressBlock> findDmaAddressBlock(IPeripheralModuleConfiguration peripheral) {
		return streamDetachedInitializableBlocks().filter(block -> block.getSourceModule() == peripheral).findAny().map(b -> (AddressBlock.DmaAddressBlock) b);
	}

	public Optional<AddressBlock.DmaAddressBlock> findDmaAddressBlock(String range, IPeripheralModuleConfiguration peripheral) {
		return findAddressBlock(range, peripheral).map(block -> ((AddressBlock.DmaAddressBlock) block));
	}

	/**
	 * Find the AddressRange in which this peripheral is listed, due to connection via exposedOn.
	 *
	 * @param exposedOn the bus over which peripheral is supposedly connected
	 * @param peripheral the peripheral which to find
	 * @return Optional, present if found, empty otherwise
	 */
	public Optional<IAddressRangeHandler> findAddressRange(Bus exposedOn, IPeripheralModuleConfiguration peripheral) {
		IBusHandler busHandler = peripheral.getBusHandler(exposedOn);
		return spaceHandlers.stream().flatMap(sh -> StreamUtil.streamOptional(sh.findRangeHandler(busHandler))).findAny();
	}

	public Optional<IAddressRangeHandler> findAddressRange(String range) {
		return spaceHandlers.stream().flatMap(sh -> StreamUtil.streamOptional(sh.findRangeHandler(range))).findAny();
	}



	public Optional<AddressBlock<?>> findAddressBlock(String range, IPeripheralModuleConfiguration peripheral) {
		return findAddressRange(range).flatMap(rh -> rh.findBlock(peripheral.getName()));
	}

	public Optional<AddressBlock<?>> findAddressBlock(Bus exposedOn, IPeripheralModuleConfiguration peripheral) {
		return findAddressRange(exposedOn, peripheral).flatMap(r -> r.findBlock(peripheral));
	}

	public AddressLayoutMapping extractStaticMappings() {
		return new AddressLayoutMapping(spaceHandlers.stream().map(space -> new AddressLayoutMapping.AddressSpaceMapping(space.getName(),
			space.getAddressBlocks().stream().filter(range -> !range.hasFixedStart()).map(range -> new AddressLayoutMapping.AddressRangeMapping(range.getName(), range.getActualStart(),
				range.getAddressBlocks().stream().filter(block -> !block.hasFixedStart()).map(block -> new AddressLayoutMapping.AddressBlockMapping(block.getSourceModule().getName(), block.getActualStart())
				).collect(Collectors.toList()))
			).collect(Collectors.toList()))).collect(Collectors.toList()));
	}

	public boolean isValid() {
		return spaceHandlers.stream().allMatch(AddressSpaceHandler::isValid);
	}

	public List<AddressSpaceHandler> getSpaceHandlers() {
		return Collections.unmodifiableList(spaceHandlers);
	}

	/**
	 * All mappings that still match are applied, rest is discarded
	 *
	 * @param layoutMapping list of mappings to apply
	 */
	public void applyMappings(AddressLayoutMapping layoutMapping) {
		layoutMapping.getSpaceMappings().stream().forEach(mm -> {
			Optional<AddressSpaceHandler> spaceHandler = findSpaceHandler(mm.getSpace());
			if (spaceHandler.isPresent()) {
				mm.getRangeMappings().forEach(rm -> {
					Optional<IAddressRangeHandler> optRangeHandler = spaceHandler.get().findRangeHandler(rm.getRange());
					if (optRangeHandler.isPresent()) {
						IAddressRangeHandler rangeHandler = optRangeHandler.get();
						rangeHandler.setActualStart(rm.getStart());
						rm.getBlockMappings().forEach(bm -> {
							Optional<? extends AddressBlock<?>> block = rangeHandler.findBlock(bm.getBlock());
							if (block.isPresent()) {
								block.get().setActualStart(bm.getStart());
							}
						});
					}
				});
			}
		});
	}

	public Stream<AddressBlock.RegAddressBlock> streamUnInitializableBlocks() {
		return streamAllBlocks().filter(b -> !b.hasAttachedFirmware() && !b.isInitializable()).map(b -> (AddressBlock.RegAddressBlock) b);
	}

	public Stream<AddressBlock.MemAddressBlock> streamFirmwareAttachedInitializableBlocks() {
		return streamAllBlocks().filter(b -> b.hasAttachedFirmware() && b.isInitializable()).map(b -> (AddressBlock.MemAddressBlock) b);
	}

	public Stream<AddressBlock<?>> streamAllBlocks() {
		return spaceHandlers.stream().flatMap(sh -> sh.getAddressBlocks().stream()).flatMap(rh -> rh.getAddressBlocks().stream());
	}

	public Stream<IMemoryAddressBlock> streamDetachedInitializableBlocks() {
		return streamAllBlocks().filter(b -> !b.hasAttachedFirmware() && b.isInitializable()).map(b -> (IMemoryAddressBlock) b);
	}

	private void autoMerge(AddressLayoutSpec.AutoMergeInfo info, List<AddressSpaceHandler> spaceHandlers) {
		Optional<AddressSpaceHandler> sourceSpace = findSpaceHandler(info.getSpace());
		Optional<IAddressRangeHandler> source = sourceSpace.flatMap(sp -> sp.findRangeHandler(info.getRange()));
		Optional<AddressSpaceHandler> targetSpace = findSpaceHandler(info.getIntoSpace());
		Optional<IAddressRangeHandler> target = targetSpace.flatMap(ts -> ts.findRangeHandler(info.getIntoRange()));


		if (source.isPresent() && target.isPresent()) {
			TreeMap<Long, AddressBlock<?>> targetBlocks = new TreeMap<>();
			target.get().getAddressBlocks().forEach(block -> {
				targetBlocks.put(block.getActualStart(), block);
			});

			List<AddressBlock<?>> merged = new ArrayList<>();

			for (AddressBlock block : source.get().getAddressBlocks()) {
				AddressBlock existing = targetBlocks.get(block.getActualStart());


				if (existing != null && existing.equals(block)) {
					merged.add(block);
				}
			}



			//TODO: Implicit limitation to merged blocks first, then split ranges, needs to be extended if they can arbitrarily occur !!
			int targetIdx = targetSpace.get().addressBlocks.indexOf(target.get());
			List<IAddressRangeHandler> sourceRanges = sourceSpace.get().addressBlocks;
			boolean haveMerged = merged.size() > 0;
			if (haveMerged) {
				MergedRangeHandler mergedRange = new MergedRangeHandler(Arrays.asList(target.get().getDescription(), source.get().getDescription()), merged);
				targetSpace.get().addressBlocks.set(targetIdx, mergedRange);
			}
			List<AddressBlock<?>> notMergedTarget = targetBlocks.values().stream().filter(block -> !merged.contains(block)).collect(Collectors.toList());
			boolean haveNotMerged = notMergedTarget.size() != 0;
			boolean haveMergedAndNotMerged = haveNotMerged && haveMerged;

			if (haveNotMerged) {
				AddressRangeHandler notMergedRange = AddressRangeHandler.create(((AddressRangeHandler) target.get()), notMergedTarget);
				if (haveMergedAndNotMerged) {
					targetSpace.get().addressBlocks.add(targetIdx + 1, notMergedRange);
				} else {
					targetSpace.get().addressBlocks.set(targetIdx, notMergedRange);
				}
			}

			List<AddressBlock<?>> notMergedSource = source.get().getAddressBlocks().stream().filter(block -> !merged.contains(block)).collect(Collectors.toList());
			if (notMergedSource.size() > 0) {
				int sourceIdx = sourceRanges.indexOf(source.get());
				sourceRanges.set(sourceIdx, AddressRangeHandler.create(((AddressRangeHandler) source.get()), notMergedSource));
			} else {
				sourceRanges.remove(source.get());
				if (sourceRanges.isEmpty()) {
					spaceHandlers.remove(sourceSpace.get());
				}
			}
		}
	}

	public ISubsystemConfiguration getSubsystem() {
		return subsystem;
	}

	/**
	 * Stream all blocks that have a configurable start (regs, dma)
	 *
	 * @return Stream
	 */
	public Stream<AddressBlock<?>> streamLayoutedBlocks() {
		return streamAllBlocks().filter(ab -> !ab.hasFixedStart());
	}

	public Stream<IMemoryAddressBlock> streamInitializableBlocks() {
		return streamAllBlocks().filter(AddressBlock::isInitializable).map(b -> ((IMemoryAddressBlock) b));
	}

	public boolean hasMemory() {
		return streamFirmwareAttachedInitializableBlocks().findAny().isPresent();
	}
}
