/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.AddressLayoutSpec;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IBusBreakoutModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * Created by Ramon on 26.03.2016.
 */
public class AddressRangeHandler extends AddressBlockContainer<AddressBlock<?>> implements IAddressRangeHandler {

	protected final AddressLayoutSpec.AddressRange range;
	protected final IBusHandler busHandler;
	protected final IGlobalAddressLayout global;

	private AddressRangeHandler(AddressLayoutSpec.AddressRange range, IBusHandler busHandler, IGlobalAddressLayout global) {
		this.range = range;
		this.busHandler = busHandler;
		this.global = global;
	}

	protected AddressRangeHandler(long nextAddress, long actualStart, List<AddressBlock<?>> addressBlocks, boolean manualChange, AddressLayoutSpec.AddressRange range, IBusHandler busHandler, IGlobalAddressLayout global) {
		super(nextAddress, actualStart, addressBlocks, manualChange);
		this.range = range;
		this.busHandler = busHandler;
		this.global = global;
	}

	static AddressRangeHandler create(AddressLayoutSpec.AddressRange range, IModuleConfiguration module, IGlobalAddressLayout layout) {
		return new AddressRangeHandler(range, module.getBusHandler(range.getImplementedBy()), layout);
	}

	/**
	 * Copy old range
	 * Must have non-empty list of blocks!
	 * @param old
	 * @param newBlocks
	 * @return
	 */
	static AddressRangeHandler create(AddressRangeHandler old, List<AddressBlock<?>> newBlocks) {
		long firstAddr = newBlocks.get(0).getActualStart();
		long nextAddr = newBlocks.get(newBlocks.size() - 1).getActualEnd() + 1;
		AddressRangeHandler newRange = new AddressRangeHandler(nextAddr, firstAddr, newBlocks, true, old.range, old.busHandler, old.global);
		return newRange;
	}

	@Override
	public Optional<AddressBlock<?>> findBlock(String name) {
		return addressBlocks.stream().filter(b -> b.getSourceModule().getName().equals(name)).findAny();
	}

	@Override
	public Optional<AddressBlock<?>> findBlock(IPeripheralModuleConfiguration peripheral) {
		return findBlock(peripheral.getName());
	}

	@Override
	public String getName() {
		return range.getName();
	}

	@Override
	public AddressLayoutSpec.AddressRange getDescription() {
		return range;
	}

	@Override
	public boolean isWritable() {
		return range.isWritable();
	}

	@Override
	public boolean isExecutable() {
		return range.isExecutable();
	}

	@Override
	public boolean isReadable() {
		return range.isReadable();
	}

	@Override
	public boolean isValid() {
		return isValid(range.getStart(), range.getEnd());
	}

	@Override
	protected RuntimeException createDoesNotFitException(AddressBlock block) {
		return new IllegalStateException(block + " does not fit into " + range + " on " + busHandler.getConnectorDescription() + "\nContents: " + addressBlocks);
	}

	@Override
	public long findNextPossibleAddress(AddressBlock block) {
		int align = block.getAlignmentExp();
		long alignMask = -1L << align; //Ex:	0b11111000
		long alignVal = 1L << align; //Ex:		0b00001000
		long alignedFloor = nextAddress & alignMask; // mask away bits below the alignment
		return (alignedFloor == nextAddress)? nextAddress : alignedFloor + alignVal; // add alignment if needed. Ex: 0b10100100 => 0b10100000 => 0b10101000
	}

	@Override
	public long getMaxEnd() {
		return range.getEnd();
	}

	@Override
	public long getMinStart() {
		return range.getStart();
	}

	@Override
	public String toString() {
		return "AddressRangeHandler{" +
				"range=" + range.getName() +
				", on=" + busHandler.getPortName() +
				", cap=" + getSpecFlags() +
				", " + super.toString();
	}

	@Override
	public int getAlignmentExp() {
		return 0;
	}

	@Override
	public void layoutSelf(AddressSpaceHandler rootHandler) {
		final Comparator<AddressBlock<?>> blockComparator =
				Comparator.<AddressBlock<?>, Boolean>comparing(AddressBlock::hasFixedStart)
						.thenComparing(AddressBlock::getActualStart)//TODO split, aligned blokcs do not have getActualStart yet, only fixed blocks do!!
						.thenComparing(Comparator.<AddressBlock<?>, Integer>comparing(AddressBlock::getAlignmentExp).reversed());
		busHandler.streamOppositeConnectors().flatMap(oppositeBus -> {
			IModuleConfiguration<?> element = (IModuleConfiguration<?>) oppositeBus.getSourceElement();
			return getAddressBlocksExposed(element, oppositeBus.getConnectorDescription()).stream();
		}).filter(Objects::nonNull).sorted(blockComparator).forEach(block -> layoutBlock(block, rootHandler));
	}

	@Override
	public AddressLayoutSpec.BlockType getBlockType() {
		return range.getType();
	}

	@Override
	public boolean hasFixedStart() {
		return false;
	}


	protected List<AddressBlock<?>> getAddressBlocksExposed(IModuleConfiguration<?> module, Bus exposing) {
		List<AddressBlock<?>> exposedAddressBlocks = new ArrayList<>();
		if (module instanceof IBusBreakoutModuleConfiguration) {
			return exposedAddressBlocks;
		}
		AddressLayoutSpec.BlockType type = range.getType();
		if (type == AddressLayoutSpec.BlockType.Unintializable) {
			IPeripheralModuleConfiguration<?> peri = (IPeripheralModuleConfiguration<?>) module;
			Optional<PeripheralModuleDescription.RegisterInfo> expRegs = peri.getRegistersExposedOn(exposing);
			expRegs.ifPresent(registerInfo -> exposedAddressBlocks.add(AddressBlock.of(registerInfo, peri, this)));
		} else if (type == AddressLayoutSpec.BlockType.Initializable) {
			IPeripheralModuleConfiguration<?> peri = (IPeripheralModuleConfiguration<?>) module;
			Optional<PeripheralModuleDescription.DmaMemory> expMem = peri.getMemoryExposedOn(exposing);
			expMem.ifPresent(dmaMemory -> exposedAddressBlocks.add(AddressBlock.of(dmaMemory, peri, this)));
		} else if (type == AddressLayoutSpec.BlockType.SelfInitializing) {
			final IBusHandler realHandler = DocumentUtil.findOriginalBusHandlerForPassthrough(module.getBusHandler(exposing), IConfigurationElement.ElementType.MemoryModule);
			IMemoryModuleConfiguration<?> mem = (IMemoryModuleConfiguration<?>) realHandler.getSourceElement();
			List<MemoryModuleDescription.ExtMemory> expMem = mem.getMemoriesExposedOn(realHandler.getConnectorDescription());
			exposedAddressBlocks.addAll(expMem.stream().map(eMem -> AddressBlock.of(eMem, realHandler.getConnectorDescription(), mem, this)).collect(Collectors.toList()));
		}
		return exposedAddressBlocks;
	}

	public int getRelevantBits() {
		return 64 - Long.numberOfLeadingZeros(getMaxEnd() - getMinStart() - 1); //ceil(log2(rangeSize))
	}

	@Override
	public Optional<? extends IGlobalAddressLayout> getGlobal() {
		return Optional.ofNullable(global);
	}
}
