/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.mappedBuses.BaseMappedBusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.mappedBuses.MergedBusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.mappedBuses.SingleMappedBusPort;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement.KEY_COMPARATOR;

/**
 * Created by Ramon on 06.04.2016.
 */
public class BusLayout {

	private static final Logger logger = LoggerFactory.getLogger(BusLayout.class);

	//Lazily populated
	private final Map<IBusHandler, Map<BusPort, List<IMappedBusPortInstance>>> reverseLookup = new HashMap<>();
	private final Map<IConfigurationElement<?, ?, ?>, List<IMappedBusPortInstance>> busPortInstances;
	private final Library library;
	private final Map<String, List<BaseMappedBusPort>> mappedBusPorts = new HashMap<>();
	private final Map<IMappedBusPortInstance, List<MappedBusConnectionStub>> busConnectionStubs = new HashMap<>();
	private final Map<MappedBusConnectionStub, MappedBusConnection> stub2conn = new HashMap<>();
	private final Map<IMappedBusPortInstance, List<MappedBusConnection>> busConnections = new HashMap<>();

	/**
	 * Create a Layout of all buses currently on the document.
	 * Time consuming operation.
	 *
	 * @param document document which to layout
	 */
	public BusLayout(IConfigurationDocument document) {
		library = document.getLibrary();
		busPortInstances = document.streamModulesRecursive().collect(Collectors.toMap(Function.identity(), this::createBusPortInstances));
		streamMappedPorts().forEach(this::getConnectionStubs);
		logger.trace("Buses Layouted!");
	}

	public List<IMappedBusPortInstance> getOppositeMappedBusPorts(BaseMappedBusPortInstance busPortInstance, IBusHandler oppositeHandler) {
		Bus myBus = busPortInstance.getHandler().getConnectorDescription();
		BusDescription busDefinition = oppositeHandler.getBusDefinition();
		BusPort myBusPort = busPortInstance.getPort();

		BusPort oppositePort = busDefinition.getOppositePort(myBus, myBusPort);

		Map<BusPort, List<IMappedBusPortInstance>> busPorts = reverseLookup.get(oppositeHandler);
		if (busPorts == null) {
			reverseMapModule(oppositeHandler.getSourceElement());
			busPorts = reverseLookup.get(oppositeHandler);
		}
		return busPorts.get(oppositePort);
	}

	private void reverseMapModule(IConfigurationElement<?, ?, ?> sourceElement) {
		List<IMappedBusPortInstance> mappedBusPorts = this.busPortInstances.get(sourceElement);

		Map<IBusHandler, Map<BusPort, List<IMappedBusPortInstance>>> reverse = mappedBusPorts.stream().flatMap(IMappedBusPortInstance::streamSingleMappedBusPorts)
			.collect(Collectors.groupingBy(BaseMappedBusPortInstance::getHandler,
				Collectors.groupingBy(BaseMappedBusPortInstance::getPort, Collectors.mapping(BaseMappedBusPortInstance::getRootBusPortInstance, Collectors.toList()))));
		reverseLookup.putAll(reverse);
	}

	private List<IMappedBusPortInstance> createBusPortInstances(IModuleConfiguration module) {
		List<BaseMappedBusPort> mappedBusPorts = this.mappedBusPorts.computeIfAbsent(module.getModuleDescriptionName(), desc -> BaseMappedBusPort.mapAllBuses(module.getModuleDescription(), library));
		return getBusPortInstances(module, mappedBusPorts);
	}

	public List<IMappedBusPortInstance> getBusPortInstances(IModuleConfiguration module) {
		return busPortInstances.get(module);
	}

	private List<IMappedBusPortInstance> getBusPortInstances(IModuleConfiguration module, List<BaseMappedBusPort> mappedBusPorts) {
		Converter converter = new Converter(module);
		return mappedBusPorts.stream().map(mbp -> mbp.apply(converter)).collect(Collectors.toList());
	}

	public List<MappedBusConnectionStub> getConnectionStubs(IMappedBusPortInstance busPortInstance) {
		return busConnectionStubs.computeIfAbsent(busPortInstance, this::generateConnectionsStubs);
	}

	private List<MappedBusConnectionStub> generateConnectionsStubs(IMappedBusPortInstance busPortInstance) {
		return streamMappedPorts().filter(p -> p == busPortInstance).flatMap(p -> {
			List<IMappedBusPortInstance> ports = new ArrayList<>();
			if (p.isPerConnectionWidth()) {
				// create virtual instances of ports if bus connections are parallel on this port
				for (int i = 0; i < p.getBusConnectionCount(busPortInstance.getSourceElement()); i++) {
					ports.add(p);
				}
			} else {
				ports.add(p);
			}
			return ports.stream();
		}).flatMap(this::getOppositePorts).flatMap(p -> {
			List<IMappedBusPortInstance> ports = new ArrayList<>();
			if (p.isPerConnectionWidth()) {
				// create virtual instances of ports if bus connections are parallel on opposite port
				for (int i = 0; i < p.getBusConnectionCount(getOppositePorts(busPortInstance).findAny().get().getSourceElement()); i++) {
					ports.add(p);
				}
			} else {
				ports.add(p);
			}
			return ports.stream();
		}).collect(new BusPortPlacer(busPortInstance));
	}

	public List<MappedBusConnection> getConnections(IMappedBusPortInstance busPortInstance) {
		return busConnections.computeIfAbsent(busPortInstance, this::generateConnections);
	}

	private List<MappedBusConnection> generateConnections(IMappedBusPortInstance busPortInstance) {
		List<MappedBusConnectionStub> connectionStubs = getConnectionStubs(busPortInstance);
		return Collections.unmodifiableList(connectionStubs.stream().map(stub -> getConnection(busPortInstance, stub)).collect(Collectors.toList()));
	}

	private MappedBusConnection getConnection(IMappedBusPortInstance busPortInstance, MappedBusConnectionStub stub) {
		MappedBusConnection connection = stub2conn.get(stub);
		if (connection == null) {
			connection = generateConnection(busPortInstance, stub);
		}

		return connection;
	}

	private MappedBusConnection generateConnection(IMappedBusPortInstance busPortInstance, MappedBusConnectionStub stub) {
		IMappedBusPortInstance targetPort = stub.getTargetPort();
		int myStart;
		int myEnd;
		List<MappedBusConnectionStub> myConnectionStubs = getConnectionStubs(busPortInstance);
		List<MappedBusConnectionStub> myStubs = getConnectionStubs(targetPort).stream().filter(rs -> rs.getTargetPort() == busPortInstance).collect(Collectors.toList());
		int myConnectionPos = myConnectionStubs.indexOf(stub); // TODO: improve position detection on mixed buses (parallel and normal connection)
		if ((busPortInstance.getBusConnectionCount(busPortInstance.getSourceElement()) == 1 || !busPortInstance.isPerConnectionWidth())
			&& (targetPort.getBusConnectionCount(busPortInstance.getSourceElement()) == 1 || !busPortInstance.isPerConnectionWidth())
			&& myStubs.size() <= myConnectionPos) {
			// not a parallel connected bus, just take first opposite port
			myConnectionPos = 0;
		}

		MappedBusConnectionStub myStub = myStubs.get(myConnectionPos);
		myStart = myStub.getStart();
		myEnd = myStub.getEnd();
		// Stub targets opposite port, but gives its location on our own port.
		// -> Stub.isPartial() checks whether
		MappedBusConnection conn = new MappedBusConnection(busPortInstance, stub.getStart(), stub.getEnd(), targetPort, myStart, myEnd, myStub.isIgnore() | stub.isIgnore(), stub.isPartial(busPortInstance), myStub.isPartial(targetPort));
		stub2conn.put(stub, conn);
		stub2conn.put(myStub, conn);
		return conn;
	}

	/**
	 * Stream all MappedBusPorts of all Modules that are known to this cache.
	 *
	 * @return all known mappedBusPortInstances
	 */
	public Stream<IMappedBusPortInstance> streamMappedPorts() {
		return busPortInstances.values().stream().flatMap(List::stream);
	}

	public Stream<IMappedBusPortInstance> streamMappedPortsSorted() {
		return busPortInstances.entrySet().stream().sorted(KEY_COMPARATOR).map(Map.Entry::getValue).flatMap(List::stream);
	}

	public Stream<IMappedBusPortInstance> streamOf(IModuleConfiguration module) {
		return busPortInstances.get(module).stream();
	}

	public Stream<IMappedBusPortInstance> getOppositePorts(IMappedBusPortInstance busPortInstance) {
		return busPortInstance.streamSingleMappedBusPorts().flatMap(single -> single.getHandler().streamOppositeConnectors().flatMap(oppositeHandler -> getOppositeMappedBusPorts(single, oppositeHandler).stream())).distinct();
	}


	private class Converter implements BaseMappedBusPort.IConverter<IMappedBusPortInstance> {

		private final IModuleConfiguration module;

		public Converter(IModuleConfiguration module) {
			this.module = module;
		}

		@Override
		public SingleMappedBusPortInstance convert(SingleMappedBusPort mappedBusPort) {
			return new SingleMappedBusPortInstance(mappedBusPort, module.getBusHandler(mappedBusPort.getBus()), BusLayout.this);
		}

		@Override
		public MergedMappedBusPortInstance convert(MergedBusPort mergedBusPort, List<SingleMappedBusPort> ports) {
			return new MergedMappedBusPortInstance(mergedBusPort, module, ports, BusLayout.this);
		}
	}

}
