/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.util;

import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorGroupDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IModuleContainer;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IToolchainAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.IIssueSink;
import de.tu_darmstadt.rs.util.StreamUtil;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Convenience methods for accessing a document and its modules
 */
public final class DocumentUtil {

	private static final String PORT = "Port '";
	private static final String DOES_NOT_EXIST = "' does not exist";

	private DocumentUtil() {
	}

	public static IPortHandler<? extends IPort> getPortHandler(IModuleConfiguration<?> module, String portName) {
		return module.getPortHandler(module.findPortDescription(portName).orElseThrow(() -> new IllegalArgumentException(PORT + portName + DOES_NOT_EXIST)));
	}

	public static IPinHandler getPinHandler(IConfigurationDocument document, String pinName) {
		return document.getPortHandler(document.getTarget().getTargetDescription().findPinDescription(pinName).orElseThrow(() -> new IllegalArgumentException(PORT + pinName + DOES_NOT_EXIST)));
	}

	public static IBusHandler getBusHandler(IModuleConfiguration module, String busName) {
		return module.getBusHandler(module.getModuleDescription().findBus(busName).orElseThrow(() -> new IllegalArgumentException("Bus '" + busName + DOES_NOT_EXIST)));
	}

	public static <P extends IParameterHandler> P getParameterHandler(IModuleConfiguration<?> module, String paramName, Class<P> type) {
		return module.getParameterHandler(type, module.findParameterDescription(paramName).orElseThrow(() -> new IllegalArgumentException("Parameter '" + paramName + DOES_NOT_EXIST)));
	}

	public static IParameterHandler getParameterHandler(IModuleConfiguration module, String paramName) {
		return getParameterHandler(module, paramName, IParameterHandler.class);
	}

	public static <PD extends IConnectorDescription, PG extends IConnectorGroupDescription<PD>, PH extends IConnectorHandler<?, ?>> Stream<PH> streamConnectorHandlers(IConfigurationElement<PD, PG, PH> element) {
		return element.getPortGroupDescriptions().stream().flatMap(g -> g.getPorts().stream()).map(element::getPortHandler);
	}

	public static Stream<IParameterHandler> streamAllParameters(IModuleConfiguration<?> module) {
		return module.getParameterGroupDescriptions().stream().flatMap(pg -> pg.getParameters().stream()).map(pd -> module.getParameterHandler(IParameterHandler.class, pd));
	}

	public static Stream<IParameterHandler> streamAllParameters(Collection<? extends IModuleConfiguration> modules) {
		return modules.stream().flatMap(DocumentUtil::streamAllParameters);
	}

	public static Map<String, String> collectParamProps(Stream<IParameterHandler> params) {
		return params.collect(Collectors.toMap(param -> "PARAM_" + param.getCanonicalName(), (t) -> t.getValue().toString()));
	}

	public static Stream<IToolchainAttributeHandler> streamTargetParamHandlers(IConfigurationDocument document) {
		IConfigurationDocument.ITargetConfiguration target = document.getTarget();
		return target.getToolchain().getParams().values().stream().map(target::getToolchainParamHandler);
	}

	public static IToolchainAttributeHandler getToolchainParamHandler(IConfigurationDocument.ITargetConfiguration target, String name) {

		final DeviceDescription.DevParam param = target.getToolchain().getParams().get(name);
		return target.getToolchainParamHandler(param);
	}

	public static List<IConfigurationElement<?, ?, ?>> getInstanceChain(IModuleConfiguration module) {
		IConfigurationDocument document = module.getDocument();
		LinkedList<IConfigurationElement<?, ?, ?>> chain = new LinkedList<>();
		IConfigurationElement<?, ?, ?> current = module;
		while (current != document) {
			chain.addFirst(current);
			current = current.getParent();
		}
		return chain;
	}

	public static Stream<IMemoryModuleConfiguration> streamMemoryModules(IConfigurationDocument document) {
		return document.streamModulesRecursive().filter(m -> m.getType() == IConfigurationElement.ElementType.MemoryModule).map(m -> ((IMemoryModuleConfiguration) m));
	}

	public static Stream<IModuleDescription> streamUsedDescriptions(IConfigurationDocument document) {
		return document.streamModulesRecursive().map(IModuleConfiguration::getModuleDescription).distinct();
	}

	public static Stream<IModuleDescription> streamUsedDescriptions(ISubsystemConfiguration<?> subsystem) {
		return subsystem.streamAllModules().map(IModuleConfiguration::getModuleDescription).distinct();
	}

	public static Stream<IBusHandler> streamBusHandlers(IModuleConfiguration<?> module) {
		return module.getBusDescriptions().stream().map(module::getBusHandler);
	}

	public static Set<IModuleConfiguration<?>> findModulesByName(IModuleContainer<?> container, String moduleName) {
		return container.streamModulesRecursive().filter(m -> m.getName().equals(moduleName)).collect(Collectors.toSet());
	}

	public static Set<IModuleConfiguration<?>> findModulesByType(IModuleContainer<?> container, String typeName) {
		return container.streamModulesRecursive().filter(m -> m.getModuleDescriptionName().equals(typeName)).collect(Collectors.toSet());
	}

	public static boolean checkModuleSupportsToolchain(IModuleConfiguration<?> module, IIssueSink issueSink, IToolchainConfiguration currentConfiguration) {
		boolean supports = module.getModuleDescription().supports(currentConfiguration);
		if (!supports) {
			issueSink.report(ConfigurationIssue.createModuleNotSupportedIssue(module, currentConfiguration, "Module Validation"));
		}
		return supports;
	}


	/**
	 * For modules that pass busses through, find the other busHandler that this connection is passed through to
	 *
	 * @param busHandler a bus handler
	 * @return the bus handler that is internally connected
	 */
	public static Optional<IBusHandler> getPassthroughBus(IBusHandler busHandler) {
		final IModuleConfiguration<?> module = (IModuleConfiguration<?>) busHandler.getSourceElement();
		Optional<Bus> passedThrough = module.getBusDescriptions().stream()
				.filter(bus -> bus.getBusDescriptionName().equals(busHandler.getBusType()) && bus != busHandler.getConnectorDescription())
				.filter(bus -> bus.getRole() != busHandler.getConnectorDescription().getRole())
				.collect(StreamUtil.atMostOne());
		return passedThrough.map(module::getBusHandler);
	}

	/**
	 * Find a connected module (optionally via passthroughs) that has the given type
	 *
	 * @param busHandler the bus handler
	 * @param expectedModuleType the expected type
	 * @return the real bushandler
	 */
	public static IBusHandler findOriginalBusHandlerForPassthrough(IBusHandler busHandler, IConfigurationElement.ElementType expectedModuleType) {
		if (busHandler.getSourceElement().getType() == expectedModuleType) {
			return busHandler;
		} else {
			final Optional<IBusHandler> passthroughBus = getPassthroughBus(busHandler);
			if (!passthroughBus.isPresent()) {
				return busHandler;
			}
			if (passthroughBus.get().getConnectionCount() != 1) {
				throw new RuntimeException("Passed through bus " + passthroughBus + " does not have exactly one connection.");
			}
			return findOriginalBusHandlerForPassthrough(passthroughBus.get().streamOppositeConnectors().iterator().next(), expectedModuleType);
		}
	}

	/**
	 * Return a function that finds a connected module (optionally via passthroughs) that has the given type
	 *
	 * @param expectedModuleType the expected type
	 * @return the real bushandler
	 */
	public static Function<IBusHandler, IBusHandler> findOriginalBusHandlerForPassthroughMapper(IConfigurationElement.ElementType expectedModuleType) {
		return bh -> findOriginalBusHandlerForPassthrough(bh, expectedModuleType);
	}
}
