/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.microStreamsImport;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPluginEngine;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IScriptManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.exceptions.ConnectionMismatchException;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IModuleContainer;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBooleanParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IFirmwareAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IIntegerParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentImporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.microStreamsImport.jaxb.Core;
import de.tu_darmstadt.rs.spartanmc.jconfig.microStreamsImport.jaxb.Hardware;
import de.tu_darmstadt.rs.spartanmc.jconfig.microStreamsImport.jaxb.Peripheral;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibraryUtil;
import de.tu_darmstadt.rs.util.LevenshteinDistance;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import javax.xml.bind.JAXBException;

import static java.util.stream.Collectors.groupingBy;

/**
 * The importer is able to create a jconfig document from the hardware.xml created by µStreams
 */
public class MicroStreamsHardwareImporter implements IDocumentImporter {

	private static final String SPARTANMC_PERI_BUS = "spartanmc-peri-bus";
	private static final Logger logger = LoggerFactory.getLogger(MicroStreamsHardwareImporter.class);
	private static MicroStreamsHardwareJaxB microStreamsJaxB;
	private final String targetDescriptionName;
	private final String deviceDescriptionName;
	private boolean scriptManagerPreviousState;

	public MicroStreamsHardwareImporter(String targetDescriptionName, String deviceDescriptionName) {
		this.targetDescriptionName = targetDescriptionName;
		this.deviceDescriptionName = deviceDescriptionName;
	}

	public MicroStreamsHardwareImporter(String targetDescriptionName) {
		this(targetDescriptionName, null);
	}

	/**
	 * Give the JaxB object
	 *
	 * @return the jaxb of µStreams if it exists, otherwise null
	 */
	public static MicroStreamsHardwareJaxB getMicroStreamsJaxB() {
		if (microStreamsJaxB == null) {
			try {
				microStreamsJaxB = new MicroStreamsHardwareJaxB();
			} catch (JAXBException e) {
				logger.error("Failed to use JAXB", e);
				throw new RuntimeException(e);
			}
		}
		return microStreamsJaxB;
	}

	private static IBusHandler getBusHandler(IPeripheralModuleConfiguration peripheral) {
		switch (peripheral.getModuleDescriptionName()) { //TODO replace hard coded stuff!!
			case "core_connector_master":
			case "core_connector_slave":
				return DocumentUtil.getBusHandler(peripheral, "connector");
			default:
				throw new IllegalArgumentException("Unsupported Peripheral");
		}
	}

	private static List<IBusHandler> getBusHandlers(IPeripheralModuleConfiguration<?> peripheral) {
		List<IBusHandler> iBusHandlers = peripheral.getBusDescriptions().stream()
					.filter(bus -> !bus.getBusDescriptionName().equalsIgnoreCase(SPARTANMC_PERI_BUS))
					.map(bus -> DocumentUtil.getBusHandler(peripheral, bus.getName()))
					.collect(Collectors.toList());
		return iBusHandlers;
	}

	private static List<IPortHandler> getPortHandlers(IPeripheralModuleConfiguration<?> peripheral) {
		final List<IPortHandler> iPortHandlers = peripheral.getPortGroupDescriptions().stream()
				.flatMap(portGroup -> portGroup.getPorts().stream())
				.filter(port -> true)
				.map(port -> DocumentUtil.getPortHandler(peripheral, port.getName()))
				.collect(Collectors.toList());
		return iPortHandlers;
	}

	@Override
	public boolean requiresLibrary() {
		return true;
	}

	@Override
	public <T extends IConfigurationDocument> T importDocument(Path file, IDocumentSupplier<T> documentSupplier) {
		MicroStreamsHardwareJaxB jaxB = getMicroStreamsJaxB();

		Library library = documentSupplier.getLibrary();

		TargetDescription target = library.getTargetDescriptionWithAdvancedSearch(targetDescriptionName);
		DeviceDescription device = LibraryUtil.getDeviceDescriptionOrCompatible(library, target, deviceDescriptionName);

		try {
			Hardware hardware = jaxB.unmarshall(file);

			T document = documentSupplier.createDocument(target, device);

			addHardware(hardware, document, library);

			return document;
		} catch (JAXBException e) {
			logger.error("Failed to read xml file {}", file, e);
			throw new RuntimeException(e);
		}
	}

	private void addHardware(Hardware hardware, IConfigurationDocument document, Library library) {
		Map<Peripheral, IPeripheralModuleConfiguration> periMap = new HashMap<>();

		hardware.getSubsystems().forEach(sub -> {
			Core core = sub.getCore();
			ISubsystemConfiguration<?> subsystem = document.addSubsystem(library.getSubsystemDescription(core.getType()), core.getName());

			//set performance counter
			subsystem.getParameterHandler(IBooleanParameterHandler.class, subsystem.findParameterDescription(Core.PERFORMANCE_COUNTER).get()).setValue(core.hasPerformanceCounter());
			subsystem.getParameterHandler(IBooleanParameterHandler.class, subsystem.findParameterDescription(Core.SHOW_INSTUCTIONS).get()).setValue(false);
			subsystem.getParameterHandler(IIntegerParameterHandler.class, subsystem.findParameterDescription(Core.EVENTCOUNTERS).get()).setValue(Core.EVENTCOUNTERS_AMOUNT);
			subsystem.getParameterHandler(IIntegerParameterHandler.class, subsystem.findParameterDescription(Core.CYCLECOUNTER).get()).setValue(Core.CYCLECOUNTER_WIDTH);

			//set the memory module to use
			if (core.hasGlobalDataMem()) {
				createAndConnectMemoryModule(document, library, subsystem, MemType.GLOBAL_DATA_MEM, core.getFirmware(), MemoryBusType.DATA);
			}
			if (core.hasGlobalCodeMem()) {
				IMemoryModuleConfiguration<? extends IModuleContainer> memory = createAndConnectMemoryModule(document, library, subsystem, MemType.GLOBAL_CODE_MEM, core.getFirmware(), MemoryBusType.CODE, MemoryBusType.DATA);
				setMemoryFirmware(memory, core.getFirmware());
			}
			//add local mem if necessary
			else {

				IMemoryModuleConfiguration<? extends IModuleContainer> localMemory;

				try {
					final Optional<? extends IMemoryModuleConfiguration<? extends ISubsystemConfiguration>> possibleLocalMem = subsystem.streamMemoryModules().filter(mem -> mem.getModuleDescriptionName().equalsIgnoreCase(MemType.LOCAL_MEM.getModuleName())).collect(StreamUtil.atMostOne());

					if (!possibleLocalMem.isPresent()) {
						localMemory = createAndConnectMemoryModule(document, library, subsystem, MemType.LOCAL_MEM, core.getFirmware(), MemoryBusType.CODE, MemoryBusType.DATA);
					} else {
						localMemory = possibleLocalMem.get();
					}
					setMemoryFirmware(localMemory, core.getFirmware());
					localMemory.getParameterHandler(IBooleanParameterHandler.class, localMemory.findParameterDescription("SHOW_MEM").get()).setValue(false);
				} catch (IllegalStateException e) {
					logger.error("More than  one local memory is used...can't handle this situation");
					throw e;
				}

			}

			sub.getPeripherals().forEach(peri -> {
				IPeripheralModuleConfiguration peripheral = subsystem.addPeripheral(library.getPeripheralDescription(peri.getType()), peri.getModuleName());
				peri.getParameters().forEach(param -> {
					final IParameterHandler parameterHandler = DocumentUtil.getParameterHandler(peripheral, param.getName());
					parameterHandler.setStoredValue(param.getValue());
				});

				periMap.put(peri, peripheral);
			});
		});

		/*hardware.getConnections().stream()
				.flatMap(c -> c.getPeripheralConnections().stream())
				.map(pc -> periMap.get(pc.getPeripheral()))
				.collect(groupingBy(Function.identity(), counting()))
				.forEach((peri, connections) -> {
					setNrOfSlaves(peri, connections.intValue());
				});
		*/

		hardware.getConnections().forEach(con -> {
			if (con.getPeripheralConnections().size() != 2) {
				throw new IllegalArgumentException("Only P2P connections are supported");
			}

			IPeripheralModuleConfiguration peripheralA = periMap.get(con.getPeripheralConnections().get(0).getPeripheral());
			IPeripheralModuleConfiguration peripheralB = periMap.get(con.getPeripheralConnections().get(1).getPeripheral());
			connect(document, peripheralA, peripheralB);
			/*if (!pHA.isAdditionalConnectionAllowed() || !otherPort.isAdditionalConnectionAllowed()) {
				setNrOfSlaves(peripheralA, peripheralB);
			}*/

		});

		periMap.values().stream()
				.filter(p -> p.getModuleDescription().getName().equalsIgnoreCase("concentrator_master")
					|| p.getModuleDescription().getName().equalsIgnoreCase("dispatcher_master"))
				.forEach(p -> activateSoftwareArbitration(p));

	}

	private void connectWithoutScript(IConfigurationDocument document, IPeripheralModuleConfiguration<?> peripheralA, IPeripheralModuleConfiguration<?> peripheralB) {
		deactivateScripts(document);
		connect(document, peripheralA, peripheralB);
		restoreScripts(document);
	}

	private void connect(IConfigurationDocument document, IPeripheralModuleConfiguration<?> peripheralA, IPeripheralModuleConfiguration<?> peripheralB) {
		//connect the ports
		List<IPortHandler> portHandlerA = getPortHandlers(peripheralA);
		List<IPortHandler> portHandlerB = getPortHandlers(peripheralB);

		if (portHandlerA.size() != portHandlerB.size()) {
			throw new IllegalArgumentException("Peripheral port counts don't match for " + peripheralA.getModuleDescriptionName() + " -> " + peripheralB.getModuleDescriptionName());
		}

		int currentConnectionOffset = -1;
		boolean currentConnectionOffsetAssigned = false;
		for (IPortHandler pHA : portHandlerA) {
			//only connect inputs with outputs
			ConnectorDirection portADirection = pHA.getDirection();
			List<IPortHandler> prunedHandlerB = portHandlerB;
			if (portADirection != ConnectorDirection.InOut) {
				prunedHandlerB = prunedHandlerB.stream().filter(ph -> ph.getDirection() != portADirection).collect(Collectors.toList());
			}
			Optional<IPortHandler> otherPort = LevenshteinDistance.matchFromList(pHA, prunedHandlerB, IPortHandler::getPortName);
			if (otherPort.isPresent()) {
				if (pHA.getConnectorWidth() != otherPort.get().getConnectorWidth()) {
					IPortHandler<?> smallPort = pHA;
					IPortHandler<?> largePort = otherPort.get();
					if (smallPort.getConnectorWidth() > largePort.getConnectorWidth()) {
						largePort = pHA;
						smallPort = otherPort.get();
					}
					//assume that the width of the small partial connections is always the same...otherwise it would be different types
					if (!currentConnectionOffsetAssigned) {
						currentConnectionOffset = largePort.getConnectionCount();
						currentConnectionOffsetAssigned = true;
					}
					final int startB = currentConnectionOffset * smallPort.getConnectorWidth();
					final int width = smallPort.getConnectorWidth();
					final IConnection<? extends IPortHandler<?>> partialConnection = document.createPartialConnection(smallPort, 0, largePort, startB, width);
					try {
						partialConnection.connect();
						logger.info("Connected: " + partialConnection);
					} catch (ConnectionMismatchException e) {
						logger.warn("Skipping creation of connection since it would be invalid " + partialConnection);
					}
				} else {
					try {
						document.createPortConnection(pHA, otherPort.get()).connect();
					} catch (ConnectionMismatchException e) {
						logger.warn("Skipping creation of connection since it would be invalid " + pHA + otherPort);
					}

				}
			} else {
				logger.error("No matching port found for " + prunedHandlerB + " in " + pHA);
			}

		}

		//connect the buses
		List<IBusHandler> busHandlerA = getBusHandlers(peripheralA).stream().filter(bh -> bh.isAdditionalConnectionAllowed()).collect(Collectors.toList());
		List<IBusHandler> busHandlerB = getBusHandlers(peripheralB).stream().filter(bh -> bh.isAdditionalConnectionAllowed()).collect(Collectors.toList());

		if (busHandlerA.size() != busHandlerB.size()) {
			throw new IllegalArgumentException("Peripheral bus counts don't match for " + peripheralA.getModuleDescriptionName() + " and " + peripheralB.getModuleDescriptionName());
		}

		for (IBusHandler bHA : busHandlerA) {
			Optional<IBusHandler> otherBus = LevenshteinDistance.matchFromList(bHA, busHandlerB, IBusHandler::getAbsoluteConnectorName);
			if (otherBus.isPresent()) {
				logger.info("Connecting " + otherBus.get() + " to " + bHA);
				document.createBusConnection(bHA, otherBus.get()).connect();
			} else {
				logger.error("No matching bus found for " + busHandlerB + " in " + bHA);
			}
		}
	}

	private void setNrOfSlaves(IPeripheralModuleConfiguration<?> peripheral, int nrOfSlaves) {
		if (nrOfSlaves < 2) {
			nrOfSlaves = 2;
		}
		final String numberOfSlaves = "NUMBER_OF_SLAVES";
		Optional<? extends IParameter> nrOfSlavesParam = peripheral.findParameterDescription(numberOfSlaves);
		IPeripheralModuleConfiguration<?> peripheralToChange = peripheral;
		if (nrOfSlavesParam.isPresent()) {
			final IIntegerParameterHandler parameterHandler = peripheralToChange.getParameterHandler(IIntegerParameterHandler.class, nrOfSlavesParam.get());
			parameterHandler.setValue(nrOfSlaves);
		}

	}

	private void activateSoftwareArbitration(IPeripheralModuleConfiguration<?> peripheral) {
		final String softwareArbitration = "SOFTWARE_ARBITRATION";
		Optional<? extends IParameter> softwareArbitrationParam = peripheral.findParameterDescription(softwareArbitration);
		IPeripheralModuleConfiguration<?> peripheralToChange = peripheral;
		if (softwareArbitrationParam.isPresent()) {
			final IBooleanParameterHandler parameterHandler = peripheralToChange.getParameterHandler(IBooleanParameterHandler.class, softwareArbitrationParam.get());
			parameterHandler.setValue(true);
		}
	}

	private IMemoryModuleConfiguration<? extends IModuleContainer> createAndConnectMemoryModule(IConfigurationDocument document, Library library, ISubsystemConfiguration<?> subsystem, MemType memoryModuleType, Path firmware, MemoryBusType... busesToConnect) {
		IMemoryModuleConfiguration<? extends IModuleContainer> memory = null;
		Optional<IMemoryModuleConfiguration> possibleGlobalMem = getGlobalMemory(document, memoryModuleType);
		if (memoryModuleType != MemType.LOCAL_MEM && possibleGlobalMem.isPresent()) {
			memory = possibleGlobalMem.get();
		} else {
			if (memoryModuleType.asCommonModule()) {
				memory = document.addMemoryModule(library.getMemoryCommonDescription(memoryModuleType.getModuleName()));
				//this is a hack to make jconfig select proper firmware for build...since it is sorted by name and spartanmc_mem_local must come alphabetically before global mem.
				if (memoryModuleType == MemType.GLOBAL_DATA_MEM) {
					memory.getNameHandler().setStoredValue("spartanmc_shared_global_data");
				}
			} else {
				memory = subsystem.addMemoryModule(library.getMemoryCommonDescription(memoryModuleType.getModuleName()));
			}
			setMemoryFirmware(memory, firmware);
		}
		//connect it
		for (int i = 0; i < busesToConnect.length; i++) {
			MemoryBusType busToConnect = busesToConnect[i];
			Bus spartanCoreMemoryDataBus = subsystem.getBusDescriptions().stream().filter(b -> b.getBusDescriptionName().equalsIgnoreCase(busToConnect.getBusName())).findAny().get();
			Bus memoryDataBus = memory.getBusDescriptions().stream().filter(b -> b.getBusDescriptionName().equalsIgnoreCase(busToConnect.getBusName())).findAny().get();
			IBusHandler spartanBusHandler = subsystem.getBusHandler(spartanCoreMemoryDataBus);
			IBusHandler memoryBusHandler = memory.getBusHandler(memoryDataBus);
			if (spartanBusHandler.isAdditionalConnectionAllowed() && memoryBusHandler.isAdditionalConnectionAllowed()) {
				IConnection connection = document.createBusConnection(spartanBusHandler, memoryBusHandler);
				connection.connect();
			}
		}
		return memory;
	}

	private Optional<IMemoryModuleConfiguration> getGlobalMemory(IConfigurationDocument document, MemType memoryModuleType) {
		return document.streamModulesRecursive()
						.filter(module -> module instanceof IMemoryModuleConfiguration)
						.map(module -> (IMemoryModuleConfiguration) module)
						.filter(module -> module.getModuleDescriptionName().equalsIgnoreCase(memoryModuleType.getModuleName()))
						.findAny();
	}

	private void setMemoryFirmware(IMemoryModuleConfiguration memory, Path firmwareLocation) {
		IFirmwareAttributeHandler firmware = memory.getFirmwareLocationHandler();
		firmware.setAbsoluteLocation(firmwareLocation.getParent());
		firmware.setFirmwareName(firmwareLocation.getFileName().toString());
	}

	@NotNull
	private Optional<IScriptManager> getScriptManager(IConfigurationDocument document) {
		IPluginEngine pluginEngine = document.getDocumentHandler().getPluginEngine();
		if (pluginEngine != null) {
			Optional<IScriptManager> scriptManager = pluginEngine.getScriptManager();
			if (scriptManager.isPresent()) {
				return scriptManager;
			}
		}
		return Optional.empty();
	}

	private void deactivateScripts(IConfigurationDocument document) {
		final Optional<IScriptManager> scriptManager = getScriptManager(document);
		if (scriptManager.isPresent()) {
			scriptManagerPreviousState = scriptManager.get().isAutoScriptEnabled();
			scriptManager.get().setAutoScriptEnabled(false);
		}
	}

	private void restoreScripts(IConfigurationDocument document) {
		final Optional<IScriptManager> scriptManager = getScriptManager(document);
		if (scriptManager.isPresent()) {
			scriptManager.get().setAutoScriptEnabled(scriptManagerPreviousState);
		}
	}

	/**
	 * An enum of the different possible mem types
	 */
	private enum MemType {
		LOCAL_MEM("spartanmc_mem_local", false),
		GLOBAL_DATA_MEM("spartanmc_mem_data_global", true),
		GLOBAL_CODE_MEM("spartanmc_mem_code_global", true);

		private final String moduleName;
		private final boolean asCommonModule;

		MemType(String moduleName, boolean asCommonModule) {
			this.moduleName = moduleName;
			this.asCommonModule = asCommonModule;
		}

		boolean asCommonModule() {
			return asCommonModule;
		}

		String getModuleName() {
			return moduleName;
		}
	}


	/**
	 * An enum of the different Memory Busses
	 */
	private enum MemoryBusType {
		DATA("spartanmc-memory-bus"),
		CODE("spartanmc-code-memory-bus");

		private final String busName;

		MemoryBusType(String busName) {
			this.busName = busName;
		}

		String getBusName() {
			return busName;
		}
	}
}
