/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.scripting;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;

import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.MalformedURLException;
import java.net.URL;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Optional;
import java.util.stream.Stream;

/**
 * Created by Ramon on 21.06.2016.
 */
public class ScriptFinder {

	public static final String IS_VALID = "isValid";
	public static final String ON_NEW_DOC = "onNewDocument";
	public static final String ON_IMPORT_DOC = "onImportDocument";
	public static final String ON_EXPORT_DOC = "onExporting";
	public static final String ON_BUILDING_DOC = "onBuilding";
	public static final String ON_SAVING = "onSaving";
	public static final String ON_DOC_RESTRUCT = "onDocumentRestruct";
	public static final String ON_MODULE_RESTRUCT = "onRestruct";
	public static final String ON_MODULE_ADDED = "onAdd";
	public static final String ON_MODULE_REMOVED = "onRemove";
	public static final String ON_MODULE_NAME_CHANGED = "onRename";
	public static final String ON_PARAM_CHANGED = "onParamChange";
	public static final String ON_TARGET_MIGRATE = "onMigrate";
	public static final String ON_CONNECTOR_CHANGED = "onConnectorChange";
	public static final String MODULE_UTILS = "moduleUtils"; //variables and functions used by other scripts of the module

	private static final Logger logger = LoggerFactory.getLogger(ScriptFinder.class);
	private static final String PROJECT_SCRIPTS_DIR = "scripts";
	private static final String GLOBAL_SCRIPTS_DIR = "spartanmc/scripts";

	private final ILibJConfigProject project;
	private final String fileExt;
	@Nullable
	private final String shippedResourcesDir;
	private final Path workingDir;

	/**
	 *
	 * @param project the current project
	 * @param fileExt the file extension to search for
	 * @param shippedResourcesDir an option directory for ressources
	 */
	public ScriptFinder(ILibJConfigProject project, String fileExt, @Nullable String shippedResourcesDir) {
		workingDir = project.getWorkingDir();
		this.project = project;
		this.fileExt = fileExt;
		this.shippedResourcesDir = shippedResourcesDir;

	}

	/**
	 * Find a moduleDescription specific script
	 * <p>
	 * Search order:
	 * <ul>
	 * <li>ProjectDir/$PROJECT_SCRIPTS_DIR/$modDescName.$reactType.$fileExt</li>
	 * <li>SourceDir/[$prefix.]$reactType.$fileExt</li>
	 * <li>resources/$shippedResourcesDir/$modDescName.$reactType.$fileExt</li>
	 * </ul>
	 *
	 * @param moduleDescription specific to this module
	 * @param reactType filename to find
	 * @return uri of file if any
	 */
	public Optional<String> findReactScript(IModuleDescription moduleDescription, String reactType) {

		String fileName = getReactScriptName(moduleDescription, reactType);
		Path prjSpecific = workingDir.resolve(PROJECT_SCRIPTS_DIR).resolve(fileName);
		if (Files.isRegularFile(prjSpecific)) {
			logger.debug("Found Project level module script for {}.{}: {}", moduleDescription.getName(), reactType, prjSpecific);
			return Optional.of(fileName);
		}

		Path modSpecific = moduleDescription.getSourceFile().resolveSibling(fileName);
		if (Files.isRegularFile(modSpecific)) {
			logger.debug("Found source level module script for {}.{}: {}", moduleDescription.getName(), reactType, modSpecific);
			return Optional.of(fileName);
		}

		return Optional.empty();
	}

	/**
	 * find a global script
	 * <p>
	 * Search order:
	 * <ul>
	 * <li>ProjectDir/$PROJECT_SCRIPTS_DIR/$reactType.$fileExt</li>
	 * <li>$SPARTANMC_ROOT/$GLOBAL_SCRIPTS_DIR/$reactType.$fileExt</li>
	 * <li>resources/$shippedResourcesDir/$reactType.$fileExt</li>
	 * </ul>
	 *
	 * @param reactType filename of script
	 * @return uri of file if any
	 */
	public Optional<String> findReactScript(String reactType) {
		String fileName = getReactScriptName(reactType);
		Path prjSpecific = getScriptInPrj(fileName);
		if (Files.isRegularFile(prjSpecific)) {
			logger.debug("Found Project level script for {}: {}", reactType, prjSpecific);
			return Optional.of(fileName);
		}

		Path global = project.getApplication().getSpartanRoot().resolve(GLOBAL_SCRIPTS_DIR).resolve(fileName);
		if (Files.isRegularFile(global)) {
			logger.debug("Found global script for {}: {}", reactType, global);
			return Optional.of(fileName);
		}

		return Optional.empty();
	}

	public String getReactScriptName(IModuleDescription moduleDescription, String reactType) {
		String name = moduleDescription.getName();
		return String.format("%s.%s.%s", name, reactType, fileExt);
	}

	public String getReactScriptName(String reactType) {
		return String.format("%s.%s", reactType, fileExt);
	}

	public Path getProjectBase() {
		return workingDir.resolve(PROJECT_SCRIPTS_DIR);
	}

	/**
	 *
	 * @param library a lib to be scanned
	 * @return the library to converted to URIs
	 */
	public URL[] getBaseURLs(Library library) {
		return Stream.concat(Stream.of(getProjectBase()), library.getModuleSourceTrees().stream()).map(p -> {
			try {
				return p.toUri().toURL();
			} catch (MalformedURLException e) {
				logger.error("Failed to convert {} to URI", p, e);
				throw new RuntimeException(e);
			}
		}).toArray(URL[]::new);
	}

	public Path getScriptInPrj(String scriptName) {
		return workingDir.resolve(PROJECT_SCRIPTS_DIR).resolve(scriptName);
	}

	public Path getScriptRelativeToSourceTree(IModuleDescription moduleDescription, String scriptName) {
		return moduleDescription.getSourceTree().toAbsolutePath().relativize(moduleDescription.getSourceFile().resolveSibling(scriptName)).normalize();
	}

	public Path getScriptInSourceTree(IModuleDescription moduleDescription, String scriptName) {
		return moduleDescription.getSourceFile().resolveSibling(scriptName);
	}

	public Path getScriptRelativeToRoot(String scriptName) {
		return Paths.get(GLOBAL_SCRIPTS_DIR, scriptName);
	}

	public Path getScriptInRoot(String scriptName) {
		return project.getApplication().getSpartanRoot().resolve(GLOBAL_SCRIPTS_DIR).resolve(scriptName);
	}
}
