# Groovy Script

## Script API

### available vars:
* document
* doc
* documentHandler
* project
* library


The **full JConfig Java API** is accessible on these instances. If the shell works correctly, the whole API of libjconfig including building logic and layouting logic is available, although it might require prior import.

### Additional Groovy API
Groovy supports properties. All Bean setters / getters will also be available under their property name. (Ex: `getSubsystems()`, is available as `subsystems`) This works for reading and writing.

In groovy, brackets on methods may be omitted. Calls can be chained without the . operator, to readable sentences.
Ex: `document find "core0" getPorts`

`ModuleExtension` extends several regular classes with very convenient methods:
* `IConfigurationDocument.each {}` applies closure to all modules in document
* `IConfigurationDocument.find(name)` find`first module matching in name (recursive)
* `IConfigurationDocument.findByName(name)` find **all** matching modules
* `IConfigurationDocument.findByType(type)` final all matching modules
* `IConfigurationDocument.modules` (LinkedHashMap of child modules)
* `IConfigurationDocument.pins`  (LinkedHashMap of pins)
* `IConfigurationDocument.connect` (create respective connection and connects it)
* `ISubsystemConfiguration.modules`  (LinkedHashMap of child modules)
* `IModuleConfiguration.ports`  (LinkedHashMap of ports)
* `IModuleConfiguration.buses`  (LinkedHashMap of buses)
* `IModuleConfiguration.parameters`  (LinkedHashMap of parameters)
* `parameters["paramName"] as Long / Double / Boolean / String`  automatically converts parameter
* `parameters["paramName"] = Long / Double / Boolean / String`  writes automatically converted parameter
* `.add(descName, moduleName defaults descName)` on doc and subsystems.
* `container + "descName"` on doc and subsystems.
* `container + [desc:"descName", name:"moduleName"]` on doc and subsystems.
* `.remove(moduleName)` on doc and subsystems
* `.remove(moduleInstance)` on doc and subsystems
* `container - "moduleName"` on doc and subsystems
* `container - moduleInstance` on doc and subsystems
* `IConfigurationElement.name = "new name"` (= setName())
* several more functions from DocumentUtil are available as extension methods

### BaseScript
Wherever the BaseScript is applied the following functions are available directly:
* all variants of `connect()`/`connectMultiple()`
* ```find()` variants
* `add()` / `remove()` variants
* every unknown identifier like `spartanmc_0` will be resolved if possible by internally using the `find()`
* Lifecycle methods such as
  * `newDocument('target[:device]')`
  * `open(file)`
  * `save([file])`
  * `validate()`
  * `build([all])`
method

### Shell
This means the following is possible
```
add 'spartamc' // add subsystem to document
doc.modules.spartanmc_0 + 'uart_light' // add uart_light modules to subsystem spartanmc_0
add 'clk_xilinx'
clk_xilinx_0.parameters.RESET_LEVEL = "HIGH_ACTIVE"
connect clk_xilinx_0.ports.clkfx, spartanmc_0.ports.clk
connect PIN_CLOCK_USER, clk_xilinx_0.ports.clkgen
spartanmc_0 + 'port_in'
port_in_0.parameters.COUNT = 4
connectMultiple port_in_0.ports.pin, doc, ["GPIO_BUTTON_0", GPIO_BUTTON_1]
port_in_0.ports.pin.disconnectAll()
connectMultiple port_in_0.ports.pin, doc, {i -> "GPIO_BUTTON_$i" }
```
### Groovy

#### Maps
In Groovy, Maps can be accessed with the subscript operator: [].
This means ```ports['rx']``` will return the IPortHandler for 'rx' if it exists.
Groovy Maps also support reading via property,
meaning ```ports.rx``` is equivalent.
The Maps just as lists support a whole bunch of different methods.

#### Elvis Operator
```x != null? x : "default"``` can be shortened to ```x?:"default"```

#### GString
In Groovy both "" and '' denote Strings. "" is a GString and can dereference vars like this: ```"Hi $name, this is ${me["name"]}"```

#### Closures
Groovys variant of Lambda functions. Can be converted to Java 8 Lambas and back automatically.
```{ var1,var2 -> code}```
As in all groovy code, last expression is automatically returned. Vars can be omitted, in which case there is a single default parameter called 'it'.
In addition to lambdas, Closures can resolve methods / properties late (execution) against arbitrary delegates.

## Examples

Here groovy Closures are combined with injected functional methods, GStrings and the original Java API
```groovy
document.each {
	println it.name
	it.buses.each {n,b ->
		println "\tbus $n: "
		b.streamOppositeConnectors().forEach { println "\t\t${it.absoluteConnectorName}" }
	}
}
```

will output for example

```
subsystem_0
	bus peri-bus: 
		/subsystem_0/intctrl_0@peri-bus
		/subsystem_0/port_in_0@peri-bus
		/subsystem_0/port_out_0@peri-bus
		/subsystem_0/uart_light_0@peri-bus
		/subsystem_0/uart_light_1@peri-bus
	bus dma-bus: 
	bus memdata-bus: 
		/subsystem_0/spartanmc_mem_local_0@data-bus
	bus memcode-bus: 
		/subsystem_0/spartanmc_mem_local_0@code-bus
clk_xilinx_0
spartanmc_mem_local_0
	bus data-bus: 
		/subsystem_0@memdata-bus
	bus code-bus: 
		/subsystem_0@memcode-bus
port_in_0
	bus peri-bus: 
		/subsystem_0@peri-bus
port_out_0
	bus peri-bus: 
		/subsystem_0@peri-bus
uart_light_0
	bus peri-bus: 
		/subsystem_0@peri-bus
uart_light_1
	bus peri-bus: 
		/subsystem_0@peri-bus
intctrl_0
	bus peri-bus: 
		/subsystem_0@peri-bus

```
