/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.scripting.groovy

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler
import de.tu_darmstadt.rs.spartanmc.jconfig.scripting.ModuleExtension
import groovy.transform.stc.ClosureParams
import groovy.transform.stc.SimpleType
import org.slf4j.Logger
import org.slf4j.LoggerFactory

import java.nio.file.Paths

/**
 * Created by Ramon on 22.06.2016.
 */
public abstract class GroovyScriptBase extends Script {

	protected final Logger logger = LoggerFactory.getLogger(this.getClass());

	//	public void println(String string) {
	//		logger.warn(PrintStreamLogger.scriptMarker, string);
	//	}

	public IConfigurationDocument getDocument() {
		return ((IConfigurationDocument) getProperty("document"));
	}

	public IConfigurationDocument getDoc() {
		return getDocument();
	}

	public IModuleConfiguration<?> find(String name) {
		return getDocument().find(name);
	}

	public Set<IModuleConfiguration<?>> findByName(String name) {
		return getDocument().findByName(name);
	}

	public Set<IModuleConfiguration<?>> findByType(String type) {
		return getDocument().findByType(type);
	}

	public boolean add(String moduleDescriptionName, String moduleName = null) {
		return getDocument().add(moduleDescriptionName, moduleName)
	}

	public boolean remove(String moduleName) {
		def mod = find(moduleName)
		return mod.parent.remove(mod)
	}

	public boolean remove(IModuleConfiguration module) {
		return module.parent.remove(module)
	}

	public void withType(String type,
						 @ClosureParams(value = SimpleType, options = ['de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration']) Closure<Void> closure) {
		for (IModuleConfiguration<?> mod : findByType(type)) {
			closure.delegate = mod;
			closure(mod)
		}
	}

	public IConnection<? extends IPortHandler<?>> connect(IPortHandler<? extends IConnectorDescription> a, IPortHandler<? extends IConnectorDescription> b) {
		if (!a || !b)
			throw new IllegalArgumentException("Need 2 ports / pins. Given: [${a?.portName}, ${b?.portName}]");
		IConnection<? extends IPortHandler<?>> connection = getDocument().createPortConnection(a, b);
		connection.connect();
		return connection;
	}

	public IConnection<? extends IBusHandler> connect(IBusHandler a, IBusHandler b) {
		IConnection<? extends IBusHandler> connection = getDocument().createBusConnection(a, b);
		connection.connect();
		return connection;
	}

	public IConnection<? extends IPortHandler<?>> connectIfConnectable(IPortHandler<? extends IConnectorDescription> a, IPortHandler<? extends IConnectorDescription> b) {
		if (a.isConnectibleTo(b) && a.isAdditionalConnectionAllowed() && b.isAdditionalConnectionAllowed()) {
			return connect(a, b);
		}
		return null;
	}

	public IConnection<? extends IBusHandler> connectIfConnectable(IBusHandler a, IBusHandler b) {
		if (a.isConnectibleTo(b) && a.isAdditionalConnectionAllowed() && b.isAdditionalConnectionAllowed()) {
			return connect(a, b);
		}
		return null;
	}

	public IConnection<? extends IPortHandler<?>> connect(IPortHandler<? extends IConnectorDescription> a, int aStart, IPortHandler<? extends IConnectorDescription> b, int bStart, int width) {
		IConnection<? extends IPortHandler<?>> connection = getDocument().createPartialConnection(a, aStart, b, bStart, width);
		connection.connect();
		return connection;
	}

	public List<IConnection<? extends IPortHandler<?>>> connectMultiple(IPortHandler<?> a, int aStart = 0, int width = ModuleExtension.AUTO, IConfigurationElement other, List<String> portNames) {
		return document.connectMultiple(a, aStart, width, other, portNames)
	}

	public List<IConnection<? extends IPortHandler<?>>> connectMultiple(IPortHandler<?> a, int aStart = 0, int width = ModuleExtension.AUTO, IConfigurationElement other,
																		@ClosureParams(value = SimpleType, options = ['int']) Closure<String> portNames) {
		return document.connectMultiple(a, aStart, width, other, portNames)
	}

	def execute(String groovyFile) {
		evaluate(new File(groovyFile))
	}

	def propertyMissing(String name) {
		if (name.startsWith('PIN_')) {
			def pinName = name - 'PIN_'
			doc.pins[pinName]
		} else {
			find(name)
		}
	}

	public IConfigurationDocument open(String file) {
		documentHandler.open(Paths.get(file))
		documentHandler.document
	}

	public IConfigurationDocument newDocument(String targetConf) {
		documentHandler.newDocument(targetConf)
		documentHandler.document
	}

	public boolean save(String file = null) {
		if (file != null)
			documentHandler.setSaveLocation(Paths.get(file))
		documentHandler.save()
	}

	public boolean validate() {
		documentHandler.isValid()
	}

	public boolean build(boolean all = false) {
		documentHandler.build(!all)
	}
}
