/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.scripting.groovy;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPluginEngine;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IScriptManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentExporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentImporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.scripting.DocumentContextCompleter;
import de.tu_darmstadt.rs.spartanmc.jconfig.scripting.ReplWriter;
import de.tu_darmstadt.rs.spartanmc.jconfig.scripting.ScriptFinder;

import groovy.lang.Binding;
import groovy.lang.GroovyClassLoader;
import groovy.lang.GroovyShell;
import groovy.util.GroovyScriptEngine;
import groovy.util.ResourceException;
import groovy.util.ScriptException;
import jline.console.completer.AggregateCompleter;
import jline.console.completer.Completer;
import org.codehaus.groovy.control.CompilerConfiguration;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.stream.Collectors;

/**
 * Created by Ramon on 21.06.2016.
 */
public class GroovyScriptPlugin implements IPluginEngine, IScriptManager {

	public static final String UTF8 = "UTF-8";
	public static final String HelpMessageFile = "/file/GroovyScriptHelpMessage.md";

	private static final Marker scriptMarker = MarkerFactory.getMarker("Script");
	private static final Marker scriptOutMarker = MarkerFactory.getMarker("Script.out");
	private static final Marker scriptErrMarker = MarkerFactory.getMarker("Script.err");
	private static final Logger logger = LoggerFactory.getLogger(GroovyScriptPlugin.class);
	private static final String DOCUMENT = "document";
	private static final String REPORT = "report";
	private static final String NOT_PROJECT_LEVEL_SCRIPT_FOR = "Not project level script for {}";
	private static final String NOT_SOURCE_TREE_LEVEL_SCRIPT_FOR = "Not sourceTree level script for {}";

	private final Binding binding;
	private GroovyShell shell;
	private ScriptFinder finder;
	private GroovyScriptEngine engine;
	private CompilerConfiguration compilerConfiguration;
	private boolean autoScript = true;
	private boolean autoFlush = false;
	private PrintStream outStream;
	private PrintStream errStream;
	private Consumer<String> eventSink;
	private Completer completer;

	public GroovyScriptPlugin() {
		binding = new Binding();
	}

	public Binding getBinding() {
		return binding;
	}

	private void logScriptOut(String s) {
		if (s != null && !s.isEmpty()) {
			logger.info(scriptOutMarker, strip(s));
		}
	}

	private void logScriptErr(String s) {
		if (s != null && !s.isEmpty()) {
			logger.error(scriptErrMarker, strip(s));
		}
	}

	private String strip(String s) {
		int last = s.length() - 1;
		if (s.charAt(last) == '\n') {
			return s.substring(0, last);
		} else {
			return s;
		}
	}

	@Override
	public void onAttached(ILibJConfigProject project) {
		IApplicationConfiguration application = project.getApplication();
		Library library = application.getLibrary();
		binding.setProperty("project", project);
		binding.setProperty("library", library);
		binding.setProperty("documentHandler", project.getDocumentHandler());
		this.finder = new ScriptFinder(project, "groovy", null);

		this.compilerConfiguration = createCompilerConfiguration();

		//String classpath = application.getSpartanRoot().resolve("bin").resolve("lib").resolve("libjconfig-3.0-alpha1.jar").toString();
		//		compilerConfiguration.setClasspathList(Collections.singletonList(classpath));

		//		ImportCustomizer ic = new ImportCustomizer();
		//		ic.addImports(AddressLayout.class.getName());
		//		compilerConfiguration.addCompilationCustomizers(ic);

		engine = new GroovyScriptEngine(finder.getBaseURLs(library), new GroovyClassLoader(this.getClass().getClassLoader(), compilerConfiguration));
		shell = new GroovyShell(engine.getGroovyClassLoader(), binding, compilerConfiguration);

		ArrayList<Completer> completers = new ArrayList<Completer>();
		completers.add(new GroovyReflectionCompleter(shell));
		completers.add(new DocumentContextCompleter(application));
		completer = new AggregateCompleter(completers);
	}

	/**
	 * @return The compiler config to be used
	 */
	public CompilerConfiguration createCompilerConfiguration() {
		CompilerConfiguration compilerConfiguration = new CompilerConfiguration();
		compilerConfiguration.setScriptBaseClass(GroovyScriptBase.class.getName());
		compilerConfiguration.setTargetBytecode("1.8"); // default apparently is 1.5
		return compilerConfiguration;
	}

	@Override
	public void onOpenedDocument(IConfigurationDocument document) {
		onAttachingToDocument(document);
	}

	@Override
	public void onImportedDocument(IConfigurationDocument document, IDocumentImporter importer) {
		onAttachingToDocument(document);
		runIfPresent(ScriptFinder.ON_IMPORT_DOC);
	}

	@Override
	public void onCreatedDocument(IConfigurationDocument document) {
		onAttachingToDocument(document);
		runIfPresent(ScriptFinder.ON_NEW_DOC);
	}

	@Override
	public boolean isModuleInUse(IModuleConfiguration module) {
		return false;
	}

	@Override
	public void onNoDocument() {
		binding.setProperty(DOCUMENT, null);
	}

	@Override
	public void onSaving(IConfigurationDocument document) {
		runIfPresent(ScriptFinder.ON_SAVING);
	}

	@Override
	public void onExportingDocument(IConfigurationDocument document, IDocumentExporter exporter) {
		runIfPresent(ScriptFinder.ON_EXPORT_DOC);
	}

	@Override
	public void onBuilding(IConfigurationDocument document) {
		runIfPresent(ScriptFinder.ON_BUILDING_DOC);
	}

	@Override
	public boolean isValid(IModuleConfiguration<?> module, boolean report) {
		binding.setProperty(REPORT, report);
		runIfPresent(module, ScriptFinder.MODULE_UTILS);
		Object result = runIfPresent(module, ScriptFinder.IS_VALID);
		try {
			return result == null || (Boolean) result;
		} catch (Exception e) {
			return false;
		}
	}

	@Override
	public boolean isValid(IConfigurationDocument configurationDocument, boolean report) {
		binding.setProperty(REPORT, report);
		Object result = runIfPresent(ScriptFinder.IS_VALID);
		try {
			return result == null || (Boolean) result;
		} catch (Exception e) {
			return true;
		}
	}

	@Override
	public Optional<IScriptManager> getScriptManager() {
		return Optional.of(this);
	}

	@Override
	public void executeScript(URI path) {
		try {
			shell.evaluate(path);
		} catch (IOException e) {
			logger.error("Failed to load script {}", path, e);
			throw new RuntimeException(e);
		}

		if (autoFlush) {
			outStream.flush();
			errStream.flush();
		}
	}

	@Override
	public Optional<Object> eval(String script) {
		Object res = shell.evaluate(script);

		if (autoFlush) {
			outStream.flush();
			errStream.flush();
		}

		return Optional.ofNullable(res);
	}

	@Override
	public int autocomplete(String input, int cursorPos, List<String> candidates) {
		List<CharSequence> props = new ArrayList<CharSequence>();
		candidates.clear();

		int identifierStart = completer.complete(input, cursorPos, props);

		candidates.addAll(props.stream().map(CharSequence::toString).collect(Collectors.toList()));

		return identifierStart;
	}

	@Override
	public String getHelpMessage() {
		InputStream in = getClass().getResourceAsStream(HelpMessageFile);
		StringBuffer message = new StringBuffer();

		try (BufferedReader reader = new BufferedReader(new InputStreamReader(in))) {
			reader.lines().forEachOrdered(l -> message.append(l + "\n"));
		} catch (IOException e) {
			throw new RuntimeException("Error reading help file", e);
		}

		return message.toString();
	}

	@Override
	public boolean supportsFile(@NotNull Path scriptFile) {
		return scriptFile.getFileName().toString().endsWith(".groovy");
	}

	@Override
	public void redirectStdIo(Consumer<String> stdOut, Consumer<String> stdErr, Consumer<String> events) {
		try {
			outStream = new PrintStream(new ReplWriter(stdOut.andThen(this::logScriptOut)), true, UTF8);
			binding.setProperty("out", outStream);
			errStream = new PrintStream(new ReplWriter(stdErr.andThen(this::logScriptErr)), true, UTF8);
			binding.setProperty("err", errStream);
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException("No UTF-8??", e);
		}
		this.eventSink = events;
		autoFlush = true;
	}

	/**
	 *  {@inheritDoc}
	 */
	@Override
	public boolean isAutoScriptEnabled() {
		return autoScript;
	}

	/**
	 *  {@inheritDoc}
	 */
	@Override
	public void setAutoScriptEnabled(boolean enabled) {
		autoScript = enabled;
	}

	/**
	 *
	 * @param name the name of the script
	 * @param module the module to run the script on
	 * @return the object returned by the script
	 */
	public Object runScript(String name, @Nullable IModuleConfiguration<?> module) {
		try {
			binding.setVariable("module", module);
			GroovyScriptBase script = (GroovyScriptBase) engine.createScript(name, binding);
			logger.info(scriptMarker, "Beginning to execute Script {}", name);
			Object res = script.run();
			logger.info(scriptMarker, "Finished executing Script {} ==> {}", name, res);
			if (eventSink != null) {
				eventSink.accept(String.format("@ Script: %s ==> %s\n", name, res));
			}
			return res;
		} catch (ScriptException e) {
			logger.error("Script {} could not be parsed", name, e);
			return e;
		} catch (ResourceException e) {
			throw new NoSuchElementException("Script not found: " + name);
		} catch (Exception e) {
			logger.error("Script {} crashed", name, e);
			if (eventSink != null) {
				eventSink.accept(String.format("!!! Script: %s crashed:\n%s\n", name, e.getMessage()));
			}
			return e;
		} finally {
			if (autoFlush) {
				outStream.flush();
				errStream.flush();
			}
		}
	}

	private Object runIfPresent(IModuleConfiguration<?> module, String reactType) {
		if (!autoScript) {
			return null;
		}

		IModuleDescription moduleDescription = module.getModuleDescription();
		String scriptName = finder.getReactScriptName(moduleDescription, reactType);
		String tryScript = scriptName;

		try {
			return runScript(tryScript, module);
		} catch (Exception e) {
			logger.trace(NOT_PROJECT_LEVEL_SCRIPT_FOR, scriptName);
		}

		tryScript = finder.getScriptRelativeToSourceTree(moduleDescription, scriptName).toString();

		try {
			return runScript(tryScript, module);
		} catch (Exception e) {
			logger.trace(NOT_SOURCE_TREE_LEVEL_SCRIPT_FOR, scriptName);
		}

		return null;
	}

	private Object runIfPresent(String reactType) {
		if (!autoScript) {
			return null;
		}

		Objects.requireNonNull(finder);
		String scriptName = finder.getReactScriptName(reactType);
		String tryScript = scriptName;

		try {
			return runScript(tryScript, null);
		} catch (Exception e) {
			logger.trace(NOT_PROJECT_LEVEL_SCRIPT_FOR, scriptName);
		}

		tryScript = finder.getScriptRelativeToRoot(scriptName).toString();

		try {
			return runScript(tryScript, null);
		} catch (Exception e) {
			logger.trace(NOT_SOURCE_TREE_LEVEL_SCRIPT_FOR, scriptName);
		}

		return null;
	}

	private void onAttachingToDocument(IConfigurationDocument document) {
		binding.setProperty(DOCUMENT, document);
	}

	@Override
	public void onParameterChanged(IConfigurationElement<?, ?, ?> element, IAttributeHandler handler) {
		if (element instanceof IModuleConfiguration) {
			runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.MODULE_UTILS);
			runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.ON_PARAM_CHANGED);
		}
	}

	@Override
	public void onFirmwareChanged(IMemoryModuleConfiguration element) {
		//		runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.ON_FIRMWARE_CHANGED);
	}

	@Override
	public void onConnectorChanged(IConfigurationElement<?, ?, ?> element, IConnectorHandler<?, ?> connector, ConnectorChangeType type, IConnection<?> connection) {
		if (element instanceof IModuleConfiguration) {
			binding.setVariable("connector", connector);
			binding.setVariable("type", type);
			binding.setVariable("connection", connection);
			runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.MODULE_UTILS);
			runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.ON_CONNECTOR_CHANGED);
			binding.setVariable("connector", null);
			binding.setVariable("type", null);
			binding.setVariable("connection", null);
		}
	}

	@Override
	public void onElementNameChanged(IConfigurationElement<?, ?, ?> element) {
		if (element instanceof IModuleConfiguration) {
			runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.MODULE_UTILS);
			runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.ON_MODULE_NAME_CHANGED);
		}
	}

	@Override
	public void onAttributeChanged(IConfigurationElement<?, ?, ?> element, String attribute) {

	}

	@Override
	public void onElementDeclarationChanged(IConfigurationElement<?, ?, ?> element) {
		if (element instanceof IModuleConfiguration) {
			runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.MODULE_UTILS);
			runIfPresent(((IModuleConfiguration<?>) element), ScriptFinder.ON_MODULE_RESTRUCT);
		} else {
			runIfPresent(ScriptFinder.ON_DOC_RESTRUCT);
		}
	}

	@Override
	public void onModuleAdded(IModuleConfiguration<?> module) {
		runIfPresent(module, ScriptFinder.MODULE_UTILS);
		runIfPresent(module, ScriptFinder.ON_MODULE_ADDED);
	}

	@Override
	public void onModuleRemoved(IConfigurationElement<?, ?, ?> parent, IModuleConfiguration module) {
		runIfPresent(module, ScriptFinder.MODULE_UTILS);
		runIfPresent(module, ScriptFinder.ON_MODULE_REMOVED);
	}

	@Override
	public void onTargetChanged() {
		runIfPresent(ScriptFinder.ON_TARGET_MIGRATE);
	}
}
