/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.scripting.groovy

import jline.console.completer.Completer

/**
 * Implements the Completer interface to provide completions for
 * GroovyShell by using reflection on global variables.
 */
class GroovyReflectionCompleter implements Completer {

	private GroovyShell shell;

	GroovyReflectionCompleter(GroovyShell shell) {
		this.shell = shell
	}

	int complete(String buffer, int cursor, List candidates) {

		int identifierStart = findIdentifierStart(buffer, cursor)
		String identifierPrefix = identifierStart != -1 ? buffer.substring(identifierStart, cursor) : ""
		int lastDot = buffer.lastIndexOf('.')

		// if there are no dots, and there is a valid identifier prefix
		if (lastDot == -1) {
			if (identifierStart != -1) {
				List myCandidates = findMatchingVariables(identifierPrefix)
				if (myCandidates.size() > 0) {
					candidates.addAll(myCandidates)
					return identifierStart

				}
			}
		} else {
			// there are 1 or more dots
			// if ends in a dot, or if there is a valid identifier prefix
			if (lastDot == cursor - 1 || identifierStart != -1) {
				// evaluate the part before the dot to get an instance
				String instanceRefExpression = buffer.substring(0, lastDot)
				def instance = shell.evaluate([instanceRefExpression])
				if (instance != null) {
					// look for public methods/fields that match the prefix
					List myCandidates = getPublicFieldsAndMethods(instance, identifierPrefix)
					if (myCandidates.size() > 0) {
						candidates.addAll(myCandidates)
						return lastDot + 1
					}
				}
			}
		}

		// no candidates
		return -1
	}

	/**
	 * Parse a buffer to determine the start index of the groovy identifier
	 * @param buffer the buffer to parse
	 * @param endingAt the end index with the buffer
	 * @return the start index of the identifier, or -1 if the buffer
	 * does not contain a valid identifier that ends at endingAt
	 */
	int findIdentifierStart(String buffer, int endingAt) {
		// if the string is empty then there is no expression
		if (endingAt == 0)
			return -1
		// if the last character is not valid then there is no expression
		char lastChar = buffer.charAt(endingAt - 1)
		if (!Character.isJavaIdentifierPart(lastChar))
			return -1
		// scan backwards until the beginning of the expression is found
		int startIndex = endingAt - 1
		while (startIndex > 0 && Character.isJavaIdentifierPart(buffer.charAt(startIndex - 1)))
			--startIndex
		return startIndex
	}

	/**
	 * Build a list of public fields and methods for an object
	 * that match a given prefix.
	 * @param instance the object
	 * @param prefix the prefix that must be matched
	 * @return the list of public methods and fields that begin with the prefix
	 */
	List getPublicFieldsAndMethods(Object instance, String prefix) {
		def rv = []
		instance.class.fields.each {
			if (it.name.startsWith(prefix))
				rv << it.name
		}
		instance.class.methods.each {
			if (it.name.startsWith(prefix))
			//rv << it.name + (it.parameterTypes.length == 0 ? "()" : "(")
				if (it.parameters.length == 0)
					rv << it.name + "()"
				else { //if method has params, add them to autocompletion proposal
					String candidateWithParams = it.name + "("
					it.parameters.each { param ->
						candidateWithParams += (param == it.parameters.last()) ?
								(param.type.simpleName + " " + param.name + ")") :
								(param.type.simpleName + " " + param.name + ", ");
					}
					rv << candidateWithParams
				}
		}
		return rv.sort().unique()
	}

	/**
	 * Build a list of variables defined in the shell that
	 * match a given prefix.
	 * @param prefix the prefix to match
	 * @return the list of variables that match the prefix
	 */
	List findMatchingVariables(String prefix) {
		def matches = []
		for (String varName in shell.getContext().getVariables().keySet())
			if (varName.startsWith(prefix))
				matches << varName
		return matches
	}
}
