/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.scripting.lua;


import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPortGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import org.luaj.vm2.Globals;
import org.luaj.vm2.LuaValue;
import org.luaj.vm2.Varargs;
import org.luaj.vm2.lib.OneArgFunction;
import org.luaj.vm2.lib.ThreeArgFunction;
import org.luaj.vm2.lib.TwoArgFunction;
import org.luaj.vm2.lib.VarArgFunction;
import org.luaj.vm2.lib.ZeroArgFunction;
import org.luaj.vm2.lib.jse.CoerceJavaToLua;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.URI;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;

/**
 * Created by melanieb on 02.03.16.
 */
public class LuaInterface {

	public static final String CURRENT_SUBSYSTEM_VAR = "currentSubsystem";
	public static final String CURRENT_ELEMENT_VAR = "currentElement";
	private static final Logger logger = LoggerFactory.getLogger(LuaInterface.class);
	public Library library;
	public IDocumentHandler documentHandler;

	private final Globals binding;
	private Consumer<URI> scripExecutor;

	/**
	 * to initialize the variabels
	 *
	 * @param library the libray needed to add elements to the configuration
	 * @param documentHandler needed to  change the configuration
	 * @params binding binding on which to declare our fuctions
	 */
	public LuaInterface(Library library, IDocumentHandler documentHandler, Globals binding, Consumer<URI> scriptExecutor) {
		this.library = library;
		this.documentHandler = documentHandler;
		this.binding = binding;
		this.scripExecutor = scriptExecutor;
		addFunctions(binding);
	}

	private void addFunctions(Globals bindings) {
		bindings.set("newdoc", new LuaNewdoc());
		bindings.set("open", new LuaOpen()); //macht mehr sinn wenn die pfade für einen rausgesucht werden.
		bindings.set("saveas", new LuaSaveas());
		bindings.set("save", new LuaSave());
		bindings.set("config", new LuaConfig());
		bindings.set("getTargetName", new LuaGetTargetName());
		bindings.set("contains", new LuaContains());
		bindings.set("getParent", new LuaGetParent());
		bindings.set("getModule", new LuaGetModule());
		bindings.set("get", new LuaGet());
		bindings.set("getSubsystem", new LuaGetSubsystem());
		bindings.set("add", new LuaAdd());
		bindings.set("remove", new LuaRemove());
		bindings.set("connectBus", new LuaConnectBus());// weis nicht ob man dies benötigt
		bindings.set("connectPort", new LuaConnectPort());
		bindings.set("connectPin", new LuaConnectPin());
		bindings.set("partialConnection", new LuaPartialConnection());
		bindings.set("disconnect", new LuaDisconnect());
		bindings.set("disconnectAll", new LuaDisconnectAll());
		bindings.set("setParam", new LuaSetParam());
		bindings.set("getParam", new LuaGetParam());
		bindings.set("script", new LuaScript());
	}

	private IModuleConfiguration<?> getCurrentElement() {
		return ((IModuleConfiguration<?>) binding.get(CURRENT_ELEMENT_VAR).touserdata());
	}

	private void setCurrentElement(IModuleConfiguration<?> moduleConfiguration) {
		binding.set(CURRENT_ELEMENT_VAR, LuaValue.userdataOf(moduleConfiguration));
	}

	private ISubsystemConfiguration<?> getCurrentSubsystem() {
		return ((ISubsystemConfiguration<?>) binding.get("currentSubsystem").touserdata());
	}

	private void setCurrentSubsystem(ISubsystemConfiguration<?> subsystemConfiguration) {
		binding.set(CURRENT_SUBSYSTEM_VAR, LuaValue.userdataOf(subsystemConfiguration));
	}


	private class LuaNewdoc extends OneArgFunction {

		/**
		 * Opens a new document with the given targethardware
		 *
		 * @param targetname The name of the target hardware
		 * @return nothing
		 */
		public LuaValue call(LuaValue targetname) {
			documentHandler.newDocument(targetname.toString());
			return CoerceJavaToLua.coerce(true);
		}
	}


	private class LuaOpen extends OneArgFunction {

		/**
		 * Opens a document
		 *
		 * @param pathtofile the working directory plus the pathtofile is the path of the document opend
		 * @return nothing
		 */
		public LuaValue call(LuaValue pathtofile) {
			documentHandler.open(documentHandler.getWorkingDir().resolve(pathtofile.checkjstring() + ".jc3"));
			return CoerceJavaToLua.coerce(true);
		}
	}


	private class LuaSave extends ZeroArgFunction {

		/**
		 * saves the document to a determined savelocation. if no location is determined the document can't be saved
		 *
		 * @return nothing
		 */
		public LuaValue call() {
			documentHandler.save();
			return CoerceJavaToLua.coerce(true);
		}
	}


	private class LuaSaveas extends OneArgFunction {

		/**
		 * Saves the document to the savelocation
		 *
		 * @param savelocation the workingdirectory plus the savelocation is the location the document is saved to
		 * @return nothing
		 */
		public LuaValue call(LuaValue savelocation) {
			Path path = documentHandler.getWorkingDir().resolve(savelocation.checkjstring() + ".jc3");
			documentHandler.setSaveLocation(path);
			documentHandler.save();
			return CoerceJavaToLua.coerce(true);
		}
	}


	private class LuaConfig extends ZeroArgFunction {

		/**
		 * to get a list of the configuration
		 *
		 * @return names of the elements in the configuration
		 */
		public LuaValue call() {
			ArrayList<String> a = new ArrayList<>();
			documentHandler.getDocument().streamModulesRecursive().forEach(e -> a.add(e.getName()));

			return CoerceJavaToLua.coerce(a.toString());
		}
	}


	private class LuaGetTargetName extends ZeroArgFunction {

		/**
		 * @return name of the target hardware
		 */
		public LuaValue call() {
			return LuaValue.valueOf(documentHandler.getDocument().getTarget().getTargetDescription().getName());
		}
	}


	private class LuaContains extends OneArgFunction {

		/**
		 * looks if elemnst exists with the moduletype in the konfiguration
		 *
		 * @param moduletype the moduletype you want to know exists how many times. If it is NIL, it  is the currentelement.
		 * @return integer how often the moduletype exists in the configuration
		 */
		public LuaValue call(LuaValue moduletype) {
			String modname;
			if (moduletype == LuaValue.NIL) {
				modname = getCurrentElement().getModuleDescriptionName();
			} else {
				modname = moduletype.checkjstring();
			}
			return LuaValue.valueOf(documentHandler.getDocument().streamModulesRecursive().filter(e -> e.getModuleDescriptionName().equals(modname)).count());
		}
	}


	private class LuaGetParent extends OneArgFunction {

		/**
		 * @param module the element you want the parent from
		 * @return the Parent of the module
		 */
		public LuaValue call(LuaValue module) {
			IModuleConfiguration mod;
			if (module == NIL) {
				mod = getCurrentElement();
			} else {
				mod = (IModuleConfiguration) module.touserdata();
			}
			IModuleConfiguration child = documentHandler.getDocument().streamModulesRecursive().filter(e -> e.equals(mod)).findFirst().get();
			return CoerceJavaToLua.coerce(child.getParent());
		}
	}


	private class LuaGetModule extends ThreeArgFunction {

		/**
		 * returns the module that meet the requirments
		 *
		 * @param moduletype the moduletype of the element wich should be returned
		 * @param subsystem the subsystem the element is in, or the currentsubsystem if subsystem==NIL,
		 * or the Configuration if subsystem is a boolean
		 * @param number if this parameter is not give the first element is used
		 * @return the element with the moduletype in the subsystem with the given number
		 */
		public LuaValue call(LuaValue moduletype, LuaValue subsystem, LuaValue number) {
			Object[] config = documentHandler.getDocument().streamModulesRecursive().toArray();
			IModuleConfiguration subsys = (IModuleConfiguration) subsystem.touserdata();
			if (subsystem == LuaValue.NIL) {
				subsys = getCurrentSubsystem();
			} else {
				if (subsystem.isint()) {
					Object[] subsystems = documentHandler.getDocument().streamModulesRecursive().
						filter(mod -> mod.getModuleDescriptionName().equals("spartanmc")).toArray();
					subsys = (IModuleConfiguration) subsystems[subsystem.checkint() - 1];
				} else if (subsystem.isboolean() || subsys == null) {
					int num = 1;
					int currentnum = 1;
					if (number != NIL) {
						num = number.checkint();
					}
					for (int i = 0; i < config.length; i++) {
						IModuleConfiguration mod = (IModuleConfiguration) config[i];
						mod.getParent().equals(documentHandler.getDocument()); //TODO unsused
						if (mod.getModuleDescriptionName().equals(moduletype.toString()) && mod.getParent().equals(documentHandler.getDocument())) {
							if (currentnum == num) {
								return CoerceJavaToLua.coerce(mod);
							}
							currentnum++;
						}
					}
				}
			}
			int num = 1;
			int currentnum = 1;
			if (number != NIL) {
				num = number.checkint();
			}
			for (int i = 0; i < config.length; i++) {
				IModuleConfiguration mod = (IModuleConfiguration) config[i];
				if (mod.getModuleDescriptionName().equals(moduletype.toString()) && mod.getParent().equals(subsys)) {
					if (currentnum == num) {
						return CoerceJavaToLua.coerce(mod);
					}
					currentnum++;
				}
			}
			return NIL;
		}
	}


	private class LuaGet extends TwoArgFunction {

		/**
		 * makes the same as getModule only you can't choose the subsystem
		 *
		 * @param moduletype the moduletype of the element wich should be returned
		 * @param number Makes a list of the elements with the given type and return the element in the place of the list from the number
		 * @return returns the element from the configuration with the moduletyp and the place of the list with the number
		 */
		public LuaValue call(LuaValue moduletype, LuaValue number) {
			if (number == NIL) {
				number = LuaValue.valueOf(1);
			}
			if (moduletype == NIL) {
				return CoerceJavaToLua.coerce(getCurrentElement());
			}
			Object[] modules = documentHandler.getDocument().streamModulesRecursive().
				filter(mod -> mod.getModuleDescriptionName().equals(moduletype.tojstring())).toArray();
			try {
				Object module = modules[number.toint() - 1];
				return CoerceJavaToLua.coerce(module);
			} catch (Exception e) {
				logger.error("There is/are not {} {} in the configuration", number.toint(), moduletype.tojstring());
			}
			return NIL;
		}
	}


	private class LuaGetSubsystem extends OneArgFunction {

		/**
		 * takes all subsystem and returns the one in the place of the number in the list
		 * same as get("spartanmc", number)
		 *
		 * @param number the number of the subsystem
		 * @return the subsystem
		 */
		public LuaValue call(LuaValue number) {
			ISubsystemConfiguration<?> subsystem = documentHandler.getDocument().streamSubsystems().skip(number.toint() - 1).findFirst().get();
			return CoerceJavaToLua.coerce(subsystem);
		}
	}


	private class LuaAdd extends TwoArgFunction {

		/**
		 * Adds a module to the configuration,
		 *
		 * @param moduletype the type of the module you want to add
		 * @param subsystem the subsystem the module should be in, If subsystem == the currentSubsystem is used
		 * if subsystem is a boolean the module is added in the Configuration
		 * @return the added element
		 */
		public LuaValue call(LuaValue moduletype, LuaValue subsystem) {
			IModuleConfiguration zw = getCurrentElement();
			ISubsystemConfiguration<?> zwsubsys = getCurrentSubsystem();
			IConfigurationElement subsys;

			if (subsystem == LuaValue.NIL) {
				subsys = getCurrentSubsystem();
			} else {
				if (subsystem.isint()) {
					Object[] subsystems = documentHandler.getDocument().streamModulesRecursive().
						filter(mod -> mod.getModuleDescriptionName().equals("spartanmc")).toArray();
					subsys = (IModuleConfiguration) subsystems[subsystem.checkint() - 1];
				} else {
					subsys = (IModuleConfiguration) subsystem.touserdata();
				}
			}
			if (subsystem.isboolean() || moduletype.toString().equals("spartanmc") || subsys == null) {
				subsys = documentHandler.getDocument();
			}
			LuaValue erg = CoerceJavaToLua.coerce(documentHandler.addModule(subsys, library.getModuleDescription(moduletype.toString())));
			setCurrentElement(zw);
			setCurrentSubsystem(zwsubsys);
			return erg;

		}
	}


	private class LuaRemove extends OneArgFunction {

		/**
		 * Removes a Module in the subsystem from the element the skript is executed or from the currentSubsystem if subsystem =NIL
		 * or from the Configuration if subsystem is boolean
		 *
		 * @param module the module to remove
		 * @return nothing
		 */
		public LuaValue call(LuaValue module) {
			IModuleConfiguration<?> mod = (IModuleConfiguration<?>) module.touserdata();
			return LuaValue.valueOf(documentHandler.removeModule(mod.getParent(), mod));
		}
	}


	private class LuaConnectBus extends ThreeArgFunction {

		/**
		 * connects two busses
		 *
		 * @param bus the name of the busses
		 * @param module1 the module of the first bus
		 * @param module2 the module of the second bus, if NIL the currentSubsystem is used
		 * @return nothing
		 */
		public LuaValue call(LuaValue bus, LuaValue module1, LuaValue module2) { //TODO not modeled this way
			IModuleConfiguration<?> mod1;
			IModuleConfiguration<?> mod;
			if (module2 == LuaValue.NIL) {
				mod = getCurrentElement();
			} else {
				mod = (IModuleConfiguration<?>) module2.touserdata();
			}
			mod1 = (IModuleConfiguration<?>) module1.touserdata();
			Bus busDescription = null;
			for (Bus b : mod.getBusDescriptions()) {
				if (bus.toString().equals(b.getName())) {
					busDescription = b;
					break;
				}
			}
			documentHandler.getDocument().createBusConnection(mod1.getBusHandler(busDescription), mod.getBusHandler(busDescription)).connect();
			return LuaValue.valueOf(true);
		}
	}


	private class LuaConnectPort extends VarArgFunction {

		/**
		 * Connects two ports together
		 *
		 * param port1 the first port to be connected
		 * param module1 the module form port1
		 * param port2 the second port to be connected
		 * param moudle2 the module form port2, if module2 is NIL the currentElement is used
		 * @return nothing
		 */
		public LuaValue invoke(Varargs vargs) {
			LuaValue port1 = vargs.arg1();
			LuaValue module1 = vargs.arg(2);
			LuaValue port2 = vargs.arg(3);
			LuaValue module2 = vargs.arg(4);
			IModuleConfiguration mod2;
			if (module2 == NIL) {
				mod2 = getCurrentElement();
			} else {
				mod2 = (IModuleConfiguration) module2.touserdata();
			}
			IModuleConfiguration mod1 = (IModuleConfiguration) module1.touserdata();
			documentHandler.getDocument().createPortConnection
				(DocumentUtil.getPortHandler(mod1, port1.toString()), DocumentUtil.getPortHandler(mod2, port2.toString())).connect();

			return LuaValue.valueOf(true);

		}
	}


	private class LuaPartialConnection extends VarArgFunction {

		/**
		 * Connects two Ports, or one Port with one Pin. If the two Elements have not the same width.
		 * it can chooses a start and a width automatically if non is given. The start is the lowest it can be and the width is the maximum it can be.
		 *
		 * param port1 the first port to be connected
		 * param module1 the module from port1
		 * param port2 the second port or pin to be connected
		 * param module2 the module from port2, or NIL if port2 is a pin
		 * param start1 start of connection from port1
		 * param start2 start of connection from port2
		 * param width of the connection
		 * return nothing
		 */
		public LuaValue invoke(Varargs vargs) {
			LuaValue port1 = vargs.arg1();
			LuaValue module1 = vargs.arg(2);
			LuaValue port2 = vargs.arg(3);
			LuaValue module2 = vargs.arg(4);
			LuaValue start1 = vargs.arg(5);
			LuaValue start2 = vargs.arg(6);
			LuaValue width = vargs.arg(7);
			IModuleConfiguration mod1 = (IModuleConfiguration) module1.touserdata();
			IModuleConfiguration mod2 = (IModuleConfiguration) module2.touserdata();
			IPortHandler<? extends IPort> a = DocumentUtil.getPortHandler(mod1, port1.toString());
			int starta = start1.toint();
			int startb = start2.toint();
			int breite = width.toint();
			int automatisch = -1;
			if (module2 == NIL) { //port2==pin
				IPinHandler b = DocumentUtil.getPinHandler(documentHandler.getDocument(), port2.toString());
				breite = automatisch;
				if (start1 == NIL) {
					if (a.getDirection().name().equals("Input")) {
						starta = a.streamConnections().map(e -> e.getMyEnd(a) + 1).max(Comparator.naturalOrder()).orElse(0);
					} else {
						starta = 0;
					}
				}
				startb = automatisch;
				documentHandler.getDocument().createPartialConnection(a, starta, b, startb, breite).connect();
			} else {
				IPortHandler<? extends IPort> b = DocumentUtil.getPortHandler(mod2, port2.toString());
				if (width == NIL) {
					breite = automatisch;
				}
				if (start1 == NIL) {
					if (a.getConnectorWidth() > b.getConnectorWidth()) {
						a.getDirection().name();
						if (a.getDirection().name().equals("Input")) {
							starta = a.streamConnections().map(e -> e.getMyEnd(a) + 1).max(Comparator.naturalOrder()).orElse(0);
						} else {
							starta = 0;
						}
					} else {
						starta = automatisch;
					}
				}
				if (start2 == NIL) {
					if (a.getConnectorWidth() > b.getConnectorWidth()) {
						startb = automatisch;
					} else {
						if (b.getDirection().name().equals("Input")) {
							startb = b.streamConnections().map(e -> e.getMyEnd(b) + 1).max(Comparator.naturalOrder()).orElse(0);
						} else {
							startb = 0;
						}
					}
				}
				documentHandler.getDocument().createPartialConnection(a, starta, b, startb, breite).connect();

			}
			return LuaValue.valueOf(true);
		}
	}


	private class LuaConnectPin extends ThreeArgFunction {

		/**
		 * connects a Pin and a Port if the Port have the width one
		 *
		 * @param pin the pin to connect
		 * @param port the port to connect
		 * @param module the module form the port, or the currentElement if the module == NIL
		 * @return nothing
		 */
		public LuaValue call(LuaValue pin, LuaValue port, LuaValue module) {
			IModuleConfiguration mod;
			if (module == NIL) {
				mod = getCurrentElement();
			} else {
				mod = (IModuleConfiguration) module.touserdata();
			}

			documentHandler.getDocument().createPortConnection(DocumentUtil.getPortHandler(mod, port.toString()),
				DocumentUtil.getPinHandler(documentHandler.getDocument(), pin.toString())).connect();
			return LuaValue.valueOf(true);
		}
	}


	private class LuaDisconnect extends TwoArgFunction {

		/**
		 * detach a connection
		 *
		 * @param port the portname you don't want to a connection for anymore
		 * @param module the module that contains the port, or the currentElement if the module == NIL
		 * @return nothing
		 */
		public LuaValue call(LuaValue port, LuaValue module) {
			IModuleConfiguration mod;
			if (module == NIL) {
				mod = getCurrentElement();
			} else {
				mod = (IModuleConfiguration) module.touserdata();
			}
			DocumentUtil.getPortHandler(mod, port.toString()).disconnectAll();
			return LuaValue.valueOf(true);
		}
	}


	private class LuaDisconnectAll extends OneArgFunction {

		/**
		 * disconnects a module form everything
		 *
		 * @param module the module to disconnect, or the currentElement if the module == NIL
		 * @return nothing
		 */
		public LuaValue call(LuaValue module) {
			IModuleConfiguration<?> mod;
			if (module == NIL) {
				mod = getCurrentElement();
			} else {
				mod = (IModuleConfiguration) module.touserdata();
			}
			List<? extends IPortGroup> portGroups = mod.getPortGroupDescriptions();
			for (IPortGroup portGroup : portGroups) {
				for (IPort port : portGroup.getPorts()) {
					mod.getPortHandler(port).disconnectAll();
				}
			}
			List<Bus> list1 = mod.getBusDescriptions();
			for (Bus bus : list1) {
				mod.getBusHandler(bus).disconnectAll();
			}
			return LuaValue.valueOf(true);
		}
	}


	private class LuaSetParam extends ThreeArgFunction {

		/**
		 * sets a parameter to the given value
		 *
		 * @param param the parameter
		 * @param value the value the parameter should have
		 * @param module the module the parameter is from, or the currentElement if the module == NIL
		 * @return nothing
		 */
		public LuaValue call(LuaValue param, LuaValue value, LuaValue module) {
			IModuleConfiguration mod;
			if (module == LuaValue.NIL) {
				mod = getCurrentElement();
			} else {
				mod = (IModuleConfiguration) module.touserdata();
			}
			DocumentUtil.streamAllParameters(mod).filter(e -> e.getParameterDescription().getName()
				.equals(param.toString())).findFirst().get().setStoredValue(value.toString());
			return LuaValue.valueOf(true);
		}
	}


	/**
	 * Gets the parameter from the device or if no device is given form the currentElement
	 */
	private class LuaGetParam extends TwoArgFunction {

		/**
		 * returns the value of the parameter
		 * @param param the parameter
		 * @param module the module the parameter is from, or the currentElement if the module == NIL
		 * @return the value of the parameter form the module
		 */
		public LuaValue call(LuaValue param, LuaValue module) {
			IModuleConfiguration mod;
			if (module == LuaValue.NIL) {
				mod = getCurrentElement();
			} else {
				mod = (IModuleConfiguration) module.touserdata();
			}
			Optional<IParameterHandler> handler =
				DocumentUtil.streamAllParameters(mod).filter(e -> e.getParameterDescription().getName().equals(param.toString())).findFirst();
			return LuaValue.valueOf(handler.get().getValue().get());
		}
	}


	/**
	 * static class routingTable extends VarArgFunction {
	 * public LuaValue invoke() {
	 * Varargs routingTable;
	 * <p>
	 * for(int i =0; i< rtable ; i++){
	 * routingTable.arg(i)= rtable[i];
	 * }
	 * return routingTable;
	 * }
	 * }
	 */


	private class LuaScript extends OneArgFunction {

		/**
		 * to call luascripts
		 *
		 * @param lPath to the luascript
		 * @return nothing
		 */
		public LuaValue call(LuaValue lPath) {
			Path path = (Path) lPath.touserdata();
			URI uri = path.toUri();
			scripExecutor.accept(uri);
			return LuaValue.valueOf(true);
		}
	}


}

