/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.scripting.lua;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPluginEngine;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IScriptManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentExporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentImporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.scripting.ReplWriter;

import org.luaj.vm2.Globals;
import org.luaj.vm2.LuaValue;
import org.luaj.vm2.lib.jse.JsePlatform;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.InputStream;
import java.io.PrintStream;
import java.net.URI;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;

/**
 * Created by melanieb on 30.05.16.
 */
public class LuaScriptPlugin implements IPluginEngine, IScriptManager {

	private static final Logger logger = LoggerFactory.getLogger(LuaScriptPlugin.class);
	private static final String LUA_FILE_EXT = ".lua";

	private final Globals globals;
	private boolean autoScript = true;
	private Library library;
	private boolean autoFlushStdIo = false;
	private Consumer<String> events;
	private LuaInterface luaInterface;

	private LuaScriptPlugin() {
		globals = JsePlatform.standardGlobals();
	}

	public static InputStream getScriptFromResources(String name) {
		return LuaScriptPlugin.class.getResourceAsStream(name);
	}

	public static void executeScriptFromResources(Globals lua, String name) {
		lua.load(getScriptFromResources(name), "@" + name, "bt", lua).call(LuaValue.valueOf(name));
	}

	/**
	 * Method to call scripts from jconfig
	 *
	 * @param path path to the script
	 */
	@Override
	public void executeScript(URI path) {
		LuaValue chunk = globals.loadfile(path.toString());
		chunk.call(LuaValue.valueOf(path.toString()));

		if (autoFlushStdIo) {
			flushStdIo();
		}
	}

	@Override
	public Optional<Object> eval(String code) {
		LuaValue chunk = globals.load(code);
		LuaValue res = chunk.call(LuaValue.valueOf(code));

		if (autoFlushStdIo) {
			flushStdIo();
		}

		return Optional.ofNullable(res.touserdata());
	}

	@Override
	public int autocomplete(String input, int cursorPos, List<String> candidates) {
		return 0;
	}

	@Override
	public String getHelpMessage() {
		return null;
	}

	@Override
	public boolean supportsFile(Path scriptFile) {
		return scriptFile.getFileName().toString().endsWith(LUA_FILE_EXT);
	}

	@Override
	public void redirectStdIo(Consumer<String> stdOut, Consumer<String> stdErr, Consumer<String> events) {
		this.events = events;
		globals.STDOUT = new PrintStream(new ReplWriter(stdOut));
		globals.STDERR = new PrintStream(new ReplWriter(stdErr));
		autoFlushStdIo = true;
	}

	@Override
	public boolean isAutoScriptEnabled() {
		return autoScript;
	}

	@Override
	public void setAutoScriptEnabled(boolean enabled) {
		autoScript = enabled;
	}

	private void flushStdIo() {
		globals.STDOUT.flush();
		globals.STDERR.flush();
	}

	@Override
	public void onTargetChanged() {

	}

	@Override
	public void onConnectorChanged(IConfigurationElement<?, ?, ?> element, IConnectorHandler<?, ?> connector, ConnectorChangeType type, IConnection<?> connection) {

	}

	@Override
	public void onElementNameChanged(IConfigurationElement<?, ?, ?> element) {

	}

	@Override
	public void onAttributeChanged(IConfigurationElement<?, ?, ?> element, String attribute) {

	}

	@Override
	public void onElementDeclarationChanged(IConfigurationElement<?, ?, ?> element) {

	}

	@Override
	public void onFirmwareChanged(IMemoryModuleConfiguration element) {

	}

	@Override
	public void onParameterChanged(IConfigurationElement<?, ?, ?> element, IAttributeHandler handler) {
		if (autoScript) {
			IModuleConfiguration module = (IModuleConfiguration) element;
			Path path = module.getModuleDescription().getSourceFile().resolveSibling("parameterchanged.lua");//TODO remove hardcoded paths. See Library.java");
			if (Files.exists(path)) {
				String script = path.toString();
				setCurrentElement(module);
				setCurrentElement(module);
				if (module.getParent().getClass() == ISubsystemConfiguration.class) {
					setCurrentSubsystem((ISubsystemConfiguration<?>) module.getParent());
				} else {
					setCurrentSubsystem(null);
				}
				logger.info("execute " + module.getModuleDescriptionName() + " parameterchanged.lua script");
				try {
					LuaValue chunk = globals.loadfile(script);
					chunk.call(LuaValue.valueOf(script));
				} catch (Exception e) {
					logger.error("Luascript failed:", e);
				}
			}
		}
	}

	private void setCurrentElement(IModuleConfiguration module) {
		globals.set("currentElement", LuaValue.userdataOf(module));
	}

	@Override
	public void onModuleAdded(IModuleConfiguration module) {
		if (autoScript) {
			try {
				String prefix = module.getModuleDescription().getSourceFilePrefix().map(e -> e + "").orElse("");
				Path path = module.getModuleDescription().getSourceFile().resolveSibling(prefix + "add.lua");
				if (Files.exists(path)) {
					String script = path.toString();
					setCurrentElement(module);
					if (module.getParent().getClass() == ISubsystemConfiguration.class) {
						setCurrentSubsystem((ISubsystemConfiguration<?>) module.getParent());
					} else {
						setCurrentSubsystem(null);
					}
					logger.info("execute " + module.getModuleDescriptionName() + " add script");
					LuaValue chunk = globals.loadfile(script);
					chunk.call(LuaValue.valueOf(script));
				}
			} catch (Exception e) {
				logger.error("Luascript failed:", e);
			}
		}
	}

	private void setCurrentSubsystem(ISubsystemConfiguration<?> subsystem) {

	}

	@Override
	public void onModuleRemoved(IConfigurationElement<?, ?, ?> parent, IModuleConfiguration module) {
		if (autoScript) {
			try {
				String prefix = module.getModuleDescription().getSourceFilePrefix().map(e -> e + "").orElse("");
				Path path = module.getModuleDescription().getSourceFile().resolveSibling(prefix + "remove.lua");
				if (Files.exists(path)) {
					setCurrentElement(module);
					if (module.getParent().getClass() == ISubsystemConfiguration.class) {
						setCurrentSubsystem((ISubsystemConfiguration<?>) module.getParent());
					} else {
						setCurrentSubsystem(null);
					}
					String script = path.toString();
					logger.info("execute " + module.getModuleDescriptionName() + " remove script");
					LuaValue chunk = globals.loadfile(script);
					chunk.call(LuaValue.valueOf(script));
				}
			} catch (Exception e) {
				logger.error("Luascript failed:", e);
			}
		}
	}

	@Override
	public void onAttached(ILibJConfigProject project) {
		this.library = project.getApplication().getLibrary();

		luaInterface = new LuaInterface(library, project.getDocumentHandler(), globals, this::executeScript);
		executeScriptFromResources(globals, "makros/makro.lua");
	}

	@Override
	public void onOpenedDocument(IConfigurationDocument document) {

	}

	@Override
	public void onImportedDocument(IConfigurationDocument document, IDocumentImporter importer) {

	}

	@Override
	public void onCreatedDocument(IConfigurationDocument document) {
		if (autoScript) {
			document.addSubsystem(library.getSubsystemDescription("spartanmc"));
		}
	}

	@Override
	public boolean isModuleInUse(IModuleConfiguration module) {
		return false;
	}

	@Override
	public void onNoDocument() {

	}

	@Override
	public void onSaving(IConfigurationDocument document) {

	}

	@Override
	public void onExportingDocument(IConfigurationDocument document, IDocumentExporter exporter) {

	}

	@Override
	public void onBuilding(IConfigurationDocument document) {
		if (autoScript) {
			logger.info("execute onBuilding script");
			try {
				executeScriptFromResources(globals, "onBuilding.lua");
			} catch (Exception e) {
				logger.error("Luascript failed:", e);
			}
		}
	}

	@Override
	public boolean isValid(IModuleConfiguration<?> module, boolean report) {
		return true;
	}

	@Override
	public boolean isValid(IConfigurationDocument configurationDocument, boolean report) {
		return true;
	}

	@Override
	public Optional<IScriptManager> getScriptManager() {
		return Optional.of(this);
	}
}
