/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IFirmwareConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.incremental.IncrementalBuildCache;
import de.tu_darmstadt.rs.spartanmc.jconfig.handlers.DocumentHandler;
import de.tu_darmstadt.rs.util.BuildInfo;
import de.tu_darmstadt.rs.util.SpartanMCFirmwareUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.FileTime;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by Ramon on 07.02.2016.
 */
public class LibJConfigProject implements ILibJConfigProject {

	private static BuildInfo buildInfo;
	private final Logger logger = LoggerFactory.getLogger(LibJConfigProject.class);
	private final Path systemDir;//configurable?
	private final Path buildDir;
	private final IApplicationConfiguration application;
	private final Path projectDir;
	private final boolean noVersionCheck;
	private final IDocumentHandler documentHandler;
	private IncrementalBuildCache buildCache;

	public LibJConfigProject(Path projectDir, IApplicationConfiguration application, boolean noVersionCheck) {
		this.application = application;
		this.noVersionCheck = noVersionCheck;
		if (projectDir!=null) {
			this.projectDir = projectDir.normalize().toAbsolutePath().normalize();
			logger.info("Project directory is {}", this.projectDir);
			buildDir = this.projectDir.resolve(BUILD_DIR);
			systemDir = this.projectDir.resolve(SYSTEM_DIR);
		} else {
			this.projectDir = null;
			systemDir = null;
			buildDir = null;
		}
		documentHandler = new DocumentHandler(application, this);
	}

	public static BuildInfo getBuildInfo() {
		if (buildInfo == null) {
			buildInfo = new BuildInfo(LibJConfigProject.class);
		}
		return buildInfo;
	}

	@Override
	public Optional<Path> getLastConfigFile() {
		return getBuildCache().getLastBuiltConfig();
	}

	@Override
	public IApplicationConfiguration getApplication() {
		return application;
	}

	@Override
	public IDocumentHandler getDocumentHandler() {
		return documentHandler;
	}

	@Override
	public Path getProjectDir() {
		return projectDir;
	}

	@Override
	public Optional<Path> findNewestAutoSave() {
		try {
			Stream<Path> pathStream = Files.list(projectDir).filter(f -> !Files.isDirectory(f) && f.toString().endsWith(AUTOSAVE_EXT));
			return pathStream.sorted(Comparator.comparing(p -> {
				try {
					return Files.getLastModifiedTime(p);
				} catch (IOException e) {
					logger.error("Could not read lastModified of {}", p, e);
					return FileTime.fromMillis(0);
				}
			})).findAny();
		} catch (IOException e) {
			logger.error("Failed to search for autosave", e);
			return Optional.empty();
		}
	}

	private Stream<Path> listDirectoriesRecursive(Path location) throws IOException {
		return Files.list(location)
				.filter(Files::isDirectory)
				.flatMap(p -> {
					try {
						return Stream.concat(Stream.of(p), listDirectoriesRecursive(p));
					} catch (IOException e) {
						throw new RuntimeException(e);
					}
				});
	}

	@Override
	public List<String> getKnownFirmwares(Path location) {
		Path realLocation;
		if (location == null) {
			realLocation = projectDir;
		} else {
			realLocation = location;
		}
		try {
			return listDirectoriesRecursive(realLocation)
					.filter(SpartanMCFirmwareUtil::isValidFirmwareFolderStructure)
					.map(realLocation::relativize)
					.map(Path::toString)
					.sorted()
					.collect(Collectors.toList());
		} catch (IOException e) {
			logger.error("Could not read firmware folders of {}", location, e);
			return new ArrayList<>();
		}
	}

	@Override
	public List<String> getKnownFirmwareSubNames(Path location, String firmwareName) {
		if (location == null) {
			location = projectDir;
		}

		Path firmwareLocation = location.resolve(firmwareName);

		if (!Files.exists(firmwareLocation)) {
			return new ArrayList<>();
		}
		try {
			return Files.list(location).filter(Files::isDirectory).filter(IFirmwareConfiguration::isSubFirmwareFolder).filter(dir -> !dir.endsWith(FIRMWARE_SRC_DIR) || !dir.endsWith(FIRMWARE_INC_DIR)).map(Path::getFileName).map(Path::toString).collect(Collectors.toList());
		} catch (IOException e) {
			logger.error("Could not read sub-firmware folders of {}", firmwareLocation, e);
			return new ArrayList<>();
		}
	}

	@Override
	public String getLibJConfigVersion() {
		return LIBJCONFIG_VERSION;
	}

	/**
	 * The projects root directory.
	 *
	 * @return path to project
	 */
	@Override
	public Path getWorkingDir() {
		return projectDir;
	}

	/**
	 * Commonly the $Proj/build directory
	 *
	 * @return path to build dir
	 */
	@Override
	public Path getBuildDir() {
		return buildDir;
	}

	/**
	 * Commonly the $Proj/system directory
	 *
	 * @return path to jconfig build dir (system)
	 */
	@Override
	public Path getIntermediaryBuildDir() {
		return systemDir;
	}

	@Override
	public boolean isNoVersionCheck() {
		return noVersionCheck;
	}

	private IncrementalBuildCache getBuildCache() {
		if (buildCache == null) {
			buildCache = IncrementalBuildCache.getInstance(this);
		}
		return buildCache;
	}
}
