/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.incremental.IncrementalBuildCache;
import de.tu_darmstadt.rs.util.FileUtil;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.Collections;
import java.util.List;

/**
 * Created by Ramon on 17.05.2016.
 */
public class IncrementalBuildManager extends BaseBuildManager {

	private static final Logger logger = LoggerFactory.getLogger(IncrementalBuildManager.class);

	public IncrementalBuildManager(DocumentBuilder builder) {
		super(builder);
	}

	@Override
	public synchronized BuildResult buildRequired(IConfigurationDocument document) {
		return build(document, true);
	}

	@Override
	public synchronized BuildResult buildAll(IConfigurationDocument document) {
		return build(document, false);
	}

	@NotNull
	private BuildResult build(IConfigurationDocument document, boolean incremental) {
		long start = System.currentTimeMillis();
		Path outputDir = getOutputDir(document);
		List<? extends DocumentBuilder.IOutputGroup> outputGroups;
		try {
			outputGroups = builder.getOutputGroups(document);
		} catch (Exception e) {
			final BuildState state = new BuildState();
			state.addException(new BuildException("Failed listing groups",e));
			return new BuildResult(state,Collections.emptyList(),Collections.emptyList(),document.getProject().getDocumentHandler().getSaveLocation(),0);
		}

		noCleanupFiles.add(IncrementalBuildCache.getCacheFile(outputDir));
		List<Path> cleanedUpFiles = cleanupUnneededFiles(outputDir, outputGroups, noCleanupFiles);

		IncrementalBuildCache incrementalBuildCache = IncrementalBuildCache.getInstance(document.getProject());

		List<? extends DocumentBuilder.IOutputGroup> upToDate;
		if (incremental) {
			logger.info("Building incrementally");
			upToDate = incrementalBuildCache.checkUpToDate(outputGroups);

			outputGroups.removeAll(upToDate);
		} else {
			logger.info("Building All");
			upToDate = Collections.emptyList();
			incrementalBuildCache.removeOldFileStates();
		}

		BuildResult result = outputGroups.stream().collect(new BuildCollector(cleanedUpFiles, upToDate, document.getProject().getDocumentHandler().getSaveLocation(), start));

		incrementalBuildCache.add(result);

		incrementalBuildCache.store();

		if (result.wasSuccessful()) {
			logger.debug(buildMarker, "Successfully built required outputs of {} with {}", document.getName(), builder.getToolchainName());
		}

		FileUtil.removeEmptyDirectories(outputDir);

		return result;
	}
}
