/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.incremental;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager.BuildResult;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.incremental.jackson.JacksonPathKeySupport;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.incremental.jackson.PathKeyDeserializer;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigApi;
import de.tu_darmstadt.rs.util.BuildInfo;
import de.tu_darmstadt.rs.util.CollectionUtil;
import de.tu_darmstadt.rs.util.DigestUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonIdentityInfo;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonTypeInfo;
import com.fasterxml.jackson.annotation.ObjectIdGenerators;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.dataformat.smile.SmileFactory;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * Created by Ramon on 17.05.2016.
 */
@JsonAutoDetect(getterVisibility = JsonAutoDetect.Visibility.NONE, setterVisibility = JsonAutoDetect.Visibility.NONE, fieldVisibility = JsonAutoDetect.Visibility.NONE)
public class IncrementalBuildCache {

	public static final Marker incMarker = MarkerFactory.getMarker("Incremental");
	private static final Logger logger = LoggerFactory.getLogger(IncrementalBuildCache.class);
	private static final boolean PRETTY_PRINT = false;
	private static final boolean DEBUG_CHANGE_REASON = logger.isDebugEnabled(incMarker);
	private static IncrementalBuildCache cache;
	private static Path currentCacheDir;
	private static ObjectMapper objectMapper;
	@JsonProperty
	@JsonDeserialize(keyUsing = PathKeyDeserializer.class)
	private Map<Path, FileState> fileStates = new HashMap<>();
	@JsonProperty
	private Map<String, GroupBuildState> groupStates = new HashMap<>();
	@JsonProperty
	private Path lastConfig;

	//Jackson serializes paths using .toUri().toString(), which makes an absolute path. so we need to use strings...
	@JsonProperty
	private String lastConfigRelative;
	@JsonIgnore
	private Path outputDir;

	@JsonProperty
	private Map<String, BuildInfo> buildInfos = collectBuildInfo();

	private IncrementalBuildCache() {

	}

	public synchronized static IncrementalBuildCache getInstance(ILibJConfigProject project) {
		Path outputDir = project.getIntermediaryBuildDir();
		if (!outputDir.equals(currentCacheDir)) {
			boolean needNewCache = true;
			try {
				Path cacheFile = getCacheFile(outputDir);
				if (Files.exists(cacheFile)) {
					cache = getObjectMapper().readerFor(IncrementalBuildCache.class).readValue(cacheFile.toFile());
					Map<String, BuildInfo> newBuildInfos = collectBuildInfo();
					needNewCache = !cache.buildInfos.equals(newBuildInfos) && !project.isNoVersionCheck(); // discard if of other version
					if (needNewCache) {
						logger.warn("Found buildCache by different libJConfig build! Discarding...");
						logger.info("Reason: {}", CollectionUtil.debugEquality(cache.buildInfos, newBuildInfos));
					} else {
						cache.setOutputDir(outputDir);
						currentCacheDir = outputDir;
					}
				}
			} catch (IOException e) {
				logger.warn("Failed to read cache!. See trace for StackTrace! "+e.getMessage());
				logger.trace("Failed to read cache!", e);
			}
			if (needNewCache) {
				logger.debug(incMarker, "Instantiating new Cache");
				IncrementalBuildCache newCache = new IncrementalBuildCache();
				newCache.setOutputDir(outputDir);
				if (cache != null) {
					newCache.lastConfig = cache.lastConfig; // but keep last config
					newCache.lastConfigRelative = cache.lastConfigRelative;
				}
				currentCacheDir = outputDir;
				cache = newCache;
			}
		}
		return cache;
	}

	private static Map<String, BuildInfo> collectBuildInfo() {
		Map<String, BuildInfo> buildInfos = new HashMap<>();
		buildInfos.put("libdevxml", Library.getBuildInfo());
		buildInfos.put("libjconfig-api", LibJConfigApi.getBuildInfo());
		buildInfos.put("libjconfig", LibJConfigProject.getBuildInfo());

		return buildInfos;
	}

	public static Path getCacheFile(Path outputDir) {
		return outputDir.resolve("buildCache.smile");
	}

	private synchronized static ObjectMapper getObjectMapper() {
		if (objectMapper == null) {
			if (PRETTY_PRINT) {
				// Alternate mapper for Indented JSON
				objectMapper = new ObjectMapper();
				objectMapper.enable(SerializationFeature.INDENT_OUTPUT);
			} else {
				//Production Mapper for Binary JSON (Smile)
				objectMapper = new ObjectMapper(new SmileFactory());
			}


			objectMapper.registerModule(new JacksonPathKeySupport());
		}
		return objectMapper;
	}

	private static FileState readFileState(Path file) {
		return new FileState(System.currentTimeMillis(), null, readDigest(file));
	}

	private static byte[] readDigest(Path file) {
		try {
			if (Files.exists(file)) {
				return DigestUtil.getDigest(file).digest();
			} else {
				return null;
			}
		} catch (IOException e) {
			throw new RuntimeException("Failed to read fileState for " + file, e);
		}
	}

	public List<? extends DocumentBuilder.IOutputGroup> checkUpToDate(@NotNull List<? extends DocumentBuilder.IOutputGroup> outputGroups) {
		return outputGroups.stream().filter(this::checkUpToDate).collect(Collectors.toList());
	}

	private boolean checkUpToDate(DocumentBuilder.IOutputGroup group) {
		long startTime = System.currentTimeMillis();
		String groupId = getGroupId(group);
		String name = group.getName();

		if (group.getInputFiles() == null && group.getInputProperties() == null) {
			logger.info("Group {}: does not support incremental building. It has neither input files, nor input properties.", name);
			return false;
		}

		GroupBuildState groupState = groupStates.get(groupId);

		if (groupState == null) {
			logger.info(incMarker, "Group {}: No previous build", name);
			return false;
		}

		boolean outputFilesChanged = !StreamUtil.containsAll(groupState.outputs, group.getOutputFiles());
		if (outputFilesChanged) {
			logger.info(incMarker, "Group {}: List of outputs changed", name);
			if (DEBUG_CHANGE_REASON) {
				logger.debug(incMarker, "Changes: {}", CollectionUtil.debugEquality(groupState.outputs.keySet(), group.getOutputFiles()));
			}
			return false;
		}
		boolean inputFilesChanged = !StreamUtil.containsAll(groupState.inputs, group.getInputFiles());
		if (inputFilesChanged) {
			logger.info(incMarker, "Group {}: List of inputs changed", name);
			if (DEBUG_CHANGE_REASON) {
				logger.debug(incMarker, "Changes: {}", CollectionUtil.debugEquality(groupState.inputs.keySet(), group.getInputFiles()));
			}
			return false;
		}
		boolean propertiesChanged = !Objects.equals(groupState.properties,group.getInputProperties());
		if (propertiesChanged) {
			logger.info(incMarker, "Group {}: List of properties changed", name);
			if (DEBUG_CHANGE_REASON) {
				logger.debug(incMarker, "Changes: {}", CollectionUtil.debugEquality(groupState.properties, group.getInputProperties()));
			}
			return false;
		}

		boolean outputDoNotExist = !group.getOutputFiles().stream().allMatch(Files::exists);
		if (outputDoNotExist) {
			logger.info(incMarker, "Group {}: Output files do not exist", name);
			if (DEBUG_CHANGE_REASON) {
				logger.debug(incMarker, "Changes: {}", group.getOutputFiles().stream().filter(f -> !Files.exists(f)).collect(Collectors.toSet()));
			}
			return false;
		}

		try {
			boolean inputDataChanged = group.getInputFiles().stream().anyMatch(path -> {
				FileState lastBuiltWith = groupState.inputs.get(path);
				boolean res;
				if (lastBuiltWith == null) {
					res = Files.exists(path);
				} else {
					res = !isNewestFileStateIdentical(path, lastBuiltWith, startTime);
				}

				if (res) {
					logger.info(incMarker, "Group {}: Input file {} changed", name, path);
				}

				return res;
			});
			if (inputDataChanged) {
				return false;
			}
		} catch (Exception e) {
			logger.error(incMarker, "Failed to digest input files of {}", groupId, e);
			return false;
		}


		try {
			boolean outputNotExpected = group.getOutputFiles().stream().anyMatch(path -> {
				FileState lastBuilt = groupState.outputs.get(path);
				return !isNewestFileStateIdentical(path, lastBuilt, startTime);
			});
			if (outputNotExpected) {
				logger.warn(incMarker, "Group {}: Outputs exist but differ in digest", name);
				return false;
			}
		} catch (Exception e) {
			logger.error(incMarker, "Failed to digest input files of {}", groupId, e);
			return false;
		}

		logger.debug(incMarker, "Group {}: Up-To-Date!", name);
		return true;
	}

	private boolean isNewestFileStateIdentical(Path file, FileState lastBuiltWith, long oldest) {
		FileState cachedState = fileStates.get(file);
		if (cachedState == null) { // Not yet seen
			cachedState = readFileState(file);
			fileStates.put(file, cachedState);
		} else if (cachedState.lastSeen < oldest) { // seen, but too old => update if still equal, overwrite if changed
			byte[] fresh = readDigest(file);
			long time = System.currentTimeMillis();
			if (Arrays.equals(lastBuiltWith.lastHash, fresh)) {
				cachedState.lastSeen = time;
				return true;
			} else {
				cachedState = new FileState(time, null, fresh);
				fileStates.put(file, cachedState);
				return false;
			}
		} else if (cachedState == lastBuiltWith) { // faster check
			return true;
		}

		return Arrays.equals(lastBuiltWith.lastHash, cachedState.lastHash);
	}

	public void add(BuildResult result) {
		if (result.getConfigurationFile()!=null) {
			lastConfig = result.getConfigurationFile().toAbsolutePath();
			lastConfigRelative = outputDir.getParent().relativize(lastConfig).toString();
		} else {
			lastConfig = null;
			lastConfigRelative = null;
		}
		result.getCleanedUpFiles().forEach(fileStates::remove);
		result.getBuilt().forEach(this::addGroupState);
	}

	private String getGroupId(DocumentBuilder.IOutputGroup group) {
		return group.getClass().getCanonicalName() + ':' + group.getName();
	}

	//	private FileState getCurrentFileState(Path input, long oldest) {
	//		FileState fileState = fileStates.get(input);
	//		if (fileState == null || fileState.lastSeen < oldest) {
	//			fileState = readFileState(input);
	//			fileStates.put(input, fileState);
	//		}
	//		return fileState;
	//	}

	public void addGroupState(BuildResult.BuiltGroup group) {
		DocumentBuilder.IOutputGroup outputGroup = group.getOutputGroup();

		Iterator<MessageDigest> it = group.getDigests().iterator();

		Map<Path, FileState> outputs = outputGroup.getOutputFiles().stream().collect(Collectors.toMap(Function.identity(), output -> updateFileStateWritten(output, group, it.next())));
		Map<Path, FileState> inputs = outputGroup.getInputFiles().stream().collect(Collectors.toMap(Function.identity(), this::getCurrentFileState));
		GroupBuildState state = new GroupBuildState(outputs, outputGroup.getInputProperties(), inputs);

		groupStates.put(getGroupId(outputGroup), state);
	}

	private FileState getCurrentFileState(Path input) {
		return fileStates.computeIfAbsent(input, IncrementalBuildCache::readFileState);
	}

	private FileState updateFileStateWritten(Path output, BuildResult.BuiltGroup group, MessageDigest digest) {
		FileState newState = new FileState(group.getTimestamp(), digest.digest());
		fileStates.put(output, newState);
		return newState;
	}

	public void store() {
		try {
			getObjectMapper().writerFor(IncrementalBuildCache.class).writeValue(getCacheFile(outputDir).toFile(), this);
		} catch (IOException e) {
			logger.error("Failed to save buildCache", e);
		}
	}

	private void setOutputDir(Path outputDir) {
		this.outputDir = outputDir;
	}

	public Optional<Path> getLastBuiltConfig() {
		if (lastConfigRelative!=null) {
			logger.info("Relative path to config is "+lastConfigRelative);
			logger.info("Output dir is "+outputDir);
			final Path resolved = outputDir.getParent().resolve(lastConfigRelative);
			if (Files.exists(resolved)) {
				return Optional.of(resolved);
			} else {
				logger.info("but does not exist: "+resolved);
			}
		}
		if (lastConfig!=null && Files.exists(lastConfig)) {
			logger.info("using absolute path: "+lastConfig);
			return Optional.of(lastConfig);
		} else {
			return Optional.empty();
		}
	}

	public void removeOldFileStates() {
		fileStates.clear();
	}

	@JsonIdentityInfo(generator = ObjectIdGenerators.IntSequenceGenerator.class, property = "@id")
	private static class FileState {

		@JsonProperty
		private long lastSeen;
		@JsonProperty
		private Long lastBuilt;
		@JsonProperty
		private byte[] lastHash;

		private FileState() {

		}

		/**
		 * FileState for written files
		 *
		 * @param timestamp
		 * @param digest
		 */
		public FileState(long timestamp, byte[] digest) {
			lastSeen = lastBuilt = timestamp;
			lastHash = digest;
		}

		public FileState(long lastSeen, Long lastBuilt, byte[] lastHash) {
			this.lastSeen = lastSeen;
			this.lastBuilt = lastBuilt;
			this.lastHash = lastHash;
		}

		@Override
		public int hashCode() {
			int result = lastBuilt != null? lastBuilt.hashCode() : 0;
			result = 31 * result + Arrays.hashCode(lastHash);
			return result;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			FileState fileState = (FileState) o;

			if (lastBuilt != null? !lastBuilt.equals(fileState.lastBuilt) : fileState.lastBuilt != null) {
				return false;
			}
			return Arrays.equals(lastHash, fileState.lastHash);

		}

		public boolean fileEquals(FileState other) {
			return this == other || Arrays.equals(lastHash, other.lastHash);
		}
	}


	private static class GroupBuildState {

		@JsonProperty
		private Map<Path, FileState> outputs;
		@JsonProperty
		@JsonTypeInfo(use = JsonTypeInfo.Id.CLASS, property = "class", include = JsonTypeInfo.As.PROPERTY)
		private Map<String, ? extends Object> properties;
		@JsonProperty
		private Map<Path, FileState> inputs;

		private GroupBuildState() {

		}

		public GroupBuildState(Map<Path, FileState> outputs, Map<String, ? extends Object> properties, Map<Path, FileState> inputs) {
			this.outputs = outputs;
			this.properties = properties;
			this.inputs = inputs;
		}
	}
}
