/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.util.DigestUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.Locale;
import java.util.Objects;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by wenzel on 06.05.16.
 */
public abstract class VerilogOutputGroup extends SingleFileBaseOutputGroup<Object, PrintWriter> {

	private static final Logger logger = LoggerFactory.getLogger(VerilogOutputGroup.class);

	private PrintWriter writer;

	public VerilogOutputGroup(Path out) {
		super(new DigestUtil.PrintWriterSupplier(), out);
	}

	protected abstract String getHeader();

	protected void buildContent(PrintWriter writer) throws IOException {
		try {
			this.writer = writer;
			// Write Header
			writer.println(getHeader());

			buildContent();
		} finally {
			writer = null;
		}
	}

	protected abstract void buildContent();

	public <T> void writeModuleStart(String name, Stream<T> pins, Function<T, ConnectorDirection> getPinDirection, Function<T, String> getPinName) {

		String pinString = pins.map(pin -> String.format("\t%-8s wire \t %s", VerilogLang.getVerilogDirection(getPinDirection.apply(pin)), getPinName.apply(pin))).collect(Collectors.joining(",\n"));

		if (pinString.isEmpty()) {
			writer.format("module %s();\n\n", name);
		} else {
			writer.format("module %s(\n", name);
			writer.println(pinString);
			writer.println(");\n\n");
		}

	}

	public void writeModuleEnd() {
		writer.println("endmodule\n\n");
	}


	public void writeDeclaration(String type, String name, int width) {
		writer.format("%-6s %-8s %s;\n", type, VerilogLang.getVerilogBus(width), name);
	}

	public void writeNewline() {
		writer.println();
	}

	public void writeComment(String s) {
		writer.println("// " + s);
	}

	public void writeDefineIfNotSimulated(String name, String value) {
		writer.print("`ifndef SIMULATION\n`define " + name);
		if (value != null) {
			writer.print(" " + value);
		}
		writer.println("\n`endif");
	}

	public void writeDefineIfNotSimulated(IConfigurationDocument.Define define) {
		writeDefineIfNotSimulated(define.getName(), define.getValue());
	}

	public void writeLatticeProgramramInclude() {
		writer.println("`ifndef SIMULATION\n`include \"spartanmc_lse_programram.v\"\n`endif\n");
	}

	public void writeLatticePurGsr() {
		writer.println("PUR PUR_INST(.PUR(1'b1));\nGSR GSR_INST(.GSR(1'b1));\n");
	}


	public void writeVerilogAssign(String leftName, String leftBus, String rightName, String rightBus, boolean inverted) {
		writer.format("assign %-50s = %s%s%s;\n", String.format("%s%s", leftName, leftBus), inverted? "~" : "", rightName, rightBus);
	}

	public void writeModuleInstanceStart(String name) {
		writer.format("%s ", name);

	}

	public void writeDefaultNettype(String type) {
		writer.format("`default_nettype %s%n", type);
	}

	public <T> void writeModuleInstanceParams(Stream<? extends OutputParameter> params) {

		String paramsString = params.map(p -> VerilogLang.getModuleInstanceParam(p.getName(), p.getValue())).collect(Collectors.joining(",\n"));
		if (!paramsString.isEmpty()) {
			writer.println("#(");
			writer.print(paramsString);
			writer.print("\t\n) ");
		}
	}

	public <T> void writeModuleInstanceNameAndPorts(String instanceName, Stream<T> ports, Function<T, String> getPortName, Function<T, String> getConnectedName, Function<T, ConnectorDirection> getDirection) {
		writer.print(instanceName);
		if (ports != null) {
			writer.println(ports.map(port -> {
				String connected = getConnectedName.apply(port);
				String wireName = connected != null? connected : "/* N/C */";
				final ConnectorDirection direction = getDirection.apply(port);
				Objects.requireNonNull(direction, () -> "Direction of " + port + " is null");
				return String.format("\t/* %-6s */ \t\t .%-20s ( %-50s )", VerilogLang.getVerilogDirection(direction), getPortName.apply(port), wireName);
			}).collect(Collectors.joining(",\n", " (\n", "\n);\n")));
		} else {
			writer.println("();");
		}
	}

	public void writeModuleInstanceNameAndPorts(String instanceName, Stream<OutputPort> ports) {
		writeModuleInstanceNameAndPorts(
				instanceName,
				ports,
				OutputPort::getName,
				OutputPort::getConnectedName,
				OutputPort::getDirection
		);
	}

	public void writeTimescale(String unit, String precision) {
		writer.println(String.format("`timescale %s / %s ", unit, precision));
	}

	protected void writeClockGeneration(String portName, double delay) {
		writer.println(String.format("initial %s=0;", portName));
		//US locale to get decimal dot right
		writer.println(String.format(Locale.US, "always #(%.3f) %s=~%s;", delay, portName, portName));

	}

	protected void writeDefparam(String name, String value) {
		final String line = String.format("defparam %s = %s;", name, value);
		logger.info("writing defparam: {}", line);
		writer.println(line);
	}

	protected void writeIfdef(String macro) {
		writer.println("`ifdef " + macro);
	}

	protected void writeEndif() {
		writer.println("`endif");
	}

	protected void writeUselib(String library) {
		if (library == null) {
			writer.println("`uselib");
		} else {
			writer.println("`uselib lib=" + library);
		}
	}

	public static class OutputParameter {

		private final String name;
		private final String value;

		public OutputParameter(IParameterHandler param) {
			this.name = param.getName();
			this.value = VerilogLang.getResolvedParameterValue(param);
		}

		public OutputParameter(String name, String value) {
			this.name = name;
			this.value = value;
		}

		public String getName() {
			return name;
		}

		public String getValue() {
			return value;
		}
	}


	public static class OutputPort {

		public final String name;
		public final String connectedName;
		public final ConnectorDirection direction;

		public OutputPort(String name, String connectedName, ConnectorDirection direction) {
			this.name = name;
			this.connectedName = connectedName;
			this.direction = direction;
		}

		public String getName() {
			return name;
		}

		public String getConnectedName() {
			return connectedName;
		}

		public ConnectorDirection getDirection() {
			return direction;
		}
	}
}
