/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.ConstraintVisitor;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.KeepHierarchy;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.TimingIgnore;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.VendorConstraint;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.initram.MemorySource;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.QuartusBramBuilderFactory;
import de.tu_darmstadt.rs.spartanmc.initram.output.quartus.MifFilesOutput;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IToolchainAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.ConstraintReferenceAnalyzer;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.HdlFilesOutput;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.quartus.AlteraQuartusDocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryMap;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Output quartus settings file
 * contains way too much:
 * 	* Part settings
 * 	* Optimization settings
 * 	* Global Verilog Defines
 * 	* Pin constraints
 * 	* User Constraints
 * 	* List of verilog files
 */
public class AlteraQuartusSettingsBuilderProvider extends BaseOutputBuilderProvider {

	private static final Logger logger = LoggerFactory.getLogger(AlteraQuartusSettingsBuilderProvider.class);
	public static final String QSF_EXT = ".qsf";
	public static final String QSF_FILENAME = "spartanmc" + QSF_EXT;
	public static final String VENDOR = "quartus";

	public AlteraQuartusSettingsBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}


	@Override
	public List<QSfOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		List<QSfOutputGroup> groups = new ArrayList<>();

		groups.add(new QSfOutputGroup(buildProcess.getDocument().getProject().getIntermediaryBuildDir(), buildProcess));
		return groups;
	}


	protected static class QSfOutputGroup extends HdlFilesOutput<QsfOutput> {

		private final List<IPinHandler> pinHandlers;

		private final Path ucFile;
		private final CalculatedValueResult<String> ucData;
		private final IConfigurationDocument document;
		private final IToolchainAttributeHandler qsfParamHandler;
		private final Set<StreamUtil.Pair<String, String>> globalAssignments;


		private String replaceMinus(String s) {
			return s.replace('-','_');
		}

		/**
		 * Get global assignments, defining verilog macros, and other options
		 * @return name + value. not a map, since duplicate keys are allowed...
		 */
		private Set<StreamUtil.Pair<String, String>> getGlobalAssignments(Path buildDir) {
			Set<StreamUtil.Pair<String, String>> result = new HashSet<>();
			//disable parallel processing, needs license
			result.add(new StreamUtil.Pair<>("NUM_PARALLEL_PROCESSORS", "1"));
			//suppress warning about disabled and unlicensed parallel processing
			result.add(new StreamUtil.Pair<>("MESSAGE_DISABLE", "20028"));
			//suppress warning about missing license for LogicLock
			result.add(new StreamUtil.Pair<>("MESSAGE_DISABLE", "292013 \n"));
			result.add(new StreamUtil.Pair<>("SDC_FILE", '"'+buildDir.resolve("spartanmc_map.sdc").toString()+'"'));
			result.add(new StreamUtil.Pair<>("TOP_LEVEL_ENTITY", VerilogLang.getVerilogName(document.getName())));
			document.getLibrary().getModuleSourceTrees().stream().map(t -> t.toAbsolutePath().toString())
					.forEach(p -> result.add(new StreamUtil.Pair<>("SEARCH_PATH",'"'+p+'"')));
			document.getTarget().streamVerilogDefines(false)
					.forEach(def -> {
						result.add(new StreamUtil.Pair<>("VERILOG_MACRO",def.format(true,false,'=')));
					});

			final DeviceDescription deviceDescription = document.getTarget().getDeviceDescription();

			result.add(new StreamUtil.Pair<>("FAMILY", deviceDescription.getFamily()));
			result.add(new StreamUtil.Pair<>("DEVICE", deviceDescription.getName()));


			return result;
		}

		public QSfOutputGroup(Path systemDir, IBuildProcess cache) {
			super(cache.getDocument(),
					Collections.singletonList(new QsfOutput(systemDir, cache.getDocument(), cache.getGlobalAddressLayout().getMemoryMap())),
					systemDir
			);

			getHdlOutputs().forEach(output->output.setGroup(this));

			this.document = cache.getDocument();

			IConfigurationDocument.ITargetConfiguration target = document.getTarget();

			assert inputs.isEmpty();
			inputs = new HashSet<>();
			inputs.add(target.getDeviceDescription().getSourceFile());
			inputs.add(target.getTargetDescription().getSourceFile());

			//TODO remove check once we are sure everyone has migrated to the new format
			document.streamModulesRecursive().forEach(module -> {
				Path qsf = module.getModuleDescription().getSourceFile().resolveSibling(module.getModuleDescription().getName() + QSF_EXT);
				if (Files.exists(qsf)) {
					cache.getIssueSink().report(ConfigurationIssue.createConstraintsFileDeprecatedIssue(module, document.getTarget().getCurrentConfiguration(), qsf, "Module Validation"));
				}
			});



			IAttributeHandler ucDataHandler = document.getTarget().getAttributeHandler(IConfigurationDocument.ITargetConfiguration.ATTRIB_UC_DATA);
			IAttributeHandler ucFileHandler = document.getTarget().getAttributeHandler(IConfigurationDocument.ITargetConfiguration.ATTRIB_UC_FILE);

			//TODO find better way to get project dir
			Path doc = systemDir.resolveSibling("..");
			ucFile = (ucFileHandler.getValue().isPresent() && !ucFileHandler.getValue().get().isEmpty())? doc.resolveSibling(ucFileHandler.getValue().get()) : null;
			ucData = ucDataHandler.getValue();

			if (ucFile != null) {
				inputs.add(ucFile);

				if (!Files.exists(ucFile)) {
					cache.getIssueSink().report(ConfigurationIssue.createMissingFileIssue(ucFile, ucFileHandler, "Building constraints", ConfigurationIssue.IssueType.Warning));
				}
			}

			properties.put("uc data", ucData.toString());

			qsfParamHandler = DocumentUtil.getToolchainParamHandler(document.getTarget(), "QSF");
			properties.put("qsf toolchain param" , qsfParamHandler.getValue().orElseThrow());

			globalAssignments = getGlobalAssignments(cache.getDocument().getProject().getBuildDir());
			final String globalAssignmentsString = globalAssignments.stream().map(Object::toString).sorted().collect(Collectors.joining("\n"));
			properties.put("Global assignments", globalAssignmentsString);

			document.streamModulesRecursive()
					.filter(m -> m.getModuleDescription().getHdl().getConstraints() != null && !m.getModuleDescription().getHdl().getConstraints().isEmpty())
					.forEach(m -> properties.put(m.getGlobalName(), ConstraintReferenceAnalyzer.makePropertyFromConstraints(m, VENDOR, false)));

			pinHandlers = DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).peek(p -> {
				properties.put(p.getPortName(), AlteraQuartusDocumentBuilder.getDefiningValues(p));
			}).collect(Collectors.toList());
		}


	}

	private static ConstraintVisitor<StreamUtil.Pair<IModuleConfiguration<?>,Path>, String> constraintVisitor = new ConstraintVisitor<StreamUtil.Pair<IModuleConfiguration<?>,Path>, String>() {


		Map<IModuleConfiguration<?>, Integer> ignoreIndices = new HashMap<>();

		@Override
		public String visitVendorConstraint(VendorConstraint constraint, StreamUtil.Pair<IModuleConfiguration<?>,Path> userData, String vendor, Boolean timing) {
			return constraint.getQuartusQsfContent().resolve(userData.getA(), VerilogLang::getVerilogName, userData.getB()).orElseThrow()
					//Remove indentation from multiline constraints
					.replaceAll("\\s*\n\\s*", "\n");
		}


		@Override
		public String visitTimingIgnore(TimingIgnore constraint, StreamUtil.Pair<IModuleConfiguration<?>,Path> userData) {
			StringBuilder result = new StringBuilder();
			return result.toString();
		}

		@Override
		public String visitKeepHierarchy(KeepHierarchy keepHierarchy, StreamUtil.Pair<IModuleConfiguration<?>,Path> userData) {
			String settingStr;
			switch (keepHierarchy.getSetting()) {
				case TRUE:
					settingStr = "on";
					break;
				case FALSE:
					settingStr = "off";
					break;
				case SOFT:
					settingStr = "off";
					break;
				default:
					throw new RuntimeException("Invalid value: " + keepHierarchy.getSetting());
			}

			return String.format("# no direct equivalent for *keep hierarchy* on quartus via constraints file%n set_instance_assignment -name PRESERVE_REGISTER -to %s \"%s\"%n", keepHierarchy.getInstance().resolve(userData.getA(), VerilogLang::getVerilogName, userData.getB()).orElseThrow(), settingStr);
		}
	};


	private static class QsfOutput extends HdlFilesOutput.HdlOutput {

		private final Path systemDir;
		private final IConfigurationDocument document;
		private final MemoryMap memoryMap;
		private QSfOutputGroup group = null;

		public QsfOutput(Path systemDir, IConfigurationDocument document, MemoryMap memoryMap) {
			super(
					systemDir.resolve(QSF_FILENAME),
					SourceRef.FileType.SYNTHESIS
			);
			this.systemDir = systemDir;
			this.document = document;
			this.memoryMap = memoryMap;
		}

		public void setGroup(QSfOutputGroup group) {
			this.group = group;
		}

		private void writeGlobalAssignments(PrintWriter writer) {
			writer.println("# General settings");

			group.globalAssignments.stream().sorted(StreamUtil.Pair.comparator())
					.forEach(pair -> {
						writer.println("set_global_assignment -name "+pair.getA()+" "+pair.getB());
					});
		}

		private void writeToolchainParameters(PrintWriter writer) {

			writer.println("# Toolchain parameter ");
			final String paramValue = group.qsfParamHandler.getValue().orElseThrow();
			writer.println(paramValue.replaceAll("\\s*\n\\s*","\n")); //remove leading/trailing whitespace

		}

		private void writeLocConstraint(PrintWriter writer, IPinHandler pinHandler) {

			writer.format("set_location_assignment -to %s %s\n", VerilogLang.getPinName(pinHandler), pinHandler.getPinDescription().getFpgaPinName());

			// IO-Standard
			String value = pinHandler.getIoStd();
			if (value != null) {
				writer.format("set_instance_assignment -name IO_STANDARD -to %s \"%s\"\n", VerilogLang.getPinName(pinHandler), value);
			}

			// Drive
			value = pinHandler.getDrive();
			if (value != null) {
				writer.format("set_instance_assignment -name CURRENT_STRENGTH_NEW %s -to %s\n", value, VerilogLang.getPinName(pinHandler) );
			}
		}

		private void writePinConstraints(PrintWriter writer) {
			writer.println("# Pins");

			group.pinHandlers.forEach(ph -> {
				writeLocConstraint(writer, ph);
			});
		}

		private void writeModuleConstraints(PrintWriter writer) {
			writer.println("# Module Constraints");

			document.streamModulesRecursive()
					.filter(m -> m.getModuleDescription().getHdl().getConstraints() != null && !m.getModuleDescription().getHdl().getConstraints().isEmpty())
					.forEach(m -> {
						writer.println();
						writer.println("# Constraints for " + m.getName());
						m.getModuleDescription().getHdl().getConstraints().forEach(constraint -> writer.println(constraint.visitIfEnabled(m,constraintVisitor, new StreamUtil.Pair<>(m,m.getModuleDescription().getSourceFile().getParent()), VENDOR, false)));
					});
		}

		private void writeUserConstraints(PrintWriter writer) throws IOException {
			writer.println("# User Constraints");
			if (group.ucFile != null) {
				writer.println();
				writer.println("# User specified file");
				writer.println("# " + group.ucFile);
				if (Files.exists(group.ucFile)) {
					Files.lines(group.ucFile).forEach(writer::println);
				} else {
					writer.println("# File not found.");
				}
			}

			if (group.ucData != null && group.ucData.isPresent() && !group.ucData.get().isEmpty()) {
				writer.println();
				writer.println("# User specified data");
				writer.println(group.ucData.get());
			}
		}

		@Override
		public void writeHead(PrintWriter writer) {
			try {
				writeGlobalAssignments(writer);
				writeToolchainParameters(writer);
				writePinConstraints(writer);
				writeModuleConstraints(writer);
				writeUserConstraints(writer);
				writeMifFilenames(writer);
				writer.println("# Source Files");
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}

		private void writeMifFilenames(PrintWriter writer) {
			writer.println("# Memory Initialization");
			QuartusBramBuilderFactory bramBuilderFactory = new QuartusBramBuilderFactory();
			MemorySource memorySource = new MemorySource(memoryMap, systemDir.getParent(), bramBuilderFactory);
			MifFilesOutput mifOutput = new MifFilesOutput(null);

			memorySource.streamAllInitializableBramNames()
					.forEach((String bramName) -> {
						final String instName = bramName.replace('/', '|');
						final String fileName = mifOutput.getBramFilename(bramName);
						writer.println("set_parameter -name init_file "+ fileName +" -to \""+ instName +"\"");
					});
		}

		@Override
		public void writeWork(PrintWriter writer, SourceRef ref) {
			String type = HdlFilesOutput.commandByFileExt(ref, "set_global_assignment -name VERILOG_FILE", "set_global_assignment -name VHDL_FILE");
			String path = ref.getPath().toString();

			writer.format("%s \"%s\"\n", type, path);
		}

		@Override
		public void writeFoot(PrintWriter writer) {

		}
	}

}

