/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider;

import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.DynamicStructureModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.DynamicModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.DigestUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Created by gregork on 07.09.16.
 */
public class LegacyCoregenXcoBuilderProvider extends BaseOutputBuilderProvider {

	private static final Logger logger = LoggerFactory.getLogger(LegacyCoregenXcoBuilderProvider.class);

	public LegacyCoregenXcoBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}

	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		IConfigurationDocument document = buildProcess.getDocument();
		Path systemDir = document.getProject().getIntermediaryBuildDir();
		return document.streamModulesRecursive().filter(m -> m instanceof DynamicModuleConfiguration)
				.map(m -> new XcoOutput((DynamicModuleConfiguration) m, systemDir)).collect(Collectors.toList());
	}

	private static class XcoOutput extends SingleFileBaseOutputGroup<Object, PrintWriter> {

		private final DynamicModuleConfiguration moduleConfiguration;

		public XcoOutput(DynamicModuleConfiguration moduleConfiguration, Path systemDir) {
			super(new DigestUtil.PrintWriterSupplier(), systemDir.resolve(moduleConfiguration.getToplevel() + ".xco"));
			inputs = Collections.singleton(moduleConfiguration.getModuleDescription().getSourceFile());

			properties = new HashMap<>();
			properties.putAll(DocumentUtil.collectParamProps(DocumentUtil.streamAllParameters(moduleConfiguration)));

			this.moduleConfiguration = moduleConfiguration;
		}

		@Override
		public String getName() {
			return super.getName() + ":" + moduleConfiguration.getGlobalName();
		}

		@Override
		protected void buildContent(PrintWriter writer) throws IOException {
			// TODO: header
			DynamicStructureModuleDescription.Coregen coregen = moduleConfiguration.getDynamicModuleDescription().getLegacyCoregen();
			writer.println("SELECT " + coregen.getSelect().getValue());

			for (DynamicStructureModuleDescription.CSet cSet : coregen.getcSets()) {
				try {
					if (cSet.getValue().getContent() == null) {
						throw new NullPointerException("No content in cset: " + cSet);
					}
					String value = cSet.getValue()
							.resolve(moduleConfiguration, VerilogLang::getVerilogName, moduleConfiguration.getDynamicModuleDescription().getSourceFile().getParent())
							.orElse(cSet.getDefaultVal());
					writer.printf("CSET %s=%s%n", cSet.getName(), value);
				} catch (NullPointerException e) {
					logger.error("Error for cset {}: {}", cSet.getName(), cSet.getValue());
					throw e;
				}
			}
			writer.println("GENERATE");
		}
	}

}
