/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.ConstraintVisitor;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.KeepHierarchy;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.MixedParamRefs;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.TimingIgnore;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.VendorConstraint;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.ConstraintReferenceAnalyzer;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.xilinx.XilinxIseDocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.HeaderGen;
import de.tu_darmstadt.rs.util.DigestUtil;
import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.UnitUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import java.util.stream.Collectors;

/**
 * Created to handle constraints like the previous jconfig did (concatenating all ucf files of modules + stuff from IOPIN Config)
 * Should be remodelled in a toolchain independent way //TODO
 * Created by ramon on 01.04.2016.
 */
public class XilinxIseConstraintsBuilderProvider extends BaseOutputBuilderProvider {

	private static final Logger logger = LoggerFactory.getLogger(XilinxIseConstraintsBuilderProvider.class);
	public static final String UCF_EXT = ".ucf";
	public static final String JCONFIG_UCF = "jconfig" + UCF_EXT;
	public static final String VENDOR = "xilinx";

	public XilinxIseConstraintsBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}

	public static Path getConstraintsFile(Path systemDir) {
		return systemDir.resolve(JCONFIG_UCF);
	}

	@Override
	public List<UcfOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		List<UcfOutputGroup> groups = new ArrayList<>();
		IConfigurationDocument document = buildProcess.getDocument();

		groups.add(new UcfOutputGroup(document.getProject().getIntermediaryBuildDir(), document, buildProcess));
		return groups;
	}


	protected static class UcfOutputGroup extends SingleFileBaseOutputGroup<Object, PrintWriter> {

		private static final String UCF_LOC_PARAM = " | %s = %s";
		private static final String UCF_LOC_VALUE = " | %2$s";
		private final List<IPinHandler> pinHandlers;

		private final Path ucFile;
		private final CalculatedValueResult<String> ucData;
		private final IConfigurationDocument document;

		public UcfOutputGroup(Path systemDir, IConfigurationDocument document, IBuildProcess cache) {
			super(new DigestUtil.PrintWriterSupplier(), systemDir.resolve(JCONFIG_UCF));
			inputs = new HashSet<>();
			properties = new HashMap<>();

			this.document = document;

			IConfigurationDocument.ITargetConfiguration target = document.getTarget();

			inputs.add(target.getDeviceDescription().getSourceFile());
			inputs.add(target.getTargetDescription().getSourceFile());

			//TODO remove check once we are sure everyone has migrated to the new format
			document.streamModulesRecursive().forEach(module -> {
				Path ucf = module.getModuleDescription().getSourceFile().resolveSibling(module.getModuleDescription().getName() + UCF_EXT);
				if (Files.exists(ucf)) {
					cache.getIssueSink().report(ConfigurationIssue.createConstraintsFileDeprecatedIssue(module, document.getTarget().getCurrentConfiguration(), ucf, "Module Validation"));
				}
			});


			IAttributeHandler ucDataHandler = document.getTarget().getAttributeHandler(IConfigurationDocument.ITargetConfiguration.ATTRIB_UC_DATA);
			IAttributeHandler ucFileHandler = document.getTarget().getAttributeHandler(IConfigurationDocument.ITargetConfiguration.ATTRIB_UC_FILE);

			//TODO find better way to get project dir
			Path doc = systemDir.resolveSibling("..");
			ucFile = (ucFileHandler.getValue().isPresent() && !ucFileHandler.getValue().get().isEmpty())? doc.resolveSibling(ucFileHandler.getValue().get()) : null;
			ucData = ucDataHandler.getValue();

			if (ucFile != null) {
				//TODO add to properties because name may change???
				inputs.add(ucFile);

				if (!Files.exists(ucFile)) {
					cache.getIssueSink().report(ConfigurationIssue.createMissingFileIssue(ucFile, ucFileHandler, "Building constraints", ConfigurationIssue.IssueType.Warning));
				}
			}

			properties.put("uc data", ucData.toString());

			document.streamModulesRecursive()
					.filter(m -> m.getModuleDescription().getHdl().getConstraints() != null && !m.getModuleDescription().getHdl().getConstraints().isEmpty())
					.forEach(m -> properties.put(m.getGlobalName(), ConstraintReferenceAnalyzer.makePropertyFromConstraints(m, VENDOR,false)));

			pinHandlers = DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).peek(p -> {
				properties.put(p.getPortName(), XilinxIseDocumentBuilder.getDefiningValues(p));
			}).collect(Collectors.toList());
		}

		private static Consumer<MixedParamRefs> resolveMixedParamRef(StreamUtil.Pair<IModuleConfiguration<?>, Path> resolver, String linePrefix, String lineSuffix, StringBuilder resultBuilder) {
			return (pr) -> resultBuilder.append(linePrefix).append(pr.resolve(resolver.getA(), VerilogLang::getVerilogName, resolver.getB()).orElseThrow()).append(lineSuffix).append('\n');
		}

		private static ConstraintVisitor<StreamUtil.Pair<IModuleConfiguration<?>,Path>, String> constraintVisitor = new ConstraintVisitor<StreamUtil.Pair<IModuleConfiguration<?>,Path>, String>() {

			private Map<String, String> timingNames = new HashMap<>();
			private Set<String> existingNames = new HashSet<>();

			/**
			 * Create or get a timing name for a net
			 * @param netName The net's name. May contain wildcards
			 * @param resultBuilder UCF Code to assign a timing name will be appended to this.
			 * @return The timing name
			 */
			private String getTimingName(String netName, StringBuilder resultBuilder) {
				String res = timingNames.get(netName);
				if (res == null) {
					res = "TS_" + VerilogLang.getVerilogName(netName);
					//Name conflict?
					if (!existingNames.add(res)) {
						//Add number at end to avoid conflict
						int i = 0;
						do {
							res = "TS_" + VerilogLang.getVerilogName(netName) + "_" + i++;
						} while (!existingNames.add(res));
					}

					resultBuilder.append(String.format("NET \"%s\" TNM_NET = \"%s\";%n", netName, res));

					timingNames.put(netName, res);
				}
				return res;
			}

			Map<IModuleConfiguration<?>, Integer> ignoreIndices = new HashMap<>();

			private int getIgnoreIndex(IModuleConfiguration<?> module) {
				int idx = ignoreIndices.getOrDefault(module, 0);
				ignoreIndices.put(module, idx + 1);
				return idx;
			}

			@Override
			public String visitVendorConstraint(VendorConstraint constraint, StreamUtil.Pair<IModuleConfiguration<?>,Path> userData, String vendor, Boolean timing) {
				return constraint.getXilinxContent().resolve(userData.getA(), VerilogLang::getVerilogName, userData.getB()).orElseThrow()
						//Remove indentation from multiline constraints
						.replaceAll("\\s*\n\\s*", "\n");
			}


			@Override
			public String visitTimingIgnore(TimingIgnore constraint, StreamUtil.Pair<IModuleConfiguration<?>,Path> userData) {
				StringBuilder result = new StringBuilder();

				constraint.getNets().forEach(resolveMixedParamRef(userData, "NET \"", "\" TIG;", result));
				constraint.getInstances().forEach(resolveMixedParamRef(userData, "INST \"", "\" TIG;", result));

				constraint.getIgnoreClockCrossings().forEach(icc -> {
					String from = icc.getFrom().resolve(userData.getA(), VerilogLang::getVerilogName, null).orElseThrow();
					String to = icc.getTo().resolve(userData.getA(), VerilogLang::getVerilogName, null).orElseThrow();

					String tnmFrom = getTimingName(from, result);
					String tnmTo = getTimingName(to, result);

					String name = "TS_ignore_" + userData.getA().getName() + "_" + getIgnoreIndex(userData.getA());

					result.append(String.format("TIMESPEC \"%s\" = FROM \"%s\" TO \"%s\" TIG;%n", name, tnmFrom, tnmTo));
					if (icc.isBidirectional()) {
						result.append(String.format("TIMESPEC \"%s\" = FROM \"%s\" TO \"%s\" TIG;%n", name + "_R", tnmTo, tnmFrom));
					}
				});


				return result.toString();
			}

			@Override
			public String visitKeepHierarchy(KeepHierarchy keepHierarchy, StreamUtil.Pair<IModuleConfiguration<?>,Path> userData) {
				String settingStr;
				switch (keepHierarchy.getSetting()) {
					case TRUE:
						settingStr = "true";
						break;
					case FALSE:
						settingStr = "false";
						break;
					case SOFT:
						settingStr = "soft";
						break;
					default:
						throw new RuntimeException("Invalid value: " + keepHierarchy.getSetting());
				}
				return String.format("INST \"%s\" KEEP_HIERARCHY=%s;%n", keepHierarchy.getInstance().resolve(userData.getA(), VerilogLang::getVerilogName, userData.getB()).orElseThrow(), settingStr);
			}
		};

		@Override
		public void buildContent(PrintWriter writer) throws IOException {
			new HeaderGen("#").getHeader("Xilinx ISE Constraints File").write(writer, output);

			writer.println();

			pinHandlers.forEach(ph -> {
				writeLocConstraint(writer, ph);
				if (ph.getFrequency().isPresent()) {
					writeTimingConstraint(writer, ph);
				}
			});

			document.streamModulesRecursive()
					.filter(m -> m.getModuleDescription().getHdl().getConstraints() != null && !m.getModuleDescription().getHdl().getConstraints().isEmpty())
					.forEach(m -> {
						writer.println();
						writer.println("# Constraints for " + m.getName());
						m.getModuleDescription().getHdl().getConstraints().forEach(constraint -> writer.println(constraint.visitIfEnabled(m,constraintVisitor, new StreamUtil.Pair<>(m,m.getModuleDescription().getSourceFile().getParent()), VENDOR, false)));
					});

			if (ucFile != null) {
				writer.println();
				writer.println("# User specified file");
				writer.println("# " + ucFile);
				if (Files.exists(ucFile)) {
					Files.lines(ucFile).forEach(writer::println);
				} else {
					writer.println("# File not found.");
				}
			}

			if (ucData != null && ucData.isPresent() && !ucData.get().isEmpty()) {
				writer.println();
				writer.println("# User specified data");
				writer.println(ucData.get());
			}


		}

		private void writeTimingConstraint(PrintWriter writer, IPinHandler pinHandler) {
			String netName = pinHandler.getPortName() + "_TNM";
			writer.format("NET \"%s\" TNM_NET = \"%s\";\n", VerilogLang.getPinName(pinHandler), netName);
			writer.format("TIMESPEC \"TS_%1$s\" = PERIOD \"%1$s\" %2$s HIGH 50%%;\n", netName, UnitUtil.getFreqLabel(pinHandler.getFrequency().orElseThrow(x -> new RuntimeException("absent for " + pinHandler))));
		}

		private void writeLocConstraint(PrintWriter writer, IPinHandler pinHandler) {
			writer.format("NET \"%s\" LOC = \"%s\"%s;\n", VerilogLang.getPinName(pinHandler), pinHandler.getPinDescription().getFpgaPinName(), getIoOptions(pinHandler));
		}

		private String getIoOptions(IPinHandler pinHandler) {
			StringBuilder sb = new StringBuilder();
			appendOnlyIfNotNull(sb, UCF_LOC_PARAM, "IOSTANDARD", pinHandler.getIoStd());
			appendOnlyIfNotNull(sb, UCF_LOC_PARAM, "SLEW", pinHandler.getSlewRate());
			appendOnlyIfNotNull(sb, UCF_LOC_PARAM, "DRIVE", pinHandler.getDrive());
			appendOnlyIfNotNull(sb, UCF_LOC_VALUE, "", XilinxIseDocumentBuilder.getUcfName(pinHandler.getTermination()));
			appendOnlyIfNotNull(sb, UCF_LOC_VALUE, "", pinHandler.getCustomUserConstraints());

			return sb.toString();
		}

		private void appendOnlyIfNotNull(StringBuilder sb, String format, String paramName, String value) {
			if (value != null) {
				sb.append(String.format(format, paramName, value));
			}
		}

	}
}

