/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider;

import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.HdlFilesOutput;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.DigestUtil;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import static java.util.stream.Collectors.toSet;

/**
 * Create the ises many config files likes XST project file (spartanmc.xst), .prj etc.
 */
public class XilinxIseProjectBuilderProvider extends BaseOutputBuilderProvider {

	public static final String SPARTANMC_XST = "spartanmc.xst";
	public static final String SPARTANMC_PRJ = "spartanmc.prj";
	public static final String HDL_SOURCES_MK = "hdl_sources.mk";
	public static final String SPARTANMC_SIM_WORKLIB = "spartanmc_worklib.fdo";
	public static final String SPARTANMC_SIM_WORKLIB_TIMESIM = "spartanmc_worklib_timesim.fdo";

	public XilinxIseProjectBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}

	private static Path getBuildRelPrj(Path systemDir) {
		return Paths.get("..").resolve(systemDir).resolve(SPARTANMC_PRJ);
	}

	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		List<DocumentBuilder.IOutputGroup> groups = new ArrayList<>();
		IConfigurationDocument document = buildProcess.getDocument();

		Path systemDir = document.getProject().getIntermediaryBuildDir();
		groups.add(new XilinxXstOutputGroup(systemDir, document));

		groups.add(new HdlFilesOutput<>(document,
				Arrays.asList(
						getPrjProjectOutput(systemDir, document),
						HdlFilesOutput.getMakeOutput(systemDir.resolve(HDL_SOURCES_MK)),
						HdlFilesOutput.getVsimOutput(systemDir.resolve(SPARTANMC_SIM_WORKLIB), SourceRef.FileType.FUNCTIONAL_SIM, document),
						HdlFilesOutput.getVsimOutput(systemDir.resolve(SPARTANMC_SIM_WORKLIB_TIMESIM), SourceRef.FileType.TIMESIM, document)
				),
				systemDir));

		return groups;
	}

	private HdlFilesOutput.DefaultHdlOutput getPrjProjectOutput(Path systemDir, IConfigurationDocument document) {

		return new HdlFilesOutput.DefaultHdlOutput(
				systemDir.resolve(SPARTANMC_PRJ),
				null,
				XilinxIseProjectBuilderProvider::writeWork,
				null,
				SourceRef.FileType.SYNTHESIS
		);
	}


	private static void writeWork(PrintWriter writer, SourceRef ref) {
		String type = HdlFilesOutput.commandByFileExt(ref, "verilog", "vhdl");
		String path = ref.getPath().toString();

		String library = ref.getLibrary();
		if (library == null) {
			library = "work";
		}

		writer.format("%s %s %s\n", type, library, path);
	}

	protected static class XilinxXstOutputGroup extends SingleFileBaseOutputGroup<Object, PrintWriter> {

		private final Path systemDir;
		private final IConfigurationDocument document;
		private final Set<Path> hdlSourceTrees;
		private final String targetOptions;
		private final Optional<String> defines;

		public XilinxXstOutputGroup(Path systemDir, IConfigurationDocument document) {
			super(new DigestUtil.PrintWriterSupplier(), systemDir.resolve(SPARTANMC_XST));
			this.systemDir = systemDir;
			this.document = document;
			inputs = new HashSet<>();
			properties = new HashMap<>();
			hdlSourceTrees = document.getLibrary().getModuleSourceTrees().stream().peek(tree -> properties.put(tree.toString(), tree.toString())).collect(toSet());
			properties.put("document", document.getName());
			targetOptions = DocumentUtil.streamTargetParamHandlers(document).filter(p -> p.getName().equals("SYNTHESIS"))
					.findAny()
					.orElseThrow(() -> new IllegalStateException("Did not find synthesis command line options"))
					.getValue().get();
			defines = DocumentUtil.streamTargetParamHandlers(document).filter(p -> p.getName().equals("DEFINES"))
					.findAny().map(o -> o.getValue().get());

			properties.put("synthesis", targetOptions);
			properties.put("targetDefines", document.getTarget().streamVerilogDefines(false).map(Object::toString).collect(Collectors.joining()));
			defines.ifPresent(def -> properties.put("defines", def));
		}

		private String buildDefines() {
			StringBuilder builder = new StringBuilder();
			builder.append("-define {");
			builder.append(document.getTarget().streamVerilogDefines(false)
					.map(def -> def.format(false,true,'='))
					.collect(Collectors.joining(" ")));
			defines.ifPresent(def -> builder.append(" ").append(def));
			builder.append("}");
			return builder.toString();
		}

		@Override
		public void buildContent(PrintWriter writer) throws IOException {
			//TODO header?
			writer.println("set -tmpdir                " + "temp");
			writer.println("set -xsthdpdir             " + "xst_lib");
			writer.println();
			writer.println("run");
			writer.println(buildDefines());
			writer.format("-vlgincdir                 %s\n", hdlSourceTrees.stream().map(p -> '"' + p.toAbsolutePath().toString() + '"').collect(Collectors.joining(" ", "{", "}")));

			writer.println("-ifn                       " + getBuildRelPrj(systemDir));
			writer.println("-ofn                       " + "spartanmc");
			writer.println("-lso                       " + "spartanmc.lso");
			writer.println("-top                       " + VerilogLang.getVerilogName(document.getName()));
			writer.println(targetOptions);
		}

	}
}
