/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.ConstraintVisitor;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.KeepHierarchy;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.TimingIgnore;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.constraints.VendorConstraint;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.ConstraintReferenceAnalyzer;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.xilinx.XilinxIseDocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.HeaderGen;
import de.tu_darmstadt.rs.util.DigestUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Created to handle constraints like the previous jconfig did (concatenating all ucf files of modules + stuff from IOPIN Config)
 * Should be remodelled in a toolchain independent way //TODO
 * Created by ramon on 01.04.2016.
 */
public class XilinxVivadoConstraintsBuilderProvider extends BaseOutputBuilderProvider {

	private static final Logger logger = LoggerFactory.getLogger(XilinxVivadoConstraintsBuilderProvider.class);
	public static final String XDC_EXT = ".xdc";
	public static final String JCONFIG_XDC = "jconfig" + XDC_EXT;
	public static final String VENDOR = "xilinx";

	public XilinxVivadoConstraintsBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}

	public static Path getConstraintsFile(Path systemDir) {
		return systemDir.resolve(JCONFIG_XDC);
	}

	@Override
	public List<XDCOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		List<XDCOutputGroup> groups = new ArrayList<>();
		IConfigurationDocument document = buildProcess.getDocument();

		groups.add(new XDCOutputGroup(document.getProject().getIntermediaryBuildDir(), document, buildProcess));
		return groups;
	}


	protected static class XDCOutputGroup extends SingleFileBaseOutputGroup<Object, PrintWriter> {

		public static final double GIGA = 1E9;
		private final List<IPinHandler> pinHandlers;

		private final Path ucFile;
		private final CalculatedValueResult<String> ucData;
		private final IConfigurationDocument document;

		public XDCOutputGroup(Path systemDir, IConfigurationDocument document, IBuildProcess cache) {
			super(new DigestUtil.PrintWriterSupplier(), systemDir.resolve(JCONFIG_XDC));
			inputs = new HashSet<>();
			properties = new HashMap<>();

			this.document = document;

			IConfigurationDocument.ITargetConfiguration target = document.getTarget();

			inputs.add(target.getDeviceDescription().getSourceFile());
			inputs.add(target.getTargetDescription().getSourceFile());

			IAttributeHandler ucDataHandler = document.getTarget().getAttributeHandler(IConfigurationDocument.ITargetConfiguration.ATTRIB_UC_DATA);
			IAttributeHandler ucFileHandler = document.getTarget().getAttributeHandler(IConfigurationDocument.ITargetConfiguration.ATTRIB_UC_FILE);

			//TODO find better way to get project dir
			Path doc = systemDir.resolveSibling("..");
			ucFile = (ucFileHandler.getValue().isPresent() && !ucFileHandler.getValue().get().isEmpty())? doc.resolveSibling(ucFileHandler.getValue().get()) : null;
			ucData = ucDataHandler.getValue();

			if (ucFile != null) {
				//TODO add to properties because name may change???
				inputs.add(ucFile);

				if (!Files.exists(ucFile)) {
					cache.getIssueSink().report(ConfigurationIssue.createMissingFileIssue(ucFile, ucFileHandler, "Building constraints", ConfigurationIssue.IssueType.Warning));
				}
			}

			properties.put("uc data", ucData.toString());

			document.streamModulesRecursive()
					.filter(m -> m.getModuleDescription().getHdl().getConstraints() != null && !m.getModuleDescription().getHdl().getConstraints().isEmpty())
					.forEach(m -> properties.put(m.getGlobalName(), ConstraintReferenceAnalyzer.makePropertyFromConstraints(m, VENDOR, false)));

			pinHandlers = DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).peek(p -> {
				properties.put(p.getPortName(), XilinxIseDocumentBuilder.getDefiningValues(p));
			}).collect(Collectors.toList());


			pinHandlers.forEach(ph -> {
				if (ph.getIoStd() == null) {
					cache.getIssueSink().report(ConfigurationIssue.createInvalidPinProperties(ph, "IO Standard cannot be empty", "Module Validation"));
				}
			});
		}

		private static ConstraintVisitor<IModuleConfiguration<?>, String> constraintVisitor = new ConstraintVisitor<IModuleConfiguration<?>, String>() {

			private Map<String, String> timingNames = new HashMap<>();
			private Set<String> existingNames = new HashSet<>();

			/**
			 * Create or get a timing name for a net
			 * @param netName The net's name. May contain wildcards
			 * @param resultBuilder UCF Code to assign a timing name will be appended to this.
			 * @return The timing name
			 */
			private String getTimingName(String netName, StringBuilder resultBuilder) {
				/*String res = timingNames.get(netName);
				if (res == null) {
					res = "TS_" + VerilogLang.getVerilogName(netName);
					//Name conflict?
					if (!existingNames.add(res)) {
						//Add number at end to avoid conflict
						int i = 0;
						do {
							res = "TS_" + VerilogLang.getVerilogName(netName) + "_" + i++;
						} while (!existingNames.add(res));
					}

					resultBuilder.append(String.format("NET \"%s\" TNM_NET = \"%s\";%n", netName, res));

					timingNames.put(netName, res);
				}
				return res;*/
				return "NOT SUPPORTED!";
			}

			Map<IModuleConfiguration<?>, Integer> ignoreIndices = new HashMap<>();

			private int getIgnoreIndex(IModuleConfiguration<?> module) {
				int idx = ignoreIndices.getOrDefault(module, 0);
				ignoreIndices.put(module, idx + 1);
				return idx;
			}

			@Override
			public String visitVendorConstraint(VendorConstraint constraint, IModuleConfiguration<?> userData, String vendor, Boolean timing) {
				return constraint.getXilinxContent().resolve(userData, VerilogLang::getVerilogName,  userData.getModuleDescription().getSourceFile().getParent()).orElseThrow()
						//Remove indentation from multiline constraints
						.replaceAll("\\s*\n\\s*", "\n");
			}


			@Override
			public String visitTimingIgnore(TimingIgnore constraint, IModuleConfiguration<?> module) {
				/*StringBuilder result = new StringBuilder();

				constraint.getNets().forEach(resolveMixedParamRef(module, "NET \"", "\" TIG;", result));
				constraint.getInstances().forEach(resolveMixedParamRef(module, "INST \"", "\" TIG;", result));

				constraint.getIgnoreClockCrossings().forEach(icc -> {
					String from = icc.getFrom().resolve(module, VerilogLang::getVerilogName);
					String to = icc.getTo().resolve(module, VerilogLang::getVerilogName);

					String tnmFrom = getTimingName(from, result);
					String tnmTo = getTimingName(to, result);

					String name = "TS_ignore_" + module.getName() + "_" + getIgnoreIndex(module);

					result.append(String.format("TIMESPEC \"%s\" = FROM \"%s\" TO \"%s\" TIG;%n", name, tnmFrom, tnmTo));
					if (icc.isBidirectional()) {
						result.append(String.format("TIMESPEC \"%s\" = FROM \"%s\" TO \"%s\" TIG;%n", name + "_R", tnmTo, tnmFrom));
					}
				});


				return result.toString();*/
				return "NOT SUPPORTED!";
			}

			@Override
			public String visitKeepHierarchy(KeepHierarchy keepHierarchy, IModuleConfiguration<?> userData) {
				/*String settingStr;
				switch (keepHierarchy.getSetting()) {
					case TRUE:
						settingStr = "true";
						break;
					case FALSE:
						settingStr = "false";
						break;
					case SOFT:
						settingStr = "soft";
						break;
					default:
						throw new RuntimeException("Invalid value: " + keepHierarchy.getSetting());
				}
				return String.format("INST \"%s\" KEEP_HIERARCHY=%s;%n", keepHierarchy.getInstance().resolve(userData, VerilogLang::getVerilogName), settingStr);*/
				return "NOT SUPPORTED!";
			}
		};

		@Override
		public void buildContent(PrintWriter writer) throws IOException {
			new HeaderGen("#").getHeader("Xilinx Vivado Constraints File").write(writer, output);

			writer.println();

			pinHandlers.forEach(ph -> {
				writeLocConstraint(writer, ph);
				if (ph.getFrequency().isPresent()) {
					writeTimingConstraint(writer, ph);
				}
			});

			document.streamModulesRecursive()
					.filter(m -> m.getModuleDescription().getHdl().getConstraints() != null && !m.getModuleDescription().getHdl().getConstraints().isEmpty())
					.forEach(m -> {
						writer.println();
						writer.println("# Constraints for " + m.getName());
						m.getModuleDescription().getHdl().getConstraints().forEach(constraint -> writer.println(constraint.visitIfEnabled(m, constraintVisitor, m, VENDOR, false)));
					});

			if (ucFile != null) {
				writer.println();
				writer.println("# User specified file");
				writer.println("# " + ucFile);
				if (Files.exists(ucFile)) {
					Files.lines(ucFile).forEach(writer::println);
				} else {
					writer.println("# File not found.");
				}
			}

			if (ucData != null && ucData.isPresent() && !ucData.get().isEmpty()) {
				writer.println();
				writer.println("# User specified data");
				writer.println(ucData.get());
			}

		}

		private void writeTimingConstraint(PrintWriter writer, IPinHandler pinHandler) {
			final Long freqHz = pinHandler.getFrequency().orElseThrow(x -> new RuntimeException("absent for " + pinHandler));
			double freqNs = GIGA / freqHz;

			writer.format(Locale.ENGLISH, "create_clock -period %2$.4f -name %1$s -waveform {0.000 5.000} [get_ports %1$s]%n", VerilogLang.getPinName(pinHandler), freqNs);
		}

		private void writeLocConstraint(PrintWriter writer, IPinHandler pinHandler) {
			writer.format("set_property PACKAGE_PIN %s [get_ports %s]%n", pinHandler.getPinDescription().getFpgaPinName(), VerilogLang.getPinName(pinHandler));
			writer.format("set_property IOSTANDARD %s [get_ports %s]%n", pinHandler.getIoStd(), VerilogLang.getPinName(pinHandler));
		}
	}
}

