/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider;

import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.HdlFilesOutput;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Created by wenzel on 13.03.17.
 */
public class XilinxVivadoProjectBuilderProvider extends BaseOutputBuilderProvider{
	private static final Logger logger = LoggerFactory.getLogger(XilinxVivadoProjectBuilderProvider.class);

	public static final String GENPROJECT_TCL = "genproject.tcl";
	public static final String HDL_SOURCES_MK = "hdl_sources.mk";
	public static final String SPARTANMC_SIM_WORKLIB = "spartanmc_worklib.fdo";
	public static final String SPARTANMC_SIM_WORKLIB_TIMESIM = "spartanmc_worklib_timesim.fdo";

	public XilinxVivadoProjectBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}

	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		IConfigurationDocument document = buildProcess.getDocument();
		Path systemDir = document.getProject().getIntermediaryBuildDir();
		return Collections.singletonList(new HdlFilesOutput(document,
				Arrays.asList(
						getGenProjectOutput(systemDir, document),
						HdlFilesOutput.getMakeOutput(systemDir.resolve(HDL_SOURCES_MK)),
						HdlFilesOutput.getVsimOutput(systemDir.resolve(SPARTANMC_SIM_WORKLIB), SourceRef.FileType.FUNCTIONAL_SIM, document),
						HdlFilesOutput.getVsimOutput(systemDir.resolve(SPARTANMC_SIM_WORKLIB_TIMESIM), SourceRef.FileType.TIMESIM, document)
				),
				systemDir));
	}


	private HdlFilesOutput.DefaultHdlOutput getGenProjectOutput(Path systemDir, IConfigurationDocument document) {
		return new HdlFilesOutput.DefaultHdlOutput(
				systemDir.resolve(GENPROJECT_TCL),
				writer -> writeTclHead(writer, systemDir, document),
				XilinxVivadoProjectBuilderProvider::writeWork,
				null,
				SourceRef.FileType.SYNTHESIS
		);
	}


	private void writeTclHead(PrintWriter writer, Path systemDir, IConfigurationDocument document) {
		writer.println("close_project -quiet");
		writer.println("set outputDir "+systemDir.getParent().resolve("build/output"));
		//prjName and projectDir are set by vivado-paths.tcl
		writer.format("create_project $prjName $projectDir -part %s -force %n", document.getTarget().getDevice().getName());
		writer.format("set_property include_dirs {%s} [current_fileset]%n",document.getLibrary().getModuleSourceTrees().stream().map(Path::toAbsolutePath).map(Object::toString).collect(Collectors.joining(" ")));
		writer.format("set_property verilog_define {%s} [current_fileset]%n",document.getTarget().streamVerilogDefines(false).map(Object::toString).collect(Collectors.joining(" ")));
		writer.format("read_xdc %s%n",systemDir.resolve(XilinxVivadoConstraintsBuilderProvider.JCONFIG_XDC));
		writer.println("set hasCheckpoint false");

		writer.println("set_property strategy Performance_Explore [get_runs impl_1]");

	}


	private static void writeWork(PrintWriter writer, SourceRef ref) {

		String type = HdlFilesOutput.commandByFileExt(ref, "read_verilog", "read_vhdl");
		String path = ref.getPath().toString();

		writer.format("%s %s\n", type, path);
	}

}
