/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common;

import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.CLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.BaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.util.DigestUtil;

import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;

/**
 * Generates outputGroups for peripheral FILE initialization
 * Created by ramon on 01.04.2016.
 */
public class CPeriBuilderProvider extends BaseOutputBuilderProvider {

	private static final String PERIPHERALS_C = "peripherals_generated_jconfig.c";

	public CPeriBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}


	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		IConfigurationDocument document = buildProcess.getDocument();
		ILibJConfigProject project = document.getProject();
		Path configFile = project.getDocumentHandler().getSaveLocation();
		Path systemDir = project.getIntermediaryBuildDir();

		List<BaseOutputGroup> groups = new ArrayList<>();

		document.getSubsystems().forEach(sub -> {
			groups.add(new CPeriGroup(configFile, systemDir, sub));
		});

		return groups;
	}

	private class CPeriGroup extends SingleFileBaseOutputGroup<PeripheralModuleDescription.CSupport, PrintWriter> {

		private final List<IPeripheralModuleConfiguration> exPeris = new ArrayList<>();
		private final List<IPeripheralModuleConfiguration> dmaPeris = new ArrayList<>();
		private final Path configFile;
		private final ISubsystemConfiguration sub;

		public CPeriGroup(Path configFile, Path systemDir, ISubsystemConfiguration<?> sub) {
			super(new DigestUtil.PrintWriterSupplier(), DocumentBuilder.getSubsystemFile(systemDir, sub.getName().toLowerCase(), PERIPHERALS_C));
			this.configFile = configFile;
			this.sub = sub;
			inputs = new HashSet<>();
			properties = new HashMap<>();

			sub.streamPeripherals().forEach(peri -> {
				PeripheralModuleDescription description = peri.getModuleDescription();
				if (description.hasRegisters()) {
					PeripheralModuleDescription.CSupport cSupport = description.getRegisters().getCSupport();
					exPeris.add(peri);
					if (cSupport != null && cSupport.getDeclareFile() != null) {
						properties.put(peri.getName() + "_regs", cSupport);
					}
				}
				if (description.hasMemory()) {
					PeripheralModuleDescription.CSupport cSupport = description.getMemory().getCSupport();
					dmaPeris.add(peri);
					if (cSupport != null && cSupport.getDeclareFile() != null) {
						properties.put(peri.getName() + "_dma", cSupport);
					}
				}
			});
		}

		@Override
		public String getName() {
			return String.format("SubsystemPeriC:%s", sub.getName());
		}

		private boolean writeVariable(PrintWriter writer, String variableName, PeripheralModuleDescription.CSupport cSupport) {
			if (cSupport != null) {
				if (cSupport.getDeclareFile() != null) {
					writer.format("FILE %1$s_file = %2$s(&%1$s);\n", variableName, cSupport.getDeclareFile());
					return true;
				}
			}
			return false;
		}

		@Override
		public void buildContent(PrintWriter writer) {
			writer.println(CLang.getCHeaderGen().getHeaderString("FILE Declarations that can be used to access peripherals", output));
			writer.print("#include <peripherals.h>\n");
			writer.print("#include <stdio.h>\n");
			writer.println();
			final boolean periWritten = exPeris.stream().map(p -> {
				final PeripheralModuleDescription.RegisterInfo registers = p.getModuleDescription().getRegisters();
				return writeVariable(writer, CLang.getExposedPeriVarName(p), registers.getCSupport());
			}).reduce(false,(a, b) -> a || b); //AnyMatch without short circuiting
			writer.println();
			final boolean dmaWritten = dmaPeris.stream().map(p -> {
				final PeripheralModuleDescription.DmaMemory memory = p.getModuleDescription().getMemory();
				return writeVariable(writer, CLang.getExposedDmaVarName(p), memory.getCSupport());
			}).reduce(false,(a, b) -> a || b); //AnyMatch without short circuiting;

			if (!periWritten && !dmaWritten) {
				//Prevent empty file
				writer.println("typedef int make_iso_compilers_happy_no_empty_file;");
			}
		}

	}
}
