/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.MakeLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SimultaneousBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.verilog.VerilogBusGlue;
import de.tu_darmstadt.rs.spartanmc.jconfig.internal.SourceFileCollector;
import de.tu_darmstadt.rs.util.DigestUtil;
import de.tu_darmstadt.rs.util.GroupWriter;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.Collection;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.function.Consumer;
import java.util.stream.Collectors;

import static de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.verilog.VerilogBuilderProvider.getTopLevelVerilogFile;
import static java.util.Collections.emptySet;

public class HdlFilesOutput<T extends HdlFilesOutput.HdlOutput> extends SimultaneousBaseOutputGroup<Object> {

	public static final String JCONFIG_HDL_SOURCES = "JCONFIG_HDL_SOURCES";


	public abstract static class HdlOutput {

		private final Path filename;
		private final SourceRef.FileType fileType;

		public HdlOutput(Path filename, SourceRef.FileType fileType) {

			this.filename = filename;
			this.fileType = fileType;
		}

		public Path getFilename() {
			return filename;
		}

		public SourceRef.FileType getFileType() {
			return fileType;
		}


		public abstract void writeHead(PrintWriter writer);

		public abstract void writeWork(PrintWriter writer, SourceRef data);

		public abstract void writeFoot(PrintWriter writer);
	}


	public static class DefaultHdlOutput extends HdlOutput {

		private final Consumer<PrintWriter> writeHead;
		private final GroupWriter.ILineMapper<SourceRef> writeContent;
		private final Consumer<PrintWriter> writeFoot;

		public DefaultHdlOutput(
				Path filename,
				Consumer<PrintWriter> writeHead,
				GroupWriter.ILineMapper<SourceRef> writeContent,
				Consumer<PrintWriter> writeFoot,
				SourceRef.FileType fileType
		) {
			super(filename, fileType);

			this.writeHead = writeHead;
			this.writeContent = writeContent;
			this.writeFoot = writeFoot;
		}


		@Override
		public void writeHead(PrintWriter writer) {
			if (writeHead != null) {
				writeHead.accept(writer);
			}
		}

		@Override
		public void writeWork(PrintWriter writer, SourceRef data) {
			writeContent.write(writer, data);
		}

		@Override
		public void writeFoot(PrintWriter writer) {
			if (writeFoot != null) {
				writeFoot.accept(writer);
			}
		}


	}


	private static final Logger logger = LoggerFactory.getLogger(HdlFilesOutput.class);

	protected final Collection<SourceRef> files;
	private final List<T> hdlOutputs;

	public HdlFilesOutput(IConfigurationDocument document, List<T> hdlOutputs, Path systemDir) {
		super(DigestUtil.PrintWriterSupplier::new);

		this.hdlOutputs = hdlOutputs;

		inputs = emptySet();

		final Set<SourceRef> busGlueSrcs = VerilogBusGlue.getBusGlueSrcs(document.getProject().getApplication().getSpartanRoot());
		files = new LinkedHashSet<>();

		// lattice toolchain requires that the top module is the first one
		files.add(new SourceRef(getTopLevelVerilogFile(systemDir, document), EnumSet.of(SourceRef.FileType.SYNTHESIS, SourceRef.FileType.FUNCTIONAL_SIM)));
		Collection<IModuleDescription> busAddressVerification = Collections.singletonList(document.getLibrary().getModuleDescription("bus_address_verification"));
		files.addAll(document.streamModulesRecursive().collect(SourceFileCollector.create(document.getProject(), document.getLibrary(), busAddressVerification,busGlueSrcs)).stream().sorted().collect(Collectors.toList()));
		files.add(new SourceRef(document.getProject().getBuildDir().resolve("spartanmc_programram.v"), "work", SourceRef.FileType.FUNCTIONAL_SIM));
		files.add(new SourceRef(document.getProject().getBuildDir().resolve("spartanmc_timesim.v"), "timesim", SourceRef.FileType.TIMESIM));

		String toolchain = document.getTarget().getCurrentConfiguration().getToolchainImplementation();
		if (toolchain.equals("vivado")) {

			files.add(new SourceRef(document.getProject().getIntermediaryBuildDir().resolve(VerilogLang.getVerilogName(document.getName()) + "_timesim_wrapper.v"), "work", SourceRef.FileType.TIMESIM));
		}

		String hdlSourceTrees = document.getLibrary().getModuleSourceTrees().stream().map(Path::toAbsolutePath).map(Object::toString).collect(Collectors.joining(","));
		final List<String> defines = document.getTarget().streamVerilogDefines(false).map(Object::toString).collect(Collectors.toList());


		properties = new HashMap<>();
		properties.put("files",
				files.stream().map(SourceRef::toString).collect(Collectors.toSet()));
		properties.put("hdlSourceTrees", hdlSourceTrees);
		properties.put("defines", defines);

	}

	public static DefaultHdlOutput getVsimOutput(Path resolve, SourceRef.FileType fileType, IConfigurationDocument document) {
		final IConfigurationDocument.ITargetConfiguration target = document.getTarget();
		final Library library = document.getLibrary();
		return getVsimOutput(resolve, fileType, target, library);
	}

	@NotNull
	public static DefaultHdlOutput getVsimOutput(Path resolve, SourceRef.FileType fileType, IConfigurationDocument.ITargetConfiguration target, Library library) {
		return new DefaultHdlOutput(
				resolve,
				null,
				HdlFilesOutput::writeVsimWork,
				null,
				fileType
		);
	}

	public static DefaultHdlOutput getMakeOutput(Path filename) {
		return new DefaultHdlOutput(
				filename,
				w -> MakeLang.startMkEntry(w, JCONFIG_HDL_SOURCES),
				(wr, ref) -> MakeLang.writePartialMkEntry(wr, ref.getPath().toString()),
				w -> MakeLang.endMkEntry(w, JCONFIG_HDL_SOURCES),
				SourceRef.FileType.SYNTHESIS
		);
	}

	public static String commandByFileExt(SourceRef ref, String verilogCommand, String vhdlCommand) {
		String path = ref.getPath().toString();
		if (path.endsWith(".v")) {
			return verilogCommand;
		} else if (path.endsWith(".vhd")) {
			return vhdlCommand;
		} else {
			throw new RuntimeException("Unknown HDL File Type " + path);
		}
	}

	private static void writeVsimWork(PrintWriter writer, SourceRef ref) {
		String command = commandByFileExt(ref, "vlog", "vcom");
		String path = ref.getPath().toString();

		String library = "spartanmc";
		if (ref.getLibrary() != null) {
			library = ref.getLibrary();
		}

		if (!command.equals("vcom")) {
			writer.println(String.format("%s {*}$CMDLINE_OPTIONS -work %s %s", command, library, path));
		} else {
			writer.println(String.format("%s -work %s %s", command, library, path));
		}
	}

	public List<T> getHdlOutputs() {
		return hdlOutputs;
	}

	@Override
	public List<Path> getOutputFiles() {
		return hdlOutputs.stream().map(HdlOutput::getFilename).collect(Collectors.toList());
	}

	private GroupWriter.ILineMapper<SourceRef> getFilteredMapper(HdlOutput output) {
		return (writer, ref) -> {
			if (ref.getFileAssociation().contains(output.getFileType())) {
				output.writeWork(writer, ref);
			}
		};
	}

	@Override
	protected void buildContent(GroupWriter writer) {

		writer.writeAll(hdlOutputs.stream().map(o -> o::writeHead));

		writer.write(files.stream(), hdlOutputs.stream().map(this::getFilteredMapper).collect(Collectors.toList()));

		writer.writeAllln();

		writer.writeAll(hdlOutputs.stream().map(o -> o::writeFoot));
	}

}

