/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.AddressLayoutSpec;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.AsmLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.CLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.BaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SimultaneousBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.DigestUtil;
import de.tu_darmstadt.rs.util.GroupWriter;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Generates outputGroups for peripheral-headers + header files containing the various module parameters in c as well as asm.
 * Created by ramon on 01.04.2016.
 */
public class HeaderBuilderProvider extends BaseOutputBuilderProvider {

	private static final Logger logger = LoggerFactory.getLogger(HeaderBuilderProvider.class);

	private static final String MODULE_PARAMETERS_H = "moduleParameters.h";
	private static final String MODULE_PARAMETERS_S = "moduleParameters.s";
	private static final String GLOBAL_PARAMETERS_H = "globalParameters.h";
	private static final String GLOBAL_PARAMETERS_S = "globalParameters.s";
	private static final String PERIPHERALS_H = "peripherals.h";

	public HeaderBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}


	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		IConfigurationDocument document = buildProcess.getDocument();
		ILibJConfigProject project = document.getProject();
		Path configFile = project.getDocumentHandler().getSaveLocation();
		Path systemDir = project.getIntermediaryBuildDir();

		List<BaseOutputGroup> groups = new ArrayList<>();

		groups.add(new GlobalParamsGroup(configFile, systemDir, document));
		document.getSubsystems().forEach(sub -> {
			groups.add(new SubParamsGroup(configFile, systemDir, sub));
			groups.add(new PeripheralsHeaderGroup(configFile, systemDir, sub));
		});

		return groups;
	}

	private static class PeripheralsHeaderGroup extends SingleFileBaseOutputGroup<PeripheralModuleDescription.CSupport, PrintWriter> {

		private final List<IPeripheralModuleConfiguration> exPeris = new ArrayList<>();
		private final List<IPeripheralModuleConfiguration> dmaPeris = new ArrayList<>();
		private final Set<String> includeFiles = new HashSet<>();
		private final Path configFile;
		private final ISubsystemConfiguration sub;
		private final boolean needFileDeclaration;

		public PeripheralsHeaderGroup(Path configFile, Path systemDir, ISubsystemConfiguration<?> sub) {
			super(new DigestUtil.PrintWriterSupplier(), DocumentBuilder.getSubsystemFile(systemDir, sub.getName().toLowerCase(), PERIPHERALS_H));
			this.configFile = configFile;
			this.sub = sub;
			inputs = new HashSet<>();
			properties = new HashMap<>();

			final boolean[] needFileDeclaration = {false};

			sub.streamPeripherals().forEach(peri -> {
				PeripheralModuleDescription description = peri.getModuleDescription();

				//				Path includeDir = Library.getCIncludeDir(description);
				if (description.hasRegisters()) {
					PeripheralModuleDescription.CSupport cSupport = description.getRegisters().getCSupport();
					exPeris.add(peri);
					if (cSupport != null) {
						includeFiles.add(cSupport.getFileName());
						properties.put(peri.getName() + "_regs", cSupport);
						if (cSupport.getDeclareFile() != null) {
							needFileDeclaration[0] = true;
						}
					}
				}
				if (description.hasMemory()) {
					PeripheralModuleDescription.CSupport cSupport = description.getMemory().getCSupport();
					dmaPeris.add(peri);
					if (cSupport != null) {
						includeFiles.add(cSupport.getFileName());
						properties.put(peri.getName() + "_dma", cSupport);
						if (cSupport.getDeclareFile() != null) {
							needFileDeclaration[0] = true;
						}
					}
				}
			});
			this.needFileDeclaration = needFileDeclaration[0];
		}

		@Override
		public String getName() {
			return String.format("SubsystemPeriHeader:%s", sub.getName());
		}

		private void writeVariable(PrintWriter writer, String variableName, PeripheralModuleDescription.CSupport cSupport, int memorySize) {
			if (cSupport!=null) {
				writer.format("extern %s %s;\n", cSupport.getTypeName(), variableName);
				if (cSupport.getDeclareFile()!=null) {
					writer.format("extern const struct file %1$s_file;\n", variableName);
				}
			} else {
				writer.format("extern volatile int %s[%d];\n", variableName, memorySize);
			}
		}

		private void writeDefine(PrintWriter writer, String defineName, String variableName, PeripheralModuleDescription.CSupport cSupport) {
			writer.format("#define %s\t\t%s\n", defineName, variableName);
			if (cSupport != null && cSupport.getDeclareFile()!=null) {
				writer.format("#define %s_FILE\t\t%s_file\n", defineName, variableName);
			}
		}

		@Override
		public void buildContent(PrintWriter writer) {
			writer.println(CLang.getCHeaderGen().getHeaderString("Variable Declarations that can be used to access peripherals", output));
			CLang.getPragmaOnce().write(writer, null);
			includeFiles.forEach((d) -> writer.format("#include <%s>\n", d));
			writer.println();

			if (needFileDeclaration) {
				writer.println("struct file;");
				writer.println();
			}

			exPeris.forEach(p -> {
				final PeripheralModuleDescription.RegisterInfo registers = p.getModuleDescription().getRegisters();
				writeVariable(writer, CLang.getExposedPeriVarName(p), registers.getCSupport(), registers.getRegisters().size());
			});
			writer.println();
			dmaPeris.forEach(p -> {
				final PeripheralModuleDescription.DmaMemory memory = p.getModuleDescription().getMemory();
				Long arraySize = (Long) memory.getSizeRef().resolve(p).get();
				Optional<AddressLayoutSpec.AddressSpace> addressSpace = sub.getModuleDescription().getAddressLayoutSpec().getAddressSpaces().stream().filter(addresspace -> addresspace.getName().equals("data")).findAny();
				if (addressSpace.isPresent()) {
					AddressLayoutSpec.AddressSpace space = addressSpace.get();
					long spaceSize = space.getEnd() - space.getStart();
					if (arraySize > spaceSize / 2) {
						logger.error("Trying to generate array larger than half of available address space. This will be rejected by gcc.");
					}
				}
				writeVariable(writer, CLang.getExposedDmaVarName(p), memory.getCSupport(), ((Long) memory.getSizeRef().resolve(p).get()).intValue());
			});
			writer.println();
			exPeris.forEach(p -> writeDefine(writer, CLang.getExposedPeriPtrDefineName(p), CLang.getExposedPeriVarName(p), p.getModuleDescription().getRegisters().getCSupport()));
			writer.println();
			dmaPeris.forEach(p -> writeDefine(writer, CLang.getExposedDmaPtrDefineName(p), CLang.getExposedDmaVarName(p), p.getModuleDescription().getMemory().getCSupport()));
		}
	}


	private abstract static class ParamsGroup extends SimultaneousBaseOutputGroup<String> {

		protected final List<GroupWriter.ILineMapper<IParameterHandler>> mappers = Arrays.asList(
			(w, p) -> {
				w.format("#define SB_%-47s\t%s\n", getParamIdentifier(p), p.getValueConverted(CLang.getParameterConverter()));
			}, (w, p) -> {
				w.format(".equ SB_%-47s, %s\n", getParamIdentifier(p), p.getValueConverted(AsmLang.getParameterConverter()));
			});
		private final Path configFile;
		private final List<Path> outputs;

		protected ParamsGroup(Path configFile, List<Path> outputs) {
			super(DigestUtil.PrintWriterSupplier::new);
			this.configFile = configFile;
			this.outputs = outputs;
		}

		@Override
		public List<Path> getOutputFiles() {
			return outputs;
		}

		private String getParamIdentifier(IParameterHandler param) {
			return CLang.getCDefineName(param.getSourceElement().getName() + '_' + param.getName());
		}

		protected void writeToFile(GroupWriter grouped, List<IParameterHandler> params, String description) {
			grouped.writeSingle(configFile, Arrays.asList(CLang.getCHeaderGen().getHeader(description), AsmLang.getSHeaderGen().getHeader(description)));
			grouped.writeSingle(null, Arrays.asList(CLang.getPragmaOnce(), null));
			grouped.write(params.stream(), mappers);
		}
	}


	private static class GlobalParamsGroup extends ParamsGroup {

		private final IConfigurationDocument document;

		public GlobalParamsGroup(Path configFile, Path systemDir, IConfigurationDocument document) {
			super(configFile, Arrays.asList(systemDir.resolve(GLOBAL_PARAMETERS_H), systemDir.resolve(GLOBAL_PARAMETERS_S)));
			this.document = document;
			//			inputs = getDescriptionSources(document.getModules().stream());
			properties = DocumentUtil.collectParamProps(DocumentUtil.streamAllParameters(document.getModules()));
			inputs = Collections.emptySet();
		}

		@Override
		public void buildContent(GroupWriter writer) {
			List<IParameterHandler> params = DocumentUtil.streamAllParameters(document.getModules()).sorted(Comparator.comparing(IParameterHandler::getCanonicalName)).collect(Collectors.toList());

			writeToFile(writer, params, "All Parameters declared in global modules");
		}

	}


	private static class SubParamsGroup extends ParamsGroup {

		private final ISubsystemConfiguration sub;

		public SubParamsGroup(Path configFile, Path systemDir, ISubsystemConfiguration sub) {
			super(configFile, resolveOutputs(systemDir, sub));
			this.sub = sub;

			inputs = DocumentUtil.streamUsedDescriptions(sub).map(IModuleDescription::getSourceFile).collect(Collectors.toSet());
			properties = DocumentUtil.collectParamProps(getAllParams(sub));
		}

		private static List<Path> resolveOutputs(Path systemDir, ISubsystemConfiguration subsystem) {
			String sub = subsystem.getName().toLowerCase();
			return Stream.of(MODULE_PARAMETERS_H, MODULE_PARAMETERS_S).map(p -> DocumentBuilder.getSubsystemFile(systemDir, sub, p)).collect(Collectors.toList());
		}

		private Stream<IParameterHandler> getAllParams(ISubsystemConfiguration<?> sub) {
			return Stream.concat(DocumentUtil.streamAllParameters(sub), DocumentUtil.streamAllParameters(sub.getModules()));
		}

		@Override
		public String getName() {
			return String.format("SubsystemParamHeader:%s", sub.getName());
		}


		@Override
		public void buildContent(GroupWriter grouped) {
			List<IParameterHandler> params = getAllParams(sub).sorted(Comparator.comparing(IParameterHandler::getCanonicalName)).collect(Collectors.toList());

			writeToFile(grouped, params, "All Parameters declared in peripherals and modules of this subsystem");
		}
	}

}
