/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HdlDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IFirmwareConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.MakeLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.PeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.AddressBlock;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.AddressLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IFirmwareBinary;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IMemoryAddressBlock;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.DigestUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.PrintWriter;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.function.Function;
import java.util.function.UnaryOperator;
import java.util.stream.Collector;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by Ramon on 14.04.2016.
 */
public class LegacyMkBuilderProvider extends BaseOutputBuilderProvider {

	public static final String PROJECT_MK = "project.mk";
	public static final String FIRMWARES_MK = "firmwares.mk";
	public static final String SUBSYSTEM_HARDWARE_MK = "hardware.mk";

	private static final Logger logger = LoggerFactory.getLogger(LegacyMkBuilderProvider.class);
	private final Function<Path, Stream<StreamUtil.Pair<String, String>>> additionalEntries;

	public LegacyMkBuilderProvider(DocumentBuilder builder, Function<Path, Stream<StreamUtil.Pair<String, String>>> additionalEntries) {
		super(builder);
		this.additionalEntries = additionalEntries;
	}

	private static Path getDescriptionParentDir(IModuleConfiguration module) {
		return module.getModuleDescription().getSourceFile().getParent();
	}

	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		List<DocumentBuilder.IOutputGroup> groups = new ArrayList<>();
		IConfigurationDocument document = buildProcess.getDocument();

		Path systemDir = document.getProject().getIntermediaryBuildDir();
		groups.add(new ProjectMkGroup(systemDir, document));
		groups.add(new FirmwareMkGroup(systemDir, document, buildProcess));
		document.streamSubsystems().forEach(sub -> {
			Path firmwareSystemDir = systemDir.resolve("firmwares").resolve(sub.getGlobalName());
			groups.add(new SubsystemFirmwareMkGroup(firmwareSystemDir, document, buildProcess.getGlobalAddressLayout().getLayout(sub), sub));
		});
		return groups;
	}


	private class FirmwareMkGroup extends MkOutputGroup {

		private final IBuildProcess cache;
		private final List<IMemoryModuleConfiguration> memories;
		private final Map<? extends IMemoryAddressBlock, ? extends IFirmwareBinary> binaries;

		public FirmwareMkGroup(Path systemDir, IConfigurationDocument document, IBuildProcess buildProcess) {
			super(systemDir, document, systemDir.resolve(FIRMWARES_MK));
			this.cache = buildProcess;
			inputs = Collections.emptySet();
			properties = new HashMap<>();

			memories = DocumentUtil.streamMemoryModules(document).peek(mm -> {
				properties.put(mm.getGlobalName(), Objects.toString(mm.getFirmware().getLocation(document.getProject())));
			}).collect(Collectors.toList());

			binaries = buildProcess.getGlobalAddressLayout().getBinaries();
		}

		@Override
		public void buildContent(PrintWriter writer) {
			this.writer = writer;

			MakeLang.getMkHeaderGen().getHeader("Legacy Make-File, containing configuration for the make part of the toolchain").write(writer, output);

			addEntry("JCONFIG_FIRMWARE_IDS", binaries.values().stream()
					.distinct()
					.map(IFirmwareBinary::getName)
					.map(MakeLang::getMakeValue)
					.collect(Collectors.joining(" ")));
			binaries.values().stream().distinct().forEach(binary -> {
				String name = binary.getName();
				String globalNameMk = MakeLang.getMakeValue(name);
				IFirmwareConfiguration firmware = binary.getFirmware();
				addEntry(String.format("JCONFIG_FW_%s_SOURCE", globalNameMk), Objects.toString(firmware.getLocation(document.getProject())));
				addEntry(String.format("JCONFIG_FW_%s_LINKERSCRIPT", globalNameMk), DocumentBuilder.getLinkerScriptFile(Paths.get("."), name).toString());
				if (binary.getType() == IFirmwareBinary.Type.SingleCore || binary.getType() == IFirmwareBinary.Type.PartOfShared) {
					ISubsystemConfiguration subsystem = binary.getLayout().getSubsystem();
					addEntry(String.format("JCONFIG_FW_%s_SUBSYSTEM", globalNameMk), MakeLang.getMakeValue(subsystem.getName()));
				}
			});
		}
	}


	private class ProjectMkGroup extends MkOutputGroup {

		public ProjectMkGroup(Path systemDir, IConfigurationDocument document) {
			super(systemDir, document, systemDir.resolve(PROJECT_MK));
			inputs = Stream.of(document.getTarget().getTargetDescription().getSourceFile(), document.getTarget().getDeviceDescription().getSourceFile()).collect(Collectors.toSet());

			properties = document.streamModulesRecursive().collect(Collectors.toMap(IConfigurationElement::getGlobalName, m -> Objects.toString(m.getModuleDescription().getSourceFile())));

			DocumentUtil.streamUsedDescriptions(document)
					.forEach(descr -> {
						properties.put("coregen constraints for " + descr.getName(), descr.getHdl().getSources().getCoregenConstraints().toString());
					});


			document.getTarget().getToolchain().getParams().values()
					.stream()
					.map(document.getTarget()::getToolchainParamHandler)
					.forEach(param -> {
						properties.put("opt_" + param.getName(), param.getValue().orElseThrow());
					});

			Path saveLocation = document.getProject().getDocumentHandler().getSaveLocation();
			if (saveLocation != null) {
				properties.put("Document", saveLocation.toAbsolutePath().toString());
			}
		}


		private Stream<StreamUtil.Pair<Path, String>> getCoregenConstraints(Collection<IModuleConfiguration> modules) {

			Path ipCoreDir = document.getProject().getBuildDir().resolve("ipcore");
			return modules.stream()
					.flatMap(module -> module.getModuleDescription().getHdl().getSources().getCoregenConstraints().stream()
							.map((HdlDescription.CoregenConstraint f) -> {
								String mod = "\"\"";
								if (f.getModifier() != null && !f.getModifier().isEmpty()) {
									mod = '"' + f.getModifier().replaceAll("\"", "\\\\\"") + '"';
								}
								return new StreamUtil.Pair<>(
										f.resolve(ipCoreDir, module.getToplevel()),
										mod
								);
							})
					);
		}

		private void appendModuleInstances() {
			Map<String, IModuleConfiguration> allModules = document.streamModulesRecursive().collect(Collectors.toMap(IConfigurationElement::getGlobalName, Function.identity(), (a, b) -> {
				throw new UnsupportedOperationException();
			}, LinkedHashMap::new));

			addEntry("JCONFIG_MODULE_IDS", allModules.keySet().stream().map(MakeLang::getMakeValue).collect(Collectors.joining(" ")));
			addEntry("JCONFIG_MODULE_DIRS", allModules.values().stream().map(module -> getDescriptionParentDir(module).toString()).distinct().map(MakeLang::getSpaceEscaped).collect(Collectors.joining(" ")));
			Collector<CharSequence, ?, String> multiLineJoiner = Collectors.joining(" \\\n", "\\\n", "");
			addEntry("JCONFIG_PATH_MAP", allModules.entrySet().stream().map(e -> String.format("\tpath_%s=%s", MakeLang.getMakeValue(e.getKey()), MakeLang.getSpaceEscaped(getDescriptionParentDir(e.getValue()).toString()))).collect(multiLineJoiner));
			//			addEntry("JCONFIG_INSTANCE_MAP", allModules.entrySet().stream().map(e -> String.format("\tinstance_%s=%s", MakeLang.getMakeValue(e.getKey()), MakeLang.getSpaceEscaped(VerilogBuilderProvider.getInstancePath(e.getValue())))).collect(multiLineJoiner));


			List<StreamUtil.Pair<Path, String>> coregenConstraints = getCoregenConstraints(allModules.values()).collect(Collectors.toList());
			addEntry("JCONFIG_COREGEN_CONSTRAINTS", coregenConstraints.stream().map(StreamUtil.Pair::getA).map(Object::toString).collect(multiLineJoiner));
			addEntry("JCONFIG_COREGEN_CONSTRAINTS_MODIFIERS", "(" + coregenConstraints.stream().map(StreamUtil.Pair::getB).map(Object::toString).collect(multiLineJoiner) + ")");
		}

		protected void buildContent(PrintWriter writer) {
			Path saveLocation = document.getProject().getDocumentHandler().getSaveLocation();
			MakeLang.getMkHeaderGen().getHeader("Legacy Make-File, containing configuration for the make part of the toolchain").write(writer, saveLocation);
			this.writer = writer;

			addEntry("JCONFIG_MODULES_ROOT", document.getLibrary().getModuleSourceTrees().stream().map(t -> MakeLang.getSpaceEscaped(t.toAbsolutePath().toString())).collect(Collectors.joining(" ")));
			addEntry("JCONFIG_MODULE_SRC_SUFFIX", "/src/");
			appendModuleInstances();
			addEntry("JCONFIG_TOP_MODULE", MakeLang.getSpaceEscaped(VerilogLang.getVerilogName(document.getName())));
			addEntry("JCONFIG_PRJNAME", "spartanmc");
			//			addEntry("JCONFIG_PROJECT_MK_FILE", getSpaceEscaped(output.toString()));


			additionalEntries.apply(systemDir).forEach(StreamUtil.Pair.accessC(this::addEntry));
			addEntry("JCONFIG_USR_CONSTRAINTS_FILE", document.getTarget().getAttributeHandler(IConfigurationDocument.ITargetConfiguration.ATTRIB_UC_FILE).getValue().orElse(""));
			addEntry("JCONFIG_VERSION", document.getProject().getLibJConfigVersion());
			addEntry("JCONFIG_SYSTEM_DIR", systemDir.toString());
			appendTargetDefines();
			addEntry("JCONFIG_XML_FILE", saveLocation != null? saveLocation.toAbsolutePath().normalize().toString() : "");
			//			appendChipScopeContent();
		}

		private void appendTargetDefines() {
			addEntry("JCONFIG_HIERARCHY_PREFIX", "");// No Support in modelling as of now //TODO: support
			IConfigurationDocument.ITargetConfiguration target = document.getTarget();
			addEntry("JCONFIG_DEVICE", target.getDeviceDescription().getName());
			addEntry("JCONFIG_FAMILY", target.getDeviceDescription().getFamily());
			addEntry("JCONFIG_TARGET", target.getTargetDescription().getName());
			addEntry("JCONFIG_TARGET_PLATFORM", MakeLang.getSpaceEscaped(String.format("%s on %s", target.getTargetDescription().getName(), target.getDeviceDescription().getName())));
			addEntry("JCONFIG_JTAGID", target.getDevice().getJtagid());
			addEntry("JCONFIG_TOOLCHAIN", MakeLang.getMakeValue(builder.getToolchainName()));


			document.getTarget().getToolchain().getParams().values()
					.stream()
					.map(document.getTarget()::getToolchainParamHandler)
					.forEach(param -> {
						addEntry("JCONFIG_OPT_" + param.getName(), param.getValue().orElseThrow());
					});
		}

	}


	private class SubsystemFirmwareMkGroup extends MkOutputGroup {

		ISubsystemConfiguration<?> subsystem;
		AddressLayout addressLayout;

		public SubsystemFirmwareMkGroup(Path systemDir, IConfigurationDocument document, AddressLayout addressLayout, ISubsystemConfiguration<?> subsystem) {
			super(systemDir, document, systemDir.resolve(SUBSYSTEM_HARDWARE_MK));
			this.subsystem = subsystem;
			this.addressLayout = addressLayout;

			inputs = subsystem.streamAllModules().map(m -> m.getModuleDescription().getSourceFile()).collect(Collectors.toSet());
			properties = new HashMap<>();
			subsystem.streamAllModules().forEach(m -> {
				properties.putAll(DocumentUtil.collectParamProps(DocumentUtil.streamAllParameters(m)));
			});
		}

		@Override
		public void buildContent(PrintWriter writer) {
			this.writer = writer;

			MakeLang.getMkHeaderGen().getHeader("Makefile for subsystem").write(writer, output);

			subsystem.streamAllModules().filter(m -> m.getModuleDescription().getFirmwareDescription() != null).forEach(m -> {
				m.getModuleDescription().getFirmwareDescription().getDefinesList().forEach(firmwareDefine -> {
					long baseAddress = 0;
					if (m instanceof PeripheralModuleConfiguration) {
						Optional<AddressBlock.RegAddressBlock> regBlock = addressLayout.findRegAddressBlock((PeripheralModuleConfiguration) m);
						if (regBlock.isPresent()) {
							baseAddress = document.getBinaryOutputFormat().targetWordsToTargetBytes(regBlock.get().getActualStart());
						}
					}
					addEntry(firmwareDefine.getName(), firmwareDefine.getValue().resolve(m, UnaryOperator.identity(), m.getModuleDescription().getSourceFile().getParent(), baseAddress).orElseThrow());
				});
			});
		}

		@Override
		public String getName() {
			return this.getClass().getSimpleName() + subsystem.getGlobalName();
		}
	}


	private abstract class MkOutputGroup extends SingleFileBaseOutputGroup<String, PrintWriter> {

		protected final Path systemDir;
		protected final IConfigurationDocument document;
		protected PrintWriter writer;

		protected MkOutputGroup(Path systemDir, IConfigurationDocument document, Path output) {
			super(new DigestUtil.PrintWriterSupplier(), output);
			this.systemDir = systemDir;
			this.document = document;
		}

		protected void addEntry(String name, String value) {
			MakeLang.startMkEntry(writer, name);
			MakeLang.writeMkEntry(writer, value);
			MakeLang.endMkEntry(writer, name);
			writer.println();
			writer.println();
		}
	}
}
