/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common;

import de.tu_darmstadt.rs.jaxbUtil.IndentingXMLStreamWriter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryMap;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryMapJaxB;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.IIssueSink;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IGlobalAddressLayout;
import de.tu_darmstadt.rs.util.DigestUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import javax.xml.bind.JAXBException;
import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;

/**
 * Created by Ramon on 15.04.2016.
 */
public class MemoryMapBuilderProvider extends BaseOutputBuilderProvider {

	public static final String MEMORY_XML = "memory.xml";

	public MemoryMapBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}

	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		IConfigurationDocument document = buildProcess.getDocument();
		ArrayList<DocumentBuilder.IOutputGroup> groups = new ArrayList<>();

		Path systemDir = document.getProject().getIntermediaryBuildDir();

		groups.add(new MemoryMapOutput(systemDir, document, buildProcess));

		return groups;
	}

	private static class MemoryMapOutput extends SingleFileBaseOutputGroup<Object, PrintWriter> {

		private static final Logger logger = LoggerFactory.getLogger(MemoryMapOutput.class);
		private final IConfigurationDocument document;
		private final IGlobalAddressLayout globalLayout;
		private final IIssueSink issueSink;

		public MemoryMapOutput(Path systemDir, IConfigurationDocument document, IBuildProcess buildProcess) {
			super(new DigestUtil.PrintWriterSupplier(), systemDir.resolve(MEMORY_XML));
			this.document = document;
			this.issueSink = buildProcess.getIssueSink();
			inputs = new HashSet<>();
			properties = new HashMap<>();
			globalLayout = buildProcess.getGlobalAddressLayout();

			globalLayout.addAllAddressLayoutDependencies(inputs, properties);
			properties.put("Target Device", document.getTarget().getDevice().getName());
			properties.put("Toolchain", document.getTarget().getToolchain().getToolchain());
		}

		@Override
		public void buildContent(PrintWriter writer) throws IOException {
			//Only to allow injection of other elements
			XMLOutputFactory xof = XMLOutputFactory.newFactory();
			IndentingXMLStreamWriter streamWriter = null;

			try {
				try {
					streamWriter = new IndentingXMLStreamWriter(xof.createXMLStreamWriter(writer));
					streamWriter.setIndentStep("\t");

					MemoryMap map = globalLayout.getMemoryMap();

					MemoryMapJaxB jaxB = new MemoryMapJaxB();
					jaxB.marshall(map, streamWriter, document.getDocumentHandler().getSaveLocation());

					streamWriter.close();
				} catch (JAXBException e) {
					logger.error("Failed to build memory map!", e);
					throw new IOException(e);
				}
			} catch (XMLStreamException e) {
				logger.error("Failed to write XMLStream", e);
				throw new IOException(e);
			} finally {
				if (streamWriter != null) {
					try {
						streamWriter.close();
					} catch (XMLStreamException e) {
						logger.error("StreamWriter failed to close", e);
					}
				}
			}
		}

	}
}
