/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.IntStream;
import java.util.stream.Stream;

/**
 * Allows access to user specified internal port names.
 *
 * Useful for signal naming in debugging tools. Often, users want to see signals internal to modules. Instead of
 * creating a module port for each signal they want to view, they often create one wide port with all signals
 * connected internally with a list of "assign" statements.
 *
 * This class can parse such a list of assigns and display the internal signal names to the user. The assign statements
 * need to be copied into a file called internal_ports.v in the module description's directory.
 */
@SuppressWarnings("checkstyle:MagicNumber")
public final class ModuleInternalNames {

	private static final Logger logger = LoggerFactory.getLogger(ModuleInternalNames.class);

	private static Map<IModuleDescription, Optional<Map<String, String[]>>> internalNamesCache = new HashMap<>();

	private ModuleInternalNames() {
	}

	private static void enterBusName(Map<String, String[]> nameMap, String busName, int num, String name) {
		String[] arr = nameMap.get(busName);
		if (arr == null) {
			arr = new String[num + 1];
		} else if (arr.length <= num) {
			arr = Arrays.copyOf(arr, num + 1);
		}
		arr[num] = name;
		nameMap.put(busName, arr);
	}

	private static String getZeroBusBit(int[] zeroCount) {
		return "nothing[" + zeroCount[0]++ + "]";
	}

	private static Optional<Map<String, String[]>> loadInternalNames(IModuleDescription peri) {
		Path defFile = peri.getSourceFile().resolveSibling("internal_ports.v");
		if (!Files.exists(defFile)) {
			return Optional.empty();
		}

		Map<String, String[]> res = new HashMap<>();


		final int[] zeroCount = {0};
		Pattern regexSingle = Pattern.compile("(\\w+)\\s*\\[(\\d+)\\]\\s*=\\s*(\\w+);\\s*(//.*)?");
		Pattern regexBus = Pattern.compile("(\\w+)\\s*\\[(\\d+):(\\d+)\\]\\s*=\\s*(\\w+);\\s*(//.*)?");
		String zeroRegex = "(\\d*'[bhd])?0+";
		try {
			Files.lines(defFile).forEach(line -> {

				Matcher regexMatcherSingle = regexSingle.matcher(line);
				if (regexMatcherSingle.find()) {
					String busName = regexMatcherSingle.group(1);
					int num = Integer.parseInt(regexMatcherSingle.group(2));
					String name = regexMatcherSingle.group(3);
					if (name.matches(zeroRegex)) {
						enterBusName(res, busName, num, getZeroBusBit(zeroCount));
					} else {
						enterBusName(res, busName, num, name);
					}
				}

				Matcher regexMatcherBus = regexBus.matcher(line);
				if (regexMatcherBus.find()) {
					String busName = regexMatcherBus.group(1);
					int end = Integer.parseInt(regexMatcherBus.group(2));
					int start = Integer.parseInt(regexMatcherBus.group(3));
					String name = regexMatcherBus.group(4);
					if (name.matches(zeroRegex)) {
						for (int i = 0; i <= end - start; i++) {
							enterBusName(res, busName, start + i, getZeroBusBit(zeroCount));
						}
					} else {
						for (int i = 0; i <= end - start; i++) {
							enterBusName(res, busName, start + i, name + "[" + i + "]");
						}
					}
				}

			});
		} catch (IOException e) {
			logger.warn("Loading internal names failed for " + defFile, e);
			return Optional.empty();
		}


		return Optional.of(res);
	}


	private static Optional<Map<String, String[]>> getInternalNames(IModuleDescription module) {
		return internalNamesCache.computeIfAbsent(module, ModuleInternalNames::loadInternalNames);
	}

	private static Optional<String[]> getInternalNames(IPortHandler<?> handler) {
		if (handler.getSourceElement() instanceof IModuleConfiguration<?>) {
			IModuleConfiguration<?> conf = (IModuleConfiguration<?>) handler.getSourceElement();
			Optional<Map<String, String[]>> internalNames = getInternalNames(conf.getModuleDescription());
			return internalNames.map(n -> n.get(handler.getPortName()));
		}
		return Optional.empty();
	}

	/**
	 * Get the internal name for the specified bit in the signal
	 *
	 * @param handler the handler
	 * @param index the index
	 * @return internal name
	 */
	public static String getFriendlyName(IPortHandler<?> handler, int index) {
		return getInternalNames(handler)
				.flatMap(a -> {
					if (index < a.length) {
						return Optional.of(handler.getSourceElement().getAbsolutePath() + "/" + a[index]);
					} else {
						return Optional.empty();
					}
				})
				.orElseGet(() -> {
					if (handler.getConnectorWidth() != 1) {
						return handler.getAbsoluteConnectorName() + "[" + index + "]";
					}
					return handler.getAbsoluteConnectorName();
				});
	}

	/**
	 * For each bit in the signal, get the internal name of the connected port
	 *
	 * @param handler the handler
	 * @return internal names of signals connected to this
	 */
	public static Stream<String> getFriendlyConnectedNames(IPortHandler<IPort> handler) {
		return handler.streamConnections().sorted(Comparator.comparing(conn -> conn.getMyStart(handler)))
				.flatMap(conn -> {
					if (conn.getOppositeConnector(handler).getConnectorWidth() == 1) {
						return Stream.of(getFriendlyName(conn.getOppositeConnector(handler), 0));
					}
					return IntStream.rangeClosed(conn.getOppositeStart(handler), conn.getOppositeEnd(handler))
							.mapToObj(i -> getFriendlyName(conn.getOppositeConnector(handler), i));
				});
	}


}
