/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HdlDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.BaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.internal.SimRootModuleCollector;
import de.tu_darmstadt.rs.util.DigestUtil;

import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.stream.Collector;
import java.util.stream.Collectors;

public class TclDeviceInfoBuilderProvider extends BaseOutputBuilderProvider {

	private static final String DEVICE_TCL = "device.tcl";

	public TclDeviceInfoBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}


	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		IConfigurationDocument document = buildProcess.getDocument();
		ILibJConfigProject project = document.getProject();
		Path configFile = project.getDocumentHandler().getSaveLocation();
		Path systemDir = project.getIntermediaryBuildDir();

		List<BaseOutputGroup> groups = new ArrayList<>();

		groups.add(new TclDeviceInfoGroup(systemDir, document));

		return groups;
	}

	public static String getVsimCmdlineOptions(IConfigurationDocument.ITargetConfiguration target, Library library) {
		String defineParameter;
		if (target != null) {
			defineParameter = target.streamVerilogDefines(true).map(s -> "+define+" + s.format(false, false, '=')).collect(Collectors.joining(" "));
		} else {
			defineParameter = "";
		}

		String incdirs = "+incdir+"+library.getModuleSourceTrees().stream().map(p -> p.toAbsolutePath().toString()).collect(Collectors.joining("+"));

		return defineParameter+" "+incdirs;
	}

	private class TclDeviceInfoGroup extends SingleFileBaseOutputGroup<String, PrintWriter> {

		private Collector<CharSequence, ?, String> tclListCollector() {
			return Collectors.joining(" ","{","}");
		}



		public TclDeviceInfoGroup(Path systemDir, IConfigurationDocument document) {
			super(new DigestUtil.PrintWriterSupplier(), systemDir.resolve(DEVICE_TCL));
			inputs = new HashSet<>();
			properties = new HashMap<>();

			final IConfigurationDocument.ITargetConfiguration target = document.getTarget();
			properties.put("TOOLCHAIN", target.getToolchain().getToolchain());
			properties.put("DEVICE", target.getDeviceDescription().getName());
			properties.put("FAMILY", target.getDeviceDescription().getFamily());
			properties.put("TARGET", target.getTargetDescription().getName());
			properties.put("DEVICE_SIMLIBS", target.getToolchain().getSimlibsAll().collect(tclListCollector()));
			properties.put("DEVICE_SIMLIBS_FUNCTIONAL", target.getToolchain().getSimlibsFunctional().collect(tclListCollector()));
			properties.put("DEVICE_SIMLIBS_TIMING", target.getToolchain().getSimlibsTiming().collect(tclListCollector()));

			properties.put("CMDLINE_OPTIONS", "{"+getVsimCmdlineOptions(target, document.getLibrary())+"}");

			//Testbench needs to be first
			final List<HdlDescription.SingletonSimRootModule> additionalRootModules = Arrays.asList(
					new HdlDescription.SingletonSimRootModule("testbench", true),
			new HdlDescription.SingletonSimRootModule("sim_initram", false)
			);
			final SimRootModuleCollector.SimRoots rootModules = document.streamModulesRecursive()
					.collect(SimRootModuleCollector.create(document.getProject(), document.getLibrary(), additionalRootModules));


			properties.put("ROOT_MODULES_FUNCTIONAL", rootModules.functionalRoots.stream().collect(tclListCollector()));
			properties.put("ROOT_MODULES_POST_PAR", rootModules.postParRoots.stream().collect(tclListCollector()));
		}

		@Override
		public void buildContent(PrintWriter writer) {
			properties.entrySet().stream()
					.sorted(Comparator.comparing(Map.Entry::getKey))
					.forEach(entry -> writer.format("variable %s %s%n",entry.getKey(), entry.getValue()));
		}

	}
}
