/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.CLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.VerilogOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.quartus.AlteraQuartusDocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.xilinx.XilinxIseDocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.xilinx.XilinxVivadoDocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.lattice.LatticeDiamondDocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections.PartialConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Creates a Verilog testbench that instantiates the toplevel design with its pins
 */
public class VerilogSimulationBuilderProvider extends BaseOutputBuilderProvider {

	private static final Logger logger = LoggerFactory.getLogger(VerilogSimulationBuilderProvider.class);

	public VerilogSimulationBuilderProvider(DocumentBuilder builder) {
		super(builder);
	}

	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		IConfigurationDocument document = buildProcess.getDocument();
		return Collections.singletonList(new VerilogSimulationOutput(document.getProject().getIntermediaryBuildDir(), document));
	}

	private static class VerilogSimulationOutput extends VerilogOutputGroup {


		private final IConfigurationDocument document;

		public VerilogSimulationOutput(Path systemDir, IConfigurationDocument document) {
			super(systemDir.resolve("module_sim_" + VerilogLang.getVerilogName(document.getName()) + ".v"));
			this.document = document;

			inputs = Collections.emptySet();
			properties = new HashMap<>();

			String toolchain = document.getTarget().getCurrentConfiguration().getToolchainImplementation();
			if(toolchain.equals("quartus")){

				DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).forEach(p -> {
					properties.put(p.getPortName(), AlteraQuartusDocumentBuilder.getDefiningValues(p) + getNet(p).toString());
				});

			}
			else if(toolchain.equals("lattice")){

				DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).forEach(p -> {
					properties.put(p.getPortName(), LatticeDiamondDocumentBuilder.getDefiningValues(p) + getNet(p).toString());
				});

			}
			else if(toolchain.equals("xilinx")){

				DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).forEach(p -> {
					properties.put(p.getPortName(), XilinxIseDocumentBuilder.getDefiningValues(p) + getNet(p).toString());
				});

			}
			else{

				DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).forEach(p -> {
					properties.put(p.getPortName(), XilinxVivadoDocumentBuilder.getDefiningValues(p) + getNet(p).toString());
				});

			}
		}

		@Override
		protected String getHeader() {
			return CLang.getCHeaderGen().getHeaderString("Simulation Testbench Verilog Module", document.getProject().getDocumentHandler().getSaveLocation());
		}

		@Override
		protected void buildContent() {
			writeTimescale("1ns", "1ps");

			writeModuleStart("testbench", Stream.empty(), null, null);

			List<IPinHandler> connectedPins = DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).sorted(IConnectorInstance.NAME_COMPARATOR).collect(Collectors.toList());

			//if(document.getTarget().getDeviceDescription().getToolchain().getName().equals("lattice")) {
			if(document.getTarget().getCurrentConfiguration().getToolchainImplementation().equals("lattice")) {
				writeComment("===================");
				writeComment("PUR / GSR instances");
				writeComment("===================");
				writeLatticePurGsr();
			}

			writeComment("======");
			writeComment("Inputs");
			writeComment("======");
			writePinsOfDirection(connectedPins, ConnectorDirection.Input, "reg");
			writeNewline();


			writeComment("=======");
			writeComment("Outputs");
			writeComment("=======");
			writePinsOfDirection(connectedPins, ConnectorDirection.Output, "wire");
			writeNewline();


			writeComment("======");
			writeComment("Inouts");
			writeComment("======");
			writePinsOfDirection(connectedPins, ConnectorDirection.InOut, "wire");
			writeNewline();


			writeComment("=============");
			writeComment("Instantiation");
			writeComment("=============");

			writeModuleInstanceStart(VerilogLang.getVerilogName(document.getName()));
			writeModuleInstanceParams(Stream.empty());//Unnecessary
			writeModuleInstanceNameAndPorts("UUT", connectedPins.stream(), VerilogLang::getPinName, p -> getNet(p).getNameWithOffset(), ph -> ph.getDirection().getReverse());

			writeComment("======");
			writeComment("Clocks");
			writeComment("======");
			connectedPins.stream().filter(ph->ph.getFrequency().isPresent()).forEach(pin -> {
				long freq = pin.getFrequency().get();
				writeComment(freq / 1E6 + "MHz clock on " + getNet(pin).name);

				//Freq in Hz to Period in ns
				double period = 1E9 / freq;

				//Clock needs to toggle every half period
				writeClockGeneration(getNet(pin).name, period / 2);

			});


			writeNewline();
			writeModuleEnd();

		}

		private void writePinsOfDirection(List<IPinHandler> pins, ConnectorDirection direction, String type) {
			pins.stream()
				.filter(p -> p.getTargetDirection() == direction)
				//Remove the offset so that multiple occurrences of the same net are equal
				.map(pin -> getNet(pin).withoutOffset())
				.distinct()
				.sorted()
				.forEach(net -> {
					writeDeclaration(type, net.name, net.width);
				});
		}

		/**
		 * Get the net to connect to the pin
		 *
		 * @param pin
		 * @return
		 */
		private Net getNet(IPinHandler pin) {
			//If pin is connected to multiple sinks, just use the pin name
			if (pin.getConnectionCount() != 1) {
				return new Net(pin.getPortName(), 1, -1);
			}

			IConnection<IPortHandler<?>> connection = pin.streamConnections().findAny()
				.orElseThrow(() -> new IllegalStateException("connection count is 1, but stream of connections is empty"));


			IPortHandler<?> connected = connection.getOppositeConnector(pin);

			String netName = getNetNameFromConnected(connected);

			switch (connection.getType()) {
				case Port:
					return new Net(netName, 1, -1);
				case Partial:
					PartialConnection c = (PartialConnection) connection;
					if (!c.isOppositePartial(pin)) {
						throw new RuntimeException("Opposite is not partial");
					}


					return new Net(netName, connected.getConnectorWidth(), c.getOppositeStart(pin));
				case Bus:
				case MappedBus:
				default:
					throw new IllegalStateException("Not supported.");
			}

		}

		private String getNetNameFromConnected(IPortHandler<?> connected) {
			return connected.getSourceElement().getName() + "_" + connected.getPortName();
		}


		private class Net implements Comparable<Net> {

			final String name;
			final int width;
			final int offset;

			private Net(String name, int width, int offset) {
				this.name = VerilogLang.getVerilogName(name);
				this.width = width;
				this.offset = offset;
			}

			public Net withoutOffset() {
				return new Net(name, width, -1);
			}

			public String getNameWithOffset() {
				if (width > 1) {
					return name + "[" + offset + "]";
				} else {
					return name;
				}
			}

			@Override
			public int compareTo(Net o) {
				return name.compareTo(o.name);
			}

			@Override
			public int hashCode() {
				int result = name.hashCode();
				result = 31 * result + width;
				result = 31 * result + offset;
				return result;
			}

			@Override
			public boolean equals(Object o) {
				if (this == o) {
					return true;
				}
				if (o == null || getClass() != o.getClass()) {
					return false;
				}

				Net net = (Net) o;

				if (width != net.width) {
					return false;
				}
				if (offset != net.offset) {
					return false;
				}
				return name.equals(net.name);

			}

			@Override
			public String toString() {
				return "Net{" +
					"name='" + name + '\'' +
					", width=" + width +
					", offset=" + offset +
					'}';
			}
		}

	}


}
