/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.linker;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.AddressLayoutSpec;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.BinaryOutputFormat;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IFirmwareConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.CLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.SingleFileBaseOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.linker.sectionLayouts.ISectionLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.linker.api.ILinkerSection;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.linker.api.PhysAddrCounter;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.IIssueSink;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.AddressLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IFirmwareBinary;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IMemoryAddressBlock;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.interrupts.IInterruptLayout;
import de.tu_darmstadt.rs.util.DigestUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

abstract class BaseGnuLinkerScriptOutput extends SingleFileBaseOutputGroup<Object, PrintWriter> {

	private static final Logger logger = LoggerFactory.getLogger(SingleCoreGnuLinkerScriptOutput.class);

	protected final BinaryOutputFormat outputFormat;
	protected final IBuildProcess buildProcess;
	protected final ISectionLayout sectionLayout;

	protected BaseGnuLinkerScriptOutput(Path systemDir, IFirmwareBinary binary, IBuildProcess buildProcess, ISectionLayout sectionLayout) {
		super(new DigestUtil.PrintWriterSupplier(), DocumentBuilder.getLinkerScriptFile(systemDir, binary.getName()));
		outputFormat = buildProcess.getDocument().getBinaryOutputFormat();
		this.buildProcess = buildProcess;

		inputs = new HashSet<>();
		properties = new HashMap<>();

		this.sectionLayout = sectionLayout;

		addSectionDependencies(sectionLayout.getAvailableSections());
	}

	/**
	 * Will utilized the sections themselves to discover all blocks into which each section could possibly be placed
	 *
	 * This does not include any restrictions based on groups of sections
	 *
	 * @param sections all available sections
	 * @return list of possible blocks for every section
	 */
	protected Map<? extends ILinkerSection, ? extends List<IMemoryAddressBlock>> discoverPossiblePlacements(List<ILinkerSection> sections, List<IMemoryAddressBlock> addressBlocks) {
		return sections.stream()
				.collect(Collectors.toMap(
						Function.identity(),
						sec -> addressBlocks.stream()
								.filter(sec::canOutputTo)
								.sorted(Comparator.comparing(IMemoryAddressBlock::getActualStart))
								.collect(Collectors.toList()),
						(a, b) -> {
							throw new RuntimeException();
						},
						LinkedHashMap::new));
	}

	private static final Comparator<StreamUtil.Pair<? extends ILinkerSection, ? extends IMemoryAddressBlock>> BLOCK_COMP = Comparator.comparing(pair -> pair.getB().getActualStart());

	protected Map<? extends IMemoryAddressBlock, ? extends List<? extends ILinkerSection>> autoBindSectionsToBlocks(Map<? extends ILinkerSection, ? extends List<? extends IMemoryAddressBlock>> possiblePlacements) {

		IIssueSink issueSink = buildProcess.getIssueSink();

		return possiblePlacements.entrySet().stream()
				.flatMap(entry -> {
					ILinkerSection section = entry.getKey();
					List<? extends IMemoryAddressBlock> possible = entry.getValue();

					if (possible != null && !possible.isEmpty()) {
						if (section.isBindMultiple() || possible.size() > 1) {
							return (Stream<? extends StreamUtil.Pair<? extends ILinkerSection, ? extends IMemoryAddressBlock>>) section.bindMultiple(possible, issueSink).stream();// Java Generics with wildcards are ugly!
						} else {
							return Stream.of(new StreamUtil.Pair<>(section, possible.get(0)));
						}
					} else if (!section.isBindMultiple()) {
						logger.warn("Section {} could not be placed!", section);
					}
					return Stream.empty();
				})
				.sorted(BLOCK_COMP)
				.collect(
						Collectors.groupingBy(
								StreamUtil.Pair::getB,
								LinkedHashMap::new,
								Collectors.mapping(
										StreamUtil.Pair::getA,
										Collectors.toList()
								)
						)
				);
	}

	protected void writeDetailedSymbol(PrintWriter writer, String name, long wordAddr) {
		writer.format("%s = 0x%x; /* internally at daddr 0x%x, caddr 0x%x */\n", name, outputFormat.targetWordsToBytes(wordAddr), outputFormat.targetWordsToTargetBytes(wordAddr), wordAddr);
	}

	private void writeStdioProvides(PrintWriter writer, String direction) {
		writer.println(CLang.MULTI_COMMENT_START);
		writer.println(CLang.MULTI_COMMENT_SPACER + " Generate better error messages for people who use stdio functions and do not set std" + direction);
		writer.println(CLang.MULTI_COMMENT_END);

		writer.format("PROVIDE(__std%s_missing = 0);%n", direction);
		writer.format("PROVIDE(std%1$s = __std%1$s_missing);%n", direction);
		writer.format("ASSERT(!DEFINED(__std%1$s_missing), \"std%1$s undefined. In order to use standard %1$sput, " +
				"please define a global variable std%1$s as FILE * and initialize it to point to your peripheral " +
				"FILE, for example: FILE * std%1$s = &UART_LIGHT_0_FILE;\");%n", direction);

		writer.println();
	}

	protected void writeStdioProvides(PrintWriter writer) {
		writeStdioProvides(writer, "out");
		writeStdioProvides(writer, "in");
	}

	protected String getRegionName(IMemoryAddressBlock block, ISubsystemConfiguration subsystem) {
		String name = block.getSourceModule().getGlobalName() + '_' + block.getName();
		return CLang.getCVarName(name);
	}

	protected String getRegionFlags(IMemoryAddressBlock block) {
		StringBuilder sb = new StringBuilder();
		if (block.getRange().isWritable()) {
			sb.append('w');
		} else {
			sb.append('r');
		}
		if (block.getRange().isExecutable()) {
			sb.append('x');
		}
		return sb.toString();
	}

	public void addSectionDependencies(List<ILinkerSection> availableSections) {
		int i = 0;
		for (ILinkerSection section : availableSections) {
			properties.put("Section " + section.getName() + ' ' + Integer.toString(i++), section.getVersion());
		}
	}

	public void addInterruptLayoutDependencies(ISubsystemConfiguration subsystem, IInterruptLayout interruptLayout) {
		properties.put("intrLayout_" + subsystem.getName(), interruptLayout.getInterruptRoutines().stream().map(IInterruptLayout.IInterruptRoutine::getName).collect(Collectors.joining(", ")));
	}

	protected void writeStackTop(PrintWriter writer, long dataMemEnd) {
		writeDetailedSymbol(writer, "STACK_TOP", dataMemEnd);
		writer.println();
	}

	protected void writeEntryVectors(PrintWriter writer) {
		writer.println();
		writer.println("PROVIDE(_trap_table = __endLoop); /* ensure that _trap_table exists, because startup writes it to SFR_TR */");
		writer.println("PROVIDE(_intr_handle = __endLoop); /* ensure that _intr_handle exists, because startup writes it to SFR_IV */");

		writer.println("ENTRY(\"__start\");");
		writer.println();
	}

	protected void writePeriAndDmaVars(PrintWriter writer, AddressLayout addressLayout) {
		addressLayout.streamLayoutedBlocks().forEach(block -> {
			long start = block.getActualStart();
			if (block.getBlockType() == AddressLayoutSpec.BlockType.Initializable) {
				writeDetailedSymbol(writer, CLang.getExposedDmaVarName((IPeripheralModuleConfiguration) block.getSourceModule()), start);
			} else {
				writeDetailedSymbol(writer, CLang.getExposedPeriVarName((IPeripheralModuleConfiguration) block.getSourceModule()), start);
			}
		});
		writer.println();
	}

	protected void writeLinkerMemories(PrintWriter writer) {
		writer.println("MEMORY {");
		writeLinkerMemoryEntries(writer);
		writer.println("}");
		writer.println();
	}

	protected abstract void writeLinkerMemoryEntries(PrintWriter writer);

	protected void writeLinkerSections(PrintWriter writer, Map<? extends IMemoryAddressBlock, ? extends IFirmwareBinary> firmwares, Map<? extends IMemoryAddressBlock, ? extends List<? extends ILinkerSection>> placed) {
		writer.println("SECTIONS {");
		writer.println("\t. = 0x0;");

		PhysAddrCounter counter = new PhysAddrCounter(false);

		writeLinkerSectionEntries(writer, firmwares, counter, placed);

		writer.println("}");
		writer.println();
	}

	protected abstract void writeLinkerSectionEntries(PrintWriter writer, Map<? extends IMemoryAddressBlock, ? extends IFirmwareBinary> firmware, PhysAddrCounter counter, Map<? extends IMemoryAddressBlock, ? extends List<? extends ILinkerSection>> placed);

	protected void writeBlockRegion(IMemoryAddressBlock block, ISubsystemConfiguration subsystem, PrintWriter writer) {
		writer.format("\t%-40s (%-2s) : ORIGIN = 0x%-10x , LENGTH = 0x%x /*  */\n", getRegionName(block, subsystem), getRegionFlags(block), outputFormat.targetWordsToBytes(block.getActualStart()), outputFormat.targetWordsToBytes(block.getLength()));

	}

	/**
	 * generates input filters for folders
	 *
	 * @param firmware
	 * @param subsystem
	 * @param global
	 * @return
	 */
	protected String getInputSubInputFilter(IFirmwareConfiguration firmware, ISubsystemConfiguration subsystem, boolean global) {
		if (global) {
			Optional<String> glob = firmware.getGlobalFirmwareName();
			return glob.map(s -> s + '/').orElseGet(() -> {
				logger.warn("Undefined Global Firmware for {}", subsystem);
				return "";
			});
		} else if (firmware.hasSubFirmwares()) {
			Optional<String> sub = firmware.getSubFirmwareName(subsystem);
			return sub.map(s -> s + '/').orElseGet(() -> {
				logger.error("Undefined Subfirmware for {}", subsystem);
				return "";
			});
		} else {
			return "";
		}
	}

	protected void writeSection(PrintWriter writer, ILinkerSection section, IMemoryAddressBlock block, ISubsystemConfiguration subsystem, IFirmwareConfiguration firmware, PhysAddrCounter counter, boolean firstInRegion) {
		String prefixName = getSectionPrefixName(block, subsystem, false);
		String regionName = getRegionName(block, subsystem);
		String inputFilter = getInputSubInputFilter(firmware, subsystem, false);

		section.writeSection(prefixName, inputFilter, regionName, writer, counter, outputFormat, subsystem, buildProcess, block, firstInRegion);
	}

	private String getSectionPrefixName(IMemoryAddressBlock block, ISubsystemConfiguration subsystem, boolean hasGlobalMem) {
		String name;
		if (block.getBlockType() == AddressLayoutSpec.BlockType.SelfInitializing && block.isShared()) {
			String dmaName = block.getSourceModule().getName();
			name = !hasGlobalMem? dmaName : (subsystem.getName() + '_' + dmaName);
		} else if (hasGlobalMem) {
			name = block.getSourceModule().getGlobalName();
		} else {
			return "";
		}
		return '.' + CLang.getCVarName(name);
	}
}
