/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.linker;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IFirmwareConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.CLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.linker.api.ILinkerSection;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.linker.api.PhysAddrCounter;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.linker.sectionLayouts.ISectionLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.AddressLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IFirmwareBinary;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IMemoryAddressBlock;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.interrupts.IInterruptLayout;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Path;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.OptionalLong;
import java.util.stream.Collectors;

public class SingleCoreGnuLinkerScriptOutput extends BaseGnuLinkerScriptOutput {

	private static final Logger logger = LoggerFactory.getLogger(SingleCoreGnuLinkerScriptOutput.class);

	private final ISubsystemConfiguration<?> subsystem;
	private final AddressLayout addressLayout;
	private final IInterruptLayout interruptLayout;
	private final IFirmwareBinary buildingBinary;
	private final Map<? extends IMemoryAddressBlock, ? extends IFirmwareBinary> binaries;

	public SingleCoreGnuLinkerScriptOutput(Path systemDir, AddressLayout addressLayout, IFirmwareBinary buildingBinary, IBuildProcess buildProcess, ISectionLayout sectionLayout) {
		super(systemDir, buildingBinary, buildProcess, sectionLayout);

		subsystem = addressLayout.getSubsystem();
		this.buildingBinary = buildingBinary;

		binaries = buildProcess.getGlobalAddressLayout().getBinaries();

		interruptLayout = buildProcess.getInterruptLayout(subsystem);

		addInterruptLayoutDependencies(subsystem, interruptLayout);

		addressLayout.addDependencies(inputs, properties);
		this.addressLayout = addressLayout;
	}

	private static IMemoryModuleConfiguration<?> getSourceModule(AddressLayout addressLayout) {
		return ((IMemoryModuleConfiguration<?>) addressLayout.streamFirmwareAttachedInitializableBlocks().findFirst().get().getSourceModule());//TODO dirty dirty code, replacement requires heavy rewrites because the linkescript names are bases on memory-modules instead of addressLayout names
	}

	@Override
	public String getName() {
		return super.getName() + ":" + subsystem.getGlobalName();
	}

	@Override
	public void buildContent(PrintWriter writer) throws IOException {
		CLang.getCHeaderGen().getHeader("Linker Script").write(writer, buildProcess.getDocument().getProject().getDocumentHandler().getSaveLocation());

//		if (firmwares.stream().anyMatch(m -> !m.getFirmware().hasFirmware())) {//TODO cannot check here, needs to be done in a new place were the common firmware -> block mapping is created
//			writer.println(CLang.MULTI_COMMENT_START);
//			writer.println(CLang.MULTI_COMMENT_SPACER+ " No linker script generated, since firmware is still missing set.");
//			writer.println(CLang.MULTI_COMMENT_END);
//
//			return;
//		}

		writer.println(CLang.MULTI_COMMENT_START);
		writer.format(CLang.MULTI_COMMENT_SPACER + " The following file is in %s.\n" +
				CLang.MULTI_COMMENT_SPACER + " 1 Target Word will occupy %d ELF-bytes in binaries. 1 Target Byte %d ELF-bytes.\n", outputFormat.getName(), outputFormat.getBytesPerTargetWord(), outputFormat.getBytesPerTargetByte());
		writer.println(CLang.MULTI_COMMENT_END);
		writer.println();

		writeStdioProvides(writer);

		writePeriAndDmaVars(writer, addressLayout);

		writer.format("INT_COUNT = %d;\n", interruptLayout.getInterruptRoutines().size());
		OptionalLong controllerAddress = interruptLayout.getControllerAddress();
		if (controllerAddress.isPresent()) {
			writeDetailedSymbol(writer, "INT_BASE", controllerAddress.getAsLong());
			writer.println("ASSERT(DEFINED(_intr_handle),\"Error: No Interrupt Handler defined. Link libinterrupt\");");
		}

		writeUndefinedSymbols(writer);

		writeEntryVectors(writer);

		writeLinkerMemories(writer);

		Map<? extends ILinkerSection, ? extends List<? extends IMemoryAddressBlock>> possiblePlacements = discoverPossiblePlacements(sectionLayout.getAvailableSections(), addressLayout.streamInitializableBlocks().collect(Collectors.toList()));
		Map<? extends IMemoryAddressBlock, ? extends List<? extends ILinkerSection>> placed = autoBindSectionsToBlocks(possiblePlacements);

		placed.forEach((block, sections) -> {
			sections.forEach(section -> logger.debug("Placing {} into {}", section, block));
		});

		writeLinkerSections(writer, binaries, placed);
	}

	private void writeUndefinedSymbols(PrintWriter writer) {
		writer.format("EXTERN( __loader );\n");
	}

	@Override
	protected void writeLinkerMemoryEntries(PrintWriter writer) {
		addressLayout.streamInitializableBlocks().sorted(Comparator.comparing(IMemoryAddressBlock::getActualStart)).forEach(block -> writeBlockRegion(block, subsystem, writer));
	}

	@Override
	protected void writeLinkerSectionEntries(PrintWriter writer, Map<? extends IMemoryAddressBlock, ? extends IFirmwareBinary> firmwares, PhysAddrCounter counter, Map<? extends IMemoryAddressBlock, ? extends List<? extends ILinkerSection>> placed) {

		placed.forEach((block, sections) -> {
			boolean first = true;
			IFirmwareConfiguration pickedFirmware = binaries.get(block).getFirmware();
			IFirmwareConfiguration firmware = buildingBinary.getFirmware();
			if (!pickedFirmware.equals(firmware)) {
				logger.error("Binary {} does not support {} as firmware source for {}", buildingBinary, pickedFirmware, block);
			}
			for (ILinkerSection section: sections) {
				writeSection(writer, section, block, subsystem, firmware, counter, first);
				first = false;
			}
		});
	}
}
