/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.linker.sections;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryContents;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.BinaryOutputFormat;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.linker.api.ILinkerSection;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.IIssueSink;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IMemoryAddressBlock;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

class DefaultSection extends BaseSection {

	private static final Logger logger = LoggerFactory.getLogger(DefaultSection.class);

	private final String section;
	private final boolean perCoreState;
	private final boolean firstStageAccess;
	protected final boolean requiresReadAccess;
	protected final boolean requiresWriteAccess;
	protected final boolean requiresExecuteAccess;
	private final String filter;

	public DefaultSection(String sectionAndFilter, boolean perCoreState, boolean firstStageAccess, boolean requiresReadAccess, boolean requiresWriteAccess, boolean requiresExecuteAccess) {
		this(sectionAndFilter, sectionAndFilter, perCoreState, firstStageAccess, requiresReadAccess, requiresWriteAccess, requiresExecuteAccess);
	}

	public DefaultSection(String section, String filter, boolean perCoreState, boolean firstStageAccess, boolean requiresReadAccess, boolean requiresWriteAccess, boolean requiresExecuteAccess) {
		this.section = section;
		this.filter = filter;
		this.perCoreState = perCoreState;
		this.firstStageAccess = firstStageAccess;
		this.requiresReadAccess = requiresReadAccess;
		this.requiresWriteAccess = requiresWriteAccess;
		this.requiresExecuteAccess = requiresExecuteAccess;
	}

	@Override
	public int getVersion() {
		return Objects.hash(section, perCoreState, requiresReadAccess, requiresWriteAccess, requiresExecuteAccess, firstStageAccess, filter, getClass().getCanonicalName());
	}

	@Override
	public String getName() {
		return section;
	}

	@Override
	public String getSectionFilter(String subPrefix) {
		return filter;
	}

	@Override
	public Collection<? extends StreamUtil.Pair<? extends ILinkerSection, ? extends IMemoryAddressBlock>> bindMultiple(List<? extends IMemoryAddressBlock> possible, IIssueSink issueSink) {
		if (!possible.isEmpty()) {
			logger.warn("Section {} could be placed into multiple blocks ({}), but does not specify a preference, choosing first possible memory", section, possible);
			issueSink.report(ConfigurationIssue.createSectionPlacementGuessed(this, possible.get(0), possible, "Writing Linkerscript"));
			return Collections.singletonList(new StreamUtil.Pair<>(this, possible.get(0)));
		} else {
			throw new IllegalArgumentException("Called with empty list of possible blocks!");
		}
//		throw new IllegalStateException("Section "+ section +" could be placed into multiple blocks ("+ possible + "), but does not specify a preference");
	}

	@Override
	public boolean canOutputTo(IMemoryAddressBlock block) {
		return block.getMemory().getContents().getContentConfiguration() instanceof MemoryContents.GeneralPurposeMemory
				&& (!firstStageAccess || block.isBitfileInitializable())
				&& (!requiresReadAccess || block.getRange().isReadable())
				&& (!requiresWriteAccess || block.getRange().isWritable())
				&& (!requiresExecuteAccess || block.getRange().isExecutable())
				&& (!perCoreState || !block.isShared());

	}

	@Override
	protected String getOutputOptions(IMemoryAddressBlock block, boolean firstInRegion, BinaryOutputFormat format) {
		if (!firstInRegion && format.hasMinimumAlignment()) {
			return String.format("ALIGN ( 0x%x )", format.getMinimumAlignment());
		}
		return "";
	}

	@Override
	public String getInputFilter() {
		return "*";
	}

	@Override
	public String toString() {
		return "Section("+ getName() +"{"+ getInputFilter() + " "+ getSectionFilter("") +"})";
	}
}
