/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.verilog;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.CombineType;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.IPortMapping;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.files.base.ILoadedFile;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IBusBreakoutModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleVisitor;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseOutputBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.CLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.VerilogOutputGroup;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.AddressBlock;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.AddressLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IAddressChunk;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.IGlobalAddressLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.BusLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.IMappedBusPortInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.MappedBusConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.jetbrains.annotations.NotNull;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.TreeMap;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by ramon on 01.04.2016.
 */
public class VerilogBuilderProvider extends BaseOutputBuilderProvider {

	private final Function<IPinHandler, String> definingValues;

	public VerilogBuilderProvider(DocumentBuilder builder, Function<IPinHandler, String> definingValues) {
		super(builder);
		this.definingValues = definingValues;
	}

	public static Path getTopLevelVerilogFile(Path systemDir, IConfigurationDocument document) {
		return systemDir.resolve(VerilogLang.getVerilogName(document.getName()) + ".v");
	}

	@Override
	public List<? extends DocumentBuilder.IOutputGroup> getOutputGroups(IBuildProcess buildProcess) {
		IConfigurationDocument document = buildProcess.getDocument();
		return Collections.singletonList(new TopLevelVerilogOutput(document.getProject().getIntermediaryBuildDir(), document, buildProcess, definingValues));
	}


	private static class TopLevelVerilogOutput extends VerilogOutputGroup {

		//private final Path out;
		private final IConfigurationDocument document;
		private final IBuildProcess cache;
		private final Function<IPinHandler, String> definingValues;
		//private final String moduleName;
		private Set<? extends IConnection> unrouted;
		private Map<IModuleConfiguration, List<IPortHandler<?>>> portsPerModule;

		public TopLevelVerilogOutput(Path systemDir, IConfigurationDocument document, IBuildProcess cache, Function<IPinHandler, String> definingValues) {
			super(getTopLevelVerilogFile(systemDir, document));
			this.document = document;
			this.cache = cache;
			this.definingValues = definingValues;
			Library library = document.getLibrary();
			inputs = new HashSet<>();
			properties = new HashMap<>();

			cache.getGlobalAddressLayout().streamLayouts().forEach(layout -> {
				properties.put("Layout_" + layout.getSubsystem().getName(), layout.extractStaticMappings());
			});

			properties.putAll(DocumentUtil.collectParamProps(
					document.streamModulesRecursive()
							.flatMap(DocumentUtil::streamAllParameters)
							.filter(p -> !p.isVirtual())
			));

			Stream.concat(document.getConnections().stream(), document.getBusConnections().stream()).map(Object::toString).forEach(e -> properties.put(e, e));

			DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).forEach(p -> {
				properties.put("PIN_" + p.getPortName(), this.definingValues.apply(p));
			});

			inputs.add(document.getTarget().getDeviceDescription().getSourceFile());
			inputs.add(document.getTarget().getTargetDescription().getSourceFile());

			Set<IModuleDescription> usedDescriptions = document.streamModulesRecursive().map(IModuleConfiguration::getModuleDescription).collect(Collectors.toSet());
			Set<BusDescription> usedBusDescriptions = usedDescriptions.stream().sorted(ILoadedFile.NAME_COMP)
					.peek(desc -> inputs.add(desc.getSourceFile()))
					.flatMap(desc -> desc.getBuses().stream()).map(busImpl -> library.getBusDescription(busImpl.getBusDescriptionName())).collect(Collectors.toSet());
			usedBusDescriptions.stream().sorted(ILoadedFile.NAME_COMP).forEach(desc -> inputs.add(desc.getSourceFile()));
		}

		private static <T, K> Stream<T> stream(Map<K, List<T>> groupedList, K key) {
			return groupedList.get(key).stream();
		}

		private static <T, M extends IConfigurationElement<?, ?, ?> & IValueResolver> Stream<T> streamSorted(Map<M, List<T>> groupedList) {
			return groupedList.entrySet().stream().sorted(Comparator.comparing(Map.Entry::getKey, IConfigurationElement.NAME_COMPARATOR)).map(Map.Entry::getValue).flatMap(List::stream);
		}

		@Override
		protected String getHeader() {
			return CLang.getCHeaderGen().getHeaderString("Toplevel Verilog Module", document.getProject().getDocumentHandler().getSaveLocation());
		}


		@Override
		public void buildContent() {

			writeDefaultNettype("none");

			writeNewline();

			// lattice specific additions
			if (document.getTarget().getCurrentConfiguration().getToolchainImplementation().equals("lattice")) {
				document.getTarget().streamVerilogDefines(false)
						.forEach(this::writeDefineIfNotSimulated);
			}

			BusLayout busLayout = cache.getBusLayout();
			WireNames wireNames = buildWireNames(busLayout);

			List<IPinHandler> connectedPins = DocumentUtil.streamConnectorHandlers(document).filter(IConnectorInstance::isConnected).sorted(IConnectorInstance.NAME_COMPARATOR).collect(Collectors.toList());

			writeModuleStart(VerilogLang.getVerilogName(document.getName()), connectedPins.stream(), IPinHandler::getTargetDirection, VerilogLang::getPinName);

			//Declare wires
			wireNames.streamUsedWireBits().map(bit -> bit.wire).distinct()
					//Do not list fpga pins, since they are already defined
					.filter(wire -> !wire.module.isEmpty())
					.sorted(Comparator.comparing(w -> w.getVerilogName(false)))
					.forEach(wire ->
							writeDeclaration("wire", wire.getVerilogName(false), wire.width)
					);

			portsPerModule = document.streamModulesRecursive().collect(Collectors.toMap(Function.identity(), module -> DocumentUtil.streamConnectorHandlers(module).sorted(IConnectorInstance.NAME_COMPARATOR).collect(Collectors.toList())));

			writeNewline();
			writeNewline();

			// write out all module instantiations, including parameters
			document.streamModulesRecursive().forEach(module -> writeModuleInstance(module, busLayout, wireNames, cache.getGlobalAddressLayout()));


			// write out bus glue, like or-gates
			writeNewline();
			writeComment("Bus Glue (Ex: Or gates that or together bus inputs)");
			writeNewline();

			VerilogBusGlue verilogBusGlue = new VerilogBusGlue(this);
			busLayout.streamMappedPorts().filter(IMappedBusPortInstance::requiresBusGlue).forEach(mp -> mp.applyBusGlue(
					verilogBusGlue,
					wireNames.glueInputMapper(),
					wireNames.glueOutputMapper()
			));


			// write out assigns
			writeNewline();
			writeComment("Assigns for toplevel outputs");
			writeNewline();

			writeAssigns(wireNames);
			writeNewline();

			writeComment("Verification for bus data");
			writeAddressVerification(cache.getGlobalAddressLayout(), wireNames, busLayout);
			writeNewline();

			// lattice specific additions
			if (document.getTarget().getCurrentConfiguration().getToolchainImplementation().equals("lattice")) {

				// defparams for Lattice programram initialization (must be done in verilog source)
				writeLatticeProgramramInclude();
			}

			// End Module
			writeNewline();
			writeModuleEnd();
		}


		private static class BusVerificationInfo {

			public final String busDescriptionName;
			public final Bus.Role role;
			private final Map<String, List<String>> ports;

			private BusVerificationInfo(String busDescriptionName, Bus.Role role, StreamUtil.Pair<String, List<String>>... ports) {
				this.busDescriptionName = busDescriptionName;
				this.role = role;
				this.ports = new TreeMap<>();
				for (StreamUtil.Pair<String, List<String>> port : ports) {
					this.ports.put(port.getA(), port.getB());
				}
			}

			private String getMappedPortNameFromBusPortName(Bus bus, String busPortName) {
				return bus.getPortMappings().stream()
						.filter(pm -> pm.getBusName().equals(busPortName))
						.collect(StreamUtil.atMostOne())
						.map(IPortMapping::getName)
						.orElse(busPortName);
			}

			public Stream<String> getMappedPortNames(Bus bus, String portName) {
				final List<String> busPortNames = ports.get(portName);
				if (busPortNames == null) {
					throw new IllegalStateException("Port " + portName + " not found on " + bus);
				}

				return busPortNames.stream().map(busPortName -> getMappedPortNameFromBusPortName(bus, busPortName));
			}

			public String getMappedPortName(Bus bus, String portName) {
				final List<String> strings = ports.get(portName);
				if (strings == null) {
					throw new IllegalStateException("Port " + portName + " not found on " + bus);
				}
				if (strings.size() != 1) {
					throw new RuntimeException("not exactly one port");
				}
				return getMappedPortNameFromBusPortName(bus, strings.get(0));
			}
		}


		private static final Map<String, BusVerificationInfo> verifiableBuses = new HashMap<>();

		private static void addVerifiableBus(BusVerificationInfo bvi) {
			verifiableBuses.put(bvi.busDescriptionName, bvi);
		}

		static {

			StreamUtil.Pair<String, List<String>>[] mem_ports = new StreamUtil.Pair[]{
					new StreamUtil.Pair("clk", Collections.singletonList("clk")),
					new StreamUtil.Pair("data", Arrays.asList("mem_do")),
					new StreamUtil.Pair("reset", Collections.singletonList("reset")),
					new StreamUtil.Pair("addr", Arrays.asList("mem_addr_high", "mem_addr_block")),
					new StreamUtil.Pair("access", Collections.singletonList("mem_access")),
					new StreamUtil.Pair("busy", Arrays.asList("mem_busy"))
			};

			StreamUtil.Pair<String, List<String>>[] codemem_ports = new StreamUtil.Pair[]{
					new StreamUtil.Pair("clk", Collections.singletonList("clk")),
					new StreamUtil.Pair("reset", Collections.singletonList("reset")),
					new StreamUtil.Pair("data", Collections.singletonList("code_do")),
					new StreamUtil.Pair("addr", Collections.singletonList("code_addr")),
					new StreamUtil.Pair("access", Collections.singletonList("code_access")),
					new StreamUtil.Pair("busy", Arrays.asList("code_busy"))
			};

			StreamUtil.Pair<String, List<String>>[] dma_ports = new StreamUtil.Pair[]{
					new StreamUtil.Pair("clk", Collections.singletonList("clk_peri")),
					new StreamUtil.Pair("reset", Collections.singletonList("reset")),
					new StreamUtil.Pair("data", Collections.singletonList("di_peri")),
					new StreamUtil.Pair("addr", Arrays.asList("addr_high", "addr_peri")),
					new StreamUtil.Pair("access", Collections.singletonList("mem_access")),
					new StreamUtil.Pair("busy", Collections.emptyList())
			};

			StreamUtil.Pair<String, List<String>>[] peri_ports = new StreamUtil.Pair[]{
					new StreamUtil.Pair("clk", Collections.singletonList("clk_peri")),
					new StreamUtil.Pair("reset", Collections.singletonList("reset")),
					new StreamUtil.Pair("data", Collections.singletonList("di_peri")),
					new StreamUtil.Pair("addr", Collections.singletonList("addr_peri")),
					new StreamUtil.Pair("access", Collections.singletonList("access_peri")),
					new StreamUtil.Pair("busy", Collections.emptyList())
			};

			addVerifiableBus(new BusVerificationInfo("spartanmc-memory-bus", Bus.Role.Master, mem_ports));
			addVerifiableBus(new BusVerificationInfo("spartanmc-memory-bus2", Bus.Role.Master, mem_ports));
			//addVerifiableBus(new BusVerificationInfo("spartanmc-code-memory-bus", Bus.Role.Master, codemem_ports));
			addVerifiableBus(new BusVerificationInfo("spartanmc-dma-bus", Bus.Role.Slave, dma_ports));
			addVerifiableBus(new BusVerificationInfo("spartanmc-peri-bus", Bus.Role.Slave, peri_ports));
		}

		private boolean isVerifiableBus(Bus bus) {
			final BusVerificationInfo busVerificationInfo = verifiableBuses.get(bus.getBusDescriptionName());
			if (busVerificationInfo == null) {
				return false;
			}
			return busVerificationInfo.role == bus.getRole();
		}

		private Stream<String> getVerificationDataOut(Bus bus) {
			final BusVerificationInfo busVerificationInfo = verifiableBuses.get(bus.getBusDescriptionName());
			return busVerificationInfo.getMappedPortNames(bus, "data");
		}


		private class BusVerifier {

			private final IModuleConfiguration<?> module;
			private final String doPort;
			private final List<Bus> buses;
			private final ISubsystemConfiguration<?> subsystem;
			private final int connectionIndex;

			public BusVerifier(IModuleConfiguration<?> module, String doPort, List<Bus> buses, ISubsystemConfiguration<?> subsystem, int connectionIndex) {
				this.module = module;
				this.doPort = doPort;
				this.buses = buses;
				this.subsystem = subsystem;
				this.connectionIndex = connectionIndex;
			}

			private Stream<? extends IAddressChunk> getAddresses(Bus bus, IGlobalAddressLayout globalAddressLayout) {
				AddressLayout layout = globalAddressLayout.getLayout(subsystem);

				return module.apply(new IModuleVisitor<Stream<? extends IAddressChunk>>() {
					@Override
					public Stream<? extends IAddressChunk> applyToPeripheralModule(IPeripheralModuleConfiguration module) {
						if (module.getParent() == null) {
							return Stream.empty();
						}

						final Stream.Builder<IAddressChunk> builder = Stream.builder();

						IPeripheralModuleConfiguration<?> peri = (IPeripheralModuleConfiguration<?>) module;
						final PeripheralModuleDescription moduleDescription = peri.getModuleDescription();

						if (moduleDescription.hasMemory() && moduleDescription.getMemory().getExposedOn() == bus) {
							layout.findAddressBlock(moduleDescription.getMemory().getExposedOn(), peri).ifPresent(builder::add);
						}

						if (moduleDescription.hasRegisters() && moduleDescription.getRegisters().getExposedOn() == bus) {
							layout.findAddressBlock(moduleDescription.getRegisters().getExposedOn(), peri).ifPresent(builder::add);
						}
						return builder.build();
					}

					@Override
					public Stream<? extends IAddressChunk> applyToSubsystemModule(ISubsystemConfiguration module) {
						//Has no memories
						return Stream.empty();
					}

					@Override
					public Stream<? extends IAddressChunk> applyToCommonModule(IModuleConfiguration<?> module) {
						//Has no memories
						return Stream.empty();
					}

					@Override
					public Stream<? extends IAddressChunk> applyToBusBreakoutModule(IBusBreakoutModuleConfiguration<?> module) {
						//Has no memories
						return Stream.empty();
					}

					@Override
					public Stream<? extends IAddressChunk> applyToMemoryModule(IMemoryModuleConfiguration<?> module) {
						IMemoryModuleConfiguration<?> mem = (IMemoryModuleConfiguration<?>) module;

						return layout.streamFirmwareAttachedInitializableBlocks()
								.filter(block -> block.getSourceModule() == mem)
								.filter(block -> mem.getMemoriesExposedOn(bus).contains(block.getMemory()));
					}
				});
			}

			private List<StreamUtil.Pair<Bus, IAddressChunk>> getAddresses(IGlobalAddressLayout globalAddressLayout) {
				final Function<Bus, Stream<StreamUtil.Pair<Bus, IAddressChunk>>> busStreamFunction = bus -> {
					return getAddresses(bus, globalAddressLayout).map((Function<IAddressChunk, StreamUtil.Pair<Bus, IAddressChunk>>) x -> new StreamUtil.Pair<>(bus, x));
				};
				return buses.stream()
						.flatMap(busStreamFunction)
						.collect(Collectors.toList());
			}

			private Stream<OutputPort> generatePorts(Bus bus, WireNames wireNames, BusLayout busLayout, int index, int watcherCount, String verificationWire) {
				final BusVerificationInfo busVerificationInfo = verifiableBuses.get(bus.getBusDescriptionName());
				final Stream<OutputPort> fromBus = busVerificationInfo.ports.keySet().stream()
						.filter(portName -> !portName.equals("data"))
						.filter(portName -> !portName.equals("busy"))
						.map(portName -> generateOutputPort(bus, wireNames, busLayout, busVerificationInfo, portName));
				//final Stream<OutputPort> clk = Stream.of(new OutputPort("clk", wireNames.getWireConnection(getClkPort(busLayout)), ConnectorDirection.Input));
				final Stream<OutputPort> select = Stream.of(new OutputPort("select", verificationWire + VerilogLang.getVerilogBus(index, index, watcherCount), ConnectorDirection.Output));
				return Stream.concat(fromBus, select);
			}

			private OutputPort concatPorts(OutputPort lowPort, OutputPort highPort) {
				if (lowPort.direction != highPort.direction) {
					throw new IllegalStateException("Cannot concat ports of different direction");
				}
				return new OutputPort(lowPort.getName(), "{" + highPort.getConnectedName() + "," + lowPort.getConnectedName() + "}", lowPort.direction);
			}


			@NotNull
			private VerilogOutputGroup.OutputPort generateOutputPort(Bus bus, WireNames wireNames, BusLayout busLayout, BusVerificationInfo busVerificationInfo, String portName) {
				final ConnectorDirection[] direction = new ConnectorDirection[1];
				direction[0] = ConnectorDirection.Input;
				final List<String> connections = busVerificationInfo.ports.get(portName).stream().map(busPortName -> {

					return generateOutputPortConnectionFromBusPortName(bus, wireNames, busLayout, busVerificationInfo, direction, busPortName);
				}).collect(Collectors.toList());
				String connectionString;
				if (connections.size() == 0) {
					connectionString = "0";
				} else if (connections.size() == 1) {
					connectionString = connections.get(0);
				} else {
					connectionString = connections.stream().collect(Collectors.joining(", ", "{", "}"));
				}


				return new OutputPort(portName, connectionString, direction[0]);
			}

			@NotNull
			private VerilogOutputGroup.OutputPort generateOutputPortFromBusPortName(Bus bus, WireNames wireNames, BusLayout busLayout, BusVerificationInfo busVerificationInfo, String portName, String busPortName) {
				final ConnectorDirection[] direction = new ConnectorDirection[1];

				String connection = generateOutputPortConnectionFromBusPortName(bus, wireNames, busLayout, busVerificationInfo, direction, busPortName);

				return new OutputPort(portName, connection, direction[0]);
			}


			private String generateOutputPortConnectionFromBusPortName(Bus bus, WireNames wireNames, BusLayout busLayout, BusVerificationInfo busVerificationInfo, ConnectorDirection[] direction, String busPortName) {
				final BusDescription busDescription = document.getLibrary().getBusDescription(bus.getBusDescriptionName());
				final BusPort busPort = busDescription.getPort(bus, busPortName);
				direction[0] = busPort.getDirection();

				IMappedBusPortInstance bpi = getBpiFromBusPortName(bus, busLayout, busPortName);
				return generateOutputPortConnectionFromBpi(wireNames, bpi);
			}

			private String generateOutputPortConnectionFromBpi(WireNames wireNames, IMappedBusPortInstance bpi) {
				final CombineType combineType = bpi.getCombineType();

				Wire wire = Wire.fromBusPortInstance(bpi);

				if (wire.width != bpi.getBaseWidth()) {

					if (combineType == CombineType.Parallel) {

						return wireNames.getWireConnection(wire, bpi.getBaseWidth() * connectionIndex, bpi.getBaseWidth() * (connectionIndex + 1) - 1);

					} else {
						throw new RuntimeException("Unexpected differing bus port widths");
					}
				} else {
					final String wireConnection = wireNames.getWireConnection(wire);
					return wireConnection;
				}
			}

			private IMappedBusPortInstance getBpiFromBusPortName(Bus bus, BusLayout busLayout, String portName) {
				final BusVerificationInfo busVerificationInfo = verifiableBuses.get(bus.getBusDescriptionName());
				String modulePortName = Objects.requireNonNull(busVerificationInfo.getMappedPortNameFromBusPortName(bus, portName));
				return getBpiFromBus(module, busLayout, modulePortName);
			}

			private IMappedBusPortInstance getBpiFromBusAbstractName(Bus bus, BusLayout busLayout, String portName) {
				final BusVerificationInfo busVerificationInfo = verifiableBuses.get(bus.getBusDescriptionName());
				String modulePortName = Objects.requireNonNull(busVerificationInfo.getMappedPortName(bus, portName));
				return getBpiFromBus(module, busLayout, modulePortName);
			}

			private IMappedBusPortInstance getBpiFromBus(IModuleConfiguration module, BusLayout busLayout, String modulePortName) {
				return busLayout.getBusPortInstances(module)
						.stream()
						.filter(p -> p.getPortName().equals(modulePortName))
						.collect(StreamUtil.exactlyOne());
			}

			private int getAddrWidth(Bus bus, BusLayout busLayout) {
				return verifiableBuses.get(bus.getBusDescriptionName()).getMappedPortNames(bus, "addr")
						.map(portName -> getBpiFromBus(module, busLayout, portName))
						.mapToInt(IMappedBusPortInstance::getBaseWidth)
						.sum();
			}

			private void writeAddressVerification(Bus bus, IAddressChunk addressChunk, String verificationWire, int index, int watcherCount, WireNames wireNames, BusLayout busLayout) {
				writeComment("\tVerifying Address range 0x" + Long.toHexString(addressChunk.getActualStart()) + " length 0x" + Long.toHexString(addressChunk.getLength()) + " on " + bus.getName());
				writeModuleInstanceStart("bus_address_verification_watcher");
				writeModuleInstanceParams(Stream.of(
						new OutputParameter("BASE_ADR", VerilogLang.getMaskedAdr(addressChunk.getActualStart(), 18)),
						new OutputParameter("WORD_COUNT", VerilogLang.getMaskedAdr(addressChunk.getLength(), 18)),
						new OutputParameter("ADDR_WIDTH", Integer.toString(getAddrWidth(bus, busLayout)))
				));
				writeModuleInstanceNameAndPorts(verificationWire + "$" + index,
						generatePorts(bus, wireNames, busLayout, index, watcherCount, verificationWire)
				);
			}

			private String formatAddresses(List<StreamUtil.Pair<Bus, IAddressChunk>> addresses, int addrWidth) {
				return addresses.stream()
						.map(StreamUtil.Pair.accessF((bus, addr) -> String.format("[%X-%X]", VerilogLang.maskAddress(addr.getActualStart(), addrWidth), VerilogLang.maskAddress(addr.getActualEnd() - 1, addrWidth))))
						.collect(Collectors.joining(", "));
			}

			private void writeVerificationTop(String verificationWire, List<StreamUtil.Pair<Bus, IAddressChunk>> addresses, WireNames wireNames, BusLayout busLayout, String locationName) {

				//Get bus with widest address
				final Bus bus = buses.stream().sorted(Comparator.comparing(b -> -getAddrWidth(b, busLayout))).limit(1)
						.findFirst().orElseThrow(() -> new RuntimeException("No bus"));

				final BusVerificationInfo busVerificationInfo = verifiableBuses.get(bus.getBusDescriptionName());

				writeModuleInstanceStart("bus_address_verification_top");
				final IMappedBusPortInstance dataPort = getBpiFromBus(module, busLayout, doPort);
				final int addrWidth = getAddrWidth(bus, busLayout);
				writeModuleInstanceParams(Stream.of(
						new OutputParameter("COUNT", Integer.toString(addresses.size())),
						new OutputParameter("LOCATION", '"' + locationName + '"'),
						new OutputParameter("DATA_WIDTH", Integer.toString(dataPort.getBaseWidth())),
						new OutputParameter("ADDR_WIDTH", Integer.toString(addrWidth)),
						new OutputParameter("ADDRESS_RANGE", '"' + formatAddresses(addresses, addrWidth) + '"')
				));
				writeModuleInstanceNameAndPorts(verificationWire + "$top",
						Stream.of(
								generateOutputPort(bus, wireNames, busLayout, busVerificationInfo, "clk"),
								generateOutputPort(bus, wireNames, busLayout, busVerificationInfo, "reset"),
								generateOutputPort(bus, wireNames, busLayout, busVerificationInfo, "addr"),
								generateOutputPort(bus, wireNames, busLayout, busVerificationInfo, "busy"),
								generateOutputPort(bus, wireNames, busLayout, busVerificationInfo, "access"),
								//generateOutputPort(bus, wireNames, busLayout, busVerificationInfo, "data"),
								//generateOutputPortFromBusPortName(bus, wireNames, busLayout, busVerificationInfo, "data", doPort),
								new OutputPort("data", generateOutputPortConnectionFromBpi(wireNames, dataPort), ConnectorDirection.Input),
								new OutputPort("select", verificationWire, ConnectorDirection.Input)
						)
				);
			}

			public void output(IGlobalAddressLayout globalAddressLayout, WireNames wireNames, BusLayout busLayout) {
				final String busNames = buses.stream().map(Bus::getName).collect(Collectors.joining(", "));
				final String locationName = busNames + " by " + module.getGlobalName() + " on " + doPort + " connected to " + subsystem.getGlobalName();
				writeComment("Verifying " + locationName);
				final List<StreamUtil.Pair<Bus, IAddressChunk>> addresses = getAddresses(globalAddressLayout);

				final String verificationWire = VerilogLang.getVerilogName("_$$verification$" + module.getGlobalName() + "$" + doPort + "$" + subsystem.getGlobalName());
				writeDeclaration("wire", verificationWire, addresses.size());

				for (int i = 0; i < addresses.size(); i++) {
					final StreamUtil.Pair<Bus, IAddressChunk> pair = addresses.get(i);
					writeAddressVerification(pair.getA(), pair.getB(), verificationWire, i, addresses.size(), wireNames, busLayout);
				}

				writeVerificationTop(verificationWire, addresses, wireNames, busLayout, locationName);
			}
		}

		private Stream<BusVerifier> createBusVerifiers(IModuleConfiguration<?> module, String doPort, List<Bus> buses) {

			final List<? extends ISubsystemConfiguration<?>> subsystems = buses.stream().map(bus ->
					//get connected subsystems
					module.getBusHandler(bus).streamOppositeConnectors()
							.map(DocumentUtil.findOriginalBusHandlerForPassthroughMapper(IConfigurationElement.ElementType.Subsystem))
							.map(IConnectorInstance::getSourceElement)
							.flatMap(StreamUtil.castOrEmpty(ISubsystemConfiguration.class))
							.map(e -> (ISubsystemConfiguration<?>) e)
							.collect(Collectors.toList())
			).collect(StreamUtil.atLeastOneOrEqual((l) -> "multiple non equal elements for " + doPort + " in " + buses + " on " + module.getGlobalName() + ": " + l, () -> "no elements for " + doPort + " in " + buses + " on " + module.getGlobalName()));


			return subsystems
					.stream()
					.map(StreamUtil.numberElements())
					.map(pair -> new BusVerifier(module, doPort, buses, pair.getB(), pair.getA()));
		}

		private void writeAddressVerificationForModule(IModuleConfiguration<?> module, WireNames wireNames, BusLayout busLayout) {
			IModuleDescription descr = module.getModuleDescription();

			//Get data out ports and buses that output on them
			final Map<String, List<Bus>> doPorts = descr.getBuses()
					.stream()
					.filter(this::isVerifiableBus)
					.filter(bus -> module.getBusHandler(bus).getConnectionCount() > 0)
					.flatMap(bus -> getVerificationDataOut(bus).map(port -> new StreamUtil.Pair<>(port, bus)))
					.collect(Collectors.groupingBy(StreamUtil.Pair::getA, Collectors.mapping(StreamUtil.Pair::getB, Collectors.toList())));

			//Get bus verifiers. One Entry in doPorts maps to one verifier for each parallel connection
			final List<BusVerifier> verifiers = doPorts.keySet().stream().sorted()
					.flatMap(doPort -> createBusVerifiers(module, doPort, doPorts.get(doPort)))
					.collect(Collectors.toList());

			verifiers.forEach(busVerifier -> busVerifier.output(cache.getGlobalAddressLayout(), wireNames, busLayout));
		}


		private void writeAddressVerification(IGlobalAddressLayout globalAddressLayout, WireNames wireNames, BusLayout busLayout) {

			writeIfdef("SIMULATION");
			document.streamModulesRecursive()
					.forEach(module -> writeAddressVerificationForModule(module, wireNames, busLayout));
			writeEndif();

		}

		private void writeAssigns(WireNames wireNames) {
			DocumentUtil.streamConnectorHandlers(document)
					.filter(IPinHandler::isConnected)
					.filter(Wire::connectPinHandlerWithAssign)
					.sorted(Comparator.comparing(IPinHandler::getPortName))
					.forEach(pinHandler -> {
						writeVerilogAssign(VerilogLang.getPinName(pinHandler), "", wireNames.getWireConnection(pinHandler), "", pinHandler.isInverted());
					});
		}

		@NotNull
		private WireNames buildWireNames(BusLayout busLayout) {
			WireNames wireNames = new WireNames();

			buildRegularConnectionNames(wireNames);
			buildBusConnectionNames(busLayout, wireNames);
			buildBusBreakoutConnectionNames(wireNames, busLayout);
			wireNames.addUnconnectedWire();
			return wireNames;
		}

		private void buildBusBreakoutConnectionNames(WireNames wireNames, BusLayout busLayout) {
			document.streamModulesRecursive()
					.forEach(module -> module.apply(new IModuleVisitor<Void>() {

						@Override
						public Void applyToPeripheralModule(IPeripheralModuleConfiguration module) {
							//Nothing to do
							return null;
						}

						@Override
						public Void applyToSubsystemModule(ISubsystemConfiguration module) {
							//Nothing to do
							return null;
						}

						@Override
						public Void applyToCommonModule(IModuleConfiguration<?> module) {
							//Nothing to do
							return null;
						}

						@Override
						public Void applyToBusBreakoutModule(IBusBreakoutModuleConfiguration<?> module) {
							buildBusBreakoutConnectionNames(wireNames, busLayout, module);
							return null;
						}

						@Override
						public Void applyToMemoryModule(IMemoryModuleConfiguration<?> module) {
							//Nothing to do
							return null;
						}
					}));
		}

		private void buildBusBreakoutConnectionNames(WireNames wireNames, BusLayout busLayout, IBusBreakoutModuleConfiguration<?> module) {
			busLayout.streamOf(module).forEach(busPort -> {
				IConnectorInstance source = module.getInternallyConnectedPort(busPort, busLayout);
				final Wire a = Wire.fromBusPortInstance(busPort);
				final Wire b = Wire.fromPort(source);
				wireNames.addWireConnection(
						a,
						b
				);

			});
		}

		private void buildRegularConnectionNames(WireNames wireNames) {
			document.getConnections().forEach(conn -> {
				Iterator<? extends IPortHandler<?>> it = conn.getPorts().iterator();
				final WirePart aWire = Wire.fromPortConnection(conn, it.next());
				final WirePart bWire = Wire.fromPortConnection(conn, it.next());
				wireNames.addWireConnection(aWire, bWire);
			});
		}

		private void buildBusConnectionNames(BusLayout busLayout, WireNames wireNames) {
			busLayout.streamMappedPortsSorted()
					.filter(IConnectorInstance::isConnected)
					.filter(p -> p.getDirection() != ConnectorDirection.Output)
					.forEach((IMappedBusPortInstance busPortInstance) -> {
						Wire wire = Wire.fromBusPortInstance(busPortInstance);
						Stream<WirePart> remoteWires = busPortInstance.streamConnections()
								.filter(c -> !c.isIgnore())
								.map((MappedBusConnection conn) -> Wire.fromBusPortConnection(busPortInstance, conn));

						if (busPortInstance.requiresBusGlue()) { // do not connect directly, but use
							Wire glueOut = Wire.fromGlueOutput(busPortInstance);
							Wire glueIn = Wire.fromGlueInput(busPortInstance);

							wireNames.addWireConnection(wire, glueOut);
							wireNames.addParallelConnections(glueIn, remoteWires);
						} else {
							if (busPortInstance.getCombineType() == CombineType.Parallel) {
								wireNames.addParallelConnections(wire, remoteWires);
							} else {
								wireNames.addWireConnections(wire, remoteWires);
							}
						}
					});
		}

		private void writePeripheralInstance(AddressLayout layout, IPeripheralModuleConfiguration<? extends ISubsystemConfiguration> peripheral, BusLayout busLayout, WireNames wireNames) {
			writeModuleInstanceStart(peripheral);
			ArrayList<OutputParameter> additional = new ArrayList<>(2);
			PeripheralModuleDescription moduleDescription = peripheral.getModuleDescription();
			if (moduleDescription.hasMemory() && layout != null) {
				Optional<AddressBlock<?>> dma = layout.findAddressBlock(moduleDescription.getMemory().getExposedOn(), peripheral);
				if (dma.isPresent()) {
					additional.add(new OutputParameter(moduleDescription.getMemory().getBaseAdrParam(), VerilogLang.getMaskedAdr(dma.get().getActualStart(), 18)));
				} else {
					cache.getIssueSink().report(ConfigurationIssue.createPeripheralExposesUnmapped(peripheral, peripheral.getParent(), "dma", "Generating Verilog Instances")); //TODO move to layout generation
				}
			}
			if (moduleDescription.hasRegisters() && layout != null) {
				Optional<AddressBlock<?>> reg = layout.findAddressBlock(moduleDescription.getRegisters().getExposedOn(), peripheral);
				if (reg.isPresent()) {
					additional.add(new OutputParameter(moduleDescription.getRegisters().getBaseAdrParam(), VerilogLang.getMaskedAdr(reg.get().getActualStart(), 18)));
				} else {
					cache.getIssueSink().report(ConfigurationIssue.createPeripheralExposesUnmapped(peripheral, peripheral.getParent(), "reg", "Generating Verilog Instances")); //TODO move to layout generation
				}
			}


			writeModuleInstanceParams(Stream.concat(additional.stream(), getModuleInstanceParams(peripheral)));
			writeModuleNameAndPorts(peripheral, busLayout, wireNames);
		}

		private void writeSubsystemInstance(ISubsystemConfiguration subsystem, BusLayout busLayout, WireNames wireNames) {
			writeModuleInstanceStart(subsystem);
			writeModuleInstanceParams(getModuleInstanceParams(subsystem));
			writeModuleNameAndPorts(subsystem, busLayout, wireNames);
		}

		private void writeModuleInstanceStart(IModuleConfiguration module) {
			if (module.getModuleDescription().getTcl() != null) { //TODO: remove quick fix by merging base-module with dynamic-structure-module
				writeModuleInstanceStart(module.getName());
			} else {
				writeModuleInstanceStart(module.getToplevel());
			}
		}

		private Stream<OutputParameter> getModuleInstanceParams(IModuleConfiguration module) {
			return DocumentUtil.streamAllParameters(module).filter(p -> !p.isVirtual()).map(OutputParameter::new);
		}

		private void writeModuleNameAndPorts(IModuleConfiguration module, BusLayout busLayout, WireNames wireNames) {
			writeModuleInstanceNameAndPorts(VerilogLang.getVerilogName(module.getGlobalName()),
					Stream.concat(stream(portsPerModule, module), busLayout.streamOf(module)),
					IConnectorInstance::getPortName,
					port -> port.isConnected()? wireNames.getWireConnection(port) : null,
					IConnectorInstance::getDirection
			);
		}

		private void writeModuleInstance(IModuleConfiguration module, BusLayout busLayout, WireNames wireNames, IGlobalAddressLayout globalAddressLayout) {
			module.apply(new IModuleVisitor<Void>() {

				private void applyRegular(IModuleConfiguration module) {
					writeModuleInstanceStart(module);
					writeModuleInstanceParams(getModuleInstanceParams(module));
					writeModuleNameAndPorts(module, busLayout, wireNames);
				}

				@Override
				public Void applyToBusBreakoutModule(IBusBreakoutModuleConfiguration<?> module) {
					//Nothing to do
					return null;
				}

				@Override
				public Void applyToMemoryModule(IMemoryModuleConfiguration module) {
					applyRegular(module);
					return null;
				}

				@Override
				public Void applyToPeripheralModule(IPeripheralModuleConfiguration module) {
					if (module.getModuleDescription().getTcl() == null ||
							!module.getModuleDescription().getTcl().omitVerilogInstatiation()) {
						AddressLayout addressLayout = null;
						if (module.getParent() instanceof ISubsystemConfiguration<?>) {
							addressLayout = globalAddressLayout.getLayout((ISubsystemConfiguration) module.getParent());
						}
						writePeripheralInstance(addressLayout, module, busLayout, wireNames);
					}
					return null;
				}

				@Override
				public Void applyToSubsystemModule(ISubsystemConfiguration module) {
					if (module.getModuleDescription().getTcl() == null ||
							!module.getModuleDescription().getTcl().omitVerilogInstatiation()) {
						writeSubsystemInstance(module, busLayout, wireNames);
					}
					return null;
				}

				@Override
				public Void applyToCommonModule(IModuleConfiguration<?> module) {
					applyRegular(module);
					return null;
				}
			});
		}


	}


}

