/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.verilog;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.BusBreakoutConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.IMappedBusPortInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.MappedBusConnection;

import java.util.Objects;

/**
 * Created by wenzel on 05.07.17.
 */
public class Wire implements IWirePart {

	public final String module;
	public final String port;
	public final int width;
	public final boolean invert;
	/**
	 * Whether the wire wants/needs to be the source of the net name. Needed for toplevel pins, wanted for outputs
	 */
	public final WireType wireType;

	public Wire(String module, String port, int width, WireType wireType, boolean invert) {
		this.module = module;
		this.port = Objects.requireNonNull(port);
		this.width = width;
		this.wireType = wireType;
		this.invert = invert;
	}

	public String getVerilogName(boolean invertedView) {
		String invertStr = (invert != invertedView? "~" : "");
		if (module.isEmpty()) {
			return invertStr + VerilogLang.getPinName(port);
		}

		return invertStr + VerilogLang.getVerilogName(module) + "$" + VerilogLang.getVerilogName(port);
	}

	@Override
	public String toString() {
		return getVerilogName(false) + VerilogLang.getVerilogBus(width);
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (!(o instanceof Wire)) {
			return false;
		}

		Wire wire = (Wire) o;

		if (width != wire.width) {
			return false;
		}
		if (module != null? !module.equals(wire.module) : wire.module != null) {
			return false;
		}
		if (!port.equals(wire.port)) {
			return false;
		}
		return wireType == wire.wireType;
	}

	@Override
	public int hashCode() {
		int result = module != null? module.hashCode() : 0;
		result = 31 * result + port.hashCode();
		result = 31 * result + width;
		result = 31 * result + wireType.hashCode();
		return result;
	}

	/**
	 * An inverted output cannot be connected directly. There needs to be an assign between.
	 *
	 * @param pinHandler the connector instance to be checked
	 * @return true if we need an assign
	 */
	public static boolean connectPinHandlerWithAssign(IPinHandler pinHandler) {
		if (pinHandler.getDirection() == ConnectorDirection.InOut && pinHandler.isInverted()) {
			throw new RuntimeException("Inverted inouts are not possible. " + pinHandler.getPortName());
		}
		//Inputs here are FPGA outputs
		return pinHandler.getDirection() == ConnectorDirection.Input;
	}

	public static Wire fromPort(IConnectorInstance port) {
		WireType type;
		boolean invert = false;
		if (port.getSourceElement() instanceof BusBreakoutConfiguration<?>) {
			type = WireType.AvoidSource;
		} else if (port.getSourceElement().getType() == IConfigurationElement.ElementType.Configuration) {
			IPinHandler pinHandler = (IPinHandler) port;
			if (connectPinHandlerWithAssign(pinHandler)) {
				type = WireType.AvoidSource;
				invert = false;
			} else {
				type = WireType.NeedSource;
				invert = pinHandler.isInverted();
			}
		} else if (port.getDirection() == ConnectorDirection.Output) {
			type = WireType.PreferSource;
		} else {
			type = WireType.Regular;
		}
		return new Wire(port.getSourceElement().getGlobalName(), port.getPortName(), port.getConnectorWidth(), type, invert);
	}

	public static WirePart fromPortConnection(IConnection<? extends IPortHandler<?>> conn, IPortHandler<?> port) {
		return new WirePart(fromPort(port), conn.getMyStart(port), conn.getMyEnd(port));
	}

	public static Wire fromBusPortInstance(IMappedBusPortInstance busPortInstance) {
		WireType wireType = WireType.Regular;

		if (busPortInstance.getSourceElement() instanceof BusBreakoutConfiguration<?>) {
			wireType = WireType.AvoidSource;
		} else if (busPortInstance.getDirection() == ConnectorDirection.Output) {
			wireType = WireType.PreferSource;
		}

		return new Wire(busPortInstance.getSourceElement().getGlobalName(), busPortInstance.getPortName(), busPortInstance.getConnectorWidth(), wireType, false);
	}

	public static WirePart fromBusPortConnection(IMappedBusPortInstance busPortInstance, MappedBusConnection connection) {
		return new WirePart(fromBusPortInstance(connection.getOppositeConnector(busPortInstance)),
				connection.getOppositeStart(busPortInstance),
				connection.getOppositeEnd(busPortInstance)
		);
	}

	private static Wire fromGlue(IMappedBusPortInstance busPortInstance, String portName, int width) {
		return new Wire(VerilogLang.getGlueName(busPortInstance), portName, width, WireType.AvoidSource, false);
	}

	public static Wire fromGlueInput(IMappedBusPortInstance busPortInstance) {
		return fromGlue(busPortInstance, "in", busPortInstance.getConnectorWidth() * busPortInstance.getConnectionCount());
	}

	public static Wire fromGlueOutput(IMappedBusPortInstance busPortInstance) {
		return fromGlue(busPortInstance, "out", busPortInstance.getConnectorWidth());
	}

	@Override
	public int getStart() {
		return 0;
	}

	@Override
	public int getEnd() {
		return width - 1;
	}

	@Override
	public Wire getWire() {
		return this;
	}
}
