/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.verilog;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.IMappedBusPortInstance;
import de.tu_darmstadt.rs.util.ConnectedGraph;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

/**
 * Compute the resulting names when connecting multiple wire bits
 */
public class WireNames extends ConnectedGraph<WireBit, WireNames.WireGroup> {


	public WireNames() {
	}

	/**
	 * Add many small connections to one wide wire
	 *
	 * @param smallWires
	 * @param wideWire
	 */
	public <T extends IWirePart> void addParallelConnections(Wire wideWire, Stream<T> smallWires) {
		int wideBit = 0;
		for (T small : (Iterable<T>) smallWires::iterator) {
			for (int smallBit = small.getStart(); smallBit <= small.getEnd(); smallBit++) {
				addWireConnection(new WireBit(wideWire, wideBit++), new WireBit(small.getWire(), smallBit));
			}
		}
	}

	private static Comparator<WireBit> byNameComparator =
			Comparator.comparing((WireBit b) -> b.wire.module)
					.thenComparing(b -> b.wire.port)
					.thenComparing(b -> b.bit);

	@Override
	protected WireBit getNewRepresentative(WireBit a, WireBit b) {
		if (a.wire.wireType == WireType.NeedSource && b.wire.wireType == WireType.NeedSource) {
			throw new RuntimeException("Multiple connected wires need to be source: " + a + " and " + b);
		} else if (a.wire.wireType.sourceAffinity == b.wire.wireType.sourceAffinity) {
			return byNameComparator.compare(a, b) >= 0? a : b;
		} else if (a.wire.wireType.sourceAffinity > b.wire.wireType.sourceAffinity) {
			return a;
		} else {
			return b;
		}
	}

	@Override
	protected WireGroup createGroup(WireBit element) {
		return new WireGroup(element);
	}

	/**
	 * Connect two wire bits
	 *
	 * @param a bit to connect
	 * @param b bit to connect
	 */
	public void addWireConnection(WireBit a, WireBit b) {
		connect(a,b);
	}

	/**
	 * Get the name of a wire bit
	 *
	 * @param in the wire bit
	 * @return the resulting name
	 */
	public WireBit getWireName(WireBit in) {
		final WireGroup wireGroup = getGroup(in);
		if (wireGroup == null) {
			throw new RuntimeException("Tried to get representative of unconnected WireBit "+in);
		}
		return wireGroup.representative;
	}

	/**
	 * Get all unconnected wireBits that have parts of their wire connected
	 *
	 * @return all unconnected wireBits
	 */
	private Stream<WireBit> getUnconnected() {
		//Find all wires
		return getActiveGroups().stream()
				.flatMap(g -> g.contained.stream())
				.map(b -> b.wire)
				.distinct()
				//Map to all bits
				.flatMap(w -> IntStream.range(0, w.width).mapToObj(i -> new WireBit(w, i)))
				//Filter out used
				.filter(b -> !containsElement(b));
	}

	/**
	 * Connect all parts of wires that are not otherwise connected to a new unused wire
	 */
	public void addUnconnectedWire() {
		final List<WireBit> unconnected = getUnconnected().collect(Collectors.toList());
		Wire unconnectedWire = new Wire("unused","unconnected",unconnected.size(),WireType.NeedSource, false);
		for (int i = 0; i < unconnected.size(); i++) {
			addWireConnection(unconnected.get(i), new WireBit(unconnectedWire, i));
		}
	}

	private Map<WireBit, Integer> unconnectedBits = new HashMap<>();

	private WireBit getUnconnectedBit(WireBit in) {
		return null;
	}

	/**
	 * Stream all wire bits that are used as name somewhere
	 *
	 * @return stream of all wireBits
	 */
	public Stream<WireBit> streamUsedWireBits() {
		return getActiveGroups().stream().map(g -> g.representative);
	}

	/**
	 * Get the names for all individual bits as one verilog concatenation
	 *
	 * @param wire the input wire
	 * @return verilog concatenation of all individual bits
	 */
	public String getWireConnection(Wire wire) {
		return getWireConnection(wire, 0, wire.width-1);
	}

	/**
	 * Get the names for sine individual bits as one verilog concatenation
	 *
	 * @param wire the input wire
	 * @param start lowest bit to include
	 * @param end highest bit to include
	 * @return verilog concatenation of all individual bits
	 */
	public String getWireConnection(Wire wire, int start, int end) {
		List<WireBitRange> conns = getWireConnectionList(wire, start, end);
		if (conns.size() == 1) {
			return conns.iterator().next().getVerilogName(wire.invert);
		} else {
			return IntStream.range(0, conns.size()).mapToObj(i -> conns.get(conns.size() - i - 1))
					.map(w -> w.getVerilogName(wire.invert)).collect(Collectors.joining(", ", "{ ", " }"));
		}
	}

	private List<WireBitRange> getWireConnectionList(Wire wire, int start, int end) {
		List<WireBitRange> result = new ArrayList<>();
		WireBitRange current = null;
		for (int bit = start; bit <= end; bit++) {
			WireBit next = getWireName(new WireBit(wire, bit));

			if (current != null && current.canAppend(next)) {
				current.append();
			} else {
				current = new WireBitRange(next);
				result.add(current);
			}
		}
		return result;
	}

	public String getWireConnection(IConnectorInstance port) {
		return getWireConnection(Wire.fromPort(port));
	}

	public void addWireConnections(IWirePart a, Stream<? extends IWirePart> bs) {
		bs.forEach(b -> addWireConnection(a, b));
	}

	public void addWireConnection(IWirePart a, IWirePart b) {
		if ((a.getEnd() - a.getStart()) != (b.getEnd() - b.getStart())) {
			throw new RuntimeException("Wires " + a + " and " + b + " are not the same width.");
		}
		for (int ai = a.getStart(), bi = b.getStart(); ai <= a.getEnd(); ai++, bi++) {
			addWireConnection(new WireBit(a.getWire(), ai), new WireBit(b.getWire(), bi));
		}
	}

	public Function<IMappedBusPortInstance, String> glueInputMapper() {
		return port -> getWireConnection(Wire.fromGlueInput(port));
	}

	public Function<IMappedBusPortInstance, String> glueOutputMapper() {
		return port -> getWireConnection(Wire.fromGlueOutput(port));
	}


	public static class WireBitRange extends WireBit {

		public int width = 1;

		public WireBitRange(Wire wire, int bit) {
			super(wire, bit);
		}

		public WireBitRange(WireBit wireBit) {
			super(wireBit.wire, wireBit.bit);
		}

		public boolean canAppend(WireBit next) {
			return next.wire.equals(wire) && next.bit == bit + width;
		}

		public void append() {
			width++;
		}

		@Override
		protected String getVerilogBus() {
			return VerilogLang.getVerilogBus(bit, bit + width - 1, width != wire.width, wire.width > 1);
		}

		@Override
		public String getVerilogName(boolean invertedView) {
			if (wire == null) {
				return width + "'bZ";
			}
			return super.getVerilogName(invertedView);
		}
	}


	/**
	 * A connected group of bits
	 */
	protected static class WireGroup extends ConnectedGraph.ConnectedComponent<WireBit> {
		public boolean hasInverted;

		public WireGroup(WireBit bit) {
			super(bit);
			hasInverted = bit.wire.invert;
		}
	}

	@Override
	protected void postMergeActions(WireGroup a, WireGroup b) {

		if (a.hasInverted && b.hasInverted) {
			throw new RuntimeException("Connection between " + a + " and " + b + " not possible. Both have inverted elements.");
		}

		a.hasInverted |= b.hasInverted;
	}
}
