/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.xilinx;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.BaseDocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.LazyBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.LegacyCoregenXcoBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.XilinxIseConstraintsBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.XilinxIseProjectBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.CPeriBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.linker.GnuLinkerMemoryBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.HeaderBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.LegacyMkBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.MemoryMapBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.TclDeviceInfoBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.VerilogSimulationBuilderProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.verilog.VerilogBuilderProvider;
import de.tu_darmstadt.rs.util.StreamUtil;

import java.nio.file.Path;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Builder for ISE/XST backed synthesis for Xilinx devices (basically everything up to (excl.) Virtex-7)
 * Created by ramon on 01.04.2016.
 */
public class XilinxIseDocumentBuilder extends BaseDocumentBuilder {

	public XilinxIseDocumentBuilder() {
		add(new VerilogBuilderProvider(this, XilinxIseDocumentBuilder::getDefiningValues));

		Function<Path, Stream<StreamUtil.Pair<String,String>>> constraints = (systemDir) -> Stream.of(
				new StreamUtil.Pair<>("JCONFIG_CONSTRAINTS_FILE", XilinxIseConstraintsBuilderProvider.getConstraintsFile(systemDir).toString())
		);
		add(new LegacyMkBuilderProvider(this, constraints));
		add(new VerilogSimulationBuilderProvider(this));
		add(new MemoryMapBuilderProvider(this));
		add(new XilinxIseConstraintsBuilderProvider(this));
		add(new GnuLinkerMemoryBuilderProvider(this));
		add(new HeaderBuilderProvider(this));
		add(new CPeriBuilderProvider(this));
		add(new XilinxIseProjectBuilderProvider(this));
		add(new LegacyCoregenXcoBuilderProvider(this));
		add(new TclDeviceInfoBuilderProvider(this));
	}

	public static String getDefiningValues(IPinHandler pinHandler) {
		CalculatedValueResult<Long> frequency = pinHandler.getFrequency();
		String freqString = frequency.map(Object::toString).orElse(null);
		return Stream.of(
				pinHandler.getCustomUserConstraints(),
				pinHandler.getDrive(), pinHandler.getIoStd(),
				pinHandler.getSlewRate(), getUcfName(pinHandler.getTermination()),
				freqString,
				pinHandler.getDirection().toString(),
				Boolean.toString(pinHandler.isInverted())
		).collect(Collectors.joining(","));
	}

	public static String getUcfName(DeviceDescription.Termination termination) {
		switch (termination) {
			case PullDown:
				return "PULLDOWN";
			case PullUp:
				return "PULLUP";
			default:
			case None:
				return null;
		}
	}

	@Override
	protected IBuildProcess createBuildProcessCache(IConfigurationDocument document) {
		return new LazyBuildProcess(document);
	}

	@Override
	public String getBuilderName() {
		return "DefaultXilinxIseBuilder";
	}

	@Override
	public String getDescription() {
		return XilinxIseToolchainProvider.XILINX_ISE_TOOLCHAIN;
	}

	@Override
	public String getToolchainName() {
		return XilinxIseToolchainProvider.ISE;
	}

}
