/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.handlers;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ResolvedConnector;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers.BasePortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers.BusHandler;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.WeakHashMap;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * @author Ramon Wirsch
 */
public class AvailablePortsHandler implements IAvailablePortsHandler {

	private final WeakHashMap<IModuleConfiguration, List<BasePortHandler>> portsPerElement = new WeakHashMap<>();
	private final ConfigurationDocument document;
	private final WeakHashMap<IConfigurationElement, List<ResolvedConnector>> resolvedPortsCache = new WeakHashMap<>();
	private final WeakHashMap<IBusHandler, List<ResolvedConnector>> resolvedBusCache = new WeakHashMap<>();
	private final List<IAvailablePortsListener> listeners = new ArrayList<>();
	private List<IModuleConfiguration<?>> portSources;
	private List<ResolvedConnector> resolvedPins;

	public AvailablePortsHandler(ConfigurationDocument document) {
		this.document = document;

		portSources = document.streamModulesRecursive().collect(Collectors.toList());

		onTargetChange();
	}

	@Override
	public void onUndefinedChange() {
		portSources = document.streamModulesRecursive().collect(Collectors.toList());

		portSources.clear();
		resolvedPortsCache.clear();
		resolvedBusCache.clear();
		portsPerElement.clear();

		onTargetChange();
	}

	@Override
	public void onTargetChange() {
		ConfigurationDocument.TargetConfiguration target = document.getTarget();
		resolvedPins = target.getTargetDescription().getIoPinGroups().stream().flatMap(ioPinGroup -> ioPinGroup.getPorts().stream()).map(document::getPortHandler).map(p -> new ResolvedConnector(p, document)).collect(Collectors.toList());
		notifyAvailablePortsChanged();
	}

	@Override
	public void moduleAdded(IModuleConfiguration<?> module) {
		portSources.add(module);
		List<BasePortHandler> newPorts = findPorts((ModuleConfiguration) module);

		resolvedPortsCache.forEach((m, ports) -> {
			List<ResolvedConnector> newRes = newPorts.stream().map(p -> new ResolvedConnector(p, m)).collect(Collectors.toList());
			ports.addAll(newRes);
			ports.sort(Comparator.naturalOrder());
		});
		resolvedBusCache.clear();
		notifyAvailablePortsChanged();
	}

	@Override
	public void moduleRemoved(IModuleConfiguration<?> module) {
		portSources.remove(module);
		resolvedPortsCache.clear();
		resolvedBusCache.clear();
		portsPerElement.remove(module);
		notifyAvailablePortsChanged();
	}

	/**
	 * Invalidate ports from that module
	 *
	 * @param module A dynamic module (like chipscope, varying count of params / ports)
	 */
	@Override
	public void moduleDeclarationChanged(IModuleConfiguration<?> module) {
		resolvedPortsCache.put(module, null);
		portsPerElement.put(module, null);
		notifyAvailablePortsChanged();
	}

	@Override
	public void busDefiningParametersChanged(IModuleConfiguration<?> module, IBusHandler bus) {
		resolvedPortsCache.remove(module);
		portsPerElement.remove(module);
		resolvedBusCache.remove(bus);
		notifyAvailablePortsChanged();
	}

	@Override
	public void addListener(IAvailablePortsListener listener) {
		listeners.add(listener);
	}

	@Override
	public void removeListener(IAvailablePortsListener listener) {
		listeners.remove(listener);
	}

	/**
	 * Filter out ports that do not match direction
	 *
	 * @param relativeTo current module
	 * @param portHandler port handler to connect
	 * @return all possible ResolvedConnectors
	 */
	@Override
	public List<ResolvedConnector> getResolvedPortsFor(IConfigurationElement<?, ?, ?> relativeTo, IPortHandler portHandler) {
		List<ResolvedConnector> resolvedPorts = getResolvedPortsFor(relativeTo);
		if (!portHandler.getConnectorDescription().isDirectionLocked()) {
			return resolvedPorts;
		} else {
			return resolvedPorts.stream().filter(r -> {
				IConnectorDescription from = portHandler.getConnectorDescription();
				IConnectorDescription to = r.getConnectorHandler().getConnectorDescription();
				return IConnectorDescription.isDirectionCompatible(from, to) && IConnectorDescription.isTypeCompatible(from, to) && portHandler!=r.getConnectorHandler();
			}).collect(Collectors.toList());
		}
	}

	@Override
	public List<ResolvedConnector> getResolvedPortsFor(IConfigurationElement<?, ?, ?> relativeTo) {
		List<ResolvedConnector> cache = resolvedPortsCache.get(relativeTo);
		if (cache == null) {
			cache = Stream.concat(
				portSources.stream().flatMap(s -> findPorts((ModuleConfiguration) s).stream().map(p -> new ResolvedConnector(p, relativeTo))).sorted(),
				resolvedPins.stream()
			).collect(Collectors.toList());
			resolvedPortsCache.put(relativeTo, cache);
		}
		return cache;
	}

	@Override
	public List<ResolvedConnector> getResolvedBusesFor(IConfigurationElement<?, ?, ?> relativeTo, IBusHandler busHandler) {
		List<ResolvedConnector> cache = resolvedBusCache.get(busHandler);
		if (cache == null) {
			cache = findBusesOfType(busHandler.getBusType(), relativeTo).filter(b -> b.isConnectibleTo(busHandler) || (b instanceof BusHandler) && ((BusHandler) b).couldBeConnectibleTo(busHandler)).map(b -> new ResolvedConnector(b, relativeTo)).sorted().collect(Collectors.toList());
			resolvedBusCache.put(busHandler, cache);
		}
		return cache;
	}

	private void notifyAvailablePortsChanged() {
		listeners.forEach(IAvailablePortsListener::onAvailablePortsChanged);
	}

	private List<BasePortHandler> findPorts(ModuleConfiguration<?> source) {
		List<BasePortHandler> portHandlers = portsPerElement.get(source);
		if (portHandlers == null) {
			portHandlers = source.getPortGroupDescriptions().stream().flatMap(pg -> pg.getPorts().stream()).map(source::getPortHandler).collect(Collectors.toList());
			portsPerElement.put(source, portHandlers);
		}
		return portHandlers;
	}

	private Stream<IBusHandler> findBusesOfType(String busType, IConfigurationElement<?, ?, ?> relativeTo) {
		return portSources.stream().filter(m -> m != relativeTo).flatMap(m -> m.getBusDescriptions().stream().filter(d -> d.getBusDescriptionName().equals(busType)).map(m::getBusHandler));
	}

}
