/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.handlers;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.INamedValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.SubsystemModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPluginEngine;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.exceptions.JConfigBuildValidationException;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IModuleContainer;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentExporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentImporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager.BaseBuildManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager.BuildResult;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager.IncrementalBuildManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.DocumentBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.toolchains.ToolchainRegistry;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.SubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.marshalling.DocumentJaxB;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.IDocumentStatusReporter;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibraryUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.TargetConf;

import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.Optional;
import java.util.function.Function;
import javax.xml.bind.JAXBException;

import static de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription.ModuleType.Subsystem;
import static java.nio.file.StandardCopyOption.REPLACE_EXISTING;

/**
 * Wrapper around a single JConfig document. Files are loaded, stored through this class.
 * Use 1 instance per possibly opened document
 * Part of a Tiered MVC Controller
 */
public class DocumentHandler implements IDocumentHandler, IDocumentImporter.IDocumentSupplier<ConfigurationDocument> {

	private static final Logger logger = LoggerFactory.getLogger(DocumentHandler.class);

	private final IApplicationConfiguration application;
	private final List<IDocumentHandlerListener> documentHandlerListeners = new ArrayList<>();
	private final LibJConfigProject project;
	private Path saveLocation;
	private Path preSaveLocation;
	private DocumentJaxB jaxB;
	private ConfigurationDocument document;
	private boolean hasChangedSinceSave;
	private IPluginEngine pluginEngine;
	private ToolchainRegistry toolchainRegistry;
	private List<IPluginEngineListener> pluginEngineListeners = new ArrayList<>();
	private IDocumentStatusReporter statusReporter;
	private List<Path> noCleanupFiles = new ArrayList<>();

	public DocumentHandler(IApplicationConfiguration application, LibJConfigProject project) {
		this.project = project;
		this.application = application;
	}

	private static Path getPreSaveFile(Path saveLocation) {
		String name = saveLocation.getFileName().toString();
		return saveLocation.resolveSibling(name + ".temp");
	}

	@Override
	public synchronized boolean needsSaving() {
		return hasChangedSinceSave;
	}

	@Override
	public synchronized ConfigurationDocument newDocument(TargetDescription target, DeviceDescription device) {
		releaseDocument();

		document = createDocument(target, device);

		//		setSaveLocation(project.getWorkingDir().resolve("new.unsaved.jc3"));
		setSaveLocation(null);
		hasChangedSinceSave = true;

		notifyDocumentSwitched();
		return document;
	}

	@Override
	public IConfigurationDocument newDocument(String targetConf) {
		TargetDescription target = getLibrary().getTargetDescriptionWithAdvancedSearch(TargetConf.extractTargetName(targetConf));
		DeviceDescription device = LibraryUtil.getDeviceDescriptionOrCompatible(getLibrary(), target, TargetConf.extractDeviceName(targetConf));
		return newDocument(target, device);
	}

	@Override
	public synchronized DocumentBuilder getDocumentBuilder() {
		logger.info("Building document {}", document);
		if (pluginEngine != null) {
			pluginEngine.onBuilding(document);
		}

		DocumentBuilder builder = DocumentBuilder.getBuilder(document);// find a builder that can build what we need

		logger.info("Building with {}", builder.getDescription());//TODO use IToolchainProvider.getDescription()

		return builder;
	}

	@Override
	public synchronized void buildIncremental() {
		build(true);
	}

	@Override
	public synchronized void build(boolean incremental) {
		buildWith(IncrementalBuildManager::new, true, incremental);
	}



	@Override
	public synchronized void buildWith(Function<DocumentBuilder, BaseBuildManager> buildManagerSupplier, boolean requireValidate, boolean incremental) {
		if (requireValidate && !isValid()) {
			throw new JConfigBuildValidationException("Pre-Build validation failed. See Issue List");
		}

		BaseBuildManager manager = buildManagerSupplier.apply(getDocumentBuilder());
		manager.setNoCleanupFiles(noCleanupFiles);

		BuildResult result = incremental? manager.buildRequired(document) : manager.buildAll(document);
		if (!result.wasSuccessful()) {
			result.getExceptions().stream().forEach(e -> logger.error("Failed to build", e));
			reportDocumentStatus("Build failed! See issues / log", true);
			throw new RuntimeException("Build Failed", result.getExceptions().get(0));
		} else {
			result.report();
			reportDocumentStatus("Build successfull", false);
		}
	}

	@Override
	public synchronized boolean isValid() {
		document.checkResolvedIssues();

		logger.info("Validating document {}", document);
		boolean valid = document.isValid(true);
		boolean modulesValid = document.streamModulesRecursive().reduce(true, (v, module) -> module.isValid(true) && v, Boolean::logicalAnd);//force evaluating all, because isValid() is where modules will create Issues, that highlight
		IToolchainConfiguration currentConfiguration = document.getTarget().getCurrentConfiguration();
		boolean toolchainValid = document.streamModulesRecursive().reduce(true, (v, desc) -> DocumentUtil.checkModuleSupportsToolchain(desc, document.getIssueSink(), currentConfiguration) && v, Boolean::logicalAnd);
		// inconsistencies in UI
		boolean noCriticalIssues = document.getConfigurationIssues().stream().noneMatch(i -> i.getLevel() == ConfigurationIssue.IssueType.Error);

		boolean allValid = valid && modulesValid && toolchainValid && noCriticalIssues;

		if (allValid) {
			reportDocumentStatus("Validation successfull!", false);
		} else {
			reportDocumentStatus("Validation failed, see issues / log", true);
		}

		return allValid;
	}

	@Override
	public synchronized boolean hasSaveLocation() {
		return saveLocation != null;
	}

	@Override
	public synchronized void save() {
		if (document == null) {
			logger.error("No document open, nothing to do");
			return;
		}

		if (pluginEngine != null) {
			pluginEngine.onSaving(document);
		}

		ensureJaxB();

		try {
			document.setJconfigVersion(getVersionString());
			jaxB.marshall(document, preSaveLocation);
			try {
				Files.move(preSaveLocation, saveLocation, REPLACE_EXISTING);
				hasChangedSinceSave = false;
				logger.info("Saved document {} to {}", document, saveLocation);
				reportDocumentStatus("Document saved", false);
			} catch (IOException e) {
				logger.error("Failed to replace existing file", e);
			}
		} catch (Exception e) {
			logger.error("Failed to save document, existing file left unmodified", e);
		}
	}

	@Override
	public synchronized void importFile(Path source, IDocumentImporter reader) {
		releaseDocument();

		if (reader.requiresLibrary() && !application.isLibraryReady()) {
			throw new IllegalStateException(String.format("Importer %s requires library", reader.toString()));
		}

		document = reader.importDocument(source, this);

		if (document == null) {
			throw new IllegalStateException(String.format("Import with %s of %s failed", reader.toString(), source.toString()));
		}

		setSaveLocation(null);
		logger.info("Imported Document {}: {}", source, document);
		this.hasChangedSinceSave = true;
		configureDocument(document);

		if (pluginEngine != null) {
			pluginEngine.onImportedDocument(document, reader);
		}

		notifyDocumentSwitched();
	}

	@Override
	public synchronized void exportFile(Path destination, IDocumentExporter writer) {
		if (pluginEngine != null) {
			pluginEngine.onExportingDocument(document, writer);
		}

		try {
			Path preSave = getPreSaveFile(destination);
			writer.exportDocument(document, preSave);
			try {
				Files.move(preSave, destination, REPLACE_EXISTING);
			} catch (IOException e) {
				logger.error("Failed to replace existing file", e);
			}
		} catch (Exception e) {
			logger.error("Failed to save destination, existing file left unmodified", e);
		}
		hasChangedSinceSave = false;
		logger.info("Exported Document {} to {}", document, destination);
	}

	@Override
	public synchronized Optional<Path> findAutosave() {
		return project.findNewestAutoSave();
	}

	@Override
	public synchronized void open(Path document) {
		releaseDocument();
		ensureJaxB();

		try {
			this.document = jaxB.unmarshall(document);
		} catch (JAXBException e) {
			throw new RuntimeException(e);
		}

		setSaveLocation(document);
		logger.info("Opened document {}: {}", document, this.document);
		hasChangedSinceSave = false;
		configureDocument(this.document);

		application.addOrCallReadyListener(this.document::attachLibrary);

		if (pluginEngine != null) {
			pluginEngine.onOpenedDocument(this.document);
		}

		notifyDocumentSwitched();
		reportDocumentStatus("Document opened", false);
	}

	@Override
	public synchronized AvailablePortsHandler getAvailablePortsHandler() {
		return document.getAvailablePortsHandler();
	}

	@Override
	public synchronized Path getSaveLocation() {
		return saveLocation;
	}

	@Override
	public synchronized void setSaveLocation(Path saveLocation) {
		this.saveLocation = saveLocation;
		if (saveLocation != null) {
			logger.info("changed save location to {}", saveLocation);
			this.preSaveLocation = getPreSaveFile(saveLocation);
		}
		notifySaveLocationChanged();
	}

	@Override
	public synchronized Path getWorkingDir() {
		return project.getWorkingDir();
	}

	@Override
	public synchronized boolean hasLoadedDocument() {
		return document != null;
	}

	@Override
	public synchronized ConfigurationDocument getDocument() {
		return document;
	}

	@Override
	public synchronized <P extends IModuleContainer & IConfigurationElement & INamedValueResolver> ModuleConfiguration<P> addModule(IConfigurationElement<?, ?, ?> parent, IModuleDescription moduleDescription) {
		switch (parent.getType()) {
			case Configuration:
				ConfigurationDocument document = (ConfigurationDocument) parent;
				if (this.document != parent) {
					break;
				}

				if (moduleDescription.getType() == Subsystem) {
					return (ModuleConfiguration<P>) document.addSubsystem(((SubsystemModuleDescription) moduleDescription));
				} else {
					return (ModuleConfiguration<P>) document.addModule(moduleDescription);
				}
			case Subsystem:
				if (moduleDescription.getType() != IModuleDescription.ModuleType.Subsystem) {
					SubsystemConfiguration subsystemConfiguration = ((SubsystemConfiguration) parent);
					if (!this.document.getSubsystems().contains(subsystemConfiguration)) {
						break;
					}
					return (ModuleConfiguration<P>) subsystemConfiguration.addModule(moduleDescription);
				}
				break;
		}
		throw new IllegalArgumentException("Incompatible elementTypes");
	}

	@Override
	public synchronized boolean removeModule(IConfigurationElement<?, ?, ?> parent, IConfigurationElement<?, ?, ?> child) {
		return removeModuleInternal(parent, child, false, false);
	}

	public synchronized boolean removeModuleInternal(IConfigurationElement<?, ?, ?> parent, IConfigurationElement<?, ?, ?> child, boolean internal, boolean invalid) {
		if (parent == null) {
			throw new IllegalArgumentException("Need a parent to remove from / Root cannot be removed.");
		}
		switch (parent.getType()) {
			case Configuration:
				return ((ConfigurationDocument) parent).removeModuleInternal((ModuleConfiguration) child, internal, invalid);
			case Peripheral:
			case CommonModule:
				throw new UnsupportedOperationException("Peripherals/GlueModules do not have children");
			case Subsystem:
				return ((SubsystemConfiguration) parent).removeModuleInternal((ModuleConfiguration) child, internal, invalid);
			default:
				throw new IllegalArgumentException("Incompatible elementTypes");
		}
	}

	@Override
	public synchronized boolean removeModule(IConfigurationElement<?, ?, ?> module) {
		return removeModule(module.getParent(), module);
	}

	@Override
	public synchronized void addDocumentHandlerListener(IDocumentHandlerListener listener) {
		documentHandlerListeners.add(listener);
	}

	@Override
	public synchronized void removeDocumentHandlerListener(IDocumentHandlerListener listener) {
		documentHandlerListeners.remove(listener);
	}

	@Override
	public synchronized boolean isModuleInUse(IModuleConfiguration module) {
		return (pluginEngine != null) && pluginEngine.isModuleInUse(module);
	}

	@Override
	public boolean hasPluginEngine() {
		return pluginEngine != null;
	}

	@Override
	public synchronized IPluginEngine getPluginEngine() {
		return pluginEngine;
	}

	@Override
	public synchronized void setPluginEngine(IPluginEngine pluginEngine) {
		this.pluginEngine = pluginEngine;

		notifyPluginEngineChanged(pluginEngine);

		if (pluginEngine != null) {
			application.addOrCallReadyListener(library -> pluginEngine.onAttached(project));
		}

		if (document != null) {
			distributePluginEngine(document, pluginEngine);
		}
	}

	private static void distributePluginEngine(ConfigurationDocument document, IPluginEngine pluginEngine) {
		document.setPluginEngine(pluginEngine);
		document.streamModulesRecursive().forEach((m) -> m.setPluginEngine(pluginEngine));
	}

	@Override
	public void addPluginEngineListener(IPluginEngineListener listener) {
		pluginEngineListeners.add(listener);
	}

	@Override
	public void removePluginEngineListener(IPluginEngineListener listener) {
		pluginEngineListeners.remove(listener);
	}

	@Override
	public synchronized void setDocumentStatusReporter(@Nullable IDocumentStatusReporter statusReporter) {
		this.statusReporter = statusReporter;
	}

	@Override
	public void doNotCleanupFile(Path file) {
		noCleanupFiles.add(file);
	}

	private void notifyPluginEngineChanged(IPluginEngine pluginEngine) {
		pluginEngineListeners.forEach(l -> l.onPluginEngineChanged(pluginEngine));
	}

	@Override
	public ConfigurationDocument createDocument(TargetDescription target, DeviceDescription device) {
		if (target.getDevices().stream().noneMatch(d -> d.getName().equals(device.getName()))) {
			throw new IllegalArgumentException("Device '" + device + "' is not compatible with Target '" + target + "'!");
		}

		ConfigurationDocument document = new ConfigurationDocument("Configuration", target, device, getVersionString(), this, application.getLibrary());


		configureDocument(document);

		if (pluginEngine != null) {
			pluginEngine.onCreatedDocument(document);
		}

		logger.info("Created new document {} of {}, {}", document, target, device);

		return document;
	}

	@Override
	public Library getLibrary() {
		return application.getLibrary();
	}

	public synchronized LibJConfigProject getProject() {
		return project;
	}

	private String getVersionString() {
		return project.getLibJConfigVersion() + ';' + application.getLibrary().getLibDevXmlVersion();
	}

	/**
	 * Attach us to document
	 * @param document
	 */
	private void configureDocument(ConfigurationDocument document) {
		document.initialize(this);
		distributePluginEngine(document, pluginEngine);
	}

	private synchronized void reportDocumentStatus(String status, boolean needsAttention) {
		logger.info("Document status: {}", status);
		if (statusReporter != null) {
			statusReporter.setStatusText(status, needsAttention);
		}
	}

	private void ensureJaxB() {
		Locale defaultLocale = Locale.getDefault();
		Locale.setDefault(Locale.ENGLISH);

		try {
			jaxB = new DocumentJaxB();
		} catch (JAXBException e) {
			throw new RuntimeException(e);
		} finally {
			Locale.setDefault(defaultLocale);
		}
	}

	private void releaseDocument() {
		if (document != null) {
			logger.debug("releasing old document {}", document);
			document.setPluginEngine(null);
			document.streamModulesRecursive().forEach(m -> m.setPluginEngine(null));
		}
	}


	private synchronized void notifySaveLocationChanged() {
		for (IDocumentHandlerListener l : documentHandlerListeners) {
			l.onSaveLocationChanged();
		}
	}


	private synchronized void notifyDocumentSwitched() {
		boolean valid = hasLoadedDocument();

		for (IDocumentHandlerListener l : documentHandlerListeners) {
			l.onDocumentSwitched(valid);
		}
	}

	public synchronized void onDocumentChange() {
		if (!hasChangedSinceSave) {
			logger.debug("Document changed since last save!");
		}
		hasChangedSinceSave = true;
		reportDocumentStatus("changed", false);
	}


}
