/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.internal;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HdlDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.TclDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.SubsystemConfiguration;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.stream.Collector;

/**
 * Created by Ramon on 17.08.2016.
 */
public class SourceFileCollector {

	private static final Logger logger = LoggerFactory.getLogger(SourceFileCollector.class);

	private final Set<IModuleDescription> handledDescription = new HashSet<>();
	private final Set<IModuleConfiguration<?>> handledDynamicConfigurations = new HashSet<>();
	private final SortedSet<String> includedModules = new TreeSet<>();
	private final Map<Path, SourceRef> sources = new HashMap<>();
	private final ILibJConfigProject project;
	private final Library library;
	private final IToolchainConfiguration currentConfig;

	public SourceFileCollector(ILibJConfigProject project, Library library, Collection<IModuleDescription> additionalModules, Collection<SourceRef> initialContent) {
		this.project = Objects.requireNonNull(project);
		this.library = library;

		final IConfigurationDocument document = project.getDocumentHandler().getDocument();
		if (document == null) {
			throw new NullPointerException("No document!");
		}
		this.currentConfig = document.getTarget().getCurrentConfiguration();

		initialContent.forEach(this::add);

		if (additionalModules != null) {
			additionalModules.forEach(this::add);
		}
	}

	public static Collector<? super IModuleDescription, SourceFileCollector, Collection<SourceRef>> createDescription(ILibJConfigProject project, Library library, Collection<SourceRef> initialContent) {
		return Collector.of(
				() -> new SourceFileCollector(project, library, null, initialContent),
				SourceFileCollector::add,
				(acc, merge) -> {
					throw new UnsupportedOperationException();
				},
				SourceFileCollector::getFiles);
	}

	public static Collector<? super IModuleConfiguration<?>, SourceFileCollector, Collection<SourceRef>> create(ILibJConfigProject project, Library library, Collection<IModuleDescription> additionalModules, Collection<SourceRef> initialContent) {
		return Collector.of(
				() -> new SourceFileCollector(project, library, additionalModules, initialContent),
				SourceFileCollector::add,
				(acc, merge) -> {
					throw new UnsupportedOperationException();
				},
				SourceFileCollector::getFiles);
	}

	private void add(SourceRef ref) {
		final SourceRef existing = this.sources.get(ref.getPath().toAbsolutePath());
		if (existing != null) {
			if (!existing.equals(ref)) {
				throw new RuntimeException("Duplicate ref to file with different props: " + existing + ", " + ref);
			}
		} else {
			this.sources.put(ref.getPath().toAbsolutePath(), ref);
		}
	}

	private void addDynamicStructureModule(IModuleConfiguration<?> module) {
		if (handledDynamicConfigurations.add(module)) {

			Path coregenPath = project.getBuildDir().resolve("ipcore");

			// find memory in subsystem
			Path firmwareBuildPath;
			if (module.getParent() != null && module.getParent() instanceof SubsystemConfiguration) {
				// parent should be Subsystem, otherwise there is no firmware
				firmwareBuildPath = project.getBuildDir().resolve("firmware").resolve(module.getParent().getGlobalName());
			} else {
				firmwareBuildPath = null;
			}
			HdlDescription.Sources sources = module.getModuleDescription().getHdl().getSources();
			String lib = sources.getLibrary();
			final List<HdlDescription.PlaceholderFile> coregenFiles = sources.getCoregenFiles();
			if (coregenFiles != null) {
				coregenFiles.stream()
						.map(f -> new SourceRef(f.resolve(coregenPath, firmwareBuildPath, module.getToplevel()), lib, SourceRef.FileType.fromBoolean(f.isSimulationOnly())))
						.forEach(this::add);
			}
		}
	}

	private void add(IModuleConfiguration<?> module) {
		addDynamicStructureModule(module);
		IModuleDescription description = module.getModuleDescription();
		add(description);
	}

	private void add(IModuleDescription description) {
		if (handledDescription.add(description)) {
			description.getHdl().getSources().streamAllResolvedPaths(description, currentConfig).forEach(this::add);

			TclDescription tclDesc = description.getTcl();
			if (tclDesc != null) {
				add(tclDesc.getBuildScriptPath(description));
			}

			includedModules.addAll(description.getHdl().getSources().getModules());
		}
	}

	private Collection<SourceRef> getFiles() {
		while (!includedModules.isEmpty()) {
			String next = includedModules.first();
			IModuleDescription description = library.getModuleDescription(next);
			if (description == null) {
				throw new IllegalArgumentException("Could not find required module " + next);
			}
			add(description);
			includedModules.remove(next);
		}

		return sources.values();
	}

}
