/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.INamedValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ParameterReference;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dynamic.IDynamicParameterResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameterGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.Parameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPortGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.DynamicStructureModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IModuleContainer;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IIntegerParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers.BaseParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import javax.xml.bind.Marshaller;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * Created by Ramon on 18.02.2016.
 */
@XmlRootElement(name = "dynamic-commonModule", namespace = "http://www.rs.tu-darmstadt.de/ns/spartanmc/configuration/jc3")
public final class DynamicModuleConfiguration<P extends IModuleContainer<?> & IConfigurationElement<?, ?, ?> & INamedValueResolver> extends CommonModuleConfiguration<P> implements IDynamicParameterResolver {

	private static final Logger logger = LoggerFactory.getLogger(DynamicModuleConfiguration.class);
	private List<? extends IParameterGroup> parameterGroupsCached;
	private List<? extends IPortGroup> portGroupsCached;
	private Set<String> parameterNamesCached;

	/**
	 * Only for JAXB
	 */
	private DynamicModuleConfiguration() {
	}

	DynamicModuleConfiguration(String name, DynamicStructureModuleDescription moduleDescription, P parent, Library library) {
		super(name, moduleDescription, parent, library);
	}

	@Override
	public ModuleDescription getModuleDescription() {
		return getDynamicModuleDescription();
	}

	public DynamicStructureModuleDescription getDynamicModuleDescription() {
		return library.getDynamicCommonDescription(getModuleDescriptionName());
	}

	@Override
	protected <T extends IAttributeHandler> T getParameterHandlerByName(Class<T> limit, IParameter parameter, String parameterName) {
		BaseParameterHandler handler = (BaseParameterHandler) super.getParameterHandlerByName(limit, parameter, parameterName);

		if (!handler.getName().equals(parameterName)) {//rebind to real description
			logger.debug("Rebinding param {} to {}", parameterName, parameter);
			handler = createNewParameterHandler(parameter, handler);
			parameterHandlers.put(parameterName, handler);
		}


		return (T) handler;
	}

	@Override
	public List<? extends IPortGroup> getPortGroupDescriptions() {
		if (portGroupsCached == null) {
			portGroupsCached = getDynamicModuleDescription().getStaticPortGroups(this);
			Set<String> ports = portGroupsCached.stream().flatMap(pg -> pg.getPorts().stream()).map(IPort::getName).collect(Collectors.toSet());
			portHandlers.keySet().stream().filter(k -> !ports.contains(k)).collect(Collectors.toList()).forEach(k -> portHandlers.remove(k).disconnectAll());
		}
		return portGroupsCached;
	}

	@Override
	public List<? extends IParameterGroup> getParameterGroupDescriptions() {
		if (parameterGroupsCached == null) {
			parameterGroupsCached = getDynamicModuleDescription().getStaticParameterGroups(this);
			parameterNamesCached = parameterGroupsCached.stream().flatMap(pg -> pg.getParameters().stream()).map(IParameter::getName).collect(Collectors.toSet());
			parameterHandlers.keySet().stream().filter(k -> !parameterNamesCached.contains(k)).collect(Collectors.toList()).forEach(k -> parameterHandlers.remove(k));
			logger.warn("Building parameters for " + getName());
		}
		return parameterGroupsCached;
	}

	@Override
	protected void notifyParameterChanged(BaseParameterHandler handler) {
		if (handler.getParameterDescription().isDynamicallyReferenced()) {
			portGroupsCached = null;
			getPortGroupDescriptions();
			parameterGroupsCached = null;
			getParameterGroupDescriptions();
		}
		super.notifyParameterChanged(handler);
	}

	@Override
	public CalculatedValueResult<? extends Number> resolveDynamicParam(ParameterReference parameter, int index) {
		if (!isIndexInRange(parameter, index)) {
			return CalculatedValueResult.customError("Index "+index+" out of range for " + parameter.getName());
		}
		return getParameterHandlerByName(IIntegerParameterHandler.class, Parameter.getBackingParameter(parameter), parameter.getName() + index).getNumberValue();
	}

	@Override
	public CalculatedValueResult<? extends String> resolveDynamicParamString(ParameterReference parameter, int index) {
		if (!isIndexInRange(parameter, index)) {
			return CalculatedValueResult.customError("Index "+index+" out of range for " + parameter.getName());
		}
		return getParameterHandlerByName(IParameterHandler.class, Parameter.getBackingParameter(parameter), parameter.getName() + index).getValue();

	}

	private boolean isIndexInRange(ParameterReference parameter, int index) {
		//Refresh list
		getParameterGroupDescriptions();

		return parameterNamesCached.contains(parameter.getName() + index);
	}

	@Override
	public String getToplevel() {
		return VerilogLang.getVerilogName(getGlobalName());
	}


	private void beforeMarshal(Marshaller m) {
		//Force creation of all parameter handlers
		//If we don't do this, parameter values are not correctly saved
		DocumentUtil.streamAllParameters(this).forEach(x -> { });

	}
}
