/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3;

import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IFirmwareConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;

import java.io.File;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;

/**
 * Created by Ramon on 13.05.2016.
 */
public class FirmwareConfiguration implements IFirmwareConfiguration {

	Path externalLocation;
	@XmlAttribute(required = true)
	String firmwareName;
	@XmlElement
	private List<SubFirmwareMapping> subFirmwareMappings = new ArrayList<>();

	@XmlAttribute(name = "externalLocation")
	private File getExternalLocationFile() {
		return externalLocation != null? externalLocation.toFile() : null;
	}

	private void setExternalLocationFile(File file) {
		this.externalLocation = file != null? file.toPath() : null;
	}

	@Override
	public boolean hasExternalLocation() {
		return externalLocation != null;
	}

	/**
	 * Only set if not in root of projectDir
	 * TODO unify with absolutePath
	 * @return Path or null
	 */
	@Override
	public Path getExternalLocation() {
		return externalLocation;
	}

	@Override
	public String getFirmwareName() {
		return firmwareName;
	}

	@Override
	public Optional<String> getSubFirmwareName(ISubsystemConfiguration subsystem) {
		return findSubFirmwareMapping(subsystem).map(SubFirmwareMapping::getSubFirmName);
	}

	@Override
	public Path getSubFirmwareLocation(ISubsystemConfiguration subsystem, ILibJConfigProject project) {
		return findSubFirmwareMapping(subsystem).map(mapping -> mapping.getLocation(project)).orElseGet(() -> {
			Path check = getLocation(project).resolve(subsystem.getName());
			if (IFirmwareConfiguration.isSubFirmwareFolder(check)) {
				return check;
			} else {
				throw new IllegalStateException("Missing Sub Firmware for " + subsystem.getName());
			}
		});
	}

	@Override
	public boolean hasFirmware() {
		return Objects.nonNull(firmwareName) || Objects.nonNull(externalLocation);
	}

	@Override
	public boolean hasSubFirmwares() {
		return false;
	}

	@Override
	public Optional<String> getGlobalFirmwareName() {
		if (!hasSubFirmwares()) {
			return Optional.empty();
		}
		return Optional.of("global");
	}

	public Optional<SubFirmwareMapping> findSubFirmwareMapping(ISubsystemConfiguration subsystem) {
		return subFirmwareMappings.stream().filter(mapping -> mapping.getSubsystemName().equals(subsystem.getName())).findAny();
	}

	@Override
	public String toString() {
		return "FirmwareConfiguration{" +
				"externalLocation=" + externalLocation +
				", firmwareName='" + firmwareName + '\'' +
				", subFirmwareMappings=" + subFirmwareMappings +
				'}';
	}

	private static class SubFirmwareMapping {

		private FirmwareConfiguration firmware;

		@XmlAttribute
		private String subsystemName;

		@XmlAttribute
		private String subFirmName;

		public SubFirmwareMapping(FirmwareConfiguration parent) {
			this.firmware = parent;
		}

		public String getSubsystemName() {
			return subsystemName;
		}

		public String getSubFirmName() {
			return subFirmName;
		}

		public Path getLocation(ILibJConfigProject project) {
			return firmware.getLocation(project).resolve(subFirmName);
		}

		private void afterUnmarshal(Unmarshaller u, java.lang.Object parent) {
			this.firmware = (FirmwareConfiguration) parent;
		}
	}


	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}
		FirmwareConfiguration that = (FirmwareConfiguration) o;
		return Objects.equals(externalLocation, that.externalLocation) &&
				Objects.equals(firmwareName, that.firmwareName) &&
				Objects.equals(subFirmwareMappings, that.subFirmwareMappings);
	}

	@Override
	public int hashCode() {

		return Objects.hash(externalLocation, firmwareName, subFirmwareMappings);
	}
}
