/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.INamedValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IFirmwareConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleVisitor;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IModuleContainer;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IFirmwareAttributeHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.function.Predicate;
import java.util.stream.Stream;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;

/**
 * Created by ramon on 17.03.2016.
 */
@XmlRootElement(name = "memoryModule", namespace = "http://www.rs.tu-darmstadt.de/ns/spartanmc/configuration/jc3")
public final class MemoryModuleConfiguration<P extends IModuleContainer<?> & de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement<?, ?, ?> & INamedValueResolver> extends CommonModuleConfiguration<P> implements IMemoryModuleConfiguration<P> {

	private static final Logger logger = LoggerFactory.getLogger(MemoryModuleConfiguration.class);
	private final List<IMemoryChangeListener> memoryListeners = new ArrayList<>();
	@XmlElement
	private FirmwareConfiguration firmware;
	private FirmwareLocationHandler firmwareLocationHandler;

	/**
	 * JaxB only
	 */
	protected MemoryModuleConfiguration() {

	}

	public MemoryModuleConfiguration(String name, MemoryModuleDescription moduleDescription, P parent, Library library) {
		super(name, moduleDescription, parent, library);
		firmware = new FirmwareConfiguration();
	}

	@Override
	public Stream<ISubsystemConfiguration> getConnectedSubsystems() {
		MemoryModuleDescription description = getModuleDescription();
		return busHandlers.values().stream()
				.flatMap(IConnectorHandler::streamOppositeConnectors)
				.map(DocumentUtil.findOriginalBusHandlerForPassthroughMapper(ElementType.Subsystem))
				.map(IConnectorInstance::getSourceElement)
				.filter(m -> m.getType() == ElementType.Subsystem)
				.map(sub -> ((ISubsystemConfiguration) sub))
				.distinct();
	}

	@Override
	public List<MemoryModuleDescription.ExtMemory> getMemoriesExposedOn(Bus exposedOn) {
		MemoryModuleDescription description = getModuleDescription();
		List<MemoryModuleDescription.ExtMemory> memory = description.getMemoryExposedOn(exposedOn);
		if (memory.isEmpty()) {
			report(ConfigurationIssue.createModuleDoesNotExpose(this, null, exposedOn, getOppositeMasterBusDescription(exposedOn), "Address Space Analysis"));
		}
		return memory;
	}

	@Override
	public void addMemoryListener(IMemoryChangeListener listener) {
		memoryListeners.add(listener);
	}

	@Override
	public void removeMemoryListener(IMemoryChangeListener listener) {
		memoryListeners.remove(listener);
	}

	@Override
	public FirmwareConfiguration getFirmware() {
		return firmware;
	}

	public IFirmwareAttributeHandler getFirmwareLocationHandler() {
		if (firmwareLocationHandler == null) {
			firmwareLocationHandler = new FirmwareLocationHandler(getDocument().getProject());
		}

		return firmwareLocationHandler;
	}

	@Override
	public long getResolvedMemorySize(MemoryModuleDescription.ExtMemory memory) {
		return memory.resolveSize(this).orElseThrow();
	}

	protected void notifyFirmwareChanged() {
		if (pluginEngine != null) {
			pluginEngine.onFirmwareChanged(this);
		}
		for (IMemoryChangeListener listener : memoryListeners) {
			listener.onFirmwareChanged(this);
		}
	}

	@Override
	public String toString() {
		return "MemoryModuleConf" + toModuleConfString();
	}

	@Override
	public ElementType getType() {
		return ElementType.MemoryModule;
	}

	@Override
	public MemoryModuleDescription getModuleDescription() {
		return library.getMemoryCommonDescription(getModuleDescriptionName());
	}

	@Override
	public <T> T apply(IModuleVisitor<T> visitor) {
		return visitor.applyToMemoryModule(this);
	}

	@Override
	public boolean isValid(boolean report) {
		boolean res = super.isValid(report);

		if (!getModuleDescription().containsFirmware()) {
			return res;
		}

		boolean firmwareValid = getFirmwareLocationHandler().isValid(report);
		return firmwareValid && res;
	}

	/**
	 * Created by Ramon on 07.02.2016.
	 */
	private class FirmwareLocationHandler implements IFirmwareAttributeHandler {

		private final LibJConfigProject project;

		private FirmwareLocationHandler(LibJConfigProject project) {
			this.project = project;
			if (MemoryModuleConfiguration.this.firmware == null) {
				firmware = new FirmwareConfiguration();
			}
		}

		@Override
		public List<String> getSuggestedFirmwares() {
			return project.getKnownFirmwares(firmware.getExternalLocation());
		}

		@Override
		public List<String> getSuggestedSubNames() {
			return project.getKnownFirmwareSubNames(firmware.getExternalLocation(), firmware.firmwareName);
		}

		@Override
		public String getFirmwareName() {
			return firmware.firmwareName;
		}

		@Override
		public void setFirmwareName(String name) {
			if (name != null && !name.equals(firmware.firmwareName)) {
				firmware.firmwareName = name;
				notifyFirmwareChanged();
			}
		}

		@Override
		public Path getAbsoluteLocation() {
			return firmware.getExternalLocation();
		}

		@Override
		public Path getSelectedFirmwareLocation() {
			Path absoluteLocation = firmware.getExternalLocation();
			if (absoluteLocation == null) {
				absoluteLocation = project.getWorkingDir();
			}
			return absoluteLocation.resolve(firmware.getFirmwareName());
		}

		@Override
		public boolean setAbsoluteLocation(Path location) {
			if ((location != null && !location.equals(firmware.externalLocation) && Files.exists(location) && Files.isDirectory(location)) || (location == null && firmware.externalLocation != null)) {
				firmware.externalLocation = location;
				notifyFirmwareChanged();// we have not accepted the new value, but notify of change anyway so the caller can react
				return true;
			}
			return false;
		}

		@Override
		public boolean isValid(boolean report) {
			IFirmwareConfiguration.FirmwareState validFirmware = firmware.isValidFirmware(project);

			if (validFirmware != IFirmwareConfiguration.FirmwareState.Valid) {
				final Predicate<MemoryModuleConfiguration<P>> actuality = m -> m.firmware.isValidFirmware(project)!=validFirmware;
				ConfigurationIssue.IssueType level;
				if (validFirmware== IFirmwareConfiguration.FirmwareState.Invalid)
					level = ConfigurationIssue.IssueType.Error;
				else
					level = ConfigurationIssue.IssueType.Warning;
				getDocument().getIssueSink().report(ConfigurationIssue.createMissingProperty(MemoryModuleConfiguration.this, actuality, "Firmware", "Checking Firmware Validity", level));
			}

			return validFirmware != IFirmwareConfiguration.FirmwareState.Invalid;
		}

		@Override
		public boolean isValidValue(String value) {
			throw new UnsupportedOperationException();
		}

		@Override
		public CalculatedValueResult<String> getValue() {
			StringBuilder sb = new StringBuilder();
			if (firmware.hasExternalLocation()) {
				sb.append(firmware.getExternalLocation()).append(':');
			}
			sb.append(firmware.firmwareName);
			//			if (firmware.hasSubFirmwares()) {
			//				sb.append('#').append(firmware.subName);
			//			}
			return CalculatedValueResult.of(sb.toString());
		}

		@Override
		public void setStoredValue(String value) {
			throw new UnsupportedOperationException();
		}

		@Override
		public CalculatedValueResult<Boolean> isEditable() {
			return CalculatedValueResult.of(true);
		}

		@Override
		public String toString() {
			return "FirmwareLocationHandler{" +
				"path=" + firmware.getFirmwareName() +
				//					"subPath="+ firmware.getSubName()+
				"location=" + firmware.getExternalLocation() +
				"}";
		}
	}
}
