/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.INamedValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.BusBreakoutModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.DynamicStructureModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IModuleContainer;

import java.util.List;
import java.util.stream.Stream;

/**
 * Created by Ramon on 20.06.2016.
 */
public class ModuleConfigurationFactory {

	public static <P extends IModuleContainer<?> & IConfigurationElement<?, ?, ?> & INamedValueResolver> ModuleConfiguration<P> createConfigurationFor(String name, IModuleDescription moduleDescription, ConfigurationDocument document, P parent, Library library) {
		switch (moduleDescription.getType()) {
			case DynamicStructure:
				return new DynamicModuleConfiguration<>(name, (DynamicStructureModuleDescription) moduleDescription, parent, library);
			case Memory:
				return new MemoryModuleConfiguration<>(name, (MemoryModuleDescription) moduleDescription, parent, library);
			case BusBreakout:
				return new BusBreakoutConfiguration<>(name, (BusBreakoutModuleDescription) moduleDescription, parent, library);
			case Peripheral:
				return new PeripheralModuleConfiguration<>(name, (PeripheralModuleDescription) moduleDescription, parent, library);
			default:
				return new CommonModuleConfiguration<>(name, moduleDescription, parent, library);
		}
	}

	public static String findUniqueName(String overrideName, IModuleDescription moduleDescription, ConfigurationDocument doc, List<? extends ModuleConfiguration<?>> existing) {
		return findUniqueName(overrideName, moduleDescription, doc, existing, null);
	}


	public static String findUniqueName(String overrideName, IModuleDescription moduleDescription, ConfigurationDocument doc, SubsystemConfiguration subsystem) {
		return findUniqueName(overrideName, moduleDescription, doc, subsystem.getModules(), subsystem);
	}

	public static String findUniqueName(String overrideName, IModuleDescription moduleDescription, ConfigurationDocument doc, List<? extends ModuleConfiguration<?>> existing, SubsystemConfiguration subsystem) {
		String baseName;
		if (overrideName != null) {
			if (isUniqueName(overrideName, subsystem, doc)) {
				return overrideName;
			} else {
				baseName = overrideName;
			}
		} else {
			baseName = moduleDescription.getName();
		}

		StringBuilder sb = new StringBuilder();
		sb.append(baseName).append('_');

		int index = (int) existing.stream().filter(m -> m.getModuleDescription() == moduleDescription).count();

		int length = sb.length();

		String name;
		do {
			sb.setLength(length);
			sb.append(index++);
			name = sb.toString();
		} while (!isUniqueName(name, subsystem, doc));

		return name;
	}

	private static boolean isUniqueName(String name, SubsystemConfiguration subsystem, ConfigurationDocument document) {
		if (subsystem != null) {
			return isSubsystemUniqueName(name, subsystem);
		} else {
			return !Stream.concat(document.getSubsystems().stream(), document.getModules().stream()).anyMatch(e -> e.getName().equals(name));
		}
	}

	public static boolean isSubsystemUniqueName(String name, SubsystemConfiguration subsystem) {
		return !subsystem.streamAllModules().anyMatch(e -> e.getName().equals(name));
	}

	public static boolean isGlobalUniqueName(String name, ConfigurationDocument doc) {
		return !doc.streamAllModules().anyMatch(e -> e.getName().equals(name));
	}

	public static boolean isUniqueName(String name, IConfigurationElement instance, List<? extends IConfigurationElement> existing) {
		return !existing.stream().filter(e -> e != instance).anyMatch(e -> e.getName().equals(name));
	}
}
