/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.IClkPortMapping;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IRedirectedResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ClkReference;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.Parameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.ClkPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPortGroup;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.IIssueSink;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.Marker;
import org.slf4j.MarkerFactory;

import java.util.LinkedList;
import java.util.List;
import java.util.Optional;

/**
 * Contains all the logic required to track clks through buses, connections and parameters
 * TODO: resolve against this, to spare creation of new instances where possible. Instead fork!
 */
public class RedirectedClkResolver implements IRedirectedResolver {

	private static final Logger logger = LoggerFactory.getLogger(RedirectedClkResolver.class);
	private static final Marker resolveMarker = MarkerFactory.getMarker("Resolve");

	private final IConfigurationElement<IPort, IPortGroup, IPortHandler<IPort>> owner;
	private final IIssueSink issueSink;
	private List<Object> history = new LinkedList<>();

	public RedirectedClkResolver(IConfigurationElement<IPort, IPortGroup, IPortHandler<IPort>> start, IIssueSink issueSink) {
		this.owner = start;
		this.issueSink = issueSink;
	}

	public CalculatedValueResult<? extends Number> resolveClk(ClkReference clkRef) {
		logHistory(clkRef);
		logger.trace(RedirectedClkResolver.resolveMarker, "resolving {}@{}", clkRef.getName(), owner.getName());

		if (clkRef.isRedirected()) { // clk routed through a bus
			return clkRef.resolveRedirected(this);
		} else {
			logger.trace(RedirectedClkResolver.resolveMarker, "freq= {}", clkRef.getFrequencyRef());
			return clkRef.resolveFrequency(owner); // module local clkPort can resolve itself against us
		}
	}

	@Override
	public CalculatedValueResult<? extends Number> resolveAgainstBus(Bus bus, IClkPortMapping clkPortMapping) {
		logHistory(clkPortMapping);
		IBusHandler busHandler = owner.getBusHandler(bus);
		if (!busHandler.isConnected()) {
			issueSink.report(ConfigurationIssue.createUnconnected(busHandler, "resolving clkPort on Bus"));
			return CalculatedValueResult.unconnected();
		}

		try {
			return busHandler.streamOppositeConnectors().map(otherHandler -> resolveClkAgainstOppositeBusHandler(otherHandler, bus, clkPortMapping)).collect(StreamUtil.atLeastOneOrEqual());
		} catch (IllegalStateException e) { // multiple conflicting clocks
			BusDescription busDefinition = busHandler.getBusDefinition();
			issueSink.report(ConfigurationIssue.createMismatchedBusClks(busHandler, busDefinition.getOppositeBusPortDescription(bus, clkPortMapping), "resolving clk through bus"));
			return CalculatedValueResult.unknown();
		}
	}

	private void logHistory(Object entry) {
		history.add(entry);
	}

	private CalculatedValueResult<? extends Number> resolveClkAgainstOppositeBusHandler(IBusHandler oppositeHandler, Bus myBus, IClkPortMapping myMapping) {
		logHistory(myMapping);
		BusPort oppositePort = oppositeHandler.getBusDefinition().getOppositeBusPortDescription(myBus, myMapping);
		Optional<Bus.SetClk> setClk = oppositeHandler.getConnectorDescription().findSetClk(oppositePort);

		if (setClk.isPresent()) {
			Bus.SetClk clk = setClk.get();
			logger.trace(resolveMarker, "replaced with busClk: {}@{}", clk.getBusName(), oppositeHandler.getSourceElement().getName());
			return clk.resolveFrequency(oppositeHandler);
		} else {
			issueSink.report(ConfigurationIssue.createConnectedButUnconfigured(oppositeHandler, oppositePort, "resolving clkPort on Bus"));
			return CalculatedValueResult.unknown();
		}
	}

	@Override
	public CalculatedValueResult<? extends Number> resolveAgainstConnectedClkPort(ClkPort clkPort) {
		logHistory(clkPort);
		IPortHandler<?> portHandler = owner.getPortHandler(clkPort);
		if (!portHandler.isConnected()) {
			issueSink.report(ConfigurationIssue.createUnconnected(portHandler, "resolving clkPort on module barrier"));
			return CalculatedValueResult.unconnected();
		}

		IConnectorHandler port = portHandler.getSingleConnectedPort(); //opposite Port

		if (port.getSourceElement() == owner) {

			issueSink.report(ConfigurationIssue.createRecursiveConnection(portHandler, portHandler.streamConnections().findAny().get(), "resolving clkPort"));
			return CalculatedValueResult.unknown();
		}

		IConnectorDescription description = port.getConnectorDescription();
		if (port.getType() == IConnectorDescription.Type.Pin) { // pins have constant freq!

			IPinHandler pinHandler = (IPinHandler) port;

			logger.trace(resolveMarker, "replaced with #{}", description.getName());
			return pinHandler.getFrequency();
		} else {
			IConfigurationElement<?, ?, ?> sourceElement = port.getSourceElement();
			logger.trace(resolveMarker, "replaced with {}@{}", description.getName(), sourceElement.getName());
			if (description instanceof ClkPort) {
				return ((ClkPort) description).resolveFrequency(sourceElement); // resolve against otherPort's module
			} else {
				return CalculatedValueResult.customError(description.getName() + " has no frequency");
			}
		}
	}

	@Override
	public CalculatedValueResult<? extends Number> resolveAgainstBusHandler(Bus bus, Parameter parameter) {
		logHistory(parameter);
		return owner.getBusHandler(bus).resolveParam(parameter);
	}
}
