/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.INamed;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.AddressLayoutSpec;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.SubsystemModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleVisitor;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers.BusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting.AddressLayoutMapping;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlElementWrapper;

import static de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription.ModuleType.Peripheral;

/**
 * Created by ramon on 04.12.2015.
 */
public class SubsystemConfiguration extends ModuleConfiguration<ConfigurationDocument> implements ISubsystemConfiguration<ConfigurationDocument> {

	private static final Logger logger = LoggerFactory.getLogger(SubsystemConfiguration.class);
	private List<ModuleConfiguration<SubsystemConfiguration>> childModules = new ArrayList<>();
	private Collection<IMemoryModuleConfiguration<?>> exportingMemory;
	@XmlElement
	private AddressLayoutMapping layoutMapping;
	/**
	 * Only for JAXB
	 */
	private SubsystemConfiguration() {}

	SubsystemConfiguration(String name, String moduleDescriptionName, ConfigurationDocument parent, Library library) {
		super(name, moduleDescriptionName, parent, library);
	}

	@Override
	public boolean isUniqueName(String name) {
		return ModuleConfigurationFactory.isUniqueName(name, this, parent.getSubsystems());
	}

	@Override
	public ConfigurationDocument getParent() {
		return parent;
	}

	@Override
	public void prepareRemove() {
		super.prepareRemove();

		childModules.forEach(ModuleConfiguration::prepareRemove);

		childModules.forEach(child -> getDocument().notifyModuleRemoved(this, child, true, false));
	}

	@Override
	public ConfigurationDocument getDocument() {
		return parent;
	}

	@Override
	public String getAbsolutePath() {
		return '/' + getName();
	}

	@Override
	public ElementType getType() {
		return ElementType.Subsystem;
	}

	@Override
	public SubsystemModuleDescription getModuleDescription() {
		return library.getSubsystemDescription(getModuleDescriptionName());
	}

	@Override
	public <T> T apply(IModuleVisitor<T> visitor) {
		return visitor.applyToSubsystemModule(this);
	}

	protected Stream<? extends IMemoryModuleConfiguration<?>> streamConnectedMemoryModules() {
		return getModuleDescription()
			.getAddressLayoutSpec()
			.streamImportingAddressRanges()
			.filter(ar -> ar.getType() == AddressLayoutSpec.BlockType.SelfInitializing)
			.map(ar -> getBusHandler(ar.getImplementedBy()))
			.flatMap(IBusHandler::streamOppositeConnectors)
			.map(bh -> DocumentUtil.findOriginalBusHandlerForPassthrough(bh, IConfigurationElement.ElementType.MemoryModule))
			.map(IConnectorInstance::getSourceElement)
			.map(mod -> mod instanceof IMemoryModuleConfiguration? ((IMemoryModuleConfiguration<?>) mod) : null)
			.filter(Objects::nonNull)
			.distinct()
			.sorted(Comparator.comparing(INamed::getName));
	}

	@Override
	@XmlElementRef
	@XmlElementWrapper(name = "childModules")
	public List<ModuleConfiguration<SubsystemConfiguration>> getModules() {
		return childModules;
	}

	@SuppressWarnings("unused")
	private void setChildModules(List<ModuleConfiguration<SubsystemConfiguration>> modules) {
		this.childModules = modules;
	}

	@Override
	public ModuleConfiguration<SubsystemConfiguration> addModule(IModuleDescription moduleDescription) {
		return addModule(moduleDescription, null);
	}

	@Override
	public ModuleConfiguration<SubsystemConfiguration> addModule(IModuleDescription moduleDescription, @Nullable String name) {
		ConfigurationDocument.verifyLibrary(library);

		if (!getModuleDescription().isSupportsPeripherals() && moduleDescription.getType() == Peripheral) {
			throw new IllegalArgumentException("This Subsystem does not support peripherals");
		}

		String uniqueName = ModuleConfigurationFactory.findUniqueName(name, moduleDescription, getDocument(), this);
		ModuleConfiguration<SubsystemConfiguration> conf = ModuleConfigurationFactory.createConfigurationFor(uniqueName, moduleDescription, getDocument(), this, library);
		childModules.add(conf);
		parent.notifyModuleAdded(conf);

		autoconnectCompatibleBuses(conf);

		return conf;
	}

	@Override
	public boolean removeModule(IModuleConfiguration module) {
		return removeModuleInternal(module, false, false);
	}

	public boolean removeModuleInternal(IModuleConfiguration module, boolean internal, boolean invalid) {
		return ConfigurationDocument.removeModule((ModuleConfiguration) module, childModules, Collections.emptyList(), this, parent, internal, invalid);
	}

	@Override
	public String toString() {
		return "SubsystemConf" + toModuleConfString();
	}

	@Override
	public PeripheralModuleConfiguration addPeripheral(PeripheralModuleDescription moduleDescription) {
		return (PeripheralModuleConfiguration) addModule(moduleDescription);
	}

	@Override
	public PeripheralModuleConfiguration addPeripheral(PeripheralModuleDescription moduleDescription, @Nullable String name) {
		return (PeripheralModuleConfiguration) addModule(moduleDescription, name);
	}

	@Override
	public ModuleConfiguration<SubsystemConfiguration> addCommonModule(IModuleDescription moduleDescription) {
		return addModule(moduleDescription);
	}

	@Override
	public MemoryModuleConfiguration<SubsystemConfiguration> addMemoryModule(MemoryModuleDescription moduleDescription) {
		return (MemoryModuleConfiguration<SubsystemConfiguration>) addCommonModule(moduleDescription);
	}

	@Override
	public Stream<PeripheralModuleConfiguration<SubsystemConfiguration>> streamPeripherals() {
		return childModules.stream().filter(p -> p.getType() == ElementType.Peripheral).map(p -> ((PeripheralModuleConfiguration) p));
	}

	@Override
	public Stream<MemoryModuleConfiguration<SubsystemConfiguration>> streamMemoryModules() {
		return childModules.stream().filter(p -> p.getType() == ElementType.MemoryModule).map(p -> ((MemoryModuleConfiguration) p));
	}

	@Override
	public Stream<ModuleConfiguration<SubsystemConfiguration>> streamCommonModules() {
		return childModules.stream().filter(p -> p.getType() != ElementType.Peripheral);
	}

	@Override
	public Stream<ModuleConfiguration<SubsystemConfiguration>> streamAllModules() {
		return childModules.stream();
	}

	@Override
	public Stream<ModuleConfiguration<SubsystemConfiguration>> streamModulesRecursive() {
		return streamAllModules();
	}

	@Override
	@Deprecated
	public Stream<IMemoryModuleConfiguration<?>> streamExportingMemoryConfiguration() {
		List<IMemoryModuleConfiguration<?>> available = streamConnectedMemoryModules().collect(Collectors.toList());

		if (exportingMemory != null && available.containsAll(exportingMemory)) {
			return exportingMemory.stream();
		} else if (available.size() > 0) {
			if (exportingMemory != null) {
				logger.warn("Exporting Memory property was overwritten with {}, because existing value {} is not longer present", available, exportingMemory);
			} //TODO needs configuration or some other method to prevent redunant and random picks for exportingFirmware
			exportingMemory = available;
			return exportingMemory.stream();
		} else {
			return Stream.empty();
		}
	}

	@Override
	public boolean hasStaticAddressSpaceMappings() {
		return layoutMapping != null;
	}

	@Override
	public AddressLayoutMapping getStaticAddressLayoutMapping() {
		return layoutMapping;
	}

	@Override
	public void setStaticAddressLayoutMapping(AddressLayoutMapping layoutMapping) {
		this.layoutMapping = layoutMapping;
	}

	private void autoconnectCompatibleBuses(ModuleConfiguration<?> conf) {
		this.getBusDescriptions().forEach(procBus -> {
			conf.getBusDescriptions().forEach(periBus -> {
				BusHandler procHandler = getBusHandler(procBus);
				BusHandler periHandler = conf.getBusHandler(periBus);
				if (IBusHandler.isCompatible(procHandler, periHandler) && procHandler.isAdditionalConnectionAllowed()) {
					getDocument().createBusConnection(procHandler, periHandler).connect();
				}
			});

		});
	}
}
