/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.OrderedConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers.BaseConnectorHandler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlIDREF;

import static de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections.BaseConnection.InitResult.ProblemA;
import static de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections.BaseConnection.InitResult.ProblemB;
import static de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections.BaseConnection.InitResult.Success;

/**
 * Created by ramon on 17.02.2016.
 */
public abstract class BaseConnection<I extends BaseConnectorHandler<I, A, ?, ?>, A extends IConnectorHandler<?, ?>> extends OrderedConnection<A> implements IConnection<A> {

	private static final Logger logger = LoggerFactory.getLogger(BaseConnection.class);

	protected ConfigurationDocument document;
	protected I portB;
	protected I portA;

	protected boolean connected;

	public BaseConnection(I portA, I portB, ConfigurationDocument document) {
		this.document = document;
		this.portA = portA;
		this.portB = portB;

		if (portA == null || portB == null) {
			throw new IllegalArgumentException("Incomplete Connection");
		}

		if (portA == portB) {
			throw new IllegalArgumentException("Not to self");
		}
	}

	protected BaseConnection() {

	}

	protected static <P extends IConnectorHandler> void verifyAttached(ConfigurationDocument document, P portA, P portB, IConnection<P> connectionA, IConnection<P> connectionB) {
		if (document == null) {
			if (!portA.isPresent(connectionA) && !portB.isPresent(connectionB)) {
				throw new IllegalStateException("Not attached to document, and not connecting stuff. Should not be reachable!!");
			} else {
				throw new IllegalStateException("Not attached to document, bus still connecting stuff!");
			}
		}
	}

	protected void afterUnmarshal(Unmarshaller u, Object parent) {
		this.document = (ConfigurationDocument) parent;
	}

	@Override
	public A getPortA() {
		return (A) portA;
	}

	@Override
	public A getPortB() {
		return (A) portB;
	}

	@Override
	public boolean isConnected() {
		return connected;
	}

	@Override
	public void connect() {
		//TODO lock
		portA.verifyAdditionalConnection();
		portB.verifyAdditionalConnection();
		portA.verifyConnectibleTo(portB);// Bidirectional

		portA.addConnection(this);
		portB.addConnection(this);
		connected = document.addConnection(this);
	}

	@Override
	public void disconnect() {
		portA.removeConnection(this);
		portB.removeConnection(this);
		document.removeConnection(this);
		document = null;
		connected = false;
	}

	@Override
	public boolean isValid(boolean report) {
		verifyAttached(document, getPortA(), getPortB(), this, this);

		boolean valid = getPortA().isConnectibleTo(getPortB());
		if (!valid && report) {
			document.report(ConfigurationIssue.createMismatchedConnection(getPortA(), this, "Port characteristics changed"));
		}
		return valid;
	}

	@XmlAttribute(name = "portA")
	@XmlIDREF
	public I getPortAInt() {
		return portA;
	}

	protected void setPortAInt(I portA) {
		this.portA = portA;
	}

	@XmlAttribute(name = "portB")
	@XmlIDREF
	public I getPortBInt() {
		return portB;
	}

	protected void setPortBInt(I portB) {
		this.portB = portB;
	}

	public void migrate(I from, I to) throws IllegalArgumentException {
		if (portA == from) {
			portA = to;
		} else if (portB == from) {
			portB = to;
		} else {
			throw new IllegalArgumentException("From Connector does not exist on this connection");
		}

		if (!isValidAndReport()) {
			throw new IllegalArgumentException("Replacement Connector is not valid for this connection");
		}
	}

	public I getOppositeConnectorInt(Object portHandler) {
		return (portHandler == portA)? portB : portA;
	}

	protected void invalidateConnector(BaseConnectorHandler portA, BaseConnectorHandler portB) {
		if (portA != null) {
			portA.invalidate();
		}
		if (portB != null) {
			portB.invalidateConnection(this); // do not us disconnect, because it will try to do a concurrent remove, and notify wrong
		}
	}

	public enum InitResult {
		Success,
		ProblemA,
		ProblemB,
		Error
	}

	public InitResult ensureInitialized() {
		if (!connected) {
			attach();
		}

		BaseConnectorHandler portA = (BaseConnectorHandler) getPortA();
		boolean successA = tryInitializeConnector(portA);
		BaseConnectorHandler portB = (BaseConnectorHandler) getPortB();
		boolean successB = tryInitializeConnector(portB);

		if (!successA) {
			invalidateConnector(portA, portB);
		}
		if (!successB) {
			invalidateConnector(portB, portA);
		}

		if (successA && successB) {
			return Success;
		} else if (successA) {
			return ProblemB;
		} else if (successB) {
			return ProblemA;
		} else {
			return InitResult.Error;
		}

	}

	protected void attach() {
		portA.attachConnection(this);
		portB.attachConnection(this);
		connected = true;
	}

	protected boolean tryInitializeConnector(BaseConnectorHandler connector) {
		try {
			connector.ensureInitialized();
			return true;
		} catch (Exception e) {
			logger.error("Connector "+connector+" could not be initialized. Removing...", e);
			return false;
		}
	}

	@Override
	public int hashCode() {
		int result = System.identityHashCode(portB);
		result += System.identityHashCode(portA);
		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		BaseConnection that = (BaseConnection) o;

		return portA == that.portA && portB == that.portB || portB == that.portA && portA == that.portB;
	}

	@Override
	public String toString() {
		String a = (getPortA() != null)? getPortA().getAbsoluteConnectorName() : "null";
		String b = (getPortB() != null)? getPortB().getAbsoluteConnectorName() : "null";
		return getType() + "Connection{" +
			"portA=" + a +
			", portB=" + b +
			'}';
	}
}
