/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.exceptions.ConnectionMismatchException;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers.BasePortHandler;

import java.util.Comparator;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlRootElement;


/**
 * Created by Ramon on 28.04.2016.
 */
@XmlRootElement(name = "partial-connection", namespace = "http://www.rs.tu-darmstadt.de/ns/spartanmc/configuration/jc3")
public class PartialConnection extends BaseConnection<BasePortHandler<?, ?>, IPortHandler<?>> implements IConnection<IPortHandler<?>> {

	@XmlAttribute
	private int startA;
	@XmlAttribute
	private int startB;
	@XmlAttribute
	private int width;
	private Mode mode;


	public PartialConnection(BasePortHandler a, int startA, BasePortHandler b, int startB, int width, ConfigurationDocument document) {
		super(a, b, document);
		this.startA = startA;
		this.startB = startB;
		this.width = width;

		analyzeMode();
	}

	private PartialConnection() {
	}

	private void analyzeMode() {
		boolean aFull = startA < 0, bFull = startB < 0;
		boolean autoWidth = width < 0;

		if (aFull && bFull) {
			throw new IllegalArgumentException("Full connection not allowed with partial connection");
		}
		if ((aFull || bFull) ^ autoWidth) {
			throw new IllegalArgumentException("width and starts must match");
		}
		if (aFull && !portB.getPartialConnectorHandler().isPresent() || bFull && !portA.getPartialConnectorHandler().isPresent()) {
			throw new IllegalArgumentException("At least 1 port must support partial connections");
		}

		if (aFull) {
			mode = Mode.AWidth;
		} else if (bFull) {
			mode = Mode.BWidth;
		} else {
			mode = Mode.Explicit;
		}
	}

	@Override
	public ConnectionType getType() {
		return IConnection.ConnectionType.Partial;
	}

	@Override
	public boolean isPartialA() {
		return startA >= 0;
	}

	@Override
	public boolean isPartialB() {
		return startB >= 0;
	}

	@Override
	public int getStartA() {
		return startA < 0? 0 : startA;
	}

	@Override
	public int getEndA() {
		return getWidth() + getStartA() - 1;
	}

	@Override
	public int getStartB() {
		return startB < 0? 0 : startB;
	}

	@Override
	public int getEndB() {
		return getWidth() + getStartB() - 1;
	}

	private void verifyWidthAvailable(BasePortHandler port, int start, int width) {
		if (!isWidthAvailable(port, start, width)) {
			throw new IllegalStateException("Width not available");
		}
	}

	private boolean isWidthAvailable(BasePortHandler port, int start, int width) {
		return port.getConnectorWidth() >= start + width;
	}

	@Override
	public void connect() { //TODO lock
		verify();

		connectTo(portA, isPortAPartial());
		connectTo(portB, isPortBPartial());

		document.addConnection(this);
	}

	private int getAndCheckConnectorWidth() {
		int width;
		switch (mode) {
			case AWidth:
				width = portA.getConnectorWidth();
				verifyWidthAvailable(portB, startB, width);
				return width;
			case BWidth:
				width = portB.getConnectorWidth();
				verifyWidthAvailable(portA, startA, width);
				return width;
			case Explicit:
				width = this.width;
				verifyWidthAvailable(portB, startB, width);
				verifyWidthAvailable(portA, startA, width);
				return width;
			default:
				throw new IllegalStateException();
		}
	}

	private boolean isPortAPartial() {
		switch (mode) {
			case AWidth:
				return false;
			case BWidth:
				return true;
			case Explicit:
				return true;
			default:
				throw new IllegalStateException();
		}
	}

	private boolean isPortBPartial() {
		switch (mode) {
			case AWidth:
				return true;
			case BWidth:
				return false;
			case Explicit:
				return true;
			default:
				throw new IllegalStateException();
		}
	}


	@Override
	public boolean isValid(boolean report) {
		verifyAttached(document, getPortA(), getPortB(), this, this);

		boolean valid;
		try {
			getAndCheckConnectorWidth();
			valid = true;
		} catch (IllegalStateException e){
			valid = false;
		}

		if (!valid && report) {
			document.report(ConfigurationIssue.createMismatchedConnection(getPortA(), this, "Port characteristics changed"));
		}
		return valid;
	}

	@Override
	protected void attach() {
		analyzeMode();

		if (startA < 0) {
			portA.attachConnection(this);
		} else {
			portA.getPartialConnectorHandler().get().attachPartialConnection(this);
		}

		if (startB < 0 || portB.getType() == IConnectorDescription.Type.Pin) {
			portB.attachConnection(this);
		} else {
			portB.getPartialConnectorHandler().get().attachPartialConnection(this);
		}

		connected = true;
	}

	@Override
	public int hashCode() {
		int result = super.hashCode();
		result = 31 * result + startA + startB;
		result = 31 * result + width;
		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}
		//		if (!super.equals(o)) return false;

		PartialConnection that = (PartialConnection) o;

		return ((portA == that.portA && portB == that.portB && startA == that.startA && startB == that.startB)
			|| (portA == that.portB && portB == that.portA && startA == that.startB && startB == that.startA))
			&& width == that.width;
	}

	@Override
	public String toString() {
		return super.toString() +
			"{startA=" + startA +
			", startB=" + startB +
			", width=" + width +
			'}';
	}

	private void verifyTypeAndDirectionCompatible(IPortHandler portA, IPortHandler portB) {
		if (!IConnectorHandler.isDirectionCompatible(portA, portB)) {
			throw new ConnectionMismatchException("Directions incompatible");
		}
		if (!IConnectorDescription.isTypeCompatible(portA.getConnectorDescription(), portB.getConnectorDescription())) {
			throw new ConnectionMismatchException("Types incompatible");
		}
	}

	private void verify(BasePortHandler<?, ?> port, boolean isPartial, int start, int width) {
		if (isPartial && port.getType() == IConnectorDescription.Type.Port) {
			port.getPartialConnectorHandler().get().verifyAdditionalPartialConnection(start, width);
		} else {
			port.verifyAdditionalConnection();
		}
	}

	private void verify() {
		verifyTypeAndDirectionCompatible(portA, portB);

		int width = getAndCheckConnectorWidth();
		verify(portA, isPortAPartial(), startA, width);
		verify(portB, isPortBPartial(), startB, width);
	}

	private void connectTo(BasePortHandler<?, ?> port, boolean isPartial) {
		if (isPartial && port.getType() == IConnectorDescription.Type.Port) {
			port.getPartialConnectorHandler().get().addPartialConnection(this);
		} else {
			port.addConnection(this);
		}
	}

	public int getWidth() {
		return mode == Mode.Explicit? width : Math.min(portA.getConnectorWidth(), portB.getConnectorWidth());
	}

	private enum Mode {
		Explicit, AWidth, BWidth
	}


	public static class PartialComparator implements Comparator<PartialConnection> {

		private final IConnectorInstance self;
		private final Comparator<IConnectorHandler> thenComp = Comparator.comparing(IConnectorHandler::getPortName);

		public PartialComparator(IConnectorInstance self) {
			this.self = self;
		}

		@Override
		public int compare(PartialConnection o1, PartialConnection o2) {
			int res = o1.getMyStart(self) - o2.getMyStart(self);
			if (res != 0) {
				return res;
			} else {
				return thenComp.compare(o1.getOppositeConnector(self), o2.getOppositeConnector(self));
			}
		}
	}
}
