/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.references.IReference;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.exceptions.ConnectionMismatchException;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections.BaseConnection;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Stream;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlID;

/**
 * @author Ramon Wirsch
 */
public abstract class BaseConnectorHandler<I extends BaseConnectorHandler<I, B, ?, ?>, B extends IConnectorHandler<?, ?>, E extends BaseConnectorHandler.IConnectorHandlerContainer & IConfigurationElement & IValueResolver, P extends IConnectorDescription> implements IConnectorHandler<B, P> {

	protected P portDescription;
	protected E sourceElement;
	protected Set<BaseConnection<I, B>> connections = new HashSet<>();
	private String tempName;

	protected BaseConnectorHandler(E sourceElement, P portDescription) {
		this.sourceElement = sourceElement;
		this.portDescription = portDescription;
	}

	protected BaseConnectorHandler() {
	}

	/**
	 * Sort by opposite sourceElement name , then opposite port name
	 *
	 * @param me this
	 */
	private static <B extends IConnectorHandler<?, ?>> Comparator<IConnection<B>> getConnectionComparator(final BaseConnectorHandler me) {
		Comparator<IConnectorHandler> parentComparator = Comparator.comparing(port -> port.getSourceElement().getGlobalName()); // compare by comparing portHandler.getSourceElement().getName()
		Comparator<IConnectorHandler> portHandlerComparator = parentComparator.thenComparing(IConnectorHandler::getPrefixedConnectorName); // 2nd level compare by portHandler.getPrefixedConnectorName()

		return Comparator.comparing(connection -> connection.getOppositeConnector(me), portHandlerComparator); // compare Connections by extracting oppositePort and then using portHandlerComparator
	}

	public boolean isInitialized() {
		return portDescription != null && sourceElement != null && getDocument() != null;
	}

	public void ensureInitialized() {
		if (!isInitialized()) {
			requestInitialize();
		}
	}

	protected abstract void requestInitialize();

	public void invalidate() {
		connections.clear();
		invalidateOnSource();
	}

	protected abstract void invalidateOnSource();

	protected void afterUnmarshal(Unmarshaller u, Object parent) {
		sourceElement = (E) parent;
	}

	protected abstract ConfigurationDocument getDocument();

	public void initialize(P portDescription) {
		this.portDescription = portDescription;
		this.tempName = null;
	}

	@Override
	public E getSourceElement() {
		return sourceElement;
	}

	@Override
	public String getPortName() {
		return (portDescription != null)? portDescription.getName() : extractPortName(tempName);
	}

	@Override
	public boolean isConnected() {
		return !connections.isEmpty();
	}

	@Override
	public int getConnectionCount() {
		return connections.size();
	}

	protected abstract String extractPortName(String tempName);

	private String getAbsolutePath() {
		return sourceElement.getAbsolutePath() + getPrefixedConnectorName();
	}

	@Override
	public abstract String getPrefixedConnectorName();

	@XmlAttribute(name = "name")
	@XmlID
	@Override
	public String getAbsoluteConnectorName() {
		return (portDescription != null)? getAbsolutePath() : tempName;
	}

	private void setAbsoluteConnectorName(String name) {
		if (this.portDescription == null) {
			this.tempName = name;
		} else {
			throw new UnsupportedOperationException("Only intended for JAXB");
		}
	}

	@Override
	public P getConnectorDescription() {
		return portDescription;
	}

	//	private void afterUnmarshal(Unmarshaller u, Object parent) {
	//		this.sourceElement = (E) parent;
	////	}// we need initialization anyway...

	@Override
	public boolean isPresent(IConnection<B> connection) {
		return connections.contains(connection);
	}

	@Override
	public boolean isMultipleConnectionsAllowed() {
		return getMaxAllowedConnections() != 1;//0 is unlimited, 1 is single, other is
	}

	@Override
	public Optional<? extends IConnection<B>> findConnection(IConnectorInstance connectorHandler) {
		return connections.stream().filter(c -> c.doesConnect(connectorHandler)).findAny();
	}

	public boolean isAdditionalConnectionAllowed() {
		int max = getMaxAllowedConnections();
		return max == 0 || max > connections.size();
	}

	@Override
	public Stream<? extends IConnection<B>> streamConnections() {
		return connections.stream().sorted(getConnectionComparator(this));
	}

	@Override
	public B getSingleConnectedPort() { //TODO remove? make Optional?
		if (getDirection() == ConnectorDirection.Input && getConnectorWidth() <= 1) {
			if (connections.isEmpty()) {
				return null;
			} else {
				return connections.iterator().next().getOppositeConnector(this);
			}
		} else {
			throw new IllegalStateException("Only allowed for single inputs, which "+getSourceElement().getName()+"."+getPortName()+" is not.");
		}
	}

	public void onReferencedParameterChanged(IReference source) { // TODO use source
		connections.forEach(IConnection::isValidAndReport);
		sourceElement.notifyConnectionPropertiesChanged(this);
	}

	protected void onReferencedClkChanged(BaseConnectorHandler source) { // TODO use source
		sourceElement.notifyConnectionPropertiesChanged(this);
	}

	@Override
	public boolean isValid(boolean report) {
		boolean valid = getMaxAllowedConnections() == 0 || getMaxAllowedConnections() >= getConnectionCount();

		if (!valid) {
			getDocument().report(ConfigurationIssue.createTooManyConnections(this, "Connector Validation"));
		}

		boolean conValid = connections.stream().reduce(true, (v, con) -> con.isValid(report) && v, Boolean::logicalAnd);
		return valid && conValid;
	}

	@Override
	public void disconnectAll() {
		new ArrayList<>(connections).forEach(BaseConnection::disconnect);// use a copy, to avoid concurrent modifications (connections remove themselves from us)
	}

	protected void notifyConnectionAdded(IConnection<B> connection) {
		sourceElement.notifyConnectionAdded(this, connection);
	}

	private void notifyConnectionRemoved(IConnection<B> connection, boolean invalid) {
		sourceElement.notifyConnectionRemoved(this, connection, invalid);
	}

	public void verifyAdditionalConnection() {
		if (!isAdditionalConnectionAllowed()) {
			throw new ConnectionMismatchException("No more connections allowed!" + getAbsoluteConnectorName());
		}
	}

	/**
	 * Internal, unchecked add. Use {@link IConnection#connect()}!
	 *
	 * @param connection valid connection
	 */
	public void addConnection(BaseConnection<I, B> connection) {
		connections.add(connection);
		notifyConnectionAdded(connection);
	}

	public void removeConnection(IConnection<B> connection) {
		connections.remove(connection);
		notifyConnectionRemoved(connection, false);
	}

	public void invalidateConnection(IConnection<B> connection) {
		connections.remove(connection);
		notifyConnectionRemoved(connection, true);
	}

	public void verifyConnectibleTo(BaseConnectorHandler otherPort) throws ConnectionMismatchException {
		if (!isConnectibleTo(otherPort)) {
			throw new ConnectionMismatchException("Ports are not compatible: " + this.portDescription + " <> " + otherPort.portDescription);
		}
	}

	/**
	 * For JAXB, unchecked add
	 *
	 * @param connection JAXB created connection, validity guaranteed through XMLIDs
	 */
	public void attachConnection(BaseConnection<I, B> connection) {
		connections.add(connection);
	}

	@Override
	public int hashCode() {
		int result = portDescription != null? portDescription.hashCode() : 0;
		result = 31 * result + (sourceElement != null? sourceElement.hashCode() : 0);
		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		BaseConnectorHandler<?, ?, ?, ?> that = (BaseConnectorHandler<?, ?, ?, ?>) o;

		if (portDescription != null? !portDescription.equals(that.portDescription) : that.portDescription != null) {
			return false;
		}
		return sourceElement != null? sourceElement.equals(that.sourceElement) : that.sourceElement == null;

	}

	@Override
	public String toString() {
		return "{" +
			"connections=" + connections +
			", tName=" + tempName +
			'}';
	}

	public interface IConnectorHandlerContainer extends IValueResolver {

		void notifyConnectionRemoved(IConnectorHandler<?, ?> connector, IConnection connection, boolean invalid);

		void notifyConnectionAdded(IConnectorHandler<?, ?> connector, IConnection connection);

		void notifyConnectionPropertiesChanged(IConnectorHandler<?, ?> handler);
	}
}
