/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ClkReference;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.analysis.ParameterReference;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.references.IReference;
import de.tu_darmstadt.rs.spartanmc.devxml.files.buses.BusDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPartialConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ModuleConfiguration;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * Created by Ramon on 10.02.2016.
 */
public class BusHandler extends BaseConnectorHandler<BusHandler, IBusHandler, ModuleConfiguration<?>, Bus> implements IBusHandler {

	private static final Logger logger = LoggerFactory.getLogger(BusHandler.class);

	private ConfigurationDocument document;
	private BusDescription busDefinition;

	public BusHandler(ModuleConfiguration sourceElement, Bus portDescription, ConfigurationDocument document, BusDescription busDefinition) {
		super(sourceElement, portDescription);
		this.document = document;
		this.busDefinition = busDefinition;
	}

	private BusHandler() {
	}

	public void initialize(Bus portDescription, ConfigurationDocument document, BusDescription busDefinition) {
		super.initialize(portDescription);
		this.busDefinition = busDefinition;
		this.document = document;
	}

	@Override
	protected void requestInitialize() {
		sourceElement.requestInitialize(this);
	}

	@Override
	protected void invalidateOnSource() {
		sourceElement.invalidateBusHandler(this);
	}

	@Override
	protected ConfigurationDocument getDocument() {
		return document;
	}

	@Override
	protected String extractPortName(String tempName) {
		return tempName.substring(tempName.lastIndexOf('@') + 1);
	}

	@Override
	public String getPrefixedConnectorName() {
		return '@' + portDescription.getName();
	}

	@Override
	public String toString() {
		return "BusHandler{" +
				"Description=" + portDescription +
				"} " + super.toString();
	}

	@Override
	public void onReferencedParameterChanged(IReference source) {
		super.onReferencedParameterChanged(source);

		streamOppositeConnectors().forEach(bh -> ((BusHandler) bh).onBusChangedProperties());
	}

	private void onBusChangedProperties() {
		sourceElement.notifyConnectionPropertiesChanged(this);
	}

	@Override
	public ConnectorDirection getDirection() {
		return getConnectorDescription().getDirection();
	}

	@Override
	public IConnectorDescription.Type getType() {
		return IConnectorDescription.Type.Bus;
	}

	@Override
	public int getConnectorWidth() {
		return 0;
	}

	@Override
	public int getMaxAllowedConnections() {
		int maxDef = busDefinition.getMaxConnections(getConnectorDescription().getRole());
		int maxImpl = getConnectorDescription().getMaxConnections();
		if (maxDef == 0) {
			return maxImpl;
		} else if (maxImpl == 0) {
			return maxDef;
		} else {
			return Math.min(maxDef, maxImpl);
		}
	}

	@Override
	public boolean isConnectibleTo(IConnectorHandler relativeToPort) {
		if (!(relativeToPort instanceof BusHandler)) {
			return false;
		}
		BusHandler otherPort = ((BusHandler) relativeToPort);
		return IBusHandler.isCompatible(this, otherPort);
	}

	public boolean couldBeConnectibleTo(IConnectorHandler relativeToPort) {
		if (!(relativeToPort instanceof BusHandler)) {
			return false;
		}
		BusHandler otherPort = ((BusHandler) relativeToPort);
		return IBusHandler.couldBeCompatible(this, otherPort);
	}

	@Override
	public Optional<IPartialConnectorHandler<IBusHandler, Bus>> getPartialConnectorHandler() {
		return Optional.empty();
	}

	@Override
	public boolean isDirectionLocked() {
		return true;
	}

	@Override
	public BusDescription getBusDefinition() {
		return busDefinition;
	}

	@Override
	public String getBusType() {
		StringBuilder sb = new StringBuilder();
		Bus bus = getConnectorDescription();
		sb.append(bus.getBusDescriptionName());
		//		List<Parameter> defParams = getBusDefinition().getDefiningParameters(); //TODO implement defParams
		//		if (!defParams.isEmpty()) {
		//			sb.append(defParams.stream().map(p -> Integer.toString(bus.findDefiningParameter(p))).collect(Collectors.joining(", ", " (", ")")));
		//		}
		return sb.toString();
	}

	@Override
	public CalculatedValueResult<? extends Number> resolveParam(ParameterReference param) {
		if (param == Bus.COUNT_PARAM) {
			return CalculatedValueResult.of((long) getConnectionCount());
		} else if (param == Bus.ID_PARAM) {
			Optional<Integer> idx = streamOppositeConnectors().collect(StreamUtil.atMostOne()).map(bh -> bh.resolveBusConnectionIndex(this));
			int actualIdx = idx.orElseGet(() -> {
				document.getIssueSink().report(ConfigurationIssue.createMissingConnection(this, "BusHandler required connection index but is unconnected"));
				return 0;
			});
			return CalculatedValueResult.of((long) actualIdx);
		} else {
			return portDescription.busResolver(sourceElement).resolveParam(param);
		}
	}

	@Override
	public int resolveBusConnectionIndex(IBusHandler other) {
		List<? extends IBusHandler> orderedConnectors = streamOppositeConnectors().collect(Collectors.toList());
		int index = orderedConnectors.indexOf(other);
		if (index < 0) {
			throw new IllegalArgumentException("There is no connection matching " + other);
		} else {
			return index;
		}
	}

	@Override
	public CalculatedValueResult<? extends Number> resolveClkPort(ClkReference clkPort) {
		return this.sourceElement.resolveClkPort(clkPort);
	}

	@Override
	public CalculatedValueResult<? extends String> resolveParamString(ParameterReference param) {
		return resolveParam(param).map(Object::toString);
	}

	@Override
	public CalculatedValueResult<Double> resolveParamFloat(ParameterReference param) {
		return this.sourceElement.resolveParamFloat(param);
	}
}
