/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.exceptions.ConnectionMismatchException;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPartialConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections.BaseConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections.PartialConnection;

import java.util.ArrayList;
import java.util.Optional;
import java.util.TreeSet;
import java.util.stream.Stream;

/**
 * Exists only, so that internal code can use internal methods on the result of {@link #getPartialConnectorHandler()}
 * <p>
 * Created by Ramon on 28.04.2016.
 */
public abstract class PartialBasePortHandler<E extends BaseConnectorHandler.IConnectorHandlerContainer & de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement & IValueResolver, PD extends IConnectorDescription> extends BasePortHandler<E, PD> implements IPartialConnectorHandler<IPortHandler<?>, PD> {

	protected TreeSet<PartialConnection> partialConnections = new TreeSet<>(new PartialConnection.PartialComparator(this));

	public PartialBasePortHandler(E sourceElement, PD portDescription) {
		super(sourceElement, portDescription);
	}

	protected PartialBasePortHandler() {
	}

	/**
	 * For JAXB, unchecked add
	 *
	 * @param connection JAXB created connection, validity guaranteed through XMLIDs
	 */
	public void attachPartialConnection(PartialConnection connection) {
		partialConnections.add(connection);
	}

	/**
	 * Internal, unchecked add. Use {@link IConnection#connect()}!
	 *
	 * @param connection valid connection
	 */
	public void addPartialConnection(PartialConnection connection) {
		if (!isAdditionalPartialConnectionAllowed()) {
			throwIllegalConnectionCombination();
		}

		partialConnections.add(connection);
		notifyConnectionAdded(connection);
	}

	protected void throwIllegalConnectionCombination() {
		throw new IllegalStateException("Cannot support both full and partial connection at the same time");
	}

	public void verifyAdditionalPartialConnection(int start, int width) { //TODO: overlapping connections
		if (getDirection() == ConnectorDirection.Output) {
			return;
		}

		if (partialConnections.isEmpty()) {
			return;
		}


		IConnection<?> lastC = null;
		int end = start + width - 1;
		int nextStart = getConnectorWidth(), lastEnd = -1;

		for (PartialConnection p : partialConnections) {
			nextStart = p.getMyStart(this);

			lastC = p;
			if (end < nextStart && start > lastEnd) {
				return; //fits
			}

			lastEnd = p.getMyEnd(this);
		}

		if (lastC.getMyEnd(this) < start && end < getConnectorWidth()) {
			return;
		}

		throw new ConnectionMismatchException("Would overlap");
	}

	public Stream<? extends IConnection<IPortHandler<?>>> streamFullConnections() {
		return super.streamConnections();
	}

	@Override
	public Stream<? extends IConnection<IPortHandler<?>>> streamPartialConnections() {
		return partialConnections.stream();
	}

	@Override
	public boolean isPartiallyConnected() { //TODO fix
		boolean notPartial = !connections.isEmpty();
		boolean partial = !partialConnections.isEmpty();
		return partial;
	}

	@Override
	public boolean isOverlappingAllowed() {
		return getDirection() == ConnectorDirection.Output;
	}

	@Override
	public boolean isAdditionalPartialConnectionAllowed() {
		return isOverlappingAllowed() || connections.isEmpty();
	}

	@Override
	public boolean isAdditionalFullConnectionAllowed() {
		return isOverlappingAllowed() || partialConnections.isEmpty();
	}

	@Override
	public boolean isPartiallyConnectibleTo(IConnectorHandler toPort) {
		IConnectorDescription toPortDescription = ((BasePortHandler) toPort).portDescription;

		return IConnectorDescription.isTypeCompatible(portDescription, toPortDescription)
				&& IConnectorDescription.isDirectionCompatible(this.portDescription, toPortDescription)
				&& (toPort.getConnectorWidth() < getConnectorWidth() || toPort.getConnectorWidth() > getConnectorWidth());// if eq. then partial is not allowed
	}

	@Override
	public boolean isConnected() {
		return super.isConnected() || !partialConnections.isEmpty();
	}

	@Override
	public int getConnectionCount() {
		if (isOverlappingAllowed()) {
			return partialConnections.size() + super.getConnectionCount();
		} else {
			return isPartiallyConnected()? partialConnections.size() : super.getConnectionCount();
		}
	}

	@Override
	public boolean isPresent(IConnection<IPortHandler<?>> connection) {
		return super.isPresent(connection) || partialConnections.contains(connection);
	}

	@Override
	public Optional<? extends IConnection<IPortHandler<?>>> findConnection(IConnectorInstance connectorHandler) {
		Optional<? extends IConnection<IPortHandler<?>>> connection = super.findConnection(connectorHandler);
		return (connection.isPresent())? connection : streamPartialConnections().filter(c -> c.doesConnect(connectorHandler)).findAny();
	}

	@Override
	public boolean isAdditionalConnectionAllowed() {
		return (!isPartiallyConnected() || isOverlappingAllowed()) && super.isAdditionalConnectionAllowed();
	}

	@Override
	public Stream<? extends IConnection<IPortHandler<?>>> streamConnections() {
		if (isOverlappingAllowed()) {
			return Stream.concat(super.streamConnections(), streamPartialConnections());
		} else {
			return isPartiallyConnected()? streamPartialConnections() : super.streamConnections();
		}
	}

	@Override
	public boolean isValid(boolean report) {
		if (isPartiallyConnected()) {
			int index = 0;
			boolean valid = true;
			boolean allowOverlapping = getDirection() == ConnectorDirection.Output;

			for (PartialConnection connection : partialConnections) {
				if (allowOverlapping) {
					if (connection.getMyStart(this) < 0 || connection.getMyEnd(this) >= getConnectorWidth()) {
						valid = false;
						if (report) {
							getDocument().report(ConfigurationIssue.createMismatchedConnection(this, connection, "Port characteristics changed"));
						}
					}
				} else {
					if (connection.getMyStart(this) < index) {
						valid = false;
						if (report) {
							getDocument().report(ConfigurationIssue.createMismatchedConnection(this, connection, "Port characteristics changed"));
						}
					}
					index = connection.getMyEnd(this);
				}
				valid &= connection.isValid(report);
			}

			valid = index <= getConnectorWidth();

			//			if (!valid && report)
			//				ConfigurationIssue.createOverlappingPartialConnections(getDocument(), this, "Port characteristics changed");//TODO

			return valid;
		} else {
			return super.isValid(report);
		}
	}

	@Override
	public void disconnectAll() {
		super.disconnectAll();
		new ArrayList<>(partialConnections).forEach(BaseConnection::disconnect);
	}

	@Override
	public void addConnection(BaseConnection<BasePortHandler<?, ?>, IPortHandler<?>> connection) {
		if (!isAdditionalFullConnectionAllowed()) {
			throwIllegalConnectionCombination();
		}

		super.addConnection(connection);
	}

	@Override
	public void removeConnection(IConnection<IPortHandler<?>> connection) {
		partialConnections.remove(connection);
		super.removeConnection(connection);
	}

	@Override
	public IPortHandler<?> getSingleConnectedPort() {
		final IPortHandler<?> singleConnectedPort = super.getSingleConnectedPort();
		if (singleConnectedPort != null) {
			return singleConnectedPort;
		}
		return partialConnections.iterator().next().getOppositeConnector(this);
	}
}
