/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.handlers;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.CalculatedValueResult;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.issues.ConfigurationIssue;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.ConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.jaxb.jc3.connections.BaseConnection;

import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Objects;
import java.util.Optional;
import javax.xml.bind.annotation.XmlAttribute;

/**
 * Created by Ramon on 27.07.2016.
 */
public class PinHandler extends BasePortHandler<ConfigurationDocument, TargetDescription.IoPin> implements IPinHandler {

	private static final Logger logger = LoggerFactory.getLogger(PinHandler.class);

	@XmlAttribute
	private String ioStd;
	private boolean lockedIoStd = false;
	@XmlAttribute
	private String slewRate;
	@XmlAttribute
	private DeviceDescription.Termination termination = DeviceDescription.Termination.None;
	@XmlAttribute
	private String customUserConstraints;
	@XmlAttribute
	private boolean inverted;
	@XmlAttribute
	private ConnectorDirection direction;
	private boolean lockedDirection = false;
	@XmlAttribute
	private String drive;
	@XmlAttribute
	private Long frequencyOverride;

	public PinHandler(ConfigurationDocument parent, TargetDescription.IoPin pinDescription) {
		super(parent, pinDescription);
		defaultAttributes();
	}

	private PinHandler() {
	}

	public static PinHandler migrate(PinHandler oldHandler, TargetDescription.IoPin pinDesc) {
		try {
			PinHandler newHandler = new PinHandler(oldHandler.sourceElement, pinDesc);

			newHandler.setInverted(oldHandler.isInverted());
			newHandler.setCustomUserConstraints(oldHandler.getCustomUserConstraints());
			newHandler.setDrive(oldHandler.getDrive());
			if (newHandler.direction != oldHandler.direction) {
				newHandler.setTargetDirection(oldHandler.getDirection());// force exception of incompatible
			}
			newHandler.setSlewRate(oldHandler.getSlewRate());
			if (!Objects.equals(newHandler.ioStd, oldHandler.ioStd)) {
				newHandler.setIoStd(oldHandler.getIoStd());
			}

			newHandler.connections = oldHandler.connections;
			newHandler.connections.removeIf(c -> {
				try {
					c.migrate(oldHandler, newHandler);
					return false;
				} catch (Exception e) {
					logger.warn("Removing connection {} during migration of {}", c, pinDesc, e);
					return true;
				}
			});

			return newHandler;

		} catch (Exception e) {
			logger.warn("Cannot migrate {}, incompatible!", oldHandler, e);
			return null;
		}
	}

	private static boolean hasChanged(String newValue, String oldValue) {
		return !Objects.equals(newValue, oldValue);
	}

	private void defaultAttributes() {
		String dIoStd = portDescription.getIostd();
		lockedIoStd = dIoStd != null;
		if (lockedIoStd && !dIoStd.equals(ioStd) && ioStd != null) {
			notifyTargetOverride("IoStd", dIoStd);
		}
		if (ioStd == null || lockedIoStd) {
			ioStd = portDescription.getIostd();
		}

		ConnectorDirection dDirection = portDescription.getTargetDirection();
		lockedDirection = portDescription.isDirectionLocked();
		if (lockedDirection && direction != dDirection) {
			if (direction != null) {
				notifyTargetOverride("Direction", dDirection);
			}
			direction = portDescription.getTargetDirection();
		} else if (direction == null) {
			direction = portDescription.getDefaultDirection();
		}
	}

	private void notifyTargetOverride(String name, Object defaultVal) {
		logger.warn("TargetDescription Override: {} attribute '{}' is now locked to '{}'", getPinDescription(), name, defaultVal);
	}

	@Override
	protected void requestInitialize() {
		getDocument().requestInitialize(this);
	}

	@Override
	protected void invalidateOnSource() {
		sourceElement.pinHandlers.remove(this.getPortName());
		logger.warn("Pin {} had become invalid and was removed", this);
		sourceElement.notifyTargetChanged();
	}

	@Override
	protected ConfigurationDocument getDocument() {
		return sourceElement;
	}

	@Override
	public void initialize(TargetDescription.IoPin portDescription) {
		super.initialize(portDescription);
		defaultAttributes();
	}

	@Override
	protected String extractPortName(String tempName) {
		return tempName.substring(tempName.lastIndexOf('#') + 1);
	}

	@Override
	public String getPrefixedConnectorName() {
		return '#' + portDescription.getName();
	}

	@Override
	public boolean isValid(boolean report) {
		boolean termValid = (getSupportedTerminations().contains(termination));
		boolean slewValid = isInputOrFamiliar()? slewRate == null : (slewRate == null || getSupportedSlewRates().contains(slewRate));
		boolean directionValid = direction != null && direction != ConnectorDirection.Inherit && (isDirectionLocked() || connections.isEmpty());
		boolean ioStdValid = lockedIoStd? ioStd.equals(portDescription.getIostd()) : ioStd == null || getSupportedIoStd().contains(ioStd);
		boolean driveValid = isInputOrFamiliar()? drive == null : (drive == null || getSupportedDrives().contains(drive));

		boolean valid = super.isValid(report);

		boolean allValid = valid && termValid && slewValid && directionValid && ioStdValid && driveValid;

		if (!allValid && report) {
			getDocument().report(ConfigurationIssue.createInvalidPinProperties(this, "Pin Validity check"));
		}

		return allValid;
	}

	@Override
	public void addConnection(BaseConnection<BasePortHandler<?, ?>, IPortHandler<?>> connection) {
		if (!isDirectionLocked()) {
			ConnectorDirection externalDirection = connection.getOppositeConnector(this).getConnectorDescription().getDirection();
			setTargetDirection(externalDirection);
			logger.debug("Reconfiguring pin {} to {}", getPrefixedConnectorName(), this.direction);
		}
		super.addConnection(connection);
	}

	@Override
	public int getMaxAllowedConnections() { //external direction.
		return (direction == ConnectorDirection.Input)? 0 : 1;
	}

	@Override
	public Optional<? extends PartialBasePortHandler<ConfigurationDocument, TargetDescription.IoPin>> getPartialConnectorHandler() {
		return Optional.empty();
	}

	@Override
	public String toString() {
		return "PinHandler{" +
			"ioStd='" + ioStd + '\'' +
			", lockedIoStd=" + lockedIoStd +
			", slewRate='" + slewRate + '\'' +
			", termination=" + termination +
			", customUserConstraints='" + customUserConstraints + '\'' +
			", inverted=" + inverted +
			", direction=" + direction +
			", lockedDirection=" + lockedDirection +
			", drive='" + drive + '\'' +
			", frequencyOverride=" + frequencyOverride +
			"} " + super.toString();
	}

	@Override
	public List<String> getSupportedIoStd() {
		return getDocument().getTarget().getDeviceDescription().getIostandards();
	}

	@Override
	public List<String> getSupportedSlewRates() {
		return getDocument().getTarget().getDeviceDescription().getSlewRates();
	}

	@Override
	public List<DeviceDescription.Termination> getSupportedTerminations() {
		return getDocument().getTarget().getDeviceDescription().getTerminations();
	}

	@Override
	public List<String> getSupportedDrives() {
		return getDocument().getTarget().getDeviceDescription().getDrives();
	}

	@Override
	public @Nullable String getIoStd() {
		return ioStd;
	}

	@Override
	public void setIoStd(String ioStd) {
		if (lockedIoStd) {
			throw new IllegalArgumentException("this pin's ioStd is locked");
		}
		if (ioStd != null && !getSupportedIoStd().contains(ioStd)) {
			throw new IllegalArgumentException("ioStd " + ioStd + " is not allowed");
		}

		if (hasChanged(ioStd, this.ioStd)) {
			this.ioStd = ioStd;
			notifyChange();
		}
	}

	@Override
	public String getSlewRate() {
		return slewRate;
	}

	@Override
	public void setSlewRate(String slewRate) {
		if (slewRate != null && isInputOrFamiliar()) {
			throw new IllegalArgumentException("Only Outputs can have a SlewRate");
		}

		if (slewRate != null && !getSupportedSlewRates().contains(slewRate)) {
			throw new IllegalArgumentException("slewRate " + slewRate + " is not allowed");
		}

		if (hasChanged(slewRate, this.slewRate)) {
			this.slewRate = slewRate;
			notifyChange();
		}
	}

	@Override
	public DeviceDescription.Termination getTermination() {
		return termination;
	}

	@Override
	public void setTermination(DeviceDescription.Termination termination) {
		if (!getSupportedTerminations().contains(termination)) {
			throw new IllegalArgumentException("termination " + termination + " is not allowed");
		}

		if (this.termination != termination) {
			this.termination = termination;
			notifyChange();
		}
	}

	@Override
	public String getCustomUserConstraints() {
		return customUserConstraints;
	}

	public void setCustomUserConstraints(String customUserConstraints) {
		if (hasChanged(customUserConstraints, customUserConstraints)) {
			this.customUserConstraints = customUserConstraints;
			notifyChange();
		}
	}

	@Override
	public boolean isInverted() {
		return inverted;
	}

	@Override
	public void setInverted(boolean inverted) {
		if (this.inverted != inverted) {
			this.inverted = inverted;
			notifyChange();
		}
	}

	@Override
	public String getDrive() {
		return drive;
	}

	@Override
	public void setDrive(String drive) {
		if (drive != null && isInputOrFamiliar()) {
			throw new IllegalArgumentException("Only Outputs can have a Drive");
		}

		if (drive != null && !getSupportedDrives().contains(drive)) {
			throw new IllegalArgumentException("drive " + drive + " is not allowed");
		}

		if (hasChanged(drive, this.drive)) {
			this.drive = drive;
			notifyChange();
		}
	}

	@Override
	public ConnectorDirection getTargetDirection() {
		return direction;
	}

	@Override
	public void setTargetDirection(ConnectorDirection direction) {
		if (isDirectionLocked()) {
			throw new IllegalArgumentException("this pin's direction is locked");
		}

		if (direction == null || direction == ConnectorDirection.Inherit) {
			throw new IllegalArgumentException("Invalid direction");
		}

		if (this.direction != direction) {
			this.direction = direction;
			if (isInputOrFamiliar()) {
				slewRate = null;
				drive = null;
			}
			notifyChange();
		}
	}

	@Override
	public boolean isIoStdLocked() {
		return lockedIoStd;
	}

	@Override
	public TargetDescription.IoPin getPinDescription() {
		return getConnectorDescription();
	}

	@Override
	public CalculatedValueResult<Long> getFrequency() {
		if (frequencyOverride != null) {
			return CalculatedValueResult.of(frequencyOverride);
		}
		long freq = getPinDescription().getFreq();
		if (freq == 0) {
			return CalculatedValueResult.customError("Pin " + getPortName() + " has no frequency");
		} else {
			return CalculatedValueResult.of(freq);
		}
	}

	@Override
	public Long getFrequencyOverride() {
		return frequencyOverride;
	}

	@Override
	public void setFrequencyOverride(Long frequencyOverride) {
		this.frequencyOverride = frequencyOverride;
	}

	@Override
	public boolean isInputOrFamiliar() {
		return lockedDirection && (direction == ConnectorDirection.Input || direction == ConnectorDirection.InOut);
	}

	private void notifyChange() {
		getDocument().notifyPinPropertiesChanged(this);
	}

	@Override
	public ConnectorDirection getDirection() {
		return direction.getReverse();
	}

	@Override
	public IConnectorDescription.Type getType() {
		return IConnectorDescription.Type.Pin;
	}

	@Override
	public int getConnectorWidth() {
		return 1;
	}

	@Override
	public boolean isDirectionLocked() {
		return lockedDirection || !connections.isEmpty();
	}
}
