/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.layouting.addressLayouting;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.Memory;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.MemoryLocator;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IFirmwareConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.linker.SingleCoreFirmwareBinary;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemFirmware;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryBlock;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryLineLayout;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryMap;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by Ramon on 18.05.2016.
 */
public class GlobalAddressLayout implements IGlobalAddressLayout {

	private static final Logger logger = LoggerFactory.getLogger(GlobalAddressLayout.class);

	private final IConfigurationDocument document;

	private Map<MemoryModuleDescription.ExtMemory, Map<Integer, AddressBlock.MemAddressBlock>> sharedMemoryIndexes = new HashMap<>();
	private Map<ISubsystemConfiguration, AddressLayout> layouts;
	private Set<AddressLayout> singleLayouts;
	private Set<SharedAddressLayout> multiLayouts;
	private Map<IMemoryAddressBlock, SingleCoreFirmwareBinary> binaries;

	public GlobalAddressLayout(@NotNull IConfigurationDocument document) {
		layouts = document.getSubsystems().stream().collect(Collectors.toMap(Function.identity(), sub -> AddressLayout.getCurrent(sub, this)));
		this.document = document;

		analyzeOverlaps();
		mapBinaries();
	}

	private static Optional<IFirmwareConfiguration> findExportingFirmware(AddressLayout layout) {
		List<IFirmwareConfiguration> availableExporters = layout.streamFirmwareAttachedInitializableBlocks()
				.sorted(Comparator.<AddressBlock.MemAddressBlock, String>comparing(mem -> mem.getSourceModule().getName())
				.thenComparing(mem -> mem.getFirmware().getFirmwareName()))
				.map(AddressBlock.MemAddressBlock::getFirmware)
				.distinct()
				.collect(Collectors.toList());
		Optional<IFirmwareConfiguration> exporter = (!availableExporters.isEmpty())? Optional.of(availableExporters.get(0)) : Optional.empty();
		if (availableExporters.size() > 1) {
			logger.warn("Subsystem '{}' exporting '{}'. (multiple exporting firmwares were found ({}), chose by alphabetical order of memoryModules and firmware name)", layout.getSubsystem().getGlobalName(), exporter.get().getFirmwareName(), availableExporters.stream().map(IFirmwareConfiguration::getFirmwareName).collect(Collectors.toList()));
		}
		return exporter;
	}

	private void mapBinaries() {
		Set<SingleCoreFirmwareBinary> binaries = Stream.concat(
				singleLayouts.stream().filter(AddressLayout::hasMemory).map(l -> new SingleCoreFirmwareBinary(l, findExportingFirmware(l).orElseThrow(UnsupportedOperationException::new), false)),
				streamFirmwareBinaries(multiLayouts)
		)
				.filter(Objects::nonNull)
				.collect(Collectors.toSet());

//		Map<IFirmwareConfiguration, List<SingleCoreFirmwareBinary>> reverseLookup = binaries.stream().collect(Collectors.groupingBy(SingleCoreFirmwareBinary::getFirmware));

		this.binaries = new HashMap<>();

		binaries.forEach(b -> {
			AddressLayout layout = b.getLayout();
			layout.streamInitializableBlocks().forEach(mem -> this.binaries.put((IMemoryAddressBlock) mem, b));
		});
	}

	@Override
	public <T> Stream<? extends T> streamMultiCoreLayouts(Function<? super IGlobalAddressLayout.SharedAddressLayout, ? extends T> handleMultiCoreLayout, Function<? super AddressLayout, ? extends T> handlePartOfSharedLayout) {
		return multiLayouts.stream().flatMap(sharedAddressLayout -> {
			if (canGenerateMultiCoreFirmware(sharedAddressLayout)) {
				return Stream.of(handleMultiCoreLayout.apply(sharedAddressLayout));
			} else {
				return sharedAddressLayout.getContainedLayouts().stream()//TODO could be possible to wrap this in a class that adds sanity checks on top
						.map(handlePartOfSharedLayout);
			}
		}).filter(Objects::nonNull);
	}



	private final Map<IFirmwareBinary, MemFirmware> firmwareCache = new HashMap<>();

	private MemFirmware getFirmware(IFirmwareBinary binary) {
		return firmwareCache.computeIfAbsent(binary, b -> {
			final Path location = b.getFirmware().getLocation(document.getProject());
			String locationString = null;
			if (location != null) {
				locationString = location.toString();
			}
			return new MemFirmware(b.getName(), locationString);
		});
	}

	private MemoryLocator getGlobalInstance(IMemoryAddressBlock block) {
		IModuleConfiguration sourceModule = block.getSourceModule();
		final MemoryLocator relativeName = block.resolveBRamInstance();
		if (relativeName==null)
			return null;
		return new MemoryLocator(sourceModule.getGlobalName() + '/', relativeName);
	}

	private MemoryMap memoryMap = null;

	@Override
	public MemoryMap getMemoryMap() {
		if (memoryMap == null) {

			Set<MemFirmware> firmwares = new HashSet<>();

			Set<MemoryBlock> memoryBlocks = streamLayouts().flatMap(AddressLayout::streamFirmwareAttachedInitializableBlocks)
					.distinct()
					.map(b -> {
						MemFirmware firmware = getFirmware(getBinaries().get(b));
						firmwares.add(firmware);
						MemoryModuleDescription.ExtMemory memory = b.getMemory();
						return new MemoryBlock(getGlobalInstance(b), b.getActualStart(), b.getLength(), b.getSourceModule().getGlobalName(), memory.getLineWidth(), memory.getBlockWidth(), firmware, MemoryLineLayout.guessMemoryLineLayout(memory));
					})
					.collect(Collectors.toSet());

			streamLayouts().flatMap(AddressLayout::streamDetachedInitializableBlocks)
					.distinct()
					.map(b -> {
						MemFirmware firmware = getFirmware(binaries.get(b));
						Memory memory = b.getMemory();
						return new MemoryBlock(getGlobalInstance(b), b.getActualStart(), b.getLength(), b.getSourceModule().getGlobalName(), memory.getLineWidth(), memory.getBlockWidth(), firmware, MemoryLineLayout.guessMemoryLineLayout(memory));
					})
					.forEach(memoryBlocks::add);

			memoryMap = new MemoryMap(memoryBlocks, firmwares, document.getBinaryOutputFormat(), document.getTarget().getDevice().getName(), document.getTarget().getToolchain().getToolchain());
		}
		return memoryMap;
	}

	private Stream<? extends SingleCoreFirmwareBinary> streamFirmwareBinaries(Set<GlobalAddressLayout.SharedAddressLayout> sharedAddressLayouts) {
		return streamMultiCoreLayouts(
				multi -> {
					throw new UnsupportedOperationException();
				}, partOfSharedLayout -> {
					Optional<IFirmwareConfiguration> exporter = findExportingFirmware(partOfSharedLayout);
					return exporter.map(fw -> new SingleCoreFirmwareBinary(partOfSharedLayout, fw, true)).orElse(null);
				});
	}

	private static boolean canGenerateMultiCoreFirmware(IGlobalAddressLayout.SharedAddressLayout sharedAddressLayout) {
		return false;//TODO we cannot, currently unimplemented, so fall back to generating a firmware for each part-taking core. user must ensure that they are equal where it counts
	}

	@Override
	public void addAllAddressLayoutDependencies(Set<Path> inputs, Map<String, Object> properties) {
		addGlobalAddressLayoutDependencies(document, layouts.values(), inputs, properties);
	}

	public static void addGlobalAddressLayoutDependencies(IConfigurationDocument document, Collection<AddressLayout> addressLayouts, Set<Path> inputs, Map<String, Object> properties) {
		List<? extends ISubsystemConfiguration> subsystems = addressLayouts.stream().map(AddressLayout::getSubsystem).collect(Collectors.toList());

		addressLayouts.forEach(layout -> {
			layout.addAddressDependencies(properties);
		});

		subsystems.forEach(subsystem -> {
			AddressLayout.addSubsystemLayoutSpecDependencies(properties, subsystem);
			AddressLayout.addAttachedMemoryModuleDependencies(inputs, properties, subsystem);
		});
	}

	@Override
	public AddressLayout getLayout(ISubsystemConfiguration subsystem) {
		return layouts.get(subsystem);
	}

	@Override
	public Stream<AddressLayout> streamLayouts() {
		return layouts.values().stream();
	}

	@Override
	public Map<Integer, AddressBlock.MemAddressBlock> getSharedMemoryIndexes(MemoryModuleDescription.ExtMemory memory) {
		return sharedMemoryIndexes.computeIfAbsent(memory, m -> new HashMap<>());
	}

	private void analyzeOverlaps() {
		Map<AddressBlock, List<AddressLayout>> overlaps = layouts.values().stream().flatMap(layout ->
				layout.streamAllBlocks().map(block -> new StreamUtil.Pair<>(block, layout))
		).collect(Collectors.groupingBy(StreamUtil.Pair::getA, Collectors.mapping(StreamUtil.Pair::getB, Collectors.toList())));

		Set<Set<AddressLayout>> multiLayouts = new HashSet<>();
		singleLayouts = new HashSet<>();

		overlaps.values().forEach(list -> {
			if (list.size() == 1) {
				singleLayouts.add(list.get(0));
			} else if (list.size() > 1) {
				multiLayouts.add(new HashSet<>(list));
			} else {
				throw new IllegalStateException("WTF?");
			}
		});

		multiLayouts.stream().flatMap(Collection::stream).forEach(singleLayouts::remove);
		this.multiLayouts = multiLayouts.stream().map(m -> new SharedAddressLayout(m, name(m))).collect(Collectors.toSet());
	}

	@Override
	public Set<AddressLayout> getSingleLayouts() {
		return singleLayouts;
	}

	@Override
	public Set<SharedAddressLayout> getMultiLayouts() {
		return multiLayouts;
	}

	@Override
	public Map<? extends IMemoryAddressBlock, SingleCoreFirmwareBinary> getBinaries() {
		return binaries;
	}

	private char nextGlobalLayoutId = 'A';

	private String name(Set<AddressLayout> layouts) {
		if (nextGlobalLayoutId >= 'Z') {
			throw new IllegalArgumentException("Too many layouts");
		}

		return "globalLayout_" + String.valueOf(nextGlobalLayoutId++);
	}

	public static class SharedAddressLayout implements IGlobalAddressLayout.SharedAddressLayout {

		private final Set<AddressLayout> containedLayouts;
		private final String name;

		public SharedAddressLayout(Set<AddressLayout> containedLayouts, String name) {
			this.containedLayouts = containedLayouts;
			this.name = name;
		}

		@Override
		public Set<AddressLayout> getContainedLayouts() {
			return containedLayouts;
		}

		@Override
		public String getName() {
			return name;
		}

		@Override
		public void addDependencies(Set<Path> inputs, Map<String, Object> properties) {
			throw new UnsupportedOperationException();
		}
	}
}
