/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.util;

import de.tu_darmstadt.rs.jcommander.validators.DirectoryExistsValidator;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.util.EnvironmentVariablesUtil;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import com.beust.jcommander.ParametersDelegate;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.URISyntaxException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

/**
 * Implements Command-Line Arguments.
 * Simplifies and Unifies configuration and creation of {@link LibJConfigProject} and {@link Library}
 */
public class LibJConfigBuilder {

	public static final String SPARTANMC_ROOT = "SPARTANMC_ROOT";
	private static final Logger logger = LoggerFactory.getLogger(LibJConfigBuilder.class);

	@Parameter(names = {"-f", "--config-file"}, description = "<File> to open at startup")
	private Path explicitConfigFile;

	@Parameter(names = {"-r", "--root"}, description = "Explicitly set the SPARTANMC_ROOT Directory", validateValueWith = DirectoryExistsValidator.class)
	private Path spartanRoot;

	@Parameter(names = {"-p", "--project-dir"}, description = "Select a folder other than the current working directory as project directory", validateValueWith = DirectoryExistsValidator.class)
	private Path projectDir = Paths.get("."); //working dir is jc3 dir by default

	@Parameter(names = {"--no-version-check"}, description = "Do not check if output files were generated by a different version of JConfig. Should only be used for debugging, as it may lead to files not being updated.")
	private boolean noVersionCheck = false;

	@Parameter(names = {"--inplace-root"}, description = "Attempts to infer SPARTANMC_ROOT based on where the executable is located. Meant for runs from within gradle or IDE", hidden = true)
	private boolean isInplaceRun = false;

	@Parameter(names = {"--no-local-hardware"}, description = "Prevent the libjconfig from automatically adding a local 'hardware' directory to the module searchpath")
	private boolean disableAutoAddLocalHardwareDir = false;

	@ParametersDelegate
	private Library.Builder libBuilder = new Library.Builder();

	private final String applicationVersion;
	private final boolean useDefaultSearchpaths;

	private List<IApplicationConfiguration.ILibraryReadyListener> libraryReadyListeners = new ArrayList<>();
	private List<Consumer<LibJConfigProject>> projectReadyListeners = new ArrayList<>();

	private boolean useLastBuildConfig = false;
	private BiConsumer<IDocumentHandler, Path> opener = null;
	private boolean requiresConfigFile = false;
	private Path configFile = null;

	protected LibJConfigProject project;
	protected Library library;

	/**
	 * Start a Builder
	 *
	 * @param applicationVersion any String describing the current application. Will be used in document
	 * @param useDefaultSearchpaths always add searchpaths derived from spartan root to Library
	 */
	public LibJConfigBuilder(String applicationVersion, boolean useDefaultSearchpaths) {
		this.applicationVersion = applicationVersion;
		this.useDefaultSearchpaths = useDefaultSearchpaths;
	}

	/**
	 * Try to read 'SPARTANMC_ROOT' from ENV
	 *
	 * @return result as Path
	 *
	 * @throws ParameterException JCommander Exception if ENV invalid
	 */
	public static Path getEnvSpartanRoot() throws ParameterException {
		return EnvironmentVariablesUtil.getEnvironmentVariableAsFolder(EnvironmentVariablesUtil.SPARTANMC_ROOT);
	}

	/**
	 * configure Builder to throw an Exception if {@link #build()} is called without an existing configFile set
	 *
	 * @return this
	 */
	public LibJConfigBuilder requiresConfigFile() {
		this.requiresConfigFile = true;

		return this;
	}

	/**
	 * configure Builder to search buildCache for last built config file, if none was explicitly set
	 *
	 * @return this
	 */
	public LibJConfigBuilder useLastBuiltConfig() {
		this.useLastBuildConfig = true;

		return this;
	}

	/**
	 * configure Library.Builder
	 *
	 * @param configurator Closure, executed immediately
	 * @return this
	 */
	public LibJConfigBuilder configureLibraryBuilder(Consumer<Library.Builder> configurator) {
		configurator.accept(libBuilder);

		return this;
	}

	/**
	 * Code that will be executed as soon as the Library has been loaded.
	 * Order of execution: order of adding (resues libraryReadyListeners)
	 *
	 * @param runWithLib Closure, executed delayed, passed Library
	 * @return this
	 */
	public LibJConfigBuilder executeWithLibrary(Consumer<Library> runWithLib) {
		libraryReadyListeners.add(runWithLib::accept);
		return this;
	}

	/**
	 * Set a closure that will delegate opening of a config file.
	 * Allows replacing default open behavior to include autosave recovery / UI etc.
	 *
	 * @param opener closure: passed documentHandler and file to open
	 * @return this
	 */
	public LibJConfigBuilder autoOpenConfig(BiConsumer<IDocumentHandler, Path> opener) {
		this.opener = opener;
		return this;
	}

	/**
	 * Auto open given config file using {@link IDocumentHandler#open(Path)}
	 *
	 * @return this
	 */
	public LibJConfigBuilder autoOpenConfig() {
		this.opener = IDocumentHandler::open;

		return this;
	}

	/**
	 * Overwrite '--root'
	 *
	 * @param root valid Spartan Root
	 * @return this
	 */
	public LibJConfigBuilder setSpartanRoot(Path root) {
		this.spartanRoot = root;
		return this;
	}

	/**
	 * Set Project Directory
	 *
	 * @param projectDir the project dir
	 */
	public void setProjectDir(Path projectDir) {
		this.projectDir = projectDir;
	}

	public LibJConfigBuilder executeWithProject(Consumer<LibJConfigProject> runWithPrj) {
		projectReadyListeners.add(runWithPrj);

		return this;
	}

	/**
	 * Build Project first, then libray
	 *
	 * @return project, when library is ready as well
	 */
	public LibJConfigProject build() {
		validate();

		project = new LibJConfigProject(projectDir, new AppConfig(), noVersionCheck);

		projectReadyListeners.forEach(c -> c.accept(project));

		detectConfigFile();

		maybeOpenFile();

		buildLibrary();

		return project;
	}

	protected void validate() {
		getSpartanRoot(); // force check
	}

	protected void maybeOpenFile() {
		if (configFile != null && opener != null) {
			if (!Files.isRegularFile(configFile)) {
				throw new ParameterException(String.format("Parameter 'config-file' must be a file! was %s", configFile));
			}
			opener.accept(project.getDocumentHandler(), configFile);
		}
	}

	protected void detectConfigFile() {
		if (explicitConfigFile == null && useLastBuildConfig) {
			configFile = project.getLastConfigFile().orElse(null);
		} else {
			configFile = explicitConfigFile;
		}

		if (requiresConfigFile && configFile == null) {
			throw new ParameterException("Config File required (either using '--config-file', or via last built)");
		}
	}

	public static Path getInplaceSpartanRoot() {
		try {
			Path path = Paths.get(LibJConfigBuilder.class.getProtectionDomain().getCodeSource().getLocation().toURI());
			path = path.getParent().getParent().getParent().getParent().getParent().getParent();
			logger.info("Path determined from JAR: {}",path);
			//One more level?
			if (path.getFileName().toString().equals("javaTools")) {
				path = path.getParent();
			}
			//One more level?
			if (path.getFileName().toString().equals("src")) {
				path = path.getParent();
			}
			return path;
		} catch (URISyntaxException e) {
			throw new ParameterException("Failed to determine SPARTANMC_ROOT based on JAR-File location!", e);
		}
	}

	public Path getSpartanRoot() {
		if (spartanRoot == null) {
			if (isInplaceRun) {
				spartanRoot = getInplaceSpartanRoot();
			} else {
				spartanRoot = getEnvSpartanRoot();
			}
		}
		return spartanRoot;
	}

	/**
	 * Whether this Builder currently knows a configFile
	 *
	 * @return true if configFile is set
	 */
	public boolean hasExplicitConfigFile() {
		return explicitConfigFile != null;
	}

	public Path getExplicitConfigFile() {
		return explicitConfigFile;
	}

	public LibJConfigBuilder setExplicitConfigFile(Path explicitConfigFile) {
		this.explicitConfigFile = explicitConfigFile;
		return this;
	}

	private Library buildLibrary() {
		//Some JAXB error messages are only sensible with english locale set...
		Locale defaultLocale = Locale.getDefault();
		Locale.setDefault(Locale.ENGLISH);

		try {

			if (useDefaultSearchpaths) {
				libBuilder.addDefaultSearchpaths(getSpartanRoot());
				Path hardwareDir = projectDir.resolve("hardware").normalize();
				if (!disableAutoAddLocalHardwareDir && Files.exists(hardwareDir)) {
					logger.info("Adding project specific folder {} to module searchpath", hardwareDir);
					libBuilder.addModuleDescriptionSearchpath(hardwareDir);
				}
			}

			synchronized (this) {
				library = libBuilder.build();
				libBuilder = null;
				Locale.setDefault(defaultLocale);
			}

		} finally {
			Locale.setDefault(defaultLocale);
		}

		notifyLibraryReady();
		return library;
	}

	private void notifyLibraryReady() {
		final List<IApplicationConfiguration.ILibraryReadyListener> readyListeners;

		//Only synchronize while getting the list
		synchronized (this) {
			readyListeners = this.libraryReadyListeners;
			this.libraryReadyListeners = null;
		}

		for (IApplicationConfiguration.ILibraryReadyListener listener : readyListeners) {
			listener.onLibraryReady(library);
		}
	}

	private class AppConfig implements IApplicationConfiguration {

		@Override
		public String getApplicationVersion() {
			return applicationVersion;
		}

		@Override
		public Library getLibrary() {
			return library;
		}

		@Override
		public Path getSpartanRoot() {
			return spartanRoot;
		}

		@Override
		public void addLibraryReadyListener(ILibraryReadyListener listener) {
			synchronized (LibJConfigBuilder.this) {
				libraryReadyListeners.add(listener);
			}
		}

		@Override
		public boolean addOrCallReadyListener(ILibraryReadyListener listener) {
			synchronized (LibJConfigBuilder.this) {
				if (isLibraryReady()) {
					listener.onLibraryReady(library);
					return true;
				} else {
					libraryReadyListeners.add(listener);
					return false;
				}
			}
		}

		@Override
		public boolean isLibraryReady() {
			synchronized (LibJConfigBuilder.this) {
				return libraryReadyListeners == null;
			}
		}
	}
}
