/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.util;

import de.tu_darmstadt.rs.jcommander.converter.FileListConverter;
import de.tu_darmstadt.rs.jcommander.validators.PathExistsValidator;
import de.tu_darmstadt.rs.jcommander.validators.PathsExistValidator;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPluginEngine;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.exceptions.JConfigBuildValidationException;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.importExport.IDocumentImporter;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.function.BiFunction;
import java.util.function.Consumer;

/**
 * Extends {@link LibJConfigBuilder} with additional functionality to run stuff / import, schedule events depending on building
 */
public class LibJConfigRunner extends LibJConfigBuilder {

	private static final Logger logger = LoggerFactory.getLogger(LibJConfigRunner.class);

	@Parameter(names = {"-u", "--update"}, description = "Build the document, last build in current project, or the one supplied with '--config-file'")
	private boolean update;

	@Parameter(names = {"-a", "--build-all"}, description = "Force  \"Build All\" instead of incremental builds. Only honored in combination with '-u'")
	private boolean buildAll = false;

	@Parameter(names = {"-i", "--import"}, description = "Try to import the given file", validateValueWith = PathExistsValidator.class)
	private Path importFile;

	@Parameter(names = {"-t", "--target"}, description = "Describe the target for which to create a configuration (<target>[:<device>]")
	private String targetConf;

	@Parameter(names = {"--script"}, description = "Execute the given script files when a document has been loaded. Multiple Values are comma separated", listConverter = FileListConverter.class, validateValueWith = PathsExistValidator.class)
	private List<Path> scriptFiles = new ArrayList<>();

	private List<Consumer<IDocumentHandler>> documentLoadedListeners = new ArrayList<>();
	private BiFunction<Path, ITargetSuggester, IDocumentImporter> importDetector;
	private boolean allowAutoCreateDocument = false;
	private boolean allowAutoScript = false;

	/**
	 * Start a Builder/Runner
	 *
	 * @param applicationVersion any String describing the current application. Will be used in document
	 * @param useDefaultSearchpaths always add searchpaths derived from spartan root to Library
	 */
	public LibJConfigRunner(String applicationVersion, boolean useDefaultSearchpaths) {
		super(applicationVersion, useDefaultSearchpaths);
	}

	/**
	 * Apply analyze, optimize, verify to builder
	 *
	 * @param builder builder
	 */
	public static void defaultLibraryConfig(Library.Builder builder) {
		builder.setVerify(true).setAnalyze(true).setOptimize(true);
	}

	/**
	 * Override cmd line arg buidlAll
	 *
	 * @param buildAll bool
	 * @return this
	 */
	public LibJConfigRunner setBuildAll(boolean buildAll) {
		this.buildAll = buildAll;
		return this;
	}

	/**
	 * Override cmd line arg importFile
	 *
	 * @param importFile file to import
	 * @return this
	 */
	public LibJConfigRunner setImportFile(Path importFile) {
		this.importFile = importFile;
		return this;
	}

	/**
	 * Override cmd line arg scriptFiles
	 *
	 * @param scriptFiles script to execute
	 * @return this
	 */
	public LibJConfigRunner setScriptFile(List<Path> scriptFiles) {
		this.scriptFiles = scriptFiles;
		return this;
	}

	/**
	 * Override cmd line arg update
	 *
	 * @param update bool
	 * @return this
	 */
	public LibJConfigRunner setUpdate(boolean update) {
		this.update = update;
		return this;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public LibJConfigRunner requiresConfigFile() {
		super.requiresConfigFile();
		return this;
	}

	@Override
	protected void maybeOpenFile() { // prevent opening if '--config-file' is meant to be target of import
		if (importFile == null) {
			super.maybeOpenFile();
		}
	}

	/**
	 * Code that will be executed as soon as a document and library have been loaded.
	 * Ignored if no document has been loaded by this runner
	 *
	 * @param runWithDoc Closure, excuted delayed. passed IDocumentHandler with document
	 * @return this
	 */
	public LibJConfigRunner executeWithLoadedDocument(Consumer<IDocumentHandler> runWithDoc) {
		documentLoadedListeners.add(runWithDoc);

		return this;
	}

	/**
	 * Let Runner utilize the '--target' flag. If specified and nothing else has been done (import / load),
	 * create a new document with targetConf
	 *
	 * @return this
	 */
	public LibJConfigRunner allowAutoCreateDocument() {
		allowAutoCreateDocument = true;

		return this;
	}

	/**
	 * Schedule a build (using {@link #buildDocument(IDocumentHandler)}), for when a document has been loaded.
	 * <p>
	 * If this should be conditional on '--update' use {@link #allowAutoBuild()}
	 *
	 * @return this
	 */
	public LibJConfigRunner autoBuildDocument() {
		return executeWithLoadedDocument(this::buildDocument);
	}

	/**
	 * If scriptFiles is set, execute automatically
	 *
	 * @return this
	 */
	public LibJConfigRunner allowAutoExecuteScript() {
		allowAutoScript = true;
		return this;
	}

	/**
	 * Enable '--import' flag. If '--import' is specified, use given detector to decide on format / importer
	 *
	 * @param importDetector logic that gets the prepared targetConf and the file to import. Needs to return a {@link IDocumentImporter} that can import the file
	 * @return this
	 */
	public LibJConfigRunner autoImport(BiFunction<Path, ITargetSuggester, IDocumentImporter> importDetector) {
		this.importDetector = importDetector;
		return this;
	}

	/**
	 * Overwrite '--target'
	 *
	 * @param targetConf valid targetConf targ[:dev]
	 * @return this
	 */
	public LibJConfigRunner setTargetConf(String targetConf) {
		this.targetConf = targetConf;

		return this;
	}

	/**
	 * Let Runner utilize the '--update' flag
	 *
	 * @return this
	 */
	public LibJConfigRunner allowAutoBuild() {
		if (update) {
			this
				.requiresConfigFile()
				.autoBuildDocument();
		}
		return this;
	}

	/**
	 * Turn on automatic script execution.
	 * if '--script' sets a script file, execute it when loaded
	 * @return this
	 */
	//	public LibJConfigRunner allowAutoScript() {
	//		allowAutoScript = true;
	//		documentLoadedListeners.add(docHandler -> {
	//			if (docHandler.hasPluginEngine() && scriptFiles != null) {
	//				docHandler.getPluginEngine().executeScript(scriptFiles);
	//			}
	//		});
	//		return this;
	//	}

	/**
	 * build a Project using {@link LibJConfigBuilder#build()} and run everything that has been configured
	 *
	 * @return project
	 */
	public LibJConfigProject buildAndRun() {
		LibJConfigProject project = build();

		IDocumentHandler documentHandler = project.getDocumentHandler();

		if (importFile != null) {
			IDocumentImporter importer = importDetector.apply(importFile, new TargetConf(targetConf == null? "3E-1600" : targetConf));
			documentHandler.importFile(importFile, importer);

			if (hasExplicitConfigFile()) {
				documentHandler.setSaveLocation(getExplicitConfigFile());
				documentHandler.save();
			}
		}

		if (!documentHandler.hasLoadedDocument() && allowAutoCreateDocument && targetConf != null) {
			logger.info("Auto Creating new Document with {}", targetConf);

			documentHandler.newDocument(targetConf);
		}

		if (documentHandler.hasLoadedDocument()) {
			if (allowAutoScript && scriptFiles != null ) {
				for (Path scriptFile : scriptFiles) {
					if (!Files.isRegularFile(scriptFile)) {
						throw new ParameterException("Script " + scriptFiles + " does not exist");
					}
					IPluginEngine pluginEngine = documentHandler.getPluginEngine();
					if (pluginEngine != null) {
						pluginEngine.getScriptManager()
								.filter(sm -> sm.supportsFile(scriptFile))
								.orElseThrow(() -> new ParameterException(String.format("Script %s not supported by current pluginEngine (%s)", scriptFiles, pluginEngine)))
								.executeScript(scriptFile.toUri());
					} else {
						logger.error("Not configured for script execution");
						throw new ParameterException("Not configured for script execution");
					}
				}
			}
			documentLoadedListeners.forEach(c -> c.accept(documentHandler));
		}

		return project;
	}

	/**
	 * build the given documentHandler, respect the '--build-all' setting
	 *
	 * @param documentHandler handler with loaded document
	 */
	public void buildDocument(IDocumentHandler documentHandler) {
		try {
			documentHandler.build(!buildAll);
		} catch (JConfigBuildValidationException e) {
			logger.error("Automatic Build aborted because of pre-build-validation failure!");
			throw new JConfigRunnerFailureException(e);
		}
	}

	public boolean isUpdate() {
		return update;
	}

}
