/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.ISelectionHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigBuilder;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigRunner;

import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.FileSystem;
import java.nio.file.FileSystems;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.Map;

/**
 * Load library only once for use in multiple tests
 */
public class TestingProjectProvider {

	private static TestingProjectProvider instance = null;

	private LibJConfigProject project;

	public static Path getSpartanRoot() {
		return LibJConfigBuilder.getInplaceSpartanRoot();
	}


	private static Map<URI, FileSystem> fileSystems = new HashMap<>();

	private static Path getResource(String resourceName) {
		try {
			URI resource = TestingProjectProvider.class.getResource(resourceName).toURI();
			if (resource.toString().startsWith("jar:")) {
				//Split into jar name and path in jar
				String[] split = resource.toString().split("!");
				URI zip = URI.create(split[0]);
				FileSystem fs = fileSystems.computeIfAbsent(zip, (x) -> {
					try {
						return FileSystems.newFileSystem(zip, new HashMap<>());
					} catch (IOException e) {
						e.printStackTrace();
						return null;
					}
				});
				return fs.getPath(split[1]);
			} else {
				return Paths.get(resource);
			}
		} catch (URISyntaxException e) {
			e.printStackTrace();
		}
		return null;
	}

	private synchronized LibJConfigProject getProject() {
		if (project == null) {
			LibJConfigBuilder builder = new LibJConfigBuilder("testing", true);
			builder
					.setSpartanRoot(getSpartanRoot())
					.configureLibraryBuilder(this::configureLibraryBuilder);
			project = builder.build();
		}
		return project;
	}

	protected void configureLibraryBuilder(Library.Builder it) {
		LibJConfigRunner.defaultLibraryConfig(it);
		it.addModuleDescriptionSearchpath(getResource("/addModuleDescriptions/"))
				.addBusDescriptionSearchpath(getResource("/addBusDescriptions/"));
	}

	public synchronized Library getLibrary() {
		return getProject().getApplication().getLibrary();
	}

	private IApplicationConfiguration application = null;

	public IApplicationConfiguration getApplication() {
		if (application == null) {
			application = new IApplicationConfiguration() {
				@Override
				public String getApplicationVersion() {
					return "testing";
				}

				@Override
				public Library getLibrary() {
					return TestingProjectProvider.this.getLibrary();
				}

				@Override
				public Path getSpartanRoot() {
					return TestingProjectProvider.getSpartanRoot();
				}

				@Override
				public void addLibraryReadyListener(ILibraryReadyListener listener) {
					//Library is already ready
				}

				@Override
				public boolean addOrCallReadyListener(ILibraryReadyListener listener) {
					listener.onLibraryReady(getLibrary());
					return true;
				}

				@Override
				public boolean isLibraryReady() {
					return true;
				}
			};
		}
		return application;
	}

	/**
	 * Create a new project, sharing the library with all other created projects
	 *
	 * @param projectDir
	 * @return
	 */
	public ILibJConfigProject newProject(Path projectDir) {
		return new LibJConfigProject(projectDir, getApplication(), false);

	}


	private static final String ISE_TARGET_NAME = "SP601";
	private static final String VIVADO_TARGET_NAME = "Nexys-Video";
	private static final String QUARTUS_TARGET_NAME = "DE0-Nano";

	public ILibJConfigProject newProjectWithTargetBoard(Path projectDir, String targetName) {
		ILibJConfigProject project = newProject(projectDir);
		project.getDocumentHandler().newDocument(targetName);
		return project;
	}

	public ILibJConfigProject newProjectWithIseToolchain(Path projectDir) {
		return newProjectWithTargetBoard(projectDir, ISE_TARGET_NAME);
	}

	public ILibJConfigProject newProjectWithVivadoToolchain(Path projectDir) {
		final ILibJConfigProject project = newProjectWithTargetBoard(projectDir, VIVADO_TARGET_NAME);
		final ISelectionHandler<DeviceDescription.Toolchain> toolchainSelectionHandler = project.getDocumentHandler().getDocument().getTarget().getToolchainSelectionHandler();
		final DeviceDescription.Toolchain vivado = toolchainSelectionHandler.getValidValues().stream().filter(tc -> tc.getToolchain().equals("vivado")).findAny().orElseThrow(RuntimeException::new);
		toolchainSelectionHandler.setValue(vivado);
		return project;
	}


	public ILibJConfigProject newProjectWithQuartusToolchain(Path projectDir) {
		return newProjectWithTargetBoard(projectDir, QUARTUS_TARGET_NAME);
	}

	public static TestingProjectProvider getInstance() {
		if (instance == null) {
			instance = new TestingProjectProvider();
		}
		return instance;
	}
}
