/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.tests

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection
import de.tu_darmstadt.rs.spartanmc.jconfig.api.exceptions.ConnectionMismatchException
import de.tu_darmstadt.rs.spartanmc.jconfig.testing.BaseLibJConfigSpec
import spock.lang.Narrative

import static de.tu_darmstadt.rs.spartanmc.jconfig.scripting.ModuleExtension.AUTO

@Narrative("Specifies behavior of connections between modules, targets and buses")
class ConnectionSpec extends BaseLibJConfigSpec {

	def "can connect ports"() {
		setup:
		def core = document.addSubsystem(CORE)
		def sysclk = document.addCommonModule(SYSCLK)

		def sysLocked = sysclk.ports["locked"]
		def coreLocked = core.ports["locked"]
		when:
		def con = document.connect(coreLocked, sysLocked)
		then:
		document.connections.size() == 1
		con.isValid(false)
		coreLocked.connected
		coreLocked.streamOppositeConnectors().allMatch({ it == sysLocked })
		sysLocked.connected
		sysLocked.streamOppositeConnectors().allMatch({ it == coreLocked })
	}

	def "refuses to connect ports with incompatible direction"() {
		setup:
		def core = document.addSubsystem(CORE)
		def test = document.addCommonModule(TEST_MOD)

		when:
		document.connect(core.ports["intr"], test.ports["in1"])
		then:
		thrown(ConnectionMismatchException)
	}

	def "refuses to connect to pin with incompatible direction"() {
		setup:
		def test = document.addCommonModule(TEST_MOD)

		when:
		document.connect(document.pins["GPIO_LED_0"], test.ports["in1"])
		then:
		thrown(ConnectionMismatchException)
	}

	def "connects port to pin with configurable direction"() {
		setup:
		def test = document.addCommonModule(TEST_MOD)

		def pin = document.pins["GPIO_HDR1"]
		when:
		def con = document.connect(pin, test.ports["in1"])
		then:
		con.isValid(false)
		pin.directionLocked
		pin.direction == ConnectorDirection.Output

		when:
		con.disconnect()
		then:
		!con.connected
		!pin.connected
		!pin.directionLocked
		!test.ports["in1"].connected

		when:
		con = document.connect(pin, test.ports["out1"])
		then:
		con.isValid(false)
		pin.directionLocked
		pin.direction == ConnectorDirection.Input
		pin.connectionCount == 1
		test.ports["out1"].streamOppositeConnectors().anyMatch({ it.is(pin) })
	}

	def "refuses to connect different size ports"() {
		setup:
		def core = document.addSubsystem(CORE)
		def test = document.addCommonModule(TEST_MOD)

		when:
		document.connect(core.ports["intr"], test.ports["out4"])
		then:
		thrown(ConnectionMismatchException)
	}

	def "refuses to connect different varsize ports"() {
		setup:
		def core = document.addSubsystem(CORE)
		def port_in = core.addPeripheral(PORT_IN)
		def test = document.addCommonModule(TEST_MOD)
		port_in.parameters["PORT_WIDTH"] = 3

		when:
		document.connect(test.ports["out4"], port_in.ports["pin"])
		then:
		thrown(ConnectionMismatchException)
	}

	def "connects matching varsize ports"() {
		setup:
		def core = document.addSubsystem(CORE)
		def port_in = core.addPeripheral(PORT_IN)
		def test = document.addCommonModule(TEST_MOD)
		port_in.parameters["PORT_WIDTH"] = 4

		when:
		def con = document.connect(test.ports["out4"], port_in.ports["pin"])
		then:
		notThrown(ConnectionMismatchException)
		con.isValid(false)
		test.ports["out4"].connectionCount == 1
		port_in.ports["pin"].connectionCount == 1
		!port_in.ports["pin"].additionalConnectionAllowed
	}

	def "partially connects port with pins (overlapping allowed)"() {
		setup:
		def core = document.addSubsystem(CORE)
		def port_out = core.addPeripheral(PORT_OUT)
		port_out.parameters["PORT_WIDTH"] = 4

		when:
		def con0 = document.connect(document.pins["GPIO_LED_0"], AUTO, port_out.ports["pin"], 0, AUTO)
		def con1 = document.connect(document.pins["GPIO_LED_1"], AUTO, port_out.ports["pin"], 1, AUTO)
		def con2 = document.connect(document.pins["GPIO_LED_2"], AUTO, port_out.ports["pin"], 2, AUTO)
		def con3 = document.connect(document.pins["GPIO_LED_3"], AUTO, port_out.ports["pin"], 3, AUTO)
		def con4 = document.connect(document.pins["GPIO_HDR0"], AUTO, port_out.ports["pin"], 3, AUTO)
		then:
		notThrown(ConnectionMismatchException)
		[con0, con1, con2, con3, con4].every { it.isValid(false) }
		port_out.ports["pin"].connectionCount == 5
		(0..3).every { document.pins["GPIO_LED_$it"].connectionCount == 1 }
		document.pins["GPIO_HDR0"].connectionCount == 1
	}

	def "partially connects port with pins (no overlapping)"() {
		setup:
		def core = document.addSubsystem(CORE)
		def port_in = core.addPeripheral(PORT_IN)
		port_in.parameters["PORT_WIDTH"] = 4

		when:
		def con0 = document.connect(document.pins["GPIO_SWITCH_0"], AUTO, port_in.ports["pin"], 0, AUTO)
		def con1 = document.connect(document.pins["GPIO_SWITCH_1"], AUTO, port_in.ports["pin"], 1, AUTO)
		def con2 = document.connect(document.pins["GPIO_SWITCH_2"], AUTO, port_in.ports["pin"], 2, AUTO)
		def con3 = document.connect(document.pins["GPIO_SWITCH_3"], AUTO, port_in.ports["pin"], 3, AUTO)
		then:
		notThrown(ConnectionMismatchException)
		[con0, con1, con2, con3].every { it.isValid(false) }
		port_in.ports["pin"].connectionCount == 4
		(0..3).every { document.pins["GPIO_SWITCH_$it"].connectionCount == 1 }
	}

	def "refuses to connect invalid partial connections"() {
		setup:
		def core = document.addSubsystem(CORE)
		def port_in = core.addPeripheral(PORT_IN)
		port_in.parameters["PORT_WIDTH"] = 4

		when:
		def con0 = document.connect(document.pins["GPIO_LED_0"], AUTO, port_in.ports["pin"], 0, AUTO) // wrong direction
		then:
		thrown(ConnectionMismatchException)

		when:
		con0 = document.connect(document.pins["GPIO_SWITCH_0"], AUTO, port_in.ports["pin"], 0, AUTO)
		def con1 = document.connect(document.pins["GPIO_SWITCH_1"], AUTO, port_in.ports["pin"], 0, AUTO) // overlapping
		then:
		thrown(IllegalArgumentException)
	}

	def "connects fully partial connection"() {
		setup:
		def test = document.addCommonModule(TEST_MOD)

		when:
		def con = document.connect(test.ports["out4"], 0, test.ports["in4"], 2, 2) // [X,X,0,0] => [0,0,X,X]
		def con1 = document.connect(test.ports["out4"], 2, document.pins["GPIO_LED_0"], AUTO, AUTO)
		def con2 = document.connect(test.ports["out4"], 3, document.pins["GPIO_LED_1"], AUTO, AUTO)
		def con3 = document.connect(test.ports["in4"], 0, document.pins["GPIO_SWITCH_0"], AUTO, AUTO)
		def con4 = document.connect(test.ports["in4"], 1, document.pins["GPIO_SWITCH_1"], AUTO, AUTO)
		then:
		notThrown(ConnectionMismatchException)
		[con, con1, con2, con3, con4].every { it.isValid(false) }
	}

	def "connects / disconnects buses"() {
		setup:
		def core = document.addSubsystem(CORE)
		def test = document.addCommonModule(TEST_MOD) // no autoconnect this way

		when:
		def con = document.connect(core.buses["peri-bus"], test.buses["peri-bus"])
		then:
		test.buses["peri-bus"].connectionCount == 1
		core.buses["peri-bus"].connectionCount == 1

		when:
		con.disconnect()
		then:
		test.buses["peri-bus"].connectionCount == 0
		core.buses["peri-bus"].connectionCount == 0
	}

	def "refuses to connect incompatible buses"() {
		setup:
		def core = document.addSubsystem(CORE)
		def test = document.addCommonModule(TEST_MOD) // no autoconnect this way

		when:
		def con = document.connect(core.buses["dma-bus"], test.buses["peri-bus"])
		then:
		thrown(ConnectionMismatchException)
		core.buses["dma-bus"].connectionCount == 0
		test.buses["peri-bus"].connectionCount == 0

		when:
		con = document.connect(core.buses["peri-bus"], test.buses["peri-master"])
		then:
		thrown(ConnectionMismatchException)
		test.buses["peri-master"].connectionCount == 0
		core.buses["peri-bus"].connectionCount == 0
	}

	def "refuses to connect mismatched parametrized buses"() {
		setup:
		def test0 = document.addCommonModule(TEST_MOD)
		def test1 = document.addCommonModule(TEST_MOD)

		when:
		def con = document.connect(test0.buses["param-master"], test1.buses["param-mismatched"])
		then:
		thrown(ConnectionMismatchException)
		test0.buses["param-master"].connectionCount == 0
		test1.buses["param-mismatched"].connectionCount == 0

		when:
		con = document.connect(test0.buses["param-master"], test1.buses["param-dynamic"])
		then:
		thrown(ConnectionMismatchException)
		test0.buses["param-master"].connectionCount == 0
		test1.buses["param-dynamic"].connectionCount == 0

		when:
		con = document.connect(test0.buses["param-master"], test1.buses["param-indirect"])
		then:
		thrown(ConnectionMismatchException)
		test0.buses["param-master"].connectionCount == 0
		test1.buses["param-indirect"].connectionCount == 0
	}

	def "connects matching dynamic parametrized buses"() {
		setup:
		def test0 = document.addCommonModule(TEST_MOD)
		def test1 = document.addCommonModule(TEST_MOD)
		test1.parameters["PARAM"] = 5

		when:
		def con = document.connect(test0.buses["param-master"], test1.buses["param-dynamic"])
		then:
		notThrown(ConnectionMismatchException)
		con.isValid(false)
		test0.buses["param-master"].connectionCount == 1
		test1.buses["param-dynamic"].connectionCount == 1
	}

	def "connects matching dynamic, indirect parametrized buses"() {
		setup:
		def test0 = document.addCommonModule(TEST_MOD)
		def test1 = document.addCommonModule(TEST_MOD)
		test1.parameters["PARAM"] = 5

		when:
		def con = document.connect(test0.buses["param-master"], test1.buses["param-indirect"])
		then:
		notThrown(ConnectionMismatchException)
		con.isValid(false)
		test0.buses["param-master"].connectionCount == 1
		test1.buses["param-indirect"].connectionCount == 1
	}

	def "bus-connection becomes invalid on param change"() {
		setup:
		def test0 = document.addCommonModule(TEST_MOD)
		def test1 = document.addCommonModule(TEST_MOD)
		test1.parameters["PARAM"] = 5

		when:
		def con = document.connect(test0.buses["param-master"], test1.buses["param-dynamic"])
		then:
		con.isValid(false)
		test0.buses["param-master"].connectionCount == 1
		test1.buses["param-dynamic"].connectionCount == 1

		when:
		test1.parameters["PARAM"] = 8
		then:
		!con.isValid(false)
		!test0.buses["param-master"].isValid(false)
		!test1.buses["param-dynamic"].isValid(false)
	}

	/*def "refuses to connect port with clkPort"() {
		setup:
		def core = document.addSubsystem(CORE)
		def sysclk = document.addCommonModule(SYSCLK)

		def intr_port = core.ports["intr"]// definitively not clkPort
		def clkfx_port = sysclk.ports["clkfx"]

		when:
		document.connect(intr_port, clkfx_port)
		then:
		thrown(ConnectionMismatchException)
	}*/

	def "connection between pin and clkPort works"() {
		setup:
		def core = document.addSubsystem(CORE)
		def sysclk = document.addCommonModule(SYSCLK)

		when:
		def con = document.connect(sysclk.ports["clk_gen"], document.pins["GPIO_SWITCH_0"])
		then:
		document.pins["GPIO_SWITCH_0"].connected
		sysclk.ports["clk_gen"].connected
		con.isValid(false)
	}

	def "connection between clkPorts work"() {
		setup:
		def core = document.addSubsystem(CORE)
		def sysclk = document.addCommonModule(SYSCLK)

		when:
		def con = document.connect(sysclk.ports["clkfx"], core.ports["clk"])
		then:
		core.ports["clk"].connected
		sysclk.ports["clkfx"].connected
		con.isValid(false)
	}
}
