/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.jconfig.tests

import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPluginEngine
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager.BaseBuildManager
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.buildManager.BuildResult
import de.tu_darmstadt.rs.spartanmc.jconfig.testing.BaseLibJConfigSpec
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibraryUtil
import spock.lang.Ignore

import java.nio.file.Files
import java.nio.file.Paths
/**
 * Created by Ramon on 03.06.2016.
 */
class PluginEngineSpec extends BaseLibJConfigSpec {

	def "can set plugin engine before document"() {
		def project = new LibJConfigProject(this.project.workingDir, this.project.application, false)
		def docHandler = project.documentHandler
		def plugin = Mock(IPluginEngine)

		when:
		docHandler.pluginEngine = null
		docHandler.pluginEngine = plugin
		docHandler.newDocument(TARGET_CONF)

		then:
		notThrown(Exception)
		docHandler.document.pluginEngine == plugin
	}

	def "can set plugin engine on existing document"() {
		def docHandler = project.documentHandler
		def test = document.addCommonModule(TEST_MOD)
		def plugin = Mock(IPluginEngine)

		when:
		docHandler.pluginEngine = null
		docHandler.pluginEngine = plugin
		then:
		notThrown(Exception)
		test.pluginEngine == plugin

		when:
		docHandler.pluginEngine = null

		then:
		notThrown(Exception)
		test.pluginEngine == null
	}

	def "plugin engine gets notified of hierarchy changes"() {
		def docHandler = project.documentHandler
		def plugin = Mock(IPluginEngine)
		docHandler.pluginEngine = plugin

		when:
		def core = document.addSubsystem(CORE)

		then:
		with(plugin) {
			1 * onModuleAdded({ it.moduleDescriptionName == CORE })
			0 * _
		}
		core.pluginEngine == plugin

		when:
		def uart = core.addPeripheral(UART)

		then:
		with(plugin) {
			1 * onModuleAdded({ it.moduleDescriptionName == UART })
			2 * onConnectorChanged(!null, !null, IConfigurationElement.IElementChangeListener.ConnectorChangeType.Added, !null)
			0 * _
		}

		when:
		document.removeModule(core)

		then:
		with(plugin) {
			1 * onModuleRemoved({ it == document }, { it.moduleDescriptionName == CORE })
			1 * onModuleRemoved({ it == core }, { it.moduleDescriptionName == UART })
			2 * onConnectorChanged(!null, !null, IConfigurationElement.IElementChangeListener.ConnectorChangeType.Removed, !null)
			0 * _
		}
	}

	def "plugin engine gets notifies of name changes"() {
		def docHandler = project.documentHandler
		def plugin = Mock(IPluginEngine)
		docHandler.pluginEngine = plugin
		def core = document.addSubsystem(CORE)

		when:
		document.nameHandler.storedValue = "changed_name"

		then:
		with(plugin) {
			1 * onElementNameChanged({ it.name == "changed_name" })
			0 * _
		}

		when:
		core.nameHandler.storedValue = "core0"

		then:
		with(plugin) {
			1 * onElementNameChanged({ it.name == "core0" })
			0 * _
		}
	}

	def "plugin engine gets notified of element changes"() {
		def docHandler = project.documentHandler
		def plugin = Mock(IPluginEngine)
		docHandler.pluginEngine = plugin
		def core = document.addSubsystem(CORE)
		def uart = core.addPeripheral(UART)

		when:
		core.parameters["HARDWARE_DEBUGGING"] = true

		then:
		1 * plugin.onParameterChanged(core, !null)

		when:
		document.connect(uart.ports['tx'], document.pins['GPIO_LED_0'])

		then:
		with(plugin) {
			1 * onConnectorChanged(uart, !null, IConfigurationElement.IElementChangeListener.ConnectorChangeType.Added, !null)
			1 * onConnectorChanged(document, !null, IConfigurationElement.IElementChangeListener.ConnectorChangeType.Added, !null)
		}

		when:
		uart.getAttributeHandler("Foo").setStoredValue("Bar")

		then:
		1 * plugin.onAttributeChanged(uart, "Foo")
	}

	@Ignore
	def "plugin engine gets notified of declaration changes"() {

	}

	def "plugin engine gets notified of lifecyle events"() {
		given: 'A DocumentHandler with pluginEngine set'
		def targ = lib.getTargetDescription("3E-1600")
		def docHandler = project.documentHandler
		def plugin = Mock(IPluginEngine)
		docHandler.pluginEngine = plugin
		def savePath = Paths.get('build/testOutput/test.jc3')
		Files.createDirectories(savePath.parent)

		def bManager = Stub(BaseBuildManager) {
			_ * buildAll(_) >> new BuildResult()
			_ * buildRequired(_) >> { IConfigurationDocument doc -> this.buildAll(doc) }
		}

		when: 'creating new document'
		def doc = docHandler.newDocument(targ, LibraryUtil.findCompatibleDevice(lib, targ))

		then: 'receive onCreateDocument in pluginEngine'
		1 * plugin.onCreatedDocument(_ as IConfigurationDocument)

		when: 'building document'
		docHandler.buildWith({ bManager }, false, false)

		then: 'receive onBuilding in pluginEngine'
		1 * plugin.onBuilding(doc)

		when: 'saving document'
		docHandler.setSaveLocation(savePath)
		docHandler.save()

		then: 'receive onSaving in pluginEngine'
		1 * plugin.onSaving(doc)

		when: 'opening document'
		docHandler.open(savePath)

		then: 'receive onOpenedDocument in pluginEngine'
		1 * plugin.onOpenedDocument(_ as IConfigurationDocument)
	}

	def "plugin engine can influence validation"() {
		def docHandler = project.documentHandler
		def plugin = Mock(IPluginEngine)
		docHandler.pluginEngine = plugin
		def core = document.addSubsystem(CORE)

		when:
		def valid = core.isValid(false)

		then:
		with(plugin) {
			1 * isValid(core, false) >> true
			0 * _
		}
		!valid

		when:
		def inUse = project.documentHandler.isModuleInUse(core)

		then:
		with(plugin) {
			1 * isModuleInUse(core) >> false
			0 * _
		}
		!inUse

		when:
		valid = document.isValid(true)

		then:
		with(plugin) {
			1 * isValid(document, true) >> false
			0 * _
		}
		!valid
	}
}
