/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package libsim;

import java.util.concurrent.Semaphore;

/**
 * This thread class implements threads that can be suspended and stopped safely.
 */
public class DThread extends Thread {

	private volatile boolean suspended = false;
	private Semaphore run_sem = new Semaphore(1);
	private Thread t = this;

	public DThread() {
		super();
	}

	public DThread(String name) {
		super(name);
	}

	public DThread(Runnable r) {
		super(r);
	}

	public void start() {
		super.start();
	}

	/**
	 * Stops this thread.
	 */
	public synchronized void _stop() {
		t = null;
		notify();
	}

	/**
	 * \brief Suspends this thread.
	 *
	 * If running, the thread is suspended after execution of _run() and
	 * _suspend_init().\par
	 * If waiting, this method does nothing.\par
	 * If new, the thread will NOT execute _run() after started via start()
	 * (but will execute _run_init() and _suspend_init()).
	 */
	public synchronized void _suspend() {
		suspended = true;
	}

	/**
	 * \brief Resumes this thread.
	 */
	public synchronized void _resume() {
		suspended = false;
		run_sem.tryAcquire();
		notify();
	}

	/**
	 * \brief Stub. Executed directly before entering wait state.
	 */
	protected void _suspend_init() {
	}

	/**
	 * \brief Executed once after started via start().
	 */
	protected void _run_init() {
	}

	/**
	 * \brief Stub. Should be overwritten to implement function.
	 */
	protected void _run() {
	}

	/**
	 * \brief Implements _stop() and _suspend() semantics. Do not
	 * overwrite! Use _run().
	 */
	public void run() {
		Thread c = Thread.currentThread();
		_run_init();
		run_sem.tryAcquire();
		while (c == t) {
			try {
				/* thread suspension */
				if (suspended) {
					synchronized (this) {
						_suspend_init();
						run_sem.release();
						while (suspended && c == t) {
							wait();
						}
						continue;
					}
				}
				_run();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Returns TRUE if the thread is (already or to be) suspended.
	 */
	public boolean suspended() {
		return suspended;
	}

	/**
	 * Blocks until the threads execution has suspended.
	 */
	public void wait_for() {
		if (Thread.currentThread() == this) {
			return;
		}
		try {
			run_sem.acquire();
			run_sem.release();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}
