/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package libsim;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Formatter;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * \class libsim.MessageType
 * \brief The message types enumeration. This class represents all possible messages
 * in the system.
 *
 * It provides the interface to deserialize and serialize messages
 * from / to a stream and to construct new messages from abstract data via
 * a Payload.
 *
 * @author Michael Raitza
 */
public enum MessageType {

    /* DEFINITION OF MESSAGE TYPES

       Each message type defines methods to construct an actual
       message from its type and user-supplied data.
       To propagate new message fields and how they are accessed,
       each message type defines a new interface extending Message
       and a class that implements this interface possibly extending
       other messages classes.
    */
	/**
	 * \brief STATE message. Signals a new simulation state.
	 *
	 * Carries a payload of one element of State (4 Byte).
	 */
	STATE(0) {
		final class M extends MsgC implements MessageState {

			private State state;

			private M(int seq, int dst, int src, MessageType t, State s) {
				super(seq, dst, src, t);
				state = s;
			}

			public State state() {
				return state;
			}

			public ByteBuffer serialize(int size) {
				ByteBuffer buf = super.serialize(4 + size);
				buf.putInt(state.state());
				return buf;
			}

			public void format_msg(Formatter f) {
				super.format_msg(f);
				f.format(" :: %s", state.toString());
			}
		}

		protected Message deserialize_payload(SocketChannel s, ByteBuffer[] msg) throws IOException, ClosedChannelException, UnknownMessage {
			try {
				if (!read_alloc(s, msg, 1, 4)) {
					return null;
				}
				msg[1].rewind();
				return new M(msg[0].getInt(), msg[0].getInt(), msg[0].getInt(), this, State.get(msg[1].getInt()));
			} catch (UnknownState e) {
				throw new UnknownMessage(e);
			}
		}

		protected Message deserialize_payload(ByteBuffer msg) {
			try {
				return new M(msg.getInt(), msg.getInt(), msg.getInt(), this, State.get(msg.getInt()));
			} catch (UnknownState e) {
				return null;
			}
		}

		public Message construct(Payload p) {
			Iterator it = p.iterator();
			return new M(p.seq(), p.dst(), p.src(), this, (State) it.next());
		}
	},
	/**
	 * \brief COND message. Signals a condition.
	 *
	 * Carries a payload of one element of Condition (4 Byte).
	 */
	COND(1) {
		final class M extends MsgC implements MessageCond {

			private Condition cond;

			private M(int seq, int dst, int src, MessageType t, Condition c) {
				super(seq, dst, src, t);
				cond = c;
			}

			public Condition cond() {
				return cond;
			}

			public ByteBuffer serialize(int size) {
				ByteBuffer buf = super.serialize(size + 4);
				buf.putInt(cond.type());
				return buf;
			}

			protected void format_msg(Formatter f) {
				super.format_msg(f);
				f.format(" :: %s", cond.toString());
			}
		}

		protected Message deserialize_payload(SocketChannel s, ByteBuffer[] msg) throws IOException, ClosedChannelException, UnknownMessage {
			try {
				if (!read_alloc(s, msg, 1, 4)) {
					return null;
				}
				msg[1].rewind();
				return new M(msg[0].getInt(), msg[0].getInt(), msg[0].getInt(), this, Condition.get(msg[1].getInt()));
			} catch (UnknownCondition e) {
				throw new UnknownMessage(e);
			}
		}

		protected Message deserialize_payload(ByteBuffer msg) {
			try {
				return new M(msg.getInt(), msg.getInt(), msg.getInt(), this, Condition.get(msg.getInt()));
			} catch (UnknownCondition e) {
				return null;
			}
		}

		public Message construct(Payload p) {
			Iterator it = p.iterator();
			return new M(p.seq(), p.dst(), p.src(), this, (Condition) it.next());
		}
	},
	/**
	 * \brief STEP message. Signals that one clock cycle happend in simulation time.
	 *
	 * Carries no payload.
	 */
	STEP(2),
	/**
	 * \brief R_MEM message. Requests a memory update.
	 *
	 * Carries no payload.
	 */
	R_MEM(3),
	/**
	 * \brief R_FIFO message. Requests a fifo update.
	 *
	 * Carries a payload of a length field (4 Byte) denoting the amount of requested bytes.
	 */
	R_FIFO(4) {
		final class M extends MsgC implements MessageRFifo {

			private int length;

			private M(int seq, int dst, int src, MessageType t, int length) {
				super(seq, dst, src, t);
				this.length = length;
			}

			private void length(int len) {
				length = len;
			}

			public int length() {
				return length;
			}

			public ByteBuffer serialize(int size) {
				ByteBuffer buf = super.serialize(4 + size);

				buf.putInt(length);
				return buf;
			}

			protected void format_msg(Formatter f) {
				super.format_msg(f);
				f.format(" :: %d", length);
			}
		}

		protected Message deserialize_payload(SocketChannel s, ByteBuffer[] msg) throws IOException, ClosedChannelException, UnknownMessage {
			if (!read_alloc(s, msg, 1, 4)) {
				return null;
			}
			msg[1].rewind();
			return new M(msg[0].getInt(), msg[0].getInt(), msg[0].getInt(), this, msg[1].getInt());
		}

		protected Message deserialize_payload(ByteBuffer msg) {
			return new M(msg.getInt(), msg.getInt(), msg.getInt(), this, msg.getInt());
		}

		public Message construct(Payload p) {
			int length = (Integer) p.iterator().next();
			return new M(p.seq(), p.dst(), p.src(), this, length);
		}
	},
	/**
	 * \brief U_MEM message. Signals a memory update.
	 *
	 * Carries a payload of a dirty flag (1 Byte) followed by a length field (4 Byte)
	 * and memory data in tuples (4 Byte Address, 4 Byte Data per tuple, at least one tuple).
	 */
	U_MEM(5) {
		final class M extends MsgC implements MessageUMem {

			private boolean dirty;
			private Collection<Tuple> data;

			private M(int seq, int dst, int src, MessageType t, boolean dirty) {
				super(seq, dst, src, t);
				this.dirty = dirty;
			}

			private void data(Collection<Tuple> d) {
				data = d;
			}

			public Collection<Tuple> data() {
				return data;
			}

			public int length() {
				return data.size();
			}

			public boolean dirty() {
				return dirty;
			}

			public ByteBuffer serialize(int size) {
				ByteBuffer buf = super.serialize(5 + (8 * data.size()) + size);

				buf.put(dirty? (byte) 1 : (byte) 0);
				buf.putInt(data.size());
				for (Tuple t : data) {
					buf.putInt(t.first);
					buf.putInt(t.second);
				}
				return buf;
			}

			protected void format_msg(Formatter f) {
				super.format_msg(f);
				f.format(" :: %s %d %s", dirty? "DIRTY" : "CLEAN", data.size(), data.toString());
			}
		}

		protected Message deserialize_payload(SocketChannel s, ByteBuffer[] msg) throws IOException, ClosedChannelException, UnknownMessage {
			if (!read_alloc(s, msg, 1, 5)) {
				return null;
			}
			msg[1].rewind();
			boolean dirty = msg[1].get() == 0? true : false;
			int length = msg[1].getInt();
			if (!read_alloc(s, msg, 2, length * 8)) {
				return null;
			}
			msg[2].rewind();

			M m = new M(msg[0].getInt(), msg[0].getInt(), msg[0].getInt(), this, dirty);
			Collection<Tuple> data = new ArrayList<Tuple>(length);
			for (int i = 0; i < length; i++)
				data.add(new Tuple(msg[2].getInt(), msg[2].getInt()));
			m.data(data);
			return m;
		}

		protected Message deserialize_payload(ByteBuffer msg) {
			M m = new M(msg.getInt(), msg.getInt(), msg.getInt(), this, true);
			boolean dirty = msg.get() == 0? true : false;
			m.dirty = dirty;
			int length = msg.getInt();
			Collection<Tuple> data = new ArrayList<Tuple>(length);
			for (int i = 0; i < length; i++)
				data.add(new Tuple(msg.getInt(), msg.getInt()));
			m.data(data);
			return m;
		}

		public Message construct(Payload p) {
			Iterator it = p.iterator();
			boolean dirty = (Boolean) it.next();
			Collection<Tuple> data = new ArrayList<Tuple>();
			while (it.hasNext()) {
				data.add((Tuple) it.next());
			}
			M m = new M(p.seq(), p.dst(), p.src(), this, dirty);
			m.data(data);
			return m;
		}
	},
	/**
	 * \brief U_FIFO message. Signals a FIFO update.
	 *
	 * Carries a payload of a length field (4 Byte) and a updated FIFO data (4 Byte per datum,
	 * at least one datum).
	 */
	U_FIFO(6) {
		final class M extends MsgC implements MessageUFifo {

			private Collection<Integer> data;

			private M(int seq, int dst, int src, MessageType t) {
				super(seq, dst, src, t);
			}

			public void data(Collection<Integer> d) {
				data = d;
			}

			public Collection<Integer> data() {
				return data;
			}

			public int length() {
				return data.size();
			}

			public ByteBuffer serialize(int size) {
				ByteBuffer buf = super.serialize(size + 4 + (4 * data.size()));

				buf.putInt(data.size());
				for (Integer i : data) {
					buf.putInt(i);
				}
				return buf;
			}

			protected void format_msg(Formatter f) {
				super.format_msg(f);
				f.format(" :: %d %s", data.size(), data.toString());
			}
		}

		protected Message deserialize_payload(SocketChannel s, ByteBuffer[] msg) throws IOException, ClosedChannelException, UnknownMessage {
			if (!read_alloc(s, msg, 1, 4)) {
				return null;
			}
			msg[1].rewind();
			int length = msg[1].getInt();
			if (!read_alloc(s, msg, 2, length * 4)) {
				return null;
			}
			msg[2].rewind();
			M m = new M(msg[0].getInt(), msg[0].getInt(), msg[0].getInt(), this);
			Collection<Integer> data = new ArrayList<Integer>(length);
			for (int i = 0; i < length; i++)
				data.add(msg[2].getInt());
			m.data(data);
			return m;
		}

		protected Message deserialize_payload(ByteBuffer msg) {
			M m = new M(msg.getInt(), msg.getInt(), msg.getInt(), this);
			int length = msg.getInt();
			Collection<Integer> data = new ArrayList<Integer>(length);
			for (int i = 0; i < length; i++)
				data.add(msg.getInt());
			m.data(data);
			return m;
		}

		@SuppressWarnings(value = "unchecked")
		public Message construct(Payload p) {
			Iterator it = p.iterator();
			Collection<Integer> data = new ArrayList<Integer>();
			while (it.hasNext()) {
				Object o = it.next();
				if (o instanceof Integer) {
					data.add((Integer) o);
				} else {
					try {
						data.addAll((Collection) o);
					} catch (ClassCastException e) {
						throw new ImplementationError("Illegal Argument to construct(). Collection<Integer> expected.");
					}
				}
			}
			M m = new M(p.seq(), p.dst(), p.src(), this);
			m.data(data);
			return m;
		}
	},
	/**
	 * \brief R_CFG message. Request a client configuration.
	 *
	 * Carries no payload.
	 */
	R_CFG(7),
	/**
	 * \brief U_CFG message. Sends a client configuration.
	 *
	 * Carries a payload of two sections.\par
	 * Section one: FIFO configuration. 1B length field
	 * followed by tuples of 10B length each.
	 * Each tuple consists of direction (1B), data width (1B),
	 * fifo depth (4B), base address (4B)\par
	 * Section two: SHM configuration. 1B length field
	 * follows by tuples of 4B length each.
	 * Each tuple consists of the base address (4B)\par
	 *
	 * Payload consists of the following elements:\par
	 * Section one: for each FIFO configuration an instance
	 * of FIFOConfig \par
	 * Section two: for each SHM configuration:
	 * Integer 'base address'
	 */
	U_CFG(8) {
		final class M extends MsgC implements MessageUCfg {

			private Collection<FIFOConfig> fifos = new LinkedList<FIFOConfig>();
			private Collection<Integer> shms = new LinkedList<Integer>();

			private M(int seq, int dst, int src, MessageType t) {
				super(seq, dst, src, t);
			}

			public void fifos(FIFOConfig fc) {
				fifos.add(fc);
			}

			public void shms(Integer i) {
				shms.add(i);
			}

			public Collection<FIFOConfig> fifos() {
				return fifos;
			}

			public Collection<Integer> shms() {
				return shms;
			}

			public ByteBuffer serialize(int size) {
				ByteBuffer buf = super.serialize(size + 2 + (10 * fifos.size()) + (4 * shms.size()));

				buf.put((byte) fifos.size());
				for (FIFOConfig fc : fifos) {
					buf.put(fc.output? (byte) 0 : (byte) 1);
					buf.put(fc.width);
					buf.putInt(fc.depth);
					buf.putInt(fc.base_address);
				}
				buf.put((byte) shms.size());
				for (Integer i : shms) {
					buf.putInt(i);
				}
				return buf;
			}

			protected void format_msg(Formatter f) {
				super.format_msg(f);
				f.format(" :: FIFOs %d; %s :: SHMs %d; [", fifos.size(), fifos.toString(), shms.size());
				boolean first = true;
				for (Integer i : shms)
					if (first) {
						f.format("0x%05x", i);
					} else {
						f.format(", 0x%05x", i);
					}

				f.format("]");
			}
		}

		protected Message deserialize_payload(SocketChannel s, ByteBuffer[] msg) throws IOException, ClosedChannelException, UnknownMessage {
			if (!read_alloc(s, msg, 1, 1)) {
				return null;
			}
			msg[1].rewind();
			int length = msg[1].get();
			if (!read_alloc(s, msg, 2, length * 10)) {
				return null;
			}
			msg[2].rewind();
			M m = new M(msg[0].getInt(), msg[0].getInt(), msg[0].getInt(), this);
			for (int i = 0; i < length; i++) {
				FIFOConfig fc = new FIFOConfig();
				fc.output = msg[2].get() == 0? true : false;
				fc.width = msg[2].get();
				fc.depth = msg[2].getInt();
				fc.base_address = msg[2].getInt();
				m.fifos(fc);
			}
			if (!read_alloc(s, msg, 3, 1)) {
				return null;
			}
			msg[3].rewind();
			length = msg[3].get();
			if (!read_alloc(s, msg, 4, 4 * length)) {
				return null;
			}
			msg[4].rewind();
			for (int i = 0; i < length; i++) {
				m.shms(msg[4].getInt());
			}
			return m;
		}

		protected Message deserialize_payload(ByteBuffer msg) {
			M m = new M(msg.getInt(), msg.getInt(), msg.getInt(), this);
			int length = msg.get();
			for (int i = 0; i < length; i++) {
				FIFOConfig fc = new FIFOConfig();
				fc.output = msg.get() == 0? true : false;
				fc.width = msg.get();
				fc.depth = msg.getInt();
				fc.base_address = msg.getInt();
				m.fifos(fc);
			}
			length = msg.get();
			for (int i = 0; i < length; i++) {
				m.shms(msg.getInt());
			}
			return m;
		}

		public Message construct(Payload p) {
			M m = new M(p.seq(), p.dst(), p.src(), this);
			Iterator it = p.iterator();
			Object o;
			boolean secOne = true;
			while (it.hasNext()) {
				o = it.next();
				if (o instanceof FIFOConfig && secOne) {
					m.fifos((FIFOConfig) o);
				} else {
					secOne = false;
					m.shms((Integer) o);
				}
			}
			return m;
		}
	};


	/* PRIMARY MESSAGE IMPLEMENTATION

	   Implements the body of a message, that consists of
	   three header fields.
	*/
	private static class MsgC implements Message {

		private Integer seq;
		private Integer src;
		private Integer dst;
		private MessageType type;

		private MsgC() {
		}

		public MsgC(Integer seq, Integer dst, Integer src, MessageType t) {
			this.seq = seq;
			this.dst = dst;
			this.src = src;
			type = t;
		}

		public Integer seq() {
			return seq;
		}

		public Integer src() {
			return src;
		}

		public Integer dst() {
			return dst;
		}

		public MessageType type() {
			return type;
		}

		public void seq(Integer s) {
			seq = s;
		}

		public void dst(Integer d) {
			dst = d;
		}

		public void src(Integer s) {
			src = s;
		}

		public ByteBuffer serialize(int size) {
			ByteBuffer buf = ByteBuffer.allocate(16 + size);
			buf.putInt(type.type());
			buf.putInt(seq);
			buf.putInt(dst);
			buf.putInt(src);
			return buf;
		}

		/**
		 * \brief Serialize a Message into stream S.
		 *
		 * @param s the output stream to serialize the Message into.
		 */
		public ByteBuffer serialize() {
			ByteBuffer buf = serialize(0);
			buf.flip();
			return buf;
		}

		public Message clone() {
			ByteBuffer buf = serialize(0);
			buf.flip();
			return MessageType.deserialize(buf);
		}

		protected void format_msg(Formatter f) {
			f.format("%d: %s - dst(0x%05x), src(0x%05x)", seq, type, dst, 0x3ffff & src);
		}

		public String format_msg() {
			Formatter f = new Formatter();
			format_msg(f);
			return f.toString();
		}
	}


	private int type;
	private static HashMap<Integer, MessageType> msg_map = null;

	private MessageType(int type) {
		this.type = type;
	}

	private static void init() {
		if (msg_map == null) {
			msg_map = new HashMap<Integer, MessageType>();
			for (MessageType m : MessageType.values()) {
				if (msg_map.put(m.type, m) != null) {
					throw new IllegalArgumentException("Duplicate elements in numeric type map");
				}
			}
		}
	}

	/* read SIZE bytes from socket channel S into the buffer msg[] at the array position BUF */
	private static boolean read_alloc(SocketChannel s, ByteBuffer[] msg, int buf, int size) throws IOException {
		if (msg[buf] == null) {
			msg[buf] = ByteBuffer.allocate(size);
		}
		if (msg[buf].remaining() == 0) {
			return true;
		}
		if (s == null) {
			throw new IOException("SocketChannel inaccessible");
		}
		if (s.read(msg[buf]) == -1) {
			throw new ClosedChannelException();
		}
		return msg[buf].remaining() == 0;
	}

	/**
	 * By default the payload is empty.
	 */
	protected Message deserialize_payload(SocketChannel s, ByteBuffer[] msg) throws IOException, ClosedChannelException, UnknownMessage {
		return construct(new Payload(msg[0].getInt(), msg[0].getInt(), msg[0].getInt()));
	}

	protected Message deserialize_payload(ByteBuffer msg) {
		return construct(new Payload(msg.getInt(), msg.getInt(), msg.getInt()));
	}

	/**
	 * \brief Constructs a message from its payload.
	 */
	public Message construct(Payload p) {
		return new MsgC(p.seq(), p.dst(), p.src(), this);
	}

	/**
	 * \brief Convenience method. Constructs a message without payload.
	 */
	public Message construct(int seq, int dst, int src) {
		return new MsgC(seq, dst, src, this);
	}

	/**
	 * \brief Deserialize a Message from stream S.
	 *
	 * The message header is read from stream S and a Message instantiated.
	 * The Message's specialized deserializer takes care of the payload (if any)
	 * beyond the header.
	 * The header's format is as follows:\par
	 * 4 Byte Type | 4 Byte Sequence | 4 Byte Destination | 4 Byte Source\par
	 * All data is in network byte order.
	 *
	 * @param s an input stream object to read the message from
	 * @return a Message of the identified type including all its data
	 *
	 * @throws UnknownMessage if the message type is unknown. Can also be caused by
	 * an UnknownCondition or UnknownState exception.
	 * @throws ClosedChannelException
	 * @throws IOException
	 */
	public static Message deserialize(SocketChannel s, ByteBuffer[] msg) throws ClosedChannelException, IOException, UnknownMessage {
		init();
		/* read the header */
		if (!read_alloc(s, msg, 0, 16)) {
			return null;
		}
		/* re-instantiate the messge */
		msg[0].rewind();
		int type = msg[0].getInt();
		MessageType m = msg_map.get(type);
		if (m == null) {
			throw new UnknownMessage(type);
		}
		return m.deserialize_payload(s, msg);
	}

	private static Message deserialize(ByteBuffer msg) {
		int type = msg.getInt();
		MessageType m = msg_map.get(type);
		return m.deserialize_payload(msg);
	}

	/**
	 * \brief Returns the numeric type of this message type.
	 */
	public int type() {
		return type;
	}

	public static ByteBuffer[] allocate_msg_buf() {
		return new ByteBuffer[5];
	}
}
