/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package libsim;

import java.util.LinkedList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

public class TimedExec {

	private class DTimerTask extends TimerTask {

		private Runnable runit;

		public DTimerTask(Runnable r) {
			runit = r;
		}

		public void run() {
			runit.run();
		}
	}


	private class DTimerThread extends DThread {

		private Runnable runit;

		public DTimerThread(Runnable r) {
			runit = r;
		}

		public void _run() {
			runit.run();
		}
	}


	private Timer tt = new Timer("Stepping Timer");
	private List<Runnable> tasks = new LinkedList<Runnable>();
	private List<Object> running_tasks = new LinkedList<Object>();
	private long delay = 0;

	/**
	 * Stops the timed execution.
	 */
	public void stop_timer() {
		if (delay < 0) {
			for (Object td : running_tasks)
				((DThread) td)._stop();
			running_tasks.clear();
		} else {
			for (Object td : running_tasks)
				((TimerTask) td).cancel();
			running_tasks.clear();
			tt.purge();
		}
	}

	/**
	 * Resumes the timed execution.
	 */
	public void resume_timer() {
		for (Runnable trtk : tasks) {
			if (delay < 0) {
				DThread td = new DTimerThread(trtk);
				running_tasks.add(td);
				td._resume();
				td.start();
			} else {
				TimerTask td = new DTimerTask(trtk);
				tt.scheduleAtFixedRate(td, 0, delay);
				running_tasks.add(td);
			}
		}
	}

	/**
	 * Executes all registered tasks once.
	 */
	public void step_timer() {
		stop_timer();
		for (Runnable trtk : tasks) {
			TimerTask td = new DTimerTask(trtk);
			tt.schedule(td, 0);
		}
	}

	/**
	 * \brief Sets the delay to DELAY and restarts the timer if necessary.
	 */
	public void delay(long delay) {
		if (running_tasks.size() > 0) {
			stop_timer();
			this.delay = delay;
			resume_timer();
		} else {
			this.delay = delay;
		}
	}

	/**
	 * Registers a task that follows the timed execution model of the dispatcher.
	 */
	public void register_task(Runnable task) {
		tasks.add(task);
	}

	/**
	 * Registers a task at this timer.
	 */
	public void schedule(TimerTask task, long delay) {
		tt.schedule(task, delay);
	}

	/**
	 * Terminates the timer thread.
	 */
	public void dissemble() {
		tt.cancel();
		tt = null;
	}
}
