/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package libsim.test;

import libsim.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import org.junit.*;

public class Tests {
    private Message msg_simple;
    private Message msg_complex;
    private String outcome_simple;
    private String outcome_complex;
    private byte[] bytestream_simple;
    private byte[] bytestream_complex;
    private ByteBuffer bytebuffer_simple;
    private ByteBuffer bytebuffer_complex;

    @Before public void setup_messages() {
        msg_simple = MessageType.STEP.construct(new Payload(0, -1, 1));
        msg_complex = MessageType.STATE.construct(new Payload(4, 5, 6) {{ add(State.RUNNING); }});
        outcome_simple = "0: STEP - dst(0xffffffff), src(0x00001)";
        outcome_complex = "4: STATE - dst(0x00005), src(0x00006) :: RUNNING";
        bytestream_simple = new byte[] { 00, 00, 00, 02,
                                         00, 00, 00, 00,
                                         (byte)0xff, (byte)0xff, (byte)0xff, (byte)0xff,
                                         00, 00, 00, 01 };
        bytebuffer_simple = ByteBuffer.allocate(16);
        bytebuffer_simple.put(bytestream_simple);
        bytestream_complex = new byte[] { 00, 00, 00, 00,
                                          00, 00, 00, 04,
                                          00, 00, 00, 05,
                                          00, 00, 00, 06,
                                          00, 00, 00, 00 };
        bytebuffer_complex = ByteBuffer.allocate(4);
        bytebuffer_complex.put(bytestream_complex, 16, 4);
    }

    /**
     * \brief Construction dispatcher.test. Builds two messages and compares their string
     * represenations to the precomputed representations for equality.
     */
    @Test public void construct_simple() {
        Assert.assertTrue(msg_simple.format_msg().equals(outcome_simple));
		Assert.assertTrue(msg_complex.format_msg().equals(outcome_complex));
    }

    /**
     * \brief Deserialisation dispatcher.test. Builds a new message from a predefined
     * byte stream and compares its formatted string representation to the
     * expected outcome for equality.
     */
    @Test public void deserialize() throws UnknownMessage,IOException,ClosedChannelException {
        ByteBuffer[] bufs = new ByteBuffer[2];
        bufs[0] = ByteBuffer.allocate(16);
        bufs[0].put(bytestream_complex, 0, 16);
        bufs[1] = bytebuffer_complex;
        msg_simple = MessageType.deserialize(null, bufs);
        Assert.assertTrue(msg_simple.format_msg().equals(outcome_complex));
    }

    /**
     * \brief Serialisation dispatcher.test. Serialises a STATE(START) message and compares
     * the output byte stream to the precomputed bytestream for equality.
     */
    @Test public void serialize_simple() throws IOException {
        ByteBuffer buf = msg_simple.serialize();
        byte[] outcome = buf.array();
        for (int i = 0; i < bytestream_simple.length; i++)
            Assert.assertTrue(bytestream_simple[i] == outcome[i]);
    }

    @Test public void serialize_complex() throws IOException {
        ByteBuffer buf = msg_complex.serialize();
        byte[] outcome = buf.array();
        for (int i = 0; i < bytestream_complex.length; i++)
            Assert.assertTrue(bytestream_complex[i] == outcome[i]);
    }

    @Test public void serialize_ucfg() throws IOException {
        byte[] outcome = new byte[] { 00, 00, 00, 8,
                                      00, 00, 00, 02,
                                      00, 00, 00, 00,
                                      -0x01, -0x01, -0x01, -0x01,
                                      02,
                                      00, 18, 00, 00, 00, 16, 00, 01, -0x60, 00,
                                      01, 18, 00, 00, 00, 16, 00, 01, -0x60, 0x10,
                                      01,
                                      00, 01, -0x60, 0x20 };
        ByteBuffer buf = MessageType.U_CFG.construct(new Payload(2, 0, -1) { {
            add(new FIFOConfig(true, (byte)18, 16, 0x1a000));
            add(new FIFOConfig(false, (byte)18, 16, 0x1a010));
            add(new Integer(0x1a020));}}).serialize();
        for (int i = 0; i < outcome.length; i++)
            Assert.assertTrue(outcome[i] == buf.get());
        /*
        System.out.printf("%n");
        for (int i = 0; i < outcome.length; i++)
            System.out.printf("%02x ", outcome[i]);
        System.out.printf("%n");

        while(buf.remaining() > 0)
            System.out.printf("%02x ", buf.get());
        System.out.printf("%n");
        */
    }
}
