/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.process;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.Component;
import java.awt.Desktop;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.file.Path;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;
import javax.swing.JOptionPane;

public final class ProcessUtil {

	private static final Logger logger = LoggerFactory.getLogger(ProcessUtil.class);

	private ProcessUtil() {
	}

	/**
	 * Run a process, redirecting its output and wait for it to finish.
	 *
	 * Allow separate redirect for stdout and stderr
	 *
	 * @param builder The ProcessBuilder to run
	 * @param stdOutConsumers StdOut consumers
	 * @param stdErrConsumers StdErr consumers
	 * @param timeout how long to wait for timeout
	 * @param unit the time unit of the timeout
	 * @param outputStreamFiller a function to fill the output stream with data. If large data is to be
	 * 				transmitted, the function should start a new thread, otherwise it will block. If it is null,
	 * 				the OutputStream will be closed immediately.
	 * @return the process's exit code
	 * @throws IOException if the process threw an exception
	 */
	public static int runAndWaitForSeparate(ProcessBuilder builder, Consumer<String> stdOutConsumers, Consumer<String> stdErrConsumers, Long timeout, TimeUnit unit, Consumer<? super OutputStream> outputStreamFiller) throws IOException {
		Process proc = builder.start();

		if (outputStreamFiller != null) {
			outputStreamFiller.accept(proc.getOutputStream());
		} else {
			//no stdin
			proc.getOutputStream().close();
		}

		StreamGobbler errorGobbler = new StreamGobbler(proc.getErrorStream(), stdErrConsumers);
		StreamGobbler outputGobbler = new StreamGobbler(proc.getInputStream(), stdOutConsumers);

		errorGobbler.start();
		outputGobbler.start();

		try {
			if (timeout != null && unit != null) {
				if (!proc.waitFor(timeout, unit)) {
					//Kill if not in time
					proc.destroy();
					throw new IOException("Process did not finish in time.");
				}
			}

			return proc.waitFor();
		} catch (InterruptedException e) {
			throw new IOException(e);
		} finally {

			try {
				errorGobbler.join();
				outputGobbler.join();
			} catch (InterruptedException e) {
				// Ignore exceptions :)
			}

		}

	}

	/**
	 * Run a process, redirecting its output and wait for it to finish.
	 *
	 * Allow separate redirects for stdout and stderr
	 *
	 * @param builder The ProcessBuilder to run
	 * @param stdOutConsumers StdOut consumers
	 * @param stdErrConsumers StdErr consumers
	 * @param outputStreamFiller a function to fill the output stream with data. If large data is to be
	 * 				transmitted, the function should start a new thread, otherwise it will block. If it is null,
	 * 				the OutputStream will be closed immediately.
	 * @return the process's exit code
	 * @throws IOException if the process threw an exception
	 */
	public static int runAndWaitForSeparate(ProcessBuilder builder, Consumer<String> stdOutConsumers, Consumer<String> stdErrConsumers, Consumer<? super OutputStream> outputStreamFiller) throws IOException {
		return runAndWaitForSeparate(builder, stdOutConsumers, stdErrConsumers, null, null, outputStreamFiller);
	}

	/**
	 * Run a process, redirecting its output and wait for it to finish.
	 *
	 * Allow separate redirects for stdout and stderr
	 *
	 * @param builder The ProcessBuilder to run
	 * @param stdOutConsumers StdOut consumers
	 * @param stdErrConsumers StdErr consumers
	 * @return the process's exit code
	 * @throws IOException if the process threw an exception
	 */
	public static int runAndWaitForSeparate(ProcessBuilder builder, Consumer<String> stdOutConsumers, Consumer<String> stdErrConsumers) throws IOException {
		return runAndWaitForSeparate(builder, stdOutConsumers, stdErrConsumers, null, null, null);
	}

	/**
	 * Run a process, redirecting its output and wait for it to finish
	 * @param builder The ProcessBuilder to run
	 * @param timeout how long to wait for timeout
	 * @param unit the time unit of the timeout
	 * @param outputStreamFiller a function to fill the output stream with data. If large data is to be
	 * 				transmitted, the function should start a new thread, otherwise it will block. If it is null,
	 * 				the OutputStream will be closed immediately.
	 * @param outputConsumers Output consumers
	 * @return the process's exit code
	 * @throws IOException if the process threw an exception
	 */
	@SafeVarargs
	public static int runAndWaitFor(ProcessBuilder builder, Consumer<? super OutputStream> outputStreamFiller, Long timeout, TimeUnit unit, Consumer<String>... outputConsumers) throws IOException {
		//Optimize away array if there is exactly one element
		if (outputConsumers.length == 1) {
			return runAndWaitForSeparate(builder, outputConsumers[0], outputConsumers[0], outputStreamFiller);
		}
		Consumer<String> c = StreamGobbler.outputToMultiple(outputConsumers);
		return runAndWaitForSeparate(builder, c, c, timeout, unit, outputStreamFiller);
	}

	/**
	 * Run a process, redirecting its output and wait for it to finish
	 * @param builder The ProcessBuilder to run
	 * @param outputStreamFiller a function to fill the output stream with data. If large data is to be
	 * 				transmitted, the function should start a new thread, otherwise it will block. If it is null,
	 * 				the OutputStream will be closed immediately.
	 * @param outputConsumers Output consumers
	 * @return the process's exit code
	 * @throws IOException if the process threw an exception
	 */
	@SafeVarargs
	public static int runAndWaitFor(ProcessBuilder builder, Consumer<? super OutputStream> outputStreamFiller, Consumer<String>... outputConsumers) throws IOException {
		return runAndWaitFor(builder, outputStreamFiller, null, null, outputConsumers);
	}

	/**
	 * Run a process, redirecting its output and wait for it to finish
	 * @param builder The ProcessBuilder to run
	 * @param outputConsumers Output consumers
	 * @return the process's exit code
	 * @throws IOException if the process threw an exception
	 */
	@SafeVarargs
	public static int runAndWaitFor(ProcessBuilder builder, Consumer<String>... outputConsumers) throws IOException {
		return runAndWaitFor(builder, null, outputConsumers);
	}

	/**
	 * Run a process, redirecting stdout to stdout and stderr to stderr. And wait for it to finish
	 * @param builder The ProcessBuilder to run
	 * @param outputStreamFiller a function to fill the output stream with data. If large data is to be
	 * 				transmitted, the function should start a new thread, otherwise it will block. If it is null,
	 * 				the OutputStream will be closed immediately.
	 * @return the process's exit code
	 * @throws IOException if the process threw an exception
	 */
	@SuppressWarnings("checkstyle:SystemOut")
	public static int runAndWaitForDefaultRedirect(ProcessBuilder builder, Consumer<? super OutputStream> outputStreamFiller) throws IOException {
		return runAndWaitForSeparate(builder, System.out::println, System.err::println, outputStreamFiller);
	}

	/**
	 * Run a process, redirecting stdout to stdout and stderr to stderr. And wait for it to finish
	 * @param builder The ProcessBuilder to run
	 * @return the process's exit code
	 * @throws IOException if the process threw an exception
	 */
	@SuppressWarnings("checkstyle:SystemOut")
	public static int runAndWaitForDefaultRedirect(ProcessBuilder builder) throws IOException {
		return runAndWaitForSeparate(builder, System.out::println, System.err::println, null);
	}

	/**
	 * Open a file with its default program. Errors are shown in a JOptionPane
	 * @param file the file to show
	 * @param parentComponent parent for the JOptionPane
	 */
	public static void openWithDefaultProgram(Path file, Component parentComponent) {
		final String errorTitle = "Error";
		if (Desktop.isDesktopSupported()) {
			try {
				Desktop.getDesktop().open(file.toFile());
			} catch (IOException ex) {
				String msg = "A default program was not registered. Can not open the file: " + file;
				logger.error(msg);
				JOptionPane.showMessageDialog(parentComponent, msg, errorTitle, JOptionPane.ERROR_MESSAGE);
			}
		} else {
			String msg = "The awt.Desktop class is not supported on the current platform.";
			logger.error(msg);
			JOptionPane.showMessageDialog(parentComponent, msg, errorTitle, JOptionPane.ERROR_MESSAGE);
		}
	}
}
