/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.process;

import java.util.Iterator;
import java.util.Spliterator;
import java.util.Spliterators;
import java.util.function.Consumer;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

/**
 * A RingBuffer that buffers the last messages saved into it.
 *
 * The number of items is configurable
 */
public class RingBuffer<T> implements Consumer<T>, Iterable<T> {

	private Object[] storage;

	private int pos;
	private boolean full;

	/**
	 * Create a RingBuffer
	 *
	 * @param size maximum number of elements
	 */
	public RingBuffer(int size) {
		storage = new Object[size];
		pos = 0;
		full = false;
	}

	@Override
	public Iterator<T> iterator() {
		int oldestItem;
		if (!full) {
			oldestItem = 0;
		} else {
			oldestItem = pos;
		}
		return new Iterator<T>() {
			int iteratorPos = oldestItem;
			boolean first = true;

			@Override
			public boolean hasNext() {
				if (full) {
					return iteratorPos != oldestItem || first;
				} else {
					return iteratorPos < pos;
				}
			}

			@Override
			public T next() {
				T t = (T) storage[iteratorPos];
				iteratorPos++;
				if (iteratorPos >= storage.length) {
					iteratorPos = 0;
				}
				first = false;
				return t;
			}
		};
	}

	@Override
	public void accept(T t) {
		storage[pos] = t;
		pos++;
		if (pos >= storage.length) {
			full = true;
			pos = 0;
		}
	}

	private Spliterator<T> spliteratorIfInBounds(int fromIndex, int toIndex) {
		if (fromIndex <= toIndex) {
			return Spliterators.spliterator(storage, fromIndex, toIndex, 0);
		} else {
			return Spliterators.emptySpliterator();
		}
	}

	public Stream<T> stream() {
		if (!full) {
			final Spliterator<T> spliterator = Spliterators.spliterator(storage, 0, pos, 0);
			return StreamSupport.stream(spliterator, false);
		}
		final Spliterator<T> older = spliteratorIfInBounds(pos, storage.length);
		final Spliterator<T> newer = spliteratorIfInBounds(0, pos);

		final Stream<T> olderStream = StreamSupport.stream(older, false);
		final Stream<T> newerStream = StreamSupport.stream(newer, false);
		return Stream.concat(olderStream, newerStream);
	}
}
