/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.process;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.function.Consumer;


public class StreamGobbler extends Thread {

	public static final String ANSI_RESET = "\u001B[0m";
	public static final String ANSI_CLEAR_LINE = "\u001B[2K";
	public static final int LINE_LENGTH = 80;
	public static final String ANSI_COLOR_RED = "\u001B[31;1m";


	private final InputStream is;
	private final Consumer<String> consumer;


	/**
	 * Create StreamGobbler
	 *
	 * @param is The input to read
	 * @param consumer Consumer that gets passed every line
	 */
	public StreamGobbler(InputStream is, Consumer<String> consumer) {
		super("Stream Gobbler for "+is.toString());
		this.is = is;
		this.consumer = consumer;
	}

	/**
	 * Run the gobbker
	 */
	public void run() {
		try {
			InputStreamReader isr = new InputStreamReader(is);
			BufferedReader br = new BufferedReader(isr);
			String line;
			while ((line = br.readLine()) != null) {
				consumer.accept(line);
			}
		} catch (IOException ioe) {
			ioe.printStackTrace();
		}
	}

	/**
	 * Prepend each line with a specified string
	 *
	 * @param prepend the string to prepend
	 * @param downstream The downstream Consimer
	 * @return A consumer that prepends each line before passing it to the downstream consumer
	 */
	public static Consumer<String> prepend(String prepend, Consumer<String> downstream) {
		return (s) -> downstream.accept(prepend + s);
	}

	/**
	 * Add ANSI Clear line and reset codes to be able to replace lines
	 *
	 * Make sure that the downstream consumer only uses print, not println!
	 *
	 * @param downstream downstream consumer
	 * @return a consumer
	 */
	public static Consumer<String> shortenToOneLine(Consumer<String> downstream) {
		return (line) -> {
			if (!line.isEmpty()) {
				if (line.length() > LINE_LENGTH) {
					line = line.substring(0, LINE_LENGTH);
				}
				downstream.accept("\r" + ANSI_CLEAR_LINE + ANSI_RESET + line);
			}
		};
	}

	@SuppressWarnings("checkstyle:SystemOut")
	public static Consumer<String> shortenToOneLineToStdOut() {
		return shortenToOneLine(System.out::print);
	}

	/**
	 * Create a single consumer from multiple ones. Every consumer is invoked when the result consumer is invoked.
	 *
	 * @param consumers the consumers
	 * @return a consumer that calls all others when invoked
	 */
	@SafeVarargs
	public static Consumer<String> outputToMultiple(Consumer<String>... consumers) {
		return (s) -> {
			for (Consumer<String> consumer : consumers) {
				consumer.accept(s);
			}
		};
	}

	/**
	 * Clear the current line after having used shortenToOneLine
	 *
	 * @param downstream The downstream consumer
	 */
	public static void clearLine(Consumer<String> downstream) {
		downstream.accept("\r" + ANSI_CLEAR_LINE + ANSI_RESET);
	}


	/**
	 * Clear the current line on stdout after having used shortenToOneLine
	 */
	public static void clearLineStdout() {
		clearLine(System.out::print);
	}
}
