/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonIgnore;
import com.fasterxml.jackson.annotation.JsonProperty;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.InputStream;
import java.io.Serializable;
import java.util.Date;
import java.util.Properties;

/**
 * @author ramonw
 */
@JsonAutoDetect(getterVisibility = JsonAutoDetect.Visibility.NONE, setterVisibility = JsonAutoDetect.Visibility.NONE, fieldVisibility = JsonAutoDetect.Visibility.NONE)
public final class BuildInfo implements Serializable {

	private static final Logger logger = LoggerFactory.getLogger(BuildInfo.class);

	private static final String FILE_PATH = "META-INF/buildInfo.properties";

	private final Class<?> clazz;
	private final String revision;
	private final String buildTime;

	@JsonCreator
	private BuildInfo(@JsonProperty("revision") String revision, @JsonProperty("buildTime") String buildTime) {
		this.revision = revision;
		this.buildTime = buildTime;
		this.clazz = null;
	}

	/**
	 * Create BuildInfo
	 * @param clazz Class to get BuildInfo for
	 */
	public BuildInfo(Class<?> clazz) {
		this.clazz = clazz;

		Properties props = getBuildInfoProps();

		String revision = props.getProperty("git.commit.id.revision");
		if (revision == null) {
			revision = "SNAPSHOT";
		}

		String buildTime = props.getProperty("git.build.time");
		if (buildTime == null) {
			buildTime = new Date().toString();
		}

		this.revision = revision;
		this.buildTime = buildTime;
	}

	@JsonIgnore
	public boolean isCurrent() {
		return clazz != null;
	}

	private Properties getBuildInfoProps() {
		Properties properties = new Properties();
		try (InputStream str = clazz.getClassLoader().getResourceAsStream(FILE_PATH)) {
			properties.load(str);

			//			Object branch = properties.get("git.commit.branch");
			//			this.branch = (branch == null)? "" : branch.toString();
		} catch (Exception e) {
			logger.error("Failed to load {} for {}", FILE_PATH, clazz);
		}
		return properties;
	}

	@JsonProperty
	public String getRevision() {
		return revision;
	}

	@JsonProperty
	public String getBuildTime() {
		return buildTime;
	}

	@Override
	public int hashCode() {
		int result = revision != null? revision.hashCode() : 0;
		result = 31 * result + (buildTime != null? buildTime.hashCode() : 0);
		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		BuildInfo buildInfo = (BuildInfo) o;

		if (revision != null? !revision.equals(buildInfo.revision) : buildInfo.revision != null) {
			return false;
		}
		return buildTime != null? buildTime.equals(buildInfo.buildTime) : buildInfo.buildTime == null;

	}

	@Override
	public String toString() {
		return "BuildInfo{" +
			"clazz=" + clazz +
			", revision='" + revision + '\'' +
			", buildTime='" + buildTime + '\'' +
			'}';
	}
}
