/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Created by Ramon on 19.05.2016.
 */
public final class CollectionUtil {

	private CollectionUtil() {
	}

	/**
	 * Get the differences between two collections
	 * @param a first collection
	 * @param b second collection
	 * @param <V> element type
	 * @return list of differences, encoded as pairs
	 */
	public static <V> List<KeyedPair<Void, V>> debugEquality(Collection<? extends V> a, Collection<? extends V> b) {
		Set<V> bCopy = new HashSet<>(b);
		List<KeyedPair<Void, V>> result = new ArrayList<>();
		for (V aVal : a) {
			if (b.contains(aVal)) {
				bCopy.remove(aVal);
			} else {
				result.add(new KeyedPair<>(null, aVal, null));
			}
		}

		for (V bVal : b) {
			result.add(new KeyedPair<>(null, null, bVal));
		}

		return result;
	}

	/**
	 * Get the differences between two maps
	 * @param a first collection
	 * @param b second collection
	 * @param <K> key type
	 * @param <V> element type
	 * @return list of differences, encoded as KeyedPair (key-double-value-pairs)
	 */
	public static <K, V> List<KeyedPair<K, V>> debugEquality(Map<K, ? extends V> a, Map<K, ? extends V> b) {
		if (a == null || b == null) {
			return null;
		}

		Map<K, V> bCopy = new HashMap<>(b);
		List<KeyedPair<K, V>> result = new ArrayList<>();
		for (Map.Entry<K, ? extends V> entry : a.entrySet()) {
			K key = entry.getKey();
			V aVal = entry.getValue();
			V bVal = b.get(key);
			bCopy.remove(key);

			if (bVal == null || !aVal.equals(bVal)) {
				result.add(new KeyedPair<>(key, aVal, bVal));
			}
		}

		for (Map.Entry<K, V> entry : bCopy.entrySet()) {
			result.add(new KeyedPair<>(entry.getKey(), null, entry.getValue()));
		}

		return result;
	}


	public static class KeyedPair<K, V> extends StreamUtil.Pair<V, V> {

		private final K key;

		public KeyedPair(K key, V a, V b) {
			super(a, b);
			this.key = key;
		}

		public K getKey() {
			return key;
		}

		@Override
		public String toString() {
			return "{" + key + ": " + getA() + ", " + getB() + "}";
		}
	}
}
