/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import difflib.DiffException;
import difflib.DiffUtils;
import difflib.Patch;
import difflib.myers.Equalizer;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.function.Predicate;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Find differences between directories, both in listing and file contents
 */
public final class DiffDirectory {

	private DiffDirectory() {
	}

	private static final Pattern REGEX_END_PATTERN = Pattern.compile("(.*)\\.regex(\\.[^.]+|)");

	private static StreamUtil.Pair<Path, Boolean> getGeneratedFile(Path referenceFile, Path referenceRoot, Path outputRoot, boolean allowRegex) {
		Path referenceFileRelative = referenceRoot.relativize(referenceFile);

		boolean isRegex = false;
		if (allowRegex) {
			final Matcher matcher = REGEX_END_PATTERN.matcher(referenceFileRelative.getFileName().toString());
			if (matcher.matches()) {
				isRegex = true;
				final Path newRelFile = referenceFileRelative.resolveSibling(matcher.group(1) + matcher.group(2));
				final Path newAbsFile = referenceRoot.resolve(newRelFile);
				if (Files.exists(newAbsFile)) {
					throw new AssertionError("Both " + referenceFile + " and " + newAbsFile + " refer to the same file "+outputRoot.resolve(newRelFile)+".");
				}
				referenceFileRelative = newRelFile;
			}
		}

		return new StreamUtil.Pair<>(outputRoot.resolve(referenceFileRelative), isRegex);
	}

	private static Optional<String> processReferenceFile(Path referenceFile, Path referenceRoot, Path outputRoot, int contextSize, boolean allowRegex) throws IOException {
		final StreamUtil.Pair<Path, Boolean> generatedFilePair = getGeneratedFile(referenceFile, referenceRoot, outputRoot, allowRegex);
		final Path generatedFile = generatedFilePair.getA();
		final boolean isRegex = generatedFilePair.getB();

		if (!Files.isRegularFile(generatedFile)) {
			return Optional.of("The file " + generatedFile + " does not exist or is not a regular file.");
		}
		final List<String> expectedLines;
		try {
			expectedLines = Files.lines(referenceFile).collect(Collectors.toList());
		} catch (IOException | UncheckedIOException e) {
			throw new IOException("Could not read expected file "+generatedFile,e);
		}
		final List<String> actualLines;
		try {
			actualLines = Files.lines(generatedFile).collect(Collectors.toList());
		} catch (IOException | UncheckedIOException e) {
			throw new IOException("Could not read actual file "+generatedFile,e);
		}

		Equalizer<String> eq = null;
		if (isRegex) {
			eq = new RegexEqualizer();
		}

		Patch<String> patch = DiffUtils.diff(expectedLines, actualLines, eq);

		if (!patch.getDeltas().isEmpty()) {
			final String diffStr = DiffUtils.generateUnifiedDiff(referenceFile.toString(), generatedFile.toString(), expectedLines, patch, contextSize)
					.stream().collect(Collectors.joining("\n"));
			return Optional.of(diffStr);
		}
		return Optional.empty();
	}

	/**
	 * Verify that all files in referenceRoot, filtered by fileFilter have an equal file relative to generatedRoot.
	 *
	 * @param referenceRoot directory with reference files
	 * @param generatedRoot directory with actual files
	 * @param fileFilter filter for files in reference directory
	 * @param contextSize context size for diffs
	 * @param allowRegex If set, files in the reference directory with an extension matching .regex.* will be
	 * referenced to the file without .regex, while each line will be handled as a regular
	 * expression that has to match the corresponding line in the generated file dir
	 * @throws IOException on broken IO
	 */
	public static void verifyFileContents(Path referenceRoot, Path generatedRoot, Predicate<Path> fileFilter, final int contextSize, boolean allowRegex) throws IOException, DiffException {
		final boolean[] hasFiles = {false};
		final Stream.Builder<String> diffBuilder = Stream.builder();
		Files.walkFileTree(referenceRoot, new SimpleFileVisitor<Path>() {
			@Override
			public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
				if (fileFilter.test(file)) {
					processReferenceFile(file, referenceRoot, generatedRoot, contextSize, allowRegex)
							.ifPresent(diffBuilder::accept);
					hasFiles[0] = true;
				}
				return FileVisitResult.CONTINUE;
			}
		});

		if (!hasFiles[0]) {
			throw new DiffException("Reference root is empty or all files were filtered out.");
		}
		final String res = diffBuilder.build().collect(Collectors.joining("\n"));
		if (!res.isEmpty()) {
			throw new DiffException("Unexpected difference:\n" + res);
		}
	}

	/**
	 * Verify that all files in referenceRoot, filtered by fileFilter have an equal file relative to generatedRoot.
	 *
	 * @param referenceRoot directory with reference files
	 * @param generatedRoot directory with actual files
	 * @param allowedExtensions file extensions that the reference dir will be filtered for
	 * @param contextSize context size for diffs
	 * @param allowRegex If set, files in the reference directory with an extension matching .regex.* will be
	 * referenced to the file without .regex, while each line will be handled as a regular
	 * expression that has to match the corresponding line in the generated file dir
	 * @throws IOException on broken IO
	 */
	public static void verifyFileContents(Path referenceRoot, Path generatedRoot, Collection<String> allowedExtensions, int contextSize, boolean allowRegex) throws IOException, DiffException {
		verifyFileContents(referenceRoot, generatedRoot, hasExtension(allowedExtensions), contextSize, allowRegex);
	}


	private static List<String> listFiles(Path root, Predicate<Path> filter) throws IOException {
		return Files.walk(root)
				.filter(Files::isRegularFile)
				.filter(filter)
				.map(root::relativize)
				.map(Object::toString)
				.sorted()
				.collect(Collectors.toList());
	}

	/**
	 * Verify that a recursive listing of files is equal
	 *
	 * @param referenceRoot reference directory
	 * @param generatedRoot generated directory
	 * @param fileFilter filter that will be applied to both listings
	 * @param contextSize context size for diffs
	 * @throws IOException on broken IO
	 */
	public static void verifyFileList(Path referenceRoot, Path generatedRoot, Predicate<Path> fileFilter, int contextSize) throws IOException, DiffException {
		final List<String> referenceFiles = listFiles(referenceRoot, fileFilter);
		final List<String> generatedFiles = listFiles(generatedRoot, fileFilter);

		Patch<String> patch = DiffUtils.diff(referenceFiles, generatedFiles);

		if (!patch.getDeltas().isEmpty()) {
			final String diffStr = DiffUtils.generateUnifiedDiff(referenceRoot.toString(), generatedRoot.toString(), referenceFiles, patch, contextSize)
					.stream().collect(Collectors.joining("\n"));
			throw new DiffException("Differing files generated:\n" + diffStr);
		}
	}

	/**
	 * Verify that a recursive listing of files is equal
	 *
	 * @param referenceRoot reference directory
	 * @param generatedRoot generated directory
	 * @param allowedExtensions filter that will be applied to both listings
	 * @param contextSize context size for diffs
	 * @throws IOException on broken IO
	 */
	public static void verifyFileList(Path referenceRoot, Path generatedRoot, Collection<String> allowedExtensions, int contextSize) throws IOException, DiffException {
		verifyFileList(referenceRoot, generatedRoot, hasExtension(allowedExtensions), contextSize);
	}

	/**
	 * Return a function that checks if a file has an allowed extension
	 *
	 * @param allowedExtensions collection of allowed extension
	 * @return a function that returns true if a supplied file has one of the specified extensions
	 */
	private static Predicate<Path> hasExtension(Collection<String> allowedExtensions) {
		return p -> {
			final String str = p.toString();
			return allowedExtensions.stream().anyMatch(e -> str.endsWith("." + e));
		};
	}

}
