/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.security.DigestOutputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

/**
 * Created by Ramon on 09.05.2016.
 */
public final class DigestUtil {

	public static final Charset UTF8 = Charset.forName("UTF-8");
	public static final int BUF_SIZE = 4096;

	private DigestUtil() {
	}

	/**
	 * Create digest for a file
	 * @param file file to creat digest for
	 * @return the digest
	 * @throws IOException on broken IO
	 */
	public static MessageDigest getDigest(Path file) throws IOException {
		MessageDigest digest = createDigest();
		byte[] buf = new byte[BUF_SIZE];
		try (InputStream stream = Files.newInputStream(file)) {
			while (stream.available() > 0) {
				int size = stream.read(buf);
				digest.update(buf, 0, size);
			}
		}
		return digest;
	}

	/**
	 * Create a MD5 digest
	 *
	 * @return new digest
	 * @throws IOException if digest cannot be created (because it is caught already by other filehandling and without digest, outputs are basically useless)
	 */
	public static MessageDigest createDigest() throws IOException {
		try {
			return MessageDigest.getInstance("MD5");
		} catch (NoSuchAlgorithmException e) {
			throw new IOException("Could not attain MD5 Digest", e);
		}
	}

	/**
	 * Make a pretty string from a digest
	 * @param digest the digest
	 * @return a pretty string
	 */
	public static String prettifyDigest(MessageDigest digest) {
		StringBuilder sb = new StringBuilder();
		for (byte b : digest.digest()) {
			sb.append(Integer.toHexString(b));
		}
		return sb.toString();
	}

	/**
	 * Creates a buffered PrintWriter that also generates a digest
	 * Used to generate hashes of generated output files for incremental building
	 * <p>
	 * File is opened as TRUNCATE, CREATE, WRITE
	 *
	 * @param path File
	 * @param md MessageDigest that will conatin the digest result
	 * @return Buffered PrintWriter
	 * @throws IOException if file cannot be opened
	 */
	public static PrintWriter createDigestPrintWriter(Path path, MessageDigest md) throws IOException {
		return new PrintWriter(new OutputStreamWriter(new DigestOutputStream(new BufferedOutputStream(Files.newOutputStream(path, StandardOpenOption.TRUNCATE_EXISTING, StandardOpenOption.WRITE, StandardOpenOption.CREATE)), md), UTF8));
	}

	public static class PrintWriterSupplier implements IWriterSupplier<PrintWriter> {

		private MessageDigest md;

		public PrintWriterSupplier() {
		}

		@Override
		public MessageDigest getDigest() {
			return md;
		}

		@Override
		public PrintWriter getWriter(Path path) throws IOException {
			md = createDigest();
			return createDigestPrintWriter(path, md);
		}
	}
}
