/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import de.tu_darmstadt.rs.process.ProcessUtil;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UncheckedIOException;
import java.io.UnsupportedEncodingException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Stream;

/**
 * Created by Ramon on 09.05.2016.
 */
public final class FileUtil {

	public static final String DEFAULT_FILE_ENCODING = "UTF-8";

	private static final Logger logger = LoggerFactory.getLogger(FileUtil.class);
	private static final String exceptionMessage = "Failed to scan existing outputs";

	private FileUtil() {
	}

	/**
	 * Delete all files in a dir that are not mentioned in a list
	 *
	 * @param outputDir the directory
	 * @param outputFiles the files to keep
	 * @param cleanedUpFiles all files that are deleted will be added to this list
	 */
	public static void cleanupFiles(Path outputDir, Set<Path> outputFiles, List<Path> cleanedUpFiles) {
		if (Files.exists(outputDir)) {
			try (Stream<Path> walkStream = Files.walk(outputDir)) {
				walkStream.forEach(p -> {
					if (Files.isRegularFile(p) && !outputFiles.contains(p)) {
						try {
							Files.delete(p);
							cleanedUpFiles.add(p);
						} catch (Exception e) {
							logger.error("Failed to clean up {}", p, e);
						}
					}
				});
				walkStream.close();
			} catch (IOException | UncheckedIOException e) {
				throw new RuntimeException(exceptionMessage, e);
			}
		}
	}

	/**
	 * Recursively delete all empty directories
	 * @param outputDir the root directory
	 */
	public static void removeEmptyDirectories(Path outputDir) {
		try (Stream<Path> walkStream = Files.walk(outputDir)) {
			walkStream.filter(Files::isDirectory).filter(p -> {
				try (Stream<Path> fileStream = Files.list(p)) {
					return fileStream.count() == 0L;
				} catch (IOException e) {
					logger.error("Failed to list directory contents", e);
					return false;
				}
			}).forEach((path) -> {
				try {
					Files.delete(path);
				} catch (IOException e) {
					logger.error("Failed to delete empty dir", e);
				}
			});
		} catch (IOException e) {
			logger.error(exceptionMessage, e);
		}
	}


	public static String replaceTildeWithUserHome(String file) {
		return file.replaceFirst("^~", System.getProperty("user.home"));
	}

	/**
	 * replaces a line in a file. If the toSearchRegex contains >1 capturing groups the toAppend is appended between the first and second group
	 * @param toSearchRegex the regex to look for in each line
	 * @param toAppend the string to be appended
	 * @param srcFile the input file
	 * @param dstFile the output file
	 * @throws IOException if something goes wrong with the files
	 */
	public static void searchLineAndAppendInStream(String toSearchRegex, String toAppend, File srcFile, File dstFile) {
		try {
			BufferedReader input = new BufferedReader(new FileReader(srcFile));
			BufferedWriter output = new BufferedWriter(new FileWriter(dstFile));
			try {

				String line;
				while ((line = input.readLine()) != null) {
					Pattern p = Pattern.compile(toSearchRegex);
					Matcher m = p.matcher(line);
					StringBuffer sb = new StringBuffer(line);
					if (m.find()) {
						if (m.groupCount() > 1) {
							sb = new StringBuffer(m.group(1));
						} else {
							sb = new StringBuffer(m.group());
						}
						sb.append(toAppend);
						for (int i=2; i<=m.groupCount(); i++) {
							sb.append(m.group(i));
						}
					}
					output.write(sb.toString());
					output.newLine();
				}
			} finally {
				input.close();
				output.close();
			}
		} catch (IOException e) {
			logger.error("IOException while processing the files: " + srcFile + " and " + dstFile);
			logger.error(e.getMessage());
		}
	}


	/**
	 * @param location a folder to be checked
	 * @return true if the directory does not exist or if it exists and is empty
	 */
	public static boolean isFolderEmpty(File location) {

		if (!location.exists()) {
			return true;
		} else if (location.isDirectory() && location.list().length == 0) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * @param folder the folder to be deleted
	 */
	public static void deleteDirectoryRecursively(File folder) {
		try {
			FileUtils.deleteDirectory(folder);
		} catch (IOException e) {
			logger.error("Can not delete the temporary folder " + folder);
		}

	}

	/**
	 * Writes to a temporary file and deletes them when the program closes
	 *
	 * @param outputFile the file to write to
	 * @param content the content of the file
	 */
	public static void writeTemporaryFile(File outputFile, String content) {
		try {
			final String fileSpace = "File ";
			if (outputFile.exists()) {
				logger.error(fileSpace + outputFile + " already exists, not overwriting existing files!");
			}
			if (outputFile.isDirectory()) {
				logger.error(fileSpace + outputFile + " is a directory, can't write into directory");
			} else {
				outputFile.deleteOnExit();
				outputFile.getAbsoluteFile().getParentFile().mkdirs();
				PrintWriter writer = new PrintWriter(outputFile, DEFAULT_FILE_ENCODING);

				writer.write(content);
				writer.flush();
				writer.close();
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}


	}

	/**
	 * Creates a temporary folder that doesn't exist yet
	 *
	 * @param prefix a prefix for that folder
	 * @return a path to the new folder
	 */
	public static Path createNewTempFolder(String prefix) {
		String program = "mktemp";
		String suffix = "_XXXXXXXXXXXX";
		if (prefix != null) {
			suffix = prefix + suffix;
		}
		String[] cmd = {program, "-d", "-t", suffix};
		ProcessBuilder pb = new ProcessBuilder(cmd);
		StringBuilder path = new StringBuilder();
		StringBuilder errorLog = new StringBuilder();
		try {
			ProcessUtil.runAndWaitForSeparate(pb, path::append, errorLog::append);
			if (errorLog.length() > 0) {
				logger.error("Temporary directory creation failed: " + errorLog.toString());
			} else {
				return Paths.get(path.toString());
			}
		} catch (IOException e) {
			logger.error("IOException while trying to create new temp folder");
			e.printStackTrace();
		}
		throw new RuntimeException("Failed to create a temp folder.");
	}
}
