/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.util;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by Ramon on 04.04.2016.
 */
public class GroupWriter implements AutoCloseable {

	private final List<PrintWriter> writers = new ArrayList<>();
	private final List<IWriterSupplier<? extends PrintWriter>> writerSuppliers = new ArrayList<>();

	public GroupWriter(List<Path> files, Supplier<IWriterSupplier<PrintWriter>> writerSupplierSupplier) throws IOException {
		for (Path file : files) {
			Files.createDirectories(file.getParent());
			IWriterSupplier<? extends PrintWriter> writerSupplier = writerSupplierSupplier.get();
			writerSuppliers.add(writerSupplier);
			writers.add(writerSupplier.getWriter(file));
		}
	}

	public void writeln(String[] lines) {
		verifyCount(lines);

		Iterator<PrintWriter> wIt = writers.iterator();
		for (String line : lines) {
			PrintWriter writer = wIt.next();
			writer.println(line);
		}
	}

	public <T> void writeSingle(T data, List<ILineMapper<T>> mappers) {
		verifyCount(mappers);

		writeInternal(data, mappers);
	}

	public <T> void writeAllSingle(T data, ILineMapper<T> mapper) {
		for (PrintWriter writer : writers) {
			mapper.write(writer, data);
		}
	}

	private <T> void writeInternal(T data, List<ILineMapper<T>> mappers) {
		Iterator<PrintWriter> wIt = writers.iterator();
		Iterator<ILineMapper<T>> mIt = mappers.iterator();
		while (wIt.hasNext()) {
			PrintWriter writer = wIt.next();
			ILineMapper<T> mapper = mIt.next();
			if (mapper != null) {
				mapper.write(writer, data);
			}
		}
	}

	/**
	 * Write the data-stream to all files in this group. Use the LineMappers to customize the representation for each file;
	 *
	 * @param data Stream of elements to write to all files
	 * @param mappers Helper function to write one T to a PrintStream
	 * @param <T> Type to write
	 */
	public <T> void write(Stream<T> data, List<ILineMapper<T>> mappers) {
		verifyCount(mappers);

		data.forEach(d -> writeInternal(d, mappers));
	}


	public void writeAll(Stream<Consumer<PrintWriter>> mappers) {
		Iterator<PrintWriter> wIt = writers.iterator();
		Iterator<Consumer<PrintWriter>> mIt = mappers.iterator();
		while (wIt.hasNext()) {
			PrintWriter writer = wIt.next();
			Consumer<PrintWriter> mapper = mIt.next();
			if (mapper != null) {
				mapper.accept(writer);
			}
		}
	}

	public <T> void writeAll(Stream<T> data, ILineMapper<T> mapper) {
		data.forEach(d -> writeAllSingle(d, mapper));
	}

	private <T> void verifyCount(T[] any) {
		if (writers.size() != any.length) {
			throw new IllegalArgumentException("We have " + writers.size() + " output files, but content specifies " + any.length + " mappers.");
		}
	}

	private <T> void verifyCount(List<ILineMapper<T>> any) {
		if (writers.size() != any.size()) {
			throw new IllegalArgumentException("We have " + writers.size() + " output files, but content specifies " + any.size() + " mappers.");
		}
	}

	@Override
	public void close() throws IOException {
		for (PrintWriter writer : writers) {
			writer.close();
		}
	}

	public void writeAllln(String line) {
		for (PrintWriter writer : writers) {
			writer.println(line);
		}
	}

	public void writeAllln() {
		for (PrintWriter writer : writers) {
			writer.println();
		}
	}


	public List<MessageDigest> getDigests() {
		return writerSuppliers.stream().map(IWriterSupplier::getDigest).collect(Collectors.toList());
	}

	/**
	 * Functional interface, with which one has full access to the underlying PrintWriter to convert T into any String representation needed
	 *
	 * @param <T> type of data to be mapped to a line
	 */
	public interface ILineMapper<T> {

		static <T> ILineMapper<T> singleLine(Function<T, String> singeLine) {
			return ((writer, data) -> writer.println(singeLine.apply(data)));
		}

		static <T> ILineMapper<T> nop() {
			return ((writer, data) -> {
			});
		}

		void write(PrintWriter writer, T data);
	}

}
