/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import java.util.Iterator;
import java.util.List;
import java.util.Optional;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * Created by Kris Heid on 19.01.17.
 */
public final class LevenshteinDistance {

	private LevenshteinDistance() {
	}


	private static int minimum(int a, int b, int c) {
		return Math.min(Math.min(a, b), c);
	}

	/**
	 * Calculates the Levenstein distance
	 *
	 * @param lhs one string
	 * @param rhs another string
	 * @return the Levenstein distance
	 */
	public static int get(CharSequence lhs, CharSequence rhs) {
		int[][] distance = new int[lhs.length() + 1][rhs.length() + 1];

		for (int i = 0; i <= lhs.length(); i++) {
			distance[i][0] = i;
		}
		for (int j = 1; j <= rhs.length(); j++) {
			distance[0][j] = j;
		}

		for (int i = 1; i <= lhs.length(); i++) {
			for (int j = 1; j <= rhs.length(); j++) {
				distance[i][j] = minimum(
						distance[i - 1][j] + 1,
						distance[i][j - 1] + 1,
						distance[i - 1][j - 1] + ((lhs.charAt(i - 1) == rhs.charAt(j - 1))? 0 : 1));
			}
		}

		return distance[lhs.length()][rhs.length()];
	}

	/**
	 * Finds the String in the list that has smallest Distance to the reference
	 *
	 * @param reference a reference String
	 * @param toMatch a list of strings to select the one with smalles distance
	 * @return the String in the list that has smallest Distance to the reference
	 */
	public static String matchFromList(String reference, List<String> toMatch) {
		int minDistance = Integer.MAX_VALUE;
		String minDistanceMatch = null;
		Iterator<String> iterator = toMatch.iterator();
		while (iterator.hasNext()) {
			String otherString = iterator.next();
			int distance = get(reference, otherString);
			if (distance < minDistance) {
				minDistance = distance;
				minDistanceMatch = otherString;
			}
		}
		return minDistanceMatch;
	}

	/**
	 * Find the best match of the reference in a list
	 * @param reference the reference whos match shall be found
	 * @param toMatch the list, the reference is compared against
	 * @param toStringFunction the function that converts the given generic type to a string
	 * @param <T> a generic type
	 * @return a match from the list with the smalles levenstein distance to the reference
	 */
	public static <T> Optional<T> matchFromList(T reference, List<T> toMatch, Function<T, String> toStringFunction) {
		String refString = toStringFunction.apply(reference);
		List<String> toMatchStrings = toMatch.stream()
				.map(x -> toStringFunction.apply(x))
				.collect(Collectors.toList());
		String match = LevenshteinDistance.matchFromList(refString, toMatchStrings);
		Optional<T> ret = toMatch.stream()
				.filter(x -> toStringFunction.apply(x).equalsIgnoreCase(match))
				.findFirst();
		return ret;
	}
}
