/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Optional;
import java.util.Set;

/**
 * Created by Kris Heid on 10.10.16.
 */
public final class SpartanMCFirmwareUtil {

	public static final String DEFAULT_INCLUDE_FOLDER = "include";
	public static final String DEFAULT_SRC_FOLDER = "src";
	public static final String CONFIG_BUILD_MAKE_FILE = "config-build.mk";

	private static final Logger log = LoggerFactory.getLogger(SpartanMCFirmwareUtil.class);

	private SpartanMCFirmwareUtil() {}

	/**
	 * @param dir a potential Firmware directory to be inspected
	 * @return true if the folder exists and contains an include and a src folder and a config-build.mk file
	 */
	public static boolean isValidFirmwareFolderStructure(Path dir) {
		Path src = dir.resolve(DEFAULT_SRC_FOLDER);
		Path configBuild = dir.resolve(CONFIG_BUILD_MAKE_FILE);
		return Files.isDirectory(src) && Files.isRegularFile(configBuild);
	}

	/**
	 * Finds the firmware root dir for given files if it exists. Each files parent directories are searched until {@link #isValidFirmwareFolderStructure(Path)} delivers true.
	 *
	 * @param sourceFiles that are located in a firmware folder or at some other location.
	 * @return the firmware root folder if at least one file resides in a firmware folder and null if none or more than one firmware folder has bee found.
	 */
	public static Path findSpartanMCFirmwareRootDir(Path... sourceFiles) {
		Set<Path> firmwareRoot = new HashSet<>();
		for (Path sourceFile : sourceFiles) {
			Path tmpFile = sourceFile;
			boolean hit = isValidFirmwareFolderStructure(tmpFile);

			while (tmpFile != null && !hit) {
				tmpFile = tmpFile.getParent();
				hit = isValidFirmwareFolderStructure(tmpFile);
			}
			if (hit) {
				firmwareRoot.add(tmpFile);
			}
		}
		if (firmwareRoot.size() != 1) {
			log.error("There were " + firmwareRoot.size() + " potentially matching firmware folders found: " + firmwareRoot + "for the source files" + Arrays.toString(sourceFiles));
			return null;
		}
		Optional<Path> res = firmwareRoot.stream().findFirst();
		return res.orElse(null);
	}

}
