/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import de.tu_darmstadt.rs.process.ProcessUtil;
import de.tu_darmstadt.rs.process.RingBuffer;
import de.tu_darmstadt.rs.process.StreamGobbler;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;


/**
 * Created by Kris Heid on 10.10.16.
 */
public final class SpartanMCMakeUtil {

	private static final Logger logger = LoggerFactory.getLogger(SpartanMCMakeUtil.class);

	public static final float NS_PER_MS = 1E-6f;
	private static final Logger log = LoggerFactory.getLogger(SpartanMCMakeUtil.class);
	private static final String MAKEFILE = "Makefile";
	private static final String MAKE = "make";
	private static final String NEW_FIRMWARE = "newfirmware";
	private static final String NEW_PROJECT = "newproject";
	private static final String PATH = "+path=";


	private SpartanMCMakeUtil() {
	}

	private static String[] newFirmwareCommand(String firmwareName) {
		return new String[]{MAKE, NEW_FIRMWARE, PATH + firmwareName};
	}

	private static String[] newProjectCommand(Path firmwareName) {
		return new String[]{MAKE, NEW_PROJECT, PATH + firmwareName.toString()};
	}


	/**
	 * @param firmwareName how the new firmware shall be named
	 * @param projectDir the project directory where the firmware shall be created
	 * @param overwriteExistingFirmware shall the firmware folder be overwritten if it already exists
	 */
	public static void generateNewFirmware(String firmwareName, String projectDir, boolean overwriteExistingFirmware) {

		String[] cmd = newFirmwareCommand(firmwareName);

		File execDir = new File(projectDir);

		Path firmwareDir = execDir.toPath().resolve(firmwareName);

		ProcessBuilder processBuilder = new ProcessBuilder(cmd).directory(execDir);

		String spartanmcRoot = EnvironmentVariablesUtil.getEnvironmentVariableAsFolder(EnvironmentVariablesUtil.SPARTANMC_ROOT).toString();
		processBuilder.environment().put(EnvironmentVariablesUtil.SPARTANMC_ROOT, spartanmcRoot);

		try {
			if (!FileUtil.isFolderEmpty(firmwareDir.toFile()) && overwriteExistingFirmware) {
				if (SpartanMCFirmwareUtil.isValidFirmwareFolderStructure(firmwareDir)) {
					log.warn("Deleting firmware folder " + firmwareName + " and create a new one");
					FileUtils.deleteDirectory(firmwareDir.toFile());
				} else {
					log.error("Will not overwrite existing location " + firmwareName + " since the given folder seems not to have a valid firmware structure.");
					return;
				}
			}
			if (FileUtil.isFolderEmpty(firmwareDir.toFile())) {
				int exitCode = ProcessUtil.runAndWaitForSeparate(processBuilder, log::debug, (e) -> {
					log.error("An error Occured while trying to create a new firmware " + firmwareName + " in the project " + execDir);
					log.error(e);
				});
				if (exitCode != 0) {
					throw new RuntimeException("Generating new firmware finished with exit code: " + exitCode);
				}
			} else {
				log.error("Firmware folder can not be created, most likely because the directory " + firmwareDir + " is not empty.");
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * @param projectDir a potential Firmware directory to be inspected
	 * @return true if the folder exists and contains a makefile
	 */
	public static boolean isValidProjectFolderStructure(File projectDir) {
		if (projectDir != null && projectDir.isDirectory()) {
			File makefile = new File(projectDir, MAKEFILE);
			return (makefile.exists() && makefile.isDirectory());
		}
		return false;
	}

	/**
	 * Generates a new Project on the specified location
	 *
	 * @param projectLocation the project location that shall be created
	 */
	public static void generateNewProject(Path projectLocation) {
		String[] cmd = newProjectCommand(projectLocation.toAbsolutePath());

		File projectFile = projectLocation.toFile();

		File spartanmcRoot = EnvironmentVariablesUtil.getEnvironmentVariableAsFolder(EnvironmentVariablesUtil.SPARTANMC_ROOT).toFile();

		ProcessBuilder processBuilder = new ProcessBuilder(cmd).directory(spartanmcRoot);


		processBuilder.environment().put(EnvironmentVariablesUtil.SPARTANMC_ROOT, spartanmcRoot.toString());

		try {
			// take care to not overwrite or delete existing projects
			if (FileUtil.isFolderEmpty(projectFile)) {
				int exitCode = ProcessUtil.runAndWaitForSeparate(processBuilder, log::debug, (e) -> {
					log.error("An error occured while trying to create a new project " + projectLocation);
					log.error(e);
				});
				if (exitCode != 0) {
					throw new RuntimeException("Generating new project finished with exit code: " + exitCode);
				}
			} else if (isValidProjectFolderStructure(projectFile)) {
				log.debug("The folder" + projectLocation + "contains a valid project and doesn't have to be newly generated");
			} else {
				log.warn("Project folder will not be created, most likely because the directory " + projectLocation + " is not empty but doesn't contain a valid project makefile.");
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Run make and return the duration it took
	 *
	 * @param projectDir directory to run in
	 * @param targets targets to execute
	 * @return duration in ms
	 *
	 * @throws IOException on broken IO
	 */
	public static float runMakeTarget(Path projectDir, String... targets) throws IOException {
		return runMakeTarget(projectDir, null, targets);
	}

	/**
	 * Run make and return the duration it took
	 *
	 * @param projectDir directory to run in
	 * @param envVars environment variables to set in the target program
	 * @param targets targets to execute
	 * @return duration in ms
	 *
	 * @throws IOException on broken IO
	 */
	public static float runMakeTarget(Path projectDir, Map<String, String> envVars, String... targets) throws IOException {
		log.info(projectDir + "> make " + Stream.of(targets).collect(Collectors.joining(" ")));


		ProcessBuilder builder;
		builder = new ProcessBuilder(Stream.concat(Stream.of("make"), Stream.of(targets)).toArray(String[]::new));

		builder.directory(projectDir.toFile());
		builder.environment().remove("MAKEFLAGS");
		if (envVars != null) {
			builder.environment().putAll(envVars);
		}

		RingBuffer<String> output = new RingBuffer<>(60);
		long before = System.nanoTime();
		int exitCode = ProcessUtil.runAndWaitForSeparate(builder,
				StreamGobbler.outputToMultiple(StreamGobbler.shortenToOneLineToStdOut(), output),
				line -> output.accept(StreamGobbler.ANSI_COLOR_RED+line));
		long after = System.nanoTime();
		StreamGobbler.clearLineStdout();


		float duration = (after - before) * NS_PER_MS;
		log.info("  {}ms", duration);

		if (exitCode != 0) {
			logger.warn("Exit code {}! End of make log: ",exitCode);
			for (String s : output) {
				logger.warn(StreamGobbler.ANSI_RESET+s);
			}
			throw new MakeErrorExitCode("make " + Arrays.toString(targets) + " in " + projectDir + " exited with exit code " + exitCode, duration);
		}

		return duration;
	}


	/**
	 * Exception caused by failing a failing make task
	 */
	public static class MakeErrorExitCode extends IOException {

		private final float duration;

		public MakeErrorExitCode(String message, float duration) {
			super(message);
			this.duration = duration;
		}

		public float getDuration() {
			return duration;
		}
	}


}
