/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.util;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.function.BinaryOperator;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Collector;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

public final class StreamUtil {


	public StreamUtil() {
	}

	/**
	 * This collector expects a stream to contain exactly one element. If it doesn't, it throws an IllegalStateException
	 *
	 * Warning: not short circuiting! The whole stream will be evaluated before the exception is thrown
	 *
	 * @param <T> The type the collector should return
	 * @return a collector
	 */
	public static <T> Collector<T, SingletonAccumulator<T>, T> exactlyOne() {
		return exactlyOne(null);
	}

	/**
	 * This collector expects a stream to contain exactly one element. If it doesn't, it throws an IllegalStateException
	 *
	 * Warning: not short circuiting! The whole stream will be evaluated before the exception is thrown
	 *
	 * @param <T> The type the collector should return
	 * @param errorMessageGen Override normal error message
	 * @return a collector
	 */
	public static <T> Collector<T, SingletonAccumulator<T>, T> exactlyOne(Function<List<T>, String> errorMessageGen) {
		return exactlyOne(errorMessageGen, null);
	}

	/**
	 * This collector expects a stream to contain exactly one element. If it doesn't, it throws an IllegalStateException
	 *
	 * Warning: not short circuiting! The whole stream will be evaluated before the exception is thrown
	 *
	 * @param <T> The type the collector should return
	 * @param multipleElementsErrorGen Override error message if multiple elements are present
	 * @param noElementsErrorGen Override error message if no elements are present
	 * @return a collector
	 */
	public static <T> Collector<T, SingletonAccumulator<T>, T> exactlyOne(Function<List<T>, String> multipleElementsErrorGen, Supplier<String> noElementsErrorGen) {
		return Collector.of(
				() -> new SingletonAccumulator<>(multipleElementsErrorGen, noElementsErrorGen),
				SingletonAccumulator::accumulate,
				SingletonAccumulator::failMerge,
				SingletonAccumulator::requireExactlyOne);
	}

	/**
	 * This collector expects a stream to contain zero or one element. If it doesn't, it throws an IllegalStateException
	 *
	 * Warning: not short circuiting! The whole stream will be evaluated before the exception is thrown
	 *
	 * @param <T> The type the collector should return
	 * @return a collector
	 */
	public static <T> Collector<T, SingletonAccumulator<T>, Optional<T>> atMostOne() {
		return Collector.of(
				SingletonAccumulator::new,
				SingletonAccumulator::accumulate,
				SingletonAccumulator::failMerge,
				SingletonAccumulator::requireAtMostOne);
	}

	/**
	 * This collector expects a stream to contain one or more elements. If it contains multiple elements, they are required to be equal. If it doesn't, it throws an IllegalStateException
	 *
	 * Warning: not short circuiting! The whole stream will be evaluated before the exception is thrown
	 *
	 * @param <T> The type the collector should return
	 * @return a collector
	 */
	public static <T> Collector<T, SingletonAccumulator<T>, T> atLeastOneOrEqual() {
		return atLeastOneOrEqual(null, null);
	}

	/**
	 * This collector expects a stream to contain one or more elements. If it contains multiple elements, they are required to be equal. If it doesn't, it throws an IllegalStateException
	 *
	 * Warning: not short circuiting! The whole stream will be evaluated before the exception is thrown
	 *
	 * @param <T> The type the collector should return
	 * @param multipleElementsErrorGen Override error message if multiple elements are present
	 * @param noElementsErrorGen Override error message if no elements are present
	 * @return a collector
	 */
	public static <T> Collector<T, SingletonAccumulator<T>, T> atLeastOneOrEqual(Function<List<T>, String> multipleElementsErrorGen, Supplier<String> noElementsErrorGen) {
		return Collector.of(
				() -> new SingletonAccumulator<>(multipleElementsErrorGen, noElementsErrorGen),
				SingletonAccumulator::accumulate,
				SingletonAccumulator::failMerge,
				SingletonAccumulator::requireAtLeastOneOrEqual);
	}

	/**
	 * This collector expects a stream to contain one or more elemente. If it doesn't, it throws an IllegalStateException
	 *
	 * Warning: not short circuiting! The whole stream will be evaluated before the exception is thrown
	 *
	 * @param <T> The type the collector should return
	 * @return a collector
	 */
	public static <T> Collector<T, SingletonAccumulator<T>, List<T>> atLeastOne() {
		return Collector.of(
				SingletonAccumulator::new,
				SingletonAccumulator::accumulate,
				SingletonAccumulator::failMerge,
				SingletonAccumulator::requireAtLeastOne);
	}

	/**
	 * Create a stream with the elements from a and b matched up into pairs
	 *
	 * @param aIt first iterator
	 * @param bIt second iterator
	 * @param <A> first element type
	 * @param <B> second element type
	 * @return stream of pairs
	 *
	 * @throws IllegalStateException if the lengths do not match
	 */
	public static <A, B> Stream<Pair<A, B>> streamZipped(Iterator<? extends A> aIt, Iterator<? extends B> bIt) {
		List<Pair<A, B>> list = new ArrayList<>();
		while (aIt.hasNext() && bIt.hasNext()) {
			list.add(new Pair<>(aIt.next(), bIt.next()));
		}

		if (aIt.hasNext() || bIt.hasNext()) {
			throw new IllegalStateException("Length does not match");
		}

		return list.stream();
	}

	public static <K> boolean containsAll(Map<K, ?> map, Collection<K> collection) {
		return map.keySet().containsAll(collection) && collection.size() == map.size();
	}

	/**
	 * flatMapper that either casts to the specified class or returns Stream.empty
	 *
	 * @param clazz the class to cast to
	 * @param <T> input type
	 * @param <R> result type
	 * @return Function that tries to cast
	 */
	public static <T, R> Function<T, Stream<R>> castOrEmpty(Class<R> clazz) {
		return (x) -> {
			if (clazz.isInstance(x)) {
				return Stream.of((R) x);
			} else {
				return Stream.empty();
			}
		};
	}

	/**
	 * If an Optional is present, return a stream containing its contents, otherwise return an empty stream
	 *
	 * Java 9 will contain this function in Optional itself (see: https://bugs.openjdk.java.net/browse/JDK-8050820).
	 *
	 * @param optional optional to stream
	 * @param <R> nested type
	 * @return Stream of optional content
	 */
	public static <R> Stream<R> streamOptional(Optional<R> optional) {
		if (optional.isPresent()) {
			return Stream.of(optional.get());
		} else {
			return Stream.empty();
		}
	}

	/**
	 * Allows to implement functionality like Stream's distinct, but for arbitrary keys
	 *
	 * @param keyExtractor Key to distinct
	 * @param <T> Content Type
	 * @return distinction function for use in filter
	 */
	public static <T> Predicate<T> distinctByKey(Function<? super T, Object> keyExtractor) {
		Map<Object, Boolean> seen = new ConcurrentHashMap<>();
		return t -> seen.putIfAbsent(keyExtractor.apply(t), Boolean.TRUE) == null;
	}

	public static <T> Stream<T> streamIterator(Iterator<T> it) {
		Iterable<T> iterable = () -> it;
		return StreamSupport.stream(iterable.spliterator(), false);
	}

	/**
	 * This can be used to do the equivalent of a Stream's findAny and findAll at the same time.
	 * It should be used like this:
	 * <pre>
	 * Stream.of(...).map(predicate)
	 * .collect(StreamUtil.AllOrAnyMatch.collector())
	 * </pre>
	 */
	public enum AllOrAnyMatch {
		NONE,
		ALL,
		SOME;

		/**
		 * Create from boolean
		 *
		 * @param b the boolean
		 * @return ALL if b is true, NONE otherwise
		 */
		public static AllOrAnyMatch fromBool(boolean b) {
			if (b) {
				return ALL;
			} else {
				return NONE;
			}
		}

		/**
		 * Merge two elements
		 *
		 * @param a first element
		 * @param b second element
		 * @return the merge
		 */
		public static AllOrAnyMatch accumulator(AllOrAnyMatch a, AllOrAnyMatch b) {
			if (a == b) {
				return a;
			} else {
				return SOME;
			}
		}

		/**
		 * Collect booleans
		 *
		 * @return the collected result
		 */
		public static Collector<Boolean, ?, Optional<AllOrAnyMatch>> collector() {
			return Collectors.mapping(AllOrAnyMatch::fromBool, Collectors.reducing(AllOrAnyMatch::accumulator));
		}

		/**
		 * Same as collector, but disallows the SOME result
		 *
		 * @return a collector
		 *
		 * @throws IllegalStateException if the result of the collector is Some
		 */
		public static Collector<Boolean, ?, Optional<Boolean>> collectorSomeForbidden() {
			return Collectors.collectingAndThen(collector(), opt -> opt.map(a -> {
				if (a == SOME) {
					throw new IllegalStateException("All or Any match is Some");
				}
				return a == ALL;
			}));
		}
	}


	public static class SingletonAccumulator<T> {

		private final Function<List<T>, String> multipleElementsErrorGen;
		private final Supplier<String> noElementsErrorGen;
		private T first;
		private List<T> multi;

		public SingletonAccumulator() {
			this(null, null);
		}

		public SingletonAccumulator(Function<List<T>, String> multipleElementsErrorGen, Supplier<String> noElementsErrorGen) {
			if (multipleElementsErrorGen != null) {
				this.multipleElementsErrorGen = multipleElementsErrorGen;
			} else {
				this.multipleElementsErrorGen = (list) -> "Stream contained multiple items: " + list;
			}
			if (noElementsErrorGen != null) {
				this.noElementsErrorGen = noElementsErrorGen;
			} else {
				this.noElementsErrorGen = () -> "Stream contained no elements";
			}
		}

		public void accumulate(T elem) {
			if (first == null) {
				first = elem;
			} else {
				if (multi == null) {
					multi = new ArrayList<>();
					multi.add(first);
				}

				multi.add(elem);
			}
		}

		public SingletonAccumulator<T> failMerge(SingletonAccumulator<T> merge) {
			throw new UnsupportedOperationException();
		}

		private T requireExactlyOne() {
			if (first == null) {
				throw new IllegalStateException(noElementsErrorGen.get());
			} else if (multi == null) {
				return first;
			} else {
				throw new IllegalStateException(multipleElementsErrorGen.apply(multi));
			}
		}

		private Optional<T> requireAtMostOne() {
			if (multi == null) {
				return Optional.ofNullable(first);
			} else {
				throw new IllegalStateException(multipleElementsErrorGen.apply(multi));
			}
		}

		private T requireAtLeastOneOrEqual() {
			if (first == null) {
				throw new IllegalStateException(noElementsErrorGen.get());
			} else {
				if (multi != null) {
					for (int i = 1; i < multi.size(); i++) {
						T current = multi.get(i);
						if (!Objects.equals(first, current)) {
							throw new IllegalStateException(multipleElementsErrorGen.apply(multi));
						}
					}
				}

				return first;
			}
		}

		private List<T> requireAtLeastOne() {
			if (multi != null) {
				return multi;
			} else if (first != null) {
				return Collections.singletonList(first);
			} else {
				throw new IllegalStateException(noElementsErrorGen.get());
			}
		}
	}


	public static class Pair<A, B> {

		private final A a;
		private final B b;

		public Pair(A a, B b) {
			this.a = a;
			this.b = b;
		}

		public A getA() {
			return a;
		}

		public B getB() {
			return b;
		}

		@Override
		public int hashCode() {
			int result = a.hashCode();
			result = 31 * result + b.hashCode();
			return result;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			Pair<?, ?> pair = (Pair<?, ?>) o;

			if (!a.equals(pair.a)) {
				return false;
			}
			return b.equals(pair.b);

		}

		@Override
		public String toString() {
			return "<" + a + "," + b + ">";
		}

		//accessF and accessC can be called the same in Java, but not in groovy...

		public static <A, B, R> Function<Pair<A, B>, R> accessF(BiFunction<A, B, R> f) {
			return
					p -> f.apply(p.getA(), p.getB());
		}

		public static <A, B> Consumer<Pair<A, B>> accessC(BiConsumer<A, B> f) {
			return
					p -> f.accept(p.getA(), p.getB());
		}

		public static <T extends Comparable<T>, U extends Comparable<U>> Comparator<Pair<T, U>> comparator() {
			return Comparator.<Pair<T, U>, T>comparing(Pair::getA)
					.thenComparing(Pair::getB);
		}
	}


	public static class IntPair<T> {

		private final int a;
		private final T b;

		public IntPair(int a, T b) {
			this.a = a;
			this.b = b;
		}

		public int getA() {
			return a;
		}

		public T getB() {
			return b;
		}

		@Override
		public int hashCode() {
			int result = a;
			result = 31 * result + b.hashCode();
			return result;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			IntPair<?> pair = (IntPair<?>) o;

			if (a != pair.a) {
				return false;
			}
			return b.equals(pair.b);

		}

		@Override
		public String toString() {
			return "<" + a + "," + b + ">";
		}

		//accessF and accessC can be called the same in Java, but not in groovy...

		public static <T, R> Function<IntPair<T>, R> accessF(BiFunction<Integer, T, R> f) {
			return
					p -> f.apply(p.getA(), p.getB());
		}

		public static <T> Consumer<IntPair<T>> accessC(BiConsumer<Integer, T> f) {
			return
					p -> f.accept(p.getA(), p.getB());
		}
	}

	/**
	 * Get a function that assigns an index to each element,
	 * starting with offset
	 *
	 * Can be used in Stream.map to assign indices to elements
	 *
	 * @param offset the number to start with
	 * @param <T> element type
	 * @return Index assigner
	 */
	public static <T> Function<T, IntPair<T>> numberElements(final int offset) {
		final int[] i = {offset};
		return (t) -> new IntPair<T>(i[0]++, t);
	}

	/**
	 * Get a function that assigns an index to each element.
	 *
	 * Can be used in Stream.map to assign indices to elements
	 *
	 * @param <T> element type
	 * @return Index assigner
	 */
	public static <T> Function<T, IntPair<T>> numberElements() {
		return numberElements(0);
	}

	/**
	 * Returns a stream of all sub elements that can be accessed through the childAccessor
	 *
	 * @param element a root element to start from
	 * @param childrenAccessor an accessor method to find children elements
	 * @param <T> the type of element
	 * @return a stream of all tree nodes
	 */
	public static <T> Stream<T> streamTree(T element, Function<T, Stream<? extends T>> childrenAccessor) {
		Stream<? extends T> children = childrenAccessor.apply(element);
		final Stream<T> subResult = children.flatMap(c -> streamTree(c, childrenAccessor));
		return Stream.concat(Stream.of(element), subResult);
	}

	/**
	 * Concat n streams.
	 *
	 * This is not optimized for very large amounts of Streams. See the warning in {@link Stream#concat}
	 *
	 * @param <T> element type
	 * @param streams streams to concat
	 * @return a stream
	 */
	@SafeVarargs
	@SuppressWarnings("unchecked")
	public static <T> Stream<T> concat(Stream<? extends T>... streams) {
		return (Stream<T>) Arrays.stream(streams).reduce(Stream::concat).orElse(Stream.empty());
	}


	/**
	 * Return a collector that applies both given collectors on the same input and returns their output as a pair
	 *
	 * @param a collector a
	 * @param b collector b
	 * @param <T> elememt type
	 * @param <A1> first collector internal type
	 * @param <A2> second collector internal type
	 * @param <R1> first collector result type
	 * @param <R2> second collector result type
	 * @return collector that returns a pair of the given collector's results
	 */
	public static <T, A1, A2, R1, R2> Collector<T, StreamUtil.Pair<A1, A2>, StreamUtil.Pair<R1, R2>> multiCollector(Collector<T, A1, R1> a, Collector<T, A2, R2> b) {
		return new Collector<T, StreamUtil.Pair<A1, A2>, StreamUtil.Pair<R1, R2>>() {
			@Override
			public Supplier<StreamUtil.Pair<A1, A2>> supplier() {
				Supplier<A1> aSupp = a.supplier();
				Supplier<A2> bSupp = b.supplier();
				return () -> new StreamUtil.Pair<>(aSupp.get(), bSupp.get());
			}

			@Override
			public BiConsumer<StreamUtil.Pair<A1, A2>, T> accumulator() {
				final BiConsumer<A1, T> aAcc = a.accumulator();
				final BiConsumer<A2, T> bAcc = b.accumulator();
				return (p, t) -> {
					aAcc.accept(p.getA(), t);
					bAcc.accept(p.getB(), t);
				};
			}

			@Override
			public BinaryOperator<Pair<A1, A2>> combiner() {
				BinaryOperator<A1> aComb = a.combiner();
				BinaryOperator<A2> bComb = b.combiner();
				return (p1, p2) -> new StreamUtil.Pair<>(
						aComb.apply(p1.getA(), p2.getA()),
						bComb.apply(p1.getB(), p2.getB())
				);
			}

			@Override
			public Function<StreamUtil.Pair<A1, A2>, StreamUtil.Pair<R1, R2>> finisher() {
				final Function<A1, R1> aFinish = a.finisher();
				final Function<A2, R2> bFinish = b.finisher();
				return (p) -> new StreamUtil.Pair<>(
						aFinish.apply(p.getA()),
						bFinish.apply(p.getB())
				);
			}

			@Override
			public Set<Characteristics> characteristics() {
				final Set<Characteristics> aChar = a.characteristics();
				final Set<Characteristics> bChar = b.characteristics();

				Set<Characteristics> res = EnumSet.noneOf(Characteristics.class);

				if (aChar.contains(Characteristics.IDENTITY_FINISH) && bChar.contains(Characteristics.IDENTITY_FINISH)) {
					res.add(Characteristics.IDENTITY_FINISH);
				}
				if (aChar.contains(Characteristics.CONCURRENT) && bChar.contains(Characteristics.CONCURRENT)) {
					res.add(Characteristics.CONCURRENT);
				}
				if (aChar.contains(Characteristics.UNORDERED) || bChar.contains(Characteristics.UNORDERED)) {
					res.add(Characteristics.UNORDERED);
				}

				return res;
			}
		};
	}

	/**
	 * Returns a {@code Collector} that accumulates the input elements into a
	 * new {@code Set}. There are no guarantees on the type, mutability,
	 * serializability, or thread-safety of the {@code Set} returned; if more
	 * control over the returned {@code Set} is required, use
	 * {@link #toCollection(Supplier)}.
	 *
	 * <p>This is an {@link Collector.Characteristics#UNORDERED unordered}
	 * Collector.
	 *
	 * @param <T> the type of the input elements
	 * @return a {@code Collector} which collects all the input elements into a
	 * {@code Set}
	 */
	public static <T>
	Collector<Optional<T>, Set<T>, Set<T>> toSetIfPresent() {
		return new Collector<Optional<T>, Set<T>, Set<T>>() {
			@Override
			public Supplier<Set<T>> supplier() {
				return HashSet::new;
			}

			@Override
			public BiConsumer<Set<T>, Optional<T>> accumulator() {
				return (set,e)->e.ifPresent(set::add);
			}

			@Override
			public BinaryOperator<Set<T>> combiner() {
				return (a,b)-> {
					a.addAll(b);
					return a;
				};
			}

			@Override
			public Function<Set<T>, Set<T>> finisher() {
				return Function.identity();
			}

			@Override
			public Set<Characteristics> characteristics() {
				return EnumSet.of(Characteristics.IDENTITY_FINISH, Characteristics.UNORDERED, Characteristics.CONCURRENT);
			}
		};

	}

	public static class FirstAndRest<T> {
		private final T first;
		private final Collection<T> rest;

		public FirstAndRest(T first, Collection<T> rest) {
			this.first = first;
			this.rest = rest;
		}

		public T getFirst() {
			return first;
		}

		public Collection<T> getRest() {
			return rest;
		}
	}

	private static class MutableFirstAndRest<T> {
		private T first;
		private Collection<T> rest;
	}

	public static <T> Collector<T, ?, FirstAndRest<T>> collectFirstAndRest() {
		return Collectors.collectingAndThen(
				Collectors.partitioningBy(isFirstPredicate()),
				map -> {
					List<T> firsts = map.get(true);
					if (firsts==null) {
						throw new IllegalStateException("Stream was empty");
					}
					return new FirstAndRest<>(firsts.get(0), map.get(false));
				}
		);
	}


	public static <T> Predicate<T> isFirstPredicate() {
		final boolean[] first = {true};
		return x -> {
			boolean firstCopy = first[0];
			first[0] = false;
			return firstCopy;
		};
	}
}
