/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Optional;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

/**
 * Match Search Strings
 *
 * The string is split into parts separated either by punctuation or by Camel Case. The search string may abbreviate
 * these parts.
 *
 * Similar to how IntelliJ's search works.
 */
public final class StringSearchMatcher {

	private StringSearchMatcher() {
	}

	/**
	 * State while generating tokens
	 */
	enum TokenState {
		Empty,
		LowerCase,
		UpperCase,
		Number
	}


	/**
	 * A Token
	 */
	static class Token {

		/**
		 * Offset of first character within sourcce
		 */
		public final int offset;
		/**
		 * Content. Always converted to lowercase
		 */
		public final String content;
		/**
		 * True if in the source text, the token was all uppercase
		 */
		public final boolean onlyUppercase;

		/**
		 * Create token
		 * @param offset the offset
		 * @param content the content
		 * @param onlyUppercase true if all characters in the source text were in range [A-Z]
		 */
		public Token(int offset, String content, boolean onlyUppercase) {
			this.offset = offset;
			this.content = content;
			this.onlyUppercase = onlyUppercase;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (!(o instanceof Token)) {
				return false;
			}

			Token token = (Token) o;

			if (offset != token.offset) {
				return false;
			}
			if (onlyUppercase != token.onlyUppercase) {
				return false;
			}
			return content.equals(token.content);
		}

		@Override
		public int hashCode() {
			int result = offset;
			result = 31 * result + content.hashCode();
			result = 31 * result + (onlyUppercase? 1 : 0);
			return result;
		}

		@Override
		public String toString() {
			return "{" + offset + "," + content + "," + onlyUppercase + "}";
		}
	}

	/**
	 * Split the input at non-alphanumerical characters
	 * @param input the input text
	 * @return stream of split parts. represented as pair of character offset and content
	 */
	private static Stream<StreamUtil.IntPair<String>> splitAtNonAlphaNum(String input) {
		final Stream.Builder<StreamUtil.IntPair<String>> resBuilder = Stream.builder();
		StringBuilder strBuilder = null;
		int firstChar = 0;
		for (int i = 0; i < input.length(); i++) {
			char c = input.charAt(i);
			final int type = Character.getType(c);
			boolean reg = Character.toString(c).matches("[A-Za-z0-9]");
			if (reg) {
				if (strBuilder == null) {
					strBuilder = new StringBuilder();
					firstChar = i;
				}
				strBuilder.append(c);
			} else {
				if (strBuilder != null) {
					resBuilder.accept(new StreamUtil.IntPair<>(firstChar, strBuilder.toString()));
					strBuilder = null;
				}
			}
		}
		if (strBuilder != null) {
			resBuilder.accept(new StreamUtil.IntPair<>(firstChar, strBuilder.toString()));
		}
		return resBuilder.build();
	}

	/**
	 * Split a string into its separate match parts
	 *
	 * @param input the input
	 * @return the input split into its match parts, in lowercase
	 */
	static Stream<Token> tokenize(String input) {
		if (input == null) {
			return Stream.empty();
		}
		return splitAtNonAlphaNum(input)
				.flatMap(StreamUtil.IntPair.accessF((Integer begin, String s) -> {
					boolean hasLowerCase = s.matches(".*[A-Z].*[a-z].*");

					final Stream.Builder<Token> builder = Stream.builder();
					StringBuilder stringBuilder = new StringBuilder();
					int firstChar = begin;

					TokenState state = TokenState.Empty;

					boolean tokenHasOnlyUppercase = true;

					for (int i = 0; i < s.length(); i++) {
						char c = s.charAt(i);
						TokenState newState;
						if (c >= 'A' && c <= 'Z') {
							newState = TokenState.UpperCase;
						} else if (c >= 'a' && c <= 'z') {
							newState = TokenState.LowerCase;
						} else if (c >= '0' && c <= '9') {
							newState = TokenState.Number;
						} else {
							throw new RuntimeException("Invalid character: " + c);
						}

						boolean finishPart;

						switch (state) {
							case Empty:
								finishPart = false;
								break;
							case UpperCase:
								if (newState == TokenState.UpperCase) {
									finishPart = hasLowerCase;
								} else {
									finishPart = newState == TokenState.Number;
								}
								break;
							case LowerCase:
								finishPart = newState != TokenState.LowerCase;
								break;
							case Number:
								finishPart = newState != TokenState.Number;
								break;
							default:
								throw new RuntimeException("Invalid state");
						}
						if (finishPart) {
							builder.accept(new Token(firstChar, stringBuilder.toString(), state == TokenState.UpperCase));
							stringBuilder = new StringBuilder();
							firstChar = i + begin;
							tokenHasOnlyUppercase = true;
						}
						state = newState;
						tokenHasOnlyUppercase = newState == TokenState.UpperCase;
						if (newState == TokenState.UpperCase) {
							c -= 'A' - 'a'; //Make lower case
						}
						stringBuilder.append(c);
					}
					builder.accept(new Token(firstChar, stringBuilder.toString(), tokenHasOnlyUppercase));

					return builder.build();
				}));
	}

	private static Function<List<Token>, List<Token>> createChangedList(BiConsumer<List<Token>, List<Token>> handler) {
		return (l) -> {
			final ArrayList<Token> res = new ArrayList<>();
			handler.accept(l, res);
			return res;
		};
	}

	/**
	 * Returns a function that prepends the given token to lists.
	 * @param prepend the token to prepend
	 * @return function that takes a list and returns a new list with the given token prepended
	 */
	private static Function<List<Token>, List<Token>> prepend(Token prepend) {
		return createChangedList((in, res) -> {
			res.add(prepend);
			res.addAll(in);
		});
	}

	/**
	 * Returns a function that prepends the given tokens to lists.
	 * @param prepend the tokens to prepend
	 * @return function that takes a list and returns a new list with the given tokens prepended
	 */
	private static Function<List<Token>, List<Token>> prependAll(List<Token> prepend) {
		return createChangedList((in, res) -> {
			res.addAll(prepend);
			res.addAll(in);
		});
	}

	/**
	 * Returns a function that appends the given token to lists.
	 * @param append the token to append
	 * @return function that takes a list and returns a new list with the given token appended
	 */
	private static Function<List<Token>, List<Token>> append(Token append) {
		return createChangedList((in, res) -> {
			res.addAll(in);
			res.add(append);
		});
	}

	/**
	 * Returns a function that appends the given tokens to lists.
	 * @param append the tokens to append
	 * @return function that takes a list and returns a new list with the given tokens appended
	 */
	private static Function<List<Token>, List<Token>> appendAll(List<Token> append) {
		return createChangedList((in, res) -> {
			res.addAll(in);
			res.addAll(append);
		});
	}


	/**
	 * From a single token that may be optionally split, return all possible ways to split it
	 * @param token the token to split
	 * @return  a stream of all possible ways to split the token, represented as list of subtokens
	 */
	static Stream<List<Token>> generateOptionalSplits(Token token) {
		final Stream<List<Token>> withoutSplit = Stream.of(Collections.singletonList(token));

		if (token.content.length() == 1) {
			return withoutSplit;
		}

		final Stream<List<Token>> withSplit = IntStream.range(1, token.content.length())
					.mapToObj(i -> i)
					.flatMap(i -> {
						Token start = new Token(token.offset, token.content.substring(0, i), token.onlyUppercase);
						Token rest = new Token(token.offset + i, token.content.substring(i), token.onlyUppercase);
						return generateOptionalSplits(rest).map(prepend(start));
					});

		return Stream.concat(withoutSplit, withSplit);
	}

	/**
	 * If the given token may be split in a needle, return a function that appends all possible splits to a list
	 * of tokens. Otherwise, return a function that appends the token,
	 * @param token the token to append
	 * @return a function that appends the token to a list, split if possibke
	 */
	private static Function<List<Token>, Stream<List<Token>>> appendMaybeSplittedToken(Token token) {
		if (token.onlyUppercase && token.content.length() > 1) {
			//Save in list so we only have to calculate it once
			final List<List<Token>> optSplits = generateOptionalSplits(token).collect(Collectors.toList());
			return (before) -> optSplits.stream().map(prependAll(before));
		} else {
			return append(token).andThen(Stream::of);
		}
	}

	/**
	 * In needles, tokens with only uppercase characters may be split into individual tokens. Return all possible
	 * tokenizations of the needle
	 *
	 * @param needle the needle
	 * @return all possible tokenizations
	 */
	static Stream<List<Token>> tokenizeNeedle(String needle) {

		return tokenize(needle).reduce(
				Stream.<List<Token>>of(new ArrayList<>()),
				(stream, token) -> stream.flatMap(appendMaybeSplittedToken(token)),
				(streamA, streamB) -> {
					throw new UnsupportedOperationException("Cannot combine");
				}
		);
	}

	/**
	 * A match between the haystack and the needle
	 */
	public static class Match {

		/**
		 * Start index of the match in the haystack
		 */
		public final int startIndex;
		/**
		 * Length of the match
		 */
		public final int length;

		public Match(int startIndex, int length) {
			this.startIndex = startIndex;
			this.length = length;
		}

		@Override
		public String toString() {
			return "{" + startIndex + "," + length + '}';
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (!(o instanceof Match)) {
				return false;
			}

			Match match = (Match) o;

			if (startIndex != match.startIndex) {
				return false;
			}
			return length == match.length;
		}

		@Override
		public int hashCode() {
			int result = startIndex;
			result = 31 * result + length;
			return result;
		}

		public static final Match INVALID = new Match(-1, 1);
	}

	/**
	 * Match two tokenizations
	 * @param haystack the tokenized haystack
	 * @param needle the tokenized needle
	 * @return Optional.empty() if there is no match, else a Match for each needle token
	 */
	private static Optional<List<Match>> matchTokenized(List<Token> haystack, List<Token> needle) {
		final Iterator<Token> haystackTokens = haystack.iterator();

		final boolean[] valid = {true};
		final List<Match> res = needle.stream().map(t -> t.content).map(token -> {
			while (true) {
				if (!haystackTokens.hasNext()) {
					valid[0] = false;
					return Match.INVALID;
				}

				Token haystackToken = haystackTokens.next();
				if (haystackToken.content.startsWith(token)) {
					return new Match(haystackToken.offset, token.length());
				}
			}
		}).collect(Collectors.toList());
		if (valid[0]) {
			return Optional.of(res);
		} else {
			return Optional.empty();
		}
	}

	/**
	 * Find a needle in a haystack, returning the positions of the match parts.
	 *
	 * Uses CamelCase-Matching like intellij, so searching like this is possible:
	 * Haystack: "SpartanMC", Needle: "SpMC"
	 *
	 * @param haystack the string to search in
	 * @param needle the string to search
	 * @return Optional.empty() if there is no match, else a Match for each needle token
	 */
	public static Optional<List<Match>> matchPositions(String haystack, String needle) {
		final List<Token> haystackTokens = tokenize(haystack).collect(Collectors.toList());

		return tokenizeNeedle(needle)
				.flatMap(n -> StreamUtil.streamOptional(matchTokenized(haystackTokens, n)))
				.findFirst();
	}

	/**
	 * Find a needle in a haystack, returning the positions of the match parts.
	 *
	 * Uses CamelCase-Matching like intellij, so searching like this is possible:
	 * Haystack: "SpartanMC", Needle: "SpMC"
	 *
	 * @param haystack the string to search in
	 * @param needle the string to search
	 * @return true if the string matches
	 */
	public static boolean match(String haystack, String needle) {
		return matchPositions(haystack, needle).isPresent();
	}
}