/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.util

import spock.lang.Shared
import spock.lang.Specification
import spock.lang.Unroll

import java.util.stream.Collectors

class StringSearchMatcherTest extends Specification {
	//Some short names for repeated constructors
	def t(int offs, String content, boolean onlyUppercase) {
		return new StringSearchMatcher.Token(offs, content, onlyUppercase);
	}

	def t(int offs, String content) {
		return t(offs, content, true);
	}

	def m(int start, int length) {
		return new StringSearchMatcher.Match(start, length);
	}

	def o(List<StringSearchMatcher.Match> m) {
		return Optional.of(m);
	}

	def "Tokenize"(String input, List<String> out) {
		expect:
		StringSearchMatcher.tokenize(input).collect(Collectors.toList()) == out

		where:
		input                  | out
		""                     | []
		"a"                    | [t(0, "a", false)]
		"a.b"                  | [t(0, "a", false), t(2, "b", false)]
		"a./b"                 | [t(0, "a", false), t(3, "b", false)]
		"aB"                   | [t(0, "a", false), t(1, "b", true)]
		"aBa"                  | [t(0, "a", false), t(1, "ba", false)]
		"aBC"                  | [t(0, "a", false), t(1, "bc", true)]
		"/spartanmc_0/intr_in" | [t(1, "spartanmc", false), t(11, "0", false), t(13, "intr", false), t(18, "in", false)]
		"/spartanmc0/intrIn"   | [t(1, "spartanmc", false), t(10, "0", false), t(12, "intr", false), t(16, "in", false)]
		"#USB_1_RX"            | [t(1, "usb", true), t(5, "1", false), t(7, "rx", true)]
		"#USB1RX"              | [t(1, "usb", true), t(4, "1", false), t(5, "rx", true)]
		"HelloWorld"           | [t(0, "hello", false), t(5, "world", false)]
		"HELLoWORLd"           | [t(0, "h", true), t(1, "e", true), t(2, "l", true), t(3, "lo", false), t(5, "w", true), t(6, "o", true), t(7, "r", true), t(8, "ld", false)]
	}

	def "Optional Splits"(StringSearchMatcher.Token input, List<List<StringSearchMatcher.Token>> out) {
		expect:
		StringSearchMatcher.generateOptionalSplits(input).collect(Collectors.toList()) == out

		where:
		input       | out
		t(0, "asd") | [[t(0, "asd")], [t(0, "a"), t(1, "sd")], [t(0, "a"), t(1, "s"), t(2, "d")], [t(0, "as"), t(2, "d")]]

	}

	//Moved out of TokenizeNeedle, because we can't use line breaks inside the table
	@Shared
	def tokenizedUSB1RX = [
			[t(1, "usb"), t(5, "1", false), t(7, "rx")],
			[t(1, "usb"), t(5, "1", false), t(7, "r"), t(8, "x")],
			[t(1, "u"), t(2, "sb"), t(5, "1", false), t(7, "rx")],
			[t(1, "u"), t(2, "sb"), t(5, "1", false), t(7, "r"), t(8, "x")],
			[t(1, "u"), t(2, "s"), t(3, "b"), t(5, "1", false), t(7, "rx")],
			[t(1, "u"), t(2, "s"), t(3, "b"), t(5, "1", false), t(7, "r"), t(8, "x")],
			[t(1, "us"), t(3, "b"), t(5, "1", false), t(7, "rx")],
			[t(1, "us"), t(3, "b"), t(5, "1", false), t(7, "r"), t(8, "x")]
	]

	def "TokenizeNeedle"(String input, List<String> out) {
		expect:
		StringSearchMatcher.tokenizeNeedle(input).collect(Collectors.toList()) == out

		where:
		input                  | out
		""                     | [[]]
		"a"                    | [[t(0, "a", false)]]
		"a.b"                  | [[t(0, "a", false), t(2, "b", false)]]
		"a./b"                 | [[t(0, "a", false), t(3, "b", false)]]
		"aB"                   | [[t(0, "a", false), t(1, "b", true)]]
		"aBC"                  | [[t(0, "a", false), t(1, "bc", true)], [t(0, "a", false), t(1, "b", true), t(2, "c", true)]]
		"aBa"                  | [[t(0, "a", false), t(1, "ba", false)]]
		"/spartanmc_0/intr_in" | [[t(1, "spartanmc", false), t(11, "0", false), t(13, "intr", false), t(18, "in", false)]]
		"/spartanmc0/intrIn"   | [[t(1, "spartanmc", false), t(10, "0", false), t(12, "intr", false), t(16, "in", false)]]
		"#USB_1_RX"            | tokenizedUSB1RX
		"HelloWorld"           | [[t(0, "hello", false), t(5, "world", false)]]
		"ASD"                  | [[t(0, "asd")], [t(0, "a"), t(1, "sd")], [t(0, "a"), t(1, "s"), t(2, "d")], [t(0, "as"), t(2, "d")]]

	}

	@Unroll
	def "Searching for \"#needle\" in \"#haystack\""(String haystack, String needle, Optional<List<StringSearchMatcher.Match>> out) {
		expect:
		StringSearchMatcher.matchPositions(haystack, needle) == out

		where:
		haystack              | needle        | out
		""                    | ""            | o([])
		"asdf"                | ""            | o([])
		"a"                   | "a"           | o([m(0, 1)])
		"a"                   | "b"           | Optional.empty()
		"HelloWorld"          | "Hello"       | o([m(0, 5)])
		"HelloWorld"          | "hello"       | o([m(0, 5)])
		"HelloWorld"          | "hElloWorld"  | Optional.empty()
		"HelloWorld"          | "helloworld"  | Optional.empty()
		"HelloTheWorld"       | "hellWorld"   | o([m(0, 4), m(8, 5)])
		"HelloTheWorld"       | "theWorld"    | o([m(5, 3), m(8, 5)])
		"HelloHellowTheWorld" | "HellowWorld" | o([m(5, 6), m(14, 5)])
		"HHTWA"               | "HHTW"        | o([m(0, 4)])
		"HHHTW"               | "HHTW"        | Optional.empty()
		"HelloHellowTheWorld" | "HHTW"        | o([m(0, 1), m(5, 1), m(11, 1), m(14, 1)])
		"HelloHellowTheWorld" | "TW"          | o([m(11, 1), m(14, 1)])
		"ASDF_JaKsLd"         | "ASDFJKL"     | o([m(0, 4), m(5, 1), m(7, 1), m(9, 1)])
		"AaSsDdFf_JKL"        | "ASDFJKL"     | o([m(0, 1), m(2, 1), m(4, 1), m(6, 1), m(9, 3)])
		"ASDF"                | "ASF"         | Optional.empty()
		"#PHY_MDC"            | "phyMD"       | o([m(1, 3), m(5, 2)])
		"#PHY_MDC"            | "#phyMD"      | o([m(1, 3), m(5, 2)])
		"Nexys-Video"         | "Nexys-Video" | o([m(0, 5), m(6, 5)])
	}
}
