/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.loopOptimizer.model;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.interfaces.AccessesVariables;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.variable.TaskVariable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;

import cetus.hir.Traversable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.stream.Collectors;

/**
 * A generic class containing a traversable and variables
 */
public class GenericAccessesVariables implements AccessesVariables {

	Collection<VariableAccess> variableAccesses = new LinkedHashSet<>();
	InnerFlowGraph innerFlowGraph = null;
	Traversable traversable = null;
	private Logger log = LoggerFactory.getLogger(GenericAccessesVariables.class);
	private Collection<TaskVariable> variables = new LinkedHashSet<>();

	public GenericAccessesVariables(Traversable traversable) {
		this.traversable = traversable;
	}

	/**
	 * Adds a variable access to the Task for dependency calculation
	 *
	 * @param access the variable access to be added
	 * @return true if the variable access could be added, false otherwise
	 */
	public boolean addVariableAccess(VariableAccess access) {
		TaskVariable taskVariable = getTaskVariable(access.getVariable());

		if (null == taskVariable) {
			log.debug("Accessed Variable not stored in this task.");
			addVariable(access.getVariable());
			taskVariable = getTaskVariable(access.getVariable());
		}

		return taskVariable.addAccess(access);
	}

	@Override
	public Collection<Variable> getVariables() {
		return variables.stream()
				.filter(v -> v.getVariable() != null)
				.map(v -> v.getVariable())
				.collect(Collectors.toSet());
	}

	/**
	 * add a variable
	 *
	 * @param variable a variable to be added
	 * @return true if the variable could be added
	 */
	public boolean addVariable(Variable variable) {
		TaskVariable taskVariable = getTaskVariable(variable);

		if (null == taskVariable) {
			taskVariable = new TaskVariable(variable);
			return variables.add(taskVariable);
		}
		return false;
	}

	/**
	 * add all variables
	 * @param access all accesses to be added
	 * @return true if all was successfully added
	 */
	public boolean addAllVariableAccess(Collection<VariableAccess> access) {
		boolean allGood = true;
		for (VariableAccess va : access) {
			if (!addVariableAccess(va)) {
				allGood = false;
			}
		}
		return allGood;
	}

	@Override
	public Traversable getTraversable() {
		return traversable;
	}

	@Override
	public void setTraversable(Traversable traversable) {
		this.traversable = traversable;
	}

	@Override
	public Model getModel() {
		return null;
	}

	@Override
	public void setModel(Model m) {

	}

	@Override
	public InnerFlowGraph getInnerFlowGraph() {
		return innerFlowGraph;
	}

	@Override
	public void setInnerFlowGraph(InnerFlowGraph graph) {
		innerFlowGraph = graph;
	}

	private TaskVariable getTaskVariable(Variable variable) {
		for (TaskVariable taskVariable : variables) {
			if (taskVariable.getVariable().equals(variable)) {
				return taskVariable;
			}
		}
		return null;
	}

	/**
	 * Returns all written variables
	 *
	 * @return written variables
	 */
	public Collection<Variable> getWriteVariables() {
		Collection<Variable> writtenVariables = new ArrayList<Variable>();

		for (TaskVariable variable : variables) {
			if (variable.isWritten()) {
				writtenVariables.add(variable.getVariable());
			}
		}

		return writtenVariables;
	}
}
