/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.loopOptimizer.model;

import cetus.hir.AnnotationStatement;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * A list of statements...it is a list with some easier modification methods.
 */
public class StatementList {

	private Set<StatementDependencyProperties> statementDependencyPropertiesList = new HashSet<>();

	private StatementDependencyProperties initialStatement = null;

	public StatementList(StatementDependencyProperties propertyDefiningRange, Set<StatementDependencyProperties> possibleOtherPropertiesToBeAdded) {
		addAllWithinRangeof(propertyDefiningRange, possibleOtherPropertiesToBeAdded);
	}

	public StatementList() {

	}

	public boolean add(StatementDependencyProperties statementDependencyProperties) {
		return statementDependencyPropertiesList.add(statementDependencyProperties);
	}

	public boolean remove(StatementDependencyProperties statementDependencyProperties) {
		return statementDependencyPropertiesList.remove(statementDependencyProperties);
	}

	public List<StatementDependencyProperties> getStatements() {
		return statementDependencyPropertiesList.stream().sorted(Comparator.comparingInt(StatementDependencyProperties::getPosition)).collect(Collectors.toList());
	}

	public boolean addAll(StatementList otherList) {
		return statementDependencyPropertiesList.addAll(otherList.getStatements());
	}

	public boolean hasNoStatements() {
		return statementDependencyPropertiesList.isEmpty();
	}


	/**
	 * Adds all that are within the propertyDefiningRage rage
	 * @param propertyDefiningRange the property defining the rage of this statement
	 * @param possibleOtherPropertiesToBeAdded possible candidates to be added
	 */
	public void addAllWithinRangeof(StatementDependencyProperties propertyDefiningRange, Set<StatementDependencyProperties> possibleOtherPropertiesToBeAdded) {
		initialStatement = propertyDefiningRange;
		statementDependencyPropertiesList.add(propertyDefiningRange);
		for (StatementDependencyProperties possibleProperty : possibleOtherPropertiesToBeAdded) {
			final int currentPosition = possibleProperty.getPosition();
			if (propertyDefiningRange != possibleProperty && currentPosition >= propertyDefiningRange.getPosition() && currentPosition <= propertyDefiningRange.getLastDepentStatementPostion()) {
				statementDependencyPropertiesList.add(possibleProperty);
			}
		}
	}

	/**
	 * Find duplicate statements
	 * @param otherStatementList the statementlist to find duplicates in
	 * @return a list of duplicates
	 */
	public List<StatementDependencyProperties> getDuplicates(StatementList otherStatementList) {
		List<StatementDependencyProperties> duplicates = new ArrayList<>();
		duplicates.addAll(statementDependencyPropertiesList);
		duplicates.retainAll(otherStatementList.getStatements());
		return duplicates;
	}

	public boolean hasDuplicates(StatementList otherStatementList) {
		return !getDuplicates(otherStatementList).isEmpty();
	}

	public boolean hasOnlyAnnotations() {
		return statementDependencyPropertiesList.stream().allMatch(p -> p.getStatement() instanceof AnnotationStatement);
	}

	public int maxPosition() {
		return statementDependencyPropertiesList.stream().max(Comparator.comparingInt(StatementDependencyProperties::getPosition)).get().getPosition();
	}

	public int minPosition() {
		return statementDependencyPropertiesList.stream().min(Comparator.comparingInt(StatementDependencyProperties::getPosition)).get().getPosition();
	}

	public int size() {
		return statementDependencyPropertiesList.size();
	}

	public String toString() {
		return "[" + initialStatement + "] -> " + getStatements().toString();
	}


}
