/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.loopOptimizer.tools;


import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.loopOptimizer.model.GenericAccessesVariables;
import de.tu_darmstadt.rs.microStreams.loopOptimizer.model.StatementDependencyProperties;
import de.tu_darmstadt.rs.microStreams.loopOptimizer.model.StatementList;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraphFractory;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowNode;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.LoopPragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.PragmaTag;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.tools.VariableAccessAnalyzer;
import de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.peripheralDetection.PeripheralDetector;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.InnerFlowLinker;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.AnnotationStatement;
import cetus.hir.ArrayAccess;
import cetus.hir.AssignmentExpression;
import cetus.hir.AssignmentOperator;
import cetus.hir.CompoundStatement;
import cetus.hir.Declaration;
import cetus.hir.DeclarationStatement;
import cetus.hir.DepthFirstIterator;
import cetus.hir.ExpressionStatement;
import cetus.hir.ForLoop;
import cetus.hir.Identifier;
import cetus.hir.Loop;
import cetus.hir.NameID;
import cetus.hir.Statement;
import cetus.hir.SymbolTools;
import cetus.hir.Traversable;
import cetus.hir.UnaryExpression;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Helper providing many different methods needed for fissioning loops
 * Created by dominikl on 11.07.17.
 */
class FissionHelper {

	private static final Logger log = LoggerFactory.getLogger(FissionHelper.class);
	private final Model model;

	public FissionHelper(Model model) {
		this.model = model;
	}

	/**
	 * Splits one CompoundStatement (Body of an Loop) into multiple CompoundStatements that are all independent from each other
	 * Tries to fulfill the requirements made in the LoopPragma
	 * Uses Autoperf Data when available
	 *
	 * @param comp initial body of the loop
	 * @param pragma pragma belonging to the loop
	 * @param autoperf autoperf data belonging to the loop, not necessary
	 * @param iterationVariable Identifier that is used as iteration Variable, as it can be ignored when finding dependencies
	 * @param peripheralDetector instance of the PeripheralDetecttor used for finding periphery. Null if no periphery shall be searched
	 * @return ArrayList of several CompoundStatements that are independent from each other
	 */
	public ArrayList<CompoundStatement> getIndependentCompounds(CompoundStatement comp, LoopPragma pragma, ArrayList<ArrayList<Integer>> autoperf, Identifier iterationVariable, PeripheralDetector peripheralDetector) {
		//This method processes the CompoundStatement comp and extracts its Statements into blocks of independent HashMaps
		//(one Hashmap is one block of statements that are dependent on each other)
		ArrayList<StatementList> statementMaps = getIndependentHashMaps(comp, iterationVariable, peripheralDetector);
		if (statementMaps == null) {
			return null;
		}

		PragmaTag<Integer> numTag = pragma.getPragmaTag(Constants.PragmaLoop.NUMBER_OF_SPLITS);
		Integer nos;
		if (numTag != null) {
			nos = (Integer) numTag.getEntryNames().toArray()[0];
		} else {
			throw new RuntimeException("Could not evaluate Number of Splits. This Should not happen.");
		}

		//if the size of statementMaps is bigger than nos, the two smallest elements of statementMaps
		//have to be collapsed / put together until nos = the size of statementMaps
		if (nos < statementMaps.size()) {
			if (autoperf == null) {
				reduceListSize(statementMaps, nos);
			} else {
				reduceListSizeWithAutoPerf(statementMaps, nos, autoperf);
			}
		} else if (nos > statementMaps.size()) {
			log.warn("Could not split the Body of the Loop  into the desired " + nos + " parts. Maximum " + statementMaps.size() + " parts are possible due to dependencies ...this value will be used now!\n" + comp.getParent().toString());
		}

		//put all Traversables from each inner map of statementMaps into one CompoundStatement
		ArrayList<CompoundStatement> compounds = new ArrayList<>();
		for (StatementList l : statementMaps) {
			CompoundStatement newCompound = new CompoundStatement();
			for (StatementDependencyProperties i : l.getStatements()) {
				Traversable t = i.getStatement();
				if (t instanceof Declaration) {
					newCompound.addDeclaration(((Declaration) t).clone());
				} else if (t instanceof DeclarationStatement) {
					newCompound.addDeclaration(((Declaration) t.getChildren().get(0)).clone());
				} else if (t instanceof Statement) {
					newCompound.addStatement(((Statement) t).clone());
				} else {
					throw new RuntimeException("Traversable could not be added to the CompoundStatement for the statementMaps. This should not happen.");
				}
			}
			compounds.add(newCompound);
		}

		return compounds;
	}

	/**
	 * returns the size of the list from getIndependentHashMapsWithReorder() for testing out how many blocks can be seperated by fission
	 *
	 * @param comp the compound to test (body of loop)
	 * @param iterationVariable the iteration variable of the loop
	 * @param peripheralDetector instance of the PeripheralDetecttor used for finding periphery. Null if no periphery shall be searched
	 * @return int as size
	 */
	public int getNumberOfIndependetBlocks(CompoundStatement comp, Identifier iterationVariable, PeripheralDetector peripheralDetector) {
		return getIndpendentHashMapsWithReorder(comp, iterationVariable, peripheralDetector).size();
	}


	/**
	 * This method processes the CompoundStatement comp and extracts its Statements into blocks of independent HashMaps
	 * (one Hashmap is one block of statements that are dependent on each other)
	 *
	 * @param comp the compound with statements to process (typically body of a loop)
	 * @param iterationVariable the iteration variable as writes on this have to be in every block
	 * @param peripheralDetector instance of the PeripheralDetecttor used for finding periphery. Null if no periphery shall be searched
	 * @return a list of HashMaps where every HashMap is independent from the others
	 */
	private ArrayList<StatementList> getIndependentHashMaps(CompoundStatement comp, Identifier iterationVariable, PeripheralDetector peripheralDetector) {

		List<Variable> allVariables = new ArrayList<>();
		final InnerFlowGraph innerFlowGraph = InnerFlowGraphFractory.getControlFlowWithoutTaskHandling(comp);
		GenericAccessesVariables loop = new GenericAccessesVariables(comp);
		loop.setInnerFlowGraph(innerFlowGraph);
		final VariableAccessAnalyzer variableAccessAnalyzer = new VariableAccessAnalyzer(model);
		List globalVariables = new ArrayList();
		final List<VariableAccess> variableAccesses = variableAccessAnalyzer.scanForIdentifier(loop, globalVariables, allVariables);
		variableAccessAnalyzer.setAccessBehaviour(variableAccesses);
		for (VariableAccess variableAccess : variableAccesses) {
			InnerFlowLinker.link(variableAccess, innerFlowGraph);
		}
		loop.addAllVariableAccess(variableAccesses);

		ArrayList<StatementList> statementMaps = new ArrayList<>();

		HashMap<Traversable, Integer> traversablePositionLookup = new HashMap();
		final Set<PeripheralModuleDescription> allPeripherals = peripheralDetector.findUsedPeripherals(comp);
		Set<StatementDependencyProperties> statementDependencyProperties = new HashSet<>();


		int count = 1;
		for (Traversable statement : comp.getChildren()) {
			traversablePositionLookup.put(statement, new Integer(count++));
		}


		for (Traversable statement : comp.getChildren()) {
			List<InnerFlowNode> statementNodes = innerFlowGraph.getInnerFlowNodesContaining(statement);
			for (InnerFlowNode statementNode : statementNodes) {
				if (statement != comp && statement instanceof Statement && statementNode != null) {
					Traversable lastTraversableWithDependencies = statement;
					final Collection<Variable> writtenVariables = statementNode.getWrittenVariables();
					//add all traversables containing reads
					List<Traversable> traversablesContainingDependencies = new ArrayList<>();
					for (Variable writtenVariable : writtenVariables) {
						for (InnerFlowNode successor : innerFlowGraph.getAllSuccessors(statementNode)) {
							if (successor.getReadVariables().contains(writtenVariable)) {
								traversablesContainingDependencies.add(successor.getTraversable());
							} else if (successor.getWrittenVariables().contains(writtenVariable)) { //write after write is not a dependency
								break;
							}
						}
					}
					//add peripheral dependencies
					if (!allPeripherals.isEmpty()) {
						final Set<PeripheralModuleDescription> peripheralsInStatement = peripheralDetector.findUsedPeripherals(statement);
						for (PeripheralModuleDescription peripheral : peripheralsInStatement) {
							for (InnerFlowNode successor : innerFlowGraph.getAllSuccessors(statementNode)) {
								final Set<PeripheralModuleDescription> otherPeripherals = peripheralDetector.findUsedPeripherals(successor.getTraversable());
								if (otherPeripherals.stream().anyMatch(p -> p.getName().equalsIgnoreCase(peripheral.getName()))) {
									traversablesContainingDependencies.add(successor.getTraversable());
								} else if (!allPeripherals.stream().anyMatch(p -> peripheral.getName().equalsIgnoreCase(p.getName()))) {
									log.error("It seems like there were falsely detected peripherals. Detected peripherals in the codeline: " + successor.getTraversable() + " -> " + otherPeripherals + ", detected peripherals in the loop: " + comp + " -> " + allPeripherals);
								}
							}
						}
					}


					//find last traversable with dependencies
					for (Traversable next : comp.getChildren()) {
						if (traversablesContainingDependencies.isEmpty()) {
							break;
						}
						if (next instanceof Statement) {
							lastTraversableWithDependencies = next;
							traversablesContainingDependencies.remove(lastTraversableWithDependencies);
						}
					}
					StatementDependencyProperties property = new StatementDependencyProperties(statement, lastTraversableWithDependencies, traversablePositionLookup.get(statement), traversablePositionLookup.get(lastTraversableWithDependencies));
					final boolean alreadyInList = statementDependencyProperties.stream().anyMatch(p -> p.getStatement().equals(statement));
					if (!alreadyInList) {
						statementDependencyProperties.add(property);
					}
				}
			}
		}


		for (StatementDependencyProperties property : statementDependencyProperties) {
			StatementList statementList = new StatementList(property, statementDependencyProperties);
			statementMaps.add(statementList);
		}

		//merge lists

		List<StreamUtil.Pair<StatementList, StatementList>> mergeList = new ArrayList<>();
		for (StatementList currentMap : statementMaps) {
			for (StatementList searchMap : statementMaps) {
				if (currentMap == searchMap) {
					break;
				} else {
					boolean merge = currentMap.hasDuplicates(searchMap);
					if (merge) {
						boolean alreadyInMergeList = mergeList.stream().anyMatch(p -> (p.getA() == searchMap && p.getB() == currentMap) || (p.getB() == searchMap && p.getA() == currentMap));
						if (!alreadyInMergeList) {
							if (searchMap.minPosition() < currentMap.minPosition()) {
								mergeList.add(new StreamUtil.Pair<>(searchMap, currentMap));
							} else {
								mergeList.add(new StreamUtil.Pair<>(currentMap, searchMap));
							}
						}
					}
				}
			}
			//Merge maps with only annotations to the successor if there is one otherwise to predecessor
			if (currentMap.hasOnlyAnnotations()) {
				final Optional<StatementList> closestSuccessor = statementMaps.stream()
						.filter(s -> s.minPosition() > currentMap.maxPosition())
						.min(Comparator.comparingInt(StatementList::minPosition));

				final Optional<StatementList> closestPredecessor = statementMaps.stream()
						.filter(s -> s.maxPosition() < currentMap.minPosition())
						.max(Comparator.comparingInt(StatementList::maxPosition));
				if (closestSuccessor.isPresent()) {
					boolean alreadyInMergeList = mergeList.stream().anyMatch(p -> (p.getA() == closestSuccessor.get() && p.getB() == currentMap) || (p.getB() == closestSuccessor.get() && p.getA() == currentMap));
					if (!alreadyInMergeList) {
						mergeList.add(new StreamUtil.Pair<>(closestSuccessor.get(), currentMap));
					}
				} else if (closestPredecessor.isPresent()) {
					boolean alreadyInMergeList = mergeList.stream().anyMatch(p -> (p.getA() == closestPredecessor.get() && p.getB() == currentMap) || (p.getB() == closestPredecessor.get() && p.getA() == currentMap));
					if (!alreadyInMergeList) {
						mergeList.add(new StreamUtil.Pair<>(closestPredecessor.get(), currentMap));
					}
				}
			}
		}

		HashMap<StatementList, StatementList> mergeLookup = new HashMap<>();
		for (StreamUtil.Pair<StatementList, StatementList> merge : mergeList) {
			StatementList statementList = mergeLookup.get(merge.getA());
			if (statementList == null) {
				statementList = merge.getA();
			}
			statementList.addAll(merge.getB());
			mergeLookup.put(merge.getB(), statementList);
			statementMaps.remove(merge.getB());
		}

		statementMaps = new ArrayList<>(statementMaps.stream()
				.filter(x -> !x.hasNoStatements())
				.sorted(Comparator.comparingInt(StatementList::minPosition))
				.collect(Collectors.toList()));

		return statementMaps;
	}


	/**
	 * This method processes the CompoundStatement comp and extracts its Statements into blocks of independent HashMaps
	 * (one Hashmap is one block of statements that are dependent on each other)
	 *
	 * @param comp the compound with statements to process (typically body of a loop)
	 * @param iterationVariable the iteration variable as writes on this have to be in every block
	 * @param peripheralDetector instance of the PeripheralDetecttor used for finding periphery. Null if no periphery shall be searched
	 * @return a list of HashMaps where every HashMap is independent from the others
	 */
	private ArrayList<HashMap<Integer, Traversable>> getIndpendentHashMapsWithReorder(CompoundStatement comp, Identifier iterationVariable, PeripheralDetector peripheralDetector) {
		//get all Identifiers that get written in that CompoundStatement
		List<Identifier> writes = getWritesFromFlowgraph(comp);
		writes.remove(iterationVariable);
		List<Traversable> children = comp.getChildren();

		//list that later should hold independent maps where all Traversables
		// from one map are independent to all other Traversables which are not in the same map
		ArrayList<HashMap<Integer, Traversable>> statementMaps = new ArrayList<>();
		//list of Traversables which are dependent on the iterationVariable
		HashMap<Integer, Traversable> ivStatements = new HashMap<>();

		//put all writes onto the iteration variable into a seperate list
		{
			int counter = 1;
			//save annotationStatements for the next iteration
			// so AnnotationStatements always stay together with the Statement in the next line
			HashMap<Integer, AnnotationStatement> annotationStatements = new HashMap<>();
			for (Traversable c : children) {
				if (c instanceof AnnotationStatement) {
					annotationStatements.put(counter, (AnnotationStatement) c);
				} else if (containsWriteOn(c, iterationVariable)) {
					if (!annotationStatements.isEmpty()) {
						ivStatements.putAll(annotationStatements);
						annotationStatements.clear();
					}
					ivStatements.put(counter, c);
				} else if (!annotationStatements.isEmpty()) {
					annotationStatements.clear();
				}
				counter++;
			}
		}

		for (Identifier identifier : writes) {
			//all HashMap Values resemble all statements that are dependent on Identifier identifier by using it
			//all HashMap Keys resemble the position that the corresponding statement had in the original body (important for autoperf)
			HashMap<Integer, Traversable> temp = new HashMap<>();
			int counter = 1;

			//save annotationStatements for the next iteration
			// so AnnotationStatements always stay together with the Statement in the next line
			HashMap<Integer, AnnotationStatement> annotationStatements = new HashMap<>();
			for (Traversable child : children) {
				if (child instanceof AnnotationStatement) {
					annotationStatements.put(counter, (AnnotationStatement) child);
				} else if (contains(child, identifier)) {
					if (!annotationStatements.isEmpty()) {
						temp.putAll(annotationStatements);
						annotationStatements.clear();
					}
					//if statement uses identifier, put counter and statement into the HashMap
					temp.put(counter, child);
				} else {
					if (!annotationStatements.isEmpty()) {
						annotationStatements.clear();
					}
				}
				counter++;
			}
			if (!annotationStatements.isEmpty()) {
				temp.putAll(annotationStatements);
			}
			statementMaps.add(temp);
		}

		//check for dependencies between writes on the iteration variable
		//and another variable, that also gets written in the body of the loop
		if (!ivStatements.isEmpty()) {
			for (HashMap hm : statementMaps) {
				if (findAndRemoveDuplicates(ivStatements, hm)) {
					model.printIgnoreableError(log, "Could not process the loop " + comp.getParent() + " because there are dependencies" +
							" between the iteration variable and another variable that gets written in this loop.");
					return null;
					//to process this loop you would have to save the dependent variable before the first loop and
					//then reset the value between all fissioned loops because the writes on this variable must be executed
					//in every loop
				}
			}
		}

		//this block adds statements that don't reference any Identifier out of writes,
		//because they are not yet saved in a statementMap. Every statement of these gets
		//saved in a new HashMap because they are independent from every other statement
		int counter = 1;

		//save annotationStatements for the next iteration
		// so AnnotationStatements always stay together with the Statement in the next line
		HashMap<Integer, AnnotationStatement> annotationStatements = new HashMap<>();
		for (Traversable t : children) {
			boolean anyIdentifier = false;
			for (Identifier i : writes) {
				if (contains(t, i)) {
					anyIdentifier = true;
				}
			}
			if (t instanceof AnnotationStatement) {
				annotationStatements.put(counter, (AnnotationStatement) t);
			} else if (!anyIdentifier && !ivStatements.containsValue(t)) {
				HashMap<Integer, Traversable> temp = new HashMap<>();
				if (!annotationStatements.isEmpty()) {
					temp.putAll(annotationStatements);
					annotationStatements.clear();
				}
				temp.put(counter, t);
				statementMaps.add(temp);
			} else {
				annotationStatements.clear();
			}
			counter++;
		}
		//statementMaps now resembles a list of maps where all elements map are dependent on one specific identifier

		//this collapses multiple inner maps into one map if they hold duplicate statements in each other
		//duplicates are a sign of identifiers that are not independent from each other
		// because multiple identifiers got referenced in one statement
		for (HashMap<Integer, Traversable> hm : statementMaps) {
			for (HashMap<Integer, Traversable> temp : statementMaps) {
				if (hm != temp && findAndRemoveDuplicates(hm, temp)) {
					hm.putAll(temp);
					temp.clear();
				}
			}
		}
		//if periphery is used, merge all HashMaps which use any periphery
		if (peripheralDetector != null) {
			//peripheryHM is the HashMap that later should hold all Elements from HashMaps which use any periphery
			Integer peripheryHMIndex = null;
			for (HashMap<Integer, Traversable> hm : statementMaps) {
				for (Traversable t : hm.values()) {
					final Set<PeripheralModuleDescription> usedPeripherals = peripheralDetector.findUsedPeripherals(t);
					if (!usedPeripherals.isEmpty()) {
						if (peripheryHMIndex == null) {
							//set Index of HashMap that will hold all periphery later
							peripheryHMIndex = statementMaps.indexOf(hm);
						} else {
							//put all Traversables from HashMap hm into peripheryHM to get one block that holds all periphery usage
							statementMaps.get(peripheryHMIndex).putAll(hm);
							hm.clear();
						}
						break;
					}
				}
			}
		}

		//delete empty maps
		statementMaps = statementMaps.stream().filter(l -> !l.isEmpty()).collect(Collectors.toCollection(ArrayList::new));

		//add ivStatements to every statementMap as they have to be executed in every fissioned loop
		for (HashMap hm : statementMaps) {
			hm.putAll(ivStatements);
		}
		return statementMaps;
	}

	/**
	 * collapses statementMaps until having the size nos while using autoperf data to find the shortest HashMaps
	 *
	 * @param statementMaps list of statement maps
	 * @param nos minimum size of statementMaps
	 * @param autoperf Autoperfn Data
	 */
	private void reduceListSizeWithAutoPerf(ArrayList<StatementList> statementMaps, int nos, ArrayList<ArrayList<Integer>> autoperf) {
		//compute average computation time of one statement over all iterations
		ArrayList<Double> averagePerf = new ArrayList<>();
		for (int i = 0; i < autoperf.get(0).size(); i++) {
			int counter = 0;
			int sum = 0;
			for (ArrayList<Integer> l : autoperf) {
				sum += l.get(i);
				counter++;
			}
			averagePerf.add((double) sum / counter);
		}

		//add all average computation times so that one element of addedPerf represents
		// the computation time of all statements in one element of statementMaps
		ArrayList<Long> addedPerf = new ArrayList<>();
		for (StatementList statementList : statementMaps) {
			//add all average computation times that belong to one statement of hm
			long sum = 0;
			for (StatementDependencyProperties entry : statementList.getStatements()) {
				//HashMap.Entry<Integer, Traversable> entry = (HashMap.Entry) iter.next();
				sum += averagePerf.get(entry.getPosition() - 1);
			}
			addedPerf.add(sum);
		}

		//collapse the two shortest elements of statementMaps in terms of computation time into one hashMap until
		//statementMaps has the size nos
		while (nos < statementMaps.size() && statementMaps.size() > 1) {
			StatementList a = statementMaps.get(0);
			StatementList b = statementMaps.get(1);
			double perfA = addedPerf.get(0);
			double perfB = addedPerf.get(1);
			if (perfA > perfB) {
				StatementList temp = b;
				b = a;
				a = temp;
			}
			int i = 0;
			for (StatementList l : statementMaps) {
				if (!l.equals(a) && !l.equals(b)) {
					if (perfA > addedPerf.get(i)) {
						b = a;
						a = l;
						perfB = perfA;
						perfA = addedPerf.get(i);
					} else if (perfB > addedPerf.get(i)) {
						b = l;
						perfB = addedPerf.get(i);
					}
				}
				i++;
			}
			//noinspection unchecked
			a.addAll(b);
			statementMaps.remove(b);
		}
	}

	/**
	 * collapse the two smallest elements of statementMaps until statementMaps has the size nos
	 *
	 * @param statementMaps list of hashMaps
	 * @param nos minimum size of statementMaps
	 */
	private void reduceListSize(ArrayList<StatementList> statementMaps, int nos) {
		//TODO: collapse Lists more advanced, e.g. by counting Children of Traversables
		//collapse the two smallest elements of statementMaps until statementMaps has the size nos
		while (nos < statementMaps.size() && statementMaps.size() > 1) {
			StatementList a = statementMaps.get(0);
			StatementList b = statementMaps.get(1);
			if (a.size() > b.size()) {
				StatementList temp = b;
				b = a;
				a = temp;
			}
			for (StatementList l : statementMaps) {
				if (!l.equals(a) && !l.equals(b)) {
					if (a.size() > l.size()) {
						b = a;
						a = l;
					} else if (b.size() > l.size()) {
						b = l;
					}
				}
			}
			//noinspection unchecked
			a.addAll(b);
			statementMaps.remove(b);
		}
	}

	/**
	 * removes duplicates of a and b in b and returns true if it found duplicates
	 *
	 * @param a HashMap
	 * @param b HashMap
	 * @return boolean if it found duplicates
	 */
	private boolean findAndRemoveDuplicates(HashMap<Integer, Traversable> a, HashMap<Integer, Traversable> b) {
		Boolean result = false;
		for (Traversable ta : a.values()) {
			Iterator iter = b.entrySet().iterator();
			while (iter.hasNext()) {
				Map.Entry pair = (Map.Entry) iter.next();
				Traversable tb = (Traversable) pair.getValue();
				if (ta.equals(tb)) {
					iter.remove();
					if (!(ta instanceof AnnotationStatement)) { // catch the rare case of duplicate annotation statements duplicates
						result = true;
					}
				}
			}
		}
		return result;
	}

	/**
	 * true if Identifier i is referenced anywhere in t
	 *
	 * @param t Traversable to search
	 * @param i Identifier to find
	 * @return boolean
	 */
	private boolean contains(Traversable t, Identifier i) {
		DepthFirstIterator iter = new DepthFirstIterator(t);
		while (iter.hasNext()) {
			if (iter.next().equals(i)) {
				return true;
			}
		}
		return false;
	}

	private boolean containsWriteOn(Traversable t, Identifier i) {
		List writes = getWritesFromFlowgraph(t);
		return writes.contains(i);
	}

	private List<Identifier> getWritesFromFlowgraph(Traversable t) {

		List<Variable> allVariables = new ArrayList<>();
		final InnerFlowGraph innerFlowGraph = InnerFlowGraphFractory.getControlFlowWithoutTaskHandling(t);
		GenericAccessesVariables loop = new GenericAccessesVariables(t);
		loop.setInnerFlowGraph(innerFlowGraph);
		final VariableAccessAnalyzer variableAccessAnalyzer = new VariableAccessAnalyzer(model);
		final List<VariableAccess> variableAccesses = variableAccessAnalyzer.scanForIdentifier(loop, Arrays.asList(), allVariables);
		variableAccessAnalyzer.setAccessBehaviour(variableAccesses);
		for (VariableAccess variableAccess : variableAccesses) {
			InnerFlowLinker.link(variableAccess, innerFlowGraph);
		}
		loop.addAllVariableAccess(variableAccesses);

		final Collection<Variable> writeVariables = loop.getWriteVariables();
		final List<Identifier> identifiers = writeVariables.stream()
				.flatMap(v -> v.getIdentifiers().stream())
				.sorted()
				.collect(Collectors.toList());
		return identifiers;
	}


	/**
	 * returns all Identifiers that got written on in the Traversable
	 *
	 * @param t Traversable
	 * @return List of Identifiers
	 */
	@Deprecated
	private ArrayList<Identifier> getWrites(Traversable t) {
		//TODO: implement with InnerFlowGraphs
		Set<Identifier> result = new HashSet<>();
		DepthFirstIterator<Traversable> iter = new DepthFirstIterator<>(t);
		Traversable currentT;
		while (iter.hasNext()) {
			currentT = iter.next();
			if (currentT instanceof ExpressionStatement) {
				if (currentT.getChildren().get(0).getChildren().get(0) instanceof ArrayAccess) {
					Traversable temp = currentT.getChildren().get(0).getChildren().get(0).getChildren().get(0);
					if (temp instanceof Identifier) {
						result.add((Identifier) temp);
					} else {
						result.add(getFirstIdentifier(temp));
					}
				} else if (currentT.getChildren().get(0) instanceof AssignmentExpression) {
					Traversable temp = currentT.getChildren().get(0).getChildren().get(0);
					if (temp instanceof Identifier) {
						result.add((Identifier) temp);
					} else {
						result.add(getFirstIdentifier(temp));
					}
				} else if (currentT.getChildren().get(0) instanceof UnaryExpression) {
					Traversable temp = currentT.getChildren().get(0).getChildren().get(0);
					if (temp instanceof Identifier) {
						result.add((Identifier) temp);
					} else {
						result.add(getFirstIdentifier(temp));
					}
				}
			} else if (currentT instanceof VariableDeclarator) {
				NameID child = (NameID) currentT.getChildren().get(0);
				Identifier i = LoopTransformer.getIdentifier(t, child.getName());
				result.add(i);
			}
		}
		return new ArrayList<>(result);
	}

	/**
	 * retrieves the old value of the iteration variable from a previously copy
	 *
	 * @param i iteration variable
	 * @return a expression statement that sets i to the old value
	 */
	public ExpressionStatement setIterationVariable(Identifier i) {
		AssignmentOperator ao = AssignmentOperator.fromString("=");
		AssignmentExpression ae = new AssignmentExpression(i.clone(), ao, new NameID(i.toString() + LoopTransformer.COPY_ITERATION_VARIABLE));
		return new ExpressionStatement(ae);
	}

	/**
	 * defines a new Variable serving as copy of the iteration value and returns a statement that sets the new variable to the value of the iteration value
	 *
	 * @param i iteration variable
	 * @param l the loop
	 * @return Statement "i_loopOptCopy = i"
	 */
	public ExpressionStatement copyIterationVariable(Identifier i, Loop l) {
		final Declaration declaration = i.findDeclaration();
		if (declaration == null) {
			throw new RuntimeException("The declaration of variable " + i + " could not be found");
		}
		VariableDeclaration decl = (VariableDeclaration) declaration.clone();
		NameID nameID = SymbolTools.getNewName(i.toString() + LoopTransformer.COPY_ITERATION_VARIABLE, i);
		VariableDeclarator vd = new VariableDeclarator(nameID);
		VariableDeclaration newDecl = new VariableDeclaration(decl.getSpecifiers().get(0), vd);
		CompoundStatement comp = (CompoundStatement) l.getParent();
		if (comp.getParent() instanceof CompoundStatement) {
			comp = (CompoundStatement) comp.getParent();
		}
		comp.addDeclarationAfterLastDeclaration(newDecl);

		AssignmentOperator ao = AssignmentOperator.fromString("=");
		AssignmentExpression ae = new AssignmentExpression(nameID.clone(), ao, i.clone());
		ExpressionStatement es = new ExpressionStatement(ae);
		return es;
	}

	/**
	 * returns Identifier if there is only one Identifier in the whole Traversable, else null
	 *
	 * @param t the Traversable to search
	 * @return Identifier
	 */
	private Identifier getUniqueIdentifier(Traversable t) {
		Traversable temp = null;
		DepthFirstIterator iter = new DepthFirstIterator(t);
		while (iter.hasNext()) {
			Traversable current = iter.next();
			if (current instanceof Identifier) {
				if (temp == null) {
					temp = current;
				} else {
					temp = null;
					break;
				}
			}
		}
		return (Identifier) temp;
	}

	private Identifier getFirstIdentifier(Traversable t) {
		Traversable temp = null;
		DepthFirstIterator iter = new DepthFirstIterator(t);
		while (iter.hasNext()) {
			Traversable current = iter.next();
			if (current instanceof Identifier) {
				temp = current;
				break;
			}
		}
		return (Identifier) temp;
	}

	/**
	 * returns the iteration variable by evaluating the pragma or checking the loop if it can be retrieved automatically
	 *
	 * @param l the loop to search into
	 * @param pragma the pragma that maybe holds a iterationVariable() Tag
	 * @return the iterationVariable as identifier
	 */
	public Identifier getIterationVariable(Loop l, LoopPragma pragma) {
		Identifier i;
		PragmaTag<String> tag = pragma.getPragmaTag(Constants.PragmaLoop.ITERATION_VARIABLE);
		if (tag != null) {
			String iv = String.valueOf(tag.getEntryNames().toArray()[0]);
			i = LoopTransformer.getIdentifier(l, iv);
			if (i == null) {
				model.printIgnoreableError(log, "The Variable " + iv + " in the  Pragma " + Constants.PragmaLoop.ITERATION_VARIABLE
						+ " could not be found in the Loop. Please check for spelling mistakes");
				return null;
			}
		} else {
			i = getUniqueIdentifier(l.getCondition());
			if (i == null && l instanceof ForLoop) {
				i = getUniqueIdentifier(((ForLoop) l).getStep());
			}
			if (i == null && l instanceof ForLoop) {
				i = getUniqueIdentifier(((ForLoop) l).getInitialStatement());
			}
			if (i == null) {
				model.printIgnoreableError(log, "Mutliple or none Identifiers were found that could be the Iteration Variable. " +
						"Please specify in the LoopPragma which of them is the true Iteration Variable!");
				return null;
			}
		}
		return i;
	}
}
