/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.loopOptimizer.tools;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.loopOptimizer.model.LoopRuntimeReport;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.LoopPragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.Pragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.PragmaTag;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.PragmaAnnotationParser;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.peripheralDetection.PeripheralDetector;
import de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.peripheralDetection.PeripheralLibrary;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.AnnotationStatement;
import cetus.hir.AssignmentExpression;
import cetus.hir.AssignmentOperator;
import cetus.hir.BinaryExpression;
import cetus.hir.BreakStatement;
import cetus.hir.CompoundStatement;
import cetus.hir.DepthFirstIterator;
import cetus.hir.DoLoop;
import cetus.hir.Expression;
import cetus.hir.ExpressionStatement;
import cetus.hir.FlatIterator;
import cetus.hir.FloatLiteral;
import cetus.hir.ForLoop;
import cetus.hir.IDExpression;
import cetus.hir.Identifier;
import cetus.hir.IntegerLiteral;
import cetus.hir.Loop;
import cetus.hir.NameID;
import cetus.hir.PragmaAnnotation;
import cetus.hir.RegularBinaryOperator;
import cetus.hir.Specifier;
import cetus.hir.Statement;
import cetus.hir.SwitchStatement;
import cetus.hir.SymbolTools;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.UnaryExpression;
import cetus.hir.UnaryOperator;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import cetus.hir.WhileLoop;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static com.google.common.primitives.UnsignedInts.max;
import static java.lang.Math.abs;

/**
 * Transform the traversable here
 */
public class LoopTransformer {
	static final String COPY_ITERATION_VARIABLE = "_loopOptInitialValue";
	static final String NEW_ITERATION_VARIABLE = "_loopOptIterationVariable";
	static final String BREAK_TRIGGERED_VARIABLE_NAME = "breakTriggered";
	static final Set<Class<? extends Traversable>> IGNORED_SUBCLASSES_FOR_BREAK_SEARCH = Stream.of(SwitchStatement.class, Loop.class).collect(Collectors.toSet());
	private static final Logger log = LoggerFactory.getLogger(LoopTransformer.class);
	private final Model model;
	private final FissionHelper fissionHelper;

	public LoopTransformer(Model model) {
		this.model = model;
		this.fissionHelper = new FissionHelper(model);
	}

	/**
	 * transforms a Loop into several splitted or fissioned loops
	 * @param l the loop
	 * @param pragma the pragma that belongs to the loop
	 * 				(holds information about how many and how the loop should be transformed)
	 * @param autoperfPointer points out which autoPerf Values should be used for this loop (optional)
	 * @param compoundStatementUsed if the LoopPragma is in front of a compound instead of a loop.
	 *                              needed in case of "C99-Hack" where for(int i=0; ;) --> {int i=0; for(;;)}
	 * @return a list of loops and other necessary Statements to replace the old loop
	 */
	public ArrayList<Traversable> transform(Loop l, LoopPragma pragma, int autoperfPointer, boolean compoundStatementUsed) {

		ArrayList<ArrayList<Integer>> autoperf = null;
		if (pragma.getPragmaTag(Constants.PragmaLoop.AUTOPERF)!=null) {
			autoperf = model.getModelComponent(LoopRuntimeReport.class).getLoop(getContainingFile(l) + ":" + autoperfPointer);
		}

		ArrayList<Traversable> result = processLoop(l, pragma, autoperf);

		transformBreakLoops(result, l);

		if (result==null) {
			return null;
		}

		//add VariableDeclaration of iterationVariable and rename iterationVariable
		if (compoundStatementUsed) {
			replaceIterationVariable(l, result);
		}
		return result;
	}

	/**
	 * replaces all ocurrences of the old iterationVariable in result with a new iterationVariable,
	 * which is not colliding with another Variable
	 * @param l the loop which was optimized
	 * @param result the result of the optimization
	 */
	public void replaceIterationVariable(Loop l, ArrayList<Traversable> result) {
		List<Traversable> children = l.getParent().getChildren();
		VariableDeclaration oldDecl = ((VariableDeclaration) children.get(0).getChildren().get(0));

		//old name of iterationVariable
		NameID oldName = (NameID) oldDecl.getChildren().get(0).getChildren().get(0);
		//new name of iterationVariable
		NameID newName = SymbolTools.getNewName(oldDecl.getDeclarator(0).getID().toString() + NEW_ITERATION_VARIABLE, l);
		VariableDeclaration newDecl = new VariableDeclaration(oldDecl.getSpecifiers().get(0), new VariableDeclarator(newName));

		//new Declaration for new name at the beginning of the method
		//getProcedureBody(l).addDeclarationAfterLastDeclaration(newDecl);
		Identifier newIdentifier = new Identifier(newDecl.getDeclaredSymbols().get(0));

		//replace all occurrences of old iterationVariable in the result with new iterationVariable
		for (Traversable t : result) {
            DepthFirstIterator iter = new DepthFirstIterator(t);
            while (iter.hasNext()) {
                Traversable current = iter.next();
                if (current.toString().equals(oldName.toString())) {
                    int index = current.getParent().getChildren().indexOf(current);
                    current.getParent().setChild(index, newIdentifier.clone());
                }
            }
        }

		//build expression which initializes the new iterationVariable in front of the first loop
		AssignmentOperator ao = AssignmentOperator.NORMAL;
		Expression rhs = (Expression) oldDecl.getChildren().get(0).getChildren().get(1).getChildren().get(0);
		if (rhs instanceof IntegerLiteral) {
            rhs = ((IntegerLiteral) rhs).clone();
        } else if (rhs instanceof FloatLiteral) {
            rhs = ((FloatLiteral) rhs).clone();
        } else if (rhs instanceof Identifier) {
            rhs = ((Identifier) rhs).clone();
        } else if (rhs instanceof NameID) {
            rhs = ((NameID) rhs).clone();
        }
		AssignmentExpression ae = new AssignmentExpression(newIdentifier, ao, rhs);
		ExpressionStatement es = new ExpressionStatement(ae);
		result.add(0, es);
		result.add(0, newDecl);
	}

	/**
	 * decides whether to use fission or splitting
	 * @param l Loop to be optimized
	 * @param pragma corresponding pragma
	 * @param autoperf autoPerf data (optional)
	 * @return a list of loops and other necessary Statements to replace the old loop
	 */
	private ArrayList<Traversable> processLoop(Loop l, LoopPragma pragma, ArrayList<ArrayList<Integer>> autoperf) {
		//search for peripherals and use fission if peripherals were found.
		PeripheralDetector peripheralDetector = new PeripheralDetector(model, new PeripheralLibrary(model.getConfig().getAdditionalSystemLibraries()));
		final Set<PeripheralModuleDescription> usedPeripherals = peripheralDetector.findUsedPeripherals(l);

		if (pragma.getPragmaTag(Constants.PragmaLoop.USE_FISSION)!=null || !usedPeripherals.isEmpty()
			|| !(l.getCondition() instanceof BinaryExpression) || !((BinaryExpression) l.getCondition()).getOperator().isCompare()
			|| ((BinaryExpression) l.getCondition()).getOperator().toString().equals("!=")
			|| ((BinaryExpression) l.getCondition()).getOperator().toString().equals("==")) {
			if (pragma.getPragmaTag(Constants.PragmaLoop.USE_SPLITTING)!=null) {
				model.printIgnoreableError(log,"Could not use Splitting because the requirements were not fulfilled. " +
					"Please check if the condition of the loop is a binary compare expression.");
				return null;
			}
			if (pragma.getPragmaTag(Constants.PragmaLoop.MAX_CYCLES)!=null) {
				model.printIgnoreableError(log,"Could not process this loop because maxCycles is only working with Splitting, but the loop does not allow using Splitting. "
					+ "Please check if the condition is a binary compare expression and no periphery is used.");
				return null;
			}
			if (usedPeripherals.isEmpty()) {
				return fissionLoop(l, pragma, autoperf, peripheralDetector);
			} else {
				final Set<PeripheralModuleDescription> condPeripherals = peripheralDetector.findUsedPeripherals(l.getCondition());
				if (!condPeripherals.isEmpty()) {
					model.printIgnoreableError(log,"Periphery was found in the condition of the loop. Cant fission loop because periphery can only be used on one core");
					return null;
				}
				if (l instanceof ForLoop) {
					final Set<PeripheralModuleDescription> isPeripherals = peripheralDetector.findUsedPeripherals(((ForLoop) l).getInitialStatement());
					final Set<PeripheralModuleDescription> stepPeripherals = peripheralDetector.findUsedPeripherals(((ForLoop) l).getStep());
					if (!isPeripherals.isEmpty() || !stepPeripherals.isEmpty()) {
						model.printIgnoreableError(log,"Periphery was found in the initial statement or step of the loop. Cant fission loop because periphery can only be used on one core");
						return null;
					}
				}
				return fissionLoop(l, pragma, autoperf, peripheralDetector);
			}
		} else {
			return  splitLoop(l, pragma, autoperf);
		}
	}

	/**
	 * fissions a loop
	 * @param l the loop
	 * @param pragma the pragma that belongs to the loop
	 * 				(holds information about how many and how the loop should be transformed)
	 * @param autoperf autoPerf data (optional)
	 * @param peripheralDetector !=null if periphery shall be searched.
	 * @return a list of loops and other necessary Statements to replace the old loop
	 */
	private ArrayList<Traversable> fissionLoop(Loop l, LoopPragma pragma, ArrayList<ArrayList<Integer>> autoperf, PeripheralDetector peripheralDetector) {
		ArrayList<Traversable> result = new ArrayList<>();

		Identifier i = fissionHelper.getIterationVariable(l, pragma);
		if (i == null) {
			return null;
		}

		//get independent compounds that can be fissioned without conflicts
		ArrayList<CompoundStatement> compounds = fissionHelper.getIndependentCompounds((CompoundStatement) l.getBody(), pragma, autoperf, i, peripheralDetector);

		if (compounds==null) {
			return null;
		}

		//add a declaration and statement that copies the iteration variable
		result.add(fissionHelper.copyIterationVariable(i, l));

		int counter = 0;
		assert compounds != null;
		boolean first = true;
		for (CompoundStatement c : compounds) {
			if (counter!=0) {
				//set iteration variable to old value so that the loop can be run again
				result.add(fissionHelper.setIterationVariable(i));
			}
			counter++;
			//clone our original loop and set body to new compound
			Loop copy = (Loop) ((Statement) l).clone();
			if (copy instanceof DoLoop) {
				copy.setChild(0, c);
			} else {
				copy.setChild(copy.getChildren().size() - 1, c);
			}
			if (first) {
				first = false;
			} else {
				//add #pragma microStreams task in front of every loop which is not the first loop
				result.add(new AnnotationStatement(new PragmaAnnotation(Constants.Pragma.MICRO_STREAMS_TASK)));
			}
			result.add(copy);
		}
		return result;
	}

	/**
	 * uses splitting to transform the loop (currently only supports For Loops)
	 * @param l the loop
	 * @param pragma corresponding LoopPragma
	 * @param autoperf autoperf data (optional)
	 * @return splitted loop
	 */
	private ArrayList<Traversable> splitLoop(Loop l, LoopPragma pragma, ArrayList<ArrayList<Integer>> autoperf) {
		PragmaTag<Integer> distributionTag = pragma.getPragmaTag(Constants.PragmaLoop.DISTRIBUTION);
		PragmaTag<Integer> numTag = pragma.getPragmaTag(Constants.PragmaLoop.NUMBER_OF_SPLITS);
		Integer nos = 0;
		PragmaTag<Integer> maxCyclesTag = pragma.getPragmaTag(Constants.PragmaLoop.MAX_CYCLES);
		if (numTag!=null) {
			//get number of splits
			nos = (Integer) numTag.getEntryNames().toArray()[0];
		}
		if (maxCyclesTag!=null && autoperf==null) {
			model.printIgnoreableError(log,"MaxCycles tag found but there was no autoPerf data. MaxCycles can't be used without autoperf data");
			return null;
		}

		//iterationPerf represents the computation time of each iteration in autoPerf
		ArrayList<Integer> iterationPerf = null;
		if (autoperf!=null) {
			//adds all perf values of one iteration to one sum
			iterationPerf = new ArrayList<>();
			for (ArrayList<Integer> list : autoperf) {
				int i = list.stream().reduce(0, (a, b) -> a + b);
				iterationPerf.add(i);
			}
		}

		Expression cond = l.getCondition();

		final SplitHelper splitHelper = new SplitHelper(model);
		Identifier iterationVariable = splitHelper.getIterationVariable(l, pragma.getPragmaTag(Constants.PragmaLoop.ITERATION_VARIABLE));
		//ConditionValue of the iterationValue
		Traversable ev = splitHelper.getEndValue(cond, iterationVariable);

		Traversable iv = null;
		if (l instanceof ForLoop) {
			//InitialValue of the iteration value
			iv = splitHelper.getInitialValue((ForLoop) l);
		} // if l is not of type ForLoop or iv can't be evaluated, iv is created later

		//TODO: Fallunterscheidung von < und <= ?
		if (maxCyclesTag!=null || distributionTag!=null || iterationPerf!=null) {
			return splitHelper.splitLoopWithPartitions(l, iv, ev, iterationVariable, nos, distributionTag, iterationPerf, maxCyclesTag);
		} else {
			return splitHelper.splitLoopSimple(l, nos, iv, ev, iterationVariable);
		}
	}

	private String getContainingFile(Traversable t) {

		final TranslationUnit translationUnit = CetusUtils.findParentOfType(t, TranslationUnit.class);
		if (translationUnit == null) {
			model.printIgnoreableError(log,"Could not find name of file because the given Statement has no TranslationUnit as parent");
		}
		return translationUnit.getOutputFilename();
	}

	/**
	 * Splits the given loop according to the given partitions. Only works on ForLoops and WhileLoops, for which all necessary information
	 * can be determined automatically. Does not evaluate the PragmaTag!
	 * @param loop the Loop to split
	 * @param partitions partitions that hold information about how big / long the split loops should be
	 * @return the split loops
	 */
	public ArrayList<Traversable> splitLoopWithPartitions(Loop loop, ArrayList<Integer> partitions) {
		Expression condition = loop.getCondition();
		SplitHelper splitHelper = new SplitHelper(model);
		Identifier iterationVariable = splitHelper.getIterationVariable(loop,null);
		Traversable iv;
		Traversable ev;
		if (loop instanceof ForLoop) {
			iv = splitHelper.getInitialValue((ForLoop) loop);
			ev = splitHelper.getEndValue(condition,iterationVariable);
		} else if (loop instanceof WhileLoop) {
			iv = splitHelper.getInitialValueWhileLoop((WhileLoop) loop, iterationVariable);
			ev = splitHelper.getEndValue(condition, iterationVariable);
		} else {
			log.error("Can only split ForLoops or WhileLoops");
			return null;
		}

		if (iv==null||ev==null||iterationVariable==null) {
			log.error("Necessary information for loop splitting can not be determined automatically");
			return null;
		}
		ArrayList<Traversable> result = splitHelper.buildLoopsWithPartitions(loop, iv, ev, iterationVariable, partitions, false);
		if (result.stream().anyMatch(l -> hasBreak(l))) {
			transformBreakLoops(result, loop);
			log.warn("Split loops contain break. Therefore split distribution might not be as scheduled");
		}
		return result;
	}

	/**
	 * how many blocks can be achieved through loop splitting
	 * supported Operators for the step: ++, --, +, -, *
	 * initial and end value of the iterationValue has to be in literal form (no variables)
	 * @param traversable a traversable to analyze
	 * @return how many blocks can be achieved through loop splitting
	 * 		   (-1 means could not be found out (not fixable with this loop),  -2 means error occurred (wrong method call etc))
	 */
	public int howOftenSplittable(Traversable traversable) {
		final int errorOccurred = -2;

		Pragma pragma = getPragma(traversable);
		if (!(pragma instanceof LoopPragma)) {
			model.printIgnoreableError(log,"No LoopPragma found at the position of the Traversable. Please specify a LoopPragma.");
			return errorOccurred;
		}
		if (!(traversable instanceof Loop)) {
			FlatIterator iter = new FlatIterator(traversable);
			while (iter.hasNext()) {
				Traversable current = iter.next();
				if (current instanceof Loop) {
					traversable = current;
					break;
				}
			}
			if (!(traversable instanceof Loop)) {
				model.printIgnoreableError(log,"No Loop found at the traversable or its immediate children. Please check the traversable.");
				return errorOccurred;
			}
		}
		if (!(traversable instanceof ForLoop)) {
			model.printIgnoreableError(log,"The given Traversable is no For Loop. Currently only supported on For Loops");
			return errorOccurred;
		}
		ForLoop l = (ForLoop) traversable;
		PragmaTag tag = pragma.getPragmaTag(Constants.PragmaLoop.ITERATION_VARIABLE);
		return howManyIterations(l, tag);
	}

	/**
	 * determines the amount of iterations of the given Loop
	 * @param loop the loop
	 * @param tag the
	 * @return the number of iterations, or -1 if it couldn't be determined
	 */
	public int howManyIterations(Loop loop, PragmaTag tag) {
		if (loop instanceof ForLoop) {
			return getIterationsOfForLoop((ForLoop) loop,tag);
		} else if (loop instanceof WhileLoop) {
			return getIterationsOfWhileLoop((WhileLoop) loop, tag);
		}
		log.error("Currently only supported on For Loops and While Loops");
		return -1;
	}

	private int getIterationsOfWhileLoop(WhileLoop loop, PragmaTag tag) {

		SplitHelper splitHelper = new SplitHelper(model);
		Identifier identifier = splitHelper.getIterationVariable(loop , null);
		Expression step = splitHelper.getStepOfWhileLoop(loop, identifier);
		Traversable initialValue = splitHelper.getInitialValueWhileLoop(loop, identifier);


		if (step == null || initialValue == null) {
			log.info("The loop can not be split automatically");
			return -1;
		}

		Traversable endValue = splitHelper.getEndValue(loop.getCondition(), identifier);

		return howManyIterations(identifier, initialValue, endValue, step);
	}

	private int getIterationsOfForLoop(ForLoop l, PragmaTag tag) {
		final SplitHelper splitHelper = new SplitHelper(model);
		Identifier iterationVariable = splitHelper.getIterationVariable(l, tag);

		//the traversable which initializes the iterationVariable
		Traversable it = splitHelper.getInitialValue(l);

		//the traversable which symbolizes the last value of the iterationVariable
		Traversable et = l.getCondition().getChildren().get(0).equals(iterationVariable) ?
				l.getCondition().getChildren().get(1) : l.getCondition().getChildren().get(0);

		Expression step = l.getStep();

		return howManyIterations(iterationVariable, it, et, step);
	}

	/**
	 * Calculates the amount of iterations with the given values
	 * @param iterationVariable the iterationVariable of the loop
	 * @param it the traversable which initializes the iterationVariable
	 * @param et the traversable which symbolizes the last value of the iterationVariable
	 * @param step the step of the loop
	 * @return how often the loop can be splitted
	 */
	private int howManyIterations(Identifier iterationVariable, Traversable it, Traversable et, Expression step) {


		if ((it instanceof IntegerLiteral || it instanceof FloatLiteral) && (et instanceof IntegerLiteral || et instanceof FloatLiteral)) {
			double itValue;
			double etValue;
			if (it instanceof IntegerLiteral) {
				itValue = ((IntegerLiteral) it).getValue();
			} else {
				itValue = ((FloatLiteral) it).getValue();
			}
			if (et instanceof IntegerLiteral) {
				etValue = ((IntegerLiteral) et).getValue();
			} else {
				etValue = ((FloatLiteral) et).getValue();
			}
			double range = abs(etValue - itValue);
			if (((BinaryExpression) et.getParent()).getOperator().toString().equals(">=")
				|| ((BinaryExpression) et.getParent()).getOperator().toString().equals("<=")) {
				range++;
			}

			if (step instanceof UnaryExpression) {
				UnaryOperator uo = ((UnaryExpression) step).getOperator();
				if (uo.toString().equals("++") || uo.toString().equals("--")) {
					return (int) range;
				} else {
					log.info("The step is of type UnaryExpression, but the Operator is not ++ or --. This is not supported");
					return -1;
				}
			} else if (step instanceof AssignmentExpression) {
				AssignmentOperator ao = ((AssignmentExpression) step).getOperator();
				if (ao.toString().equals("=") && step.getChildren().get(1) instanceof BinaryExpression) {
					BinaryExpression be = (BinaryExpression) step.getChildren().get(1);
					if (be.getOperator().toString().equals("+") || be.getOperator().toString().equals("-")) {
						if (be.getLHS().equals(iterationVariable) && be.getRHS() instanceof IntegerLiteral) {
							return (int) (range / ((IntegerLiteral) be.getRHS()).getValue());
						} else if (be.getRHS().equals(iterationVariable) && be.getLHS() instanceof IntegerLiteral) {
							return (int) (range / ((IntegerLiteral) be.getLHS()).getValue());
						}
					} else if (be.getOperator().toString().equals("*")) {
						log.warn("LoopOpt normally expects even steps. By using * the splitting could be uneven");
						if (be.getLHS().equals(iterationVariable) && be.getRHS() instanceof IntegerLiteral) {
							return (int) (Math.log(((IntegerLiteral) et).getValue()/((IntegerLiteral) it).getValue())
								/ Math.log(((IntegerLiteral) be.getRHS()).getValue()));
						} else if (be.getRHS().equals(iterationVariable) && be.getLHS() instanceof IntegerLiteral) {
							return (int) (Math.log(((IntegerLiteral) et).getValue()/((IntegerLiteral) it).getValue())
								/ Math.log(((IntegerLiteral) be.getLHS()).getValue()));
						}
					}
					model.printIgnoreableError(log, "The step could not be evaluated. This loop can not be evaluated.");
					return -1;
				}
			}
		} else {
			log.info("Could not find out how many splits can be made because start value " +
				"and / or end value of the iterationVariable are initialised with a variable. ");
			return -1;
		}
		return -1;
	}

	/**
	 * Method for finding out how many times a loop can be fissioned
	 * @param t Traversable which is the loop or in special cases a compoundStatement surrounding the loop
	 * @param periphery true if periphery shall be searched and processed (all periphery is put in one loop-body)
	 * @return how many times the loop can be fissioned
	 */
	public int howOftenFissionizable(Traversable t, boolean periphery) {
		//search for periphery if wanted
		PeripheralDetector peripheralDetector = null;
		if (periphery) {
			peripheralDetector = new PeripheralDetector(model,
				new PeripheralLibrary(model.getConfig().getAdditionalSystemLibraries()));
		}

		Pragma pragma = getPragma(t);
		if (pragma instanceof LoopPragma) {
			if (t instanceof Loop) {
				return howOftenFissionizable((Loop) t, (LoopPragma) pragma, peripheralDetector);
			} else {
				FlatIterator iter = new FlatIterator(t);
				while (iter.hasNext()) {
					Traversable current = iter.next();
					if (current instanceof Loop) {
						return howOftenFissionizable((Loop) current, (LoopPragma) pragma, peripheralDetector);
					}
				}
				log.warn("No Loop found at Traversable t or its immediate children. Please check the Traversable.");
				return 0;
			}
		} else {
			log.warn("No LoopPragma found at the location of the loop. Please specify a LoopPragma in front of Traversable t.");
			return 0;
		}
	}

	/**
	 * Method for finding out how many times a loop can be fissioned
	 * @param l the loop to test
	 * @param pragma the pragma belonging to the loop
	 * @param peripheralDetector instance of the PeripheralDetecttor used for finding periphery.
	 *                           null if no periphery shall be searched
	 * @return number of fissions possible
	 */
	private int howOftenFissionizable(Loop l, LoopPragma pragma, PeripheralDetector peripheralDetector) {
		Identifier i = fissionHelper.getIterationVariable(l, pragma);
		if (i == null) {
			return -1;
		}
		return fissionHelper.getNumberOfIndependetBlocks((CompoundStatement) l.getBody(), i, peripheralDetector);
	}

	private Pragma getPragma(Traversable traversable) {
		PragmaAnnotation pragmaAnnotation = CetusUtils.getAnnotation(traversable);
		// Check whether the PragmaAnnotation is a valid UStreams Annotation.
		return new PragmaAnnotationParser(model).parse(pragmaAnnotation, traversable);
	}

	/**
	 * how many blocks can be achieved through all loop opt methods
	 *
	 * @param traversable a traversable to analyze
	 * @param periphery if periphery has to be regarded
	 * @return how many blocks can be achieved through any loop division method
	 */
	public int howOftenDividable(Traversable traversable, boolean periphery) {
		if (periphery) {
			PeripheralDetector peripheralDetector = new PeripheralDetector(model, new PeripheralLibrary(model.getConfig().getAdditionalSystemLibraries()));
			final Set<PeripheralModuleDescription> usedPeripherals = peripheralDetector.findUsedPeripherals(traversable);
			if (usedPeripherals.isEmpty()) {
				return max(howOftenSplittable(traversable), howOftenFissionizable(traversable, false));
			} else {
				return howOftenFissionizable(traversable, true);
			}
		} else {
			return max(howOftenSplittable(traversable), howOftenFissionizable(traversable, false));
		}
	}

	/**
	 * returns a Identifier inside the loop that has the name i
	 * @param t the loop to search inside
	 * @param i the name of the identifier as String
	 * @return Identifier or null if none was found
	 */
	public static Identifier getIdentifier(Traversable t, String i) {
		DepthFirstIterator iter = new DepthFirstIterator(t);
		while (iter.hasNext()) {
			Object o = iter.next();
			if (o instanceof Identifier && i.equals(o.toString())) {
				return (Identifier) o;
			}
		}
		return null;
	}

	/**
	 * returns the Body of the Procedure wrapping the Traversable t
	 * @param t the Traversable to search
	 * @return CompoundStatement whose parent is the Procedure
	 */
	public CompoundStatement getProcedureBody(Traversable t) {
		Traversable ret = t;
		while (!(ret.getParent() instanceof CompoundStatement)) {
			ret = ret.getParent();
		}
		return (CompoundStatement) ret.getParent();
	}

	public boolean hasBreak(Traversable traversable) {
		return findBreak(traversable).findAny().isPresent();
	}

	public Stream<BreakStatement> findBreak(Traversable traversable) {
		return CetusUtils.getStreamOfAllChildren(traversable, IGNORED_SUBCLASSES_FOR_BREAK_SEARCH).flatMap(StreamUtil.castOrEmpty(BreakStatement.class));
	}

	/**
	 * Transforms loops containing a break statement
	 * @param transformedLoops loops to be transformed
	 * @param formerLoop the original loop
	 */
	public void transformBreakLoops(List<Traversable> transformedLoops, Loop formerLoop) {

		boolean loopsContainBreak = transformedLoops.stream().anyMatch(l -> hasBreak(l));
		if (loopsContainBreak) {
			//create break triggered variable
			final VariableDeclaration breakTriggeredVariableDeclaration = createBreakTriggeredVariable(formerLoop);
			final IDExpression breakTriggeredVariableID = breakTriggeredVariableDeclaration.getDeclaredIDs().get(0);
			CetusUtils.findParentOfType(formerLoop, CompoundStatement.class).addDeclarationAfterLastDeclaration(breakTriggeredVariableDeclaration);

			transformedLoops.stream().flatMap(StreamUtil.castOrEmpty(Loop.class)).forEach(x -> transformBreakLoop(x, breakTriggeredVariableID));
		}

	}


	/**
	 * Transforms a loop with a break statemtn
	 * @param loop the original loop
	 * @param breakTriggeredVariableID the identifier triggering the break
	 */
	public void transformBreakLoop(Loop loop, IDExpression breakTriggeredVariableID) {
		//add trigger for variable before each relevant break
		final List<BreakStatement> breakStatements = findBreak(loop).collect(Collectors.toList());
		for (BreakStatement bs : breakStatements) {
			final ExpressionStatement expressionStatement = new ExpressionStatement(new AssignmentExpression(breakTriggeredVariableID.clone(), AssignmentOperator.NORMAL, new IntegerLiteral(1)));
			CetusUtils.findParentOfType(bs, CompoundStatement.class).addStatementBefore(bs, expressionStatement);
		}

		//modify loop condition

		final BinaryExpression newCondition = new BinaryExpression(loop.getCondition().clone(), RegularBinaryOperator.LOGICAL_AND, new UnaryExpression(UnaryOperator.LOGICAL_NEGATION, breakTriggeredVariableID.clone()));
		loop.getCondition().swapWith(newCondition);
	}

	static VariableDeclaration createBreakTriggeredVariable(Traversable traversable) {
		final NameID breakTriggeredVariable = SymbolTools.getNewName(BREAK_TRIGGERED_VARIABLE_NAME, traversable);
		final VariableDeclaration variableDeclaration = CetusUtils.createVariable(Arrays.asList(Specifier.CHAR), breakTriggeredVariable, new IntegerLiteral(0));
		return variableDeclaration;
	}
}
