/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.loopOptimizer.tools;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.LoopPragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.PragmaTag;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.VariableAccessAnalyzer;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.AnnotationStatement;
import cetus.hir.AssignmentExpression;
import cetus.hir.BinaryExpression;
import cetus.hir.CompoundStatement;
import cetus.hir.ConditionalExpression;
import cetus.hir.DFIterator;
import cetus.hir.DeclarationStatement;
import cetus.hir.DoLoop;
import cetus.hir.EvaluateExpression;
import cetus.hir.Expression;
import cetus.hir.ExpressionStatement;
import cetus.hir.FloatLiteral;
import cetus.hir.ForLoop;
import cetus.hir.Identifier;
import cetus.hir.IfStatement;
import cetus.hir.Initializer;
import cetus.hir.IntegerLiteral;
import cetus.hir.Loop;
import cetus.hir.NameID;
import cetus.hir.NullStatement;
import cetus.hir.PragmaAnnotation;
import cetus.hir.Procedure;
import cetus.hir.RegularBinaryOperator;
import cetus.hir.Statement;
import cetus.hir.SwitchStatement;
import cetus.hir.Symbolic;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.UnaryExpression;
import cetus.hir.UnaryOperator;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import cetus.hir.WhileLoop;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static java.lang.Math.abs;

/**
 * Helper providing many different methods needed for splitting loops
 * Created by dominikl on 14.07.17.
 */
public class SplitHelper {
	private static final Logger log = LoggerFactory.getLogger(SplitHelper.class);
	private final Model model;

	public SplitHelper(Model model) {
		this.model = model;
	}

	/**
	 * split the loop into nos loops while nor using distribution, autoperf or maxCycles
	 * @param l the loop to split
	 * @param nos Number of Splits
	 * @param iv initial value of the iterationVariable, either as Identifier saving the value, or as Literal containing the value
	 * @param ev same as iv, only with the last (end) value of the iterationVariable
	 * @param iterationVariable the iterationVariable of the loop
	 * @return a CompoundStatement containing the nos loops
	 */
	public ArrayList<Traversable> splitLoopSimple(Loop l, Integer nos, Traversable iv, Traversable ev, Identifier iterationVariable) {
		ArrayList<Traversable> result = new ArrayList<>();

		//copy the iterationVariable in front of the loop to save the initial value
		if (iv==null) {
			Statement expression = new FissionHelper(model).copyIterationVariable(iterationVariable, l);
			iv = expression.getChildren().get(0).getChildren().get(0);
			result.add(expression);
		}

		for (int i = 1; i <= nos; i++) {
			Expression newCond = l.getCondition().clone();
			Expression be = buildSimpleExpression(iv, ev, i, nos);
			if (newCond.getChildren().get(1).equals(iterationVariable)) {
				newCond.setChild(0, be);
			} else {
				newCond.setChild(1, be);
			}
			Loop newLoop = buildLoop(l, i, newCond);
			if (i!=1) {
				result.add(new AnnotationStatement(new PragmaAnnotation(Constants.Pragma.MICRO_STREAMS_TASK)));
			}
			result.add( newLoop);
		}
		return result;
	}

	/**
	 * split the loop while using (maxCycles && autoperf) || (autoperf || distribution)
	 * this method just constructs the partitions for buildLoopsWithPartitions and then calls this method with the partitions
	 * @param l the loop to split
	 * @param iv initial value of the iterationVariable, either as Identifier saving the value, or as Literal containing the value
	 * @param ev same as iv, only with the last (end) value of the iterationVariable
	 * @param iterationVariable the iterationVariable of the loop
	 * @param nos Number of Splits
	 * @param distributionTag PragmaTag containg the distribution data
	 * @param iterationPerf ArrayList containing the autoperf data
	 * @param maxCyclesTag PragmaTag containing the maxCycles data
	 * @return a CompoundStatement containing the nos loops
	 */
	public ArrayList<Traversable> splitLoopWithPartitions(Loop l, Traversable iv, Traversable ev, Identifier iterationVariable, Integer nos, PragmaTag<Integer> distributionTag, ArrayList<Integer> iterationPerf, PragmaTag<Integer> maxCyclesTag) {
		if (maxCyclesTag!=null) {
			//maximum of cycles that should be taken in one loop
			Integer maxCycles = (Integer) maxCyclesTag.getEntryNames().toArray()[0];
			ArrayList<Integer> partitions = getMaxCyclesPartitions(iterationPerf, maxCycles);

			return buildLoopsWithPartitions(l, iv, ev, iterationVariable, partitions, true);
		} else if (iterationPerf!=null) {
			ArrayList<Integer> partitions;
			if (distributionTag!=null) {
				//get partitions accordingly to distribution
				partitions = getUnevenPartitions(iterationPerf, new ArrayList<>(distributionTag.getEntryNames()));
			} else {
				partitions = getEvenPartitions(iterationPerf, nos);
			}
			return buildLoopsWithPartitions(l, iv, ev, iterationVariable, partitions, true);
		} else {
			ArrayList<Integer> distribution = new ArrayList<Integer>(distributionTag.getEntryNames());
			return buildLoopsWithPartitions(l, iv, ev, iterationVariable, distribution, true);
		}
	}

	/**
	 * split the loop into partitions.size() parts
	 * @param l the loop to split
	 * @param iv initial value of the iterationVariable, either as Identifier saving the value, or as Literal containing the value
	 * @param ev same as iv, only with the last (end) value of the iterationVariable
	 * @param iterationVariable the iterationVariable of the loop
	 * @param partitions partitions that hold information about how big / long the splitted loops should be later
	 * @param insertTaskPragma true, if taskPragmas should be inserted between new loops
	 * @return a CompoundStatement containing the nos loops
	 */
	public ArrayList<Traversable> buildLoopsWithPartitions(Loop l, Traversable iv, Traversable ev, Identifier iterationVariable, ArrayList<Integer> partitions, boolean insertTaskPragma) {
		ArrayList<Traversable> result = new ArrayList<>();

		if (iv==null) {
			Statement expression = new FissionHelper(model).copyIterationVariable(iterationVariable, l);
			iv = expression.getChildren().get(0).getChildren().get(0);
			result.add(expression);
		}

		for (int i = 1; i <= partitions.size(); i++) {
			Expression newCond = l.getCondition().clone();
			Expression be = buildDistributionExpression(iv, ev, partitions, i);
			if (newCond.getChildren().get(1).equals(iterationVariable)) {
				newCond.setChild(0, be);
			} else {
				newCond.setChild(1, be);
			}
			Loop newLoop = buildLoop(l, i, newCond);
			if (i!=1 && insertTaskPragma) {
				result.add(new AnnotationStatement(new PragmaAnnotation(Constants.Pragma.MICRO_STREAMS_TASK)));
			}
			result.add( newLoop);
		}
		return result;
	}

	/**
	 * builds a loop that is a copy of l with a new Condition and InitialStatement is only copied if i==1
	 * @param l the loop to copy
	 * @param i condition i==1 for copying the initial statement
	 * @param newCond condition to set into the new loop
	 * @return a new loop
	 */
	private Loop buildLoop(Loop l, int i, Expression newCond) {
		Loop newLoop;
		if (l instanceof ForLoop) {
			Statement newIS = null;
			if (i == 1) {
				newIS = ((ForLoop) l).getInitialStatement().clone();
			}
			Expression newStep = ((ForLoop) l).getStep().clone();
			newLoop = new ForLoop(newIS, newCond, newStep, l.getBody().clone());
		} else if (l instanceof WhileLoop) {
			newLoop = new WhileLoop(newCond, l.getBody().clone());
		} else {
			newLoop = new DoLoop(l.getBody().clone(), newCond);
		}
		return newLoop;
	}

	/**
	 * builds an expression that is later set into the condition of the loop, so that it splits the loop according to partitions
	 * @param iv initial value of the iterationVariable, either as Identifier saving the value, or as Literal containing the value
	 * @param ev same as iv, only with the last (end) value of the iterationVariable
	 * @param partitions partitions that hold information about how big / long the splitted loops should be
	 * @param i the current iteration (first splitted loop is i = 1)
	 * @return a BinaryExpression that can be set into the condition of the loop
	 */
	private Expression buildDistributionExpression(Traversable iv, Traversable ev, ArrayList<Integer> partitions, int i) {
		iv = cloneTraversable(iv);
		ev = cloneTraversable(ev);

		Optional<Long> ivLong = Optional.empty();
		Optional<Long> evLong = Optional.empty();


		if (iv instanceof Expression) {
			ivLong = ((Expression) iv).visit(new EvaluateExpression()).map(BigInteger::longValueExact);
		}

		if (ev instanceof Expression) {
			evLong = ((Expression) ev).visit(new EvaluateExpression()).map(BigInteger::longValueExact);
		}

		if (ivLong.isPresent() && evLong.isPresent()) {
			final int mult = partitions.stream().limit(i).mapToInt(Integer::intValue).sum();
			final int div = partitions.stream().mapToInt(Integer::intValue).sum();
			long val = ivLong.get().longValue()+((evLong.get().longValue()-ivLong.get().longValue())*mult)/div;
			final IntegerLiteral valLiteral = new IntegerLiteral(val);
			return valLiteral;
		}

		IntegerLiteral multLiteral = new IntegerLiteral(partitions.stream().limit(i).mapToInt(Integer::intValue).sum(), "l");
		IntegerLiteral divLiteral = new IntegerLiteral(partitions.stream().mapToInt(Integer::intValue).sum(), "l");
		BinaryExpression be = new BinaryExpression((Expression) ev, RegularBinaryOperator.SUBTRACT, (Expression) iv);
		be = new BinaryExpression(be, RegularBinaryOperator.MULTIPLY, multLiteral);
		be = new BinaryExpression(be, RegularBinaryOperator.DIVIDE, divLiteral);

		iv = cloneTraversable(iv);
		be = new BinaryExpression((Expression) iv, RegularBinaryOperator.fromString("+"), be);
		return be;
	}

	/**
	 * builds an expression that is later set into the condition of the loop, so that the loop is split into even blocks of iterationVariable
	 * @param iv initial value of the iterationVariable, either as Identifier saving the value, or as Literal containing the value
	 * @param ev same as iv, only with the last (end) value of the iterationVariable
	 * @param i the current iteration (first splitted loop is i = 1)
	 * @param nos Number of Splits
	 * @return a BinaryExpression that can be set into the condition of the loop
	 */
	private Expression buildSimpleExpression(Traversable iv, Traversable ev, int i, int nos) {
		iv = cloneTraversable(iv);
		ev = cloneTraversable(ev);

		Optional<Long> ivLong = Optional.empty();
		Optional<Long> evLong = Optional.empty();


		if (iv instanceof Expression) {
			ivLong = ((Expression) iv).visit(new EvaluateExpression()).map(BigInteger::longValueExact);
		}

		if (ev instanceof Expression) {
			evLong = ((Expression) ev).visit(new EvaluateExpression()).map(BigInteger::longValueExact);
		}


		if (ivLong.isPresent() && evLong.isPresent()) {
			final int mult = i;
			final int div = nos;
			long val = ivLong.get().longValue()+((evLong.get().longValue()-ivLong.get().longValue())*mult)/div;
			final IntegerLiteral valLiteral = new IntegerLiteral(val);
			return valLiteral;
		}

		IntegerLiteral iLiteral = new IntegerLiteral(i, "l");
		IntegerLiteral nosLiteral = new IntegerLiteral(nos, "l");
		BinaryExpression be = new BinaryExpression((Expression) ev, RegularBinaryOperator.SUBTRACT, (Expression) iv);
		be = new BinaryExpression(be, RegularBinaryOperator.MULTIPLY, iLiteral);
		be = new BinaryExpression(be, RegularBinaryOperator.DIVIDE, nosLiteral);

		iv = cloneTraversable(iv);
		be = new BinaryExpression((Expression) iv, RegularBinaryOperator.fromString("+"), be);
		return be;
	}

	/**
	 * construct a list that holds sums of perf
	 * with as few as possible elements, while no element shall be bigger than maxCycles
	 * @param perf performance counters
	 * @param maxCycles maximum of cycles that one loop should run
	 * @return list
	 */
	private ArrayList<Integer> getMaxCyclesPartitions(ArrayList<Integer> perf, int maxCycles) {
		ArrayList<Integer> result = new ArrayList<>();
		int currentPos = 0;
		int value = 0;
		while (true) {
			if (value+perf.get(currentPos)>maxCycles) {
				result.add(currentPos);
				value = 0;
			}
			value += perf.get(currentPos);
			currentPos++;
			if (currentPos>=perf.size()) {
				result.add(currentPos-1);
				break;
			}
		}
		return result;
	}

	private Traversable cloneTraversable(Traversable t) {
		if (t instanceof Identifier) {
			return ((Identifier) t).clone();
		} else if (t instanceof IntegerLiteral) {
			return ((IntegerLiteral) t).clone();
		} else if (t instanceof FloatLiteral) {
			return ((FloatLiteral) t).clone();
		} else if (t instanceof NameID) {
			return ((NameID) t).clone();
		} else if (t instanceof Expression) {
			return ((Expression) t).clone();
		} else {
			throw new RuntimeException("The initialization or condition of the iteration value could not be cloned. Please check if it is of type FloatLiteral, IntegerLiteral, Expression  or an Identifier.");
		}
	}

	/**
	 * construct a list that holds sums of perf Integers. The size of the list shall be nos
	 * and the distribution of the new list shall be as close to distribution as possible while maintaining the order
	 * @param perf the measured performance of each iteration
	 * @param distribution targeted distribution of iterations
	 * @return list of Integers representing the partitions
	 */
	private ArrayList<Integer> getUnevenPartitions(ArrayList<Integer> perf, ArrayList<Integer> distribution) {
		ArrayList<Integer> result = new ArrayList<>();
		ArrayList<Integer> target = new ArrayList<>();
		int perfSum = perf.stream().mapToInt(Integer::intValue).sum();
		int distributionSum = distribution.stream().mapToInt(Integer::intValue).sum();
		for (Integer d : distribution) {
			target.add(perfSum*d/distributionSum);
		}
		int perfPos = 0;
		int targetPos = 0;
		int value = 0;
		while (true) {
			if (abs(value-target.get(targetPos)) < abs(value+perf.get(perfPos)-target.get(targetPos))) {
				result.add(perfPos);
				targetPos++;
				value = 0;
			}
			value += perf.get(perfPos);
			perfPos++;
			if (perfPos>=perf.size()) {
				result.add(perfPos-1);
				break;
			}
		}
		return result;
	}

	/**
	 * construct a list that holds sums of perf Integers. The size of the list shall be nos and
	 * all sums shall be as close together as possible while maintaining order
	 * @param perf the measured performance of each iteration
	 * @param nos the number of splits that is wanted
	 * @return a list of integers representing the partitions
	 */
	private ArrayList<Integer> getEvenPartitions(ArrayList<Integer> perf, int nos) {
		ArrayList<Integer> result = new ArrayList<>();
		int target = perf.stream().mapToInt(Integer::intValue).sum()/nos;
		int currentPos = 0;
		int value = 0;
		while (true) {
			if (abs(value-target)  < abs(value+perf.get(currentPos)-target)) {
				result.add(currentPos);
				value = 0;
			}
			value += perf.get(currentPos);
			currentPos++;
			if (currentPos>=perf.size()) {
				result.add(currentPos-1);
				break;
			}
		}
		return result;
	}

	/**
	 * returns the Literal or Identifier that is saving the starting value for the iteration variable
	 * @param l the loop
	 * @return IntegerLiteral
	 */
	public Traversable getInitialValue(ForLoop l) {
		Statement is =  l.getInitialStatement();
		if (is instanceof NullStatement) {
			if (l.getParent() instanceof CompoundStatement) {
				List<Traversable> children = l.getParent().getChildren();
				if (children.size()>=2 && children.get(0) instanceof DeclarationStatement
					&& children.get(0).getChildren().get(0) instanceof VariableDeclaration
					&& children.get(0).getChildren().get(0).getChildren().get(0) instanceof VariableDeclarator
					&& children.get(0).getChildren().get(0).getChildren().get(0).getChildren().get(1) instanceof Initializer
					&& children.get(0).getChildren().get(0).getChildren().get(0).getChildren().get(1).getChildren().get(0) instanceof Traversable
					&& children.get(1)==l) {
					return children.get(0).getChildren().get(0).getChildren().get(0).getChildren().get(1).getChildren().get(0);
				} else {
					return null;
				}
			} else {
				return null;
			}
		} else {
			AssignmentExpression ae  = ((AssignmentExpression) is.getChildren().get(0));
			return ae.getChildren().get(1);
		}
	}

	/**
	 * Tries to determine the initial value of the iteration variable of the given WhileLoop.
	 * Looks at statements within the same compound parent as the loop. The last statement to change
	 * the iteration variable has to an assignment or the initialization
	 * @param loop the while loop
	 * @param iterationVariable the iteration variable
	 * @return the Initial Value if it could be determined, otherwise null
	 */
	public Traversable getInitialValueWhileLoop(WhileLoop loop, Identifier iterationVariable) {

		if (loop == null) {
			log.error("The given loop is null");
			return null;
		}

		if (iterationVariable == null) {
			log.info("Can't determine the initial Value, without the iteration variable");
			return null;
		}

		Traversable parent = loop.getParent();
		if (parent instanceof CompoundStatement) {

			//Find Last Assignment value
			List<Identifier> iterationVariableOccurrences = getAllIdentifierOccurrencesBefore(iterationVariable, parent, loop);
			List<Identifier> relevantOccurrences = getWriteOccurrences(iterationVariableOccurrences);
			if (relevantOccurrences == null) {
				log.info("Can not determine initial value of the iteration variable, if not all accesses are known");
				return null;
			}

			if (relevantOccurrences.isEmpty()) {
				//No changes between Loop and declaration of variable
				return getDeclarationValue(iterationVariable,parent);

			}

			boolean addressOperatorOccurs = relevantOccurrences.stream()
					.map(Traversable::getParent)
					.anyMatch(p -> p instanceof UnaryExpression && ((UnaryExpression)p).getOperator().equals(UnaryOperator.ADDRESS_OF));
			if (addressOperatorOccurs) {
				log.info("Value change could be unnoticed when accessed with pointer");
				return null;
			}

			Statement lastStatement = relevantOccurrences.stream()
					.map(id -> CetusUtils.findParentOfType(id, Statement.class))
					.max(Comparator.comparing(Statement::where))
					.get();

			if (lastStatement.getParent().equals(parent)) {
				Traversable init;
				if (lastStatement instanceof DeclarationStatement) {
					try {
						init = lastStatement.getChildren().stream()
								.filter(t -> t instanceof VariableDeclaration)
								.flatMap(t -> t.getChildren().stream())
								.filter(t -> t instanceof VariableDeclarator)
								.flatMap(t -> t.getChildren().stream())
								.filter(t -> t instanceof Initializer)
								.flatMap(t -> t.getChildren().stream())
								.collect(StreamUtil.exactlyOne());
					} catch (IllegalStateException e) {
						log.info("Initializer of the iteration Variable could not be found or evaluated");
						return null;
					}

				} else if (lastStatement instanceof ExpressionStatement) {
					try {
						init = lastStatement.getChildren().stream()
								.filter(t -> t instanceof AssignmentExpression)
								.map(t -> ((AssignmentExpression)t).getRHS())
								.collect(StreamUtil.exactlyOne());
					} catch (IllegalStateException e) {
						log.info("The iteration variable changes between last assignment and loop");
						return null;
					}
				} else {
					log.info("Initial Value of the iteration variable could not be determined");
					return null;
				}
				return init;
			}
			//TODO : changes between init and loop
		}
		log.info("Loop not inside a compound statement, can not find the initial value of the iteration variable");
		return null;

	}

	/**
	 * Compares to statements according to their line number,
	 * if both statements are inside the same Procedure of the same TranslationUnit
	 * @param statement the statement to check
	 * @param boundary the statement to check against
	 * @return true, if both statements are in the same procedure and the statement to check is before the statement to check against
	 */
	private boolean isBeforeStatement(Statement statement, Statement boundary) {

		//same Translation Unit?
		TranslationUnit statementTransUnit = CetusUtils.getLocation(statement);
		TranslationUnit boundaryTransUnit = CetusUtils.getLocation(boundary);
		if (statementTransUnit == null || boundaryTransUnit == null || !statementTransUnit.equals(boundaryTransUnit)) {
			return false;
		}

		//same Procedure?
		Procedure statementProcedure = CetusUtils.findParentOfType(statement, Procedure.class);
		Procedure boundaryProcure = CetusUtils.findParentOfType(boundary, Procedure.class);
		if (statementProcedure == null || boundaryProcure == null || !statementProcedure.equals(boundaryProcure)) {
			return false;
		}

		return statement.where() < boundary.where();
	}

	/**
	 * Filters the List to get all write occurrences. Returns null, if at least one occurrence type
	 * can not be determinded
	 * @param allOccurrences the occurrences of Identifier to filter
	 * @return the occurrences with write access or null, if at least one occurrence is of unknown type
	 */
	private List<Identifier> getWriteOccurrences(List<Identifier> allOccurrences) {
		VariableAccessAnalyzer variableAccessAnalyzer = new VariableAccessAnalyzer(model);

		boolean containsUnknownAccessBehavior = allOccurrences.stream()
				.anyMatch(t -> variableAccessAnalyzer.determineAccessBehavior(t).equals(VariableAccess.AccessBehavior.UNKNOWN));

		if (containsUnknownAccessBehavior) {
			log.info("Can't determine the access behaviour of all occurrences");
			return null;
		}

		return allOccurrences.stream()
				.filter(t -> variableAccessAnalyzer.determineAccessBehavior(t).equals(VariableAccess.AccessBehavior.WRITE)
						||variableAccessAnalyzer.determineAccessBehavior(t).equals(VariableAccess.AccessBehavior.READWRITE) )
				.collect(Collectors.toList());
	}

	/**
	 * Searches for all occurrences of the given Identifier inside the given Traversable,
	 * that are in before the given Statement.
	 * @param identifier the identifier to look for
	 * @param traversable the traversable to look in
	 * @param boundary the statement inside the same traversable, all occurrences are in front of
	 * @return the identifiers inside the Traversable, that are before the given statement
	 */
	private List<Identifier> getAllIdentifierOccurrencesBefore(Identifier identifier, Traversable traversable, Statement boundary) {
		List<Identifier> iterationVariableOccurrences = new ArrayList<>();
		DFIterator occurrences = new DFIterator<>(traversable, Identifier.class);
		while (occurrences.hasNext()) {
			Identifier currIdentifier = (Identifier) occurrences.next();
			if (currIdentifier.equals(identifier)) {
				Statement parentStatement = CetusUtils.findParentOfType(currIdentifier,Statement.class);
				if (isBeforeStatement(parentStatement, boundary)) {
					iterationVariableOccurrences.add(currIdentifier);
				}
			}
		}
		return iterationVariableOccurrences;
	}

	/**
	 * Searches the given Traversable for the declaration and the initial value (IntegerLiteral)
	 * @param identifier The Identifier being initialized
	 * @param traversable the traversable where to look
	 * @return the initial Value, if its an IntegerLiteral, otherwise null
	 */
	private IntegerLiteral getDeclarationValue(Identifier identifier, Traversable traversable) {
		DFIterator varDecsIterator = new DFIterator(traversable, VariableDeclarator.class);
		List<VariableDeclarator> allVarDecs = new ArrayList<>();
		while (varDecsIterator.hasNext()) {
			VariableDeclarator currVarDeclarator = (VariableDeclarator) varDecsIterator.next();
			allVarDecs.add(currVarDeclarator);
		}

		//Search for exactly one VariableDeclarator for the identifier
		VariableDeclarator declarator;
		try {
			declarator = allVarDecs.stream().filter( vd -> {
				return vd.getChildren().stream()
						.flatMap(StreamUtil.castOrEmpty(NameID.class))
						.anyMatch(id -> id.equals(identifier));
			}).collect(StreamUtil.exactlyOne());
		} catch (IllegalStateException e) {
			log.info("Can't identify the declaration statement of the iteration variable of the loop");
			return null;
		}

		//Get the initial Value, which has to be an integer literal
		IntegerLiteral initialValue;
		try {
			initialValue = declarator.getChildren().stream()
					.flatMap(t -> t.getChildren().stream())
					.flatMap(StreamUtil.castOrEmpty(IntegerLiteral.class))
					.collect(StreamUtil.exactlyOne());
		} catch (IllegalStateException e) {
			log.info("Initial value of the iteration variable is not an IntegerLiteral");
			return null;
		}
		return initialValue;
	}

	/**
	 * Searches the body of the given {@link WhileLoop} for the step.
	 * Can only evaluate the step if it is not inside a nested loop or conditional statement.
	 * Can't evaluate the step, if it consists of multiple statements (for now)
	 * @param loop the while loop
	 * @param iterationVariable the iteration variable
	 * @return The step, or null if step cant be evaluated
	 */
	public Expression getStepOfWhileLoop(WhileLoop loop, Identifier iterationVariable) {
		if (iterationVariable == null) {
			log.info("Step can not be evaluated automatically");
			return null;
		}

		//Find all occurrences of the iterationVariable inside the loops body
		List<Identifier> iterationVariableOccurrences = new ArrayList<>();
		DFIterator iterator = new DFIterator<>(loop.getBody(), Identifier.class);
		while (iterator.hasNext()) {
			Identifier currIdentifier = (Identifier) iterator.next();
			if (currIdentifier.equals(iterationVariable)) {
				iterationVariableOccurrences.add(currIdentifier);
			}
		}

		//Filter for relevant occurrences (write access)
		VariableAccessAnalyzer variableAccessAnalyzer = new VariableAccessAnalyzer(model);
		List<Identifier> relevantIdentifiers = iterationVariableOccurrences.stream()
				.filter(t -> variableAccessAnalyzer.determineAccessBehavior(t).equals(VariableAccess.AccessBehavior.WRITE)
						||variableAccessAnalyzer.determineAccessBehavior(t).equals(VariableAccess.AccessBehavior.READWRITE) )
				.collect(Collectors.toList());

		boolean containsUnknownAccesBehavior = iterationVariableOccurrences.stream()
				.anyMatch(t -> variableAccessAnalyzer.determineAccessBehavior(t).equals(VariableAccess.AccessBehavior.UNKNOWN));
		if (containsUnknownAccesBehavior) {
			log.info("Can't determine the access behaviour of all occurrences of the iteration Variable inside the loops body");
			return null;
		}

		//Step can not be evaluated if inside conditional statement
		boolean stepNested = relevantIdentifiers.stream()
				.map(Traversable::getParent)
				.anyMatch(parent -> {
					while (parent != null && !(parent.equals(loop.getBody()))) {
						if (parent instanceof IfStatement || parent instanceof SwitchStatement|| parent instanceof ConditionalExpression) {
							log.info("Step inside conditional statement, can't determine step");
							return true;
						}
						if (parent instanceof Loop) {
							log.info("Identifier inside nested loop, can't determine step"); //TODO : evaluate if step is still consistent for each iteration
							return true;
						}
						parent = parent.getParent();
					}
					if (parent == null) {
						log.debug("Identifier not inside Loop body, this should not happen");
						return true;
					}
					return false;
				});
		if (stepNested) {
			return null;
		}

		//Evaluate changes to iteration variable
		if (relevantIdentifiers.isEmpty()) {
			log.info("Can't find step inside loop body");
			return null;
		} else if (relevantIdentifiers.size() == 1) {
			Identifier stepIdentifier = relevantIdentifiers.stream().collect(StreamUtil.exactlyOne());
			if (stepIdentifier.getParent() instanceof UnaryExpression || stepIdentifier.getParent() instanceof AssignmentExpression) {
				return  (Expression) stepIdentifier.getParent();
			}
		} else {
			log.info("Found multiple changes to the iteration variable inside the loop body. Iteration step can't be determined"); //TODO: combine steps to one step if possible
			return null;
		}
		return null;
	}


	/**
	 * returns the Literal or Identifier that is saving the last (end) value for the iteration variable
	 * @param cond Expression being the condition of the loop
	 * @param i Identifier representing the iterationVariable
	 * @return Literal or Identifier containing the last value
	 */
	public Traversable getEndValue(Expression cond, Identifier i) {
		Traversable result;
		if (cond.getChildren().get(1).equals(i)) {
			result = cond.getChildren().get(0);
		} else {
			result = cond.getChildren().get(1);
		}
		return result;
	}

	/**
	 * returns the iterationVariable of the loop. retrieves it from the LoopPragma if it is specified there
	 * @param l the loop to search
	 * @param tag PragmaTag of type Constants.PragmaLoop.ITERATION_VARIABLE or null if tag doesn't hold the iteration variable
	 * @return the identifier
	 */
	public Identifier getIterationVariable(Loop l, PragmaTag tag) {

		Identifier iterationVariable = null;

		if (tag!=null) {
			String ivString = String.valueOf(tag.getEntryNames().toArray()[0]);
			iterationVariable = LoopTransformer.getIdentifier(l, ivString);
			if (iterationVariable == null) {
				model.printIgnoreableError(log,"The Variable " + ivString + " in the  Pragma " + Constants.PragmaLoop.ITERATION_VARIABLE
					+ " could not be found in the Loop. Please check for spelling mistakes");
				return null;
			}
		} else if (l instanceof ForLoop) {
			Expression step = ((ForLoop) l).getStep();
			if (step.getChildren().get(0) instanceof Identifier) {
				iterationVariable = (Identifier) step.getChildren().get(0);
			}
		} else {
			Expression cond = l.getCondition();
			if (!(cond.getChildren().get(0) instanceof Identifier && cond.getChildren().get(1) instanceof Identifier)) {
				iterationVariable = (cond.getChildren().get(0) instanceof Identifier) ?
					((Identifier) cond.getChildren().get(0)) : ((Identifier) cond.getChildren().get(1));
			}
		}

		if (iterationVariable==null) {
			model.printIgnoreableError(log,"IterationVariable could not be found automatically. Please specify a PragmaTag iterationVariable().");
		}
		return iterationVariable;
	}
}
