/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.loopOptimizer.transformationPasses;


import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.loopOptimizer.model.LoopRuntimeReport;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.LoopPragma;
import de.tu_darmstadt.rs.microStreams.loopOptimizer.tools.LoopTransformer;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.Pragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.PragmaAnnotationParser;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;

import cetus.hir.CompoundStatement;
import cetus.hir.Declaration;
import cetus.hir.DepthFirstIterator;
import cetus.hir.FlatIterator;
import cetus.hir.Loop;
import cetus.hir.PragmaAnnotation;
import cetus.hir.Program;
import cetus.hir.Statement;
import cetus.hir.Traversable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Optimize the loops here
 */
public class LoopOptimizerPass extends AbstractTransformationPass {

	private Model model;
	private final Logger log = LoggerFactory.getLogger(LoopOptimizerPass.class);

	public LoopOptimizerPass() {
	}

	@Override
	public boolean init() {
		return true;
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				CetusProgram.class, LoopRuntimeReport.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				CetusProgram.class
		);
	}

	@Override
	public boolean process(Model model) {
		//search for split pragmas and call the transformer

		this.model = model;
		Program program = CetusUtils.getProgram(model);

		/* Search all pragmas. */
		searchAnnotations(program);

		return true;

	}

	/**
	 * Recursive search of of all microStreams Task pragmas in the program.
	 *
	 * @param parentTraversable a traversable of the parent
	 */
	private void searchAnnotations(Traversable parentTraversable) {

		DepthFirstIterator<Traversable> iter = new DepthFirstIterator<>(parentTraversable);
		int autoperfCounter = 0;

		while (iter.hasNext()) {

			Traversable tChild = iter.next();
			Pragma pragma = getPragma(tChild);

			if (pragma instanceof LoopPragma) {
				if (iter.hasNext()) {
					Traversable loop = iter.next();
					//if loop is pointing on a compound Statement outside the loop, find the loop
					Boolean compoundStatementUsed = false;
					if (!(loop instanceof Loop)) {
						FlatIterator flatIterator = new FlatIterator(loop);
						while (flatIterator.hasNext()) {
							Traversable elem = flatIterator.next();
							if (elem instanceof Loop) {
								loop = elem;
								compoundStatementUsed = true;
								break;
							}
						}
						if (!(loop instanceof Loop)) {
							model.printIgnoreableError(log,"There is no loop behind the LoopPragma Statement.");
							break;
						}
					}
					ArrayList<Traversable> result = new LoopTransformer(model).transform((Loop) loop, (LoopPragma) pragma, autoperfCounter, compoundStatementUsed);
					if (pragma.getPragmaTag(Constants.PragmaLoop.AUTOPERF)!=null) {
						autoperfCounter++;
					}
					if (result != null) {
						if (compoundStatementUsed) {
							CompoundStatement compound = (CompoundStatement) loop.getParent();
							setResult(compound, result);
						} else {
							setResult((Statement) loop, result);
						}
					}
				} else {
					log.warn("LoopPragma found but not followed by loop: " + CetusUtils.getStringFromPrintable(tChild) + " Skipping for now ...");
				}
				Traversable parent = tChild.getParent();
				parent.removeChild(tChild);
			}
		}
	}

	/**
	 * replaces one Statement by all Traversable in result
	 * @param toReplace the Statement to replace
	 * @param result the traversables to set
	 */
	public static void setResult(Statement toReplace, ArrayList<Traversable> result) {
		if (toReplace.getParent() instanceof CompoundStatement) {
			CompoundStatement parent = (CompoundStatement) toReplace.getParent();
			for (Traversable t : result) {
				if (t instanceof Declaration) {
					parent.addDeclarationAfterLastDeclaration((Declaration)t);
				} else if (t instanceof Statement) {
					parent.addStatementBefore(toReplace, (Statement) t);
				} else {
					throw new RuntimeException("Could not add a Traversable because it was not of type Statement. This should not happen.");
				}
			}
			parent.removeChild(toReplace);
		} else {
			throw new RuntimeException("Could not replace toReplace loop because it is not wrapped by a CompoundStatement. This should not happen");
		}
	}

	private Pragma getPragma(Traversable traversable) {
		PragmaAnnotation pragmaAnnotation = CetusUtils.getAnnotation(traversable);
		// Check whether the PragmaAnnotation is a valid UStreams Annotation.
		return new PragmaAnnotationParser(model).parse(pragmaAnnotation, traversable);
	}
}
