/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;

import edu.uci.ics.jung.graph.DirectedOrderedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Runs several workers in the order defined by the workers processable method
 */
public class TransformationPassRunner {

	private static final Logger log = LoggerFactory.getLogger(TransformationPassRunner.class);
	Graph<TransformationPass, CustomEdge> g = new DirectedOrderedSparseMultigraph<>();
	private Model model;


	public TransformationPassRunner(Model model) {
		this.model = model;
	}

	/**
	 * starts the workers
	 *
	 * @return true if everything went fine
	 */
	@SuppressWarnings({"checkstyle:MultipleStringLiterals"})
	public boolean process() {

		generateGraph();

		for (TransformationPass worker : model.getAllWorkers()) {
			worker.init();
		}

		while (model.getAllWorkers().size() - model.countAppliedWorkers() > 0) {

			final List<TransformationPass> executableWorkers = model.getAllWorkers().stream()
					.filter(w -> !model.isWorkerApplied(w.getClass()))
					.filter(w -> w.processable(model))
					.collect(Collectors.toList());
			sortWorkers(executableWorkers);
			if (executableWorkers.size() == 0) {
				printWorkersNotProcessableError();
				return false;
			} else {
				if (!runWorker(executableWorkers.stream().findFirst().get())) {
					return false;
				}
			}
		}
		return true;
	}

	private void sortWorkers(List<TransformationPass> executableWorkers) {
		Comparator<TransformationPass> isExecutable = (TransformationPass w1, TransformationPass w2) -> Boolean.compare(w1.processable(model), w2.processable(model));
		Comparator<TransformationPass> isOptimizationPass = Comparator.comparingInt(w -> workerOptimises(w).size());
		Comparator<TransformationPass> providesRequirement = (TransformationPass w1, TransformationPass w2) -> {
			long providedMCFromW2ToW1 = w1.requiredModelComponents().stream().filter(mc -> w2.providedModelComponents().contains(mc)).count();
			long providedMCFromW1ToW2 = w2.requiredModelComponents().stream().filter(mc -> w1.providedModelComponents().contains(mc)).count();
			if (providedMCFromW2ToW1 > providedMCFromW1ToW2) {
				return 1;
			} else if (providedMCFromW2ToW1 == providedMCFromW1ToW2) {
				return 0;
			} else {
				return -1;
			}
		};
		Comparator<TransformationPass> providesMore = Comparator.comparingInt(w -> w.providedModelComponents().size());
		Comparator<TransformationPass> leadsToOptimizationPass = Comparator.comparingInt(w -> workerLeadsToOptimizationPass(w));
		Comparator<TransformationPass> name = Comparator.comparing(w -> w.getClass().getSimpleName());
		Comparator<TransformationPass> workerComparator = (TransformationPass w1, TransformationPass w2) -> isExecutable.thenComparing(providesRequirement).thenComparing(providesMore.reversed()).thenComparing(leadsToOptimizationPass).thenComparing(name).compare(w1, w2);
		Collections.sort(executableWorkers, workerComparator);
	}

	private void generateGraph() {
		for (TransformationPass worker : model.getAllWorkers()) {
			g.addVertex(worker);

		}
		for (TransformationPass worker : model.getAllWorkers()) {
			for (Class<? extends ModelComponent> mc : worker.requiredModelComponents()) {
				model.getAllWorkers().stream()
						.filter(w -> w.providedModelComponents().contains(mc))
						.forEach(w -> {
							//try {
							g.addEdge(new ModelComponentEdge(mc), worker, w);
							//} catch (IllegalArgumentException e) {

							//}
						});
			}

			worker.requiredTransformationPasses().stream().forEach(w -> g.addEdge(new DirectDependencyEdge(), worker, model.getAllWorkers().stream().filter(worker1 -> worker1.getClass().equals(w)).findAny().get()));
		}
	}


	private void printWorkersNotProcessableError() {
		List<Class> notAppliedWorkerClasses = model.getNotAppliedWorkers().stream().map(x -> x.getClass()).collect(Collectors.toList());
		log.error(String.format("Cannot finish proccessing %d workers, because they are not processable: ", model.getAllWorkers().size() - model.countAppliedWorkers()));
		model.getNotAppliedWorkers().stream()
				.forEach(w -> {
					log.error("");
					log.error(String.format("\t\t%-35s Executable(%b)", w.getClass().getSimpleName(), w.processable(model)));
					log.error(String.format("\t\t\t\trequired Workers: \t\t\t%s", w.requiredTransformationPasses().stream().map(c -> c.getSimpleName()).collect(Collectors.joining(", "))));
					log.error(String.format("\t\t\t\trequired ModelComponents: \t%s", w.requiredModelComponents().stream().map(c -> c.getSimpleName()).collect(Collectors.joining(", "))));
					log.error(String.format("\t\t\t\tprovided ModelComponents: \t%s", w.providedModelComponents().stream().map(c -> c.getSimpleName()).collect(Collectors.joining(", "))));

					log.error(String.format("\t\t\t\tnot fulfilled Requirements: %s %s", w.requiredTransformationPasses().stream().filter(x -> !model.isWorkerApplied(x)).map(x -> x.getSimpleName()).collect(Collectors.joining(", ")), w.requiredModelComponents().stream().filter(mc -> !model.hasModelComponent(mc)).map(x -> x.getSimpleName()).collect(Collectors.joining(", "))));

				});

		log.error("The following workers are missing: " + model.getNotAppliedWorkers().stream()
				.flatMap(w -> w.requiredTransformationPasses().stream())
				.filter(w -> !model.isWorkerApplied(w))
				.filter(w -> !notAppliedWorkerClasses.contains(w))
				.map(c -> c.getSimpleName())
				.collect(Collectors.toList()));
	}


	private boolean runWorker(TransformationPass worker) {
		log.info(String.format("The Worker\t %-30s started...", worker.getClass().getSimpleName()));
		long startTime   = System.currentTimeMillis();
		if (!worker.process(model)) {
			log.error("The Worker "
					+ worker.getClass().getSimpleName()
					+ " did not finish successfully. See log for further information...");
			return false;
		} else {
			model.addAppliedWorker(worker.getClass());
			long endTime   = System.currentTimeMillis();
			long totalTime = endTime - startTime;
			log.info(String.format("\t %-42s ...and finished successfully after \t%-4s ms.", "", totalTime));
			return true;
		}
	}

	private Set<Class<? extends ModelComponent>> workerOptimises(TransformationPass worker) {
		Set<Class<? extends ModelComponent>> modelComponentsOptimizedInThisPass = new HashSet(worker.providedModelComponents());
		modelComponentsOptimizedInThisPass.retainAll(worker.requiredModelComponents());
		return modelComponentsOptimizedInThisPass;
	}

	private int workerLeadsToOptimizationPass(TransformationPass worker) {
		return g.getSuccessors(worker).stream().mapToInt(w -> workerOptimises(w).size()).sum();
	}

	/**
	 * represents a dummy edge
	 */
	private interface CustomEdge {

		String print();
	}


	/**
	 * a special edge
	 */
	private class ModelComponentEdge implements CustomEdge {

		Class<? extends ModelComponent> mc;

		public ModelComponentEdge(Class<? extends ModelComponent> mc) {
			this.mc = mc;
		}

		@Override
		public String print() {
			return mc.getSimpleName();
		}
	}

	/**
	 * a special edge
	 */
	private class DirectDependencyEdge implements CustomEdge {

		public DirectDependencyEdge() {

		}

		@Override
		public String print() {
			return "";
		}
	}

}
