/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.config;

import de.tu_darmstadt.rs.jcommander.converter.FileListConverter;
import de.tu_darmstadt.rs.jcommander.converter.ReplaceTildeWithUserHomeConverter;
import de.tu_darmstadt.rs.jcommander.validators.PathsExistValidator;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.util.FileUtil;

import com.beust.jcommander.Parameter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * holds the current configuration of the program, especially the cmd line arguments
 *
 * Any program that wants more options, should extend the Configuration and override the {@link IConfiguration#getProgramName()} function.
 * Additionally the Configuration should implement {@link IConfiguration} in order to be found by ArgumentParser.
 * If the extended Configuration should be found by {@link Model#getConfig()} it should write its own interface extending {@link IConfiguration}.
 */
public class Configuration implements IConfiguration {

	Model model;

	/* collection of project files */
	@Parameter(required = true, description = "sourceFirmware folder/files", listConverter = ReplaceTildeWithUserHomeConverter.class)
	private List<String> sourceFiles = new ArrayList<>();

	/*
	 * Holds the project location where Firmwares and Hardware description will
	 * be written
	 */
	@Parameter(names = {"-p" , "--project"},required = false, description = "Should point to the directory where the spartanmc project to be modified resides.", converter = ReplaceTildeWithUserHomeConverter.class)
	private String targetRootDir = ".";

	/*
	 * Holds all C libraries for C headers that will be used in the firmware.
	 * Especially for peri headers
	 */
	@Parameter(names = {"-l" , "--libs"}, description = "Give path to additional libraries i.e. header files. Paths are separated by comma.", listConverter = FileListConverter.class, validateValueWith = PathsExistValidator.class)
	private List<File> additionalSystemLibraries = new ArrayList<>();

	/* Tells weather we want performance measurement or not. Creates HW and SW. */
	@Parameter(names = {"-e" , "--evaluate"}, description = "Add constructs for a performance measurement evaluation on each core.")
	private boolean createPerformanceEvaluation = false;

	/* For a simulated target simprint instead of uart shall be used */
	@Parameter(names = {"-s" , "--sim"}, description = "Generate hardware for a simulated target.")
	private boolean createSimulatedTarget = false;

	/* Tells weather we want global Memory or not. Creates HW and SW. */
	@Parameter(names = {"-g" , "--no-global-mem"}, description = "Tells weather we want to move variables to global memory or not. Creates HW and SW.")
	private boolean noGlobalMem;

	/* Tells wether the spartanMC libraries shall be automatically added */
	@Parameter(names = {"-a" , "--autolibspartanmc"}, description = "Automatically detect SpartanMC lib folder.")
	private boolean spartanMCAutoLibDetection = false;

	@Parameter(names = {"-d", "--dma"} , description = "Use DMA Interconnects")
	private boolean useDMA = false;

	@Parameter(names = {"--ignoreErrors"} , description = "Try to ignore Erros")
	private boolean ignoreErrors = false;

	private Logger log = LoggerFactory.getLogger(Configuration.class);


	public Configuration() {
		super();
	}

	public Collection<File> getAdditionalSystemLibraries() {
		return additionalSystemLibraries.stream().map(l -> Paths.get("").toAbsolutePath().resolve(l.toPath()).toFile()).collect(Collectors.toList());
	}

	/**
	 * sets additional libs
	 * @param libsCperi additional libs to add
	 */
	public void setAdditionalSystemLibraries(String... libsCperi) {

		for (String libloc : libsCperi) {
			File f = new File(FileUtil.replaceTildeWithUserHome(libloc));
			if (!f.isDirectory()) {
				throw new IllegalArgumentException(libloc + " is not a directory");
			} else if (!f.exists()) {
				throw new IllegalArgumentException(libloc + " does not exist");
			} else if (!f.canRead()) {
				throw new IllegalArgumentException(libloc + " can not be read");
			} else {
				additionalSystemLibraries.add(f);
			}
		}
	}

	public String getTargetRootDir() {
		return FileUtil.replaceTildeWithUserHome(targetRootDir);
	}

	public boolean isCreatePerformanceEvaluation() {
		return createPerformanceEvaluation;
	}

	@Override
	public void setCreatePerformanceEvaluation(boolean createPerformanceEvaluation) {
		this.createPerformanceEvaluation = createPerformanceEvaluation;
	}

	public boolean isCreateSimulatedTarget() {
		return createSimulatedTarget;
	}

	public boolean isSpartanMCAutoLibDetection() {
		return spartanMCAutoLibDetection;
	}

	public boolean useGlobalMemory() {
		return !noGlobalMem;
	}

	public boolean useDMA() {
		return useDMA;
	}

	public boolean addSourceFileAsPath(List<Path> files) {
		return addSourceFileAsString(files.stream().map(f -> f.toAbsolutePath().toString()).collect(Collectors.toList()));
	}

	public boolean addSourceFileAsString(List<String> files) {
		return sourceFiles.addAll(files);
	}

	public void clearSourceFiles() {
		sourceFiles.clear();
	}

	public Collection<String> getSourceFiles() {
		return sourceFiles.stream().map(s -> FileUtil.replaceTildeWithUserHome(s)).collect(Collectors.toList());
	}

	public Collection<Path> getSourceFilesAsPath() {
		return sourceFiles.stream().map(s -> Paths.get(s)).collect(Collectors.toList());
	}

	public boolean ignoreErrors() {
		return ignoreErrors;
	}

	@Override
	public String getProgramName() {
		return Constants.MICROSTREAMS;
	}

	@Override
	public Model getModel() {
		return model;
	}

	@Override
	public void setModel(Model m) {
		model = m;
	}
}
