/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.config;

import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.tools.templates.TemplateFile;
import de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.EvaluationPeripheralCreator;
import de.tu_darmstadt.rs.util.FileUtil;

import java.io.File;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.regex.Pattern;

/**
 * Used Constants
 */
public final class Constants {

	public static final String REGEX_ENDS_WITH_NUMBER = "\\([0-9]+\\)$";
	public static final String REGEX_CASE_INSENSITIVE_STARTSWITH = "(?i)^";
	public static String MICROSTREAMS = "microStreams";
	public static String AUTOPERF = "AutoPerf";
	public static String LOOPOPTIMIZER = "LoopOptimizer";

	/**
	 *  Constants for Task extractor
	 */
	public static final class Pragma {

		public static final String PRAGMA = "pragma";
		public static final String SYMBOL = "#" + PRAGMA;
		public static final String AUTOPERF = SYMBOL + " " + Constants.AUTOPERF.toLowerCase();
		public static final String IDENTIFIER = MICROSTREAMS.toLowerCase();
		public static final String END = "end";
		public static final String MICROSTREAMS_REGEX = SYMBOL + " +" + IDENTIFIER;
		public static final String C_COMMENT_WITH_PRAGMA_REGEX = "(?:/\\* *" + MICROSTREAMS_REGEX + "(?:[^*]|(?:\\*+[^*/]))*\\*+/)|(?://.*" + MICROSTREAMS_REGEX + ".*\n)";
		public static final String TASK = "task";
		public static final String TASKGROUP = "taskgroup";
		public static final String VARIABLE_IN = "in";
		public static final String VARIABLE_OUT = "out";
		public static final String REPLICATE = "replicate";
		public static final String REGEX_VALID_REPLICATE = "^" + REPLICATE + REGEX_ENDS_WITH_NUMBER;
		public static final String REGEX_VALID_VARIABLE_LIST = "[-\\w]+(?:,\\s?[-\\w]+)*";
		public static final String REGEX_VALID_IN = "^" + VARIABLE_IN + "\\(" + REGEX_VALID_VARIABLE_LIST + "\\)$";
		public static final String REGEX_VALID_OUT = "^" + VARIABLE_OUT + "\\(" + REGEX_VALID_VARIABLE_LIST + "\\)$";
		public static final String MICRO_STREAMS_TASK = Constants.MICROSTREAMS + " " + Constants.Pragma.TASK;

		private Pragma() {
		}
	}


	/**
	 * Constants for loopOpt Parser
	 */
	public static final class PragmaLoop {
		public static final String LOOPOPT = "loopOpt";
		public static final String USE_FISSION = "useFission";
		public static final String USE_SPLITTING = "useSplitting";
		public static final String REGEX_VALID_NUM = "[0-9]+";
		public static final String AUTOPERF = "autoperf";
		public static final String NUMBER_OF_SPLITS = "NoS";
		public static final String DISTRIBUTION = "distribution";
		public static final String REGEX_DISTRIBUTION = REGEX_CASE_INSENSITIVE_STARTSWITH + DISTRIBUTION + "\\(\\d+(,\\d+)*\\)$";
		public static final String MAX_CYCLES = "maxCycles";
		public static final String REGEX_MAX_CYCLES = REGEX_CASE_INSENSITIVE_STARTSWITH + MAX_CYCLES + REGEX_ENDS_WITH_NUMBER;
		public static final String ITERATION_VARIABLE = "iterationVariable";
		public static final String REGEX_IV = REGEX_CASE_INSENSITIVE_STARTSWITH + ITERATION_VARIABLE + "\\([\\w]+\\)$";

		private PragmaLoop() {
		}
	}


	/**
	 * C Files Constants
	 */
	public static final class C {

		public static final String SYSTEM_INCLUDE_PREFIX = "#include <";
		public static final String USER_INCLUDE_PREFIX = "#include \"";
		public static final String SYSTEM_INCLUDE_SUFFIX = ">";
		public static final String USER_INCLUDE_SUFFIX = "\"";
		public static final String MAIN_FUNCTION_NAME = "main";
		public static final String UNDEFINED = "undefined";
		public static final String VOID = "void";
		public static final String GLOBAL_DECLARE = "DECLARE_GLOBAL";
		public static final String GLOBAL_DEFINE = "DEFINE_GLOBAL";
		public static final String WRAP_INIT_VALUES = "WRAP_INIT_VALUES";
		public static final String PRINTF = "printf";
		public static final String SCANF = "scanf";

		private C() {
		}

		public static String buildSystemInclude(String headerfile) {
			return SYSTEM_INCLUDE_PREFIX + headerfile + SYSTEM_INCLUDE_SUFFIX;
		}

		public static String buildUserInclude(String headerfile) {
			return USER_INCLUDE_PREFIX + headerfile + USER_INCLUDE_SUFFIX;
		}
	}


	/**
	 * CMD Line Arg Parser Constants
	 */
	public static final class CommandlineParser {

		public static final String CFILES_ID = "Files";
		public static final String JASPXML = "jsap.xml";
		public static final String LIBS = "Libs";
		public static final String LIBS_AUTODETECTION_SPARTANMC = "LibsAutodetect";
		public static final String PROJECTPATH = "ProjectPath";
		public static final String PERFORMANCEMEASUREMENT = "performanceEvaluation";
		public static final String DISABLEGLOBALMEM = "disableGlobalMemory";
		public static final String INSPECT_GENERATED_MODEL = "inspectGeneratedModel";

		private CommandlineParser() {
		}
	}


	/**
	 * constants for indent
	 */
	public static final class Indent {

		private static final String[] INDENT_COMMAND = {"indent", "inputFile", "-o", "outputFile", "-nbad", "-nbc", "-bbo", "-br", "-bli4", "-ncdb", "-nce", "-cp1", "-cs", "-di2", "-ndj", "-nfc1", "-nfca", "-hnl", "-i4", "-ip4", "-lp", "-pcs", "-nprs", "-npcs", "-psl", "-saf", "-sai", "-saw", "-nsc", "-nsob", "-npsl", "-brf", "-fc1"};
		private static final int COMMAND_INPUT_POS = 1;
		private static final int COMMAND_OUTPUT_POS = 3;


		private Indent() {
		}

		/**
		 * Cretes an indent Command
		 *
		 * @param inputFile the input file to be formatted
		 * @param outputFile the output file with the formatted code
		 * @return the indent command
		 */
		public static String[] getIndentCommand(File inputFile, File outputFile) {
			String[] command = INDENT_COMMAND;
			command[COMMAND_INPUT_POS] = inputFile.getAbsolutePath();
			command[COMMAND_OUTPUT_POS] = outputFile.getAbsolutePath();
			return command;
		}
	}


	/**
	 * SpartanMC specific Constants
	 */
	public static final class SpartanMC {

		public static final String CONFIG_BUILD_MK_LIBRARY_REGEX = "^(LIB_OBJ_FILES:=)(.*)$";
		public static final Pattern PREPROCESSOR_NOT_FOUND_INCLUDES_TO_IGNORE = Pattern.compile("((^subsystems\\/.+\\/|^)peripherals.h)$");


		/**
		 * Represents C Libs used in spartanmc
		 */
		public enum CLibraries {
			PERF(),
			MPSOC(),
			PERIPHERALS(),
			GLOBALVARIABLES(),
			STDIO(),
			STRING();

			private final String header;

			CLibraries() {
				this.header = null;
			}

			CLibraries(String header) {
				this.header = header;
			}

			public String toLowerCase() {
				return this.toString().toLowerCase();
			}

			/**
			 * gives the header file either constructed from the name of the lib + .h or a given name
			 * @return the according header file
			 */
			public String getHeader() {
				if (header == null) {
					return this.toLowerCase() + "." + FilesAndFolders.HEADERFILES_EXTENSION;
				} else {
					return header;
				}
			}

			public String buildSystemInclude() {
				return C.buildSystemInclude(this.getHeader());
			}

			public String buildUserInclude() {
				return C.buildUserInclude(this.getHeader());
			}
		}

		private SpartanMC() {
		}



	}


	/**
	 * Constants used with files and Folders
	 */
	public static final class FilesAndFolders {

		public static final Path TMP_FOLDER_MICROSTREAMS_USER = FileUtil.createNewTempFolder(MICROSTREAMS);
		public static final Path TEMPORARY_INCLUDE_FOR_ALL_PERIPHERALS = Paths.get(TMP_FOLDER_MICROSTREAMS_USER.toString(), "allHeaders.h");
		public static final String SOURCEFILES_EXTENSION = "c";
		public static final String HEADERFILES_EXTENSION = "h";
		public static final String C_FILES_EXT_PATTERN = "{*." + SOURCEFILES_EXTENSION + ", *." + HEADERFILES_EXTENSION + "}";
		public static final String TARGET_SUBDIR_PREFIX = "core";
		public static final String TEMPLATE_SUFFIX = ".ftl";
		public static final boolean OVERWRITE_EXISTING_FILES = true;
		public static final Include PERIPHERALS = new Include(SpartanMC.CLibraries.PERIPHERALS.getHeader(), true, false);
		public static final Include GLOBAL_VARIABLES = new Include(SpartanMC.CLibraries.GLOBALVARIABLES.getHeader(), true, false);
		public static final Include STRING = new Include(SpartanMC.CLibraries.STRING.getHeader(), true, false);

		public static final List<Include> DEFAULT_HEADERS = Arrays.asList(
				PERIPHERALS
		);
		public static final Include EVALUATION_HEADER = new Include("evaluation.h", true, SpartanMC.CLibraries.PERF.toLowerCase());

		private FilesAndFolders() {
		}
	}


	/**
	 * C language cetus constructs for Cetus
	 */
	public static final class Cetus {

		public static final String PREPROCESSOR_COMMAND = "spartanmc-cpp -C -I.";
		public static final String PREPROCESSOR_INCLUDE = " -I";
		private static final String startinclude = " startinclude #include *";
		private static final String endinclude = " endinclude";
		private static final String libIdentifier = "(\"[^\"]+\")";
		private static final String systemLibIdentifier = "(<[^>]+>)";
		public static final String SYSTEM_LIB_INCLUDE_REGEX = Pragma.SYMBOL + startinclude + systemLibIdentifier;
		public static final String LIB_INCLUDE_REGEX = Pragma.SYMBOL + startinclude + "(" + systemLibIdentifier + "|" + libIdentifier + ") *";
		public static final String END_INCLUDE_REGEX = Pragma.SYMBOL + endinclude + " *";

		private Cetus() {
		}
	}


	/**
	 * JConfig peripheral types
	 */
	public static final class JConfigTypes {

		public static final String DISPATCHER_MASTER = "dispatcher_master";
		public static final String DISPATCHER_SLAVE = "dispatcher_slave";
		public static final String DISPATCHER_MASTER_DMA = "dispatcher_master_dma";
		public static final String DISPATCHER_SLAVE_DMA = "dispatcher_slave_dma";
		public static final String CORE_CONNECTOR_SLAVE = "core_connector_slave";
		public static final String CORE_CONNECTOR_MASTER = "core_connector_master";
		public static final String CORE_CONNECTOR_SLAVE_DMA = "core_connector_slave_dma";
		public static final String CORE_CONNECTOR_MASTER_DMA = "core_connector_master_dma";
		public static final String CORE_CONNECTOR_DUPLEX = "core_connector_duplex";
		public static final String CONCENTRATOR_MASTER = "concentrator_master";
		public static final String CONCENTRATOR_SLAVE = "concentrator_slave";
		public static final String CONCENTRATOR_MASTER_DMA = "concentrator_master_dma";
		public static final String CONCENTRATOR_SLAVE_DMA = "concentrator_slave_dma";
		public static final String MEM_SWAP_SLAVE = "mem_swap_slave";
		public static final String MEM_SWAP_DUAL_MASTER = "mem_swap_dual_master";
		public static final String MEM_SWAP_MULTI_MASTER = "mem_swap_multi_master";
		public static final String UART = "uart";
		public static final String UART_LIGHT = "uart_light";
		public static final String SIM_PRINT = "sim_print";
		public static final String ADDITIONAL_DMA_APPENDIX = "_DMA";

		public static final String CORE_CONNECTOR_DMA_PERI_TYPE = "core_connector_dma_base_t";
		public static final String CONCENTRATOR_DMA_PERI_TYPE = "inter_core_communication_rx_dma_t";
		public static final String DISPATCHER_DMA_PERI_TYPE = "inter_core_communication_tx_dma_t";

		public static final List<String> PRINTF_PERIPHERALS = Arrays.asList(UART, UART_LIGHT, SIM_PRINT);

		public static final List<String> DMA_TYPES = Arrays.asList(CONCENTRATOR_MASTER_DMA, CONCENTRATOR_SLAVE_DMA, DISPATCHER_MASTER_DMA, DISPATCHER_SLAVE_DMA, CORE_CONNECTOR_MASTER_DMA, CORE_CONNECTOR_SLAVE_DMA);

		private JConfigTypes() {
		}
	}


	/**
	 * All freemarker templates
	 */
	public static final class Freemarker {

		public static final String GLOBAL_VARIABLES_HEADER = "single_task/globalvariabledefinitions.h.ftl";
		/* Software Builder template files */
		public static final Collection<TemplateFile> SOFTWARE_TEMPLATES = new HashSet<>(
				Arrays.asList(
						new TemplateFile("single_task/main.c.ftl"),
						new TemplateFile("evaluation/evaluation.c.ftl"),
						new TemplateFile("single_task/main.h.ftl"),
						new TemplateFile("single_task/base_types.h.ftl"),
						//TODO: only gen if globel mem there
						new TemplateFile(GLOBAL_VARIABLES_HEADER),
						new TemplateFile("evaluation/evaluation.h.ftl")
				)
		);

		/* Hardware builder template files */
		public static final Collection<TemplateFile> HARDWARE_TEMPLATES = new HashSet<>(
				Arrays.asList(
						new TemplateFile("hardware.xml.ftl"),
						new TemplateFile("generateHardware.sh.ftl")
				)
		);

		private Freemarker() {

		}
	}

	//TODO: move to JConfigTypes


	/**
	 * All peripherals for evaluation
	 */
	public static final class EvaluationPeripherals {
		private static final String PERIPHERAL_PACKAGE = "de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.";
		public static final Collection<EvaluationPeripheralCreator.EvaluationComponent> EVALUATION_COMPONENTS = new HashSet<>(
				Arrays.asList(
						new EvaluationPeripheralCreator.EvaluationComponent(PERIPHERAL_PACKAGE + "Uart", EvaluationPeripheralCreator.EvaluationComponentType.BASE_SUB)
				)
		);

		public static final Collection<EvaluationPeripheralCreator.EvaluationComponent> EVALUATION_COMPONENTS_SIM = new HashSet<>(
				Arrays.asList(
						new EvaluationPeripheralCreator.EvaluationComponent(PERIPHERAL_PACKAGE + "SimPrint", EvaluationPeripheralCreator.EvaluationComponentType.BASE_SUB)
				)
		);

		private EvaluationPeripherals() {

		}
	}

	private Constants() {
	}
}
