/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.freemarker;

import de.tu_darmstadt.rs.microStreams.model.Evaluation;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConnectionPeripheralMasterDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConnectionPeripheralSlaveDMA;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;

import cetus.hir.ClassDeclaration;
import cetus.hir.Enumeration;
import cetus.hir.Program;
import cetus.hir.Traversable;
import com.google.common.collect.Sets;
import freemarker.ext.beans.BeansWrapper;
import freemarker.template.ObjectWrapper;
import freemarker.template.TemplateHashModel;
import freemarker.template.TemplateModel;
import freemarker.template.TemplateModelException;
import freemarker.template.WrappingTemplateModel;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.List;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * A Wrapper for the Software to access java stuff from freemarker
 */
public class SoftwareTemplateAdapter extends WrappingTemplateModel implements TemplateHashModel {

	private Core core;

	/**
	 * the constructor
	 * @param core core to create software for
	 * @param objectWrapper a wrapper
	 */
	public SoftwareTemplateAdapter(Core core, ObjectWrapper objectWrapper) {
		super(objectWrapper);
		this.core = core;

	}

	@Override
	public TemplateModel get(String key) throws TemplateModelException {
		switch (key) {
			case "include_files":
				return wrap(core.getIncludeFilesForMain());

			case "peripherals":
				return wrap(core.getPeripherals());

			case "masterDMAConnection":
				return wrap(core.getPeripherals(ConnectionPeripheralMasterDMA.class));

			case "slaveDMAConnection":
				return wrap(core.getPeripherals(ConnectionPeripheralSlaveDMA.class));

			case "core_id":
				return wrap(core.getId());

			case "hasGlobalMemory":
				return wrap(core.hasGlobalMemory());

			case "task":
				return wrap(core.getTasks().iterator().next());

			case "tasks":
				return wrap(core.getTasks());

			case "evaluation":
				return wrap(new Evaluation(core.getPeripherals()));

			case "structs":
				final Program program = core.getModel().getModelComponent(CetusProgram.class).getProgram();
				final List<Traversable> traversables = CetusUtils.getStreamOfAllChildren(program)
						.filter(CetusUtils.childInsideInclude().negate())
						.filter(CetusUtils.ignoreChildrenTypes(Sets.newHashSet(ClassDeclaration.class, Enumeration.class)).negate())
						.collect(Collectors.toList());
				return wrap(traversables);

			case "isEvaluated":
				return wrap(core.isEvaluated());
			case "configAllowGlobalMemory":
				return wrap(core.getModel().getConfig().useGlobalMemory());
			case "core_count":
				return wrap(core.getModel().getModelComponents(Core.class).size());
			case "containsISRs":
				return wrap(core.containsISRs());
			case "printInterruptServiceRoutines":
				return wrap(core.printInterruptServiceRoutines());
			case "JConfigTypes":
				return getStaticModels().get("de.tu_darmstadt.rs.microStreams.config.Constants$JConfigTypes");
			case "BodyType":
				return getStaticModels().get("de.tu_darmstadt.rs.microStreams.model.software.freemarker.Body$BodyType");
			case "PeripheralUsage":
				return getStaticModels().get("de.tu_darmstadt.rs.microStreams.model.software.drivers.PeripheralUsage");
			case "Measurement":
				return getStaticModels().get("de.tu_darmstadt.rs.microStreams.model.software.freemarker.Measurement");
			case "statics":
				return getStaticModels();
			case "hasReplicatedTasks":
				return wrap(core.getModel().getModelComponents(Task.class).stream().anyMatch(t -> t.isReplica()));
			default:
				throw new RuntimeException("The key: " + key + " is nowhere to be found in the template");

		}
	}

	@Override
	public boolean isEmpty() throws TemplateModelException {
		return false;
	}

	private TemplateHashModel getStaticModels() {
		return ((BeansWrapper) this.getObjectWrapper()).getStaticModels();
	}

}
