/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model;

import de.tu_darmstadt.rs.microStreams.config.IConfiguration;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareComponentFactory;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelComponent;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskFactory;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * <p>
 * This is the root class of the µStreams Source-To-Source Compiler data model.
 * In the µStreams Source-To-Source Compiler Model and the Operation are
 * disjoint. While the {@link Model} is representing the entire data model the
 * µStreams Source-To-Source Compiler is working on, implementations of the
 * {@link TransformationPass} interface are Operations working on this data model.
 * </p>
 * <p>
 * <p>
 * {@link ModelComponent ModelComponents} can be added to the {@link Model}
 * instance with the {@link Model#addModelComponent(ModelComponent...)} method
 * respectively removed with the
 * {@link Model#removeModelComponent(ModelComponent)} method.
 * </p>
 * <p>
 * <p>
 * You can invoke the {@link Model#getModelComponents(Class)} method in order to
 * return already added ModelComponents. The argument can be used to filter for
 * a specific component type.
 * </p>
 * <p>
 * <p>
 * The {@link Model#isWorkerApplied(Class)} can be used to check whether the
 * given worker was already applied to this Model. This information could be
 * useful, if you are implementing the {@link TransformationPass#processable(Model)
 * processable(Model)} method of the {@link TransformationPass} interface in your own Worker
 * implementations.
 * </p>
 *
 * @author Jan Weber
 */
public class Model {

	/**
	 * slf4j Logger
	 */
	Logger log = LoggerFactory.getLogger(this.getClass());

	/**
	 * This collection of ModelComponents contains all model components in this
	 * Model.
	 */
	private Collection<ModelComponent> components = new ArrayList<>();

	/**
	 * HardwareComponentFactory associated with this Model.
	 */
	private HardwareComponentFactory hwFactory = new HardwareComponentFactory(
			this);

	/**
	 * This collection of Workers contains all {@link TransformationPass Workers} applied to
	 * this {@link Model}.
	 */
	private Collection<Class<? extends TransformationPass>> appliedWorker = new HashSet<>();

	private Collection<TransformationPass> workers = new HashSet<>();

	/**
	 * Constructor. Expecting a Collection of {@link SourceFile SourceFiles} of
	 * the original C Firmware this Model belongs to.
	 */
	public Model() {

	}

	/**
	 * Adds a Worker to the list of already applied {@link TransformationPass Workers} to
	 * this {@link Model}. This is done to let other workers check whether all
	 * pre-requirements for their execution are met.
	 *
	 * @param worker Class of the worker that was applied.
	 */
	public void addAppliedWorker(Class<? extends TransformationPass> worker) {
		appliedWorker.add(worker);
	}

	public boolean addWorkers(TransformationPass... workers) {
		return this.workers.addAll(Arrays.asList(workers));
	}

	public Collection<TransformationPass> getWorkers(Predicate<TransformationPass> pruneOn) {
		return workers.stream().filter(pruneOn).collect(Collectors.toSet());
	}

	public Collection<TransformationPass> getAllWorkers() {
		return workers;
	}

	public Collection<TransformationPass> getNotAppliedWorkers() {
		return getWorkers(w -> !isWorkerApplied(w.getClass()));
	}

	public Collection<TransformationPass> getAppliedWorkers() {
		return getWorkers(w -> isWorkerApplied(w.getClass()));
	}

	public boolean allWorkersApplied(List<Class<? extends TransformationPass>> requestedWorkers) {
		return requestedWorkers.stream().allMatch(w -> isWorkerApplied(w));
	}

	public boolean removeWorker(TransformationPass... workers) {
		return this.workers.removeAll(Arrays.asList(workers));
	}

	/**
	 * Adds a new {@link ModelComponent} to this {@link Model}.
	 *
	 * @param c ModelComponent which is added to this Model.
	 * @return true if this Model changed as a result of the call.
	 */
	public boolean addModelComponent(ModelComponent... c) {
		boolean allAdd = true;
		for (ModelComponent tmp : c) {
			tmp.setModel(this);
			if (components.contains(tmp)) {
				log.debug("Not adding component " + tmp + " since it is already in the model.");
				allAdd = false;
			} else {
				final boolean successfulAdd = components.add(tmp);
				if (!successfulAdd) {
					allAdd = false;
				}
			}
		}
		return allAdd;
	}

	/**
	 * Returns the number of {@link TransformationPass Workers} already applied to this
	 * {@link Model}.
	 *
	 * @return the number of Workers applied to this Model.
	 */
	public int countAppliedWorkers() {
		return appliedWorker.size();
	}

	/**
	 * Returns the {@link HardwareComponentFactory} for this {@link Model}. Use
	 * this HardwareComponentFactory to generate {@link HardwareModelComponent
	 * HardwareModelComponents} in order to get a unique id for each
	 * HardwareModelComponent.
	 *
	 * @return Returns the HardwareComponentFactory for this {@link Model}.
	 */
	public HardwareComponentFactory getHardwareComponentFactory() {
		return hwFactory;
	}

	/**
	 * Returns a Collection of all {@link ModelComponent ModelComponents} stored
	 * within this {@link Model} from the given type.
	 *
	 * @param t type of ModelComponents that shall be returned.
	 * @param <T> t
	 * @return Collection of {@link ModelComponent ModelComponents}.
	 */
	@SuppressWarnings("unchecked")
	public <T extends ModelComponent> Collection<T> getModelComponents(Class<T> t) {

		if (t == null) {
			return null;
		}
		return components.stream()
				.filter(c -> t.isAssignableFrom(c.getClass()))
				.map(c -> (T) c)
				.collect(Collectors.toCollection(ArrayList::new));
	}


	/**
	 * Returns one model component if only one exists
	 *
	 * @param t type of ModelComponents that shall be returned.
	 * @param <T> t
	 * @return {@link ModelComponent ModelComponent} or an exception if more than one model component of this type exists
	 */
	@SuppressWarnings("unchecked")
	public <T extends ModelComponent> T getModelComponent(Class<T> t) {
		return getModelComponents(t).stream().collect(StreamUtil.exactlyOne());
	}

	/**
	 * Returns one model component if only one exists
	 *
	 * @param t the model component to be found
	 * @param <T> t
	 * @return {@link ModelComponent ModelComponent} or an exception if more than one model component of this type exists
	 */
	public <T extends ModelComponent> Optional<T> getModelComponent(T t) {
		return components.stream()
				.filter(c -> t.equals(c))
				.map(c -> (T) c)
				.collect(StreamUtil.atMostOne());
	}

	public boolean hasModelComponent(ModelComponent modelComponent) {
		return getModelComponent(modelComponent).isPresent();
	}

	public <T extends ModelComponent> boolean hasModelComponent(Class<T> t) {
		return !getModelComponents(t).isEmpty();
	}

	public boolean allComponentsOnModel(List<Class<? extends ModelComponent>> requestedComponents) {
		return requestedComponents.stream().allMatch(mc -> !getModelComponents(mc).isEmpty());
	}

	/**
	 * The {@link Model#isWorkerApplied(Class)} can be used to check whether the
	 * given worker was already applied to this Model. This information could be
	 * useful, if you are implementing the {@link TransformationPass#processable(Model)
	 * processable(Model)} method of the {@link TransformationPass} interface in your own
	 * Worker implementations.
	 *
	 * @param worker the {@link Class} of the worker that shall be checked.
	 * @return true if the {@link TransformationPass} was applied, false otherwise.
	 */
	public boolean isWorkerApplied(Class<? extends TransformationPass> worker) {

		for (Class<? extends TransformationPass> applWorkerName : appliedWorker) {
			if (applWorkerName.equals(worker)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Removes a {@link ModelComponent} from this Model. Returns true if this
	 * Model contained the specified element (or equivalently, if this
	 * collection changed as a result of the call).
	 *
	 * @param component ModelComponent to be removed from this Model, if present.
	 * @return true if an element was removed as a result of this call.
	 */
	public boolean removeModelComponent(ModelComponent component) {
		return components.remove(component);

	}

	public void printIgnoreableError(Logger log, String msg) {
		printIgnoreableError(log, msg, null);
	}

	/**
	 * Prints error messages and throws an exception if strict checking is enabled in the config
	 * @param log a logger to use
	 * @param msg the message to print
	 * @param e an optional exception
	 */
	public void printIgnoreableError(Logger log, String msg, Exception e) {
		if (getConfig().ignoreErrors()) {
			log.error(msg);
		} else {
			throw new RuntimeException(msg, e);
		}
	}

	public IConfiguration getConfig() {
		return getModelComponent(IConfiguration.class);
	}

	public TaskFactory getTaskFactory() {
		return getModelComponent(TaskFactory.class);
	}

}
