/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.hardware;

import de.tu_darmstadt.rs.microStreams.model.Model;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.Map;

/**
 * The {@link HardwareComponentFactory} generates {@link HardwareModelComponent
 * HardwareModelComponents}.
 * <p>
 * This factory class is necessary to manage the unique ids of the components.
 *
 * @author Jan Weber
 */
public class HardwareComponentFactory {

	/**
	 * Slf4j Logger
	 */
	Logger log = LoggerFactory.getLogger(this.getClass());

	/**
	 * Contains the unique id counter of all top level hardware components.
	 */
	private Map<Type, Integer> componentIds = new HashMap<Type, Integer>();

	/**
	 * Contains the unique id counter of all peripheral types in each core.
	 */
	private Map<Integer, Map<String, Integer>> subComponentIds = new HashMap<>();

	private Model m;

	public HardwareComponentFactory(Model m) {
		this.m = m;
	}

	/**
	 * Tries to instanciate a {@link HardwareModelComponent} of the given
	 * {@link Class} type.
	 * <p>
	 * The Component has to contain a public non-argument constructor
	 *
	 * @param type - the {@link Class} type of the {@link HardwareModelComponent} that shall be instanciated.
	 * @param <T> t
	 * @return Returns the new instanciated {@link HardwareModelComponent}.
	 */
	@SuppressWarnings("unchecked")
	public <T extends HardwareModelComponent> T createComponent(Class<T> type) {

		try {
			HardwareModelComponent hc = createComponent(type.getName());
			return (T) hc;
		} catch (ClassNotFoundException e) {
			log.error("Could not create Component because class not found", e);
		}
		return null;
	}

	/**
	 * Tries to instanciate a {@link HardwareModelComponent} of the given type
	 * as {@link String} .
	 * <p>
	 * The SubComponent has to contain a public non-argument constructor and
	 * implmenting the {@link HardwareModelSubComponent} interface.
	 *
	 * @param componentName - the {@link Class} type of the {@link HardwareModelComponent} that shall be instanciated
	 * @return Returns the new instanciated {@link HardwareModelComponent}
	 * @throws ClassNotFoundException if class with the given name was not found
	 */
	public HardwareModelComponent createComponent(String componentName)
		throws ClassNotFoundException {
		Class<?> clazz = Class.forName(componentName);

		HardwareModelComponent hc;
		try {
			hc = (HardwareModelComponent) clazz.newInstance();
		} catch (InstantiationException | IllegalAccessException e) {
			log.error(
				"Could not instanciate Component. Make sure there is an public non-argument constructor.",
				e);
			return null;

		}

		int componentId = 0;

		if (componentIds.containsKey(clazz)) {
			componentId = componentIds.get(clazz) + 1;
		}

		componentIds.put(clazz, componentId);

		hc.setId(componentId);

		m.addModelComponent(hc);

		return hc;
	}

	/**
	 * Tries to instanciate a {@link HardwareModelSubComponent} for the given
	 * {@link HardwareModelComponent} of the given {@link Class} type.
	 * <p>
	 * The SubComponent has to contain a public non-argument constructor and
	 * implmenting the {@link HardwareModelSubComponent} interface.
	 *
	 * @param type - the {@link Class} type of the {@link HardwareModelComponent} that shall be instanciated.
	 * @param parentComponent the parent component
	 * @param <T> t
	 * @return Returns the new instanciated {@link HardwareModelComponent}.
	 */
	@SuppressWarnings("unchecked")
	public <T extends HardwareModelSubComponent> T createSubComponent(
		Class<T> type, HardwareModelComponent parentComponent) {

		try {
			HardwareModelComponent hc = createSubComponent(type.getName(),
				parentComponent);
			return (T) hc;
		} catch (ClassNotFoundException e) {
			log.error("Could not create SubComponent because class not found",
				e);
		}
		return null;

	}

	/**
	 * Tries to instanciate a {@link HardwareModelSubComponent} for the given
	 * {@link HardwareModelComponent} of the given type as {@link String} .
	 * <p>
	 * The SubComponent has to contain a public non-argument constructor and
	 * implmenting the {@link HardwareModelSubComponent} interface.
	 *
	 * @param componentName - the {@link Class} type of the {@link HardwareModelComponent} that shall be instanciated.
	 * @param parentComponent the parent component
	 * @return Returns the new instanciated {@link HardwareModelComponent}.
	 * @throws ClassNotFoundException if the component name has not been found
	 */
	public HardwareModelSubComponent createSubComponent(String componentName,
														HardwareModelComponent parentComponent)
		throws ClassNotFoundException {

		Class<?> clazz = Class.forName(componentName);

		HardwareModelSubComponent hsc;

		try {
			hsc = (HardwareModelSubComponent) clazz.newInstance();
		} catch (InstantiationException | IllegalAccessException e) {
			log.error(
				"Could not instanciate SubComponent. Make sure there is an public non-argument constructor.",
				e);
			return null;
		}

		int peripheralId = 0;
		int coreId = parentComponent.getId();

		if (!subComponentIds.containsKey(coreId)) {
			subComponentIds.put(coreId, new HashMap<>());
		}

		Map<String, Integer> idCounterForCore = subComponentIds.get(coreId);

		if (!idCounterForCore.containsKey(hsc.getType())) {
			idCounterForCore.put(hsc.getType(), 0);
		} else {
			peripheralId = idCounterForCore.get(hsc.getType()) + 1;
			idCounterForCore.put(hsc.getType(), peripheralId);
		}

		hsc.setId(peripheralId);
		hsc.setParentHardwareComponent(parentComponent);

		m.addModelComponent(hsc);
		parentComponent.addSubComponent(hsc);

		return hsc;

	}

}
