/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.hardware.core;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelSubComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.Peripheral;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.DirectConnection;
import de.tu_darmstadt.rs.microStreams.model.interfaces.NeedsIncludes;
import de.tu_darmstadt.rs.microStreams.model.other.HeaderFile;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;

import cetus.hir.Initializer;
import cetus.hir.NameID;
import cetus.hir.Procedure;

import java.io.File;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Represents a processing core holding tasks and peripherals
 */
public class Core implements NeedsIncludes, HardwareModelComponent {

	/**
	 * All tasks running on this core.
	 */
	private Collection<Task> tasks = new LinkedHashSet<Task>();

	/**
	 * All peripherals connected to this core.
	 */
	private Collection<Peripheral> peripherals = new LinkedHashSet<Peripheral>();

	/**
	 * The unique Id of this core.
	 */
	private int id;

	/**
	 * The folder the firmware of this core takes place.
	 */
	private Path firmwareFolder;

	/**
	 * The project this Core belongs to.
	 */
	private Model project;

	private Set<Procedure> interruptServiceRoutines = new LinkedHashSet<>();
	private Map<Variable, Initializer> globalInitialisation = new HashMap<Variable, Initializer>();

	public Core() {
	}

	/**
	 * Adds the given {@link Peripheral} to the {@link Core}.
	 *
	 * @param component a component to be added
	 * @return true or false, depending whether the operation was successful or not
	 */
	@Override
	public boolean addSubComponent(HardwareModelSubComponent component) {
		if (component instanceof Peripheral) {
			return peripherals.add((Peripheral) component);
		} else {
			throw new UnsupportedOperationException("Subcomponent has to be Peripheral.");
		}

	}

	/**
	 * Returns the unique ID of this task.
	 *
	 * @return Returns a Integer value that represents the unique ID of this task.
	 */
	public int getId() {
		return id;
	}

	@Override
	public void setId(int id) {
		this.id = id;

	}

	/**
	 * Returns the name of the Core.
	 *
	 * @return Returns the name of the Core.
	 */
	@Override
	public String getName() {
		return getType() + "_" + getId();

	}

	@Override
	public String getType() {
		return "spartanmc";
	}

	@Override
	public boolean isEvaluationComponent() {
		return false;
	}

	@Override
	public void setEvaluationComponent(boolean isEvaluationComponent) {
		throw new UnsupportedOperationException("Not part of Evaluation");

	}

	/**
	 * Adds a new {@link Task} task to this Core.
	 *
	 * @param task the task to be added
	 * @return Returns true or false, depending on adding was successfull.
	 */
	public boolean addTask(Task task) {
		return tasks.add(task);
	}

	/**
	 * Tells if the core has evaluation hardware
	 * @return true if this Core contains Hardware for evaluation purpose
	 */
	public boolean containsEvaluationHw() {
		for (HardwareModelComponent peripheral : getPeripherals()) {
			if (peripheral.isEvaluationComponent()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Tells if the core is being evaluated
	 * @return true if this Core should be evaluated
	 */
	public boolean isEvaluated() {
		return project.getConfig().isCreatePerformanceEvaluation();
	}

	/**
	 * return the firmware path
	 * @return Returns the folder were the firmware of the core is placed represented as String.
	 */
	public Path getFirmware() {
		Path p = Paths.get(project.getConfig().getTargetRootDir()).toAbsolutePath().normalize().relativize(firmwareFolder.toAbsolutePath().normalize());
		return p;
	}

	/**
	 * Sets the folder where the firmware for this Core is placed to the given
	 * String firmwareFolder.
	 *
	 * @param firmwareFolder folder for the FW
	 */
	public void setFirmware(Path firmwareFolder) {
		this.firmwareFolder = firmwareFolder;
	}

	/**
	 * Returns a Collection of Strings that represents all include files.
	 *
	 * @return A collection of strings that represent the include files.
	 */
	@Override
	public Collection<Include> getIncludeFiles() {
		Collection<Include> includeFiles = new HashSet<Include>();

		for (Task task : tasks) {
			includeFiles.addAll(task.getIncludeFiles());
		}

		for (Peripheral peripheral : getPeripherals()) {
			includeFiles.addAll(peripheral.getIncludeFiles());
		}

		if (project.getConfig().isCreatePerformanceEvaluation()) {
			includeFiles.add(Constants.FilesAndFolders.EVALUATION_HEADER);
		}

		return includeFiles;
	}

	/**
	 * Returns all include files that are needed for the core that are not system includes.
	 *
	 * @return A Collection of all local include files.
	 */
	public Collection<File> getLocalIncludeFiles() {
		Collection<File> includeFiles = new HashSet<>();

		for (Task task : tasks) {
			for (HeaderFile h : task.getHeaderFiles()) {
				includeFiles.add(h.getFile());
			}
			for (Include inc : task.getIncludeFiles()) {
				if (inc.getHeaderFile() != null) {
					includeFiles.add(inc.getHeaderFile().getFile());
				}
			}
		}

		return includeFiles;
	}

	/**
	 * Returns all source files that are needed for the core.
	 *
	 * @return A Collection of all local source files.
	 */
	public Collection<File> getLocalSourceFiles() {
		Collection<File> sourceFiles = new HashSet<>();

		for (Task task : tasks) {
			for (SourceFile sf : task.getSourceFiles()) {
				sourceFiles.add(sf.getFile());
			}
		}

		return sourceFiles;
	}

	/**
	 * Returns a Collection of Strings that represents all include files that
	 * should be included in this core's main src file in order to work
	 * properly.
	 *
	 * @return A collection of strings that represent the include files without duplicates determined by the print method.
	 */
	public Collection<Include> getIncludeFilesForMain() {
		Collection<Include> includeFiles = new LinkedHashSet<Include>();

		for (Task task : tasks) {
			includeFiles.addAll(task.getIncludeFiles());
		}

		peripherals.stream().forEach(p -> {
			Include include = p.getStdioInclude();
			if (include != null) {
				includeFiles.add(include);
			}
		});
		includeFiles.addAll(Constants.FilesAndFolders.DEFAULT_HEADERS);

		if (project.getConfig().isCreatePerformanceEvaluation()) {
			includeFiles.add(Constants.FilesAndFolders.EVALUATION_HEADER);
		}

		//filter duplicates by their print method
		Collection<Include> includeFilesNoDupl = includeFiles.stream()
				.collect(Collectors.toMap(Include::print, p -> p, (p, q) -> p))
				.values().stream()
				.collect(Collectors.toList());
		return includeFilesNoDupl;
	}

	/**
	 * Returns a Collection of all {@link Peripheral}s of this {@link Core}.
	 *
	 * @return Returns a Collection of all {@link Peripheral}s of this {@link Core}.
	 */
	public List<Peripheral> getPeripherals() {
		List<Peripheral> tmp = new ArrayList<>(peripherals);
		tmp.sort(Comparator.comparing(p -> p.getName()));
		return tmp;
	}

	/**
	 * Returns a Collection of all {@link Peripheral}s of this {@link Core}. The
	 * returned {@link Peripheral}s are from the given Classtype type.
	 *
	 * @param type typ
	 * @param <T> t
	 * @return Returns a Collection of all {@link Peripheral}s of this {@link Core} from the given Classtype type.
	 */
	@SuppressWarnings("unchecked")
	public <T extends Peripheral> Collection<T> getPeripherals(Class<T> type) {
		List<T> peripherals = new ArrayList<>();

		for (Peripheral peripheral : getPeripherals()) {
			if (type.isAssignableFrom(peripheral.getClass())) {
				peripherals.add((T) peripheral);
			}
		}

		return peripherals;

	}

	/**
	 * Finds Peripherals used by both cores
	 *
	 * @param otherCore another core to compare with
	 * @return A list of common Peripherals
	 */
	public List<Peripheral> getCommonPeripherals(Core otherCore) {
		Collection<Peripheral> otherPeripherals = otherCore.getPeripherals();
		return getPeripherals().stream().filter(p -> otherPeripherals.contains(p)).collect(Collectors.toList());
	}

	/**
	 * Returns if this core has tasks which are child tasks of the other cores tasks.
	 *
	 * @param otherCore another core whose child tasks shall be searched for a dependency to this cores tasks
	 * @return true if a dependency was detected, flase otherwise
	 */
	public boolean hasChildTasksOf(Core otherCore) {
		return otherCore.getTasks().stream().filter(task -> task.getContainedTasks().stream().filter(containedTask -> getTasks().contains(containedTask)).count() > 0).count() > 0;
	}

	@Override
	public Model getModel() {
		return this.project;
	}

	@Override
	public void setModel(Model project) {
		this.project = project;

	}

	/**
	 * returns all contained tasks
	 * @return Returns a Collection of {@link Task Tasks} that are running on this {@link Core}.
	 */
	public Collection<Task> getTasks() {
		return tasks;

	}

	@Override
	public String toString() {
		return "Core#" + id + " [tasks=" + tasks + "] [peripherals= " + peripherals + "]";
	}

	/**
	 * Removes the given {@link Peripheral} from the {@link Core}. Returns true
	 * or false, depending whether the operation was successfull or not.
	 *
	 * @param peripheral peripheral to be removed
	 * @return true if ther peripheral was successfully removed
	 */
	public boolean removePeripheral(Peripheral peripheral) {
		return peripherals.remove(peripheral);
	}

	/**
	 * Removes the given {@link Task} task from the {@link Core}
	 *
	 * @param task the task to be removed
	 * @return true or false depending this was successfull or not (Task was no part of Core).
	 */
	public boolean removeTask(Task task) {
		return tasks.remove(task);
	}

	public boolean containsISRs() {
		return !interruptServiceRoutines.isEmpty();
	}

	public boolean addInterruptServiceRoutine(Procedure procedure) {
		return interruptServiceRoutines.add(procedure);
	}

	public boolean addInterruptServiceRoutines(Set<Procedure> procedures) {
		return interruptServiceRoutines.addAll(procedures);
	}

	public boolean removeInterruptServiceRoutine(Procedure procedure) {
		return interruptServiceRoutines.remove(procedure);
	}

	/**
	 * Adds an global init section to the core
	 * @param var the variable to be initialized
	 * @param init the initializer of the variable
	 * @return true if addign was successful and the variable is not already to be initialized
	 */
	public boolean addGlobalInitialisation(Variable var, Initializer init) {
		if (globalInitialisation.containsKey(var)) {
			return false;
		}
		globalInitialisation.put(var, init);
		return true;
	}

	public boolean removeGlobalInitialisation(NameID identifier, Initializer init) {
		return globalInitialisation.remove(identifier, init);
	}

	public Map<Variable, Initializer> getGlobalInitialisation() {
		return globalInitialisation;
	}

	public boolean hasGlobalMemory() {
		return getModel().getModelComponents(DirectConnection.class).stream().anyMatch(con -> con.isConnected(this));
	}

	/**
	 * Prints ISRs to String
	 *
	 * @return ISRs for the template filling
	 */
	public String printInterruptServiceRoutines() {
		StringBuilder stringBuilder = new StringBuilder();
		interruptServiceRoutines.stream().forEach(isr -> {
			stringBuilder.append(isr.getDeclaration().toString());
			stringBuilder.append("\n\n");
		});
		return stringBuilder.toString();
	}
}
