/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection;

import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.PeripheralParameter;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;

import cetus.hir.AccessExpression;
import cetus.hir.AccessOperator;
import cetus.hir.ClassDeclaration;
import cetus.hir.Expression;
import cetus.hir.Identifier;
import cetus.hir.Initializer;
import cetus.hir.NameID;
import cetus.hir.Traversable;
import cetus.hir.UserSpecifier;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * A DMA connection peripherals Master component
 * Created by Kris Heid on 12.12.17.
 */
public abstract class ConnectionPeripheralMasterDMA extends ConnectionPeripheralDMA implements MasterPeripheral {

	private Map<Variable, Integer> variablesDMAMemoryMap = new HashMap<>();

	public abstract String getPeriTypeName();

	@Override
	public boolean addVariable(Variable variable) {
		if (variablesDMAMemoryMap.containsKey(variable)) {
			return false;
		}

		int lowestFreeOffset = getFirstFreeMemoryOffset();
		variablesDMAMemoryMap.put(variable, lowestFreeOffset);
		return true;
	}

	@Override
	public Set<Variable> getVariables() {
		return variablesDMAMemoryMap.keySet();
	}

	@Override
	public Integer getDMAOffsetOf(Variable variable) {
		return variablesDMAMemoryMap.get(variable);
	}

	/**
	 * Gives the first free memory offset in the DMA mapping
	 *
	 * @return the first free memory offset in the DMA mapping
	 */
	public int getFirstFreeMemoryOffset() {
		final Optional<Map.Entry<Variable, Integer>> variableWithHighestOffset = variablesDMAMemoryMap.entrySet().stream()
				.max(Comparator.comparing(Map.Entry::getValue));
		Integer lowestFreeOffset;
		if (variableWithHighestOffset.isPresent()) {
			lowestFreeOffset = variableWithHighestOffset.get().getValue() + variableWithHighestOffset.get().getKey().getSizeInWords();
		} else {
			lowestFreeOffset = 0;
		}
		return lowestFreeOffset;
	}

	/**
	 * Provides the declaration statement for the struct that holds all variables sent over the peripheral.
	 *
	 * @return A declaration statement.
	 */
	public ClassDeclaration getStructDeclaration() {
		ClassDeclaration decl = new ClassDeclaration(ClassDeclaration.STRUCT, new NameID(getStructName()));

		if (getDMAPeripheralControlRegionSize() > 0) {
			decl.addDeclaration(new VariableDeclaration(new VariableDeclarator(new NameID(getPeriTypeName() + " peri"))));
		}

		for (Variable v: variablesDMAMemoryMap.keySet().stream().sorted().collect(Collectors.toList())) {
			final VariableDeclaration declarationWithoutInitializer = CetusUtils.getDeclarationWithoutInitializer((VariableDeclaration) v.getVariableDeclarator().getDeclaration(), v.getVariableDeclarator());
			decl.addDeclaration(declarationWithoutInitializer);
		}
		return decl;
	}

	/**
	 * Returns the definition of the struct, including initializers for variables where possible when enabled.
	 *
	 * @param withInit Whether the definition should include initializers.
	 * @return The definition statement.
	 */
	public VariableDeclaration getStructDefinition(boolean withInit) {
		VariableDeclarator varDecl = new VariableDeclarator(new NameID(getStructInstanceName()));

		if (withInit) {
			List<Initializer> initializers = new ArrayList<>();
			for (Variable v : variablesDMAMemoryMap.keySet().stream().sorted().collect(Collectors.toList())) {
				// Check if this variable is received on the same core, then we don't need an initializer.
				List<Variable> varsReceived = getCore().getPeripherals(ConnectionPeripheralSlaveDMA.class).stream()
						.flatMap(c -> c.getVariables().stream())
						.collect(Collectors.toList());
				if (varsReceived.contains(v)) {
					continue;
				}

				Initializer i = v.getInitializer();
				final Optional<Identifier> identifierInInitializer = CetusUtils.getStreamOfAllChildren(i, Identifier.class).findAny();
				if (i == null || identifierInInitializer.isPresent()) {
					continue;
				} else {
					i = i.clone();
				}

				//List<Expression> initClones = i.getChildren().stream().map(c -> ((Expression)c).clone()).collect(Collectors.toList());
				List<Traversable> initClones = new ArrayList<>();
				for (Traversable t : i.getChildren()) {
					if (t instanceof Initializer) {
						initClones.add(((Initializer) t).clone());
					} else if (t instanceof Expression) {
						initClones.add(((Expression) t).clone());
					} else {
						throw new UnsupportedOperationException("Unknown initializer.");
					}
				}


				Initializer initWithDesignator;
				if (i.isList() || initClones.get(0) instanceof Initializer) {
					initWithDesignator = new Initializer(new AccessExpression(new NameID(""), AccessOperator.MEMBER_ACCESS, new NameID(v.getName())), initClones);
				} else {
					initWithDesignator = new Initializer(new AccessExpression(new NameID(""), AccessOperator.MEMBER_ACCESS, new NameID(v.getName())), (Expression) initClones.get(0));
				}

				initializers.add(initWithDesignator);
			}

			if (initializers.size() > 0) {
				Initializer finalInit = new Initializer(initializers);
				varDecl.setInitializer(finalInit);
			}
		}

		VariableDeclaration varDecn = new VariableDeclaration(new UserSpecifier(new NameID("struct " + getStructName())), varDecl);
		varDecn.setSemiColon(true);

		return varDecn;
	}

	@Override
	public List<PeripheralParameter> getParameters() {
		final int usedBytes = getFirstFreeMemoryOffset() + getDMAPeripheralControlRegionSize();
		return Arrays.asList(new PeripheralParameter("DMA_SIZE", Integer.toString(roundUpToBrams(usedBytes))));
	}

}
