/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection;

import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.Peripheral;

import cetus.hir.Identifier;

import java.util.Set;

/**
 * <p>
 * The {@link IConnectionPeripheral} interface represents the interconnect
 * peripheral of a {@link Core} through which a {@link IndirectConnection} for task
 * communication is settled.
 * <p>
 * <p>
 * <p>
 * All {@link IConnectionPeripheral ConnectionPeripherals} have to be added
 * through the {@link IndirectConnection#addConnectionPeripheral(IConnectionPeripheral)}
 * method to the corresponding connection. This method stores the peripheral
 * internally in a field and uses the {@link IConnectionPeripheral#validate()}
 * method to check whether all {@link IConnectionPeripheral
 * ConnectionPeripherals} are compatible to each other.
 * <p>
 * <p>
 * <p>
 * If you are implementing the {@link IConnectionPeripheral} interface make sure
 * to store the {@link Connection} handed in through the
 * {@link IConnectionPeripheral#addConnection(IndirectConnection)} inside a field and use
 * it in the {@link IConnectionPeripheral#validate() validate()} method
 * implementation to verify if your peripheral is compatible to all other
 * peripherals stored within the {@link Connection}. If you detect some
 * incompatibility you have to throw a
 * {@link ConnectionPeripheralsIncompatibleException}.
 * </p>
 *
 * @author Jan Weber
 */
public interface IConnectionPeripheral extends Peripheral {

	/**
	 * Returns the {@link Connection}, which connects this
	 * {@link IConnectionPeripheral} with one or multiple other
	 * {@link IConnectionPeripheral ConnectionPeripherals}.
	 *
	 * @return The {@link Connection} for this {@link IConnectionPeripheral}
	 */
	Set<IndirectConnection> getConnections();

	/**
	 * Sets the {@link IndirectConnection}, which connects this
	 * {@link IConnectionPeripheral} with one or multiple other
	 * {@link IConnectionPeripheral ConnectionPeripherals}.
	 * @param connection the connection
	 */
	void addConnection(IndirectConnection connection);

	/**
	 * Checks wether the this {@link IConnectionPeripheral} is correctly
	 * connected to the {@link Connection}.
	 *
	 * @throws ConnectionPeripheralsIncompatibleException A {@link ConnectionPeripheralsIncompatibleException} is thrown if any {@link IConnectionPeripheral} stored within the corresponding {@link Connection} that is incompatible to this {@link IConnectionPeripheral}
	 */
	void validate() throws ConnectionPeripheralsIncompatibleException;

	/**
	 * Gives the identifier for this peripheral to be used internally with Cetus. This should deliver the same as getName()
	 * @return an according identifier to this peripherals name
	 */
	Identifier getIdentifier();

}
