/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection;

import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelSubComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;

/**
 * Indirect connection between cores
 */
public class IndirectConnection extends Connection {
	/**
	 * {@link IConnectionPeripheral ConnectionPeripherals} that are connected
	 * with this connection;
	 */
	private Collection<IConnectionPeripheral> connectionPeripherals = new LinkedHashSet<IConnectionPeripheral>();

	public IndirectConnection() {
		super();
	}

	/**
	 * Adds an {@link IConnectionPeripheral} to this {@link Connection}. Returns
	 * true or false, depending whether the operation was successful or not
	 * (Peripheral already added).
	 *
	 * @param peripheral the {@link IConnectionPeripheral} that shall be added
	 * @return true if this {@link Connection} changed as a result of the call
	 * @throws ConnectionPeripheralsIncompatibleException a {@link ConnectionPeripheralsIncompatibleException} is thrown if any {@link IConnectionPeripheral} stored by this {@link Connection} is incompatible to the given given {@link IConnectionPeripheral}. The {@link IConnectionPeripheral} is not added in this case.
	 */
	public boolean addConnectionPeripheral(IConnectionPeripheral peripheral)
			throws ConnectionPeripheralsIncompatibleException {
		boolean operationSuccessfull = false;
		if (connectionPeripherals.add(peripheral)) {
			peripheral.addConnection(this);
			for (IConnectionPeripheral cp : connectionPeripherals) {
				try {
					cp.validate();
				} catch (ConnectionPeripheralsIncompatibleException e) {
					connectionPeripherals.remove(peripheral);
					cp.addConnection(null);
					throw e;
				}
			}
			operationSuccessfull = true;
		}
		return operationSuccessfull;

	}

	@Override
	public boolean addSubComponent(HardwareModelSubComponent component) {
		throw new UnsupportedOperationException("Has no subcomponents.");
	}

	@Override
	public String getType() {
		return "IndirectConnection";
	}

	/**
	 * give the connection peripheral
	 * @param core the core which shall be searched for connection peripherals
	 * @return all found connection peripherals in the given core
	 */
	public IConnectionPeripheral getConnectionPeripheral(Core core) {
		for (IConnectionPeripheral peripheral : connectionPeripherals) {
			if (peripheral.getCore().equals(core)) {
				return peripheral;
			}
		}
		return null;
	}

	public Collection<IConnectionPeripheral> getConnectionPeripherals() {
		return connectionPeripherals;
	}

	/**
	 * tells if a connection is possible
	 * @param sendingCore the sending core
	 * @param receivingCore the receiving core
	 * @return true if a connection is possible
	 */
	public boolean isConnecting(Core sendingCore, Core receivingCore) {

		for (IConnectionPeripheral sendPeripheral : connectionPeripherals) {
			if (sendPeripheral.getCore().equals(sendingCore)
					&& sendPeripheral instanceof SendPeripheral) {
				for (IConnectionPeripheral receivePeripheral : connectionPeripherals) {
					if (receivePeripheral.getCore().equals(receivingCore)
							&& receivePeripheral instanceof ReceivePeripheral) {
						return true;
					}
				}
			}
		}
		return false;
	}

	@Override
	public String toString() {

		ArrayList<Core> sending = new ArrayList<Core>();
		ArrayList<Core> receiving = new ArrayList<Core>();

		for (IConnectionPeripheral peri : connectionPeripherals) {
			if (peri instanceof SendPeripheral) {
				sending.add(peri.getCore());
			}

			if (peri instanceof ReceivePeripheral) {
				receiving.add(peri.getCore());
			}

		}

		return getName() + " [sendingCores = " + sending
				+ " receivingCores = " + receiving + " connectionPeripherals="
				+ connectionPeripherals + ", id=" + getId() + "]\n";
	}

	/**
	 * Removes an {@link IConnectionPeripheral} from the {@link Connection}.
	 * Returns true or false, depending whether the operation was successful or
	 * not.
	 *
	 * @param peripheral peripheral to be disconnected
	 * @return true if remove is successful
	 */
	public boolean removeConnectionPeripheral(IConnectionPeripheral peripheral) {
		return connectionPeripherals.remove(peripheral);

	}

}
