/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.controlflow;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.interfaces.HasCetusComponent;
import de.tu_darmstadt.rs.microStreams.model.software.SoftwareModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.DepthFirstIterator;
import cetus.hir.FunctionCall;
import cetus.hir.Traversable;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Optional;
import java.util.function.BiFunction;
import java.util.function.BiPredicate;
import java.util.function.BooleanSupplier;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * <p>
 * Represents the control flow graph of a {@link Task Task's} or
 * {@link UFunction Function's} {@link Traversable}.
 * <p>
 * <p>
 * <p>
 * A {@link InnerFlowGraph} consists of interconnected {@link InnerFlowNode
 * InnerFlowNodes} representing source code lines within the original C firmware
 * and is stored its task or function. Each {@link InnerFlowNode} holds a
 * collection of direct successor nodes, predecessor nodes, the
 * {@link Traversable} of the source code line and a list of
 * {@link VariableAccess VariableAccesses} that takes place within the nodes
 * source code line. There are several inheriting classes for special purposes:
 * <p>
 * <p>
 * <ul>
 * <li>{@link InnerFlowStart} - does not represent an actual source code line it
 * just marks the entry point of the whole program flow graph</li>
 * <li>{@link InnerFlowEnd} - does not represent an actual source code line it
 * just marks the exit point of the whole program flow graph</li>
 * <li>{@link InnerFlowTask} - represents a child task access within the program
 * flow</li>
 * </ul>
 * <p>
 * <p>
 * <p>
 * The {@link InnerFlowGraph} provides some helper methods to find
 * {@link Traversable traversables} and {@link Task Tasks} (child tasks) within
 * the graph respectively returning the nodes representing them.
 * </p>
 *
 * @author Jan Weber
 */
public class InnerFlowGraph implements SoftwareModelComponent,
		HasCetusComponent {

	private InnerFlowNode startNode;

	private InnerFlowNode endNode;

	private List<InnerFlowNode> allNodes;

	private Model model;

	private Traversable traversable;

	public InnerFlowGraph(InnerFlowNode startNode, InnerFlowNode endNode) {
		this.startNode = startNode;
		this.endNode = endNode;
	}

	/**
	 * Searches and returns all possible successors of the given
	 * {@link InnerFlowNode} node.
	 *
	 * @param node the node whose successors shall be returned.
	 * @return collection of {@link InnerFlowNode InnerFlowNodes} that are successors of the given node.
	 */
	public static Collection<InnerFlowNode> getAllSuccessors(InnerFlowNode node) {
		return getAllNodes(node,InnerFlowNode::getSucccessors, InnerFlowEnd.class);
		/*Collection<InnerFlowNode> foundNodes = new LinkedHashSet<>();
		ArrayList<InnerFlowNode> leftNodes = new ArrayList<>(Collections.singleton(node));

		boolean endReached = false;

		while (!leftNodes.isEmpty() && !endReached) {
			InnerFlowNode curNode = leftNodes.remove(0);

			Collection<InnerFlowNode> newNodes = new HashSet<>(curNode.getSucccessors());

			//break graph examining if end node reached
			for (InnerFlowNode newNode : newNodes) {
				boolean taskWraparound = InnerFlowTask.isTaskIDLess(newNode, node);
				if (newNode instanceof InnerFlowEnd || taskWraparound) {
					if (newNodes.size() == 1 && leftNodes.isEmpty() || taskWraparound) {
						endReached = true;
					}
					newNodes.remove(newNode);
					break;
				}
			}

			newNodes.removeAll(foundNodes);
			leftNodes.addAll(newNodes);
			foundNodes.addAll(newNodes);
		}
		return foundNodes;*/

	}

	/**
	 * Searches and returns all possible predecessors of the given
	 * {@link InnerFlowNode} node.
	 *
	 * @param node the node whose predecessors shall be returned.
	 * @return collection of {@link InnerFlowNode InnerFlowNodes} that are predecessors of the given node.
	 */
	public static Collection<InnerFlowNode> getAllPredecessors(
			InnerFlowNode node) {
		return getAllNodes(node, InnerFlowNode::getPredecessors, InnerFlowStart.class);
		/*Collection<InnerFlowNode> foundNodes = new LinkedHashSet<InnerFlowNode>();
		ArrayList<InnerFlowNode> leftNodes = new ArrayList<InnerFlowNode>(
				Collections.singleton(node));

		boolean startReached = false;

		while (!leftNodes.isEmpty() && !startReached) {
			InnerFlowNode curNode = leftNodes.remove(0);

			Collection<InnerFlowNode> newNodes = new HashSet<InnerFlowNode>(
					curNode.getPredecessors());

			//break graph examining if start node reached
			for (InnerFlowNode newNode : newNodes) {
				final boolean taskWraparound = InnerFlowTask.isTaskIDGreater(newNode, node);
				if (newNode instanceof InnerFlowStart || taskWraparound) {
					if (newNodes.size() == 1 && leftNodes.isEmpty() || taskWraparound) {
						startReached = true;
					}
					newNodes.remove(newNode);
					break;
				}
			}

			newNodes.removeAll(foundNodes);
			leftNodes.addAll(newNodes);
			foundNodes.addAll(newNodes);
		}
		return foundNodes;*/

	}




	public static Collection<InnerFlowNode> getAllNodes(InnerFlowNode node, Function<InnerFlowNode, Collection<InnerFlowNode>> accessor, Class<? extends InnerFlowNode> terminatingClassType) {
		Collection<InnerFlowNode> foundNodes = new LinkedHashSet<>();
		ArrayList<InnerFlowNode> leftNodes = new ArrayList<>(Collections.singleton(node));

		boolean endReached = false;
		boolean hasTwoEntryPoints;

		while (!leftNodes.isEmpty() && !endReached) {
			InnerFlowNode curNode = leftNodes.remove(0);

			Collection<InnerFlowNode> newNodes = new HashSet<>(accessor.apply(curNode));

			hasTwoEntryPoints = newNodes.size() > 1;
			//break graph examining if start node reached
			for (InnerFlowNode newNode : newNodes) {
				final boolean taskWraparound = hasTwoEntryPoints && newNode instanceof InnerFlowTask;
				if (newNode.getClass().isInstance(terminatingClassType) || taskWraparound) {
					if (newNodes.size() == 1 && leftNodes.isEmpty() || taskWraparound) {
						endReached = true;
					}
					newNodes.remove(newNode);
					break;
				}
			}

			newNodes.removeAll(foundNodes);
			leftNodes.addAll(newNodes);
			foundNodes.addAll(newNodes);
		}
		return foundNodes;

	}

	/**
	 * Returns the control flow graph's start node.
	 *
	 * @return InnerFlowNode
	 */
	public InnerFlowNode getInnerFlowStart() {
		return startNode;
	}

	/**
	 * Returns the control flow graph's end node.
	 *
	 * @return InnerFlowNode
	 */
	public InnerFlowNode getInnerFlowEnd() {
		return endNode;
	}

	@Override
	public Model getModel() {
		return model;
	}

	@Override
	public void setModel(Model model) {
		this.model = model;
	}

	@Override
	public Traversable getTraversable() {
		return traversable;
	}

	@Override
	public void setTraversable(Traversable traversable) {
		this.traversable = traversable;

	}

	/**
	 * Creates and returns a {@link List} of all {@link InnerFlowNode
	 * InnerFlowNodes} within this {@link InnerFlowGraph}.
	 *
	 * @return list of all {@link InnerFlowNode InnerFlowNodes}
	 */
	public List<InnerFlowNode> getAllNodes() {
		if (null != allNodes) {
			return allNodes;
		}

		allNodes = new ArrayList<>(Collections.singleton(startNode));

		ArrayList<InnerFlowNode> toVisit = new ArrayList<InnerFlowNode>(
				Collections.singleton(startNode));

		while (!toVisit.isEmpty()) {
			Collection<InnerFlowNode> newNodes = new HashSet<InnerFlowNode>(
					toVisit.remove(0).getSucccessors());
			/* do not process already discovered nodes */
			newNodes.removeAll(allNodes);

			/* newly discovered nodes are added to the toVisit-queue and result */
			Iterator<InnerFlowNode> newIter = newNodes.iterator();
			while (newIter.hasNext()) {
				InnerFlowNode newNode = newIter.next();
				if (!(newNode instanceof InnerFlowEnd)
						&& !(newNode instanceof InnerFlowStart)) {
					allNodes.add(newNode);
					toVisit.add(newNode);
				}

			}
		}

		return allNodes;
	}

	/**
	 * Creates and returns a {@link List} of all {@link InnerFlowNode
	 * InnerFlowNodes} within this {@link InnerFlowGraph}, starting with the given InnerFlowNode e.g. InnerFlowTask containing the given task.
	 *
	 * @param fromThisTaskOn the task to be found in InnerFlowTask that servers as Starting point
	 * @return list of all {@link InnerFlowNode InnerFlowNodes}
	 */
	public List<InnerFlowNode> getAllNodes(Task fromThisTaskOn) {
		List<InnerFlowNode> all = new ArrayList<>();
		boolean nodePassed = false;
		for (InnerFlowNode node : getAllNodes()) {
			if (node instanceof InnerFlowTask && ((InnerFlowTask) node).getTask() == fromThisTaskOn) {
				nodePassed = true;
			}
			if (nodePassed) {
				all.add(node);
			}
		}
		return all;
	}

	/**
	 * find inner flow node for traversable
	 *
	 * @param t traversable
	 * @return the {@link InnerFlowNode} that represents the given {@link Traversable} t
	 */
	public InnerFlowNode getInnerFlowNodeWith(Traversable t) {
		for (InnerFlowNode n : getAllNodes()) {
			if (!(n instanceof InnerFlowStart) && !(n instanceof InnerFlowEnd)) {
				if (n.getTraversable().equals(t)) {
					return n;
				}
			}
		}
		return null;
	}

	/**
	 * Returns the {@link InnerFlowNode} that represents the given {@link Task} t.
	 *
	 * @param t the {@link Task} whose {@link InnerFlowNode} shall be returned.
	 * @return the {@link InnerFlowNode} of the given {@link Task}, <code>null</code> if the {@link Task} is not contained in this {@link InnerFlowGraph}.
	 */
	public Optional<InnerFlowNode> getInnerFlowNodeWith(Task t) {
		for (InnerFlowNode n : getAllNodes()) {
			if (n instanceof InnerFlowTask
					&& ((InnerFlowTask) n).getTask().equals(t)) {
				return Optional.of(n);
			} else if (n.getClass().equals(InnerFlowNode.class)) {
				final List<FunctionCall> functionCalls = CetusUtils.getStreamOfAllChildren(n.getTraversable(), FunctionCall.class).collect(Collectors.toList());
				for (FunctionCall fc : functionCalls) {
					final Optional<UFunction> uFunction = model.getModelComponents(UFunction.class).stream()
							.filter(f -> f.isAccordingFunction(fc))
							.collect(StreamUtil.atMostOne());
					if (uFunction.isPresent()) {
						final Optional<InnerFlowNode> innerFlowNodeWith = uFunction.get().getInnerFlowGraph().getInnerFlowNodeWith(t);
						if (innerFlowNodeWith.isPresent()) {
							return innerFlowNodeWith;
						}
					}
				}
			}
		}
		return Optional.empty();
	}

	/**
	 * Returns the InnerFlowNode that is containing the given
	 * {@link Traversable} t.
	 *
	 * @param t the {@link Traversable} whose {@link InnerFlowNode} shall be returned.
	 * @return the {@link InnerFlowNode} of the given {@link Traversable}, <code>null</code> if the {@link Traversable} is not contained in this {@link InnerFlowGraph}.
	 */
	public InnerFlowNode getInnerFlowNodeContaining(Traversable t) {

		Collection<Traversable> allTraversables = new ArrayList<Traversable>();

		for (InnerFlowNode n : getAllNodes()) {
			if (!(n instanceof InnerFlowStart) && !(n instanceof InnerFlowEnd)) {
				allTraversables.add(n.getTraversable());
			}
		}

		while (t != null) {
			if (allTraversables.contains(t)) {
				return getInnerFlowNodeWith(t);
			}
			t = t.getParent();
		}
		return null;
	}

	/**
	 * Returns the InnerFlowNodes that is containing the given
	 * {@link Traversable} t.
	 *
	 * @param t the {@link Traversable} whose {@link InnerFlowNode} shall be returned.
	 * @return the {@link InnerFlowNode} of the given {@link Traversable}, <code>null</code> if the {@link Traversable} is not contained in this {@link InnerFlowGraph}.
	 */
	public List<InnerFlowNode> getInnerFlowNodesContaining(Traversable t) {
		List<InnerFlowNode> ret = new ArrayList<>();
		DepthFirstIterator<Traversable> iterator = new DepthFirstIterator<>(t);

		while (iterator.hasNext()) {
			Traversable traversable = iterator.next();
			final InnerFlowNode innerFlowNode = getInnerFlowNodeContaining(traversable);
			if (!ret.contains(innerFlowNode)) {
				ret.add(innerFlowNode);
			}
		}
		return ret;
	}
}
