/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.controlflow;

import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.BreakStatement;
import cetus.hir.Case;
import cetus.hir.CompoundStatement;
import cetus.hir.ContinueStatement;
import cetus.hir.DoLoop;
import cetus.hir.FlatIterator;
import cetus.hir.ForLoop;
import cetus.hir.GotoStatement;
import cetus.hir.IfStatement;
import cetus.hir.Label;
import cetus.hir.Loop;
import cetus.hir.Procedure;
import cetus.hir.ReturnStatement;
import cetus.hir.SwitchStatement;
import cetus.hir.Traversable;
import cetus.hir.WhileLoop;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

/**
 * This is a Helper Class to generate a inner Flow Graph
 * @author Jan Weber, Kris Heid
 */
public class InnerFlowGraphFractory {

	/* slf4j logger */
	protected static final Logger log = LoggerFactory.getLogger(InnerFlowGraphFractory.class);

	public InnerFlowGraphFractory() {
	}

	/**
	 * Connects two nodes.
	 *
	 * @param pred the predecessor
	 * @param succ the successor
	 */
	private static void connect(InnerFlowNode pred, InnerFlowNode succ) {
		if (!pred.addSuccecessor(succ)) {
			log.warn("successor already added");
		}
		if (!succ.addPredecessor(pred)) {
			log.warn("predecessor already added.");
		}
	}

	/**
	 * Disconnects two nodes.
	 *
	 * @param pred the predecessor
	 * @param succ the successor
	 */
	private static void disconnect(InnerFlowNode pred, InnerFlowNode succ) {
		if (!pred.removeSuccecessor(succ)) {
			log.warn("Could not remove successor");
		}
		if (!succ.removePredecessor(pred)) {
			log.warn("Could not remove predecessor");
		}
	}

	/**
	 * Generates an {@link InnerFlowGraph} for the given {@link InnerFlowNode}
	 * representing a compound statement
	 *
	 * @param curNode the node containing the compound statement
	 * @param leftNodes collection of nodes to be analyzed, to add special child nodes
	 * @return Returns the last node within the compound statement.
	 */
	private static InnerFlowGraph generateCompoundStatement(InnerFlowNode curNode,
													 Collection<InnerFlowNode> leftNodes,
													 Collection<InnerFlowNode> jumpNodes, List<Task> tasks) {
		InnerFlowNode start;
		InnerFlowNode end;
		InnerFlowNode lastNode;
		InnerFlowNode newNode;

		FlatIterator<Traversable> iter = new FlatIterator<Traversable>(
			curNode.getTraversable());

		start = new InnerFlowStart();
		lastNode = start;

		while (iter.hasNext()) {
			Traversable tChild = iter.next();
			newNode = new InnerFlowNode(tChild);
			connect(lastNode, newNode);

			/*
			 * Check whether one of the nodes within the body of this
			 * CompoundLoop is a special node that needs further inspection
			 */
			handleSpecialNodes(newNode, leftNodes, jumpNodes, tasks);

			lastNode = newNode;

		}
		end = new InnerFlowEnd();
		connect(lastNode, end);

		return new InnerFlowGraph(start, end);
	}

	/**
	 * Generates an {@link InnerFlowGraph} for the given {@link InnerFlowNode}
	 * representing a do loop
	 *
	 * @param curNode the current node
	 * @param leftNodes left nodes
	 * @param jumpNodes jumpnodes
	 * @return the inner flow graph
	 */
	private static InnerFlowGraph generateDoLoop(InnerFlowNode curNode,
										  Collection<InnerFlowNode> leftNodes,
										  Collection<InnerFlowNode> jumpNodes, List<Task> tasks) {
		return generateLoop(curNode, leftNodes, jumpNodes, tasks);
	}

	/**
	 * Generates an {@link InnerFlowGraph} for the given {@link InnerFlowNode}
	 * representing a for loop
	 *
	 * @param curNode the current node
	 * @param leftNodes all left nodes
	 * @param jumpNodes jump nodes
	 * @return the inner flow graph
	 */
	private static InnerFlowGraph generateForLoop(InnerFlowNode curNode,
										   Collection<InnerFlowNode> leftNodes,
										   Collection<InnerFlowNode> jumpNodes, List<Task> tasks) {

		return generateLoop(curNode, leftNodes, jumpNodes, tasks);
	}

	private static InnerFlowGraph generateLoop(InnerFlowNode curNode, Collection<InnerFlowNode> leftNodes, Collection<InnerFlowNode> jumpNodes, List<Task> tasks) {
		InnerFlowNode start = new InnerFlowStart();
		InnerFlowNode end = new InnerFlowEnd();

		InnerFlowNode nLoop = new InnerFlowNode(curNode.getTraversable());

		InnerFlowNode nBody = new InnerFlowNode(
				((Loop) nLoop.getTraversable()).getBody());

		/*
		 * Check whether the body of this Loop is a special node that needs further inspection
		 */
		handleSpecialNodes(nBody, leftNodes, jumpNodes, tasks);

		connect(start, nLoop);
		connect(nLoop, nBody);
		connect(nBody, nLoop);
		connect(nLoop, end);

		return new InnerFlowGraph(start, end);
	}

	/**
	 * Generates an {@link InnerFlowGraph} for the given {@link InnerFlowNode}
	 * representing an if statement
	 *
	 * @param curNode the current node
	 * @param leftNodes all left nodes
	 * @param jumpNodes jump nodes
	 * @return the inner flow graph
	 */
	private static InnerFlowGraph generateIfStatement(InnerFlowNode curNode,
											   Collection<InnerFlowNode> leftNodes,
											   Collection<InnerFlowNode> jumpNodes, List<Task> tasks) {

		InnerFlowNode start = new InnerFlowStart();
		InnerFlowNode end = new InnerFlowEnd();

		InnerFlowNode nIfClause = new InnerFlowNode(curNode.getTraversable());

		connect(start, nIfClause);

		Traversable tIf = ((IfStatement) curNode.getTraversable())
			.getThenStatement();
		Traversable tElse = ((IfStatement) curNode.getTraversable())
			.getElseStatement();

		if (null == tElse) {
			/* if there is no else block */
			InnerFlowNode nIf = new InnerFlowNode(tIf);

			connect(nIfClause, nIf);
			connect(nIf, end);
			connect(nIfClause, end);

			/*
			 * Check whether the if body of this IfStatement is a special node
			 * that needs further inspection
			 */
			handleSpecialNodes(nIf, leftNodes, jumpNodes, tasks);

			return new InnerFlowGraph(start, end);
		} else {
			/* if there is an else block */

			InnerFlowNode nIf = new InnerFlowNode(tIf);
			InnerFlowNode nElse = new InnerFlowNode(tElse);

			connect(nIfClause, nIf);
			connect(nIfClause, nElse);
			connect(nIf, end);
			connect(nElse, end);

			/*
			 * Check whether the if body of this IfStatement is a special node
			 * that needs further inspection
			 */
			handleSpecialNodes(nIf, leftNodes, jumpNodes, tasks);

			/*
			 * Check whether the else body of this IfStatement is a special node
			 * that needs further inspection
			 */
			handleSpecialNodes(nElse, leftNodes, jumpNodes, tasks);

			return new InnerFlowGraph(start, end);
		}
	}

	/**
	 * Generates an {@link InnerFlowGraph} for the given {@link InnerFlowNode}
	 * representing a switch statement
	 *
	 * @param curNode the current node
	 * @param leftNodes all left nodes
	 * @param jumpNodes jump nodes
	 * @return the inner flow graph
	 */
	private static InnerFlowGraph generateSwitchStatement(InnerFlowNode curNode,
												   Collection<InnerFlowNode> leftNodes,
												   Collection<InnerFlowNode> jumpNodes, List<Task> tasks) {

		/* Get the compound statement's graph */
		InnerFlowGraph curGraph = generateCompoundStatement(new InnerFlowNode(
				((SwitchStatement) curNode.getTraversable()).getBody()),
			leftNodes, jumpNodes, tasks);

		/* Generate a node for the switch statement */
		InnerFlowNode switchNode = new InnerFlowNode(curNode.getTraversable());

		InnerFlowNode compoundFirst = curGraph.getInnerFlowStart()
			.getSucccessors().iterator().next();

		/* Reconnect the start of the compound statement with the switch */
		connect(curGraph.getInnerFlowStart(), switchNode);
		connect(switchNode, compoundFirst);
		disconnect(curGraph.getInnerFlowStart(), compoundFirst);

		/* Scan for case statements */
		InnerFlowNode curCompound = compoundFirst;
		while (curCompound != curGraph.getInnerFlowEnd()) {

			if (curCompound.getTraversable() instanceof Case) {
				connect(switchNode, curCompound);
			}

			curCompound = curCompound.getSucccessors().iterator().next();
		}

		return curGraph;

	}

	/**
	 * Generates an {@link InnerFlowGraph} for the given {@link InnerFlowNode}
	 * representing a while loop
	 *
	 * @param curNode the current node
	 * @param leftNodes all left nodes
	 * @param jumpNodes jump nodes
	 * @return the inner flow graph
	 */
	private static InnerFlowGraph generateWhileLoop(InnerFlowNode curNode,
											 Collection<InnerFlowNode> leftNodes,
											 Collection<InnerFlowNode> jumpNodes, List<Task> tasks) {

		return generateLoop(curNode, leftNodes, jumpNodes, tasks);
	}

	/**
	 * Returns the {@link InnerFlowGraph} of the given {@link Traversable} t.
	 *
	 * @param t the {@link Traversable} whose {@link InnerFlowGraph} shall be returned.
	 * @param tasks contains all tasks in the current model
	 * @param errorOnReturnStatement stops and throws Exception if a return statement is used. Usually return statements are only valid for functions and base tasks
	 * @return the {@link InnerFlowGraph} of the given {@link Traversable}.
	 */
	public static InnerFlowGraph getControlFlowForTask(Traversable t, List<Task> tasks, boolean errorOnReturnStatement) {
		return getControlFlow(t, false, tasks, errorOnReturnStatement);
	}

	/**
	 * Returns the {@link InnerFlowGraph} of the given {@link Traversable} t.
	 *
	 * @param t the {@link Traversable} whose {@link InnerFlowGraph} shall be returned.
	 * @param tasks contains all tasks in the current model
	 * @return the {@link InnerFlowGraph} of the given {@link Traversable}.
	 */
	public static InnerFlowGraph getControlFlowForFunction(Traversable t, List<Task> tasks) {
		return getControlFlow(t, true, tasks, true);
	}


	/**
	 * Returns the {@link InnerFlowGraph} of the given {@link Traversable} t.
	 *
	 * @param t the {@link Traversable} whose {@link InnerFlowGraph} shall be returned.
	 * @param isFunction tells if a function shall be processed
	 * @param tasks contains all tasks in the current model
	 * @param errorOnReturnStatement stops and throws Exception if a return statement is used. Usually return statements are only valid for functions and base tasks
	 * @return the {@link InnerFlowGraph} of the given {@link Traversable}.
	 */
	public static InnerFlowGraph getControlFlow(Traversable t, boolean isFunction, List<Task> tasks, boolean errorOnReturnStatement) {

		return getInnerFlowGraph(t, isFunction, tasks, errorOnReturnStatement, true);
	}

	/**
	 * Returns the {@link InnerFlowGraph} of the given {@link Traversable} t. Without handling Task specific stuff
	 * @param t a traversable to generate a inner flow graph for
	 * @return the {@link InnerFlowGraph} of the given {@link Traversable}.
	 */
	public static InnerFlowGraph getControlFlowWithoutTaskHandling(Traversable t) {

		return getInnerFlowGraph(t, false, Arrays.asList(), true, false);
	}

	private static InnerFlowGraph getInnerFlowGraph(Traversable t, boolean isFunction, List<Task> tasks, boolean errorOnReturnStatement, boolean taskSpecificJumpNodeHandling) {
		InnerFlowNode curNode = null;

		ArrayList<InnerFlowNode> leftNodes = new ArrayList<InnerFlowNode>(
			Collections.singleton(new InnerFlowNode(t)));

		ArrayList<InnerFlowNode> jumpNodes = new ArrayList<InnerFlowNode>();

		InnerFlowGraph graph = null;
		InnerFlowGraph curGraph = null;

		while (!leftNodes.isEmpty()) {
			curNode = leftNodes.remove(0);

			Traversable curNodeTraversable = curNode.getTraversable();
			Optional<Task> taskWithCurrentTraversable = tasks.stream().filter(tsk -> tsk.getTraversable().equals(curNodeTraversable)).collect(StreamUtil.atMostOne());
			if (t != curNode.getTraversable()
				&& taskWithCurrentTraversable.isPresent()) {
				/* if it is a task ... */
				InnerFlowNode tNode = new InnerFlowTask(taskWithCurrentTraversable.get());

				curGraph = new InnerFlowGraph(new InnerFlowStart(),
					new InnerFlowEnd());
				connect(curGraph.getInnerFlowStart(), tNode);
				connect(tNode, curGraph.getInnerFlowEnd());

			} else if (curNode.getTraversable() instanceof CompoundStatement) {

				/*
				 * if the current node represents a compound statement generate
				 * its content and inner flow:
				 */
				curGraph = generateCompoundStatement(curNode, leftNodes, jumpNodes, tasks);

			} else if (curNode.getTraversable() instanceof Procedure) {
				// use compund statement for now
				curGraph = generateCompoundStatement(curNode, leftNodes, jumpNodes, tasks);

			} else if (curNode.getTraversable() instanceof SwitchStatement) {

				/*
				 * if the current node represents a switch statement generate
				 * its content and inner flow:
				 */
				curGraph = generateSwitchStatement(curNode, leftNodes,
					jumpNodes, tasks);

			} else if (curNode.getTraversable() instanceof IfStatement) {

				/*
				 * if the current node represents a if statement generate its
				 * content and inner flow:
				 */
				curGraph = generateIfStatement(curNode, leftNodes, jumpNodes, tasks);

			} else if (curNode.getTraversable() instanceof WhileLoop) {

				/*
				 * if the current node represents a while statement generate its
				 * content and inner flow:
				 */
				curGraph = generateWhileLoop(curNode, leftNodes, jumpNodes, tasks);

			} else if (curNode.getTraversable() instanceof ForLoop) {

				/*
				 * if the current node represents a for loop statement generate
				 * its content and inner flow:
				 */
				curGraph = generateForLoop(curNode, leftNodes, jumpNodes, tasks);

			} else if (curNode.getTraversable() instanceof DoLoop) {

				/*
				 * if the current node represents a do loop statement generate
				 * its content and inner flow:
				 */
				curGraph = generateDoLoop(curNode, leftNodes, jumpNodes, tasks);

			} else {

				if (!curNode.getTraversable().equals(t)) {
					log.warn("single statement should not handled here!");
				}
				/*
				 * if the given traversable is just one statement, generate its
				 * linear inner flow
				 */
				InnerFlowStart start = new InnerFlowStart();
				InnerFlowEnd end = new InnerFlowEnd();
				connect(start, curNode);
				connect(curNode, end);
				curGraph = new InnerFlowGraph(start, end);
			}

			/*
			 * Replace InnerFlowNode with new developed InnerFlowGraph. If it is
			 * root traversable save as result value
			 */
			if (null == graph) {
				graph = curGraph;
			} else {
				replace(curNode, curGraph);
			}

		}

		/* handling all jump nodes */
		if (taskSpecificJumpNodeHandling) {
			for (InnerFlowNode nJump : jumpNodes) {
				Traversable tJump = nJump.getTraversable();

				if (tJump instanceof BreakStatement) {
					handleBreakStatement(nJump, tJump, graph);
				} else if (tJump instanceof ContinueStatement) {
					handleContinueStatement(nJump, tJump, graph);
				} else if (tJump instanceof GotoStatement) {
					handleGotoStatement(nJump, tJump, graph);
				} else if (tJump instanceof ReturnStatement) {
					handleReturnStatement(nJump, tJump, isFunction, graph, errorOnReturnStatement);

				}
			}
		}

		return graph;
	}

	/**
	 * Adds additional connectivity to a break statement {@link InnerFlowNode}.
	 *
	 * @param nJump InnerFlowNode of the break statement
	 * @param tJump Traversable of the break statement
	 * @param graph The corresponding InnerFlowGraph.
	 */
	private static void handleBreakStatement(InnerFlowNode nJump, Traversable tJump,
									  InnerFlowGraph graph) {

		Traversable tChild = tJump;
		Traversable tParent = tJump.getParent();

		while (null != tParent) {

			if (tParent instanceof ForLoop || tParent instanceof WhileLoop
				|| tParent instanceof DoLoop) {

				boolean found = false;
				for (InnerFlowNode node : graph.getAllNodes()) {
					if (node.getTraversable() == tParent) {
						found = true;
						/*
						 * the successor of the loop that is not the loops body
						 * have to be the next statement.
						 */
						for (InnerFlowNode succ : node.getSucccessors()) {
							if (succ != tChild) {
								for (InnerFlowNode jumpSucc : nJump
									.getSucccessors()) {
									disconnect(nJump, jumpSucc);
								}
								connect(nJump, succ);
								return;
							}
						}

						break;
					}
				}
				if (!found) {
					//CHECKSTYLE.OFF: MultipleStringLiterals
					final String message = "target loop " + tParent.getClass().getSimpleName() + " of continue not inside same task";
					//CHECKSTYLE.ON: MultipleStringLiterals
					throw new RuntimeException(message);
				}
				break;

			}

			tChild = tParent;
			tParent = tParent.getParent();
		}

	}

	/**
	 * Adds additional connectivity to a continue statement
	 * {@link InnerFlowNode}.
	 *
	 * @param nJump InnerFlowNode of the continue statement
	 * @param tJump Traversable of the continue statement
	 * @param graph The corresponding InnerFlowGraph.
	 */
	private static void handleContinueStatement(InnerFlowNode nJump,
										 Traversable tJump, InnerFlowGraph graph) {
		Traversable tParent = tJump.getParent();

		while (null != tParent) {

			if (tParent instanceof ForLoop || tParent instanceof WhileLoop
				|| tParent instanceof DoLoop) {

				boolean found = addShortcutJump(nJump, graph.getAllNodes(), tParent);
				if (!found) {
					throw new RuntimeException("target loop " + tParent.getClass().getSimpleName() + " of continue not inside same task");
				}
				break;

			}

			tParent = tParent.getParent();
		}

	}

	private static boolean addShortcutJump(InnerFlowNode nJump, Collection<InnerFlowNode> allNodes, Traversable tParent) {
		boolean found = false;
		for (InnerFlowNode node : allNodes) {
			if (node.getTraversable() == tParent) {
				found = true;
				for (InnerFlowNode succ : nJump.getSucccessors()) {
					disconnect(nJump, succ);
				}
				connect(nJump, node);
				break;
			}
		}
		return found;
	}

	/**
	 * Adds additional connectivity to a goto statement {@link InnerFlowNode}.
	 *
	 * @param nJump InnerFlowNode of the goto statement
	 * @param tJump Traversable of the goto statement
	 * @param graph The corresponding InnerFlowGraph.
	 */
	private static void handleGotoStatement(InnerFlowNode nJump, Traversable tJump,
									 InnerFlowGraph graph) {
		Label tLabel = ((GotoStatement) tJump).getTarget();

		boolean found = addShortcutJump(nJump, graph.getAllNodes(), tLabel);
		if (!found) {
			throw new RuntimeException("target label " + tLabel.getName() + " of goto not inside same task");
		}

	}

	/**
	 * Adds additional connectivity to a return statement {@link InnerFlowNode}.
	 *
	 * @param nJump InnerFlowNode of the return statement
	 * @param tJump Traversable of the return statement
	 * @param isFunction indicates whether the InnerFlowGraph represents a Function or not.
	 * @param graph The corresponding InnerFlowGraph.
	 */
	private static void handleReturnStatement(InnerFlowNode nJump, Traversable tJump,
									   boolean isFunction, InnerFlowGraph graph, boolean errorOnReturnStatement) {
		if (!isFunction) {
			if (errorOnReturnStatement) {
				throw new RuntimeException("cannot handle return statement inside task");
			}
		}
		/* if it is a function connect with the end */
		for (InnerFlowNode succ : nJump.getSucccessors()) {
			disconnect(nJump, succ);
		}
		connect(nJump, graph.getInnerFlowEnd());

	}

	/**
	 * Checks whether the given {@link InnerFlowNode} is a "special node". A
	 * special node is a node that represents a {@link Traversable} (Cetus IR)
	 * of a {@link Task}, a statement that contains nested statements like
	 * {@link SwitchStatement}, {@link CompoundStatement}, {@link IfStatement},
	 * {@link ForLoop}, {@link DoLoop}, {@link WhileLoop} or statements
	 * representing a non linear program flow within the source code code like
	 * {@link GotoStatement}, {@link ContinueStatement}, {@link ReturnStatement}
	 * , {@link BreakStatement}. If the given statement one of these special
	 * nodes it is added to one of the two given collections leftNodes and
	 * jumpNodes for further processing.
	 *
	 * @param newNode the node which is checked whether it belongs to the special nodes.
	 * @param leftNodes a collection in which all nodes which needs further inspection are collected.
	 * @param jumpNodes a collection in which all jump nodes are collected.
	 */
	private static void handleSpecialNodes(InnerFlowNode newNode,
									Collection<InnerFlowNode> leftNodes,
									Collection<InnerFlowNode> jumpNodes, List<Task> tasks) {
		Traversable tChild = newNode.getTraversable();

		/* Is it a Task? */
		boolean isTask = tasks.stream().anyMatch(t -> t.getTraversable().equals(tChild));
		if (isTask) {
			leftNodes.add(newNode);
			return;
		}

		/* Special Nodes that are needed to be analysed in detail. */
		if ((tChild instanceof CompoundStatement
			|| tChild instanceof SwitchStatement
			|| tChild instanceof IfStatement || tChild instanceof DoLoop
			|| tChild instanceof ForLoop || tChild instanceof WhileLoop /*
																			 * TODO
																			 * more
																			 * ?
																			 */)) {
			leftNodes.add(newNode);

			return;
		}

		/*
		 * break, continue, goto, return are connected at the end and saved
		 * until then in the jumpNodes collection.
		 */
		if (tChild instanceof BreakStatement || tChild instanceof GotoStatement
			|| tChild instanceof ContinueStatement
			|| tChild instanceof ReturnStatement) {
			jumpNodes.add(newNode);
			return;
		}

	}

	/**
	 * Replaces the given {@link InnerFlowNode} node with the given
	 * {@link InnerFlowGraph} graph.
	 *
	 * @param node the {@link InnerFlowNode} within a graph that shall be substituted with another graph.
	 * @param graph the graph which shall be substitude the given node within it's graph.
	 */
	private static void replace(InnerFlowNode node, InnerFlowGraph graph) {
		if (null == graph || null == node) {
			return;
		}

		Collection<InnerFlowNode> predecessors = new ArrayList<InnerFlowNode>(
			node.getPredecessors());

		for (InnerFlowNode predecessor : predecessors) {
			disconnect(predecessor, node);
		}

		InnerFlowNode start = graph.getInnerFlowStart();
		Collection<InnerFlowNode> startSuccessors = new ArrayList<InnerFlowNode>(
			start.getSucccessors());

		for (InnerFlowNode first : startSuccessors) {
			disconnect(start, first);
			for (InnerFlowNode predecessor : predecessors) {
				connect(predecessor, first);
			}
		}

		Collection<InnerFlowNode> successors = new ArrayList<InnerFlowNode>(
			node.getSucccessors());

		for (InnerFlowNode successor : successors) {
			disconnect(node, successor);
		}

		InnerFlowNode end = graph.getInnerFlowEnd();
		Collection<InnerFlowNode> endPredecessors = new ArrayList<InnerFlowNode>(
			end.getPredecessors());

		for (InnerFlowNode last : endPredecessors) {
			disconnect(last, end);
			for (InnerFlowNode successor : successors) {
				connect(last, successor);
			}
		}

		// TODO: debug remove later
		if (node.getPredecessors().size() != 0
			|| node.getSucccessors().size() != 0) {
			throw new RuntimeException("Replacement failed!");
		}
	}

}
