/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.controlflow;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.interfaces.HasCetusComponent;

import cetus.hir.Traversable;

import de.tu_darmstadt.rs.microStreams.model.software.SoftwareModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess.AccessBehavior;

import java.util.Collection;
import java.util.LinkedHashSet;

/**
 * <p>
 * The {@link InnerFlowNode} represents a node within the {@link InnerFlowGraph}
 * (a control flow graph of a {@link Task Task's} or {@link Function Function's}
 * {@link Traversable}).
 * </p>
 * <p>
 * <p>
 * A {@link InnerFlowGraph} consists of interconnected {@link InnerFlowNode
 * InnerFlowNodes} representing source code lines within the original C firmware
 * and is stored its task or function. Each {@link InnerFlowNode} holds a
 * collection of direct successor nodes, predecessor nodes, the
 * {@link Traversable} of the source code line and a list of
 * {@link VariableAccess VariableAccesses} that takes place within the nodes
 * source code line. There are several inheriting classes for special purposes:
 * <p>
 * <p>
 * <ul>
 * <li>{@link InnerFlowStart} - does not represent an actual source code line it
 * just marks the entry point of the whole program flow graph</li>
 * <li>{@link InnerFlowEnd} - does not represent an actual source code line it
 * just marks the exit point of the whole program flow graph</li>
 * <li>{@link InnerFlowTask} - represents a child task access within the program
 * flow</li>
 * </ul>
 * <p>
 *
 * @author Jan Weber
 */
public class InnerFlowNode implements SoftwareModelComponent, HasCetusComponent {

	/*
	 * All predecessing InnerFlowNodes.
	 */
	private Collection<InnerFlowNode> predecessors = new LinkedHashSet<InnerFlowNode>();
	/*
	 * All succeeding InnerFlowNodes.
	 */
	private Collection<InnerFlowNode> successors = new LinkedHashSet<InnerFlowNode>();
	/*
	 * the model.
	 */
	private Model model;
	/*
	 * The corresponding traversable from the Cetus IR.
	 */
	private Traversable traversable;
	/*
	 * If in this InnerFlowNode Variables are accessed these accesses are stored
	 * here.
	 */
	private Collection<VariableAccess> access = new LinkedHashSet<VariableAccess>();

	/**
	 * Constructor expecting the corresponding {@link Traversable} (Cetus IR).
	 *
	 * @param traversable the traversable
	 */
	public InnerFlowNode(Traversable traversable) {
		this.traversable = traversable;
	}

	@Override
	public String toString() {

		//String s = CetusUtils.getStringFromPrintable(traversable);

		return traversable.toString();
	}

	/**
	 * Adds the predecessing {@link InnerFlowNode} node.
	 *
	 * @param node a node
	 * @return Returns true if the method call has changed this {@link InnerFlowNode}
	 */
	public boolean addPredecessor(InnerFlowNode node) {
		return predecessors.add(node);

	}

	/**
	 * Adds the succeeding {@link InnerFlowNode} node.
	 *
	 * @param node a node
	 * @return Returns true if the method call has changed this {@link InnerFlowNode}
	 */
	public boolean addSuccecessor(InnerFlowNode node) {
		return successors.add(node);
	}

	/**
	 * Removes the predecessing {@link InnerFlowNode} node from this
	 * {@link InnerFlowNode}.
	 *
	 * @param node a node
	 * @return true if the predecessor was removed from this {@link InnerFlowNode} as a result of this call.
	 */
	public boolean removePredecessor(InnerFlowNode node) {
		return predecessors.remove(node);

	}

	/**
	 * Removes the succeeding {@link InnerFlowNode} node from this
	 * {@link InnerFlowGraph}.
	 *
	 * @param node a node
	 * @return true if the successor was removed from this {@link InnerFlowNode} as a result of this call.
	 */
	public boolean removeSuccecessor(InnerFlowNode node) {
		return successors.remove(node);
	}

	/**
	 * Returns all predecessing {@link InnerFlowNode InnerFlowNodes}.
	 *
	 * @return Returns all predecessing {@link InnerFlowNode}s.
	 */
	public Collection<InnerFlowNode> getPredecessors() {
		return predecessors;
	}

	/**
	 * Returns all succeeding {@link InnerFlowNode InnerFlowNodes}.
	 *
	 * @return Returns all succeeding {@link InnerFlowNode}s.
	 */
	public Collection<InnerFlowNode> getSucccessors() {
		return successors;
	}

	@Override
	public Model getModel() {
		return model;
	}

	@Override
	public void setModel(Model m) {
		this.model = m;
	}

	@Override
	public Traversable getTraversable() {
		return traversable;
	}

	@Override
	public void setTraversable(Traversable traversable) {
		this.traversable = traversable;

	}

	/**
	 * Adds a {@link VariableAccess} to this {@link InnerFlowNode} that is
	 * accessed within this {@link InnerFlowNode InnerFlowNode's}
	 * {@link Traversable} (Cetus IR).
	 *
	 * @param access the
	 */
	public void addVariableAccess(VariableAccess access) {
		this.access.add(access);
	}

	/**
	 * give all read variables
	 * @return all read variables
	 */
	public Collection<Variable> getReadVariables() {

		Collection<Variable> result = new LinkedHashSet<Variable>();

		for (VariableAccess access : this.access) {
			if (AccessBehavior.READ == access.getBehavior()
				|| AccessBehavior.READWRITE == access.getBehavior()) {
				result.add(access.getVariable());
			}
		}

		return result;

	}

	/**
	 * give all written variables
	 * @return all written variables
	 */
	public Collection<Variable> getWrittenVariables() {
		Collection<Variable> result = new LinkedHashSet<Variable>();

		for (VariableAccess access : this.access) {
			if (AccessBehavior.WRITE == access.getBehavior()
				|| AccessBehavior.READWRITE == access.getBehavior()) {
				result.add(access.getVariable());
			}
		}

		return result;

	}


}
