/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.controlflow;

import cetus.hir.Traversable;
import de.tu_darmstadt.rs.microStreams.model.software.SoftwareModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;

import java.util.Collection;
import java.util.Comparator;

/**
 * <p>
 * A {@link InnerFlowGraph} consists of interconnected {@link InnerFlowNode
 * InnerFlowNodes} representing source code lines within the original C firmware
 * and is stored its task or function. Each {@link InnerFlowNode} holds a
 * collection of direct successor nodes, predecessor nodes, the
 * {@link Traversable} of the source code line and a list of
 * {@link VariableAccess VariableAccesses} that takes place within the nodes
 * source code line. There are several inheriting classes for special purposes:
 * <p>
 * <p>
 * <ul>
 * <li>{@link InnerFlowStart} - does not represent an actual source code line it
 * just marks the entry point of the whole program flow graph</li>
 * <li>{@link InnerFlowEnd} - does not represent an actual source code line it
 * just marks the exit point of the whole program flow graph</li>
 * <li>{@link InnerFlowTask} - represents a child task access within the program
 * flow</li>
 * </ul>
 * <p>
 *
 * @author Jan Weber
 */
public class InnerFlowTask extends InnerFlowNode implements
	SoftwareModelComponent {

	private Task task;

	private static final Comparator<InnerFlowTask> idComparator = Comparator.comparingInt(x -> x.getTask().getID());

	public InnerFlowTask(Task task) {
		super(task.getTraversable());
		this.task = task;
	}

	@Override
	public String toString() {
		return "Call " + task;
	}

	@Override
	public void addVariableAccess(VariableAccess access) {
		/*
		 * Variable informations are accessed directly through task.
		 */
		throw new UnsupportedOperationException();
	}

	@Override
	public Collection<Variable> getReadVariables() {
		return task.getCombinedReadVariables();

	}

	@Override
	public Collection<Variable> getWrittenVariables() {
		return task.getCombinedWriteVariables();

	}

	/**
	 * Task ID comparison
	 * @param n1 reference
	 * @param n2 to compare against
	 * @return true if the both arguments are #InnerFlowTask and the second task ID is less than the one of the first argument
	 */
	public static boolean isTaskIDLess(InnerFlowNode n1, InnerFlowNode n2) {
		return n1 instanceof InnerFlowTask && n2 instanceof InnerFlowTask && InnerFlowTask.idComparator.compare((InnerFlowTask) n1, (InnerFlowTask) n2) < 0;
	}

	/**
	 * Task ID comparison
	 * @param n1 reference
	 * @param n2 to compare against
	 * @return true if the both arguments are #InnerFlowTask and the second task ID is greater than the one of the first argument
	 */
	public static boolean isTaskIDGreater(InnerFlowNode n1, InnerFlowNode n2) {
		return n1 instanceof InnerFlowTask && n2 instanceof InnerFlowTask && InnerFlowTask.idComparator.compare((InnerFlowTask) n1, (InnerFlowTask) n2) > 0;
	}

	public Task getTask() {
		return task;
	}

}
