/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.drivers;

import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IConnectionPeripheral;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

/**
 * Model of Peripherals accessing the same resource...for example UART MUX
 */
public class CombinedConnectionPeripheralAccess {

	private int id;

	private ArrayList<ConnectionPeripheralAccess> accesses = new ArrayList<ConnectionPeripheralAccess>();

	private Logger log = LoggerFactory.getLogger(this.getClass());


	public CombinedConnectionPeripheralAccess(int id) {
		this.id = id;
	}

	public boolean addConnectionPeripheralAccess(ConnectionPeripheralAccess access) {

		return accesses.add(access);
	}

	/**
	 * find all peripherals
	 * @return all bundled peripherals
	 */
	public Collection<IConnectionPeripheral> getPeripherals() {

		ArrayList<IConnectionPeripheral> peripherals = new ArrayList<IConnectionPeripheral>();

		Collections.sort(accesses);

		for (ConnectionPeripheralAccess access : accesses) {
			if (!peripherals.contains(access.getPeripheral())) {
				peripherals.add(access.getPeripheral());
			}
		}

		return peripherals;

	}

	/**
	 * all variables for this connection
	 * @return all variables send through the connection
	 */
	public List<Variable> getVariables() {
		List<Variable> variables = new ArrayList<Variable>();

		for (ConnectionPeripheralAccess access : accesses) {
			Variable variable = access.getVariable();
			if (!variables.contains(variable)) {
				variables.add(variable);
			}
		}

		Collections.sort(variables);

		return variables;

	}

	/**
	 * The amount of parallel tasks
	 * @return amount of producers/consumers in the connection
	 */
	public int getReplicationFactor() {
		if (accesses == null || accesses.size() == 0) {
			return 1;
		}

		return accesses.get(0).getDependency().getReplica().size() + 1;
	}

	public int getId() {
		return id;
	}

	/**
	 * gives the peripheral usage type
	 * @return if it's send/rec type as a string
	 */
	public PeripheralUsage getType() {
		if (!accesses.isEmpty()) {
			return accesses.stream().findFirst().get().getType();
		}
		throw new UnsupportedOperationException("No accesses.");
	}

	public boolean isEmpty() {
		return accesses.isEmpty();
	}

}
