/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.drivers;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.Connection;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IConnectionPeripheral;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency.DependencyRelation;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Set;

/**
 * <p>
 * The {@link ConnectionPeripheralAccess} indicates by which connection a
 * TaskDependency is realized. Thus it represents a kind of mapping between
 * {@link Connection Connections} and {@link TaskDependency TaskDependencies}.
 * </p>
 *
 * @author Jan Weber
 */
public class ConnectionPeripheralAccess implements ModelComponent,
		Comparable<ConnectionPeripheralAccess> {

	/**
	 * Indicates whether this {@link ConnectionPeripheralAccess} is located in
	 * the parent task or in the child task respectively one of the child tasks.
	 */
	boolean isInParent;
	private Logger log = LoggerFactory
			.getLogger(ConnectionPeripheralAccess.class);
	/**
	 * {@link TaskDependency} this {@link ConnectionPeripheralAccess}
	 * corresponds to.
	 */
	private TaskDependency dependency;
	/**
	 * {@link Connection} this {@link ConnectionPeripheralAccess} uses.
	 */
	private IConnectionPeripheral peripheral;
	/**
	 * Indicates whether this {@link ConnectionPeripheralAccess} is a read or
	 * write access.
	 */
	private boolean isReceive;
	private Model model;

	private Task task;

	/**
	 * the constructor
	 * @param dependency the task dependency to be modeled
	 * @param task an according task
	 * @param peripheral the peripheral to be used
	 * @param isReceive if the task is receiving data
	 */
	public ConnectionPeripheralAccess(TaskDependency dependency, Task task, IConnectionPeripheral peripheral, boolean isReceive) {
		this.dependency = dependency;
		this.peripheral = peripheral;
		this.isReceive = isReceive;
		this.task = task;

		/*
		 * Determine whether this ConnectionAccess is located within parent or
		 * child task.
		 */
		if (DependencyRelation.CHILD_CHILD == dependency.getRelation()) {
			isInParent = false;
		} else if (DependencyRelation.PARENT_CHILD == dependency.getRelation()) {
			if (isReceive) {
				isInParent = false;
			} else {
				isInParent = true;
			}
		} else if (DependencyRelation.CHILD_PARENT == dependency.getRelation()) {
			if (isReceive) {
				isInParent = true;
			} else {
				isInParent = false;
			}
		} else {
			throw new RuntimeException("Could not determine location of task dependency " + dependency);
		}
	}

	/**
	 * Returns the corresponding {@link Connection}.
	 *
	 * @return Returns the corresponding {@link Connection}.
	 */
	public Set<? extends Connection> getConnection() {
		return peripheral.getConnections();
	}

	/**
	 * Returns the corresponding {@link TaskDependency}.
	 *
	 * @return Returns the corresponding {@link TaskDependency}.
	 */
	public TaskDependency getDependency() {
		return dependency;
	}

	public Task getTask() {
		return task;
	}

	/**
	 * Returns the corresponding {@link IConnectionPeripheral}.
	 *
	 * @return Returns the corresponding {@link IConnectionPeripheral}.
	 */
	public IConnectionPeripheral getPeripheral() {
		return peripheral;
	}

	/**
	 * Returns the accessed {@link Variable}.
	 *
	 * @return Returns the accessed {@link Variable}.
	 */
	public Variable getVariable() {
		if (dependency.getProvider() == task) {
			return dependency.getVariableOfProvider();
		} else {
			return dependency.getVariableOfConsumer();
		}
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
				+ ((dependency == null)? 0 : dependency.hashCode());
		result = prime * result + (isInParent? 1231 : 1237);
		result = prime * result + (isReceive? 1231 : 1237);
		result = prime * result + ((log == null)? 0 : log.hashCode());
		result = prime * result
				+ ((peripheral == null)? 0 : peripheral.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		ConnectionPeripheralAccess other = (ConnectionPeripheralAccess) obj;
		if (dependency == null) {
			if (other.dependency != null) {
				return false;
			}
		} else if (!dependency.equals(other.dependency)) {
			return false;
		}
		if (isInParent != other.isInParent) {
			return false;
		}
		if (isReceive != other.isReceive) {
			return false;
		}
		if (log == null) {
			if (other.log != null) {
				return false;
			}
		} else if (!log.equals(other.log)) {
			return false;
		}
		if (peripheral == null) {
			if (other.peripheral != null) {
				return false;
			}
		} else if (!peripheral.equals(other.peripheral)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "ConnectionPeripheralAccess [\n dependency=" + dependency
				+ ",\n peripheral=" + peripheral + ",\n isReceive=" + isReceive
				+ ", task=" + task + "]\n\n";
	}

	@Override
	public int compareTo(ConnectionPeripheralAccess other) {

		if ((isReceive && !other.isReceive)) {
			return 1;
		} else if (!isReceive && other.isReceive) {
			return -1;
		} else if ((isInParent && !other.isInParent)) {
			return 1;
		} else if (!isInParent && other.isInParent) {
			return -1;
		} else if ((null != peripheral && null != other.peripheral)
				&& (peripheral.getId() != other.peripheral.getId())) {

			return peripheral.getId() - other.peripheral.getId();

		} else {
			return dependency.compareTo(other.dependency);

		}
	}

	public PeripheralUsage getType() {
		return isReceive? PeripheralUsage.RECEIVE : PeripheralUsage.SEND;

	}

	public boolean isInParent() {
		return isInParent;
	}

	public boolean isInChild() {
		return !isInParent;
	}

	public boolean isReceive() {
		return isReceive;
	}

	public boolean isSend() {
		return !isReceive;
	}

	@Override
	public Model getModel() {
		return this.model;
	}

	@Override
	public void setModel(Model model) {
		this.model = model;

	}

}
