/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.function;

import de.tu_darmstadt.rs.microStreams.model.Model;

import cetus.hir.Declaration;
import cetus.hir.FunctionCall;
import cetus.hir.Identifier;
import cetus.hir.Procedure;
import cetus.hir.Specifier;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;

import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.interfaces.AccessesVariables;
import de.tu_darmstadt.rs.microStreams.model.interfaces.CallsFunctions;
import de.tu_darmstadt.rs.microStreams.model.interfaces.ContainsChildTask;
import de.tu_darmstadt.rs.microStreams.model.interfaces.FreemarkerHasBody;
import de.tu_darmstadt.rs.microStreams.model.interfaces.HasCetusComponent;
import de.tu_darmstadt.rs.microStreams.model.interfaces.NeedsIncludes;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.ConnectionPeripheralAccess;
import de.tu_darmstadt.rs.microStreams.model.software.freemarker.Body;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.variable.TaskVariable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.tools.BodyBuilder;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.TreeSet;
import java.util.stream.Collectors;

/**
 * Respresents functions called by tasks that are located within the same source
 * code file as the task.
 *
 * @author jan
 */
public class UFunction implements CallsFunctions, NeedsIncludes,
		ContainsChildTask, FreemarkerHasBody, HasCetusComponent,
		ModelComponent, AccessesVariables {

	private Logger log = LoggerFactory.getLogger(UFunction.class);

	private Procedure traversable;

	private Collection<UFunctionCall> calls = new LinkedHashSet<UFunctionCall>();

	private Collection<Task> containedTasks;

	private Collection<TaskVariable> variables;

	private Collection<ConnectionPeripheralAccess> connectionPeripheralAccesses = new TreeSet<ConnectionPeripheralAccess>();

	private List<Body> body;

	private boolean isBodyGenerated = false;

	private boolean evaluation = false;

	private Model project;

	private InnerFlowGraph graph;

	/**
	 * The Constructor
	 * @param traversable a traversable for this function
	 * @param containedTasks the tasks contained in the function
	 * @param model the model
	 */
	public UFunction(Procedure traversable, Collection<Task> containedTasks, Model model) {
		this.traversable = traversable;
		this.containedTasks = containedTasks;
		this.variables = new LinkedHashSet<>();
		setModel(model);
		evaluation = getModel().getConfig().isCreatePerformanceEvaluation();
	}

	@Override
	public boolean addFunctionCall(UFunctionCall function) {
		return calls.add(function);
	}

	@Override
	public boolean removeFunctionCall(UFunctionCall function) {
		return calls.remove(function);
	}

	@Override
	public Collection<UFunctionCall> getFunctionCalls() {
		return calls;
	}

	@Override
	public boolean addTask(Task task) {
		return containedTasks.add(task);
	}

	@Override
	public Collection<Task> getContainedTasks() {
		return containedTasks;
	}

	@Override
	public boolean addConnectionPeripheralAccess(ConnectionPeripheralAccess connectionAccess) {
		return connectionPeripheralAccesses.add(connectionAccess);

	}

	/**
	 * adds a variable access
	 * @param access a variable access to be added
	 * @return true if it could be added
	 */
	public boolean addVariableAccess(VariableAccess access) {
		TaskVariable taskVariable = getTaskVariable(access.getVariable());

		if (null == taskVariable) {
			log.debug("Accessed Variable not stored in this function.");
			addVariable(access.getVariable());
			taskVariable = getTaskVariable(access.getVariable());
		}

		return taskVariable.addAccess(access);
	}

	@Override
	public Collection<Variable> getVariables() {
		return variables.stream()
				.filter(v -> v.getVariable() != null)
				.map(v -> v.getVariable())
				.collect(Collectors.toSet());
	}

	/**
	 * add a variable
	 * @param variable a variable to be added
	 * @return true if the variable could be added
	 */
	public boolean addVariable(Variable variable) {
		TaskVariable taskVariable = getTaskVariable(variable);

		if (null == taskVariable) {
			taskVariable = new TaskVariable(variable);
			return variables.add(taskVariable);
		}
		return false;
	}

	/**
	 * Tells if this function is the according function to the function call
	 * @param functionCall a function call
	 * @return true if the given function call is this function
	 */
	public boolean isAccordingFunction(FunctionCall functionCall) {
		if (functionCall.getProcedure() != null) {
			if (functionCall.getProcedure().equals((Procedure) getTraversable())) {
				return true;
			}
		}
		return false;
	}

	@Override
	public Collection<Include> getIncludeFiles() {
		Collection<Include> includes = new LinkedHashSet<>();
		for (UFunctionCall call : calls) {
			includes.addAll(call.getIncludeFiles());
		}
		return includes;
	}

	private TaskVariable getTaskVariable(Variable variable) {
		for (TaskVariable taskVariable : variables) {
			if (taskVariable.getVariable().equals(variable)) {
				return taskVariable;
			}
		}
		return null;
	}

	@Override
	public Traversable getTraversable() {
		return traversable;
	}

	public Procedure getProcedure() {
		return traversable;
	}

	@Override
	public void setTraversable(Traversable traversable) {
		if (traversable instanceof Procedure) {
			this.traversable = (Procedure) traversable;
		} else {
			throw new RuntimeException("Can not set Traversable that is not of type Procedure");
		}
	}

	/**
	 * Gives the Declaration
	 * @return the function declaration as a string
	 */
	public String getDeclaration() {
		@SuppressWarnings("unchecked")
		List<Specifier> returnTypes = (traversable).getReturnType();

		StringBuilder sb = new StringBuilder();

		for (Specifier returnType : returnTypes) {
			sb.append(returnType.toString());
			sb.append(" ");
		}

		sb.append((traversable).getDeclarator().toString());

		return sb.toString();
	}

	/**
	 * Gives identifiers matching the declaration
	 * @param declaration a declaration
	 * @return all identifiers contained in this function matching the given declaration
	 */
	public Collection<Identifier> getIdentifiersMatchingDeclaration(Declaration declaration) {
		Collection<Identifier> identifiers = new LinkedList<Identifier>();
		for (Variable var : getVariables()) {
			if (var.getVariableDeclarator().getDeclaration() == declaration) {
				TaskVariable tv = getTaskVariable(var);
				for (VariableAccess va : tv.getVariableAccess()) {
					identifiers.add(va.getIdentifier());
				}
			}
		}
		return identifiers;
	}

	/**
	 * tells if the function declaration resides in the main c file
	 * @return true if the function is defined in the main C-file
	 */
	public boolean isDefinedInMain() {
		final TranslationUnit parentTranslationUnit = CetusUtils.findParentOfType(traversable, TranslationUnit.class);
		final TranslationUnit mainTranslationUnit = CetusUtils.findParentOfType(CetusUtils.getMainProcedure(CetusUtils.getProgram(project)), TranslationUnit.class);
		return parentTranslationUnit.equals(mainTranslationUnit);
	}

	public Collection<Declaration> getFunctionArguments() {
		return traversable.getDeclarations();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((calls == null)? 0 : calls.hashCode());
		result = prime * result
				+ ((containedTasks == null)? 0 : containedTasks.hashCode());
		result = prime * result
				+ ((traversable == null)? 0 : traversable.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		UFunction other = (UFunction) obj;
		if (calls == null) {
			if (other.calls != null) {
				return false;
			}
		} else if (!calls.equals(other.calls)) {
			return false;
		}
		if (containedTasks == null) {
			if (other.containedTasks != null) {
				return false;
			}
		} else if (!containedTasks.equals(other.containedTasks)) {
			return false;
		}
		if (traversable == null) {
			if (other.traversable != null) {
				return false;
			}
		} else if (!traversable.equals(other.traversable)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "Function [traversable=" + traversable + ", calls=" + calls
				+ "]";
	}

	@Override
	public List<Body> getBody() {
		if (!isBodyGenerated) {
			body = new BodyBuilder(getTraversable(), connectionPeripheralAccesses, containedTasks, false, getModel().getConfig().isCreatePerformanceEvaluation()).generateBody();
			isBodyGenerated = true;
		}
		return body;
	}

	@Override
	public void insertEvaluationAccess(boolean evaluation) {
		this.evaluation = evaluation;
	}

	@Override
	public Model getModel() {
		return this.project;
	}

	@Override
	public void setModel(Model project) {
		this.project = project;

	}

	@Override
	public InnerFlowGraph getInnerFlowGraph() {
		return graph;
	}

	public void setInnerFlowGraph(InnerFlowGraph graph) {
		this.graph = graph;

	}


}
