/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.include;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.interfaces.FreemarkerPrintable;
import de.tu_darmstadt.rs.microStreams.model.other.HeaderFile;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;

/**
 * Represents a c header include directive.
 * <p>
 * It implements the
 * {@link FreemarkerPrintable} interface.
 *
 * @author Jan Weber
 */
public class Include implements FreemarkerPrintable, ModelComponent {

	/* file name specified in the include directive */
	private String fileName;

	/* denotes whether the included header is a system library or a local include */
	private boolean isSystemLibrary;

	/* if it is a system library does it need to be included in the config-build.mk */
	private boolean needsToBeAddedToGCCLibObjFiles;

	/* the string that is included in the config-build.mk */
	private String libraryName;

	/* the parent c file of the include (where the include ) */
	private SourceFile location;

	/* the header file that is included. */
	private HeaderFile hFile;

	/* reference to the model root */
	private Model model;

	/**
	 * the constructor
	 * @param fileName a name of a file
	 * @param isSystemLibrary if the file is a system lib
	 * @param needsToBeAddedToGCCLibObjFiles if it is a system library does it need to be included in the config-build.mk
	 */
	public Include(String fileName, boolean isSystemLibrary, boolean needsToBeAddedToGCCLibObjFiles) {
		this(fileName, isSystemLibrary, needsToBeAddedToGCCLibObjFiles, null);
	}

	/**
	 * the constructor
	 * @param fileName a name of a file
	 * @param isSystemLibrary if the file is a system lib
	 * @param libraryName set a specific library name for config-build.mk
	 */
	public Include(String fileName, boolean isSystemLibrary, String libraryName) {
		this(fileName, isSystemLibrary, true, libraryName);
	}

	private Include(String fileName, boolean isSystemLibrary,  boolean needsToBeAddedToGCCLibObjFiles, String libraryName) {
		this.isSystemLibrary = isSystemLibrary;
		this.needsToBeAddedToGCCLibObjFiles = needsToBeAddedToGCCLibObjFiles;
		String headerEnding = "." + Constants.FilesAndFolders.HEADERFILES_EXTENSION;
		if (fileName.endsWith(headerEnding)) {
			this.fileName = fileName;
			if (libraryName == null || libraryName.isEmpty()) {
				//use the filename as libname
				this.libraryName = fileName.substring(0, fileName.lastIndexOf(headerEnding));
			} else {
				this.libraryName = libraryName;
			}
		} else {
			throw new RuntimeException("The given filename " + fileName + "does not match the expected File type: " + headerEnding);
		}
	}

	public HeaderFile getHeaderFile() {
		return hFile;
	}

	public void setHeaderFile(HeaderFile hFile) {
		this.hFile = hFile;

	}

	public String getLibrary() {
		return libraryName;
	}

	public SourceFile getLocation() {
		return location;
	}

	public void setLocation(SourceFile srcFile) {
		this.location = srcFile;

	}

	@Override
	public Model getModel() {
		return this.model;
	}

	@Override
	public void setModel(Model project) {
		this.model = project;

	}

	public String getName() {
		return fileName;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
			+ ((fileName == null)? 0 : fileName.hashCode());
		result = prime * result + ((hFile == null)? 0 : hFile.hashCode());
		result = prime * result + (needsToBeAddedToGCCLibObjFiles? 1231 : 1237);
		result = prime * result + (isSystemLibrary? 1231 : 1237);
		result = prime * result
			+ ((libraryName == null)? 0 : libraryName.hashCode());
		result = prime * result
			+ ((location == null)? 0 : location.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		Include other = (Include) obj;
		if (fileName == null) {
			if (other.fileName != null) {
				return false;
			}
		} else if (!fileName.equals(other.fileName)) {
			return false;
		}
		if (hFile == null) {
			if (other.hFile != null) {
				return false;
			}
		} else if (!hFile.equals(other.hFile)) {
			return false;
		}
		if (needsToBeAddedToGCCLibObjFiles != other.needsToBeAddedToGCCLibObjFiles) {
			return false;
		}
		if (isSystemLibrary != other.isSystemLibrary) {
			return false;
		}
		if (libraryName == null) {
			if (other.libraryName != null) {
				return false;
			}
		} else if (!libraryName.equals(other.libraryName)) {
			return false;
		}
		if (location == null) {
			if (other.location != null) {
				return false;
			}
		} else if (!location.equals(other.location)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "Include [fileName=" + fileName + ", isSystemLibrary="
			+ isSystemLibrary + ", needsToBeAddedToGCCLibObjFiles=" + needsToBeAddedToGCCLibObjFiles
			+ ", location=" + location + ", hFile=" + hFile + "]";
	}

	public boolean inConfigBuild() {
		return needsToBeAddedToGCCLibObjFiles;
	}

	public boolean isSystemLibrary() {
		return isSystemLibrary;
	}

	public static String printUserInclude(String fileName) {
		return Constants.C.USER_INCLUDE_PREFIX + fileName + Constants.C.USER_INCLUDE_SUFFIX;
	}

	public static String printSystemInclude(String fileName) {
		return Constants.C.SYSTEM_INCLUDE_PREFIX + fileName + Constants.C.SYSTEM_INCLUDE_SUFFIX;
	}

	@Override
	public String print() {
		if (isSystemLibrary) {
			return printSystemInclude(fileName);
		} else {
			return printUserInclude(fileName);
		}
	}
}
