/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.task;

import cetus.hir.CompoundStatement;
import cetus.hir.Enumeration;
import cetus.hir.Identifier;
import cetus.hir.Procedure;
import cetus.hir.Traversable;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConnectionPeripheralDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IConnectionPeripheralDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ReceivePeripheral;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.SendPeripheral;
import de.tu_darmstadt.rs.microStreams.model.interfaces.AccessesVariables;
import de.tu_darmstadt.rs.microStreams.model.interfaces.CallsFunctions;
import de.tu_darmstadt.rs.microStreams.model.interfaces.ContainsChildTask;
import de.tu_darmstadt.rs.microStreams.model.interfaces.FreemarkerHasBody;
import de.tu_darmstadt.rs.microStreams.model.interfaces.HasCetusComponent;
import de.tu_darmstadt.rs.microStreams.model.other.HeaderFile;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.ConnectionPeripheralAccess;
import de.tu_darmstadt.rs.microStreams.model.software.freemarker.Body;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunctionCall;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.EndPragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.TaskPragma;
import de.tu_darmstadt.rs.microStreams.model.software.variable.TaskVariable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.tools.BodyBuilder;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.Replicate;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * Implements a Task that contains a piece of source code and manages variables inside that code as well as dependencies to other Tasks when those variables are used in the other task
 *
 * @author Jan Weber
 */
public class Task implements HasCetusComponent, ContainsChildTask,
		FreemarkerHasBody, CallsFunctions, ModelComponent, AccessesVariables {

	Logger log = org.slf4j.LoggerFactory.getLogger(Task.class);
	/**
	 * corresponding parsed pragma annotation
	 */
	private TaskPragma pragma;
	private EndPragma endPragma;
	/**
	 * cetus task core
	 */
	private Traversable traversable;
	/**
	 * All variables accessed by this Task.
	 */
	private Collection<TaskVariable> variables = new LinkedHashSet<TaskVariable>();
	/**
	 * Indicates dependencies between different tasks. This means if a task
	 * reads a variable that is written by another task before. See
	 * {@link TaskDependency} for further informations.
	 */
	private Collection<TaskDependency> dependencies = new LinkedHashSet<>();
	/**
	 * Stores a Collection of {@link Task}s containing child {@link Task}s
	 * of this Task. The Childtask can be a direct children in the Cetus
	 * Internal Representation hierarchy (They have to be removed from the
	 * hierachy before writing out the program). Or they are external children
	 * that are contained in a function.
	 */
	private Collection<Task> childTasks = new LinkedHashSet<Task>();
	/**
	 * Unique identifier for this task.
	 */
	private int id;
	/**
	 * Functions called by this task.
	 */
	private Collection<UFunctionCall> functionCalls = new LinkedHashSet<UFunctionCall>();
	/**
	 * Source files needed for this task.
	 */
	private Collection<SourceFile> sourceFiles = new LinkedHashSet<>();
	/**
	 * Header files needed for this task.
	 */
	private Collection<HeaderFile> headerFiles = new LinkedHashSet<>();
	/**
	 * Indicates that this task is the base task.
	 */
	private boolean isBaseTask;
	private boolean isReplica = false;
	private Task originalTask = null;
	/**
	 * Contains the parent task. This field is null, if this task is the base
	 * task.
	 */
	private Task parentTask;
	/**
	 * Contains the direct parent of this task in the source code. This may be
	 * another {@link Task} or a {@link UFunction}
	 */
	private ContainsChildTask directParent;
	/**
	 * Collection of all {@link ConnectionPeripheralAccess ConnectionAccesses}
	 * done by this task.
	 */
	private Collection<ConnectionPeripheralAccess> connectionPeripheralAccesses = new LinkedHashSet<ConnectionPeripheralAccess>();
	/**
	 * Indicates whether the {@link Body}-{@link List} of this {@link Task} is
	 * already generated or not.
	 */
	private boolean isBodyGenerated = false;
	/**
	 * The {@link Body}-{@link List} of this {@link Task} for the software
	 * generation process.
	 */
	private List<Body> body;
	private boolean evaluation = false;
	/**
	 * The corresponding streaming model.
	 */
	private Model model;
	/**
	 * The control flow graph of the code inside this tasks traversable.
	 */
	private InnerFlowGraph graph;

	private Set<CompoundStatement> addedStatements;

	private Core core;

	/*
	 * set constructor to protected to be invisible outside this package.
	 */
	public Task(int id, Traversable cetusTaskCore, Model model) {
		this(id, cetusTaskCore, null, null, model);
		this.isBaseTask = true;
	}

	/*
	 * set constructor to protected to be invisible outside this package.
	 */
	protected Task(int id, Traversable cetusTaskCore, TaskPragma taskPragma, EndPragma endPragma, Model model) {
		this.id = id;
		this.traversable = cetusTaskCore;
		this.pragma = taskPragma;
		this.endPragma = endPragma;
		this.isBaseTask = false;
		setModel(model);
		this.evaluation = getModel().getConfig().isCreatePerformanceEvaluation();
	}

	/**
	 * adds a dependency to this task and creates the VariableAccess if necessary
	 *
	 * @param taskDependency the {@link TaskDependency} to add to this task
	 * @return true if adding was successful, false if task dependcy already exists
	 */
	public boolean addDependency(TaskDependency taskDependency) {
		Variable variable;
		if (this == taskDependency.getProvider()) {
			variable = taskDependency.getVariableOfProvider();
		} else {
			variable = taskDependency.getVariableOfConsumer();
		}

		TaskVariable taskvariable = getTaskVariable(variable);
		if (taskvariable == null && taskDependency.getLocation() instanceof UFunction) {
			UFunction function = (UFunction) taskDependency.getLocation();
			Procedure procedure = ((Procedure) function.getTraversable());
			boolean variableFoundInFunctionArgument = procedure.getDeclarations().stream().filter(vd -> vd.getDeclaredIDs().containsAll(variable.getIdentifiers())).findAny().isPresent();
			if (variableFoundInFunctionArgument) {
				return dependencies.add(taskDependency);
			}
		}
		if (taskvariable == null || taskvariable.isNotUsed() && taskDependency.isForced()) {
			//a forced task dependency might require to add a variable access that wasn't created earlier
			VariableAccess.AccessBehavior behavior = taskDependency.getProvider() == this? VariableAccess.AccessBehavior.WRITE : VariableAccess.AccessBehavior.READ;
			addVariableAccess(new VariableAccess(this, variable, behavior, taskDependency.isForced()));
		}

		return dependencies.add(taskDependency);
	}

	@Override
	public boolean addFunctionCall(UFunctionCall call) {
		return functionCalls.add(call);
	}

	@Override
	public boolean removeFunctionCall(UFunctionCall call) {
		return functionCalls.remove(call);
	}

	@Override
	public Collection<UFunctionCall> getFunctionCalls() {
		return functionCalls;
	}

	/**
	 * get all function calls
	 * @return all function calls ever
	 */
	public Collection<UFunctionCall> getAllFunctionCalls() {
		Collection<UFunctionCall> functionCalls = new HashSet<>();
		functionCalls.addAll(getFunctionCalls());
		int prevFunctionCallSize = 0;
		while (prevFunctionCallSize < functionCalls.size()) {
			prevFunctionCallSize = functionCalls.size();
			functionCalls.addAll(functionCalls.stream()
					.map(fc -> fc.getFunction())
					.filter(Objects::nonNull)
					.flatMap(f -> f.getFunctionCalls().stream())
					.collect(Collectors.toList()));
		}
		return functionCalls;
	}

	public boolean addSourceFile(SourceFile sf) {
		return sourceFiles.add(sf);
	}

	public boolean addHeaderFile(HeaderFile hf) {
		return headerFiles.add(hf);
	}

	public Collection<SourceFile> getSourceFiles() {
		return sourceFiles;
	}

	public Collection<HeaderFile> getHeaderFiles() {
		return headerFiles;
	}

	@Override
	public boolean addTask(Task childTask) {
		return childTasks.add(childTask);
	}

	/**
	 * Returns a {@link Collection} of all {@link Task direct Childtasks} of
	 * this {@link Task}.
	 *
	 * @return Returns a {@link Collection} of all {@link Task direct
	 * Childtasks} of this {@link Task}.
	 */
	@Override
	public Collection<Task> getContainedTasks() {
		Collection<Task> containedTasks = new LinkedHashSet<Task>();

		for (Task childTask : childTasks) {
			if (childTask.getDirectParent().equals(this)) {
				containedTasks.add(childTask);
			}
		}
		return containedTasks;
	}

	public Collection<Task> getChildTasks() {
		return childTasks;
	}

	public Collection<Task> getAllChildTasks() {
		return StreamUtil.streamTree(this, task -> task.getChildTasks().stream()).collect(Collectors.toList());
	}

	@Override
	public boolean addConnectionPeripheralAccess(
			ConnectionPeripheralAccess access) {
		return connectionPeripheralAccesses.add(access);
	}

	/**
	 * Adds declarations to traversable
	 *
	 * @param compoundStatements a set of compund statements to add
	 * @return true if adding was successful, false otherwise
	 */
	public boolean addDeclarations(Set<CompoundStatement> compoundStatements) {
		if (traversable instanceof CompoundStatement) {
			CompoundStatement cs = ((CompoundStatement) traversable);
			compoundStatements.stream().forEach(statement -> {
				CompoundStatement compoundStatement = statement.clone();
				cs.addStatement(compoundStatement);
				if (addedStatements == null) {
					addedStatements = new HashSet<CompoundStatement>();
				}
				addedStatements.add(compoundStatement);
			});
			return true;
		}
		return false;
	}

	/**
	 * Removes declarations of the traversable
	 *
	 * @return true if removing was successful, false otherwise
	 */
	public boolean removeAddedDeclarations() {
		if (traversable instanceof CompoundStatement) {
			CompoundStatement cs = ((CompoundStatement) traversable);
			addedStatements.stream().forEach(s -> {
				cs.removeStatement(s);
			});
			addedStatements.clear();
			return true;
		}
		return false;
	}

	/**
	 * Adds a variable access to the Task for dependency calculation
	 *
	 * @param access the variable access to be added
	 * @return true if the variable access could be added, false otherwise
	 */
	public boolean addVariableAccess(VariableAccess access) {
		TaskVariable taskVariable = getTaskVariable(access.getVariable());

		if (null == taskVariable) {
			log.debug("Accessed Variable not stored in this task.");
			addVariable(access.getVariable());
			taskVariable = getTaskVariable(access.getVariable());
		}

		return taskVariable.addAccess(access);
	}

	/**
	 * Variables of the Task
	 *
	 * @return all variables contained
	 */
	public Collection<Variable> getVariables() {
		Collection<Variable> allVariables = new LinkedHashSet<Variable>();

		for (TaskVariable variable : variables) {
			allVariables.add(variable.getVariable());
		}

		allVariables.addAll(getAllFunctionVariables().stream().filter(v -> v.isGlobal()).collect(Collectors.toList()));

		return allVariables;
	}

	/**
	 * Adds a variable to the Task
	 *
	 * @param variable variable to be added
	 * @return true if the variable could be added, false otherwise
	 */
	public boolean addVariable(Variable variable) {
		TaskVariable taskVariable = getTaskVariable(variable);

		if (null == taskVariable) {
			taskVariable = new TaskVariable(variable);
			return variables.add(taskVariable);
		}
		return false;
	}

	/**
	 * returns a list of all variables not going into global memory
	 *
	 * @return the list of {@link Variable}s
	 */
	public List<Variable> getLocalVariables() {
		List<Variable> globals = getSortedGlobalVariables();
		return variables.stream()
				.map(v -> v.getVariable())
				.filter(v -> !globals.contains(v))
				.collect(Collectors.toList());

	}


	/**
	 * returns a list of all variables to go into global memory
	 * sorting is done alphabetically
	 *
	 * @return the list of {@link Variable}s
	 */
	public List<Variable> getSortedGlobalVariables() {
		List<Variable> allVariables = new LinkedList<Variable>();

		if (!getModel().getConfig().useGlobalMemory()) {
			return allVariables;
		}

		//base task needs some special handling since it contains all variableaccesses for child taks
		//therefore we first need to determine if there is an actual used global variable
		if (isBaseTask()) {
			boolean globalVars = false;
			for (TaskVariable variable : variables) {
				if ((variable.isUsed() || isGlobalVariableUsedInFunction(variable.getVariable())) && variable.getVariable().isGlobalShared() && !variable.getVariable().isGlobalWritten()) {
					globalVars = true;
				}
			}
			if (!globalVars) {
				return allVariables;
			}
		}


		for (TaskVariable variable : variables) {
			Variable var = variable.getVariable();
			if (var.isGlobalShared() && !var.isGlobalWritten()) {
				allVariables.add(var);
			}
		}

		allVariables.sort(Comparator.comparing(x -> x.getVariableDeclarator().getSymbolName()));
		return allVariables;
	}

	@Override
	public List<Body> getBody() {
		if (!isBodyGenerated) {
			BodyBuilder bodyBuilder = new BodyBuilder(getTraversable(),
					connectionPeripheralAccesses, getContainedTasks(),true, getModel().getConfig().isCreatePerformanceEvaluation());
			body = bodyBuilder.generateBody();
			isBodyGenerated = true;
		}
		return body;
	}

	@Override
	public void insertEvaluationAccess(boolean evaluation) {
		this.evaluation = evaluation;

	}

	/**
	 * Returns all {@link Variable}s read by this {@link Task} and it's child
	 * tasks. If any {@link Variable} access is indicated through task pragma
	 * tags by the user, this indicated behavior is used instead of the behavior
	 * detected through µStreams.
	 *
	 * @return All read variables.
	 */
	public Collection<Variable> getCombinedReadVariables() {
		Collection<Variable> result = new LinkedHashSet<Variable>();
		Collection<Variable> readVariables = getReadVariables();
		Collection<Variable> forcedReadVariables = getForcedReadVariables();
		Collection<Variable> forcedWriteVariables = getForcedWriteVariables();

		result.addAll(forcedReadVariables);

		for (Variable variable : readVariables) {
			if (!(forcedWriteVariables.contains(variable) && !forcedReadVariables
					.contains(variable))) {
				result.add(variable);
			}
		}

		return result;

	}

	/**
	 * Returns all {@link Variable}s written by this {@link Task} and it's child
	 * tasks. If any {@link Variable} access is indicated through task pragma
	 * tags by the user, this indicated behavior is used instead of the behavior
	 * detected through µStreams.
	 *
	 * @return All written variables.
	 */
	public Collection<Variable> getCombinedWriteVariables() {
		Collection<Variable> result = new LinkedHashSet<Variable>();
		Collection<Variable> writeVariables = getWriteVariables();
		Collection<Variable> forcedReadVariables = getForcedReadVariables();
		Collection<Variable> forcedWriteVariables = getForcedWriteVariables();

		result.addAll(forcedWriteVariables);

		for (Variable variable : writeVariables) {
			if (!(!forcedWriteVariables.contains(variable) && forcedReadVariables
					.contains(variable))) {
				result.add(variable);
			}
		}

		return result;

	}

	public Collection<TaskDependency> getDependencies() {
		return dependencies;
	}

	/**
	 * Returns a Collection of Variables this Task uses, that are also used by
	 * the given preceding Task.
	 *
	 * @param dependsOn - The preceding Task.
	 * @return all variables this task depends on
	 */
	public Collection<Variable> getDependendVariables(Task dependsOn) {
		Collection<Variable> dependendVariables = new LinkedHashSet<Variable>();

		for (TaskDependency dependency : dependencies) {
			if (dependency.getProvider().equals(dependsOn)) {
				dependendVariables.add(dependency.getVariableOfConsumer());
			}
		}

		return dependendVariables;
	}

	/**
	 * Returns a Collection of Tasks that are depending on this Task. Generally
	 * this means that this Tasks writes one ore more Variables that are read by
	 * one of the returned Tasks.
	 *
	 * @return Collection of Tasks that are depending on this task.
	 */
	public Collection<Task> getConsumers() {
		if (this.isReplica) {
			return originalTask.getProviders();
		} else {
			Collection<Task> consumers = new ArrayList<Task>();
			for (TaskDependency dependency : dependencies) {
				if (dependency.getProvider() == this) {
					consumers.add(dependency.getConsumer());

					if (TaskDependency.DependencyType.ONE_TO_MANY == dependency.getType()) {
						consumers.addAll(dependency.getReplica());
					}

				}
			}
			return consumers;
		}
	}

	/**
	 * Returns a Collection of Tasks, this Task depends on. Generally this means
	 * that this Task reads one or more Variables that are written by one of the
	 * returned Task.
	 *
	 * @return Collection of Tasks, this Task depends on.
	 */
	public Collection<Task> getProviders() {
		if (this.isReplica) {
			return originalTask.getProviders();
		} else {
			Collection<Task> providers = new ArrayList<Task>();
			for (TaskDependency dependency : dependencies) {
				if (dependency.getConsumer() == this) {
					providers.add(dependency.getProvider());

					if (TaskDependency.DependencyType.MANY_TO_ONE == dependency.getType()) {
						providers.addAll(dependency.getReplica());
					}

				}
			}
			return providers;

		}
	}

	public ContainsChildTask getDirectParent() {
		return directParent;
	}

	public void setDirectParent(ContainsChildTask directParent) {
		this.directParent = directParent;
	}

	/**
	 * Returns all read variables enforced by the user
	 *
	 * @return read variables enforced by the user
	 */
	public Collection<Variable> getForcedReadVariables() {
		Collection<Variable> readVariables = new ArrayList<Variable>();

		for (TaskVariable variable : variables) {
			if (variable.isForcedRead()) {
				readVariables.add(variable.getVariable());
			}
		}

		return readVariables;
	}

	/**
	 * Returns all write variables enforced by the user
	 *
	 * @return write variables enforced by the user
	 */
	public Collection<Variable> getForcedWriteVariables() {
		Collection<Variable> writtenVariables = new ArrayList<Variable>();

		for (TaskVariable variable : variables) {
			if (variable.isForcedWritten()) {
				writtenVariables.add(variable.getVariable());
			}
		}

		return writtenVariables;
	}

	/**
	 * Returns all variables that are contained in the function calls of non system library function
	 *
	 * @return a list of variables from function calls
	 */
	public Collection<Variable> getAllFunctionVariables() {
		Collection<Variable> functionvars = new ArrayList<>();
		getAllFunctionCalls().stream()
				.filter(f -> !f.isExtern())
				.forEach(f -> functionvars.addAll(f.getFunction().getVariables()));
		return functionvars;
	}

	public boolean isGlobalVariableUsedInFunction(Variable variable) {
		return getAllFunctionVariables().stream().filter(v -> v.isGlobal() && v.equals(variable)).findAny().isPresent();
	}

	public int getID() {
		return id;
	}

	/**
	 * Returns all include files
	 *
	 * @return the needed include files
	 */
	public Collection<Include> getIncludeFiles() {

		Collection<Include> includes = new LinkedHashSet<Include>();

		for (UFunctionCall call : functionCalls) {
			if (call.isExtern()) {
				/* Add all includes of extern calls of this task */
				includes.addAll(call.getIncludeFiles());
			}
		}

		for (UFunction localFunction : getLocalFunctions()) {
			/*
			 * Also add all includes of extern calls from local functions called
			 * by this task.
			 */
			includes.addAll(localFunction.getIncludeFiles());
		}
		/*Add variables declared in other headers*/
		Collection<Variable> allVariables = variables.stream()
				.filter(v -> !v.isNotUsed())
				.map(v -> v.getVariable())
				.collect(Collectors.toList());
		allVariables.addAll(getAllFunctionVariables());
		final List<Variable> globalVariables = getSortedGlobalVariables();
		allVariables.stream()
				.filter(v -> !globalVariables.contains(v))
				.forEach(v -> {
					Include include = v.getInclude();
					if (include != null) {
						includes.add(include);
					}
				});

		return includes;
	}

	@Override
	public InnerFlowGraph getInnerFlowGraph() {
		return graph;
	}

	public void setInnerFlowGraph(InnerFlowGraph graph) {
		this.graph = graph;

	}

	/**
	 * Returns a collection of all {@link UFunction}s that are needed by this
	 * tasks an should be included in the tasks output file and/or can not be
	 * included through c header file includes.
	 *
	 * @return A Collection of all {@link UFunction}s that are stored in the same file as the Task.
	 */
	public Collection<UFunction> getLocalFunctions() {
		Collection<UFunction> localFunctions = new LinkedHashSet<UFunction>();

		for (UFunctionCall call : functionCalls) {
			if (!call.isExtern()) {
				getLocalFunctionsRecursion(call.getFunction(), localFunctions);
			}
		}
		return localFunctions;
	}

	private void getLocalFunctionsRecursion(UFunction visit,
											Collection<UFunction> visited) {
		visited.add(visit);
		for (UFunctionCall call : visit.getFunctionCalls()) {
			if (!call.isExtern() && !visited.contains(call.getFunction())) {
				getLocalFunctionsRecursion(call.getFunction(), visited);
			}
		}
	}

	public Task getParentTask() {
		return parentTask;
	}

	public void setParentTask(Task parentTask) {
		this.parentTask = parentTask;
	}

	@Override
	public Model getModel() {
		return this.model;
	}

	@Override
	public void setModel(Model model) {
		this.model = model;

	}

	/**
	 * Returns all read variables
	 *
	 * @return read variables
	 */
	public Collection<Variable> getReadVariables() {
		Collection<Variable> readVariables = new ArrayList<Variable>();

		for (TaskVariable variable : variables) {
			if (variable.isRead()) {
				readVariables.add(variable.getVariable());
			}
		}

		return readVariables;
	}

	/**
	 * Returns all used variables that should be used locally
	 * - excludes all variable not used
	 * - exludes globalshared variables (those go into global mem)
	 * - excludes variables declared in a include file
	 * - excludes variables received via DMA peripherals
	 * @return used variables
	 */
	public List<Variable> getUsedGlobalVariables() {
		final List<Variable> ret = this.variables.stream()
					.filter(taskVariable -> taskVariable.isUsed())
					.map(var -> var.getVariable())
					.filter(var -> var.isGlobalWritten() || ((!getModel().getConfig().useGlobalMemory() || !var.isGlobalShared()) && var.isGlobal()))
					.filter(var -> !(var.getVariableDeclarator().getDeclaration() instanceof Enumeration))
					.filter(var -> !var.isDeclaredInExternalInclude())
					.collect(Collectors.toList());
		ret.removeAll(getVariablesSendReceivedViaDMAPeripherals());
		return ret.stream().sorted().collect(Collectors.toList());
	}

	/**
	 * Returns all used variables that should be used locally
	 * - excludes all variable not used
	 * - exludes globalshared variables (those go into global mem)
	 * - excludes variables declared in a include file
	 * - excludes variables received via DMA peripherals
	 * @return used variables
	 */
	public List<Variable> getUsedLocalVariables() {
		Predicate<TaskVariable> declaredInMain =  tv -> {
			final Procedure procedure = CetusUtils.findParentOfType(tv.getVariable().getVariableDeclarator(), Procedure.class);
			if (procedure != null) {
				return procedure.getDeclaration().getDeclaredIDs().stream().anyMatch(id -> id.toString().equalsIgnoreCase(Constants.C.MAIN_FUNCTION_NAME));
			}
			return false;
		};
		Collection<TaskVariable> vars = this.variables;
		Task task = this;
		if (isReplica) {
			vars = originalTask.variables;
			task = originalTask;
		}
		Task finalTask = task;
		final List<Variable> ret = vars.stream()
				.filter(taskVariable -> taskVariable.isUsed())
				.filter(declaredInMain)
				.map(var -> var.getVariable())
				.filter(var -> !var.isGlobal())
				.filter(var -> !var.isDeclaredInExternalInclude())
				.filter(var -> !CetusUtils.contains(var.getVariableDeclarator(), finalTask, model) || var.getVariableDeclarator().getInitializer() == null)
				.collect(Collectors.toList());
		ret.removeAll(getVariablesSendReceivedViaDMAPeripherals());
		ret.removeAll(getUsedVariablesFromPredecessors());
		return ret.stream().sorted().collect(Collectors.toList());
	}

	/**
	 * Returns all used variables
	 *
	 * @return used variables
	 */
	public List<Variable> getUsedVariables() {
		List<Variable> usedVariables = new ArrayList<>();

		for (TaskVariable variable : variables) {
			if (!variable.isNotUsed()) {
				usedVariables.add(variable.getVariable());
			}
		}

		return usedVariables.stream().sorted().collect(Collectors.toList());
	}

	/**
	 * Returns all used variables that don't have an initializer and thus can be shifted to globals
	 *
	 * @return used variables without declaration
	 */
	public List<Variable> getUsedVariablesWithoutInitAtDeclaration() {
		List<Variable> vars = getUsedVariables().stream()
					.filter(var -> var.getInitializer() == null)
					.filter(var -> !var.isDeclaredInExternalInclude())
					.sorted()
					.collect(Collectors.toList());
		return vars;
	}

	/**
	 * Returns all used variables that are not alredy returned from getUsedVariablesWithoutInitAtDeclaration()
	 *
	 * @return used variables
	 */
	public List<Variable> getUsedVariablesWithInitializer() {
		Collection<Variable> vars = getUsedVariables();
		vars.removeAll(getUsedVariablesWithoutInitAtDeclaration());
		return vars.stream().sorted().collect(Collectors.toList());
	}

	/**
	 * Variables provided by predecessors
	 * @return all variables that are received through communication infrastructure
	 */
	public List<Variable> getVariablesProvidedByPredecessors() {
		Set<Variable> deps = new LinkedHashSet<>();
		getDependencies().stream()
					.filter(td -> (td.getConsumer().equals(this) || (this.isReplica() && td.getConsumer().equals(getOriginal()))))
					.forEach(td -> deps.addAll(getDependendVariables(td.getProvider())));
		return deps.stream().sorted().collect(Collectors.toList());
	}

	/**
	 * Variables for globals without initialization
	 * @return combines all variables that don't have to be initialized and can be provided as globals
	 */
	public List<Variable> getUsedVariablesFromPredecessors() {
		Collection<Variable> vars = new LinkedHashSet<>();
		//prune DMA send/rec variables
		final List<Variable> dmaVariables = getVariablesSendReceivedViaDMAPeripherals();
		final Collection<Variable> variablesProvidedByPredecessors = getVariablesProvidedByPredecessors();

		vars.addAll(variablesProvidedByPredecessors);
		vars.removeAll(dmaVariables);
		final List<Variable> ret = vars.stream()
				.filter(v -> !v.isGlobal())
				.sorted()
				.collect(Collectors.toList());
		return ret;
	}

	public Set<ConnectionPeripheralAccess> getConnectionPeripheralAccesses() {
		return new HashSet<>(connectionPeripheralAccesses);
	}

	/**
	 * Get all variables that are received via a DMA peripheral in this task
	 * @return all received variables via dma
	 */
	public List<Variable> getVariablesReceivedViaDMAPeripherals() {
		return connectionPeripheralAccesses.stream()
					.filter(c -> c.getPeripheral() instanceof ConnectionPeripheralDMA)
					.filter(c -> c.getDependency().getConsumer().equals(this))
					.map(c -> c.getDependency().getVariableOfConsumer())
					.distinct()
					.collect(Collectors.toList());
	}

	/**
	 * Get all variables that are send via a DMA peripheral in this task
	 * @return all send variables via dma
	 */
	public List<Variable> getVariablesSendViaDMAPeripherals() {
		return connectionPeripheralAccesses.stream()
				.filter(c -> c.getPeripheral() instanceof ConnectionPeripheralDMA)
				.filter(c -> c.getDependency().getProvider().equals(this))
				.map(c -> c.getDependency().getVariableOfProvider())
				.distinct()
				.collect(Collectors.toList());
	}

	/**
	 * Get all variables that are send/received via a DMA peripheral in this task
	 * @return all send variables via dma
	 */
	public List<Variable> getVariablesSendReceivedViaDMAPeripherals() {
		Set<Variable> ret = getVariablesSendViaDMAPeripherals().stream().collect(Collectors.toSet());
		ret.addAll(getVariablesReceivedViaDMAPeripherals());
		return ret.stream().sorted().collect(Collectors.toList());
	}

	/**
	 * This should just be used for FTL
	 * @return x
	 */
	@Deprecated
	public List<String> getUsedVariablesFromDMATransfersProvidedByPredecessorsAsDefinition() {
		final Set<Variable> variablesProvidedByPredecessors = new HashSet<>(getVariablesProvidedByPredecessors());
		variablesProvidedByPredecessors.retainAll(getVariablesReceivedViaDMAPeripherals());
		final List<Variable> variablesProvidedByPredecessors_clones = variablesProvidedByPredecessors.stream().map(v -> new Variable(v)).collect(Collectors.toList());
		return variablesProvidedByPredecessors_clones.stream()
				.map(v -> v.getDeclaration() + " = " + findInitializerForVariableDMATransfer(v) + ";")
				.sorted()
				.collect(Collectors.toList());
	}

	private Identifier findInitializerForVariableDMATransfer(Variable variable) {
		final Optional<IConnectionPeripheralDMA> provider = getDMAPeripheralsProviding(variable);
		final Optional<IConnectionPeripheralDMA> consumer = getDMAPeripheralsConsuming(variable);
		if (provider.isPresent()) {
			return provider.get().getIdentifierFor(variable);
		} else if (consumer.isPresent()) {
			return consumer.get().getIdentifierFor(variable);
		}
		return null;
	}

	public Optional<IConnectionPeripheralDMA> getDMAPeripheralsConsuming(Variable variable) {
		return getDMAPeripheralsWithFilter(variable, SendPeripheral.class);
	}

	public Optional<IConnectionPeripheralDMA> getDMAPeripheralsProviding(Variable variable) {
		return getDMAPeripheralsWithFilter(variable, ReceivePeripheral.class);
	}

	private Optional<IConnectionPeripheralDMA> getDMAPeripheralsWithFilter(Variable variable, Class clazz) {
		return getDMAPeripheralsTransferring(variable).stream()
				.filter(p -> clazz.isAssignableFrom(p.getClass()))
				.collect(StreamUtil.atMostOne());
	}

	private List<IConnectionPeripheralDMA> getDMAPeripheralsTransferring(Variable variable) {
		return getCore().getPeripherals(IConnectionPeripheralDMA.class).stream()
				.filter(p -> p.getVariables().contains(variable))
				.collect(Collectors.toList());
	}

	public TaskPragma getTaskPragma() {
		return pragma;
	}

	public EndPragma getEndPragma() {
		return endPragma;
	}

	public boolean hasEndPragma() {
		return endPragma != null;
	}

	private TaskVariable getTaskVariable(Variable variable) {
		for (TaskVariable taskVariable : variables) {
			if (taskVariable.getVariable().equals(variable)) {
				return taskVariable;
			}
		}

		return null;
	}

	@Override
	public Traversable getTraversable() {
		return traversable;
	}

	@Override
	public void setTraversable(Traversable traversable) {
		this.traversable = traversable;

	}

	/**
	 * Returns all variable accesses
	 *
	 * @return variable accesses
	 */
	public Collection<VariableAccess> getVariableAccesses() {
		Collection<VariableAccess> result = new LinkedHashSet<VariableAccess>();

		for (TaskVariable taskVariable : variables) {
			result.addAll(taskVariable.getVariableAccess());
		}
		return result;
	}

	/**
	 * Returns all written variables
	 *
	 * @return written variables
	 */
	public Collection<Variable> getWriteVariables() {
		Collection<Variable> writtenVariables = new ArrayList<Variable>();

		for (TaskVariable variable : variables) {
			if (variable.isWritten()) {
				writtenVariables.add(variable.getVariable());
			}
		}

		return writtenVariables;
	}

	@Override
	public int hashCode() {
		/*
		 * equals(...) and hashCode(...) are limited to id, pragma and
		 * traversable in order to avoid endless equals-call-loops among
		 * different tasks.
		 */
		final int prime = 31;
		int result = 1;
		result = prime * result + id;
		result = prime * result + ((pragma == null)? 0 : pragma.hashCode());
		result = prime * result
				+ ((traversable == null)? 0 : traversable.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		/*
		 * equals(...) and hashCode(...) are limited to id, pragma and
		 * traversable in order to avoid endless equals-call-loops among
		 * different tasks.
		 */
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		Task other = (Task) obj;
		if (id != other.id) {
			return false;
		}
		if (pragma == null) {
			if (other.pragma != null) {
				return false;
			}
		} else if (!pragma.equals(other.pragma)) {
			return false;
		}
		if (traversable == null) {
			if (other.traversable != null) {
				return false;
			}
		} else if (!traversable.equals(other.traversable)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "Task#" + id + (isReplica()? "(Replica of " + getOriginal() + ")" : "");
	}

	public boolean isBaseTask() {
		return isBaseTask;
	}

	/**
	 * Returns whether this Task depends on any other Task or not.
	 *
	 * @return Returns whether this Task depends on any other Task or not.
	 */
	public boolean isIndependent() {
		for (TaskDependency dependency : dependencies) {
			if (!dependency.getProvider().equals(this)) {
				return false;
			}
		}
		return true;
	}

	public boolean removeDependency(TaskDependency dependency) {
		return dependencies.remove(dependency);
	}

	public Task replicate() {
		return replicate(null);
	}

	/**
	 * Clones the task for the {@link Replicate} worker
	 *
	 * @param parentTask the parent task
	 * @return the task
	 */
	private Task replicate(Task parentTask) {

		final Traversable traversableClone = CetusUtils.deepClone((CompoundStatement) this.traversable);
		traversableClone.setParent(traversable.getParent());
		Task replica = model.getTaskFactory().createTask(traversableClone, this.pragma, model);

		if (null == parentTask) {
			replica.parentTask = this.parentTask;
			replica.directParent = this.directParent;
		} else {
			replica.parentTask = parentTask;
			if (this.directParent == this.parentTask) {
				replica.directParent = parentTask;
			} else {
				replica.directParent = this.directParent;
			}
		}

		replica.childTasks = new LinkedHashSet<Task>();
		for (Task child : childTasks) {
			Task childClone = child.replicate(replica);
			replica.childTasks.add(childClone);
		}

		replica.isReplica = true;
		replica.originalTask = this;

		if (null != this.functionCalls) {
			replica.functionCalls = new LinkedHashSet<UFunctionCall>(
					this.functionCalls);
		} else {
			replica.functionCalls = new LinkedHashSet<UFunctionCall>();
		}

		if (null != this.variables) {
			replica.variables = new LinkedHashSet<TaskVariable>(this.variables);
		} else {
			replica.variables = new LinkedHashSet<TaskVariable>();
		}

		replica.directParent = this.directParent;
		replica.graph = this.graph;
		replica.isBaseTask = this.isBaseTask;
		replica.isBodyGenerated = this.isBodyGenerated;
		replica.model = this.model;
		replica.traversable = traversableClone;

		replica.headerFiles = new LinkedHashSet<>(this.headerFiles);
		replica.sourceFiles = new LinkedHashSet<>(this.sourceFiles);

		return replica;

	}

	public boolean isReplica() {
		return isReplica;
	}

	/**
	 * Is the task is replicated
	 *
	 * @return true if the task shall be replicated, false otherwise
	 */
	public boolean hasReplica() {
		if (this.isReplica) {
			return true;
		}

		for (Task task : getModel().getModelComponents(Task.class)) {
			if (task.isReplica && task.originalTask == this) {
				return true;
			}
		}

		return false;
	}

	/**
	 * The amount of replications for this task
	 *
	 * @return amount of replications
	 */
	public int countReplica() {
		int count = 0;

		for (Task task : getModel().getModelComponents(Task.class)) {
			if (this.isReplica) {
				if (this != task
						&& ((this.originalTask == task.originalTask) || (this.originalTask == task))) {
					count++;
				}

			} else if (this == task.originalTask) {
				count++;
			}
		}

		return count + 1;
	}

	public Task getOriginal() {
		return !this.isReplica? this : originalTask;
	}

	/**
	 * Collects all replicas of this task
	 *
	 * @return all replicated tasks
	 */
	public Collection<Task> getReplica() {
		Collection<Task> replica = new ArrayList<Task>();

		for (Task task : getModel().getModelComponents(Task.class)) {
			if (this.isReplica) {
				if (this.originalTask == task.originalTask
						|| this.originalTask == task) {
					replica.add(task);
				}
			} else if (task.originalTask == this) {
				replica.add(task);
			}
		}

		return replica;
	}

	public Core getCore() {
		return core;
	}

	public void setCore(Core core) {
		this.core = core;
	}

}
