/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.task;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.interfaces.ContainsChildTask;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Optional;

/**
 * A {@link TaskDependency} indicates that one tasks
 * {@link TaskDependency#consumer} uses a variable
 * {@link TaskDependency#variableOfConsumer} that is written by a predecessor task
 * {@link TaskDependency#provider}.
 *
 * @author Jan Weber
 */
public class TaskDependency implements ModelComponent,
	Comparable<TaskDependency> {

	private Logger log = LoggerFactory.getLogger(this.getClass());
	/**
	 * Task that uses a variable, written by another Task before.
	 */
	private Task consumer;
	/**
	 * Task that writes these variable.
	 */
	private Task provider;
	/**
	 * Replica for MANY_TO_ONE or ONE_TO_MANY DependencyType.
	 */

	private Collection<Task> replica;
	/**
	 * The Variable that is written.
	 */
	private Variable variableOfConsumer;

	/**
	 * The Variable that is written.
	 */
	private Variable variableOfProvider;

	/**
	 * The relation of the dependency describes the relationship between the
	 * {@link Task} consumer and the {@link Task} provider:
	 * <ul>
	 * <li>
	 * ONE_TO_ONE</li>
	 * <li>
	 * ONE_TO_MANY</li>
	 * <li>
	 * MANY_TO_ONE</li>
	 * </ul>
	 */
	private DependencyRelation relation;
	/**
	 * The type of the dependency describes if there are replica of one of the
	 * tasks
	 * <ul>
	 * <li>
	 * PARENT_CHILD</li>
	 * <li>
	 * CHILD_PARENT</li>
	 * <li>
	 * CHILD_CHILD</li>
	 * </ul>
	 */
	private DependencyType type;
	/**
	 * The location indicates the code position within the parent task at which
	 * the shared variable have to be transfered to the child task. This field
	 * is not used if the {@link TaskDependency}s type is set to
	 * {@link DependencyRelation#CHILD_CHILD}, due to the transfer takes place
	 * at the beginning respectively end of the corresponding tasks.
	 */
	private ContainsChildTask location;
	private Model model;

	private boolean forced = false;

	/**
	 * Constructor expecting all fields as arguments.
	 *
	 * @param consumer - Task that uses a variable, written by another Task before.
	 * @param provider - Task that writes these variable.
	 * @param variable - The Variable that is written.
	 * @param relation - the type of relationship
	 * @param location - the location in the parent task, if neccessary.
	 */
	public TaskDependency(Task consumer, Task provider, Variable variable,
						  DependencyRelation relation, ContainsChildTask location) {
		this(consumer, provider, variable, Optional.empty(), relation ,location);
	}

	/**
	 * Constructor expecting all fields as arguments.
	 *
	 * @param consumer - Task that uses a variable, written by another Task before.
	 * @param provider - Task that writes these variable.
	 * @param variableOfProvider - The Variable that is written.
	 * @param variableOfConsumer - The Variable that is written at the consumer if empty then provider is taken
	 * @param relation - the type of relationship
	 * @param location - the location in the parent task, if neccessary.
	 */
	public TaskDependency(Task consumer, Task provider, Variable variableOfProvider, Optional<Variable> variableOfConsumer,
						  DependencyRelation relation, ContainsChildTask location) {

		if (consumer.hasReplica() && provider.hasReplica()) {
			throw new IllegalArgumentException("consumer and provider have both replica. This TaskDependency cannot model this behavior.");

		}

		this.consumer = consumer.getOriginal();
		this.provider = provider.getOriginal();
		if (variableOfConsumer.isPresent()) {
			this.variableOfConsumer = variableOfConsumer.get();
		} else {
			this.variableOfConsumer = variableOfProvider;
		}
		this.variableOfProvider = variableOfProvider;
		this.relation = relation;
		this.location = location;

		if (consumer.hasReplica()) {

			replica = new ArrayList<Task>(this.consumer.getReplica());

			this.type = DependencyType.ONE_TO_MANY;

		} else if (provider.hasReplica()) {

			replica = new ArrayList<Task>(this.provider.getReplica());

			this.type = DependencyType.MANY_TO_ONE;

		} else {

			this.type = DependencyType.ONE_TO_ONE;

		}

	}

	/**
	 * Returns the other {@link Task} {@link TaskDependency#provider} the
	 * {@link Task} {@link TaskDependency#consumer} depends on.
	 *
	 * @return Returns the other {@link Task} {@link TaskDependency#provider} the {@link Task} {@link TaskDependency#consumer} depends on.
	 */
	public Task getProvider() {
		return provider;
	}

	public Collection<Task> getReplica() {
		return replica;
	}

	public ContainsChildTask getLocation() {
		return location;
	}

	/**
	 * Returns the {@link Task} {@link TaskDependency#consumer} that depends on
	 * another {@link Task} {@link TaskDependency#provider}.
	 *
	 * @return Returns the {@link Task} {@link TaskDependency#consumer} that depends on another {@link Task} {@link TaskDependency#provider}.
	 */
	public Task getConsumer() {
		return consumer;
	}

	public DependencyRelation getRelation() {
		return relation;
	}

	/**
	 * Returns true if the given {@link Task} task is the parent task in this
	 * dependency otherwise false.
	 *
	 * @param task a task to be checked
	 * @return Returns true if the given {@link Task} task is the parent task is this dependency otherwise false.
	 */
	public boolean isParent(Task task) {
		return (((DependencyRelation.PARENT_CHILD == relation) && this.provider
			.equals(task)) || DependencyRelation.CHILD_PARENT == relation
			&& this.consumer.equals(task));
	}

	/**
	 * Returns true if the given {@link Task} task is the or one of the child
	 * task in this dependency otherwise false.
	 *
	 * @param task a task to be checked
	 * @return Returns true if the given {@link Task} task is the parent task in this dependency otherwise false.
	 */
	public boolean isChildren(Task task) {
		return ((DependencyRelation.CHILD_PARENT == relation && this.provider
			.equals(task))
			|| (DependencyRelation.PARENT_CHILD == relation && this.consumer
			.equals(task)) || (DependencyRelation.CHILD_CHILD == relation && (this.consumer
			.equals(task) || this.provider.equals(task))));
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
			+ ((consumer == null)? 0 : consumer.hashCode());
		result = prime * result
			+ ((location == null)? 0 : location.hashCode());
		result = prime * result
			+ ((provider == null)? 0 : provider.hashCode());
		result = prime * result
			+ ((relation == null)? 0 : relation.hashCode());
		result = prime * result + ((replica == null)? 0 : replica.hashCode());
		result = prime * result + ((type == null)? 0 : type.hashCode());
		result = prime * result
			+ ((variableOfProvider == null)? 0 : variableOfProvider.hashCode());
		result = prime * result
				+ ((variableOfConsumer == null)? 0 : variableOfConsumer.hashCode());
		result = 2 * result + (forced ? 0 : 1);
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		TaskDependency other = (TaskDependency) obj;
		if (consumer == null) {
			if (other.consumer != null) {
				return false;
			}
		} else if (!consumer.equals(other.consumer)) {
			return false;
		}
		if (location == null) {
			if (other.location != null) {
				return false;
			}
		} else if (!location.equals(other.location)) {
			return false;
		}
		if (provider == null) {
			if (other.provider != null) {
				return false;
			}
		} else if (!provider.equals(other.provider)) {
			return false;
		}
		if (relation != other.relation) {
			return false;
		}
		if (forced != other.forced) {
			return false;
		}
		if (replica == null) {
			if (other.replica != null) {
				return false;
			}
		} else if (!replica.equals(other.replica)) {
			return false;
		}
		if (type != other.type) {
			return false;
		}
		if (variableOfConsumer == null) {
			if (other.variableOfConsumer != null) {
				return false;
			}
		} else if (!variableOfConsumer.equals(other.variableOfConsumer)) {
			return false;
		}
		if (variableOfProvider == null) {
			if (other.variableOfProvider != null) {
				return false;
			}
		} else if (!variableOfProvider.equals(other.variableOfProvider)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		final String child = "CHILD";
		final String parent = "PARENT";
		String consumerType = null;
		String providerType = null;
		switch (relation) {
			case CHILD_CHILD:
				consumerType = child;
				providerType = child;
				break;
			case PARENT_CHILD:
				consumerType = child;
				providerType = parent;
				break;
			case CHILD_PARENT:
				consumerType = parent;
				providerType = child;
				break;
			default:
				break;
		}
		//CHECKSTYLE.OFF: MultipleStringLiterals
		switch (type) {
			case ONE_TO_ONE:
				return "TaskDependency = {" + provider + "[" + providerType
					+ "] -> " + consumer + "[" + consumerType + "]: "
					+ variableOfProvider + "->" + variableOfConsumer + " }";

			case ONE_TO_MANY:
				return "TaskDependency = {" + provider + "[" + providerType
					+ "] -> " + consumer + " | " + replica + "[" + consumerType
					+ "]: " + variableOfProvider + "->" + variableOfConsumer + " }";

			case MANY_TO_ONE:
				return "TaskDependency = {" + provider + " | " + replica + "["
					+ providerType + "] -> " + consumer + "[" + consumerType
					+ "]: " + variableOfProvider + "->" + variableOfConsumer + " }";
			default:
				break;
		}
		return "TaskDependency = {ERROR: No DependencyType set!}";
		//CHECKSTYLE.ON: MultipleStringLiterals
	}

	@Override
	public Model getModel() {
		return this.model;
	}

	@Override
	public void setModel(Model project) {
		this.model = project;

	}

	public DependencyType getType() {
		return type;
	}

	@Override
	public int compareTo(TaskDependency other) {
		if (other == this) {
			return 0;
		}

		if (this.consumer.getID() != other.getConsumer().getID()) {
			return this.consumer.getID() - other.getConsumer().getID();
		} else if (this.provider.getID() != other.getProvider().getID()) {
			return this.provider.getID() - other.getProvider().getID();
		} else {
			return variableOfProvider.getName().compareTo(other.variableOfProvider.getName());
		}
	}

	/**
	 * Returns true if the given {@link Task} task has the same target.
	 *
	 * @param other A task to be checked
	 * @return Returns true if the given {@link Task} task has the same target.
	 */
	public boolean isSameTarget(TaskDependency other) {
		if (other == this) {
			return true;
		}

		if (this.consumer.getID() != other.getConsumer().getID()) {
			return false;
		} else if (this.provider.getID() != other.getProvider().getID()) {
			return false;
		}

		return true;
	}

	public void setForced(boolean forced) {
		this.forced = forced;
	}

	public boolean isForced() {
		return forced;
	}
	/**
	 * A {@link DependencyRelation} describes the relationship between two
	 * tasks. The relation can be
	 * <ul>
	 * <li>
	 * PARENT_CHILD - The parent sends the variable to a child task.</li>
	 * <li>
	 * CHILD_PARENT - The child task sends the variable back to the parent task.
	 * </li>
	 * <li>
	 * CHILD_CHILD - The tasks are both childtask of the same parent task and
	 * the variable is sent directly from one task to the other due
	 * optimization.</li>
	 * </ul>
	 *
	 * @author Jan Weber
	 */
	public enum DependencyRelation {
		PARENT_CHILD, CHILD_PARENT, CHILD_CHILD
	}


	/**
	 * The dependency relation to one or multiple components
	 */
	public enum DependencyType {
		ONE_TO_ONE, ONE_TO_MANY, MANY_TO_ONE
	}

	public Variable getVariableOfConsumer() {
		return variableOfConsumer;
	}

	public Variable getVariableOfProvider() {
		return variableOfProvider;
	}

}
