/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.task;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.other.OtherModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.EndPragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.TaskPragma;

import cetus.hir.Traversable;

/**
 * This {@link TaskFactory} is the factory class for {@link Task Tasks}. It
 * ensures that every generated task gets an unique id and only one base
 * task is generated.
 *
 * @author Jan Weber
 */
public class TaskFactory implements OtherModelComponent {

	/*
	 * Counts the created cores in order to assign a unique id to each core.
	 */
	private int idCounter;

	/*
	 * Ensures that only one base task is generated.
	 */
	private boolean isBaseTaskGenerated;

	private Model model;

	public TaskFactory() {
		idCounter = 0;
		isBaseTaskGenerated = false;
	}

	/**
	 * Generates the base task with the given core (containing the programms
	 * main procedure) and a Unique ID that is neccessary for communication
	 * between Tasks. Note that the base task can only generated once.
	 *
	 * @param cetusTaskCore traversable
	 * @param model a model
	 * @return The generated Task.
	 */
	public Task createTask(Traversable cetusTaskCore, Model model) {
		if (isBaseTaskGenerated) {
			throw new UnsupportedOperationException(
					"Base task already generated. Cannot be generated twice.");
		}
		isBaseTaskGenerated = true;

		return new Task(getTaskID(), cetusTaskCore, model);
	}

	/**
	 * Generates a task with the given core and pragma and a Unique ID that
	 * is neccessary for communication between Tasks.
	 *
	 * @param cetusTaskCore the traversable
	 * @param taskPragma an according task pragma
	 * @param model the model
	 * @return The generated Task.
	 */
	public Task createTask(Traversable cetusTaskCore, TaskPragma taskPragma, Model model) {
		return createTask(cetusTaskCore, taskPragma, null, model);
	}

	/**
	 * Generates a task with the given core and pragma and a Unique ID that
	 * is neccessary for communication between Tasks.
	 *
	 * @param cetusTaskCore the traversable
	 * @param taskPragma an according task pragma
	 * @param model th model
	 * @param endPragma if existing, en end pragma otherwise null
	 * @return The generated Task.
	 */
	public Task createTask(Traversable cetusTaskCore, TaskPragma taskPragma, EndPragma endPragma, Model model) {
		return new Task(getTaskID(), cetusTaskCore, taskPragma, endPragma, model);
	}

	/**
	 * Returns the next unique identifier. The ids are counted up
	 * internally.
	 *
	 * @return Next unique identifier.
	 */
	private synchronized int getTaskID() {
		return (idCounter++);
	}

	@Override
	public Model getModel() {
		return model;
	}

	@Override
	public void setModel(Model m) {
		model = m;
	}
}