/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.task.pragma;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.config.LoopOptimizerConfiguration;
import de.tu_darmstadt.rs.microStreams.model.Model;

import cetus.hir.PragmaAnnotation;
import cetus.hir.Traversable;
import org.slf4j.Logger;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Parses the pragma and checks for correctness of the annotation
 */
public final class PragmaLoopParser {
	private static Logger log = org.slf4j.LoggerFactory.getLogger(PragmaLoopParser.class);

	private final Model model;

	public PragmaLoopParser(Model model) {
		this.model = model;
	}

	/**
	 * Creates a {@link TaskPragma}-Object with the the given
	 * {@link PragmaAnnotation}.
	 *
	 * @param pa - the {@link PragmaAnnotation} that should be parsed.
	 * @param parentTraversable the parent traversable
	 * @return JSPragmaAnnotation
	 */
	@SuppressWarnings("checkstyle:magicNumber")
	public LoopPragma parse(PragmaAnnotation pa, Traversable parentTraversable) {
		final String bracketSplitRegex = "[\\(, \\)]";
		if (pa != null && parentTraversable != null) {
			validatePragma(pa);
			String pragma = pa.getName().toLowerCase();
			String[] pragmas = splitPragma(pragma);

			Collection<PragmaTag> tags = new ArrayList<PragmaTag>();
			if (!pragmas[2].matches(Constants.PragmaLoop.REGEX_MAX_CYCLES)) {
				PragmaTag<Integer> num = new PragmaTag<Integer>(Constants.PragmaLoop.NUMBER_OF_SPLITS);
				num.addEntry(Integer.valueOf(pragmas[2]));
				tags.add(num);
			} else {
				String[] parts = pragmas[2].split(bracketSplitRegex);
				PragmaTag<Integer> tag = new PragmaTag<>(Constants.PragmaLoop.MAX_CYCLES);
				tag.addEntry(Integer.valueOf(parts[1]));
				tags.add(tag);
			}

			if (pragmas.length >= 4) {
				if (Arrays.stream(pragmas).anyMatch(s -> s.equalsIgnoreCase(Constants.PragmaLoop.AUTOPERF))) {
					tags.add(new PragmaTag<String>(Constants.PragmaLoop.AUTOPERF));
				}
				if (Arrays.stream(pragmas).anyMatch(s -> s.equalsIgnoreCase(Constants.PragmaLoop.USE_FISSION))) {
					tags.add(new PragmaTag<String>(Constants.PragmaLoop.USE_FISSION));
				}
				if (Arrays.stream(pragmas).anyMatch(s -> s.equalsIgnoreCase(Constants.PragmaLoop.USE_SPLITTING))) {
					tags.add(new PragmaTag<String>(Constants.PragmaLoop.USE_SPLITTING));
				}
				if (Arrays.stream(pragmas).anyMatch(s -> s.matches(Constants.PragmaLoop.REGEX_IV))) {
					String iv = String.valueOf(Arrays.stream(pragmas).filter(s -> s.matches(Constants.PragmaLoop.REGEX_IV)).toArray()[0]);
					String[] parts = iv.split(bracketSplitRegex);
					PragmaTag<String> tag = new PragmaTag<String>(Constants.PragmaLoop.ITERATION_VARIABLE);
					tag.addEntry(parts[1]);
					tags.add(tag);
				}
				if (Arrays.stream(pragmas).anyMatch(s -> s.matches(Constants.PragmaLoop.REGEX_DISTRIBUTION))) {
					//TODO: allow spaces between distribution values (e.g. distribution(30, 70))
					if (!(Arrays.stream(pragmas).anyMatch(s -> s.matches("^" + Constants.PragmaLoop.USE_FISSION + "$")))) {
						for (String s : pragmas) {
							if (s.matches(Constants.PragmaLoop.REGEX_DISTRIBUTION)) {
								PragmaTag<Integer> tag = new PragmaTag<Integer>(Constants.PragmaLoop.DISTRIBUTION);
								String[] splits = s.split("[\\( , \\)]");
								if (splits.length != Integer.valueOf(pragmas[2]) + 1) {
									throw new RuntimeException("The Count of Numbers in the distribution is not equal to the Number of splits. "
										+ (splits.length - 1) + "Numbers were found in the distribution, but " + pragmas[2] +
										"splits were found. The full pragma is " + pa.toString() + ".");
								}
								Arrays.stream(splits).skip(1).forEach(string -> tag.addEntry(Integer.valueOf(string)));
								tags.add(tag);
							}
						}
					}
				}
			}
			return new LoopPragma(pa, tags, parentTraversable);
			//CHECKSTYLE.ON: MagicNumber
		}
		return null;
	}
	@SuppressWarnings({"checkstyle:magicNumber", "checkstyle:MultipleStringLiterals"})
	private void validatePragma(PragmaAnnotation pragmaAnnotation) {
		String[] pragmas = splitPragma(pragmaAnnotation.getName().toLowerCase());
		if (pragmas.length > 2) {
			if (pragmas[0].equalsIgnoreCase(Constants.Pragma.IDENTIFIER) && pragmas[1].equalsIgnoreCase(Constants.PragmaLoop.LOOPOPT)) {
				if (pragmas[2].matches(Constants.PragmaLoop.REGEX_VALID_NUM) || pragmas[2].matches(Constants.PragmaLoop.REGEX_MAX_CYCLES)) {
					if (pragmas.length > 3) {
						List<String> leftovers = Arrays.stream(pragmas).skip(3).collect(Collectors.toList());
						leftovers.stream().filter(s -> !s.equalsIgnoreCase(Constants.PragmaLoop.USE_FISSION))
								.filter(s -> !s.equalsIgnoreCase(Constants.PragmaLoop.USE_SPLITTING))
								.filter(s -> !s.matches(Constants.PragmaLoop.REGEX_DISTRIBUTION))
								.filter(s -> !s.matches(Constants.PragmaLoop.REGEX_IV))
								.filter(s -> !s.matches("^" + Constants.PragmaLoop.AUTOPERF + "$"))
								.forEach(s -> model.printIgnoreableError(log,"The Pragma " + pragmaAnnotation.getName() + " could not be parsed. The sequence " + s + " is not known."));
						if (leftovers.stream().anyMatch(s -> s.matches("^" + Constants.PragmaLoop.USE_FISSION + "$"))
								&& leftovers.stream().anyMatch(s -> s.matches("^" + Constants.PragmaLoop.USE_SPLITTING + "$"))) {
							model.printIgnoreableError(log,"The Pragma " + pragmaAnnotation.getName() + " could not be parsed. "
									+ Constants.PragmaLoop.USE_SPLITTING + " and " + Constants.PragmaLoop.USE_FISSION +
									" can not be used in one pragma statement");
						}
						if (pragmas[2].matches(Constants.PragmaLoop.REGEX_MAX_CYCLES)
								&& leftovers.stream().anyMatch(s -> s.matches(Constants.PragmaLoop.REGEX_DISTRIBUTION))) {
							model.printIgnoreableError(log,"The Pragma " + pragmaAnnotation.getName() + " could not be parsed. "
									+ Constants.PragmaLoop.MAX_CYCLES + " and " + Constants.PragmaLoop.DISTRIBUTION
									+ " should not be used in the same Pragma, this is not supported");
						}
						if (pragmas[2].matches(Constants.PragmaLoop.REGEX_MAX_CYCLES)
								&& model.getModelComponent(LoopOptimizerConfiguration.class).getApplicationProfile().equals("")) {
							model.printIgnoreableError(log,"The Pragma " + pragmaAnnotation.getName() + " could not be parsed. "
									+ Constants.PragmaLoop.MAX_CYCLES + " can only be used together with autoPerf data. " +
									"Please set the profile option when calling loopOpt");
						}
						if (leftovers.stream().anyMatch(s -> s.matches(Constants.PragmaLoop.REGEX_DISTRIBUTION))
							&& leftovers.stream().anyMatch(s -> s.matches("^" + Constants.PragmaLoop.USE_FISSION + "$"))) {
							model.printIgnoreableError(log, "The Pragma " + pragmaAnnotation.getName() + " is faulty. "
								+ Constants.PragmaLoop.DISTRIBUTION + " and " + Constants.PragmaLoop.USE_FISSION +
								" are not supported together. The distribution tag will be ignored.");
						}
					}
				} else {
					model.printIgnoreableError(log,"The Pragma \"" + pragmaAnnotation.getName() + "\" could not be parsed. " +
							"The Number of Splits or the maxCycles() Tag is not a valid.");
				}
			}
		} else {
			model.printIgnoreableError(log,"The Pragma \"" + pragmaAnnotation.getName() + "\" could not be parsed. " +
					"The Pragma ustreams loopOpt should be followed by at least an number of splits or a maxCycles() Tag");
		}
	} //CHECKSTYLE.ON: MagicNumber

	/**
	 * splits the String at all spaces except one that are enclosed by "( )" brackets
	 * @param pragma the String to split
	 * @return	the result of the splitting
	 */
	private String[] splitPragma(String pragma) {
		String[] split = pragma.split(" ");
		ArrayList<String> result = new ArrayList<>();
		String temp = null;
		for (String s : split) {
			if (temp == null && s.contains("(") && !s.contains(")")) {
				temp = s;
			} else if (temp != null && !s.contains("(") && s.contains(")")) {
				temp = temp + s;
				result.add(temp);
				temp = null;
			} else if (temp!=null && !s.contains("(") && !s.contains(")")) {
				temp = temp + s;
			} else if (temp==null && ((!s.contains("(") && !s.contains(")")) || (s.contains("(") && s.contains(")")))) {
				result.add(s);
			} else {
				model.printIgnoreableError(log,"Error occurred while searching through the LoopPragma. Please check the Placement of brackets in the LoopPragma");
			}
		}
		return result.toArray(new String[result.size()]);
	}
}

