/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.task.pragma;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.tools.Utils;

import cetus.hir.PragmaAnnotation;
import cetus.hir.Traversable;
import org.slf4j.Logger;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Parses the ustreams pragmas
 */
public class PragmaTaskParser {

	private static Logger log = org.slf4j.LoggerFactory.getLogger(PragmaTaskParser.class);

	private final Model model;

	public PragmaTaskParser(Model model) {
		this.model = model;
	}

	/**
	 * Creates a {@link TaskPragma}-Object with the the given
	 * {@link PragmaAnnotation}.
	 *
	 * @param pa                - the {@link PragmaAnnotation} that should be parsed.
	 * @param parentTraversable the parent traversable
	 * @return JSPragmaAnnotation
	 */
	@SuppressWarnings("checkstyle:magicNumber")
	public Pragma parse(PragmaAnnotation pa, Traversable parentTraversable) {
		if (pa != null && parentTraversable != null) {
			validatePragma(pa);
			Collection<PragmaTag<String>> ppList = null;
			String pragma = pa.getName().toLowerCase();
			String[] pragmas = pragma.split(" ", 3);
			if (pragmas.length >= 2 && pragmas[0].equalsIgnoreCase(Constants.Pragma.IDENTIFIER)) {
				String optionals = "";
				if (pragmas.length >= 3) {
					optionals = pragmas[2];
				}
				if (pragmas[1].equalsIgnoreCase(Constants.Pragma.TASKGROUP)) {
					ppList = parseParameter(optionals);
					return new TaskGroupPragma(pa, ppList, parentTraversable);
				} else if (pragmas[1].equalsIgnoreCase(Constants.Pragma.TASK)) {
					ppList = parseParameter(optionals);
					return new TaskPragma(pa, ppList, parentTraversable);
				} else if (pragmas[1].equalsIgnoreCase(Constants.Pragma.END)) {
					return new EndPragma(pa, new ArrayList<>(), parentTraversable);
				} else {
					return null;
				}
			}
			//CHECKSTYLE.ON: MagicNumber
		}
		return null;
	}

	private Collection<PragmaTag<String>> parseParameter(String ppString) {

		Collection<PragmaTag<String>> ppCollection = new ArrayList<PragmaTag<String>>();

		Collection<String> ppStrings = Utils.getMatches(ppString, "[a-zA-Z]+\\(.*?\\)");

		for (String parameter : ppStrings) {

			/* entries part of the parameter */
			String ppeString = parameter.substring(parameter.indexOf('(') + 1,
				parameter.indexOf(')'));

			/* name part of the parameter */
			String ppName = parameter.substring(0, parameter.indexOf('(')).trim();

			PragmaTag<String> pp = new PragmaTag<String>(ppName.toLowerCase());

			/* parsing entries part */
			if (parseParameterEntries(pp, ppeString)) {
				ppCollection.add(pp);
			} else {
				model.printIgnoreableError(log,"Could not parse parameter entries of parameter: \"" + ppName + "\" Ignore ...");
			}
		}
		return ppCollection;
	}

	/**
	 * Parses the entries of an pragma parameter. Returned boolean indicates
	 * whether this was successful or not.
	 *
	 * @param pp        - PragmaParameter
	 * @param ppeString - String containing the entries.
	 * @return boolean
	 */
	private boolean parseParameterEntries(PragmaTag<String> pp,
												 String ppeString) {
		Collection<String> ppeStrings = Utils.getMatches(ppeString, ",?\\s*\\w+\\s*");

		for (String entry : ppeStrings) {
			if (entry.startsWith(",")) {
				pp.addEntry(entry.substring(1).trim());
			} else {
				pp.addEntry(entry.trim());
			}
		}
		return true;
	}

	private void validatePragma(PragmaAnnotation pragmaAnnotation) {
		String[] pragmas = pragmaAnnotation.getName().toLowerCase().split("\\s+");
		if (pragmas.length > 0) {
			if (pragmas[0].equalsIgnoreCase(Constants.Pragma.IDENTIFIER)) {
				List<String> leftovers = Arrays.stream(pragmas)
					.filter(part -> !part.matches("^" + Constants.Pragma.IDENTIFIER + "$"))
					.collect(Collectors.toList());
				if (leftovers.isEmpty()) {
					model.printIgnoreableError(log,"The pragma " + Constants.Pragma.IDENTIFIER + " should at least be followed by a " + Constants.Pragma.TASK);
				}
				leftovers.stream()
						.filter(part -> !part.matches("^" + Constants.Pragma.TASK + "$"))
						.filter(part -> !part.matches("^" + Constants.Pragma.END + "$"))
						.filter(part -> !part.matches(Constants.Pragma.REGEX_VALID_REPLICATE))
						.filter(part -> !part.matches(Constants.Pragma.REGEX_VALID_IN))
						.filter(part -> !part.matches(Constants.Pragma.REGEX_VALID_OUT))
						.forEach(option -> {
							model.printIgnoreableError(log,"The Pragma \"" + pragmaAnnotation.getName() + "\" could not be parsed. The sequence \"" + option + "\" is not known");
						});
			}
		}
	}

}
