/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.task.pragma;

import de.tu_darmstadt.rs.microStreams.model.Model;

import cetus.hir.PragmaAnnotation;
import cetus.hir.Traversable;

import de.tu_darmstadt.rs.microStreams.model.interfaces.HasCetusComponent;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskCoreType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;

/**
 * Represents the the pragma of a task
 */
public class TaskPragma implements Pragma, HasCetusComponent {

	Logger log = LoggerFactory.getLogger(TaskPragma.class);

	private PragmaAnnotation annotation;

	private Collection<PragmaTag<String>> ptCollection;

	private Traversable traversable;

	/**
	 * indicates whether the Task kernel is a function, compound or single
	 * statement
	 */
	private TaskCoreType coreType;

	private Traversable taskTraversable;

	private boolean isTransformed;

	private Model project;

	/**
	 * The constructor
	 * @param annotation the according annotation
	 * @param ptCollection a pragma tag
	 * @param traversable the according traversable
	 */
	public TaskPragma(PragmaAnnotation annotation, Collection<PragmaTag<String>> ptCollection, Traversable traversable) {
		this.annotation = annotation;
		this.traversable = traversable;
		this.ptCollection = ptCollection;
		this.setTransformed(false);
	}

	/**
	 * The Constructor
	 * @param annotation the according annotation
	 * @param ptCollection a pragma tag
	 * @param traversable the according traversable
	 * @param taskTraversable an according task traversable
	 */
	public TaskPragma(PragmaAnnotation annotation,
					  Collection<PragmaTag<String>> ptCollection, Traversable traversable,
					  Traversable taskTraversable) {
		this.annotation = annotation;
		this.traversable = traversable;
		this.ptCollection = ptCollection;
		this.taskTraversable = taskTraversable;
		this.setTransformed(true);
	}


	public Traversable getTaskTraversable() {
		return taskTraversable;
	}


	@Override
	public void addPragmaParameter(PragmaTag<String> pp) {
		ptCollection.add(pp);
	}

	@Override
	public PragmaTag<String> getPragmaTag(String ptName) {
		for (PragmaTag<String> pt : ptCollection) {
			if (ptName.equals(pt.getName())) {
				return pt;
			}
		}
		return null;
	}

	public Collection<PragmaTag<String>> getPragmaTags() {
		return ptCollection;
	}


	public PragmaAnnotation getAnnotation() {
		return annotation;
	}

	public void setAnnotation(PragmaAnnotation annotation) {
		this.annotation = annotation;

	}

	@Override
	public Traversable getTraversable() {
		return traversable;
	}

	@Override
	public void setTraversable(Traversable traversable) {
		this.traversable = traversable;
	}

	public TaskCoreType getCoreType() {
		return coreType;
	}

	public void setCoreType(TaskCoreType coreType) {
		this.coreType = coreType;

	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result
			+ ((annotation == null)? 0 : annotation.hashCode());
		result = prime * result
			+ ((coreType == null)? 0 : coreType.hashCode());
		result = prime * result + ((log == null)? 0 : log.hashCode());
		result = prime * result
			+ ((ptCollection == null)? 0 : ptCollection.hashCode());
		result = prime * result
			+ ((traversable == null)? 0 : traversable.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		TaskPragma other = (TaskPragma) obj;
		if (annotation == null) {
			if (other.annotation != null) {
				return false;
			}
		} else if (!annotation.equals(other.annotation)) {
			return false;
		}
		if (coreType != other.coreType) {
			return false;
		}
		if (ptCollection == null) {
			if (other.ptCollection != null) {
				return false;
			}
		} else if (!ptCollection.equals(other.ptCollection)) {
			return false;
		}
		if (traversable == null) {
			if (other.traversable != null) {
				return false;
			}
		} else if (!traversable.equals(other.traversable)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return annotation.toString();
	}

	public boolean isTransformed() {
		return isTransformed;
	}


	public void setTransformed(boolean isTransformed) {
		this.isTransformed = isTransformed;
	}

	@Override
	public Model getModel() {
		return this.project;
	}

	@Override
	public void setModel(Model project) {
		this.project = project;

	}


}
