/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.variable;

import java.util.Collection;
import java.util.LinkedHashSet;

/**
 * A variable used inside a task
 */
public class TaskVariable {

	private Variable variable;

	private Collection<VariableAccess> access;

	private Collection<VariableAccess> forced;

	/**
	 * Creates a new task variable from a variable
	 *
	 * @param variable the variable to be contained in the task variable
	 */
	public TaskVariable(Variable variable) {
		this.variable = variable;
		this.access = new LinkedHashSet<VariableAccess>();
		this.forced = new LinkedHashSet<VariableAccess>();
	}

	public boolean addAccess(VariableAccess access) {
		return access.isForced()? this.forced.add(access) : this.access.add(access);
	}

	public Variable getVariable() {
		return variable;
	}

	/**
	 *Tells if the variable is read
	 * @return true is the variable access is a read
	 */
	public boolean isRead() {
		for (VariableAccess access : this.access) {
			if (access.isRead()) {
				return true;
			}
		}
		return false;
	}

	/**
	 *Tells if the variable is written
	 * @return true is the variable access is a write
	 */
	public boolean isWritten() {
		for (VariableAccess access : this.access) {
			if (access.isWritten()) {
				return true;
			}
		}
		return false;
	}

	/**
	 *Tells if the variable is forced access
	 * @return true is the variable access is a forced read
	 */
	public boolean isForcedRead() {
		for (VariableAccess access : this.forced) {
			if (access.isRead()) {
				return true;
			}
		}
		return false;
	}

	/**
	 *Tells if the variable is forced written
	 * @return true is the variable access is a forced write
	 */
	public boolean isForcedWritten() {
		for (VariableAccess access : this.forced) {
			if (access.isWritten()) {
				return true;
			}
		}
		return false;
	}

	/**
	 *Tells if the variable is not used
	 * @return true is the task variable is not used
	 */
	public boolean isNotUsed() {
		for (VariableAccess access : this.access) {
			if (!access.isNotUsed()) {
				return false;
			}
		}
		for (VariableAccess access : this.forced) {
			if (!access.isNotUsed()) {
				return false;
			}
		}
		return true;
	}

	public boolean isUsed() {
		return !isNotUsed();
	}

	public boolean isForced() {
		return !forced.isEmpty();
	}

	public boolean removeAccess(VariableAccess access) {
		return this.access.remove(access);
	}

	/**
	 *Gives all variable accesses
	 * @return all variable accesses
	 */
	public Collection<VariableAccess> getVariableAccess() {
		Collection<VariableAccess> result = new LinkedHashSet<VariableAccess>();

		result.addAll(access);
		result.addAll(forced);

		return result;

	}

	public String toString() {
		return variable.toString() + " " + getVariableAccess();
	}

}
