/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.variable;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.VariableUtils;

import cetus.hir.ArraySpecifier;
import cetus.hir.CodeAnnotation;
import cetus.hir.Identifier;
import cetus.hir.Initializer;
import cetus.hir.InlineAnnotation;
import cetus.hir.PointerSpecifier;
import cetus.hir.PrintTools;
import cetus.hir.Specifier;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;

import java.util.ArrayList;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;

/**
 * Represents a variable in the Program
 */
public class Variable implements ModelComponent, Comparable<Variable> {

	private VariableDeclarator variableDeclarator;

	private Collection<Identifier> identifiers = new LinkedHashSet<Identifier>();

	//this one should not be changeable
	private Identifier initialIdentifier;

	private int size = 0;

	private Model project;

	private Scope scope;

	private Include include;

	private Initializer init;

	/**
	 * a constructor
	 * @param variableDeclarator a variable declarator to use
	 * @param identifier the identifier of the variable
	 */
	public Variable(VariableDeclarator variableDeclarator, Identifier identifier) {
		this.variableDeclarator = variableDeclarator;
		initialIdentifier = identifier;
		identifiers.add(identifier);
		scope = Scope.UNKNOWN;
		VariableUtils.extractMissingArrayDimensions(variableDeclarator);
	}

	/**
	 * Copy constructor
	 * @param variable a variable to copy
	 */
	public Variable(Variable variable) {
		variableDeclarator = variable.getVariableDeclarator();
		identifiers = new ArrayList<>(variable.getIdentifiers());
		initialIdentifier = variable.getInitialIdentifier();
		size = variable.getSize();
		project = variable.project;
		scope = variable.scope;
		include = variable.include;
		init = variable.init;
	}

	protected Variable(){

	}

	public VariableDeclarator getVariableDeclarator() {
		return variableDeclarator;
	}

	/**
	 * The Scope that a Variable has
	 */
	public enum Scope {
		UNKNOWN, GLOBAL, LOCAL, GLOBALSHARED, GLOBALWRITTEN
	}

	public void setVariableDeclarator(VariableDeclarator variableDeclarator) {
		this.variableDeclarator = variableDeclarator;
	}

	/**
	 * Returns the name of this {@link Variable} as String.
	 *
	 * @return A String representing the name of this {@link Variable}.
	 */
	public String getName() {
		return variableDeclarator.getSymbolName();
	}

	/**
	 * Gives the initial initializer
	 * @return the initializer of this variable
	 */
	public Initializer getInitializer() {
		return init;
	}


	/**
	 * Returns the declaration of this {@link Variable} as String.
	 *
	 * @return A String representing the declaration of this {@link Variable}.
	 */
	public String getDeclaration() {

		StringBuilder sb = new StringBuilder();

		@SuppressWarnings("rawtypes")
		List specifiers = variableDeclarator.getTypeSpecifiers();

		for (Object specifier : specifiers) {
			sb.append(specifier.toString());
			if (!(specifier instanceof PointerSpecifier)) {
				sb.append(" ");
			}
		}

		sb.append(variableDeclarator.getSymbolName());

		@SuppressWarnings("rawtypes")
		List<Specifier> arraySpecifiers = variableDeclarator.getTrailingSpecifiers();

		for (Specifier specifier : arraySpecifiers) {
			sb.append((CetusUtils.getStringFromPrintable(specifier)));
		}
		return sb.toString();
	}

	public void collectDataFromModel() {
		init = CetusUtils.findGlobalInitializer(CetusUtils.getProgram(project), variableDeclarator);
		include = CetusUtils.getInclude(project, variableDeclarator);
	}

	/**
	 * gives the global declaration
	 * @return the global declaration of a variable going into the global memory
	 */
	public String getGlobalDeclaration() {
		StringBuilder sb = new StringBuilder();
		if (init != null) {
			sb.append(Constants.C.GLOBAL_DEFINE);
		} else {
			sb.append(Constants.C.GLOBAL_DECLARE);
		}

		sb.append("(");

		@SuppressWarnings("rawtypes")
		List specifiers = variableDeclarator.getTypeSpecifiers();


		for (Object specifier : specifiers) {
			sb.append(specifier.toString());
			if (!(specifier instanceof PointerSpecifier)) {
				sb.append(" ");
			}
		}
		sb.deleteCharAt(sb.length()-1);
		sb.append(", ");
		sb.append(variableDeclarator.getSymbolName());

		@SuppressWarnings("rawtypes")
		List arraySpecifiers = variableDeclarator.getArraySpecifiers();

		for (Object specifier : arraySpecifiers) {
			if (specifier instanceof ArraySpecifier) {
				sb.append(" ");
				sb.append((CetusUtils.getStringFromPrintable((ArraySpecifier) specifier)));
			}
		}

		if (init != null) {
			sb.append(", ");
			if (init.getChildren().size() > 1) {
				sb.append(Constants.C.WRAP_INIT_VALUES);
				sb.append("(");
				sb.append(PrintTools.listToString(init.getChildren(), ", "));
				sb.append(")");
			} else {
				sb.append(PrintTools.listToString(init.getChildren(), ", "));
			}
		}

		sb.append(")");
		return sb.toString();
	}

	/**
	 * give declaration with assignment
	 * @return the assignment declaration
	 */
	public String getDeclarationWithAssignment() {
		StringBuilder sb = new StringBuilder();
		sb.append(getDeclaration());
		Initializer init = variableDeclarator.getInitializer();
		if (init != null) {
			sb.append(init);
		}
		return sb.toString();
	}

	/**
	 * Returns the included header file that declares the variable.
	 *
	 * @return the include file.
	 */
	public Include getInclude() {
		return include;
	}

	public boolean isDeclaredInExternalInclude() {
		return getInclude() != null;
	}

	public boolean isGlobal() {
		return scope==Scope.GLOBAL || scope==Scope.GLOBALSHARED || scope==Scope.GLOBALWRITTEN;
	}

	/**
	 * sets the variable to go to the global section
	 * @param global if the variable should go the global memory
	 */
	public void setGlobal(boolean global) {
		if (global && (scope != Scope.GLOBALSHARED || scope != Scope.GLOBALWRITTEN)) {
			scope = Scope.GLOBAL;
		}
		if (!global) {
			scope = Scope.LOCAL;
		}
	}

	/**
	 * the size of a variable
	 * @return the size of this {@link Variable} in Bytes
	 */
	public int getSize() {
		return size;
	}

	/**
	 * the size of a variable
	 * @return the size of this {@link Variable} in Bytes
	 */
	public int getSizeInWords() {
		return (int) Math.ceil((double)getSize()/2);
	}

	public void setSize(int size) {
		this.size = size;
	}

	public boolean addIdentifier(Identifier identifier) {
		return identifiers.add(identifier);
	}

	public Identifier getInitialIdentifier() {
		return initialIdentifier;
	}

	/**
	 * if the variable as an identifier with the given name
	 * @param identifierName an identifier name
	 * @return true if the variable holds such an identifier
	 */
	public boolean hasIdentifierNamed(String identifierName) {
		return identifiers.stream()
			.filter(i -> i.getName().equalsIgnoreCase(identifierName))
			.count() > 0;
	}

	public Collection<Identifier> getIdentifiers() {
		return identifiers;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + size;
		result = prime * result + initialIdentifier.hashCode();
		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) {
			return true;
		}
		if (o == null || getClass() != o.getClass()) {
			return false;
		}

		Variable variable = (Variable) o;

		if (size != variable.size) {
			return false;
		}
		//don't use equals here since the overloaded Identifier.equals() is to weak by just comparing the toString
		return initialIdentifier == variable.initialIdentifier;
	}

	@Override
	public String toString() {
		return variableDeclarator.getSymbolName();
	}

	@Override
	public Model getModel() {
		return this.project;
	}

	@Override
	public void setModel(Model project) {
		this.project = project;
		collectDataFromModel();
	}

	@Override
	public int compareTo(Variable o) {
		int comp = getName().compareTo(o.getName());

		if (0 == comp) {
			//
		}

		return comp;
	}

	public boolean isGlobalShared() {
		return scope==Scope.GLOBALWRITTEN || scope==Scope.GLOBALSHARED;
	}

	public boolean isGlobalWritten() {
		return scope==Scope.GLOBALWRITTEN;
	}

	/**
	 * set when the variable is written
	 * @param globalWritten true if the variable is global and written
	 */
	public void setGlobalWritten(boolean globalWritten) {
		if (globalWritten) {
			scope = Scope.GLOBALWRITTEN;
		}
		if (!globalWritten && scope==Scope.GLOBALWRITTEN) {
			scope = Scope.GLOBALSHARED;
		}
	}

	/**
	 * set when the variable is shared among several tasks
	 * @param globalShared true if it is a shared variable
	 */
	public void setGlobalShared(boolean globalShared) {
		if (globalShared && scope != Scope.GLOBALWRITTEN) {
			scope = Scope.GLOBALSHARED;
		}
		if (!globalShared) {
			if (scope == Scope.GLOBALSHARED || scope == Scope.GLOBALWRITTEN) {
				scope = Scope.GLOBAL;
			}
		}
	}
}
