/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.model.software.variable;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.interfaces.AccessesVariables;
import de.tu_darmstadt.rs.microStreams.model.software.SoftwareModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;

import cetus.hir.Identifier;
import cetus.hir.NameID;

import de.tu_darmstadt.rs.microStreams.model.software.task.Task;

/**
 * <p>
 * The {@link VariableAccess} class describes the {@link Variable} access trough
 * a {@link Task} or {@link UFunction}.
 * <p>
 * <p>
 * <p>
 * It contains the type of access (READ, WRITE or both) (described with with
 * {@link AccessBehavior} enumeration) and the {@link Identifier} (Cetus IR) to
 * determine the exact location, e.g. if it is accessed before or after a child
 * tasks execution within the {@link Task} or {@link UFunction} (direct parent).
 * <p>
 * <p>
 * <p>
 * In addition to that it is marked as "forced" with the forced field, if the
 * {@link VariableAccess} is set through a #pragma tag instead of autodetection.
 * </p>
 *
 * @author Jan Weber
 */
public class VariableAccess implements SoftwareModelComponent {

	/**
	 * The Variable that is accessed through this VariableAccess
	 */
	private Variable variable;
	/**
	 * The AccessesVariables in which this VariableAccess takes place.
	 */
	private AccessesVariables directParent;
	/**
	 * The type of access (e.g. READ, WRITE, both)
	 */
	private AccessBehavior behavior;
	/**
	 * Cetus IR Identifier used to determine code location.
	 */
	private Identifier identifier;
	private NameID nameid;
	private Model model;
	private boolean forced;

	/**
	 * The Constructor
	 * @param directParent the direct parent
	 * @param identifier the identifier
	 * @param variable the according variable
	 * @param forced if it is enforced by the user
	 */
	public VariableAccess(AccessesVariables directParent,
						  Identifier identifier, Variable variable, boolean forced) {
		this.identifier = identifier;
		this.directParent = directParent;
		this.variable = variable;
		this.forced = forced;
	}

	/**
	 * The Constructor
	 * @param directParent the direct parent
	 * @param nameID the name identifier
	 * @param variable the according variable
	 * @param forced if it is enforced by the user
	 */
	public VariableAccess(AccessesVariables directParent,
						  NameID nameID, Variable variable, boolean forced) {
		this.nameid = nameID;
		this.directParent = directParent;
		this.variable = variable;
		this.forced = forced;
	}

	/**
	 * The Constructor
	 * @param directParent the direct parent
	 * @param variable a variable
	 * @param behavior the access behaviour of the variable
	 * @param forced if it is enforced by the user
	 */
	public VariableAccess(AccessesVariables directParent, Variable variable,
						  AccessBehavior behavior, boolean forced) {
		this.behavior = behavior;
		this.directParent = directParent;
		this.variable = variable;
		this.forced = forced;
	}

	/**
	 * Returns the AccessBehavior behavior
	 *
	 * @return AccessBehavior behavior
	 */
	public AccessBehavior getBehavior() {
		return behavior;
	}

	/**
	 * Sets the AcccessBehavior behavior to the given value.
	 *
	 * @param behavior - the value the AccessBehavior behavior is set to.
	 */
	public void setBehavior(AccessBehavior behavior) {
		this.behavior = behavior;
	}

	/**
	 * Returns the {@link AccessesVariables} that contains this
	 * {@link VariableAccess}.
	 *
	 * @return {@link AccessesVariables}
	 */
	public AccessesVariables getDirectParent() {
		return directParent;
	}

	/**
	 * Returns the Cetus IR identifier.
	 *
	 * @return Cetus IR identifier.
	 */
	public Identifier getIdentifier() {
		return identifier;
	}

	public NameID getNameID() {
		return nameid;
	}

	@Override
	public Model getModel() {
		return this.model;
	}

	@Override
	public void setModel(Model m) {
		this.model = m;
	}

	/**
	 * return the according variable
	 * @return the Variable that is accessed through this VariableAccess
	 */
	public Variable getVariable() {
		return variable;
	}

	/**
	 * Returns whether this {@link VariableAccess} is forced through pragma
	 * tags.
	 *
	 * @return true or false
	 */
	public boolean isForced() {
		return forced;
	}

	/**
	 * Returns true if the VariableAccess describes a read access otherwise
	 * false.
	 *
	 * @return true or false.
	 */
	public boolean isRead() {
		return ((AccessBehavior.READ == behavior) || (AccessBehavior.READWRITE == behavior));
	}

	/**
	 * Returns true if the VariableAccess describes a write access otherwise
	 * false.
	 *
	 * @return true or false.
	 */
	public boolean isWritten() {
		return ((AccessBehavior.WRITE == behavior) || (AccessBehavior.READWRITE == behavior));
	}

	/**
	 * Returns true if the VariableAccess is never used in this context otherwise
	 * false.
	 *
	 * @return true or false.
	 */
	public boolean isNotUsed() {
		return (AccessBehavior.UNTOUCHED == behavior);
	}

	@Override
	public String toString() {
		return "VariableAccess [variable=" + variable + ", behavior="
			+ behavior + ", identifier=" + identifier + ", forced="
			+ forced + "]";
	}

	/**
	 * <p>
	 * An enumeration that describes the type variable access. It can take the
	 * following values:
	 * <p>
	 * <p>
	 * <ul>
	 * <li>UNKOWN - if the value of the variable could not determined and ended
	 * with an error.</li>
	 * <li>UNTOUCHED - the variable is not accessed at all.</li>
	 * <li>READ - the variable is read.</li>
	 * <li>WRITE - the variables is written.</li>
	 * <li>READWRITE - the variable is both read and written.</li>
	 * </ul>
	 *
	 * @author Jan Weber
	 */
	public enum AccessBehavior {
		UNKNOWN, UNTOUCHED, READ, WRITE, READWRITE
	}

}
