/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.tools;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.Peripheral;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConcentratorMaster;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConnectionPeripheralDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.DispatcherMaster;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IConnectionPeripheralDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.MemSwapMultiMasterDMA;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.CombinedConnectionPeripheralAccess;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.ConnectionPeripheralAccess;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.PeripheralUsage;
import de.tu_darmstadt.rs.microStreams.model.software.freemarker.Body;
import de.tu_darmstadt.rs.microStreams.model.software.freemarker.CodeSegment;
import de.tu_darmstadt.rs.microStreams.model.software.freemarker.ConnectionPeripheralWait;
import de.tu_darmstadt.rs.microStreams.model.software.freemarker.EvaluationHardwareAccess;
import de.tu_darmstadt.rs.microStreams.model.software.freemarker.Measurement;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency.DependencyRelation;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency.DependencyType;

import cetus.hir.CompoundStatement;
import cetus.hir.DepthFirstIterator;
import cetus.hir.Statement;
import cetus.hir.Traversable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Creator for the freemarker printables
 */
public class BodyBuilder {

	private Logger log = LoggerFactory.getLogger(this.getClass());

	private Traversable traversable;
	private CompoundStatement convertedTraversable;
	private Collection<ConnectionPeripheralAccess> connectionAccesses;
	private Collection<Task> containedTasks;

	private String completeBody;

	private Collection<BodyDummy> dummies = new TreeSet<BodyDummy>();

	private LinkedList<Body> body = new LinkedList<Body>();

	private boolean evaluation;
	private boolean pruneVariableDeclarations;

	private int combinedPeriAccessID;

	/**
	 * The constructor
	 * @param traversable a traversable
	 * @param connectionAccesses the connection access
	 * @param containedTasks the contained tasks
	 * @param pruneVariableDeclarations if all VariableDeclarations should be discarded
	 * @param createPerformanceEvaluation if evaluation is desired
	 */
	public BodyBuilder(Traversable traversable,
					   Collection<ConnectionPeripheralAccess> connectionAccesses,
					   Collection<Task> containedTasks, boolean pruneVariableDeclarations, boolean createPerformanceEvaluation) {
		this.traversable = traversable;
		this.connectionAccesses = connectionAccesses;
		this.containedTasks = containedTasks;
		this.evaluation = createPerformanceEvaluation;
		this.pruneVariableDeclarations = pruneVariableDeclarations;
		this.combinedPeriAccessID = 0;

		if (pruneVariableDeclarations) {
			if (traversable instanceof Statement) {
				convertedTraversable = CetusUtils.toCompundStatement((Statement) traversable);
				completeBody = CetusUtils.getStringFromPrintable(convertedTraversable, CetusUtils::getListOfSplittedVariableDeclaratorsWithoutInstantiation);
			} else {
				throw new IllegalArgumentException("If pruneVariableDeclarations is switched on, the traversable has to be a Statement");
			}

		} else {
			completeBody = CetusUtils.getStringFromPrintable(traversable);
		}
	}

	private static int getIndexOfFirstPragma(String code, String pragma) {
		StringBuffer stringBuffer = deletePragmaComments(code);
		return stringBuffer.toString().indexOf(pragma);
	}

	private static int getIndexOfFirstPragma(String code, String pragma, int before) {
		StringBuffer stringBuffer = deletePragmaComments(code);

		int index = getIndexOfFirstPragma(code, pragma);
		int savedIndex = index;

		while (index < before && index > 0) {
			savedIndex = index;
			index = stringBuffer.toString().indexOf(pragma, index + 1);
		}

		return savedIndex;
	}

	private static StringBuffer deletePragmaComments(String code) {
		Pattern commentPragma = Pattern.compile(Constants.Pragma.C_COMMENT_WITH_PRAGMA_REGEX);
		Matcher m = commentPragma.matcher(code);
		StringBuffer stringBuffer = new StringBuffer();

		while (m.find()) {
			m.appendReplacement(stringBuffer, new String(new char[m.group(0).length()]));
		}
		m.appendTail(stringBuffer);
		return stringBuffer;
	}

	/**
	 * generates the new method body
	 *
	 * @return new sourcecode
	 */
	public List<Body> generateBody() {

		if (connectionAccesses.isEmpty() && containedTasks.size() == 0) {
			String bodyContent = null;
			if (pruneVariableDeclarations) {
				bodyContent = CetusUtils.getStringFromPrintable(convertedTraversable, CetusUtils::getListOfSplittedVariableDeclaratorsWithoutInstantiation);
			} else {
				bodyContent = CetusUtils.getStringFromPrintable(traversable);
			}
			body.add(new Body(new CodeSegment(bodyContent)));
		} else {

			/*
			 * 1. Create a sorted set of all Tasks and their position within the
			 * body. (Sorted by their position index within traversable)
			 */
			for (Task t : containedTasks) {

				if (t.isReplica()) {
					continue;
				}

				BodyDummy bd;
				if (t.hasReplica()) {
					bd = new BodyDummyReplicatedTask(t);
				} else {
					bd = new BodyDummyTask(t);
				}
				ArrayList<ConnectionPeripheralDMA> handledDMAPeripherals = new ArrayList<>();
				for (ConnectionPeripheralAccess ca : connectionAccesses) {

					TaskDependency d = ca.getDependency();

					if (ca.isInParent() && !handledDMAPeripherals.contains(ca.getPeripheral())) {
						if ((d.getRelation() == DependencyRelation.CHILD_PARENT && d.getProvider() == t)
								|| (d.getRelation() == DependencyRelation.PARENT_CHILD && d.getConsumer() == t)) {
							bd.addConnectionAccess(ca);
							if (ca.getPeripheral() instanceof ConnectionPeripheralDMA) {
								handledDMAPeripherals.add((ConnectionPeripheralDMA) ca.getPeripheral());
							}

							/*
							 * if there are replicas, add the corresponding
							 * connection peri access to body
							 */
							if (t.hasReplica()) {

								for (ConnectionPeripheralAccess rca : connectionAccesses) {

									if (d == rca.getDependency()
											&& t == rca.getTask() && rca != ca) {
										bd.addConnectionAccess(rca);
										if (ca.getPeripheral() instanceof ConnectionPeripheralDMA) {
											handledDMAPeripherals.add((ConnectionPeripheralDMA) ca.getPeripheral());
										}
									}
								}

							}

						}
					}
				}
				dummies.add(bd);
			}

			/*
			 * 2. Get receiving ConnectionAccesses at the beginning of a child's
			 * body. (Receiving in child)
			 */

			boolean evaluationHardwareAccessAdded = false;

			TreeSet<SendReceiveGroup> groups = getReceiveGroups();

			SendReceiveGroup lastGroup = null;
			boolean measuring = false;
			for (SendReceiveGroup bdrt : groups) {
				if (evaluation) {
					if (lastGroup == null || !lastGroup.isSameTarget(bdrt)) {
						if (measuring) {
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.RECEIVE,
									Measurement.STOP)));
						}
						if (!(bdrt.connectionAccesses.iterator().next().getPeripheral() instanceof IConnectionPeripheralDMA)) {
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.RECEIVE_WAIT,
									Measurement.START)));

							body.add(new Body(new ConnectionPeripheralWait(
									PeripheralUsage.RECEIVE_WAIT,
									bdrt.connectionAccesses.iterator().next().getPeripheral().getHwAddress())));

							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.RECEIVE_WAIT,
									Measurement.STOP)));
						}
						body.add(new Body(new EvaluationHardwareAccess(
								PeripheralUsage.RECEIVE,
								Measurement.START)));
						measuring = true;
					}
				}
				body.addAll(bdrt.getRealBody());
				lastGroup = bdrt;
			}

			if (measuring && evaluation) {
				body.add(new Body(new EvaluationHardwareAccess(
						PeripheralUsage.RECEIVE,
						Measurement.STOP)));
			}
			if (evaluation) {
				body.add(new Body(new EvaluationHardwareAccess(
						PeripheralUsage.RUNTIME,
						Measurement.START)));
			}

			/*
			 * 3. Generate CodeSegments and insert the ConnectionAccesses for
			 * each task
			 */
			int currentPosition = 0;
			int lastPosition = 0;

			//update indices of tasks having the same body. Indices could not be calculated during creation since at this point duplicates can not be detected.
			HashMap<BodyDummy, Integer> taskDuplicates = new HashMap<>();
			for (BodyDummy dummy : dummies) {
				long index = taskDuplicates.keySet().stream()
						.filter(d -> CetusUtils.getStringFromPrintable(d.getTask())
								.equalsIgnoreCase(CetusUtils.getStringFromPrintable(dummy.getTask())))
						.count();
				taskDuplicates.put(dummy, Math.toIntExact(index));
			}
			taskDuplicates.entrySet().stream().filter(entry -> entry.getValue().intValue() > 0).forEach(entry -> {
				entry.getKey().updateIndices(entry.getValue().intValue());
			});
			for (BodyDummy dummy : dummies) {
				currentPosition = dummy.getStartIndexPragma();

				if (lastPosition < currentPosition) {
					String bodyString = completeBody.substring(lastPosition, currentPosition);
					if (!bodyString.matches("[\\n]+")) {
						// Only add bodies that contain more that only newlines.
						body.add(new Body(new CodeSegment(bodyString)));
					}
				}

				body.addAll(dummy.getConnectionCode());

				lastPosition = dummy.getEndIndex();
			}

			if (lastPosition < completeBody.length()) {
				body.add(new Body(new CodeSegment(completeBody
						.substring(lastPosition))));
			}

			/*
			 * 4. Get sending ConnectionAccesses at the end of child's body.
			 * (Sending in child tasks)
			 */
			if (evaluation) {
				body.add(new Body(new EvaluationHardwareAccess(
						PeripheralUsage.RUNTIME,
						Measurement.STOP)));
			}

			evaluationHardwareAccessAdded = false;

			TreeSet<SendReceiveGroup> groups2 = getSendGroups();

			measuring = false;
			lastGroup = null;
			for (SendReceiveGroup bdrt : groups2) {
				if (evaluation) {
					if (lastGroup == null || !lastGroup.isSameTarget(bdrt)) {
						if (measuring) {
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.SEND,
									Measurement.STOP)));
						}
						if (!(bdrt.connectionAccesses.iterator().next().getPeripheral() instanceof IConnectionPeripheralDMA)) {
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.SEND_WAIT,
									Measurement.START)));
							body.add(new Body(new ConnectionPeripheralWait(
									PeripheralUsage.SEND_WAIT,
									bdrt.connectionAccesses.iterator().next().getPeripheral().getHwAddress())));
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.SEND_WAIT,
									Measurement.STOP)));
						}
						body.add(new Body(new EvaluationHardwareAccess(
								PeripheralUsage.SEND,
								Measurement.START)));
						measuring = true;
					}
				}
				body.addAll(bdrt.getRealBody());
				lastGroup = bdrt;
			}

			if (measuring && evaluation) {
				body.add(new Body(new EvaluationHardwareAccess(
						PeripheralUsage.SEND,
						Measurement.STOP)));
			}

		}

		/* Removes unessecary eval acceses */
		ListIterator<Body> iter = body.listIterator();
		Body body = null;
		EvaluationHardwareAccess access = null;
		EvaluationHardwareAccess oldAccess;

		while (iter.hasNext()) {
			oldAccess = access;
			body = iter.next();

			access = body.getEvaluationHardwareAccess();
			if (null != access && null != oldAccess
					&& oldAccess.getMeasurement().equals(access.getMeasurement())
					&& !oldAccess.getPeripheralUsage().equals(access.getPeripheralUsage())) {
				iter.previous();
				iter.remove();
				iter.next();
				iter.remove();
			}

		}

		return this.body;
	}

	private TreeSet<SendReceiveGroup> getReceiveGroups() {
		return getSendReceiveGroups(true);
	}

	private TreeSet<SendReceiveGroup> getSendGroups() {
		return getSendReceiveGroups(false);
	}

	private TreeSet<SendReceiveGroup> getSendReceiveGroups(boolean generateReceiveGroup) {
		TreeSet<SendReceiveGroup> groups = new TreeSet<>();
		ArrayList<ConnectionPeripheralAccess> handled = new ArrayList<>();
		ArrayList<ConnectionPeripheralDMA> handledDMAPeripherals = new ArrayList<>();

		for (ConnectionPeripheralAccess ca : connectionAccesses) {
			if (!handled.contains(ca) && !handledDMAPeripherals.contains(ca.getPeripheral()) && ca.isInChild() && (ca.isReceive() == generateReceiveGroup)) {
				TaskDependency d = ca.getDependency();
				if (DependencyType.MANY_TO_ONE == d.getType() && generateReceiveGroup || DependencyType.ONE_TO_MANY == d.getType() && !generateReceiveGroup) {
					SendReceiveGroup bdrt = new SendReceiveGroup(d);
					bdrt.addConnectionAccess(ca);
					for (ConnectionPeripheralAccess ca2 : connectionAccesses) {
						if (!handled.contains(ca2)) {
							Task task;
							Task task2;
							if (generateReceiveGroup) {
								task = d.getProvider();
								task2 = ca2.getDependency().getProvider();
							} else {
								task = d.getConsumer();
								task2 = ca2.getDependency().getConsumer();
							}

							if (ca2.getDependency().equals(d) || (ca2.isInChild() && ca2.isReceive() && task2.equals(task))) {
								bdrt.addConnectionAccess(ca2);
								handled.add(ca2);
								if (ca2.getPeripheral() instanceof ConnectionPeripheralDMA) {
									handledDMAPeripherals.add((ConnectionPeripheralDMA) ca2.getPeripheral());
								}
							}
						}
					}
					groups.add(bdrt);
					handled.add(ca);
				} else {
					SendReceiveGroup bdrt = new SendReceiveGroup(d);
					bdrt.addConnectionAccess(ca);
					groups.add(bdrt);
					handled.add(ca);
					if (ca.getPeripheral() instanceof ConnectionPeripheralDMA) {
						handledDMAPeripherals.add((ConnectionPeripheralDMA) ca.getPeripheral());
					}
				}
			}
		}
		return groups;
	}

	/**
	 * Abstract Body
	 */
	private abstract class BodyDummy implements Comparable<BodyDummy> {

		protected Task task;
		protected int startIndex;
		protected int finishIndex;
		protected String taskString;
		protected String pragmaString;
		protected int startIndexPragma;
		protected Collection<ConnectionPeripheralAccess> connectionAccesses = new TreeSet<ConnectionPeripheralAccess>();
		protected boolean taskTransformed;

		BodyDummy(Task task) {
			this.task = task;

			taskString = CetusUtils.getStringFromPrintable(task.getTraversable());
			taskTransformed = task.getTaskPragma().isTransformed();

			pragmaString = taskTransformed? "" : CetusUtils.getStringFromPrintable(task.getTaskPragma().getTraversable());

			updateIndices(0);
		}

		abstract List<Body> getConnectionCode();

		public int getStartIndexPragma() {
			return startIndexPragma;
		}

		public int getStartIndex() {
			return startIndex;
		}

		public int getEndIndex() {
			return finishIndex;
		}

		public boolean addConnectionAccess(ConnectionPeripheralAccess ca) {
			return this.connectionAccesses.add(ca);
		}

		public Task getTask() {
			return task;
		}

		public void updateIndices(int occurancesToIgnore) {
			for (int fromIndex = 0; occurancesToIgnore >= 0; occurancesToIgnore--) {
				startIndex = completeBody.indexOf(taskString, fromIndex);
				if (startIndex < 0) {
					throw new IndexOutOfBoundsException("The string of the contained task can not be found inside the current task " + getTask().toString());
				}
				fromIndex = startIndex + 1;
			}
			finishIndex = startIndex + taskString.length();
			if (task.hasEndPragma()) {
				String endPragma = CetusUtils.getStringFromPrintable(task.getEndPragma().getTraversable());
				finishIndex = completeBody.indexOf(endPragma, finishIndex) + endPragma.length();
			}
			startIndexPragma = taskTransformed? startIndex : getIndexOfFirstPragma(completeBody, pragmaString, startIndex);
		}

		@Override
		public int compareTo(BodyDummy other) {
			if (task.getTraversable().equals(other.getTask().getTraversable())) {
				return 0;
			}
			Traversable thisLocation = CetusUtils.getLocation(task.getTraversable());
			Traversable otherLocation = CetusUtils.getLocation(other.getTask().getTraversable());
			if (thisLocation != otherLocation) {
				throw new RuntimeException("The traverables to be compared are not in the same Translation unit");
			}
			DepthFirstIterator<Traversable> locationIterator = new DepthFirstIterator<>(thisLocation);
			while (locationIterator.hasNext()) {
				Traversable current = locationIterator.next();

				if (current == task.getTraversable()) {
					return -1;
				}
				if (current == other.getTask().getTraversable()) {
					return 1;
				}
			}
			throw new RuntimeException("The two objects con not be compared");
		}

		@Override
		public int hashCode() {
			int result = startIndex;
			result = 31 * result + finishIndex;
			result = 31 * result + startIndexPragma;
			result = 31 * result + task.hashCode();
			result = 31 * result + taskString.hashCode();
			result = 31 * result + pragmaString.hashCode();
			result = 31 * result + (taskTransformed? 1 : 0);
			result = 31 * result + (connectionAccesses != null? connectionAccesses.hashCode() : 0);
			return result;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (o == null || getClass() != o.getClass()) {
				return false;
			}

			BodyDummyTask that = (BodyDummyTask) o;

			if (startIndex != that.startIndex) {
				return false;
			}
			if (finishIndex != that.finishIndex) {
				return false;
			}
			if (startIndexPragma != that.startIndexPragma) {
				return false;
			}
			if (taskTransformed != that.taskTransformed) {
				return false;
			}
			if (!task.equals(that.task)) {
				return false;
			}
			if (!taskString.equals(that.taskString)) {
				return false;
			}
			if (!pragmaString.equals(that.pragmaString)) {
				return false;
			}
			return connectionAccesses != null? connectionAccesses.equals(that.connectionAccesses) : that.connectionAccesses == null;
		}
	}


	/**
	 * A group of core interconnect receivers
	 */
	private class SendReceiveGroup implements Comparable<SendReceiveGroup> {

		private Collection<ConnectionPeripheralAccess> connectionAccesses = new TreeSet<ConnectionPeripheralAccess>();

		private TaskDependency taskdependency;

		public SendReceiveGroup(TaskDependency taskdependency) {
			this.taskdependency = taskdependency;
		}

		private boolean isCombined() {

			Peripheral p = null;

			for (ConnectionPeripheralAccess ca : connectionAccesses) {
				if (ca.getPeripheral() instanceof DispatcherMaster || ca.getPeripheral() instanceof ConcentratorMaster || ca.getPeripheral() instanceof MemSwapMultiMasterDMA) {
					return true;
				} else if (null == p) {
					p = ca.getPeripheral();
				} else {
					if (!p.equals(ca.getPeripheral())) {
						return true;
					}
				}

			}

			return false;
		}

		private CombinedConnectionPeripheralAccess getCombined() {
			CombinedConnectionPeripheralAccess cmcpa = new CombinedConnectionPeripheralAccess(combinedPeriAccessID++);

			for (ConnectionPeripheralAccess ca : this.connectionAccesses) {

				cmcpa.addConnectionPeripheralAccess(ca);

			}

			return cmcpa;

		}

		public boolean addConnectionAccess(ConnectionPeripheralAccess ca) {
			return this.connectionAccesses.add(ca);

		}

		public List<Body> getRealBody() {
			LinkedList<Body> body = new LinkedList<Body>();

			if (!connectionAccesses.isEmpty()) {

				if (isCombined()) {
					body.add(new Body(getCombined()));
				} else {
					for (ConnectionPeripheralAccess ca : connectionAccesses) {
						body.add(new Body(ca));
					}
				}

			}
			return body;
		}

		@Override
		public int compareTo(SendReceiveGroup o) {
			return taskdependency.compareTo(o.taskdependency);
		}

		public boolean isSameTarget(SendReceiveGroup o) {
			return taskdependency.isSameTarget(o.taskdependency);
		}
	}


	/**
	 * A dummy for a replicated task
	 */
	private class BodyDummyReplicatedTask extends BodyDummy {


		public BodyDummyReplicatedTask(Task task) {
			super(task);
		}

		private CombinedConnectionPeripheralAccess getCombinedSend() {
			CombinedConnectionPeripheralAccess cmcpa = new CombinedConnectionPeripheralAccess(combinedPeriAccessID++);

			for (ConnectionPeripheralAccess ca : this.connectionAccesses) {
				if (ca.isSend()) {
					cmcpa.addConnectionPeripheralAccess(ca);
				}
			}

			return cmcpa;

		}

		private CombinedConnectionPeripheralAccess getCombinedReceive() {
			CombinedConnectionPeripheralAccess cmcpa = new CombinedConnectionPeripheralAccess(combinedPeriAccessID++);

			for (ConnectionPeripheralAccess ca : this.connectionAccesses) {
				if (ca.isReceive()) {
					cmcpa.addConnectionPeripheralAccess(ca);
				}
			}

			return cmcpa;

		}

		@Override
		public List<Body> getConnectionCode() {

			boolean isRuntimeStopped = false;

			LinkedList<Body> body = new LinkedList<Body>();

			CombinedConnectionPeripheralAccess cmcpa = getCombinedSend();

			if (!cmcpa.isEmpty()) {

				if (evaluation) {
					isRuntimeStopped = true;
					body.add(new Body(new EvaluationHardwareAccess(
							PeripheralUsage.RUNTIME,
							Measurement.STOP)));
					if (!(cmcpa.getPeripherals().iterator().next() instanceof IConnectionPeripheralDMA)) {
						body.add(new Body(new EvaluationHardwareAccess(
								PeripheralUsage.SEND_WAIT,
								Measurement.START)));

						body.add(new Body(new ConnectionPeripheralWait(
								PeripheralUsage.SEND_WAIT,
								cmcpa.getPeripherals().iterator().next().getHwAddress())));

						body.add(new Body(new EvaluationHardwareAccess(
								PeripheralUsage.SEND_WAIT,
								Measurement.STOP)));
					}
					body.add(new Body(new EvaluationHardwareAccess(
							PeripheralUsage.SEND,
							Measurement.START)));
				}
				body.add(new Body(cmcpa));
				if (evaluation) {
					body.add(new Body(new EvaluationHardwareAccess(
							PeripheralUsage.SEND,
							Measurement.STOP)));

				}
			}

			cmcpa = getCombinedReceive();

			if (!cmcpa.isEmpty()) {

				if (evaluation) {
					if (!isRuntimeStopped) {
						isRuntimeStopped = true;
						body.add(new Body(new EvaluationHardwareAccess(
								PeripheralUsage.RUNTIME,
								Measurement.STOP)));
					}
					if (!(cmcpa.getPeripherals().iterator().next() instanceof IConnectionPeripheralDMA)) {
						body.add(new Body(new EvaluationHardwareAccess(
								PeripheralUsage.RECEIVE_WAIT,
								Measurement.START)));
						body.add(new Body(new ConnectionPeripheralWait(
								PeripheralUsage.RECEIVE_WAIT,
								cmcpa.getPeripherals().iterator().next().getHwAddress())));
						body.add(new Body(new EvaluationHardwareAccess(
								PeripheralUsage.RECEIVE_WAIT,
								Measurement.STOP)));
					}
					body.add(new Body(new EvaluationHardwareAccess(
							PeripheralUsage.RECEIVE,
							Measurement.START)));
				}
				body.add(new Body(cmcpa));

				if (evaluation) {
					body.add(new Body(new EvaluationHardwareAccess(
							PeripheralUsage.RECEIVE,
							Measurement.STOP)));
				}

			}
			if (isRuntimeStopped && evaluation) {
				body.add(new Body(new EvaluationHardwareAccess(
						PeripheralUsage.RUNTIME,
						Measurement.START)));
			}
			return body;
		}
	}


	/**
	 * A dummy for a task body
	 */
	private class BodyDummyTask extends BodyDummy {

		public BodyDummyTask(Task task) {
			super(task);
		}

		public List<Body> getConnectionCode() {

			boolean isRuntimeStopped = false;
			boolean measuringSend = false;
			boolean measuringReceive = false;
			ConnectionPeripheralAccess lastAccess = null;

			LinkedList<Body> body = new LinkedList<Body>();

			for (ConnectionPeripheralAccess connectionAccess : connectionAccesses) {
				if (connectionAccess.getDependency().getConsumer().equals(task)) {
					if (evaluation) {
						if (!isRuntimeStopped) {
							isRuntimeStopped = true;
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.RUNTIME,
									Measurement.STOP)));
						}

						if (lastAccess == null || !lastAccess.getDependency().isSameTarget(connectionAccess.getDependency())) {
							if (measuringSend) {
								body.add(new Body(new EvaluationHardwareAccess(
										PeripheralUsage.SEND,
										Measurement.STOP)));
							}
							if (!(connectionAccess.getPeripheral() instanceof IConnectionPeripheralDMA)) {
								body.add(new Body(new EvaluationHardwareAccess(
										PeripheralUsage.SEND_WAIT,
										Measurement.START)));

								body.add(new Body(new ConnectionPeripheralWait(
										PeripheralUsage.SEND_WAIT,
										connectionAccess.getPeripheral().getHwAddress())));

								body.add(new Body(new EvaluationHardwareAccess(
										PeripheralUsage.SEND_WAIT,
										Measurement.STOP)));
							}
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.SEND,
									Measurement.START)));
							measuringSend = true;
						}
					}
					body.add(new Body(connectionAccess));
					lastAccess = connectionAccess;
				}
			}

			if (measuringSend && evaluation) {
				body.add(new Body(new EvaluationHardwareAccess(
						PeripheralUsage.SEND,
						Measurement.STOP)));

			}

			lastAccess = null;


			for (ConnectionPeripheralAccess connectionAccess : connectionAccesses) {

				if (connectionAccess.getDependency().getProvider().equals(task)) {
					if (evaluation) {
						if (!isRuntimeStopped) {
							isRuntimeStopped = true;
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.RUNTIME,
									Measurement.STOP)));
						}

						if (lastAccess == null || !lastAccess.getDependency().isSameTarget(connectionAccess.getDependency())) {
							if (measuringReceive) {
								body.add(new Body(new EvaluationHardwareAccess(
										PeripheralUsage.RECEIVE,
										Measurement.STOP)));
							}
							if (!(connectionAccess.getPeripheral() instanceof IConnectionPeripheralDMA)) {
								body.add(new Body(new EvaluationHardwareAccess(
										PeripheralUsage.RECEIVE_WAIT,
										Measurement.START)));
								body.add(new Body(new ConnectionPeripheralWait(
										PeripheralUsage.RECEIVE_WAIT,
										connectionAccess.getPeripheral().getHwAddress())));
								body.add(new Body(new EvaluationHardwareAccess(
										PeripheralUsage.RECEIVE_WAIT,
										Measurement.STOP)));
							}
							body.add(new Body(new EvaluationHardwareAccess(
									PeripheralUsage.RECEIVE,
									Measurement.START)));
							measuringReceive = true;
						}
					}
					body.add(new Body(connectionAccess));
					lastAccess = connectionAccess;
				}
			}

			if (measuringReceive && evaluation) {
				body.add(new Body(new EvaluationHardwareAccess(
						PeripheralUsage.RECEIVE,
						Measurement.STOP)));
			}
			if (isRuntimeStopped && evaluation) {
				body.add(new Body(new EvaluationHardwareAccess(
						PeripheralUsage.RUNTIME,
						Measurement.START)));
			}

			return body;
		}
	}
}
