/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.tools;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.util.FileUtil;

import cetus.exec.Driver;
import cetus.hir.PreprocessorException;
import cetus.hir.Program;
import cetus.hir.Tools;
import cetus.hir.UnknownIncludeException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;


/**
 * This is an adapted Cetus {@link cetus.exec.Driver Driver} implementation. It
 * Instantiates, runs the Cetus Source-To-Source Compiler Infrastructure and
 * allows to extract the the {@link cetus.hir.Program Program} for further
 * handling. In order to get the {@link Program} from a {@link Collection} of
 * {@link File Files} simply the static method
 * {@link CetusDriver#parse()} can be used.
 *
 * @author Jan Weber
 */
public final class CetusDriver extends Driver {

	private static Logger log = LoggerFactory.getLogger(CetusDriver.class);
	private Collection<File> files;
	private List<File> searchPathes;
	private boolean usePreprocessor;
	private Collection<File> additionalSystemLibraries;



	/**
	 * A constructor
	 *
	 * @param files A {@link Collection} of C program files to be parsed.
	 * @param searchPathes A {@link List} of directories to scan for C header files in descending priority order.
	 * @param additionalSystemLibraries libs to search for defs
	 */
	public CetusDriver(Collection<File> files, List<File> searchPathes, Collection<File> additionalSystemLibraries) {
		this(files, searchPathes, true, additionalSystemLibraries);
	}

	/**
	 * A constructor
	 *
	 * @param files A {@link Collection} of C program files to be parsed.
	 * @param searchPathes A {@link List} of directories to scan for C header files in descending priority order.
	 * @param usePreprocessor if a preprocessor shall be used
	 * @param additionalSystemLibraries libs to search for defs
	 */
	public CetusDriver(Collection<File> files, List<File> searchPathes, boolean usePreprocessor, Collection<File> additionalSystemLibraries) {
		super();
		filenames = new ArrayList<>(files.size());
		this.files = new ArrayList<>(files);
		this.searchPathes = new ArrayList<>(searchPathes);
		this.usePreprocessor = usePreprocessor;
		this.additionalSystemLibraries = new ArrayList<>(additionalSystemLibraries);
	}


	/**
	 * Parses the given {@link Collection} of {@link File Files} with the aid of
	 * Cetus and returns the result for further usage.
	 * @return the parsed program from the given files.
	 */
	public Program parse() {

		/* Add search pathes to the preprocessor */
		if (usePreprocessor) {
			//use the temp folder if missing headers have to be generated
			searchPathes.add(Constants.FilesAndFolders.TMP_FOLDER_MICROSTREAMS_USER.toFile());
			setOptionValue("preprocessor", buildPreprocessorCmd(searchPathes, additionalSystemLibraries));
		}

		for (File file : files) {
			String fileName = file.getAbsolutePath();

			if (!filenames.contains(fileName)) {
				filenames.add(fileName);
			}

		}

		Tools.exitThrowsException(true);

		boolean retry = true;

		while (retry) {
			try {
				parseFiles();
				retry = false;
				log.info("Done parsing files: " + files);
			} catch (PreprocessorException e) {
				if (e instanceof UnknownIncludeException) {
					String missingFile = ((UnknownIncludeException) e).getMissingFile();
					if (Constants.SpartanMC.PREPROCESSOR_NOT_FOUND_INCLUDES_TO_IGNORE.matcher(missingFile).matches()) {
						generateMissingDummyIncludeFile(missingFile);
						log.info("Restarting parser with replaced dummy header " + missingFile);
					} else {
						throw e;
					}

				} else {
					throw e;
				}
			}
		}

		return program;

	}


	private void generateMissingDummyIncludeFile(String file) {
		File outFile = Constants.FilesAndFolders.TMP_FOLDER_MICROSTREAMS_USER.resolve(file).toFile();
		FileUtil.writeTemporaryFile(outFile, "");
	}

	private String buildPreprocessorCmd(List<File> searchPathes, Collection<File> additionalSystemLibraries) {
		StringBuilder sb = new StringBuilder(Constants.Cetus.PREPROCESSOR_COMMAND);

		for (File searchPath : searchPathes) {
			sb.append(Constants.Cetus.PREPROCESSOR_INCLUDE);
			sb.append(searchPath.getAbsolutePath());
		}
		for (File searchPath : additionalSystemLibraries) {
			sb.append(Constants.Cetus.PREPROCESSOR_INCLUDE);
			sb.append(searchPath.getAbsolutePath());
		}

		return sb.toString();
	}

}
