/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.tools;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.other.HeaderFile;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.FunctionCallgraph;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.Annotatable;
import cetus.hir.AnnotationDeclaration;
import cetus.hir.AnnotationStatement;
import cetus.hir.ArraySpecifier;
import cetus.hir.BreadthFirstIterator;
import cetus.hir.ClassDeclaration;
import cetus.hir.CloneableTraversable;
import cetus.hir.CompoundStatement;
import cetus.hir.Declaration;
import cetus.hir.DeclarationStatement;
import cetus.hir.DepthFirstIterator;
import cetus.hir.Expression;
import cetus.hir.FunctionCall;
import cetus.hir.IDExpression;
import cetus.hir.Identifier;
import cetus.hir.Initializer;
import cetus.hir.IntegerLiteral;
import cetus.hir.Literal;
import cetus.hir.NameID;
import cetus.hir.PragmaAnnotation;
import cetus.hir.Printable;
import cetus.hir.Procedure;
import cetus.hir.Program;
import cetus.hir.Specifier;
import cetus.hir.Statement;
import cetus.hir.Symbol;
import cetus.hir.SymbolTable;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Cetus Utilities
 *
 * @author Jan Weber
 */
public class CetusUtils {

	/* slf4j logger */
	private static Logger log = LoggerFactory.getLogger(CetusUtils.class);

	public CetusUtils() {
	}

	/**
	 * Returns the string that a Cetus {@link Printable} would print.
	 *
	 * @param p the {@link Printable} whose String shall be returned.
	 * @return Returns the string that would be printed by calling the print()-method of the given {@link Printable} object
	 */
	public static String getStringFromPrintable(Printable p) {
		Objects.requireNonNull(p);

		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);

		p.print(pw);
		return sw.toString();
	}

	/**
	 * Returns the string that a Cetus {@link Printable} would print and prunes the printable with the accept method
	 *
	 * @param compoundStatement the {@link Traversable} whose String shall be returned.
	 * @param transformer a filter/trasformator function
	 * @return Returns the string that would be printed by calling the print()-method of the given {@link Traversable} object while pruning with accept method
	 */
	public static String getStringFromPrintable(CompoundStatement compoundStatement, Function<Traversable, Optional<List<? extends Traversable>>> transformer) {
		Objects.requireNonNull(compoundStatement);
		Objects.requireNonNull(transformer);

		StringWriter sw = new StringWriter();
		PrintWriter pw = new PrintWriter(sw);
		CompoundStatement.defaultPrint(compoundStatement, pw, transformer);
		return sw.toString();
		/*StringBuilder stringBuilder = new StringBuilder();
		for (Traversable current: (Iterable<Traversable>) () -> new FlatIterator<>(traversable)) {
			if (!accept.test(current)) {
				stringBuilder.append(CetusUtils.getStringFromPrintable(current));
			}
		}
		return stringBuilder.toString();*/
	}

	/**
	 * Returns the string that a Cetus {@link Printable} would print and prunes the printable with the accept method
	 *
	 * @param task the {@link Task} whose String shall be returned.
	 * @return Returns the string that would be printed by calling the print()-method of the given {@link Task}s traversable object while pruning with {@link CetusUtils#isDeclarationWithoutInstantiation}
	 */
	public static String getStringFromPrintable(Task task) {
		if (task.getTraversable() instanceof Statement) {
			return getStringFromPrintable(toCompundStatement((Statement) task.getTraversable()), CetusUtils::getListOfSplittedVariableDeclaratorsWithoutInstantiation);
		}
		throw new IllegalArgumentException(task + "s traversable is not of type compound statement.");
	}

	/**
	 * Tells if the traversable is a declaration statement and is instantiated upon declaration
	 *
	 * @param t a traversable to be analyzed
	 * @return if the traversable is a declaration statement and is not instantiated upon declaration false otherwise
	 */
	public static boolean isDeclarationWithoutInstantiation(Traversable t) {
		Declaration declaration = null;
		if (t instanceof VariableDeclaration) {
			declaration = (VariableDeclaration) t;
		} else if (t instanceof DeclarationStatement) {
			declaration = ((DeclarationStatement) t).getDeclaration();
		}
		if (declaration != null) {
			boolean hasInitializer = declaration.getChildren().stream()
							.flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class))
							.anyMatch(varDec -> varDec.getInitializer() != null);
			if (!hasInitializer) {
				return true;
			}
		}
		return false;
	}

	/**
	 * A function to prune unwanted traversables in the generated code of the body builder
	 * @param t a traversable to be transformed
	 * @return Optional.empty() if the traversable should be dismissed, else the same or a bunch of other traversables that should be put instead of the original
	 */
	public static Optional<List<? extends Traversable>> getListOfSplittedVariableDeclaratorsWithoutInstantiation(Traversable t) {
		VariableDeclaration variableDeclaration = null;
		if (t instanceof VariableDeclaration) {
			variableDeclaration = (VariableDeclaration) t;
		} else if (t instanceof DeclarationStatement) {
			variableDeclaration = (VariableDeclaration)((DeclarationStatement) t).getDeclaration();
		}

		if (variableDeclaration != null) {
			if (variableDeclaration.getNumDeclarators() > 1) {
				final List<DeclarationStatement> variableDeclarations = CetusUtils.variableDeclarationSplitter(variableDeclaration);
				variableDeclarations.removeIf(CetusUtils::isDeclarationWithoutInstantiation);
				return Optional.of(variableDeclarations);
			} else if (isDeclarationWithoutInstantiation(variableDeclaration)) {
				return Optional.empty();
			}
		}
		return Optional.of(Arrays.asList(t));
	}

	/**
	 * Splits a variable declaration with multiple identifiers into multiple declarations each with its own specifier
	 * @param variableDeclaration a VariableDeclaration to be transformed
	 * @return multiple declarations each with its own specifier
	 */
	public static List<DeclarationStatement> variableDeclarationSplitter(VariableDeclaration variableDeclaration) {
		List<DeclarationStatement> ret = new ArrayList<>();
		if (variableDeclaration.getNumDeclarators() > 1) {
			final List<VariableDeclarator> variableDeclarators = variableDeclaration.getDeclarators().stream().flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class)).collect(Collectors.toList());
			for ( VariableDeclarator variableDeclarator : variableDeclarators) {
				VariableDeclaration vd = variableDeclaration.clone();
				final List<VariableDeclarator> variableDeclarators2 = vd.getDeclarators().stream().flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class)).collect(Collectors.toList());
				for (VariableDeclarator variableDeclarator2 : variableDeclarators2) {
					if (!variableDeclarator2.getSymbolName().equalsIgnoreCase(variableDeclarator.getSymbolName())) {
						CetusUtils.removeChild(variableDeclarator2);
					}
				}
				ret.add(new DeclarationStatement(vd));
			}
		} else {
			ret.add(new DeclarationStatement(variableDeclaration));
		}
		return ret;
	}

	/**
	 * Returns a Set of all Symbols contained in parental
	 * SymbolTables visible to the scope of the given
	 * {@link Traversable}
	 *
	 * @param t the returned symbols are visible in the scope of this traversable
	 * @return Returns the Set of Symbols
	 */
	public static Set<Symbol> getVisibleSymbols(Traversable t) {

		Set<Symbol> ss = new LinkedHashSet<Symbol>();

		while (null != t) {
			if (t instanceof SymbolTable) {
				ss.addAll(((SymbolTable) t).getSymbols());
			}
			t = t.getParent();
		}

		return ss;
	}

	/**
	 * Generates the search path (list of directories), the result contains only
	 * directories that are relevant for the given {@link SourceFile}.
	 *
	 * @param m the current µStreams {@link Model} as search context.
	 * @param file file for which the search path is generated.
	 * @return search path (list of directories).
	 */
	public static List<File> getSearchPathFor(Model m, SourceFile file) {

		Collection<SourceFile> srcFiles = m
				.getModelComponents(SourceFile.class);

		List<File> searchPath = new ArrayList<File>();

		/* first look at the c files folder */
		if (null != file) {
			File parent = file.getFile().getParentFile();
			if (!searchPath.contains(parent)) {
				searchPath.add(parent);
			}
			File folder = file.getFolder();
			if (null != folder && !searchPath.contains(folder)) {
				searchPath.add(folder);
			}
		}

		/* then at the given folders */
		for (SourceFile srcFile : srcFiles) {

			File folder = srcFile.getFolder();

			if (null != folder && !searchPath.contains(folder)) {
				searchPath.add(folder);
			}
		}

		searchPath.addAll(m.getConfig().getAdditionalSystemLibraries());

		return searchPath;

	}

	/**
	 * Generates the search path (List of {@link File Files}) for the given
	 * Collection of {@link SourceFile SourceFiles}.
	 *
	 * @param m the current µStreams {@link Model} as search context.
	 * @return search path (aka list of files).
	 */
	public static List<File> getSearchPath(Model m) {

		Collection<SourceFile> srcFiles = m
				.getModelComponents(SourceFile.class);

		List<File> searchPath = new ArrayList<File>();

		/* generate search path for C header files */
		for (SourceFile srcFile : srcFiles) {

			File folder = srcFile.getFolder();

			File file = srcFile.getFile();

			if (null != folder && !searchPath.contains(folder)) {
				searchPath.add(folder);
			}

			if (null != file && !searchPath.contains(file.getParentFile())) {
				searchPath.add(file.getParentFile());
			}

		}

		return searchPath;
	}

	/**
	 * Searches and returns the {@link HeaderFile} for the given headerName.
	 * This method fetches the search path for the given {@link SourceFile} and
	 * scans for the {@link HeaderFile}. If the {@link HeaderFile} is not
	 * present in the given {@link Model}, it is added as {@link ModelComponent}
	 * for further usage .
	 *
	 * @param m the current µStreams {@link Model} as search context.
	 * @param srcFile the sourcecode file the header is searched for.
	 * @param headerName the name of the header that is searched.
	 * @return the {@link HeaderFile} for the given headerName
	 */
	public static HeaderFile getHeaderFile(Model m, SourceFile srcFile, String headerName) {

		Collection<HeaderFile> hFiles = m.getModelComponents(HeaderFile.class);

		Collection<File> searchPath = getSearchPathFor(m, srcFile);

		for (File dir : searchPath) {
			File file = new File(dir, headerName);

			if (file.exists()) {

				/* checking whether header already present in model. */
				for (HeaderFile hFile : hFiles) {
					if (null != hFile.getFile() && hFile.getFile().equals(file)) {
						return hFile;
					}
				}

				/* .. otherwise create new one. */
				HeaderFile hFile = new HeaderFile(file);
				m.addModelComponent(hFile);
				return hFile;
			}
		}

		return null;
	}

	/**
	 * Extracts the included file from the given {@link AnnotationDeclaration}.
	 * Returns null if no include is found.
	 *
	 * @param model the current µStreams {@link Model} as search context.
	 * @param ad Cetus AnnotationDeclaration
	 * @return the {@link Include}.
	 */
	public static Include getIncludeFromAnnotation(Model model,
												   AnnotationDeclaration ad) {

		boolean isSystemLibrary = ad.toString().trim().matches(Constants.Cetus.SYSTEM_LIB_INCLUDE_REGEX);

		String s = Utils.firstMatch(ad.toString().trim(), Constants.Cetus.LIB_INCLUDE_REGEX, 1);

		if (null != s) {
			Include include = new Include(s.replaceAll("<|>|\"", ""),
					isSystemLibrary, false);

			if (!isSystemLibrary) {

				/* 1. store in which c source file this include is placed */
				SourceFile location = getSourceFile(model, ad);

				if (null == location) {
					model.printIgnoreableError(log, "Could not determine the source file this include takes place. " + ad);
					return null;
				}

				include.setLocation(location);

				/* 2. find the c header file */
				HeaderFile header = getHeaderFile(model, location, include.getName());

				if (null == header) {
					log.warn("Could not locate the header file which is included by this directive.");
					return null;
				}

				include.setHeaderFile(header);

			}
			return include;
		}
		return null;
	}

	/**
	 * Returns the {@link SourceFile} the given {@link Traversable} is located
	 * in.
	 *
	 * @param m the current {@link Model} as search context.
	 * @param t the traversable.
	 * @return the {@link SourceFile}.
	 */
	public static SourceFile getSourceFile(Model m, Traversable t) {

		File file = new File(getLocation(t).getInputFilename());

		Collection<SourceFile> srcFiles = m
				.getModelComponents(SourceFile.class);

		for (SourceFile srcFile : srcFiles) {
			if (null != srcFile.getFile() && srcFile.getFile().equals(file)) {
				return srcFile;
			}
		}

		return null;
	}

	/**
	 * Returns the {@link TranslationUnit} the given {@link Traversable} t is
	 * contained.
	 *
	 * @param t the Traversable the TranslationUnit is searched for.
	 * @return The Translation Unit
	 */
	public static TranslationUnit getLocation(Traversable t) {
		return findParentOfType(t, TranslationUnit.class);
	}

	/**
	 * Returns other Translation Units that might as well define the variable Declaration
	 *
	 * @param variableDeclaration a variable declaration
	 * @return a list of other locations where this variable is defined. Ideally this should be only one.
	 */
	public static List<TranslationUnit> getOtherTranslationUnitsContaining(VariableDeclaration variableDeclaration) {
		TranslationUnit origin = getLocation(variableDeclaration);
		BiFunction<Traversable, Traversable, Boolean> isSameIDCast = (t1, t2) -> {
			if (t1 instanceof VariableDeclaration && t2 instanceof VariableDeclaration) {
				return Boolean.valueOf(isSameID((VariableDeclaration) t1, (VariableDeclaration) t2));
			}
			return Boolean.FALSE;
		};
		Comparator<VariableDeclaration> bodyOccuranceBeforeIncludeOccurance = Comparator.comparingInt(v -> occurancePointOf(findParentOfType(v, TranslationUnit.class), v, isSameIDCast).ordinal());
		Predicate<Traversable> occurancePoint = v -> {
			final OccurancePlaceInTranslationUnit occurancePlaceInTranslationUnit = occurancePointOf(findParentOfType(v, TranslationUnit.class), v, isSameIDCast);
			return occurancePlaceInTranslationUnit != OccurancePlaceInTranslationUnit.NOWHERE && occurancePlaceInTranslationUnit != OccurancePlaceInTranslationUnit.INCLUDE;
		};
		return origin.getParent().getChildren().stream()
					.filter(tu -> tu != origin)
					.flatMap(tu -> tu.getChildren().stream())
					.flatMap(StreamUtil.castOrEmpty(VariableDeclaration.class))
					.filter(c -> isSameID(variableDeclaration, c))
					.sorted(bodyOccuranceBeforeIncludeOccurance.reversed())
					.filter(occurancePoint)
					.map(CetusUtils::getLocation)
					.collect(Collectors.toList());
	}

	/**
	 * finds the occurance of the searchable in a translation unit telling if it resides in the body, include, both or none
	 * @param container a translation unit to search in
	 * @param searchable a traversable to search
	 * @param equals a custom equals method or null to use the usual equals method of the object
	 * @return if the searchable resides in the body, include, both or nowhere
	 */
	public static OccurancePlaceInTranslationUnit occurancePointOf(TranslationUnit container, Traversable searchable, BiFunction<Traversable, Traversable, Boolean> equals) {
		if (equals == null) {
			equals = 	(x, y) -> x.equals(y);
		}
		final Iterator<Traversable> iterator = container.getChildren().iterator();
		Set<OccurancePlaceInTranslationUnit> ret = EnumSet.noneOf(OccurancePlaceInTranslationUnit.class);
		boolean insideInclude = false;
		while (iterator.hasNext()) {
			Traversable t = iterator.next();
			if (t instanceof AnnotationDeclaration) {
				if (t.toString().matches(Constants.Cetus.LIB_INCLUDE_REGEX)) {
					insideInclude = true;
				} else if (t.toString().matches(Constants.Cetus.END_INCLUDE_REGEX)) {
					insideInclude = false;
				}
			}
			if (equals.apply(t, searchable)) {
				if (insideInclude) {
					ret.add(OccurancePlaceInTranslationUnit.INCLUDE);
				} else {
					ret.add(OccurancePlaceInTranslationUnit.BODY);
				}
			}
		}
		if (ret.isEmpty()) {
			log.warn("The searched traversable " + searchable + " can not be found inside the container " + container);
		}
		return OccurancePlaceInTranslationUnit.occuranceFromSet(ret);
	}
	/**
	 * Checks if two variable declarations have the same specifiers and the same declared ids
	 *
	 * @param d1 one variable declaration
	 * @param d2 another variable declaration
	 * @return true if they have an equal declaration and name
	 */
	public static boolean isSameDefinition(VariableDeclaration d1, VariableDeclaration d2) {
		return isSameSpecifier(d1, d2) && isSameID(d1, d2) && isSameDeclarator(d1, d2);
	}

	/**
	 * Checks if two variable declarations have the same specifiers
	 *
	 * @param d1 one variable declaration
	 * @param d2 another variable declaration
	 * @return true if they have an equal declaration and name
	 */
	public static boolean isSameSpecifier(VariableDeclaration d1, VariableDeclaration d2) {
		List<Specifier> specifiers1 = d1.getSpecifiers();
		List<Specifier> specifiers2 = d2.getSpecifiers();
		if (specifiers1.containsAll(specifiers2) && specifiers1.size() == specifiers2.size()) {
			return true;
		}
		return false;
	}

	/**
	 * Checks if two variable declarations have the same declarator
	 *
	 * @param d1 one variable declaration
	 * @param d2 another variable declaration
	 * @return true if they have an equal declaration and name
	 */
	public static boolean isSameDeclarator(VariableDeclaration d1, VariableDeclaration d2) {
		List<String> symbols1 = d1.getDeclaredSymbols().stream().map(Symbol::toString).collect(Collectors.toList());
		List<String> symbols2 = d2.getDeclaredSymbols().stream().map(Symbol::toString).collect(Collectors.toList());

		if (symbols1.containsAll(symbols2) && symbols1.size() == symbols2.size()) {
			return true;
		}
		return false;
	}

	/**
	 * Checks if two variable declarations have the same declared ids
	 *
	 * @param d1 one variable declaration
	 * @param d2 another variable declaration
	 * @return true if they have an equal name
	 */
	public static boolean isSameID(VariableDeclaration d1, VariableDeclaration d2) {
		List<IDExpression> declaredIDs1 = d1.getDeclaredIDs();
		List<IDExpression> declaredIDs2 = d2.getDeclaredIDs();
		if (declaredIDs1.containsAll(declaredIDs2) && declaredIDs1.size() == declaredIDs2.size()) {
			return true;
		}
		return false;
	}


	/**
	 * Returns whether the {@link Traversable} traversableA is contained the
	 * given other {@link Traversable} traversableB.
	 *
	 * @param traversableA {@link Traversable} to find.
	 * @param traversableB {@link Traversable} which is scanned.
	 * @return Returns true or false, depending whether the first
	 * {@link Traversable} is contained in the second one.
	 */
	public static boolean contains(Traversable traversableA, Traversable traversableB) {
		while (traversableA != null) {
			traversableA = traversableA.getParent();

			if (traversableB.equals(traversableA)) {
				return true;
			}

		}
		return false;

	}

	/**
	 * Returns whether the {@link Traversable} traversable is contained the
	 * given {@link Task} task.
	 *
	 * @param traversable {@link Traversable} to find.
	 * @param task {@link Task} which is scanned.
	 * @param model a model to search for other tasks
	 * @return Returns true or false, depending whether the first
	 * {@link Traversable} is contained in the second one.
	 */
	public static boolean contains(Traversable traversable, Task task, Model model) {
		final Collection<Task> otherTasks = model.getModelComponents(Task.class);
		otherTasks.remove(task);
		final List<Traversable> otherTasksTraversable = otherTasks.stream().map(t -> t.getTraversable()).collect(Collectors.toList());
		Predicate<? super Traversable> noTraversablesFromOtherTasks = t -> !otherTasksTraversable.contains(t);
		final boolean containedInTask = getStreamOfAllChildren(task.getTraversable(), noTraversablesFromOtherTasks).anyMatch(t -> t.equals(traversable));
		return containedInTask;

	}

	/**
	 * Compares if two traversables have the same ancestor hierarchy
	 *
	 * @param t1 first traversable
	 * @param t2 second traversable
	 * @return true if they have all identical ancestors until the root, false otherwise
	 */
	public static boolean compareIdentifiersAncestors(Traversable t1, Traversable t2) {
		if (t1 != null && t2 != null && t1.equals(t2)) {
			return compareIdentifiersAncestors(t1.getParent(), t2.getParent());
		} else if (t1 == t2) {
			return true;
		} else {
			return false;
		}
	}


	/**
	 * Returns the included header file that declares the variable.
	 *
	 * @param project a project to search on
	 * @param traversable a traversable to search for
	 * @return the include file.
	 */
	public static Include getInclude(Model project, Traversable traversable) {

		TranslationUnit translationUnit = CetusUtils.getLocation(traversable);
		Optional<VariableDeclarator> variableDeclarator = Stream.of(traversable).flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class)).findFirst();
		Optional<TranslationUnit> alternativetranslationUnitContainingDefinition = Optional.empty();
		if (variableDeclarator.isPresent() && variableDeclarator.get().getDeclaration() instanceof VariableDeclaration && ((VariableDeclaration)variableDeclarator.get().getDeclaration()).isExtern()) {
			alternativetranslationUnitContainingDefinition = CetusUtils.getOtherTranslationUnitsContaining(((VariableDeclaration)variableDeclarator.get().getDeclaration())).stream().distinct().collect(StreamUtil.atMostOne());
		}
		if (translationUnit != null) {
			DepthFirstIterator<Traversable> iter = new DepthFirstIterator<>(translationUnit);

			Include include = null;

			while (iter.hasNext()) {
				final Traversable t = iter.next();

				if (FunctionCallgraph.isStartIncludePragma(t)) {
					include = CetusUtils.getIncludeFromAnnotation(project, (AnnotationDeclaration) t);
				} else if (FunctionCallgraph.isEndIncludePragma(t)) {
					include = null;
				} else if (null != include) {
					if (t instanceof VariableDeclarator && variableDeclarator.isPresent() && ((VariableDeclaration)variableDeclarator.get().getDeclaration()).isExtern() && ((VariableDeclarator) t).getSymbolName().equalsIgnoreCase(variableDeclarator.get().getSymbolName())) {
						if (alternativetranslationUnitContainingDefinition.isPresent()) {
							include.setLocation(CetusUtils.getSourceFile(project, alternativetranslationUnitContainingDefinition.get()));
						}
						return include;
					}
					if (t.getParent() instanceof TranslationUnit && t instanceof VariableDeclaration) {
						if (t == traversable || matchFunctionParamAndNames(traversable, (VariableDeclaration) t)) {
							//if the variable comes from an include, but then it must also be declared in another translation unit and not in this tu. Set it if we find other file otherwise keep the main
							if (traversable instanceof FunctionCall && ((FunctionCall) traversable).getProcedure() != null) {
								final SourceFile location = CetusUtils.getSourceFile(project, ((FunctionCall) traversable).getProcedure());
								include.setLocation(location);
								return include;
							} else {
								List<TranslationUnit> otherTUs = CetusUtils.getOtherTranslationUnitsContaining((VariableDeclaration) t);
								if (!otherTUs.isEmpty()) {
									SourceFile sourceFileAlternative = CetusUtils.getSourceFile(project, otherTUs.stream().findFirst().get());
									include.setLocation(sourceFileAlternative);
								}
								return include;

							}
						}
					}
				}
			}
		}
		return null;
	}


	private static boolean matchFunctionParamAndNames(Traversable traversable, VariableDeclaration variableDeclaration) {
		if (traversable instanceof FunctionCall) {
			FunctionCall functionCall = (FunctionCall) traversable;
			if (functionCall.getName() instanceof Identifier) {
				Identifier identifier = (Identifier) functionCall.getName();
				if (variableDeclaration.getDeclaredSymbols().size() > 0 && identifier.getSymbol() != null && identifier.getSymbol().getDeclaration() != null) {
					Declaration dec1 = variableDeclaration.getDeclaredSymbols().stream().findFirst().get().getDeclaration();
					Declaration dec2 = identifier.getSymbol().getDeclaration();
					if (dec1 == dec2) {
						return true;
					}
				}
			}

		}
		return false;
	}

	/**
	 * wraps the statement in a compound statemtent
	 *
	 * @param statement a cetus statemant
	 * @return the wraped statement in a compund statement
	 */
	public static CompoundStatement toCompundStatement(Statement statement) {
		if (statement instanceof CompoundStatement) {
			return (CompoundStatement) statement;
		} else {
			CompoundStatement convertedTraversable = new CompoundStatement();
			Statement s = (statement).clone();
			s.setParent(null);
			convertedTraversable.addStatement(s);
			s.setParent(convertedTraversable);
			return convertedTraversable;
		}
	}

	/**
	 * Gives the next parent of the given type
	 *
	 * @param traversable a traversable whos parents shall be searched
	 * @param searchedParentType the type to be searched
	 * @param <T> the searched type
	 * @return the parent of the type or null if there is no such type
	 */
	public static <T extends Traversable> T findParentOfType(Traversable traversable, Class<T> searchedParentType) {
		while (traversable != null) {
			if (searchedParentType.isInstance(traversable)) {
				return (T) traversable;
			} else {
				traversable = traversable.getParent();
			}
		}
		return null;
	}


	/**
	 * Gives the next child of the given type using breadth first search
	 *
	 * @param traversable a traversable whos children shall be searched
	 * @param searchedParentType the type to be searched
	 * @param <T> the searched type
	 * @return the child of the type or null if there is no such type
	 */
	public static <T extends Traversable> T findChildOfType(Traversable traversable, Class<T> searchedParentType) {
		BreadthFirstIterator iterator = new BreadthFirstIterator(traversable);
		while (iterator.hasNext()) {
			Traversable t = iterator.next();
			if (searchedParentType.isInstance(t) && traversable != t) {
				return (T) t;
			}
		}
		return null;
	}

	/**
	 * gives the dimensions of an array
	 *
	 * @param variableDeclarator a variable declarator
	 * @return Empty Optional if the size can not be determined or the size
	 *
	 * @throws IllegalArgumentException if the variableDeclaration is not an array
	 */
	public static Optional<Long> getArrayElements(VariableDeclarator variableDeclarator) throws IllegalArgumentException {
		long arraySize = 1;
		final List arraySpecifiers = variableDeclarator.getArraySpecifiers();
		if (arraySpecifiers != null && arraySpecifiers.size() > 0) {

			final ArraySpecifier arraySpecifier = (ArraySpecifier) arraySpecifiers.stream().collect(StreamUtil.exactlyOne());

			for (int i = 0; i < arraySpecifier.getNumDimensions(); i++) {

				if (arraySpecifier.getDimension(i) instanceof IntegerLiteral) {
					arraySize *= ((IntegerLiteral) arraySpecifier.getDimension(i)).getValue();
				} else {
					return Optional.empty();
				}
			}
			return Optional.of(new Long(arraySize));
		}
		throw new IllegalArgumentException("The Variable declaration is not an array");
	}

	/**
	 * Finds the main procedure in the traversables Children
	 *
	 * @param start a traversable housing the main in its children
	 * @return the main traversable
	 */
	public static Procedure getMainProcedure(Program start) {

		Traversable tMain = null;

		DepthFirstIterator<Traversable> iter = new DepthFirstIterator<Traversable>(
				start);

		while (iter.hasNext()) {
			Traversable t = iter.next();

			if (t instanceof Procedure) {
				Procedure p = (Procedure) t;

				if (p.getName().toString().equals(Constants.C.MAIN_FUNCTION_NAME)) {
					if (null == tMain) {
						tMain = p;
					} else {
						log.warn("More than one main task procedure \""
								+ Constants.C.MAIN_FUNCTION_NAME + "\" found.");
					}
				}
			}
		}

		if (null == tMain) {
			log.warn("No main task procedure \"" + Constants.C.MAIN_FUNCTION_NAME
					+ "\"  found.");
		}

		return (Procedure) tMain;
	}

	/**
	 * Returns the Program from the model
	 *
	 * @param m a model
	 * @return a program if one exists
	 */
	public static Program getProgram(Model m) {
		Collection<CetusProgram> p = m.getModelComponents(CetusProgram.class);

		if (!p.isEmpty() && p.size() == 1) {
			return p.iterator().next().getProgram();
		}

		return null;
	}

	/**
	 * see if the traversable is a annotation statement
	 *
	 * @param traversable a traversable to check
	 * @return true if it is a annotation, false otherwise
	 */
	public static boolean isAnnotation(Traversable traversable) {
		return (traversable instanceof AnnotationStatement || traversable instanceof AnnotationDeclaration);
	}

	/**
	 * get the pragma annotation of the traversable if it is an annotation
	 *
	 * @param traversable the traversable to transform
	 * @return a casted traversable or null
	 */
	public static PragmaAnnotation getAnnotation(Traversable traversable) {
		if (isAnnotation(traversable)) {
			return ((Annotatable) traversable).getAnnotation(PragmaAnnotation.class, Constants.Pragma.PRAGMA);
		}
		return null;
	}

	/**
	 * Returns all includes from a translation unit
	 *
	 * @param translationUnit a translation unit to be searched for includes
	 * @param model a model to work on
	 * @return a list of includes
	 */
	public static List<Include> getAllIncludes(TranslationUnit translationUnit, Model model) {
		DepthFirstIterator<Traversable> iter = new DepthFirstIterator<>(translationUnit);

		Traversable t = null;
		List<Include> includes = new ArrayList<>();

		while (iter.hasNext()) {
			t = iter.next();

			if (FunctionCallgraph.isStartIncludePragma(t)) {
				Include include = CetusUtils.getIncludeFromAnnotation(model, (AnnotationDeclaration) t);
				if (include != null) {
					includes.add(include);
				}
			}
		}
		return includes;
	}

	/**
	 * returns the direct children of a traversable in a stream
	 *
	 * @param t a traversable
	 * @return the children as a stream
	 */
	public static Stream<Traversable> getStreamOfChildren(Traversable t) {
		if (t == null || t.getChildren() == null) {
			return Stream.empty();
		}
		return t.getChildren().stream().filter(Objects::nonNull);
	}

	public static <T extends Traversable> Stream<T> getStreamOfChildren(Traversable t, Class<T> childrenType) {
		return streamChildrenOfType(t, childrenType);
	}

	/**
	 * returns the parent in a stream
	 * @param t a traversable
	 * @return the parent in a stream
	 */
	public static Stream<Traversable> getStreamOfParent(Traversable t) {
		if (t == null || t.getParent() == null) {
			return Stream.empty();
		}
		return Stream.of(t.getParent());
	}

	public static Stream<Traversable> getStreamOfAllParents(Traversable t) {
		return StreamUtil.streamTree(t, CetusUtils::getStreamOfParent);
	}

	public static <T extends Traversable> Stream<T> getStreamOfAllParents(Traversable traversable, Class<T> onlyWith) {
		return getStreamOfAllParents(traversable).flatMap(StreamUtil.castOrEmpty(onlyWith));
	}

	public static Stream<? extends Traversable> getStreamOfChildrenWithout(Traversable t, Set<Class<? extends Traversable>> ignoreChildrenTypes) {
		return getStreamOfChildren(t).filter(ignoreChildrenTypes(ignoreChildrenTypes));
	}

	public static Stream<? extends Traversable> getStreamOfChildrenWithout(Traversable t, Predicate<? super Traversable> filter) {
		return getStreamOfChildren(t).filter(filter);
	}

	public static <T extends Traversable> Stream<T> streamChildrenOfType(Traversable t, Class<T> childrenType) {
		return streamChildrenOfType(t, Collections.singleton(childrenType));
	}

	/**
	 * streams direct children of a sepecific type
	 * @param t a traversable whos children shall be converted to a stream
	 * @param childrenTypes the types of children to filter
	 * @param <T> return type
	 * @return all direct children having the specified type
	 */
	public static <T extends Traversable> Stream<T> streamChildrenOfType(Traversable t, Set<Class<? extends T>> childrenTypes) {
		return getStreamOfChildren(t).flatMap(c -> {
			for (Class<? extends T> type : childrenTypes) {
				if (type.isAssignableFrom(c.getClass())) {
					return Stream.of(type.cast(c));
				}
			}
			return Stream.empty();
		});
	}

	public static Stream<Traversable> getStreamOfAllChildren(Traversable traversable, Set<Class<? extends Traversable>> ignoreChildrenTypes) {
		return StreamUtil.streamTree(traversable, t -> CetusUtils.getStreamOfChildrenWithout(t, ignoreChildrenTypes));
	}

	public static Stream<Traversable> getStreamOfAllChildren(Traversable traversable, Predicate<? super Traversable> filter) {
		return StreamUtil.streamTree(traversable, t -> CetusUtils.getStreamOfChildrenWithout(t, filter));
	}

	public static Stream<Traversable> getStreamOfAllChildren(Traversable traversable) {
		return StreamUtil.streamTree(traversable, CetusUtils::getStreamOfChildren);
	}

	public static <T extends Traversable> Stream<T> getStreamOfAllChildren(Traversable traversable, Class<T> onlyWith) {
		return getStreamOfAllChildren(traversable).flatMap(StreamUtil.castOrEmpty(onlyWith));
	}

	/**
	 * Predicate to see if a child is between include annotations
	 * @param <T> type
	 * @return true if the Traversable is between include annotations
	 */
	public static <T extends Traversable> Predicate<T> childInsideInclude() {
		final boolean[] insideInclude = {false};
		return t -> {
			if (FunctionCallgraph.isStartIncludePragma(t)) {
				insideInclude[0] = true;
			} else if (FunctionCallgraph.isEndIncludePragma(t)) {
				insideInclude[0] = false;
				return true;
			}
			return insideInclude[0];
		};
	}

	public static <T extends Traversable> Predicate<? super Traversable> ignoreChildrenTypes(Set<Class<? extends Traversable>> ignoreChildrenTypes) {
		return c -> !ignoreChildrenTypes.stream().anyMatch(ignoredTypes -> ignoredTypes.isAssignableFrom(c.getClass()));
	}

	public static <T extends Traversable> Predicate<? super Traversable> ignoreChildrenTypes(Class<T> ignoreChildrenType) {
		return ignoreChildrenTypes(Collections.singleton(ignoreChildrenType));
	}




	/**
	 * Returns the translation Unit from a source file
	 * @param sourceFile a source file
	 * @param model a model to operate on
	 * @return a translation unit to the respective source file
	 */
	public static TranslationUnit getTranslationUnitFromSourceFile(SourceFile sourceFile, Model model) {
		final CetusProgram program = model.getModelComponent(CetusProgram.class);
		return getStreamOfAllChildren(program.getProgram(), TranslationUnit.class)
				.filter(tu -> tu.getInputFilename().equals(sourceFile.toString()))
				.collect(StreamUtil.exactlyOne());
	}

	/**
	 * Returns the translation Unit from a source file
	 * @param file a source file
	 * @param model a model to operate on
	 * @return a translation unit to the respective source file
	 */
	public static TranslationUnit getTranslationUnitFromFile(File file, Model model) {
		final SourceFile cSourceFile = model.getModelComponents(SourceFile.class).stream()
				.filter(sf -> sf.getFile().equals(file))
				.collect(StreamUtil.exactlyOne());
		return getTranslationUnitFromSourceFile(cSourceFile, model);
	}

	/**
	 * Creates a variable
	 * @param specifiers the type
	 * @param identifier an identifier
	 * @param initExpression an initialization
	 * @return a variable declaration with the given elements
	 */
	public static VariableDeclaration createVariable(List<Specifier> specifiers, NameID identifier, Expression initExpression) {
		VariableDeclarator variableDeclarator = new VariableDeclarator(identifier);
		variableDeclarator.setInitializer(new Initializer(initExpression));
		return new VariableDeclaration(specifiers, variableDeclarator);
	}

	/**
	 * finds the global initalizer of a variable declaration in a program
	 * @param program a program to search in
	 * @param variableDeclarator a variable declaration to be found
	 * @return a global initializer of the variable declaration or null if nothing is found
	 */
	public static Initializer findGlobalInitializer(Program program, VariableDeclarator variableDeclarator) {
		final Initializer initializer = variableDeclarator.getInitializer();
		if (initializer != null) {
			return initializer;
		}
		final Declaration declaration = variableDeclarator.getDeclaration();
		if (declaration instanceof VariableDeclaration) {
			VariableDeclaration variableDeclaration = (VariableDeclaration) declaration;
			final boolean isExtern = variableDeclaration.isExtern();
			final boolean isGlobal = variableDeclaration.getParent() instanceof TranslationUnit;

			if (isExtern) {
				final Optional<Initializer> initializerOptional = program.getChildren().stream()
						.flatMap(x -> CetusUtils.streamChildrenOfType(x, VariableDeclaration.class))
						.filter(v -> isSameID(v, variableDeclaration))
						.flatMap(v -> v.getDeclarators().stream())
						.flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class))
						.filter(Objects::nonNull)
						.filter(d -> d.getSymbolName().equalsIgnoreCase(variableDeclarator.getSymbolName()))
						.map(d -> d.getInitializer())
						.filter(Objects::nonNull)
						.collect(StreamUtil.atMostOne());
				if (initializerOptional.isPresent()) {
					return initializerOptional.get();
				}
			}
		}
		return null;
	}

	/**
	 * Prunes a list of duplicate elements. This method uses == comparison stead of .equals() since Cetus is sometimes too sloppy with equals
	 * @param list a list to prune off elements
	 * @param <T> the list type
	 * @return a list containing no duplicate instances
	 */
	public static <T extends Traversable> List<T> distinctByInstance(List<T> list) {
		List<T> ret = new ArrayList<>();
		for (T element : list) {
			boolean inRet = false;
			for (T elementInRet : ret) {
				if (element == elementInRet) {
					inRet = true;
				}
			}
			if (inRet == false) {
				ret.add(element);
			}
		}
		return ret;
	}

	/**
	 * Removes the initializers from a declaration.
	 *
	 * Transforms "int i = 0;" into "int i;" for example.
	 * The returned VariableDeclaration is a new object instance, the original VariableDeclaration
	 * is not changed.
	 *
	 * @param d VariableTransformation to remove the initializer from.
	 * @param variableDeclarator the variable to return an initializer for
	 * @return A new VariableDeclaration without the initializer.
	 */
	public static VariableDeclaration getDeclarationWithoutInitializer(VariableDeclaration d, VariableDeclarator variableDeclarator) {
		VariableDeclaration decl = d.clone();
		if (decl.getNumDeclarators() == 0) {
			decl.addDeclarator(variableDeclarator);
		}
		if (decl.getDeclarators().size() > 0) {
			final List<VariableDeclarator> variableDeclarators = decl.getDeclarators().stream().flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class)).collect(Collectors.toList());
			for (VariableDeclarator declarator : variableDeclarators) {
				if (!declarator.getSymbolName().equalsIgnoreCase(variableDeclarator.getSymbolName())) {
					CetusUtils.removeChild(declarator);
				}
			}
			if (!variableDeclarators.stream().anyMatch(vd -> vd.getSymbolName().equalsIgnoreCase(variableDeclarator.getSymbolName()))) {
				decl.addDeclarator(variableDeclarator);
			}
			//log.error("More than one declaration in variableDeclaration!");

			//return null;
		}
		if (decl.getNumDeclarators() > 0) {
			decl.getDeclarator(0).getChildren().removeIf(c -> !(c instanceof NameID));
		}
		decl.getSpecifiers().removeIf(s -> s == Specifier.EXTERN);

		return decl;
	}

	/**
	 * Checks whether a Traversable only has literals as leafs.
	 *
	 * This is useful when checking if an expression can be evaluated by the compiler.
	 *
	 * Example:
	 *   1 + 2 + 3; // returns true
	 *   (2 * i) + 5; // returns false, i is not a literal
	 *
	 * @param t Traversable to check, typically an expression.
	 * @return True if the Traversable has only literals as leafs, false otherwise.
	 */
	public static boolean onlyLiteralLeafs(Traversable t) {
		boolean literal = true;

		if (t.getChildren().size() == 0) {
			literal = t instanceof Literal;
		} else {
			for (Traversable c : t.getChildren()) {
				literal = onlyLiteralLeafs(c);

				if (!literal) {
					break;
				}
			}
		}

		return literal;
	}

	/**
	 * Returns the Initializer of a Declaration if it contains only literals.
	 *
	 * Example:
	 *   int i = 1 + 2; // Returns "1 + 2"
	 *   int j = i * 3; // Returns null, i is not a literal.
	 * @param d The Declaration to process.
	 * @return A copy of the initializer of d.
	 */
	public static Initializer getLiteralInitializer(VariableDeclaration d) {
		if (d.getDeclarators().size() > 1) {
			log.error("More than one declaration in VariableDeclaration!");

			return null;
		}

		if (d.getSpecifiers().contains(Specifier.EXTERN)) {
			Initializer init = findGlobalInitializer(findParentOfType(d, Program.class), (VariableDeclarator)d.getDeclarator(0));

			if (init != null) {
				return init.clone();
			}

			return null;
		}

		Optional<Traversable> init = d.getDeclarator(0).getChildren().stream().filter(c -> c instanceof Initializer).findFirst();
		if (!init.isPresent()) {
			return null;
		}

		Initializer i = (Initializer)init.get();

		if (onlyLiteralLeafs(i)) {
			return i.clone();
		}

		return null;
	}

	/**
	 * Removes a child from the parent.
	 *
	 * @param child Child to remove.
	 */
	public static void removeChild(Traversable child) {
		Traversable parent = child.getParent();
		int index = parent.getChildren().indexOf(child);

		parent.getChildren().remove(index);
	}

	/**
	 * Replaces a child.
	 *
	 * @param old The child to be replaced.
	 * @param n The new child.
	 */
	public static void replaceChild(Traversable old, Traversable n) {
		Traversable parent = old.getParent();
		int index = parent.getChildren().indexOf(old);

		parent.getChildren().set(index, n);
	}

	public static boolean isDeclarationInStruct(VariableDeclaration variableDeclaration) {
		Traversable traversable = variableDeclaration.getParent();
		while ( traversable != null ) {
			if (traversable instanceof ClassDeclaration && ((ClassDeclaration)traversable).containsDeclaration(variableDeclaration)) {
				return true;
			}
			traversable = traversable.getParent();
		}
		return false;
	}

	public static Traversable deepClone(CloneableTraversable traversable) {
		Traversable newTraversable = (Traversable) traversable.clone();
		int i = 0;
		for (Traversable child : traversable.getChildren()) {
			Traversable newChild = deepClone((CloneableTraversable)child);
			try {
				newTraversable.setChild(i++, newChild);
			} catch (UnsupportedOperationException e) {
				log.warn(newTraversable.getClass() + " can not be cloned");
			}
		}
		return newTraversable;
	}

}
