/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.tools;

import org.apache.commons.io.FileUtils;
import org.apache.commons.io.filefilter.DirectoryFileFilter;
import org.apache.commons.io.filefilter.IOFileFilter;
import org.apache.commons.io.filefilter.RegexFileFilter;
import org.slf4j.Logger;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

/**
 * Widely used utilities.
 *
 * @author Jan Weber
 */
public final class Utils {

	private static Logger log = org.slf4j.LoggerFactory.getLogger(Utils.class);


	private Utils(){}

	/**
	 * Sees if the onject is of a specific class
	 * @param o an object
	 * @param t t
	 * @param <T> T
	 * @return true if this is implementing the class
	 */
	public static <T> boolean isImplementing(Object o, Class<T> t) {
		for (Class<?> parentType : o.getClass().getInterfaces()) {
			if (parentType.equals(t)) {
				return true;
			}
		}
		return false;
	}


	/**
	 * Find the first match according to a pattern in a string
	 * @param string string to be matched
	 * @param pattern a pattern to be matched against
	 * @param groupIndex a group index for the matcher
	 * @return the first matched string
	 */
	public static String firstMatch(String string, String pattern,
									int groupIndex) {

		Matcher matcher = Pattern.compile(pattern).matcher(string);

		if (matcher.find()) {

			return matcher.group(groupIndex);

		}

		return null;
	}

	/**
	 * Gives all matches of a pattern in a string
	 * @param string string to be matched
	 * @param pattern a pattern to be matched against
	 * @return all matched strings
	 */
	public static Collection<String> getMatches(String string, String pattern) {
		Collection<String> matches = new ArrayList<String>();

		Matcher matcher = Pattern.compile(pattern).matcher(string);

		while (matcher.find()) {
			matches.add(matcher.group());
		}

		return matches;
	}


	/**
	 * Read input stream into a string
	 * @param is input stream
	 * @return conversts a stream to a string
	 */
	public static String readStreamToString(InputStream is) {

		BufferedReader br = null;
		StringBuilder sb = new StringBuilder();

		String line;
		try {

			br = new BufferedReader(new InputStreamReader(is));
			while ((line = br.readLine()) != null) {
				sb.append(line);
			}

		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			if (br != null) {
				try {
					br.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}

		return sb.toString();
	}

	/**
	 * writes a string into a file
	 * @param f file
	 * @param string string
	 * @param overwrite content overwrite
	 */
	public static void writeStringToFile(File f, String string,
										 boolean overwrite) {

		if (f.exists() && f.isFile() && !overwrite) {
			log.warn("File already exists: " + f.getPath());
		}

		FileWriter fw = null;
		try {
			fw = new FileWriter(f, false);
			fw.write(string);
			fw.flush();
		} catch (IOException e) {
			log.warn("Unable to write File: " + f.getPath());
			log.debug(e.getMessage());
		} finally {
			if (fw != null) {
				try {
					fw.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}

	}

	/**
	 * searches files by extension in a folder
	 * @param parentDir parent
	 * @param ext file extension
	 * @param scanSubDir subdir
	 * @return All files matching the extension
	 */
	public static List<File> searchFilesByExt(File parentDir, String ext, boolean scanSubDir) {

		IOFileFilter fileFilter = null;
		if (scanSubDir) {
			fileFilter = DirectoryFileFilter.DIRECTORY;
		}

		String fileExtRegex = "^(.*)\\." + ext + "$";
		return FileUtils.listFiles(
				parentDir,
				new RegexFileFilter(fileExtRegex),
				fileFilter)
				.stream()
				.sorted()
				.collect(Collectors.toList());
	}
}
