/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.tools;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.interfaces.AccessesVariables;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.VariableAccessCreator;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.AccessExpression;
import cetus.hir.ArrayAccess;
import cetus.hir.AssignmentExpression;
import cetus.hir.AssignmentOperator;
import cetus.hir.BinaryExpression;
import cetus.hir.CompoundStatement;
import cetus.hir.DFIterator;
import cetus.hir.Declaration;
import cetus.hir.Declarator;
import cetus.hir.Expression;
import cetus.hir.FunctionCall;
import cetus.hir.IDExpression;
import cetus.hir.Identifier;
import cetus.hir.IfStatement;
import cetus.hir.Initializer;
import cetus.hir.NameID;
import cetus.hir.PointerSpecifier;
import cetus.hir.Procedure;
import cetus.hir.ReturnStatement;
import cetus.hir.SwitchStatement;
import cetus.hir.Symbol;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.UnaryExpression;
import cetus.hir.UnaryOperator;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Modifies&analyzes Variable Accesses
 */
public class VariableAccessAnalyzer {

	private final Logger log = LoggerFactory.getLogger(VariableAccessAnalyzer.class);
	private final Model model;

	public VariableAccessAnalyzer(Model model) {
		this.model = model;
	}

	/**
	 * set the access behaviour for all variable accesses
	 * @param variableAccesses variable accesses to scan for variable accesses
	 */
	public void setAccessBehaviour(Collection<VariableAccess> variableAccesses) {
		for (VariableAccess variableAccess : variableAccesses) {

			/*
			 * Determine whether the variable is read, written or both in the
			 * context of the current identifier.
			 */
			if (variableAccess.getBehavior() == null) {
				VariableAccess.AccessBehavior accessbehavior = determineAccessBehavior(variableAccess
						.getIdentifier());
				variableAccess.setBehavior(accessbehavior);
			}
		}
	}

	/**
	 * Determines the access type of the given {@link Identifier}.
	 *
	 * @param identifier {@link Identifier} whose access behavior shall be determined.
	 * @return The {@link VariableAccess.AccessBehavior} of the given {@link Identifier}.
	 */
	public VariableAccess.AccessBehavior determineAccessBehavior(Identifier identifier) {
		/* traversable to walk through identifier's parents. */
		Traversable parent = identifier.getParent();



		VariableAccess.AccessBehavior ab = getAccessBehaviourOfFunctionCall(parent, identifier);
		if (ab != null) {
			return ab;
		}

		/* store all parents */
		List<Traversable> history = new ArrayList<>();
		history.add(identifier);

		/* scan the identifiers context to determine access behavior. */
		while (null != parent) {
			if (parent instanceof Initializer) {
				/* e.g. int foo = i; -> r access (TODO: other side?) */
				log.debug("found initialisation " + identifier.toString());
				if (identifier.getSymbol().getArraySpecifiers().size() != 0) {
					return VariableAccess.AccessBehavior.READWRITE;
				} else {
					return VariableAccess.AccessBehavior.READ;
				}
			} else if (parent instanceof UnaryExpression) {
				/* if post-/pre- de-/increment -> r/w access */
				final UnaryOperator operator = ((UnaryExpression) parent).getOperator();
				if (UnaryOperator.hasSideEffects(operator) || operator.equals(UnaryOperator.ADDRESS_OF)) {
					return VariableAccess.AccessBehavior.READWRITE;
				}
			} else if (parent instanceof AssignmentExpression) {
				AssignmentExpression assignmentExpression = (AssignmentExpression) parent;
				if (assignmentExpression.getLHS() == history.get(history.size() - 1)) {
					if (assignmentExpression.getRHS().getChildren().contains(identifier) || assignmentExpression.getOperator() != AssignmentOperator.NORMAL) {
						/* if identifier is on left and right handside -> rw access */
						return VariableAccess.AccessBehavior.READWRITE;
					} else {
						/* if identifier on left hand side -> w access */
						return VariableAccess.AccessBehavior.WRITE;
					}
				} else {
					/* if identifier on right hand side -> r access */
					return VariableAccess.AccessBehavior.READ;
				}
			} else if (parent instanceof AccessExpression) {
				//go to next parent

			} else if (parent instanceof BinaryExpression) {
				/*
				 * if one of these operators +, &, ^, |, ==, >=, >, <=, <, !=,
				 * /, &&, ||, %, *, <<, >>, - -> r access. //TODO: is it always?
				 * -> & in function call maybe write
				 */
				return VariableAccess.AccessBehavior.READ;
			} else if (parent instanceof ArrayAccess) {
				for (Traversable indice : ((ArrayAccess) parent).getIndices()) {
					if (indice == history.get(history.size() - 1)) {
						/* if identifier part of array indice -> r access */
						return VariableAccess.AccessBehavior.READ;
					}
				}
				/* ... if not continue. */
			} else if (parent instanceof IfStatement) {
				if (!(history.get(history.size() - 1) instanceof CompoundStatement)) {
					return VariableAccess.AccessBehavior.READ;
				}
			} else if (parent instanceof SwitchStatement) {
				if (!(history.get(history.size() - 1) instanceof CompoundStatement)) {
					return VariableAccess.AccessBehavior.READ;
				}
			} else if (parent instanceof ReturnStatement) {
				return VariableAccess.AccessBehavior.READ;
			} else if (parent instanceof FunctionCall) {
				return getAccessBehaviourOfFunctionCall(parent, identifier);
			}

			history.add(parent);
			parent = parent.getParent();
		}

		log.warn("Could not determine variable access type of " + identifier);

		return VariableAccess.AccessBehavior.UNKNOWN;
	}

	/**
	 * Checks whether there is already a {@link Variable} for the given
	 * identifier stored in the current {@link Model}. If there is a
	 * {@link Variable}, the {@link Identifier} is added to the variable and it
	 * is returned, otherwise a new variable is created, stored within the model
	 * and within the given {@link AccessesVariables} and returned.
	 *
	 * @param identifier The identifier containing the variable.
	 * @param accessesvariable The {@link AccessesVariables} the variable should be stored in.
	 * @param globalVariables the global variables
	 * @param allVariables all variables
	 * @return The {@link Variable}
	 */
	public Variable getVariable(Identifier identifier,
										AccessesVariables accessesvariable, Collection<Identifier> globalVariables, Collection<Variable> allVariables) {
		Symbol symbol = identifier.getSymbol();

		if (symbol instanceof VariableDeclarator /*
		 * Add other types if
		 * nessecary...
		 */) {

			/*
			 * Check whether the variable was already created
			 */
			VariableDeclarator variableDeclarator = (VariableDeclarator) symbol;
			for (Variable variable : allVariables) {
				if (variable.getVariableDeclarator().equals(variableDeclarator)) {
					/* if so add it to the current task if nesseccary ... */
					if (!accessesvariable.getVariables().contains(variable)) {
						accessesvariable.addVariable(variable);
					}
					variable.addIdentifier(identifier);
					return variable;
				}
			}
			/* Otherwise create it ... */
			if (variableDeclarator.getDeclaration() instanceof VariableDeclaration && CetusUtils.isDeclarationInStruct((VariableDeclaration)variableDeclarator.getDeclaration())) {
				return  null;
			}
			Variable variable = new Variable(variableDeclarator, identifier);

			/* a very simple detection of global variables, and warning */
			Declaration declaration = identifier.findDeclaration();
			if (null != declaration) {
				Traversable parent = declaration.getParent();
				while (null != parent) {
					if (parent instanceof Procedure) {
						break;
					}
					if (parent instanceof TranslationUnit) {
						variable.setGlobal(true);
						if (!globalVariables.contains(identifier)) {
							globalVariables.add(identifier);
						}
					}
					parent = parent.getParent();
				}

			}

			/* and add it to the current task */
			accessesvariable.addVariable(variable);
			//model.addModelComponent(variable);
			allVariables.add(variable);
			variable.addIdentifier(identifier);

			return variable;
		}

		return null;
	}

	/**
	 * Checks if the identifier is a pointer or is used as a pointer with the
	 * given functioncall
	 *
	 * @param traversable a traversable
	 * @param identifier an identifier
	 * @return TRUE: Is the identifier is used as a pointer or is a pointer itself FALSE: otherwise
	 */
	private boolean isIdentifierUsedAsPointerinFunctionCall(
			Traversable traversable, Identifier identifier) {
		Traversable parent = identifier.getParent();
		while (parent != null) {
			if (parent instanceof FunctionCall) {
				FunctionCall functionCall = (FunctionCall) parent;
				/*
				 * If it is an array access on the pointer (without [0] for
				 * example)
				 */
				if (!identifier.getSymbol().getArraySpecifiers().isEmpty()) {
					for (Object argument : functionCall.getArguments()) {
						if (argument instanceof Expression) {
							Expression e = (Expression) argument;
							if (e != null
									&& !e.findExpression(identifier).isEmpty()) {
								// for( Expression exp :
								// e.findExpression(identifier) ){
								DFIterator<Identifier> iter = new DFIterator<Identifier>(
										e, Identifier.class);
								while (iter.hasNext()) {
									Identifier identifier1 = iter.next();
									if (CetusUtils.compareIdentifiersAncestors(identifier1, identifier)) {
										if (identifier1.getParent() != null && !(identifier1.getParent() instanceof ArrayAccess)) {
											return true;
										}
									}
								}
							}
						}

					}
				}
				/*
				 * If the identifier itself is of a pointer type
				 */
				Set<Integer> argumentPositions = new HashSet<>();
				for (Object specifier : identifier.getSymbol()
						.getTypeSpecifiers()) {
					if (specifier instanceof PointerSpecifier) {
						return true;
					}
				}
				/*
				 * If the variable is dereferenced (foo(&variable))
				 */
				for (Object argument : functionCall.getArguments()) {
					if (argument instanceof Expression) {
						Expression e = (Expression) argument;
						if (e != null
								&& !e.findExpression(identifier).isEmpty()) {
							for (Expression exp : e.findExpression(identifier)) {
								if (exp.equals(identifier)) {
									DFIterator<UnaryExpression> iterator = new DFIterator<UnaryExpression>(
											e, UnaryExpression.class);
									while (iterator.hasNext()) {
										UnaryExpression ue = iterator.next();
										DFIterator<Identifier> iter = new DFIterator<Identifier>(
												ue, Identifier.class);
										while (iter.hasNext()) {
											Identifier id = iter.next();
											if (id.equals(identifier)) {
												return true;
											}
										}
									}
								}
							}
						}
					}
				}

				/*
				 * If the variable is none of the above, but the function
				 * expects a pointer at this position
				 *
				 * Edit: It is hard to determine which variable is
				 * meant...arbitrary mem access
				 */
				/*
				 * if
				 * (functionCall.getProcedure()!=null&&functionCall.getProcedure
				 * ()!=null) { Set<Integer>
				 * positions=getArgumentPositions(functionCall, identifier);
				 * Integer counter=-1; for (Object argument :
				 * functionCall.getProcedure().getParameters()) { counter++; if
				 * (argument instanceof VariableDeclaration) { for (Symbol
				 * symbol : ((VariableDeclaration)
				 * argument).getDeclaredSymbols()) { for (Object spec :
				 * symbol.getTypeSpecifiers()){ if (spec instanceof
				 * PointerSpecifier){ if (positions.contains(counter)){ return
				 * true; } } } } } } }
				 */
			}
			parent = parent.getParent();
		}
		return false;
	}

	/**
	 * Returns the access behaviour of a function call
	 *
	 * @param traversable any traversable containing a function call
	 * @param identifier the identifier to be found
	 * @return The determined access behaviour null if no function call has been found
	 */
	private VariableAccess.AccessBehavior getAccessBehaviourOfFunctionCall(
			Traversable traversable, Identifier identifier) {
		if (isIdentifierUsedAsPointerinFunctionCall(traversable, identifier)) {
			return VariableAccess.AccessBehavior.READWRITE;
		} else if (traversable instanceof FunctionCall) {
			return VariableAccess.AccessBehavior.READ;
		}
		return null;
	}


	private Variable getKnownVariable(IDExpression variableName, AccessesVariables accessesvariable, Collection<Variable> allVariables) {
		Collection<Variable> foundVars = allVariables.stream()
				.filter(v -> v.getVariableDeclarator() instanceof VariableDeclarator)
				.filter(v -> v.getVariableDeclarator().getDeclaration() == variableName.findDeclaration())
				.collect(Collectors.toList());
		Optional<Variable> optional;
		if (foundVars.size() > 1) {
			optional = foundVars.stream()
					.filter(x -> x.equals(variableName))
					.collect(StreamUtil.atMostOne());
		} else {
			optional = foundVars.stream().collect(StreamUtil.atMostOne());
		}

		Variable variable = null;
		if (optional.isPresent()) {
			variable = optional.get();
			if (!accessesvariable.getVariables().contains(variable)) {
				accessesvariable.addVariable(variable);
			}
		}
		return variable;
	}


	/**
	 * Scans the {@link Traversable} of the given {@link AccessesVariables}
	 * accessesVariables for Variables respectively {@link Identifier
	 * Identifiers} (Cetus IR) and generates {@link Variable} and
	 * {@link VariableAccess} {@link ModelComponent ModelComponents}, which are
	 * added to the {@link Model}.
	 * <p>
	 * {@link VariableAccess VariableAccesses} are also stored in the field
	 * {@link VariableAccessCreator} for determining the
	 * {@link VariableAccess.AccessBehavior} later on.
	 *
	 * @param accessesvariables the {@link AccessesVariables} that shall be scanned.
	 * @param globalVariables all global variables
	 * @param allVariables all variables that exist
	 * @return a list of all variable accesses and additionally allVariables might have changed
	 */
	public List<VariableAccess> scanForIdentifier(AccessesVariables accessesvariables, Collection<Identifier> globalVariables, Collection<Variable> allVariables) {

		List<VariableAccess> variableAccesses = new ArrayList<>();

		DFIterator<Identifier> iter = new DFIterator<Identifier>(
				accessesvariables.getTraversable(), Identifier.class);

		while (iter.hasNext()) {
			Identifier identifier = iter.next();

			if (CetusUtils.contains(identifier, accessesvariables.getTraversable())) {
				Variable variable = getVariable(identifier, accessesvariables, globalVariables, allVariables);
				if (null != variable) {
					boolean variableLifesInChild = false;
					if (accessesvariables instanceof Task) {
						Task t = (Task) accessesvariables;
						for (Task childtask : t.getContainedTasks()) {
							if (CetusUtils.contains(identifier,
									childtask.getTraversable())) {
								variableLifesInChild = true;
							}
						}
					}
					log.debug("creating variable access for " + variable);
					VariableAccess access;
					if (variableLifesInChild) {
						access = new VariableAccess(accessesvariables,
								identifier, variable, false);
						access.setBehavior(VariableAccess.AccessBehavior.UNTOUCHED);
					} else {
						access = new VariableAccess(accessesvariables,
								identifier, variable, false);
					}
					variableAccesses.add(access);
					accessesvariables.addVariableAccess(access);
				}

			}
		}

		DFIterator<VariableDeclaration> it = new DFIterator<VariableDeclaration>(
				accessesvariables.getTraversable(), VariableDeclaration.class);
		while (it.hasNext()) {
			VariableDeclaration dec = it.next();
			final boolean declaredInsideOtherTask = model.getModelComponents(Task.class).stream()
					.filter(t -> !t.equals(accessesvariables))
					.flatMap(t -> CetusUtils.getStreamOfAllChildren(t.getTraversable(), VariableDeclaration.class))
					.anyMatch(vd -> vd.equals(dec));
			if (!declaredInsideOtherTask) {
				for (int i = 0; i < dec.getNumDeclarators(); i++) {
					Declarator d = dec.getDeclarator(i);
					Symbol symbol = dec.getDeclaredSymbols().get(i);
					if (d.getInitializer() != null) {
						if (symbol instanceof VariableDeclarator) {
							IDExpression name = ((VariableDeclarator) symbol).getID();
							Variable var = getKnownVariable(name, accessesvariables, allVariables);
							if (var != null && name instanceof NameID) {
								VariableAccess access = new VariableAccess(accessesvariables, (NameID) name, var, false);
								access.setBehavior(VariableAccess.AccessBehavior.WRITE);
								variableAccesses.add(access);
								accessesvariables.addVariableAccess(access);
							}

						}
					}
				}
			}
		}

		return variableAccesses;

	}

}
