/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.microStreams.tools;

import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.FunctionCallgraph;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.ArraySpecifier;
import cetus.hir.ClassDeclaration;
import cetus.hir.Declarator;
import cetus.hir.DepthFirstIterator;
import cetus.hir.EvaluateExpression;
import cetus.hir.IntegerLiteral;
import cetus.hir.Literal;
import cetus.hir.PointerSpecifier;
import cetus.hir.Program;
import cetus.hir.Specifier;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.UserSpecifier;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.math.BigInteger;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Helper methods to determine the size of Variables
 */
public final class VariableUtils {

	public static final int BYTES_PER_WORD = 2;
	private static final Logger log = LoggerFactory.getLogger(VariableUtils.class);
	private static final String STRUCT_STRING = "struct ";

	private VariableUtils() {

	}

	/**
	 * Returns the data size of this variable in terms of data/word size of SpartanMC 18 bit words if the CoreConnector functions
	 * for sending and receiving data are used.
	 *
	 * @param var variable whose size in spmc words to determine
	 * @param program the current program to analyze
	 * @return spartanMC words necessary to send this variable
	 */
	public static int getSpartanMCWordsCount(Variable var, Program program) {
		int sizeInBytes = getSizeInSpartanMCBytes(var);
		return getSpartanMCWordsCount(sizeInBytes);
	}

	/**
	 * Returns the data size of this variable in terms of data/word size of SpartanMC 18 bit words if the CoreConnector functions
	 * for sending and receiving data are used.
	 *
	 * @param sizeInBytes variable size in bytes
	 * @return spartanMC words necessary to send this variable
	 */
	public static int getSpartanMCWordsCount(int sizeInBytes) {
		return (sizeInBytes / BYTES_PER_WORD) + (sizeInBytes % BYTES_PER_WORD);
	}

	/**
	 * Returns the size of a variable in bytes.
	 *
	 * @param var variable
	 * @return size of the variable in SpartanMC bytes
	 */
	public static int getSizeInSpartanMCBytes(Variable var) {
		return getSizeInSpartanMCBytes(var.getVariableDeclarator());
	}

	/**
	 * This algorithm analyses the size of a variable recursively and distinguishes between different types such as
	 * primitives, struct, arrays or typedef-types.
	 *
	 * @param varDecl the declaration of the variable
	 * @return the amount of bytes used for this variable
	 */
	public static int getSizeInSpartanMCBytes(VariableDeclarator varDecl) {

		List<Specifier> specifiers = varDecl.getTypeSpecifiers();
		int typeSize = 0;

		//first case - the type of the variable is user-defined, i.e. there must be a typedef statement defining it
		if (isTypeDef(varDecl)) {
			Optional<UserSpecifier> typeSpecifier = specifiers.stream().flatMap(StreamUtil.castOrEmpty(UserSpecifier.class)).findFirst();
			//find declaration where the user type is defined
			if (typeSpecifier.isPresent()) {
				VariableDeclarator vd = findTypeDeclaration(typeSpecifier.get(), varDecl);
				ClassDeclaration cd = findStructClassDeclaration(typeSpecifier.get(), varDecl);
				if (vd != null) {
					typeSize = getSizeInSpartanMCBytes(vd);
				} else if (cd != null) {
					typeSize = getSizeInSpartanMCBytes(cd);
				} else {
					log.warn("No type declaration found for " + varDecl + "!");
					return -1;
				}
			} else {
				typeSize = getSpartanMCSizeFromSpecifiers(specifiers);
			}

		}

		//second case - variable is a struct (not using an alias): there must be a class declaration defining its members
		else if (isStruct(varDecl)) {
			UserSpecifier typeSpecifier = specifiers.stream().flatMap(StreamUtil.castOrEmpty(UserSpecifier.class)).findFirst().get();
			//isolate class name from "struct "

			ClassDeclaration structClass = findStructClassDeclaration(typeSpecifier, varDecl);
			if (structClass == null) {
				log.warn("No struct class declaration found for struct " + varDecl.toString() + "!");
				return -1;
			}
			typeSize = getSizeInSpartanMCBytes(structClass);
		}

		//third case - a user type is defined here (declaration begins with "typedef")
		else if (isTypeDef(varDecl)) {
			List<Specifier> underlyingType = specifiers.subList(1, specifiers.size());
			final int size = getSpartanMCSizeFromSpecifiers(underlyingType);
			if (size == 0) {
				//not a primitive type - "renamed" type is EITHER a struct or a typedef user-type (however, must be a UserSpecifier in either way)
				final UserSpecifier underlying = underlyingType.stream().flatMap(StreamUtil.castOrEmpty(UserSpecifier.class)).collect(StreamUtil.exactlyOne());
				if (underlying.isStructure()) {
					//find structure class declaration
					ClassDeclaration structClass = findStructClassDeclaration(underlying, varDecl);
					if (structClass == null) {
						log.warn("No underlying struct class declaration found for type declaration " + varDecl.toString() + "!");
						return -1;
					}
					typeSize = getSizeInSpartanMCBytes(structClass);
				} else {
					//find declaration of the underlying type
					VariableDeclarator userTypeDeclaration = findTypeDeclaration(underlying, varDecl);
					if (userTypeDeclaration == null) {
						log.warn("No underlying type declaration found for type declaration " + varDecl.toString() + "!");
						return -1;
					}
					typeSize = getSizeInSpartanMCBytes(userTypeDeclaration);
				}
			} else {
				//primitive type
				typeSize = size;
			}
		}

		//fourth case - variable if of a primitive type (DataTypeC)
		else if (isPrimitive(varDecl)) {

			final List<Specifier> typeSpecifiers = varDecl.getTypeSpecifiers();
			typeSize = getSpartanMCSizeFromSpecifiers(typeSpecifiers);
		} else if (isExtern(varDecl)) {
			final Program program = CetusUtils.findParentOfType(varDecl, Program.class);
			final TranslationUnit translationUnit = CetusUtils.findParentOfType(varDecl, TranslationUnit.class);
			final String identifier = varDecl.getID().toString();
			final VariableDeclarator extVarDecl = CetusUtils.getStreamOfAllChildren(program, TranslationUnit.class)
					.filter(t -> !t.equals(translationUnit))
					.flatMap(t -> CetusUtils.streamChildrenOfType(t, VariableDeclaration.class))
					.flatMap(vd -> vd.getDeclarators().stream())
					.flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class))
					.filter(vd -> vd.getID().toString().equals(identifier))
					.filter(vd -> !isExtern(vd))
					.collect(StreamUtil.exactlyOne());
			return getSizeInSpartanMCBytes(extVarDecl);
		}

		//error
		else {
			log.warn("Size could not be determined for the declaration " + varDecl.toString() + "!");
			return -1;
		}

		//additional to the type, the variable of the declaration can be an array
		if (isArray(varDecl)) {
			//multiply the size of the data type with the total number of entries
			int arrLength = getLength(varDecl);
			if (arrLength < 0) {
				typeSize = -1;
			} else {
				typeSize *= arrLength;
			}
		}

		//end result should contain the correct size
		return typeSize;
	}

	/**
	 * NOTE: Only apply this on struct class declarations.
	 *
	 * @param structClass the structure class
	 * @return the size in bytes
	 */
	public static int getSizeInSpartanMCBytes(ClassDeclaration structClass) {
		int typeSize = 0;
		//recursive calls for VariableDeclarations which are members of the struct class
		List<VariableDeclarator> members = getMemberVariableDeclarators(structClass);
		for (VariableDeclarator curr : members) {
			final int size = getSizeInSpartanMCBytes(curr);
			if (size >= BYTES_PER_WORD && typeSize % BYTES_PER_WORD != 0) {
				typeSize = typeSize + BYTES_PER_WORD - (typeSize % BYTES_PER_WORD);
			}
			typeSize += size;
		}
		//pad struct size to full words
		if (typeSize % BYTES_PER_WORD != 0) {
			typeSize = typeSize + BYTES_PER_WORD - (typeSize % BYTES_PER_WORD);
		}
		return typeSize;
	}

	public static boolean isExtern(VariableDeclarator varDecl) {
		final List<Specifier> typeSpecifiers = varDecl.getTypeSpecifiers();
		return typeSpecifiers.stream().anyMatch(s -> s.equals(Specifier.EXTERN));
	}

	/**
	 * Returns the {@link VariableDeclarator} to a given userSpecifier.
	 * Traverses the preceding traversables of start, from inner scopes to outer ones.
	 *
	 * @param userSpecifier the userSpecifier of the declaration to find
	 * @param start starting point, predecessors will be traversed
	 * @return the VariableDeclarator to userSpecifier, or null if no declaration can be found in the range of searching (start to end)
	 */
	public static VariableDeclarator findTypeDeclaration(UserSpecifier userSpecifier, Traversable start) {
		Traversable parent = start.getParent();
		if (parent == null) {
			//no variable declaration found
			return null;
		}
		final Optional<VariableDeclarator> ret = CetusUtils.streamChildrenOfType(parent, VariableDeclaration.class)
				.flatMap(vd -> vd.getDeclarators().stream())
				.flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class))
				.filter(vd -> isTypeDef(vd))
				.filter(vd -> vd.toString().equals(userSpecifier.toString()))
				.collect(StreamUtil.atMostOne());
		if (ret.isPresent()) {
			return ret.get();
		}
		return findTypeDeclaration(userSpecifier, parent);
	}

	/**
	 * Returns the {@link ClassDeclaration} to a given class name.
	 * Traverses the preceding traversables of start, from inner scopes to outer ones.
	 *
	 * @param className the class name of the declaration to find
	 * @param start starting point, predecessors will be traversed
	 * @return the ClassDeclaration to the class name, or null if no declaration can be found in the range of searching (start to end)
	 */
	public static ClassDeclaration findStructClassDeclaration(UserSpecifier className, Traversable start) {
		Traversable parent = start.getParent();
		if (parent == null) {
			//no variable declaration found
			return null;
		}
		final Optional<ClassDeclaration> classDeclaration = CetusUtils.streamChildrenOfType(parent, ClassDeclaration.class)
				.filter(c -> c.getKey().equals(ClassDeclaration.STRUCT))
				.filter(c -> c.getName().toString().equals(className.toString().replace(STRUCT_STRING, "")))
				.findFirst();
		if (classDeclaration.isPresent()) {
			return classDeclaration.get();
		}
		return findStructClassDeclaration(className, parent);
	}

	/**
	 * Returns the member variable declarations of a given struct declaration.
	 *
	 * @param classDecl struct declaration
	 * @return member variable declarations
	 */
	public static List<VariableDeclarator> getMemberVariableDeclarators(ClassDeclaration classDecl) {
		return CetusUtils.getStreamOfAllChildren(classDecl, VariableDeclarator.class).collect(Collectors.toList());

		/*classDecl.getChildren().stream()
		.flatMap(c -> CetusUtils.streamChildrenOfType(c, VariableDeclaration.class))
		.flatMap(vd -> vd.getVariableDeclarators().stream())
		.distinct()
		.collect(Collectors.toList());*/
	}

	/**
	 * Checks if the variable in this declaration is of a primitive data type (or of an array of those).
	 *
	 * @param varDecl the declaration
	 * @return true if and only if the type is primitive
	 */
	public static boolean isPrimitive(VariableDeclarator varDecl) {

		final List<Specifier> typeSpecifiers = varDecl.getTypeSpecifiers();
		int typeSize = getSpartanMCSizeFromSpecifiers(typeSpecifiers);
		return typeSize > 0;
	}

	/**
	 * Checks if the declared variable is an array (may be of any type: primitive, struct, user-type...).
	 *
	 * @param varDecl the variable declaration
	 * @return true if and only if variable is an array
	 */
	public static boolean isArray(VariableDeclarator varDecl) {
		return (getArraySpecifiers(varDecl).size() > 0);
	}

	/**
	 * Checks if the declared variable is an array (may be of any type: primitive, struct, user-type...).
	 *
	 * @param declarator the declarator
	 * @return true if and only if the declarator is a variabledeclaration and an array
	 */
	public static boolean isArray(Declarator declarator) {
		if (declarator instanceof VariableDeclarator) {
			return isArray((VariableDeclarator) declarator);
		}
		return false;
	}

	public static boolean isArray(Variable var) {
		return isArray(var.getVariableDeclarator());
	}

	public static boolean isPointer(Variable var) {
		return isPointer(var.getVariableDeclarator());
	}

	public static boolean isPointer(VariableDeclarator var) {
		return var.getSpecifiers().stream().map(StreamUtil.castOrEmpty(PointerSpecifier.class)).findAny().isPresent();

	}

	/**
	 * Tells if the Variable Declaration is of struct type
	 *
	 * @param varDecl a variable declaration
	 * @return true if the variable declaration has a struct type
	 */
	public static boolean isStruct(VariableDeclarator varDecl) {

		Optional<UserSpecifier> s = varDecl.getTypeSpecifiers().stream()
				.flatMap(StreamUtil.castOrEmpty(UserSpecifier.class))
				.collect(StreamUtil.atMostOne());
		if (s.isPresent()) {
			return s.get().isStructure();
		}
		return false;
	}

	public static boolean isStruct(Variable var) {
		return isStruct(var.getVariableDeclarator());
	}


	//TODO: unterscheiden zwischen typedef und istypedeclaration

	/**
	 * Determines if the declarated variable is of a user-defined type (typedef-type).
	 *
	 * @param varDecl the declaration
	 * @return true if and only if the type is user-defined and no struct
	 */
	public static boolean isTypeDef(VariableDeclarator varDecl) {

		final List<Specifier> typeSpecifiers = varDecl.getTypeSpecifiers();
		boolean hasTypedefSpecifier = typeSpecifiers.stream().anyMatch(s -> s.equals(Specifier.TYPEDEF));
		boolean hasOnlyUserSpecifiers = typeSpecifiers.stream().filter(s -> !(s instanceof PointerSpecifier)).allMatch(s -> s instanceof UserSpecifier);
		if (hasTypedefSpecifier) {
			return true;
		} else if (hasOnlyUserSpecifiers) {
			return !isStruct(varDecl);
		}
		return false;
	}

	/**
	 * Determines if the declarated variable is of a user-defined type (typedef-type).
	 *
	 * @param var variable to examine
	 * @return true if and only if var is user-defined and no struct
	 */
	public static boolean isTypeDef(Variable var) {
		return isTypeDef(var.getVariableDeclarator());
	}

	/**
	 * Determines if the declaration describes a type declaration.
	 *
	 * @param varDecl declaration
	 * @return true if and only if the statement is a typedef statement (begins with "typedef")
	 */
	@Deprecated
	public static boolean isTypeDeclaration(VariableDeclarator varDecl) {

		final List<Specifier> typeSpecifiers = varDecl.getTypeSpecifiers();
		final boolean hasTypeSpecifiers = !typeSpecifiers.isEmpty();
		return (hasTypeSpecifiers && typeSpecifiers.get(0) == Specifier.TYPEDEF);
	}

	/**
	 * Returns the type specifiers of a variable which contain information if a variable is an int, float, etc.
	 * Can also be a "compound" type like "long long" etc.
	 *
	 * @param var the variable whose type to determine
	 * @return the specifiers determining the type
	 */
	public static List<Specifier> getTypeSpecifiers(Variable var) {
		final List typeSpecifiers = var.getVariableDeclarator().getTypeSpecifiers();
		if (typeSpecifiers.isEmpty()) {
			log.warn("No type specifiers found for variable: " + var);
			return null;
		}
		return typeSpecifiers;
	}

	/**
	 * Gives a list of Array specifiers from a variable declaration
	 *
	 * @param varDecl a variable declaration whose array specifiers should be found
	 * @return all array specifiers or an empty list if none exist
	 */
	public static List<ArraySpecifier> getArraySpecifiers(VariableDeclarator varDecl) {
		return varDecl.getArraySpecifiers().stream()
				.flatMap(StreamUtil.castOrEmpty(ArraySpecifier.class))
				.collect(Collectors.toList());
	}

	public static VariableDeclaration getVariableDeclaration(Variable var) {
		return ((VariableDeclaration) var.getVariableDeclarator().getDeclaration());
	}

	public static int getLength(Variable var) {
		return getLength(var.getVariableDeclarator());
	}

	/**
	 * Returns the length of the given array variable. If the array is multidimensional,
	 * the single dimension sizes are multiplied with each other in order to get the total length.
	 * ONLY apply this method on arrays.
	 *
	 * @param varDecl the declaration of the array
	 * @return count of array fields, -1 if var is not an array variable
	 */
	public static int getLength(VariableDeclarator varDecl) {
		if (isArray(varDecl)) {
			ArraySpecifier specifier = getArraySpecifiers(varDecl).stream().collect(StreamUtil.exactlyOne());
			if (specifier.getDimensions().contains(null)) {
				final Optional<Integer> arrayLengthFromInitializer = getArrayLengthFromInitializer(varDecl);
				if (arrayLengthFromInitializer.isPresent()) {
					return arrayLengthFromInitializer.get();
				}
			} else {
				final Optional<Integer> lengthFromDimensionFields = getLengthFromDimensionFields(specifier, false);
				if (lengthFromDimensionFields.isPresent()) {
					return lengthFromDimensionFields.get();
				} else {
					return -1;
				}
			}
		} else {
			log.warn("Variable " + varDecl + " is not an array!");
		}
		return -1;
	}

	private static Optional<Integer> getArrayLengthFromInitializer(VariableDeclarator varDecl) {
		if (isArray(varDecl)) {
			if (varDecl.getInitializer() != null) {
				final int count = (int) CetusUtils.getStreamOfAllChildren(varDecl.getInitializer(), Literal.class).count();
				if (count != 0) {
					return Optional.of(Integer.valueOf(count));
				}
			}
		}
		return Optional.empty();
	}

	private static int nullDimensionCount(ArraySpecifier specifier) {
		return (int) specifier.getDimensions().stream().filter(e -> e == null).count();
	}

	private static Optional<Integer> getLengthFromDimensionFields(ArraySpecifier specifier, boolean ignoreEmptyDimensionFields) {
		final long nullElements = nullDimensionCount(specifier);
		if (nullElements == 0 || nullElements == 1 && ignoreEmptyDimensionFields) {
			final List<BigInteger> values = specifier.getDimensions().stream()
					.filter(e -> e != null)
					.map(d -> d.visit(new EvaluateExpression()))
					.peek(v -> {
						if (!v.isPresent()) {
							log.error("Can not resolve dimensions for specifier " + specifier);
						}
					})
					.flatMap(o -> o.isPresent()? Stream.of(o.get()) : Stream.empty())
					.collect(Collectors.toList());
			final boolean valueForEachDiminsionPresent = values.size() == specifier.getNumDimensions();
			if (!values.isEmpty() && (ignoreEmptyDimensionFields || valueForEachDiminsionPresent)) {
				final int len = values.stream()
						.reduce(BigInteger.ONE, BigInteger::multiply)
						.intValue();
				return Optional.of(Integer.valueOf(len));
			}
		} else {
			log.warn("Can not analyze dimensions with more than one dimension missing");
		}
		return Optional.empty();
	}

	/**
	 * Read the dimensions of an array into a List while beeing able to add missing information in the array declaration (e.g. int myarray[]) from the initializer size
	 *
	 * @param varDecl an array variable declaration
	 * @return a list of array dimensions
	 */
	public static List<Integer> getDimenstionsOfArray(VariableDeclarator varDecl) {
		if (isArray(varDecl)) {
			ArraySpecifier specifier = getArraySpecifiers(varDecl).stream().collect(StreamUtil.exactlyOne());

			final Optional<Integer> lengthFromDimensionFields = getLengthFromDimensionFields(specifier, true);
			Optional<Integer> arrayLengthFromInitializer = getArrayLengthFromInitializer(varDecl);
			if (!arrayLengthFromInitializer.isPresent() && ((VariableDeclaration)varDecl.getDeclaration()).isExtern()) {
				final Optional<VariableDeclarator> alternativeVariableDeclarator = VariableUtils.getAlternativeVariableDeclarator(varDecl);
				if (alternativeVariableDeclarator.isPresent()) {
					arrayLengthFromInitializer = getArrayLengthFromInitializer(alternativeVariableDeclarator.get());
				}
			}


			if (!lengthFromDimensionFields.isPresent() && !arrayLengthFromInitializer.isPresent() && nullDimensionCount(specifier) != 0) {
				return null;
			}


			final Optional<Integer> finalArrayLengthFromInitializer = arrayLengthFromInitializer;
			final List<Integer> dimensions = specifier.getDimensions().stream()
					.map(d -> {
						if (d == null) {
							Integer divider;
							if (!lengthFromDimensionFields.isPresent() && nullDimensionCount(specifier) == 1 && specifier.getNumDimensions() == 1) {
								divider = new Integer(1);
							} else {
								divider = lengthFromDimensionFields.get();
							}
							if (!finalArrayLengthFromInitializer.isPresent()) {
								return null;
							}
							Integer missingDimension = finalArrayLengthFromInitializer.get() / divider;
							return missingDimension;
						} else {
							return new Integer(d.visit(new EvaluateExpression()).get().intValue());
						}
					})
					.collect(Collectors.toList());

			return dimensions;
		}
		return null;
	}

	/**
	 * Add missing dimensions to a array if possible
	 * @param variableDeclarator the variable declaration tp be expanded
	 */
	public static void extractMissingArrayDimensions(VariableDeclarator variableDeclarator) {
		final List<Integer> dimenstionSizes = getDimenstionsOfArray(variableDeclarator);
		if (dimenstionSizes == null || dimenstionSizes.contains(null)) {
			return;
		}
		final List<IntegerLiteral> dimensionSizesAsExpression = dimenstionSizes.stream().map(i -> new IntegerLiteral(i)).collect(Collectors.toList());
		final ArraySpecifier arraySpecifier = variableDeclarator.getArraySpecifiers().stream().flatMap(StreamUtil.castOrEmpty(ArraySpecifier.class)).collect(StreamUtil.atMostOne()).get();

		for (int dimension = 0; dimension<arraySpecifier.getNumDimensions(); dimension++) {
			if (arraySpecifier.getDimension(dimension) == null) {
				arraySpecifier.setDimension(dimension, dimensionSizesAsExpression.get(dimension));
			}

		}
	}

	/**
	 * Returns a string representation of this variable collection
	 *
	 * @param variables the variables
	 * @return information about the variables as string
	 */
	public static String variablesToString(Collection<Variable> variables) {
		return variables.stream().map(var -> {
			String out = var.toString() + " (typeSize: " + getSpartanMCSizeFromSpecifiers(getTypeSpecifiers(var));
			out = isArray(var)? out + "; arrLength: " + getLength(var) + ")" : out + ")";
			return out;
		})
				.collect(Collectors.toList()).toString();
	}

	public static int getSpartanMCSizeFromSpecifiers(List<Specifier> specifiers) {
		return specifiers.stream().mapToInt(s -> getSpartanMCSizeFromSpecifier(s)).sum();
	}

	/**
	 * gives the size in bytes of a specifier in spartanmc
	 * @param specifier a specifier
	 * @return the size in bytes
	 */
	@SuppressWarnings("magicNumber")
	public static int getSpartanMCSizeFromSpecifier(Specifier specifier) {
		if (specifier == Specifier.BYTE || specifier == Specifier.CHAR) {
			return 1;
		} else if (specifier == Specifier.SHORT || specifier == Specifier.INT) {
			return 2;
		} else if (specifier == Specifier.LONG || specifier == Specifier.FLOAT) {
			return 4;
		} else if (specifier == Specifier.DOUBLE) {
			return 8;
		} else {
			return 0;
		}
	}

	/**
	 * searches for alternative variable declarators in other files if a variable is declared extern
	 * @param variableDeclarator declarator who'S alternative should be found
	 * @return Optional.empty() if the variableDeclarator is not extern or no alternative can be found and an alternative declaration otherwise
	 */
	public static Optional<VariableDeclarator> getAlternativeVariableDeclarator(VariableDeclarator variableDeclarator) {

		final VariableDeclaration variableDeclaration = (VariableDeclaration) variableDeclarator.getDeclaration();
		if (!variableDeclaration.isExtern()) {
			return Optional.empty();
		}
		Traversable decl = variableDeclarator.getDeclaration().getParent();
		Optional<VariableDeclarator> alternativeVariableDeclarator;
		final Program program = CetusUtils.findParentOfType(variableDeclarator, Program.class);
		alternativeVariableDeclarator = CetusUtils.streamChildrenOfType(program, TranslationUnit.class)
				.filter(tu -> !tu.equals(CetusUtils.findParentOfType(decl, TranslationUnit.class)))
				.flatMap(tu -> CetusUtils.streamChildrenOfType(tu, VariableDeclaration.class))
				.filter(vd -> !vd.isExtern())
				.flatMap(ds -> CetusUtils.getStreamOfAllChildren(ds, VariableDeclarator.class))
				.filter(vd -> vd.getSymbolName().equalsIgnoreCase(variableDeclarator.getSymbolName()))
				.collect(StreamUtil.atMostOne());

		if (alternativeVariableDeclarator.isPresent()) {
			return alternativeVariableDeclarator;
		}

		DepthFirstIterator<Traversable> iter = new DepthFirstIterator<>(CetusUtils.findParentOfType(decl, TranslationUnit.class));

		boolean insideInclude = false;

		while (iter.hasNext()) {
			final Traversable t = iter.next();

			if (FunctionCallgraph.isStartIncludePragma(t)) {
				insideInclude = true;
			} else if (FunctionCallgraph.isEndIncludePragma(t)) {
				insideInclude = false;
			} else if (insideInclude) {
				final Optional<VariableDeclarator> optionalVariableDeclarator = Stream.of(t).flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class)).findFirst();
				if (optionalVariableDeclarator.isPresent() && optionalVariableDeclarator.get().getSymbolName().equalsIgnoreCase(variableDeclarator.getSymbolName()) && !((VariableDeclaration) optionalVariableDeclarator.get().getDeclaration()).isExtern()) {
					return optionalVariableDeclarator;
				}
			}
		}

		return Optional.empty();
	}
}
