/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.tools.spartanmc;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.tools.CetusDriver;
import de.tu_darmstadt.rs.util.FileUtil;
import de.tu_darmstadt.rs.util.SpartanMCDefaultFolders;

import cetus.hir.DFIterator;
import cetus.hir.Declaration;
import cetus.hir.Identifier;
import cetus.hir.Program;
import cetus.hir.TranslationUnit;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.nio.file.FileAlreadyExistsException;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Some helpful classes in the spartanmc environment
 */
public final class SpartanMCEnvUtils {

	private static Library library;
	private static final Logger log = LoggerFactory.getLogger(SpartanMCEnvUtils.class);
	private static final Comparator<ModuleDescription> DESC_COMPARATOR = Comparator.comparing(ModuleDescription::getName).thenComparing(p -> -p.getName().length());

	private SpartanMCEnvUtils() {
	}

	public static boolean matchesExpectedPeriIdentifier(Identifier id) {
		return matchesExpectedPeriIdentifier(id.getName());
	}

	public static boolean matchesExpectedPeriIdentifier(String id) {
		return id.matches("^(?!SB_|SBI_)[A-Z][A-Z_0-9]+$");
	}

	public static boolean matchesExpectedISRIdentifier(String id) {
		return id.matches("^isr_[0-9]+$");
	}

	public static boolean matchesExpectedInterruptPeripheralName(String peripheralName) {
		return peripheralName.matches("^(intctrl)p?$");
	}

	/**
	 * finds a matching peripheral identifier in the declaration
	 *
	 * @param declaration a declaration to be searched
	 * @return An identifier if a matching one has been found, null otherwise
	 */
	public static Identifier getPeriIdentifierFromDeclaration(Declaration declaration) {
		DFIterator<Identifier> it = new DFIterator(declaration, Identifier.class);
		while (it.hasNext()) {
			Identifier identifier = it.next();
			if (SpartanMCEnvUtils.matchesExpectedPeriIdentifier(identifier)) {
				return identifier;
			}
		}
		return null;
	}


	/**
	 * Get all peripherals found by libdevxml
	 * @param peripheralBlacklist a blacklist of peripherals to ignore for this scan
	 * @return All PeripheralModuleDescription found by libdevxml
	 */
	public static List<PeripheralModuleDescription> getPeripheralLibrary(Set<String> peripheralBlacklist) {
		List<PeripheralModuleDescription> periConstants = getLibdevxmlLibraryPeripherals(peripheralBlacklist).stream()
				.sorted(DESC_COMPARATOR)
				.collect(Collectors.toList());
		return periConstants;
	}

	/**
	 * map of includepath to peripheral
	 * @param peripheralBlacklist a blacklist of peripherals to ignore for this scan
	 * @return For each include Path a Set of Peripherals
	 */
	public static Map<Path, Set<PeripheralModuleDescription>> getIncludeFileToPeripheralMap(Set<String> peripheralBlacklist) {
		Map<Path, Set<PeripheralModuleDescription>> includeLookup = new HashMap<>();

		getLibdevxmlLibraryPeripherals(peripheralBlacklist).stream()
				.forEach(peri -> {
					if (peri.hasRegisters() && peri.getRegisters().getCSupport() != null) {
						Path regInclude = Library.getCIncludeDir(peri).resolve(peri.getRegisters().getCSupport().getFileName());
						addIncludeLookup(includeLookup, regInclude, peri);
					}
					if (peri.hasMemory() && peri.getMemory().getCSupport() != null) {
						Path dmaInclude = Library.getCIncludeDir(peri).resolve(peri.getMemory().getCSupport().getFileName());
						addIncludeLookup(includeLookup, dmaInclude, peri);
					}
				});
		return includeLookup;
	}

	private static boolean addIncludeLookup(Map<Path, Set<PeripheralModuleDescription>> includeLookup, Path headerPath, PeripheralModuleDescription peripheralModuleDescription) {
		if (includeLookup.containsKey(headerPath)) {
			return includeLookup.get(headerPath).add(peripheralModuleDescription);
		} else {
			Set<PeripheralModuleDescription> tmp = new HashSet<>();
			tmp.add(peripheralModuleDescription);
			includeLookup.put(headerPath, tmp);
			return true;
		}
	}

	private static Library getLibdevxmlLibrary() {
		if (library == null) {
			Library.Builder libBuilder = new Library.Builder();
			Map<Path, Set<PeripheralModuleDescription>> includeLookup = new HashMap<>();
			libBuilder.addDefaultSearchpaths(SpartanMCDefaultFolders.getSpartanMCRootFromEnvironmentVariable());
			libBuilder.setVerify(true).setAnalyze(true).setOptimize(true);
			library = libBuilder.build();
		}
		return library;
	}

	private static Set<PeripheralModuleDescription> getLibdevxmlLibraryPeripherals(Set<String> peripheralBlacklist) {
		return getLibdevxmlLibrary().getAvailableModules().stream()
				.filter(m -> m.getType() == IModuleDescription.ModuleType.Peripheral)
				.map(m -> (PeripheralModuleDescription) m)
				.filter(peri -> peri.getName() != null && !peri.getName().isEmpty())
				.filter(peri -> peripheralBlacklist.stream().noneMatch(hl -> peri.getName().equalsIgnoreCase(hl)))
				.collect(Collectors.toSet());
	}


	/**
	 * Finds all header files defined by libdevxml or scans an include folder and creates a traversable to find peripheral function even though they might not be included by main.c and thus could not be traversed
	 *
	 * @param includeLookup the generated includelookup from Peripheral Detector containing header file path and a associated peripheral module
	 * @param additionalSystemLibraries place to search for additional libs
	 * @return true if scanning was successful, false otherwise
	 */
	public static TranslationUnit initPeripheralIncludesTranslationUnit(Map<Path, Set<PeripheralModuleDescription>> includeLookup, Collection<File> additionalSystemLibraries) {
		Path includeDir = SpartanMCDefaultFolders.cHeaders();
		Path tmpHeader = Constants.FilesAndFolders.TEMPORARY_INCLUDE_FOR_ALL_PERIPHERALS;
		boolean scanPeriHeadersByLibDevXML = true;

		try {
			List<String> includes = null;
			if (scanPeriHeadersByLibDevXML) {
				includes = includeLookup.keySet().stream()
						.map(f -> includeDir.relativize(f).toString())
						.map(path -> Include.printSystemInclude(path))
						.collect(Collectors.toList());
			} else {
				File folder = includeDir.toFile();
				List<File> listOfFiles = Arrays.asList(folder.listFiles());
				includes = listOfFiles.stream()
						.filter(f -> f.toString().endsWith(String.valueOf(Constants.FilesAndFolders.HEADERFILES_EXTENSION)))
						.map(f -> Include.printSystemInclude(f.getName()))
						.collect(Collectors.toList());
			}

			String content = includes.stream().collect(Collectors.joining("\n")) + "\n" + "\n";
			if (!tmpHeader.toFile().exists()) {
				FileUtil.writeTemporaryFile(tmpHeader.toFile(), content);
			}

			LinkedList<File> searchpaths = new LinkedList<>();
			searchpaths.add(includeDir.toFile());

			SourceFile sf = new SourceFile(tmpHeader.toFile(), includeDir.toFile(), ".h");
			HashSet<File> files = new HashSet<>();
			files.add(sf.getFile());
			log.info("Starting Cetus parser again with all system headers in " + includeDir.toString() + " to auto-create all includes");
			CetusDriver cetusDriver = new CetusDriver(Collections.singleton(sf.getFile()), searchpaths, additionalSystemLibraries);
			Program p = cetusDriver.parse();

			return (TranslationUnit) p.getChildren().stream().findFirst().get();
		} catch (Exception e) {
			if (e instanceof FileAlreadyExistsException) {
				log.warn("File already exists won't overwrite {}", tmpHeader);
			} else if (e instanceof FileNotFoundException) {
				log.warn("Could no write to file {}", tmpHeader);
			} else if (e instanceof UnsupportedEncodingException) {
				log.warn("UTF-8 Encoding not fupported for file {}", tmpHeader);
			} else if (e instanceof IOException) {
				log.warn("Unable to delete file {}", tmpHeader);
			}
			throw e;
		}
	}

}
