/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Implements default methods to check if TransformationPass is executable
 */
public abstract class AbstractTransformationPass implements TransformationPass {

	@Override
	public boolean init() {
		return true;
	}

	@Override
	public List<Class<? extends TransformationPass>> requiredTransformationPasses()  {
		return new ArrayList<>();
	};

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents()  {
		return new ArrayList<>();
	};

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return new ArrayList<>();
	}

	@Override
	public boolean processable(Model model) {
		final boolean workersApplied = model.allWorkersApplied(requiredTransformationPasses());
		boolean requiredModelComponentsThereOrNoOutstandingWorkerProvidesComponents = requiredModelComponents().stream().allMatch(mc -> checkRequirementsFor(mc, model));
		return workersApplied && requiredModelComponentsThereOrNoOutstandingWorkerProvidesComponents;
	}

	protected boolean isOptimizationPass() {
		return !getOptimizationComponents().isEmpty();
	}

	protected boolean isOptimizationPassFor(Class<? extends ModelComponent> modelComponent) {
		return getOptimizationComponents().contains(modelComponent);
	}

	/**
	 * get components optimized by this worker
	 * @return all components optimized by this worker
	 */
	protected Set<Class<? extends ModelComponent>> getOptimizationComponents() {
		return requiredModelComponents().stream()
				.filter(mc -> providedModelComponents().contains(mc))
				.collect(Collectors.toSet());
	}

	/**
	 * Checks if requirements for a model component are fulfilled
	 * @param modelComponent the ModelComponent to check
	 * @param model the current model
	 * @return true if the ModelComponent is already on the Model or all passes providing this ModelComponent are already executed, false otherwise.
	 */
	protected boolean checkRequirementsFor(Class<? extends ModelComponent> modelComponent, Model model) {
		final boolean componentExists = model.hasModelComponent(modelComponent);

		final Set<Class<? extends ModelComponent>> providedComponentsOfNotAppliedWorkers = model.getNotAppliedWorkers().stream()
				.filter(naw -> !naw.equals(this))
				.flatMap(w -> w.providedModelComponents().stream())
				.collect(Collectors.toSet());
		boolean allPassesProvidingRequirementsExecuted = !providedComponentsOfNotAppliedWorkers.contains(modelComponent);

		return componentExists || allPassesProvidingRequirementsExecuted;
	}
}
