/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;

import java.util.List;

/**
 * <p>
 * This is the {@link TransformationPass} interface. In the µStreams Source-To-Source
 * Compiler Model and the Operation are disjoint. While {@link Model} is
 * representing the entire Datamodel the µStreams Source-To-Source Compiler is
 * working on, implementations of the {@link TransformationPass} interface are Operations
 * working on this Datamodel.
 * </p>
 * <p>
 * The other two methods of the {@link TransformationPass} interface are
 * {@link TransformationPass#process(Model) process(Model)} and
 * {@link TransformationPass#processable(Model) processable(Model)}. During runtime the
 * µStreams Source-To-Source Compile will check whether a worker is ready to
 * process the Datamodel or not. This is done by invoking the
 * {@link TransformationPass#processable(Model) processable(Model)} method with the current
 * {@link Model}. You can use the given {@link Model} to verify, if your
 * {@link TransformationPass} implementation is ready to run. You can use the
 * {@link Model#isWorkerApplied(Class)} method to check whether a specific
 * {@link TransformationPass} has already processed the Model. The
 * {@link TransformationPass#process(Model) process(Model)} method should contain the actual
 * Operations modifying the given {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */

public interface TransformationPass {

	/**
	 * Initializes the {@link TransformationPass}.
	 *
	 * @return Returns whether the initialization was successful or not.
	 */
	boolean init();

	/**
	 * Worker predecessors. The usage of this method should be avoided. Better use {@link TransformationPass ::reqiredModelComponents()} if possible.
	 *
	 * @return all workers required to be processed before this worker is processed
	 */
	List<Class<? extends TransformationPass>> requiredTransformationPasses();

	/**
	 * Required Model Componentes. If a Worker modifies/optimizes a {@link ModelComponent} then it should be listed as provided and required.
	 *
	 * @return a list of model components required by this worker
	 */
	List<Class<? extends ModelComponent>> requiredModelComponents();

	/**
	 * Provided Model Components for the current model. If a Worker modifies/optimizes a {@link ModelComponent} then it should be listed as provided and required.
	 *
	 * @return all additional model components that will be on the model after this worker completes or components beeing optimized/modified.
	 */
	List<Class<? extends ModelComponent>> providedModelComponents();

	/**
	 * Checks whether the given {@link Model} project is contains all necessary
	 * data to be processed by this {@link TransformationPass}.
	 *
	 * @param model The {@link Model} that is checked.
	 * @return Returns whether the {@link TransformationPass} can be executed or not
	 */
	boolean processable(Model model);

	/**
	 * Processes the given {@link Model} m with this {@link TransformationPass} .
	 *
	 * @param model The {@link Model} that is processed.
	 * @return Returns whether the execution was successful or not.
	 */
	boolean process(Model model);
}
