/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.Connection;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IConnectionPeripheral;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IndirectConnection;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ReceivePeripheral;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.SendPeripheral;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.ConnectionPeripheralAccess;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency.DependencyRelation;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency.DependencyType;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependencyGroup;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.DependencyCreator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;

/**
 * <p>
 * This Worker generates {@link ConnectionPeripheralAccess
 * ConnectionPeripheralAccesses} in the {@link Model}.
 * </p>
 * <p>
 * <p>
 * The {@link ConnectionPeripheralAccess} indicates by which connection a
 * TaskDependency is realized. Thus it represents a kind of mapping between
 * {@link Connection Connections} and {@link TaskDependency TaskDependencies}
 * respectively the actual access on the connection peripheral's driver
 * libraries within the finally generated firmware.
 * </p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> All {@link Task Tasks} are extracted an mapped on a
 * {@link Core Core}. All {@link TaskDependency TaskDependencies} have to be
 * generated and added to the {@link Model}.
 * {@link ConnectionAccessCreator#processable(Model)} method checks whether the
 * {@link CoreConnectionCreator} and the {@link DependencyCreator} are applied
 * to the {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class ConnectionAccessCreator extends AbstractTransformationPass {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(ConnectionAccessCreator.class);

	private Model model;

	private Collection<ConnectionPeripheralAccess> accesses;

	private Collection<TaskDependency> dependencies;


	public ConnectionAccessCreator() {
	}

	@Override
	public boolean init() {
		accesses = new LinkedHashSet<>();
		dependencies = new LinkedHashSet<>();
		return true;
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Task.class,
				Core.class,
				TaskDependencyGroup.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				ConnectionPeripheralAccess.class
		);
	}

	@Override
	public boolean process(Model model) {
		this.model = model;

		accesses.clear();

		/* Collect all TaskDependencies */
		dependencies.clear();
		dependencies.addAll(model.getModelComponents(TaskDependency.class));

		/* Create ConnectionPeripheralAccesses for each dependency */

		for (TaskDependency dependency : dependencies) {
			if (!createPeripheralAccess(dependency)) {
				return false;
			}
		}

		return true;
	}

	private boolean createPeripheralAccess(TaskDependency dependency) {

		Collection<Task> providers;// = new ArrayList<Task>();
		Collection<Task> consumers;// = new ArrayList<Task>();

		if (DependencyType.ONE_TO_ONE == dependency.getType()) {
			providers = Collections.singletonList(dependency.getProvider());
			consumers = Collections.singletonList(dependency.getConsumer());
		} else if (DependencyType.ONE_TO_MANY == dependency.getType()) {
			consumers = new ArrayList<Task>(dependency.getReplica());
			consumers.add(dependency.getConsumer());
			providers = Collections.singletonList(dependency.getProvider());
		} else if (DependencyType.MANY_TO_ONE == dependency.getType()) {
			providers = new ArrayList<Task>(dependency.getReplica());
			providers.add(dependency.getProvider());
			consumers = Collections.singletonList(dependency.getConsumer());
		} else {
			model.printIgnoreableError(log,"Cannot handle MANY_TO_MANY dependencies.");
			return false;
		}

		for (Task provider : providers) {
			for (Task consumer : consumers) {
				IndirectConnection connection = getConnection(provider, consumer);

				IConnectionPeripheral sendingPeripheral = getSendingPeripheral(
						connection, provider);
				IConnectionPeripheral receivingPeripheral = getReceivingPeripheral(
						connection, consumer);

				ConnectionPeripheralAccess receivingAccess = new ConnectionPeripheralAccess(
						dependency, consumer, receivingPeripheral, true);

				ConnectionPeripheralAccess sendingAccess = new ConnectionPeripheralAccess(
						dependency, provider, sendingPeripheral, false);

				if (DependencyRelation.PARENT_CHILD == dependency.getRelation()) {

					dependency.getLocation().addConnectionPeripheralAccess(sendingAccess);
					consumer.addConnectionPeripheralAccess(receivingAccess);

				} else if (DependencyRelation.CHILD_PARENT == dependency
						.getRelation()) {

					dependency.getLocation().addConnectionPeripheralAccess(receivingAccess);
					provider.addConnectionPeripheralAccess(sendingAccess);

				} else if (DependencyRelation.CHILD_CHILD == dependency
						.getRelation()) {

					consumer.addConnectionPeripheralAccess(receivingAccess);
					provider.addConnectionPeripheralAccess(sendingAccess);

				}

				model.addModelComponent(receivingAccess);
				model.addModelComponent(sendingAccess);
			}
		}

		return true;
	}

	// private boolean createPeripheralAccess(TaskDependency dependency) {
	// Connection connection = getConnections(dependency.getDependsOn(),
	// dependency.getTask());
	//
	// if (null != connection) {
	//
	// ConnectionPeripheral sendingPeripheral = getSendingPeripheral(
	// connection, dependency);
	// ConnectionPeripheral receivingPeripheral = getReceivingPeripheral(
	// connection, dependency);
	//
	// ConnectionPeripheralAccess accessParent = null;
	//
	// ConnectionPeripheralAccess accessChild = null;
	//
	// if (DependencyRelation.CHILD_CHILD == dependency.getRelation()) {
	//
	// accessParent = new ConnectionPeripheralAccess(dependency,
	// sendingPeripheral, false);
	//
	// accessChild = new ConnectionPeripheralAccess(dependency,
	// receivingPeripheral, true);
	//
	// if (!dependency.getDependsOn().addConnectionPeripheralAccess(
	// accessParent)) {
	// log.warn("ConnectionAccess refused!");
	// }
	//
	// if (!dependency.getTask().addConnectionPeripheralAccess(
	// accessChild)) {
	// log.warn("ConnectionAccess refused!");
	// }
	//
	// return true;
	//
	// } else if (DependencyRelation.PARENT_CHILD == dependency.getRelation()) {
	//
	// accessParent = new ConnectionPeripheralAccess(dependency,
	// sendingPeripheral, false);
	//
	// accessChild = new ConnectionPeripheralAccess(dependency,
	// receivingPeripheral, true);
	//
	// if (!dependency.getLocation().addConnectionPeripheralAccess(
	// accessParent)) {
	// log.warn("ConnectionAccess refused!");
	// }
	//
	// if (!dependency.getTask().addConnectionPeripheralAccess(
	// accessChild)) {
	// log.warn("ConnectionAccess refused!");
	// }
	//
	// return true;
	//
	// } else if (DependencyRelation.CHILD_PARENT == dependency.getRelation()) {
	//
	// accessParent = new ConnectionPeripheralAccess(dependency,
	// receivingPeripheral, true);
	//
	// accessChild = new ConnectionPeripheralAccess(dependency,
	// sendingPeripheral, false);
	//
	// if (!dependency.getLocation().addConnectionPeripheralAccess(
	// accessParent)) {
	// log.warn("ConnectionAccess refused!");
	// }
	//
	// if (!dependency.getDependsOn().addConnectionPeripheralAccess(
	// accessChild)) {
	// log.warn("ConnectionAccess refused!");
	// }
	//
	// return true;
	//
	// }
	//
	// return false;
	//
	// } else {
	// return false;
	// }
	//
	// }
	//
	// private boolean createMultiplePeripheralAccess(
	// MultipleTaskDependency dependency) {
	//
	// if (DependencyRelation.CHILDS_PARENT == dependency.getRelation()) {
	//
	// Collection<ConnectionPeripheral> peripherals = new
	// ArrayList<ConnectionPeripheral>();
	// Task parent = dependency.getTask();
	//
	// for (Task child : dependency.getDependsOns()) {
	// Connection connection = getConnections(child, parent);
	//
	// if (null != connection) {
	//
	// ConnectionPeripheral childPeripheral = connection
	// .getConnectionPeripheral(getCore(child));
	// ConnectionPeripheral parentPeripheral = connection
	// .getConnectionPeripheral(getCore(parent));
	//
	// peripherals.add(parentPeripheral);
	// child.addConnectionPeripheralAccess(new ConnectionPeripheralAccess(
	// dependency, childPeripheral, false));
	//
	// } else {
	// return false;
	// }
	//
	// }
	//
	// dependency.getLocation().addConnectionPeripheralAccess(
	// new MultipleConnectionPeripheralAccess(dependency,
	// peripherals, true));
	//
	// return true;
	// } else if (DependencyRelation.PARENT_CHILDS == dependency.getRelation())
	// {
	//
	// Collection<ConnectionPeripheral> peripherals = new
	// ArrayList<ConnectionPeripheral>();
	// Task parent = dependency.getProviders();
	//
	// for (Task child : dependency.getTasks()) {
	// Connection connection = getConnections(parent, child);
	//
	// if (null != connection) {
	//
	// ConnectionPeripheral childPeripheral = connection
	// .getConnectionPeripheral(getCore(child));
	// ConnectionPeripheral parentPeripheral = connection
	// .getConnectionPeripheral(getCore(parent));
	//
	// peripherals.add(parentPeripheral);
	// child.addConnectionPeripheralAccess(new ConnectionPeripheralAccess(
	// dependency, childPeripheral, true));
	//
	// } else {
	// return false;
	// }
	//
	// }
	//
	// dependency.getLocation().addConnectionPeripheralAccess(
	// new MultipleConnectionPeripheralAccess(dependency,
	// peripherals, false));
	//
	// return true;
	// } else if (DependencyRelation.CHILDS_CHILDS == dependency.getRelation())
	// {
	//
	// log.error("Childs-childs not supported yet.");
	// return false;
	// }
	//
	// return false;
	//
	// }

	/**
	 * This helper method returns the {@link IConnectionPeripheral} on the
	 * receiving side which is able to realize the given {@link TaskDependency}
	 * over the given {@link Connection}.
	 *
	 * @param connection the connection to which the peripheral shall belong to.
	 * @return the {@link IConnectionPeripheral} which is able to realize the given {@link TaskDependency}.
	 */
	private IConnectionPeripheral getReceivingPeripheral(IndirectConnection connection,
														 Task consumer) {
		IConnectionPeripheral peripheral = connection
				.getConnectionPeripheral(consumer.getCore());

		if (!(peripheral instanceof ReceivePeripheral)) {
			model.printIgnoreableError(log,"Receiving peripheral cannot send.");
			return null;
		}

		return peripheral;
	}

	/**
	 * This helper method returns the {@link IConnectionPeripheral} on the
	 * sending side which is able to realize the given {@link TaskDependency}
	 * over the given {@link Connection}.
	 *
	 * @param connection the connection to which the peripheral shall belong to.
	 * @return the {@link IConnectionPeripheral} which is able to realize the given {@link TaskDependency}.
	 */
	private IConnectionPeripheral getSendingPeripheral(IndirectConnection connection,
													   Task provider) {
		IConnectionPeripheral peripheral = connection
				.getConnectionPeripheral(provider.getCore());

		if (!(peripheral instanceof SendPeripheral)) {
			model.printIgnoreableError(log,"Sending peripheral cannot send.");
			return null;
		}

		return peripheral;
	}

	/**
	 * This helper method returns the {@link Connection} which is able to
	 * realize the given {@link TaskDependency}.
	 *
	 * @return the {@link Connection} which is able to realize the given
	 * {@link TaskDependency}.
	 */
	private IndirectConnection getConnection(Task sender, Task receiver) {

		LinkedHashSet<IndirectConnection> foundConnections = new LinkedHashSet<IndirectConnection>();

		Collection<IndirectConnection> connections = model
				.getModelComponents(IndirectConnection.class);

		Core sendingCore = sender.getCore();

		Core receivingCore = receiver.getCore();

		if (null == receivingCore || null == sendingCore) {
			return null;
		}

		for (IndirectConnection connection : connections) {
			if (connection.isConnecting(sendingCore, receivingCore)) {
				foundConnections.add(connection);
			}
		}

		if (foundConnections.isEmpty()) {
			model.printIgnoreableError(log,"No connection for this tasks found: " + sender + " -> "
					+ receiver);
			return null;
		} else if (foundConnections.size() > 1) {
			log.warn("More than one connection for this dependeny found: "
					+ sender + " -> " + receiver);
		}

		return foundConnections.iterator().next();
	}

}
