/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConcentratorMaster;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConcentratorSlave;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.Connection;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IConnectionPeripheral;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConnectionPeripheralsIncompatibleException;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.CoreConnectorMaster;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.CoreConnectorSlave;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.DirectConnection;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.DispatcherMaster;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.DispatcherSlave;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IndirectConnection;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.MemSwapDualMasterDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.MemSwapMultiMasterReceiveDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.MemSwapMultiMasterSendDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.MemSwapSlaveReceiveDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.MemSwapSlaveSendDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ReceivePeripheral;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.SendPeripheral;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.DependencyCreator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * <p>
 * This Worker generates {@link Connection Connections} between {@link Core
 * Cores} according to the {@link TaskDependency TaskDependencies} stored in the
 * {@link Model}. It also generates suitable {@link IConnectionPeripheral
 * ConnectionPeripherals} within the {@link Core Cores}. This Worker uses
 * {@link CoreConnectorMaster} and {@link CoreConnectorSlave} peripherals for as
 * {@link IConnectionPeripheral}.
 * </p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> All {@link Task Tasks} are extracted an mapped on a
 * {@link Core Core}. All {@link TaskDependency TaskDependencies} have to be
 * generated and added to the {@link Model}.
 * {@link CoreConnectionCreator#processable(Model)} method checks whether the
 * {@link CoreCreator} and the {@link DependencyCreator} are applied to the
 * {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class CoreConnectionCreator extends AbstractTransformationPass {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(CoreConnectionCreator.class);

	private Model model;


	public CoreConnectionCreator() {
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Core.class,
				TaskDependency.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				Core.class,
				Connection.class
		);
	}

	@Override
	public boolean process(Model model) {
		this.model = model;

		return processWithConnection();
	}


	private boolean processWithConnection() {
		Collection<Core> cores = model.getModelComponents(Core.class);
		DirectConnection connection = createGlobalMemoryConnection();
		for (Core core : cores) {
			for (Task task : core.getTasks()) {
				for (Task provider : task.getProviders()) {

					if (!hasIndirectConnection(provider, task)) {
						Connection con = null;

						//TODO: select connection type intelligently
						if (provider.hasReplica()) {
							//N-to-1 connection
							con = createConcentratorConnection(core, getCore(provider));
						} else if (task.hasReplica()) {
							//1-to-N connection
							con = createDispatcherConnection(core, getCore(provider));
						} else {
							//1-to-1 connection
							con = createSimpleConnection(core, getCore(provider));
						}

						if (null == con) {
							return false;
						}
						model.addModelComponent(con);
					}
				}
				if (task.getSortedGlobalVariables().size() > 0 && !connection.isConnected(getCore(task))) {
					createGlobalMemoryInterface(getCore(task), connection);
				}
			}
		}

		model.addModelComponent(connection);
		return true;
	}

	private boolean addConnectiontToBaseTask(DirectConnection con) {
		Collection<Core> cores = model.getModelComponents(Core.class);
		for (Core core : cores) {
			for (Task task : core.getTasks()) {
				if (task.isBaseTask()) {
					if (!hasDirectConnection(task, task)) {
						createGlobalMemoryInterface(core, con);
					}
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * This helper method returns true if there is already a connection for the
	 * given tasks
	 */
	private boolean hasIndirectConnection(Task sender, Task receiver) {

		Collection<IndirectConnection> connections = model
				.getModelComponents(IndirectConnection.class);

		Core sendingCore = getCore(sender);

		Core receivingCore = getCore(receiver);

		if (null == receivingCore || null == sendingCore) {
			//CHECKSTYLE.OFF: MultipleStringLiterals
			model.printIgnoreableError(log,"Could not find cores.");
			//CHECKSTYLE.ON: MultipleStringLiterals
			return true;
		}

		for (Connection connection : connections) {
			if (connection.isConnecting(sendingCore, receivingCore)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * This helper method returns true if there is already a connection for the
	 * given tasks
	 */
	private boolean hasDirectConnection(Task sender, Task receiver) {

		Collection<DirectConnection> connections = model
				.getModelComponents(DirectConnection.class);

		Core sendingCore = getCore(sender);

		Core receivingCore = getCore(receiver);

		if (null == receivingCore || null == sendingCore) {
			//CHECKSTYLE.OFF: MultipleStringLiterals
			model.printIgnoreableError(log,"Could not find cores.");
			//CHECKSTYLE.ON: MultipleStringLiterals
			return true;
		}

		for (Connection connection : connections) {
			if (connection.isConnecting(sendingCore, receivingCore)) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Returns the {@link Core} hosting the given {@link Task}.
	 *
	 * @param task the {@link Task} whose {@link Core} shall be returned
	 * @return the {@link Core} hosting the given {@link Task}.
	 */
	private Core getCore(Task task) {
		for (Core core : model.getModelComponents(Core.class)) {
			if (core.getTasks().contains(task)) {
				return core;
			}
		}
		return null;
	}

	/**
	 * Creates a simple {@link Connection} between the two given {@link Core
	 * Cores} utilizing the {@link CoreConnectorMaster} and
	 * {@link CoreConnectorSlave} {@link IConnectionPeripheral}.
	 *
	 * @param receivingCore the receiving core.
	 * @param sendingCore the sending core.
	 * @return the generated {@link Connection}, <code>null</code> if the connection could not be generated.
	 */
	private Connection createSimpleConnection(Core receivingCore,
											  Core sendingCore) {

		if (null == model || null == receivingCore || null == sendingCore) {
			//CHECKSTYLE.OFF: MultipleStringLiterals
			model.printIgnoreableError(log,"project, sendingCore, receivingCore must not be null");
			//CHECKSTYLE.ON: MultipleStringLiterals
			return null;
		}

		Class<? extends SendPeripheral> coreConnectorSendClass;
		Class<? extends ReceivePeripheral> coreConnectorReceiveClass;
		if (model.getConfig().useDMA()) {
			coreConnectorSendClass = MemSwapDualMasterDMA.class;
			coreConnectorReceiveClass = MemSwapSlaveReceiveDMA.class;
		} else {
			coreConnectorSendClass = CoreConnectorMaster.class;
			coreConnectorReceiveClass = CoreConnectorSlave.class;
		}

		IndirectConnection connection = getIndirectConnectionFrom(sendingCore, receivingCore, coreConnectorSendClass, coreConnectorReceiveClass, Connection.ConnectionType.POINT_TO_POINT);
		return connection;
	}

	/**
	 * Creates a dispatcher {@link Connection} between the two given {@link Core
	 * Cores} utilizing the DispatcherMaster and
	 * DispatcherSlave {@link IConnectionPeripheral}.
	 *
	 * @param receivingCore the receiving core.
	 * @param sendingCore the sending core.
	 * @return the generated {@link Connection}, <code>null</code> if the connection could not be generated.
	 */
	private Connection createDispatcherConnection(Core receivingCore,
												  Core sendingCore) {

		if (null == model || null == receivingCore || null == sendingCore) {
			//CHECKSTYLE.OFF: MultipleStringLiterals
			model.printIgnoreableError(log,"project, sendingCore, receivingCore must not be null");
			//CHECKSTYLE.ON: MultipleStringLiterals
			return null;
		}
		Class<? extends SendPeripheral> dispatcherSendClass;
		Class<? extends ReceivePeripheral> dispatcherReceiveClass;
		if (model.getConfig().useDMA()) {
			dispatcherSendClass = MemSwapMultiMasterSendDMA.class;
			dispatcherReceiveClass = MemSwapSlaveReceiveDMA.class;
		} else {
			dispatcherSendClass = DispatcherMaster.class;
			dispatcherReceiveClass = DispatcherSlave.class;
		}
		IndirectConnection connection = getIndirectConnectionFrom(sendingCore, receivingCore, dispatcherSendClass, dispatcherReceiveClass, Connection.ConnectionType.DISPATCHER);
		return connection;
	}

	private IndirectConnection getIndirectConnectionFrom(Core sendCore, Core receiveCore, Class<? extends SendPeripheral> sendPeripheralClass, Class<? extends ReceivePeripheral> receivePeripheralClass, Connection.ConnectionType connectionType) {
		IndirectConnection connection = model.getHardwareComponentFactory()
				.createComponent(IndirectConnection.class);
		connection.setConnectionType(connectionType);
		try {
			IConnectionPeripheral sendPeripheral;
			IConnectionPeripheral receivePeripheral;

			//test if sending core already has a master
			if (sendCore.getPeripherals(sendPeripheralClass).isEmpty() || connectionType == Connection.ConnectionType.POINT_TO_POINT) {
				sendPeripheral = model
						.getHardwareComponentFactory().createSubComponent(
								sendPeripheralClass, sendCore);
			} else {
				sendPeripheral = sendCore.getPeripherals(sendPeripheralClass).stream().findFirst().get();
			}

			if (receiveCore.getPeripherals(receivePeripheralClass).isEmpty() || connectionType == Connection.ConnectionType.POINT_TO_POINT) {
				receivePeripheral = model
						.getHardwareComponentFactory().createSubComponent(
								receivePeripheralClass, receiveCore);
			} else {
				receivePeripheral = receiveCore.getPeripherals(receivePeripheralClass).stream().findFirst().get();
			}

			//IConnectionPeripheral receivePeripheral = model
			//		.getHardwareComponentFactory().createSubComponent(
		//					receivePeripheralClass, receiveCore);

			connection.addConnectionPeripheral(sendPeripheral);
			connection.addConnectionPeripheral(receivePeripheral);

		} catch (ConnectionPeripheralsIncompatibleException e) {
			model.printIgnoreableError(log,"Can not create indirect connection", e);
			return null;
		}
		return connection;
	}

	/**
	 * Creates a concentrator {@link Connection} between the two given {@link Core
	 * Cores} utilizing the ConcentratorMaster and
	 * ConcentratorSlave {@link IConnectionPeripheral}.
	 *
	 * @param receivingCore the receiving core.
	 * @param sendingCore the sending core.
	 * @return the generated {@link Connection}, <code>null</code> if the connection could not be generated.
	 */
	private Connection createConcentratorConnection(Core receivingCore, Core sendingCore) {

		if (null == model || null == receivingCore || null == sendingCore) {
			model.printIgnoreableError(log,"project, sendingCore, receivingCore must not be null");
			return null;
		}

		Class<? extends ReceivePeripheral> concentratorReceiveClass;
		Class<? extends SendPeripheral> concentratorSendClass;
		if (model.getConfig().useDMA()) {
			concentratorReceiveClass = MemSwapMultiMasterReceiveDMA.class;
			concentratorSendClass = MemSwapSlaveSendDMA.class;
		} else {
			concentratorReceiveClass = ConcentratorMaster.class;
			concentratorSendClass = ConcentratorSlave.class;
		}

		IndirectConnection connection = getIndirectConnectionFrom(sendingCore, receivingCore, concentratorSendClass, concentratorReceiveClass, Connection.ConnectionType.CONCENTRATOR);
		return connection;
	}

	private void createGlobalMemoryInterface(Core core, DirectConnection connection) {
		if (null == model || null == core) {
			model.printIgnoreableError(log,"project, Core must not be null");
			return;
		}

		try {
			/*ConnectionPeripheral memInterface = model
				.getHardwareComponentFactory().createSubComponent(
					GlobalMemInterface.class, core);
			*/
			connection.addConnection(core);

		} catch (Exception e) {
			model.printIgnoreableError(log, "", e);
			return;
		}
	}

	/**
	 * Creates a globel memory connection for the component
	 * @return the created connection
	 */
	private DirectConnection createGlobalMemoryConnection() {

		if (null == model) {
			model.printIgnoreableError(log,"project must not be null");
			return null;
		}

		DirectConnection connection = model.getHardwareComponentFactory()
				.createComponent(DirectConnection.class);
		connection.setConnectionType(Connection.ConnectionType.NETWORK);
		return connection;
	}
}
