/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ConnectionPeripheralDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.IConnectionPeripheralDMA;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.ReceivePeripheral;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.SendPeripheral;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.ConnectionPeripheralAccess;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunctionCall;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.VariableUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.AccessExpression;
import cetus.hir.AccessOperator;
import cetus.hir.ArrayAccess;
import cetus.hir.AssignmentExpression;
import cetus.hir.AssignmentOperator;
import cetus.hir.CompoundStatement;
import cetus.hir.DeclarationStatement;
import cetus.hir.Expression;
import cetus.hir.ExpressionStatement;
import cetus.hir.FunctionCall;
import cetus.hir.Identifier;
import cetus.hir.Initializer;
import cetus.hir.IntegerLiteral;
import cetus.hir.NameID;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Modifies the send variables via DMA connections to properly initialize them
 * Created by Kris Heid on 04.12.17.
 */
public class CoreConnectionDMAModifier extends AbstractTransformationPass {

	private final Logger log = LoggerFactory.getLogger(CoreConnectionDMAModifier.class);
	private Model model = null;

	public CoreConnectionDMAModifier() {

	}

	@Override
	public boolean init() {
		return true;
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				ConnectionPeripheralAccess.class,
				CetusProgram.class,
				Task.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				CetusProgram.class,
				Task.class
		);
	}

	@Override
	public boolean process(Model model) {
		this.model = model;
		if (!model.getConfig().useDMA()) {
			return true;
		}

		addVariablesToDMAConnectionPeripherals();

		final List<Variable> variables = model.getModelComponents(IConnectionPeripheralDMA.class).stream()
				.flatMap(c -> c.getVariables().stream())
				.distinct()
				.sorted()
				.collect(Collectors.toList());

		for (Variable variable : variables) {
			removeOrTransformInitializers(variable);
			transformIdentifierAccessesToStructAccesses(variable);
		}

		addMemcopyToTasks();

		return true;
	}

	private void addMemcopyToTasks() {
		for (Task task : model.getModelComponents(Task.class)) {

			final List<Variable> variablesReceivedAndSendViaDMA = getVariablesReceivedAndSendViaDMA(task);

			for (Variable variable : variablesReceivedAndSendViaDMA) {
				final Optional<IConnectionPeripheralDMA> dmaPeripheralsConsuming = task.getDMAPeripheralsConsuming(variable);
				final Optional<IConnectionPeripheralDMA> dmaPeripheralsProviding = task.getDMAPeripheralsProviding(variable);
				final IntegerLiteral size = new IntegerLiteral(variable.getSize());

				if (dmaPeripheralsConsuming.isPresent() && dmaPeripheralsProviding.isPresent()) {
					final Identifier idFrom = dmaPeripheralsProviding.get().getIdentifierFor(variable);
					final Identifier idTo = dmaPeripheralsConsuming.get().getIdentifierFor(variable);
					addMemcopyToTask(task, idFrom, idTo, size);
				} else {
					model.printIgnoreableError(log,"No consuming identifier for variable " + variable + " in task " + task + "found");
				}

			}
		}
	}

	private void addMemcopyToTask(Task task, Identifier src, Identifier dst, IntegerLiteral size) {
		if (task.getTraversable() instanceof CompoundStatement) {
			final CompoundStatement compound = (CompoundStatement) task.getTraversable();

			src.setParent(null);
			FunctionCall functionCall = new FunctionCall(new Identifier("memcpy"), dst, src, size);
			compound.addStatement(new ExpressionStatement(functionCall));

			//add include in task
			UFunctionCall call = new UFunctionCall(Constants.FilesAndFolders.STRING);
			task.addFunctionCall(call);
		} else {
			model.printIgnoreableError(log,"Task traversable is no compound statement");
		}
	}

	private List<Variable> getVariablesReceivedAndSendViaDMA(Task task) {
		final Set<ConnectionPeripheralAccess> connectionPeripheralAccessesDMA = task.getConnectionPeripheralAccesses().stream()
				.filter(cpa -> cpa.getPeripheral() instanceof IConnectionPeripheralDMA)
				.collect(Collectors.toSet());

		final List<Variable> receivedVariables = connectionPeripheralAccessesDMA.stream()
				.filter(cpa -> cpa.getDependency().getConsumer().equals(task) || cpa.getDependency().getConsumer().getReplica().contains(task))
				.map(cpa -> cpa.getDependency().getVariableOfConsumer())
				.collect(Collectors.toList());
		final List<Variable> sendVariables = connectionPeripheralAccessesDMA.stream()
				.filter(cpa -> cpa.getDependency().getProvider().equals(task) || cpa.getDependency().getProvider().getReplica().contains(task))
				.map(cpa -> cpa.getDependency().getVariableOfProvider())
				.collect(Collectors.toList());

		receivedVariables.retainAll(sendVariables);
		receivedVariables.sort(Comparator.naturalOrder());
		return receivedVariables;
	}

	private void addVariablesToDMAConnectionPeripherals() {

		final Map<IConnectionPeripheralDMA, Set<Variable>> peripheralVariablesMap = model.getModelComponents(Task.class).stream()
				.flatMap(t -> t.getDependencies().stream())
				.map(d -> getConnectionPeripheralAccessOfDependencyProviderTo(d, model))
				.filter(cpa -> cpa.getPeripheral() instanceof IConnectionPeripheralDMA)
				.distinct()
				.collect(Collectors.groupingBy(cpa -> (IConnectionPeripheralDMA) cpa.getPeripheral(), Collectors.mapping(ConnectionPeripheralAccess::getVariable, Collectors.toSet())));

		for (Map.Entry<IConnectionPeripheralDMA, Set<Variable>> entry : peripheralVariablesMap.entrySet()) {
			final IConnectionPeripheralDMA peripheral = entry.getKey();
			final List<Variable> variables = entry.getValue().stream().sorted().collect(Collectors.toList());
			for (Variable variable : variables) {
				peripheral.addVariable(variable);
			}
		}
	}

	private void removeOrTransformInitializers(Variable sharedVariable) {
		Initializer i = sharedVariable.getInitializer();//CetusUtils.getLiteralInitializer((VariableDeclaration) sharedVariable.getVariableDeclarator().getDeclaration());
		Traversable decl = sharedVariable.getVariableDeclarator().getDeclaration().getParent();
		if (decl instanceof DeclarationStatement) {
			final VariableDeclaration variableDeclaration = (VariableDeclaration) sharedVariable.getVariableDeclarator().getDeclaration();
			// Variable is declared in a function. Replace declaration with assignment or remove declaration if no initializer is found.
			final Optional<VariableDeclarator> variableDeclarator = ((DeclarationStatement) decl).getDeclaration().getChildren().stream()
					.flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class))
					.filter(c -> c.equals(sharedVariable.getVariableDeclarator()))
					.findFirst();
			Optional<VariableDeclarator> alternativeVariableDeclarator = Optional.empty();
			if (variableDeclaration.isExtern()) {

				alternativeVariableDeclarator = VariableUtils.getAlternativeVariableDeclarator(sharedVariable.getVariableDeclarator());
			}
			final boolean isVariableWritten = model.getModelComponents(VariableAccess.class).stream()
					.anyMatch(vac -> vac.getVariable() == sharedVariable && vac.isWritten());

			if (i != null && isVariableWritten) {
				List<ExpressionStatement> statements = new ArrayList<>();
				if (i.isList()) {
					final List<Integer> dimenstionsOfArray = VariableUtils.getDimenstionsOfArray(sharedVariable.getVariableDeclarator());
					final int totalIndices = dimenstionsOfArray.stream().mapToInt(Integer::intValue).reduce(1, Math::multiplyExact);
					final List<Expression> initializers = CetusUtils.getStreamOfAllChildren(i, Expression.class).flatMap(StreamUtil.castOrEmpty(Expression.class)).collect(Collectors.toList());
					for (int index = 0; index<totalIndices; index++) {
						Expression initExpr;
						if (index < initializers.size()) {
							initExpr = initializers.get(index);
						} else {
							initExpr = new IntegerLiteral(0);
						}
						final ArrayAccess arrayAccess = new ArrayAccess(sharedVariable.getInitialIdentifier().clone(), new IntegerLiteral(index));
						AssignmentExpression assignment = new AssignmentExpression(arrayAccess, AssignmentOperator.NORMAL, initExpr.clone());
						ExpressionStatement statement = new ExpressionStatement(assignment);
						statements.add(statement);
					}

				} else {
					Expression initExpr = (Expression) i.getChildren().get(0);
					AssignmentExpression assignment = new AssignmentExpression(sharedVariable.getInitialIdentifier().clone(), AssignmentOperator.NORMAL, initExpr.clone());
					ExpressionStatement statement = new ExpressionStatement(assignment);
					statements.add(statement);
				}
				CetusUtils.removeChild(variableDeclarator.get());
				final CompoundStatement compoundStatement = CetusUtils.getStreamOfParent(decl).flatMap(StreamUtil.castOrEmpty(CompoundStatement.class)).collect(StreamUtil.exactlyOne());
				for (ExpressionStatement statement : statements) {
					compoundStatement.addStatementAfter((DeclarationStatement) decl, statement);
				}
				if (((DeclarationStatement) decl).getDeclaration().getDeclaredIDs().isEmpty()) {
					CetusUtils.removeChild(decl);
				}
			} else {
				if (((DeclarationStatement) decl).getDeclaration().getDeclaredIDs().size() > 1 ) {
					CetusUtils.removeChild(variableDeclarator.get());
				} else {
					CetusUtils.removeChild(decl);
				}
				if (alternativeVariableDeclarator.isPresent()) {
					Optional<VariableDeclaration> alternativeDeclaration = Optional.of(CetusUtils.findParentOfType(alternativeVariableDeclarator.get(), VariableDeclaration.class));
					if (alternativeDeclaration.get().getDeclaredIDs().size() > 1) {
						CetusUtils.removeChild(alternativeVariableDeclarator.get());
					} else {
						CetusUtils.removeChild(alternativeDeclaration.get());
					}
				}
			}
		} else if (decl instanceof TranslationUnit) {
			// Variable is declared globally, remove declaration because it is moved to struct.
			CetusUtils.removeChild(sharedVariable.getVariableDeclarator().getDeclaration());
		} else {
			throw new UnsupportedOperationException("Variable is declared or initialized in an unsupported way.");
		}
	}

	private void replaceIdentifiers(List<Identifier> identifiersToReplace, AccessExpression structAccess) {
		for (Identifier identifer : identifiersToReplace) {
			final Traversable parent = identifer.getParent();
			int index = 0;
			for (Traversable c : parent.getChildren()) {
				if (c.equals(identifer)) {
					parent.setChild(index, structAccess.clone());
				}
				index++;
			}
		}
	}

	private void transformIdentifierAccessesToStructAccesses(Variable sharedVariable) {
		Map<Task, ConnectionPeripheralDMA> varOccurences = new HashMap<>();

		// Find all DMA peripherals where the variable is used.
		for (Task t : model.getModelComponents(Task.class)) {
			boolean varFromSender = false;

			final List<ConnectionPeripheralDMA> receiveDMAPeripherals = t.getCore().getPeripherals(ReceivePeripheral.class).stream()
					.flatMap(StreamUtil.castOrEmpty(ConnectionPeripheralDMA.class))
					.collect(Collectors.toList());
			final List<ConnectionPeripheralDMA> sendDMAPeripherals = t.getCore().getPeripherals(SendPeripheral.class).stream()
					.flatMap(StreamUtil.castOrEmpty(ConnectionPeripheralDMA.class))
					.collect(Collectors.toList());
			// If the variable is received from another core, it is in the DMA memory of the slave.
			for (ConnectionPeripheralDMA conn : receiveDMAPeripherals) {
				if (conn.getVariables().contains(sharedVariable)) {
					varOccurences.put(t, conn);
					varFromSender = true;
				}
			}

			// If the variable is instantiated in task t it is put in the master DMA memory.
			if (!varFromSender) {
				for (ConnectionPeripheralDMA conn : sendDMAPeripherals) {
					if (conn.getVariables().contains(sharedVariable)) {
						varOccurences.put(t, conn);
					}
				}
			}
		}

		// The basetask has to be handled at the end because its traversable contains all other tasks.
		// This hack can be removed once the baseTask has its own compound statement and does no longer
		// contain all other tasks.
		// TODO: Remove hack once base task has own compound statement.
		Task baseTask = null;

		final Identifier originalIdentifier = sharedVariable.getIdentifiers().stream().findFirst().get();
		for (Map.Entry<Task, ConnectionPeripheralDMA> e : varOccurences.entrySet()) {
			List<Identifier> identifiersToReplace = e.getKey().getTraversable().getChildren().stream()
					.flatMap(t -> CetusUtils.getStreamOfAllChildren(t, Identifier.class))
					.filter(i -> i.equals(originalIdentifier))
					.collect(Collectors.toList());

			final String structName = e.getValue().getStructInstanceName();
			final AccessExpression structAccess = new AccessExpression(new NameID(structName), AccessOperator.MEMBER_ACCESS, new NameID(sharedVariable.getName()));

			if (e.getKey().isBaseTask()) {
				baseTask = e.getKey();
				continue;
			}

			replaceIdentifiers(identifiersToReplace, structAccess);
		}

		if (baseTask != null) {
			List<Identifier> baseTaskIdentifiers = baseTask.getTraversable().getChildren().stream()
					.flatMap(t -> CetusUtils.getStreamOfAllChildren(t, Identifier.class))
					.filter(i -> i.equals(originalIdentifier))
					.collect(Collectors.toList());

			final String structName = varOccurences.get(baseTask).getStructInstanceName();
			final AccessExpression baseTaskStructAccess = new AccessExpression(new NameID(structName), AccessOperator.MEMBER_ACCESS, new NameID(sharedVariable.getName()));

			replaceIdentifiers(baseTaskIdentifiers, baseTaskStructAccess);
		}
	}

	private ConnectionPeripheralAccess getConnectionPeripheralAccessOfDependencyProviderTo(TaskDependency taskDependency, Model model) {
		return model.getModelComponents(ConnectionPeripheralAccess.class).stream()
				.filter(x -> x.getTask().equals(taskDependency.getProvider()))
				.filter(x -> x.getDependency().equals(taskDependency))
				.collect(StreamUtil.exactlyOne());
	}
}
