/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelSubComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.connection.DirectConnection;
import de.tu_darmstadt.rs.microStreams.model.interfaces.FreemarkerHasBody;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * This {@link TransformationPass} generates the nessesary hardware to evaluate µStreams
 * generated software
 *
 * @author Jan Weber
 */
public class EvaluationPeripheralCreator extends AbstractTransformationPass {

	private static final Logger log = LoggerFactory.getLogger(EvaluationPeripheralCreator.class);
	private Boolean createEvaluationHardware = false;


	public EvaluationPeripheralCreator() {
	}

	/**
	 * Checks whether the given {@link Core} core contains the base task. And
	 * returns true if so, false otherwise.
	 *
	 * @param core a core
	 * @return Returns true if the given {@link Core} core contains the base task, false otherwise.
	 */
	private boolean containsBaseTask(Core core) {

		boolean containsBaseTask = false;

		for (Task task : core.getTasks()) {
			if (task.isBaseTask()) {
				containsBaseTask = true;
				break;
			}
		}

		return containsBaseTask;

	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Task.class,
				Core.class,
				FreemarkerHasBody.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				HardwareModelSubComponent.class
		);
	}

	@Override
	public boolean process(Model project) {
		createEvaluationHardware = project.getConfig().isCreatePerformanceEvaluation();
		if (!createEvaluationHardware) {
			log.info("Evaluation hardware deactivated. Skipping creation ...");
			return true;
		}

		Collection<Core> cores = project.getModelComponents(Core.class);
		Collection<EvaluationPeripheralCreator.EvaluationComponent> evalComponents = null;
		if (project.getConfig().isCreateSimulatedTarget()) {
			evalComponents = Constants.EvaluationPeripherals.EVALUATION_COMPONENTS_SIM;
		} else {
			evalComponents = Constants.EvaluationPeripherals.EVALUATION_COMPONENTS;
		}

		HardwareModelSubComponent subComponent;

		// Add all components from Constants.EvaluationPeripherals.EVALUATION_COMPONENTS.
		try {
			for (EvaluationComponent component : evalComponents) {
				for (Core core : cores) {
					if (containsBaseTask(core)
							&& component.type == EvaluationComponentType.BASE_SUB) {
						subComponent = project.getHardwareComponentFactory()
								.createSubComponent(component.componentName,
										core);

						subComponent.setEvaluationComponent(true);

					}
					if (!containsBaseTask(core)
							&& component.type == EvaluationComponentType.OTHER_SUB) {
						subComponent = project.getHardwareComponentFactory()
								.createSubComponent(component.componentName,
										core);

						subComponent.setEvaluationComponent(true);
					}
				}
				if (component.type == EvaluationComponentType.TOP_LEVEL) {
					project.getHardwareComponentFactory().createComponent(
							component.componentName);
				}
			}

			Collection<FreemarkerHasBody> hasBodys = project
					.getModelComponents(FreemarkerHasBody.class);

			for (FreemarkerHasBody hasBody : hasBodys) {
				hasBody.insertEvaluationAccess(true);
			}

		} catch (IllegalArgumentException | SecurityException
				| ClassNotFoundException e) {
			log.error(
					"There was an error while creating the evaluation hardware.",
					e);
			return false;
		}

		// Add a global data memory to all cores.
		DirectConnection conn = new DirectConnection();
		for (Core core : cores) {
			conn.addConnection(core);
		}
		project.addModelComponent(conn);

		return true;
	}

	/**
	 * The Type of a Evaluation component
	 */
	public enum EvaluationComponentType {
		BASE_SUB, OTHER_SUB, TOP_LEVEL
	}


	/**
	 * Represents an evaluation component
	 */
	public static class EvaluationComponent {

		public String componentName;

		public EvaluationComponentType type;

		public EvaluationComponent(String componentName, EvaluationComponentType type) {
			this.componentName = componentName;
			this.type = type;
		}

	}

}
