/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.peripheralDetection;

import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.tools.spartanmc.SpartanMCEnvUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.FunctionCallgraph;

import cetus.hir.DFIterator;
import cetus.hir.FunctionCall;
import cetus.hir.IDExpression;
import cetus.hir.Identifier;
import cetus.hir.NameID;
import cetus.hir.Symbol;
import cetus.hir.Traversable;
import cetus.hir.VariableDeclarator;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

/**
 * Searches for globally declared peripheral pointers
 */
class GlobalPeripheralNames {

	private Map<NameID, Identifier> globalPeriPointers;

	private Model project;

	private PeripheralLibrary peripheralLibrary;

	private UninferableConstants uninferableConstants;

	private SoCPeripheralsManager soCPeripheralsManager;

	public GlobalPeripheralNames(PeripheralLibrary peripheralLibrary, UninferableConstants uninferableConstants, SoCPeripheralsManager soCPeripheralsManager) {
		globalPeriPointers = new HashMap<>();
		this.peripheralLibrary = peripheralLibrary;
		this.uninferableConstants = uninferableConstants;
		this.soCPeripheralsManager = soCPeripheralsManager;
	}

	public Map<NameID, Identifier> findGlobalPeriPointers(Model project) {
		this.project = project;
		return findGlobalPeriPointers();
	}

	public Map<NameID, Identifier> findGlobalPeriPointers() {
		Traversable t = project.getModelComponents(CetusProgram.class).stream()
				.findFirst().get()
				.getProgram();
		DFIterator<VariableDeclarator> itVariableDeclarator = new DFIterator<VariableDeclarator>(t, VariableDeclarator.class);
		while (itVariableDeclarator.hasNext()) {
			VariableDeclarator variableDeclarator = itVariableDeclarator.next();
			Optional<NameID> name = variableDeclarator.getChildren().stream()
					.filter(cc -> cc instanceof NameID)
					.map(cc -> (NameID) cc)
					.findFirst();
			if (name.isPresent()) {
				Identifier identifier = SpartanMCEnvUtils.getPeriIdentifierFromDeclaration(variableDeclarator.getDeclaration());
				if (identifier != null) {
					PeripheralModuleDescription pmd = peripheralLibrary.findPeriToConstant(identifier);
					if (pmd != null) {
						uninferableConstants.add(identifier.getName());
						globalPeriPointers.put(name.get(), identifier);
					}
				}


			}
		}
		return globalPeriPointers;
	}

	public void findUsagesOfGlobalPeriPointers(Traversable traversable) {
		if (traversable == null) {
			return;
		}
		DFIterator<Identifier> it = new DFIterator<>(traversable, Identifier.class);
		while (it.hasNext()) {
			Identifier id = it.next();
			addPeripheralIfPeriPointer(id);
		}
		//recursively call all other functions in the traversable to be scanned for constant defines
		Collection<FunctionCall> functionCalls = FunctionCallgraph.getNestedFunctionCalls(traversable, project.getModelComponents(Task.class));
		for (FunctionCall functionCall : functionCalls) {
			if (functionCall.getProcedure() != null) {
				Traversable tr = functionCall.getProcedure();
				findUsagesOfGlobalPeriPointers(tr);
			}
		}
	}


	private void addPeripheralIfPeriPointer(Identifier id) {
		Symbol symbol = id.getSymbol();
		if (symbol != null && symbol instanceof VariableDeclarator) {
			IDExpression nameID = ((VariableDeclarator) symbol).getID();
			globalPeriPointers.keySet().stream()
					.filter(key -> key.equals(nameID))
					.forEach(key -> {
						Identifier identifier = globalPeriPointers.get(key);
						PeripheralModuleDescription peripheralModuleDescription = peripheralLibrary.findPeriToConstant(identifier);
						soCPeripheralsManager.add(peripheralModuleDescription, identifier.getName());
					});
		}

	}
}
