/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.peripheralDetection;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelSubComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.Peripheral;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.Utils;
import de.tu_darmstadt.rs.microStreams.tools.spartanmc.SpartanMCEnvUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel.FunctionCallgraph;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.util.LevenshteinDistance;

import cetus.hir.BreadthFirstIterator;
import cetus.hir.DFIterator;
import cetus.hir.Expression;
import cetus.hir.FunctionCall;
import cetus.hir.Identifier;
import cetus.hir.Procedure;
import cetus.hir.ProcedureDeclarator;
import cetus.hir.Specifier;
import cetus.hir.Symbol;
import cetus.hir.Traversable;
import cetus.hir.UserSpecifier;
import cetus.hir.VariableDeclaration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Detects peripheral accesses within the source code (maybe use pragma)
 *
 * @author Kris Heid
 */
public class PeripheralDetector extends AbstractTransformationPass {

	private static final Logger log = LoggerFactory.getLogger(PeripheralDetector.class);

	private GlobalPeripheralNames globalPeriPointers;

	private PeripheralLibrary peripheralLibrary;

	private SoCPeripheralsManager soCPeripheralsManager;

	private UninferableConstants uninferableConstants;

	private Model model;

	public PeripheralDetector() {
		super();
	}

	/**
	 * Constructor
	 * @param model the current model
	 * @param peripheralLibrary a peripheral library
	 */
	public PeripheralDetector(Model model, PeripheralLibrary peripheralLibrary) {
		this.model = model;
		this.peripheralLibrary = peripheralLibrary;
		uninferableConstants = new UninferableConstants();
		soCPeripheralsManager = new SoCPeripheralsManager(model, uninferableConstants);
		globalPeriPointers = new GlobalPeripheralNames(peripheralLibrary, uninferableConstants, soCPeripheralsManager);
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Core.class,
				TaskDependency.class,
				CetusProgram.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				HardwareModelSubComponent.class
		);
	}

	/**
	 * This Methode recognize all used Peripheries which are defined in a .C file and assigns them to the Cores respectively
	 *
	 * @param project the current project
	 * @return true if the process was being successfully
	 */
	@Override
	public boolean process(Model project) {
		//init stuff
		model = project;
		peripheralLibrary = new PeripheralLibrary(project.getConfig().getAdditionalSystemLibraries());
		uninferableConstants = new UninferableConstants();
		soCPeripheralsManager = new SoCPeripheralsManager(project, uninferableConstants);
		globalPeriPointers = new GlobalPeripheralNames(peripheralLibrary, uninferableConstants, soCPeripheralsManager);

		globalPeriPointers.findGlobalPeriPointers(project);
		List<Core> cores = new ArrayList<Core>(project.getModelComponents(Core.class));
		for (Core core : cores) {
			log.debug("Detecting peripheries for core" + core.getId() + "...");
			soCPeripheralsManager.reset();
			//Access to the assigned Task
			Collection<Task> tasks = new HashSet<Task>(core.getTasks());
			for (Task task : tasks) {
				runAnalysisPasses(task.getTraversable());

				boolean hasOnlyOneTask = (cores.size() == 1 && tasks.size() == 1);
				Set<Procedure> isrs = findISRs(project);
				if ((hasOnlyOneTask && isrs.size() > 0) || soCPeripheralsManager.containsInterrupt(peripheralLibrary.getInterruptPeripherals())) {
					if (!soCPeripheralsManager.containsInterrupt(peripheralLibrary.getInterruptPeripherals())) {
						PeripheralModuleDescription interrupt = peripheralLibrary.getInterruptPeripheral();
						soCPeripheralsManager.add(interrupt);
					}
					//Parse isrs since peri could be used in interrupts
					isrs.stream().forEach(isr -> runAnalysisPasses(isr));
					core.addInterruptServiceRoutines(isrs);
				} else if (!hasOnlyOneTask && isrs.size() > 0){
					model.printIgnoreableError(log, "Can not infer to which core the the ISR method(s) should be mapped. You can force build and copy ISRs manually later.");
				}
			}
			soCPeripheralsManager.addAllPeriesToCore(core);
			log.info("Found peripherals for core " + core.getId() + ": " + soCPeripheralsManager.getPeripherals().stream().map(p -> p.getName()).collect(Collectors.toList()));
		}
		removeDuplicatePeripheralsInChildTasks(project);
		//TODO: search all api calls and see if in the current core an according peripheral was found
		if (!Verifier.verifyProject(project, uninferableConstants)) {
			log.warn("There seems to be a problem with the detected peripherals...continue anyways...check the output!");
		}
		return true;
	}

	/**
	 * Removes peripherals from a parent tasks which were also detected in child tasks
	 *
	 * @param project the current Project
	 */
	private void removeDuplicatePeripheralsInChildTasks(Model project) {
		//TODO: search from what expression the peripheral was detected
		List<Core> cores = new ArrayList<Core>(project.getModelComponents(Core.class));
		for (Core core : cores) {
			cores.stream()
					.filter(c -> c != core)
					.filter(c -> c.hasChildTasksOf(core))
					.forEach(c -> {
						List<Peripheral> commonPeripherals = c.getCommonPeripherals(core);
						commonPeripherals.stream().forEach(peripheral -> core.removePeripheral(peripheral));
					});
		}
	}

	/*private void searchAndAddInterrupts(boolean hasMultipleTasks, Task task, Core core) {

		Set<Procedure> isrs = findISRs(project);
		if ((hasMultipleTasks && isrs.size() > 0) || SoCPeripheralsManager.getInstance().containsInterrupt()) {
			if (!SoCPeripheralsManager.getInstance().containsInterrupt()) {
				PeripheralModuleDescription interrupt = PeripheralLibrary.getInstance().getInterruptPeripheral();
				SoCPeripheralsManager.getInstance().add(interrupt);
			}
			//Add statements of declaration to task for isrXX to be parsed as well for Peri
			task.addDeclarations(isrs.stream().map(p -> p.getBody()).collect(Collectors.toSet()));
			//Parse twice since peri could be used in interrupts
			runAnalysisPasses(task.getTraversable());
			//remove falsely added function calls from traversable again after parsing
			task.removeAddedDeclarations();
			core.addInterruptServiceRoutines(isrs);
		}
	}*/

	private void runAnalysisPasses(Traversable traversable) {
		findDefinedConstants(traversable);
		soCPeripheralsManager.preRunDone();
		findFunctionCalls(traversable);
		findDefinedConstants(traversable);
		globalPeriPointers.findUsagesOfGlobalPeriPointers(traversable);
	}

	/**
	 * Searches for peripherals in the given traversable
	 * @param traversable a traversable
	 * @return the found peripherals
	 */
	public Set<PeripheralModuleDescription> findUsedPeripherals(Traversable traversable) {
		uninferableConstants.clear();
		soCPeripheralsManager.reset();
		findDefinedConstants(traversable);
		soCPeripheralsManager.preRunDone();
		findFunctionCalls(traversable);
		findDefinedConstants(traversable);
		return soCPeripheralsManager.getPeripherals();
	}


	/**
	 * Iterates over the traversable and finds defined constants. Additionally all function calls
	 * in this traversable are found and scanned for constants as well
	 *
	 * @param traversable The traversable to be scanned
	 * @return The peripheral module along with the name to be
	 */
	private void findDefinedConstants(Traversable traversable) {
		if (traversable == null) {
			return;
		}
		//TODO: Scan used global variable definition. if a global variable was used which was priorly assigned the value of peripheral constant.
		//iterate over defined constants
		DFIterator<Identifier> it = new DFIterator<>(traversable, Identifier.class);
		while (it.hasNext()) {
			Identifier id = it.next();
			PeripheralModuleDescription pmd;
			if (id.getSymbol() == null) {
				pmd = peripheralLibrary.findPeriToConstant(id);
				soCPeripheralsManager.add(pmd, id.getName());
			}
		}
		//recursively call all other functions in the traversable to be scanned for constant defines
		Collection<cetus.hir.FunctionCall> functionCalls = FunctionCallgraph.getNestedFunctionCalls(traversable, model.getModelComponents(Task.class));
		for (FunctionCall functionCall : functionCalls) {
			if (functionCall.getProcedure() != null) {
				Traversable tr = functionCall.getProcedure();
				findDefinedConstants(tr);
			}
		}

	}


	/**
	 * Finds Peripheries by function calls
	 *
	 * @param traversable a traversable to search for function calls
	 */
	private void findFunctionCalls(Traversable traversable) {
		if (traversable == null) {
			return;
		}
		Collection<cetus.hir.FunctionCall> functionCalls = FunctionCallgraph.getNestedFunctionCalls(traversable, model.getModelComponents(Task.class));
		for (FunctionCall functionCall : functionCalls) {
			findPeriToFunctionCall(functionCall);
			if (functionCall.getProcedure() != null) {
				Traversable tr = functionCall.getProcedure();
				findFunctionCalls(tr);
			}
		}

	}

	/**
	 * Finds peripherals by system function calls
	 *
	 * @param functionCall a functioncall to be matched against the system library
	 */
	private void findPeriToFunctionCall(FunctionCall functionCall) {
		//TODO: peripherie anhand des erwarteten Parameters in den Argumenten identifizieren
		if (functionCall.getName().toString().equalsIgnoreCase(Constants.C.PRINTF)) {
			Optional<PeripheralModuleDescription> pmd = peripheralLibrary.getAllPeripherals().stream()
					.filter(p -> model.getConfig().isCreateSimulatedTarget()? p.getName().equalsIgnoreCase(Constants.JConfigTypes.SIM_PRINT) : true)
					.filter(p -> Constants.JConfigTypes.PRINTF_PERIPHERALS.stream()
							.anyMatch(printf -> printf.equalsIgnoreCase(p.getName())))
					.sorted(Comparator.comparing(ModuleDescription::getName).reversed())
					.findFirst();
			if (pmd.isPresent() && !soCPeripheralsManager.hasPeripheralAndName(pmd.get(), pmd.get().getName() + "_0")) {
				soCPeripheralsManager.add(pmd.get());
			}
		}

		Comparator<PeripheralModuleDescription> peripheralNameAgainstFunctionName = Comparator.comparing((PeripheralModuleDescription p) -> LevenshteinDistance.get(p.getName(), functionCall.getName().toString()));
		Map<Include, Set<String>> includes = getIncludes(functionCall);
		includes.keySet().forEach(include -> {
			if (include.isSystemLibrary()) {
				includes.get(include).stream().forEach(type -> {
					List<PeripheralModuleDescription> matchingIncludes = peripheralLibrary.findMatchingPeripheralsSorted(type, include.getName(), functionCall, peripheralNameAgainstFunctionName);
					try {
						Set<PeripheralModuleDescription> includesMatchingPreRun = new HashSet<>();
						if (matchingIncludes.size() > 1) {
							includesMatchingPreRun = matchingIncludes.stream()
									.filter(peri -> soCPeripheralsManager.hasPeripheralAndName(peri, getPossiblePeriConstantFromArgs(functionCall)))
									.collect(Collectors.toSet());
						}
						PeripheralModuleDescription pmd = null;
						//CHECKSTYLE.OFF: MultipleStringLiterals
						if (!includesMatchingPreRun.isEmpty()) {
							pmd = includesMatchingPreRun.stream().findFirst().get();
							if (includesMatchingPreRun.size() > 1) {
								log.warn("The include " + include.getName() + " matches with more than one system library and was used several times with constants. Taking first in the list: " + pmd.getName());
							}
						} else {
							pmd = matchingIncludes.stream().sorted((a, b) -> (((Integer) b.getName().compareTo(include.getName())).compareTo(a.getName().compareTo(include.getName())))).findFirst().get();
							if (matchingIncludes.size() > 1) {
								log.warn("The include " + include.getName() + " matches with more than one system library. Taking first in the list: " + pmd.getName());
							}
						}
						//CHECKSTYLE.ON: MultipleStringLiterals
						String name = getPossiblePeriConstantFromArgs(functionCall);
						soCPeripheralsManager.add(pmd, name);
					} catch (NoSuchElementException e) {
						return;
					}
				});

			}
		});
	}

	/**
	 * finds a name to be used as a constant for the function call without args
	 *
	 * @param functionCall a functioncall with arguments
	 * @return a name to be used as a constant for this peripheral and null if no possible name has been found
	 */
	private String getPossiblePeriConstantFromArgs(FunctionCall functionCall) {
		String constant = null;
		if (functionCall.getArguments().size() > 0) {
			for (Expression expression : functionCall.getArguments()) {
				BreadthFirstIterator<Traversable> iterator = new BreadthFirstIterator<>(expression);
				while (iterator.hasNext()) {
					Traversable traversable = iterator.next();
					if (traversable instanceof Identifier) {
						Identifier identifier = (Identifier) traversable;
						if (identifier.findDeclaration() == null) {
							if (SpartanMCEnvUtils.matchesExpectedPeriIdentifier(identifier)) {
								return identifier.getName();
							} else {
								log.warn("The name: " + identifier.getName() + " found as possible Peripheral Constant in function: " + functionCall.toString() + " is not a valid peripheral name");
							}
						} else {
							Identifier i = SpartanMCEnvUtils.getPeriIdentifierFromDeclaration(identifier.getSymbol().getDeclaration());

							if (i != null) {
								return i.getName();
							}

							return null;
						}
					}
				}

			}
		}
		return constant;
	}


	/**
	 * Finds all the system-header file where the function is declared and along with the types of the peripherals
	 *
	 * @param functionCall the function call to be found in the headers
	 * @return all found heads where this function call is located along with their parameter types
	 */
	private Map<Include, Set<String>> getIncludes(FunctionCall functionCall) {

		Map<Include, Set<String>> possibleIncludes = new HashMap<>();

		DFIterator<Traversable> iter = new DFIterator<>(peripheralLibrary.getPeripheralIncludesTranslationUnit());

		Include include = null;

		while (iter.hasNext()) {
			Traversable t = iter.next();

			if (FunctionCallgraph.isStartIncludePragma(t)) {
				boolean isSystemLibrary = t.toString().trim().matches(Constants.Cetus.SYSTEM_LIB_INCLUDE_REGEX);
				String name = Utils.firstMatch(t.toString().trim(), Constants.Cetus.LIB_INCLUDE_REGEX, 1);

				include = new Include(name.replaceAll("<|>|\"", ""), isSystemLibrary, false);
			} else if (FunctionCallgraph.isEndIncludePragma(t)) {
				include = null;
			} else if (null != include) {
				if (t instanceof VariableDeclaration) {
					VariableDeclaration variableDeclaration = (VariableDeclaration) t;
					if ((variableDeclaration.getDeclaredIDs().contains(functionCall.getName()))) {
						Symbol symbol = variableDeclaration.getDeclaredSymbols().stream().findFirst().get();
						if (symbol instanceof ProcedureDeclarator) {
							final Include finalInclude = include;
							((ProcedureDeclarator) symbol).getParameters().stream().forEach(f -> {
								if (f instanceof VariableDeclaration) {
									((VariableDeclaration) f).getSpecifiers().forEach(s -> {
										if (s instanceof UserSpecifier) {
											String argumentType = ((UserSpecifier) s).getIDExpression().getName();
											if (possibleIncludes.containsKey(finalInclude)) {
												possibleIncludes.get(finalInclude).add(argumentType);
											} else {
												Set<String> argTypes = new HashSet<>();
												argTypes.add(argumentType);
												possibleIncludes.put(finalInclude, argTypes);
											}

										} else if (s instanceof Specifier) {
											if (((Specifier) s).toString().equalsIgnoreCase(Constants.C.VOID)) {
												if (!possibleIncludes.containsKey(finalInclude)) {
													Set<String> argTypes = new HashSet<>();
													argTypes.add(Constants.C.UNDEFINED);
													possibleIncludes.put(finalInclude, argTypes);
												}
											}
										}
									});
								}
							});
						}
					}
				}
			}
		}
		return possibleIncludes;
	}

	private Set<Procedure> findISRs(Model project) {
		Set<Procedure> isrProcedures;
		try {
			isrProcedures = CetusUtils.getStreamOfAllChildren(CetusUtils.getProgram(model), Procedure.class)
					.filter(f -> SpartanMCEnvUtils.matchesExpectedISRIdentifier(f.getDeclarator().getID().getName()))
					.collect(Collectors.toSet());
		} catch (NullPointerException e) {
			model.printIgnoreableError(log,"Nullpointer while parsing ISR methods", e);
			return null;
		}
		if (isrProcedures != null && isrProcedures.isEmpty()) {
			log.debug("No ISR methods have been found");
		}
		return isrProcedures;
	}


}


