/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.peripheralDetection;

import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.tools.spartanmc.SpartanMCEnvUtils;

import cetus.hir.FunctionCall;
import cetus.hir.Identifier;
import cetus.hir.TranslationUnit;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * A wrapper for the libdevxml library that helps to search for peripherals by constants and function calls
 */
public final class PeripheralLibrary {

	private static final Logger log = LoggerFactory.getLogger(PeripheralLibrary.class);
	//TODO: Fix compilation of USB ...antlr grammar should support attribute packed
	private static final Set<String> peripheralHeaderBlacklist = new HashSet<>(Arrays.asList("usb","usb11"));
	private static final Set<String> peripheralConstantsBlacklist = new HashSet<>();

	private List<PeripheralModuleDescription> peripheralLibrary;
	private TranslationUnit peripheralIncludesTranslationUnit;
	private Map<Path, Set<PeripheralModuleDescription>> includeFileToPeripheralLookup;

	/**
	 * The constructor and initializer
	 * @param additionalSystemLibraries additional c libs for init
	 */
	public PeripheralLibrary(Collection<File> additionalSystemLibraries) {
		includeFileToPeripheralLookup = SpartanMCEnvUtils.getIncludeFileToPeripheralMap(peripheralHeaderBlacklist);
		peripheralLibrary = SpartanMCEnvUtils.getPeripheralLibrary(peripheralConstantsBlacklist);
		log.info("Peripheral lib scanned: found " + includeFileToPeripheralLookup.keySet().size() + " C-headers");
		log.info("                          and " + peripheralLibrary.size() + " peripherals for define scan");
		peripheralIncludesTranslationUnit = SpartanMCEnvUtils.initPeripheralIncludesTranslationUnit(includeFileToPeripheralLookup, additionalSystemLibraries);
	}

	/**
	 * Finds a PeriperalModuleDescription by matching the name of the constants with all peripheral names
	 *
	 * @param constantIdentifier The found constants to be matched against the peripheral library
	 * @return null if no fitting peripheral has been found and the PeripheralModuleDescription otherwise
	 */
	public PeripheralModuleDescription findPeriToConstant(Identifier constantIdentifier) {
		if (Constants.JConfigTypes.PRINTF_PERIPHERALS.stream().anyMatch(p -> constantIdentifier.toString().startsWith(p.toUpperCase()))) {
			return null;
		}
		if (SpartanMCEnvUtils.matchesExpectedPeriIdentifier(constantIdentifier)) {
			Optional<PeripheralModuleDescription> longestMatch = peripheralLibrary.stream()
					.filter(p -> constantIdentifier.getName().contains(p.getName().toUpperCase()))
					.sorted(Comparator.comparing(p -> -p.getName().length()))
					.findFirst();
			if (longestMatch.isPresent()) {
				return longestMatch.get();
			}
		}
		return null;
	}

	/**
	 * Finds a PeriperalModuleDescription by matching the name of the constants with all peripheral names
	 *
	 * @param constantIdentifier The found constants to be matched against the peripheral library
	 * @return null if no fitting peripheral has been found and the PeripheralModuleDescription otherwise
	 */
	public PeripheralModuleDescription findPeriToConstant(String constantIdentifier) {
		if (SpartanMCEnvUtils.matchesExpectedPeriIdentifier(constantIdentifier)) {
			Optional<PeripheralModuleDescription> longestMatch = peripheralLibrary.stream()
					.filter(p -> constantIdentifier.contains(p.getName().toUpperCase()))
					.sorted(Comparator.comparing(p -> -p.getName().length()))
					.findFirst();
			if (longestMatch.isPresent()) {
				return longestMatch.get();
			}
		}
		return null;
	}

	public Set<PeripheralModuleDescription> getAllPeripherals() {
		return peripheralLibrary.stream().collect(Collectors.toSet());
	}

	public Set<PeripheralModuleDescription> getInterruptPeripherals() {
		return getAllPeripherals().stream().filter(peri -> SpartanMCEnvUtils.matchesExpectedInterruptPeripheralName(peri.getName())).collect(Collectors.toSet());
	}

	/**
	 * returns a peripheral used for interrupts
	 * @return An peripheral used for interrupts
	 */
	public PeripheralModuleDescription getInterruptPeripheral() {
		Set<PeripheralModuleDescription> matchingPeripherals = getInterruptPeripherals();
		PeripheralModuleDescription pmd = matchingPeripherals.stream().findFirst().get();
		if (matchingPeripherals.size() > 1) {
			log.warn("More than one interrupt peripheral found taking:" + pmd.getName());
		}
		return pmd;
	}

	/**
	 * Searches for matching peripherals
	 * @param type the type of the parameter of the function call
	 * @param name the constants name
	 * @param functionCall a function call to be searched
	 * @return matching peripherals to the parameters
	 */
	public HashSet<PeripheralModuleDescription> findMatchingPeripherals(String type, String name, FunctionCall functionCall) {
		HashSet<PeripheralModuleDescription> matchingIncludes = new HashSet<>();
		includeFileToPeripheralLookup.keySet().stream()
				.filter(f -> f.getFileName().toString().equalsIgnoreCase(name))
				.forEach(f -> {
					includeFileToPeripheralLookup.get(f).forEach(peri -> {
						if (functionCall.getArguments().size() > 0) {
							if ((peri.hasRegisters() && peri.getRegisters().getCSupport().getTypeName().equalsIgnoreCase(type.toString())) || (peri.hasMemory() && peri.getMemory().getCSupport()!=null && peri.getMemory().getCSupport().getTypeName().equalsIgnoreCase(type.toString()))) {
								matchingIncludes.add(peri);
							}
						} else if (type.equalsIgnoreCase(Constants.C.UNDEFINED)) {
							matchingIncludes.add(peri);
						}
					});
				});
		return matchingIncludes;
	}

	/**
	 * Searches for matching peripherals and sort the output
	 * @param type the type of the parameter of the function call
	 * @param name the constants name
	 * @param functionCall a function call to be searched
	 * @param compare a comparator for sorting
	 * @return matching peripherals to the parameters sorted by given comparator
	 */
	public List<PeripheralModuleDescription> findMatchingPeripheralsSorted(String type, String name, FunctionCall functionCall, Comparator compare) {
		List<PeripheralModuleDescription> periList = new ArrayList<>(findMatchingPeripherals(type, name, functionCall));
		Collections.sort(periList, compare);
		return periList;
	}

	public TranslationUnit getPeripheralIncludesTranslationUnit() {
		return peripheralIncludesTranslationUnit;
	}
}
