/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.peripheralDetection;

import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelSubComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.hardware.peripherals.CommonPeripheral;
import de.tu_darmstadt.rs.microStreams.tools.spartanmc.SpartanMCEnvUtils;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

/**
 * Data Structure of all SoC peripherals on all cores
 */
class SoCPeripherals implements SoCPeripheralsInterface {

	private static final Logger log = LoggerFactory.getLogger(SoCPeripherals.class);
	private HashMap<PeripheralModuleDescription, HashSet<String>> usedPeris;
	private Model model;
	private UninferableConstants uninferableConstants;

	public SoCPeripherals(Model model, UninferableConstants uninferableConstants) {
		usedPeris = new HashMap<>();
		this.model = model;
		this.uninferableConstants = uninferableConstants;
	}


	/**
	 * adds all recognized peripheries to the core respectively
	 *
	 * @param core    current Core
	 */
	public void addAllPeriesToCore(Core core) {
		for (Map.Entry<PeripheralModuleDescription, HashSet<String>> entry : usedPeris.entrySet()) {
			int count = 0;
			for (String value : entry.getValue()) {
				String valueTemp = value.toLowerCase();
				HardwareModelSubComponent component = new CommonPeripheral(entry.getKey().getName(), valueTemp, core, count++, model, false);
				final boolean equalEvalComponentAlreadyExists = core.getPeripherals().stream().filter(p -> p.isEvaluationComponent()).anyMatch(p -> p.getType().equals(component.getType()));
				if (equalEvalComponentAlreadyExists) {
					log.warn("Equivalent component to " + component + " already added to core as evaluation component, not adding another");
				} else {
					core.addSubComponent(component);
				}
			}
		}
	}

	public boolean containsInterrupt(Set<PeripheralModuleDescription> interruptPeripherals) {
		return usedPeris.keySet().stream()
				.filter(p -> interruptPeripherals.contains(p))
				.count() > 0;
	}

	public String getNameOf(PeripheralModuleDescription peripheralModuleDescription) {
		//TODO: what if there are several names?
		if (usedPeris.containsKey(peripheralModuleDescription)) {
			return usedPeris.get(peripheralModuleDescription).stream().findFirst().get();
		} else {
			return peripheralModuleDescription.getName() + "_0";
		}
	}

	public boolean hasPeripheralAndName(PeripheralModuleDescription pmd, String name) {
		return usedPeris.containsKey(pmd) && usedPeris.get(pmd).contains(name);
	}

	public PeripheralModuleDescription getPeripheralWithName(String name) {
		for (PeripheralModuleDescription pmd : usedPeris.keySet()) {
			Optional<String> foundName = usedPeris.get(pmd).stream()
					.filter(periName -> periName.equalsIgnoreCase(name))
					.findFirst();
			if (foundName.isPresent()) {
				return pmd;
			}
		}
		return null;
	}

	public void add(PeripheralModuleDescription pmd, String name) {
		if (name != null && !name.isEmpty()) {
			if (pmd == null) {
				if (SpartanMCEnvUtils.matchesExpectedPeriIdentifier(name) && getPeripheralWithName(name) == null) {
					uninferableConstants.add(name);
				}
			} else {
				//UninferableConstants.getInstance().remove(name);
				PeripheralModuleDescription pmdDuplicate = null;
				for (PeripheralModuleDescription existingPmd : usedPeris.keySet()) {
					for (String existingName : usedPeris.get(existingPmd)) {
						if (existingName.equalsIgnoreCase(name)) {
							pmdDuplicate = existingPmd;
						}
					}
				}
				if (pmdDuplicate != null) {
					if (!pmdDuplicate.equals(pmd)) {
						log.warn("The peripheral named " + name + " should be added as " + pmd.getName() + " but it is already recognized as " + pmdDuplicate);
					}
				} else if (usedPeris.containsKey(pmd)) {
					usedPeris.get(pmd).add(name);
				} else {
					HashSet<String> periNames = new HashSet<>();
					periNames.add(name);
					usedPeris.put(pmd, periNames);
				}
			}
		} else {
			add(pmd);
		}
	}

	@Override
	public void add(PeripheralModuleDescription pmd) {
		if (pmd != null) {
			add(pmd, getNameOf(pmd));
		}
	}

	@Override
	public Set<PeripheralModuleDescription> getPeripherals() {
		return usedPeris.keySet();
	}

	public void reset() {
		usedPeris.clear();
	}
}
