/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.microStreams.transformationPasses.input;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.config.IConfiguration;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;
import de.tu_darmstadt.rs.microStreams.tools.Utils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;

import org.apache.commons.io.FilenameUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * Collects all input files and adds them to the Model
 */
public class InputCollector extends AbstractTransformationPass {

	private static final Logger log = LoggerFactory.getLogger(InputCollector.class);

	private Model model = null;

	public InputCollector() {

	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				IConfiguration.class
		);
	}


	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				SourceFile.class
		);
	}

	@Override
	public boolean process(Model m) {
		this.model = m;
		List<SourceFile> foundSourceFiles = getSourceFiles(m.getConfig().getSourceFiles());

		log.info("");
		log.info("Using the following C-Files:");
		log.info("");
		foundSourceFiles.forEach(sf -> {
			m.addModelComponent(sf);
			log.info("  + " + sf.toString());
		});
		log.info("");

		return true;
	}


	/**
	 * Generates a List of Source Files from a given list of Filenames encoded as Strings
	 * @param fileNames files
	 * @return a List of Source Files from a given list of Filenames
	 */
	private List<SourceFile> getSourceFiles(Collection<String> fileNames) {
		List<SourceFile> foundSourceFiles = new ArrayList<>();

		for (String fileName : fileNames) {
			File file = new File(fileName);
			/* distinguish between files and folders */
			if (file.exists() && !file.isDirectory()) {

				boolean extFound = false;

				/* check whether given file extension is supported */
				if (FilenameUtils.getExtension(file.getName()).equalsIgnoreCase(Constants.FilesAndFolders.SOURCEFILES_EXTENSION)) {
					SourceFile sf = new SourceFile(file);
					foundSourceFiles.add(sf);
					extFound = true;
				}

				if (!extFound) {
					model.printIgnoreableError(log,"Type of given file is not supported: "
							+ file.getAbsolutePath());
					System.exit(-1);
				}

			} else if (file.exists() && file.isDirectory()) {

				/* scan for files with one of the supported extensions */
				List<File> foundFiles = Utils.searchFilesByExt(file, Constants.FilesAndFolders.SOURCEFILES_EXTENSION, true);

				for (File foundFile : foundFiles) {
					foundSourceFiles.add(new SourceFile(foundFile));
				}

			} else {
				throw new RuntimeException("The given file or directory \"" + fileName
						+ "\" does not exist");
			}

		}
		return foundSourceFiles;
	}

}
