/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.output;

import de.tu_darmstadt.rs.microStreams.tools.templates.TemplateFile;
import de.tu_darmstadt.rs.microStreams.tools.templates.TemplateLocation;

import freemarker.cache.FileTemplateLoader;
import freemarker.core.ParseException;
import freemarker.template.Configuration;
import freemarker.template.MalformedTemplateNameException;
import freemarker.template.Template;
import freemarker.template.TemplateNotFoundException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;

/**
 * Utils for the BodyBuilder
 */
public final class BuilderUtil {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(BuilderUtil.class);

	private BuilderUtil() {
	}


	/**
	 * give template file loader
	 *
	 * @param templatesRootDir the templates directory
	 * @return The File template loader to the given templates directory
	 */
	static FileTemplateLoader getFileTemplateLoader(String templatesRootDir) {
		try {
			if (null != templatesRootDir) {
				File dir = new File(templatesRootDir);
				if (dir.exists() && dir.isDirectory()) {
					return new FileTemplateLoader(new File(
							templatesRootDir));
				} else {
					throw new RuntimeException("Given template root dir \"" + templatesRootDir + " \" does not exist or is not a directory. Only using class loader");
				}
			}

		} catch (IOException e) {
			throw new RuntimeException("Unable to setup template loader to the given directory. Only using class loader", e);
		}
		return null;
	}

	static String getFileName(TemplateFile templateFile, String templatesRootDir) {
		if (TemplateLocation.FILESYSTEM == TemplateLocation
				.valueOf(templateFile.getLocation())) {
			return new File(templatesRootDir, templateFile.getName()).getName();
		} else if (TemplateLocation.RESSOURCES == TemplateLocation
				.valueOf(templateFile.getLocation())) {
			return templateFile.getName().substring(
					templateFile.getName().lastIndexOf("/") + 1,
					templateFile.getName().length());
		} else {
			throw new RuntimeException("Unknown Template location: "
					+ templateFile.getLocation());
		}

	}

	/**
	 * Loads the template with the Freemarker template system.
	 *
	 * @param templateFile the given tamplate
	 * @return The loaded template. Null if an error occurs.
	 */
	static Template loadTemplate(Configuration fmConfig, TemplateFile templateFile) {
		try {
			return fmConfig.getTemplate(templateFile.getName());
		} catch (TemplateNotFoundException e) {
			log.warn("Template not found: " + templateFile, e);
		} catch (MalformedTemplateNameException e) {
			log.warn("Malformed template name: " + templateFile, e);
		} catch (ParseException e) {
			log.warn(
					"Parse Exception while parsing template: " + templateFile,
					e);
		} catch (IOException e) {
			log.warn("Unable to open file: " + templateFile, e);
		}
		return null;
	}
}
