/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.output;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.config.IConfiguration;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.util.FileUtil;
import de.tu_darmstadt.rs.util.SpartanMCFirmwareUtil;

import org.apache.commons.io.FileUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * This class represents a firmware directory. The directory contains the firmware that
 * should be modified by autoPerf.
 *
 * This class also provides functions to parse the source code with cetus, transform the
 * parsed firmware with autoPerf and write the new firmware to a new directory.
 */
public abstract class FirmwareDuplicator extends AbstractTransformationPass {

	private Logger log = LoggerFactory.getLogger(FirmwareDuplicator.class);
	private Model model;

	/**
	 * Creates a new instance.
	 *
	 */
	public FirmwareDuplicator() {
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				IConfiguration.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				IConfiguration.class
		);
	}

	/**
	 * Delivers all libs that should be added to configBuild.mk file
	 * @return A list of library names to add in the configBuild.mk if nothing has to be added, return null or empty list
	 */
	protected abstract List<String> getConfigBuildMkLibsToAdd();

	/**
	 * Delivers an appendix for the newly created firmware
	 * @return A string, which is appended to the newly created folder
	 */
	protected abstract String getAppendixForNewFirmware();

	@Override
	public boolean process(Model m) {
		model = m;


		if (!write()) {
			model.printIgnoreableError(log,"Writing new firmware failed!");
			return false;
		}


		log.info("Firmware transformed successfully!");

		return true;
	}



	/**
	 * Writes the generated firmware to disk.
	 *
	 * @return true on success, false otherwise.
	 */
	private boolean write() {
		final Collection<Path> sourceFiles = model.getConfig().getSourceFilesAsPath();
		final Path sourceFolder = SpartanMCFirmwareUtil.findSpartanMCFirmwareRootDir(sourceFiles.toArray(new Path[sourceFiles.size()]));

		boolean allSourcesContainedInSourceFolder = sourceFiles.stream().allMatch(sf -> sf.toAbsolutePath().startsWith(sourceFolder.toAbsolutePath()));
		if (!allSourcesContainedInSourceFolder) {
			model.printIgnoreableError(log,"Can only copy firmware if all source files reside in the same firmware folder");
			return false;
		}

		String appendix = getAppendixForNewFirmware();
		if (appendix == null) {
			appendix = "";
		}

		final Path targetFolder = Paths.get(model.getConfig().getTargetRootDir()).resolve(sourceFolder.getFileName().toString() + appendix);
		final File configMkSrc = sourceFolder.resolve(SpartanMCFirmwareUtil.CONFIG_BUILD_MAKE_FILE).toFile();
		final File configMkDst = targetFolder.resolve(SpartanMCFirmwareUtil.CONFIG_BUILD_MAKE_FILE).toFile();
		try {
			FileUtils.copyDirectory(sourceFolder.toFile(), targetFolder.toFile());

			final List<String> configBuildMkLibsToAdd = getConfigBuildMkLibsToAdd();
			if (configBuildMkLibsToAdd != null && !configBuildMkLibsToAdd.isEmpty()) {
				String replacement = configBuildMkLibsToAdd.stream().collect(Collectors.joining(" ", " " , " "));
				FileUtil.searchLineAndAppendInStream(Constants.SpartanMC.CONFIG_BUILD_MK_LIBRARY_REGEX, replacement, configMkSrc, configMkDst);
			}
		} catch (IOException e) {
			model.printIgnoreableError(log,"Can not copy firmware directory!" + e);
			return false;
		}

		model.getConfig().clearSourceFiles();
		model.getConfig().addSourceFileAsPath(Arrays.asList(targetFolder));
		//model.addModelComponent(new FirmwareFolder(firmwareFolder.getTargetPath()));
		return true;
	}
}
