/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.output;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.freemarker.HardwareObjectWrapper;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelSubComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.tools.templates.FilePlacingAndNamingStrategy;
import de.tu_darmstadt.rs.microStreams.tools.templates.HardwareDefaultPlacing;
import de.tu_darmstadt.rs.microStreams.tools.templates.TemplateFile;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.util.FileUtil;

import freemarker.cache.ClassTemplateLoader;
import freemarker.cache.MultiTemplateLoader;
import freemarker.cache.TemplateLoader;
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Arrays;
import java.util.List;

/**
 * Builds the hardware xml output
 */
public class HardwareBuilder extends AbstractTransformationPass {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(HardwareBuilder.class);

	/**
	 * Firmware Target Directory
	 */
	private String targetRootDir;

	/**
	 * Directory containing the c templates
	 */
	private String templatesRootDir = null;

	/**
	 * The placing and naming strategy for the templates.
	 */
	private FilePlacingAndNamingStrategy placing;

	/**
	 * Freemarker configuration
	 */
	private Configuration fmConfig;


	public HardwareBuilder() {
	}

	@Override
	public boolean init() {

		placing = new HardwareDefaultPlacing();

		fmConfig = new Configuration(Configuration.VERSION_2_3_22);

		fmConfig.setTemplateExceptionHandler(TemplateExceptionHandler.DEBUG_HANDLER);

		fmConfig.setObjectWrapper(new HardwareObjectWrapper(fmConfig
				.getIncompatibleImprovements()));

		TemplateLoader[] loaders = new TemplateLoader[2];

		loaders[0] = new ClassTemplateLoader(this.getClass().getClassLoader(),
				"templates/hardware");

		loaders[1] = BuilderUtil.getFileTemplateLoader(templatesRootDir);

		if (null != loaders[1]) {
			fmConfig.setTemplateLoader(new MultiTemplateLoader(loaders));
		} else {
			/* just use the class loader */
			fmConfig.setTemplateLoader(loaders[0]);
		}

		fmConfig.setDefaultEncoding(FileUtil.DEFAULT_FILE_ENCODING);

		return true;
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Core.class,
				HardwareModelSubComponent.class
		);
	}

	@Override
	public boolean process(Model model) {

		/*
		 * Check whether target directory is given as argument, overwrite value
		 * from config in this case.
		 */
		String targetArg = model.getConfig().getTargetRootDir();

		if (null != targetArg) {
			targetRootDir = targetArg;
		}

		File targetFolder = new File(targetRootDir);

		if (!targetFolder.exists()) {
			targetFolder.mkdirs();
		}

		for (TemplateFile templateFile : Constants.Freemarker.HARDWARE_TEMPLATES) {

			Template t = BuilderUtil.loadTemplate(fmConfig, templateFile);

			String fileName = BuilderUtil.getFileName(templateFile, templatesRootDir);

			File targetFile = placing.place(fileName, targetFolder);

			if (null == t || null == targetFile) {
				return false;
			}

			try {
				OutputStream targetFileStream = new FileOutputStream(targetFile);

				if (null != targetFileStream) {
					try {
						t.process(model, new OutputStreamWriter(
								targetFileStream));

					} catch (TemplateException e) {
						model.printIgnoreableError(log, "Template error while processing "
								+ templateFile + ".");
					} catch (IOException e) {
						model.printIgnoreableError(log,"Could not write to output file "
								+ targetFile.getName() + ".");
					}
				}

			} catch (FileNotFoundException e) {
				// Should not happen...
				model.printIgnoreableError(log,"Output target file not found "
						+ targetFile.getName());
			}
		}

		return true;
	}
}
