/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.output;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.freemarker.SoftwareObjectWrapper;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.HardwareModelSubComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.other.HeaderFile;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.ConnectionPeripheralAccess;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.tools.CUtils;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.templates.FilePlacingAndNamingStrategy;
import de.tu_darmstadt.rs.microStreams.tools.templates.SoftwareDefaultPlacing;
import de.tu_darmstadt.rs.microStreams.tools.templates.TemplateFile;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.EvaluationPeripheralCreator;
import de.tu_darmstadt.rs.util.FileUtil;
import de.tu_darmstadt.rs.util.SpartanMCFirmwareUtil;
import de.tu_darmstadt.rs.util.SpartanMCMakeUtil;

import cetus.hir.TranslationUnit;
import freemarker.cache.ClassTemplateLoader;
import freemarker.cache.MultiTemplateLoader;
import freemarker.cache.TemplateLoader;
import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

/**
 * Builds c and h output files
 */
public class SoftwareBuilder extends AbstractTransformationPass {

	/* slf4j logger */
	private Logger log = LoggerFactory.getLogger(SoftwareBuilder.class);

	/**
	 * The placing and naming strategy for the templates.
	 */
	private FilePlacingAndNamingStrategy placing;

	/**
	 * Root directory for externally stored template files.
	 */
	private String templatesRootDir = null;

	/**
	 * Freemarker configuration
	 */
	private Configuration fmConfig;

	/**
	 * The target root directory.
	 */
	private String targetRootDir;

	private Model model;


	public SoftwareBuilder() {
	}

	@Override
	public boolean init() {

		placing = new SoftwareDefaultPlacing();

		fmConfig = new Configuration(Configuration.VERSION_2_3_22);

		fmConfig.setTemplateExceptionHandler(TemplateExceptionHandler.DEBUG_HANDLER);

		fmConfig.setObjectWrapper(new SoftwareObjectWrapper(fmConfig
				.getIncompatibleImprovements()));

		TemplateLoader[] loaders = new TemplateLoader[2];

		loaders[0] = new ClassTemplateLoader(this.getClass().getClassLoader(),
				"templates/software");

		loaders[1] = BuilderUtil.getFileTemplateLoader(templatesRootDir);

		if (null != loaders[1]) {
			fmConfig.setTemplateLoader(new MultiTemplateLoader(loaders));
		} else {
			/* just use the class loader */
			fmConfig.setTemplateLoader(loaders[0]);
		}

		fmConfig.setDefaultEncoding(FileUtil.DEFAULT_FILE_ENCODING);

		return true;
	}


	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Core.class,
				CetusProgram.class,
				ConnectionPeripheralAccess.class,
				HardwareModelSubComponent.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				Core.class
		);
	}


	@Override
	public List<Class<? extends TransformationPass>> requiredTransformationPasses() {
		return Arrays.asList(
				EvaluationPeripheralCreator.class
		);
	}

	@Override
	public boolean process(Model project) {

		model = project;
		/*
		 * Check whether target directory is given as argument, overwrite value
		 * from config in this case.
		 */
		String targetArg = project.getConfig().getTargetRootDir();

		if (null != targetArg) {
			targetRootDir = targetArg;
		}

		for (Core core : project.getModelComponents(Core.class)) {

			File targetFolder = new File(targetRootDir,
					Constants.FilesAndFolders.TARGET_SUBDIR_PREFIX + core.getId());

			SpartanMCMakeUtil.generateNewFirmware(targetFolder.getName(), project.getConfig().getTargetRootDir(), Constants.FilesAndFolders.OVERWRITE_EXISTING_FILES);

			/* save folder path for hardware generation */
			core.setFirmware(targetFolder.toPath());

			/* Copy source files to firmware folder */
			for (File sourceFile : core.getLocalSourceFiles()) {
				// copy the C file
				File targetFile = placing.place(sourceFile.getName(), targetFolder);
				final TranslationUnit translationUnit = CetusUtils.getTranslationUnitFromFile(sourceFile, project);
				writeTranslationUnitToFile(translationUnit, targetFile);
			}

			final Collection<Include> coreIncludeFiles = core.getIncludeFiles();
			for (Include include : coreIncludeFiles) {
				if (!include.isSystemLibrary()) {
					File headerFile = include.getHeaderFile().getFile();
					File cFile = include.getLocation().getFile();
					File targetFile = null;
					// copy the header file ...copy all headers anyways later
					/*File targetFile = placing.place(include.getName(), targetFolder);
					copyFiles(targetFolder, headerFile, targetFile);*/

					// copy the according c-File as well since it is not in a system lib
					targetFile = placing.place(include.getLocation().getFile().getName(), targetFolder);

					final TranslationUnit translationUnit = CetusUtils.getTranslationUnitFromFile(cFile, project);
					writeTranslationUnitToFile(translationUnit, targetFile);
				}
			}

			//copy all remaining headers since headers without functions won't be detected. This should be safe even though the headers are not used.
			project.getModelComponents(HeaderFile.class).stream()
					.forEach(header -> {
						final File targetFile = placing.place(header.getFile().getName(), targetFolder);
						copyFiles(targetFolder, header.getFile(), targetFile);
					});

			final Collection<TemplateFile> softwareTemplates = Constants.Freemarker.SOFTWARE_TEMPLATES.stream()
					.filter(templateFile -> !(templateFile.isEvaluationComponent() && !project.getConfig().isCreatePerformanceEvaluation()))
					.filter(templateFile -> !templateFile.isGlobalVariablesHeader() || (project.getConfig().useGlobalMemory() && core.hasGlobalMemory()))
					.collect(Collectors.toList());
			for (TemplateFile templateFile : softwareTemplates) {
				if (!processTemplateFile(core, templateFile, targetFolder)) {
					return false;
				}
			}

			//copy config-build.mk from the source firmware to the new firmware
			List<Path> sourceFiles = project.getModelComponents(SourceFile.class).stream().map(sourceFile -> sourceFile.getFile().toPath()).collect(Collectors.toList());
			File configMkSrc = SpartanMCFirmwareUtil.findSpartanMCFirmwareRootDir(sourceFiles.toArray(new Path[sourceFiles.size()])).resolve(SpartanMCFirmwareUtil.CONFIG_BUILD_MAKE_FILE).toFile();
			File configMkDst = targetFolder.toPath().resolve(SpartanMCFirmwareUtil.CONFIG_BUILD_MAKE_FILE).toFile();

			String libs = coreIncludeFiles.stream().filter(i -> i.isSystemLibrary() && i.inConfigBuild()).map(i -> i.getLibrary()).collect(Collectors.joining(" "));
			FileUtil.searchLineAndAppendInStream(Constants.SpartanMC.CONFIG_BUILD_MK_LIBRARY_REGEX, " " + libs + " ", configMkSrc, configMkDst);
		}
		return true;
	}

	private void writeTranslationUnitToFile(TranslationUnit translationUnit, File targetFile) {
		try {
			PrintWriter pw =  new PrintWriter(Files.newBufferedWriter(targetFile.toPath()));
			translationUnit.print(pw);
			pw.close();
			CUtils.reformatCode(targetFile);
		} catch (IOException e) {
			model.printIgnoreableError(log,"Something went wrong writing file: " + targetFile, e);
		}
	}

	private void copyFiles(File targetFolder, File file, File targetFile) {
		try {
			if (targetFile.exists()) {
				if (Constants.FilesAndFolders.OVERWRITE_EXISTING_FILES) {
					//CHECKSTYLE.OFF: MultipleStringLiterals
					log.warn("Already existing file " + targetFile.getName() + " will be deleted now.");
				} else {
					model.printIgnoreableError(log,"Already existing file " + targetFile.getName() + " can not be overwritten due to the set overwrite policy");
					return;
					//CHECKSTYLE.ON: MultipleStringLiterals
				}
			}
			targetFile.mkdirs();
			Files.copy(file.toPath(), targetFile.toPath(), java.nio.file.StandardCopyOption.REPLACE_EXISTING);
		} catch (IOException e) {
			model.printIgnoreableError(log,"Could not copy file " + file.getName() + " to " + targetFolder);
		}
	}

	private boolean processTemplateFile(Core core, TemplateFile templateFile,
										File targetFolder) {
		Template t = BuilderUtil.loadTemplate(fmConfig, templateFile);

		String fileName = BuilderUtil.getFileName(templateFile, templatesRootDir);

		if (null == t && null == fileName) {
			return false;
		}

		File targetFile = placing.place(fileName, targetFolder);

		if (null == targetFile) {
			return false;
		}

		if (targetFile.exists()) {
			//CHECKSTYLE.OFF: MultipleStringLiterals
			log.warn("Already existing output file " + targetFile.getName()
					+ " will be deleted now.");
			//CHECKSTYLE.ON: MultipleStringLiterals
			targetFile.delete();
		}

		try {
			targetFile.getParentFile().mkdirs();
			targetFile.createNewFile();
		} catch (IOException e) {
			model.printIgnoreableError(log,"Could not create new file.", e);
			return false;
		}

		try {
			OutputStream targetFileStream = new FileOutputStream(targetFile);

			if (null != targetFileStream) {
				try {
					t.process(core, new OutputStreamWriter(targetFileStream));

					CUtils.reformatCode(targetFile);

				} catch (TemplateException e) {
					model.printIgnoreableError(log,"Template error while processing " + templateFile + ".", e);
				} catch (IOException e) {
					model.printIgnoreableError(log,"Could not write to output file " + targetFile.getName() + ".", e);
				}
			}

		} catch (FileNotFoundException e) {
			// Should not happen...
			model.printIgnoreableError(log,"Output target file not found " + targetFile.getName());
		}

		return true;
	}


}
