/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.output.graph;

import de.tu_darmstadt.rs.microStreams.config.IGuiInspectorConfiguration;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowEnd;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowNode;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowStart;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowTask;
import de.tu_darmstadt.rs.microStreams.model.software.drivers.ConnectionPeripheralAccess;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency.DependencyType;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;

import cetus.hir.AnnotationStatement;
import edu.uci.ics.jung.algorithms.layout.FRLayout;
import edu.uci.ics.jung.algorithms.layout.KKLayout;
import edu.uci.ics.jung.algorithms.layout.Layout;
import edu.uci.ics.jung.graph.DirectedOrderedSparseMultigraph;
import edu.uci.ics.jung.graph.Graph;
import edu.uci.ics.jung.graph.util.EdgeType;
import edu.uci.ics.jung.visualization.VisualizationImageServer;
import edu.uci.ics.jung.visualization.VisualizationViewer;
import edu.uci.ics.jung.visualization.control.DefaultModalGraphMouse;
import edu.uci.ics.jung.visualization.control.ModalGraphMouse;
import edu.uci.ics.jung.visualization.decorators.ToStringLabeller;
import edu.uci.ics.jung.visualization.renderers.VertexLabelAsShapeRenderer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import javax.imageio.ImageIO;

/**
 * Draws the dependencies of µStreams tasks
 */
public final class MicroStreamsDrawer extends AbstractTransformationPass {

	private static final int dimension = 500;
	private int count = 0;
	private boolean showGUI = false;

	public MicroStreamsDrawer() {

	}

	/**
	 * Draws the task dependencies in a new frame with all variable dependencies
	 *
	 * @param m a model
	 */
	public void drawDependecies(Model m) {
		Collection<Task> tasks = m.getModelComponents(Task.class);
		Graph<Task, TaskDependency> g = new DirectedOrderedSparseMultigraph<Task, TaskDependency>();

		for (Task t : tasks) {
			g.addVertex(t);

		}

		for (Task t : tasks) {
			for (TaskDependency d : t.getDependencies()) {
				if (DependencyType.ONE_TO_ONE == d.getType()) {
					if (d.getProvider() == t) {
						g.addEdge(d, t, d.getConsumer());
					}
				} else if (DependencyType.MANY_TO_ONE == d.getType()) {
					if (d.getProvider() == t) {
						g.addEdge(d, t, d.getConsumer());
					} else if ((t.isReplica() && d.getProvider() == t
							.getOriginal())) {
						g.addEdge(new TaskDependencyReplica(d), t, d.getConsumer());

					}
				} else if (DependencyType.ONE_TO_MANY == d.getType()) {
					if (d.getProvider() == t) {
						g.addEdge(d, t, d.getConsumer());
						for (Task r : d.getReplica()) {
							g.addEdge(new TaskDependencyReplica(d), t, r);
						}
					}

				}
			}
		}

		// The Layout<V, E> is parameterized by the vertex and edge types
		Layout<Task, TaskDependency> layout = new FRLayout<>(g);
		layout.setSize(new Dimension(dimension, dimension)); // sets the initial size of the
		// space
		// The BasicVisualizationServer<V,E> is parameterized by the edge types
		//VisualizationViewer<Task, TaskDependency> vv = new VisualizationViewer<>(layout);
		VisualizationImageServer<Task, TaskDependency> vv = new VisualizationImageServer<>(layout, new Dimension(dimension, dimension));

		//vv.setVertexToolTipTransformer(Task::toString);

		vv.setPreferredSize(new Dimension(dimension, dimension)); // Sets the viewing area
		// size

		DefaultModalGraphMouse<InnerFlowNode, InnerFlowEdge> gm = new DefaultModalGraphMouse<>();

		gm.setMode(ModalGraphMouse.Mode.PICKING);
		//vv.setGraphMouse(gm);
		//vv.addKeyListener(gm.getModeKeyListener());

		gm.add(new PopupGraphMousePlugin());   // Add our new plugin to the mouse

		vv.getRenderContext().setVertexFillPaintTransformer(task -> {
			if (task.isBaseTask()) {
				return Color.GREEN;
			} else if (task.isReplica()) {
				return Color.LIGHT_GRAY;
			} else if (task.hasReplica()) {
				return Color.GRAY;
			} else if (task.isIndependent()) {
				return Color.ORANGE;
			} else if (task.getConsumers().size() == 0) {
				return Color.RED;
			}
			return Color.CYAN;
		});

		VertexLabelAsShapeRenderer<Task, TaskDependency> vlasr = new VertexLabelAsShapeRenderer<>(vv.getRenderContext());

		vv.getRenderContext().setVertexLabelTransformer(task -> "<html><center>" + task.toString() + "<p><br><br>");
		vv.getRenderContext().setVertexShapeTransformer(vlasr);
		vv.getRenderer().setVertexLabelRenderer(vlasr);


		vv.getRenderContext().setEdgeLabelTransformer(taskDependency -> {
			if (taskDependency.getVariableOfProvider() != taskDependency.getVariableOfConsumer()) {
				return taskDependency.getVariableOfProvider().getName() + "->" + taskDependency.getVariableOfConsumer().getName();
			} else {
				return taskDependency.getVariableOfProvider().getName();
			}
		});




		BufferedImage image = (BufferedImage)vv.getImage(new Point2D() {
			@Override
			public double getX() {
				return 0;
			}

			@Override
			public double getY() {
				return 0;
			}

			@Override
			public void setLocation(double x, double y) {

			}
		}, new Dimension(dimension, dimension));

		File outputfile = new File(m.getConfig().getTargetRootDir() + "/pipeline.png");
		try {
			ImageIO.write(image, "png", outputfile);
		} catch (IOException e) {
			e.printStackTrace();
		}


		MicroStreamsGraphGUI window = new MicroStreamsGraphGUI(vv);
		window.title("Task dependency graph " + (++count));
		if (showGUI) {
			window.show();
		}
	}

	/**
	 * Draws the inner flow graph in an extra frame
	 *
	 * @param task a task to be drawn
	 */
	public static void drawInnerFlow(Task task) {

		InnerFlowGraph ifg = task.getInnerFlowGraph();
		Graph<InnerFlowNode, InnerFlowEdge> g = new DirectedOrderedSparseMultigraph<InnerFlowNode, InnerFlowEdge>();
		Collection<InnerFlowTaskReplicate> replicates = new ArrayList<InnerFlowTaskReplicate>();

		for (InnerFlowNode n : ifg.getAllNodes()) {
			if (n instanceof InnerFlowTask) {
				Task t = ((InnerFlowTask) n).getTask();
				if (t.hasReplica()) {
					for (Task r : t.getReplica()) {
						InnerFlowTaskReplicate replicate = new InnerFlowTaskReplicate(
								(InnerFlowTask) n, r);
						replicates.add(replicate);
						g.addVertex(replicate);
					}
				}
			}

			g.addVertex(n);
		}

		for (InnerFlowNode n : ifg.getAllNodes()) {

			for (InnerFlowNode s : n.getSucccessors()) {

				for (InnerFlowTaskReplicate rr : replicates) {
					if (rr.getOriginal() == s) {
						g.addEdge(new InnerFlowEdge(n, rr), n, rr,
								EdgeType.DIRECTED);
					}
				}

				for (InnerFlowTaskReplicate rr : replicates) {
					if (rr.getOriginal() == n) {
						g.addEdge(new InnerFlowEdge(rr, s), rr, s,
								EdgeType.DIRECTED);
					}
				}

				g.addEdge(new InnerFlowEdge(n, s), n, s, EdgeType.DIRECTED);

			}

		}

		// The Layout<V, E> is parameterized by the vertex and edge types
		Layout<InnerFlowNode, InnerFlowEdge> layout = new KKLayout<>(g);
		layout.setSize(new Dimension(dimension, dimension)); // sets the initial size of the
		// space
		// The BasicVisualizationServer<V,E> is parameterized by the edge types
		VisualizationViewer<InnerFlowNode, InnerFlowEdge> vv = new VisualizationViewer<>(layout);

		vv.setVertexToolTipTransformer(n1 -> {
			if (n1 instanceof InnerFlowTask) {
				return "Call of the ChildTask Task #"
						+ ((InnerFlowTask) n1).getTask().getID();
			} else if (n1 instanceof InnerFlowStart) {
				return "This is the start node of the graph";
			} else if (n1 instanceof InnerFlowEnd) {
				return "This is the end node of the graph";
			} else if (n1 instanceof InnerFlowNode) {

				return CetusPrinter.print(n1.getTraversable());
			}
			return "ERROR";
		});

		vv.setPreferredSize(new Dimension(dimension, dimension)); // Sets the viewing area
		// size

		DefaultModalGraphMouse<InnerFlowNode, InnerFlowEdge> gm = new DefaultModalGraphMouse<>();
		gm.setMode(edu.uci.ics.jung.visualization.control.ModalGraphMouse.Mode.TRANSFORMING);
		vv.setGraphMouse(gm);

		vv.getRenderContext().setVertexLabelTransformer(new ToStringLabeller());

		vv.getRenderContext().setVertexFillPaintTransformer(n -> {
			if (n instanceof InnerFlowTask) {
				return Color.BLACK;
			} else if (n instanceof InnerFlowStart) {
				return Color.GREEN;
			} else if (n instanceof InnerFlowEnd) {
				return Color.RED;
			} else if (n instanceof InnerFlowNode) {

				if (n.getTraversable() instanceof AnnotationStatement) {
					return Color.LIGHT_GRAY;
				}

				return Color.WHITE;
			}
			return Color.PINK;
		});

		MicroStreamsGraphGUI window = new MicroStreamsGraphGUI(vv);
		window.title("InnerFlowGraph of Task #" + task.getID());
		window.show();

	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				ConnectionPeripheralAccess.class
		);
	}

	@Override
	public boolean process(Model m) {
		if (m.getModelComponent(IGuiInspectorConfiguration.class).inspectGeneratedModel()) {
			drawDependecies(m);
		}
		return true;
	}

	/**
	 * An extension to task dependency to handle replicas
	 */
	private static class TaskDependencyReplica extends TaskDependency {

		public TaskDependencyReplica(TaskDependency d) {
			super(d.getConsumer(), d.getProvider(), d.getVariableOfProvider(), Optional.of(d.getVariableOfConsumer()), d
					.getRelation(), d.getLocation());

		}

	}


	/**
	 * An edge between inner flow nodes
	 */
	private static class InnerFlowEdge {

		private InnerFlowNode start;
		private InnerFlowNode ende;

		public InnerFlowEdge(InnerFlowNode start, InnerFlowNode ende) {
			this.start = start;
			this.ende = ende;
		}

	}


	/**
	 * Extension for the replica inner flow task
	 */
	private static class InnerFlowTaskReplicate extends InnerFlowTask {

		private InnerFlowNode orig;

		public InnerFlowTaskReplicate(InnerFlowTask orig, Task task) {
			super(task);
			this.orig = orig;
		}

		public InnerFlowNode getOriginal() {
			return orig;
		}

	}

}
