/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.interfaces.ContainsChildTask;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunctionCall;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;

import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

/**
 * <p>
 * This {@link TransformationPass} generates dependencies between {@link Task}s by
 * determining which task contains other tasks. This parent child relationship
 * is stored within the {@link Model}. After running this {@link TransformationPass} each
 * task stores its parent task and a list of its child tasks. Special case: the
 * base task does not have a parent task. So its corresponding field will be
 * <code>null</code>.
 * </p>
 * <p>
 * <p>
 * Further on it is necessary to distinguished between parent task and the so
 * called direct parent. In addition to the parent task and child tasks each
 * task also stores its direct parent after the execution of this {@link TransformationPass}
 * . It denotes whether the task is nested directly in its parent task or a
 * function called by its parent task. The direct parent has to implement the
 * {@link ContainsChildTask} interface, which is done by {@link Task} and
 * {@link UFunction}. The base task's direct parent field will be also
 * <code>null</code> due to the lack of parent task.
 * </p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> The {@link Model} already contains all {@link Task
 * Tasks} and {@link UFunction Functions}. The
 * {@link ChildtaskCallgraph#processable(Model)} method checks whether the
 * {@link TaskCreator} and the {@link FunctionCallgraph} is applied to the
 * {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class ChildtaskCallgraph extends AbstractTransformationPass {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(ChildtaskCallgraph.class);

	private Collection<Task> tasks;

	private Map<Task, Collection<Task>> indirectChildTasks;

	private Model model;


	public ChildtaskCallgraph() {
	}

	/**
	 * Utilizes internally the {@link ChildtaskCallgraph#getContainedTasks(Task)} )}
	 * method to extract all direct child tasks and stores them in to given
	 * parent task.
	 *
	 * @param task The parent task whose child task shall be found and added.
	 */
	private void addAllDirectChildren(Task task) {
		Collection<Task> childTasks = getContainedTasks(task);

		for (Task childTask : childTasks) {
			/* Add child to parent task, */
			task.addTask(childTask);
			/* set the direct parent of the child task ... */
			childTask.setDirectParent(task);
			/* ... and the parent of the child task to the parent task. */
			childTask.setParentTask(task);
		}
	}

	/**
	 * Recursively searches all indirect child tasks of the given
	 * {@link UFunction} and adds them to the given Collection of {@link Task}s
	 * foundTasks.
	 *
	 * @param function The function that shall be scanned for child tasks.
	 * @param foundTasks A collection of already found tasks and the final result after execution.
	 * @param visited A collection of already visited {@link UFunction Functions} to avoid endless recursion. At initial call hand in an empty collection.
	 */
	private void addAllIndirectChildren(UFunction function,
										Collection<Task> foundTasks, Collection<UFunction> visited) {
		/*
		 * Store all already visited functions to avoid endless recursion.
		 */
		if (!visited.contains(function)) {
			visited.add(function);

			Collection<UFunctionCall> calls = function.getFunctionCalls();

			Collection<Task> childTasks = function.getContainedTasks();

			foundTasks.addAll(childTasks);

			for (UFunctionCall call : calls) {
				if (!call.isExtern()) {
					UFunction childFunction = call.getFunction();
					addAllIndirectChildren(childFunction, foundTasks, visited);
				}
			}
		}
	}

	/**
	 * Checks whether all child task collections in the
	 * {@link ChildtaskCallgraph#indirectChildTasks} field are disjoint, and
	 * returns the result as boolean.
	 *
	 * @return Returns <code>true</code> if the collections are disjoint,
	 * <code>false</code> otherwise.
	 */
	private boolean areIndirectChildTasksDisjoint() {
		for (Task task1 : indirectChildTasks.keySet()) {
			Collection<Task> childTasks1 = indirectChildTasks.get(task1);
			for (Task task2 : indirectChildTasks.keySet()) {
				Collection<Task> childTasks2 = indirectChildTasks.get(task2);
				if (!task1.equals(task2)
						&& !Collections.disjoint(childTasks1, childTasks2)) {
					Collection<Task> notDisjoint = new LinkedHashSet<Task>(
							childTasks1);
					notDisjoint.retainAll(childTasks2);

					// TODO: replicate task instead of forbidding... Possible
					// problem: globals or statics inside of the task.

					model.printIgnoreableError(log,notDisjoint
							+ ((notDisjoint.size() > 1)? "are" : "is")
							+ " indirect Childtasks of Task " + task1
							+ " as well as task " + task2
							+ ". This is not allowed.");

					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Scans the given task's {@link Traversable} for contained child tasks and
	 * returns the result. Note: that only direct tasks are returned, children
	 * contained in function calls are not returned, children's children are
	 * also not returned.
	 *
	 * @param task - scanned Task
	 * @return cans the given task's {@link Traversable} for contained child tasks and returns the result
	 */
	private Collection<Task> getContainedTasks(Task task) {
		Collection<Task> childTasks = new LinkedHashSet<Task>();

		for (Task currentTask : tasks) {
			if (!currentTask.equals(task)
					&& isContainedInTask(currentTask.getTraversable(), task)) {
				childTasks.add(currentTask);
			}
		}

		return childTasks;
	}

	/**
	 * Returns the task containing the given {@link Traversable} traversable.
	 *
	 * @param traversable - the {@link Traversable} the task is searched for.
	 * @return Returns the task that contains the given {@link Traversable} traversable.
	 */
	private Task getContainingTask(Traversable traversable) {
		Traversable t = traversable;

		final TranslationUnit mainTranslationUnit = CetusUtils.getLocation(CetusUtils.getMainProcedure(model.getModelComponent(CetusProgram.class).getProgram()));
		boolean traversableNotInMain = !CetusUtils.findParentOfType(traversable, TranslationUnit.class).equals(mainTranslationUnit);
		while (t != null) {
			t = t.getParent();
			for (Task task : tasks) {
				final boolean isBaseTaskEqualsOriginalTraversable = task.isBaseTask() && model.getModelComponent(CetusProgram.class).getProgram().equals(t) && traversableNotInMain;
				if (task.getTraversable().equals(t) || isBaseTaskEqualsOriginalTraversable) {
					return task;
				}
				/*if (task.getTraversable().equals(t)) {
					return task;
				}*/
			}
		}
		return null;
	}

	/**
	 * Checks whether the the given {@link Traversable} traversable is contained
	 * in the given {@link Task} task. If so true is returned, otherwise false.
	 *
	 * @param traversable the traversable.
	 * @param task the tasks that is checked wether it contains the given traversable.
	 * @return true or false depending whether the traversable is contained in the given task.
	 */
	private boolean isContainedInTask(Traversable traversable, Task task) {
		Task containingTask = getContainingTask(traversable);
		return (null != containingTask && containingTask.equals(task));
	}

	@Override
	public boolean init() {

		this.indirectChildTasks = new HashMap<Task, Collection<Task>>();

		return true;
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Task.class,
				UFunction.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				Task.class
		);
	}

	@Override
	public boolean process(Model model) {

		this.model = model;

		this.tasks = model.getModelComponents(Task.class);

		this.indirectChildTasks.clear();

		/*
		 * 1. First add all direct child tasks to all task. Remark: Functions'
		 * child tasks are already added to the corresponding functions.
		 */
		for (Task task : this.tasks) {
			addAllDirectChildren(task);
		}

		/*
		 * 2. Searching all indirect child tasks (Childtask by call). Note that
		 * a Task cannot have more than one indirect Parenttask. This will
		 * ensured later on.
		 */
		for (Task task : this.tasks) {
			indirectChildTasks.put(task, new LinkedHashSet<Task>());
			for (UFunctionCall call : task.getFunctionCalls()) {
				if (!call.isExtern()) {
					addAllIndirectChildren(call.getFunction(),
							indirectChildTasks.get(task),
							new LinkedHashSet<UFunction>());
				}
			}
		}

		/*if (!areIndirectChildTasksDisjoint()) {
			return false;
		}*/

		for (Task task : this.tasks) {
			for (Task indirectChildTask : indirectChildTasks.get(task)) {
				/* Add indirect child to parent task */
				task.addTask(indirectChildTask);
				/*
				 * and set the parent task of the indirect child task to parent
				 * task. (Direct parent was set to the function containing the
				 * child task before, see FunctionCallgraph.class.)
				 */
				indirectChildTask.setParentTask(task);
			}
		}

		return true;
	}
}
