/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.interfaces.ContainsChildTask;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowNode;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowTask;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunctionCall;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency.DependencyRelation;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependencyGroup;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.VariableUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.Declaration;
import cetus.hir.Identifier;
import cetus.hir.Traversable;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.function.BiPredicate;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * <p>
 * This {@link TransformationPass} implementation creates the {@link TaskDependency} between
 * parent task and child task depending on the variables read and written from
 * the children and the children's children.
 * <p>
 * <p>
 * These {@link TaskDependency TaskDependencies} still can be optimized. This is
 * done later on in other {@link TransformationPass}s, e.g. the {@link TaskPipeline}.
 * </p>
 * <p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> All {@link Task Tasks} and {@link UFunction Functions}
 * in the {@link Model} contain already their {@link InnerFlowGraph}, also all
 * {@link VariableAccess VariableAccesses} have to be detected and added to the
 * {@link Task Tasks} and {@link UFunction Functions} and their
 * {@link InnerFlowGraph InnerFlowGraphs}. The
 * {@link DependencyCreator#processable(Model)} method checks whether the
 * {@link InnerFlowLinker} and the {@link VariableAccessCreator} are applied to
 * the {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class DependencyCreator extends AbstractTransformationPass {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(DependencyCreator.class);


	public DependencyCreator() {
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				InnerFlowGraph.class,
				Task.class,
				VariableAccess.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				TaskDependencyGroup.class
		);
	}

	@Override
	public boolean process(Model project) {
		Collection<Task> tasks = project.getModelComponents(Task.class);

		for (Task task : tasks) {
			/*
			 * Base task has no dependencies/parent task, don't handle replica
			 * just the original tasks.
			 */
			if (task.isBaseTask()) {
				baseTaskDependencyAnalysis(task, project);
			}

			if (!task.isBaseTask() && !task.isReplica()) {
				Task parent = task.getParentTask();

				Collection<Task> replica = task.getReplica();

				Collection<Variable> readVariables = task
						.getCombinedReadVariables();
				Collection<Variable> writtenVariables = task
						.getCombinedWriteVariables();

				TaskDependencyGroup group = new TaskDependencyGroup();

				// TODO: calls without var! DUMMYVARIABLE!
				for (Variable readVariable : readVariables) {
					/*
					 * if the variable is forced read or written by other tasks
					 * before, add dependency
					 */
					boolean forced = task.getForcedReadVariables().contains(readVariable);
					final VariableOccurancePoint beforeVariableAnalysis = beforeVariableAnalysis(task, readVariable, InnerFlowNode::getWrittenVariables);
					if (forced || beforeVariableAnalysis.isOccurancePointFound()) {
						//CHECKSTYLE.OFF: MultipleStringLiterals
						log.debug("creating task dependency on " + readVariable + ", depending task: " + task);
						//CHECKSTYLE.ON: MultipleStringLiterals
						TaskDependency dependency;
						if (beforeVariableAnalysis.isOccurancePointFound() && beforeVariableAnalysis.getTask().isPresent()) {
							dependency = new TaskDependency(task,
									beforeVariableAnalysis.getTask().get(), readVariable,
									DependencyRelation.CHILD_CHILD,
									task.getDirectParent());
							setDependencies(dependency, task, beforeVariableAnalysis.getTask().get(), group, forced, project);
							//TODO differentiate getA true and b is not present
						} else {
							dependency = new TaskDependency(task,
									parent, readVariable,
									DependencyRelation.PARENT_CHILD,
									task.getDirectParent());
							setDependencies(dependency, task, parent, group, forced, project);
						}


						/*
						 * if the task was replicated add the dependency also to
						 * the replica.
						 */
						if (task.hasReplica()) {
							for (Task t : replica) {
								t.addDependency(dependency);
							}
						}

					} else {
						//CHECKSTYLE.OFF: MultipleStringLiterals
						log.debug("Not adding READ dependency for variable "
								+ readVariable
								+ " (CHILD:"
								+ task
								+ " and  PARENT:"
								+ parent
								+ "), because it seems it is not neccessary (not beforeVariableAnalysis).");
						//CHECKSTYLE.ON: MultipleStringLiterals
					}

				}

				if (!group.isEmpty()) {
					project.addModelComponent(group);
				}

				group = new TaskDependencyGroup();


				for (Variable writtenVariable : writtenVariables) {
					/*
					 * if the variable is forced write or read after other tasks
					 * before, add dependency
					 */
					boolean forced = task.getForcedWriteVariables().contains(writtenVariable);
					VariableOccurancePoint isArrayAndWrittenBefore =  isArrayOrStructAndWrittenBefore(task, writtenVariable);
					final BiPredicate<InnerFlowNode, Variable> isVariableWrittenAndNotRead = (node, variable) -> {
						final boolean isMyOwnDeclaration = CetusUtils.getStreamOfAllChildren(node.getTraversable(), VariableDeclaration.class).anyMatch(vd -> vd == variable.getVariableDeclarator().getDeclaration());
						return (node.getWrittenVariables().contains(variable) && !node.getReadVariables().contains(variable)) || isMyOwnDeclaration;
					};
					final VariableOccurancePoint afterVariableAnalysis = afterVariableAnalysis(task, writtenVariable, InnerFlowNode::getReadVariables, isVariableWrittenAndNotRead);
					if (forced || afterVariableAnalysis.isOccurancePointFound() || isArrayAndWrittenBefore.isOccurancePointFound()) {
						List<TaskDependency> dependencies = new ArrayList<>();
						if (isArrayAndWrittenBefore.isOccurancePointFound()) {
							if (isArrayAndWrittenBefore.getTask().isPresent()) {
								TaskDependency dependency = new TaskDependency(task, isArrayAndWrittenBefore.getTask().get(),
										writtenVariable, isArrayAndWrittenBefore.getVariableAtTask(),
										DependencyRelation.CHILD_CHILD,
										task.getDirectParent());
								//CHECKSTYLE.OFF: MultipleStringLiterals
								log.debug("creating task dependency on " + writtenVariable + ", depending task: " + task);
								//CHECKSTYLE.ON: MultipleStringLiterals
								setDependencies(dependency, isArrayAndWrittenBefore.getTask().get(), task, group, forced, project);
								dependencies.add(dependency);
							} else {
								TaskDependency dependency = new TaskDependency(task, parent,
										writtenVariable,
										DependencyRelation.PARENT_CHILD,
										task.getDirectParent());
								//CHECKSTYLE.OFF: MultipleStringLiterals
								log.debug("creating task dependency on " + writtenVariable + ", depending task: " + task);
								//CHECKSTYLE.ON: MultipleStringLiterals
								setDependencies(dependency, parent, task, group, forced, project);
								dependencies.add(dependency);
							}
						}
						if (afterVariableAnalysis.isOccurancePointFound()) {
							if (afterVariableAnalysis.getTask().isPresent()) {
								DependencyRelation relation = DependencyRelation.CHILD_CHILD;
								if (parent.equals(afterVariableAnalysis.getTask().get())) {
									relation = DependencyRelation.CHILD_PARENT;
								}
								TaskDependency dependency = new TaskDependency(afterVariableAnalysis.getTask().get(),
										task, writtenVariable, afterVariableAnalysis.getVariableAtTask(),
										relation,
										task.getDirectParent());
								//CHECKSTYLE.OFF: MultipleStringLiterals
								log.debug("creating task dependency on " + writtenVariable + ", depending task: " + task);
								//CHECKSTYLE.ON: MultipleStringLiterals
								setDependencies(dependency, afterVariableAnalysis.getTask().get(), task, group, forced, project);
								dependencies.add(dependency);
							} else {
								TaskDependency dependency = new TaskDependency(parent,
										task, writtenVariable,
										DependencyRelation.CHILD_PARENT,
										task.getDirectParent());
								//CHECKSTYLE.OFF: MultipleStringLiterals
								log.debug("creating task dependency on " + writtenVariable + ", depending task: " + task);
								//CHECKSTYLE.ON: MultipleStringLiterals
								setDependencies(dependency, parent, task, group, forced, project);
								dependencies.add(dependency);
							}

						}
						if (forced) {
							TaskDependency dependency = new TaskDependency(parent,
									task, writtenVariable,
									DependencyRelation.CHILD_PARENT,
									task.getDirectParent());
							//CHECKSTYLE.OFF: MultipleStringLiterals
							log.debug("creating task dependency on " + writtenVariable + ", depending task: " + task);
							//CHECKSTYLE.ON: MultipleStringLiterals
							setDependencies(dependency, parent, task, group, forced, project);
							dependencies.add(dependency);
						}


						/*
						 * if the task was replicated create one dependency for
						 * all replica
						 */
						if (task.hasReplica()) {
							for (Task t : replica) {
								for (TaskDependency dependency : dependencies) {
									t.addDependency(dependency);
								}
							}
						}
					} else {
						//CHECKSTYLE.OFF: MultipleStringLiterals
						log.debug("Not adding WRITE dependency for variable "
								+ writtenVariable
								+ " (CHILD:"
								+ task
								+ " and  PARENT:"
								+ parent
								+ "), because it seems it is not neccessary (not read after it).");
						//CHECKSTYLE.ON: MultipleStringLiterals
					}
				}

				if (!group.isEmpty()) {
					project.addModelComponent(group);
				}
			}
		}
		if (!project.hasModelComponent(TaskDependencyGroup.class)) {
			project.addModelComponent(new TaskDependencyGroup());
		}

		return true;
	}

	private void setDependencies(TaskDependency dependency, Task task, Task parent, TaskDependencyGroup group, boolean forced, Model project) {
		dependency.setForced(forced);
		parent.addDependency(dependency);
		task.addDependency(dependency);
		project.addModelComponent(dependency);
		group.addDependency(dependency);
	}

	private void baseTaskDependencyAnalysis(Task baseTask, Model project) {
		if (baseTask.isBaseTask()) {

			TaskDependencyGroup group = new TaskDependencyGroup();
			final InnerFlowGraph innerFlowGraph = baseTask.getInnerFlowGraph();

			for (InnerFlowNode node : innerFlowGraph.getAllNodes()) {

				if (!(node instanceof InnerFlowTask)) {

					for (Variable variable : node.getReadVariables()) {

						for (InnerFlowNode predecessor : InnerFlowGraph.getAllPredecessors(node)) {

							if (predecessor.getWrittenVariables().contains(variable)) {

								if (predecessor instanceof InnerFlowTask) {

									InnerFlowTask predecessorTask = (InnerFlowTask) predecessor;
									TaskDependency dependency = new TaskDependency(baseTask, predecessorTask.getTask(), variable, DependencyRelation.CHILD_PARENT, baseTask);
									final List<Task> replicateTasks = project.getModelComponents(Task.class).stream()
											.filter(Task::isReplica)
											.filter(t -> t.getOriginal() == predecessorTask.getTask())
											.collect(Collectors.toList());

									for (Task t : replicateTasks) {
										setDependencies(dependency, baseTask, t, group, false, project);
									}
									setDependencies(dependency, baseTask, predecessorTask.getTask(), group, false, project);
								}
								else {
									break;
								}
								if (!group.isEmpty()) {
									project.addModelComponent(group);
								}
							}
						}
					}
				}
			}

		}
	}

	private VariableOccurancePoint beforeVariableAnalysis(Task task, Variable variable, Function<InnerFlowNode, Collection<Variable>> variablesToCompareTo) {


		VariableOccurancePoint fail = new VariableOccurancePoint(false);
		if (task.isBaseTask()) {
			return fail;
		}

		ContainsChildTask location = task.getDirectParent();
		InnerFlowGraph graph = location.getInnerFlowGraph();
		/* Locate task within location */
		InnerFlowNode node = graph.getInnerFlowNodeWith(task).get();

		for (InnerFlowNode predecessor : InnerFlowGraph.getAllPredecessors(node)) {
			/* If it is the examined task itself, ignore ... */
			final Optional<Task> otherTask = Stream.of(predecessor).flatMap(StreamUtil.castOrEmpty(InnerFlowTask.class)).map(InnerFlowTask::getTask).findFirst();
			if (!(otherTask.isPresent() && otherTask.get().equals(task))) {
				if (variablesToCompareTo.apply(predecessor).contains(variable)) {
					return new VariableOccurancePoint(true, otherTask);
				}
			}
		}

		/*
		 * if it is not found within the parent's traversable. Look at the
		 * parents parents.
		 */
		if (location instanceof Task) {
			return beforeVariableAnalysis((Task) location, variable, variablesToCompareTo);
		} else if (location instanceof UFunction) {
			UFunction function = (UFunction) location;

			final Declaration variableDeclaration = variable.getIdentifiers().stream()
					.map(i -> i.getSymbol().getDeclaration())
					.collect(StreamUtil.exactlyOne());

			boolean variableDeclaredInFunctionArguments = function.getFunctionArguments().stream().anyMatch(a -> a == variableDeclaration);

			if (variableDeclaredInFunctionArguments) {
				return new VariableOccurancePoint(true);
			}

			log.warn("Before Analysis beyond Function boundaries not fully implemented yet! Check Dependencies for variable " + variable + " in " + function.getDeclaration());
			// TODO: in return statement or arguments (pointer)? analyse
		}

		return fail;

	}

	private VariableOccurancePoint afterVariableAnalysis(Task task, Variable variable, Function<InnerFlowNode, Collection<Variable>> variablesToCompareTo, BiPredicate<InnerFlowNode, Variable> abortCriteria) {

		VariableOccurancePoint fail = new VariableOccurancePoint();
		if (task.isBaseTask()) {
			return fail;
		}

		ContainsChildTask location = task.getDirectParent();
		InnerFlowGraph graph = location.getInnerFlowGraph();
		/* Locate task within location */
		InnerFlowNode node = graph.getInnerFlowNodeWith(task).get();

		final Collection<InnerFlowNode> allSuccessors = InnerFlowGraph.getAllSuccessors(node);
		for (InnerFlowNode succ : allSuccessors) {
			//see if we are inside a loop and have a backward dependency to us
			if (abortCriteria.test(succ, variable)) {
				return fail;
			}

			for (InnerFlowNode n : succ.getSucccessors()) {
				if (n instanceof InnerFlowTask) {
					InnerFlowTask innerFlowTask = (InnerFlowTask) n;
					if (variablesToCompareTo.apply(innerFlowTask).contains(variable)) {
						if (innerFlowTask.getTask() != task) {
							return new VariableOccurancePoint(true, innerFlowTask.getTask());
						} else {
							return new VariableOccurancePoint(true);
						}
					}
				}
			}
			/*boolean readVariablesInTask = succ.getSucccessors().stream()
					.filter(s -> s instanceof InnerFlowTask)
					.map(s -> (InnerFlowTask) s)
					.flatMap(t -> variablesToCompareTo.apply(t).stream())
					.anyMatch(var -> var.equals(variable));
			if (readVariablesInTask) {
				return true;
			}*/
			/* If it is the examined task itself, ignore ... */
			final Optional<Task> otherTask = Stream.of(succ).flatMap(StreamUtil.castOrEmpty(InnerFlowTask.class)).map(InnerFlowTask::getTask).findFirst();
			if (!(otherTask.isPresent() && otherTask.get().equals(task))) {
				if (variablesToCompareTo.apply(succ).contains(variable)) {
					return new VariableOccurancePoint(true, otherTask);
				}
			}
		}

		/*
		 * if it is not found within the parent's traversable. Look at the
		 * parents parents.
		 */
		if (location instanceof Task) {
			return afterVariableAnalysis((Task) location, variable, variablesToCompareTo, abortCriteria);
		} else if (location instanceof UFunction) {
			final Optional<InnerFlowNode> anySuccessorContainsVariable = InnerFlowGraph.getAllSuccessors(node).stream()
					.filter(successors -> variablesToCompareTo.apply(successors).contains(variable))
					.findAny();
			final List<UFunctionCall> parentTaskFunctionCall = task.getParentTask().getFunctionCalls().stream()
					.filter(fc -> fc.getFunction().equals(location))
					.collect(Collectors.toList());
			ArrayList<Declaration> args = new ArrayList<>(((UFunction)location).getFunctionArguments());
			final int argumentIndexOfVariable = args.indexOf(variable.getVariableDeclarator().getDeclaration());
			final boolean isPointertype = VariableUtils.isArray(variable) || VariableUtils.isPointer(variable);
			if (!parentTaskFunctionCall.isEmpty() && argumentIndexOfVariable != -1 && isPointertype) {

				final List<VariableDeclarator> parentTaskVariableDeclarators = task.getParentTask().getVariables().stream()
						.map(v -> v.getVariableDeclarator())
						.collect(Collectors.toList());

				List<VariableDeclarator> argumentVariableDeclarators = parentTaskFunctionCall.stream()
						.map(fc -> fc.getCetusFunctionCall().getArgument(argumentIndexOfVariable))
						.flatMap(StreamUtil.castOrEmpty(Traversable.class))
						.flatMap(arg -> CetusUtils.getStreamOfAllChildren(arg, Identifier.class))
						.filter(i -> parentTaskVariableDeclarators.contains(i.getSymbol()))
						.map(i -> i.getSymbol())
						.flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class))
						.collect(Collectors.toList());
				final List<Variable> variables = task.getParentTask().getVariables().stream()
						.filter(v -> argumentVariableDeclarators.contains(v.getVariableDeclarator()))
						.collect(Collectors.toList());
				for (Variable var : variables) {
					return new VariableOccurancePoint(true, task.getParentTask(), var);

				}
			}
			return new VariableOccurancePoint(anySuccessorContainsVariable.isPresent());

		}

		return fail;

	}

	private VariableOccurancePoint isArrayOrStructAndWrittenBefore(Task task, Variable variable) {
		if (VariableUtils.isArray(variable) || VariableUtils.isStruct(variable)) {
			return beforeVariableAnalysis(task, variable, InnerFlowNode::getWrittenVariables);
		}
		else {
			return new VariableOccurancePoint(false);
		}

	}

	/**
	 * Helper class to specify the occurance point of variables in successors/predecessors
	 */
	private class VariableOccurancePoint {

		private boolean occurancePointFound;

		private Task task;

		private Variable variableAtTask;

		public VariableOccurancePoint(boolean occurancePointFound, Task task, Variable variableAtTask) {
			this.occurancePointFound = occurancePointFound;
			this.task = task;
			this.variableAtTask = variableAtTask;
		}

		public VariableOccurancePoint(boolean occurancePointFound, Task task) {
			this.occurancePointFound = occurancePointFound;
			this.task = task;
			this.variableAtTask = null;
		}

		public VariableOccurancePoint(boolean occurancePointFound, Optional<Task> task) {
			this.occurancePointFound = occurancePointFound;
			this.task = null;
			if (task.isPresent()) {
				this.task = task.get();
			}
			this.variableAtTask = null;
		}

		public VariableOccurancePoint(boolean occurancePointFound) {
			this.occurancePointFound = occurancePointFound;
			this.task = null;
			this.variableAtTask = null;
		}

		public VariableOccurancePoint() {
			this.occurancePointFound = false;
			this.task = null;
			this.variableAtTask = null;
		}

		public boolean isOccurancePointFound() {
			return occurancePointFound;
		}

		public Optional<Task> getTask() {
			if (task == null) {
				return Optional.empty();
			}
			return Optional.of(task);
		}

		public Optional<Variable> getVariableAtTask() {
			if (variableAtTask == null) {
				return Optional.empty();
			}
			return Optional.of(variableAtTask);
		}

		public void setOccurancePointFound(boolean occurancePointFound) {
			this.occurancePointFound = occurancePointFound;
		}

		public void setTask(Task task) {
			this.task = task;
		}

		public void setVariableAtTask(Variable variableAtTask) {
			this.variableAtTask = variableAtTask;
		}

		public String toString() {
			return "Found " + isOccurancePointFound() + " | Task " + task + " | Variable " + variableAtTask;
		}
	}

}
