/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.interfaces.CallsFunctions;
import de.tu_darmstadt.rs.microStreams.model.interfaces.ContainsChildTask;
import de.tu_darmstadt.rs.microStreams.model.other.HeaderFile;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunctionCall;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;

import cetus.hir.AnnotationDeclaration;
import cetus.hir.DepthFirstIterator;
import cetus.hir.FunctionCall;
import cetus.hir.Procedure;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;

/**
 * <p>
 * This {@link TransformationPass} implementation generates a call graph for each
 * {@link Task} already found by the {@link TaskCreator} {@link TransformationPass}. This
 * call graph shows up which functions are called by each {@link Task}. For all
 * functions found on this way a new {@link UFunction} instance is added to the
 * {@link Model} and linked with the corresponding task trough a
 * {@link UFunctionCall}. This is also done recursively for all other functions
 * nested within found functions. For Functions in extern C source files the
 * necessary header file is searched and added to the {@link Model} as
 * {@link HeaderFile}. Caller (Task or intern Function) and header file are
 * linked through a {@link Include} instance.
 * </p>
 * <p>
 * <p>
 * In addition to that each function discovered on this way is scanned for
 * directly nested tasks. Each task has a field to store its direct parent. It
 * denotes whether the task is nested directly in its parent task or a function
 * called by its parent task. The direct parent has to implement the
 * {@link ContainsChildTask} interface, which is done by {@link Task} and
 * {@link UFunction}. The direct parent field of all tasks nested directly within
 * a function are set within this {@link FunctionCallgraph}. (The whole parent task -
 * child task relationship is set up in the {@link ChildtaskCallgraph} Worker
 * interface implementation.)
 * </p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> The {@link Model} already contains all {@link Task
 * Tasks}. The {@link FunctionCallgraph#processable(Model)} method checks
 * whether the {@link TaskCreator} is applied to the {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class FunctionCallgraph extends AbstractTransformationPass {

	private static final Logger logger = LoggerFactory.getLogger(FunctionCallgraph.class);

	private Collection<Task> tasks;
	private Model model;
	private Collection<UFunction> visitedFunctions;
	private Collection<UFunction> allFunctions;


	public FunctionCallgraph() {
	}

	/**
	 * Returns the included header file that declares the function corresponding
	 * to the given {@link UFunctionCall}.
	 *
	 * @param functionCall a function call
	 * @return the include file.
	 */
	/*public static Include getInclude(cetus.hir.FunctionCall functionCall) {

		Traversable traversable = functionCall.getProcedure();
		if (traversable == null) {
			traversable = functionCall;
		}

		TranslationUnit translationUnit = CetusUtils.getLocation(traversable);

		DepthFirstIterator<Traversable> iter = new DepthFirstIterator<>(
			translationUnit);

		Traversable t = null;
		Include include = null;

		while (iter.hasNext()) {
			t = iter.next();

			if (isStartIncludePragma(t)) {
				include = CetusUtils.getIncludeFromAnnotation(model, (AnnotationDeclaration) t);
			} else if (isEndIncludePragma(t)) {
				include = null;
			} else if (null != include) {
				if (t instanceof VariableDeclaration) {
					if (((VariableDeclaration) t).getDeclaredIDs().contains(
						functionCall.getName())) {
						return include;
					}
				}
			}
		}
		return null;
	}*/

	/**
	 * Returns a Collection of all FunctionCalls inside the given Traversable t
	 *
	 * @param t a traversable
	 * @param tasks the tasks to operate on
	 * @return Collection of all FunctionCalls
	 */
	public static Collection<cetus.hir.FunctionCall> getNestedFunctionCalls(Traversable t, Collection<Task> tasks) {
		Collection<cetus.hir.FunctionCall> nestedFunctionCalls = new LinkedHashSet<>();
		DepthFirstIterator<Traversable> iter = new DepthFirstIterator<>(t);
		while (iter.hasNext()) {
			Traversable tmp = iter.next();
			if (tmp instanceof FunctionCall) {
				cetus.hir.FunctionCall functionCall = (FunctionCall) tmp;

				/*
				 * Check whether the function call is contained in a Child task of
				 * the traversable. if so it is not added to the nested function
				 * calls.
				 */
				boolean skip = false;
				Traversable tParent = functionCall.getParent();
				while (null != tParent && !t.equals(tParent)) {
					for (Task task : tasks) {
						if (tParent.equals(task.getTraversable())) {
							skip = true;
							break;
						}
					}
					if (skip) {
						break;
					}
					tParent = tParent.getParent();
				}
				if (!skip) {
					nestedFunctionCalls.add(functionCall);
				}
			}
		}
		return nestedFunctionCalls;
	}

	/**
	 * Checks whether the given {@link Traversable} t is a "#pragma endinclude"
	 * {@link AnnotationDeclaration}.
	 *
	 * @param t a traversable
	 * @return true if the given {@link Traversable} t is a "#pragma endinclude"
	 */
	public static boolean isEndIncludePragma(Traversable t) {
		if (t instanceof AnnotationDeclaration
				&& t.toString().startsWith("#pragma endinclude")) {
			return true;
		}
		return false;
	}

	/**
	 * Checks whether the given Traversable t is a "#pragma startinclude"
	 * AnnotationDeclaration.
	 *
	 * @param t traversable
	 * @return true if the given {@link Traversable} t is a "#pragma startinclude"
	 */
	public static boolean isStartIncludePragma(Traversable t) {
		if (t instanceof AnnotationDeclaration
				&& t.toString().startsWith("#pragma startinclude")) {
			return true;
		}
		return false;
	}

	/**
	 * Returns true if the given {@link Traversable} contains at least one Task,
	 * false otherwise.
	 *
	 * @param traversable checked {@link Traversable}
	 * @return Returns true if the given {@link Traversable} contains a Task, false otherwise.
	 */
	private boolean containsTasks(Traversable traversable) {
		return (!getContainedTasks(traversable).isEmpty());
	}

	/**
	 * Scans the given {@link Traversable} t for contained ChildTasks and
	 * returns the result. Note that only direct tasks are returned, children
	 * contained in functions called within this {@link Traversable} are not
	 * returned, children's children are also not returned.
	 *
	 * @param t scanned Traversable
	 * @return Collection of Childtasks
	 */
	private Collection<Task> getContainedTasks(Traversable t) {
		Collection<Task> childs = new LinkedHashSet<Task>();

		for (Task task : tasks) {
			/* Add only direct childs */
			if (CetusUtils.contains(task.getTraversable(), t)
					&& null == getTaskContaining(task.getTraversable())) {
				childs.add(task);
			}
		}

		return childs;
	}

	/**
	 * Returns the task containing the given {@link Traversable} traversable.
	 *
	 * @param traversable - Traversable the containing task is searched for.
	 * @return Returns the {@link Task} containing the given {@link Traversable} traversable.
	 */
	private Task getTaskContaining(Traversable traversable) {
		while (traversable != null) {
			traversable = traversable.getParent();
			for (Task task : tasks) {
				if (task.getTraversable().equals(traversable)) {
					return task;
				}
			}
		}
		return null;
	}

	/**
	 * Returns the µStreams {@link UFunction} instance for the given Cetus IR
	 * {@link Procedure}. If there is no instance yet, it is generated.
	 *
	 * @param procedure Cetus IR {@link Procedure}
	 * @return the {@link UFunction} instance for the given Cetus IR Procedure.
	 */
	private UFunction getFunction(Procedure procedure) {

		if (procedure == null) {
			return null;
		}
		/* Check whether function was already created. And return it. */
		for (UFunction function : allFunctions) {
			if (function.getTraversable().equals(procedure)) {
				return function;
			}
		}

		/* If not: */

		/* Get all direct children of function */
		Collection<Task> directChildTasks = getContainedTasks(procedure);

		/* Create function and add it to collection */
		UFunction function = new UFunction(procedure, directChildTasks, model);

		allFunctions.add(function);

		/* Set the direct parent of all found child task to the function */
		for (Task directChildTask : directChildTasks) {
			directChildTask.setDirectParent(function);
		}

		return function;

	}

	/**
	 * Starts a scan for all Calls inside a Task. It also scans recursive after
	 * function calls inside called functions. It collects all necessary
	 * function bodies and includes for the given {@link CallsFunctions caller
	 * (Task)}.
	 *
	 * @param caller the caller
	 * @param taskLocation the task location to be scanned
	 */
	private void scanForFunctionCalls(CallsFunctions caller, TranslationUnit taskLocation) {
		Collection<cetus.hir.FunctionCall> functionCalls = getNestedFunctionCalls(caller.getTraversable(), tasks);

		for (cetus.hir.FunctionCall functionCall : functionCalls) {
			Procedure procedure = functionCall.getProcedure();
			TranslationUnit location = CetusUtils.getLocation(procedure);
			UFunction function = getFunction(procedure);
			if (null != location && taskLocation.equals(location)) {
				/*
				 * Function is located within same TranslationUnit as the Task.
				 * Add Function to task an restart scan inside this function.
				 */

				UFunctionCall call = new UFunctionCall(function, functionCall);

				caller.addFunctionCall(call);

				/*
				 * In order to avoid endless loops track all functions visited
				 * in the HashSet visitedFunctions.
				 */
				if (!visitedFunctions.contains(function)) {
					visitedFunctions.add(function);
					scanForFunctionCalls(function, taskLocation);
				}
			} else {
				/*
				 * Function is located within another TranslationUnit or is not
				 * found. Function access will be ensured trough header include.
				 */

				if (null != procedure && containsTasks(procedure)) {
					/*
					 * TODO: if it contains other tasks ... it is necessary to
					 * put this function in the main file! That also means that
					 * all function calls from other files have to be
					 * redirected, too.
					 */
					throw new UnsupportedOperationException(
							"Child Tasks in included files are not supported yet.");
				} else {
					Include include = CetusUtils.getInclude(model, functionCall);
					if (include != null) {
						UFunctionCall call = new UFunctionCall(include);
						caller.addFunctionCall(call);
					} else {
						logger.warn("Unable to detect declaration site of call: {}", functionCall);
					}
				}
			}
		}
	}

	@Override
	public List<Class<? extends TransformationPass>> requiredTransformationPasses() {
		return Arrays.asList(

		);
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Task.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				UFunction.class
		);
	}

	@Override
	public boolean process(Model m) {

		this.tasks = m.getModelComponents(Task.class);

		this.model = m;

		allFunctions = new LinkedHashSet<>();
		for (Task task : tasks) {
			Traversable taskCore = task.getTraversable();
			visitedFunctions = new LinkedHashSet<>();
			scanForFunctionCalls(task, CetusUtils.getLocation(taskCore));
		}

		for (UFunction function : allFunctions) {
			m.addModelComponent(function);
		}

		return true;
	}

}
