/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraphFractory;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowNode;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

/**
 * This Worker generates a control flow graph for all tasks and functions. Each
 * control flow graph is represented by a {@link InnerFlowGraph}.
 * <p>
 * <p>
 * A {@link InnerFlowGraph} consists of interconnected {@link InnerFlowNode
 * InnerFlowNodes} representing source code lines within the original C firmware
 * and is stored its task or function.
 * </p>
 * <p>
 * <p>
 * The {@link InnerFlowLinker} {@link TransformationPass} implementation uses this
 * {@link InnerFlowGraph} and adds all {@link VariableAccess VariableAccesses}
 * to the corresponding graph nodes.
 * </p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> The {@link Model} already contains a parsed Cetus IR
 * and all {@link Task Tasks} and {@link UFunction Functions} are already
 * extracted. The {@link InnerFlowCreator#processable(Model)} method checks
 * whether the {@link ChildtaskCallgraph} and the {@link FunctionCallgraph} are
 * applied to the {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class InnerFlowCreator extends AbstractTransformationPass {

	/* slf4j logger */
	protected static final Logger log = LoggerFactory.getLogger(InnerFlowCreator.class);

	protected boolean operationSuccessful;


	public InnerFlowCreator() {
	}





	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Task.class,
				UFunction.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				InnerFlowGraph.class
		);
	}

	@Override
	public boolean process(Model m) {

		this.operationSuccessful = true;

		/*
		 * Collecting all task traversables, for detecting tasks within
		 * programflow.
		 */
		List<Task> tasks = m.getModelComponents(Task.class).stream().filter(t -> !t.isReplica()).collect(Collectors.toList());

		/* Generate a inner flow graph for all Task */
		for (Task task : m.getModelComponents(Task.class)) {
			InnerFlowGraph graph = InnerFlowGraphFractory.getControlFlowForTask(task.getTraversable(), tasks, !task.isBaseTask());
			task.setInnerFlowGraph(graph);
			m.addModelComponent(graph);
		}
		/* Generate a inner flow graph for all Functions */
		for (UFunction function : m.getModelComponents(UFunction.class)) {
			InnerFlowGraph graph = InnerFlowGraphFractory.getControlFlowForFunction(function.getTraversable(),tasks);
			function.setInnerFlowGraph(graph);
			m.addModelComponent(graph);
		}

		return operationSuccessful;
	}

}
