/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.interfaces.HasControlFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowNode;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowTask;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;

/**
 * Binds the {@link VariableAccess VariableAccesses} to the
 * {@link InnerFlowNode}, which contains them.
 * <p>
 * This informations are used later in the {@link DependencyCreator}
 * {@link TransformationPass} implementation to optimize the data flow between the
 * {@link Task}s to generate the so called TaskPipeline (further information to
 * the TaskPipline concept, see.: master's thesis
 * "Design und Implementierung eines parallelisierenden Source-To-Source Compiler für SpartanMC"
 * ).
 * <p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> All {@link Task Tasks} and {@link UFunction Functions}
 * in the {@link Model} contain already their {@link InnerFlowGraph}, also all
 * {@link VariableAccess VariableAccesses} have to be detected and added to the
 * {@link Task Tasks} and {@link UFunction Functions}. The
 * {@link InnerFlowLinker#processable(Model)} method checks whether the
 * {@link InnerFlowCreator} and the {@link VariableAccessCreator} are applied to
 * the {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class InnerFlowLinker extends AbstractTransformationPass {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(InnerFlowLinker.class);

	public InnerFlowLinker() {
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				InnerFlowGraph.class,
				VariableAccess.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				InnerFlowGraph.class
		);
	}

	@Override
	public boolean process(Model m) {
		Collection<VariableAccess> variableAccess = m
				.getModelComponents(VariableAccess.class);

		for (VariableAccess va : variableAccess) {
			if (!va.isForced()) {
				/* get direct parent's control flow graph. */
				HasControlFlowGraph parent = va.getDirectParent();
				InnerFlowGraph graph = parent.getInnerFlowGraph();
				if (!link(va, graph)) {
					return false;
				}


			}
		}

		return true;
	}

	/**
	 * Links a variable access in a InnerFlowGraph
	 * @param va a variable access
	 * @param graph a inner flow graoh to link on
	 * @return true if a variable access is found in the graph and was linked
	 */
	public static boolean link(VariableAccess va, InnerFlowGraph graph) {
		/* get the node the access takes place */
		InnerFlowNode accessNode;
		if (va.getIdentifier() != null) {
			accessNode = graph.getInnerFlowNodeContaining(va.getIdentifier());
		} else if (va.getNameID() != null) {
			accessNode = graph.getInnerFlowNodeContaining(va.getNameID());
		} else {
			accessNode = graph.getInnerFlowNodeContaining(va.getVariable().getVariableDeclarator().getDeclaration());
		}


		if (null != accessNode) {
			/*
			 * ... and add the access. (For InnerFlowTasks this is not
			 * necessary because the information is already accessible
			 * through the actual Task stored within the InnerFlowTask.)
			 */
			if (!(accessNode instanceof InnerFlowTask)) {
				accessNode.addVariableAccess(va);
			}
		} else {
			log.warn("VariableAccess " + va.getIdentifier()
					+ " not found within InnerFlowGraph");
			return false;
		}
		return true;
	}

}
