/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.other.SourceFile;
import de.tu_darmstadt.rs.microStreams.model.software.include.Include;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.hwmodel.CoreConnectionDMAModifier;

import cetus.hir.AnnotationDeclaration;
import cetus.hir.DFIterator;
import cetus.hir.Declaration;
import cetus.hir.FunctionCall;
import cetus.hir.Identifier;
import cetus.hir.NameID;
import cetus.hir.Procedure;
import cetus.hir.Program;
import cetus.hir.Specifier;
import cetus.hir.Symbol;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.VariableDeclaration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;

/**
 * Worker that scans all tasks and determines all source and include files that
 * need to be copied for the task to compile.
 */
public class NecessaryFileSearcher extends AbstractTransformationPass {

	/* slf4j logger */
	private Logger log = LoggerFactory.getLogger(NecessaryFileSearcher.class);
	private Model model;

	public NecessaryFileSearcher() {

	}

	@Override
	public boolean init() {
		return false;
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Task.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				Task.class
		);
	}

	@Override
	public List<Class<? extends TransformationPass>> requiredTransformationPasses() {
		return Arrays.asList(CoreConnectionDMAModifier.class);
	}

	// TODO: Remove dependency to FunctionCallgraph.getNestedFunctionCalls().
	private void scanForFunctions(Traversable t, Collection<Task> tasks, Collection<FunctionCall> functionCalls, Collection<Procedure> procedures) {
		Collection<FunctionCall> newCalls = FunctionCallgraph.getNestedFunctionCalls(t, tasks);

		for (FunctionCall call : newCalls) {
			Procedure procedure = call.getProcedure();

			if (procedure != null && !procedures.contains(procedure)) {
				procedures.add(procedure);
				functionCalls.add(call);
				scanForFunctions(procedure.getBody(), tasks, functionCalls, procedures);
			}
		}

		return;
	}

	private Collection<FunctionCall> scanForFunctions(Traversable t, Collection<Task> tasks) {
		Collection<FunctionCall> functionCalls = new LinkedHashSet<>();
		Collection<Procedure> procedures = new LinkedHashSet<>();

		scanForFunctions(t, tasks, functionCalls, procedures);

		return functionCalls;
	}

	private boolean hasExternDeclaration(Identifier id) {
		Symbol s = id.getSymbol();

		if (s != null) {
			return id.getSymbol().getTypeSpecifiers().contains(Specifier.EXTERN);
		}

		return false;
	}

	private boolean isRootDeclaration(VariableDeclaration dec) {
		Traversable p = dec.getParent();

		while (p != null) {
			if (p instanceof Declaration) {
				return false;
			}

			p = p.getParent();
		}

		return true;
	}

	private boolean isDeclarationOf(VariableDeclaration dec, Identifier id) {
		List<NameID> decNames = new DFIterator<NameID>(dec, NameID.class).getList();
		List<NameID> idNames = new DFIterator<NameID>(id.getSymbol().getDeclaration(), NameID.class).getList();

		return decNames.get(0).equals(idNames.get(0));
	}

	private TranslationUnit findDefinition(Identifier id) {
		Program p = CetusUtils.getProgram(model);
		DFIterator<TranslationUnit> iterTU = new DFIterator<>(p, TranslationUnit.class);

		while (iterTU.hasNext()) {
			TranslationUnit tu = iterTU.next();
			DFIterator<VariableDeclaration> iterDec = new DFIterator<>(tu, VariableDeclaration.class);

			while (iterDec.hasNext()) {
				VariableDeclaration dec = iterDec.next();

				if (isRootDeclaration(dec)
						&& !dec.getSpecifiers().contains(Specifier.EXTERN)
						&& !dec.getSpecifiers().contains(Specifier.STATIC)
						&& isDeclarationOf(dec, id)) {
					return tu;
				}
			}
		}

		return null;
	}

	@Override
	public boolean process(Model model) {
		this.model = model;

		Collection<Task> tasks = model.getModelComponents(Task.class);
		for (Task task : tasks) {
			Traversable taskBody = task.getTraversable();
			TranslationUnit taskLocation = CetusUtils.getLocation(taskBody);

			// Scan for all used functions.
			Collection<FunctionCall> functions = scanForFunctions(taskBody, tasks);
			Map<TranslationUnit, SourceFile> sourceFiles = new HashMap<>();

			for (FunctionCall f : functions) {
				Procedure procedure = f.getProcedure();
				TranslationUnit location = CetusUtils.getLocation(procedure);

				if (!taskLocation.equals(location)) {
					SourceFile sf = CetusUtils.getSourceFile(model, location);
					if (sf != null) {
						sourceFiles.put(location, sf);
					}
				}
			}

			// Scan base task for extern variables.
			Traversable taskBase = task.getTraversable();
			DFIterator<cetus.hir.Identifier> baseTaskIter = new DFIterator<>(taskBase, cetus.hir.Identifier.class);

			while (baseTaskIter.hasNext()) {
				Identifier id = baseTaskIter.next();

				// Check whether this variable is used in another task.
				boolean skip = false;
				Traversable tParent = id.getParent();
				while (null != tParent && !taskBase.equals(tParent)) {
					for (Task t : tasks) {
						if (tParent.equals(t.getTraversable())) {
							skip = true;
							break;
						}
					}
					if (skip) {
						break;
					}
					tParent = tParent.getParent();
				}

				if (!skip && hasExternDeclaration(id)) {
					TranslationUnit defFile = findDefinition(id);

					if (defFile != null) {
						SourceFile sf = CetusUtils.getSourceFile(model, defFile);
						if (sf != null) {
							sourceFiles.put(defFile, sf);
						}
					}
				}
			}

			// Scan called function for extern variables.
			for (FunctionCall f : functions) {
				Procedure procedure = f.getProcedure();
				DFIterator<Identifier> iter = new DFIterator<>(procedure, Identifier.class);

				while (iter.hasNext()) {
					Identifier id = iter.next();

					if (hasExternDeclaration(id)) {
						TranslationUnit defFile = findDefinition(id);

						if (defFile != null) {
							SourceFile sf = CetusUtils.getSourceFile(model, defFile);
							if (sf != null) {
								sourceFiles.put(defFile, sf);
							}
						}
					}
				}
			}

			// Add all necessary source files to task.
			for (SourceFile sf : sourceFiles.values()) {
				task.addSourceFile(sf);
			}

			// Scan for include files.
			for (TranslationUnit tu : sourceFiles.keySet()) {
				for (Traversable t : tu.getChildren()) {
					if (t instanceof AnnotationDeclaration) {
						Include i = CetusUtils.getIncludeFromAnnotation(model, (AnnotationDeclaration) t);
						if (i != null && !i.isSystemLibrary()) {
							task.addHeaderFile(i.getHeaderFile());
						}
					}
				}
			}

		}
		return true;
	}
}
