/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskCoreType;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.EndPragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.Pragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.PragmaAnnotationParser;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.PragmaTag;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.TaskPragma;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.input.CetusParser;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.analysis.CallGraph;
import cetus.analysis.CallGraph.Caller;
import cetus.analysis.CallGraph.Node;
import cetus.hir.CompoundStatement;
import cetus.hir.Declaration;
import cetus.hir.DeclarationStatement;
import cetus.hir.DepthFirstIterator;
import cetus.hir.FlatIterator;
import cetus.hir.FunctionCall;
import cetus.hir.Identifier;
import cetus.hir.PragmaAnnotation;
import cetus.hir.Procedure;
import cetus.hir.Program;
import cetus.hir.Statement;
import cetus.hir.Symbol;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.VariableDeclaration;
import com.google.common.collect.Lists;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * <p>
 * This {@link TransformationPass} implementation extracts all Tasks from the
 * {@link Program} (Cetus IR) stored in the given {@link Model}. Therefore the
 * programmer has to place microstreams pragma annotations in front of the code
 * components that shall be transformed into tasks.
 * </p>
 * <p>
 * <p>
 * Pragmas can be placed
 * <ul>
 * <li>in front of a single statement or procedure call,</li>
 * <li>a compound statement (two enclosing brackets: { and }) or</li>
 * <li>in front of a procedure declaration.</li>
 * </ul>
 * </p>
 * <p>
 * <p>
 * If a Pragma is placed in front of a procedure declaration, the pragma will be
 * transformed to multiple procedure call pragmas, before the tasks are created,
 * so that each procedure call represents an individual task. Note: the
 * main(...) function of the Firmware is also interpreted as task and is referred as
 * Basetask.
 * </p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> The {@link Model} already contains a parsed
 * {@link Program} (Cetus IR). The {@link TaskCreator#processable(Model)} method
 * checks whether the {@link CetusParser} is applied to the {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class TaskCreator extends AbstractTransformationPass {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(TaskCreator.class);

	/* Stores the project instance the worker is currently working on. */
	private Model model;

	/*
	 * Instance of a Cetus callgraph, which is utilized to transform PROCEDURE
	 * pragmas into PROCEDURE_CALL pragmas.
	 */
	private CallGraph callGraph;


	/* Instance of the base task */
	private Task baseTask;

	private Program program;

	private List<Procedure> taskProcedures;


	public TaskCreator() {
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				CetusProgram.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				Task.class
		);
	}

	@Override
	public boolean process(Model project) {
		this.model = project;
		taskProcedures = new ArrayList<>();
		program = CetusUtils.getProgram(project);
		callGraph = new CallGraph(program);

		/* Generate a Base Task */
		Traversable tMain = CetusUtils.getMainProcedure(program);
		if (null != tMain) {
			/* Use the main procedures compound statement to generate Base Task. */
			Traversable tBaseTask = tMain.getChildren().get(0);
			baseTask = model.getTaskFactory().createTask(tBaseTask, model);
			project.addModelComponent(baseTask);
		}

		/* Search all pragmas. */
		searchAnnotations(program);

		/* Inform the user that there are no Task pragmas in the source code */
		if (project.getModelComponents(Task.class).isEmpty()) {
			log.warn("No tasks pragmas found!");
		}

		/* Task generation was successful, if at least one task was generated. */
		return (!project.getModelComponents(Task.class).isEmpty() && null != baseTask);

	}

	/**
	 * Recursive search of of all µStreams Task pragmas in the program.
	 *
	 * @param parentTraversable a traversable of the parent
	 */
	private void searchAnnotations(Traversable parentTraversable) {

		DepthFirstIterator<Traversable> iter = new DepthFirstIterator<>(parentTraversable);

		Set<TaskPragma> totalTaskPragmas = Lists.newArrayList(iter).stream()
				.map(p -> getPragma(p))
				.filter(p -> p instanceof TaskPragma)
				.map(p -> (TaskPragma)p)
				.collect(Collectors.toSet());
		Set<Traversable> handledTaskPragmas = new HashSet<>();
		for (int pragmaSearchRuns = 0;pragmaSearchRuns < totalTaskPragmas.size(); pragmaSearchRuns++) {
			iter = new DepthFirstIterator<>(parentTraversable);
			while (iter.hasNext()) {

				Traversable tChild = iter.next();
				Pragma pragma = getPragma(tChild);

				if (pragma instanceof TaskPragma && !handledTaskPragmas.contains(((TaskPragma) pragma).getTraversable())) {
					if (iter.hasNext()) {
						Traversable taskTraversable = iter.next();

						TaskCoreType coreType = TaskCoreType.getTypeOf(taskTraversable);

						TaskPragma taskPragma = (TaskPragma) pragma;
						taskPragma.setCoreType(coreType);
						ArrayList<Task> tasks = new ArrayList<>();
						// If it is a PROCEDURE-pragma, it has to be transformed.
						if (coreType == TaskCoreType.PROCEDURE) {
							Collection<TaskPragma> transformedPragmas = transform(taskPragma, taskTraversable);
							// Create for all results of the transformed pragma a new task.
							for (TaskPragma transformedPragma : transformedPragmas) {
								tasks.add(model.getTaskFactory().createTask(transformedPragma.getTaskTraversable(), transformedPragma, model));
								taskProcedures.add((Procedure) taskTraversable);
							}
						} else {
							boolean isInMainFile = CetusUtils.findParentOfType(CetusUtils.getMainProcedure(CetusUtils.getProgram(model)), TranslationUnit.class).equals(CetusUtils.findParentOfType(taskTraversable, TranslationUnit.class));
							if (!isInMainFile) {
								model.printIgnoreableError(log, "The task is declared outside the main C file inside a function. This is currently not supported. Place functions holding tasks inside main.c or write the pragma above the function declaration. The tasks traversable is: " + taskTraversable);
							}

							StreamUtil.Pair<Pragma, CompoundStatement> result = getTaskContent(taskPragma);
							CompoundStatement cs = result.getB();
							EndPragma endPragma = null;
							if (result.getA() instanceof EndPragma) {
								endPragma = (EndPragma) result.getA();
							}
							tasks.add(model.getTaskFactory().createTask(cs, taskPragma, endPragma, model));
						}
						if (tasks.isEmpty() || !tasks.stream().noneMatch(t -> t==null)) {
							throw new RuntimeException("Got null task, this should never be the case!");
						}
						handledTaskPragmas.add(taskPragma.getTraversable());
						for (Task t : tasks) {
							model.addModelComponent(t);
						}
						break;
					} else {
						log.warn("TaskPragma found but not followed by task core: " + CetusUtils.getStringFromPrintable(tChild) + " Skipping for now ...");
					}
				}
			}
		}
		if (handledTaskPragmas.size() != totalTaskPragmas.size()) {
			model.printIgnoreableError(log,"Discovered " + totalTaskPragmas.size() + " task pragmas, but only handled " + handledTaskPragmas.size() + " of them!");
		}
	}

	private StreamUtil.Pair<Pragma, CompoundStatement> getTaskContent(TaskPragma taskPragma) {
		CompoundStatement cs = new CompoundStatement();
		FlatIterator<Traversable> seeker = new FlatIterator<>(taskPragma.getTraversable().getParent());
		//FWD to current position
		while (seeker.hasNext() && seeker.next() != taskPragma.getTraversable()) {
		}
		boolean stop = false;
		List<Statement> addedStatements = new ArrayList<>();
		Pragma nextPragma = null;
		while (seeker.hasNext() && !stop) {
			Traversable t = seeker.next();
			nextPragma = getPragma(t);

			boolean isFunctionCallWithTaskPragma = CetusUtils.getStreamOfAllChildren(t, FunctionCall.class)
					.map(fc -> fc.getProcedure())
					.filter(p -> taskProcedures.contains(p))
					.findAny().isPresent();
			if ((nextPragma != null && (nextPragma instanceof EndPragma || nextPragma instanceof TaskPragma))||isFunctionCallWithTaskPragma) {
				stop = true;
			} else if (t instanceof DeclarationStatement) {
				Declaration tmp = ((DeclarationStatement) t).getDeclaration().clone();
				tmp.setParent(null);
				cs.addDeclaration(tmp);
				updateIdentifierSymbols(((DeclarationStatement) t).getDeclaration(), tmp);
				addedStatements.add((DeclarationStatement) t);

			} else if (t instanceof Statement) {
				Statement s = ((Statement) t).clone();
				s.setParent(null);
				cs.addStatement(s);
				s.setParent(cs);
				addedStatements.add((Statement) t);
			}
		}

		Traversable parent = taskPragma.getTraversable().getParent();
		if (!addedStatements.isEmpty()) {
			Statement leftover = addedStatements.remove(0);
			addedStatements.forEach(s -> {
				parent.removeChild(s);
			});
			int pos = 0;
			for (Traversable c : parent.getChildren()) {
				if (c == leftover) {
					parent.setChild(pos, cs);
				}
				pos++;
			}
		}
		cs.setParent(parent);
		return new StreamUtil.Pair<>(nextPragma, cs);
	}

	private void updateIdentifierSymbols(Declaration oldDeclaration, Declaration newDeclaration) {
		if (newDeclaration instanceof VariableDeclaration && oldDeclaration instanceof VariableDeclaration) {
			final List<Symbol> newSymbol = ((VariableDeclaration) newDeclaration).getDeclaredSymbols().stream().collect(StreamUtil.atLeastOne());
			final List<Symbol> oldSymbol = ((VariableDeclaration)  oldDeclaration).getDeclaredSymbols().stream().collect(StreamUtil.atLeastOne());
			if (newSymbol.size() != oldSymbol.size()) {
				throw new RuntimeException("Can not replace old symbols since sizes don't match!");
			}

			final List<StreamUtil.Pair<Symbol, Symbol>> oldNewSymbols = StreamUtil.streamZipped(oldSymbol.iterator(), newSymbol.iterator()).collect(Collectors.toList());
			for (StreamUtil.Pair<Symbol, Symbol> pair : oldNewSymbols) {
				CetusUtils.getStreamOfAllChildren(CetusUtils.getProgram(model), Identifier.class)
						.filter(i -> i.getSymbol() != null)
						.filter(i -> i.getSymbol().equals(pair.getA()))
						.forEach(i -> i.setSymbol(pair.getB()));
			}
		}
	}

	/**
	 * Transforms the given PragmaTag of a PROCEDURE-Task that is transformed to
	 * a PROCEDURE_CALL-Task.
	 *
	 * @param pragmaTags - A collection of PragmaTags that will be transformed.
	 * @return A collection of PragmaTags as result of the transformation.
	 */
	private Collection<PragmaTag<String>> transform(Collection<PragmaTag<String>> pragmaTags) {
		/*
		 * TODO: If there are any PragmaTags that should also be transformed, do
		 * it here!
		 */
		return pragmaTags;
	}

	/**
	 * Transforms TaskPragmas with TaskCoreType == PROCEDURE into TaskCoreType
	 * == PROCEDURE_CALL ones by creating new TaskPragmas in front of each
	 * function call of the corresponding Task function.
	 *
	 * @param pragma - The pragma of the task that will be transformed.
	 * @return A collection of pragmas as result of the transformation.
	 */
	private Collection<TaskPragma> transform(TaskPragma pragma,
											 Traversable taskTraversable) {
		Collection<TaskPragma> transformedPragmas = new ArrayList<TaskPragma>();
		Traversable t = taskTraversable;

		if (t instanceof Procedure) {
			Procedure p = (Procedure) t;

			@SuppressWarnings("unchecked")
			Map<Procedure, Node> cgMap = callGraph.getCallGraph();

			if (cgMap.containsKey(p)) {
				List<Caller> callers = cgMap.get(p).getCallers();

				for (Caller caller : callers) {
					Statement tCall = caller.getCallSite();

					TaskPragma callPragma = new TaskPragma(
							pragma.getAnnotation(),
							transform(pragma.getPragmaTags()),
							pragma.getTraversable(), tCall);

					transformedPragmas.add(callPragma);
				}
			}

		} else {
			log.warn("Could not be transformed no \"Procedure\"-TaskPragma");
		}

		return transformedPragmas;
	}



	private Pragma getPragma(Traversable traversable) {
		PragmaAnnotation pragmaAnnotation = CetusUtils.getAnnotation(traversable);
		// Check whether the PragmaAnnotation is a valid µStreams Annotation.
		return new PragmaAnnotationParser(model).parse(pragmaAnnotation, traversable);
	}

}
