/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowEnd;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowGraph;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowNode;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowStart;
import de.tu_darmstadt.rs.microStreams.model.software.controlflow.InnerFlowTask;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunctionCall;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency.DependencyRelation;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependencyGroup;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.tools.VariableUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.AnnotationStatement;
import cetus.hir.Declaration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * Tries to transform the dependencies into a pipeline especially it tries to remove backwards dependencies
 */
public class TaskPipeline extends AbstractTransformationPass {

	public static final String COULD_NOT_REMOVE_DEPENDENCY = "Could not remove dependency.";
	public static final String COULD_NOT_ADD_OPTIMIZED_DEPENDENCY = "Could not add optimized dependency.";
	public static final String DETECTED_BASE_TASK_CONSUMER_DEPENDENCY = "Detected dependency with base task as consumer. The backward dependency avoids proper pipeline generation. Try placing pragmas differnt.";
	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(TaskPipeline.class);

	private Model model;


	public TaskPipeline() {
	}

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Task.class,
				InnerFlowGraph.class,
				TaskDependencyGroup.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				TaskDependency.class
		);
	}

	@Override
	public boolean process(Model m) {

		this.model = m;

		for (Task task : m.getModelComponents(Task.class)) {

			Collection<Shortcut> shortcuts = new ArrayList<>();

			Collection<Task> childs = task.getAllChildTasks();

			List<InnerFlowGraph> graphs = new ArrayList<>();
			graphs.add(task.getInnerFlowGraph());
			for (UFunctionCall fc : task.getAllFunctionCalls()) {
				if (!fc.isExtern()) {
					graphs.add(fc.getFunction().getInnerFlowGraph());
				}
			}

			for (Task childTask1 : childs) {
				if (childTask1.isReplica()) {
					continue;
				}
				Task modifiedChildTask = childTask1;

				if (childTask1.getParentTask() != null && childTask1.getDirectParent() instanceof UFunction) {
					ArrayList<Declaration> args = new ArrayList<>(((UFunction)childTask1.getDirectParent()).getFunctionArguments());
					boolean modifiesArgumentAndArgumentIsPointer = childTask1.getDependencies().stream()
							.filter(d -> d.getConsumer() == childTask1)
							.map(d -> d.getVariableOfConsumer())
							.filter(v -> args.indexOf(v.getVariableDeclarator().getDeclaration()) > -1)
							.filter(v -> VariableUtils.isArray(v) || VariableUtils.isPointer(v))
							.count() > 0;
					if (modifiesArgumentAndArgumentIsPointer) {
						modifiedChildTask = childTask1.getParentTask();
						graphs.add(modifiedChildTask.getInnerFlowGraph());
						for (UFunctionCall fc : modifiedChildTask.getAllFunctionCalls()) {
							if (!fc.isExtern()) {
								graphs.add(fc.getFunction().getInnerFlowGraph());
							}
						}
					}
				}
				Task finalC = modifiedChildTask;
				Optional<InnerFlowNode> n1 = graphs.stream()
						.flatMap(g -> StreamUtil.streamOptional(g.getInnerFlowNodeWith(finalC)))
						.distinct()
						.collect(StreamUtil.atMostOne());
				for (Task childTask2 : childs) {
					if (childTask2.isReplica()) {
						continue;
					}
					Optional<InnerFlowNode> n2 = graphs.stream()
							.flatMap(g -> StreamUtil.streamOptional(g.getInnerFlowNodeWith(childTask2)))
							.distinct()
							.collect(StreamUtil.atMostOne());;
					if (n1.isPresent() && n2.isPresent()) {
						if (followsDirect(n1.get(), n2.get())) {
							shortcuts.add(new Shortcut(childTask1, childTask2));
						} else {
							final Optional<Collection<InnerFlowNode>> followsIndirect = followsIndirect(n1.get(), n2.get());
							if (followsIndirect.isPresent()) {
								Shortcut shortcut = new Shortcut(childTask1, childTask2, followsIndirect.get());
								if (shortcut.onlyAnnotations()) {
									shortcuts.add(shortcut);
								}
							}
						}
					}

				}
			}



			boolean finished = false;
			int iterations = 0;
			final int maxIterations = 1 + shortcuts.size()*10;
			while (!finished) {
				iterations++;
				if (iterations > maxIterations) {
					model.printIgnoreableError(log, "Aborting dependency optimization since no solution can be found.");
				}
				finished = true;
				for (Shortcut s : shortcuts) {
					if (s.isDirect() || s.onlyAnnotations()) {
						shortcut(s, task);
					}
					// TODO: can be further optimized, if there are nodes in
					// between and do
					// not use relevant variables.

					while (loopTroughParentChild(s, task)) {
						finished = false;
					}

					while (loopTroughChildParent(s, task)) {
						finished = false;
					}

					while (loopTroughChildChild(s)) {
						finished = false;
					}
				}

			}
		}

		//remove WrittenBeforeRead dependencies
		//it seems not used anymore / not neccessary
		removeWrittenBeforeReadDependencies();

		//remove dependencies with base task as consumer
		detectParentConsumerDependencies();

		return true;
	}


	private void detectParentConsumerDependencies() {
		for (TaskDependency dependency : this.model.getModelComponents(TaskDependency.class)) {

			if (dependency.getConsumer().isBaseTask()) {

				log.warn(DETECTED_BASE_TASK_CONSUMER_DEPENDENCY + " Provider: " + dependency.getProvider().toString() +
						". Variable: " + dependency.getVariableOfProvider().toString());
			}
		}
	}

	private void removeWrittenBeforeReadDependencies() {
		for (TaskDependency dependency : this.model.getModelComponents(TaskDependency.class)) {
			Variable variable = dependency.getVariableOfProvider();
			Task provider = dependency.getProvider();
			Task consumer = dependency.getConsumer();

			InnerFlowGraph graph = consumer.getInnerFlowGraph();

			//walk through the consumers graph and check if the variable is written before read
			for (InnerFlowNode node : graph.getAllNodes(provider)) {

				if (node.getWrittenVariables().contains(variable) && !node.getReadVariables().contains(variable)) {
					provider.removeDependency(dependency);
					consumer.removeDependency(dependency);
					model.removeModelComponent(dependency);
				}
				if (node.getReadVariables().contains(variable)) {
					break;
				}
			}
		}
	}

	private boolean loopTroughChildParent(Shortcut shortcut, Task currentTask) {

		boolean modified = false;

		for (TaskDependency currentDependency : new ArrayList<>(shortcut.getTask1().getDependencies())) {
			if (currentDependency.getProvider() == shortcut.getTask1() && currentDependency.getRelation() == DependencyRelation.CHILD_PARENT) {

				TaskDependency optimizedDependency1 = new TaskDependency(shortcut.getTask2(), shortcut.getTask1(), currentDependency.getVariableOfProvider(), Optional.of(currentDependency.getVariableOfConsumer()), DependencyRelation.CHILD_CHILD, currentTask);
				TaskDependency optimizedDependency2 = new TaskDependency(currentDependency.getConsumer(), shortcut.getTask2(), currentDependency.getVariableOfProvider(), Optional.of(currentDependency.getVariableOfConsumer()), DependencyRelation.CHILD_PARENT, currentTask);

				optimizedDependency2.setForced(currentDependency.isForced());
				optimizedDependency1.setForced(currentDependency.isForced());

				replaceDependencies(shortcut.getTask2(), currentDependency, optimizedDependency1, optimizedDependency2);

				modified = true;

			}
		}
		return modified;

	}

	private void replaceDependencies(Task intermediateTask, TaskDependency dependencyToRemove, TaskDependency dependencyToAdd1, TaskDependency dependencyToAdd2) {
		removeDependency(dependencyToRemove.getProvider(), dependencyToRemove);
		removeDependency(dependencyToRemove.getConsumer(), dependencyToRemove);

		addDependency(dependencyToRemove.getProvider(), dependencyToAdd1);
		addDependency(intermediateTask, dependencyToAdd2);
		addDependency(intermediateTask, dependencyToAdd1);
		addDependency(dependencyToRemove.getConsumer(), dependencyToAdd2);

		model.removeModelComponent(dependencyToRemove);
		model.addModelComponent(dependencyToAdd1);
		model.addModelComponent(dependencyToAdd2);
	}

	private boolean loopTroughParentChild(Shortcut shortcut, Task currentTask) {


		boolean modified = false;

		for (TaskDependency currentDependency : new ArrayList<>(shortcut.getTask2().getDependencies())) {
			if (currentDependency.getConsumer() == shortcut.getTask2()	&& currentDependency.getRelation() == DependencyRelation.PARENT_CHILD) {

				TaskDependency optimizedDependency2 = new TaskDependency(shortcut.getTask2(), shortcut.getTask1(), currentDependency.getVariableOfProvider(), Optional.of(currentDependency.getVariableOfConsumer()), DependencyRelation.CHILD_CHILD, currentTask);
				TaskDependency optimizedDependency1 = new TaskDependency(shortcut.getTask1(), currentDependency.getProvider(), currentDependency.getVariableOfProvider(), Optional.of(currentDependency.getVariableOfConsumer()), DependencyRelation.PARENT_CHILD, currentTask);
				optimizedDependency2.setForced(currentDependency.isForced());
				optimizedDependency1.setForced(currentDependency.isForced());

				replaceDependencies(shortcut.getTask1(), currentDependency, optimizedDependency1, optimizedDependency2);

				modified = true;

			}
		}
		return modified;
	}

	private boolean loopTroughChildChild(Shortcut shortcut) {

		boolean modified = false;

		for (TaskDependency currentDependency : new ArrayList<>(shortcut.getTask2().getDependencies())) {

			if (currentDependency.getConsumer() == shortcut.getTask2()
					&& currentDependency.getProvider() != shortcut.getTask1()
					&& currentDependency.getRelation() == DependencyRelation.CHILD_CHILD) {

				TaskDependency optimizedDependency2 = new TaskDependency(shortcut.getTask2(), shortcut.getTask1(), currentDependency.getVariableOfProvider(), Optional.of(currentDependency.getVariableOfConsumer()) , DependencyRelation.CHILD_CHILD, currentDependency.getProvider());

				TaskDependency optimizedDependency1 = new TaskDependency(shortcut.getTask1(), currentDependency.getProvider(), currentDependency.getVariableOfProvider(), Optional.of(currentDependency.getVariableOfConsumer()), DependencyRelation.CHILD_CHILD, currentDependency.getProvider());

				optimizedDependency2.setForced(currentDependency.isForced());
				optimizedDependency1.setForced(currentDependency.isForced());

				replaceDependencies(shortcut.getTask1(), currentDependency, optimizedDependency1, optimizedDependency2);

				modified = true;

			}
		}
		return modified;
	}

	private void removeDependency(Task task, TaskDependency dependency) {
		if (!task.removeDependency(dependency)) {
			model.printIgnoreableError(log,COULD_NOT_REMOVE_DEPENDENCY);
		}

		if (task.hasReplica()) {
			for (Task r : task.getReplica()) {
				if (!r.removeDependency(dependency)) {
					model.printIgnoreableError(log,COULD_NOT_REMOVE_DEPENDENCY);
				}
			}
		}
	}

	private void addDependency(Task task, TaskDependency dependency) {
		if (!task.addDependency(dependency)) {
			logNoOptDependency(dependency.toString(), task.toString());
		}

		if (task.hasReplica()) {
			for (Task r : task.getReplica()) {
				if (!r.addDependency(dependency)) {
					logNoOptDependency(r.toString(), task.toString());
				}
			}
		}
	}

	private void logNoOptDependency(String from, String to) {
		log.warn("Could not add optimized dependency " + from + "  to " + to);
	}

	private boolean shortcut(Shortcut shortcut, Task currentTask) {

		for (TaskDependency currentDependencyTask2 : new ArrayList<>(shortcut.getTask2().getDependencies())) {

			if (currentDependencyTask2.getConsumer() == shortcut.getTask2() ) {

				for (TaskDependency currentDependencyTask1 : new ArrayList<>(shortcut.getTask1().getDependencies())) {

					if (currentDependencyTask1.getProvider() == shortcut.getTask1()
							&& currentDependencyTask1.getRelation() == DependencyRelation.CHILD_PARENT
							&& currentDependencyTask1.getVariableOfConsumer().equals(currentDependencyTask2.getVariableOfProvider())) {

						TaskDependency optimizedDependency = new TaskDependency(shortcut.getTask2(), shortcut.getTask1(), currentDependencyTask1.getVariableOfProvider(), Optional.of(currentDependencyTask1.getVariableOfConsumer()), DependencyRelation.CHILD_CHILD, currentTask);

						optimizedDependency.setForced(currentDependencyTask2.isForced() || currentDependencyTask1.isForced());

						removeDependency(shortcut.getTask1(), currentDependencyTask1);
						removeDependency(shortcut.getTask2(),currentDependencyTask2);

						removeDependency(currentDependencyTask1.getConsumer(),currentDependencyTask1);
						removeDependency(currentDependencyTask2.getProvider(),currentDependencyTask2);

						addDependency(shortcut.getTask1(), optimizedDependency);
						addDependency(shortcut.getTask2(), optimizedDependency);

						model.removeModelComponent(currentDependencyTask2);
						model.removeModelComponent(currentDependencyTask1);
						model.addModelComponent(optimizedDependency);
					}
				}
			}
		}
		return true;
	}


	private Optional<Collection<InnerFlowNode>> followsIndirect(InnerFlowNode n1, InnerFlowNode n2) {

		Collection<InnerFlowNode> foundNodes = new LinkedHashSet<>(Collections.singleton(n1));
		ArrayList<InnerFlowNode> leftNodes = new ArrayList<>(Collections.singleton(n1));

		while (!leftNodes.isEmpty()) {
			InnerFlowNode n = leftNodes.remove(0);

			Collection<InnerFlowNode> newNodes = new HashSet<InnerFlowNode>(
					n.getSucccessors());

			for (InnerFlowNode newNode : newNodes) {
				if (newNode != n2) {
					if (newNode instanceof InnerFlowStart
							|| newNode instanceof InnerFlowEnd
							|| newNode instanceof InnerFlowTask) {
						return Optional.empty();
					}
				}
			}

			newNodes.remove(n2);
			newNodes.removeAll(foundNodes);
			leftNodes.addAll(newNodes);
			foundNodes.addAll(newNodes);
		}

		foundNodes.remove(n1);
		foundNodes.remove(n2);

		return Optional.of(foundNodes);
	}

	private boolean followsDirect(InnerFlowNode n1, InnerFlowNode n2) {
		return n1.getSucccessors().contains(n2)
				&& n1.getSucccessors().size() == 1;
	}

	/**
	 * A shortcut at a task dependency. Eliminates an intermediate task on the dependency list
	 */
	private class Shortcut {

		private Task t1;
		private Task t2;
		private Collection<InnerFlowNode> vT;

		public Shortcut(Task t1, Task t2) {
			this.t1 = t1;
			this.t2 = t2;
		}

		public Shortcut(Task t1, Task t2, Collection<InnerFlowNode> vT) {
			this.t1 = t1;
			this.t2 = t2;
			this.vT = vT;
		}

		@Override
		public String toString() {
			return "Shortcut [t1=" + t1 + ", t2=" + t2 + ", vT=" + vT + "]";
		}

		public boolean onlyAnnotations() {
			if (null == vT) {
				return true;
			}

			for (InnerFlowNode n : vT) {
				if (!(n.getTraversable() instanceof AnnotationStatement)) {
					return false;
				}
			}
			return true;
		}

		public Task getTask1() {
			return t1;
		}

		public Task getTask2() {
			return t2;
		}

		public boolean isDirect() {
			return vT == null;
		}
	}

}
