/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.TaskDependency;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;

import cetus.hir.Specifier;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;

import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * Adds alignment annotations to variables if they are transferred with core connectors and 1 byte types (core_connector functions are not able to transfer unaligned chars efficiently)
 */
public class TransferredCharAddressAligner extends AbstractTransformationPass {

	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(TaskDependency.class, CetusProgram.class);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(CetusProgram.class);
	}

	@Override
	public boolean process(Model model) {
		if (model.getConfig().useDMA()) {
			return true;
		}
		final Map<Variable, Set<Task>> taskDependencyVariables = new HashMap<>();

		for (TaskDependency taskDependency : model.getModelComponents(TaskDependency.class)){
			if (taskDependencyVariables.containsKey(taskDependency.getVariableOfProvider())) {
				final Collection<Task> replica = taskDependency.getProvider().getReplica();
				final Set<Task> tasks = taskDependencyVariables.get(taskDependency.getVariableOfProvider());
				tasks.addAll(replica);
			} else {
				final Collection<Task> replica = taskDependency.getProvider().getReplica();
				taskDependencyVariables.put(taskDependency.getVariableOfProvider(), new HashSet<>(replica));
			}
		}

		for (Map.Entry<Variable, Set<Task>> entry : taskDependencyVariables.entrySet()) {
			final boolean tooSmallSpecifier = ((VariableDeclaration) entry.getKey().getVariableDeclarator().getDeclaration()).getSpecifiers().stream().anyMatch(s -> s == Specifier.BYTE || s == Specifier.CHAR);
			if (tooSmallSpecifier) {
				String originalVariableDeclarator = entry.getKey().getVariableDeclarator().toString();
				entry.getKey().getVariableDeclarator().addTrailingSpecifier(Specifier.ATTRIBUTE);
				entry.getKey().getVariableDeclarator().addTrailingSpecifier(Specifier.ALIGNED);

				for (Task task : entry.getValue()) {
					if (task.isReplica()) {
						CetusUtils.getStreamOfAllChildren(task.getTraversable(), VariableDeclarator.class)
								.filter(vd -> vd.toString().equalsIgnoreCase(originalVariableDeclarator))
								.forEach(vd -> {
									vd.addTrailingSpecifier(Specifier.ATTRIBUTE);
									vd.addTrailingSpecifier(Specifier.ALIGNED);});

					}
				}
			}
		}

		return true;
	}
}
