/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.transformationPasses.swmodel;

import de.tu_darmstadt.rs.microStreams.config.Constants;
import de.tu_darmstadt.rs.microStreams.config.IConfiguration;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.microStreams.model.ModelComponent;
import de.tu_darmstadt.rs.microStreams.model.hardware.core.Core;
import de.tu_darmstadt.rs.microStreams.model.interfaces.AccessesVariables;
import de.tu_darmstadt.rs.microStreams.model.other.CetusProgram;
import de.tu_darmstadt.rs.microStreams.model.software.function.UFunction;
import de.tu_darmstadt.rs.microStreams.model.software.task.Task;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.Pragma;
import de.tu_darmstadt.rs.microStreams.model.software.task.pragma.PragmaTag;
import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.model.software.variable.VariableAccess;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.VariableAccessAnalyzer;
import de.tu_darmstadt.rs.microStreams.tools.VariableUtils;
import de.tu_darmstadt.rs.microStreams.transformationPasses.AbstractTransformationPass;
import de.tu_darmstadt.rs.microStreams.transformationPasses.TransformationPass;

import cetus.hir.DFIterator;
import cetus.hir.Expression;
import cetus.hir.FunctionCall;
import cetus.hir.Identifier;
import cetus.hir.Initializer;
import cetus.hir.Procedure;
import cetus.hir.Program;
import cetus.hir.Symbol;
import cetus.hir.TranslationUnit;
import cetus.hir.Traversable;
import cetus.hir.VariableDeclarator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Collectors;

/**
 * <p>
 * This {@link TransformationPass} extracts the {@link Variable Variables} read and written
 * by each Task. For each Identifier within a {@link Task} a
 * {@link VariableAccess} is generated. After that they are checked against
 * the "in" and "out" pragma tags. Note: that the pragma tags have higher
 * priority and overwriting the detected {@link VariableAccess}.
 * </p>
 * <p>
 * <p>
 * After executing this {@link TransformationPass} the Model contains all {@link Variable}s
 * that are relevant for the parallelized firmware and the
 * {@link VariableAccess VariableAccesses} with the exact access position within
 * the firmware.
 * </p>
 * <p>
 * <p>
 * <b>Prerequirements:</b> The {@link Model} already contains a parsed Cetus IR
 * and all {@link Task Tasks} and {@link UFunction Functions}. The
 * {@link VariableAccessCreator#processable(Model)} method checks whether the
 * {@link ChildtaskCallgraph} and the {@link FunctionCallgraph} are applied to
 * the {@link Model}.
 * </p>
 *
 * @author Jan Weber
 */
public class VariableAccessCreator extends AbstractTransformationPass {

	/* slf4j logger */
	private static final Logger log = LoggerFactory.getLogger(VariableAccessCreator.class);

	private Model model;
	private VariableAccessAnalyzer variableAccessAnalyzer;


	public VariableAccessCreator() {
	}

	public static <T> Predicate<T> distinctByKey(Function<? super T, Object> keyExtractor) {
		Map<Object, Boolean> seen = new ConcurrentHashMap<>();
		return t -> seen.putIfAbsent(keyExtractor.apply(t), Boolean.TRUE) == null;
	}


	@Override
	public List<Class<? extends ModelComponent>> requiredModelComponents() {
		return Arrays.asList(
				Task.class,
				CetusProgram.class,
				Core.class,
				IConfiguration.class,
				UFunction.class
		);
	}

	@Override
	public List<Class<? extends ModelComponent>> providedModelComponents() {
		return Arrays.asList(
				VariableAccess.class,
				Variable.class,
				Core.class
		);
	}

	@Override
	public boolean process(Model model) {

		Collection<VariableAccess> variableAccesses = new LinkedHashSet<>();

		Collection<Identifier> globalVariables = new LinkedHashSet<>();

		this.model = model;
		this.variableAccessAnalyzer = new VariableAccessAnalyzer(model);

		/*
		 * Scan for all identifiers in the current task, create all variables
		 * and create for each identifier a VariableAccess.
		 */
		for (Task task : model.getModelComponents(Task.class)) {
			variableAccesses.addAll(scanForIdentifierAndAddToModel(globalVariables,task));
		}


		Collection<CetusProgram> p = model
				.getModelComponents(CetusProgram.class);

		if (p.isEmpty() || p.size() != 1) {
			model.printIgnoreableError(log,"found more than one program");
			return false;
		}




		/*
		 * Also scan for identifiers in all relevant functions, create the
		 * variables and create for each identifier a VariableAccess.
		 */
		for (UFunction function : model.getModelComponents(UFunction.class)) {
			variableAccesses.addAll(scanForIdentifierAndAddToModel(globalVariables, function));
		}

		variableAccessAnalyzer.setAccessBehaviour(variableAccesses);


		model.getModelComponents(Core.class).stream()
				.filter(core -> core.getTasks().stream().anyMatch(task -> task.isBaseTask()))
				.findAny()
				.ifPresent(core -> {
					Program program = p.iterator().next().getProgram();
					scanForInitialisation(program, core, variableAccesses, model.getModelComponents(Variable.class));
		});

		/*
		 * Determine if a warning about global variable usage in several tasks
		 * should be given or not
		 */
		analyzeGlobalVariableUsage(variableAccesses, model.getModelComponents(Task.class));

		if (model.getConfig().useGlobalMemory()) {
			for (Task task : model.getModelComponents(Task.class)) {
				final Collection<Variable> allVariables = model.getModelComponents(Variable.class);
				List<VariableAccess> vas = addGlobalVariables(task, globalVariables, allVariables);
				model.addModelComponent(allVariables.toArray(new Variable[0]));
				variableAccesses.addAll(vas);
				model.addModelComponent(vas.toArray(new VariableAccess[0]));
			}
		}
		/*
		 * Add "forced" VariableAccessBehavior to all Tasks
		 */
		for (Task task : model.getModelComponents(Task.class)) {
			if (!task.isBaseTask()) {
				setAccessBehaviorThroughPragma(task);
			}
		}
		/*
		 * Cross-check "forced" and normal VariableAccessBehavior and warn user
		 * if the behavior differs.
		 */
		crosscheckResults(model.getModelComponents(Task.class));


		return true;
	}



	private Collection<VariableAccess> scanForIdentifierAndAddToModel(Collection<Identifier> globalVariables, AccessesVariables accessesVariables) {
		Collection<VariableAccess> variableAccesses = new LinkedHashSet<>();
		final Collection<Variable> allVariables = this.model.getModelComponents(Variable.class);
		allVariables.stream().filter(v -> v.getSize() == 0).forEach(v -> v.setSize(VariableUtils.getSizeInSpartanMCBytes(v)));
		List<VariableAccess> vas = variableAccessAnalyzer.scanForIdentifier(accessesVariables, globalVariables, allVariables);
		model.addModelComponent(allVariables.toArray(new Variable[0]));
		variableAccesses.addAll(vas);
		model.addModelComponent(vas.toArray(new VariableAccess[0]));
		return variableAccesses;
	}

	/**
	 * Cross-check "forced" and normal VariableAccessBehavior and warn user if
	 * the behavior differs.
	 */
	@SuppressWarnings({"checkstyle:MultipleStringLiterals"})
	private void crosscheckResults(Collection<Task> allTasks) {
		for (Task t : allTasks) {

			Collection<Variable> variables = t.getVariables();
			Collection<Variable> readVariables = t.getReadVariables();
			Collection<Variable> writtenVariables = t.getWriteVariables();
			Collection<Variable> forcedReadVariables = t
					.getForcedReadVariables();
			Collection<Variable> forcedWrittenVariables = t
					.getForcedWriteVariables();

			for (Variable v : forcedReadVariables) {
				if (!readVariables.contains(v)) {
					log.warn("Variable \""
							+ v.getName()
							+ "\" is specified in \""
							+ Constants.Pragma.VARIABLE_IN
							+ "\" pragma tag, but it seems that it is not read within task "
							+ t + ".");
				}
			}
			for (Variable v : forcedWrittenVariables) {
				if (!writtenVariables.contains(v)) {
					log.warn("Variable \""
							+ v.getName()
							+ "\" is specified in \""
							+ Constants.Pragma.VARIABLE_OUT
							+ "\" pragma tag, but it seems that it is not written within task "
							+ t + ".");
				}
			}

			for (Variable v : variables) {
				if (forcedReadVariables.contains(v)
						&& !forcedWrittenVariables.contains(v)
						&& writtenVariables.contains(v)) {
					log.warn("Variable "
							+ v.getName()
							+ " is specified in \""
							+ Constants.Pragma.VARIABLE_IN
							+ "\", but not in \""
							+ Constants.Pragma.VARIABLE_OUT
							+ "\" pragma tag, but it seems that it is also written within task "
							+ t + ".");
				}

				if (!forcedReadVariables.contains(v)
						&& forcedWrittenVariables.contains(v)
						&& readVariables.contains(v)) {
					log.warn("Variable "
							+ v.getName()
							+ " is specified in \""
							+ Constants.Pragma.VARIABLE_OUT
							+ "\", but not in \""
							+ Constants.Pragma.VARIABLE_IN
							+ "\" pragma tag, but it seems that it is also read within task "
							+ t + ".");
				}
			}
		}
	}

	private List<VariableAccess> addGlobalVariables(AccessesVariables accessesvariables, Collection<Identifier> globalVariables, Collection<Variable> allVariables) {
		List<VariableAccess> returnedAccessesVariables = new ArrayList<>();
		Iterator<Identifier> iter = globalVariables.iterator();
		final Collection<Variable> taskVariables = accessesvariables.getVariables();
		if (taskVariables.stream().anyMatch(var -> var.isGlobalShared() && !var.isGlobalWritten())) {
			for (Identifier ident : globalVariables) {
				Variable variable = variableAccessAnalyzer.getVariable(ident, accessesvariables, globalVariables, allVariables);
				if (!variable.isGlobalWritten() && variable.isGlobalShared()) {
					log.debug("here creating variable access for " + ident.toString());
					VariableAccess access = new VariableAccess(accessesvariables, ident, variable, true);
					access.setBehavior(VariableAccess.AccessBehavior.UNTOUCHED);
					returnedAccessesVariables.add(access);
					accessesvariables.addVariableAccess(access);
				}

			}
		} else {
			taskVariables.stream()
					.filter(var -> var.isGlobalShared() && var.isGlobalWritten())
					.forEach(variable -> {
						log.warn("Global variable " + variable + " is globally shared but also globally written. The variable will be transformed to a local one and send through the pipline.");
					});
		}
		return returnedAccessesVariables;
	}

	private void scanForInitialisation(Traversable program, Core core, Collection<VariableAccess> variableAccesses, Collection<Variable> variables) {

		DFIterator<Initializer> iter = new DFIterator<Initializer>(program, Initializer.class);
		while (iter.hasNext()) {
			Initializer init = iter.next();
			Traversable parent = init.getParent();


			DFIterator<Traversable> i = new DFIterator<Traversable>(init.getParent(), VariableDeclarator.class);


			while (null != parent) {
				if (parent instanceof Procedure) {
					break;
				}
				if (parent instanceof Initializer) {
					// Skip nested initializations.
					break;
				}
				if (parent instanceof TranslationUnit) {
					Symbol symbol;

					if (i.hasNext()) {
						symbol = (Symbol) i.next();
					} else {
						log.error("could not find symbol for initialisation of " + init.toString());
						break;
					}

					for (Variable variable : variables) {
						if (variable.getVariableDeclarator().equals(symbol)) {
							if (variableAccesses.stream().filter(access -> access.getVariable().equals(variable)).anyMatch(access -> access.isWritten())) {
								core.addGlobalInitialisation(variable, init);
							}

						}
					}

				}
				parent = parent.getParent();
			}
		}

	}

	/**
	 * Generates {@link VariableAccess VariableAccesses} for all <code>in</code>
	 * and <code>out</code> {@link PragmaTag PragmaTags} set for the given
	 * {@link Task}
	 *
	 * @param task a task
	 */
	private void setAccessBehaviorThroughPragma(Task task) {

		Pragma pragma = task.getTaskPragma();
		if (null != pragma) {

			Collection<Variable> forcedReadVariables = new LinkedHashSet<Variable>();
			Collection<Variable> forcedWrittenVariables = new LinkedHashSet<Variable>();

			PragmaTag<String> ptInput = pragma
					.getPragmaTag(Constants.Pragma.VARIABLE_IN);

			if (null != ptInput) {
				for (Variable v : task.getVariables()) {
					if (ptInput.getEntryNames().contains(v.getName())) {
						forcedReadVariables.add(v);
						VariableAccess access = new VariableAccess(task, v,
								VariableAccess.AccessBehavior.READ, true);
						this.model.addModelComponent(access);
						task.addVariableAccess(access);
					}
				}

				if (forcedReadVariables.size() != ptInput.getEntryNames()
						.size()) {
					log.warn("Some Variables in the Pragma Tag \""
							+ Constants.Pragma.VARIABLE_IN
							+ "\" were not found inside the task's Body.");
				}
			}

			PragmaTag<String> ptOutput = pragma
					.getPragmaTag(Constants.Pragma.VARIABLE_OUT);

			if (null != ptOutput) {
				for (Variable v : task.getVariables()) {
					if (ptOutput.getEntryNames().contains(v.getName())) {
						forcedWrittenVariables.add(v);
						VariableAccess access = new VariableAccess(task, v,
								VariableAccess.AccessBehavior.WRITE, true);
						this.model.addModelComponent(access);
						task.addVariableAccess(access);
					}
				}

				if (forcedWrittenVariables.size() != ptOutput.getEntryNames()
						.size()) {

					log.warn("Some Variables in the Pragma Tag \""
							+ Constants.Pragma.VARIABLE_OUT
							+ "\" were not found inside the Task's Body.");
				}
			}

		}

	}

	/**
	 * Returns the positions in the arguments of the functioncall where the
	 * given identifier was used. Numbering starts from left to right with the
	 * first argument being 0.
	 *
	 * @param functionCall The functioncall whose arguments should be searched for the identifier.
	 * @param identifier The identifier that is searched in the arguments
	 * @return empty Set: If no arguments were found empty Set: If no arguments with the given identifier were found <p> The argument positions where the identifier was found.
	 */
	private Set<Integer> getArgumentPositions(FunctionCall functionCall,
											  Identifier identifier) {
		Set<Integer> argumentPositions = new LinkedHashSet<>();
		Integer counter = -1;
		for (Object argument : functionCall.getArguments()) {
			counter++;
			if (argument instanceof Expression) {
				Expression expression = (Expression) argument;
				DFIterator<Identifier> iter = new DFIterator<Identifier>(
						expression, Identifier.class);
				while (iter.hasNext()) {
					Identifier id = iter.next();
					if (id.equals(identifier)) {
						argumentPositions.add(counter);
					}
				}
			}
		}
		return argumentPositions;
	}

	/**
	 * Looks through all found global variables and sees if the variable is
	 * read&written by more then one task and gives an error in the log if so.
	 */
	private void analyzeGlobalVariableUsage(Collection<VariableAccess> variableAccesses, Collection<Task> allTasks) {
		Map<Variable, List<VariableAccess>> globals = new LinkedHashMap<Variable, List<VariableAccess>>();
		for (VariableAccess access : variableAccesses) {
			if (access.getVariable().isGlobal()) {
				if (!access.isNotUsed()) {
					log.debug("in task " + access.getDirectParent() + " access to " + access);
					if (globals.containsKey(access.getVariable())) {
						globals.get(access.getVariable()).add(
								access);
					} else {
						List<VariableAccess> list = new LinkedList<VariableAccess>();
						list.add(access);
						globals.put(access.getVariable(), list);
					}

				}
			}

		}
		log.debug("globals size: " + globals.size());

		globals.keySet().forEach(x -> log.debug(x.toString()));

		for (Variable var : globals.keySet()) {
			List<VariableAccess> accesses = globals.get(var).stream().filter(distinctByKey(a -> a.getDirectParent())).collect(Collectors.toList());

			//if variable access is in function, see in how many tasks this function is accessed from
			Identifier identifier = var.getIdentifiers().stream().findFirst().get();
			Procedure procedure = CetusUtils.findParentOfType(identifier, Procedure.class);
			long functionAccessSize = allTasks.stream()
					.filter(t -> t.getAllFunctionCalls().stream()
							.filter(fc -> fc.getCetusFunctionCall() != null
									&& fc.getCetusFunctionCall().getProcedure() != null
									&& fc.getCetusFunctionCall().getProcedure().equals(procedure))
							.count()>0
					).count();

			if (accesses.size() > 1 || functionAccessSize > 1) {
				var.setGlobalShared(true);
				log.debug("globalshared: " + var);
				if (accesses.stream().filter(access -> access.isWritten()).count() > 1) {
					var.setGlobalWritten(true);
				}
			}
		}
	}
}
