/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.integration_test;

import de.tu_darmstadt.rs.microStreams.system_test.test_reference.TransformationPassCollectorRunner;
import de.tu_darmstadt.rs.util.DiffDirectory;
import de.tu_darmstadt.rs.util.FileUtil;
import de.tu_darmstadt.rs.util.SpartanMCDefaultFolders;

import difflib.DiffException;
import org.apache.commons.io.FileUtils;
import org.junit.Assert;
import org.junit.Assume;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

/**
 * transforms test inputs with the given Transformation Pass Collection and compares it against gold files
 */
@RunWith(Parameterized.class)
public abstract class TransformationPassOutputTester {

	public static final Path outputRoot = Paths.get("build/testOutputs");
	private static final String ignoreFile = "ignore";
	private static final int CONTEXT_SIZE = 3;
	private static Logger logger;
	@Parameterized.Parameter(0)
	public Path inputFirmware;
	@Parameterized.Parameter(1)
	public Path verifiedOutput;
	@Parameterized.Parameter(2)
	public String name;
	@Parameterized.Parameter(3)
	public Path testFolder;

	public TransformationPassOutputTester() {
		logger = LoggerFactory.getLogger(this.getClass());
	}

	/**
	 * create data for testcases
	 *
	 * @return create data for testcases
	 *
	 * @throws IOException if something with the files fails
	 */

	@Parameterized.Parameters(name = "{index}: {2}")
	public static Collection<Object[]> data() throws IOException {
		throw new RuntimeException("Override this method!!!");
	}

	protected static Collection<Object[]> getTestProjectsFrom(Path microStreamsIntegrationtestProjects) throws IOException {
		final Path fwFolder = SpartanMCDefaultFolders.getIntegrationtestProjectsFWInput();
		final Path referenceProjects = SpartanMCDefaultFolders.getIntegrationtestProjectsReferenceProjects();
		return StreamSupport.stream(Files.newDirectoryStream(microStreamsIntegrationtestProjects).spliterator(), false)
				.filter(dir -> Files.exists(dir.resolve(fwFolder)))
				.filter(dir -> Files.exists(dir.resolve(referenceProjects)))
				.sorted()
				.map(p -> new Object[]{
						p.resolve(fwFolder),
						p.resolve(referenceProjects),
						p.getFileName().toString(),
						p
				})
				.collect(Collectors.toList());
	}

	public static Path getTestDir(Path testFolder) {
		return outputRoot.resolve(testFolder.getFileName());
	}

	public abstract TransformationPassCollectorRunner getRunnableTool(Path outputDir);

	/**
	 * The test method
	 */
	@Test
	public void testOutput() throws IOException, DiffException {
		Assume.assumeFalse("Project " + testFolder + "skipped due to ignore file", Files.exists(testFolder.resolve(ignoreFile)));
		Path outputDir = getTestDir(testFolder);
		if (Files.exists(outputDir)) {
			FileUtil.deleteDirectoryRecursively(outputDir.toFile());
		}
		String expectedExceptionMessage = getExpectedExceptionMessage();

		TransformationPassCollectorRunner objectHolder = getRunnableTool(outputDir);
		boolean runFinished = false;
		try {
			runFinished = objectHolder.run();
			assertTrue("Program finished successfully, but an thrown exception is expected for this test",runFinished && expectedExceptionMessage == null);
			assertTrue("Runnning " + objectHolder.getToolName() + " failed! ", runFinished);
			if (runFinished) {
				testProject(outputDir, verifiedOutput);
			}
		} catch (Exception e) {
			logger.info("Caught exception...now checking if there is a exception expected and the exception message is as expected");
			if (expectedExceptionMessage != null) {
				assertEquals("The expected exception message differs from the actual", expectedExceptionMessage, e.getMessage());
			} else {
				logger.error("Exception thrown but the test expects/specifies no exception");
				throw e;
			}
		}
	}

	private String getExpectedExceptionMessage() {
		String expectedExceptionMessage = null;
		if (testFolder != null) {
			Path exceptionFile = testFolder.resolve("exception.thrown");
			if (exceptionFile.toFile().exists() && exceptionFile.toFile().isFile()) {
				try {
					expectedExceptionMessage = com.google.common.io.Files.readFirstLine(exceptionFile.toFile(), StandardCharsets.UTF_8);
				} catch (IOException e) {
					logger.error("Exception file " + exceptionFile + " not readable.", e);
				}
			}
		}
		return expectedExceptionMessage;
	}

	protected void testProject(Path generatedProject, Path referenceProject) throws IOException, DiffException {
		List<String> extensions = Arrays.asList("c", "h", "xml", "mk", "sh");
		DiffDirectory.verifyFileList(referenceProject, generatedProject, extensions, 1);
		try {
			DiffDirectory.verifyFileContents(referenceProject, generatedProject, extensions, CONTEXT_SIZE, false);
		} catch (DiffException e) {
			e.printStackTrace();
			Collection<File> referenceFiles = FileUtils.listFiles(referenceProject.toFile(), extensions.toArray(new String[extensions.size()]), true);
			Collection<File> generatedFiles = FileUtils.listFiles(generatedProject.toFile(), extensions.toArray(new String[extensions.size()]), true);
			referenceFiles.stream().forEach(reference -> {
				Path referencePathSuffix = reference.toPath().subpath(referenceProject.getNameCount(), reference.toPath().getNameCount());
				Path generatedFile = generatedProject.resolve(referencePathSuffix);

				try {
					Assert.assertEquals("The reference file " + reference.toString() + " differs from the generated file " + generatedFile.toString(), FileUtils.readLines(reference, StandardCharsets.UTF_8), FileUtils.readLines(generatedFile.toFile(), StandardCharsets.UTF_8));
				} catch (IOException ex) {
					fail("Check if the generated file " + generatedFile.toString() + " exists");
				}
			});
		}
	}


}
