/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.module_test;

import de.tu_darmstadt.rs.microStreams.model.software.variable.Variable;
import de.tu_darmstadt.rs.microStreams.tools.CetusDriver;
import de.tu_darmstadt.rs.microStreams.tools.CetusUtils;
import de.tu_darmstadt.rs.microStreams.tools.VariableUtils;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.Declaration;
import cetus.hir.Declarator;
import cetus.hir.IDExpression;
import cetus.hir.Identifier;
import cetus.hir.Program;
import cetus.hir.VariableDeclaration;
import cetus.hir.VariableDeclarator;
import org.apache.commons.io.IOUtils;
import org.junit.Assert;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.slf4j.LoggerFactory;

import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.junit.Assert.assertEquals;

/**
 * Tests µStreams outputs
 * tests should be in resources/varableSizeSamples. The Variables to test must be used in main function.
 */
@RunWith(Parameterized.class)
public class VariableSizeTest {

	private final static String ARRAY = "array.c";
	private final static String ARRAY_WITH_STRUCT = "arrayWithStruct.c";
	private final static String SIMPLE_DATA_TYPES = "simpleDataTypes.c";
	private final static String STRUCT = "struct.c";
	private final static String TYPEDEF = "typedef.c";
	private final static String VARIABLE_SIZE_SAMPLES_FOLDER = "variableSizeSamples";

	public VariableSizeTest() {
		super();
	}

	@Rule
	public TemporaryFolder inputTemp = new TemporaryFolder();

	@Parameterized.Parameter(0)
	public String inputFile;
	@Parameterized.Parameter(1)
	public Map<String, Integer> variableSizes;
	@Parameterized.Parameter(2)
	public String name;

	/**
	 * create data for testcases
	 *
	 * @return create data for testcases
	 * @throws IOException on file error
	 */

	@Parameterized.Parameters(name = "{index}: {2}")
	public static Collection<Object[]> data() throws IOException {
		List<Object[]> ret = new ArrayList<>();
		ret.add(getsimpleDataTypes());
		ret.add(getArray());
		ret.add(getStruct());
		ret.add(getArrayWithStruct());
		ret.add(getTypedef());
		return ret;
	}

	private static Object[] getsimpleDataTypes() {
		Map<String, Integer> variableSizes = new HashMap<>();
		variableSizes.put("character" , 1);
		variableSizes.put("anotherchar1", 1);
		variableSizes.put("anotherchar2", 1);
		variableSizes.put("shortType" , 2);
		variableSizes.put("integer" , 2);
		variableSizes.put("longType" , 4);
		variableSizes.put("longlong" , 8);
		variableSizes.put("extInt" , 2);
		variableSizes.put("pointer" , 2);
		variableSizes.put("unsignedinteger" , 2);

		String path = getSourceDir(SIMPLE_DATA_TYPES);
		return new Object[]{path, variableSizes, SIMPLE_DATA_TYPES};
	}

	private static Object[] getArray() {
		Map<String, Integer> variableSizes = new HashMap<>();
		variableSizes.put("intarray" , 20);
		variableSizes.put("doubledec1", 6);
		variableSizes.put("doubledec2", 2);
		variableSizes.put("chararray" , 100);
		variableSizes.put("longlongarray" , 32);
		variableSizes.put("pointerarray" , 4);
		variableSizes.put("multidim", 320);
		variableSizes.put("chararraynodim1", 3);
		variableSizes.put("chararraynodim2", 18);
		variableSizes.put("enumarray", -1);
		variableSizes.put("enumarray2", -1);
		variableSizes.put("constarray", -1);

		String path = getSourceDir(ARRAY);
		return new Object[]{path, variableSizes, ARRAY};
	}

	private static Object[] getStruct() {
		Map<String, Integer> variableSizes = new HashMap<>();
		variableSizes.put("even1" , 4);
		variableSizes.put("pad" , 4);
		variableSizes.put("alignment" , 18); //allign long long to word borders
		variableSizes.put("structception", 22);
		variableSizes.put("cc",66);
		String path = getSourceDir(STRUCT);
		return new Object[]{path, variableSizes, STRUCT};
	}

	private static Object[] getArrayWithStruct() {
		Map<String, Integer> variableSizes = new HashMap<>();
		variableSizes.put("structarray" , 40);
		variableSizes.put("arrstruct", 14);
		String path = getSourceDir(ARRAY_WITH_STRUCT);
		return new Object[]{path, variableSizes, ARRAY_WITH_STRUCT};
	}

	private static Object[] getTypedef() {
		Map<String, Integer> variableSizes = new HashMap<>();
		variableSizes.put("structarray" , 100);
		String path = getSourceDir(TYPEDEF);
		return new Object[]{path, variableSizes, TYPEDEF};
	}

	private static String getSourceDir(String filename) {
		return "/" + VARIABLE_SIZE_SAMPLES_FOLDER + "/" + filename;
	}

	@Test
	public void testOutput() throws IOException {

		InputStream inputStream = getClass().getResourceAsStream(inputFile);
		final Path outPath = inputTemp.getRoot().toPath().resolve(Paths.get(inputFile).getFileName());

		OutputStream outputStream = new FileOutputStream(outPath.toFile());
		IOUtils.copy(inputStream, outputStream);
		outputStream.close();

		CetusDriver cetusDriver = new CetusDriver(Arrays.asList(outPath.toFile()), Arrays.asList(), true, Arrays.asList());
		final Program program = cetusDriver.parse();

		final List<VariableDeclarator> variableDeclarations = CetusUtils.getStreamOfAllChildren(CetusUtils.getMainProcedure(program), Identifier.class)
				.map(i -> i.getSymbol())
				.flatMap(StreamUtil.castOrEmpty(VariableDeclarator.class))
				.collect(Collectors.toList());
		final List<String> variableDeclarationsAsString = variableDeclarations.stream()
				.map(vd -> vd.getSymbolName())
				.distinct()
				.collect(Collectors.toList());
		assertArrayEquals(variableSizes.keySet().toArray(), variableDeclarationsAsString.toArray(), true);
		for (VariableDeclarator vd: variableDeclarations ) {
			final int calculatedSize = VariableUtils.getSizeInSpartanMCBytes(vd);
			final IDExpression varableName = vd.getID();
			Assert.assertNotNull("Can not get Identifier of " + vd , varableName);
			final Integer expectedVariableSize = variableSizes.get(varableName.toString());
			Assert.assertNotNull("Variable " + varableName.toString() + " was not found in the expected variables list", expectedVariableSize);
			assertEquals("Sizes for variable "+ varableName + " don't match", expectedVariableSize.intValue(),  calculatedSize);
		}


	}

	public static void assertArrayEquals(Object[] expected, Object[] actual, boolean ignoreOrder) {
		final List<Object> expectedList = new ArrayList<>(Arrays.asList(expected));
		final List<Object> actualList = new ArrayList<>(Arrays.asList(actual));

		final List<Object> actualElementsNotinExpected = new ArrayList<>(actualList);
		actualElementsNotinExpected.removeAll(actualList);

		final List<Object> expectedElementsNotinActual = new ArrayList<>(expectedList);
		expectedElementsNotinActual.removeAll(actualList);

		if (expected.length > actual.length) {
			throw new AssertionError("Array lengths don't match (expected " + expected.length + ", actual " + actual.length + ").\n The following expected elements were not found in the actual elements: " + expectedElementsNotinActual);
		} else if (expected.length < actual.length) {
			throw new AssertionError("Array lengths don't match (expected " + expected.length + ", actual " + actual.length + ").\nThe following actual elements were not found in the expected elements: " + actualElementsNotinExpected);
		} else {
			if (ignoreOrder) {
				if (!expectedElementsNotinActual.isEmpty() && !actualElementsNotinExpected.isEmpty()) {
					throw new AssertionError("Expected elements not in actual: " +expectedElementsNotinActual + "\nActual elements not in expected: "+ actualElementsNotinExpected);
				} else if (!expectedElementsNotinActual.isEmpty()) {
					throw new AssertionError("Expected elements not in actual: " +expectedElementsNotinActual);
				} else if (!actualElementsNotinExpected.isEmpty()) {
					throw new AssertionError("Actual elements not in expected: "+ actualElementsNotinExpected);
				}
			} else {
				Assert.assertEquals(expectedList, actualList);
			}
		}
	}
}
