/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.microStreams.system_test.test_reference;

import de.tu_darmstadt.rs.microStreams.TransformationPassCollectorLauncher;
import de.tu_darmstadt.rs.microStreams.model.Model;
import de.tu_darmstadt.rs.util.SpartanMCMakeUtil;

import com.google.common.io.Files;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * An abstract to provide some common functions for a transformation pass Collector runner
 */
public abstract class TransformationPassCollectorRunner {

	private static final String microStreamsArgFileName = "test.args";
	private static final String resourcesFolder = "resources";
	private static final String resourcesPlaceholder = "\\$TEST_RESOURCES";
	private static Logger log;
	private Set<String> args;
	private TransformationPassCollectorLauncher streamsToolStarter;


	public TransformationPassCollectorRunner(TransformationPassCollectorLauncher starterMain, Path testFolder, String inputFile, Path projectOutput) {
		this(starterMain, testFolder, Arrays.stream(new String[]{inputFile}).collect(Collectors.toSet()), projectOutput);
	}

	/**
	 * Configures microStremas
	 *
	 * @param starterMain the main of the according transformation pass
	 * @param testFolder the folder where to search for an args file
	 * @param inputFiles the input files... c files or a firmware directory
	 * @param projectOutput where the project shall be written to
	 */
	public TransformationPassCollectorRunner(TransformationPassCollectorLauncher starterMain, Path testFolder, Set<String> inputFiles, Path projectOutput) {
		log = LoggerFactory.getLogger(starterMain.getClass());
		log.info(getToolName() + " Test reference generator started ...");

		streamsToolStarter = starterMain;
		args = new LinkedHashSet<>();

		args.addAll(addArgsFromFile(testFolder));

		//default args that don't need to be specified by the args file
		args.add("-a");
		args.add("-p");
		args.add(projectOutput.toString());
		args.addAll(inputFiles);

		SpartanMCMakeUtil.generateNewProject(projectOutput);
	}

	/**
	 * runs microStreams
	 *
	 * @return true if microStreams ran through
	 */
	public boolean run() {
		if (streamsToolStarter.run(args.toArray(new String[0]))) {
			log.info(getToolName() + " test reference generator finished successfully ...");
			return true;
		} else {
			log.warn(getToolName() + " test reference generator was not finished successfull, see log for further details.");
			return false;
		}
	}

	public Model getModel() {
		return streamsToolStarter.getWorkerCollector().getModel();
	}


	/**
	 * gets args from a file in the project
	 * @param searchpath where to search for the arg file
	 * @return a list of parameters from the file
	 */
	public List<String> addArgsFromFile(Path searchpath) {
		if (searchpath != null) {
			Path argFile = searchpath.resolve(microStreamsArgFileName);
			if (argFile.toFile().exists() && argFile.toFile().isFile()) {
				try {
					List<String> args = Arrays.asList(Files.readFirstLine(argFile.toFile(), StandardCharsets.UTF_8).split("\\s+"));
					args = args.stream().map(a -> a.replaceAll(resourcesPlaceholder,searchpath.resolve(resourcesFolder).toString())).collect(Collectors.toList());
					return args;
				} catch (IOException e) {
					e.printStackTrace();
					getModel().printIgnoreableError(log,"Can not read arg file in project " + searchpath);
				}
			}
		}
		return new ArrayList<>();
	}

	public abstract String getToolName();

}
