/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc;

import edu.byu.ece.rapidSmith.device.Tile;

import java.awt.Point;
import java.util.function.ToIntFunction;
import java.util.stream.Collector;
import java.util.stream.Collectors;

/**
 * BoundingBox
 */
public class BoundingBox {

	public final int minX;
	public final int minY;
	public final int maxX;
	public final int maxY;

	/**
	 * Create an empty bounding box
	 */
	public BoundingBox() {
		minX = Integer.MAX_VALUE;
		minY = Integer.MAX_VALUE;
		maxX = Integer.MIN_VALUE;
		maxY = Integer.MIN_VALUE;
	}

	/**
	 * Create the specified bounding box
	 * @param minX Minimum included X
	 * @param minY Minimum included Y
	 * @param maxX Maximum included X
	 * @param maxY Maximum included X
	 */
	public BoundingBox(int minX, int minY, int maxX, int maxY) {

		this.minX = minX;
		this.minY = minY;
		this.maxX = maxX;
		this.maxY = maxY;
	}

	/**
	 * Create a bounding box from a point
	 * @param x X coordinate
	 * @param y Y coordinate
	 */
	public BoundingBox(int x, int y) {
		this(x, y, x, y);
	}

	public BoundingBox(Point p) {
		this(p.x, p.y, p.x, p.y);
	}

	public BoundingBox(Tile t) {
		this(t.getColumn(), t.getRow());
	}

	/**
	 * Extend the bounding box to include the given point
	 * @param x X coordinate
	 * @param y Y coordinate
	 * @return This bounding box extended to include the point
	 */
	public BoundingBox extend(int x, int y) {
		return new BoundingBox(
				Math.min(minX, x),
				Math.min(minY, y),
				Math.max(maxX, x),
				Math.max(maxY, y)
		);
	}
	/**
	 * Extend the bounding box to include the given point
	 * @param p the point
	 * @return This bounding box extended to include the point
	 */
	public BoundingBox extend(Point p) {
		return extend(p.x, p.y);
	}

	/**
	 * Return the union of this bounding box with the given one
	 * @param b the other bounding box
	 * @return the union
	 */
	public BoundingBox union(BoundingBox b) {
		return new BoundingBox(
				Math.min(minX, b.minX),
				Math.min(minY, b.minY),
				Math.max(maxX, b.maxX),
				Math.max(maxY, b.maxY)
		);
	}

	/**
	 * Shift the bounding box by a given amount
	 * @param x x delta
	 * @param y y delta
	 * @return the shifted bounding box
	 */
	public BoundingBox shift(int x, int y) {
		return new BoundingBox(
				minX + x,
				minY + y,
				maxX + x,
				maxY + y
		);
	}



	/**
	 * Accumulator for use in Stream.reduce
	 * @param a first bounding box
	 * @param b second bounding box
	 * @return union of the bounding boxes
	 */
	public static BoundingBox accumulator(BoundingBox a, BoundingBox b) {
		return a.union(b);
	}

	public static <T> Collector<T, ?, BoundingBox> getBoundingBox(ToIntFunction<T> colMapper, ToIntFunction<T> rowMapper) {
		return Collectors.mapping(t -> new BoundingBox(colMapper.applyAsInt(t), rowMapper.applyAsInt(t)), Collectors.reducing(new BoundingBox(), BoundingBox::accumulator));
	}

	public static Collector<Tile, ?, BoundingBox> getTileBoundingBox() {
		return getBoundingBox(Tile::getColumn, Tile::getRow);
	}

	/**
	 * Check if an intersection exists between this bounding box and another one
	 * @param other the other bounding box
	 * @return true if there is an intersection
	 */
	public boolean intersectionExists(BoundingBox other) {
		return !(this.minX > other.maxX
				|| this.maxX < other.minX
				|| this.minY > other.maxY
				|| this.maxY < other.minY);
	}

	public int circumference() {
		return (maxX - minX) + (maxY - minY);
	}

	public String toString() {
		return "[[" + minX + ";" + minY + "];[" + maxX + ";" + maxY + "]]";
	}
}
