/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc;

import de.tu_darmstadt.rs.rapidSoc.directories.ModuleRootPath;
import de.tu_darmstadt.rs.rapidSoc.module_synthesis.constraints.ConstraintCreatorFactory;
import de.tu_darmstadt.rs.rapidSoc.module_synthesis.constraints.ThorstenConstraintCreator;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigBuilder;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParametersDelegate;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.util.DeviceDatabaseProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class RapidSocProject {
	private static final Logger logger = LoggerFactory.getLogger(RapidSocProject.class);


	private static final String RAPIDSOC_VERSION = "RapidSoc-1";

	private final boolean noCheckUpToDate;

	private final boolean noSynthesize;


	private final ILibJConfigProject project;

	private final ModuleRootPath moduleRootPath;

	private final String partName;

	private final IConfigurationDocument document;

	private final TargetDescription target;

	private final ConstraintCreatorFactory constraintCreatorFactory;


	public RapidSocProject(boolean noCheckUpToDate, boolean noSynthesize, ILibJConfigProject project, ModuleRootPath moduleRootPath, TargetDescription targetDescription, ConstraintCreatorFactory constraintCreatorFactory) {
		this.noCheckUpToDate = noCheckUpToDate;
		this.noSynthesize = noSynthesize;
		this.project = project;
		this.moduleRootPath = moduleRootPath;
		this.constraintCreatorFactory = constraintCreatorFactory;
		this.document = project.getDocumentHandler().getDocument();
		this.target = targetDescription;
		this.partName = partNameToRapidSmith(targetDescription.getDevices().get(0).getName());
	}


	public RapidSocProject(boolean noCheckUpToDate, boolean noSynthesize, ILibJConfigProject project, ModuleRootPath moduleRootPath, TargetDescription targetDescription) {
		this(noCheckUpToDate, noSynthesize, project, moduleRootPath, targetDescription, ThorstenConstraintCreator::new);
	}

	public ILibJConfigProject getProject() {
		return project;
	}

	public IApplicationConfiguration getApplication() { return project.getApplication(); }

	public boolean isNoSynthesize() {
		return noSynthesize;
	}


	public boolean isNoCheckUpToDate() {
		return noCheckUpToDate;
	}

	public ModuleRootPath getModuleRootPath() {
		return moduleRootPath;
	}

	public IConfigurationDocument getDocument() {
		return document;
	}

	public TargetDescription getTarget() {
		return target;
	}

	/**
	 * Change the formatting of a part name from something like XC6SLX16-2CSG324 to xc6slx16csg324-2
	 */
	@SuppressWarnings("checkstyle:MagicNumber")
	public static String partNameToRapidSmith(String value) {
		Pattern p = Pattern.compile("(.*)-([0-9])?(.*)");

		Matcher m = p.matcher(value.toLowerCase());
		if (!m.matches()) {
			logger.warn(LogMarkers.rapidSoc, "part name matcher does not match");
			return value;
		}

		String part = m.group(1);
		String speed = m.group(2);
		String pack = m.group(3);
		return part + pack + '-' + speed;
	}

	public String getPartName() {
		return partName;
	}

	public ConstraintCreatorFactory getConstraintCreatorFactory() {
		return constraintCreatorFactory;
	}

	private Device device;

	public Device getDevice() {
		if (device == null) {
			device = DeviceDatabaseProvider.getDeviceDatabase().loadDevice(getPartName());
		}
		return device;
	}

	public static class Builder {

		@Parameter(names = {"--noCheckUpToDate", "-c"}, description = "Do not check whether modules are up to date. Saves time, but you may end up with modules that do not correspond to the current verilog sources.")
		private boolean noCheckUpToDate;

		@Parameter(names = {"--noSynthesize", "-s"}, description = "Do not synthesize modules that are not up to date or not synthesized.")
		private boolean noSynthesize;

		@ParametersDelegate
		private LibJConfigBuilder projectBuilder = new LibJConfigBuilder(RAPIDSOC_VERSION, true);


		public RapidSocProject build(ModuleRootPath moduleRootPath) {

			projectBuilder.getSpartanRoot(); // force check ROOT before start
			projectBuilder.requiresConfigFile();
			projectBuilder.autoOpenConfig();
			LibJConfigProject project = projectBuilder.build();

			IConfigurationDocument document = project.getDocumentHandler().getDocument();
			TargetDescription target = document.getTarget().getTargetDescription();

			return new RapidSocProject(noCheckUpToDate, noSynthesize, project, moduleRootPath, target);
		}
	}
}
