/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.design;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.OrderedConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.IMappedBusPortInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.MappedBusConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import java.util.HashMap;
import java.util.Map;
import java.util.stream.Stream;

/**
 * Fake connections to Bus Glue. They are not present in the design
 */
public class FakeBusGlueConnectorInstance implements IConnectorInstance {

	private static Map<IMappedBusPortInstance, Map<IMappedBusPortInstance, Integer>> connectionIndices = new HashMap<>();
	private final IMappedBusPortInstance bpi;
	private final Library library;
	private final IConfigurationDocument document;

	/**
	 * Create FakeBusGlueConnectorInstance
	 * @param bpi The busPortInstance
	 * @param library The library
	 * @param document The document
	 */
	public FakeBusGlueConnectorInstance(IMappedBusPortInstance bpi, Library library, IConfigurationDocument document) {
		this.bpi = bpi;
		this.library = library;
		this.document = document;
	}

	@Override
	public IConfigurationElement<?, ?, ?> getSourceElement() {
		return bpi.getSourceElement();
	}

	@Override
	public String getPortName() {
		return bpi.getPortName();
	}

	@Override
	public ConnectorDirection getDirection() {
		return bpi.getDirection();
	}

	@Override
	public Stream<? extends IConnection<?>> streamConnections() {
		return bpi.streamConnections()
				.map(conn -> {
					int index = getConnectionIndex(conn);
					return new OrderedConnection<IConnectorInstance>() {

						@Override
						public ConnectionType getType() {
							return ConnectionType.Partial;
						}

						@Override
						public IConnectorInstance getPortA() {
							return FakeBusGlueConnectorInstance.this;
						}

						@Override
						public IConnectorInstance getPortB() {
							IModuleConfiguration<?> module = BusGlueToModuleInstance.getOrgate(conn.getOppositeConnector(bpi), library, document);
							return DocumentUtil.getPortHandler(module, "a");
						}

						@Override
						public boolean isConnected() {
							return conn.isConnected();
						}

						@Override
						public void connect() {
							throw new UnsupportedOperationException();
						}

						@Override
						public void disconnect() {
							throw new UnsupportedOperationException();

						}

						@Override
						public boolean isValid(boolean report) {
							return conn.isValid(report);
						}

						@Override
						public boolean isPartialA() {
							return false;
						}

						@Override
						public boolean isPartialB() {
							return true;
						}

						@Override
						public int getStartA() {
							return 0;
						}

						@Override
						public int getEndA() {
							return IConnection.FULL_CONNECTOR_WIDTH;
						}

						@Override
						public int getStartB() {
							return index * bpi.getConnectorWidth();
						}

						@Override
						public int getEndB() {
							return (index + 1) * bpi.getConnectorWidth() - 1;
						}
					};
				});
	}

	@Override
	public boolean isConnected() {
		return bpi.isConnected();
	}

	@Override
	public IConnectorDescription.Type getType() {
		return bpi.getType();
	}

	@Override
	public int getConnectionCount() {
		return bpi.getConnectionCount();
	}

	@Override
	public int getConnectorWidth() {
		return bpi.getConnectorWidth();
	}

	private int getConnectionIndex(MappedBusConnection conn) {
		IMappedBusPortInstance connector = conn.getOppositeConnector(bpi);

		Map<IMappedBusPortInstance, Integer> indexMap = connectionIndices.computeIfAbsent(connector, (k) -> new HashMap<>());

		return indexMap.computeIfAbsent(bpi, (k) -> indexMap.size());
	}

	@Override
	public String toString() {
		return bpi.toString();
	}
}
