/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.design;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;
import de.tu_darmstadt.rs.rapidSoc.LogMarkers;
import de.tu_darmstadt.rs.rapidSoc.RapidSocProject;
import de.tu_darmstadt.rs.rapidSoc.directories.ModuleDirectories;
import de.tu_darmstadt.rs.rapidSoc.directories.ModulePath;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescription;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionInstance;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionLoader;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionVariantPort;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionVariant;
import de.tu_darmstadt.rs.rapidSoc.module.ModulePlacement;
import de.tu_darmstadt.rs.rapidSoc.module_synthesis.ModuleToVerilog;
import de.tu_darmstadt.rs.rapidSoc.parameters.ParameterReplaceFactory;
import de.tu_darmstadt.rs.rapidSoc.parameters.ReplaceRam;
import de.tu_darmstadt.rs.rapidSoc.placer.ModulePlacer;
import de.tu_darmstadt.rs.rapidSoc.placer.cost.HpwlCost;
import de.tu_darmstadt.rs.rapidSoc.placer.overlap.TileOverlapDetector;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IMemoryModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.LazyBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.IBusGlue;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.IMappedBusPortInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.BuildInfo;
import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.table.Column;
import de.tu_darmstadt.rs.util.table.TableFormat;
import de.tu_darmstadt.rs.util.table.TableFormatter;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.Parameters;
import com.beust.jcommander.ParametersDelegate;
import edu.byu.ece.rapidSmith.design.Attribute;
import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.ModuleInstance;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.PrimitiveType;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.xml.bind.JAXBException;

/**
 * Main Entry point for RapidSoc
 */
//We are in a command line program, help and errors are printed to stdout.
@SuppressWarnings("checkstyle:SystemOut")
@Parameters(commandNames = {"design", "moduledesign"}, commandDescription = "RapidSoc main application")
public class ModuleDesign implements Runnable {


	private static final Logger logger = LoggerFactory.getLogger(ModuleDesign.class);

	public static final String visualizeWhenExplanation = "Possible values for visualization settings are (case-insensitive):\n"
			+ VisualizeWhen.explainPossible()
			+ "\nParameter can be specified multiple times to enable multiple settings.";


	@Parameter(names = {"--route"}, description = "Route Design")
	private boolean route = false;


	@Parameter(names = {"--visugui", "-g"}, description = "Set when to show Visualization GUI. For possible values, see below", converter = VisualizeWhen.AbbreviationConverter.class)
	private List<VisualizeWhen> visuGUI;

	@Parameter(names = {"--visufile", "-vf"}, description = "Set when to save Visualization Image to subfolder \"visu\" of project dir. For possible values, see below", converter = VisualizeWhen.AbbreviationConverter.class)
	private List<VisualizeWhen> visuFile;

	@Parameter(names = {"--simulation"}, description = "Write simulation Verilog output")
	private boolean simulation;

	@ParametersDelegate
	private RapidSocProject.Builder projectBuilder = new RapidSocProject.Builder();

	private RapidSocProject project;


	private IConfigurationDocument document;
	private final Map<StreamUtil.Pair<Collection<IModuleDescription>, Map<String, String>>, Module> moduleCache = new HashMap<>();
	private final Map<StreamUtil.Pair<Collection<IModuleDescription>, Map<String, String>>, Set<String>> modulesRequiredBy = new HashMap<>();

	public ModuleDesign() {

	}

	/**
	 * Get the suffix that a port that connects to this one has
	 *
	 * @param s the current port's suffix
	 * @return suffix of a connected port
	 */
	public static char getConnectedSuffix(char s) {
		switch (s) {
			case 'O':
				return 'I';
			case 'I':
				return 'O';
			case 'T':
				return 'T';
			default:
				throw new RuntimeException("invalid suffix " + s);
		}
	}

	@Override
	public void run() {

		Tracer.start();
		try {

			Tracer.enter("Library load");

			project = projectBuilder.build(ModuleDirectories.getRootDir());
			document = project.getDocument();
			Tracer.exit();

			Tracer.enter("Build design");
			Design d = buildDesign();
			Tracer.exit();

		} finally {
			System.out.println();
			Tracer.printStats(System.out);

			if (project != null && project.getProject() != null) {
				try {
					Tracer.serialize(new FileOutputStream(new File(project.getProject().getWorkingDir().toFile(), "runtime.xml")));
				} catch (JAXBException | FileNotFoundException e) {
					throw new RuntimeException(e);
				}
			}

		}

	}
/*
	private Module getModule(Collection<IModuleConfiguration<?>> modules, TargetDescription targetBoard) {
		Collection<IModuleDescription> md = modules.stream().map(IModuleConfiguration::getModuleDescription).collect(Collectors.toSet());
		Map<String, String> attrMap = ModuleDirectories.getAttributeMap(modules);

		StreamUtil.Pair<Collection<IModuleDescription>, Map<String, String>> p = new StreamUtil.Pair<>(md, attrMap);

		final String nameInDesign = moduleDirectories.getModuleName(modules);
		modulesRequiredBy.computeIfAbsent(p, x -> new HashSet<>()).add(nameInDesign);


		//We are using containsKey, because the map may contain null values if a module was not loadable
		if (moduleCache.containsKey(p)) {
			return moduleCache.get(p);
		} else {
			Module module = extractModule(modules, targetBoard, md, attrMap);
			moduleCache.put(p, module);
			return module;
		}
	}*/

	/**
	 * Build the design
	 *
	 * @return the finished design
	 */
	private Design buildDesign() {

		Design design = createDesign();

		LazyBuildProcess buildProcessCache = new LazyBuildProcess(document);

		ModuleDescriptionInstance iobModuleDescriptionInstance = buildIOBModule(design);

		Tracer.enter("Load module descriptions");
		Map<IModuleConfiguration<?>, ModuleDescriptionInstance> moduleDescriptionInstanceMap = loadModuleDescriptions(design, buildProcessCache);
		Tracer.exit();

		Tracer.enter("Connect modules 1");
		logger.info("PortConnector start");
		PortConnector portConnector = new PortConnector(document, design, moduleDescriptionInstanceMap, iobModuleDescriptionInstance, buildProcessCache, project.getApplication().getLibrary());
		Set<PortNet> portNets = portConnector.connectPorts();
		//TODO: remove debug output
		portNets.stream()
				.sorted(
						Comparator.comparing((PortNet a) -> a.getSourceModule().toString())
								.thenComparing(a -> a.getSourcePort().getName())
				)
				.map(PortNet::toString)
				.forEach(logger::trace);
		logger.info("PortConnector done");
		Tracer.exit();

		Tracer.enter("Place modules");
		ModulePlacer placer = placeModules(design, moduleDescriptionInstanceMap, portNets);
		Tracer.exit();

		Tracer.enter("Instantiate Modules");
		Map<ModuleDescriptionInstance, ModuleInstance> miMap = placer.instantiateModules(design, moduleDescriptionInstanceMap.values().stream().distinct().collect(Collectors.toSet()));
		//Map<IModuleConfiguration<?>, ModuleInstance> moduleMap = instantiateModules(design, buildProcessCache);
		Map<IModuleConfiguration<?>, ModuleInstance> moduleMap = moduleDescriptionInstanceMap.keySet().stream().collect(Collectors.toMap(
				Function.identity(),
				p -> miMap.get(moduleDescriptionInstanceMap.get(p))
		));
		Map<String, edu.byu.ece.rapidSmith.design.Port> fpgaPins = buildPinModule(design);
		Tracer.exit();

		Tracer.enter("Connect modules");
		ModuleConnector connector = new ModuleConnector(document, design, moduleMap, fpgaPins, buildProcessCache, project.getApplication().getLibrary());
		Set<Net> interModuleNets = connector.connectModules();
		interModuleNets.stream().sorted(Comparator.comparing(Net::getName)
		).map(net -> net.toString(design.getWireEnumerator())).forEach(logger::trace);
		Tracer.exit();

		Tracer.enter("Replace parameters");
		replaceParameters(design, moduleMap, buildProcessCache);
		Tracer.exit();

		//instanceInfo(design, moduleMap);

		//findRoutingConflicts(design);

		Tracer.enter("Place BUFG");
		placeBufg(design, moduleMap);
		Tracer.exit();


		Tracer.enter("Flatten");
		design.flattenDesign();
		Tracer.exit();

		fixShapes(design);

		if (route) {
			routeDesign(design);
		}

		Tracer.enter("Saving");
		String filename;
		if (route) {
			filename = "rapidSoc_ownrouter.xdl";
		} else {
			filename = "rapidSoc.xdl";
		}
		design.saveXDLFile(project.getProject().getWorkingDir().resolve("build").resolve(filename));

		generatePCF(design, fpgaPins);
		Tracer.exit();

		if (simulation) {
			try (Tracer ignored = Tracer.enter("Create Simulation Verilog files")) {
				new ModuleToVerilog(design, moduleMap).toVerilog(project.getProject().getBuildDir());
			}
		}

		return design;
	}

	@NotNull
	private ModuleDescriptionInstance buildIOBModule(Design design) {
		Tracer.enter("Create IO module description");
		ModuleDescription description = buildIOBModuleDescription(design);
		ModuleDescriptionInstance mdi = new ModuleDescriptionInstance(description, "IO Module");//TODO name
		//TODO placing outside of placer?
		final ModulePlacement placement = mdi.getDescription().getPlacements().iterator().next();
		if (placement==null)
			throw new IllegalStateException("no placement for iob module");
		mdi.setPlacement(placement);
		Tracer.exit();
		return mdi;
	}

	private ModulePlacer placeModules(
			Design design,
			//Map<IModuleConfiguration<?>, ModuleInstance> moduleMap,
			Map<IModuleConfiguration<?>, ModuleDescriptionInstance> modulePlacementMap,
			//Map<String, edu.byu.ece.rapidSmith.design.Port> fpgaPins,
			//Set<Net> interModuleNets,
			Set<PortNet> portNets
	) {
		//Filter out clock nets, as they are routed over the clock distribution network
		//TODO same thing for PortNets in ModulePlacer
		/*Set<Net> routeNets = interModuleNets.stream()
				.filter(net -> net.getSource() != null)
				.filter(net -> net.getPins().stream().noneMatch(pin -> pin.getName().equals("CLK")))
				.collect(Collectors.toSet());*/

		final ModulePlacer placer = new ModulePlacer(design,
				//TODO: this is becoming a mess
				new HpwlCost(),
				//new CostFunctionLogger(new HpwlCost()),
				//new OverlapDetectorCombiner.RequireAll(
				//		new BoundingBoxOverlapDetector(false),
				//		new TileOverlapDetector(true)
				//),
				//new OverlapDetectorCombiner.RequireAny(
				//		new BoundingBoxOverlapDetector(true),
				//		new TileOverlapDetector(true)
				//),
				//new BoundingBoxOverlapDetector(true),
				new TileOverlapDetector(true),
				//new OverlapDetectorLogger(new TileOverlapDetector(true)),
				project.getProject().getWorkingDir(),
				visuFile,
				visuGUI
		);
		placer.place(
				modulePlacementMap.values().stream().distinct().collect(Collectors.toSet()),
				portNets,
				1000 //TODO config
		);
		return placer;
	}

	/**
	 * Create empty RapidSmith design with correct part name
	 *
	 * @return the design
	 */
	private Design createDesign() {
		Design design = new Design();
		design.setName(VerilogLang.getVerilogName(document.getName()));
		//Does not work without these...
		final String propKey = "_DESIGN_PROP";
		design.getAttributes().add(new Attribute(propKey, "P3_PLACE_OPTIONS", "EFFORT_LEVEL:high"));
		design.getAttributes().add(new Attribute(propKey, "", "P3_PLACED:"));

		design.addComment(new BuildInfo(ModuleDesign.class).toString());


		IConfigurationDocument.ITargetConfiguration target = document.getTarget();
		design.setPartName(project.getPartName());
		return design;
	}

	private void placeBufg(Design design, Map<IModuleConfiguration<?>, ModuleInstance> moduleMap) {
		Iterator<PrimitiveSite> bufgSites = Arrays.stream(design.getDevice().getAllPrimitiveSitesOfType(PrimitiveType.BUFGMUX)).iterator();
		design.getInstances().forEach(instance -> {
			if (!instance.isPlaced()) {
				if (instance.getType() != PrimitiveType.BUFG && instance.getType() != PrimitiveType.BUFGMUX) {
					throw new RuntimeException("Found unplaced instance " + instance.getName() + " of type " + instance.getType() + " in " + instance.getModuleInstanceName());
				}

				if (!bufgSites.hasNext()) {
					throw new RuntimeException("not enough BUFGs");
				} else {
					PrimitiveSite site = bufgSites.next();
					logger.info(LogMarkers.rapidSoc, "placing " + instance.getName() + " at " + site);
					instance.place(site);
				}
			}
		});
	}

	private Map<IModuleConfiguration<?>, ModuleDescriptionInstance> loadModuleDescriptions(Design design, LazyBuildProcess buildProcessCache) {
		Map<IModuleConfiguration<?>, ModuleDescriptionInstance> result = new HashMap<>();

		ModuleDescriptionLoader loader = new ModuleDescriptionLoader(project);

		boolean allFound = streamModuleConfigurations(buildProcessCache)
				.map(jModules -> {
					final ModuleDescription moduleDescription = loader.loadAllVariants(jModules);

					if (moduleDescription!=null) {
						final ModuleDescriptionInstance mdi = new ModuleDescriptionInstance(moduleDescription, ModuleDirectories.getModuleConfigurationName(jModules));
						jModules.forEach(jModule ->	result.put(jModule, mdi));
						return true;
					}
					return false;


				}).reduce(true, Boolean::logicalAnd); //same as allMatch without short circuiting

		//TODO adjust messages
		if (!allFound) {
			if (project.isNoSynthesize()) {
				logger.warn(LogMarkers.rapidSoc, "Consider removing the command line switch --noSynthesize to automate the work");
			}

			throw new RuntimeException("One or more modules need to be (re-) synthesized before continuing. See log for details.");
		}

		return result;
	}
/*
	private Map<IModuleConfiguration<?>, ModuleInstance> instantiateModules(Design design, LazyBuildProcess buildProcessCache) {
		IConfigurationDocument document = project.getDocumentHandler().getDocument();

		TargetDescription target = document.getTarget().getTargetDescription();

		Map<IModuleConfiguration<?>, ModuleInstance> result = new HashMap<>();

		//Create dummy orgates for all bus glue
		IBusGlue<IModuleConfiguration<?>> instantiateBusGlue = new IBusGlue<IModuleConfiguration<?>>() {
			@Override
			public IModuleConfiguration<?> glueOrMergedBus(IMappedBusPortInstance busPortInstance, Function<IMappedBusPortInstance, String> glueInputName, Function<IMappedBusPortInstance, String> glueOutputName) {
				return BusGlueToModuleInstance.getOrgate(busPortInstance, project.getApplication().getLibrary(), document);
			}
		};

		Stream<Collection<IModuleConfiguration<?>>> glueModules = buildProcessCache.getBusLayout().streamMappedPorts()
				.filter(IMappedBusPortInstance::requiresBusGlue)
				.<IModuleConfiguration<?>>map(port -> port.applyBusGlue(instantiateBusGlue, x -> null, x -> null))
				.map(Collections::singleton);


		Set<IModuleConfiguration<?>> handled = new HashSet<>();

		//Put subsystems and local memories together
		Set<Collection<IModuleConfiguration<?>>> subsystemsWithMem = new HashSet<>();

		document.getSubsystems().forEach(subsystem -> {
			IConfigurationElement<?, ?, ?> connected = DocumentUtil.getBusHandler(subsystem, "memcode-bus").streamOppositeConnectors().collect(StreamUtil.exactlyOne()).getSourceElement();
			if (connected instanceof IMemoryModuleConfiguration<?>) {
				IMemoryModuleConfiguration<?> mem = (IMemoryModuleConfiguration<?>) connected;
				if (mem.getModuleDescriptionName().equals("spartanmc_mem_local")) {
					handled.add(subsystem);
					handled.add(mem);

					List<IModuleConfiguration<?>> l = new ArrayList<>(2);
					l.add(subsystem);
					l.add(mem);
					subsystemsWithMem.add(l);
				}
			}
		});


		Stream<? extends List<IModuleConfiguration<?>>> regularModules = document.streamModulesRecursive().filter(m -> !handled.contains(m)).map(Collections::singletonList);

		boolean allFound = Stream.concat(subsystemsWithMem.stream(), Stream.concat(regularModules, glueModules))
				.map(jModules -> {
					//Module rModule = getModule(jModules, target);
					Module rModule = null;// = getModule(jModules, target);

					String name = ModuleDirectories.getModuleName(jModules);
					try {
						System.out.println("hash: " + moduleRootPath.getHashDir(jModules.stream().map(IModuleConfiguration::getModuleDescription).collect(Collectors.toSet()), ModuleDirectories.getAttributeMap(jModules)));
					} catch (IOException e) {
						e.printStackTrace();
					}
					//TODO grep all variants
					try {
						System.out.println(moduleRootPath.getModuleDirectories(jModules).map(Path::toString).reduce((a, b) -> a + ";" + b));
					} catch (IOException e) {
						e.printStackTrace();
					}
					try {
						System.out.println(moduleRootPath.getHashDir(jModules).resolve("0").resolve("build").resolve("spartanmc.mod").toFile().exists());
						//System.out.println(ModuleDirectories.getHashDir(jModules.stream().map(IModuleConfiguration::getModuleDescription).collect(Collectors.toSet()), ModuleDirectories.getAttributeMap(jModules)).toPath().resolve("0").resolve("build").resolve(ModuleDirectories.getModuleCompoundName(jModules.stream().map(IModuleConfiguration::getModuleDescription).collect(Collectors.toSet()))+".mod").toFile());					rModule = JacksonUtil.loadModule(ModuleDirectories.getHashDir(jModules).toPath().resolve("0").resolve("build").resolve("spartanmc.mod").toFile(),true);
					} catch (IOException e) {
						e.printStackTrace();
					}

					Pattern pattern = Pattern.compile("\\d+");
					try {
						for (Path entry : Files.newDirectoryStream(moduleRootPath.getHashDir(jModules))) {
							if (Files.isDirectory(entry) && pattern.matcher(entry.getFileName().toString()).matches()) {
								logger.info("Found variant: " + entry.getFileName().toString() + " (" + entry.toString() + ")");
								logger.warn("Module: " + entry.resolve("build").resolve("spartanmc.mod").toString());
								logger.warn("Exists: " + Files.exists(entry.resolve("build").resolve("spartanmc.mod")));
							}
						}
					} catch (IOException e) {
						e.printStackTrace();
					}


					Tracer.enterLoop("Create Instance");
					ModuleInstance m = design.createModuleInstance(name, rModule);
					if (m == null) {
						logger.error("Could not create Module Instance for " + name);
						return Tracer.exit(false);
					} else {

						jModules.forEach(jModule -> result.put(jModule, m));
						return Tracer.exit(true);
					}
				}).reduce(true, Boolean::logicalAnd); //same as allMatch without short circuiting


		outputUsageTable();


		if (!allFound) {
			if (noSynthesize) {
				logger.warn(LogMarkers.rapidSoc, "Consider removing the command line switch --noSynthesize to automate the work");
			}

			throw new RuntimeException("One or more modules need to be (re-) synthesized before continuing. See log for details.");
		}

		return result;
	}*/

	private Stream<Collection<IModuleConfiguration<?>>> streamModuleConfigurations(LazyBuildProcess buildProcessCache) {
		//modules that have been handled and have to be ignored to avoid duplicates
		Set<IModuleConfiguration<?>> handled = new HashSet<>();

		//create modules for bus glue
		Stream<Collection<IModuleConfiguration<?>>> glueModules = createBusGlue(buildProcessCache, document);

		//Put subsystems and local memories together and add them to the handled modules
		Set<Collection<IModuleConfiguration<?>>> subsystemsWithMem = createSubsystemsWithMem(document);
		subsystemsWithMem.stream().flatMap(Collection::stream).forEach(handled::add);

		//modules that have not been handled
		Stream<? extends List<IModuleConfiguration<?>>> regularModules = document.streamModulesRecursive().filter(m -> !handled.contains(m)).map(Collections::singletonList);

		return StreamUtil.concat(subsystemsWithMem.stream(), regularModules, glueModules);
	}

	/**
	 * merges spartanmc with its local memory
	 *
	 * @param document the jconfig configuration document
	 * @return a set of the merged modules
	 */
	private Set<Collection<IModuleConfiguration<?>>> createSubsystemsWithMem(IConfigurationDocument document) {
		//TODO reenable subsystems with mem
		//return Collections.emptySet();

		Set<Collection<IModuleConfiguration<?>>> subsystemsWithMem = new HashSet<>();
		document.getSubsystems().forEach(subsystem -> {
			IConfigurationElement<?, ?, ?> connected = DocumentUtil.getBusHandler(subsystem, "memcode-bus").streamOppositeConnectors().collect(StreamUtil.exactlyOne()).getSourceElement();
			if (connected instanceof IMemoryModuleConfiguration<?>) {
				IMemoryModuleConfiguration<?> mem = (IMemoryModuleConfiguration<?>) connected;
				if (mem.getModuleDescriptionName().equals("spartanmc_mem_local")) {
					subsystemsWithMem.add(Arrays.asList(subsystem, mem));
				}
			}
		});
		return subsystemsWithMem;
	}

	/**
	 * Create dummy orgates for all bus glue
	 */
	private Stream<Collection<IModuleConfiguration<?>>> createBusGlue(LazyBuildProcess buildProcessCache, IConfigurationDocument document) {
		IBusGlue<IModuleConfiguration<?>> instantiateBusGlue = new IBusGlue<IModuleConfiguration<?>>() {
			@Override
			public IModuleConfiguration<?> glueOrMergedBus(IMappedBusPortInstance busPortInstance, Function<IMappedBusPortInstance, String> glueInputName, Function<IMappedBusPortInstance, String> glueOutputName) {
				return BusGlueToModuleInstance.getOrgate(busPortInstance, project.getApplication().getLibrary(), document);
			}
		};

		return buildProcessCache.getBusLayout().streamMappedPorts()
				.filter(IMappedBusPortInstance::requiresBusGlue)
				.<IModuleConfiguration<?>>map(port -> port.applyBusGlue(instantiateBusGlue, x -> null, x -> null))
				.map(Collections::singleton);
	}

	private void outputUsageTable() {
		final TableFormatter table = new TableFormatter(TableFormat.CHARACTERS,
				new Column("%s", Column.Alignment.LEFT),
				new Column("%s", Column.Alignment.LEFT),
				new Column("%s", Column.Alignment.LEFT),
				new Column("%s", Column.Alignment.LEFT)
		);
		table.addHeaderRow("Instance Name", "Module", "Hash", "State");
		final Comparator<StreamUtil.Pair<Collection<IModuleDescription>, Map<String, String>>> compareModuleNames = Comparator.comparing(p -> ModuleDirectories.getModuleDescriptionName(p.getA()));
		final Comparator<StreamUtil.Pair<Collection<IModuleDescription>, Map<String, String>>> compareModuleHashes = Comparator.comparing(p -> ModulePath.getHashName(p.getB()));
		moduleCache.keySet().stream()
				.sorted(compareModuleNames.thenComparing(compareModuleHashes))
				.forEach(p -> {
					final boolean loadable = moduleCache.get(p) != null;
					String modName = ModuleDirectories.getModuleDescriptionName(p.getA());
					String state = loadable? "Ok" : "Not loadable";
					final String hashName = ModulePath.getHashName(p.getB());
					modulesRequiredBy.get(p).stream()
							.sorted()
							.forEach(nameInDesign -> table.addContentRow(nameInDesign, modName, hashName, state));
				});
		table.output().forEach(logger::info);
	}

	private void routeDesign(Design design) {
		logger.info(LogMarkers.rapidSoc, "Routing");
		Tracer.enter("Router");
		/*Pathfinder router = new Pathfinder();
		router.setDesign(design);
		router.setDev(design.getDevice());
		router.setWe(design.getWireEnumerator());

		// route design
		router.routeDesign();*/
		//TODO
		Tracer.exit();
	}

	private void replaceParameters(Design design, Map<IModuleConfiguration<?>, ModuleInstance> moduleMap, LazyBuildProcess buildProcessCache) {
		File projectPath = project.getProject().getWorkingDir().toFile();
		moduleMap.forEach((jModule, rModule) -> {
			Stream.of(ParameterReplaceFactory.replacers).forEach(replacer -> {
				if (replacer.hasChangeableParameter(jModule)) {
					logger.info(LogMarkers.parameters, "Running " + replacer + " on " + jModule);
					replacer.replaceParameter(design, jModule, rModule, projectPath, buildProcessCache);
				}
			});
		});

		ReplaceRam.replace(design, projectPath, moduleMap);
	}

	private void generatePCF(Design design, Map<String, edu.byu.ece.rapidSmith.design.Port> fpgaPins) {
		try (PrintWriter out = new PrintWriter(new FileWriter(new File(project.getProject().getWorkingDir().toFile(), "build/rapidSoc.pcf")))) {
			DocumentUtil.streamConnectorHandlers(document)
					.filter(p -> p.getFrequency().isPresent())
					.filter(IConnectorInstance::isConnected)
					.forEach(fpgaPin -> {

						edu.byu.ece.rapidSmith.design.Port port = fpgaPins.get(fpgaPin.getPortName() + ".O");
						String name = port.getPin().getNet().getName();

						//PCF files don't seem to support specifying clock frequencies, so calculate period
						final double gigaPrefix = 1E9;
						double periodNs = gigaPrefix / fpgaPin.getFrequency().orElseThrow();
						out.println(String.format(Locale.ROOT, "NET \"%s\" PERIOD = %f nS HIGH 50.000 %% ;", name, periodNs));


					});
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	private void createPort(ConnectorDirection direction, Instance pinInst, IPinHandler fpgaPin, Map<String, edu.byu.ece.rapidSmith.design.Port> result) {
		Pin pin = new Pin();
		if (direction.equals(ConnectorDirection.Input)) {
			pin.setPinName("I");
			pin.setIsOutputPin(true);

		} else if (direction.equals(ConnectorDirection.Output)) {
			pin.setPinName("O");
			pin.setIsOutputPin(false);

		} else if (direction.equals(ConnectorDirection.InOut)) {
			pin.setPinName("T");
			pin.setIsOutputPin(false);
		} else {
			throw new RuntimeException("unsupported port type: " + direction);
		}


		pin.setInstance(pinInst);
		edu.byu.ece.rapidSmith.design.Port port = new edu.byu.ece.rapidSmith.design.Port(fpgaPin.getPortName() + "." + pin.getName(), pin);


		result.put(fpgaPin.getPortName() + "." + getConnectedSuffix(pin.getName().charAt(0)), port);
	}

	public static final char getDirectionSuffix(ConnectorDirection direction) {
		switch (direction) {
			case Input:
				return 'I';
			case Output:
				return 'O';
			case InOut:
				return 'T';
			default:
				throw new IllegalStateException("No suffix for connector direction " + direction);
		}
	}

	private ModuleDescription buildIOBModuleDescription(Design design) {
		Tracer.enter("Build IOB module description");
		Device device = design.getDevice();

		Map<String, ModuleDescriptionVariantPort> portMap = DocumentUtil.streamConnectorHandlers(document)
				.filter(IConnectorInstance::isConnected)
				.flatMap(port -> {
					ConnectorDirection direction = port.getDirection();
					switch (direction) {
						case Input:
						case Output:
							return Stream.of(new StreamUtil.Pair<>(port, getDirectionSuffix(direction)));
						case InOut:
							return Stream.of(
									new StreamUtil.Pair<>(port, getDirectionSuffix(ConnectorDirection.Input)),
									new StreamUtil.Pair<>(port, getDirectionSuffix(ConnectorDirection.Output)),
									new StreamUtil.Pair<>(port, getDirectionSuffix(ConnectorDirection.InOut))
							);
						default:
							throw new IllegalStateException("Connector direction not handled for " + port);
					}
				})
				.collect(Collectors.toMap(
						pair -> pair.getA().getPortName() + "." + pair.getB(),
						pair -> new ModuleDescriptionVariantPort(pair.getA().getPortName() + "." + pair.getB(), new int[]{device.getPrimitiveSite(pair.getA().getPinDescription().getFpgaPinName()).getTile().getUniqueAddress()})
				));


		final ModuleDescriptionVariant variant = new ModuleDescriptionVariant(
				"IO Module",
				device,
				new int[]{0},
				0,
				portMap.values().stream().flatMapToInt(ModuleDescriptionVariantPort::streamTileNumbers).distinct().sorted().toArray(),
				new int[]{},
				new BoundingBox(0, 0),
				new BoundingBox(0, 0),
				portMap,
				null
		);
		return Tracer.exit(new ModuleDescription("IOBS",Collections.singleton(variant), null));
	}

	@SuppressWarnings("checkstyle:MultipleStringLiterals")
	private Map<String, edu.byu.ece.rapidSmith.design.Port> buildPinModule(Design design) {

		Tracer.enter("Build FPGA pins");
		Map<String, edu.byu.ece.rapidSmith.design.Port> result = new HashMap<>();

		DocumentUtil.streamConnectorHandlers(document)
				.filter(IConnectorInstance::isConnected)
				.forEach(fpgaPin -> {

					Instance pinInst = new Instance(fpgaPin.getPortName(), PrimitiveType.IOB);


					PrimitiveSite site = design.getDevice().getPrimitiveSite(fpgaPin.getPinDescription().getFpgaPinName());
					pinInst.place(site);


					design.addInstance(pinInst);

					ConnectorDirection direction = fpgaPin.getTargetDirection();


					pinInst.addAttribute("DIFFI_INUSED", "", "#OFF");
					pinInst.addAttribute("DIFF_TERM", "", "#OFF");
					pinInst.addAttribute("DRIVE_0MA", "", "#OFF");
					pinInst.addAttribute("OUT_TERM", "", "#OFF");
					pinInst.addAttribute("PADOUTUSED", "", "#OFF");
					pinInst.addAttribute("PAD", "pin_" + fpgaPin.getPortName(), "");
					pinInst.addAttribute("PCI_RDYUSED", "", "#OFF");
					pinInst.addAttribute("PULLTYPE", "", "#OFF");
					pinInst.addAttribute("IN_TERM", "", "#OFF");

					if (direction == ConnectorDirection.Input || direction == ConnectorDirection.InOut) {
						pinInst.addAttribute("BYPASS_MUX", "", "I");
						//Invert if necessary
						if (fpgaPin.isInverted()) {
							pinInst.addAttribute("IMUX", fpgaPin.getPortName() + ".IMUX", "I_B");
						} else {
							pinInst.addAttribute("IMUX", fpgaPin.getPortName() + ".IMUX", "I");
						}
						pinInst.addAttribute("INBUF", "pin_" + fpgaPin.getPortName() + "_IBUF", "");
						pinInst.addAttribute("ISTANDARD", "", "LVCMOS25");//TODO use configured
					} else {
						pinInst.addAttribute("BYPASS_MUX", "", "#OFF");
						pinInst.addAttribute("IMUX", "", "#OFF");
					}

					if (direction == ConnectorDirection.Output || direction == ConnectorDirection.InOut) {
						if (fpgaPin.isInverted()) {
							throw new RuntimeException("Inverted outputs are not supported. Consider adding an inverted net on " + fpgaPin.getPortName());
						}
						pinInst.addAttribute("DRIVEATTRBOX", "", "12");//TODO use configured
						pinInst.addAttribute("OSTANDARD", "", "LVCMOS25");//TODO use configured
						pinInst.addAttribute("OUSED", "", "0");
						pinInst.addAttribute("OUTBUF", "pin_" + fpgaPin.getPortName() + "_OBUF", "");//OUTPUT
						pinInst.addAttribute("PRE_EMPHASIS", "", "#OFF");//XDL bug, we cannot add this to inputs or they are considered as outs
						pinInst.addAttribute("SLEW", "", "SLOW");//TODO use configured
						pinInst.addAttribute("SUSPEND", "", "3STATE");
					} else {
						pinInst.addAttribute("DRIVEATTRBOX", "", "#OFF");
						pinInst.addAttribute("OUSED", "", "#OFF");
						pinInst.addAttribute("SLEW", "", "#OFF");
						pinInst.addAttribute("SUSPEND", "", "#OFF");

					}

					if (direction == ConnectorDirection.InOut) {
						pinInst.addAttribute("TUSED", "", "0");
					} else {
						pinInst.addAttribute("TUSED", "", "#OFF");
					}


					if (direction == ConnectorDirection.InOut) {
						createPort(ConnectorDirection.InOut, pinInst, fpgaPin, result);
						createPort(ConnectorDirection.Input, pinInst, fpgaPin, result);
						createPort(ConnectorDirection.Output, pinInst, fpgaPin, result);
					} else {
						createPort(direction, pinInst, fpgaPin, result);
					}


				});
		return Tracer.exit(result);
	}

	public static void fixShapes(Design design) {
		design.getInstances().forEach(inst -> {
			//TODO can we really just remove them?
			inst.removeAttribute("_INST_PROP");
		});
	}


}
