/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.design;

import de.tu_darmstadt.rs.rapidSoc.LogMarkers;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionInstance;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionPort;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionVariant;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionVariantPort;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.ClkPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.LazyBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.layouting.buses.IMappedBusPortInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.StreamUtil;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.ModuleInstance;
import edu.byu.ece.rapidSmith.design.Port;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import static de.tu_darmstadt.rs.rapidSoc.design.ModuleDesign.getConnectedSuffix;

/**
 * Connect the RapidSmith modules as specified by the spmc modules.
 */
public class PortConnector {

	private static final Logger logger = LoggerFactory.getLogger(PortConnector.class);

	@NotNull private final LazyBuildProcess buildProcessCache;
	@NotNull private final IConfigurationDocument document;
	@NotNull private final Design design;
//	@Deprecated private final Map<? extends IConfigurationElement<?, ?, ?>, ModuleInstance> moduleMap;//TODO remove
	@NotNull private final Map<? extends IConfigurationElement<?, ?, ?>, ModuleDescriptionInstance> moduleDescriptionInstanceMap;
//	@Deprecated private final Map<String, Port> fpgaPins;//TODO remove
	@NotNull private final ModuleDescriptionVariant ioModule;
	@NotNull private final ModuleDescriptionInstance ioModuleDescriptionInstance;
	@NotNull private final Library library;
	@NotNull private static final String prefixGeneratedNets = "genNet_";
//	@Nonnull private final Map<IConnectorInstance, PortNet> portNetMap;
	@NotNull private final Map<StreamUtil.Pair<ModuleDescriptionInstance, String>, PortNet> portNetMap;


	private Set<PortBitSuffix> passthroughPorts;

	/**
	 * Create ModuleConnector instance
	 * @param document the jc document
	 * @param design the rapidsmith design
	 * @param moduleDescriptionInstanceMap map from jc module to module placement
	 * @param ioModuleDescriptionInstance The instantiated FPGA pins
	 * @param buildProcessCache The build cache
	 * @param library The library
	 */
	public PortConnector(
			@NotNull final IConfigurationDocument document,
			@NotNull final Design design,
			@NotNull final Map<? extends IConfigurationElement<?, ?, ?>, ModuleDescriptionInstance> moduleDescriptionInstanceMap,
			@NotNull final ModuleDescriptionInstance ioModuleDescriptionInstance,
			@NotNull final LazyBuildProcess buildProcessCache,
			@NotNull final Library library
	) {
		this.buildProcessCache = buildProcessCache;
		this.document = document;

		this.design = design;
//		this.moduleMap = moduleMap;
		this.moduleDescriptionInstanceMap = moduleDescriptionInstanceMap;
//		this.fpgaPins = fpgaPins;
		this.library = library;
		this.ioModuleDescriptionInstance = ioModuleDescriptionInstance;
		this.ioModule = ioModuleDescriptionInstance.getDescription().getVariants().iterator().next();
		this.portNetMap = new HashMap<>();
	}

	/**
	 * Creates the connections between modules from port to port
	 * @return a set of all {@link PortNet}s connecting the modules
	 */
	@NotNull
	public Set<PortNet> connectPorts() {
		passthroughPorts = new HashSet<>();

		Stream<PortNet> busNets = connectBusPorts();//TODO FIXME
		Stream<PortNet> regularNets = connectRegularPorts();
		connectPassthroughs();

		return Stream.concat(busNets, regularNets).collect(Collectors.toSet());
	}

/*	private Port getPort(IConnectorInstance jPort, int bit, char suffix) {
		String name = getPortBitName(jPort, bit, suffix);

		if (jPort.getSourceElement() instanceof IConfigurationDocument) {
			return fpgaPins.get(name);
		} else {

			ModuleInstance moduleInstance = getModuleInstance(jPort);
			if (moduleInstance == null) {
				throw new RuntimeException("module instance for " + jPort.getSourceElement() + " not found.");
			}
			return moduleInstance.getModule().getPort(name);
		}
	}*/

	private ModuleDescriptionPort getMDPort(IConnectorInstance jPort, int bit, char suffix) {
		String name = getPortBitName(jPort, bit, suffix);

		if (jPort.getSourceElement() instanceof IConfigurationDocument) {
			return ioModule.getPort(name);
		} else {

			ModuleDescriptionInstance moduleDescriptionInstance = getModuleInstance(jPort);
			if (null == moduleDescriptionInstance) {
				throw new RuntimeException("module instance for " + jPort.getSourceElement() + " not found.");
			}
			return moduleDescriptionInstance.getDescription().getPort(name);
		}
	}

	@NotNull
	private String getPortBitName(@NotNull final IConnectorInstance jPort, final int bit, final char suffix) {
		return jPort.getPortName() + (jPort.getConnectorWidth() > 1? "[" + bit + "]" : "") + "." + suffix;
	}

/*	@Deprecated
	private ModuleInstance getModuleInstance(IConnectorInstance connectorInstance) {
		return moduleMap.get(connectorInstance.getSourceElement());
	}*/

	private ModuleDescriptionInstance getModuleInstance(IConnectorInstance connectorInstance) {
		if (connectorInstance.getSourceElement() instanceof IConfigurationDocument) {
			return ioModuleDescriptionInstance;
		} else {
			ModuleDescriptionInstance res = moduleDescriptionInstanceMap.get(connectorInstance.getSourceElement());
			Objects.requireNonNull(res, "Module not found");
			return res;
		}
	}


	private Stream<StreamUtil.Pair<ModuleInstance, Port>> findConnectedPortsByPassthroughInput(IConfigurationElement<?, ?, ?> module, Port port) {
		return port.getPassthroughPorts().stream().flatMap(out -> null);
	}

	/**
	 *
	 * @param port
	 * @param bit
	 * @param suffix
	 * @return
	 */
	private Stream<StreamUtil.Pair<ModuleDescriptionInstance, String>> findConnectedPorts(IConnectorInstance port, int bit, char suffix) {
		char connSuffix = getConnectedSuffix(suffix);

		return port.streamConnections().flatMap(conn -> {
			if (conn.isMyPartial(port) && (bit < conn.getMyStart(port) || bit > conn.getMyEnd(port))) {
				return Stream.empty();
			}

			int connectedBit = bit - conn.getMyStart(port) + conn.getOppositeStart(port);


//			Port rConnPort = getPort(conn.getOppositeConnector(port), connectedBit, connSuffix);
			String rConnPortName = getPortBitName(conn.getOppositeConnector(port), connectedBit, connSuffix);

			ModuleDescriptionInstance connModPlace = getModuleInstance(conn.getOppositeConnector(port));
			Objects.requireNonNull(conn.getOppositeConnector(port), "Port has no opposite connector");
			Objects.requireNonNull(connModPlace, "Opposite connector has no module placement");
			String name = getPortBitName(conn.getOppositeConnector(port), connectedBit, connSuffix);
			final String arrow = "    -> ";
			if (null == rConnPortName) {
				//Missing self connections are assumed to be connected internally, so only debug
				//TODO equals?
				if (connModPlace == getModuleInstance(port)) {
					logger.debug(LogMarkers.connections, arrow + name + " at " + connModPlace + " not found, is self connection");
				} else {
					logger.warn(LogMarkers.connections, arrow + name + " at " + connModPlace + " not found");
				}
				return Stream.empty();
			} else {
				logger.debug(LogMarkers.connections, arrow + name + " at " + connModPlace);

				return Stream.of(new StreamUtil.Pair<>(connModPlace, name));
			}
		});
	}

	private Stream<PortNet> connectPortBit(IConnectorInstance port, int bit, char suffix) {
//		@Deprecated
//		Port rPort = getPort(port, bit, suffix);
		ModuleDescriptionPort mdPort = getMDPort(port, bit, suffix);
		String portName = getPortBitName(port, bit, suffix);
//		@Deprecated
//		ModuleInstance modInst = getModuleInstance(port);
		ModuleDescriptionInstance moduleDescriptionInstance = getModuleInstance(port);
		logger.debug(LogMarkers.connections, portName + " at " + moduleDescriptionInstance);

		// if the module does not have the required port
		if (null == mdPort) {
			if (port.isConnected()) {

//TODO
				boolean allInternal = port.streamOppositeConnectors().map(this::getModuleInstance).allMatch(mi -> mi == moduleDescriptionInstance);

				if (!allInternal) {
					throw new RuntimeException("output port " + getPortBitName(port, bit, suffix) + " not found in " + port.getSourceElement().getName() + ", but connections are present to " +
							port.streamOppositeConnectors().map(c -> c.getSourceElement().toString() + "." + c.getPortName()).collect(Collectors.toList())
					);
				} else { // if the port is not used outside of the module, ignore it
					return Stream.empty();
				}
			} else { // if the port is not connected in jconfig, ignore it
				return Stream.empty();
			}
		}

//		Pin pin = rPort.streamModuleInstancePins(modInst).collect(StreamUtil.atMostOne()).orElse(null);

//		logger.debug(LogMarkers.connections, getPortBitName(port, bit, suffix) + " at " + modInst + ": " + pin + " (" + (pin != null? pin.getInstance().getType() : "") + ")");


		if (mdPort.getPassthroughFrom() != null) {
			passthroughPorts.add(new PortBitSuffix(port, bit, suffix, portName));

			logger.info(LogMarkers.connections, "Ignoring port " + port.getSourceElement().getName()+"."+port.getPortName() + ", because is passes through " + mdPort.getPassthroughFrom());//TODO getName
			return Stream.empty();
		}

/*		if (pin == null && rPort.getNetType() == NetType.WIRE) {
			throw new RuntimeException("Port " + port + " has no pin");
		}*/

//		String moduleName = modInst != null ? modInst.getName() : "root";

/*		Net net;
		if (pin == null || pin.getNet() == null) {
			net = new Net(prefixGeneratedNets + moduleName + "." + getPortBitName(port, bit, suffix), rPort.getNetType());
			design.addNet(net);

			if (pin != null) {
				net.addPin(pin);
			}
		} else {
			net = pin.getNet();
		}*/

		boolean isClockOutput = (port.getType() == IConnectorDescription.Type.Clk) && port.getDirection() == ConnectorDirection.Output;
		boolean fpgaPinConnectedToClockInputs = null == moduleDescriptionInstance && port.streamOppositeConnectors().allMatch((IConnectorInstance c) -> c instanceof ClkPort);


/*		if ((isClockOutput || fpgaPinConnectedToClockInputs) && port.isConnected()) {
			logger.info(LogMarkers.connections, port.getPortName() + " is clock, inserting bufg");

			Instance bufgInst = new Instance(prefixGeneratedNets + moduleName + "." + getPortBitName(port, bit, suffix) + "_bufg", PrimitiveType.BUFG);
			design.addInstance(bufgInst);

			Pin bufgI = new Pin(false, "I0", bufgInst);
			Pin bufgO = new Pin(true, "O", bufgInst);

			net.addPin(bufgI);

			net = new Net(prefixGeneratedNets + bufgInst.getName(), NetType.WIRE);
			net.addPin(bufgO);
			design.addNet(net);
		}*/

//		final Net netToConnect = net;
		final PortNet portNet = getPortNet(port, portName);

		if(port.getType() != IConnectorDescription.Type.Clk
			&&port.streamOppositeConnectors().allMatch(c -> c.getType() != IConnectorDescription.Type.Clk))//(IConnectorInstance c) -> c instanceof ClkPort))
				doConnection(port, bit, suffix, portNet, moduleDescriptionInstance);

		return Stream.of(portNet);//TODO
	}

	private PortNet getPortNet(IConnectorInstance port, String name) {
		ModuleDescriptionInstance mp = getModuleInstance(port);
//		return portNetMap.computeIfAbsent(name, (n) -> new PortNet(getModulePlacement(port), n.getB()));
		return portNetMap.computeIfAbsent(new StreamUtil.Pair<>(mp, name), (p) -> new PortNet(getModuleInstance(port), name));
	}

	private void doConnection(IConnectorInstance port, int bit, char suffix, @NotNull final PortNet portNet, ModuleDescriptionInstance modPlace) {
		findConnectedPorts(port, bit, suffix).forEach(pair -> {
			ModuleDescriptionInstance connModPlace = pair.getA();


			String rConnPort = pair.getB();


			if(
					Objects.nonNull(modPlace.getDescription().getPort(getPortBitName(port, bit, suffix)))
					&&
					Objects.nonNull(connModPlace.getDescription().getPort(rConnPort))
			) {
				portNet.addSink(connModPlace, rConnPort);
				//this is used for passthrough connecting
				portNetMap.put(new StreamUtil.Pair<>(connModPlace, rConnPort), portNet);
			} else {
				handleMissingPort(modPlace, getPortBitName(port, bit, suffix), connModPlace, rConnPort);
			}


/*			rConnPort.streamModuleInstancePins(connModInst)
					.forEach(connPin -> {
						if (connPin.getNet() != null) {
							throw new RuntimeException("Pin to connect to already has a net: " + connPin + " in " + rConnPort.getName() + " in " + connModInst + ". Currently connected to " + connPin.getNet() + ". New connection to " + rPort);
						}
						logger.debug(LogMarkers.connections, "        -> " + connPin);
						netToConnect.addPin(connPin);
					});*/
		});
	}

	private void handleMissingPort(ModuleDescriptionInstance sourceModule, String sourcePortName, ModuleDescriptionInstance destModule, String destPortName) {
		if (sourceModule == destModule) {
/*				if (rConnPort.streamModuleInstancePins(modInst).map(p -> p.getNet() != null).collect(StreamUtil.AllOrAnyMatch.collectorSomeForbidden()).orElse(false)) {
					Net connNet = rConnPort.streamModuleInstancePins(modInst).map(Pin::getNet).distinct().collect(StreamUtil.exactlyOne());
					if (connNet != netToConnect) {
						throw new RuntimeException("wrong net: " + connNet + " instead of " + netToConnect);
					}
					logger.info(LogMarkers.connections, "Ignoring self connection: " + port.getPortName() + " to " + rConnPort.getName());
					return;
				} else {
					logger.warn(LogMarkers.connections, "Self connection is not correctly entered for " + port);
				}*/
			logger.info(LogMarkers.connections, "Ignoring self connection: " + sourcePortName + " to " + destPortName);
		} else {
			if(null == destModule.getDescription().getPort(destPortName)) {
				logger.warn("Destination port " + destModule + ":" + destPortName + " does not exist");
				logger.warn("Can't connect " + sourceModule + ":" + sourcePortName + " to " + destModule + ":" + destPortName);
			}
			if(null == sourceModule.getDescription().getPort(sourcePortName)) {
				logger.error("Source port " + sourceModule + ":" + sourcePortName + " does not exist");
				logger.error("Can't connect " + sourceModule + ":" + sourcePortName + " to " + destModule + ":" + destPortName);
				throw new IllegalStateException("Source (output) port not found");
			}
		}
	}

		private Stream<? extends IConnectorHandler<?, ?>> streamAllConnectorHandlers(IConfigurationDocument document) {
		Stream<IPortHandler<IPort>> modulePorts = document.streamModulesRecursive().flatMap(DocumentUtil::streamConnectorHandlers);
		Stream<IPinHandler> pins = DocumentUtil.streamConnectorHandlers(document);
		return Stream.concat(modulePorts, pins);
	}

	private Stream<PortNet> connectBusPorts() {
		return Stream.concat(
				connectNonGlueBusPorts(),
				connectGlueBusPorts()
		);
	}

	private Stream<PortNet> connectGlueBusPorts() {
		return streamBusGluePorts()
				.flatMap(port -> {
					//Glue module's output
					FakeBusGlueOutputConnectorInstance glueOutput = new FakeBusGlueOutputConnectorInstance(port, library, document);

					//Get the outputs
					Stream<FakeBusGlueConnectorInstance> inputs = port.streamOppositeConnectors()
							//Get fake connections that connect to an explicit glue element
							.map(bpi -> new FakeBusGlueConnectorInstance(bpi, library, document))
							.peek(x -> {
								logger.info(LogMarkers.connections, "connecting glue bus port " + x.getPortName());

							});

					return Stream.concat(Stream.of(glueOutput), inputs);
				}).flatMap(this::connectPort);
	}

	/**
	 * Stream all inputs that require Bus glue
	 *
	 * @return inputs that require bus glue
	 */
	private Stream<IMappedBusPortInstance> streamBusGluePorts() {
		return buildProcessCache.getBusLayout()
				.streamMappedPorts()
				.filter(IConnectorInstance::isConnected)
				.filter(p -> p.getDirection() == ConnectorDirection.Input)
				.filter(IMappedBusPortInstance::requiresBusGlue);
	}

	private Stream<PortNet> connectNonGlueBusPorts() {
		return buildProcessCache.getBusLayout()
				.streamMappedPorts()
				.filter(IConnectorInstance::isConnected)
				.filter(p -> p.getDirection() != ConnectorDirection.Input)
				//Check if the input requires bus glue. Throws an exception if we mix inputs requiring glue or not
				.filter(p -> !p.streamConnections()
						.map(con -> con.getOppositeConnector(p).requiresBusGlue())
						.collect(StreamUtil.AllOrAnyMatch.collectorSomeForbidden())
						.orElse(false)
				)
				.flatMap(this::connectPort);
	}

	private Stream<PortNet> connectRegularPorts() {
		return streamAllConnectorHandlers(document)
				.filter(IConnectorInstance::isConnected)
				.filter(port -> port.getDirection() != ConnectorDirection.Input)
				.flatMap(this::connectPort);
	}

/*	private void staticallySourceUnconnected() {
		streamAllConnectorHandlers(document)
				.filter(p -> p.getDirection() == ConnectorDirection.Input)
				.forEach(port -> {
					ModuleInstance modInst = getModuleInstance(port);

					for (int i = 0; i < port.getConnectorWidth(); i++) {
						Port rPort = getPort(port, i, 'I');
						if (rPort == null) {
							continue;
						}
						if (rPort.streamModuleInstancePins(modInst).anyMatch(p -> p.getNet() == null)) {

							Net n = new Net();
							n.setName("staticSource_" + port.getSourceElement().getName() + "." + getPortBitName(port, i, 'I'));
							n.setType(NetType.GND);
							rPort.streamModuleInstancePins(modInst).forEach(n::addPin);
							design.addNet(n);
						}
					}
				});
	}*/

	private Stream<PortNet> connectPort(IConnectorInstance port) {
		return IntStream.range(0, port.getConnectorWidth()).boxed().flatMap(bit -> {

			if (port.getDirection() == ConnectorDirection.InOut) {
				Stream<PortNet> t = connectPortBit(port, bit, 'O');
				//if we don't filter here, we connect all Ts twice, once for each direction.
				//check if the port is a fpga pin
				if (!(port.getSourceElement() instanceof IConfigurationDocument)) {
					Stream<PortNet> t2 = connectPortBit(port, bit, 'T');
					return Stream.concat(t, t2);
				}
				return t;
			} else {
				return connectPortBit(port, bit, 'O');
			}

		});
	}

	private void connectPassthroughs() {
		while (!passthroughPorts.isEmpty()) {
			Set<PortBitSuffix> unknown = new HashSet<>();

			passthroughPorts.forEach(p -> {
				//Get net
//				ModuleInstance modInst = getModuleInstance(p.port);
				ModuleDescriptionInstance mdi = getModuleInstance(p.port);
				ModuleDescriptionVariantPort passthroughFrom = mdi.getDescription().getPort(p.portName).getPassthroughFrom();
				ModuleDescriptionPort x = portNetMap.values().stream()
//						.filter()
//						.flatMap(PortNet::streamSinkPorts)
//						.map(PortNet::getSinks)
						.filter(net -> net.getSinks().contains(new StreamUtil.Pair<>(mdi, passthroughFrom.getName())))
						.collect(StreamUtil.exactlyOne())
						.getSourcePort();
//						.filter(p_ -> p_.equals(passthrough));

//				Net n = p.portName.getPassthroughFrom().streamModuleInstancePins(modInst).map(Pin::getNet).distinct().collect(StreamUtil.atLeastOneOrEqual());
				PortNet net = portNetMap.get(new StreamUtil.Pair<>(mdi, passthroughFrom.getName()));

				if (null != net) {
//					logger.debug(LogMarkers.connections, "Passthrough port " + p.port.getPortName() + " in " + p.port.getSourceElement().getName() + " is sourced by " + n);

					//Add the pins to the net
//					doConnection(p.port, p.bit, p.suffix, p.portName, n, modInst);
					doConnection(p.port, p.bit, p.suffix, net, mdi);
				} else {
					//Net has to be analyzed again
					unknown.add(p);
				}
			});

			if (unknown.size() == passthroughPorts.size()) {
				throw new RuntimeException("Could not determine sources of passthrough ports. Loop in passthrough ports?");
			}

			//The rest has to be analyzed again
			passthroughPorts = unknown;
		}
	}


	/**
	 * Represents a single direction in a single bit in a single port
	 */
	private static class PortBitSuffix {

		public final IConnectorInstance port;
		public final int bit;
		public final char suffix;

		public final String portName;

		public PortBitSuffix(IConnectorInstance port, int bit, char suffix, String portName) {
			this.port = port;
			this.bit = bit;
			this.suffix = suffix;
			this.portName = portName;
		}
	}
}
