/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.design;

import com.beust.jcommander.IStringConverter;
import com.beust.jcommander.ParameterException;

import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Set when to show visualizations
 */
public enum VisualizeWhen {
	BacktrackStep("When Backtracking in initial Placement", "b", "bs", "backtrack", "backtractstep"),
	AfterInitial("After initial placement", "i", "ai", "initial", "afterinitial"),
	EachStep("After each placement change", "e", "es", "step", "eachstep"),
	Final("Final Placement", "f", "final");

	final List<String> abbreviations;
	final String explanation;


	VisualizeWhen(String explanation, String... abbreviations) {
		this.explanation = explanation;
		this.abbreviations = Arrays.asList(abbreviations);
	}

	/**
	 * Gen an instance from an abbreviation
	 * @param s the abbreviation
	 * @return A matching VisualizeWhen instance or null if nothing matches
	 */
	public static VisualizeWhen fromAbbreviation(String s) {
		s = s.toLowerCase();
		for (VisualizeWhen visualizeWhen : values()) {
			if (visualizeWhen.abbreviations.contains(s)) {
				return visualizeWhen;
			}
		}
		return null;
	}


	/**
	 * Explanation text for the possible values
	 * @return explanation
	 */
	public static String explainPossible() {
		return Stream.of(values())
				.map(v -> "\t" + v.abbreviations.stream().map(a -> "\"" + a + "\"").collect(Collectors.joining(", ")) + ": " + v.explanation)
				.collect(Collectors.joining("\n"));
	}

	/**
	 * Convert abbreviated VisualizeWhen to instances
	 */
	public static class AbbreviationConverter implements IStringConverter<VisualizeWhen> {

		private String optionName;

		public AbbreviationConverter(String optionName) {
			this.optionName = optionName;
		}

		public String getOptionName() {
			return this.optionName;
		}


		@Override
		public VisualizeWhen convert(String value) {
			VisualizeWhen result = fromAbbreviation(value);
			if (result == null) {
				throw new ParameterException("\"" + getOptionName() + "\": Invalid value \"" + value + "\"");
			}
			return result;
		}
	}
}
