/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.directories;

import de.tu_darmstadt.rs.rapidSoc.LogMarkers;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class HashPath extends PathWrapper {
	private static final Logger logger = LoggerFactory.getLogger(HashPath.class);
	private static final Pattern VARIANT_PATTERN = Pattern.compile("^\\d+$");
	static final String hashCheckName = "attributes.txt";
	private final ModulePath modulePath;


	public HashPath(Path sourcePath, ModulePath modulePath) {
		super(sourcePath);
		this.modulePath = modulePath;
	}

	/**
	 * Get a directory to place a new module in
	 * @return a new directory
	 * @throws IOException on broken IO
	 */
	public VariantPath getNewVariantDirectory() throws IOException {
		int i = 0;
		do {
			Path r = getPath().resolve(Integer.toString(i));
			if (!Files.exists(r)) {
				return new VariantPath(r, this);
			}
			i++;
		}
		while (true);
	}


	@NotNull
	public Stream<VariantPath> getModuleDirectories() {
		try {
			return Files.list(getPath())
					.filter(Files::isDirectory)
					.filter(p -> VARIANT_PATTERN.matcher(p.getFileName().toString()).matches())
					.map(sourcePath -> new VariantPath(sourcePath, this));
		} catch (IOException e) {
			logger.error(e.getMessage());
			return Stream.empty();
		}
	}

	public VariantPath resolveVariant(String variant) {
		return new VariantPath(getPath().resolve(variant), this);
	}



	/**
	 * Load the attribute map
	 * @return the attribute map
	 */
	public Map<String, String> loadAttributeMap() {
		if (!Files.exists(getPath())) {
			logger.warn(LogMarkers.rapidSoc, "Hash dir not found.");
			return null;
		}

		Path mapFile = getPath().resolve(hashCheckName);
		return loadAttributeMap(mapFile);
	}

	private static Map<String, String> loadAttributeMap(Path file) {
		if (!Files.exists(file)) {
			logger.warn(LogMarkers.rapidSoc, "Map file not found.");
			return null;
		}

		try {
			return Files.lines(file).map(x -> {
				int eqIdx = x.lastIndexOf('=');
				String name = x.substring(0, eqIdx);
				String val = x.substring(eqIdx + 1);
				return new StreamUtil.Pair<>(name, val);
			}).collect(Collectors.toMap(StreamUtil.Pair::getA, StreamUtil.Pair::getB));
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}



	/**
	 * Write the hash checking file
	 * @param hashDir the hash directory
	 * @param attrMap the attribute map
	 */
	private void writeHashCollisionCheck(Path hashDir, Map<String, String> attrMap) throws IOException {
		Path hashCheck = hashDir.resolve(hashCheckName);
		Files.write(hashCheck, hashCheckContents(attrMap).collect(Collectors.toList()));
	}

	void doHashCheck(Map<String,String> attrMap) throws IOException {
		if (!Files.exists(getPath())) {
			Files.createDirectories(getPath());
			writeHashCollisionCheck(getPath(), attrMap);
		} else {
			checkHashCollision(attrMap);
		}
	}



	/**
	 * Sort the map by key and map lines to key=value
	 * @param attrMap the map
	 * @return a sorted stream of strings in the form "&lt;key>=&lt;value>"
	 */
	public static Stream<String> hashCheckContents(Map<String, String> attrMap) {
		return attrMap.keySet().stream().sorted().map(k -> k + "=" + attrMap.get(k));
	}

	/**
	 * Check for a hash collision
	 * @param attrMap the map of attributes
	 */
	private void checkHashCollision(Map<String, String> attrMap) throws IOException {

		Path attrFile = getPath().resolve(hashCheckName);
		List<String> lines = Files.readAllLines(attrFile);

		int[] i = {0};
		hashCheckContents(attrMap).forEach(line -> {
			String readLine = lines.get(i[0]++);
			if (!line.equals(readLine)) {
				throw new RuntimeException("Hash collision! Existing attr line is \"" + readLine + "\", new line is \"" + line + "\"");
			}
		});
	}

	public ModulePath getModulePath() {
		return modulePath;
	}

	public String getHash() {
		return getPath().getFileName().toString();
	}
}
