/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.directories;


import de.tu_darmstadt.rs.rapidSoc.parameters.ParameterReplaceFactory;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.dependencies.IValueResolver;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Paths;
import java.util.Collection;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Map a module to directories where the module is implemented
 *
 * The first subdirectory is the module name. The second is a hash of a map
 * attr name -> attr value. Collisions in the hash are only detected, but not
 * actively avoided. A third subdirectory is used to allow for multiple
 * versions of a module that differ in area constraints.
 *
 */
public final class ModuleDirectories {
	private static final Logger logger = LoggerFactory.getLogger(ModuleDirectories.class);
	static final ModuleRootPath moduleRootDir;

	private ModuleDirectories() {
	}

	public static Collection<IModuleDescription> getModuleDescriptions(Collection<IModuleConfiguration<?>> mcs) {
		return mcs.stream().map(IModuleConfiguration::getModuleDescription).collect(Collectors.toSet());
	}

	/**
	 * For a module, get a map of attribute name -> attribute value, filtered
	 * to attributes that cannot be changed in the synthesized module
	 * @param modules the modules
	 * @return map attr name -> attr value
	 */
	public static Map<String, String> getAttributeMap(Collection<IModuleConfiguration<?>> modules) {
		return modules.stream()
				.flatMap(module ->
						DocumentUtil.streamAllParameters(module)
								.filter(attr -> !canBeReplaced(module.getModuleDescription(), attr.getParameterDescription(), module))
				).collect(Collectors.toMap(IParameterHandler::getName, (t) -> t.getValue().get()));
	}


	private static String getDescriptionsName(Stream<IModuleDescription> stream) {
		return stream.map(IModuleDescription::getName).sorted().collect(Collectors.joining("+"));
	}

	public static String getModuleDescriptionNameFromConf(Collection<IModuleConfiguration<?>> mcs) {
		return getDescriptionsName(mcs.stream().map(IModuleConfiguration::getModuleDescription));
	}


	public static String getModuleConfigurationName(Collection<IModuleConfiguration<?>> mcs) {
		return mcs.stream().map(IConfigurationElement::getGlobalName).collect(Collectors.joining("+"));
	}


	public static String getModuleDescriptionName(Collection<IModuleDescription> md) {
		return getDescriptionsName(md.stream());
	}


	/**
	 * Tests if a parameter can be changed in the synthesized module
	 * @param parameter the parameter to test
	 * @return true if the param can be replaced
	 */
	private static boolean canBeReplaced(IModuleDescription md, IParameter parameter, IValueResolver resolver) {
		if (parameter.isSimulationOnly()) {
			return true;
		}

		if (parameter.isVirtual()) {
			return true;
		}

		if (resolver != null) {
			if (!parameter.resolveRelevantIf(resolver).get()) {
				return true;
			}
		}

		return ParameterReplaceFactory.findReplacerForAttribute(md, parameter) != null;
	}

	public static Map<String, String> getDefaultAttributeMap(IModuleDescription md) {
		return md.getParameterGroups().stream().flatMap(p -> p.getParameters().stream())
				.filter(attr -> !canBeReplaced(md, attr, null)).collect(Collectors.toMap(IParameter::getName, IParameter::getDefaultValue));
	}



	static {
		String dir = System.getenv("SPARTANMC_MODULES");
		if (dir != null) {
			moduleRootDir = new ModuleRootPath(Paths.get(dir.trim()));
		} else {
			moduleRootDir = null;
		}
	}

	public static ModuleRootPath getRootDir() {
		if (moduleRootDir == null) {
			throw new RuntimeException("Please set the SPARTANMC_MODULES environment variable to the modules directory.");
		}
		return moduleRootDir;
	}
}
