/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module;

import de.tu_darmstadt.rs.util.tracer.Tracer;

import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.BeanDescription;
import com.fasterxml.jackson.databind.DeserializationConfig;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.JsonDeserializer;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.databind.deser.BeanDeserializerModifier;
import com.fasterxml.jackson.databind.deser.ResolvableDeserializer;
import com.fasterxml.jackson.databind.deser.std.StdDeserializer;
import com.fasterxml.jackson.databind.module.SimpleModule;
import com.fasterxml.jackson.databind.ser.std.StdSerializer;
import com.fasterxml.jackson.dataformat.smile.SmileFactory;
import com.fasterxml.jackson.dataformat.smile.SmileGenerator;
import com.fasterxml.jackson.module.afterburner.AfterburnerModule;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.jackson.JacksonDeserListener;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.Tile;
import edu.byu.ece.rapidSmith.util.DeviceDatabaseProvider;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Objects;



/**
 * Utility class for full module serialization/deserialization
 */
public final class JacksonUtil {

	private static final Logger logger = LoggerFactory.getLogger(JacksonUtil.class);

	private static final JacksonUtil INSTANCE = new JacksonUtil();
	private static final String ATTRIBUTE_DEVICE = "device";

	private static final StdSerializer<Device> deviceSerializer = new StdSerializer<Device>(Device.class) {
		@Override
		public void serialize(Device device, JsonGenerator gen, SerializerProvider provider) throws IOException {
			gen.writeString(device.getPartName());
		}
	};
	private static final StdSerializer<Tile> tileSerializer = new StdSerializer<Tile>(Tile.class) {
		@Override
		public void serialize(Tile tile, JsonGenerator gen, SerializerProvider provider) throws IOException {
			gen.writeString(tile.getName());
		}
	};
	private static final StdSerializer<PrimitiveSite> siteSerializer = new StdSerializer<PrimitiveSite>(PrimitiveSite.class) {
		@Override
		public void serialize(PrimitiveSite site, JsonGenerator gen, SerializerProvider provider) throws IOException {
			gen.writeString(site.getName());
		}
	};
	private static final StdDeserializer<Device> deviceDeserializer = new StdDeserializer<Device>(Device.class) {
		@Override
		public Device deserialize(JsonParser p, DeserializationContext ctxt) throws IOException {
			Device device = DeviceDatabaseProvider.getDeviceDatabase().loadDevice(p.getValueAsString());
			ctxt.setAttribute(ATTRIBUTE_DEVICE, device);
			return device;
		}
	};
	private static final StdDeserializer<Tile> tileDeserializer = new StdDeserializer<Tile>(Tile.class) {
		@Override
		public Tile deserialize(JsonParser p, DeserializationContext ctxt) throws IOException {
			//p.getCurrentName()
			Object device = ctxt.getAttribute(ATTRIBUTE_DEVICE);
			if (device instanceof Device) {
				return ((Device) device).getTile(p.getValueAsString());
			} else {
				throw new IllegalStateException("Unknown device: can not deserialize Tile");
			}
		}
	};
	private static final StdDeserializer<PrimitiveSite> siteDeserializer = new StdDeserializer<PrimitiveSite>(PrimitiveSite.class) {
		@Override
		public PrimitiveSite deserialize(JsonParser p, DeserializationContext ctxt) throws IOException {
			Object device = ctxt.getAttribute(ATTRIBUTE_DEVICE);
			if (device instanceof Device) {
				return ((Device) device).getPrimitiveSite(p.getValueAsString());
			} else {
				throw new IllegalStateException("Unknown device: can not deserialize PrimitiveSite");
			}
		}
	};
/*
	private static final StdConverter<Object, Instance> instanceConverter = new StdConverter<Object, Instance>() {

		@Override
		public Instance convert(Object obj) {
			Instance inst = (Instance) obj;
			for (Pin pin : inst.getPins()) {
				pin.setInstanceInternal(inst);
			}
			return inst;
		}
	};
	*/
	private static final class NotifyingDeserializer extends JsonDeserializer<JacksonDeserListener> implements ResolvableDeserializer {

		private final JsonDeserializer<? extends JacksonDeserListener> src;

		public NotifyingDeserializer(JsonDeserializer<? extends JacksonDeserListener> src) {
			this.src = Objects.requireNonNull(src);
		}

		@Override
		public JacksonDeserListener deserialize(JsonParser p, DeserializationContext ctxt) throws IOException, JsonProcessingException {
			final JacksonDeserListener deserialize = src.deserialize(p, ctxt);
			deserialize.afterDeserialization();
			return deserialize;
		}

	@Override
	public void resolve(DeserializationContext ctxt) throws JsonMappingException {
			if (src instanceof ResolvableDeserializer)
				((ResolvableDeserializer)src).resolve(ctxt);
	}
}
/*
	private static final StdDeserializer<Instance> defaultInstanceDeserializer = ...;

	private static final StdDeserializer<Instance> instanceDeserializer = new StdDelegatingDeserializer<Instance>() {
		@Override
		public Instance deserialize(JsonParser p, DeserializationContext ctxt) throws IOException, JsonProcessingException {
			return null;
		}
	};

*/
	private SimpleModule jacksonModule;
	private ObjectMapper jsonMapper;
	private ObjectMapper smileMapper;

	private JacksonUtil() {}

	/**
	 * load an extracted module from file
	 * @param path the file the module is stored in
	 * @param binary whether or not the file is a "smile" binary file. If false it is assumed to be a JSON file
	 * @return the pre-extracted module
	 * @throws IOException on broken IO
	 */
	public static Module loadModule(@NotNull Path path, boolean binary) throws IOException {
		Tracer.enterLoop("Load modules");
		return Tracer.exit((binary?getSmileMapper() : getJsonMapper()).readValue(Files.newInputStream(path), Module.class));
	}

	/**
	 * write an extracted module to file
	 * @param module the module to save
	 * @param path the file the module will be stored in
	 * @param binary whether or not the file is supposed to be a "smile" binary file. If false a JSON file is created.
	 * @throws IOException on broken IO
	 */
	public static void writeModule(@NotNull Module module, @NotNull Path path, boolean binary) throws IOException {
		Tracer.enterLoop("Write modules");
		(binary?getSmileMapper().writer() : getJsonMapper().writerWithDefaultPrettyPrinter()).writeValue(Files.newOutputStream(path), module);
		Tracer.exit();
	}

	public static void printModule(@NotNull Module module) throws IOException {
		getJsonMapper().writerWithDefaultPrettyPrinter().writeValue(System.out, module);
	}


	/**
	 * load a set of module descriptions from a file
	 * @param path a path to the file to read
	 * @param binary whether or not the file is a "smile" binary file. If false it is assumed to be a JSON file
	 * @return the deserialized module description
	 * @throws IOException on broken IO
	 */
	public static ModuleDescriptionVariant loadModuleDescription(@NotNull Path path, boolean binary) throws IOException {
		Tracer.enterLoop("Load module description");
		//logger.info("Loading module descriptions from {}", path);
		return Tracer.exit((binary?getSmileMapper() : getJsonMapper()).readValue(Files.newInputStream(path), ModuleDescriptionVariant.class));
	}

	/**
	 * write a set of module descriptions to a file
	 * @param mdv a module description variant
	 * @param path a path to the file to create
	 * @param binary whether or not the file is supposed to be a "smile" binary file. If false a JSON file is created.
	 * @throws IOException on broken IO
	 */
	public static void writeModuleDescription(@NotNull ModuleDescriptionVariant mdv, @NotNull Path path, boolean binary) throws IOException {
		Tracer.enterLoop("Write module descriptions");
		(binary?getSmileMapper().writer() : getJsonMapper().writerWithDefaultPrettyPrinter()).writeValue(Files.newOutputStream(path), mdv);
		Tracer.exit();
	}



	@NotNull
	private static ObjectMapper getJsonMapper() {
		if (Objects.isNull(INSTANCE.jsonMapper)) {
			//Tracer.enter("Construct JSON Mapper");
			ObjectMapper jsonMapper = new ObjectMapper();
			jsonMapper.registerModule(getJacksonModule());
			//TODO register AfterburnerModule
			INSTANCE.jsonMapper = jsonMapper;
			//Tracer.exit();
		}
		return INSTANCE.jsonMapper;
	}

	@NotNull
	private static ObjectMapper getSmileMapper() {
		if (Objects.isNull(INSTANCE.smileMapper)) {
			//Tracer.enter("Construct smile Mapper");
			ObjectMapper smileMapper = new ObjectMapper(new SmileFactory().enable(SmileGenerator.Feature.CHECK_SHARED_STRING_VALUES));
			smileMapper.registerModule(getJacksonModule());
			smileMapper.registerModule(new AfterburnerModule());
			INSTANCE.smileMapper = smileMapper;
			//Tracer.exit();
		}
		return INSTANCE.smileMapper;
	}

	@NotNull
	private static SimpleModule getJacksonModule() {
		if (null==INSTANCE.jacksonModule) {
			SimpleModule jacksonModule = new SimpleModule();
			jacksonModule.addSerializer(deviceSerializer);
			jacksonModule.addSerializer(tileSerializer);
			jacksonModule.addSerializer(siteSerializer);
			jacksonModule.addDeserializer(Device.class, deviceDeserializer);
			jacksonModule.addDeserializer(Tile.class, tileDeserializer);
			jacksonModule.addDeserializer(PrimitiveSite.class, siteDeserializer);
			//jacksonModule.addDeserializer(Instance.class, instanceDeserializer);
			jacksonModule.setDeserializerModifier(new BeanDeserializerModifier()
			{
				@Override public JsonDeserializer<?> modifyDeserializer(DeserializationConfig config, BeanDescription beanDesc, JsonDeserializer<?> deserializer)
				{
					if (JacksonDeserListener.class.isAssignableFrom(beanDesc.getBeanClass()))
						return new NotifyingDeserializer((JsonDeserializer<? extends JacksonDeserListener>) deserializer);
					return deserializer;
				}
			});
			INSTANCE.jacksonModule = jacksonModule;
		}
		return INSTANCE.jacksonModule;
	}
}
