/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.ModuleInstance;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.PIP;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.Tile;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;


/**
 */
public final class ModuleDescriptionFactory {

	private static final Logger logger = LoggerFactory.getLogger(ModuleDescriptionFactory.class);

	private ModuleDescriptionFactory() {

	}

	/**
	 * Create a new {@link ModuleDescriptionVariant} for the given module
	 *
	 * @param module the module to describe
	 * @param device the device this description is for
	 * @return the new ModuleDescriptionVariant or null if the module is not placeable on the target device
	 */
	@Nullable
	public static ModuleDescriptionVariant getModuleDescription(@NotNull Module module, @NotNull Device device) {
		int[] possAnchors = calcPossibleAnchorsForDevice(module, device);
		if (0 == possAnchors.length) {
			return null;
		}

		final int defaultAnchor;
		if (module.getDevice().getPartName().equals(device.getPartName())) {
			defaultAnchor = module.getAnchor().getPrimitiveSite().getTile().getUniqueAddress();
		} else {
			defaultAnchor = possAnchors[0];
		}
		Tile defaultAnchorTile = device.getTile(defaultAnchor);


		return new ModuleDescriptionVariant(
				module.getName(),
				device,
				possAnchors,
				defaultAnchor,
				getTileNumbers(module, defaultAnchorTile, true, false),
				getTileNumbers(module, defaultAnchorTile, false, true),
				getBoundingBox(module, false),
				getBoundingBox(module, true),
				getPortMap(module, defaultAnchorTile),
				module
		);
	}

	@NotNull
	private static int[] calcPossibleAnchorsForDevice(@NotNull Module module, @NotNull Device device) {
		@NotNull Design tempDesign = new Design("temp", device.getPartName());
		@NotNull ModuleInstance mi = tempDesign.createModuleInstance("temp", module);

		int centerX = tempDesign.getDevice().getColumns() / 2;
		int centerY = tempDesign.getDevice().getRows() / 2;

		return mi.getAllValidPlacements((a, b) -> {
		}).stream()
				.sorted(Comparator.comparing((PrimitiveSite primitiveSite) -> {
					int diffX = centerX - primitiveSite.getInstanceX();
					int diffY = centerY - primitiveSite.getInstanceY();
					return -diffX * diffX - diffY * diffY;
				}))
				.mapToInt(site -> site.getTile().getUniqueAddress())
				.toArray();
	}

	@NotNull
	private static Stream<Tile> streamPlacedTiles(@NotNull Module module) {
		return module
				.getInstances().stream()
				.filter(Instance::isPlaced)
				.map(Instance::getTile)
				.distinct();
	}

	@NotNull
	private static Stream<Tile> streamRoutingTiles(@NotNull Module module) {
		@NotNull Collection<Integer> coreTiles = streamPlacedTiles(module)
				.map(Tile::getUniqueAddress).collect(Collectors.toSet());
		return module.getNets().stream()
				.flatMap(n -> n.getPIPs().stream())
				.map(PIP::getTile)
				.filter(tile -> !coreTiles.contains(tile.getUniqueAddress()))
				.distinct();
	}

	@NotNull
	private static Stream<Tile> streamTiles(@NotNull Module module, boolean includeCoreTiles, boolean includeRoutingTiles) {
		return Stream.concat(
				includeCoreTiles? streamPlacedTiles(module) : Stream.empty(),
				includeRoutingTiles? streamRoutingTiles(module) : Stream.empty()
		);
	}

	@NotNull
	private static int[] getTileNumbers(@NotNull Module module, Tile newAnchor, boolean includeCoreTiles, boolean includeRoutingTiles) {
		//final int offset = module.getAnchor().getTile().getUniqueAddress();
		// anchor.getDevice()
		return streamTiles(module, includeCoreTiles, includeRoutingTiles)
				.map(tile -> getActualTile(module, tile, newAnchor))
				.mapToInt(Tile::getUniqueAddress)
				//.map(i -> i-offset)
				.sorted()
				.toArray();
	}

	//TODO passthrough
	@NotNull
	private static Map<String, ModuleDescriptionVariantPort> getPortMap(@NotNull Module module, Tile newAnchor) {
		@NotNull HashMap<String, ModuleDescriptionVariantPort> portMap = new HashMap<>();
		module.getPorts().forEach(p -> {
			try {
				Stream<Tile> source = p.getPins().stream()
						.map(Pin::getNet)
						.filter(Objects::nonNull)
						.map(Net::getSourceTile);

				Stream<Tile> all = Stream.concat(
						p.getPins().stream(),
						p.getPins().stream()
								.map(Pin::getNet)
								.filter(Objects::nonNull)
								.map(Net::getPins)
								.flatMap(Collection::stream)
				)
						.map(Pin::getTile);

				@NotNull ModuleDescriptionVariantPort port = new ModuleDescriptionVariantPort(
						p.getName(),
						Stream.concat(
								source,
								all
						)

								.filter(Objects::nonNull)
								.map(tile -> getActualTile(module, tile, newAnchor))
								.mapToInt(Tile::getUniqueAddress)
								.distinct()
								.toArray());
				portMap.put(p.getName(), port);
				p.getAliases().forEach(a -> portMap.put(a, port));
			} catch (IllegalStateException e) {
				throw new IllegalStateException("Failed for "+p, e);
			}
		});

		//fill passthroughs
		module.getPorts().forEach(p -> {
			if (Objects.nonNull(p.getPassthroughFrom())) {
				ModuleDescriptionVariantPort ptp = portMap.get(p.getPassthroughFrom().getName());
				portMap.get(p.getName()).setPassthroughFrom(ptp);
			}
		});
		return Collections.unmodifiableMap(portMap);
	}

	@NotNull
	private static BoundingBox getBoundingBox(@NotNull Module module, boolean includeRoutingTiles) {
		return streamTiles(module, true, includeRoutingTiles)
				.collect(BoundingBox.getTileBoundingBox())
				.shift(0 - module.getAnchor().getTile().getColumn(), 0 - module.getAnchor().getTile().getRow());//shift to origin
	}

	@NotNull
	private static Tile getActualTile(Module module, Tile template, Tile newAnchor) {
		Tile correspondingTile = module.getCorrespondingTile(template, newAnchor, newAnchor.getDevice());
		if (correspondingTile == null) {
			boolean containedInInstances = module.getInstances().stream().map(Instance::getPrimitiveSite).map(PrimitiveSite::getTile).anyMatch(t -> t.getUniqueAddress() == template.getUniqueAddress());
			boolean containedInRouting = module.getNets().stream().flatMap(net -> net.getPIPs().stream().peek(pip -> {

				if (pip.getTile().getName().equals("CLEXM_X10Y30")) {
					logger.info("there it is in "+net.getName());
				}
			})).map(PIP::getTile)

					.anyMatch(t -> t.getUniqueAddress() == template.getUniqueAddress());
			if (containedInInstances || containedInRouting) {
				throw new IllegalStateException("Did not find mapping for tile " + template.getDevice().getPartName() + "." + template.getName() + " for anchor " + newAnchor.getDevice().getPartName() + "." + newAnchor.getName() + " but why did ModuleInstance.place not sort this out?");
			} else {
				throw new IllegalStateException("Tried to find non existent mapping for tile " + template + " that is not part of the module!");
			}

		}
		return correspondingTile;

	}
}
