/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module;

import de.tu_darmstadt.rs.rapidSoc.RapidSocProject;
import de.tu_darmstadt.rs.rapidSoc.directories.HashPath;
import de.tu_darmstadt.rs.rapidSoc.directories.ModuleDirectories;
import de.tu_darmstadt.rs.rapidSoc.directories.VariantPath;
import de.tu_darmstadt.rs.rapidSoc.module_synthesis.ModuleState;
import de.tu_darmstadt.rs.rapidSoc.module_synthesis.ModuleSynthesis;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

public class ModuleDescriptionLoader {

	private static final Logger logger = LoggerFactory.getLogger(ModuleDescriptionLoader.class);


	final Map<HashPath, ModuleDescription> moduleCache;
	private final RapidSocProject project;

	public ModuleDescriptionLoader(RapidSocProject project) {
		this.project = project;
		moduleCache = new HashMap<>();
	}

	public ModuleDescription loadAllVariants(Collection<IModuleDescription> descriptions, Map<String, String> attrMap) throws IOException {
		return moduleCache.computeIfAbsent(project.getModuleRootPath().resolveHashDir(descriptions, attrMap), hashDir -> {
			logger.info("Loading from {}", hashDir);
			logger.info("module dirs: " + hashDir.getModuleDirectories().collect(Collectors.toList()));
			final StreamUtil.Pair<Boolean, Set<ModuleDescriptionVariant>> variantPair = hashDir.getModuleDirectories()
					.map(variantDir -> loadModuleVariant(variantDir, descriptions, attrMap))
					.peek(v -> logger.info("loaded variant: " + v))
					.collect(StreamUtil.multiCollector(
							Collectors.mapping(Optional::isPresent, Collectors.reducing(true, Boolean::logicalAnd)),
							StreamUtil.toSetIfPresent()
					));


			final Boolean hasAllAvailable = variantPair.getA();
			Set<ModuleDescriptionVariant> variants = variantPair.getB();
			if (!hasAllAvailable) {
				logger.error("Module {} has one or more variants that are unavailable. Cannot continue", hashDir.getPath());
				return null;
			}

			if (variants.isEmpty()) {
				Optional<ModuleDescriptionVariant> newVariant = createNewVariant(descriptions, attrMap, hashDir);
				if (!newVariant.isPresent()) {
					logger.error("created initial variant for {}, but not available yet", hashDir.getPath());
					return null;
				}
				variants = Collections.singleton(newVariant.get());
			}

			return new ModuleDescription(ModuleDirectories.getModuleDescriptionName(descriptions), variants, hashDir.getPath().getFileName().toString());
		});
	}


	public ModuleDescription loadAllVariants(Collection<IModuleConfiguration<?>> jModules) {
		try {
			Map<String, String> attrMap = ModuleDirectories.getAttributeMap(jModules);
			Collection<IModuleDescription> descriptions = ModuleDirectories.getModuleDescriptions(jModules);
			return loadAllVariants(descriptions, attrMap);
/*

			Path hashPath = moduleDirectories.getHashDir(jModules);
			Path var0Path = hashPath.resolve("0");
			if (!Files.isDirectory(var0Path)) {
				Files.createDirectories(var0Path);
			}*/
			//
			//
			//			List<ModuleDescriptionVariant> variants = new ArrayList<>();
			//			List<ModulePlacement> placements = new ArrayList<>();
			//
			//
			//
			//			moduleDirectories.getModuleDirectories(jModules)
			//					.map(path ->
			//
			//					)
			//					.allMatch(path -> {
			//						logger.info("Processing module {} located at {}", jModules, path);
			//						boolean avail = synthesizeModule(
			//								ModuleDirectories.getModuleDescriptions(jModules),
			//								ModuleDirectories.getAttributeMap(jModules),
			//								path.getFileName().toString()
			//						);
			//						if (!avail) {
			//							logger.error("Failed to synthesize module {} located in {}", jModules, path);
			//							return false;
			//						}
			//						ModuleDescriptionVariant md = moduleDescriptionMap.computeIfAbsent(
			//								path,
			//								modPath -> {
			//									try {
			//										return JacksonUtil.loadModuleDescriptionFile(path.resolve("build").resolve("spartanmc.dscr"), true)
			//												.getModuleDescription(design.getDevice().getPartName());
			//									} catch (IOException e) {
			//										throw new UncheckedIOException(e);
			//									}
			//								}
			//						);
			//						if (Objects.isNull(md)) {
			//							logger.info("Module {} is not placeable on target device {}", jModules, design.getDevice().getPartName());
			//							return false;
			//						}
			//						md.setModulePath(path.resolve("build").resolve("spartanmc.mod"));
			//						md.setVariantList(variants);
			//						variants.add(md);
			//						Arrays.stream(md.getPlacements())
			//								.mapToObj(i -> new ModulePlacement(md, md.getDevice().getTile(i)))
			//								.forEach(placements::add);
			//						md.setAllPlacements(placements);
			//
			//						return true;
			//					});
			//			if (!allVarsLoaded) {
			//				return false;
			//			}

			//			ModuleDescriptionVariant md = Objects.requireNonNull(moduleDescriptionMap.get(hashPath.resolve("0")));
			//			ModuleDescriptionInstance mdi = new ModuleDescriptionInstance(md, name);
			//			jModules.forEach(jModule -> result.put(jModule, mdi));
		} catch (IOException e) {
			logger.error("Failed to load module {}", jModules, e);
			//			return false;
			return null;
		}
		//		return true;
	}

	private Optional<ModuleDescriptionVariant> createNewVariant(Collection<IModuleDescription> descriptions, Map<String, String> attrMap, HashPath hashDir) {
		try {
			final VariantPath variantPath = hashDir.getNewVariantDirectory();

			return loadModuleVariant(variantPath, descriptions, attrMap);

		} catch (IOException e) {
			logger.error(e.getMessage());
			return Optional.empty();
		}
	}


	private boolean synthesizeModule(Collection<IModuleDescription> md, Map<String, String> attrMap, String variant) {
		ModuleSynthesis synthesis = new ModuleSynthesis(md, attrMap, variant, project);
		final ModuleState ms = synthesis.tryMakeAvailable();

		//If it's still not synthesized, we can't help it
		if (ms != ModuleState.AVAILABLE) {
			String areaConstraints = ms.hasConstraints? "" : " create area constraints";
			logger.error("module " + ModuleDirectories.getModuleDescriptionName(md) + " is " + ms + "! Go to "
					+ synthesis.getProjectPath() + areaConstraints + " and run \"make " + (ms.hasConstraints? "" : "jconfig_update ") + "placeroute xdl\" before continuing!");
			return false;
		} else {
			return true;
		}
	}

	private Optional<ModuleDescriptionVariant> loadModuleVariant(VariantPath variantDir, Collection<IModuleDescription> descriptions, Map<String, String> attrMap) {
		logger.info("Processing module {} located at {}", descriptions, variantDir);
		boolean avail = synthesizeModule(
				descriptions,
				attrMap,
				variantDir.getVariantName()
		);
		if (!avail) {
			logger.error("Failed to synthesize module {} located in {}", descriptions, variantDir);
			return Optional.empty();
		}

		try {
			final ModuleDescriptionVariant md = JacksonUtil.loadModuleDescription(variantDir.resolveSmileDescription(project.getDevice()), true);

			if (Objects.isNull(md)) {
				logger.info("Module {} is not placeable on target device {}", descriptions, project.getPartName());
				return Optional.empty();
			}
			md.setModulePath(variantDir);
			return Optional.of(md);
		} catch (IOException e) {
			logger.error("Failed to load module: " + e.getMessage());
			return Optional.empty();
		}
	}

}
