/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;
import de.tu_darmstadt.rs.rapidSoc.directories.VariantPath;

import com.fasterxml.jackson.annotation.JsonAutoDetect;
import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.databind.annotation.JsonSerialize;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.jackson.MapToSetConverter;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.Tile;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.IntStream;
import java.util.stream.Stream;


//TODO adjust json names


/**
 * a very reduced summary of the module used for choosing and placing the module
 */
@JsonAutoDetect(fieldVisibility = JsonAutoDetect.Visibility.NONE, getterVisibility = JsonAutoDetect.Visibility.NONE, isGetterVisibility = JsonAutoDetect.Visibility.NONE, setterVisibility = JsonAutoDetect.Visibility.NONE)
public class ModuleDescriptionVariant {

	private static final String PROPERTY_NAME = "name";
	private static final String PROPERTY_DEVICE = "device";
	private static final String PROPERTY_PLACEMENTS = "placements";
	private static final String PROPERTY_ANCHOR = "anchor";
	private static final String PROPERTY_TILES_CORE = "tiles-core";
	private static final String PROPERTY_TILES_ROUTING = "tiles-routing";
	private static final String PROPERTY_BB_CORE = "bb-core";
	private static final String PROPERTY_BB_FULL = "bb-full";
	private static final String PROPERTY_PORTS = "ports";

	@NotNull
	private static final Logger logger = LoggerFactory.getLogger(ModuleDescriptionVariant.class);

	@NotNull
	@JsonProperty(PROPERTY_NAME)
	private final String name;

	/**target device*/
	@NotNull
	@JsonProperty(PROPERTY_DEVICE)
	private final Device device;

	private VariantPath modulePath;
	/**Full module. null if not loaded*/
	@Nullable
	private transient Module module;

	/**Tiles (not sites) the anchor can be placed in*/
	@NotNull
	@JsonProperty(PROPERTY_PLACEMENTS)
	private final int[] placements;

	/**The original anchor tile*/
	@JsonProperty(PROPERTY_ANCHOR)
	private final int anchor;

	/**Tiles used by instances of the module relative to the anchor 0/0*/
	@NotNull
	@JsonProperty(PROPERTY_TILES_CORE)
	private final int[] coreTiles;

	/**Tiles used by routing only relative to the anchor 0/0*/
	@NotNull
	@JsonProperty(PROPERTY_TILES_ROUTING)
	private final int[] routingTiles;

	/**Bounding box around tiles containing instances of this module*/
	@NotNull
	@JsonProperty(PROPERTY_BB_CORE)
	private final BoundingBox coreBb;

	/**Bounding box around all tiles related to this module*/
	@NotNull
	@JsonProperty(PROPERTY_BB_FULL)
	private final BoundingBox routingBb;
	/**Maps port names (and their aliases) to the actual ports*/

	@NotNull
	@JsonProperty(PROPERTY_PORTS)
	@JsonSerialize(converter = MapToSetConverter.class)
	@JsonDeserialize(converter = ModuleDescriptionVariantPort.ModuleDescriptionVariantPortSetToMapConverter.class)
	private final Map<String, ModuleDescriptionVariantPort> portMap;

	private final Map<Integer, int[]> coreTileCache = new HashMap<>();
	private final Map<Integer, int[]> routingTileCache = new HashMap<>();

	public ModuleDescriptionVariant(
			@NotNull final String name,
			@NotNull final Device device,
			@NotNull final int[] placements,
			final int anchor,
			@NotNull final int[] coreTiles,
			@NotNull final int[] routingTiles,
			@NotNull final BoundingBox coreBb,
			@NotNull final BoundingBox fullBb,
			@NotNull final Map<String, ModuleDescriptionVariantPort> portMap,
			@Nullable final Module module
	) {
		this.name = name;
		this.device = device;
		this.module = null;
		this.placements = placements;
		this.anchor = anchor;
		this.coreTiles = coreTiles;
		this.routingTiles = routingTiles;
		this.coreBb = coreBb;
		this.routingBb = fullBb;
		this.portMap = portMap;
		this.module = module;
		this.portMap.values().forEach(p -> p.setModuleDescriptionVariant(this));
	}

	/**
	 * The constructor intended to be used on deserialization
	 * All tiles are referred to by their unique address
	 * @param device the device this description is for
	 * @param placements the tiles the anchor can be placed in
	 * @param anchor the original anchor tile
	 * @param coreTiles the tiles of the core module ie. the tiles containing instances of this module
	 * @param routingTiles the tiles used by routing only
	 * @param coreBb a bounding box around the core tiles shifted to 0/0
	 * @param fullBb a bounding box around all tiles shifted to 0/0
	 * @param portMap a map from all port names to all ports
	 */
	@JsonCreator
	public ModuleDescriptionVariant(
			@JsonProperty(PROPERTY_NAME) @NotNull final String name,
			@JsonProperty(PROPERTY_DEVICE) @NotNull final Device device,
			@JsonProperty(PROPERTY_PLACEMENTS) @NotNull final int[] placements,
			@JsonProperty(PROPERTY_ANCHOR) final int anchor,
			@JsonProperty(PROPERTY_TILES_CORE) @NotNull final int[] coreTiles,
			@JsonProperty(PROPERTY_TILES_ROUTING) @NotNull final int[] routingTiles,
			@JsonProperty(PROPERTY_BB_CORE) @NotNull final BoundingBox coreBb,
			@JsonProperty(PROPERTY_BB_FULL) @NotNull final BoundingBox fullBb,
			@JsonProperty(PROPERTY_PORTS) @NotNull final Map<String, ModuleDescriptionVariantPort> portMap
	) {
		this(
				name,
				device,
				placements,
				anchor,
				coreTiles,
				routingTiles,
				coreBb,
				fullBb,
				portMap,
				null
		);
	}


	public void setModulePath(@NotNull VariantPath modulePath) {
		this.modulePath = modulePath;
	}

	@NotNull
	public int[] getPlacements() {
		return placements;
	}

	/**
	 * Get the full module this description belongs to. Loads the module if necessary
	 * @return the rapidsmith Module
	 */
	@NotNull
	public Module getModule() {
		if (Objects.isNull(module)) {
			return loadModule();
		}
		return module;
	}

	/**
	 * load module from disk using {@link JacksonUtil}
	 * @return the rapidsmith module
	 * @throws IllegalStateException on broken io
	 */
	@NotNull
	private Module loadModule() {
		//TODO test
		Objects.requireNonNull(modulePath, "can't load module; module path is null");
		try {
			module = JacksonUtil.loadModule(modulePath.resolveSmileModule(),true);
			return module;
		} catch (IOException e) {
			logger.error("failed to load module",e);
			throw new IllegalStateException("unable to load module",e);
		}
	}

	/**
	 * set the module. does not check anything
	 * @param module the module
	 * @return this
	 */
	@NotNull
	public ModuleDescriptionVariant setModule(@NotNull final Module module) {
		this.module = module;
		return this;
	}

	@NotNull
	public Device getDevice() {
		return device;
	}

	@NotNull
	public Tile getAnchor() {
		return device.getTile(anchor);
	}

	/**
	 * stream the unique numbers of the tiles used by the module relative to {@link ModuleDescriptionVariant#getAnchor}
	 * @param includeCoreTiles whether or not to include tiles used by instances
	 * @param includeRoutingTiles whether or not to include tiles used for routing only
	 * @return the stream
	 */
	@NotNull
	public IntStream streamTileNumbers(final boolean includeCoreTiles, final boolean includeRoutingTiles) {
		return IntStream.concat(
				includeCoreTiles?Arrays.stream(coreTiles) : IntStream.empty(),
				includeRoutingTiles?Arrays.stream(routingTiles) : IntStream.empty()
		);
	}

	@NotNull
	public Stream<Tile> streamTiles(final boolean includeCoreTiles, final boolean includeRoutingTiles) {
		return streamTileNumbers(includeCoreTiles, includeRoutingTiles)
				.mapToObj(device::getTile);
	}

	@NotNull
	public Stream<Tile> streamTilesForAnchor(Tile anchor, final boolean includeCoreTiles, final boolean includeRoutingTiles) {
		return streamTileNumbersForAnchor(anchor.getUniqueAddress(), includeCoreTiles, includeRoutingTiles)
				.mapToObj(device::getTile);
	}

	private int getCorrespondingTileNumber(final int tile, final int anchor) {
		Tile anchorTile = Objects.requireNonNull(device.getTile(anchor));
		return Module.getCorrespondingTile(
			Objects.requireNonNull(device.getTile(tile)),
			anchorTile.getTileXCoordinate() - getAnchor().getTileXCoordinate(),
			anchorTile.getTileYCoordinate() - getAnchor().getTileYCoordinate(),
			device).getUniqueAddress();
	}

	public IntStream streamTileNumbersForAnchor(final int anchor, final boolean includeCoreTiles, final boolean includeRoutingTiles) {
		return IntStream.concat(
				includeCoreTiles ?
						Arrays.stream(coreTileCache.computeIfAbsent(
							anchor,
							(key) -> streamTileNumbers(true, false).map(i -> getCorrespondingTileNumber(i, key)).toArray())
						) : IntStream.empty(),
				includeRoutingTiles ?
						Arrays.stream(routingTileCache.computeIfAbsent(
								anchor,
								(key) -> streamTileNumbers(false, true).map(i -> getCorrespondingTileNumber(i, key)).toArray())
						) : IntStream.empty()
		);
	}

	/**
	 * Returns a bounding box around 0,0
	 * @param includeRoutingTiles - weather or not routing tiles should be enclosed
	 * @return an instance of {@link BoundingBox}
	 */
	//TODO BUG: tiles can't be properly moved by row/column, so shifting tile row/column based bbs isn't accurate either
	@NotNull
	@Deprecated
	public BoundingBox getBoundingBox(final boolean includeRoutingTiles) {
		return includeRoutingTiles ? routingBb : coreBb;
	}

	//TODO the modules routing resources used by exact routing overlap detector. that is a lot of data without much benefit, so just load the module in case those are needed?

	@NotNull
	public Set<String> getPortNames() {
		return portMap.keySet();
	}


	public ModuleDescriptionVariantPort getPort(String name) {
		return portMap.get(name);
	}

	@NotNull
	public String toString() {
		return name;
	}

	@NotNull
	public String getStats() {
		return
				"device: " +
				device.getPartName() +
				System.lineSeparator() +
				"placements (" +
				placements.length +
				"): " +
				Arrays.toString(placements) +
				System.lineSeparator() +
				"core tiles (" +
				coreTiles.length +
				"): " +
				Arrays.toString(coreTiles) +
				System.lineSeparator() +
				"routing tiles (" +
				routingTiles.length +
				"): " +
				Arrays.toString(routingTiles) +
				System.lineSeparator() +
				"core bb: " +
				coreBb +
				System.lineSeparator() +
				"full bb: " +
				routingBb +
				System.lineSeparator() +
				"ports: " +
				portMap +
				System.lineSeparator();
	}

	public Stream<Tile> streamPlacements() {
		return IntStream.of(placements)
				.mapToObj(device::getTile);
	}

	public Stream<ModulePlacement> streamModulePlacements() {
		return streamPlacements()
				.map(tile -> new ModulePlacement(this, tile));
	}

	public Stream<ModuleDescriptionVariantPort> streamPorts() {
		return portMap.values().stream();
	}
}
