/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module;

import de.tu_darmstadt.rs.rapidSoc.RapidSocProject;
import de.tu_darmstadt.rs.rapidSoc.directories.HashPath;
import de.tu_darmstadt.rs.rapidSoc.directories.ModuleDirectories;
import de.tu_darmstadt.rs.rapidSoc.directories.ModuleRootPath;
import de.tu_darmstadt.rs.rapidSoc.directories.VariantPath;
import de.tu_darmstadt.rs.rapidSoc.module_synthesis.JacksonModuleWriter;
import de.tu_darmstadt.rs.rapidSoc.module_synthesis.ModuleSynthesis;
import de.tu_darmstadt.rs.rapidSoc.placer.ModulePlacer;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.targets.TargetDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IApplicationConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigBuilder;
import de.tu_darmstadt.rs.util.table.Column;
import de.tu_darmstadt.rs.util.table.TableFormat;
import de.tu_darmstadt.rs.util.table.TableFormatter;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;
import com.beust.jcommander.Parameters;
import com.beust.jcommander.ParametersDelegate;
import com.beust.jcommander.converters.PathConverter;
import edu.byu.ece.rapidSmith.design.Attribute;
import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.design.Port;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.IntSummaryStatistics;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

/**
 * This command extracts a module and serializes it as a smile or json file. It guesses module name, hash and variant based on the current working directory
 */
@Parameters(commandNames = {"mod", "module"}, commandDescription = "Converts a module from xdl to our own format")
public class ModuleExtractionCommand implements Runnable {

	private static final Logger logger = LoggerFactory.getLogger(ModuleExtractionCommand.class);

	@Parameter(names = "--possible-placements-dir", description = "directory to store images of possible placements on all supported FPGAs", converter = PathConverter.class)
	private Path possiblePlacementsDir;

	@Parameter(names = "--verify-outputs")
	private boolean verifyOutputs = true;

	@ParametersDelegate
	private LibJConfigBuilder projectBuilder = new LibJConfigBuilder("1", true);

	public ModuleExtractionCommand() {
	}

	@Override
	public void run() {
		try {
			Tracer.start();

			Tracer.enter("Load library");
			LibJConfigProject project = projectBuilder.build();
			final IApplicationConfiguration application = project.getApplication();
			Library library = application.getLibrary();
			Tracer.exit();


			ModuleRootPath rootPath = ModuleDirectories.getRootDir();
			final VariantPath varPath = rootPath.getAsVariantPath(project.getProjectDir());
			final HashPath hashPath = varPath.getHashPath();
			final String moduleName = hashPath.getModulePath().getModuleName();
			logger.info("Using module \"" + moduleName + ":" + hashPath.getHash() + ":" + varPath.getVariantName() + "\"");

			Map<String, String> attributeMap = hashPath.loadAttributeMap();

			logger.info("Loading modules {}", moduleName);

			List<IModuleDescription> moduleDescriptions = Arrays.stream(moduleName.split("\\+")).map(name -> {
				final IModuleDescription d = library.getModuleDescription(name);
				if (d == null) {
					throw new ParameterException("Module \"" + name + "\" not found.");
				}
				return d;
			}).collect(Collectors.toList());

			TargetDescription target = library.getTargetDescription("SP601");


			RapidSocProject rapidProject = new RapidSocProject(false, false, project, ModuleDirectories.getRootDir(), target);

			ModuleSynthesis modSyn = new ModuleSynthesis(moduleDescriptions, attributeMap, varPath.getPath().getFileName().toString(), rapidProject);

			Module module = modSyn.extractModule(false, new HashSet<>(), moduleDescriptions.stream().anyMatch(x -> x.getType() == IModuleDescription.ModuleType.Subsystem));


			JacksonModuleWriter.writeAll(module, varPath, possiblePlacementsDir);

			if (verifyOutputs) {
				final Module deser = JacksonUtil.loadModule(varPath.resolveSmileModule(), true);
				Objects.requireNonNull(deser);
				verifyDeser(deser);
				verifyPlacements(deser, varPath);
			}

			System.out.flush();
			Tracer.printStats(System.out);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private void verifyPlacements(Module module, VariantPath variantPath) throws IOException {
		TableFormatter table = new TableFormatter(TableFormat.CHARACTERS, Column.STRING,
				new Column("%d", Column.Alignment.RIGHT),
				new Column("%d", Column.Alignment.RIGHT),
				new Column("%.2f%%", Column.Alignment.RIGHT)
		);
		table.addHeaderRow("Device", "Failing Placements", "Total Placements", "Failing Ratio");
		final IntSummaryStatistics statsGlobal = variantPath.streamAvailableDevices()
				.map(device -> {
					try {
						ModuleDescriptionVariant variant = JacksonUtil.loadModuleDescription(variantPath.resolveSmileDescription(device), true);
						variant.setModulePath(variantPath);
						variant.setModule(module);


						logger.info("trying " + variant.getDevice().getPartName());
						final IntSummaryStatistics stats = variant.streamPlacements().mapToInt(placement -> {
							Design design = new Design("verification", variant.getDevice().getPartName());
							ModuleDescription md = new ModuleDescription("verification", Collections.singleton(variant), null);

							ModuleDescriptionInstance mdi = new ModuleDescriptionInstance(md, "verification", new ModulePlacement(variant, placement));
							try {
								ModulePlacer.instantiateModules(design, Collections.singleton(mdi));
							} catch (RuntimeException e) {
								logger.error(e.toString());
								return 1;
							}
							return 0;
						}).summaryStatistics();
						addStatsToTable(table, stats, variant.getDevice().getPartName());
						return stats;
					} catch (IOException e) {
						throw new RuntimeException(e);
					}
				}).reduce((l, r) -> {
					l.combine(r);
					return l;
				}).orElse(new IntSummaryStatistics());
		boolean ok = statsGlobal.getSum() == 0;
		if (!ok) {
			addStatsToTable(table, statsGlobal, "Total");
			table.output().forEach(logger::warn);
			throw new IllegalStateException("There are invalid placements");
		}
	}

	private void addStatsToTable(TableFormatter table, IntSummaryStatistics stats, String rowHeader) {
		table.addContentRow(rowHeader, stats.getSum(), stats.getCount(), 100f * stats.getSum() / stats.getCount());
	}

	private void verifyDeser(Module module) {
		boolean fail = false;

		int pinCount = 0;
		for (Instance instance : module.getInstances()) {
			for (Pin pin : instance.getPins()) {
				if (pin.getInstance() != instance) {
					logger.error("In {}.{}, inst is set to {}", instance.getName(), pin.getName(), pin.getInstance() != null? pin.getInstance().getName() : "null");
					fail = true;
				}
				pinCount++;
			}
		}

		logger.info("saw {} pins", pinCount);

		for (Net net : module.getNets()) {
			for (Pin pin : net.getPins()) {
				if (pin.getNet() != net) {
					logger.error("In {}.{}, net is set to {}", net, pin, pin.getNet());
					fail = true;
				}

				if (pin.getInstance() != null) {
					if (!pin.getInstance().getNetList().contains(net)) {
						logger.error("Net {} is not contained in {}'s netList", net, pin.getInstance().getName());
						fail = true;
					}
				}

				if (pin.getPort() != null) {
					if (!pin.getPort().getPins().contains(pin)) {
						logger.error("Pin {} is not contained in port {}'s pin list", pin, pin.getPort().getName());
						fail = true;
					}
				}
			}
		}

		for (Port port : module.getPorts()) {
			if (module.getPort(port.getName()) != port) {
				logger.error("Port name not registered for {}", port.getName());
				fail = true;
			}

			if (port.getPins().size() != port.getPins().stream().distinct().count()) {
				logger.error("pin contained multiple times in {}", port);
				fail = true;
			}
		}

		for (Instance instance : module.getInstances()) {
			for (Attribute attribute : instance.getAttributes()) {
				if (instance.getAttribute(attribute.getPhysicalName()) != attribute) {
					logger.error("Attribute name not registered for {}", attribute.getName());
					fail = true;
				}
			}
			for (Pin pin : instance.getPins()) {
				if (instance.getPin(pin.getName()) != pin) {
					logger.error("Pin name not registered for {}", pin.getName());
					fail = true;

				}
			}
		}


		if (fail) {
			throw new RuntimeException("Module is invalid");
		}
	}
}
