/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;

import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.Tile;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.stream.IntStream;
import java.util.stream.Stream;

/**
 */
public class ModulePlacement {
	@NotNull
	private final Device device;
	@NotNull
	private final ModuleDescriptionVariant modDescr;
	@Nullable
	private final Tile anchor;

	/**
	 * Construct a new {@link ModulePlacement}
	 *
	 * @param description a description of the module
	 * @param anchor the tile the anchor is supposed to be located in
	 */
	public ModulePlacement(
			@NotNull final ModuleDescriptionVariant description,
			@Nullable final Tile anchor
	) {
		this.modDescr = description;
		this.device = description.getDevice();
		this.anchor = anchor;
	}

	public ModulePlacement(@NotNull final ModuleDescriptionVariant description) {
		this(description, null);
	}

	/**
	 * Copy constructor
	 *
	 * @param other the object to copy
	 */
	public ModulePlacement(@NotNull final ModulePlacement other) {
		this(
				other.modDescr,
				other.anchor
		);
	}

	/**
	 * Get the {@link ModuleDescriptionVariant}
	 *
	 * @return the {@link ModuleDescriptionVariant}
	 */
	@NotNull
	public ModuleDescriptionVariant getModuleDescription() {
		return modDescr;
	}

	/**
	 * Get the anchor {@link Tile} currently used. null if this is considered unplaced
	 *
	 * @return the anchor
	 */
	@Nullable
	public Tile getAnchor() {
		return this.anchor;
	}

	/**
	 * returns a bounding box around the module placed using the anchor
	 *
	 * @param includeRoutingTiles - weather or not routing tiles should be enclosed
	 * @return an instance of {@link BoundingBox}
	 */
	//TODO see comment in ModuleDescriptionVariant
	@NotNull
	public BoundingBox getBoundingBox(final boolean includeRoutingTiles) {
		return modDescr.getBoundingBox(includeRoutingTiles).shift(anchor.getColumn(), anchor.getRow());
	}

	/**
	 * create a stream containing all tiles with respect to the currently selected anchor
	 *
	 * @param includeCoreTiles whether or not to include tiles used by instances
	 * @param includeRoutingTiles whether or not to include tiles used by routing only
	 * @return the resulting stream
	 */
	@NotNull
	public Stream<Tile> streamTiles(final boolean includeCoreTiles, final boolean includeRoutingTiles) {
		return streamTileNumbers(includeCoreTiles, includeRoutingTiles).mapToObj(device::getTile);
		//return modDescr.streamTiles(includeCoreTiles, includeRoutingTiles)
		//		.map(this::getActualTile);
	}

	/**
	 * calculate what tile things originally placed at tile would be placed in with the currently selected anchor
	 *
	 * @param tile the tile to "move"
	 * @return the resulting tile
	 */
	@NotNull
	public Tile getActualTile(@NotNull Tile tile) {
		Objects.requireNonNull(anchor);
		final Tile correspondingTile = Module.getCorrespondingTile(
				tile,
				anchor.getTileXCoordinate() - modDescr.getAnchor().getTileXCoordinate(),
				anchor.getTileYCoordinate() - modDescr.getAnchor().getTileYCoordinate(),
				device);
		if (correspondingTile == null) {
			throw new IllegalStateException("Selected anchor or given tile is invalid. Did not find " + tile.getName());
		}
		return correspondingTile;
	}

	/**
	 * a stream containing the unique address of all tiles with respect to the currently selected anchor
	 *
	 * @param includeCoreTiles whether or not to include tiles used by instances
	 * @param includeRoutingTiles whether or not to include tiles used by routing only
	 * @return the resulting stream
	 */
	@NotNull
	public IntStream streamTileNumbers(final boolean includeCoreTiles, final boolean includeRoutingTiles) {
		return modDescr.streamTileNumbersForAnchor(anchor.getUniqueAddress(), includeCoreTiles, includeRoutingTiles);
		//return streamTiles(includeCoreTiles, includeRoutingTiles)
		//		.mapToInt(Tile::getUniqueAddress);
	}

	//TODO broken but something like this would be performing better
	/*@Nonnull
	public IntStream streamTileNumbers(boolean includeCoreTiles, boolean includeRoutingTiles) {
		if(null==anchor)
			throw new IllegalStateException("can't get correct tiles for an unplaced module");//TODO own exception
		int offset = anchor.getUniqueAddress();
		return modDescr.streamTileNumbers(includeCoreTiles, includeRoutingTiles)
				.map(i -> i+offset);
	}*/

	public ModuleDescriptionVariantPort getPort(@NotNull String name) {
		return modDescr.getPort(name);
	}

	public Stream<Tile> streamPortTiles(@NotNull String name) {
		return getPort(name).streamTiles().map(this::getActualTile);
	}

	/*public Stream<Tile> streamPortTiles(String name) {
		if(null == getPort(name))
			Objects.requireNonNull(getPort(name));
		if(hasAnchor())
			return getPort(name).streamTiles().map(this::getActualTile);
		else
			return Stream.empty();
	}*/

	/**
	 * create a string describing this {@link ModulePlacement}
	 * contains current module description and anchor
	 *
	 * @return the resulting string
	 */
	@NotNull
	public String toString() {
		return modDescr.toString() + " at " + anchor.getName();
	}
}
