/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module_synthesis;

import de.tu_darmstadt.rs.util.StreamUtil;

import edu.byu.ece.rapidSmith.design.Attribute;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.device.PrimitiveType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class InstanceInsideChecker {

	private static final Logger logger = LoggerFactory.getLogger(InstanceInsideChecker.class);


	public enum InclusionType {
		InModule(false, "#FFFF80"),
		OutModule(false, "#8080FF"),
		OtherOutside(false, "#FF8080"),
		IobOutside(false, "#80FF80"),
		InOutModuleIob(false, "#80FFFF"),
		Inside(true, "");

		private final boolean inside;
		private final String htmlColor;

		InclusionType(boolean inside, String htmlColor) {
			this.inside = inside;
			this.htmlColor = htmlColor;
		}

		public boolean isInside() {
			return inside;
		}

		public static Stream<InclusionType> valuesOutside() {
			return Arrays.stream(values())
					.filter(InclusionType::isInside);
		}

		public String toHtmlColor() {
			return htmlColor;
		}
	}


	private enum Exclude {
		EXCLUDE,
		NEUTRAL,
		INCLUDE;

		public static Exclude accumulator(Exclude a, Exclude b) {
			if (a == EXCLUDE || b == EXCLUDE) {
				return EXCLUDE;
			}

			if (a == INCLUDE || b == INCLUDE) {
				return INCLUDE;
			}

			return NEUTRAL;
		}
	}

	private static Exclude isExcludeAttribute(Attribute attribute) {
		//Never use LUT attributes to exclude stuff, as they may sit at the edge between exclude stuff and included
		//Flip flops
		//e.g. Enable in timer is most often a 5FF, sourced by a LUT with DI[x] as single input and reg_write as CE
		if (attribute.getPhysicalName().endsWith("LUT")) {
			if (attribute.getLogicalName().contains("_exclude") || attribute.getLogicalName().isEmpty()) {
				return Exclude.NEUTRAL;
			} else {
				return Exclude.INCLUDE;
			}
		}

		//ohne den "/" am Ende tauchen auch Instanzen auf, die nichts in der Gruppe der teilweise enthaltenen Netze
		//zu suchen hat, weil es - warum auch immer - bei den attributen der instanzen verweise zu "_exclude" gibt...
		if (attribute.getLogicalName().contains("_exclude/")) {
			return Exclude.EXCLUDE;
		}
		//ohne den "/" am Ende tauchen auch Instanzen auf, die nichts in der Gruppe der teilweise enthaltenen Netze
		//zu suchen hat, weil es - warum auch immer - bei den attributen der instanzen verweise zu "_exclude" gibt...
		if (attribute.getLogicalName().contains("_exclude$")) {
			return Exclude.EXCLUDE;
		}

		if (attribute.getLogicalName().contains("exclude_instance_from_module")) {
			return Exclude.EXCLUDE;
		}

		if (!attribute.getLogicalName().isEmpty()) {
			return Exclude.INCLUDE;
		}

		return Exclude.NEUTRAL;
	}

	public static boolean checkOnlyByType(Instance instance) {
		if (instance.getName().contains("_exclude/")) {
			return false;
		}

		final Exclude excludeByAttribute = instance.getAttributes().stream().map(InstanceInsideChecker::isExcludeAttribute).reduce(Exclude.NEUTRAL, Exclude::accumulator);
		if (excludeByAttribute == Exclude.EXCLUDE || excludeByAttribute == Exclude.NEUTRAL) {
			return false;
		}

		//Include everything except io logic
		switch (instance.getType()) {

			//A BUFG is inside if the source is inside and at least one sink is inside
			case BUFG:
				return instance.getPins().stream().allMatch(pin -> {
					if (pin != pin.getNet().getSource()) {
						return checkOnlyByType(pin.getNet().getSource().getInstance());
					} else {
						return pin.getNet().getPins().stream().filter(p -> p != pin).anyMatch(i -> checkOnlyByType(i.getInstance()));
					}
				});

			//IO stuff is never inside
			case ILOGIC:
			case ILOGIC2:
			case IBUF:
			case IOB:
			case OLOGIC:
			case OLOGIC2:
			case BUFH:
				return false;

			//the rest is always inside
			default:
				return true;
		}
	}

	public static boolean isStaticSourceInst(Instance i) {
		//TODO check for routethroughs?
		//if (i.getAttribute("_ROUTETHROUGH")!=null)
		//    return false;
		return i.getPins().stream().allMatch(pin ->
				Stream.of(i.getAttribute("_GND_SOURCE"), i.getAttribute("_VCC_SOURCE"))
						.filter(Objects::nonNull)
						.flatMap(a -> Stream.of(a.getMultiValueValues()))
						.anyMatch(s -> s.equals(pin.getName())));
	}


	public static boolean isInstanceIncluded(Instance i) {
		//For tieoffs, we include them if their connected instances are included
		//Get connected pins
		//And check if all match
		//If there was no connection, we don't want to include it
		//Split tieoffs that are partially included

		if (i.getType().equals(PrimitiveType.TIEOFF) || isStaticSourceInst(i)) {
			StreamUtil.AllOrAnyMatch include = i.getNetList().stream()
					//Get connected pins
					.flatMap(n -> n.getPins().stream().filter(p -> p != n.getSource()))
					//And check if all match
					.map(p -> checkOnlyByType(p.getInstance()))
					.collect(StreamUtil.AllOrAnyMatch.collector())
					//If there was no connection, we don't want to include it
					.orElse(StreamUtil.AllOrAnyMatch.NONE);


			//Split tieoffs that are partially included
			if (include == StreamUtil.AllOrAnyMatch.SOME) {
				/*logger.warn("Tieoff {} is partially included. Splitting its connections...", i.getName());
				for (Net net : i.getNetList()) {
					logger.info("Create new Net for: " + net.getName());
					Net newNet = new Net(net.getName() + "second", net.getType());
					design.addNet(newNet);

					net.getPins().forEach(temp -> {
						if (!checkInside(temp.getInstance())) {
							newNet.addPin(temp);
						}
					});

					net.getPins().removeIf(pin -> checkInside(pin.getInstance()));
				}*/
				throw new RuntimeException("Partially included tieoff");
				//return true;
			}
			return include == StreamUtil.AllOrAnyMatch.ALL;

		} else {
			return checkOnlyByType(i);
		}
	}


	public static InclusionType getInclusionType(Instance instance, Set<Instance> inOutModuleIobs) {

		if (inOutModuleIobs.contains(instance)) {
			return InclusionType.InOutModuleIob;
		}

		if (instance.getType().equals(PrimitiveType.TIEOFF) || isStaticSourceInst(instance)) {
			return getInclusionTypeByNeighbours(instance, inOutModuleIobs);
		}

		if (isInstanceIncluded(instance)) {
			return InclusionType.Inside;
		}

		if (isInModuleInstance(instance)) {
			return InclusionType.InModule;
		}

		if (isOutModuleInstance(instance)) {
			return InclusionType.OutModule;
		}

		if (instance.isIOB()) {
			return InclusionType.IobOutside;
		}

		return InclusionType.OtherOutside;
	}

	private static InclusionType getInclusionTypeByNeighbours(Instance instance, Set<Instance> inOutModuleIobs) {
		//For tieoffs, we include them if their connected instances are included
		//Get connected pins
		//And check if all match
		Set<InclusionType> include = instance.getNetList().stream()
				//Get connected pins
				.flatMap(n -> n.getPins().stream().filter(p -> p != n.getSource()))
				//And check if all match
				.map(p -> getInclusionType(p.getInstance(), inOutModuleIobs))
				.collect(Collectors.toSet());

		//If there was no connection, we don't want to include it
		if (include.isEmpty()) {
			return InclusionType.OtherOutside;
		}

		if (include.size() > 1) {
			throw new IllegalStateException("Partially included tieoff");
				/*logger.warn("Tieoff {} is partially included. Splitting its connections...", i.getName());
				for (Net net : i.getNetList()) {
					logger.info("Create new Net for: " + net.getName());
					Net newNet = new Net(net.getName() + "second", net.getType());
					design.addNet(newNet);

					net.getPins().forEach(temp -> {
						if (!checkInside(temp.getInstance())) {
							newNet.addPin(temp);
						}
					});

					net.getPins().removeIf(pin -> checkInside(pin.getInstance()));
				}*/
		}

		return include.iterator().next();

	}

	private static boolean isInModuleInstance(Instance instance) {
		return instance.getName().startsWith("inmodule_exclude");
	}

	private static boolean isOutModuleInstance(Instance instance) {
		return instance.getName().startsWith("outmodule_exclude");
	}
}
