/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module_synthesis;

import de.tu_darmstadt.rs.rapidSoc.LogMarkers;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.IPortMapping;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.INamed;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IIntegerParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.design.Port;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.PrimitiveType;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public abstract class ModuleExtractor {

	protected static final String portSpace = "port ";
	private static final Logger logger = LoggerFactory.getLogger(ModuleExtractor.class);

	protected final Design design;
	protected final IConfigurationDocument document;
	protected Collection<IModuleConfiguration<?>> jmodules;
	private List<String> unrouteNets = Arrays.asList(
			//Core
			"spmc/ifid/N48",
			"spmc/pc_trap[8]",
			"addr_ram[17]",
			"di_ram_code[8]",
			"spmc/exmem/alu_inst/differenz[11]",
			"spmc/exmem/alu_inst/differenz[6]",
			"spmc/exmem/alu_inst/Sh741",
			"spmc/ifid/pc_new[10]3",
			"spmc/ifid/pc_new[11]2",
			"spmc/pipe_1[64]",
			"spmc/pipe_1[65]",
			"spmc/pipe_1[66]",

			//USB
			"DataIn[5]",
			"DataIn[7]",
			"rx_data[1]",
			"rx_data[0]",
			"phy/i_rx_phy/hold_reg[2]",
			"phy/i_rx_phy/hold_reg[5]",
			"phy/i_rx_phy/hold_reg[7]",
			"pl/br/ctrl/set_address",
			"pl/pd/crc16_sum[3]",
			"pl/pd/crc16_sum[4]",

			//uart_light
			"uart0/clock_generator0/divisor_tx[4]",
			"uart0/uart_tx0/tx_dp0/yaf_tx0/wbin[2]",
			"uart0/uart_rx0/rx_dp0/yaf_rx0/yaf_sync_r2w/rinc_s_winc_OR_53_o_inv1_cepot",
			"decod1/reg_read_1",

			//core+mem
			"spartanmc_0/Madd_n0132[10:0]_lut[4]",
			"spartanmc_0/pipe_1[63]",
			"spartanmc_0/pipe_1[68]",
			"spartanmc_0/exmem/alu_inst/Mmux_result123",
			"spartanmc_0_spartanmc_mem_local_0/code_do_all[8]",
			"spartanmc_mem_local_0/code_do_all[42]",
			"spartanmc_0/ifid/pc[5]",
			"spartanmc_0/exmem/alu_op_a[1]",
			"spartanmc_0/exmem/alu_op_a[3]",
			"spartanmc_0/exmem/alu_op_a[16]",
			"spartanmc_0/exmem/sfr/_n0117_inv",

			//Ethernet_rx
			"Madd_n0200[13:0]_Madd6",

			//DCM
			"reset_dcm"
	);

	public ModuleExtractor(Design design, Collection<IModuleConfiguration<?>> jmodules, IConfigurationDocument document) {
		this.design = design;
		this.jmodules = jmodules;
		this.document = document;
	}

	/**
	 * Get the local name for a bus port at the local module
	 *
	 * @param busHandler the bus handler
	 * @param portName the port name
	 * @return the local name
	 */
	public static String findLocalNameForBusPort(IBusHandler busHandler, String portName) {
		return busHandler.getConnectorDescription().getPortMappings().stream()
				.filter(m -> m.getBusName().equals(portName)).collect(StreamUtil.atMostOne()).map(IPortMapping::getName).orElse(portName);
	}

	protected void generateAnchor(Module rmodule) {
		Objects.requireNonNull(design.getDevice());
		Instance inst = rmodule.getInstances().stream()
				.sorted(Comparator.comparingInt((Instance i) -> {
					PrimitiveSite[] sites = design.getDevice().getAllCompatibleSites(i.getType());
					if (sites == null) {
						logger.warn(LogMarkers.rapidSoc, "did not find sites for " + i + "(" + i.getType() + ")");
						return Integer.MAX_VALUE;
					}
					return sites.length;
				}))
				.findFirst()
				.orElseThrow(() -> new RuntimeException("module is empty"));

		rmodule.setAnchor(inst);
		design.addModule(rmodule);
	}

	protected void setPortNames(Port port, Stream<String> names) {
		StreamUtil.FirstAndRest<String> namesCollect = names.collect(StreamUtil.collectFirstAndRest());
		//TODO unify name and aliases???
		port.setName(namesCollect.getFirst());
		port.getAliases().addAll(namesCollect.getRest());

	}

	protected void generatePortName(Port port) {
		//Find connected iob

		Net net = port.getPins().iterator().next().getNet();

		Stream<String> names = getPortNames(net, port.isOutPort());


		setPortNames(port, names);

	}

	@NotNull
	protected Stream<String> getPortNames(Net net, boolean isOutPort) {
		Stream<Pin> iobPins;
		if (isOutPort) {

			iobPins = net.getPins().stream()
					.filter(pin -> net.getSource() != pin)
					.filter(pin -> !InstanceInsideChecker.checkOnlyByType(pin.getInstance()));
		} else {
			iobPins = Stream.of(net.getSource());
		}


		return iobPins.map(this::findPortNameFromOutsidePin);
	}


	protected Optional<String> findPortNameFromInOutModuleInstance(Instance instance, String moduleName, String suffix) {
		String moduleNameLower = moduleName.toLowerCase();
		String instanceName = instance.getName().toLowerCase();

		if (!instanceName.contains(moduleNameLower)) {
			return Optional.empty();
		}

		boolean withoutBracket = !instanceName.contains("[");

		int remoteConnectionIndex;
		if (withoutBracket) {
			remoteConnectionIndex = 0;
		} else {

			final Pattern regex = Pattern.compile("(" + moduleNameLower + "[_$/](?:pin|temp)\\[)(\\d+)(\\])");
			Matcher matcher = regex.matcher(instanceName);
			if (!matcher.matches()) {
				throw new RuntimeException("Pattern " + regex.pattern() + " does not match text " + instanceName);
			}
			String index = matcher.group(2);
			remoteConnectionIndex = Integer.parseInt(index);
		}

		IModuleConfiguration<? extends IConfigurationDocument> module = findInOrOutModule(moduleNameLower);
		IPortHandler<? extends IPort> portHandler = DocumentUtil.getPortHandler(module, "pin");


		if (remoteConnectionIndex >= ((IIntegerParameterHandler) DocumentUtil.getParameterHandler(module, "PORT_WIDTH")).getIntegerValue().get()) {
			throw new IllegalStateException("Index out of range for " + instanceName + ", while creating port name");
		}


		IConnection<IPortHandler<?>> connection = portHandler.streamConnections()
				.filter(c -> (remoteConnectionIndex >= c.getMyStart(portHandler) && remoteConnectionIndex <= c.getMyEnd(portHandler)))
				.collect(StreamUtil.exactlyOne());

		int indexWithinConnection = remoteConnectionIndex - connection.getMyStart(portHandler);

		//PortHandler2 von Module outModule_exclude
		//Connection zwischen unserem übrig gebliebenen Pin von outModule zu dem Pin im eigentlichen Modul
		//IConnection tempConn = portHandler.streamConnections().filter(conns -> conns.getPortB() == connection.getPortB()).collect(StreamUtil.exactlyOne());
		//Local Name vom eigentlich Modul finden und zurück geben
		//String localName = findNameFromConnection2(connection.getPortA(), tempConn, indexWithinConnection);
		String localName = findNameFromConnection2(portHandler, connection, indexWithinConnection);
		logger.debug(instanceName + " -> " + localName + "." + suffix);
		return Optional.of(localName + "." + suffix);
/*
		regexSLICES = Pattern.compile("(.+)(\\[\\d+\\])");
		matcherSLICES = regexSLICES.matcher(localName);
		if (matcherSLICES.find()) {
			localName = matcherSLICES.group(1);
		}

		if (connection.getPortB().getConnectorWidth() > 1) {
			logger.debug(instanceName + " -> " + localName + "[" + tempReturn + "]." + suffix);
			return localName + "[" + tempReturn + "]." + suffix;
		} else {
			logger.debug(instanceName + " -> " + localName + "." + suffix);
			return localName + "." + suffix;
		}*/
	}

	protected IModuleConfiguration<? extends IConfigurationDocument> findInOrOutModule(String moduleNameLower) {
		return document.streamCommonModules()
				.filter(m -> m.getName().equalsIgnoreCase(moduleNameLower))
				.collect(StreamUtil.exactlyOne(
						(l) -> "Multiple modules: " + l,
						() -> "not found. available: " + document.streamCommonModules().map(INamed::getName).collect(Collectors.joining(", "))
				));
	}

	protected Optional<String> findPortNameFromInOutModuleInstance(Instance instance) {

		Optional<String> name = findPortNameFromInOutModuleInstance(instance, ModuleSynthesis.IN_MODULE_EXCLUDE, "I");
		if (name.isPresent()) {
			return name;
		}

		name = findPortNameFromInOutModuleInstance(instance, ModuleSynthesis.OUT_MODULE_EXCLUDE, "O");
		if (name.isPresent()) {
			return name;
		}

		//return Optional.of(findNameFromInOutModulePinOld(inOutModulePin));
		return Optional.empty();


	}

	/*
		@Deprecated
		@NotNull
		private String findNameFromInOutModulePinOld(Pin inOutModulePin) {
			//auch die neuen Module pseudoIn und pseudoOut, werden als inmodule_exclude bzw.
			//outmodule_exclude erstellt und duch den suffix _exclude entfernt und müssen hier
			//berücksichtigt werden
			//inModule hat ebenfalls exakt 3 pins: clock, in, out. Interessant ist der Out-Pin, es
			//wird ermittelt, woran dieser angeschlossen ist
			//outModule hat 6 pins, 2 interne pins, die flipflop und or verbinden, clock, flip_in,
			//or_in und out

			boolean inModuleWithoutBracket = inOutModulePin.getInstanceName().contains("inmodule_exclude") && !inOutModulePin.getInstanceName().contains("[");

			Pattern regexSLICES = Pattern.compile("(inmodule_exclude[_$]pin\\[)(\\d+)(\\])");
			Matcher matcherSLICES = regexSLICES.matcher(inOutModulePin.getInstanceName());
			if (matcherSLICES.find() || inModuleWithoutBracket) {
				String suffix = "I";
				String str;
				if (!inModuleWithoutBracket) {
					str = matcherSLICES.group(2);
				} else {
					str = "0";
				}
				//Portname versuchen zu finden und zurück zu geben...
				IModuleConfiguration<?> module = DocumentUtil.findModulesByName(document, "inModule_exclude").stream().collect(StreamUtil.exactlyOne());
				IPortHandler<? extends IPort> portHandler = DocumentUtil.getPortHandler(module, "pin");


				if (Integer.parseInt(str) >= ((IIntegerParameterHandler) DocumentUtil.getParameterHandler(module, "PORT_WIDTH")).getIntegerValue().get()) {
					throw new RuntimeException("Index out of range for " + inOutModulePin.getInstanceName() + ", while creating port name for " + inOutModulePin);
				}


				IConnection<IPortHandler<?>> ph = portHandler.streamConnections()
						.filter(temp_ph -> {
							return (Integer.parseInt(str) >= temp_ph.getStartA() && Integer.parseInt(str) <= temp_ph.getEndA());
						})
						.collect(StreamUtil.exactlyOne());

				int tempReturn = Integer.parseInt(str);
				tempReturn -= ph.getStartA();

				//PortHandler2 von Module outModule_exclude
				//Connection zwischen unserem übrig gebliebenen Pin von outModule zu dem Pin im eigentlichen Modul
				IConnection tempConn = portHandler.streamConnections().filter(conns -> conns.getPortB() == ph.getPortB()).collect(StreamUtil.exactlyOne());
				//Local Name vom eigentlich Modul finden und zurück geben
				String localName = findNameFromConnection2(ph.getPortA(), tempConn, null);

				regexSLICES = Pattern.compile("(.+)(\\[\\d+\\])");
				matcherSLICES = regexSLICES.matcher(localName);
				if (matcherSLICES.find()) {
					localName = matcherSLICES.group(1);
				}

				if (ph.getPortB().getConnectorWidth() > 1) {
					logger.debug(inOutModulePin.getInstanceName() + " -> " + localName + "[" + tempReturn + "]." + suffix);
					return localName + "[" + tempReturn + "]." + suffix;
				} else {
					logger.debug(inOutModulePin.getInstanceName() + " -> " + localName + "." + suffix);
					return localName + "." + suffix;
				}

			}

			boolean outModuleWithoutBracket = inOutModulePin.getInstanceName().contains("outmodule_exclude") && !inOutModulePin.getInstanceName().contains("[");

			regexSLICES = Pattern.compile("(outmodule_exclude.+\\[)(\\d+)(\\])");
			matcherSLICES = regexSLICES.matcher(inOutModulePin.getInstanceName());
			if (matcherSLICES.find() || outModuleWithoutBracket) {

				String suffix = "O";

				String str;
				if (!outModuleWithoutBracket) {
					str = matcherSLICES.group(2);
				} else {
					str = "0";
				}

				//Portname versuchen zu finden und zurück zu geben...
				final IModuleConfiguration<?> module = DocumentUtil.findModulesByName(document, "outModule_exclude").stream().collect(StreamUtil.exactlyOne());
				final IPortHandler<? extends IPort> portHandler = DocumentUtil.getPortHandler(module, "pin");


				if (Integer.parseInt(str) >= ((IIntegerParameterHandler) DocumentUtil.getParameterHandler(module, "PORT_WIDTH")).getIntegerValue().get()) {
					throw new RuntimeException("Index out of range for " + inOutModulePin.getInstanceName() + ", while creating port name for " + inOutModulePin);
				}

				IConnection<IPortHandler<?>> ph = portHandler.streamConnections()
						.filter(temp_ph -> {
							return (Integer.parseInt(str) >= temp_ph.getStartA() && Integer.parseInt(str) <= temp_ph.getEndA());
						})
						.collect(StreamUtil.exactlyOne());

				int tempReturn = Integer.parseInt(str);
				tempReturn -= ph.getStartA();

				//PortHandler von Module outModule_exclude
				//Connection zwischen unserem übrig gebliebenen Pin von outModule zu dem Pin im eigentlichen Modul
				IConnection tempConn = portHandler.streamConnections().filter(conns -> conns.getPortB() == ph.getPortB()).collect(StreamUtil.exactlyOne());
				//Local Name vom eigentlich Modul finden und zurück geben
				String localName = findNameFromConnection2(ph.getPortA(), tempConn, null);

				regexSLICES = Pattern.compile("(.+)(\\[\\d+\\])");
				matcherSLICES = regexSLICES.matcher(localName);
				if (matcherSLICES.find()) {
					localName = matcherSLICES.group(1);
				}

				if (ph.getPortB().getConnectorWidth() > 1) {
					logger.debug(inOutModulePin.getInstanceName() + " -> " + localName + "[" + tempReturn + "]." + suffix);
					return localName + "[" + tempReturn + "]." + suffix;
				} else {
					logger.debug(inOutModulePin.getInstanceName() + " -> " + localName + "." + suffix);
					return localName + "." + suffix;
				}

			}
			throw new RuntimeException("Could not determine port name in in/out module");
		}
	*/
	protected String findPortNameFromOutsidePin(Pin origiobPin) {
		Pin iobPin = origiobPin;


		String suffix = "";
		while (iobPin.getInstance().getType() != PrimitiveType.IOB) {

			switch (iobPin.getInstance().getType()) {
				//TODO mark that a ff needs to be inserted
				case ILOGIC2:
							/*System.out.print("setting iobPin from " + iobPin);
							suffix = iobPin.getName();
							iobPin = iobPin.getInstance().getPin("D").getNet().getSource();
							System.out.println(" to "+iobPin);
							break;*/
					throw new RuntimeException("Registers joined to IOBs are not currently supported. Error in " + iobPin);
				case OLOGIC2:
					//TODO Check which pin is used...
					//System.out.print("ologic setting iobPin from " + iobPin);
					suffix = iobPin.getName();
					Net n = iobPin.getInstance().getPin("OQ").getNet();
					iobPin = findNetSingleSink(n);
					//Remove suffix for clkout
					if ("CLK0".equals(suffix)) {
						suffix = "";
					}
					//System.out.println(" to " + iobPin);

					break;


				case ADC:
					break;
				case BLOCKRAM:
					break;
				case BSCAN:
					break;
				case BUFDS:
					break;
				case BUFG:
					suffix = iobPin.getName();
					switch (suffix) {
						case "O":
							Pin p = iobPin.getInstance().getPin("I0");
							if (p == null) {
								throw new RuntimeException("BUFG does not have I0. It has: " + iobPin.getInstance().getPins());
							}
							if (p.getNet() == null) {
								throw new RuntimeException("I0 does not have a net.");
							}
							iobPin = p.getNet().getSource();
							suffix = "";
							break;
						case "I0":
							iobPin = getNetSinks(iobPin.getInstance().getPin("O").getNet()).filter(x -> x.getName().equals("CLK0")).collect(StreamUtil.exactlyOne());
							break;
						default:
							throw new RuntimeException("Unsupported BUFG Pin found: " + suffix);
					}
					break;

				case SLICEX:
					//Slice that was excluded by attribute named "exclude_instance_from_module"
					//we assume it has exactly three pin: clock, in and out. change to the other side.

					final Pin temp = iobPin;

					suffix = iobPin.getInstanceName();

							/*logger.error(" ");
							logger.error("module " + module);
							logger.error("port " + port);
							logger.error("iobPin " + iobPin);
							logger.error("iobPin.getNet " + iobPin.getNet());
							logger.error("iobPin.getPort " + iobPin.getPort());
							logger.error("iobPin.getInstanceName " + iobPin.getInstanceName());

							logger.error("-------------- Es folgen Pins der Instance -------------- ");
							logger.error("iobPin.getInstance.getName " + iobPin.getInstance().getName());

							iobPin.getInstance().getPins().forEach(p -> {
								logger.error(" ");
								logger.error("pin: " + p);
								logger.error("pin.getNet(): " + p.getNet());
								if(p.getNet()!= null) {
									p.getNet().getPins().forEach(mt -> {
										logger.error("Pins vom Netz des Pins: " + mt);
									});
								}
								logger.error("pin.getInstance().getName(): " + p.getInstance().getName());
							});

							/*
							//Dump erstellen
							PrimitiveDefList primitiveDefs = FileTools.loadPrimitiveDefs(design.getFamilyType());
							try {
								DotDumper.dumpInstanceOnce(iobPin.getInstance(), primitiveDefs);
							} catch (FileNotFoundException e) {
								throw new RuntimeException(e);
							}*/

					Optional<String> name = findPortNameFromInOutModuleInstance(iobPin.getInstance());
					if (name.isPresent()) {
						return name.get();
					}

					Pin otherSidePin;
					otherSidePin = iobPin.getInstance().getPins().stream().filter(p -> p != temp && !p.getName().equals("CLK")).collect(StreamUtil.exactlyOne());

					if (otherSidePin.isOutPin()) {
						iobPin = otherSidePin.getNet().getPins().stream().filter(p -> p != otherSidePin).collect(StreamUtil.exactlyOne());
					} else {
						iobPin = otherSidePin.getNet().getSource();
						break;
					}

				default:
					//throw new RuntimeException("unexpected primitive type: " + port.getPins().get(0).getNet().getName() + " has type " + iobPin.getInstance().getType());
					throw new RuntimeException("unexpected primitive type: " + iobPin.getInstanceName() + " has type " + iobPin.getInstance().getType() + " while generating port for " + iobPin);
			}

		}


		String name = mapIobPosition(iobPin.getInstance().getPrimitiveSiteName());
		return name + suffix + "." + iobPin.getName();

	}

	private Pin findNetSingleSink(Net n) {
		return getNetSinks(n)
				.collect(StreamUtil.exactlyOne());
	}

	private Stream<Pin> getNetSinks(Net n) {
		return n.getPins().stream()
				.filter(pin -> !(pin == n.getSource()));
	}

	protected String mapIobPosition(String primitiveSiteName) {
		IPinHandler pin = DocumentUtil.streamConnectorHandlers(document).filter(p -> p.getPinDescription().getFpgaPinName().equals(primitiveSiteName)).collect(StreamUtil.exactlyOne());
		return mapIobPosition(pin);
	}

	protected String mapIobPosition(IPinHandler pin) {
		try {
			return pin.streamConnections().map(conn -> findNameFromConnection(pin, conn)).collect(StreamUtil.atLeastOneOrEqual());
		} catch (RuntimeException e) {
			throw new RuntimeException("Error in " + pin + " in " + jmodules + ". Pin has " + pin.getConnectionCount() + " connections", e);
		}
	}

	private String findNameFromConnection(IPinHandler pin, IConnection<IPortHandler<?>> conn) {
		//Is the module a reg or clkout?
		IConfigurationElement<?, ?, ?> ce = conn.getOppositeConnector(pin).getSourceElement();
		if (!(ce instanceof IModuleConfiguration<?>)) {
			throw new RuntimeException("Pin is not connected to module but to " + ce);
		}

		IModuleConfiguration<?> module = (IModuleConfiguration<?>) ce;

		IPortHandler<?> moduleConnector = conn.getOppositeConnector(pin);

		if (!jmodules.contains(module)) {
			if (module.getModuleDescription().getName().equals("clkout")
					|| module.getModuleDescription().getName().equals("reg")) {
				IPortHandler<? extends IPort> port = DocumentUtil.getPortHandler(module, "i");
				conn = port.streamConnections().collect(StreamUtil.exactlyOne());

				module = (IModuleConfiguration<?>) conn.getOppositeConnector(port).getSourceElement();
				moduleConnector = conn.getOppositeConnector(port);
			}
		}

		String offsetSuffix = "";
		if (conn.isMyPartial(moduleConnector)) {
			offsetSuffix = "[" + conn.getMyStart(moduleConnector) + "]";
		}

		if (module.getModuleDescription().getType() == IModuleDescription.ModuleType.BusBreakout) {
			//Find out which bus this connects to
			Bus bbBus = module.getBusDescriptions().stream().collect(StreamUtil.exactlyOne());
			IBusHandler bbHandler = module.getBusHandler(bbBus);
			IBusHandler mHandler = bbHandler.streamOppositeConnectors().collect(StreamUtil.exactlyOne());

			BusPort bbPort = bbHandler.getBusDefinition().getMatchingPorts(bbBus).getPort(moduleConnector.getPortName()).orElseThrow(() -> new RuntimeException(portSpace + bbHandler.getPortName() + " not found"));

			String mName = bbHandler.getBusDefinition().getOppositePort(bbBus, bbPort).getName();

			//Find name of port in module

			String localName = findLocalNameForBusPort(mHandler, mName);
			return localName + offsetSuffix;
		} else {
			return moduleConnector.getPortName() + offsetSuffix;
		}
	}

	protected String findNameFromConnection2(IPortHandler moduleConnector, IConnection<IPortHandler<?>> conn, Integer indexWithinConnection) {
		//Is the module a reg or clkout?
		IConfigurationElement<?, ?, ?> ce = conn.getOppositeConnector(moduleConnector).getSourceElement();

		IModuleConfiguration<?> module = (IModuleConfiguration<?>) ce;

		if (!jmodules.contains(module)) {
			if (module.getModuleDescription().getName().equals("clkout")
					|| module.getModuleDescription().getName().equals("reg")) {
				IPortHandler<? extends IPort> port = DocumentUtil.getPortHandler(module, "i");
				conn = port.streamConnections().collect(StreamUtil.exactlyOne());

				module = (IModuleConfiguration<?>) conn.getOppositeConnector(port).getSourceElement();
				moduleConnector = conn.getOppositeConnector(port);
			}
		}

		String offsetSuffix = "";
		/*if (conn.isMyPartial(moduleConnector)) {
			offsetSuffix = "[" + conn.getMyStart(moduleConnector) + "]";
		}*/
		if (conn.getOppositeConnector(moduleConnector).getConnectorWidth() > 1) {
			int index;
			if (indexWithinConnection == null) {
				if (conn.getMyStart(moduleConnector) == conn.getMyEnd(moduleConnector)) {
					index = conn.getOppositeStart(moduleConnector);
				} else {
					throw new RuntimeException("need an index!");
				}
			} else {
				index = conn.getOppositeStart(moduleConnector) + indexWithinConnection;
			}
			offsetSuffix = "[" + index + "]";
		} else if (indexWithinConnection != null && indexWithinConnection != 0) {
			throw new RuntimeException("Given nonzero index for connector of width 1");
		}

		if (module.getModuleDescription().getType() == IModuleDescription.ModuleType.BusBreakout) {
			//Find out which bus this connects to
			Bus bbBus = module.getBusDescriptions().stream().collect(StreamUtil.exactlyOne());

			IBusHandler bbHandler = module.getBusHandler(bbBus);
			IBusHandler mHandler = bbHandler.streamOppositeConnectors().collect(StreamUtil.exactlyOne());

			IPortHandler finalModuleConnector = moduleConnector;
			BusPort bbPort = bbHandler.getBusDefinition()
					.getMatchingPorts(bbBus)
					.getPort(conn.getOppositeConnector(moduleConnector).getPortName())
					.orElseThrow(() -> new RuntimeException(portSpace + bbHandler.getPortName() + " " + finalModuleConnector.getPortName() + " not found"));

			String mName = bbHandler.getBusDefinition().getOppositePort(bbBus, bbPort).getName();

			//Find name of port in module
			String localName = findLocalNameForBusPort(mHandler, mName);
			return localName + offsetSuffix;
		} else {
			return conn.getOppositeConnector(moduleConnector).getPortName() + offsetSuffix;
		}
	}


	public Module generateModule(String name, boolean isCore) {
		return generateModule(name, false, new HashSet<>(), isCore);
	}

	/**
	 * Generate a module
	 *
	 * @param name the module name
	 * @param noDisconnect disable disconnecting the module
	 * @param partialNets the set of nets crossing the module boundary
	 * @param isCore flag whether this is the core
	 * @return the created module
	 */
	public abstract Module generateModule(String name, boolean noDisconnect, Set<Net> partialNets, boolean isCore);

	protected void unrouteFailNets(Module rModule) {
		//Some nets contain weird irregular pips. until we have figured out which they are, we unroute
		//TODO find out which pips cause the weird behaviour


		unrouteNets.forEach(netName -> {
			Net net = rModule.getNet(netName);
			if (net != null) {
				logger.info(LogMarkers.rapidSoc, "unrouted " + net + " in " + rModule.getName());
				net.unroute();
			}
		});
	}

	protected void unplaceSpecialInstances(Module module) {
		Tracer.enterLoop("Unplace BUFG");
		module.getInstances().forEach(i -> {
			if (i.getType() == PrimitiveType.BUFG || i.getType() == PrimitiveType.BUFGMUX) { //TODO more prims?
				i.unPlace();
				i.getNetList().forEach(Net::unroute);
			}
		});
		Tracer.exit();
	}
}
