/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module_synthesis;

import de.tu_darmstadt.rs.rapidSoc.Util;
import de.tu_darmstadt.rs.rapidSoc.design.ModuleDesign;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationElement;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IConnectorHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPinHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.ConnectedGraph;
import de.tu_darmstadt.rs.util.StreamUtil;

import edu.byu.ece.rapidSmith.design.Attribute;
import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.NetType;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.design.Port;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.PrimitiveType;
import edu.byu.ece.rapidSmith.timing.DotDumper;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Collection;
import java.util.EnumMap;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class ModuleExtractorNew extends ModuleExtractor {

	private static final Logger logger = LoggerFactory.getLogger(ModuleExtractorNew.class);

	private static final Supplier<RuntimeException> UNKNOWN_NAME = () -> new RuntimeException("no name found");

	private final Path dumpPath;

	/**
	 * Create ModuleExtractor
	 *
	 * @param design the design
	 * @param jmodules the jconfig modules
	 * @param document the document
	 */
	public ModuleExtractorNew(Design design, Collection<IModuleConfiguration<?>> jmodules, IConfigurationDocument document, Path dumpPath) {
		super(design, jmodules, document);
		this.dumpPath = dumpPath;
	}

	public ModuleExtractorNew(Design design, Collection<IModuleConfiguration<?>> jmodules, IConfigurationDocument document) {
		this(design, jmodules, document, null);
	}

	/**
	 * Generate a module
	 *
	 * @param name the module name
	 * @param noDisconnect disable disconnecting the module
	 * @param partialNets the set of nets crossing the module boundary
	 * @param isCore flag whether this is the core
	 * @return the created module
	 */
	public Module generateModule(String name, boolean noDisconnect, Set<Net> partialNets, boolean isCore) {
		Set<Instance> inOutModuleIobs = findInOutModuleIobs();

		final Map<InstanceInsideChecker.InclusionType, Set<Instance>> instancesByInclusion =
				design.getInstances().stream()
						.collect(Collectors.groupingBy(
								(instance) -> InstanceInsideChecker.getInclusionType(instance, inOutModuleIobs),
								() -> new EnumMap<>(InstanceInsideChecker.InclusionType.class),
								Collectors.toSet()
						));


		try {
			if (dumpPath != null) {
				DotDumper.dumpConnectedInstances(design.getInstances(), design.getNets(), true, i -> InstanceInsideChecker.getInclusionType(i, inOutModuleIobs).toHtmlColor(), dumpPath);
			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

		Module module = new Module();
		module.setName(name);


		instancesByInclusion.get(InstanceInsideChecker.InclusionType.Inside)
				.forEach(module::addInstance);

		disconnectStaticIOModuleNets(module, instancesByInclusion);


		handleInModule(instancesByInclusion, module, inOutModuleIobs);
		handleOutModule(instancesByInclusion, module, inOutModuleIobs);

		handleIobPorts(module, inOutModuleIobs);

		unifyMultiports(module);

		addInsideNets(module, inOutModuleIobs);

		disconnectDesign(module);

		generateAnchor(module);

		verifyDisconnected(module);

		unplaceSpecialInstances(module);

		return module;
	}

	/**
	 * Static nets may be packed in the in/out modules. Disconnect these
	 *
	 * @param module
	 * @param instancesByInclusion
	 */
	private void disconnectStaticIOModuleNets(Module module, Map<InstanceInsideChecker.InclusionType, Set<Instance>> instancesByInclusion) {
		Stream.of(InstanceInsideChecker.InclusionType.InModule, InstanceInsideChecker.InclusionType.OutModule)
				.map(instancesByInclusion::get)
				.flatMap(Collection::stream)
				.map(Instance::getPins)
				.flatMap(Collection::stream)
				.filter(Pin::isOutPin)
				.forEach(pin -> {
					if (pin.getNet().isStaticNet()) {
						pin.getNet().removePin(pin);
					}
				});
	}

	private Set<Instance> findInOutModuleIobs() {
		IModuleConfiguration<?> inModule = findInOrOutModule(ModuleSynthesis.IN_MODULE_EXCLUDE.toLowerCase());
		IModuleConfiguration<?> outModule = findInOrOutModule(ModuleSynthesis.OUT_MODULE_EXCLUDE.toLowerCase());
		return Stream.of(inModule, outModule)
				.flatMap(module -> module.getPortGroupDescriptions().stream()
						.flatMap(group -> group.getPorts().stream())
						.map(module::getPortHandler))
				.filter(handler -> !handler.getPortName().equals("pin"))
				.flatMap(IConnectorHandler::streamOppositeConnectors)
				.flatMap(StreamUtil.castOrEmpty(IPinHandler.class))
				.map(pinHandler -> pinHandler.getPinDescription().getFpgaPinName())
				.map(location -> design.getDevice().getPrimitiveSite(location))
				.map(design::getInstanceAtPrimitiveSite)
				.collect(Collectors.toSet());
	}

	private boolean verifyNetDisconnected(Net net, Module module) {
		if (net.getPins() == null) {
			throw new RuntimeException("Net " + net.getName() + " has null pins");
		}
		return net.getPins().stream()
				.allMatch(pin -> {
					boolean contains = module.getInstances().contains(pin.getInstance());
					if (!contains) {
						logger.error("Net " + net.getName() + " is connected to " + pin + " outside module");
					}
					return contains;
				});
	}

	private void verifyDisconnected(Module module) {
		boolean ok = true;
		for (Net net : module.getNets()) {
			ok &= verifyNetDisconnected(net, module);
		}

		for (Port port : module.getPorts()) {
			for (Pin pin : port.getPins()) {
				if (pin.getNet() != null) {
					ok &= verifyNetDisconnected(pin.getNet(), module);
				}
			}
		}
		if (!ok) {
			throw new IllegalStateException("Module is not completely disconnected");
		}
	}


	private static class PortGroup extends ConnectedGraph.ConnectedComponent<Port> {

		public PortGroup(Port element) {
			super(element);
		}
	}


	private static class PortGraph extends ConnectedGraph<Port, PortGroup> {

		@Override
		protected Port getNewRepresentative(Port a, Port b) {
			return a;
		}

		@Override
		protected PortGroup createGroup(Port element) {
			return new PortGroup(element);
		}

		public void addPortConnections(Module module) {

			for (Port port : module.getPorts()) {
				getGroupOrCreate(port);
			}

			Map<Pin, List<Port>> portsByPin = module.getPorts().stream().collect(Util.multiGroupingBy(p -> p.getPins().stream()));
			portsByPin.values().forEach(list -> {
				if (list.size() > 1) {
					Port first = list.get(0);
					for (int i = 1; i < list.size(); i++) {
						connect(first, list.get(i));
					}
				}
			});
		}

		@Override
		public Set<PortGroup> getActiveGroups() {
			return super.getActiveGroups();
		}

		@Override
		public PortGroup getGroup(Port element) {
			return super.getGroup(element);
		}
	}

	/**
	 * Right now, multiple ports may contain the same pins. fix that
	 *
	 * @param module
	 */
	private void unifyMultiports(Module module) {
		PortGraph graph = new PortGraph();
		graph.addPortConnections(module);


		for (PortGroup group : graph.getActiveGroups()) {
			for (Port port : group.contained) {
				if (port != group.representative) {
					for (Pin pin : port.getPins()) {
						if (!group.representative.getPins().contains(pin)) {
							group.representative.addPin(pin);
						}
						pin.setPort(group.representative);
					}
					if (group.representative.getNetType() != port.getNetType()) {
						throw new IllegalStateException("Differing nettypes while unifying ports");
					}
					if (group.representative.getPassthroughFrom() != port.getPassthroughFrom()) {
						throw new IllegalStateException("Differing passthroughFroms while unifying ports");
					}
					group.representative.addAlias(port.getName());
					for (String s : port.getAliases()) {
						group.representative.addAlias(s);
					}
					for (Port pt : port.getPassthroughPorts()) {
						group.representative.addPassthroughPort(graph.getGroup(pt).representative);
					}
				}
			}
		}

		module.setPorts(graph.getActiveGroups().stream().map(g -> g.representative).collect(Collectors.toSet()));
	}

	private void disconnectDesign(Module module) {
		for (Port port : module.getPorts()) {
			port.getPins().stream()
					.map(Pin::getNet)
					.distinct()
					.collect(StreamUtil.atMostOne())
					.ifPresent(net -> {

						/*if (port.getPins().contains(net.getSource())) {
							if (port.getPins().size() != 1) {
								throw new IllegalStateException("not exactly one pin in output port " + port);
							}
							net.removePin(port.getPin());
							module.addNet(net);
						} else {
							net.getPins().removeIf(p -> !port.getPins().contains(p));
						}*/

						List<Pin> pinsToRemove;

						if (port.getPins().contains(net.getSource())) {
							//Remove all external pins
							pinsToRemove = net.getPins().stream().filter(pin -> !module.getInstances().contains(pin.getInstance())).collect(Collectors.toList());
							net.unroute();
							module.addNet(net);
						} else {
							//Disconnect net
							pinsToRemove = net.getPins().stream().filter(p -> !port.getPins().contains(p)).collect(Collectors.toList());
							net.unroute();
						}

						for (Pin pin : pinsToRemove) {
							net.removePin(pin);
						}
					});
		}
	}

	private void addInsideNets(Module module, Set<Instance> inOutModuleIobs) {
		for (Net net : design.getNets()) {
			if (!net.getPins().isEmpty()
					&& net.getPins().stream()
					.map(Pin::getInstance)
					.distinct()
					.map((Instance instance) -> InstanceInsideChecker.getInclusionType(instance, inOutModuleIobs))
					.allMatch(InstanceInsideChecker.InclusionType::isInside)
					) {
				module.addNet(net);
			}
		}
	}


	private void handleOutModule(Map<InstanceInsideChecker.InclusionType, Set<Instance>> instancesByInclusion, Module module, Set<Instance> inOutModuleIobs) {
		instancesByInclusion.get(InstanceInsideChecker.InclusionType.OutModule)
				.forEach(instance -> {
					@NotNull String name = findPortNameFromInOutModuleInstance(instance).orElseThrow(UNKNOWN_NAME);
					Port port = new Port();
					setPortNames(port, Stream.of(name));

					final boolean[] isStaticOutput = {true};

					instance.getPins().stream()
							.filter(((Predicate<Pin>) Pin::isOutPin).negate())
							.forEach(modulePin -> {
								Pin pin = modulePin.getNet().getSource();

								InstanceInsideChecker.InclusionType inclusion = InstanceInsideChecker.getInclusionType(pin.getInstance(), inOutModuleIobs);
								switch (inclusion) {
									case Inside:
										port.addPin(pin);
										pin.setPort(port);
										isStaticOutput[0] = false;
										break;
									case OtherOutside:
										//ignoring other outside
										break;
									case InModule:
										//reverse passthrough, already added
										isStaticOutput[0] = false;
										break;
									case OutModule:
										//ignoring connection to other outmodule inst
										break;
									case IobOutside:
										addPassthrough(port, pin, module, inclusion, true);
										isStaticOutput[0] = false;
										break;
									case InOutModuleIob:
										//ignoring connection to inoutmodule iob
										break;
								}
							});

					if (isStaticOutput[0]) {
						detectStaticOutput(instance, port);
					}
					if (isPortRelevant(port)) {
						module.addPort(port);
					}

				});
	}


	/**
	 * Check if the port is relevant, i.e. not empty
	 *
	 * @param port
	 * @return
	 */
	private boolean isPortRelevant(Port port) {
		final boolean hasPins = !port.getPins().isEmpty();
		final boolean isStatic = port.getNetType() != NetType.WIRE;
		final boolean isPassthroughIn = !port.getPassthroughPorts().isEmpty();
		final boolean isPassthroughOut = port.getPassthroughFrom() != null;

		return hasPins || isStatic || isPassthroughIn || isPassthroughOut;
	}

	private NetType getStaticOutputType(Instance instance) {
		int lutInputCount = 0;
		int xInputCount = 0;
		int qOutputCount = 0;
		int clkInputCount = 0;
		for (Pin pin : instance.getPins()) {
			String name = pin.getName();
			if (name.matches(".Q")) {
				qOutputCount++;
			} else if (name.matches(".X")) {
				xInputCount++;
			} else if (name.matches("[ABCD][123456]")) {
				lutInputCount++;
			} else if (name.equals("CLK")) {
				clkInputCount++;
			}
		}

		try {
			if (clkInputCount != 1) {
				throw new IllegalStateException("Instance " + instance.getName() + " does not have exactly one clk port");
			}
			if (qOutputCount != 1) {
				throw new IllegalStateException("Instance " + instance.getName() + " does not have exactly one q output port");
			}
			if ((lutInputCount + xInputCount) != 1) {
				throw new IllegalStateException("Instance " + instance.getName() + " does not have exactly one lut input or x input port");
			}
		} catch (IllegalStateException e) {
			//More infos in case of errors
			logger.info("Handling " + instance.getName());
			for (Pin pin : instance.getPins()) {
				logger.info("{} is {} connected to {}", pin.getName(), pin.getPinType(), pin.getNet().getName());
			}
			throw e;
		}
		//If we have a lut, it's inverting -> static 1
		if (lutInputCount == 1) {
			return NetType.VCC;
		} else {
			return NetType.GND;
		}
	}

	private void detectStaticOutput(Instance instance, Port port) {
		port.setNetType(getStaticOutputType(instance));

	}

	private void handleInModule(Map<InstanceInsideChecker.InclusionType, Set<Instance>> instancesByInclusion, Module module, Set<Instance> inOutModuleIobs) {
		instancesByInclusion.get(InstanceInsideChecker.InclusionType.InModule)
				.forEach(instance -> {
					@NotNull String name = findPortNameFromInOutModuleInstance(instance).orElseThrow(UNKNOWN_NAME);
					Port inPort = new Port();
					setPortNames(inPort, Stream.of(name));


					instance.getPins().stream()
							.filter(Pin::isOutPin)
							//Static sources may get packed into the same instance. they already have their net removed. filter them out
							.filter(pin -> !Objects.isNull(pin.getNet()))
							.forEach(source -> {
								for (Pin pin : source.getNet().getPins()) {
									if (pin != source) {
										InstanceInsideChecker.InclusionType inclusion = InstanceInsideChecker.getInclusionType(pin.getInstance(), inOutModuleIobs);
										if (inclusion == InstanceInsideChecker.InclusionType.OutModule) {
											addPassthrough(inPort, pin, module, inclusion, false);
										} else if (inclusion == InstanceInsideChecker.InclusionType.IobOutside) {
											addPassthrough(inPort, pin, module, inclusion, false);
										} else if (inclusion != InstanceInsideChecker.InclusionType.InModule) {
											if (!inclusion.isInside()) {
												throw new RuntimeException("unexpected inclusion type " + inclusion);
											}
											inPort.addPin(pin);
											pin.setPort(inPort);
										}
									}
								}

							});
					if (isPortRelevant(inPort)) {
						module.addPort(inPort);
					}
				});
	}

	private void addPassthrough(Port inPort, Pin passthroughPin, Module module, InstanceInsideChecker.InclusionType inclusion, boolean reverse) {
		Port outPort;
		switch (inclusion) {
			case OutModule:
				outPort = new Port(findPortNameFromInOutModuleInstance(passthroughPin.getInstance()).orElseThrow(UNKNOWN_NAME), null);
				break;
			case IobOutside:
				String portName = mapIobPosition(passthroughPin.getInstance().getPrimitiveSiteName());
				outPort = new Port(portName + "." + passthroughPin.getName(), null);
				break;
			default:
				throw new IllegalStateException("Unsupported inclusion type for passthrough: " + inclusion);
		}
		if (reverse) {
			outPort.addPassthroughPort(inPort);
		} else {
			inPort.addPassthroughPort(outPort);
		}
		module.addPort(outPort);
	}


	private boolean isStaticSourceInst(Instance i) {
		//TODO check for routethroughs?
		//if (i.getAttribute("_ROUTETHROUGH")!=null)
		//    return false;
		return i.getPins().stream().allMatch(pin ->
				Stream.of(i.getAttribute("_GND_SOURCE"), i.getAttribute("_VCC_SOURCE"))
						.filter(Objects::nonNull)
						.flatMap(a -> Stream.of(a.getMultiValueValues()))
						.anyMatch(s -> s.equals(pin.getName())));
	}

	/**
	 * Find instances at the edge of the design (like IOBs)
	 *
	 * @return
	 */
	private Stream<Instance> findEdgeInstances() {
		return design.getInstances().stream()
				.filter(Instance::isIOB);
	}

	private Stream<Instance> getConnectedInstances(boolean searchForward, Net net) {
		if (!searchForward) {
			return Stream.of(net.getSource().getInstance());
		}
		return net.getPins().stream()
				.filter(p -> p != net.getSource())
				.map(Pin::getInstance);
	}


	private Predicate<Net> isPartialNetPredicate(Instance current, boolean searchForward, Set<Instance> inOutModuleIobs) {
		return net -> getConnectedInstances(true, net)
				.anyMatch(next -> isPartialNet(current, next, inOutModuleIobs));
	}

	private boolean isPartialNet(Instance current, Instance next, Set<Instance> inOutModuleIobs) {
		final InstanceInsideChecker.InclusionType currentInclusion = InstanceInsideChecker.getInclusionType(current, inOutModuleIobs);
		final InstanceInsideChecker.InclusionType nextInclusion = InstanceInsideChecker.getInclusionType(next, inOutModuleIobs);

		if (currentInclusion.isInside() != nextInclusion.isInside()) {
			return true;
		}

		if (currentInclusion == InstanceInsideChecker.InclusionType.OutModule && nextInclusion == InstanceInsideChecker.InclusionType.InModule) {
			return true;
		}

		return currentInclusion == InstanceInsideChecker.InclusionType.InModule && nextInclusion == InstanceInsideChecker.InclusionType.OutModule;
	}

	private Instance getInstanceFromJcPin(IPinHandler port) {
		String fpgaPinName = port.getPinDescription().getFpgaPinName();
		PrimitiveSite primitiveSite = design.getDevice().getPrimitiveSite(fpgaPinName);
		Instance res = design.getInstanceAtPrimitiveSite(primitiveSite);
		if (res == null) {
			throw new NullPointerException("Did not find instance for " + port + " at " + fpgaPinName);
		}
		return res;
	}

	private Pin getPinForDirection(IPinHandler port, ConnectorDirection direction) {
		Instance instance = getInstanceFromJcPin(port);
		return instance.getPin(Character.toString(ModuleDesign.getDirectionSuffix(direction)));
	}

	private String getPortName(IPinHandler pinHandler, ConnectorDirection direction) {
		String name = mapIobPosition(pinHandler);
		return name + "." + ModuleDesign.getDirectionSuffix(direction);
	}

	private StreamUtil.Pair<Port, Pin> getPortInfos(IPinHandler port, ConnectorDirection direction) {
		Port rport = new Port();
		rport.setName(getPortName(port, direction));
		return new StreamUtil.Pair<>(rport, getPinForDirection(port, direction));
	}

	private Stream<StreamUtil.Pair<Port, Pin>> findIobPorts() {

		return DocumentUtil.streamConnectorHandlers(document)
				.filter(IConnectorInstance::isConnected)
				.filter(pin -> !isInOrOutModulePin(pin))
				.flatMap(port -> {
					ConnectorDirection direction = port.getDirection();
					switch (direction) {
						case Input:
						case Output:
							return Stream.of(getPortInfos(port, direction.getReverse()));
						case InOut:
							return Stream.of(
									getPortInfos(port, ConnectorDirection.Input),
									getPortInfos(port, ConnectorDirection.Output),
									getPortInfos(port, ConnectorDirection.InOut)
							);
						default:
							throw new IllegalStateException("Connector direction not handled for " + port);
					}
				});
	}

	private boolean isInOrOutModulePin(IPinHandler pin) {
		int connectionCount = pin.getConnectionCount();
		if (connectionCount != 1) {
			throw new IllegalStateException("Expected exactly one connection for " + pin);
		}
		IConfigurationElement<?, ?, ?> sourceElement = pin.streamOppositeConnectors().iterator().next().getSourceElement();
		return isInOrOutModule(sourceElement);
	}

	private boolean isInOrOutModule(IConfigurationElement<?, ?, ?> module) {
		String name = module.getName();
		return name.equals(ModuleSynthesis.IN_MODULE_EXCLUDE) || name.equals(ModuleSynthesis.OUT_MODULE_EXCLUDE);
	}

	private void handleIobPorts(Module module, Set<Instance> inOutModuleIobs) {
		Map<Pin, Port> iobPorts = findIobPorts()
				.collect(Collectors.toMap(StreamUtil.Pair::getB, StreamUtil.Pair::getA));


		iobPorts.forEach((iobPin, port) -> {
			walkInFromEdge(iobPin, port, iobPorts, inOutModuleIobs);
			if (isPortRelevant(port)) {
				module.addPort(port);
			}
		});
	}

	private void walkInFromEdge(Pin outsidePin, Port port, Map<Pin, Port> iobPorts, Set<Instance> inOutModuleIobs) {
		outsidePin.getNet().streamOppositePins(outsidePin)
				.forEach(closerPin -> {
					Instance inst = closerPin.getInstance();
					InstanceInsideChecker.InclusionType inctype = InstanceInsideChecker.getInclusionType(inst, inOutModuleIobs);
					switch (inctype) {
						//We found the module, we are finished
						case Inside:
							port.addPin(closerPin);
							closerPin.setPort(port);
							break;
						//Get other side of outside instance and call ourselves again
						case OtherOutside:
							getOtherSideOfOutsideInst(closerPin).forEach(p -> walkInFromEdge(p, port, iobPorts, inOutModuleIobs));
							break;
						case IobOutside:
							addIobPassthrough(port, iobPorts, closerPin);
							break;
						//Unimplemented
						case InModule:
						case OutModule:
							//Reverse passthrough, already added
							break;

						default:
							throw new IllegalStateException("Unexpected inclusion type: " + inctype);
					}
				});

	}

	private void addIobPassthrough(Port port, Map<Pin, Port> iobPorts, Pin pin) {

		Port otherPort = iobPorts.get(pin);
		if (otherPort == null) {
			throw new IllegalStateException("Passthrough, but did not find other port");
		}

		Port source;
		Port destination;

		if (pin.isOutPin()) {
			source = otherPort;
			destination = port;
		} else {
			source = port;
			destination = otherPort;
		}

		source.addPassthroughPort(destination);

	}

	private Stream<Pin> getOtherSideOfOutsideInst(Pin pin) {
		Instance instance = pin.getInstance();
		if (instance.getType() == PrimitiveType.BUFG) {
			return instance.getPins().stream().filter(p -> p != pin);
		} else if (isClkout(instance)) {
			if (pin.isOutPin()) {
				return Stream.of(instance.getPin("CLK0"));
			} else {
				return Stream.of(instance.getPin("OQ"));
			}
		}
		throw new IllegalStateException("Unknown outside instance, cannot get other side pin: " + pin);
	}

	private static final Map<String, String> clkoutReferenceAttributes = new HashMap<>();

	static {
		clkoutReferenceAttributes.put("D1USED", "0");
		clkoutReferenceAttributes.put("CLK0INV", "CLK0");
		clkoutReferenceAttributes.put("OUTFFTYPE", "DDR");
		clkoutReferenceAttributes.put("SRTYPE_OQ", "#OFF");
		clkoutReferenceAttributes.put("OREVUSED", "#OFF");
		clkoutReferenceAttributes.put("TFFTYPE", "#OFF");
		clkoutReferenceAttributes.put("SRINIT_OQ", "0");
		clkoutReferenceAttributes.put("OT1USED", "#OFF");
		clkoutReferenceAttributes.put("O1USED", "#OFF");
		clkoutReferenceAttributes.put("D2USED", "0");
		clkoutReferenceAttributes.put("MISR_ENABLE_CLK", "#OFF");
		clkoutReferenceAttributes.put("SRTYPE_TQ", "#OFF");
		clkoutReferenceAttributes.put("DDR_ALIGNMENT", "NONE");
		clkoutReferenceAttributes.put("MISRATTRBOX", "#OFF");
		clkoutReferenceAttributes.put("OMUX", "OUTFF");
		clkoutReferenceAttributes.put("T2USED", "#OFF");
		clkoutReferenceAttributes.put("TCEUSED", "#OFF");
		clkoutReferenceAttributes.put("CLK1INV", "CLK1_B");
		clkoutReferenceAttributes.put("TSRUSED", "#OFF");
		clkoutReferenceAttributes.put("T1USED", "#OFF");
		clkoutReferenceAttributes.put("TREVUSED", "#OFF");
		clkoutReferenceAttributes.put("OUTFF", "");
		clkoutReferenceAttributes.put("SRINIT_TQ", "#OFF");
		clkoutReferenceAttributes.put("OSRUSED", "#OFF");
		clkoutReferenceAttributes.put("OCEUSED", "#OFF");
		clkoutReferenceAttributes.put("TMUX", "#OFF");
		clkoutReferenceAttributes.put("TDDR_ALIGNMENT", "#OFF");
	}

	private boolean isClkout(Instance instance) {
		if (instance.getType() != PrimitiveType.OLOGIC2) {
			return false;
		}

		//Both clock pins connected to the same source?
		if (instance.getPin("CLK0").getNet() != instance.getPin("CLK1").getNet()) {
			return false;
		}

		return clkoutReferenceAttributes.keySet().stream()
				.allMatch(name -> {
					Attribute attr = instance.getAttribute(name);
					return attr.getValue().equals(clkoutReferenceAttributes.get(name));
				});
	}

}
