/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.module_synthesis;

import de.tu_darmstadt.rs.rapidSoc.LogMarkers;
import de.tu_darmstadt.rs.rapidSoc.directories.ModuleDirectories;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.Bus;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.buses.BusPort;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IPort;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IBusHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;
import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.NetType;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.design.Port;
import edu.byu.ece.rapidSmith.device.PrimitiveType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.function.Consumer;
import java.util.function.Supplier;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Creates modules from full xdl designs
 *
 * Created by jakobw on 26.10.15.
 */
public class ModuleExtractorThorsten extends ModuleExtractor {

	private static final Logger logger = LoggerFactory.getLogger(ModuleExtractorThorsten.class);

	/**
	 * Create ModuleExtractor
	 *
	 * @param design the design
	 * @param jmodules the jconfig modules
	 * @param document the document
	 */
	public ModuleExtractorThorsten(Design design, Collection<IModuleConfiguration<?>> jmodules, IConfigurationDocument document) {
		super(design, jmodules, document);
	}

	/**
	 * Get the local name for a port connected to the given port
	 *
	 * @param remoteBusHandler the opposite bus handler
	 * @param remotePortName the opposite port name
	 * @return the local name
	 */
	public static String findLocalNameForRemoteBusPort(IBusHandler remoteBusHandler, String remotePortName) {

		IBusHandler localBusHandler = remoteBusHandler.streamOppositeConnectors().collect(StreamUtil.exactlyOne());

		Bus remoteBus = remoteBusHandler.getConnectorDescription();

		BusPort remotePort = remoteBusHandler
				.getBusDefinition()
				.getMatchingPorts(remoteBus)
				.getPort(remotePortName)
				.orElseThrow(() -> new RuntimeException(portSpace + remotePortName + " not found in " + remoteBusHandler.getPortName() + " in " + remoteBusHandler.getSourceElement().getName()));

		String mName = remoteBusHandler.getBusDefinition().getOppositePort(remoteBus, remotePort).getName();

		//Find name of port in module
		String localName = findLocalNameForBusPort(localBusHandler, mName);

		return findLocalNameForBusPort(localBusHandler, localName);
	}

	/**
	 * Generate a module
	 *
	 * @param name the module name
	 * @param noDisconnect disable disconnecting the module
	 * @param partialNets the set of nets crossing the module boundary
	 * @param isCore flag whether this is the core
	 * @return the created module
	 */
	public Module generateModule(String name, boolean noDisconnect, Set<Net> partialNets, boolean isCore) {
		Tracer.enterLoop("Find included instances");
		Module rmodule = new Module();
		rmodule.setName(name);

		//check for unplaced instances
		for (Instance instance : design.getInstances()) {
			if (!instance.isPlaced()) {
				throw new RuntimeException("This design is not fully placed.  Did PAR fail? " + instance);
			}
		}

		Set<Instance> instancesToMove = findIncludedInstances();
		if (instancesToMove.isEmpty()) {
			throw new RuntimeException("No instances found in " + name + "!");
		}
		Tracer.exit();

		Tracer.enterLoop("Analyze Nets");
		partialNets.clear();
		Set<Net> fullNets = new HashSet<>();
		findIncludedNets(instancesToMove, fullNets, partialNets);

		Tracer.exit();
		Tracer.enterLoop("Disconnect Instances");
		design.getInstances().removeAll(instancesToMove);
		design.getNets().removeAll(fullNets);

		Tracer.exit();

		Tracer.enterLoop("Generate Ports");

		String prefix = ModuleDirectories.getModuleDescriptionNameFromConf(jmodules) + '/';

		renameInstances(prefix, instancesToMove);
		instancesToMove.forEach(rmodule::addInstance);

		fullNets.forEach((net) -> {
			removePrefix(prefix, net::getName, net::setName);
			rmodule.addNet(net);
		});

		generatePorts(rmodule, partialNets, design.getInstances(), prefix, noDisconnect);
		generateAnchor(rmodule);

		Tracer.exit();


		Tracer.enterLoop("Disconnect Nets");
		if (!noDisconnect) {
			rmodule.disconnectDesign();
		} else {
			instancesToMove.forEach(design::addInstance);
			fullNets.forEach(design::addNet);
			partialNets.forEach(Net::unroute);
		}
		Tracer.exit();

		unplaceSpecialInstances(rmodule);

		unrouteFailNets(rmodule);

		passthroughNets(rmodule, isCore);

		//hier ist der FINISH Teil
		return rmodule;
	}

	private void passthroughNets(Module rmodule, boolean isCore) {
		//TODO these connections should be auto detected
		//TODO remove isCore parameter
		if (isCore) {
			Port in = rmodule.getPort("clk.I");
			Port out = new Port("bus_clk.O", null);
			rmodule.addPort(out);
			in.addPassthroughPort(out);
		}
	}

	private boolean removePrefix(String prefix, Supplier<String> nameGetter, Consumer<String> nameSetter) {
		String name = nameGetter.get();
		if (name.startsWith(prefix)) {
			String newName = name.substring(prefix.length());
			nameSetter.accept(newName);
			return true;
		} else {
			return false;
		}
	}

	private void renameInstances(String prefix, Set<Instance> instancesToMove) {
		instancesToMove.forEach(instance -> {
			removePrefix(prefix, instance::getName, instance::setName);

			instance.getAttributes().forEach(attr -> {
				removePrefix(prefix, attr::getLogicalName, attr::setLogicalName);
			});
		});
	}

	private void findIncludedNets(Set<Instance> instancesToMove, Set<Net> fullNets, Set<Net> partialNets) {
		design.getNets()
				.forEach(net -> {

					StreamUtil.AllOrAnyMatch res = checkNetInside(instancesToMove, net);

					if (res == StreamUtil.AllOrAnyMatch.SOME) {
						if (net.getType() != NetType.WIRE) {
							net.removePin(net.getSource());
							res = checkNetInside(instancesToMove, net);
						} else {
							partialNets.add(net);
						}
					}

					if (res == StreamUtil.AllOrAnyMatch.ALL) {
						fullNets.add(net);
					}
				});
	}

	private StreamUtil.AllOrAnyMatch checkNetInside(Set<Instance> instancesToMove, Net net) {
		return net.getPins().stream()
				.map(pin -> instancesToMove.contains(pin.getInstance()))
				.collect(StreamUtil.AllOrAnyMatch.collector())
				.orElse(StreamUtil.AllOrAnyMatch.NONE);
	}

	private Set<Instance> findIncludedInstances() {
		return design.getInstances().stream()
				.filter(InstanceInsideChecker::isInstanceIncluded)
				.collect(Collectors.toSet());
	}

	private void generatePorts(Module module, Set<Net> partialNets, Collection<Instance> outsideInsts, String prefix, boolean noDisconnect) {

		partialNets.forEach(net -> {
			Port port = new Port();

			Set<Pin> pinsToRemoveFromNet;

			//If the net's source is contained, we don't need to add any sinks
			if (module.getInstances().contains(net.getSource().getInstance())) {
				//logger.error("contains: " + net.getSource().getInstance());

				port.setPin(net.getSource());
				net.getSource().setPort(port);

				removePrefix(prefix, net::getName, net::setName);
				module.addNet(net);

				//Remove all pins that do not belong to the module
				pinsToRemoveFromNet = net.getPins().stream()
						.filter(pin -> !module.getInstances().contains(pin.getInstance()))
						.collect(Collectors.toSet());

				//logger.error("pinsToRemoveFromNet: " + pinsToRemoveFromNet);

			} else {
				/*net.getPins().forEach(pin -> {
					logger.error("Pin: " + pin);
					logger.error("Pin in Module? -> " + module.getInstances().contains(pin.getInstance()));
				});*/

				//Remove all pins that belong to the module
				pinsToRemoveFromNet = net.getPins().stream()
						.filter(pin -> module.getInstances().contains(pin.getInstance()))
						.collect(Collectors.toSet());

				//logger.error("pinsToRemoveFromNet: " + pinsToRemoveFromNet);

				//Add Pins to port
				pinsToRemoveFromNet.forEach(pin -> {
					port.addPin(pin);
					pin.setPort(port);
				});
			}

			if (port.getPins().isEmpty()) {
				throw new RuntimeException("empty port!");
			}

			generatePortName(port);

			if (net.getType() != NetType.WIRE) {
				port.getPins().clear();
				port.setNetType(net.getType());
				throw new RuntimeException("WTF WARUM CLEAR????");
			}

			if (!noDisconnect) {
				pinsToRemoveFromNet.forEach(pin -> {
					pin.getNet().removePin(pin);
				});
			}

			if (port.getName() != null) {
				module.addPort(port);
			} else {
				logger.warn(LogMarkers.rapidSoc, "discarding port because of null name");
			}
		});

		buildStaticOutputPorts(module, outsideInsts);
	}

	/**
	 * Statically sourced output ports are not contained in our partial nets. Find them and add them
	 * @param module
	 * @param outsideInsts
	 */
	private void buildStaticOutputPorts(Module module, Collection<Instance> outsideInsts) {

		final Set<IModuleConfiguration<?>> outModuleSet = DocumentUtil.findModulesByName(this.document, "outModule_exclude");

		if (outModuleSet.size()>1)
			throw new IllegalStateException("multiple outmodules???");

		if (outModuleSet.size()==0)
			return;

		IModuleConfiguration<?> outModule = outModuleSet.iterator().next();
		Map<Integer, String> missingPort = identifyMissingOutputs(module, outModule);

		//diese Ports fehlen, muss natürlich automatisch ermittelt werden...
		Port newPort = new Port();

		//Find statically sourced outputs
		outsideInsts.forEach(i -> {

			//Fehlende Port suchen und hinzufügen
			i.getNetList().forEach(net -> {
				if (net.getSource() != null) {
					Pattern myRegex = Pattern.compile("(outmodule_exclude.+\\[)(\\d+)(\\])");
					Matcher myMatcher = myRegex.matcher(net.getSource().getInstanceName());

					if (myMatcher.find()) {
						/*//Dumpen der Instanz
						PrimitiveDefList primitiveDefs = FileTools.loadPrimitiveDefs(design.getFamilyType());
						try {
							DotDumper.dumpInstanceOnce(net.getSource().getInstance(), primitiveDefs);
						} catch (FileNotFoundException e) {
							throw new RuntimeException(e);
						}*/

						String str = myMatcher.group(2);
						int tempPort = Integer.parseInt(str);

						if (missingPort.containsKey(tempPort)) {
							/*logger.error("---------- Start der Instance --------- ");
							logger.error(net.getSource().getInstance().getName());
							net.getSource().getInstance().getPins().forEach(p -> {
								logger.error("pin: " + p);

								if(p.getNet() != null) {
									if (p.getNet().getAttributes() != null) {
										p.getNet().getAttributes().forEach(attr -> {
											logger.error("Attributes: " + attr);
										});
									}
								}
							});
							logger.error("---------- Ende der Instance ----------");
							logger.error("");*/

							Port newPort2 = new Port();
							newPort2.setName(missingPort.get(tempPort));

							//Invertierung in den Attributen gefunden? Invertierung in XDL ist "~"
							if (net.getSource().getInstance().getAttributes().stream().anyMatch(a -> a.getValue().contains("~"))) {
								newPort2.setNetType(NetType.VCC);
								logger.debug(newPort2.getName() + " -> VCC");
							} else {
								newPort2.setNetType(NetType.GND);
								logger.debug(newPort2.getName() + " -> GND");
							}

							//Port hinzufügen
							module.addPort(newPort2);

							//Aus Liste der missingPort entfernen
							missingPort.remove(tempPort);
						}
					}
				}
			});

			if (i.getType().equals(PrimitiveType.TIEOFF) || InstanceInsideChecker.isStaticSourceInst(i)) {
				i.getNetList().forEach(net -> {
					Port port = new Port();

					port.addPin(net.getSource());

					generatePortName(port);

					//Remove pin because we don't want that explicit source
					port.getPins().clear();

					port.setNetType(net.getType());

					if (port.getName() != null) {
						module.addPort(port);
					}
				});
			}

		});
	}

	private Map<Integer, String> identifyMissingOutputs(Module module, IModuleConfiguration<?> outModule) {
		Map<Integer, String> missingPort = new HashMap<>();
		IPortHandler<? extends IPort> outmodulePortHandler = DocumentUtil.getPortHandler(outModule, "pin");

		final int[] missingIndex = {0};
		outmodulePortHandler.streamConnections().forEach(tempConnection -> {
			String localName = findNameFromConnection2(outmodulePortHandler, tempConnection, null);

			Pattern myRegex = Pattern.compile("(.+)(\\[\\d+\\])");
			Matcher myMatcher = myRegex.matcher(localName);
			if (myMatcher.find()) {
				localName = myMatcher.group(1);
			}

			if (((IConnection) tempConnection).getOppositeConnector(outmodulePortHandler).getConnectorWidth() == 1) {
				logger.debug("Prüfe " + localName + ".O  -->  " + module.getPort(localName + ".O"));
				if (module.getPort(localName + ".O") == null) {
					missingPort.put(missingIndex[0], localName + ".O");
				}
				missingIndex[0]++;
			} else {
				for (int i = 0; i < tempConnection.getOppositeConnector(outmodulePortHandler).getConnectorWidth(); i++) {
					logger.debug("Prüfe " + localName + "[" + i + "].O  -->  " + module.getPort(localName + "[" + i + "].O"));
					if (module.getPort(localName + "[" + i + "].O") == null) {
						missingPort.put(missingIndex[0], localName + "[" + i + "].O");
					}
					missingIndex[0]++;
				}

			}

		});
		return missingPort;
	}


	private String getNewName(String name, String prefix) {
		if (name.startsWith(prefix)) {
			return name.substring(prefix.length());
		} else {
			return name;
		}
	}

	private Stream<Instance> getIobs(Design design) {
		return design.getInstances().stream().filter(i -> i.getType() == PrimitiveType.IOB);
	}


}
