/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module_synthesis;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.util.StreamUtil;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.INamed;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.util.SetHelpers;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collection;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

public class ModuleExtractorVerifying extends ModuleExtractor {

	private static final Logger logger = LoggerFactory.getLogger(ModuleExtractorVerifying.class);

	private final Collection<ModuleExtractor> childExtractors;

	public ModuleExtractorVerifying(Design design, Collection<IModuleConfiguration<?>> jmodules, IConfigurationDocument document, Path xdl) {
		super(design, jmodules, document);
		childExtractors = Arrays.asList(
				new ModuleExtractorNew(new Design(xdl), jmodules, document),
				new ModuleExtractorTesting(new Design(xdl), jmodules, document)
				//new ModuleExtractorNew(design, jmodules, document)
		);
	}


	private <T extends INamed> void explainDifferences(String name, Module a, Module b, Function<Module, Collection<T>> getter) {
		Map<String, T> aMap = collectNamed(getter.apply(a));
		Map<String, T> bMap = collectNamed(getter.apply(b));

		if (!aMap.keySet().equals(bMap.keySet())) {

			String[] aKeys = aMap.keySet().stream().sorted().toArray(String[]::new);
			String[] bKeys = bMap.keySet().stream().sorted().toArray(String[]::new);
			logger.info("Difference in " + name + ": " + SetHelpers.dumpVs(aKeys, bKeys));
		}

		aMap.keySet().stream().filter(bMap::containsKey).sorted().forEach(s -> {
			T ae = aMap.get(s);
			T be = bMap.get(s);
			if (!ae.equals(be)) {
				logger.info(s + ": " + ae + " / " + be);
			}
		});

	}

	private <T extends INamed> Map<String, T> collectNamed(Collection<T> elements) {
		return elements.stream().distinct().collect(Collectors.toMap(INamed::getName, Function.identity()));
	}

	private void moduleEqualsExplanation(Module a, Module b) {
		if (Float.compare(a.getMinClkPeriod(), b.getMinClkPeriod()) != 0) {
			logger.info("getMinClkPeriod() not equal");
		}
		if (!a.getName().equals(b.getName())) {
			logger.info("getName() not equal");
		}
		if (!a.getAttributes().equals(b.getAttributes())) {
			logger.info("getAttributes() not equal");
		}
		if (!a.getAnchor().equals(b.getAnchor())) {
			logger.info("getAnchor() not equal");
		}

		explainDifferences("Ports", a, b, Module::getPorts);
		explainDifferences("Instances", a, b, Module::getInstances);
		explainDifferences("Nets", a, b, Module::getNets);

		if (!Objects.equals(a.getMetaDataMap(), b.getMetaDataMap())) {
			logger.info("getMetaDataMap() not equal");
		}
		if (!a.getAllValidPlacements().equals(b.getAllValidPlacements())) {
			logger.info("getAllValidPlacements() not equal");
		}


	}

	@Override
	public Module generateModule(String name, boolean noDisconnect, Set<Net> partialNets, boolean isCore) {
		StreamUtil.FirstAndRest<Module> modules = childExtractors.stream()
				.map(ex -> ex.generateModule(name, noDisconnect, partialNets, isCore))
				.collect(StreamUtil.collectFirstAndRest());

		Module template = modules.getFirst();

		for (Module module : modules.getRest()) {
			if (!module.equals(template)) {
				moduleEqualsExplanation(template, module);
				throw new RuntimeException("Module does not equal template");
			}
		}

		return template;
	}

}
