/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module_synthesis;

import de.tu_darmstadt.rs.process.ProcessUtil;
import de.tu_darmstadt.rs.rapidSoc.design.ModuleDesign;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.table.Column;
import de.tu_darmstadt.rs.util.table.TableFormat;
import de.tu_darmstadt.rs.util.table.TableFormatter;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import edu.byu.ece.rapidSmith.design.Attribute;
import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.ModuleInstance;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.NetType;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.design.Port;
import edu.byu.ece.rapidSmith.util.FileConverter;
import edu.byu.ece.rapidSmith.util.UnrouteNets;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Spliterator;
import java.util.Spliterators;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

public class ModuleToVerilog {

	private static final Logger logger = LoggerFactory.getLogger(ModuleToVerilog.class);
	private final Design originaldesign;
	private final Map<IModuleConfiguration<?>, ModuleInstance> moduleMap;


	public ModuleToVerilog(Design design, Map<IModuleConfiguration<?>, ModuleInstance> moduleMap) {

		this.originaldesign = design;
		this.moduleMap = moduleMap;
	}

	private Function<String, String> mapPortName(int i) {
		return s -> s + "___" + i;
	}

	private Stream<Port> makeSingularPort(Port port) {
		if (port.getPins().size() == 1) {
			return Stream.of(port);
		}

		if (port.getPassthroughFrom() != null) {
			logger.warn("Dropping passthrough port {}", port.getName());
			return Stream.empty();
		}

		if (port.getPins().isEmpty()) {
			logger.warn("Dropping port without pins {}", port.getName());
			return Stream.empty();
		}

		List<Pin> pins = new ArrayList<>(port.getPins());


		return IntStream.range(0, port.getPins().size()).mapToObj(i -> {
			Function<String, String> nameMapper = mapPortName(i);
			Port p = new Port(nameMapper.apply(port.getName()), pins.get(i));
			p.getPin().setPort(p);
			p.setAliases(port.getAliases().stream().map(nameMapper).collect(Collectors.toList()));
			return p;
		});


	}

	/**
	 * Our ports have multiple pins, which is not representable in xdl --> make singular ports
	 */
	private void makeSingularPorts(Module module) {
		logger.info("Singular ports");
		final List<Port> newPorts = module.getPorts().stream().flatMap(this::makeSingularPort).collect(Collectors.toList());
		for (Port port : newPorts) {
			port.getPin();
			if (port.getPins().size() != 1) {
				throw new RuntimeException(port.getName());
			}
		}
		module.setPorts(newPorts);
		logger.info("Singular ports finished");
	}

	private void makeXdl(ModuleInstance modinst, Path xdlFile) {

		Design design = modinst.getDesign();
		for (Port port : modinst.getModule().getPorts()) {
			Net net = new Net("Net_for_" + port.getName(), NetType.WIRE);
			final boolean[] needed = {false};
			port.streamModuleInstancePins(modinst).forEach(pin -> {
				if (pin.getNet() == null) {
					net.addPin(pin);
					needed[0] = true;
				}
			});
			if (needed[0]) {
				design.addNet(net);
			}
		}

		design.setNets(UnrouteNets.combineStaticNets(design.getNets()));
		ModuleDesign.fixShapes(design);

		design.flattenDesign();

		design.saveXDLFile(xdlFile);
	}

	private void convertNCD2Verilog(Path ncdFile, Path verilogFile) {
		try {
			String[] command = {"netgen", "-ofmt", "verilog", "-w", ncdFile.toAbsolutePath().toString(), verilogFile.toAbsolutePath().toString()};
			logger.info("converting ncd 2 verilog: {}", (Object) command);
			ProcessBuilder pb = new ProcessBuilder(command);
			final int exitCode = ProcessUtil.runAndWaitFor(pb, (Consumer<String>) System.out::println);
			if (exitCode != 0) {
				throw new RuntimeException("Exit code = "+exitCode);
			}
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	enum VerilogMangleState {
		BeforeModule,
		InPorts,
		AfterPorts,
		AfterWires
	}

	private Stream<String> insertCommas(Stream<String> stream) {
		Iterator<String> iterator = stream.iterator();
		Spliterator<String> spliterator = new Spliterators.AbstractSpliterator<String>(Long.MAX_VALUE, 0) {
			@Override
			public boolean tryAdvance(Consumer<? super String> action) {
				if (iterator.hasNext()) {
					String item = iterator.next();
					boolean last = !iterator.hasNext();

					if (last) {
						action.accept(item);
					} else {
						action.accept(item + ",");
					}
					return true;
				}
				return false;
			}
		};
		return StreamSupport.stream(spliterator, false);
	}

	private static Pattern portPattern = Pattern.compile("(\\w+)(?:\\[([0-9]+)\\])?\\.([IOT])");


	static class PortInfo {

		final Port port;

		final int index;
		final String baseName;
		final char direction;

		public PortInfo(Port port) {

			this.port = port;

			final Matcher matcher = portPattern.matcher(port.getName());
			if (!matcher.matches()) {
				throw new RuntimeException("Does not match pattern: " + port.getName());
			}

			baseName = matcher.group(1);
			final String indexString = matcher.group(2);
			if (indexString != null) {
				index = Integer.parseInt(indexString);
			} else {
				index = 0;
			}
			direction = matcher.group(3).charAt(0);
		}

	}

	private Stream<String> genPorts(Map<String, List<PortInfo>> portInfo) {
		final Stream<String> stream = portInfo.keySet().stream().sorted().map(baseName -> {
			List<PortInfo> ports = portInfo.get(baseName);
			final Character direction = ports.stream().map(p -> p.direction).collect(StreamUtil.atLeastOneOrEqual());
			int maxIdx = ports.stream().mapToInt(p -> p.index).max().orElseThrow(RuntimeException::new);

			String directionStr;
			switch (direction) {
				case 'I':
					directionStr = "input";
					break;
				case 'O':
					directionStr = "output";
					break;
				default:
					throw new RuntimeException("Unknown direction: " + direction);
			}
			if (maxIdx == 0) {
				return directionStr + " wire " + baseName;
			} else {
				return directionStr + " wire [" + maxIdx + ":0] " + baseName;
			}
		});
		return insertCommas(stream.map(s -> "  " + s));
	}

	private Stream<String> genAssigns(ModuleInstance modinst, Map<String, List<PortInfo>> portInfo) {
		return portInfo.keySet().stream().sorted().flatMap(baseName -> {
			List<PortInfo> ports = portInfo.get(baseName);
			final Character direction = ports.stream().map(p -> p.direction).collect(StreamUtil.atLeastOneOrEqual());
			int maxIdx = ports.stream().mapToInt(p -> p.index).max().orElseThrow(RuntimeException::new);

			return ports.stream().sorted(Comparator.comparing(p -> p.index)).flatMap(port -> {
				String a, b;

				String outsideName;
				if (maxIdx == 0) {
					outsideName = baseName;
				} else {
					outsideName = baseName + "[" + port.index + "]";
				}

				String insideName;

				if (port.port.getPassthroughFrom() != null) {
					insideName = "\\"+port.port.getPassthroughFrom().streamModuleInstancePins(modinst).map(Pin::getNet).collect(StreamUtil.atLeastOneOrEqual()).getName();
				} else {
					if (port.port.getNetType()==NetType.WIRE && port.port.getPins().isEmpty()) {
						logger.warn("No pins for " + port.port);
						return Stream.empty();
					}

					switch (port.port.getNetType()) {
						case VCC: insideName = "1'b1";
							break;
						case GND: insideName = "1'b0";
							break;
						case WIRE: try {
							if (port.port.getPins().isEmpty()) {
								logger.warn("No pins for " + port.port);
								//return Stream.empty();
								throw new RuntimeException("no pins");
							}
							insideName = "\\"+port.port.streamModuleInstancePins(modinst).map(Pin::getNet).collect(StreamUtil.atLeastOneOrEqual()).getName();
							break;
						} catch (IllegalStateException e) {
							throw new RuntimeException(e);
						}
						default:
							throw new RuntimeException("unknown wire type: "+port.port.getNetType());
					}
				}

				if (port.direction == 'O') {
					a = outsideName;
					b = insideName;
				} else {
					a = insideName;
					b = outsideName;
				}
				return Stream.of("assign " + a + " = " + b + " ;");
			});
		}).map(s -> "  " + s);
	}

	private Function<String, Stream<String>> mangleVerilogLine(ModuleInstance modinst) {
		final Map<String, List<PortInfo>> portInfo = modinst.getModule().getPorts().stream().map(PortInfo::new).collect(Collectors.groupingBy(p -> p.baseName));
		final VerilogMangleState[] state = {VerilogMangleState.BeforeModule};
		return line -> {
			switch (state[0]) {
				case BeforeModule:
					if (line.startsWith("module ")) {
						state[0] = VerilogMangleState.InPorts;
					}
					return Stream.of(line);
				case InPorts:
					state[0] = VerilogMangleState.AfterPorts;
					return Stream.concat(genPorts(portInfo), Stream.of(line));
				case AfterPorts:
					if (line.trim().startsWith("initial")) {
						state[0] = VerilogMangleState.AfterWires;
						//return Stream.concat(genAssigns(portInfo), Stream.of(line));
						//Leave out import of sdf file
						return genAssigns(modinst, portInfo);
					} else {
						return Stream.of(line);
					}
				case AfterWires:
					return Stream.of(line);
			}
			throw new RuntimeException("Unknown state: " + state[0]);
		};
	}

	private void mangleVerilog(ModuleInstance modinst, Path verilogInput, Path verilogOutput) {
		try {
			final Stream<String> lines = Files.lines(verilogInput)
					.flatMap(mangleVerilogLine(modinst));
			Files.write(verilogOutput, (Iterable<String>) lines::iterator);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	public boolean toVerilog(ModuleInstance modinst, Path path) {
		try {
			String baseName = VerilogLang.getVerilogName(modinst.getName());
			final Path tempFolder = path.resolve("temp");
			Files.createDirectories(tempFolder);
			final Path xdlFile = tempFolder.resolve(baseName + ".xdl");
			final Path ncdFile = tempFolder.resolve(baseName + ".ncd");
			final Path verilogFile = tempFolder.resolve(baseName + ".v");
			final Path verilogFinalFile = path.resolve(baseName + ".v");

			try (Tracer ignored = Tracer.enterLoop("save xdl")) {
				makeXdl(modinst, xdlFile);
			}
			try (Tracer ignored = Tracer.enterLoop("convert to ncd")) {
				logger.info("Converting {} to {}",xdlFile, ncdFile);
				FileConverter.convertXDL2NCD(xdlFile, ncdFile);
			}
			try (Tracer ignored = Tracer.enterLoop("convert to verilog")) {
				convertNCD2Verilog(ncdFile, verilogFile);
			}
			try (Tracer ignored = Tracer.enterLoop("generate verilog ports")) {
				mangleVerilog(modinst, verilogFile, verilogFinalFile);
			}
			return true;
		} catch (RuntimeException | IOException e) {
			e.printStackTrace();
			return false;
		}
	}

	public void toVerilog(Path directory) {
		Path p = directory.resolve("rapidSocSimulation");
		try {
			Files.createDirectories(p);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

		TableFormatter table = new TableFormatter(TableFormat.CHARACTERS, Column.STRING, new Column("%b", Column.Alignment.LEFT));
		table.addHeaderRow("Module", "Conversion ok?");

		moduleMap.values().stream().distinct() .forEach(modinst -> {
			ModuleInstance clonedInstance = createClonedInstance(modinst);
			boolean ok = toVerilog(clonedInstance, p);
			table.addContentRow(modinst.getName(), ok);
		});
		table.output().forEach(logger::info);
	}

	private ModuleInstance createClonedInstance(ModuleInstance modinst) {
		Module module = modinst.getModule().clone();
		Design design = new Design(modinst.getName(), this.originaldesign.getPartName());
		design.addModule(module);
		ModuleInstance clonedInstance = design.createModuleInstance(modinst.getName(), module);


		Map<String, Instance> origInstancesByName = modinst.getInstances().stream().collect(Collectors.toMap(Instance::getName, Function.identity()));
		clonedInstance.getInstances().forEach(inst -> {
			Instance origInstance = origInstancesByName.get(inst.getName());
			inst.setAttributes(origInstance.getAttributes().stream().map(Attribute::clone).collect(Collectors.toMap(Attribute::getName, Function.identity())));
		});

		clonedInstance.place(modinst.getAnchor().getPrimitiveSite(), design.getDevice(), (x, y) -> { });
		return clonedInstance;
	}

}
