/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module_synthesis.constraints;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;
import de.tu_darmstadt.rs.rapidSoc.RapidSocProject;
import de.tu_darmstadt.rs.rapidSoc.directories.VariantPath;
import de.tu_darmstadt.rs.rapidSoc.visualization.BaseVisualization;
import de.tu_darmstadt.rs.rapidSoc.visualization.VisualizationFrame;

import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.PrimitiveType;
import edu.byu.ece.rapidSmith.device.Tile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collector;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public abstract class RectangularConstraintCreator extends ConstraintCreator {
	private static final Logger logger = LoggerFactory.getLogger(RectangularConstraintCreator.class);
	private Device device;

	protected RectangularConstraintCreator(VariantPath projectPath, RapidSocProject project) {
		super(projectPath, project);
		device = getProject().getDevice();
	}

	private IntStream streamColumns(BoundingBox bb) {
		final int count = bb.maxX - bb.minX + 1;
		return IntStream.range(0, count).map(i -> bb.maxX - i);
	}

	protected String constrainType(PrimitiveType type, BoundingBox constraintCorners) {
		//First, we separate the boundingBox into columns
		final List<BoundingBox> boxes = streamColumns(constraintCorners).mapToObj(x -> new BoundingBox(x, constraintCorners.minY, x, constraintCorners.maxY))
				//Shrink Y range to only contain actually existing tiles of correct type
				.flatMap(bb -> shrinkYRange(type, bb))

				//Combine neighbouring bbs
				.collect(bbColumnCollector(type));
		//visualizeExpectation(constraintCorners, boxes);
		return boxes.stream()
				.map(box -> getBoxDescription(type, box))
				.collect(Collectors.joining(", "));
	}

	private String getSiteName(PrimitiveType type, Tile tile, boolean higherName) {
		if (tile.getPrimitiveSites() == null) {
			return null;
		}

		final Stream<String> nameStream = Arrays.stream(tile.getPrimitiveSites())
				.filter(ps -> ps.isCompatiblePrimitiveType(type))
				.map(PrimitiveSite::getName);

		final Stream<String> sortedNames;
		if (higherName) {
			sortedNames = nameStream.sorted(Comparator.reverseOrder());
		} else {
			sortedNames = nameStream.sorted();
		}

		return sortedNames.findFirst().orElse(null);
	}

	private String getSiteName(PrimitiveType type, int x, int y, boolean higherName) {
		Tile tile = device.getTile(y, x);
		return getSiteName(type, tile, higherName);
	}

	private String getBoxDescription(PrimitiveType type, BoundingBox bb) {
		String lowerLeft = getSiteName(type, bb.minX, bb.maxY, false);
		String upperRight = getSiteName(type, bb.maxX, bb.minY, true);
		if (lowerLeft != null && upperRight != null) {
			return lowerLeft + ":" + upperRight;
		}

		String upperLeft = getSiteName(type, bb.minX, bb.minY, false);
		String lowerRight = getSiteName(type, bb.maxX, bb.maxY, true);
		return upperLeft + ":" + lowerRight;
	}

	private boolean isDescribable(PrimitiveType type, BoundingBox bb) {
		boolean upperLeft = matches(type, bb.minX, bb.minY);
		boolean upperRight = matches(type, bb.maxX, bb.minY);
		boolean lowerLeft = matches(type, bb.minX, bb.maxY);
		boolean lowerRight = matches(type, bb.maxX, bb.maxY);

		return (upperLeft && lowerRight) || (lowerLeft && upperRight);
	}

	private Collector<BoundingBox, ?, List<BoundingBox>> bbColumnCollector(PrimitiveType type) {
		return Collectors.collectingAndThen(Collectors.toList(), list -> {
			//Try to combine neighbouring elements
			boolean changes;
			do {
				changes = false;
				int i = 0;
				while (i < list.size() - 1) {
					BoundingBox current = list.get(i);
					BoundingBox next = list.get(i + 1);
					BoundingBox union = current.union(next);
					if (isDescribable(type, union)) {
						list.remove(i + 1);
						list.remove(i);
						list.add(i, union);
						changes = true;
						//Do not increment, so we can try to combine resulting union with new next element
					} else {
						i++;
					}
				}
			} while (changes);
			return list;
		});
	}

	private boolean matches(PrimitiveType type, int x, int y) {
		final Tile tile = device.getTile(y, x);
		if (tile == null) {
			return false;
		}
		if (tile.getPrimitiveSites() == null) {
			return false;
		}
		return Stream.of(tile.getPrimitiveSites()).anyMatch(ps -> ps.isCompatiblePrimitiveType(type));
	}

	private Stream<BoundingBox> shrinkYRange(PrimitiveType type, BoundingBox bb) {
		while (!matches(type, bb.minX, bb.minY) && bb.minY < device.getRows()) {
			bb = new BoundingBox(bb.minX, bb.minY + 1, bb.maxX, bb.maxY);
		}
		while (!matches(type, bb.minX, bb.maxY) && bb.maxY >= bb.minY) {
			bb = new BoundingBox(bb.minX, bb.minY, bb.maxX, bb.maxY - 1);
		}

		if (bb.minY > bb.maxY) {
			return Stream.empty();
		}
		return Stream.of(bb);
	}


	private void visualizeExpectation(BoundingBox constraintCorners, List<BoundingBox> actual) {
		Set<Tile> usedTiles = new HashSet<>();
		//List<BoundingBox> expectedBBs = parseExpected(expect, usedTiles);

		final BaseVisualization visu = new BaseVisualization(device, false, false) {

			@Override
			protected boolean isUsed(Tile tile) {
				return usedTiles.contains(tile) || true;
			}

			@Override
			protected Map<Tile, Set<Net>> calcRoutingOnlyTiles() {
				return null;
			}

			@Override
			public Instance getInstanceAtPrimitiveSite(PrimitiveSite primitiveSite) {
				return null;
			}

			@Override
			public void paint(Graphics2D g2d, int tileSize) {

				super.paint(g2d, tileSize);
				g2d.setColor(Color.RED);
				g2d.setStroke(new BasicStroke(3));
				drawBB(g2d, constraintCorners, tileSize, 3);

				g2d.setColor(Color.WHITE);
				/*for (BoundingBox bb : expectedBBs) {
					drawBB(g2d, bb);
				}*/
				g2d.setColor(Color.GREEN);
				for (BoundingBox bb : actual) {
					drawBB(g2d, bb, tileSize);
				}
			}
		};
		new VisualizationFrame<>(visu).openBlockingUI();
		Path dir = Paths.get("constraintRef");
		try {
			Files.createDirectories(dir);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}
/*
	private List<BoundingBox> parseExpected(String expect, Set<Tile> usedTiles) {
		return Arrays.stream(expect.split(",")).map(rect -> {
			final String[] split = rect.trim().split(":");
			PrimitiveSite a = device.getPrimitiveSite(split[0]);
			PrimitiveSite b = device.getPrimitiveSite(split[1]);

			usedTiles.add(a.getTile());
			usedTiles.add(b.getTile());

			return new BoundingBox(a.getTile()).union(new BoundingBox(b.getTile()));
		}).collect(Collectors.toList());
	}*/

	@Override
	protected Stream<String> getFullConstraints(ResourceUsage usage) {
		final BoundingBox constraintCorners = getConstraintRectangle(usage);

		logger.info("Constraint rectangle: "+constraintCorners);

		final Stream<String> boxes = Arrays.stream(new PrimitiveType[]{
				PrimitiveType.SLICEX,
				PrimitiveType.DSP48A1,
				PrimitiveType.RAMB8BWER,
				PrimitiveType.RAMB16BWER
		})
				.map(pt -> constrainType(pt, constraintCorners))
				.filter(s -> !s.isEmpty())
				.map(s -> "AREA_GROUP \"pblock\" RANGE = " + s+";");


		return Stream.concat(
				Stream.of("INST \"*\" AREA_GROUP = \"pblock\";"),
				boxes
		);
	}

	/**
	 * Get Constraints as BoundingBox of tile's row and column
	 *
	 * @param usage
	 * @return
	 */
	protected abstract BoundingBox getConstraintRectangle(ResourceUsage usage);
}
