/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.parameters;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.ModuleInstance;
import edu.byu.ece.rapidSmith.design.Pin;

import java.io.File;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

/**
 * Allows to change the value an Equality tester checks against
 *
 * This is abstract, methods to find the specific instance that does the checking have to be implemented in extending
 * classes
 */
public abstract class EqTester extends ParameterReplace {

	Pattern bitNumberPattern = Pattern.compile(".*\\/([a-z_A-Z0-9]+)\\[(\\d+)\\](.[IOT])?$");

	@Override
	public void replaceParameter(Design design, IModuleConfiguration jModule, ModuleInstance rModule, File projectPath, IBuildProcess buildProcessCache) {
		Instance matchInstance = getMatchInstance(jModule, rModule);

		long testValue = getTestValue(jModule, buildProcessCache);

		matchInstance.getAttributes().stream().filter(a -> a.getPhysicalName().endsWith("LUT")).forEach(lut -> {
			if (lut.getValue().equals("#OFF")) {
				return;
			}

			char lutName = lut.getPhysicalName().charAt(0);
			int lutSize = lut.getPhysicalName().charAt(1) - '0';
			final int supportedLutSize = 6;
			if (lutSize != supportedLutSize) {
				throw new RuntimeException("only 6-input luts supported, but we got " + lutSize + " " + lut + " in " + matchInstance);
			}


			MatchOption[] matchOptions = new MatchOption[lutSize];
			for (int i = 1; i <= lutSize; i++) {
				String pinName = lutName + String.valueOf(i);
				Pin p = matchInstance.getPin(pinName);
				MatchOption opt;
				if (p == null) {
					opt = MatchOption.DONTCARE;
				} else {
					int bit = getBitNumber(p);
					if (bit == -1) {
						opt = MatchOption.REGULAR;
					} else if (((testValue >> bit) & 1) != 0) {
						opt = MatchOption.REGULAR;
					} else {
						opt = MatchOption.INVERT;
					}
				}
				matchOptions[i - 1] = opt;
			}

			String function = genLutFunction(matchOptions);
			lut.setValue(function);
		});
	}

	protected String genLutFunction(MatchOption[] matchOptions) {
		//IntStream for some reason does not have flatMapToObj, so map to obj first...
		return "#LUT:O6=" + IntStream.rangeClosed(1, matchOptions.length).boxed().flatMap(i -> {
			MatchOption m = matchOptions[i - 1];

			if (m == MatchOption.DONTCARE) {
				return Stream.empty();
			}

			String name = "A" + i;
			return Stream.of(m.prefix + name);
		}).collect(Collectors.joining("*"));
	}

	/**
	 * Get the bit that a pin represents
	 *
	 * @param pin the pin to analyze
	 * @return the bit. -1 if it is an additional signal to test (eg. enable). those will never be inverted
	 */
	protected int getBitNumber(Pin pin) {
		if (pin == null) {
			throw new RuntimeException("null");
		}
		if (pin.getNet() == null) {
			throw new RuntimeException("net null for " + pin);
		}

		String name = pin.getNet().getName();

		Matcher m = bitNumberPattern.matcher(name);
		if (!m.matches()) {
			return -1;
		} else {
			String netBaseName = m.group(1);
			int offs = getBusOffset(netBaseName);
			return offs + Integer.parseInt(m.group(2));
		}
	}

	/**
	 * The new value the instance should test against
	 * @param jModule the module
	 * @return the value to test against
	 */
	protected abstract long getTestValue(IModuleConfiguration jModule, IBuildProcess buildProcessCache);

	/**
	 * The Rapidsmith instance that is doing the testing
	 * @param jModule the jc module
	 * @param rModule the rapidsmith module
	 * @return the testing instance
	 */
	protected abstract Instance getMatchInstance(IModuleConfiguration jModule, ModuleInstance rModule);

	/**
	 * Get the offset a bus has
	 * @param name The bus name
	 * @return the offset
	 */
	protected abstract int getBusOffset(String name);

	/**
	 * Options in how a bit may be used in matching in equality testers
	 */
	private enum MatchOption {
		REGULAR(""),
		INVERT("~"),
		DONTCARE("X");

		String prefix;

		MatchOption(String prefix) {
			this.prefix = prefix;
		}
	}
}
