/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.parameters;

import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.initram.MemorySource;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.XilinxBram;
import de.tu_darmstadt.rs.spartanmc.initram.bramLayout.XilinxBramBuilderFactory;
import de.tu_darmstadt.rs.spartanmc.initram.parameter.MemoryMapConverter;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.documentBuilder.IBuildProcess;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.interchange.memory.MemoryMap;
import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import edu.byu.ece.rapidSmith.design.Attribute;
import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.ModuleInstance;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * Replaces the ram of a processor
 */
public class ReplaceRam extends ParameterReplace {
	private static final Logger logger = LoggerFactory.getLogger(ReplaceRam.class);

	public ReplaceRam() {
	}

	@Override
	public boolean hasChangeableParameter(IModuleConfiguration module) {
		return module.getModuleDescription().getType() == IModuleDescription.ModuleType.Subsystem;
	}

	@Override
	public void replaceParameter(Design design, IModuleConfiguration jModule, ModuleInstance rModule, File projectPath, IBuildProcess buildProcessCache) {
	}


	@Override
	public boolean canChangeAttribute(IModuleDescription md, IParameter parameter) {
		//Ram is not modeled as attribute, so return false
		return false;
	}

	/**
	 * Replace the ram
	 * @param design the design
	 * @param projectPath the project's path
	 * @param moduleMap the module map
	 */
	public static void replace(Design design, File projectPath, Map<IModuleConfiguration<?>, ModuleInstance> moduleMap) {
		Tracer.enterLoop("Replace RAM");
		MemoryMap map = new MemoryMapConverter().convert(new File(projectPath, "system/memory.xml").toString());
		MemorySource memorySource = new MemorySource(map, projectPath.toPath(), new XilinxBramBuilderFactory());
		List<XilinxBram> initData = memorySource.streamAllBRams().map(bram -> (XilinxBram) bram).collect(Collectors.toList());

		initData.forEach(bram -> {

			//Since processor and mem are often joined in a single module, we can't match by module name

			//Find the jconfig module's name
			final String[] instanceHierarchy = bram.getInstanceName().split("/");
			final String moduleName = instanceHierarchy[0];

			//Find the jconfig module
			final IModuleConfiguration<?> jModule = moduleMap.keySet().stream().filter(m -> VerilogLang.getVerilogName(m.getGlobalName()).equals(moduleName)).collect(StreamUtil.exactlyOne());
			//Get the rapidsmith module
			final ModuleInstance moduleInstance = moduleMap.get(jModule);

			//Get the name suffix
			final String nameInModule = bram.getInstanceName().substring(instanceHierarchy[0].length());

			//Search for it
			final Instance inst = moduleInstance.getInstances().stream().filter(i -> i.getName().endsWith(nameInModule)).collect(StreamUtil.exactlyOne());

			bram.streamLines().forEach(d -> {
				Attribute a = inst.getAttribute(d.getName());
				if (a == null) {
					throw new RuntimeException("attribute " + d.getName() + " not found on " + inst.getName());
				}
				a.setValue(String.format("%064X", d.getContents()));
			});
		});
		Tracer.exit();
	}

}
