/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.placer;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.PrimitiveType;
import edu.byu.ece.rapidSmith.device.Tile;
import edu.byu.ece.rapidSmith.router.Node;
import edu.byu.ece.rapidSmith.router.RoutingElement;
import edu.byu.ece.rapidSmith.timing.DesignWalker;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Place Clock Infrastructure
 */
@SuppressWarnings("checkstyle:magicNumber")
public class ClockInfrastructurePlacer {
	private static final Logger logger = LoggerFactory.getLogger(ClockInfrastructurePlacer.class);

	private final Design design;
	private final int maxFollowDepth = 10;
	private Map<Tile, Map<Node, PrimitiveSitePin>> pinsOfTile = new HashMap<>();
	private Instance loopInstance = new Instance("LOOP", null);
	private Pin loopPin = new Pin(true, "LOOP", loopInstance);

	public ClockInfrastructurePlacer(Design design) {
		this.design = design;
	}

	/**
	 * Main
	 * @param args arguments
	 */
	public static void main(String[] args) {
		Design design = new Design(Paths.get("test/uart/build/rapidSoc.xdl"));
		ClockInfrastructurePlacer clockInfrastructurePlacer = new ClockInfrastructurePlacer(design);
		clockInfrastructurePlacer.analyzeDevice();
		//clockInfrastructurePlacer.analyzeDesign();
	}

	/**
	 * Testing...
	 */
	public void analyzeDevice() {


		/*Design.iobTypes.stream()
			.flatMap(t-> {
				PrimitiveSite[] sites = design.getDevice().getAllPrimitiveSitesOfType(t);
				if (sites==null)
					return Stream.empty();
				return Arrays.stream(sites);
			})


			.filter(p->p.getName().equals("V10"))

			.forEach(this::analyzeSite);*/

		findConnectedSites(PrimitiveType.BUFIO2FB, PrimitiveType.DCM);
	}

	private String pad(int length) {
		StringBuilder pad = new StringBuilder(" ");

		StringBuilder result = new StringBuilder("");

		while (length > 0) {
			if ((length & 1) != 0) {
				result.append(pad);
			}

			pad.append(pad);
			length >>= 1;
		}

		return result.toString();

	}

	private Map<Node, PrimitiveSitePin> getPinsOfTile(Tile t) {
		return pinsOfTile.computeIfAbsent(t, this::computePinsOfTile);
	}

	private Map<Node, PrimitiveSitePin> computePinsOfTile(Tile tile) {
		if (tile.getPrimitiveSites() == null) {
			return new HashMap<>();
		}

		return Arrays.stream(tile.getPrimitiveSites())
				.flatMap(site -> site.getPins().keySet().stream().map(pin -> new PrimitiveSitePin(site, pin)))
				.collect(Collectors.toMap(PrimitiveSitePin::toNode, Function.identity()));
	}

	private Stream<PrimitiveSitePin> followRouting(Node n, int indent) {

		PrimitiveSitePin psp = getPinsOfTile(n.getTile()).get(n);
		if (psp != null && n.getLevel() > 0) {
			return Stream.of(psp);
		}


		if (indent > maxFollowDepth) {
			return Stream.empty();
		}

		if (n.getConnections() != null) {
			return Arrays.stream(n.getConnections()).flatMap(wireConnection -> {
				Node next = wireConnection.createNode(n);
				return followRouting(next, indent + 1);
			});
		} else {
			return Stream.empty();
		}
	}

	private Stream<PrimitiveSitePin> findReachableSites(PrimitiveSite start, String pin, int maxDepth) {
		Node startNode = new Node(start.getTile(), start.getExternalPinWireEnum(pin));
		return DesignWalker.<Node, PrimitiveSitePin>walkPathDepth(startNode,
			//isTerminal
			(node, depth) -> (depth > 0 && getPinsOfTile(node.getTile()).containsKey(node)),
			//Next elements
			(node, depth) -> {
				if (depth >= maxDepth) {
					return Stream.empty();
				}

				if (node.getConnections() == null) {
					return Stream.empty();
				} else {
					return Arrays.stream(node.getConnections()).map(conn -> conn.createNode(node));
				}
			},
			//Accumulator
			(partialResult, node) -> {
				if (partialResult == null) {
					return getPinsOfTile(node.getTile()).get(node);
				} else {
					return partialResult;
				}
			},
			//No reducer
			null,
			//Loop
			loop -> null);
	}

	private Stream<PrimitiveSitePin> findReachableSites(PrimitiveSite start, int maxDepth) {
		return start.getPins().keySet().stream().flatMap(pin -> findReachableSites(start, pin, maxDepth));
	}

	private Stream<PrimitiveSite> findReachableSitesOfType(PrimitiveSite start, PrimitiveType target) {
		return findReachableSites(start, 8)
				.map(x -> x.primitiveSite)
				.distinct()
				.filter(x -> x.getType() == target);

	}

	private void findConnectedSites(PrimitiveType start, PrimitiveType target) {
		for (PrimitiveSite primitiveSite : design.getDevice().getAllPrimitiveSitesOfType(start)) {
			logger.info("{}",primitiveSite);
			findReachableSitesOfType(primitiveSite, target).forEach(x -> logger.info("    " + x));
		}
	}

	private void analyzeDesign() {
		Pin pin = design.getInstance("CLOCK_USER").getPin("I");
		pin.getNet().getPins().forEach(p -> {
			logger.info("{}",p);
		});
		logger.info("walk");
		Stream<PathNode<RoutingElement>> path = DesignWalker.walkPathDepth(pin,
			(re, depth) -> {
				PrimitiveType instType = re.getInstance().getType();
				return instType.toString().startsWith("SLICE") || instType.toString().startsWith("RAM") || depth > 9;
			}, (re, depth) -> {
				Pin p = (Pin) re;
				if (p.isOutPin()) {
					return ((Pin) re).getNet().getPins().stream().filter(x -> x != p);
				}

				return re.getInstance().getPins().stream().filter(x -> x.isOutPin());

			},
			(partialResult, current) -> {
				if (partialResult != null) {
					return new PathNode<>(Collections.singletonList(partialResult), current);
				}
				return new PathNode<>(Collections.emptyList(), current);
			},
			(a, b) -> {
				if (a.element != b.element) {
					throw new RuntimeException("Different current elements");
				}
				List<PathNode<RoutingElement>> r = new ArrayList<>(a.children);
				r.addAll(b.children);
				return new PathNode<>(r, a.element);
			},
			(x) -> new PathNode<RoutingElement>(Collections.emptyList(), loopPin)
		);

		path.forEach(p -> {
			logger.info(p.toString(x -> x.toString() + " " + x.getInstance().getType()));
		});

	}


	/**
	 * Pin of a primitive site
	 */
	private final class PrimitiveSitePin {

		public final PrimitiveSite primitiveSite;
		public final String pin;

		private PrimitiveSitePin(PrimitiveSite primitiveSite, String pin) {
			this.primitiveSite = primitiveSite;
			this.pin = pin;
		}

		public Node toNode() {
			return new Node(primitiveSite.getTile(), primitiveSite.getExternalPinWireEnum(pin));
		}

		@Override
		public String toString() {
			return "Pin " + pin + " of " + primitiveSite;
		}
	}


	/**
	 * Node in a path
	 * @param <T> element type
	 */
	public class PathNode<T> {

		public final Collection<PathNode<T>> children;
		public final T element;

		public PathNode(Collection<PathNode<T>> children, T element) {
			this.children = children;
			this.element = element;
		}

		private void toString(StringBuilder builder, Function<T, String> mapper, int level) {
			builder.append(pad(level * 4));
			builder.append(mapper.apply(element));
			builder.append('\n');

			children.forEach(c -> c.toString(builder, mapper, level + 1));
		}

		/**
		 * Make string with custom mapper
		 * @param mapper Map elements to string
		 * @return the string
		 */
		public String toString(Function<T, String> mapper) {
			StringBuilder builder = new StringBuilder();
			toString(builder, mapper, 0);
			return builder.toString();
		}

		@Override
		public String toString() {
			return toString(Object::toString);
		}
	}

}
