/*
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.placer;

import de.tu_darmstadt.rs.rapidSoc.LogMarkers;
import de.tu_darmstadt.rs.rapidSoc.design.PortNet;
import de.tu_darmstadt.rs.rapidSoc.design.VisualizeWhen;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionInstance;
import de.tu_darmstadt.rs.rapidSoc.module.ModulePlacement;
import de.tu_darmstadt.rs.rapidSoc.placer.cost.CostFunction;
import de.tu_darmstadt.rs.rapidSoc.placer.overlap.OverlapDetector;
import de.tu_darmstadt.rs.rapidSoc.visualization.PlacementVisualization;
import de.tu_darmstadt.rs.rapidSoc.visualization.VisualizationFrame;
import de.tu_darmstadt.rs.util.table.Column;
import de.tu_darmstadt.rs.util.table.TableFormat;
import de.tu_darmstadt.rs.util.table.TableFormatter;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.ModuleInstance;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.PrimitiveType;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Random;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * Place modules close together
 */
public class ModulePlacer {

	@NotNull
	private static final Logger logger = LoggerFactory.getLogger(ModulePlacer.class);
	@NotNull
	private static final String TRACER_APPLY_PLACEMENT = "Apply placement";
	@NotNull
	private static final String TRACER_PLACER = "Find placement";
	@NotNull
	private static final String TRACER_INITIAL = "Initial placement";
	@NotNull
	private static final String TRACER_IMPROVE = "Improve placement";
	@NotNull
	private static final String TRACER_INSTANTIATE = "Create and place module instances";

	@NotNull
	private final List<ProposedMoveListener> proposedMoveListeners = new ArrayList<>();
	@NotNull
	private final OverlapDetector overlapDetector;
	@NotNull
	private final CostFunction costFunction;

	@NotNull
	private final Design design;
	@NotNull
	private final Path projectPath;

	@NotNull
	private final PlacementVisualization visualization;
	@NotNull
	private final List<VisualizeWhen> visuFile;
	@NotNull
	private final List<VisualizeWhen> visuGUI;
	private int visNumber = 0;


	/**
	 * Create ModulePlacer
	 *
	 * @param design the design
	 * @param costFunction the cost function
	 * @param overlapDetector the overlap detector
	 * @param projectPath the project path
	 * @param visuFile when to render visualizations to write to file
	 * @param visuGUI when to render visualizations to show gui
	 */
	public ModulePlacer(
			@NotNull final Design design,
			@NotNull final CostFunction costFunction,
			@NotNull final OverlapDetector overlapDetector,
			@NotNull final Path projectPath,
			@Nullable final List<VisualizeWhen> visuFile,
			@Nullable final List<VisualizeWhen> visuGUI
	) {
		this.overlapDetector = Objects.requireNonNull(overlapDetector);
		this.costFunction = Objects.requireNonNull(costFunction);

		proposedMoveListeners.add(this.overlapDetector);
		proposedMoveListeners.add(this.costFunction);

		this.projectPath = Objects.requireNonNull(projectPath);
		this.design = Objects.requireNonNull(design);

		this.visuFile = Objects.isNull(visuFile)? Collections.emptyList() : visuFile;
		this.visuGUI = Objects.isNull(visuGUI)? Collections.emptyList() : visuGUI;

		visualization = new PlacementVisualization(design, false, true, true, true);
		proposedMoveListeners.add(visualization);
	}


	/**
	 * Place the design
	 *
	 * @param moduleDescriptionInstances module description instances
	 * @param allPortNets nets connecting different modules
	 * @param iterations how many attempts to improve the placement will be done
	 * @return final cost
	 */
	public double place(
			@NotNull Set<ModuleDescriptionInstance> moduleDescriptionInstances,
			@NotNull Set<PortNet> allPortNets,
			int iterations
	) {
		visualization.setHlNets(allPortNets);

		Set<PortNet> portNets = allPortNets.stream()
				.filter(PortNet::hasSinks)
				.collect(Collectors.toSet());

		proposedMoveListeners.forEach(l -> l.initialize(moduleDescriptionInstances, portNets));

		final long seed = System.nanoTime();
		logger.info(LogMarkers.placer, "seed = " + seed);
		@NotNull final Random random = new Random(seed);

		design.addComment("PlacementSeed: " + seed);

		printPlacementStatistics(moduleDescriptionInstances);

		Tracer.enter(TRACER_PLACER);

		//try placing modules with fewer possibilities first
		@NotNull final List<ModuleDescriptionInstance> placementsSorted = moduleDescriptionInstances.stream()
				.sorted(Comparator.comparingInt(p -> p.getDescription().getPlacements().size()))
				.collect(Collectors.toList());

		initialPlacement(placementsSorted, random);
		final double initialCost = costFunction.getCost();

		renderVisualisation(VisualizeWhen.AfterInitial);

		//don't attempt to change placement for modules with only one possibility
		@NotNull final List<ModuleDescriptionInstance> placementsFiltered = placementsSorted.stream()
				.filter(p -> p.getDescription().getPlacements().size() > 1)
				.collect(Collectors.toList());

		final double finalCost = improvePlacement(placementsFiltered, iterations, initialCost, random);

		Tracer.exit();

		renderVisualisation(VisualizeWhen.Final);

		return finalCost;
	}

	/**
	 * finds an initial placement using backtracking
	 *
	 * @param placements list of modules to place
	 * @param random a {@link Random} used to generate random changes
	 * @throws IllegalStateException if there is no valid initial placement
	 */
	private void initialPlacement(
			@NotNull List<ModuleDescriptionInstance> placements,
			@NotNull Random random
	) {
		Tracer.enter(TRACER_INITIAL);
		if (placements.stream()
				.filter(p -> p.getDescription().getPlacements().size() < 1)
				.peek(p -> logger.error("Module {} has no possible placements", p))
				.count() > 1
				) {
			throw new IllegalStateException("one or more modules are not placeable on the target device");
		}
		if (!initialPlaceModule(0, placements, random)) {
			throw new IllegalStateException("could not find initial placement");
		}
		Tracer.exit();
	}


	/**
	 * recursively places modules using backtracking
	 *
	 * @param moduleIdx offset to start from in moduleDescriptionInstances
	 * @param moduleDescriptionInstances all moduleDescriptionInstances to place
	 * @param random a {@link Random} used to generate random moduleDescriptionInstances
	 * @return whether or not the placement succeeded
	 */
	private boolean initialPlaceModule(
			int moduleIdx,
			@NotNull List<ModuleDescriptionInstance> moduleDescriptionInstances,
			@NotNull Random random
	) {

		@NotNull ModuleDescriptionInstance mdiToChange = Objects.requireNonNull(moduleDescriptionInstances.get(moduleIdx), "can not place null module");

		logger.debug(LogMarkers.placer, padding(moduleIdx * 2) + "trying to place " + mdiToChange);

		//we want a copy here, because we shuffle the list. Multiple module instances would interfere otherwise
		List<ModulePlacement> placements = new ArrayList<>(mdiToChange.getDescription().getPlacements());

		Collections.shuffle(placements, random);

		for (ModulePlacement newPlacement : placements) {

			overlapDetector.tryMove(mdiToChange, newPlacement);
			costFunction.tryMove(mdiToChange, newPlacement);

			if (!overlapDetector.isMoveOk()) {
				overlapDetector.moveRejected();
				//costFunction.moveRejected();
			} else {
				overlapDetector.moveAccepted();
				//costFunction.moveAccepted();

				logger.debug(LogMarkers.placer, padding(moduleIdx * 2) + "found placement for " + mdiToChange + " at " + newPlacement);

				Tracer.enterLoop(TRACER_APPLY_PLACEMENT);
				mdiToChange.setPlacement(newPlacement);
				Tracer.exit();

				if (moduleIdx >= moduleDescriptionInstances.size() - 1) {
					return true;
				}

				if (initialPlaceModule(moduleIdx + 1, moduleDescriptionInstances, random)) {
					return true;
				}

			}

		}


		overlapDetector.tryMove(mdiToChange, null);
		//costFunction.tryMove(instToPlace, null);

		if (!overlapDetector.isMoveOk()) {
			throw new RuntimeException("unplace should always be ok");
		}
		mdiToChange.setPlacement(null);

		logger.debug(LogMarkers.placer, padding(moduleIdx * 2) + "fail for " + mdiToChange);

		renderVisualisation(VisualizeWhen.BacktrackStep);

		return false;
	}


	/**
	 * improve module placement using a random walk
	 *
	 * @param moduleDescriptionInstances the module description instances we will try to change
	 * @param iterations how many moves will be done
	 * @param initialCost the cost of the initial placement
	 * @param random a {@link Random} used to generate random moves
	 * @return the cost
	 */
	private double improvePlacement(
			@NotNull final List<ModuleDescriptionInstance> moduleDescriptionInstances,
			final int iterations,
			final double initialCost,
			@NotNull final Random random
	) {
		Tracer.enter(TRACER_IMPROVE);
		double cost = initialCost;

		for (int i = 0; i < iterations; ++i) {
			logger.trace("Improve placement iteration {}", i);

			@NotNull ModuleDescriptionInstance mdiToChange = moduleDescriptionInstances.get(random.nextInt(moduleDescriptionInstances.size()));

			@NotNull List<ModulePlacement> possPlacements = mdiToChange.getDescription().getPlacements();
			@NotNull ModulePlacement newPlacement = possPlacements.get(random.nextInt(possPlacements.size()));

			proposedMoveListeners.forEach(l -> l.tryMove(mdiToChange, newPlacement));

			if (!overlapDetector.isMoveOk()) {
				proposedMoveListeners.forEach(ProposedMoveListener::moveRejected);
			} else {

				double newCost = costFunction.getCost();
				if (newCost < cost) {

					proposedMoveListeners.forEach(ProposedMoveListener::moveAccepted);
					cost = newCost;

					logger.info(LogMarkers.placer, "found placement with cost " + cost);
					Tracer.enterLoop(TRACER_APPLY_PLACEMENT);
					mdiToChange.setPlacement(newPlacement);
					Tracer.exit();

					renderVisualisation(VisualizeWhen.EachStep);

				} else {
					proposedMoveListeners.forEach(ProposedMoveListener::moveRejected);
				}
			}
		}
		return Tracer.exit(cost);
	}


	private static PrimitiveSite getNewAnchorSite(Module module, ModuleDescriptionInstance mdi, Design design) {
		@NotNull PrimitiveSite moduleAnchorSite = module.getAnchor().getPrimitiveSite();

		final PrimitiveSite newSite = Device.getCorrespondingPrimitiveSite(
				moduleAnchorSite,
				module.getAnchor().getType(),
				mdi.getPlacement().getAnchor());


		if (newSite != null) {
			return newSite;
		}

		final String availableSites = Arrays.stream(mdi.getPlacement().getAnchor().getPrimitiveSites()).map(site -> site + " (" + site.getType() + ")").collect(Collectors.joining(", "));

		logger.info("Did not find site for " + moduleAnchorSite + " ("+moduleAnchorSite.getType()+") in "+mdi.getName()+" at " + mdi.getPlacement().getAnchor()+". available: "+availableSites);

		PrimitiveType[] compatibleTypes = PrimitiveSite.compatibleTypesArray[design.getFamilyType().ordinal()].get(module.getAnchor().getType());
		if (compatibleTypes == null) {
			return null;
		}
		for (PrimitiveType compatibleType : compatibleTypes) {

			logger.info("trying compatible type " + compatibleType);
			final PrimitiveSite site = Device.getCorrespondingPrimitiveSite(
					moduleAnchorSite,
					compatibleType,
					mdi.getPlacement().getAnchor());

			if (site != null) {
				return site;
			}
		}
		return null;
	}

	/**
	 * instantiates and places rapidsmith modules
	 *
	 * @param design the design to place the modules in
	 * @param moduleDescriptionInstances the module description instances that should be properly instantiated
	 * @return a map from the moduleDescriptionInstances to the newly created module instances
	 */
	public static Map<ModuleDescriptionInstance, ModuleInstance> instantiateModules(
			@NotNull Design design,
			@NotNull Set<ModuleDescriptionInstance> moduleDescriptionInstances
	) {
		Tracer.enterLoop(TRACER_INSTANTIATE);
		@NotNull Device device = design.getDevice();
		if (moduleDescriptionInstances.stream()
				.filter(p -> !p.isPlaced())
				.peek(p -> logger.error("Module p is not placed"))
				.count() > 0
				) {
			throw new IllegalStateException("Can't instantiate unplaced modules");
		}

		return Tracer.exit(moduleDescriptionInstances.stream().collect(Collectors.toMap(
				Function.identity(),
				mdi -> {
					@NotNull Module module = mdi.getPlacement().getModuleDescription().getModule();
					@NotNull PrimitiveSite moduleAnchorSite = module.getAnchor().getPrimitiveSite();

					@NotNull ModuleInstance mi = design.createModuleInstance(mdi.getName(), module);
					final PrimitiveSite newSite = getNewAnchorSite(module, mdi, design);
					if (newSite == null) {
						throw new NullPointerException("New site is null!");
					}
					if (!mi.place(
							newSite,
							device,
							(a, b) -> {
							}
					)) {
						throw new IllegalStateException("Could not place module at " + newSite + ", but description said we could!");
					}
					return mi;
				}
		)));
	}


	/**
	 * generate a {@link String} of spaces
	 *
	 * @param length the amount of spaces
	 * @return the string
	 */
	@NotNull
	private String padding(int length) {
		@NotNull StringBuilder r = new StringBuilder();
		for (int i = 0; i < length; i++) {
			r.append(' ');
		}
		return r.toString();
	}


	private void printPlacementStatistics(Collection<ModuleDescriptionInstance> mdis) {
		final TableFormatter table = new TableFormatter(TableFormat.CHARACTERS,
				new Column("%s", Column.Alignment.LEFT),
				new Column("%s", Column.Alignment.LEFT),
				new Column("%s", Column.Alignment.LEFT),
				new Column("%s", Column.Alignment.LEFT)
		);
		table.addHeaderRow("Instance Name", "Hash", "Variants", "Placements");
		mdis.forEach(
				mdi -> table.addContentRow(mdi.getName(), mdi.getDescription().getHash(), mdi.getDescription().getVariants().size(), mdi.getDescription().getPlacements().size())
		);
		mdis.forEach(mdi -> mdi.getDescription().getPlacements().stream().map(ModulePlacement::toString).forEach(logger::trace));
		table.output().forEach(logger::info);
	}

	/**
	 * render visualizations if enabled for the current situation
	 *
	 * @param current the current situation
	 */
	private void renderVisualisation(VisualizeWhen current) {
		boolean file = visuFile.contains(current);
		boolean gui = visuGUI.contains(current);
		if (!file && !gui) {
			return;
		}

		Tracer.enterLoop("Visualisation");

		design.rebuildPrimitiveSitesUsed();

		if (gui) {
			new VisualizationFrame<>(visualization).openBlockingUI();
		}
		if (file) {
			visualization.saveImgNum(projectPath, visNumber++);
		}

		Tracer.exit();
	}
}
