/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.placer.cost;

import java.util.Arrays;
import java.util.Collection;
import java.util.Objects;
import java.util.Set;

import org.jetbrains.annotations.NotNull;

import de.tu_darmstadt.rs.rapidSoc.design.PortNet;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionInstance;
import de.tu_darmstadt.rs.rapidSoc.module.ModulePlacement;
import de.tu_darmstadt.rs.rapidSoc.placer.ProposedMoveListener;

/**
 * Combines multiple instances of {@link CostFunction} to form a single cost function.
 * How the costs are combined needs to be defined in overriding classes
 */
public abstract class CostFunctionCombiner implements CostFunction {

	@NotNull
	protected final CostFunction[] costFunctions;

	/**
	 * create a new meta-cost function
	 *
	 * @param costFunctions the child cost functions
	 */
	public CostFunctionCombiner(@NotNull CostFunction... costFunctions) {
		this.costFunctions = costFunctions;
		if (costFunctions.length < 1) {
			throw new IllegalArgumentException("CostFunctionCombiner requires at least one overlap detector");
		}
	}

	@Override
	public void tryMove(@NotNull ModuleDescriptionInstance mdiToChange, @NotNull ModulePlacement newPlacement) {
		for (ProposedMoveListener detector : costFunctions) {
			detector.tryMove(mdiToChange, newPlacement);
		}
	}

	@Override
	public void moveAccepted() {
		for (ProposedMoveListener pml : costFunctions) {
			pml.moveAccepted();
		}
	}

	@Override
	public void moveRejected() {
		for (ProposedMoveListener pml : costFunctions) {
			pml.moveRejected();
		}
	}

	@Override
	public void initialize(
			@NotNull Collection<ModuleDescriptionInstance> moduleDescriptionInstances,
			@NotNull Set<PortNet> portNets
	) {
		for (ProposedMoveListener pml : costFunctions) {
			pml.initialize(moduleDescriptionInstances, portNets);
		}
	}

	/**
	 * Meta-{@link CostFunction} that returns the highest cost returned by a child functions
	 */
	public static class Max extends CostFunctionCombiner {

		public Max(CostFunction... costFunctions) {
			super(costFunctions);
		}

		@Override
		public double getCost() {
			return Arrays.stream(costFunctions)
					.mapToDouble(CostFunction::getCost)
					.max().orElse(Double.MAX_VALUE);
		}
	}


	/**
	 * Meta-{@link CostFunction} that returns the lowest cost returned by a child functions
	 */
	public static class Min extends CostFunctionCombiner {

		public Min(CostFunction... costFunctions) {
			super(costFunctions);
		}

		@Override
		public double getCost() {
			return Arrays.stream(costFunctions)
					.mapToDouble(CostFunction::getCost)
					.min().orElse(Double.MAX_VALUE);
		}
	}


	/**
	 * Meta-{@link CostFunction} that returns the sum of all costs returned by child functions
	 */
	public static class Sum extends CostFunctionCombiner {

		public Sum(CostFunction... costFunctions) {
			super(costFunctions);
		}

		@Override
		public double getCost() {
			return Arrays.stream(costFunctions)
					.mapToDouble(CostFunction::getCost)
					.sum();
		}
	}


	/**
	 * Meta-{@link CostFunction} that requires all child costFunctions to return the same result
	 * Useful for testing multiple implementations of the same function only
	 */
	public static class RequireEqual extends CostFunctionCombiner {

		private final double tolerance;

		public RequireEqual(double tolerance, CostFunction... costFunctions) {
			super(costFunctions);
			this.tolerance = tolerance;
		}

		public RequireEqual(CostFunction... costFunctions) {
			this(0, costFunctions);
		}

		@Override
		public double getCost() {
			Double first = null;
			for (CostFunction costFunction : costFunctions) {
				double res = costFunction.getCost();
				if (Objects.isNull(first)) {
					first = res;
				} else if (Math.abs(first - res) > tolerance) {
					throw new IllegalStateException("the cost functions did not return equal results. first=" + first + ", current=" + res);
				}
			}
			Objects.requireNonNull(first);
			return first;
		}
	}
}
