/*
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.placer.overlap;

import java.util.Collection;
import java.util.Objects;
import java.util.Set;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.tu_darmstadt.rs.rapidSoc.design.PortNet;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionInstance;
import de.tu_darmstadt.rs.rapidSoc.module.ModulePlacement;
import de.tu_darmstadt.rs.rapidSoc.placer.ProposedMoveListener;


/**
 * Combines multiple instances of {@link OverlapDetector} to form a single detector.
 * Which detectors need to accept in order for the meta-detector to match needs to be defined in overriding classes
 *
 * Created by jwenzel on 04.01.16.
 */
public abstract class OverlapDetectorCombiner implements OverlapDetector {

	@NotNull
	private static final Logger logger = LoggerFactory.getLogger(OverlapDetectorCombiner.class);
	@NotNull
	protected final OverlapDetector[] detectors;

	/**
	 * create a new meta-cost function
	 *
	 * @param detectors the child detectors
	 */
	protected OverlapDetectorCombiner(@NotNull OverlapDetector... detectors) {
		Objects.requireNonNull(detectors, "detectors to combine may not be null");
		this.detectors = detectors;
		if (detectors.length<1) {
			throw new IllegalArgumentException("OverlapCombiner requires at least one overlap detector");
		}
	}

	@Override
	public void tryMove(@NotNull ModuleDescriptionInstance mdiToChange, @NotNull ModulePlacement newPlacement) {
		for (ProposedMoveListener detector : detectors) {
			detector.tryMove(mdiToChange, newPlacement);
		}
	}

	@Override
	public void moveAccepted() {
		for (ProposedMoveListener detector : detectors) {
			detector.moveAccepted();
		}
	}

	@Override
	public void moveRejected() {
		for (ProposedMoveListener detector : detectors) {
			detector.moveRejected();
		}
	}

	@Override
	public void initialize(
			@NotNull Collection<ModuleDescriptionInstance> moduleDescriptionInstances,
			@NotNull Set<PortNet> portNets
	) {
		for (ProposedMoveListener detector : detectors) {
			detector.initialize(moduleDescriptionInstances, portNets);
		}
	}

	/**
	 * Meta-Overlap detector that requires all child detectors to match
	 */
	public static class RequireAll extends OverlapDetectorCombiner {

		public RequireAll(@NotNull OverlapDetector... detectors) {
			super(detectors);
		}

		@Override
		public boolean isMoveOk() {
			for (OverlapDetector detector : detectors) {
				if (!detector.isMoveOk()) {
					return false;
				}
			}
			return true;
		}
	}


	/**
	 * Meta-Overlap detector that matches if any child detector matches
	 */
	public static class RequireAny extends OverlapDetectorCombiner {

		public RequireAny(@NotNull OverlapDetector... detectors) {
			super(detectors);
		}

		@Override
		public boolean isMoveOk() {
			for (OverlapDetector detector : detectors) {
				if (detector.isMoveOk()) {
					return true;
				}
			}
			return false;
		}
	}

	/**
	 * Meta-Overlap detector that requires all child detectors to return the same result
	 * Useful for testing multiple implementations of the same detector only
	 */
	public static class RequireEqual extends OverlapDetectorCombiner {

		public RequireEqual(@NotNull OverlapDetector... detectors) {
			super(detectors);
		}

		@Override
		public boolean isMoveOk() {
			Boolean first = null;
			for (OverlapDetector detector : detectors) {
				boolean res = detector.isMoveOk();
				if (Objects.isNull(first)) {
					first = res;
				} else if (first != res) {
					throw new IllegalStateException("the overlap costFunctions did not return equal results. first=" + first + ", current=" + res);
				}
			}
			Objects.requireNonNull(first);
			return first;
		}
	}
}
