/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.placer.overlap;

import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionInstance;
import de.tu_darmstadt.rs.rapidSoc.placer.ProposedMoveAdapter;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

/**
 * Implementation of {@link OverlapDetector}, checking whether any FPGA Tiles are used by multiply modules
 */
public class TileOverlapDetector extends ProposedMoveAdapter implements OverlapDetector {

	private final boolean includeRoutingTiles;

	/**
	 * construct a new {@link TileOverlapDetector}
	 * @param includeRoutingTiles whether or not tiles used by routing only should by considered
	 */
	public TileOverlapDetector(boolean includeRoutingTiles) {
		this.includeRoutingTiles = includeRoutingTiles;
	}

	@Override
	public boolean isMoveOk() {
		Tracer.enterLoop("Tile Overlap");
		// always accept un-place
		if (newPlacement == null) {
			return Tracer.exit(true);
		}
		Objects.requireNonNull(mdiToChange, "no move specified");

		//TODO is a set the best option here?
		Set<Integer> newTiles = newPlacement.streamTileNumbers(true, includeRoutingTiles).boxed().collect(Collectors.toSet());

		/*moduleDescriptionInstances.stream()
				//exclude the mdi we want to change
				.filter(mdi -> mdi != mdiToChange)
				//only placed modules
				.filter(ModuleDescriptionInstance::hasAnchor)
				//no collisions in any of these moduleDescriptionInstances
				.forEach(mdi -> mdi.streamTileNumbers(true, includeRoutingTiles)
						.filter(newTiles::contains)
						.forEach(i -> System.out.println("Collision with tile "+ mdi.getModuleDescription().getDevice().getTile(i) + " from module" + mdi))
				);*/

		//no intersection with any tile from any module already "placed"
		return Tracer.exit(moduleDescriptionInstances.stream()
						//exclude the mdi we want to change
						.filter(mdi -> mdi != mdiToChange)
						//only placed modules
						.filter(ModuleDescriptionInstance::isPlaced)
						//no collisions in any of these moduleDescriptionInstances
						.noneMatch(mdi -> mdi.getPlacement().streamTileNumbers(true, includeRoutingTiles)
								.anyMatch(newTiles::contains))
		);
	}
}
