/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.visualization;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;
import de.tu_darmstadt.rs.rapidSoc.design.PortNet;

import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.Tile;
import edu.byu.ece.rapidSmith.device.TileType;
import edu.byu.ece.rapidSmith.device.Utils;
import edu.byu.ece.rapidSmith.gui.TileColorsJava;
import org.apache.batik.dom.GenericDOMImplementation;
import org.apache.batik.svggen.SVGGraphics2D;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Point;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.TreeSet;
import java.util.function.BiFunction;
import java.util.function.Function;

/**
 * Base For all FPGA visualizations
 */
//Many pixel offsets. Therefore we disable the magic number check.
@SuppressWarnings("checkstyle:MagicNumber")
public abstract class BaseVisualization {

	public static final Color[] colors = new Color[]{
			Color.BLUE,
			Color.GRAY,
			Color.CYAN,
			Color.MAGENTA,
			Color.GREEN,
			Color.YELLOW,
			Color.ORANGE,
			Color.PINK
	};
	/** Number of tile columns being referenced on the device */
	public final int cols;
	/** Number of tile rows being referenced on the device */
	public final int rows;
	/** When hiding tiles, this contains the grid of drawn tiles */
	public Tile[][] drawnTiles;
	/** Gets the X coordinate of the tile in the drawnTiles grid */
	public HashMap<Tile, Integer> tileXMap;
	/** Gets the Y coordinate of the tile in the drawnTiles grid */
	public HashMap<Tile, Integer> tileYMap;
	/** Width of the lines drawn in between tiles when columns/rows are hidden */
	public double lineWidth = 1;
	protected final boolean drawPrimitives;
	/** The current design associated with this scene */
	protected final Device device;
	/** This is the set of column tile types which should not be drawn */
	protected final HashSet<TileType> tileColumnTypesToHide;
	/** This is the set of row tile types which should not be drawn */
	protected final HashSet<TileType> tileRowTypesToHide;
	protected final TreeSet<Integer> colsToSkip;
	protected final TreeSet<Integer> rowsToSkip;

	private Map<Tile, Set<Net>> routingOnlyTiles;

	private final String unsupportedFamilyType = "Unsupported Family Type";

	/**
	 * Create Visualization
	 * @param device The device
	 * @param hideTiles hide clock distribution cols/rows
	 * @param drawPrimitives draw primitives in more detail
	 */
	public BaseVisualization(Device device, boolean hideTiles, boolean drawPrimitives) {
		this.drawPrimitives = drawPrimitives;
		this.device = device;
		Objects.requireNonNull(device);


		tileColumnTypesToHide = new HashSet<>();
		tileRowTypesToHide = new HashSet<>();
		colsToSkip = new TreeSet<>();
		rowsToSkip = new TreeSet<>();

		if (hideTiles) {
			populateTileTypesToHide();
		}

		int tempRows = device.getRows();
		int tempCols = device.getColumns();

		// Determine which columns and rows to not draw
		determineColsRowsToSkip();


		rows = tempRows - rowsToSkip.size();
		cols = tempCols - colsToSkip.size();

		// Create new tile layout without hidden tiles
		int i = 0;
		int j = 0;
		drawnTiles = new Tile[rows][cols];
		tileXMap = new HashMap<Tile, Integer>();
		tileYMap = new HashMap<Tile, Integer>();
		for (int row = 0; row < tempRows; row++) {
			if (rowsToSkip.contains(row)) {
				continue;
			}
			for (int col = 0; col < tempCols; col++) {
				if (colsToSkip.contains(col)) {
					continue;
				}
				Tile tile = device.getTile(row, col);
				Objects.requireNonNull(tile);
				drawnTiles[i][j] = tile;
				tileXMap.put(tile, j);
				tileYMap.put(tile, i);
				j++;
			}
			i++;
			j = 0;
		}
	}

	public Point getSceneSize(int tileSize) {
		return new Point(cols * tileSize, rows * tileSize);
	}

	protected void determineColsRowsToSkip() {
		for (Tile[] tileRow : device.getTiles()) {
			for (Tile tile : tileRow) {
				TileType type = tile.getType();
				if (tileColumnTypesToHide.contains(type)) {
					colsToSkip.add(tile.getColumn());
				}
				if (tileRowTypesToHide.contains(type)) {
					rowsToSkip.add(tile.getRow());
				}
			}
		}
	}

	protected boolean showNetInRoutingOnly(Net n) {
		return true;
	}

	public Map<Tile, Set<Net>> getRoutingOnlyTiles() {
		initRoutingOnlyTiles();
		return routingOnlyTiles;
	}

	public void paint(Graphics2D g2d, int tileSize) {
		drawFpgaFabric(drawPrimitives, g2d, tileSize);
	}

	public void svgToStream(OutputStream outputStream) throws IOException {
		svgToStream(outputStream, null);
	}

	/**
	 * Write SVG Data to stream
	 * @param outputStream the stream to write to
	 * @param viewBox Bounding Box to limit the view to
	 * @throws IOException When Writing to the stream fails
	 */
	public void svgToStream(OutputStream outputStream, BoundingBox viewBox, int tileSize) throws IOException {

		// Get a DOMImplementation.
		DOMImplementation domImpl =
				GenericDOMImplementation.getDOMImplementation();

		// Create an instance of org.w3c.dom.Document.
		String svgNS = "http://www.w3.org/2000/svg";
		Document document = domImpl.createDocument(svgNS, "svg", null);

		// Create an instance of the SVG Generator.
		SVGGraphics2D svgGenerator = new SVGGraphics2D(document);


		// Ask the test to render into the SVG Graphics2D implementation.
		paint(svgGenerator, tileSize);

		Element root = svgGenerator.getRoot();
		if (viewBox != null) {
			root.setAttributeNS(null, "viewBox", String.format("%d %d %d %d", viewBox.minX, viewBox.minY, viewBox.maxX, viewBox.maxY));
		}


		// Finally, stream out SVG to the standard output using
		// UTF-8 encoding.
		boolean useCSS = true; // we want to use CSS style attributes
		Writer out = new OutputStreamWriter(outputStream, "UTF-8");
		//svgGenerator.stream(out, useCSS);

		svgGenerator.stream(root, out, useCSS, false);
	}


	public void svgToStream(OutputStream outputStream, BoundingBox viewBox) throws IOException {
		svgToStream(outputStream, viewBox, 20);
	}

	protected abstract boolean isUsed(Tile tile);

	protected void drawFpgaFabric(boolean drawPrimitives, Graphics2D g2d, int tileSize) {

		initRoutingOnlyTiles();

		g2d.setColor(Color.black);
		g2d.setPaint(Color.black);

		g2d.fillRect(0, 0, getSceneSize(tileSize).x, getSceneSize(tileSize).y);


		//Draw dashed lines where rows/columns have been removed
		drawRemoveLines(g2d, tileSize);


		g2d.setStroke(new BasicStroke());

		// Draw the tile layout
		int offset = (int) Math.ceil((lineWidth / 2.0));

		for (int y = 0; y < rows; y++) {
			for (int x = 0; x < cols; x++) {
				Tile tile = drawnTiles[y][x];
				TileType tileType = tile.getType();

				boolean used = isUsed(tile);
				boolean routingOnly = routingOnlyTiles != null && routingOnlyTiles.keySet().contains(tile);


				Color color = getTileColor(tile, used, routingOnly);


				g2d.setColor(color);
				g2d.setPaint(color);


				int rectX = x * tileSize;
				int rectY = y * tileSize;
				int rectSide = tileSize - 2 * offset;


				int tileHeight;
				if (Utils.isBRAM(tileType) || Utils.isDSP(tileType)) {
					tileHeight = 4;
				} else {
					tileHeight = 1;
				}

				if (drawPrimitives) {
					if (Utils.isCLB(tileType)) {
						drawCLB(g2d, rectX, rectY, rectSide, tileType);
					} else if (Utils.isSwitchBox(tileType)) {
						drawSwitchBox(g2d, rectX, rectY, rectSide);
					} else if (Utils.isBRAM(tileType)) {
						drawBram(g2d, rectX, rectY, rectSide, offset, color, tileSize);
					} else if (Utils.isDSP(tileType)) {
						drawDSP(g2d, rectX, rectY, rectSide, offset, color, tileSize);
					} else { // Just fill the tile in with a color
						colorTile(g2d, x, y, offset, color, tileHeight, tileSize);
					}
				} else {
					colorTile(g2d, x, y, offset, color, tileHeight, tileSize);
				}
			}
		}

	}

	protected abstract Map<Tile, Set<Net>> calcRoutingOnlyTiles();

	private void initRoutingOnlyTiles() {
		if (routingOnlyTiles == null) {
			routingOnlyTiles = calcRoutingOnlyTiles();
		}

	}

	protected void drawRemoveLines(Graphics2D g2d, int tileSize) {
		//QPen missingTileLinePen = new QPen(QColor.lightGray, 2, Qt.PenStyle.DashLine);
		g2d.setColor(Color.lightGray);
		g2d.setStroke(new BasicStroke(2,
				BasicStroke.CAP_BUTT,
				BasicStroke.JOIN_MITER,
				1.0f, new float[]{10.0f}, 0.0f));
		int i = 0;
		for (int col : colsToSkip) {
			int realCol = col - i;
			g2d.drawLine(tileSize * realCol - 1, 0, tileSize * realCol - 1, rows * tileSize - 3);
			i++;
		}
		i = 0;
		for (int row : rowsToSkip) {
			int realRow = row - i;
			g2d.drawLine(0, tileSize * realRow - 1, cols * tileSize - 3, tileSize * realRow - 1);
			i++;
		}
	}

	protected Color getTileColor(Tile tile, boolean used, boolean routingOnly) {
		// Set pen color based on current tile
		Color color = TileColorsJava.getSuggestedTileColor(tile);
		if (color==null)
			color = Color.CYAN;

		if (routingOnly) {
			color = Color.red;
		} else if (!used) {
			color = color.darker().darker().darker().darker();
		}

		return color;
	}

	protected Color colorWithAlpha(Color c, int a) {
		return new Color(c.getRed(), c.getGreen(), c.getBlue(), a);
	}

	/**
	 * Gets the tile based on the x and y coordinates given (typically from mouse input)
	 * @param x The x location on the screen.
	 * @param y The y location on the screen.
	 * @return The tile at the x,y location or null if none exist.
	 */
	public Tile getTile(double x, double y, int tileSize) {
		int currX = (int) Math.floor(x / tileSize);
		int currY = (int) Math.floor(y / tileSize);
		if (currX >= 0 && currY >= 0 && currX < cols && currY < rows) { // && (currX != prevX || currY != prevY)){
			return drawnTiles[currY][currX];
		}
		return null;
	}

	private void drawCLB(Graphics2D g2d, int rectX, int rectY, int rectSide, TileType tileType) {

		g2d.drawRect(rectX, rectY + rectSide / 2, rectSide / 2 - 1, rectSide / 2 - 1);
		g2d.drawRect(rectX + rectSide / 2, rectY, rectSide / 2 - 1, rectSide / 2 - 1);
		switch (device.getFamilyType()) {
			case SPARTAN3:
			case SPARTAN3A:
			case SPARTAN3ADSP:
			case SPARTAN3E:
			case VIRTEX2:
			case VIRTEX2P:
			case VIRTEX4:
			case SPARTAN6:

				char c = tileType.toString().charAt(tileType.toString().length() - 1);
				if (c == 'L') {
					g2d.drawOval(rectX, rectY, rectSide / 2 - 1, rectSide / 2 - 1);
					g2d.drawOval(rectX + rectSide / 2, rectY + rectSide / 2, rectSide / 2 - 1, rectSide / 2 - 1);
				}

				break;
			default:
		}
	}

	private void drawBram(Graphics2D g2d, int rectX, int rectY, int rectSide, int offset, Color color, int tileSize) {
		switch (device.getFamilyType()) {
			case SPARTAN6:
				g2d.drawRect(rectX, rectY - 3 * tileSize, rectSide - 1, 4 * rectSide + 3 * 2 * offset - 1);
				g2d.setColor(color.darker());
				g2d.drawRect(rectX + 2, rectY - 3 * tileSize + 2, rectSide - 1 - 4, 2 * rectSide + 2 * offset - 1 - 2);
				g2d.drawRect(rectX + 2, rectY - tileSize, rectSide - 1 - 4, 2 * rectSide + 2 * offset - 1 - 2);
				break;
			case VIRTEX5:
				g2d.drawRect(rectX, rectY - 4 * tileSize, rectSide - 1, 5 * rectSide + 3 * 2 * offset - 1);
				g2d.setColor(color.darker());
				g2d.drawRect(rectX + 2, rectY - 4 * tileSize + 2, rectSide - 5, 5 * rectSide + 3 * 2 * offset - 5);
				break;
			case KINTEX7:
			case VIRTEX6:
			case VIRTEX7:
			case ARTIX7:
				g2d.drawRect(rectX, rectY - 4 * tileSize, rectSide - 1, 5 * rectSide + 3 * 2 * offset - 1);
				g2d.setColor(color.darker());
				g2d.drawRect(rectX + 2, rectY - 4 * tileSize + 2, rectSide - 5, ((int) (2.5 * rectSide)) + 3 * 2 * offset - 5);
				g2d.drawRect(rectX + 2, (rectY - 2 * tileSize) + 7, rectSide - 5, ((int) (2.5 * rectSide)) + 3 * 2 * offset - 5);
				break;
			case VIRTEXE:
			case SPARTAN2:
			case SPARTAN2E:
			case SPARTAN3:
			case SPARTAN3A:
			case SPARTAN3ADSP:
			case SPARTAN3E:
			case VIRTEX:
			case VIRTEX2:
			case VIRTEX2P:
			case VIRTEX4:
				g2d.drawRect(rectX, rectY - 3 * tileSize, rectSide - 1, 4 * rectSide + 3 * 2 * offset - 1);
				g2d.setColor(color.darker());
				g2d.drawRect(rectX + 2, rectY - 3 * tileSize + 2, rectSide / 2 - 4, 4 * (rectSide + offset) - 3);
				g2d.drawRect(rectX + rectSide / 2 + 1, rectY - 3 * tileSize + 2, rectSide / 2 - 4, 4 * (rectSide + offset) - 3);
				break;
			default: throw new IllegalStateException(unsupportedFamilyType);
		}
	}

	private void drawDSP(Graphics2D g2d, int rectX, int rectY, int rectSide, int offset, Color color, int tileSize) {
		switch (device.getFamilyType()) {
			case SPARTAN6:
				g2d.drawRect(rectX, rectY - 3 * tileSize, rectSide - 1, 4 * rectSide + 3 * 2 * offset - 1);
				g2d.setColor(color.darker());
				g2d.drawRect(rectX + 2, rectY - 3 * tileSize + 2, rectSide - 5, 4 * rectSide + 3 * 2 * offset - 5);
				break;
			case VIRTEX5:
			case KINTEX7:
			case VIRTEX6:
			case VIRTEX7:
			case ARTIX7:
				g2d.drawRect(rectX, rectY - 4 * tileSize, rectSide - 1, 5 * rectSide + 3 * 2 * offset - 1);
				g2d.setColor(color.darker());
				g2d.drawRect(rectX + 2, rectY - 4 * tileSize + 2, rectSide - 5, ((int) (2.5 * rectSide)) + 3 * 2 * offset - 5);
				g2d.drawRect(rectX + 2, (rectY - 2 * tileSize) + 7, rectSide - 5, ((int) (2.5 * rectSide)) + 3 * 2 * offset - 5);
				break;
			case VIRTEXE:
			case SPARTAN2:
			case SPARTAN2E:
			case SPARTAN3:
			case SPARTAN3A:
			case SPARTAN3ADSP:
			case SPARTAN3E:
			case VIRTEX:
			case VIRTEX2:
			case VIRTEX2P:
			case VIRTEX4:
				g2d.drawRect(rectX, rectY - 3 * tileSize, rectSide - 1, 4 * rectSide + 3 * 2 * offset - 1);
				g2d.setColor(color.darker());
				g2d.drawRect(rectX + 2, rectY - 3 * tileSize + 2, rectSide - 1 - 4, 2 * rectSide + 2 * offset - 1 - 2);
				g2d.drawRect(rectX + 2, rectY - tileSize, rectSide - 1 - 4, 2 * rectSide + 2 * offset - 1 - 2);
				break;
			default: throw new IllegalStateException(unsupportedFamilyType);
		}

	}

	private void drawSwitchBox(Graphics2D g2d, int rectX, int rectY, int rectSide) {
		g2d.drawRect(rectX + rectSide / 6, rectY, 4 * rectSide / 6 - 1, rectSide - 1);
	}

	private void colorTile(Graphics2D g2d, int x, int y, int offset, Color color, int tileHeight, int tileSize) {
		g2d.setPaint(color);
		g2d.fillRect(x * tileSize, (y - tileHeight + 1) * tileSize,
				tileSize - 2 * offset, tileSize * tileHeight - 2 * offset);
	}

	protected void populateTileTypesToHide() {
		switch (device.getFamilyType()) {
			case VIRTEX4:
				tileColumnTypesToHide.add(TileType.CLB_BUFFER);
				tileColumnTypesToHide.add(TileType.CLK_HROW);
				tileColumnTypesToHide.add(TileType.CFG_VBRK_FRAME);
				tileRowTypesToHide.add(TileType.HCLK);
				tileRowTypesToHide.add(TileType.BRKH);
				break;
			case VIRTEX5:
				tileColumnTypesToHide.add(TileType.CFG_VBRK);
				tileColumnTypesToHide.add(TileType.CLKV);
				tileColumnTypesToHide.add(TileType.INT_BUFS_L);
				tileColumnTypesToHide.add(TileType.INT_BUFS_R);
				tileColumnTypesToHide.add(TileType.INT_BUFS_R_MON);
				tileColumnTypesToHide.add(TileType.INT_INTERFACE);
				//tileColumnTypesToHide.add(TileType.IOI);
				tileRowTypesToHide.add(TileType.HCLK);
				tileRowTypesToHide.add(TileType.BRKH);
				break;
			case VIRTEX6:
				tileRowTypesToHide.add(TileType.HCLK);
				tileRowTypesToHide.add(TileType.BRKH);
				tileColumnTypesToHide.add(TileType.INT_INTERFACE);
				tileColumnTypesToHide.add(TileType.VBRK);
				break;
			case SPARTAN6:
				tileColumnTypesToHide.add(TileType.INT_INTERFACE);
				tileColumnTypesToHide.add(TileType.VBRK);
				tileRowTypesToHide.add(TileType.HCLK_CLB_XL_CLE);
				tileRowTypesToHide.add(TileType.REGH_CLEXL_CLE);
				break;
			default:
				throw new IllegalStateException(unsupportedFamilyType);
		}
	}

	/**
	 * Save numbered image
	 * @param path Path to save to
	 * @param i Image number
	 */
	public void saveImgNum(Path path, int i) {
		try {
			final Path dir = path.resolve("visu");
			Files.createDirectories(dir);
			saveImage(dir.resolve(String.format("%02d.svg", i)));
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void saveImage(Path file) {
		try {
			svgToStream(Files.newOutputStream(file));
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	protected void drawNets(Graphics2D g2d, Collection<Net> nets, Function<Tile, Tile> tileMapper, BiFunction<Net, Pin, Color> netColor, int tileSize) {
		nets.forEach(net -> {
			if (net.getSource() == null) {
				return;
			}

			if (net.getSourceTile() == null) {
				return;
			}

			net.getPins().stream()
					.filter(p -> p != net.getSource())
					//.filter(p -> p.getInstance().getModuleInstance()!=net.getSource().getInstance().getModuleInstance())
					.forEach(sink -> {
						if (sink.getTile() != null) {
							Color c = netColor.apply(net, sink);
							if (c != null) {
								g2d.setColor(c);
								Tile sourceTile = tileMapper.apply(net.getSourceTile());
								Tile sinkTile = tileMapper.apply(sink.getTile());
								g2d.drawLine(
										tileSize / 2 + tileSize * tileXMap.get(sourceTile),
										tileSize / 2 + tileSize * tileYMap.get(sourceTile),
										tileSize / 2 + tileSize * tileXMap.get(sinkTile),
										tileSize / 2 + tileSize * tileYMap.get(sinkTile)
								);
							}
						}
					});
		});
	}


	protected void drawNets(Graphics2D g2d, Collection<Net> nets, BiFunction<Net, Pin, Color> netColor, int tileSize) {
		drawNets(g2d, nets, Function.identity(), netColor, tileSize);
	}

	protected void drawPortNets(Graphics2D g2d, Collection<PortNet> nets, BiFunction<PortNet, Tile, Color> netColor, int tileSize) {
		nets.forEach(net -> {
			if (null == net.getSourcePort()) {
				return;
			}

			if (!net.getSourceModule().isPlaced()) {
				return;
			}

			//TODO make sure the first tile contains the source pin
			Optional<Tile> sourceTileOpt = net.getSourceModule().getPlacement().streamPortTiles(net.getSourcePort().getName()).findFirst();
			if (!sourceTileOpt.isPresent()) {
				//TODO log
				return;
			}
			Tile sourceTile = sourceTileOpt.get();

			//draw a line from the source tile to all sink tiles
			net.streamTiles()
					.filter(Objects::nonNull)
					.filter(t -> !t.equals(sourceTile))
					//.filter(p -> p.getInstance().getModuleInstance()!=net.getSource().getInstance().getModuleInstance())
					.forEach(sink -> {
						Color c = netColor.apply(net, sink);
						if (Objects.nonNull(c)) {
							g2d.setColor(c);
							g2d.drawLine(
									tileSize / 2 + tileSize * tileXMap.get(sourceTile),
									tileSize / 2 + tileSize * tileYMap.get(sourceTile),
									tileSize / 2 + tileSize * tileXMap.get(sink),
									tileSize / 2 + tileSize * tileYMap.get(sink)
							);
						}
					});
		});
	}


	protected boolean drawBB(Graphics2D g2d, BoundingBox bb, int tileSize, int extend) {
		Tile tl = device.getTile(bb.minY, bb.minX);
		Tile br = device.getTile(bb.maxY, bb.maxX);

		if (tileXMap.get(tl) == null || tileXMap.get(br) == null) {
			return false;
		}

		int minX = tileSize * tileXMap.get(tl) - extend;
		int minY = tileSize * tileYMap.get(tl) - extend;
		int maxX = tileSize * (tileXMap.get(br) + 1) + extend - 2;
		int maxY = tileSize * (tileYMap.get(br) + 1) + extend - 2;


		g2d.drawRect(minX, minY, maxX - minX, maxY - minY);

		return true;

	}

	protected boolean drawBB(Graphics2D gd2, BoundingBox bb, int tileSize) {
		return drawBB(gd2, bb, tileSize, 0);
	}

	public abstract Instance getInstanceAtPrimitiveSite(PrimitiveSite primitiveSite);
}
