/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.visualization;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.ModuleInstance;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.PIP;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.Tile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.Color;
import java.awt.Graphics2D;
import java.util.List;
import java.util.Set;
import java.util.stream.Collector;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Visualization for a single module still embedded in its source design
 */
public class ModuleInDesignVisualization extends DesignVisualization {
	private static final Logger logger = LoggerFactory.getLogger(ModuleInDesignVisualization.class);

	private final Module module;
	private final boolean includeRoutingTiles;
	private final boolean showPossiblePlacements;
	private final List<PrimitiveSite> placements;

	/**
	 * Create a module visualization
	 * @param design the design to visualize
	 * @param hideTiles hide clock tiles
	 * @param drawPrimitives draw primitives in more detail
	 * @param module the module to visualize
	 * @param includeRoutingTiles show routing tiles
	 * @param showPossiblePlacements show possible placements
	 */
	public ModuleInDesignVisualization(Design design, boolean hideTiles, boolean drawPrimitives, Module module, boolean includeRoutingTiles, boolean showPossiblePlacements) {
		super(design, hideTiles, drawPrimitives);
		this.module = module;
		this.includeRoutingTiles = includeRoutingTiles;
		this.showPossiblePlacements = showPossiblePlacements;

		if (showPossiblePlacements && module != null) {
			this.placements = calcPlacements();
		} else {
			this.placements = null;
		}
	}

	private List<PrimitiveSite> calcPlacements() {
		Design tempDesign = new Design("temp", design.getPartName());
		ModuleInstance mi = tempDesign.createModuleInstance("temp", module);
		List<PrimitiveSite> allValidPlacements = mi.getAllValidPlacements((a, b) -> {
		});
		logger.info("module has " + allValidPlacements.size() + " placements");
		return allValidPlacements;
	}

	protected Color getNetColor(Net net, boolean insideNet, boolean outsideNet) {
		if (outsideNet) {
			return Color.RED;
		}
		final int netAlpha = 30;
		if (insideNet) {
			return colorWithAlpha(Color.GREEN, netAlpha);
		}

		//Net is partial
		return colorWithAlpha(Color.WHITE, netAlpha);
	}


	protected boolean showNetInRoutingOnly(Net n) {
		return module != null && module.getNets().contains(n);
	}

	@Override
	protected void drawFpgaFabric(boolean drawPrimitives, Graphics2D g2d, int tileSize) {
		super.drawFpgaFabric(drawPrimitives, g2d, tileSize);


		if (module != null) {

			Collector<Tile, ?, BoundingBox> bbCollector = BoundingBox.getTileBoundingBox();

			BoundingBox bb = null;
			Stream<Tile> tileStream = module.getInstances().stream()
					.map(Instance::getTile)
					.filter(t -> t != null);


			if (includeRoutingTiles) {
				Stream<Tile> fromRouting = module.getNets().stream()
						.flatMap(n -> n.getPIPs().stream())
						.map(PIP::getTile);
				tileStream = Stream.concat(tileStream, fromRouting);
			}

			Set<Tile> tiles = tileStream.collect(Collectors.toSet());

			bb = tiles.stream().collect(bbCollector);


			g2d.setColor(Color.WHITE);


			final int[] i = {0};

			if (showPossiblePlacements) {
				placements.forEach(anchor -> {

					i[0]++;

					Color color = BaseVisualization.colors[i[0] % BaseVisualization.colors.length];
					g2d.setColor(color);

					BoundingBox b = tiles.stream().map(tile -> module.getCorrespondingTile(tile, anchor.getTile(), design.getDevice())).collect(bbCollector);

					if (!drawBB(g2d, b, tileSize)) {
						logger.warn("could not draw bb for " + anchor.getTile());
					}
				});
			}
			g2d.setColor(Color.BLUE);

			if (bb != null) {
				if (!drawBB(g2d, bb, tileSize)) {
					logger.warn("could not draw bb");
				}
			}

		}


		drawNets(g2d, design.getNets(), (net, pin) -> {
			boolean sourceInside = module == null || module.getInstances().contains(net.getSource().getInstance());
			boolean sinkInside = module == null || module.getInstances().contains(pin.getInstance());

			boolean outsideNet = !sourceInside && !sinkInside;
			boolean insideNet = sourceInside && sinkInside;

			return getNetColor(net, insideNet, outsideNet);

		}, tileSize);
	}

	@Override
	public String toString() {
		if (module!=null)
			return module.getName();
		else return "Visualization";
	}
}
