/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.visualization;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionVariant;
import de.tu_darmstadt.rs.rapidSoc.module.ModulePlacement;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.Tile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collector;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Visualization for a single module, possibly on a different FPGA than it was created with
 */
public class ModuleVisualization extends BaseVisualization {

	private static final Logger logger = LoggerFactory.getLogger(ModuleVisualization.class);

	@org.jetbrains.annotations.NotNull
	private final ModuleDescriptionVariant moduleDescriptionVariant;
	private final boolean includeRoutingTiles;
	private final boolean showPossiblePlacements;
	private final List<Tile> placements;
	private Set<Tile> usedTiles;
	private ModulePlacement modulePlacement;

	/**
	 * Create a module visualization
	 *
	 * @param moduleDescriptionVariant the module to visualize
	 * @param hideTiles hide clock tiles
	 * @param drawPrimitives draw primitives in more detail
	 * @param includeRoutingTiles show routing tiles
	 * @param showPossiblePlacements show possible placements
	 */
	public ModuleVisualization(ModuleDescriptionVariant moduleDescriptionVariant, boolean hideTiles, boolean drawPrimitives, boolean includeRoutingTiles, boolean showPossiblePlacements) {
		super(Objects.requireNonNull(moduleDescriptionVariant).getDevice(), hideTiles, drawPrimitives);
		this.moduleDescriptionVariant = moduleDescriptionVariant;
		this.includeRoutingTiles = includeRoutingTiles;
		this.showPossiblePlacements = showPossiblePlacements;


		Tile placement = moduleDescriptionVariant.getAnchor();

		this.modulePlacement = new ModulePlacement(moduleDescriptionVariant, placement);


		calcUsedTiles();

		if (showPossiblePlacements) {
			this.placements = calcPlacements();
		} else {
			this.placements = null;
		}
	}

	private void calcUsedTiles() {
		this.usedTiles = modulePlacement.streamTiles(true, false).collect(Collectors.toSet());
	}

	private List<Tile> calcPlacements() {
		return moduleDescriptionVariant.streamPlacements()
				.collect(Collectors.toList());
	}

	protected Color getNetColor(Net net, boolean insideNet, boolean outsideNet) {
		if (outsideNet) {
			return Color.RED;
		}
		final int netAlpha = 30;
		if (insideNet) {
			return colorWithAlpha(Color.GREEN, netAlpha);
		}

		//Net is partial
		return colorWithAlpha(Color.WHITE, netAlpha);
	}


	protected boolean showNetInRoutingOnly(Net n) {
		return moduleDescriptionVariant.getModule().getNets().contains(n);
	}

	@Override
	protected boolean isUsed(Tile tile) {
		return usedTiles.contains(tile);
	}

	@Override
	protected void drawFpgaFabric(boolean drawPrimitives, Graphics2D g2d, int tileSize) {
		super.drawFpgaFabric(drawPrimitives, g2d, tileSize);


		if (showPossiblePlacements) {
			drawPossiblePlacements(g2d, tileSize);
		}



		drawNets(g2d, moduleDescriptionVariant.getModule().getNets(), tile -> {
			//Do not use description, as it uses a different original anchor (one for the current device)
			return moduleDescriptionVariant.getModule().getCorrespondingTile(tile, modulePlacement.getAnchor(),device);
		}, (net, pin) -> {
				boolean sourceInside = moduleDescriptionVariant.getModule().getInstances().contains(net.getSource().getInstance());
				boolean sinkInside = moduleDescriptionVariant.getModule().getInstances().contains(pin.getInstance());

				boolean outsideNet = !sourceInside && !sinkInside;
				boolean insideNet = sourceInside && sinkInside;

				return getNetColor(net, insideNet, outsideNet);


		}, tileSize);

		//Draw bounding box highlighting currently selected placement
		g2d.setColor(Color.RED);
		g2d.setStroke(new BasicStroke(3));
		drawBoundingBox(g2d, modulePlacement.getAnchor(), tileSize);
	}

	private void drawPossiblePlacements(Graphics2D g2d, int tileSize) {
		for (int i = 0; i < placements.size(); i++) {
			Tile anchor = placements.get(i);

			Color color = BaseVisualization.colors[i % BaseVisualization.colors.length];
			g2d.setColor(color);

			drawBoundingBox(g2d, anchor, tileSize);
		}
	}

	private void drawBoundingBox(Graphics2D g2d, Tile anchor, int tileSize) {

		Collector<Tile, ?, BoundingBox> bbCollector = BoundingBox.getTileBoundingBox();
		BoundingBox b = moduleDescriptionVariant.streamTilesForAnchor(anchor, true, includeRoutingTiles).collect(bbCollector);

		if (!drawBB(g2d, b, tileSize)) {
			logger.warn("could not draw bb for " + anchor);
		}
	}

	@Override
	protected Map<Tile, Set<Net>> calcRoutingOnlyTiles() {
		return null;
	}

	@Override
	public Instance getInstanceAtPrimitiveSite(PrimitiveSite primitiveSite) {
		Tracer.enterLoop("instance at primitive site");

			int tileXOffset = modulePlacement.getAnchor().getTileXCoordinate() - moduleDescriptionVariant.getAnchor().getTileXCoordinate();
			int tileYOffset = modulePlacement.getAnchor().getTileYCoordinate() - moduleDescriptionVariant.getAnchor().getTileYCoordinate();
		final Tile correspondingTile = Module.getCorrespondingTile(primitiveSite.getTile(), -tileXOffset, -tileYOffset, moduleDescriptionVariant.getDevice());
		PrimitiveSite newSite = Device.getCorrespondingPrimitiveSite(primitiveSite, primitiveSite.getType(), correspondingTile);

		return Tracer.exit(getInstanceByOriginalSite(newSite));
	}

	private Map<PrimitiveSite, Instance> instancesBySite = null;
	private Instance getInstanceByOriginalSite(PrimitiveSite site) {
		if (instancesBySite==null) {
			instancesBySite = moduleDescriptionVariant.getModule().getInstances().stream().collect(Collectors.toMap(Instance::getPrimitiveSite, Function.identity()));
		}
		return instancesBySite.get(site);
	}

	@Override
	public String toString() {
		return moduleDescriptionVariant.getModule().getName() + " on "+device.getPartName()+" at "+modulePlacement.getAnchor().getName();
	}

	public Stream<Tile> streamPossiblePlacements() {
		return moduleDescriptionVariant.streamPlacements();
	}

	public Tile getPlacement() {
		return modulePlacement.getAnchor();
	}

	public void setPlacement(Tile tile) {
		modulePlacement = new ModulePlacement(moduleDescriptionVariant, tile);
		calcUsedTiles();
	}
}
