/*
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.visualization;

import de.tu_darmstadt.rs.rapidSoc.BoundingBox;
import de.tu_darmstadt.rs.rapidSoc.design.PortNet;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionInstance;
import de.tu_darmstadt.rs.rapidSoc.module.ModulePlacement;
import de.tu_darmstadt.rs.rapidSoc.placer.ProposedMoveListener;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.device.Tile;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * Visualize Module placements
 */
public class PlacementVisualization extends DesignVisualization implements ProposedMoveListener {
	private static final Logger logger = LoggerFactory.getLogger(PlacementVisualization.class);

	private final boolean showBBs;

	private Collection<ModuleDescriptionInstance> moduleDescriptionInstances;
	private Collection<PortNet> hlNets;
	private Map<ModuleDescriptionInstance, Color> mdiColor;

	private boolean pendingMove = false;
	private ModuleDescriptionInstance changedMDI;
	private Tile oldAnchor;

	private final boolean includeRoutingTiles;

	/**
	 * Creates a new tile scene with a design.
	 * @param design The design and device to associate with this scene.
	 * @param hideTiles A flag to hide/show certain tiles to make the fabric appear more homogeneous.
	 * @param drawPrimitives A flag to draw boxes to represent primitives.
	 * @param showBBs show bounding boxes
	 * @param includeRoutingTiles Highlight tiles only touched by routing
	 */
	public PlacementVisualization(Design design, boolean hideTiles, boolean drawPrimitives, boolean showBBs, boolean includeRoutingTiles) {
		super(design, hideTiles, drawPrimitives);

		this.hlNets = Collections.emptySet();
		this.showBBs = showBBs;
		this.includeRoutingTiles = includeRoutingTiles;
	}

	/**
	 * set nets to highlight
	 * @param hlNets highlight nets
	 */
	public void setHlNets(Collection<PortNet> hlNets) {
		this.hlNets = hlNets;
	}

	private void drawHlNets(Graphics2D g2d, int tileSize) {
		drawPortNets(g2d, hlNets, (net, tile) -> {
			//TODO
/*T			if (pin.getInstance().getModuleInstance() == net.getSource().getInstance().getModuleInstance()) {
				return null;
			}*/

			ModuleDescriptionInstance mdi = findModuleDescriptionInstance(tile);

			Color c;
			if (Objects.nonNull(mdi)) {
				c = modulePlacementColor(mdi);
			} else {
				c = Color.red;
			}

			final int halfTransparent = 128;
			c = colorWithAlpha(c, halfTransparent);

/*			if (net.getType() != NetType.WIRE) {
				c = Color.WHITE;
			}*/

			return c;
		}, tileSize);
	}

	private ModuleDescriptionInstance findModuleDescriptionInstance(Tile tile) {
		List<ModuleDescriptionInstance> found = moduleDescriptionInstances.stream()
				.filter(ModuleDescriptionInstance::isPlaced)
				.filter(mdi ->
			mdi.getPlacement().streamTiles(true, false)
				.filter(tile::equals)
				.count() > 0
		).collect(Collectors.toList());
		if (found.size() < 1) {
			return null;
		} else {
			return found.get(0);
		}
	}

	private Color modulePlacementColor(ModuleDescriptionInstance placement) {
		return mdiColor.get(placement);
	}

	private void drawBBs(Graphics2D g2d, int tileSize) {
		int[] modIdx = {0};
		moduleDescriptionInstances.forEach(mdi -> {
			if (mdi.isPlaced()) {

				g2d.setColor(modulePlacementColor(mdi));

				if (!drawBB(g2d, mdi.getPlacement().getBoundingBox(includeRoutingTiles), tileSize)) {
					logger.warn("fail for " + mdi);
				}
			}

			modIdx[0]++;
		});
	}

	@Override
	protected void drawFpgaFabric(boolean drawPrimitives, Graphics2D g2d, int tileSize) {
		super.drawFpgaFabric(drawPrimitives, g2d, tileSize);

		if (hlNets != null) {
			drawHlNets(g2d, tileSize);
		}

		if (showBBs) {
			drawBBs(g2d, tileSize);
		}


		drawModuleMove(g2d, tileSize);

	}

	@Override
	protected Color getTileColor(Tile tile, boolean used, boolean routingOnly) {
		Color color = modulePlacementColor(findModuleDescriptionInstance(tile));
		if (color != null) {
			return color;
		}
		return super.getTileColor(tile, used, routingOnly);
	}

	//FROM http://stackoverflow.com/questions/4112701/drawing-a-line-with-arrow-in-java
	@SuppressWarnings("checkstyle:MagicNumber")
	private void drawArrow(Graphics g1, int x1, int y1, int x2, int y2) {
		final int arrowSize = 20;

		Graphics2D g = (Graphics2D) g1.create();

		double dx = x2 - x1;
		double dy = y2 - y1;
		double angle = Math.atan2(dy, dx);
		int len = (int) Math.sqrt(dx * dx + dy * dy);
		AffineTransform at = AffineTransform.getTranslateInstance(x1, y1);
		at.concatenate(AffineTransform.getRotateInstance(angle));
		g.transform(at);

		// Draw horizontal arrow starting in (0, 0)
		g.drawLine(0, 0, len - arrowSize, 0);
		g.fillPolygon(new int[]{len, len - 2 * arrowSize, len - 2 * arrowSize, len},
				new int[]{0, -arrowSize, arrowSize, 0}, 4);
	}

	private void drawModuleMove(Graphics2D g2d, int tileSize) {
		if (Objects.nonNull(changedMDI) && !pendingMove) {

			BoundingBox bb = changedMDI.getPlacement().getBoundingBox(includeRoutingTiles);

			int offsAnchorToBBBorderX = bb.minX - changedMDI.getPlacement().getAnchor().getColumn();
			float offsAnchorToMiddleX = offsAnchorToBBBorderX + (bb.maxX - bb.minX + 1) / 2F;

			int offsAnchorToBBBorderY = bb.minY - changedMDI.getPlacement().getAnchor().getRow();
			float offsAnchorToMiddleY = offsAnchorToBBBorderY + (bb.maxY - bb.minY + 1) / 2F;

			g2d.setColor(Color.RED);
			final int strokeWidth = 10;
			g2d.setStroke(new BasicStroke(strokeWidth));


			Tile startTile = oldAnchor;
			Tile endTile = changedMDI.getPlacement().getAnchor();

			int startX = (int) (tileSize * tileXMap.get(startTile) + offsAnchorToMiddleX * tileSize);
			int startY = (int) (tileSize * tileYMap.get(startTile) + offsAnchorToMiddleY * tileSize);

			int endX = (int) (tileSize * tileXMap.get(endTile) + offsAnchorToMiddleX * tileSize);
			int endY = (int) (tileSize * tileYMap.get(endTile) + offsAnchorToMiddleY * tileSize);

			drawArrow(g2d, startX, startY, endX, endY);

		}
	}

	@Override
	public void tryMove(
			@NotNull ModuleDescriptionInstance mdiToChange,
			@NotNull ModulePlacement newPlacement
	) {
		if (mdiToChange.isPlaced()) {
			pendingMove = true;
			changedMDI = mdiToChange;
			oldAnchor = changedMDI.getPlacement().getAnchor();
		}
	}

	@Override
	public void moveAccepted() {
		pendingMove = false;
	}

	@Override
	public void moveRejected() {
		pendingMove = false;
		changedMDI = null;
	}

	@Override
	public void initialize(
			@NotNull Collection<ModuleDescriptionInstance> moduleDescriptionInstances,
			@NotNull Set<PortNet> portNets
	) {
		this.moduleDescriptionInstances = moduleDescriptionInstances;
		pendingMove = false;
		changedMDI = null;
		int[] id = {0};
		mdiColor = moduleDescriptionInstances.stream().collect(Collectors.toMap(
				Function.identity(),
				m -> colors[id[0]++ % colors.length])
		);
	}
}
