/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.rapidSoc.visualization;

import de.tu_darmstadt.rs.util.StreamUtil;
import de.tu_darmstadt.rs.util.tracer.Tracer;

import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.Tile;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.DisplayMode;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.Point;
import java.awt.Transparency;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.lang.ref.SoftReference;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.stream.Stream;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.SwingUtilities;

/**
 * UI for Visualizations
 */
//Since we are in ui, there are many constant numbers...
@SuppressWarnings("checkstyle:MagicNumber")
public class VisualizationFrame<T extends BaseVisualization> extends JFrame {

	private static final Logger logger = LoggerFactory.getLogger(VisualizationFrame.class);

	protected final T visualization;

	private int displayedTileSize = 20;


	Integer drawRequest = null;
	private Thread drawerThread = null;
	boolean stopping = false;
	private final Object drawerSynchronization = new Object();


	/**
	 * Create a frame
	 *
	 * @param visualization the visualization to show
	 */
	public VisualizationFrame(T visualization) {
		this.visualization = visualization;
		setTitle(visualization.toString());
		setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
		addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosed(WindowEvent e) {
				synchronized (VisualizationFrame.this) {
					VisualizationFrame.this.notify();
				}
			}
		});
		drawerThread = new Thread(() -> {
			while (!stopping) {
				if (drawRequest == null) {
					try {
						synchronized (drawerSynchronization) {
							drawerSynchronization.wait();
						}
					} catch (InterruptedException e) {
					}
				}
				if (drawRequest != null && !stopping) {
					int tileSize;
					synchronized (drawerSynchronization) {
						tileSize = drawRequest;
						drawRequest = null;
					}


					logger.info("drawer running for {}", tileSize);
					BufferedImage img = createImage(tileSize);
					oldCachedImages.put(tileSize, new SoftReference<>(img));

					logger.info("drawer finished for {}", tileSize);
					SwingUtilities.invokeLater(() -> {
						//Only repaint if still the one to be displayed
						if (displayedTileSize == tileSize) {
							repaint();
						}
					});
				}
			}
		}, "Visualization drawer");
	}


	/**
	 * Open blocking UI
	 */
	public synchronized void openBlockingUI() {
		Tracer.enterLoop("Visualization UI");
		display();

		while (isVisible()) {
			try {
				wait();
			} catch (InterruptedException e) {
			}
		}
		stopping = true;
		synchronized (drawerSynchronization) {
			drawerSynchronization.notifyAll();
		}
		try {
			drawerThread.join();
		} catch (InterruptedException e) {
		}
		Tracer.exit();
	}

	/**
	 * Display the frame
	 */
	public void display() {
		setLayout(new BorderLayout());
		add(new PainterComponent(), BorderLayout.CENTER);
		Component bottomComponent = createBottomComponent();
		if (bottomComponent != null) {
			add(bottomComponent, BorderLayout.PAGE_END);
		}
		pack();
		setMinimumSize(getSize());
		DisplayMode dm = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice().getDisplayMode();
		setSize(new Dimension(dm.getWidth(), dm.getHeight() * 3 / 4));

		applyCustomization();

		drawerThread.start();
		setVisible(true);
	}

	protected void applyCustomization() {
	}

	protected Component createBottomComponent() {
		return null;
	}

	private Map<Integer, SoftReference<BufferedImage>> oldCachedImages = Collections.synchronizedMap(new HashMap<>());

	private BufferedImage createImage(int tileSize) {
		GraphicsDevice gd = GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice();
		final Point sceneSize = visualization.getSceneSize(tileSize);
		BufferedImage image = gd.getDefaultConfiguration().createCompatibleImage(sceneSize.x, sceneSize.y, Transparency.OPAQUE);

		Graphics2D g2d = image.createGraphics();
		visualization.paint(g2d, tileSize);

		g2d.dispose();

		return image;
	}

	/**
	 * The component that shows the Visualization
	 */
	private class PainterComponent extends JComponent implements MouseListener, MouseWheelListener, MouseMotionListener {

		private static final long serialVersionUID = 1L;
		int offsX = 0;
		int offsY = 0;
		private Point dragStart;

		public PainterComponent() {
			addMouseListener(this);
			addMouseWheelListener(this);
			addMouseMotionListener(this);
		}

		private BufferedImage cachedImage = null;


		private BufferedImage getCachedImage2(int tileSize) {
			final SoftReference<BufferedImage> ref = oldCachedImages.get(tileSize);
			if (ref != null) {
				BufferedImage img = ref.get();
				if (img != null) {
					return img;
				}
			}
			return null;
		}

		private StreamUtil.Pair<BufferedImage, Integer> getPlaceholderImage(int tileSize) {
			return oldCachedImages.entrySet().stream()
					.flatMap(entry -> {
						BufferedImage img = entry.getValue().get();
						if (img == null) {
							return Stream.empty();
						}
						return Stream.of(new StreamUtil.Pair<>(img, entry.getKey()));
					})
					.min(Comparator.comparing(pair -> Math.abs(tileSize - pair.getB())))
					.orElse(new StreamUtil.Pair<>(null, 0));
		}

		private StreamUtil.Pair<BufferedImage, Integer> getOrCalcImage(int tileSize) {
			Tracer.enterLoop("get cached image 2");
			BufferedImage cached = getCachedImage2(tileSize);
			Tracer.exit();
			if (cached != null) {
				//logger.info("using cached image for {}", tileSize);
				return new StreamUtil.Pair<>(cached, null);
			}
			logger.info("need new image for {}", tileSize);

			Tracer.enterLoop("notify drawer");
			synchronized (drawerSynchronization) {
				drawRequest = tileSize;
				drawerSynchronization.notifyAll();
			}
			Tracer.exit();
			return getPlaceholderImage(tileSize);
		}


		@Override
		public void paintComponent(Graphics g) {
			super.paintComponent(g);

			Tracer.enterLoop("paint");

			Graphics2D g2d = (Graphics2D) g.create();

			//AffineTransform t = g2d.getTransform();

			//AffineTransform nt = AffineTransform.getTranslateInstance(offsX, offsY);
			//g2d.setTransform(nt);


			BufferedImage img;
			Integer rescale = null;


			//logger.info("Getting image for " + displayedTileSize);
			if (cachedImage == null) {
				Tracer.enterLoop("Get image");
				final StreamUtil.Pair<BufferedImage, Integer> res = getOrCalcImage(displayedTileSize);
				Tracer.exit();
				img = res.getA();
				rescale = res.getB();
				if (rescale == null) {
					cachedImage = img;
				}
			} else {
				img = cachedImage;
			}


			if (rescale == null) {
				Tracer.enterLoop("paint correct image");
				//logger.info("drawing correct image");
				g2d.drawImage(img, offsX, offsY, null);
				Tracer.exit();
			} else {
				Tracer.enterLoop("paint backup image");
				//logger.info("drawing backup image");
				final Point sceneSize = visualization.getSceneSize(displayedTileSize);
				g2d.drawImage(img, offsX, offsY, sceneSize.x, sceneSize.y, null);
				Tracer.exit();
			}

			Tracer.exit();
			//g2d.setTransform(t);
		}

		@Override
		public Dimension getPreferredSize() {
			return new Dimension(400, 300);
		}

		@Override
		public Dimension getMinimumSize() {
			return new Dimension(100, 100);
		}

		@Override
		public void mouseWheelMoved(MouseWheelEvent mouseWheelEvent) {
			if (mouseWheelEvent.getWheelRotation() == 0) {
				return;
			}

			int oldTileSize = displayedTileSize;

			//Scale by factor
			int newTileSize = (int) ((mouseWheelEvent.getWheelRotation() > 0)? oldTileSize * 1.02 : oldTileSize / 1.02);

			//If the change gets rounded away, add/subtract a pixel
			if (newTileSize == oldTileSize) {
				newTileSize = oldTileSize + ((mouseWheelEvent.getWheelRotation() > 0)? 1 : -1);
			}

			//Go no smaller than 2
			if (newTileSize < 2) {
				newTileSize = 2;
			}


			int x = mouseWheelEvent.getX();
			int y = mouseWheelEvent.getY();

			offsX = (offsX - x) * newTileSize / oldTileSize + x;
			offsY = (offsY - y) * newTileSize / oldTileSize + y;

			VisualizationFrame.this.displayedTileSize = newTileSize;

			cachedImage = null;
			repaint();
		}

		@Override
		public void mouseDragged(MouseEvent mouseEvent) {
			Point p = new Point(mouseEvent.getPoint());
			offsX = p.x - dragStart.x;
			offsY = p.y - dragStart.y;
			repaint();
		}

		@Override
		public void mouseMoved(MouseEvent mouseEvent) {

		}

		@Override
		@SuppressWarnings("checkstyle:SystemOut")
		public void mouseClicked(MouseEvent mouseEvent) {
			Tracer.enterLoop("clicked");
			int x = mouseEvent.getX() - offsX;
			int y = mouseEvent.getY() - offsY;
			Tile t = visualization.getTile(x, y, displayedTileSize);

			if (t != null) {
				System.out.println(t.getName());
				if (t.getPrimitiveSites() != null) {
					for (PrimitiveSite primitiveSite : t.getPrimitiveSites()) {
						Instance i = visualization.getInstanceAtPrimitiveSite(primitiveSite);
						System.out.println(primitiveSite + ": " + (i != null? i.getName() : "empty"));
					}
				} else {
					System.out.println("no primitive sites");
				}

				final Map<Tile, Set<Net>> routingOnlyTiles = visualization.getRoutingOnlyTiles();
				if (routingOnlyTiles != null) {
					Set<Net> routethroughs = routingOnlyTiles.get(t);
					if (routethroughs != null) {
						System.out.println("Routethrough for " + routethroughs);
					}
				}
				System.out.println();

			}
			Tracer.exit();

		}

		@Override
		public void mousePressed(MouseEvent mouseEvent) {
			dragStart = new Point(mouseEvent.getPoint());
			dragStart.translate(-offsX, -offsY);
		}

		@Override
		public void mouseReleased(MouseEvent mouseEvent) {
			dragStart = null;
		}

		@Override
		public void mouseEntered(MouseEvent mouseEvent) {

		}

		@Override
		public void mouseExited(MouseEvent mouseEvent) {

		}
	}

}
