/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

import de.tu_darmstadt.rs.spartanmc.jconfig.util.LibJConfigBuilder;

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.device.Device;
import edu.byu.ece.rapidSmith.device.PrimitiveSite;
import edu.byu.ece.rapidSmith.device.WireConnection;
import edu.byu.ece.rapidSmith.device.WireEnumerator;
import edu.byu.ece.rapidSmith.primitiveDefs.PrimitiveDefList;
import edu.byu.ece.rapidSmith.router.Node;
import edu.byu.ece.rapidSmith.util.DeviceDatabaseProvider;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.stream.Stream;


/**
 * Generate JConfig device descriptions from RapidSmith devices.
 * TODO migrate to JC3!
 *
 * Created by wenzel on 03.06.16.
 */
public final class DeviceDescriptionGenerator {

	private static PrimitiveDefList primitiveDefs;
	private static String devName;

	private DeviceDescriptionGenerator() {

	}

	/**
	 * Stream an array that may be null. In that case the stream will be empty.
	 * @param values the array
	 * @param <T> element type
	 * @return stream of array elements
	 */
	//TODO move to libutil after jc3 merge
	public static <T> Stream<T> streamOfNullable(T[] values) {
		if (values == null) {
			return Stream.empty();
		}
		return Arrays.stream(values);
	}

	@SuppressWarnings("checkstyle:SystemOut")
	private static void explore(Node n, WireEnumerator we) {
		System.out.println(n.toString(we));
		final int maxLevel = 8;
		if (n.getLevel() > maxLevel) {
			return;
		}

		if (n.getConnections() != null) {
			for (WireConnection wireConnection : n.getConnections()) {
				explore(wireConnection.createNode(n), we);
			}
		}
	}

	private static boolean isClockCapable(PrimitiveSite p, WireEnumerator we) {
		if (devName.toLowerCase().contains("fgg484")) {
			return p.getName().matches("[AB]1[02]");
		} else if (devName.toLowerCase().contains("fgg900")) {
			return p.getName().matches("[ABCD]15");
		} else {
			throw new RuntimeException("unknown");
		}
	}

	private static void writeTarget(Device device) throws IOException {
		String name = device.getPartName();

		Path file = LibJConfigBuilder.getEnvSpartanRoot().resolve("lib/targets/" + name + ".xml");


		try (PrintWriter p = new PrintWriter(Files.newBufferedWriter(file))) {
			p.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
			p.println("<target xmlns=\"soc-library\" document-type=\"soc-target-1.0\" name=\"" + name + "\">");
			p.println("  <description>Raw " + name + "</description>");
			p.println("  <device name=\"" + name + "\" jtagid=\"1\" />");
			p.println("  <iopins name=\"P\">");

			Design.iobTypes.stream()
					.flatMap(type -> streamOfNullable(device.getAllPrimitiveSitesOfType(type)))
					.forEach(ps -> {
						String clock = isClockCapable(ps, device.getWireEnumerator())? "freq=\"100000000\" " : "";
						p.println("    <pin name=\"" + ps + "\" fpgapin=\"" + ps + "\" descr=\"" + ps + "\" " + clock + "/>");
					});

			p.println("  </iopins>");
			p.println("</target>");
		}
	}

	private static void writeDevice(Device device) throws IOException {
		String name = device.getPartName();
		Path file = LibJConfigBuilder.getEnvSpartanRoot().resolve( "lib/devices/" + name + ".xml");
		try (PrintWriter p = new PrintWriter(Files.newBufferedWriter(file))) {
			p.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n" +
					"<?xml-model href=\"device.xsd\" schematypens=\"http://www.w3.org/2001/XMLSchema\"?>\n" +
					"<device xmlns=\"soc-library\" document-type=\"soc-device-1.0\" name=\"" + name + "\">\n" +
					"  <toolchain name=\"xilinx\">\n" +
					"    <param name=\"SYNTHESIS\" value=\"-p " + name + " -ifmt mixed -ofmt NGC -opt_mode Speed -opt_level 1 -iuc NO -keep_hierarchy YES -glob_opt ALLClockNets -rtlview YES -read_cores YES -write_timing_constraints NO -cross_clock_analysis NO -hierarchy_separator _ -bus_delimiter [] -case maintain -slice_utilization_ratio 100 -fsm_extract YES -fsm_encoding Auto -fsm_style lut -ram_extract Yes -ram_style Auto -rom_extract Yes -rom_style Auto -shreg_extract YES -resource_sharing YES -mult_style auto -iobuf YES -max_fanout 500 -bufg 8 -register_duplication YES -equivalent_register_removal YES -register_balancing No -optimize_primitives No -iob auto -slice_utilization_ratio_maxmargin 5\" />\n" +
					"    <param name=\"TRANSLATE\" value=\"-nt on -p " + name + "\" />\n" +
					"    <param name=\"MAP\" value=\"-timing -p " + name + " -logic_opt on -pr b -register_duplication -xe n -ol high\" />\n" +
					"    <param name=\"PAR\" value=\"-ol high -xe n\" />\n" +
					"    <param name=\"NETGEN\" value=\"-s 3\" />\n" +
					"    <param name=\"BITGEN\" value=\"-m\" />\n" +
					"    <param name=\"PROGRAM\" value=\"\" />\n" +
					"  </toolchain>\n" +
					"  <iostandard>\n" +
					"    LVTTL, LVCMOS33, LVCMOS25, LVCMOS18, LVCMOS15,\n" +
					"    LVCMOS12, PCI, I2C, SMBUS, SDIO,\n" +
					"    HSTL_I, HSTL_III, HSTL_IV, HSTL_I_18, HSTL_III_18,\n" +
					"    HSTL_IV_18, SSTL3_I, SSTL3_II, SSTL2_I, SSTL2_II,\n" +
					"    SSTL18_I, SSTL18_II, DIFF_SSTL18_II, SSTL15_II, LVDS_25\n" +
					"  </iostandard>\n" +
					"  <drive> 2,4,6,8,12,16,24\n" +
					"  </drive>\n" +
					"</device>\n");
		}
	}

	/**
	 * Main
	 * @param args arguments
	 * @throws IOException on io exception
	 */
	public static void main(String[] args) throws IOException {
		devName = args[0];
		Device device = DeviceDatabaseProvider.getDeviceDatabase().loadDevice(args[0]);
		DeviceDatabaseProvider.getDeviceDatabase().loadWireEnumerator(device.getFamilyType());
		primitiveDefs = DeviceDatabaseProvider.getDeviceDatabase().loadPrimitiveDefs(device.getFamilyType());

		writeTarget(device);
		writeDevice(device);

	}
}
