/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

import edu.byu.ece.rapidSmith.design.Design;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Net;
import edu.byu.ece.rapidSmith.design.PIP;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.device.WireConnection;
import edu.byu.ece.rapidSmith.device.WireEnumerator;
import edu.byu.ece.rapidSmith.primitiveDefs.PrimitiveDefList;
import edu.byu.ece.rapidSmith.router.Node;
import edu.byu.ece.rapidSmith.util.DeviceDatabaseProvider;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Stream;

/**
 * Testing: Print predecessors of select pins
 * Created by jwenzel on 30.12.15.
 */
@SuppressWarnings("checkstyle:SystemOut")
public class PrintPredecessors {
	private static PrimitiveDefList primitiveDefs;

	public PrintPredecessors() {

	}

	/**
	 * Entry point
	 * @param args arguments
	 */
	public static void main(String[] args) {
		Design design = new Design();
		design.loadXDLFile(Paths.get(args[0]));


		String[][] elementsOfInterest = new String[][]{
				{"SLICE_X36Y37", "CLK"},
				{"SLICE_X33Y38", "CLK"},
				{"SLICE_X34Y37", "CLK"},
				{"RAMB16_X1Y14", "CLKB"}
		};


		primitiveDefs = DeviceDatabaseProvider.getDeviceDatabase().loadPrimitiveDefs(design.getDevice().getFamilyType());

		for (String[] strings : elementsOfInterest) {
			String site = strings[0];
			String pin = strings[1];


			Instance inst = design.getInstanceAtPrimitiveSite(design.getDevice().getPrimitiveSite(site));
			if (inst == null) {
				throw new RuntimeException("instance at site " + site + " not found");
			}

			Pin p = inst.getPin(pin);
			Net n = p.getNet();

			try (PrintStream out = new PrintStream(new FileOutputStream("out.dot"))) {

				printConnections(n, p, design, out);
			} catch (FileNotFoundException e) {
				e.printStackTrace();
			}


		}

	}

	private static String niceName(Node n) {
		WireEnumerator we = n.getTile().getDevice().getWireEnumerator();
		Objects.requireNonNull(we);
		return n.getTile() + "." + we.getWireName(n.getWire());
	}

	private static void printNode(Node n, Predicate<Node> enabled, Set<Node> seen, PrintStream out, Node sinkNode) {

		/*if (n.getParent()!=null)
			out.println(niceName(n.getParent())+" -> "+niceName(n)+";");*/
		if (n.equals(sinkNode)) {
			sinkFound(out, n);
		}
		if (seen.add(n)) {
			getNextNodes(n).filter(enabled).forEach(next -> printNode(next, enabled, seen, out, sinkNode));
		}
	}

	private static void sinkFound(PrintStream out, Node sinkNode) {
		Node n = sinkNode;
		List<Node> nodes = new ArrayList<>();
		while (n != null) {
			nodes.add(n);
			n = n.getParent();
		}
		Collections.reverse(nodes);
		nodes.stream().forEach(node -> {
			System.out.println(niceName(node));
		});
	}

	private static void printConnections(Net net, Pin pin, Design design, PrintStream out) {
		out.println("digraph G {");
		/*net.getPIPs().forEach(pip -> {
			WireEnumerator we = pip.getTile().getDevice().getWireEnumerator();
			System.out.println( "  pip " + pip.getTile().getName() + " " + we.getWireName(pip.getStartWire()) +
					" -> " + we.getWireName(pip.getEndWire()));

		});*/
		System.out.println("Sink: " + pin);

		Node sourceNode = new Node(net.getSource().getTile(), design.getDevice().getPrimitiveExternalPin(net.getSource()), null, 0);

		Node sinkNode = new Node(pin.getTile(), design.getDevice().getPrimitiveExternalPin(pin), null, 0);



		/*getNextNodes(sourceNode).forEach(node ->{
			System.out.println(node.toString());
		});*/
		printNode(sourceNode, isNodeEnabled(net), new HashSet<>(), out, sinkNode);

		//Node sinkNode = new Node(sink.getTile(), device.getPrimitiveExternalPin(sink), null, 0); //parent is null?
		out.println("}");
		System.out.println("==");
		System.out.println("");
	}

	private static Predicate<Node> isNodeEnabled(Net net) {
		return node -> {
			if (!node.isPIP()) {
				return true;
			}

			PIP fwPip = new PIP(node.getTile(), node.getParent().getWire(), node.getWire());
			PIP bwPip = new PIP(node.getTile(), node.getWire(), node.getParent().getWire());
			return net.getPIPs().contains(fwPip) || net.getPIPs().contains(bwPip);
		};
	}

	private static Stream<Node> getNextNodes(Node node) {
		WireConnection[] conns = node.getConnections();
		if (conns == null) {
			return Stream.empty();
		}

		return Stream.of(conns).map(conn -> {
			if (conn == null) {
				throw new RuntimeException("conn is null!!!");
			}
			return conn.createNode(node);
		});
	}
}
