/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.rapidSoc.module_synthesis;

import de.tu_darmstadt.rs.rapidSoc.RapidSocProject;
import de.tu_darmstadt.rs.rapidSoc.directories.ModuleRootPath;
import de.tu_darmstadt.rs.rapidSoc.directories.VariantPath;
import de.tu_darmstadt.rs.rapidSoc.module.JacksonUtil;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescription;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionLoader;
import de.tu_darmstadt.rs.rapidSoc.module.ModuleDescriptionVariant;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.util.StreamUtil;

import edu.byu.ece.rapidSmith.design.Attribute;
import edu.byu.ece.rapidSmith.design.Instance;
import edu.byu.ece.rapidSmith.design.Module;
import edu.byu.ece.rapidSmith.design.NetType;
import edu.byu.ece.rapidSmith.design.Pin;
import edu.byu.ece.rapidSmith.design.Port;
import org.apache.commons.io.FileUtils;
import org.jetbrains.annotations.NotNull;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.function.Consumer;
import java.util.stream.Collectors;

public class ModuleExtractorTest {
	private static final Logger logger = LoggerFactory.getLogger(ModuleExtractorTest.class);

	ModuleRootPath moduleRootPath;
	private RapidSocTestingProjectProvider projectProvider;
	private RapidSocProject project;

	@Before
	public void init() throws IOException {
		moduleRootPath = new ModuleRootPath(temporaryFolder.newFolder("modules").toPath());
		projectProvider = RapidSocTestingProjectProvider.getInstance();
		project = projectProvider.getRapidSocProject(null, "SP601", moduleRootPath, UnconstrainedConstraintCreator::new);

	}

	@Rule
	public TemporaryFolder temporaryFolder = new TemporaryFolder();

	private String getExpectedLutFunction(Port in1, Port in2) {
		String name1 = getLutName(in1);
		String name2 = getLutName(in2);
		return "#LUT:O6=(" + name1 + "@" + name2 + ")";
	}

	@NotNull
	private String getLutName(Port port) {
		return port.getPinName().replaceAll("[BCD]","A");
	}


	private Module buildAndLoadModule(String moduleName) throws IOException {
		final IModuleDescription moduleDescription = projectProvider.getLibrary().getModuleDescription(moduleName);
		HashMap<String, String> attrMap = new HashMap<>();
		Set<IModuleDescription> descriptions = Collections.singleton(moduleDescription);

		//Force running extractor again by removing its output
		VariantPath variantPath = moduleRootPath.resolveHashDir(descriptions, attrMap).resolveVariant("0");
		FileUtils.deleteDirectory(variantPath.resolveSmilePath().toFile());


		final ModuleDescriptionLoader loader = new ModuleDescriptionLoader(project);

		ModuleDescription md = loader.loadAllVariants(descriptions, attrMap);


		ModuleDescriptionVariant mdv = md.getVariants().stream().collect(StreamUtil.exactlyOne());
		return mdv.getModule();
	}

	private Port ensurePortExists(Module module, String portName) {
		Port port = module.getPort(portName);
		Assert.assertNotNull("Did not find Port "+portName, port);
		return port;
	}

	private Port ensurePortExistsOnInstance(Module module, String portName, Collection<Instance> instances) {
		Port port = ensurePortExists(module, portName);
		Assert.assertEquals(new HashSet<>(instances), port.getPins().stream().map(Pin::getInstance).collect(Collectors.toSet()));
		return port;
	}

	private Port ensurePortExistsOnInstance(Module module, String portName, Instance... instances) {
		return ensurePortExistsOnInstance(module, portName, Arrays.asList(instances));
	}


	private void testAndDumpOnError(String moduleName, Consumer<Module> test) throws IOException {
		Module module = buildAndLoadModule(moduleName);
		try {
			test.accept(module);
		} catch (Exception | AssertionError e) {
			logger.error("Error in Testing module "+ module.getName()+", dumping info:");
			JacksonUtil.printModule(module);
			throw e;
		}
	}

	@Test
	public void testExtractorSimple() throws IOException {
		testAndDumpOnError("rapidsocTestSimple", module -> {

			Assert.assertEquals(3, module.getInstances().size());

			Instance xor = module.getInstances().stream().filter(inst -> inst.getName().endsWith("$out")).collect(StreamUtil.exactlyOne());
			Instance reg1 = module.getInstances().stream().filter(inst -> inst.getName().endsWith("$outreg")).collect(StreamUtil.exactlyOne());
			Instance reg2 = module.getInstances().stream().filter(inst -> inst.getName().endsWith("$outreg2")).collect(StreamUtil.exactlyOne());


			Port clk = ensurePortExistsOnInstance(module, "clk.I", reg1, reg2);
			Port in1 = ensurePortExistsOnInstance(module, "in1.I", xor);
			Port in2 = ensurePortExistsOnInstance(module, "in2.I", xor);
			Port out = ensurePortExistsOnInstance(module, "out.O", xor);
			Port outDuplicate = ensurePortExistsOnInstance(module, "outDuplicate.O", xor);
			Assert.assertEquals(out, outDuplicate);
			Port outReg = ensurePortExistsOnInstance(module, "outReg.O", reg1);
			Port outReg2 = ensurePortExistsOnInstance(module, "outReg2.O", reg2);

			Port static1 = ensurePortExists(module, "static1.O");
			Port static0 = ensurePortExists(module, "static0.O");


			Assert.assertEquals(NetType.GND, static0.getNetType());
			Assert.assertEquals(NetType.VCC, static1.getNetType());

			Assert.assertEquals("CLK", clk.getPins().stream().map(Pin::getName).collect(StreamUtil.atLeastOneOrEqual()));

			Assert.assertEquals(out.getPinName(), in1.getPinName().substring(0, 1));
			Assert.assertEquals(out.getPinName(), in2.getPinName().substring(0, 1));
			Assert.assertTrue(outReg.getPinName().endsWith("Q"));
			Assert.assertTrue(outReg2.getPinName().endsWith("Q"));

			Attribute lut6 = xor.getAttribute(out.getPinName() + "6LUT");
			Attribute lut5 = xor.getAttribute(out.getPinName() + "5LUT");


			Assert.assertEquals("#OFF", lut5.getValue());
			Assert.assertEquals(getExpectedLutFunction(in1, in2), lut6.getValue());
		});
	}


	private void verifyPassthrough(Port in, Port out) {
		Assert.assertEquals(in, out.getPassthroughFrom());
		Assert.assertTrue(in.getName()+" should have "+out.getName()+" in its passthrough ports", in.getPassthroughPorts().contains(out));
	}

	private void verifyRegister(Port clk, Port in, Port out) {
		Pin inPin = in.getPin();
		Pin outPin = out.getPin();

		Assert.assertSame(inPin.getInstance(), outPin.getInstance());
		Assert.assertTrue(clk.getPins().contains(inPin.getInstance().getPin("CLK")));

		char subSliceIndex = inPin.getName().charAt(0);

		Assert.assertEquals(subSliceIndex+"X", inPin.getName());
		Assert.assertEquals(subSliceIndex+"Q", outPin.getName());

		Attribute ffAttr = inPin.getInstance().getAttribute(subSliceIndex + "FF");
		Assert.assertTrue(!ffAttr.getValue().isEmpty());
		Assert.assertNotEquals("#OFF",ffAttr.getValue());
	}


	@Test
	public void testExtractorPassthrough() throws IOException {
		testAndDumpOnError("rapidsocTestPassthrough", module -> {

			Assert.assertEquals(1, module.getInstances().size());

			Instance inst = module.getInstances().stream().collect(StreamUtil.exactlyOne());


			Port clk = ensurePortExistsOnInstance(module, "clk.I");
			Port in1 = ensurePortExistsOnInstance(module, "in1.I", inst);
			Port in2 = ensurePortExistsOnInstance(module, "in2.I", inst);
			Port passthrough1 = ensurePortExistsOnInstance(module, "passthrough1.O");
			Port passthrough2 = ensurePortExistsOnInstance(module, "passthrough2.O");
			Port passthroughClk = ensurePortExistsOnInstance(module, "passthroughClk.O");

			verifyPassthrough(in1, passthrough1);
			verifyPassthrough(in1, passthrough2);
			verifyPassthrough(clk, passthroughClk);

		});
	}



	@Test
	public void testExtractorTristate() throws IOException {
		testAndDumpOnError("rapidsocTestTristate", module -> {

			Assert.assertEquals(3, module.getInstances().size());

			Port clk = ensurePortExistsOnInstance(module, "clk.I", module.getInstances());
			Port tristateI = ensurePortExists(module, "tristate.I");
			Port tristateO = ensurePortExists(module, "tristate.O");
			Port tristateT = ensurePortExists(module, "tristate.T");

			Port tristateDirectI = ensurePortExists(module, "tristateDirect.I");
			Port tristateDirectO = ensurePortExists(module, "tristateDirect.O");
			Port tristateDirectT = ensurePortExists(module, "tristateDirect.T");

			Port tristateData = ensurePortExists(module, "tristateData.I");
			Port tristateDataDirect = ensurePortExists(module, "tristateDataDirect.I");
			Port tristateSelect = ensurePortExists(module, "tristateSelect.I");
			Port tristateSelectDirect = ensurePortExists(module, "tristateSelectDirect.I");
			Port tristateRead = ensurePortExists(module, "tristateRead.O");
			Port tristateReadDirect = ensurePortExists(module, "tristateReadDirect.O");


			verifyPassthrough(tristateDataDirect, tristateDirectO);
			verifyPassthrough(tristateSelectDirect, tristateDirectT);
			verifyPassthrough(tristateDirectI, tristateReadDirect);

			verifyRegister(clk, tristateData, tristateO);
			verifyRegister(clk, tristateSelect, tristateT);
			verifyRegister(clk, tristateI, tristateRead);

		});
	}

	@Test
	public void testExtractorOnePin() throws IOException {
		testAndDumpOnError("rapidsocTestOnePin", module -> {

			Assert.assertEquals(1, module.getInstances().size());

			Port clk = ensurePortExistsOnInstance(module, "clk.I", module.getInstances());
			Port in = ensurePortExistsOnInstance(module, "in.I", module.getInstances());
			Port out = ensurePortExistsOnInstance(module, "out.O", module.getInstances());

			verifyRegister(clk, in, out);

		});
	}


	@Test
	public void testExtractorWidePort() throws IOException {
		testAndDumpOnError("rapidsocTestWidePort", module -> {

			Port clk = ensurePortExistsOnInstance(module, "clk.I", module.getInstances());

			for (int i=0;i<5;i++) {
				Port in1 = ensurePortExists(module, "in1["+i+"].I");
				Port in2 = ensurePortExists(module, "in2["+i+"].I");
				Port out1 = ensurePortExists(module, "out1["+i+"].O");
				Port out2 = ensurePortExists(module, "out2["+i+"].O");

				verifyRegister(clk, in1, out1);
				verifyRegister(clk, in2, out2);
			}

		});
	}

}
