package dispatcher.gui;

import java.util.*;
import java.net.*;
import com.trolltech.qt.*;
import com.trolltech.qt.core.*;
import com.trolltech.qt.gui.*;

import dispatcher.model.*;

import libsim.SimulationSpeed_I;

import libjconfig.common.Log;
import libjconfig.common.LogTitle;

/**
 * Controls the GUI via signals from the dispatcher model.
 *
 * @author Michael Raitza
 */
public class GuiController extends QSignalEmitter {

    private Dispatcher_widget ui;
    private Model dm;
    private Qt.MatchFlag[] exact_matcher = new Qt.MatchFlag[]{ Qt.MatchFlag.MatchExactly };
    private LogTitle lt = Log.get("GuiController");
    private static final int gui_role = Qt.ItemDataRole.UserRole + 1;
    private static final int ui_role = Qt.ItemDataRole.UserRole + 2;
    private static final int flags_role = Qt.ItemDataRole.UserRole + 3;
    private QListWidgetItem log_item;
    private QColor unconn_clr = new QColor(255, 159, 159, 255);

    public GuiController(Dispatcher_widget ui, Model dm) {
        this.ui = ui;
        this.dm = dm;
        ui.splitter.setSizes(new ArrayList<Integer>(){ { add(20); add(0); } });

        /* initialise combo box */
        int longest_item = 0;
        for (SimulationSpeed_I.Speed ss : SimulationSpeed_I.Speed.values()) {
            ui.mode_combobox.insertItem(ss.ordinal(), ss.toString());
            longest_item = longest_item < ss.toString().length() ? ss.toString().length() : longest_item;
        }
        ui.mode_combobox.setCurrentIndex(0);
        ui.mode_combobox.setMinimumContentsLength(longest_item);

        /* instantiate the log item */
        log_item = new QListWidgetItem();
        log_item.setText("Log");
        QWidget lvw = new QWidget(ui.main_pane);
        ui.main_pane_layout.addWidget(lvw);
        log_view lv = new log_view();
        lv.setupUi(lvw);
        log_item.setData(gui_role, lvw);
        log_item.setData(ui_role, lv);
        log_item.setData(flags_role, new Integer(1));
        ui.main_list.addItem(log_item);

        /* fill items with data */
        doc_load(false, null, null, null);
        sim_state(SimState.UNINITIALIZED);
        sim_speed(ui.mode_combobox.currentIndex());
    }

    private void update_client_view(client_view cv, ClientModel c) {
        /* setup labels */
        cv.name_value_label.setText(c.name());
        cv.id_value_label.setText("0x" + Integer.toHexString(c.id()));
        String conn_value;
        if (c.address() != null) {
            conn_value = (new Formatter()).format("%s:%s", c.address().getHostName(), c.port()).toString();
        } else {
            conn_value = "- unconnected -";
        }
        cv.conn_value_label.setText(conn_value);
    }

    /**
     * Reacts on simulation state changes.
     */
    public void sim_state(SimState state) {
        switch(state) {
        case UNINITIALIZED:
            ui.run_button.setEnabled(false);
            ui.stop_button.setEnabled(false);
            ui.step_button.setEnabled(false);
            ui.reset_button.setEnabled(false);
            ui.mode_combobox.setEnabled(false);
            break;
        case UNCONNECTED:
        case RESET:
            ui.run_button.setEnabled(false);
            ui.stop_button.setEnabled(false);
            ui.step_button.setEnabled(false);
            ui.reset_button.setEnabled(false);
            ui.mode_combobox.setEnabled(true);
            break;
        case STOPPED:
            ui.run_button.setEnabled(true);
            ui.stop_button.setEnabled(false);
            ui.step_button.setEnabled(true);
            ui.reset_button.setEnabled(true);
            ui.mode_combobox.setEnabled(true);
            break;
        case RUNNING:
            ui.run_button.setEnabled(false);
            ui.stop_button.setEnabled(true);
            ui.step_button.setEnabled(true);
            ui.reset_button.setEnabled(true);
            ui.mode_combobox.setEnabled(true);
            break;
        default:
        }
    }

    public void doc_load(boolean loaded, Collection<ClientModel> clients, Collection<InterconnectModel> ics, Collection<SharedMemoryModel> shms) {
        /* remove all items; close widgets of normal items (clients) */
        while (ui.main_list.count() > 0) {
            Integer special = (Integer)ui.main_list.item(0).data(flags_role);
            if (special.equals(0))
                ((QWidget)ui.main_list.item(0).data(gui_role)).close();
            ui.main_list.takeItem(0);
        }
        if (loaded) {
            for (ClientModel c : clients) {
                /* initialise basic gui elements */
                QListWidgetItem item = new QListWidgetItem();
                item.setData(Qt.ItemDataRole.UserRole, c);
                QWidget cvw = new QWidget(ui.main_pane);
                cvw.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose);
                ui.main_pane_layout.addWidget(cvw);
                cvw.setVisible(false);
                client_view cv = new client_view();
                cv.setupUi(cvw);

                /* initialise FIFO table */
                cv.fifo_table.horizontalHeader().setHighlightSections(false);
                cv.fifo_table.horizontalHeader().setResizeMode(QHeaderView.ResizeMode.ResizeToContents);
                cv.fifo_table.horizontalHeader().setClickable(false);
                cv.fifo_table.verticalHeader().setHighlightSections(false);
                cv.fifo_table.verticalHeader().setResizeMode(QHeaderView.ResizeMode.Fixed);
                cv.fifo_table.verticalHeader().setClickable(false);
                int min_height = cv.fifo_table.verticalHeader().fontMetrics().height() + 4;
                if (min_height < 16)
                    min_height = 16;
                cv.fifo_table.verticalHeader().setDefaultSectionSize(min_height);
                cv.fifo_table.setModel(new GuiICModel(c.interconnects(), c));
                cv.fifo_table.resizeColumnsToContents();
                cv.fifo_table.horizontalHeader().setResizeMode(3, QHeaderView.ResizeMode.Stretch);

                /* initialise SHM table */
                cv.shm_table.horizontalHeader().setHighlightSections(false);
                cv.shm_table.horizontalHeader().setResizeMode(QHeaderView.ResizeMode.ResizeToContents);
                cv.shm_table.horizontalHeader().setClickable(false);
                cv.shm_table.verticalHeader().setHighlightSections(false);
                cv.shm_table.verticalHeader().setResizeMode(QHeaderView.ResizeMode.Fixed);
                cv.shm_table.verticalHeader().setClickable(false);
                min_height = cv.shm_table.verticalHeader().fontMetrics().height() + 4;
                if (min_height < 16)
                    min_height = 16;
                cv.shm_table.verticalHeader().setDefaultSectionSize(min_height);
                cv.shm_table.setModel(new GuiShmModel(c.shms(), c));
                cv.shm_table.resizeColumnsToContents();
                cv.shm_table.horizontalHeader().setResizeMode(1, QHeaderView.ResizeMode.Stretch);

                update_client_view(cv, c);
                if (c.address() == null)
                    item.setData(Qt.ItemDataRole.BackgroundRole, unconn_clr);
                else
                    item.setData(Qt.ItemDataRole.BackgroundRole, null);

                item.setData(gui_role, cvw);
                item.setData(ui_role, cv);
                item.setData(flags_role, new Integer(0));
                item.setText(c.name());
                ui.main_list.addItem(item);
            }
        }
        ui.main_list.addItem(log_item);
        ui.main_list.setCurrentItem(log_item);
        ui.main_list.setMinimumWidth(ui.main_list.sizeHintForColumn(0) + 4);
    }

    public void main_list_item(QListWidgetItem current, QListWidgetItem prev) {
        if (prev != null && !prev.equals(current))
            ((QWidget)prev.data(gui_role)).hide();
        if (current != null)
            ((QWidget)current.data(gui_role)).show();
    }

    public void connection(ClientModel c, boolean connect) {
        List items = ui.main_list.findItems(c.name(), exact_matcher);
        Iterator it = items.iterator();
        while (it.hasNext()) {
            QListWidgetItem item = (QListWidgetItem)it.next();
            if (((Integer)item.data(flags_role)).equals(0))
                update_client_view((client_view)item.data(ui_role), c);
            if (connect)
                item.setData(Qt.ItemDataRole.BackgroundRole, null);
            else
                item.setData(Qt.ItemDataRole.BackgroundRole, unconn_clr);
            if (item.equals(ui.main_list.currentItem()))
                main_list_item(item, item);
        }
    }

    public void start_sim(boolean checked) {
        dm.sim_state(SimState.RUNNING);
    }

    public void stop_sim(boolean checked) {
        dm.sim_state(SimState.STOPPED);
    }

    public void sim_speed(int i) {
        dm.timer().delay(SimulationSpeed_I.Speed.values()[i].delay());
    }

    public void append_log(String log_text) {
        ((log_view)log_item.data(ui_role)).log_text.appendPlainText(log_text);
    }
}
