package dispatcher.model;

import java.io.File;
import java.util.*;
import java.net.*;

import dispatcher.exception.ApplicationError;
import libsim.*;

/**
 * The application's data model and controller.
 *
 * @author Michael Raitza
 */
public class DispatcherModel implements Model {

    private static class ApplicationModel {
        private boolean network_error = false;
        private Collection<ClientModel> clients = null;
        private Collection<InterconnectModel> interconnects = null;
        private Collection<SharedMemoryModel> shms = null;
    }

    private static class SimulationModel {
        private SimState state = SimState.UNINITIALIZED;
        private Map<ClientModel,ClientModel> connected = new HashMap<ClientModel,ClientModel>();
        private List<Message> inbound_queue = Collections.synchronizedList(new LinkedList<Message>());
        private TimedExec timer = new TimedExec();
    }

    private ApplicationModel app = new ApplicationModel();
    private SimulationModel sim = new SimulationModel();

    public DispatcherModel() {
    }

    public void network_error(boolean err) {
        if (app.network_error != err)
            app.network_error = err;
    }

    public boolean network_error() {
        return app.network_error;
    }

    public void load(Collection<ClientModel> clients, Collection<InterconnectModel> ics, Collection<SharedMemoryModel> shms) {
        if (clients != null && ics != null && shms != null) {
            app.clients = clients;
            app.interconnects = ics;
            app.shms = shms;
            sim.state = SimState.UNCONNECTED;
        } else {
            app.clients = null;
            app.interconnects = null;
            app.shms = null;
            sim.state = SimState.UNINITIALIZED;
        }
    }

    public void client_connection(ClientModel c, boolean connect) throws ApplicationError {
        ClientModel registered = sim.connected.get(c);
        if (connect) {
            if (registered != null)
                throw new ApplicationError("Client '" + c.name() + "' already connected.");
            sim.connected.put(c,c);
            if (sim.connected.size() == app.clients.size() && sim.state.equals(SimState.UNCONNECTED))
                sim.state = SimState.STOPPED;
        } else {
            c.address(null);
            sim.connected.remove(c);
            sim.state = SimState.UNCONNECTED;
        }
    }

    public Collection<ClientModel> clients() {
        return app.clients;
    }

    public Collection<InterconnectModel> interconnects() {
        return app.interconnects;
    }

    public Collection<SharedMemoryModel> shms() {
        return app.shms;
    }

    public boolean document_loaded() {
        return app.clients != null && app.interconnects != null && app.shms != null;
    }

    public SimState sim_state() {
        return sim.state;
    }

    public void sim_state(SimState state) {
        sim.state = state;
    }

    public Map<ClientModel,ClientModel> connected() {
        return sim.connected;
    }

    public void incoming_msg(Message m) {
        sim.inbound_queue.add(m);
    }

    public List<Message> incoming_queue() {
        return sim.inbound_queue;
    }

    public void client_died(Integer id) {
        try {
            client_connection(findClient(id), false);
        } catch (ApplicationError e) {
        }
    }

    public ClientModel findClient(int id) {
        for (ClientModel c : app.clients) {
            if (c.id() == id)
                return c;
        }
        return null;
    }

    public TimedExec timer() {
        return sim.timer;
    }
}
