/**
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2011
 */
package dispatcher.model;

import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import libjconfig.data.document.DocumentController;
import libjconfig.data.document.attribute.NumericAttribute;
import libjconfig.data.document.element.GlueLogic;
import libjconfig.data.document.element.ModulePort;
import libjconfig.data.document.element.Peripheral;
import libjconfig.data.document.element.ProcessorModule;
import libjconfig.data.document.element.VerilogParameter;
import libjconfig.data.document.event.DocumentEvent;
import libjconfig.data.document.event.DocumentEvent.EventCode;
import libjconfig.data.document.node.ConnectionEndpoint;
import libjconfig.data.document.node.ParentDocumentNode;
import libjconfig.data.document.node.AddressBlock;
import libjconfig.data.document.node.DocumentNode.Id;
import libjconfig.data.spartanmc.Environment;
import libjconfig.common.ProgressIndicator;

import dispatcher.exception.ApplicationError;

/**
 * Parse a configuration file from JConfig and create a model of
 * the multicore system.
 *
 * @author Matthias Hofmann, adapted Michael Raitza
 */
public class JConfigReader {

    private boolean valid = true;
    private String errMsg = "";


    //the DocumentController should only be created once, use initDocumentControllerOnce!!
    private DocumentController documentController = null;

    private ProgressIndicator progressIndicator;

    private List<ClientModel> clients;
    private List<InterconnectModel> interconnects;

    private Map<String, InterconnectModel> incompleteIcs;
    /* only one shared memory supported so far */
    private SharedMemoryModel the_shared_memory = new SharedMemoryModel();

    /**
     * create a new JConfigReader, with no progress indicator
     * @param xmlFilePath the path to the configuration to import
     */
    public JConfigReader() {
        this(new GenericProgressIndicatorDummy());
    }


    /**
     * create a new JConfigReader
     * @param xmlFilePath the path to the configuration to import
     * @param progressIndicator the progress indicator, which is displayed during loading the jconfig
     * environment.
     */
    public JConfigReader(ProgressIndicator progressIndicator) {
        assert(progressIndicator != null);

        clients = new LinkedList<ClientModel>();
        interconnects = new LinkedList<InterconnectModel>();

        incompleteIcs = new HashMap<String, InterconnectModel>();

        this.progressIndicator = progressIndicator;

        try {
            documentController = initDocumentControllerOnce();
        } catch (ApplicationError e) {
            valid = false;
            errMsg = e.getMessage();
            return;
        }
    }

    public void parseConfiguration(final String xmlFilePath) throws ApplicationError {
        final File confFile = new File(xmlFilePath);
        final ParentDocumentNode rootNode;

        clients = new LinkedList<ClientModel>();
        interconnects = new LinkedList<InterconnectModel>();

        incompleteIcs = new HashMap<String, InterconnectModel>();

        //parse the configuration
        documentController.enqueueEvent(new DocumentEvent(EventCode.LOAD_DOCUMENT, confFile));
        documentController.processPendingEvents();
        rootNode = documentController.getRootNode();
        if (rootNode == null)
            throw new ApplicationError();

        //extract information from processor module
        final Collection<ProcessorModule> subsystemNodes = getSubsystemNodes(rootNode);

        for(ProcessorModule subsystem : subsystemNodes){
            final ClientModel currentClient = new ClientModel();
            clients.add(currentClient);
            currentClient.changeId(subsystem.getCoreId());

            //fetch subsystem name
            final String clientName = (String) subsystem.getAttribute(Id.NAME).getValue();
            currentClient.setName(clientName);

            //fetch all peripheral nodes
            Collection<Peripheral> peripheralNodes = subsystem.getChildrenRecursive(Peripheral.class);

            //find interconnects and ddr memory
            for(Peripheral pNode : peripheralNodes){
                String peripheralType = pNode.getModuleType();

                //interconnects
                final boolean isMasterIc = peripheralType.equals("master_core_connector");
                final boolean isSlaveIc = peripheralType.equals("slave_core_connector");

                if(isMasterIc || isSlaveIc){
                    //determine connection id (for connection between master and slave)
                    final Collection<ModulePort> ports = pNode.getConnectedPorts();

                    final String connectionId = getIcConnectionId(ports);
                    if(connectionId == null){
                        //no connection exists
                        valid = false;
                        errMsg = "Interconnect is missing connection to other endpoint";
                        return;
                    }

                    //search if interconnect with same connection id is known
                    InterconnectModel ic = incompleteIcs.get(connectionId);
                    if(ic == null){
                        //new interconnect
                        ic = new InterconnectModel();
                    }

                    //update connection model
                    if(isMasterIc){
                        //ic.setMaster(currentClient);

                        //set hardware parameters, the parameters from the slave part are ignored
                        setInterconnectParameters(ic ,pNode, isMasterIc, currentClient);
                    }else{
                        assert(isSlaveIc);
                        setInterconnectParameters(ic ,pNode, isMasterIc, currentClient);
                        //ic.setSlave(currentClient);
                    }

                    if(ic.isValid()){
                        incompleteIcs.remove(connectionId);
                        interconnects.add(ic);
                    }else{
                        //ic is incomplete (slave or master is missing)
                        //should be completed in another run
                        incompleteIcs.put(connectionId, ic);
                    }

                    //shared memory peripheral
                } else if (peripheralType.equals("ddr_mem_port_18")) {
                    NumericAttribute base = pNode.<AddressBlock>getChild(Id.BASE_ADR).<NumericAttribute>getChild(Id.VALUE);
                    the_shared_memory.addClient(currentClient, base.getValue().intValue());
                }
            }
        }

        //check if any incomplete connections exists
        if(incompleteIcs.size() > 0){
            valid = false;
            errMsg = "Some interconnects are not completely described";
            return;
        }
    }

    /**
     * search through the ModulePorts after the data connection of the master or slave interconnect.
     * @param ports the searched ModulePorts
     * @return the connection id, or null if no interconnect data connection was found.
     */
    private String getIcConnectionId(Collection<ModulePort> ports){
        for(ModulePort port : ports){
            final String portId = port.getId();

            if(portId.equals("#PORT.data_from_master") || portId.equals("#PORT.data_to_slave")){
                //module port owns the connection -> grep connection
                final Collection<ConnectionEndpoint> endpoints = port.getConnections();

                for(ConnectionEndpoint connection : endpoints){
                    final String connectionId = connection.getId();
                    return connectionId;
                }
            }
        }

        return null;
    }

    /**
     * extract the hardware parameters FIFO_WIDTH and FIFO_DEPTH from the PeripheralNode and
     * configure the InterconnectModel accordingly
     * @param ic
     * @param pNode
     */
    private void setInterconnectParameters(InterconnectModel ic, Peripheral pNode, boolean isMaster, ClientModel c){
        Collection <VerilogParameter> params = pNode.<VerilogParameter>getChildrenRecursive(VerilogParameter.class);

        for(VerilogParameter p : params){
            if(p.getName().equals("FIFO_WIDTH")){
                final int fifoWidth = p.<NumericAttribute>getChild(Id.VALUE).getValue().intValue();

                ic.setFifoWidth(fifoWidth);

            }else if(p.getName().equals("FIFO_DEPTH")){
                final int fifoDepth = p.<NumericAttribute>getChild(Id.VALUE).getValue().intValue();

                ic.setFifoDepth(fifoDepth);
            }
        }
        NumericAttribute base = pNode.<AddressBlock>getChild(Id.BASE_ADR).<NumericAttribute>getChild(Id.VALUE);
        if (isMaster)
            ic.setMaster(c, base.getValue().intValue());
        else
            ic.setSlave(c, base.getValue().intValue());
    }

    /**
     * search through he ModulePorts for the mem clock connection of the mcb peripheral interface
     * @param ports the searched ModulePorts
     * @return the connection id, or null if no mem clock connection was found
     */
    private void getSharedMemoryConnectionId(Peripheral pNode){
        NumericAttribute base = pNode.<AddressBlock>getChild(Id.BASE_ADR).<NumericAttribute>getChild(Id.VALUE);
    }

    /**
     * @return all Subsystem nodes, sorted by name
     */
    private Collection<ProcessorModule> getSubsystemNodes(ParentDocumentNode rootNode){
        assert(rootNode != null);
        Collection<ProcessorModule> subsystemNodes = rootNode.getChildrenRecursiveSorted(ProcessorModule.class, ProcessorModule.SORT_BY_NAME);
        return subsystemNodes;
    }

    /**
     * The initialization of the jconfiglib take some time. So we do not want to do it
     * on startup.
     * @return the DocumentController for the jconfig file
     * @throws ApplicationError throw when init the jconfig environment failed
     */
    private DocumentController initDocumentControllerOnce() throws ApplicationError {
        if(documentController == null){
            boolean success = Environment.getInstance().init(progressIndicator);
            if(!success) {
                throw new ApplicationError("Initialisation of JConfig environment failed");
            }

            documentController = new DocumentController();
        }

        return documentController;
    }

    /**
     * checks if the configuration file was read without error and
     * if it was malformed
     * @return true if all is ok
     */
    public boolean isValid(){
        return valid;
    }

    /**
     * @return the message of the last occurred error.
     */
    public String getErrorMessage(){
        return errMsg;
    }

    /**
     * @return all clients found in the xml file
     */
    public List<ClientModel> getClients(){
        return clients;
    }

    /**
     * @return all interconnects found in the xml file
     */
    public List<InterconnectModel> getInterconnects(){
        return interconnects;
    }

    /**
     * @return all shared memories found in the xml file
     */
    public List<SharedMemoryModel> getSharedMemory(){
        List<SharedMemoryModel> ll = new LinkedList<SharedMemoryModel>();
        ll.add(the_shared_memory);
        return ll;
    }

}
