package dispatcher.model;

import java.util.*;
import java.net.*;

import dispatcher.exception.*;

import libsim.*;

/**
 * The application model.
 */
public interface Model {
    /**
     * Signals the network error condition.
     * @param err the error state. TRUE iff there is a networking error
     */
    public void network_error(boolean err);

    /**
     * Returns the network error state.
     * @return the network error state
     */
    public boolean network_error();

    /**
     * Loads a new project into the model.
     * @param clients the collection of clients in the project
     * @param ics the collection of interconnects in the project
     * @param shms the collection of shared memories in the project
     */
    public void load(Collection<ClientModel> clients, Collection<InterconnectModel> ics, Collection<SharedMemoryModel> shms);

    /**
     * \brief Marks a client (dis-)connected in the model.
     *
     * Ensures that a client is not connected twice.
     * @param c the client to be (dis-)connected
     * @param connected flag. TRUE iff the client is connected, FALSE otherwise
     * @throws ApplicationError in the attempt to connect an already connected client
     */
    public void client_connection(ClientModel c, boolean connect) throws ApplicationError;

    /**
     * Returns the collection of clients in the project.
     * @return the collection of clients in the project
     */
    public Collection<ClientModel> clients();

    /**
     * Returns the collection of interconnects in the project.
     * @return the collection of interconnects in the project
     */
    public Collection<InterconnectModel> interconnects();

    /**
     * Returns the collection of shared memories in the project.
     * @return the collection of shared memories in the project
     */
    public Collection<SharedMemoryModel> shms();

    /**
     * Returns TRUE iff a projects is loaded.
     * @return TRUE iff a projects is loaded
     */
    public boolean document_loaded();

    /**
     * Returns the simulation state.
     * @return the simulation state
     */
    public SimState sim_state();

    /**
     * Sets the simulation state to STATE.
     */
    public void sim_state(SimState state);

    /**
     * Returns the collection of connected clients.
     * @return the collection of connected clients
     */
    public Map<ClientModel,ClientModel> connected();

    /**
     * Enqueues an incoming message.
     */
    public void incoming_msg(Message m);

    /**
     * Returns the queue that hols incoming messages. Thread safe.
     */
    public List<Message> incoming_queue();

    /**
     * Client with ID unexpectedly died.
     */
    public void client_died(Integer id);

    public ClientModel findClient(int id);

    public TimedExec timer();
}
