package dispatcher.model;

import java.util.*;
import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.net.*;

import libsim.*;

import libjconfig.common.Log;
import libjconfig.common.LogTitle;

import dispatcher.exception.*;

/**
 * \brief The multi-threaded server.
 *
 * The server handles all network I/O through Communication threads. The communication threads
 * are held sleeping in a ThreadPool and are activated when new data arrives or can be sent.
 * The server manages the network connection state of all clients through the model.
 *
 * @author Michael Raitza
 */
public class NewServer extends DThread {

    private int srv_port = 8886;
    private LogTitle lt = Log.get("Server");
    private Model model;
    private ThreadPool pool = new ThreadPool();
    private Selector sel = null;
    private SelectionKey srv_key = null;
    private ServerSocketChannel srv_sc = null;
    private Map<Integer,SelectionKey> clients = Collections.synchronizedMap(new HashMap<Integer,SelectionKey>());
    private Map<SelectionKey,Transaction> reading = Collections.synchronizedMap(new HashMap<SelectionKey,Transaction>());
    private Map<SelectionKey,Communication> reading_thread = Collections.synchronizedMap(new HashMap<SelectionKey,Communication>());
    private Map<SelectionKey,Transaction> writing = Collections.synchronizedMap(new HashMap<SelectionKey,Transaction>());
    private Map<SelectionKey,Communication> writing_thread = Collections.synchronizedMap(new HashMap<SelectionKey,Communication>());
    private List<Message> outbound_queue = Collections.synchronizedList(new LinkedList<Message>());

    public NewServer(Model m) {
        super("NewServer");
        model = m;
        try {
            sel = Selector.open();
        } catch (IOException e) {
            Log.debug(lt, "Cannot initialize server");
        }
    }

    private void close_srv_socket() {
        if (srv_sc == null)
            return;
        try {
            srv_sc.close();
        } catch (IOException e) {
            Log.error(lt, "Error closing server socket");
            Log.exception(lt, e);
        } finally {
            srv_sc = null;
        }
    }

    private void open_srv_socket(int port) throws IOException {
        srv_sc = ServerSocketChannel.open();
        srv_sc.configureBlocking(false);
        ServerSocket srv_s = srv_sc.socket();
        srv_s.bind(new InetSocketAddress(port));
        srv_key = srv_sc.register(sel, SelectionKey.OP_ACCEPT);
    }

    private void close_channel(SelectionKey key) {
        SocketChannel scc = (SocketChannel)key.channel();
        Log.debug(lt, "Connection closed: %s:%d", scc.socket().getInetAddress().getHostAddress(), scc.socket().getPort());
        Integer id = prune_client(key);
        if (id != null)
            model.client_died(id);
        try {
            ((SocketChannel)key.channel()).socket().close();
            key.channel().close();
        } catch (IOException e) {
            Log.debug(lt, "Error closing channel");
        }
        key.cancel();
        key.selector().wakeup();
    }

    /* maybe re-implement as a reverse map */
    private Integer prune_client(SelectionKey ck) {
        synchronized (clients) {
            Iterator<Map.Entry<Integer,SelectionKey>> it = clients.entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry<Integer,SelectionKey> e = it.next();
                if (e.getValue() != ck)
                    continue;
                it.remove();
                return e.getKey();
            }
            return null;
        }
    }

    protected void _run_init() {
        sim_state(model.sim_state());
    }

    protected void _suspend_init() {
        close_srv_socket();
    }

    protected void _run() {
        /* server (re-)connection */
        if (srv_sc == null) {
            Log.debug(lt, "Restarting");
            try {
                open_srv_socket(srv_port);
            } catch (IOException e) {
                Log.error(lt, "Unable to bind server socket to port %d", srv_port);
                Log.exception(lt, e);
                _suspend();
            }
        }
        /* normal duty */
        try {
            int num = sel.select();
            /* Prepare outbound transactions */
            synchronized (outbound_queue) {
                Iterator<Message> it = outbound_queue.iterator();
                while (it.hasNext()) {
                    Message m = it.next();
                    SelectionKey k = clients.get(m.dst());
                    if (k == null)
                        continue;
                    if (writing.containsKey(k))
                        continue;
                    k.interestOps(SelectionKey.OP_WRITE | k.interestOps());
                    writing.put(k, Transaction.send(m));
                    it.remove();
                }
            }

            /* Handle connections */
            Iterator<SelectionKey> it = sel.selectedKeys().iterator();
            while (it.hasNext()) {
                SelectionKey key = it.next();
                if (!key.isValid()) {
                    it.remove();
                    continue;
                }
                /* register new client connection */
                if (key == srv_key && key.isAcceptable()) {
                    try {
                        ServerSocketChannel sc = (ServerSocketChannel)key.channel();
                        SocketChannel scc = sc.accept();
                        scc.configureBlocking(false);
                        SelectionKey ck = scc.register(sel, SelectionKey.OP_READ);
                        Log.debug(lt, "Connection opened: %s:%d", scc.socket().getInetAddress().getHostAddress(), scc.socket().getPort());
                    } catch (ClosedChannelException e) {
                        Log.error(lt, "Channel closed prematurely");
                        _stop();
                    } catch (IOException e) {
                        Log.error(lt, "Error establishing connection. Terminating");
                        _stop();
                        return;
                    }
                    it.remove();
                    continue;
                }
                /* channel ready to be read, initiate/continue transaction */
                if (key.isReadable() && key.attachment() == null) {
                    if (reading_thread.get(key) != null) {
                        it.remove();
                        continue;
                    }
                    Transaction tn;
                    tn = reading.get(key);
                    if (tn == null) {
                        tn = Transaction.recv();
                        reading.put(key, tn);
                    }
                    Communication cc = pool.get();
                    if (cc == null) {
                        /* all threads busy, come back later */
                        it.remove();
                        continue;
                    }
                    //System.err.printf("Registering new thread %s%n", cc);
                    cc.transaction(key, tn);
                    cc.server(this);
                    reading_thread.put(key, cc);
                    if (cc.getState().equals(Thread.State.NEW))
                        cc.start();
                    cc._resume();
                    it.remove();
                    continue;
                }
                /* channel ready to be written, initiate/continue transaction */
                if (key.isWritable()) {
                    Communication cc = writing_thread.get(key);
                    if (cc != null) {
                        it.remove();
                        continue;
                    }

                    Transaction tn = writing.get(key);
                    if (tn == null) {
                        key.interestOps(key.interestOps() & ~SelectionKey.OP_WRITE);
                        it.remove();
                        continue;
                    }

                    cc = pool.get();
                    if (cc == null) {
                        /* all threads busy, come back later */
                        it.remove();
                        continue;
                    }

                    cc.transaction(key, tn);
                    cc.server(this);
                    writing_thread.put(key, cc);
                    if (cc.getState().equals(Thread.State.NEW))
                        cc.start();
                    cc._resume();
                    it.remove();
                    continue;
                }
            }
        } catch (IOException e) {
            Log.error(lt, "Error polling sockets. Terminating");
            _stop();
            return;
        } catch (CancelledKeyException k) {
            System.err.printf("Cancelled key%n");
        } catch (ClosedSelectorException e) {
        }
    }

    public void _stop() {
        try {
            sel.close();
        } catch (IOException e) {
        }
        pool.dissemble();
        super._stop();
    }

    public void sim_state(SimState state) {
        switch (state) {
        case UNINITIALIZED:
            _suspend();
            clients.clear();
            break;
        case UNCONNECTED:
            _resume();
            break;
        default:
            break;
        }
    }

    /* methods called from ServerLogic thread */
    public void send_message(Message msg) {
        outbound_queue.add(msg);
        sel.wakeup();
    }

    public Socket connection(int id) {
        return ((SocketChannel)clients.get(id).channel()).socket();
    }

    public void terminate_link(int id) {
        SelectionKey key = clients.get(id);
        if (key != null)
            close_channel(key);
    }

    /* methods called from Communication threads */
    /**
     * Client terminated its network connection.
     */
    public void client_terminated(SelectionKey ck, boolean write) {
        Transaction tn;
        ck.attach(new Boolean(true));
        if (write)
            clients.remove(writing.remove(ck));
        else {
            tn = reading.remove(ck);
            if (tn.msg != null) {
                model.incoming_msg(tn.msg);
            }
        }
        close_channel(ck);
    }

    /**
     * Communication thread finished transaction.
     */
    public void worker_finished(SelectionKey key, boolean write) {
        Transaction tn;
        if (write) {
            tn = writing.remove(key);
        } else {
            tn = reading.remove(key);
            if (!clients.containsKey(tn.msg.src()))
                clients.put(tn.msg.src(), key);
            else if (!clients.get(tn.msg.src()).equals(key)) {
                /* close the channel if a client with the same id is already connected */
                close_channel(key);
                return;
            }
            if (tn.msg != null)
                model.incoming_msg(tn.msg);
        }
    }

    /**
     * Communication thread signaled a networking error. Terminate connection.
     */
    public void worker_net_error(SelectionKey key, boolean write) {
        Log.debug(lt, "Channel closed");
        client_terminated(key, write);
    }

    /**
     * Communication thread goes to sleep. Transaction unfinished.
     */
    public void worker_deregister(SelectionKey key, boolean write) {
        if (write) {
            pool.ready(writing_thread.get(key));
            writing_thread.remove(key);
            sel.wakeup();
        } else {
            pool.ready(reading_thread.get(key));
            reading_thread.remove(key);
            sel.wakeup();
        }
    }
}
