package dispatcher.model;

import com.trolltech.qt.*;
import java.io.File;
import java.util.*;
import java.net.*;

import libjconfig.common.Log;
import libjconfig.common.LogTitle;

import dispatcher.exception.ApplicationError;

import libsim.*;

/**
 * Wrapper class around DispatcherModel that defines signals and slots
 *
 * @author Michael Raitza
 */
public class QtDispatcherModel extends QSignalEmitter implements Model {

    private LogTitle lt = Log.get("Dispatcher");
    private DispatcherModel dm;

    /* SIGNALS */
    public Signal1<Boolean> network_error = new Signal1<Boolean>();
    public Signal1<Boolean> document_loaded = new Signal1<Boolean>();
    public Signal4<Boolean,
        Collection<ClientModel>,
        Collection<InterconnectModel>,
        Collection<SharedMemoryModel>> document_data = new Signal4<Boolean,
        Collection<ClientModel>,
        Collection<InterconnectModel>,
        Collection<SharedMemoryModel>>();
    public Signal1<SimState> sim_state_changed = new Signal1<SimState>();
    public Signal0 incoming_message = new Signal0();
    public Signal1<Integer> client_dies = new Signal1<Integer>();
    public Signal2<ClientModel,Boolean> client_connection = new Signal2<ClientModel,Boolean>();

    public QtDispatcherModel() {
        dm = new DispatcherModel();
    }

    public void network_error(boolean err) {
        if (dm.network_error() != err) {
            dm.network_error(err);
            network_error.emit(err);
        }
    }

    @QtBlockedSlot
    public boolean network_error() {
        return dm.network_error();
    }

    public void load(Collection<ClientModel> clients, Collection<InterconnectModel> ics, Collection<SharedMemoryModel> shms) {
        sim_state_changed.emit(SimState.UNINITIALIZED);
        dm.load(clients, ics, shms);
        if (clients != null && ics != null && shms != null) {
            document_loaded.emit(true);
            document_data.emit(true, clients, ics, shms);
            sim_state_changed.emit(dm.sim_state());
        } else {
            document_loaded.emit(false);
            document_data.emit(false, clients, ics, shms);
            sim_state_changed.emit(dm.sim_state());
        }
    }

    public void client_connection(ClientModel c, boolean connect) throws ApplicationError {
        SimState old_state = dm.sim_state();
        dm.client_connection(c, connect);
        if (c != null)
            client_connection.emit(c, connect);
        if (!old_state.equals(dm.sim_state()))
            sim_state_changed.emit(dm.sim_state());
    }

    @QtBlockedSlot
    public Collection<ClientModel> clients() {
        return dm.clients();
    }

    @QtBlockedSlot
    public Collection<InterconnectModel> interconnects() {
        return dm.interconnects();
    }

    @QtBlockedSlot
    public Collection<SharedMemoryModel> shms() {
        return dm.shms();
    }

    @QtBlockedSlot
    public boolean document_loaded() {
        if (signalSender() != this) {
            document_loaded.emit(dm.document_loaded());
            document_data.emit(dm.document_loaded(), dm.clients(), dm.interconnects(), dm.shms());
            sim_state_changed.emit(dm.sim_state());
        }
        return dm.document_loaded();
    }

    @QtBlockedSlot
    public SimState sim_state() {
        return dm.sim_state();
    }

    public void sim_state(SimState state) {
        if (!state.equals(dm.sim_state())) {
            dm.sim_state(state);
            sim_state_changed.emit(state);
        } else
            dm.sim_state(state);
    }

    @QtBlockedSlot
    public Map<ClientModel,ClientModel> connected() {
        return dm.connected();
    }

    @QtBlockedSlot
    public void incoming_msg(Message m) {
        dm.incoming_msg(m);
        incoming_message.emit();
    }

    @QtBlockedSlot
    public void client_died(Integer id) {
        ClientModel c = findClient(id);
        try {
            if (c == null) {
                return;
            }
            client_connection(c, false);
        } catch (ApplicationError e) {
        }
    }

    @QtBlockedSlot
    public List<Message> incoming_queue() {
        return dm.incoming_queue();
    }

    @QtBlockedSlot
    public ClientModel findClient(int id) {
        return dm.findClient(id);
    }

    @QtBlockedSlot
    public TimedExec timer() {
        return dm.timer();
    }
}
