package dispatcher.model;

import java.util.*;
import java.io.*;
import java.net.*;
import java.nio.*;
import java.nio.channels.*;
import libjconfig.common.Log;
import libjconfig.common.LogTitle;
import libsim.*;

import dispatcher.exception.*;
import dispatcher.*;

public class Server extends Thread {
    private Selector sel = null;
    private SelectionKey srv_key = null;
    private ServerSocketChannel srv_sc = null;
    private ServerSocket srv_s = null;
    private volatile boolean suspended = false;
    private Thread t;
    private boolean terminate = false;
    private int srv_port = 8886;
    private LogTitle lt = Log.get("Server");
    private Model model;
    private ThreadPool pool = new ThreadPool();
    private Collection<ClientModel> pending = new LinkedList<ClientModel>();
    private Queue<Message> queue = new LinkedList<Message>();
    private Queue<Message> recv_msg = new LinkedList<Message>();
    private Map<SelectionKey,Transaction> transactions = new HashMap<SelectionKey,Transaction>();
    private Map<SelectionKey,Communication> running = new HashMap<SelectionKey,Communication>();
    private SpinLock TLock = new SpinLock();

    /* shadow map from the application model */
    private Map<Channel,Channel> connected;

    public Server(Model m) {
        model = m;
        connected = model.connected();
        try {
            sel = Selector.open();
        } catch (IOException e) {
            Log.debug(lt, "Cannot initialize server");
        }
    }

    private Channel connexion(ClientModel c) {
        for (Channel cc : connected.keySet()) {
            if (cc.c.equals(c))
                return cc;
        }
        return null;
    }

    private void close_srv_socket() {
        if (srv_s == null)
            return;
        try {
            srv_s.close();
        } catch (IOException e) {
            Log.error(lt, "Error closing server socket");
            Log.exception(lt, e);
        } finally {
            srv_s = null;
        }
    }

    private void open_srv_socket(int port) throws IOException {
        srv_sc = ServerSocketChannel.open();
        srv_sc.configureBlocking(false);
        srv_s = srv_sc.socket();
        srv_s.bind(new InetSocketAddress(port));
        srv_key = srv_sc.register(sel, SelectionKey.OP_ACCEPT);
    }

    private Communication setup_worker(SelectionKey key, Transaction tn) {
        Communication c = pool.get();
        c.server(this);
        transactions.put(key, tn);
        c.put_data(key, tn);
        if (!c.isAlive())
            c.start();
        return c;
    }

    private void terminate_all_workers() {
        for (Channel chan : connected.keySet()) {
            //Log.debug(lt, "Terminating worker thread: %s", chan.t);
            //chan.t._stop();
        }
    }

    public synchronized void _resume() {
        suspended = false;
        notify();
    }

    public void run() {
        Thread c = Thread.currentThread();
        sim_state(model.sim_state());
        while (c == t) {
            try {
                /* thread suspension */
                if (suspended) {
                    close_srv_socket();
                    Log.debug(lt, "Suspending");
                    synchronized(this) {
                        while (suspended && c == t) {
                            wait();
                        }
                    }
                }
                /* thread (re-)connection */
                if (srv_s == null) {
                    Log.debug(lt, "Restarting");
                    try {
                        open_srv_socket(srv_port);
                    } catch (IOException e) {
                        Log.error(lt, "Unable to bind server socket to port %d", srv_port);
                        Log.exception(lt, e);
                        _suspend();
                    }
                }
                /* normal duty */
                int num = sel.select();
                Iterator<SelectionKey> it = sel.selectedKeys().iterator();
                while (it.hasNext()) {
                    SelectionKey key = it.next();
                    /* new client connection */
                    if (key == srv_key && key.isAcceptable()) {
                        ServerSocketChannel sc = (ServerSocketChannel)key.channel();
                        SocketChannel scc = sc.accept();
                        scc.configureBlocking(true);
                        SelectionKey ck = scc.register(sel, SelectionKey.OP_READ);
                        Communication cc = setup_worker(ck, Transaction.recv());
                        running.put(ck, cc);
                        //model.client_connection(new Channel(null, ck), false);
                    } else if (key.isReadable()) {
                        /* run worker */
                        Communication cc = running.get(key);
                        if (cc == null) {
                            cc = setup_worker(key, Transaction.recv());
                            running.put(key, cc);
                        } else
                            cc._resume();
                    }
                    it.remove();
                }
            } catch (SocketTimeoutException e) {
                continue;
            } catch (InterruptedException e) {
                continue;
                /*
            } catch (ApplicationError e) {
                Log.exception(lt, e);
                */
            } catch (IOException e) {
                model.network_error(true);
            }
        }
        terminate_all_workers();
        close_srv_socket();
    }

    public void start() {
        t = new Thread(this);
        t.start();
    }

    public synchronized void _stop() {
        t = null;
        notify();
    }

    public synchronized void _suspend() {
        suspended = true;
    }

    public Collection<ClientModel> clients() {
        return model.clients();
    }

    /**
     * Client has sent INIT condition.
     */
    public void client_register(SelectionKey ck, ClientModel c) throws ApplicationError {
        if (connexion(c) == null) {
            model.client_connection(new Channel(c, ck), true);
            Log.debug(lt, "Registering new client: %s%n", c);
        }
    }

    /**
     * Client has sent NUPD condition. To be called by a communication thread.
     */
    public void client_clear(ClientModel c) {
        pending.remove(c);
    }

    public void client_error(ClientModel c) {
    }

    public void client_terminated(ClientModel c) throws ApplicationError {
        Channel chan = connexion(c);
        Log.debug(lt, "Unregistering client: %s", c);
        model.client_connection(chan, false);
    }

    public void stop_simulation() {
    }

    public void continue_simulation() {
    }

    public void worker_error(SelectionKey ck) {
        while (true) {
            try {
                TLock.lock();
                Communication cc = running.remove(ck);
                pool.remove(cc);
                transactions.remove(ck);
                break;
            } catch (InterruptedException e) {
            } finally {
                TLock.unlock();
            }
        }
        model.network_error();
    }

    public void worker_sleep(SelectionKey ck) {
        while (true) {
            try {
                TLock.lock();
                Communication cc = running.remove(ck);
                pool.ready(cc);
                break;
            } catch (InterruptedException e) {
            } finally {
                TLock.unlock();
            }
        }
    }

    public void worker_finished(SelectionKey ck) {
        while (true) {
            try {
                TLock.lock();
                Communication cc = running.remove(ck);
                pool.ready(cc);
                Transaction tn = transactions.remove(ck);
                if (!tn.send)
                    recv_msg.add(tn.msg);
                break;
            } catch (InterruptedException e) {
            } finally {
                TLock.unlock();
            }
        }
    }

    public void sim_state(SimState state) {
        switch (state) {
        case UNINITIALIZED:
            terminate_all_workers();
            _suspend();
            break;
        case UNCONNECTED:
            _resume();
            break;
        default:
            break;
        }
    }
}
