package dispatcher.model;

import java.util.*;
import java.net.*;

import libjconfig.common.Log;
import libjconfig.common.LogTitle;
import libsim.*;

import dispatcher.exception.*;

public class ServerLogic extends DThread {
    private NewServer srv;
    private Model model;
    private Set<ClientModel> pending = new HashSet<ClientModel>();
    private List<Message> inbound_queue;
    private LogTitle lt = Log.get("ServerLogic");
    private SimState sim_state;
    private int active_sequence = 0;
    private boolean error = false;
    private static int SERVER_ID = -1;
    private boolean send_step;
    private boolean run_once;
    private static final Collection<Tuple> empty_tuples = new LinkedList<Tuple>();

    public ServerLogic(Model model, NewServer server) {
        super("ServerLogic");
        this.srv = server;
        inbound_queue = model.incoming_queue();
        this.model = model;
    }

    private void handle_condition(MessageCond m) {
        ClientModel sender = model.findClient(m.src());
        Payload pp = new Payload(active_sequence, 0, SERVER_ID);
        switch(m.cond()) {
        case INIT:
            /* should never happen... */
            try {
                Socket s = srv.connection(m.src());
                sender.address(s.getInetAddress());
                sender.port(s.getPort());
                model.client_connection(sender, true);
            } catch(ApplicationError e) {
                srv.terminate_link(m.src());
                Log.error(lt, "Re-connection");
            }
            break;
        case NUPD:
            pending.remove(sender);
            break;
        case EUPD:
            error = true;
            break;
        case TERM:
            try {
                sender.address(null);
                sender.port(0);
                model.client_connection(sender, false);
            } catch (ApplicationError e) {
                Log.error(lt, "Error terminating client.");
            }
            break;
        case STOP:
            model.sim_state(SimState.STOPPED);
            Log.debug(lt, "Simulation state changed to: STOPPED");
            break;
        case CONT:
            model.sim_state(SimState.RUNNING);
            Log.debug(lt, "Simulation state changed to: RUNNING");
            break;
        case RESET:
            reset_sim(false);
            Log.debug(lt, "Simulation RESET");
            break;
        case STEP:
            step_sim(false);
            break;
        case CLEAR:
            break;
        default:
            throw new ImplementationError("Unexpected condition code: " + m.cond());
        }
    }

    /* forward FIFO requests and responses */
    private void handle_fifo(Message m) {
        ClientModel sender = model.findClient(m.src());
        ClientModel receiver;
        try {
            receiver = sender.get_partner(m.dst());
        } catch (ApplicationError e) {
            Log.error(lt, "Destination unknown. Message dropped: %s", m.format_msg());
            model.sim_state(SimState.STOPPED);
            return;
        }
        Message n = m.clone();
        n.dst(receiver.id());
        try {
            n.src(sender.ic(m.dst()).base(receiver));
        } catch (ApplicationError e) {
            Log.error(lt, "Destination unknown. Message dropped: %s", m.format_msg());
            model.sim_state(SimState.STOPPED);
            return;
        }
        srv.send_message(n);
    }

    /* Memory updates */
    private void handle_mem(MessageUMem m) {
        ClientModel sender = model.findClient(m.src());
        SharedMemoryModel mem;

        try {
            mem = sender.get_shm(m.dst());
        } catch (ApplicationError e) {
            Log.error(lt, "%s. Message dropped: %s", e.getMessage(), m.format_msg());
            model.sim_state(SimState.STOPPED);
            return;
        }

        /* memory clean, but update marked as dirty; propagate dirty state */
        if (!mem.dirty() && m.dirty()) {
            for (ClientModel recv : mem.clients()) {
                if (sender.equals(recv))
                    continue;
                Message n = m.clone();
                n.dst(recv.id());
                n.src(mem.base(recv));
                srv.send_message(n);
            }
        }

        /* store updates */
        mem.dirty(sender, m.dirty());
        mem.updates(sender, m.data());

        Payload p = new Payload(active_sequence, 0, 0);
        /* reply pending requests */
        if (!mem.pending()) {
            for (ClientModel target : mem.targets()) {
                p.reset();
                p.dst(target.id());
                p.src(mem.base(target));
                p.add(mem.dirty());
                p.add(mem.updates(target));
                srv.send_message(MessageType.U_MEM.construct(p));
            }
            mem.clear_targets();
        }

        /* send high volume updates */
        for (ClientModel recv : mem.high_vol()) {
            p.reset();
            p.dst(recv.id());
            p.src(mem.base(recv));
            p.add(mem.dirty());
            p.add(mem.updates(recv));
            srv.send_message(MessageType.U_MEM.construct(p));
        }
    }

    /* Memory requests */
    private void handle_rmem(Message m) {
        ClientModel sender = model.findClient(m.src());
        SharedMemoryModel mem;

        try {
            mem = sender.get_shm(m.dst());
        } catch (ApplicationError e) {
            Log.error(lt, "%s. Message dropped: %s", e.getMessage(), m.format_msg());
            model.sim_state(SimState.STOPPED);
            return;
        }
        /* forward the request, if necessary or reply with empty update */
        Payload pp = new Payload(m.seq(), 0, 0);
        if (mem.dirty()) {
            for (ClientModel supply : mem.dirty_clients()) {
                if (sender.equals(supply))
                    continue;
                pp.dst(supply.id());
                pp.src(mem.base(supply));
                srv.send_message(MessageType.R_MEM.construct(pp));
                mem.set_pending(supply);
            }
            mem.set_target(sender);
        } else {
            pp.reset();
            pp.dst(sender.id());
            pp.src(mem.base(sender));
            pp.add(mem.dirty());
            pp.add(empty_tuples);
            srv.send_message(MessageType.U_MEM.construct(pp));
        }
    }

    /* handle messages */
    private void handle_msg(Message m) {
        switch(m.type()) {
        case COND:
            handle_condition((MessageCond)m);
            break;
        case R_FIFO:
        case U_FIFO:
            handle_fifo(m);
            break;
        case R_MEM:
            handle_rmem(m);
        case U_MEM:
            handle_mem((MessageUMem)m);
            break;
        default:
            Log.error(lt, "Unexpected message type: %s", m.type());
        }
    }

    /* some clients are unconnected, don't handle any messages besides conditions */
    private void handle_init(Message m) {
        ClientModel sender = model.findClient(m.src());
        switch (m.type()) {
        case COND:
            switch (((MessageCond)m).cond()) {
            case INIT:
                try {
                    Socket s = srv.connection(m.src());
                    sender.address(s.getInetAddress());
                    sender.port(s.getPort());
                    model.client_connection(sender, true);
                } catch(ApplicationError e) {
                    srv.terminate_link(m.src());
                    Log.error(lt, "Re-connection");
                }
                break;
            default:
                Log.debug(lt, "Message ignored");
                break;
            }
            break;
        case R_CFG:
            /* do something veery clever here */
            Payload p = new Payload(m.seq(), m.src(), SERVER_ID);
            for (Map.Entry<Integer,InterconnectModel> ic : sender.ic_links().entrySet()) {
                InterconnectModel icc = ic.getValue();
                p.add(new FIFOConfig(sender.equals(icc.master()), (byte)icc.width(), icc.depth(), ic.getKey()));
            }
            for (Integer shm_base : sender.shm_links().keySet())
                p.add(shm_base);
            srv.send_message(MessageType.U_CFG.construct(p));
            break;
        default:
            Log.debug(lt, "Message ignored");
        }
    }
    private void handle_reset(Message m) {
        ClientModel sender = model.findClient(m.src());
        switch (m.type()) {
        case COND:
            switch (((MessageCond)m).cond()) {
            case CLEAR:
                pending.remove(sender);
                if (pending.size() == 0)
                    model.sim_state(SimState.STOPPED);
                break;
            default:
                Log.debug(lt, "Message ignored: %s", m.format_msg());
                break;
            }
            break;
        default:
            Log.debug(lt, "Message ignored: %s", m.format_msg());
        }
    }

    private void handle_incoming(Message m) {
        //Log.debug(lt, "Receiving message: %s", m.format_msg());
        if (model.findClient(m.src()) == null) {
            Log.debug(lt, "Unknown source (0x%x). Rejecting packet.\n%s", m.src(), m.format_msg());
            srv.terminate_link(m.src());
            return;
        }
        switch(sim_state) {
        case RUNNING:
        case STOPPED:
            handle_msg(m);
            break;
        case UNCONNECTED:
            handle_init(m);
            break;
        case RESET:
            handle_reset(m);
            break;
        case UNINITIALIZED:
        default:
            break;
        }
    }

    private void send_to_all(MessageType t, Payload p) {
        for (ClientModel c : model.clients()) {
            p.dst(c.id());
            srv.send_message(t.construct(p));
        }
    }

    protected void _run() {
        if (sim_state.equals(SimState.UNINITIALIZED))
            return;

        while(inbound_queue.size() > 0) {
            Message m = inbound_queue.remove(0);
            handle_incoming(m);
        }

        /* No pending clients, send STEP */
        //System.err.printf("%s %d %s%n", sim_state, pending.size(), send_step);
        if (sim_state.equals(SimState.RUNNING) && pending.size() == 0 && send_step) {
            send_step = false;
            if (!error)
                active_sequence++;
            else
                error = false;
            Payload p = new Payload(active_sequence, 0, SERVER_ID);
            for (ClientModel c : model.clients()) {
                p.dst(c.id());
                srv.send_message(MessageType.STEP.construct(p));
                pending.add(c);
            }
        }

        if (run_once && sim_state.equals(SimState.RUNNING)) {
            run_once = false;
            model.sim_state(SimState.STOPPED);
        }

        _suspend();
    }

    public void sim_state(SimState state) {
        sim_state = state;
        Payload pp;
        switch (state) {
        case UNINITIALIZED:
            pending.clear();
            _suspend();
            break;
        case UNCONNECTED:
            pending.clear();
            _resume();
            break;
        case RESET:
            for (ClientModel c : model.clients())
                pending.add(c);
            pp = new Payload(active_sequence, 0, SERVER_ID);
            pp.add(libsim.State.RESET);
            send_to_all(MessageType.STATE, pp);
            break;
        case RUNNING:
            pp = new Payload(active_sequence, 0, SERVER_ID);
            pp.add(libsim.State.RUNNING);
            send_to_all(MessageType.STATE, pp);
            break;
        case STOPPED:
            pp = new Payload(active_sequence, 0, SERVER_ID);
            pp.add(libsim.State.STOPPED);
            send_to_all(MessageType.STATE, pp);
            break;
        default:
            break;
        }
    }

    public void server(NewServer server) {
        srv = server;
    }

    /* called from TimedExec */
    public void send_step() {
        send_step = true;
        _resume();
    }

    /* called from GUI */
    public void reset_sim(boolean checked) {
        model.sim_state(SimState.RESET);
        Log.debug(lt, "Simulation RESET");
    }

    public void step_sim(boolean checked) {
        switch(sim_state) {
        case RUNNING:
            run_once = true;
            send_step = true;
            break;
        case STOPPED:
            run_once = true;
            send_step = true;
            model.sim_state(SimState.RUNNING);
            break;
        default: break;
        }
    }

    /**
     * Client with CLIENT_ID has no further pending updates.
     */
    public void client_clear(int client_id) {
    }

    /**
     * Client with CLIENT_ID set the error condition.
     */
    public void client_error(int client_id) {
    }

    /**
     * Client terminated its execution.
     */
    public void client_terminated(int client_id) throws ApplicationError {
    }

    /**
     * Client requests stop of simulation.
     */
    public void client_stopped(int client_id) {
    }

    /**
     * Client requests continuation of simulation.
     */
    public void client_continued(int client_id) {
    }

    /**
     * Registers a new client with id CLIENT_ID.
     */
    public void client_register(int client_id) throws ApplicationError {
    }
}
