package dispatcher.model;

import java.util.*;
import libsim.Tuple;

public class SharedMemoryModel {
    private Map<ClientModel,Integer> clients = new HashMap<ClientModel,Integer>();
    private Set<ClientModel> dirty = new HashSet<ClientModel>();
    private Map<ClientModel,Set<Tuple>> updates = new HashMap<ClientModel,Set<Tuple>>();
    private Set<ClientModel> high_vol = new HashSet<ClientModel>();
    private Set<ClientModel> memr_pending = new HashSet<ClientModel>();
    private Set<ClientModel> memr_targets = new HashSet<ClientModel>();
    private static final int HIGH_VOL = 64;

    /**
     * \brief Adds the client C to this model.
     *
     * Does the registration in the client via ClientModel.addShm().
     */
    public void addClient(ClientModel c, int base) {
        clients.put(c, base);
        c.addShm(this, base);
        updates.put(c, new HashSet<Tuple>());
    }

    /**
     * Returns the base address with which the client C is connected to this model.
     */
    public Integer base(ClientModel c) {
        return clients.get(c);
    }

    /**
     * \brief (Un-)Marks this model dirty for client C.
     *
     * From the model's perspective this means that all the clients that have marked
     * this model dirty have unpropagated memory updates.
     */
    public void dirty(ClientModel c, boolean d) {
        if (d)
            dirty.add(c);
        else
            dirty.remove(c);
    }

    /**
     * Stores the updates of client C for deliveral to all other connected clients.
     */
    public void updates(ClientModel c, Collection<Tuple> uds) {
        for (ClientModel uc : updates.keySet()) {
            if (!c.equals(uc)) {
                updates.get(uc).addAll(uds);
                if (updates.get(uc).size() > HIGH_VOL)
                    high_vol.add(uc);
            }
        }
    }

    /**
     * Retrieves all updates for client C.
     */
    public Collection<Tuple> updates(ClientModel c) {
        Collection<Tuple> uds = updates.get(c);
        updates.put(c, new HashSet<Tuple>());
        high_vol.remove(c);
        return uds;
    }

    /**
     * Returns TRUE if at least one client has unpropagated memory updates.
     */
    public boolean dirty() {
        return dirty.size() > 0;
    }

    /**
     * Returns the collection of clients connected to this model.
     */
    public Collection<ClientModel> clients() {
        return clients.keySet();
    }

    /**
     * Returns the collection of clients that might have unpropagated memory updates.
     */
    public Collection<ClientModel> dirty_clients() {
        return dirty;
    }

    /**
     * Returns the collection of clients that have reached a certain volume of
     * uncommited changes but have not asked for it yet.
     */
    public Collection<ClientModel> high_vol() {
        return high_vol;
    }

    /**
     * Mark client c as pending.
     */
    public void set_pending(ClientModel c) {
        memr_pending.add(c);
    }

    /**
     * Clear 'pending' mark.
     */
    public void clear_pending(ClientModel c) {
        memr_pending.remove(c);
    }

    /**
     * Returns TRUE iff there are memory requests pending for this model.
     */
    public boolean pending() {
        return memr_pending.size() == 0;
    }

    /**
     * Mark client C as target for the next memory update.
     */
    public void set_target(ClientModel c) {
        memr_targets.add(c);
    }

    /**
     * Return the targets for the next memory update.
     */
    public Collection<ClientModel> targets() {
        return memr_targets;
    }

    /**
     * Clear the collection of targets for the next memory update.
     */
    public void clear_targets() {
        memr_targets.clear();
    }
}
