/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 *
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 *
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

/**
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 */
package de.tu_darmstadt.rs.sim.engine;

import de.tu_darmstadt.rs.sim.signal.SignalModified;

import java.util.BitSet;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

/**
 * Each component which has an internal state to be displayed (esp. by GUI) must
 * be derived by the AbstractHardwareComponent. If the component's state
 * changes, a refresh event must be fired. RefreshListeners may register for
 * events, and will be called whenever such one occurs.
 */
public abstract class AbstractHardwareComponent implements
		SpartanDefinitions_I, RefreshTrigger_I {

	/**
	 * listeners for data changes
	 */
	protected final List<RefreshListener_I> listenerList = new LinkedList<RefreshListener_I>();
	private BitSet fDirtyAddresses = new BitSet();
	private int fCompleteRefresh = Integer.MAX_VALUE;
	private boolean fHasRefreshListeners = false;
	/**
	 * de.tu_darmstadt.rs.sim.signal object for posting modified data events
	 */
	protected final SignalModified sigModified;

	/**
	 * abstract constructor for subclasses
	 *
	 * @param refreshHandler
	 *            a {@link RefreshHandler_I} to which refresh requests should be
	 *            sent
	 */
	public AbstractHardwareComponent(RefreshHandler_I refreshHandler) {
		sigModified = new SignalModified();
		refreshHandler.registerTrigger(this);
		refreshHandler.registerDataProducerChangeSignal(sigModified);
	}

	/**
	 * Adds a listener to the event queue.
	 *
	 * @param listener
	 *            the listener
	 */
	public void addRefreshListener(final RefreshListener_I listener) {
		listenerList.add(listener);
		fHasRefreshListeners = true;
	}

	/**
	 * Removes a listener from the event queue.
	 *
	 * @param listener
	 *            the listener
	 */
	public void removeRefreshListener(final RefreshListener_I listener) {
		listenerList.remove(listener);
		fHasRefreshListeners = (!listenerList.isEmpty());
	}

	/**
	 * Will be invoked after the component's state has changed.
	 *
	 * @param address
	 *            the values's address which has changed, or COMPLETE_REFRESH if
	 *            no differentiation is made
	 */
	protected final void fireRefreshEvent(final int address) {
		if (!fHasRefreshListeners) {
			return;
		}
		if (address >= fCompleteRefresh) {
			return;
		}
		if (address < 0) {
			fCompleteRefresh = address;
			fDirtyAddresses.set(0);
			return;
		}
		fDirtyAddresses.set(address);
	}

	@Override
	public Collection<RefreshListener_I> getRefreshListeners() {
		return listenerList;
	}

	@Override
	public int[] getDirtyAddresses() {
		int cnt = fDirtyAddresses.cardinality();
		int[] addr = new int[cnt];
		int a = fDirtyAddresses.nextSetBit(0);
		while ((a >= 0) && (cnt > 0)) {
			cnt--;
			addr[cnt] = a;
			fDirtyAddresses.clear(a);
			a = fDirtyAddresses.nextSetBit(a + 1);
		}
		while (cnt-- > 0) {
			addr[cnt] = 0;
		}
		if (fCompleteRefresh != Integer.MAX_VALUE) {
			// TODO: implement a way to avoid the race condition?
			// if(addr.length==0){
			// addr=new int[1];
			// }
			addr[0] = fCompleteRefresh;
			fCompleteRefresh = Integer.MAX_VALUE;
		}
		return addr;
	}
}
