/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine;

import de.tu_darmstadt.rs.sim.preferences.ProjectComponent_I;

import java.util.BitSet;
import java.util.HashMap;
import java.util.Map;
import java.util.prefs.Preferences;

/**
 * Class for checking accesses to any memory units for watch or breakpoint handling and stopping the simulation accordingly.
 * Singleton.
 */
public class AccessManager implements ProjectComponent_I {

	/**
	 * used as index in the start address array for the de.tu_darmstadt.rs.sim.main address space
	 */
	public static final int MAIN_ADRESS_SPACE = 0;
	/**
	 * used as index in the start address array for the register address space
	 */
	public static final int REGISTER_ADDRESS_SPACE = 1;
	/**
	 * used as index in the start address array for the special function register address space
	 */
	public static final int SFR_REGISTER_ADDRESS_SPACE = 2;
	private SimulationController_I simulationController;
	private final BitSet tagsOnAddresses;
	private final int[] startAddresses;
	private final Map<Integer, AccessType> addressAccessTypeMap;
	private SimulationManager simulationManager;


	/**
	 * access types to check
	 */
	enum AccessType {
		BREAKPOINT, WATCHPOINT
	}


	private static class InstanceHolder {

		private static final AccessManager theInstance = new AccessManager();
	}

	/**
	 * create a new {@link AccessManager}.
	 * visible only for subclasses (which allows MockObjects for testing purposes)
	 */
	protected AccessManager() {
		tagsOnAddresses = new BitSet(0x40408);
		addressAccessTypeMap = new HashMap<Integer, AccessType>();
		startAddresses = new int[3];
		startAddresses[MAIN_ADRESS_SPACE] = 0;
		startAddresses[REGISTER_ADDRESS_SPACE] = 0x40000;
		startAddresses[SFR_REGISTER_ADDRESS_SPACE] = 0x40400;
	}

	/**
	 * set the {@link SimulationController_I}. This is necessary for stopping the simulation.
	 *
	 * @param simulationController the reference to the {@link SimulationController_I}
	 */
	public void setSimulationController(final SimulationController_I simulationController) {
		this.simulationController = simulationController;
	}

	/**
	 * set the {@link SimulationManager}. This is necessary to push error messages.
	 *
	 * @param simulationManager the reference to the {@link SimulationManager}
	 */
	public void setSimulationManager(final SimulationManager simulationManager) {
		this.simulationManager = simulationManager;
	}

	/**
	 * retrieve the single instance
	 *
	 * @return the singleton instance
	 */
	public static AccessManager getTheInstance() {
		return InstanceHolder.theInstance;
	}

	private void registerEventOnAddress(int address, AccessType accessType) {
		tagsOnAddresses.set(address);
		AccessType typeFromMap = addressAccessTypeMap.get(address);
		if (typeFromMap == null) {
			addressAccessTypeMap.put(address, accessType);
		} else {
			if (typeFromMap != accessType) {
				if (accessType == AccessType.BREAKPOINT) {
					addressAccessTypeMap.put(address, accessType);
				}
			}
		}
	}


	private void unregisterEventOnAddress(final int address) {
		tagsOnAddresses.clear(address);
		addressAccessTypeMap.remove(address);
	}

	/**
	 * check if there is a watch or breakpoint
	 *
	 * @param address address to check for
	 */
	public void checkForEvent(final int address) {
		if (tagsOnAddresses.get(address)) {
			simulationController.stopSimulation();

		}
	}

	/**
	 * push an error message
	 *
	 * @param message the message to be shown
	 */
	public void pushErrorMessage(String message) {
		simulationManager.pushErrorMessage(message);
	}

	@Override
	public void loadConfiguration(Preferences prefs) {
		int theAddress;
		int i = 0;
		for (int j = tagsOnAddresses.nextSetBit(0); j >= 0; j = tagsOnAddresses.nextSetBit(j + 1)) {
			theAddress = prefs.getInt("" + i, -1);
			if (theAddress != -1) {
				toggleBreakpoint(theAddress);
			}
			i++;
		}
	}

	@Override
	public void storeConfiguration(Preferences prefs) {
		int i = 0;
		for (int j = tagsOnAddresses.nextSetBit(0); j >= 0; j = tagsOnAddresses.nextSetBit(j + 1)) {
			prefs.putInt("" + i, j);
			i++;
		}
	}

	private AccessType getAccessTypeOfAddress(final int address) {
		return addressAccessTypeMap.get(address);
	}

	/**
	 * Checks if a watchpoint is connected to a global address.
	 *
	 * @param address the address to be checked
	 * @return true, if there is a watchpoint on that address
	 */
	public boolean hasWatchpoint(final int address) {
		return getAccessTypeOfAddress(address) == AccessType.WATCHPOINT;
	}

	/**
	 * Checks if a breakpoint is connected to a global address.
	 *
	 * @param address the address to be checked
	 * @return true, if there is a breakpoint on that address
	 */
	public boolean hasBreakpoint(final int address) {
		return getAccessTypeOfAddress(address) == AccessType.BREAKPOINT;
	}

	/**
	 * Toggles the breakpoint at a global address.
	 *
	 * @param address the address
	 */
	public void toggleBreakpoint(int address) {
		if (hasBreakpoint(address)) {
			unregisterEventOnAddress(address);
		} else {
			registerEventOnAddress(address, AccessType.BREAKPOINT);
		}
	}

	/**
	 * Toggles the watchpoint at a global address.
	 *
	 * @param address the address
	 */
	public void toggleWatchpoint(final int address) {
		if (hasWatchpoint(address)) {
			unregisterEventOnAddress(address);
		} else {
			registerEventOnAddress(address, AccessType.WATCHPOINT);
		}
	}

}
