/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine;

import de.tu_darmstadt.rs.sim.engine.memory.InstructionPort;
import de.tu_darmstadt.rs.sim.engine.memory.MainMemory;
import de.tu_darmstadt.rs.sim.engine.peripheral.JConfigPeripheral;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.PipelineStage;
import de.tu_darmstadt.rs.sim.utilities.MemoryAccess_I;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.TreeMap;

/**
 * This class represents the entire SpartanMC address space. It provides methods
 * for registering peripherals to addresses, watch and breakpoint checking and
 * toggling, getting and setting memory data.
 */
public class AddressSpaceManager implements SimulationComponent_I,
		RefreshTrigger_I, MemoryAccess_I {

	/**
	 * value used if there is no valid address
	 */
	public final static int NO_IO_ADDRESS = -1;
	private final static int MIN_IO_ADDRESS = (Integer) Settings
			.getSetting(Settings.IO_ADDRESS_MIN);

	private final MainMemory memory;
	private final InstructionPort instructionPort;
	private final AccessManager accessManager;
	private final ArrayList<JConfigPeripheral> peripherals;
	private final ArrayList<RefreshListener_I> refreshListenerList;

	private final TreeMap<Integer, List<AddressWithComponent>> mappedPeripherals;
	private Map<Integer, Integer> memoryBreakpoints = new HashMap<Integer, Integer>();


	private boolean hasModifiedData;


	private class AddressWithComponent {

		public AddressWithComponent(int size, JConfigPeripheral peripheral) {
			this.size = size;
			this.peripheral = peripheral;
		}

		public int size;
		public JConfigPeripheral peripheral;
	}

	/**
	 * create a new {@link AddressSpaceManager}
	 *
	 * @param refreshHandler a {@link RefreshHandler_I} to which refresh requests should be
	 * sent
	 * @param accessManager a reference to the {@link AccessManager} (used to check for
	 * watch or breakpoints)
	 */
	public AddressSpaceManager(final RefreshHandler_I refreshHandler,
							   final AccessManager accessManager) {
		this.accessManager = accessManager;
		this.memory = new MainMemory(refreshHandler);
		peripherals = new ArrayList<JConfigPeripheral>();
		mappedPeripherals = new TreeMap<Integer, List<AddressWithComponent>>();
		instructionPort = new InstructionPort(this);
		refreshListenerList = new ArrayList<RefreshListener_I>();
		refreshHandler.registerTrigger(this);
	}

	/**
	 * get the {@link InstructionPort} which provides reading of instructions
	 *
	 * @return the instance of the {@link InstructionPort}
	 */
	public InstructionPort getInstructionPort() {
		return instructionPort;
	}

	/**
	 * perform a write access to the address
	 *
	 * @param defaultAddress the value's address
	 * @param data the data which should be written at defaultAddress
	 */
	public void writeTo(final int defaultAddress, final int data) {
		if (defaultAddress < MIN_IO_ADDRESS) {
			memory.setValueAt(defaultAddress, data);
		} else {
			Entry<Integer, List<AddressWithComponent>> entry = mappedPeripherals
					.floorEntry(defaultAddress);
			if (entry != null) {
				List<AddressWithComponent> peripheralList = entry.getValue();
				if (peripheralList != null && peripheralList.size() != 0) {
					for (AddressWithComponent peripheral : peripheralList) {
						if (entry.getKey() + peripheral.size > defaultAddress) {
							peripheral.peripheral.writeToWithSideEffects(
									defaultAddress, data);
						}
					}
				} else {
					if ((MIN_IO_ADDRESS <= defaultAddress)
							&& (defaultAddress <= Settings.IO_ADDRESS_MAX)) {
						accessManager.pushErrorMessage(String.format(
								"No periphery unit on IO address 0x%05x",
								defaultAddress));
					}
				}
			}
		}
		hasModifiedData = true;
	}

	/**
	 * get a value from a given address
	 *
	 * @param defaultAddress the value's address
	 * @return the value at that address
	 */
	@Override
	public int getValue(final int defaultAddress) {
		if (defaultAddress < MIN_IO_ADDRESS) {
			return memory.getValueAt(defaultAddress);
		} else {
			Entry<Integer, List<AddressWithComponent>> entry = mappedPeripherals
					.floorEntry(defaultAddress);
			if (entry != null) {
				List<AddressWithComponent> peripheralList = entry.getValue();
				if (peripheralList != null && peripheralList.size() != 0) {
					int value = 0;
					for (AddressWithComponent peripheral : peripheralList) {
						if (entry.getKey() + peripheral.size > defaultAddress) {
							value |= peripheral.peripheral
									.getValueAt(defaultAddress);
						}
					}
					return value;
				}
			}
			return SpartanDefinitions_I.INVALID;
		}
	}

	/**
	 * set a value from a given address
	 *
	 * @param defaultAddress the value's address
	 * @param data the data which should be written at defaultAddress
	 */
	@Override
	public void setValue(final int defaultAddress, final int data) {
		if (defaultAddress < MIN_IO_ADDRESS) {
			memory.setValueAt(defaultAddress, data);
		} else {
			Entry<Integer, List<AddressWithComponent>> entry = mappedPeripherals
					.floorEntry(defaultAddress);
			if (entry != null) {
				List<AddressWithComponent> peripheralList = entry.getValue();
				if (peripheralList != null && peripheralList.size() != 0) {
					for (AddressWithComponent peripheral : peripheralList) {
						if (entry.getKey() + peripheral.size > defaultAddress) {
							peripheral.peripheral.setValueAt(defaultAddress,
									data);
						}
					}
				}
			}
		}
		hasModifiedData = true;
	}

	/**
	 * perform a read access to the address
	 *
	 * @param defaultAddress the value's address
	 * @return the value at that address
	 */
	public int readFrom(final int defaultAddress) {
		if (defaultAddress < MIN_IO_ADDRESS) {
			return memory.getValueAt(defaultAddress);
		} else {
			Entry<Integer, List<AddressWithComponent>> entry = mappedPeripherals
					.floorEntry(defaultAddress);
			if (entry != null) {
				List<AddressWithComponent> peripheralList = entry.getValue();
				if (peripheralList != null && peripheralList.size() != 0) {
					int value = 0;
					for (AddressWithComponent peripheral : peripheralList) {
						if (entry.getKey() + peripheral.size > defaultAddress) {
							value |= peripheral.peripheral
									.readFromWithSideEffects(defaultAddress);
						}
					}
					return value;
				} else {
					if ((MIN_IO_ADDRESS <= defaultAddress)
							&& (defaultAddress <= Settings.IO_ADDRESS_MAX)) {
						accessManager.pushErrorMessage(String.format(
								"No periphery unit on IO address 0x%05x",
								defaultAddress));
					}
				}
			}
			return SpartanDefinitions_I.INVALID;
		}
	}

	/**
	 * check if there is a watchpoint on defaultAddress
	 *
	 * @param defaultAddress the address to check
	 * @return true if there is a watchpoint on the address, false otherwise
	 */
	@Override
	public boolean hasWatchpoint(final int defaultAddress) {
		return accessManager.hasWatchpoint(defaultAddress);
	}

	/**
	 * check if there is a breakpoint on defaultAddress
	 *
	 * @param defaultAddress the address to check
	 * @return true if there is a breakpoint on the address, false otherwise
	 */
	public boolean hasBreakpoint(final int defaultAddress) {
		return accessManager.hasBreakpoint(defaultAddress);
	}

	/**
	 * set a watchpoint on defaultAddress if there was not any, otherwise remove
	 * the watchpoint. Note: If there is a breakpoint on the address, there
	 * could not be a watchpoint.
	 *
	 * @param defaultAddress the address whose watchpoint should be toggled
	 */
	@Override
	public void toggleWatchpoint(final int defaultAddress) {
		accessManager.toggleWatchpoint(defaultAddress);
	}

	/**
	 * set a breakpoint on defaultAddress if there was not any, otherwise remove
	 * the breakpoint. Note: If there is a breakpoint on the address, there
	 * could not be a watchpoint.
	 *
	 * @param defaultAddress the address whose breakpoint should be toggled
	 */
	public void toggleBreakpoint(final int defaultAddress) {
		accessManager.toggleBreakpoint(defaultAddress);
	}

	/**
	 * Adds a listener to the event queue.
	 *
	 * @param listener the listener
	 */
	public void addRefreshListener(final RefreshListener_I listener) {
		if (!refreshListenerList.contains(listener)) {
			refreshListenerList.add(listener);
		}
	}

	/**
	 * Removes a listener from the event queue.
	 *
	 * @param listener the listener
	 */
	public void removeRefreshListener(final RefreshListener_I listener) {
		refreshListenerList.remove(listener);
	}

	/**
	 * Retrieves the pipeline stage which is executing the address' instruction
	 * if any.
	 *
	 * @param defaultAddress the address to check for
	 * @return the pipeline stage, if any
	 */
	public PipelineStage getPipeStage(final int defaultAddress) {
		return memory.getPipeStage(defaultAddress);
	}

	/**
	 * Marks a certain address whose instruction is loaded by the pipelines IFID
	 * stage.
	 *
	 * @param defaultAddress the address
	 */
	public void setIfIdStage(final int defaultAddress) {
		memory.setIfIdStage(defaultAddress);
	}

	/**
	 * delete the de.tu_darmstadt.rs.sim.main memory
	 */
	public void delete() {
		memory.delete();
		memory.restart();
	}

	// public boolean containsAddress(final int defaultAddress){
	// // if(memory.containsAddr(defaultAddress)){
	// // return true;
	// // }else{
	// // final Collection<AbstractPeripheral> units =
	// ioMap.get(defaultAddress);
	// // if ((units != null) && (units.size() != 0)){
	// // return true;
	// // }else{
	// // return false;
	// // }
	// // }
	// return true;
	// }

	/**
	 * register a peripheral in the address space.
	 *
	 * @param defaultAddress the base I/O address of the peripheral
	 * @param size the number of addresses this peripheral uses
	 * @param peripheral the peripheral
	 */
	public void registerPeripheral(final int defaultAddress, final int size,
								   final JConfigPeripheral peripheral) {
		List<AddressWithComponent> peripheralList = mappedPeripherals
				.get(defaultAddress);
		if (peripheralList == null) {
			peripheralList = new ArrayList<AddressSpaceManager.AddressWithComponent>();
			mappedPeripherals.put(defaultAddress, peripheralList);
		}
		peripheralList.add(new AddressWithComponent(size, peripheral));
		if (!peripherals.contains(peripheral)) {
			peripherals.add(peripheral);
		}
	}

	/**
	 * unregister a peripheral in the address space.
	 *
	 * @param defaultAddress the base I/O address of the peripheral
	 * @param peripheral the peripheral
	 */
	public void unregisterPeripheral(final int defaultAddress,
									 final JConfigPeripheral peripheral) {

		Entry<Integer, List<AddressWithComponent>> entry = mappedPeripherals
				.floorEntry(defaultAddress);
		if (entry != null) {
			List<AddressWithComponent> peripheralList = entry.getValue();
			for (int i = 0; i < peripheralList.size(); i++) {
				if (peripheralList.get(i).peripheral == peripheral) {
					peripheralList.remove(i);
					return;
				}
			}
		}
		peripherals.remove(peripheral);
	}

	/**
	 * returns the address space's size
	 *
	 * @return size of the address space
	 */
	@Override
	public int getSize() {
		return (Integer) Settings.getSetting(Settings.IO_ADDRESS_MAX)
				- (Integer) Settings.getSetting(Settings.MEMORY_ADDRESS_MIN)
				+ 1;
	}

	/**
	 * Asks all periphery devices to restart().
	 */
	@Override
	public void restart() {

		// ask each periphery unit to restart()
		for (final JConfigPeripheral p : peripherals) {
			p.restart();
		}
		memory.restart();
	}

	/**
	 * Performs setup procedures. Each periphery unit will be asked to setup()
	 * for itself. The periphery list will be rebuilt.
	 */
	@Override
	public void setup() {

		mappedPeripherals.clear();

		// Note: while a new IO map will be established, the clients list will
		// be recalculated, too
		// >> we have to copy it to iterate the clients correctly with their
		// setup() invocations

		for (final JConfigPeripheral p : new LinkedList<JConfigPeripheral>(
				peripherals)) {
			registerPeripheral(p.getBaseAddr(), p.getRegisters().length, p);
			p.setup();
		}
	}

	@Override
	public Collection<RefreshListener_I> getRefreshListeners() {
		return refreshListenerList;
	}

	@Override
	public int[] getDirtyAddresses() {
		if (hasModifiedData) {
			return new int[]{SpartanDefinitions_I.COMPLETE_REFRESH};
		} else {
			return new int[0];
		}
	}

	public void clearAllWatchpoints() {
		for (int address = 0; address < memory.getSize(); address++) {
			if (hasWatchpoint(address)) {
				toggleWatchpoint(address);
			}
		}
	}

	public void clearAllBreakpoints() {
		for (int address = 0; address < memory.getSize(); address++) {
			if (hasBreakpoint(address)) {
				toggleBreakpoint(address);
			}
		}
		if (!memoryBreakpoints.isEmpty()) {
			memoryBreakpoints.clear();
		}

	}

	public void saveInstruction(int address) {
		memoryBreakpoints.put(address, memory.getValueAt(address));
	}

	public void clearInstruction(int address) {
		if (memory.getValueAt(address) != memoryBreakpoints.get(address)) {
			memory.setValueAt(address, memoryBreakpoints.get(address));
		}
		memoryBreakpoints.remove(address, memoryBreakpoints.get(address));
	}
}
