/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine;

import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IDocumentHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;

import java.nio.file.Paths;
import java.util.List;

/**
 * Class for jconfig file import handling for peripherals and DCM
 */
public class JConfigDocumentManager {

	//private DocumentController documentController;
	private final IDocumentHandler handler;
	private IConfigurationDocument document;

	private SimulationManager fSimMgr;

	/**
	 * Constructor
	 *
	 * @param simMgr reference to the simulation manager
	 * @param documentHandler
	 */
	public JConfigDocumentManager(SimulationManager simMgr, IDocumentHandler documentHandler) {
		fSimMgr = simMgr;
		this.handler = documentHandler;
		//rootNode=builder.getSpartanRoot();
		//handler=documentHandler;
		//project=libJConfigProject;
		//progressIndicator = new GenericProgressIndicatorDummy();
	}

	/**
	 * check if a rootNode is set.
	 */
	private void nodeGuard() {
		if (handler == null) {
			fSimMgr.pushErrorMessage("JConfig document: load methods called before parsing the configuration");
		}
	}

	/**
	 * Set the indicator used when the jconfiglib environment is initialized.
	 * This method have only some effect before set/parse any jconfig xml file.
	 *
	 * @param indicator
	 *            the progress indicator
	 */
	/*public void setProgressIndicator(GenericProgressIndicator_I indicator) {
		if (indicator != null) {
			progressIndicator = indicator;
		}
	}*/

	/**
	 * Set the file from which the configuration should be read.
	 *
	 * @param filepath the path to the configuration file
	 */
	public void setJConfigFile(String filepath) {
		//remember jconfig file path
		Settings.setSetting(Settings.JCONFIG_FILE, filepath);

		handler.open(Paths.get(filepath));

		document = handler.getDocument();
	}

	/**
	 * retrieve the number of subsystems
	 *
	 * @return the number of subsystems in the jconfig xml file.
	 */
	public int getNumberOfSubsystems() {
		nodeGuard();

		final int size = getSubsystemNodes().size();

		return size;
	}

	/**
	 * retrieve the names of all subsystems
	 *
	 * @return an array with the names of all subsystems. The position in the
	 * array is equal to the subsystem number.
	 */
	public String[] getSubsystemNames() {
		nodeGuard();

		List<? extends ISubsystemConfiguration> subsystemNodes = getSubsystemNodes();

		String ret[] = new String[subsystemNodes.size()];
		int index = 0;

		for (IModuleConfiguration pm : subsystemNodes) {
			String name = pm.getName();

			ret[index] = name;
			index++;
		}

		return ret;
	}

	/**
	 * Load the configuration from the specified subsystem from the jconfig
	 * document tree. During the process peripherals and DCM are created and the
	 * global settings are imported.
	 *
	 * @param subsystemNumber number of the subsystem
	 */
	public void loadSubsystem(int subsystemNumber) {
		nodeGuard();

		Settings.setSetting(Settings.SUBSYSTEM_NUMBER, subsystemNumber);

		//get chosen subsystem node
		List<? extends ISubsystemConfiguration> subsystemNodes = getSubsystemNodes();
		ISubsystemConfiguration selectedSubsystem = null;
		int currentNumber = 0;
		for (ISubsystemConfiguration pm : subsystemNodes) {
			if (currentNumber >= subsystemNumber) {
				selectedSubsystem = pm;
				break;
			}
			currentNumber++;
		}

		assert (currentNumber == subsystemNumber);
		if (selectedSubsystem != null) {

			//import global settings
			//List<ISubsystemConfiguration> processorList = subsystemNodes;
			if (subsystemNodes.isEmpty()) {
				fSimMgr.pushErrorMessage("malformed JConfig document, processor is missing");
			}
			ISubsystemConfiguration selectedProcessor = subsystemNodes.iterator().next();
			Settings.importParameter(selectedProcessor);

			fSimMgr.getPeripheralManager().onLoadSubsystem(selectedSubsystem);
			fSimMgr.getDCMManager().onLoadSubsystem(document, selectedSubsystem);
			fSimMgr.getMpsocManager().setCoreId(selectedSubsystem.getName());
			fSimMgr.resetSimulation();
		}
	}

	/**
	 * reload the subsystem
	 */
	public void reloadSubsystem() {
		if (handler.hasLoadedDocument()) {
			fSimMgr.pushErrorMessage("reload is not possible: no configuration file specified");
			return;
		}
		if (Settings.SUBSYSTEM_NUMBER == Settings.SUBSYSTEM_INVALID) {
			fSimMgr.pushErrorMessage("reload is not possible: never loaded any subsystem before");
			return;
		}

		/*if(documentController == null){
			progressIndicator.start();
			Library.getInstance().load(progressIndicator);
			progressIndicator.cleanup();
			documentController = new DocumentController();
		return documentController;*/

		document = handler.getDocument();




		/*DocumentController controller = initDocumentControllerOnce();
		controller.enqueueEvent(new DocumentEvent(EventCode.LOAD_DOCUMENT,
				confFile));
		controller.processPendingEvents();
		rootNode = controller.getRootNode();*/

		loadSubsystem(Settings.SUBSYSTEM_NUMBER);
	}

	/**
	 * @return all Subsystem nodes, sorted by name
	 */
	private List<? extends ISubsystemConfiguration> getSubsystemNodes() {
		assert (document != null);
		//DocumentUtil.streamAllModules(document).collect(Collectors.toList());
		return document.getSubsystems();
	}

}
