/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

package de.tu_darmstadt.rs.sim.engine;

import libsim.DThread;
import libsim.SimulationSpeed_I;
import libsim.SimulationSpeed_I.Speed;
import libsim.TimedExec;

/**
 * Thread which is used to run the simulation
 */
public class SimulationThread extends DThread implements
		SimulationController_I, SettingListener_I {

	private SimulationManager fSimMgr;
	private final TimedExec fTimedExec;
	private volatile boolean fReset;
	private volatile boolean fRefresh;
	private volatile boolean fFreeRunning;
	private boolean fIsRunning;
	private SimulationSpeed_I.Speed fRunSpeed;
	private Object fCmdLock;

	/**
	 * create a new thread for the simulation
	 */
	public SimulationThread() {
		super("Main Application Thread");
		fRunSpeed = (SimulationSpeed_I.Speed) Settings
				.getSetting(Settings.SIMULATION_SPEED);
		Settings.registerListener(Settings.SIMULATION_SPEED, this);
		fTimedExec = new TimedExec();
		fTimedExec.register_task(new Runnable() {
			@Override
			public void run() {
				_resume();
			}
		});
		fCmdLock = new Object();
		_suspend();
	}

	/**
	 * sets the simulation manager
	 *
	 * @param sm the manager of the simulation
	 */
	public void setSimulationManager(SimulationManager sm) {
		fSimMgr = sm;
	}

	private void setRunning(boolean value) {
		if (fIsRunning == value) {
			return;
		}
		fIsRunning = value;
		if (fIsRunning) {
			fFreeRunning = fRunSpeed
					.equals(SimulationSpeed_I.Speed.FREE_RUNNING);
			if (fFreeRunning) {
				_resume();
			} else {
				fTimedExec.delay(fRunSpeed.delay());
				fTimedExec.resume_timer();
			}
		} else {
			fTimedExec.stop_timer();
			fFreeRunning = false;
			wait_for();
		}
	}

	@Override
	public void runSimulation() {
		synchronized (fCmdLock) {
			setRunning(true);
		}
	}

	@Override
	public void stepSimulation() {
		synchronized (fCmdLock) {
			setRunning(false);
			_resume();
		}
	}

	@Override
	public void stopSimulation() {
		synchronized (fCmdLock) {
			setRunning(false);
			fRefresh = true;
		}
	}

	@Override
	public void resetSimulation() {
		synchronized (fCmdLock) {
			setRunning(false);
			fReset = true;
			_resume();
		}
	}

	@Override
	public void _run() {
		try {
			if (fReset) {
				fReset = false;
				fSimMgr.resetSimulation();
			} else {
				fSimMgr.executeStep();
			}
		} catch (Exception e) {
			stopSimulation();
			e.printStackTrace();
		} finally {
			if (fRefresh) {
				fSimMgr.requestCompleteRefresh();
				fRefresh = false;
			}
			if (!fFreeRunning) {
				_suspend();
			}
		}
	}

	@Override
	public void onSettingChange(int settingType, Object newValue) {
		final SimulationSpeed_I.Speed spd = (SimulationSpeed_I.Speed) newValue;
		if (spd.equals(fRunSpeed)) {
			return;
		}
		fRunSpeed = spd;
		if (fIsRunning) {
			setRunning(false);
			setRunning(true);
		}
	}

	/**
	 * called whenever the simulation's speed is changed
	 *
	 * @param spd the new speed of the simulation
	 */
	@Override
	public void setSimulationSpeed(Speed spd) {
		if (spd.equals(fRunSpeed)) {
			return;
		}
		fRunSpeed = spd;
		Settings.setSetting(Settings.SIMULATION_SPEED, spd);
		if (fIsRunning) {
			setRunning(false);
			setRunning(true);
		}
	}

}
