/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.assembly;

import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.RefreshListener_I;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.memory.MainMemory;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.PipelineStage;

import java.util.ArrayList;
import java.util.List;

/**
 * This class is responsible for the management of the {@link AssemblyLine_I}
 * objects. This class is a singleton. To get the instance use
 * {@link #getTheInstance()}.
 */
public class AssemblyLineManager implements SpartanDefinitions_I,
		AssemblyLineDefinitions_I {

	private static AssemblyLineManager theInstance;
	private DummyAssemblyLine dummyAssemblyLine;
	private final AddressSpaceManager addressSpaceManager;
	private AssemblyLineProvider_I provider;
	private List<AssemblyLineProvider_I> availableProvider;
	private List<RefreshListener_I> refreshListenerList;
	private List<String> providerNames;

	private AssemblyLineManager(AddressSpaceManager addressSpaceManager) {
		dummyAssemblyLine = new DummyAssemblyLine();
		dummyAssemblyLine.data = "-- No Instruction --";
		this.addressSpaceManager = addressSpaceManager;
		availableProvider = new ArrayList<AssemblyLineProvider_I>();
		refreshListenerList = new ArrayList<RefreshListener_I>();
		providerNames = new ArrayList<String>();
	}

	/**
	 * retrieves the single instance of the {@link AssemblyLineManager}
	 *
	 * @return the single instance of this class
	 */
	public static AssemblyLineManager getTheInstance() {
		return theInstance;
	}

	/**
	 * get the {@link AssemblyLine_I} object at the given rowIndex
	 *
	 * @param rowIndex the row index where the line object should be searched
	 * @return the line object or a dummy line containing almost no useful
	 * information
	 */
	public AssemblyLine_I getLineObject(int rowIndex) {
		AssemblyLine_I assemblyLine;
		final int dataAddress;
		final int numberOfRows = provider.getNumberOfLines();
		if ((rowIndex < 0) || (rowIndex >= numberOfRows)) {
			assemblyLine = dummyAssemblyLine;
			if (numberOfRows == 0) {
				((DummyAssemblyLine) assemblyLine).dataAddress = rowIndex * 2;
			} else {
				((DummyAssemblyLine) assemblyLine).dataAddress = (provider
						.getLastDefinedAddress() & ~0x1)
						+ ((rowIndex - provider
						.getIndexOfLineWithLastDefinedAddress()) * 2);
			}
		} else {
			assemblyLine = provider.getAssemblyLine(rowIndex);
		}
		dataAddress = assemblyLine.getDataAddress();
		if (addressSpaceManager.hasWatchpoint(dataAddress >> 1)) {
			assemblyLine.setBreakOrWatchValue(WATCH_VAL);
		} else {
			assemblyLine.setBreakOrWatchValue(NONE_VAL);
		}
		if (addressSpaceManager.hasBreakpoint(dataAddress >> 1)) {
			assemblyLine.setBreakOrWatchValue(BREAK_VAL);
		}
		if (dataAddress != INVALID) {
			assemblyLine.setStage(addressSpaceManager
					.getPipeStage(dataAddress >> 1));
		} else {
			assemblyLine.setStage(PipelineStage.NOT_IN_PIPE);
		}
		return assemblyLine;
	}

	/**
	 * calculate the row index for an address (given in default addressing mode)
	 *
	 * @param defaultAddress the address
	 * @return the row index
	 */
	public int findRowIndexOfAddress(int defaultAddress) {
		final int dataAddress = defaultAddress << 1;
		final int numberOfRows = provider.getNumberOfLines();
		int index = dataAddress;
		if (numberOfRows == 0) {
			return -1;
		}
		if (index >= numberOfRows) {
			index = numberOfRows - 1;
		}
		AssemblyLine_I guessedLine = provider.getAssemblyLine(index);
		// search backward
		while ((guessedLine.getDataAddress() == INVALID)
				|| (guessedLine.getDataAddress() > dataAddress)) {
			index--;
			if (index < 0) {
				index++;
				break;
			}
			guessedLine = provider.getAssemblyLine(index);
		}
		// search forward
		while ((guessedLine.getDataAddress() == INVALID)
				|| (guessedLine.getDataAddress() < dataAddress)) {
			index++;
			if (index >= numberOfRows) {
				index--;
				break;
			}
			guessedLine = provider.getAssemblyLine(index);
		}
		return index;
	}

	/**
	 * toggle a watch or breakpoint. Whether a watch or breakpoint is used is
	 * determined using the {@link AssemblyLine_I#isData()} method: watchpoints
	 * are set on data lines, breakpoints on instructions.
	 *
	 * @param rowIndex the row's index
	 */
	public void toggle(int rowIndex) {
		AssemblyLine_I assemblyLine = getLineObject(rowIndex);
		final int dataAddress = assemblyLine.getDataAddress();
		if (dataAddress == INVALID) {
			return;
		}
		assert (MainMemory.containsAddress(dataAddress >> 1));
		if (assemblyLine.isData()) {
			addressSpaceManager.toggleWatchpoint(dataAddress >> 1);
		} else {
			addressSpaceManager.toggleBreakpoint(dataAddress >> 1);
		}
	}

	/**
	 * add a provider for the {@link AssemblyLine_I} objects
	 *
	 * @param lineObjectProvider a new provider for the data
	 */
	public void addLineObjectProvider(AssemblyLineProvider_I lineObjectProvider) {
		if (!availableProvider.contains(lineObjectProvider)) {
			availableProvider.add(lineObjectProvider);
			providerNames.add(lineObjectProvider.getProviderID());
		}
		if (provider == null) {
			provider = availableProvider.get(0);
			sendRefreshEvent();
		}
	}

	/**
	 * select a provider via its index from GUI
	 *
	 * @param index the index of the provider
	 */
	public void selectAssemblyLineProvider(int index) {
		provider = availableProvider.get(index);
		sendRefreshEvent();
	}

	/**
	 * retrieve the name of the selected {@link AssemblyLineProvider_I} object
	 *
	 * @return name of the selected provider
	 */
	public String getSelectedAssemblyLineProvider() {
		return provider.getProviderID();
	}

	/**
	 * retrieve a list with the names of the {@link AssemblyLineProvider_I}
	 * objects. This list should not be modified.
	 *
	 * @return names of the providers
	 */
	public List<String> getAssemblyLineProviderNames() {
		return providerNames;
	}

	/**
	 * register a listener which should be invoked whenever the data changes
	 * (this happens if another {@link AssemblyLineProvider_I} is selected or
	 * added)
	 *
	 * @param listener the listener to register
	 */
	public void registerRefreshListener(RefreshListener_I listener) {
		if (!refreshListenerList.contains(listener)) {
			refreshListenerList.add(listener);
		}
		listener.refresh(COMPLETE_REFRESH);
	}

	/**
	 * unregister a listener previously registered with
	 * {@link #registerRefreshListener(RefreshListener_I)}
	 *
	 * @param listener the listener which should not be invoked any more
	 */
	public void unregisterRefreshListener(RefreshListener_I listener) {
		refreshListenerList.remove(listener);
	}

	private void sendRefreshEvent() {
		for (RefreshListener_I listener : refreshListenerList) {
			listener.refresh(COMPLETE_REFRESH);
		}
	}

	/**
	 * create the single instance of the {@link AssemblyLineManager}
	 *
	 * @param addressSpaceManager reference to the {@link AddressSpaceManager}
	 */
	public static void createTheInstance(AddressSpaceManager addressSpaceManager) {
		if (theInstance == null) {
			theInstance = new AssemblyLineManager(addressSpaceManager);
		}
	}
}
