/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.assembly;

import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.RefreshListener_I;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * The AssemblyVariablesManager provides access to the variables from the .lst
 * program. It manages the list of variables provides a mechanism to ask for a
 * variables value.
 */
public class AssemblyVariablesManager implements SpartanDefinitions_I {

	private static final int STOP_VALUE = 20;
	private List<AssemblyVariable> variableList;
	private final AddressSpaceManager addressSpace;
	private AssemblyVariable variableToBeCreated;
	private int currentDataAddress;
	private boolean isCreationOfVariablesEnabled;
	private List<AssemblyVariableInterpretation_I> interpretations;
	private Map<AssemblyVariable, AssemblyVariableInterpretation_I> selectedInterpretations;

	/**
	 * create a new {@link AssemblyVariablesManager}
	 *
	 * @param addressSpace reference to the {@link AddressSpaceManager} which is used to
	 * get the value of variables
	 */
	public AssemblyVariablesManager(final AddressSpaceManager addressSpace) {
		this.addressSpace = addressSpace;
		variableList = new ArrayList<AssemblyVariable>();
		interpretations = new ArrayList<AssemblyVariableInterpretation_I>();
		interpretations.add(new Hex18Interpretation());
		interpretations.add(new StringInterpretation());
		interpretations.add(new Dec18Interpretation());
		interpretations.add(new Byte9Interpretation());
		selectedInterpretations = new HashMap<AssemblyVariable, AssemblyVariableInterpretation_I>();
	}

	/**
	 * retrieves how many variables are available
	 *
	 * @return the number of variables
	 */
	public int getNumberOfVariables() {
		return variableList.size();
	}

	/**
	 * add a new variable to the {@link AssemblyVariablesManager}
	 *
	 * @param variable the new variable
	 */
	public void add(AssemblyVariable variable) {
		if (!variableList.contains(variable)) {
			variableList.add(variable);
			selectedInterpretations.put(variable, interpretations.get(0));
		}
	}

	/**
	 * remove all the variables. When called the
	 * {@link AssemblyVariablesManager} does not contain any variables any more.
	 */
	public void removeAllVariables() {
		variableList.clear();
		selectedInterpretations.clear();
	}

	/**
	 * get a variable from the list of all variables
	 *
	 * @param index index of the variable on the list
	 * @return the variable at the index
	 */
	public AssemblyVariable getVariableAtIndex(int index) {
		return variableList.get(index);
	}

	private String getValueOf(AssemblyVariable variable,
							  AssemblyVariableInterpretation_I interpretation) {
		int theDataAddress = variable.getDataAddress();
		int singleValueData;
		final int upperBound = variable.getDataAddress()
				+ variable.getLengthIn9Bytes();
		singleValueData = getMemoryValue(theDataAddress,
				interpretation.is9BitInterpretation());
		final int increment;
		int numberOfReadValues = 1;
		if (isSingleValue(variable, interpretation)) {
			interpretation.workOnSingleValue(singleValueData);
		} else {
			interpretation.prepareMultiValue(singleValueData);
			if (interpretation.is9BitInterpretation()) {
				increment = 1;
			} else {
				increment = 2;
			}
			theDataAddress += increment;
			for (; theDataAddress < upperBound; theDataAddress += increment) {
				if (numberOfReadValues >= STOP_VALUE) {
					interpretation.onStopRequested();
					break;
				}
				interpretation.workOnValue(getMemoryValue(theDataAddress,
						interpretation.is9BitInterpretation()));
				numberOfReadValues++;
			}
			interpretation.finishMultiValue();
		}
		return interpretation.getValue();
	}

	private boolean isSingleValue(AssemblyVariable variable,
								  AssemblyVariableInterpretation_I interpretation) {
		return ((variable.getLengthIn9Bytes() == 1) && (interpretation
				.is9BitInterpretation()))
				|| ((variable.getLengthIn9Bytes() == 2) && (!interpretation
				.is9BitInterpretation()));
	}

	private int getMemoryValue(final int dataAddress, final boolean is9Bit) {
		int returnValue = addressSpace.getValue(dataAddress >> 1);
		final int mask;
		final int shiftValue;
		if (is9Bit) {
			if ((dataAddress & 0x1) == 0) {
				mask = BITMASK_9 << 9;
				shiftValue = 9;
			} else {
				mask = BITMASK_9;
				shiftValue = 0;
			}
		} else {
			mask = BITMASK_18;
			shiftValue = 0;
		}
		returnValue = (returnValue & mask) >> shiftValue;
		return returnValue;
	}

	/**
	 * begin creating a variable with a start address and a name
	 *
	 * @param dataAddress the start address of the variable
	 * @param name the name of the variable
	 */
	public void beginCreationOfVariable(int dataAddress, String name) {
		if (isCreationOfVariablesEnabled) {
			variableToBeCreated = new AssemblyVariable(dataAddress, name);
		}
	}

	/**
	 * add the data address to the address space of the variable
	 *
	 * @param dataAddress the next address which is part of the address space of the
	 * variable
	 */
	public void workOnCreationOfVariable(int dataAddress) {
		currentDataAddress = dataAddress;
	}

	/**
	 * end adding data addresses and finish the creation of a variable
	 */
	public void finishCreationOfVariable() {
		if (isCreationOfVariablesEnabled) {
			if (variableToBeCreated != null) {
				variableToBeCreated.setLengthIn9Bytes(currentDataAddress
						- variableToBeCreated.getDataAddress() + 1);
				add(variableToBeCreated);
				variableToBeCreated = null;
			}
		}
	}

	/**
	 * allow creation of variables. In combination with
	 * {@link #disableCreationOfVariables()} this method is used to prevent
	 * creation of variables in some sections of the assembly.
	 */
	public void enableCreationOfVariables() {
		isCreationOfVariablesEnabled = true;
	}

	/**
	 * disallow creation of variables. In combination with
	 * {@link #enableCreationOfVariables()} this method is used to prevent
	 * creation of variables in some sections of the assembly.
	 */
	public void disableCreationOfVariables() {
		isCreationOfVariablesEnabled = false;
	}

	/**
	 * retrieve the number of possible interpretations of a variable
	 *
	 * @return the number of variable interpretations
	 */
	public int getNumberOfInterpretations() {
		return interpretations.size();
	}

	/**
	 * get the name of an interpretation at the list index given by parameter
	 * index
	 *
	 * @param index the index on the list
	 * @return the name of the interpretation
	 */
	public String getInterpretation(int index) {
		return interpretations.get(index).getName();
	}

	/**
	 * retrieve the value(s) of a variable
	 *
	 * @param variable the variable to which a value should be found
	 * @return the value of selected interpretation for the variable as a string
	 */
	public String getInterpretedValueOf(AssemblyVariable variable) {
		return getValueOf(variable, selectedInterpretations.get(variable));
	}

	/**
	 * select an interpretation for a variable
	 *
	 * @param index the list index of the variable
	 * @param interpretationName the name of the interpretation to select
	 */
	public void setSelectedInterpretation(int index, String interpretationName) {
		AssemblyVariable variable = variableList.get(index);
		AssemblyVariableInterpretation_I interpretation = searchInterpretation(interpretationName);
		selectedInterpretations.put(variable, interpretation);
	}

	private AssemblyVariableInterpretation_I searchInterpretation(
			String interpretationName) {
		for (AssemblyVariableInterpretation_I theInterpretation : interpretations) {
			if (theInterpretation.getName().equals(interpretationName)) {
				return theInterpretation;
			}
		}
		AssemblyVariableInterpretation_I interpretation = interpretations
				.get(0);
		return interpretation;
	}

	/**
	 * the name of the selected interpretation of a variable
	 *
	 * @param variable the variable
	 * @return name of the selected interpretation
	 */
	public String getInterpretationOfVariable(AssemblyVariable variable) {
		return selectedInterpretations.get(variable).getName();
	}

	/**
	 * register a listener which should be called whenever the variable's value
	 * is changed (possible by every memory change)
	 *
	 * @param listener the listener
	 */
	public void registerRefreshListener(RefreshListener_I listener) {
		addressSpace.addRefreshListener(listener);
	}

	/**
	 * unregister a previously registered {@link RefreshListener_I}
	 *
	 * @param listener the listener
	 */
	public void unregisterRefreshListener(RefreshListener_I listener) {
		addressSpace.removeRefreshListener(listener);
	}
}
