/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.dcm;


import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.TearDownListener_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.SignalHelper;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;

import java.util.ArrayList;
import java.util.List;

/**
 * Abstract base class for  {@link SystemClockDCM})
 */
public abstract class AbstractDCM {

	/**
	 * name of the jconfig parameter of a DCM's CLKDV_DIVIDE value
	 */
	public static final String PARAM_CLKDV_DIVIDE = "CLKDV_DIVIDE";
	/**
	 * name of the jconfig parameter of a DCM's CLKFX_DIVIDE value
	 */
	public static final String PARAM_CLKFX_DIVIDE = "CLKFX_DIVIDE";
	/**
	 * name of the jconfig parameter of a DCM's CLKIN_DIVIDE_BY_2 value
	 */
	public static final String PARAM_CLKIN_DIVIDE_BY_2 = "CLKIN_DIVIDE_BY_2";
	/**
	 * name of the jconfig parameter of a DCM's CLKFX_MULTIPLY value
	 */
	public static final String PARAM_CLKFX_MULTIPLY = "CLKFX_MULTIPLY";

	/**
	 * zzzzz
	 * reference to the DCMManager
	 */
	protected DCMManager dcmManager;
	/**
	 * reference to the DCMSignalInvocationManager
	 */
	private DCMSignalInvocationManager invocationManager;

	/**
	 * the DCM's source
	 */
	private AbstractDCMSignal inputSignal;

	/**
	 * the DCM's name
	 */
	private String name;

	/**
	 * a unique identification
	 */
	private String globalUniqueID;

	/**
	 * the value of the CLKIN_DIVIDE_BY_2 property
	 */
	private boolean clkInDivideBy2;

	/**
	 * the value of the CLKFX_MULTIPLY property
	 */
	private CLKFX_MULTIPLY clkFxMultiply;

	/**
	 * the value of the CLKFX_DIVIDE property
	 */
	private CLKFX_DIVIDE clkFxDivide;

	/**
	 * the value of the CLKDV_DIVIDE property
	 */
	//private CLKDV_DIVIDE clkDvDivide;

	/**
	 * the CLK0 output
	 */
	//protected CLK0 clk0;
	/**
	 * the CLK90 output
	 */
	//protected CLK90 clk90;
	/**
	 * the CLK180 output
	 */
	//protected CLK180 clk180;
	/**
	 * the CLK270 output
	 */
	//protected CLK270 clk270;
	/**
	 * the CLK2X output
	 */
	//protected CLK2X clk2X;
	/**
	 * the CLK2X180 output
	 */
	//protected CLK2X180 clk2X180;
	/**
	 * the CLKDV output
	 */
	//protected CLKDV clkDV;
	/**
	 * the CLKFX output
	 */
	protected CLKFX clkFX;
	/**
	 * the CLKFX180 output
	 */
	//protected CLKFX180 clkFX180;

	private List<TearDownListener_I> teardownListeners;

	/**
	 * abstract constructor for instances of subclasses
	 *
	 * @param simulationManager a reference to the simulation manager
	 * @param name the name of this DCM
	 */
	public AbstractDCM(SimulationManager simulationManager, String name) {
		this.name = name;
		this.dcmManager = simulationManager.getDCMManager();
		this.invocationManager = simulationManager.getDCMSignalInvocationManager();
		//default values
		clkInDivideBy2 = false;
		clkFxMultiply = CLKFX_MULTIPLY.DEFAULT_VALUE;
		clkFxDivide = CLKFX_DIVIDE.DEFAULT_VALUE;
		//clkDvDivide=CLKDV_DIVIDE.DEFAULT_VALUE;
		createSignals();
		teardownListeners = new ArrayList<TearDownListener_I>();
	}

	/**
	 * This callback method should create the signals
	 * because the signals could have special names.
	 * It is invoked by the constructor of {@link AbstractDCM}.
	 */
	protected abstract void createSignals();

	/**
	 * retrieve the DCM's name
	 *
	 * @return name of the DCM
	 */
	public String getDCMName() {
		return name;
	}

	/**
	 * register a DCMSignalInvoker object
	 *
	 * @param signalInvoker DCMSignalInvoker object
	 */
	public void registerDCMSignalInvoker(DCMSignalInvoker signalInvoker) {
		invocationManager.registerDCMSignalInvoker(signalInvoker);
	}

	/**
	 * unregister a DCMSignalInvoker object
	 *
	 * @param signalInvoker DCMSignalInvoker object
	 */
	public void unregisterDCMSignalInvoker(DCMSignalInvoker signalInvoker) {
		invocationManager.unregisterDCMSignalInvoker(signalInvoker);
	}

	/**
	 * get the used FX_MULTIPLY value
	 *
	 * @return the FX_MULTIPLY value
	 */
	public CLKFX_MULTIPLY getClkFxMultiply() {
		return clkFxMultiply;
	}

	/**
	 * set a different FX_MULTIPLY value and starts the preparation of the clocking mechanism
	 *
	 * @param clkFxMultiply the new FX_MULTIPLY value which should be used
	 */
	public void setClkFxMultiply(CLKFX_MULTIPLY clkFxMultiply) {
		this.clkFxMultiply = clkFxMultiply;
		invocationManager.prepareClocking();
	}

	/**
	 * get the used FX_DIVIDE value
	 *
	 * @return the FX_DIVIDE value
	 */
	public CLKFX_DIVIDE getClkFxDivide() {
		return clkFxDivide;
	}

	/**
	 * set a different FX_DIVIDE value and starts the preparation of the clocking mechanism
	 *
	 * @param clkFxDivide the new FX_DIVIDE value which should be used
	 */
	public void setClkFxDivide(CLKFX_DIVIDE clkFxDivide) {
		this.clkFxDivide = clkFxDivide;
		invocationManager.prepareClocking();
	}

	/**
	 * get the used DV_DIVIDE value
	 *
	 * @return the DV_DIVIDE value
	 */
	/*public CLKDV_DIVIDE getClkDvDivide() {
		return clkDvDivide;
	}

	/**
	 * set a different DV_DIVIDE value and starts the preparation of the clocking mechanism
	 * @param clkDvDivide the new DV_DIVIDE value which should be used
	 */
	/*public void setClkDvDivide(CLKDV_DIVIDE clkDvDivide) {
		this.clkDvDivide = clkDvDivide;
		invocationManager.prepareClocking();
	}

	/**
	 * retrieve the input signal
	 * @return the input signal or null if the DCM is the first in the chain of DCMs
	 */
	public AbstractDCMSignal getInputSignal() {
		return inputSignal;
	}

	/**
	 * set a new input signal for the DCM. Performs the necessary register and unregister calls.
	 *
	 * @param signal the new input signal
	 */
	public void setInputSignal(AbstractDCMSignal signal) {
		if (inputSignal != null) {
			inputSignal.unregisterDCM(this);
		}
		inputSignal = signal;
		inputSignal.registerDCM(this);
	}

	/**
	 * retrieves whether the DCM's input is divided by 2
	 * (if this is the case the divided value is used for calculation of the output signals)
	 *
	 * @return true if the DCM's input is divided by 2
	 */
	public boolean isClkInDivideBy2() {
		return clkInDivideBy2;
	}

	/**
	 * set whether the DCM's input is divided by 2
	 * (if this is the case the divided value is used for calculation of the output signals)
	 *
	 * @param clkInDivideBy2 should be set to true if the DCM's input is divided by 2
	 */
	public void setClkInDivideBy2(boolean clkInDivideBy2) {
		this.clkInDivideBy2 = clkInDivideBy2;
		invocationManager.prepareClocking();
	}

	/**
	 * retrieve the unique identification
	 *
	 * @return a unique identification
	 */
	public String getGlobalUniqueID() {
		return globalUniqueID;
	}

	/**
	 * set the unique identification
	 *
	 * @param globalUniqueID the identification value
	 */
	public void setGlobalUniqueID(String globalUniqueID) {
		this.globalUniqueID = globalUniqueID;
	}

	/**
	 * import hardware parameters from a section of the jconfig file
	 *
	 * @param module the jconfig file's section
	 */
	public void importHardwareParameters(IModuleConfiguration module) {
		/**
		 * the frequency can be directly extracted, so no need for division or multiplication
		 */
		importConnection(module);
		invocationManager.prepareClocking();
		/*Collection<VerilogParameter> parameters=documentNode.getChildren(VerilogParameter.class);
		for(VerilogParameter parameter:parameters){
			if(parameter.getName().equals(PARAM_CLKIN_DIVIDE_BY_2)){
				clkInDivideBy2=parameter.getBooleanValue();
			}else if(parameter.getName().equals(PARAM_CLKDV_DIVIDE)){
				double value=parameter.getNumericValue().doubleValue();
				clkDvDivide=CLKDV_DIVIDE.getByValue(value);
			}else if(parameter.getName().equals(PARAM_CLKFX_DIVIDE)){
				int value=parameter.getNumericValue().intValue();
				clkFxDivide=CLKFX_DIVIDE.getByValue(value);
			}else if(parameter.getName().equals(PARAM_CLKFX_MULTIPLY)){
				int value=parameter.getNumericValue().intValue();
				clkFxMultiply=CLKFX_MULTIPLY.getByValue(value);
			}
		}*/
		/*IBusHandler busHandler=DocumentUtil.getBusHandler(module,"clkFX");

			if(module.getModuleDescription().findParameterDescription(PARAM_CLKIN_DIVIDE_BY_2).isPresent()){
				clkInDivideBy2=module.getModuleDescription().findParameterDescription(PARAM_CLKIN_DIVIDE_BY_2).get().hasCalculatedValue();
			}
			/*else if(subsystem.getModuleDescription().findParameterDescription(PARAM_CLKDV_DIVIDE).isPresent()){
				double value=Double.parseDouble(subsystem.getModuleDescription().findParameterDescription(PARAM_CLKDV_DIVIDE).get().getDefaultValue());
				clkDvDivide=CLKDV_DIVIDE.getByValue(value);
			}
			else if(module.getModuleDescription().findParameterDescription(PARAM_CLKFX_DIVIDE).isPresent()){
				int value=Integer.parseInt(module.getModuleDescription().findParameterDescription(PARAM_CLKFX_DIVIDE).get().getDefaultValue());
				clkFxDivide=CLKFX_DIVIDE.getByValue(value);
			}
			else if(module.getModuleDescription().findParameterDescription(PARAM_CLKFX_MULTIPLY).isPresent()){
				int value=Integer.parseInt(module.getModuleDescription().findParameterDescription(PARAM_CLKFX_MULTIPLY).get().getDefaultValue());
				clkFxMultiply=CLKFX_MULTIPLY.getByValue(value);
			}*/


	}

	/**
	 * import the connection from the DCM's source from the jconfig file's section
	 *
	 * @param module
	 */
	private void importConnection(IModuleConfiguration module) {
		String signalName;
		String connectedSignal;
		IPortHandler cNode = getInputPort(module);
		signalName = cNode.getPortName();
		connectedSignal = SignalHelper.getTimerNameFromConnectedSignal(signalName);
		dcmManager.connectDCMToSignal(connectedSignal, this);
	}

	/**
	 * get the input port from documentNode. This method is a callback called when the hardware
	 * parameters are imported.
	 *
	 * @param subsystem the subsystem node representing this DCM
	 * @return the input port
	 */
	protected abstract IPortHandler getInputPort(IModuleConfiguration subsystem);

	/**
	 * register a tear down listener
	 *
	 * @param listener a tear down listener
	 */
	public void registerTearDownListener(TearDownListener_I listener) {
		assert (listener != null);
		teardownListeners.add(listener);
	}

	/**
	 * unregister a tear down listener
	 *
	 * @param listener a tear down listener
	 */
	public void unregisterTearDownListener(TearDownListener_I listener) {
		assert (teardownListeners.contains(listener));
		assert (listener != null);
		teardownListeners.remove(listener);
	}

	/**
	 * retrieve whether the DCM has tear down listeners
	 *
	 * @return true if the DCM has tear down listeners
	 */
	public boolean hasTearDownListeners() {
		return teardownListeners.size() != 0;
	}
}
