/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.dcm;

import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerObserver_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerSource_I;

import java.util.ArrayList;
import java.util.List;

/**
 * abstract base class for all the DCM's output signals
 */
public abstract class AbstractDCMSignal implements TimerSource_I {

	private List<TimerObserver_I> dcmObservers;
	/**
	 * the DCM which has the signal as an output
	 */
	protected AbstractDCM dcm;
	/**
	 * the DCMSignal's name
	 */
	private String name;

	/**
	 * how often is this DCMSignal emitted in a period?
	 */
	private long howOftenClockedPerPeriod;
	/**
	 * how many clock steps are done in a period?
	 */
	private long clockStepsInOnePeriod;
	/**
	 * the phase shift of the DCMSignal
	 */
	//private QuadrantPhaseShift shift;

	/**
	 * the DCMSignalInvoker used by this instance
	 */
	protected DCMSignalInvoker dcmSignalInvoker;

	/**
	 * abstract constructor used by subclasses. The signal is related to the input signal of the DCM
	 * by the fraction of <code>howOftenClockedPerPeriod / clockStepsInOnePeriod</code>.
	 *
	 * @param dcm backward reference to the DCM that has this signal
	 * @param name the name of this signal
	 * @param howOftenClockedPerPeriod how often is the signal clocked
	 * @param clockStepsInOnePeriod how often is a period clocked
	 */
	public AbstractDCMSignal(AbstractDCM dcm, String name, long howOftenClockedPerPeriod, long clockStepsInOnePeriod) {
		this.dcm = dcm;
		this.howOftenClockedPerPeriod = howOftenClockedPerPeriod;
		this.clockStepsInOnePeriod = clockStepsInOnePeriod;
		//this.shift=shift;
		this.name = name;
		dcmObservers = new ArrayList<TimerObserver_I>();
	}

	/**
	 * retrieve how often the signal is clocked in a period
	 *
	 * @return how often
	 */
	public long getHowOftenClockedPerPeriod() {
		return howOftenClockedPerPeriod;
	}

	/**
	 * retrieve how many clock steps are done
	 *
	 * @return number of clock steps
	 */
	public long getClockStepsInOnePeriod() {
		long retVal = clockStepsInOnePeriod;
		if (dcm.isClkInDivideBy2()) {
			retVal *= 2;
		}
		return retVal;
	}

	/**
	 * get the DCMSignal's phase shift
	 *
	 * @return phase shift
	 */
	/*public QuadrantPhaseShift getPhaseShift(){
		return shift;
	}*/
	@Override
	public String getUniqueTimerName() {
		return dcm.getDCMName() + "_" + name;
	}

	@Override
	public void registerTimerObserver(TimerObserver_I timerObserver) {
		dcmObservers.add(timerObserver);
		if (dcmSignalInvoker == null) {
			dcmSignalInvoker = new DCMSignalInvoker(this);
			dcm.registerDCMSignalInvoker(dcmSignalInvoker);
		}
	}

	@Override
	public void unregisterTimerObserver(TimerObserver_I timerObserver) {
		dcmObservers.remove(timerObserver);
		if (dcmObservers.isEmpty()) {
			dcm.unregisterDCMSignalInvoker(dcmSignalInvoker);
			dcmSignalInvoker = null;
		}
	}

	/**
	 * inform all Observers that a clock is executed
	 */
	public void informAllObservers() {
		for (TimerObserver_I timerObserver : dcmObservers) {
			timerObserver.onClock(1);
			timerObserver.onClock(0);
		}
	}

	@Override
	public int getOutputBitCount() {
		return 1;
	}

	/**
	 * get the DCM that the DCMSignal belongs to
	 *
	 * @return DCM
	 */
	public AbstractDCM getDcm() {
		return dcm;
	}

	/**
	 * retrieve the DCMSignalInvoker that is used for this Signal
	 *
	 * @return a DCMSignalInvoker
	 */
	public DCMSignalInvoker getDCMSignalInvoker() {
		return dcmSignalInvoker;
	}

	/**
	 * called when a DCM connects to this DCMSignal
	 *
	 * @param otherDCM reference to the DCM that connects (unused)
	 */
	public void registerDCM(AbstractDCM otherDCM) {
		if (dcmSignalInvoker == null) {
			dcmSignalInvoker = new DCMSignalInvoker(this);
			dcm.registerDCMSignalInvoker(dcmSignalInvoker);
		}
	}

	/**
	 * called when a DCM disconnects from this DCMSignal
	 *
	 * @param otherDCM reference to the DCM that connects (unused)
	 */
	public void unregisterDCM(AbstractDCM otherDCM) {
		if (dcmObservers.isEmpty()) {
			dcm.unregisterDCMSignalInvoker(dcmSignalInvoker);
			dcmSignalInvoker = null;
		}
	}

	/**
	 * retrieve whether this DCMSignal has observers
	 *
	 * @return true if the DCMSignal has observers
	 */
	public boolean hasTimerObservers() {
		return !dcmObservers.isEmpty();
	}
}
