/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.dcm;

import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.JConfigPeripheral;
import de.tu_darmstadt.rs.sim.engine.peripheral.OnLoadedListener_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.SignalHelper;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.ISubsystemConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * manages all DCM instances<br>
 * this includes imports from a jconfig file and also GUI access
 */
public class DCMManager {

	/**
	 * used to identify the input port of the processor
	 */
	public static final String PORT_PROCESSOR_INPUT = "clk";
	private List<AbstractDCM> dcms;
	private List<AbstractDCMSignal> dcmSignals;
	private Map<String, List<AbstractDCM>> dcmConnections;
	private final SimulationManager simMgr;
	private final List<OnLoadedListener_I> onLoadedListenerList;

	/**
	 * create a new {@link DCMManager}
	 *
	 * @param simMgr the {@link SimulationManager} required for further
	 * functionality
	 */
	public DCMManager(SimulationManager simMgr) {
		this.simMgr = simMgr;
		dcms = new ArrayList<AbstractDCM>();
		dcmSignals = new ArrayList<AbstractDCMSignal>();
		dcmConnections = new HashMap<String, List<AbstractDCM>>();
		onLoadedListenerList = new ArrayList<OnLoadedListener_I>();
	}

	/**
	 * @return all GlueLogic nodes, sorted by name
	 */
	private Collection<IModuleConfiguration> getGlueLogics(IConfigurationDocument rootNode) {
		return rootNode.getModules().stream().collect(Collectors.toList());
	}

	/**
	 * search through the list looking for a DCM with matching id.
	 *
	 * @param list the DCM list
	 * @param id the id
	 * @return null if no matching DCM was found, else the DCM with given id
	 */
	private static AbstractDCM searchDCM(List<AbstractDCM> list, String id) {
		AbstractDCM ret = null;

		for (AbstractDCM p : list) {
			if (p.getDCMName().equals(id)) {
				ret = p;
				break;
			}
		}

		return ret;
	}

	/**
	 * create new DCM instance and initialize it.
	 *
	 * @param node GlueLogic node describing the glue logic
	 * @return null if DCM is not supported, the new DCM instance otherwise
	 */
	private AbstractDCM createGlueLogic(IModuleConfiguration node) {
		AbstractDCM instance = null;
		String moduleType = node.getType().name();
		String name = node.getNameHandler().getValue().orElseThrow();
		if (moduleType.contains("clk")) {
			instance = new SystemClockDCM(simMgr, name);
		}
		/**
		 * the new jconfig does not differentiate between user and system clock
		 */
		/*else if (moduleType.equals("userclk")) {
			instance = new UserClockDCM(simMgr, name);
		}*/
		if (instance != null) {
			instance.importHardwareParameters(node);
		}
		return instance;
	}

	/**
	 * load the selected processor's clock source
	 *
	 * @param rootNode jconfig file representation
	 * @param selectedSubsystem the selected subsystem
	 */
	public void onLoadSubsystem(IConfigurationDocument rootNode, ISubsystemConfiguration selectedSubsystem) {
		loadGlueLogics(rootNode);
		loadProcessorClockSource(selectedSubsystem);
		for (OnLoadedListener_I listener : onLoadedListenerList) {
			listener.onLoaded(this);
		}
	}

	/**
	 * Load the configuration from the specified subsystem from the jconfig
	 * document tree. During the process the DCM is created and the global
	 * settings are imported.
	 *
	 * @param rootNode the document node which contains the glue logic elements
	 */
	public void loadGlueLogics(IConfigurationDocument rootNode) {
		List<AbstractDCM> oldDCMS = dcms;
		dcms = new ArrayList<AbstractDCM>();
		String id = "";

		// fetch all glueLogic nodes
		Collection<IModuleConfiguration> glueLogicNodes = getGlueLogics(rootNode);

		// update/create glueLogic
		for (IModuleConfiguration pNode : glueLogicNodes) {
			id = pNode.getModuleDescriptionName();
			AbstractDCM dcm = searchDCM(oldDCMS, id);
			if (dcm == null) {
				// DCM is new
				dcm = createGlueLogic(pNode);

				// GlueLogic is not a dcm
				if (dcm == null) {
					continue;
				}

				// important set the global dcm id
				dcm.setGlobalUniqueID(id);

				dcms.add(dcm);

			} else {
				// dcm already exists -> update dcm
				dcm.importHardwareParameters(pNode);

				if (oldDCMS.remove(dcm)) {
					dcms.add(dcm);
				}
			}
		}
	}

	/**
	 * fetch the DCM with matching id. If an invalid id is passed no search is
	 * done.
	 *
	 * @param id the id of the desired DCM
	 * @return null if no matching DCM was found, otherwise the DCM with
	 * matching id
	 */
	public AbstractDCM getDCM(String id) {
		if (id.equals(JConfigPeripheral.INVALID_ID)) {
			return null;
		}
		return searchDCM(dcms, id);
	}


	/**
	 * load the processor's clock source
	 *
	 * @param node the jconfig processor representation
	 */
	public void loadProcessorClockSource(IModuleConfiguration node) {
		String signalName;
		String connectedTimerName;
		BitSet connectedBits;
		int inputFilterBitIndex;
		/*
		 * TODO:check if this piece of code can take advantage of new libjconfig
		 * method SpartanModule.getConnectedModule()
		 */

		IPortHandler cNode = DocumentUtil.getPortHandler(node, PORT_PROCESSOR_INPUT);
		/* HACK to prevent np exception when there is no connection to DCM */
		if (cNode.isConnected()) {
			// assuming that the first Connection is the correct one
			//IConnection cep = (IConnection)cNode.streamConnections().iterator().next();
			signalName = cNode.getPortName();
			connectedTimerName = SignalHelper
					.getTimerNameFromConnectedSignal(signalName);
			TimerManager.addTimerConnection(connectedTimerName,
					simMgr.getPipelineClock());
			connectedBits = new BitSet(cNode.getConnectorWidth());
			inputFilterBitIndex = SignalHelper
					.getIndexOfHighestConnectedBit(connectedBits);
			simMgr.getPipelineClock().setInputFilterBit(inputFilterBitIndex);
		}
	}

	/**
	 * register a DCM signal
	 *
	 * @param signal a DCM signal
	 */
	public void registerDCMSignal(AbstractDCMSignal signal) {
		List<AbstractDCM> dcmList;
		if (!dcmSignals.contains(signal)) {
			dcmSignals.add(signal);
			// check if the DCMManager remembers connections
			dcmList = dcmConnections.get(signal.getUniqueTimerName());
			if (dcmList != null) {
				// it remembers connections, so connect it
				for (AbstractDCM dcm : dcmList) {
					// signal.registerDCM(dcm);
					dcm.setInputSignal(signal);
				}
			}
			// every dcm signal is also a timerSource, so register it
			TimerManager.addTimerSource(signal);
		}
	}

	/**
	 * unregister a DCM signal
	 *
	 * @param signal a DCM signal
	 */
	public void unregisterDCMSignal(AbstractDCMSignal signal) {
		assert (dcmSignals.contains(signal));
		dcmSignals.remove(signal);
		TimerManager.removeTimerSource(signal);
	}

	/**
	 * connect a DCM's output with another DCM
	 *
	 * @param signalName name of the DCM's output
	 * @param dcm the other DCM
	 */
	public void connectDCMToSignal(String signalName, AbstractDCM dcm) {
		AbstractDCMSignal signal = getDCMSignalByName(signalName);
		if (signal == null) {
			List<AbstractDCM> dcmList = dcmConnections.get(signalName);
			if (dcmList == null) {
				dcmList = new ArrayList<AbstractDCM>();
				dcmConnections.put(signalName, dcmList);
			}
			dcmList.add(dcm);
		} else {
			dcm.setInputSignal(signal);
		}
	}

	/**
	 * get a DCMSignal by name
	 *
	 * @param name the DCMSignal's name
	 * @return the DCMSignal if found, null otherwise
	 */
	private AbstractDCMSignal getDCMSignalByName(String name) {
		for (AbstractDCMSignal signal : dcmSignals) {
			if (signal.getUniqueTimerName().equals(name)) {
				return signal;
			}
		}
		return null;
	}

	/**
	 * get a DCM which has no representation in the GUI
	 *
	 * @return an unrepresented DCM
	 */
	public AbstractDCM fetchDCM() {
		for (AbstractDCM dcm : dcms) {
			if (!dcm.hasTearDownListeners()) {
				return dcm;
			}
		}
		return null;
	}

	/**
	 * register a listener which will be invoked when the XML information are
	 * loaded
	 *
	 * @param listener the listener which should be invoked
	 */
	public void registerOnLoadedListener(OnLoadedListener_I listener) {
		if (!onLoadedListenerList.contains(listener)) {
			onLoadedListenerList.add(listener);
		}
	}

	/**
	 * tell the {@link DCMManager} that a previously registered listener should
	 * not be invoked when the XML information are loaded
	 *
	 * @param listener the listener which should not be invoked any more
	 * @return true if the listener was registered before, false otherwise
	 */
	public boolean unregisterOnLoadedListener(OnLoadedListener_I listener) {
		return onLoadedListenerList.remove(listener);
	}

	/**
	 * retrieve a DCM by its name
	 *
	 * @param name the name of the DCM
	 * @return the DCM if found or null if the DCM could not be found
	 */
	public AbstractDCM getDCMByName(String name) {
		for (AbstractDCM dcm : dcms) {
			if (dcm.getDCMName().equals(name)) {
				return dcm;
			}
		}
		return null;
	}

	/**
	 * retrieve a sorted list with the names of all DCMs
	 *
	 * @return list with the names of all DCMs
	 */
	public List<String> getSortedDCMNameList() {
		List<String> dcmNameList = new ArrayList<String>(dcms.size());
		for (AbstractDCM dcm : dcms) {
			dcmNameList.add(dcm.getDCMName());
		}
		Collections.sort(dcmNameList);
		return dcmNameList;
	}

}
