/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.dcm;

import de.tu_darmstadt.rs.sim.engine.RefreshListener_I;
import de.tu_darmstadt.rs.sim.engine.RefreshTrigger_I;
import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.PipelineClock;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * Class controlling the DCM output signals' order of invocation. E.g. if signal
 * A is twice as often clocked as signal B the
 * {@link DCMSignalInvocationManager} clocks first A and in a second step
 * signals A and B (assuming their order in the second step is not relevant).
 */
public class DCMSignalInvocationManager implements RefreshTrigger_I {

	/**
	 * reference to the simulation manager
	 */
	private SimulationManager simulationManager;

	/**
	 * list of all signal invokers
	 */
	private List<DCMSignalInvoker> signalInvokers;
	/**
	 * the maximum of microsteps
	 */
	private long maxMicroSteps;

	/**
	 * list representing the order of signals which can be invoked
	 */
	private List<DCMSignalInvocation> orderedSignalsToBeInvoked;
	/**
	 * length of the orderedSignalsToBeInvoked list
	 */
	private int listSize;
	/**
	 * index of the signal which was invoked last
	 */
	private int currentSignalIndex;
	/**
	 * refresh listeners
	 */
	protected List<RefreshListener_I> listenerList = new ArrayList<RefreshListener_I>();

	private final PipelineClock pipelineClock;

	private final List<Integer> dirtyAddresses;

	/**
	 * create a new {@link DCMSignalInvocationManager}
	 *
	 * @param simulationManager the {@link SimulationManager} required for further
	 * functionality
	 */
	public DCMSignalInvocationManager(SimulationManager simulationManager) {
		this.simulationManager = simulationManager;
		signalInvokers = new ArrayList<DCMSignalInvoker>();
		orderedSignalsToBeInvoked = new ArrayList<DCMSignalInvocation>();
		this.pipelineClock = simulationManager.getPipelineClock();
		dirtyAddresses = new ArrayList<Integer>();
		simulationManager.getRefreshHandler().registerTrigger(this);
	}

	/**
	 * simulate the DCM signals until a pipeline clock starts
	 */
	public void doClockingUntilStartOfNextPipelineClock() {
		if (listSize == 0) {
			// this should never happen
			pipelineClock.onClock(1);
			return;
		}
		// this loop is not endless if the pipeline is connected to a DCM signal
		while (true) {
			currentSignalIndex++;
			// if the end of the signal list is reached then start again
			if (currentSignalIndex >= listSize) {
				currentSignalIndex = 0;
			}
			DCMSignalInvocation signalInvocation = orderedSignalsToBeInvoked
					.get(currentSignalIndex);
			signalInvocation.fire();
			// stop clocking if the pipeline was clocked
			// so the simulation manager can do its stuff
			if (isCurrentCycleThePipelineClockCycle()) {
				return;
			}
		}
	}

	/**
	 * check if a pipeline clock cycle was started
	 *
	 * @return true if a pipeline clock cycle was started, false otherwise
	 */
	private boolean isCurrentCycleThePipelineClockCycle() {
		return simulationManager.getPipelineClock().wasClocked();
	}

	/**
	 * register a DCMSignalInvoker
	 *
	 * @param signalInvoker the DCMSignalInvoker
	 */
	public void registerDCMSignalInvoker(DCMSignalInvoker signalInvoker) {
		signalInvokers.add(signalInvoker);
		prepareClocking();
	}

	/**
	 * unregister a DCMSignalInvoker
	 *
	 * @param signalInvoker the DCMSignalInvoker
	 */
	public void unregisterDCMSignalInvoker(DCMSignalInvoker signalInvoker) {
		signalInvokers.remove(signalInvoker);
		prepareClocking();
	}

	/**
	 * prepare the signal emitting
	 */
	protected void prepareClocking() {
		long clockStepsPerPeriod = 1;
		long howManySteps = 1;
		long dcmPathClockStepsPerPeriod;
		long dcmPathHowOftenClocked;
		maxMicroSteps = 1;
		// compute the LCM of the denominators of the fractions
		for (DCMSignalInvoker signalInvoker : signalInvokers) {
			// we begin a new computation so restore values from before the
			// first change
			signalInvoker.restoreValues();
			dcmPathClockStepsPerPeriod = signalInvoker
					.getClockStepsInOnePeriod();
			clockStepsPerPeriod = getLCM(clockStepsPerPeriod,
					dcmPathClockStepsPerPeriod);
		}
		// distribute the LCM and calculate the number of steps
		for (DCMSignalInvoker signalInvoker : signalInvokers) {
			// change the values
			signalInvoker.adjustClockStepsInOnePeriod(clockStepsPerPeriod);
			dcmPathHowOftenClocked = signalInvoker
					.getHowOftenClockedPerPeriod();
			howManySteps = getLCM(howManySteps, dcmPathHowOftenClocked);
		}
		maxMicroSteps = 4 * howManySteps; // for the phase shifts
		// let every invoker do some preparation
		for (DCMSignalInvoker signalInvoker : signalInvokers) {
			signalInvoker.adjustToMaxCounterValue(maxMicroSteps);
		}
		// adjust the clock offset of every invoker - must be extra loop because
		// this may require the clock offset of previous components in the chain
		/**
		 * this loop is no longer valid for there are no phase shifts needed
		 */
		/*for (DCMSignalInvoker signalInvoker : signalInvokers) {
			signalInvoker.adjustClockOffset();
		}*/
		// we were assuming that we begin counting with 1, so normalize it
		maxMicroSteps--;
		prepareOrderedSignalInvocation();
	}

	/**
	 * prepare the list with ordered signal invocations
	 */
	private void prepareOrderedSignalInvocation() {
		DCMSignalInvocation signalInvocation = new DCMSignalInvocation();
		orderedSignalsToBeInvoked.clear();
		for (int i = 0; i <= maxMicroSteps; i++) {
			for (DCMSignalInvoker signalInvoker : signalInvokers) {
				if (signalInvoker.isActivatedAtStep(i)
						&& signalInvoker.canFire()) {
					signalInvocation.addSignalInvoker(signalInvoker);
				}
			}
			if (signalInvocation.canFire()) {
				orderedSignalsToBeInvoked.add(signalInvocation);
				signalInvocation = new DCMSignalInvocation();
			}
		}
		listSize = orderedSignalsToBeInvoked.size();
		fireRefreshEvent(0);
	}

	/**
	 * Adds a listener to the event queue.
	 *
	 * @param listener the listener
	 */
	public void addRefreshListener(final RefreshListener_I listener) {
		listenerList.add(listener);
	}

	/**
	 * Removes a listener from the event queue.
	 *
	 * @param listener the listener
	 */
	public void removeRefreshListener(final RefreshListener_I listener) {
		listenerList.remove(listener);
	}

	/**
	 * Will be invoked after the component's state has changed. All registered
	 * listeners will be called to refresh() -- unless Settings.REFRESH_ENABLE
	 * is set to false.
	 *
	 * @param address the values's address which has changed, or COMPLETE_REFRESH if
	 * no differentiation is made
	 */
	protected void fireRefreshEvent(final int address) {

		// Note: address may be COMPLETE_REFRESH
		dirtyAddresses.add(address);
	}

	/**
	 * get the names of all invoked signals during one cycle
	 *
	 * @return list with names of all invoked signals during one cycle
	 */
	public List<String> getOrderOfInvokedSignals() {
		List<String> signalList = new ArrayList<String>();
		for (DCMSignalInvocation signalInvocation : orderedSignalsToBeInvoked) {
			signalList.add(signalInvocation.toString());
		}
		return signalList;
	}

	/**
	 * compute the Least Common Multiple of a and b
	 *
	 * @param a
	 * @param b
	 * @return
	 */
	private long getLCM(long a, long b) {
		return (a * b) / getGCD(a, b);
	}

	/**
	 * compute the Greatest Common Divisor of n and m
	 *
	 * @param n
	 * @param m
	 * @return
	 */
	private long getGCD(final long n, final long m) {
		long a = n;
		long b = m;
		long new_m = m;
		// n must be greater than m so swap if necessary
		if (m > n) {
			long save = n;
			a = m;
			b = save;
		}
		while (b != 0) {
			new_m = a % b;
			a = b;
			b = new_m;
		}
		return a;
	}

	@Override
	public Collection<RefreshListener_I> getRefreshListeners() {
		return listenerList;
	}

	@Override
	public int[] getDirtyAddresses() {
		int size = dirtyAddresses.size();
		int[] dirty = new int[size];
		for (int i = 0; i < size; i++) {
			dirty[i] = dirtyAddresses.get(i);
		}
		return dirty;
	}
}
