/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.dcm;

/**
 * Class for the calculation of a single signal invocation.
 */
public class DCMSignalInvoker {

	private long howOftenClockedPerPeriod;
	private long clockStepsInOnePeriod;
	//	private QuadrantPhaseShift shift;
	private AbstractDCMSignal signal;

	private long moduloStep;
	/**
	 * clock offset is no longer needed in the new jconfig since the phase shift is never used
	 */
	private long clkOffset;

	/**
	 * create a DCMSignalInvoker
	 *
	 * @param signal the signal that should be invoked
	 */
	public DCMSignalInvoker(AbstractDCMSignal signal) {
		this.signal = signal;
		restoreValues();
	}

	/**
	 * is the signal invoked at the microstep?
	 *
	 * @param stepValue the index of the microstep
	 * @return true if invoked, false otherwise
	 */
	public boolean isActivatedAtStep(long stepValue) {
		return (stepValue % moduloStep) == (clkOffset % moduloStep);
	}

	/**
	 * clean up values which could have been changed during preparation phase
	 */
	public void restoreValues() {
		howOftenClockedPerPeriod = signal.getHowOftenClockedPerPeriod();
		clockStepsInOnePeriod = signal.getClockStepsInOnePeriod();
		//		shift=signal.getPhaseShift();
		//		clkOffset=shift.getValue()*signal.getClockStepsInOnePeriod()>>2;
	}

	/**
	 * get the number of clock steps performed in one period
	 *
	 * @return clock steps
	 */
	public long getClockStepsInOnePeriod() {
		return clockStepsInOnePeriod * getClockStepsOfDCMPath();
	}

	/**
	 * check how often the {@link DCMSignalInvoker} is clocked in a period
	 *
	 * @return how often is this element clocked
	 */
	public long getHowOftenClockedPerPeriod() {
		return howOftenClockedPerPeriod * getHowOftenClockedInDCMPath();
	}

	/**
	 * retrieve the clock offset given by the phase shifts
	 *
	 * @return offset in microsteps
	 */
	/*public long getClkOffset(){
		return clkOffset+getClkOffsetInPath();
	}

	/**
	 * adjust the maximum number of steps in the signals period
	 * @param newNumberOfClockStepsInOnePeriod value to be used as the new maximum number of clock steps
	 */
	public void adjustClockStepsInOnePeriod(long newNumberOfClockStepsInOnePeriod) {
		long factor;
		factor = newNumberOfClockStepsInOnePeriod / getClockStepsInOnePeriod();
		multiplyBy(factor);
	}

	/**
	 * multiply all used microstep values with the factor
	 *
	 * @param factor the factor used
	 */
	private void multiplyBy(long factor) {
		howOftenClockedPerPeriod *= factor;
		clockStepsInOnePeriod *= factor;
		//clkOffset*=factor ;
	}

	/**
	 * retrieve the DCMSignal which can be invoked by the DCMSignalInvoker
	 *
	 * @return the DCMSignal
	 */
	public AbstractDCMSignal getDCMSignal() {
		return signal;
	}

	/**
	 * adjust to the maximum number of microsteps
	 *
	 * @param counterValue the new maximum
	 */
	public void adjustToMaxCounterValue(long counterValue) {
		moduloStep = counterValue / howOftenClockedPerPeriod;
		//		int phaseShiftValue=shift.getValue();
		//		clkOffset=(moduloStep>>2)*phaseShiftValue;
	}

	/**
	 * adjust the clock offset
	 */
	/*public void adjustClockOffset(){
		clkOffset=getClkOffset();
	}*/
	private long getClockStepsOfDCMPath() {
		final AbstractDCMSignal signal = getDCMSignal();
		long retVal = 1;
		AbstractDCMSignal previousSignal = signal.getDcm().getInputSignal();
		if (previousSignal != null) {
			retVal = previousSignal.getClockStepsInOnePeriod();
		}
		return retVal;
	}


	private long getHowOftenClockedInDCMPath() {
		final AbstractDCMSignal signal = getDCMSignal();
		long retVal = 1;
		AbstractDCMSignal previousSignal = signal.getDcm().getInputSignal();
		if (previousSignal != null) {
			retVal = previousSignal.getHowOftenClockedPerPeriod();
		}
		return retVal;
	}

	/*private long getClkOffsetInPath(){
		final AbstractDCMSignal signal=getDCMSignal();
		long retVal=0;
		AbstractDCMSignal previousSignal=signal.getDcm().getInputSignal();
		if(previousSignal!=null){
			retVal=previousSignal.getDCMSignalInvoker().getPhaseValue();
		}
		return retVal;
	}

	/*private long getPhaseValue(){
		long retVal=shift.getValue();
		retVal*=getClockStepsInOnePeriod()>>2;
		return retVal;
	}*/

	/**
	 * can the DCMSignalInvoker emit the signal?
	 *
	 * @return true if the signal has observers, false otherwise
	 */
	public boolean canFire() {
		return signal.hasTimerObservers();
	}

	/**
	 * emit the signal
	 */
	public void fire() {
		signal.informAllObservers();
	}
}
