/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
*/

package de.tu_darmstadt.rs.sim.engine.net;

import de.tu_darmstadt.rs.sim.engine.MessageHandler_I;
import de.tu_darmstadt.rs.sim.engine.SimControlSwitcher_I;
import de.tu_darmstadt.rs.sim.engine.SimulationComponent_I;
import de.tu_darmstadt.rs.sim.engine.SimulationController_I;
import de.tu_darmstadt.rs.sim.engine.SimulationStepListener_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.mpsoc.MasterConnector;
import de.tu_darmstadt.rs.sim.engine.peripheral.mpsoc.MpsocConnector;
import de.tu_darmstadt.rs.sim.engine.peripheral.mpsoc.SlaveConnector;

import libsim.Condition;
import libsim.Message;
import libsim.MessageRFifo;
import libsim.MessageState;
import libsim.MessageType;
import libsim.MessageUFifo;
import libsim.Payload;
import libsim.SimulationSpeed_I.Speed;
import libsim.State;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.UnknownHostException;
import java.nio.channels.SocketChannel;
import java.nio.channels.UnresolvedAddressException;
import java.util.HashMap;
import java.util.Map;

public class MpsocManager implements MpsocMessageHandler_I, SimulationController_I, SimulationStepListener_I, SimulationComponent_I {

	public enum ConnState {
		DISABLED,
		OFFLINE,
		CONNECTED,
		RESET,
		ONLINE,
		SYNCHRONIZED
	}


	private String fHost;
	private int fPort;
	private SocketChannel fSocketChannel;
	private final Map<Integer, MpsocConnector> fConnectors;
	private String fStatusMsg;
	private String fCoreId;
	private ConnState fConnState;
	private final SimControlSwitcher_I fSimControlSwitcher;
	private final SimulationController_I fLocalSimControl;
	private final MessageHandler_I fMsgHandler;
	private int fRxSeq;
	//    private int fTxSeq = (int)System.currentTimeMillis();
	private NetworkThread fNetworkThread;
	private final Object fWaitResponse;
	private volatile boolean fResponsePending;

	public MpsocManager(String host, int port, SimControlSwitcher_I sw, SimulationController_I lsc, MessageHandler_I mh) {
		fHost = host;
		fPort = port;
		fConnState = ConnState.OFFLINE;
		fSimControlSwitcher = sw;
		fLocalSimControl = lsc;
		fMsgHandler = mh;
		fWaitResponse = new Object();
		fConnectors = new HashMap<Integer, MpsocConnector>();
	}

	private void connect() {
		if (!fConnState.equals(ConnState.OFFLINE)) {
			return;
		}
		System.out.printf("  connecting\n");
		Exception thrownException = null;
		try {
			fSocketChannel = SocketChannel.open(new InetSocketAddress(fHost, fPort));
			fSocketChannel.configureBlocking(false);
		} catch (UnresolvedAddressException e) {
			thrownException = e;
		} catch (UnknownHostException e) {
			thrownException = e;
		} catch (IOException e) {
			thrownException = e;
		} finally {
			if (thrownException != null) {
				fStatusMsg = String.format(
						"Not connected: %s",
						fHost,
						fPort,
						thrownException.getMessage()
				);
				fConnState = ConnState.OFFLINE;
			} else {
				fStatusMsg = String.format(
						"Connected to %s:%d",
						fHost,
						fPort
				);
				fConnState = ConnState.CONNECTED;
			}
			System.out.printf("%s\n", fStatusMsg);
		}
	}

	private boolean sendMsg(MessageType mt, int dst, Object data) {
		if (fConnState.compareTo(ConnState.CONNECTED) < 0) {
			return false;
		}
		char[] convert = fCoreId.toCharArray();
		int ID = 0;
		for (int i = 0; i < convert.length; i++) {
			ID = +(int) convert[i];
		}
		Payload p = new Payload(fRxSeq, dst, ID);
		if (data != null) {
			p.add(data);
		}
		fNetworkThread.enqueueSend(mt.construct(p));
		return true;
	}

	private void initialize(ConnState state) {
		if (!fConnState.equals(ConnState.CONNECTED)) {
			return;
		}
		System.out.printf("  initializing\n");
		fNetworkThread = new NetworkThread(this, fSocketChannel);
		if (sendMsg(MessageType.COND, 0, Condition.INIT)) {
			fConnState = state;
		} else {
			fNetworkThread = null;
		}
	}

	private void recvFifoUpdate(MessageUFifo msg) throws MpsocException {
		MpsocConnector conn = fConnectors.get(msg.src());
		if (conn == null) {
			throw new MpsocException("No multicore connector peripheral at base address 0x%05X", msg.src());
		}
		if (fConnState.equals(ConnState.SYNCHRONIZED)) {
			synchronized (fWaitResponse) {
				fResponsePending = false;
				fWaitResponse.notify();
			}
		}
		conn.onRecvUpdate(msg.data());
	}

	public void setConnState(final ConnState state) {
		if (fConnState.equals(state)) {
			return;
		}
		System.out.printf("trying state: %s\n", state);
		if (state.compareTo(ConnState.OFFLINE) <= 0) {
			fConnState = state;
		}
		if (state.compareTo(ConnState.CONNECTED) >= 0) {
			connect();
		}
		if (state.compareTo(ConnState.ONLINE) >= 0) {
			initialize(state);
		}
		if (fConnState.equals(ConnState.SYNCHRONIZED)) {
			fSimControlSwitcher.switchToSimController(this);
		} else {
			fSimControlSwitcher.switchToSimController(fLocalSimControl);
			fConnState = state;
		}
		System.out.printf("state: %s\n", fConnState);
	}

	public synchronized void sendFifoUpdate(final int base, final int reqSize) throws MpsocException {
		MpsocConnector conn = fConnectors.get(base);
		if (conn == null) {
			throw new MpsocException("No multicore connector peripheral at base address 0x%05X", base);
		}
		sendMsg(MessageType.U_FIFO, base, conn.getEntriesPending(reqSize));

	}

	public void requestFifoUpdate(final MpsocConnector conn) {
		synchronized (fWaitResponse) {
			if (!sendMsg(MessageType.R_FIFO, conn.getBaseAddr(), conn.getRequestSize())) {
				return;
			}
			if (!fConnState.equals(ConnState.SYNCHRONIZED)) {
				return;
			}
			try {
				fResponsePending = true;
				fWaitResponse.wait(3000);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}

	public ConnState getConnState() {
		return fConnState;
	}

	public void registerMaster(final MasterConnector m) {
		fConnectors.put(m.getBaseAddr(), m);
	}

	public void registerSlave(final SlaveConnector s) {
		fConnectors.put(s.getBaseAddr(), s);
	}

	public void unregister(final MpsocConnector conn) {
		fConnectors.remove(conn.getBaseAddr());
	}

	public void setCoreId(final String id) {
		fCoreId = id;
	}

	@Override
	public void onMessageReceived(final Message msg) {
		if (msg == null) {
			setConnState(ConnState.OFFLINE);
			return;
		}
		try {
			switch (msg.type()) {
				case STEP:
					//                    System.out.printf("executing step\n");
					if (fRxSeq != msg.seq()) {
						fRxSeq = msg.seq();
						if (fConnState.equals(ConnState.SYNCHRONIZED)) {
							fLocalSimControl.stepSimulation();
						} else {
							sendMsg(MessageType.COND, 0, Condition.NUPD);
						}
					} else {
						sendMsg(MessageType.COND, 0, Condition.NUPD);
					}
					break;
				case STATE:
					State st = ((MessageState) msg).state();
					if (st.equals(State.RESET)) {
						fLocalSimControl.resetSimulation();
						synchronized (fWaitResponse) {
							fResponsePending = false;
							fWaitResponse.notify();
						}
						sendMsg(MessageType.COND, 0, Condition.CLEAR);
					}
					if (fConnState.equals(ConnState.SYNCHRONIZED)) {
						break;
					}
					switch (st) {
						case RUNNING:
							fLocalSimControl.runSimulation();
							break;
						case STOPPED:
							fLocalSimControl.stopSimulation();
							break;
					}
					break;
				case R_FIFO:
					fRxSeq = msg.seq();
					//                    try {
					//                        System.out.printf("zZzZzZzzz ...\n");
					//                        Thread.sleep(5000);
					//                    } catch (InterruptedException e) {
					//                        e.printStackTrace();
					//                    }
					sendFifoUpdate(msg.src(), ((MessageRFifo) msg).length());
					break;
				case U_FIFO:
					recvFifoUpdate((MessageUFifo) msg);
					break;
			}
		} catch (MpsocException e) {
			fLocalSimControl.stopSimulation();
			fMsgHandler.postMessage(e.getMessage(), MessageHandler_I.MsgType.ERR);
		}
	}

	@Override
	public void runSimulation() {
		sendMsg(MessageType.COND, 0, Condition.CONT);
	}

	@Override
	public void stepSimulation() {
		sendMsg(MessageType.COND, 0, Condition.STEP);
	}

	@Override
	public void stopSimulation() {
		sendMsg(MessageType.COND, 0, Condition.STOP);
	}

	@Override
	public void resetSimulation() {
		sendMsg(MessageType.COND, 0, Condition.RESET);
	}

	@Override
	public void simulationStepComplete(final Result r) {
		if (fConnState.equals(ConnState.SYNCHRONIZED)) {
			if (fResponsePending) {
				sendMsg(MessageType.COND, 0, Condition.EUPD);
			}
			sendMsg(MessageType.COND, 0, Condition.NUPD);
		}
	}

	@Override
	public void restart() {
		setConnState(MpsocManager.ConnState.ONLINE);
	}

	@Override
	public void setup() {
		// nothing to do here
	}

	@Override
	public void setSimulationSpeed(final Speed spd) {
		//nothing to do here
	}
}
