/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
*/

package de.tu_darmstadt.rs.sim.engine.net;

import libsim.Message;
import libsim.MessageType;
import libsim.UnknownMessage;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.SocketChannel;
import java.util.Queue;
import java.util.concurrent.ConcurrentLinkedQueue;

/**
 * Class for the sending/receiving messages
 */
public class NetworkThread extends Thread {

	private final SocketChannel fSocketChannel;
	private final MpsocMessageHandler_I fMsgHandler;
	private Selector fSel;
	private SelectionKey fSelKey;
	private ByteBuffer fBytesTx;
	private Queue<Message> fTx;
	private ByteBuffer[] fBytesRx;

	/**
	 * create a new {@link NetworkThread}
	 *
	 * @param mh the handler for the received messages
	 * @param sock connection to a socket
	 */
	public NetworkThread(MpsocMessageHandler_I mh, SocketChannel sock) {
		super("Mpsoc Network Thread");
		fMsgHandler = mh;
		fSocketChannel = sock;
		fTx = new ConcurrentLinkedQueue<Message>();
		try {
			fSel = Selector.open();
			fSelKey = fSocketChannel.register(fSel, SelectionKey.OP_READ);
			start();
		} catch (ClosedChannelException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
		setPriority(Thread.MAX_PRIORITY);
	}

	@Override
	public void run() {
		try {
			while (true) {
				fSel.select();
				if (fSelKey.isReadable()) {
					receive();
				}
				if (fSelKey.isWritable()) {
					send();
				}
			}
		} catch (ClosedChannelException e) {
			/* do nothing, just terminate */
		} catch (IOException e) {
			e.printStackTrace();
		} finally {
			try {
				if (fSel != null) {
					fSel.close();
				}
				fSocketChannel.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
			System.out.printf("disconnected\n");
			fMsgHandler.onMessageReceived(null);
		}
	}

	private void receive() throws IOException {
		if (fBytesRx == null) {
			fBytesRx = MessageType.allocate_msg_buf();
		}
		try {
			Message msg = MessageType.deserialize(fSocketChannel, fBytesRx);
			if (msg != null) {
				//                if (msg.type().equals(MessageType.R_FIFO) || msg.type().equals(MessageType.U_FIFO)) {
				System.out.printf("RX: %s\n", msg.format_msg());
				//                }
				fBytesRx = null;
				fMsgHandler.onMessageReceived(msg);
			}
		} catch (UnknownMessage e) {
			e.printStackTrace();
		}
	}

	private void send() throws IOException {
		if (fBytesTx == null) {
			Message msg = fTx.poll();
			if (msg == null) {
				fSelKey.interestOps(SelectionKey.OP_READ);
				return;
			}
			fBytesTx = msg.serialize();
			//            if (msg.type().equals(MessageType.R_FIFO) || msg.type().equals(MessageType.U_FIFO)) {
			System.out.printf("TX: %s\n", msg.format_msg());
			//            }
		}
		fSocketChannel.write(fBytesTx);
		if (fBytesTx.remaining() == 0) {
			fBytesTx = null;
		}
	}

	/**
	 * enqueue a message in the send queue.
	 *
	 * @param msg the message to be sent
	 */
	public synchronized void enqueueSend(Message msg) {
		//        if (msg.type().equals(MessageType.R_FIFO) || msg.type().equals(MessageType.U_FIFO)) {
		//            System.out.printf("TX (pending): %s\n", msg.format_msg());
		//        }
		fTx.offer(msg);
		fSelKey.interestOps(SelectionKey.OP_READ | SelectionKey.OP_WRITE);
		fSel.wakeup();
	}

}
