/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.parser;


import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;

/**
 * Abstract base class for classes which are used to read a .lst file
 */
public abstract class AbstractLstLineMatcher implements SpartanDefinitions_I {

	/**
	 * the current line as a string
	 */
	protected String line;

	/**
	 * Checks whether this object thinks it can handle the .lst line. This
	 * method also tries to find the information for later use.
	 *
	 * @param line the listing's line
	 * @return true if it can handle the line, false otherwise
	 */
	public boolean matches(final String line) {
		return false;
	}

	/**
	 * Gets the address (in data addressing mode) where the data should be
	 * stored in memory.
	 *
	 * @return data address
	 */
	public int getAddress() {
		return INVALID;
	}

	/**
	 * Checks whether this matcher holds a memory value if it has matched.
	 *
	 * @return true if a value should be stored in memory, false otherwise
	 */
	public boolean hasCode() {
		return false;
	}

	/**
	 * Checks whether this matcher can only hold 9 bit memory values.
	 *
	 * @return true if a value to be stored is a 9 bit value, false if it is 18
	 * bit
	 */
	public boolean isHalfCode() {
		return false;
	}

	/**
	 * Gets the value to be stored in memory.
	 *
	 * @return the value to be stored in memory (This may be an INVALID value.)
	 */
	public int getCode() {
		return INVALID;
	}

	/**
	 * Checks whether this matcher holds data or an instruction.
	 *
	 * @return true if the matcher holds data, false if it holds an instruction
	 */
	public boolean createsData() {
		return false;
	}

	/**
	 * Gets the line in an transformed representation (as an array of strings).
	 *
	 * @return the transformed representation
	 */
	public abstract String[] getTransformedLineData();

	/**
	 * Helper method for parsing.<br/>
	 * Gets the integer value of a string which contains exactly 1 byte in hex.
	 *
	 * @param hexStr the string which contains exactly 1 byte
	 * @return the integer value of the byte
	 */
	protected int getElementalHex(final String hexStr) {
		return Integer.parseInt(hexStr, 16) & 0xff;
	}

	/**
	 * Helper method for parsing.<br/>
	 * Gets the integer value of a string which contains exactly an address.
	 *
	 * @param addressString the string containing the address
	 * @return the integer value of the address
	 */
	protected int getAddressFromString(final String addressString) {
		return Integer.parseInt(addressString, 16);
	}
}
