/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.parser;

import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.assembly.AssemblyLineDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.assembly.AssemblyLineProvider_I;
import de.tu_darmstadt.rs.sim.engine.assembly.AssemblyLine_I;
import de.tu_darmstadt.rs.sim.engine.assembly.AssemblyVariablesManager;
import de.tu_darmstadt.rs.sim.gui.assembly.AssemblyView;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;


/**
 * Class for parsing listing files (.lst) (useful for program symbols in the
 * assembly window)
 */
public class LstFileParser implements SpartanDefinitions_I,
		AssemblyLineDefinitions_I, AssemblyLineProvider_I {

	private final AddressSpaceManager memory;
	private static LstFileParser theInstance = null;

	private List<AbstractLstLineMatcher> lineMatcher;
	private List<LstAssemblyLine> assemblyLineList;
	private AssemblyVariablesManager variablesManager;
	private int lastDefinedAddress;
	private int lastDefinedAddressLineIndex;

	private LstFileParser(final AddressSpaceManager memory) {
		this.memory = memory;
		lineMatcher = new ArrayList<AbstractLstLineMatcher>();
		variablesManager = new AssemblyVariablesManager(memory);
		lineMatcher.add(new LstProgramLineMatcher(variablesManager));
		lineMatcher.add(new LstDataLineMatcher(variablesManager));
		lineMatcher.add(new LstSectionLineMatcher(variablesManager));
		lineMatcher.add(new LstSymbolLineMatcher(variablesManager));
		lineMatcher.add(new LstAnythingMatcher(variablesManager));
		assemblyLineList = new ArrayList<LstAssemblyLine>();
	}

	/**
	 * load a listing file and parse it
	 *
	 * @param file the file
	 * @throws ParsingException an exception will be thrown when the file cannot be parsed
	 */
	public void loadLstFile(File file) throws ParsingException {
		List<String> listOfLines = readLstFile(file);
		parseListOfLines(listOfLines);
	}

	/**
	 * read a file and create a list with lines
	 *
	 * @param file the file to read
	 * @return a list of the file's lines
	 *
	 * @throws ParsingException an exception will be thrown when the file cannot be parsed
	 */
	private List<String> readLstFile(File file) throws ParsingException {
		List<String> listOfLines = new ArrayList<String>();
		BufferedReader reader = null;
		try {
			reader = new BufferedReader(new FileReader(file));

			String line;
			while ((line = reader.readLine()) != null) {
				listOfLines.add(line);
			}
		} catch (Exception e) {
			throw new ParsingException(e.getMessage());
		} finally {
			if (reader != null) {
				try {
					reader.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}
		}
		return listOfLines;
	}

	/**
	 * find the information within the lines
	 *
	 * @param listOfLines the lines of a .lst file
	 */
	private void parseListOfLines(List<String> listOfLines) {
		LstAssemblyLine assemblyLine;
		assemblyLineList.clear();
		// for each line
		for (String line : listOfLines) {
			// try for each matcher ...
			for (AbstractLstLineMatcher matcher : lineMatcher) {
				// ... if the matcher can handle the line ...
				if (matcher.matches(line)) {
					// ... if so, read the information ...
					assemblyLine = new LstAssemblyLine();
					assemblyLine.dataAddress = matcher.getAddress();
					assemblyLine.code = matcher.getCode();
					assemblyLine.data = matcher.getTransformedLineData();
					if (matcher.createsData()) {
						assemblyLine.isData = true;
					}
					// ... remember the line ...
					addLineObject(assemblyLine);
					// ... and store the value (if existing) in memory.
					if (matcher.hasCode()) {
						setValueInMemory(assemblyLine.dataAddress,
								assemblyLine.code, matcher.isHalfCode());
					}
					// Stop on the first match, don't try to find other
					// matchers.
					break;
				}
			}
		}
	}

	/**
	 * Remember an assembly line for later being able to show it via
	 * {@link AssemblyView}.
	 *
	 * @param assemblyLine the assembly line to remember
	 */
	private void addLineObject(LstAssemblyLine assemblyLine) {
		if (assemblyLine.getDataAddress() != INVALID) {
			lastDefinedAddress = assemblyLine.getDataAddress();
			lastDefinedAddressLineIndex = assemblyLineList.size();
		}
		assemblyLineList.add(assemblyLine);
	}

	/**
	 * Store the 18 bit value which was read from the assembly line in memory.
	 *
	 * @param dataAddress the address where the value should be stored
	 * @param value the value to be stored
	 */
	private void setValueInMemory(int dataAddress, int value) {
		if (value != INVALID) {
			memory.setValue(dataAddress >> 1, value);
		}
	}

	/**
	 * Store a value which was read from the assembly line in memory.
	 *
	 * @param dataAddress the address where the value should be stored
	 * @param value the value to be stored
	 * @param isHalfWord if true the value to be stored is a 9 bit value, otherwise an
	 * 18 bit value
	 */
	private void setValueInMemory(int dataAddress, int value, boolean isHalfWord) {
		if (isHalfWord) {
			int shiftedValue = value;
			int mask = BITMASK_9 << 9;
			if (dataAddress % 2 == 0) {
				shiftedValue <<= 9;
				mask = BITMASK_9;
			}
			int memValue = memory.getValue(dataAddress >> 1);
			memValue &= mask;
			memValue += shiftedValue;
			memory.setValue(dataAddress >> 1, memValue);
		} else {
			setValueInMemory(dataAddress, value);
		}
	}

	/**
	 * Returns the single instance.
	 *
	 * @return the instance
	 */
	public static LstFileParser getTheInstance() {
		return theInstance;
	}

	/**
	 * Create a single instance. This method is not thread safe.
	 *
	 * @param memory the memory to store values which have been read from the file
	 */
	public static void createTheInstance(final AddressSpaceManager memory) {
		if (theInstance == null) {
			theInstance = new LstFileParser(memory);
		}
	}

	@Override
	public AssemblyLine_I getAssemblyLine(int rowIndex) {
		return assemblyLineList.get(rowIndex);
	}

	@Override
	public int getNumberOfLines() {
		return assemblyLineList.size();
	}

	@Override
	public int getLastDefinedAddress() {
		return lastDefinedAddress;
	}

	@Override
	public int getIndexOfLineWithLastDefinedAddress() {
		return lastDefinedAddressLineIndex;
	}

	@Override
	public String getProviderID() {
		return "Code From Listing";
	}

	/**
	 * Get the manager of the assembly's variables.
	 *
	 * @return the instance of the {@link AssemblyVariablesManager}
	 */
	public AssemblyVariablesManager getAssemblyVariablesManager() {
		return variablesManager;
	}
}
