/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

package de.tu_darmstadt.rs.sim.engine.peripheral;

import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.PipelineClock;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerObserver_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerSource_I;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;

import java.io.IOException;
import java.io.InputStream;

/**
 * This class represents a hardware FIFO for input.
 */
public class FifoIn extends JConfigPeripheral implements TimerObserver_I {

	private static final int REG_DATA_READ = 0;
	private static final int REG_STATUS = 1;

	private InputStream fInputStream;
	private int fDataReadValue;
	private boolean fDataWasRead;
	private TimerSource_I timerSource;

	/**
	 * create a new {@link FifoIn}
	 *
	 * @param peri the hardware description
	 * @param simulationManager reference to the {@link SimulationManager}
	 */
	public FifoIn(final IPeripheralModuleConfiguration peri,
				  final SimulationManager simulationManager, final int baseAddress) {
		super(peri, simulationManager, baseAddress);
		TimerManager
				.addTimerConnection(PipelineClock.PIPELINE_CLOCK_NAME, this);
	}

	private void readNext() {
		try {
			int v;
			if ((fInputStream != null) && (fInputStream.available() > 0)) {
				v = fInputStream.read();
			} else {
				v = 0;
			}
			if (v != fDataReadValue) {
				fDataReadValue = v;
				fireRefreshEvent(REG_DATA_READ);
			}
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	@Override
	public int getRegisterValue(int offset) {
		switch (offset) {
			case REG_DATA_READ:
				return fDataReadValue;
			case REG_STATUS:
				if (fInputStream == null) {
					return 0;
				}
				try {
					if (fInputStream.available() > 0) {
						return 1;
					}
				} catch (IOException e) {
					e.printStackTrace();
				}
				return 0;
			default:
				return 0;
		}
	}

	@Override
	protected int readRegisterValue(int offset) {
		int unchangedRegisterValue = getRegisterValue(offset);
		switch (offset) {
			case REG_DATA_READ:
				if (!fDataWasRead) {
					readNext();
					fDataWasRead = true;
				}
				return fDataReadValue;
			default:
				return unchangedRegisterValue;
		}
	}

	@Override
	public void setRegisterValue(int offset, int value) {
		//no writable registers available
	}

	@Override
	public boolean setInputStream(InputStream stream) {
		fInputStream = stream;
		fireRefreshEvent(REG_STATUS);
		return true;
	}

	@Override
	public void onClock(int clockValue) {
		fDataWasRead = false;
	}

	@Override
	public void setTimerSource(TimerSource_I timerSource) {
		this.timerSource = timerSource;
		timerSource.registerTimerObserver(this);
	}

	@Override
	public void tearDown() {
		super.tearDown();
		timerSource.unregisterTimerObserver(this);
	}

}
