/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

package de.tu_darmstadt.rs.sim.engine.peripheral;

import de.tu_darmstadt.rs.sim.engine.AbstractHardwareComponent;
import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.RefreshHandler_I;
import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.LinkedList;
import java.util.List;


/**
 * Abstract base class for the simulation of one peripheral. Provides access to
 * meta information from jconfig and implements some default behavior.
 */
public abstract class JConfigPeripheral extends AbstractHardwareComponent
		implements RegisterAccess_I {

	/**
	 * invalid id indicating that a value should not be used (e.g. method
	 * results or parameters)
	 */
	public static final String INVALID_ID = "invalid";

	/**
	 * the list of all listeners which must be informed when this object is
	 * teared down
	 */
	protected List<TearDownListener_I> listeners = new LinkedList<TearDownListener_I>();
	/**
	 * id of this peripheral, by default {@link #INVALID_ID}
	 */
	protected String id = INVALID_ID;

	private final String fPeriName;
	private final String fPeriType;
	private final int fBaseAddr;
	private final AddressSpaceManager fAddressSpace;
	private final PeripheralModuleDescription.Register[] fRegisters;
	private final SimulationManager fSimulationManager;

	/**
	 * retrieve objects which contain meta data about registers (e.g. name,
	 * access, ...)
	 *
	 * @return a register object
	 */
	public final PeripheralModuleDescription.Register[] getRegisters() {
		return fRegisters;
	}

	/**
	 * create a {@link JConfigPeripheral}
	 *
	 * @param peri the hardware description
	 * @param simulationManager reference to the {@link SimulationManager}
	 */
	public JConfigPeripheral(IPeripheralModuleConfiguration peri,
							 final SimulationManager simulationManager, int address) {
		this(peri.getName(), peri.getModuleDescriptionName(), address, simulationManager.getAddressSpaceManager(),
				simulationManager.getRefreshHandler(), simulationManager, peri.getModuleDescription().getRegisters().getRegisters());
	}

	/**
	 * create a {@link JConfigPeripheral} without using information from the
	 * jconfig environment. It is recommended to use the constructor
	 * {@link #JConfigPeripheral(IPeripheralModuleConfiguration, SimulationManager, int)}
	 * instead of this constructor.
	 *
	 * @param periName the name of the peripheral
	 * @param periType the peripheral type as string
	 * @param baseAddress the base address of the peripheral
	 * @param addressSpace reference to the {@link AddressSpaceManager}
	 * @param refreshHandler a {@link RefreshHandler_I} to which refresh requests should be
	 * sent
	 * @param simulationManager reference to the {@link SimulationManager}
	 * @param registers the definition of all the registers of the peripheral
	 */
	public JConfigPeripheral(final String periName, final String periType,
							 final int baseAddress, AddressSpaceManager addressSpace,
							 RefreshHandler_I refreshHandler,
							 final SimulationManager simulationManager,
							 final List<PeripheralModuleDescription.Register> registers) {
		super(refreshHandler);
		fPeriName = periName;
		fPeriType = periType;
		fBaseAddr = baseAddress;
		fRegisters = new PeripheralModuleDescription.Register[registers.size()];
		registers.toArray(fRegisters);
		fAddressSpace = addressSpace;
		fSimulationManager = simulationManager;
		registerIoAddresses();

	}

	public JConfigPeripheral(final String periName, final String periType,
							 final int baseAddress, AddressSpaceManager addressSpace,
							 RefreshHandler_I refreshHandler,
							 final SimulationManager simulationManager,
							 final PeripheralModuleDescription.Register[] registers) {
		super(refreshHandler);
		fPeriName = periName;
		fPeriType = periType;
		fBaseAddr = baseAddress;
		fRegisters = registers;
		fAddressSpace = addressSpace;
		fSimulationManager = simulationManager;
		registerIoAddresses();

	}

	/**
	 * Registers the address range covered by this peripheral in the address
	 * space.
	 */
	private void registerIoAddresses() {
		fAddressSpace.registerPeripheral(fBaseAddr, fRegisters.length, this);
	}

	/**
	 * Unregisters the address range covered by this peripheral in the address
	 * space.
	 */
	private void unregisterIoAddresses() {
		fAddressSpace.unregisterPeripheral(fBaseAddr, this);
	}

	private void validateRegisterOffset(int offset) {
		if (offset >= fRegisters.length) {
			fSimulationManager.pushErrorMessage(String.format(
					"%s: Register offset out of bounds (%d)\n", fPeriName,
					offset));
		}
	}

	/**
	 * Returns the peripheral type as specified by jConfig.
	 *
	 * @return the peripheral type
	 */
	public final String getPeriType() {
		return fPeriType;
	}

	/**
	 * Returns the peripheral name as specified by jConfig.
	 *
	 * @return the peripheral name
	 */
	public final String getPeriName() {
		return fPeriName;
	}

	/**
	 * Returns the peripherals base address as specified by jConfig.
	 *
	 * @return the peripheral name
	 */
	public final int getBaseAddr() {
		return fBaseAddr;
	}

	/**
	 * Set the parameter with parameter id the the given value.
	 *
	 * @param parId the parameter id.
	 * @param value the new value
	 * @return true if the parameter was set
	 */
	public boolean setParameter(int parId, int value) {
		return false;
	}

	/**
	 * Return all editable parameters of the peripheral.
	 *
	 * @return an array with the names of the parameters. The index in the array
	 * is equal to the parameter id. If the peripheral contains no
	 * parameters an empty array has to be returned. No null values
	 * allowed!
	 */
	public String[] getParameters() {
		return new String[0];
	}

	/**
	 * Will be called, whenever the simulation's settings are altered. Must
	 * check for inconsistencies and must re-register to the address space. Must
	 * also include all configurations made in restart(). May be called several
	 * times, e.g. after reconfigurations.
	 */
	public void setup() {
		//must be implemented in sub classes if needed
	}

	/**
	 * Will be called once if the peripheral is to be removed. Must unregister
	 * from the address space, inform all registered TearDownListeners and
	 * execute all necessary shut down procedures. All TearDownListener can be
	 * informed by calling informTearDownListeners. If once tear-down'ed, the
	 * peripheral unit will definitely never be re-used.
	 */
	public void tearDown() {
		unregisterIoAddresses();
		informTearDownListeners();
	}

	/**
	 * Will be called, whenever the simulation restarts.
	 */
	public void restart() {
		fireRefreshEvent(REFRESH_AND_CLEAR);
	}

	/**
	 * Performs a write access to the given IO address. Will be called by the
	 * address space. Must NOT be called by any (transparent) GUI, since there
	 * may be side-effects. Any write access will be validated against the
	 * address space of this peripheral (number of registers).
	 *
	 * @param address a valid IO address which the peripheral has registered to
	 * @param value the value to write
	 */
	public final void writeToWithSideEffects(final int address, final int value) {
		int offset = address - fBaseAddr;
		validateRegisterOffset(offset);
		if (fRegisters[offset].getAccess() == PeripheralModuleDescription.Register.AccessType.Read) {
			fSimulationManager.pushErrorMessage(String.format(
					"%s: Register %s (offset %d) is not writeable\n",
					fPeriName, getRegisters()[offset].getName(), offset));
		}
		writeRegisterValue(offset, value);
	}

	/**
	 * Performs a read access to the given IO address. Will be called by the
	 * address space. Must NOT be called by any (transparent) GUI, since there
	 * may be side-effects. Any read access will be validated against the
	 * address space of this peripheral (number of registers).
	 *
	 * @param address a valid IO address which the peripheral has registered to
	 * @return the value to be read
	 */
	public final int readFromWithSideEffects(final int address) {
		int offset = address - (int) fBaseAddr;
		validateRegisterOffset(offset);
		if (fRegisters[offset].getAccess() == PeripheralModuleDescription.Register.AccessType.Write) {
			fSimulationManager.pushErrorMessage(String.format(
					"%s: Register %s (offset %d) is not readable\n", fPeriName,
					getRegisters()[offset].getName(), offset));
		}
		return readRegisterValue(offset);
	}

	/**
	 * Set the hardware parameters for the peripheral.
	 *
	 * @param node peripheral module instance from libjconfig
	 */
	public void importHardwareParameters(final IPeripheralModuleConfiguration<?> node) {
		//must be implemented in sub classes if needed
	}

	/**
	 * Some peripheral have a bus like channel for communication with the
	 * environment. The peripheral will read input data from the supplied
	 * InputStream.
	 *
	 * @param stream the input stream
	 * @return true if the peripheral supports this feature.
	 */
	public boolean setInputStream(InputStream stream) {
		return false;
	}

	/**
	 * Some peripheral have a bus like channel for communication with the
	 * environment. The peripheral will write output data into the supplied
	 * OutputStream.
	 *
	 * @param stream the output stream
	 * @return true if the peripheral supports this feature.
	 */
	public boolean setOutputStream(OutputStream stream) {
		return false;
	}

	@Override
	public abstract int getRegisterValue(int offset);

	@Override
	public abstract void setRegisterValue(int offset, int value);

	/**
	 * The {@link #readRegisterValue(int)} method is used only(!) from the core
	 * and should not be used by any GUI method because some additional
	 * functions can take place here (e.g. resetting interrupts). The default
	 * implementation is the same as {@link #getRegisterValue(int)} which may be
	 * not intended by some sub class implementations (if so override this
	 * method).
	 *
	 * @param offset the offset of the register to the base address of the
	 * peripheral
	 * @return the value read
	 */
	protected int readRegisterValue(int offset) {
		return getRegisterValue(offset);
	}

	/**
	 * The {@link #writeRegisterValue(int, int)} method is used only(!) from the
	 * core and should not be used by any GUI method because some additional
	 * functions can take place here (e.g. resetting interrupts). The default
	 * implementation is the same as {@link #setRegisterValue(int, int)} which
	 * may be not intended by some sub class implementations (if so override
	 * this method).
	 *
	 * @param offset the offset of the register to the base address of the
	 * peripheral
	 * @param value the value to be written
	 */
	protected void writeRegisterValue(int offset, int value) {
		setRegisterValue(offset, value);
	}

	/**
	 * The {@link #getValueAt(int)} method should be used only by GUI methods.
	 * The default implementation is hence the same as
	 * {@link #getRegisterValue(int)}.
	 *
	 * @param defaultAddress the address where to get the value from
	 * @return the value at the address
	 */
	public int getValueAt(final int defaultAddress) {
		// TODO: check if getValueAt works for all JConfigPeripherals
		return getRegisterValue(defaultAddress - fBaseAddr);
	}

	/**
	 * The {@link #setValueAt(int, int)} method should be used only by GUI
	 * methods. The default implementation is hence the same as
	 * {@link #setRegisterValue(int, int)}.
	 *
	 * @param defaultAddress the address where to get the value from
	 * @param data the data to be set at the address
	 */
	public void setValueAt(final int defaultAddress, final int data) {
		// TODO: check if setValueAt works for all JConfigPeripherals
		setRegisterValue(defaultAddress - fBaseAddr, data);
	}

	/**
	 * Helper method: Sends the TearDownEvent to all registered
	 * TearDownListener.
	 */
	protected void informTearDownListeners() {
		TearDownListener_I[] temp = new TearDownListener_I[listeners.size()];
		temp = listeners.toArray(temp);

		// a for loop iterate directly over listeners does not allow concurrent
		// modifications
		for (TearDownListener_I l : temp) {
			l.performTearDownEvent(this);
		}
	}

	/**
	 * Register the TearDownListener. Receive the TearDownEvent when tearDown is
	 * called.
	 *
	 * @param listener the TearDownListener
	 * @return true if successfully registered
	 */
	public boolean registerTearDownListener(TearDownListener_I listener) {
		return listeners.add(listener);
	}

	/**
	 * Remove the TearDownListener from the notification list.
	 *
	 * @param listener the listener to be removed
	 * @return true if successfully removed
	 */
	public boolean removeTearDownListener(TearDownListener_I listener) {
		return listeners.remove(listener);
	}

	/**
	 * @return the number of registered TearDownListeners
	 */
	public int getTearDownListenerCount() {
		return listeners.size();
	}

	/**
	 * Sets the id for the peripheral. It have to be a unique id, as it is used
	 * for identify the peripheral.
	 *
	 * @param newId the new id of the peripheral
	 */
	public void setId(String newId) {
		id = newId;
	}

	/**
	 * Retrieves the unique id of the peripheral.
	 *
	 * @return the peripheral id
	 */
	public String getId() {
		return id;
	}

}
