/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 *
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2012
 *
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2012
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
*/

/**
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2012
 */
package de.tu_darmstadt.rs.sim.engine.peripheral;

import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;

import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * The peripheral class for the small MCB interface.
 * <p/>
 * <b>IMPORTANT NOTE:</b> the address variable in the fireRefreshEvent is the address for
 * the ddr memory. It is not the address for the mcb interface itself!
 *
 * Register should be updated when REFRESH_REGISTER_ONLY or COMPLETE_REFRESH is fired.
 */
public class McbInterfaceSmall extends JConfigPeripheral implements SharedMemory_I {

	private final static int MB = 1048576;

	private final static int MAX_DATA_FIFO_SIZE = 64;

	private final static Set<Integer> USED_DDR_IDS = new HashSet<Integer>();

	//register offsets
	/**
	 * offset of the COMMAND register (to the base address)
	 */
	public final static int COMMAND_OFFSET = 0;
	/**
	 * offset of the ADDR_LOW register (to the base address)
	 */
	public final static int ADDR_LOW_OFFSET = 1;
	/**
	 * offset of the ADDR_HIGH register (to the base address)
	 */
	public final static int ADDR_HIGH_OFFSET = 2;
	/**
	 * offset of the DATA_WRITE register (to the base address)
	 */
	public final static int DATA_WRITE_OFFSET = 3;
	/**
	 * offset of the DATA_READ register (to the base address)
	 */
	public final static int DATA_READ_OFFSET = 4;
	/**
	 * offset of the STATUS register (to the base address)
	 */
	public final static int STATUS_OFFSET = 5;
	/**
	 * offset of the FILL_WRITE register (to the base address)
	 */
	public final static int FILL_WRITE_OFFSET = 6;
	/**
	 * offset of the FILL_READ register (to the base address)
	 */
	public final static int FILL_READ_OFFSET = 7;

	//register values
	private int commandRegister = 0;
	private int addressLowRegister = 0;
	private int addressHighRegister = 0;

	//the maximum size of the memory in MB
	private int maxMemorySize = 1;
	private final Map<Integer, Integer> memory;

	private final int ddrId;

	private final List<Integer> readFifo;
	private final List<Integer> writeFifo;

	//status flags, only reseted on peripheral reset
	boolean rdOverflow = false;
	boolean wrUnderrun = false;

	private final SimulationManager simulationManager;


	/**
	 * create a new {@link McbInterfaceSmall}
	 *
	 * @param peri
	 *            the hardware description
	 * @param simulationManager
	 *            reference to the {@link SimulationManager}
	 */
	public McbInterfaceSmall(IPeripheralModuleConfiguration peri, final SimulationManager simulationManager, final int address) {
		super(peri, simulationManager, address);
		memory = new HashMap<Integer, Integer>();
		readFifo = new LinkedList<Integer>();
		writeFifo = new LinkedList<Integer>();
		this.simulationManager = simulationManager;

		ddrId = getFreeId();
	}


	/**
	 * retrieve an unused id for the view
	 * @return the lowest free ViewId
	 */
	private int getFreeId() {
		for (int i = 0; i < USED_DDR_IDS.size(); i++) {
			if (!USED_DDR_IDS.contains(i)) {
				USED_DDR_IDS.add(i);
				return i;
			}
		}
		int newId = USED_DDR_IDS.size();
		USED_DDR_IDS.add(newId);
		return newId;
	}

	/**
	 * determine if the commandRegister is set to read access.
	 * @return
	 */
	private boolean isReadCommand() {
		//[2:0] cmd:  Read  			001
		//	Read with Auto Precharge 	011

		final int operation = commandRegister & 7;

		if ((operation == 1) || (operation == 3)) {
			//precharge is an access optimisation, but the end effect is like normal read
			return true;
		}

		return false;
	}


	/**
	 * determine if the commandRegister is set to write command
	 * @return
	 */
	private boolean isWriteCommand() {
		//[2:0] cmd:   write 		000
		//Write with Auto Precharge 010

		final int operation = commandRegister & 7;

		if ((operation == 0) || (operation == 2)) {
			//precharge is an access optimisation, but the end effect is like normal write
			return true;
		}

		return false;
	}


	/**
	 * extract the burst length from the commandRegister
	 * @return the burst length (1 - 64)
	 */
	private int getBurstLength() {
		int burstLength = (commandRegister >> 3) + 1;

		if (burstLength > 64) {
			burstLength = 64;
		}

		return burstLength;
	}

	/**
	 * return the memory at the given address
	 * @param addr where to find the memory value
	 * @return value from the memory
	 */
	public int getMemoryAt(int addr) {
		Integer val = memory.get(addr);
		return val != null?
				val.intValue() :
				0;
	}

	@Override
	public void setMemorySize(int newSize) {
		maxMemorySize = newSize;
	}


	/**
	 * @return the size of the memory in MB
	 */
	public int getMemorySize() {
		return maxMemorySize;
	}


	@Override
	public int getMemoryId() {
		return ddrId;
	}


	/**
	 * reads the specified number of values from the memory and store it in the read fifo
	 * @param address the complete 30 bit address
	 * @param length the number of values to read
	 */
	private void readMemory(int address, int length) {

		assert (length <= 64);

		if (address + length > maxMemorySize * MB) {
			simulationManager.pushErrorMessage(this + ": Address outside of the memory range");
		}

		for (int i = 0; i < length; i++) {
			if (readFifo.size() < MAX_DATA_FIFO_SIZE) {
				Integer value = memory.get(address + i);
				if (value == null) {
					readFifo.add(0);
				} else {
					readFifo.add(value);
				}
			} else {
				//readFifo is full
				rdOverflow = true;
				break;
			}
		}

		fireRefreshEvent(REFRESH_REGISTER_ONLY);
	}


	@Override
	public int writeMemory(int offset, int[] values) {
		if (offset + values.length > maxMemorySize * MB) {
			return -1;
		}

		for (int i = 0; i < values.length; i++) {
			final int address = offset + i;
			memory.put(address, values[i]);

			fireRefreshEvent(address);
		}

		return 0;
	}


	/**
	 * @return the current value of the global status register
	 */
	private int calcStatusRegister() {
		//[0] -> calibration done, in simulation always done
		//[1] -> command fifo empty (cmd_fifo 4 Enträge)
		//[2] -> command fifo full

		//commands are executed immediately -> commandFifo always empty
		int reg = (1 << 0) | (1 << 1); // calibration done and command empty
		int fill = writeFifo.size();
		assert (fill <= 64);
		if (fill == 64) {
			reg |= (1 << 3);
		} else if (reg == 0) {
			reg |= (1 << 4);
		}
		if (wrUnderrun) {
			reg |= (1 << 5);
		}

		fill = readFifo.size();
		assert (fill <= 64);
		if (fill == 64) {
			reg |= (1 << 7);
		} else if (reg == 0) {
			reg |= (1 << 8);
		}
		if (rdOverflow) {
			reg |= (1 << 9);
		}

		return reg;
	}

	//**********************
	// AbstractPeripheral methods
	//******************

	@Override
	public void setup() {
		restart();
	}

	@Override
	public void tearDown() {
		//important, or frontend can break
		informTearDownListeners();
	}

	@Override
	public void restart() {
		//reset memory
		memory.clear();
		readFifo.clear();
		writeFifo.clear();

		//reset registers
		commandRegister = 0;
		addressLowRegister = 0;
		addressHighRegister = 0;

		//reset status flags
		rdOverflow = false;
		wrUnderrun = false;

		fireRefreshEvent(COMPLETE_REFRESH);
	}

	@Override
	public synchronized void writeRegisterValue(int offset, int value) {

		// DO NEVER CALL THIS METHOD BY THE GUI!!

		switch (offset) {
			case COMMAND_OFFSET:
				commandRegister = value;
				fireRefreshEvent(REFRESH_REGISTER_ONLY);
				return;
			case ADDR_LOW_OFFSET:
				addressLowRegister = value;

				final int completeAddress = (addressHighRegister << 18) | addressLowRegister;

				if (isReadCommand()) {
					readMemory(completeAddress, getBurstLength());

				} else if (isWriteCommand()) {
					final int burstLength = getBurstLength();
					final int diff = burstLength - writeFifo.size();

					//missing values are written as 0
					if (diff > 0) {
						wrUnderrun = true;

						for (int i = 0; i < diff; i++) {
							writeFifo.add(0);
						}
					}

					//remove written values from writeFifo
					int[] values = new int[burstLength];
					for (int i = 0; i < burstLength; i++) {
						values[i] = writeFifo.remove(0);
					}

					//apply changes
					writeMemory(completeAddress, values);
				}

				//the third possibility is refresh, nothing change -> nothing to do
				return;
			case ADDR_HIGH_OFFSET:
				addressHighRegister = value;
				fireRefreshEvent(REFRESH_REGISTER_ONLY);
				return;
			case DATA_WRITE_OFFSET:
				if (writeFifo.size() < MAX_DATA_FIFO_SIZE) {
					writeFifo.add(value);
				}
				fireRefreshEvent(REFRESH_REGISTER_ONLY);
				return;
			case DATA_READ_OFFSET:
				simulationManager.pushErrorMessage(this + ": The data read register is read only");
				return;
			default:
				//the remaining register are status register
				simulationManager.pushErrorMessage(this + ": Status registers are not writable");
		}
	}

	@Override
	public void setRegisterValue(int offset, int value) {
		switch (offset) {
			case COMMAND_OFFSET:
				commandRegister = value;
				break;
			case ADDR_LOW_OFFSET:
				//TODO: implement the setValue for McbInterfaceSmall's ADDR_LOW_OFFSET
				break;
			case ADDR_HIGH_OFFSET:
				//TODO: implement the setValue for McbInterfaceSmall's ADDR_HIGH_OFFSET
				break;
			case DATA_WRITE_OFFSET:
				//TODO: implement the setValue for McbInterfaceSmall's DATA_WRITE_OFFSET
				break;
			case DATA_READ_OFFSET:
				//TODO: implement the setValue for McbInterfaceSmall's DATA_READ_OFFSET
				break;
			case STATUS_OFFSET:
				//TODO: implement the setValue for McbInterfaceSmall's STATUS_OFFSET
				break;
			case FILL_WRITE_OFFSET:
				//TODO: implement the setValue for McbInterfaceSmall's FILL_WRITE_OFFSET
				break;
			case FILL_READ_OFFSET:
				//TODO: implement the setValue for McbInterfaceSmall's FILL_READ_OFFSET
				break;
			default:
				break;
		}
		fireRefreshEvent(REFRESH_REGISTER_ONLY);
	}

	@Override
	public synchronized int readRegisterValue(int offset) {

		// DO NEVER CALL THIS METHOD BY THE GUI!!

		switch (offset) {
			case DATA_WRITE_OFFSET:
				simulationManager.pushErrorMessage(this + ": The data write register is write only");
				return 0;
			case DATA_READ_OFFSET:
				if (readFifo.size() > 0) {
					final int value = readFifo.remove(0);
					fireRefreshEvent(REFRESH_REGISTER_ONLY);
					return value;
				}

				return 0;
			default:
				return getRegisterValue(offset);
		}
	}

	@Override
	public synchronized int getRegisterValue(final int offset) {
		switch (offset) {
			case COMMAND_OFFSET:
				return commandRegister;
			case ADDR_LOW_OFFSET:
				return addressLowRegister;
			case ADDR_HIGH_OFFSET:
				return addressHighRegister;
			case DATA_WRITE_OFFSET:
				if (writeFifo.size() > 0) {
					return writeFifo.get(0);
				}
				return 0;
			case DATA_READ_OFFSET:
				if (readFifo.size() > 0) {
					final int value = readFifo.get(0);
					return value;
				}
				return 0;
			case STATUS_OFFSET:
				return calcStatusRegister();
			case FILL_WRITE_OFFSET:
				return writeFifo.size();
			case FILL_READ_OFFSET:
				return readFifo.size();

			default:
				return 0;
		}
	}

	@Override
	public void importHardwareParameters(IPeripheralModuleConfiguration node) {
		setup();

		//other parameters are not required for the simulation or have to be set manually
	}

	/**
	 * provides access to the values of the DATA_WRITE and DATA_READ register
	 * @param offset either DATA_WRITE_OFFSET or DATA_READ_OFFSET (anything else will return always 0)
	 * @param repeat value of the register if read repeat times
	 * @return the value of the registers
	 */
	public synchronized Integer readAtOffset(int offset, int repeat) {
		switch (offset) {
			case DATA_WRITE_OFFSET:
				if (writeFifo.size() > repeat) {
					return writeFifo.get(repeat);
				} else {
					return 0;
				}

			case DATA_READ_OFFSET:
				if (readFifo.size() > repeat) {
					return readFifo.get(repeat);
				} else {
					return null;
				}
		}
		return 0;
	}

	@Override
	public String toString() {
		return "DDR Peripheral Interface " + ddrId;
	}


	@Override
	public String[] getParameters() {
		//no editable parameters
		return new String[0];
	}


	@Override
	public boolean setParameter(int parId, int value) {
		return false;
	}
}
