/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 *
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 *
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

/**
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.interrupts;

import de.tu_darmstadt.rs.sim.engine.SimulationComponent_I;
import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.JConfigPeripheral;
import de.tu_darmstadt.rs.sim.engine.peripheral.OnLoadedListener_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.PeripheralManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConnection;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.features.IConnectorInstance;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IParameterHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;


/**
 * The InterruptController causes external hardware interrupts. It may have
 * several interrupt sources, each defined by a name and a priority. If any
 * interrupt is triggered, its priority value is stored at the
 * InterruptController's IO address. The simulation will jump on the defined
 * interrupt jump address. The simulated application may read the IRQ IO address
 * to identify the source. Note: There should be only one single controller
 * instance at a time.
 */
public class InterruptController extends JConfigPeripheral implements
		Comparator<InterruptSource_I>, SimulationComponent_I,
		OnLoadedListener_I {

	private List<InterruptSource_I> interruptSourceList;
	private HashMap<InterruptSource_I, Integer> interruptSourcePriorities;
	private HashMap<Integer, String> periPriorities;
	private int expectedNumberOfInterrupts;

	private int allInterruptsIoValue;
	private int priorityInterruptIoValue;

	/**
	 * offset of the ALL_IRQ register (to the base address)
	 */
	public final static int ALL_IRQ_ADDRESS_OFFSET = 0x1;
	/**
	 * offset of the PRIORITY register (to the base address)
	 */
	public final static int PRIORITY_ADDRESS_OFFSET = 0x0;

	/**
	 * Constructor.
	 *
	 * @param peri
	 *            the hardware description
	 * @param simulationManager
	 *            reference to the {@link SimulationManager}
	 */
	public InterruptController(final IPeripheralModuleConfiguration peri,
							   final SimulationManager simulationManager, int address) {
		super(peri, simulationManager, address);
		setup();
		interruptSourceList = new ArrayList<InterruptSource_I>();
		interruptSourcePriorities = new HashMap<InterruptSource_I, Integer>();
		simulationManager.getPeripheralManager().registerOnLoadedListener(this);
	}

	/**
	 * Check if any interrupt is requested. No recalculation is done!
	 *
	 * @return true, if at least one source requests an interrupt
	 */
	public boolean hasInterrupt() {
		return (allInterruptsIoValue != 0x0);
	}

	/**
	 * Iterates all interrupt sources and checks if any interrupt is requested.
	 * All interrupts are recorded. Must be executed each CPU clock, immediately
	 * before hasInterrupt() & Co. are invoked.
	 */
	/**
	 public void calculateInterrupt() {

	 allInterruptsIoValue = 0x0;
	 priorityInterruptIoValue = 0x0;

	 for (InterruptSource_I source : interruptSourceList) {
	 if (source.hasInterruptRequest()) {
	 priorityInterruptIoValue = interruptSourcePriorities
	 .get(source);
	 allInterruptsIoValue |= 0x1 << (getIndexOfPriority(priorityInterruptIoValue - 1));
	 }
	 }
	 fireRefreshEvent(COMPLETE_REFRESH);
	 }
	 /**/

	/**
	 * post an interrupt request for the specified source
	 * @param source the source of the interrupt request
	 */
	public void postInterruptRequest(final InterruptSource_I source) {
		final int interruptPriority = interruptSourcePriorities.get(source);
		final int bitMask = 0x1 << (getIndexOfPriority(interruptPriority));
		if (priorityInterruptIoValue < interruptPriority) {
			priorityInterruptIoValue = interruptPriority;
		}
		if ((allInterruptsIoValue & bitMask) == 0) {
			allInterruptsIoValue |= bitMask;
			fireRefreshEvent(COMPLETE_REFRESH);
		}
	}

	/**
	 * reset an interrupt request of the specified source
	 * @param source the source of the previously requested interrupt
	 */
	public void resetInterruptRequest(final InterruptSource_I source) {
		final int interruptPriority = interruptSourcePriorities.get(source);
		final int bitMask = ~(0x1 << (getIndexOfPriority(interruptPriority)));
		allInterruptsIoValue &= bitMask;
		if (interruptPriority < priorityInterruptIoValue) {
			fireRefreshEvent(COMPLETE_REFRESH);
			return;
		}
		if (allInterruptsIoValue == 0x0) {
			priorityInterruptIoValue = 0x0;
			fireRefreshEvent(COMPLETE_REFRESH);
			return;
		}
		for (int i = priorityInterruptIoValue; i >= 0; i--) {
			if ((allInterruptsIoValue & (0x1 << i)) != 0) {
				priorityInterruptIoValue = getPriorityOfIndex(i);
				break;
			}
		}
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * Executes setup procedures.
	 */
	@Override
	public void setup() {
		restart();
	}

	/**
	 * Restarts the controller.
	 */
	@Override
	public void restart() {
		allInterruptsIoValue = 0x0;
		priorityInterruptIoValue = 0x0;
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * Retrieves the register value containing all interrupt requests.
	 *
	 * @return the register value
	 */
	public int getInterruptIoValue() {
		return allInterruptsIoValue;
	}

	/**
	 * Retrieves the register value containing the interrupt request with the
	 * highest priority value.
	 *
	 * @return the register value
	 */
	public int getPriorityIoValue() {
		return priorityInterruptIoValue;
	}

	/**
	 * retrieve the interrupt name which has the given priority
	 *
	 * @param priority
	 *            the priority to look for
	 * @return the name of the interrupt
	 */
	public String getInterruptNameAtPriority(final int priority) {
		if ((priority < 1) || (priority > getHighestPriority())) {
			return "<empty>";
		}
		InterruptSource_I source = interruptSourceList
				.get(getIndexOfPriority(priority));
		return source.getInterruptName();
	}

	/**
	 * Increments an interrupt priority and reorders the other ones.
	 *
	 * @param oldPriority
	 *            priority to be incremented
	 */
	public void incrementInterruptPriority(final int oldPriority) {

		assert (oldPriority != getHighestPriority());
		swapInterruptSources(oldPriority, oldPriority + 1);
	}

	/**
	 * Decrements an interrupt priority and reorders the other ones.
	 *
	 * @param oldPriority
	 *            priority to be decremented
	 */
	public void decrementInterruptPriority(final int oldPriority) {

		assert (oldPriority != 1);
		swapInterruptSources(oldPriority - 1, oldPriority);
	}

	private void swapInterruptSources(final int priorityOfLowerInterrupt,
									  final int priorityOfHigherInterrupt) {
		if ((priorityOfLowerInterrupt < 1)
				|| (priorityOfHigherInterrupt > getHighestPriority())) {
			return;
		}
		InterruptSource_I lowerInterrupt = interruptSourceList
				.get(getIndexOfPriority(priorityOfLowerInterrupt));
		InterruptSource_I higherInterrupt = interruptSourceList
				.get(getIndexOfPriority(priorityOfHigherInterrupt));
		interruptSourcePriorities
				.put(lowerInterrupt, priorityOfHigherInterrupt);
		interruptSourcePriorities
				.put(higherInterrupt, priorityOfLowerInterrupt);
		Collections.sort(interruptSourceList, this);
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	private int getIndexOfPriority(final int priority) {
		return priority - 1;
	}

	private int getPriorityOfIndex(final int index) {
		return index + 1;
	}

	/**
	 * Retrieves the highest interrupt priority.
	 *
	 * @return the priority
	 */
	public int getHighestPriority() {
		return expectedNumberOfInterrupts;
	}

	/**
	 * Will be invoked if a RFE instruction has been loaded into the IFID
	 * pipeline stage. NOT IMPLEMENTED YET.
	 */
	public void RfeInIfId() {
		// nothing
		// subject to change
	}

	@Override
	public String[] getParameters() {
		return new String[0];
	}

	@Override
	public void importHardwareParameters(IPeripheralModuleConfiguration node) {
		if (periPriorities == null) {
			periPriorities = new HashMap<Integer, String>();
		}
		IParameterHandler irSources = DocumentUtil.getParameterHandler(node, "IR_SOURCES");
		expectedNumberOfInterrupts = Integer.parseInt(irSources.getValue().orElseThrow());
		IPortHandler cNode = DocumentUtil.getPortHandler(node, "intr_in");
		IPeripheralModuleConfiguration peri;
		String uid;
		int prio;
		IConnectorInstance connectedPort;
		for (IConnection cep : (IConnection[]) cNode.streamConnections().toArray()) {
			connectedPort = cep.getOppositeConnector(cNode);
			if (connectedPort == null) {
				continue;
			}
			peri = (IPeripheralModuleConfiguration) cNode.getSourceElement();
			uid = peri.getModuleDescriptionName();
			prio = cep.getOppositeConnector(cNode).getConnectorWidth();
			// assuming that the id of an AbstractPeripheral is identical to the
			// uniqueId18 of a Peripheral
			periPriorities.put(prio, uid);
		}
		// TODO implement full import of hardware parameters for the int ctrl

	}

	@Override
	public boolean setParameter(int parId, int value) {
		return false;
	}

	@Override
	public void tearDown() {
		// do nothing
	}

	@Override
	public void onLoaded(Object loadedObject) {
		final PeripheralManager periMgr = (PeripheralManager) loadedObject;
		int lastPrio = 0;
		String peripheryID;
		int realPriority;
		JConfigPeripheral peripheral;
		InterruptSource_I source;
		DummyInterruptSource dummySource;
		interruptSourceList.clear();

		for (int priority = 0; priority < expectedNumberOfInterrupts; priority++) {
			peripheryID = periPriorities.get(priority);
			realPriority = priority + 1;
			peripheral = periMgr.getPeripheral(peripheryID);
			if ((peripheral != null)
					&& (peripheral instanceof InterruptSource_I)) {
				source = (InterruptSource_I) peripheral;
				source.setInterruptController(this);
				interruptSourcePriorities.put(source, realPriority);
				interruptSourceList.add(source);
				lastPrio = realPriority;
			} else {
				dummySource = new DummyInterruptSource();
				lastPrio++;
				interruptSourcePriorities.put(dummySource, lastPrio);
				interruptSourceList.add(dummySource);
			}
		}
		Collections.sort(interruptSourceList, this);
		fireRefreshEvent(COMPLETE_REFRESH);
		periPriorities = null;
	}

	@Override
	public int compare(InterruptSource_I o1, InterruptSource_I o2) {
		int priority1 = interruptSourcePriorities.get(o1);
		int priority2 = interruptSourcePriorities.get(o2);
		return priority1 - priority2;
	}

	@Override
	public int getRegisterValue(int offset) {
		switch (offset) {
			case ALL_IRQ_ADDRESS_OFFSET:
				return allInterruptsIoValue;
			case PRIORITY_ADDRESS_OFFSET:
				return priorityInterruptIoValue;
		}
		return 0;
	}

	@Override
	public void setRegisterValue(int offset, int value) {
		switch (offset) {
			case ALL_IRQ_ADDRESS_OFFSET:
				allInterruptsIoValue = value;
				break;
			case PRIORITY_ADDRESS_OFFSET:
				priorityInterruptIoValue = value;
				break;
		}
	}
}
