/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.ports;


import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.JConfigPeripheral;
import de.tu_darmstadt.rs.sim.engine.peripheral.interrupts.InterruptController;
import de.tu_darmstadt.rs.sim.engine.peripheral.interrupts.InterruptSource_I;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;


/**
 * Class representing the port_in peripheral. It is used to control input pins.
 */
public class PortIn extends JConfigPeripheral implements InterruptSource_I {

	/**
	 * offset of the DAT register (to the base address)
	 */
	public static final int DAT_OFFSET = 0;
	/**
	 * offset of the IE register (to the base address)
	 */
	public static final int IE_OFFSET = 1;
	/**
	 * offset of the EDGSEL register (to the base address)
	 */
	public static final int EDGSEL_OFFSET = 2;
	/**
	 * offset of the IR_STATUS register (to the base address)
	 */
	public static final int IR_STATUS_OFFSET = 3;

	private int datRegisterValue;
	private int ieRegisterValue;
	private int edgselRegisterValue;
	private int ir_statusRegisterValue;

	private int numberOfPins;
	private int numberOfPinsMask;

	private boolean hasInterrupt;
	private InterruptController interruptController;

	/**
	 * create a new {@link PortIn}
	 *
	 * @param peri the hardware description
	 * @param simulationManager reference to the {@link SimulationManager}
	 */
	public PortIn(IPeripheralModuleConfiguration peri,
				  final SimulationManager simulationManager, int baseAddress) {
		super(peri, simulationManager, baseAddress);
	}

	@Override
	public void importHardwareParameters(IPeripheralModuleConfiguration<?> node) {
		super.importHardwareParameters(node);
		numberOfPins = node.getPortHandler(node.findPortDescription("pin").get()).getConnectorWidth();
		numberOfPinsMask = (1 << numberOfPins) - 1;
	}

	@Override
	public int readRegisterValue(int offset) {
		if (offset == DAT_OFFSET || offset == IE_OFFSET
				|| offset == EDGSEL_OFFSET) {
			ir_statusRegisterValue = 0;
			changeInterrupt();
		}
		return getRegisterValue(offset);
	}

	@Override
	public int getRegisterValue(int offset) {
		switch (offset) {
			case DAT_OFFSET:
				return datRegisterValue;
			case IE_OFFSET:
				return ieRegisterValue;
			case EDGSEL_OFFSET:
				return edgselRegisterValue;
			case IR_STATUS_OFFSET:
				return ir_statusRegisterValue;
		}
		return 0;
	}

	@Override
	public void setRegisterValue(int offset, int value) {
		switch (offset) {
			case DAT_OFFSET:
				changeValue(value & numberOfPinsMask);
				break;
			case IE_OFFSET:
				ieRegisterValue = value & numberOfPinsMask;
				break;
			case EDGSEL_OFFSET:
				edgselRegisterValue = value & numberOfPinsMask;
				break;
			case IR_STATUS_OFFSET:
				ir_statusRegisterValue = value & numberOfPinsMask;
				break;
		}
		changeInterrupt();
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	@Override
	public void writeRegisterValue(int offset, int value) {
		switch (offset) {
			case IE_OFFSET:
				ieRegisterValue = value & numberOfPinsMask;
				break;
			case EDGSEL_OFFSET:
				edgselRegisterValue = value & numberOfPinsMask;
				break;
		}
		changeInterrupt();
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	@Override
	public void restart() {
		ieRegisterValue = 0;
		edgselRegisterValue = 0;
		ir_statusRegisterValue = 0;
		if (interruptController != null) {
			changeInterrupt();
		}
	}

	/**
	 * change the input pins' value. In the simulator, this is done via changing
	 * register bits in the GUI.
	 *
	 * @param value the new register value
	 */
	public void changeValue(int value) {
		int changedBits = value ^ datRegisterValue;
		ir_statusRegisterValue = edgselRegisterValue
				^ ((datRegisterValue & changedBits) | (edgselRegisterValue & ~changedBits));
		datRegisterValue = value;
	}

	@Override
	public String getInterruptName() {
		return getPeriName() + " IRQ";
	}

	private void changeInterrupt() {
		final boolean hasInterrupt_old = hasInterrupt;
		hasInterrupt = (ir_statusRegisterValue & ieRegisterValue) != 0;
		if (interruptController != null) {
			if (hasInterrupt_old ^ hasInterrupt) {
				if (hasInterrupt) {
					interruptController.postInterruptRequest(this);
				} else {
					interruptController.resetInterruptRequest(this);
				}
			}
		}
	}

	@Override
	public void setInterruptController(InterruptController interruptController) {
		this.interruptController = interruptController;
	}
}
