/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.register;


import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;

/**
 * Helper class for evaluating the bit group values held by a peripheral's registers.
 */
public class PeripheralRegisterUtil {

	/**
	 * get a mask for the given bit group.<br />
	 * The result is a value that could be used like this:<br />
	 * <code>bitGroupValue = registerValue & PeripheralRegisterUtil.getBitGroupMask(theBitGroup);</code>
	 *
	 * @param bitGroup the bit group whose mask bits should be found
	 * @return the bit mask for the given bit group
	 */
	private static int getBitGroupMask(PeripheralModuleDescription.BitGroup bitGroup) {
		final int lsb = bitGroup.getLsb();
		final int msb = lsb + bitGroup.getWidth() - 1;
		int returnValue = 0;
		for (int i = lsb; i <= msb; i++) {
			returnValue += (1 << i);
		}
		return returnValue;
	}

	/**
	 * helper method to set the bit group's value within the given register value.<br />
	 * The bit group's value is automatically shifted and masked and must therefore not been done by the caller.
	 * The method can be used like this:<br />
	 * <code>newRegisterValue = PeripheralRegisterUtil.putBitGroupValue(oldRegisterValue, theBitGroup, 0x1);</code>
	 *
	 * @param registerValue the old register value
	 * @param group the group whose value should be changed
	 * @param groupValue the new bit group's value
	 * @return the new register value
	 */
	public static int putBitGroupValue(final int registerValue, final PeripheralModuleDescription.BitGroup group, final int groupValue) {
		final int mask = getBitGroupMask(group);
		final int registerValueWithoutGroup = registerValue & ~mask;
		final int realGroupValue = (groupValue << group.getWidth()) & mask;
		return registerValueWithoutGroup + realGroupValue;
	}

	/**
	 * helper method to get a bit group's value from within the given register value.<br />
	 * The method can be used like this:<br />
	 * <code>bitGroupValue = PeripheralRegisterUtil.getShiftedBitGroupValue(registerValue, theBitGroup);</code>
	 *
	 * @param registerValue the register value
	 * @param group the bit group
	 * @return the bit group's value
	 */
	public static int getShiftedBitGroupValue(final int registerValue, final PeripheralModuleDescription.BitGroup group) {
		final int mask = getBitGroupMask(group);
		final int returnValue = (registerValue & mask) >> group.getWidth();
		return returnValue;
	}
}
