/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.timer;

import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.JConfigPeripheral;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IPortHandler;
import de.tu_darmstadt.rs.spartanmc.jconfig.util.DocumentUtil;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.List;


/**
 * abstract base class for {@link BasicTimer} and {@link RTI}
 */
public abstract class AbstractTimer extends JConfigPeripheral implements
		TimerSource_I, TimerObserver_I {

	/**
	 * the source of this timer
	 */
	private TimerSource_I timerSource;
	/**
	 * the timers connected to the output of this one
	 */
	private List<TimerObserver_I> relatedTimers;

	/**
	 * used for the division of the input de.tu_darmstadt.rs.sim.signal
	 */
	protected Pre pre;

	/**
	 * used to filter the clock input de.tu_darmstadt.rs.sim.signal
	 */
	protected int inputFilterBit = 17;

	private int lastValue = 0;

	/**
	 * abstract constructor used from subclasses of {@link AbstractTimer}
	 * remember to call {@link #setup()} from subclasses of
	 * {@link AbstractTimer}
	 *
	 * @param peri the {@link de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration} to get hardware information from
	 * @param mgr reference to the {@link SimulationManager}
	 */
	public AbstractTimer(final IPeripheralModuleConfiguration peri, final SimulationManager mgr, final int address) {
		super(peri, mgr, address);
		pre = new Pre();
		relatedTimers = new ArrayList<TimerObserver_I>();
	}

	@Override
	public void importHardwareParameters(final IPeripheralModuleConfiguration node) {
		String signalName;
		String connectedTimerName;
		BitSet connectedBits;
		// connections to this timer
		IPortHandler cNode = DocumentUtil.getPortHandler(node, "clk_i");
		//TODO make sure that signalName is in fact the port name
		signalName = cNode.getPortName();
		connectedTimerName = SignalHelper.getTimerNameFromConnectedSignal(signalName);
		TimerManager.addTimerConnection(connectedTimerName, this);
		if (cNode.isConnected()) {
			connectedBits = new BitSet(cNode.getConnectorWidth());
			inputFilterBit = SignalHelper.getIndexOfHighestConnectedBit(connectedBits);
		}
	}

	@Override
	public final void registerTimerObserver(final TimerObserver_I timerObserver) {
		relatedTimers.add(timerObserver);
	}

	@Override
	public final void unregisterTimerObserver(final TimerObserver_I timerObserver) {
		assert (relatedTimers.contains(timerObserver));
		relatedTimers.remove(timerObserver);
	}

	/**
	 * inform all components which are connected to the timer's output about a
	 * performed clock
	 *
	 * @param value the bits of the output de.tu_darmstadt.rs.sim.signal
	 */
	protected void informTimerObservers(final int value) {
		for (TimerObserver_I timerObserver : relatedTimers) {
			timerObserver.onClock(value);
		}
	}

	@Override
	public final void setTimerSource(final TimerSource_I timerSource) {
		if (this.timerSource != null) {
			this.timerSource.unregisterTimerObserver(this);
		}
		this.timerSource = timerSource;
		timerSource.registerTimerObserver(this);
		if (timerSource.getOutputBitCount() <= inputFilterBit) {
			inputFilterBit = timerSource.getOutputBitCount() - 1;
		}
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * retrieve the name of the element that the timer is connected to
	 *
	 * @return the name of the timer's source
	 */
	public final String getTimerSourceString() {
		return TimerManager.getName(timerSource);
	}

	@Override
	public String getUniqueTimerName() {
		return getPeriName();
	}

	/**
	 * use another bit of the input de.tu_darmstadt.rs.sim.signal
	 *
	 * @param bitIndex the new bit to use
	 */
	public void setInputFilterBit(final int bitIndex) {
		inputFilterBit = bitIndex;
	}

	/**
	 * retrieve which bit of the input de.tu_darmstadt.rs.sim.signal is used
	 *
	 * @return index of the bit
	 */
	public int getInputFilterBit() {
		return inputFilterBit;
	}

	/**
	 * check whether the input de.tu_darmstadt.rs.sim.signal's value clocks this instance
	 *
	 * @param value input de.tu_darmstadt.rs.sim.signal bits from a source
	 * @return true if the value of the input de.tu_darmstadt.rs.sim.signal should clock, false
	 * otherwise
	 */
	protected boolean shouldValueClock(final int value) {
		final int internalValue = (value & (1 << inputFilterBit));
		boolean returnValue = false;
		if (internalValue != 0 && lastValue == 0) {
			returnValue = true;
		}
		lastValue = internalValue;
		return returnValue;
	}

	@Override
	public String toString() {
		String str = String.format("%s \"%s\" @ 0x%05x", getClass()
				.getSimpleName(), getUniqueTimerName(), super.getBaseAddr());
		return str;
	}

	@Override
	public void tearDown() {
		if (this.timerSource != null) {
			this.timerSource.unregisterTimerObserver(this);
		}
		super.tearDown();
	}
}
