/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.timer;

import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.register.PeripheralRegisterUtil;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;


/**
 * Class representing the Basic Timer module (timer)
 */
public class BasicTimer extends AbstractTimer {

	/**
	 * offset of the CTRL register to the base IO address
	 */
	public final static int TIMER_CTRL_IO_ADDRESS_OFFSET = 0x0;
	/**
	 * offset of the DAT register to the base IO address
	 */
	public final static int TIMER_DAT_IO_ADDRESS_OFFSET = 0x1;
	/**
	 * offset of the VALUE register to the base IO address
	 */
	public final static int TIMER_VALUE_IO_ADDRESS_OFFSET = 0x2;

	/**
	 * mask for all used bits of the CTRL register
	 */
	public final static int CONTROL_MASK = 0x0001f;
	/**
	 * mask for the TI_EN bit
	 */
	public final static int CONTROL_TIMER_ENABLE = 0x00001;
	/**
	 * mask for the PRE_EN bit
	 */
	public final static int CONTROL_PRE_ENABLE = 0x00002;
	/**
	 * mask for the PRE value
	 */
	public final static int CONTROL_PRE_MASK = 0x0001c;

	/**
	 * timer control register (CTRL)
	 */
	private int timerCtrl;
	/**
	 * maximum timer value register (DAT)
	 */
	private int timerDat;
	/**
	 * current timer value register (VALUE)
	 */
	private int timerValue;
	/**
	 * is the timer enabled?
	 */
	private boolean timerEnabled;
	/**
	 * is the precounter enabled?
	 */
	private boolean preEnabled;

	/**
	 * create a new {@link BasicTimer}.
	 *
	 * @param peri the {@link de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration} to get hardware information from
	 * @param simulationManager reference to the {@link SimulationManager}
	 */
	public BasicTimer(final IPeripheralModuleConfiguration peri,
					  final SimulationManager simulationManager, final int baseAddress) {
		super(peri, simulationManager, baseAddress);
		TimerManager.addTimerSource(this);
		setup();
	}

	@Override
	public void setup() {
		restart();
	}

	@Override
	public void restart() {
		initTimerValue();
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * initialize the timer
	 */
	private void initTimerValue() {
		timerValue = 0;
		timerDat = 0;
		timerCtrl = 0;
		timerEnabled = false;
		preEnabled = false;
		pre.setDivider(0);
		pre.setValue(0);
	}

	@Override
	public void onClock(int value) {
		performClockingIfPossible(value);
	}

	/**
	 * check conditions for clocking and perform the clocking
	 */
	private void performClockingIfPossible(int value) {
		if (timerEnabled) {
			if (preEnabled) {
				if (shouldValueClock(value)) {
					if (pre.clock()) {
						performClocking();
					}
				}
			} else {
				if (shouldValueClock(value)) {
					performClocking();
				}
			}
		}
	}

	/**
	 * increase the timer value and inform the observers if necessary
	 */
	private void performClocking() {
		timerValue++;
		if (timerValue == timerDat) {
			timerValue = 0;
		}
		informTimerObservers(timerValue);
		fireRefreshEvent(REFRESH_REGISTER_ONLY);
	}

	/**
	 * retrieve the CTRL (control) register's value
	 *
	 * @return value of the CTRL register
	 */
	public int getCTRLValue() {
		return timerCtrl;
	}

	/**
	 * retrieve the DAT (limit) register's value
	 *
	 * @return value of the DAT register
	 */
	public int getDATValue() {
		return timerDat;
	}

	/**
	 * retrieve the VALUE (current counter value) register's value
	 *
	 * @return value of the VALUE register
	 */
	public int getVALUEValue() {
		return timerValue;
	}

	/**
	 * retrieve whether the basic timer is enabled
	 *
	 * @return true if enabled, false otherwise
	 */
	public boolean getBasicTimerEnabled() {
		return timerEnabled;
	}

	/**
	 * retrieve whether the basic timer's precounter is enabled
	 *
	 * @return true if enabled, false otherwise
	 */
	public boolean getPreEnabled() {
		return preEnabled;
	}

	/**
	 * retrieve the precounter's division value
	 *
	 * @return divisor
	 */
	public int getPreDivider() {
		return pre.getDivider();
	}

	@Override
	public String[] getParameters() {
		return new String[0];
	}

	@Override
	public boolean setParameter(int parId, int value) {
		return false;
	}

	@Override
	public int getOutputBitCount() {
		return 18;
	}

	@Override
	public int getRegisterValue(int offset) {
		switch (offset) {
			case TIMER_CTRL_IO_ADDRESS_OFFSET:
				return timerCtrl;
			case TIMER_DAT_IO_ADDRESS_OFFSET:
				return timerDat;
			case TIMER_VALUE_IO_ADDRESS_OFFSET:
				return timerValue;
		}
		return SpartanDefinitions_I.INVALID;
	}

	@Override
	public void setRegisterValue(int offset, int value) {
		switch (offset) {
			case TIMER_CTRL_IO_ADDRESS_OFFSET:
				timerCtrl = value;
				final PeripheralModuleDescription.BitGroup[] bitGroups = (PeripheralModuleDescription.BitGroup[]) getRegisters()[0].getBitGroups().toArray();
				timerEnabled = PeripheralRegisterUtil.getShiftedBitGroupValue(
						value, bitGroups[0]) != 0;
				preEnabled = PeripheralRegisterUtil.getShiftedBitGroupValue(value,
						bitGroups[1]) != 0;
				pre.setDivider(Pre.getDividerForTimer(PeripheralRegisterUtil
						.getShiftedBitGroupValue(value, bitGroups[2])));
				pre.setValue(0);
				break;
			case TIMER_DAT_IO_ADDRESS_OFFSET:
				timerDat = value;
				break;
			case TIMER_VALUE_IO_ADDRESS_OFFSET:
				timerValue = value;
				break;
		}
		fireRefreshEvent(0);
	}
}
