/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.timer;

import java.util.ArrayList;
import java.util.List;

/**
 * Class representing the pipeline's clock
 */
public class PipelineClock implements TimerSource_I, TimerObserver_I {

	/**
	 * the name of this clock
	 */
	public static final String PIPELINE_CLOCK_NAME = "Pipeline clock";
	private List<TimerObserver_I> timerObservers;
	private boolean wasClocked;
	private TimerSource_I timerSource;
	private int inputFilterBitIndex;

	/**
	 * Constructor.
	 */
	public PipelineClock() {
		timerObservers = new ArrayList<TimerObserver_I>();
		TimerManager.addTimerSource(this);
	}

	@Override
	public String getUniqueTimerName() {
		return PIPELINE_CLOCK_NAME;
	}

	@Override
	public void registerTimerObserver(TimerObserver_I timerObserver) {
		timerObservers.add(timerObserver);
	}

	@Override
	public void unregisterTimerObserver(TimerObserver_I timerObserver) {
		assert (timerObservers.contains(timerObserver));
		timerObservers.remove(timerObserver);
	}

	/**
	 * checks whether the {@link PipelineClock} was clocked.
	 * NOTE: This method resets the internal flag that indicates this property.
	 *
	 * @return true when it was clocked and not reset
	 */
	public boolean wasClocked() {
		boolean retVal = wasClocked;
		wasClocked = false;
		return retVal;
	}

	@Override
	public void onClock(int value) {
		if (shouldValueClock(value)) {
			informClockObservers();
			wasClocked = true;
		}
	}

	/**
	 * check whether the given value should start a clock cycle
	 *
	 * @param value the microstep value
	 * @return true if value starts a clock cycle, false otherwise
	 */
	private boolean shouldValueClock(int value) {
		return (value & (1 << inputFilterBitIndex)) != 0;
	}

	/**
	 * inform all the observers
	 */
	private void informClockObservers() {
		for (TimerObserver_I clkObserver : timerObservers) {
			clkObserver.onClock(1);
			clkObserver.onClock(0);
		}
	}

	@Override
	public void setTimerSource(TimerSource_I timerSource) {
		if (this.timerSource != null) {
			this.timerSource.unregisterTimerObserver(this);
		}
		this.timerSource = timerSource;
		timerSource.registerTimerObserver(this);
	}

	@Override
	public int getOutputBitCount() {
		return 1;
	}

	/**
	 * set the input filter bit
	 *
	 * @param index the used bit's index from the source
	 */
	public void setInputFilterBit(int index) {
		inputFilterBitIndex = index;
	}
}
