/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.timer;

/**
 * Pre is representing a physical precounter (as used in the BasicTimer and RTI module)
 */
public class Pre {

	/**
	 * return a division value that is used as the limit in a BasicTimer precounter
	 *
	 * @param preBits the bit group from the BasicTimer's CTRL register value
	 * @return the limit to which the precounter counts
	 */
	public static int getDividerForTimer(final int preBits) {
		return getDividerValue(preBits + 1);
	}

	/**
	 * return a division value that is used as the limit in a RTI precounter
	 *
	 * @param preBits the bit group from the RTI's CTRL register value
	 * @return the limit to which the precounter counts
	 */
	public static int getDividerForRTI(final int preBits) {
		return getDividerValue(preBits);
	}

	private static int getDividerValue(final int preBits) {
		int dividerExponent = preBits;
		int divider = 1 << dividerExponent;
		return divider;
	}

	/**
	 * current value of the precounter
	 */
	private int value;
	/**
	 * the precounter's limit
	 */
	private int divider;

	/**
	 * create a precounter
	 */
	public Pre() {
		value = 0;
	}

	/**
	 * set the precounter's limit to which it counts before transmitting a de.tu_darmstadt.rs.sim.signal
	 *
	 * @param divider the new limit
	 */
	public void setDivider(int divider) {
		this.divider = divider;
	}

	/**
	 * get the precounter's limit
	 *
	 * @return the limit
	 */
	public int getDivider() {
		return divider;
	}

	/**
	 * set the precounter's value (this method is especially used on reset)
	 *
	 * @param value the new counter value
	 */
	public void setValue(int value) {
		this.value = value;
	}

	/**
	 * get the precounter's current count value
	 *
	 * @return the current value
	 */
	public int getValue() {
		return value;
	}

	/**
	 * simulate an input clock
	 *
	 * @return true if this results in an output clock, false otherwise
	 */
	public boolean clock() {
		boolean retVal = false;
		value++;
		if (value == divider) {
			retVal = true;
			value = 0;
		}
		return retVal;
	}
}