/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.timer;


import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.interrupts.InterruptController;
import de.tu_darmstadt.rs.sim.engine.peripheral.interrupts.InterruptSource_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.register.PeripheralRegisterUtil;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;


/**
 * Class representing the Real Time Interrupt (timer_rti)
 */
public class RTI extends AbstractTimer implements InterruptSource_I {

	private InterruptController interruptController;
	/**
	 * offset of the CTRL register to the base IO address
	 */
	public final static int RTI_CTRL_IO_ADDRESS_OFFSET = 0x0;

	/**
	 * mask for all used bits of the CTRL register
	 */
	public final static int CONTROL_MASK = 0x0003f;
	/**
	 * mask for the RTI_EN bit
	 */
	public final static int CONTROL_RTI_ENABLE = 0x00001;
	/**
	 * mask for the INT_EN bit
	 */
	public final static int CONTROL_INT_ENABLE = 0x00002;
	/**
	 * mask for the PRE value
	 */
	public final static int CONTROL_PRE_MASK = 0x0003c;

	/**
	 * timer control register
	 */
	private int rtiCtrl;
	/**
	 * is the RTI enabled?
	 */
	private boolean rtiEnabled;
	/**
	 * is the interrupt enabled?
	 */
	private boolean interruptEnabled;
	/**
	 * is there an interrupt?
	 */
	private boolean interruptSignal;

	/**
	 * create a new {@link RTI}.
	 *
	 * @param peri the {@link de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration} to get hardware information from
	 * @param simulationManager reference to the {@link SimulationManager}
	 */
	public RTI(final IPeripheralModuleConfiguration peri,
			   final SimulationManager simulationManager, final int baseAddress) {
		super(peri, simulationManager, baseAddress);
		TimerManager.addTimerSource(this);
		setup();
	}

	/**
	 * initialize the timer
	 */
	private void initTimerValue() {
		pre.setDivider(0);
		pre.setValue(0);
		rtiCtrl = 0;
		rtiEnabled = false;
		interruptEnabled = false;
		interruptSignal = false;
		if (interruptController != null) {
			interruptController.resetInterruptRequest(this);
		}
	}

	@Override
	public void setup() {
		restart();
	}

	@Override
	public void restart() {
		initTimerValue();
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * reset the interrupt de.tu_darmstadt.rs.sim.signal
	 */
	private void resetInterruptSignal() {
		if (interruptController != null) {
			interruptController.resetInterruptRequest(this);
		}
		interruptSignal = false;
		fireRefreshEvent(REFRESH_REGISTER_ONLY);
	}

	@Override
	public void onClock(int value) {
		if (shouldValueClock(value)) {
			performClocking();
		}
	}

	/**
	 * check conditions for clocking and perform the clocking
	 */
	private void performClocking() {
		if (rtiEnabled) {
			if (pre.clock()) {
				if (interruptEnabled) {
					if (interruptController != null) {
						interruptController.postInterruptRequest(this);
					}
					interruptSignal = true;
				}
				informTimerObservers(1);
			} else {
				informTimerObservers(0);
			}
			fireRefreshEvent(REFRESH_REGISTER_ONLY);
		}
	}

	/**
	 * for the GUI
	 *
	 * @return true if there is an interrupt request, false otherwise
	 */
	public boolean hasInterruptRequest() {
		return interruptSignal;
	}

	@Override
	public void setInterruptController(final InterruptController interruptController) {
		this.interruptController = interruptController;
	}

	@Override
	public String getInterruptName() {
		return getPeriName() + " IRQ";
	}

	/**
	 * retrieve the CTRL (control) register's value
	 *
	 * @return value of the CTRL register
	 */
	public int getCTRLRegisterValue() {
		return rtiCtrl;
	}

	/**
	 * retrieve whether the RTI is enabled
	 *
	 * @return true if the RTI is enabled, false otherwise
	 */
	public boolean getRTIEnable() {
		return rtiEnabled;
	}

	/**
	 * retrieve whether the RTI can create interrupts
	 *
	 * @return true if the RTI creates interrupts, false otherwise
	 */
	public boolean getInterruptEnabled() {
		return interruptEnabled;
	}

	/**
	 * retrieve the precounter's division value
	 *
	 * @return divisor
	 */
	public int getPreDivider() {
		return pre.getDivider();
	}

	@Override
	public int getOutputBitCount() {
		return 1;
	}

	/**
	 * retrieve the precounter's current value
	 *
	 * @return the current value of the precounter
	 */
	public int getPrecounterValue() {
		return pre.getValue();
	}

	@Override
	public int getRegisterValue(int offset) {
		return rtiCtrl;
	}

	@Override
	public int readRegisterValue(int offset) {
		resetInterruptSignal();
		return rtiCtrl;
	}

	@Override
	public void setRegisterValue(int offset, int value) {
		assert (offset == RTI_CTRL_IO_ADDRESS_OFFSET);
		rtiCtrl = value;
		final PeripheralModuleDescription.BitGroup[] bitGroups = (PeripheralModuleDescription.BitGroup[]) getRegisters()[0].getBitGroups().toArray();
		rtiEnabled = PeripheralRegisterUtil.getShiftedBitGroupValue(value,
				bitGroups[0]) != 0;
		interruptEnabled = PeripheralRegisterUtil.getShiftedBitGroupValue(
				value, bitGroups[1]) != 0;
		pre.setDivider(Pre.getDividerForRTI(PeripheralRegisterUtil
				.getShiftedBitGroupValue(value, bitGroups[2])));
		pre.setValue(0);
		fireRefreshEvent(0);
	}
}
