/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.timer;

import de.tu_darmstadt.rs.sim.engine.RefreshHandler_I;
import de.tu_darmstadt.rs.sim.engine.RefreshListener_I;
import de.tu_darmstadt.rs.sim.engine.RefreshTrigger_I;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;


/**
 * Single instance to remember all Timers<br />
 * Provides the timers' names to the GUI's timer views (BasicTimer and RTI).
 * Because the order of construction of timers from the preference file is
 * unknown, this class rebuilds the connections between them lazily with
 * addTimerConnection.
 */
public final class TimerManager implements RefreshTrigger_I {

	private static class InstanceHolder {

		public static TimerManager instance = null;
	}

	// public static final String PIPELINE_CLOCK_NAME="Pipeline clock";

	/**
	 * to be able to connect timers via GUI, remember the names and the
	 * associated instances ...
	 */
	private Map<String, TimerSource_I> timerNameMapping;

	/**
	 * ... and the mapping from instance to name
	 */
	private Map<TimerSource_I, String> nameTimerMapping;

	/**
	 * to remember connections when the construction of a timer is not finished
	 * (especially when loaded from preference or jconfig file)
	 */
	private Map<String, List<TimerObserver_I>> timerConnections;

	/**
	 * to reflect data changes to the GUI
	 */
	private List<RefreshListener_I> refreshListeners;

	private boolean refresh;

	/**
	 * private constructor due to singleton pattern
	 */
	private TimerManager(final RefreshHandler_I refreshHandler) {
		timerNameMapping = new HashMap<String, TimerSource_I>();
		nameTimerMapping = new HashMap<TimerSource_I, String>();
		timerConnections = new HashMap<String, List<TimerObserver_I>>();
		refreshListeners = new ArrayList<RefreshListener_I>();
		refresh = false;
		refreshHandler.registerTrigger(this);
	}

	/**
	 * initialize the {@link TimerManager}
	 *
	 * @param refreshHandler the refresh handler
	 */
	public synchronized static void init(final RefreshHandler_I refreshHandler) {
		InstanceHolder.instance = new TimerManager(refreshHandler);
	}

	/**
	 * remove a timer input de.tu_darmstadt.rs.sim.signal
	 *
	 * @param timerSource the timer that should be removed
	 */
	public static void removeTimerSource(TimerSource_I timerSource) {
		InstanceHolder.instance.removeTimerSourceInternal(timerSource);
	}

	/**
	 * internal method to remove a timer input de.tu_darmstadt.rs.sim.signal
	 *
	 * @param timerSource the timer that should be removed
	 * @see TimerManager#removeTimerSource(TimerSource_I)
	 */
	private void removeTimerSourceInternal(TimerSource_I timerSource) {
		String name = timerSource.getUniqueTimerName();
		TimerSource_I oldValue = timerNameMapping.remove(name);
		assert (oldValue != null);
		nameTimerMapping.remove(oldValue);
	}

	/**
	 * register a new timer input de.tu_darmstadt.rs.sim.signal so that it can be chosen from the GUI
	 *
	 * @param timerSource the new timer input de.tu_darmstadt.rs.sim.signal
	 */
	public static void addTimerSource(TimerSource_I timerSource) {
		InstanceHolder.instance.addTimerSourceInternal(timerSource);
	}

	/**
	 * internal method to register a new timer input de.tu_darmstadt.rs.sim.signal so that it can be
	 * chosen from the GUI
	 *
	 * @param timerSource the new timer input de.tu_darmstadt.rs.sim.signal
	 * @see TimerManager#addTimerSource(TimerSource_I)
	 */
	private void addTimerSourceInternal(TimerSource_I timerSource) {
		String name = timerSource.getUniqueTimerName();
		TimerSource_I oldValue = timerNameMapping.put(name, timerSource);
		// assert that there was no timer registered with the same name
		assert (oldValue == null);
		// register the mapping vice versa
		nameTimerMapping.put(timerSource, name);
		// look for connections
		List<TimerObserver_I> connectedTimers = timerConnections.get(name);
		// if there are connections from the timer with this name
		if (connectedTimers != null) {
			// set this timer as the source of each connection
			for (TimerObserver_I connectedTimer : connectedTimers) {
				connectedTimer.setTimerSource(timerSource);
			}
			// remove the connections from the memory of the TimerManager
			connectedTimers.clear();
			timerConnections.remove(name);
		}
		// inform the GUI that there is a new Timer which could be connected to
		informTheListeners();
	}

	/**
	 * connect to the timer which has the name nameOfTimerSource or remember the
	 * connection if this timer does not exist yet
	 *
	 * @param nameOfTimerSource the source timer of the connection
	 * @param timerDestination the destination timer of the connection
	 */
	public static void addTimerConnection(String nameOfTimerSource,
										  TimerObserver_I timerDestination) {
		InstanceHolder.instance.addTimerConnectionInternal(nameOfTimerSource,
				timerDestination);
	}

	/**
	 * internal method to connect to the timer which has the name
	 * nameOfTimerSource or remember the connection if this timer does not exist
	 * yet
	 *
	 * @param nameOfTimerSource the source timer of the connection
	 * @param timerDestination the destination timer of the connection
	 * @see TimerManager#addTimerConnection(String, TimerObserver_I)
	 */
	private void addTimerConnectionInternal(String nameOfTimerSource,
											TimerObserver_I timerDestination) {
		TimerSource_I theSource = timerNameMapping.get(nameOfTimerSource);
		if (theSource == null) {
			List<TimerObserver_I> connectedTimers = timerConnections
					.get(nameOfTimerSource);
			if (connectedTimers == null) {
				connectedTimers = new ArrayList<TimerObserver_I>();
				timerConnections.put(nameOfTimerSource, connectedTimers);
			}
			if (!connectedTimers.contains(timerDestination)) {
				connectedTimers.add(timerDestination);
			}
		} else {
			timerDestination.setTimerSource(theSource);
		}
	}

	/**
	 * get all the timers' names as a list
	 *
	 * @return a list with the names of all timers
	 */
	public static List<String> getAllTimerNames() {
		return InstanceHolder.instance.getAllTimerNamesInternal();
	}

	/**
	 * internal method to get all the timers' names as a list
	 *
	 * @return a list with the names of all timers
	 */
	private List<String> getAllTimerNamesInternal() {
		List<String> timerNames = new ArrayList<String>();
		for (String timerName : timerNameMapping.keySet()) {
			timerNames.add(timerName);
		}
		return timerNames;
	}

	/**
	 * get the name of a timer
	 *
	 * @param timerSource the timer
	 * @return the name of that timer
	 */
	public static String getName(TimerSource_I timerSource) {
		return InstanceHolder.instance.getNameInternal(timerSource);
	}

	/**
	 * internal method get the name of a timer
	 *
	 * @param timerSource the timer
	 * @return the name of that timer
	 */
	private String getNameInternal(TimerSource_I timerSource) {
		return nameTimerMapping.get(timerSource);
	}

	/**
	 * inform all the listeners that the data has changed
	 */
	private void informTheListeners() {
		refresh = true;
	}

	/**
	 * Adds a listener to the change of the set of timers
	 *
	 * @param listener the listener
	 */
	public static void addRefreshListener(final RefreshListener_I listener) {
		InstanceHolder.instance.refreshListeners.add(listener);
	}

	/**
	 * Removes a listener from the change of the set of timers
	 *
	 * @param listener the listener
	 */
	public static void removeRefreshListener(final RefreshListener_I listener) {
		InstanceHolder.instance.refreshListeners.remove(listener);
	}

	/**
	 * gets the number of bits that a timer with the given name outputs (a
	 * {@link BasicTimer} has 18)
	 *
	 * //@param timerName
	 * the name of the timer
	 *
	 * @return the number of output bits
	 */
	/*public static int getOutputBitsOfTimer(final String timerName) {
		return InstanceHolder.instance.getOutputBitsOfTimerInternal(timerName);
	}

	/**
	 * internal method which gets the number of bits that a timer with the given
	 * name outputs (a {@link BasicTimer} has 18)
	 * 
	 * //@param timerName
	 *            the name of the timer
	 * @return the number of output bits
	 */
	/*private int getOutputBitsOfTimerInternal(final String timerName) {
		TimerSource_I source = timerNameMapping.get(timerName);
		assert (source != null);
		return source.getOutputBitCount();
	}*/
	@Override
	public Collection<RefreshListener_I> getRefreshListeners() {
		return refreshListeners;
	}

	@Override
	public int[] getDirtyAddresses() {
		if (refresh) {
			refresh = false;
			return new int[]{SpartanDefinitions_I.COMPLETE_REFRESH};
		}
		return new int[0];
	}
}
