/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.timer;

/**
 * Every clocking mechanism which could be used as a timer input de.tu_darmstadt.rs.sim.signal should implement the {@link TimerSource_I} interface.
 * <p>This is not restricted to DCM modules but can be applied to basic timers and timer real time interrupt (RTI).
 * The TimerSource_I represents the component whose output is connected to the input of a TimerObserver_I component.
 * E.g. for chaining of timers, it is possible to register a RTI as an observer to an basic timer
 * whereas the basic timer is for example registered to the system clock.
 * Whenever the basic timer reaches its maximum value it informs the registered RTI so the RTI can increment its (internal) value or even interrupt the processor.
 * Remember to inform registered TimerObserver_Is if you implement this interface.
 */
public interface TimerSource_I extends IdentifiableTimer_I {

	/**
	 * Registers a {@linkTimerObserver_I} (BasicTimer or RTI).
	 * <p>Note: A TimerObserver_I should not be registered to more than one TimerSource_I.
	 *
	 * @param timerObserver the next module in the timer chain
	 */
	public void registerTimerObserver(TimerObserver_I timerObserver);

	/**
	 * Unregisters a {@linkTimerObserver_I}.
	 * <p>Afterwards, the timer chain does not contain the TimerObserver_I.
	 * <p>Note: If a TimerObserver_I is unregistered which is not at the end of a chain
	 * the TimerObserver_Is after the unregistered one do not become informed anymore.
	 *
	 * @param timerObserver the {@link TimerObserver_I} to unregister
	 */
	public void unregisterTimerObserver(TimerObserver_I timerObserver);

	/**
	 * get the number of output bits (a TimerObserver_I connects usually to one bit)
	 *
	 * @return number of output bits
	 */
	public int getOutputBitCount();
	/*private void informTimerObservers();*/
}