/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.uart;

import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.JConfigPeripheral;
import de.tu_darmstadt.rs.sim.engine.peripheral.interrupts.InterruptController;
import de.tu_darmstadt.rs.sim.engine.peripheral.interrupts.InterruptSource_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.PipelineClock;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerObserver_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.timer.TimerSource_I;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;

/**
 * Interface for all classes implementing a UART-like peripheral
 */
public abstract class AbstractUart extends JConfigPeripheral implements InterruptSource_I, TimerObserver_I {


	/**
	 * bit group mask for the four baud rate bits
	 */
	public static final int CONTROL_BAUDRATE_MASK = 0x01e00;
	/**
	 * bit group value for baud rate = 115200 Bd
	 */
	public static final int CONTROL_BAUDRATE_115200 = 0x00000; // default value
	/**
	 * bit group value for baud rate = 57600 Bd
	 */
	public static final int CONTROL_BAUDRATE_57600 = 0x00200;
	/**
	 * bit group value for baud rate = 38400 Bd
	 */
	public static final int CONTROL_BAUDRATE_38400 = 0x00400;
	/**
	 * bit group value for baud rate = 31250 Bd
	 */
	public static final int CONTROL_BAUDRATE_31250 = 0x00600;
	/**
	 * bit group value for baud rate = 19200 Bd
	 */
	public static final int CONTROL_BAUDRATE_19200 = 0x00800;
	/**
	 * bit group value for baud rate = 9600 Bd
	 */
	public static final int CONTROL_BAUDRATE_9600 = 0x00a00;
	/**
	 * bit group value for baud rate = 4800 Bd
	 */
	public static final int CONTROL_BAUDRATE_4800 = 0x00c00;
	/**
	 * bit group value for baud rate = 2400 Bd
	 */
	public static final int CONTROL_BAUDRATE_2400 = 0x00e00;
	/**
	 * bit group value for baud rate = 1200 Bd
	 */
	public static final int CONTROL_BAUDRATE_1200 = 0x01000;
	/**
	 * bit group value for baud rate = 600 Bd
	 */
	public static final int CONTROL_BAUDRATE_600 = 0x01200;
	/**
	 * bit group value for baud rate = 300 Bd
	 */
	public static final int CONTROL_BAUDRATE_300 = 0x01400;
	/**
	 * bit group value for baud rate = 150 Bd
	 */
	public static final int CONTROL_BAUDRATE_150 = 0x01600;
	/**
	 * bit group value for baud rate = 75 Bd
	 */
	public static final int CONTROL_BAUDRATE_75 = 0x01800;
	/**
	 * bit group value for baud rate = 50 Bd
	 */
	public static final int CONTROL_BAUDRATE_50 = 0x01a00;

	/**
	 * offset of the STATUS register to the base IO address
	 */
	public static final int STATUS_IO_ADDRESS_OFFSET = 0x0;
	/**
	 * offset of the FIFO_READ register to the base IO address
	 */
	public static final int FIFO_READ_IO_ADDRESS_OFFSET = 0x1;
	/**
	 * offset of the FIFO_WRITE register to the base IO address
	 */
	public static final int FIFO_WRITE_IO_ADDRESS_OFFSET = 0x2;

	/**
	 * bit group mask for the TX FIFO full flag
	 */
	public static final int STATUS_TX_FIFO_FULL = 0x08;
	/**
	 * bit group mask for the TX FIFO empty flag
	 */
	public static final int STATUS_TX_FIFO_EMPTY = 0x04;
	/**
	 * bit group mask for the RX FIFO full flag
	 */
	public static final int STATUS_RX_FIFO_FULL = 0x02;
	/**
	 * bit group mask for the RX FIFO empty flag
	 */
	public static final int STATUS_RX_FIFO_EMPTY = 0x01;

	private static final int PARAM_ID_WRITE_TO_STDOUT = 0;
	private static final int DEFAULT_COUNTER_MAX = 200;
	private static final int UART_FIFO_DEPTH = 7;

	/**
	 * reference to the interrupt controller for receiving interrupt requests
	 */
	protected InterruptController interruptController;

	/**
	 * the address space to be used
	 */
	protected final AddressSpaceManager addressSpace;
	/**
	 * the stream multiplexer for the UART. This holds one input stream and multiple output streams.
	 */
	protected final UartStreamMultiplexer streamable;
	/**
	 * the output stream which prints to standard out
	 */
	protected UartStdout uartStdOut;

	/**
	 * counter for the baud rate simulation
	 */
	protected int counter;
	/**
	 * the maximum counter value for the baud rate simulation
	 */
	protected int counterMax;

	/**
	 * value indicating if an interrupt for the RX de.tu_darmstadt.rs.sim.signal has occurred
	 */
	protected boolean rxInterruptSignal;
	/**
	 * value indicating if an interrupt for the TX de.tu_darmstadt.rs.sim.signal has occurred
	 */
	protected boolean txInterruptSignal;

	/**
	 * the value of the RX register
	 */
	protected int rxShiftRegister;
	/**
	 * the data in the RX FIFO
	 */
	protected int[] fifoRxValues;
	/**
	 * index of the next byte to be read. The value is in the interval [0,8]. 0 means FIFO empty, 8 FIFO full.
	 */
	protected int fifoRxByteIndex;

	/**
	 * the value of the TX register
	 */
	protected int txShiftRegister;
	/**
	 * the data in the TX FIFO
	 */
	protected int[] fifoTxValues;
	/**
	 * index of the next byte to be sent. The value is in the interval [0,8]. 0 means FIFO empty, 8 FIFO full.
	 */
	protected int fifoTxByteIndex;
	/**
	 * the clock source for the baud rate simulation. This should be the pipeline clock.
	 */
	protected TimerSource_I timerSource;

	/**
	 * status register value
	 */
	protected int statusIoValue;

	/**
	 * create an Abstract UART
	 *
	 * @param peri the peripheral hardware specification
	 * @param simulationManager the simulation manager
	 */
	public AbstractUart(final IPeripheralModuleConfiguration peri, final SimulationManager simulationManager, final int baseAddress) {
		super(peri, simulationManager, baseAddress);
		this.addressSpace = simulationManager.getAddressSpaceManager();
		this.streamable = new UartStreamMultiplexer();
		this.counterMax = DEFAULT_COUNTER_MAX;

		setup();
	}

	@Override
	public void setup() {

		// in case, we are already registered -- does not harm anyway:
		TimerManager
				.addTimerConnection(PipelineClock.PIPELINE_CLOCK_NAME, this);
		restart();
	}

	@Override
	public void tearDown() {
		timerSource.unregisterTimerObserver(this);
		super.tearDown();
	}

	@Override
	public void restart() {

		counter = 0;

		rxInterruptSignal = false;
		txInterruptSignal = false;

		rxShiftRegister = INVALID;
		fifoRxValues = new int[8];
		fifoRxByteIndex = -1;
		txShiftRegister = INVALID;
		fifoTxValues = new int[8];
		fifoTxByteIndex = -1;
		if (interruptController != null) {
			interruptController.resetInterruptRequest(this);
		}

		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * for the GUI
	 *
	 * @return true if there is an interrupt request, false otherwise
	 */
	public abstract boolean hasInterruptRequest();

	@Override
	public void setInterruptController(final InterruptController interruptController) {
		this.interruptController = interruptController;
	}

	/**
	 * Retrieves the CONTROL value.
	 *
	 * @return the register value
	 */
	public abstract int getControlIoValue();

	/**
	 * Retrieves the STATUS value.
	 *
	 * @return the register value
	 */
	public int getStatusIoValue() {
		return statusIoValue;
	}

	/**
	 * Retrieves a byte from the Rx FIFO register.
	 *
	 * @param i the byte index
	 * @return the byte value
	 */
	public int getFifoRxValueAt(final int i) {
		return fifoRxValues[i];
	}

	/**
	 * Retrieves a byte from the Tx FIFO register.
	 *
	 * @param i the byte index
	 * @return the byte value
	 */
	public int getFifoTxValueAt(final int i) {
		return fifoTxValues[i];
	}

	/**
	 * Retrieves the current Rx FIFO byte (pointer) index.
	 *
	 * @return the index
	 */
	public int getFifoRxByteIndex() {
		return fifoRxByteIndex;
	}

	/**
	 * Retrieves the current Tx FIFO byte (pointer) index.
	 *
	 * @return the index
	 */
	public int getFifoTxByteIndex() {
		return fifoTxByteIndex;
	}

	/**
	 * Retrieves the content of the Tx shift register.
	 *
	 * @return the register value
	 */
	public int getTxShiftRegister() {
		return txShiftRegister;
	}

	/**
	 * Retrieves the content of the Rx shift register.
	 *
	 * @return the register value
	 */
	public int getRxShiftRegister() {
		return rxShiftRegister;
	}

	/**
	 * Retrieves the (internal!) de.tu_darmstadt.rs.sim.signal for a Rx IRQ. To generate a real
	 * interrupt, the Rx IRQ enable bit must be set, too.
	 *
	 * @return true, if the IRQ conditions are fullfilled
	 */
	public boolean getInternalRxInterruptSignal() {
		return rxInterruptSignal;
	}

	/**
	 * Retrieves the (internal!) de.tu_darmstadt.rs.sim.signal for a Tx IRQ. To generate a real
	 * interrupt, the Tx IRQ enable bit must be set, too.
	 *
	 * @return true, if the IRQ conditions are fullfilled
	 */
	public boolean getInternalTxInterruptSignal() {
		return txInterruptSignal;
	}

	/**
	 * Changes the CONTROL register value.
	 *
	 * @param value the new value
	 */
	public abstract void setControlIoValue(final int value);

	@Override
	public String getInterruptName() {
		return getPeriName() + " IRQ";
	}

	/**
	 * set the source of the UART input
	 *
	 * @param source the source of the UART
	 */
	public void setInputSource(UartStream_I source) {
		streamable.setUartInputSource(source);
	}

	@Override
	public String[] getParameters() {
		String[] ret = new String[]{"writeToStdOut"};
		return ret;
	}

	@Override
	public boolean setParameter(final int parId, final int value) {
		switch (parId) {
			case PARAM_ID_WRITE_TO_STDOUT:
				setWriteToStdOut(value != 0);
				return true;
		}
		return false;
	}

	/**
	 * add a destination to the UART output
	 *
	 * @param destination a new destination
	 */
	public void addOutputDestination(UartStream_I destination) {
		streamable.addUartOutputDestination(destination);
	}

	/**
	 * remove the source of the UART input
	 */
	public void removeInputSource() {
		streamable.setUartInputSource(null);
	}

	/**
	 * remove a destination from the UART output
	 *
	 * @param destination a destination to be removed
	 */
	public void removeOutputDestination(UartStream_I destination) {
		streamable.removeUartOutputDestination(destination);
	}

	/**
	 * set if the UART should write to standard output
	 *
	 * @param shouldWrite true if the UART should write to standard output, false
	 * otherwise
	 */
	public void setWriteToStdOut(boolean shouldWrite) {
		if (shouldWrite) {
			if (uartStdOut == null) {
				uartStdOut = new UartStdout();
			}
			addOutputDestination(uartStdOut);
			System.out.println(getPeriName() + " write to stdOut set");
		} else {
			removeOutputDestination(uartStdOut);
		}
	}

	@Override
	public abstract int getRegisterValue(int offset);

	@Override
	public abstract void setRegisterValue(int offset, int value);

	/**
	 * Puts a byte to the Tx FIFO register.
	 *
	 * @param value the new byte
	 */
	protected void addByteToFifoTx(final int value) {

		statusIoValue &= ~STATUS_TX_FIFO_EMPTY; // definetely not empty
		// (anymore)

		if (fifoTxByteIndex < UART_FIFO_DEPTH) {
			fifoTxValues[++fifoTxByteIndex] = value;
			if (fifoTxByteIndex == UART_FIFO_DEPTH) {
				statusIoValue |= STATUS_TX_FIFO_FULL;
			}

		} else {
			assert (fifoTxByteIndex == UART_FIFO_DEPTH);
			// Tx FIFO is full >> new data is lost
		}

	}

	/**
	 * Removes a byte from the Tx FIFO register.
	 *
	 * @return the removed byte
	 */
	protected int removeByteFromFifoTx() {

		if (fifoTxByteIndex > -1) {

			final int r = fifoTxValues[0];
			assert ((r & ~BITMASK_8) == 0);

			System.arraycopy(fifoTxValues, 1, fifoTxValues, 0, UART_FIFO_DEPTH);
			fifoTxByteIndex--;
			statusIoValue &= ~STATUS_TX_FIFO_FULL; // definetely not full
			// (anymore)

			if (fifoTxByteIndex == -1) {
				// Tx FIFO empty
				statusIoValue |= STATUS_TX_FIFO_EMPTY;
			}
			return r;

		} else {
			assert (fifoTxByteIndex == -1);
			// Tx FIFO is empty

			// never end up here:
			// the only function which uses removeByteFromFifoTx() is onClock(),
			// and we check
			// the FIFO before we remove any byte
			assert (false);
			return 0;
		}
	}

	/**
	 * Puts a byte in the FIFO Rx register.
	 *
	 * @param value the new byte
	 */
	protected void addByteToFifoRx(final int value) {

		statusIoValue &= ~STATUS_RX_FIFO_EMPTY; // definetely not empty
		// (anymore)

		if (fifoRxByteIndex < UART_FIFO_DEPTH) {

			System.arraycopy(fifoRxValues, 0, fifoRxValues, 1, UART_FIFO_DEPTH);
			fifoRxValues[0] = value;
			fifoRxByteIndex++;

			if (fifoRxByteIndex == UART_FIFO_DEPTH) {
				// Rx FIFO full
				statusIoValue |= STATUS_RX_FIFO_FULL;
			}

		} else {
			assert (fifoRxByteIndex == UART_FIFO_DEPTH);
			// Rx FIFO is full >> new data is lost
		}
	}

	/**
	 * Removes a byte from the FIFO Rx register.
	 *
	 * @return the byte
	 */
	protected int removeByteFromFifoRx() {

		statusIoValue &= ~STATUS_RX_FIFO_FULL; // definetely not full (anymore)

		if (fifoRxByteIndex > -1) {
			int r = fifoRxValues[fifoRxByteIndex--];
			assert ((r & ~BITMASK_8) == 0);
			if (fifoRxByteIndex == -1) {
				statusIoValue |= STATUS_RX_FIFO_EMPTY;
			}

			return r;

		} else {
			assert (fifoRxByteIndex == -1);
			// Rx FIFO is empty >> no byte to read
			// throw new SpartanException("Cannot read from empty Rx FIFO");
			return SpartanDefinitions_I.INVALID;
		}
	}

	/**
	 * Changes a byte in the Rx FIFO.
	 *
	 * @param index the byte's index
	 * @param value the new value
	 */
	public void setFifoRxValueAt(final int index, final int value) {

		assert ((value & ~BITMASK_8) == 0);
		fifoRxValues[index] = value;

		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * Changes a byte in the Tx FIFO.
	 *
	 * @param index the byte's index
	 * @param value the new value
	 */
	public void setFifoTxValueAt(final int index, final int value) {

		assert ((value & ~BITMASK_8) == 0);
		fifoTxValues[index] = value;

		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * Retrieves the baudrate counter.
	 *
	 * @return the counter's value
	 */
	public int getCounterMax() {
		return counterMax;
	}

	/**
	 * Changes the baudrate counter.
	 *
	 * @param value the new counter value
	 */
	public void setCounterMax(final int value) {
		assert (value >= 1);
		counterMax = value;
		counter = 0;

		fireRefreshEvent(COMPLETE_REFRESH);
	}

	@Override
	public void setTimerSource(TimerSource_I timerSource) {
		this.timerSource = timerSource;
		timerSource.registerTimerObserver(this);
	}
}
