/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.peripheral.uart;

import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IPeripheralModuleConfiguration;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.handlers.IIntegerParameterHandler;

/**
 * This UART class simulates a UART hardware interface. It has a status
 * register, a control register and two FIFO buffers. The Uart uses a
 * {@link UartStream_I} interface to simulate IO behavior, i.e. to receive and
 * to transmit bytes.
 */
public class UartLight extends AbstractUart {

	/**
	 * bit group mask for the RX interrupt enable flag
	 */
	public final static int STATUS_RX_IRQ_ENABLE = 0x02000;
	/**
	 * bit group mask for the TX interrupt enable flag
	 */
	public final static int STATUS_TX_IRQ_ENABLE = 0x04000;

	/**
	 * create a new {@link UartLight}.
	 *
	 * @param peri the {@link IPeripheralModuleConfiguration} to get hardware information from
	 * @param simulationManager reference to the {@link SimulationManager}
	 */
	public UartLight(final IPeripheralModuleConfiguration peri, final SimulationManager simulationManager, int baseAddress) {
		super(peri, simulationManager, baseAddress);
	}

	@Override
	public void restart() {

		statusIoValue = STATUS_TX_FIFO_EMPTY | STATUS_RX_FIFO_EMPTY;
		super.restart();
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	@Override
	public void importHardwareParameters(IPeripheralModuleConfiguration node) {
		super.importHardwareParameters(node);
		IIntegerParameterHandler parameter = (IIntegerParameterHandler) node.findParameterDescription("BAUDRATE").get();
		counterMax = (int)(long) parameter.getIntegerValue().orElseThrow();
	}


	/**
	 * check if there are interrupts
	 */
	private void calculateInterrupt() {

		// ////////////////////////////////////////////////////////////////////////////////////////////
		// 1) Rx interrupt:

		if ((statusIoValue & STATUS_RX_FIFO_EMPTY) == 0) {
			// Rx is NOT empty >> set the interrupt flag
			rxInterruptSignal = true;

			fireRefreshEvent(COMPLETE_REFRESH);
		}
		// Note: interrupt de.tu_darmstadt.rs.sim.signal will be deleted on any Rx FIFO read access,
		// see getValueAt()

		// ///////////////////////////////////////////////////////////////////////////////////////////
		// 2) Tx interrupt

		if ((statusIoValue & STATUS_TX_FIFO_FULL) == 0) {

			// Tx is NOT full >> set the interrupt flag
			txInterruptSignal = true;

			fireRefreshEvent(COMPLETE_REFRESH);
		}
		// Note: interrupt de.tu_darmstadt.rs.sim.signal will be deleted on any STATUS read access or
		// any Tx FIFO write access,
		// see getValueAt() and setValueAt()

		if (interruptController != null) {
			if ((((statusIoValue & STATUS_RX_IRQ_ENABLE) != 0) && rxInterruptSignal) || (((statusIoValue & STATUS_TX_IRQ_ENABLE) != 0) && txInterruptSignal)) {
				interruptController.postInterruptRequest(this);
			} else {
				interruptController.resetInterruptRequest(this);
			}
		}
	}

	@Override
	public boolean hasInterruptRequest() {
		return (((statusIoValue & STATUS_RX_IRQ_ENABLE) != 0) && rxInterruptSignal) || (((statusIoValue & STATUS_TX_IRQ_ENABLE) != 0) && txInterruptSignal);
	}

	/**
	 * Changes the Status register value.
	 *
	 * @param value the new value
	 */
	public void setStatusIoValue(final int value) {
		assert ((value & ~BITMASK_18) == 0);
		statusIoValue = value;

		fireRefreshEvent(COMPLETE_REFRESH);
	}

	/**
	 * Cuts - if necessary - leading digits according to the selected word
	 * length.
	 *
	 * @param value the byte which may be cut off
	 * @return the last 5, 6, 7 or all 8 bits
	 */
	private int cutOff(final int value) /* throws SpartanException */ {

		assert ((value & ~BITMASK_8) == 0);

		return value; // default = 8 bit length
	}

	@Override
	public int readRegisterValue(int offset) {
		int unchangedRegisterValue = getRegisterValue(offset);
		switch (offset) {
			case STATUS_IO_ADDRESS_OFFSET:
				txInterruptSignal = false;
				fireRefreshEvent(COMPLETE_REFRESH);
				return unchangedRegisterValue;
			case FIFO_READ_IO_ADDRESS_OFFSET:
				final int rxData = removeByteFromFifoRx();
				rxInterruptSignal = false; // important!

				fireRefreshEvent(COMPLETE_REFRESH);
				return rxData;
			default:
				return 0;
		}
	}

	@Override
	public void writeRegisterValue(int offset, int value) {
		switch (offset) {
			case FIFO_WRITE_IO_ADDRESS_OFFSET:
				addByteToFifoTx((value & BITMASK_8));

				//			txInterruptSignal = false; // important!
				break;
			default:
				setRegisterValue(offset, value);
				break;
		}
		fireRefreshEvent(COMPLETE_REFRESH);
	}

	@Override
	public void onClock(int clockValue) {
		boolean pleaseRefreshAfterwards = false;

		calculateInterrupt();

		// //////////////////////////////////////////////////////////////////////////
		// simulate a "baudrate" with a counter:
		counter++;

		if ((rxShiftRegister != INVALID)) {
			// shift register has something for the FIFO
			assert ((rxShiftRegister & ~BITMASK_8) == 0);
			addByteToFifoRx(rxShiftRegister);
			rxShiftRegister = INVALID;
			pleaseRefreshAfterwards = true;
			// Note: moving a byte from the shift register to the FIFO is
			// independent from any
			// baudrate -- that's why it must be executed here, i.e. before the
			// counter-check
		}

		if (counter == counterMax) {

			counter = 0;

			// //////////////////////////////////////////////////////////////////////////
			// reception:
			if (streamable.hasNextByte()) {
				// there is something to be received
				final int data = cutOff(streamable.onReceiveNextByte()
						& BITMASK_8);
				rxShiftRegister = data;
				pleaseRefreshAfterwards = true;

			} else {
				// no data to receive
				assert (rxShiftRegister == INVALID);
			}

			// //////////////////////////////////////////////////////////////////////////
			// transmission:
			if (txShiftRegister != INVALID) {
				// shift register has something to be sent
				assert ((txShiftRegister & ~BITMASK_8) == 0);
				streamable
						.onTransmitNextByte((byte) cutOff(txShiftRegister));
			}

			if ((statusIoValue & STATUS_TX_FIFO_EMPTY) == 0) {
				// Tx FIFO not empty, i.e. something there
				// to be pushed into the shift register
				final int data = removeByteFromFifoTx();
				assert ((data & ~BITMASK_8) == 0);
				txShiftRegister = data;
			} else {
				// nothing more to transmit >> shift register is empty
				txShiftRegister = INVALID;
			}
			pleaseRefreshAfterwards = true;

		} // end if counter

		if (pleaseRefreshAfterwards) {
			fireRefreshEvent(COMPLETE_REFRESH);
		}
	}

	@Override
	public int getRegisterValue(int offset) {
		switch (offset) {
			case STATUS_IO_ADDRESS_OFFSET:
				return statusIoValue;
			case FIFO_READ_IO_ADDRESS_OFFSET:
				if (fifoRxByteIndex >= 0) {
					return fifoRxValues[fifoRxByteIndex];
				} else {
					return 0;
				}
			case FIFO_WRITE_IO_ADDRESS_OFFSET:
				if (fifoTxByteIndex >= 0) {
					return fifoTxValues[fifoTxByteIndex];
				} else {
					return 0;
				}
			default:
				return 0;
		}
	}

	@Override
	public void setRegisterValue(int offset, int value) {
		switch (offset) {
			case STATUS_IO_ADDRESS_OFFSET:
				statusIoValue = value;
				break;
			case FIFO_WRITE_IO_ADDRESS_OFFSET:
				if (fifoTxByteIndex >= 0) {
					fifoTxValues[fifoTxByteIndex] = value;
				}
				break;
		}
	}

	@Override
	public int getControlIoValue() {
		return 0;
	}

	@Override
	public void setControlIoValue(int value) {
		// unused
	}

}
