/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
*/

package de.tu_darmstadt.rs.sim.engine.peripheral.uart;

import java.util.Collection;
import java.util.HashSet;

/**
 * This class allows an uart's data output to
 * any number of destinations and to switch between
 * several sources for uart input.
 *
 * @author Markus Vogt
 * @version 31.08.2011
 */
public class UartStreamMultiplexer implements UartStream_I {

	private UartStream_I fUartInputSource;
	private Collection<UartStream_I> fUartOutputDestinations;

	/**
	 * create a new {@link UartStreamMultiplexer}
	 */
	public UartStreamMultiplexer() {
		fUartOutputDestinations = new HashSet<UartStream_I>();
	}

	/**
	 * Sets the data input source that should be used
	 * for transmitting data from the uart
	 *
	 * @param source the new source
	 */
	public void setUartInputSource(UartStream_I source) {
		fUartInputSource = source;
	}

	/**
	 * Adds an output destination to this multiplexer
	 * that will receive all data received from the uart.
	 *
	 * @param destination Destination to add
	 */
	public void addUartOutputDestination(UartStream_I destination) {
		fUartOutputDestinations.add(destination);
	}

	/**
	 * Removes an output destination from this multiplexer.
	 *
	 * @param destination Destination to remove
	 */
	public void removeUartOutputDestination(UartStream_I destination) {
		fUartOutputDestinations.remove(destination);
	}

	/**
	 * Retrieves the next incoming byte from the
	 * current uart input source.
	 *
	 * @return a byte
	 */
	@Override
	public byte onReceiveNextByte() {
		return fUartInputSource.onReceiveNextByte();
	}


	/**
	 * Checks if there is any byte to be received on the
	 * currently selected uart input source.
	 *
	 * @return true, if at least one byte can be received
	 */
	@Override
	public boolean hasNextByte() {
		return (fUartInputSource != null) && (fUartInputSource.hasNextByte());
	}

	/**
	 * Puts the next byte out to any of the registered
	 * uart output destinations.
	 *
	 * @param nextByte any byte
	 */
	@Override
	public void onTransmitNextByte(byte nextByte) {
		for (UartStream_I dst : fUartOutputDestinations) {
			dst.onTransmitNextByte(nextByte);
		}
	}


}
