/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.processor;

import de.tu_darmstadt.rs.sim.engine.RefreshHandler_I;
import de.tu_darmstadt.rs.sim.engine.SimulationComponent_I;
import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.interrupts.InterruptController;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.PipelineUI;

import java.util.Collection;

/**
 * Processor core simulation class with an interrupt controller embedded (for
 * performance reasons there exists also an processor core without interrupt
 * controller called {@link SimpleProcessorCore}). It contains a
 * {@link SimpleProcessorCore} which does the instruction processing while this
 * class is only responsible for the additional interrupt handling.
 */
public class InterruptableProcessorCore implements ProcessorCore_I {

	private final SimpleProcessorCore simpleCore;
	private final SfRegisters sfr;
	private InterruptController irqController;

	/**
	 * create a new {@link InterruptableProcessorCore}
	 *
	 * @param simMgr reference to the {@link SimulationManager}
	 * @param simpleCore the {@link SimpleProcessorCore}
	 */
	public InterruptableProcessorCore(SimulationManager simMgr,
									  SimpleProcessorCore simpleCore) {
		this.sfr = simMgr.getSfRegisters();
		this.simpleCore = simpleCore;
		setInterruptController(simMgr.getInterruptController());
	}

	@Override
	public void setInterruptController(InterruptController irqController) {
		this.irqController = irqController;
		simpleCore.setInterruptController(irqController);
	}

	@Override
	public void onExecuteStep(final boolean alwaysFalse) {
		// check if irq ctrl wants to interrupt & if irq's are enabled
		final boolean doInterrupt = irqController.hasInterrupt() && areInterruptsEnabled();

		simpleCore.onExecuteStep(doInterrupt);
	}

	@Override
	public void onResetSimulation() {
		//unused
	}

	private final boolean areInterruptsEnabled() {
		return (sfr.getInterruptEnable() == 1);
	}

	@Override
	public boolean isAddressInDelayWindow(int address) {
		return simpleCore.isAddressInDelayWindow(address);
	}

	@Override
	public PipelineUI getPipelineUI() {
		return simpleCore.getPipelineUI();
	}

	@Override
	public Collection<SimulationComponent_I> getSimulationComponents() {
		return simpleCore.getSimulationComponents();
	}

	@Override
	public void setRefreshHandler(RefreshHandler_I refreshHandler) {
		simpleCore.setRefreshHandler(refreshHandler);
	}

}
