/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.engine.processor;

import de.tu_darmstadt.rs.sim.engine.AccessManager;
import de.tu_darmstadt.rs.sim.engine.RefreshHandler_I;
import de.tu_darmstadt.rs.sim.engine.Settings;
import de.tu_darmstadt.rs.sim.engine.SimulationComponent_I;
import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.peripheral.interrupts.InterruptController;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.ExMemStage;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.IfIdStage;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.PipelineUI;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.WbStage;

import java.util.ArrayList;
import java.util.Collection;

/**
 * Processor core class which handles the basic instruction processing with
 * pipeline stages.
 */
public class SimpleProcessorCore implements ProcessorCore_I {

	private final IfIdStage ifId;
	private final ExMemStage exMem;
	private final WbStage wb;
	private final SfRegisters sfr;

	private final int[] currentWindowStages;
	private final boolean[] isJump;
	private static final int STAGE_IFID = 0;
	private static final int STAGE_EXMEM = 1;
	private static final int STAGE_WB = 2;
	private PipelineUI pipeUI;

	/**
	 * create a new {@link SimpleProcessorCore}
	 *
	 * @param simMgr reference to the {@link SimulationManager}
	 * @param accessManager reference to the {@link AccessManager}
	 */
	public SimpleProcessorCore(SimulationManager simMgr,
							   AccessManager accessManager) {
		this.sfr = simMgr.getSfRegisters();
		ifId = new IfIdStage(simMgr.getAddressSpaceManager(),
				simMgr.getRegisterFile(), simMgr.getInterruptController(),
				simMgr.getSfRegisters(), accessManager);
		exMem = new ExMemStage(ifId, simMgr.getAddressSpaceManager(),
				simMgr.getSfRegisters(), simMgr.getRegisterFile(),
				accessManager);
		wb = new WbStage(exMem, simMgr.getRegisterFile());
		currentWindowStages = new int[3];
		isJump = new boolean[3];
	}

	@Override
	public void onExecuteStep(final boolean doInterrupt) {
		final int wbStageWindow = getDelayWbStageWindow();
		wb.executeStep(wbStageWindow); // may throw SpartanException

		// Note: IfId needs the cc register parallelly to ExMem. Changes on cc
		// caused by ExMem do NOT infer
		// IfId, that's why we need to remember the "old" cc.
		final int old_cc = sfr.getCC();

		final int ifIdStageWindow = getDelayIfIdStageWindow();
		exMem.executeStep(wb.getOverwrittenRegNo(), wb.getOverwrittenRegValue()); // may
		// throw
		final int new_cc = sfr.getCC();                                                                            // SpartanException

		ifId.executeStep(old_cc, new_cc, doInterrupt, ifIdStageWindow); // may throw SpartanException
		currentWindowStages[STAGE_WB] = currentWindowStages[STAGE_EXMEM];
		currentWindowStages[STAGE_EXMEM] = currentWindowStages[STAGE_IFID];
		currentWindowStages[STAGE_IFID] = sfr.getWindow();
		isJump[STAGE_WB] = isJump[STAGE_EXMEM];
		isJump[STAGE_EXMEM] = isJump[STAGE_IFID];
		isJump[STAGE_IFID] = ifId.hasDelayedJump();

		pipeUI.requestRefresh(SpartanDefinitions_I.COMPLETE_REFRESH); // let the
		// Pipeline
		// Refresh
		// Listeners
		// (GUI)
		// refresh
	}

	/**
	 * get the window for the wb stage. The window may be delayed if there was a
	 * delayed jump before.
	 *
	 * @return the window for the wb stage
	 */
	public int getDelayWbStageWindow() {
		return isJump[STAGE_WB]? currentWindowStages[STAGE_WB] : sfr
				.getWindow();
	}

	private int getDelayIfIdStageWindow() {
		return isJump[STAGE_EXMEM]? currentWindowStages[STAGE_IFID] : sfr.getWindow();
	}

	@Override
	public void onResetSimulation() {
		currentWindowStages[STAGE_IFID] = 0;
		currentWindowStages[STAGE_EXMEM] = 0;
		currentWindowStages[STAGE_WB] = 0;
		isJump[STAGE_IFID] = false;
		isJump[STAGE_EXMEM] = false;
		isJump[STAGE_WB] = false;

		ifId.initiate((Integer) Settings.getSetting(Settings.PC_INITIAL));
		ifId.setFrame((Integer) Settings.getSetting(Settings.PC_FRAME));
	}

	@Override
	public boolean isAddressInDelayWindow(int address) {
		final int localRegExMem = address
				- (isJump[STAGE_EXMEM]? currentWindowStages[STAGE_EXMEM] : sfr
				.getWindow()) * 8;
		final boolean isLocalRegInExMem = (localRegExMem >= 4)
				&& (localRegExMem <= 15);
		final int localRegWb = address - getDelayWbStageWindow() * 8;
		final boolean isLocalRegInWb = (localRegWb >= 4) && (localRegWb <= 15);
		return isLocalRegInExMem || isLocalRegInWb;
	}

	@Override
	public void setInterruptController(InterruptController irqController) {
		ifId.setInterruptController(irqController);
	}

	@Override
	public PipelineUI getPipelineUI() {
		return pipeUI;
	}

	@Override
	public Collection<SimulationComponent_I> getSimulationComponents() {
		ArrayList<SimulationComponent_I> componentList = new ArrayList<SimulationComponent_I>(
				3);
		componentList.add(ifId);
		componentList.add(exMem);
		componentList.add(wb);
		return componentList;
	}

	@Override
	public void setRefreshHandler(final RefreshHandler_I refreshHandler) {
		pipeUI = new PipelineUI(refreshHandler, ifId, exMem, wb, sfr);
	}

}
