/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 *
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 *
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

/**
 * Spartan MC Simulator
 * Stefan Spickereit
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2006
 */
package de.tu_darmstadt.rs.sim.engine.statistics;

import de.tu_darmstadt.rs.sim.engine.RefreshHandler_I;
import de.tu_darmstadt.rs.sim.engine.processor.Instruction;
import de.tu_darmstadt.rs.sim.engine.processor.InstructionType;
import de.tu_darmstadt.rs.sim.model.StatisticsModel;
import de.tu_darmstadt.rs.sim.signal.Signal1;
import de.tu_darmstadt.rs.sim.signal.SignalModified;

/**
 * The pipeline (instruction) statistics are recorded and administered by the
 * Statistics class. It stores single instructions and consecutive instructions
 * (predecessor-successor-pairs).
 */
public class Statistics implements Statistics_I {

	private long totalStepsExecuted;
	private final long[] instructionsCount;
	private int lastInstructionOrd;
	private final long[][] contiguousInstructions; // [predecessor][successor]

	private final SignalModified fSigModified;
	private final Signal1<StatisticsModel> sigStatisticsData;

	/**
	 * Constructor.
	 *
	 * @param refreshHandler
	 *            a {@link RefreshHandler_I} to which refresh requests should be
	 *            sent
	 */
	public Statistics(final RefreshHandler_I refreshHandler) {
		fSigModified = new SignalModified();
		fSigModified.registerFlushHandler(this, "flush");
		sigStatisticsData = new Signal1<StatisticsModel>();
		refreshHandler.registerDataProducerChangeSignal(fSigModified);
		instructionsCount = new long[InstructionType.values().length];
		contiguousInstructions = new long[InstructionType.values().length][InstructionType
				.values().length];
		restart();
	}

	/**
	 * Clears all statistics.
	 */
	@Override
	public void restart() {
		final int len = InstructionType.values().length;
		for (int i = 0; i < len; i++) {
			instructionsCount[i] = 0L;
			for (int j = 0; j < len; j++) {
				contiguousInstructions[i][j] = 0L;
			}
		}
		totalStepsExecuted = 0;
		lastInstructionOrd = -1;
		fSigModified.emit();
	}

	@Override
	public void setup() {
		// nothing to do here
	}

	@Override
	public void addInstruction(final Instruction instruction) {
		if (instruction == null) {
			return;
		}
		final int ord = instruction.type.ordinal();
		instructionsCount[ord]++;
		totalStepsExecuted++;

		if (lastInstructionOrd != -1) {
			contiguousInstructions[lastInstructionOrd][ord]++;
		}

		lastInstructionOrd = ord;
		fSigModified.emit();
	}

	@Override
	public void flush() {
		StatisticsModel sm = new StatisticsModel();
		sm.totalInstrCount = totalStepsExecuted;
		sm.instrCountByType = instructionsCount.clone();
		sm.instrCountByPair = contiguousInstructions.clone();
		for (int i = 0; i < sm.instrCountByPair.length; i++) {
			sm.instrCountByPair[i] = sm.instrCountByPair[i].clone();
		}
		sigStatisticsData.emit(sm);
	}

	/**
	 * register a handler method for new data (method must have exactly 1
	 * parameter/argument of type {@link StatisticsModel}
	 *
	 * @param o
	 *            the object having the method
	 * @param dataHandlerMethodName
	 *            name of the handler method
	 * @param argClass
	 *            the type of the parameter of the method
	 */
	public void registerDataHandlerMethod(final Object o,
										  final String dataHandlerMethodName,
										  final Class<StatisticsModel> argClass) {
		sigStatisticsData.registerMethodCallOnEmit(o, dataHandlerMethodName,
				argClass);
	}

	/**
	 * unregister a previously registered handler method for data
	 *
	 * @param o
	 *            the object having the method
	 */
	public void unregisterDataHandlerMethod(final Object o) {
		sigStatisticsData.unregisterMethodCallOnEmit(o);
	}

}
