/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gdb;

import java.util.Arrays;
import java.util.Vector;

/**
 * Created by heid on 10.12.15.
 */
public enum Command {
	qSupported("^qSupported:.*"), //OK
	OK("^Hg0|^Hc-1"), //OK
	qTStatus("^qTStatus$"), //OK
	questionMark("^\\?$"), //OK
	qfThreadInfo("^qfThreadInfo$|^qsThreadInfo$"), //OK
	qAttached("^qAttached[:]?"), //OK
	vCont("^vCont?$"),
	qSymbol("^qSymbol::$"),
	qOffsets("^qOffsets$"),
	qC("^qC$"), //OK
	g("^g$"), //OK
	G("^G[1-9a-fA-F]+"),
	p("^p[0-9a-fA-F]{1,6}"), //OK
	P("^P[0-9a-fA-F]+=[0-9a-fA-F]+"),
	m("^m[0-9a-fA-F]+,[0-9a-fA-F]+"), //OK
	M("^M[0-9a-fA-F]+,[0-9a-fA-F]+:[0-9a-fA-F]+"), //OK
	c("^c[0-9a-fA-F]*|C[0-9a-fA-F]*"),
	s("^s[0-9a-fA-F]*|S[0-9a-fA-F]*"),
	z("^z[0-4],[0-9a-fA-F]+,[0-9a-fA-F]+"),
	Z("^Z[0-4],[0-9a-fA-F]+,[0-9a-fA-F]+"),
	w("^w[0-124]$"),
	k("^k$"), //OK
	r("^r$"),
	qT("^qT[f|s][V|P]$"), //OK
	NULL("XXXXXXXXXXXX"); //OK

	private static final int RADIX16 = 16;

	private final String regex;

	Command(final String regex) {
		this.regex = regex;
	}

	/**
	 * @param cmdAsString a string to be transformed to a commend
	 * @return command if it is described in the above list of commands,otherwise null
	 */
	public static Command find(String cmdAsString) {
		for (Command p : Command.values()) {
			if (cmdAsString.matches(p.regex)) {
				return p;
			}
		}
		return NULL;
	}

	/**
	 * @param pkt find the according command to a packet
	 * @return command if it is described in the above list of commands,otherwise null
	 */
	public static Command find(Packet pkt) {
		try {
			return find(pkt.getBody());
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	/**
	 * @param p the packet to be extracted
	 * @return the first field of the packet as an integer number
	 */
	public static int getFirstFieldtoInt(Packet p) {
		int ret = Integer.valueOf(getFirstField(p), RADIX16).intValue();
		return ret;
	}

	/**
	 * @param p  the packet to be extracted
	 * @return first element of the packet
	 */
	public static String getFirstField(Packet p) {
		Vector<String> sv = getFields(p);
		return sv.firstElement();
	}

	/**
	 * splits packet into multiple strings
	 *
	 * @param p  the packet to be extracted
	 * @return vector of strings containing the request,address,and sometimes data after removing all commas
	 */
	public static Vector<String> getFields(Packet p) {
		Vector<String> ret = null;
		String body = null;
		try {
			body = p.getBody();
		} catch (Exception e) {
			e.printStackTrace();
		}
		Command cmd = find(body);
		switch (cmd) {
			case qSupported: {
				body = removeCmdBeforeColon(body);
				ret = split(";", body);
				break;
			}
			case qAttached: {
				body = removeCmdBeforeColon(body);
				ret = toStringVector(body);
				break;
			}
			case w:
			case s:
			case c:
			case m:
			case M:
			case G:
			case p:
			case P: {
				//remove first letter the rest is the first field
				body = body.substring(1);
				ret = toStringVector(body);
				break;
			}

			case z:
			case Z: {
				body = body.substring(1);
				ret = split(",", body);
				break;
			}
			default:
				return null;
		}
		return ret;
	}

	private static String removeCmdBeforeColon(String s) {
		String[] sa = s.split(":", 1);
		if (sa.length == 2) {
			return sa[1];
		}
		return null;
	}

	private static Vector<String> split(String splitChar, String str) {
		return new Vector<String>(Arrays.asList(str.split(splitChar)));
	}

	private static Vector<String> toStringVector(String str) {
		String[] s = new String[]{str};
		return new Vector<String>(Arrays.asList(s));
	}

	private static Vector<String> toStringVector(String[] str) {
		return new Vector<String>(Arrays.asList(str));
	}


}

