/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gdb;


import org.slf4j.Logger;

import java.util.Vector;

/**
 * Created by heid on 08.12.15.
 */
public final class CommandHandler {

	/**
	 * the class variables are the Server and the SimulationManager to enable communication between
	 * simulator and gdb (or any other client)
	 * The Server is the instance enabling the transfer of packets
	 * The SimulationManager is the variable accessing and forwarding the data to the server
	 * The logger variable is only added for the programmer to monitor the packet transfer
	 */
	private static final Logger logger = org.slf4j.LoggerFactory.getLogger(CommandHandler.class);
	private static final SimulationManagerGDBInterface SIMULATION_MANAGER_GDB_INTERFACE = SimulationManagerGDBInterface.getInstance();
	private static final int RADIX16 = 16;
	private static final int REG_AMOUNT = 16;
	private static final String ERROR_REPLY = "E 00";

	private CommandHandler(){}

	/**
	 * filters the received packet, if command is not supported the response is empty,
	 * else if the command is supported the appropriate method is called and the packet is transferred
	 * to the method in order to extract the needed data for processing
	 *
	 * @param pkt a received packet
	 */

	public static void process(Packet pkt) {
		Command cmd = Command.find(pkt);
		switch (cmd) {

			case OK: {
				logger.info("OK");
				sendWithAck(PacketBuilder.OK());
				break;
			}

			case questionMark: {
				logger.info(PacketBuilder.stopReason().toString());
				sendWithAck(PacketBuilder.stopReason());
				break;
			}

			case g: {
				readGeneralRegisters(pkt);
				break;
			}
			case G: {
				writeGeneralRegisters(pkt);
				break;
			}
			case p: {
				readRegister(pkt);
				break;
			}
			case P: {
				writeRegister(pkt);
				break;
			}
			case m: {
				readMemory(pkt);
				break;
			}
			case M: {
				writeMemory(pkt);
				break;
			}
			case c: {
				continueExec(pkt);
				break;
			}
			case s: {
				step(pkt);
				break;
			}
			case z: {
				clearBreakOrWatchpoint(pkt);
				break;
			}
			case Z: {
				setBreakOrWatchpoint(pkt);
				break;
			}
			case w: {
				readRegisterWindow(pkt);
				break;
			}
			case k: {
				kill();
				break;
			}
			case r: {
				reset();
				break;
			}
			default: {
				logger.info("send empty packet");
				sendWithAck(PacketBuilder.emptyPacket());
				break;
			}
		}
	}

	/**
	 * Read the values stored in memory starting at the given address:
	 * numberOfBytes is the addressable memory length in SpartanMC
	 * each address contains 18 bits which is then converted into 4 bytes
	 * @param pkt a received packet
	 */

	private static void readMemory(Packet pkt) {
		String region = Command.getFirstField(pkt);
		String[] ra = region.split(",");
		String addr = ra[0];
		String length = ra[1];
		int address = Integer.parseInt(addr, RADIX16);
		int numberOfBytes = Integer.parseInt(length, RADIX16);
		String memdata;
		String data = null;
		for (int i = 0; i < numberOfBytes; i += 4) {
			memdata = Integer.toHexString(SIMULATION_MANAGER_GDB_INTERFACE.getMemory(address));
			address += 2;
			if (i == 0) {
				data = adaptdatalength(memdata);
			} else {
				data = data + adaptdatalength(memdata);
			}
		}
		logger.info("Read memory at {} for {} with {}", Integer.parseInt(addr, RADIX16), numberOfBytes, data);
		sendWithAck(data);
	}

	/**
	 * Write the given hex values starting at the given address:
	 * numberOfBytes is the addressable memory length in SpartanMC
	 * each address contains 18 bits which is then converted into 4 bytes
	 *
	 * @param pkt a received packet
	 */
	private static void writeMemory(Packet pkt) {
		String region = Command.getFirstField(pkt);
		String[] ra = region.split(":");
		String data = ra[1];
		ra = ra[0].split(",");
		String addr = ra[0];
		String length = ra[1];
		String memdata;
		String temp;
		int address = Integer.parseInt(addr, RADIX16);
		int numberOfBytes = Integer.parseInt(length, RADIX16);
		for (int i = 0; i < numberOfBytes * 2; i += 8) {
			temp = data;
			memdata = temp.substring(i, i + 8);
			SIMULATION_MANAGER_GDB_INTERFACE.setMemory(address, Integer.parseInt(memdata, RADIX16));

		}

		logger.info("Write memory at {} for {} with {}", Integer.parseInt(addr, RADIX16), numberOfBytes, data);
		sendWithAck(PacketBuilder.OK());
	}
	/**
	 * Write the given binary values starting at the given address:
	 * numberOfBytes is the addressable memory length in SpartanMC
	 * each address contains 18 bits which is then converted into 4 bytes
	 */
	/*private static void writeDataToMemory(Packet pkt) {
		Vector<String> fields = Command.getFields(pkt);
        String addr = fields.get(0);
        String length = fields.get(1);
        //TODO:DONE Write Data to memory
        int address=Integer.parseInt(addr,16);
        int numberOfBytes=Integer.parseInt(length,16);
        if (address==0){
            logger.info("Binary data downloading is supported");
            sendWithAck(PacketBuilder.OK());
            return;
        }
        String data = fields.get(2);
        //String [] memdata= data.split();
        for (int i=0;i<data.length();i+=32){
            instance.setMemory(address,Integer.parseInt(data,2));
            address+=2;
        }
        logger.info("Write memory at {} for {} with {}",Integer.parseInt(addr,16),numberOfBytes,data);
        sendWithAck(PacketBuilder.OK());
    }*/

	/**
	 * Read the register window:
	 * only the 12 local register values of the given register window base are sent back
	 *
	 * @param pkt a received packet
	 */

	private static void readRegisterWindow(Packet pkt) {
		int registerWindow = Command.getFirstFieldtoInt(pkt);
		String data = null;
		String display = null;
		String memdata;
		for (int i = 4; i < REG_AMOUNT; i++) {
			if (i == 4) {
				memdata = Integer.toHexString(SIMULATION_MANAGER_GDB_INTERFACE.getRegisterWindowBaseValues(i, registerWindow));
				data = adaptdatalength(memdata);
				display = Integer.toHexString(SIMULATION_MANAGER_GDB_INTERFACE.getRegisterWindowBaseValues(i, registerWindow));
			} else {
				memdata = Integer.toHexString(SIMULATION_MANAGER_GDB_INTERFACE.getRegisterWindowBaseValues(i, registerWindow));
				data += adaptdatalength(memdata);
				display += SIMULATION_MANAGER_GDB_INTERFACE.getRegisterWindowBaseValues(i, registerWindow);
			}
		}
		if (data.length() == 96) {
			logger.info("read Register Window {} with data {}", registerWindow, display);
			sendWithAck(data);
		} else {
			sendWithAck(ERROR_REPLY);
		}
	}

	/**
	 * Read data from the desired register
	 *
	 * @param pkt a received packet
	 */
	private static void readRegister(Packet pkt) {
		int regNr = Command.getFirstFieldtoInt(pkt);
		int regData = SIMULATION_MANAGER_GDB_INTERFACE.getRegister(regNr);
		String data = Integer.toHexString(regData);
		data = adaptdatalength(data);
		if (data.length() == 8) {
			logger.info("Read register nr={} with value {}", regNr, data);
			sendWithAck(data);
		} else {
			sendWithAck(ERROR_REPLY);
		}
	}

	/**
	 * Write the given data to the desired register
	 *
	 * @param pkt a received packet
	 */
	private static void writeRegister(Packet pkt) {
		String regDump = Command.getFirstField(pkt);
		String[] ra = regDump.split("=");
		String regNr = ra[0];
		String regData = ra[1];
		int number = Integer.parseInt(regNr, RADIX16);
		int data = Integer.parseInt(regData, RADIX16);
		SIMULATION_MANAGER_GDB_INTERFACE.setRegister(number, data);
		logger.info("Write Register {} with data {}", number, data);
		if (SIMULATION_MANAGER_GDB_INTERFACE.getRegister(number) == data) {
			sendWithAck(PacketBuilder.OK());
		} else {
			sendWithAck(PacketBuilder.buildHeader("E01"));
		}
	}

	/**
	 * Write all 16 general registers: global and local
	 *
	 * @param pkt a received packet
	 */
	private static void writeGeneralRegisters(Packet pkt) {
		String regDump = Command.getFirstField(pkt);
		int j = 0;
		String display = null;
		String temp;
		String regData;
		for (int i = 0; i < REG_AMOUNT; i++) {
			temp = regDump;
			regData = temp.substring(j, j + 8);
			SIMULATION_MANAGER_GDB_INTERFACE.setRegister(i, Integer.parseInt(regData, RADIX16));
			display += Integer.parseInt(regData, RADIX16);
			j += 8;
		}
		logger.info("Write general regs with data={}", display);
		if (regDump.length() == 128) {
			sendWithAck(PacketBuilder.OK());
		} else {
			sendWithAck(ERROR_REPLY);
		}
	}

	/**
	 * Read all 16 general registers, global and local
	 *
	 * @param pkt a received packet
	 */
	private static void readGeneralRegisters(Packet pkt) {
		String data = null;
		String memdata;
		for (int i = 0; i < REG_AMOUNT; i++) {
			memdata = Integer.toHexString(SIMULATION_MANAGER_GDB_INTERFACE.getRegister(i));
			if (i == 0) {
				data = adaptdatalength(memdata);
			} else {
				data += adaptdatalength(memdata);
			}
		}
		if (data.length() == 128) {
			sendWithAck(data);
			logger.info("General registers' values :{} ", data);
		} else {
			sendWithAck(ERROR_REPLY);
		}
	}

	/**
	 * Setting a break-or a watchpoint (depending on type) at the given address:
	 * for both types of breakpoints(memory & hardware) kind has to be 4 bytes,
	 * for read,write and access watchpoints : kind is either 1,2,3 or 4 bytes
	 *
	 * @param pkt a received packet
	 */
	private static void setBreakOrWatchpoint(Packet pkt) {
		Vector<String> fields = Command.getFields(pkt);
		String type = fields.get(0);
		String address = fields.get(1);
		String kind = fields.get(2);
		int counter = Integer.parseInt(address, RADIX16);
		int numberOfBytes = Integer.parseInt(kind, RADIX16);
		if (type.equalsIgnoreCase("0")) {
			if (numberOfBytes == 4) {
				if (SIMULATION_MANAGER_GDB_INTERFACE.setMemoryBreakpoint(counter)) {
					logger.info("Memory breakpoint is set");
					sendWithAck(PacketBuilder.OK());
				} else {
					sendWithAck(ERROR_REPLY);
				}
			}
		} else if (type.equalsIgnoreCase("1")) {
			if (numberOfBytes == 4) {
				if (SIMULATION_MANAGER_GDB_INTERFACE.setHardwareBreakpoint(counter)) {
					logger.info("Hardware breakpoint is set");
					sendWithAck(PacketBuilder.OK());
				} else {
					sendWithAck(ERROR_REPLY);
				}
			}
		} else if (type.equalsIgnoreCase("2")) {
			if (numberOfBytes < 1 || numberOfBytes > 4) {
				sendWithAck(ERROR_REPLY);
			} else {
				if (SIMULATION_MANAGER_GDB_INTERFACE.setWatchpoint(counter)) {
					logger.info("Write watchpoint is set");
					sendWithAck(PacketBuilder.OK());
				}
			}
		} else if (type.equalsIgnoreCase("3")) {
			if (numberOfBytes < 1 || numberOfBytes > 4) {
				sendWithAck(ERROR_REPLY);
			} else {
				if (SIMULATION_MANAGER_GDB_INTERFACE.setWatchpoint(counter)) {
					logger.info("Read watchpoint is set");
					sendWithAck(PacketBuilder.OK());
				}
			}
		} else if (type.equalsIgnoreCase("4")) {
			if (numberOfBytes < 1 || numberOfBytes > 4) {
				sendWithAck(ERROR_REPLY);
			} else {
				if (SIMULATION_MANAGER_GDB_INTERFACE.setWatchpoint(counter)) {
					logger.info("Access watchpoint is set");
					sendWithAck(PacketBuilder.OK());
				}
			}
		} else {
			sendWithAck(PacketBuilder.emptyPacket());
		}
	}

	/**
	 * Clear the previously set break- or watchpoint (depending on type) at the given address,
	 * kind is not being considered in this case
	 *
	 * @param pkt a received packet
	 */
	private static void clearBreakOrWatchpoint(Packet pkt) {
		Vector<String> fields = Command.getFields(pkt);
		String type = fields.get(0);
		String address = fields.get(1);
		String kind = fields.get(2);
		int counter = Integer.parseInt(address, RADIX16);
		if (type.equalsIgnoreCase("0")) {
			if (SIMULATION_MANAGER_GDB_INTERFACE.clearMemoryBreakpoint(counter)) {
				logger.info("Memory breakpoint cleared");
				sendWithAck(PacketBuilder.OK());
			} else {
				sendWithAck(ERROR_REPLY);
			}

		} else if (type.equalsIgnoreCase("1")) {
			if (SIMULATION_MANAGER_GDB_INTERFACE.clearHardwareBreakpoint(counter)) {
				logger.info("Hardware breakpoint cleared");
				sendWithAck(PacketBuilder.OK());
			} else {
				sendWithAck(ERROR_REPLY);
			}
		} else if (type.equalsIgnoreCase("2")) {
			if (SIMULATION_MANAGER_GDB_INTERFACE.getAddressSpaceManager().hasWatchpoint(counter)) {
				SIMULATION_MANAGER_GDB_INTERFACE.getAddressSpaceManager().toggleWatchpoint(counter);
				logger.info("Write watchpoint cleared");
				sendWithAck(PacketBuilder.OK());
			} else {
				sendWithAck(ERROR_REPLY);
			}
		} else if (type.equalsIgnoreCase("3")) {
			if (SIMULATION_MANAGER_GDB_INTERFACE.clearWatchpoint(counter)) {
				logger.info("Read watchpoint cleared");
				sendWithAck(PacketBuilder.OK());
			} else {
				sendWithAck(ERROR_REPLY);
			}

		} else if (type.equalsIgnoreCase("4")) {
			if (SIMULATION_MANAGER_GDB_INTERFACE.clearWatchpoint(counter)) {
				logger.info("Access watchpoint cleared");
				sendWithAck(PacketBuilder.OK());
			} else {
				sendWithAck(ERROR_REPLY);
			}

		} else {
			sendWithAck(PacketBuilder.emptyPacket());
		}

	}

	/**
	 * Stepping the program counter by one from the current address or from a specified address
	 *
	 * @param pkt a received packet
	 */
	private static void step(Packet pkt) {
		String address = Command.getFirstField(pkt);
		if (address == null || address.isEmpty()) {
			SIMULATION_MANAGER_GDB_INTERFACE.stepFromCurrentAddress();
			logger.info("step from current address");
		} else {
			int addr = Integer.parseInt(address, RADIX16);
			SIMULATION_MANAGER_GDB_INTERFACE.stepFromGivenAddress(addr);
			logger.info("step from {}", addr);

		}
		sendWithAck(PacketBuilder.stopReason());
	}

	/**
	 * Continue program execution at the given address or at the current address
	 *
	 * @param pkt a received packet
	 */
	private static void continueExec(Packet pkt) {
		String address = Command.getFirstField(pkt);
		if (address == null || address.isEmpty()) {
			SIMULATION_MANAGER_GDB_INTERFACE.continueFromCurrentAddress();
			logger.info("Continue Exec from current address");
		} else {
			int addr = Integer.parseInt(address, RADIX16);
			SIMULATION_MANAGER_GDB_INTERFACE.continueAtGivenAddress(addr);
			logger.info("Continue Exec from {}", addr);
		}
		sendWithAck(PacketBuilder.stopReason(Symbols.GDBSignal.GDB_SIGNAL_CONT));
	}

	/**
	 * Removing all break- and watchpoints, resetting the system and disconnecting the server
	 */
	private static void kill() {
		SIMULATION_MANAGER_GDB_INTERFACE.kill();
		logger.info("Process killed");
	}

	/**
	 * Resetting the system: PC is set back to its initial value,so is the register window base
	 */
	private static void reset() {
		SIMULATION_MANAGER_GDB_INTERFACE.reset();
		logger.info("Resetting the System");
	}

	/**
	 * This method adapts the data string to the standard 4 byte GDB size(8 hex digits)if this is not the case
	 *
	 * @param data the data to be expanded to 4 byte
	 * @return the expanded string
	 */
	private static String adaptdatalength(String data) {
		switch (data.length()) {
			case 1: {
				return "0000000" + data;
			}
			case 2: {
				return "000000" + data;
			}
			case 3: {
				return "00000" + data;
			}
			case 4: {
				return "0000" + data;
			}
			case 5: {
				return "000" + data;
			}
			case 6: {
				return "00" + data;
			}
			case 7: {
				return "0" + data;
			}
			default: {
				return data;
			}
		}
	}

	/**
	 * puts the reply in the correct format by building the packet through the class PacketBuilder and then passes
	 * the packet to sendWithAck
	 *
	 * @param body the body of the message without the header
	 */
	private static void sendWithAck(String body) {
		sendWithAck(PacketBuilder.buildHeader(body));
	}

	/**
	 * replies with the requested packet(in the correct format) to gdb
	 *
	 * @param pkt a received packet
	 */

	private static void sendWithAck(Packet pkt) {
		Server.getInstance().send(pkt.toString());
	}

	private static void send(Packet pkt) {
		Server.getInstance().send(pkt.toString());
	}

}

