/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gdb;

import de.tu_darmstadt.rs.sim.engine.AccessManager;
import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.SimControlSelector;
import de.tu_darmstadt.rs.sim.engine.SimulationThread;
import de.tu_darmstadt.rs.sim.engine.net.MpsocManager;
import de.tu_darmstadt.rs.sim.gui.GuiMessageHandler;
import de.tu_darmstadt.rs.sim.gui.RefreshThread;
import de.tu_darmstadt.rs.sim.preferences.ProjectManager;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;


public class GdbServer extends Thread {

	private static final Logger log = LoggerFactory.getLogger(GdbServer.class);

	private volatile boolean running = true;


	public GdbServer() {
	}


	/**
	 * starts server and links the CommandHandler class to the server ,
	 * so the received packets are processed
	 */

	public void parse() {
		log.info("Starting...");
		//Server server = Server.getInstance();
		//getSimulator();
		Packet pkt;
		log.info("Server Up");
		while (running) {
			try {
				if (!Server.getInstance().isAlive()) {
					Server.getInstance().awaitNextConnection();
				}
				while ((pkt = Server.getInstance().getNextPacket()) != null) {
					log.info("<-\t" + pkt);
					CommandHandler.process(pkt);
				}
			} catch (IOException e) {
				if (!running) {
					log.info("The connection collapsed due to shutdown");
				} else {
					e.printStackTrace();
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * initializes all needed variables to construct a simulator instance and return it
	 *
	 * @return the simulator instance
	 */
	public SimulationManagerGDBInterface getSimulator() {
		final ProjectManager projectManager = new ProjectManager();
		final GuiMessageHandler guiMessageHandler = new GuiMessageHandler();
		final RefreshThread refreshThread = new RefreshThread(100);
		final SimulationThread simulationThread = new SimulationThread();
		final SimControlSelector simControlSelector = new SimControlSelector(simulationThread);
		final MpsocManager mpsocManager = new MpsocManager("localhost", 8886, simControlSelector, simulationThread, guiMessageHandler);
		final AddressSpaceManager addressSpaceManager = new AddressSpaceManager(refreshThread, AccessManager.getTheInstance());
		final SimulationManagerGDBInterface simulationManagerGDBInterface = new SimulationManagerGDBInterface(mpsocManager, refreshThread, simControlSelector, mpsocManager, guiMessageHandler, projectManager, addressSpaceManager, simulationThread);
		simulationManagerGDBInterface.setupSimulation();
		simulationThread.setSimulationManager(simulationManagerGDBInterface);
		//simulationManagerGDBInterface.setInstance(simulationManagerGDBInterface);
		return simulationManagerGDBInterface;
	}

	/**
	 * this method is needed to run the class in an independent thread
	 */

	@Override
	public void run() {
		parse();
	}

	/**
	 * Stops the Server
	 */
	public void terminate() {
		try {
			running = false;
			Server.getInstance().stopServer();
		} catch (SocketException e) {
			log.info("Server stopped!");
		} catch (IOException e) {
			e.printStackTrace();
		}

	}

	/**
	 * The main to run as standalone
	 * @param args none
	 */
	public static void main(String[] args) {
		GdbServer gdbServer = new GdbServer();
		gdbServer.getSimulator();
		gdbServer.parse();

	}
}
