/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gdb;

import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.ServerSocket;
import java.net.Socket;

/**
 * Created by heid on 08.12.15.
 */
public final class Server {

	protected static final int PORT = 6789;
	private static Server server = null;
	private ServerSocket welcomeSocket = null;
	private Socket connectionSocket = null;
	private BufferedReader inFromClient = null;
	private DataOutputStream outToClient = null;


	private Server() {
		startServer();
	}

	/**
	 *
	 * @return The instance of the Server
	 */
	public static Server getInstance() {
		if (server == null) {
			server = new Server();

		}
		return server;
	}

	/**
	 * sets up connection with the gdb(client)
	 */
	public void startServer() {
		if (welcomeSocket == null) {
			try {
				//start inputstream
				welcomeSocket = new ServerSocket(PORT);
			} catch (IOException e) {
				e.printStackTrace();
			}

		}
	}

	/**
	 * waits for the coming packets and forwards these to be processed,
	 * also sends the output packets back to the server
	 *
	 * @return BufferReader, the input coming from the client
	 * @throws IOException if the underlying connection thows an exception
	 */
	public BufferedReader awaitNextConnection() throws IOException {
		connectionSocket = welcomeSocket.accept();
		InputStreamReader isr = new InputStreamReader(connectionSocket.getInputStream());
		inFromClient = new BufferedReader(isr);
		outToClient = new DataOutputStream(connectionSocket.getOutputStream());
		return inFromClient;
	}

	/**
	 * disconnects port
	 *
	 * @return true if socket successfully closed, false if an exception is thrown
	 * @throws IOException if the underlying connection thows an exception
	 */
	public boolean stopServer() throws IOException {
		try {

			welcomeSocket.close();
			if (connectionSocket != null) {
				connectionSocket.close();
			}
			server = null;
			return true;
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}

	}

	/**
	 * @return input coming from client
	 */
	private BufferedReader getInput() {
		return inFromClient;
	}

	/**
	 * @return true if connection is still active, false otherwise
	 * @throws IOException if the underlying connection thows an exception
	 */
	public boolean isAlive() throws IOException {
		if (inFromClient != null) {
			return inFromClient.ready();
		}

		return false;
	}

	/**
	 * transforms the ASCII code received from the client into commands that can be read by the implementation
	 * (gdb format)
	 * checks if first character received is '$'
	 * then checks if the last character is '#'
	 * lastly checks if the checksum digits are two
	 *
	 * @return null if packet format is incorrect
	 * @throws IOException if the underlying connection thows an exception
	 */
	public Packet getNextPacket() throws IOException {
		StringBuffer ret = new StringBuffer();
		int ch = -1;
		boolean hashDetected = false;
		boolean firstCharacter = true;
		int passedHash = 0;
		try {
			while (passedHash < 3 && (ch = inFromClient.read()) != -1) {
				char c = (char) ch;
				if (firstCharacter) {
					if (Character.compare(c, '$') == 0) {
						ret.append(c);
						firstCharacter = false;
					}
				} else {
					ret.append(c);
					if (Character.compare(c, '#') == 0) {
						hashDetected = true;
					}
					if (hashDetected) {
						passedHash++;
					}
				}
			}
		} catch (IOException e) {
			return null;
		}
		if (ch == -1) {
			return null;
		}
		Packet p = new Packet(ret.toString());
		acknowledgePacket(p);
		return p;
	}

	/**
	 * acknowledges received packet after examining the checksum,
	 * '+' is for correctly received
	 * '-' otherwise
	 *
	 * @param p a packet to be acked
	 */
	public void acknowledgePacket(Packet p) {
		if (p.isHashVerified(p.getBody())) {
			send("+");
		} else {
			send("-");
		}

        /*try {
			//outToClient.write("+");
            outToClient2.writeByte('+');
        } catch (IOException e) {
            e.printStackTrace();
        }*/

	}

	/**
	 * transforms packet contents into its ASCII code and then gives it back as outputStream
	 *
	 * @param pkt a packet to be sent
	 */
	public void send(String pkt) {
		try {
			LoggerFactory.getLogger(Server.class).debug("->\t" + pkt);
			outToClient.writeBytes(pkt);
		} catch (IOException e) {
			e.printStackTrace();
		}

	}
}
