/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gdb;


import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.MessageHandler_I;
import de.tu_darmstadt.rs.sim.engine.RefreshHandler_I;
import de.tu_darmstadt.rs.sim.engine.Settings;
import de.tu_darmstadt.rs.sim.engine.SimulationController_I;
import de.tu_darmstadt.rs.sim.engine.SimulationManager;
import de.tu_darmstadt.rs.sim.engine.SimulationStepListener_I;
import de.tu_darmstadt.rs.sim.engine.SimulationThread;
import de.tu_darmstadt.rs.sim.engine.net.MpsocManager;
import de.tu_darmstadt.rs.sim.engine.processor.RegisterFile;
import de.tu_darmstadt.rs.sim.preferences.ProjectManager;

/**
 * Created by heid on 14.12.15.
 */
public class SimulationManagerGDBInterface extends SimulationManager {

	private static SimulationManagerGDBInterface simulationManagerGDBInterface = null;
	private int currentAddress = 0;
	private SimulationThread simulationThread;

	/**
	 * Constructor.
	 *
	 * @param mpsocManager {@link MpsocManager}
	 * @param refreshHandler a {@link RefreshHandler_I} to which refresh requests should be sent
	 * @param simulationControllerI controller of the simulation state
	 * @param simulationStepListenerI a listener for the correctly executed simulation steps
	 * @param messageHandlerI handler for occurring messages
	 * @param projectManager project's component de.tu_darmstadt.rs.sim.preferences manager
	 * @param addressSpaceManager the address space manager
	 * @param registerFile the reg file
	 * @param simulationThread the sim thread
	 */
	public SimulationManagerGDBInterface(final MpsocManager mpsocManager,
										 final RefreshHandler_I refreshHandler,
										 final SimulationController_I simulationControllerI,
										 final SimulationStepListener_I simulationStepListenerI,
										 final MessageHandler_I messageHandlerI,
										 final ProjectManager projectManager,
										 AddressSpaceManager addressSpaceManager,
										 RegisterFile registerFile,
										 SimulationThread simulationThread) {
		super(mpsocManager, refreshHandler, simulationControllerI, simulationStepListenerI, messageHandlerI, projectManager, addressSpaceManager);
		regFile = registerFile;
		this.simulationThread = simulationThread;
		refreshHandler.registerTrigger(this);

	}

	/**
	 * Constructor.
	 *
	 * @param mpsocManager {@link MpsocManager}
	 * @param refreshHandler a {@link RefreshHandler_I} to which refresh requests should be sent
	 * @param simulationControllerI controller of the simulation state
	 * @param simulationStepListenerI a listener for the correctly executed simulation steps
	 * @param messageHandlerI handler for occurring messages
	 * @param projectManager project's component de.tu_darmstadt.rs.sim.preferences manager
	 * @param addressSpaceManager the address space manager
	 * @param simulationThread the sim thread
	 */
	public SimulationManagerGDBInterface(final MpsocManager mpsocManager,
										 final RefreshHandler_I refreshHandler,
										 final SimulationController_I simulationControllerI,
										 final SimulationStepListener_I simulationStepListenerI,
										 final MessageHandler_I messageHandlerI,
										 final ProjectManager projectManager,
										 AddressSpaceManager addressSpaceManager,
										 SimulationThread simulationThread) {
		super(mpsocManager, refreshHandler, simulationControllerI, simulationStepListenerI, messageHandlerI, projectManager, addressSpaceManager);
		this.simulationManagerGDBInterface = this;
		this.simulationThread = simulationThread;
	}

	public static SimulationManagerGDBInterface getInstance() {
		return simulationManagerGDBInterface;
	}

	public void setInstance(SimulationManagerGDBInterface sm) {
		this.simulationManagerGDBInterface = sm;
		simulationThread.setSimulationManager(sm);
	}

	/**
	 * saves the data at the starting at memory address 0
	 *
	 * @param address the address
	 * @param data the data to be written
	 */
	public void setMemory(int address, int data) {
		addressSpaceManager.setValue(currentAddress, data);
		currentAddress += 2;

	}

	/**
	 * @param address the address
	 * @return integer value saved at the given memory address
	 */
	public int getMemory(int address) {
		return addressSpaceManager.getValue(address);
	}

	/**
	 * @param regNr the register nr
	 * @return integer value saved at the given register number
	 */
	public int getRegister(int regNr) {
		RegisterFile.RegisterDescriptor rd = regFile.getRegisterDescriptor(regNr, regFile.getRegisterWindowBase());
		return regFile.getRegister(rd);
	}

	/**
	 * saves the data in the given register number
	 *
	 * @param regNr the register nr
	 * @param data the data to be written
	 */
	public void setRegister(int regNr, int data) {
		RegisterFile.RegisterDescriptor rd = regFile.getRegisterDescriptor(regNr, regFile.getRegisterWindowBase());
		regFile.setRegister(rd, data);
	}

	/**
	 * @param registerNumber the register number
	 * @param regWindow the register window
	 * @return the value saved in a specific register of the given window base
	 */
	public int getRegisterWindowBaseValues(int registerNumber, int regWindow) {
		int address = regFile.getRegisterDescriptor(registerNumber, regWindow).getAddress();
		return getRegister(address);
	}

	/**
	 * @param address the address
	 * @return true if the address is a valid one,false otherwise
	 */

	public boolean containsAddress(int address) {
		return regFile.isAddressGlobalRegister(address);
	}

	/**
	 * @return the address currently in the fetch/decode stage of the processor
	 */
	public int getAddress() {
		return this.getPipelineUI().getCurrentAddress();
	}

	/**
	 * sets a memory breakpoint at an address
	 *
	 * @param address the address
	 * @return true if the address has no breakpoint currently set,the breakpoint is successfully set and instruction at that address is saved, false otherwise
	 */
	public boolean setMemoryBreakpoint(int address) {
		if (!addressSpaceManager.hasBreakpoint(address)) {
			addressSpaceManager.toggleBreakpoint(address);
			addressSpaceManager.saveInstruction(address);
			return true;
		}
		return false;
	}

	/**
	 * clears a memory breakpoint set earlier
	 *
	 * @param address the address
	 * @return true if the address has a breakpoint set earlier,the breakpoint is successfully cleared and instruction at that address is cleared, false otherwise
	 */
	public boolean clearMemoryBreakpoint(int address) {
		if (addressSpaceManager.hasBreakpoint(address)) {
			addressSpaceManager.toggleBreakpoint(address);
			addressSpaceManager.clearInstruction(address);
			return true;
		}
		return false;
	}

	/**
	 * sets a hardware breakpoint at an address
	 *
	 * @param address the address
	 * @return true if the address has no breakpoint currently set and the breakpoint is successfully set, false otherwise
	 */
	public boolean setHardwareBreakpoint(int address) {
		if (!addressSpaceManager.hasBreakpoint(address)) {
			addressSpaceManager.toggleBreakpoint(address);
			return true;
		}
		return false;
	}

	/**
	 * clears a hardware breakpoint set earlier
	 *
	 * @param address the address
	 * @return true if the address has a breakpoint set earlier and the breakpoint is successfully removed, false otherwise
	 */
	public boolean clearHardwareBreakpoint(int address) {
		if (addressSpaceManager.hasBreakpoint(address)) {
			addressSpaceManager.toggleBreakpoint(address);
			return true;
		}
		return false;
	}

	/**
	 * sets a watchpoint at an address
	 *
	 * @param address the address
	 * @return true if the address has no watchpoint currently set and the watchpoint is successfully set, false otherwise
	 */
	public boolean setWatchpoint(int address) {
		if (!addressSpaceManager.hasWatchpoint(address)) {
			addressSpaceManager.toggleWatchpoint(address);
			return true;
		}
		return false;
	}

	/**
	 * clears a watchpoint set earlier
	 *
	 * @param address the address
	 * @return true if the address has a watchpoint set earlier and the watchpoint is successfully cleared, false otherwise
	 */
	public boolean clearWatchpoint(int address) {
		if (addressSpaceManager.hasWatchpoint(address)) {
			addressSpaceManager.toggleWatchpoint(address);
			return true;
		}
		return false;
	}

	/**
	 * sets the PC to the next instruction address
	 */
	public void stepFromCurrentAddress() {
		if (getPipelineUI().getCurrentAddress() == 1) {
			getPipelineUI().setNewPC(0);
		}
		simulationThread._run();
	}

	/**
	 * sets PC to the given address and continues execution from there
	 *
	 * @param address the address
	 */
	public void stepFromGivenAddress(int address) {
		while (getAddress() <= address) {
			simulationThread._run();
		}

	}

	/**
	 * continue execution from the same address
	 */
	public void continueFromCurrentAddress() {
		this.getPipelineUI().setNewPC(getAddress());
		while (true) {
			try {
				simulationThread._run();
			} catch (Exception e) {
				return;
			}
		}

	}

	/**
	 * continue execution at address
	 *
	 * @param address the address
	 */

	public void continueAtGivenAddress(int address) {
		this.getPipelineUI().setNewPC(address);
		try {
			while (true) {

				simulationThread._run();
			}
		} catch (Exception e) {
			return;
		}

	}


	/**
	 * sets the register window base to 0 and the PC address in the processor back to its initial value
	 */
	public void reset() {
		sfr.setWindow(0);
		currentAddress = 0;
		this.getPipelineUI().setNewPC(Settings.PC_INITIAL);
		resetSimulation();
		simulationThread.resetSimulation();
	}

	/**
	 * clears all break-and watchpoints and calls the reset method
	 */
	public void kill() {
		addressSpaceManager.clearAllBreakpoints();
		addressSpaceManager.clearAllWatchpoints();
		reset();
	}
}
