/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gui;

import de.tu_darmstadt.rs.sim.engine.MessageHandler_I;
import de.tu_darmstadt.rs.sim.engine.SimulationController_I;

import java.awt.Component;
import java.io.File;
import java.io.IOException;
import javax.swing.JFileChooser;
import javax.swing.filechooser.FileFilter;

/**
 * Helper class for file saving/loading
 */
public class FileHelper {

	private final Component parent;
	private final FileHelperFilter filter;
	private final MessageHandler_I fMsgHandler;
	private final SimulationController_I simCtrl;

	/**
	 * create a new {@link FileHelper}
	 *
	 * @param parent the window which shows the windows of the helper
	 * @param mh a message handler to inform the user about exceptions
	 * @param simCtrl the simulation controller
	 */
	public FileHelper(final Component parent, final MessageHandler_I mh,
					  final SimulationController_I simCtrl) {
		this.parent = parent;
		fMsgHandler = mh;
		filter = new FileHelperFilter();
		this.simCtrl = simCtrl;
	}

	private class FileHelperFilter extends FileFilter {

		private FileHelperMechanism_I helper;

		@Override
		public boolean accept(final File f) {
			if (f.isDirectory()) {
				return true;
			} else if (f.getPath().endsWith(helper.getFileSuffix())) {
				return true;
			} else {
				return false;
			}
		}

		@Override
		public String getDescription() {
			return helper.getFileSuffix();
		}

	}

	/**
	 * show a load dialog
	 *
	 * @param helper a mechanism which provides the data for the {@link FileHelper}
	 * and performs the actual loading
	 */
	public void showLoadDialog(final FileHelperMechanism_I helper) {
		simCtrl.stopSimulation();
		final JFileChooser fileChooser = new JFileChooser(
				helper.getDefaultDirectory());
		filter.helper = helper;
		fileChooser.setFileFilter(filter);
		final int returnVal = fileChooser.showOpenDialog(parent);
		try {
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				helper.onApprovedLoad(fileChooser.getCurrentDirectory(),
						fileChooser.getSelectedFile());
			}
		} catch (IOException e) {
			fMsgHandler.postMessage(String.format(
					"Load from file '%s' failed: %s\n",
					fileChooser.getSelectedFile(), e.getMessage()),
					MessageHandler_I.MsgType.ERR);
		}
	}

	/**
	 * show a save dialog
	 *
	 * @param helper a mechanism which provides the data for the {@link FileHelper}
	 * and performs the actual saving
	 */
	public void showSaveDialog(final FileHelperMechanism_I helper) {
		simCtrl.stopSimulation();
		final JFileChooser fileChooser = new JFileChooser(
				helper.getDefaultDirectory());
		filter.helper = helper;
		fileChooser.setFileFilter(filter);
		final int returnVal = fileChooser.showSaveDialog(parent);
		try {
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				helper.onApprovedSave(fileChooser.getCurrentDirectory(),
						fileChooser.getSelectedFile());
			}
		} catch (IOException e) {
			fMsgHandler.postMessage(
					String.format("Save to file '%s' failed: %s\n",
							fileChooser.getSelectedFile(), e.getMessage()),
					MessageHandler_I.MsgType.ERR);
		}
	}
}
