/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gui;

import de.tu_darmstadt.rs.sim.engine.JConfigDocumentManager;
import de.tu_darmstadt.rs.sim.engine.Settings;
import de.tu_darmstadt.rs.sim.engine.peripheral.PeripheralManager;
import de.tu_darmstadt.rs.sim.utilities.MessageBox;

import java.io.File;
import java.util.List;

/**
 * Mechanism which saves and loads a configuration files
 */
public class JConfigMechanism implements FileHelperMechanism_I,
		GuiDefinitions_I {

	/**
	 * file extension for jconfig files
	 */
	private final static String JCONFIG_ENDING = ".jc3";
	private final MainWindow mainWnd;
	private final PeripheralManager peripheralManager;
	private final JConfigDocumentManager documentManager;

	/**
	 * create a new {@link JConfigMechanism}
	 *
	 * @param mainWnd window which must be updated when loaded
	 * @param peripheralManager {@link PeripheralManager} for the peripherals
	 * @param documentManager the document manager
	 */
	public JConfigMechanism(final MainWindow mainWnd,
							final PeripheralManager peripheralManager,
							final JConfigDocumentManager documentManager) {
		this.mainWnd = mainWnd;
		this.peripheralManager = peripheralManager;
		this.documentManager = documentManager;
	}

	@Override
	public String getFileSuffix() {
		return JCONFIG_ENDING;
	}

	@Override
	public String getDefaultDirectory() {
		return ((File) Settings.getSetting(Settings.LAST_PROJECT_DIR)).getAbsolutePath();
	}

	@Override
	public void onApprovedLoad(final File directory, final File approvedFile) {
		final String filepath = approvedFile.getAbsolutePath();
		documentManager.setJConfigFile(filepath);

		final int subsystems = documentManager.getNumberOfSubsystems();

		if (subsystems == 0) {
			MessageBox.showError("No valid JConfig file chosen",
					"Import JConfig");
			return;
		}

		if (subsystems == 1) {
			documentManager.loadSubsystem(0);
			mainWnd.setTitleExtention(""); // clear previous extensions
		} else { // multiple subsystems
			assert (subsystems > 1);

			final String[] subsystemNames = documentManager.getSubsystemNames();
			String infoText = "Multiple Subsystems detected:\n";
			int index = 0;

			for (String name : subsystemNames) {
				infoText += index + ") " + name + "\n";
				index++;
			}
			infoText += "Choose which Subsystem should be imported.";

			final String answer = MessageBox.showInput(infoText, "ImportJConfig");

			if (answer == null) {
				return;
			}

			int selectedSubsystem = 0;
			try {
				selectedSubsystem = Integer.parseInt(answer);
			} catch (NumberFormatException ne) {
				MessageBox.showError("Entered value is not a number",
						"ImportJConfig");
				return;
			}

			if (selectedSubsystem >= subsystems) {
				MessageBox.showError("Entered value is too big",
						"Import JConfig");
				return;
			}

			// update MainWindow name and perform import
			mainWnd.setTitleExtention(subsystemNames[selectedSubsystem]);
			documentManager.loadSubsystem(selectedSubsystem);
		}

		// remember project directory
		Settings.setSetting(Settings.LAST_PROJECT_DIR, directory);

		// open views for all new peripheries
		final List<String> sortedPeripheralNamesList = peripheralManager
				.getSortedPeripheralNameList();
		for (String name : sortedPeripheralNamesList) {
			SubWindow subWindow = SubWindow.createSubWindow(name);
			if (subWindow == null) {
				MessageBox
						.showError(
								"That should never happen. Creation of new views failed after import",
								"JConfig import errror");
				break;
			}
			mainWnd.attachNewSubWindow(subWindow);
		}
	}

	@Override
	public void onApprovedSave(final File directory, final File approvedFile) {
		// never used
	}

	/**
	 * reload the data
	 */
	public void onReloadJConfig() {
		documentManager.reloadSubsystem();
	}
}
