/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4s. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

package de.tu_darmstadt.rs.sim.gui;

import de.tu_darmstadt.rs.sim.engine.RefreshHandler_I;
import de.tu_darmstadt.rs.sim.engine.RefreshListener_I;
import de.tu_darmstadt.rs.sim.engine.RefreshTrigger_I;
import de.tu_darmstadt.rs.sim.signal.Signal0;
import de.tu_darmstadt.rs.sim.signal.SignalModified;

import java.util.Collection;
import java.util.LinkedList;
import java.util.concurrent.ConcurrentLinkedQueue;
import javax.swing.SwingUtilities;

/**
 * Thread for posting refresh events from the engine's components to GUI
 * components
 */
public class RefreshThread extends Thread implements RefreshHandler_I {

	private int fInterval;
	private final Collection<RefreshTrigger_I> fTriggers = new LinkedList<RefreshTrigger_I>();
	private final ConcurrentLinkedQueue<Signal0> fChanges;
	private boolean fRefreshInProgress;

	/**
	 * create a new {@link RefreshThread}. This is usually required only once.
	 *
	 * @param interval the time in milliseconds which the thread should sleep between
	 * two refreshes (this value is not an exact value, but used for
	 * orientation)
	 */
	public RefreshThread(int interval) {
		super("GUI Refresh Thread");
		fInterval = interval;
		fChanges = new ConcurrentLinkedQueue<Signal0>();
		start();
	}

	@Override
	public void registerTrigger(RefreshTrigger_I trig) {
		fTriggers.add(trig);
	}

	@Override
	public void registerDataProducerChangeSignal(SignalModified sigModified) {
		sigModified.registerModifiedHandler(this, "producerChanged", Signal0.class);
	}

	/**
	 * called when a data producer (registered via
	 * {@link #registerDataProducerChangeSignal(SignalModified)}) fires the 'modified'
	 * de.tu_darmstadt.rs.sim.signal
	 *
	 * @param sigFlush de.tu_darmstadt.rs.sim.signal which should be emitted when the data producer changes
	 */
	public void producerChanged(Signal0 sigFlush) {
		fChanges.add(sigFlush);
	}

	@Override
	public void run() {
		while (true) {
			try {
				Thread.sleep(fInterval);
				runGUIUpdater();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}

	private void runGUIUpdater() {
		if (fRefreshInProgress) {
			return;
		}
		final Collection<Signal0> ch;
		synchronized (fChanges) {
			ch = new LinkedList<Signal0>(fChanges);
			fChanges.clear();
		}
		fRefreshInProgress = true;
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				for (RefreshTrigger_I trig : fTriggers) {
					int[] addr = trig.getDirtyAddresses();
					if (addr.length == 0) {
						continue;
					}
					for (RefreshListener_I l : trig.getRefreshListeners()) {
						for (Integer a : addr) {
							l.refresh(a.intValue());
						}
					}
				}
				for (Signal0 refresh : ch) {
					refresh.emit();
				}
				fRefreshInProgress = false;
			}
		});
	}
}
