/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gui.assembly.variables;

import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.assembly.AssemblyVariablesManager;
import de.tu_darmstadt.rs.sim.engine.parser.LstFileParser;
import de.tu_darmstadt.rs.sim.gui.AbstractView;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.util.HashMap;
import java.util.Map;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;
import javax.swing.DefaultCellEditor;
import javax.swing.JComboBox;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableColumn;

/**
 * The VariableView shows variables which are read from the .data section of a
 * listing file.
 */
@SuppressWarnings("serial")
public class VariablesView extends AbstractView implements
		VariableViewDefinitions_I {

	private final AssemblyVariablesManager manager;
	private final VariablesTableModel tableModel;
	private final DefaultTableColumnModel columnModel;
	private final JTable table;
	private final JScrollPane scrollPane;
	private boolean hasBeenGlobalRefreshed;
	private Map<String, String> knownVariableInterpretations;

	/**
	 * create a new {@link VariablesView}
	 */
	public VariablesView() {
		this.manager = LstFileParser.getTheInstance()
				.getAssemblyVariablesManager();
		manager.registerRefreshListener(this);
		knownVariableInterpretations = new HashMap<String, String>();
		tableModel = new VariablesTableModel(manager);
		setName("Variables");
		setPreferredSize(new Dimension(500, 200));

		columnModel = new DefaultTableColumnModel();

		String[] tableColumnNames = new String[]{"Address", "Name",
				"Width (x 9bit)", "", "Value"};

		TableColumn column;
		for (int i = 0; i < NUMBER_OF_COLUMNS; i++) {
			column = new TableColumn(i);
			column.setHeaderValue(tableColumnNames[i]);
			column.setPreferredWidth(60);
			if (i < (NUMBER_OF_COLUMNS - 1)) {
				// Resizing will only happen in columns without a maximum width
				// Value of 400 was chosen arbitrarily and can be adjusted if required
				column.setMaxWidth(400);
			}
			columnModel.addColumn(column);
		}

		table = new JTable(tableModel, columnModel);
		table.setDefaultRenderer(Object.class, new VariablesCellRenderer());
		table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		table.setAutoResizeMode(JTable.AUTO_RESIZE_LAST_COLUMN);
		table.setShowGrid(false);
		// table.setShowHorizontalLines(true);
		table.setIntercellSpacing(new Dimension(10, 0));
		JComboBox<String> chooser = new JComboBox<String>(
				new VariableInterpretationChooserModel());
		TableCellEditor editor = new DefaultCellEditor(chooser);
		columnModel.getColumn(3).setCellEditor(editor);

		scrollPane = new JScrollPane(table);
		scrollPane.getViewport().setBackground(Color.WHITE);

		setLayout(new BorderLayout());
		add(scrollPane, BorderLayout.CENTER);
	}

	@Override
	public void refresh(int address) {
		tableModel.fireTableDataChanged();
		if (isFirstGlobalRefresh(address)) {
			changeVariableInterpretations();
		}
	}

	private void changeVariableInterpretations() {
		String varName;
		String interpretation;
		for (int i = 0; i < tableModel.getRowCount(); i++) {
			varName = (String) tableModel.getValueAt(i, NAME_COLUMN);
			interpretation = knownVariableInterpretations.get("var_" + varName);
			if ((interpretation != null) && (!interpretation.equals(""))) {
				tableModel.setValueAt(interpretation, i, VALUE_CHOOSE_COLUMN);
			}

		}
	}

	private boolean isFirstGlobalRefresh(int address) {
		boolean globalRefresh = (address == SpartanDefinitions_I.COMPLETE_REFRESH);
		if (globalRefresh && !hasBeenGlobalRefreshed) {
			hasBeenGlobalRefreshed = true;
			return true;
		} else {
			return false;
		}
	}

	@Override
	public void storeConfiguration(Preferences prefs) {
		TableColumn column;
		final int columnCount = columnModel.getColumnCount();
		String columnWidthString = "" + columnModel.getColumn(0).getWidth();
		String varName;
		String chosenInterpretation;
		// columns
		for (int i = 1; i < columnCount; i++) {
			column = columnModel.getColumn(i);
			columnWidthString += "," + column.getWidth();
		}
		prefs.put("columns", columnWidthString);
		// variables
		for (int i = 0; i < tableModel.getRowCount(); i++) {
			varName = (String) tableModel.getValueAt(i, NAME_COLUMN);
			chosenInterpretation = (String) tableModel.getValueAt(i,
					VALUE_CHOOSE_COLUMN);
			prefs.put("var_" + varName, chosenInterpretation);
		}
	}

	@Override
	public void loadConfiguration(Preferences prefs) {
		if (prefs == null) {
			return;
		}
		TableColumn column;
		final int columnCount = columnModel.getColumnCount();
		String columnWidthString = prefs.get("columns", "");
		String[] columnWidth = columnWidthString.split(",");
		// columns
		if (columnWidth.length == columnCount) {
			for (int i = 0; i < columnCount; i++) {
				column = columnModel.getColumn(i);
				column.setWidth(Integer.parseInt(columnWidth[i], 10));
				column.setPreferredWidth(Integer.parseInt(columnWidth[i], 10));
			}
		}
		// do variables later:
		try {
			for (String key : prefs.keys()) {
				if (!key.equals("columns")) {
					knownVariableInterpretations.put(key, prefs.get(key, ""));
				}
			}
		} catch (BackingStoreException e) {
			e.printStackTrace();
		}
	}

	@Override
	public void onViewDispose() {
		manager.unregisterRefreshListener(this);
	}
}
