/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gui.memory;

import de.tu_darmstadt.rs.sim.engine.Settings;
import de.tu_darmstadt.rs.sim.gui.GuiDefinitions_I;
import de.tu_darmstadt.rs.sim.utilities.Formatter;
import de.tu_darmstadt.rs.sim.utilities.MemoryAccess_I;
import de.tu_darmstadt.rs.sim.utilities.ToolTipGuru;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JViewport;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.border.LineBorder;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;

/**
 * Panel which shows a view on the memory's data
 */
@SuppressWarnings("serial")
public class MemoryPanel extends JScrollPane implements
		MemoryViewDefinitions_I, MouseListener, GuiDefinitions_I {

	/**
	 * color for the 3rd and 4th column in each group of 4 (if unselected)
	 */
	private final static Color VERY_LIGHT_GRAY_COLOR = new Color(240, 240, 240);
	final JTable table;
	final DefaultTableModel tableModel;
	private final MemoryAccess_I memory;
	private final MemoryPanel_I viewType;

	/**
	 * create a new {@link MemoryPanel}
	 *
	 * @param memoryAccess the memory to get the data from
	 * @param viewType the meta data of the memory panel
	 */
	public MemoryPanel(final MemoryAccess_I memoryAccess, final MemoryPanel_I viewType) {
		this.viewType = viewType;
		memory = memoryAccess;
		setName(viewType.getName());
		setPreferredSize(new Dimension(600, 200));
		final int divider = viewType.isDataView()? 3 : 4;
		final int columnCount = (viewType.isDataView()? 1 : 2) * (viewType.is9BitView()? 16 : 8);
		tableModel = new DefaultTableModel() {
			@Override
			public int getRowCount() {
				return memory.getSize() >> divider;
			}

			@Override
			public int getColumnCount() {
				return columnCount;
			}

			@Override
			public Object getValueAt(final int row, final int col) {

				switch (col) {

					case ADR_COL:
						return String
								.format("%06x",
										row
												* 16
												+ (Integer) Settings
												.getSetting(Settings.MEMORY_ADDRESS_MIN));

					default:
						final int addr = viewType.translateToAddress(row, col);
						int value = memory.getValue(addr);

						if (viewType.is9BitView()) {
							if ((col - 1) % 2 == 0) {
								value = (value & (BITMASK_9 << 9)) >> 9; // High
							} else {
								value = value & BITMASK_9; // Low
							}

							return String.format("%03x", value);
						} else {
							return String.format("%05x", value);
						}
				}
			}

			@Override
			public boolean isCellEditable(final int row, final int col) {

				if (col == ADR_COL) {
					return false;
				} else {
					return true;
				}
			}

			@Override
			public void setValueAt(final Object aValue, final int row,
								   final int col) {

				assert (col > 0);

				final int addr = viewType.translateToAddress(row, col);
				int value;
				if (viewType.is9BitView()) {
					value = Formatter
							.parseHexString((String) aValue, BITMASK_9);
					if (value == INVALID) {
						return;
					}

					final int oldValue = memory.getValue(addr);

					if ((col - 1) % 2 == 0) {
						value = (value << 9) | (oldValue & BITMASK_9); // High
					} else {
						value = (oldValue & (BITMASK_9 << 9)) | value; // Low
					}
				} else {
					value = Formatter.parseHexString((String) aValue,
							BITMASK_18);
					if (value == INVALID) {
						return;
					}
				}

				memory.setValue(addr, value);
			}
		};

		final int width = (viewType.is9BitView()? 45 : 55);
		final DefaultTableColumnModel columnModel = new DefaultTableColumnModel();
		final TableColumn col0 = new TableColumn(ADR_COL);
		col0.setHeaderValue("Address");
		col0.setMinWidth(60);
		col0.setMaxWidth(60);
		columnModel.addColumn(col0);

		for (int i = 0; i < columnCount; ++i) {
			final TableColumn col1 = new TableColumn(i + 1);
			col1.setHeaderValue(String.format(viewType.getHeaderPattern(i), i));
			col1.setMinWidth(width);
			col1.setMaxWidth(width);
			columnModel.addColumn(col1);
		}
		table = new JTable(tableModel, columnModel);
		table.setDefaultRenderer(Object.class, new DefaultTableCellRenderer() {
			@Override
			public Component getTableCellRendererComponent(final JTable t,
														   final Object value, final boolean isSelected,
														   final boolean hasFocus, final int row, final int col) {

				final Component component = super
						.getTableCellRendererComponent(t, value, isSelected,
								hasFocus, row, col);
				final JLabel label = (JLabel) component;
				label.setHorizontalAlignment(JLabel.CENTER);
				label.setToolTipText(ToolTipGuru.create((String) value, true));

				switch (col) {

					case ADR_COL:
						if (!isSelected) {
							label.setBackground(LIGHT_GRAY_COLOR);
						}
						label.setForeground(Color.BLACK);
						label.setFont((Font) Settings
								.getSetting(Settings.REGISTER_FONT_PLAIN));
						break;

					default:
						if (!isSelected) {
							if (((col - 1) / 2) % 2 == 0) {
								label.setBackground(Color.WHITE);
							} else {
								label.setBackground(VERY_LIGHT_GRAY_COLOR);
							}
							label.setForeground(Color.BLACK);
						}
						String text = label.getText();
						if ((viewType.is9BitView() && text.equals("000"))
								|| ((!viewType.is9BitView()) && text
								.equals("00000"))) {
							label.setFont((Font) Settings
									.getSetting(Settings.REGISTER_FONT_PLAIN));
						} else {
							label.setFont((Font) Settings
									.getSetting(Settings.REGISTER_FONT_BOLD));
						}
						final int defaultAddress = viewType.translateToAddress(row,
								col);
						if (memory.hasWatchpoint(defaultAddress)) {
							label.setBorder(new LineBorder(WATCHPOINT_COLOR));
						} else if (!isSelected) {
							label.setBorder(null);
						}
						break;
				}
				return label;
			}
		});

		table.setSelectionMode(ListSelectionModel.SINGLE_INTERVAL_SELECTION);
		table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		table.getTableHeader().setReorderingAllowed(false);
		table.setRowHeight(15);
		table.addMouseListener(this);

		setViewportView(table);
	}

	/**
	 * scroll to an address. If the toTopParameter is set to true then the line
	 * with containing the address will be the top line in the panel.
	 *
	 * @param address the address to which the memory panel should scroll
	 * @param toTop true if the memory panel should show the line containing the
	 * line as the topmost line
	 */
	public void scrollToAddress(int address, boolean toTop) {
		JViewport viewport = getViewport();
		int rowIndex = findRowIndexForScrolling(address);
		Rectangle rectangle = table.getCellRect(rowIndex, ADR_COL, true);
		Rectangle viewportRectangle = viewport.getViewRect();
		int newHeight = rectangle.y - viewportRectangle.y;
		if (toTop && newHeight > 0) {
			newHeight += (viewportRectangle.height - rectangle.height);
		}
		rectangle.y = newHeight;
		viewport.scrollRectToVisible(rectangle);
	}

	private int findRowIndexForScrolling(final int address) {
		return address >> 4;
	}

	/**
	 * Deletes the memory table's currently selected line.
	 */
	public void deleteSelection() {

		for (int row : table.getSelectedRows()) {

			for (int col = 1; col < tableModel.getColumnCount(); ++col) {
				tableModel.setValueAt("0", row, col);
			}
		}

	}

	/**
	 * refresh the data at a particular address
	 *
	 * @param defaultAddress the address which data has changed
	 */
	public void refresh(final int defaultAddress) {
		final int theRow;
		if (defaultAddress < 0) {
			tableModel.fireTableDataChanged();
			return;
		}
		if (viewType.isDataView()) {
			theRow = defaultAddress >> 3;
		} else {
			theRow = defaultAddress >> 4;
		}
		tableModel.fireTableRowsUpdated(theRow, theRow);
	}

	@Override
	public void mouseClicked(final MouseEvent event) {

		// react only to right mouse click
		if (!SwingUtilities.isRightMouseButton(event)) {
			return;
		}

		int row = table.rowAtPoint(event.getPoint());
		int col = table.columnAtPoint(event.getPoint());

		if (col == ADR_COL) {
			return; // do not care about address column
		}

		final int addr = viewType.translateToAddress(row, col);
		memory.toggleWatchpoint(addr);
	}

	@Override
	public void mousePressed(MouseEvent e) {
		// TODO Auto-generated method stub

	}

	@Override
	public void mouseReleased(MouseEvent e) {
		// TODO Auto-generated method stub

	}

	@Override
	public void mouseEntered(MouseEvent e) {
		// TODO Auto-generated method stub

	}

	@Override
	public void mouseExited(MouseEvent e) {
		// TODO Auto-generated method stub

	}
}
