/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gui.peripheralRegisters;

import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.List;
import javax.swing.Box;
import javax.swing.JPanel;

/**
 * Cell containing a binary value of a register with bit value selectors
 */
@SuppressWarnings("serial")
public class BinRegisterCell extends JPanel implements RegisterCell_I,
		ActionListener {

	private BitValueSelector[] bitValueSelectorList;
	private List<ActionListener> actionListenerList;
	private int howManyBits;

	/**
	 * create a new {@link BinRegisterCell} containing howManyBits
	 *
	 * @param howManyBits the number of bits contained in the cell
	 */
	public BinRegisterCell(int howManyBits) {
		this.howManyBits = howManyBits;
		setOpaque(true);
		GridBagLayout layout = new GridBagLayout();
		setLayout(layout);
		GridBagConstraints constraints = new GridBagConstraints();
		constraints.weightx = 1.0;
		constraints.gridheight = 2;
		Component comp = Box.createHorizontalStrut(5);
		add(comp);
		layout.setConstraints(comp, constraints);
		constraints = new GridBagConstraints();
		constraints.gridheight = GridBagConstraints.REMAINDER;

		bitValueSelectorList = new BitValueSelector[howManyBits];
		BitValueSelector bitValueSelector;
		for (int i = 0; i < howManyBits; i++) {
			if ((howManyBits - i) % 8 == 0) {
				comp = Box.createHorizontalStrut(5);
				add(comp);
				layout.setConstraints(comp, constraints);
			}
			bitValueSelector = new BitValueSelector();
			bitValueSelector.addActionListener(this);
			bitValueSelector.setOpaque(true);
			layout.setConstraints(bitValueSelector, constraints);
			bitValueSelectorList[howManyBits - 1 - i] = bitValueSelector;
			add(bitValueSelector);
		}
		comp = Box.createHorizontalStrut(5);
		add(comp);
		layout.setConstraints(comp, constraints);
		setBorder(null);
		setBackground(Color.WHITE);
		actionListenerList = new ArrayList<ActionListener>();
	}

	/**
	 * set a (new) value
	 *
	 * @param value the value to be set
	 */
	public void setValue(final int value) {
		int mask;
		boolean bvalue;
		for (int i = 0; i < howManyBits; i++) {
			mask = (1 << i);
			bvalue = (value & mask) != 0;
			bitValueSelectorList[i].setValue(bvalue);
		}
	}

	@Override
	public Object getValue() {
		int returnValue = 0;
		for (int i = 0; i < howManyBits; i++) {
			returnValue += (bitValueSelectorList[i].getValue() << i);
		}
		return returnValue;
	}

	@Override
	public void actionPerformed(ActionEvent e) {
		informActionListeners(e);
	}

	private void informActionListeners(ActionEvent e) {
		for (ActionListener actionListener : actionListenerList) {
			actionListener.actionPerformed(e);
		}
	}

	/**
	 * add an {@link ActionListener} to this component. This is used to add a
	 * cell editor.
	 *
	 * @param actionListener the listener which should receive events
	 */
	public void addActionListener(ActionListener actionListener) {
		if (!actionListenerList.contains(actionListener)) {
			actionListenerList.add(actionListener);
		}
	}

	/**
	 * remove an {@link ActionListener} from this component. This is used to
	 * remove a cell editor.
	 *
	 * @param actionListener the listener which should not receive events any more
	 * @return true if the actionListener was previously added, false otherwise
	 */
	public boolean removeActionListener(ActionListener actionListener) {
		return actionListenerList.remove(actionListener);
	}
}
