/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gui.peripheralRegisters;

import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;

import java.util.regex.Matcher;
import java.util.regex.Pattern;
import javax.swing.JTextField;

/**
 * Cell containing a hexadecimal value of a register
 */
@SuppressWarnings("serial")
public class HexRegisterCell extends JTextField implements RegisterCell_I {

	private Pattern octPattern = Pattern.compile("(-{0,1})0{0,1}([0-7]*)");
	private Pattern decPattern = Pattern.compile("(-{0,1})(\\p{Digit}*)");
	private Pattern hexPattern = Pattern
			.compile("(-{0,1})(0x){0,1}(\\p{XDigit}*)");
	private Pattern binPattern = Pattern.compile("(-{0,1})(0b){0,1}([0,1]*)");
	private int howManyBits;
	private int howManyHex;

	/**
	 * create a new {@link HexRegisterCell}
	 *
	 * @param howManyBits the number of bits which are represented with this cell
	 */
	public HexRegisterCell(int howManyBits) {
		this.howManyBits = howManyBits;
		this.howManyHex = howManyBits / 4;
		if (this.howManyHex * 4 < howManyBits) {
			this.howManyHex++;
		}
	}

	/**
	 * set the (new) value
	 *
	 * @param value the value
	 */
	public void setValue(final int value) {
		String shownValue = String.format("0x%0" + howManyHex + "x", value);
		setText(shownValue);
	}

	@Override
	public Object getValue() {
		String shownValue = getText().toLowerCase();
		int theValue = SpartanDefinitions_I.INVALID;
		Matcher matcher = octPattern.matcher(shownValue);
		int returnValue = SpartanDefinitions_I.INVALID;
		if (matcher.matches()) {
			theValue = parseInt(matcher.group(1), matcher.group(2), 8);
		} else {
			matcher = decPattern.matcher(shownValue);
			if (matcher.matches()) {
				theValue = parseInt(matcher.group(1), matcher.group(2), 10);
			} else {
				matcher = binPattern.matcher(shownValue);
				if (matcher.matches()) {
					theValue = parseInt(matcher.group(1), matcher.group(3), 2);
				} else {
					matcher = hexPattern.matcher(shownValue);
					if (matcher.matches()) {
						theValue = parseInt(matcher.group(1), matcher.group(3),
								16);
					}
				}
			}
		}
		if (theValue != SpartanDefinitions_I.INVALID) {
			returnValue = 0;
			for (int i = 0; i < howManyBits; i++) {
				returnValue += theValue & (1 << i);
			}
		}
		return returnValue;
	}

	private String limitLengthFor18Bit(String theString, int radix) {
		final int length = theString.length();
		final int lengthRelativeToRadix;
		if (radix == 16) {
			lengthRelativeToRadix = 5;
		} else if (radix == 10) {
			lengthRelativeToRadix = 6;
		} else if (radix == 8) {
			lengthRelativeToRadix = 6;
		} else {
			lengthRelativeToRadix = 18;
		}
		if (length <= lengthRelativeToRadix) {
			return theString;
		} else {
			return theString.substring(length - lengthRelativeToRadix);
		}
	}

	private int parseInt(String sign, String theString, int radix) {
		return Integer.parseInt(sign + limitLengthFor18Bit(theString, radix),
				radix) & SpartanDefinitions_I.BITMASK_18;
	}
}
