/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gui.peripheralRegisters;

import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.peripheral.JConfigPeripheral;
import de.tu_darmstadt.rs.sim.gui.AbstractView;

import java.awt.BorderLayout;
import java.util.prefs.Preferences;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;
import javax.swing.table.DefaultTableColumnModel;
import javax.swing.table.TableColumn;
import javax.swing.table.TableColumnModel;

/**
 * View for a peripherals registers
 */
@SuppressWarnings("serial")
public class PeripheralRegisterView extends AbstractView implements
		TableModelListener {

	private RegisterModel tableModel;
	private JConfigPeripheral peripheral;
	private TableColumnModel columnModel;
	private JTable table;

	/**
	 * create a new {@link PeripheralRegisterView}
	 *
	 * @param peripheral the peripheral to get the information from
	 * @param addressSpaceManager the addressSpaceManager which toggles watchpoints
	 */
	public PeripheralRegisterView(JConfigPeripheral peripheral, AddressSpaceManager addressSpaceManager) {
		this.peripheral = peripheral;
		columnModel = new DefaultTableColumnModel();
		TableColumn column;
		tableModel = new RegisterModel(peripheral, peripheral.getRegisters());
		peripheral.addRefreshListener(this);
		int columnWidth;
		for (int i = 0; i < RegisterTableDefinitions_I.COLUMN_COUNT; i++) {
			column = new TableColumn(i);
			column.setHeaderValue(tableModel.getColumnName(i));
			columnWidth = getFixedWidth(i);
			if (columnWidth > 0) {
				column.setMinWidth(columnWidth);
				column.setMaxWidth(columnWidth);
			} else {
				column.setWidth(30);
			}
			columnModel.addColumn(column);
		}
		table = new JTable(tableModel, columnModel);
		tableModel.addTableModelListener(this);
		table.setDefaultEditor(Object.class, new RegisterCellEditor());
		table.setDefaultRenderer(Object.class, new RegisterCellRenderer());
		table.setRowHeight(30);
		table.addMouseListener(new PeripheralRegisterMouseListener(this, peripheral, addressSpaceManager));
		JScrollPane pane = new JScrollPane(table);
		setLayout(new BorderLayout());
		add(pane, BorderLayout.CENTER);
	}

	@Override
	public void refresh(int address) {
		tableModel.fireTableDataChanged();
	}

	@Override
	public void storeConfiguration(Preferences prefs) {
		TableColumn column;
		final int columnCount = columnModel.getColumnCount();
		String columnWidthString = "" + columnModel.getColumn(0).getWidth();
		// columns
		for (int i = 1; i < columnCount; i++) {
			column = columnModel.getColumn(i);
			columnWidthString += "," + column.getWidth();
		}
		prefs.put("columns", columnWidthString);
	}

	@Override
	public void loadConfiguration(Preferences prefs) {
		if (prefs == null) {
			return;
		}
		TableColumn column;
		final int columnCount = columnModel.getColumnCount();
		String columnWidthString = prefs.get("columns", "");
		String[] columnWidth = columnWidthString.split(",");
		// columns
		if (columnWidth.length == columnCount) {
			for (int i = 0; i < columnCount; i++) {
				column = columnModel.getColumn(i);
				column.setWidth(Integer.parseInt(columnWidth[i], 10));
				column.setPreferredWidth(Integer.parseInt(columnWidth[i], 10));
			}
		}
	}

	@Override
	public void onViewDispose() {
		peripheral.removeRefreshListener(this);
	}

	@Override
	public String getName() {
		return this.getClass().getSimpleName();
	}

	@Override
	public void tableChanged(TableModelEvent e) {
		table.repaint();
	}

	/**
	 * get a column's width
	 *
	 * @param columnIndex the index of the column
	 * @return width of that column
	 */
	public int getFixedWidth(int columnIndex) {
		switch (columnIndex) {
			case RegisterTableDefinitions_I.EXPAND_COLUMN:
				return 20;
			case RegisterTableDefinitions_I.OFFSET_COLUMN:
				return 20;
			case RegisterTableDefinitions_I.HEX_COLUMN:
				return -1;
			default:
				return -1;
		}
	}

	JTable getTable() {
		return table;
	}
}
