/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package de.tu_darmstadt.rs.sim.gui.peripheralRegisters;

import de.tu_darmstadt.rs.sim.engine.peripheral.JConfigPeripheral;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;

import java.util.ArrayList;
import java.util.List;


/**
 * Line in the {@link PeripheralRegisterView} containing information about a
 * register
 */
public class RegisterNode implements RegisterNode_I, RegisterTableDefinitions_I {

	private boolean isExpanded;
	private List<RegisterNode_I> childList;
	private PeripheralModuleDescription.Register register;
	private JConfigPeripheral peripheral;
	private int registerOffset;

	/**
	 * create a new {@link RegisterNode}
	 *
	 * @param peripheral the peripheral whose register is displayed
	 * @param register the register to display
	 * @param registerOffset the offset of the register to the peripheral's base I/O
	 * address
	 */
	public RegisterNode(JConfigPeripheral peripheral, PeripheralModuleDescription.Register register,
						int registerOffset) {
		this.register = register;
		this.peripheral = peripheral;
		this.registerOffset = registerOffset;
		childList = new ArrayList<RegisterNode_I>();
		BitGroupNode node;
		if (register.getBitGroups() != null) {
			for (int i = 0; i < register.getBitGroups().size(); i++) {
				node = new BitGroupNode(this, register.getBitGroups().get(i));
				childList.add(node);
			}
		}
	}

	@Override
	public void setExpanded(boolean expanded) {
		isExpanded = expanded;
	}

	@Override
	public boolean isExpanded() {
		return isExpanded;
	}

	@Override
	public boolean canExpand() {
		return true;
	}

	@Override
	public void addChild(RegisterNode_I childNode) {
		childList.add(childNode);
	}

	@Override
	public Object getValue(int columnIndex) {
		Object returnValue = null;
		int value = getTheRegisterValue();
		switch (columnIndex) {
			case OFFSET_COLUMN:
				returnValue = registerOffset;
				break;
			case NAME_COLUMN:
				returnValue = register.getName();
				break;
			case BIT_INDEX_COLUMN:
				returnValue = "0..17";
				break;
			case HEX_COLUMN:
			case BIN_COLUMN:
				returnValue = value;
				break;
			case SYMBOL_COLUMN:
				String symbolString = register.toString();
				if (symbolString != null) {
					returnValue = symbolString;
				}
				break;
			default:
				return register;
		}
		return returnValue;
	}

	@Override
	public void setValue(Object theValue, int columnIndex) {
		switch (columnIndex) {
			case HEX_COLUMN:
			case BIN_COLUMN:
				int intValue = (Integer) theValue;
				setTheRegisterValue(intValue);
				break;
			case SYMBOL_COLUMN:
				String strValue = (String) theValue;
				if (strValue != null && !strValue.equals("")) {
					setTheRegisterValue((int) register.getDefaultvalue());
				}
				break;
		}
	}

	@Override
	public int getChildCount() {
		return childList.size();
	}

	@Override
	public RegisterNode_I getChildAt(int index) {
		return childList.get(index);
	}

	/**
	 * retrieve the value of the register
	 *
	 * @return register value
	 */
	public int getTheRegisterValue() {
		return peripheral.getRegisterValue(registerOffset);
	}

	/**
	 * set the (new) register value
	 *
	 * @param value the value to be set
	 */
	public void setTheRegisterValue(int value) {
		peripheral.setRegisterValue(registerOffset, value);
	}

	@Override
	public int getBitWidth() {
		return 18;
	}

	@Override
	public int getLevel() {
		return 0;
	}

	@Override
	public String getDescription() {
		return register.getDescription();
	}
}
