/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

package de.tu_darmstadt.rs.sim.gui.uart;

import de.tu_darmstadt.rs.sim.engine.peripheral.uart.AbstractUart;
import de.tu_darmstadt.rs.sim.engine.peripheral.uart.Uart;
import de.tu_darmstadt.rs.sim.engine.peripheral.uart.UartStream_I;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;

/**
 * This class allows to send any file for reception by the SpartanMC uart. The
 * transfer is accomplished by reading the file one byte after another and
 * sending it via the uart line. This process is synchronous, which means, that
 * the next byte is sent only if the RX fifo has free space left.
 */
public class UartBinaryFileTransfer implements UartStream_I {

	private File fFile;
	private FileInputStream fIn;
	private UartView fView;
	private AbstractUart fUart;

	/**
	 * create and initialize an input stream from the file to the Uart
	 *
	 * @param view the view the binary transfer was started from
	 * @param uart the {@link Uart}
	 * @param f the file which contains the actual data
	 */
	public UartBinaryFileTransfer(UartView view, AbstractUart uart, File f) {
		fFile = f;
		fView = view;
		fUart = uart;
		try {
			fIn = new FileInputStream(f);
		} catch (IOException e) {
			System.err
					.println("error while opening " + fFile.getAbsolutePath());
			System.err.println(e);
		}
	}

	/**
	 * Receives the next byte from the file being transmitted to the uart.
	 *
	 * @return a byte
	 */
	@Override
	public byte onReceiveNextByte() {
		try {
			byte x = (byte) fIn.read();
			// System.out.format("byte from file: %02X\n", (int)(x & 0xFF));
			return x;
		} catch (IOException e) {
			System.err.println("error while reading character from "
					+ fFile.getAbsolutePath());
			System.err.println(e);
			return 0x0;
		}
	}

	/**
	 * Checks if there is any byte left in the file to be received by the uart.
	 *
	 * @return true, if at least one byte can be received
	 */
	@Override
	public boolean hasNextByte() {
		boolean has_byte;
		boolean eof;
		try {
			eof = (fIn == null) || (fIn.available() == 0);
			has_byte = (!eof)
					&& ((fUart.getStatusIoValue() & Uart.STATUS_RX_FIFO_FULL) == 0);
		} catch (IOException e) {
			System.err.println(e);
			has_byte = false;
			eof = true;
		}
		if (eof) {
			fView.stopBinaryTransfer();
		}
		return has_byte;
	}

	/**
	 * File transfer does not implement an uart transmit channel.
	 *
	 * @param nextByte any byte
	 */
	@Override
	public void onTransmitNextByte(byte nextByte) {
		/* file transfer does not return anything */
	}

}
