/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 *
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2011
 *
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2011
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
*/

/**
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2011
 */
package de.tu_darmstadt.rs.sim.preferences;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;
import java.util.prefs.BackingStoreException;
import java.util.prefs.Preferences;
import java.util.prefs.PreferencesFactory;

/**
 * Class for managing the project components which should be able to store their de.tu_darmstadt.rs.sim.preferences
 */
public class ProjectManager implements PreferencesFactory {

	private static ProjectPreferences rootPreference = null;

	private final List<ProjectComponent_I> components;

	/**
	 * create a new {@link ProjectManager}
	 */
	public ProjectManager() {
		components = new LinkedList<ProjectComponent_I>();
	}

	/**
	 * Add the ProjectComponent to the project.
	 * @param comp the project component
	 */
	public void registerComponent(ProjectComponent_I comp) {
		components.add(comp);
	}

	/**
	 * Remove the ProjectComponent from the project.
	 * @param comp the project component
	 * @return true, if the component was removed and false, if component could not be found.
	 */
	public boolean removeComponent(ProjectComponent_I comp) {
		return components.remove(comp);
	}

	/**
	 * remove all child de.tu_darmstadt.rs.sim.preferences from root.
	 */
	private void clearPreferences() {
		try {
			for (String nodeName : rootPreference.childrenNames()) {
				rootPreference.node(nodeName).removeNode();
			}
		} catch (BackingStoreException e) {
			//should never happen
			assert (false);
		}
	}

	/**
	 * Reads the configuration file and order all ProjectComponents to load
	 * the new configuration. The components load their configuration, in the order
	 * they were registered.
	 * @param filename path to the configuration file.
	 * @throws IOException exception thrown when the project manager fails to load the file
	 */
	public void loadProject(String filename) throws IOException {
		if (rootPreference == null) {
			userRoot(); //set rootPreference
		}

		assert (rootPreference != null);

		clearPreferences(); //is necessary as nodes are cached by AbstractPreferences

		rootPreference.importProjectPreferences(new FileInputStream(filename));

		for (ProjectComponent_I comp : components) {
			comp.loadConfiguration(rootPreference.node(comp.getClass().getName()));
		}
	}

	/**
	 * Order all registered ProjectComponents to store their current configuration and
	 * write it to the configuration file. The components store their configuration, in
	 * the order they were registered.
	 * @param filename path to the configuration file.
	 * @throws IOException exception thrown when the project manager fails to save the file
	 */
	public void saveProject(String filename) throws IOException {
		if (rootPreference == null) {
			userRoot(); //set rootPreference
		}

		assert (rootPreference != null);

		clearPreferences(); //start with an empty root node

		for (ProjectComponent_I comp : components) {
			comp.storeConfiguration(rootPreference.node(comp.getClass().getName()));
		}

		rootPreference.exportProjectPreferences(new FileOutputStream(filename));
	}


	@Override
	public Preferences systemRoot() {
		//no support for systemRoot
		return userRoot();
	}

	@Override
	public Preferences userRoot() {
		if (rootPreference == null) {
			rootPreference = new ProjectPreferences(null, "");
		}
		return rootPreference;
	}
}
