/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 *
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2012
 *
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2012
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

/**
 * Spartan MC Simulator
 * Matthias Hofmann
 * Department for Computer Science, Chair for Embedded Systems
 * TU Dresden, 2012
 */
package de.tu_darmstadt.rs.sim.preferences;

import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.StringReader;
import java.util.HashMap;
import java.util.Map;
import java.util.prefs.AbstractPreferences;
import java.util.prefs.BackingStoreException;

/**
 * Preference class, which does not have a backing store. If you do not use Spi
 * methods directly, no BackingStoreException will be thrown by any of the
 * ProjectPreference methods.
 */
public class ProjectPreferences extends AbstractPreferences {

	private Map<String, String> values;
	private Map<String, ProjectPreferences> childs;

	final private ProjectPreferences parent;

	/**
	 * Create an new instance of the ProjectPreferences. Root de.tu_darmstadt.rs.sim.preferences are
	 * created by passing NULL as parent and "" as name.
	 *
	 * @param parent
	 *            the preference that the contains this preference
	 * @param name
	 *            the name of this preference
	 */
	protected ProjectPreferences(ProjectPreferences parent, String name) {
		super(parent, name);

		this.parent = parent;

		values = new HashMap<String, String>();
		childs = new HashMap<String, ProjectPreferences>();
	}

	/**
	 * Import ProjectPreferences from the InputStream. The Subtree is appended
	 * to this node. It is recommend to use this method only with the root node.
	 * If the subtree contains data, their values are only overwritten, if the
	 * InputStream assigns new values to them.
	 *
	 * @param is
	 *            the input stream to read data from
	 * @throws IOException
	 *             exception thrown if the import fails
	 */
	protected void importProjectPreferences(InputStream is) throws IOException {
		PreferencesXmlHandler handler = new PreferencesXmlHandler(this);

		try {
			XMLReader xmlReader = XMLReaderFactory.createXMLReader();

			InputSource inputSource = new InputSource(is);

			// disable online xml schema lookup
			xmlReader.setEntityResolver(new EntityResolver() {
				@Override
				public InputSource resolveEntity(String publicId,
												 String systemId) throws SAXException, IOException {
					return new InputSource(new StringReader(""));
				}
			});

			xmlReader.setContentHandler(handler);
			xmlReader.parse(inputSource);
		} catch (SAXException e) {
			try {
				clear();
			} catch (BackingStoreException be) {
				// should never happen
				assert (false);
			}
		}
	}

	/**
	 * Export the current ProjectPreferences to the OutputStream. Only this node
	 * and its children are exported.
	 *
	 * @param os
	 *            the output stream to write data to
	 * @throws IOException
	 *             exception thrown if the export fails
	 */
	protected void exportProjectPreferences(OutputStream os) throws IOException {
		try {
			exportSubtree(os);
		} catch (BackingStoreException e) {
			// TODO: check: really ignore?
		}
	}

	@Override
	public boolean isUserNode() {
		// only user nodes supported
		return true;
	}

	@Override
	protected AbstractPreferences childSpi(String name) {
		ProjectPreferences child = childs.get(name);

		if (child == null) {
			child = new ProjectPreferences(this, name);
		}

		return child;
	}

	@Override
	protected String[] childrenNamesSpi() throws BackingStoreException {
		return childs.keySet().toArray(new String[childs.keySet().size()]);
	}

	@Override
	public void flush() throws BackingStoreException {
		// no central BackingStore -> do nothing
	}

	@Override
	protected void flushSpi() throws BackingStoreException {
		throw new BackingStoreException("flusSpi called, should never happen");
	}

	@Override
	protected String getSpi(String key) {
		return values.get(key);
	}

	@Override
	protected String[] keysSpi() throws BackingStoreException {
		return values.keySet().toArray(new String[values.keySet().size()]);
	}

	@Override
	protected void putSpi(String key, String value) {
		values.put(key, value);
	}

	@Override
	protected void removeNodeSpi() throws BackingStoreException {
		if (parent != null) {
			parent.childs.remove(this);
		}

	}

	@Override
	protected void removeSpi(String key) {
		values.remove(key);

	}

	@Override
	public void sync() throws BackingStoreException {
		// no central BackingStore -> do nothing
	}

	@Override
	protected void syncSpi() throws BackingStoreException {
		throw new BackingStoreException("syncSpi called, should never happen");
	}

}
