/**
 * _______________________________________________________________________________
 *
 * Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 * Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 * (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 * must display the following acknowledgement: "This product includes
 * software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 * its contributors."
 *
 * 4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 * names of its contributors may be used to endorse or promote products
 * derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 * BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 * FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
/*
_______________________________________________________________________________

 Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 (http://www.mr.inf.tu-dresden.de) All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

 3. All advertising materials mentioning features or use of this software
    must display the following acknowledgement: "This product includes
    software developed by the TU Dresden Chair for Embedded Systems and
    its contributors."

 4. Neither the name of the TU Dresden Chair for Embedded Systems nor the
    names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS AND
 CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
_______________________________________________________________________________
 */

package de.tu_darmstadt.rs.sim.signal;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * base class for all signals. This class provides the basic functionality to
 * emit/fire signals and to connect methods (which are called when the de.tu_darmstadt.rs.sim.signal is
 * emitted). This class uses reflections to call the methods.
 */
public abstract class Signal {

	/**
	 * internal class for remembering the object and the method for the emit
	 */
	protected class ObjectMethod {

		/**
		 * method which should be called
		 */
		public final Method method;
		/**
		 * object which has the method
		 */
		public final Object object;

		/**
		 * create a {@link ObjectMethod}
		 *
		 * @param o object having the method
		 * @param m the method to be called
		 */
		public ObjectMethod(Object o, Method m) {
			object = o;
			method = m;
		}

		@Override
		public boolean equals(Object o) {
			if (!(o instanceof ObjectMethod)) {
				return false;
			}
			ObjectMethod om = (ObjectMethod) o;
			if (System.identityHashCode(om.object) != System
					.identityHashCode(object)) {
				return false;
			}
			if (!om.method.equals(method)) {
				return false;
			}
			return true;
		}
	}


	private final LinkedList<ObjectMethod> fConn;

	/**
	 * create a new {@link Signal} can be emitted later
	 */
	public Signal() {
		fConn = new LinkedList<ObjectMethod>();
	}

	/**
	 * register a method call with some arguments
	 *
	 * @param o the object having the method
	 * @param methodName the name of the method to be invoked
	 * @param args the parameters' classes of the method
	 */
	protected final void registerMethodCallOnEmitInternal(final Object o,
														  final String methodName, Class<?>... args) {
		try {
			ObjectMethod om = new ObjectMethod(o, o.getClass().getMethod(
					methodName, args));
			if (!fConn.contains(om)) {
				fConn.add(om);
			}
		} catch (NoSuchMethodException e) {
			//try to find another method with the name which fits to the argument types
			Method[] methods = o.getClass().getMethods();
			ObjectMethod om = null;
			boolean wasFound = false;
			for (Method method : methods) {
				if (method.getName().equals(methodName)) {
					Class<?>[] classes = method.getParameterTypes();
					int i;
					for (i = 0; i < args.length; i++) {
						if (!classes[i].isAssignableFrom(args[i])) {
							break;
						}
					}
					if (i == args.length) {
						om = new ObjectMethod(o, method);
						wasFound = true;
						if (!fConn.contains(om)) {
							fConn.add(om);
						}
						break;
					}
				}
			}
			if (!wasFound) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * unregister a previously registered call
	 *
	 * @param o the object having the method(s) not to be invoked any more
	 */
	public final void unregisterMethodCallOnEmit(final Object o) {
		Iterator<ObjectMethod> it = fConn.iterator();
		ObjectMethod om;
		while (it.hasNext()) {
			om = it.next();
			if (om.object == o) {
				it.remove();
			}
		}
	}

	/**
	 * fire the de.tu_darmstadt.rs.sim.signal
	 *
	 * @param args arguments to be transported to de.tu_darmstadt.rs.sim.signal handlers
	 */
	protected final void emitInternal(Object... args) {
		for (ObjectMethod om : fConn) {
			try {
				om.method.invoke(om.object, args);
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				e.getCause().printStackTrace();
			}
		}
	}

	/**
	 * unregister all previously registered calls
	 */
	public void unregisterAllMethodCalls() {
		fConn.clear();
	}
}
