/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package tests.engine;

import static org.junit.Assert.*;
import junit.framework.TestSuite;

import org.junit.BeforeClass;
import org.junit.Test;

import tests.mocks.SimulationControllerMock;

import de.tu_darmstadt.rs.sim.engine.AccessManager;
import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.gui.RefreshThread;

/**
 * Tests for the {@link AccessManager}
 */
public class AccessManagerTest extends TestSuite {
	private static AccessManager accessManager;
	private static AddressSpaceManager addressSpaceManager;
	private static SimulationControllerMock simulationController;

	/**
	 * called before the first dispatcher.test is started
	 */
	@BeforeClass
	public static void setUpBeforeClass() {
		simulationController = new SimulationControllerMock();
		accessManager = AccessManager.getTheInstance();
		accessManager.setSimulationController(simulationController);
		RefreshThread refreshThread = new RefreshThread(1);
		addressSpaceManager = new AddressSpaceManager(refreshThread,
				accessManager);
	}

	/**
	 * Test of the {@link AccessManager#getTheInstance()} method
	 */
	@Test
	public void testGetTheInstance() {
		accessManager = AccessManager.getTheInstance();
		assertNotNull(accessManager);
	}

	/**
	 * Test of the {@link AccessManager#checkForEvent(int)} method
	 */
	@Test
	public void testCheckForEvent() {
		int[] addresses = new int[] { 0x0, 0xaaaa, 0x1a001 };
		int numberOfElements;
		for (int i = 0; i < addresses.length; i++) {
			assertFalse(accessManager.hasWatchpoint(addresses[i]));
			accessManager.toggleWatchpoint(addresses[i]);
			accessManager.checkForEvent(addresses[i]);
			assertEquals("stopSimulation",
					simulationController.getLastCommand());
			simulationController.resetLastCommand();
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				accessManager.checkForEvent(j);
				if (!"none".equals(simulationController.getLastCommand())) {
					numberOfElements++;
				}
				simulationController.resetLastCommand();
			}
			assertEquals(1, numberOfElements);
			accessManager.toggleWatchpoint(addresses[i]);
			accessManager.checkForEvent(addresses[i]);
			assertEquals("none", simulationController.getLastCommand());
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				accessManager.checkForEvent(j);
				if (!"none".equals(simulationController.getLastCommand())) {
					numberOfElements++;
				}
			}
			assertEquals(0, numberOfElements);
		}
	}

	/**
	 * Test of the {@link AccessManager#toggleWatchpoint(int)} and
	 * {@link AccessManager#hasWatchpoint(int)} methods
	 */
	@Test
	public void testToggleWatchpointAndHasWatchpoint() {
		int[] addresses = new int[] { 0x0, 0xaaaa, 0x1a001 };
		int numberOfElements;
		for (int i = 0; i < addresses.length; i++) {
			assertFalse(accessManager.hasWatchpoint(addresses[i]));
			accessManager.toggleWatchpoint(addresses[i]);
			assertTrue(String.format(
					"watchpoint at address 0x%05x was not toggeled",
					addresses[i]), accessManager.hasWatchpoint(addresses[i]));
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				if (accessManager.hasWatchpoint(j)) {
					numberOfElements++;
				}
			}
			assertEquals(1, numberOfElements);
			accessManager.toggleWatchpoint(addresses[i]);
			assertFalse(accessManager.hasWatchpoint(addresses[i]));
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				if (accessManager.hasWatchpoint(j)) {
					numberOfElements++;
				}
			}
			assertEquals(0, numberOfElements);
		}
	}

	/**
	 * Test of the {@link AccessManager#toggleBreakpoint(int)} and
	 * {@link AccessManager#hasBreakpoint(int)} methods
	 */
	@Test
	public void testToggleBreakpointAndHasBreakpoint() {
		int[] addresses = new int[] { 0x0, 0xaaaa, 0x1a001 };
		int numberOfElements;
		for (int i = 0; i < addresses.length; i++) {
			assertFalse(accessManager.hasBreakpoint(addresses[i]));
			accessManager.toggleBreakpoint(addresses[i]);
			assertTrue(String.format(
					"breakpoint at address 0x%05x was not toggeled",
					addresses[i]), accessManager.hasBreakpoint(addresses[i]));
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				if (accessManager.hasBreakpoint(j)) {
					numberOfElements++;
				}
			}
			assertEquals(1, numberOfElements);
			accessManager.toggleBreakpoint(addresses[i]);
			assertFalse(accessManager.hasBreakpoint(addresses[i]));
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				if (accessManager.hasBreakpoint(j)) {
					numberOfElements++;
				}
			}
			assertEquals(0, numberOfElements);
		}
	}
}
