/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package tests.engine;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.Random;

import junit.framework.TestSuite;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Test;

import tests.mocks.PeripheralMock;
import tests.mocks.SimulationControllerMock;
import de.tu_darmstadt.rs.sim.engine.AccessManager;
import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.Settings;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.PipelineStage;
import de.tu_darmstadt.rs.sim.gui.RefreshThread;

/**
 * Tests for the {@link AddressSpaceManager}
 */
public class AddressSpaceManagerTest extends TestSuite {
	private static RefreshThread refreshThread;
	private static AddressSpaceManager addressSpaceManager;
	private static AccessManager accessManager;
	private static SimulationControllerMock simulationController;
	private static PeripheralMock fPeripheralMock;

	/**
	 * called after each dispatcher.test
	 */
	@After
	public void afterTest() {
		addressSpaceManager.delete();
		for (int i = 0; i < addressSpaceManager.getSize(); i++) {
			if (addressSpaceManager.hasBreakpoint(i)) {
				addressSpaceManager.toggleBreakpoint(i);
			}
			if (addressSpaceManager.hasWatchpoint(i)) {
				addressSpaceManager.toggleWatchpoint(i);
			}
		}
	}

	/**
	 * called before the first dispatcher.test is started
	 */
	@BeforeClass
	public static void setUpBeforeClass() {
		simulationController = new SimulationControllerMock();
		accessManager = AccessManager.getTheInstance();
		accessManager.setSimulationController(simulationController);
		refreshThread = new RefreshThread(1);
		addressSpaceManager = new AddressSpaceManager(refreshThread,
				accessManager);
		fPeripheralMock = new PeripheralMock(0x1a000, addressSpaceManager,
				refreshThread);
		fPeripheralMock.setup();
		fPeripheralMock.resetLastCommand();
	}

	/**
	 * called after the last dispatcher.test has finished
	 */
	@AfterClass
	public static void afterClass() {
		fPeripheralMock.tearDown();
	}

	/**
	 * Test of
	 * {@link AddressSpaceManager#AddressSpaceManager(engine.RefreshHandler_I, AccessManager)}
	 */
	@Test
	public void testAddressSpaceManager() {
		AddressSpaceManager addressSpaceManagerObj = new AddressSpaceManager(
				refreshThread, accessManager);
		assertNotNull(addressSpaceManagerObj);
	}

	/**
	 * Test of the {@link AddressSpaceManager#getInstructionPort()} method
	 */
	@Test
	public void testGetInstructionPort() {
		assertNotNull(addressSpaceManager.getInstructionPort());
	}

	/**
	 * Test of the {@link AddressSpaceManager#writeTo(int, int)} and
	 * {@link AddressSpaceManager#readFrom(int)} methods
	 */
	@Test
	public void testWriteToAndReadFrom() {
		fPeripheralMock.resetLastCommand();
		Random random = new Random();
		int aValue;
		int watchpointAddress = 0x15555;
		int breakpointAddress = 0xaaaa;
		addressSpaceManager.toggleWatchpoint(watchpointAddress);
		addressSpaceManager.toggleBreakpoint(breakpointAddress);
		assertTrue(addressSpaceManager.hasWatchpoint(watchpointAddress));
		assertTrue(addressSpaceManager.hasBreakpoint(breakpointAddress));
		for (int i = 0; i < addressSpaceManager.getSize(); i++) {
			aValue = random.nextInt(1 << 18);
			assertEquals("none", fPeripheralMock.getLastCommand());
			// simulation controller should not be called since each component
			// should know if it checks for events
			assertEquals("none", simulationController.getLastCommand());
			addressSpaceManager.writeTo(i, aValue);
			if (i < (Integer) Settings.getSetting(Settings.IO_ADDRESS_MIN)) {
				assertEquals(aValue, addressSpaceManager.readFrom(i));
			} else if (i >= 0x1a000 && i < 0x1a004) {
				assertEquals("write at " + i + ": " + aValue,
						fPeripheralMock.getLastCommand());
				fPeripheralMock.resetLastCommand();
				addressSpaceManager.readFrom(i);
				assertEquals("read at " + i, fPeripheralMock.getLastCommand());
				fPeripheralMock.resetLastCommand();
			}
		}
		addressSpaceManager.toggleWatchpoint(watchpointAddress);
		addressSpaceManager.toggleBreakpoint(breakpointAddress);
	}

	/**
	 * Test of the {@link AddressSpaceManager#setValue(int, int)} and
	 * {@link AddressSpaceManager#getValue(int)} methods
	 */
	@Test
	public void testSetValueAndGetValue() {
		Random random = new Random();
		int aValue;
		for (int i = 0; i < addressSpaceManager.getSize(); i++) {
			aValue = random.nextInt(1 << 18);
			assertEquals("none", fPeripheralMock.getLastCommand());
			addressSpaceManager.setValue(i, aValue);
			if (i < (Integer) Settings.getSetting(Settings.IO_ADDRESS_MIN)) {
				assertEquals(aValue, addressSpaceManager.getValue(i));
			} else if (i >= 0x1a000 && i < 0x1a004) {
				assertEquals("set value at " + i + ": " + aValue,
						fPeripheralMock.getLastCommand());
				fPeripheralMock.resetLastCommand();
				addressSpaceManager.getValue(i);
				assertEquals("get value at " + i,
						fPeripheralMock.getLastCommand());
				fPeripheralMock.resetLastCommand();
			}
		}
	}

	/**
	 * Test of the {@link AddressSpaceManager#toggleWatchpoint(int)} and
	 * {@link AddressSpaceManager#hasWatchpoint(int)} methods
	 */
	@Test
	public void testToggleWatchpointAndHasWatchpoint() {
		int[] addresses = new int[] { 0x0, 0xaaaa, 0x1a001 };
		int numberOfElements;
		for (int i = 0; i < addresses.length; i++) {
			assertFalse(addressSpaceManager.hasWatchpoint(addresses[i]));
			addressSpaceManager.toggleWatchpoint(addresses[i]);
			assertTrue(String.format(
					"watchpoint at address 0x%05x was not toggeled",
					addresses[i]),
					addressSpaceManager.hasWatchpoint(addresses[i]));
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				if (addressSpaceManager.hasWatchpoint(j)) {
					numberOfElements++;
				}
			}
			assertEquals(1, numberOfElements);
			addressSpaceManager.toggleWatchpoint(addresses[i]);
			assertFalse(addressSpaceManager.hasWatchpoint(addresses[i]));
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				if (addressSpaceManager.hasWatchpoint(j)) {
					numberOfElements++;
				}
			}
			assertEquals(0, numberOfElements);
		}
	}

	/**
	 * Test of the {@link AddressSpaceManager#toggleBreakpoint(int)} and
	 * {@link AddressSpaceManager#hasBreakpoint(int)} methods
	 */
	@Test
	public void testToggleBreakpointAndHasBreakpoint() {
		int[] addresses = new int[] { 0x0, 0xaaaa, 0x1a001 };
		int numberOfElements;
		for (int i = 0; i < addresses.length; i++) {
			assertFalse(addressSpaceManager.hasBreakpoint(addresses[i]));
			addressSpaceManager.toggleBreakpoint(addresses[i]);
			assertTrue(String.format(
					"breakpoint at address 0x%05x was not toggeled",
					addresses[i]),
					addressSpaceManager.hasBreakpoint(addresses[i]));
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				if (addressSpaceManager.hasBreakpoint(j)) {
					numberOfElements++;
				}
			}
			assertEquals(1, numberOfElements);
			addressSpaceManager.toggleBreakpoint(addresses[i]);
			assertFalse(addressSpaceManager.hasBreakpoint(addresses[i]));
			numberOfElements = 0;
			for (int j = 0; j < addressSpaceManager.getSize(); j++) {
				if (addressSpaceManager.hasBreakpoint(j)) {
					numberOfElements++;
				}
			}
			assertEquals(0, numberOfElements);
		}
	}

	/**
	 * Test of the {@link AddressSpaceManager#getPipeStage(int)} and
	 * {@link AddressSpaceManager#setIfIdStage(int)} methods
	 */
	@Test
	public void testGetPipeStageAndSetIfIdStage() {
		addressSpaceManager.restart();
		for (int i = 0; i < addressSpaceManager.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE,
					addressSpaceManager.getPipeStage(i));
		}
		addressSpaceManager.setIfIdStage(0x0);
		for (int i = 1; i < addressSpaceManager.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE,
					addressSpaceManager.getPipeStage(i));
		}
		assertEquals(PipelineStage.IFID_STAGE,
				addressSpaceManager.getPipeStage(0));
		addressSpaceManager.setIfIdStage(0x1);
		for (int i = 2; i < addressSpaceManager.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE,
					addressSpaceManager.getPipeStage(i));
		}
		assertEquals(PipelineStage.EXMEM_STAGE,
				addressSpaceManager.getPipeStage(0));
		assertEquals(PipelineStage.IFID_STAGE,
				addressSpaceManager.getPipeStage(1));
		addressSpaceManager.setIfIdStage(0x2);
		for (int i = 3; i < addressSpaceManager.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE,
					addressSpaceManager.getPipeStage(i));
		}
		assertEquals(PipelineStage.WB_STAGE,
				addressSpaceManager.getPipeStage(0));
		assertEquals(PipelineStage.EXMEM_STAGE,
				addressSpaceManager.getPipeStage(1));
		assertEquals(PipelineStage.IFID_STAGE,
				addressSpaceManager.getPipeStage(2));
		addressSpaceManager.setIfIdStage(0x3);
		for (int i = 4; i < addressSpaceManager.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE,
					addressSpaceManager.getPipeStage(i));
		}
		assertEquals(PipelineStage.NOT_IN_PIPE,
				addressSpaceManager.getPipeStage(0));
		assertEquals(PipelineStage.WB_STAGE,
				addressSpaceManager.getPipeStage(1));
		assertEquals(PipelineStage.EXMEM_STAGE,
				addressSpaceManager.getPipeStage(2));
		assertEquals(PipelineStage.IFID_STAGE,
				addressSpaceManager.getPipeStage(3));
	}

	/**
	 * Test of the {@link AddressSpaceManager#delete()} method
	 */
	@Test
	public void testDelete() {
		Random rand = new Random();
		int aValue;
		for (int j = 0; j < addressSpaceManager.getSize(); j++) {
			aValue = rand.nextInt(addressSpaceManager.getSize())
					& SpartanDefinitions_I.BITMASK_18;
			addressSpaceManager.setValue(j, aValue);
		}
		addressSpaceManager.delete();
		for (int i = 0; i < addressSpaceManager.getSize(); i++) {
			assertEquals(
					"expected <0> but was <" + addressSpaceManager.getValue(i)
							+ "> (i=" + i + ")", 0,
					addressSpaceManager.getValue(i));
		}
	}

	/**
	 * Test of the {@link AddressSpaceManager#getSize()} method
	 */
	@Test
	public void testGetSize() {
		assertEquals(0, Settings.getSetting(Settings.MEMORY_ADDRESS_MIN));
		assertEquals(0x1ffff, Settings.getSetting(Settings.IO_ADDRESS_MAX));
		assertEquals(
				(Integer) Settings.getSetting(Settings.IO_ADDRESS_MAX) + 1,
				addressSpaceManager.getSize());
	}

	/**
	 * Test of the {@link AddressSpaceManager#restart()} method
	 */
	@Test
	public void testRestart() {
		addressSpaceManager.restart();
		assertEquals("restarted", fPeripheralMock.getLastCommand());
		fPeripheralMock.resetLastCommand();
	}

	/**
	 * Test of the {@link AddressSpaceManager#setup()} method
	 */
	@Test
	public void testSetup() {
		addressSpaceManager.setup();
		assertEquals("setup", fPeripheralMock.getLastCommand());
		fPeripheralMock.resetLastCommand();
	}
}
