/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package tests.engine.memory;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.Random;

import junit.framework.TestSuite;

import org.junit.After;
import org.junit.BeforeClass;
import org.junit.Test;

import de.tu_darmstadt.rs.sim.engine.Settings;
import de.tu_darmstadt.rs.sim.engine.SpartanDefinitions_I;
import de.tu_darmstadt.rs.sim.engine.memory.MainMemory;
import de.tu_darmstadt.rs.sim.engine.processor.pipeline.PipelineStage;
import de.tu_darmstadt.rs.sim.gui.RefreshThread;

/**
 * Tests for the {@link MainMemory}
 */
public class MainMemoryTest extends TestSuite {
	private static MainMemory mainMemory;
	private static RefreshThread refreshThread;

	/**
	 * called before the first dispatcher.test is started
	 */
	@BeforeClass
	public static void beforeClass() {
		refreshThread = new RefreshThread(10000);
		mainMemory = new MainMemory(refreshThread);
	}

	/**
	 * called after each dispatcher.test
	 */
	@After
	public void afterTest() {
		mainMemory.delete();
	}

	/**
	 * Test of the {@link MainMemory#MainMemory(engine.RefreshHandler_I)} method
	 */
	@Test
	public void testMainMemory() {
		mainMemory = new MainMemory(refreshThread);
		assertTrue(mainMemory != null);
	}

	/**
	 * Test of the {@link MainMemory#getSize()} method
	 */
	@Test
	public void testGetSize() {
		assertEquals(0, Settings.getSetting(Settings.MEMORY_ADDRESS_MIN));
		assertEquals(0x18fff, Settings.getSetting(Settings.MEMORY_ADDRESS_MAX));
		assertEquals(
				(Integer) Settings.getSetting(Settings.MEMORY_ADDRESS_MAX) + 1,
				mainMemory.getSize());
	}

	/**
	 * Test of the {@link MainMemory#delete()} method
	 */
	@Test
	public void testDelete() {
		Random rand = new Random();
		int aValue;
		for (int j = 0; j < 500; j++) {
			aValue = rand.nextInt(mainMemory.getSize())
					& SpartanDefinitions_I.BITMASK_18;
			mainMemory.setValueAt(j, aValue);
		}
		mainMemory.delete();
		for (int i = 0; i < mainMemory.getSize(); i++) {
			assertEquals(0, mainMemory.getValueAt(i));
		}
	}

	/**
	 * Test of the {@link MainMemory#setValueAt(int, int)} and
	 * {@link MainMemory#getValueAt(int)} methods
	 */
	@Test
	public void testSetValueAtAndGetValueAt() {
		for (int i = 0; i < mainMemory.getSize(); i++) {
			assertEquals(0, mainMemory.getValueAt(i));
			mainMemory.setValueAt(i, 0xaaaa);
			assertEquals(0xaaaa, mainMemory.getValueAt(i));
		}
	}

	/**
	 * Test of the {@link MainMemory#setIfIdStage(int)} and
	 * {@link MainMemory#getPipeStage(int)} methods
	 */
	@Test
	public void testSetIfIdStageAndGetPipeStage() {
		mainMemory.restart();
		for (int i = 0; i < mainMemory.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE, mainMemory.getPipeStage(i));
		}
		mainMemory.setIfIdStage(0x0);
		for (int i = 1; i < mainMemory.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE, mainMemory.getPipeStage(i));
		}
		assertEquals(PipelineStage.IFID_STAGE, mainMemory.getPipeStage(0));
		mainMemory.setIfIdStage(0x1);
		for (int i = 2; i < mainMemory.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE, mainMemory.getPipeStage(i));
		}
		assertEquals(PipelineStage.EXMEM_STAGE, mainMemory.getPipeStage(0));
		assertEquals(PipelineStage.IFID_STAGE, mainMemory.getPipeStage(1));
		mainMemory.setIfIdStage(0x2);
		for (int i = 3; i < mainMemory.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE, mainMemory.getPipeStage(i));
		}
		assertEquals(PipelineStage.WB_STAGE, mainMemory.getPipeStage(0));
		assertEquals(PipelineStage.EXMEM_STAGE, mainMemory.getPipeStage(1));
		assertEquals(PipelineStage.IFID_STAGE, mainMemory.getPipeStage(2));
		mainMemory.setIfIdStage(0x3);
		for (int i = 4; i < mainMemory.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE, mainMemory.getPipeStage(i));
		}
		assertEquals(PipelineStage.NOT_IN_PIPE, mainMemory.getPipeStage(0));
		assertEquals(PipelineStage.WB_STAGE, mainMemory.getPipeStage(1));
		assertEquals(PipelineStage.EXMEM_STAGE, mainMemory.getPipeStage(2));
		assertEquals(PipelineStage.IFID_STAGE, mainMemory.getPipeStage(3));
	}

	/**
	 * Test of the {@link MainMemory#containsAddress(int)} method
	 */
	@Test
	public void testContainsAddress() {
		for (int i = 0; i < mainMemory.getSize(); i++) {
			assertTrue(MainMemory.containsAddress(i));
		}
	}

	/**
	 * Test of the {@link MainMemory#setup()} and {@link MainMemory#restart()}
	 * method
	 */
	@Test
	public void testSetupAndRestart() {
		mainMemory.setIfIdStage(0x0);
		mainMemory.setup();
		for (int i = 0; i < mainMemory.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE, mainMemory.getPipeStage(i));
		}
		mainMemory.setIfIdStage(0x0);
		mainMemory.restart();
		for (int i = 0; i < mainMemory.getSize(); i++) {
			assertEquals(PipelineStage.NOT_IN_PIPE, mainMemory.getPipeStage(i));
		}
	}

}
