/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package tests.engine.parser;

import static org.junit.Assert.*;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import junit.framework.TestSuite;

import org.junit.BeforeClass;
import org.junit.Test;

import tests.mocks.RefreshHandlerMock;
import tests.mocks.SimulationControllerMock;

import de.tu_darmstadt.rs.sim.engine.AccessManager;
import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.parser.AssemblyParser;
import de.tu_darmstadt.rs.sim.engine.parser.ParsingException;

/**
 * Tests for the {@link AssemblyParser}
 */
public class AssemblyParserTest extends TestSuite {
	private static AddressSpaceManager addressSpaceManager;
	private static RefreshHandlerMock refreshHandler;
	private static AccessManager accessManager;
	private static SimulationControllerMock simulationController;
	private static AssemblyParser assemblyParser;
	private static final String[] sphCode = new String[] { "15000", "1b06e",
			"02105", "15200", "1b25f", "02122", "02015", "00000", "00026",
			"0004a", "0002c", "00032", "00038", "0003e", "00044", "34070",
			"34040", "00000", "151a0", "1b010", "03c7f", "25300", "1921f",
			"15400", "1b40e", "27d40", "09206", "29201", "03351", "25320",
			"02125", "01965", "15400", "1b40e", "25d40", "03c7e", "02160",
			"00057", "15800", "1b822", "12800", "26980", "02164", "02015",
			"15800", "1b822", "12800", "26980", "02164", "02015", "15800",
			"1b822", "12800", "26980", "02164", "02015", "15800", "1b822",
			"12800", "26980", "02164", "02015", "15800", "1b822", "12800",
			"26980", "02164", "02015", "15800", "1b822", "12800", "26980",
			"02164", "02015", "101fe", "15800", "1b820", "25980", "25980",
			"27800", "15a00", "1ba22", "259a0", "11801", "1280f", "0188c",
			"279a0", "0e004", "12800", "269a0", "01885", "0602b", "10002",
			"02164", "02015", "0601e", "15800", "1b81e", "25980", "1280f",
			"26980", "12803", "26982", "15800", "1b820", "25980", "1280f",
			"26980", "0600e", "04000", "14000", "1a024", "2c001", "0209e",
			"14047", "1a1ff", "2c001", "02164", "28001", "0289e", "02164",
			"02015", "00057", "02164", "02015", "00056", "02164", "02015",
			"0283e", "02164", "02015", "0283f", "02164", "02015", "3880f",
			"02015", "0c009", "14a00", "1ab2a", "028b0", "20880", "189ff",
			"0283e", "02164", "02015", "12800", "0283e", "02164", "02015",
			"07e06", "0b64f", "0cc6d", "0fa07", "0fe6f", "0ee7c", "0b05e",
			"0f271" };

	/**
	 * called before the first dispatcher.test is started
	 */
	@BeforeClass
	public static void setUpBeforeClass() {
		refreshHandler = new RefreshHandlerMock();
		accessManager = AccessManager.getTheInstance();
		simulationController = new SimulationControllerMock();
		accessManager.setSimulationController(simulationController);
		addressSpaceManager = new AddressSpaceManager(refreshHandler,
				accessManager);
		assemblyParser = new AssemblyParser(addressSpaceManager);
	}

	/**
	 * Test of the {@link AssemblyParser#AssemblyParser(AddressSpaceManager)}
	 * method
	 */
	@Test
	public void testAssemblyParser() {
		assemblyParser = new AssemblyParser(addressSpaceManager);
		assertNotNull(assemblyParser);
	}

	/**
	 * Test of the {@link AssemblyParser#loadAssemblyCode(File)} method
	 */
	@Test
	public void testLoadAssemblyCode() {
		try {
			File listingFile = File.createTempFile("listing", null);
			FileWriter writer = new FileWriter(listingFile);
			for (String line : sphCode) {
				writer.write(line + "\n");
			}
			writer.close();
			addressSpaceManager.restart();
			assemblyParser.loadAssemblyCode(listingFile);
			for (int i = 0; i < sphCode.length; i++) {
				assertEquals(sphCode[i],
						String.format("%05x", addressSpaceManager.getValue(i)));
			}
			listingFile.deleteOnExit();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ParsingException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Test of the {@link AssemblyParser#storeAssemblyCode(File)} method
	 */
	@Test
	public void testStoreAssemblyCode() {
		try {
			File listingFile = File.createTempFile("listing", null);
			FileWriter writer = new FileWriter(listingFile);
			for (String line : sphCode) {
				writer.write(line + "\n");
			}
			writer.close();
			addressSpaceManager.restart();
			assemblyParser.loadAssemblyCode(listingFile);
			listingFile.delete();
			listingFile = File.createTempFile("listing", null);
			assemblyParser.storeAssemblyCode(listingFile);
			BufferedReader reader = new BufferedReader(new FileReader(
					listingFile));
			int index = 0;
			String line;
			while ((line = reader.readLine()) != null) {
				assertEquals(sphCode[index], line);
				index++;
			}
			listingFile.deleteOnExit();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ParsingException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

}
