/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */
package tests.engine.parser;

import static org.junit.Assert.*;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;

import junit.framework.TestSuite;

import org.junit.BeforeClass;
import org.junit.Test;

import tests.mocks.RefreshHandlerMock;
import tests.mocks.SimulationControllerMock;

import de.tu_darmstadt.rs.sim.engine.AccessManager;
import de.tu_darmstadt.rs.sim.engine.AddressSpaceManager;
import de.tu_darmstadt.rs.sim.engine.assembly.AssemblyLine_I;
import de.tu_darmstadt.rs.sim.engine.parser.LstFileParser;
import de.tu_darmstadt.rs.sim.engine.parser.ParsingException;

/**
 * Tests for the {@link LstFileParser}
 */
public class LstFileParserTest extends TestSuite {
	private static LstFileParser lstFileParser;
	private static AddressSpaceManager addressSpaceManager;
	private static RefreshHandlerMock refreshHandler;
	private static AccessManager accessManager;
	private static SimulationControllerMock simulationController;
	private static final String[] programListing = new String[] {
			"\n",
			"main.elf:     file format elf32-spmc\n",
			"\n",
			"\n",
			"Disassembly of section .startup:\n",
			"\n",
			"00000000 <_startup>:\n",
			"   0:	00 01 50 00 	lhi	r8, 0	; 0x00\n",
			"   2:	00 01 b0 6e 	ori	r8, 110	; 0x6e\n",
			"   4:	00 00 21 05 	jalrs	r8\n",
			"   6:	00 01 52 00 	lhi	r9, 0	; 0x00\n",
			"   8:	00 01 b2 5f 	ori	r9, 95	; 0x5f\n",
			"   a:	00 00 21 22 	jr	r9\n",
			"   c:	00 00 20 15 	or	r0, r0\n",
			"\n",
			"Disassembly of section .data:\n",
			"\n",
			"0000000e <_intr_tab>:\n",
			"   e:	00 00       	.w9	0x00\n",
			"   f:	00 00       	.w9	0x00\n",
			"  10:	00 00       	.w9	0x00\n",
			"  11:	00 26       	.w9	0x26	; '&'\n",
			"  12:	00 00       	.w9	0x00\n",
			"  13:	00 4a       	.w9	0x4a	; 'J'\n",
			"  14:	00 00       	.w9	0x00\n",
			"  15:	00 2c       	.w9	0x2c	; ','\n",
			"  16:	00 00       	.w9	0x00\n",
			"  17:	00 32       	.w9	0x32	; '2'\n",
			"  18:	00 00       	.w9	0x00\n",
			"  19:	00 38       	.w9	0x38	; '8'\n",
			"  1a:	00 00       	.w9	0x00\n",
			"  1b:	00 3e       	.w9	0x3e	; '>'\n",
			"  1c:	00 00       	.w9	0x00\n",
			"  1d:	00 44       	.w9	0x44	; 'D'\n",
			"\n",
			"0000001e <t>:\n",
			"  1e:	01 a0       	.w9	0x1a0\n",
			"  1f:	00 70       	.w9	0x70	; 'p'\n",
			"\n",
			"00000020 <rt0>:\n",
			"  20:	01 a0       	.w9	0x1a0\n",
			"  21:	00 40       	.w9	0x40	; '@'\n",
			"\n",
			"00000022 <i>:\n",
			"  22:	00 00       	.w9	0x00\n",
			"  23:	00 00       	.w9	0x00\n",
			"\n",
			"Disassembly of section interrupt:\n",
			"\n",
			"00000024 <_intr_handle>:\n",
			"  24:	00 01 51 a0 	lhi	r8, 416	; 0x1a0\n",
			"  26:	00 01 b0 10 	ori	r8, 16	; 0x10\n",
			"  28:	00 00 3c 7f 	movs2i	r14, sfr_cc\n",
			"  2a:	00 02 53 00 	l18	r9, 0(r8)\n",
			"  2c:	00 01 92 1f 	andi	r9, 31	; 0x1f\n",
			"  2e:	00 01 54 00 	lhi	r10, 0	; 0x00\n",
			"  30:	00 01 b4 0e 	ori	r10, 14	; 0x0e\n",
			"  32:	00 02 7d 40 	s18	0(r10), r14\n",
			"  34:	00 00 92 06 	beqz	r9, 0x00040	; 0x40 <_intr_exit>\n",
			"  36:	00 02 92 01 	slli	r9, 1	; 0x01\n",
			"  38:	00 00 33 51 	addu	r9, r10\n",
			"  3a:	00 02 53 20 	l18	r9, 0(r9)\n",
			"  3c:	00 00 21 25 	jalrs	r9\n",
			"  3e:	00 00 19 65 	mov	r12, r11\n",
			"\n",
			"00000040 <_intr_exit>:\n",
			"  40:	00 01 54 00 	lhi	r10, 0	; 0x00\n",
			"  42:	00 01 b4 0e 	ori	r10, 14	; 0x0e\n",
			"  44:	00 02 5d 40 	l18	r14, 0(r10)\n",
			"  46:	00 00 3c 7e 	movi2s	sfr_cc, r14\n",
			"  48:	00 00 21 60 	rfe	r11\n",
			"  4a:	00 00 00 57 	sbits	INT\n",
			"\n",
			"Disassembly of section .text:\n",
			"\n",
			"0000004c <isr00>:\n",
			"  4c:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  4e:	00 01 b8 22 	ori	r12, 34	; 0x22\n",
			"  50:	00 01 28 00 	movi	r4, 0x00	; 0x0 <SBBOOL_NO>\n",
			"  52:	00 02 69 80 	s18	0(r12), r4\n",
			"  54:	00 00 21 64 	jrs	r11\n",
			"  56:	00 00 20 15 	or	r0, r0\n",
			"\n",
			"00000058 <isr02>:\n",
			"  58:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  5a:	00 01 b8 22 	ori	r12, 34	; 0x22\n",
			"  5c:	00 01 28 00 	movi	r4, 0x00	; 0x0 <SBBOOL_NO>\n",
			"  5e:	00 02 69 80 	s18	0(r12), r4\n",
			"  60:	00 00 21 64 	jrs	r11\n",
			"  62:	00 00 20 15 	or	r0, r0\n",
			"\n",
			"00000064 <isr03>:\n",
			"  64:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  66:	00 01 b8 22 	ori	r12, 34	; 0x22\n",
			"  68:	00 01 28 00 	movi	r4, 0x00	; 0x0 <SBBOOL_NO>\n",
			"  6a:	00 02 69 80 	s18	0(r12), r4\n",
			"  6c:	00 00 21 64 	jrs	r11\n",
			"  6e:	00 00 20 15 	or	r0, r0\n",
			"\n",
			"00000070 <isr04>:\n",
			"  70:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  72:	00 01 b8 22 	ori	r12, 34	; 0x22\n",
			"  74:	00 01 28 00 	movi	r4, 0x00	; 0x0 <SBBOOL_NO>\n",
			"  76:	00 02 69 80 	s18	0(r12), r4\n",
			"  78:	00 00 21 64 	jrs	r11\n",
			"  7a:	00 00 20 15 	or	r0, r0\n",
			"\n",
			"0000007c <isr05>:\n",
			"  7c:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  7e:	00 01 b8 22 	ori	r12, 34	; 0x22\n",
			"  80:	00 01 28 00 	movi	r4, 0x00	; 0x0 <SBBOOL_NO>\n",
			"  82:	00 02 69 80 	s18	0(r12), r4\n",
			"  84:	00 00 21 64 	jrs	r11\n",
			"  86:	00 00 20 15 	or	r0, r0\n",
			"\n",
			"00000088 <isr06>:\n",
			"  88:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  8a:	00 01 b8 22 	ori	r12, 34	; 0x22\n",
			"  8c:	00 01 28 00 	movi	r4, 0x00	; 0x0 <SBBOOL_NO>\n",
			"  8e:	00 02 69 80 	s18	0(r12), r4\n",
			"  90:	00 00 21 64 	jrs	r11\n",
			"  92:	00 00 20 15 	or	r0, r0\n",
			"\n",
			"00000094 <isr01>:\n",
			"  94:	00 01 01 fe 	addi	r0, -2	; 0x1fe\n",
			"  96:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  98:	00 01 b8 20 	ori	r12, 32	; 0x20\n",
			"  9a:	00 02 59 80 	l18	r12, 0(r12)\n",
			"  9c:	00 02 59 80 	l18	r12, 0(r12)\n",
			"  9e:	00 02 78 00 	s18	0(r0), r12\n",
			"  a0:	00 01 5a 00 	lhi	r13, 0	; 0x00\n",
			"  a2:	00 01 ba 22 	ori	r13, 34	; 0x22\n",
			"  a4:	00 02 59 a0 	l18	r12, 0(r13)\n",
			"  a6:	00 01 18 01 	addi	r12, 1	; 0x01\n",
			"  a8:	00 01 28 0f 	movi	r4, 0x0f	; 0xf <SBI_SECTION_RODATA_BYTES>\n",
			"  aa:	00 00 18 8c 	sleu	r12, r4\n",
			"  ac:	00 02 79 a0 	s18	0(r13), r12\n",
			"  ae:	00 00 e0 04 	bnezc	0x000b6	; 0xb6 <L15>\n",
			"  b0:	00 01 28 00 	movi	r4, 0x00	; 0x0 <SBBOOL_NO>\n",
			"  b2:	00 02 69 a0 	s18	0(r13), r4\n",
			"  b4:	00 00 18 85 	mov	r12, r4\n",
			"\n",
			"000000b6 <L15>:\n",
			"  b6:	00 00 60 2b 	jals	0x0010c	; 0x10c <led7_hex>\n",
			"  b8:	00 01 00 02 	addi	r0, 2	; 0x02\n",
			"  ba:	00 00 21 64 	jrs	r11\n",
			"  bc:	00 00 20 15 	or	r0, r0\n",
			"\n",
			"000000be <main>:\n",
			"  be:	00 00 60 1e 	jals	0x000fa	; 0xfa <interrupt_disable>\n",
			"  c0:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  c2:	00 01 b8 1e 	ori	r12, 30	; 0x1e\n",
			"  c4:	00 02 59 80 	l18	r12, 0(r12)\n",
			"  c6:	00 01 28 0f 	movi	r4, 0x0f	; 0xf <SBI_SECTION_RODATA_BYTES>\n",
			"  c8:	00 02 69 80 	s18	0(r12), r4\n",
			"  ca:	00 01 28 03 	movi	r4, 0x03	; 0x3 <SBI_VERSION+0x1>\n",
			"  cc:	00 02 69 82 	s18	2(r12), r4\n",
			"  ce:	00 01 58 00 	lhi	r12, 0	; 0x00\n",
			"  d0:	00 01 b8 20 	ori	r12, 32	; 0x20\n",
			"  d2:	00 02 59 80 	l18	r12, 0(r12)\n",
			"  d4:	00 01 28 0f 	movi	r4, 0x0f	; 0xf <SBI_SECTION_RODATA_BYTES>\n",
			"  d6:	00 02 69 80 	s18	0(r12), r4\n",
			"  d8:	00 00 60 0e 	jals	0x000f4	; 0xf4 <interrupt_enable>\n",
			"\n", "000000da <L20>:\n",
			"  da:	00 00 40 00 	j	0x000da	; 0xda <L20>\n", "\n",
			"000000dc <_init>:\n", "  dc:	00 01 40 00 	lhi	r0, 0	; 0x00\n",
			"  de:	00 01 a0 24 	ori	r0, 36	; 0x24\n",
			"  e0:	00 02 c0 01 	srli	r0, 1	; 0x01\n",
			"  e2:	00 00 20 9e 	movi2s	sfr_iv, r0\n",
			"  e4:	00 01 40 47 	lhi	r0, 71	; 0x47\n",
			"  e6:	00 01 a1 ff 	ori	r0, 511	; 0x1ff\n",
			"  e8:	00 02 c0 01 	srli	r0, 1	; 0x01\n",
			"  ea:	00 00 21 64 	jrs	r11\n",
			"  ec:	00 02 80 01 	slli	r0, 1	; 0x01\n", "\n",
			"000000ee <set_iv>:\n", "  ee:	00 00 28 9e 	movi2s	sfr_iv, r4\n",
			"  f0:	00 00 21 64 	jrs	r11\n", "  f2:	00 00 20 15 	or	r0, r0\n",
			"\n", "000000f4 <interrupt_enable>:\n",
			"  f4:	00 00 00 57 	sbits	INT\n", "  f6:	00 00 21 64 	jrs	r11\n",
			"  f8:	00 00 20 15 	or	r0, r0\n", "\n",
			"000000fa <interrupt_disable>:\n",
			"  fa:	00 00 00 56 	cbits	INT\n", "  fc:	00 00 21 64 	jrs	r11\n",
			"  fe:	00 00 20 15 	or	r0, r0\n", "\n", "00000100 <led7_set>:\n",
			" 100:	00 00 28 3e 	movi2s	sfr_leds, r4\n",
			" 102:	00 00 21 64 	jrs	r11\n", " 104:	00 00 20 15 	or	r0, r0\n",
			"\n", "00000106 <led7_get>:\n",
			" 106:	00 00 28 3f 	movs2i	r4, sfr_leds\n",
			" 108:	00 00 21 64 	jrs	r11\n", " 10a:	00 00 20 15 	or	r0, r0\n",
			"\n", "0000010c <led7_hex>:\n",
			" 10c:	00 03 88 0f 	slei	r4, 15	; 0x0f\n",
			" 10e:	00 00 20 15 	or	r0, r0\n",
			" 110:	00 00 c0 09 	beqzc	0x00122	; 0x122 <L10>\n",
			" 112:	00 01 4a 00 	lhi	r5, 0	; 0x00\n",
			" 114:	00 01 ab 2a 	ori	r5, 298	; 0x12a\n",
			" 116:	00 00 28 b0 	add	r4, r5\n",
			" 118:	00 02 08 80 	l9	r4, 0(r4)\n",
			" 11a:	00 01 89 ff 	andi	r4, 511	; 0x1ff\n",
			" 11c:	00 00 28 3e 	movi2s	sfr_leds, r4\n",
			" 11e:	00 00 21 64 	jrs	r11\n", " 120:	00 00 20 15 	or	r0, r0\n",
			"\n", "00000122 <L10>:\n",
			" 122:	00 01 28 00 	movi	r4, 0x00	; 0x0 <SBBOOL_NO>\n",
			" 124:	00 00 28 3e 	movi2s	sfr_leds, r4\n",
			" 126:	00 00 21 64 	jrs	r11\n", " 128:	00 00 20 15 	or	r0, r0\n",
			"\n", "Disassembly of section rodata:\n", "\n",
			"0000012a <led7>:\n", " 12a:	00 3f       	.w9	0x3f	; '?'\n",
			" 12b:	00 06       	.w9	0x06\n",
			" 12c:	00 5b       	.w9	0x5b	; '['\n",
			" 12d:	00 4f       	.w9	0x4f	; 'O'\n",
			" 12e:	00 66       	.w9	0x66	; 'f'\n",
			" 12f:	00 6d       	.w9	0x6d	; 'm'\n",
			" 130:	00 7d       	.w9	0x7d	; '}'\n",
			" 131:	00 07       	.w9	0x07\n", " 132:	00 7f       	.w9	0x7f\n",
			" 133:	00 6f       	.w9	0x6f	; 'o'\n",
			" 134:	00 77       	.w9	0x77	; 'w'\n",
			" 135:	00 7c       	.w9	0x7c	; '|'\n",
			" 136:	00 58       	.w9	0x58	; 'X'\n",
			" 137:	00 5e       	.w9	0x5e	; '^'\n",
			" 138:	00 79       	.w9	0x79	; 'y'\n",
			" 139:	00 71       	.w9	0x71	; 'q'\n" };

	/**
	 * called before the first dispatcher.test is started
	 */
	@BeforeClass
	public static void setUpBeforeClass() {
		refreshHandler = new RefreshHandlerMock();
		accessManager = AccessManager.getTheInstance();
		simulationController = new SimulationControllerMock();
		accessManager.setSimulationController(simulationController);
		addressSpaceManager = new AddressSpaceManager(refreshHandler,
				accessManager);
		LstFileParser.createTheInstance(addressSpaceManager);
		lstFileParser = LstFileParser.getTheInstance();
	}

	/**
	 * Test of the {@link LstFileParser#loadLstFile(File)} method
	 */
	@Test
	public void testDataAfterLoadLstFile() {
		try {
			File listingFile = File.createTempFile("listing", null);
			FileWriter writer = new FileWriter(listingFile);
			for (String line : programListing) {
				writer.write(line);
			}
			writer.close();
			lstFileParser.loadLstFile(listingFile);
			assertTrue(lstFileParser.getNumberOfLines() != 0);
			assertEquals(241, lstFileParser.getNumberOfLines());
			assertEquals(0x139, lstFileParser.getLastDefinedAddress());
			assertEquals(240,
					lstFileParser.getIndexOfLineWithLastDefinedAddress());
			AssemblyLine_I assemblyLine;
			for (int i = 0; i < 241; i++) {
				assemblyLine = lstFileParser.getAssemblyLine(i);
				assertNotNull(assemblyLine);
				if (programListing[i].equals("\n")) {
					assertEquals(programListing[i], lstFileParser
							.getAssemblyLine(i).getData() + "\n");
				} else {
					assertFalse(lstFileParser.getAssemblyLine(i).getData()
							.equals(""));
				}
			}
			assemblyLine = lstFileParser.getAssemblyLine(240);
			assertNotNull(assemblyLine);
			assertEquals(0x139, assemblyLine.getDataAddress());
			assertEquals("00 71", assemblyLine.getData().trim().substring(0, 5));
			listingFile.deleteOnExit();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ParsingException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
	}

	/**
	 * Test of the {@link LstFileParser#getTheInstance()} method
	 */
	@Test
	public void testGetTheInstance() {
		lstFileParser = LstFileParser.getTheInstance();
		assertNotNull(lstFileParser);
	}

	/**
	 * Test of the {@link LstFileParser#createTheInstance(AddressSpaceManager)}
	 * method
	 */
	@Test
	public void testCreateTheInstance() {
		LstFileParser.createTheInstance(addressSpaceManager);
		lstFileParser = LstFileParser.getTheInstance();
		assertNotNull(lstFileParser);
	}

	/**
	 * Test of the {@link LstFileParser#getProviderID()} method
	 */
	@Test
	public void testGetProviderID() {
		assertEquals("Code From Listing", lstFileParser.getProviderID());
	}

	/**
	 * Test of the {@link LstFileParser#getAssemblyVariablesManager()} method
	 */
	@Test
	public void testGetAssemblyVariablesManager() {
		assertNotNull(lstFileParser.getAssemblyVariablesManager());
	}

}
