/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.header_check;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.IOException;
import java.util.Spliterator;
import java.util.Spliterators;
import java.util.function.Consumer;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

/**
 * Created by wenzel on 12.01.17.
 */
public class CCommentRemover extends Spliterators.AbstractSpliterator<String> {
	private static final Logger logger = LoggerFactory.getLogger(CCommentRemover.class);

	private final Spliterator<String> input;
	private final BackSlashCommentEOL backSlashCommentEOL;
	private StringBuilder rawText = new StringBuilder();
	private ParserState parserState = ParserState.REGULAR;


	public CCommentRemover(Spliterator<String> input, BackSlashCommentEOL backSlashCommentEOL) {
		super(Long.MAX_VALUE, input.characteristics() & ~(Spliterator.SIZED|Spliterator.SUBSIZED));

		this.input = input;
		this.backSlashCommentEOL = backSlashCommentEOL;

		if ((input.characteristics()& Spliterator.ORDERED) ==0)
			throw new IllegalArgumentException("cannot remove comments from a unordered input");
	}

	@Override
	public long estimateSize() {
		return input.estimateSize();
	}

	public CCommentRemover(BufferedReader input, BackSlashCommentEOL backSlashCommentEOL) {
		this(new Spliterators.AbstractSpliterator<String>(Long.MAX_VALUE, Spliterator.ORDERED) {
			@Override
			public boolean tryAdvance(Consumer<? super String> action) {
				try {
					final String s = input.readLine();
					if (s != null) {
						action.accept(s);
						return true;
					}
					return false;
				} catch (IOException e) {
					throw new RuntimeException(e);
				}
			}
		},backSlashCommentEOL);
	}

	private void processLine(String line, StringBuilder res) {
		rawText.append(line);
		rawText.append('\n');


		if (parserState == ParserState.IN_SINGLELINE_COMMENT_SAW_BACKSLASH) {
			switch (backSlashCommentEOL) {
				case EndComment:
					parserState = ParserState.REGULAR;
					break;
				case ContinueComment:
					parserState = ParserState.IN_SINGLELINE_COMMENT;
					break;
				case Error:
					throw new CommentException("Backslashes at end of line in single line comments are forbidden.");
				case ErrorIfSignificant:
					parserState = ParserState.MAYBE_COMMENT;
			}
		} else if (parserState == ParserState.IN_SINGLELINE_COMMENT || parserState == ParserState.SAW_SLASH) {
			parserState = ParserState.REGULAR;
		} else if (parserState == ParserState.IN_MULTILINE_COMMENT_SAW_STAR) {
			parserState = ParserState.IN_MULTILINE_COMMENT;
		}

		for (int i = 0; i < line.length(); i++) {
			char c = line.charAt(i);

			switch (parserState) {
				case REGULAR:
					if (c == '/') {
						parserState = ParserState.SAW_SLASH;
					} else {
						res.append(c);
					}
					break;
				case SAW_SLASH:
					if (c == '/') {
						parserState = ParserState.IN_SINGLELINE_COMMENT;
					} else if (c == '*') {
						parserState = ParserState.IN_MULTILINE_COMMENT;
					} else {
						res.append('/');
						res.append(c);
						parserState = ParserState.REGULAR;
					}
					break;
				case IN_SINGLELINE_COMMENT:
				case IN_SINGLELINE_COMMENT_SAW_BACKSLASH:
					if (c == '\\') {
						parserState = ParserState.IN_SINGLELINE_COMMENT_SAW_BACKSLASH;
					} else {
						parserState = ParserState.IN_SINGLELINE_COMMENT;
					}
					break;
				case IN_MULTILINE_COMMENT:
					if (c == '*') {
						parserState = ParserState.IN_MULTILINE_COMMENT_SAW_STAR;
					}
					break;
				case IN_MULTILINE_COMMENT_SAW_STAR:
					if (c == '/') {
						parserState = ParserState.REGULAR;
					} else if (c != '*') {
						parserState = ParserState.IN_MULTILINE_COMMENT;
					}
					break;
				case MAYBE_COMMENT:
					if (c == '/') {
						parserState = ParserState.MAYBE_COMMENT_SAW_SLASH;
					} else {
						throw CommentException.backslashEOLSignificant();
					}
					break;
				case MAYBE_COMMENT_SAW_SLASH:
					if (c == '/') {
						parserState = ParserState.IN_SINGLELINE_COMMENT;
					} else if (c == '*') {
						parserState = ParserState.IN_MULTILINE_COMMENT;
					} else {
						throw CommentException.backslashEOLSignificant();
					}
					break;

			}
		}
	}

	public String getRawText() {
		String s = rawText.toString();
		rawText = new StringBuilder();
		return s;
	}

	@Override
	public boolean tryAdvance(Consumer<? super String> action) {
		StringBuilder res = new StringBuilder();
		while (res.length() == 0) {
			boolean couldAdvance = input.tryAdvance(l -> processLine(l, res));
			if (!couldAdvance) {
				return false;
			}
		}

		action.accept(res.toString());
		return true;
	}

	public Stream<String> stream() {
		return StreamSupport.stream(this, false);
	}

	/**
	 * What to do if a singleline comment ends in a backslash
	 */
	public enum BackSlashCommentEOL {
		/**
		 * Comment continues on next line
		 */
		ContinueComment,
		/**
		 * Comment does not continue on next line
		 */
		EndComment,
		/**
		 * Raise Exception
		 */
		Error,
		/**
		 * Raise exception if the next line does not immediately begin with a comment
		 */
		ErrorIfSignificant
	}


	private enum ParserState {
		REGULAR,
		SAW_SLASH,
		IN_MULTILINE_COMMENT,
		IN_SINGLELINE_COMMENT,
		IN_SINGLELINE_COMMENT_SAW_BACKSLASH,
		IN_MULTILINE_COMMENT_SAW_STAR,
		/**
		 * Backslash at end of previous line
		 */
		MAYBE_COMMENT,
		/**
		 * Backslash at end of previous line and we saw a slash
		 */
		MAYBE_COMMENT_SAW_SLASH
	}


	public static class CommentException extends RuntimeException {

		public CommentException(String s) {
			super(s);
		}

		public static CommentException backslashEOLSignificant() {
			return new CommentException("Backslash at end of line without a comment starting in the following line");
		}
	}
}
