/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.header_check;

import de.tu_darmstadt.rs.spartanmc.jconfig.TestingProjectProvider;

import org.junit.Test;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.FileVisitOption;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.Spliterators;
import java.util.function.Consumer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Created by wenzel on 12.01.17.
 */
public class HeaderChecker {

	private final String pragmaOnce = "\n#pragma once";

	private Pattern preprocessorDirective = Pattern.compile("^#([a-z]+)\\s+(\\w+)");
	private Pattern unaryDirective = Pattern.compile("^#([a-z]+)");
	private Pattern externC = Pattern.compile("^extern\\s+\"C\"\\s*\\{");

	protected boolean checkExternC(Iterator<String> supplier) {
		int state = 0;
		while (true) {
			if (!supplier.hasNext()) {
				return false;
			}
			String line = supplier.next();

			Matcher matcher = preprocessorDirective.matcher(line);
			if (matcher.matches()) {
				if (matcher.group(1).equals("include")) {
					continue;
				}
			}

			if (state == 0) {
				if (!matcher.matches() || !matcher.group(1).equals("ifdef") || !matcher.group(2).equals("__cplusplus")) {
					return false;
				}
			} else if (state == 1) {
				if (!externC.matcher(line).matches()) {
					return false;
				}
			} else if (state == 2) {
				Matcher unaryMatcher = unaryDirective.matcher(line);
				if (!unaryMatcher.matches() || !unaryMatcher.group(1).equals("endif")) {
					return false;
				}
			} else {
				return true;
			}

			state++;
		}
	}

	private Optional<String> getIncludeGuards(Iterator<String> it) {
		if (!it.hasNext()) {
			return Optional.empty();
		}
		String firstLine = it.next().trim();

		Matcher firstLineMatcher = preprocessorDirective.matcher(firstLine);

		if (!firstLineMatcher.matches()) {
			return Optional.empty();
		}

		String firstLineArg = firstLineMatcher.group(2);
		String firstLineDirective = firstLineMatcher.group(1);

		if (firstLineDirective.equals("pragma") && firstLineArg.equals("once")) {
			return Optional.of(pragmaOnce);
		}

		if (firstLineDirective.equals("ifndef")) {
			if (!it.hasNext()) {
				return Optional.empty();
			}
			String secondLine = it.next().trim();
			Matcher secondLineMatcher = preprocessorDirective.matcher(secondLine);
			if (!secondLineMatcher.matches()) {
				return Optional.empty();
			}
			String secondLineArg = secondLineMatcher.group(2);
			String secondLineDirective = secondLineMatcher.group(1);
			if (!secondLineDirective.equals("define") || !secondLineArg.equals(firstLineArg)) {
				return Optional.empty();
			}

			return Optional.of(firstLineArg);
		} else {
			return Optional.empty();
		}
	}

	protected Optional<String> processFile(Path file, Consumer<String> errConsumer) {
		try (BufferedReader bufferedReader = Files.newBufferedReader(file)) {
			CCommentRemover contents = new CCommentRemover(bufferedReader, CCommentRemover.BackSlashCommentEOL.ContinueComment);

			Iterator<String> it = Spliterators.iterator(contents);

			Optional<String> includeGuards = getIncludeGuards(it);

			if (!checkExternC(it)) {
				errConsumer.accept(file + " is not encapsulated by an extern \"C\" block");
			}

			return includeGuards;
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	@Test
	public void checkHeader() throws IOException {
		List<String> externCErrors = new ArrayList<>();
		Path envSpartanRoot = TestingProjectProvider.getSpartanRoot();
		Path includeDir = envSpartanRoot.resolve("spartanmc/include");
		Map<Optional<String>, Set<Path>> collect = Files.walk(includeDir, FileVisitOption.FOLLOW_LINKS)
				.filter(f -> Files.isRegularFile(f))
				.collect(Collectors.groupingBy(f -> processFile(f, externCErrors::add), Collectors.toSet()));

		Stream<String> guardErrors = collect.keySet().stream()
				.flatMap(guard -> {
					if (!guard.isPresent()) {
						return collect.get(guard).stream().map(f -> f.toString() + " has no include guard");
					} else {
						String g = guard.get();
						Set<Path> files = collect.get(guard);
						if (!g.equals(pragmaOnce) && files.size() > 1) {
							return Stream.of("The include guard \"" + g + "\" is used multiple times: " + files);
						} else {
							return Stream.empty();
						}
					}
				});
		String res = Stream.concat(guardErrors, externCErrors.stream()).collect(Collectors.joining("\n"));
		if (!res.isEmpty()) {
			throw new AssertionError(res);
		}
	}

}
