/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.integration_test;

import de.tu_darmstadt.rs.process.ProcessUtil;
import de.tu_darmstadt.rs.spartanmc.jconfig.TestingProjectProvider;
import de.tu_darmstadt.rs.util.SpartanMCDefaultFolders;

import com.googlecode.junittoolbox.ParallelParameterized;
import org.apache.commons.io.output.StringBuilderWriter;
import org.jetbrains.annotations.Nullable;
import org.junit.Assert;
import org.junit.Assume;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

/**
 * Run integration tests in $SPARTANMC_ROOT/tests by issuing make test in each test subfolder
 */
@RunWith(ParallelParameterized.class)
public class IntegrationTest {

	private static final Object monitor = new Object();

	@Nullable
	private static Path getLogOutputDir() throws IOException {

		final String reportDir = System.getProperty("STDOUT_LOG_LOCATION");
		if (reportDir == null) {
			return null;
		}

		final Path dir = Paths.get(reportDir).resolve(IntegrationTest.class.getCanonicalName());
		Files.createDirectories(dir);
		return dir;

	}

	@Parameterized.Parameters(name = "{2}")
	public static Iterable<?> data() throws IOException {
		Path tests = SpartanMCDefaultFolders.getIntegrationtestsGeneral(TestingProjectProvider.getSpartanRoot());
		return StreamSupport.stream(Files.newDirectoryStream(tests).spliterator(), false)
				//Only subdirs
				.filter(x -> Files.isDirectory(x))
				//Not the hardware directory
				.filter(x -> !x.endsWith("hardware"))
				//When a testcase gets renamed, untracked temporary files may remain. We need to filter those out.
				//Therefore, we check if a Makefile exists
				.filter(x -> Files.exists(x.resolve("Makefile")))
				//Sort
				.sorted(Comparator.comparing(Object::toString))
				//Append testcase name
				.map(IntegrationTest::folderToParameters)
				//To set
				.collect(Collectors.toList());
	}

	@Parameterized.Parameter
	public Path testcase;

	@Parameterized.Parameter(value = 1)
	public String name;

	@Parameterized.Parameter(value = 2)
	public String nameWithTrackingInfo;

	private static boolean isTracked(Path x) {
		ProcessBuilder pb = new ProcessBuilder("git", "ls-files");
		pb.directory(x.toFile());

		List<String> output = new ArrayList<>();
		try {

			final int exitCode = ProcessUtil.runAndWaitFor(pb, (Consumer<String>) output::add);

			//If git errors out, we are probably outside a repository
			if (exitCode!=0) {
				return true;
			}

			return !output.isEmpty();

		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
	}

	private static Object[] folderToParameters(Path x) {
		String name = x.getFileName().toString();
		String nameWithTracking;
		if (!isTracked(x)) {
			nameWithTracking = name + " (untracked)";
		} else {
			nameWithTracking = name;
		}
		return new Object[]{x, name, nameWithTracking};
	}

	@Test
	public void test() throws InterruptedException, IOException, TimeoutException {
		// Skip test if there is an ignore file
		Assume.assumeFalse("Test is ignored by ignore-file", Files.exists(testcase.resolve("ignore")));

		final Path logOutputDir = getLogOutputDir();
//		Consumer<String> logWriter;
//		PrintWriter printWriter = null;
		final Path logFile = (logOutputDir != null)? logOutputDir.resolve(name + ".txt") : null;
//			final BufferedWriter bufferedWriter = Files.newBufferedWriter(logFile);
//			printWriter = new PrintWriter(bufferedWriter);
//			logWriter = printWriter::println;
//		} else {
//			logWriter = (s) -> {
//			};
//		}

		StringBuilderWriter stringWriter = new StringBuilderWriter(); // faster bc. unsynchronized

		ModelsimOutputParser outputParser = new ModelsimOutputParser();

		String errorMessage = null;
		try {

			ProcessBuilder pb = new ProcessBuilder("make", "test");
			pb.directory(testcase.toFile());
			pb.environment().put("SPARTANMC_ROOT", TestingProjectProvider.getSpartanRoot().toString());

			final String jacocoArg = System.getProperty("jacoco");
			if (jacocoArg != null) {
				pb.environment().put("JAVA_OPTS", jacocoArg);
			}

			int exitCode;
			try (PrintWriter printWriter = new PrintWriter(stringWriter)) {
				Consumer<String> logWriter = printWriter::println;
				exitCode = ModelsimParallelismLimiter.getInstance().withModelsim(() -> ProcessUtil.runAndWaitFor(pb, null, 60L, TimeUnit.MINUTES, logWriter, outputParser::parseLine));
			}

			errorMessage = outputParser.getErrors().stream().collect(Collectors.joining("\n"));

			if (exitCode != 0) {
				Assert.assertTrue("Unknown error, see transcript", outputParser.hasErrors());
				throw new AssertionError('\n' + errorMessage);
			}

		} finally {
			String stdout = stringWriter.toString();

			synchronized (monitor) {
				System.out.println("Output of Testcase: "+ testcase);
				System.out.println("################ Start #####################");
				System.out.print(stdout);
				System.out.println("################ End   #####################");
				if (errorMessage != null) {
					System.out.println("############### Error ##################");
					System.out.println(errorMessage);
					System.out.println("\n\n\n");
				}
			}

			if (logOutputDir != null) {
				try (BufferedWriter logOut = Files.newBufferedWriter(logFile)) {
					logOut.write(stdout);
				} catch (IOException e) {
					e.printStackTrace();
				}
			}

		}



	}

}
