/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.integration_test;

import de.tu_darmstadt.rs.process.ProcessUtil;
import de.tu_darmstadt.rs.process.RingBuffer;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.jconfig.TestingProjectProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.HdlFilesOutput;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputProvider.common.TclDeviceInfoBuilderProvider;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.apache.commons.io.output.CloseShieldOutputStream;
import org.apache.commons.io.output.TeeOutputStream;
import org.junit.Assert;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class ModelsimSimulation {

	private static final Logger logger = LoggerFactory.getLogger(ModelsimSimulation.class);

	private final Set<StreamUtil.Pair<Path, String>> sourceFiles;
	private final Set<String> libraries;
	private final IConfigurationDocument.ITargetConfiguration target;
	private final Consumer<PrintWriter> additionalCommands;
	private final Path runDirectory;
	private final Library library;
	private final boolean quiet;
	private final Collection<String> toplevel;
	private final boolean fatalWarnings;
	private final String simulationRuntime;
	private final boolean wrapInCatch;
	private final Path logFile;
	private final boolean loadWaveDoIfExists;
	public static final Function<Path, String> defaultSourceFileMapper = sourceFileMapper("\tvlog -pedanticerrors \"%s\"", "\tvcom \"%s\"");

	private ModelsimSimulation(
			Set<StreamUtil.Pair<Path, String>> sourceFiles,
			Path runDirectory,
			IConfigurationDocument.ITargetConfiguration target,
			Library library,
			boolean quiet,
			Collection<String> toplevel,
			boolean fatalWarnings,
			String simulationRuntime,
			Consumer<PrintWriter> additionalCommands,
			boolean wrapInCatch,
			Path logFile,
			boolean loadWaveDoIfExists) {
		this.sourceFiles = Objects.requireNonNull(sourceFiles);
		this.target = target;
		this.additionalCommands = additionalCommands;
		this.runDirectory = Objects.requireNonNull(runDirectory);
		this.library = Objects.requireNonNull(library);
		this.quiet = quiet;
		this.toplevel = toplevel;
		this.fatalWarnings = fatalWarnings;
		this.simulationRuntime = simulationRuntime;
		this.wrapInCatch = wrapInCatch;
		this.logFile = logFile;
		this.loadWaveDoIfExists = loadWaveDoIfExists;
		this.libraries = sourceFiles.stream().map(StreamUtil.Pair::getB).collect(Collectors.toSet());
	}


	public static String getModulesRoot(Library library) {
		return library.getModuleSourceTrees().stream()
				.map(Object::toString)
				.filter(r -> !r.contains("addModule"))
				.findAny()
				.orElseThrow(() -> new RuntimeException("No matching root found"));
	}

	private String getModulesRoot() {
		return getModulesRoot(library);
	}


	public void writeModelsimInitializationCommands(PrintWriter pw, DeviceDescription.Toolchain toolchain) {
		if (toolchain != null) {
			pw.println("do $::env(SPARTANMC_ROOT)/src/scripts/tcl/functions.tcl");
			pw.println("variable TOOLCHAIN " + toolchain.getToolchain());
			pw.println("variable DEVICE_SIMLIBS {" + toolchain.getSimlibsFunctional().collect(Collectors.joining(" ")) + "}");

			pw.println("variable CMDLINE_OPTIONS {" + TclDeviceInfoBuilderProvider.getVsimCmdlineOptions(target, library) + "}");
			pw.println("set libroot [ getLibroot ]");
			pw.println("echo \"libraries are at $libroot\"");
			pw.println("updateLibraries $libroot");
		} else {
			logger.info("Not including simlibs, since no toolchain was given");
		}
	}


	public void createModelsimConfig() throws IOException {
		if (!Files.exists(runDirectory.resolve("modelsim.ini"))) {
			Path makefile = runDirectory.resolve("Makefile");
			Files.write(makefile, Arrays.asList(
					quiet? "" : "NOQUIET:=yes",
					"SIM_HOME:=$(abspath .)",
					"JCONFIG_MODULES_ROOT:=" + getModulesRoot(),
					"include $(SPARTANMC_ROOT)/src/scripts/make/project/main.mk",
					"init: $(MODELSIM_INI) $(VLOG_OPT) | $(VSIM_DIR)\n"
			));
			ProcessBuilder pb = new ProcessBuilder("make", "init");
			pb.directory(runDirectory.toFile());
			pb.environment().put("SPARTANMC_ROOT", TestingProjectProvider.getSpartanRoot().toString());
			RingBuffer<String> rb = new RingBuffer<>(100);
			final int exitCode = ProcessUtil.runAndWaitFor(pb, null, 10L, TimeUnit.SECONDS, rb);
			if (exitCode != 0) {
				rb.iterator().forEachRemaining(logger::warn);
				throw new AssertionError("Failed creating modelsim configuration, error code: " + exitCode);
			}
		}
	}


	public static Function<Path, String> sourceFileMapper(String verilogFormat, String vhdlFormat) {
		return source -> {
			if (!Files.isRegularFile(source)) {
				Assert.fail("Module Description references missing file: " + source);
			}
			String fn = source.toString();
			if (fn.endsWith(".v")) {
				return String.format(verilogFormat, fn);
			} else if (fn.endsWith(".vhd")) {
				return String.format(vhdlFormat, fn);
			} else {
				Assert.fail("Module Description references file of unknown type: " + fn);
			}
			//Cannot be reached
			return null;
		};

	}

	public static Function<SourceRef, String> sourceRefMapper(String verilogFormat, String vhdlFormat) {
		Function<Path, String> sfm = sourceFileMapper(verilogFormat, vhdlFormat);
		return source -> sfm.apply(source.getPath());
	}


	public void run(Long timeout, TimeUnit timeUnit) throws IOException {

		ModelsimOutputParser outputParser = new ModelsimOutputParser();
		RingBuffer<String> ringBuffer = new RingBuffer<>(100);


		int exitCode = runWithOutputConsumers(timeout, timeUnit, outputParser::parseLine, ringBuffer);

		if (exitCode != 0 && !outputParser.hasErrors()) {
			throw new AssertionError("Unknown error, see transcript:\n" + ringBuffer.stream().collect(Collectors.joining("\n")));
		}

		if (outputParser.hasErrors()) {
			throw new AssertionError(outputParser.getErrors().stream().collect(Collectors.joining("\n")));
		}

		if (outputParser.hasWarnings()) {
			if (fatalWarnings) {
				throw new AssertionError(outputParser.getWarnings().stream().collect(Collectors.joining("\n")));
			} else {
				outputParser.getWarnings().forEach(logger::warn);
			}
		}
	}

	@SafeVarargs
	public final int runWithOutputConsumers(Long timeout, TimeUnit timeUnit, Consumer<String>... consumers) throws IOException {
		createModelsimConfig();

		ProcessBuilder builder = new ProcessBuilder("vsim");
		builder.directory(runDirectory.toFile());


		//The output stream is written in a second thread. If that crashes, the original thread may
		//not notice. We need to save the second thread's exception and rethrow it on the main thread.
		final Throwable[] outputGenException = {null};

		int exitCode = ModelsimParallelismLimiter.getInstance().withModelsim(() -> ProcessUtil.runAndWaitFor(builder, (OutputStream processOs) -> {
			new Thread(() -> {
				//				OutputStream os = processOs;
				OutputStream os = new TeeOutputStream(processOs, new CloseShieldOutputStream(System.out));
				OutputStream fileOs = null;
				if (logFile != null) {
					try {
						fileOs = Files.newOutputStream(logFile);
					} catch (IOException e) {
						throw new RuntimeException(e);
					}
					os = new TeeOutputStream(os, fileOs);
				}
				try (PrintWriter pw = new PrintWriter(os)) {
					if (wrapInCatch) {
						pw.println("set err \"\"");
						pw.println("catch { ");
					}


					for (String library : libraries) {
						pw.println("vlib vsim/" + library);
						pw.println("vmap " + library + " ./vsim/" + library);
					}

					if (additionalCommands != null) {
						additionalCommands.accept(pw);
					}
					writeModelsimInitializationCommands(pw, target != null? target.getToolchain() : null);

					final HdlFilesOutput.DefaultHdlOutput vsimOutput = HdlFilesOutput.getVsimOutput(null, SourceRef.FileType.FUNCTIONAL_SIM, target, library);
					vsimOutput.writeHead(pw);
					sourceFiles.forEach(f -> vsimOutput.writeWork(pw, new SourceRef(f.getA(), f.getB(), SourceRef.FileType.FUNCTIONAL_SIM)));
					vsimOutput.writeFoot(pw);

					pw.println("\tvlog " + System.getenv("XILINX") + "/verilog/src/glbl.v");

					StringBuilder libraries = new StringBuilder();
					if (target != null) {
						libraries = new StringBuilder(target.getToolchain().getSimlibsFunctional().map(s -> "-L " + s).collect(Collectors.joining(" ")));
					}
					for (String library : this.libraries) {
						if (!library.equals(Builder.defaultLibrary)) {
							libraries.append(" -L ").append(library);
						}
					}

					if (toplevel != null && simulationRuntime != null) {
						String toplevelStr = toplevel.stream().collect(Collectors.joining(" "));
						pw.println("\tvsim -voptargs=+acc -pedanticerrors " + libraries + " " + toplevelStr);

						if (loadWaveDoIfExists) {
							pw.println("# load saved wave display");
							pw.println("if {[file exists \"wave.do\"]} {");
							pw.println("\tdo {wave.do}");
							pw.println("}");
						}

						pw.println("\trun " + simulationRuntime);
						logger.info("Starting simulation of {} for {}", toplevelStr, simulationRuntime);
					} else {
						logger.info("Not starting simulation, since no toplevel and/or no runtime was given");
					}
					if (wrapInCatch) {
						pw.println("} err ");

						pw.println("if [expr  {${err}!=\"\"}] then { echo $err;  exit -code 10 }");
					}
					pw.flush();
				} catch (Throwable e) {
					outputGenException[0] = e;
					throw e;
				}

				if (logFile != null) {
					try {
						fileOs.close();
					} catch (IOException e) {
						throw new RuntimeException(e);
					}
				}
			}).start();
		}, timeout, timeUnit, consumers));

		if (outputGenException[0] != null) {
			throw new AssertionError(outputGenException[0]);
		}
		return exitCode;
	}

	public static Builder builder() {
		return new Builder();
	}


	public static class Builder {

		private static final String defaultLibrary = "work";

		private Set<StreamUtil.Pair<Path, String>> sourceFiles = new HashSet<>();
		private Path runDirectory = null;
		private Library library;
		private boolean quiet = false;
		private Collection<String> toplevel;
		private boolean fatalWarnings;
		private String simulationRuntime;
		private Consumer<PrintWriter> additionalCommands;
		private boolean wrapInCatch = true;
		private IConfigurationDocument.ITargetConfiguration target = null;
		private Path logFile;
		private boolean loadWaveDoIfExists;

		private Builder() {
		}

		public Builder addSourceFile(Path file, String library) {
			sourceFiles.add(new StreamUtil.Pair<>(file, library));
			return this;
		}

		public Builder addSourceFile(Path file) {
			addSourceFile(file, defaultLibrary);
			return this;
		}

		public Builder addSourceFiles(Collection<Path> files, String library) {
			for (Path file : files) {
				sourceFiles.add(new StreamUtil.Pair<>(file, library));
			}
			return this;
		}

		public Builder addSourceFiles(Collection<Path> files) {
			addSourceFiles(files, defaultLibrary);
			return this;
		}


		public Builder addSourceFiles(Stream<Path> files, String library) {
			files.forEach(f -> addSourceFile(f, library));
			return this;
		}

		public Builder addSourceFiles(Stream<Path> files) {
			addSourceFiles(files, defaultLibrary);
			return this;
		}

		public Builder setRunDirectory(Path directory) {
			this.runDirectory = directory;
			return this;
		}

		public Builder setTarget(IConfigurationDocument.ITargetConfiguration target) {
			this.target = target;
			return this;
		}

		public Builder setModuleRootDir(Library library) {
			this.library = library;
			return this;
		}

		public Builder setQuiet(boolean quiet) {
			this.quiet = quiet;
			return this;
		}

		public Builder setToplevel(Collection<String> toplevel) {
			this.toplevel = toplevel;
			return this;
		}


		public Builder setToplevel(String toplevel) {
			this.toplevel = Collections.singleton(toplevel);
			return this;
		}

		public Builder setFatalWarnings(boolean fatalWarnings) {
			this.fatalWarnings = fatalWarnings;
			return this;
		}

		public Builder setSimulationRuntime(String runtime) {
			this.simulationRuntime = runtime;
			return this;
		}

		public Builder setAdditionalOutputStreamCommands(Consumer<PrintWriter> consumer) {
			this.additionalCommands = consumer;
			return this;
		}

		public Builder setWrapInCatch(boolean wrapInCatch) {
			this.wrapInCatch = wrapInCatch;
			return this;
		}

		public Builder logCommandsToFile(Path logFile) {
			this.logFile = logFile;
			return this;
		}

		public Builder loadWaveDoIfExists(boolean b) {
			this.loadWaveDoIfExists = true;
			return this;
		}

		public ModelsimSimulation create() {
			return new ModelsimSimulation(sourceFiles, runDirectory, target, library, quiet, toplevel, fatalWarnings, simulationRuntime, additionalCommands, wrapInCatch, logFile, loadWaveDoIfExists);
		}
	}


}
