/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.integration_test;

import de.tu_darmstadt.rs.process.ProcessUtil;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.TestingProjectProvider;
import de.tu_darmstadt.rs.util.StreamUtil;

import cetus.hir.ClassDeclaration;
import cetus.hir.Declaration;
import cetus.hir.DepthFirstIterator;
import cetus.hir.Program;
import cetus.hir.Symbol;
import cetus.hir.VariableDeclaration;
import org.junit.Assert;
import org.junit.Assume;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.TimeUnit;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

/**
 * Verify that module's headers are correct
 *
 * Not parallel, since Cetus is broken :(
 */
@RunWith(Parameterized.class)
public final class ModuleHeaders {


	private static final Logger logger = LoggerFactory.getLogger(ModuleHeaders.class);

	@Parameterized.Parameters(name = "{1}")
	public static Iterable<?> data() throws IOException {

		Library library = TestingProjectProvider.getInstance().getLibrary();

		return library.getAvailableModules().stream()
				.filter(module -> module instanceof PeripheralModuleDescription)
				.sorted(Comparator.comparing(ModuleDescription::getLabel))
				.map(module -> new Object[]{
						module,
						module.getLabel() + " (" + module.getName() + ")",
						library
				}).collect(Collectors.toList());
	}

	@Parameterized.Parameter
	public PeripheralModuleDescription module;

	@Parameterized.Parameter(value = 1)
	public String name;

	@Parameterized.Parameter(value = 2)
	public Library library;

	@Rule
	public TemporaryFolder temporaryFolder = new TemporaryFolder();

	/**
	 * Check that the supplied headers really define the peripheral registers
	 *
	 * @throws IOException On broken IO
	 */
	@Test
	public void testHeaders() throws IOException {

		StringBuilder testCode = new StringBuilder();

		final Stream.Builder<String> builder = Stream.builder();

		if (module.getRegisters() != null && module.getRegisters().getCSupport() != null) {
			testCode.append(String.format("#include <%s>\n", module.getRegisters().getCSupport().getFileName()));
			testCode.append(String.format("extern %s %s;\n", module.getRegisters().getCSupport().getTypeName(), "registers"));
			builder.accept(module.getRegisters().getCSupport().getTypeName());
		}
		if (module.getMemory() != null && module.getMemory().getCSupport() != null) {
			testCode.append(String.format("#include <%s>\n", module.getMemory().getCSupport().getFileName()));
			testCode.append(String.format("extern %s %s;\n", module.getMemory().getCSupport().getTypeName(), "memory"));
			builder.accept(module.getMemory().getCSupport().getTypeName());
		}

		String testString = testCode.toString();

		Assume.assumeFalse("No CSupport defined", testString.isEmpty());

		Path tempDirectory = temporaryFolder.newFolder().toPath();


		final Path testFile = tempDirectory.resolve(module.getName() + ".c");
		Files.write(testFile, Collections.singletonList(testString));

		final Path searchPath = TestingProjectProvider.getSpartanRoot().resolve("spartanmc/include");
		ProcessBuilder pb = new ProcessBuilder("spartanmc-gcc", "-c", "-I", searchPath.toString(), testFile.toString());
		pb.directory(tempDirectory.toFile());

		List<String> output = new ArrayList<>();

		int exitCode = ProcessUtil.runAndWaitFor(pb, null, 10L, TimeUnit.SECONDS, (Consumer<String>) output::add);

		if (exitCode != 0) {
			throw new AssertionError(output.stream().collect(Collectors.joining("\n")));
		}

		checkFieldsVolatile(testFile, Collections.singletonList(searchPath.toFile()), builder.build());

	}

	private Optional<VariableDeclaration> getTypedef(Program program, String toFind) {
		return StreamUtil.streamIterator(new DepthFirstIterator<>(program))
				.flatMap(StreamUtil.castOrEmpty(VariableDeclaration.class))
				.filter(vd -> vd.getDeclaredSymbols().stream().anyMatch(s -> s.getSymbolName().equals(toFind)))
				.filter(vd -> vd.getSpecifiers().get(0).toString().equals("typedef"))
				.collect(StreamUtil.atMostOne());
	}


	private ClassDeclaration getStructDefinition(Program program, String structName) {

		return StreamUtil.streamIterator(new DepthFirstIterator<>(program))
				.flatMap(StreamUtil.castOrEmpty(ClassDeclaration.class))
				.filter(cd -> cd.getDeclaredIDs().stream().anyMatch(x -> x.toString().equals(structName)))
				.collect(StreamUtil.atMostOne())
				.orElseThrow(() -> new RuntimeException("did not find " + structName));
	}

	private void verifyTypeSpecifier(Symbol symbol, VariableDeclaration variable, Program program) {
		final List typeSpecifiers = symbol.getTypeSpecifiers();
		Object last = typeSpecifiers.get(typeSpecifiers.size() - 1);

		//Non volatile pointer is error
		if (last.toString().equals("*")) {
			Assert.fail("Non volatile pointer: " + variable);
		}

		//Volatile pointer is ok
		if (last.toString().equals("* volatile")) {
			return;
		}

		if (typeSpecifiers.size() == 1 && !isPrimitiveType(typeSpecifiers.get(0).toString())) {
			final String type = typeSpecifiers.get(0).toString();
			logger.info("checking referenced struct: " + type);
			verifyTypeAllVolatile(program, type);
			return;
		}


		//Non volatile regular field is an error
		if (typeSpecifiers.stream().noneMatch(e -> e.toString().equals("volatile"))) {
			Assert.fail("Non volatile regular field: " + variable);
		}


	}

	private boolean isPrimitiveType(String s) {
		return s.equals("char") || s.equals("int") || s.equals("long") || s.equals("byte");
	}

	private void verifyFieldVolatile(Declaration decl, Program program) {
		if (!(decl instanceof VariableDeclaration)) {
			Assert.fail("Not a variable Declaration: " + decl);
		}
		VariableDeclaration variable = (VariableDeclaration) decl;
		final boolean noFunctionPointers = IntStream.range(0, variable.getNumDeclarators())
				.mapToObj(variable::getDeclarator)
				.noneMatch(declarator -> declarator.toString().startsWith("("));
		if (!noFunctionPointers) {
			Assert.fail("Function pointers not supported in peri regs: " + decl);
		}

		logger.info("checking " + variable);

		variable.getDeclaredSymbols().forEach(sym -> {
			verifyTypeSpecifier(sym, variable, program);
		});

	}

	private Program parse(Path inFile, List<File> searchPaths) {
		try {
			return new ModuleHeaderDriver(Collections.singletonList(inFile.toFile()), searchPaths).parse();
		} catch (RuntimeException e) {
			Assume.assumeNoException(e);
			//Not reachable, previous line always throws
			return null;
		}
	}

	private void checkFieldsVolatile(Path inFile, List<File> searchPaths, Stream<String> typesToTest) {
		Program program = parse(inFile, searchPaths);
		typesToTest.forEach(type -> {
			verifyTypeAllVolatile(program, type);
		});
	}

	private void verifyTypeAllVolatile(Program program, String type) {
		Optional<VariableDeclaration> typedef = getTypedef(program, type);
		String realName;
		if (typedef.isPresent()) {
			if (typedef.get().getSpecifiers().get(1).toString().equals("volatile")) {
				logger.info("typedef for " + type + " is volatile, no need to check individual fields");
				return;
			}
			realName = typedef.get().getSpecifiers().get(typedef.get().getSpecifiers().size() - 1).toString();
		} else {
			realName = type;
		}
		if (!realName.startsWith("struct ") && !realName.startsWith("union ")) {
			logger.info(type + " is typedef'd to " + realName + ", so it's ok");
			return;
		}
		final ClassDeclaration structDefinition = getStructDefinition(program, realName);
		structDefinition.getDeclarations().forEach(decl -> verifyFieldVolatile(decl, program));
	}
}
