/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.integration_test;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.TestingProjectProvider;
import de.tu_darmstadt.rs.util.StreamUtil;

import org.junit.Assert;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.file.FileVisitOption;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Check that no two SpartanMC modules contain verilog modules with the same name.
 * If they did and a user builds a project with both SpartanMC modules, errors will follow
 */
public class ModuleNameUniqueness {

	private static final Logger logger = LoggerFactory.getLogger(ModuleNameUniqueness.class);

	@Test
	public void verifyModuleNameUniqueness() throws IOException {
		Library library = TestingProjectProvider.getInstance().getLibrary();

		Map<Path, Set<ModuleDescription>> filesWithModules = getHdlFiles(library);
		Map<String, Set<Path>> moduleDefiningFiles = getModulesByHdlFiles(library, filesWithModules.keySet());

		String res = moduleDefiningFiles.keySet().stream()
				.filter(m -> moduleDefiningFiles.get(m).size() > 1)
				.sorted()
				.flatMap((String m) -> {
					//Map from file -> module description to module description -> file
					final Map<ModuleDescription, Set<Path>> descrsContainingFile = moduleDefiningFiles.get(m).stream()
							.flatMap(p -> filesWithModules.get(p).stream().map(descr -> new StreamUtil.Pair<>(descr, p)))
							.collect(Collectors.groupingBy(
									StreamUtil.Pair::getA,
									Collectors.mapping(StreamUtil.Pair::getB, Collectors.toSet())
							));

					//If one module description contains all versions of a module, we assume it knows what it is doing
					//(e.g. it provides alternatives for different FPGAs)
					if (descrsContainingFile.size() == 1) {
						return Stream.empty();
					}


					final String containing = descrsContainingFile.entrySet().stream().sorted(Comparator.comparing(x -> x.getKey().getName())).map(entry ->
							"  * In Module " + entry.getKey().getName() + "\n"
									+ entry.getValue().stream().sorted(Comparator.comparing(Object::toString))
									.map(p -> "    * " + p.toString())
									.collect(Collectors.joining("\n"))
					).collect(Collectors.joining("\n"));

					return Stream.of("There are multiple HDL modules named " + m + ":\n" + containing);
				}).collect(Collectors.joining("\n\n"));

		if (!res.isEmpty()) {
			Assert.fail("\n"+res);
		}
	}

	/**
	 * For every hdl file, find out which module names it contains by compiling it with modelsim
	 */
	private Map<String, Set<Path>> getModulesByHdlFiles(Library library, Collection<Path> hdlFiles) throws IOException {
		Map<String, Set<Path>> moduleDefiningFiles = new HashMap<>();
		Path tempDirectory = Files.createTempDirectory("spmc-name-uniqueness-");
		try {

			CompilingLister compilingLister = new CompilingLister();

			final ModelsimSimulation modelsimSimulation = ModelsimSimulation.builder()
					.setRunDirectory(tempDirectory)
					.setQuiet(false)
					.setModuleRootDir(library)
					.setWrapInCatch(false)
					.setAdditionalOutputStreamCommands(pw -> {

						hdlFiles.forEach(file -> {
							final String line = ModelsimSimulation.defaultSourceFileMapper.apply(file);
							pw.println(line);
							pw.flush();
							//TODO detect errors, break
							synchronized (compilingLister) {
								while (!compilingLister.isFinishedParsing()) {
									try {
										compilingLister.wait();
									} catch (InterruptedException e) {
										e.printStackTrace();
									}
								}
							}
							compilingLister.getModuleNames().forEach(m -> {
								moduleDefiningFiles.computeIfAbsent(m, (x) -> new HashSet<>()).add(file);
							});

							compilingLister.reset();
						});

					})
					.create();

			int exitCode = modelsimSimulation.runWithOutputConsumers(100L, TimeUnit.SECONDS, compilingLister::parseLine);

			if (exitCode != 0) {
				throw new RuntimeException("exit code is " + exitCode);
			}
		} finally {
			Files.walk(tempDirectory, FileVisitOption.FOLLOW_LINKS)
					.sorted(Comparator.reverseOrder())
					.map(Path::toFile)
					.forEach(File::delete);
		}
		return moduleDefiningFiles;
	}

	/**
	 * Get all HDL files mapped to the module they are used in
	 */
	private Map<Path, Set<ModuleDescription>> getHdlFiles(Library library) {
		Map<Path, Set<ModuleDescription>> filesWithModules = new HashMap<>();
		library.getAvailableModules().forEach(module -> {
			module.getHdl().getSources().streamAllResolvedPaths(module, IToolchainConfiguration.ANY).forEach(source -> {
				filesWithModules.computeIfAbsent(source.getPath(), (x) -> new HashSet<>()).add(module);
			});
		});
		return filesWithModules;
	}

	private static class CompilingLister {

		private Set<String> moduleNames = new HashSet<>();
		private boolean finishedParsing = false;

		private static final Pattern moduleNamePattern = Pattern.compile("#?\\s*-- Compiling (module|entity) (.*)");

		public void parseLine(String line) {
			Matcher m = moduleNamePattern.matcher(line);
			if (m.matches()) {
				addModule(m.group(2));
			}
			if (line.contains("End time: ")) {
				finishParsing();
			}
		}

		private synchronized void finishParsing() {
			finishedParsing = true;
			notify();
		}

		public boolean isFinishedParsing() {
			return finishedParsing;
		}

		private synchronized void addModule(String module) {
			moduleNames.add(module);
		}

		public synchronized void reset() {
			moduleNames.clear();
			finishedParsing = false;
		}

		public Set<String> getModuleNames() {
			return Collections.unmodifiableSet(moduleNames);
		}
	}
}
