/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.integration_test;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HdlDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.Verification;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.LibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.TestingProjectProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.internal.SimRootModuleCollector;
import de.tu_darmstadt.rs.spartanmc.jconfig.internal.SourceFileCollector;

import org.junit.Assume;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;
import java.util.stream.Stream;

@RunWith(Parameterized.class)
public class ModuleTestbenchRunner {

	private static final Logger logger = LoggerFactory.getLogger(ModuleTestbenchRunner.class);

	@Parameterized.Parameters(name = "{2}")
	public static Iterable<?> data() throws IOException {

		Library library = TestingProjectProvider.getInstance().getLibrary();

		return library.getAvailableModules().stream()
				.sorted(Comparator.comparing(ModuleDescription::getLabel))
				.flatMap(module ->
						module.getHdl().getTestbenches().stream().map(testbench -> new Object[]{
								module,
								testbench,
								testbench.getName() + " for " + module.getLabel() + " (" + module.getName() + ")",
								library
						})
				).collect(Collectors.toList());
	}

	@Parameterized.Parameter
	public ModuleDescription module;

	@Parameterized.Parameter(value = 1)
	public HdlDescription.ModuleTestbench testbench;

	@Parameterized.Parameter(value = 2)
	public String name;

	@Parameterized.Parameter(value = 3)
	public Library library;

	private Path getRunDirectory(ILibJConfigProject project) throws IOException {
		final String toolchainName = ((LibJConfigProject) project).getDocumentHandler().getDocument().getTarget().getToolchain().getName();
		final Path buildDir = project.getBuildDir().resolve(testbench.getName()).resolve(toolchainName);
		if (Files.exists(buildDir)) {
			//Clear contents if it exists
			Files.walkFileTree(buildDir, new SimpleFileVisitor<Path>() {
				@Override
				public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
					Files.delete(file);
					return FileVisitResult.CONTINUE;
				}

				@Override
				public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
					if (!buildDir.equals(dir)) {
						Files.delete(dir);
					}
					return FileVisitResult.CONTINUE;
				}
			});
		} else {
			Files.createDirectories(buildDir);
		}
		return buildDir;
	}

	public void runTestbench(ILibJConfigProject project) throws IOException {
		final IConfigurationDocument.ITargetConfiguration target = project.getDocumentHandler().getDocument().getTarget();

		Assume.assumeTrue("Module not compatible with toolchain", module.supports(target.getCurrentConfiguration()));

		Path runDirectory = getRunDirectory(project);


		Collection<SourceRef> sources = Stream.of(module)
				.collect(SourceFileCollector.createDescription(
						project,
						library,
						Collections.emptyList()
				));


		final SimRootModuleCollector.SimRoots simRoots = Stream.of(module).collect(SimRootModuleCollector.createDescription(project, library, Collections.singletonList(new HdlDescription.SingletonSimRootModule(testbench.getName(), false))));


		final ModelsimSimulation sim = ModelsimSimulation.builder()
				.setRunDirectory(runDirectory)
				.addSourceFiles(
						sources.stream()
								.filter(sourceRef -> sourceRef.getFileAssociation().contains(SourceRef.FileType.FUNCTIONAL_SIM))
								.map(SourceRef::getPath)
				)
				.setModuleRootDir(library)
				.setTarget(target)
				.setFatalWarnings(module.getHdl().getVerification() == Verification.FULL)
				.setToplevel(simRoots.functionalRoots)
				.setSimulationRuntime(testbench.getRuntime())
				.logCommandsToFile(runDirectory.resolve("test.do"))
				.loadWaveDoIfExists(true)
				.create();

		sim.run(40L, TimeUnit.SECONDS);
	}

	@Test
	public void runTestbenchQuartus() throws IOException {
		runTestbench(TestingProjectProvider.getInstance().newProjectWithQuartusToolchain(module.getSourceFile().getParent()));
	}

	@Test
	public void runTestbenchIse() throws IOException {
		runTestbench(TestingProjectProvider.getInstance().newProjectWithIseToolchain(module.getSourceFile().getParent()));
	}

	@Test
	public void runTestbenchVivado() throws IOException {
		runTestbench(TestingProjectProvider.getInstance().newProjectWithVivadoToolchain(module.getSourceFile().getParent()));
	}

}
