/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.integration_test;

import de.tu_darmstadt.rs.process.ProcessUtil;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.HdlDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.Verification;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.jconfig.ILibJConfigProject;
import de.tu_darmstadt.rs.spartanmc.jconfig.TestingProjectProvider;
import de.tu_darmstadt.rs.spartanmc.jconfig.api.IConfigurationDocument;
import de.tu_darmstadt.rs.spartanmc.jconfig.internal.SimRootModuleCollector;
import de.tu_darmstadt.rs.spartanmc.jconfig.internal.SourceFileCollector;

import com.googlecode.junittoolbox.ParallelParameterized;
import org.junit.Assert;
import org.junit.Assume;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Verify that each module can be independently compiled by modelsim and xst
 */
@RunWith(ParallelParameterized.class)
public class ModuleVerification {

	private static final Logger logger = LoggerFactory.getLogger(ModuleVerification.class);

	@Parameterized.Parameters(name = "{1}")
	public static Iterable<?> data() throws IOException {

		Library library = TestingProjectProvider.getInstance().getLibrary();

		return library.getAvailableModules().stream()
				.filter(module -> module.getType() != IModuleDescription.ModuleType.Code)
				.sorted(Comparator.comparing(ModuleDescription::getLabel))
				//.filter(md -> md.getName().equals("sim_stopper"))
				.map(module -> new Object[]{
						module,
						module.getLabel() + " (" + module.getName() + ")",
						library
				}).collect(Collectors.toList());
	}

	@Parameterized.Parameter
	public ModuleDescription module;

	@Parameterized.Parameter(value = 1)
	public String name;

	@Parameterized.Parameter(value = 2)
	public Library library;

	@Rule
	public TemporaryFolder temporaryFolder = new TemporaryFolder();

	/**
	 * Check that the module can be compiled and linked independently from other modules with XST.
	 *
	 * Without a part set, XST defaults to the same weird mode as for Spartan 3, where it gets confused about some
	 * valid language constructs.
	 *
	 * @throws InterruptedException
	 * @throws IOException
	 * @throws TimeoutException
	 */
	@Test
	public void testXstSynthesizable() throws InterruptedException, IOException, TimeoutException {
		final ILibJConfigProject project = TestingProjectProvider.getInstance().newProjectWithIseToolchain(null);
		checkVerifiable(project.getDocumentHandler().getDocument().getTarget());

		final Collection<SourceRef> sources = Stream.of(module).collect(SourceFileCollector.createDescription(project, library, Collections.emptyList()));

		Path tempDirectory = temporaryFolder.newFolder().toPath();


		//The elaborate command only does the equivalent of compiling, not linking. There is no checking done
		//to see if all referenced modules are known. However, the Modelsim test already does the linking. Also,
		//the run command which would do that checking does way too much in that it produces a netlist. That wastes
		//much time for little gain, so we don't do it.
		Files.write(tempDirectory.resolve("verify.xst"), Arrays.asList(
				"elaborate",
				"-vlgincdir                 {\"" + ModelsimSimulation.getModulesRoot(library) + "\"}",
				"-ifn verify.prj"
		));

		final Stream<CharSequence> stringStream = sources.stream()
				//Filter for synthesis files, as testbenches need not conform to XST's verilog subset
				.filter(sourceRef -> sourceRef.getFileAssociation().contains(SourceRef.FileType.SYNTHESIS))
				.map(ModelsimSimulation.sourceRefMapper("verilog work \"%s\"", "vhdl work \"%s\""));
		Files.write(tempDirectory.resolve("verify.prj"), stringStream::iterator);

		String xstFilename = System.getenv("XST_EXECUTABLE");
		if (xstFilename == null || xstFilename.isEmpty()) {
			xstFilename = "xst";
		}

		ProcessBuilder builder = new ProcessBuilder(xstFilename, "-intstyle", "ise", "-ifn", "verify.xst");
		builder.directory(tempDirectory.toFile());


		OutputParser outputParser = new XstOutputParser();

		int exitCode = ProcessUtil.runAndWaitFor(builder, (Consumer<String>) outputParser::parseLine);

		if (exitCode != 0) {
			//Internal XST error. We can't say if the module is valid, so ignore it...
			Assume.assumeFalse("XST should not have internal errors...", exitCode == 134);
			Assert.assertTrue("Exit code: " + exitCode + " Unknown error, " + outputParser.getFullLog(), outputParser.hasErrors());
			throw new AssertionError("Exit code: " + exitCode + '\n' + outputParser.getErrors().stream().collect(Collectors.joining("\n")));
		}

	}

	private void checkVerifiable(IConfigurationDocument.ITargetConfiguration target) {
		Assume.assumeFalse("User disabled verification", module.getHdl().getVerification() == Verification.NONE);
		Assume.assumeFalse("Dynamic Structure modules cannot currently be verified", module.getType() == IModuleDescription.ModuleType.DynamicStructure);
		Assume.assumeFalse("Modules built by TCL scripts cannot currently be verified", module.getTcl() != null); //TODO: merge with dynamic structure modules
		Assume.assumeTrue("Module not compatible with toolchain", module.supports(target.getCurrentConfiguration()));
	}


	private void testModelsimSynthesizable(ILibJConfigProject project) throws InterruptedException, IOException, TimeoutException{
		final IConfigurationDocument.ITargetConfiguration target = project.getDocumentHandler().getDocument().getTarget();

		checkVerifiable(target);

		Collection<SourceRef> sources = Stream.of(module)
				.collect(SourceFileCollector.createDescription(
						project,
						library,
						Collections.emptyList()
				));

		final SimRootModuleCollector.SimRoots simRoots = Stream.of(module).collect(SimRootModuleCollector.createDescription(project, library, Collections.singletonList(new HdlDescription.SingletonSimRootModule(TestingToplevelDesign.toplevelName, false))));

		Path tempDirectory = temporaryFolder.newFolder().toPath();
		final TestingToplevelDesign testingToplevelDesign = new TestingToplevelDesign(module, library, target.getToolchain());
		Path toplevel = testingToplevelDesign.createToplevelDesign(tempDirectory);

		ModelsimSimulation sim = ModelsimSimulation.builder()
				.setRunDirectory(tempDirectory)
				.addSourceFiles(
						sources.stream()
								//Filter for synthesis files, as we want to check if all referenced modules are there
								//and some may be wrongly associated as simulation ones
								.filter(sourceRef -> sourceRef.getFileAssociation().contains(SourceRef.FileType.SYNTHESIS))
								.map(SourceRef::getPath),
						"spartanmc"
				)
				.addSourceFile(toplevel)
				.setModuleRootDir(library)
				.setTarget(target)
				.setFatalWarnings(module.getHdl().getVerification() == Verification.FULL)
				.setToplevel(simRoots.functionalRoots)
				.setSimulationRuntime("1ps")
				.create();

		sim.run(40L, TimeUnit.SECONDS);
	}

	/**
	 * Check that the module can be compiled and linked independently from other modules with modelsim on a Xilinx FPGA
	 *
	 * @throws InterruptedException
	 * @throws IOException
	 * @throws TimeoutException
	 */
	@Test
	public void testModelsimSynthesizableIse() throws InterruptedException, IOException, TimeoutException {
		testModelsimSynthesizable(TestingProjectProvider.getInstance().newProjectWithIseToolchain(null));
	}


	/**
	 * Check that the module can be compiled and linked independently from other modules with modelsim on a Xilinx FPGA
	 *
	 * @throws InterruptedException
	 * @throws IOException
	 * @throws TimeoutException
	 */
	@Test
	public void testModelsimSynthesizableVivado() throws InterruptedException, IOException, TimeoutException {
		testModelsimSynthesizable(TestingProjectProvider.getInstance().newProjectWithVivadoToolchain(null));
	}

	/**
	 * Check that the module can be compiled and linked independently from other modules with modelsim on an Altera FPGA
	 *
	 * @throws InterruptedException
	 * @throws IOException
	 * @throws TimeoutException
	 */
	@Test
	public void testModelsimSynthesizableQuartus() throws InterruptedException, IOException, TimeoutException {
		testModelsimSynthesizable(TestingProjectProvider.getInstance().newProjectWithQuartusToolchain(null));
	}


}
