/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.integration_test;

import de.tu_darmstadt.rs.spartanmc.devxml.AttributeUtil;
import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.generic.ConnectorDirection;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.memory.Memory;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.parameters.IParameter;
import de.tu_darmstadt.rs.spartanmc.devxml.descriptions.ports.IConnectorDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.devices.DeviceDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.MemoryModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.PeripheralModuleDescription;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.languages.VerilogLang;
import de.tu_darmstadt.rs.spartanmc.jconfig.generation.outputGroups.VerilogOutputGroup;
import de.tu_darmstadt.rs.util.table.Column;
import de.tu_darmstadt.rs.util.table.TableFormat;
import de.tu_darmstadt.rs.util.table.TableFormatter;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class TestingToplevelDesign {

	private static final Logger logger = LoggerFactory.getLogger(TestingToplevelDesign.class);
	public static final String toplevelName = "testingToplevel";


	private static class MinimalConnectorDescription implements IConnectorDescription {

		private final String name;
		private final ConnectorDirection direction;
		private final boolean directionLocked;
		private final Type type;

		public MinimalConnectorDescription(IConnectorDescription source, String overrideName) {
			name = overrideName;
			direction = source.getDirection();
			directionLocked = source.isDirectionLocked();
			type = source.getType();

		}

		public MinimalConnectorDescription(IConnectorDescription source) {
			name = source.getName();
			direction = source.getDirection();
			directionLocked = source.isDirectionLocked();
			type = source.getType();
		}

		@Override
		public String getName() {
			return name;
		}

		@Override
		public ConnectorDirection getDirection() {
			return direction;
		}

		@Override
		public boolean isDirectionLocked() {
			return directionLocked;
		}

		@Override
		public Type getType() {
			return type;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) {
				return true;
			}
			if (!(o instanceof MinimalConnectorDescription)) {
				return false;
			}

			MinimalConnectorDescription that = (MinimalConnectorDescription) o;

			if (directionLocked != that.directionLocked) {
				return false;
			}
			if (!name.equals(that.name)) {
				return false;
			}
			if (direction != that.direction) {
				return false;
			}
			return type == that.type;
		}

		@Override
		public int hashCode() {
			int result = name.hashCode();
			result = 31 * result + direction.hashCode();
			result = 31 * result + (directionLocked? 1 : 0);
			result = 31 * result + (type != null? type.hashCode() : 0);
			return result;
		}
	}


	private final ModuleDescription module;
	private final Library library;
	private final DeviceDescription.Toolchain toolchain;

	public TestingToplevelDesign(ModuleDescription module, Library library, DeviceDescription.Toolchain toolchain) {

		this.module = module;
		this.library = library;
		this.toolchain = toolchain;
	}

	public Path createToplevelDesign(Path directory) {
		Path out = directory.resolve(toplevelName + ".v");
		VerilogOutputGroup group = new VerilogOutputGroup(out) {
			@Override
			protected String getHeader() {
				return "";
			}

			@Override
			protected void buildContent() {
				writeModuleStart(toplevelName, Stream.empty(), null, null);

				writeModuleInstanceStart(module.getHdl().getToplevel());
				TableFormatter table = new TableFormatter(TableFormat.CHARACTERS, Column.STRING, Column.STRING);
				table.addHeaderRow("Parameter Name", "Value");
				getParams().forEach(op -> table.addContentRow(op.getName(), op.getValue()));
				table.output().forEach(logger::info);
				writeModuleInstanceParams(getParams());

				Stream<? extends IConnectorDescription> fromBusses =
						module.getBuses().stream().flatMap(bus -> {
							final Map<String, String> nameMapping = bus.getPortMappings().stream().collect(Collectors.toMap(m -> m.getBusName(), m -> m.getName()));
							return library.getBusDescription(bus.getBusDescriptionName()).getMatchingPorts(bus.getRole())
									.getPorts().stream()
									.map(port -> new MinimalConnectorDescription(port, nameMapping.getOrDefault(port.getName(), port.getName())));
						}).map(MinimalConnectorDescription::new).distinct();

				Stream<IConnectorDescription> fromPorts =
						module.getPortGroups().stream().flatMap(p -> p.getPorts().stream());

				writeModuleInstanceNameAndPorts(
						"UUT",
						Stream.concat(fromBusses, fromPorts),
						IConnectorDescription::getName,
						p -> null,
						IConnectorDescription::getDirection
				);

				writeMemoryInitalization();

				writeModuleEnd();
			}

			private String intParamValToString(IParameter param, int radix) {
				if (param.getDefaultValue() == null) {
					//TODO this is ugly...
					//Probably clock frequency...
					if (param.hasUiUnit()) {
						return "10000000";
					}
					if (param.getRange()!=null) {
						return param.getRange().getMin().toString();
					}
					return "1";
				}
				long val = AttributeUtil.parseLong(param.getDefaultValue(), radix);
				return VerilogLang.VerilogParamConverter.convertInt(val, radix, param.getName());
			}

			private String paramValToString(IParameter param) {
				try {
					switch (param.getType()) {

						case HexType:
							return intParamValToString(param, 16);
						case IntType:
							return intParamValToString(param, 10);
						case BinType:
							return intParamValToString(param, 2);
						case FloatType:
							if (param.getDefaultValue() == null) {
								return "1";
							}
							return String.format(Locale.US, "%.6f", Float.parseFloat(param.getDefaultValue()));
						case BoolType:
							if ("1".equals(param.getDefaultValue()) || "0".equals(param.getDefaultValue())) {
								return param.getDefaultValue();
							}
							//Fallthrough
						case StringType:
							return VerilogLang.VerilogParamConverter.toWrappedString(param.getDefaultValue());
						default:
							throw new RuntimeException("unknown type");
					}
				} catch (RuntimeException e) {
					throw new RuntimeException("failed to process " + param, e);
				}
			}

			private Stream<OutputParameter> getParams() {
				List<OutputParameter> additional = new ArrayList<>();
				if (module instanceof PeripheralModuleDescription) {
					PeripheralModuleDescription peri = (PeripheralModuleDescription) module;
					if (peri.hasMemory()) {
						additional.add(new OutputParameter(peri.getMemory().getBaseAdrParam(), "0"));
					}
					if (peri.hasRegisters()) {
						additional.add(new OutputParameter(peri.getRegisters().getBaseAdrParam(), "0"));
					}

				}
				return Stream.concat(
						module.getParameterGroups().stream()
								.flatMap(pg -> pg.getParameters().stream())
								.filter(p -> !p.isVirtual())
								.map(p -> new OutputParameter(p.getName(), paramValToString(p))),
						additional.stream()
				);
			}

			private void writeMemoryInitalization() {
				List<? extends Memory> memories = null;
				if (module instanceof PeripheralModuleDescription) {
					PeripheralModuleDescription peri = (PeripheralModuleDescription) module;
					if (peri.hasMemory()) {
						memories = Collections.singletonList(peri.getMemory());
					}
				}

				if (module instanceof MemoryModuleDescription) {
					MemoryModuleDescription memModule = (MemoryModuleDescription) module;
					memories = memModule.getMemories();
				}

				if (memories == null) {
					return;
				}

				for (Memory memory : memories) {
					if (memory.getInstance() != null) {
						Integer bramColumnIndex = null;
						if (memory.getBlockWidth() == 9) {
							bramColumnIndex = 0;
						}
						final String instance = "UUT." + memory.getInstance().resolveMemoryIndex(0).resolve(0, bramColumnIndex);
						writeMemoryInit(instance.replace('/', '.'));
					}
				}

			}
			private void writeMemoryInit(String memoryInstanceName) {
				switch (toolchain.getToolchain()) {
					case "ise":
					case "vivado":
						writeDefparam(memoryInstanceName+".INIT_00", "0");
						break;
					case "quartus":
						writeDefparam(memoryInstanceName+".init_file","\"\"");
						break;
					default:
						throw new RuntimeException("Unsupported toolchain");
				}

			}
		};
		try {
			group.build();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return out;
	}

}
