/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.nettype;

import de.tu_darmstadt.rs.spartanmc.devxml.Library;
import de.tu_darmstadt.rs.spartanmc.devxml.configurations.IToolchainConfiguration;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.IModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.files.modules.ModuleDescription;
import de.tu_darmstadt.rs.spartanmc.devxml.sources.SourceRef;
import de.tu_darmstadt.rs.spartanmc.header_check.CCommentRemover;
import de.tu_darmstadt.rs.spartanmc.jconfig.TestingProjectProvider;

import org.junit.Assert;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;

import java.io.BufferedReader;
import java.io.IOException;
import java.nio.file.Files;
import java.util.Comparator;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * Check that all verilog files contain a default nettype directive
 */
@RunWith(Parameterized.class)
public class DefaultNettypeCheck {

	@Parameterized.Parameters(name = "{1}")
	public static Iterable<?> data() throws IOException {

		Library library = TestingProjectProvider.getInstance().getLibrary();

		return library.getAvailableModules().stream()
				.sorted(Comparator.comparing(ModuleDescription::getLabel))
				.map(module -> new Object[]{
						module,
						module.getLabel() + " (" + module.getName() + ")",
						library
				}).collect(Collectors.toList());
	}

	@Parameterized.Parameter
	public IModuleDescription module;

	@Parameterized.Parameter(value = 1)
	public String name;

	@Parameterized.Parameter(value = 2)
	public Library library;

	@Test
	public void checkNettype() {
		final String invalidFiles = module.getHdl().getSources().streamAllResolvedPaths(module, IToolchainConfiguration.ANY)
				.map(SourceRef::getPath)
				.filter(f -> f.toString().endsWith(".v"))
				.filter(file -> {
					try (BufferedReader bufferedReader = Files.newBufferedReader(file)) {
						CCommentRemover contents = new CCommentRemover(bufferedReader, CCommentRemover.BackSlashCommentEOL.ErrorIfSignificant);

						final Optional<String> firstContentLine =  contents.stream()
								.filter(line -> !line.trim().isEmpty())
								.filter(line -> !isOtherPreprocessorDirective(line))
								.findFirst();

						//Empty file is ok
						if (!firstContentLine.isPresent()) {
							return false;
						}

						return !hasNettype(firstContentLine.get());
					} catch (IOException | RuntimeException e) {
						throw new RuntimeException("Error in " + file, e);
					}
				})
				.map(Object::toString)
				.collect(Collectors.joining("\n"));
		if (!invalidFiles.isEmpty()) {
			Assert.fail("Some files are missing a default_nettype or declare it at the wrong position:\n" + invalidFiles);
		}
	}

	private boolean isOtherPreprocessorDirective(String line) {
		Stream<String> directives = Stream.of(
				"`ifndef",
				"`ifdef",
				"`define",
				"`include",
				"`timescale"
		);
		return directives.anyMatch(line::startsWith);
	}

	private boolean hasNettype(String s) {
		return s.contains("`default_nettype ");
	}

}
