/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc;

import de.tu_darmstadt.rs.jcommander.JCommanderRunner;
import de.tu_darmstadt.rs.spartanmc.client.Client;
import de.tu_darmstadt.rs.spartanmc.server.Server;
import de.tu_darmstadt.rs.spartanmc.server.VivadoInstance;
import de.tu_darmstadt.rs.util.BuildInfo;

import com.beust.jcommander.Parameter;
import com.beust.jcommander.ParameterException;

import java.nio.file.Path;
import java.util.List;

/**
 * Main class
 */
public class VivadoWrapper implements Runnable {

	public static final int PORT = 0x5D3C;
	@Parameter(names = "--server")
	private boolean server;
	@Parameter(names = "--executable", required = true, description = "Vivado executable")
	private Path executable;
	@Parameter(names = "--project", description = "Project dir to run in")
	private Path project;
	@Parameter(names = "--script", description = "Script to execute. May be passed multiple times")
	private List<Path> script;
	@Parameter(names = "--stopForProject", description = "Stop all vivado instances for project")
	boolean stopForProject;
	@Parameter(names = "--stopServer", description = "Stop server completely")
	boolean stopServer;

	@Parameter(names = "--no-daemon", description = "Directly start Vivado without a Server. Useful for CI")
	boolean noDaemon = System.getenv("VIVADO_WRAPPER_NO_DAEMON")!=null;

	public VivadoWrapper() {
	}

	/**
	 * Entry point
	 *
	 * @param args cmd line arguments
	 * @throws Exception on errors...
	 */
	public static void main(String[] args) throws Exception {
		JCommanderRunner.create(new VivadoWrapper()).run(args);
	}

	public void run() {
		if (server) {
			if (noDaemon) {
				throw new ParameterException("Cannot have both --server and --no-daemon");
			}
			if (stopForProject || stopServer) {
				throw new ParameterException("Cannot have stop together with server");
			}
			try {
				new Server().start(executable);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		} else if (noDaemon) {
			String errorInfo = VivadoInstance.runWithoutDaemon(executable, project, script);
			if (errorInfo!=null) {
				System.err.println(errorInfo);
				System.exit(1);
			}
		} else {
			if (project == null && !stopServer) {
				throw new ParameterException("Client needs project");
			}
			if (stopForProject || stopServer) {
				if (script != null) {
					throw new ParameterException("Cannot have stop together with script");
				}
			} else if (script == null) {
				throw new ParameterException("Client needs script");
			}
			try {
				new Client().start(project, script, executable, stopForProject, stopServer);
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	}

	private static BuildInfo buildInfo = null;

	public static BuildInfo getBuildInfo() {
		if (buildInfo == null) {
			buildInfo = new BuildInfo(VivadoWrapper.class);
		}
		return buildInfo;
	}
}
