/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.client;

import de.tu_darmstadt.rs.spartanmc.VivadoWrapper;
import de.tu_darmstadt.rs.spartanmc.message.CommandMessage;
import de.tu_darmstadt.rs.spartanmc.message.Message;
import de.tu_darmstadt.rs.spartanmc.message.StopMessage;
import de.tu_darmstadt.rs.spartanmc.message.VersionMessage;

import io.netty.bootstrap.Bootstrap;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.codec.serialization.ClassResolvers;
import io.netty.handler.codec.serialization.ObjectDecoder;
import io.netty.handler.codec.serialization.ObjectEncoder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zeroturnaround.process.Processes;

import java.io.File;
import java.io.IOException;
import java.net.ConnectException;
import java.nio.file.Path;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * The client. Sends commands and displays output
 */
public final class Client {

	public static final int CONNECT_WAIT_TIME_MILLIS = 1000;
	public static final int NEW_SERVER_TRY_COUNT = 3;
	private static final Logger logger = LoggerFactory.getLogger(Client.class);

	public Client() {
	}

	/**
	 * Start the client
	 *
	 * @param projectPath project path
	 * @param scriptPath path to scripts to execute
	 * @param vivadoExecutable path to server executable
	 * @param stopProject stop vivado instances for project
	 * @param stopServer stop server
	 * @throws Exception on errors...
	 */
	public void start(Path projectPath, List<Path> scriptPath, Path vivadoExecutable, boolean stopProject, boolean stopServer) throws Exception {

		final List<Message> firstMessage;
		if (stopProject || stopServer) {
			firstMessage = Arrays.asList(new VersionMessage(), new StopMessage(projectPath, stopServer));
		} else {
			firstMessage = Arrays.asList(new VersionMessage(), new CommandMessage(projectPath, scriptPath));
		}

		final CommandException[] commandException = {null};
		EventLoopGroup group = new NioEventLoopGroup();
		try {
			Bootstrap b = new Bootstrap();
			b.group(group)
					.channel(NioSocketChannel.class)
					.handler(new ChannelInitializer<SocketChannel>() {
						@Override
						public void initChannel(SocketChannel ch) throws Exception {
							ChannelPipeline p = ch.pipeline();
							p.addLast(
									new ObjectEncoder(),
									new ObjectDecoder(ClassResolvers.cacheDisabled(null)),
									new ClientHandler(firstMessage, (e) -> commandException[0] = e)
							);
						}
					});
			// Start the connection attempt.
			try {
				doConnection(b);
			} catch (ConnectException e) {
				if (!stopProject && !stopServer) {
					logger.info("Could not connect to server, starting one...");
					startServer(vivadoExecutable);

					for (int tries = 0; tries < NEW_SERVER_TRY_COUNT; tries++) {
						try {
							logger.info("Connecting to newly created server");
							Thread.sleep(CONNECT_WAIT_TIME_MILLIS);
							doConnection(b);
							break;
						} catch (ConnectException e2) {
							logger.warn("Could not connect: {}",e2.getMessage());

							//Swallow exception --> retry
						}
					}
				} else {
					logger.warn("There is no server to stop");
				}
			}
		} finally {
			group.shutdownGracefully(0, 0, TimeUnit.MILLISECONDS);
		}
		if (commandException[0] != null) {
			logger.error(commandException[0].getMessage());
			System.exit(1);
		}
	}

	private void startServer(Path vivadoExecutable) throws IOException {
		final File tempFile = File.createTempFile("spmc-vivado-wrapper-", ".txt");
		logger.info("Logging server output to {}", tempFile);
		String separator = System.getProperty("file.separator");
		String classpath = System.getProperty("java.class.path");
		String path = System.getProperty("java.home")
				+ separator + "bin" + separator + "java";
		if (!new File(path).isFile()) {
			throw new RuntimeException("Java executable does not exist at " + path);
		}
		final File tempDir = tempFile.getParentFile();
		if (!tempDir.isDirectory()) {
			throw new RuntimeException("tempDir is not a directory at " + tempDir);
		}
		ProcessBuilder processBuilder =
				new ProcessBuilder("setsid", "-w", "--", path, "-cp",
						classpath,
						VivadoWrapper.class.getName(),
						"--server",
						"--executable",
						vivadoExecutable.toAbsolutePath().toString());
		processBuilder.directory(tempDir);
		processBuilder.redirectErrorStream(true);
		processBuilder.redirectOutput(tempFile);

		final Process process = processBuilder.start();

		logger.info("Started server with PID " + Processes.newPidProcess(process).getPid());

		process.getOutputStream().close();
	}

	private void doConnection(Bootstrap b) throws InterruptedException, ConnectException {
		b.connect("127.0.0.1", VivadoWrapper.PORT).sync().channel().closeFuture().sync();
	}

}
