/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.client;

import de.tu_darmstadt.rs.spartanmc.message.AbortCommandMessage;
import de.tu_darmstadt.rs.spartanmc.message.EndOfCommandMessage;
import de.tu_darmstadt.rs.spartanmc.message.LogMessage;
import de.tu_darmstadt.rs.spartanmc.message.Message;
import de.tu_darmstadt.rs.spartanmc.message.MessageVisitorAdapter;
import de.tu_darmstadt.rs.spartanmc.message.NewInstanceMessage;

import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelInboundHandlerAdapter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.function.Consumer;

/**
 * Handler implementation for the client.
 */
public class ClientHandler extends ChannelInboundHandlerAdapter {

	private static final Logger logger = LoggerFactory.getLogger(ClientHandler.class);
	private final List<Message> firstMessage;
	private final Consumer<CommandException> commandExceptionConsumer;
	private boolean isActive = false;

	private Thread handlerThread;

	/**
	 * Creates a client-side handler.
	 *
	 * @param firstMessage initial message to send to the server
	 * @param commandExceptionConsumer consumer for commandExceptions
	 */
	public ClientHandler(List<Message> firstMessage, Consumer<CommandException> commandExceptionConsumer) {
		this.firstMessage = firstMessage;
		this.commandExceptionConsumer = commandExceptionConsumer;
	}

	@Override
	public void channelActive(ChannelHandlerContext ctx) {
		firstMessage.forEach(ctx::writeAndFlush);
		isActive = true;
		Runtime.getRuntime().addShutdownHook(new Thread(() -> {
			if (isActive) {
				ctx.writeAndFlush(new AbortCommandMessage());
				while (isActive) {
					try {
						/*System.out.println("handlerThread = " + handlerThread.isAlive());
						synchronized (shutdownSynchronize) {
							shutdownSynchronize.wait();
						}*/
						handlerThread.join();
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
				}
			}
		}, "Shutdown hook"));

		ctx.channel().closeFuture().addListener(future -> {
			if (isActive) {
				isActive = false;
				final CommandException commandException = new CommandException("Connection lost");
				commandExceptionConsumer.accept(commandException);
			}

		});
	}

	@Override
	public void channelRead(ChannelHandlerContext ctx, Object msg) {
		if (!(msg instanceof Message)) {
			throw new RuntimeException("Received object is not a message: " + msg);
		}

		handlerThread = Thread.currentThread();

		Message message = (Message) msg;
		message.visit(new MessageVisitorAdapter<RuntimeException>() {
			@Override
			public void visitLog(LogMessage message) throws RuntimeException {
				//CHECKSTYLE.OFF: SystemOut - Log messages from Vivado should not be prefixed with our stuff
				System.out.println(message.message);
				//CHECKSTYLE.ON: SystemOut
			}

			@Override
			public void visitEndOfCommand(EndOfCommandMessage message) throws RuntimeException {
				isActive = false;
				logger.info("command end");
				if (message.errorInfo == null) {
					logger.info("Command finished ok");
				} else {
					throw new CommandException("Error: " + message.errorInfo);
				}
				ctx.close();
			}

			@Override
			public void visitNewInstance(NewInstanceMessage newInstanceMessage) throws RuntimeException {
				logger.info("New Vivado instance started, please wait");
			}

			@Override
			public void visitDefault(Message message) {
				throw new RuntimeException("Unexpected message: " + message);
			}
		});
	}

	@Override
	public void channelReadComplete(ChannelHandlerContext ctx) {
		ctx.flush();
	}

	@Override
	public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) {
		if (cause instanceof CommandException) {
			commandExceptionConsumer.accept((CommandException) cause);
		} else {
			logger.warn("Client error", cause);
		}
		ctx.close();
	}
}