/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.server;

import de.tu_darmstadt.rs.spartanmc.VivadoWrapper;

import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.Channel;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.ChannelPipeline;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.SocketChannel;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.handler.codec.serialization.ClassResolvers;
import io.netty.handler.codec.serialization.ObjectDecoder;
import io.netty.handler.codec.serialization.ObjectEncoder;
import io.netty.handler.logging.LogLevel;
import io.netty.handler.logging.LoggingHandler;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.nio.file.Path;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.OptionalLong;
import java.util.function.BiConsumer;

/**
 * The Server
 */
public class Server {

	public static final int MILLIS = 1000;
	private static final Logger logger = LoggerFactory.getLogger(Server.class);
	/**
	 * Time after which the server will shut down after all Vivado instances have been closed
	 */
	private static final long MAX_SERVER_AGE_SECONDS = 10;
	private Channel channel;
	private Instant serverEmptyInstant;

	public Server() {
	}

	/**
	 * Start the server
	 * @param executable vivado executable
	 * @throws Exception on errors...
	 */
	public void start(Path executable) throws Exception {

		String waitObject = "asdf";
		serverEmptyInstant = Instant.now();

		final BiConsumer<VivadoInstance, VivadoInstanceFactory> onInstCountChange = (inst, factory) -> {
			if (!factory.hasInstances()) {
				serverEmptyInstant = Instant.now();
			}
			synchronized (waitObject) {
				waitObject.notify();
			}
		};
		VivadoInstanceFactory factory = new VivadoInstanceFactory(executable, onInstCountChange, onInstCountChange);

		//Instance pruning
		final Thread pruningThread = new Thread("Instance pruning") {
			public void run() {
				while (true) {
					try {
						final OptionalLong highestInstanceAge = factory.getHighestInstanceAge();
						long waitSeconds;
						if (highestInstanceAge.isPresent()) {
							waitSeconds = VivadoInstanceFactory.MAX_INSTANCE_AGE_SECONDS - highestInstanceAge.getAsLong();
						} else {
							waitSeconds = MAX_SERVER_AGE_SECONDS - ChronoUnit.SECONDS.between(serverEmptyInstant, Instant.now());
						}

						logger.info("Pruning Thread waits for {} seconds", waitSeconds);
						synchronized (waitObject) {
							if (waitSeconds > 0 && !factory.isStopServer()) {
								waitObject.wait(MILLIS * waitSeconds);
							}
						}
						if (isInterrupted()) {
							return;
						}
						if (highestInstanceAge.isPresent()) {
							factory.pruneInstances();
							if (!factory.hasActiveInstances()) {
								serverEmptyInstant = Instant.now();
							}
						} else {
							//Still empty?, using millis get sub-second-accuracy
							final boolean timeout = ChronoUnit.MILLIS.between(serverEmptyInstant, Instant.now()) > MILLIS * MAX_SERVER_AGE_SECONDS;
							if (!factory.hasInstances() && (timeout || factory.isStopServer())) {
								logger.info("Shutting down due to lack of use");
								channel.close();
								return;
							}
						}
					} catch (InterruptedException e) {
						if (isInterrupted()) {
							return;
						}
					}
				}
			}
		};
		pruningThread.start();

		EventLoopGroup bossGroup = new NioEventLoopGroup(1);
		EventLoopGroup workerGroup = new NioEventLoopGroup();
		try {
			ServerBootstrap b = new ServerBootstrap();
			b.group(bossGroup, workerGroup)
					.channel(NioServerSocketChannel.class)
					.handler(new LoggingHandler(LogLevel.INFO))
					.childHandler(new ChannelInitializer<SocketChannel>() {
						@Override
						public void initChannel(SocketChannel ch) throws Exception {
							ChannelPipeline p = ch.pipeline();
							p.addLast(
									new ObjectEncoder(),
									new ObjectDecoder(ClassResolvers.cacheDisabled(null)),
									new ServerHandler(factory));
						}
					});

			// Bind and start to accept incoming connections.
			final Channel channel = b.bind(VivadoWrapper.PORT).sync().channel();
			this.channel = channel;
			channel.closeFuture().sync();
		} finally {
			pruningThread.interrupt();
			bossGroup.shutdownGracefully();
			workerGroup.shutdownGracefully();
		}

	}
}
