/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.server;

import de.tu_darmstadt.rs.spartanmc.VivadoWrapper;
import de.tu_darmstadt.rs.spartanmc.message.AbortCommandMessage;
import de.tu_darmstadt.rs.spartanmc.message.CommandMessage;
import de.tu_darmstadt.rs.spartanmc.message.EndOfCommandMessage;
import de.tu_darmstadt.rs.spartanmc.message.LogMessage;
import de.tu_darmstadt.rs.spartanmc.message.Message;
import de.tu_darmstadt.rs.spartanmc.message.MessageVisitorAdapter;
import de.tu_darmstadt.rs.spartanmc.message.NewInstanceMessage;
import de.tu_darmstadt.rs.spartanmc.message.StopMessage;
import de.tu_darmstadt.rs.spartanmc.message.VersionMessage;

import io.netty.channel.ChannelFutureListener;
import io.netty.channel.ChannelHandlerContext;
import io.netty.channel.ChannelInboundHandlerAdapter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Paths;

/**
 * Handler for the Vivado server
 */
public class ServerHandler extends ChannelInboundHandlerAdapter {


	private static final Logger logger = LoggerFactory.getLogger(ServerHandler.class);

	private VivadoInstance instance;
	private final VivadoInstanceFactory factory;

	public ServerHandler(VivadoInstanceFactory factory) {
		this.factory = factory;
	}

	private void instanceOutput(String s) {
		logger.info("Sending log to client: " + s);
	}

	@Override
	public void channelRead(ChannelHandlerContext ctx, Object msg) {
		try {
			if (!(msg instanceof Message)) {
				throw new RuntimeException("Received object is not a message: " + msg);
			}

			Message message = (Message) msg;
			logger.info("server got {}", message);
			message.visit(new MessageVisitorAdapter<IOException>() {
				@Override
				public void visitCommand(CommandMessage message) throws IOException {
					instance = factory.getInstance(Paths.get(message.projectPath));
					if (instance == null) {
						ctx.writeAndFlush(new EndOfCommandMessage("Server is shutting down. Cannot get new instances."));
						ctx.close();
						return;
					}
					if (instance.isNewlyCreated()) {
						ctx.writeAndFlush(new NewInstanceMessage());
					}
					instance.setConsumer(new OutputConsumer() {
						@Override
						public void output(String s) {
							ctx.writeAndFlush(new LogMessage(s));
						}

						@Override
						public void endOfCommand(String error) {
							logger.info("Command finished.");
							ctx.writeAndFlush(new EndOfCommandMessage(error))
									.addListener(ChannelFutureListener.CLOSE);
						}
					});
					instance.sendCommand(message.scriptPath);
				}

				@Override
				public void visitAbortCommand(AbortCommandMessage message) throws IOException {
					instance.abort();
					ctx.writeAndFlush(new LogMessage("Sending SIGINT..."));
				}

				@Override
				public void visitStopServer(StopMessage message) throws IOException {
					if (message.stopServer) {
						final int count = factory.stopServer();
						ctx.write(new LogMessage("Stopped " + count + " instances"));
						if (factory.hasActiveInstances()) {
							ctx.writeAndFlush(new EndOfCommandMessage("There are vivado "+factory.countAllInstances()+" instances still executing commands. Server will be running until they are finished."));
						} else {
							ctx.writeAndFlush(new EndOfCommandMessage(null));
						}
						ctx.close();
					} else {
						final int count = factory.stopInstances(Paths.get(message.projectPath));
						ctx.write(new LogMessage("Stopped " + count + " instances"));
						long remaining = factory.countAllInstancesForProject(Paths.get(message.projectPath));
						if (remaining > 0) {
							ctx.writeAndFlush(new EndOfCommandMessage("Some instances are still processing commands. "+remaining+" instances remaining."));
						} else {
							ctx.write(new LogMessage("All instances are gone."));
							ctx.writeAndFlush(new EndOfCommandMessage(null));
						}
						ctx.close();
					}
				}

				@Override
				public void visitVersion(VersionMessage message) throws IOException {
					final boolean ok = VivadoWrapper.getBuildInfo().equals(message.buildInfo);
					if (!ok) {
						ctx.writeAndFlush(new EndOfCommandMessage("Version mismatch! Server: "+VivadoWrapper.getBuildInfo()+", client: "+message.buildInfo));
						ctx.close();
					} else {
						ctx.writeAndFlush(new LogMessage("Versions match."));
					}
				}

				@Override
				public void visitDefault(Message message) {
					throw new RuntimeException("Unexpected message: " + message);
				}
			});
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public void channelReadComplete(ChannelHandlerContext ctx) {
		ctx.flush();
	}

	@Override
	public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) {
		logger.warn("server error", cause);
		ctx.close();
	}
}