/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.server;

import de.tu_darmstadt.rs.process.StreamGobbler;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zeroturnaround.exec.ProcessExecutor;
import org.zeroturnaround.exec.stream.slf4j.Slf4jStream;
import org.zeroturnaround.process.PidProcess;
import org.zeroturnaround.process.Processes;

import java.io.IOException;
import java.io.PrintStream;
import java.nio.file.Path;
import java.time.Instant;
import java.util.List;
import java.util.stream.Collectors;

/**
 * An instance of Vivado
 */
public class VivadoInstance {

	public static final String END_OF_COMMAND = "__vivadowrapper_cmd_finished__";
	private static final Logger logger = LoggerFactory.getLogger(VivadoInstance.class);
	private final Process process;
	private final PrintStream printStream;
	private final VivadoInstanceFactory factory;
	private final Path project;
	private OutputConsumer consumer;
	private Instant lastAccess = null;
	private boolean newlyCreated = true;
	private boolean exiting = false;

	private BufferingConsumer bufferingConsumer = new BufferingConsumer();

	private Thread shutdownHook;

	/**
	 * Create an instance of vivado
	 *
	 * @param vivadoExecutable the executable
	 * @param project project to run in
	 * @param factory Instance Factory
	 * @throws IOException on broken IO
	 */
	public VivadoInstance(Path vivadoExecutable, Path project, VivadoInstanceFactory factory) throws IOException {
		lastAccess = Instant.now();
		this.project = project;
		this.factory = factory;
		logger.info("starting " + vivadoExecutable + " at " + project);
		//Create new process group and session by wrapping with setsid
		final ProcessBuilder builder = new ProcessBuilder("setsid", "-w", vivadoExecutable.toAbsolutePath().toString(), "-mode", "tcl");
		//final ProcessBuilder builder = new ProcessBuilder("cat");
		builder.directory(project.resolve("build").toFile());
		process = builder.start();
		final StreamGobbler errorGobbler = new StreamGobbler(process.getErrorStream(), this::consumeLine);
		errorGobbler.start();
		final StreamGobbler outGobbler = new StreamGobbler(process.getInputStream(), this::consumeLine);
		outGobbler.start();
		printStream = new PrintStream(process.getOutputStream());

		shutdownHook = new Thread(() -> {
			logger.info("shutdown hook called");
			if (consumer != null) {
				consumer.output("Server is shutting down! Command is being canceled");
			}
			abortSynchronous();
			exit();


			System.out.println("process exited with code "+waitSynchronous());
		});

		Runtime.getRuntime().addShutdownHook(shutdownHook);

		new Thread(() -> {
			while (process.isAlive()) {
				try {
					process.waitFor();
				} catch (InterruptedException e) {
					//Ignore interruptions
				}
			}
			Runtime.getRuntime().removeShutdownHook(shutdownHook);

			//Call waitFor again to get exit code
			int exitCode = 0;
			try {
				exitCode = process.waitFor();
			} catch (InterruptedException e) {
				throw new RuntimeException("Interrupted when process was not even running");
			}
			instanceExited(exitCode, outGobbler, errorGobbler);
		}, "Process exit watch").start();
	}

	public static String runWithoutDaemon(Path executable, Path project, List<Path> script) {
		try {
			final String[] errorInfo = {null};
			VivadoInstance vivadoInstance = new VivadoInstance(executable, project, null);
			vivadoInstance.setConsumer(new OutputConsumer() {
				@Override
				public void output(String s) {
					System.out.println(s);
				}

				@Override
				public void endOfCommand(String errorInfoParam) {
					errorInfo[0] = errorInfoParam;
					vivadoInstance.exit();
				}
			});
			vivadoInstance.sendCommand(script.stream().map(Path::toAbsolutePath).map(Path::toString).collect(Collectors.toList()));

			vivadoInstance.exit();
			vivadoInstance.waitSynchronous();

			return errorInfo[0];
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	private OutputConsumer getConsumerOrBuffer() {
		if (consumer != null) {
			return consumer;
		}
		return bufferingConsumer;
	}

	private synchronized void instanceExited(int exitCode, Thread outGobbler, Thread errorGobbler) {
		logger.warn("Process exited with exit code {}", exitCode);


		try {
			outGobbler.join();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		logger.info("out gobbler joined");
		try {
			errorGobbler.join();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		logger.info("error gobbler joined");
		if (factory != null) {
			factory.notifyInstanceExit(this);
		}
		logger.info("factory informed");
		final String exitInfo = "Process exited with exit code " + exitCode;
		getConsumerOrBuffer().endOfCommand(exitInfo);
	}


	private synchronized void consumeLine(String s) {
		//Only log if we are running a real server
		if (factory != null) {
			logger.info("output: " + s);
		}


		OutputConsumer consumerOrBuffer = getConsumerOrBuffer();
		if (s.startsWith(END_OF_COMMAND)) {
			logger.info("command finished");
			lastAccess = Instant.now();

			String errorOnly = s.substring(END_OF_COMMAND.length()).replace("_newline_", "\n").trim();
			if (errorOnly.isEmpty()) {
				consumerOrBuffer.endOfCommand(null);
				logger.info("Command finished ok");
			} else {
				consumerOrBuffer.endOfCommand(errorOnly);
				logger.info("Command finished with error {}", errorOnly);
			}
			consumer = null;
			if (factory != null) {
				factory.notifyInstanceFree(this);
			}
		} else {
			consumerOrBuffer.output(s);
		}
	}

	/**
	 * Send a command to vivado
	 *
	 * @param scriptPaths List of tcl files to execute
	 */
	public synchronized void sendCommand(List<String> scriptPaths) {
		newlyCreated = false;
		lastAccess = Instant.now();

		logger.info("Sending to instance: " + scriptPaths);
		//Clear error
		printStream.println("set errorInfo \"\";");
		//Output commands
		scriptPaths.forEach(p -> {
			printStream.println("source " + p);
		});
		//Print special string plus error info without newlines
		printStream.println("puts [concat \"" + END_OF_COMMAND + "\" [string map {\\n _newline_} $errorInfo]]");
		printStream.flush();
	}

	public synchronized void setConsumer(OutputConsumer consumer) {
		this.consumer = consumer;
		bufferingConsumer.outputToConsumer(consumer);
	}

	public Path getProject() {
		return project;
	}

	/**
	 * Return the last access, or the specified now if there is currently an access running
	 *
	 * @param now Now, sampled once before to avoid differences
	 * @return time of last access
	 */
	public Instant getLastAccess(Instant now) {
		if (consumer == null) {
			return lastAccess;
		} else {
			return now;
		}
	}

	/**
	 * Request instance to exit
	 */
	public synchronized void exit() {
		printStream.println("exit");
		printStream.flush();
		exiting = true;
	}

	/**
	 * Check if there has been any command executed on this instance
	 *
	 * @return true if there has never been any command executed
	 */
	public boolean isNewlyCreated() {
		return newlyCreated;
	}

	/**
	 * Check if the instance is currently exiting
	 *
	 * @return true if exit has been called
	 */
	public boolean isExiting() {
		return exiting;
	}

	/**
	 * Abort the currently running command by sending a SIGINT to the process group
	 */
	public synchronized void abort() {
		new Thread(() -> {
			abortSynchronous();
		}, "Vivado abort").start();
	}

	private void abortSynchronous() {
		final PidProcess pidProcess = Processes.newPidProcess(process);
		System.out.println("sending sigint to process with PID "+pidProcess.getPid());
		try {
			new ProcessExecutor()
					.commandSplit(String.format("kill -INT -- -%d", pidProcess.getPid()))
					.redirectOutput(Slf4jStream.ofCaller().asDebug()).exitValueNormal().executeNoTimeout();
		} catch (IOException | InterruptedException e) {
			e.printStackTrace();
		}
	}

	public boolean isAlive() {
		return process.isAlive();
	}

	private int waitSynchronous() {
		try {
			return process.waitFor();
		} catch (InterruptedException e) {
			throw new RuntimeException(e);
		}
	}
}
