/*
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */

package de.tu_darmstadt.rs.spartanmc.server;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Path;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.OptionalLong;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.stream.Collectors;
import java.util.stream.LongStream;

/**
 * Creates and manages Vivado Instances
 */
public class VivadoInstanceFactory {

	public static final long MAX_INSTANCE_AGE_SECONDS = 5*60;
	public static final int MILLIS = 1000;
	private static final Logger logger = LoggerFactory.getLogger(VivadoInstanceFactory.class);
	private final Path executable;
	private final BiConsumer<VivadoInstance, VivadoInstanceFactory> onInstanceStart;
	private final BiConsumer<VivadoInstance, VivadoInstanceFactory> onInstanceExit;

	private Set<VivadoInstance> allInstances = new HashSet<>();

	private Map<Path, Set<VivadoInstance>> freeInstances = new HashMap<>();

	private boolean stopServer = false;

	/**
	 * Create a factory
	 * @param executable the vivado executable
	 * @param onInstanceStart listener to call for new instances
	 * @param onInstanceExit listener to call for instances that have exited
	 */
	public VivadoInstanceFactory(Path executable, BiConsumer<VivadoInstance, VivadoInstanceFactory> onInstanceStart, BiConsumer<VivadoInstance, VivadoInstanceFactory> onInstanceExit) {
		this.executable = executable;
		this.onInstanceStart = onInstanceStart;
		this.onInstanceExit = onInstanceExit;
	}

	private Set<VivadoInstance> getInstancesForProject(Path project) {
		return freeInstances.computeIfAbsent(project, (x) -> new HashSet<>());
	}


	public long countAllInstancesForProject(Path project) {
		return allInstances.stream().filter(i -> i.getProject().equals(project))
				.filter(i -> !i.isExiting())
				.count();
	}
	public long countAllInstances() {
		return allInstances.stream()
				.filter(i -> !i.isExiting())
				.count();
	}
	/**
	 * Get a free Vivado instance for a project
	 *
	 * @param project The project
	 * @return a Vivado Instance, either reused or new
	 *
	 * @throws IOException on broken IO
	 */
	public synchronized VivadoInstance getInstance(Path project) throws IOException {
		if (stopServer) {
			return null;
		}
		final Set<VivadoInstance> instsForProject = getInstancesForProject(project);
		if (instsForProject.isEmpty()) {
			final VivadoInstance inst = new VivadoInstance(executable, project, this);
			allInstances.add(inst);
			onInstanceStart.accept(inst, this);
			return inst;
		} else {
			VivadoInstance inst = instsForProject.iterator().next();
			instsForProject.remove(inst);
			return inst;
		}
	}

	/**
	 * Called by instances that are free again after command execution
	 * @param vivadoInstance the instance that is free again
	 */
	public synchronized void notifyInstanceFree(VivadoInstance vivadoInstance) {
		if (stopServer) {
			vivadoInstance.exit();
		} else {
			getInstancesForProject(vivadoInstance.getProject()).add(vivadoInstance);
		}
	}

	/**
	 * Called by instances that have exited
	 * @param vivadoInstance the instance that has exited
	 */
	public synchronized void notifyInstanceExit(VivadoInstance vivadoInstance) {
		allInstances.remove(vivadoInstance);
		getInstancesForProject(vivadoInstance.getProject()).remove(vivadoInstance);
		onInstanceExit.accept(vivadoInstance, this);
	}

	/**
	 * Stop instances that have been unused for longer than MAX_INSTANCE_AGE_SECONDS
	 */
	public synchronized void pruneInstances() {
		Instant now = Instant.now();
		logger.info("Pruning instances. {} instances before. now: {}", allInstances.size(),now);

		final List<VivadoInstance> toClean = freeInstances.values().stream()
				.flatMap(Collection::stream)
				.filter(inst -> !inst.isExiting())
				//Use millis to be able to prune stuff that has just timed out
				.filter(inst -> ChronoUnit.MILLIS.between(inst.getLastAccess(now), now) >= MAX_INSTANCE_AGE_SECONDS * MILLIS)
				.collect(Collectors.toList());
		toClean.forEach(inst -> {
			logger.info("Closing instance for " + inst.getProject() + " because of long unused duration: "+inst.getLastAccess(now));
			inst.exit();
		});
	}

	private synchronized LongStream getInstanceAges() {
		Instant now = Instant.now();
		return allInstances.stream()
				.filter(inst -> !inst.isExiting())
				.mapToLong(inst -> ChronoUnit.SECONDS.between(inst.getLastAccess(now), now));
	}

	public synchronized OptionalLong getHighestInstanceAge() {
		return getInstanceAges().max();
	}

	public synchronized OptionalLong getLowestInstanceAge() {
		return getInstanceAges().min();
	}

	public boolean hasInstances() {
		return !allInstances.isEmpty();
	}

	public boolean hasActiveInstances() {
		return allInstances.stream().anyMatch(i -> !i.isExiting());
	}

	public synchronized int stopInstances(Path path) {
		final Set<VivadoInstance> instancesForProject = getInstancesForProject(path);
		int size = instancesForProject.size();
		instancesForProject.forEach(VivadoInstance::exit);
		instancesForProject.clear();
		return size;
	}

	public synchronized int stopServer() {
		stopServer = true;
		return freeInstances.keySet().stream().mapToInt(this::stopInstances).sum();
	}

	public boolean isStopServer() {
		return stopServer;
	}
}
