#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#ifdef CONFIG_HAVE_TERMCAP
#  include <termcap.h>
#endif

/**
 * Buffer holding current error message.
 * \see \ref action_errmsg()
 */
static char *buf_error_msg;

/**
 * Buffer holding current action message.
 * \see \ref action_start()
 */
static char *buf_action_msg;

/**
 * Buffer holding current progress message.
 * \see \ref action_progress()
 */
static char *buf_action_progress;

/**
 * Buffer used for message formatting.
 */
static char *buf_action_fmt;

/**
 * With of the terminal in columns.
 */
static int term_cols = 80;


/**
 * Initializes static buffers.
 */
void action_init() {
#ifdef CONFIG_HAVE_TERMCAP
    char *termtype = getenv("TERM");
    if (tgetent(NULL, termtype) == 1) {
        term_cols = tgetnum("co");
    }
#endif
    buf_error_msg = calloc(255, sizeof(char));
    buf_action_msg = calloc(term_cols, sizeof(char));
    buf_action_progress = calloc(term_cols, sizeof(char));
    buf_action_fmt = calloc(term_cols, sizeof(char));
}

/**
 * Free static buffers.
 */
void action_finalize() {
    free(buf_error_msg);
    free(buf_action_msg);
    free(buf_action_progress);
    free(buf_action_fmt);
}

/**
 * Print an action message. This function prints a message on stdout of the
 * following form:
 * ~~~~~~~~~~~~~
 * * MESSAGE<NEWLINE>
 * ~~~~~~~~~~~~~
 * \param   *action     Message to print (format string for printf)
 * \param   ...         Arguments for format specifiers in message string
 */
void action_print(char *action, ...) {
    va_list argp;
    va_start(argp, action);
    printf("* ");
    vprintf(action, argp);
    printf("\n");
    va_end(argp);
}

/**
 * Start an action message. This function prints a message on stdout of the
 * following form:
 * ~~~~~~~~~~~~~
 * * MESSAGE...
 * ~~~~~~~~~~~~~
 * Later use of \ref action_progress() and \ref action_end() will modify/append
 * more text to the current line.
 * \param   *action     Message to print (format string for printf)
 * \param   ...         Arguments for format specifiers in message string
 */
void action_start(char *action, ...) {
    va_list argp;
    snprintf(buf_action_fmt, term_cols, "* %s...", action);
    va_start(argp, action);
    vsnprintf(buf_action_msg, term_cols, buf_action_fmt, argp);
    va_end(argp);
    buf_action_progress[0] = '\0';
    printf("%s", buf_action_msg);
    fflush(stdout);
}

/**
 * Update progress information. This function updates the progress text printed
 * in the most recent status message printed with \ref action_start.
 * \param   *fmt    Progress text to print (format string for printf)
 * \param   ...     Arguments for format specifiers in progress text
 */
void action_progress(char *fmt, ...) {
    int msg_width = term_cols - strlen(buf_action_msg) - 9;
    va_list argp;
    va_start(argp, fmt);
    vsnprintf(buf_action_progress, term_cols, fmt, argp);
    va_end(argp);
    printf("\r%s%*s", buf_action_msg, msg_width, buf_action_progress);
    fflush(stdout);
    
}

/**
 * Set error message. This function sets the error message that will be printed
 * on the next call of \ref action_end() passing a result other than \c 0.
 * \param   *fmt    Error message (format string for printf)
 * \param   ...     Arguments for format specifiers in error message
 */
void action_errmsg(char *fmt, ...) {
    va_list argp;
    va_start(argp, fmt);
    vsnprintf(buf_error_msg, 255, fmt, argp);
    va_end(argp);
}

/**
 * Reports the result of an action. This function finalizes the output started
 * with \ref action_start() and possibly modified with \ref action_progress().
 * If \p status is 0, e.g. the action was successful, the string "[ ok ]" is
 * printed at the end of the current line. Otherwise, the string "[ !! ]" is
 * printed followed by the error message earlier set with \ref action_errmsg().
 * \param   status  The result of a previously performed operation, 0 means
 *                  success, other value means error.
 * \return          The value passed by \c status
 */
int action_end(int status) {
    if (strlen(buf_action_progress) == 0) {
        action_progress("");
    }
    if (status == 0) {
        printf("%9s\n", "[ ok ]");
    } else {
        printf("%9s\n\nERROR: %s\n\n", "[ !! ]", buf_error_msg);
    }
    return status;
}


