#include "loader.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <elf.h>
#include <endian.h>
#include <signal.h>
#ifdef CONFIG_HAVE_TERMCAP
#  include <termcap.h>
#endif
#include "../../../spartanmc/include/loader_cmd.h"

static volatile int abort_test = 0;

/**
 * Check if data is currently arriving. This function returns 0 if no data
 * arrives within an interval of \p timeout_ms milliseconds, -1 otherwise.
 * \returns     0 if no data has been receiving, -1 otherwise
 */
static int check_silence(int timeout_ms) {
    int dummy;
    return (uart_read(&dummy, 1, timeout_ms) == 0) ? -1 : 0;
}

/**
 * Waits for serial silence. This function waits until no data has been receive
 * on the serial line within an interval of \p timeout_ms milliseconds.
 * \param   timeout_ms  Required time with no data received 
 */
static void wait_silence(int timeout_ms) {
    while (check_silence(timeout_ms) != 0);
}

/**
 * Waits for certain message from target. This function expects the message \p
 * *fmt from the target. The message is interpreted as format string to \c
 * scanf(), and all additional arguments are filled with data matched against
 * format specifiers found in \c *fmt. The number of matched items is returned.
 * If no items have been matched (or no format specifiers were contained in \c
 * *fmt), 0 is returned if \c *fmt matches the ,essage received form the
 * target. This allows to check for static responses with no format specifiers.
 * On mismatch, -1 is returned.
 * \param   *fmt    Format string to match with message from target
 * \param   ...     Arguments filled with matched format specifiers from \c *fmt
 * \return          Number of items matched, -1 if static message did not match
 */
static int expect_response(char *fmt, ...) {
    char line[256];
    va_list argp;
    int res;
    while (uart_readln(line, sizeof(line), 1000) <= 0);
    va_start(argp, fmt);
    res = vsscanf(line, fmt, argp);
    va_end(argp);
    if ((res == 0) && (strcmp(fmt, line) != 0)) {
        action_progress("unexpected response");
        return -1;
    }
    action_progress("");
    return res;
}

/**
 * Send 18-bit value to the target. This function sends an 18-bit value \p data
 * to the target device using command \p cmd. Three bytes (24 bit) are used for
 * transmission. The lower 18 bit are filled with the data value, the upper 6
 * bit are filled with the command. The assembled value is returned in \p *tx
 * if not \c NULL. If \c cmd is not \ref LOADER_CMD_NOP, the same data is
 * expected to be sent back to verify commucation. If the received data does
 * not match the data originally sent, -1 is returned. The received 24-bit
 * value stored in \p *rx if not \c NULL.
 * \param   data    Data value
 * \param   cmd     Command
 * \param   *tx     Returns sent data packet, may be NULL
 * \param   *rx     Returns received data packet for verification, may be NULL
 * \returned        0 on success, -1 in case of error or data mismatch
 */
static int send18(uint32_t data, uint32_t cmd, uint32_t *tx, uint32_t *rx) {
    uint32_t verify;
    data &= MASK_18BIT;
    cmd &= 0x3F;
    data |=  (cmd << 18);
    uart_write(&data, 3);
    if (tx) {
        *tx = data;
    }
    if (cmd == LOADER_CMD_NOP) {
        return 0;
    }
    if (uart_read(&verify, 3, 1000) != 0) {
        action_progress("timeout");
        return -1;
    }
    verify &= 0xFFFFFF;
    if (rx) {
        *rx = verify;
    }
    if (data != verify) {
        action_progress("%02x != %02x", data, verify);
        return -1;
    }
    return 0;
}

/**
 * Wait for boot loader on device. This function waits until the boot loader is
 * executed on the target after system reset.
 * \return  0 on success, -1 if laoder version on target does not match current
 *          version
 */
static int wait_device() {
    char version[9];
    while (expect_response(VERSION_PATTERN, version) < 1);
    if (strncmp(VERSION_NUMBER, version, 9) != 0) {
        action_errmsg("Loader version check failed: device is %s, host requires %s", version, VERSION_NUMBER);
        return -1;
    }
    return 0;
}

/**
 * Perform serial initilization sequence. This function send a sequence of \ref
 * LOADER_SYNQSEQ_LEN bytes of value \ref LOADER_SYNC and expects the same
 * number of bytes of value \ref LOADER_OK sent back. 
 * \return  0 on success, -1 on error
 */
static int send_init() {
    unsigned char c = LOADER_SYNC;
    unsigned char reply;
    int i;
    for (i=0; i<LOADER_SYNQSEQ_LEN; i++) {
        uart_write(&c, 1);
        if (uart_read(&reply, 1, 1000) != 0) {
            action_errmsg("Timeout during initialization");
            return -1;
        }
        if (reply != LOADER_OK) {
            action_errmsg(
                    "Initialization sequence failed\n  expected :  %02x\n  received :  %02x",
                    LOADER_OK,
                    reply);
            return -1;
        }
    }
    return 0;
}

/**
 * Read current hardware design ID from ELF file. This function reads the
 * hardware design ID from the ELF file represented by \p *elf. The ID is
 * stored in symbol \c SBI_CORE_ID. The value is returned in \p *id.
 * \param   *elf    ELF file representing current firmware image
 * \param   *id     Returns the hardware ID found in ELF file
 * \return          0 on success, -1 in case of error
 */
static int read_hw_id(struct elf_file *elf, uint32_t *id) {
    Elf32_Sym *sym = elf_find_symbol(elf, "SBI_CORE_ID");

    action_start("reading design ID from ELF file");
    if (sym == NULL) {
        action_errmsg("Invalid ELF file: No hardware design ID");
        action_end(1);
        return -1;
    }

    *id = be32toh(sym->st_value) & MASK_18BIT;
    action_progress("SBI_CORE_ID = %05x", *id);
    action_end(0);

}

/**
 * Verify hardware design ID on target. This function receives a hardware
 * design ID from the target device and compares the value against the one
 * passed by \p id_elf. On match, 0 is returned, -1 otherwise.
 * \param   id_elf      Hardware ID to match
 * \return              0 on match, -1 on mismatch or in case of error
 */
static int verify_hw_id(uint32_t id_elf) {
    uint32_t id_device;
    action_start("checking design ID");
    if (uart_read((char*)&id_device, 3, 1000) != 0) {
        action_errmsg("Timeout while waiting for hardware design ID");
        return -1;
    }
    id_device &= MASK_18BIT;
    if (id_elf != id_device) {
        action_errmsg("Hardware implementation present on FPGA does not match firmware contained in ELF file.");
        action_progress("mismatch: %05x", id_device);
        action_end(-1);
        return -1;
    }
    action_progress("%05x", id_device);
    action_end(0);
    return 0;
}

/**
 * Transfer reads/writes data from/to target. This function transfers \p
 * num_bytes bytes between the target and the host. The mode of operation is
 * defined by \p cmd. 
 *
 * If \p cmd is \ref LOADER_CMD_WRITE_DATA, \p num_bytes bytes stored in \p
 * *buf are written to the target device memory starting at address \p addr. If
 * \p cmd is \ref LOADER_CMD_READ_DATA, \p num_bytes bytes are read from the
 * target device starting at address \p addr and compared with the bytes stored
 * in \p *buf.
 *
 * In case of failure, message \p *errmsg is printed. During transfer, the text
 * given by \p *progmsg is prepended to the progress showing the number of
 * transferred bytes. 
 * \param   addr        The start address of the memory area on the target
 * \param   *buf        Pointer to buffer containing data to transfer/verify
 * \param   num_bytes   Number of bytes to transfer/verify
 * \param   cmd         Transfer command to use. Can be \ref
 *                      LOADER_CMD_WRITE_DATA or \ref LOADER_CMD_READ_DATA
 * \param   *errmsg     Message printed in case of failure
 * \param   *progmsg    Text prepended to the progress output
 */
static int transfer_bytes(uint32_t addr, uint32_t *buf, uint32_t num_bytes, int
        cmd, char *errmsg, char *progmsg) {
    int i;
    uint32_t tx, rx;
    /* Round up number of words to avoid losing the last byte of an unaligned
     * section end. Data has been properly allocated and prepared for that
     * procdure by elf_get_section_data(). */
    uint32_t num_words = (num_bytes + 1) / 2;
    if (send18(addr, LOADER_CMD_SET_ADDR, NULL, NULL) != 0) {
        action_errmsg("Failed to set start address");
        return -1;
    }
    for (i=0; i<num_words; i++) {
        action_progress("%s 0x%05x (%5d / %d bytes)", progmsg, addr, i*2, num_bytes);
        if (send18(buf[i], cmd, &tx, &rx) != 0) {
            if ((tx >> 18) & 0x3F != (rx >> 18) & 0x3F) {
                action_errmsg("Communication error\n  sent cmd     :  %02x\n  received cmd :  %02x",
                    (tx >> 18) & 0x3F,
                    (rx >> 18) & 0x3F);
                return -1;
            }
            action_errmsg(
                    "%s at 0x%05x\n  expected :  %05x / %03x %03x\n  received :  %05x / %03x %03x",
                    errmsg,
                    addr,
                    tx & MASK_18BIT,
                    (tx >> 9) & 0x1FF,
                    (tx) & 0x1FF,
                    rx & MASK_18BIT,
                    (rx >> 9) & 0x1FF,
                    (rx) & 0x1FF);
            return -1;
        }
        addr += 2;
    }
    return 0;
}

/**
 * Initialize serial communication. This function waits until the loader is
 * executed on target and initializes serial communication afterwards.
 */
static int init() {
    if (check_silence(500) != 0) {
        action_start("device is sending data, waiting for silence");
        wait_silence(1500);
        action_end(0);
    }
    action_start("waiting for device (push reset)");
    if (action_end(wait_device()) != 0) {
        return -1;
    }
    action_start("initializing");
    if (action_end(send_init()) != 0) {
        return -1;
    }

    return 0;
}

/**
 * Check if section start or end address has changed. This function checks if
 * the position of section \p sec_name in the ELF file is the same as in the
 * memory of the target. Therefore it receives two address values from the
 * target device via serial line and compares them with the current addresses
 * found in the ELF file. On match, 0 is returned, -1 otherwise.
 * \param   *elf        ELF file representing current firmware image
 * \param   *sec_name   Name of section to check
 * \return              0 if section addresses match, -1 if not or an error
 *                      occured
 */
static int verify_section_addrs(struct elf_file *elf, char *sec_name) {
    int i;
    uint32_t addr[2];
    uint32_t sec_start;
    uint32_t sec_words;
    uint32_t sec_end;
    Elf32_Shdr *sec_hdr;
    for (i=0; i<2; i++) {
        if (uart_read(&addr[i], 3, 1000) != 0) {
            action_errmsg("Timeout while waiting for section addresses");
        }
        addr[i] &= MASK_18BIT;
    }
    sec_hdr = elf_find_section(elf, sec_name);
    if (sec_hdr == NULL) {
        action_errmsg("Invalid ELF file: No section %s", sec_name);
        return -1;
    }
    sec_start = be32toh(sec_hdr->sh_addr);
    sec_words = (be32toh(sec_hdr->sh_size) + (sizeof(uint32_t)/2)) / sizeof(uint32_t);
    sec_end = sec_start + sec_words * 2 - 1;
    if ((addr[0] != sec_start) || (addr[1] != sec_end)) {
        action_errmsg("Current startup/loader code does not match code present "
                "on device. Firmware update not possible.");
        return -1;
    }
    return 0;
}

/**
 * Check for valid section layout on target. This function verifies if the
 * sections containing loader code and data in the ELF file match their
 * counterparts present in memory of the target device.
 * \param   *elf    ELF file representing current firmware image
 * \return          0 on success, -1 otherwise
 */
static int verify_section_layout(struct elf_file *elf) {
    int i;
    uint32_t addr[4];
    action_start("checking section layout");
    if (verify_section_addrs(elf, ".loader.text") != 0) {
        action_end(1);
        return -1;
    }
    if (verify_section_addrs(elf, ".loader.data") != 0) {
        action_end(1);
        return -1;
    }
    action_end(0);
    return 0;
}

/**
 * Signal handler to abort serial test mode. This function is called as signal
 * handler for SIGINT and SIGTERM to exit from the infinite loop of the serial
 * communication test routine
 * \see \ref loader_test_uart();
 */
static void sig_handler(int signal) {
    abort_test = 1;
}

/**
 * Run test on serial communication. This function enters an infinite loop
 * sending the commands \ref LOADER_CMD_ECHO and \ref LOADER_CMD_NOP to the
 * target device. The number of errors is counted and continuously reported to
 * the user. The procedure can be aborted by \c SIGINT or \c SIGTERM (Ctrl+C).
 * \return      After receiving \c SIGINT or \c SIGTERM, 0 if no errors
 *              occured, -1 otherwise
 */
int loader_test_uart() {
    
    if (init() != 0) {
        return -1;
    }

    int cnt = 0;
    int err = 0;
    int data;
    int i;
    uint32_t tx;
    uint32_t rx;
    action_start("running serial communication test");
    
    signal(SIGINT, sig_handler);
    signal(SIGTERM, sig_handler);

    wait_silence(500);

    while(!abort_test) {
        for (i=0; i<64; i++) {
            data = random() & 0x3FFFF;
            send18(data, LOADER_CMD_NOP, NULL, NULL);
        }
        data = random() & 0x3FFFF;
        cnt++;   
        if (send18(data, LOADER_CMD_ECHO, &tx, &rx) != 0) {
            err++;
        }
        action_progress("%d words sent, %d errors", cnt, err);
    }
    
    signal(SIGINT, SIG_DFL);
    signal(SIGTERM, SIG_DFL);

    action_errmsg("Transfer of %d out of %d words failed", err, cnt);
    action_end(err);
}

/**
 * Transfer binary data between target and host. This function is the main
 * transfer routine. For each section tha is to be transferred, first the start
 * address of the section is sent, followed by the actual data. After a whole
 * section has been transferred its content is read back from the target to
 * verify the upload process. If \p verify_only is not zero, the transfer-part
 * is skipped and only data verification is performed.
 * \param   *elf            ELF file structure representing program to upload
 * \param   verify_only     Verify data in ELF file against memory contents of
 *                          target, no data is uploaded
 * \return                  0 on success, 1 otherwise
 */
static int transfer_sections(struct elf_file *elf, int verify_only) {
    int i;
    uint32_t rx;
    uint32_t tx;
    for (i=0; i<elf->elf_num_sections; i++) {
        Elf32_Shdr *sec_hdr = &elf->elf_section_hdrs[i];
        /* skip sections containing no program data */
        uint32_t type = be32toh(sec_hdr->sh_type);
        if (type != SHT_PROGBITS) {
            continue;
        }
        /* skip zero-sized sections */
        uint32_t size = be32toh(sec_hdr->sh_size);
        if (size == 0) {
            continue;
        }
        /* skip masked sections (see elf_select_sections(), options -s and -e) */
        if (elf->elf_masked_sections[i] != 0) {
            continue;
        }
        uint32_t addr = be32toh(sec_hdr->sh_addr)/2;
        uint32_t *buf = NULL;
        uint32_t num_bytes = size / (sizeof(uint32_t) / 2);
        char *name = elf_get_section_name(elf, i);
        buf = elf_get_section_data(elf, i);
        action_start("processing section %s", name);
        if (!verify_only) {
            if (transfer_bytes(addr, buf, num_bytes, LOADER_CMD_WRITE_DATA, "Transfer failed", "writing") != 0) {
                action_end(1);
                return -1;
            }
        }
        if (transfer_bytes(addr, buf, num_bytes, LOADER_CMD_READ_DATA, "Data mismatch", "verifying") != 0) {
            action_end(1);
            return -1;
        }
        action_progress("%d bytes @ 0x%05x", num_bytes, addr);
        action_end(0);
    }
    printf("\nOperation complete :)\n\n");
    return 0;
}

/**
 * Upload contents from ELF file. This function uploads sections from the
 * specified ELF file into the memory of the target device. The procedure is as
 * follows:
 *   - Read hardware ID from ELF file
 *   - Initialize serial communication 
 *   - Check if hardware ID matches hardware design on target
 *   - Check if loader- or startup-code has changed
 *   - Upload sections with respect to options -s and -e
 *
 * \param   *elf            ELF file structure representing program to upload
 * \param   verify_only     Verify data in ELF file against memory contents of
 *                          target, no data is uploaded
 * \return                  0 on success, -1 on failure
 */
int loader_upload_file(struct elf_file *elf, int verify_only) {

    uint32_t elf_id;
   
    if (read_hw_id(elf, &elf_id) != 0) {
        return -1;
    }

    if (init() != 0) {
        return -1;
    }

    if (verify_hw_id(elf_id) != 0) {
        return -1;
    }

    /*if (verify_section_layout(elf) != 0) {
        return -1;
    }*/

    return transfer_sections(elf, verify_only);

}


/**
 * Reset SpartanMC core
 *
 * \return                  0 on success, -1 on failure
 */
int loader_reset() {
    if (send18(0, LOADER_CMD_RESET, NULL, NULL) != 0) {
        action_errmsg("Failed to reset SpartanMC");
        return -1;
    }
    return 0;
}

