/* Read ELF file for SpartanMC and produce a textual dump of the
 * .text and .data sections in SPH format. (linear text output of 18-bit words)
 * The main purpose is to generate input files for the SpartanMC Simulator. */

#include <assert.h>
#include <endian.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "loader.h"

#define MACH_SPMC   0x5d2c

#define UPPER_BYTE(val) (((val) >> 9) & 0x1FF)
#define LOWER_BYTE(val) ((val) & 0x1FF)

// Macros for spreading 18 bits over two 16-bit bytes and vice versa.
#define _18TO32(val) (((val) & 0x1FF) | ((((val) >> 9) & 0x1FF) << 16))
#define _32TO18(val) (((val) & 0x1FF) | ((((val) >> 16) & 0x1FF) << 9))

/**
 * Align raw data of an ELF section. This function aligns the native data from
 * an ELF section pointed to by \p *buf according to section flags \p flags to
 * simplify further processing.
 * \param   *buf    Buffer containing raw section data
 * \param   size    Number of bytes in buffer
 * \param   flags   Flags from the corresponding section header
 */
static void align_words(uint32_t *buf, uint32_t size, uint32_t flags) {
    int i;
    for (i=0; i<size/sizeof(uint32_t); i++) {
        uint32_t val = be32toh(buf[i]);
        if (flags & SHF_EXECINSTR) {
            val &= MASK_18BIT;
        } else {
            val = _32TO18(val);
        }
        buf[i] = val;
    }
    /* handle last byte if section ends unaligned, this may only occur in data
     * sections, so we don't need to check for SHF_EXECINSTR */
    if (size % sizeof(uint32_t) != 0) {
        uint16_t val = be16toh(buf[i]);
        val = _32TO18(val);
        /* keep only high order byte (big endian) */
        val &= (0x1FF << 9);
        buf[i] = val;
    }
}

/**
 * Read all section headers. This function reads all section headers from the
 * ELF file represented by \p *elf and updates its internal data accordingly.
 * \param   *elf    ELF file to read sections from
 * \return          0 on success, -1 on failure    
 */
static int read_section_headers(struct elf_file *elf) {
    
    Elf32_Ehdr *elf_hdr = &elf->elf_header;
    Elf32_Shdr sec_hdr;

    uint32_t sec_table_off = be32toh(elf_hdr->e_shoff);
    uint32_t sec_ent_size = be16toh(elf_hdr->e_shentsize);

    elf->elf_num_sections = be16toh(elf_hdr->e_shnum);
    elf->elf_section_data = calloc(elf->elf_num_sections, sizeof(char*));

    assert(sec_ent_size == sizeof(sec_hdr));
   
    elf->elf_section_hdrs = calloc(elf->elf_num_sections, sec_ent_size);
    elf->elf_masked_sections = calloc(elf->elf_num_sections, sizeof(uint32_t));

    fseek(elf->elf_infile, sec_table_off, SEEK_SET);
    if (fread(elf->elf_section_hdrs, sec_ent_size, elf->elf_num_sections, elf->elf_infile) != elf->elf_num_sections) {
        fprintf(stderr, "failed to read section headers\n");
        return -1;
    }

    return 0;
}

/**
 * Read ELF section data. This function reads the data associated with section
 * number \p sec_idx of the ELF file represented by \p *elf into buffer \p
 * **buf. The buffer will be allocated if necessary.
 * \param   *elf        ELF file to get section data from
 * \param   sec_idx     Index number of section to read
 * \param   **buf       Buffer to store data, is allocated if required
 * \return              0 on success, -1 on failure
 */
static int read_section_data(struct elf_file *elf, uint32_t sec_idx, uint32_t **buf) {
    Elf32_Shdr *sec_hdr = &elf->elf_section_hdrs[sec_idx];
    uint32_t offs = be32toh(sec_hdr->sh_offset);
    uint32_t size = be32toh(sec_hdr->sh_size);
    uint32_t flags = be32toh(sec_hdr->sh_flags);
    uint32_t type = be32toh(sec_hdr->sh_type);

    fseek(elf->elf_infile, offs, SEEK_SET);
    if (*buf == NULL) {
        /* allocate whole words (sizeof(uint32_t)) for uniform handling of
         * aligned and unaligned section ends */
        *buf = malloc(((size + (sizeof(uint32_t) / 2)) / sizeof(uint32_t)) * sizeof(uint32_t));
    }
    if (fread(*buf, 1, size, elf->elf_infile) < size) {
        return -1;
    }
    if (type != SHT_PROGBITS) {
        return 0;
    }
    align_words(*buf, size, flags);
    return 0;
}

/**
 * Get ELF section data. This function returns a pointer to a buffer holding
 * the data associated with section number \p sec_idx of the ELF file
 * represented by \p *elf.
 * \param   *elf        ELF file to get section data from
 * \param   sec_idx     Index number of section to access
 * \return              Pointer to buffer holding section data, NULL on failure
 */
uint32_t *elf_get_section_data(struct elf_file *elf, uint32_t sec_idx) {
    if (elf->elf_section_data[sec_idx] == NULL) {
        if (read_section_data(elf, sec_idx, &elf->elf_section_data[sec_idx]) != 0) {
            fprintf(stderr, "failed to read section %d\n", sec_idx);
            return NULL;
        }
    }
    return elf->elf_section_data[sec_idx];
}

/**
 * Read header from ELF file. This function reads the ELF header from the file
 * represented by \p *elf and checks whether it is valid for the purpose of
 * this program.
 * \param   *elf        ELF file to process
 */
static int fetch_elf_header(struct elf_file *elf) {

    Elf32_Ehdr *hdr = &elf->elf_header;

    // read header
    int num_read = fread(hdr, 1, sizeof(*hdr), elf->elf_infile);
    if (num_read != sizeof(*hdr)) {
        fprintf(stderr, "Error: incomplete ELF header (%d bytes)\n", num_read);
        return -1;
    }

    // check magic
    if (strncmp(ELFMAG, (char*)hdr->e_ident, 4)) {
        fprintf(stderr, "Error: ELF magic mismatch\n");
        return -1;
    }

    // check class and machine
    if ((hdr->e_ident[EI_CLASS] != ELFCLASS32)
            || (be16toh(hdr->e_machine) != MACH_SPMC)) {
        fprintf(stderr, "Error: file not in elf32-spmc format\n");
        fprintf(stderr, "       (class: %#x, machine: %#x)\n",
                         hdr->e_ident[EI_CLASS], be16toh(hdr->e_machine));
        return -1;
    }

    return 0;
}

/**
 * Get section name from global string table.
 * \param   *elf        ELF file to process
 * \param   *sec_idx    Index of section to get name for
 */
char *elf_get_section_name(struct elf_file *elf, uint32_t sec_idx) {
    uint32_t strtab_idx = be16toh(elf->elf_header.e_shstrndx);
    char *strtab = (char*)elf_get_section_data(elf, strtab_idx);
    Elf32_Shdr *sec_hdr = &elf->elf_section_hdrs[sec_idx];
    return strtab + be32toh(sec_hdr->sh_name);
}

/**
 * Get symbol name from symbol string table.
 * \param   *elf        ELF file to process
 * \param   *sec_hdr    Section containing symbol \c *elf_sym
 * \param   *elf_sym    Symbol to get name for
 */
char *elf_get_symbol_name(struct elf_file *elf,  Elf32_Shdr *sec_hdr, Elf32_Sym *elf_sym) {
    uint32_t strtab_idx = be32toh(sec_hdr->sh_link);
    char *strtab = (char*)elf_get_section_data(elf, strtab_idx);
    return strtab + be32toh(elf_sym->st_name);
}

/**
 * Find symbol in ELF file. This function returns the symbol named \p *sym_name, if found.
 * \param   *elf        ELF file to process
 * \param   *sym_name   Name of symbol to look for
 * \return              Structure representing symbol \c *sym_name or NULL if
 *                      not found
 */
Elf32_Sym *elf_find_symbol(struct elf_file *elf, char *sym_name) {
    int i;
    for (i=0; i<elf->elf_num_sections; i++) {
        Elf32_Shdr *sec_hdr = &elf->elf_section_hdrs[i];
        uint32_t type = be32toh(sec_hdr->sh_type);
        if (type != SHT_SYMTAB) {
            continue;
        }
        uint32_t size = be32toh(sec_hdr->sh_size);
        uint32_t num_symbs = size / sizeof(Elf32_Sym);
        Elf32_Sym *elf_sym = (Elf32_Sym*)elf_get_section_data(elf, i);
        int j;
        char *name;
        for (j=0; j<num_symbs; j++) {
            name = elf_get_symbol_name(elf, sec_hdr, &elf_sym[j]);
            if (strcmp(name, sym_name) != 0) {
                continue;
            }
            return &elf_sym[j];
        }
    }
    return NULL;
}

/**
 * Find section in ELF file. This function returns the header for the section
 * named \p *sec_name, if found.
 * \param   *elf        ELF file to process
 * \param   *sec_name   Name of section to look for
 * \return              Section header of section named \c *sec_name or NULL if
 *                      not found
 */
Elf32_Shdr *elf_find_section(struct elf_file *elf, char *sec_name) {
    int i;
    char *curr_name;
    for (i=0; i<elf->elf_num_sections; i++) {
        curr_name = elf_get_section_name(elf, i);
        if (strcmp(curr_name, sec_name) == 0) {
            return &elf->elf_section_hdrs[i];
        }
    }
    return NULL;
}

/**
 * Checks if section is in list. This function checks whether section name \p
 * *sec_name is contained in the list \p **names_wanted. If \p exclude is not
 * 0, the function checks if the name is NOT in the list.
 * \param   *sec_name       Section name to check
 * \param   **names_wanted  List of section names to compare against
 * \param   exclude         Compare mode, see above
 * \return                  0 on positive result with respective to \c exclude
 */
static int section_wanted(char *sec_name, char **names_wanted, int exclude) {
    int i = 0;
    char *n;
    while ((n = names_wanted[i++])) {
        if (strcmp(sec_name, n) != 0 && strncmp(sec_name, ".debug", 6) != 0) {
            continue;
        }
        return !exclude;
    }
    return exclude;
}

/**
 * Select sections for processing. This function selects sections for forther
 * processing. If \p exclude is 0, \p **names specifies the names of all
 * section to select. Otherwise, all sections NOT contained in \c **names are
 * selected.
 * \param   *elf        ELF file to select sections from
 * \param   **names     Array of section names
 * \param   exclude     Selection mode, see above
 */
void elf_select_sections(struct elf_file *elf, char **names, int exclude) {
    int i;

    if (names == NULL) {
        return;
    }

    for (i=0; i<elf->elf_num_sections; i++) {
        Elf32_Shdr *sec_hdr = &elf->elf_section_hdrs[i];
        if (!section_wanted(elf_get_section_name(elf, i), names, exclude)) {
            elf->elf_masked_sections[i] = 1;
        } else {
            elf->elf_masked_sections[i] = 0;
        }
    }
}

/**
 * Open an ELF file. This function opens the ELF file denoted by \p *filename.
 * The structure \p *elf is initializing accordingly.
 * elf_open_file().
 * \param   *elf        The ELF structure representing the file opened
 * \param   *filename   Name of to file to open
 * \return              0 on success, -1 otherwise
 */
int elf_open_file(struct elf_file *elf, char *filename) {
    if ((elf->elf_infile = fopen(filename, "rb")) == NULL) {
        perror("Error opening input file");
        return EXIT_FILEIO;
    }

    if (fetch_elf_header(elf) != 0) {
        fprintf(stderr, "failed to read elf header\n");
        return EXIT_INVALIDELF;
    }

    if (read_section_headers(elf) != 0) {
        return EXIT_INVALIDELF;
    }

    return 0;
}

/**
 * Close ELF file. This function closes an ELF file previously opened with \ref
 * elf_open_file().
 * \param   *elf    The ELF file to close
 */
void elf_close_file(struct elf_file *elf) {
    int i;
    free(elf->elf_section_hdrs);
    for (i=0; i<elf->elf_num_sections; i++) {
        free(elf->elf_section_data[i]);
    }
    free(elf->elf_section_data);
    fclose(elf->elf_infile);
}

