#include "loader.h"

#include <getopt.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

struct cmdline_opts {
    int opt_dump;
    int opt_exclude;
    int opt_test;
    int opt_verify;
    char *opt_infile;
    char *opt_uart_device;
    char **opt_sections;
};

static void print_usage(const char *prog_name) {
    printf("Usage: %s [OPTIONS] FILE\n", prog_name);
    printf("Upload SpartanMC ELF file to processor.\n");
    printf("\n");
    printf(" -d, --device          uart device\n");
    printf(" -e, --exclude         changes option -s to specify sections NOT to be processed\n");
    printf(" -h, --headers         dump all section headers and exit\n");
    printf(" -s, --sections NAMES  specify sections to process (comma separated)\n");
    printf(" -t, --test            run serial communication test\n");
    printf(" -v, --verify          verify memory contents, don't upload any data\n");
    printf("     --help            display this help\n");
    printf(" -D, --dump            dump all sections and exit\n");
}

static int parse_cmdline_opts(struct cmdline_opts *opts, int argc, char **argv) {
    int opt;

    static struct option long_options[] = {
        { "headers",  no_argument      ,  NULL, 'h' },
        { "dump",     no_argument      ,  NULL, 'D' },
        { "device",   required_argument,  NULL, 'd' },
        { "test",     no_argument      ,  NULL, 't' },
        { "help",     no_argument      ,  NULL, 'H' },
        { "sections", required_argument,  NULL, 's' },
        { "exclude",  no_argument,        NULL, 'e' },
        { "verify",   no_argument,        NULL, 'v' },
        { NULL,       0,                  NULL,  0 }
    };

    while (1) {

        opt = getopt_long(argc, argv, "d:s:heDvt", long_options, NULL);
        if (opt == -1) {
            break;
        }
        switch (opt) {
            case 'H':
                print_usage(argv[0]);
                exit(EXIT_SUCCESS);
                break;

            case 'e':
                opts->opt_exclude = 1;
                break;

            case 's': ;
                // parse argument string into section names
                char *section_name = strtok(optarg, ",");
                char **sections = opts->opt_sections;
                int num_sections = 0;

                while (sections && sections[num_sections]) {
                    num_sections++;
                }

                do {
                    num_sections++;
                    sections = realloc(sections, num_sections * sizeof (char *));
                    sections[num_sections - 1] = section_name;
                }
                while ((section_name = strtok(NULL, ",")) != NULL);

                // terminate with null pointer
                sections = realloc(sections, (num_sections + 1) * sizeof (char *));
                sections[num_sections] = NULL;

                opts->opt_sections = sections;

                break;

            case 'h':
                opts->opt_dump = OPT_DUMP_HEADER;
                break;

            case 'D':
                opts->opt_dump = OPT_DUMP_CONTENT;
                break;

            case 'd':
                opts->opt_uart_device = optarg;
                break;

            case 't':
                opts->opt_test = 1;
                break;

            case 'v':
                opts->opt_verify = 1;
                break;

            default:
                exit(EXIT_OPTIONS);
        }

    }

    if (!opts->opt_dump && (opts->opt_uart_device == NULL)) {
        printf("Missing uart device (-d)!\n\n");
        print_usage(argv[0]);
        exit(EXIT_OPTIONS);
    }

    if (opts->opt_test) {
        return 0;
    }

    if (optind >= argc) {
        printf("Missing input file!\n\n");
        print_usage(argv[0]);
        exit(EXIT_OPTIONS);
    }

    opts->opt_infile = argv[optind];

    return 0;
}

static void dump_sections(struct elf_file *elf, int opt_dump) {
    int i,j;
    if (opt_dump == OPT_DUMP_HEADER) {
        printf("\n%-20s %-12s  %-13s %-8s\n\n", "SECTION", "ADDR (9 bit)", "SIZE (bytes)", "FLAGS");
    }
    for (i=0; i<elf->elf_num_sections; i++) {
        Elf32_Shdr *sec_hdr = &elf->elf_section_hdrs[i];
        uint32_t type = be32toh(sec_hdr->sh_type);
        if (type != SHT_PROGBITS) {
            continue;
        }
        uint32_t size = be32toh(sec_hdr->sh_size);
        if (size == 0) {
            continue;
        }
        uint32_t num_bytes = size / (sizeof(uint32_t) / 2);
        char *name = elf_get_section_name(elf, i);
        uint32_t addr = be32toh(sec_hdr->sh_addr);
        if (opt_dump == OPT_DUMP_HEADER) {
            uint32_t flags = be32toh(sec_hdr->sh_flags);
            printf("%-20s %08x      %08x      %08x\n", name, addr, num_bytes, flags);
        }
        if (opt_dump == OPT_DUMP_CONTENT) {
            uint32_t *buf = elf_get_section_data(elf, i);
            uint32_t flags = be32toh(sec_hdr->sh_flags);
            printf("\n%s\n", name);
            if (flags & SHF_EXECINSTR) {
                for (j=0; j<num_bytes/2; j++) {
                    printf("%5x:   %05x\n", addr + j*2, buf[j]);
                }
            } else {
                for (j=0; j<num_bytes/2; j++) {
                    printf("%5x:   %03x\n", addr + j*2, buf[j] & 0x1FF);
                    printf("%5x:   %03x\n", addr + j*2 + 1, (buf[j] >> 9) & 0x1FF);
                }
                if (num_bytes % 2 == 1) {
                    printf("%5x:   %03x\n", addr + j*2, buf[j] & 0x1FF);
                }
            }
        }
    }
}

int main(int argc, char *argv[]) {
    struct cmdline_opts opts = {0};
    struct elf_file elf = {0};
    int err = 0;

    printf("\nSpartanMC ELF Loader v20140908\n");

    action_init();

    parse_cmdline_opts(&opts, argc, argv);

    if (!opts.opt_test) {
        action_start("loading ELF file");

        if ((err = elf_open_file(&elf, opts.opt_infile)) != 0) {
            action_end(1);
            goto loader_done;
        }

        action_end(0);

        elf_select_sections(&elf, opts.opt_sections, opts.opt_exclude);

        if (opts.opt_dump) {
            dump_sections(&elf, opts.opt_dump);
            goto loader_done;
        }
    }
    
    if ((err = uart_device_open(opts.opt_uart_device)) != 0) {
        goto loader_done;
    }

    if (opts.opt_test) {
        err = loader_test_uart();
        goto loader_done;
    } else {
        err = loader_upload_file(&elf, opts.opt_verify);
    }
    
    loader_reset();

    uart_device_close();

loader_done:
    if (elf.elf_infile) {
        elf_close_file(&elf);
    }
    free(opts.opt_sections);
    action_finalize();
    exit(err);
}

