#include <pthread.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <termios.h>
#include <sys/stat.h>
#include <sys/time.h>

//#define UART_DEBUG

static int uart_fd;

/**
 * Wait for data. This function waits until data is available for reception on
 * the serial device or the timeout \p timeout_ms has passed.
 * \param   timeout_ms  Time in milliseconds to wait for data
 * \return              0 if data is available, -1 on error or if timeout
 *                      occured
 */
static int uart_select(int timeout_ms) {
    struct timeval t = {0};
    fd_set fds;
    FD_ZERO(&fds);
    FD_SET(uart_fd, &fds);
    while (timeout_ms >= 1000) {
        t.tv_sec++;
        timeout_ms -= 1000;
    }
    t.tv_usec = timeout_ms * 1000;
    if (select(uart_fd + 1, &fds, NULL, NULL, &t) == 0) {
        return -1;
    }
    return 0;
}

/**
 * Reads a line from serial device. This function reads a line from the serial
 * device. Data is stored into \p *buf until a newline or a carriage return is
 * received as termination character. That character is not appended to the
 * buffer. The maximum number of characters to receive is \p max_size. The
 * number of received characters is returned or -1 in case of error or timeout. 
 * \param   *buf        Buffer to store line
 * \param   max_size    Maximum number of characters to store in buffer
 * \param   timeout_ms  Return if no data is received within that interval (in
 *                      milliseconds)
 * \return              Number of chracters received, -1 in case of error or timeout.
 */
int uart_readln(unsigned char *buf, int max_size, int timeout_ms) {
    int cnt;
    int line_len = 0;
    *buf = '\0';
    while (max_size > 0) {
        if (uart_select(timeout_ms) != 0) {
            return -1;
        }
        cnt = read(uart_fd, buf, 1);
        if (cnt <= 0) {
            if ((errno != EAGAIN) && (errno != EINTR)) {
                return -1;
            }
            continue;
        }
        if (cnt == 0) {
            continue;
        }
#ifdef UART_DEBUG
        printf("uart read: %02x %c\n", *buf, *buf);
#endif
        if ((*buf == '\n') || (*buf == '\r')) {
            break;
        }
        max_size -= cnt;
        buf += cnt;
        line_len += cnt;
    }
    *buf = '\0';
    return line_len;
}

/**
 * Read bytes from serial device. This function reads \p cnt bytes from the
 * serial device and stores them into \p *buf. If no data is receive within \p
 * timeout_ms milliseconds, en error is returned.
 * \param   *buf        Buffer to store data
 * \param   cnt         Number of bytes to receive
 * \param   timeout_ms  Timeout while waiting for new data (in milliseconds)
 * \return              0 on success, -1 in case of error or timeout
 */
int uart_read(void *buf, int cnt, int timeout_ms) {
#ifdef UART_DEBUG
    int orig_cnt = cnt;
#endif
    unsigned char *ptr = buf;
    int rcnt;
    while (cnt > 0) {
        if (uart_select(timeout_ms) != 0) {
            return -1;
        }
        rcnt = read(uart_fd, ptr, cnt);
        if (rcnt >= 0) {
            cnt -= rcnt;
            ptr += rcnt;
            continue;
        }
        if ((errno != EAGAIN) && (errno != EINTR)) {
            return -1;
        }
    }
#ifdef UART_DEBUG
    ptr = buf;
    printf("uart read:");
    while (orig_cnt--) {
        printf(" %02x %c", *ptr, *ptr);
        ptr++;
    }
    printf("\n");
#endif
    return 0;
}

/**
 * Send bytes to serial device. This function sends \p cnt bytes to the serial
 * device stored in \p *buf.
 * \param   *buf        Buffer with data ti send
 * \param   cnt         Number of bytes to send
 * \return              0 on success, -1 in case of error
 */
int uart_write(void *buf, int cnt) {
    unsigned char *ptr = buf;
#ifdef UART_DEBUG
    int orig_cnt = cnt;
#endif
    int wcnt;
    while (cnt) {
        wcnt = write(uart_fd, buf, cnt);
        if (wcnt < 0) {
            if ((errno != EAGAIN) && (errno != EINTR)) {
                return -1;
            }
            continue;
        }
        ptr += wcnt;
        cnt -= wcnt;
    }
#ifdef UART_DEBUG
    ptr = buf;
    printf("uart write:");
    while (orig_cnt--) {
        printf(" %02x %c", *ptr, *ptr);
        ptr++;
    }
    printf("\n");
#endif
    fsync(uart_fd);
    tcdrain(uart_fd);
    return wcnt;
}

/**
 * Open serial device. This function opens the serial device \p *devname for
 * use with the function from this module.
 * \see \ref uart_device_close()
 * \param   *devname    Device to open (devnode)
 * \return              0 on success, -1 on error
 */
int uart_device_open(char *devname) {
    struct stat buf;
    uart_fd = open(devname, O_RDWR);
    if (stat(devname, &buf) != 0) {
        perror(devname);
        return -1;
    }
    if ((buf.st_mode & S_IFCHR) == 0) {
        fprintf(stderr, "%s: not a character device\n", devname);
        return -1;
    }
    if (uart_fd <= 0) {
        perror("failed to open uart device");
        return -1;
    }
    return 0;
}

/**
 * Close serial device. This function closes a serial device previously opened
 * with \ref uart_device_open().
 */
void uart_device_close() {
    close(uart_fd);
}
