#!/bin/bash

prepare_package_dirs() {
    src_dir=$work_dir/src/$package_name
    bin_dir=$work_dir/bin/$package_name
    package_dir=$work_dir/package/$package_name
    mkdir -p $package_dir $bin_dir $src_dir
}

git_get_version() {
    revision=$(cd ${src_dir} && git log --pretty=format:'%h' -n 1)
    date="$(cd ${src_dir} && git log --pretty=format:'%cD' HEAD -n 1)"
    date="$(date --date="${date}" +%Y%m%d)"
    package_version="$date-$revision"
}

git_checkout() {
    if (cd ${src_dir} && git status 2>/dev/null 1>/dev/null); then
        (cd ${src_dir} && git checkout ${git_branch})
        (cd ${src_dir} && git pull origin ${git_branch})
        (cd ${src_dir} && git clean -dxf)
    else
        mkdir -p ${src_dir}
        (git clone $git_url ${src_dir})
    fi
}

create_debian_package() {
    package_size=`du -hks $package_dir | grep -o '[0-9]\{1,\}'`
(cat << EOF
Package: $package_name
Version: $package_version
Section: embedded
Priority: extra
Architecture: $machine
Depends: $package_depends 
Installed-Size: $package_size
Maintainer: Markus Vogt <markus.vogt1@mailbox.tu.dresden.de>
Homepage: http://www.spartanmc.de
Description: $package_description
EOF
) > $package_dir/DEBIAN/control
    (cd package && fakeroot dpkg -b $package_dir ${package_filename}.deb)
}

create_tarball() {
    (cd $package_dir && tar -cvzf ../${package_filename}.tar.gz *)
}

gcc_build() {
    local bin_dir_abs=$(readlink -f ${bin_dir})
    rm -f $bin_dir/gcc/cc1
    (
        cd $src_dir
        ./build-spmc-gcc.sh ${bin_dir_abs} || true
    )
    if ! test -r $bin_dir/gcc/cc1; then
        echo "cc1 build failed"
        false
    fi
    strip $bin_dir/gcc/cc1
}

gcc_package_deb() {
    mkdir -p $package_dir/usr/bin $package_dir/DEBIAN
    cp $bin_dir/gcc/cc1 $package_dir/usr/bin/spmc-cc1
    create_debian_package
}

gcc_package_targz() {
    mkdir -p $package_dir/bin
    cp $bin_dir/gcc/cc1 $package_dir/bin/spmc-cc1
    create_tarball
}

binutils_build() {
    gcc_major=`expr \`gcc -dumpversion | cut -d. -f1\``
    gcc_minor=`expr \`gcc -dumpversion | cut -d. -f2\``
    (
        cd $src_dir/binutils-spmc
        export CFLAGS="-g -O2"
        if test $gcc_major -ge 4 && test $gcc_minor -ge 6; then
            export CFLAGS="$CFLAGS -Wno-error=unused-but-set-variable -Wno-error=unused-but-set-parameter" 
        fi
        ./configure --target=spmc --prefix=$bin_dir
        echo "MAKEINFO = :" >> Makefile
        make $MAKEOPTS
        make install
	cd $src_dir
        make tools
        mv $src_dir/install/bin/elf2sph $bin_dir/bin/
        rm -rf $bin_dir/spmc
        strip $bin_dir/bin/*
    )
}

binutils_package_deb() {
    mkdir -p $package_dir/usr $package_dir/DEBIAN
    cp -r $bin_dir/* $package_dir/usr/
    create_debian_package
}

binutils_package_targz() {
    mkdir -p $package_dir
    cp -r $bin_dir/* $package_dir
    create_tarball
}

build_package() {
    prepare_package_dirs
    git_checkout
    git_get_version
    package_filename=${package_name}_${package_version}_${build_system}_${machine}
    ${package_prefix}_build
    rm -rf $package_dir/*
    ${package_prefix}_package_${package_type}
}

main() {
    work_dir=$(readlink -f $1)
    if [ -z "${work_dir}" ]; then
        printf "usage: $0 WORKDIR\n"
        return
    fi
    if [ ! -d "${work_dir}" ]; then
        printf "not existing or not a directory: %s\n" "${work_dir}"
        return
    fi
    if ! dist=`grep -io 'Ubuntu\|Fedora\|CentOS\|Arch' /etc/issue`; then
        printf "Unsupported Linux distribution\n"
        return
    fi
    machine=`uname -m`
#    if test $machine = "x86_64"; then
#        machine="amd64"
#    else
#        machine="i386"
#    fi

### always build tar.gz packages

#    if test -z "$dist"; then
#        echo "Package build for your Linux disitribution is not (yet) supported (currently Fedora and Ubuntu are supported"
#        exit 1
#    fi
#
    case $dist in
        "Ubuntu")
            package_type="targz"
            ubuntu_major=`lsb_release -rs | cut -d. -f1`
            ubuntu_minor=`lsb_release -rs | cut -d. -f2`
            build_system="ubuntu_$ubuntu_major-$ubuntu_minor";;
        "Fedora") 
            package_type="targz"
            build_system="fedora_`rpm -q --queryformat '%{VERSION}' fedora-release`"
            ;;
        "CentOS")
            package_type="targz"
            build_system="centos_`rpm -q --queryformat '%{VERSION}' centos-release`"
            ;;
        "Arch")
            package_type="targz"
            build_system="arch"
            ;;
        *)
            printf "Unsupported Linux distribution: %s\n" "${dist}"
            return
    esac

    package_prefix="gcc"
    package_name="spmc-gcc"
    git_url="ssh://git@www.mr.inf.tu-dresden.de/global/spartanmc-gcc"
    git_branch="master"
#    package_description="Port of GCC for SpartanMC"
#    package_depends="libc6 (>=2.12.1), libmpfr4 (>=3.0.0)"
#    if test "$dist" = "Ubuntu" && test $ubuntu_major -ge 11 && test $ubuntu_minor -ge 10; then
#        package_depends="$package_depends, libgmp10 (>=2.5.0.1)"
#    else
#        package_depends="$package_depends, libgmp3c2 (>=2:4.3.2)"
#    fi
    build_package 

    package_prefix="binutils"
    package_name="spmc-binutils"
    git_url="ssh://git@www.mr.inf.tu-dresden.de/global/spartanmc-binutils"
    git_branch="master"
#    package_description="Port of GNU Binutils for SpartanMC"
#    package_depends="libc6 (>=2.12.1), zlib1g (>=1:1.2.3.4)"
    build_package 

}

set -e
main $*
