#-------------------------------------------------------------------
# Gets the name of the driving cell for the net with the given name.
# parameter 1: The name of the net.
#-------------------------------------------------------------------
proc get_driving_cell {argNetName} {
	# get pins connected to the net
	set pins [get_pins -leaf -of_objects [get_nets $argNetName]]
	
	# find driving pin
	foreach pin $pins {
		if {[get_property DIRECTION $pin] == "OUT"} {
			set cellName [get_property PARENT_CELL $pin]
            set cellGroup [get_property PRIMITIVE_GROUP [get_cells -hierarchical $cellName]]
            if {$cellGroup == {FLOP_LATCH}} {
                return $cellName
            } else {
                return ""
            }
        }
	}
};

#-------------------------------------------------------------------
# Gets the name of the cell which is placed on the specified latch position.
# parameter 1: The name of the slice where the latch resides.
# parameter 2: The letter defining the latch within the slice.
#-------------------------------------------------------------------
proc get_latch_by_position {argSlice argLetter} {
    switch $argLetter {
        A {
            return [get_cells -hierarchical -filter "LOC == $argSlice && (BEL == SLICEM.AFF || BEL == SLICEL.AFF)"]
        }
        B {
            return [get_cells -hierarchical -filter "LOC == $argSlice && (BEL == SLICEM.BFF || BEL == SLICEL.BFF)"]
        }
        C {
            return [get_cells -hierarchical -filter "LOC == $argSlice && (BEL == SLICEM.CFF || BEL == SLICEL.CFF)"]
        }
        D {
            return [get_cells -hierarchical -filter "LOC == $argSlice && (BEL == SLICEM.DFF || BEL == SLICEL.DFF)"]
        }
    }
}

#-------------------------------------------------------------------
# Modifies a line of the logic location file.
# parameter 1: The line to modify.
#-------------------------------------------------------------------
proc modify_llfile_line {argLine} {
    
    set netStartIndex [string first "Net=" $argLine]
    
    # no net name in this line
    if {$netStartIndex == -1} {
        set splitLine [regexp -all -inline {\S+} $argLine]
        set block [lindex $splitLine 4]
        set bit [lindex $splitLine 5]
        
        # 36k RAM block -> copy only first bit
        if {[string first "RAMB36" $block] != -1} {
            if {$bit == "Ram=B:BIT0"} {
                return $argLine
            } else {
                return ""
            }
        }
        
        # 18k RAM block -> copy only first bit
        if {[string first "RAMB18" $block] != -1} {
            if {$bit == "Ram=B:BIT0" || $bit == "Ram=B:BIT1"} {
                return $argLine
            } else {
                return ""
            }
        }
        
        # some other line -> copy
        return $argLine
        
    # net name in this line -> replace net name by name of driving latch
    } else {
        set netStartIndex [expr {$netStartIndex + 4}]
        set netStopIndex [string length $argLine]
        #set latchStartIndex [expr {[string first "Latch=" $argLine] + 6}]
        #set sliceStartIndex [expr {[string first "Block=" $argLine] + 6}]
        #set sliceStopIndex [string first " " $argLine $sliceStartIndex]
        set netName [string range $argLine $netStartIndex $netStopIndex]
        set cellName [get_driving_cell $netName]
        if {$cellName == ""} {
            return ""
        }
        #set cellName [get_latch_by_position [string range $argLine $sliceStartIndex $sliceStopIndex] [string index $argLine $latchStartIndex]]
        return [string replace $argLine $netStartIndex $netStopIndex $cellName]
    }
}

#-------------------------------------------------------------------
# Modifies the logic location file by replacing the net names by
# the names of the driving cells.
# parameter 1: path of the logic location file output
#-------------------------------------------------------------------
proc modify_llfile {argLlOutFileName} {
    
	# open logic location file (input)
	set llFileName [get_property TOP [current_design]]
	append llFileName ".ll"
	set fileInput [open $llFileName r]

	# open logic location file (output)
	set fileOutput [open $argLlOutFileName w 0600]

	# read input file line by line
	while {-1 != [gets $fileInput lineInput]} {
		
		# skip all lines which do not begin with "Bit"
		if {0 == [string first "Bit" $lineInput]} {
			set lineOutput [modify_llfile_line $lineInput]
            # skip invalid lines
            if {$lineOutput != ""} {
                puts $fileOutput $lineOutput
            }
		}
	}

	#close files
	close $fileInput
	close $fileOutput

	# delete input file
	file delete $llFileName
}

#-------------------------------------------------------------------
# Creates the file with placing information of RAM elements.
# parameter 1: path of the placement file output
#-------------------------------------------------------------------
proc create_locfile {argLocOutFileName} {

	# open file
	set fileOutput [open $argLocOutFileName w 0600]	

	# write header (just as information)
	puts $fileOutput "; Type Block BEL Name Width(BRAM) Start_Addr(BRAM) Stop_Addr(BRAM) Start_Slice(BRAM) Start_Ext_Slice(BRAM)"

	# get all cells which use LUTs as RAM (includes shift registers)
	set lut_mems [get_cells -hierarchical -filter {PRIMITIVE_GROUP == DMEM}]

	# write their placing information to the file
	foreach element $lut_mems {
		puts -nonewline $fileOutput [get_property PRIMITIVE_TYPE $element]
		puts -nonewline $fileOutput " "
		puts -nonewline $fileOutput [get_property LOC $element]
		puts -nonewline $fileOutput " "
		puts -nonewline $fileOutput [get_property BEL $element]
		puts -nonewline $fileOutput " "
		puts $fileOutput [get_property NAME $element]
	}

	# get all block RAM cells
	set block_mems [get_cells -hierarchical -filter {PRIMITIVE_GROUP == BMEM}]

	# write their placing information to the file
	foreach element $block_mems {
		puts -nonewline $fileOutput [get_property PRIMITIVE_TYPE $element]
		puts -nonewline $fileOutput " "
		puts -nonewline $fileOutput [get_property LOC $element]
		puts -nonewline $fileOutput " "
		puts -nonewline $fileOutput [get_property BEL $element]
		puts -nonewline $fileOutput " "
		puts -nonewline $fileOutput [get_property NAME $element]
		puts -nonewline $fileOutput " "
        set width [get_property READ_WIDTH_A $element]
        if {$width == ""} {
            set width [get_property READ_WIDTH_B $element]
        }
        if {$width == ""} {
            set width [get_property DATA_WIDTH $element]
        }
        puts -nonewline $fileOutput $width
        puts -nonewline $fileOutput " "
        set startAddr [get_property bram_addr_begin $element]
        if {$startAddr == ""} {
            set startAddr 0
        }
        puts -nonewline $fileOutput $startAddr
        puts -nonewline $fileOutput " "
		set stopAddr [get_property bram_addr_end $element]
        if {$stopAddr == ""} {
            set type [get_property PRIMITIVE_TYPE $element]
            if {$type == "BMEM.bram.RAMB36E1"} {
                set stopAddr [expr {36864/$width}]
            } else {
                set stopAddr [expr {18432/$width}]
            }
        }
        puts -nonewline $fileOutput $stopAddr
        puts -nonewline $fileOutput " "
        set startSlice [get_property bram_slice_begin $element]
        if {$startSlice == ""} {
            set startSlice 0;
        }
        puts -nonewline $fileOutput $startSlice
        puts -nonewline $fileOutput " "
        puts $fileOutput [get_property bram_ext_slice_begin $element]
	}

	# close file
	close $fileOutput
}
source $ensureCheckpointScript
create_locfile $primloc
