#include "filter.h"
#include "buffer.h"

static int memcmp(const void *d1, const void *d2, int len)
{
	const unsigned char *c1 = d1;
	const unsigned char *c2 = d2;

	while(len) {
		if(*c1 != *c2) {
			return *c1 - *c2;
		}

		c1++;
		c2++;
		len--;
	}

	return 0;
}

static const int rule_length[] = {
		2, // PROTOCOL
		2, // PORT_TCP
		2, // PORT_UDP
		4, // PORT_IP_TCP
		4, // PORT_IP_UDP
		10, // PORT_IP6_TCP
		10, // PORT_IP6_UDP
		3, // IP_DEST
		3, // IP_SRC
		9, // IP6_DEST
		9, // IP6_SRC
		5, // IPRANGE_DEST
		5, // IPRANGE_SRC
};

static const struct firewall_conf config = {
	DENY_ALL,
	26,
	{
		RULE_PORT_IP_TCP(10,42,0,1,80),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,1,80),
		RULE_PORT_IP_TCP(10,42,0,1,443),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,1,443),
		RULE_PORT_IP_TCP(10,42,0,2,110),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,2,110),
		RULE_PORT_IP_TCP(10,42,0,2,995),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,2,995),
		RULE_PORT_IP_TCP(10,42,0,2,143),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,2,143),
		RULE_PORT_IP_TCP(10,42,0,2,993),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,2,993),
		RULE_PORT_IP_TCP(10,42,0,2,25),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,2,25),
		RULE_PORT_IP_TCP(10,42,0,2,465),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,2,465),
		RULE_PORT_IP_TCP(10,42,0,2,587),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,2,587),
		RULE_PORT_IP_TCP(10,42,0,3,22),
		RULE_PORT_IP6_TCP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,3,22),
		RULE_PORT_IP_UDP(10,42,0,3,20),
		RULE_PORT_IP6_UDP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,3,20),
		RULE_PORT_IP_UDP(10,42,0,3,21),
		RULE_PORT_IP6_UDP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,3,21),
		RULE_PORT_IP_UDP(10,42,0,3,53),
		RULE_PORT_IP6_UDP(0xfe,0x80,0,0,0,0,0,0,0,0,0,0,10,42,0,3,53),
	}
};

#define ETHERTYPE_ARP   0x0806
#define ETHERTYPE_IPV4  0x0800
#define ETHERTYPE_IPV6  0x86DD

#define IP_ICMP         0x01
#define IP_TCP          0x06
#define IP_UDP          0x11

#define ETHERTYPE(d) (d)[OFF_ETHERTYPE]
#define PORT(d) (ETHERTYPE((d))==ETHERTYPE_IPV4?(d)[OFF_IPV4_PORT]:ETHERTYPE((d))==ETHERTYPE_IPV6?(d)[OFF_IPV6_PORT]:-1)
#define PROTO(d) (ETHERTYPE((d))==ETHERTYPE_IPV4?(d)[OFF_IPV4_PROTO]:ETHERTYPE((d))==ETHERTYPE_IPV6?((d)[OFF_IPV6_PROTO]>>8):-1)
#define OFF_ETHERTYPE    6
#define OFF_IPV4_PROTO  11
#define OFF_IPV6_PROTO  10
#define OFF_IPV4_PORT   18
#define OFF_IPV6_PORT   28
#define OFF_IPV4_SRC    13
#define OFF_IPV4_DST    15
#define LEN_IPV4_ADDR    4
#define OFF_IPV6_SRC    11
#define OFF_IPV6_DST    19
#define LEN_IPV6_ADDR   16

static int check_protocol(unsigned int *data, const unsigned int *rule)
{
	unsigned int protocol = rule[0];
	
	switch(protocol) {
		case IPV4:
			return ETHERTYPE(data) == ETHERTYPE_IPV4;
		case IPV6:
			return ETHERTYPE(data) == ETHERTYPE_IPV6;
		case ARP:
			return ETHERTYPE(data) == ETHERTYPE_ARP;
		case ICMP:
			return (ETHERTYPE(data) == ETHERTYPE_IPV4 && PROTO(data) == IP_ICMP)
					|| (ETHERTYPE(data) == ETHERTYPE_IPV6 && PROTO(data) == IP_ICMP);
	}

	return 0;
}

static int check_destport(unsigned int *data, const unsigned int *rule, unsigned int proto)
{
	if(ETHERTYPE(data) == ETHERTYPE_IPV4) {
		return PROTO(data) == proto && PORT(data) == rule[0];
	} else if(ETHERTYPE(data) == ETHERTYPE_IPV6) {
		return PROTO(data) == proto && PORT(data) == rule[0];
	}

	return 0;
}

static int check_destport_addr(unsigned int *data, const unsigned int *rule, unsigned int proto)
{
	return ETHERTYPE(data) == ETHERTYPE_IPV4
			&& PROTO(data) == proto
			&& PORT(data) == rule[2]
			&& memcmp(data + OFF_IPV4_DST, rule, LEN_IPV4_ADDR) == 0;
}

static int check_destport_addr6(unsigned int *data, const unsigned int *rule, unsigned int proto)
{
	return ETHERTYPE(data) == ETHERTYPE_IPV6
			&& PROTO(data) == proto
			&& PORT(data) == rule[8]
			&& memcmp(data + OFF_IPV6_DST, rule, LEN_IPV6_ADDR) == 0;
}

static int check_ipdest(unsigned int *data, const unsigned int *rule)
{
	return ETHERTYPE(data) == ETHERTYPE_IPV4
			&& memcmp(data + OFF_IPV4_DST, rule, LEN_IPV4_ADDR) == 0;
}

static int check_ip6dest(unsigned int *data, const unsigned int *rule)
{
	return ETHERTYPE(data) == ETHERTYPE_IPV6
			&& memcmp(data + OFF_IPV6_DST, rule, LEN_IPV6_ADDR) == 0;
}

static int check_ipsrc(unsigned int *data, const unsigned int *rule)
{
	return ETHERTYPE(data) == ETHERTYPE_IPV4
			&& memcmp(data + OFF_IPV4_SRC, rule, LEN_IPV4_ADDR) == 0;
}

static int check_ip6src(unsigned int *data, const unsigned int *rule)
{
	return ETHERTYPE(data) == ETHERTYPE_IPV6
			&& memcmp(data + OFF_IPV6_SRC, rule, LEN_IPV6_ADDR) == 0;
}

static int check_iprangedest(unsigned int *data, const unsigned int *rule)
{
	unsigned char ruleaddrmin1, ruleaddrmin2, ruleaddrmin3, ruleaddrmin4;
	unsigned char ruleaddrmax1, ruleaddrmax2, ruleaddrmax3, ruleaddrmax4;
	unsigned char addr1, addr2, addr3, addr4;

	ruleaddrmin1 = rule[0] >> 8;
	ruleaddrmin2 = rule[0] & 0xff;
	ruleaddrmin3 = rule[1] >> 8;
	ruleaddrmin4 = rule[1] & 0xff;

	ruleaddrmax1 = rule[2] >> 8;
	ruleaddrmax2 = rule[2] & 0xff;
	ruleaddrmax3 = rule[3] >> 8;
	ruleaddrmax4 = rule[3] & 0xff;

	addr1 = data[15] >> 8;
	addr2 = data[15] & 0xff;
	addr3 = data[16] >> 8;
	addr4 = data[16] & 0xff;

	if(ETHERTYPE(data) != ETHERTYPE_IPV4) {
		return 0;
	}

	return addr1 >= ruleaddrmin1 && addr1 <= ruleaddrmax1 &&
	       addr2 >= ruleaddrmin2 && addr2 <= ruleaddrmax2 &&
	       addr3 >= ruleaddrmin3 && addr3 <= ruleaddrmax3 &&
	       addr4 >= ruleaddrmin4 && addr4 <= ruleaddrmax4;
}

static int check_iprangesrc(unsigned int *data, const unsigned int *rule)
{
	unsigned char ruleaddrmin1, ruleaddrmin2, ruleaddrmin3, ruleaddrmin4;
	unsigned char ruleaddrmax1, ruleaddrmax2, ruleaddrmax3, ruleaddrmax4;
	unsigned char addr1, addr2, addr3, addr4;

	ruleaddrmin1 = rule[0] >> 8;
	ruleaddrmin2 = rule[0] & 0xff;
	ruleaddrmin3 = rule[1] >> 8;
	ruleaddrmin4 = rule[1] & 0xff;

	ruleaddrmax1 = rule[2] >> 8;
	ruleaddrmax2 = rule[2] & 0xff;
	ruleaddrmax3 = rule[3] >> 8;
	ruleaddrmax4 = rule[3] & 0xff;

	addr1 = data[13] >> 8;
	addr2 = data[13] & 0xff;
	addr3 = data[14] >> 8;
	addr4 = data[14] & 0xff;

	if(ETHERTYPE(data) != ETHERTYPE_IPV4) {
		return 0;
	}

	return addr1 >= ruleaddrmin1 && addr1 <= ruleaddrmax1 &&
	       addr2 >= ruleaddrmin2 && addr2 <= ruleaddrmax2 &&
	       addr3 >= ruleaddrmin3 && addr3 <= ruleaddrmax3 &&
	       addr4 >= ruleaddrmin4 && addr4 <= ruleaddrmax4;
}

void filter(int buf_index)
{
	int i, matching_rules = 0;
	unsigned int headers[32], *data;
	
	data = buffer_getdata(buf_index);
	
	// Copy headers from shared memory to local memory. Avoids memory bottleneck with multiple cores.
	for(i = 0; i < 32; i++) {
		headers[i] = data[i];
	}
	
	// Apply all rules to the packet.
	const int *ruleptr = config.rules;
	for(i = 0; i < config.num_rules; i++) {
		int ruletype;
		const unsigned int *ruledata;
		
		ruletype = *ruleptr;
		ruledata = ruleptr + 1;

		ruleptr += rule_length[ruletype];

		switch(ruletype) {
			case PROTOCOL:
				matching_rules += check_protocol(headers, ruledata);
			break;
			case PORT_TCP:
				matching_rules += check_destport(headers, ruledata, IP_TCP);
			break;
			case PORT_UDP:
				matching_rules += check_destport(headers, ruledata, IP_UDP);
			break;
			case PORT_IP_TCP:
				matching_rules += check_destport_addr(headers, ruledata, IP_TCP);
			break;
			case PORT_IP_UDP:
				matching_rules += check_destport_addr(headers, ruledata, IP_UDP);
			break;
			case PORT_IP6_TCP:
				matching_rules += check_destport_addr6(headers, ruledata, IP_TCP);
			break;
			case PORT_IP6_UDP:
				matching_rules += check_destport_addr6(headers, ruledata, IP_UDP);
			break;
			case IP_DEST:
				matching_rules += check_ipdest(headers, ruledata);
			break;
			case IP_SRC:
				matching_rules += check_ipsrc(headers, ruledata);
			break;
			case IP6_DEST:
				matching_rules += check_ip6dest(headers, ruledata);
			break;
			case IP6_SRC:
				matching_rules += check_ip6src(headers, ruledata);
			break;
			case IPRANGE_DEST:
				matching_rules += check_iprangedest(headers, ruledata);
			break;
			case IPRANGE_SRC:
				matching_rules += check_iprangesrc(headers, ruledata);
			break;
		}

		if(matching_rules > 0) {
			break;
		}
	}

	if(config.type == DENY_ALL) {
		if(matching_rules > 0) {
			buffer_setstate(buf_index, BUF_ALLOWED);
		} else {
			buffer_setstate(buf_index, BUF_UNUSED);
		}
	} else if(config.type == ALLOW_ALL) {
		if(matching_rules > 0) {
			buffer_setstate(buf_index, BUF_UNUSED);
		} else {
			buffer_setstate(buf_index, BUF_ALLOWED);
		}
	}

	return;
}
