#include <stdio.h>
#include <ethlib.h>
#include <peripherals.h>
#include <memdump.h>

#include "network.h"
#include "util.h"

#define RX_DMA_SIZE 1024

int counter;

void network_init(const struct netdev *dev)
{
	unsigned int regval;
	
	/* Initialize DMA buffers. */
	if(dev->rx) {
		ethernet_rx_init(dev->rx, dev->rx_dma);
		dev->rx->control |= 1; // Receive all frames.
	}
	
	/* Configure PHY. */
	// Wait for PHY to finish reset.
	do {
		mdio_read(dev->mdio, dev->phy_addr, 0, &regval);
	} while(regval & 0x8000);
	
	//printf("PHY%d released from reset.\n", dev->id);
	
	mdio_write(dev->mdio, dev->phy_addr, 0x09, 0x00); // Disable gigabit
	regval = (0 << 6) | (1 << 13) | // 100 Mbit speed
	         (1 << 12) | // auto-negotiation.
	         (1 << 9) | // Trigger autoneg
	         (1 << 8); // Full duplex mode.
	mdio_write(dev->mdio, dev->phy_addr, 0, regval);
	mdio_write(dev->mdio, dev->phy_addr, 0x1F, 1 << 14); // Restart software

	return;
}

void network_wait_ready(const struct netdev *dev)
{
	unsigned int regval;
	
	mdio_read(dev->mdio, dev->phy_addr, 1, &regval);
	if(!(regval & 0x04)) {
		//printf("No link on eth%d, waiting for link...\n", dev->id);
		
		do {
			mdio_read(dev->mdio, dev->phy_addr, 1, &regval);
		} while(!(regval & 0x04));
		
		//printf("Link established on eth%d!\n", dev->id);
	}
	
	return;
}

void read_from_ringbuffer(unsigned int *rbuf_start, unsigned int *src, unsigned int *dest, unsigned int len)
{
	// Calculate limit of ringbuffer.
	unsigned int *rbuf_end = rbuf_start + RX_DMA_SIZE;
	unsigned int space_left = (rbuf_end - src);
	unsigned int limit_first_loop = space_left < len ? space_left : len;

	unsigned int i;

	// Copy packet, there might be data left at the start of the ringbuffer after this.
	for(i = 0; i < limit_first_loop; i++) {
		*dest = *src;
		dest++;
		src++;
	}

	// Wrap around ringbuffer.
	src = rbuf_start;
	for(; i < len; i++) {
		*dest = *src;
		dest++;
		src++;
	}

	return;
}

void fast_read_from_ringbuffer(unsigned int *rbuf_start, unsigned int *src, unsigned int *dest, unsigned int len)
{
	/*
	 * r4: rbuf_start
	 * r5: src
	 * r6: dest
	 * r7: len
	 * r8: loop limit
	 * r9: data buffer.
	 */

	__asm__ volatile(
			// Copy until src is 8 word aligned.
			"check_alignmentf:\n"
			"mov r9, r5\n"
			"andi r9, 0x0f\n"
			"beqz r9, alignedf\n"
			"l18 r9, 0(r5)\n"
			"s18 0(r6), r9\n"
			"addi r5, 2\n"
			"addi r6, 2\n"
			"addi r7, -1\n"
			"j check_alignmentf\n"

			// Calculate loop limit to avoid overshooting buffer
			"alignedf:\n"
			"movi r8, 1\n"
			"slli r8, 11\n"
			"add r8, r4\n"
			"subu r8, r5\n"
			"srli r8, 1\n"
			"sleu r8, r7\n"
			"bnezc check_limitf\n"
			"nop\n"
			"mov r8, r7\n"
			"j check_limitf\n"

			"first_loopf:\n"
			"l18 r9, 0(r5)\n"
			"s18 0(r6), r9\n"
			"l18 r9, 2(r5)\n"
			"s18 2(r6), r9\n"
			"l18 r9, 4(r5)\n"
			"s18 4(r6), r9\n"
			"l18 r9, 6(r5)\n"
			"s18 6(r6), r9\n"
			"l18 r9, 8(r5)\n"
			"s18 8(r6), r9\n"
			"l18 r9, 10(r5)\n"
			"s18 10(r6), r9\n"
			"l18 r9, 12(r5)\n"
			"s18 12(r6), r9\n"
			"l18 r9, 14(r5)\n"
			"s18 14(r6), r9\n"

			"addi r5, 16\n"
			"addi r6, 16\n"
			"addi r8, -8\n"
			"addi r7, -8\n"
			"check_limitf:\n"
			"slei r8, 0\n"
			"beqzc first_loopf\n"
			"slei r7, 0\n"
			"bnezc endf\n"
			"mov r8, r7\n"
			"mov r5, r4\n"
			"j first_loopf\n"
			"nop\n"

			"endf:\n" ::: "r4", "r5", "r6", "r7", "r8", "r9"
	);

	return;
}

void fast_copy_to_ringbuffer(unsigned int *rbuf_start, unsigned int *src, unsigned int *dest, unsigned int len)
{
	/*
	 * r4: rbuf_start
	 * r5: src
	 * r6: dest
	 * r7: len
	 * r8: loop limit
	 * r9: data buffer.
	 */

	__asm__ volatile(
			// Copy until dest is 8 word aligned.
			"check_alignmentt:\n"
			"mov r9, r6\n"
			"andi r9, 0x0f\n"
			"beqz r9, alignedt\n"
			"l18 r9, 0(r5)\n"
			"s18 0(r6), r9\n"
			"addi r5, 2\n"
			"addi r6, 2\n"
			"addi r7, -1\n"
			"j check_alignmentt\n"

			// Calculate loop limit to avoid overshooting buffer
			"alignedt:\n"
			"movi r8, 1\n"
			"slli r8, 11\n"
			"add r8, r4\n"
			"subu r8, r6\n"
			"srli r8, 1\n"
			"sleu r8, r7\n"
			"bnezc check_limitt\n"
			"nop\n"
			"mov r8, r7\n"
			"j check_limitt\n"

			"first_loopt:\n"
			"l18 r9, 0(r5)\n"
			"s18 0(r6), r9\n"
			"l18 r9, 2(r5)\n"
			"s18 2(r6), r9\n"
			"l18 r9, 4(r5)\n"
			"s18 4(r6), r9\n"
			"l18 r9, 6(r5)\n"
			"s18 6(r6), r9\n"
			"l18 r9, 8(r5)\n"
			"s18 8(r6), r9\n"
			"l18 r9, 10(r5)\n"
			"s18 10(r6), r9\n"
			"l18 r9, 12(r5)\n"
			"s18 12(r6), r9\n"
			"l18 r9, 14(r5)\n"
			"s18 14(r6), r9\n"

			"addi r5, 16\n"
			"addi r6, 16\n"
			"addi r8, -8\n"
			"addi r7, -8\n"
			"check_limitt:\n"
			"slei r8, 0\n"
			"beqzc first_loopt\n"
			"slei r7, 0\n"
			"bnezc endt\n"
			"mov r8, r7\n"
			"mov r6, r4\n"
			"j first_loopt\n"
			"nop\n"

			"endt:\n" ::: "r4", "r5", "r6", "r7", "r8", "r9"
	);

	return;
}
/*
int network_receive(const struct netdev *dev, unsigned int *buf)
{
	unsigned int frame_size, frame_size_words, *rx_data, i;
	unsigned int offset = dev->rx->offset;

	if(!dev->rx_dma->x[offset]) {
		return 0;
	}

	frame_size = dev->rx_dma->x[offset];
	frame_size_words = frame_size / 2 + frame_size % 2;
	rx_data = dev->rx_dma->x + offset + 1;

	fast_read_from_ringbuffer(dev->rx_dma->x, rx_data, buf, frame_size_words);

	dev->rx->offset += frame_size_words + 1;

	return frame_size;
}
*/
int network_receive(const struct netdev *dev, unsigned int *buf)
{
	unsigned int frame_size, *rx_data, i;

	// Receive packet on eth0 and relay to eth1.
	rx_data = ethernet_rx_get_frame_start(dev->rx, dev->rx_dma, &frame_size);

	if(rx_data) {
		/*
		counter++;
		printf("%d %d %d %d %d %d\n", dev->rx->rxCount, counter, dev->rx->wrongCrcCount, frame_size, dev->rx->offset, dev->rx->control & 0x10);
		if(dev->rx->control & 0x10) {
			dev->rx->control &= ~0x10;
		}
		*/
		ethernet_rx_frame_free(dev->rx, dev->rx_dma);

		//read_from_ringbuffer(dev->rx_dma->x, rx_data, buf, frame_size / 2 + frame_size % 2);
		fast_read_from_ringbuffer(dev->rx_dma->x, rx_data, buf, frame_size / 2 + frame_size % 2);

		return frame_size;
	}
	
	return 0;
}

void fast_copy(unsigned int *src, unsigned int *dest, unsigned int len)
{
	/*
	 * r4: src
	 * r5: dest
	 * r6: len
	 * r7: count
	 * r8: buffer
	 */
	__asm__ volatile(
			"movi r7, 0x00\n"
			"startloop:\n"
			"l18 r8, 0(r4)\n"
			"s18 0(r5), r8\n"
			"l18 r8, 2(r4)\n"
			"s18 2(r5), r8\n"
			"l18 r8, 4(r4)\n"
			"s18 4(r5), r8\n"
			"l18 r8, 6(r4)\n"
			"s18 6(r5), r8\n"
			"l18 r8, 8(r4)\n"
			"s18 8(r5), r8\n"
			"l18 r8, 10(r4)\n"
			"s18 10(r5), r8\n"
			"l18 r8, 12(r4)\n"
			"s18 12(r5), r8\n"
			"l18 r8, 14(r4)\n"
			"s18 14(r5), r8\n"
			"addi r7, 8\n"
			"sltu r7, r6\n"
			"addi r4, 16\n"
			"bnezc startloop\n"
			"addi r5, 16\n" ::: "r4", "r5", "r6", "r7", "r8"
	);

	return;
}

void copy(unsigned int *src, unsigned int *dest, unsigned int len)
{
	unsigned int count = 0;

	while(count < len) {
		*dest = *src;
		dest++;
		src++;
		count++;
	}

	return;
}

void network_send(const struct netdev *dev, unsigned int *data, unsigned int size)
{
	unsigned int * tx_data;
	unsigned int i = 0;

	unsigned int *rbuf_start = dev->tx_dma->x;
	unsigned int *rbuf_end = rbuf_start + 1024;

	unsigned int *start;

	// Wait until enough space is free in the ringbuffer.
	while(1) {
		start = rbuf_start + (dev->tx->offset / 2);
		int freebytes = 2048;

		while(freebytes > 0) {
			unsigned int framesize = (*start / 2) + (*start % 2);
			if(framesize == 0) {
				break;
			}
			start += framesize + 1;
			freebytes -= (framesize + 1) * 2;
		}

		if(freebytes >= size) {
			break;
		}
	}

	// Copy data to ringbuffer.
	int framesize = (size / 2) + (size % 2);
	fast_copy_to_ringbuffer(dev->tx_dma->x, data, start + 1, framesize);

	// Mark end of frame.
	unsigned int *frameend = start + framesize + 1;
	if(frameend >= rbuf_end) {
		frameend -= 1024;
	}
	
	*frameend = 0;
	
	// Commit frame for sending.
	*start = size;
	
	return;
}
