/***************************************************************************/
/*                                                                         */
/*  Copyright (September 1997) Honeywell Inc. Unpublished - All rights     */
/*  reserved. This material may be reproduced by, or for, the U.S.         */
/*  Government pursuant to the copyright license under the clause at       */
/*  DFARS 252.227-7013 (Oct. 1988).                                        */
/*                                                                         */
/***************************************************************************/

/***************************************************************************
 * compressimage
 * Version 1.0
 *
 * File         : compress.c
 * Company      : Honeywell Technology Center, Minneapolis, MN 55412
 * Date         : November 10 1997
 * Authors      : Luiz Pires and Deepa Pandalai
 * Description  : This file contains all the routines required for a simple
 *              : wavelet based image compression of a 512x512 8-bit pixel
 *              : image in PGM format. The compression process consists of
 *              : four basic steps: Wavelet transform, quantization,
 *              : run length encoding and entropy coding. Each of these 
 *              : steps is implemented as a function and is described below. 
 *              : This program will expect the name of the pgm image to
 *              : be compressed as a required input. It will also accept an
 *              : optional compression-factor parameter specifying how 
 *              : aggresively the image should be compressed: A compression
 *              : factor indicates minimal compression and maximum: image
 *              : quality. Conversely, a compression factor of 255 indicates
 *              : maximum compression with higher degradation to the image
 *              : quality. If the compression factor is not indicated, a  
 *              : default compression rate of 128 will be used. The compressed
 *              : image file format is specific to this program a  will be
 *              : indicated by a ".cmp" extension.
 *              : NOTE: The wavelet transform routine implements a modified
 *              : version of the biorthogonal (2-2) Cohen-Daubechies-Feauveau
 *              : wavelet. The routine employs the lifting scheme to simplify
 *              : the wavelet implementation. As it stands it requires only
 *              : integer adds and shifts. A reference on the transform used
 *              : is "A. Cohen, I. Daubechies, J. Feauveau, ``Biorthogonal
 *              : Bases of Compactly Supported Wavelets,'' Comm. Pure Appl.
 *              : Math., 45 (1992). More effective wavelet filters may be found
 *              : in "J. Villasenor, B. Belzer, J. Liao, ``Wavelet Filter
 *              : Evaluation for Image Compression,'' IEEE Transactions on
 *              : Image Processing, vol. 2, pp. 1053-1060, August 1995. 
 * Modifications:
 *  Yanbing Li
 * Various optimizations (memory usage, procedure inling, loop optimizations, 
 * strength reduction etc.)
 * Modify for running on ACEII
 *************************************************************************/

#define debugPrintf(x,...) if (x) { printf(__VA_ARGS__);}
#define debugForwardWavelet2 0
#define debugBlockQuantize 0

//#include<sys/time.h>

#include "globals.h"
#ifdef spartanmc
#include <subsystems/spartanmc_0/peripherals.h>
#include <perf.h>
#endif
#include "memdump.h"
#include <stdio.h>


#if WIDTH==64
#include "img64.h"
#define int_data_g int_data_g64
#elif WIDTH==128
#include "img128.h"
#define int_data_g int_data_g128
#elif WIDTH==256
#include "img256.h"
#define int_data_g int_data_g256
#elif WIDTH==512
#include "img512.h"
#define int_data_g int_data_g512
#endif

#define MAX_PARALLEL 0


char in_name[80],out_name[80];
int nbytes_g;



/* Quantization variables */
/* 
   The image will be divided into 10 blocks, the first four will be of
   size 64x64 pixels (4096 pixels), then three of size 128x128 (16384 pixels)
   and the remaining three of size 256x256 pixels (65536 pixels).
   These blocks will be arranged as represented below:

   +---+---+-------+---------------+
   | 4K| 4K|       |               |
   +---+---|  16K  |               |
   | 4K| 4K|       |               |
   +---+---+-------+      65K      |
   |       |       |               |
   |  16K  |  16K  |               |
   |       |       |               |
   +-------+-------+---------------+
   |               |               |
   |               |               |
   |               |               |
   |      65K      |      65K      |
   |               |               |
   |               |               |
   |               |               |
   +---------------+---------------+
*/


int blockminval_g[NBLKS], blockmaxval_g[NBLKS];

int quant_buf_g[IMG_SIZE];

//struct timeval tv1, tv2;
//struct timezone tz1, tz2;
/* RLE variables */
unsigned char rle_buf_g[IMG_SIZE]; /*reuse int_data*/
unsigned int rle_size_g[NBLKS];

/* Huffman Encoding variables */
unsigned char codep_g[IMG_SIZE]; /*reuse quant_buf */



/* Function Declarations */
int read_image(char *);
void forward_wavelet();
void quantization(imgtype * int_data, int * quant_buf, int * blockminval, int * blockmaxval);
void RLE_encode(int * quant_buf, unsigned char * rle_buf, int * rle_size);
void entropy_encode(unsigned char * rle_buf, int * rle_size, unsigned char * codep, int* nbytes);
void forward_wavelet1(int, imgtype *);
void forward_wavelet2(int, imgtype *);
void block_quantize(int x, int y, int size, int num, imgtype * int_data, int * quant_buf, int * blockminval, int * blockmaxval);
void write_compressed_file();


static int classifyThreshold(const int *thresh, int val);

void innerLoop(int num, const imgtype *int_data, const int *thresh, int i, int j, int val, int qsize,
	        int *quant_buf_pt, int tmp_classify, const int *blockthresh);

void
loops(int x, int y, int size, int num, const imgtype *int_data, const int *thresh, int i, int j, int val, int qsize,
       int *quant_buf_pt, int tmp_classify, const int *blockthresh);

void dump_int_data() {
    printf("image dump\n");
    const unsigned int wordcount = sizeof(int_data_g)/sizeof(int_data_g[0]);
    unsigned char intasshort = sizeof(int_data_g[0])==4 ? MEMDUMP_BM_INT_AS_SHORT : 0;
    memdump(0,int_data_g,wordcount,MEMDUMP_BM_UPPER_AT_8|MEMDUMP_BM_IGNORE_BIT_9|MEMDUMP_BM_SWAP|intasshort);
}

void dump_quantization() {
#if debugBlockQuantize
    unsigned char intasshort = sizeof(quant_buf_g[0])==4 ? MEMDUMP_BM_INT_AS_SHORT : 0;

    printf("quant buf dump\n");
    const unsigned int wordcountQuantBuf = sizeof(quant_buf_g)/sizeof(quant_buf_g[0]);
    memdump(0,quant_buf_g,wordcountQuantBuf,MEMDUMP_BM_UPPER_INVALID|MEMDUMP_BM_SWAP|intasshort);

    printf("blockminval dump\n");
    const unsigned int wordcountBlockminval = sizeof(blockminval_g)/sizeof(blockminval_g[0]);
    memdump(0,blockminval_g,wordcountBlockminval,MEMDUMP_BM_UPPER_INVALID|MEMDUMP_BM_SWAP|intasshort);

    printf("blockmaxval dump\n");
    const unsigned int wordcountBlockmaxval = sizeof(blockmaxval_g)/sizeof(blockmaxval_g[0]);
    memdump(0,blockmaxval_g,wordcountBlockminval,MEMDUMP_BM_UPPER_INVALID|MEMDUMP_BM_SWAP|intasshort);
#endif
}

/********************************************************************
 * Routine: main(int argc, char **argv)
 * Expects image file name and the scale factor for compression as
 * input parameters. Calls all the required routines for compression.
 * The compressed file is the output of write_compressed_file().
 * This file format is specific to this program and includes all image
 * information required for decompression. 
 *********************************************************************/
main(int argc, char **argv)
{
	int s[256],d[256];
	#if (debugForwardWavelet2 == 1)
    debugPrintf(debugForwardWavelet2,"1. nt = %d\n",WIDTH);
	#endif
#pragma microstreams loopOpt 3
    for (int i=0;i<WIDTH*COL;i+=COL) {
        int mid=(WIDTH>>1)-1;
        imgtype* x=&int_data_g[i];
        s[0] = x[0];
        d[0] = x[ROW];
        s[1] = x[2];
        s[mid] = x[2*mid];
        d[mid] = x[2*mid+ROW];

        d[0]=(d[0]<<1)-s[0]-s[1];
        s[0]=s[0]+(d[0]>>2);

        int d_tmp0 = d[0];
        int s_tmp0 = s[1];

        for(int ii=1; ii<mid; ii++)
        {  // k-s: yes 
            int s_tmp1 = x[2*ii+2]; //TODO memory error here
            int d_tmp1 =((x[2*ii+ROW])<<1) - s_tmp0 - s_tmp1;
            d[ii] = d_tmp1;
            s[ii]= s_tmp0+((d_tmp0+d_tmp1)>>3);
            d_tmp0 = d_tmp1;
            s_tmp0 = s_tmp1;
        }

        d[mid]=(d[mid]-s[mid])<<1;
        s[mid]=s[mid]+((d[mid-1]+d[mid])>>3);

        for(int ii=0; ii<=mid; ii++)
        { // k-s: no not working correctly individually, don't select it 
            x[ii]=s[ii];
            x[ii+mid+1]=d[ii];
        }
    }

#pragma microstreams task
	int s2[256],d2[256];
#pragma microstreams loopOpt 3
    for (int i=0;i<WIDTH;i++) {
        // fcdf22(&int_data[i],nt,COL); inline
        int mid=(WIDTH>>1)-1;
        imgtype* x=&int_data_g[i];
        #if (debugForwardWavelet2 == 1)
		debugPrintf(debugForwardWavelet2,"x=data[%d]\n", i);
        debugPrintf(debugForwardWavelet2, "mid = %d\n", mid);
		#endif        
		s2[0] = x[0];
		#if (debugForwardWavelet2 == 1)        
		debugPrintf(debugForwardWavelet2, "s2[0] = %d\n", s2[0]);
		#endif
        d2[0] = x[COL];
		#if (debugForwardWavelet2 == 1)
        debugPrintf(debugForwardWavelet2, "d2[0] = %d\n", d2[0]);
		#endif
        s2[1] = x[COL<<1];
		#if (debugForwardWavelet2 == 1)
        debugPrintf(debugForwardWavelet2, "s2[1] = %d\n", s2[1]);
		#endif
        s2[mid] = x[(COL<<1)*mid];
        d2[mid] = x[(COL<<1)*mid +COL];

        d2[0]=(d2[0]<<1)-s2[0]-s2[1];
        s2[0]=s2[0]+(d2[0]>>2);

        int d_tmp0 = d2[0];
		#if (debugForwardWavelet2 == 1)
        debugPrintf(debugForwardWavelet2, "d_tmp0 = %d\n", d_tmp0);
		#endif
        int s_tmp0 = s2[1];
		#if (debugForwardWavelet2 == 1)        
		debugPrintf(debugForwardWavelet2, "s_tmp0 = %d\n", s_tmp0);
        debugPrintf(debugForwardWavelet2, "test 1 data[683] = %d\n", int_data_g[683]);
		#endif
        for(int ii=1; ii<mid; ii++)
        {  // k-s: yes 
			#if (debugForwardWavelet2 == 1)            
			debugPrintf(debugForwardWavelet2,"ii = %d\n", ii);
			#endif
            int idx = 2 * COL * (ii + 1);
            int idxi = idx+i;
			#if (debugForwardWavelet2 == 1)
            debugPrintf(debugForwardWavelet2,"2*COL*(ii+1) = %d, +i = %d\n", idx, idxi);
			#endif
            int s_tmp1 = x[idx];
			#if (debugForwardWavelet2 == 1)
            debugPrintf(debugForwardWavelet2, "s_tmp1 = %d\n", s_tmp1);
			#endif
            int d_tmp1 =(x[2*COL*ii+COL]<<1) - s_tmp0 - s_tmp1;
			#if (debugForwardWavelet2 == 1)
            debugPrintf(debugForwardWavelet2, "d_tmp1 = %d\n", d_tmp1);
			#endif            
			d2[ii] = d_tmp1;
            s2[ii]= s_tmp0+((d_tmp0+d_tmp1)>>3);
			#if (debugForwardWavelet2 == 1)
            debugPrintf(debugForwardWavelet2, "s2[ii] = %d\n", s2[ii]);
			#endif            
			d_tmp0 = d_tmp1;
            s_tmp0 = s_tmp1;
        }

        d2[mid]=(d2[mid]<<1) -(s2[mid]<<1);
        s2[mid]=s2[mid]+((d2[mid-1]+d2[mid])>>3);
		#if (debugForwardWavelet2 == 1)
        debugPrintf(debugForwardWavelet2, "d2[mid] = %d\n", d2[mid]);
        debugPrintf(debugForwardWavelet2, "s2[mid] = %d\n", s2[mid]);
		#endif
        for(int ii=0; ii<=mid; ii++)
        { // k-s: no not working correctly individually, don't select it
			#if (debugForwardWavelet2 == 1)            
			debugPrintf(debugForwardWavelet2, "data[%d]=int_data_g[%d]=s2[%d]=%d\n",i+ii*COL,ii*COL,ii,s2[ii]);
            debugPrintf(debugForwardWavelet2, "data[%d]=int_data_g[%d]=d2[%d]=%d\n",(ii+mid+1)*COL+i, (ii+mid+1)*COL,ii,d2[ii]);
			#endif
            x[ii*COL]=s2[ii];
            x[(ii+mid+1)*COL]=d2[ii];
        }
    }

#pragma microStreams task
    forward_wavelet1(WIDTH/2,&int_data_g[0]);
   // dump_int_data();
#if MAX_PARALLEL == 1
#pragma microstreams task
#endif
    forward_wavelet2(WIDTH/2,&int_data_g[0]);

   // dump_int_data();
    forward_wavelet1(WIDTH/4,&int_data_g[0]);
   // dump_int_data();
    forward_wavelet2(WIDTH/4,&int_data_g[0]);
  //  dump_int_data();

	#pragma microStreams task

     //printf("-- End wavelet. \n-- Quantization ... ...\n");
    quantization(& int_data_g[0], &quant_buf_g[0], &blockminval_g[0], &blockmaxval_g[0]);
#if MAX_PARALLEL == 1
#pragma microstreams task
#endif
     //printf("-- End quantization. \n-- RLE encoding ... ...\n");
    RLE_encode(&quant_buf_g[0], &rle_buf_g[0], &rle_size_g[0]);
#pragma microstreams task
   entropy_encode(&rle_buf_g[0], &rle_size_g[0], &codep_g[0], &nbytes_g);

     //printf("-- End entropy encoding. \n-- Write compressed file ... ...\n");
    write_compressed_file();
}


/********************************************************************
 * Routine: void quantization() //input: int_data,     output: quant_buf, blockminval, blockmaxval
 * 
 *********************************************************************/
void quantization(imgtype * int_data, int * quant_buf, int * blockminval, int * blockmaxval)
{
    int bl,num;
    num=0;
    block_quantize(0,0,BLOCK_SIZE>>1, num++, int_data, quant_buf, blockminval, blockmaxval); /*yb*/
	#if (debugBlockQuantize == 1)    
	debugPrintf(debugBlockQuantize, "before loop\n");
	#endif
    dump_quantization();
    for(bl=BLOCK_SIZE>>1; bl<COL>>1; bl<<=1) /*yb*/
    {
        block_quantize(0,bl,bl,num++, int_data, quant_buf, blockminval, blockmaxval);
        block_quantize(bl,0,bl,num++, int_data, quant_buf, blockminval, blockmaxval);
        block_quantize(bl,bl,bl,num++, int_data, quant_buf, blockminval, blockmaxval);
	#if (debugBlockQuantize == 1) 
	debugPrintf(debugBlockQuantize, "after bl=%d\n",bl);
	#endif
	dump_quantization();
    }
}


void __attribute__((noinline,noclone))
loopsTemp(int x, int y, int size, int qsize);

void __attribute__((noinline,noclone))
getMinMax(int x, int y, int size, const imgtype *int_data, int *minval,
	  int *maxval);

/********************************************************************
 * Routine: block_quantize(int x, int y, int size, int num)
 *********************************************************************/
void block_quantize(int x, int y, int size, int num, imgtype * int_data, int * quant_buf, int * blockminval, int * blockmaxval)
{
	static int thresh[17]; /* valid: 1-15 */
#if WIDTH==64
    int buf_pts[NBLKS]= {0,64,128,192,256,512,768,1024,2048,3072};
#elif WIDTH==128
    int buf_pts[NBLKS]= {0,256,512,768,1024,2048,3072,4096,8192};
#elif WIDTH==256
    int buf_pts[NBLKS]= {0,1024,2048,3072,4096,8192,12288,16384,32768,49152};
#elif WIDTH==512
    int buf_pts[NBLKS]= {0,4096,8192,12288,16384,32768,49152,65536,131072,196608};
#endif
    int i,j,val;
    int minval,maxval;
    int qsize;
	#if (debugBlockQuantize == 1) 
    debugPrintf(debugBlockQuantize, "buf_pts[%d] = %d\n", num, buf_pts[num]);
	#endif
    int *quant_buf_pt = quant_buf + buf_pts[num];
    int tmp_classify =0;
    int blockthresh[NBLKS]={0, 39, 27, 104,
                            79, 50,191, 99999,
                            99999, 99999};

    minval = 10000; maxval =-10000;

    getMinMax(x, y, size, int_data, &minval, &maxval);
	#if (debugBlockQuantize == 1) 
    debugPrintf(debugBlockQuantize, "minval = %d\n", minval);
    debugPrintf(debugBlockQuantize, "maxval = %d\n", maxval);
	#endif
    for (i=1; i<=16; i++) /*yb: 16->15? since 16 not used*/ {
        thresh[i]= minval+((i*(maxval-minval)+8)>>4);
	#if (debugBlockQuantize == 1) 
	debugPrintf(debugBlockQuantize, "thresh[%d] = %d\n", i, thresh[i]);
	#endif
    }

    qsize=0;

    loops(x, y, size, num, int_data, thresh, i, j, val, qsize, quant_buf_pt, tmp_classify, blockthresh);

    blockminval[num]=minval;
    blockmaxval[num]=maxval;

}

void
getMinMax(int x, int y, int size, const imgtype *int_data, int *minval,
int *maxval) {
    int xsize= x + size;
    int ysize= y+size;
		#if (debugBlockQuantize == 1) 
    debugPrintf(debugBlockQuantize, "running from x %d to %d and y %d to %d\n",x,xsize,y,ysize);
	#endif
    for(int i=x; i<xsize; i++) {
for(int j =y; (j) < ysize; (j) = (j) + 1)
{ /* k-s: yes */
int val =int_data[i + (j) * COL]; /*no need to change, shifts are free in FPGA*/
	#if (debugBlockQuantize == 1)   
	debugPrintf(debugBlockQuantize, "val = %d\n", (val));
	#endif
(*minval) = min((val), (*minval));
(*maxval) = max((val), (*maxval));
	#if (debugBlockQuantize == 1) 
  debugPrintf(debugBlockQuantize, "loop minval = %d\n", (*minval));
  debugPrintf(debugBlockQuantize, "loop maxval = %d\n", (*maxval));
	#endif
}
/* printf("out of hw kernel %d\n", i); */
}
}

void
loops(int x, int y, int size, int num, const imgtype *int_data, const int *thresh, int i, int j, int val, int qsize,
      int *quant_buf_pt, int tmp_classify, const int *blockthresh) {
    for(i=x; i < x + size; i++)
        for(j=y; j<y+size; j++)
        {/* ks: no */ /* 115 rows, too big to fit in hardware */
				#if (debugBlockQuantize == 1) 
            debugPrintf(debugBlockQuantize, "0 qsize = %d\n",qsize);
			#endif
            innerLoop(num, int_data, thresh, i, j, val, qsize, quant_buf_pt, tmp_classify, blockthresh);
			#if (debugBlockQuantize == 1) 
            debugPrintf(debugBlockQuantize, "4 qsize = %d\n",qsize);
			#endif
            qsize++;
			#if (debugBlockQuantize == 1) 
            debugPrintf(debugBlockQuantize, "5 qsize = %d\n",qsize);
			#endif
        }
}

void innerLoop(int num, const imgtype *int_data, const int *thresh, int i, int j, int val, int qsize,
               int *quant_buf_pt, int tmp_classify, const int *blockthresh) {
	#if (debugBlockQuantize == 1)     
	debugPrintf(debugBlockQuantize, "1 qsize = %d, i=%d, j=%d\n", qsize, i, j);
	#endif    
	val=int_data[i+COL*j];

    if(my_abs(val)<blockthresh[num]) {
        quant_buf_pt[qsize]=ZERO_MARK;
		#if (debugBlockQuantize == 1) 		
		debugPrintf(debugBlockQuantize, "zero quant_buf_pt[%d] = %d\n", qsize, quant_buf_pt[qsize]);
		#endif
            }
            else
            {
			#if (debugBlockQuantize == 1) 		
		debugPrintf(debugBlockQuantize, "2 qsize = %d\n",qsize);
		#endif        
        tmp_classify = classifyThreshold(thresh, val);

			#if (debugBlockQuantize == 1) 
                debugPrintf(debugBlockQuantize, "3 qsize = %d\n",qsize);
		debugPrintf(debugBlockQuantize, "tmp quant_buf_pt[%d] = %d\n", qsize, tmp_classify);
		#endif        
        quant_buf_pt[qsize] = tmp_classify;
            }
}

static int __attribute__((noinline,noclone)) classifyThreshold(const int *thresh, int val) {
    int tmp_classify;
    if(val > thresh[8]) {
                    if(val>thresh[12]) {
                        if(val>thresh[14]) {
                            if(val>thresh[15]) tmp_classify = 15;
                            else tmp_classify = 14;

                        } else {
                            if(val>thresh[13]) tmp_classify = 13;
                            else tmp_classify = 12;
                        } } else {
                        if(val>thresh[10]) {
                            if(val>thresh[11]) tmp_classify = 11;
                            else tmp_classify = 10;
                        } else {
                            if(val>thresh[9]) tmp_classify = 9;
                            else tmp_classify = 8;
                        } } } else {
                    if(val>thresh[4]) {
                        if(val>thresh[6]) {
                            if(val>thresh[7]) tmp_classify = 7;
                            else tmp_classify = 6;
                        } else {
                            if(val>thresh[5]) tmp_classify = 5;
                            else tmp_classify = 4;
                        } } else {
                        if(val>thresh[2]) {
                            if(val>thresh[3]) tmp_classify = 3;
                            else tmp_classify = 2;
                        } else {
                            if(val>thresh[1]) tmp_classify = 1;
                            else tmp_classify = 0;
                        } } }
    return tmp_classify;
}

/********************************************************************
 * Routine: void  RLE_encode()
 *********************************************************************/
void  RLE_encode(int * quant_buf, unsigned char * rle_buf, int * rle_size) //input: quand_buf    output: rle_buf, rle_size
{
    //
//
//
//
#if WIDTH==64
    int blocksize[NBLKS]= {64,64,64,64,256,256,256,1024,1024,1024};
    int buf_pts[NBLKS]= {0,64,128,192,256,512,768,1024,2048,3072};
#elif WIDTH==128
    int blocksize[NBLKS]= {256,256,256,256,1024,1024,1024,4096,4096,4096};
    int buf_pts[NBLKS]= {0,256,512,768,1024,2048,3072,4096,8192};
#elif WIDTH==256
    int blocksize[NBLKS]= {1024,1024,1024,1024,4096,4096,4096,16384,16384,16384};
    int buf_pts[NBLKS]= {0,1024,2048,3072,4096,8192,12288,16384,32768,49152};
#elif WIDTH==512
    int blocksize[NBLKS]= {4096,4096,4096,4096,16384,16384,16384,65536,65536,65536};
    int buf_pts[NBLKS]= {0,4096,8192,12288,16384,32768,49152,65536,131072,196608};
#endif
    unsigned int bl;
    int i, j, count;
    int *quant_buf_pt;
    unsigned char *rle_buf_pt;
    for (bl = 0; bl < BLKS_SAVED; /*NBLKS;*/ bl++) {
        /*block_RLE_encode(bl); inlined */
        quant_buf_pt = quant_buf + buf_pts[bl];
        rle_buf_pt = rle_buf + buf_pts[bl];

        i = j = 0;
        while(i<blocksize[bl])
        {
            if(quant_buf_pt[i]!=ZERO_MARK)
                rle_buf_pt[j]=quant_buf_pt[i];
            else {
                count=0;

                while(quant_buf_pt[i]==ZERO_MARK) {  /* k-s: no */
                    i++;
                    count++;
                    /* if((count==256-ZERO_MARK)||(i==IMG_SIZE)) break;*/
                    if((count==256-ZERO_MARK)||(i==blocksize[bl]))
                        break;
                }
                i--;
                rle_buf_pt[j]=count+ZERO_MARK-1;
            }
            j++;
            i++;
        }
        rle_size[bl]=j;
    }
}




/********************************************************************
 * Routine: void entropy_encode()
 *********************************************************************/
void entropy_encode(unsigned char * rle_buf, int * rle_size, unsigned char * codep, int* nbytes) //input: rle_buf, rle_size   output:codep, nbytes...done here
{
#if WIDTH==64
    int buf_pts[NBLKS]= {0,64,128,192,256,512,768,1024,2048,3072};
#elif WIDTH==128
    int buf_pts[NBLKS]= {0,256,512,768,1024,2048,3072,4096,8192};
#elif WIDTH==256
    int buf_pts[NBLKS]= {0,1024,2048,3072,4096,8192,12288,16384,32768,49152};
#elif WIDTH==512
    int buf_pts[NBLKS]= {0,4096,8192,12288,16384,32768,49152,65536,131072,196608};
#endif
    /* Parameters for a fixed huffman encoding tree. These parameters are  */
/* expected to work reasonably well with most images.                  */
    int
            HufSize[256]={ 9 , 7 , 6 , 5 , 4 , 4 , 3 , 4 , 3 , 3 , 4 , 5 , 6 , 6 , 7 , 8 ,
                           4 , 5 , 5 , 6 , 6 , 7 , 7 , 7 , 8 , 8 , 8 , 8 , 9 , 9 , 9 , 9 ,
                           9 , 9 , 9 ,10 , 9 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,10 ,11 ,
                           10 ,11 ,11 ,10 ,10 ,11 ,11 ,11 ,11 ,11 ,11 ,12 ,12 ,12 ,12 ,11 ,
                           11 ,12 ,11 ,12 ,12 ,12 ,12 ,11 ,13 ,12 ,12 ,12 ,12 ,13 ,12 ,13 ,
                           13 ,13 ,14 ,13 ,13 ,14 ,13 ,13 ,13 ,13 ,14 ,13 ,13 ,13 ,13 ,13 ,
                           13 ,13 ,13 ,13 ,14 ,13 ,14 ,13 ,13 ,14 ,15 ,13 ,13 ,14 ,13 ,14 ,
                           13 ,14 ,13 ,14 ,14 ,14 ,14 ,13 ,13 ,13 ,13 ,14 ,13 ,13 ,13 ,13 ,
                           14 ,13 ,15 ,14 ,15 ,14 ,14 ,13 ,14 ,16 ,14 ,14 ,13 ,12 ,12 ,13 ,
                           15 ,15 ,17 ,18 ,18 ,18 ,17 ,17 ,15 ,17 ,18 ,16 ,17 ,17 ,16 ,16 ,
                           18 ,18 ,16 ,18 ,16 ,18 ,18 ,18 ,16 ,16 ,16 ,18 ,15 ,16 ,18 ,15 ,
                           16 ,18 ,16 ,16 ,18 ,18 ,18 ,18 ,16 ,18 ,16 ,18 ,18 ,18 ,16 ,18 ,
                           18 ,18 ,18 ,18 ,18 ,18 ,18 ,18 ,16 ,16 ,18 ,18 ,17 ,18 ,18 ,18 ,
                           18 ,17 ,15 ,18 ,18 ,15 ,15 ,18 ,17 ,16 ,16 ,15 ,16 ,16 ,18 ,18 ,
                           16 ,17 ,18 ,18 ,15 ,18 ,16 ,16 ,18 ,18 ,18 ,18 ,16 ,18 ,18 ,16 ,
                           18 ,16 ,18 ,18 ,18 ,18 ,18 ,18 ,18 ,16 ,18 ,18 ,17 ,18 ,18 , 9 };

    int
            HufVal[256]={0x00097,0x00066,0x00007,0x00016,0x00000,0x0000B,0x00001,0x0000F,
                         0x00002,0x00004,0x0000E,0x00015,0x00037,0x00025,0x00038,0x000ED,
                         0x00003,0x0001D,0x00008,0x0000D,0x00018,0x00067,0x00005,0x00006,
                         0x000D7,0x0002D,0x000C6,0x000F8,0x001A7,0x00197,0x0016D,0x00127,
                         0x001C5,0x001A6,0x00146,0x00357,0x00178,0x00257,0x000AD,0x0006D,
                         0x000A7,0x003AD,0x001AD,0x00017,0x000C5,0x002C5,0x00145,0x00117,
                         0x000A6,0x00427,0x002A7,0x00246,0x00226,0x002AD,0x00217,0x00027,
                         0x006A6,0x002A6,0x00445,0x00A27,0x00A6D,0x00717,0x00D57,0x00345,
                         0x00446,0x00057,0x00726,0x006A7,0x00026,0x00EA7,0x00E17,0x00126,
                         0x0026D,0x00D26,0x006AD,0x00B17,0x00845,0x00F17,0x00317,0x00617,
                         0x01457,0x01326,0x02526,0x00B26,0x01D17,0x01F17,0x00627,0x01C57,
                         0x0166D,0x00E6D,0x00957,0x01E6D,0x01E45,0x01526,0x00517,0x01846,
                         0x00745,0x01617,0x00157,0x01627,0x00E27,0x01EAD,0x00227,0x01745,
                         0x00245,0x01957,0x06957,0x01517,0x01157,0x0326D,0x01645,0x02557,
                         0x00A45,0x03957,0x01046,0x03045,0x00457,0x02227,0x00557,0x01A45,
                         0x00046,0x00C57,0x00326,0x02457,0x00EAD,0x01B26,0x00645,0x01227,
                         0x01426,0x0066D,0x02E45,0x01557,0x03F17,0x02E27,0x03426,0x01245,
                         0x00D17,0x02957,0x03557,0x02D17,0x00045,0x00857,0x00F45,0x00846,
                         0x0526D,0x00426,0x1A957,0x06C26,0x26C26,0x14426,0x1D045,0x09E27,
                         0x00526,0x19E27,0x04426,0x06E45,0x05E27,0x15E27,0x02C26,0x04526,
                         0x24426,0x0CC26,0x0EE45,0x2CC26,0x0C526,0x16C26,0x36C26,0x0AC26,
                         0x07E27,0x01826,0x09826,0x2AC26,0x00C26,0x05826,0x0A826,0x01045,
                         0x0D826,0x2A826,0x04C26,0x0CE45,0x1A826,0x3A826,0x1AC26,0x3AC26,
                         0x0C426,0x1CC26,0x0126D,0x1EC26,0x3EC26,0x1DC26,0x05C26,0x3DC26,
                         0x0DC26,0x2DC26,0x1C826,0x3C826,0x04826,0x24826,0x0C826,0x2C826,
                         0x03826,0x0B826,0x16426,0x36426,0x0DE27,0x06426,0x0E826,0x2E826,
                         0x14826,0x1DE27,0x03E27,0x34826,0x0A426,0x07826,0x00826,0x2A426,
                         0x0FE27,0x0BC26,0x02826,0x00E45,0x07C26,0x0926D,0x05045,0x25045,
                         0x0FC26,0x1FE27,0x15045,0x35045,0x07F17,0x1A426,0x06826,0x0E426,
                         0x3A426,0x19C26,0x39C26,0x09C26,0x02426,0x29C26,0x26426,0x01C26,
                         0x13C26,0x01E27,0x33C26,0x0D045,0x2D045,0x03C26,0x3CC26,0x23C26,
                         0x1E826,0x04E45,0x3E826,0x0EC26,0x0A957,0x2EC26,0x34426,0x00078};
    int i,j,nb;
    int rlecount;
    unsigned char *rle_buf_pt;
    unsigned int ii, nbits;
    int val;

    int bytn,bitn;
    unsigned char ich;


    rlecount=0;
    nb=0;
    for (i = 0; i < BLKS_SAVED; i++)
    {
        rle_buf_pt = rle_buf + buf_pts[i];
        for(j = 0; j < rle_size[i]; j++) { /* k-s: no */
            /* outer loop, not considered for hardware */
            /* hufenc(rle_buf_pt[j],&nb); inline */
            ich = rle_buf_pt[j];
            nbits=HufSize[ich];
            val=HufVal[ich];

            for(ii=0; ii<nbits; ii++)
            { /* k-s: no  */
                bytn=nb>>3;
                bitn=nb&7;
                if((val&(1<<ii))>0)
                    codep[bytn]|=(1<<bitn);
                nb++;
            }

        }
        rlecount+=rle_size[i];
    }
    *nbytes = (nb>>3);
    int c = (*nbytes)+88;
	//printf("    Compressed size is %d bytes. Expecting 849 bytes\n",c);
}

#ifdef spartanmc
FILE * stdout = &UART_LIGHT_0_FILE;
#endif


/********************************************************************
 * Routine: int write_compressed_file()
 ********************************************************s*************/

void write_compressed_file()
{
#ifdef spartanmc
    //memdump(0,codep_g,nbytes_g,MEMDUMP_BM_IGNORE_BIT_9|MEMDUMP_BM_COUNT_BYTES);
#else
    //memdump(0,codep_g,nbytes_g,MEMDUMP_BM_IGNORE_BIT_9|MEMDUMP_BM_UPPER_AT_8|MEMDUMP_BM_COUNT_BYTES);
#endif
    printf("Compressed size is %d bytes. Expecting 849 bytes\n",nbytes_g+88);
}


imgtype TRUNCATE16(imgtype x) {
#ifdef spartanmc
    return x&0xFFFF;
#else
    return x;
#endif
}

/********************************************************************
 * Routine: void forward_wavelet()
 *********************************************************************/

void forward_wavelet1(int nt, imgtype * data){//inout int_data
    int i;
    int d_tmp0, d_tmp1, s_tmp0, s_tmp1;
    int mid, ii;
    imgtype* x;
    int s[256],d[256];
	#if (debugForwardWavelet2 == 1) 
    debugPrintf(debugForwardWavelet2,"1. nt = %d\n",nt);
	#endif
    for (i=0;i<nt*COL;i+=COL) {
        mid=(nt>>1)-1;
        x=&data[i];
        s[0] = x[0];
        d[0] = x[ROW];
        s[1] = x[2];
        s[mid] = x[2*mid];
        d[mid] = x[2*mid+ROW];

        d[0]=(d[0]<<1)-s[0]-s[1];
        s[0]=s[0]+(d[0]>>2);

        d_tmp0 = d[0];
        s_tmp0 = s[1];

        for(ii=1; ii<mid; ii++)
        {  /* k-s: yes */
            s_tmp1 = x[2*ii+2]; //TODO memory error here
            d_tmp1 =((x[2*ii+ROW])<<1) - s_tmp0 - s_tmp1;
            d[ii] = d_tmp1;
            s[ii]= s_tmp0+((d_tmp0+d_tmp1)>>3);
            d_tmp0 = d_tmp1;
            s_tmp0 = s_tmp1;
        }

        d[mid]=(d[mid]-s[mid])<<1;
        s[mid]=s[mid]+((d[mid-1]+d[mid])>>3);

        for(ii=0; ii<=mid; ii++)
        { /* k-s: no */ /* not working correctly individually, don't select it */
            x[ii]=s[ii];
            x[ii+mid+1]=d[ii];
        }
    }

}

void forward_wavelet2(int nt,imgtype * data){//inout int_data
    int i;
    int d_tmp0, d_tmp1, s_tmp0, s_tmp1;
    int mid, ii;
    imgtype* x;
    int s[256],d[256];
	#if (debugForwardWavelet2 == 1) 
    debugPrintf(debugForwardWavelet2,"2. nt = %d\n",nt);
	#endif    
	for (i=0;i<nt;i++) {
        /* fcdf22(&int_data[i],nt,COL); inline */
        mid=(nt>>1)-1;
        x=&data[i];
	#if (debugForwardWavelet2 == 1) 	
	debugPrintf(debugForwardWavelet2,"x=data[%d]\n", i);
	debugPrintf(debugForwardWavelet2, "mid = %d\n", mid);
	#endif
        s[0] = x[0];
	#if (debugForwardWavelet2 == 1) 	
	debugPrintf(debugForwardWavelet2, "s[0] = %d\n", s[0]);
	#endif    
	    d[0] = x[COL];
	#if (debugForwardWavelet2 == 1) 
	debugPrintf(debugForwardWavelet2, "d[0] = %d\n", d[0]);
	#endif    
	    s[1] = x[COL<<1];
	#if (debugForwardWavelet2 == 1) 	
	debugPrintf(debugForwardWavelet2, "s[1] = %d\n", s[1]);
	#endif
        s[mid] = x[(COL<<1)*mid];
        d[mid] = x[(COL<<1)*mid +COL];

        d[0]=(d[0]<<1)-s[0]-s[1];
        s[0]=s[0]+(d[0]>>2);

        d_tmp0 = d[0];
	#if (debugForwardWavelet2 == 1) 	
	debugPrintf(debugForwardWavelet2, "d_tmp0 = %d\n", d_tmp0);
	#endif
        s_tmp0 = s[1];
	#if (debugForwardWavelet2 == 1) 	
	debugPrintf(debugForwardWavelet2, "s_tmp0 = %d\n", s_tmp0);
	debugPrintf(debugForwardWavelet2, "test 1 data[683] = %d\n", data[683]);
	#endif
        for(ii=1; ii<mid; ii++)
        {  /* k-s: yes */
		#if (debugForwardWavelet2 == 1) 	    
		debugPrintf(debugForwardWavelet2,"ii = %d\n", ii);
		#endif
	    int idx = 2 * COL * (ii + 1);
	    int idxi = idx+i;
		#if (debugForwardWavelet2 == 1) 	    
		debugPrintf(debugForwardWavelet2,"2*COL*(ii+1) = %d, +i = %d\n", idx, idxi);
		#endif
            s_tmp1 = x[idx];
		#if (debugForwardWavelet2 == 1) 	    
		debugPrintf(debugForwardWavelet2, "s_tmp1 = %d\n", s_tmp1);
		#endif
            d_tmp1 =(x[2*COL*ii+COL]<<1) - s_tmp0 - s_tmp1;
		#if (debugForwardWavelet2 == 1) 	    
		debugPrintf(debugForwardWavelet2, "d_tmp1 = %d\n", d_tmp1);
		#endif
            d[ii] = d_tmp1;
            s[ii]= s_tmp0+((d_tmp0+d_tmp1)>>3);
		#if (debugForwardWavelet2 == 1) 	    
		debugPrintf(debugForwardWavelet2, "s[ii] = %d\n", s[ii]);
		#endif        
		    d_tmp0 = d_tmp1;
            s_tmp0 = s_tmp1;
        }

        d[mid]=(d[mid]<<1) -(s[mid]<<1);
        s[mid]=s[mid]+((d[mid-1]+d[mid])>>3);
	#if (debugForwardWavelet2 == 1) 
	debugPrintf(debugForwardWavelet2, "d[mid] = %d\n", d[mid]);
	debugPrintf(debugForwardWavelet2, "s[mid] = %d\n", s[mid]);
	#endif
        for(ii=0; ii<=mid; ii++)
        { /* k-s: no */ /* not working correctly individually, don't select it */
		#if (debugForwardWavelet2 == 1) 	    
		debugPrintf(debugForwardWavelet2, "data[%d]=x[%d]=s[%d]=%d\n",i+ii*COL,ii*COL,ii,s[ii]);
	    debugPrintf(debugForwardWavelet2, "data[%d]=x[%d]=d[%d]=%d\n",(ii+mid+1)*COL+i, (ii+mid+1)*COL,ii,d[ii]);
		#endif
            x[ii*COL]=s[ii];
            x[(ii+mid+1)*COL]=d[ii];
        }
    }
}
