/**
 * _______________________________________________________________________________
 *
 *  Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
 *  Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
 *  (http://www.rs.tu-darmstadt.de) All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *  3. All advertising materials mentioning features or use of this software
 *     must display the following acknowledgement: "This product includes
 *     software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
 *     its contributors."
 *
 *  4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
 *     names of its contributors may be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
 *  CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
 *  BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 *  TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * _______________________________________________________________________________
 */


`timescale      1ns / 10ps
`default_nettype none

module stall_control(
	clk_peri,
	do_peri,
	di_peri,
	addr_peri,
	access_peri,
	wr_peri,
	reset,
	stallcontrol,
	accessstats
	);

parameter BASE_ADR	= 10'h0;
parameter PORT_WIDTH	= 18;

parameter ALL_PORTS	= 7;
parameter DECODSIM_DEB	= "NO";

// inputs & outputs des Systembus
input	wire				clk_peri;
input	wire	[17:0]	do_peri;	// Datenbus 
output	wire	[17:0]			di_peri;	// Datenbus
input	wire	[9:0]			addr_peri;	// Adressbus
input	wire				access_peri;	// chipselect Peripherie
input	wire				wr_peri;	// Peripherie-Schreibzugriff
input	wire				reset;
// Signale des Ports
output	wire	[64:0] stallcontrol;
input wire [35:0] accessstats;


wire					select;
wire		[ALL_PORTS-1:0]		reg_read;
wire		[ALL_PORTS-1:0]		reg_write;

// Addressdekoder fuer alle Geraeteregister
reg_access_decoder	#(
		.REG_COUNT(	ALL_PORTS	),
		.BASE_ADR(	BASE_ADR	),
		.SIM_DEBUG(	DECODSIM_DEB	)
			)
decod1			(
		.reset(		reset		),
		.clk_peri(	clk_peri	),
		.addr_peri(	addr_peri	),
		.access_peri(	access_peri	),
		.wr_peri(	wr_peri		),

		.peri_select(	select		),
		.reg_read(	reg_read	),
		.reg_write(	reg_write	),
		.reg_select()
			);

//0: enables
//1: stall data addr
//2: stall code addr
//3: offset select
//4: data count (clear all on write)
//5: code count (clear all on write)
//6: stall durations lower 9 data, upper 9 code


reg [3:0]stall_data_en,stall_code_en;
reg [17:0] stall_data_addr;
reg [17:0] stall_code_addr;

reg [8:0] stall_data_duration;
reg [8:0] stall_code_duration;

wire reset_counts;

reg [1:0] select_stats;

assign reset_counts = reg_write[4] | reg_write[5];


always @ (posedge clk_peri)
	if(reset) begin
		stall_code_en <= 0;
		stall_code_addr <= 0;

		stall_data_en <= 0;
		stall_data_addr<= 0;
		
		stall_code_duration <= 0;
		stall_data_duration <= 0;
	end else begin	
		if (reg_write[0]) begin
			stall_data_en <= do_peri[3:0];
			stall_code_en <= do_peri[7:4];
		end
		if (reg_write[1]) begin
			stall_data_addr <= do_peri>>1;
		end
		if (reg_write[2]) begin
			stall_code_addr <= do_peri;
		end
		if (reg_write[3]) begin
			select_stats <= do_peri [1:0];
		end
		if (reg_write[6]) begin
			stall_data_duration <= do_peri[8:0];
			stall_code_duration <= do_peri[17:9];
		end
end

assign stallcontrol[17:0] = stall_data_addr;
assign stallcontrol[35:18] = stall_code_addr;
assign stallcontrol[39:36] = stall_data_en;
assign stallcontrol[43:40] = stall_code_en;
assign stallcontrol[45:44] = select_stats;
assign stallcontrol[46] = reset_counts;
assign stallcontrol[55:47] = stall_data_duration;
assign stallcontrol[64:56] = stall_code_duration;

assign	di_peri		=  reg_read[0]?{13'b0, stall_code_en, stall_data_en}:
			   reg_read[1]?stall_data_addr:
			   reg_read[2]?stall_code_addr:
			   reg_read[3]?{16'b0, select_stats}:
			   reg_read[4]?accessstats[17:0]:
			   reg_read[5]?accessstats[35:18]: 
			   reg_read[6]?{stall_code_duration, stall_data_duration}:
			   18'b0;


endmodule
