`default_nettype none
module stalling_memory #(
	parameter	BASE_ADDR	=	0,
	parameter	RAMBLOCKS	=	4,
    parameter	SHOW_MEM	=	1,
    parameter	SHOW_PERI	=	1
) (
	input wire							clk,
	input wire							reset,

	input wire							code_access,
	input wire							code_jmp,
	input wire		[17:0]				code_addr,
	output wire		[17:0]				code_do,
	output wire							code_busy,

	input wire		[9:0]				mem_addr_block,
	input wire		[7:0]				mem_addr_high,
	input wire		[17:0]				mem_di,
	input wire							mem_wr_low,
	input wire							mem_wr_high,
	input wire							mem_access,
	output wire		[17:0]				mem_do,
	output wire							mem_busy,

	input wire [64:0] stallcontrol,
	output wire [35:0] accessstats
);



wire[17:0] mem_do_orig;
wire[17:0] code_do_orig;

spartanmc_mem_local  #(
	.BASE_ADDR(BASE_ADDR),
	.RAMBLOCKS(RAMBLOCKS),
	.SHOW_MEM(SHOW_MEM),
	.SHOW_PERI(SHOW_PERI)
) mem (
	.clk(clk),
	.reset(reset),
	.code_access(code_access),
	.code_jmp(code_jmp),
	.code_addr(code_addr),
	.code_do(code_do_orig),
	.code_busy(),
	.mem_addr_block(mem_addr_block),
	.mem_addr_high(mem_addr_high),
	.mem_di(mem_di),
	.mem_wr_low(mem_wr_low),
	.mem_wr_high(mem_wr_high),
	.mem_access(mem_access),
	.mem_do(mem_do_orig),
	.mem_busy()
);



wire[17:0] stall_data_addr;
wire[17:0] stall_code_addr;
wire[3:0] stall_data_en,stall_code_en;

wire[8:0] stall_data_duration, stall_code_duration;


wire [1:0] select_stats;
wire reset_counts;


assign stall_data_addr = stallcontrol[17:0];
assign stall_code_addr = stallcontrol[35:18];
assign stall_data_en = stallcontrol[39:36];
assign stall_code_en = stallcontrol[43:40];

assign select_stats = stallcontrol[45:44];
assign reset_counts = stallcontrol[46];

assign stall_data_duration = stallcontrol[55:47];
assign stall_code_duration = stallcontrol[64:56];


wire [17:0] access_count_code, access_count_data;

stalllogic stalllogic_code(
	.clk(clk),
	.reset(reset),
	.current_address(code_addr),
	.stall_address(stall_code_addr),
	.access(code_access),
	.reset_counts(reset_counts),
	.stall_en(stall_code_en),
	.stall_duration(stall_code_duration),
	.do_orig(code_do_orig),
	.do_stalled(code_do),
	.accesscount_out(access_count_code),
	.accesscount_select(select_stats),
	.busy(code_busy)
);

stalllogic stalllogic_data(
	.clk(clk),
	.reset(reset),
	.current_address({mem_addr_high, mem_addr_block}),
	.stall_address(stall_data_addr),
	.access(mem_access),
	.reset_counts(reset_counts),
	.stall_en(stall_data_en),
	.stall_duration(stall_data_duration),
	.do_orig(mem_do_orig),
	.do_stalled(mem_do),
	.accesscount_out(access_count_data),
	.accesscount_select(select_stats),
	.busy(mem_busy)
);


assign accessstats = { access_count_code, access_count_data};

endmodule

module stalllogic (
	input wire clk,
	input wire reset,
	input wire [17:0] current_address,
	input wire [17:0] stall_address,
	input wire access,
	input wire reset_counts,
	input wire [8:0] stall_duration,
	input wire [17:0] do_orig,
	input wire [1:0] accesscount_select,
	input wire [3:0] stall_en,
	output wire [17:0] do_stalled,
	output wire [17:0] accesscount_out,
	output reg busy
);




reg[17:0] do_fromstall, do_buffered;
reg busy_last;

assign do_stalled = busy ? 18'hXX : busy_last ? do_buffered : do_orig;


reg [17:0] accesscount [0:3];
assign accesscount_out = accesscount[accesscount_select];

always @(posedge clk) begin
	busy_last <= busy;
	if (!busy_last)
		do_buffered <= do_orig;
	if (reset | reset_counts) begin: resetcounts
		integer i;
		for (i=0; i<4; i=i+1) begin
			accesscount[i] <= 0;
		end

	end else begin
		if (access && (current_address[17:2] == stall_address[17:2]) && ~busy) begin
			accesscount[current_address[1:0]] <= accesscount[current_address[1:0]] + 1;
		end
	end
end

reg [9:0] current_stall_duration;

always @(posedge clk) begin
	if (reset) begin
		busy <= 0;
		current_stall_duration <= 0;
		do_fromstall <= 0;
	end else begin
		if (busy) begin
			busy <= current_stall_duration != 0;
			current_stall_duration <= current_stall_duration - 1;
		end else if (access && stall_en[current_address[1:0]] && (current_address[17:2] == stall_address[17:2])) begin
			busy <= 1;
			current_stall_duration <= stall_duration-1;
		end else
			do_fromstall <= do_orig;

	end

end



endmodule
