/*
   _______________________________________________________________________________

   Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
   Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
   (http://www.rs.tu-darmstadt.de) All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement: "This product includes
   software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
   its contributors."

   4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
   names of its contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
   CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   _______________________________________________________________________________
   *//*****************************************************************************/
/** \file main.c
 ** Main component of the µStream Application.
 **
 ** History:
 **   - 2015-05-04 Version 1.0 Jan Weber: Base Version.
 **
 *****************************************************************************/
/*****************************************************************************/
/* Include files                                                             */
/*****************************************************************************/

#include "base_types.h"
#include <peripherals.h>

/*****************************************************************************/
/* Local pre-processor symbols/macros ('#define')                            */
/*****************************************************************************/

/*****************************************************************************/
/* Global variable definitions (declared in header file with 'extern')       */
/*****************************************************************************/

#define CORE_ID 1
#define TASK_ID 1
#define IS_BASETASK false

/*****************************************************************************/
/* Local type definitions ('typedef')                                        */
/*****************************************************************************/

/*****************************************************************************/
/* Local variable definitions ('static')                                     */
/*****************************************************************************/


/*****************************************************************************/
/* Local variables without initialization definitions ('static')             */
/*****************************************************************************/


/*****************************************************************************/
/* Local function prototypes ('static')                                      */
/*****************************************************************************/

void calcAutocorrelation(int samplesForAdaptivePass, long *correlationSum,
			 int *correlationValue, long *abs, long *maxAbs);
void extractEqu(int *correlationValue, long *A, int *k, long *maxAbs);
void solveEquGauss(int r, int k, long abs, long maxAbs, int *correlationValue,
		   long *A);
long backSubstitution(long *coefficients, int k, long abs,
		      int *correlationValue, long *A);
int writeCoefficients(signed char *outBytesAdaptive, long *coefficients,
		      int k);
signed int signExtend8toInt(signed char x);

/*****************************************************************************/
/* Function implementation - global ('extern') and local ('static')          */
/*****************************************************************************/

void calcAutocorrelation(int samplesForAdaptivePass, long *correlationSum,
			 int *correlationValue, long *abs, long *maxAbs) {
    {
	int i = 0;
	for (; i <= 9; i++) {
	    correlationValue[i] =
		((int) (correlationSum[i] / (samplesForAdaptivePass - i)));
	    (*abs) =
		((correlationValue[i] <
		  0) ? (-correlationValue[i]) : correlationValue[i]);
	    if (((*abs) > (*maxAbs))) {
		(*maxAbs) = (*abs);
	    }
	}
    }
}

void extractEqu(int *correlationValue, long *A, int *k, long *maxAbs) {
    /* reduce values to 16bit (to avoid overflows in integer arithmetic) */
    while ((*maxAbs) >= (1 << 15)) {
	(*maxAbs) /= 2;
	{
	    int i = 0;
	    for (; i <= 9; i++) {
		correlationValue[i] >>= 1;
	    }
	}
    }
    /* derive linear equation system Ax=B from autocorrelation values, B[i]=correlationValue[i+1] */
    {
	int i = 0;
	for (; i < 9; i++) {
	    for ((*k) = 0; (*k) < 9; (*k)++) {
		int n = (i * 9) + (*k);
		int c =
		    correlationValue[((i > (*k)) ? (i - (*k)) : ((*k) - i))];
		A[n] = c;
	    }
	}
    }
}

void solveEquGauss(int r, int k, long abs, long maxAbs, int *correlationValue,
		   long *A) {
    {
	int i = 0;
	for (; i < (9 - 1); i++) {
	    maxAbs = 0;
	    for (r = (i + 1); r < 9; r++) {
		for (k = (i + 1); k < 9; k++) {
		    long x = A[((r * 9) + i)];
		    long y = A[((i * 9) + k)];
		    long z = A[((i * 9) + i)];
		    /* long mul = x y; */
		    long mul = x * y;
		    long delta = mul / z;
		    A[((r * 9) + k)] -= delta;
		    abs = A[((r * 9) + k)];
		    if ((abs < 0)) {
			abs = (-abs);
		    }
		    if ((abs > maxAbs)) {
			maxAbs = abs;
		    }
		}
		correlationValue[(r + 1)] -=
		    ((A[((r * 9) + i)] * correlationValue[(i + 1)]) /
		     A[((i * 9) + i)]);
		abs = correlationValue[(r + 1)];
		if ((abs < 0)) {
		    abs = (-abs);
		}
		if ((abs > maxAbs)) {
		    maxAbs = abs;
		}
		A[((r * 9) + i)] = 0;
		/* reduce values to 16bit */
		while (maxAbs >= (1 << 15)) {
		    maxAbs /= 2;
		    correlationValue[(r + 1)] >>= 1;
		    for (k = (i + 1); k <= 9; k++) {
			A[((r * 9) + k)] >>= 1;
		    }
		}
	    }
	}
    }
}

long backSubstitution(long *coefficients, int k, long abs,
		      int *correlationValue, long *A) {
    {
	int i = 9;
	for (; (i--) != 0;) {
	    coefficients[i] = correlationValue[(i + 1)];
	    for (k = (9 - 1); k > i; k--) {
		long delta = A[((i * 9) + k)] * coefficients[k];
		long shiftedDelta = delta >> 12;
		coefficients[i] -= shiftedDelta;
	    }
	    long shifted = coefficients[i] << 12;
	    coefficients[i] = (shifted / A[((i * 9) + i)]);
	    /* fallback to simple dpcm if coefficients do not fit into fixedpoint format bits */
	    abs = coefficients[i];
	    if ((abs < 0)) {
		abs = (-abs);
	    }
	    if ((abs >= (1 << ((4 + 12) - 1)))) {
		coefficients[0] = (1 << 12);
		for (k = 1; k < 9; k++) {
		    coefficients[k] = 0;
		}
		break;
	    }
	}
    }
    return abs;
}

int writeCoefficients(signed char *outBytesAdaptive, long *coefficients,
		      int k) {
    int byteCounterForAdaptivePass = 0;
    int bufferForAdaptive = 0;
    int bitsForAdaptive = 0;
    for (k = 0; k < 9; k++) {
	int i = 4 + 12;
	while (i != 0) {
	    if (((i + bitsForAdaptive) >= 8)) {
		bitsForAdaptive = (8 - bitsForAdaptive);
		i = (i - bitsForAdaptive);
		int shiftedA = bufferForAdaptive << bitsForAdaptive;
		signed char shiftedB = (1 << bitsForAdaptive) - 1;
		long shiftedC = coefficients[k] >> i;
		signed char anded = shiftedB & shiftedC;
		signed int andedSigex = signExtend8toInt(anded);
		outBytesAdaptive[(byteCounterForAdaptivePass++)] =
		    ((signed char) (shiftedA | andedSigex));
		bitsForAdaptive = 0;
	    }
	    else {
		bitsForAdaptive += i;
		bufferForAdaptive =
		    ((bufferForAdaptive << i) |
		     (((1 << i) - 1) & coefficients[k]));
		i = 0;
	    }
	}
    }
    return k;
}

signed int signExtend8toInt(signed char x) {
    signed int r;
    /* Sign extend from 8 bits */
  __asm__("sigex %0, 8": "=r"(r):"0"(x));
    return r;
}




/*****************************************************************************/
/* Global variables definitions and initialisation                           */
/*****************************************************************************/


/*****************************************************************************/
/* main                                                                      */
/*****************************************************************************/

void main() {


    while (TRUE) {
	//local variables
	long A[(9 * 9)];
	long abs;
	long coefficients[9];
	int correlationValue[(9 + 1)];
	int k;
	signed char outBytesAdaptive[((9 * (4 + 12)) / 8)]
	    __attribute__ ((aligned));
	int r;
	//local variables originally having initializers received from interconnect peripherals
	long correlationSum[(9 + 1)];
	int inByte;
	char lastBlock __attribute__ ((aligned));
	long maxAbs;
	int outByte;
	int samplesForAdaptivePass;
	// receives the variable correlationSum from another core.
	core_connector_slave_receive_data(&CORE_CONNECTOR_SLAVE_0,
					  &correlationSum,
					  (sizeof correlationSum) / 2 +
					  (sizeof correlationSum) % 2);
	// receives the variable inByte from another core.
	core_connector_slave_receive_data(&CORE_CONNECTOR_SLAVE_0, &inByte,
					  (sizeof inByte) / 2 +
					  (sizeof inByte) % 2);
	// receives the variable lastBlock from another core.
	core_connector_slave_receive_data(&CORE_CONNECTOR_SLAVE_0, &lastBlock,
					  (sizeof lastBlock) / 2 +
					  (sizeof lastBlock) % 2);
	// receives the variable maxAbs from another core.
	core_connector_slave_receive_data(&CORE_CONNECTOR_SLAVE_0, &maxAbs,
					  (sizeof maxAbs) / 2 +
					  (sizeof maxAbs) % 2);
	// receives the variable outByte from another core.
	core_connector_slave_receive_data(&CORE_CONNECTOR_SLAVE_0, &outByte,
					  (sizeof outByte) / 2 +
					  (sizeof outByte) % 2);
	// receives the variable samplesForAdaptivePass from another core.
	core_connector_slave_receive_data(&CORE_CONNECTOR_SLAVE_0,
					  &samplesForAdaptivePass,
					  (sizeof samplesForAdaptivePass) /
					  2 +
					  (sizeof samplesForAdaptivePass) %
					  2);

	/* calculate autocorrelation values */
	calcAutocorrelation(samplesForAdaptivePass, correlationSum,
			    correlationValue, (&abs), (&maxAbs));
	extractEqu(correlationValue, A, (&k), (&maxAbs));
	/* solve linear equation system by gaussian elimination (without pivot search) to calculate coefficients */
	solveEquGauss(r, k, abs, maxAbs, correlationValue, A);
	/* backsubstitution */
	backSubstitution(coefficients, k, abs, correlationValue, A);
	/* write coefficients to output stream */
	writeCoefficients(outBytesAdaptive, coefficients, k);
	/* predict sample and encode prediction error (except for the first unencoded samples) */

	// sends the variable coefficients to another core.
	core_connector_master_send_data(&CORE_CONNECTOR_MASTER_0,
					&coefficients,
					(sizeof coefficients) / 2 +
					(sizeof coefficients) % 2);
	// sends the variable inByte to another core.
	core_connector_master_send_data(&CORE_CONNECTOR_MASTER_0, &inByte,
					(sizeof inByte) / 2 +
					(sizeof inByte) % 2);
	// sends the variable lastBlock to another core.
	core_connector_master_send_data(&CORE_CONNECTOR_MASTER_0, &lastBlock,
					(sizeof lastBlock) / 2 +
					(sizeof lastBlock) % 2);
	// sends the variable outByte to another core.
	core_connector_master_send_data(&CORE_CONNECTOR_MASTER_0, &outByte,
					(sizeof outByte) / 2 +
					(sizeof outByte) % 2);
	// sends the variable outBytesAdaptive to another core.
	core_connector_master_send_data(&CORE_CONNECTOR_MASTER_0,
					&outBytesAdaptive,
					(sizeof outBytesAdaptive) / 2 +
					(sizeof outBytesAdaptive) % 2);
    }

}
