/*
   _______________________________________________________________________________

   Copyright (c) 2012 TU Dresden, Chair for Embedded Systems
   Copyright (c) 2013-2016 TU Darmstadt, Computer Systems Group
   (http://www.rs.tu-darmstadt.de) All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

   1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.

   3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement: "This product includes
   software developed by the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group and
   its contributors."

   4. Neither the name of the TU Dresden Chair for Embedded Systems, TU Darmstadt Computer Systems Group nor the
   names of its contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.

   THIS SOFTWARE IS PROVIDED BY TU DRESDEN CHAIR FOR EMBEDDED SYSTEMS, TU DARMSTADT COMPUTER SYSTEMS GROUP AND
   CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
   BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
   REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
   TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   _______________________________________________________________________________
   *//*****************************************************************************/
/** \file timercontrol.c
 ** Contains functions to control the evaluation timer.
 **
 ** History:
 **   - 2015-09-01 Version 1.0 Jan Weber: Base Version.
 **   - 2016-12-20 Version 2.0 Laurenz Kamp: Updated to use performance counter.
 **
 *****************************************************************************/
/*****************************************************************************/
/* Include files                                                             */
/*****************************************************************************/
#include "base_types.h"
#include "evaluation.h"
#include <subsystems/spartanmc_1/peripherals.h>
#include <stdio.h>
#include <perf.h>

/*****************************************************************************/
/* Local pre-processor symbols/macros ('#define')                            */
/*****************************************************************************/
#define CORE_COUNT 4
#define RESULT_WORDS_PER_CORE 11
#define RESULT_TOTAL_WORDS ((CORE_COUNT-1)*RESULT_WORDS_PER_CORE)
#define RESULT_LENGTH 36
#define CORE_ID 1
#define TASK_ID 1
#define IS_BASETASK 0
#define RESULT_STRING "%02d,%010d,%010d,%010d;\n"
/*****************************************************************************/
/* Global variable definitions (declared in header file with 'extern')       */
/*****************************************************************************/

/*****************************************************************************/
/* Local type definitions ('typedef')                                        */
/*****************************************************************************/

/*****************************************************************************/
/* Local variable definitions ('static')                                     */
/*****************************************************************************/

static uint36_t au9EvalResults[5] = { 0u, 0u, 0u, 0u, 0u };

static en_evalTarget_t enEvalTarget = NONE;
static struct evalResults *resultAddress;

/*****************************************************************************/
/* Local function prototypes ('static')                                      */
/*****************************************************************************/

/*****************************************************************************/
/* Function implementation - global ('extern') and local ('static')          */
/*****************************************************************************/

/**
 ******************************************************************************
 ** \brief Initializes the evaluation module within Child Tasks.
 **
 ** \param  [in]     pstcUart              pointer to the hardware address of
 **                                        the uart that should be used for
 **                                        collecting evaluation results.
 ** \param  [in/out] pstcPortIn            pointer to the hardware address of
 **                                        the port in that is used to receive
 **                                        the exit signal from the BaseTask.
 **
 ******************************************************************************/
void evaluation_initChild(void *resAddrBase) {
    struct evalResults *ptr = (struct evalResults *) resAddrBase;
    ptr -= (CORE_ID + 1);
    resultAddress = ptr;

    int i;
    resultAddress->valid = 0;
    for (i = 0; i < 5; i++) {
	au9EvalResults[i] = 0;
    }

    microstreams_eval_init();
}

/**
 ******************************************************************************
 ** \brief Starts the measurement of the given target.
 **
 ** \param  [in]     enTarget              measurement target.
 **
 ******************************************************************************/
void evaluation_startMeasurement(en_evalTarget_t enTarget) {
    enEvalTarget = enTarget;
    microstreams_eval_restart();
}

/**
 ******************************************************************************
 ** \brief Stops the running measurement and accumulates for the final result.
 **
 ******************************************************************************/
void evaluation_stopMeasurement() {
    uint36_t u9counterResult;

    u9counterResult = microstreams_eval_stop();
    au9EvalResults[enEvalTarget] += u9counterResult;
    enEvalTarget = NONE;
}

/**
 ******************************************************************************
 ** \brief Sends the results of the evaluation to the core hosting the base task.
 ** 	   The core hosting the base task receives the evaluation results with
 ** 	   this procedure and sends them to a PC terminal.
 **
 ******************************************************************************/
void evaluation_sendResults() {


    /* Store results to global memory. */
    int i;
    for (i = 0; i < 5; i++) {
	resultAddress->results[i] += au9EvalResults[i];
    }
    resultAddress->valid = 1;

}
